;;;-*- Mode: Common-LISP; Package: FILE-SYSTEM; Base: 10. -*- 

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985-1989 Texas Instruments Incorporated. All rights reserved.
;;; Directory Hacks program by HANSON -- feel free to use it.


;;; 05-11-88 DAB Fixed balance-directory to convert file types to canonical types.

(PROCLAIM '(SPECIAL *DEFAULT-PATHNAME-DEFAULTS*)) 
;;;; Pathname Operations



(DEFSUBST UNSPECIFIED-PATHNAME-COMPONENT? (COMPONENT)
  (MEMBER COMPONENT '(NIL :UNSPECIFIC) :TEST #'EQ))  



(DEFSUBST WILD-PATHNAME-COMPONENT? (COMPONENT)
  (MEMBER COMPONENT '(NIL :WILD) :TEST #'EQ))  



(DEFUN DEFAULT-EMPTY-PATHNAME-COMPONENTS (PATH1 PATH2)
  "Fill in the unspecified parts of PATH1 with the corresponding parts of PATH2."
  (FAST-NEW-PATHNAME PATH1
		     (AND (UNSPECIFIED-PATHNAME-COMPONENT? (PATHNAME-DEVICE PATH1))
			(PATHNAME-DEVICE PATH2))
		     (AND (UNSPECIFIED-PATHNAME-COMPONENT? (PATHNAME-DIRECTORY PATH1))
			(PATHNAME-DIRECTORY PATH2))
		     (AND (UNSPECIFIED-PATHNAME-COMPONENT? (PATHNAME-NAME PATH1))
			(PATHNAME-NAME PATH2))
		     (AND (UNSPECIFIED-PATHNAME-COMPONENT? (PATHNAME-TYPE PATH1))
			(PATHNAME-TYPE PATH2))
		     (AND (UNSPECIFIED-PATHNAME-COMPONENT? (PATHNAME-VERSION PATH1))
			(PATHNAME-VERSION PATH2))))  



(DEFUN DEFAULT-WILD-PATHNAME-COMPONENTS (PATH1 PATH2)
  "Fill in the unspecified or wild parts of PATH1 with the corresponding parts of PATH2."
  (FAST-NEW-PATHNAME PATH1
		     (AND (WILD-PATHNAME-COMPONENT? (PATHNAME-DEVICE PATH1))
			(PATHNAME-DEVICE PATH2))
		     (AND (WILD-PATHNAME-COMPONENT? (PATHNAME-DIRECTORY PATH1))
			(PATHNAME-DIRECTORY PATH2))
		     (AND (WILD-PATHNAME-COMPONENT? (PATHNAME-NAME PATH1)) (PATHNAME-NAME PATH2))
		     (AND (WILD-PATHNAME-COMPONENT? (PATHNAME-TYPE PATH1)) (PATHNAME-TYPE PATH2))
		     (AND (WILD-PATHNAME-COMPONENT? (PATHNAME-VERSION PATH1))
			(PATHNAME-VERSION PATH2))))  



(DEFSUBST MATCHING-PATHNAME-COMPONENTS? (ACCESS-MSG PATHNAME1 PATHNAME2 CONVERTED-PATHNAME1 CONVERTED-PATHNAME2)
  "Two components match if one maps into the other."
  (OR (EQUALP (SEND PATHNAME1 ACCESS-MSG) (SEND CONVERTED-PATHNAME2 ACCESS-MSG))
     (EQUALP (SEND PATHNAME2 ACCESS-MSG) (SEND CONVERTED-PATHNAME1 ACCESS-MSG))))  

(DEFSUBST MATCHING-PATHNAME-canonical-COMPONENTS? (ACCESS-MSG PATHNAME1 PATHNAME2)  ;05-11-88 DAB
  "Two components match if one maps into the other."
   (EQUALP (SEND PATHNAME1 ACCESS-MSG) (SEND PATHNAME2 ACCESS-MSG)))
      

(DEFUN MATCHING-NAME-AND-TYPE? (PATHNAME1 PATHNAME2 &OPTIONAL CANONICAL-TYPES-P)
  (AND
   (LET ((NAME1 (PATHNAME-NAME PATHNAME1))
	 (NAME2 (PATHNAME-NAME PATHNAME2)))
     (OR (EQUALP NAME1 NAME2)
	(AND (STRINGP NAME1) (CONSP NAME2) (NULL (CDR NAME2)) (EQUALP NAME1 (CAR NAME2)))
	(AND (STRINGP NAME2) (CONSP NAME1) (NULL (CDR NAME1)) (EQUALP NAME2 (CAR NAME1)))))
   (LET ((HOST1 (PATHNAME-HOST PATHNAME1))
	 (HOST2 (PATHNAME-HOST PATHNAME2))
	 (CONVERTED-PATHNAME1 PATHNAME1)
	 (CONVERTED-PATHNAME2 PATHNAME2))
     (UNLESS (EQ HOST1 HOST2)
       (SETQ CONVERTED-PATHNAME1 (SEND PATHNAME1 :NEW-PATHNAME :HOST HOST2)
	     CONVERTED-PATHNAME2 (SEND PATHNAME2 :NEW-PATHNAME :HOST HOST1)))
     (OR
      (MATCHING-PATHNAME-COMPONENTS? :TYPE PATHNAME1 PATHNAME2 CONVERTED-PATHNAME1
				     CONVERTED-PATHNAME2)
      (AND CANONICAL-TYPES-P
	 (MATCHING-PATHNAME-canonical-COMPONENTS? :CANONICAL-TYPE PATHNAME1 PATHNAME2))))))   ;05-11-88 DAB 


  



(DEFSUBST MATCHING-OR-MISSING-PATHNAME-COMPONENTS? (ACCESS-MSG PATHNAME1 PATHNAME2)
  "Two components match if one maps into the other or if either is NIL or :WILD."
  (OR (MEMBER (SEND PATHNAME1 ACCESS-MSG) '(NIL :WILD) :TEST #'EQ)
     (MEMBER (SEND PATHNAME2 ACCESS-MSG) '(NIL :WILD) :TEST #'EQ)
     (EQUALP (SEND PATHNAME1 ACCESS-MSG) (SEND PATHNAME2 ACCESS-MSG))))  

;Used only to compare against the :IGNORE filenames.

(DEFUN MATCHING-PATHNAMES? (PATHNAME1 PATHNAME2)
  (AND (MATCHING-OR-MISSING-PATHNAME-COMPONENTS? :DEVICE PATHNAME1 PATHNAME2)
     (MATCHING-OR-MISSING-PATHNAME-COMPONENTS? :DIRECTORY PATHNAME1 PATHNAME2)
     (MATCHING-OR-MISSING-PATHNAME-COMPONENTS? :NAME PATHNAME1 PATHNAME2)
     (MATCHING-OR-MISSING-PATHNAME-COMPONENTS? :TYPE PATHNAME1 PATHNAME2)
     (MATCHING-OR-MISSING-PATHNAME-COMPONENTS? :VERSION PATHNAME1 PATHNAME2)))  



(DEFUN MATCHING-PATHNAME-COMPONENTS-WITH-WILD? (ACCESS-MSG PATHNAME1 PATHNAME2 CONVERTED-PATHNAME1 CONVERTED-PATHNAME2)
  "Two components match if one maps into the other or one is WILD."
  (LET ((P1-COMP (SEND PATHNAME1 ACCESS-MSG))
	(P2-C-COMP (SEND CONVERTED-PATHNAME2 ACCESS-MSG))
	(P2-COMP (SEND PATHNAME2 ACCESS-MSG))
	(P1-C-COMP (SEND CONVERTED-PATHNAME1 ACCESS-MSG)))
    (OR (EQ P1-COMP :WILD) (EQ P2-C-COMP :WILD) (EQUAL P1-COMP P2-C-COMP) (EQ P2-COMP :WILD)
       (EQ P1-C-COMP :WILD) (EQUAL P2-COMP P1-C-COMP))))  



(DEFUN MATCHING-PATHNAMES-WITH-WILD? (PATHNAME1 PATHNAME2)
  (LET ((CONVERTED-PATHNAME1 (SEND PATHNAME1 :NEW-PATHNAME :HOST (PATHNAME-HOST PATHNAME2)))
	(CONVERTED-PATHNAME2 (SEND PATHNAME2 :NEW-PATHNAME :HOST (PATHNAME-HOST PATHNAME1))))
    (AND
     (MATCHING-PATHNAME-COMPONENTS-WITH-WILD? :DEVICE PATHNAME1 PATHNAME2 CONVERTED-PATHNAME1
					      CONVERTED-PATHNAME2)
     (MATCHING-PATHNAME-COMPONENTS-WITH-WILD? :DIRECTORY PATHNAME1 PATHNAME2 CONVERTED-PATHNAME1
					      CONVERTED-PATHNAME2)
     (MATCHING-PATHNAME-COMPONENTS-WITH-WILD? :NAME PATHNAME1 PATHNAME2 CONVERTED-PATHNAME1
					      CONVERTED-PATHNAME2)
     (MATCHING-PATHNAME-COMPONENTS-WITH-WILD? :TYPE PATHNAME1 PATHNAME2 CONVERTED-PATHNAME1
					      CONVERTED-PATHNAME2)
     (MATCHING-PATHNAME-COMPONENTS-WITH-WILD? :VERSION PATHNAME1 PATHNAME2 CONVERTED-PATHNAME1
					      CONVERTED-PATHNAME2))))  



;;;; CFile Abstraction
;;; Covers up the structure of the object returned by DIRECTORY-LIST.


(DEFSUBST CFILE-PLIST (CFILE)
  CFILE)  


(DEFSUBST CFILE-TRUENAME (CFILE)
  (CAR CFILE))  


(DEFSUBST CFILE-HOST (CFILE)
  (PATHNAME-HOST (CFILE-TRUENAME CFILE)))  


(DEFSUBST CFILE-DEVICE (CFILE)
  (PATHNAME-DEVICE (CFILE-TRUENAME CFILE)))  


(DEFSUBST CFILE-DIRECTORY (CFILE)
  (PATHNAME-DIRECTORY (CFILE-TRUENAME CFILE)))  


(DEFSUBST CFILE-NAME (CFILE)
  (PATHNAME-NAME (CFILE-TRUENAME CFILE)))  


(DEFSUBST CFILE-TYPE (CFILE)
  (PATHNAME-TYPE (CFILE-TRUENAME CFILE)))  


(DEFSUBST CFILE-CANONICAL-TYPE (CFILE)
  (SEND (CFILE-TRUENAME CFILE) :CANONICAL-TYPE))  


(DEFSUBST CFILE-VERSION (CFILE)
  (PATHNAME-VERSION (CFILE-TRUENAME CFILE)))  



(DEFSUBST CFILE-CREATION-DATE (CFILE)
  (GET (CFILE-PLIST CFILE) :CREATION-DATE))  



(DEFSUBST CFILE-AUTHOR (CFILE)
  (GET (CFILE-PLIST CFILE) :AUTHOR))  



(DEFSUBST CFILE-TRANSFER-DESTINATIONS (CFILE)
  (GET (CFILE-PLIST CFILE) 'TRANSFER-DESTINATIONS))  



(DEFSUBST CFILE-TRANSFER-MODE (CFILE)
  (GET (CFILE-PLIST CFILE) 'TRANSFER-MODE))  



(DEFSUBST CFILE-CLASS-REPRESENTATIVE (CFILE)
  (GET (CFILE-PLIST CFILE) 'CLASS-REPRESENTATIVE))  



(DEFSUBST CFILE-REPRESENTED-CLASS (REPRESENTATIVE)
  (GET (CFILE-PLIST REPRESENTATIVE) 'REPRESENTATIVE-CLASS))  



(DEFSUBST CFILE-BYTE-SIZE (CFILE)
  (GET (CFILE-PLIST CFILE) :BYTE-SIZE))  



(DEFSUBST CFILE-IS-SUBDIRECTORY? (CFILE)
  (GET (CFILE-PLIST CFILE) :DIRECTORY))  



(DEFSUBST SAME-NAME-AND-TYPE? (CFILE1 CFILE2)
  (MATCHING-NAME-AND-TYPE? (CFILE-TRUENAME CFILE1) (CFILE-TRUENAME CFILE2)))  



(DEFSUBST SAME-NAME-AND-CANONICAL-TYPE? (CFILE1 CFILE2)
  (MATCHING-NAME-AND-TYPE? (CFILE-TRUENAME CFILE1) (CFILE-TRUENAME CFILE2) T))  



(DEFSUBST SAME-VERSION? (CFILE1 CFILE2)
  (EQUAL (CFILE-VERSION CFILE1) (CFILE-VERSION CFILE2)))  



(DEFSUBST SAME-CREATION-DATE? (CFILE1 CFILE2)
  (= (CFILE-CREATION-DATE CFILE1) (CFILE-CREATION-DATE CFILE2)))  



(DEFUN SAME-NAME-TYPE-AND-VERSION? (CFILE1 CFILE2)
  (AND (SAME-NAME-AND-TYPE? CFILE1 CFILE2) (SAME-VERSION? CFILE1 CFILE2)))  

;;;; CDirectory Abstraction



(DEFSTRUCT (CDIRECTORY (:TYPE :NAMED-ARRAY) (:CONC-NAME "CDIRECTORY-") (:CALLABLE-CONSTRUCTORS NIL)
  (:ALTERANT ALTER-CDIRECTORY) (:PREDICATE NIL) (:COPIER NIL))
  PATHNAME;Specification of directory.
  TRUENAME;Pathname, with wildcards inserted.
  PLIST;Directory-List information.
  SUBDIRECTORIES;Subdirectory plists.
  CFILES;Component files.
)  



(DEFUN CREATE-CDIRECTORY (PATHNAME REMOVABLE-NAMES ERROR-P)
  (LET* ((WILDCARDED-PATHNAME (MAKE-DIRECTORY-LIST-PATHNAME PATHNAME))
	 (DIRECTORY
	  (IF (NOT ERROR-P)
	    (DIRECTORY-LIST WILDCARDED-PATHNAME :SORTED :NOERROR)
	    (DIRECTORY-LIST WILDCARDED-PATHNAME :SORTED))))
    (IF (ERRORP DIRECTORY)
      DIRECTORY
      (LET ((SUBDIRS (QUOTE NIL))
	    (CFILES (QUOTE NIL)))
	(DOLIST (CFILE (CDR DIRECTORY))
	  (COND
	    ((CFILE-IS-SUBDIRECTORY? CFILE) (PUSH CFILE SUBDIRS))
	    ((NOT
	      (DOLIST (PATHNAME REMOVABLE-NAMES)
		(IF (MATCHING-PATHNAMES-WITH-WILD? (CFILE-TRUENAME CFILE) PATHNAME)
		  (RETURN T))))
	     (PUSH CFILE CFILES))))
	(LET ((CDIRECTORY
	       (MAKE-CDIRECTORY PATHNAME PATHNAME TRUENAME WILDCARDED-PATHNAME PLIST
		  (CAR DIRECTORY) SUBDIRECTORIES (NREVERSE SUBDIRS) CFILES (NREVERSE CFILES))))
	  (SETF (GET (CAR DIRECTORY) 'ASSOCIATED-CDIRECTORY) CDIRECTORY)
	  CDIRECTORY)))))  



(DEFUN CREATE-FAKE-CDIRECTORY (PATHNAME)
  (MAKE-CDIRECTORY PATHNAME PATHNAME TRUENAME PATHNAME PLIST () SUBDIRECTORIES () CFILES ()))  



(DEFUN CDIRECTORY-LIST (DIRECTORY)
  (CONS (CDIRECTORY-PLIST DIRECTORY) (CDIRECTORY-CFILES DIRECTORY)))  



(DEFUN ASSOCIATED-CDIRECTORY (DIRECTORY-PLIST)
  (GET DIRECTORY-PLIST 'ASSOCIATED-CDIRECTORY))  



(DEFUN MAKE-DIRECTORY-LIST-PATHNAME (PATHNAME)
  (FAST-NEW-PATHNAME PATHNAME
		     ;; DIRECTORY-LIST should hack the device/directory being wild.
		     (PATHNAME-DEVICE PATHNAME) (PATHNAME-DIRECTORY PATHNAME)
		     (AND (UNSPECIFIED-PATHNAME-COMPONENT? (PATHNAME-NAME PATHNAME)) :WILD)
		     (AND (UNSPECIFIED-PATHNAME-COMPONENT? (PATHNAME-TYPE PATHNAME)) :WILD)
		     (AND (UNSPECIFIED-PATHNAME-COMPONENT? (PATHNAME-VERSION PATHNAME)) :NEWEST)))  

;;;; IsoSet Abstraction
;;; A set of functions for manipulating sets and equivalence classes sorted
;;; by means of an isomorphism into their intersection/difference parts.


(DEFSTRUCT (ISOSET (:TYPE :LIST) (:CONC-NAME "ISOSET-") (:CALLABLE-CONSTRUCTORS NIL)
  (:ALTERANT ALTER-ISOSET) (:PREDICATE NIL) (:COPIER NIL))
  IN-1-ONLY;difference set1 set2
  IN-BOTH;intersection set1 set2
  IN-2-ONLY;difference set2 set1
)  



(DEFUN SORT-ISOMORPHIC-CLASSES (ISOMORPHISM CROSS-ISOMORPHISM SET1 SET2)
  "Given an isomorphism and two sets, return three values:
1.  A list of equivalence classes from SET1 that were not isomorphic
    to any element in SET2.
2.  A list of pairs of isomorphic equivalence classes; car=SET1, cdr=SET2.
3.  A list of equivalence classes from SET2 that were not isomorphic
    to any element in SET1.
ISOMORPHISM is used to make equivalence classes, while
CROSS-ISOMORPHISM is used to compare an element of SET1 with one of SET2."
  (SORT-ISOMORPHIC-ELEMENTS
   (LET-CLOSED ((ISOMORPHISM CROSS-ISOMORPHISM))
      #'(LAMBDA (CLASS1 CLASS2)
	  (FUNCALL ISOMORPHISM (CAR CLASS1) (CAR CLASS2))))
   (EQUIVALENCE-CLASSES-OF ISOMORPHISM SET1) (EQUIVALENCE-CLASSES-OF ISOMORPHISM SET2)))  


(DEFUN EQUIVALENCE-CLASSES-OF (ISOMORPHISM SET)
  "Give a set and an isomorphism, return the set of equivalence-classes."
  (LET ((EQV-CLASSES (QUOTE NIL)))
    (DOLIST (ITEM SET)
      (DO ((SUBSET EQV-CLASSES (CDR SUBSET)))
	  ((NULL SUBSET)
	   (PUSH (LIST ITEM) EQV-CLASSES))
	(COND
	  ((FUNCALL ISOMORPHISM ITEM (CAAR SUBSET)) (PUSH ITEM (CAR SUBSET)) (RETURN T)))))
    (DO ((SUBSET EQV-CLASSES (CDR SUBSET)))
	((NULL SUBSET)
	 (NREVERSE EQV-CLASSES))
      (RPLACA SUBSET (NREVERSE (CAR SUBSET))))))  



(DEFUN SORT-ISOMORPHIC-ELEMENTS (ISOMORPHISM SET1 SET2)
  "Given an isomorphism and two sets, return three values:
1.  A list of elements from SET1 that were not isomorphic
    to any element in SET2.
2.  A list of pairs of isomorphic elements; car=SET1, cdr=SET2.
3.  A list of elements from SET2 that were not isomorphic
    to any element in SET1.
It is assumed that no element in an input set is isomorphic to
any other element in the same set.  If this is possible, use
SORT-ISOMORPHIC-CLASSES instead."
  (LET ((ELEMENTS-IN-SET1-ONLY (QUOTE NIL))
	(ELEMENTS-IN-SET2-ONLY (COPY-LIST SET2))
	(ELEMENTS-IN-BOTH-SETS (QUOTE NIL)))
    (DOLIST (ELEMENT-IN-SET1 SET1)
      (LET ((ISOMORPHIC-ELEMENT
	     (DOLIST (ELEMENT-IN-SET2 ELEMENTS-IN-SET2-ONLY)
	       (IF (FUNCALL ISOMORPHISM ELEMENT-IN-SET1 ELEMENT-IN-SET2)
		 (RETURN ELEMENT-IN-SET2)))))
	(COND
	  ((NOT (NULL ISOMORPHIC-ELEMENT))
	   (PUSH (CONS ELEMENT-IN-SET1 ISOMORPHIC-ELEMENT) ELEMENTS-IN-BOTH-SETS)
	   (SETQ ELEMENTS-IN-SET2-ONLY
		 (DELETE ISOMORPHIC-ELEMENT (THE LIST ELEMENTS-IN-SET2-ONLY) :TEST #'EQ)))
	  (T (PUSH ELEMENT-IN-SET1 ELEMENTS-IN-SET1-ONLY)))))
    (MAKE-ISOSET IN-1-ONLY (NREVERSE ELEMENTS-IN-SET1-ONLY) IN-BOTH
       (NREVERSE ELEMENTS-IN-BOTH-SETS) IN-2-ONLY ELEMENTS-IN-SET2-ONLY)))  

;;;; Transfer Operators


(DEFUN SETUP-CDIRECTORY-TRANSFER-MODE (DIRECTORY &OPTIONAL COPY-MODE)
  (DOLIST (CFILE (CDIRECTORY-CFILES DIRECTORY))
    (IF (AND (NOT (NULL (CFILE-TRANSFER-DESTINATIONS CFILE))) (NULL (CFILE-TRANSFER-MODE CFILE)))
      (SETF (CFILE-TRANSFER-MODE CFILE)
	    (LET ((BYTE-SIZE (CFILE-BYTE-SIZE CFILE))
		  (INTYPE (CFILE-CANONICAL-TYPE CFILE))
		  (TRUENAME (CFILE-TRUENAME CFILE)))
	      (COND
		((MEMBER COPY-MODE '(:CHARACTERS BINARY) :TEST #'EQ) COPY-MODE)
		((EQ COPY-MODE :ASK)
		 (IF (FQUERY () "~&Is ~A a text file? " TRUENAME)
		   :CHARACTERS
		   :BINARY))
		((MEMBER BYTE-SIZE '(7. 8.) :TEST #'EQ) :CHARACTERS)
		((EQ BYTE-SIZE 16.) :BINARY)
		((MEMBER INTYPE *COPY-FILE-KNOWN-TEXT-TYPES* :TEST #'EQUAL) :CHARACTERS)
		((MEMBER INTYPE *COPY-FILE-KNOWN-BINARY-TYPES* :TEST #'EQUAL) :BINARY)
		((EQ COPY-MODE :NEVER-ASK) :CHARACTERS)
		((FQUERY '(:BEEP T) "~&Is ~A a text file? " TRUENAME) :CHARACTERS)
		(T :BINARY)))))))  



(DEFUN MARK-CFILE-FOR-TRANSFER (CFILE TARGET-DIRECTORY &OPTIONAL IGNORE)
  (LET ((CURRENT-BINDING (ASSOC TARGET-DIRECTORY (CFILE-TRANSFER-DESTINATIONS CFILE) :TEST #'EQ))
	(TARGET-PATHNAME
	 (DEFAULT-WILD-PATHNAME-COMPONENTS (CDIRECTORY-PATHNAME TARGET-DIRECTORY)
					   (CFILE-TRUENAME CFILE))))
    (IF (NULL CURRENT-BINDING)
      (PUSH (CONS TARGET-DIRECTORY TARGET-PATHNAME) (CFILE-TRANSFER-DESTINATIONS CFILE))
      (RPLACD CURRENT-BINDING TARGET-PATHNAME))))  



(DEFVAR CDIRECTORY-TRANSFER-QUERY-CHOICES
   `(,@SI:Y-OR-N-P-CHOICES ((:PROCEED "Proceed.") #\P #\)))  



(DEFUN CDIRECTORY-TRANSFER-QUERY (DIRECTORY)
  (CATCH 'CDIRECTORY-TRANSFER-QUERY-PROCEEDED
    (DOLIST (CFILE (CDIRECTORY-CFILES DIRECTORY))
      (LET ((TRUENAME (CFILE-TRUENAME CFILE)))
	(DOLIST (TARGET-PAIR (CFILE-TRANSFER-DESTINATIONS CFILE))
	  (LET ((QRESULT
		 (FQUERY `(:CHOICES ,CDIRECTORY-TRANSFER-QUERY-CHOICES) "~&Copy ~A to ~A? "
			 TRUENAME (CDR TARGET-PAIR))))
	    (COND
	      ((NULL QRESULT)
	       (SETF (CFILE-TRANSFER-DESTINATIONS CFILE)
		     (DELETE TARGET-PAIR (THE LIST (CFILE-TRANSFER-DESTINATIONS CFILE)) :TEST
			     #'EQ)))
	      ((EQ QRESULT :PROCEED) (THROW 'CDIRECTORY-TRANSFER-QUERY-PROCEEDED
					    T)))))))))  



(DEFUN CDIRECTORY-TRANSFER (DIRECTORY &OPTIONAL (ERROR-P T))
  (DOLIST (CFILE (CDIRECTORY-CFILES DIRECTORY))
    (DO ()
	(NIL)
      (LET ((DESTINATIONS (CFILE-TRANSFER-DESTINATIONS CFILE)))
	(IF (NULL (CFILE-TRANSFER-DESTINATIONS CFILE))
	  (RETURN T))
	(LET ((TARGET-PAIR (CAR DESTINATIONS)))
	  (CFILE-TRANSFER-1 CFILE (CDR TARGET-PAIR) (CAR TARGET-PAIR) ERROR-P)
	  (SETF (CFILE-TRANSFER-DESTINATIONS CFILE) (CDR DESTINATIONS)))))))  

(DEFUN CFILE-TRANSFER-1 (CFILE DESTINATION TARGET-DIRECTORY ERROR-P)
  (LET ((CHARACTERS? (EQ (CFILE-TRANSFER-MODE CFILE) :CHARACTERS))
	(new-canonicalized-pathname (send (send DESTINATION :new-type nil)   ;05-11-88 DAB
				     :NEW-PREFERRED-CANONICAL-TYPE (CFILE-TRUENAME CFILE)) ))
    (LET ((RESULT (CFILE-TRANSFER-2 CFILE new-canonicalized-pathname CHARACTERS? TARGET-DIRECTORY ERROR-P)))
      (IF (ERRORP RESULT)
	(FORMAT T "~&Copy Error: ~A" RESULT)
	(FORMAT T "~&Copied ~A to ~A (~:[binary~;text~])" (CFILE-TRUENAME CFILE) new-canonicalized-pathname 
		CHARACTERS?)))))

  



(DEFUN CFILE-TRANSFER-2 (INPUT-CFILE OUTPUT-TRUENAME CHARACTERS? TARGET-DIRECTORY ERROR-P)
  TARGET-DIRECTORY;Use this to decide which props to change.
  (WITH-OPEN-FILE (INSTREAM (CFILE-TRUENAME INPUT-CFILE) :DIRECTION :INPUT :CHARACTERS CHARACTERS? :ERROR
    ERROR-P)
    (IF (ERRORP INSTREAM)
      INSTREAM
      (WITH-OPEN-FILE (OUTSTREAM OUTPUT-TRUENAME :DIRECTION :OUTPUT :CHARACTERS CHARACTERS? :ERROR ERROR-P)
	(COND
	  ((ERRORP OUTSTREAM) OUTSTREAM)
	  (T
	   (SEND OUTSTREAM :CHANGE-PROPERTIES () :CREATION-DATE
	      (CFILE-CREATION-DATE INPUT-CFILE) :AUTHOR (CFILE-AUTHOR INPUT-CFILE))
	   (STREAM-COPY-UNTIL-EOF INSTREAM OUTSTREAM) (CLOSE OUTSTREAM)))))))  

;;;; Directory Comparison



(DEFUN COMPARE-CDIRECTORIES (CDIR1 CDIR2 &OPTIONAL TAPE-LOSSAGE)
  (LET ((ISOCLASSES
	 (SORT-ISOMORPHIC-CLASSES 'SAME-NAME-AND-TYPE? 'SAME-NAME-AND-CANONICAL-TYPE?
				  (CDIRECTORY-CFILES CDIR1) (CDIRECTORY-CFILES CDIR2)))
	(NEWEST?
	 (NOT
	  (OR (EQ (PATHNAME-VERSION (CDIRECTORY-PATHNAME CDIR1)) :WILD)
	     (EQ (PATHNAME-VERSION (CDIRECTORY-PATHNAME CDIR2)) :WILD)))))
    (DOLIST (CLASS (ISOSET-IN-1-ONLY ISOCLASSES))
      (COMPARE-CDIRECTORIES-1 CLASS CDIR2 NEWEST? TAPE-LOSSAGE))
    (DOLIST (CLASS (ISOSET-IN-2-ONLY ISOCLASSES))
      (COMPARE-CDIRECTORIES-1 CLASS CDIR1 NEWEST? TAPE-LOSSAGE))
    (DOLIST (CLASS-PAIR (ISOSET-IN-BOTH ISOCLASSES))
      (LET ((REP1 (CLASS-REPRESENTATIVE (CAR CLASS-PAIR)))
	    (REP2 (CLASS-REPRESENTATIVE (CDR CLASS-PAIR)))
	    (ISOSET (SORT-ISOMORPHIC-ELEMENTS #'SAME-VERSION? (CAR CLASS-PAIR) (CDR CLASS-PAIR))))
	(COND
	  (NEWEST?
	   (LET ((V1 (CFILE-VERSION REP1))
		 (V2 (CFILE-VERSION REP2)))
	     (COND
	       ((OR (NOT (INTEGERP V1)) (NOT (INTEGERP V2)))
		(LET ((CD1 (CFILE-CREATION-DATE REP1))
		      (CD2 (CFILE-CREATION-DATE REP2)))
		  (COND
		    ((> CD1 CD2) (MARK-CFILE-FOR-TRANSFER REP1 CDIR2 TAPE-LOSSAGE))
		    ((> CD2 CD1) (MARK-CFILE-FOR-TRANSFER REP2 CDIR1 TAPE-LOSSAGE)))))
	       ((> V1 V2) (MARK-CFILE-FOR-TRANSFER REP1 CDIR2 TAPE-LOSSAGE))
	       ((> V2 V1) (MARK-CFILE-FOR-TRANSFER REP2 CDIR1 TAPE-LOSSAGE)))))
	  (T
	   (DOLIST (CFILE (ISOSET-IN-1-ONLY ISOSET))
	     (MARK-CFILE-FOR-TRANSFER CFILE CDIR2 TAPE-LOSSAGE))
	   (DOLIST (CFILE (ISOSET-IN-2-ONLY ISOSET))
	     (MARK-CFILE-FOR-TRANSFER CFILE CDIR1 TAPE-LOSSAGE))))))))  



(DEFUN COMPARE-CDIRECTORIES-1 (CLASS TARGET-DIRECTORY NEWEST? TAPE-LOSSAGE)
  (LET ((REP (CLASS-REPRESENTATIVE CLASS)))
    (COND
      (NEWEST? (MARK-CFILE-FOR-TRANSFER REP TARGET-DIRECTORY TAPE-LOSSAGE))
      (T (DOLIST (CFILE CLASS)
	   (MARK-CFILE-FOR-TRANSFER CFILE TARGET-DIRECTORY TAPE-LOSSAGE))))))  



(DEFUN CLASS-REPRESENTATIVE (CFILE-CLASS)
  (LET ((REP (FIND-NEWEST-CFILE CFILE-CLASS)))
    (DOLIST (CFILE CFILE-CLASS)
      (SETF (CFILE-CLASS-REPRESENTATIVE CFILE) REP))
    (SETF (CFILE-REPRESENTED-CLASS REP) CFILE-CLASS)
    REP))  



(DEFUN FIND-NEWEST-CFILE (CFILES)
  (LET ((NEWEST NIL)
	(NV 0.))
    (DOLIST (CFILE CFILES)
      (LET ((V (CFILE-VERSION CFILE)))
	(COND
	  ((NOT (INTEGERP V)) (SETQ NEWEST CFILE) (RETURN ()))
	  ((> V NV) (SETQ NEWEST CFILE
			  NV V)))))
    NEWEST))  


 

;;;; Balance Two Directories
;;; A user level function that ties together everything.

(DEFUN BALANCE-DIRECTORIES (DIRSPEC1 DIRSPEC2 &KEY &OPTIONAL ((:IGNORE REMOVABLE-NAMES)) DIRECTION QUERY-MODE COPY-MODE
  (DEFAULTS "lm:~;*.*") (ERROR T));IGNORE-DESTINATION-DIRECTORY
  "Balance the files on two different directories.  Files are compared
by name and type. Keywords are:

:IGNORE <list> is a list of pathname specifications which should
 be ignored.  These specs are parsed relative to each directory [default: '()].

:DIRECTION <flag> specifies which way to transfer:
 NIL = both ways [the default].
 :1->2 or :2->1 = only one way.

:QUERY-MODE <flag> is designed to allow the system to inform the user if there are any files in the
  destination directory that are more recent than those in the source directory.
  For example, if :direction :1->2 :query-mode :1->2 is specified, the system will move any files
  from 1 to 2 without asking and if there are any files in 2 that are more recent than their
  counterparts in 1, the system will ask the user if they wish to transfer them as well.

    If you want to transfer files in only one direction, but want the system to query you about each
  of those transfers, :direction :1->2 :query-mode :2->1 is the appropriate syntax.

 NIL = don't ask [the default].
 :1->2 = ask about any files that would transfer from DIRSPEC2 to DIRSPEC1 only.
 :2->1 = ask about any files that would transfer from DIRSPEC1 to DIRSPEC2 only.
 :ALWAYS = ask about every file that could be transferred.

:COPY-MODE <flag> has the values:
 :CHARACTERS = copy all files as characters.
 :BINARY = copy all files as binary data.
 :ASK = ask about each file.
 :NEVER-ASK = try to figure out which mode, defaulting to characters.
  NIL = try to figure out which mode, asking if can't [the default].

:DEFAULTS <defaults> is a defaults argument for FS:MERGE-PATHNAME-DEFAULTS
 [default: FS:*DEFAULT-PATHNAME-DEFAULTS*].

:ERROR <boolean> if false, don't cause errors, just print string [default: T].";":IGNORE-DESTINATION-DIRECTORY <boolean> if true considers destination directory
  ;    to be null for balancing purposes.  :DIRECTION must be non null! Use for magtape."
  ;  
  ;  (if (and ignore-destination-directory (null direction))
  ;      (ferror nil "direction must be non-NIL if ignore-destination-directory is non-NIL"))


         ; check for valid arguments
  
  (CHECK-ARG DIRECTION (MEMBER DIRECTION '(NIL :1->2 :2->1) :TEST #'EQ) "NIL, :1->2 or :2->1")
  
  (CHECK-ARG QUERY-MODE (MEMBER QUERY-MODE '(NIL :ALWAYS :1->2 :2->1) :TEST #'EQ)
                         "NIL, :ALWAYS, :1->2 or :2->1")
  
  (CHECK-ARG COPY-MODE (MEMBER COPY-MODE '(:CHARACTERS :BINARY :ASK :NEVER-ASK NIL) :TEST #'EQ)
                        ":CHARACTERS, :BINARY, :ASK, :NEVER-ASK or NIL")
  
  (SETQ REMOVABLE-NAMES (MAPCAR 'PARSE-PATHNAME REMOVABLE-NAMES))
  
  (LET* ((W-DEFAULTS
	  (COND
	    ((CONSP DEFAULTS) DEFAULTS)
	    ((STRINGP DEFAULTS) (PATHNAME DEFAULTS))
	    ((PATHNAMEP DEFAULTS) DEFAULTS)
	    (T NIL)))
	 (DIRPATH1 (MERGE-PATHNAME-DEFAULTS DIRSPEC1 W-DEFAULTS () ())); create directory path
	 (DIR1 (CREATE-CDIRECTORY DIRPATH1 REMOVABLE-NAMES ERROR))     ; create directory without
	                                                               ; the files specified by the
	                                                               ; :IGNORE
	 (DIRPATH2 (MERGE-PATHNAME-DEFAULTS DIRSPEC2 DIRPATH1 () ()))
	 (DIR2 (CREATE-CDIRECTORY DIRPATH2 REMOVABLE-NAMES ERROR))); end of let declarations
    (PRINT DIRPATH1)
    (PRINT DIRPATH2)
    (COND
      ((ERRORP DIR1) DIR1)
      ((ERRORP DIR2) DIR2)
      (T (COMPARE-CDIRECTORIES DIR1 DIR2)                            ; find which files in each
	                                                             ; directory are more recent 
                                                                     ; than their counterparts
	                                                  
       (IF (MEMBER QUERY-MODE '(:ALWAYS :2->1) :TEST #'EQ)           ; query about any possible
	 (CDIRECTORY-TRANSFER-QUERY DIR1))                           ; transfers from the second
                                                                     ; directory to the first
       
       (IF (MEMBER QUERY-MODE '(:ALWAYS :1->2) :TEST #'EQ)           ; query about any possible
	 (CDIRECTORY-TRANSFER-QUERY DIR2))                           ; transfers from the first
                                                                     ; directory to the second
       
       (SETUP-CDIRECTORY-TRANSFER-MODE DIR1 COPY-MODE)
       (SETUP-CDIRECTORY-TRANSFER-MODE DIR2 COPY-MODE)
       (WHEN (OR (MEMBER DIRECTION '(NIL :1->2) :TEST #'EQ)          ; make the file transfer from
		 (MEMBER QUERY-MODE '(:ALWAYS :2->1) :TEST #'EQ))    ; the first to the second if  
	 (CDIRECTORY-TRANSFER DIR1))                                 ; specified by the direction or
                                                                     ; if the user answered yes to
                                                                     ; a query.
       
       (WHEN (OR (MEMBER DIRECTION '(NIL :2->1) :TEST #'EQ)          ; same for files being 
		 (MEMBER QUERY-MODE '(:ALWAYS :1->2) :TEST #'EQ))    ; transferred from the second
	 (CDIRECTORY-TRANSFER DIR2))                                 ; to the first.
       T); otherwise clause
    ); cond
  ); let
)


(DEFUN BALANCE-TO-NON-DIRECTORY-DEVICE (DIRSPEC1 TO-DEV &KEY &OPTIONAL ((:IGNORE REMOVABLE-NAMES)) QUERY-MODE COPY-MODE
  (DEFAULTS *DEFAULT-PATHNAME-DEFAULTS*) (ERROR T))
  "Balance the files on a directory to tape.  Selects most recent versions
and supports ':IGNORE option.  The keywords are:

:IGNORE <list> is a list of pathname specifications which should
    be ignored.  These specs are parsed relative to each directory [default: '()].

:QUERY-MODE <flag> has the values:
    NIL == don't ever ask [the default].
    :1->2 == ask about any files transferred from DIRSPEC2 to DIRSPEC1 only.
    :2->1 == ask about any files transferred from DIRSPEC1 to DIRSPEC2 only.
    :ALWAYS == ask about any files transferred.

:COPY-MODE <flag> has the values:
    :CHARACTERS == copy all files as characters.
    :BINARY == copy all files as binary data.
    :ASK == ask about each file.
    :NEVER-ASK == try to figure out which mode, defaulting to characters.
    NIL == try to figure out which mode, asking if can't [the default].

:DEFAULTS <defaults> is a defaults argument for FS:MERGE-PATHNAME-DEFAULTS
    [default: FS:*DEFAULT-PATHNAME-DEFAULTS*].

:ERROR <boolean> if false, don't cause errors, just print string [default: T]."
  (LET* ((DIRPATH1 (MERGE-PATHNAME-DEFAULTS DIRSPEC1 DEFAULTS () ()))
	 (DIR1 (CREATE-CDIRECTORY DIRPATH1 REMOVABLE-NAMES ERROR))
	 (DIR2 (CREATE-FAKE-CDIRECTORY TO-DEV)))
    (COND
      ((ERRORP DIR1) DIR1)
      (T (COMPARE-CDIRECTORIES DIR1 DIR2 T)
       (IF (MEMBER QUERY-MODE '(:ALWAYS :1->2) :TEST #'EQ)
	 (CDIRECTORY-TRANSFER-QUERY DIR2))
       (SETUP-CDIRECTORY-TRANSFER-MODE DIR1 COPY-MODE)
       (SETUP-CDIRECTORY-TRANSFER-MODE DIR2 COPY-MODE) (CDIRECTORY-TRANSFER DIR1)
       (LIST DIR1 DIR2)))))  
