;;; -*- Mode: Common-Lisp; Base: 10.; Package: File-System -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985-1989 Texas Instruments Incorporated. All rights reserved.



;;; REVSIONS:
;;;     06-06-88 ab     Moved COPY-NEWER-FILES in here from PUBLIC.
;;;     03-18-88 DAB    Fixed delete-directory-files when deleting directory on microExplorer and LX machine.
;;;     02-22-88 DAB    Added support for remote-tape.
;;;     02-17-88 DAB    Fixed delete-directory-files to work on UNIX subdirectories.
;;;     12-07-87 DAB    Fixed Copy-directory to not accumulate NIL (empty directories).
;;;     06.29.87 cac    Added a temporary variable to hold the value of
;;;                     :selective so that when it becomes nil for a
;;;                     particular subdirectory, it regains its
;;;                     original value for the next subdirectory 	
;;;  	01.28.87 MBC	Make sure copy-directory do things alphabetically
;;;  			[call directory-list with :SORTED keyword]
;;;     02.27.87 DAB    Add translated-pathname to Delete-directory to handle logical pathnames
;;;	03.26.87 MBC	Fix delete directory to leave LOCAL-HOST alone.
;;;	04.03.87 MBC    Fix delete directory to check host for :Undeletable-P and
;;;			send :expunge if T, else send :delete-and-expunge.
;;;			Also speed it up by removing fs:file-properties call
;;;			and use info from dired instead.
;;;	04.13.87 MBC	Use old calling sequence to :delete-and-expunge; No pathname.

;; New file copying stuff.
;;
;; Entrypoints:
;; COPY-FILES files to &rest options
;; COPY-DIRECTORY from to &rest options
;; COPY-DIRS list-of-directories from-host to &rest options
;; DELETE-DIRECTORY PATHNAME &keywords
			 
(DEFF FS-COPY-FILE 'COPY-FILE)  

(DEFSUBST PARSE-MULTIPLE-FILE-KEYWORDS ()
  (DO ((O OPTIONS (CDDR O)))
      ((NULL O))
    (CASE (CAR O)
      (:COPY-ONLY (SETQ COPY-ONLY (CADR O)))
      (:SELECTIVE (SETQ SELECTIVE (CADR O)))
      (:CREATE-DIRECTORY (SETQ CREATE-DIRECTORY (CADR O))))))


(DEFUN COPY-FILES (FILES TO &REST OPTIONS &AUX COPY-ONLY SELECTIVE (CREATE-DIRECTORY T) TEM)
  (PARSE-MULTIPLE-FILE-KEYWORDS)
  (DOLIST (FILE FILES)
    RETRY
    (COND
      ((AND (OR (NOT SELECTIVE) (PROGN
				  (FORMAT *QUERY-IO* "Copy ~A ?" FILE)
				  (Y-OR-N-P)))
	  (OR (NULL COPY-ONLY) (CHECK-COPY-ONLY FILE COPY-ONLY)))
       (COND
	 ((AND (ERRORP (SETQ TEM (APPLY #'FS-COPY-FILE FILE TO OPTIONS)))
	     (OR (STRING-EQUAL TEM " ERROR NSD" :end1 10. :end2 10.)
		(STRING-EQUAL TEM " ERROR FNF" :end1 10. :end2 10.))
	     (EQUALP TO "LM"))
	  (LMFS-CREATE-DIRECTORY
	   (SEND (SEND (MERGE-PATHNAME-DEFAULTS FILE) :TRANSLATED-PATHNAME) :DIRECTORY))
	  (GO RETRY)))))))  

(defvar MT-FILEHANDLE nil)  ;02-01-88 DAB
(defvar *REMOTE-TAPE-STREAM* nil) 
(export 'MT-FILEHANDLE 'FS)  ;02-01-88 DAB

(defun COPY-DIRECTORY (SOURCE DESTINATION &REST OPTIONS &KEY SET-BACKED-UP-FLAG ERROR (VERBOSE T)
  DO-NOT-COPY-PROPERTIES &ALLOW-OTHER-KEYS)
  "Copy files and subdirectories specified by SOURCE to DESTINATION.  Wildcarding is supported.
 :SET-BACKED-UP-FLAG - the file system's backed up bit will be set for files backed up.
  Other keyword options allowed are the same as COPY-FILE's.
  Returns four values: 1) A list of default pathname objects for each output file,
                       2) A list of truename objects for each input file,
                       3) A list of outcomes. Outcome will be either a truename object of the output file
                          if the copy was successful, a condition object if successful and :error is nil,
                          or nil if the user was asked whether to copy this file and said no.
                       4) A list of modes used to copy each file."
  (declare (ARGLIST SOURCE DESTINATION &REST OPTIONS &KEY SET-BACKED-UP-FLAG ERROR (VERBOSE T)
		    DO-NOT-COPY-PROPERTIES (COPY-CREATION-DATE T)
		    (COPY-AUTHOR T) REPORT-STREAM (CREATE-DIRECTORIES ':QUERY)
		    )
	   (VALUES TARGET-PATHNAME TARGET-TRUENAME RESULT-PATHNAME COPY-MODE)
	   (special mt-filehandle))
  (FORCE-USER-TO-LOGIN)
  (WHEN VERBOSE
    (SETF (GETF OPTIONS :VERBOSE) T))
  (UNLESS DO-NOT-COPY-PROPERTIES
    (SETF (GETF OPTIONS :DO-NOT-COPY-PROPERTIES) ()))
  (LET* ((OUTNAME)                        ;07-21-87 DAB
	 (INPUT-TRUENAME)
	 (outcome)
	 (mode)
	 (BACKED-UP-FILES-LIST
	  (IF (CONSP SOURCE)
	    (PROGN
	      (SETF (GETF OPTIONS :ERROR) ERROR
		    (GETF OPTIONS :SET-BACKED-UP-FLAG) ())
	      (DOLIST (NEW-SOURCE SOURCE input-truename)
		(multiple-value-bind (v1 v2 v3 v4)          ;07-21-87 DAB
		    (APPLY #'COPY-DIRECTORY NEW-SOURCE DESTINATION OPTIONS)
		  (setq outname  (NCONC outname v1))
		  (setq input-truename (NCONC INPUT-TRUENAME v2))
		  (setq outcome (nconc outcome v3))
		  (setq mode (nconc mode v4)))
		)
	      INPUT-TRUENAME)
	    (IF (OR (STRINGP SOURCE) (TYPEP SOURCE 'PATHNAME) (MAGTAPE-FILEHANDLE SOURCE));Not a stream
	      (FILE-RETRY-NEW-PATHNAME-IF (MEMBER ERROR '(:RETRY :REPROMPT) :TEST #'EQ)
		 (SOURCE FILE-ERROR)
		 (LET* ((PARSED-PATHNAME (PARSE-PATHNAME SOURCE))
			(MERGED-PATHNAME
			 (SEND PARSED-PATHNAME :NEW-PATHNAME
			       :DIRECTORY (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :DIRECTORY) :WILD)
			       :NAME (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :NAME) :WILD)
			       :TYPE (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :TYPE) :wild)
			       :VERSION (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :VERSION) :NEWEST)))
			(PARSED-DESTINATION (PARSE-PATHNAME DESTINATION))
			(ODIR
			 (SEND PARSED-DESTINATION :NEW-PATHNAME :DIRECTORY
			    (IF (MEMBER (PATHNAME-DIRECTORY PARSED-DESTINATION)
				     '(NIL :WILD :ROOT :UNSPECIFIC) :TEST #'EQ);If dir unknown
			      (PATHNAME-DIRECTORY MERGED-PATHNAME);Use source's
			      (PATHNAME-DIRECTORY PARSED-DESTINATION));else own
			    :NAME :WILD :TYPE :WILD :VERSION :WILD)))
		   (UNLESS (and (type-specifier-p 'mt-filehandle)
				(OR (TYPEp ODIR 'MT-FILEHANDLE) (MAGTAPE-FILEHANDLE SOURCE))) ;02-01-88 DAB
		     (SETQ ODIR (DIRECTORY-LIST ODIR :NOERROR :SORTED))	;MBC 1.28.87
		     (AND (NOT (ERRORP ODIR)) (SETF (GETF OPTIONS :OUTPUT-DIRECTORY-LIST) ODIR)))
		   (SETF (GETF OPTIONS :SET-BACKED-UP-FLAG) ()
			 (GETF OPTIONS :PASS-DIRECTORIES-ON) T)
		   (let ((v1    ;07-21-87 DAB
			   (APPLY MERGED-PATHNAME :WILDCARD-MAP #'PRIMITIVE-COPY-DIRECTORY :MAYBE ()
				  MERGED-PATHNAME
				  (IF (STRINGP DESTINATION)
				      (PARSE-PATHNAME DESTINATION)
				      DESTINATION)
				  () () OPTIONS)))
		     ;;v1 may be a list of atoms, or a list of list returned from a recursive called to copy-directory.
		     (dolist (v2 v1 input-truename)
		       (if (not (consp v2))		;; Tung debug -- 12/11/87
			   (setq outname (append outname (list v2)))
			   (IF (consp (car v2))
			       (progn
				 (setq outname (nconc outname (first v2)))
				 (setq INPUT-TRUENAME (nconc INPUT-TRUENAME (second v2)))
				 (setq outcome  (nconc outcome  (third v2)))
				 (setq mode (nconc mode (fourth v2))))
			       (when  (car v2)	;don't add NIL , empty directories. 12-07-87 DAB
				 (setq outname (append outname (list (first v2))))
				 (setq INPUT-TRUENAME (append INPUT-TRUENAME (list (second v2))))
				 (setq outcome (append outcome (list (third v2))))
				 (setq mode (append mode (list (fourth v2)))))
			       ))))		;let
		     ))
	      ;else
	      (MULTIPLE-VALUE-setq (outname INPUT-TRUENAME outcome mode)
		(APPLY #'COPY-FILE SOURCE DESTINATION
		       (PROGN
			 (SETF (GETF OPTIONS :SET-BACKED-UP-FLAG) ()
			       (GETF OPTIONS :PASS-DIRECTORIES-ON) ())
			 OPTIONS)))
		(IF (OR (NULL INPUT-TRUENAME) (ERRORP INPUT-TRUENAME))
		  ()
		    INPUT-TRUENAME)
		))))
    
    (WHEN (and SET-BACKED-UP-FLAG  BACKED-UP-FILES-LIST)
      (REALLY-SET-BACKED-UP-FLAG BACKED-UP-FILES-LIST))
    (values outname INPUT-TRUENAME outcome mode)))


(DEFUN PRIMITIVE-COPY-DIRECTORY (INPUT-PLIST-OR-PATHNAME MAPPED-PATHNAME OUTPUT-SPEC INPUT-STREAM &OPTIONAL IGNORE &REST
				 OPTIONS &KEY SELECTIVE &ALLOW-OTHER-KEYS &AUX INPUT-PATHNAME INPUT-PLIST)
  (declare (special *REMOTE-TAPE-STREAM*))
  (IF (ATOM INPUT-PLIST-OR-PATHNAME)
      (SETQ INPUT-PATHNAME INPUT-PLIST-OR-PATHNAME
	    INPUT-PLIST ())
     (SETQ INPUT-PATHNAME (CAR INPUT-PLIST-OR-PATHNAME)
	    INPUT-PLIST INPUT-PLIST-OR-PATHNAME))
  (UNLESS INPUT-PLIST
    (WHEN INPUT-PATHNAME
      (SETQ INPUT-PLIST (FILE-PROPERTIES INPUT-PATHNAME)
	    INPUT-PATHNAME (CAR INPUT-PLIST))))
  (LET ((BACKED-UP-FILES-LIST
	  (IF (and (MEMBER :DIRECTORY (CDR INPUT-PLIST) :TEST #'EQ)
		   (second (MEMBER :DIRECTORY (CDR INPUT-PLIST) :TEST #'EQ)))   ;02-03-88 DAB must be :directory t
	      (LET* ((DES-DIR (SEND OUTPUT-SPEC :DIRECTORY))
		     (SOURCE (CAR INPUT-PLIST))
		     (ORIGINAL-SELECTIVE (GETF OPTIONS :SELECTIVE))	; save the :selective value 6/26/87 cc
		     (SRC-DIR
		       (SEND (SEND SOURCE :PATHNAME-AS-DIRECTORY) :NEW-PATHNAME :NAME () :TYPE ()
			     :VERSION ())))
		(SETF (GETF OPTIONS :SET-BACKED-UP-FLAG) ())
		(WHEN (OR (NOT SELECTIVE)
			  (CASE (FQUERY
				 '(:CHOICES
				    (((:SELECT "Selective.") #\s) ((T "Yes.") #\y #\t #\SPACE #\)
				     ((NIL "No.") #\n #\RUBOUT #\))
				    :HELP-FUNCTION SELECTIVE-HELPER)
				 "~&Copy Subdirectory ~A ? " SRC-DIR)
			    ((T) (SETF (GETF OPTIONS :SELECTIVE) ()) T)
			    ((NIL) NIL)
			    ((:SELECT) (SETF (GETF OPTIONS :SELECTIVE) T) T)))
		  (prog1
		    (multiple-value-bind (v1 v2 v3 v4)
			(APPLY #'COPY-DIRECTORY (MERGE-PATHNAME-DEFAULTS SRC-DIR MAPPED-PATHNAME)
			       (SEND OUTPUT-SPEC :NEW-PATHNAME :DIRECTORY
				     (IF (OR (MEMBER DES-DIR '(:ROOT :WILD NIL) :TEST #'EQ)
					     (AND (CONSP DES-DIR) (= (LENGTH DES-DIR) 1.)
						  (MEMBER (CAR DES-DIR) '(:ROOT :WILD NIL) :TEST #'EQ)))
					 (LET ((SSS (SEND SOURCE :DIRECTORY)))
					   (IF (OR
						 (MEMBER (SEND SOURCE :DIRECTORY) '(:ROOT :WILD NIL) :TEST #'EQ)
						 (AND (CONSP SSS) (= (LENGTH SSS) 1.)
						      (MEMBER (CAR SSS) '(:ROOT :WILD NIL) :TEST #'EQ)))
					       (SEND SOURCE :NAME)
					       (APPEND (IF (ATOM SSS)
							   (LIST SSS)
							   SSS)
						       (CONS (SEND SOURCE :NAME) ()))))
					 (APPEND (IF (ATOM DES-DIR)
						     (LIST DES-DIR)
						     DES-DIR)
						 (CONS (SEND SOURCE :NAME) ()))))
			       OPTIONS)
		      (list  v1  v2 v3 v4)

		      )				;bind
		    (SETF (GETF OPTIONS :SELECTIVE) ORIGINAL-SELECTIVE)
		    )				;prog1
		  
		  ))				; reset :selective 6/26/87 cc
	      (LET* (COPIED-FILES)
		(SETF (GETF OPTIONS :SET-BACKED-UP-FLAG) ())
		(SETQ COPIED-FILES
		      (APPLY #'PRIMITIVE-COPY-FILE INPUT-PLIST-OR-PATHNAME MAPPED-PATHNAME
			     OUTPUT-SPEC INPUT-STREAM OPTIONS)) 
		(when *REMOTE-TAPE-STREAM*
		  (reject-command *REMOTE-TAPE-STREAM* (if (pathnamep (fourth copied-files))  ;06-06-88 DAB
							   (send (FOURTH COPIED-FILES) :string-for-printing)
							   (fourth copied-files))))
		(list (third COPIED-FILES) (FOURTH COPIED-FILES) (fifth COPIED-FILES) (sixth COPIED-FILES))
		)
	      )))
    BACKED-UP-FILES-LIST))


(defun reject-command (&rest arg)
   "Dummy function. It will be redefined by the MT subsystem when loaded."
  (declare (ignore arg))
  ())


(DEFUN DELETE-DIRECTORY (PATHNAME
			 &OPTIONAL
			 &KEY
                         (verbose nil)
			 (QUERY NIL)
			 (SUBDIRECTORIES T)
			 (DONT-DELETE-NEWEST NIL)
			 (DELETE-PROTECTED-FILES NIL)
			 (DELETE-SELF t)
			 (are-you-sure nil)
			 &AUX DIRECTORY-PATHNAME-AS-FILE
			 local-host-p)		;3.26.87
  "Delete and expunge the files in the specified directory.
PATHNAME is the directory to delete.  The format is \"HOST:DIRECTORY;\".        
Options available:
:verbose is t, a listing is displayed.
:QUERY is T, the caller will be prompted on whether or not to delete each file.
:SUBDIRECTORIES is T, the subdirectories will be deleted as well.
:DONT-DELETE-NEWEST is T, the new version of each file will not be deleted.
:DELETE-PROTECTED-FILES is T, delete protected files will be deleted.
:DELETE-SELF is T, the directory file itself will be deleted.
:are-you-sure is T, the prompt asking the user ARE YOU SURE (Y/N) is suppressed."
  (setf pathname (merge-pathname-defaults (translated-pathname pathname)	;handled logical pathnames
					  (make-pathname :name :wild :type :wild :version :wild)))	;02.27.87
  (when (eq (send pathname :directory) :root)	;dont let them delete the root directory
    (ferror nil "Delete-Directory is not allowed on the ROOT directory."))
  (when (or are-you-sure			;verify this is what the user really whats
	    (y-or-n-p "Deleting directory ~a. Are you sure?" pathname))
    
    (SETF DIRECTORY-PATHNAME-AS-FILE		; Create "host:x;y.directory" from "host:x.y;".
	  (SEND (pathname PATHNAME) :DIRECTORY-PATHNAME-AS-FILE))
    (setq local-host-p (if (eq si:local-host (pathname-host pathname))	;3.26.87
			   t nil))
    ;; Delete files and subdirectories.
    (DELETE-DIRECTORY-FILES pathname
			    verbose QUERY
			    SUBDIRECTORIES
			    DONT-DELETE-NEWEST
			    DELETE-PROTECTED-FILES
			    local-host-p)	;3.26.87
    ;; Delete the directory itself?
    (WHEN DELETE-SELF
      (AND DELETE-PROTECTED-FILES	   
	   (FS:CHANGE-FILE-PROPERTIES (TRUENAME DIRECTORY-PATHNAME-AS-FILE) NIL :DONT-DELETE NIL))	;Unprotect
      (send (truename directory-pathname-as-file) :delete)
      (if (send (send DIRECTORY-PATHNAME-AS-FILE :host) :undeletable-p)	;4.3.87 Use the correct method for host type
	  (send DIRECTORY-PATHNAME-AS-FILE :expunge :error nil)	
	  (send DIRECTORY-PATHNAME-AS-FILE :delete-and-expunge nil)))	;4.13.87
    nil))


(DEFUN DELETE-DIRECTORY-FILES (PATHNAME
			       verbose
			       QUERY
			       SUBDIRECTORIES
			       DONT-DELETE-NEWEST
			       DELETE-PROTECTED-FILES
			       local-host-p)	;3.26.87
  (declare (special lm-signal-pathname-object))
  (let-if (not (boundp 'lm-signal-pathname-object))	;this variable is used in lm-signal-error
	  ((lm-signal-pathname-object pathname))	;The value may be already binded. If so, do not rebind 
    (let ((ENTRIES (cdr (DIRECTORY-list  PATHNAME :sorted)))	;list of elements in this directory
	  return-list)			;list of the files to be deleted form directory
      (DO* ((dir-list entries  (cdr dir-list))	;take one entry at a time and check attributes, then delete
	    (entry (caar dir-list) (caar dir-list))
	    (properties (car dir-list) (car dir-list))
	    (next-entry (car (second dir-list)) (car (second dir-list)))	;used to check dont-delete-newest
	    (ok nil nil)
	    (ok-sub t t)
	    (dont-delete nil nil))
	   ((null dir-list))			; exit do when directory list is completed
	(if (GETF (CDR PROPERTIES) :DIRECTORY)	; If it is a directory
	    (progn 
	      (if SUBDIRECTORIES		;... and the caller wants subdirectories deleted
		(DELETE-DIRECTORY-FILES (SEND (SEND ENTRY :PATHNAME-AS-DIRECTORY)
					      :NEW-PATHNAME
					      :NAME :WILD
					      :TYPE :WILD
					      :VERSION :WILD)
					verbose
					QUERY	;... call ourself with the subdirectory as the pathname
					SUBDIRECTORIES
					DONT-DELETE-NEWEST
					DELETE-PROTECTED-FILES
					local-host-p)	;3.26.87
		(setq ok-sub nil))
	      (when (directory (send (send entry :pathname-as-directory) :new-name "*"))   ;02-17-88 DAB
		(setq ok-sub nil)))		;check the subdirectory, it must be empty to delete this entry
	    (WHEN  DONT-DELETE-NEWEST	    
	      (IF next-entry			;check the next entry. If it has the same pathname we can delete this entry
		  (when (not (string-equal (STRING-RIGHT-TRIM "0 1 2 3 4 5 6 7 8 9" ENTRY)	;ignore version numbers
					   (STRING-RIGHT-TRIM "0 1 2 3 4 5 6 7 8 9" next-entry)))
		    (setq dont-delete t))
		  (setq dont-delete t))))	; Yes, don't delete since last entry in the directory is the newest
	(if (GETF (CDR PROPERTIES) :DONT-DELETE)	;is entry delete protected?
	    (when delete-protected-files	;do you what to delete protected files?
	      (FS:CHANGE-FILE-PROPERTIES ENTRY NIL :DONT-DELETE NIL)	;ok, undelete protect, then delete
	      (setq ok t))
	    (setq ok t))
	(when (and ok				;file is not delete protected
		   ok-sub			;A subdirectory entry is empty
		   (null dont-delete))		;this is not the newest entry - if on.
	  (setf return-list (nconc return-list (list entry)))))	;do  - add this entry to the list
      ;;Delete all entries in the list.
      (if (or verbose				;Do you what a listing or be prompted?
	      query				;Yes, handle each file individually, otherwise delete entire list at once 
	      (null local-host-p))		;delet multiple file is not supported across the network.3.26.87
	  (dolist (entry return-list)		;slower way
	    (if verbose
		(format t "~%~s" (FS:Delete-file ENTRY :QUERY QUERY))
		(FS:Delete-file ENTRY :QUERY QUERY)))
	  (send pathname :delete-multiple-files t return-list))	;faster way
      (when (send (send pathname :host) :undeletable-p)	;4.3.87 Use the correct method for host type
	(send pathname :expunge :error nil)  ;03-18-88 DAB	
	;(send pathname :delete-and-expunge nil)
	  )	;4.13.87
      )))

					



(DEFUN REALLY-SET-BACKED-UP-FLAG (THE-LIST)
  (DOLIST (FILE-ELEMENT THE-LIST)
    (IF (CONSP FILE-ELEMENT)
      (REALLY-SET-BACKED-UP-FLAG FILE-ELEMENT)
      (SEND FILE-ELEMENT :CHANGE-PROPERTIES T :NOT-BACKED-UP ()))))



(DEFUN SELECTIVE-HELPER (STREAM IGNORE IGNORE)
  (FORMAT STREAM
	  "~% S - copy sub-directory, WITH :selective option.~% ~
                     Y - copy sub-directory, WITHOUT :selective option.~% ~
                     N - don't copy sub-directory."))



(DEFUN ELIMINATE-BACKED-UP-FILES (DIRLIST)
  "Returns DIRLIST with only files not backed up"
  (DO* ((FILES (CDR DIRLIST) (CDR FILES))
	(FILE (CAR FILES) (CAR FILES))
	(RETURNABLES (LIST (CAR DIRLIST))))
       ((NULL FILES)
	RETURNABLES)
    (IF (CADR (MEMBER :NOT-BACKED-UP FILE :TEST #'EQ))
      (SETQ RETURNABLES (APPEND RETURNABLES (LIST FILE))))))  



(DEFUN ELIMINATE-DATED-FILES (DATE DIRLIST)
  "Returns DIRLIST with only files created since DATE (Universal Time)."
  (DO* ((FILES (CDR DIRLIST) (CDR FILES))
	(FILE (CAR FILES) (CAR FILES))
	(RETURNABLES (LIST (CAR DIRLIST))))
       ((NULL FILES)
	RETURNABLES)
    (IF (>= (CADR (MEMBER :CREATION-DATE FILE :TEST #'EQ)) DATE)
      (SETQ RETURNABLES (APPEND RETURNABLES (LIST FILE))))))  



(DEFUN COPY-DIRS (DIRS FROM TO &REST OPTIONS)
  (DOLIST (DIR DIRS)
    (APPLY #'COPY-DIRECTORY (MAKE-PATHNAME :HOST FROM :DIRECTORY DIR) TO OPTIONS)))  



(DEFUN COPY-ALL-FILES (FROM TO &REST OPTIONS)
 ;; KKLLUUDDGGEE
  (COND
    ((MEMBER FROM '("MT" "MT:") :TEST 'EQUALP)
     (LOOP UNTIL (ERRORP (APPLY #'FS-COPY-FILE "MT:*" TO OPTIONS))))
    (T
     (LOOP FOR (DIRECTORY) IN (ALL-DIRECTORIES FROM) DO
	(SETQ DIRECTORY (SEND DIRECTORY :NEW-PATHNAME :NAME :WILD :TYPE :WILD :VERSION :WILD))
	(APPLY #'COPY-DIRECTORY DIRECTORY TO OPTIONS)))))  



(DEFUN ELIMINATE-OLDER-FILES (DIRLIST &AUX NU)
  "Returns DIRLIST with only the latest versions of the files."
  (DO ((ELLS DIRLIST (CDR ELLS))
       EL
       FILE
       OTHER)
      ((NULL ELLS)
       (REVERSE NU))
    (SETQ EL (FIRST ELLS))
    (IF (NULL (SETQ FILE (FIRST EL)))
      (PUSH EL NU)
      (PROGN
	(SETQ OTHER
	      (MEMBER FILE NU :TEST
		      #'(LAMBDA (P1 P2)
			  (AND (SETQ P2 (CAR P2)); get pathname, ensure non-()
			     (EQUAL (SEND P1 :TYPE) (SEND P2 :TYPE))
			     (EQUAL (SEND P1 :NAME) (SEND P2 :NAME))))))
	(IF OTHER
	  (IF (> (GET EL :CREATION-DATE) (GET (FIRST OTHER) :CREATION-DATE))
	    (SETF (FIRST OTHER) EL))
	  (PUSH EL NU))))))  


;; MAG TAPE SUPPORT FOR PRIMITIVE COPY FILE
;; IT IS HERE BECAUSE IT CAN NOT BE IN IO;OPEN
;; FOR BUILD REASONS.

(DEFUN MAGTAPE-FILEHANDLE (SPEC)
  (and (type-specifier-p 'mt-filehandle) (TYPEp SPEC 'MT-FILEHANDLE)))    ;02-01-88 DAB

(DEFUN MAGTAPE-EOT-STREAM-COPY (&rest ignore)
  (ferror NIL "Copy continuation at EOT not supported."))

  

;;(comment					;MAGTAPE STUFF THAT IS NOT COMPLETE

;;;; Found this exactly as it is... didn't have time to investigate why it
;;;; was left like this or how to best fix...
;;;;(defun same-ignoring-version-and-directory (p1 p2)
;;;;  (and (equal (send p1 ':type) (send p2 ':type))
;;;;       (equal (send p1 ':name) (send p2 ':name))



;;(DEFUN COPY-EOT-HANDLER (&REST IGNORE)
;;  (COND
;;    ((EQ PROCESSOR-TYPE-CODE SI::CHAPARRAL-TYPE-CODE) (MT::WRITE-EOF)
;;     (FORMAT T "~%~%*** End of Tape: rewinding... ") (MT::REWIND)
;;     (FORMAT T "~&*** Time to install a new tape cartridge ***"))
;;    (T (MT::SPACE-REV-TO-BOF 0. 1.) (MT::WRITE-EOF) (MT-OFFLINE)
;;     (FORMAT T "~%~%*** Time to mount a new tape ***")))
;;  (BEEP)
;;  (FORMAT T " Press return when tape is ready: ")
;;  (READLINE)
;;  (FORMAT T "~&")
;;  (THROW 'MAGTAPE-EOT
;;	 :MAGTAPE-EOT))  


;;(DEFUN COPY-MOUNT-NEXT-TAPE (TAPE-DUMP)
;;  (IF TAPE-DUMP
;;    (MT-WRITE-EOF))
;;  (MT-REWIND)
;;  (MT-OFFLINE)
;;  (BEEP)
;;  (FORMAT T "~&>>> Time to mount a new tape <<<")
;;  (READLINE))

;;(DEFUN MAGTAPE-EOT-STREAM-COPY (IN OUT INPUT-PLIST-OR-PATHNAME MAPPED-PATHNAME OUTPUT-SPEC INPUT-STREAM ERROR
;;  COPY-CREATION-DATE COPY-AUTHOR REPORT-STREAM VERBOSE CREATE-DIRECTORIES CHARACTERS BYTE-SIZE
;;  OVERWRITE DELETE AFTER DIRECTORY-LIST DEFAULT-BYTE-SIZE OUTPUT-DIRECTORY-LIST SELECTIVE
;;  NOT-BACKED-UP-ONLY COPY-ONLY)
;;  (CONDITION-BIND
;;   (('MT::END-OF-TAPE (FUNCTION EOT-HANDLER) INPUT-PLIST-OR-PATHNAME MAPPED-PATHNAME OUTPUT-SPEC
;;     INPUT-STREAM ERROR COPY-CREATION-DATE COPY-AUTHOR REPORT-STREAM VERBOSE CREATE-DIRECTORIES
;;     CHARACTERS BYTE-SIZE OVERWRITE DELETE AFTER DIRECTORY-LIST DEFAULT-BYTE-SIZE
;;     OUTPUT-DIRECTORY-LIST SELECTIVE NOT-BACKED-UP-ONLY COPY-ONLY))
;;   (STREAM-COPY-UNTIL-EOF IN OUT)))  



;;(DEFUN EOT-HANDLER (INPUT-PLIST-OR-PATHNAME MAPPED-PATHNAME OUTPUT-SPEC INPUT-STREAM ERROR COPY-CREATION-DATE
;;  COPY-AUTHOR REPORT-STREAM VERBOSE CREATE-DIRECTORIES CHARACTERS BYTE-SIZE OVERWRITE DELETE
;;  AFTER DIRECTORY-LIST DEFAULT-BYTE-SIZE OUTPUT-DIRECTORY-LIST SELECTIVE NOT-BACKED-UP-ONLY
;;  COPY-ONLY)
;; ;;;
;; ;;;  TO GET EOT HANDLER WORKING:                               9-25-85 MBC
;; ;;;  
;; ;;;  Several things need to be done to get magtape EOT handling working for
;; ;;;  copying file to tape.
;; ;;;
;; ;;;  1.  SI:HANDLE-TAPE-ERROR needs to recognize the error code returned in 
;; ;;;      the RQB and signal the proper condition, MT:END-OF-TAPE.
;; ;;;
;; ;;;  2.  The condition handler for EOT is set up by MAGTAPE-EOT-STREAM-COPY, but the open
;; ;;;      and close of the magtape stream occur before and after this.  Since both the
;; ;;;      open (writes header to tape) and the close (flushes last output buffer) do 
;; ;;;      I/O to the tape, an EOT encountered here is not within the condition handler's scope.
;; ;;;
;; ;;;  2.  In PRIMITIVE-COPY-FILE the function MAGTAPE-EOT never calls MAGTAPE-EOT-STREAM-COPY 
;; ;;;      because the predicate MAGTAPE-FILEHANDLE is called with a stream of type
;; ;;;      MT:MT-FILE-CHARACTER-OUTPUT-STREAM.
;; ;;;  
;; ;;;  3.  EOT-HANDLER is called with too many arguments.  Apparently the condition bind
;; ;;;      is pushing the condition object onto the argument list.
;; ;;;
;; ;;;  4.  When EOT-HANDLER is called properly, it calls PRIMITIVE-COPY-FILE, who calls
;; ;;;      DETERMINE-CHARACTERS-AND-BYTE-SIZE with a character input stream.  This stream
;; ;;;      can't handle the TYPE message it is sent.
;; ;;;
;;  (MT::WRITE-EOF)
;;  (FORMAT T "~%~%*** End of Tape: rewinding... ")
;;  (MT::REWIND)
;;  (FORMAT T "~&*** Time to install a new tape cartridge ***")
;;  (BEEP)
;;  (FORMAT T " Press return when tape is ready: ")
;;  (READLINE)
;;  (FORMAT T "~&")
;;  (PRIMITIVE-COPY-FILE INPUT-PLIST-OR-PATHNAME MAPPED-PATHNAME OUTPUT-SPEC INPUT-STREAM :ERROR
;;		       ERROR :COPY-CREATION-DATE COPY-CREATION-DATE :COPY-AUTHOR COPY-AUTHOR
;;		       :REPORT-STREAM REPORT-STREAM :VERBOSE VERBOSE :CREATE-DIRECTORIES
;;		       CREATE-DIRECTORIES :CHARACTERS CHARACTERS :BYTE-SIZE BYTE-SIZE :OVERWRITE
;;		       OVERWRITE :DELETE DELETE :AFTER AFTER :DIRECTORY-LIST DIRECTORY-LIST
;;		       :DEFAULT-BYTE-SIZE DEFAULT-BYTE-SIZE :OUTPUT-DIRECTORY-LIST
;;		       OUTPUT-DIRECTORY-LIST :SELECTIVE SELECTIVE :NOT-BACKED-UP-ONLY
;;		       NOT-BACKED-UP-ONLY :COPY-ONLY COPY-ONLY)
;;  :RE-COPY)  
;;)



;;;   *-----------------------------------------------------------*
;;;   |			   COPY-NEWER-FILES			  |
;;;   *-----------------------------------------------------------*

;;  12/04/87 DNG - Original.
;;  12/22/87 DNG - Rewritten to avoid using unreliable FS:BALANCE-DIRECTORIES.
;;   1/09/88 DNG - Add support for copying sub-directories.
;;   5/26/88 CLM - Fixed problem in call to fs:directory-list (options to function
;;                 are keywords without following value).

(defun copy-newer-files (from-directory to-directory &rest options)
  "For each file in FROM-DIRECTORY that is more recent than the corresponding 
file in TO-DIRECTORY, copy the file from FROM-DIRECTORY to TO-DIRECTORY.
This is particularly useful for updating copies on Unix or Macintosh where 
there aren't any version numbers.  The keyword options have the same meaning 
as for COPY-FILE, although many of the default values are different."
  (declare (arglist from-directory to-directory &key (:selective t) (:report-stream t) :after
		    (:create-directories t) (:characters :maybe-ask)
		    (:not-backed-up-only nil) (:set-backed-up-flag nil)))
  (when (= (length options) 1)
    ;; for compatibility with old calling sequence
    (setf options (list ':selective (first options))))
  (let* ((from-path (pathname from-directory))
	 (to-path (pathname to-directory))
	 ;; Carefully avoid using pathname merging because it is too hard to
	 ;; get it to do what I want.
	 (from-dir (send from-path :new-pathname
			  :name (or (send from-path :name) :wild)
			  :type (or (send from-path :type) :wild)
			  :version :newest))
	 (to-dir (send (send to-path :new-pathname
			     :directory (or (send to-path :directory)
					    (send from-path :directory))
			     :name nil
			     :type nil
			     :version nil)
		       :translated-pathname))
	 (from-list (fs:directory-list from-dir))
	 (to-list (fs:directory-list (send to-dir :new-pathname
					   :name (or (send from-path :name) :wild)
					   :type (or (send from-path :canonical-type) :wild) ;ab 3/18/88
					   :version :newest)
				     :noerror ))	;ab 3/17/88
	 (files-to-copy '())
	 (directories-to-copy '())
	 (source-versions '?)
	 (destination-versions '?))
    (declare (list files-to-copy))
   (multiple-value-bind (query-p after copy-file-options)
	(apply #'(lambda (&key (selective t) (report-stream t) (after most-negative-fixnum)
			       (create-directories t) (characters :maybe-ask)
			       (not-backed-up-only nil) (set-backed-up-flag nil))
		   (values selective
			   (if (integerp after) after (time:parse-universal-time after 0 nil nil))
			   (list :report-stream report-stream :characters characters
				 :create-directories create-directories
				 :not-backed-up-only not-backed-up-only
				 :set-backed-up-flag set-backed-up-flag)))
	       options)
    (flet ((newer-pathname-p (from-file from-plist to-file to-plist)
	      ;; is from-file newer than to-file?
	      (let ((from-version (send from-file :version))
		    (to-version (send to-file :version)))
		(if (and (numberp from-version)
			 (numberp to-version))
		    (> from-version to-version)
		  (> (getf from-plist :creation-date 0)
		     (getf to-plist :creation-date 0))))))
      (dolist (from-info from-list)
	(let ((from-file (car from-info))
	      (from-plist (cdr from-info)))
	  (cond ((null from-file)) ; information about current directory, ignore.
		((getf from-plist :directory nil)	; a sub-directory
		 (let* ((from-name (send from-file :name))
			(sub-dir (send from-file :new-pathname
				       :directory (append (send from-file :directory)
							  (list from-name))
				       :name nil :type nil :version nil)))
		   (when (or (not query-p)
			     (y-or-n-p "Copy sub-directory \"~A\"?" sub-dir))
		     (push (list sub-dir
				 (send to-dir :new-pathname
				       :directory (append (send to-dir :directory)
							  (list from-name))
				       ))
			   directories-to-copy)) ))
		(t				; else a file.
		 (when (eq source-versions '?)
		   (setq source-versions (numberp (send from-file :version))))
		 (let* ((from-name (send from-file :name))
			(from-type (send from-file :canonical-type))
			(to-pathname (send to-dir :new-pathname
					   :name from-name
					   :canonical-type (send from-file :canonical-type)
					   :version (if (and source-versions destination-versions)
							(send from-file :version)
						      :newest)))
			(to-file nil)
			(to-plist nil)
			)
		   (dolist (to-info to-list)
		     (unless (null (car to-info))
		       (when (eq destination-versions '?)
			 (setq destination-versions (numberp (send (car to-info) :version))))
		       (when (and (string-equal from-name
						(send (car to-info) :name))
				  (equal from-type (send (car to-info) :canonical-type)))
			 (setq to-file (car to-info))
			 (setq to-plist (cdr to-info))
			 (return))))
		   (when (and (or (null to-file)
				  (newer-pathname-p from-file from-plist to-file to-plist))
			      (>= (getf from-plist :creation-date 0) after)
			      (or (not query-p)
				  (y-or-n-p "Copy \"~A\" to \"~A\"?" from-file to-pathname)))
		     (push (list from-file to-pathname)
			   files-to-copy))))))))
    (setq directories-to-copy (nreverse directories-to-copy))
    (dolist (elt directories-to-copy)
      (CATCH-ERROR-RESTART (ERROR "Give up copying sub-directory \"~A\"" (first elt))
	(apply #'copy-newer-files (first elt) (second elt) options)
	))
    (setq files-to-copy (nreverse files-to-copy))
    (dolist (elt files-to-copy)
      (CATCH-ERROR-RESTART (ERROR "Give up copying \"~A\" and continue with the next file."
				  (first elt))
	(apply #'copy-file (first elt) (second elt) copy-file-options)
	)))
    (values)))




