;;; -*- Mode:Common-LISP; Package:SYSTEM-INTERNALS; Lowercase:T; Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151

;; Copyright (c) 1980 Massachusetts Institute of Technology 
;; Copyright (c) 1984-1989 Texas Instruments Incorporated.  All Rights Reserved.

;;; 12-16-88 DAB  Fixed DRIBBLE-STREAM-IO to make sure *file-stream* is open before processing OP. Also
;;;               added :item to OP.
;;; 12-15-88 DAB  Updated the documentation line in dribble-all.
;;; ************* Rel 5.0 OS 12-14-88 DAB
;;; 09-15-87 DAB Fixed Dribble-start and Dribble-all to bind *standard-output* and *standard-input* inside the
;;;              Unwind-protect boby. You could get dual file allocation if you are using format statements in low 
;;;              level file and disk functions.
;;; 04.15.87 DAB Fixed Dribble-end error trying to throw with no catch.


(DEFUN DRIBBLE (&OPTIONAL FILENAME)
  "Sends input and output to a dribble file, FILENAME by rebinding 
*STANDARD-OUTPUT* and *STANDARD-INPUT*.  
With no arg, exits a previously entered DRIBBLE and closes the dribble file.
To rebind *TERMINAL-IO*, see DRIBBLE-ALL"
  (IF FILENAME
    (DRIBBLE-START FILENAME)
    (DRIBBLE-END)))  

;This binds STANDARD-OUTPUT and STANDARD-INPUT and enters a new read-eval-print
;loop.  SETQ'ing them would be global for all processes and would leave you
;totally shafted if the file connection broke.



(DEFUN DRIBBLE-START (FILENAME &OPTIONAL EDITOR-P)
  "Sends input and output to a file or an file editor buffer, 
if second arg is T.  Rebinds *STANDARD-OUTPUT* and *STANDARD-INPUT*.  
With no arg, exits a previously entered DRIBBLE and closes the dribble file.
To rebind *TERMINAL-IO*, see DRIBBLE-ALL"
  (IF (AND (NULL FILENAME) (NULL EDITOR-P))
      (DRIBBLE-END)
      (PROGN
	(FORMAT T "~&Entering Dribble Read-Eval-Print Loop.  Type (DRIBBLE) to exit.")
	(LET ((DRIBBLE-STREAM
		(MAKE-DRIBBLE-STREAM *TERMINAL-IO*
				     (IF (NOT EDITOR-P)
					 (OPEN FILENAME :DIRECTION :OUTPUT :ERROR :REPROMPT)
					 (ZWEI:MAKE-FILE-BUFFER-STREAM FILENAME)))))
	  (UNWIND-PROTECT (CATCH
			    'DRIBBLE-END
			    (let ((*STANDARD-INPUT* DRIBBLE-STREAM)    ;09-15-87 DAB
				  (*STANDARD-OUTPUT* DRIBBLE-STREAM))
			      (IF (AND (GET 'UCL::BASIC-COMMAND-LOOP 'FLAVOR)
				       (TYPEP *TERMINAL-IO* 'UCL::BASIC-COMMAND-LOOP)
				       (ZEROP EH::ERROR-DEPTH))
				  (UCL-DRIBBLE-LOOP *TERMINAL-IO*)
				  (LISP-TOP-LEVEL1 *TERMINAL-IO*))))
	    (SEND  DRIBBLE-STREAM :DRIBBLE-END))))))  ;09-15-87 DAB



(DEFUN DRIBBLE-ALL (&OPTIONAL FILENAME EDITOR-P)
  "Sends input and output to a file or a file editor buffer,
if second arg is T.  This differs from DRIBBLE and DRIBBLE-START in that it 
rebinds *TERMINAL-IO*, not just *STANDARD-INPUT* and *STANDARD-OUTPUT*,
so that queries, break loops, etc. are also included.
With no argument, exits a previously entered DRIBBLE-ALL and closes the file.
Entering (DRIBBLE) or (DRIBBLE-END) will exit DRIBBLE-ALL and closes the file."   ;12-14-88 DAB
  (declare (special  UCL::THIS-APPLICATION))
  (IF (AND (NULL FILENAME) (NULL EDITOR-P))
      (DRIBBLE-END)
      (PROGN
	(FORMAT T "~&Entering Dribble Read-Eval-Print Loop.  Type (DRIBBLE-END) or (DRIBBLE) to exit.")
	(LET ((DRIBBLE-STREAM
		(MAKE-DRIBBLE-STREAM *TERMINAL-IO*
				     (IF (NOT EDITOR-P)
					 (OPEN FILENAME :DIRECTION :OUTPUT :ERROR :REPROMPT)
					 (ZWEI:MAKE-FILE-BUFFER-STREAM FILENAME)))))
	  (UNWIND-PROTECT (CATCH 'DRIBBLE-END
			    (let ((*STANDARD-INPUT* DRIBBLE-STREAM)    ;09-15-87 DAB
				  (*STANDARD-OUTPUT* DRIBBLE-STREAM))
			      (COND
				((AND (GET 'UCL::BASIC-COMMAND-LOOP 'FLAVOR)
				      (TYPEP *TERMINAL-IO* 'UCL::BASIC-COMMAND-LOOP)
				      (ZEROP EH::ERROR-DEPTH))
				 (UCL-DRIBBLE-LOOP DRIBBLE-STREAM))
				((AND (BOUNDP 'UCL::THIS-APPLICATION) UCL::THIS-APPLICATION)
				 (UCL-DRIBBLE-LOOP DRIBBLE-STREAM UCL::THIS-APPLICATION))
				(T (LISP-TOP-LEVEL1 DRIBBLE-STREAM)))))
	    ;; Do this with old definition of *terminal-io*, in case it gets an error.
	    (SEND DRIBBLE-STREAM :DRIBBLE-END))))))  



(DEFUN UCL-DRIBBLE-LOOP (*TERMINAL-IO* &OPTIONAL (UCL-APPLICATION *TERMINAL-IO*))
  (CATCH 'UCL::EXIT-COMMAND-LOOP
    (ERROR-RESTART-LOOP ((ERROR ABORT) "Return to the Dribble command loop.")
       (CONDITION-CASE () (CATCH 'UCL::COMMAND-ABORT
			    (SEND UCL-APPLICATION :FETCH-AND-EXECUTE))
	  (ABORT (SEND UCL-APPLICATION :HANDLE-PROMPT))))))  



(DEFUN DRIBBLE-END ()
  "Exit from the recursive read-eval-print loop entered by DRIBBLE-START or DRIBBLE-ALL."
  (and (closurep *standard-input*)  ;check to make sure you are in a dribble loop. 04.15.87 DAB
       (member 'dribble-stream (closure-alist *standard-input*) :key #'car :test #'eq)
    (THROW 'DRIBBLE-END
	   ())))  



(DEFUN MAKE-DRIBBLE-STREAM (TV-STREAM FILE-STREAM)
  (LET ((*RUBOUT-HANDLER-BUFFER* (MAKE-ARRAY 100. :ELEMENT-TYPE 'STRING-CHAR :FILL-POINTER 0.))
	(DRIBBLE-STREAM)
	(*TV-STREAM* TV-STREAM)
	(*FILE-STREAM* FILE-STREAM))
    (DECLARE (SPECIAL *TV-STREAM* *FILE-STREAM* DRIBBLE-STREAM *RUBOUT-HANDLER-BUFFER*))
    (SETQ DRIBBLE-STREAM
	  (CLOSURE '(*TV-STREAM* *FILE-STREAM* *RUBOUT-HANDLER-BUFFER* DRIBBLE-STREAM)
		   'DRIBBLE-STREAM-IO))))  


(DEFUN DRIBBLE-STREAM-IO (OP &REST ARGS)
  (DECLARE (SPECIAL *TV-STREAM* *FILE-STREAM* DRIBBLE-STREAM *RUBOUT-HANDLER-BUFFER*))
  (when (eq (send *file-STREAM* :Send-if-handles :status) ; DAB 01-19-89 may be a zmacs buffer.
	    :closed) ;make-sure file has not been closed out from under us. 12-16-88 DAB
    (unless (eq op :dribble-end) (dribble-end)))
  (CASE
    OP
    ((:ANY-TYO :TYO) (APPLY *TV-STREAM* OP ARGS) (FUNCALL *FILE-STREAM* OP (CAR ARGS)))
    ((:STRING-OUT :LINE-OUT) (APPLY *TV-STREAM* OP ARGS)
			     (APPLY *FILE-STREAM* OP (STRING (CAR ARGS)) (CDR ARGS)))
    ((:item) (APPLY *TV-STREAM* OP ARGS)  ;12-16-88 DAB [4302]
	     (if (third args)
		 (send *FILE-STREAM* :string-out (apply #'format () (cddr args)))
		 (princ *FILE-STREAM* (second args))))
    (:FRESH-LINE (FUNCALL *TV-STREAM* OP) (FUNCALL *FILE-STREAM* OP))
    ((:ANY-TYI :TYI :TYI-NO-HANG :ANY-TYI-NO-HANG)
     (BLOCK
       NIL
       (OR RUBOUT-HANDLER (SEND *FILE-STREAM* :SEND-IF-HANDLES :FORCE-OUTPUT))
       (CATCH
	 (IF RUBOUT-HANDLER 'RUBOUT-HANDLER 'DUMMY-TAG)
	 (LET ((CH (SEND *TV-STREAM* OP)))
	   (AND CH
		RUBOUT-HANDLER
		(VECTOR-PUSH-EXTEND (IF (LISTP CH) (CADR CH) CH)
				    *RUBOUT-HANDLER-BUFFER*))
	   (RETURN CH)))
       (STORE-ARRAY-LEADER 0 *RUBOUT-HANDLER-BUFFER* 0)
       (THROW 'RUBOUT-HANDLER NIL)
       NIL))
    (:UNTYI (SEND *TV-STREAM* :UNTYI (CAR ARGS))
	    (AND RUBOUT-HANDLER
		 (PLUSP (LENGTH *RUBOUT-HANDLER-BUFFER*))
		 (DECF (ARRAY-LEADER *RUBOUT-HANDLER-BUFFER* 0))))
    ((:RUBOUT-HANDLER :PREEMPTABLE-READ)
     (STORE-ARRAY-LEADER 0 *RUBOUT-HANDLER-BUFFER* 0)
     (PROG (VALS)
	   (SETQ VALS (MULTIPLE-VALUE-LIST (APPLY *TV-STREAM* OP ARGS)))
	   (SEND *FILE-STREAM* :STRING-OUT *RUBOUT-HANDLER-BUFFER*)
	   (SEND *FILE-STREAM* :SEND-IF-HANDLES :FORCE-OUTPUT)
	   (RETURN (VALUES-LIST VALS))))
    (:DRIBBLE-END (FORMAT *TV-STREAM*
			  "~&Closing dribble file, ~A.~%"
			  (SEND *FILE-STREAM* :TRUENAME))
		  (CLOSE *FILE-STREAM*)
		  (SEND *FILE-STREAM* :SEND-IF-HANDLES :TRUENAME))
    (:NOTICE (IF (SEND *TV-STREAM* :OPERATION-HANDLED-P :NOTICE)
		 (APPLY *TV-STREAM* :NOTICE ARGS)
		 'COLD-LOAD-STREAM))
    (:INCREMENT-CURSORPOS (COND ((EQ (CADDR ARGS) :CHARACTER)
				 (DOTIMES (Y-INCREMENT  (CADR ARGS))
				   (SEND *FILE-STREAM* :TYO #\NEWLINE))
				 (DOTIMES (X-INCREMENT  (CAR ARGS))
				   (SEND *FILE-STREAM* :TYO #\SPACE))))
			  (APPLY *TV-STREAM* OP ARGS))
    ((:FINISH :FORCE-OUTPUT) (APPLY *FILE-STREAM* :SEND-IF-HANDLES OP ARGS)
			     (APPLY *TV-STREAM* OP ARGS))
    (:SEND-IF-HANDLES (APPLY *FILE-STREAM* :SEND-IF-HANDLES OP ARGS)
		      (APPLY *TV-STREAM* OP ARGS))
    (OTHERWISE (APPLY *TV-STREAM* OP ARGS))))
