;-*- Mode:Common-LISP; Package:FILE-SYSTEM; Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1980, Massachusetts Institute of Technology
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;;  Define the error flavors used by all file errors.

;;;  REVISION:
;;;	01.05.87 MBC	47 DEFSIGNALs moved here from CHAOS;QFILE-DEFINITIONS.
;;;     02-26-87 DAB    Modified proceed types file-operation-failure DIRED,CLEAN-DIRECTORY EXPUNGE-DIRECTORY.
;;;     03.04.87 DAB    Added a defsignal for rename-directory-version-not-1.
(DEFFLAVOR FILE-ERROR (PATHNAME OPERATION) (NO-ACTION-MIXIN ERROR)
  :ABSTRACT-FLAVOR
  :GETTABLE-INSTANCE-VARIABLES
  :INITTABLE-INSTANCE-VARIABLES) 


(DEFMETHOD (FILE-ERROR :AFTER :INIT) (IGNORE)
  (SETQ PATHNAME (GETF SI:PROPERTY-LIST :PATHNAME))
  (OR (VARIABLE-BOUNDP OPERATION) (SETQ OPERATION (GETF SI:PROPERTY-LIST :OPERATION)))) 


(DEFMETHOD (FILE-ERROR :SET-FORMAT-ARGS) (NEW-ARGS)
  (SETQ EH::FORMAT-ARGS NEW-ARGS)) 

(DEFMETHOD (FILE-ERROR :CASE :PROCEED-ASKING-USER :RETRY-FILE-OPERATION) (CONTINUATION IGNORE)
  "Proceeds, trying the file operation again."
  (FUNCALL CONTINUATION :RETRY-FILE-OPERATION))

(DEFMETHOD (FILE-ERROR :CASE :PROCEED-ASKING-USER :NEW-PATHNAME) (CONTINUATION READ-OBJECT-FUNCTION)
  "Proceeds, reading a new filename and using that instead."
  (FUNCALL CONTINUATION :NEW-PATHNAME
	   (FUNCALL READ-OBJECT-FUNCTION `(:PATHNAME :DEFAULTS ,PATHNAME)
		    "Pathname to use instead: (default ~A) " PATHNAME)))

(DEFFLAVOR FILE-REQUEST-FAILURE () (FILE-ERROR)) 

(DEFFLAVOR FILE-OPERATION-FAILURE () (FILE-ERROR)
  (:DOCUMENTATION "The file system understood an operation but decided it was erroneous.")) 


(DEFMETHOD (FILE-OPERATION-FAILURE :CASE :PROCEED-ASKING-USER :DIRED) (&REST IGNORE)
  "Runs DIRED. Press TERM-B  to return to the debugger."
  (DIRED (if (SEND SELF :PATHNAME)
	     (SEND (SEND SELF :PATHNAME) :NEW-PATHNAME :NAME :WILD :TYPE :WILD :VERSION :WILD)
	     (make-PATHNAME :NAME :WILD :TYPE :WILD :VERSION :WILD)))
  ())


(DEFMETHOD (FILE-OPERATION-FAILURE :CASE :PROCEED-ASKING-USER :CLEAN-DIRECTORY) (&REST IGNORE)
  "Cleans the directory, then returns to debugger."
  (ZWEI::CLEAN-DIRECTORY
    (if (SEND SELF :PATHNAME)
	(SEND (SEND SELF :PATHNAME) :NEW-PATHNAME :NAME :WILD :TYPE :WILD :VERSION :WILD)
	(make-PATHNAME :NAME :WILD :TYPE :WILD :VERSION :WILD)))
  ()) 


(DEFMETHOD (FILE-OPERATION-FAILURE :CASE :PROCEED-ASKING-USER :EXPUNGE-DIRECTORY) (&REST IGNORE)
  "Expunges the directory, then returns to debugger."
  (EXPUNGE-DIRECTORY
    (if (SEND SELF :PATHNAME)
	(SEND (SEND SELF :PATHNAME) :NEW-PATHNAME :NAME :WILD :TYPE :WILD :VERSION :WILD)
	(make-PATHNAME :NAME :WILD :TYPE :WILD :VERSION :WILD)))
  ()) 


(DEFFLAVOR NO-FILE-SYSTEM () (FILE-REQUEST-FAILURE)) 


(DEFMETHOD (NO-FILE-SYSTEM :USER-PROCEED-TYPES) (REAL-PROCEED-TYPES)
  (APPEND (INTERSECTION '(:NO-ACTION :RETRY-FILE-OPERATION) REAL-PROCEED-TYPES :TEST #'EQ)
	  '(:BOOT-FILE-SYSTEM-AND-RETRY)
	  (REMOVE-IF
	   #'(LAMBDA (ELT)
	       (MEMBER ELT '(:NO-ACTION :RETRY-FILE-OPERATION :BOOT-FILE-SYSTEM-AND-RETRY) :TEST
		       #'EQ))
	   REAL-PROCEED-TYPES))) 


(DEFMETHOD (NO-FILE-SYSTEM :CASE :PROCEED-ASKING-USER :BOOT-FILE-SYSTEM-AND-RETRY) (CONTINUATION IGNORE)
  "Boot the local file system and try again."
  (BOOT-FILE-SYSTEM)
  (FUNCALL CONTINUATION :RETRY-FILE-OPERATION)) 




(DEFFLAVOR NO-MORE-ROOM-ERROR () (FILE-OPERATION-FAILURE)) 


(DEFMETHOD (NO-MORE-ROOM-ERROR :USER-PROCEED-TYPES) (REAL-PROCEED-TYPES)
  (APPEND (INTERSECTION '(:NO-ACTION :RETRY-FILE-OPERATION) REAL-PROCEED-TYPES :TEST #'EQ)
	  '(:DIRED :CLEAN-DIRECTORY)
	  (REMOVE-IF
	   #'(LAMBDA (ELT)
	       (MEMBER ELT '(:NO-ACTION :RETRY-FILE-OPERATION :DIRED :CLEAN-DIRECTORY) :TEST
		       #'EQ))
	   REAL-PROCEED-TYPES))) 



(DEFFLAVOR FILE-LOOKUP-ERROR () (FILE-OPERATION-FAILURE)) 


(DEFFLAVOR DIRECTORY-NOT-FOUND-ERROR () (FILE-LOOKUP-ERROR)) 

(DEFMETHOD (DIRECTORY-NOT-FOUND-ERROR :USER-PROCEED-TYPES) (REAL-PROCEED-TYPES)
  (APPEND (INTERSECTION '(:NO-ACTION :RETRY-FILE-OPERATION) REAL-PROCEED-TYPES :TEST #'EQ)
	  '(:CREATE-DIRECTORY-AND-RETRY)
	  (REMOVE-IF
	   #'(LAMBDA (ELT)
	       (MEMBER ELT '(:NO-ACTION :RETRY-FILE-OPERATION :CREATE-DIRECTORY-AND-RETRY) :TEST
		       #'EQ))
	   REAL-PROCEED-TYPES))) 


(DEFMETHOD (DIRECTORY-NOT-FOUND-ERROR :CASE :PROCEED-ASKING-USER :CREATE-DIRECTORY-AND-RETRY) (CONTINUATION IGNORE)
  "Creates the directory and tries again."
  (CREATE-DIRECTORY (SEND SELF :PATHNAME) :RECURSIVE T)
  (FUNCALL CONTINUATION :RETRY-FILE-OPERATION)) 

(DEFFLAVOR RENAME-FAILURE () (FILE-OPERATION-FAILURE)
  (:DEFAULT-INIT-PLIST :OPERATION ':RENAME)) 


(DEFFLAVOR CHANGE-PROPERTY-FAILURE () (FILE-OPERATION-FAILURE)
  (:DEFAULT-INIT-PLIST :OPERATION ':CHANGE-PROPERTIES)) 


(DEFFLAVOR DELETE-FAILURE () (FILE-OPERATION-FAILURE)
  (:DEFAULT-INIT-PLIST :OPERATION ':DELETE)) 


(COMPILE-FLAVOR-METHODS FILE-ERROR FILE-REQUEST-FAILURE FILE-OPERATION-FAILURE FILE-LOOKUP-ERROR
			DIRECTORY-NOT-FOUND-ERROR NO-MORE-ROOM-ERROR RENAME-FAILURE
			CHANGE-PROPERTY-FAILURE DELETE-FAILURE no-file-system)


;;; DEFSIGNALs moved here from CHAOS;QFILE-DEFINITIONS

(defsignal data-error file-request-failure (pathname operation)
   "Inconsistent data found in the file system.") 


(defsignal host-not-available file-request-failure (pathname operation)
   "A file system is refusing to listen to requests.") 


(defsignal no-file-system no-file-system (pathname operation)
   "The file system does not seem to exist.") 


(defsignal network-lossage file-request-failure (pathname operation)
   "Misc. network problems during file accessing.
In particular, failure to open data connection.") 



(defsignal not-enough-resources file-request-failure (pathname operation)
   "Shortage of resources (network or file server) to transmit operation.
This is if the file server itself says it hasn't got enough resources.") 


(defsignal unknown-operation file-request-failure (pathname operation)
   "OPERATION is not supported on the particular file system used.") 


(defsignal login-problems file-request-failure () "Some sort of failure to log in.") 


(defsignal unknown-user
   (file-request-failure login-problems correctable-login-problems unknown-user) ()
   "USER is not a known login-name at HOST.") 


(defsignal invalid-password
   (file-request-failure login-problems correctable-login-problems invalid-password) ()
   "PASSWORD is not USER's password at HOST.") 


(defsignal not-logged-in (file-request-failure not-logged-in) (pathname operation)
   "The file server said you were not logged in when the Lispm thought you should be.") 


(defsignal file-operation-failure-1 file-operation-failure (pathname operation)
   "Unrecognized file error codes signal this.") 


(defsignal file-open-for-output file-operation-failure (pathname operation)
   "How is this different from FILE-LOCKED?") 


(defsignal file-locked file-operation-failure (pathname operation)
   "File cannot be accessed because someone else is using it.") 


(defsignal circular-link file-operation-failure (pathname operation)
   "A link pointed to a link which pointed to a link ... too deeply.") 

;;UNIMPLEMENTED-OPTION


(defsignal unimplemented-option file-operation-failure (pathname operation)
   "Specified some option this file system doesn't understand.") 


(defsignal invalid-byte-size (file-operation-failure unimplemented-option invalid-byte-size)
   (pathname operation) "Specified a byte size that the file system cannot handle.") 


(defsignal inconsistent-options file-operation-failure (pathname operation)
   "Specified a byte size that the file system cannot handle.") 


(defsignal no-more-room no-more-room-error (pathname operation)
   "Out of resources within the file system.") 


(defsignal filepos-out-of-range file-operation-failure (pathname operation)
   "Access pointer set to bad value, outside 0 to length of file.") 


(defsignal not-available file-operation-failure (pathname operation)
   "File, pack etc. exists but is currently down.") 


(defsignal invalid-file-attribute error (pathname attribute value)
   "An attribute in the file attribute list had a bad value.
This is detected in the Lisp machine, not in the file server.") 

;;; Subclasses of FILE-OPERATION-FAILURE.

(defsignal file-not-found file-lookup-error (pathname operation)
   "The file was not found in the containing directory.") 


(defsignal multiple-file-not-found file-lookup-error (operation pathname pathnames)
   "None of the files was found in the containing directory.") 


(defsignal directory-not-found directory-not-found-error (pathname operation)
   "Containing directory is not found.") 


(defsignal device-not-found file-lookup-error (pathname operation)
   "Device specified in pathname not found.") 

(defsignal link-target-not-found file-lookup-error (pathname operation)
   "Failure in looking up the target of a link that was found.") 

;ACCESS-ERROR means some kind of protection failure.
(defsignal access-error file-operation-failure (pathname operation)
   "Some sort of protection screwed you.") 
		  

(defsignal incorrect-access-to-file
   (file-operation-failure access-error incorrect-access-to-file) (pathname operation)
   "File protection screwed you.") 


(defsignal incorrect-access-to-directory
   (file-operation-failure access-error incorrect-access-to-directory) (pathname operation)
   "Directory protection screwed you.") 

;;INVALID-PATHNAME-SYNTAX

(defsignal invalid-pathname-syntax file-operation-failure (pathname operation)
   "Some weird syntax got through pathname parsing but file system didn't like it.") 

(defsignal invalid-wildcard (file-operation-failure invalid-pathname-syntax invalid-wildcard)
   (pathname operation)
   "Wildcard that got through pathname parsing but file system didn't like it.") 


(defsignal wildcard-not-allowed
   (file-operation-failure invalid-pathname-syntax wildcard-not-allowed) (pathname operation)
   "Wildcard is ok but you did something to it that doesn't like wildcards.") 

;;WRONG-KIND-OF-FILE

(defsignal wrong-kind-of-file file-operation-failure (pathname operation)) 


(defsignal invalid-operation-for-link
   (file-operation-failure wrong-kind-of-file invalid-operation-for-link) (pathname operation)) 


(defsignal invalid-operation-for-directory
   (file-operation-failure wrong-kind-of-file invalid-operation-for-directory)
   (pathname operation)) 

;; CREATION-FAILUREs


(defsignal file-already-exists (file-operation-failure creation-failure file-already-exists)
   (pathname operation) "A file of this name already exists.") 


(defsignal superior-not-directory
   (file-operation-failure wrong-kind-of-file creation-failure superior-not-directory)
   (pathname operation) "The \"directory\" was not a directory.") 


(defsignal create-directory-failure
   (file-operation-failure creation-failure create-directory-failure) (pathname operation) "") 


(defsignal directory-already-exists
   (file-operation-failure creation-failure create-directory-failure file-already-exists
    directory-already-exists)
   (pathname operation) "A directory of this name already exists.") 


(defsignal create-link-failure (file-operation-failure creation-failure create-link-failure)
   (pathname operation) "Some problem creating a link.") 


;Should have property :NEW-PATHNAME.

(defsignal rename-to-existing-file rename-failure (pathname new-pathname)
   "NEW-PATHNAME already exists.") 

(defsignal rename-directory-version-not-1 rename-failure (pathname new-pathname) ;03.04.87 DAB
   "A directory file must have a version number of 1.")


(defsignal rename-across-directories rename-failure (pathname new-pathname)
   "PATHNAME and NEW-PATHNAME are on different directories.") 


(defsignal unknown-property change-property-failure (pathname property)
   "PROPERTY isn't one of the properties this file system handles.") 


(defsignal invalid-property-value change-property-failure (pathname property value)
   "VALUE is not a legal value for PROPERTY.") 


(defsignal invalid-property-name change-property-failure (pathname property)
   "Property name is syntactically bad.") 


(defsignal directory-not-empty delete-failure (pathname)
   "Deleting a directory that is not empty.") 


(defsignal dont-delete-flag-set delete-failure (pathname) "File's dont-delete flag was set.") 
