;;; -*- Mode:COMMON-LISP; Package:SYSTEM-INTERNALS; Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985-1989 Texas Instruments Incorporated. All rights reserved.
;;; 
;;;    Enhancements (C) Copyright 1981, Massachusetts Institute of Technology
;;; The Massachusetts Institute of Technology has acquired the rights from Symbolics
;;; to include the Software covered by the foregoing notice of copyright with its
;;; licenses of the Lisp Machine System **

;;; Flavor basis for streams

;;; To make a buffered stream:
;;; For input, define :NEXT-INPUT-BUFFER &OPTIONAL NO-HANG-P => ARRAY START END
;;;		      :DISCARD-INPUT-BUFFER ARRAY
;;; For output, define :NEW-OUTPUT-BUFFER => ARRAY START END
;;;		       :SEND-OUTPUT-BUFFER ARRAY END
;;;		       :DISCARD-OUTPUT-BUFFER ARRAY
;;; For IO, define both requires for input and output.

;;; To make :SET-POINTER work, define :SET-BUFFER-POINTER NEW-POINTER => REAL-NEW-POINTER
;;; where REAL-NEW-POINTER should be such that the next :NEXT-INPUT-BUFFER will access the
;;; desired position somewhere in the buffer.

;;; To make binary streams, mix with
;;; for input	SI:BUFFERED-INPUT-STREAM
;;; for output	SI:BUFFERED-OUTPUT-STREAM
;;; for IO      SI:BUFFERED-IO-STREAM
;;; for bidirectional SI:BUFFERED-STREAM
;;; Bidirectional vs IO. They both support input and output operations. Input and output operations work
;;;                      independently in bidirectional streams, resources are shared in IO streams. IO streams
;;;                      should be used with files.

;;; To make character streams, mix with
;;; for input	SI:BUFFERED-INPUT-CHARACTER-STREAM
;;; for output 	SI:BUFFERED-OUTPUT-CHARACTER-STREAM
;;; for IO 	SI:BUFFERED-IO-CHARACTER-STREAM
;;; for bidirectional SI:BUFFERED-CHARACTER-STREAM

;;; To make a character stream that does character set translation or compression or something
;;; like that, define a wrapper on :TYI (:TYO) that does the translation.  Then mix with
;;; for input	SI:BUFFERED-TYI-INPUT-STREAM
;;; for output	SI:BUFFERED-TYO-OUTPUT-STREAM
;;; for both	SI:BUFFERED-TYI-TYO-STREAM

;;; There are mixins in this file (SI:ASCII-TRANSLATING-INPUT-STREAM-MIXIN and
;;;	SI:ASCII-TRANSLATING-OUTPUT-STREAM-MIXIN) for translating between ASCII and Lisp
;;;	machine character set.

;;; For file streams, the mixin SI:STREAM-WITH-PATHNAME-MIXIN handles the pathname and
;;; printing nicely.  Use this for :PROBE openings (since they cannot do I/O).
;;; For input file streams, use SI:INPUT-FILE-STREAM-MIXIN.
;;; For output file streams, use SI:OUTPUT-FILE-STREAM-MIXIN.
;;; These make you interact with the who-line correctly.
;;; 2-2-87 (BASIC-BUFFERED-INPUT-STREAM :TYI). Changed ferror on EOF to use EOF as the error message if it is a string.
;;; 2-16-87 DAB output-pointer-remembering-mixin. Added changes to fix :set-pointer
;;; 03.05.87 DAB Modified TYI-from-ascii-stream to handle a nil charcater.
(PROCLAIM '(OPTIMIZE (SPEED 3.) (SAFETY 0.)))  
;;; Base flavors
;;; 12-13-88 DAB Fixed (INPUT-POINTER-REMEMBERING-MIXIN :SET-POINTER) to return a non-nil value if it successfully set the new pointer.[9079]
;;; 01-27-88 DAB Added the variable  *new-line-delimiter*. This will be used in all :line-in and :string-line-in to 
;;;              determine new line seperators.
;;; 07-23-87 DAB Added code to support IO streams.
;;; 04.08.87 DAB fixed (DEFMETHOD (OUTPUT-POINTER-REMEMBERING-MIXIN :BEFORE :SEND-CURRENT-OUTPUT-BUFFER). Causing lost 
;;;          of block if set-pointer in same block but greater than saved-stream-output-limit
;;; 07.15.87 DAB Fixed OUTPUT-POINTER-REMEMBERING-MIXIN :AFTER :SETUP-NEW-OUTPUT-BUFFER and :set-pointer to handle
;;;              yet another case.


;;;
(defvar *new-line-delimiter* '(#\newline))   ;Define the delimiters that may seperate lines. Some FTP implements
                                             ; will use the linefeed. This will allow FTP to redefine this variable to
                                             ; suit its application. 

(DEFFLAVOR STREAM () () (:REQUIRED-METHODS :DIRECTION)
   (:DOCUMENTATION :BASE-FLAVOR
    "All streams are built on this.
    This flavor is mostly for TYPEP, but also provides default methods for messages
    which all streams, input or output, are required to handle."))  



(DEFMETHOD (STREAM :CLOSE) (&OPTIONAL MODE)
  MODE;ignored
  ())  

;;; Streams are binary until proven otherwise


(DEFMETHOD (STREAM :CHARACTERS) ()
  ())  



(DEFMETHOD (STREAM :ELEMENT-TYPE) ()
  (LET ((VALUE (SEND SELF :SEND-IF-HANDLES :BYTE-SIZE)))
    (IF VALUE
      `(UNSIGNED-BYTE ,VALUE)
      'FIXNUM)))  



(DEFFLAVOR CHARACTER-STREAM () () (:INCLUDED-FLAVORS STREAM))  



(DEFMETHOD (CHARACTER-STREAM :CHARACTERS) ()
  T)  



(DEFMETHOD (CHARACTER-STREAM :ELEMENT-TYPE) ()
  'GLOBAL:CHARACTER)  



(DEFFLAVOR INPUT-STREAM () (STREAM) (:REQUIRED-METHODS :TYI :UNTYI)
   (:SELECT-METHOD-ORDER :TYI :UNTYI)
   (:DOCUMENTATION :BASE-FLAVOR "All input streams are built on this."))  



(DEFMETHOD (INPUT-STREAM :DIRECTION) ()
  :INPUT)  



(DEFMETHOD (INPUT-STREAM :LISTEN) ()
  (LET ((TEM (SEND SELF :TYI-NO-HANG ())))
    (COND
      (TEM (SEND SELF :UNTYI TEM) T))))  



(DEFMETHOD (INPUT-STREAM :TYIPEEK) (&OPTIONAL EOF)
  (LET ((TEM (SEND SELF :TYI EOF)))
    (COND
      (TEM (SEND SELF :UNTYI TEM) TEM))))  



(DEFMETHOD (INPUT-STREAM :TYI-NO-HANG) (&OPTIONAL EOF)
  (SEND SELF :TYI EOF))  



(DEFMETHOD (INPUT-STREAM :ANY-TYI) (&OPTIONAL EOF)
  (SEND SELF :TYI EOF))  



(DEFMETHOD (INPUT-STREAM :ANY-TYI-NO-HANG) (&OPTIONAL EOF)
  (SEND SELF :TYI EOF))  
  


(DEFMETHOD (INPUT-STREAM :CLEAR-INPUT) ()
  ())  



(DEFMETHOD (INPUT-STREAM :READ-UNTIL-EOF) ()
  (LOOP WHILE (SEND SELF :TYI)))  



(DEFMETHOD (INPUT-STREAM :STRING-IN) (EOF STRING &OPTIONAL (START 0.) END)
  (OR END (SETQ END (ARRAY-TOTAL-SIZE STRING)))
  (LOOP WHILE (< START END) AS CH = (SEND SELF :TYI) WHILE CH DO
     (SETF (AREF STRING (PROG1
			  START
			  (INCF START))) CH) FINALLY
     (AND (ARRAY-HAS-LEADER-P STRING) (STORE-ARRAY-LEADER START STRING 0.))
     (AND (NULL CH) EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF))
     (RETURN (VALUES START (NULL CH)))))  



(DEFFLAVOR OUTPUT-STREAM () (STREAM) (:REQUIRED-METHODS :TYO) :ABSTRACT-FLAVOR
   (:SELECT-METHOD-ORDER :TYO)
   (:DOCUMENTATION :BASE-FLAVOR "All output streams are built on this."))  



(DEFMETHOD (OUTPUT-STREAM :DIRECTION) ()
  :OUTPUT)  



(DEFMETHOD (OUTPUT-STREAM :FRESH-LINE) ()
  (SEND SELF :TYO #\NEWLINE)
  T)  



(DEFMETHOD (OUTPUT-STREAM :STRING-OUT) (STRING &OPTIONAL (START 0.) END)
  (OR END (SETQ END (ARRAY-ACTIVE-LENGTH STRING)))
  (DO ((I START (1+ I)))
      ((>= I END))
    (SEND SELF :TYO (AREF STRING I))))  



(DEFMETHOD (OUTPUT-STREAM :CLEAR-OUTPUT) ()
  ())  



(DEFMETHOD (OUTPUT-STREAM :FORCE-OUTPUT) ()
  ())  



(DEFMETHOD (OUTPUT-STREAM :FINISH) ()
  ())  



(DEFMETHOD (OUTPUT-STREAM :BEFORE :FINISH) ()
  (SEND SELF :FORCE-OUTPUT))  



(DEFMETHOD (OUTPUT-STREAM :EOF) ()
  (SEND SELF :FINISH))  



(DEFFLAVOR BIDIRECTIONAL-STREAM () () (:INCLUDED-FLAVORS INPUT-STREAM OUTPUT-STREAM))  



(DEFMETHOD (BIDIRECTIONAL-STREAM :DIRECTION) ()
  :BIDIRECTIONAL)  

;;; Buffered streams



(DEFFLAVOR BASIC-BUFFERED-INPUT-STREAM
   ((STREAM-INPUT-BUFFER NIL) (STREAM-INPUT-INDEX NIL) STREAM-INPUT-LIMIT) (INPUT-STREAM)
   :GETTABLE-INSTANCE-VARIABLES (:REQUIRED-METHODS :NEXT-INPUT-BUFFER :DISCARD-INPUT-BUFFER)
   (:DOCUMENTATION :COMBINATION
    "Input stream with a buffer.  Defines only a :TYI method.
Requires methods :NEXT-INPUT-BUFFER, which takes a no hang argument and returns three
values, an array, a starting index, and an ending index, or NIL at EOF or no input available
right away.  And :DISCARD-INPUT-BUFFER takes the array back and throws it away someplace."))  

;;; Returns non-NIL if any input was to be found.


(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :SETUP-NEXT-INPUT-BUFFER) (&OPTIONAL NO-HANG-P)
  (SEND SELF :DISCARD-CURRENT-INPUT-BUFFER)
  (MULTIPLE-VALUE-SETQ (STREAM-INPUT-BUFFER STREAM-INPUT-INDEX STREAM-INPUT-LIMIT)
    (SEND SELF :NEXT-INPUT-BUFFER NO-HANG-P)))  



(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :DISCARD-CURRENT-INPUT-BUFFER) ()
  (COND
    (STREAM-INPUT-BUFFER
     ;; Pretend that the index has reached the end, so that input remembering will work.
     (SETQ STREAM-INPUT-INDEX STREAM-INPUT-LIMIT)
     (SEND SELF :DISCARD-INPUT-BUFFER (PROG1
					STREAM-INPUT-BUFFER
					(SETQ STREAM-INPUT-BUFFER ()))))))  



(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :TYI) (&OPTIONAL EOF)
  (DO ()
      ((AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT))
       (PROG1
	 (CHAR-INT (AREF STREAM-INPUT-BUFFER STREAM-INPUT-INDEX))
	 (INCF STREAM-INPUT-INDEX)))
    (UNLESS (SEND SELF :SETUP-NEXT-INPUT-BUFFER ())
      (IF EOF
	(FERROR 'END-OF-FILE-1 (if (stringp eof) eof "End of file on ~S.") SELF)
	(RETURN NIL)))))


(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :READ-CHAR) (&OPTIONAL EOF-ERROR-P EOF-VALUE RECURSIVE-P )
  (DECLARE (IGNORE RECURSIVE-P))
  (DO ()
      ((AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT))
       (PROG1
	 (AREF STREAM-INPUT-BUFFER STREAM-INPUT-INDEX)
	 (INCF STREAM-INPUT-INDEX)))
    (UNLESS (SEND SELF :SETUP-NEXT-INPUT-BUFFER ())
      (IF EOF-ERROR-P
	  (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)
	  (RETURN EOF-VALUE)))))


(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :TYI-NO-HANG) (&OPTIONAL EOF)
  (LOOP UNTIL (AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT)) UNLESS
     (SEND SELF :SETUP-NEXT-INPUT-BUFFER T) RETURN
     (AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY
     (RETURN (PROG1
	       (CHAR-INT (AREF STREAM-INPUT-BUFFER STREAM-INPUT-INDEX))
	       (INCF STREAM-INPUT-INDEX)))))  



(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :TYIPEEK) (&OPTIONAL EOF)
  (LOOP UNTIL (AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT)) UNLESS
     (SEND SELF :SETUP-NEXT-INPUT-BUFFER ()) RETURN
     (AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY
     (RETURN (CHAR-INT (AREF STREAM-INPUT-BUFFER STREAM-INPUT-INDEX)))))


(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :UNTYI) (CH)
  (LET ((NEW-INDEX (AND STREAM-INPUT-BUFFER (1- STREAM-INPUT-INDEX))))
    (COND
      ((AND NEW-INDEX
	    (>= NEW-INDEX 0.)
	    ;; Put char-int arount ch here. *BJ*
	    (EQL (CHAR-INT (AREF STREAM-INPUT-BUFFER NEW-INDEX)) (char-int CH)))
       (SETQ STREAM-INPUT-INDEX NEW-INDEX))
      (T (FERROR () "Attempt to :UNTYI something different than last :TYI'ed.")))))  

(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :UNREAD-CHAR) (CH)
  (LET ((NEW-INDEX (AND STREAM-INPUT-BUFFER (1- STREAM-INPUT-INDEX))))
    (COND
      ;; Put char-int arount aref here. *BJ*
      ((AND NEW-INDEX (>= NEW-INDEX 0.) (EQL (INT-CHAR (AREF STREAM-INPUT-BUFFER NEW-INDEX)) CH))
       (SETQ STREAM-INPUT-INDEX NEW-INDEX))
      (T (FERROR () "Attempt to :UNTYI something different than last :TYI'ed.")))))

(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :LISTEN) ()
  (OR (AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT))
     (SEND SELF :SETUP-NEXT-INPUT-BUFFER T)))  



(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :READ-UNTIL-EOF) ()
  (LOOP WHILE (SEND SELF :SETUP-NEXT-INPUT-BUFFER)))  



(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :CLEAR-INPUT) ()
  (SEND SELF :DISCARD-CURRENT-INPUT-BUFFER))  



(DEFMETHOD (BASIC-BUFFERED-INPUT-STREAM :BEFORE :CLOSE) (&OPTIONAL IGNORE)
  (SEND SELF :DISCARD-CURRENT-INPUT-BUFFER))  



(DEFFLAVOR BUFFERED-INPUT-STREAM () (BASIC-BUFFERED-INPUT-STREAM)
   (:DOCUMENTATION :COMBINATION
    "Buffered input stream with :READ-INPUT-BUFFER.
Usef for all buffered input streams for which :TYI doesn't have wrappers to do
translation or such."))  



(DEFMETHOD (BUFFERED-INPUT-STREAM :GET-INPUT-BUFFER) (&OPTIONAL EOF)
  (LOOP UNTIL (AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT)) UNLESS
     (SEND SELF :SETUP-NEXT-INPUT-BUFFER) RETURN
     (AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY
     (RETURN
      (VALUES STREAM-INPUT-BUFFER STREAM-INPUT-INDEX (- STREAM-INPUT-LIMIT STREAM-INPUT-INDEX)))))  



(DEFMETHOD (BUFFERED-INPUT-STREAM :READ-INPUT-BUFFER) (&OPTIONAL EOF)
  (LOOP UNTIL (AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT)) UNLESS
     (SEND SELF :SETUP-NEXT-INPUT-BUFFER) RETURN
     (AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY
     (RETURN (VALUES STREAM-INPUT-BUFFER STREAM-INPUT-INDEX STREAM-INPUT-LIMIT))))  



(DEFMETHOD (BUFFERED-INPUT-STREAM :STRING-IN) (EOF STRING &OPTIONAL (START 0.) END)
  (OR END (SETQ END (ARRAY-TOTAL-SIZE STRING)))
  (LOOP WHILE (< START END) WHILE
     (LOOP UNTIL (AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT));Out of input, get some more
;Out of input, get some more
	UNLESS (SEND SELF :SETUP-NEXT-INPUT-BUFFER) RETURN
	(AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY (RETURN T))
     AS AMT = (MIN (- END START) (- STREAM-INPUT-LIMIT STREAM-INPUT-INDEX)) DO
     (COPY-ARRAY-PORTION STREAM-INPUT-BUFFER STREAM-INPUT-INDEX
			 (SETQ STREAM-INPUT-INDEX (+ STREAM-INPUT-INDEX AMT)) STRING START
			 (SETQ START (+ START AMT)))
     FINALLY (AND (ARRAY-HAS-LEADER-P STRING) (STORE-ARRAY-LEADER START STRING 0.))
     (RETURN (VALUES START (NULL STREAM-INPUT-BUFFER)))))  




(DEFMETHOD (BUFFERED-INPUT-STREAM :ADVANCE-INPUT-BUFFER) (&OPTIONAL NEW-INDEX)
  (COND
    (NEW-INDEX
     (OR (AND (>= NEW-INDEX 0.) (<= NEW-INDEX STREAM-INPUT-LIMIT))
	(FERROR () "New index out of range"))
     (SETQ STREAM-INPUT-INDEX NEW-INDEX))
    (T (SEND SELF :DISCARD-CURRENT-INPUT-BUFFER))))  



  

(DEFFLAVOR BASIC-BUFFERED-OUTPUT-STREAM
   ((STREAM-OUTPUT-BUFFER NIL) (STREAM-OUTPUT-INDEX NIL) STREAM-OUTPUT-LIMIT (Stream-output-byte-limit nil))
   (OUTPUT-STREAM)
   :GETTABLE-INSTANCE-VARIABLES
   (:REQUIRED-METHODS :NEW-OUTPUT-BUFFER :SEND-OUTPUT-BUFFER :DISCARD-OUTPUT-BUFFER)
   (:DOCUMENTATION :COMBINATION
    "Output stream with a buffer.  Only gives a :TYO method.
Required methods are :NEW-OUTPUT-BUFFER, which returns three values, an array, starting index,
and ending index into which characters can be stuffed and ,optionally, a fourth argument of the minimal number of characters
to write back.  And :SEND-OUTPUT-BUFFER takes the array and the ending output index reached or the minimal number of 
characters,whichever is greater, and transmit to the particular device.
:DISCARD-OUTPUT-BUFFER takes the array and should forget about sending the buffered data."))

  

(DEFMETHOD (BASIC-BUFFERED-OUTPUT-STREAM :SETUP-NEW-OUTPUT-BUFFER) ()
  (SEND SELF :SEND-CURRENT-OUTPUT-BUFFER)
   (MULTIPLE-VALUE-SETQ (STREAM-OUTPUT-BUFFER STREAM-OUTPUT-INDEX
			 STREAM-OUTPUT-LIMIT Stream-output-byte-limit) ;06.25.87 DAB
    (SEND SELF :NEW-OUTPUT-BUFFER)))

(DEFMETHOD (BASIC-BUFFERED-OUTPUT-STREAM :SEND-CURRENT-OUTPUT-BUFFER) ()
  (COND
    (STREAM-OUTPUT-BUFFER
     (SEND SELF
	   ;; If aborted out of write, prefer losing data to
	   ;; getting links circular.
	:SEND-OUTPUT-BUFFER (PROG1
			      STREAM-OUTPUT-BUFFER
			      (SETQ STREAM-OUTPUT-BUFFER ()))
	STREAM-OUTPUT-INDEX))))  

(DEFMETHOD (BASIC-BUFFERED-OUTPUT-STREAM :DISCARD-CURRENT-OUTPUT-BUFFER) ()
  (AND STREAM-OUTPUT-BUFFER
     (SEND SELF :DISCARD-OUTPUT-BUFFER
	(PROG1
	  STREAM-OUTPUT-BUFFER
	  (SETQ STREAM-OUTPUT-BUFFER ())))))  

(DEFMETHOD (BASIC-BUFFERED-outPUT-STREAM :CLEAR-outPUT) () 
  (SEND SELF :DISCARD-CURRENT-output-BUFFER))

(DEFMETHOD (BASIC-BUFFERED-OUTPUT-STREAM :TYO) (CH)
  (LOOP UNTIL (AND STREAM-OUTPUT-BUFFER (< STREAM-OUTPUT-INDEX STREAM-OUTPUT-LIMIT)) DO
     (SEND SELF :SETUP-NEW-OUTPUT-BUFFER)
     FINALLY
     (SETF (AREF STREAM-OUTPUT-BUFFER STREAM-OUTPUT-INDEX) CH) (INCF STREAM-OUTPUT-INDEX)))  



(DEFMETHOD (BASIC-BUFFERED-OUTPUT-STREAM :PAD-AND-SEND-BUFFER) (&OPTIONAL (PAD 0.))
  (COND
    (STREAM-OUTPUT-BUFFER
     (LOOP UNTIL (NOT (< STREAM-OUTPUT-INDEX STREAM-OUTPUT-LIMIT)) DO
	(PROGN
	  (SETF (AREF STREAM-OUTPUT-BUFFER STREAM-OUTPUT-INDEX) PAD)
	  (INCF STREAM-OUTPUT-INDEX))
	FINALLY (SEND SELF :SEND-CURRENT-OUTPUT-BUFFER)))))  




(DEFMETHOD (BASIC-BUFFERED-OUTPUT-STREAM :FORCE-OUTPUT) ()
  (SEND SELF :SEND-CURRENT-OUTPUT-BUFFER))  



(DEFMETHOD (BASIC-BUFFERED-OUTPUT-STREAM :BEFORE :CLOSE) (&OPTIONAL ABORTP)
  (SEND SELF (IF ABORTP
	       :DISCARD-CURRENT-OUTPUT-BUFFER
	       :SEND-CURRENT-OUTPUT-BUFFER)))  



(DEFMETHOD (BASIC-BUFFERED-OUTPUT-STREAM :LAST-CHAR-OUTPUT) ()
  (AND STREAM-OUTPUT-BUFFER (PLUSP STREAM-OUTPUT-INDEX)
     (AREF STREAM-OUTPUT-BUFFER (1- STREAM-OUTPUT-INDEX))))  



(DEFFLAVOR BUFFERED-OUTPUT-STREAM () (BUFFERED-OUTPUT-STREAM-MIXIN BASIC-BUFFERED-OUTPUT-STREAM)
   (:DOCUMENTATION :COMBINATION
    "Buffered output stream with :STRING-OUT.
Should be used for all output streams which do not have wrappers on :TYO to do translation
or such like."))  



(DEFFLAVOR BUFFERED-OUTPUT-STREAM-MIXIN () () (:REQUIRED-FLAVORS BASIC-BUFFERED-OUTPUT-STREAM))  



(DEFMETHOD (BUFFERED-OUTPUT-STREAM-MIXIN :STRING-OUT) (STRING &OPTIONAL (START 0.) END)
  (OR END (SETQ END (ARRAY-ACTIVE-LENGTH STRING)))
  (LOOP WHILE (< START END) UNLESS STREAM-OUTPUT-BUFFER DO (SEND SELF :SETUP-NEW-OUTPUT-BUFFER)
     AS AMT = (MIN (- END START) (- STREAM-OUTPUT-LIMIT STREAM-OUTPUT-INDEX)) DO
     (COPY-ARRAY-PORTION STRING START (SETQ START (+ START AMT)) STREAM-OUTPUT-BUFFER
			 STREAM-OUTPUT-INDEX
			 (SETQ STREAM-OUTPUT-INDEX (+ STREAM-OUTPUT-INDEX AMT)))
     WHEN (< START END) DO (SEND SELF :SEND-CURRENT-OUTPUT-BUFFER)))  

;; Since this stream has enough knowledge to execute this message intelligently, let it.
;; Note that this message can't go in BASIC-BUFFERED-OUTPUT-STREAM, because wrappers may
;; be defined on :TYO messages which would cause it to lose.


(DEFMETHOD (BUFFERED-OUTPUT-STREAM-MIXIN :FRESH-LINE) ()
  (IF (CHAR-EQUAL #\NEWLINE (OR (SEND SELF :LAST-CHAR-OUTPUT) 0.))
    ()
    (PROGN
      (SEND SELF :TYO #\NEWLINE)
      T)))  

;;; New code to support IO streams.  07-23-87 DAB

(DEFFLAVOR BASIC-BUFFERED-IO-STREAM-mixin
   ()  
   (BASIC-BUFFERED-OUTPUT-STREAM bidirectional-stream)
   (:required-flavors output-pointer-remembering-mixin)
   (:REQUIRED-METHODS :NEXT-INPUT-BUFFER :DISCARD-INPUT-BUFFER
		      :NEW-OUTPUT-BUFFER :SEND-OUTPUT-BUFFER :DISCARD-OUTPUT-BUFFER)
   (:DOCUMENTATION :COMBINATION
    "Input/output stream with a buffer.  Defines only a :TYI and a :TYO methods.
     Requires methods are :NEXT-INPUT-BUFFER and :NEW-OUTPUT-BUFFER. :NEXT-INPUT-BUFFER takes a no hang argument
     and returns three values, an array, a starting index, and an ending index, or NIL at EOF or no input available
     right away. And :DISCARD-INPUT-BUFFER takes the array back and throws it away someplace.
     :NEW-OUTPUT-BUFFER returns three values, an array, starting index and ending index into which characters
     can be stuffed and ,optionally, a fourth argument of the minimal number of characters to write back.  And
     :SEND-OUTPUT-BUFFER takes the array and the ending output index reached or the minimal number of characters,
     whichever is greater, and transmit to the particular device.
     :DISCARD-OUTPUT-BUFFER takes the array and should forget about sending the buffered data."))  


(defmethod (BASIC-BUFFERED-IO-STREAM-MIXIN :direction)
	   ()
  :IO)

(defmethod (BASIC-BUFFERED-IO-STREAM-MIXIN :check-buffer-eof)
	   ()
  ;;;THis method is necessary because-of the way new outptu buffers are generated. The EOF is more dynamic in
  ;;; IO stream. Returns T if the index is not at eof, otherwise nil.
  (cond ((not stream-output-buffer) nil)
	(Stream-output-byte-limit  (< STREAM-outPUT-INDEX Stream-output-byte-limit))
	(saved-STREAM-OUTPUT-INDEX (< STREAM-outPUT-INDEX (max saved-STREAM-OUTPUT-INDEX stream-output-index)))
	(t ())   ;09-10-87 DAB
	   ))



(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :SETUP-NEXT-INPUT-BUFFER) (&OPTIONAL NO-HANG-P)
  (when stream-output-buffer ;maybe the first time 
    (INCF OUTPUT-POINTER-BASE (- STREAM-OUTPUT-INDEX STREAM-OUTPUT-LOWER-LIMIT))
    (SETQ STREAM-OUTPUT-LOWER-LIMIT STREAM-OUTPUT-INDEX)
    (setq saved-relative-pointer nil)
    (setq saved-STREAM-OUTPUT-INDEX  nil)
    (send self :SEND-CURRENT-OUTPUT-BUFFER))
  (MULTIPLE-VALUE-SETQ (STREAM-outPUT-BUFFER STREAM-outPUT-INDEX STREAM-outPUT-LIMIT Stream-output-byte-limit)
    (SEND SELF :NEXT-INPUT-BUFFER NO-HANG-P))
  (AND STREAM-outPUT-BUFFER (SETQ STREAM-outPUT-LOWER-LIMIT STREAM-outPUT-INDEX))
  (setq saved-relative-pointer 0) ;07.15.87 DAB
  (setf saved-stream-output-index STREAM-outPUT-LIMIT)
  STREAM-outPUT-BUFFER ;return buffer
  )  


(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-MIXIN :DISCARD-CURRENT-INPUT-BUFFER) ()
  (when stream-output-buffer (send self :SEND-CURRENT-OUTPUT-BUFFER))  ;Save the output buffer first.
  (when
    STREAM-outPUT-BUFFER
     ;; Pretend that the index has reached the end, so that input remembering will work.
    (SETQ STREAM-outPUT-INDEX STREAM-outPUT-LIMIT)
    (SETQ STREAM-outPUT-BUFFER ()))
  )


(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :TYI) (&OPTIONAL EOF)
  (DO ()
      ((AND STREAM-outPUT-BUFFER (send self :check-buffer-eof))
       (PROG1
	 (CHAR-INT (AREF STREAM-outPUT-BUFFER STREAM-outPUT-INDEX))
	 (INCF STREAM-outPUT-INDEX)))
    (UNLESS (SEND SELF :SETUP-NEXT-INPUT-BUFFER ())
      (IF EOF
	(FERROR 'END-OF-FILE-1 (if (stringp eof) eof "End of file on ~S.") SELF)
	(RETURN NIL)))))


(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :READ-CHAR) (&OPTIONAL EOF-ERROR-P EOF-VALUE RECURSIVE-P )
  (DECLARE (IGNORE RECURSIVE-P))
  (DO ()
      ((AND STREAM-outPUT-BUFFER (send self :check-buffer-eof))
       (PROG1
	 (AREF STREAM-outPUT-BUFFER STREAM-outPUT-INDEX)
	 (INCF STREAM-outPUT-INDEX)))
    (UNLESS (SEND SELF :SETUP-NEXT-INPUT-BUFFER ())
      (IF EOF-ERROR-P
	  (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)
	  (RETURN EOF-VALUE)))))


(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :TYI-NO-HANG) (&OPTIONAL EOF)
  (LOOP UNTIL (AND STREAM-outPUT-BUFFER (send self :check-buffer-eof))
	UNLESS
	(SEND SELF :SETUP-NEXT-INPUT-BUFFER T) RETURN
	(AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY
	(RETURN (PROG1
		  (CHAR-INT (AREF STREAM-outPUT-BUFFER STREAM-outPUT-INDEX))
		  (INCF STREAM-outPUT-INDEX)))))  



(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :TYIPEEK) (&OPTIONAL EOF)
  (LOOP UNTIL (AND STREAM-outPUT-BUFFER (send self  :check-buffer-eof))
	UNLESS
	(SEND SELF :SETUP-NEXT-outPUT-BUFFER ()) RETURN
	(AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY
	(RETURN (CHAR-INT (AREF STREAM-outPUT-BUFFER STREAM-outPUT-INDEX)))))


(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :UNTYI) (CH)
  (LET ((NEW-INDEX (AND STREAM-outPUT-BUFFER (1- STREAM-outPUT-INDEX))))
    (COND
      ((AND NEW-INDEX
	    (>= NEW-INDEX 0.)
	    ;; Put char-int arount ch here. *BJ*
	    (EQL (CHAR-INT (AREF STREAM-outPUT-BUFFER NEW-INDEX)) (char-int CH)))
       (SETQ STREAM-outPUT-INDEX NEW-INDEX))
      (T (FERROR () "Attempt to :UNTYI something different than last :TYI'ed.")))))  

(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :UNREAD-CHAR) (CH)
  (LET ((NEW-INDEX (AND STREAM-outPUT-BUFFER (1- STREAM-outPUT-INDEX))))
    (COND
      ;; Put char-int arount aref here. *BJ*
      ((AND NEW-INDEX (>= NEW-INDEX 0.) (EQL (INT-CHAR (AREF STREAM-outPUT-BUFFER NEW-INDEX)) CH))
       (SETQ STREAM-outPUT-INDEX NEW-INDEX))
      (T (FERROR () "Attempt to :UNTYI something different than last :TYI'ed.")))))

(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :LISTEN) ()
  (OR (AND STREAM-outPUT-BUFFER (send self  :check-buffer-eof))
     (SEND SELF :SETUP-NEXT-INPUT-BUFFER T)))  



(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :READ-UNTIL-EOF) ()
  (LOOP WHILE (SEND SELF :SETUP-NEXT-INPUT-BUFFER)))  



(DEFMETHOD (BASIC-BUFFERED-IO-STREAM-mixin :CLEAR-INPUT) ()
  (SEND SELF :DISCARD-CURRENT-INPUT-BUFFER))  


(DEFFLAVOR BUFFERED-IO-STREAM ()
	   (BASIC-BUFFERED-IO-STREAM-mixin BUFFERED-OUTPUT-STREAM output-pointer-remembering-mixin)
   (:DOCUMENTATION :COMBINATION
    "Buffered IO stream with :READ-INPUT-BUFFER, :string-in and :string-out.
Used for all buffered IO streams for which :TYI and TYO doesn't have wrappers to do
translation or such."))


(DEFMETHOD (BUFFERED-IO-STREAM :GET-INPUT-BUFFER) (&OPTIONAL EOF)
  (LOOP UNTIL (AND STREAM-outPUT-BUFFER (send self :check-buffer-eof))
	UNLESS
	(SEND SELF :SETUP-NEXT-INPUT-BUFFER) RETURN
	(AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY
	(RETURN
	  (VALUES STREAM-outPUT-BUFFER STREAM-outPUT-INDEX (- STREAM-outPUT-LIMIT STREAM-outPUT-INDEX)))))  



(DEFMETHOD (BUFFERED-IO-STREAM :READ-INPUT-BUFFER) (&OPTIONAL EOF)
  (LOOP UNTIL (AND STREAM-outPUT-BUFFER  (send self   :check-buffer-eof))
	UNLESS
	(SEND SELF :SETUP-NEXT-INPUT-BUFFER) RETURN
	(AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY
	(RETURN (VALUES STREAM-outPUT-BUFFER STREAM-outPUT-INDEX STREAM-outPUT-LIMIT))))  



(DEFMETHOD (BUFFERED-IO-STREAM :STRING-IN) (EOF STRING &OPTIONAL (START 0.) END)
  (OR END (SETQ END (ARRAY-TOTAL-SIZE STRING)))
  (LOOP WHILE (< START END) WHILE
     (LOOP UNTIL (AND STREAM-outPUT-BUFFER (send self :check-buffer-eof));Out of input, get some more
	   ;Out of input, get some more
	UNLESS (SEND SELF :SETUP-NEXT-INPUT-BUFFER) RETURN
	(AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY (RETURN T))
     AS AMT = (MIN (- END START) (- (if Stream-output-byte-limit
					 Stream-output-byte-limit
					(if saved-STREAM-outPUT-index
						 (max STREAM-outPUT-index
						      saved-STREAM-outPUT-index)
						  STREAM-outPUT-index))
				    STREAM-outPUT-INDEX)) DO
     (COPY-ARRAY-PORTION STREAM-outPUT-BUFFER STREAM-outPUT-INDEX
			 (SETQ STREAM-outPUT-INDEX (+ STREAM-outPUT-INDEX AMT)) STRING START
			 (SETQ START (+ START AMT)))
     FINALLY (AND (ARRAY-HAS-LEADER-P STRING) (STORE-ARRAY-LEADER START STRING 0.))
     (RETURN (VALUES START (NULL STREAM-outPUT-BUFFER)))))  




(DEFMETHOD (BUFFERED-IO-STREAM :ADVANCE-INPUT-BUFFER) (&OPTIONAL NEW-INDEX)
  (COND
    (NEW-INDEX
     (OR (AND (>= NEW-INDEX 0.) (<= NEW-INDEX
				    (if Stream-output-byte-limit
					Stream-output-byte-limit
					(if saved-STREAM-outPUT-index
					    (max STREAM-outPUT-index
						 saved-STREAM-outPUT-index)
					    STREAM-outPUT-index))))
	 (FERROR () "New index out of range"))
     (SETQ STREAM-outPUT-INDEX NEW-INDEX))
    (T (SEND SELF :DISCARD-CURRENT-INPUT-BUFFER))))




; LINE input and output

; This comes in two different flavors, depending on whether the stream is buffered
; or unbuffered


(DEFFLAVOR UNBUFFERED-LINE-INPUT-STREAM 
	   ((readstring nil))
	   (CHARACTER-STREAM INPUT-STREAM)
  (:SETTABLE-INSTANCE-VARIABLES readstring)
  (:DOCUMENTATION :COMBINATION "Input stream with :LINE-IN but no buffering.
Used with input streams which only support :TYI."))


(DEFMETHOD (UNBUFFERED-LINE-INPUT-STREAM :after :CLOSE) (&OPTIONAL ignore)
  (when readstring (return-readstring readstring)	;MBC 8-13-86
	(setf readstring nil)))


(DEFMETHOD (UNBUFFERED-LINE-INPUT-STREAM :STRING-LINE-IN) (EOF STRING &OPTIONAL (START 0.) END)
  (declare (special *new-line-delimiter* ))  ;01-27-88 DAB
  (OR END (SETQ END (ARRAY-TOTAL-SIZE STRING)))
  (LOOP WHILE (< START END) AS CH = (SEND SELF :TYI)
	WHILE (AND CH (Not (member CH *new-line-delimiter* :test #'EQL))) DO
     (SETF (AREF STRING (PROG1
			  START
			  (INCF START))) CH) FINALLY
     (AND (ARRAY-HAS-LEADER-P STRING) (STORE-ARRAY-LEADER START STRING 0.))
     (AND (NULL CH) EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF))
     (fill string 0 :start start)
     (RETURN (VALUES START (NULL CH) (Not (member CH *new-line-delimiter* :test #'EQL))))))  


(DEFMETHOD (UNBUFFERED-LINE-INPUT-STREAM :LINE-IN) (&OPTIONAL LEADER)
  (declare (special *new-line-delimiter* ))  ;01-27-88 DAB
  (if readstring
      (setf (fill-pointer readstring) 0)    ;03-01-88 DAB
      (setf readstring (get-readstring)))
  (do ((ch (SEND SELF :TYI) (SEND SELF :TYI)))
      ((or (null ch) (member (int-char CH) *new-line-delimiter* :test #'EQ))  ;01-27-88  DAB
       (values
	 (cond ((fixnump leader)
		(make-array (length readstring)
			    :initial-contents readstring :leader-length leader))	; need an array with a leader.
	       ((null leader ) readstring)
	       (t (string-append readstring)))	; need a copy-of the array.
	 (null ch)))
    (vector-push-extend ch readstring)))


(DEFFLAVOR LINE-OUTPUT-STREAM-MIXIN () () (:REQUIRED-METHODS :STRING-OUT)
   (:INCLUDED-FLAVORS CHARACTER-STREAM OUTPUT-STREAM)
   (:DOCUMENTATION :MIXIN
    "Output stream with :LINE-OUT.
Used for buffered and unbuffered streams."))  



(DEFMETHOD (LINE-OUTPUT-STREAM-MIXIN :LINE-OUT) (LINE &OPTIONAL (START 0.) END)
  (SEND SELF :STRING-OUT LINE START END)
  (SEND SELF :TYO #\NEWLINE))  




(DEFFLAVOR BUFFERED-LINE-INPUT-STREAM
	   ((readstring nil))
	   (CHARACTER-STREAM BUFFERED-INPUT-STREAM)
  (:SETTABLE-INSTANCE-VARIABLES readstring)
  (:DOCUMENTATION :COMBINATION "Input stream with buffering and :LINE-IN."))  

(DEFMETHOD (BUFFERED-LINE-INPUT-STREAM :after :CLOSE) (&OPTIONAL ignore)
  (when readstring (return-readstring readstring)	;MBC 8-13-86
	(setf readstring nil)))


(DEFMETHOD (BUFFERED-LINE-INPUT-STREAM :STRING-LINE-IN) (EOF STRING &OPTIONAL (START 0.) END)
  (declare (special *new-line-delimiter* ))  ;01-27-88 DAB
  (OR END (SETQ END (ARRAY-TOTAL-SIZE STRING)))
  (LOOP WHILE (< START END) WHILE
     (LOOP UNTIL (AND STREAM-INPUT-BUFFER (< STREAM-INPUT-INDEX STREAM-INPUT-LIMIT));Out of input, get some more
;Out of input, get some more
	UNLESS (SEND SELF :SETUP-NEXT-INPUT-BUFFER) RETURN
	(AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY (RETURN T))
     AS LINE-END-INDEX =
     (STRING-SEARCH-SET *new-line-delimiter* STREAM-INPUT-BUFFER STREAM-INPUT-INDEX STREAM-INPUT-LIMIT) ;01-27-88 DAB
     AS AMT = (MIN (- END START) (- (OR LINE-END-INDEX STREAM-INPUT-LIMIT) STREAM-INPUT-INDEX))
     DO
     (COPY-ARRAY-PORTION STREAM-INPUT-BUFFER STREAM-INPUT-INDEX
			 (SETQ STREAM-INPUT-INDEX (+ STREAM-INPUT-INDEX AMT)) STRING START
			 (SETQ START (+ START AMT))
			 )
     (WHEN (EQ STREAM-INPUT-INDEX LINE-END-INDEX)
       (AND (ARRAY-HAS-LEADER-P STRING) (STORE-ARRAY-LEADER START STRING 0.))  ;03-01-88 DAB p7300
       (fill string 0 :start start)  ;03-01-88 DAB clear junk from end of array.
       (INCF STREAM-INPUT-INDEX)
       (RETURN (VALUES START () ())))
     FINALLY (AND (ARRAY-HAS-LEADER-P STRING)
		  (STORE-ARRAY-LEADER START STRING 0.)
		  (fill string 0 :start start))   ;03-01-88 DAB
     (RETURN (VALUES START (NULL STREAM-INPUT-BUFFER) T))))  



(DEFMETHOD (BUFFERED-LINE-INPUT-STREAM :LINE-IN) (&OPTIONAL LEADER)
  (declare (special *new-line-delimiter* ))  ;01-27-88 DAB
  (let ((String-Index 0.)
	(New-String-Index)
	(New-Buffer-Index)
	(CR-Index))
    (unless readstring
      (setf readstring (get-readstring)))
    (loop
      
      (unless (and  stream-input-buffer (< stream-input-index stream-input-limit))	;Out of input, get some more
	(unless (SEND SELF :SETUP-NEXT-INPUT-BUFFER)	;EOF if none that way
	  (setf (fill-pointer readstring) String-Index)
	  (Return (Values
		    (cond ((fixnump leader)
			   (make-array (length readstring)
				       :element-type 'string-char
				       :initial-contents readstring
				       :leader-length leader
				       :fill-pointer string-index))	; need a leader.
			  ((null leader ) readstring)
			  (t (string-append readstring)))
		    T))))
      
      (setf CR-Index
	    (STRING-SEARCH-SET *new-line-delimiter*    ;01-27-88 DAB
			       STREAM-INPUT-BUFFER STREAM-INPUT-INDEX STREAM-INPUT-LIMIT))
      
      (if CR-Index
	  (SETQ NEW-BUFFER-INDEX (1+ CR-INDEX)
		NEW-STRING-INDEX (+ STRING-INDEX (- CR-INDEX STREAM-INPUT-INDEX))) 
	  (SETQ NEW-BUFFER-INDEX STREAM-INPUT-LIMIT
		NEW-STRING-INDEX (+ STRING-INDEX (- STREAM-INPUT-LIMIT STREAM-INPUT-INDEX))))
      
      (When (and CR-Index (= String-Index 0) (fixnump leader))	;Saves one copy of the array within buffer boundary
	(return (values
		  (let ((read-string (make-array New-String-Index
					  :element-type 'string-char
					  :leader-length leader
					  :fill-pointer New-string-index)))	; need a leader.
		    (COPY-ARRAY-PORTION STREAM-INPUT-BUFFER STREAM-INPUT-INDEX NEW-BUFFER-INDEX
					Read-string 0 NEW-STRING-INDEX)
		    (setf (fill-pointer read-string) New-String-Index
			  STREAM-INPUT-INDEX NEW-BUFFER-INDEX)
		    read-string)
		  ())))
      
      (When (> new-string-index (array-total-size readstring))
	(ADJUST-ARRAY Readstring NEW-STRING-INDEX))
      
      (COPY-ARRAY-PORTION STREAM-INPUT-BUFFER STREAM-INPUT-INDEX NEW-BUFFER-INDEX
			  Readstring STRING-INDEX NEW-STRING-INDEX)
      
      (SETQ STREAM-INPUT-INDEX NEW-BUFFER-INDEX
	    String-INdex New-String-Index)
      
      (When CR-Index
	(setf (fill-pointer readstring) String-Index)
	(return (values
		  (cond ((fixnump leader)
			 (make-array (length readstring)
				     :element-type 'string-char
				     :initial-contents readstring
				     :leader-length leader
				     :fill-pointer String-Index))	; need an  a leader.
			((null leader ) readstring)
			(t (string-append readstring)))	; need a copy-of the array.
		  ()))))))


;;; Less basic stream flavors

(DEFFLAVOR ASCII-TRANSLATING-INPUT-STREAM-MIXIN (CRFF) () (:REQUIRED-METHODS :TYI)
   (:INCLUDED-FLAVORS INPUT-STREAM)
   (:SETTABLE-INSTANCE-VARIABLES CRFF)
   (:DOCUMENTATION :MIXIN
    "An input stream that translates characters from ASCII into lisp machine character set for :TYI method."))  



(DEFWRAPPER (ASCII-TRANSLATING-INPUT-STREAM-MIXIN :TYI) (IGNORE . BODY)
   `(PROGN
      .DAEMON-CALLER-ARGS.
      (TYI-FROM-ASCII-STREAM
       #'(LAMBDA (&REST .DAEMON-CALLER-ARGS. &AUX (.DAEMON-MAPPING-TABLE. SELF-MAPPING-TABLE))
	   .DAEMON-MAPPING-TABLE.
	   ,@BODY))))  

(DEFUN-method TYI-FROM-ASCII-STREAM ASCII-TRANSLATING-INPUT-STREAM-MIXIN (ASCII-STREAM &AUX CH) ;03.05.87 DAB
  (setf CRFF nil)
    (CASE (SETQ CH (FUNCALL ASCII-STREAM :TYI))
	  (8. #.(char-int #\BACKSPACE))
	  (9. #.(char-int #\TAB))
	  (10. #.(char-int #\LINEFEED))
	  (12. #.(char-int #\PAGE))
	  (13.
	   (LET ((CH1 (FUNCALL ASCII-STREAM :TYI)))
	     (OR (and (= CH1 10.) (setf CRFF t)) (FUNCALL ASCII-STREAM :UNTYI CH1)))
	   #.(char-int #\NEWLINE))
	  (127. #.(char-int #\RUBOUT))
	  (nil nil)
	  (T (char-int CH))))


  

(DEFWRAPPER (ASCII-TRANSLATING-INPUT-STREAM-MIXIN :UNTYI) (ch . BODY)
   `(PROGN
      .DAEMON-CALLER-ARGS.
      (UNTYI-FROM-ASCII-STREAM
	ch
       #'(LAMBDA (&REST .DAEMON-CALLER-ARGS. &AUX (.DAEMON-MAPPING-TABLE. SELF-MAPPING-TABLE))
	   .DAEMON-MAPPING-TABLE.
	   ,@BODY))))

(DEFun-method UNTYI-FROM-ASCII-STREAM ASCII-TRANSLATING-INPUT-STREAM-MIXIN (CH ASCII-STREAM)
    (setf ch (case (car ch)
	     (136. (int-char 8.))   ;#\backspace 
	     (137. (int-char 9.))   ;#\TAB
	     (138. (int-char 10.))  ;#\LINEFEED
	     (140. (int-char 12.))  ;#\PAGE
	     (141. (prog1 (int-char 13.)
			  (when CRFF (FUNCALL ASCII-STREAM :UNTYI (int-char 10.)))))  ;#\NEWLINE
	     (135. (int-char 127.)) ;#\RUBOUT
	     (t (int-char (car ch)))))
  (FUNCALL ASCII-STREAM :UNTYI CH)
  )


(DEFFLAVOR ASCII-TRANSLATING-OUTPUT-STREAM-MIXIN () () (:REQUIRED-METHODS :TYO)
   (:INCLUDED-FLAVORS OUTPUT-STREAM)
   (:DOCUMENTATION :MIXIN
    "An output stream that translates characters from lisp machine character set into ASCII for :TYO method."))  



(DEFWRAPPER (ASCII-TRANSLATING-OUTPUT-STREAM-MIXIN :TYO) ((CH) . BODY)
   `(TYO-TO-ASCII-STREAM
     #'(LAMBDA (&REST .DAEMON-CALLER-ARGS. &AUX (.DAEMON-MAPPING-TABLE. SELF-MAPPING-TABLE))
	 .DAEMON-MAPPING-TABLE.
	 ,@BODY)
     CH))  



(DEFUN TYO-TO-ASCII-STREAM (ASCII-STREAM CH)
  (FUNCALL ASCII-STREAM :TYO
	   (CASE (CHAR-CODE CH);rla - fix so that either fixnums (e.g. from write-line) or
	    ;      characters (e.g. from format) will match
	     
	     (#.(char-int #\BACKSPACE) 8.)
	     (#.(char-int #\TAB) 9.)
	     (#.(char-int #\LINEFEED) 10.)
	     (#.(char-int #\PAGE) 12.)
	     (#.(char-int #\NEWLINE) (FUNCALL ASCII-STREAM :TYO 13.) 10.)
	     (#.(char-int #\RUBOUT) 127.)
	     (T CH))))  



(DEFFLAVOR INPUT-POINTER-REMEMBERING-MIXIN
   ((INPUT-POINTER-BASE 0.) (STREAM-INPUT-LOWER-LIMIT 0.)) ()
   (:INCLUDED-FLAVORS BASIC-BUFFERED-INPUT-STREAM)
   ;; :SET-BUFFER-POINTER is sent when a :SET-POINTER request goes beyond the current buffer.
   ;; It should return the real position set and arrange for the next :NEXT-INPUT-BUFFER
   ;; to contain the desired position in it someplace.
   (:REQUIRED-METHODS :SET-BUFFER-POINTER)
   (:DOCUMENTATION :MIXIN "Buffered input stream with :SET-POINTER and :READ-POINTER methods."))  



(DEFMETHOD (INPUT-POINTER-REMEMBERING-MIXIN :SET-BUFFER-POINTER) (NEW-POINTER)
  (FERROR () "Cannot set pointer on ~S to ~D" SELF NEW-POINTER))  

;;; Obsolete shorthand message, but in the manual, so keep for a while.


(DEFMETHOD (INPUT-POINTER-REMEMBERING-MIXIN :REWIND) ()
  (SEND SELF :SET-POINTER 0.))  



(DEFMETHOD (INPUT-POINTER-REMEMBERING-MIXIN :SET-POINTER) (NEW-POINTER)
  (LOOP AS NEW-RELATIVE-POINTER =
     (+ (- NEW-POINTER INPUT-POINTER-BASE) STREAM-INPUT-LOWER-LIMIT) WHEN
     (AND STREAM-INPUT-INDEX (= NEW-RELATIVE-POINTER STREAM-INPUT-INDEX)) RETURN T UNTIL
     (IF STREAM-INPUT-BUFFER
	 (AND (>= NEW-RELATIVE-POINTER STREAM-INPUT-LOWER-LIMIT)
	      (< NEW-RELATIVE-POINTER STREAM-INPUT-LIMIT))
	 (= NEW-RELATIVE-POINTER STREAM-INPUT-LOWER-LIMIT))
     DO (SEND SELF :DISCARD-CURRENT-INPUT-BUFFER)
     (SETQ INPUT-POINTER-BASE (SEND SELF :SET-BUFFER-POINTER NEW-POINTER))
     (SEND SELF :SETUP-NEXT-INPUT-BUFFER)
     FINALLY (return (SETQ STREAM-INPUT-INDEX NEW-RELATIVE-POINTER)) ;12-06-88 DAB return non-nil if successful.
     ))  



(DEFMETHOD (INPUT-POINTER-REMEMBERING-MIXIN :BEFORE :DISCARD-INPUT-BUFFER) (IGNORE)
  (INCF INPUT-POINTER-BASE (- STREAM-INPUT-INDEX STREAM-INPUT-LOWER-LIMIT))
  (SETQ STREAM-INPUT-LOWER-LIMIT STREAM-INPUT-INDEX))  



(DEFMETHOD (INPUT-POINTER-REMEMBERING-MIXIN :AFTER :SETUP-NEXT-INPUT-BUFFER) (&OPTIONAL IGNORE)
  (AND STREAM-INPUT-BUFFER (SETQ STREAM-INPUT-LOWER-LIMIT STREAM-INPUT-INDEX)))  



(DEFMETHOD (INPUT-POINTER-REMEMBERING-MIXIN :READ-POINTER) ()
  (+ INPUT-POINTER-BASE
     (IF STREAM-INPUT-INDEX
       (- STREAM-INPUT-INDEX STREAM-INPUT-LOWER-LIMIT)
       0.)))  



(DEFFLAVOR OUTPUT-POINTER-REMEMBERING-MIXIN
	((OUTPUT-POINTER-BASE 0)
	 (STREAM-OUTPUT-LOWER-LIMIT 0)
	 (saved-relative-pointer nil)
	 (saved-stream-output-index nil))
	()
  (:SETTABLE-INSTANCE-VARIABLES OUTPUT-POINTER-BASE)
  (:INCLUDED-FLAVORS BASIC-BUFFERED-OUTPUT-STREAM)
  (:DOCUMENTATION :MIXIN "Buffered output stream with :READ-POINTER method."))  



(DEFMETHOD (OUTPUT-POINTER-REMEMBERING-MIXIN :BEFORE :SEND-OUTPUT-BUFFER) (IGNORE IGNORE)
  (INCF OUTPUT-POINTER-BASE (- STREAM-OUTPUT-INDEX STREAM-OUTPUT-LOWER-LIMIT))
  (SETQ STREAM-OUTPUT-LOWER-LIMIT STREAM-OUTPUT-INDEX)
  (setq saved-relative-pointer nil)
  (setq saved-STREAM-OUTPUT-INDEX  nil))  

(DEFMETHOD (OUTPUT-POINTER-REMEMBERING-MIXIN :AFTER :SETUP-NEW-OUTPUT-BUFFER) ()
  (SETQ STREAM-OUTPUT-LOWER-LIMIT STREAM-OUTPUT-INDEX)
  (setq saved-relative-pointer 0) ;07.15.87 DAB
  (when Stream-output-byte-limit (setf saved-stream-output-index Stream-output-byte-limit)))


(DEFMETHOD (OUTPUT-POINTER-REMEMBERING-MIXIN :READ-POINTER) ()
  (+ OUTPUT-POINTER-BASE
     (IF STREAM-OUTPUT-INDEX
       (- STREAM-OUTPUT-INDEX STREAM-OUTPUT-LOWER-LIMIT)
       0.)))  

;We send a :GET-OLD-DATA (args are buffer and ending index) if it is
;necessary to fill the beginning of the output buffer with the current
;contents of the file in order to avoid clobbering data that we are
;supposed to be skipping over.


(DEFMETHOD (OUTPUT-POINTER-REMEMBERING-MIXIN :SET-POINTER) (NEW-POINTER)
  (DO ((BUFFER-RELOADED NIL T))
      (())
    (LET ((allocated-new-space nil) 
	  (NEW-RELATIVE-POINTER (+ (- NEW-POINTER OUTPUT-POINTER-BASE)
				   STREAM-OUTPUT-LOWER-LIMIT)))
      ;; Does the buffer we have now include the desired pointer?
      (IF (IF STREAM-OUTPUT-BUFFER
	      (AND (>= NEW-RELATIVE-POINTER STREAM-OUTPUT-LOWER-LIMIT)
		   (< NEW-RELATIVE-POINTER STREAM-OUTPUT-LIMIT))
	      (= NEW-RELATIVE-POINTER STREAM-OUTPUT-LOWER-LIMIT))
	  (PROGN
	    (if saved-STREAM-OUTPUT-INDEX	    	;last pointer in this buffered allocation.
		(and (> STREAM-OUTPUT-INDEX saved-STREAM-OUTPUT-INDEX)
		     (not (eql saved-relative-pointer stream-output-index)) ;07.15.87 DAB
		     (SETQ saved-STREAM-OUTPUT-INDEX STREAM-OUTPUT-INDEX))
		(unless (or (null saved-relative-pointer)
			    (eql saved-relative-pointer stream-output-index))
		       (SETQ saved-STREAM-OUTPUT-INDEX STREAM-OUTPUT-INDEX)))
	    (SETQ STREAM-OUTPUT-INDEX NEW-RELATIVE-POINTER)
	    (setq saved-relative-pointer new-relative-pointer)
	    (AND BUFFER-RELOADED
		 (send SELF :GET-OLD-DATA
		       STREAM-OUTPUT-BUFFER STREAM-OUTPUT-LOWER-LIMIT))
	    (RETURN (+ OUTPUT-POINTER-BASE
		       (IF STREAM-OUTPUT-INDEX
			   (- STREAM-OUTPUT-INDEX STREAM-OUTPUT-LOWER-LIMIT)
			   0)))))
      ;; No, get another buffer after specifying where in the file we want.
      (send SELF :SEND-CURRENT-OUTPUT-BUFFER)
      (multiple-value-SETQ (OUTPUT-POINTER-BASE allocated-new-space)
	    (send SELF :SET-BUFFER-POINTER NEW-POINTER))
      (send SELF :SETUP-NEW-OUTPUT-BUFFER)
      (when allocated-new-space
	  (setf saved-STREAM-OUTPUT-INDEX  nil)
	  (setf Stream-output-byte-limit  nil)
	  ))))


(DEFMETHOD (OUTPUT-POINTER-REMEMBERING-MIXIN :BEFORE :SEND-CURRENT-OUTPUT-BUFFER) (&rest IGNORE)
  (cond ((and saved-relative-pointer
	      (eql saved-relative-pointer stream-output-index))
	 (if saved-stream-output-index
	     (setf stream-output-index saved-stream-output-index)
	     (setf stream-output-index 0)))
	((and saved-stream-output-index   ;04.08.87 DAB
	      (> saved-stream-output-index 0)
	      (> saved-stream-output-index stream-output-index))
	 (setf stream-output-index saved-stream-output-index))
	(t ())))


(DEFMETHOD (OUTPUT-POINTER-REMEMBERING-MIXIN :DEFAULT :GET-OLD-DATA) (BUFFER-ARRAY OUTPUT-LOWER-LIMIT)
  BUFFER-ARRAY
  OUTPUT-LOWER-LIMIT
  ())    

;;; Some useful combinations


;; New version - LSS -  1/18/88 - for Symbolics font support
(DEFFLAVOR BUFFERED-INPUT-CHARACTER-STREAM
	   ((symbolics-fat-mode nil)  ;; use symbolics fonts - LSS 
	    (symbolics-thin-mode nil)) ;; don't convert epsilons - LSS
   (INPUT-POINTER-REMEMBERING-MIXIN BUFFERED-LINE-INPUT-STREAM)
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES
  (:DOCUMENTATION :COMBINATION "A buffered input character stream, gives :LINE-IN."))


;; Old version
;(DEFFLAVOR BUFFERED-INPUT-CHARACTER-STREAM ()
;   (INPUT-POINTER-REMEMBERING-MIXIN BUFFERED-LINE-INPUT-STREAM)
;   (:DOCUMENTATION :COMBINATION "A buffered input character stream, gives :LINE-IN."))  



(DEFFLAVOR BUFFERED-OUTPUT-CHARACTER-STREAM ()
   (LINE-OUTPUT-STREAM-MIXIN CHARACTER-STREAM BUFFERED-OUTPUT-STREAM)
   (:DOCUMENTATION :COMBINATION "A buffered output character stream, gives :LINE-OUT."))  



(DEFFLAVOR BUFFERED-CHARACTER-STREAM ()
   (BIDIRECTIONAL-STREAM INPUT-POINTER-REMEMBERING-MIXIN BUFFERED-LINE-INPUT-STREAM
    LINE-OUTPUT-STREAM-MIXIN BUFFERED-OUTPUT-STREAM)
   (:DOCUMENTATION :COMBINATION "A bidrection character stream, :LINE-IN and :LINE-OUT."))  



(DEFFLAVOR BUFFERED-STREAM ()
   (BIDIRECTIONAL-STREAM BUFFERED-INPUT-STREAM BUFFERED-OUTPUT-STREAM)
   (:DOCUMENTATION :COMBINATION "A bidirection buffered stream."))  



(DEFFLAVOR BUFFERED-TYI-INPUT-STREAM ()
   (INPUT-POINTER-REMEMBERING-MIXIN UNBUFFERED-LINE-INPUT-STREAM BASIC-BUFFERED-INPUT-STREAM)
   (:DOCUMENTATION :COMBINATION "A buffered character input stream for use with :TYI wrappers."))  



(DEFFLAVOR BUFFERED-TYO-OUTPUT-STREAM ()
   (LINE-OUTPUT-STREAM-MIXIN CHARACTER-STREAM BASIC-BUFFERED-OUTPUT-STREAM)
   (:DOCUMENTATION :COMBINATION "A buffered character output stream for use with :TYO wrappers."))  



(DEFFLAVOR BUFFERED-TYI-TYO-STREAM ()
   (BIDIRECTIONAL-STREAM INPUT-POINTER-REMEMBERING-MIXIN UNBUFFERED-LINE-INPUT-STREAM
    BASIC-BUFFERED-INPUT-STREAM LINE-OUTPUT-STREAM-MIXIN BASIC-BUFFERED-OUTPUT-STREAM)
   (:DOCUMENTATION :COMBINATION
    "A bidirectional buffered character stream, for use with :TYI and :TYO wrappers."))  

;;;new code to support IO streams. 07-23-87 DAB

(DEFFLAVOR BUFFERED-IO-character-STREAM ((readstring nil))
   (BUFFERED-IO-STREAM
    BUFFERED-OUTPUT-CHARACTER-STREAM
    OUTPUT-POINTER-REMEMBERING-MIXIN 
   )
   (:DOCUMENTATION :COMBINATION "A IO character stream, :LINE-IN and :LINE-OUT.")
   )

(DEFMETHOD (BUFFERED-IO-character-STREAM  :LENGTH) ()
  ())

(DEFMETHOD (BUFFERED-IO-character-STREAM :after :CLOSE) (&OPTIONAL ignore)
  (when readstring (return-readstring readstring)	;MBC 8-13-86
	(setf readstring nil)))

(DEFMETHOD (BUFFERED-IO-character-STREAM :STRING-LINE-IN) (EOF STRING &OPTIONAL (START 0.) END)
  (declare (special *new-line-delimiter* )) ;01-27-88 DAB
  (OR END (SETQ END (ARRAY-TOTAL-SIZE STRING)))
     (LOOP WHILE (< START END) WHILE
	  (LOOP UNTIL (AND STREAM-outPUT-BUFFER (send self :check-buffer-eof))	;Out of input, get some more
;Out of input, get some more
		UNLESS (SEND SELF :SETUP-NEXT-INPUT-BUFFER) RETURN
		(AND EOF (FERROR 'END-OF-FILE-1 "End of file on ~S." SELF)) FINALLY (RETURN T))
	  as real-buffer-limit = (if Stream-output-byte-limit
				   Stream-output-byte-limit
				   (if saved-STREAM-outPUT-index
				       (max STREAM-outPUT-index
					    saved-STREAM-outPUT-index)
				       STREAM-outPUT-index))
	  AS LINE-END-INDEX =
	  (STRING-SEARCH-SET *new-line-delimiter*    ;01-27-88  DAB
			     STREAM-outPUT-BUFFER STREAM-outPUT-INDEX real-buffer-limit)
	  AS AMT = (MIN (- END START) (- (OR LINE-END-INDEX real-buffer-limit)
					 STREAM-outPUT-INDEX))
	  DO
	  (COPY-ARRAY-PORTION STREAM-outPUT-BUFFER STREAM-outPUT-INDEX
			      (SETQ STREAM-outPUT-INDEX (+ STREAM-outPUT-INDEX AMT)) STRING START
			      (SETQ START (+ START AMT)))
	  (WHEN (EQ STREAM-outPUT-INDEX LINE-END-INDEX)
	    (INCF STREAM-outPUT-INDEX)
	    (RETURN (VALUES START () ())))
	  FINALLY (AND (ARRAY-HAS-LEADER-P STRING) (STORE-ARRAY-LEADER START STRING 0.))
	  (RETURN (VALUES START (NULL STREAM-outPUT-BUFFER) T))))


(DEFMETHOD (BUFFERED-IO-character-STREAM :LINE-IN) (&OPTIONAL LEADER)
  (declare (special *new-line-delimiter* )) ;01-27-88 DAB
  (let ((String-Index 0.)
	(New-String-Index)
	(New-Buffer-Index)
	(CR-Index)
	(real-buffer-limit))
    (unless readstring
      (setf readstring (get-readstring)))
    (loop
      (unless (and  stream-output-buffer (send self :check-buffer-eof))	;Out of input, get some more
	(unless (SEND SELF :SETUP-NEXT-INPUT-BUFFER)	;EOF if none that way
	  (setf (fill-pointer readstring) String-Index)
	  (Return (Values
		    (cond ((fixnump leader)
			   (make-array (length readstring)
				       :element-type 'string-char
				       :initial-contents readstring
				       :leader-length leader
				       :fill-pointer string-index))	; need a leader.
			  ((null leader ) readstring)
			  (t (string-append readstring)))
		    T))))
      (setq real-buffer-limit (if Stream-output-byte-limit
					      Stream-output-byte-limit
					      (if saved-STREAM-outPUT-index
						  (max STREAM-outPUT-index
						       saved-STREAM-outPUT-index)
						  STREAM-outPUT-index)))
      (setf CR-Index (STRING-SEARCH-SET *new-line-delimiter*   ;01-27-88 DAB
					STREAM-outPUT-BUFFER
					STREAM-outPUT-INDEX
					real-buffer-limit))
      
      (if CR-Index
	  (SETQ NEW-BUFFER-INDEX (1+ CR-INDEX)
		NEW-STRING-INDEX (+ STRING-INDEX (- CR-INDEX STREAM-outPUT-index))) 
	  (SETQ NEW-BUFFER-INDEX real-buffer-limit
		NEW-STRING-INDEX (+ STRING-INDEX (- real-buffer-limit STREAM-outPUT-INDEX))))
      
      (When (and CR-Index (= String-Index 0) (fixnump leader))	;Saves one copy of the array within buffer boundary
	(return (values
		  (let ((read-string (make-array New-String-Index
					  :element-type 'string-char
					  :leader-length leader
					  :fill-pointer New-string-index)))	; need a leader.
		    (COPY-ARRAY-PORTION STREAM-outPUT-BUFFER STREAM-outPUT-INDEX NEW-BUFFER-INDEX
					Read-string 0 NEW-STRING-INDEX)
		    (setf (fill-pointer read-string) New-String-Index
			  STREAM-outPUT-INDEX NEW-BUFFER-INDEX)
		    read-string)
		  ())))
      
      (When (> new-string-index (array-total-size readstring))
	(ADJUST-ARRAY Readstring NEW-STRING-INDEX))
      
      (COPY-ARRAY-PORTION STREAM-outPUT-BUFFER STREAM-outPUT-INDEX NEW-BUFFER-INDEX
			  Readstring STRING-INDEX NEW-STRING-INDEX)
      
      (SETQ STREAM-outPUT-INDEX NEW-BUFFER-INDEX
	    String-INdex New-String-Index)
      
      (When CR-Index
	(setf (fill-pointer readstring) String-Index)
	(return (values
		  (cond ((fixnump leader)
			 (make-array (length readstring)
				     :element-type 'string-char
				     :initial-contents readstring
				     :leader-length leader
				     :fill-pointer String-Index))	; need an  a leader.
			((null leader ) readstring)
			(t (string-append readstring)))	; need a copy-of the array.
		  ()))))))


(DEFMETHOD (BUFFERED-IO-character-STREAM  :REWIND) ()
  (SEND SELF :SET-POINTER 0.))  



;;; Flavors for use with file computers

;;; For use with :PROBE OPEN calls


(DEFFLAVOR FILE-STREAM-MIXIN (PATHNAME) () (:INCLUDED-FLAVORS STREAM)
   (:REQUIRED-METHODS :TRUENAME :PLIST) (:INITABLE-INSTANCE-VARIABLES PATHNAME)
   (:GETTABLE-INSTANCE-VARIABLES PATHNAME)
   (:DOCUMENTATION :MIXIN "Streams for use with file computers, as returned by OPEN."))  



(DEFMETHOD (FILE-STREAM-MIXIN :PRINT-SELF) (STREAM IGNORE IGNORE)
  (PRINTING-RANDOM-OBJECT (SELF STREAM :TYPEP) (PRIN1 (STRING PATHNAME) STREAM)))  



(DEFMETHOD (FILE-STREAM-MIXIN :GET) (INDICATOR)
  (LET ((PLIST (SEND SELF :PLIST)))
    (GETF PLIST INDICATOR)))  



(DEFMETHOD (FILE-STREAM-MIXIN :CREATION-DATE) ()
  (SEND SELF :GET :CREATION-DATE))  



(DEFMETHOD (FILE-STREAM-MIXIN :INFO) ()
  (CONS (SEND SELF :TRUENAME) (SEND SELF :CREATION-DATE)))  



(DEFMETHOD (FILE-STREAM-MIXIN :GENERIC-PATHNAME) ()
  (FUNCALL PATHNAME :GENERIC-PATHNAME))  

;;; For use with :READ and :WRITE OPEN calls


(DEFFLAVOR FILE-DATA-STREAM-MIXIN () (FILE-STREAM-MIXIN) (:INCLUDED-FLAVORS STREAM)
   (:REQUIRED-METHODS :READ-POINTER :LENGTH :QFASLP)
   (:DOCUMENTATION :MIXIN "Streams which can actually do file IO."))  



(DEFMETHOD (FILE-DATA-STREAM-MIXIN :AFTER :INIT) (IGNORE)
  (FUNCALL TV:WHO-LINE-FILE-STATE-SHEET :ADD-STREAM SELF))  



(DEFMETHOD (FILE-DATA-STREAM-MIXIN :AFTER :CLOSE) (&OPTIONAL IGNORE)
  (FUNCALL TV:WHO-LINE-FILE-STATE-SHEET :DELETE-STREAM SELF))  



(DEFMETHOD (FILE-DATA-STREAM-MIXIN :WHO-LINE-INFORMATION) (&AUX COUNT LENGTH PERCENT DIRECTION)
  (SETQ COUNT (SEND SELF :READ-POINTER))
  (SETQ DIRECTION (SEND SELF :DIRECTION))
  (AND (SETQ LENGTH (SEND SELF :LENGTH)) (NOT (ZEROP LENGTH))
     (SETQ PERCENT (ROUND (* 100. COUNT) LENGTH)))
  (LET ((MAYBE-NAME (SEND SELF :TRUENAME)))
	;; directory streams (and maybe others) don't have truenames; show
	;; their pathnames instead.
    (IF (NULL MAYBE-NAME)
      (SETQ MAYBE-NAME (SEND SELF :PATHNAME)))
    (VALUES MAYBE-NAME DIRECTION COUNT PERCENT)))  



(DEFFLAVOR INPUT-FILE-STREAM-MIXIN () (FILE-DATA-STREAM-MIXIN)
   (:INCLUDED-FLAVORS INPUT-POINTER-REMEMBERING-MIXIN)
   (:DOCUMENTATION :MIXIN "Streams for use with input files."))  



(DEFFLAVOR OUTPUT-FILE-STREAM-MIXIN () (FILE-DATA-STREAM-MIXIN)
   (:INCLUDED-FLAVORS OUTPUT-POINTER-REMEMBERING-MIXIN)
   (:DOCUMENTATION :MIXIN "Streams for use with output files."))  



(DEFMETHOD (OUTPUT-FILE-STREAM-MIXIN :LENGTH) ()
  ())

(DEFFLAVOR IO-FILE-STREAM-MIXIN () (FILE-DATA-STREAM-MIXIN)
   (:INCLUDED-FLAVORS OUTPUT-POINTER-REMEMBERING-MIXIN)
   (:DOCUMENTATION :MIXIN "Streams for use with IO files."))

(DEFMETHOD (IO-FILE-STREAM-MIXIN :REWIND) ()
  (SEND SELF :SET-POINTER 0.))






(DEFSTRUCT (INDENTING-STREAM (:CONC-NAME NIL) (:CALLABLE-CONSTRUCTORS NIL)
  (:ALTERANT ALTER-INDENTING-STREAM) (:PREDICATE NIL) (:COPIER NIL) (:TYPE :ARRAY))
  INDENTING-STREAM-BASE-STREAM
  INDENTING-STREAM-INDENTATION
  INDENTING-STREAM-BEGINNING-OF-LINE?)  



(DEFUN MAKE-STREAM-INDENTABLE (STREAM)
  "Return an indenting stream which passes output to STREAM.
If STREAM is already an indenting stream, it is returned unchanged.
The indenting stream inserts a specified indentation before each line.
It supports the :SET-INDENTATION and :INDENT-RELATIVE operations
to specify the amount of indentation."
  (IF (FUNCALL STREAM :OPERATION-HANDLED-P :INDENT-RELATIVE)
    STREAM
    (LET-CLOSED
     ((*SELF
       (MAKE-INDENTING-STREAM INDENTING-STREAM-BASE-STREAM STREAM INDENTING-STREAM-INDENTATION
	  0. INDENTING-STREAM-BEGINNING-OF-LINE? ())))
     #'(LAMBDA (&REST STUFF)
	 (APPLY #'INDENTING-STREAM-INTERFACE *SELF STUFF)))))  



(DEFUN INDENTING-STREAM-INTERFACE (SELF MESSAGE &REST ARGS)
  (LET ((HANDLER (AND (SYMBOLP MESSAGE) (GET MESSAGE 'INDENTING-STREAM-OPERATIONS))))
    (IF (NULL HANDLER)
      (APPLY (INDENTING-STREAM-BASE-STREAM SELF) MESSAGE ARGS)
      (FUNCALL HANDLER SELF ARGS))))  



(DEFUN INDENTING-STREAM-WHICH-OPERATIONS (SELF)
  (LET ((MY-OPS NIL)
	(BASE-OPS (FUNCALL (INDENTING-STREAM-BASE-STREAM SELF) :WHICH-OPERATIONS)))
    (SETQ MY-OPS
	  (LIST* () :WHICH-OPERATIONS :OPERATION-HANDLED-P :SEND-IF-HANDLES :SET-INDENTATION
		 :INDENT-RELATIVE :TYO :STRING-OUT :FRESH-LINE BASE-OPS))
    (DO ((TAIL MY-OPS (CDR TAIL)))
	((EQ (CDR TAIL) BASE-OPS)
	 (CDR MY-OPS))
      (COND
	((MEMBER (CADR TAIL) BASE-OPS :TEST #'EQ) (RPLACD TAIL (CDDR TAIL))
	 (IF (EQ (CDR TAIL) BASE-OPS)
	   (RETURN (CDR MY-OPS))))))))  



(DEFUN INDENTING-STREAM-INDENT (SELF)
  (IF (FUNCALL (INDENTING-STREAM-BASE-STREAM SELF) :OPERATION-HANDLED-P :INCREMENT-CURSORPOS)
    (FUNCALL (INDENTING-STREAM-BASE-STREAM SELF) :INCREMENT-CURSORPOS
	     (INDENTING-STREAM-INDENTATION SELF) 0. :CHARACTER)
    (DOTIMES (I (INDENTING-STREAM-INDENTATION SELF))
      (FUNCALL (INDENTING-STREAM-BASE-STREAM SELF) :TYO #\SPACE))))  



(DEFUN (:PROPERTY :WHICH-OPERATIONS INDENTING-STREAM-OPERATIONS) (SELF IGNORE)
  (INDENTING-STREAM-WHICH-OPERATIONS SELF))  



(DEFUN (:PROPERTY :OPERATION-HANDLED-P INDENTING-STREAM-OPERATIONS) (SELF ARGS)
  (MEMBER (CAR ARGS) (INDENTING-STREAM-WHICH-OPERATIONS SELF) :TEST #'EQ))  



(DEFUN (:PROPERTY :SEND-IF-HANDLES INDENTING-STREAM-OPERATIONS) (SELF ARGS)
  (IF (MEMBER (CAR ARGS) (INDENTING-STREAM-WHICH-OPERATIONS SELF) :TEST #'EQ)
    (APPLY SELF (CAR ARGS) (CDR ARGS))
    ()))  



(DEFUN (:PROPERTY :SET-INDENTATION INDENTING-STREAM-OPERATIONS) (SELF ARGS)
  (SETF (INDENTING-STREAM-BEGINNING-OF-LINE? SELF) ())
  (SETF (INDENTING-STREAM-INDENTATION SELF) (CAR ARGS)))  



(DEFUN (:PROPERTY :INDENT-RELATIVE INDENTING-STREAM-OPERATIONS) (SELF ARGS)
  (SETF (INDENTING-STREAM-BEGINNING-OF-LINE? SELF) ())
  (INCF (INDENTING-STREAM-INDENTATION SELF) (CAR ARGS)))  



(DEFUN (:PROPERTY :TYO INDENTING-STREAM-OPERATIONS) (SELF ARGS)
  (APPLY (INDENTING-STREAM-BASE-STREAM SELF) :TYO (CAR ARGS) (CDR ARGS))
  (SETF (INDENTING-STREAM-BEGINNING-OF-LINE? SELF) (CHAR-EQUAL (CAR ARGS) #\NEWLINE))
  (IF (NOT (NULL (INDENTING-STREAM-BEGINNING-OF-LINE? SELF)))
    (INDENTING-STREAM-INDENT SELF)))  



(DEFUN (:PROPERTY :STRING-OUT INDENTING-STREAM-OPERATIONS) (SELF ARGS)
  (LET ((STRING (CAR ARGS))
	(START (CADR ARGS))
	(END (CADDR ARGS)))
    (IF (NULL START)
      (SETQ START 0.))
    (IF (NULL END)
      (SETQ END (ARRAY-ACTIVE-LENGTH STRING)))
    (COND
      ((> END START) (SETF (INDENTING-STREAM-BEGINNING-OF-LINE? SELF) ())
       (LET ((MARK START))
	 (DO ()
	     ((NOT (SETQ MARK (%STRING-SEARCH-CHAR #\NEWLINE STRING MARK END)))
	      (FUNCALL (INDENTING-STREAM-BASE-STREAM SELF) :STRING-OUT STRING START END)
	      (RETURN STRING))
	   (FUNCALL (INDENTING-STREAM-BASE-STREAM SELF) :STRING-OUT STRING START MARK)
	   (INDENTING-STREAM-INTERFACE SELF :TYO #\NEWLINE)
	   (COND
	     ((= (SETQ MARK (1+ MARK)) END) (SETF (INDENTING-STREAM-BEGINNING-OF-LINE? SELF) T)
	      (RETURN STRING)))
	   (SETQ START MARK)))))))  



(DEFUN (:PROPERTY :FRESH-LINE INDENTING-STREAM-OPERATIONS) (SELF IGNORE)
  (IF (NULL (INDENTING-STREAM-BEGINNING-OF-LINE? SELF))
    (INDENTING-STREAM-INTERFACE SELF :TYO #\NEWLINE)))  


(compile-flavor-methods stream  CHARACTER-STREAM  INPUT-STREAM   OUTPUT-STREAM BIDIRECTIONAL-STREAM
			BASIC-BUFFERED-INPUT-STREAM  BUFFERED-INPUT-STREAM  BASIC-BUFFERED-OUTPUT-STREAM
			BUFFERED-OUTPUT-STREAM    UNBUFFERED-LINE-INPUT-STREAM
			BUFFERED-LINE-INPUT-STREAM  
			BUFFERED-INPUT-CHARACTER-STREAM
			BUFFERED-OUTPUT-CHARACTER-STREAM BUFFERED-CHARACTER-STREAM BUFFERED-STREAM
			BUFFERED-io-CHARACTER-STREAM  buffered-io-stream
			BUFFERED-TYI-INPUT-STREAM BUFFERED-TYO-OUTPUT-STREAM  BUFFERED-TYI-TYO-STREAM
			)



