;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


;;  8/10/88 DNG - Moved definition of methods for STANDARD-TYPE-CLASS to new 
;;		file "standard-type-class".
;;  8/12/88 DNG - Fixed SINGLE-FLOAT etc. to be subclass of FLOAT.
;;  8/24/88 DNG - Add class for STACK-GROUP .
;;  8/25/88 DNG - Add FUNCTION class.
;;  9/12/88 DNG - STACK-GROUP not a subclass of VECTOR since it supports very 
;;		few of the array functions.  MICROCODE-FUNCTION not a subclass of 
;;		COMPILED-FUNCTION for compatibility with TYPEP.
;; 10/25/88 DNG - Changed metaclass name from STANDARD-TYPE-CLASS to BUILT-IN-CLASS.
;; 11/09/88 DNG - Add class LOCATIVE.
;; 12/20/88 DNG - Fixed discrimination between VECTOR and ARRAY.  Added class 
;;		description entries for DTP-CLOSURE and DTP-LEXICAL-CLOSURE.
;; 12/21/88 DNG - Added support vector entry for class NULL.
;;  4/07/89 DNG - Add class REAL.
;;  4/22/89 DNG - Make LEXICAL-CLOSURE a subclass of CLOSURE for consistency 
;;		with the type hierarchy.
    
(defclass t () () (:metaclass built-in-class))
(defvar *t-class* (class-named 't))

(defclass array	(t) ()(:metaclass built-in-class))
(defclass character  (t) () (:metaclass built-in-class))
(defclass number (t) () (:metaclass built-in-class))
(defclass sequence (t) () (:metaclass built-in-class))
(defclass symbol (t) () (:metaclass built-in-class))
(defclass list (sequence ) () (:metaclass built-in-class))
(defclass null (symbol list) () (:metaclass built-in-class))
(defclass complex (number) () (:metaclass built-in-class))
(defclass cons	  (list) () (:metaclass built-in-class))
(defclass real	  (number) () (:metaclass built-in-class))
(defclass float	  (real) () (:metaclass built-in-class))
(defclass rational (real) () (:metaclass built-in-class))
(defclass integer (rational) () (:metaclass built-in-class))
(defclass ratio	(rational ) () (:metaclass built-in-class))
(defclass vector (array sequence) () (:metaclass built-in-class))
(defclass string (vector) () (:metaclass built-in-class))
(defclass bit-vector (vector ) ()  (:metaclass built-in-class))

;; NIL is not a class name, but make sure someone doesn't accidently define it as one.
(sys:record-source-file-name 'nil 'defclass)
;;;;

(defclass function (t) () (:metaclass built-in-class))

;;;;Our extensions:

(defclass fixnum (integer)() (:metaclass built-in-class))
(defclass bignum (integer)() (:metaclass built-in-class))
(defclass single-float (float) () (:metaclass built-in-class))
(defclass double-float (float) () (:metaclass built-in-class))
(defclass short-float (float) () (:metaclass built-in-class))
(defclass closure (function) () (:metaclass built-in-class))
(defclass lexical-closure (closure) () (:metaclass built-in-class))
(defclass compiled-function (function) () (:metaclass built-in-class))
(defclass microcode-function (function) () (:metaclass built-in-class))

(defclass stack-group (function)() (:metaclass built-in-class))
(defclass locative (t) () (:metaclass built-in-class))

;;Support for microcode
;(defsysconst sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE #o60)  	;#o60 -#o117 - Clos data-type class descriptions
;(Defsysconst sys:%SVC-CLOS-ARRAY-DESCRIPTION-BASE #o120)  	;#o120-#o157 - Clos data-type array descriptions
;(Defsysconst %SVC-CLOS-CLASS-DESCRIPTION-NIL      #o162)  	;#o162- Clos class description for NIL

(proclaim '(special sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE sys:%SVC-CLOS-ARRAY-DESCRIPTION-BASE))

(defun %extended-number-description (object)
     (case (sys:%p-ldb-offset sys:%%header-type-field object 0)
       (#.sys:%header-type-double-float '#,(class-description (class-named 'double-float)))
       (#.sys:%header-type-bignum '#,(class-description (class-named 'bignum)))
       (#.sys:%header-type-rational '#,(class-description (class-named 'ratio)))
       (#.sys:%header-type-complex '#,(class-description (class-named 'complex)))
       (otherwise nil)))


(defun %fef-class-description (object)
  (if (zerop (sys:%p-ldb-offset sys:%%FEF-Storage-Length-Generic-Function-Flag object 1))
      (sys:%class-description (sys:get-debug-info-field (sys:get-debug-info-struct object ) :generic-function))
      '#,(ticlos:class-description (ticlos:class-named  'compiled-function ))))

(dotimes (i 32)
  (si:%add-to-support-vector (ticlos:class-description *t-class*)
			     (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   i) T))

(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named 'fixnum))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-fix) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'single-float ))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-single-float) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'short-float ))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-short-float) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'character ))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-character) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'symbol ))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-symbol) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'cons ))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-list) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'cons ))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-stack-list) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'microcode-function ))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-u-entry) T)
(si:%add-to-support-vector '%fef-class-description
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-function) t)
(si:%add-to-support-vector '%extended-number-description
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   sys:dtp-extended-number) T)

(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named 'stack-group))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-stack-group) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named 'locative))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-locative) T)

(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named 'lexical-closure))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-lexical-closure) T)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named 'closure))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   si:dtp-closure) T)

(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named 'null))
                           sys:%SVC-CLOS-CLASS-DESCRIPTION-NIL T)


;; 11/10/88 DNG - Extended to support classes STRING, VECTOR, and BIT-VECTOR.  
;;		Use FIND-CLASS so that it can create the class if necessary.
;;  3/22/89 DNG - Fix for something that looks like a named structure but 
;;		doesn't have an associated structure definition.  Needed for arrays 
;;		created by FS:MAP-CREATE and NFS:MAKE-FHANDLE .
(defun %defstruct-class (object)
  (declare (optimize (speed 3) (safety 0) (compilation-speed 0))
	   (inline find-class))
  (let ((type (sys:named-structure-p object)) class)
    (or (and type
	     (symbolp type)
	     (setq class (or (get type class-property nil)
			     (class-named type :build)))
	     (class-description class))
	(if (vectorp object)
	    (if (stringp object)
		'#,(ticlos:class-description (find-class 'string))
	      (if (typep object 'bit-vector)
		  '#,(ticlos:class-description (find-class 'bit-vector))
		'#,(ticlos:class-description (find-class 'vector))))
	  '#,(ticlos:class-description (find-class 'array))))))

;; This slot is used for arrays whose number of dimensions is not 1.
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named 'array ))
                           (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE   sys:dtp-array) T)

;; For one-dimensional arrays:
(dotimes (i 32)
  (si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'vector )) 
			   (+ sys:%SVC-CLOS-ARRAY-DESCRIPTION-BASE  i) t))

(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'string )) 
			   (+ sys:%SVC-CLOS-ARRAY-DESCRIPTION-BASE  (ldb si:%%ARRAY-TYPE-FIELD art-string)) t)
(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'bit-vector )) 
			   (+ sys:%SVC-CLOS-ARRAY-DESCRIPTION-BASE  (ldb si:%%ARRAY-TYPE-FIELD art-1b)) t)

(si:%add-to-support-vector (ticlos:class-description (ticlos:class-named  'stack-group )) 
			   (+ sys:%SVC-CLOS-ARRAY-DESCRIPTION-BASE  (ldb si:%%ARRAY-TYPE-FIELD art-stack-group-head)) t)

;; For named structures:
(si:%add-to-support-vector '%defstruct-class
			   (+ sys:%SVC-CLOS-ARRAY-DESCRIPTION-BASE (ldb si:%%ARRAY-TYPE-FIELD art-error)) t)


;; Temporary fixup until the microcode is changed to use the array type table 
;; only for one-dimensional arrays.  [in EXP2-UCODE#411] -- DNG 12/21/88
(unless (eq (sys:%class-description (make-array '(2 2) :element-type 'bit))
	    (class-description (find-class 'array)))
  (dotimes (i 32)
    (unless (= i (ldb si:%%ARRAY-TYPE-FIELD art-stack-group-head))
      (si:%add-to-support-vector '%defstruct-class
				 (+ sys:%SVC-CLOS-ARRAY-DESCRIPTION-BASE i) t)))
  ;;Support for named structures.
  (si:%add-to-support-vector '%defstruct-class 
			     (+ sys:%SVC-CLOS-CLASS-DESCRIPTION-BASE sys:dtp-array) T)
  )

