;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988,1989 Texas Instruments Incorporated. All rights reserved.

;;  5/7/89 DNG - Added this variable for use by the class inspector.
(defvar *all-standard-class-names* nil
  "List of symbols which are the names of instances of STANDARD-CLASS.")
(proclaim '(type list *all-standard-class-names*))

;; 10/03/88 DNG - Modified to test UNDO-DECLARATIONS-FLAG instead of QC-FILE-IN-PROGRESS.
;; 10/06/88 DNG - Redesigned to use environments.
;; 03/16/89 clm - Integrated into Kernel.
;; 03/21/89 clm - moved to this file from "KERNEL;TYPES" - necessary because of build
;;                dependencies (i.e. we can't handle this in the cold band).
;;  5/07/89 DNG - Remove name from *all-standard-class-names* list.
(defun (setf class-named ) (class class-name &optional noerrorp environment)
  (declare (ignore noerrorp))
  (check-type class-name (and symbol (not null)) "a non-nil symbol")
  (compiler:putprop-in-environment class-name class class-property environment)
  (when (and (null class)
	     (not (compiler:environment-remote-p environment)))
    (setf *all-standard-class-names*
	  (delete class-name *all-standard-class-names* :test #'eq :count 1)))
  class)

(defvar method-combination-keywords '(:before :after :around))
;;;This is a compatibility hack. The system is going to send messages to
;;;CLOS instances, and we have to do something. The trick is to provide a method 
;;;hash table in the class description, so the microcode will be happy.
(defflavor xxxx () ())
;;;(ticl:defmethod (xxxx :fasd-form)
;;;   (let* ((size (class-instance-size self))
;;;      `(let ((i (sys:%allocate-and-initialize-instance
;;;		  ',(class-description self)
;;;		  sys:default-cons-area
;;;		  ',(if (= size 1) 2 size))))
;;;	 (sys:%p-store-cdr-code i 1)
;;;	 ))))
(compile-flavor-methods xxxx)
(defvar *compatibility-flavor-hash-table* (sys:flavor-method-hash-table  (get 'xxxx 'sys:flavor)))

(defstruct (class-description :named :array  (:alterant nil)
  ;;(:make-array (:area permanent-storage-area)) ; removed 11/23/88 by DNG
  (:callable-constructors nil)
  (:predicate nil) (:copier nil)
  (:print-function (lambda (cd str ignore)
		     (declare (notinline obsolete-class-description-p))
		     (format str "#<Class-description ~s"
			     (class-description-name cd))
		     ;; 11/29/88 DNG - added flagging of obsolete class.
		     (when (and (fboundp 'obsolete-class-description-p)
				(obsolete-class-description-p cd))
		       (format str "~a" " [obsolete]"))
		     (let ((class (class-named (class-description-name cd) t)))
		       (unless (and class
				    (eq cd (class-description class)))
			 ;; Not the currently installed class; show the address to avoid ambiguity.
			 (format str " ~o" (sys:%pointer cd))))
		     (format str ">"))))
  instance-size
      ;1+ the number of instance variables
  flavor-bindings
      ;List of locatives to instance variable
      ; internal value cells.  MUST BE CDR-CODED!!
      ;Fixnums can also appear.  They say to skip
      ;whatever number of instance variable slots.
  (method-hash-table *compatibility-flavor-hash-table* )
      ;The hash table for methods of this flavor.
      ; NIL means method-combination not composed yet.
      ; T means abstract flavor with COMPILE-FLAVOR-METHODS done.
  name
      ; Symbol which is the name of the class. This is returned by TYPE-OF and CLASS-NAME.
  mapping-tables ;flavor-component-mapping-table-alist
      ;Alist of component flavor names vs.
      ;locatives into vector containing mapping tables.
  direct-slots

  instance-slot-names
	; Names of all instance slots in the order allocated in the instance.  
	; Used by the microcode for LOCATE-IN-INSTANCE.
  method-table					;Defined below.
  ;; End of locations depended on in many other files.
  direct-supers
      ;List of names of classes incorporated into this class.
  direct-subclasses
      ;List of class objects which incorporate this one.
      ;The above are only immediate dependencies.
  class-variable-structure #| flavor-includes |#
      ;; Points to the class-variable array

  class-object    ; Formerly the package in which the DEFFLAVOR was done.
      ;Pointer to the class-object.
  class-precedence-list
      ; Class objects of all classes depended on, to all levels, including
      ; this class itself.  NIL means this class not composed yet.
      ; This is used by TYPEP of 2 arguments.
  (flavor-which-operations ())
      ;List of operations handled, created when needed.
      ; This is NIL if it has not been computed yet.
  (class-slot-alist #| mapped-instance-variables |# ())

  (default-handler nil)
      ; This slot must be NIL because it is used by function SYS:INSTANCE-HASH-FAILURE.
  (all-slot-initargs ())
      ;Alist from init keyword to name of variable
  (init-keywords ())
  (plist ()) ; property list for additional information.

)  

;; 12/06/88 DNG - Moved the cache-alist into the plist because the slot it was 
;;		occupying needs to be kept empty for proper error reporting by 
;;		SYS:INSTANCE-HASH-FAILURE.
(defsubst class-description-update-instance-cache-alist (class-description)
  ;; Used only in RESTRUCTURE-INSTANCE.
  (getf (class-description-plist class-description) 'update-instance-cache-alist))



(defconstant %class-object 13.)

(eval-when (eval compile)
  (assert (equal (macroexpand '(class-description-class-object x))
		 `(aref x ,(- %class-object 1)))))

(defsubst obsolete-class-description-p (class-description)
  (not (zerop (logand (sys:%p-cdr-code class-description)
		      '#.(ldb sys:%%q-cdr-code
			      (dpb 1 sys:%%CLOS-OBSOLETE-CLASS-DESCRIPTION-FLAG 0))))))

(defsubst obsolete-instance-p (instance)
  (obsolete-class-description-p (sys:%class-description instance)))

(defun set-obsolete-class-description (class-description)
  (sys:%p-store-cdr-code class-description
			 '#.(ldb sys:%%q-cdr-code
				 (dpb 1 sys:%%CLOS-OBSOLETE-CLASS-DESCRIPTION-FLAG 0))))

;;Misc accessors for class structure
(defsubst class-slot-descriptions (class)
  (getf (class-description-plist (send class :class-description)) 'slot-descriptions))

(defsubst class-description-finalize-status (class-description)
  (getf (class-description-plist class-description)
	:finalize-status 0))

(DEFSUBST class-description-finalized-p (class-description)
  (EQL 7 (LDB (BYTE 3 0) (class-description-finalize-status class-description))))

;;  5/7/89 DNG - Changed name so as not to conflict with the MOP.
(DEFSUBST internal-class-finalized-p (class)
  (class-description-finalized-p (class-description class)))

(defsubst class-description-class-prototype (class-description)
  (getf (class-description-plist class-description) 'prototype))



;;  9/13/88 DNG - Added documentation slot.

(defstruct ( slot-description (:constructor internal-make-slot-description)
	    (:conc-name slot-)
	    (:print-function (lambda (cd str ignore )
			      (format str "#<slot ~s>" (slot-name cd)))))
  name 
  writers
  readers
  initform
  (allocation :instance)
  (type t)
  initclosure
  initargs
  documentation)
;;initform is a consp if it is provided, or nil otherwise.
;;It can be tested by non-empty-p

(defun non-empty-p (form) (and (consp form) form))

;; The following names are the ones specified in CLOS chapter 3 [88-003 p. 3/21]
(deff SLOT-DESCRIPTION-NAME #'slot-name)
(defun SLOT-DESCRIPTION-INITFORM (slot-description)
  (let ((init (slot-initform slot-description)))
    (and init (or (car init) `(quote nil)))))
(defun SLOT-DESCRIPTION-INITFUNCTION (slot-description)
  (declare (inline non-empty-p))
  (or (slot-initclosure slot-description)
      (let ((init (slot-initform slot-description)))
	(and (non-empty-p init)
	     `(lambda () ,(car init))))))
(deff SLOT-DESCRIPTION-TYPE #'slot-type)
(deff SLOT-DESCRIPTION-ALLOCATION #'slot-allocation)
(deff SLOT-DESCRIPTION-INITARGS #'slot-initargs)
(deff SLOT-DESCRIPTION-READERS #'slot-readers)
(deff SLOT-DESCRIPTION-WRITERS #'slot-writers)

;; Insulate these user-callable names from future 
;; changes to the data structure.  -- DNG 9/21/88
(proclaim '(notinline SLOT-DESCRIPTION-NAME SLOT-DESCRIPTION-INITFORM 
		      SLOT-DESCRIPTION-INITFUNCTION SLOT-DESCRIPTION-TYPE 
		      SLOT-DESCRIPTION-ALLOCATION SLOT-DESCRIPTION-INITARGS 
		      SLOT-DESCRIPTION-READERS SLOT-DESCRIPTION-WRITERS))

;; from page 3-90 of 89-003

(deff SLOT-DEFINITION-NAME #'slot-name)
(defun SLOT-DEFINITION-INITFORM (slot-DEFINITION)
  (let ((init (slot-initform slot-DEFINITION)))
    (and init (or (car init) `(quote nil)))))
(defun SLOT-DEFINITION-INITFUNCTION (slot-DEFINITION)
  (declare (inline non-empty-p))
  (or (slot-initclosure slot-DEFINITION)
      (let ((init (slot-initform slot-DEFINITION)))
	(and (non-empty-p init)
	     `(lambda () ,(car init))))))
(deff SLOT-DEFINITION-TYPE #'slot-type)
(deff SLOT-DEFINITION-ALLOCATION #'slot-allocation)
(deff SLOT-DEFINITION-INITARGS #'slot-initargs)
(deff SLOT-DEFINITION-READERS #'slot-readers)
(deff SLOT-DEFINITION-WRITERS #'slot-writers)

;; Insulate these user-callable names from future 
;; changes to the data structure.
(proclaim '(notinline SLOT-DEFINITION-NAME SLOT-DEFINITION-INITFORM 
		      SLOT-DEFINITION-INITFUNCTION SLOT-DEFINITION-TYPE 
		      SLOT-DEFINITION-ALLOCATION SLOT-DEFINITION-INITARGS 
		      SLOT-DEFINITION-READERS SLOT-DEFINITION-WRITERS))

;;				88-003			89-003
(compiler:make-obsolete SLOT-DESCRIPTION-NAME SLOT-DEFINITION-NAME)
(compiler:make-obsolete SLOT-DESCRIPTION-INITFORM  SLOT-DEFINITION-INITFORM )
(compiler:make-obsolete SLOT-DESCRIPTION-INITFUNCTION SLOT-DEFINITION-INITFUNCTION)
(compiler:make-obsolete SLOT-DESCRIPTION-TYPE  SLOT-DEFINITION-TYPE )
(compiler:make-obsolete SLOT-DESCRIPTION-ALLOCATION SLOT-DEFINITION-ALLOCATION)
(compiler:make-obsolete SLOT-DESCRIPTION-INITARGS  SLOT-DEFINITION-INITARGS )
(compiler:make-obsolete SLOT-DESCRIPTION-READERS SLOT-DEFINITION-READERS)
(compiler:make-obsolete SLOT-DESCRIPTION-WRITERS SLOT-DEFINITION-WRITERS)

;; 11/23/88 DNG - Added metaclass-flavor option.
;; 12/02/88 DNG - Added direct-slots and class-options.
;; 12/05/88 DNG - Send :INIT-PRIMITIVE message.
;; 12/08/88 DNG - Permit replacing a flavor-class definition.
;;  5/07/89 DNG - Add use of *ALL-STANDARD-CLASS-NAMES* .

;;  5/07/89 DNG - Added new class METAOBJECT.  [ref p. 3-9 of 89-003]
(defflavor metaobject () ())
;(defclass-primitive metaobject (t) ()
;  (:metaclass hybrid-class))

;;  5/7/89 DNG - Include METAOBJECT instead of OBJECT.
(defflavor class (class-description) (metaobject)
  :inittable-instance-variables
  :outside-accessible-instance-variables
  (:ordered-instance-variables class-description) ; added 11/17/88 DNG
  (:accessor-prefix ||))

(unless (fboundp 'class-name) ; redefined later in file "metaclass"
  (p-defgeneric-options  class-name (class)
			 (:documentation "Given a class object, return its name."))
  )

(proclaim '(ftype (function (class) list)
		  ;; standard accessors
		  class-default-initargs class-direct-default-initargs 
		  class-direct-slots class-direct-subclasses
		  class-direct-superclasses class-precedence-list class-slots
		  ;; internal accessors
		  class-direct-supers class-all-slots class-all-slot-initargs
		  ))

(p-defgeneric-options  class-direct-supers (class))

(p-defgeneric-options  set-class-direct-supers (class value))

(defsetf class-direct-supers set-class-direct-supers )

(p-defgeneric-options  class-direct-subclasses (class)  
	(:documentation "Returns a list of the direct subclasses of CLASS."))

(p-defgeneric-options  class-precedence-list (class)
  (:documentation "Returns the class precedence list for CLASS.
This list includes the class itself and all superclasses, both direct and 
inherited.  The list is ordered with the most specific class first, and least 
specific last.  The class must have been finalized."))

;; 12/19/88 DNG - Show address only if the name is not a proper name.
;;  2/14/89 DNG - Show address if name is not an interned symbol.
(defmethod (class :print-self ) (stream &rest ignore)
  (declare (type stream stream))
  (write-string "#<" stream)
  (princ (type-of self) stream)
  (let ((name nil))
    (when (variable-boundp  class-description)
      (setq name (class-description-name class-description))
      (princ #\space stream)
      (prin1 name stream))
    (unless (and (symbolp name)
		 (symbol-package name)
		 (eq self (class-named name t)))
      (format stream " ~o" (sys:%pointer self))))
  (write-char #\> stream))


(defmethod (class :class-name) ()
  (class-description-name class-description))

(defmethod (class :class-direct-supers) ()
  (class-description-direct-supers class-description))
(defmethod (class :set-class-direct-supers) (v)
  (setf (class-description-direct-supers class-description) v))

(defmethod (class :class-direct-subclasses) ()
  (class-description-direct-subclasses class-description))

(defsubst internal-class-precedence-list (class)
  ;; Like CLASS-PRECEDENCE-LIST except returns NIL if class is not finalized.
  ;; 10/28/88 DNG - Original.
  (class-description-class-precedence-list (class-description class)))

(defsubst class-composed-p (class)
  ;; 10/28/88 DNG - Original.
  (internal-class-precedence-list class))

;; 10/28/88 DNG - Add error report.
(defmethod (class :class-precedence-list) ()
  (or (class-description-class-precedence-list class-description)
      (unless (class-description-finalized-p class-description)
	;; document 88-003 page 3-16 requires error to be signalled
	(error "Class is not finalized."))))

;; 10/17/88 DNG - Added to speed up SUBCLASSP .
(defmethod (class :subclassp) (other-class)
  ;; Is the other-class contained in the class-precedence list for this class?
  (member other-class
	  (or (class-description-class-precedence-list class-description)
	      (progn (compose-class self)
		     (class-description-class-precedence-list class-description)))
	  :test #'eq))
  

(defflavor standard-class () (class))
(defflavor hybrid-class (flavor) (standard-class))
(defflavor hybrid-class-with-flavor-instances () (hybrid-class))
(defflavor class-class () (hybrid-class-with-flavor-instances))
(defflavor funcallable-standard-class () (hybrid-class-with-flavor-instances))

(defmethod (hybrid-class :compute-instance-size) ()
  ;; 12/02/88 DNG - Original.
  ;;  1/05/89 DNG - Fix to permit composing flavor defined in compile-time environment only.
  ;;  4/29/89 DNG - Add binding of *JUST-COMPILING*.
  ;;  5/08/89 DNG - Ensure that the instance slot names list is cdr-coded.
  ;; Let Flavors figure out the order of the slots so that any ordered 
  ;; instance variables will be handled correctly, but need to add any slots 
  ;; from non-flavor super-classes.
  (when (null (sys:flavor-depends-on-all flavor))
    (let ((si:*just-compiling* (si:just-compiling)) ; needed by SYS:MAP-OVER-COMPONENT-FLAVORS
	  (loc (locf (get (sys:flavor-name flavor) 'sys:flavor))))
      (if (eq flavor (contents loc))
	  (sys:compose-flavor-combination flavor)
	(without-interrupts
	  ;; Temporarily install the flavor definition globally so that it can be 
	  ;; composed.  (Much of the flavor system code assumes that
	  ;; compile-time flavors will never be instantiated.)
	  (sys:%bind loc flavor)
	  (sys:compose-flavor-combination flavor)))))
  (let ((ivars (sys:flavor-all-instance-variables flavor))
	(slots (class-description-instance-slot-names class-description)))
    (if (equal ivars slots)
	(sys:flavor-instance-size flavor)
      (progn (dolist (x slots)
	       (unless (member x ivars :test #'eq)
		 ;; must have been inherited from a non-flavor super-class.
		 (push-end x ivars)))
	     (setf ivars (copy-list ivars)) ; the microcode requires that this be cdr-coded.
	     (setf (sys:flavor-all-instance-variables flavor) ivars)
	     (setf (class-description-instance-slot-names class-description) ivars)
	     (setf (sys:flavor-instance-size flavor) (1+ (length ivars)))))))

(defmethod (hybrid-class :init-primitive) (&rest ignore)
  ;; Invoked by DEFINE-PRIMITIVE-CLASS .
  ;; 12/02/88 DNG - Original (:method class-class :init).
  ;; 12/05/88 DNG - Changed to (:method hybrid-class :init-primitive).
  (when (variable-boundp class-description)	; unless prototype instance
    (let* ((name (class-description-name class-description))
	   (fl (sys:get-flavor name
			       (class-description-environment class-description))))
      (unless (null fl)
	(setf flavor fl)
	(setf (class-description-class-object fl) self)
	(setf (class-description-instance-size class-description)
	      (send self :compute-instance-size))
	)))
  (values))

;; 11/23/88 DNG - Store environment in class description before calling 
;;		INSTANTIATE-FLAVOR so that it is available to an :AFTER :INIT method.
(defun create-class (class-name class-flavor direct-supers &optional direct-slots class-options environment)
  (declare (values class-object class-description))
  (let* ((class-descript
	   (make-class-description :name class-name
				   :direct-supers direct-supers
				   :direct-slots direct-slots
				   :plist class-options))
	 (init-options (list :class-description class-descript))
	 class-object)
    (when (compiler:environment-remote-p environment)
      (setf (class-description-environment class-descript)
	    environment))
    (setf class-object (instantiate-flavor class-flavor (locf init-options) t))
    (setf (class-description-class-object class-descript) class-object)
    (values class-object class-descript)))

(defun define-primitive-class (class-name direct-supers &optional (metaclass-flavor 'standard-class)
			       direct-slots class-options)
  (let ((old-class (class-named class-name t)))
    (when (and (or (null old-class)
		   (sys:typep-structure-or-flavor old-class 'flavor-class))
	       (sys:RECORD-SOURCE-FILE-NAME class-name 'defclass))
      (let ((class-object (create-class class-name metaclass-flavor direct-supers direct-slots class-options)))
	(setf (class-named class-name) class-object)
	(when (typep class-object 'standard-class)
	  (pushnew class-name *all-standard-class-names* :test #'eq))
	(send class-object :init-primitive)
	class-object))))

;; 12/02/88 DNG - Original.
;; 12/05/88 DNG - Permit initial values on flavor slots.
(defmacro defclass-primitive (class-name includes &optional slot-options &rest class-options &environment env)
  ;; This is a primitive version of DEFCLASS used for bootstrapping the pre-defined classes.
  ;; It has the following limitations:
  ;;  * Does nothing if the class is already defined.
  ;;  * The super-class list must not be nil.
  ;;  * Slot options can only be a symbol, not a list.
  ;;  * The only meaningful class options are :METACLASS and :DOCUMENTATION.
  ;;  * For a hybrid-class, the associated flavor must be explicitly defined first.
  (check-type includes cons)
  (let ((metaclass-flavor 'standard-class)
	(direct-options '())
	(flavor (sys:get-flavor class-name env)))
    (dolist (x class-options)
      (if (eq (first x) ':metaclass)
	  (setq metaclass-flavor (second x))
	(push-key-and-value (first x) (second x) direct-options)))
    (dolist (slot slot-options)
      (check-type slot symbol))
    (unless (null flavor)
      (setq slot-options (union slot-options (mapcar #'(lambda (slot)
							 (if (atom slot) slot (car slot)))
						     (sys:flavor-local-instance-variables flavor)))))
    `(define-primitive-class ',class-name ',includes ',metaclass-flavor
       (list . ,(loop for slot in slot-options
		      collect `(internal-make-slot-description :name ',slot)))
       ',direct-options)))

;;;generic functions for standard class.

(defclass-primitive metaobject (t) ()
  (:metaclass hybrid-class))

(def class-direct-slots
     (p-defgeneric-options  class-direct-slots (class )
       (:documentation "Returns a list of the slots defined by the class (not including any inherited
slots).  Each element of the list is an instance of class DIRECT-SLOT-DEFINITION.")))

(p-defgeneric-options  set-class-direct-slots (class value ))

(defsetf class-direct-slots set-class-direct-slots)

(p-defgeneric-options  class-all-slots (class ))
(p-defgeneric-options  set-class-all-slots (class v))
(defsetf class-all-slots  set-class-all-slots )

(p-defgeneric-options  class-class-slot-alist (class ))
(p-defgeneric-options  set-class-class-slot-alist  (class v))
(defsetf class-class-slot-alist   set-class-class-slot-alist  )

(p-defgeneric-options  class-cached-handlers (class))
(p-defgeneric-options  set-class-cached-handlers  (class value))
(defsetf class-cached-handlers  set-class-cached-handlers)

(def class-prototype
     (p-defgeneric-options class-prototype (class)
	(:documentation "Returns a prototype instance of CLASS.  Usually the same instance will be
returned by each call on the same class.  The instance will not be 
initialized, and its slots should not be altered; it is intended for use as an 
argument to a generic function in order to select the method appropriate for 
the class.")))

(p-defgeneric-options class-mapped-slot-names (class))
(p-defgeneric-options set-class-mapped-slot-names (class value))
(defsetf class-mapped-slot-names set-class-mapped-slot-names)

(p-defgeneric-options class-mapping-tables (class))
(p-defgeneric-options set-class-mapping-tables (class value))
(defsetf class-mapping-tables set-class-mapping-tables )

(p-defgeneric-options class-mapped-supers (class))
(p-defgeneric-options set-class-mapped-supers  (class value))
(defsetf class-mapped-supers  set-class-mapped-supers )

(p-defgeneric-options  class-constructors (class))
;;;(p-defgeneric-options class-options (class))
;;;Look in class.lisp

(p-defgeneric-options class-instance-size (class))
(p-defgeneric-options set-class-instance-size (class value))
(defsetf class-instance-size set-class-instance-size)

(defmethod (standard-class :class-direct-slots) ()
  (class-description-direct-slots class-description))

(defmethod (standard-class :set-class-direct-slots) (value)
  ;;must do some work and call supers-changed
  (setf (class-description-direct-slots class-description) value))

(defmethod (standard-class :class-all-slots) ()
  (getf (class-description-plist class-description) 'all-slots))
(defsubst class-description-all-slots (class-description)
  (getf (class-description-plist class-description) 'all-slots))
(defmethod (standard-class :set-class-all-slots) (v)
  (setf (getf (class-description-plist class-description) 'all-slots) v))

(defmethod (standard-class :class-class-slot-alist) ()
  (getf (class-description-plist class-description) 'class-slot-alist))
(defmethod (standard-class :set-class-class-slot-alist) (v)
  (setf (getf (class-description-plist class-description) 'class-slot-alist) v))



(defmethod (standard-class   :class-cached-handlers )()
  (getf (class-description-plist class-description ) 'cached-handlers))
(defmethod (standard-class   :set-class-cached-handlers ) (value)
  (setf (getf (class-description-plist class-description ) 'cached-handlers) value))


(defmethod (standard-class :class-constructors ) ()
  (getf (class-description-plist class-description) :constructors))

;(defmethod (standard-class :class-options ) ()
;  (getf (class-description-plist class-description) :class-options))

;;Since prototype is an anonymous class, its name should be nil.

;; 12/02/88 DNG - simplified by moving flavor handling to new method for class-class.
(defmethod (standard-class :class-prototype) ()
  (or (class-description-class-prototype class-description)
      (setf (class-description-class-prototype class-description)
	    (allocate-standard-instance self))))

;; 12/05/88 DNG - Hack to make compile-time flavor instantiable.
;; 12/06/88 DNG - Changed method from flavor class-class.
;; 12/08/88 DNG - Use without-interrupts to avoid a loophole.
(defmethod (hybrid-class-with-flavor-instances :class-prototype) ()
  (or (class-description-class-prototype class-description)
      (setf (class-description-class-prototype class-description)
	    (let ((loc (locf (get (sys:flavor-name flavor) 'sys:flavor))))
	      (if (eq flavor (contents loc))
		  (sys:make-flavor-instance flavor)
		(without-interrupts
		  ;; Temporarily install the flavor definition globally so that it can be 
		  ;; instantiated.  (Much of the flavor system code assumes that
		  ;; compile-time flavors will never be instantiated.)
		  (sys:%bind loc flavor)
		  (sys:make-flavor-instance flavor)))))))

(defmethod (standard-class :class-mapped-slot-names) ()
  (getf (class-description-plist class-description) 'mapped-slot-names))

(defmethod (standard-class :set-class-mapped-slot-names) (value)
  (setf (getf (class-description-plist class-description) 'mapped-slot-names) value))

(defmethod (standard-class :class-mapping-tables ) ()
  (class-description-mapping-tables class-description))

(defmethod (standard-class :set-class-mapping-tables ) (value)
  (setf (class-description-mapping-tables class-description)value))

(defmethod (standard-class :class-mapped-supers  ) ()
  (getf (class-description-plist class-description) 'class-mapped-supers  ))
(defmethod (standard-class :set-class-mapped-supers  ) (value)
  (setf (getf (class-description-plist class-description) 'class-mapped-supers  ) value))

(defmethod (standard-class :class-instance-size) ()
  (class-description-instance-size class-description))
(defmethod (standard-class :set-class-instance-size) (size)
  (setf (class-description-instance-size class-description) size))

(defsubst class-description-environment (class-description)
  ;; 10/26/88 DNG - Original.
  (getf (class-description-plist class-description) 'definition-environment))

(defsubst class-environment (class)
  ;; Return the environment in which this class was defined.
  ;; 10/26/88 DNG - Original.
  (class-description-environment (class-description class)))				

;; 10/25/88 DNG - Changed name from STANDARD-TYPE-CLASS to BUILT-IN-CLASS .
(defflavor built-in-class () (class)
  :inittable-instance-variables
  :outside-accessible-instance-variables
  (:accessor-prefix class-))

(comment ; removed 4/5/89
;; Temporary for compatibility with old name
(when (sys:record-source-file-name 'standard-type-class 'defflavor)
  (setf (get 'standard-type-class 'sys:flavor)
	(get 'built-in-class 'sys:flavor)))
)

(proclaim '(inline class-of))

(defun class-of (object)
  "Return the class object of which the given object is an instance."
  (class-description-class-object (sys:%class-description object)))




;;;Fake classes for bootstrapping
;; 11/17/88 DNG - removed instance variable NAME.
(defflavor object ( #| name |# )
	   ()
  :inittable-instance-variables
  :outside-accessible-instance-variables
  )

(comment ; obsolete
(ticl:defmethod (object :class-name) () name)

(ticl:defmethod  (object :print-self ) (stream &rest ignore)
  (format stream "#<Object ~s>" name))
 )


;(defmethod (hybrid-class :compute-instance-size) ()
;  ;; 12/02/88 DNG - Original.
;  ;;  1/05/89 DNG - Fix to permit composing flavor defined in compile-time environment only.
;  ;;  4/29/89 DNG - Add binding of *JUST-COMPILING*.
;  ;;  5/08/89 DNG - Ensure that the instance slot names list is cdr-coded.
;  ;; Let Flavors figure out the order of the slots so that any ordered 
;  ;; instance variables will be handled correctly, but need to add any slots 
;  ;; from non-flavor super-classes.
;  (when (null (sys:flavor-depends-on-all flavor))
;    (let ((si:*just-compiling* (si:just-compiling)) ; needed by SYS:MAP-OVER-COMPONENT-FLAVORS
;	  (loc (locf (get (sys:flavor-name flavor) 'sys:flavor))))
;      (if (eq flavor (contents loc))
;	  (sys:compose-flavor-combination flavor)
;	(without-interrupts
;	  ;; Temporarily install the flavor definition globally so that it can be 
;	  ;; composed.  (Much of the flavor system code assumes that
;	  ;; compile-time flavors will never be instantiated.)
;	  (sys:%bind loc flavor)
;	  (sys:compose-flavor-combination flavor)))))
;  (let ((ivars (sys:flavor-all-instance-variables flavor))
;	(slots (class-description-instance-slot-names class-description)))
;    (if (equal ivars slots)
;	(sys:flavor-instance-size flavor)
;      (progn (dolist (x slots)
;	       (unless (member x ivars :test #'eq)
;		 ;; must have been inherited from a non-flavor super-class.
;		 (push-end x ivars)))
;	     (setf ivars (copy-list ivars)) ; the microcode requires that this be cdr-coded.
;	     (setf (sys:flavor-all-instance-variables flavor) ivars)
;	     (setf (class-description-instance-slot-names class-description) ivars)
;	     (setf (sys:flavor-instance-size flavor) (1+ (length ivars)))))))

(defmethod (class :default :init-primitive) ignore)

;(defmethod (hybrid-class :init-primitive) (&rest ignore)
;  ;; Invoked by DEFINE-PRIMITIVE-CLASS .
;  ;; 12/02/88 DNG - Original (:method class-class :init).
;  ;; 12/05/88 DNG - Changed to (:method hybrid-class :init-primitive).
;  (when (variable-boundp class-description)	; unless prototype instance
;    (let* ((name (class-description-name class-description))
;	   (fl (sys:get-flavor name
;			       (class-description-environment class-description))))
;      (unless (null fl)
;	(setf flavor fl)
;	(setf (class-description-class-object fl) self)
;	(setf (class-description-instance-size class-description)
;	      (send self :compute-instance-size))
;	)))
;  (values))


(defstruct (arglist-description (:type list)(:callable-constructors nil))
  (n-required 0)
  (n-optional 0)
  (&allow-other-keys-p nil)
  (keywords '())
  (&rest-p nil)
  (&key-p nil))

;;;This one is an internal generic-function.
;;;Generic functions are implemented as functions whose
;;;debug-info-field :generic-function points to a generic-function-object.
;;  5/7/89 DNG - Include METAOBJECT instead of OBJECT.
(defflavor generic-function
	   ((method-list nil)
	    discriminator-code)
	   (metaobject)
  :inittable-instance-variables
  :outside-accessible-instance-variables
  (:default-handler default-generic-function-handler))

(defun default-generic-function-handler (&rest args)
  ;; This enables a generic function object to be funcallable.
  ;; 12/06/88 DNG - Original.
  (let ((generic-function self)) ; so this will show up in the debugger
    (apply (generic-function-discriminator-code generic-function) args)))

(defflavor standard-generic-function
	   (lambda-list
	    (argument-precedence-order nil)
	    (declare nil)
	    (documentation nil)
	    (method-combination '(standard))
	    ;; (generic-function-class) ; removed 12/8/88
	    method-class
	    (reorder-specs nil)
	    arglist-description
	    (name nil)
	    (initial-methods)
	    (method-hash-table nil))
	   (generic-function)
  :inittable-instance-variables
  :outside-accessible-instance-variables
  (:accessor-prefix generic-function-)
  (:default-handler default-generic-function-handler))

(defmethod  (standard-generic-function :print-self ) (stream &rest ignore)
  (format stream "#<~a ~s>" (type-of self) name))

;;  5/07/89 DNG - Include METAOBJECT.
(defflavor method
	   ((generic-function nil)
	    parameter-specializers
	    (function nil)
	    )
	   (metaobject)
  :gettable-instance-variables
  :inittable-instance-variables
  (:outside-accessible-instance-variables parameter-specializers))

;;  4/21/89 DNG - Added the following 2 functions for use internally.  They 
;;		are redefined as generic functions later in file "meta-object".
(proclaim '(notinline *method-generic-function *method-function))
(unless (fboundp '*method-generic-function)
  (defun *method-generic-function (method) (send method :generic-function)))
(unless (fboundp '*method-function)
  (defun *method-function (method) (send method :function)))

;; 5/7/89 DNG - Include METAOBJECT.
(defclass-primitive method (metaobject)
  ()
  (:metaclass hybrid-class-with-flavor-instances))

(proclaim '(notinline method-qualifiers)) ; don't expand the temporary definition.

(defflavor standard-method
	   (qualifiers
	    arglist-description
	    (plist nil)
	    function-location)
	   (method)
  :inittable-instance-variables
  :outside-accessible-instance-variables
  (:settable-instance-variables arglist-description)
  (:init-keywords :lambda-list)
  (:accessor-prefix method-))

;;  4/10/89 DNG - Show address of method without a proper name.
(defmethod  ( standard-method :print-self) (stream &rest ignore)
  (declare (type stream stream))
  (let ((name (function-name function)))
    (format stream "#<method ~s" name)
    (unless (and (sys:validate-function-spec name)
		 (eq (si:fdefinition-safe name t) function))
      ;; Not the currently installed method; show the address to avoid ambiguity.
      (format stream " ~o" (sys:%pointer self))))
  (write-char #\> stream))

;  (format stream "#<method (method ~s ~{~s ~}~s)>"
;	  (and generic-function (generic-function-name generic-function))
;	  qualifiers parameter-specializers))

(defclass-primitive standard-method (method) ()
  (:metaclass hybrid-class-with-flavor-instances))


(defflavor combined-method
	   ((derivation nil))
	   (standard-method)
  :inittable-instance-variables
  :outside-accessible-instance-variables
  (:accessor-prefix method-))

(defflavor null-combined-method (forwarded-method) (combined-method)
  :inittable-instance-variables
  :outside-accessible-instance-variables
  (:accessor-prefix method-))

(defmethod  ( null-combined-method  :print-self) (stream &rest ignore)
  (format stream "#<forwarded-method ~s>"
	  (function-name (*method-function forwarded-method))))

(defsubst combined-method-p (obj)
  (typep obj 'combined-method))

;; 1/3/89 DNG - added next 2 methods for use by HANDLER function spec.
(defmethod (standard-method :method-fef-location) () function-location)
(defmethod (null-combined-method :method-fef-location) ()
  (send forwarded-method :method-fef-location))

;;  5/08/89 DNG - remove the accessor for METHOD-SLOT-NAME since it conflicts with the MOP generic function.
(defflavor standard-accessor-method (slot-name) (combined-method)
  :inittable-instance-variables)

(defflavor standard-reader-method () (standard-accessor-method))
(defflavor standard-writer-method () (standard-accessor-method))

(defsubst slot-accessor-p (obj)
  (typep obj 'standard-accessor-method))

(defsubst handlerp (obj)
  (or (typep obj 'combined-method)
      (typep obj 'standard-accessor-method)))

;;  3/13/89 DNG - Original.
(defsubst combined-effective-method-p (obj)
  (and (typep obj 'combined-method)
       (not (typep obj 'standard-accessor-method))))

(defstruct (method-spec-object (:type :list*))
  function-spec qualifiers method function plist)



;;;Now defined in method-hash-table
;;;(defstruct individual-dispatch-hash-table
;;;  (ht (sys:make-hash-array  :number-of-values 2 :funcallable-p t :test #'eq))
;;;  default-value)

(proclaim '(inline individual-typep ))
(defun individual-typep (type)
  (and (eq (car-safe type) 'eql) (cdr type)))

;;  4/21/89 DNG - For efficiency, changed to a DEFSUBST, and assume that the 
;;		argument is an individual type instead of checking.
(defsubst individual-type (type)
  (second type))

(deftype eql (x) `(member ,x))

;;  4/21/89 DNG - New function.
(proclaim '(inline specializer-class))
(defun specializer-class (specializer)
  (if (individual-typep specializer)
      (class-of (individual-type specializer))
    specializer))

(proclaim '(special *t-class*)) ; defined later in "built-in-classes".
