;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


(defmacro collect-body ( &body body )
  "iteration support"
  (let ((tail (gensym))
	(result (gensym)))
  `(macrolet ((collect (item)
	        `(rplacd ,',tail (setf ,',tail (cons ,item nil))))
	      (join (item)
		`(and (cdr (rplacd ,',tail  ,item))
		      (setf ,',tail (last (cdr ,',tail))))))
     (let* ((,result nil)
	    (,tail (locf ,result)))
       (declare (list ,result))
       ,@body
       ,result))))

(defmacro multiple-values-collect-body (results &body body )
  "iteration support.
    (multiple-values-collect-body (a b c) ... (collect-into a ...))"
  (let ((tails (collect-body (dolist (i  results) (collect (gensym))))))
  `(macrolet ((collect-into  (var item)
		 (let ((tail (nth (position var ',results :test #'eq) ',tails)))
		   `(rplacd ,tail (setf ,tail (cons ,item nil)))))
	      (join-into (var item)
		 (let ((tail (nth (position var', results :test #'eq) ',tails)))
		   `(and (cdr (rplacd ,tail  ,item))
		      (setf ,tail (last (cdr ,tail)))))))
     (let* (,@(mapcar #'(lambda (result) `(,result nil)) results)
	    ,@(mapcar #'(lambda (tail result) `(,tail (locf ,result))) tails results))
       (declare (list ,@results))
       ,@body
       (values ,@results)))))

(defmacro with-pattern-matching (object clauses &body body)
  (let* ((sys:binding-list ())
	 (genvar (gensym))
	 (clauses (mapcar #'sys:select-match-clause clauses (circular-list genvar))))
    (declare (special sys:binding-list))
    `(let ((,genvar ,object))
       (cond
	 . ,clauses)
       . ,body)))

;;  8/30/88 DNG - Added support for DECLAREs.
;; 10/28/88 DNG - Add checking for proper argument type.
(defmacro p-defgeneric-options (name arglist &rest options)
  (let ((options 
	  (mapcan #'(lambda (option)
		      (case (car option )
			(:documentation  (cdr option))
			(declare (list option))
			(t nil)))
		  options))
	(true-arglist (cons (if (consp (first arglist)) (car (first arglist)) (first arglist)) (rest arglist))))
  `(defun ,name ,true-arglist
     ,@options
     ,@(and (compiler:external-symbol-p name) ; user-callable
	    (let ((arg (first arglist)))
	      (and (symbolp arg)
		   (if (eq arg 'class)
		       `((or (sys:classp class)
			     (error "The argument CLASS was ~S, which is not a class object." class)))
		     `((or (instancep ,arg)
			   (error "The argument ~A was ~S, which is not an instance."
				  ',arg ,arg)))))))
     (,(if (member '&rest arglist :test #'eq)
	   'apply
	   'send)
      ,(if (eq (car arglist)(car true-arglist)) (car arglist) (cadar arglist))
      ,(intern name *keyword-package*)
      ,@(mapcan #'(lambda (arg)
		    (if (member arg lambda-list-keywords :test #'eq)
			nil
			(cons (if (consp arg) (car arg) arg)
			      nil)))
		(cdr arglist))))))




(defmacro multiple-value-prog2 (form1 form2 &body body)
  `(progn
     ,form1
     (multiple-value-prog1 ,form2
			   ,@body)))



(defmacro with-lambda-list (lambda-list &key required  optionals optionals-before
			    rest keys keys-before allow-other-keys-key aux aux-before end)
   "parses a lambda-list"
  (let ((ll (gensym))
	(l (gensym)))
  `(flet ((bug()
	     (error "invalid lambda-list  ~s" ,lambda-list)))
     (let ((,ll ,lambda-list))
       (macrolet ((exit () '(go end)))
       (tagbody
	   (do((,l ,ll (cdr ,l)))
	      ((or (null ,l) (member  (car ,l) '(&optional &rest &key &aux)))
	       (setf ,ll (cdr ,l))
	       (case (car ,l)
		 (&optional (go opt))
		 (&rest (go rest))
		 (&key (go key))
		 (&aux (go aux))
		 (t (go end))))
	     ,@(and required `((funcall ,required (car ,l))))
	     )
	opt
	   ,@(and optionals-before `((funcall ,optionals-before)))
	   (do ((,l ,ll (cdr ,l)))
	       ((or (null ,l) (member  (car ,l) '(&rest &key &aux)))
		(setf ,ll (cdr ,l))
		(case (car ,l)
		  (&rest (go rest))
		  (&key (go key))
		  (&aux (go aux))
		  (t (go end))))
	     ,@(and optionals `((funcall ,optionals (car ,l)))))
	rest
	   ,@(and rest `((funcall ,rest (car ,ll))))
	   (setf ,ll (cdr ,ll))
	   (case (pop ,ll)
	     (&key (go key))
	     (&aux (go aux))
	     (nil (go end))
	     (t (bug)))
	key
	   ,@(and keys-before `((funcall ,keys-before)))
	   (do ((,l ,ll (cdr ,l)))
	       ((or (null ,l) (member  (car ,l) '(&aux &allow-other-keys)))
		(setf ,ll (cdr ,l))
		(case (car ,l)
		  (&aux (go aux))
		  (&allow-other-keys (go allow))
		  (nil (go end))
		  (t )))
	     ,@(and keys `((funcall ,keys (car ,l))))
	     )
	allow
	   ,@(and allow-other-keys-key `((funcall ,allow-other-keys-key )))
	   (case (car ,ll)
	     (&aux (setf ,ll (cdr ,ll)) (go aux))
	     (nil (go end))
	     (t (bug)))
	aux
	   ,@(and aux-before `((funcall ,aux-before)))
	   (do ((,l ,ll (cdr ,l)))
	       ((null ,l) (go end))
	     ,@(and aux `((funcall ,aux (car ,l))))
	     )
	end
	   ,@(and end `((funcall ,end)))
	   ))))))


;;(defmacro with-slots (slot-entry-list instance-form &body body)
;;  (let ((gs (gensym)))
;;    `(let ((,gs ,instance-form))
;;       (symbol-macrolet
;;	 (
;;	  ,@(collect-body
;;	    (dolist (se slot-entry-list)
;;	      (collect `(,(if (consp se) (car se) se) (slot-value ,gs ',(if (consp se) (cadr se) se)))))))
;;	 ,@body))))

(defmacro with-slots (slot-entry-list instance-form &body body)
  "Create a lexical context for referring to specified slots of an object as though
they were variables.  Each element of slot-entry-list is either a slot-name or a
list (variable-name slot-name).  For example:
\(WITH-SLOTS (X Y) OBJ (SETQ X Y)) ==> (SETF (SLOT-VALUE OBJ 'X) (SLOT-VALUE OBJ 'Y))"

  (once-only (instance-form)
    `(symbol-macrolet
       (,@(collect-body
	    (dolist (se slot-entry-list)
	      (collect `(,(if (consp se) (car se) se) (slot-value ,instance-form ',(if (consp se) (cadr se) se)))))))
       ,@body)))

(defmacro with-accessors (slot-entry-list instance-form &body body)
  "Create a lexical environment in which specified slots of an object are lexically
available through their accessors as if they were variables. The slot-entry-list
has the form:  ((variable-name accessor-function)...)   For example:
 (WITH-ACCESSORS ((X XA)(Y YA)) OBJ (SETQ X Y)) ==> (SETF (XA OBJ) (YA OBJ))"

  (once-only (instance-form)
    `(symbol-macrolet
       (,@(collect-body
	    (dolist (se slot-entry-list)
	      (collect `(,(car se) (,(cadr se) ,instance-form))))))
       ,@body)))

;; for PPRINT	11/23/88 DNG
;; 12/12/88 change to use DEFPRINT macro.
(sys:defprint with-slots -2) ; print the same as MULTIPLE-VALUE-BIND.
(sys:defprint with-accessors -2)


;;  Added 9/20/88 by DNG
(defflavor method-combination-error () (ferror))
(defsignal method-combination-error (method-combination-error) ()
   "Signaled for failure in CLOS method combination.")

(defmacro ignore-method-combination-error (&body body)
  ;; This is like IGNORE-ERRORS but only ignores METHOD-COMBINATION-ERROR .
  (let ((tag (gensym)))
    `(catch-continuation ',tag #'(lambda () (values nil t)) nil
       (condition-bind ((method-combination-error 'sys::ignore-errors-handler ',tag))
	 (values (progn . ,body) nil)))))

;;  Added 9/23/88 by DNG
(defmacro push-key-and-value (key value list)
  "Push KEY and VALUE onto a property list."
  `(setf ,list (list* ,key ,value ,list)))

;;  3/18/89 DNG - Moved CLASS-PROPERTY to file "KERNEL;TYPES".

