;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.


;;Prototypes for class and standard-class

;;  5/7/89 DNG - Include METAOBJECT.
(defclass-primitive class (metaobject standard-object) (class-description)
  (:metaclass class-class)
  (:documentation "All class objects are members of this class."))
(defclass-primitive standard-class (class) ()
  (:metaclass class-class)
  (:documentation "Default metaclass for CLOS objects."))

(defclass-primitive hybrid-class (standard-class) (flavor)
  (:metaclass class-class)
  (:documentation "Metaclass for classes that combine flavors and CLOS classes."))
(defclass-primitive hybrid-class-with-flavor-instances (hybrid-class) ()
  (:metaclass class-class)
  (:documentation "Like hybrid-class except that instances are flavor instances."))
(defclass-primitive class-class (hybrid-class-with-flavor-instances) (class-description)
  (:metaclass class-class)
  (:documentation "The metaclass of class objects."))

(defclass-primitive built-in-class  (class) () (:metaclass class-class))

(defclass-primitive standard-object (t))

(defclass-primitive funcallable-standard-class (hybrid-class-with-flavor-instances)
  (class-description flavor)
  (:metaclass class-class)
  (:documentation "The metaclass for generic functions."))

;;  5/7/89 DNG - Include METAOBJECT.
(defclass-primitive generic-function (metaobject function
				       #| standard-object |#	; should be added someday
				       ) ()
  (:metaclass funcallable-standard-class))

(defclass-primitive standard-generic-function
		    (generic-function
		      ;; The next line may need to be removed if generic functions become standard-objects.
		      compiled-function 
		      ) ()
  (:metaclass funcallable-standard-class))

(defflavor init-generic-function () (standard-generic-function))
(defclass-primitive init-generic-function (standard-generic-function) ()
  (:metaclass funcallable-standard-class))

;;; The next 4 added 1/30/89 by DNG
(defclass-primitive combined-method (standard-method) (derivation)
  (:metaclass hybrid-class-with-flavor-instances))
(defclass-primitive standard-accessor-method (combined-method) (slot-name)
  (:metaclass hybrid-class-with-flavor-instances))
(defclass-primitive standard-reader-method (standard-accessor-method) ()
  (:metaclass hybrid-class-with-flavor-instances))
(defclass-primitive standard-writer-method (standard-accessor-method) ()
  (:metaclass hybrid-class-with-flavor-instances))

;; 12/08/88 DNG - Added to enable compiling methods specialized on (EQL (FIND-CLASS ...)).
(defmethod (class :fasd-form) ()
  (let ((name (class-proper-name self)))
    (if (symbolp name)
	`(class-named ',name)
      (error "Can't dump ~S to object file because it doesn't have a proper name." self))))

;;;external symbol-class support
;;Symbol-class is obsolete.
(defun symbol-class (name &optional environment)
  (class-named name nil environment))
(compiler:make-obsolete symbol-class find-class)

(eval-when (compile) (proclaim '(inline find-class))) ; to save interp. def. for local inline

;;  9/27/88 DNG - Pass ENVIRONMENT on to CLASS-NAMED.
;; 10/26/88 DNG - Shortcut simple cases.
(defun find-class (name &optional (errorp t) environment)
  "Returns the class object named by the given symbol in the given environment.
Returns NIL if no such class and the second argument is NIL.  The third 
argument should be an environment object received as the &ENVIRONMENT argument 
of a macro if it is desired to access the compile-time environment."
  (or (and (not (compiler:environment-remote-p environment))
	   (get name class-property nil))	; quick way
      (class-named name (not errorp) environment)))

(proclaim '(notinline find-class))

;; 10/06/88 DNG - Redesigned to use environments.
;;  5/03/89 DNG - When undefining a hybrid-class, undefine the associated flavor also.
;;  5/07/89 DNG - Remove name from *all-standard-class-names* list when CLASS is NIL.
(defun (setf find-class) (class class-name &optional ignore environment)
  (if (and (null class)
	   (not (compiler:environment-remote-p environment)))
      (let ((old (car (remprop class-name class-property))) fl)
	(when (and (typep old 'hybrid-class)
		   (setq fl (sys:get-flavor class-name))
		   (slot-boundp (the hybrid-class old) 'flavor)
		   (eq fl (slot-value (the hybrid-class old) 'flavor)))
	  (undefflavor class-name))
	(setf *all-standard-class-names*
	      (delete class-name *all-standard-class-names* :test #'eq :count 1)))
    (progn (check-type class-name (and symbol (not null)) "a non-nil symbol")
	   (check-type class (or ticlos:class null) "a class object or NIL")
	   (compiler:putprop-in-environment class-name class class-property environment)))
  class)

;; obsolete
(defun set-symbol-class (name second &optional (third nil third-p))
  (let ((class (if third-p third second)))
    (setf (class-named name ) class)))

;; obsolete
(defsetf symbol-class set-symbol-class)

;;obsolete
(defun cboundp (name &optional environment)
  (and (class-named name t environment)
       t))
(compiler:make-obsolete cboundp "use FIND-CLASS with a second argument of NIL")

;;obsolete
(defun cmakunbound (name &optional environment)
  (declare (ignore environment))
  (remprop name class-property)
  name)
(compiler:make-obsolete cmakunbound "use (SETF (FIND-CLASS name NIL) NIL)")

(defun record-documentation-p ()
  (and (boundp 'compiler:*SUPPRESS-DEBUG-INFO*)
       (not compiler:*SUPPRESS-DEBUG-INFO*)))

(defun get-default-class-metaclass-prototype ()
  (class-prototype (class-named 'standard-class)))

(compiler-let ((local-declarations '((:expr-sxhash 1216958.))))
(defmacro defclass (&whole defclass-form class-name includes slot-options &rest class-options
		    &environment environment)
  (declare (arglist class-name includes slot-options &rest class-options))
  "Define a named CLOS class.  The new class object is returned.  Syntax:
\(DEFCLASS name ( {superclass}* ) ( {slot-specifier}* ) {class-option}* )
name and superclass are class name symbols.
slot-specifier ::= slot-name | ( slot-name {slot-option}* )
slot-option ::= { :ACCESSOR | :READER | :WRITER } function-spec | 
		:INITARG symbol | :INITFORM form | :TYPE type-specifier |
		:DOCUMENTATION \"string\" | :ALLOCATION { :INSTANCE | :CLASS } 
class-option ::= (:METACLASS class-name) | (:DOCUMENTATION \"string\") |
		 (:DEFAULT-INITARGS {key value-form}*)"

  (let ((forms (defclass1 defclass-form class-name includes slot-options class-options environment)))
    (if (and (null (cdr forms))
	     (member (caar forms) '(clos:add-named-class add-named-class-internal)
		     :test #'eq))
	(car forms)
      `(progn
	 ,@forms
	 (class-named ',class-name)))))
)

(defun defclass1 (defclass-form class-name includes slot-options class-options environment)
  (declare (ignore includes slot-options))
  (unless (symbolp class-name)
    (non-fatal-error ':impossible
		      "The first argument of DEFCLASS is ~S which is not a symbol." class-name))
  ;;check :metaclass-name from class-options
  (let ((metaclass nil ) )
    (dolist (co class-options)
      (when (or (eq (first co) :metaclass)
		(and (eq (first co) :metaclass-name)
		     (progn (non-fatal-error ':obsolete
					      "Class option :METACLASS-NAME is obsolete; use :METACLASS.")
			    t)))
	(when metaclass
	  (non-fatal-error ':impossible ":METACLASS option provided more than once."))
	(setf metaclass (class-named (second co) nil environment))))
    (expand-defclass (if metaclass
			 (class-prototype metaclass)
			 (get-default-class-metaclass-prototype))
		     defclass-form
		     environment)))

(p-defgeneric-options expand-defclass (class-prototype defclass-form &optional environment))

;; 10/06/88 DNG - Add support for compile-time definition.
;; 10/26/88 DNG - Look up metaclass in compile-time environment.
;; 12/01/88 DNG - Use function ADD-NAMED-CLASS-INTERNAL instead of TICLOS::ADD-NAMED-CLASS.
(defmethod (standard-class :expand-defclass) (defclass-form &optional environment)
  (let ((class-name (progn (pop defclass-form)(pop defclass-form)))
	(includes (pop defclass-form))
	(slot-options (pop defclass-form))
	(class-options defclass-form)
	(def-initargs nil)
	(def-initarg-alist nil)
;	constructor 
	metaclass-name doc-string doc-string-provided-p
	metaclass 
	;;metaclass-provided-p
	)
    (dolist (option class-options)
      (with-pattern-matching  option
	  ((`(:documentation ,doc-string) t
	    (check-type doc-string string)
	    (when doc-string-provided-p
	      (non-fatal-error ':ignorable-mistake
				"Class option ~S provided more than once." (car option)))
	    (setf doc-string-provided-p t))
	   (`(:default-initargs . ,def-initargs) t
	    (do* ((di def-initargs (cddr di))
		  (def-initarg (car di)(car di))
		  (form (second di)(second di)))
		 ((endp di))
	      (when (assoc def-initarg def-initarg-alist :test #'eq)
		(non-fatal-error ':impossible ":DEFAULT-INITARG ~s is specified twice."
				  def-initarg))
	      (push (list def-initarg  `(freeze ,form) form)
		    def-initarg-alist)))
	   (`(:metaclass ,metaclass-name) t
	    (check-type metaclass symbol)
	    (comment ; already checked in DEFCLASS1
	      (assert (null metaclass-provided-p) nil
		      "option ~s provided more than once" option))
	    (comment ; not used here
	      (setf metaclass-provided-p t)
	      (setf metaclass (class-named metaclass-name nil environment))))
	   (t (non-fatal-error ':impossible "Unrecognized class option ~s" option)))))
    (multiple-value-bind ( slot-description-list slot-forms)
	(parse-slot-options self slot-options class-name)
      (let ((add-form `(add-named-class-internal
			 ',(type-of self)
			 ',class-name
			 ',includes
			 (list ,@slot-forms)
			 ,(let ((class-options '()))
			    (when (and doc-string
				       (record-documentation-p))
			      (push-key-and-value :documentation doc-string class-options))
			    (comment ; not needed here -- DNG 11/18/88
			      (unless (or (null metaclass-name)
					  (eq metaclass-name 'standard-class))
				(push-key-and-value :metaclass `',metaclass-name class-options)) )
			    (unless (null def-initarg-alist)
			      (push-key-and-value :default-initargs
						  ``(,,@(mapcar #'(lambda (x)
								    ``(,',(first x)
								       ,,(second x)
								       ,',(third x)))
								def-initarg-alist))
						  class-options))
			    `(list . ,class-options)))))
	(collect-body
	  (collect add-form)
	  ;; Note: in COMPILE-FILE, need to dump the form to the XLD file first 
	  ;; before doing the compile-time evaluation because ADD-NAMED-CLASS does a 
	  ;; destructive modification of the slot description for slots with :ALLOCATION :CLASS.
	  (when (and (compiler:environment-remote-p environment)
		     compiler:undo-declarations-flag)
	    (collect `(eval-when (compile)
			(,@add-form
			  ',environment))))
	  (join (generate-accessors-and-readers self class-name
						slot-description-list)))))))



(p-defgeneric-options  parse-slot-options (class slot-options class-name))


(defmethod (standard-class :parse-slot-options) ( slot-options class-name)
  (MULTIPLE-VALUES-COLLECT-BODY (a b)
    (dolist (slot slot-options )
      (multiple-value-bind (a b)
	  (make-slot-description self slot  class-name)
	(collect-into a a)
	(collect-into b b )))))

(p-defgeneric-options make-slot-description (class slot class-name)
		      (declare (values slot-description form)))


;;;Note: initform should be a closure in order to capture the lexical environment. 
;;;We need to know when the lexical environment is null so we can optimize the closure out.
;;;Knowing if the initform is a constant allows for some more optimization is the instance creation.
(defmethod (standard-class :make-slot-description) (slot  class-name)
  ;;  8/17/88 DNG - Added support for :DOCUMENTATION option. Default :TYPE to T.
  ;;  9/12/88 DNG - Fix initclosure optimization.
  ;;  9/21/88 DNG - Fix initclosure optimization for macro defined in same file.
  ;;		Record documentation in slot description.
  ;; 11/03/88 DNG - Add check for invalid slot name.
  ;; 11/17/88 DNG - Cleaned up handling of options.
  ;;  4/26/89 DNG - Add warning for misplaced option keyword.
  ;;  5/06/89 DNG - Add warnings for invalid type specifier and initform inconsistent with type.
  (let (slot-name slot-options reader-list writer-list 
	reader writer accessor initform (initform-provided-p nil)
	(allocation :instance) (allocation-provided-p nil)
	(type t) (type-provided-p nil)
	initarg (initargs nil)
	documentation)
    (if (consp slot)
	(setf slot-name (car slot)
	      slot-options (rest slot))
	(progn (when (and compiler:qc-file-in-progress
			  (member slot '(:accessor :reader :writer :initarg :initform) :test #'eq))
		 ;; Probably missing a left parenthesis; a common mistake.
		 (compiler:warn 'make-slot-description ':implausible
				"Option keyword ~S found where a slot name or slot specifier list was expected.
Check your parentheses." slot))
	       (setf slot-name slot slot-options nil)))
    (unless (symbolp slot-name)
      (non-fatal-error ':impossible "Non-symbol ~S found where a slot name was expected."
			slot-name))
    (do ((slot-options slot-options (cddr slot-options)))
	((endp slot-options))
      (with-pattern-matching slot-options
	 ((`(:accessor ,accessor . ,ignore) t
	   (check-type accessor symbol)
	   (push accessor reader-list)
	   (push `(setf ,accessor) writer-list))
	  (`(:reader ,reader . ,ignore) t
	   (check-type reader symbol)
	   (push reader reader-list))
	  (`(:writer ,writer . ,ignore) t
	   ;(check-type writer symbol)
	   (push writer writer-list))
	  (`(:initform ,initform . ,ignore ) t
	   (when initform-provided-p
	     (non-fatal-error ':impossible "Option ~S provided more than once for slot ~S."
			       (car slot-options) slot-name))
	   (setf initform-provided-p  t))
	  (`(:allocation ,allocation . ,ignore ) t
	   (when allocation-provided-p
	     (non-fatal-error ':impossible "Option ~S provided more than once for slot ~S."
			       (car slot-options) slot-name))
	   (unless (member allocation '(:instance :class))
	     (non-fatal-error ':impossible
		   "Invalid value ~S for ~S option."
		   (second slot-options) (car slot-options)))
	   (setf allocation-provided-p t))
	  (`(:type ,type . ,ignore ) t
	   (when type-provided-p
	     (non-fatal-error ':impossible "Option ~S provided more than once for slot ~S."
			       (car slot-options) slot-name))
	   (setf type-provided-p  t)
	   (unless (or (symbolp type)
		       (not compiler:qc-file-in-progress))
	     ;; Warn if not a valid type.
	     (compiler:canonicalize-type-for-compiler type slot-name)
	   ))
	  (`(:initarg ,initarg . ,ignore) t
	   (assert (typep initarg 'symbol) nil ":INITARG ~s should be a symbol" initarg )
	   (push initarg initargs))
	  (`(:documentation ,documentation . ,ignore ) t)
	  (t (non-fatal-error ':impossible "Unrecognized slot option ~s" (car slot-options))))))
    (when (eq allocation :class)
      (setq allocation (list :class class-name)))
    (when (and initform-provided-p
	       (not (eq type 't))
	       compiler:qc-file-in-progress
	       (constantp initform)
	       (sys:type-specifier-p type nil)
	       (not (typep (eval initform) type)))
      (non-fatal-error ':ignorable-mistake
		       "The init form for slot ~S is ~S which is inconsistent with its type declaration of ~S."
		       slot-name initform type))
    (values
      (internal-make-slot-description :name slot-name
				      :writers writer-list
				      :readers reader-list
				      :initform (and initform-provided-p  (list initform))
				      :allocation allocation
				      :type type
				      :initargs initargs
				      )
      (let ((opts '()))
	(when (and documentation (record-documentation-p))
	  (push-key-and-value :documentation documentation opts))
	(unless (null initargs)
	  (push-key-and-value :initargs `',initargs opts))
	(unless (eq type 't)
	  (push-key-and-value :type `',type opts))
	(unless (eq allocation ':instance)
	  (push-key-and-value :allocation `',allocation opts))
	(when initform-provided-p
	  (unless (and (constantp initform)
		       (not (member initform '(nil t 0))))
	    (push-key-and-value :initclosure `(freeze ,initform) opts))
	  (push-key-and-value :initform `',(list initform) opts))
	(unless (null reader-list)
	  (push-key-and-value :readers `',reader-list opts))
	(unless (null writer-list)
	  (push-key-and-value :writers `',writer-list opts))
	`(internal-make-slot-description :name ',slot-name . ,opts)
	))))

(defmacro freeze (exp &environment env)
  ;; Return a function that, when called, will evaluate EXP.
  ;; Might be better to do this optimization in COMPILER:BREAKOFF ?
  ;;   9/23/88 DNG - Original.
  ;;  10/05/88 DNG - Add use of ALWAYS.
  ;;   4/13/89 DNG - Pass environment to DECLARED-DEFINITION so that it sees local macros.
  ;;   4/20/89 DNG - Fixed to handle local functions correctly.
  (let ((expr (macroexpand exp env)))
    (cond ((cdr (assoc (if (and (consp expr) (eq (car expr) 'quote))
			   (second expr)
			 expr)
		       '((nil . #'false)
			 (t   . #'true)
			 (0   . #'zero)))))
	  ((and (consp expr) (null (cdr expr))
		(symbolp (car expr)))
	   ;; function call with no arguments
	   (let ((def (compiler:declared-definition (car expr) env)))
	     (if (eq def (compiler:declared-definition
			   (car expr) (compiler:env-global-env env)))
		 `(quote ,(car expr)) ; global function
	       `(function ,(car expr)))))
	  ((constantp expr)
	   `(always ,expr))
	  (t `(function (lambda () ,exp))))))

(defun zero () 0)
(defun always (value)
  "Return a function that returns VALUE when it is called."
  #'(lambda () value))

;;;(defun generate-constructors (class)
;;;  (warn "~%~S not Implemented Yet" 'generate-constructors )
;;;  nil)


;;; GBK 4/24/89 Added the two following functions and rewrote
;;;             (standard-class :generate-accessors-and-readers)
;;;             to call RECORD-SOURCE-FILE-NAME.

(DEFUN make-reader (name slot-name class-name)
  (ensure-generic-fn name '(instance))
  (LET ((method-spec `(method ,name (,class-name))))
    (WHEN (SYS:RECORD-SOURCE-FILE-NAME method-spec)
      (add-method (FDEFINITION name)
		  (make-instance
		    'standard-reader-method
		    :slot-name slot-name
		    :qualifiers nil
		    :parameter-specializers
		      (list (class-named class-name))
		    :function
		      (compile nil
			       `(sys:named-lambda
				  (,method-spec (instance)
				   (slot-value instance ',slot-name))))
		    :derivation
		      `((standard) (,method-spec) nil))))))

(DEFUN make-writer (name slot-name class-name)
  (declare (special *t-class*))			; defined later in file "built-in-classes"
  (ensure-generic-fn name '(x instance))
  (LET ((method-spec `(method ,name (t ,class-name))))
    (WHEN (SYS:RECORD-SOURCE-FILE-NAME method-spec)
      (add-method (FDEFINITION name)
		  (make-instance
		    'standard-writer-method
		    :slot-name slot-name
		    :qualifiers nil
		    :parameter-specializers
		      (list *t-class* (class-named class-name))
		    :function
		      (compile nil
			       `(sys:named-lambda
				  (,method-spec) (value instance)
				   (SETF (slot-value instance ',slot-name) value)))
		    :derivation
		      `((standard) (,method-spec) nil))))))

(p-defgeneric-options  generate-accessors-and-readers (class class-name slot-descriptions))

(defmethod (standard-class :generate-accessors-and-readers) (class-name slot-descriptions)
  (collect-body
    (dolist (slot-description slot-descriptions )
      (let ((slot-name (slot-name slot-description)))
	(dolist (reader (slot-readers slot-description))
	  (collect `(make-reader ',reader ',slot-name ',class-name)))
	(dolist (writer (slot-writers slot-description))
	  (collect `(make-writer ',writer ',slot-name ',class-name)))))))
  

(p-defgeneric-options class-options (class))
;;regenerate the class options from the internal structure.
;; This is only used by the object editor.

(defmethod (standard-class :class-options) ()
  (let ((plist (class-description-plist class-description)))
    `((:documentation ,(getf plist :documentation nil))
      (:metaclass ,(class-name (class-of self))) ; formerly (getf plist :metaclass 'standard-class))
      (:default-initargs ,@(mapcan #'(lambda (v) (list* (first v) (third v) nil))
				    (getf plist :default-initargs nil))))))


;;Add-named-class & related.

;; Used in the expansion of DEFCLASS.
;; 11/23/88 DNG - Original.
(defun add-named-class-internal (metaclass class-name &optional direct-supers 
				 direct-slots class-options environment)
  (let ((class-prototype (class-prototype (class-named metaclass nil environment))))
    (if (fboundp 'ADD-NAMED-CLASS)
	(ADD-NAMED-CLASS class-prototype :name class-name
			      :superclass-names direct-supers :slot-specifications direct-slots 
			      :direct-options class-options :environment environment)
      (send class-prototype :add-named-class class-name direct-supers 
	    direct-slots class-options environment))))

;;  9/27/88 DNG - Added environment argument.
;; 10/21/88 DNG - Redesigned using CLOS:CLASS-FOR-REDEFINITION and CREATE-CLASS .
;; 10/28/88 DNG - Remove :ENVIRONMENT argument from  REINITIALIZE-INSTANCE.
;; 11/01/88 DNG - Added call to DEFFLAVOR1 when defining a metaclass.
;; 11/23/88 DNG - Use VALID-SUPERCLASS-P.
(defmethod (standard-class :add-named-class) (class-name direct-supers direct-slots class-options &optional environment)
  ;; call FIND-CLASS first in order to get an error if the class-name argument is not a symbol.
  (let ((old-class-object (find-class class-name nil environment))
	(super-metaclass nil))
    (dolist (super-name direct-supers)
      (let ((super-class (class-named super-name t environment)))
	(cond ((null super-class))
	      ((and (class-composed-p super-class)
		    (subclassp super-class '#,(class-named 'class)))
	       (setq super-metaclass super-name))
	      ((valid-superclass-p self super-class))
	      (t (non-fatal-error :probable-error "Class ~S can't inherit from class ~S."
				  class-name super-name))
	      )))
    (when (null direct-supers)
      (setq direct-supers '(standard-object))) ; is this right???
    (when (or (compiler:environment-remote-p environment)
	      (sys:RECORD-SOURCE-FILE-NAME class-name 'defclass))
      (let ((class-object
	      (if (and old-class-object ; There is a class of the same name already.
		       (or (null environment)
			   (compiler:same-environment-p (class-environment old-class-object)
							environment)))
		  ;; Try to re-use the existing class.
		  (reinitialize-instance 
		    (clos:class-for-redefinition self old-class-object
						 :direct-superclasses direct-supers
						 :direct-slots direct-slots
						 :direct-options class-options
						 :environment environment)
		    :direct-superclasses direct-supers :direct-slots direct-slots
		    :direct-options class-options)
		;; Else create a new class object.
		(create-class class-name (type-of self) direct-supers direct-slots 
			      class-options environment))))
	;; Update the slot-class-pointer to point the the class-object.
	(LET ((sys:%inhibit-read-only t))
	  (dolist (ds direct-slots)
	    (when (eq (car-safe (slot-allocation ds)) :class)
	      (setf (cadr (slot-allocation ds)) class-object))))
	(unless (eq class-object old-class-object)
	  ;; Delete any old type definitions having the same name.
	  (compiler:remprop-from-environment class-name 'sys:defstruct-description environment)
	  (compiler:remprop-from-environment class-name 'sys:flavor environment)
	  (compiler:remprop-from-environment class-name 'sys:type-expander environment)
	  ;; Install the definition.
	  (if (and super-metaclass
		   (compiler:same-environment-p environment
						compiler:*compile-file-environment*)
		   (null old-class-object)
		   (sys:RECORD-SOURCE-FILE-NAME class-name 'defclass))
	      ;; When defining a subclass of CLASS, need to also define the flavor 
	      ;; which will be the type of the class objects.  This definition has to 
	      ;; be done in the running environment in order to be able to instantiate 
	      ;; the flavor if it is used in a DEFCLASS later in the same file.  This 
	      ;; is a temporary hack until class objects are bootstrapped to really be
	      ;; standard objects.
	      (let ((sys::*just-compiling* nil))
		(setf (class-named class-name t nil) class-object)
		(sys::defflavor1 class-name '() (list super-metaclass) '()))
	    (setf (class-named class-name t environment)
		  class-object))
	  (when old-class-object
	    ;; Now, make sure that we don't find somebody that depends on old-class-object 
	    ;; in the current environment.
	    (update-subclass-cpl old-class-object environment)
	    (when (class-composed-p old-class-object)
	      (compose-class class-object))))
	class-object))))

(unless (fboundp 'valid-superclass-p)
  ;; Temporary definition until the real one is loaded.
  (setf (symbol-function 'valid-superclass-p)
	#'(lambda (&rest ignore) t)))

(comment ; this is not used
(p-defgeneric-options class-different-p (class other-class  direct-supers direct-slots class-options #| constructors |#))

(defmethod (standard-class :class-different-p)
		(other-class  direct-supers direct-slots class-options #| constructors |#)
  (declare (ignore other-class  direct-supers direct-slots class-options #| constructors |#))
  ;;Test if main slots are the same
  (format t "~%class-different-p not implemented yet" )
  t) )

;;Class-for-redefinition & related
(comment ; not used anymore - DNG 10/28/88
(p-defgeneric-options class-for-redefinition
		      (class other-class class-name direct-supers direct-slots class-options #| constructors |#))
(defmethod (standard-class :class-for-redefinition)
		(other-class class-name direct-supers direct-slots class-options #| constructors |#)
  (declare (ignore class-name))
  (update-class other-class direct-supers direct-slots class-options #| constructors |#)
  other-class)
)

(p-defgeneric-options update-class
		      (class  direct-supers direct-slots class-options &optional environment))



(defun reinit-effective-slots (class new-slots temp-class)
  ;;We need to reinit the slots like the new-slots
  (collect-body 
    (dolist (s (class-all-slots class))
      (let ((new-slot (find (slot-name s) new-slots :key #'slot-name :test #'eq)))
	(when new-slot
	  (setf (slot-type s) (slot-type new-slot)
		(slot-allocation s) (subst class temp-class (slot-allocation new-slot))
		(slot-writers s) (slot-writers new-slot)
		(slot-readers s) (slot-readers new-slot)
		(slot-initclosure s) (slot-initclosure new-slot)
		(slot-initform s) (slot-initform new-slot)
		(slot-initargs s) (slot-initargs new-slot))
	  (collect s)
	  (setf new-slots (delete new-slot new-slots :count 1 :test #'eq)))))
    (join new-slots)))

;;; This method needs to be overhauled to reflect the new class-redefinition protocol.
;; Used by UPDATE-CLASS for ADD-NAMED-CLASS.
;; 10/26/88 DNG - Pass environment to compute-class-precedence-list.
;; 12/20/88 DNG - Fix to update the precedence list of the new class 
;;		description instead of the old one.
(defun re-init-class (class temp-class-description )
  (declare (special temp-class-description))
  (let* ((temp-class (sys:make-flavor-instance (class-name (class-of class))
					       :class-description temp-class-description))
	 (old-cpl (internal-class-precedence-list class))
	 (old-class-description (class-description class))
	 (environment (class-description-environment old-class-description))
	 (new-cpl (loop for super in (compute-class-precedence-list temp-class environment)
			collect (class-named super nil environment))))
    ;;check if the class need to be made obsolete
    (setf (class-description-class-precedence-list temp-class-description) new-cpl)
    (let ((new-slots (collect-slot-descriptions temp-class)))
      (flet ((instance-slot-names (slots)
		(collect-body (dolist (s slots)
				(when (eq :instance (slot-allocation s))
				  (collect (slot-name s)))))))
	(if (set-exclusive-or (instance-slot-names (class-all-slots class))
			      (instance-slot-names  new-slots))
	    
	    (progn
	      ;;the class needs to be made obsolete
	      (make-instances-obsolete class)
	      (setf (class-all-slots class) nil))
	    (progn
	      (setf (class-all-slots class)
		    (reinit-effective-slots class new-slots temp-class))))
	(setf (class-description-class-precedence-list (class-description class))
	      (cons class (cdr new-cpl)))
	(unless (equal (cdr old-cpl) (cdr new-cpl))
	  ;;All the combined method are to be recomputed
	  ;;However, how can we make sure it is done more than it needs to be?
	  (dolist (gf (specializer-direct-generic-functions class))
	    (uncache-effective-methods gf class)))
	class))))

;; 10/28/88 DNG - Fixed to update plist when adding a new class option.
;; 12/01/88 DNG - Just mark it not composed and not finalized instead of 
;;		calling COMPOSE-CLASS and FINALIZE-INHERITANCE now.
;; 12/05/88 DNG - Reinstate call to COMPOSE-CLASS because use of old MAKE-INSTANCE method avoids calling it.
;; 12/14/88 DNG - Reinstate call to FINALIZE-INHERITANCE for same reason.
(defmethod (standard-class :update-class)
		(direct-supers direct-slots class-options
		 &optional (environment (class-description-environment class-description)))
  ;;We need a way to return the obsolete class
  ;;We probably need to record the fact that the obsolete class is replaced by the new one.
  ;;This function is a critical section.
  (flet ((update ()
		 ;;update the class
		 (setf (class-direct-supers self) (or direct-supers '(standard-object)))
		 (setf (class-direct-slots self)  direct-slots)
		 (do ((co  class-options (cddr co))
		      (plist (class-description-plist class-description)))
		     ((null co)
		      (setf (class-description-plist class-description) plist))
		   (setf (getf plist (first co)) (second co)))
		 (values)))
      (if (class-composed-p self)
	  (let* ((temp-class-description
		   (make-class-description :name (class-name self)
					   :direct-supers (or direct-supers '(standard-object))
					   :direct-slots direct-slots
					   :direct-subclasses (class-direct-subclasses self )
					   :class-precedence-list (class-precedence-list self)
					   :class-object nil
					   :plist class-options))
		 (old-direct-supers (class-direct-supers self)))
	    (re-init-class self temp-class-description)
	    (update)
	    ;;Now take care of the superclasses we removed from what we had before.
	    (dolist (old old-direct-supers)
	      (unless (member old direct-supers :test #'eq)
		(remove-direct-subclass (class-named old nil environment) self)))
	    (compose-class self)
	    (when (internal-class-finalized-p self)
	      ;; Need to re-finalize because there could be a specialized 
	      ;; MAKE-INSTANCE method which does not check for finalization.
	      (finalize-inheritance-internal self :slots t :methods t :initargs t))
	    (dolist (c (class-direct-subclasses self))
	      (supers-changed c))
	    self)
	(update)))
  self)

;; 12/20/88 DNG - Moved MAKE-INSTANCES-OBSOLETE to file "CLASS-REDEFINITION" 
;;		so that it can be a real generic function now.
(comment ; old definition
(def make-instances-obsolete)
(p-defgeneric-options make-instances-obsolete (class)
		      (:documentation 
 "This generic function is called when a class has been redefined such that the
list of slot names is different.  It constructs a new class description and 
marks the old one obsolete so that the next time an attempt is made to access 
one of the slots in an old instance, UPDATE-INSTANCE-FOR-REDEFINED-CLASS will 
be called to update the instance."))


;; 11/30/88 DNG - Add code to remove MAKE-INSTANCE and SHARED-INITIALIZE methods.
(defmethod (standard-class :make-instances-obsolete)()
  (let* ((new-class-description
	   (or (and (boundp 'temp-class-description)
		    temp-class-description)
	       (sys:copy-object class-description )))
	 (obsolete-class-description (class-description self))
	 (obs-mapped-supers (class-mapped-supers self ))
	 (obs-mapped-slots (class-mapped-slot-names self )))
    (declare (special temp-class-description))
    ;;make sure that we don't share the plist with the obsolete class description
    (setf (class-description-plist new-class-description) nil)
    (obsolete-mapping-table self)
    (uncache-accessors self)
    (setf (class-description self) new-class-description)
    (set-obsolete-class-description obsolete-class-description )
    (setf (class-description-class-object new-class-description ) self )
    ;;transfer information from obsolete to new class-description
    (setf (class-mapped-supers self) obs-mapped-supers )
    (setf (class-mapped-slot-names self) obs-mapped-slots )
    ;;clear mapping tables slot so we don't pick them by mistake.
    (setf (class-mapping-tables self) nil)
    ;; Make sure that we don't call an obsolete method that was automatically 
    ;; generated from the old class description.
    (let ((old-method (find-method #'make-instance nil (list `(eql ,self)) nil)))
       (when (and old-method (getf (method-plist old-method) 'computed-method))
	 (remove-method #'make-instance old-method)))
    (let ((old-method (find-method #'shared-initialize nil (list self *t-class*) nil)))
       (when (and old-method (getf (method-plist old-method) 'computed-method))
	 (remove-method #'shared-initialize old-method)))
    nil))
)



;; fix Class Precedence List (used only by ADD-NAMED-CLASS)
(defun update-subclass-cpl (class environment)
  (dolist (sub (class-direct-subclasses class))
    (if (and (class-composed-p sub)
	     (compiler:same-environment-p (class-environment sub) environment))
	(compose-class sub)
      (update-subclass-cpl sub environment)))
  (values))


(p-defgeneric-options supers-changed (class &rest rest))

(defmethod (standard-class :supers-changed) (&rest ignore)
  (when (internal-class-precedence-list self)
    (let* ((temp-class-description (sys:copy-object class-description)))
      (re-init-class self temp-class-description))
    ;;Now take care of the superclasses we removed from what we had before.
    (compose-class self)
    (when (internal-class-finalized-p self)
      (finalize-inheritance-internal self :slots t :methods t :initargs t)))
  ;;We must do something else for compile evironment
  ;;If a subclass is in the other environment, create a new class in the current environment.
  (dolist (sub (class-direct-subclasses self ))
    (supers-changed sub)))

;; 5/8/89 DNG - moved DELETE-CLASS to the "extensions" file.
