;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


(p-defgeneric-options compose-class (class))

;; 10/26/88 DNG - Updated to use CLASS-DESCRIPTION-ENVIRONMENT.
;; 11/03/88 DNG - Return no values.
(ticl:defmethod (standard-class :compose-class) ()
  ;;This thing should do all of what flavor does:
  ;;Come up with the instance size...
  (let ((environment (class-description-environment class-description)))
    ;;The next two operations must be done atomically
    (unless (class-description-class-precedence-list class-description)
      (setf (class-description-class-precedence-list class-description)
	    (loop for super in (compute-class-precedence-list self environment)
		  collect (class-named super nil environment))))
    (dolist (class (class-description-class-precedence-list class-description))
      (dolist (super (class-direct-supers class))
	(add-direct-subclass (class-named super nil environment) class nil)))
    )
  (unless (class-all-slots self)
    (setf (class-all-slots self)
	  (collect-slot-descriptions self)))
  (build-class-slots self)
  ;;Compute the instance size
  (setf (class-instance-size self) (compute-instance-size self))
  (values))


(p-defgeneric-options add-direct-subclass (class subclass &optional errorp))

(ticl:defmethod (class :add-direct-subclass ) (subclass &optional (errorp t))
  (if (and errorp (member subclass (class-description-direct-subclasses class-description) :test #'eq))
      (error "This subclass ~S is already a subclass of ~S" subclass self)
      (pushnew subclass (class-description-direct-subclasses class-description))))

(p-defgeneric-options remove-direct-subclass (class subclass &optional errorp))

(ticl:defmethod (class :remove-direct-subclass ) (subclass &optional (errorp t))
  (if (and errorp (not (member subclass (class-description-direct-subclasses class-description) :test #'eq)))
      (error "This subclass ~S is not a subclass of ~S" subclass self)
      (setf (class-description-direct-subclasses class-description)
	    (delete  subclass (class-description-direct-subclasses class-description)
		     :test #'eq :count 1))))

;; 10/20/88 DNG - Added :TEST #'EQ to ASSOC and MEMBER calls.
;;  5/05/89 DNG - Added :TEST #'EQ to UNION call.
(defun compute-class-precedence-list (root &optional environment)
  (let ((cpl ())
	(must-precede-alist ()))
    ;; We start by computing two values.
    ;;   CPL
    ;;     The depth-first left-to-right up to joins walk of the supers tree.
    ;;   MUST-PRECEDE-ALIST
    ;;     An alist of the must-precede relations. The car of each element of the
    ;;;    must-precede-alist is a class, the cdr is all the classes which must
    ;;     precede that class in the CPL.
    ;;
    ;; Note that the actual walk is breadth-first, right-to-left.
    ;;
    (labels ((walk-supers (class &optional precedence)
	         (let ((elem (assoc class must-precede-alist :test #'eq)))
		   (if elem
		       (setf (cdr elem) (union (cdr elem) precedence :test #'eq))
		     (push (cons class precedence) must-precede-alist)))
			  (let* ((rsupers (reverse (cons class (class-direct-supers
								 (class-named class nil environment)))))
				 (precedence (cdr rsupers)))
			    (do ((sup rsupers (cdr sup))
				 (pre precedence (cdr pre)))
				((null pre))
			      (walk-supers (car sup) pre)))
			  (unless (member class cpl :test #'eq) (push class cpl)))
	     (must-move-p (element list &aux move)
	         (dolist (must-precede (cdr (assoc element must-precede-alist :test #'eq)))
		   (when (setq move (member must-precede (cdr list) :test #'eq))
		     (return move))))
	     (find-farthest-move (element move)
	         (let ((closure (compute-must-precedes-closure element)))
		   (dolist (must-precede closure)
		     (setq move (or (member must-precede move :test #'eq) move)))
		   move))
	     (compute-must-precedes-closure (class)
	         (let ((closure ()))
		   (labels ((walk (element path)
			        (when (member element path :test #'eq)
				  (class-ordering-error
				    root element path must-precede-alist))
				  (dolist (precede (cdr (assoc element must-precede-alist
							       :test #'eq)))
				    (unless (member precede closure :test #'eq)
				      (push precede closure)
				      (walk precede (cons element path))))))
		     (walk class nil)
		     closure))))
      (walk-supers root)			;Do the walk
      ;; For each class in the cpl, make sure that there are no classes after it which
      ;; should be before it.  We do this by cdring down the list, making sure that for
      ;; each element of the list, none of its must-precedes come after it in the list.
      ;; If we find one, we use the transitive closure of the must-precedes (using
      ;; find-farthest-move) to see where the class must really be moved. We use a
      ;; hand-coded loop so that we can splice things in and out of the CPL as we go.
      (let ((tail cpl)
	    (element nil)
	    (move nil))
	(loop (when (null tail) (return))
	      (setq element (car tail)
		    move (must-move-p element tail))
	      (cond (move
		     (setq move (find-farthest-move element move))
		     (setf (cdr move) (cons element (cdr move)))
		     (setf (car tail) (cadr tail))	;OK to use Interlisp delete trick since it
		     (setf (cdr tail) (cddr tail))	;will never be the last element of the list.
		     )
		    (t
		     (setq tail (cdr tail)))))
	(copy-list cpl)))))
		  
(defun class-ordering-error (root element path must-precede-alist)
  (setq path (cons element (reverse (member element (reverse path)))))
  (let ((explanations ()))
    (do ((tail path (cdr tail)))
	((null (cdr tail)))
      (let ((after (cadr tail))
	    (before (car tail)))
	(if (member after (class-direct-supers (class-named before )))
	    (push (format nil "~% ~A must precede ~A -- ~A is in the local supers of ~A."
			  before after after before)
		  explanations)
	    (dolist (common-precede (intersection
				      (cdr (assoc after must-precede-alist))
				      (cdr (assoc before must-precede-alist))))
	      (when (member after (member before (class-direct-supers  (class-named common-precede ))))
		(push (format nil "~% ~A must precede ~A -- ~A has local supers ~S."
			      before after
			      common-precede
			      (class-direct-supers  (class-named common-precede )))
		      explanations))))))
    (error "While computing the class-precedence-list for the class ~A:~%~
             There is a circular constraint through the classes:~{ ~A~}.~%~
             This arises because:~{~A~}"
	   root path (reverse explanations))))
		  



(p-defgeneric-options collect-slot-descriptions (class))

(ticl:defmethod (standard-class :collect-slot-descriptions) ()
  (let ((slot-alist nil))
    (dolist (class (class-precedence-list self))
      (dolist (slot (class-direct-slots class ))
	(let ((value (assoc (slot-name slot ) slot-alist :test #'eq)))
	  (if value
	      (nconc value (cons slot nil))
	      (setf slot-alist (nconc slot-alist (cons (cons (slot-name slot) (cons slot nil)) nil)))))))
    (collect-body
      (dolist (slot-acons slot-alist)
	(collect (compute-effective-slot-description self (cdr slot-acons)))))))
    

;;We need to come up with the real slot description, i.e we have to update (or create) a new
;;slot description where all the values are computed, like allocation and type. This needs to 
;;be done even if we have only one slot description.

(p-defgeneric-options compute-effective-slot-description (class slot-list))

;;  8/17/88 DNG - Cleaned up calculation of :TYPE to yield T instead of (AND T).
(ticl:defmethod (standard-class :compute-effective-slot-description) (slot-list)
  (declare (list slot-list))
  (internal-make-slot-description
    :name (slot-name (first slot-list))
    :writers nil ;These are not relevant for an effective slot.
    :readers nil   ;
    :initform  (let ((slot (car (member-if  #'non-empty-p slot-list :key #'slot-initform))))
		 (and slot (slot-initform slot)))
    :initclosure (let ((slot (car (member-if  #'non-empty-p slot-list :key #'slot-initform))))
		   (and slot (slot-initclosure slot)))		    
    :allocation (compute-slot-allocation self slot-list)
    :type (let ((res-type '()))
	    (dolist (slot slot-list (cond ((null res-type) t)
					  ((null (rest res-type)) (first res-type))
					  (t (cons 'and res-type))))
	      (let ((type (slot-type slot)))
		(when (and type (not (eq type 't)))
		  (pushnew type res-type :test #'equal)))))))

(p-defgeneric-options compute-slot-allocation (class slot-list))

(ticl:defmethod (standard-class :compute-slot-allocation) (slot-list)
  (let ((allocation (slot-allocation (car slot-list))))
    (if (eq (car-safe allocation) :class )
	`(:class ,(class-named (cadr allocation)))
	:instance)))


(ticl:defmethod (standard-class :after :set-class-all-slots) (v)
  
  (setf (class-description-instance-slot-names  class-description)
	(copy-list
	  (mapcan #'(lambda (s)
		      (and (eq :instance (slot-allocation s)) (cons (slot-name s) nil)))
		  v))))

(p-defgeneric-options compute-instance-size  (class))

(ticl:defmethod (standard-class :compute-instance-size ) (&aux (size 0))
  (dolist (slot (class-all-slots self))
    (when (eq (slot-allocation slot) :instance)
      (incf size)))
  (1+ size))

(defstruct (class-slot-structure (:type :named-array-leader) (:callable-constructors nil))
  class-description)

(p-defgeneric-options build-class-slots (class))
(defmethod (standard-class :build-class-slots) ()
  (let* ((old-alist (class-class-slot-alist self))
	 (all-slots (class-all-slots self))
	 (new-class-slots (collect-body (dolist (s all-slots)
					  (when (and (eq :class (car-safe (slot-allocation s)))
						     (eq self (cadr (slot-allocation s))))
					    (collect (slot-name s))))))
	 (array (class-description-class-variable-structure class-description))
	 (size (or (and (arrayp array)
			(length array))  0)))
    (dolist (s old-alist)
      (if (member (car s) new-class-slots :test #'eq)
	  (setf new-class-slots (delete (car s) new-class-slots :test #'eq :count 1))
	  (setf (cdr s) nil)))
    (setf (class-class-slot-alist self)
	  (append old-alist
		  (collect-body
		    (dolist (s new-class-slots )
		      (collect (cons s size))
		      (incf size)))))
    (when (plusp size)
      (setf array (if array
		      (adjust-array (class-description-class-variable-structure class-description)
				    size)
		      (make-class-slot-structure :make-array (:length size) :class-description  class-description)))
      (without-interrupts
	(dolist (s (class-class-slot-alist self))
	  (unless (assoc (car s) old-alist :test #'eq)
;;The follwing is done when the first instance gets created.
;;	    (let ((sl (find (car s) all-slots :key #'slot-name :test #'eq)))
;;	      (if (non-empty-p (slot-initform sl))
;;		  (setf (aref array (cdr s))
;;			(if (slot-initclosure sl)
;;			    (funcall (slot-initclosure  sl))
;;			    (eval (car (slot-initform  sl)))))
	    (sys:%P-STORE-DATA-TYPE-AND-POINTER (locf (aref array (cdr s)))  sys:DTP-Null 0))))
      
      (setf (class-description-class-variable-structure class-description)
	    array))
    (setf (class-description-class-slot-alist  class-description)
	  (copy-list
	    (mapcan #'(lambda (s)
			(and (eq :class (car-safe (slot-allocation s)))
			     (let ((slot-class (cadr (slot-allocation s))))
			       ;;The slot-class must be composed, so the class-slot
			       ;;can exist
			       (unless (or (eql slot-class self)
					   (class-composed-p slot-class))
				 (compose-class slot-class))
			       (cons (cons (slot-name s)(class-slot-location slot-class
									     (slot-name s)))
				     nil))))
		    (class-all-slots self))))))
  
  

(p-defgeneric-options  class-slot-location (class slot-name))

(ticlos:defmethod (standard-class :class-slot-location) (slot-name)
  (locf (aref (class-description-class-variable-structure class-description)
	      (cdr (assoc slot-name (class-class-slot-alist self) :test #'eq)))))













