;;; -*- Mode:Common-Lisp; Package:Compiler; Base:10; Cold-Load:T; Patch-file: T -*-


;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights
;;; in Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.


;;;	Disassembler modifications for support of CLOS


(DEFSUBST DTP-PRINTABLE-P (DTP)
  ;; Is this a data type that can be "in the machine"?
  ;; 12/05/88 DNG - Add DTP-Lexical-Closure.
  ;;  2/07/89 DNG - Add DTP-Stack-List.
  (MEMBER DTP '( #.DTP-Symbol #.DTP-Fix #.DTP-Extended-Number #.DTP-Locative
		#.DTP-List #.DTP-Stack-List #.DTP-U-Entry #.DTP-FEF-Pointer #.DTP-Array-Pointer
		#.DTP-Closure #.DTP-Lexical-Closure #.DTP-Small-Flonum
		#.DTP-Instance #.DTP-Character #.DTP-Single-Float)
	  :TEST #'EQ))

(DEFUN DISASSEMBLE-POINTER (FEF DISP &OPTIONAL PC SUPPRESS-QUOTE-P)
 ;;  7/25/85 - Avoid trying to print illegal data types.
 ;; 11/16/85 - Use PRINC instead of PRIN1 for instance variable names.
 ;;  7/09/86 - Handle pointer to :INTERNAL function in another FEF.
 ;;  9/08/86 - New argument SUPPRESS-QUOTE-P.
 ;;  4/25/88 DNG - Updated to recognize pointers to function cells for CLOS 
 ;;		methods and (SETF ...) functions and show the right function spec.
 ;;  5/05/88 DNG - Fix 4/25 change to work right in the inspector.  Avoid 
 ;;		crashing trying to print a list of DTP-SELF-REF-POINTERs.
 ;;		Preliminary handling for CLOS instance references.
 ;;  5/09/88 DNG - Show names of CLOS instance variables.
 ;;  8/09/88 DNG - Fix handling of DTP-SELF-REF-POINTER that can't be decoded.
 ;; 11/08/88 DNG - Fix to not error on a list having a DTP-SELF-REF-POINTER as its second element.
  (LET ((LOC (%MAKE-POINTER-OFFSET DTP-LOCATIVE FEF DISP))
	(DTP (SI:%P-DATA-TYPE-OFFSET FEF DISP))
	CELL PTR OFFSET TEM)
    (COND
      ((= DTP DTP-SELF-REF-POINTER)
       (LET ((NUMBER (SI:%P-POINTER-OFFSET FEF DISP)))
	 (IF (ZEROP (%LOGLDB SYS:%%SELF-REF-TYPE-FLAG NUMBER))
	     ;; Flavors reference
	     (MULTIPLE-VALUE-BIND (PTR COMPONENT-FLAVOR-FLAG)
		 (SI:FLAVOR-DECODE-SELF-REF-POINTER
		   (OR (DISASSEMBLE-CURRENT-FLAVOR FEF PC)
		       (SI:FEF-FLAVOR-NAME FEF))
		   NUMBER)
	       (IF (NULL PTR)
		   (FORMAT T "'#<~A ~O>" 'DTP-SELF-REF-POINTER (SI:CONVERT-TO-UNSIGNED NUMBER))
		 (PROGN
		   (SETQ CELL (IF COMPONENT-FLAVOR-FLAG "mapping table for " ""))
		   (IF DISASSEMBLE-OBJECT-OUTPUT-FUN
		       (FUNCALL DISASSEMBLE-OBJECT-OUTPUT-FUN PTR CELL LOC T)
		     (PROGN
		       (PRINC CELL)
		       (PRINC PTR)
		       (WHEN (EQUAL CELL "")
			 (PRINC " in SELF")))))))
	   ;; Else CLOS reference.
	   (MULTIPLE-VALUE-BIND (NAME MAP-FLAG)
	       (DECODE-CLOS-SELF-REF-POINTER FEF NUMBER)
	     (LET ((ARG (IF (ZEROP (LDB SYS:%%CLOS-SELF-REF-INSTANCE-REF-ADDRESSING-MODE NUMBER))
			   (DISASSEMBLE-ARG-NAME
			     FEF (LDB SYS:%%CLOS-SELF-REF-INSTANCE-REF-INDEX NUMBER))
			 (DISASSEMBLE-LOCAL-NAME
			   FEF (LDB SYS:%%CLOS-SELF-REF-INSTANCE-REF-INDEX NUMBER)))))
	       (IF (AND NAME ARG)
		   (FORMAT T (IF MAP-FLAG "map for ~S in ~A" "slot ~S in ~A")
			   NAME ARG)
		 (FORMAT T "'#<~A ~O>" 'DTP-SELF-REF-POINTER (SI:CONVERT-TO-UNSIGNED NUMBER)))
	   )))))
      ((= DTP DTP-EXTERNAL-VALUE-CELL-POINTER)
       (SETQ PTR (%FIND-STRUCTURE-HEADER (SETQ TEM (%P-CONTENTS-AS-LOCATIVE-OFFSET FEF DISP)))
	     OFFSET (%POINTER-DIFFERENCE TEM PTR))
       (LET (NAME)
	 (COND
	   ((SYMBOLP PTR)
	    (SETQ CELL (NTH OFFSET '("@+0?? " "" "#'" "@PLIST-HEAD-CELL " "@PACKAGE-CELL ")))
	    (SETQ NAME PTR))
	   ((CONSP PTR)
	    (COND ((AND (CONSP (CAR PTR))
			(SYMBOLP (CAAR PTR))
			(GET (CAAR PTR) 'FUNCTION-SPEC-HANDLER))
		   ;; The list could be a flavor "meth" list, or a CLOS
		   ;; "method-spec-object", or a SETF-GENERIC-FUNCTION property.
		   ;; In all of these cases, the function spec is the first
		   ;; element of the list.
		   (SETQ NAME (CAR PTR)))
		  ;; If the cell holds a named function, use its name.
		  ((AND (FUNCTIONP (NTH OFFSET PTR) T)
			(SETQ NAME (FUNCTION-NAME (NTH OFFSET PTR)))))
		  ;; Else can't figure out the name for this cell.
		  (T (SETQ NAME `(:LOCATION ,TEM))))
	    (SETQ CELL "#'"))
	   ((COMPILED-FUNCTION-P PTR) ; probably an :INTERNAL function in the parent FEF
	    (RETURN-FROM DISASSEMBLE-POINTER
	      (DISASSEMBLE-POINTER PTR OFFSET NIL)))
	   (T (SETQ CELL "" NAME PTR)))
	 (IF DISASSEMBLE-OBJECT-OUTPUT-FUN
	     (FUNCALL DISASSEMBLE-OBJECT-OUTPUT-FUN NAME CELL LOC T)
	   (PROGN
	     (PRINC CELL)
	     (PRIN1 NAME)))))
      ((NOT (DTP-PRINTABLE-P DTP))
       ;; Data type that cannot be "in the machine"; don't try to print it.
       (FORMAT T "#<~A ~O>" (OR (NTH DTP Q-DATA-TYPES) DTP)
	       (SI:%P-POINTER-OFFSET FEF DISP)))
      (T
       (LET ((VALUE (CONTENTS LOC)))
	 (SETQ CELL (IF (AND SUPPRESS-QUOTE-P
			     (OR (NUMBERP VALUE) (KEYWORDP VALUE)))
			""
		      "'"))
	 (IF (AND (CONSP VALUE)
		  (OR (NOT (DTP-PRINTABLE-P (SI:%P-DATA-TYPE-OFFSET VALUE 0)))
		      (AND (EQL (SI:%P-CDR-CODE-OFFSET VALUE 0) SI:CDR-NEXT)
			   (NOT (DTP-PRINTABLE-P (SI:%P-DATA-TYPE-OFFSET VALUE 1))))))
	     ;; Careful handling of lists of DTP-SELF-REF-POINTERs used by CLOS.
	     (PROGN (PRINC CELL)
		    (WRITE-CHAR #\()
		    (DO ((TAIL VALUE (CDR TAIL)))
			((NULL TAIL))
		      (LET ((DTP (SI:%P-DATA-TYPE-OFFSET TAIL 0)))
			(IF (DTP-PRINTABLE-P DTP)
			    (IF DISASSEMBLE-OBJECT-OUTPUT-FUN
				(FUNCALL DISASSEMBLE-OBJECT-OUTPUT-FUN
					 (CAR TAIL) "" (LOCF (CAR TAIL)) NIL)
			      (PRIN1 (CAR TAIL)))
			  (FORMAT T "#<~A ~O>" (NTH DTP Q-DATA-TYPES)
				  (SI:CONVERT-TO-UNSIGNED (SI:%P-POINTER-OFFSET TAIL 0)))))
		      (UNLESS (NULL (CDR TAIL)) (WRITE-CHAR #\SPACE)) )
		    (WRITE-CHAR #\)))
	   (IF DISASSEMBLE-OBJECT-OUTPUT-FUN
	       (FUNCALL DISASSEMBLE-OBJECT-OUTPUT-FUN VALUE CELL LOC NIL)
	     (PROGN
	       (PRINC CELL)
	       (PRIN1 VALUE)))))))))


(defun decode-clos-self-ref-pointer (FEF pointer-number)
  "Decode the pointer field of a DTP-SELF-REF-POINTER.
Values are a slot name and NIL, or a component class name and T."
  ;;  5/09/88 DNG - Original (adapted from FLAVOR-DECODE-SELF-REF-POINTER).
  ;;  2/21/89 DNG - Fix to handle (EQL #) specializers.
  (declare (values instance-var-or-component-class t-if-component-class))
  (let* ((LOCAL-SLOT (LDB SYS:%%CLOS-SELF-REF-MAPPING-TABLE-LOCAL-INDEX POINTER-NUMBER))
	 (ARG-SLOT (IF (= LOCAL-SLOT SYS:LOCAL-FOR-FIRST-MAPPING-TABLE)
		       0
		     (- LOCAL-SLOT (- SYS:LOCALS-FOR-MAPPING-TABLE-BASE 1))))
	 (CLASS-NAME (AND (>= ARG-SLOT 0)
			  (NTH ARG-SLOT (FUNCTION-SPECIALIZERS FEF)))))
    (UNLESS (OR (NULL CLASS-NAME)
		(NOT (FBOUNDP 'ticlos:class-named)))
      (LET ((class-object (if (ticlos:individual-typep class-name)
			      (ticlos:class-of (ticlos:individual-type class-name))
			    (ticlos:class-named class-name t)))
	    (offset (ldb sys:%%CLOS-SELF-REF-SLOT-OFFSET pointer-number)))
	(cond
	  ((null class-object) nil)
	  ((ldb-test sys:%%CLOS-SELF-REF-MAP-LEADER-FLAG pointer-number)
	   (values (ticlos:class-name (nth offset (ticlos:class-mapped-supers class-object)))
		   t))
	  ((ldb-test sys:%%CLOS-SELF-REF-RELOCATE-FLAG pointer-number)
	   (nth offset (ticlos:class-mapped-slot-names class-object)))
	  (t NIL))))))

(DEFUN FUNCTION-SPECIALIZERS (FCT)
  ;; Given a function, which should be for a CLOS method, return the list of class names.
  ;;  2/10/89 DNG - Add use of debug info if name is not a method.
  (LET ((FNAME (FUNCTION-NAME FCT)))
    (IF (EQ (CAR-SAFE FNAME) 'TICLOS:METHOD)
	(CAR (LAST FNAME))
      (GET-DEBUG-INFO-FIELD (GET-DEBUG-INFO-STRUCT FCT) 'ARG-CLASSES)
      )))


(DEFUN DISASSEMBLE-CALL-INFO-WORD (CALL-INFO)
  ;;  9/06/86 DNG - Original.
  ;;  8/09/88 DNG - Added recognition of the CLOS bit.
  (WHEN (PLUSP (LDB (SYMEVAL-FOR-TARGET 'SI:%%call-info-lexpr-funcall-flag) CALL-INFO))
    (PRINC 'APPLY) (WRITE-CHAR #\SPACE))
  (FORMAT T "~A arg~:P, "
	  (LDB (SYMEVAL-FOR-TARGET 'SI:%%call-info-number-of-arguments) CALL-INFO))
  (WHEN (PLUSP (LDB (SYMEVAL-FOR-TARGET 'SI:%%call-info-self-map-table-provided) CALL-INFO))
    (PRINC "self-map, "))
  (WHEN (PLUSP (LDB (SYMEVAL-FOR-TARGET 'si:%%call-info-clos-info-provided) CALL-INFO))
    (PRINC "next-method-list, maps, "))
  (LET ((RETURN-TYPE (LDB (SYMEVAL-FOR-TARGET 'SI:%%call-info-return-type) CALL-INFO)))
    (DECLARE (FIXNUM RETURN-TYPE))
    (COND ((EQL RETURN-TYPE (SYMEVAL-FOR-TARGET 'SI:%only-one-result-needed))
	   (PRINC "1 value"))
	  ((EQL RETURN-TYPE (SYMEVAL-FOR-TARGET 'SI:%normal-return))
	   (FORMAT T "~A values"
		   (LDB (SYMEVAL-FOR-TARGET 'SI:%%call-info-number-of-results) CALL-INFO)))
	  ((EQL RETURN-TYPE (SYMEVAL-FOR-TARGET 'SI:%multiple-value-list-return))
	   (PRINC 'multiple-value-list))
	  ((EQL RETURN-TYPE (SYMEVAL-FOR-TARGET 'SI:%return-all-values-with-count-on-stack))
	   (PRINC "return values and count")))))


(DEFUN DISASSEMBLE-ONE-INSTRUCTION (WD &OPTIONAL SECOND-WORD FEF PC)
  "Print on STANDARD-OUTPUT the disassembly of the instruction."
  ;;  8/10/85 DNG - Fix to correctly display CEILING, TRUNCATE, and ROUND. [SPR 233]
  ;;  8/21/85 DNG - Fix to not error on old-style long branch in V2 mode.
  ;;  9/17/85 DNG - Display PC in default base instead of octal.
  ;; 10/23/85 DNG - Change names ADD-IMMEDIATE etc. to ADD-IMMED etc.
  ;; 11/09/85 DNG - Fix to show correct variable name for STACK-CLOSURE-UNSHARE.
  ;; 11/23/85 DNG - Support module-op instructions.
  ;; 12/09/85 CLM - Support the new aux-op long branches; changed to print PDL-PUSH
  ;;                instead of PDL-POP in those cases where instruction DEST property
  ;;                equals D-STORE and the displacement equals #o777.
  ;; 12/18/85 CLM - For Rel.3, added code to print the contents of a dispatch table
  ;;                in the comment field.
  ;;  1/09/86 DNG - Re-design to key off of NO-REG and DISP properties instead of name.
  ;;  1/20/86 CLM - Corrected long-branch addressing to absolute instead of relative.
  ;;  2/17/86 DNG - Use MISC-OP-NAME-TABLE for VM2 native mode.
  ;;  7/11/86 DNG - Modify to show "(MISC) PUSH OP" instead of "(MISC) OP D-PDL" etc.
  ;;  7/16/86 DNG - Show name of variable accessed by "higher-context" instructions.
  ;;  9/06/86 DNG - Add special handling for SELECT and complex call instructions.
  ;;  2/10/87 CLM - Fixed to handle UNBIND's and POP-PDL's of greater than 16 (SPR 3111).
  ;;  1/18/88 DNG - Enable displaying call-info description on COMPLEX-CALL to %FUNCTION-INSIDE-SELF.
  ;; 11/28/88 DNG - Enable displaying call-info when CALL-NEXT-METHOD used PUSH-CAR for the function.
  (LET ( OP SUBOP DEST DISP REG )
    (BLOCK NIL
      (SETQ OP    (LDB (BYTE 4  9) WD)
	    SUBOP (LDB (BYTE 3 13) WD)
	    DEST  (LDB (BYTE 2 14) WD)
	    DISP  (LDB (BYTE 9  0) WD)
	    REG   (LDB (BYTE 3  6) WD))
      (WHEN SECOND-WORD
       ;; If a two-word insn has a source address, it must be an extended address,
       ;; so set up REG and DISP to be right for that.
	(UNLESS (OR (COMPILING-FOR-V2)
		    (= OP #o14))
	  (SETQ REG (LDB (BYTE 3 6) SECOND-WORD)
		DISP (DPB (LDB (BYTE 4 9) SECOND-WORD)
			  (BYTE 4 6)
			  (LDB (BYTE 6 0) SECOND-WORD)))))
      (WHEN (< OP #o11)
	(SETQ OP (LDB (BYTE 5 9) WD)))
      (COND
	((COMPILING-FOR-V2)
	 (SETQ OP (ASH WD -9))
	 (LET* ((NAME (AREF (INSTRUCTION-DECODE-TABLE) OP))
		(NO-REG (GET NAME 'NO-REG)))
	  (FLET (( DESTINATION-FOR-PRINTING (NAME)
		  (LET (( D (GET NAME 'DEST) ))
		    (COND ((EQ D 'D-PDL) 'PUSH)
			  ((EQ D 'D-INDS) 'TEST)
			  ((EQ D 'D-RETURN) 'RETURN)
			  (T D) )) ))
	   (COND
	     ((EQ NO-REG 'MISC)
	      (LET ((MISC-NAME (AREF (MISC-OP-NAME-TABLE) DISP)))
		(FORMAT T "(~A) ~A ~A "
			'MISC (DESTINATION-FOR-PRINTING NAME)
			(IF (NULL MISC-NAME) DISP MISC-NAME))
		(WHEN (AND (MEMBER MISC-NAME '(LOAD-FROM-HIGHER-CONTEXT LOCATE-IN-HIGHER-CONTEXT))
			   PC)
		  (LET (( NUM (PUSH-NUMBER-VALUE FEF (1- PC)) ))
		    (UNLESS (NULL NUM)
		      (DISASSEMBLE-LEXICAL-VAR-COMMENT
			FEF
			(LDB SI:%%CONTEXT-DESC-REL-LEVEL NUM)
			(LDB SI:%%CONTEXT-DESC-SLOT NUM)
			T) )))))
	     ((EQ NO-REG 'AREFI)
	      (FORMAT T "~A ~A (~A) "
		      (DESTINATION-FOR-PRINTING NAME)
		      (NTH REG
			   '(GLOBAL:AR-1 ARRAY-LEADER %INSTANCE-REF COMMON-LISP-AR-1 SET-AR-1
					 SET-ARRAY-LEADER SET-%INSTANCE-REF UNUSED-AREFI))
		      (+ (LDB (BYTE 6 0) DISP)
			 (IF (MEMBER REG '(2 6) :TEST #'EQL) 1 0))))
	     ((NULL NAME) (FORMAT T "#o~6O " WD))
	     ((EQ NO-REG 'AUX)
	      (IF (OR (= REG 4)
		      (= REG 5))
		  (FORMAT T "(~A) ~A ~16,2T~S" 'AUX (IF (= REG 4) 'UNBIND 'POP-PDL)
			  (1+ (LDB (BYTE 6 0) WD)))
		(LET ((AUX-NAME (AREF (AUX-OP-NAME-TABLE) DISP)))
		  (FORMAT T "(~A) ~A " 'AUX (IF (NULL AUX-NAME) DISP AUX-NAME))
		  ;;12/09/85 CLM added long branches
		  (COND ((<= #O160 DISP #O177)
			 (UNLESS (NULL SECOND-WORD)
			   (SETQ DISP SECOND-WORD)
			   (WHEN (>= DISP #O100000)
			     (SETQ DISP (LOGIOR #O-100000 DISP)))
			   (WRITE-CHAR #\SPACE)
			   (PRINC DISP) ))
			((EQ AUX-NAME 'STORE-IN-HIGHER-CONTEXT)
			 (UNLESS (NULL PC)
			   (LET (( NUM (PUSH-NUMBER-VALUE FEF (1- PC)) ))
			     (UNLESS (NULL NUM)
			       (DISASSEMBLE-LEXICAL-VAR-COMMENT
				 FEF
				 (LDB SI:%%CONTEXT-DESC-REL-LEVEL NUM)
				 (LDB SI:%%CONTEXT-DESC-SLOT NUM)
				 T) ))))
			((<= #O100 DISP #O103)	   ; complex call
			 (UNLESS
			   (OR (NULL PC)
			       (LET* ((TEM (DISASSEMBLE-FETCH FEF (- PC 1)))
				      (NAME (AREF (INSTRUCTION-DECODE-TABLE)
						  (LDB (SYMEVAL-FOR-TARGET '%%QMI-FULL-OPCODE)
						       TEM))))
				 (NOT (OR (MEMBER NAME '(PUSH PUSH-LONG-FEF))
					  (EQL TEM '#.(+ (LAP-VALUE 'PUSH-MISC-GROUP)
							 (MISC-OP-EVAL '%FUNCTION-INSIDE-SELF)))
					  (AND (EQ (GET NAME 'DEST) 'D-PDL)
					       (EQ (GET NAME 'NO-REG) 'NIL)
					       (NOT (EQL (LDB (BYTE 9 0) TEM) (LAP-VALUE 'PDL-POP))))
					  ))))
			   (LET (( CALL-INFO (PUSH-NUMBER-VALUE FEF (- PC 2)) ))
			     (UNLESS (NULL CALL-INFO)
			       (FORMAT T "~30,2T; ")
			       (DISASSEMBLE-CALL-INFO-WORD CALL-INFO)
			       ))))
			))  ) )
	     ((EQ NO-REG 'MODULE)
	      (LET ((TEM (MODULE-OP-NAME-TABLE)))
		(IF (AND TEM
			 (SETQ TEM
			       (AREF TEM
				     (LDB (SYMEVAL-FOR-TARGET '%%QMI-EXTERNAL-MODULE-NUMBER)
					  DISP))))
		    (LET ((OPNUM (LDB (SYMEVAL-FOR-TARGET '%%QMI-MODULE-OP)
				      DISP)))
		      (FORMAT T "(~A) ~A ~A "
			      (ARRAY-LEADER TEM 0)
			      (DESTINATION-FOR-PRINTING NAME)
			      (OR (AREF TEM OPNUM) OPNUM)))
		  (FORMAT T "~A ~20,1T~S" NAME DISP))))
	     ((EQ NO-REG 'CALL)
	      (PRINC (NTH (LDB (SYMEVAL-FOR-TARGET '%%QMI-CALL-DEST) WD)
			  '(TEST PUSH RETURN TAIL-REC)))
	      (WRITE-CHAR #\SPACE)
	      (PRINC NAME)
	      (DISASSEMBLE-ADDRESS FEF REG DISP SECOND-WORD))
	     ((SYMBOLP NAME)
	      (PRINC NAME)
	      (COND
		((EQ NO-REG 'NIL) ; does use register
		 ;;12/09/85 CLM now prints PDL-PUSH instead of PDL-POP
		 (IF (AND (EQ (GET NAME 'DEST) 'D-STORE)
			  (EQ DISP (LAP-VALUE 'PDL-PUSH)))
		     (FORMAT T " ~20,1T~A" 'PDL-PUSH)
		   (DISASSEMBLE-ADDRESS FEF REG DISP SECOND-WORD)))
		((EQ NO-REG 'BRANCH)
		 (FORMAT T " ~16,1T")
		 (WHEN (> DISP #o400)
		   (SETQ DISP (LOGIOR #o-400 DISP)))	   ;Sign-extend
		 (IF (NULL PC)
		     (PRINC DISP)
		   (PRINC (+ PC DISP 1))))
		((EQ NO-REG 'IMMED)
		 (FORMAT T " ~20,1T~S" (IF (> DISP #o377)
					   (DPB DISP (BYTE 9 0) -1)
					 DISP)))
		((EQ NO-REG 'NOTHING)
		 (COND
		   ((EQ NAME 'DISPATCH)
		    (DISASSEMBLE-DISPATCH-TABLE FEF DISP))
		   ((EQ NAME 'PUSH-LONG-FEF)
		    (DISASSEMBLE-ADDRESS FEF 0 DISP NIL))
		   ((EQ NAME 'SELECT)
		    (DISASSEMBLE-SELECT-TABLE FEF DISP))
		   (T
		    (FORMAT T " ~20,1T~S" DISP)
		    (COND
		      ((EQ NAME 'LDB-IMMED)
		       (DISASSEMBLY-COMMENT)
		       (FORMAT T "(~A ~A ~A)" 'BYTE
			       (LDB (BYTE 4 0) DISP) ; 4-bit length
			       (LDB (BYTE 5 4) DISP) ; 5-bit position
			       ))
		      ((EQ NAME 'LEXICAL-UNSHARE)
		       (UNLESS (NULL FEF)
			 (LET ((VARNAME
				(NTH DISP
				     (SI:GET-DEBUG-INFO-FIELD
				       (FUNCTION-DEBUGGING-INFO FEF)
				       :VARIABLES-USED-IN-LEXICAL-CLOSURES))))
			   (UNLESS (NULL VARNAME)
			     (DISASSEMBLY-COMMENT VARNAME)))))))))
		(T (FORMAT T " ~20,1T~S" DISP))))
	     (T (FORMAT T "#o~O" OP)))))) ; end of compiling for V2
	((ZEROP WD) (PRINC "0"))
	((< OP 9) ;DEST/ADDR
	 (PRINC (NTH OP '(CALL CALL0 MOVE CAR CDR CADR CDDR CDAR CAAR)))
	 (WRITE-CHAR #\SPACE)
	 (PRINC (NTH DEST '(D-IGNORE D-PDL D-RETURN D-LAST)))
	 (DISASSEMBLE-ADDRESS FEF REG DISP SECOND-WORD))
	((= OP 9) ;ND1
	 (PRINC (NTH SUBOP '(ND1-UNUSED + - * / LOGAND LOGXOR LOGIOR)))
	 (DISASSEMBLE-ADDRESS FEF REG DISP SECOND-WORD))
	((= OP 10) ;ND2
	 (PRINC (NTH SUBOP '(= > < EQ SETE-CDR SETE-CDDR SETE-1+ SETE-1-)))
	 (DISASSEMBLE-ADDRESS FEF REG DISP SECOND-WORD))
	((= OP 11) ;ND3
	 (PRINC
	  (NTH SUBOP '(BIND-OBSOLETE? BIND-NIL BIND-POP SET-NIL SET-ZERO PUSH-E MOVEM POP)))
	 (DISASSEMBLE-ADDRESS FEF REG DISP SECOND-WORD))
	((= OP 12) ;BRANCH
	 (PRINC
	  (NTH SUBOP
	       '(BR BR-NIL BR-NOT-NIL BR-NIL-POP BR-NOT-NIL-POP BR-ATOM BR-NOT-ATOM BR-ILL-7)))
	 (FORMAT T " ~16,1T")
	 (AND (> DISP 256) (SETQ DISP (LOGIOR -256 DISP))) ;Sign-extend
	 (IF (NEQ DISP -1)
	     ;; One word
	     (IF (NULL PC)
		 (PRINC DISP)
	       (PRINC (+ PC DISP 1)))
	   ;; Long branch
	   (PROGN
	     (SETQ DISP SECOND-WORD)
	     (AND (>= DISP 32768) (SETQ DISP (LOGIOR -32768 DISP)))
	     (IF (NULL PC)
		 (PRINC DISP)
	       (FORMAT T "LONG ~A" (+ PC DISP 1)))
	     (RETURN))))
	((= OP 13) ;MISC
	 (PRINC "(MISC) ") ;Moon likes to see this
	 (WHEN (LOGTEST 1 SUBOP)
	   (SETQ DISP (+ DISP 512)))
	 (COND
	   ((< DISP 128)
	    (FORMAT T "~A (~A) "
		    (NTH (LDB (BYTE 3 4) DISP)
			 '(GLOBAL:AR-1 ARRAY-LEADER %INSTANCE-REF UNUSED-AREFI-3 AS-1
				       STORE-ARRAY-LEADER %INSTANCE-SET UNUSED-AREFI-7))
		    (+ (LDB (BYTE 4 0) DISP) (IF (= (LDB (BYTE 2 4) DISP) 2)
					       1
					       0))))
	   ((< DISP 144) (FORMAT T "UNBIND ~A binding~:P " (- DISP 127)) ;200 does 1 unbind.
	    (AND (ZEROP DEST) (RETURN)))
	   ((< DISP 160) (FORMAT T "POP-PDL ~A time~:P " (- DISP 144)) ;220 does 0 pops.
	    (AND (ZEROP DEST) (RETURN)))
	   ((= DISP 304) ;(GET 'INTERNAL-FLOOR-1 'QLVAL)
	    (PRINC (NTH DEST '(FLOOR CEILING TRUNCATE ROUND))) (PRINC " one value to stack")
	    (SETQ DEST NIL))
	   ((= DISP 328) ;(GET 'INTERNAL-FLOOR-2 'QLVAL)
	    (PRINC (NTH DEST '(FLOOR CEILING TRUNCATE ROUND))) (PRINC " two values to stack")
	    (SETQ DEST NIL))
	   (T
	    (LET ((OP (MICRO-CODE-SYMBOL-NAME-AREA (- DISP 128))))
	      (IF (NULL OP)
		  (FORMAT T "#~O " DISP)
		(FORMAT T "~A " OP)))))
	 (WHEN DEST
	   (PRINC (NTH DEST '(D-IGNORE D-PDL D-RETURN D-LAST)))))
	((= OP 14) ;ND4
	 (CASE SUBOP
	   (0 (FORMAT T "STACK-CLOSURE-DISCONNECT  local slot ~A" DISP))
	   (1
	    (LET* ((LIST (%P-CONTENTS-OFFSET
			   FEF (- (%P-LDB (SYMEVAL-FOR-TARGET 'SI:%%FEFH-PC-IN-WORDS) FEF) 2)))
		   (CODE (NTH (- (LENGTH LIST) 1 DISP) LIST))
		   (LOCALNUM (LDB (BYTE 10 0) CODE)))
	      (FORMAT T "STACK-CLOSURE-UNSHARE ~A" DISP)
	      (DISASSEMBLY-COMMENT)
	      (IF (< CODE 0)
		  (PROGN (FORMAT T "LOCAL|~A  " LOCALNUM)
			 (LET ((TEM (DISASSEMBLE-LOCAL-NAME FEF LOCALNUM)))
			   (UNLESS (NULL TEM)
			     (PRINC TEM))))
		#+compiler:debug ; this case shouldn't ever happen
		(PROGN
		  (FORMAT T "ARG|~A  " LOCALNUM)
		  (LET ((TEM (DISASSEMBLE-ARG-NAME FEF LOCALNUM)))
		    (UNLESS (NULL TEM)
		      (PRINC TEM)))))))
	   (2 (FORMAT T "MAKE-STACK-CLOSURE  local slot ~A" DISP))
	   (3 (FORMAT T "PUSH-NUMBER ~S" DISP))
	   (4 (FORMAT T "STACK-CLOSURE-DISCONNECT-FIRST  local slot ~A" DISP))
	   (5 (FORMAT T "PUSH-CDR-IF-CAR-EQUAL ")
	      (DISASSEMBLE-ADDRESS FEF REG DISP SECOND-WORD PC))
	   (6 (FORMAT T "PUSH-CDR-STORE-CAR-IF-CONS ")
	      (DISASSEMBLE-ADDRESS FEF REG DISP SECOND-WORD PC))
	   (T (FORMAT T "UNDEF-ND4-~A ~A" SUBOP DISP))))
	((= OP 16)
	 (FORMAT T "~A (~A) "
		 (NTH REG
		      '(GLOBAL:AR-1 ARRAY-LEADER %INSTANCE-REF COMMON-LISP-AR-1 SET-AR-1
			SET-ARRAY-LEADER SET-%INSTANCE-REF UNUSED-AREFI))
		 (+ (LDB (BYTE 6 0) DISP)
		    (IF (MEMBER REG '(2 6) :TEST #'EQL) 1 0)))
	 (PRINC (NTH DEST '(D-IGNORE D-PDL D-RETURN D-LAST))))
	(T ;UNDEF
	 (FORMAT T "UNDEF-~O" OP))))
    (VALUES) ))
