;;;-*- Mode:Common-Lisp; Package:SYSTEM-INTERNALS; Base:10 -*-

;;;	Common Lisp Object System for Explorer and microExplorer
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


;;  8/10/88 DNG - Added new file "firstdefs"; moved "da-changes" after loading
;;		"compiler-support".  Load "generic" to define DEFMETHOD before 
;;		compiling "class" and "compose-class".  These changes needed to 
;;		build without having the system loaded first.
;;  8/12/88 DNG - Add STANDARD-CLASS, STANDARD-GENERIC-FUNCTION, and 
;;		STANDARD-METHOD to the EXPORT list.  Remove :COMPILE-LOAD-INIT dependency 
;;		on compiler files.
;;  8/15/88 DNG - Define "CLOS" package for use without TICL.
;;  8/23/88 DNG - Make the CLOS symbols visible from the USER package.
;;  9/02/88 DNG - Export some of the meta-class protocol symbols that are currently defined.
;;  9/12/88 DNG - Export METHOD-COMBINATION.
;;  9/13/88 DNG - Export SLOT-DESCRIPTION-... symbols.
;;  9/23/88 DNG - Removed use of :COMPILE-LOAD-INIT since it was more of a nuisance than a help.
;;  9/27/88 DNG - Export CLOS:ADD-NAMED-CLASS.
;; 10/26/88 DNG - Added files "environment" and "compile-file".
;; 12/02/88 DNG - Load flavor-metaclass before built-in-classes.
;; 12/05/88 DNG - Use a logical host.
;; 12/06/88 DNG - Include documentation files as an :AUXILIARY module for use by COPY-SYSTEM.
;; 12/07/88 DNG - Export some more accessors from chapter 3.
;; 12/13/88 DNG - Add file "extensions".  Import some accessors into the USER package.
;; 12/28/88 DNG - Directly export ADD-NAMED-CLASS and CLASS-FOR-REDEFINITION now.
;; 12/29/88 DNG - Export STANDARD.
;;  1/04/89 CLM - Add PRINT-OBJECT support.
;;  1/04/89 DNG - MAKE-INSTANCE and DEFMETHOD are now the same symbol in TICL and TICLOS.
;;		Changed file name from "metaclass" to "meta-object".
;;  3/13/89 DNG - Export MAKE-METHOD from the CLOS package.
;;  3/15/89 DNG - Conditionalize to no longer load kernel and compiler patches under release 6.
;;  3/16/89 DNG - Move package creation to new file "package".
;;  3/31/89 DNG - File "zmacs-changes" no longer needed for release 6.
;;; 5/01/89 DNG - Load "meta-object" before "extensions".
;;  5/09/89 DNG - Use SYS host instead of TICLOS logical host.

(comment ; removed 5/9/89
;; Assign a logical host to point to the directory containing this file.
(let* ((this-file (send sys:fdefine-file-pathname :translated-pathname))
       (device (send this-file :device))
       (directory (send this-file :directory)))
  (unless (listp directory) (setq directory (list directory)))
  (net:set-logical-pathname-host "TICLOS"
    :physical-host (send this-file :host)
    :translations 
    `(("TICLOS" ,device ,directory)
      ("PATCH" ,device ,directory)
      ("TEST" ,device ,(append (butlast directory 1) '("TEST")))
      ("TRANSL" ,device ,(append (butlast directory 1) '("TRANSL")))
      ("EDITOR" ,device ,(append (butlast directory 1) '("EDITOR"))))    
    :local-only t))
)

(eval-when (eval compile)
  (when (and (>= (si:get-system-version :system) 6)
	     (fboundp 'compiler:get-from-environment))
    (pushnew :REL6 *features*)) ; temporary, just for use in the DEFSYSTEM below.
  )

(defsystem ti-clos
  (:name "ti Common Lisp Object System")	
  (:short-name "TICLOS")
  (:nicknames "CLOS")
  ;;(:compile-defsystem t)
  (:pathname-default "SYS:CLOS;")
  (:PATCHABLE "SYS:PATCH.CLOS;" PATCH)
  (:warnings-pathname-default "SYS: CWARNS; clos.lisp")  ;; 03/27/89 las
  (:module first ("package" "firstdefs" #-Rel6 "first-rel4"))
  #-Rel6 (:module com-sup ("environment" "compiler-support" "loader" "debugger-changes"))
  #-Rel6 (:module com-sup2 ("da-changes" "optimize"))
  (:module macros ("class-macros"))
  (:module definitions (#-Rel6 "kernel-changes" "class-defs" #-Rel6 "walker"
			#-Rel6 "error-condition-changes"))
  (:module defmethod ("generic"))
  (:module code
	   ("class" 
	    "function-spec"
	    "compose-class"
	    "method-combination"
	    "method-hash-table"
	    "mapping-tables"
	    "standard-type-class"
	    ))
  (:module high-level
	   ("flavor-metaclass"
	    "built-in-classes"
	    "structure-metaclass"
	    "make-instance"
	    "class-redefinition"
	    "slot-generic-functions"
	    "high-opt"
	    "errors"
	    "new-print-object"
	    #-Rel6 "compile-file"
	    "high-level-kernel"
	    "describe"))     ;;moved DESCRIBE here from Kernel-Aux defsystem - clm 03/20/89
  #-Rel6 (:module others ("zmacs-changes"))
  (:module meta ("meta-object" "extensions" #-Rel6 "rel4" "new-print"))
  (:module documentation ("intro.text" "reference.text" "clos-vs-flavors.text"
			  "release-notes.text"))
  #-Rel6
  (:module other-patches ("documenter.lisp"
			  "documenter.xld")) ; conditionally loaded by "rel4" file.
  (:compile-load first)
  #-Rel6 (:compile-load com-sup (:fasload first))
  #-Rel6 (:compile-load com-sup2 (:fasload com-sup first))
  (:compile-load macros (:fasload first #-Rel6 com-sup))
  (:compile-load definitions (:fasload first #-Rel6 com-sup macros))

  (:compile-load defmethod (:fasload first #-Rel6 com-sup macros definitions))
  (:compile-load code (:fasload #-Rel6 com-sup macros definitions defmethod))
  (:compile-load high-level (:fasload first #-Rel6 com-sup #-Rel6 com-sup2 
				      macros definitions code defmethod))
#|
  (:compile-load-init defmethod ( definitions macros)
		      (:fasload #-Rel6 com-sup macros definitions))
  (:compile-load-init code ( definitions macros)
		      (:fasload #-Rel6 com-sup macros definitions defmethod))
  (:compile-load-init high-level (code defmethod)
		      (:fasload #-Rel6 com-sup #-Rel6 com-sup2 code defmethod))
 |#
  #-Rel6 (:compile-load others (:fasload #-Rel6 com-sup code defmethod))
  (:compile-load meta (:fasload #-Rel6 com-sup macros definitions code defmethod high-level))
  (:auxiliary documentation)
  #-Rel6 (:auxiliary other-patches)
  )

(eval-when (eval compile)
  (when (eq (car *features*) ':REL6) ; remove temporary indicator
    (pop *features*)))