;;; -*- Mode:Common-Lisp; Package:SYSTEM-INTERNALS; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985-1989 Texas Instruments Incorporated.  All rights reserved.

; ** (c) Copyright 1980 Massachusetts Institute of Technology **

;;; 03/20/89 clm - This file contains both the new DESCRIBE and DOCUMENTATION
;;;                generic functions.

;;;03/28/89 clm - need STANDARD-SLOT-DESCRIPTION defined at this point - moved here from meta-object
(eval-when (eval compile load)
;; Temporary hack to enable users to use STANDARD-SLOT-DESCRIPTION as a method specializer.
(setf (ticlos:find-class 'ticlos:standard-slot-description)
      (ticlos:find-class 'ticlos:slot-description)) )



;;; 03/20/89 clm - for CLOS, remove old DESCRIBE.
(eval-when (eval compile load)
  (fmakunbound 'describe))

;;;03/20/89 clm - Integrated CLOS DESCRIBE into the Kernel
;;; 05/08/89 DNG - Updated DESCRIBE to use the new generic function DESCRIBE-OBJECT.

(eval-when (eval compile load)
  (import '(ticlos:describe-object))
  (when (and (not (eq (find-package "LISP") (find-package "CL")))
	     (eq (find-symbol "DESCRIBE" "LISP") (find-symbol "DESCRIBE" "CL")))
    (unintern (find-symbol "DESCRIBE" "CL") "CL")
    (export (intern "DESCRIBE" "CL") "CL")))

(ticlos:defgeneric describe (object) ; DESCRIBE according to 88-002R
  (:documentation "Describe the value or components of any Lisp object.
This is a good way to find out more than the printed representation says."))

(defmethod describe ((object t))
  (cl:describe object *standard-output*))

(defvar *describe-depth* 0)

(add-initialization "DESCRIBE warm boot" '(setq *describe-depth* 0) '(:WARM))

(defun cl:describe (object &optional output-stream) ; DESCRIBE as adopted by X3J13 March 1989
  "Describe the value or components of any Lisp object.
This is a good way to find out more than the printed representation says."
  (let* ((stream (decode-print-arg output-stream))
	 (stream2 stream))
    (when (> *describe-depth* 0)
      (when (or (null object)			;Don't recursively describe boring things
		(numberp object) (symbolp object) (stringp object) (consp object)
		(and *print-level* (> *describe-depth* *print-level*)))
	(return-from cl:describe (values)))
      (fresh-line stream)
      (setq stream2				;Arrange for indentation by 5 spaces
	    #'(lambda (&rest args)
		(when (eql (send stream :send-if-handles :read-cursorpos) 0)
		  (write-string "     " stream))
		(lexpr-send stream args))))
    (let ((*describe-depth* (1+ *describe-depth*)))
      (describe-object object stream2)))
  (values))

(ticlos:defgeneric describe-object (object stream)
  (:documentation "Used by DESCRIBE to describe a Lisp object."))

(defmethod describe-object ((dobj ticlos:standard-object) stream)
  (declare (type stream stream))
  (format stream "~&~S, an object of class ~S,~% has slot values of:~%"
	  dobj
	  (ticlos:class-of dobj) )
  (dolist
    (slot (ticlos:class-description-instance-slot-names
	    (ticlos:class-description (ticlos:class-of dobj))) )
    (format stream "~%	~S:~27T" slot)
    (if (ticlos:slot-boundp dobj slot)
	(prin1 (ticlos:slot-value dobj slot) stream)
      (write-string "#< unbound >" stream)))
  (values))

(defmethod describe-object ((dobj lexical-closure) stream)
  (let ((*standard-output* stream))
    (describe-lexical-closure dobj)))

(defmethod describe-object ((dobj closure) stream)
  (let ((*standard-output* stream))
    (describe-closure dobj)))

(defmethod describe-object ((dobj compiled-function) stream)
  (let ((*standard-output* stream))
    (describe-fef dobj)))

(defmethod describe-object ((dobj symbol) stream)
  (let ((*standard-output* stream))
    (describe-symbol dobj)))

(defmethod describe-object ((dobj cons) stream)
  (declare (type stream stream))
  (let ((*print-circle* t))
    (format stream "~%~S is a list" dobj)))

(defmethod describe-object ((dobj stack-group) stream)
  (let ((*standard-output* stream))
    (describe-stack-group dobj)))

(defmethod describe-object ((dobj short-float) stream)
  (let ((*standard-output* stream))
    (describe-short-float dobj)))

(defmethod describe-object ((dobj single-float) stream)
  (let ((*standard-output* stream))
    (describe-single-float dobj)))

(defmethod describe-object ((dobj double-float) stream)
  (let ((*standard-output* stream))
    (describe-double-float dobj)))

(defmethod describe-object ((dobj character) stream)
  (let ((*standard-output* stream))
    (describe-character dobj)))

(defmethod describe-object ((dobj fixnum) stream)
  (declare (type stream stream))
  (format stream "~%~R is ~[even~;odd~]" dobj (ldb (byte 1 0) dobj) (values)))

(defmethod describe-object ((dobj integer) stream)
  (let ((*standard-output* stream))
    (describe-bignum dobj)))

(defmethod describe-object ((number rational) stream)
  (declare (type stream stream))
  (format stream "~&~S is a rational number with numerator ~S and denominator ~S" number
	  (numerator number) (denominator number)))

(defmethod describe-object ((object complex) stream)
  (declare (type stream stream))
  (let ((number object))
    (format stream "~&~S is a complex number with real part ~S and imaginary part ~S." number
	    (realpart number) (imagpart number))))

(defmethod describe-object ((dobj locative) stream)
  (let ((*standard-output* stream))
    (describe-locative dobj)))

(defmethod describe-object ((dobj array) stream)
  (let ((*standard-output* stream))
    (describe-array dobj)))

(defmethod describe-object ((dobj ticlos:standard-generic-function) stream)
  (let ((*standard-output* stream))
    (typecase dobj
      (compiled-function (describe-fef dobj))
      (instance (send dobj :send-if-handles :describe)))
    ))

;; 12/05/88 DNG
(defmethod describe-object :after ((dobj ticlos:class) stream)
  (ticlos:with-slots (ticlos:class-description) dobj
    (when (and (ticlos:slot-boundp dobj 'ticlos:class-description)
	       (typep ticlos:class-description 'ticlos:class-description))
      (cl:describe ticlos:class-description stream))))

;; 5/9/89 DNG Separated structure-object from the T method.  Use describe-array if all else fails.
(defmethod describe-object ((dobj ticlos:structure-object) stream)
  (declare (type stream stream))
  (let ((*standard-output* stream))
    (cond
      ((and (get (named-structure-p dobj) 'named-structure-invoke)
	    (member :describe (named-structure-invoke dobj :which-operations) :test #'eq))
       (named-structure-invoke dobj :describe))
      ;; phd 3/13/86 remove support for the old old defstruct.
      ((get (named-structure-p dobj) 'defstruct-description)
       (describe-defstruct dobj))
      (t (describe-array dobj)))))

;;clm 01/04/89 this fixes the "I don't know how to describe ..." problem
;;DNG 05/09/89 separated out named structure handling.
(defmethod describe-object (dobj stream)
  (declare (type stream stream))
  (let ((*standard-output* stream))
    (typecase dobj
      (instance (send dobj :send-if-handles :describe))
      (t (format stream "~%I don't know how to describe ~S" dobj)))))

(defun describe-fef (fef &aux name  length dbi)
  ;; 11/18/88 DNG - Add indication for when the function is generic.
  ;;  4/19/88 DNG - Add indication for when the FEF is an encapsulation.
  (cond
    ((symbolp fef) (describe-fef (symbol-function fef)))
    ((not (typep fef 'compiled-function)) (ferror () "~S is not a FEF" fef))
    (t 
     (setq length (%p-contents-offset fef %fef-storage-length-word))
     (setq dbi (%p-contents-offset fef %fef-debugging-info-word))
     (if dbi
	 (setq name (SI:GET-DEBUG-INFO-FIELD dbi :name))
	 (setf name "Unknown"))
     (format t "~%FEF for")
     (when (get-debug-info-field dbi 'sys::encapsulated-definition)
       (write-string " encapsulation on"))
     (when (ticlos:generic-function-p fef) (write-string " generic"))
     (format t " function ~S~%" name)
     (unless (zerop (%p-ldb %%FEF-HEADER-Self-Mapping-Table fef))
       (format t "This is a method of flavor ~S.~%"
	       (%p-contents-offset fef
				   (if (= (%p-ldb si::%%fef-header-call-type
						     fef)
					  %fef-call-long)
				   %fef-second-optional-word
				   %fef-first-optional-word))))
     (format t "Initial relative PC: ~S full words.~%" (%p-ldb %%FEF-HEADER-Location-Counter-Offset  fef))
     (format t "There is:") (describe-numeric-descriptor-word fef)
     (format t "~%The length of the local block is ~S~%"
	     (if (= (%p-ldb si::%%fef-header-call-type  fef)
		    %fef-call-long)
		 (%p-ldb-offset %%fef-long-args-number-of-locals fef %fef-first-optional-word )
		 (%p-ldb %%FEF-HEADER-Number-Locals  fef)))
     (format t "The total storage length of the FEF is ~S~%" length)
     ;; -- Special variables
;;     (cond
;;       ((zerop (ldb %%fefh-sv-bind header)) (princ "There are no special variables present."))
;;       (t (princ "There are special variables, ") (terpri)
;;	(cond
;;	  ((zerop (ldb %%fefhi-svm-active sv)) (princ "but the S-V bit map is not active. "))
;;	  (t
;;	   (format t "and the S-V bit map is active and contains: ~O" (ldb %%fefhi-svm-bits sv))))))
;;     (terpri)
     ;; -- ADL.
;;     (cond
;;       ((not (zerop (ldb %%fefhi-ms-bind-desc-length misc)))
;;	(format t "There is an ADL:  It is ~S long, and starts at ~S"
;;		(ldb %%fefhi-ms-bind-desc-length misc) (ldb %%fefhi-ms-arg-desc-org misc))
;;	(describe-fef-adl fef))
;;       (t (princ "There is no ADL.")))
;;     (terpri)
     ;; -- Debugging Info
     (when dbi
	(format t "Debugging info:~%")
	(describe-1 dbi)
	;; Do the known fields first,
        ;; Do misc fields
;;	(dolist (alist-item ....)
;;	  (format t "  ~S: ~S~%" (car alist-item ) (cdr alist-item)))
	))))




(defun describe-stack-group (sg)
  (format t "~&Stack Group ~S~2%" (sg-name sg))
  (let ((d (get 'stack-group 'defstruct-description)))
    (when d					; Which had better be there.
      (loop for (slot-name . slot-description) in (defstruct-description-slot-alist d)
	    doing (format t "~&~S:~35T" slot-name)
	    doing (let ((pointer (locf (array-leader sg (defstruct-slot-description-number slot-description)))))
		    (if (%p-contents-safe-p pointer)
			(let ((slot-value (funcall (defstruct-slot-description-ref-macro-name slot-description) sg)))
			  (format t "~S" slot-value)
			  (when (get slot-name 'sg-description-annotation)
			    (funcall (get slot-name 'sg-description-annotation) slot-value)))
			(format t "#<~S ~O>"
				(or (nth (%p-data-type pointer) q-data-types)
				    (%p-data-type pointer))
				(%p-pointer pointer))))))))

(defun (:property sg-flags-car-sym-mode sg-description-annotation) (slot-value)
   (format t " (CAR of a symbol is ~[an error~;NIL if the symbol is NIL, otherwise an error~;NIL~;its print-name~])"
	   slot-value))

(defun (:property sg-flags-cdr-sym-mode sg-description-annotation) (slot-value)
   (format t " (CDR of a symbol is ~[an error~;NIL if the symbol is NIL, otherwise an error~;NIL~;its property list~])"
	   slot-value))

(defun (:property sg-flags-car-num-mode sg-description-annotation) (slot-value)
   (format t " (CAR of a number is ~[an error~;NIL~:;in an unknown state~])"
	   slot-value))

(defun (:property sg-flags-cdr-num-mode sg-description-annotation) (slot-value)
   (format t " (CDR of a number is ~[an error~;NIL~:;in an unknown state~])"
	   slot-value))

(defun (:property sg-flags-trap-enable sg-description-annotation) (slot-value)
   (format t " (Trapping is ~[dis~;en~]abled)"
	   slot-value))

(defun (:property sg-current-state sg-description-annotation) (slot-value)
   (format t " (~S)" (nth slot-value sg-states)))

(defun (:property sg-inst-disp sg-description-annotation) (slot-value)
   (format t " (~[Normal~;Debug~;Single-Step~;Single-Step Done~])" slot-value))

(defun (:property sg-in-swapped-state sg-description-annotation) (slot-value)
   (format t " (Variables~[ not~] currently swapped out)" slot-value))

(defun (:property sg-foothold-executing-flag sg-description-annotation) (slot-value)
   (format t " (Foothold~[ not~] currently executing)" slot-value))

(defun (:property sg-processing-error-flag sg-description-annotation) (slot-value)
   (format t " (~[Not c~:;C~]urrently processing an error)" slot-value))

(defun (:property sg-processing-interrupt-flag sg-description-annotation) (slot-value)
   (format t " (~[Not c~:;C~]urrently processing an interrupt)" slot-value))

(defun (:property sg-flags-mar-mode sg-description-annotation) (slot-value)
   (format t " (~[Inactive~;Read~;Write~;Read&Write~])" slot-value))


;;(defun describe-stack-group (sg &aux tem)
;;  (format t "~%Stack Group; name is ~S, current state ~S" (sg-name sg)
;;	  (nth (sg-current-state sg) sg-states))
;;  (cond
;;    ((not (zerop (sg-in-swapped-state sg))) (format t "~%  Variables currently swapped out")))
;;  (cond
;;    ((not (zerop (sg-foothold-executing-flag sg))) (format t "~%  Foothold currently executing")))
;;  (cond
;;    ((not (zerop (sg-processing-error-flag sg))) (format t "~% Currently processing an error")))
;;  (cond
;;    ((not (zerop (sg-processing-interrupt-flag sg)))
;;     (format t "~% Currently processing an interrupt")))
;;  (let ((em (sg-saved-m-flags sg)))
;;    (format t "~%ERROR-MODE:")
;;    (format t "~%  CAR of a number is ~A." (case (ldb %%m-flags-car-num-mode em)
;;						 (0 "an error")
;;						 (1 "NIL")
;;						 (otherwise "in an unknown state")))
;;    (format t "~%  CDR of a number is ~A." (case (ldb %%m-flags-cdr-num-mode em)
;;						 (0 "an error")
;;						 (1 "NIL")
;;						 (otherwise "in an unknown state")))
;;    (format t "~%  CAR of a symbol is ~A." (case (ldb %%m-flags-car-sym-mode em)
;;						 (0 "an error")
;;						 (1 "NIL if the symbol is NIL, otherwise an error")
;;						 (2 "NIL")
;;						 (3 "its print-name")))
;;    (format t "~%  CDR of a symbol is ~A." (case (ldb %%m-flags-cdr-sym-mode em)
;;						 (0 "an error")
;;						 (1 "NIL if the symbol is NIL, otherwise an error")
;;						 (2 "NIL")
;;						 (3 "its property list")))
;;    (format t "~%  Trapping is ~A." (case (ldb %%m-flags-trap-enable em)
;;					  (0 "disabled")
;;					  (1 "enabled"))))
;;  (format t "~%SG-SAFE ~D, SG-SWAP-SV-ON-CALL-OUT ~D, SG-SWAP-SV-OF-SG-THAT-CALLS-ME ~D"
;;	  (sg-safe sg) (sg-swap-sv-on-call-out sg) (sg-swap-sv-of-sg-that-calls-me sg))
;;  (format t "~%SG-INST-DISP: ~D (~:*~[Normal~;Debug~;Single-step~;Single-step done~])"
;;	  (sg-inst-disp sg))
;;  (format t "~%SG-PREVIOUS-STACK-GROUP ~S, SG-CALLING-ARGS-NUMBER ~S, SG-CALLING-ARGS-POINTER ~S"
;;	  (sg-previous-stack-group sg) (sg-calling-args-number sg) (sg-calling-args-pointer sg))
;;  (format t "~%Regular PDL pointer ~D, ~D available, ~D limit" (sg-regular-pdl-pointer sg)
;;	  (array-total-size (sg-regular-pdl sg)) (sg-regular-pdl-limit sg))
;;  (format t "~%Special PDL pointer ~D, ~D available, ~D limit" (sg-special-pdl-pointer sg)
;;	  (array-total-size (sg-special-pdl sg)) (sg-special-pdl-limit sg))
;;  (cond ((setq tem (sg-recovery-history sg)) (format t "~%Recovery history ~S" tem)))
;;  (cond ((setq tem (sg-ucode sg)) (format t "~%SG-UCODE ~S" tem)))) 


(defun describe-numeric-descriptor-word (fef &aux min max)
  (terpri)
  (princ "   ")
  (case (%p-ldb %%FEF-HEADER-Call-Type	fef )
    (0 (format t "~D. required arguments, no optional or locals (simple call)"
	       (%p-ldb %%FEF-HEADER-Number-Args fef)))
    (1 (format t "~D. optional arguments" (%p-ldb %%FEF-HEADER-Number-Optional-Args fef)))
    (2 (format t "~D. local variables" (%p-ldb %%FEF-HEADER-Number-Locals fef)))
    (3 (format t "&REST arg and ~D. local variables" (%p-ldb %%FEF-HEADER-Number-Locals fef)))
    (4 (format t "~D. optional arguments and ~D. local variables"
	       (%p-ldb %%FEF-HEADER-Number-Optional-Args fef)
	       (%p-ldb %%FEF-HEADER-Number-Locals fef)))
    (5 (format t "~D. optional arguments and &REST arg" (%p-ldb %%FEF-HEADER-Number-Optional-Args fef)))
    (7
     (when (%p-ldb-offset %%fef-long-args-rest-arg fef %fef-first-optional-word)
       (princ "&rest argument"))
     (setq max (%p-ldb-offset  %%fef-long-args-min-args fef  %fef-first-optional-word))
     (setq min (%p-ldb-offset  %%fef-long-args-max-args fef  %fef-first-optional-word))
     (cond
       ((= max min) (princ max) (princ " arg(s)."))
       (t (princ "Takes between ") (princ min) (princ " and ") (princ max) (princ " args."))))))


(defun describe-array (array &aux arraydims ndims long-length-flag)
  (cond
    ((symbolp array)
     (cond
       ((and (boundp array) (arrayp (symbol-value array))) (describe-array (symbol-value array)))
       ((and (fboundp array) (arrayp (symbol-function array)))
	(describe-array (symbol-function array)))
       (t nil)))
    ((arrayp array) (format *standard-output* "~%This is an ~S type array." (array-type array))
		    (setq arraydims (array-dimensions array)) (setq ndims (length arraydims))
		    (setq long-length-flag (%p-ldb-offset %%array-long-length-flag array 0))
		    (cond
		      ((> ndims 1) (format *standard-output* "~%It is ~D-dimensional, with dimensions " ndims)
				   (do ((l arraydims (cdr l)))
				       ((null l)
					nil)
				     (format *standard-output* "~S " (car l))))
		      (t (format *standard-output* "~%It is ~S long." (array-total-size array))
			 (if arraydims		;don't barf on 0-rank arrays
			     (and (< (array-active-length array) (car arraydims))
				  (format *standard-output* "  Active length is ~S" (array-active-length array)))
			     (format t " It is of zero rank."))))
		    (and (array-has-leader-p array)
			 (format *standard-output* "~%It has a leader, of length ~S" (array-leader-length array)))
		    (cond
		      ((array-displaced-p array)
		       (cond
			 ((array-physical-p array)
			  (format *standard-output* "~%The array is displaced to the physical address ~S"
				  (dpb (%p-ldb-offset (byte 16 16) array (+ ndims long-length-flag)) (byte 16 16)
				       (%p-ldb-offset (byte 16 0) array (+ ndims long-length-flag))) ))
			 ((array-indirect-p array)
			  (format *standard-output* "~%The array is indirected to ~S"
				  (%p-contents-offset array (+ ndims long-length-flag)))
			  (and (array-indexed-p array)
			       (format *standard-output* ", with index-offset ~S"
				       (%p-contents-offset array (+ ndims long-length-flag 2))))
			  (format *standard-output* "~%Description:")
			  (describe-array (%p-contents-offset array (+ ndims long-length-flag))))
			 (t
			  (format *standard-output* "~%The array is displaced to ~S"
				  (%p-contents-offset array (+ ndims long-length-flag))))))))
    (t (ferror () "~S is not an array" array)))) 




;;11/11/88 clm - removed optional argument T (no-complaints) in call to describe;
;;new clos describe doesn't use it
;;03/20/89 clm - Integrated CLOS changes into Kernel.
(defun describe-1 (thing);An internal subroutine
  (unless (or (null thing)			;Don't recursively describe boring things
	      (numberp thing) (symbolp thing) (stringp thing) (consp thing))
    (fresh-line)
    (let ((*standard-output*			;Arrange for indentation by 5 spaces
	    (closure '(*standard-output*)
		     #'(lambda (&rest args)
			 (and (eql (send *standard-output* :send-if-handles :read-cursorpos) 0)
			      (write-string "     "))
			 (lexpr-send *standard-output* args)))))
      (describe thing)))) 

;;PAD 4/8/87 Prevent inherited symbols from being described as interned symbols 
;;           SPR 4578 (CR:PHD)
;;DNG 4/19/89 Don't describe pathnames appearing as symbol properties since 
;;		the verbose display rarely shows anything of interest that is not included 
;;		in the name.  Say "in the FOO package" instead of "in FOO package".
(defun describe-symbol (sym)
  (if (SYMBOL-PACKAGE sym)
      (format t "~%Symbol ~S is in the ~A package." sym (package-name (symbol-package sym)))
      (format t "~%Symbol ~S is Uninterned." sym))
  (let ((tem nil))
    (do-all-packages (p)
      (multiple-value-bind (s flag)
	  (find-symbol sym p)
	(when (and (neq :inherited flag) (eq s sym) (not (eq p (symbol-package sym)))
		   (not (member p (package-used-by-list (symbol-package sym)) :test #'eq)))
	  (push p tem))))
    (when tem
      (format t "~% It is also interned in package~P ~{~A~^, ~}" (length tem) tem)))
  (when (and (boundp sym) (not (keywordp sym)))
    (let ((*print-level* 2)
	  (*print-length* 3))
      (format t "~%The value of ~S is ~S" sym (symbol-value sym)))
    (describe-1 (symbol-value sym)))
  (when (fboundp sym)
    (let ((*print-level* 2)
	  (*print-length* 3))
      (ignore-errors
	(format t "~%The function definition of ~S is ~S: ~S" sym (symbol-function sym)
		(arglist sym))))
    (describe-1 (symbol-function sym)))
  (do ((pl (symbol-plist sym) (cddr pl))
       (*print-level* 2)
       (*print-length* 3))
      ((null pl))
    (format t "~%~S has property ~S: ~S" sym (car pl) (cadr pl))
    (unless (pathnamep (cadr pl))
      (describe-1 (cadr pl))))
  (if (not (or (boundp sym) (fboundp sym) (symbol-plist sym)))
      (format t "~%It has no value, definition or properties"))
  ())
 



(defun describe-locative (x)
  (let ((area (%area-number x)))
    (cond
      (area (format t "~%~S is a locative pointer into area ~S~%It points " x (area-name area))
       (let* ((struc (%find-structure-header (follow-structure-forwarding x)))
	      (basep (%pointer (%find-structure-leader struc)))
	      (bound (+ (%structure-total-size struc) basep)))
	 (if (and (<= basep (%pointer x)) (< (%pointer x) bound))
	   (format t "to word ~O of ~S~%" (%pointer-difference x struc) struc)
	   (format t "at some sort of forwarded version of ~S~%" struc))
	 (describe-1 struc)))
      (t (format t "~%~S is a locative pointer not into any area." x))))) 



(defun describe-defstruct (x &optional defstruct-type &aux description)
  (setq description
	(get (or defstruct-type (if (consp x)
				  (car x)
				  (named-structure-p x)))
	     'defstruct-description))
  (let ((*print-structure* nil))
    (format t "~%~S is a structure of type ~S~%" x (defstruct-description-name)))
  (do ((l (defstruct-description-slot-alist) (cdr l)))
      ((null l)
       nil)
    (format t "   ~30A~S~%" (string-append (caar l) ":")
	    (catch-error (eval1 `(,(defstruct-slot-description-ref-macro-name (cdar l)) ',x)) ())))
  x) 


(defun describe-closure (cl)
  (let ((c (%make-pointer dtp-list cl))
	(sym nil)
	(offset nil))
    (format t "~%~S is a closure of ~S:~%" cl (car c))
    (do ((l (cdr c) (cddr l)))
	((null l)
	 nil)
      (setq sym (%find-structure-header (car l))
	    offset (%pointer-difference (car l) sym))
      (format t "   ~A cell of ~S: ~40T~:[unbound~;~S~]~%"
	      (case offset
		(0 "Print name")
		(1 "Value")
		(2 "Function")
		(3 "Property list")
		(4 "Package"))
	      sym (location-boundp (cadr l)) (and (location-boundp (cadr l)) (caadr l))))
    (describe-1 (car c))))

;;;
;;;  Lexical-closure code adapted from Inspector code.  Thanks, Dan.

(defun describe-lexical-closure (lcl)
   (LET* ((c (convert-closure-to-list lcl))
	  (env (CDR c))
	  (env-vec (%make-pointer dtp-list env))
	  (dbi (get-debug-info-struct (CAR c)))
	  (parent-info (get-debug-info-field dbi :lexical-parent-debug-info)))
     (format t "~%~S is a lexical closure of ~S:~%" lcl (car c))
     (DO* ((i 0 (1+ i))
	   (levels (closure-levels? env-vec))
	   (env-vec env-vec (%make-pointer dtp-list (CAR env-vec)))
	   (pnt-info parent-info (get-debug-info-field pnt-info :lexical-parent-debug-info))
	   (vars (get-debug-info-field pnt-info :variables-used-in-lexical-closures)
		 (get-debug-info-field pnt-info :variables-used-in-lexical-closures)))
	  ((> i levels))
       (format t "~2&Context ~A (function ~S):"
	       i (get-debug-info-field pnt-info :name))
       (DO* ((j 0 (1+ j))
	     (vals (CDR env-vec) (CDR vals))
	     (val (CAR vals) (CAR vals))
	     (vrs vars (CDR vrs))
	     (var (CAR vrs) (CAR vrs)))
	    ((NULL vrs))
	 (format t "~&   Lexical variable ~S: ~40T~S~%" var val)))
     (describe-1 (car c))))

(DEFUN closure-levels? (LIST)
  "Looks at the top level of a closure's environment vector to see if it
has nested levels of environment vectors. Returns 0 if no nested levels,
otherwise returns the number of nested levels."
  (LET ((parent (CAR list)))
    (IF parent 
        (DO ((i 0 (1+ i))
             (prnt parent (CAR (%make-pointer dtp-list prnt))))
            ((NULL prnt) (RETURN i)))
        (VALUES 0))))


;(defun describe-stack-closure (scl)
;  (let ((sc (%make-pointer dtp-list scl)))
;    (format t "~%~S is a stack closure of ~S:~%" scl (car sc))
;    ;;   (FORMAT T "~%   Entries: ~A" (CAADR SC))   ;; improve this -- mary
;    (describe-1 (car sc)))) 


(defun describe-character (character)
  (setq character (character character))
  (format t
	  "~&~S is a character with code ~D.
Its control-bits are ~D~:[~4*~; (~@[Control~*~]~@[Meta~*~]~@[Super~*~]~@[Hyper~*~])~]. Its font is ~D"
	  character (char-code character) (char-bits character) (/= 0 (char-bits character))
	  (char-bit character :control) (char-bit character :meta) (char-bit character :super)
	  (char-bit character :hyper) (char-font character))
  character) 


(defun describe-short-float (x)
  (format t "~&~S is a short floating-point number.~&  Sign ~O, "
	  x (if (minusp (%pointer x)) 1 0))
  (if (zerop x)
      (format t "exponent 0, 16-bit fraction 0 (zero)")
      (format t "exponent #o~O, 16-bit fraction #o~O (not including hidden bit)"
	  (ldb (byte 8. 16.) (%pointer x))
	  (ldb (byte 16. 0) (%pointer x))))) 


(defun describe-single-float (x)
  (format t "~&~S is a single-precision floating-point number.~&  Sign ~O, "
	  x (%p-ldb-offset (byte 1 31.) x 1))
  (if (zerop x)
      (format t "exponent 0, 23-bit fraction 0 (zero)")
      (format t "exponent #o~O, 23-bit fraction #o~O (not including hidden bit)"
	  (%p-ldb-offset (byte 8. 23.) x 1)
	  (%p-ldb-offset (byte 23. 0) x 1))))


(defun describe-double-float (x)
  (format t "~&~S is a double-precision floating-point number.~&  Sign ~O, "
	  x (%p-ldb-offset (byte 1 31.) x 2))
  (if (zerop x)
      (format t "exponent 0, 23-bit fraction 0 (zero)")
      (format t "exponent #o~O, 52-bit fraction #o~O (not including hidden bit)"
	  (%p-ldb-offset (byte 11. 20.) x 2)
	  (dpb (%p-ldb-offset (byte 20. 0) x 2) (byte 20. 32.)
	       (dpb (%p-ldb-offset (byte 16. 16.) x 1) (byte 16. 16.)
		    (%p-ldb-offset (byte 16. 0) x 1))))))  


(defun describe-bignum (x)
  (let ((len (%p-ldb-offset (BYTE 18 0) X  0))
	(barf nil))
    (format t
	    "~&~S is a bignum.~&It is ~R word~:P long.  It is ~[positive~;negative~].  ~
                 It is stored starting at location: ~O~&Its contents:~2%"
	    x len (%p-ldb-offset (BYTE 1 18) X  0) (%pointer x))
    (do ((i 1 (1+ i)))
	((> i len))
      (or (zerop (%p-ldb-offset (BYTE 1 31) X  i)) (setq barf t))
      (format t "~&~3O: ~[ ~;*~]" i (%p-ldb-offset (BYTE 1 31) X  i))
      (do ((ppss (BYTE 1 30) (- ppss (BYTE 0 1)) ))
	  ((< ppss 1))
	(write-char (int-char (+ #\0 (%p-ldb-offset ppss x i)))))
      (format t "  ~O," (%p-ldb-offset (BYTE 1 30) X  i))
      (do ((ppss (BYTE 3 27)   (- ppss  (BYTE 0 3)) ))
	  ((< ppss (BYTE 3 0) ))
	(write-char (int-char (+ #\0 (%p-ldb-offset ppss x i)))))
      (princ "  ")
      (do ((ppss (BYTE 3 28) (- ppss (BYTE 0 3) )))
	  ((< ppss (BYTE 3 1) ))
	(write-char (int-char (+ #\0 (%p-ldb-offset ppss x i)))))
      (format t ",~O  ~O," (%p-ldb-offset 1 x i) (%p-ldb-offset (BYTE 2 29) X  i))
      (do ((ppss (BYTE 3 26) (- ppss (BYTE 0 3) )))
	  ((< ppss (BYTE 3 2) ))
	(write-char (int-char (+ #\0 (%p-ldb-offset ppss x i)))))
      (format t ",~O" (%p-ldb-offset (BYTE 2 0) X  i)))
    (if barf
      (format t "~2&* = high order bit illegally 1, bug in bignum microcode?"))
    (terpri))
  x) 
 



;;;DOCUMENTATION

;;;03/20/89 clm - Integrated CLOS DOCUMENTATION into the Kernel.

(defvar *documentation-types* '(variable function defflavor flavor structure type setf value profile:profile-variable)
  "This is a list of valid documentation-types for the DOCUMENTATION generic function.")

(eval-when (compile load)
  (unless (ticlos:generic-function-p 'documentation)
    (fmakunbound 'documentation))
)

;;;GENERIC-FUNCTIONS
(ticlos:defgeneric documentation (symbol &optional doc-type )
  (:documentation   "Try to return the documentation string for SYMBOL, else return NIL.
Standard values of DOC-TYPE are: FUNCTION, VARIABLE, TYPE, STRUCTURE and SETF,
but you can put on and retrieve documentation for any DOC-TYPE.
Documentation strings are installed by SETFing a call to DOCUMENTATION."))

(ticlos:defgeneric (setf documentation) (new-value dobj &optional doc-type)
  (:argument-precedence-order dobj new-value))

;;this is removed by the defgeneric; adding it back in to prevent problem sited in 
;;SET-DOCUMENTATION.
(defsetf documentation set-documentation)


;;;STANDARD-METHOD
(defmethod documentation ((dobj ticlos:standard-method) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to DOCUMENTATION 
of a method object.~%" doc-type))
  (let ((dbi (get-debug-info-struct (ticlos:method-function dobj))))
    (get-debug-info-field dbi :documentation)))


(defmethod (setf documentation) (new-value (dobj ticlos:standard-method) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to DOCUMENTATION 
of a method object.~%" doc-type))
  (let ((dbi (get-debug-info-struct (ticlos:method-function dobj)))
	(%inhibit-read-only t))
    (setf (get-debug-info-field dbi :documentation) new-value)))



;;;STANDARD-GENERIC-FUNCTION
(defmethod documentation ((dobj ticlos:standard-generic-function) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to DOCUMENTATION 
of a generic function object.~%" doc-type))
  (if (compiled-function-p dobj)
      (let ((dbi (get-debug-info-struct dobj)))
	(get-debug-info-field dbi :documentation))
      (ticlos:generic-function-documentation dobj) ))

(defmethod (setf documentation) (new-value (dobj ticlos:standard-generic-function) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to (SETF DOCUMENTATION)
of a generic function object.~%" doc-type))
  (if (compiled-function-p dobj)
      (let ((dbi (get-debug-info-struct dobj))
	    (%inhibit-read-only t))
	(setf (get-debug-info-field dbi :documentation) new-value))
      (setf (ticlos:generic-function-documentation dobj) new-value)))



;;;STANDARD-CLASS
(defmethod documentation ((dobj ticlos:standard-class) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to DOCUMENTATION 
of a class object.~%" doc-type))
  (let ((plist (ticlos:class-description-plist (ticlos:class-description dobj))))
    (getf plist :documentation)))

(defmethod (setf documentation) (new-value (dobj ticlos:standard-class) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to (SETF DOCUMENTATION) 
of a class object.~%" doc-type))
    (setf (getf (ticlos:class-description-plist (ticlos:class-description dobj)) :documentation) new-value))

;;BUILT-IN-CLASS
(defmethod documentation ((dobj ticlos:built-in-class) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to DOCUMENTATION 
of a built-in-class object.~%" doc-type))
  (let ((plist (ticlos:class-description-plist (ticlos:class-description dobj))))
    (getf plist :documentation)))

(defmethod (setf documentation) (new-value (dobj ticlos:built-in-class) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to (SETF DOCUMENTATION) 
of a built-in-class object.~%" doc-type))
    (setf (getf (ticlos:class-description-plist (ticlos:class-description dobj)) :documentation) new-value))

;;;STRUCTURE-CLASS
(defmethod documentation ((dobj ticlos:structure-class) &optional doc-type)
  (when doc-type
   (error "~%A second argument, ~s, was supplied in a call to DOCUMENTATION 
of a structure-class object.~%" doc-type))
  (getf (get (ticlos:class-name dobj) 'documentation-property)  'structure))

(defmethod (setf documentation) (new-value (dobj ticlos:structure-class) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to (SETF DOCUMENTATION) 
of a structure-class object.~%" doc-type))
  (setf (getf (get (ticlos:class-name dobj) 'documentation-property)  'structure) new-value))

;;FLAVOR-CLASS  -- 1/24/89 DNG
(defmethod documentation ((dobj ticlos:flavor-class) &optional doc-type)
  (declare (ignore doc-type))
  (or (getf (get (ticlos:class-name dobj) 'documentation-property) 'defflavor)
      ;; work-around for SPR 9245
      (values (find-if #'stringp (the list (getf (sys:flavor-plist (ticlos:class-description dobj))
						 :documentation))))))

;;;METHOD-COMBINATION
;;1/11/89 method combination is not a defined class yet
;(defmethod documentation ((dobj ticlos:method-combination) &optional doc-type)
;  (when doc-type
;    (error "~%A second argument, ~s, was supplied in a call to DOCUMENTATION 
;of a method combination object.~%" doc-type))
;  ())


;;;STANDARD-SLOT-DESCRIPTION
(defmethod documentation ((dobj ticlos:standard-slot-description) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to DOCUMENTATION 
of a slot description object.~%" doc-type))
  (ticlos:slot-documentation dobj))

(defmethod (setf documentation) (new-value (dobj ticlos:standard-slot-description) &optional doc-type)
  (when doc-type
    (error "~%A second argument, ~s, was supplied in a call to (SETF DOCUMENTATION) 
of a slot description object.~%" doc-type))
  (setf (ticlos:slot-documentation dobj) new-value))


;;;SYMBOL
(defmethod documentation ((dobj symbol) &optional (doc-type 'function))
  (cond ((and (eq doc-type 'value)    
	      (get dobj :documentation)))
	((getf (get dobj 'documentation-property)
		(if (string-equal doc-type 'flavor)
		    ;; PMH: allow 'flavor to map to 'defflavor
		    'defflavor doc-type)))
	((eq doc-type 'type)
	 (let ((obj (ticlos:class-named dobj t)))
	   (or (and obj
		    (documentation obj))
	       (and (get dobj 'type-expander)
		    (documentation (get dobj 'type-expander) 'function)))
	     ))
	((eq doc-type 'function)
	 (and (fboundp dobj)
	      (not (microcode-function-p (symbol-function dobj)))
	      (not (vectorp (symbol-function dobj)))
	      (documentation (fdefinition (unencapsulate-function-spec dobj)) )))
	(t (unless (member doc-type *documentation-types* :test #'eq)
	     (error "~s is not an acceptable value for the optional argument DOC-TYPE." doc-type)))
	))

;; clm 01/23/89 - changed to not call SET-DOCUMENTATION.
(defmethod (setf documentation) (new-value (dobj symbol) &optional (doc-type 'function))
  (pushnew doc-type *documentation-types*)
  (setf (getf (get dobj 'documentation-property)
		(if (string-equal doc-type 'flavor)
		    ;; PMH: allow 'flavor to map to 'defflavor
		    'defflavor doc-type)) new-value))

;;;LIST
;;1/30/89 clm - removed optional arg in call to documentation using fdefinition-safe;
;;problem occurred if fdefinition-safe returned a generic-function object
(defmethod documentation ((dobj cons) &optional (doc-type 'function))
  (if (functionp dobj t)
      (if (eq (car dobj) 'macro)
	  (documentation (cdr dobj) doc-type)
	  (multiple-value-bind (nil nil doc)
	      (parse-body (cdr (lambda-exp-args-and-body dobj)) nil t)
	    doc))
      (documentation (fdefinition-safe dobj t))
	   ))

(defmethod (setf documentation) (new-value (dobj cons) &optional (doc-type 'function))
  (if (functionp dobj t)
      (if (eq (car dobj) 'macro)
	  (setf (documentation (cdr dobj) doc-type) new-value)
	  (error "~s is not a valid argument for (SETF DOCUMENTATION)" dobj))
      (and (fdefinedp dobj)
	   (setf (documentation (fdefinition (unencapsulate-function-spec dobj))) new-value))
      ))



;;;COMPILED-FUNCTION
(defmethod documentation ((dobj ticlos:compiled-function) &optional doc-type)
  (declare (ignore doc-type))   
  (let ((dbi (get-debug-info-struct dobj)))
    (get-debug-info-field dbi :documentation)))


(defmethod (setf documentation) (new-value (dobj compiled-function) &optional doc-type)
  (declare (ignore doc-type))
  (let ((dbi (get-debug-info-struct dobj))
	(%inhibit-read-only t))
    (setf (get-debug-info-field dbi :documentation) new-value))
  )

;;LEXICAL-CLOSURE
(defmethod documentation ((dobj ticlos:lexical-closure) &optional doc-type)
  (declare (ignore doc-type))
  (documentation (closure-function dobj)))


(defmethod (setf documentation) (new-value (dobj ticlos:lexical-closure) &optional doc-type)
  (declare (ignore doc-type))
  (setf (documentation (closure-function dobj)) new-value))


;;CLOSURE
(defmethod documentation ((dobj ticlos:closure) &optional doc-type)
  (declare (ignore doc-type))
  (documentation (closure-function dobj)))


(defmethod (setf documentation) (new-value (dobj ticlos:closure) &optional doc-type)
  (declare (ignore doc-type))
  (setf (documentation (closure-function dobj)) new-value))

;; clm 01/23/89 - changed SET-DOCUMENTATION to call the setf function; problem
;;occurred when loading files compiled with the new documentation on an older system.
(defun set-documentation (symbol doc-type &optional (value doc-type value-supplied))
  (unless value-supplied
    (setf doc-type 'function))
  (funcall #'(setf documentation)
	   value symbol doc-type ))


