;;; -*- cold-load:t; Mode:Common-Lisp; Package:EH; Fonts:(CPTFONT HL12B HL12BI); Base:10;Patch-file: T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.


(defun trans-trap-decode (sg)
  "Given the address on which a trans-trap occurred, determine where the contents of
that address is stored now, and if it is a null pointer, what cell of which symbol was unbound. 
The symbol can actually be a function-spec."
  (declare (values original-address current-address cell-type symbol))
  (let* ((original-address (sg-saved-vma sg))
	 (current-address (cell-location-in-stack-group original-address sg)))
    ;; If NULL on top of stack (because trans-trap pushes the pointer and then the data-type).
    (if (= (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length)) 
	   dtp-null)
	(let ((cell-type :unknown)		; Means we don't know what's unbound.
	      symbol)
	  (multiple-value-bind (containing-structure explanation)
	      (if *use-fsh-safe*
		  (find-structure-header-safe original-address)
		  (%find-structure-header original-address))
	    (if (null explanation)		; No problems.
		;; Is it a symbol, list, function, ...
		(cond ((symbolp containing-structure)
		       (case (%pointer-difference original-address containing-structure)
			     (1 (setq cell-type :value))
			     (2 (setq cell-type :function)))
		       (setq symbol containing-structure))
;;  This appears to be an unbound cell that points at a list whose car is a :method function-spec.
;;  I don't understand how such a case could come about, and don't think that it does anymore
;;  so I'm commenting it out.  - pf, Aug 3, 1986
;		       ((and (listp symbol)		; I don't understand this case.  - pf, Aug 1, 1986
;		             (= (%p-data-type symbol) dtp-list)
;		             (si:%p-contents-eq (car symbol) :method))
;		        (setq cell-type :function
;		              symbol    (si:meth-function-spec symbol)))
		      ((ticlos:clos-instance-p  containing-structure) 
		       (setq cell-type :clos-instance)
		      (let ((flavor (instance-flavor containing-structure))
			     (index (1- (%pointer-difference original-address containing-structure))))
			 (if flavor
			     (setq symbol (nth index (flavor-all-instance-variables flavor)))
			     (setq symbol (intern (format nil "INSTANCE-VAR-~A" index))))))
		      ((eq (named-structure-p containing-structure)
			   'ticlos:class-slot-structure )
		       (setf cell-type :class-slot)
		       (let ((class-description (array-leader  containing-structure 0)))
			 (setf symbol
			       (car (rassoc (%make-pointer sys:dtp-locative original-address )
				       (ticlos:class-description-class-slot-alist class-description ))))))
		      ((instancep containing-structure)
		       (setq cell-type :instance)
		       (let ((flavor (instance-flavor containing-structure))
			     (index (1- (%pointer-difference original-address containing-structure))))
			 (if flavor
			     (setq symbol (nth index (flavor-all-instance-variables flavor)))
			     (setq symbol (intern (format nil "INSTANCE-VAR-~A" index))))))
		      ((listp containing-structure)	; Because closures are built in list space.
		       (let ((contents (%make-pointer dtp-locative
						      ;; 2nd from top of stack ..
						      (aref (sg-regular-pdl sg)
							    (- (sg-regular-pdl-pointer sg) %call-state-length 1))))
			     explanation)
			 (multiple-value-setq (symbol explanation)
			     (if *use-fsh-safe*
				 (find-structure-header-safe contents)
				 (%find-structure-header contents)))
			 (if (null explanation)	; No problems with the symbol.
			     (setq cell-type :closure)
			     (setq cell-type :unknown
				   symbol explanation)))))
		(setq symbol explanation))	; Couldn't find the structure, cell-type is still :unknown.
	    (values original-address current-address cell-type symbol containing-structure)))
	(values original-address current-address))))

(defun (:property trans-trap make-ucode-error-function) (ignore sg ignore)
  (without-interrupts
    (multiple-value-bind (original-address current-address cell-type symbol containing-structure)
	(trans-trap-decode sg)
      (if (or (eq cell-type :clos-instance) (eq cell-type :class-slot))
	  (let ((ce-type cell-type)
		(or-address original-address )
		(cu-address current-address )
		(cont-structure containing-structure))
	    (declare (special cont-structure  ce-type or-address cu-address ))
	    (clos-unbound  containing-structure symbol sg))
	  (make-instance (case cell-type
			   ((:value :closure :clos-instance :instance :class-slot) 'unbound-variable)
			   (:function 'undefined-function)
			   (otherwise 'cell-contents-error))
			 :address original-address
			 :current-address current-address
			 :cell-type cell-type
			 :symbol symbol
			 :data-type 
			 ;; look at top of stack
			 (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length)) 
			 :pointer 
			 ;; look at 2nd from top of stack
			 (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length 1))
			 :containing-structure containing-structure
			 :condition-names (case cell-type
					    (:value    '(unbound-symbol))
					    (:closure  '(unbound-closure-variable))
					    (:instance '(unbound-instance-variable))
					    (:function '(undefined-function))
					    (otherwise '(bad-data-type-in-memory))))))))


1;; INSTANCE-LACKS-INSTANCE-VARIABLE 
;; INSTANCE-LACKS-INSTANCE-VARIABLE M-C M-A
;; First arg is instance variable
;; Second arg is flavor instance
;; Signaled by LOCATE-IN-INSTANCE*

(defun (:property instance-lacks-instance-variable  make-ucode-error-function) (ignore sg ete)
  (let ((instance (sg-contents sg (third ete)))
	(slot-name (sg-contents sg (second ete))))
    (if (ticlos:clos-instance-p instance)	
	(let* ((pdl (sg-regular-pdl sg))
	       (frame (sg-top-frame sg))
	       (fef (rp-function-word  pdl frame)))
	  (if (eq fef #'ticlos:locate-slot)
	      (eh-slot-missing (ticlos:class-of instance) instance  slot-name (sg-frame-arg-value sg frame 2)
			       (sg-frame-arg-value sg frame 3) sg)
	      (MAKE-INSTANCE (QUOTE ERROR)
		       :CONDITION-NAMES (QUOTE (INSTANCE-LACKS-INSTANCE-VARIABLE))
		       :FORMAT-STRING "There is no slot ~S in instance ~S."
		       :FORMAT-ARGS (LIST slot-name instance))))
	(MAKE-INSTANCE (QUOTE ERROR)
		       :CONDITION-NAMES (QUOTE (INSTANCE-LACKS-INSTANCE-VARIABLE))
		       :FORMAT-STRING "There is no instance variable ~S in instance ~S."
		       :FORMAT-ARGS (LIST slot-name instance)))))

(defmethod (cell-contents-error :report) (stream)
  (let* ((contents-changed
	   (or (neq (%p-data-type current-address) data-type)
	       (neq (%p-pointer current-address) pointer)))
	 (verb (if contents-changed "was" "is")))
    (case cell-type
      (:value    (format stream "The variable ~S ~A unbound." symbol verb))
      (:function (format stream "The function ~S ~A undefined." symbol verb))
      (:closure  (format stream "The variable ~S ~A unbound (in a closure value-cell)." symbol verb))
      (:instance (format stream "The instance variable ~S ~A unbound in ~S." symbol verb containing-structure))
      (:clos-instance (format stream "The instance slot ~S ~A unbound in ~S." symbol verb containing-structure))
      (:class-slot (format stream "The class slot ~S ~A unbound in an instance of class ~S."
			   symbol verb (ticlos:class-description-class-object (array-leader containing-structure 0))))
      (otherwise
       (format stream "The word #<~S ~O> was read from location #o~O ~@[(in ~S)~].~:[~;~@
		       Problem in find-structure-header -- ~A.~]"
	       (q-data-types data-type)
	       pointer
	       (%pointer address)
	       (let ((area (%area-number address)))
		 (if area (area-name area)))
	       (and (eq cell-type :unknown)
		    (not (null symbol)))
	       symbol)))))	

(defmethod (cell-contents-error :after :print-error-message) (sg brief stream)
  (declare (special *error-locus-frame*))
  (unless brief
    (let ((contents-changed
	    (or (neq (%p-data-type current-address) data-type)
		(neq (%p-pointer current-address) pointer))))
      (case cell-type
	((:value :closure :instance :clos-instance :class-slot)
	 (when contents-changed
	   (print-new-cell-contents "It now has the value" current-address stream)))
	(:function
	 (when contents-changed
	   (print-new-cell-contents "It now has the definition" current-address stream))
	 (when (and (symbolp symbol)
		    (get symbol 'compiler:opcode))
	   (let ((fn (rp-function-word (sg-regular-pdl sg)
				       (sg-out-to-interesting-frame sg *error-locus-frame*))))
	     (format stream
		     "Note: ~S is supported as a function when used in compiled code.~%"
		     symbol)
	     (if (consp fn)
		 (format stream
			 "You may have evaluated the definition of ~S, which is now not compiled.~%"
			 (function-name fn)))))
	 )
	(otherwise
	 (when contents-changed
	   (print-new-cell-contents "The cell now contains" current-address stream)))))))
(eval-when (compile)
(defun build-decode-table ()
  (let ((arr (make-array 128 :element-type 'bit))
	(ar(compiler:INSTRUCTION-DECODE-TABLE)))
    (dotimes (i 128)
      (if (or (null (AREF ar i)) (get (AREF ar i) 'compiler:no-reg))
	  (setf (aref arr i ) 0)
	  (setf (aref arr i ) 1)))
    arr)))
;;;02/21/89 clm - problem occurred when trying to access an unbound class (i.e., shared) 
;;;slot.  In this case the local instance was being set to NIL; we then called SLOT-UNBOUND
;;;with the class of NIL as the class argument and NIL as the instance argument.  This would 
;;;result in a no-applicable method error for SLOT-UNBOUND on built-in-class NIL.  The fix: if
;;;we can't obtain the instance, pass the class obtained from the containing structure
;;;and NIL for the instance.   
(defun clos-unbound ( containing-structure symbol sg)
  (declare (special ce-type))
  (let* ((instance
	  (if (neq :class-slot ce-type)
	      containing-structure 
	      (let* ((pdl (sg-regular-pdl sg))
		     (frame (sg-top-frame sg))
		     (fef (rp-function-word  pdl frame))
		     (lc (1- (rp-location-counter-offset pdl frame)))
		     (wd (FEF-INSTRUCTION fef lc)))
		(if (aref #.(build-decode-table) (ASH WD -9))
		    (if ( < (ldb (byte 3 6) wd) 3)
			;;fef address mode
			(let ((self-ref (SYSTEM:%P-POINTER-OFFSET fef (ldb (byte 6 0) wd))))
			  (if (zerop (ldb sys:%%CLOS-SELF-REF-INSTANCE-REF-ADDRESSING-MODE self-ref ))
			      ;;arg number
			      (sg-frame-arg-value sg frame
						  (ldb sys:%%CLOS-SELF-REF-INSTANCE-REF-INDEX self-ref ))
			      ;;local
			      (sg-frame-local-value sg frame
						    (ldb sys:%%CLOS-SELF-REF-INSTANCE-REF-INDEX self-ref ))))


			)))))
	 (class (if instance (ticlos:class-of instance)
		    (when (eq (named-structure-p containing-structure)
				  'ticlos:class-slot-structure )
			  (let ((class-description (array-leader  containing-structure 0)))
			    (ticlos:class-description-class-object class-description )))))
	 )
      
    (let* ((*sg* sg)
	   (new-value (ticlos:slot-unbound class instance symbol)))
      (declare (special *sg*))
      (sg-regpdl-insert sg new-value)  ;; ** used to be (SG-REGPDL-PUSH VALUE SG)
      (sg-proceed-micro-pc sg 'trans-trap-restart))
    (setf (sg-current-state sg) sg-state-resumable)
    (proceed-sg sg)))

(defun eh-slot-missing (class instance slot-name operation new-value sg)
  (let* ((*sg* sg)
	 (new-value (ticlos:slot-missing class instance slot-name operation new-value)))
    (declare (special *sg*))
    (sg-regpdl-insert sg new-value)  ;; ** used to be (SG-REGPDL-PUSH VALUE SG)
    (sg-proceed-micro-pc sg 'trans-trap-restart))
  (setf (sg-current-state sg) sg-state-resumable)
  (proceed-sg sg))

(defun default-slot-missing (ignore instance slot-name operation)
  (declare (special eh:*sg*)
	   (ignore operation))
  (if (boundp '*sg*)
      (let ((sg *sg*)
	    (co (MAKE-INSTANCE (QUOTE ERROR)
			       :CONDITION-NAMES (QUOTE (INSTANCE-LACKS-INSTANCE-VARIABLE))
			       :FORMAT-STRING "There is no slot  ~S in instance ~S."
			       :FORMAT-ARGS (LIST slot-name instance))))
	(let ((signal-result
		(eh:sg-funcall-no-restart
		  sg 'signal-condition
		  co (send co :ucode-proceed-types)
		  t )))
	  ;; Try to proceed with a proceed type
	  (when (car signal-result) 
	    (lexpr-send co :proceed-ucode-with-args
			(car signal-result) sg (cdr signal-result))
	    (setf (sg-current-state sg) sg-state-resumable)
	    (proceed-sg sg))
	  (ferror nil "Proceed-type ~S~%not handled by microcode error ~S."
		  (car signal-result) co)))
      (multiple-value-bind (ignore value)
	  (signal 'error
		  :CONDITION-NAMES (QUOTE (INSTANCE-LACKS-INSTANCE-VARIABLE))
		  :FORMAT-STRING "There is no slot  ~S in instance ~S."
		  :FORMAT-ARGS (LIST slot-name instance))
	value)))

(defun default-slot-unbound (ignore instance name)
  (declare (special eh:*sg*  ce-type cont-structure or-address cu-address))
  (declare (ignore instance))
  (if (boundp 'eh:*sg*)
      (let ((eh:sg eh:*sg*))
	(let ((co (make-instance 'unbound-variable
				 :address or-address
				 :current-address cu-address
				 :cell-type ce-type
				 :symbol name
				 :data-type 
				 ;; look at top of stack
				 (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length)) 
				 :pointer 
				 ;; look at 2nd from top of stack
				 (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length 1))
				 :containing-structure cont-structure
				 :condition-names '(unbound-instance-variable))))
	  (let ((signal-result
		  (eh:sg-funcall-no-restart
		    sg 'signal-condition
		    co (send co :ucode-proceed-types)
		    t )))
	    ;; Try to proceed with a proceed type
	    (when (car signal-result) 
	      (lexpr-send co :proceed-ucode-with-args
			  (car signal-result) sg (cdr signal-result))
	      (setf (sg-current-state sg) sg-state-resumable)
	      (proceed-sg sg))
	    (ferror nil "Proceed-type ~S~%not handled by microcode error ~S."
		    (car signal-result) co))))
      (multiple-value-bind (ignore value)
	  (signal 'unbound-variable
		  :address or-address
		  :current-address cu-address
		  :cell-type ce-type
		  :data-type (sys:%p-data-type or-address)
		  :pointer (%p-pointer or-address)
		  :symbol name
		  :containing-structure cont-structure
		  :condition-names '(unbound-instance-variable))
	value)))

;;12/15/88 clm - changed so that number of locals is taken from the fef header, not 
;;the debug-info-struct
(defun fef-number-of-locals (function)
  2"Get number of locals in function from the fef header."*
  (if (= (%p-ldb si::%%fef-header-call-type function)
		    %fef-call-long)
		 (%p-ldb-offset %%fef-long-args-number-of-locals function %fef-first-optional-word )
		 (%p-ldb %%FEF-HEADER-Number-Locals  function)) )


