;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10. -*-

;;;			      RESTRICTED RIGHTS LEGEND
;;;
;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.


;;  8/10/88 DNG - Moved FLAVOR-CLASS flavor definition to file 
;;		"standard-type-class" because it is also needed while
;;		compiling file "make-instance".
;;  3/18/89 DNG - Removed kernel patches - now included in the release 6 source.

(defclass-primitive flavor-class    (class) ()
  (:metaclass class-class)
  (:documentation "All class objects for flavors are members of this class."))

(defmethod (flavor-class :class-name) ()
  (sys:flavor-name class-description))
 
;;  9/23/88 DNG - Fixed to not append T.
;; 11/21/88 DNG - Fixed to default to vanilla-flavor.
(defmethod (flavor-class :class-direct-supers) ()
  (or (sys:flavor-depends-on class-description)
      (if (getf (sys:flavor-plist class-description) :no-vanilla-flavor)
	  '(t)
	'(sys:vanilla-flavor))))

(defmethod (flavor-class :class-direct-subclasses) ()
  (let ((environment (class-description-environment class-description)))
    (loop for flavor-name in (sys:flavor-depended-on-by  class-description)
	  collect (class-named flavor-name nil environment))))

;;  9/23/88 DNG - Fixed to work right on an uncomposed flavor [such as PATHNAME].
;;		Redesigned to use COLLECT-BODY instead of NCONC for speed.
;; 10/28/88 DNG - Updated to use environment.
(defmethod (flavor-class :class-precedence-list) ()
  (let ((environment (class-description-environment class-description)))
    (collect-body
      (collect self)
      (let ((all (sys:flavor-depends-on-all class-description)))
	(if (null all)				; uncomposed abstract flavor
	    ;; This will not be a complete list, but maybe it is good enough.
	    (dolist (x (sys:flavor-depends-on class-description))
	      (collect (class-named x nil environment)))
	  (dolist (x (cdr all))			; all except self
	    (collect (class-named x nil environment)))))
      (collect *t-class*))))

;; 10/17/88 DNG - Added to speed up SUBCLASSP by avoiding consing and  
;;		minimizing function calls.
;; 10/28/88 DNG - Updated to use environment.
;; 12/07/88 DNG - Fix to use environment in test for proper name.
(defmethod (flavor-class :subclassp) (other-class)
  ;; Is the other-class contained in the class-precedence list for this class?
  (declare (optimize (speed 3) (safety 0))) ; because we spend a lot of time here.
  (declare (inline class-name find-class))
  (or (eq other-class *t-class*)
      (and (typep other-class 'flavor-class)
	   (let ((other-name (send other-class :class-name))
		 (environment (class-description-environment class-description)))
	     (and (eq other-class (find-class other-name nil environment)) ; a proper name
		  (let ((all (sys:flavor-depends-on-all class-description)))
		    (if all	
			(member other-name all :test #'eq)
		      ;; else uncomposed abstract flavor
		      (or (member other-name (sys:flavor-depends-on class-description) :test #'eq)
			  (dolist (x (sys:flavor-depends-on class-description) nil)
			    (when (subclassp (class-named x nil environment) other-class)
			      (return t)))))
		    ))))))

(defmethod (flavor-class :compose-class)()
  (SYSTEM:COMPOSE-FLAVOR-COMBINATION class-description ))


(defmethod (flavor-class :class-direct-slots) ()
  (flet ((get-initargs ( iv)
		       (let* ((acons (rassoc iv (sys:flavor-inittable-instance-variables class-description) :test #'eq)))
			 (if acons `(,(car acons)) nil))))
    (mapcar
      #'(lambda (iv &aux (initform (and (consp iv) (cdr iv))) (iv (if (consp iv) (car iv) iv)))
	  (internal-make-slot-description :name iv
					  :initform initform
					  :allocation :instance
					  :type t
					  :initargs (get-initargs iv)
					  ))
      (sys:flavor-local-instance-variables class-description))))

;;   3/18/89 DNG - Moved SYS:MAKE-FLAVOR-INSTANCE to "KERNEL;RUNTIME-FLAVOR".

;;Since prototype is an anonymous class, its name should be nil.
(defmethod (flavor-class :class-prototype ) ()
  (or (class-description-class-prototype class-description)
      (let ((proto (sys:make-flavor-instance (class-name self))))
	(setf (class-description-class-prototype class-description)
	      proto)
	proto)))

(defmethod (flavor-class :add-method-on-specializer) ignore)

(defmethod (flavor-class :remove-method-on-specializer) ignore)

(defmethod (flavor-class :slot-value-using-class) (instance slot-name)
  (contents (sys:locate-in-instance  instance slot-name)))

(defmethod (flavor-class :set-slot-value-using-class) (instance slot-name value)
  (setf (contents (sys:locate-in-instance instance slot-name )) value))

;;  2/25/89 DNG - Added the next 4 methods.
(defmethod (flavor-class :class-mapped-slot-names) ()
  (getf (class-description-plist class-description) 'mapped-slot-names))

(defmethod (flavor-class :set-class-mapped-slot-names) (list)
  (setf (getf (class-description-plist class-description) 'mapped-slot-names) list))

(defmethod (flavor-class :class-mapped-supers) ()
  (getf (class-description-plist class-description) 'class-mapped-supers))

(defmethod (flavor-class :set-class-mapped-supers) (value)
  (setf (getf (class-description-plist class-description) 'class-mapped-supers) value))

;;  2/25/89 DNG - Added this method to enable efficient access to flavor 
;;		instance slots in a CLOS method.
;; Note that this is the same as the method for STANDARD-CLASS; someday might 
;; want to unify these under a common metaclass.
(defmethod (flavor-class :slot-reference-index) (slot-name)
  (or (position slot-name (the list (class-mapped-slot-names self)) :test #'eq)
      (prog1 
	(progn
	  (setf (class-mapped-slot-names self)
		(nconc (class-mapped-slot-names self) (cons slot-name nil)))
	  (1- (length (class-mapped-slot-names self))))
	(update-mapping-tables self))))


;; 11/21/88 DNG - Added the following four methods.
(defmethod class-direct-initarg-defaults ((class flavor-class))
  (let ((def-initarg-alist '()))
    (declare (list def-initarg-alist))
    (do ((tail (getf (sys:flavor-plist (class-description class))
		     :default-init-plist)
	       (cddr tail)))
	((null tail))
      (let ((form (second tail)))
	(push (list (first tail)  `(lambda () ,form) form)
	      def-initarg-alist)))
    (nreverse def-initarg-alist)))

(defmethod class-direct-slot-initargs ((class flavor-class))
  (mapcar #'(lambda (x)
	      (list (car x) (cdr x)))
	  (sys:flavor-inittable-instance-variables class-description) ; ((:key . name)...)
	  ))

(defmethod class-direct-initargs ((class flavor-class))
  (mapcar #'car (sys:flavor-inittable-instance-variables (class-description class))))
(defmethod class-initargs ((class flavor-class))
  (sys:flavor-all-inittable-instance-variables (class-description class)))


;;  8/24/88 DNG - Updated to handle a string in the FLAVOR-PACKAGE slot; this 
;;	occurs in the PROCESS and PATHNAME flavors.
;;  3/18/89 DNG - Changed from a function named FIX-FLAVOR-CLASSES to 
;;		top-level code so that it doesn't stay around.
(eval-when (load)
  (dolist (f sys:*all-flavor-names*)
    (let((fl (get f 'sys:flavor))
	 (class (class-named f t)))
      (when (and fl  (typep fl 'sys:flavor ))
	(when (null class)
	  (setf class (sys:make-flavor-instance 'flavor-class :class-description fl))
	  (setf (class-named f ) class))
	;; Genasys version 3 still records the package name in the old place, so the 
	;; following is needed to fix up any cold-loaded flavors.  -- DNG 4/10/89
	(let ((pkg (class-description-class-object fl)))
	  (when (or (packagep pkg)
		    (and (stringp pkg) (setq pkg (find-package pkg))))
	    (setf (si:flavor-definition-package fl)
		  pkg)))
	(setf (class-description-class-object fl) class)))))

;;  3/18/89 DNG - patches for DEFFLAVOR1, COMPILE-AT-APPROPRIATE-TIME, etc. 
;;		integrated into file "KERNEL;FLAVOR."
;; Patches for GET-FLAVOR-TRACING-ALIASES, etc. moved to "KERNEL;RUNTIME-FLAVOR".

;; 11/29/88 DNG - Original.
(defmethod (flavor-class :set-flavor) (fl)
  ;; Invoked by PERFORM-FLAVOR-REDEFINITION when the flavor is changed incompatibly.
  (setf (ticlos:class-named (sys:flavor-name fl) t (ticlos:class-description-environment fl))
	(setf (class-description-class-object fl)
	      (sys:make-flavor-instance 'ticlos:flavor-class :class-description fl)))
  fl)



(defmethod (flavor-class :expand-defclass) (defclass-form &optional environment)
  ;; 11/19/88 DNG - Original.
  (declare (ignore environment))
  (destructuring-bind (ignore class-name includes slot-options &rest class-options)
		      defclass-form
    (let ((def-initargs nil)
	  (def-initarg-plist nil)
	  metaclass-name 
	  (defflavor-options '()))
      (declare (list def-initargs def-initarg-plist defflavor-options))
      (dolist (option class-options)
	(with-pattern-matching option
	    ((`(:default-initargs . ,def-initargs) t
	      (do* ((di def-initargs (cddr di))
		    (def-initarg (car di)(car di))
		    (form (second di)(second di)))
		   ((endp di))
		(when (getf def-initarg-plist def-initarg)
		  (non-fatal-error ':impossible ":DEFAULT-INITARG ~s is specified twice."
				   def-initarg))
		(push-key-and-value def-initarg form def-initarg-plist)))
	     (`(:metaclass ,metaclass-name) t)
	     (t (push option defflavor-options)))))
      (setq defflavor-options (nreverse defflavor-options))
      (unless (null def-initarg-plist)
	(push `(:default-init-plist . ,def-initarg-plist) defflavor-options))
      (let* ((slot-description-list (parse-slot-options (get-default-class-metaclass-prototype)
							slot-options class-name))
	     (inittable '())
	     (ivars (mapcar #'(lambda (sd)
				(declare (type slot-description sd))
				(let ((name (slot-name sd)))
				  (unless (eq (slot-allocation sd) :instance)
				    (non-fatal-error :impossible
						     "Slot option ~S not supported by meta class ~S."
						     :allocation (type-of self)))
				  (dolist (i (slot-initargs sd))
				    (unless (and (keywordp i)
						 (equal (symbol-name i) (symbol-name name)))
				      (non-fatal-error :impossible
						       "Slot ~S option ~S ~S not valid for meta class ~S; must be ~S ~S."
						       name :initarg i (type-of self)
						       :initarg (intern name sys:*keyword-package*)))
				    (push name inittable))
				  (if (non-empty-p (slot-initform sd))
				      (list name (slot-definition-initform sd))
				    name)))
			    slot-description-list))
	     )
	(declare (list slot-description-list inittable ivars))
	(unless (null inittable)
	  (push `(:inittable-instance-variables . ,(nreverse inittable)) defflavor-options))
	(collect-body
	  (collect `(defflavor ,class-name ,ivars ,includes
		      . ,defflavor-options))
	  (join (generate-accessors-and-readers self class-name
						slot-description-list)))))))

(defmethod generate-accessors-and-readers ((class class) class-name slot-descriptions)
  ;; Default method, currently only used for flavor-class.
  ;; 11/19/88 DNG - Original.
  (collect-body
    (dolist (slot-description slot-descriptions )
      (let ((writers (slot-writers slot-description))
	    (readers (slot-readers slot-description))
	    (slot-name (slot-name slot-description)))
	(dolist (reader readers )
	  (collect
	    `(defmethod ,reader ((object ,class-name))
	       (slot-value object ',slot-name))
	    ))
	(dolist (writer writers )
	  (collect
	    `(defmethod ,writer (value (object ,class-name))
	       (setf (slot-value object ',slot-name) value))
	    ))))))

;;; hybrid class support, added 12/2/88

(defmethod (flavor-class :add-direct-subclass) ignore)
(defmethod (flavor-class :remove-direct-subclass) ignore)

(defmethod (hybrid-class :set-flavor) (fl)
  ;; Invoked by PERFORM-FLAVOR-REDEFINITION when the flavor is changed incompatibly.
  (setf flavor fl)
  (setf (class-composed-p self) nil) ; if already composed, need to re-do it.
  fl)

;; 1/3/89 Moved (METHOD ALLOCATE-INSTANCE :BEFORE (HYBRID-CLASS)) and 
;;		(METHOD INITIALIZE-INSTANCE :AFTER (SI:VANILLA-FLAVOR)) to file 
;;		"make-instance" because of bootstrapping problems.


;;  2/09/89 DNG - Added the following two methods.  Needed for a CLOS method 
;;		with an argument which is an instance of a flavor that inherits from a 
;;		hybrid class.
(defmethod (flavor-class :class-mapping-tables ) ()
  (getf (sys:flavor-plist class-description) 'class-mapping-tables '()))
(defmethod (flavor-class :set-class-mapping-tables ) (value)
  (setf (getf (sys:flavor-plist class-description) 'class-mapping-tables) value))


;;  5/3/89 DNG - Added this method.
;;  5/5/89 DNG - Add binding of *JUST-COMPILING*.
(defmethod (hybrid-class :after :compute-all-initargs) ()
  ;; Include :INIT-KEYWORDS from any included flavors.
  (when (null (si:flavor-remaining-init-keywords flavor))
    (let* ((compiler:*compile-file-environment* (class-description-environment class-description))
	   (sys::*just-compiling* compiler:*compile-file-environment*))
      (sys:compose-flavor-initializations flavor)))
  (let ((more (delete ':allow-other-keys
		      (union (getf (si:flavor-plist flavor) 'si:required-init-keywords nil)
			     (si:flavor-remaining-init-keywords flavor))
		      :count 1 :test #'eq)))
    (unless (null more)
      (setf (getf (class-description-plist class-description) :all-initargs)
	    (initargs-nunion (getf (class-description-plist class-description) :all-initargs)
			     more)))))
