;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


;; 11/30/88 DNG - Improve error message.
;;  1/07/89 DNG - Signal SYS:INVALID-FUNCTION-SPEC so that ZWEI:SECTIONIZE-FILE-BUFFER can recover.
;;  1/09/89 DNG - Added NOERRORP argument.
(defun get-method-spec-object (gfunc qualifiers specializers &optional fspec-to-create noerrorp)
  (unless (= (length specializers) (length (generic-function-argument-precedence-order gfunc)))
    (if noerrorp
	(return-from get-method-spec-object nil)
      (ferror 'sys:invalid-function-spec
	      "The number of specializers given to FIND-METHOD does not match the number
expected by the generic function ~s,
which was defined with lambda-list ~a in file \"~a\"."
	      gfunc (generic-function-lambda-list gfunc)
	      (let ((name (generic-function-name gfunc))
		    (code (generic-function-discriminator-code gfunc)))
		(and (or (eq code name)
			 (eq code (sys:fdefinition-safe (generic-function-name gfunc) t)))
		     (sys:get-source-file-name name 'defun))))))
  (let* ((keys (reorder-parameter-specializers gfunc  specializers)))
    (do ((ml (find-method-spec-objects  (generic-function-method-list gfunc) keys) (cdr ml)))
	((null ml) (and fspec-to-create
			(insert-method-spec-object gfunc keys nil qualifiers  fspec-to-create)
			(get-method-spec-object gfunc qualifiers specializers nil)))
      (when (same-qualifier-p qualifiers  
			      (method-spec-object-qualifiers (car ml)))
	(return (car ml))))))

;; 11/08/88 DNG - Modify to error in GET-GENERIC-FUNCTION-OBJECT instead of 
;;	FIND-METHOD on NIL.  Allow the second element of the function spec to be 
;;	the actual generic function object instead of it's name, and allow the 
;;	argument to be an :INTERNAL function spec - these are needed for 
;;	handling anonymous generic functions.
;; 12/07/88 DNG - Allow receiving the generic function as an argument - for 
;;		use by MAKE-COMBINED-METHOD.
;;  1/31/89 DNG - Change ERRORP default to NIL for use by DEFMETHOD - don't 
;;		need another error if the method creation failed.
(defun method-named (method-spec &optional (errorp nil) generic-function)
  (if (eq (car-safe method-spec) 'method)
      (let* ((rest (nthcdr 2 method-spec))
	     (qualifiers
	       (collect-body
		 (loop (when (listp (car rest ))
			 (return))
		       (collect (pop rest)))))
	     (specializers (mapcar #'canonicalize-class-spec (pop rest))))
	(*find-method (or generic-function
			  (let ((gname (second method-spec)))
			    (if (sys:typep-structure-or-flavor gname 'generic-function)
				gname
			      (get-generic-function-object (sys:fdefinition-safe gname t)))))
		      qualifiers specializers errorp ))
    (or (sys:get-debug-info-field (sys:get-debug-info-struct method-spec t)
				  :method)
	(error "~S is not the name of a method." method-spec))))

;;  8/11/88 DNG - Return NIL instead of signalling error when operation is FDEFINEDP.
;;		Added support for FUNDEFINE.   
;; 11/08/88 DNG - Allow the second element of the function spec to be 
;;	the actual generic function object instead of it's name - this is needed 
;;	for handling anonymous generic functions.
;; 11/14/88 DNG - Clarify error messages.  Return second value for FDEFINEDP.
;;  1/03/89 DNG - Redesigned for DWIMIFY support.  Signal 
;;		SYS:INVALID-FUNCTION-SPEC when reporting error so 
;;		ZWEI:SECTIONIZE-FILE-BUFFER can recover.  Use METHOD-SPEC-OBJECT-PLIST for 
;;		the GET and PUTPROP operations.
;;  1/09/89 DNG - Modify to not error on GET with wrong number of specializers 
;;		-- needed for (:METHOD ZWEI:ZMACS-BUFFER :AFTER :KILL).
;;  1/16/89 DNG - Fix VALIDATE-FUNCTION-SPEC for generic function object as 2nd element of fspec.
;;  2/16/89 DNG - Extend DWIMIFY support to look for methods on superclasses of the classes specified.
;;  3/16/89 DNG - Use updated DWIMIFY-PACKAGE-2.
(defun (:property method sys:function-spec-handler) (function function-spec &optional arg1 arg2)
  (let* ((gname (second function-spec))
	 (rest (nthcdr 2 function-spec))
	 (qualifiers
	   (collect-body
	     (loop (when (listp (car rest))
		     (return))
		   (collect  (pop rest)))))
	 (specializer-names (pop rest)))
    (if (eq function 'sys:validate-function-spec)
	;; This checks for syntactic validity without requiring that the generic 
	;; function and classes be defined.  Thus, this indicates whether the name 
	;; is acceptable to FDEFINEDP but does not necessarily mean that FDEFINE 
	;; would be valid.  This interpretation is necessary for 
	;; SYS:DWIMIFY-PACKAGE-1 to work right.
	(and (null rest)
	     (consp specializer-names)
	     (or (sys:typep-structure-or-flavor gname 'generic-function)
		 (sys:validate-function-spec gname))
	     (every #'(lambda (x) (or (symbolp x)
				      (sys:classp x)
				      (individual-typep x)))
		    (the list specializer-names)))
      (let ((gfunc (if (sys:typep-structure-or-flavor gname 'generic-function)
		       gname
		     (let ((def (sys:fdefinition-safe gname t)))
		       (and def
			    (generic-function-p def)
			    (get-generic-function-object def)))))
	    (specializers (ignore-errors (mapcar #'canonicalize-class-spec specializer-names))))
	(cond
	  ((eq function 'sys:dwimify)
	   (catch 'sys:dwimify-package
	     (flet ((try-method-spec (new-spec)
			;; Ask the user if this is OK; throws to 'sys:dwimify-package if yes.
			(sys:dwimify-package-2 new-spec nil arg1 arg2 t)
			))
	       (let* ((fname (if gfunc gname (sys:dwimify-package gname 'generic-function-p)))
		      (class-names
			(mapcar #'(lambda (name)
				    (if (or (sys:classp name) (individual-typep name))
					name
				      (sys:dwimify-package name 'find-class)))
				(let ((n (length (generic-function-argument-precedence-order
						   (get-generic-function-object fname)))))
				  (cond ((< n (length specializer-names))
					 (firstn n specializer-names))
					((> n (length specializer-names))
					 (append specializer-names
						 (make-list (- n (length specializer-names))
							    :initial-element 't)))
					(t specializer-names))))))
		 (declare (list class-names))
		 (try-method-spec `(method ,fname ,@qualifiers ,class-names))
		 (when qualifiers
		   (try-method-spec `(method ,fname ,class-names)))
		 (let ((combination (clos:generic-function-method-combination gfunc)))
		   (unless (equal combination '(standard))
		     (try-method-spec `(method ,fname ,(first combination) ,class-names))))
		 (try-method-spec `(method ,fname :around ,class-names))
		 (try-method-spec `(method ,fname :before ,class-names))
		 (try-method-spec `(method ,fname :after ,class-names))
		 (dolist (class-name class-names)
		   (dolist (super (if (consp class-name)
				      (and (individual-typep class-name)
					   (class-precedence-list (class-of (second class-name))))
				    (cdr (class-precedence-list (class-named class-name)))))
		     (try-method-spec `(method ,fname ,(substitute (class-name super)
								   class-name class-names
								   :count 1 :test #'eq)))))
		 (multiple-value-bind (handlerp defn)
		     (fdefinedp `(handler ,fname ,class-names))
		   (when handlerp
		     (try-method-spec (function-name defn))))
		 nil)))) ; end of dwimify
	  ((not (and gfunc specializers))
	   (cond  ((eq function 'sys:fdefinedp) nil)
		  ((eq function 'get) arg2)
		  ((null gfunc)
		   (ferror 'sys:invalid-function-spec
			   "The generic function ~s is not defined." (second function-spec)))
		  (t (when (consp specializer-names)
		       (dolist (name specializer-names)
			 (when (and (symbolp name)
				    (not (class-named name t)))
			   (ferror 'sys:invalid-function-spec
				   "Undefined class ~S in function spec ~S." name function-spec))))
		     (ferror 'sys:invalid-function-spec
			     "Invalid qualifier or specializer syntax in function spec ~S"
			     function-spec))))
	  ((eq function 'sys:fdefine)
	   (add-method gfunc
		       (make-method qualifiers specializers arg1
				    ;; Note: we need the original DEFMETHOD lambda-list which
				    ;; was stored by PARSE-METHOD in the :DESCRIPTIVE-ARGLIST
				    ;; field of the debug info, not the real arglist of the FEF.
				    (arglist arg1 nil)))
	   function-spec)
	  (t (let ((meth-spec (get-method-spec-object
				gfunc qualifiers specializers
				(and (member function '( sys:fdefinition-location
							putprop sys:push-property) :test #'eq)
				     function-spec)
				(member function '(sys:fdefinedp get sys:function-parent) :test #'eq ))))
	       (case function
		 (sys:fdefinedp
		  (and meth-spec
		       (let ((def (method-spec-object-function meth-spec)))
			 (and def (values t def)))))
		 (sys:fdefinition
		  (or (method-spec-object-function meth-spec)
		      (ferror 'sys:invalid-function-spec
			      "The method ~s is not defined." function-spec)))
		 (sys:fdefinition-location (locf (method-spec-object-function meth-spec)))
		 (get (if (null meth-spec)
			  arg2
			(getf (method-spec-object-plist meth-spec) arg1 arg2)))
		 (putprop
		  (let ((default-cons-area sys:background-cons-area))
		    (setf (getf (method-spec-object-plist meth-spec) arg2) arg1)))
		 (sys:push-property
		  (let ((default-cons-area sys:background-cons-area))
		    (push arg1 (getf (method-spec-object-plist meth-spec) arg2))))
		 (sys:fundefine (remove-method gfunc (method-spec-object-method meth-spec)))
		 (otherwise (sys:function-spec-default-handler function
							       function-spec
							       arg1 arg2))))))))))

(defprop generic-function-p ( sys:validate-function-spec generic-function-p 
			     fdefinition-location "definition" sys:function-spec-dwimify)
	 sys:dwimify)
(defprop generic-function-p "a valid function spec defined as a generic function"
	 sys:dwimify-error-message)

(setf (get 'find-class 'sys:dwimify)
      (list #'symbolp					; name validator
	    #'(lambda (name) (find-class name nil))	; test for whether defined
	    #'(lambda (name) (locf (get name class-property)))	; location of definition
	    "class definition"				; description
	    #'ignore					; aux dwimifier
	    ))

(defprop find-class "a valid, defined class name" sys:dwimify-error-message)

;; (HANDLER generic-function (class-name ...)) refers to the FEF that is called 
;; when the generic function is called with arguments of the specified classes.
;; Note that the syntax is like METHOD function specs except that qualifiers 
;; are not applicable here.
;; Storing into this changes the value in the method table
;;  which should make it possible to trace and so forth.

(defun (:property handler sys:function-spec-handler) (function function-spec &optional arg1 arg2)
  ;; 12/21/88 DNG - Original.
  (let* ((gname (second function-spec))
	 (gfunc (if (sys:typep-structure-or-flavor gname 'generic-function)
		    gname
		  (let ((fdef (sys:fdefinition-safe gname t)))
		    (and fdef
			 (generic-function-p fdef)
			 (get-generic-function-object fdef)))))
	 (specializer-names (car (last function-spec)))
	 (specializers (ignore-errors (mapcar #'canonicalize-class-spec specializer-names ))))
    (if (not (and gfunc specializers (= (length function-spec) 3)))
	(if (member function '(sys:validate-function-spec sys:fdefinedp) :test #'eq)
	    nil
	  (IF (NULL gfunc)
	      (error "The generic function ~s is not defined." (second function-spec))
	    (progn (when (consp specializer-names)
		     ;; Do it again, letting the error be reported this time.
		     (mapc #'canonicalize-class-spec specializer-names))
		   (error "Invalid qualifier or specializer syntax in function spec ~S" function-spec))))
      (or (eq function 'sys:validate-function-spec)
	  (let ((handler (find-handler gfunc specializers)))
	    (if (null handler)
		(if (eq function 'fdefinedp)
		    nil
		  (error "No applicable method found for ~S." function-spec))
	      (let ((loc (send handler :method-fef-location)))
		(case function
		  (fdefine (setf (contents loc) arg1))
		  (fdefinition (contents loc))
		  (fdefinedp (values t (contents loc)))
		  (fdefinition-location loc)
		  (otherwise (sys:function-spec-default-handler function function-spec arg1 arg2))))))))))