;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


;;;Missing: environment.
  
;; 11/18/88 DNG - Added shortcut for FEF -- just check the flag bit.
(defun generic-function-p (object &aux def)
  (cond ((typep object 'generic-function) t)
	((typep object 'compiled-function)
	 (zerop (sys:%p-ldb-offset sys:%%FEF-Storage-Length-Generic-Function-Flag object 1)))
	((setf def (sys:fdefinition-safe object t))
	 (typep (sys:get-debug-info-field (sys:get-debug-info-struct def t) :generic-function)
		'generic-function))
	((functionp object)
	 (typep (sys:get-debug-info-field (sys:get-debug-info-struct object t) :generic-function)
	      'generic-function))))

;; 11/18/88 DNG - Simplified error check for efficiency.
(defun get-generic-function-object (object)
  (or (sys:get-debug-info-field (sys:get-debug-info-struct object t) :generic-function)
      (error "~S is not a generic function." object)))

;;  4/6/89 DNG - Original.
;; 4/20/89 DNG - Add THE to declare result type.
(defsubst generic-function-named (gfun)
  (declare (arglist generic-function))
  ;; Given a generic function's function spec, FEF, or structure, return the structure.
  (the generic-function
       (if (sys:typep-structure-or-flavor gfun 'generic-function)
	   gfun
	 (get-generic-function-object gfun))))

(defun lambda-list-arguments (ll &aux (rest-p nil))
  ;; Takes a function lambda-list and extract all its arguments names.
  ;; It is used in call-next-method.
  (values
    (collect-body
      (prog (item )
	 start
	    (setf item (pop ll))
	    (when (null item) (go end))
	    (if (member item lambda-list-keywords :test #'eq)
		(case  item 
		  (&aux   (go end))
		  (&rest (setf rest-p t)
			 (collect (pop ll))
			 (go end))
		  (t (go next)))
		(progn
		  (collect item)
		  (go start)))
	 next
	    (setf item (pop ll))
	    (when (null item) (go end))
	    (if (member item lambda-list-keywords :test #'eq)
		(case  item 
		  (&aux   (go end))
		  (&rest (setf rest-p t)
			 (collect (pop ll))
			 (go end))
		  (t (go next)))
		(progn
		  (collect (if (symbolp item)
			       item
			       (if (consp (car item))
				   (second (car item))
				   (car item))))
		  (go next)))
	 end))
    rest-p))

(defun lambda-list-specializer (lambda-list)
  ;; return all the specializers of a method lambda-list.
  ;; Those that can be specialized but where not are specialized with T.
  (collect-body
    (dolist (item lambda-list)
      (when (member item '(&rest &optional &key &aux) :test #'eq)
	(return))
      (unless (member item lambda-list-keywords :test #'eq)
	(collect (if (consp item) item (cons item '(t))))))))

;;  5/8/89 DNG - Change name from UNSPECIALIZE-LAMBDA-LIST to EXTRACT-LAMBDA-LIST to conform to 89-003.
(defun extract-lambda-list (lambda-list)
  (declare (arglist specialized-lambda-list))
  "Given a specialized lambda list (such as accepted by DEFMETHOD), returns the
lambda list with the specializers removed (such as accepted by DEFUN)."
  ;;given a method lambda-list, returns a non specialized one.
  ;;The lambda-list-keywords are preserved.
  (collect-body
    (do* ((items lambda-list (cdr items))
	  (item (car items)(car items)))
	((null items))
      (when (member item '(&rest &optional &key &aux) :test #'eq)
	(join items)
	(return))
      (collect (if (consp item) (car item) item)))))

(deff unspecialize-lambda-list 'extract-lambda-list) ; old name
(compiler:make-obsolete unspecialize-lambda-list extract-lambda-list)

(defun specializable-arguments (lambda-list)
  ;;Return the list of all the arguments that can be specialized.
  ;;lambda-list is a function lambda-list
  (collect-body
    (dolist (item lambda-list)
      (when (member item '(&rest &optional &key &aux) :test #'eq)
	(return))
      (unless (member item lambda-list-keywords :test #'eq)
	(collect item)))))

(defun specialized-arguments (lambda-list)
  ;;Return the list of all the arguments that are specialized.
  ;;lambda-list is a function lambda-list
  (collect-body
    (dolist (item lambda-list)
      (when (member item '(&rest &optional &key &aux) :test #'eq)
	(return))
      (unless (member item lambda-list-keywords :test #'eq)
	(when (consp item ) (collect (first item)))))))
(defun compute-generic-lambda-list-from-method (lambda-list)
  ;;Returns a lambda-list suitable for defgeneric.
  ;;lambda-list is not specialized.
  ;;The rule is : 
  ;;If there is a rest arg and either no &key or a &allow-other-keys,
  ;;            no need to change the lambda-list.
  ;;Otherwise add &key &allow-other-keys.
  (collect-body
    (with-lambda-list lambda-list
	:required #'(lambda (l) (collect l))
	:optionals-before #'(lambda () (collect '&optional))
	:optionals #'(lambda (l) (collect (if (consp l ) (car l) l )))
	:rest #'(lambda (l) (collect '&rest ) (collect l))
	:keys-before #'(lambda () (collect '&key) (exit)))))


(def call-next-method)
;; CALL-NEXT-METHOD is a local function defined by an FLET generated by 
;; function PARSE-METHOD below, using the function body created by 
;; GENERATE-CALL-NEXT-METHOD below.
(defprop call-next-method (&rest method-args) arglist)
(setf (documentation 'call-next-method 'function)
      "A function defined locally within a method.  Calling it invokes the next-most
specific applicable method.  See also NEXT-METHOD-P.  If no arguments are
specified, the original arguments to the current method are passed.")

(def next-method-p)
;; NEXT-METHOD-P is a local function defined by an FLET generated by 
;; function PARSE-METHOD below, using the function body created by 
;; GENERATE-CALL-NEXT-METHOD immediately below.
(defprop next-method-p () arglist) ; doesn't really help until SPR 8644 is fixed
(setf (documentation 'next-method-p 'function)
      "A function defined locally within a method.  It tests whether there is another
method available for CALL-NEXT-METHOD to use.")

;;  3/15/89 DNG - Modified to use NO-NEXT-METHOD instead of ERROR.
;;  4/24/89 DNG - Use new function NO-NEXT-METHOD-ERROR to minimize the amount of inline code.
(defun generate-call-next-method (env lambda-list)
  `((call-next-method (&rest args)
      (declare (notinline method-generic-function))
      (or ,env (apply #'no-next-method-error args))
      (if (null args)
	  (sys:%apply-method (car ,env) (list* ,@lambda-list) (cadr ,env) (cddr ,env))
	  (sys:%apply-method (car ,env) args (cadr ,env) (cddr ,env))))
    (next-method-p ()
       (not (null ,env)))))

    
;; 10/18/88 DNG - Fixed to return gen. fn. object when :method option is 
;;	used.  Don't use DEFMETHOD in order to avoid putting redundant forms into an XLD file.
;; 11/09/88 DNG - Use new function PROCESS-GENERIC-OPTIONS.
;;  1/18/89 DNG - Add use of ADDITIONAL-FORMS.
;;  4/21/89 DNG - Always compile methods that use CALL-NEXT-METHOD.
(defmacro defgeneric (name lambda-list &rest options &environment env)
  (declare (arglist function-spec lambda-list &rest options))
  "Define a generic function.  Syntax:
\(DEFGENERIC function-spec lambda-list {option}* )
option ::= (:ARGUMENT-PRECEDENCE-ORDER {parameter-name}+) |
	   (DECLARE {declaration}+) |
	   (:DOCUMENTATION string) |
	   (:METHOD-COMBINATION symbol {arg}*) |
	   (:GENERIC-FUNCTION-CLASS class-name) |
	   (:METHOD-CLASS class-name) |
           (:METHOD {qualifier}* lambda-list . body)
Predefined method combinations are +, AND, APPEND, LIST, MAX, MIN, NCONC, OR,
and PROGN, with an arg of :MOST-SPECIFIC-FIRST or :MOST-SPECIFIC-LAST."
  (declare (:expr-sxhash 4270956.))
  (multiple-value-bind (option-list method-list)
      (process-generic-options options)
    (let ((form `(ensure-generic-function ',name :lambda-list ',lambda-list
					  ,@option-list)))
      (if (null method-list)
	  form
	`(prog1 ,form
		,@(collect-body
		    (dolist (method method-list)
		      (multiple-value-bind ( ignore ignore function ignore fspec 
					    call-next-method-used additional-forms)
			  (parse-method method name env)
			(join additional-forms)
			(collect `(fdefine ',fspec #',function t))
			(when call-next-method-used (collect `(eval-when (eval) (compile ',fspec))))
			))))))))

;;;(defmacro generic ( lambda-list &rest options)
;;;  ;;We could use a little of error checking at macroexpand time
;;;  (let ((sym-fun (gensym))
;;;	(method-list))
;;;    (setf options
;;;	  (mapcan #'(lambda (option)
;;;		      (case (car option )
;;;			(declare `(:declare ,(cdr option)))
;;;			(:method (push (cdr option) method-list) nil)
;;;			(:method-combination`(:method-combination ,(cdr option)))
;;;			(t  `(,(car option) . ,(cdr option)))))
;;;		  options))
;;;    `(let* ((gf (apply 'make-generic-function  :lambda-list ',lambda-list
;;;		       ,options))
;;;	    (object (get-generic-function-object gf)))
       
;;;       ,@(mapcar #'(lambda(m) `(defmethod object ,@m)) method-list))))

(defun process-generic-options (options)
  (declare (values option-list method-list))
  ;; This function is used by the macros DEFGENERIC and GENERIC-FUNCTION.
  ;; 11/09/88 DNG - Original version of this function separated from DEFGENERIC.
  (let* ((method-list '())
	 (option-list 
	  (mapcan #'(lambda (option)
		      (case (car-safe option )
			(declare `(:declare ',(cdr option)))
			((:argument-precedence-order :method-combination)
			 `(,(car option) ',(if (and (consp (second option))
						    (null (cddr option)))
					       (progn (non-fatal-error ':impossible
									"Obsolete option syntax: ~S; use ~S."
									option `(,(first option) . ,(second option)))
						 (second option))
					     (cdr option))))
			(:method (push (cdr option) method-list)
				 nil)
			((:documentation :generic-function-class :method-class)
			 (unless (eql (length option) 2)
			   (non-fatal-error ':impossible
					     "Wrong number of arguments in generic function option ~S"
					     option))
			 `(,(car option) ',(second option)))
			(t (non-fatal-error ':impossible
					     "Unrecognized generic function option: ~S"
					     option)
			   (and (consp option)
				`(',(car option) ',(second option))))))
		  options)))
    (declare (list option-list method-list))
    (values option-list (nreverse method-list))))

(deff defgeneric-options 'defgeneric)
(compiler:make-obsolete defgeneric-options defgeneric) ; DNG 11/3/88

;; 12/17/88 DNG - Simplify handling of generic-function-class, and use the CLOS MAKE-INSTANCE.
;; 12/17/88 DNG - Report error for invalid argument-precedence-order name.
;; 12/20/88 DNG - Fix for bootstrapping MAKE-INSTANCE.
;; 12/29/88 DNG - Add :ENVIRONMENT argument for use by COMPILER::PREPARE-LOCAL-GENERIC .
(defun create-generic-function (&rest options &key name lambda-list argument-precedence-order
			      declare (documentation nil)
			      (generic-function-class 'standard-generic-function)
			      (method-class (class-named 'standard-method))
			      environment
			      &allow-other-keys)
  (declare (values lambda-expression instance))
  ;;Create the generic function template, puts the object part in the debug-info
  ;;of the resulting object.
  (let* ((arg-desc (check-lambda-list-for-generic-function lambda-list))
	 (ins (apply (if (fboundp 'make-instance) #'make-instance
		       ;; this is needed for creating the MAKE-INSTANCE generic-function.
		       #'sys:make-flavor-instance)
		     (if environment
			 (class-named generic-function-class nil environment)
		       generic-function-class)
		     :method-class method-class :allow-other-keys t options))
	 (specializable-ll (specializable-arguments lambda-list))
	 (rest-p (or (plusp (arglist-description-n-optional arg-desc))
		     (arglist-description-&rest-p arg-desc)
		     (arglist-description-&key-p arg-desc))))
    (declare (list specializable-ll))
    (if  argument-precedence-order
	 (setf (generic-function-reorder-specs ins)
	       (if (= (length specializable-ll) (length argument-precedence-order))
		   (mapcar #'(lambda (x)
			       (or (position x specializable-ll)
				   (error "Generic function ~S option ~S mentions
 ~S, which is not among the required arguments ~S."
					  name :argument-precedence-order x specializable-ll)
				   ))
			   argument-precedence-order)
		   (error "The length of argument-precedence-order does not match the number of required arguments
for generic-function ~S" ins)))
	 (setf (generic-function-argument-precedence-order ins)
	       specializable-ll)) 
    (setf (generic-function-arglist-description ins) arg-desc)
    (values
      `(named-lambda (,name  (:generic-function ,ins :DESCRIPTIVE-ARGLIST ,lambda-list))
		      ;;The lambda-list needs to be changed. We must be sure to accept any keywords
		      ;;At this point. The checking is done by the lambda-list of the handler.
		      ;;The original lambda-list can be put as a declared arglist.
		      (,@specializable-ll ,@(and rest-p '(&rest .rest.)))
		      ,.(if documentation (cons documentation nil) nil)
		      ,@(if declare `((declare ,@declare)) nil)
		      (declare (unspecial ,@specializable-ll ))
		      ,@(and rest-p '(.rest.))
		      . ,(compute-main-body ins))
      ins)))

(defun make-generic-function (&rest options)
   (compile nil (apply #'create-generic-function options)))

;;  8/15/88 DNG - Fixed to not lose original arglist.
;; 12/14/88 DNG - Use CLOS:MAKE-INSTANCE instead of TICL:MAKE-INSTANCE.
;; 12/17/88 DNG - Removed FBOUNDP check because MAKE-INSTANCE is now defined after SHARED-INITIALIZE.
;; 12/29/88 DNG - Add use of ENVIRONMENT and NAME arguments.
(defun modify-generic-function (old-fef &rest options
				&key (generic-function-class nil generic-function-class-p)
				(lambda-list nil llp) name environment &allow-other-keys)
  (let* ((ins (get-generic-function-object old-fef))
	 gfun-class)
    (when (and generic-function-class-p
	       (not (class-eql (setq gfun-class (class-named generic-function-class nil environment))
			       (class-of ins))))
      (let ((new-ins (make-instance gfun-class
				    :method-list (generic-function-method-list ins)
				    :name (generic-function-name ins)
				    :method-class (generic-function-method-class ins)
				    :lambda-list (generic-function-lambda-list  ins)
				    :argument-precedence-order (generic-function-argument-precedence-order  ins)
				    :declare (generic-function-declare  ins)
				    :documentation (generic-function-documentation  ins)
				    :method-combination (generic-function-method-combination  ins)
				    :arglist-description (generic-function-arglist-description ins)
				    :reorder-specs (generic-function-reorder-specs ins))))
	(sys:structure-forward ins new-ins)
	(setf ins new-ins)))
    (apply ins  :modify-generic-function options)
    (let* ((declare (generic-function-declare ins))
	   (documentation (generic-function-documentation ins))
	   (arg-desc (generic-function-arglist-description ins))
	   (rest-p (or (plusp (arglist-description-n-optional arg-desc))
		       (arglist-description-&rest-p arg-desc)
		       (arglist-description-&key-p arg-desc)))
	   (new-fef (compile name
			 `(named-lambda (,name ( :generic-function ,ins
						    :descriptive-arglist
						    ,(if llp lambda-list
						       (generic-function-lambda-list ins))))
					;;The lambda-list needs to be changed. We must be sure to accept any keywords
					;;At this point. The checking is done by the lambda-list of the handler.
					;;The original lambda-list can be put as a declared arglist.
					(,@(specializable-arguments (generic-function-lambda-list ins))
					 ,@(and rest-p '(&rest .rest.)))
					,.(if documentation (cons documentation nil) nil)
					,@(if declare `((declare ,@declare)) nil)
					(declare (unspecial ,@(specializable-arguments (generic-function-lambda-list ins)) ))
					,@(and rest-p '(.rest.))
					. ,(compute-main-body ins))))
	   )
      ;; This is now done in COMPILER:LAP-MFEF
      ;;(sys:%p-dpb-offset 0 sys:%%FEF-Storage-Length-Generic-Function-Flag new-fef 1)
      (setf (generic-function-discriminator-code ins) new-fef)
      new-fef)))


(p-defgeneric-options compute-main-body (ins)
		      ;;This needs to include the real lambda-list of the generic-function
		      ;;And a restp flag. We won't need to call LAMBDA-LIST-ARGUMENTS
		      ) 
;;  8/26/88 DNG - Use new function DISPATCH-ERROR to save space.
;;  4/21/89 DNG - Use new function PREPARE-EFFECTIVE-CODE to save space.  
;;		Removed code to call the result of COMPUTE-EFFECTIVE-CODE because it 
;;		always returns NIL.
(defmethod (standard-generic-function :compute-main-body ) ()
  (let* ((ll (specializable-arguments lambda-list))
    #|
	 (argdesc (generic-function-arglist-description self))
	 (rest-p (or (plusp (arglist-description-n-optional argdesc))
		     (arglist-description-&rest-p argdesc)
		     (arglist-description-&key-p argdesc)))
    |#
         )
    (flet ((natural-order-p  (l)
	     (do ((l l (cdr l))
		  (i 0 (1+ i)))
		  ((endp l) t)
		 (unless (= (car l) i)
		   (return nil)))))
      `((compiler:%dispatch-method
	  (sys:%generic-function-hash-table)
	  ',(or (copy-list (generic-function-reorder-specs self))
		(length (generic-function-argument-precedence-order self))))
     #|
	(let ((function 
		(compute-effective-code ,self
					,@(collect-body
					    (let ((ll ll))
					      (dolist (i  argument-precedence-order)
						(collect ``(eql ,,(pop ll)))))))))
	  (and function (,(if rest-p 'apply 'funcall) function ,@ll ,@(and rest-p '(.rest.)))))
     |#
        (prepare-effective-code ,self . ,(reorder-parameter-specializers self ll))
	(compiler:%dispatch-method
	  (sys:%generic-function-hash-table)
	  ',(if (natural-order-p  (generic-function-reorder-specs self))
		(length (generic-function-argument-precedence-order self))
		(copy-list (generic-function-reorder-specs self))))
	(dispatch-error)
	))))

(defun dispatch-error ()
  (ferror nil "Internal CLOS Error, Can't dispatch"))
(defprop dispatch-error t :error-reporter)

;;;(ticl:defmethod (standard-generic-function :compute-main-body ) ()
;;;  (let* ((ll (specializable-arguments lambda-list))
;;;	(argdesc (generic-function-arglist-description self))
;;;	(rest-p (or (plusp (arglist-description-n-optional argdesc))
;;;		    (arglist-description-&rest-p argdesc)
;;;		    (arglist-description-&key-p argdesc))))
;;;    `((let ((ht (generic-function-method-hash-table ,self)))
;;;	(unless ht
;;;	  (setf ht (build-method-hash-table ,self)))
;;;	(let ((m (find-right-method ht ,@(generic-function-argument-precedence-order self))))
;;;	  (if m
;;;	      (,(if rest-p 'apply 'funcall) (contents m) ,@ll ,@(and rest-p '(.rest.)))
;;;	      (,(if rest-p 'apply 'funcall) (compute-effective-code ,self
;;;					     ,@(collect-body
;;;						 (let ((ll ll))
;;;						   (dolist (i  argument-precedence-order)
;;;						     (collect ``(eql ,,(pop ll)))))))
;;;		     ,@ll ,@(and rest-p '(.rest.)))
;;;	      ))))))

;; 12/17/88 DNG - Set the rest-p flag to the name of the rest argument instead of T.
(defun check-lambda-list (oll &optional (method-p))
  ;;Check the lambda-list of a generic-function or a method
  ;;returns number of required, number of optionals, restp, 
  ;;keyword-list and allow-other-keys-p
  (declare (values nreq nopt rest-p allow-other-keys-p key-list keyp))
  (let ((nreq 0)
	(nopt 0)
	(allow-other-keys-p nil)
	(rest-p)
	(key-list)
	(keyp nil))
    (with-lambda-list oll
	:required #'(lambda (l) (when (not (member  l lambda-list-keywords :test #'eq))
				  (incf nreq)))
	:optionals #'(lambda (l)
		       (when (not (member  l lambda-list-keywords :test #'eq))
			 (if (and (not method-p )(not (symbolp  l)))
			     (bug)
			     (incf nopt))))
	:rest #'(lambda (name) (setf rest-p name))
	:keys-before #'(lambda () (setf keyp t))
	:keys #'(lambda (l)
		  (when (not (member  l lambda-list-keywords :test #'eq))
		    (if (and (not method-p )(not (symbolp l)))
			(bug)
			(let ((key (if (consp l)
				       (car l)
				       l)))
			  (push (if (consp key) (car key) (intern key *keyword-package*)) key-list)))))
	:allow-other-keys-key #'(lambda() (setf allow-other-keys-p t)))
    (values nreq nopt rest-p allow-other-keys-p key-list keyp)))



(defun  check-lambda-list-for-generic-function  (ll)
  (multiple-value-call #'(lambda (n-req n-opt rest-p allow-p keylist keyp)
			   (if (or rest-p allow-p keylist keyp)
			       (make-arglist-description :n-required n-req
							 :n-optional n-opt
							 :&allow-other-keys-p allow-p
							 :&rest-p rest-p
							 :&key-p keyp
							 :keywords keylist)
			     ;; abbreviate to save space
			     (list n-req n-opt)))
		       (check-lambda-list ll)))
			   

(unless (fboundp 'ensure-generic-function) ; redefined in file "meta-object".
 ;; 12/29/88 DNG - Pass :NAME option to MODIFY-GENERIC-FUNCTION.
 (defun ensure-generic-function (fspec &rest options
				&key lambda-list argument-precedence-order
				declare documentation generic-function-class
				method-combination method-class environment)
  "Define a generic function or update its options."
  (declare (ignore lambda-list argument-precedence-order
		   declare documentation generic-function-class
		   method-combination method-class environment))
  (flet ((create-generic-function ()
	   (let ((form (apply #'make-generic-function :name fspec :allow-other-keys t
			      options)))
	     (fdefine fspec form t)
	     form)))
    (let ((def (sys:fdefinition-safe fspec t)))
      (if def
	  (if (generic-function-p def)
	      (apply #'modify-generic-function def :name fspec options)
	      (progn
		(cerror "Discard the current definition of ~s and create a generic function."
			"~s is not a generic function." fspec fspec)
		(create-generic-function)))
	  (create-generic-function)))))
 )


;; 11/16/88 DNG - Fix to insert BLOCK to allow RETURN-FROM.  Use 
;;		EVAL-FOR-TARGET instead of EVAL1 and report error on invalid parameter 
;;		name or specializer syntax.
;; 11/28/88 DNG - Fix handling of declarations when CALL-NEXT-METHOD is used.
;; 12/14/88 DNG - Avoid getting confused by a user-supplied ARGLIST declaration.
;;  1/07/89 DNG - Fix ARGLIST check to permit default values on optionals.
;;  1/17/89 DNG - Evaluate EQL type forms at load time instead of macro expansion time.
;;  1/18/89 DNG - Optimize class (EQL NIL) [which didn't work] to NULL.
;;  2/01/89 DNG - Change EQL handling to use new special form LOAD-TIME-VALUE.
;;  4/18/89 DNG - Fix so that CALL-NEXT-METHOD with no arguments will use the 
;;		original, unaltered, arguments.  [Part 2 of SPR 9570]
;;		Add IGNORE-ERRORS around the call to CODE-WALK so that macro 
;;		expansion errors don't get reported as problems with DEFMACRO.
;;  4/21/89 DNG - Don't call COMPILER:PRE-OPTIMIZE when within Evaluate Buffer.
;;  4/28/89 DNG - Add use of DEFAULT-TYPE declaration to help the compiler 
;;		generate more efficient code.

;;;Should check the arglist better. A nil will be accepted here!
(defun parse-method (things &optional name environment)
  (declare (values method-combination-identifiers specializers function
		   specialized-lambda-list fspec call-next-method-used additional-forms))
  (let* ((additional-forms '())
	 (method-combination-identifiers
	   (collect-body
	     (loop (when(listp (car things ))
		     (return))
		   (collect (pop things)))))
	 (specialized-lambda-list (pop things))
	 (body things)
	 (parameter-and-specializers (lambda-list-specializer  specialized-lambda-list))
	 (lambda-list (extract-lambda-list  specialized-lambda-list))
	 (specializers
	   (mapcar #'(lambda (sp)
		       (unless (symbolp (first sp))
			 (error "Non-symbol ~S found where a method parameter name was expected."
				(first sp)))
		       (let ((spec (second sp)))
			 (cond ((symbolp spec) spec)
			       ((and (consp spec)
				     (eq (car spec) 'eql)
				     (null (cddr spec)))
				(let ((form (second spec)))
				  (when (and compiler:qc-file-in-progress
					     (fboundp 'compiler:macro-cons-area))
				    (setq form
					  (let ((compiler:p1value 'compiler:single-value))
					    (compiler:pre-optimize form t))))
				  (cond ((constantp form)
					 ;; Evaluate it now in current lexical environment.
					 (let ((value (compiler:eval-for-target form environment)))
					   (if (null value)
					       'null
					     `(eql ,value))))
					((or (and compiler:qc-file-in-progress compiler:vars)
					     (compiler:env-vars environment)
					     (compiler:env-functions environment))
					 ;; Generate code to evaluate in proper lexical environment.
					 (setq spec (list 'eql form))
					 ;; Use load-time-value to prevent constant folding or write-protection.
					 (push-end `(rplaca (cdr (compiler:load-time-value ',spec nil)) ,form)
						   additional-forms)
					 spec)
					((or (not compiler:qc-file-in-progress)
					     compiler:qc-file-load-flag)
					 ;; Evaluate it now in current lexical environment.
					 (let ((value (compiler:eval-for-target form environment)))
					   (if (null value)
					       'null
					     `(eql ,value))))
					(t	; Have the loader evaluate it.
					 `(eql ,(cons compiler:eval-at-load-time-marker form))))))
			       ((sys:classp spec) spec)
			       (t (error "Invalid specializer syntax in method parameter ~S." sp)))))
		   parameter-and-specializers))
	 (fspec `(method ,name ,@method-combination-identifiers
			 ,specializers))
	 (argl (gensym))
	 (spec-args  (specializable-arguments lambda-list)))
    (multiple-value-bind (undeclared-body declarations documentation)
	(sys:parse-body body environment t)
      ;; Find and remove any user-supplied ARGLIST declaration which is not 
      ;; congruent with the actual lambda list, because we need to use the 
      ;; :descriptive-arglist to remember the original DEFMETHOD args, and an 
      ;; ARGLIST declaration is not very useful here anyway.  -- DNG 12/14/88
      (do ((decls declarations (rest decls)))
	  ((atom decls))
	(let ((temp (assoc 'arglist (rest (first decls)))))
	  (when (and temp
		     (not (equal (multiple-value-list (check-lambda-list (rest temp) t))
				 (multiple-value-list (check-lambda-list lambda-list t)))))
	    (setf (first decls) (remove temp (the list (first decls)) :test #'eq)))))
      (when (and documentation (record-documentation-p))
	(push documentation declarations))
      (let* ((call-next-method-used
	       (catch 'walk
		 (ignore-errors
		   (compiler:code-walk
		     `(progn . ,undeclared-body)
		     #'(lambda (form)
			 (when (member (car form) '(call-next-method next-method-p) :test #'eq)
			   (throw 'walk t))
			 form)
		     #'identity
		     nil
		     environment
		     :function #'(lambda (form)
				   (when (member (cadr form) '(call-next-method next-method-p) :test #'eq)
				     (throw 'walk t))
				   form) )
		   ;; successfully walked the tree without finding anything.
		   (throw 'walk nil))
		 ;; if there was an error, assume the worst.
		 t))
	     (block-form (cond ((and (symbolp name) (not (null name)))
				`(block ,name . ,undeclared-body))
			       ((and (consp name)
				     (eq (first name) 'setf)
				     (symbolp (second name)))
				`(block ,(second name) . ,undeclared-body))
			       (t `(progn . ,undeclared-body))))
	     (function (if call-next-method-used
			   (let ((subargl (compute-lambda-list-for-method lambda-list nil))
				 (reqargs (mapcar #'(lambda (x) (if (consp x) (car x) x)) spec-args)))
			     `(,(if name
				    'named-lambda
				  'lambda)
			       ,@(and name `((,fspec  (:method nil :DESCRIPTIVE-ARGLIST ,lambda-list))))
			       (,@spec-args ,@(and subargl `(&rest ,argl)))
			       (DECLARE (TICLOS::SPECIALIZERS ,@specializers))
			       ;; Because the same declarations are being put in two places, the magic 
			       ;; flags .ARG. and .AUX. tell the compiler to suppress warnings on 
			       ;; declarations referring to variables not present.
			       ,@(and declarations `((declare (compiler::.ARG.)) . ,declarations))
			       (flet (,@(generate-call-next-method '(SYS:NEXT-METHOD-LIST)
								   `(,@spec-args ,(and subargl  argl))))
				 (declare (inline call-next-method next-method-p))
				 #'call-next-method
				 #'next-method-p
				 (apply #'(lambda ,(append reqargs subargl)
					    ,@(and declarations `((declare (compiler::.AUX.)) . ,declarations))
					    (declare . ,(loop for x in parameter-and-specializers
							      unless (eq (second x) 't)
							      collect `(compiler:default-type
									 ,(second x) ,(first x))))
					    ,@reqargs	; to suppress compiler warnings on unused variables
					    ,block-form)
					,@reqargs ,(and subargl  argl)))))
			 `(,(if name
				'named-lambda
			      'lambda) ,@(and name `((,fspec  (:method nil :DESCRIPTIVE-ARGLIST ,lambda-list
								       :next-methods-not-used t	    ))))
			   ( ,@(compute-lambda-list-for-method lambda-list))
			   (DECLARE (TICLOS::SPECIALIZERS ,@specializers))
			   ,@declarations
			   ,@(specialized-arguments specialized-lambda-list )
			   ,block-form ))))
	(values method-combination-identifiers specializers function
		specialized-lambda-list fspec call-next-method-used additional-forms)))))

(eval-when (eval compile load)
(defun flavor-method-spec-p (spec)
  ;; Given the first argument of a DEFMETHOD, return true if this is a flavor 
  ;; method, false if this is the name of a CLOS generic function.
  (and (consp spec)
       (neq (car spec) 'setf)
       (not (and (symbolp (car spec))
		 (get (car spec) 'si:function-spec-handler)
		 (null (sys:compilation-flavor (car spec) t))))))
)
(define-without-redefinition-query DEFMETHOD
;;  8/26/88 DNG - Use new function ENSURE-GENERIC-FN to simplify the code dumped to an object file.
;; 11/02/88 DNG - Use new function FLAVOR-METHOD-SPEC-P to avoid duplication 
;;		of code with UNDEFMETHOD and the translator.
;; 11/03/88 DNG - Use new function EXPAND-KLUDGE-METHOD .
;; 12/19/88 DNG - Use new function SYS:DEFMETHOD1 to enable this to replace the old DEFMETHOD.
;;  1/17/89 DNG - Add ADDITIONAL-FORMS to enable evaluating EQL type forms at load time.
;;  4/21/89 DNG - Force compilation of methods using CALL-NEXT-METHOD.
 (defmacro DEFMETHOD (name &rest things &environment env)
	 "For CLOS:
  (DEFMETHOD function-spec {method-qualifier}* lambda-list . body)
  Defines a method on a generic function named by function-spec.
  Standard method-qualifiers are :BEFORE, :AFTER, and :AROUND.
  Required arguments in the lambda-list may have the form (var-name class-name)
  indicating that this method applies when the argument is of that class.
For Flavors:
  (DEFMETHOD (flavor-name [daemon-type] operation [:case-sub-operation]) lambda-list . body)
  Defines the method for flavor: flavor-name for the message operation,
  Daemon-type can be one of: :BEFORE :AFTER :AROUND :INVERSE-AROUND :CASE :DEFAULT
	:OR :AND :OVERRIDE :PROGN :LIST :INVERSE-LIST :PASS-ON :APPEND :NCONC.
  :case-sub-operation must be provided for :CASE daemon-type, and is illegal otherwise."
	 
	 (declare (arglist spec lambda-list &body body))
	 (declare (:expr-sxhash 3542393.))	; match releases 3,4,5
	 (if (flavor-method-spec-p name)
	     ;;old flavor defmethod
	     (sys::defmethod1 name (first things) (rest things))
	   ;;Parse out things into
	   ;;method-combination-identifier
	   ;;specialized-lambda-list
	   ;;body
	   (multiple-value-bind ( method-combination-identifiers specializers 
				 function lambda-list fspec call-next-method-used additional-forms)
	       (parse-method things name env)
	     (or (expand-kludge-method name method-combination-identifiers specializers function lambda-list env)
		 ;; Else a real CLOS method.
		 `(progn
		    (ensure-generic-fn ',name ',(compute-generic-lambda-list-from-method
						  (extract-lambda-list lambda-list)))
		    ,.additional-forms
		    (fdefine ',fspec #',function t)
		    ,@(and call-next-method-used ; won't work unless compiled
			   `((eval-when (eval) (compile ',fspec))))
		    (method-named ',fspec ))))))
       )

;; 11/03/88 DNG - Original.
;; 11/17/88 DNG - Fix handling of declarations.
;; 11/23/88 DNG - Use get-flavor.
(defun expand-kludge-method (name method-combination-identifiers specializers 
			     function lambda-list environment)
  (if (and (symbolp name)
	   (fboundp name)
	   (symbolp (first specializers))
	   (eq (symbol-package name) '#,(symbol-package 'defclass))
	   (sys:get-flavor (first specializers) environment)
	   (not (generic-function-p name))
	   (every #'(lambda (x) (eq x 't)) (the list (rest specializers))))
      ;; Pseudo-generic function for something that is currently really a flavor.
      ;; Generate a flavor method.
      (let ((arg1 (first (first lambda-list)))
	    (body (cddr (sys:lambda-exp-args-and-body function))))
	(when (subtypep (first specializers) 'class)
	  (multiple-value-bind (forms decls)
	      (parse-body body environment)
	    (setq body `(,@decls
			 (macrolet ((class-description (&whole form class)
				        (if (eq class ',arg1)
					    'class-description	; use instance variable
					  (macroexpand form))))
			   . ,forms)))))
	`(zlc:defmethod (,(first specializers) ,@method-combination-identifiers
			  ,(intern (symbol-name name) *keyword-package*))
			 (,@(rest lambda-list) &aux (,arg1 self))
	   . ,body))
    nil))

(defun ensure-generic-fn (name lambda-list)
  (unless (generic-function-p name)
    (ensure-generic-function name :lambda-list lambda-list)))

;;This macro is designed to define methods that the system recognizes to be one 
;;of its own. Some optimizations are based on the fact that no user method are applicable
(defmacro defmethod-system (name &rest things)
  (let ((meth-sym (gensym)))
  `(let ((,meth-sym  (defmethod ,name ,@things)))
     (setf (getf (method-plist ,meth-sym) 'system-defined) t)
     ,meth-sym)))

(defun compute-lambda-list-for-method (declared-lambda-list &optional (required t))
  ;;The rule is :
  ;;If there is a &key, then &allow-other-keys must be there too
  (let ((rest-p)
	(keyp)
	(allowp))
    (collect-body
      (with-lambda-list declared-lambda-list
	 :required #'(lambda (l) (when required (collect l)))
	 :optionals-before #'(lambda () (collect '&optional))
	 :optionals #'(lambda(l) (collect l))
	 :rest #'(lambda (l) (collect '&rest ) (collect l) (setf rest-p t))
	 :keys-before #'(lambda () (collect '&key) (setf keyp t))
	 :keys #'(lambda (l) (collect l))
	 :allow-other-keys-key #'(lambda () (collect '&allow-other-keys) (setf allowp t))
	 :aux-before #'(lambda () (if (and keyp (not allowp)) (collect '&allow-other-keys))(collect '&aux))
	 :aux #'(lambda (l) (collect l))
         :end #'(lambda () (if (and keyp (not allowp)) (collect '&allow-other-keys)))))))

;;  3/15/89 DNG - add call to FEF-CONSTANT-REPLACE.
(defun make-method (method-combination-identifiers
		    specializers function lambda-list &optional environment)
  (let* ((m (sys:make-flavor-instance (if (equal '(:combined) method-combination-identifiers)
					  'combined-method
					'standard-method)
				      :lambda-list lambda-list
				      :parameter-specializers
				      (loop for x in specializers
					    collect (canonicalize-class-spec x environment))
				      :qualifiers method-combination-identifiers
				      :function function))
	 (sys:%inhibit-read-only t)
	 (ds (sys:get-debug-info-struct function t)))
    (setf (sys:get-debug-info-field ds :method) m)
    (if (sys:get-debug-info-field ds ':next-methods-not-used)
	(setf (getf (method-plist m) 'next-methods-not-used ) t)
      ;; Plug the method object into the FEF in place of the marker left by GENERATE-CALL-NEXT-METHOD .
      ;; This is obsolete, remove after the next build. -- DNG 4/24/89
      #.(and (< (time:get-universal-time) (time:parse-universal-time "5/10/89"))
	     `(function-constant-replace m '%method% function)))
    m))

(defun function-constant-replace (new old function)
  ;; In the constants table of FEF, replace OLD with NEW.
  ;; 3/9/89 DNG - Original.
  (typecase function
    ( compiled-function 
     (do ((i si:%fef-header-length (1+ i))
	  (lim (truncate (si:fef-initial-pc function) 2))
	  (dtp (sys:%data-type old)))
	 ((>= i lim) nil)
       ;; Beware of DTP-EXTERNAL-VALUE-CELL-POINTER and DTP-SELF-REF-POINTER.
       (when (and (eql (sys:%p-data-type-offset function i) dtp)
		  (equal (sys:%p-contents-offset function i) old))
	 ;; Found it.
	 (setf (sys:%p-contents-offset function i) new)
	 ;; We don't need to look for more than one occurrence because the 
	 ;; compiler will have coalesced EQUAL values.
	 (return t) )))
    ( closure (function-constant-replace new old (closure-function function)))
    ( symbol (function-constant-replace new old (symbol-function function)))
    ( cons ;; not safe to do this because it would destructively modify a 
     ;;		constant in GENERATE-CALL-NEXT-METHOD
     ;;(nsubst new old function)
     )
    ( t (error "~S is not a function." function))))

(defmethod (standard-method :init ) (init-plist)
  (apply
    #'(lambda (&key lambda-list &allow-other-keys)
	(multiple-value-call
	  #'(lambda (n-req n-opt rest-p allow-p keylist keyp)
	      (setf (method-arglist-description self)
		    (if (or rest-p allow-p keylist keyp)
			(make-arglist-description :n-required n-req
						  :n-optional n-opt
						  :&allow-other-keys-p allow-p
						  :&rest-p rest-p
						  :&key-p keyp
						  :keywords keylist )
		      ;; shortcut to save space
		      (list n-req n-opt))))
	      (check-lambda-list lambda-list t)))
    (contents init-plist)))


;;; Method list management.
;;; Its structure is a resursive alist. The top-level is an alist whose key is the class of 
;;; the first argument in argument-precedence-order. The contents in an alist whose key is the 
;;; class of the next argument in argument-precedence-order. When all the arguments are exhausted, 
;;; then the contents is a list of method-spec-objects that hold the methods.

(defun class-eql ( x y)
  ;; eql for class and 'individual
  (or (eq x y)
      (and (individual-typep x) 
	   (individual-typep y)
	   (eql (individual-type x)(individual-type y)))))

(defun canonicalize-class-spec (class-name &optional environment)
  ;;We probably want more error checking at this point.
  (if (individual-typep class-name)
      class-name
      (class-named class-name nil environment)))

(defun find-method-spec-objects (alist args)
  ;; At this point the args have been reordered according to argument-precedence-order. 
  (if (null args)
      alist
      (let ((sublist (assoc (car args) alist :test #'class-eql)))
	(and sublist (find-method-spec-objects  (cdr sublist) (cdr args))))))

(defun insert-method-spec-object (gfunc keys method qualifiers fspec &rest prop)
  (setf (generic-function-method-list gfunc)
	(apply
	  #'insert-in-method-list
	  keys (generic-function-method-list gfunc) method qualifiers fspec prop)))

(defun insert-in-method-list (keys alist method  qualifiers fspec &rest prop)
  (if (null keys)
      (dolist (m alist
		 (let ((spec (make-method-spec-object
			       :function-spec fspec
			       :method method :plist prop
			       :function (and method (*method-function method))
			       :qualifiers qualifiers )))
		   (when method 
		     (setf (method-function-location method ) (locf (method-spec-object-function spec))))
		   (nconc alist (cons spec nil))))
	(when (same-qualifier-p qualifiers (method-spec-object-qualifiers m))
	  (when method
	    (setf (method-spec-object-method m) method)
	    (setf (method-spec-object-function m) (*method-function method))
	    (setf (method-function-location method) (locf (method-spec-object-function m))))
	  (when prop (setf (getf (method-spec-object-plist  m) (car prop)) (cadr prop)))
	  (return alist)))
      (let* ((next-assoc (assoc (car keys) alist :test #'class-eql))
	     (next-alist (cdr next-assoc)))
	(if next-alist
	    (progn
	      (apply #'insert-in-method-list (cdr keys) next-alist method qualifiers fspec prop)
	      alist)
	    (let ((spec (make-method-spec-object
			  :function-spec fspec
			  :method method :plist prop
			  :function (and method (*method-function method))
			  :qualifiers qualifiers )))
	      (when method
		(setf (method-function-location method) (locf (method-spec-object-function spec))))
	      (if next-assoc
		  (progn
		    (nconc next-assoc (cons (build-key  (cdr keys) spec) nil))
		    alist)
		  (nconc alist (cons (build-key  keys spec) nil))))))))


(defun build-key ( keys method)
  (if (null keys)
        method
      (cons (car keys) (cons (build-key  (cdr keys)  method) nil))))


(unless (fboundp 'add-method) ; redefined later in file "metaclass"
  (p-defgeneric-options add-method ((generic-function
				      (if (typep generic-function 'generic-function)
					  generic-function
					(get-generic-function-object generic-function)))
				    method)
			(:documentation "Add a method to a generic function.
The generic function is destructively modified and returned as the result.")) )

;; 11/30/88 DNG - Update combined method that pointed to the method being replaced.
;;  1/17/89 DNG - Fix for adding method to anonymous generic function.
;;  3/13/89 DNG - Call ADD-METHOD-ON-SPECIALIZER for slot accessors.
;;  4/10/89 DNG - Remove old method if it's arg list is not compatible.
(defmethod (standard-generic-function :add-method) ( method)
  (let ((gfunc self))
	 ;;We need to check the lambda-list congruence with the generic-function.
    (unless (= (length (method-parameter-specializers method)) (length argument-precedence-order))
      (error "The number of specializers of method ~s does not match the 
number expected by the generic function ~s." method gfunc))
    (unless (combined-method-p  method)
      (check-lambda-list-for-method method  gfunc))
    (let ((keys (reorder-parameter-specializers gfunc (method-parameter-specializers method)))
	  (existing-method (*find-method gfunc
				       (method-qualifiers method )
				       (method-parameter-specializers method) nil))
	  (method-name (function-name (*method-function method ))))
      ;;at this point, we should go from the class-name to the real class object.
      ;;[first class things]
      (setf method-list (insert-in-method-list keys method-list method (method-qualifiers method)
					       method-name))
      ;;Side effect the method so it points to its generic function.
      (setf (symeval-in-instance method 'generic-function) gfunc)
      (unless (or (not (eq (car-safe method-name) 'method))
		  (eq gfunc (second method-name))
		  (and (variable-boundp discriminator-code)
		       (or (eq discriminator-code (second method-name))
			   (eq discriminator-code
			       (sys:fdefinition-safe (second method-name) t)))))
	(setf (second method-name) gfunc))
       (when (generic-function-method-hash-table gfunc)
	 (unless (handlerp method)
	   (when (and existing-method
		      (incompatible-arg-desc-p
			(method-arglist-description method)
			(method-arglist-description existing-method)))
	     ;; Here when replacing a method that was previously defined with the 
	     ;; wrong number of optional arguments.  Need to make sure that the 
	     ;; combined method is re-generated.
	     (remove-method gfunc existing-method)
	     (setq existing-method nil))
	   (if (and existing-method (same-arglist-signature-p existing-method method)
			(not (and (getf (method-plist existing-method ) 'next-methods-not-used )
				  (null (getf (method-plist method) 'next-methods-not-used )))))
	       (let ((forward (getf (method-plist existing-method) 'forwarded-from)))
		  (when (and forward
			      (eq existing-method (method-forwarded-method forward)))
		    ;; Update combined method that pointed to the method being replaced.
		    (setf (method-forwarded-method forward) method)
		    (setf (getf (method-plist method) 'forwarded-from) forward)))
	     ;;Don't recombine if is a redefinition of the same method.
	     ;;This can be refined if we do inlining of methods inside combined methods.
	     ;;With the new congruence rule, if the new method does not have the same arglist
	     ;; the old one has, then we must redo the combined method.
	     (recombine-methods gfunc keys))
	   ))
       (unless (or existing-method (combined-effective-method-p method))
	 (dolist (k keys )
	   (unless (consp k) 
	     (add-method-on-specializer k  method))))
      gfunc)))

(unless (fboundp 'remove-method) ; redefined later in file "metaclass"
  (p-defgeneric-options remove-method ((generic-function
					 (if (typep generic-function 'generic-function)
					     generic-function
					   (get-generic-function-object generic-function)))
				       method)
			(:documentation 
			  "Remove a method from a generic function.
The generic function is destructively modified and returned as the result.")) )

;; 11/30/88 DNG - Remove combined method that pointed to the method being removed.
;; 12/06/88 DNG - Add type check on method argument.
;;  3/01/89 DNG - Don't RECOMBINE-METHODS if the specializer class is not composed.
(defmethod (standard-generic-function :remove-method) (method)
  (check-type method method)
  (let ((gfunc self))
    (unless (= (length (method-parameter-specializers method)) (length argument-precedence-order))
      (error "The number of specializers of method ~s does not match the 
number expected by the generic function ~s." method gfunc))
    (let* ((keys (reorder-parameter-specializers gfunc (method-parameter-specializers method)))
	   (recombine-methods t))
      (labels ((find-method-spec-objects (alist args)
		(let ((sublist (assoc (car args) alist :test #'class-eql)))
		  (and sublist (if
				 (cdr args)
				 (find-method-spec-objects  (cdr sublist) (cdr args))
				 sublist)))))
	(let((ml (find-method-spec-objects (generic-function-method-list gfunc)  keys)))
	  (unless ml 
	    (error "The method ~s has not been not found in the generic-function ~s." method gfunc))
	  (setf (cdr ml) (delete method (cdr ml) :test #'eq :key #'TICLOS:METHOD-SPEC-OBJECT-METHOD :count 1))))
      (dolist (k keys )
	 (unless (consp k)
	   (remove-method-on-specializer k  method)
	   (unless (or (typep k 'built-in-class)
		       (class-description-instance-size (class-description k)))
	     ;; When REMOVE-METHOD is called from MAKE-INSTANCES-OBSOLETE, the new 
	     ;; class description has not yet been composed, so we can't build 
	     ;; mapping tables for it yet.
	     (setq recombine-methods nil))))
       (let ((forward (getf (method-plist method) 'forwarded-from)))
	 (when (and forward
		      (eq method (method-forwarded-method forward)))
	    ;; Remove combined method that pointed to the method being removed.
	   (remove-method gfunc forward)))
      (when (and (generic-function-method-hash-table gfunc) (not (handlerp method)))
	(if recombine-methods
	    (recombine-methods gfunc keys)
	  (dolist (k keys)
	    (unless (or (consp k) (typep k 'built-in-class))
	      (uncache-effective-methods gfunc k))))))
    gfunc))

(p-defgeneric-options add-method-on-specializer (class method))
(defmethod (standard-class :add-method-on-specializer) (method)
  (let* ((gfunc (*method-generic-function method))
	 (df (getf (class-description-plist (class-description self)) 'direct-generic-functions))
	 (acons (assoc gfunc
		      df)))
    (if  acons
	 (incf (cdr acons))
	 (setf (getf (class-description-plist (class-description self)) 'direct-generic-functions)
	       (cons (cons gfunc 1) df)))))

(p-defgeneric-options remove-method-on-specializer (class method))
(defmethod (standard-class :remove-method-on-specializer) (method)
  (let* ((gfunc (*method-generic-function method))
	 (df (getf (class-description-plist (class-description self)) 'direct-generic-functions))
	 (acons (assoc gfunc
		      df)))
    (when  acons
      (if (> (cdr acons) 1)
	  (decf (cdr acons))
	  (setf (getf (class-description-plist (class-description self)) 'direct-generic-functions)
		(delete acons df :count 1 :test #'eq))))))

(p-defgeneric-options  class-direct-generic-functions (class))
;; 12/07/88 DNG - Fixed to use correct keyword and changed from class standard-class to class.
(defmethod (class :class-direct-generic-functions)()
  (mapcar #'first (getf (class-description-plist class-description) 'direct-generic-functions)))

		      
(p-defgeneric-options reorder-parameter-specializers (gfunc list))

(defmethod (standard-generic-function :reorder-parameter-specializers ) (list)
  (if (generic-function-reorder-specs self)
      (mapcar #'(lambda (item) (nth item list))
	      (generic-function-reorder-specs self))
      list))

(p-defgeneric-options order-parameter-specializers (gfunc list))

(defmethod (standard-generic-function :order-parameter-specializers ) (list)
  (if (generic-function-reorder-specs self)
      (let ((l (make-list (length list))))
	(mapc #'(lambda (it or) (setf (nth or l) it))
	      list (generic-function-reorder-specs self) )
	l)
      list))

;; 12/13/88 DNG - Fix to check for error before starting to change anything, 
;;		and to correctly show the old lambda list in the error message.
;; 12/29/88 DNG - Accept :environment argument.
(defmethod (standard-generic-function :modify-generic-function)
	   (&key ((:name nam) nil name-p) ((:lambda-list ll) nil lambda-list-p)
	    ((:argument-precedence-order apo) nil argument-precedence-order-p)
	    ((:declare decl) nil declare-p) ((:documentation doc ) nil documentation-p)
	    ((:generic-function-class gfc))
	    ((:method-class mc) nil method-class-p)
	    ((:method-combination mcomb) nil method-combination-type-p)
	    ((:environment env)))
  (declare (ignore gfc env))
  (without-interrupts
    (let* ((specializable-ll (specializable-arguments ll))
	   (rorder-specs (if (or (null apo) (= (length specializable-ll) (length apo)))
			     (mapcar #'(lambda (x) (position x specializable-ll))
				     (or apo specializable-ll))
			     (error "The length of argument-precedence-order does not match the number of required
arguments for generic-function ~S" self)))
	   (old-reorder-spec (or (generic-function-reorder-specs self)
				 (collect-body (dotimes (i (length (generic-function-argument-precedence-order self)) )
						 (collect i))))))
      (declare (list specializable-ll))
      (when lambda-list-p
	(let ((new-arg-desc (check-lambda-list-for-generic-function ll)))
	  (when (and (generic-function-method-list self)
		     (incompatible-arg-desc-p new-arg-desc (generic-function-arglist-description self)))
	    (cerror "Remove all the methods"
		    "The new lambda-list of ~s, ~s is not congruent with the old one: ~s."
		    self  ll (generic-function-lambda-list self))
	    (setf (generic-function-method-list self) nil)
	    (setf (generic-function-method-hash-table self ) nil))
	  (setf (generic-function-lambda-list self) ll)
	  (setf (generic-function-arglist-description self) new-arg-desc)))
      (when name-p (setf (generic-function-name self ) nam))
      (when argument-precedence-order-p
	(setf (generic-function-argument-precedence-order self) apo ))
      (when declare-p (setf (generic-function-declare self) decl))
      (when documentation-p (setf (generic-function-documentation self) doc ))
      (when method-class-p (setf (generic-function-method-class self) mc ))
      (when method-combination-type-p (setf (generic-function-method-combination self) mcomb))
      (if  apo
	   (setf (generic-function-reorder-specs self)
		 rorder-specs)
	   (setf (generic-function-argument-precedence-order self)
		 specializable-ll))
      (unless (equal rorder-specs old-reorder-spec )
	;;We have to rebuild the method-list
	(dolist (m (prog1 (generic-function-direct-methods self)
			  (setf (generic-function-method-list self ) nil)))
	  (add-method self m))
	(setf (generic-function-method-hash-table self ) nil)))))

;; 11/14/88 DNG - To save space in object file, modified to omit parameters with default values.
(defmethod (standard-generic-function :fasd-form) ()
  (let ((options `( :method-class (class-named ',(class-name method-class))
		   :lambda-list ',lambda-list
		   :argument-precedence-order ',argument-precedence-order 
		   :arglist-description ',arglist-description
		   )))
    (unless (null reorder-specs)
      (push-key-and-value :reorder-specs `(quote ,reorder-specs) options))
    (unless (equal method-combination '(standard))
      (push-key-and-value :method-combination `(quote ,method-combination) options))
    (unless (null declare)
      (push-key-and-value :declare `(quote ,declare) options))
    (unless (null documentation)
      (push-key-and-value :documentation documentation options))
    (unless (null name)
      (push-key-and-value :name `(quote ,name) options))
    `(sys:make-flavor-instance ',(type-of self) . ,options)))

(defun same-qualifier-p (qu1 qu2)
  (and (= (length qu1 ) (length qu2))
       (every #'eql qu1 qu2)))

;;  4/10/89 DNG - Add check for matching number of optional args.
(defun same-arglist-signature-p (m1 m2)
  (let ((as1 (method-arglist-description m1))
	(as2 (method-arglist-description m2)))
    (or (equal as1 as2)
	(and (eq (arglist-description-&rest-p as1)
		 (arglist-description-&rest-p as2))
	     (eq (arglist-description-&key-p as1)
		 (arglist-description-&key-p as2))
	     (eq (arglist-description-&allow-other-keys-p as1)
		 (arglist-description-&allow-other-keys-p as2))
	     (eql (arglist-description-n-optional as1)
		  (arglist-description-n-optional as2))
	     (eql (arglist-description-n-required as1)
		  (arglist-description-n-required as2))
	     (null (set-exclusive-or (arglist-description-keywords as1)
				     (arglist-description-keywords as2)))))))

;; Initial version of FIND-METHOD used internally to avoid bootstrapping problems.
;;  1/4/89 DNG
(unless (fboundp '*find-method) ; redefined later in file "metaclass"
  (defun *find-method (generic-function meth-qualifiers specializers &optional (errorp t))
    (let ((gfunc (if (typep generic-function 'generic-function)
		     generic-function
		   (get-generic-function-object generic-function))))
      (unless (= (length specializers) (length (generic-function-argument-precedence-order gfunc)))
	(error "The number of specializer given to FIND-METHOD does not match the number expected by the
	generic function ~s"  gfunc))
      (let* ((keys (reorder-parameter-specializers gfunc  specializers)))
	(do ((ml (find-method-spec-objects  (generic-function-method-list gfunc) keys) (cdr ml)))
	    ((null ml) (if errorp (error "No method with qualifiers ~S and specializers ~S
	found for the generic function ~S" meth-qualifiers specializers gfunc)))   
	  (when (same-qualifier-p meth-qualifiers  
				  (method-spec-object-qualifiers  (car ml)))
	    (return (method-spec-object-method (car ml))))))) ))

(comment ; old way
(unless (fboundp 'find-method) ; redefined later in file "metaclass"
  (p-defgeneric-options find-method
			((generic-function (if (typep generic-function 'generic-function)
					       generic-function
					     (get-generic-function-object generic-function)))
			 method-qualifiers specializers &optional (errorp t))
			(:documentation "Given a generic function, return the method object that agrees with the method
qualifiers and parameter specializers given."))
  )

(deff get-method 'find-method)
(compiler:make-obsolete get-method find-method)

(defmethod (standard-generic-function :find-method) ( meth-qualifiers specializers &optional (errorp t))
  (let* ((gfunc self))
    (unless (= (length specializers) (length argument-precedence-order))
      (error "the number of specializer given to find-method does not match the number expected by the
              generic function ~s"  gfunc))
    (let* ((keys (reorder-parameter-specializers gfunc  specializers)))
      (do ((ml (find-method-spec-objects  (generic-function-method-list gfunc) keys) (cdr ml)))
	  ((null ml) (if errorp (error "No method with qualifiers ~S and specializers ~S
                                        found for the generic function ~S" meth-qualifiers specializers gfunc)))   
	(when (same-qualifier-p meth-qualifiers  
				(method-spec-object-qualifiers  (car ml)))
	  (return (method-spec-object-method (car ml))))))))
) ; end comment

;;This function is internal, and should not be called by portable programs.
(p-defgeneric-options get-method-location (method))
(defmethod (standard-method :get-method-location) ()
  (or function-location
      (ferror nil "internal error: the method ~s does not have its function-location slot initialized")))


;; 12/17/88 DNG - Added a .REST. arg kludge needed by EXTEND-GENERIC-FUNCTION .
;;  4/10/89 DNG - Compare first with EQUAL to save time.
(defun incompatible-arg-desc-p (arg-desc1 arg-desc2)
  (if (equal arg-desc1 arg-desc2)
      nil
    (let* ((n-req (arglist-description-n-required arg-desc1 ))
	   (n-opt (arglist-description-n-optional arg-desc1 ))
	   (rest-p (arglist-description-&rest-p arg-desc1 ))
	   (allow-p(arglist-description-&allow-other-keys-p  arg-desc1 ))
	   (keylist (arglist-description-keywords  arg-desc1 ))
	   (keyp (arglist-description-&key-p  arg-desc1 ))
	   (n-opt2 (arglist-description-n-optional arg-desc2)))
      (not (and (= (arglist-description-n-required arg-desc2) n-req)
		(or (eql n-opt2 n-opt)
		    (and (>= n-opt2 n-opt)
			 (eq rest-p '.rest.)
			 (progn (setq rest-p (arglist-description-&rest-p arg-desc2))
				t)))
		(not (xor (or (arglist-description-&rest-p arg-desc2)
			      (arglist-description-&key-p arg-desc2))
			  (or rest-p keyp)))
		(if (arglist-description-keywords  arg-desc2)
		    (or allow-p
			(subsetp (arglist-description-keywords  arg-desc2)
				 keylist :test #'eq)
			(and rest-p (null keylist)))
		  t))))))

;; 12/19/88 DNG - Use NON-FATAL-ERROR instead of ERROR to permit proceeding.
(defun check-lambda-list-for-method (method gfunc)
  (when (incompatible-arg-desc-p
	  (method-arglist-description method)
	  (generic-function-arglist-description gfunc))
    (non-fatal-error ':ignorable-mistake
		     "~s lambda-list: ~s is not congruent with~%~s lambda-list: ~s "
	   method (method-lambda-list method)
	   gfunc (generic-function-lambda-list gfunc) )))

;;Should be a generic function.
(defun generic-function-direct-methods (generic-function)
  (let ((gfunc (if (typep generic-function   'generic-function)
		   generic-function  
		   (get-generic-function-object generic-function  ))))
    (labels ((coll (ml n)
		   (if (= n 0)
		       (collect-body
			 (dolist (m ml)
			   (let ((method (method-spec-object-method m)))
			     (cond ((typep method 'standard-accessor-method)
				    (collect method ))
				   ((combined-method-p method) nil)
				   (t 
				    (collect method))))))
		       (collect-body
			 (dolist (sublist ml)
			 (and sublist
			      (join (coll (cdr sublist) (1- n)))))))))
      (coll (generic-function-method-list gfunc) (length (generic-function-argument-precedence-order gfunc))))))

;;Should be a generic function.
;;  3/31/89 DNG - Added type check for ARGS.
;;  4/21/89 DNG - Use GET-SORTED-ALISTS-BY-VALUE instead of GET-SORTED-ALISTS to 
;;		avoid having to cons up EQL specializers.
(defun compute-applicable-methods (generic-function  args)
  "Return the list of methods that are applicable for the given generic function
and list of argument values.  The result is sorted by precedence order."
  ;;note that reorder-parameter-specializer will throw away non
  ;;required arguments.
  ;;The result is sorted by precedence order.
  (let ((gfunc (generic-function-named generic-function)))
    (check-type args list)
    (labels ((cam (alist args)
		  (if (null args)
		      (collect-body
			(dolist (m alist)
			  (let ((method (method-spec-object-method m)))
			    (cond ((typep method 'standard-accessor-method)
				   (collect method))
				  ((combined-method-p method)
				   nil)
				  (t 
				   (collect method))))))
		      (let ((sublist (get-sorted-alists-by-value (car args) alist)))
			(and sublist
			     (collect-body 
			       (dolist (ael sublist)
				 (join (cam (cdr ael) (cdr args))))))))))
      (cam (generic-function-method-list gfunc)
	   (reorder-parameter-specializers gfunc args)))))

(defun user-defined-methods-p (generic-function args)
  (let ((methods (compute-applicable-methods generic-function args)))
    (and methods
	 (null (getf (method-plist (first methods)) 'system-defined)) ; from defmethod-system
	 )))

;;  1/04/89 DNG - Disregard methods whose definition is nil.
;;  4/22/89 DNG - add sort-function option.
(defun compute-method-precedence-list (alist args sort-function)
  (if (null args)
      (multiple-values-collect-body (methods combined-methods) 
	(dolist (m alist)
	  (let ((method (method-spec-object-method m)))
	    (cond ((typep method 'standard-accessor-method)
		   (collect-into combined-methods method )
		   (collect-into methods method))
		  ((combined-method-p method)
		   (collect-into combined-methods method))
		  ((null method))
		  (t (collect-into methods method))))))
      (let ((sublist (funcall sort-function (car args) alist)))
	(and sublist
	     (multiple-values-collect-body (methods combined-methods) 
	       (dolist (al sublist)
		 (multiple-value-call #'(lambda (m &optional cm)
					  (join-into  methods m)
					  (join-into combined-methods cm))
				      (compute-method-precedence-list
					(cdr al) (cdr args) sort-function))))))))


;; 10/31/88 DNG - Test CLASS-COMPOSED-P instead of (OR (CLASS-PRECEDENCE-LIST CLASS) ...) 
(defun get-sorted-alists (arg alist)
  ;;Args are types, they can be individual types.
  (let* ((sorted-list ())
	 (class (if (individual-typep arg)
		    (let ((instance (individual-type arg)))
		      (when (obsolete-instance-p instance )
			(restructure-instance instance (class-of instance)))
		      (class-of instance))
		  arg))
	 (cpl (progn (unless (class-composed-p class)
		       (compose-class class))
		     (class-precedence-list class))))
    (dolist (ael alist sorted-list)
      (when (if (individual-typep (car ael))
		(and (individual-typep arg)
		     (eql (individual-type arg) (individual-type (car ael))))
	      (member (car ael) cpl  :test #'eq))
	(setf sorted-list
	      (sort-insert ael sorted-list
			   #'(lambda (x y) (if (individual-typep x)
					       (individual-type x)
					     (and (not (individual-typep y))
						  (first-in-cpl-p cpl x y))
					     ))))))))

;;  4/21/89 DNG - Original version; adapted from function GET-SORTED-ALISTS.
(defun get-sorted-alists-by-value (value alist)
  ;;Args are actual values.
  (let* ((sorted-list ())
	 (class (let ((instance value))
		  (when (obsolete-instance-p instance )
		    (restructure-instance instance (class-of instance)))
		  (class-of instance)))
	 (cpl (progn (unless (class-composed-p class)
		       (compose-class class))
		     (class-precedence-list class))))
    (dolist (ael alist sorted-list)
      (when (if (individual-typep (car ael))
		(eql value (individual-type (car ael)))
	      (member (car ael) cpl  :test #'eq))
	(setf sorted-list
	      (sort-insert ael sorted-list
			   #'(lambda (x y) (if (individual-typep x)
					       (individual-type x)
					     (and (not (individual-typep y))
						  (first-in-cpl-p cpl x y))
					     ))))))))

(defun first-in-cpl-p (cpl x y)
  (dolist (i cpl)
    (when (eq x i) (return t))
    (when (eq y i) (return nil))))



(defun sort-insert (ael sorted-list pred)
  (if (null sorted-list)
      (cons ael nil)
      (if (funcall pred  (car ael) (car (car sorted-list)))
	  (cons ael sorted-list)
	  (do* ((x sorted-list (cdr x))
		(y (cdr sorted-list) (cdr y)))
	       ((endp y) (setf (cdr x) (cons ael nil)))
	    (when  (funcall pred  (car ael) (car (car  y)))
	      (setf (cdr x) (cons ael y))
	      (return )))
	  sorted-list)))



(defmacro generic-function (lambda-list &rest options &environment env)
  (declare (arglist lambda-list &rest options))
  "Create an anonymous generic function.  The options are the same as for DEFGENERIC."
  ;; 11/09/88 DNG - Use new function PROCESS-GENERIC-OPTIONS.  This fixes DECLARE option.
  ;;  1/18/89 DNG - Add use of ADDITIONAL-FORMS.
  (multiple-value-bind (option-list method-list)
      (process-generic-options options)
    (let* ((gf (gensym)))
      `(let ((,gf (make-generic-function :lambda-list ',lambda-list ,@option-list)))
	 . ,(collect-body
	      (dolist (m method-list)
		(multiple-value-bind (qualifiers specializers function lambda-list ignore 
				      ignore additional-forms)
		    (parse-method m nil env)
		  (join additional-forms)
		  (collect `(add-method ,gf (make-method ',qualifiers ',specializers
							 (function ,function) ',lambda-list)))))
	      (collect gf))))))


;; This is called by the code that the compiler generates for WITH-ADDED-METHODS.
;; 12/17/88 - Original version.
;; 12/19/88 - Simplify by using function reorder-parameter-specializers.
(defun extend-generic-function (&rest options &key name lambda-list argument-precedence-order
				declare documentation generic-function-class method-class
				&allow-other-keys)
  (let ((defn (sys:fdefinition-safe name t))
	lambda-expression instance)
    (if (or (null defn)
	    (not (generic-function-p defn)))
	(progn
	  (multiple-value-setq (lambda-expression instance)
	    (apply #'create-generic-function options))
	  (unless (null defn)
	    (let ((method (make-method nil (make-list (length (generic-function-argument-precedence-order instance))
						    :initial-element t)
				     (compile nil `(lambda (&rest .rest.)
						     (declare (optimize compilation-speed))
						     (apply #',name .rest.)))
				     lambda-list)))
	      (setf (getf (method-plist method) 'next-methods-not-used) t)
	      (add-method instance method))))
      (let ((gfun (get-generic-function-object defn)))
	(multiple-value-setq (lambda-expression instance)
	  (apply #'create-generic-function
		 :declare (or declare (generic-function-declarations gfun))
		 :documentation (or documentation (generic-function-documentation gfun))
		 :generic-function-class (or generic-function-class
					     (class-of gfun))
		 :method-class (or method-class (generic-function-method-class gfun))
		 :argument-precedence-order (or argument-precedence-order
						(reorder-parameter-specializers
						  gfun (specializable-arguments lambda-list)))
		 options))
	(setf (generic-function-method-list instance)
	      (copy-tree (generic-function-method-list gfun)))
	))
    (compile nil lambda-expression)))

;;03/21/89 clm - moved here from the the "KERNEL;FLET" file.
(DEFMACRO WITH-ADDED-METHODS ((function-specifier lambda-list &rest options) &body body)
  "Syntax:  (WITH-ADDED-METHODS (function-spec lambda-list {option}*) {form}*)
Like GENERIC-LABELS except that the local function inherits methods from any 
previous definition of the function."
  ;; 12/08/88 DNG - Initial crude version - not completely correct.
  ;; 12/19/88 DNG - Fixed for (SETF ...) functions, but still not as complete as the compiler's version.
  (let ((required-args (loop for x in lambda-list
			     until (member x '(&rest &optional &key &aux))
			     unless (member x lambda-list-keywords :test #'eq)
			     collect (if (atom x) x (car x))))
	(defined (if (and (compiler:external-symbol-p function-specifier)
		       (fboundp function-specifier)
		       (member (symbol-package function-specifier)
			       '#,(mapcar #'find-package '("LISP" "TICL" "TICLOS" "SYS"))
			       :test #'eq))
		  't
		(if (symbolp function-specifier)
		    `(fboundp ',function-specifier)
		  `(fdefinedp ',function-specifier))))
	(outer-function (if (symbolp function-specifier)
			    `(quote ,function-specifier)
			  `(cdr (eval-at-lo