;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


(defun initargs-nunion (l1 l2)
  (if (or (eq l1 '&allow-other-keys)
	  (eq l2 '&allow-other-keys))
      '&allow-other-keys
      (nunion l1 l2 :test #'eq)))

(defsubst slot-location-boundp (loc)
  (/= sys:DTP-Null (sys:%P-DATA-TYPE loc)))

(p-defgeneric-options class-direct-slot-initargs (class))

(defmethod (standard-class :class-direct-slot-initargs) ()
  (collect-body
    (dolist (s (class-direct-slots self))
      (dolist (initarg (slot-initargs s))
	(collect (list initarg (slot-name s)))))))
(defmethod (built-in-class :class-direct-slot-initargs) ()
  nil)

(p-defgeneric-options class-all-slot-initargs (class))

(defmethod (standard-class :class-all-slot-initargs) ()
  (class-description-all-slot-initargs class-description))

(p-defgeneric-options compute-all-slot-initargs (class))

(defmethod (standard-class :compute-all-slot-initargs) (&aux initalist)
  (dolist (c (class-precedence-list self))
    (dolist (initcons (class-direct-slot-initargs c))
      (unless (assoc initcons initalist :test (function eq))
	(setf initalist (cons initcons initalist)))))
  (setf (class-description-all-slot-initargs class-description) initalist))

(p-defgeneric-options class-direct-initarg-defaults  (class))
(defmethod (standard-class :class-direct-initarg-defaults) ()
  (getf (class-description-plist class-description) :default-initargs))
(defmethod (built-in-class :class-direct-initarg-defaults) ()
  nil)

(p-defgeneric-options class-all-initarg-defaults  (class))
(defmethod (standard-class :class-all-initarg-defaults) ()
  (getf (class-description-plist class-description) :all-initarg-defaults))


(p-defgeneric-options default-initargs (class init-list))

(defmethod (standard-class :default-initargs) (init-list)
  (let* ((init-names
	  (collect-body
	    (do ((inits init-list (cddr inits)))
		((endp inits))
	      (collect (car inits)))))
	 (additional-defaults
	   (collect-body
	     (dolist (init-pair (class-all-initarg-defaults  self))
	       (unless (member (car init-pair) init-names :test #'eq)
		 (collect (car init-pair))
		 ;;collect the value, the lexical context has been preserved in the closure
		 (collect (funcall (second init-pair))))))))
    (if additional-defaults
	(append init-list additional-defaults)
	init-list)))

(p-defgeneric-options check-initargs (class init-list))

(defmethod (standard-class :check-initargs) (init-list)
  (let ((all-init-names (class-all-initargs self))
	(allow-other-keys-p nil)
	(allow-other-keys nil))
    (or (eq '&allow-other-keys all-init-names )
	(do ((inits init-list (cddr inits)))
	    ((endp inits))
	  (unless (member (car inits) all-init-names :test #'eq)
	    (or allow-other-keys
			 (and (eq (car inits) :allow-other-keys)
			      (progn (unless  allow-other-keys-p
				       (setf allow-other-keys (cadr inits))
				       (setf allow-other-keys-p t))
				     t))
			 (and (null allow-other-keys-p)
			      (let ((p (getf inits  :allow-other-keys  inits)))
				(when (neq p inits)
				  (setf  allow-other-keys-p t)
				  (setf allow-other-keys p))))
			 (error "invalid initarg ~s" (car inits))))))))

(p-defgeneric-options compute-all-initarg-defaults (class))

;;  3/07/89 DNG - Fixed test so that shadowing works right.
(defmethod (standard-class :compute-all-initarg-defaults) ()
  ;;We could check that all the default-initargs are valid and warn if they are not.
  (let ((all-def-initargs nil))
    (dolist (c (class-precedence-list self))
      (dolist (init-pair (class-direct-initarg-defaults c))
	(unless (assoc (car init-pair) all-def-initargs :test (function eq))
	  (setf all-def-initargs (cons init-pair all-def-initargs)))))
    (setf (getf (class-description-plist class-description) :all-initarg-defaults) all-def-initargs)))

(p-defgeneric-options compute-all-initargs (class))

(defmethod (standard-class :compute-all-initargs) (&aux initargs)
  (dolist (c (class-precedence-list self))
    (setf initargs (initargs-nunion initargs (copylist* (class-direct-initargs c)))))
  (setf (getf (class-description-plist class-description) :all-initargs) initargs))

(p-defgeneric-options class-all-initargs (class))
(defmethod (standard-class :class-all-initargs) ()
  (getf (class-description-plist class-description) :all-initargs))

(p-defgeneric-options class-direct-initargs (class))
(defmethod (standard-class :class-direct-initargs) ()
  (initargs-nunion (mapcar #'car (class-direct-slot-initargs  self))
	  (let ((initargs nil))
	    ;;This shouldn't be an exact search 
 	    (dolist (ms (compute-applicable-methods 'allocate-instance (list self)))
	      (setf initargs (initargs-nunion initargs
					      (copylist* (method-keyword-names ms)))))
	    ;; do an exact search for this one.
	    (dolist (ms (find-method-spec-objects  (generic-function-method-list
						     (get-generic-function-object 'shared-initialize))
						   (list self *t-class*)))
	      (when (not (combined-method-p (method-spec-object-method  ms)))
		(setf initargs (initargs-nunion initargs
				       (copylist* (method-keyword-names (method-spec-object-method ms)))))))
	    (dolist (ms (find-method-spec-objects  (generic-function-method-list
						     (get-generic-function-object 'initialize-instance))
						   (list self)))
	      (when (not (combined-method-p (method-spec-object-method  ms)))
		(setf initargs (initargs-nunion initargs
				       (copylist* (method-keyword-names (method-spec-object-method ms)))))))
	    #|(dolist (ms (compute-applicable-methods 'initialize-instance (list self)))
	      (setf initargs (initargs-nunion initargs
					      (copylist* (method-keyword-names ms))))) |#
	    initargs)
	  ))
(defmethod (built-in-class :class-direct-initargs) ()
  nil)

;;  5/7/89 DNG - Renamed this to not conflict with the MOP.
;;(p-defgeneric-options finalize-inheritance (class &rest options ))
(defun finalize-inheritance-internal (class &rest options)
  (lexpr-send class :finalize-inheritance options))

;; 12/23/88 DNG - Don't compute SHARED-INITIALIZE method if there are too may 
;;		  initargs to handle as &KEY arguments.
;; 12/23/88 DNG - Use new function CLASS-DESCRIPTION-FINALIZE-STATUS .
;; 12/28/88 DNG - Fix initarg test to check CLASS-ALL-SLOT-INITARGS instead of
;;                CLASS-ALL-INITARGS .
;;  4/ 6/89 GBK - Don't compute SHARED-INITIALIZE method if there are any slots
;;                with more than one :INITARG
(defmethod (standard-class :finalize-inheritance) (&rest options
						   &key slots methods initargs)
  ;; If changing slots or initargs, compute lists
  (when (or slots initargs )
    (compute-all-initarg-defaults self)
    (setf (class-description-finalize-status (class-description self))
	  (LOGIOR (class-description-finalize-status (class-description self)) 1))
    (compute-all-slot-initargs self)
    (setf (class-description-finalize-status (class-description self))
	  (logior (class-description-finalize-status (class-description self)) 2)))

  (when (or methods slots)
    (progn 
      (compute-all-initargs self)
      (setf (class-description-finalize-status (class-description self))
	    (logior (class-description-finalize-status (class-description self)) 4))))

  (unless (let ((m (*find-method #'make-instance  nil (list `(eql ,self)) nil)))
	    (and m (null (getf (method-plist m) 'computed-method))))
    (compute-make-instance-method self))

  (if (and (let ((ms (compute-applicable-methods #'shared-initialize
						 (list (class-prototype self) nil))))
	     (dolist (m ms t)
	       (when (and (null (method-qualifiers m))
			  (and (null (getf (method-plist m) 'system-defined))
			       (null (getf (method-plist m) 'computed-method))))
		 (return nil))))
	   ;; The specialized method can't be compiled if there are too many keyword
	   ;; arguments.
	   (<= (length (class-all-slot-initargs self))
	       #.(- compiler:MAX-LOCAL-SLOTS sys:LOCALS-FOR-MAPPING-TABLE-BASE))
	   (LET ((init-list (class-all-slot-initargs self)))
	     (EVERY #'(lambda (x)
			(LET (first-initarg-found
			      (name-of-x (slot-name x)))
			  (DO* ((remainder-of-init-list init-list
							(REST remainder-of-init-list))
				(init-pair (CAR remainder-of-init-list)
					   (CAR remainder-of-init-list)))
			       ;; T if no duplicates found
			       ((ENDP remainder-of-init-list) t)
			    ;; Pause when slot name found
			    (WHEN (EQ name-of-x (SECOND init-pair))
			      (LET ((this-initarg (FIRST init-pair)))
				(IF (BOUNDP first-initarg-found)
				    ;; Bail out when a different one is found
				    (UNLESS (EQ this-initarg first-initarg-found)
				      (RETURN nil))
				    ;; Save the first
				    (SETQ first-initarg-found this-initarg)))))))
		    (class-all-slots self))))
      
      (compute-shared-initialize-method self)
      (let ((old-method (*find-method #'shared-initialize nil (list self *t-class*)
				      nil)))
	(when (and old-method (getf (method-plist old-method ) 'computed-method))
	  (remove-method #'shared-initialize old-method)))))


(ticl:defmethod (built-in-class :finalize-inheritance) ignore)

;;;(defun make-instance (class &rest initargs)
;;;  (let* ((class (class-named class)))
;;;    (unless (class-precedence-list class)
;;;      (compose-class class))
;;;    (unless (class-finalized-p class)
;;;      (finalize-inheritance class :methods t :initargs T :slots t))
;;;    (setf initargs (default-initargs class initargs))
;;;    (check-initargs class initargs)
;;;    (let ((instance (apply 'allocate-instance class initargs)))
;;;      (apply 'initialize-instance instance initargs)
;;;      instance)))

;; 12/17/88 DNG - Moved definition of MAKE-INSTANCE to the end of this file so 
;;		that all of the functions it uses will be installed first, to avoid 
;;		bootstrapping problems with calling it too soon.

;; 12/07/88 moved flavor definition for init-generic-function to file "class".

;  8/09/88 DNG - Corrected second call to FINALIZE-INHERITANCE.
;  2/03/89 DNG - Compose subclass if necessary.
(defmethod (init-generic-function :after :add-method) (method)
  (LET((first-class (first (method-parameter-specializers method))))
    (when (AND (not (combined-method-p method))
	       (not (getf (method-plist method ) 'computed-method))
	       (not (individual-typep first-class))
	       (internal-class-finalized-p first-class ))
      (finalize-inheritance-internal (first (method-parameter-specializers method)) :methods t)
      (dolist (c (class-direct-subclasses (first (method-parameter-specializers method))))
	(if (or (not (internal-class-finalized-p c))
		(not (class-composed-p c)))
	    ;; We shouldn't ever get here, but just in case the data structures have 
	    ;; gotten into an inconsistent state, this should clean things up.
	    (clos:finalize-inheritance c)
	  (finalize-inheritance-internal c :methods t))))))


;;Allocate instance stuff

(defun allocate-standard-instance (class &key (area sys:default-cons-area) &allow-other-keys)
  (unless (class-composed-p class)
    (compose-class class))
  (let* ((size (class-instance-size class))
	 (i (sys:%allocate-and-initialize-instance
	      (class-description class)
	      area
	      (if (= size 1) 2 size))))
    (sys:%p-store-cdr-code i '#.(ldb sys:%%q-cdr-code
				     (dpb 1 sys:%%CLOS-INSTANCE-HEADER-FLAG 0)))
    i))

(defgeneric allocate-instance (class &key &allow-other-keys)
  (:documentation "Called by MAKE-INSTANCE to create a new, uninitialized instance of a class.")
  (:generic-function-class init-generic-function))

;; 12/30/88 DNG - Fixed to pass area correctly.
;;  2/14/89 DNG - Fix to not use &ALLOW-OTHER-KEYS so that error checking is not suppressed.
(defmethod-system allocate-instance ((class standard-class) &rest initargs)
  ;;Temporary thing, should go in the body of make-instance.
  (unless (internal-class-finalized-p class)
    (finalize-inheritance-internal class :slots t :methods t :initargs t ))
  (apply #'allocate-standard-instance class initargs))


(defgeneric initialize-instance (instance &key &allow-other-keys)
  (:generic-function-class init-generic-function)
  (declare (arglist instance &rest initargs))
  (:documentation "Called by MAKE-INSTANCE to initialize a newly created instance.
The modified instance is returned as the result.
The default method does:  (APPLY #'SHARED-INITIALIZE instance T initargs)"))

(defmethod-system initialize-instance ((object t) &rest initlist)
  (apply #'shared-initialize object t initlist))


(defgeneric shared-initialize (instance slots-for-initform &rest initargs)
  (:generic-function-class init-generic-function)
  (:documentation "Used to fill the slots of an instance using initialization arguments and
:INITFORM forms.  It is called by the default methods for INITIALIZE-INSTANCE, 
REINITIALIZE-INSTANCE, UPDATE-INSTANCE-FOR-REDEFINED-CLASS, and 
UPDATE-INSTANCE-FOR-DIFFERENT-CLASS.  The second argument is a list of the 
names of slots which are to be defaulted from their :INITFORM, or T for all 
slots."))

(defmethod-system shared-initialize ((instance standard-object) slots-for-initform &rest initlist)
  (let ((init-alist (class-all-slot-initargs (class-of instance)))
	(all-slots (class-all-slots (class-of instance)))
	(init-number 0))
    (do* ((inits initlist (cddr inits))
	  (init-name (car inits) (car inits))
	  (init-value (second inits)(second inits)))
	 ((endp  inits))
      (let ((acons (assoc init-name init-alist)))
	(when acons
	  (let ((pos (position (second acons ) all-slots :key #'slot-name :test #'eq)))
	    (when (zerop (ldb (byte 1 pos) init-number))
	      (setf (slot-value instance (second acons)) init-value)
	      (setf init-number (dpb 1 (byte 1 pos) init-number)))))))
    (if (eq t slots-for-initform)
	(without-interrupts
	  (let ((slots all-slots)
		(current (sys:follow-structure-forwarding instance))
		(i 1))
	    (dolist (s slots)
	      (if (eq :instance (slot-allocation s))
		  (progn 
		    (when (and (slot-initform s)
			       (not (slot-location-boundp (sys:%make-pointer-offset  sys:dtp-locative current  i))))
		      (sys:%p-store-contents-offset (if (slot-initclosure s)
							(funcall (slot-initclosure s ))
							(eval (car (slot-initform s ))))
						    current  i))
		    (incf i))
		  (when (slot-initform s) 
		    (let ((loc (locate-slot current (slot-name s))))
		      (unless (slot-location-boundp loc)
			(setf (contents loc)
			      (if (slot-initclosure s)
				  (funcall (slot-initclosure s))
				  (eval (car (slot-initform s))))))))))))
	(shared-initialize-default instance slots-for-initform all-slots))
    instance))

(defun shared-initialize-default (instance slots-for-initform &optional (all-slots (class-all-slots (class-of instance))))
  (let ((instance (sys:follow-structure-forwarding instance)))
    (dolist (s slots-for-initform )
      (let ((loc (locate-slot instance s)))
	(unless  (slot-location-boundp loc)
	  (let ((sl (find s (the list all-slots) :key #'slot-name :test #'eq)))
	    (unless sl
	      (error "~s s is not a slot in instance ~s" s instance))
	    (when (slot-initform sl)
	      (setf (contents loc)
		    (if (slot-initclosure sl)
			(funcall (slot-initclosure sl))
			(eval (car (slot-initform sl))))))))))))


;;Reinitialize-instance

(defgeneric reinitialize-instance (instance &rest initargs ))

(defmethod reinitialize-instance ((instance standard-object) &rest initargs)
  ;; 11/11/88 DNG - Fix generation of valid-keys list.
  ;;  5/05/89 DNG - Add :TEST #'EQ to the UNION call for speed.
  (setq initargs (default-initargs (class-of instance) initargs))
  (let* ((methods 
	   (compute-applicable-methods #'shared-initialize `(,instance nil))))
    (block check-keys
      (let ((valid-keys (mapcar #'first (class-all-slot-initargs (class-of instance)))))
	(dolist (method methods)
	  (multiple-value-bind (keys allow-other-keys-p)
	      (function-keywords method)
	    (if  allow-other-keys-p
		 (return-from check-keys t)
		 (setf valid-keys (union keys valid-keys :test #'eq)))))
	(check-keyargs initargs valid-keys)))
	(apply #'shared-initialize instance nil  initargs)
	instance))


;;Make-instance optimization, A specific individual method is added to make-instance
;;This method precomputes the check-initarg.
;;Further optimizations include cloning a class prototype that has all the constant
;;slot initform filled in, and maybe an individual method for initialize-instance.

(defun precompute-defaults (class)
  (collect-body
    (dolist (init-pair (class-all-initarg-defaults  class))
      (collect `((,(car init-pair) ,(gensym)) (funcall ',(second init-pair)))))))

;;  5/3/89 DNG - Use CLASS-ALL-INITARGS instead of CLASS-ALL-SLOT-INITARGS.  
;;		This is needed to pick up initargs defined in flavors included in hybrid 
;;		classes, but is also simpler.
;;  5/05/89 DNG - Allow for possibility of CLASS-ALL-INITARGS returning the 
;;		symbol &ALLOW-OTHER-KEYS.  Add :TEST #'EQ to the UNION call for speed.
(defun precompute-valid-keys (class)
  (let ((valid-keys (class-all-initargs class)))
    (if (eq valid-keys '&allow-other-keys)
	'(&allow-other-keys)
      (let* ((proto (class-prototype class))
	     (methods 
	       (append
		 (compute-applicable-methods #'allocate-instance `(,class))
		 (compute-applicable-methods #'initialize-instance `(,proto))
		 (compute-applicable-methods #'shared-initialize `(,proto nil)))))
	(dolist (method methods)
	  (multiple-value-bind (keys allow-other-keys-p)
	      (function-keywords method)
	    (if allow-other-keys-p
		(return-from precompute-valid-keys '(&allow-other-keys))
	      (setf valid-keys (union keys valid-keys :test #'eq)))))
	valid-keys))))


;;Before generating the code for init-code and default-code, we must make sure that
;;There is no user provided methods on shared-initialize and initialize-instance.
;;Same thing for cloning the instance.
;;  2/14/89 DNG - Fix to put :AREA on the VALID-KEYS list.
;;  2/24/89 DNG - Don't include class slots in the constant-slots list.
;;  3/22/89 GBK - Change :area to sys:area to avoid user conflict
;;  4/10/89 GBK - Fix order in default initarg list
;;  4/19/89 GBK - Remove %BLT-TYPED initialization
(defun compute-make-instance-method (class)
  (LET* ((user-alloc-code (user-defined-methods-p #'allocate-instance (list class)))
	 (keylist (precompute-defaults  class))
	 (keypars (mapcan #'(lambda (x) (list* (LIST 'QUOTE (caar x)) (cadar x) nil))
			  keylist))                         ; Looks like a property list
	 (valid-keys (precompute-valid-keys class))
	 
;;; The local variables SLOTS and CONSTANT-SLOTS were only used
;;; to initialize the constant slots using %BLT-TYPED.  For now
;;; this has been removed.  This was done because
;;; SHARED-INITIALIZE does this anyway.  Doing it in both
;;; places is not efficient.  The CLOS spec does allow the
;;; optimization of doing it only here, but for now we are
;;; following the formal specification.
	 
;;;	 (slots (class-all-slots class))
;;;	 (constant-slots (collect-body (dolist (s slots)
;;;					 (when (and (slot-initform s)
;;;						    (constantp (car (slot-initform s)))
;;;						    (eq (slot-allocation s) :instance))
;;;					   (collect s)))))
	 )
    (unless user-alloc-code
      (LET ((keylist-tail (member 'sys:area keylist :key #'caar)))
	(IF keylist-tail
	    (PROGN
	      (SETF (CADAAR keylist-tail) 'area)  ; Toss out GENSYMed symbol. We say,
						  ; area
	      (SETF (SECOND
		      (MEMBER '(QUOTE sys:area) keypars :test #'equal)) 'area)
	      (PUSH 'sys:area valid-keys))
	    (PROGN
	      (push '((sys:area area) sys:default-cons-area) keylist)
	      (push 'sys:area valid-keys)))))
    (multiple-value-bind (qualifiers specializers function lambda-list)
	(parse-method
	  `(((class (eql ,class)) &rest initargs &key ,@keylist &allow-other-keys)
	    ,@(and (not (member '&allow-other-keys valid-keys))
		   `((check-keyargs initargs ',valid-keys )))
	    ,@(mapcar #'cadar keylist)
	    (let ((instance ,(if user-alloc-code
				 (IF (NULL keypars)
				     `(apply #'allocate-instance class initargs)
				     `(IF (NULL initargs)
					  (allocate-instance class ,@keypars)
					  (apply #'allocate-instance
						 class
						 (APPEND initargs (LIST ,@keypars)))))
				 `(let* ((size ',(class-instance-size class))
					 (i (sys:%allocate-and-initialize-instance
					      (sys:eval-at-load-time (class-description
								       ',class))
					      area
					      (if (= size 1) 2 size))))
				    (sys:%p-store-cdr-code
				      i
				      '#.(ldb sys:%%q-cdr-code
					      (DPB 1
						   sys:%%CLOS-INSTANCE-HEADER-FLAG
						   0)))
				    i))))

;;; This optimization is being removed for now because it is being redone in
;;; SHARED-INITIALIZE.  For now it is easier to take it out here.  To put it back we
;;; would have to omit the constant INITFORMs from the list considered by
;;; SHARED-INITIALIZE.
	      
;;;	      ,@(and constant-slots
;;;		     `((sys:%blt-typed ',(let ((i (sys:copy-object
;;;						    (class-prototype class))))
;;;					   (dolist (s constant-slots)
;;;					     (setf (slot-value i (slot-name s))
;;;						   (eval (car (slot-initform s)))))
;;;					   i)
;;;				       instance
;;;				       ',(sys:%STRUCTURE-BOXED-SIZE
;;;					   (class-prototype class))
;;;				       1)))
	      ,(IF (NULL keypars)
		   `(apply #'initialize-instance instance initargs)
		   `(IF (NULL initargs)
			(initialize-instance instance ,@keypars)
			(APPLY #'initialize-instance
			       instance
			       (APPEND initargs (LIST ,@keypars)))))
	      instance))
	  'make-instance nil)
      (let ((method (make-method qualifiers
				 specializers
				 (compile nil function)
				 lambda-list)))
	(setf (getf (method-plist method ) 'computed-method) t)
	(add-method #'make-instance
		    method )))))

(defun compute-shared-initialize-method (class)
  (let* ((slots (class-all-slots class))
	 (init-alist (class-all-slot-initargs class))
	 (init-keylist (mapcar #'(lambda (x) `((,(car x) ,(gensym)) nil ,(gensym))) init-alist))
	 (instance-slot-names (class-description-instance-slot-names (class-description class)))
	 (init-code (mapcar #'(lambda (x y)
				 (let ((offs (slot-offset-or-locative (second y) slots instance-slot-names)))
				   `(when ,(third x)
				      (setf ,(if (numberp offs)
						 `(sys:%instance-ref instance ,(1+ offs))
						 `(contents ,offs))
					    ,(second (first x))))))
			     init-keylist init-alist))
	 (loc-sym (gensym))
	 (default-code
	   (collect-body
	     (dolist (s slots)
	       (when (slot-initform s)
		 (let ((offs (slot-offset-or-locative (slot-name s) slots instance-slot-names)))
		   (collect
		     `(let ((,loc-sym ,(if (numberp offs)
					   `(locf (sys:%instance-ref  instance ',(1+ offs)))
					   offs)))
			(unless (slot-location-boundp ,loc-sym)
			  (setf (contents ,loc-sym) 
				,(if (and (slot-initclosure s)
					  (not (constantp (car (slot-initform s)))))
				     `(funcall ',(slot-initclosure s))
				     (car (slot-initform s)))))))))))))

    
    (multiple-value-bind (qualifiers specializers function lambda-list)
	(parse-method
	  `(((instance ,class) slots-for-initform  &rest initargs )
	    (apply #'(lambda (&key ,@init-keylist &allow-other-keys)
		       ,@init-code
		       (if (eq slots-for-initform t)
			   (progn
			     ,@default-code)
			   ;;the following code could be put in a function, shared by shared-initialize and this
			   ;;specialized method.
			   (shared-initialize-default instance slots-for-initform )))

		   initargs))
		   'shared-initialize nil)
      (let ((method (make-method qualifiers specializers
				 (compile nil function)
				 lambda-list)))
	(setf (getf (method-plist method ) 'computed-method) t)
	(add-method #'shared-initialize
		    method )))))

(define-without-redefinition-query MAKE-INSTANCE
  (unless (generic-function-p 'make-instance)
    (fmakunbound 'make-instance))
  (defgeneric make-instance (class-or-name &rest initargs)
    (:documentation "Creates and returns a new instance of a class or flavor.
The first argument is a class name, class object, or flavor name.
The remaining arguments are alternating initialization keys and values."))
  )

(defmethod make-instance ((name symbol) &rest initargs)
  (declare (inline find-class))
  (apply #'make-instance (find-class name) initargs))

;; 10/17/88 DNG - removed redundant call to CLASS-NAMED.
;; 10/28/88 DNG - Use CLASS-COMPOSED-P to avoid new error in CLASS-PRECEDENCE-LIST.
(defmethod make-instance ((class standard-class) &rest initargs)
  (unless (class-composed-p class)
    (compose-class class))
  (if (not (internal-class-finalized-p class))
      (progn 
	(finalize-inheritance-internal class :methods t :initargs T :slots t)
	(apply #'make-instance class initargs))
    (progn
      (setf initargs (default-initargs class initargs))
      (check-initargs class initargs)
      (let ((instance (apply #'allocate-instance class :allow-other-keys t initargs)))
	(apply #'initialize-instance instance initargs)
	instance))))

;; Added 8/15/88 by DNG; optimized to use SEND 8/30/88.
;; 10/20/88 DNG - Changed to pass flavor structure instead of name.
(defmethod make-instance ((class flavor-class) &rest initargs)
  (declare (optimize speed (safety 0)))
  (instantiate-flavor (class-description class) (locf initargs) t))

;; Added 8/30/88 by DNG
(defmethod make-instance ((object si:vanilla-flavor) &rest initargs)
  ;; re-initialize a flavor instance
  (instantiate-flavor object (locf initargs) t))

;; Added 9/23/88 by DNG 
(defmethod make-instance ((class structure-class) &rest initargs)
  (declare (ignore initargs))
  ;; Page 1-15 of 88-002R requires an error to be signalled for this.
  (error "MAKE-INSTANCE is not permitted for a class defined by DEFSTRUCT."))

;; Added 9/27/88 by DNG to make sure the VANILLA-FLAVOR method isn't used.
(defmethod make-instance ((class built-in-class) &rest initargs)
  (declare (ignore initargs))
  (error "MAKE-INSTANCE is not valid for built-in class ~S." (class-name class)))


(defmethod allocate-instance ((class class-class) &key name 
			      direct-superclasses direct-slots direct-options documentation environment)
  (create-class name (slot-value class 'flavor)
		direct-superclasses direct-slots
		(if documentation
		    (list* :documentation documentation direct-options)
		  direct-options)
		environment))

;; 12/29/88 DNG - Added &ALLOW-OTHER-KEYS to defer initarg checking to INSTANTIATE-FLAVOR. 
(defmethod allocate-instance ((class hybrid-class-with-flavor-instances)
			      &rest initargs &key &allow-other-keys)
  (apply #'sys:make-flavor-instance (slot-value class 'flavor) initargs))


(defmethod allocate-instance :before ((class hybrid-class) &rest ignore)
  (let ((fl (slot-value class 'flavor))
	(update nil))
    (assert (typep fl 'sys:flavor))
    (when (null (sys:flavor-depends-on-all fl))
      (sys:compose-flavor-combination fl)
      (setf (class-description-class-precedence-list (class-description class))
	    nil)
      (compose-class class)
      (setq update t))
    (typecase (sys:flavor-method-hash-table fl)
      (cons (sys:make-method-hash-table fl) (setq update t))
      (null (sys:compose-method-combination fl) (setq update t)))
    (when update
      (let ((cd (class-description class)))
	(setf (class-description-method-hash-table cd)
	      (sys:flavor-method-hash-table fl))
	(setf (class-description-method-table cd)
	      (sys:flavor-method-table fl))
	(setf (class-description-flavor-which-operations cd)
	      (sys:flavor-which-operations fl))
	(unless (member (first (sys:flavor-component-mapping-table-alist fl))
			(sys:flavor-component-mapping-table-alist cd)
			:test #'eq)
	  (setf (sys:flavor-component-mapping-table-alist cd)
		(append (sys:flavor-component-mapping-table-alist cd)
			(sys:flavor-component-mapping-table-alist fl))))
	))))

;;  5/3/89 DNG - Pass locative to plist instead of the list itself.
(defmethod initialize-instance :after ((instance si:vanilla-flavor) &rest args)
  ;; Invoke any :INIT methods inherited from included flavors.
  (send instance :init (locf args)))
