;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


;;  8/29/88 DNG - Separate error message for no primary method.
;;  3/13/89 DNG - Fix to call match predicate on qualifiers rather than the method object.
(defmacro match-method-qualifiers (ml pat order requiredp desc
				   &aux (ord-ml `(if (eq ,order :most-specific-last)
						     (reverse ,ml) ,ml )))
  (declare (arglist method-list pattern order requiredp desc)
	   (values remaining-methods-list selected-methods-list))
  (declare (ignore desc))
  (if (symbolp pat)
      `(multiple-value-bind (ml1 ml2)
	   (multiple-values-collect-body (ml1 ml2)
	     (dolist (ml ,ord-ml)
	       (if (,pat (method-qualifiers ml))
		   (collect-into ml1 ml)
		   (collect-into ml2 ml))))
	 ,@(if requiredp
	       `((unless ml1
		   (method-combination-error
		     "There is no method whose qualifiers satisfies the pattern: ~s" ',pat)))
	       nil)
	 (values ml2  ml1))
      (let (code pcode)
	(if (dolist (p pat)
	      (case p
		(* (return t ))
		(t (do ((pe p (cdr pe)))
		       ((atom pe )
			(if pe
			    (progn 
			      (unless (eq '* pe)
				(error "invalid qualifier pattern ~s" p ))
			      (push p pcode))
			    (push p code)))
		     ))))
	    `(values nil
		     ,(if requiredp
			  `(or ,ord-ml (method-combination-error
					 "There is no method whose qualifiers satisfies the pattern: ~s" ',pat))
			  ord-ml))
	    `(multiple-value-bind (ml1 ml2)
		 (multiple-values-collect-body (ml1 ml2)
		   (dolist (m ,ord-ml)
		     (cond ,@(mapcar #'(lambda (p)
					 `((same-qualifier-p (method-qualifiers m) ',p)
					   (collect-into ml1  m)))
				     code)
			   ,@(mapcar
			      #'(lambda (p)
				  `(,((lambda (pe)
					(labels ((m (pe)
						    (if(consp pe)
						       `(lambda (obj)
							  (sys:match-car-cdr
							    obj
							    #'(lambda (x) (eq x ',(car pe)))
							    #',(m (cdr pe))))
						       `#'(lambda (ignore) t))))
					  `(,(m pe) (method-qualifiers m))))
				      p)
				    (collect-into ml1  m)))
			      pcode)
			   (t (collect-into ml2 m)))))
	       ,@(if requiredp
		     `((unless ml1
			 ,(if (equal pat '(nil))
			      `(method-combination-error "There is no primary method.")
			    `(method-combination-error
			       "There is no method whose qualifiers satisfies the pattern: ~s" ',pat))))
		   nil)
	       (values ml2 ml1))))))
(defun trivial-method-combination-p (combination-name)
   (get combination-name 'trivial-combination))

;; 12/29/88 DNG - Improved the error message.
;;  1/31/89 DNG - Use new function METHODS-LEFT-OVER .  Fix short form to use 
;;		as the qualifier the combination name instead of the operator.
;;  4/03/89 DNG - Moved the MACROLET defining CALL-METHOD from here to FIND-HANDLER.
(defmacro define-method-combination
	  (&whole form name &optional (lambda-list nil lambda-list-p) method-group-specifiers
	   &body body &environment environment
	   &aux method-groups (method-left (gensym)) generic-function-name arguments
	   (method-list (gensym)))
  "Define a new type of method combination.
\(DEFINE-METHOD-COMBINATION name {short-form-option}*)
    short-form-option ::= :DOCUMENTATION string | :IDENTITY-WITH-ONE-ARGUMENT boolean | :OPERATOR operator
\(DEFINE-METHOD-COMBINATION name lambda-list ({method-group-specifier}*)
	[(:ARGUMENTS . lambda-list)] [(:GENERIC-FUNCTION parameter-symbol)]
	{ declaration | doc-string }* {form}*)
    method-group-specifier ::= (variable {{qualifier-pattern}+ | predicate} {long-form-option}*)
    long-form-option ::= :DESCRIPTION format-string | :REQUIRED boolean |
		:ORDER { :MOST-SPECIFIC-LAST | :MOST-SPECIFIC-FIRST }"
  (declare (arglist name &optional lambda-list method-group-specifiers &body body))

  (if (or (null  lambda-list-p )
	  (not (listp lambda-list)))
      (multiple-value-bind ( doc   identity-with-one-argument operator )
	  (parse-short-define-method-combination-form  form)
	`(progn
	   (define-method-combination ,name (&optional (order :most-specific-first))
				      ((around (:around))
				       (primary (,name) :order order :required t))
	     ,@(if doc (list doc))
	     (let ((form (if (or ,(null identity-with-one-argument )
				 (rest primary))
			     `(,',(or operator name)
			       ,@(mapcar #'(lambda(method)
					     `(call-method ,method nil))
					 primary))
			     `(call-method ,(first primary) nil))))
	       (if around
		   `(call-method ,(first around)
				 (,@(rest around)
				  (make-method ,form)))
		   form)))
	   (defprop ,name t trivial-combination)))
      (loop 
	(case (car-safe (car body))
	  (:generic-function (setf generic-function-name (cadr  (pop body))))
	  (:arguments (setf arguments (cdr (pop body))))
	  (t (return))))
      (multiple-value-bind (body decl doc-string)
	  (parse-body body environment t)
	(let((code (mapcar
		     #'(lambda (method-group-specifier )
			 (multiple-value-bind (mv mp order req doc)
			     (parse-method-group-specifier method-group-specifier)
			   (push mv method-groups)
			   `(multiple-value-setq (,method-left ,mv)
			      (match-method-qualifiers ,method-left ,mp ,order ,req ,doc))))
		     method-group-specifiers))
	     required other-arguments rest-of-the-arglist
	     (arglist (gensym)))
	  (setf rest-of-the-arglist
		(collect-body
		(with-lambda-list arguments
				  :required #'(lambda (l) (push-end l required))
				  :optionals-before #'(lambda () (collect '&optional))
				  :optionals #'(lambda (l &aux (sym (gensym)))
						 (if (consp l )
						     (progn
						       (collect (list sym (cadr l)))
						       (push `(,(car l) ',sym)  other-arguments))
						     (progn 
						       (collect sym)
						       (push `(,l ',sym) other-arguments))))
				  :rest #'(lambda (l &aux (sym (gensym)))
					    (collect '&rest ) (collect sym)
					    (push `(,l ',sym) other-arguments))
				  :keys-before #'(lambda () (collect '&key))
				  :keys #'(lambda (l &aux (sym(gensym)))
					    (if (consp l)
						(progn
						  (collect (list* (if (consp (car l))
								     (list (caar l) sym )
								     (list (intern (car l) *keyword-package* )
									   sym))
								 (rest l)))
						  (push `(,(if (consp (car l)) (cadar l) (car l)) ',sym)
							 other-arguments))
						(progn
						  (collect (list (list (intern l *keyword-package*) sym)))
						  (push `(,l ',sym) other-arguments))))
				  :allow-other-keys-key #'(lambda () (collect '&allow-other-keys))
				  :aux-before #'(lambda () (collect '&aux))
				  :aux #'(lambda (l) (collect l)))))
				  
	  `(defun (:property ,name method-combiner)
		  (,method-list ,@lambda-list)
	     ,@(and doc-string (list doc-string))
	     ,@(and decl (list decl))
	     (declare (special generic-function))
	     (let* ((,method-left ,method-list)
		   ,@(and generic-function-name `((,generic-function-name generic-function)))
		   ,@(and arguments `((,arglist (generic-function-lambda-list generic-function))))
		   ,@(collect-body (dolist (req required) (collect `(,req (pop ,arglist)))))
		   ,@other-arguments
		   ,@method-groups)
	       ,@code
	         ;;(when ,method-left
		 ;;  (method-combination-error
		 ;;    ;; used to say "There are some methods left"
		 ;;    "In ~S method combination, the following applicable methods have~% unrecognized qualifiers: ~S"
		 ;;    ',name ,method-left)
	       (methods-left-over ,method-left ',(mapcar #'second method-group-specifiers))
	       (let ((generated-code (progn ,@body)))
		 ,(if other-arguments
		      ``(apply #'(lambda (,',@rest-of-the-arglist)
				   ,generated-code)
			       .rest.)
		    'generated-code))))))))

;;  1/31/89 DNG - Original version.
(defun methods-left-over (methods patterns)
  (dolist (method methods (values))
    (invalid-method-error method "  qualifier ~S does not match any of the valid patterns:  ~{~S~^, ~}"
			  (method-qualifiers method) patterns)))

;;  1/31/89 DNG - Original version.
;;  2/01/89 DNG - Use new function METHOD-COMBINATION-NAME .
(defun invalid-method-error (method format-string &rest format-args)
  "Called by method combination to report an error for an applicable method with invalid qualifiers."
  (declare (special generic-function))
  (let ((combination (generic-function-method-combination generic-function)))
    (catch-error-restart ((method-combination-error) "Remove ~S from ~S and continue."
			  method generic-function)
      (method-combination-error
	"~S is not valid for ~S method combination;~%~?"
	method (method-combination-name combination) format-string format-args)))
  (remove-method generic-function method)
  (values))
(defprop invalid-method-error t :error-reporter)

;;;Here is an example of arguments use.
;;;(define-method-combination progn-with-lock ()
;;;			   ((methods ()))
;;;  (:arguments a object &optional (foo t) &rest f &key x &allow-other-keys &aux (s 3))
;;;  `(unwind-protect
;;;       (progn (lock ( object-lock ,object ,foo))
;;;	      ,@(mapcar #'(lambda (method)
;;;			    `(call-method ,method ()))
;;;			methods))
;;;     (unlock (object-lock ,object))))

(eval-when (eval compile load) ; needed by DEFINE-METHOD-COMBINATION macro later in this file.
(defun parse-method-group-specifier (gs &aux (ogs gs))
  (let* ((mv (pop gs))
	(mp (if (and (neq (car gs) '*)
		     (not (listp (car gs)))
		     (not (keywordp (car gs))))
		(pop gs)
		(collect-body
		  (loop
		    (when (or (null gs)
			      (and (not (listp(car gs)))
				   (neq (car gs) '*)))
		      (return ))
		    (collect (pop gs))))))
	desc order req required-p )
    (unless mp
      (error "missing qualifier pattern in this method-group-specifier: ~s" ogs))
    (loop
      (unless (cdr gs)
	(if (car gs)
	    (error "syntax error in this method-group-specifier: ~s" ogs)
	    (return)))
      (case (car gs)
	(:description (unless (stringp (second gs))
			(error ":description option for method-group-specifier ~s must be a string" ogs))
		      (when desc
			(error ":description option specified more than once  for method-group-specifier ~s" ogs))
		      (setf desc (second gs)))
	(:order (unless (second gs)
		  (error ":order option for method-group-specifier ~s must be a non null expression" ogs))
		(when order
		  (error ":order option specified more than once  for method-group-specifier ~s" ogs))
		(setf order (second gs)))
	(:required (unless (or (eq t (second gs))
			       (eq nil (second gs)))
		     (error ":required option for method-group-specifier ~s must be NIL or T" ogs))
		   (when required-p
		     (error ":required option specified more than once  for method-group-specifier ~s" ogs))
		   (setf  required-p t
			  req (second gs)))
	
	(t (error "syntax error in this method-group-specifier: ~s" ogs)))
      (pop gs) (pop gs))
    (values mv mp (or order :most-specific-first) req desc)))
 )

(eval-when (compile load eval)
(defun parse-short-define-method-combination-form (form)
  (let ((ogs (cddr form))
	(gs (cddr form))
	 doc operator identity-with-one-argument-p  identity-with-one-argument)
    (loop
      (unless (cdr gs)
	(if (car gs)
	    (error "syntax error in this method-group-specifier: ~s" ogs)
	    (return)))
      (case (car gs)
	(:documentation (unless (stringp (second gs))
			(error ":documentation option for method-group-specifier ~s must be a string" ogs))
		      (when doc
			(error ":documentation option specified more than once  for method-group-specifier ~s" ogs))
		      (setf doc (second gs)))
	(:operator (unless (second gs)
		  (error ":operator  option for method-group-specifier ~s must be a non null symbol" ogs))
		(when operator 
		  (error ":operator  option specified more than once  for method-group-specifier ~s" ogs))
		(setf operator (second gs)))
	(:identity-with-one-argument (unless (or (eq t (second gs))
			       (eq nil (second gs)))
		     (error ":identity-with-one-argument  option for method-group-specifier ~s must be NIL or T" ogs))
		   (when identity-with-one-argument-p 
		     (error ":identity-with-one-argument  option specified more than once  for method-group-specifier ~s" ogs))
		   (setf  identity-with-one-argument-p  t
			  identity-with-one-argument  (second gs)))
	
	(t (error "syntax error in this method-group-specifier: ~s" ogs)))
      (pop gs) (pop gs))
    (values doc   identity-with-one-argument operator ))))

;;  9/20/88 DNG - Use new condition name.
(defun method-combination-error (format-string &rest args)
  (apply #'ferror 'method-combination-error format-string args))
(defprop method-combination-error t :error-reporter)

;;  2/22/89 DNG - Changed for EQL types to return the type the expression will 
;;		evaluate to, since it is no longer evaluated in DEFMETHOD.
(defun type-name (class)
  (cond ((symbolp class) class)
	((individual-typep class)
	 (let ((exp (individual-type class)))
	   (if (constantp exp)
	       (class-name (class-of (compiler:eval-for-target exp)))
	     (compiler:type-of-expression exp))))
	(t (class-name class))))

;; 12/20/88 DNG - Original.
(defun ref-method-function (method eval-now-p)
  ;; Return a form for referencing the method's code.
  (let* ((mf (*method-function method))
	 (name (function-name mf)))
    (if (and (consp name)
	     (eq (first name) 'method)
	     (second name))
	;; reference it by name.
	(if eval-now-p
	    name
	  `(function ,name))
      ;; reference the function object itself.
      (if (or eval-now-p (and (atom mf) (not (symbolp mf))))
	  mf
	`(quote ,mf)))))

(def clos:make-method) ; not really a function, handled by the code below.
(defprop clos:make-method (&quote form) arglist)
(setf (documentation 'clos:make-method 'function)
      "A local macro that can be used within CALL-METHOD to create a method.
The argument is a form that will be used as the body of the method function.")

;;Version of the code using sys:eval-at-load-time
(defun call-next-method-code (generic-function)
  ;; 11/30/88 DNG - Add use of MAKE-UNSPECIAL.
  ;; 12/20/88 DNG - Use new function ref-method-function in order to not 
  ;;		reference a method by an inappropriate name, such as using an 
  ;;		(:INTERNAL ...) FEF instead of its lexical closure.
  ;;  3/13/89 DNG - Add recognition of CLOS:MAKE-METHOD.
  (let* ((arg-desc (generic-function-arglist-description generic-function))
	 (rest-p (or (plusp (arglist-description-n-optional arg-desc))
		     (arglist-description-&rest-p arg-desc)
		     (arglist-description-&key-p arg-desc)))
	 (specializable-arguments (specializable-arguments (generic-function-lambda-list  generic-function)))
	 (args (append specializable-arguments (and rest-p '(.rest.))))
	 (arglist (append specializable-arguments (and rest-p '(&rest .rest.)))))
    
    `(if (member (car-safe method) '(make-method clos:make-method))
	 (second method)
	 `(sys:%apply-method
	    ,(ref-method-function method nil)
	   ,',(if rest-p `(list* ,.args) `(list ,.args))
	   ,,(if (= (length specializable-arguments ) 1)
		``(sys:%clos-map-ref ,',(car specializable-arguments )
				       ',(type-name (car (method-parameter-specializers method))))
		  ``(sys:%make-clos-map-list
		     ,@(mapcan #'(lambda ( x y) (list* x (list 'quote (type-name y)) nil))
			       ',specializable-arguments
			     (method-parameter-specializers method))))
	   ,(and next-method-list (not (slot-accessor-p method)) (not (getf (method-plist  method) 'next-methods-not-used ))
		 `(sys:%make-clos-continuation-list
		    ,@(collect-body
			(do ((ml  next-method-list (cdr ml))
			     (prevm method (car ml))
			     (prev-types nil))
			    ((null ml))
			  (let ((m (car ml)))
			    (if (member (car-safe m) '(make-method clos:make-method))
				(progn
				  (collect
				    (compile
				      nil
				      `(lambda ( ,',@arglist)
					 (declare (specializers ,@(or prev-types
								      (method-parameter-specializers prevm)))
						  ,',@(cdar (make-unspecial arglist)))
					 (macrolet ((call-method (method next-method-list)
					  			,(call-next-method-code ',generic-function)))
					    ,(second m)))))
				  (collect (mapcan #'(lambda ( x z )
						       (list* x (and z (type-name z)) (and z (type-name z)) nil))
						   ',specializable-arguments
						   (setf prev-types
							 (or prev-types (method-parameter-specializers prevm))))))
				(progn
				  (collect (ref-method-function m t))
				  (collect (mapcan #'(lambda ( x y z) (list* x (type-name y) (and z (type-name z)) nil))
						   ',specializable-arguments
						   (method-parameter-specializers m)
						   (or prev-types (method-parameter-specializers prevm))))
				  (setf prev-types nil))))))
			  
			 ))))))

(defun %clos-continuation-list (methods-and-maps)
  (let ((list (make-list (length methods-and-maps))))
    (do ((result-tail list (cddr result-tail))
	 (specs-tail methods-and-maps (cddr specs-tail)))
	((null specs-tail))
      (if (atom (first specs-tail))
	  (setf (first result-tail) (first specs-tail))
	  (sys:%p-store-data-type-and-pointer
	    (locf (first result-tail))
	    sys:dtp-external-value-cell-pointer
	    (fdefinition-location (first specs-tail))))
      (if (= (length (second specs-tail)) 1)
	  (let ((number (apply #'clos-var-pointer (first (second specs-tail)))))
	    (unless (null number)
	      (sys:%p-store-data-type-and-pointer
		(locf (second result-tail))
		sys:dtp-self-ref-pointer
		number)))
	  (setf (second result-tail)
		(apply #'%clos-map-list (second specs-tail)))))
    list))


(define-method-combination and :identity-with-one-argument  t)

(define-method-combination progn :identity-with-one-argument  t)

(define-method-combination or :identity-with-one-argument  t)

(define-method-combination + :identity-with-one-argument  t)

(define-method-combination max :identity-with-one-argument  t)

(define-method-combination min :identity-with-one-argument  t)

(define-method-combination list)

(define-method-combination nconc)

;;  9/07/88 DNG - Avoid unnecessary use of MULTIPLE-VALUE-PROG1 to not 
;;		prevent optimization in MAKE-COMBINED-METHOD .
(define-method-combination standard ()
			   ((around (:around))
			    (before (:before))
			    (primary () :required t)
			    (after (:after)))
  (flet ((call-methods (methods)
	     (mapcar #'(lambda (method) `(call-method ,method nil))
		     methods)))
    (let ((form (if (or before after)
		    `(multiple-value-prog1
		       (progn ,@(call-methods before)
			      (call-method ,(first primary)
					   ,(rest primary)))
		       ,@(call-methods (reverse after)))
		  `(call-method ,(first primary) ,(rest primary)))))
      (if around
	  `(call-method ,(first around)
			(,@(rest around) (make-method ,form)))
	form))))
(defprop standard t trivial-combination)

(comment ; removed 4/3/89
  (deff (:PROPERTY STANDARD-METHOD-COMBINATION METHOD-COMBINER)	; this may be obsolete? -- DNG 9/7/88
	#'(:PROPERTY STANDARD METHOD-COMBINER))
  (defprop standard-method-combination t trivial-combination)
  )

(def call-method) ; a local macro defined by the MACROLET in FIND-HANDLER below.
(defprop call-method (method next-method-list) arglist)
(setf (documentation 'call-method 'function)
      "A local macro that can be used within DEFINE-METHOD-COMBINATION to call a method.
The first argument is the method object to be invoked, and the second argument
is a list of method objects for use by CALL-NEXT-METHOD.")

;;;We should be smarter about the derivation. for standard method combination and 
;;;simple method combinations, the list of method and the combination should be enough.
;;;If we do that, we don't have to execute the method combination, we just have to
;;;compute the method precedence list.
;;;--------------------------------------------------------------------------------
;;;With the new congruence rule, the derivation must include a trace of the lambda list
;;;for the methods used. Keylist is used for this.
;;;
;;  4/03/89 DNG - Modified to use compute-effective-method.
(defun find-handler (gen-function arg-types &optional (sort-function #'get-sorted-alists))
  ;;Arg-types is supposed to be ordered as arguments-precedence-order
  ;;Returns one to four values:
  ;;First value is the handler, if found;
  ;;Second value is the code of the combined method (if the handler is a combined method)
  ;;Third value is the combined method derivation.
  ;;Fourth value is the applicable methods, so make-combined-method can compute its arglist.
  (declare (values handler code derivation applicable-methods))
  (let* ((generic-function gen-function)
	 (ml (generic-function-method-list  generic-function))
	 (combination (generic-function-method-combination generic-function))
	 (derivation))
    (declare (special generic-function derivation))
    (multiple-value-bind (methods combined-methods)
	(compute-method-precedence-list ml arg-types sort-function)
      (if methods
	  (multiple-value-bind (ignore keylist)
	      (compute-combined-lambda-list gen-function methods)
	    (when (trivial-method-combination-p (car combination))
	      (setf derivation (list  combination
				      (mapcar #'(lambda (m) (function-name (*method-function m))) methods)
				      keylist))
	      (let* ((handler (have-combined-method combined-methods derivation)))
		(when handler (return-from find-handler handler nil derivation))))
	    (let ((code `(macrolet ((call-method (method next-method-list)
						 ,(call-next-method-code gen-function)))
			   ,(compute-effective-method gen-function combination methods))))
	      (if (trivial-method-combination-p (car combination))
		  (values nil code derivation methods)
		(values  (have-combined-method combined-methods derivation)
			 code
			 (list combination (copy-tree code) keylist)
			 methods))))))))


(unless (fboundp 'compute-effective-method) ; redefined later in "meta-object" file.
  (defun compute-effective-method (generic-function combination methods)
    (declare (special generic-function))
    (apply (or (get (car combination) 'method-combiner)
	       (error "Undefined method combination: ~S" (car combination)))
	   methods (rest combination))
    ))

(defun compute-initial-points (method-list n)
  (let ((res (make-list n)))
    (dolist (m method-list)
      (let ((specializers (method-parameter-specializers m)))
	(dotimes (i n)
	  (let ((c (nth i specializers )))
	    (unless (individual-typep c)
	      (pushnew c (nth i res)))))))
    ;;remove common superclasses from initial set.
    (dotimes  (i n)
      (let ((l (nth i res)))
	(setf (nth i res)
	      (if (= (length l) 1)
		  l
		  (collect-body
		    (dolist (c l)
		      (unless (dolist (x l t)
				(unless (subclassp x c)
				  (return nil)))
			(collect c)))))))
    res)))

(defflavor no-applicable-method-handler ()(method))
(defmethod (no-applicable-method-handler :print-self) (stream &rest ignore)
	   (format stream "#<No-applicable-method-handler ~o>" (sys:%pointer self)))

;; 11/23/88 DNG - Fix to inherit from METHOD instead of nothing.
;; 12/05/88 DNG - Use HYBRID-CLASS.
(defclass-primitive no-applicable-method-handler (method) ()
  (:metaclass hybrid-class-with-flavor-instances))

(defmethod (no-applicable-method-handler   :get-hash-entries-from-handler) (ignore ignore)
  (values (locf function) (locf (generic-function-discriminator-code generic-function))))

(comment ; superseded by PREPARE-EFFECTIVE-CODE below  4/21/89
;;  8/09/88 DNG - Fixed LAMBDA to use ,ARGLIST instead of (,ARGLIST).
;; 11/21/88 DNG - Add use of MAKE-UNSPECIAL.
(defun compute-effective-code (gen-function &rest arg-types)
  ;;arg-type are of the form (eql argument). this allows to
  ;;pick methods on individuals. They are in the same order they were passed to the generic function.
  ;;At this point, it shouldn't have to make new combined methods.
  ;;Combined method are done when the methods are defined or the generic function called the
  ;;first time.
  (let* ((rargs (reorder-parameter-specializers gen-function arg-types))
	(mloc (apply #'find-right-method (generic-function-method-hash-table gen-function)
		     (mapcar #'individual-type  rargs))))
    (if mloc
	nil ;(contents mloc)
	(multiple-value-bind (handler code ignore methods)
	    (find-handler gen-function  rargs)
	  (if (null handler)
	      (if (null code)
		  (let* ((arglist (compute-combined-lambda-list gen-function nil))
			 (fn (compile nil `(lambda ,arglist
					     ,@(make-unspecial arglist)
					     (apply #'no-applicable-method
						    (contents (COMPILER:%LOCAL-SLOT
								SYS:LOCAL-FOR-FIRST-MAPPING-TABLE))
						    ,@(if (member '&rest arglist)
							  (remove '&rest arglist)
							  (append arglist '(nil)))))))
			 (handler (sys:make-flavor-instance
				    'no-applicable-method-handler :function fn
				    :generic-function gen-function
				    :parameter-specializers
				    (mapcar #'(lambda(x) (class-of (individual-type  x))) rargs))))
		    (cache-handler-for-arguments  gen-function handler)
		    nil)
		  (PROGN 
		    (recombine-methods gen-function (mapcar #'(lambda (x) (class-of (individual-type x))) rargs)
				       (compute-initial-points methods (length rargs)))
		    (APPLY #'compute-effective-code gen-function  arg-types)
		    nil))
	      (progn 
		(when (generic-function-method-hash-table gen-function)
		  (cache-handler-for-arguments  gen-function
						handler
						(mapcar #'individual-type  rargs)))
		nil #| (*method-function handler) |# ))))))
)
;;  4/21/89 DNG - Original.  Adapted from COMPUTE-EFFECTIVE-CODE, differing by 
;;		taking arg values instead of types and by expecting the args to already be 
;;		in precedence order.
(defun prepare-effective-code (gen-function &rest arg-values)
  ;;arg-values are the actual argument values, reordered in precedence order.
  ;;At this point, it shouldn't have to make new combined methods.
  ;;Combined method are done when the methods are defined or the generic function called the
  ;;first time.
  (let ((mloc (apply #'find-right-method (generic-function-method-hash-table gen-function)
		     arg-values)))
    (when (null mloc)
      (multiple-value-bind (handler code ignore methods)
	  (find-handler gen-function arg-values #'get-sorted-alists-by-value)
	(if (null handler)
	    (if (null code)
		(let* ((arglist (compute-combined-lambda-list gen-function nil))
		       (fn (compile nil `(lambda ,arglist
					   ,@(make-unspecial arglist)
					   (apply #'no-applicable-method
						  (contents (COMPILER:%LOCAL-SLOT
							      SYS:LOCAL-FOR-FIRST-MAPPING-TABLE))
						  ,@(if (member '&rest arglist)
							(remove '&rest arglist)
						      (append arglist '(nil)))))))
		       (handler (sys:make-flavor-instance
				  'no-applicable-method-handler :function fn
				  :generic-function gen-function
				  :parameter-specializers
				  (mapcar #'class-of arg-values))))
		  (cache-handler-for-arguments  gen-function handler)
		  nil)
	      (PROGN 
		(recombine-methods gen-function (mapcar #'class-of arg-values)
				   (compute-initial-points methods (length arg-values)))
		(APPLY #'prepare-effective-code gen-function  arg-values)
		nil))
	  (progn 
	    (when (generic-function-method-hash-table gen-function)
	      (cache-handler-for-arguments  gen-function handler arg-values))
	    nil)))))
  (values))

(defun make-unspecial (arglist)
  ;; 11/21/88 DNG - Original.
  (let ((vars '()))
    (dolist (x arglist)
      (let ((name (if (consp x) (car x) x)))
	(when (compiler:specialp name)
	  (push name vars))))
    (and vars
	 `((declare (unspecial . ,vars))))))

(defun have-combined-method (combined-methods derivation)
  (find derivation combined-methods :key #'method-derivation :test #'equal))

(SYSTEM:ADD-ITEMS-TO-SUPPORT-VECTOR '(:allow-other-keys) #o34)

(proclaim '(inline check-keyargs ))
(DEFUN check-keyargs (ARGS KEYKEYS)
  (declare (arglist plist-of-actual-keys-and-values list-of-valid-keys))
  ;; Note: passing NIL as the third argument to %STORE-KEY-WORD-ARGS changes 
  ;; the meaning of the second argument.
  (SYSTEM:%STORE-KEY-WORD-ARGS keykeys args nil))

;;;  (DO ((ARGS-LEFT ARGS (CDDR ARGS-LEFT)))
;;;      ((NULL ARGS-LEFT))
;;;    (LET ((KEYWORD (CAR ARGS-LEFT)))
;;;      (COND (ALLOW-OTHER-KEYS (RETURN))
;;;	    ((POSITION KEYWORD (THE LIST KEYKEYS) :TEST #'EQ)
;;;	       )
;;;	    ((SETQ ALLOW-OTHER-KEYS
;;;		   (GETF ARGS-LEFT :ALLOW-OTHER-KEYS)))	;treat as a disembodied plist
;;;	    (t (fERROR 'SYS:UNDEFINED-KEYWORD-ARGUMENT
;;;			 "Keyword arg keyword ~S unrecognized."
;;;			 KEYWORD ))))))

;; 11/30/88 DNG - Add FORWARDED-FROM property.
;; 12/07/88 DNG - Pass the generic function as an additional argument to 
;;		METHOD-NAMED; this saves it the time of having to look it up itself and 
;;		also ensures looking at the right thing if it doesn't have a proper name.
;;  1/04/88 DNG - Use SYS:MAKE-FLAVOR-INSTANCE instead of MAKE-INSTANCE in 
;;		order to be able to build combined methods for MAKE-INSTANCE.
(defun make-combined-method (gen-function arg-types code derivation methods)
  (multiple-value-bind (lambda-list keylist keyp)
      (compute-combined-lambda-list gen-function methods)
    (let ((exp (sys:macroexpand-all code )) 
	  method m)
      (if (and (eq (car exp)
		   'sys:%apply-method)
	       (eq (car-safe (cadr exp))
		   'function)
	       (eq (nth-safe 4 exp )
		   nil)
	       (or (null keyp) (not (listp keylist)))
	       (setf m (if (consp (cadr exp))
			   (method-named (cadr (cadr exp)) t gen-function)
			   (sys:get-debug-info-field (sys:get-debug-info-struct (cadr exp)) :method)))
	       ;;Can't dispatch to interpreted functions.
	       (typep (*method-function m) 'compiled-function)
	       (let ((arg-desc (TICLOS:METHOD-ARGLIST-DESCRIPTION m)))
		 (= 0 (ARGLIST-DESCRIPTION-N-OPTIONAL arg-desc))))
	  (progn (setf method (sys:make-flavor-instance 'null-combined-method
					     :function nil
					     :forwarded-method m
					     :parameter-specializers arg-types
					     :qualifiers '(:combined)
					     :lambda-list lambda-list
					     :derivation derivation))
		  (setf (getf (method-plist m) 'forwarded-from) method)
		  method)
	  
	  (let* ((function
		   (compile nil 
                    `(named-lambda ((method ,(generic-function-name gen-function) :combined ,arg-types )
				    (:DESCRIPTIVE-ARGLIST ,(generic-function-lambda-list gen-function)))
				   ( ,@lambda-list)
				   (declare (unspecial ,@(generic-function-argument-precedence-order gen-function)))
				   ,@(and keyp (listp keylist)
					  `((check-keyargs (nthcdr ,(arglist-description-n-optional
								      (generic-function-arglist-description 
									gen-function ))
								   .rest. )
							   ',keylist)))
				   ,code)))
		 (sys:%inhibit-read-only t)
		 (ds (sys:get-debug-info-struct function t)))
	    (setf method (sys:make-flavor-instance 'combined-method
					:parameter-specializers arg-types
					:qualifiers '(:combined)
					:lambda-list lambda-list
					:function function
					:derivation derivation))
	    (setf (sys:get-debug-info-field ds :method) method))
	  )
      (add-method gen-function  method)
      (when (generic-function-method-hash-table gen-function)
	(put-method-in-hash-table gen-function  method)
	;; remember that this handler is cached in the hash-table
	;; so we can remove it when the class gets redefined.
	;;	(dolist (c arg-types)
	;;	  (handler-cached (if (individual-typep c) (class-of (individual-type c)) c) method)))
	    )
      method)))

(p-defgeneric-options  method-keyword-names (method))
(defmethod (standard-method :method-keyword-names) ()
  (let ((argdesc (method-arglist-description self )))
    (if (ARGLIST-DESCRIPTION-&ALLOW-OTHER-KEYS-P argdesc )
	'&allow-other-keys
	(arglist-description-keywords argdesc))))

(def function-keywords
(p-defgeneric-options  function-keywords (method)
		       (declare (values keyword-list allow-other-keys-p)))
(defmethod (standard-method :function-keywords) ()
  (let ((argdesc (method-arglist-description self )))
    (if (ARGLIST-DESCRIPTION-&ALLOW-OTHER-KEYS-P argdesc )
	(values nil t)
	(values (arglist-description-keywords argdesc)))))
)

(defun compute-combined-lambda-list (gen-function methods)
  (declare (values arglist keylist))
  (let* ((argdesc (generic-function-arglist-description gen-function ))
	 (keyp (arglist-description-&key-p argdesc))
	 (keylist (if (arglist-description-&allow-other-keys-p argdesc)
		      '&allow-other-keys
		      (arglist-description-keywords argdesc)))
	 (rest-p (or (plusp (arglist-description-n-optional argdesc))
		     (arglist-description-&rest-p argdesc)
		     (arglist-description-&key-p argdesc))))
    (unless (eq keylist '&allow-other-keys)  
      (dolist (m methods)
	(unless keyp
	  (setf keyp (arglist-description-&key-p (method-arglist-description m))))
	(if (eq (method-keyword-names m) '&allow-other-keys)
	    (return (setf keylist '&allow-other-keys))
	    (setf keylist (union keylist (method-keyword-names m) :test #'eq)))))
    (values
      (nconc (firstn (arglist-description-n-required argdesc)
		     (generic-function-lambda-list gen-function))
	     (and rest-p '(&rest .rest.)))
      (if (consp keylist )
	  (sort keylist #'string-lessp)
	  keylist)
      keyp)))

;;  4/21/89 DNG - Use BACKGROUND-CONS-AREA.
(defun combine-method  (gen-function arg-types)
  ;;Arg-types is supposed to be ordered as arguments-precedence-order
  (let ((default-cons-area sys:background-cons-area))
    (multiple-value-bind (handler code derivation methods)
	(find-handler gen-function  arg-types)
      (when (and handler (generic-function-method-hash-table gen-function))
	(put-method-in-hash-table  gen-function  handler arg-types))
      (or handler
	  (if methods
	      (make-combined-method gen-function
				    (order-parameter-specializers gen-function arg-types)
				    code derivation methods)
	    (let ((m (*find-method gen-function '(:combined) (REORDER-PARAMETER-SPECIALIZERS gen-function arg-types) nil)))
	      (when m (remove-method gen-function m))
	      (when (generic-function-method-hash-table gen-function)
		(remove-method-in-hash-table gen-function arg-types))
	      nil))))))

;;  8/12/88 DNG - Use CATCH-ERROR-RESTART instead of IGNORE-ERRORS.
;;  9/20/88 DNG - Use IGNORE-METHOD-COMBINATION-ERROR instead of CATCH-ERROR-RESTART.
(defun combine-methods (generic-function)
  ;;Take the method-alist, reorder it by increasing specialization,
  ;;then combine each node of the tree.
  (let* ((n-args (length
		  (generic-function-argument-precedence-order generic-function)))
	 (points (make-list n-args)))
    (declare (special *t-class*))
    (labels ((ncm (rargs tree points seen)
		(if (null points)
		    (ignore-method-combination-error
		      (combine-method generic-function (reverse rargs)))
		    (dolist (c (car points))
		      (unless (member c (car seen) :test #'class-eql)
			(push c  (car seen))
			(let ((subtree (check-tree c tree points)))
			  (and subtree
			       (ncm (cons c rargs) subtree (cdr points) (list nil))))
			(unless (or (individual-typep c) (eql *t-class* c) (= (length tree) 1))
			  (with-stack-list* (new-points (class-direct-subclasses c) (cdr points))
			    (ncm rargs tree new-points seen))))
		      )))
	     (cm1(tree points)
		 (if (null points)
		     nil 
		     (dolist (ael tree)
		       (pushnew (car ael) (car points) :test #'class-eql)
		       (cm1 (cdr ael)  (cdr points))))))
      (let* ((ml (generic-function-method-list  generic-function))
	     (tree ml ))
	(when tree
	  (cm1 tree  points)
	  (ncm nil tree points(list nil)))
	))))

(defun check-tree (arg tree points)
  (let ((res nil))
    (and 
      (let ((f nil))
	  (dolist (Ci tree f)
	    (when (and (subclassp arg (car Ci))
		       (prog1 t
			      (setf f t)
			      (when (cdr points)
				(dolist (acons (cdr ci) t)
				  (let ((ac (assoc (car acons) res :test #'class-eql)))
				    (if ac (setf (cdr ac ) (append (cdr acons) (cdr ac)))
					(push (cons (car acons) (cdr acons))  res))))))
		       (not (dolist (Zi tree nil)
			      (when (or (eq (car zi) arg)
					(and (subclassp arg (car zi))
					     (not (subclassp (car Ci ) (car Zi)))))
				(return t)))))
	      (return nil))))
      (or res t))))




(defun superclassp (ael1 ael2)
  (subclassp ael2 ael1))



;;  9/20/88 DNG - Add use of IGNORE-METHOD-COMBINATION-ERROR .
(defun recombine-methods (generic-function arg-classes &optional initial-points)
  ;;Starting at arg-classes, refigures out the handler for subclasses
  ;;Combine-method is the expensive call.  However, combining method for
  ;;only cached classes will tend to make combined methods too specific.
  ;;Arg-classes is supposed to be ordered by argument-precedence-order.
  ;;Take the method-alist, 
  ;;then combine each node of the tree.
  (let* ((ht (generic-function-method-hash-table generic-function))
	 (points (mapcar #'list arg-classes))
;;	 (n-args (length arg-classes))
	 )
    (declare (special *t-class*))
    (labels ((cm1 (tree points args-c)
		  (if (null points)
		      nil 
		      (dolist (ael tree)
			(when (subclassp (car ael) (car args-c))
			  (pushnew (car ael) (car points) :test #'class-eql))
			(cm1 (cdr ael)  (cdr points) (cdr args-c)))))
	     (ncm (rargs tree points seen)
		(if (null points)
		    (ignore-method-combination-error
		      (combine-method generic-function (reverse rargs)))
		    (dolist (c (car points))
		      (unless (member c (car seen) :test #'class-eql)
			(push c  (car seen))
			(let ((subtree (check-tree c tree points)))
			  (and subtree
			       (ncm (cons c rargs) subtree (cdr points) (list nil))))
			(unless (or (individual-typep c) (eql *t-class* c) (= (length tree) 1))
			  (with-stack-list* (new-points (class-direct-subclasses c) (cdr points))
			    (ncm rargs tree new-points seen))))
		      )))
	     
	     )
      ;;First combine the method for arg-classes, recombine-method is called
      ;;when a new method is added.
      ;;This will prevent a more specific set of arguments that would inherit this
      ;;combined method to make a combined method of their own, just because they were
      ;;in the hashtable
      (unless initial-points
	(ignore-method-combination-error (combine-method generic-function arg-classes)))
      (let* ((ml (generic-function-method-list  generic-function))
	     (tree ml ))
	(when tree
	  (unless initial-points (cm1 tree  (mapcar #'nreverse points) arg-classes))
	  (ncm nil tree (or initial-points points) (list nil))
	  (when (and ht (null initial-points))
	    (fix-cached-methods generic-function  arg-classes))
	  )))))

(defun fix-cached-methods (gfunc arg-classes)
  (map-all-hash-entries
    gfunc
    #'(lambda (key mloc mtl ht rargs)
	(declare (ignore mtl ))
	;;remove obsolete class-descriptions
	(if (and (arrayp key) (obsolete-class-description-p key))
	    (remhash key ht)
	    (when (locativep mloc)
	      (let* ((classes (let ((l nil))
				(dolist (x (if key (cons key rargs) rargs) l)
				  (push (if (consp x) x
					    (class-description-class-object x))
					l)))))
		(when				;(every #'subclassp classes arg-classes)
		  (do ((cl classes (cdr cl))
		       (argc arg-classes (cdr argc)))
		      ((null cl) t)
		    (unless (subclassp (car cl) (car argc))
		      (return nil)))
		  (combine-method gfunc (nconc classes
					       (nthcdr (length classes) arg-classes))))))))))


(defun subclassp (x y)
  ;;  8/12/88 DNG - Added  :test #'eq  for efficiency.
  ;; 10/17/88 DNG - Changed to use :SUBCLASSP message instead of calling 
  ;;		CLASS-PRECEDENCE-LIST so that we don't have to cons up the list for 
  ;;		flavors.
  ;;  4/21/89 DNG - Speed up handling of EQL specializers and fix bug on (EQL NIL).
  (declare (optimize speed (safety 0))) ; because we spend a lot of time here.
  (cond ((individual-typep x)
	 (if (individual-typep y)
	     (eql (individual-type x) (individual-type y))
	   (subclassp (class-of (individual-type x)) y )))
	((individual-typep y)
	 (and (equal y '(eql nil))
	      (subclassp x (class-of nil)))) ; the only class that contains just one element
	(t (or (eql x y)
	       (send x :subclassp y)))))













