;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;Individual hash tables
(defconstant %%array-leader-clos-individual-hash-table (byte 1 19))
(defconstant %%array-leader-clos-rehash-needed-flag (byte 1 20))

(defvar *individual-hash-tables* nil)


(defun individual-hash-table-rehash (hash-table grow)
  (sys:hash-table-rehash hash-table grow (* 2 (sys:hash-table-modulus hash-table)))
  (without-interrupts
    (let ((ht (sys:follow-structure hash-table)))
      (sys:%p-dpb-offset 1 %%array-leader-clos-individual-hash-table ht -1)
      (put-default-value-in-hash-table ht )
      (setf (sys:hash-table-rehash-function ht) #'individual-hash-table-rehash )
      (sys:%p-dpb-offset 0 %%array-leader-clos-rehash-needed-flag  ht -1)
      ht)))

(defun make-individual-dispatch-hash-table (&key (size nil size-p) default-value)
  (without-interrupts
    (let ((ht  (sys:make-hash-array  :number-of-values 2 :funcallable-p t :test #'eq
				     :rehash-function #'individual-hash-table-rehash
				     :size (or size 16)
				     :instance default-value)))
      (sys:%p-dpb-offset 1 %%array-leader-clos-individual-hash-table ht -1)
      (put-default-value-in-hash-table ht default-value)
      (setf (sys:hash-table-rehash-function ht) #'individual-hash-table-rehash )
      (push ht *individual-hash-tables* )
      ht)))

(proclaim '(inline individual-hash-table-p
		   individual-dispatch-hash-table-default-value
		   individual-dispatch-hash-table-ht ))

(defun individual-hash-table-p (hash-table &optional  (follow-forwarding t))
  (and (arrayp hash-table)
       (array-has-leader-p hash-table)
       (ldb-test %%array-leader-clos-individual-hash-table   
		(sys:%p-pointer
		  (sys:%make-pointer-offset sys:dtp-locative
					    (if follow-forwarding
						(sys:follow-structure hash-table)
						hash-table)
					    -1)))))

(defun individual-dispatch-hash-table-default-value (ht)
  (sys:hash-table-instance ht))

(defsetf individual-dispatch-hash-table-default-value put-default-value-in-hash-table )

(defun individual-dispatch-hash-table-ht (ht)
  ht)

(defun mark-individual-table ()
  (dolist (ht *individual-hash-tables* )
    (sys:%p-dpb-offset 1 %%array-leader-clos-rehash-needed-flag  ht -1)))

(push '(mark-individual-table ) sys:gc-after-flip-list)

(defun put-default-value-in-hash-table (hash-table &optional value-list)
  (with-lock ((sys:hash-table-lock hash-table) :whostate "Hash Table Lock")
    ;; Set all of hash table to NIL with cdr-next.
    (without-interrupts
      (setf hash-table (sys:follow-structure hash-table))
      (if value-list
	  (setf (sys:hash-table-instance hash-table ) value-list )
	  (setf value-list (sys:hash-table-instance hash-table )))
      ;;Do a do loop in order to preserve the existing entries.
      (let ((hlen (array-total-size hash-table)))
	(do ((p (locf (aref hash-table 0)) (sys:%make-pointer-offset sys:dtp-locative p 3)))
	    ((> (sys:%pointer-difference p hash-table) hlen))
	  (when (= (sys:%p-data-type p) sys:dtp-null)
	    (sys:%p-store-contents (sys:%make-pointer-offset sys:dtp-locative p 1) (car value-list))
	    (sys:%p-store-contents (sys:%make-pointer-offset sys:dtp-locative p 2) (cadr value-list))
	    )))
      value-list
      )))


(p-defgeneric-options class-hash-key (class))
;; 10/31/88 DNG - added flavor-class and structure-class to the list.
;; 12/02/88 DNG - remove meta classes from the list.
;; 12/06/88 DNG - no more special cases needed.
(defmethod (class :class-hash-key) ()
  class-description)

(comment ; old way
(defmethod (class :class-hash-key) ()
  (if (member self '#,(list ;;(class-named 'standard-class)
			    ;;(class-named 'built-in-class)
			    ;;(class-named 'class)
			    (class-named 'method)
			    (class-named 'standard-method)
			    ;;(class-named 'standard-generic-function)
			    ;;(class-named 'generic-function)
			    ;;(class-named 'flavor-class)
			    ;;(class-named 'structure-class)
			    )
	      :test #'eq)
      (sys:%class-description (class-prototype self))
    class-description)))

;; 12/02/88 - orginal
;; 12/06/88 changed from flavor class-class.
(defmethod (hybrid-class-with-flavor-instances :class-hash-key) ()
  flavor)

(p-defgeneric-options get-hash-entries-from-handler (handler handler-specializers entry-specializers)
		       (declare (values method-location map-list)))

;;  2/23/89 DNG - Fix for single specializer which is (EQL #).
;;		Also modified to be more efficient by not using MAPCAR and COPY-LIST.
;;		Use BACKGROUND-CONS-AREA.
(defmethod (standard-method :get-hash-entries-from-handler) (handler-specializers entry-specializers)
  (flet ((specializer-map-class (specializer)
	     ;; Given a method specializer [not its name], return the class object for which a mapping table is needed.
	     (if (individual-typep specializer)
		 (class-of (individual-type specializer))
	       specializer)))
    (values (get-method-location self)
	    (if (= (length handler-specializers ) 1)
		(get-mapping-table (specializer-map-class (car handler-specializers))
				   (specializer-map-class (car  entry-specializers )))
	      (let ((map-list (make-list (length handler-specializers) :area sys:background-cons-area)))
		(do ((ml map-list (cdr ml))
		     (hs handler-specializers (cdr hs))
		     (es entry-specializers (cdr es)))
		    ((null ml))
		  (setf (car ml)
			(get-mapping-table (specializer-map-class (car hs))
					   (specializer-map-class (car es)))))
		map-list)))))

;;  2/23/89 DNG - Send :GET-HASH-ENTRIES-FROM-HANDLER message to the forwarded method 
;;		instead of duplicating the code here.
(defmethod (null-combined-method  :get-hash-entries-from-handler) (handler-specializers entry-specializers)
  (declare (ignore handler-specializers))
  (send forwarded-method :get-hash-entries-from-handler
	(method-parameter-specializers forwarded-method) entry-specializers))

(comment ; old way
(defmethod (standard-method :get-hash-entries-from-handler) (handler-specializers entry-specializers)
  (values (get-method-location self)
	  (if (= (length handler-specializers ) 1)
	      (get-mapping-table (car handler-specializers) (car  entry-specializers ))
	      (copy-list
		(mapcar #'(lambda (x y)
			    (let ((x (if (individual-typep x)
					 (class-of (individual-type x))
					 x))
				  (y (if (individual-typep y)
					 (class-of (individual-type y))
					 y)))
			      (get-mapping-table x y)))
			handler-specializers  entry-specializers)))))

(defmethod (null-combined-method  :get-hash-entries-from-handler) (handler-specializers entry-specializers)
  (let ((m forwarded-method))
    (values (get-method-location m)
	    (if (= (length handler-specializers ) 1)
		(get-mapping-table (car (method-parameter-specializers  m)) (car  entry-specializers ))
		(copy-list
		  (mapcar #'(lambda (x y)
			      (let ((x (if (individual-typep x)
					   (class-of (individual-type x))
					   x))
				    (y (if (individual-typep y)
					   (class-of (individual-type y))
					   y)))
				(get-mapping-table x y)))
			  (method-parameter-specializers  m)  entry-specializers))))))
)

(defmethod (standard-reader-method :get-hash-entries-from-handler) (handler-specializers entry-specializers)
  (declare (ignore  handler-specializers))
  (let ((offset-or-locative (slot-offset-or-locative slot-name
						     (class-all-slots (car entry-specializers))
						     (class-description-instance-slot-names
						       (class-description (car entry-specializers))))))
    (if (numberp offset-or-locative )
	(VALUES (locf #'read-instance-slot) (1+ offset-or-locative))
	(VALUES (locf #'read-class-slot) offset-or-locative))))

(defun read-instance-slot (instance)
  (sys:%instance-ref instance
		     (COMPILER:%LOCAL-SLOT
		       SYS:LOCAL-FOR-FIRST-MAPPING-TABLE)))

(defun read-class-slot (instance)
  (declare (ignore instance))
  (contents (COMPILER:%LOCAL-SLOT
	      SYS:LOCAL-FOR-FIRST-MAPPING-TABLE)))

(defmethod (standard-writer-method :get-hash-entries-from-handler) (handler-specializers entry-specializers)
  (declare (ignore  handler-specializers))
  (let* ((class (cadr entry-specializers))
	 (offset-or-locative (slot-offset-or-locative slot-name
						      (class-all-slots class)
						      (class-description-instance-slot-names
							(class-description class)))))
    (if (numberp offset-or-locative )
	(VALUES (locf #'write-instance-slot) (1+ offset-or-locative))
      (VALUES (locf #'write-class-slot)     offset-or-locative))))

(defun write-instance-slot (value instance)
  ;;We use the first mapping table, because both of them have the same value.
  (setf
    (sys:%instance-ref instance
		       (COMPILER:%LOCAL-SLOT
			 SYS:LOCAL-FOR-FIRST-MAPPING-TABLE)) value))

(defun write-class-slot (instance value)
  (declare (ignore instance))
  ;;We use the first mapping table, because both of them have the same value.
  (setf (contents (COMPILER:%LOCAL-SLOT
		    SYS:LOCAL-FOR-FIRST-MAPPING-TABLE)) value))

;;  8/12/88 DNG - Use CATCH-ERROR-RESTART instead of IGNORE-ERRORS.
;;  9/20/88 DNG - Use IGNORE-METHOD-COMBINATION-ERROR instead of CATCH-ERROR-RESTART.
;;		Don't call PUTHASH with a NIL value.
;;  2/20/89 DNG - Change hash array size from (length ml) to (truncate (* (+ (length ml) 1) 10) 7)
;;		to ensure that we don't try to make a table of size zero, and to not 
;;		create a table that will need to be rehashed before we finish filling it.
(defun build-method-hash-table (gfunc)
  (let ((nargs (length (generic-function-argument-precedence-order gfunc))))
    (labels
      ((find-handler-carefully (gfunc args)
           (ignore-method-combination-error (find-handler gfunc args)))
       (bht (ml n rargs)
	    (if (= n 0)
		(let* ((args (reverse rargs))
		       (handler (find-handler-carefully gfunc args)))
		  (and handler
		       (get-hash-entries-from-handler
			 handler (method-parameter-specializers handler)
			 (order-parameter-specializers gfunc args))))
	      (if (and (not (= n nargs)) (only-t  ml *t-class* n))
		  (let* ((args (nconc (reverse rargs)
				      (make-list n :initial-element
						 *t-class*)))
			 (handler (find-handler-carefully gfunc args)))
		    (and handler
			 (get-hash-entries-from-handler
			   handler (method-parameter-specializers handler)
			   (order-parameter-specializers gfunc args))))
		;; Size of hash table is current number of methods plus room for one more, 
		;; divided by the rehash threshold.
		(let ((ht (sys:make-hash-array :size (truncate (* (+ (length ml) 1) 10) 7)
					       :number-of-values 2
					       :funcallable-p t :test #'eq)))
		  (dolist (msubl ml ht)
		    (if (individual-typep (car msubl))
			;;Nasty case, we have to have an individual-dispatch hash-table
			(let* ((individual-type (class-of (individual-type (car msubl))))
			       dht)
			  (multiple-value-bind (oe ignore vals)
			      (gethash (class-hash-key individual-type) ht )
			    (if (individual-hash-table-p oe)
				(setq dht oe)
			      (setq dht (make-individual-dispatch-hash-table 
					  :default-value (or (copy-list (cdr vals ))
							     (list nil nil)))))
			    (puthash (class-hash-key individual-type) dht ht t)
			    (multiple-value-bind (subh mtl)
				(bht (cdr msubl) (1- n) (cons (car msubl) rargs))
			      (unless (null subh)
				(puthash (individual-type (car msubl))
					 subh
					 (individual-dispatch-hash-table-ht dht)
					 mtl)))))
		      (multiple-value-bind (oe ignore )
			  (gethash (class-hash-key (car msubl)) ht )
			(if (individual-hash-table-p oe)
			    (multiple-value-bind  (subh mtl)
				(bht (cdr msubl) (1- n)(cons (car msubl) rargs))
			      (setf (individual-dispatch-hash-table-default-value oe)
				    (list subh mtl)))
			  (multiple-value-bind  (subh mtl)
			      (bht (cdr msubl) (1- n) (cons (car msubl) rargs))
			    (unless (null subh)
			      (puthash (class-hash-key (car msubl))
				       subh
				       ht
				       mtl))))))))))))
      (combine-methods gfunc)
      (setf (generic-function-method-hash-table gfunc)
	    (bht (generic-function-method-list  gfunc)
		 nargs
		 nil)))))

;;This function is called out from the microcode.
(defvar sys:%SVC-CLOS-find-generic-hash-table #o57)
(si:%add-to-support-vector '%build-method-hash-table  
			   sys:%SVC-CLOS-find-generic-hash-table t)

(defun %build-method-hash-table  (fef)
  (let* ((ht (build-method-hash-table (get-generic-function-object fef)))
	 (offset (sys:get-debug-info-field (sys:get-debug-info-struct fef)
				       :generic-function-hash-table-offset)))
    (when offset
      (save-hash-table-in-fef fef ht offset))
    ht))

(defun save-hash-table-in-fef (fef ht offset)
  (let ((sys:%inhibit-read-only t))
    (sys:%p-store-contents-offset ht fef offset)))

;;  4/06/89 DNG - Original.
;;  4/29/89 DNG - Invoke PREPARE-EFFECTIVE-CODE if arguments are supplied.
;;  5/02/89 DNG - Fix to always pass the right number of arguments to PREPARE-EFFECTIVE-CODE.
(defun prepare-generic-function (generic-function &rest arguments)
  "Build the generic-function's method hash table if it hasn't been done already.
If arguments are supplied, any necessary combined methods are prepared for 
those arguments.  Only the specialized arguments need be supplied; optional 
arguments will have no effect here."
  (catch-error-restart ((error break) "Give up preparing generic function ~S." generic-function)
    (let ((object (generic-function-named generic-function)))
      (send object :prepare-hash-table)
      (unless (null arguments)
	(apply #'prepare-effective-code object
	       (reorder-parameter-specializers
		 object
		 (firstn (length (specializable-arguments (clos:generic-function-lambda-list object)))
			 arguments))))))
  (values))

;;  4/24/89 DNG - Fix for discriminator-code being a symbol.
(defmethod (standard-generic-function :prepare-hash-table) ()
  (when (and (null method-hash-table)
	     (variable-boundp discriminator-code))
    (let ((fn discriminator-code))
      (when (symbolp fn) (setq fn (symbol-function fn)))
      (when (typep fn 'compiled-function)
	(%build-method-hash-table fn))))
  (values))

(defmethod (generic-function :default :prepare-hash-table) ignore)

(defun only-t (ml class-t n )
  (if (= n 0 ) t
      (and (= 1 (length  ml))
	   (class-eql class-t (caar ml))
	   (only-t (cdar ml) class-t (1- n)))))

(defun put-method-in-hash-table (gfunc method &optional
				 (arg-specializers  nil))
  ;;arg-specializers are already ordered by arguments-precedence-order
  (let*((ht (generic-function-method-hash-table gfunc))
	m
	(specializers (reorder-parameter-specializers
			gfunc
			(method-parameter-specializers method))))
    (multiple-value-bind (meth-loc mtl)
	(get-hash-entries-from-handler method (method-parameter-specializers method)
				       (order-parameter-specializers gfunc (or arg-specializers specializers )))
      (labels
	((put-arg-in-hash-table (args specs ht &optional parent)
	   (declare (list args specs))
	   (if (null args)
	       (values meth-loc mtl)
	       (progn
		 (when (or (null ht) (not (hash-table-p ht)))
		   (if (and parent
			    (every #'(lambda (c) (class-eql c *t-class*)) specs)
			    (locativep (gethash (class-hash-key *t-class*) parent)))
		       ;; This is a default handler, we don't need to cons up new hash tables.
		       (return-from put-arg-in-hash-table meth-loc mtl)
		       (progn
			 ;; ;;There is no right place, since it was a default.
			 ;;(unless (null ht)
			 ;; ;;handler is in there. record it so it can be put back in the right place.
			 ;;(setf m (sys:get-debug-info-field (sys:get-debug-info-struct (contents ht)) :method))
			 ;; (unless m (ferror nil "unable to get the method object from the function ~s" ht))) 
			 (setf ht (sys:make-hash-array  :number-of-values 2 :funcallable-p t :test #'eq)))))
		 (if (individual-typep (car args))
		     (let ((dht (multiple-value-bind (ov ignore vals)
				    (gethash (class-hash-key (class-of (individual-type (car args)))) ht)
				  (if (individual-hash-table-p ov)
				      ov
				      (make-individual-dispatch-hash-table 
					:default-value (or (copy-list (cdr vals)) (list nil nil)))))))
		       (puthash (class-hash-key (class-of (individual-type (car args)))) dht ht t)
		       (multiple-value-bind (subh mtl)
			   (put-arg-in-hash-table (cdr args) (cdr specs)
						  (gethash (individual-type (car args))
							   (individual-dispatch-hash-table-ht dht)))
			 (puthash (individual-type (car args))
				  subh
				  (individual-dispatch-hash-table-ht dht)
				  mtl))
		       ht)
		     (multiple-value-bind (ov ignore )
			 (gethash (class-hash-key (car args)) ht)
		       (if (individual-hash-table-p ov)
			   (multiple-value-bind  (subh mtl)
			       (put-arg-in-hash-table (cdr args) (cdr specs)
						      (car (individual-dispatch-hash-table-default-value ov)))
			     (setf (individual-dispatch-hash-table-default-value ov)
				   (list subh mtl)
				   ))
			   (multiple-value-bind (subh mtl)
			       (put-arg-in-hash-table (cdr args) (cdr specs)
						      (gethash (class-hash-key (car args)) ht)
						      ht)
			     (puthash (class-hash-key (car args))
				      subh
				      ht
				      mtl)))
		       ht))))))
	(put-arg-in-hash-table (or arg-specializers specializers ) specializers  ht)
	(when m
	  (put-method-in-hash-table gfunc m)))))
  (values))



(defun remove-method-in-hash-table  (gen-function arg-classes)
  (let ((ht (generic-function-method-hash-table gen-function))
	value hash-key arg-class)
    (when ht
      (loop
	(setf arg-class (pop arg-classes))
	(setf hash-key (class-hash-key (if (individual-typep arg-class)
					   (class-of (individual-type arg-class))
					   arg-class)))
	(setf value  (gethash hash-key ht))
	(cond ((null value ) (return nil))
	      ((locativep value) (remhash hash-key ht)(return value ))
	      ((individual-hash-table-p value)
	       (setf ht value)
	       (setf value (if (individual-typep arg-class)
			       (or (gethash (individual-type arg-class) (individual-dispatch-hash-table-ht ht))
				   (car (individual-dispatch-hash-table-default-value ht)))
			       (car (individual-dispatch-hash-table-default-value ht))))
	       (cond ((locativep value)
		      (if (individual-typep arg-class)
			  (remhash (individual-type arg-class) (individual-dispatch-hash-table-ht ht))
			  (setf (individual-dispatch-hash-table-default-value ht) (list nil nil)))
		      (return nil))
		     ((null ht) (return nil)))))
	(setf ht value)
	(when (null arg-classes) (return nil))))))

(defun find-right-method (ht &rest args &aux arg  ignor)
  (loop
    (multiple-value-setq (ht ignor )
      (gethash (sys:%class-description  (setf arg (pop args))) ht))
    (cond ((null ht ) (return nil))
	  ((locativep ht) (return ht))
	  ((individual-hash-table-p ht)
	   (setf ht (or (gethash arg (individual-dispatch-hash-table-ht ht))
			(car (individual-dispatch-hash-table-default-value ht))))
	   (unless ht (return nil))))
    (when (null args) (return ht))))

(defun find-handler-for-classes (ht arg-classes &aux arg-class  ignor)
  ;;This function does not return the handler if the handler is gotten 
  ;;by default (optimization of the trailing t)
  (loop
    (setf arg-class (pop arg-classes))
    (multiple-value-setq (ht ignor )
      (gethash (class-hash-key (if (individual-typep arg-class)
				      (class-of (individual-type arg-class))
				      arg-class)) ht))
    (cond ((null ht ) (return nil))
	  ((locativep ht) (return ht ) );(if arg-classes (return nil) (return ht)))
	  ((individual-hash-table-p ht)
	   (setf ht (if (individual-typep arg-class)
			(or (gethash (individual-type arg-class) (individual-dispatch-hash-table-ht ht))
			    (car (individual-dispatch-hash-table-default-value ht)))
			(car (individual-dispatch-hash-table-default-value ht))))
	   (cond ((locativep ht) (return ht))
		 ((null ht) (return nil)))))
    (when (null arg-classes) (return ht))))


(p-defgeneric-options cache-handler-for-arguments (gfunc method &optional args))

(ticl:defmethod (standard-generic-function :cache-handler-for-arguments) (meth &optional args)
  (let* ((gfunc self)
	 (specializers (method-parameter-specializers meth)))
    ;;If it is an invididual method, stick it where it was intended for.
    (if (null args)
	(put-method-in-hash-table gfunc meth)
	;;try to generalize, based on the type of the args.
	(put-method-in-hash-table gfunc meth (mapcar #'(lambda (arg  specializer)
							     (if (individual-typep specializer)
								  specializer
								 (class-of arg)))
							 args  specializers)))))


(defun handler-equal (handler1 handler2)
  (and (eq (*method-generic-function handler1)
	   (*method-generic-function handler2))
       (equal (method-parameter-specializers handler1 )
	      (method-parameter-specializers handler2 ))))

(p-defgeneric-options handler-cached (class handler))

(ticl:defmethod (standard-class :handler-cached) (handler)
  (pushnew handler (class-cached-handlers self) :test #'handler-equal))

(ticl:defmethod (built-in-class :handler-cached) (handler)
  (pushnew handler (class-cached-handlers self) :test #'handler-equal))




(defun map-all-hash-entries (gfunc hf)
  "Go through the hash tables of a generic function
and map on the entries specialized on class.
HF will be called with the following args:
  Hash-Key, or nil if the entry is the default value of an individual hashtable.
  First value, Second value of the hash entry.
  The hash table this entry comes from.
  The reverse of keys so far, does not include key."
  (let ((ht (generic-function-method-hash-table gfunc)))
    (labels ((map-table (ht args)
		 (declare (optimize (speed 3) (safety 0))) ; because we spend a lot of time here.
		 (maphash #'(lambda (key mloc mtl)
			      (funcall hf key mloc mtl ht args)
			      (unless (locativep mloc)
				  ;;method found
				(if (individual-hash-table-p mloc)
				    (progn
				      (maphash #'lf mloc mloc args)
				      (when (car (individual-dispatch-hash-table-default-value mloc))
					(with-stack-list* (args key args) 
					  (funcall #'lf nil
						   (car (individual-dispatch-hash-table-default-value mloc))
						   (cadr (individual-dispatch-hash-table-default-value mloc))
						   mloc args))))
				  (unless (null mloc) ; check added 8/25/88 by DNG
				    (with-stack-list* (args key args) 
				      (map-table mloc args))))))
			 ht))
	     (lf (key mloc mtl ht args)
		 (if key
		     (with-stack-list  (key  'eql key)
		       (funcall hf key mloc mtl ht args))
		     (funcall hf key mloc mtl ht args))
		 (unless (locativep mloc)
		   ;;method found
		   (if key
		     (with-stack-list  (key  'eql key)
		       (with-stack-list* (args key args) 
			 (map-table mloc args)))
		     (map-table mloc args)))))
      (when ht
	(map-table ht nil)))))




(p-defgeneric-options uncache-effective-methods (gfunc class) )

(defmethod (standard-generic-function :uncache-effective-methods ) (class)
  (let ((cd (class-description class)))
    (map-all-hash-entries self
			  #'(lambda (key  ignore ignore ht ignore)
			      (when (eq cd key)
				(remhash key ht))))))