; -*- Mode:Common-Lisp; Package:TICLOS;  Base:10.-*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.


;;This file contains the new CLOS version of PRINT-OBJECT.

(defvar si:*prindepth* nil
  "Used to pass print depth information to new print-object methods")
(defvar si:*which-operations* nil
  "Used by streams which have a print handler and recursively call
print-object.")


(defgeneric print-object (object stream)
  (:documentation "Write the printed representation of OBJECT to STREAM."))

(defmethod print-object ((pobj standard-object) stream)
  (sys:print-object pobj 0 stream)
  pobj)

(defmethod print-object ((pobj cons) stream)
  (if (and *print-level* (>= (or si:*prindepth* 0) *print-level*))
      (si:print-raw-string (si:pttbl-prinlevel *readtable*) stream t )
      (progn
	(if *print-pretty*
	    (if *print-escape*
		(si:pprin1 pobj stream)
		(si:pprinc pobj stream))
	    (si:print-list pobj (or si:*prindepth* 0) stream si:*which-operations* ))))
  pobj)

(defmethod print-object ((pobj string) stream)
  (if (<= (array-active-length pobj) (array-total-size pobj))
      (si:print-quoted-string pobj stream t)
      (si:print-random-object pobj stream t (or si:*prindepth* 0)
				    si:*which-operations* ))
  pobj)

(defmethod print-object ((pobj array) stream)
  (if (stringp pobj)
      (if (<= (array-active-length pobj) (array-total-size pobj))
	  (si:print-quoted-string pobj stream t)
	  (si:print-random-object pobj stream t (or si:*prindepth* 0)
				  si:*which-operations* ))
      (si:print-array pobj stream t  (or si:*prindepth* 0) si:*which-operations* ))
  pobj)
  
(defmethod print-object ((pobj fixnum) stream)
  (si:print-fixnum pobj stream)
  pobj)

(defmethod print-object ((pobj bignum) stream)
  (si:print-bignum pobj stream t)
  pobj)

(defmethod print-object ((pobj symbol) stream)
  (si:print-pname-string pobj stream t)
  pobj)

(defmethod print-object ((pobj float) stream)
  (si:print-flonum pobj stream ())
  pobj)

(defmethod print-object ((pobj rational) stream)
  (si:print-rational pobj stream t)
  pobj)

(defmethod print-object ((pobj complex) stream)
  (si:print-complex pobj stream t)
  pobj)

(defmethod print-object ((pobj number) stream)
  (si:print-raw-string (si:pttbl-open-random *readtable*) stream t)
  (si:print-raw-string (symbol-name (data-type pobj)) stream t)
  (send stream :tyo (si:pttbl-space *readtable*))
  (let ((*print-base* 8.)
	(*print-radix* NIL))
    (si:print-fixnum (si:%pointer pobj) stream))
  (si:print-raw-string (si:pttbl-close-random *readtable*) stream t)
  pobj)

(defmethod print-object ((pobj character) stream)
  (if (not *print-escape*)
      (write-char pobj stream)
      (progn
	(send stream :string-out
	      (si:pttbl-character-before-font *readtable*))
	(if (ldb-test si:%%ch-font pobj)
	    (let ((*print-base* 10.)
		  (*print-radix* NIL)
		  (*nopoint t))
	      (prin1 (ldb si:%%ch-font pobj) stream)))
	(send stream :string-out
	      (si:pttbl-character-prefix *readtable*))
	(let ((real-bits (ldb  si:%%kbd-control-meta pobj))
	      (char (char-code pobj)))
	  (send stream :string-out
		(nth real-bits
		     '("" "c-" "m-" "c-m-" "s-" "c-s-" "m-s-"
		       "c-m-s-" "h-" "c-h-" "m-h-" "c-m-h-" "s-h-"
		       "c-s-h-" "m-s-h-" "c-m-s-h-")))
	  (let ((chname (si:ochar-get-character-name
			  (dpb 0 si:%%kbd-control-meta
			       (dpb 0 si:%%ch-font pobj)))))
	    (if chname (send stream :string-out chname)
		(progn
		  (when (char-bit pobj :mouse)
		    (write-string "mouse-" stream))
		  (when (char-bit pobj :keypad)
		    (write-string "keypad-" stream))
		  (when (and (/= (char-bits pobj) 0) (si:character-needs-quoting-p char))
		    (princ (si:pttbl-slash *readtable*)  stream))
		  (send stream :tyo char)))))))
  pobj)


(defmethod print-object (pobj stream)
  (let ((si:*which-operations*  (si:which-operations-for-print stream)))
    (catch-continuation-if t 'print-object
	#'(lambda ()
	    (format stream "...error printing ")
	    (si:printing-random-object (pobj stream :typep :fastp t ))
	    (format stream "..."))
	()
      (condition-resume
	'((error) :abort-printing t ("Give up trying to print this object.")
	  si:catch-error-restart-throw print-object)
	(let (nss)
	  (typecase pobj
	    (named-structure
	     (ignore-errors (setq nss (named-structure-p pobj)))
	     (cond
	       ((and (symbolp nss)
		     (or (get nss 'named-structure-invoke)
			 (get nss :named-structure-invoke))
		     (member :print-self
			     (named-structure-invoke pobj :which-operations)
			     :test #'eq))
		(named-structure-invoke pobj :print-self stream (or si:*prindepth* 0)
					*print-escape*))
	       (t				   ;Named structure that doesn't print itself
		(si:print-named-structure nss pobj (or si:*prindepth* 0) stream
					  si:*which-operations* ))))
	    (instance (send pobj :print-self stream (or si:*prindepth* 0) *print-escape*))
	    (t (si:print-random-object pobj stream t (or si:*prindepth*  0) si:*which-operations* )))
	  )
	pobj) ) ) )


;this should really be :around the default method, but that isn't ready yet
(defmethod print-object :around ((pobj standard-object) stream)
  (let ((si:*which-operations*  (si:which-operations-for-print stream)))
    (catch-continuation-if t 'print-object
	#'(lambda ()
	    (format stream "...error printing ")
	    (si:printing-random-object (pobj stream :typep :fastp t ))
	    (format stream "..."))
	()
      (condition-resume
	'((error) :abort-printing t ("Give up trying to print this object.")
	  si:catch-error-restart-throw print-object)
      
	(or
	  (and (member :print si:*which-operations*  :test #'eq)
	       ;Allow stream to intercept print operation
	       (send stream :print pobj (or si:*prindepth* 0) *print-escape*))
	  
	  (and *print-circle* (si:%pointerp pobj)
	     (or (not (symbolp pobj)) (not (symbol-package pobj)))
	     ;; This is a candidate for circular or shared structure printing.
	     ;; See what the hash table says about the object:
	     ;; NIL - occurs only once.
	     ;; T - occurs more than once, but no occurrences printed yet.
	     ;; Allocate a label this time and print #label= as prefix.
	     ;; A number - that is the label.  Print only #label#.
	     
	     (catch 'label-printed
	       (si:modifyhash pobj si:print-hash-table 
			   #'(lambda (key
				       value
				       key-found-p
				       stream)
			       key
			       key-found-p
			       (cond
				 ((null value) NIL)
				 ((eq value t)
				  (let ((label (incf si:print-label-number))
					(*print-base* 10.)
					(*print-radix* NIL)
					(*nopoint t))
				    (send stream :tyo #\#)
				    (si:print-fixnum label stream)
				    (send stream :tyo #\=)
				    label))
				 (t
				  (let ((*print-base* 10.)
					(*print-radix* NIL)
					(*nopoint t))
				    (send stream :tyo #\#)
				    (si:print-fixnum value stream)
				    (send stream :tyo #\#)
				    (throw 'label-printed
					   t)))))
			   stream)
	       ()))
	  (when (next-method-p)
	    (call-next-method))  ) )) )
  pobj)

;;; 05/08/89 clm - added the following two forms to make CLEH use CLOS.
(defmethod print-object ((pobj eh:condition) stream)
   (if *print-escape*
       (call-next-method)
       (send pobj :report stream))
  pobj) ;default case - no handlers - give control back to EH :report function


eh:
(defmethod (condition :print-error-message) (sg brief stream)
  (send self :print-error-message-prefix sg brief stream)
  (princ self stream)
  (terpri stream))
 

;;dummy call to print-object to get combined-method generated
;;using *standard-output* here may be incorrect, not sure what it will 
;;be bound to
(eval-when (eval compile load)
  (print-object "Generating the combined-method for PRINT-OBJECT." #'si:null-stream))