; -*- Mode:Common-Lisp; Package:si; Cold-load:t; Base:8; Patch-file:T -*-

;;The following system functions have been changed to call the new CLOS:PRINT-OBJECT.
;;These functions are taken from the file "sys:kernel;print".

(defun print (object &optional stream)
  "Print OBJECT on STREAM with quoting if needed, with a Return before and a Space after."
  (setq stream (decode-print-arg stream))
  (funcall stream :tyo (pttbl-newline *readtable*))
  (let ((*print-escape* t)
	(character-attribute-table (character-attribute-table *readtable*)))
    (if *print-pretty*
	(pprint object stream)
	;(progn
	(let ((*prindepth* 0))
	  (print-circle (ticlos:print-object object stream))
	  (funcall stream :tyo (pttbl-space *readtable*)))))
  object)

(defun prin1 (object &optional stream)
  "Print OBJECT on STREAM with quoting if needed."
  (setq stream (decode-print-arg stream))
  (let ((*print-escape* t)
	(character-attribute-table (character-attribute-table *readtable*)))
    (if *print-pretty*
	(pprin1 object stream)
	(let ((*prindepth* 0))
	      (print-circle (ticlos:print-object object stream )))))
  object) 


(defun princ (object &optional stream)
  "Print OBJECT with no quoting, on STREAM.
Strings and characters print just their contents with no delimiters or quoting.
Pathnames, editor buffers, host objects, and many other hairy things
 print as their names with no delimiters."
  (setq stream (decode-print-arg stream))
  (let ((*print-escape* NIL)
	(character-attribute-table (character-attribute-table *readtable*)))
    (if *print-pretty*
	(pprinc object stream)
	(let ((*prindepth* 0))
	  (print-circle (ticlos:print-object object stream)))))
  object)


;;1/3/89 clm - Fix for SPR 9077 included (wrapped PRINT-CIRCLE around the call to OUTPUT-PRETTY-OBJECT). 
(defun write (object &key &optional (stream *standard-output*) ((:escape *print-escape*) *print-escape*)
  ((:radix *print-radix*) *print-radix*) ((:base *print-base*) *print-base*)
  ((:circle *print-circle*) *print-circle*) ((:pretty *print-pretty*) *print-pretty*)
  ((:level *print-level*) *print-level*) ((:length *print-length*) *print-length*)
  ((:case *print-case*) *print-case*) ((:gensym *print-gensym*) *print-gensym*)
  ((:array *print-array*) *print-array*))
  "Print OBJECT on STREAM.  Keyword args control parameters affecting printing.
The argument ESCAPE specifies the value for the flag *PRINT-ESCAPE*, and so on.
For any flags not specified by keyword arguments, the current special binding is used."
  (let ((character-attribute-table (character-attribute-table *readtable*)))
    (if *print-pretty*
	(let ((*standard-output* (decode-print-arg stream)))
	  (print-circle (output-pretty-object object)))
	(let ((*prindepth* 0))
	  (print-circle (ticlos:print-object object (decode-print-arg stream))) ) )
  object))


;;the new ticlos version of prin1-then-space 
(defun prin1-then-space (object &optional stream)
  "Print OBJECT on STREAM with quoting if needed, followed by a Space character."
  (setq stream (decode-print-arg stream))
  (let ((*print-escape* t)
	(character-attribute-table (character-attribute-table *readtable*)))
    (if *print-pretty*
	(pprin1 object stream)
	(let ((*prindepth* 0))
	  (print-circle (ticlos:print-object object stream))))
  (funcall stream :tyo (pttbl-space *readtable*)))
  object)

;;the old prin1-then-space 
;(defun prin1-then-space (object &optional stream)
;  "Print OBJECT on STREAM with quoting if needed, followed by a Space character."
;  (setq stream (decode-print-arg stream))
;  (let ((*print-escape* t)
;	(character-attribute-table (character-attribute-table *readtable*)))
;    (if *print-pretty*
;	(pprin1 object stream)
;	(print-circle (print-object object 0 stream)))
;  (funcall stream :tyo (pttbl-space *readtable*)))
;  object)


;;;new clos version
(defun print-random-object (exp stream fastp i-prindepth which-operations)
  (declare (ignore fastp))
  (printing-random-object (exp stream :fastp t)
                          (print-raw-string (string (data-type exp)) stream t)
                          (typecase exp
                            (microcode-function
			     (unless (>= (%pointer exp)
					 (Array-Total-Size #'MICRO-CODE-ENTRY-DEBUG-INFO-AREA))
			       (funcall stream :tyo (pttbl-space *readtable*))
			       (let ((*prindepth* i-prindepth)
				     (*which-operations* which-operations))
				 (ticlos:print-object (function-name exp) stream))))
			    ((or compiled-function stack-group)
			     (funcall stream :tyo (pttbl-space *readtable*))
                             (let ((*prindepth* i-prindepth)
				   (*which-operations* which-operations))
			       (ticlos:print-object (function-name exp) stream))))))


(defun print-list (exp i-prindepth stream which-operations)
  (funcall stream :tyo #\()
  (do ((i-prinlength 1 (1+ i-prinlength))
       (first t ()))
      ((or (atom exp) (and *print-circle* (not first) (gethash exp print-hash-table)))
       (cond
	 ((not (null exp))
	  (print-raw-string " . " stream t)
	  (let ((*prindepth* (1+ i-prindepth))
		(*which-operations* which-operations)) 
	    (ticlos:print-object exp stream))))
       (funcall stream :tyo #\)))
    (or first (funcall stream :tyo #\space))
    (let ((*prindepth* (1+ i-prindepth))
	  (*which-operations* which-operations)) 
      (ticlos:print-object (car exp) stream))
    (setq exp (cdr exp))
    (and *print-length* (>= i-prinlength *print-length*);One frob gets printed before test.
                     (not (atom exp));Don't do it uselessly
                                                  
         (progn
           (send stream :tyo #\space)
           (print-raw-string "..." stream t)
           (send stream :tyo #\))
           (return ())))))


(defun print-vector (exp i-prindepth stream which-operations)
  (funcall stream :string-out (pttbl-open-vector *readtable*))
  (do ((i-prinlength 0 (1+ i-prinlength))
       (length (length exp))
       (first t ()))
      ((= i-prinlength length) (funcall stream :string-out (pttbl-close-vector *readtable*)))
    (or first (funcall stream :tyo (pttbl-space *readtable*)))
    (let ((*prindepth* (1+ i-prindepth))
	  (*which-operations* which-operations))
      (ticlos:print-object (aref exp i-prinlength) stream))
    (and *print-length* (>= i-prinlength *print-length*);One frob gets printed before test.
                    
         (progn
           (send stream :tyo (pttbl-space *readtable*))
           (print-raw-string (pttbl-prinlength *readtable*) stream t)
           (send stream :tyo (pttbl-close-paren *readtable*))
           (return ())))))


(defun print-array-contents (array dimension index-so-far i-prindepth stream w-o &aux tem)
  (if (and *print-level* (>= i-prindepth *print-level*))
      (send stream :string-out (pttbl-prinlevel *readtable*))
      (if (zerop (array-rank array))
	  (let ((*prindepth* i-prindepth)
		(*which-operations* w-o))
	    (ticlos:print-object (aref array) stream))
          (let ((index (* index-so-far (array-dimension array dimension)))
                (mode
                 (car (member (array-type array) '(art-1b art-string art-fat-string) :test #'eq)))
                (length (array-dimension array dimension))
                (rank (array-rank array)))
            (cond
              ((and mode (= (1+ dimension) rank))
               (let ((kludge
                      (if
                       (and
                        (%store-conditional (locf array-contents-array)
                                            (setq tem array-contents-array) ())
                        tem)
                       (change-indirect-array tem mode length array index)
                       (make-array length :type mode :displaced-to array
                                   :displaced-index-offset index))))
                 (if (eq mode 'art-1b) (print-bit-vector kludge stream);
                                       
                     (print-quoted-string kludge stream (member :string-out w-o :test #'eq)))
                 (setq array-contents-array kludge)));massachusetts
              
              (t (send stream :tyo (pttbl-open-paren *readtable*))
               (dotimes (i (array-dimension array dimension))
                 (or (zerop i) (send stream :tyo (pttbl-space *readtable*)))
                 (cond
                   ((and *print-length* (= i *print-length*))
                    (send stream :string-out (pttbl-prinlength *readtable*)) (return))
                   ((= (1+ dimension) (array-rank array))
                    (let ((*prindepth* (1+ i-prindepth))
			  (*which-operations* w-o))
		      (ticlos:print-object (ar-1-force array (+ index i)) stream)))
                   ((and *print-level* (>= (1+ i-prindepth) *print-level*))
                    (send stream :string-out (pttbl-prinlevel *readtable*)))
                   (t
                    (print-array-contents array (1+ dimension) (+ index i) (1+ i-prindepth)
                                          stream w-o))))
               (send stream :tyo (pttbl-close-paren *readtable*))))))))


(defun print-named-structure (nss exp i-prindepth stream which-operations)
  (declare (special *print-structure*))
  (let ((description (get nss 'defstruct-description)))
    (if (or (not description)
	    (if (boundp '*print-structure* )
		 (null *print-structure*)
		 (null *print-array*)))
	(printing-random-object (exp stream :typep))
        (progn
          (funcall stream :string-out "#S")
          (let ((slot-alist (defstruct-description-slot-alist))
                (l (list nss)))
            (dolist (s slot-alist)
	      (unless (defstruct-slot-description-name-slot-p (cdr s))
		(let* ((kwd (intern (symbol-name (car s)) pkg-keyword-package))
		       (fun (defstruct-slot-description-ref-macro-name (cdr s)))
		       (init (defstruct-slot-description-init-code (cdr s)))
		       (val (eval1 `(,fun ,exp))));watch out for macros!
		  
		  (unless (equal val init)
		    (push kwd l)
		    (push val l)))))
            (let ((*prindepth* i-prindepth)
		  (*which-operations* which-operations))
	      (ticlos:print-object (nreverse l) stream))))
        )))


;;the following functions appear in file "SYS:KERNEL;PPRINT"

;;1/9/89 clm - changed to use CLOS:PRINT-OBJECT
(defun pp-objify (object location &optional (currlevel 0))
  "Returns a PP-Obj structure which is used to prettily print the Object."
  (declare (special pprint-string))
  (flet ((objify (object currlevel )
		 (typecase object
		   (cons (pp-objify-list object location currlevel))
		   (null pp-nil-obj)
		   (structure (pp-objify-atom object location ))
		   (string (pp-objify-atom object location ))
		   (vector (pp-objify-vector object location currlevel))
		   (array (pp-objify-array object location currlevel))
		   (t (pp-objify-atom object location )))))
    (if (and *print-level* (> currlevel (the fixnum *print-level*)))
	pp-sharp-obj
	(if (and *print-circle* 
		 (or (consp object) (and (arrayp object) (not (named-structure-p object)))))
	     ;; This is a candidate for circular or shared structure printing.
	     ;; See what the hash table says about the object:
	     ;; NIL - occurs only once.
	     ;; T - occurs more than once, but no occurrences printed yet.
	     ;;  Allocate a label this time and print #label= as prefix.
	     ;; A number - that is the label.  Print only #label#.
	    (catch 'label-printed
	      (let ((lab
		      (modifyhash  object print-hash-table 
			    #'(lambda (key
				       value
				       key-found-p)
				key
				key-found-p
				(cond
				  ((null value)  nil)
				  ((eq value t) 
				   (incf print-label-number))
				  (t  (let ((*print-base* 10.)
					    (*print-radix* NIL)
					    (*print-circle* nil)
					    (*print-pretty* nil)
					    (*nopoint t)
					    (start (length pprint-string))
					    (*prindepth* 0)  ;;clm - added for CLOS
					    (*which-operations* '(:string-out)))
					(write-char #\#)
					(ticlos:print-object value *standard-output*)
					(write-char #\#)
					(throw 'label-printed
					       (make-pp-obj
						 :length (- (length pprint-string)
							    start)
						 :object start))))))))

		    ob ob1)
		(setf ob1 (objify object currlevel ))
		(if lab
		    (progn
		      (setf ob
			    (let ((*print-base* 10.)
				  (*print-radix* NIL)
				  (*print-circle* nil)
				  (*print-pretty* nil)
				  (*nopoint t)
				  (start (length pprint-string))
				  (*prindepth* 0)
				  (*which-operations* '(:string-out)))
			      (write-char #\#)
			      (ticlos:print-object lab *standard-output*)
			      (write-char #\=)
			      (make-pp-obj
				:length (- (length pprint-string)
					   start)
				:object start
				:location location)))
		      (make-pp-obj :length (+ (pp-obj-length ob)
					      (pp-obj-length ob1))
				   :type 'complex
				   :object (list ob ob1)
				   :location location))
		      ob1)))
	    (objify object currlevel )))))

;;1/9/89 clm - changed to use CLOS:PRINT-OBJECT
(defun pp-objify-atom (object location )
  "Makes a PP-Obj for an atom."
  (declare (special pprint-string))
  (let ((start (length pprint-string))
	(*prindepth* 0)
	(*which-operations* '(:string-out)))
      (ticlos:print-object object *standard-output*)
     ;(output-object object)
    (make-pp-obj :length (- (length pprint-string)
			    start)
		 :object start
		 :location location)))

;;1/9/89 clm - changed to use CLOS:PRINT-OBJECT
(defun master-grind (object charpos)
  (declare (special pprint-string))
  ;;;phd 11/25/85 Fixed printing of NIL.
  (cond ((eq (pp-obj-type object) 'simple)
	 
	 (let ((start (pp-obj-object object))
	       (len (pp-obj-length object)))
	   (when *pprint-notify-fun* 
	     (funcall *pprint-notify-fun* object
		      (pp-obj-location object)
		      t 
		      len))
	   (write-string pprint-string
			 *standard-output* :start start
			 :end (+ start len)))
	 nil)
	((eq (pp-obj-type object) 'string)
	 (if (eq object pp-nil-obj)
	     (progn
	       (when  *pprint-notify-fun* 
		 (funcall *pprint-notify-fun* object
			  (pp-obj-location object)
			  t 
			  3))
	       (let ((*prindepth* 0)
		     (*which-operations* '(:string-out)))
		 (ticlos:print-object nil *standard-output*)))
	     (write-string (pp-obj-object object)))
	 nil)
	((eq (pp-obj-type object) 'comment)
	 (comment-grind object 'past-name charpos))
	((pp-obj-callish object)
	 (when *pprint-notify-fun* 
	   (funcall *pprint-notify-fun* 'start-of-object
		    (pp-obj-location object)
		    nil
		    nil))
	 (cond ((get (pp-obj-callish object) 'specially-grind)
		(specially-grind object charpos))
	       ((get (pp-obj-callish object) 'hairily-grind)
		(hairily-grind object charpos))
	       ;; phd 2/4/86 tried to have a better default for macros
	       ((or (special-form-p (pp-obj-callish object))
		    (macro-function (pp-obj-callish object)))
		(basically-grind object () t 3 charpos))
	       (t
		(basically-grind object () t  'past-name charpos)))
	 (when *pprint-notify-fun* 
	   (funcall *pprint-notify-fun* 'end-of-object
		    nil
		    nil
		    nil )))
	(t
	 (when *pprint-notify-fun* 
	   (funcall *pprint-notify-fun* 'start-of-object
		    (pp-obj-location object)
		    nil
		    nil))
	 (basically-grind object () t  'normal charpos)
	 (when *pprint-notify-fun* 
	   (funcall *pprint-notify-fun* 'end-of-object
		    nil
		    nil
		    nil )))))