;;;-*- Mode:Common-Lisp; Package:System-Internals; Base:10 -*-

;;;	Common Lisp Object System for Explorer and microExplorer
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.

;;; This file defines the packages used.

;;  3/16/89 DNG - Original version of this file (previously in DEFSYSTEM file).
;;  4/07/89 DNG - Export MAKE-LOAD-FORM, MAKE-LOAD-FORM-SAVING-SLOTS,  
;;		LOAD-TIME-VALUE, and DESCRIBE-OBJECT from CLOS.
;;  4/19/89 DNG - Export CLOS-INSTANCE-P from TICLOS.
;;  4/20/89 DNG - Export REAL and FDEFINITION.  Remove obsolete function RECONSTRUCTION-FORM.  
;;  4/21/89 DNG - Export COPY from TICLOS since it is no longer in LISP.
;;  4/24/89 DNG - Export FUNCTION-LAMBDA-EXPRESSION .
;;  4/29/89 DNG - Export PREPARE-GENERIC-FUNCTION from TICLOS.
;;  5/08/89 DNG - Export DELETE-CLASS from TICLOS.

;; Temporary for use under release 4.
(unless (fboundp 'sys:make-flavor-instance)
  (setf (symbol-function 'sys:make-flavor-instance)
	(if (fboundp 'zlc:make-instance)
	    #'zlc:make-instance
	  #'ticl:make-instance)))

(defpackage "TICLOS" (:use "TICL" "LISP") (:prefix-name "TICLOS"))

(defpackage "CLOS" (:use "LISP"))

;; Temporary for release 4:
;; Make a new symbol for MAKE-INSTANCE instead of redefining the old one 
;; because too many things use it to be able to have it momentarily undefined 
;; while changing the definition.
(let (old)
  (setq old (find-symbol "MAKE-INSTANCE" "TICL"))
  (when (eq old (find-symbol "MAKE-INSTANCE" "ZLC")) ; once only
    (unintern old "TICL")
    (shadowing-import old "ZLC")
    (let ((new (intern "MAKE-INSTANCE" "TICL")))
      (setf (symbol-plist new) (copy-list (symbol-plist old)))
      (export (list new) "TICL")
      (export (list new) "TICLOS")
      (setf (symbol-function new) #'sys:make-flavor-instance) )))

;; export the documented symbols
(dolist (name '(
	  ;; Symbols from chapters 1 and 2 of the CLOS spec [88-002R]:
	   "ADD-METHOD"
	   "BUILT-IN-CLASS"
	   "CALL-METHOD" "CALL-NEXT-METHOD"  "CHANGE-CLASS" "CLASS-NAME"
           "CLASS-OF"  "COMPUTE-APPLICABLE-METHODS"
	   "DEFCLASS" "DEFGENERIC" "DEFINE-METHOD-COMBINATION" "DEFMETHOD"
	   "ENSURE-GENERIC-FUNCTION"
	   "FIND-CLASS" "FIND-METHOD" "FUNCTION-KEYWORDS"
	   "GENERIC-FLET" "GENERIC-FUNCTION" "GENERIC-LABELS"
           "INITIALIZE-INSTANCE" "INVALID-METHOD-ERROR"
	   "MAKE-INSTANCE" "MAKE-INSTANCES-OBSOLETE" "METHOD-COMBINATION"
	   "METHOD-COMBINATION-ERROR" "METHOD-QUALIFIERS" 
           "NEXT-METHOD-P" "NO-APPLICABLE-METHOD" "NO-NEXT-METHOD"
	   "PRINT-OBJECT"
	   "REINITIALIZE-INSTANCE" "REMOVE-METHOD"
	   "SHARED-INITIALIZE"
	   "SLOT-BOUNDP" "SLOT-EXISTS-P" "SLOT-MAKUNBOUND" "SLOT-MISSING" 
	   "SLOT-UNBOUND" "SLOT-VALUE"
	   "STANDARD" "STANDARD-CLASS" "STANDARD-GENERIC-FUNCTION"
	   "STANDARD-METHOD" "STANDARD-OBJECT" "STRUCTURE-CLASS" "SYMBOL-MACROLET"
	   "UPDATE-INSTANCE-FOR-DIFFERENT-CLASS" "UPDATE-INSTANCE-FOR-REDEFINED-CLASS"
	   "WITH-ADDED-METHODS" "WITH-SLOTS" "WITH-ACCESSORS"

	  ;; Incomplete list of symbols from chapter 3, March 1988 draft [88-003]:
	   ;; 	[some of these are now obsolete]
	   "ADD-METHOD-ON-SPECIALIZER" "ADD-NAMED-CLASS" "ALLOCATE-INSTANCE"
	   "CHECK-INITARGS"
	   ;; "CLASS" ; can't export from TICLOS yet because conflicts with ZLC:CLASS in USER.
	   "CLASS-DIRECT-GENERIC-FUNCTIONS" "CLASS-DIRECT-INITARG-DEFAULTS"
	   "CLASS-DIRECT-METHODS" "CLASS-DIRECT-SLOT-INITARGS" "CLASS-DIRECT-SLOTS"
	   "CLASS-DIRECT-SUBCLASSES" "CLASS-DIRECT-SUPERCLASSES"
	   "CLASS-FINALIZED-P" "CLASS-FOR-REDEFINITION" "CLASS-INITARGS"
	   "CLASS-INITARG-DEFAULTS" "CLASS-PRECEDENCE-LIST" "CLASS-PROTOTYPE"
	   "COMPUTE-CLASS-PRECEDENCE-LIST" "CLASS-SLOTS"
	   "DEFAULT-INITARGS" "DEFINE-INSTANCE-ACCESS-OPTIMIZATION"
	   "GENERIC-FUNCTION-DECLARATIONS" "GENERIC-FUNCTION-METHODS"
	   ;; Don't export yet; exist but are not generic functions.
	   ;;"GENERIC-FUNCTION-ARGUMENT-PRECEDENCE-ORDER" "GENERIC-FUNCTION-INITIAL-METHODS"
	   ;;"GENERIC-FUNCTION-LAMBDA-LIST" "GENERIC-FUNCTION-METHOD-CLASS"
	   ;;"GENERIC-FUNCTION-METHOD-COMBINATION" "GENERIC-FUNCTION-NAME"
	   "METHOD" "METHOD-APPLICABLE-CLASS" "METHOD-FUNCTION" "METHOD-GENERIC-FUNCTION"
	   "METHOD-LAMBDA-LIST" "METHOD-QUALIFIERS" "METHOD-SPECIALIZERS"
	   "OPTIMIZE-SLOT-VALUE" "OPTIMIZE-SETF-SLOT-VALUE"
	   "REMOVE-METHOD-ON-SPECIALIZER"
	   "SLOT-DESCRIPTION"
	   "SLOT-DESCRIPTION-NAME" "SLOT-DESCRIPTION-INITFORM" "SLOT-DESCRIPTION-INITFUNCTION"
	   "SLOT-DESCRIPTION-TYPE" "SLOT-DESCRIPTION-ALLOCATION" "SLOT-DESCRIPTION-INITARGS"
	   "SLOT-DESCRIPTION-READERS" "SLOT-DESCRIPTION-WRITERS"
	   "SLOT-BOUNDP-USING-CLASS" "SLOT-EXISTS-P-USING-CLASS"
	   "SLOT-MAKUNBOUND-USING-CLASS" "SLOT-VALUE-USING-CLASS"
	   "STANDARD-ACCESSOR-METHOD" "STANDARD-INSTANCE-ACCESS" "STANDARD-SLOT-DESCRIPTION"
	   "STANDARD-STRUCTURE"
	   "VALID-SUPERCLASS-P"

	   ;; More symbols from December 1988 draft of chapter 3 [89-003]:
	   ;;	new function names
	   "ADD-DEPENDENT" "ADD-DIRECT-METHOD" "ADD-DIRECT-SUBCLASS" "APPLY-METHOD-FUNCTION"
	   "CLASS-DEFAULT-INITARGS" "CLASS-DIRECT-DEFAULT-INITARGS"
	   "COMPUTE-DISCRIMINATING-FUNCTION" "COMPUTE-EFFECTIVE-METHOD"
	   "COMPUTE-EFFECTIVE-SLOT-DEFINITION" "COMPUTE-SLOTS"
	   "DIRECT-SLOT-DEFINITION-CLASS" "EFFECTIVE-SLOT-DEFINITION-CLASS"
	   "ENSURE-CLASS" "ENSURE-CLASS-USING-CLASS" "ENSURE-GENERIC-FUNCTION-USING-CLASS"
	   "EXTRACT-LAMBDA-LIST" "EXTRACT-SPECIALIZER-NAMES"
	   "FIND-METHOD-COMBINATION"
	   "FUNCALLABLE-STANDARD-CLASS" "FUNCALLABLE-STANDARD-INSTANCE-ACCESS"
	   "MAKE-METHOD-LAMBDA" "MAP-DEPENDENTS"
	   "METHOD-COMBINATION-NAME" "METHOD-COMBINATION-OPTIONS"
	   "METHOD-COMBINATION-ORDER" "METHOD-COMBINATION-OPERATOR"
	   "METHOD-COMBINATION-IDENTITY-WITH-ONE-ARGUMENT"
	   "REMOVE-DEPENDENT" "REMOVE-DIRECT-METHOD" "REMOVE-DIRECT-SUBCLASS"
	   "SET-FUNCALLABLE-INSTANCE-FUNCTION"
	   "SLOT-DEFINITION-ELIDE-ACCESS-METHOD-P" ; page 3-25
	   "SPECIALIZER-DIRECT-GENERIC-FUNCTIONS" "SPECIALIZER-DIRECT-METHODS"
	   "UPDATE-DEPENDENT" "UPDATE-DISCRIMINATING-FUNCTION"
	   "VALIDATE-GENERIC-FUNCTION-CLASS-CHANGE" "VALIDATE-METACLASS-CHANGE" "VALIDATE-SUPERCLASS"
	   ;;	class names from page 3-9
	   "DEPENDEE-MIXIN" "DIRECT-SLOT-DEFINITION"
	   "EFFECTIVE-SLOT-DEFINITION"
	   "FORWARD-REFERENCED-CLASS"
	   "METAOBJECT"
	   "SLOT-DEFINITION" "SLOT-DEFINITION-NAME" "SLOT-DEFINITION-INITFORM"
	   "SLOT-DEFINITION-INITFUNCTION" "SLOT-DEFINITION-TYPE" "SLOT-DEFINITION-ALLOCATION"
	   "SLOT-DEFINITION-INITARGS" "SLOT-DEFINITION-READERS" "SLOT-DEFINITION-WRITERS"
	   "SLOT-DEFINITION-LOCATION"
	   "SPECIALIZER-DIRECT-GENERIC-FUNCTIONS" "SPECIALIZER-DIRECT-METHODS"
	   "STANDARD-DIRECT-SLOT-DEFINITION" "STANDARD-EFFECTIVE-SLOT-DEFINITION"
	   "STANDARD-READER-METHOD" "STANDARD-SLOT-DEFINITION" "STANDARD-WRITER-METHOD"
	   "STRUCTURE-DIRECT-SLOT-DEFINITION" "STRUCTURE-EFFECTIVE-SLOT-DEFINITION"
	   "STRUCTURE-INSTANCE-ACCESS" "STRUCTURE-OBJECT" "STRUCTURE-SLOT-DEFINITION"
	   
	   ;; New features adopted by X3J13 in 1989
	   "DESCRIBE-OBJECT" "FDEFINITION" "MAKE-LOAD-FORM" "MAKE-LOAD-FORM-SAVING-SLOTS" 
	   ))
  (let ((symbol (intern name "TICLOS")))
    (export symbol "TICLOS")
    (export symbol "CLOS")))

(export '(TICLOS::CLASS) "CLOS")
(use-package '("TICLOS") "USER")
(shadowing-import '(CLOS:CLASS) "USER")

;; The following have internal definitions not suitable for export.
(dolist (name '("GENERIC-FUNCTION-ARGUMENT-PRECEDENCE-ORDER" "GENERIC-FUNCTION-INITIAL-METHODS"
		"GENERIC-FUNCTION-LAMBDA-LIST" "GENERIC-FUNCTION-METHOD-CLASS"
		"GENERIC-FUNCTION-METHOD-COMBINATION" "GENERIC-FUNCTION-NAME"
		"MAKE-METHOD" ; page 2-11
		))
  (let ((symbol (intern name "CLOS")))
    (export symbol "CLOS")
    (import symbol "USER")))

;; These are ok now
(dolist (name '("FINALIZE-INHERITANCE" ; 5/7/89
		"METHOD-SLOT-NAME" ; 5/8/89
		))
  (unless (find-symbol name "TICLOS")
    (let ((symbol (intern name "TICLOS")))
      (export symbol "TICLOS")
      (export symbol "CLOS"))))

(import '(SYS:CLASSP SYS:NON-FATAL-ERROR) "TICLOS")

;; User-defined stream support that may be part of the second edition ANSI Common Lisp:
(export 'ticlos:(;; classes
		 FUNDAMENTAL-STREAM FUNDAMENTAL-INPUT-STREAM 
		 FUNDAMENTAL-OUTPUT-STREAM FUNDAMENTAL-CHARACTER-STREAM 
		 FUNDAMENTAL-BINARY-STREAM FUNDAMENTAL-CHARACTER-INPUT-STREAM 
		 FUNDAMENTAL-CHARACTER-OUTPUT-STREAM FUNDAMENTAL-BINARY-INPUT-STREAM 
		 FUNDAMENTAL-BINARY-OUTPUT-STREAM
		 ;; input functions
		 STREAM-READ-CHAR STREAM-UNREAD-CHAR STREAM-READ-CHAR-NO-HANG 
		 STREAM-PEEK-CHAR STREAM-LISTEN STREAM-READ-LINE  
		 STREAM-CLEAR-INPUT STREAM-READ-BYTE
		 ;; output functions
		 STREAM-WRITE-CHAR STREAM-LINE-COLUMN STREAM-START-LINE-P 
		 STREAM-WRITE-STRING STREAM-TERPRI STREAM-FRESH-LINE STREAM-FINISH-OUTPUT 
		 STREAM-FORCE-OUTPUT STREAM-CLEAR-OUTPUT STREAM-ADVANCE-TO-COLUMN
		 STREAM-WRITE-BYTE
		 ;; other functions
		 OPEN-STREAM-P
		 )
	"TICLOS")

;; TI extensions to CLOS
(export '( TICLOS::FLAVOR-CLASS TICLOS::HYBRID-CLASS TICL:UNDEFMETHOD
	  TICLOS::PROPERTY-MIXIN TICLOS::GET-PROPERTY TICLOS::REMOVE-PROPERTY
	  TICLOS::FASDUMP-MIXIN  TICLOS::HANDLER
	  TICLOS::CLOS-INSTANCE-P  COPY  TICLOS::STREAM-RUBOUT-HANDLER
	  TICLOS::PREPARE-GENERIC-FUNCTION
	  TICLOS::SOURCE-FILE-PATHNAME TICLOS::DEFINITION-GENERIC-PATHNAME
	  TICLOS::DELETE-CLASS
	  ) "TICLOS")


;; preview of ANSI Common Lisp
(let ((ansi '( COMPILER::LOAD-TIME-VALUE TICL:REAL SYS:FUNCTION-LAMBDA-EXPRESSION )))
  (export ansi "TICLOS")
  (export ansi "CLOS"))