;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


(defflavor slot-access-generic-function ()(standard-generic-function))

;;  1/4/89 DNG - added class definition.
(defclass-primitive slot-access-generic-function
		    (standard-generic-function) ()
  (:metaclass funcallable-standard-class))

(defmethod (slot-access-generic-function :after :add-method) (method)
  (when (null (getf (method-plist method ) 'system-defined))
    (let ((specs (method-parameter-specializers method)))
      ;;We have to assess the extent of the damage.
      )))

(defgeneric slot-unbound (class instance slot-name)
  (:documentation "Called when an unbound slot is read.  The default method signals an error."))

(defmethod slot-unbound ((class standard-class) instance slot-name)
  (eh:default-slot-unbound class instance slot-name))


(defgeneric slot-missing (class object slot-name operation &optional new-value))

;;  2/27/89 DNG - Changed specializer from STANDARD-CLASS to T to enable use for flavors also.
(defmethod slot-missing ((class t) object slot-name operation &optional new-value)
  (declare (ignore new-value))
  (eh:default-slot-missing class object slot-name operation))
;;ferror nil "slot ~s missing in instance ~s for operation ~s" slot-name object operation))

(proclaim '(inline locate-slot))
(defun locate-slot (instance slot-name )
  (when (obsolete-instance-p instance)
    (restructure-instance instance (class-of instance)))
  (sys:locate-in-instance instance slot-name))
	     

(defun slot-value (instance slot-name)
  "Returns the value contained in a particular slot of an object."
  (declare (optimize speed (safety 0)) (inline class-of))
  (slot-value-using-class (class-of instance) instance slot-name))

(defun (setf slot-value) (new-value instance slot-name)
  (declare (optimize speed (safety 0)) (inline class-of))
  (funcall #'(setf slot-value-using-class) new-value (class-of instance) instance slot-name))

(unless (generic-function-p 'slot-value-using-class)
  (fmakunbound 'slot-value-using-class))
;;  4/18/89 DNG - Added doc string.
(defgeneric slot-value-using-class (class instance slot-name)
  (:documentation "Returns the value contained in the slot named SLOT-NAME in INSTANCE. 
The first argument is (CLASS-OF INSTANCE), used for dispatching to the 
appropriate method.  This is called by SLOT-VALUE.  If you write a method for 
this, you probably need to also write corresponding methods for
\(SETF SLOT-VALUE-USING-CLASS), OPTIMIZE-SLOT-VALUE, and OPTIMIZE-SETF-SLOT-VALUE.")
  (:generic-function-class slot-access-generic-function  ))

;;  8/09/88 DNG - Fixed to pass CLASS instead of SELF to SLOT-UNBOUND.
(defmethod  slot-value-using-class ((class standard-class) instance slot-name)
  (let ((loc (locate-slot instance slot-name)))
    (if loc
	(if (location-boundp loc)
	    (contents loc)
	    (let* ((eh:cont-structure (sys:find-structure-header loc))
		   (eh:ce-type (if (instancep eh:cont-structure) :clos-instance :class-slot))
		   (eh:or-address loc)
		   (eh:cu-address loc))
	      (declare (special  eh:ce-type eh:cont-structure eh:or-address eh:cu-address))
	      (slot-unbound class instance slot-name)))
	(slot-missing class instance slot-name 'slot-value ))))

;; 12/09/88 DNG - replaced SET-SLOT-VALUE-USING-CLASS with (SETF SLOT-VALUE-USING-CLASS).

(defgeneric (setf slot-value-using-class) (new-value class instance slot-name)
  (:documentation "Stores NEW-VALUE in the slot named SLOT-NAME in INSTANCE. 
CLASS is (CLASS-OF INSTANCE).  This is called by function (SETF SLOT-VALUE).
If you write a method for this, you probably need to also write a 
corresponding method for OPTIMIZE-SETF-SLOT-VALUE.")
  (:argument-precedence-order class instance new-value slot-name))

(defmethod (setf slot-value-using-class) (new-value (class standard-class) instance slot-name)
  (let ((loc (locate-slot instance slot-name)))
    (if loc
	(setf (contents loc) new-value)
      (slot-missing class instance slot-name 'setf new-value))))

(unless (generic-function-p 'slot-boundp)
  (fmakunbound 'slot-boundp))
;;  8/16/88 DNG - Redefined to be a generic function.
(defgeneric slot-boundp (instance slot-name)
  (:documentation "Tests whether a specific slot in an instance has a value.
The default method invokes SLOT-BOUNDP-USING-CLASS."))
(defmethod slot-boundp ((instance t) slot-name)
  (slot-boundp-using-class (class-of instance) instance slot-name))

(unless (generic-function-p 'slot-boundp-using-class)
  (fmakunbound 'slot-boundp-using-class))
(defgeneric slot-boundp-using-class (class instance slot-name)
  (:documentation "Called by SLOT-BOUNDP with CLASS being (CLASS-OF INSTANCE) for dispatching to a
method appropriate to the metaclass."))

(defmethod slot-boundp-using-class ((class standard-class) instance slot-name)
  (let ((loc (locate-slot instance slot-name )))
    (if loc 
	(location-boundp loc)
	(slot-missing class instance slot-name 'slot-boundp))))

(unless (generic-function-p 'slot-makunbound)
  (fmakunbound 'slot-makunbound))
;;  12/31/88 DNG - Redefined to be a generic function.
(defgeneric slot-makunbound (instance slot-name)
  (:documentation "Cause the indicated slot in the instance to become unbound.")
  (:method ((instance t) slot-name)
	   (slot-makunbound-using-class (class-of instance) instance slot-name)))

(unless (generic-function-p 'slot-makunbound-using-class)
  (fmakunbound 'slot-makunbound-using-class ))
(defgeneric slot-makunbound-using-class (class instance slot-name))

;; 12/31/88 DNG - Fix to return the instance as the result.
(defmethod slot-makunbound-using-class ((class standard-class) instance slot-name)
  (let ((loc (locate-slot instance slot-name)))
    (if loc 
	(progn (location-makunbound loc)
	       instance)
      (slot-missing class instance slot-name 'slot-makunbound))))

(defun standard-instance-access (instance slot-name)
  (let ((loc (locally (declare (optimize speed))
	        (locate-slot instance slot-name ))))
    (if loc
	(contents loc)
      (slot-missing (class-of instance) instance slot-name 'slot-value))))

(defun (setf standard-instance-access) (value instance slot-name)
  (let ((loc (locally (declare (optimize speed))
		      (locate-slot instance slot-name ))))
    (if loc
	(setf (contents loc) value)
      (slot-missing (class-of instance) instance slot-name 'setf))))

;;  Added 8/18/88 by DNG.
;;  Modified 9/2/88 to use slot-exists-p-using-class.
(defgeneric slot-exists-p (object slot-name)
  (:documentation "Tests whether an object has a slot of the given name."))
(defmethod slot-exists-p ((object t) slot-name)
  (slot-exists-p-using-class (class-of object) object slot-name))
(defmethod slot-exists-p-using-class ((class standard-class) object slot-name)
  (not (null (locate-slot object slot-name))))

;; 5/7/89 DNG
(defmethod slot-exists-p-using-class ((class built-in-class) object slot-name)
  ;; built-in classes never have slots.
  (declare (ignore object slot-name))
  nil)

;;; TI extension to support locatives  (added 9/2/88 by DNG)
;; 12/09/88 use (LOCF SLOT-VALUE-USING-CLASS) instead of LOC-SLOT-VALUE-USING-CLASS.

(defun (locf slot-value) (object slot-name)
  (declare (optimize speed (safety 0)) (inline class-of))
  (locf (slot-value-using-class (class-of object) object slot-name)))
(defmethod (locf slot-value-using-class) ((class standard-class) object slot-name)
  (or (locate-slot object slot-name)
      (slot-missing class object slot-name 'locf)))


;;; Extensions for Flavors  (added 9/2/88 by DNG)

(defmethod slot-value-using-class ((class flavor-class) instance slot-name)
  (symeval-in-instance instance slot-name))

;; These are not inline because they are expanded by a compiler post-optimizer instead.
(proclaim '(notinline flavor-instance-access (setf flavor-instance-access)))
;;  2/27/89 DNG - Original.
(defun flavor-instance-access (instance slot-name)
  (symeval-in-instance instance slot-name))
(defun (setf flavor-instance-access) (value instance slot-name)
  (setf (symeval-in-instance instance slot-name) value))

;;  4/18/89 DNG - Add doc string.
(defgeneric optimize-slot-value (class form)
  (:documentation "Called by the compiler to optimize calls to SLOT-VALUE.
FORM looks like (SLOT-VALUE object-expression 'slot-name).  
CLASS is the class of the object.  Returns a form to be compiled in place of 
the original form."))

;; 10/20/88 DNG - Use %INSTANCE-REF for ordered instance variables.
;;  2/27/89 DNG - Use FLAVOR-INSTANCE-ACCESS instead of SYMEVAL-IN-INSTANCE .
(defmethod optimize-slot-value ((class flavor-class) form)
  ;; FORM is (SLOT-VALUE object slot-name)
  ;; Optimize to the same code that would be generated by an 
  ;;	:OUTSIDE-ACCESSIBLE-INSTANCE-VARIABLES macro.
  (let* ((name-form (third form))
	 (n (and (compiler:quotep name-form)
		 (position (second name-form) (the list (sys:flavor-unmapped-instance-variables
							    (class-description class)))
			   :test #'eq))))
    (if (and n ; ordered instance variable
	     compiler:(>= (OPT-SPEED OPTIMIZE-SWITCH) (OPT-SAFETY OPTIMIZE-SWITCH)))
	`(sys:%instance-ref ,(second form) ,(1+ n))
      `(flavor-instance-access . ,(rest form)))))

(defmethod (setf slot-value-using-class) (new-value (class flavor-class) instance slot-name)
  ;; equivalent to inline expansion of (set-in-instance instance slot-name new-value)
  (setf (contents (locate-in-instance instance slot-name)) new-value)
  new-value)

(defgeneric optimize-setf-slot-value (class form)
  (:documentation "Called by the compiler to optimize SETF on SLOT-VALUE.
FORM looks like (FUNCALL #'(SETF SLOT-VALUE) value-form object-form 'slot-name).
CLASS is the class of the object.  Returns a form to be compiled in place of 
the original form."))

;;  2/27/89 DNG - Use #'(SETF FLAVOR-INSTANCE-ACCESS) instead of 
;;		SET-IN-INSTANCE and add special handling for ordered instance variables.
(defmethod optimize-setf-slot-value ((class flavor-class) form)
  ;; FORM is (FUNCALL #'(SETF SLOT-VALUE) value object slot-name)
  (let* ((name-form (fifth form))
	 (n (and (compiler:quotep name-form)
		 (position (second name-form) (the list (sys:flavor-unmapped-instance-variables
							  (class-description class)))
			   :test #'eq))))
    (if (and n ; ordered instance variable
	     compiler:(>= (OPT-SPEED OPTIMIZE-SWITCH) (OPT-SAFETY OPTIMIZE-SWITCH)))
	(let ((value (gensym)))
	  `(let ((,value ,(third form))) ; so args are evaluated in right order
	     (sys:set-%instance-ref ,(fourth form) ,(1+ n) ,value)))
      `(funcall #'(setf flavor-instance-access) . ,(cddr form)))))

(defmethod slot-boundp-using-class ((class flavor-class) instance slot-name)
  (not (eql (sys:%p-data-type (locate-in-instance instance slot-name))
	    sys:dtp-null)))

(defmethod slot-makunbound-using-class ((class flavor-class) instance slot-name)
  (location-makunbound (locate-in-instance instance slot-name))
  instance)

(defmethod slot-exists-p-using-class ((class flavor-class) object slot-name)
  (and (ignore-errors (locate-in-instance object slot-name))
       t))

(defmethod (locf slot-value-using-class) ((class flavor-class) object slot-name)
  (locate-in-instance object slot-name))

(defmethod optimize-locf-slot-value ((class flavor-class) form)
  ;; FORM is (FUNCALL #'(LOCF SLOT-VALUE) object slot-name)
  `(locate-in-instance . ,(cddr form)))


;; This is used in file "meta-object" but needs to be loaded before compiling that file.
(defclass gfun-generic-function (standard-generic-function) ()
  (:metaclass funcallable-standard-class))

(defwhopper (gfun-generic-function :compute-main-body) ()
  (let ((body (continue-whopper))
	(arg (car argument-precedence-order)))
    (cons `(and (typep ,arg 'compiled-function)
		(setq ,arg (get-generic-function-object ,arg)))
	  body)))