;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


;; This file needs to be loaded before compiling "built-in-classes".

;;  8/10/88 DNG - Original version of this file separated from "built-in-classes".
;; 10/25/88 DNG - Changed flavor name from STANDARD-TYPE-CLASS to BUILT-IN-CLASS.

(defmethod (built-in-class  :class-direct-slots) ignore)

;; 12/01/88 DNG - Use function ADD-NAMED-CLASS-INTERNAL instead of TICLOS::ADD-NAMED-CLASS.
(defmethod (built-in-class :expand-defclass) (defclass-form &optional environment)
  (declare (ignore environment))
  (let ((class-name (progn (pop defclass-form)(pop defclass-form)))
	(includes (pop defclass-form))
	(slot-options (pop defclass-form))
	(class-options defclass-form)
	)
    (when (and class-options (or (not (= (length class-options ) 1))
				 (neq (caar class-options) :metaclass)))
      
      (error "Built-in-class cannot specify class-options"))
    (when slot-options
      (error "Built-in-class cannot specify slot-options"))
    `((add-named-class-internal ',(type-of self)
				',class-name
				',includes))))

(defmethod (built-in-class :add-named-class)
		(class-name direct-supers direct-slots class-options &optional environment)
  (declare (ignore direct-slots class-options))
  (let ((existing-class (class-named class-name t environment)))
    (if existing-class
	(let ((class-descript (class-description existing-class )))
	  (setf (class-description-direct-supers class-descript)
		direct-supers))
      (let ((class-object (create-class class-name (type-of self) direct-supers nil nil environment)))
	(setf (class-named class-name t environment) class-object)
	(setf existing-class class-object)))
    (compose-class existing-class)
    existing-class))

(defmethod (built-in-class :compose-class)()
  ;;This thing should do all of what flavor does:
  ;;Come up with the instance size...
  (let ((environment (class-description-environment class-description)))
    ;;The next two operations must be done atomically
    (when (class-description-class-precedence-list class-description)
      ;;clean up things. we might be computing this because
      ;;of an incompatible change in the supers.
      ;;May be this is not useful after all.
      (dolist (super (class-direct-supers self))
	(remove-direct-subclass (class-named super) self nil)))
    (setf (class-description-class-precedence-list class-description)
	  (loop for super in (compute-class-precedence-list self environment)
		collect (class-named super nil environment)))
    (dolist (class (class-description-class-precedence-list class-description))
      (dolist (super (class-direct-supers class))
	(add-direct-subclass (class-named super nil environment) class nil)))
    (values)))

(defmethod (built-in-class  :class-cached-handlers )()
  (getf (class-description-plist class-description ) 'cached-handlers))

(defmethod (built-in-class  :set-class-cached-handlers ) (value)
  (setf (getf (class-description-plist class-description ) 'cached-handlers) value))
(defmethod (built-in-class :add-method-on-specializer) (method)
  (let* ((gfunc (*method-generic-function method))
	 (df (getf (class-description-plist (class-description self)) 'direct-generic-functions))
	 (acons (assoc gfunc
		      df)))
    (if  acons
	 (incf (cdr acons))
	 (setf (getf (class-description-plist (class-description self)) 'direct-generic-functions)
	       (cons (cons gfunc 1) df)))))

(defmethod (built-in-class :remove-method-on-specializer) (method)
  (let* ((gfunc (*method-generic-function method))
	 (df (getf (class-description-plist (class-description self)) 'direct-generic-functions))
	 (acons (assoc gfunc
		      df)))
    (when  acons
      (if (> (cdr acons) 1)
	  (decf (cdr acons))
	  (setf (getf (class-description-plist (class-description self)) 'direct-generic-functions)
		(delete acons df :count 1 :test #'eq))))))

(defmethod (built-in-class :class-mapped-supers  ) ()
  (getf (class-description-plist class-description) 'class-mapped-supers  ))
(defmethod (built-in-class :set-class-mapped-supers  ) (value)
  (setf (getf (class-description-plist class-description) 'class-mapped-supers  ) value))

(defflavor flavor-class () (class)
    :inittable-instance-variables
  :outside-accessible-instance-variables
  (:accessor-prefix ||))