;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10. -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;  3/18/89 DNG - Moved SYS:DEFSTRUCT-1 patch to "KERNEL;STRUCTURE".

(defflavor structure-class () (class)
    :inittable-instance-variables
  :outside-accessible-instance-variables
  (:accessor-prefix ||))

(remprop 'structure-class class-property)

(defclass-primitive structure-class (class) ()
  (:metaclass class-class)
  (:documentation "All class objects for DEFSTRUCT types are members of this class."))


;; 1/3/89 DNG - Added use of STANDARD-STRUCTURE class. [ref 88-003 p. 3-11]
;;1/18/89 DNG - Name changed to STRUCTURE-OBJECT. [ref 89-003 p. 3-9]
(defclass STRUCTURE-OBJECT (t) () (:metaclass built-in-class)
	  ;;(:documentation "This class includes all classes defined by DEFSTRUCT.")
	  )

(defmethod (structure-class :add-named-class)
		(class-name direct-supers direct-slots class-options &optional environment)
  ;; 10/31/88 DNG - Redesigned using CLOS:CLASS-FOR-REDEFINITION and CREATE-CLASS .
  (declare (ignore class-options))
  (let ((existing-class (find-class class-name nil environment)))
    (if (and existing-class
	     (or (null environment)
		 (compiler:same-environment-p (class-environment existing-class) environment)))
	(reinitialize-instance
	  ;; Make sure the old definition was also a structure-class.
	  (CLOS:CLASS-FOR-REDEFINITION self existing-class :direct-superclasses direct-supers
				       :direct-slots direct-slots :environment environment)
	  :direct-superclasses direct-supers :direct-slots direct-slots)
      (let ((class-object (create-class class-name (type-of self) 
					direct-supers direct-slots nil environment)))
	(setf (class-named class-name t environment) class-object)))))

(defmethod clos:ADD-NAMED-CLASS ((class-prototype structure-class) &key name 
				 superclass-names slot-specifications direct-options environment)
  (send class-prototype :add-named-class name superclass-names slot-specifications direct-options environment))

;;  4/18/89 DNG - Removed obsolete (:method standard-type-class :class-direct-slots);
;;		this is handled in file "standard-type-class".
 
(defmethod (structure-class :class-direct-subclasses) ()
  nil) ; This doesn't look right!  -- DNG 3/18/89

(defmethod (structure-class :compose-class)()
  (let* ((environment (class-description-environment class-description))
	 (super (class-named (car (class-direct-supers self)) nil environment)))
    (setf (class-description-class-precedence-list class-description)
	  (cons self (or (internal-class-precedence-list super)
			 (progn (compose-class super)
				(internal-class-precedence-list super))))))
  (values))

(defmethod (structure-class  :class-cached-handlers )()
  (getf (class-description-plist class-description ) 'cached-handlers))

(defmethod (structure-class  :set-class-cached-handlers ) (value)
  (setf (getf (class-description-plist class-description ) 'cached-handlers) value))

(defmethod (structure-class :class-direct-slots) ()
  (class-description-direct-slots class-description))

;; 1/16/89 DNG - Added for use by DOC:DOCUMENT-CLASS .
(defmethod (structure-class :class-all-slots) ()
  (class-description-direct-slots class-description))

system:
(defun make-class-definition ()
  (using-defstruct-special-variables)
  (when (eq type 'common-lisp-structure)
    (push `(and (fboundp 'add-structure-class) ; this line needed for backward comptibility of XLD files
		(add-structure-class 
		  ',name
		  ,(and include `(quote ,include))
		  ',(mapcar #'(lambda (x)
				(let ((slot-name (car x))
				      (initform (defstruct-slot-description-init-code (cdr x)))
				      (slot-type (defstruct-slot-description-type (cdr x))))
				  (if (emptyp slot-type)
				      (if (emptyp initform)
					  slot-name
					(list slot-name initform))
				    (list slot-name
					  initform
					  (if (emptyp slot-type) 't slot-type)))))
			    slot-alist)))
	  returns)))

;; added 9/23/88 by DNG to minimize the amount of data needed in an XLD file.
;; 12/01/88 DNG - Use function ADD-NAMED-CLASS-INTERNAL instead of TICLOS::ADD-NAMED-CLASS.
;;  1/03/89 DNG - Use STANDARD-STRUCTURE class as default superclass instead of T.
;;  1/18/89 DNG - Change name from STANDARD-STRUCTURE to STRUCTURE-OBJECT.
system:
(defun add-structure-class (class-name direct-supers slots)
  (when (fboundp 'ticlos:add-named-class-internal)
    (ticlos:add-named-class-internal
      'ticlos:structure-class
      class-name
      (or direct-supers '(ticlos:structure-object))
      (mapcar #'(lambda (x) 
		  (if (atom x)
		      (ticlos:internal-make-slot-description
			:name x :initform nil :allocation :instance :type 't)
		    ;; x is (slot-name initform [ slot-type ] )
		    (ticlos:internal-make-slot-description
		      :name (first x)
		      :initform (and (cdr x) (not (emptyp (second x))) (list (second x)))
		      :allocation :instance
		      :type (if (null (cddr x)) 't (third x))
		      )))
	      slots)
      ;; Note: no need to include :DOCUMENTATION here because the DOCUMENTATION 
      ;; method for STRUCTURE-CLASS will know where to find it.
      nil
      )))


;; added 8/24/88 by DNG
(defmethod (structure-class :add-method-on-specializer) ignore)
(defmethod (structure-class :remove-method-on-specializer) ignore)

;;  8/24/88 DNG - Original.
;; 11/10/88 DNG - Changed from a macro to a function so it can be used by CLASS-NAMED.
(defun make-class-for-structure (structure-name)
  ;; Given the name of a named-structure type, create a class definition for it.
  sys:
  (let* ((name ticlos:structure-name)
	 (dd (get name 'defstruct-description))
	 (returns '())
	 (type 'common-lisp-structure)
	 (include (defstruct-description-include dd))
	 (slot-alist (defstruct-description-slot-alist dd)))
    (declare (list returns))
    (using-defstruct-special-variables)
    (when (null dd)
      (error "~S is not the name of a named structure type." name))
    (make-class-definition)
    (mapc #'eval (nreverse returns))
    name))
  
;; pre-defined classes from page 1-17 of chapter 1 of 88-002R which are 
;; implemented here as structures:
(make-class-for-structure 'hash-table)
(make-class-for-structure 'package)
(make-class-for-structure 'random-state)
(make-class-for-structure 'readtable)

;;  5/7/89 DNG - Patch up slot description class to look like a subclass of METAOBJECT.
(let* ((class (find-class 'slot-description))
       (cd (class-description class))
       (mo (find-class 'metaobject))
       (cpl (class-description-class-precedence-list cd)))
  (unless (member mo cpl :test #'eq)
    (setf (class-description-class-precedence-list cd)
	  (list* class mo (rest cpl)))
    (add-direct-subclass mo class)))
