;;; -*- Mode:Lisp; Package:XLIB; Base:10; Lowercase:YES; Patch-file:T; Syntax:Common-Lisp -*-

;; This file integrates CLX with CLOS

;;;
;;;			 TEXAS INSTRUMENTS INCORPORATED
;;;				  P.O. BOX 2909
;;;			       AUSTIN, TEXAS 78769
;;;
;;; Copyright (C) 1987 Texas Instruments Incorporated.
;;;
;;; Permission is granted to any individual or institution to use, copy, modify,
;;; and distribute this software, provided that this complete copyright and
;;; permission notice is maintained, intact, in all copies and supporting
;;; documentation.
;;;
;;; Texas Instruments Incorporated provides this software "as is" without
;;; express or implied warranty.
;;;

(in-package 'xlib :use '(lisp clos))

;; Temporary crock: undo the previous structure definition to avoid problems
;; with bogus compiler optimizations.
#+explorer
(EVAL-WHEN (COMPILE load eval)
  (remprop  'drawable 'sys:defstruct-description)
  (remprop  'window   'sys:defstruct-description)
  (remprop  'pixmap   'sys:defstruct-description)
  )

;;
;; Drawables
;;

;; Allow change in metaclass (structure-class to standard-class)
(SETF (find-class 'drawable t) nil)

;; Allow these to be defined as generic functions
(fmakunbound 'drawable-id)
(fmakunbound 'drawable-display)
(fmakunbound 'drawable-plist)

(defclass drawable ()
  ((id      :type     resource-id
	    :initform 0
	    :accessor drawable-id
	    :initarg  :id)
   
   (display :type     (or null display)
	    :initform nil
	    :accessor drawable-display
	    :initarg  :display)
   
   (plist   :type     list
	    :initform nil
	    :accessor drawable-plist
	    :initarg  :plist))		; Extension hook

  (:documentation "The class of CLX drawable objects."))


(defun make-drawable (&rest initargs)
  (apply #'make-instance 'drawable initargs))

(defun drawable-p (object)
  (typep object 'drawable))

;;
;; Windows
;;

;; Allow change in metaclass (structure-class to standard-class)
(SETF (find-class 'window t) nil)

;; Allow these to be defined as generic functions
(fmakunbound 'window-id)
(fmakunbound 'window-display)
(fmakunbound 'window-plist)

(defclass window (drawable)
  ((id      :type     resource-id
	    :initform 0
	    :accessor window-id
	    :initarg  :id)
   
   (display :type     (or null display)
	    :initform nil
	    :accessor window-display
	    :initarg  :display)
   
   (plist   :type     list
	    :initform nil
	    :accessor window-plist
	    :initarg  :plist))		; Extension hook

  (:documentation "The class of CLX window objects."))


(defun make-window (&rest initargs)
  (apply #'make-instance 'window initargs))

(defun window-p (object)
  (typep object 'window))


;;
;; Pixmaps
;;

;; Allow change in metaclass (structure-class to standard-class)
(SETF (find-class 'pixmap t) nil)

;; Allow these to be defined as generic functions
(fmakunbound 'pixmap-id)
(fmakunbound 'pixmap-display)
(fmakunbound 'pixmap-plist)

(defclass pixmap (drawable)
  ((id      :type     resource-id
	    :initform 0
	    :accessor pixmap-id
	    :initarg  :id)
   
   (display :type     (or null display)
	    :initform nil
	    :accessor pixmap-display
	    :initarg  :display)
   
   (plist   :type     list
	    :initform nil
	    :accessor pixmap-plist
	    :initarg  :plist))		; Extension hook

  (:documentation "The class of CLX pixmap objects."))


(defun make-pixmap (&rest initargs)
  (apply #'make-instance 'pixmap initargs))

(defun pixmap-p (object)
  (typep object 'pixmap))
