;;;  -*- Mode:Common-Lisp; Package:color; Fonts:(MEDFNT HL12B HL12BI HL10B CPTFONTB); Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987, 1988 Texas Instruments Incorporated. All rights reserved.



(DEFCOMMAND (color-editor :DISPLAY-COLOR-MAP)()
  '(:DESCRIPTION "Activate the Color Map Display with the current color map of the CME."
		 :NAMES ("Display Color Map")
		 :KEYS (#\control-D))
  
  (LET (index local-color-map )
    (when (and *color-map-selector* (<= (send *color-map-selector* :width) (send self :x-offset)))
      (send *color-map-selector* :set-superior (send self :superior))
      (send *color-map-selector* :set-position 
	    (- (send self :x-offset) (send *color-map-selector* :width))
	    (send self :y-offset)))
    (multiple-value-setq (index local-color-map)
      (color:show-map current-color-map nil)) 
    (when (null index)
      (setq index (if editing-background-flag
		      col-ed-background-color
		      col-ed-foreground-color)))
    (send self :set-color-map local-color-map)
    (setf current-color-map local-color-map)
    (send self :select)
    (if editing-background-flag                              
	(send self :update-from-above-for-background index)
	(send self :update-from-above index))
    ))
	  
(defcommand (color-editor :SCREEN-OBJECT) ()
  '(:description "Left click on a color on the screen to be edited."
		 :names "Screen Object"
		 :keys (#\ctrl-s)
		 :menus *color-editor-instance*
		 )
  (let (color x y wind wind-map r g b char offset-x offset-y)
    (setq char (send w:mouse-blinker :character))
    (multiple-value-setq (offset-x offset-y) (send w:mouse-blinker :offsets))
    (send w:mouse-blinker :set-character w:mouse-glyph-hollow-circle-pointer)
    (send w:mouse-blinker :set-offsets 8 8)	   ;Find the REAL value for these offsets...
    (w:with-mouse-grabbed
      (setq w:who-line-mouse-grabbed-documentation "Pick a color from the screen")
      (process-wait "Pick a Color" #'(lambda () (not (= (w:mouse-buttons) 1))))
      (process-wait "Pick a Color" #'(lambda () (= (w:mouse-buttons) 1)))
      (setq x sys:mouse-x
	    y sys:mouse-y))
    ;;Restore the original mouse-blinker character
    (send w:mouse-blinker :set-character char)
    (send w:mouse-blinker :set-offsets offset-x offset-y)
    (setq wind (w:window-under-mouse))
    (setq wind-map (send wind :color-map))
    (multiple-value-setq (x y) (mouse-offsets wind x y))
    (setq color (aref (send wind :screen-array) y x)) 
    (unless (eq wind-map current-color-map)
      (setf current-color-map wind-map)
      (setf original-color-map (tv:copy-color-map wind-map))
      (when *color-map-selector*
	(send  *color-map-selector* :set-color-map current-color-map)))
    (multiple-value-setq (r g b)
      (tv:read-color-map wind-map color))
    (if editing-background-flag                              
	(send self :update-from-above-for-background color)
	(send self :update-from-above color))
    ))

(defun mouse-offsets (window x y)
  (let ((ms w:mouse-sheet)
	(xx x)
	(yy y))
    (loop
      (when (eq ms window) (return (values xx yy)))
      (setq xx (- xx (send window :x-offset))
	    yy (- yy (send window :y-offset))
	    window (send window :superior)))))

(DEFUN display-help (help-msg)
  (let (char)
    (w:expose-window-near *help-window* `(:point ,(send *color-editor-instance* :x-offset)
						 ,(send *color-editor-instance* :y-offset)))
    (send *help-window* :select)
    (SEND *help-window* :string-out help-msg)
    (TERPRI *help-window*)
    (FORMAT *help-window* "3~& ~A*" tv:*REMOVE-TYPEOUT-STANDARD-MESSAGE*)
    (SETQ char (READ-CHAR *help-window*))
    (send *help-window* :clear-screen)
    (send *help-window* :deactivate)))

(defcommand (color-editor :HELP) ()
  '(:description "Display documentation on how to use the Color Map Editor."
		 :names ("Help")
		 :keys (#\help))
  (display-help *color-editor-help*))


(defcommand (color-editor :UNDO) ()
  '(:description "Undo changes made to the current color since it was selected."
		 :names ("Undo")
		 :keys (#\undo))
  (if editing-background-flag            
      (send self :undo-from-above-for-background col-ed-background-color)
      (send self :undo-from-above col-ed-foreground-color)))

(defcommand (color-editor :reset-map) ()
  `(:description "Undo all changes made to the current color map of the CME."
		 :names "Reset Map"
		 :keys (#\ctrl-r)
		 :documentation "Undo all changes made to the current color map of the CME.")
    (setf (tv:color-map-table current-color-map) (copy (tv:color-map-table original-color-map)))
    (send self :set-color-map current-color-map)
    (when *color-map-selector*
      (send  *color-map-selector* :set-color-map current-color-map))
    (send self :select))

(defcommand (color-editor :install-this-map) ()  
  '(:description "Left click on a window to set its color map to the current color map of the CME."
		 :names "Install Map"
		 :keys (#\ctrl-i)
		 :documentation "Set the color map of a window to the current color map of the CME."
		 )
  (let (wind char offset-x offset-y)
    (setq char (send w:mouse-blinker :character))
    (multiple-value-setq (offset-x offset-y) (send w:mouse-blinker :offsets))
    (send w:mouse-blinker :set-character w:mouse-glyph-hollow-circle-pointer)
    (send w:mouse-blinker :set-offsets 8 8)	   ;Find the REAL value for these offsets...
    (w:with-mouse-grabbed
      (setq w:who-line-mouse-grabbed-documentation "Pick a window from the screen.")
      (process-wait "Pick a Window" #'(lambda () (not (= (w:mouse-buttons) 1))))
      (process-wait "Pick a Window" #'(lambda () (= (w:mouse-buttons) 1)))
      (send (setq wind (w:window-under-mouse)) :set-color-map current-color-map)
      )
    ;;Keep track of the window we just put this color map into.
    (setf wind-under-mouse wind)
    ;;Restore the original mouse-blinker character
    (send w:mouse-blinker :set-character char)
    (send w:mouse-blinker :set-offsets offset-x offset-y)
    (send self :select)
    ))

(defcommand (color-editor :toggle-edit-contrast) ()           
  '(:description "Switch between editing the inner and outer rectangle colors."
		 :names ("Toggle Edit Contrast")
		 :keys (#\control-t))
    (cond ((eq toggle-flag :foreground)
	   (send self :update-from-above col-ed-foreground-color)
	   (setq toggle-flag :background))
	  (t
	   (send self :update-from-above-for-background col-ed-background-color)
	   (setq toggle-flag :foreground))))

(defcommand (color-editor :move-this-window) ()
  '(:description "Reposition the Color Map Editor by moving it with the mouse."
		 :names ("Move This Window")
		 :keys (#\c-m))
  (tv:mouse-set-window-position self))

(defcommand (color-editor :exit) ()
  '(:description "Exit the CME and kill the command loop."
		 :names ("Exit")
		 :keys (#\End))
  (send self :bury)
  (send self :quit))

(BUILD-COMMAND-TABLE 'color-editor-command-table 'color:COLOR-EDITOR
  '(:display-color-map
     :undo
     :screen-object
     :reset-map
     :toggle-edit-contrast
     :install-this-map
     :move-this-window
     :help
     :exit
     ))

(BUILD-MENU 'color-editor-menu 'color:COLOR-EDITOR
  :item-list-order
  '(:display-color-map
     :undo     
     :screen-object
     :reset-map
     :toggle-edit-contrast
     :install-this-map
     :move-this-window
     :help
     ) 
  :default-item-options '(:font fonts:cptfontb))


