;;;  -*- Mode:Common-Lisp; Package:color; Fonts:(MEDFNT HL12B HL12BI HL10B CPTFONTB); Base:10 -*-
;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987, 1988 Texas Instruments Incorporated. All rights reserved.



(DEFFLAVOR rgb-converter ()
	   ()
  (:settable-instance-variables)
  (:documentation " It converts TO and FROM rgb"))

(DEFMETHOD (rgb-converter :convert-to-rgb) (r g b)
  (VALUES (round r) (round g) (round b)))

(DEFMETHOD (rgb-converter :convert-from-rgb) (r g b)
  (VALUES (round r) (round g) (round b)))

(DEFFLAVOR hsv-converter ()
	   ()
  (:settable-instance-variables)
  (:documentation " It converts TO and FROM hsv"))

(defmethod (hsv-converter :convert-to-rgb) (hue saturation value)
  (let* (rred ggreen bblue)
    (if (= saturation 0)
	(setf rred value
	      ggreen value
	      bblue value) 
	(let* ((h (mod (/ hue 60) 6))
	       (h-integer (floor h))
	       (fract (- h h-integer))
	       (p (* (- 1.0 saturation) value))
	       (q (* (- 1.0 (* saturation fract)) value))
	       (tt (* (- 1.0 (* saturation (- 1.0 fract))) value)))
	  (multiple-value-setq (rred ggreen bblue)
	    (case h-integer
	      (0 (values value tt p))
	      (1 (values q value p))
	      (2 (values p value tt))
	      (3 (values p q value))
	      (4 (values tt p value))
	      (5 (values value p q))))))
    (values (round (* 255 rred)) (round (* 255 ggreen)) (round (* 255 bblue)))))


(DEFMETHOD  (hsv-converter :convert-from-rgb) (rr gg bb)  
  "2Convert R G B (range 0 to 255) to H S V* 2(range 0-360 for H, 0.0 to 1.0 for s,v)*"
  (LET* ((rred (/ rr 255))		  
	 (ggreen (/ gg 255))
	 (bblue (/ bb 255))
	 (max-rgb (MAX rred ggreen bblue))     
	 (min-rgb (MIN rred ggreen bblue))
	 (delta (- max-rgb min-rgb))
	 (value max-rgb)
	 (saturation (if (zerop max-rgb) 0 (/ delta max-rgb))) 
	 (hue 360))
    (UNLESS (= saturation 0)
      (LET* ((rc (/ (- max-rgb rred) delta ))
	     (gc (/ (- max-rgb ggreen) delta ))
	     (bc (/ (- max-rgb bblue) delta )))
	(setq hue
	      (mod (* 60
		      (cond ((= max-rgb rred) (- bc gc))
			    ((= max-rgb ggreen) (- (+ 2 rc) bc))
			    ((= max-rgb bblue) (- (+ 4 gc) rc))
			    (t (error "max-rgb not equal to red, green or blue"))))
		   360))))
    (values (ROUND hue) saturation value)))
 




