;;; -*- Mode:Common-Lisp; Package:USER; Base:10 -*-
;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987, 1988 Texas Instruments Incorporated. All rights reserved.



(DEFUN TV:WRITE-COLOR-MAP-FILE (MAP)
  "Save this color map to a file which defaults to LM:COLOR-MAPS;name-of-map.XLD.
The user is prompted for a name; that symbol will be set to the color map, and the 
name slot of the map will be set to the stringified symbol. Subsequent loading of 
the file (via the LOAD command) sets the name-of-map symbol back to this value." 
  (UNLESS MAP
    (BEEP) 
    (FORMAT *QUERY-IO* "There is no current color map.~%")
    (RETURN-FROM TV:WRITE-COLOR-MAP-FILE))
  (CHECK-TYPE MAP TV:COLOR-MAP)
  (LET* ((MAP-NAME (PROMPT-AND-READ :READ "What will be the name of your color map? ")) 
	 (FILENAME (FS:MAKE-PATHNAME :HOST "LM"
				     :DIRECTORY "COLOR-MAPS"
				     :NAME (SYMBOL-NAME MAP-NAME))))
    (SETF (SYMBOL-VALUE MAP-NAME) MAP)
    (COMPILER:FASD-COLOR-MAP MAP-NAME FILENAME)))

(DEFUN COMPILER:FASD-COLOR-MAP (MAP-SYMBOL FILENAME)
  (LET ((TEM (SYMBOL-VALUE MAP-SYMBOL))
	TRUENAME)
    (UNWIND-PROTECT
	(PROGN
	  (SETF (TV:COLOR-MAP-NAME TEM) (STRING MAP-SYMBOL))
	  (SETF (SYMBOL-VALUE MAP-SYMBOL)
		(DO ((M MAP-SYMBOL (SYMBOL-VALUE M)))
		    ((OR (NULL M) (NOT (SYMBOLP M)) (NOT (BOUNDP M)))
		     M)))
	  (SETQ TRUENAME
		 (COMPILER:DUMP-FORMS-TO-FILE FILENAME
					      `((SETF ,MAP-SYMBOL ,TEM)
						(PUSHNEW ',MAP-SYMBOL tv:*COLOR-MAPS*))
					      `(:COLOR-MAP-SYMBOL ,MAP-SYMBOL :PACKAGE :USER))))
      (SETF (SYMBOL-VALUE MAP-SYMBOL) TEM)
      (PUSHNEW MAP-SYMBOL TV:*COLOR-MAPS*))
    TRUENAME))

