;;; -*- Mode:Common-Lisp; Package:SYSTEM-INTERNALS; Base:8; Fonts:(CPTFONT HL12B HL12BI) -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; 04-24-89 DAB Added all function, variables, flavors and instances that are documented to export list.



 ;;; Ultra-simple stepper for lisp-machine.
 ;;; Wins with multiple values
 ;;; Does not attempt to win with editor top level
 ;;; Compile with QC-FILE.


;NOTES:
; The way it decides whether it needs to reprint the form when showing
; you the values is pretty kludgey right now.  Can it check the cursorpos
; or ask itself whether it typed anything or something?
;
; Would like to be able to evaluate and/or substitute in atoms and forms
; without having to break first.
;
; Would like to be able to type A and have it stop after evaluating the
; args, before calling the function.
;
; Raid registers
;
; Hook up to DDT?
; 
; If an error happens, user should be able to throw back into the stepper.

(defvar *step-level* nil "Depth within STEP-EVAL, minus one, within call to STEP")
(defvar *step-array* nil "Holds forms to evaluate, indexed by *STEP-LEVEL* value.")
(defvar *step-apply-p-array* nil
  "Holds the APPLY-P flag for each level, indexed by *STEP-LEVEL* value.")
(defvar *step-max* nil "Do not tell user about evaluations with *STEP-LEVEL* deeper than this.")
(defvar *step-form* nil "Form to be or just evaluated, in STEP command loop")
(defvar *step-value* nil "First value just computed.  May be changed in a breakpoint.")
(defvar *step-values* nil "List of values just computed.  May be changed in a breakpoint.")

(proclaim '(SPECIAL STEP-FORM STEP-VALUE STEP-VALUES))
(FORWARD-VALUE-CELL 'STEP-FORM '*STEP-FORM*)
(FORWARD-VALUE-CELL 'STEP-VALUE '*STEP-VALUE*)
(FORWARD-VALUE-CELL 'STEP-VALUES '*STEP-VALUES*)

(defvar	 *step-auto* nil)	       ;if non NIL, simulate cntrl-n at step-cmdr.
			       ; normal printout produced unless NO-PRINT.
			       ;  User's program can turn on auto mode by
			       ;  (si:step-auto-on &optional (mode 'no-print))
			       ; and (si:step-auto-off) to reable stepping.

(defvar *ucl-stepper* nil 1"Holds the currently executing instance of stepper."*)

(DEFVAR *step-function-not-recursive* t)   ;1Is this the first level call to this function?*


;1;;NEW SUGGESTIONS!!!*
(sys:declare-suggestions-for 'si:step-eval :around (sys:make-default-around-form
						si:step-eval single-stepper-top-menu
						single-stepper-bottom-menu
						single-stepper-commands-menu))

(DEFFLAVOR ucl-step
	   ((apply-p)
	    (*step-max*)
	    (*step-level*)
	    ;;;(*step-auto*)
	    (form)
	    (values)
	    (*step-array*)
	    (*step-apply-p-array*)
	    (*step-form*)
	    (*step-value*)
	    (*step-values*)
            name                                   ;NEW
	    value-to-return)
	   (UCL:BASIC-COMMAND-LOOP)
  :gettable-instance-variables                     ;NEW
  :SETTABLE-INSTANCE-VARIABLES
  (:SPECIAL-INSTANCE-VARIABLES
    apply-p *step-max* *step-level* form values *step-array* *step-apply-p-array* 
    *step-form* *step-value* *step-values*)
  (:DEFAULT-INIT-PLIST
    :active-command-tables '(step-cmd-table)
    :all-command-tables '(step-cmd-table)
    :menu-panes  nil
    :typein-handler :handle-typein-input
    :name        "Step"                            ;NEW
    :basic-help '(documentation-cmd)))

(DEFRESOURCE ucl-step ()
  :constructor (MAKE-INSTANCE 'ucl-step)
  :initial-copies 0)

;;Redefine this function to call the ucl command loop.
;print whatever is necessary, read a command, set special variables
;and return how to proceed:  *eval (just *eval), evalhook (recurse), more options later.
;If calling for *eval, *step-values* is nil, otherwise calling for return.
(DEFUN STEP-CMDR (FORM VALUES PRINT-FORM-P &OPTIONAL APPLY-P)
  (DECLARE (SPECIAL APPLY-P *STEP-MAX* *STEP-LEVEL*  FORM VALUES))
  (BLOCK ()
    (IF *STEP-AUTO*
       (IF (EQUAL *STEP-AUTO* :NO-PRINT);1 Change no-print to keyword to avoid package conflict*
	  (PROGN
	    (SETQ *STEP-MAX* (1+ *STEP-LEVEL*))
	    (RETURN 'EVALHOOK));1 Add else clause to handle *step-auto* being non-nil but not*
	  (SEND TV:SELECTED-WINDOW :FORCE-KBD-INPUT #\c-N)));1 eq to :no-print*
    (AND PRINT-FORM-P (STEP-PRINT-FORM FORM *STEP-LEVEL* APPLY-P))
    (PRINT-VALUES)
    ;;The following code has been reworked to make the stepper run in its own temporary stack group, 
    ;;so that its UCL instance variables are not bound when user evaluates a form.  This is particularly
    ;;important if the user is stepping through a UCL application, where he expects to find
    ;;*his* instance vars bound.  The three tricks below are a mechanism of communication between
    ;;two stack groups--the current one in which this function is called, and the temporary stack group
    ;;which is used (in UCL:IN-OWN-STACK-GROUP) to execute the stepper command loop.  In the current
    ;;sg, we have bound a group of special vars.  Trick#1 gives the stepper sg access to these 
    ;;variables' current values.  Trick#2 makes sure the current sg gets any modifications the stepper 
    ;;sg (stepper commands) make to the vars.
    (USING-RESOURCE (*UCL-STEPPER* UCL-STEP)
		    ;;Trick #1
       (DOLIST (IVAR
	 '(APPLY-P *STEP-MAX* *STEP-LEVEL*  FORM VALUES *STEP-ARRAY* *STEP-APPLY-P-ARRAY*
	   *STEP-FORM* *STEP-VALUE* *STEP-VALUES*))
	 (SET-IN-INSTANCE *UCL-STEPPER* IVAR (SYMBOL-VALUE IVAR)))
       ;;Trick#3: UCL:IN-OWN-STACK-GROUP can not return any values, so we
       ;;return a value through an instance variable.
       (UCL::IN-OWN-STACK-GROUP (*UCL-STEPPER*) ()
	  (SEND *UCL-STEPPER* :SET-VALUE-TO-RETURN (SEND *UCL-STEPPER* :COMMAND-LOOP)))
       ;;Trick#2: some of these vars may have been modified during the command loop.
       ;;set them to the values of the corresponding instance variables.
       (DOLIST (SPECIAL-VAR
	 '(APPLY-P *STEP-MAX* *STEP-LEVEL* FORM VALUES *STEP-ARRAY* *STEP-APPLY-P-ARRAY*
	   *STEP-FORM* *STEP-VALUE* *STEP-VALUES*))
	 (SET SPECIAL-VAR (SYMEVAL-IN-INSTANCE *UCL-STEPPER* SPECIAL-VAR)))
       (LET ((RET-VAL (SEND *UCL-STEPPER* :VALUE-TO-RETURN)));1 If return value is 'exiting due to quit-cmd, then throw to*
	 (IF (EQ RET-VAL 'EXITING);1 top level of step, otherwise continue*
              (THROW 'exit-stepper ())
	    (RETURN RET-VAL))))))       

(DEFMETHOD (UCL-STEP :AROUND :FETCH-AND-EXECUTE) (CONT MT IGNORE)
1  "Intercept a ucl method to break out of the command loop for some of the commands.
   This is necessary due to the nature of the implementation of the stepper.
   It sets a var in eval as a hook to regain control of the evaluation."*
  (CATCH 'UCL::COMMAND-ABORT
    (LET ((TV:KBD-INTERCEPTED-CHARACTERS;Let this code handle the break key.
	   (REMOVE (ASSOC #\BREAK TV:KBD-INTERCEPTED-CHARACTERS :TEST #'EQ)
		   (THE LIST TV:KBD-INTERCEPTED-CHARACTERS) :TEST #'EQUAL)))
      (FUNCALL-WITH-MAPPING-TABLE CONT MT :FETCH-AND-EXECUTE)
      (WHEN (AND (NOT UCL::PREEMPTING?) UCL::COMMAND-ENTRY (TYPEP UCL::COMMAND-ENTRY 'UCL::COMMAND)
						      ;;Several stepper commands are given this property in their DEFCOMMANDs.
	  (SEND UCL::COMMAND-ENTRY :GET 'EXIT-LOOP?))
	(THROW 'UCL::EXIT-COMMAND-LOOP
	       (CAR
		/)))))) 

(DEFMETHOD (UCL-STEP :AROUND :HANDLE-TYPEIN-INPUT) (CONT MT IGNORE &OPTIONAL (UNTYI-FIRST-CHAR? T))
1  "Do the lisp prompt and afterwards output the stepper forms for Lisp typein."*
  (DECLARE (SPECIAL *STEP-ARRAY* *STEP-LEVEL* *STEP-APPLY-P-ARRAY*))
  (FORMAT T "    Eval: ")
  (UNWIND-PROTECT (FUNCALL-WITH-MAPPING-TABLE
		   CONT
		   MT
		   :HANDLE-TYPEIN-INPUT
		   UNTYI-FIRST-CHAR?)
    (DO ((I (MAX 0 (- *STEP-LEVEL* 10)) (1+ I)))
	((> I *STEP-LEVEL*)
	 NIL)
      (STEP-PRINT-FORM (AREF *STEP-ARRAY* I) I (AREF *STEP-APPLY-P-ARRAY* I)))
    (PRINT-VALUES))) 

(DEFMETHOD (ucl-step :handle-prompt) ()
  1"Redefine UCL method to do nothing so that a prompt is not printed."* nil)

(DEFMETHOD (ucl-step :designate-io-streams) ()
  "Redefine a ucl method to set up the correct bindings."
  (SETQ *standard-input* *terminal-io*
        *standard-output* *terminal-io*))

(DEFCOMMAND next-thing-cmd ()
  '(:DESCRIPTION 1"Proceed to next thing evaled."*
    :NAMES ("Next thing")
    :KEYS (#\CONTROL-N)
    :PROPERTY-LIST (exit-loop? T))
  (DECLARE (SPECIAL *step-max* *step-level*))
  (SETQ *step-max* (1+ *step-level*))
  'evalhook)

(DEFCOMMAND next-this-level-cmd () 
  '(:DESCRIPTION 1"Proceed to next thing evaled at same level."*
    :NAMES ("Next this level") 
    :KEYS (#\SPACE)
    :PROPERTY-LIST (exit-loop? T :CANT-PREEMPT T)) ;;Keep space from preempting type-in.
  (DECLARE (SPECIAL *step-max* *step-level*))
  (SETQ *step-max* *step-level*)
  '*eval)    ;1eval1 in rel 2*

(DEFCOMMAND EVAL-ARGS-CMD ()
   '(:DESCRIPTION1 "Eval the args without stepping; stop before applying the function."*
     :NAMES ("Eval args")
     :KEYS (#\c-A)
     :PROPERTY-LIST (EXIT-LOOP? T))
   (DECLARE (SPECIAL APPLY-P *STEP-MAX* *STEP-LEVEL*))
   (IF (NOT APPLY-P)
       (PROGN (SETQ *STEP-MAX* (1+ *STEP-LEVEL*))
              'APPLYHOOK)
       (PROGN (BEEP)
              (THROW 'UCL::COMMAND-ABORT T)))) 

(DEFCOMMAND proceed-up-cmd ()
  '(:DESCRIPTION 1"Proceed to first thing up one level."*
    :NAMES ("Proceed up")
    :KEYS (#\CONTROL-u)
    :PROPERTY-LIST (exit-loop? T))
  (DECLARE (SPECIAL *step-max* *step-level*))
  (SETQ *step-max* (max 0 (1- *step-level*)))
  '*eval)   ;1eval1 in rel 2*

(DEFCOMMAND execute-no-stepping-cmd ()
  '(:DESCRIPTION 1"Continue executing without further stepping."*
    :NAMES ("Exit Stepper")
    :KEYS (#\END #\CONTROL-x)
    :PROPERTY-LIST (exit-loop? T))
  (DECLARE (SPECIAL *step-max*))
  (SETQ *step-max* -1)
  '*eval)   ;1eval1 in rel 2*

(DEFCOMMAND QUIT-CMD ()
   '(:DESCRIPTION1 "Throw out of Stepper without further execution of program."* :NAMES
     ("Abort Program") :KEYS (#\s-END))
   (THROW  'UCL::EXIT-COMMAND-LOOP
	  'EXITING))                           ;1Throw 'exiting so we can check at next level up*

(DEFCOMMAND ESCAPE-TO-EDITOR-CMD ()
   '(:DESCRIPTION1 "Escape to editor."* :NAMES ("Escape to editor") :KEYS (#\c-E))
   (DECLARE (SPECIAL *STEP-LEVEL* *STEP-APPLY-P-ARRAY* *STEP-ARRAY* *UCL-STEPPER*)) (ED)
   (SEND *STANDARD-OUTPUT* :CLEAR-SCREEN)
   (DO ((I (MAX 0 (- *STEP-LEVEL* 10)) (1+ I)))
       ((> I *STEP-LEVEL*)
	NIL)
     (STEP-PRINT-FORM (AREF *STEP-ARRAY* I) I (AREF *STEP-APPLY-P-ARRAY* I)))
   (PRINT-VALUES)) 

(DEFCOMMAND retype-cmd ()
  '(:DESCRIPTION 1"Retype current form in full."*
    :NAMES ("Retype")
    :KEYS (#\C-T))
  (DECLARE (SPECIAL values form *ucl-stepper*))  
  (IF (NULL values) (PRINT form)
    (DOLIST (value values) (PRINT value))))

(DEFCOMMAND grind-cmd ()
  '(:DESCRIPTION 1"Grind current form."*
    :NAMES ("Grind")
    :KEYS (#\C-G))
  (DECLARE (SPECIAL values form *ucl-stepper*))
  (IF (NULL values) (GRIND-TOP-LEVEL form)
    (DOLIST (value values) (GRIND-TOP-LEVEL value))))

(DEFCOMMAND BREAKPOINT-CMD ()
   '(:DESCRIPTION1 "Enter breakpoint with STEP-FORM, STEP-VALUE and STEP-VALUES bound."*
     :NAMES ("Breakpoint")
     :KEYS (#\BREAK #\c-B))
   (DECLARE (SPECIAL *STEP-FORM* *STEP-ARRAY* *STEP-LEVEL* APPLY-P *STEP-APPLY-P-ARRAY* *UCL-STEPPER*))
   (UNWIND-PROTECT (PROGN
		    (BREAK "step - with *STEP-FORM*, *STEP-VALUE* and *STEP-VALUES* bound." T)
		    (SETF (AREF *STEP-ARRAY* *STEP-LEVEL*) *STEP-FORM*)
		    (SETF (AREF *STEP-APPLY-P-ARRAY* *STEP-LEVEL*) APPLY-P))
     (DO ((I (MAX 0 (- *STEP-LEVEL* 10)) (1+ I)))
	 ((> I *STEP-LEVEL*)
	  NIL)
       (STEP-PRINT-FORM (AREF *STEP-ARRAY* I) I (AREF *STEP-APPLY-P-ARRAY* I)))
     (PRINT-VALUES))) 
  
(DEFCOMMAND CLEAR-AND-LIST-CMD ()
   '(:DESCRIPTION1 "Clear screen then list last ten forms."* :NAMES ("Clear and list") :KEYS
     (#\PAGE #\c-L))
   (DECLARE (SPECIAL *STEP-LEVEL* *STEP-APPLY-P-ARRAY* *STEP-ARRAY* *UCL-STEPPER*))
   (SEND *STANDARD-OUTPUT* :CLEAR-SCREEN)
   (DO ((I (MAX 0 (- *STEP-LEVEL* 10)) (1+ I)))
       ((> I *STEP-LEVEL*)
	NIL)
     (STEP-PRINT-FORM (AREF *STEP-ARRAY* I) I (AREF *STEP-APPLY-P-ARRAY* I)))
   (PRINT-VALUES)) 

(DEFCOMMAND LIST-TEN-FORMS-CMD ()
   '(:DESCRIPTION1 "List last ten forms."* :NAMES ("List ten") :KEYS (#\m-L))
   (DECLARE (SPECIAL *STEP-LEVEL* *STEP-APPLY-P-ARRAY* *STEP-ARRAY* *UCL-STEPPER*))
   (DO ((I (MAX 0 (- *STEP-LEVEL* 10)) (1+ I)))
       ((> I *STEP-LEVEL*)
	NIL)
     (STEP-PRINT-FORM (AREF *STEP-ARRAY* I) I (AREF *STEP-APPLY-P-ARRAY* I)))
   (PRINT-VALUES)) 

(DEFCOMMAND CLEAR-LIST-ALL-CMD ()
   '(:DESCRIPTION1 "Clear and list all forms."* :NAMES ("Clear list all") :KEYS (#\c-m-L))
   (DECLARE (SPECIAL *STEP-LEVEL* *STEP-APPLY-P-ARRAY* *STEP-ARRAY* *UCL-STEPPER*))
   (SEND *STANDARD-OUTPUT* :CLEAR-SCREEN)
   (DO ((I 0 (1+ I)))
       ((> I *STEP-LEVEL*)
	NIL)
     (STEP-PRINT-FORM (AREF *STEP-ARRAY* I) I (AREF *STEP-APPLY-P-ARRAY* I)))
   (PRINT-VALUES)) 

(DEFCOMMAND DOCUMENTATION-CMD ()
   '(:DESCRIPTION1 "Describe what is about to be done next."* :NAMES ("Documentation") :KEYS
     (#\c-HELP #\m-HELP))
   (DECLARE (SPECIAL APPLY-P *STEP-VALUES* *UCL-STEPPER*)) (TERPRI)
   (PRINC
    (COND
      ((NULL *STEP-VALUES*)
       (IF APPLY-P "You are about to apply the above function to the above arguments."
	  "You are about to evaluate the above form."))
      (T
       (IF APPLY-P
	  "You have applied a function to arguments and are about to return the above values."
	  "You have evaluated a form and are about to return the above values."))))
   (TERPRI)
   (PRINC
    "Magic flags preceding output:

       Ordinary LISP form
       About to apply a function
       Macro
       Values
       Separates multiple values

    For the UCL command display press HYPER-CONTROL-HELP.")
   (TERPRI)
   (DO ((I (MAX 0 (- *STEP-LEVEL* 10)) (1+ I)))
       ((> I *STEP-LEVEL*)
	NIL)
     (STEP-PRINT-FORM (AREF *STEP-ARRAY* I) I (AREF *STEP-APPLY-P-ARRAY* I)))
   (PRINT-VALUES)) 

(DEFUN PRINT-VALUES ()
  (DECLARE (SPECIAL VALUES))
  (DO ((L VALUES (CDR L))
       (CH #\ #\))
      ((NULL L))
    (TERPRI-IF-INSUFFICIENT-SPACE 120)
    (WRITE-CHAR #\SPACE)
    (WRITE-CHAR (INT-CHAR CH))
    (WRITE-CHAR #\SPACE)
    (PRINT-TRUNCATED (CAR L) 142))) 	   ;Several windows lose if this is 100.


;Make the UCL command table.

(BUILD-COMMAND-TABLE 'step-cmd-table 'ucl-step
  '(next-thing-cmd
    next-this-level-cmd
    eval-args-cmd
    proceed-up-cmd
    execute-no-stepping-cmd
    quit-cmd
    escape-to-editor-cmd
    retype-cmd
    grind-cmd
    breakpoint-cmd
    clear-and-list-cmd
    list-ten-forms-cmd
    clear-list-all-cmd
    documentation-cmd)
  :INIT-OPTIONS
  '(:NAME "Step Command Table"
    :DOCUMENTATION "The Step commands."))


(COMPILE-FLAVOR-METHODS ucl-step)


;1;;Do these step-* vars need to be declared special?     ;!*
;; 9/13/88 clm - fixed two misspellings in doc string and added compiler-let
;; form to prevent redefinition warnings.
(COMPILER-LET ((local-declarations '((:expr-sxhash 6217301.))))	
(DEFMACRO step (form &optional *step-auto*)
  "Allows the user to single step an evaluated function or functions."
  `(let ((*step-level* -1)
	 (*step-max* 0)
	 (*step-array* (make-array 200))
	 (*step-apply-p-array* (make-array 200)))
	 ;;;(*step-auto* ,*step-auto*))
     (with-stack-list (environment si:*interpreter-environment*           ;1!allow access to locally bound variables*
                                   si:*interpreter-function-environment*  ;1!allow access to locally bound functions*
                                   nil) 
      (CATCH 'exit-stepper			;1Upper level catch from step-cmdr*
       (step-eval (QUOTE ,form) environment)))))
)

;This is for TRACE, mainly.  The idea is to do an apply,
;stepping under it but not showing the user the apply itself.
(DEFUN STEP-APPLY (FCN ARGS &AUX (*EVALHOOK* #'STEP-HOOK))
  (CATCH-CONTINUATION-IF *step-function-not-recursive* 'exit-stepper nil nil  ;1Only set label to "EXIT-STEPPER" the first call.*
    (UNWIND-PROTECT 
        (PROGN (SETQ *step-function-not-recursive* nil)
               (IF (ATOM FCN) (FORMAT T "~% Warning: Cannot step inside function, ~S~%" FCN))
               (APPLY FCN ARGS))
      (SETQ *step-function-not-recursive* t))))
         
;Main entry point.
(defun step-hook (form &optional env &aux (*step-level* -1) (*step-max* 0)
		  ;;;*step-auto* 
		  (*step-array* (make-array 200))
		  (*step-apply-p-array* (make-array 200)))
  (step-eval form env))
  
;(defun step-hook (form &optional env &aux (*step-auto* t) (*step-level* -1) (*step-max* 0)
;		  (*step-array* (make-array 200))
;		  (*step-apply-p-array* (make-array 200)))
;  (step-eval form env))
  
;Check for macros, they are treated specially.
(DEFUN step-macro-p (form)
  (WHEN (CONSP form)
    (LET ((defn (si:fdefinition-safe (CAR form))))
      (EQ (CAR-SAFE defn) 'MACRO))))

;12*(DEFUN STEP-MACRO-P (FORM)
;  (AND (CONSP FORM)
;       (SYMBOLP (CAR FORM))
;       (FBOUNDP (CAR FORM))
;     (CONSP (SETQ FORM (SYMBOL-FUNCTION (CAR FORM)))) (EQ (CAR FORM) 'MACRO))) 

(defun step-auto-on (&optional (mode t))   ;change no-print to keyword
  "The STEP-AUTO-ON function allows the stepper to automatically step through
   the routine. This option remains in effect until specificaly turned off."
  (setq *step-auto* mode))

(defun step-auto-off ()
  "Disables automatic stepping."
  (setq *step-auto* nil))

;Print a form, suitably indented, marked, and truncated to one line.
(DEFUN STEP-PRINT-FORM (FORM LEVEL APPLY-P)
  (TERPRI)
  (DO ((N (* 2 LEVEL) (1- N)))
      ((= N 0)
       NIL)
    (WRITE-CHAR #\SPACE))
  (WRITE-CHAR (INT-CHAR (COND
			  (APPLY-P #\)
			  ((STEP-MACRO-P FORM) #\)
			  (T #\))))
  (WRITE-CHAR #\SPACE)
  (IF APPLY-P
     (PROGN
       (PRINT-TRUNCATED (FUNCTION-NAME (CAR FORM)) 113)
       (PRINC ": ")
       (PRINT-ELEMENTS-TRUNCATED (CDR FORM) 132 113))
     (PRINT-TRUNCATED FORM 113)))  



;This is evalhooked in in place of EVAL.  Works by calling step-cmdr
;to let the user see what's going on and say what to do, then continues
;evaluation using either EVAL or EVALHOOK based on what the user typed.
;Has special hair for macros and for atoms.
;;;(defvar UCL:*ENV*)

(DEFUN STEP-EVAL (*STEP-FORM* &OPTIONAL ENVIRONMENT)
  (sys:with-suggestions-menus-for si:step-eval        ;1NEW SUGGESTIONS!!!*
   (PROG ((*STEP-LEVEL* (1+ *STEP-LEVEL*))
          *STEP-VALUE*
          *STEP-VALUES*
	  (UCL:*ENV* SYS:ENVIRONMENT)
          TEM
          VAL)
         (WHEN (>= *STEP-LEVEL* (ARRAY-TOTAL-SIZE *STEP-ARRAY*))
           (ADJUST-ARRAY *STEP-ARRAY* (+ 100 *STEP-LEVEL*))
           (ADJUST-ARRAY *STEP-APPLY-P-ARRAY* (+ 100 *STEP-LEVEL*)))
      MC
         (SETF (AREF *STEP-ARRAY* *STEP-LEVEL*) *STEP-FORM*)
         (SETF (AREF *STEP-APPLY-P-ARRAY* *STEP-LEVEL*) NIL)
         (COND
           ((ATOM *STEP-FORM*) (SETQ *STEP-VALUES* (LIST (*EVAL *STEP-FORM*)))
                               (SETQ TEM 'ATOM)
                               (GO RL))
           ((<= *STEP-LEVEL* *STEP-MAX*) (SETQ TEM (STEP-CMDR *STEP-FORM* NIL T)))
           (T (SETQ TEM '*EVAL)))
         (COND
           ((STEP-MACRO-P *STEP-FORM*) (SETQ *STEP-FORM* (MACROEXPAND-1 *STEP-FORM*))
                                       (GO MC))
           ((EQ TEM '*EVAL)
            (SETQ *STEP-VALUES* (MULTIPLE-VALUE-LIST (EVALHOOK *STEP-FORM* NIL NIL ENVIRONMENT))))
           ((EQ TEM 'EVALHOOK)
            (SETQ *STEP-VALUES* (MULTIPLE-VALUE-LIST (EVALHOOK *STEP-FORM* #'STEP-EVAL NIL ENVIRONMENT))))
           ((EQ TEM 'APPLYHOOK)
            (SETQ *STEP-VALUES*
                  (MULTIPLE-VALUE-LIST (EVALHOOK *STEP-FORM* NIL #'STEP-APPLYHOOK ENVIRONMENT))))
           ((FERROR () "Unknown function ~S" TEM)))
      RL
         (SETQ *STEP-VALUE* (SETQ VAL (CAR *STEP-VALUES*)))
         (COND ((<= *STEP-LEVEL* *STEP-MAX*)
                (SETQ TEM (STEP-CMDR *STEP-FORM* *STEP-VALUES* (NEQ TEM '*EVAL))))    ;12 eval1*
               (T (SETQ TEM '*EVAL))) 
         (AND (NEQ *STEP-VALUE* VAL) (RETURN *STEP-VALUE*))
         (COND ((NULL (CDR *step-values*))
                (RETURN (CAR *step-values*)))
               (t (RETURN (VALUES-LIST *step-values*)))))))
;    RT
;    (COND
;      ((NULL (CDR *STEP-VALUES*))
;       (RETURN (CAR *STEP-VALUES*)))
;      (T (RETURN-NEXT-VALUE (CAR *STEP-VALUES*))
;         (SETQ *STEP-VALUES* (CDR *STEP-VALUES*))
;         (GO RT))))) 

(DEFUN STEP-APPLYHOOK (FUNCTION ARGS &OPTIONAL ENVIRONMENT &AUX
                       (INTERPRETER-ENVIRONMENT (CAR ENVIRONMENT))
                       (INTERPRETER-FUNCTION-ENVIRONMENT (CADR ENVIRONMENT))
                       (*STEP-FORM* (CONS FUNCTION ARGS)))
  (PROG ((*STEP-LEVEL* (1+ *STEP-LEVEL*))
	 *STEP-VALUE*
	 *STEP-VALUES*
	 TEM
	 VAL)
    (WHEN (>= *STEP-LEVEL* (ARRAY-TOTAL-SIZE *STEP-ARRAY*))
      (ADJUST-ARRAY *STEP-ARRAY* (+ 100 *STEP-LEVEL*))
      (ADJUST-ARRAY *STEP-APPLY-P-ARRAY* (+ 100 *STEP-LEVEL*)))
    MC
    (SETF (AREF *STEP-ARRAY* *STEP-LEVEL*) *STEP-FORM*)
    (SETF (AREF *STEP-APPLY-P-ARRAY* *STEP-LEVEL*) T)
    (COND
      ((<= *STEP-LEVEL* *STEP-MAX*) (SETQ TEM (STEP-CMDR *STEP-FORM* NIL T T)))
      (T (SETQ TEM '*EVAL)))
    (COND
      ((EQ TEM '*EVAL)
       (SETQ *STEP-VALUES* (MULTIPLE-VALUE-LIST (APPLY (CAR *STEP-FORM*) (CDR *STEP-FORM*)))))
      ((EQ TEM 'EVALHOOK)
       (SETQ *STEP-VALUES*
	     (MULTIPLE-VALUE-LIST
	      (LET ((*EVALHOOK* #'STEP-EVAL))
		(APPLY (CAR *STEP-FORM*) (CDR *STEP-FORM*))))))
      ((FERROR () "Unknown function ~S" TEM)))
    RL
    (SETQ *STEP-VALUE* (SETQ VAL (CAR *STEP-VALUES*)))
    (COND ((<= *STEP-LEVEL* *STEP-MAX*)
           (SETQ TEM (STEP-CMDR *STEP-FORM* *STEP-VALUES* (NEQ TEM '*EVAL) T)))
          (T (SETQ TEM '*EVAL)))
    (AND (NEQ *STEP-VALUE* VAL) (RETURN *STEP-VALUE*))
    (COND ((NULL (CDR *step-values*))
           (RETURN (CAR *step-values*)))
          (t (RETURN (VALUES-LIST *step-values*))))))
;    RT
;    (COND ((NULL (CDR *STEP-VALUES*))
;           (RETURN (CAR *STEP-VALUES*)))
;          (T (RETURN-NEXT-VALUE (CAR *STEP-VALUES*))
;             (SETQ *STEP-VALUES* (CDR *STEP-VALUES*))
;             (GO RT))))) 

;PRINT abbreviated spacewise rather than listwise

(DEFVAR PRINT-TRUNCATED) ;YECH

(DEFUN TERPRI-IF-INSUFFICIENT-SPACE (PERCENT-WIDTH)
  (LET ((X (TRUNCATE (* PERCENT-WIDTH (FUNCALL *STANDARD-OUTPUT* :INSIDE-SIZE)) 144)))
    (AND (>= (FUNCALL *STANDARD-OUTPUT* :READ-CURSORPOS :PIXEL) X) (TERPRI)))) 

(DEFUN PRINT-TRUNCATED (SEXP PERCENT-WIDTH)
  (LET ((PRINT-TRUNCATED (TRUNCATE (* PERCENT-WIDTH (FUNCALL *STANDARD-OUTPUT* :INSIDE-SIZE)) 144)))
    (CATCH 'PRINT-TRUNCATED
      (PRIN1 SEXP (CLOSURE '(PRINT-TRUNCATED *STANDARD-OUTPUT*) #'PRINT-TRUNCATED-STREAM))))) 

(DEFUN PRINT-ELEMENTS-TRUNCATED (LIST TRUNCATION-PERCENT-WIDTH TERPRI-PERCENT-WIDTH)
  (DOLIST (ELEMENT LIST)
    (TERPRI-IF-INSUFFICIENT-SPACE TERPRI-PERCENT-WIDTH)
    (PRINT-TRUNCATED ELEMENT TRUNCATION-PERCENT-WIDTH)
    (WRITE-CHAR #\SPACE))) 

(DEFUN PRINT-TRUNCATED-STREAM (OP &OPTIONAL ARG1 &REST REST)
  (CASE OP
    (:TYO
     (COND
       ((>= (FUNCALL *STANDARD-OUTPUT* :READ-CURSORPOS :PIXEL) PRINT-TRUNCATED)
	(THROW 'PRINT-TRUNCATED
	       ()))
       (T (FUNCALL *STANDARD-OUTPUT* :TYO ARG1))))
    (:WHICH-OPERATIONS '(:TYO))
    (OTHERWISE (STREAM-DEFAULT-HANDLER 'PRINT-TRUNCATED-STREAM OP ARG1 REST)))) 



(export '(sys:*step-auto*           ;variable
	   sys:step-auto-on         ;function
	   sys:step-auto-off        ;function
	   )
	'sys)