;; -*- Mode:Common-Lisp; Package:SYSTEM-INTERNALS; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (c) 1980 Massachusetts Institute of Technology 
;;

;; Trace package

;;	"There is always a place for debugging.  No matter how
;;	 hard you try to think of everything in advance, you
;;	 will always find that there is something else that you
;;	 hadn't thought of."
;;			- My Life as a Mathematician
;;			  by Hfpsh Dboups
;;; Items in this menu are lists of the form:
;;;  ("name" :VALUE (S-expr-arg-p . what-to-append-into-trace-options))
;;;			^-- if this is UNTRACE, QUIT, or DO-IT, that
;;;                         special function.
;;;			if NIL, nothing special.
;;;			otherwise is prompt for reading what goes into
;;;                         trace options.
;;; Try to keep this so it comes out in 3 columns.
tv:
(DEFVAR TRACE-ITEM-LIST
   '(("Break before" :VALUE (NIL :BREAK T) :DOCUMENTATION
      "Call BREAK before entering the function, binding ARGLIST.")
     ("Break after" :VALUE (NIL :EXITBREAK T) :DOCUMENTATION
      "Call BREAK after leaving the function, binding VALUES.")
     ("Step" :VALUE (NIL :STEP) :DOCUMENTATION
      "Single-step through the body of the (interpreted) function.")
     ("Cond Step" :VALUE ("Predicate for stepping calls" :STEPCOND) :DOCUMENTATION
      "Asks for a predicate which controls whether stepping happens.")
     ("Error" :VALUE (NIL :ERROR) :DOCUMENTATION
      "Enter the error handler's debugger when the function is called.")
     ("Print" :VALUE ("Form to evaluate and print in trace messages" :PRINT) :DOCUMENTATION
      "Asks for a form to be evaluated and printed in trace messages.")
     ("Print before" :VALUE ("Form to evaluate and print before calling" :ENTRYPRINT)
      :DOCUMENTATION "Asks for a form to be evaluated and printed upon function entry.")
     ("Print after" :VALUE ("Form to evaluate and print after returning" :EXITPRINT)
      :DOCUMENTATION "Asks for a form to be evaluated and printed upon function exit.")
     ("Conditional" :VALUE ("Predicate for tracing" :COND) :DOCUMENTATION
      "Asks for a predicate which controls whether tracing happens.")
     ("Cond before" :VALUE ("Predicate for tracing calls" :ENTRYCOND) :DOCUMENTATION
      "Asks for a predicate which controls tracing of function entry.")
     ("Cond after" :VALUE ("Predicate for tracing returns" :EXITCOND) :DOCUMENTATION
      "Asks for a predicate which controls tracing of function exit.")
     ("Cond break before" :VALUE ("Predicate for breaking before" :BREAK) :DOCUMENTATION
      "Asks for a predicate which controls BREAKing on function entry.")
     ("Cond break after" :VALUE ("Predicate for breaking after" :EXITBREAK) :DOCUMENTATION
      "Asks for a predicate which controls BREAKing on function exit.")
     ("ARGPDL" :VALUE ("Arg pdl variable" :ARGPDL) :DOCUMENTATION
      "Asks for a variable which gets list of recursive argument lists.")
     ("Wherein" :VALUE ("Function within which to trace" :WHEREIN) :DOCUMENTATION
      "Asks for a function, within which tracing will be active.")
     ("Untrace" :VALUE (UNTRACE) :FONT :MENU-STANDOUT :DOCUMENTATION
      "Instead of tracing this function, stop tracing it.")
     ("Abort" :VALUE (QUIT) :FONT :MENU-STANDOUT :DOCUMENTATION
      "Click here to get out of this without doing anything.")
     ("Do It" :VALUE (DO-IT) :FONT :MENU-STANDOUT :DOCUMENTATION
      "Click here to do the tracing with the selected options."))) 

tv:
(DEFWINDOW-RESOURCE TRACE-POP-UP-MENU () :MAKE-WINDOW
   (DYNAMIC-TEMPORARY-MENU
     :NAME "Trace Options"
     :ITEM-LIST-POINTER 'TRACE-ITEM-LIST)
   :REUSABLE-WHEN :DEEXPOSED) 

tv:
(DEFUN TRACE-VIA-MENUS (&OPTIONAL FCN)
  "This function is invoked in the momentary menu process when the
user clicks TRACE and in the editor process by the editor's Trace
command.  If the function isn't supplied as an argument the user is
asked for it."
  (USING-RESOURCE (TRACE-POP-UP-WINDOW POP-UP-TEXT-WINDOW)
     (USING-RESOURCE (TRACE-POP-UP-MENU TRACE-POP-UP-MENU)
	(FUNCALL TRACE-POP-UP-WINDOW :SET-LABEL "Trace")
	(FUNCALL TRACE-POP-UP-WINDOW :SET-SIZE
		 (min 1000			;Accomadate small mac screens  PMH 3/11
		      (tv:sheet-width (tv:sheet-superior trace-pop-up-window))) 
		 (min 300			;Accomadate small mac screens  PMH 3/11
		      (floor (tv:sheet-height (tv:sheet-superior trace-pop-up-window)) 2)))
	(FUNCALL TRACE-POP-UP-WINDOW :CENTER-AROUND MOUSE-X MOUSE-Y)
	(WINDOW-CALL (TRACE-POP-UP-WINDOW :DEACTIVATE)
	 (UNWIND-PROTECT (LET
			  ((BLINKER (CAR (SHEET-BLINKER-LIST TRACE-POP-UP-WINDOW))))
			  (COND
			    ((NULL FCN)
			     ;; Make sure blinker is blinking.
			     (BLINKER-SET-VISIBILITY BLINKER :BLINK)
			     (FORMAT TRACE-POP-UP-WINDOW
				     "Type in name of function to be traced or untraced.
  Abort quits.~%")
			     (DO ((*TERMINAL-IO* TRACE-POP-UP-WINDOW)
				  (*STANDARD-INPUT* TRACE-POP-UP-WINDOW)
				  (*STANDARD-OUTPUT* TRACE-POP-UP-WINDOW))
				 (NIL)
			       (SETQ FCN (READ))
                               (setq fcn (dwimify-arg-package fcn 'function-spec))     ;!
			       (IF (FDEFINEDP FCN)
				 (RETURN ())
				 (FORMAT T " ;not a defined function, try again~%")))))
			  (FUNCALL TRACE-POP-UP-MENU :MOVE-NEAR-WINDOW TRACE-POP-UP-WINDOW)
			  (DO ((FORM (IF (ATOM FCN)
				       `(TRACE (,FCN))
				       `(TRACE (:FUNCTION ,FCN))))
			       (CHOICE)
			       (OPTION)
			       (ARG))
			      (NIL)
			       ;; Put the current status on the text window.
			    (FUNCALL TRACE-POP-UP-WINDOW :CLEAR-SCREEN)
			    ;; 76 is width in characters.
			    (GRIND-TOP-LEVEL FORM 76 TRACE-POP-UP-WINDOW)
			    ;; Not listening to the keyboard any more, shut off blinker.
			    (BLINKER-SET-VISIBILITY BLINKER NIL)
			    ;; Get input from the menu.
			    (SETQ CHOICE (FUNCALL TRACE-POP-UP-MENU :CHOOSE)
				  OPTION (FIRST CHOICE))
			    (COND
			      ((NULL CHOICE))            ;Try again if outside menu
			      ((EQ OPTION 'UNTRACE) (EVAL `(UNTRACE ,FCN)) (RETURN ()))
			      ((EQ OPTION 'QUIT) (RETURN ()))
			      ((EQ OPTION 'DO-IT) (EVAL FORM) (RETURN ()))
			      (T
			       ;; let's make it smarter so that we can undo choices.
			       (LET ((PREVIOUS-ENTRY-POSITION
				      (POSITION (SECOND CHOICE) (THE LIST (SECOND FORM)) :TEST
						#'EQ)))
				 (COND
				   ((AND PREVIOUS-ENTRY-POSITION
				       (OR (EQ (SECOND CHOICE) :BREAK)
					  (EQ (SECOND CHOICE) :EXITBREAK))
				       (CDDR CHOICE))
				    ;; Then we should either remove it from
				    ;; the list if it is the same or just
				    ;; change the argument.
				    (IF (EQ (THIRD CHOICE)
					 (NTH (1+ PREVIOUS-ENTRY-POSITION) (SECOND FORM)))
				     ;; Remove the choice
				      (IF (EQ PREVIOUS-ENTRY-POSITION 0)
					(SETF (SECOND FORM) (CDDR (SECOND FORM)))
					;;ELSE
					(SETF (NTHCDR PREVIOUS-ENTRY-POSITION (SECOND FORM))
					      (NTHCDR (+ 2 PREVIOUS-ENTRY-POSITION)
						      (SECOND FORM))))
				      ;;ELSE
				      ;; Fix the choice
				      (SETF (NTH (1+ PREVIOUS-ENTRY-POSITION) (SECOND FORM))
					    (THIRD CHOICE))))
				   ((AND PREVIOUS-ENTRY-POSITION (NULL OPTION))
				    ;; Delete the choice.
				    (IF (EQ PREVIOUS-ENTRY-POSITION 0)
				      (SETF (SECOND FORM) (CDR (SECOND FORM)))
				      (SETF (NTHCDR PREVIOUS-ENTRY-POSITION (SECOND FORM))
					    (NTHCDR (1+ PREVIOUS-ENTRY-POSITION) (SECOND FORM)))))
				   ((NULL OPTION)
				    ;; Just add the object.
				    (SETF (SECOND FORM) (APPEND (SECOND FORM) (CDR CHOICE))))
				   (T
				    ;; Needs an arg, get it.
				    (FORMAT TRACE-POP-UP-WINDOW "~2%~A:~%" OPTION)
				    ;; Turn on blinker.
				    (BLINKER-SET-VISIBILITY BLINKER :BLINK)
				    (LET ((*TERMINAL-IO* TRACE-POP-UP-WINDOW)
					  (*STANDARD-INPUT* TRACE-POP-UP-WINDOW)
					  (*STANDARD-OUTPUT* TRACE-POP-UP-WINDOW)
					  (FLAG))
				      (MULTIPLE-VALUE-SETQ (ARG FLAG)
					(FUNCALL *TERMINAL-IO* :RUBOUT-HANDLER
						 '((:FULL-RUBOUT :FULL-RUBOUT))
						 #'READ-FOR-TOP-LEVEL))
				      (UNLESS FLAG
				       ;; If previous entry replace the argument
				       ;; otherwise add on new argument.
					(IF (NULL PREVIOUS-ENTRY-POSITION)
					  (SETF (SECOND FORM)
						(APPEND (SECOND FORM) (CDR CHOICE) (CONS ARG ())))
					  (SETF (NTH (1+ PREVIOUS-ENTRY-POSITION) (SECOND FORM))
						ARG)))))))))))
	   (FUNCALL TRACE-POP-UP-MENU :DEACTIVATE)))))) 


;tv:
;(ADD-TO-SYSTEM-MENU-COLUMN :DEBUG "Trace" '(TRACE-VIA-MENUS)
;			   "Trace a function.  Options selected from menu."
;			   :SORTED)
