;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987 Texas Instruments Incorporated. All rights reserved.

;;; Patch file for DECNET version 1.12
;;; Reason: Prevents the echoing of control characters and eliminates the
;;; erroneous output suspended state in VMS EMACS.
;;; Code read by CRW and BRG.
;;; Written 07/22/87 13:10:57 by WALTON,
;;; while running on CHARLIE from band LODA
;;; With IO 3.24, PATHNAME 3.16, FILE 3.9, SYSTEM 3.47, METER 3.3, SERIAL 3.1, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.9, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.7, PROFILE 3.1,
;;;  SUGGESTIONS 3.3, UCL 3.1, ZWEI 3.14, STREAMER-TAPE 3.7, DEBUG-TOOLS 3.1, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.4, NAMESPACE-EDITOR 3.4, NVRAM 3.2,
;;;  TELNET 3.5, TV 3.9, NAMESPACE 3.7, COMPILER 3.20, PRINTER 3.2, SYSLOG 3.0, VT100 3.9,
;;;  IP 2.44, Experimental BUG 7.0, DECNET 1.11,  microcode 400, Band Name: REL3+DECnet

#!C
; From file CTERMINAL.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; CTERMINAL.#"


(defmethod (CTERM-TERMINAL-WINDOW :process-terminal-input)(&optional (special-p nil))
  (without-interrupts			               ; Setup terminal flags
    (let ((echo            (send stream :echo-flag))   ; controls local echoing - nil for passwords
	  (echo-terminator (send stream :echo-terminator))
	  (system-type     (send stream :system-type))
	  (timeout         (send stream :time-out-flag))
	  (termset         (send stream :termset))
	  (max-len         (send stream :max-len))
	  key)

      (when (eql (send stream :system-type) ':VMS4)
      	(send stream :set-raw-mode telnet:keypad-mode))
      
      ; begin by translating keyboard char to ascii
      (if (or special-p                       ; no translation necessary for pretranslated prompt-data
	      (and (keypad-char-p lisp-char) telnet:keypad-mode (not echo)))
	  (setf key lisp-char)
	  (if (meta-char-p lisp-char)         ; EMACS uses these 
	      (send stream :tyo lisp-char)
	      (setf key (map-lisp-char-to-ascii lisp-char))))      
    
      (unless (or (null key) (zerop key))
	;; Map linefeeds and carriage returns:
      	(cond
	  ;; CTRL-S (stop) key.
	  ((= key ASCII-DC3)
	   ; kludge for EMACS
	   (unless (and telnet:keypad-mode
			(not echo)
			(or
			  (and (eql system-type ':VMS4)
			       (eql termset 1))
			  (and (eql system-type ':UNIX-UCB)
			       timeout)))
	     (setf output-suspended t))
	   (send stream :tyo key))
	  ;; CTRL-Q (resume) key.
	  ((and output-suspended (= key ASCII-DC1))
	   (setf output-suspended nil)
	   (send stream :tyo key))
	  ;; CTRL-Z (cancel) key or CTRL-C (interrupt) key
	  ((member key *interrupt-chars*)
	   (send stream :tyo key))	  
	  
	  ;; Non-Canonical Processing.
	  ((and (not output-suspended)
		(or		  
		  (and (not echo)(eql system-type ':VMS4) telnet:keypad-mode)
		  (and (not echo)(eql system-type ':VMS4)
		       (not timeout)(not telnet:keypad-mode)(= termset 1)) ; updspr kludge
		  (eql system-type ':UNIX-UCB)
		  (eql max-len 1)))
	   (cond
	     ;; Rubout key
	     ((= key ASCII-DEL)         
	      (when ECHO
	      	(send self :backward-char)
	      	(send self :clear-char))
	      (send stream :tyo key))	
	     ;; All other keys
	     (t (when (and ECHO (not (ctrl-char-p lisp-char)))
		  (send self :test-for-escape-sequence key))
		(send stream :tyo key))))

	  
	  ;; Canonical Processing.
	  ((and (not output-suspended)
		(or (and echo (eql system-type ':VMS4) telnet:keypad-mode)
		    (not telnet:keypad-mode)))
	   (send self :COOKED-MODE)
	   (cond
	     ;; Handle escaped character
	     (escaped
	      (setf escaped nil)
	      (vector-push key read-buffer)
	      (when ECHO
		(send self :test-for-escape-sequence key)))
	    ;; Command line editing 
	     ((member key *cmdline-edit-keys*)
	      (send self :process-cmdline-edit key))
	    ;; Newline key
	     ((= key ASCII-CR)
	      (vector-push key read-buffer)
	      (send stream :string-out read-buffer)
	      (send stream :force-output)
	      (setf (fill-pointer read-buffer) 0)
	      (setf cmd-edit-in-prog nil)
	      (setf inserting nil)
	      (setf keycount 0)
	      (when (or (eql system-type ':UNIX-UCB)
			(and  ECHO ECHO-TERMINATOR)
			(and telnet:keypad-mode (not echo-terminator)))
		(send self :test-for-escape-sequence ASCII-CR)
		(send self :test-for-escape-sequence ASCII-LF)))
	     ;; Rubout key
	     ((= key  ASCII-DEL)
	      (send self :read-buffer-delete))
	     ((= key ASCII-TAB)
	      (send self :tab-with-spaces))
	     ;; All other keys
	     (t (if (and (not inserting)(not cmd-edit-in-prog))
		    (progn
		      (vector-push key read-buffer)
		      (when ECHO (send self :test-for-escape-sequence key))
		      )
		    ;inserts
		    (if (and cmd-edit-in-prog inserting)
			(progn
			  (send self :insert-char)
			  (send self :clear-char)
			  (send self :test-for-escape-sequence key)
			  (send self :read-buffer-insert key)
			  )
			;overstrikes
			(send self :clear-char)
			(send self :test-for-escape-sequence key)
			(if (or (zerop (fill-pointer read-buffer))         ;beginning of the line
				(>= keycount (fill-pointer read-buffer)))  ;end of the line
			    (vector-push key read-buffer)
			    (setf (aref read-buffer keycount) key))))
	      (incf keycount))
	     )))))))

))
