;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987 Texas Instruments Incorporated. All rights reserved.

;;; Written 08/05/87 14:37:47 by GRAHAM,
;;; Reason: Add default who-line Net Input and Net Output when blocked waiting on transmit or receive. Also changed default
;;; Dna address from DNA:xxxx to DNA|xxxx. SPR #6106. Code read by BRG and LER.
;;; while running on Oppenheimer from band LOD1
;;; With IO 3.18, PATHNAME 3.13, FILE 3.7, SYSTEM 3.36, METER 3.1, SERIAL 3.0, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.7, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.6, PROFILE 3.0,
;;;  SUGGESTIONS 3.3, UCL 3.0, ZWEI 3.6, STREAMER-TAPE 3.5, DEBUG-TOOLS 3.0, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.1, NAMESPACE-EDITOR 3.2, NVRAM 3.0,
;;;  TELNET 3.0, TV 3.5, NAMESPACE 3.7, COMPILER 3.9, PRINTER 3.1, SYSLOG 3.0, VT100 3.1,
;;;  IP 2.22, Experimental BUG 4.18, DECNET 1.0,  microcode 395, Band Name: 6/25+DECnet

#!C
; From file NSP_PORTS.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; NSP_PORTS.#"


(defmethod (nsp-port :foreign-host) ()
  (or (si:get-host-from-address NODEaddr :dna nil)
      (format nil "DNA|~d" NODEaddr)))

))

#!C
; From file NSP_PORTS.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; NSP_PORTS.#"


(defmethod (nsp-port :wait-for-connect-accept) ()
  (process-allow-schedule)
  (let ((connect-success nil)
	(connect-increment 30.)
	(connect-delay 30.))
    (unwind-protect
	(dotimes (i 12.)
	  (when (process-wait-with-timeout Who-line-state
					   connect-delay
					   'port-state-accept self)
	    (setf connect-success t)
	    (return t))
	  (setf connect-delay (+ connect-delay connect-increment))
	  (send self :retransmit-connect)
	  (when (= i 1) (setf who-line-state (string-append "Retry " who-line-state)))
	  )
	(if connect-success
	    (progn
	      (send self :remove-retransmitted-connect)
	      (setf who-line-state nil))
	    (send self :reset)))	   ;if rejected, reset the port to refuse further communication to this port.
    (if (eql state :rejected) nil t)))
))

#!C
; From file NSP_PORTS.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; NSP_PORTS.#"


(defmethod (nsp-port :internal-data-rcv) (data-request-sent)
  (block outer-loop
    (dotimes (i 100. nil)
      (progn
	(process-wait-with-timeout (or Who-line-state "Net Input")
				    1800.
				   #'(lambda (instance)
				       (or (send instance :receive-packet-queue)
					   (not (or (eql (send instance :state) :RUNNING)
						    (eql (send instance :state) :CONNECT-CONFIRM)))
					   ))
				   self)		      
	(if receive-packet-queue
	    (let ((entry receive-packet-queue))
	      (setf receive-packet-queue (send entry :next))
	      (setf (send entry :next) nil)
	      (return-from outer-loop entry))
	    (if (not (or (eql state :RUNNING)
			 (eql state :CONNECT-CONFIRM)))
		(return-from outer-loop nil)	   ;We will get this if the other end disconnected.
		(progn
		  (setf data-request-sent (or (send self :retransmit-next-message)
					      (send self :send-link-service-message (nsp-get-pkt))
					      ))
		  (unless data-request-sent
		    (send self :send-data-request-message (nsp-get-pkt))
		    (setf data-request-sent t)))))))))
))

#!C
; From file NSP_TRANSMIT.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; NSP_TRANSMIT.#"


(defmethod (NSP-port :segmentation-send) (pkt eom bom)
  (declare (special *setup-both*
		    *setup-neither*))
  (let ((header (get-nsp-header pkt))
	(timeout nil))
    (set-msgflag-data-message eom bom header)
    (setf (dstaddr header) ADDRrem)
    (setf (srcaddr header) ADDRloc)
    (unless *setup-neither*
      (setf (acknum-number header) NUMdata-received)
      (setf (acknum-qual header) qual-ack)
      (set-acknum-presence header t)
      (when *setup-both*
	(setf (ackoth-number header) NUMoth-received)
	(setf (ackoth-qual header) cross-subchannel-ack)
	(set-ackoth-presence header t)))
    (setf (segnum header) NUMdat)
    (setf state :RUNNING)
    (if (eql flowrem-sw :SEND)
	(if (<=segnum numdat (+segnum remote-window-size numdat-remote))
	    (progn
	      (nsp-display "~%TRANSMIT normal data pkt. Number ~d." numdat)
	      (send self :transmit-buffer pkt))
	    (loop
	      (setf timeout
		    (process-wait-with-timeout (or Who-line-state "Net Output") 1800.
		     #'(lambda (instance)
				(or ;(send instance :try-sending)
				    (not (eql (send instance :state) :RUNNING))
				    (= (send instance :numdat)
				       (+segnum 1 (send instance :numdat-remote)))))
			    self))
	      (if timeout 
		  (progn
		    (setf try-sending nil)
		    (when (eql state :RUNNING)
		      (send self :transmit-buffer pkt)
		      (nsp-display "~%TRANSMIT AFTER WAIT normal data pkt. Number ~d." numdat))
		    (return nil))
		  (send self :send-other-data-message (nsp-get-pkt) t))))
	(send self :queue-for-retransmission pkt))
    (setf NUMdat (+segnum 1 NUMdat))
    (incf Data-segment-message-transmitted)))
))
