;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987 Texas Instruments Incorporated. All rights reserved.

;;; Patch file for DECNET version 1.18
;;; Reason: Multitudes of fixes to DECnet PMU. SPR #6106. Code read by BRG and LER.
;;; Written 08/05/87 16:15:45 by GRAHAM,
;;; while running on Oppenheimer from band LOD1
;;; With IO 3.18, PATHNAME 3.13, FILE 3.7, SYSTEM 3.36, METER 3.1, SERIAL 3.0, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.7, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.6, PROFILE 3.0,
;;;  SUGGESTIONS 3.3, UCL 3.0, ZWEI 3.6, STREAMER-TAPE 3.5, DEBUG-TOOLS 3.0, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.1, NAMESPACE-EDITOR 3.2, NVRAM 3.0,
;;;  TELNET 3.0, TV 3.5, NAMESPACE 3.7, COMPILER 3.9, PRINTER 3.1, SYSLOG 3.0, VT100 3.1,
;;;  IP 2.22, Experimental BUG 4.18, DECNET 1.0,  microcode 395, Band Name: 6/25+DECnet

#!C
; From file DECNET_SERVERS.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DECNET_SERVERS.#"



(defun pmu-error (host stream message address)
  (send stream :setup-next-input-buffer)
  (let ((string (send stream :packet-string)))
    (send message :dispose-address address :rejected host string))
  (send stream :packet-string))		                    ;This is to flush the next buffer which contains #x00

(defun send-zero-message (stream)
  (send stream :tyo (int-char #x00))
  (send stream :force-output))

(defun clean-up-transmission (stream &optional (any-error-p nil))
  (send-zero-message stream)
  (let ((returned-value (char-int (send stream :tyi))))
    (when (eql returned-value #x01)
      (dotimes (i 3) (send stream :tyi))                    ; Use up rest of this message; and get acknowledge.
      (unless any-error-p (send stream :tyi)))))

(defun start-pmu (host stream from message address-list total-to-list)
  (let ((any-valid-addresses nil)
	(any-bad-addresses t))
    (send stream :string-out (string-upcase from))          ;The first packet contains the user-id field
    (send stream :force-output)
    (dolist (to address-list (send-zero-message stream))
      (send stream :string-out                              ;Until we see a #x00, TO: fields will follow.
	    (string-upcase (send to :local-part-string)))
      (send stream :force-output)
      (if (= (char-int (send stream :tyi)) 1)
	  (progn 
	    (dotimes (i 3) (send stream :tyi))
	    (setf any-valid-addresses t)
	    (send message :dispose-address to :delivered host))
	  (progn
	    (pmu-error host stream message to)              ; One of the addresses failed.
	    (setf any-bad-addresses t)
	    )))
    (when any-valid-addresses
      (if total-to-list                                     ;send total address list
	  (send stream :string-out total-to-list)                          
	   (dolist (address address-list)                    ;If TO: line not available(?), send addresses from address-list.
	     (send stream :string-out (send address :string-for-message))
	     (send stream :string-out ",")))
      (send stream :force-output))
    (values any-valid-addresses any-bad-addresses)))

;; Most of this code was ripped from the Chaos/SMTP mail interface.
(defun pmu-mail-client (host stream host-queue)
  (let ((*print-base* 10)
	(*read-base* 10)
	(message-alist (mail:host-queue-message-alist host-queue))
	 message address-list header-type)
    (unwind-protect
	(loop
	  with max = (length (mail:host-queue-message-alist host-queue))
	  when (minusp (decf max)) return nil	            ;Avoid getting stuck here in case of a mail loop.
	  for message-item in message-alist
	  do
	  (block do-message
		  (setq message (car message-item))
		  (setq address-list (cdr message-item))
		  (when (null address-list)
		    (return-from do-message))
		  (let ((from "")
			(subject "")
			(total-to nil)
			(any-messages-sent nil)
			(any-address-errors)
			(msg-stream (mail:make-message-input-stream message)))
		      (loop
			(multiple-value-bind (line eof)
			    (read-line msg-stream nil nil)
			  (when (or eof (null line))
			    (return nil))
			  (when (mail:string-blank-p line) (return nil))
			  (setq header-type (mail:header-line-type line))
			  (cond ((eq header-type :subject)
				 (setq subject (mail:header-line-body-string line)))
				((eq header-type :from)
				 (setq from (mail:header-line-body-string line)))
				((eq header-type :to)
				 (setq total-to (mail:header-line-body-string line)))
				)))
		      (send msg-stream :close)
		      (when (search "@" from)
			(setq from (subseq from 0 (search "@" from))))
		      (multiple-value-setq (any-messages-sent any-address-errors)
			 (start-pmu host stream from message address-list total-to))
		      (if any-messages-sent
			  (progn
			    (send stream :string-out (string-upcase subject))	   ;send SUBJECT: field
			    (send stream :force-output)

			    (setq msg-stream (mail:make-message-input-stream message))
			    (when (not (streamp msg-stream))
			      (return-from do-message))
			    (loop
			      for line = (read-line msg-stream nil nil)
			      until (null line)
			      do
			      (if (string-equal "" line)
				  (send stream :string-out " ")
				  (send stream :string-out line))
			      (send stream :force-output)
			      finally	                                    ;send #x00 to signify end of message.
			      (clean-up-transmission stream any-address-errors)))
			  (send msg-stream :close))
		      (send message :update))))
      (send stream :close)
      (mail:dispose-host-queue-message host message-alist))))
    
;; All of the DECnet PMU (personal message utility) SERVER code resides here.

(defun PMU-force-acknowledge (stream)
  (send stream :tyo (int-char #x01))
  (send stream :tyo (int-char #x00))
  (send stream :tyo (int-char #x00))
  (send stream :tyo (int-char #x00))
  (send stream :force-output))

(defun PMU-reject (stream string)
  (send stream :tyo (int-char #xFA))
  (send stream :tyo (int-char #x10))
  (send stream :tyo (int-char #x7E))
  (send stream :tyo (int-char #x00))
  (send stream :force-output)
  (send stream :string-out string)
  (send stream :force-output))

(defun pmu-read (stream)
  (send stream :packet-string))

(defun pmu-end-session (stream &optional (any-error-p nil))
  (unless any-error-p
    (send stream :tyo (int-char #x00))
    (send stream :force-output)))

(defun process-from-field (mail-stream message foreign-host)
  (let ((text-line (pmu-read mail-stream)))
    (when text-line
      (when (search " " text-line)                                    ;(go ahead and strip any trailing blanks.)
	(setf text-line (subseq text-line 0 (search " " text-line))))
      (setf text-line (format nil "~a@~a" text-line foreign-host))
      (send message :set-return-path (mail:parse-address text-line))
      (send message :line-out (format nil "~a ~a" "From:    " text-line)))
    text-line))

(defun process-to-field (mail-stream message)
  (let* ((text-line nil)
	 (verified nil)
	 (verify-reason nil)
	 (offset nil)
	 (message-delivered nil)
	 (any-error nil)
	 (address nil))
    (loop
      (if (and (setq text-line (pmu-read mail-stream))	       ;These are TO: field(s), process until we get a #x00
	       (not (eql (int-char 0) (aref text-line 0))))
	  (progn
	    (setq address (mail:parse-address text-line  (+  (or (prog1 (search "::" text-line)
								    (setf offset 2))
							     (prog1 (search "@" text-line)
								    (setf offset 1))
							     (prog1 0 (setf offset 0))) offset)
					      nil nil :mailbox))
	    (setq address (send address :remove-local-host))
	    (multiple-value-setq (verified verify-reason)
	      (send address :verify))
	    (if verified
		(progn
		  (send message :add-recipient address)
		  (pmu-force-acknowledge mail-stream)
		  (setf message-delivered t))
		(progn
		  (pmu-reject mail-stream
			      (format nil "%Error ~s for: ~s"
				      verify-reason
				      (send address :address-string)))
		  (pmu-end-session mail-stream)
		  (setf any-error t))))
	  (return)))
    (when message-delivered 
      (send message :line-out (format nil "~a ~a" "To:      " (pmu-read mail-stream))))
    (values message-delivered any-error)))

(defun process-subject-field (mail-stream message)
  (let ((text-line (pmu-read mail-stream)))
    (when text-line
      (send message :set-subject text-line)
      (send message :line-out (format nil "~a ~a" "Subject: " text-line))
      (send message :line-out ""))
    text-line))

(defun pmu-server-function ()
  (let (message server-object foreign-host text-line data-p)
    (unwind-protect
	(with-open-stream
	  (mail-stream (host:listen-for-connection-on-medium
			 :nsp-stream "PMU"
			 :stream-type :ascii-translating-character-stream))
	  (setq foreign-host (or (send (send mail-stream :foreign-host) :name)
				 "UNKNOWN"))
	  (setq server-object (add-server-to-wholine mail-stream "MAIL"))
	  (setq message (allocate-resource 'mail:message))
	  (send message :line-out (format nil "~a ~a" "Date:" (time:print-current-time nil)))
	  (when (process-from-field mail-stream message foreign-host)
	    (multiple-value-bind (any-message-received any-message-error)
	      (process-to-field mail-stream message)
	      (if any-message-received
		  (when (process-subject-field mail-stream message)
		    (loop		                                  ;This is the text of the message.
		      (setq text-line (pmu-read mail-stream))
		      (when (eql (int-char 0) (aref text-line 0)) (return))
		      (setf data-p t)
		      (send message :line-out text-line))
		    (when data-p
		      (send message :force-output)
		      (mail:enqueue-message message)
		      (pmu-force-acknowledge mail-stream)
		      (pmu-end-session mail-stream any-message-error)))
	      (pmu-end-session mail-stream)))))
      (when server-object
	(send w:who-line-file-state-sheet :delete-server server-object))
      )))

))
