;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987 Texas Instruments Incorporated. All rights reserved.

;;; Written 08/06/87 10:36:06 by REINER,
;;; Reason: Fix parse-dna-address-spec to parse a host who has an address in dotted notation.  [SPR#6176]
;;; Code Readers:  BRG, LER
;;; while running on NOVEMBER from band LOD1
;;; With IO 3.25, PATHNAME 3.16, FILE 3.10, SYSTEM 3.50, METER 3.5, SERIAL 3.1, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.9, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.7, PROFILE 3.1,
;;;  SUGGESTIONS 3.3, UCL 3.3, ZWEI 3.14, STREAMER-TAPE 3.7, DEBUG-TOOLS 3.2, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.4, NAMESPACE-EDITOR 3.4, NVRAM 3.3,
;;;  TELNET 3.5, TV 3.15, NAMESPACE 3.7, COMPILER 3.22, PRINTER 3.2, SYSLOG 3.0, VT100 3.12,
;;;  IP 2.46, Experimental BUG 7.0, DECNET 1.14,  microcode 169, Band Name: 7/27 + DECNET + ENET patch

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(Defun PARSE-DNA-ADDRESS-SPEC (address-spec &optional (check-tag nil) &aux (start 0) (address 0))
  "Returns the address corresponding to address-spec.  If check-tag is T, this must be a string beginning with
   'IP|'.  Otherwise a number is also accepted (and simply returned).  Nil is returned if check tag fails.
   An error is signalled if the string is formatted incorrectly." 
  (if address-spec
      (COND
	((SYMBOLP address-spec) (parse-dna-address-spec (dna:map-name-to-node (string address-spec))))
	
	((TYPEP address-spec 'net:host) (parse-dna-address-spec (dna:map-name-to-node (string address-spec))))
	
	((FLOATP address-spec) (parse-dna-address-spec (format nil "~a" address-spec) check-tag))
	
	((NUMBERP address-spec) (UNLESS check-tag address-spec))
	
	((STRINGP address-spec)
	 (WHEN (STRING-EQUAL address-spec "DECnet|" :start1 0 :end1 7 :start2 0 :end2 7) (SETQ start 7))
	 
	 (WHEN (STRING-EQUAL address-spec "DNA|" :start1 0 :end1 4 :start2 0 :end2 4)    (SETQ start 4))
	 
	 (WHEN (OR (> start 0) (NOT check-tag))
	   (SETQ address (COND 
			   ;; form = #
			   ((EQL (AREF address-spec start) #\#)
			    (CASE (aref address-spec (1+ start))
			      ((#\X #\x) (PARSE-INTEGER address-spec :start (+ start 2) :radix 16))
			      ((#\O #\o) (PARSE-INTEGER address-spec :start (+ start 2) :radix 8))
			      ((#\B #\b) (PARSE-INTEGER address-spec :start (+ start 2) :radix 2))))
			   
			   ;; form = d.d
			   ((SEARCH "." address-spec) (parse-dna-byte-spec address-spec start))
			   
			   ;; form = number (decimal)
			   (t (PARSE-INTEGER address-spec :start start :radix 10 :junk-allowed t))))
	   
	   (COND ((or address (parse-dna-address-spec (dna:map-name-to-node (string address-spec)))))
		 (t (FERROR nil "Illegal DECnet address ~A" address-spec)))))	   
	
	(t (UNLESS check-tag (FERROR nil "Illegal DECnet address: ~A" address-spec))))
      (FERROR nil "Illegal DECnet address: ~A" address-spec))
  
  )

))
