;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987 Texas Instruments Incorporated. All rights reserved.

;;; Patch file for DECNET version 1.21
;;; Reason: Routing fix to utilize *designated-router* when no message entry found for intended destination.
;;; [SPR#6109]  Code Readers:  BRG, LER
;;; Written 08/06/87 11:00:01 by REINER,
;;; while running on NOVEMBER from band LOD1
;;; With IO 3.25, PATHNAME 3.16, FILE 3.10, SYSTEM 3.50, METER 3.5, SERIAL 3.1, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.9, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.7, PROFILE 3.1,
;;;  SUGGESTIONS 3.3, UCL 3.3, ZWEI 3.14, STREAMER-TAPE 3.7, DEBUG-TOOLS 3.2, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.4, NAMESPACE-EDITOR 3.4, NVRAM 3.3,
;;;  TELNET 3.5, TV 3.15, NAMESPACE 3.7, COMPILER 3.22, PRINTER 3.2, SYSLOG 3.0, VT100 3.12,
;;;  IP 2.46, Experimental BUG 7.0, DECNET 1.14,  microcode 169, Band Name: 7/27 + DECNET + ENET patch

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(DEFMETHOD (routing-handler :transmit-pkt)
	   (pkt destination-address)
  "Invoke to send a canonical packet to the given destination decnet address"
  (LET* ((default-cons-area *decnet-area*)
	 (INHIBIT-SCHEDULING-FLAG T)
	 (start-time      (sys:%fixnum-microsecond-time))
	 (header-preamble (net:get-datalink-header pkt))
	 (header          (net:get-network-header pkt))
	 )
    
    (setf (net:packet-datalink-nbytes pkt) (net:packet-logical-length pkt))
    
    (setf (flags-packet-size header-preamble) (net:get-datalink-data-nbytes pkt))
    
    
    (setf (route-padded              header) 0)
    (setf (route-version             header) 0)
    (setf (route-intra-ethernet      header) 1)  
    (setf (route-return-trip         header) 0)
    (setf (route-return-to-sender    header) 0)
    (setf (route-format              header) 6)
    (setf (route-dst-area            header) 0)
    (Setf (route-dst-subarea         header) 0)
    (setf (route-dst-node-hiord      header) HIORD)
    (setf (route-dst-node            header) destination-address)
    (setf (route-src-area            header) 0)
    (setf (route-src-subarea         header) 0)
    (setf (route-src-node-hiord      header) HIORD)
    (setf (route-src-node            header) my-address)
    (setf (route-next-level-2-router header) 0)
    (setf (route-visit-count         header) 0)
    (setf (route-service-class       header) 0)
    (setf (route-protocol-type       header) 0)
    
    (UNLESS (MEMBER (area destination-address) dna:my-areas)
      (INCF packets-to-other-area))

;    (pad-pkt pkt)

    (when (and (null (lookup-on-ethernet-cache-element destination-address))
		     (not (null *designated-router*)))
	    (setf destination-address *designated-router*))

    (routing-transmit (net:get-physical-data pkt)
		      (net:get-physical-data-nbytes pkt)
		      destination-address)
    (Incf time-sending        
	  (round
	    (time-difference
	      (sys:%fixnum-microsecond-time) start-time) 1000.)
	  )
    )
  )

))
