;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987 Texas Instruments Incorporated. All rights reserved.

;;; Written 08/06/87 14:17:44 by graham,
;;; Reason: Fixes bug report 6183. Merges a type of :UNSPECIFIC when no type is returned from the remote host. Type used to 
;;; be NIL which allowed merging to put in any old type. Code read by BRG and LER.
;;; while running on Oppenheimer from band LOD1
;;; With IO 3.18, PATHNAME 3.13, FILE 3.7, SYSTEM 3.36, METER 3.1, SERIAL 3.0, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.7, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.6, PROFILE 3.0,
;;;  SUGGESTIONS 3.3, UCL 3.0, ZWEI 3.6, STREAMER-TAPE 3.5, DEBUG-TOOLS 3.0, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.1, NAMESPACE-EDITOR 3.2, NVRAM 3.0,
;;;  TELNET 3.0, TV 3.5, NAMESPACE 3.7, COMPILER 3.9, PRINTER 3.1, SYSLOG 3.0, VT100 3.1,
;;;  IP 2.22, Experimental BUG 4.18, Inconsistent DECNET 1.21,  microcode 395, Band Name: 6/25+DECnet

#!C
; From file DAP_UTILITIES.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_UTILITIES.#"


(defun parse-returned-pathname (namestring pname)
  (let ((tname (fs:parse-namestring namestring (send pname :host) pname)))
    (if (null (send tname :type))
	(send tname :new-type :unspecific)
	tname)))
))

#!C
; From file DAP_UTILITIES.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_UTILITIES.#"


(defmethod (DAP-basic-stream :open) (pname direction new-version error-if-does-not-exist
				     data-type overwrite append supersede byte-size 
				     ascii-stream batch estimated-length)
  (setf current-direction direction)
  (when (eql direction :output) (setf (send (send self :port) :remote-window-size) 1))
  (let ((condition nil)
	(returned-name-string)
	(signaled-error t))
    (unwind-protect
	(setf condition
	      (catch 'dap-catch-errors
		(when send-config?
		  (send self :send-config-message)
		  (send self :process-config-message)
		  (setf send-config? nil))
		(if (eql direction :input)
		    (setf returned-name-string
			  (send self :process-open-input pname data-type byte-size ascii-stream))
		    (setf returned-name-string
			  (send self :process-open-output pname new-version error-if-does-not-exist
			                                  data-type overwrite append supersede byte-size
				                          estimated-length ascii-stream batch)))
		(setf signaled-error nil)))
      (when signaled-error (setf dap-internal-state :CLOSED)))	   ;If error on open, dont send access complete on close.
    (if (eql (type-of condition) 'dna-base-error-flavor)
	(values nil condition)
	(progn
	  (send self :putprop (parse-returned-pathname returned-name-string pname) :truename)
	  (values (send self :get :truename) nil)))))
))

#!C
; From file DAP_UTILITIES.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_UTILITIES.#"


(defmethod (DAP-basic-stream :pathname-delete) (del-pathname)
  (let ((condition nil)
	(returned-name-string nil))
    (setf condition (catch 'dap-catch-errors
		      (when send-config?
			(send self :send-config-message)
			(send self :process-config-message))
		      (send self :send-access-message accfunc-erase
			    0
			    del-pathname
			    nil
			    nil
			    #x111
			    nil)
		      (send self :process-attributes-message)
		      (multiple-value-bind (cdt rdt)
			  (send self :process-date-time-attributes)
			(when cdt (send self :putprop (time:parse-universal-time cdt) :creation-date))
			(when rdt (send self :putprop (time:parse-universal-time rdt) :last-updated)))
		      (setf returned-name-string (send self :process-name-message))
		      (send self :process-access-complete-response t)))
    
    (if (eql (type-of condition) 'dna-base-error-flavor)
	(progn
	  (send self :send-continue-transfer-message confunc-skip)
	  (values nil condition))
	(send self :putprop (parse-returned-pathname returned-name-string pathname) :truename)
	(send self :clear-receive-queue)
	(add-port-to-released-list port)
	(values (send self :get :truename) nil))))
))

#!C
; From file DAP_UTILITIES.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_UTILITIES.#"


(defmethod (DAP-basic-stream :pathname-rename) (old-pathname new-pathname)
  (let ((pname nil)
	(display (loop for display-list in '(#.display-date-time #.display-name)
		       with display = 0
		       do  (setq display (dpb 1 (byte 1 display-list) display))
		       finally (return display)))	
	(condition nil))
					   ;ask for allocation-attribs, date-time, and name msgs in response
	(setf condition
	      (catch 'dap-catch-errors
					   ;setup and request msg series
		(when send-config?
		  (send self :send-config-message)
		  (send self :process-config-message))
		
		(send self :send-access-message accfunc-rename 0 old-pathname nil nil display nil)
		(send self :send-name-message filespec-nametype new-pathname)  
		
					   ;response msgs
		(send self :process-date-time-attributes)  ;for resultant old-pathname
		(setf pname (send self :process-name-message))
		(send self :process-acknowledge-message)
		
		(send self :process-date-time-attributes)  ;for resultant new-pathname
		(setf pname (send self :process-name-message))
		(send self :process-acknowledge-message)
		
		(send self :process-access-complete-response)))

      (if (eql (type-of condition) 'dna-base-error-flavor)
	  (progn
	    (values nil condition))
	  (send self :putprop (parse-returned-pathname pname pathname) :truename)
	  (send self :clear-receive-queue)
	  (add-port-to-released-list port)
	  (values (send self :get :truename) nil))))
))

#!C
; From file DAP_UTILITIES.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_UTILITIES.#"


(defmethod (DAP-basic-stream :pathname-truename) (&optional (pname (send self :pathname)))
  (let ((returned-name-string)
	(display (loop for display-list in '(#.display-main #.display-date-time #.display-name)
		       with display = 0
		       do  (setq display (dpb 1 (byte 1 display-list) display))
		       finally (return display)))
	(condition nil))
    (setf condition
	  (catch 'dap-catch-errors
	    (progn
	      (when send-config?
		(send self :send-config-message)
		(send self :process-config-message))
	      (send self :send-attributes-message) 
	      (send self :send-access-message
		    accfunc-open	   ;Access function
		    0			   ;accopt (access options)
		    (send pname :string-for-host)  ;pathname
		    fac-get		   ;File access operations.
		    nil			   ;Operations shared with other users.
		    display		   ;display attributes to be returned 
		    nil)		   ;password
	      (send self :process-attributes-message)
	      (multiple-value-bind (cdt rdt)
		  (send self :process-date-time-attributes)
		(when cdt (send self :putprop (time:parse-universal-time cdt) :creation-date))
		(when rdt (send self :putprop (time:parse-universal-time rdt) :last-updated)))
	      (setf returned-name-string (send self :process-name-message))
	      (send self :process-acknowledge-message)
	      (send self :send-access-complete-message cmpfunc-close)
	      (send self :process-access-complete-response)
	      (setf dap-internal-state :CLOSED))))
    
    (if (eql (type-of condition) 'dna-base-error-flavor)
	(values nil condition)
	(send self :putprop (parse-returned-pathname returned-name-string pathname) :truename)
	(send self :clear-receive-queue)
	(add-port-to-released-list port)
	(values (send self :get :truename) nil))))
))
