;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Written 08/17/87 11:03:30 by reiner,
;;; Reason: Partial fix to DECnet's problem of dropping packets.  [SPR#6121] Code Readers: BRG, LER
;;; Reason: Partial fix to DECnet's problem of dropping packets.  [SPR#6121]  Code Readers:  BRG, LER
;;; while running on NOVEMBER from band LOD1
;;; With IO 3.25, PATHNAME 3.16, FILE 3.10, SYSTEM 3.50, METER 3.5, SERIAL 3.1, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.9, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.7, PROFILE 3.1,
;;;  SUGGESTIONS 3.3, UCL 3.3, ZWEI 3.14, STREAMER-TAPE 3.7, DEBUG-TOOLS 3.2, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.4, NAMESPACE-EDITOR 3.4, NVRAM 3.3,
;;;  TELNET 3.5, TV 3.15, NAMESPACE 3.7, COMPILER 3.22, PRINTER 3.2, SYSLOG 3.0, VT100 3.12,
;;;  IP 2.46, Experimental BUG 7.0, DECNET 1.24,  microcode 409, Band Name: 7/27+DECNET+ENET patch

#!C
; From file SC_HOST.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; SC_HOST.#"


(defun sc-establish-connection
	   (&key dstnode nbytes format name usrcode
	    grpcode descrpt userid password accounting objtype)
  (declare (special *nsp-object*))
  (let ((port nil)
	(retry-p t))        ; first try has to be t
    (loop      
      (if retry-p
	  (let* ((reject-pkt nil)(reject-reason nil)
		 (pkt (allocate-special-pkt
			    :kind :CONNECT :len nbytes :format format :name name
			    :usrcode usrcode :grpcode grpcode :descrpt descrpt
			    :rqstrid userid :password password
			    :account accounting :objtyp objtype)))
	    
	    (setf port (send *NSP-object* :connect-xmt
			     (parse-dna-address-spec dstnode) pkt))
	    
	    (if (send port :wait-for-connect-accept)
		(setf retry-p nil)	                ;don't need to retry - have succeeded
		(setf reject-pkt (send port :get-reject-data)))
	    
	    ; handle the case of no response from the other end
	    ; give the caller a second chance
	    (when (and retry-p (null reject-pkt))       ;failed on first try, but no reject packet was received
	      (if (send port :wait-for-connect-accept)
		  (setf retry-p nil)	                ; succeeded on second try
		  (progn
		    (setf reject-pkt (send port :get-reject-data))
		    (send port :close)
		    (when (null reject-pkt)               ; host did not respond
		      (ferror 'dna:host-not-responding-during-connection "Host ~A not responding."
			      dstnode)))))
	    (when reject-pkt
	      (unwind-protect
		  (progn
		    (send port :close)
		    (setf reject-reason (get-two-byte-field (get-session-header reject-pkt :DNA-SESSION 2) 0))
		    (when reject-reason
		      (cond
			((zerop reject-reason))
			((= reject-reason inv-dst-name)
			 (cerror  "Tries the connection again with a different end user name."
				  "Connection rejected because ~s is an invalid end user name."
				  (sc-get-DECnet-name dstnode))
			 (setf dstnode (prompt-and-read :string-trim " Enter a new value:")))
			
			((= reject-reason inv-pwd-rqstid)
			 (multiple-value-setq (userid password)
			   (fs:file-get-password-etc (sc-get-DECnet-name dstnode) userid nil t)))
			
			((= reject-reason inv-acct)
			 (cerror "Tries the connection again with a different account number."
				 "Connection rejected because of an invalid account number or balance.")
			 (setf accounting (prompt-and-read :string-trim " Enter a new account number:")))
			
			(t (multiple-value-bind (signal-name signal-string)(map-reject-to-signal reject-reason)
			     (sc-error signal-name signal-string))))))	    
		(return-pkt reject-pkt)))
	    )     
	  (return nil)))
    port))

))

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(defun PROCESS-ROUTER-HELLO (pkt dst-addr)
  "Handle a just-received router hello packet"
  (let ((entry (update-on-ethernet-cache pkt dst-addr)))
    (unwind-protect
	(progn
	  (incf (routing-handler-router-hello-packets-received *routing-handler*))
	  (incf (on-ethernet-cache-router-hello-packets-received entry))
	  (when (eql dst-addr ALL-ENDNODES)
	    (let ((header  (net:get-network-header pkt)))
	      (when (eql my-area (area (src-node header)))
		(setf *designated-router* (src-node header))))
	    )
	  )
      (send *routing-handler* :deallocate-pkt pkt)
      )))

))

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(defun PROCESS-ENDNODE-HELLO (pkt dst-addr)
  "Handle a just-received endnode hello packet"
  (let* ((entry (update-on-ethernet-cache pkt dst-addr)))
    (unwind-protect
	(progn
	  (incf (routing-handler-endnode-hello-packets-received *routing-handler*))
	  (incf (on-ethernet-cache-endnode-hello-packets-received entry)))
      (send *routing-handler* :deallocate-pkt pkt)))
  )

))

#!C
; From file DAP_STREAM_FLAVOR.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_STREAM_FLAVOR.#"


(defflavor dap-input-character-mixin
	   ((return-cr-array nil)
	    (return-first-header nil)
	    (first-offset nil)
	    (first-length nil)
	    (first-header nil)
	    (basic-stream nil)
	    (pathname nil))
	   (si:property-list-mixin)
  (:required-flavors si:buffered-input-character-stream)
  (:settable-instance-variables)
  (:gettable-instance-variables)
  (:documentation "DAP sequential stream"))

))


#!C
; From file DAP_STREAM_FLAVOR.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_STREAM_FLAVOR.#"


(defmethod (dap-input-character-mixin :after :close) (&optional abort-p)
  (when (and abort-p si:stream-input-buffer)
    (progn
      (send basic-stream :return-header si:stream-input-buffer)
      (setf si:stream-input-buffer nil)))
  (when basic-stream (send basic-stream :close abort-p))
  (setf basic-stream nil))

))

#!C
; From file DAP_STREAM_FLAVOR.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_STREAM_FLAVOR.#"


(defmethod (dap-output-character-mixin :discard-output-buffer) (&rest ignore)
  (when si:stream-output-buffer
    (send basic-stream :return-header si:stream-output-buffer)))

))


#!C
; From file DAP_SERVER.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_SERVER.#"


(defmethod (dap-server-stream :output-close) ()
  (when (eql (send port :state) :RUNNING)
    (send self :set-signal-error t)
    (catch 'dap-catch-errors
      (setf (send port :who-line-state) "Close")
      (send self :server-send-error 5 #o47)
      (unwind-protect
	  (let ((func (send self :process-access-complete-response)))
	    (send self :send-access-complete-message cmpfunc-response)
	    (when (eql func cmpfunc-end-of-stream)
	      (send self :process-access-complete-response)
	      (send self :send-access-complete-message cmpfunc-close)))
	(when current-write-header
	  (send self :return-header current-write-header))
	(when current-read-header
	  (send self :return-header current-read-header))
	(send self :clear-receive-queue))
      )))

))

#!C
; From file DAP_UTILITIES.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_UTILITIES.#"


(defmethod (DAP-basic-stream :close) (&optional abort)
  (when (eql dap-internal-state :OPEN)
    (setf dap-internal-state :CLOSED)	   ;This is to prevent multiple closes to a stream.
    (when (eql (send port :state) :RUNNING)
      (send self :set-signal-error t)
      (unwind-protect
	  (catch 'dap-catch-errors
	    (setf (send port :who-line-state) "Close")	    
	    (send self :send-access-complete-message
		  (if (and (eql current-direction :OUTPUT) abort)
		      cmpfunc-purge cmpfunc-close))
	    (when (eql (send self :process-access-complete-response) 6)
	      (send self :process-access-complete-response)))
	(send self :clear-receive-queue)
	(when current-write-header
	  (send self :return-header current-write-header))
	(when current-read-header
	  (send self :return-header current-read-header))
	(add-port-to-released-list port)))
    ))

))
