;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Adds additional checks for ill-formed packets.  
;;; Prevents NUBUS controller error. 
;;; Fixes [SPR 8190] Code readers: CRW RAF.
;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;; Written 05/24/88 14:31:29 by WALTON,
;;; while running on CHARLIE from band LODA
;;; With SYSTEM 4.64, VIRTUAL-MEMORY 4.4, EH 4.5, MAKE-SYSTEM 4.5, MICRONET 4.5, Experimental LOCAL-FILE 4.2,
;;;  BASIC-PATHNAME 4.13, NETWORK-SUPPORT-COLD 4.1, NAMESPACE 4.22, NETWORK-NAMESPACE 4.3,
;;;  DISK-IO 4.12, DISK-LABEL 4.0, BASIC-FILE 4.11, MAC-PATHNAME 4.7, NETWORK-PATHNAME 4.1,
;;;  COMPILER 4.12, TV 4.87, DATALINK 4.15, CHAOSNET 4.19, GC 4.3, MEMORY-AUX 4.1,
;;;  NVRAM 4.6, SYSLOG 4.0, Experimental STREAMER-TAPE 4.5, UCL 4.1, INPUT-EDITOR 4.0,
;;;  METER 4.3, ZWEI 4.21, DEBUG-TOOLS 4.2, NETWORK-SUPPORT 4.6, NETWORK-SERVICE 4.0,
;;;  DATALINK-DISPLAYS 4.0, FONT-EDITOR 4.0, SERIAL 4.0, PRINTER 4.9, PRINTER-TYPES 4.4,
;;;  IMAGEN 4.0, SUGGESTIONS 4.0, MAIL-DAEMON 4.9, MAIL-READER 4.9, TELNET 4.1, VT100 4.5,
;;;  NAMESPACE-EDITOR 4.5, PROFILE 4.4, VISIDOC 4.4, IP 3.19, Experimental BUG 10.1,
;;;  DECNET 1.34, KERMIT 6.0,  microcode 534, Band Name: Rel 4.1 + IP 5/9

#!C
; From file ROUTING.LISP#> DECNET; DECNET:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "DECNET: DECNET; ROUTING.#"


(defun ROUTING-RECEIVER (circuit data-array nbytes type src-addr dst-addr
			 &optional (free-buffer t))
  "Receive a DECnet packet from the datalink handler.  This function must never block, go into the error
    handler, or take excessive CPU time.  It is executed in the receiver process for the datalink."
  (declare (special *nsp-object*))
  (unwind-protect
      (when (and *receive-enabled*
		 (not (eql 0 nbytes))
		 (not (null dna:my-addresses)))
	(let* ((INHIBIT-SCHEDULING-FLAG T)
	       (start-time  (sys:%fixnum-microsecond-time))
	       (default-cons-area *decnet-area*)
	       (pkt (send *routing-handler* :allocate-pkt :input))
	       (header-preamble
		 (progn (WITHOUT-INTERRUPTS
			  (SYS:%BLT (SYS:%MAKE-POINTER-OFFSET
				  SYS:DTP-FIX data-array
				  (SI:ARRAY-DATA-OFFSET data-array))
				(SYS:%MAKE-POINTER-OFFSET
				  SYS:DTP-FIX (send pkt :base-array)
				  (SI:ARRAY-DATA-OFFSET
				    (send pkt :base-array)))
				(min (CEILING NBYTES 4)
				     (FLOOR (length (send pkt :base-array)) 4))
				1))
			(unless (<= (* 4 (CEILING NBYTES 4))	   ;rwf 10/31/86
				    (length (send pkt :base-array)))
			  (when *packet-trace-enable*
			    (record-unusual-packet :too-big
						   (send pkt :base-array) :receive nbytes
						   type src-addr dst-addr circuit)))
			(setf (net:packet-time-received     pkt) (time))
			(setf (net:packet-datalink-src-addr pkt) src-addr) 
			(setf (net:packet-datalink-dst-addr pkt) dst-addr) 
			(setf (net:packet-datalink-nbytes   pkt)
			      (min nbytes (+ 2 (FLAGS-PACKET-SIZE (send pkt :base-array)))))
			(setf (net:packet-datalink-type     pkt) type)
			(setf (net:packet-datalink-circuit  pkt) circuit)
			(net:get-datalink-header
			  pkt :pre-routing
			  (DNA:ROUTE-HEADER-OFFSET (send pkt :base-array)))))
	       (datalink-data (net:get-datalink-data
				pkt (min (- nbytes (length header-preamble))
					 (flags-packet-size header-preamble))
				)
			      )
	       (header        (net:get-network-header
				pkt
				(DNA:PARSE-ROUTE-MESSAGE-TYPE  datalink-data)
				(DNA:PARSE-ROUTE-HEADER-LENGTH datalink-data)))
	       (data          (net:get-network-data
				pkt (min (- nbytes
					    (length header-preamble)
					    (length header))
					 (- (flags-packet-size header-preamble)
					    (length header)))
				)
			      )
	       )

	      (if (prog1 (or (null header)(null data)(zerop (length header))) ;CRW 5-24-88
			 (when *packet-trace-enable*
			   (record-packet (send pkt :base-array) :receive (net:packet-datalink-nbytes pkt)
					  type src-addr dst-addr circuit)))
		  (incf (routing-handler-ill-formed-packets-received *routing-handler*)) 
		  (progn
		    (unless (member (area (src-node header)) dna:my-areas)
		      (incf (routing-handler-packets-from-other-area *routing-handler*))
		      (when *packet-trace-enable*
			(record-unusual-packet :foreign-area
					       (send pkt :base-array) :receive nbytes
					       type src-addr dst-addr circuit))
		      )
		    (incf (routing-handler-packets-received        *routing-handler*))
		    
		    
		    (when (< 2 (length header-preamble))
		      (incf (routing-handler-padded-packets-received *routing-handler*)))
		    (incf (routing-handler-bytes-received   *routing-handler*) nbytes)
		    (let* ((source (src-node header)))
		      (when source
			(let* ((entry (LOOKUP-NODE-COUNTERS source)))
			  (when entry 
			    (case (net:get-network-header-kind pkt)
				  (:ENDNODE-HELLO-MESSAGE
				   (incf (node-counters-ENDNODE-HELLO-packets-received   entry)))
				  (:ROUTER-HELLO-MESSAGE
				   (incf (node-counters-ROUTER-HELLO-packets-received    entry)))   
				  (:LEVEL-2-ROUTING-MESSAGE
				   (incf (node-counters-level-2-routing-packets-received entry)))
				  (:LEVEL-1-ROUTING-MESSAGE
				   (incf (node-counters-level-1-routing-packets-received entry)))
				  (:HELLO-AND-TEST-MESSAGE
				   (incf (node-counters-hello-and-test-packets-received  entry)))
				  (:ROUTING
		                   (setf (node-counters-last-receive-time entry) (get-universal-time))
				   (incf (node-counters-long-route-packets-received      entry))
				   (incf (node-counters-packets-received entry))
				   (incf (node-counters-bytes-received   entry) nbytes)
				   )
				  (otherwise
				   (setf (node-counters-last-receive-time entry) (get-universal-time))
				   (incf (node-counters-unknown-type-packets-received    entry))
				   (incf (node-counters-packets-received entry))
				   (incf (node-counters-bytes-received   entry) nbytes)
				   )
				  )	   ;case
			    )		   ;when
			  )		   ;let*
			)		   ;when
		      )			   ;let*
		    (case (net:get-network-header-kind pkt)
			  (:ENDNODE-HELLO-MESSAGE (process-endnode-hello pkt dst-addr))
			  (:ROUTER-HELLO-MESSAGE  (process-router-hello  pkt dst-addr))   
			  (:LEVEL-2-ROUTING-MESSAGE
			   (incf (routing-handler-level-2-routing-packets-received *routing-handler*))
			   (send *routing-handler* :deallocate-pkt pkt))
			  (:LEVEL-1-ROUTING-MESSAGE
			   (incf (routing-handler-level-1-routing-packets-received *routing-handler*))
			   (send *routing-handler* :deallocate-pkt pkt))
			  (:HELLO-AND-TEST-MESSAGE
			   (incf (routing-handler-hello-and-test-packets-received  *routing-handler*))
			   (send *routing-handler* :deallocate-pkt pkt))
			  (:ROUTING
			   (incf (routing-handler-long-route-packets-received      *routing-handler*))
			   (when (and (not *passive-nsp*)
				      (variable-boundp *nsp-object*)
				      *nsp-object*)
			     (send *nsp-object* :receive-message pkt (src-node header))))
			  (otherwise
			   (incf (routing-handler-unknown-type-packets-received    *routing-handler*))
			   (when *packet-trace-enable*
			     (record-unusual-packet :unknown-packet-type
						    (send pkt :base-array) :receive nbytes
						    type src-addr dst-addr circuit))
			   (send *nsp-object* :receive-odd-message pkt
				 (net:swap-bytes (logxor src-addr hiord_16))))			   
			  )		   ;case
		    )			   ;else (progn)
		  )			   ;if
	  (incf (routing-handler-time-receiving *routing-handler*)        
		(round (time-difference (sys:%fixnum-microsecond-time) start-time) 1000.)
		)))
	;; clean up resources.
	(when free-buffer
	  (ethernet::deallocate-net-packet type data-array)))
    )

))
