;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1987 Texas Instruments Incorporated. All rights reserved.

;;; Patch file for DECNET version 1.4
;;; Reason: Dap server now checks for record format of ascii-stream to help determine :characters property for opening
;;; output files. Code read by BRG and CRW.
;;; Written 07/08/87 11:35:28 by WALTON,
;;; while running on CHARLIE from band LODA
;;; With IO 3.22, PATHNAME 3.13, FILE 3.8, SYSTEM 3.41, METER 3.1, SERIAL 3.0, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.7, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.7, PROFILE 3.0,
;;;  SUGGESTIONS 3.3, UCL 3.0, ZWEI 3.6, STREAMER-TAPE 3.5, DEBUG-TOOLS 3.1, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.2, NAMESPACE-EDITOR 3.4, NVRAM 3.0,
;;;  TELNET 3.3, TV 3.5, NAMESPACE 3.7, COMPILER 3.12, PRINTER 3.1, SYSLOG 3.0, VT100 3.5,
;;;  IP 2.42, Experimental BUG 4.10, DECNET 1.1,  microcode 355, Band Name: Release 3.0 - 6/17+ip

#!C
; From file dap_server.LISP.NEWEST >REL3-DNA> SA:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; DAP_SERVER.#"


(defmethod (dap-server-stream :server-open-for-input) (path display datatype rfm rat create-p)
  (server-display "~% Open for input called on ~a. Datatype = ~a, rfm = ~a, rat = ~a." path datatype rfm rat)
  (let ((local-stream
	  (open path :direction nil :error nil))
	(any-error t))
    (if (and (errorp local-stream)	   ;Return an error unless file is not found and we want to create it.
	     (not (and create-p
		       (member 'fs:file-not-found (send local-stream :condition-names)))))
	(send self :server-map-error local-stream)
	(unwind-protect
	    (progn
	      (when create-p		   ;Open the file now if it is for a new version, to send proper display attributes.
		(setf local-stream
		      (open path :direction :output :error nil
			    :if-does-not-exist (if create-p :create nil)
			    :if-exists :new-version
			    :byte-size (if (and (eql rfm rfm-fixed) (eql rat rat-blk)) 16. 8)
			    :characters (if (or (and (eql rfm rfm-variable)
						     (eql rat rat-implied-crlf))
						(eql rfm rfm-asciistream))
					    t nil))))
	      (setf end-of-file nil)
	      (send self :process-display display
		                          (if (errorp local-stream) path (send local-stream :truename)) t)
	      (send self :send-acknowledge-message)
	      (multiple-value-bind (ignore ignore rop)
		  (send self :server-process-control-message)	   ;this is the INITIATE message
		(if end-of-file		   ;or it could be an ACCESS COMPLETE if only a file create.
		    (progn
		      (send local-stream :close)
		      (send self :send-acknowledge-message)
		      (send self :send-access-complete-message 7))
		    (progn
		      (unless create-p	   ;If existing file, open now to check for Append or Supersede in ROP.
			(setf local-stream
			      (open path :direction :output :error nil
				    :if-does-not-exist (if create-p :create nil)
				    :if-exists (cond ((test-rop-for-append rop) :append)
						     ((test-rop-for-supersede rop) :supersede)
						     (t (if create-p :new-version
							    (send self :send-not-at-eof-error))))
				    :byte-size (if (and (eql rfm rfm-fixed) (eql rat rat-blk)) 16. 8)
				    :characters (if (or (and (eql rfm rfm-variable)
							     (eql rat rat-implied-crlf))
							(eql rfm rfm-asciistream))
						    t nil))))
		      
		      (send self :send-acknowledge-message)
		      (if rat (when (ldb-test (byte 1 1) rat) (setf (send self :implied-crlf?) t))
			  (setf (send self :implied-crlf?) nil))
		      (let ((server-stream
			      (make-instance
				(if (or (and (eql rfm rfm-variable) ;Character file
					     (eql rat rat-implied-crlf))
					(eql rfm rfm-asciistream))
				    (decnet-server-stream-flavor-type :input 8. t)
				    (if (and (eql rfm rfm-fixed)   ;Binary file
					     (eql rat rat-blk))
					(decnet-server-stream-flavor-type :input 16. nil)
					(decnet-server-stream-flavor-type :input 8. nil))) ;Image
				:pathname (send local-stream :truename)
				:basic-stream self)))
			(send self :server-process-control-message)	   ;this is the PUT message
			(condition-case (condition)
			    (stream-copy-until-eof server-stream local-stream)
			  ('error (send self :server-map-error condition :during)))
			(send server-stream :close)	   ;Send this to the "new" stream
			(send self :input-close  :access-complete-code)	   ;Send this to ourself.
			(setf any-error nil))))))
	  (send local-stream :close any-error))
	  )))
))
