;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Written 11/09/88 14:46:52 by JONES,
;;; Reason: Make Decnet tolerate having a null controller-list. This happens when booting a network band
;;; on a machine without an ethernet board.
;;; while running on HOBBS from band LOD1
;;; With SYSTEM 5.17, GC 5.3, VIRTUAL-MEMORY 5.5, MICRONET 5.5, MICRONET-COMM 5.11,
;;;  DISK-IO 5.8, BASIC-PATHNAME 5.2, MAC-PATHNAME 5.0, NETWORK-SUPPORT-COLD 5.1,
;;;  BASIC-NAMESPACE 5.5, BASIC-FILE 5.3, RPC 5.4, NFS 5.8, EH 5.2, MAKE-SYSTEM 5.1,
;;;  MEMORY-AUX 5.1, MACTOOLBOX 1.20, COMPILER 5.1, TV 5.15, NVRAM 5.1, UCL 5.0, INPUT-EDITOR 5.0,
;;;  METER 5.0, ZWEI 5.8, DEBUG-TOOLS 5.0, WINDOW-MX 5.24, PRINTER 5.10, MAC-PRINTER-TYPES 5.2,
;;;  NETWORK-PATHNAME 5.0, NETWORK-NAMESPACE 5.0, DATALINK 5.6, CHAOSNET 5.4, NETWORK-SUPPORT 5.0,
;;;  NETWORK-SERVICE 5.0, DATALINK-DISPLAYS 5.0, NAMESPACE-EDITOR 5.1, IP 3.32, NFS-SERVER 5.3,
;;;  PRINTER-TYPES 5.2, IMAGEN 5.1, MAIL-DAEMON 5.1, MAIL-READER 5.3, TELNET 5.0,
;;;  VT100 5.0, STREAMER-TAPE 5.6, DECNET 1.40, VISIDOC 5.4, PROFILE 5.1, DISK-LABEL 5.1,
;;;   microcode 92, Band Name: nb04*1

#!C
; From file ROUTING.LISP#> DECNET; MR-X:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :Common-lisp)
                          (*READTABLE* Sys:Common-lisp-readtable)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* Sys::*common-lisp-symbol-substitutions*))
  (COMPILER#:PATCH-SOURCE-FILE "Sys: DECNET; ROUTING.#"


(defun ROUTING-TRANSMIT (output-array nbytes destination &optional (self-receive nil))
  "The transmit routine for DNA Routing."
  (when (and *transmit-enabled*
	     (not (null dna:my-addresses))
	     net:controller-list)
    (let* ((pkt-nbytes  nbytes)
	   (controller  (first net:controller-list))	   ;TEMPORARY!!!!!
	   (Destination-Ether-Address (if (< destination #x10000)
					  (logior HIORD_16 (net:swap-bytes destination))
					  destination))
	   (Source-Ether-Address      (send controller :ethernet-address))
	   (output-word-array nil))
      (when (eql (flags-padded byte-pkt) 1)
	(INCF (routing-handler-padded-packets-sent *ROUTING-HANDLER*)))
      (when *packet-trace-enable*
	(record-packet output-array :transmit pkt-nbytes
		       *ROUTE-ETHERNET-TYPE*
		       Source-Ether-Address
		       Destination-Ether-Address
		       Controller))
      
      (INCF (routing-handler-packets-sent *ROUTING-HANDLER*))
      (INCF (routing-handler-bytes-sent   *ROUTING-HANDLER*) pkt-nbytes)
      
      (when (< destination #x10000)
	(let* ((entry (LOOKUP-NODE-COUNTERS destination)))
	  (when entry 
	    (case (DNA:ROUTE-MESSAGE-TYPE  byte-pkt)
	      (:ENDNODE-HELLO-MESSAGE
	       (incf (node-counters-ENDNODE-HELLO-packets-sent   entry)))
	      (:ROUTER-HELLO-MESSAGE
	       (incf (node-counters-ROUTER-HELLO-packets-sent    entry)))   
	      (:LEVEL-2-ROUTING-MESSAGE
	       (incf (node-counters-level-2-routing-packets-sent entry)))
	      (:LEVEL-1-ROUTING-MESSAGE
	       (incf (node-counters-level-1-routing-packets-sent entry)))
	      (:HELLO-AND-TEST-MESSAGE
	       (incf (node-counters-hello-and-test-packets-sent  entry)))
	      (:ROUTING
	       (setf (node-counters-last-send-time entry) (get-universal-time))
	       (incf (node-counters-long-route-packets-sent      entry))
	       (incf (node-counters-packets-sent entry))
	       (incf (node-counters-bytes-sent   entry) pkt-nbytes)
	       )
	      (otherwise
	       (setf (node-counters-last-send-time entry) (get-universal-time))
	       (incf (node-counters-packets-sent entry))
	       (incf (node-counters-bytes-sent   entry) pkt-nbytes)
	       )
	      )
	    )
	  )
	)
      (without-interrupts
	(setq output-word-array (if (= 16. (si:array-element-size output-array))
				    output-array       
				    (when (SI:ARRAY-INDIRECT-P output-array)
				      (ferror nil "~& Illegal (indirect) input to net."))
				    (assert (<= (* 4 (CEILING pkt-nbytes 4))	   ;rwf 10/31/86
						(* 2 (length real-word-pkt))))
				    
				    (SYS:%BLT (SYS:%MAKE-POINTER-OFFSET
						SYS:DTP-FIX
						output-array
						(SI:ARRAY-DATA-OFFSET output-array))
					      (SYS:%MAKE-POINTER-OFFSET
						SYS:DTP-FIX
						real-word-pkt
						(SI:ARRAY-DATA-OFFSET real-word-pkt))
					      (CEILING pkt-nbytes 4)
					      1)
				    real-word-pkt))
	
	(unless (MEMBER destination dna:my-addresses)
	  (SEND controller :TRANSMIT-FAST
		*ROUTE-ETHERNET-TYPE*
		Destination-Ether-Address
		output-word-array
		pkt-nbytes)
	  )
	
	(when (or (MEMBER destination dna:my-addresses) self-receive)
	  (ROUTING-RECEIVER
	    controller
	    output-word-array
	    pkt-nbytes
	    *ROUTE-ETHERNET-TYPE*
	    Source-Ether-Address      
	    Destination-Ether-Address 
	    nil)
	  (process-allow-schedule))
	)
      )					   ;when
    )					   ;let
  )

(defun ROUTING-RESET (&key (enable t) (hard t))
  "Put the DECnet module into a (hopefully) runnable state.  Note that this host's ETHERNET
    address may change if this is the first time that the machine was reset with DECnet installed."
  (when net:controller-list
    (LET* ((newdriverp t)
	   (c (CAR net:controller-list))
	   (enable-p (SEND c :enabled)))
      (dna:disable)
      (dna:reset-on-ethernet-cache)
      (dna:reset-node-counters)
      (WHEN (si:mx-p)
	(SEND c :disable)
	(send c :set-eaddr (+ hiord_16 (net:swap-bytes my-address)))
	(send c :close)
	(SEND c :initialize)
	(WHEN enable-p
	  (SEND c :enable))
	(setf newdriverp       
	      (equalp (send c :read-eadr)
		      (send c :read-rom-id))))
      (if newdriverp
	  (progn
	    (dna:setup-my-address)		       ;possibly obsolete ethernet address
	    (if hard
		(progn
		  (ethernet:add-protocol dna:*ROUTE-ETHERNET-TYPE* 'Routing-Receiver)
		  (Ethernet:Add-Net-Packet-Allocator
		    dna:*ROUTE-ETHERNET-TYPE* 'CHAOS::ALLOCATE-INT-PKT 'CHAOS::FREE-INT-PKT)
		  (net:reset-packets)
		  (setf *ROUTING-HANDLER* (make-instance 'routing-handler))
		  (net:send-all-controllers :disable)  ;need to permanently flush old controllers using...
		  (Dolist (Cont Net:Controller-List)
		    (Setup-Ethernet-Address Cont))
		  (ethernet:reset-address-translations)	       ;force rebroadcasts of our (new) ethernet address
		  (net:send-all-controllers :reset)
		  (net:send-all-controllers :enable))	
		(send *routing-handler* :reset))
	    (when (and my-address (not (eql 0 my-address)))
	      (set-host-addresses)
	      (reset-hello-xmitter))
	    (when enable
	      (dna:enable))
	    )
	  (tv:notify nil " *** WARNING ***   Unable to initialize or reset DECnet. ~%          You must first install EtherTalk version 1.1")
	  )
      )))					   ;defun

))
