;;; -*- Mode: Common-Lisp; Package: DNA; Base: 10.; Patch-File: T -*-

;;; Reason: Changed PARSE-DNA-ADDRESS-SPEC to use :fully-qualified-name to instead of name to parse address spec.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/03/89 08:21:35 by BERGER,
;;; while running on ARIES from band LOD1
;;; With SYSTEM 5.24, VIRTUAL-MEMORY 5.5, EH 5.3, MICRONET 5.5, LOCAL-FILE 5.0, BASIC-PATHNAME 5.4,
;;;  NETWORK-SUPPORT-COLD 5.1, BASIC-NAMESPACE 5.8, NETWORK-NAMESPACE 5.0, DISK-IO 5.9,
;;;  DISK-LABEL 5.1, BASIC-FILE 5.4, MAC-PATHNAME 5.0, NETWORK-PATHNAME 5.0, COMPILER 5.2,
;;;  MAKE-SYSTEM 5.3, TV 5.24, DATALINK 5.7, CHAOSNET 5.6, GC 5.4, MEMORY-AUX 5.1,
;;;  NVRAM 5.1, SYSLOG 5.2, STREAMER-TAPE 5.7, UCL 5.0, INPUT-EDITOR 5.0, METER 5.1,
;;;  ZWEI 5.12, DEBUG-TOOLS 5.1, Experimental NETWORK-SUPPORT 5.1, Experimental NETWORK-SERVICE 5.1,
;;;  DATALINK-DISPLAYS 5.0, FONT-EDITOR 5.0, SERIAL 5.0, PRINTER 5.12, PRINTER-TYPES 5.5,
;;;  IMAGEN 5.2, SUGGESTIONS 5.0, MAIL-DAEMON 5.5, MAIL-READER 5.6, TELNET 5.2, VT100 5.1,
;;;  Experimental NAMESPACE-EDITOR 5.1, PROFILE 5.1, VISIDOC 5.4, Experimental BUG 10.6,
;;;  IP 3.41, VISIDOC-SERVER 5.2, DECNET 1.47,  microcode 413, Band Name: Pre 5.0 +1p 9/23 +visidoc

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(Defun PARSE-DNA-ADDRESS-SPEC (address-spec &optional (check-tag nil) &aux (start 0) (address 0))
  "Returns the address corresponding to address-spec.  If check-tag is T, this must be a string beginning with
   'IP|'.  Otherwise a number is also accepted (and simply returned).  Nil is returned if check tag fails.
   An error is signalled if the string is formatted incorrectly." 
  (if address-spec
      (COND
	((SYMBOLP address-spec) (parse-dna-address-spec (dna:map-name-to-node (string address-spec))))
	
	((TYPEP address-spec 'net:host) (parse-dna-address-spec
					  (dna:map-name-to-node
					    (send address-spec :fully-qualified-name)))) ; DAB 02-03-89
	
	((FLOATP address-spec) (parse-dna-address-spec (format nil "~a" address-spec) check-tag))
	
	((NUMBERP address-spec) (UNLESS check-tag address-spec))
	
	((STRINGP address-spec)
	 (WHEN (STRING-EQUAL address-spec "DECnet|" :start1 0 :end1 7 :start2 0 :end2 7) (SETQ start 7))
	 
	 (WHEN (STRING-EQUAL address-spec "DNA|" :start1 0 :end1 4 :start2 0 :end2 4)    (SETQ start 4))
	 
	 (WHEN (OR (> start 0) (NOT check-tag))
	   (SETQ address (COND 
			   ;; form = #
			   ((EQL (AREF address-spec start) #\#)
			    (CASE (aref address-spec (1+ start))
			      ((#\X #\x) (PARSE-INTEGER address-spec :start (+ start 2) :radix 16))
			      ((#\O #\o) (PARSE-INTEGER address-spec :start (+ start 2) :radix 8))
			      ((#\B #\b) (PARSE-INTEGER address-spec :start (+ start 2) :radix 2))))
			   
			   ;; form = d.d
			   ((SEARCH "." address-spec) (parse-dna-byte-spec address-spec start))
			   
			   ;; form = number (decimal)
			   (t (PARSE-INTEGER address-spec :start start :radix 10 :junk-allowed t))))
	   
	   (COND ((or address (parse-dna-address-spec (dna:map-name-to-node (string address-spec)))))
		 (t (FERROR nil "Illegal DECnet address ~A" address-spec)))))	   
	
	(t (UNLESS check-tag (FERROR nil "Illegal DECnet address: ~A" address-spec))))
      (FERROR nil "Illegal DECnet address: ~A" address-spec))
  
  )
))
