;;; -*- Mode: Common-Lisp; Package: DNA; Base: 10.; Patch-File: T -*-

;;; Reason: Add support for dormant nsp port GC process. This will remove idle connections. HOST-UNIT-LIFETIME defines how long a port connection can be idle before disconnecting. DORMANT-NSP-CONNECTION-GC-WAIT-INTERVAL determines the wakeup interval of the bakcground process that will check for idle connections.  *** A DNA:RESET is done at the end of this patch.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/15/89 13:57:56 by BERGER,
;;; while running on ARIES from band LOD1
;;; With Experimental SYSTEM 6.0, Experimental VIRTUAL-MEMORY 6.0, Experimental EH 6.0,
;;;  Experimental MAKE-SYSTEM 6.0, Experimental MICRONET 6.0, Experimental LOCAL-FILE 6.0,
;;;  Experimental BASIC-PATHNAME 6.0, Experimental NETWORK-SUPPORT-COLD 6.0, Experimental BASIC-NAMESPACE 6.0,
;;;  Experimental NETWORK-NAMESPACE 6.0, Experimental DISK-IO 6.0, Experimental DISK-LABEL 6.0,
;;;  Experimental BASIC-FILE 6.0, Experimental MAC-PATHNAME 6.0, Experimental NETWORK-PATHNAME 6.0,
;;;  Experimental COMPILER 6.0, Experimental TV 6.0, Experimental DATALINK 6.0, Experimental CHAOSNET 6.0,
;;;  Experimental GC 6.0, Experimental MEMORY-AUX 6.0, Experimental NVRAM 6.0, Experimental SYSLOG 6.0,
;;;  Experimental STREAMER-TAPE 6.0, Experimental CLEH 1.0, Experimental UCL 6.0,
;;;  Experimental INPUT-EDITOR 6.0, Experimental METER 6.0, Experimental ZWEI 6.0,
;;;  Experimental DEBUG-TOOLS 6.0, Experimental NETWORK-SUPPORT 6.0, Experimental NETWORK-SERVICE 6.0,
;;;  DATALINK-DISPLAYS 6.0, Experimental FONT-EDITOR 6.0, Experimental SERIAL 6.0,
;;;  Experimental PRINTER 6.0, Experimental MAC-PRINTER-TYPES 6.0, Experimental PRINTER-TYPES 6.0,
;;;  Experimental IMAGEN 6.0, Experimental SUGGESTIONS 6.0, Experimental MAIL-DAEMON 6.0,
;;;  Experimental MAIL-READER 6.0, Experimental TELNET 6.0, Experimental VT100 6.0,
;;;  Experimental NAMESPACE-EDITOR 6.0, Experimental PROFILE 6.0, VISIDOC 6.0, Experimental BUG 10.6,
;;;  IP 3.40, Experimental TI-CLOS 14.16, Experimental CLX 3.0, Experimental X11M 1.0,
;;;  Experimental CLUE 15.0, RPC 4.19, NFS 3.4, DECNET 1.49,  microcode 414, Band Name: REL 6.0 2/2 +DNA

#!C
; From file NSP_NSP.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; NSP_NSP.#"


(defparameter host-unit-lifetime 72000 "Time in 60th's of a second for a NSP server connection 
                                          to remain idle before it is closed.")    ;20 MINUTES.
	
(defvar dormant-nsp-connection-gc-wait-interval (* 30 3600)
  "Time in 60th's of a second for a NSP server connection process
   to remain idle before it rechecks for idle server connections.") ; DAB 02-15-89 30 Minutes.

(defvar dormant-nsp-gc-process nil) 

(add-initialization "Kill dormant-nsp-gc-process" '(init-dormant-nsp-gc-process nil)
		    '(:before-cold :full-gc)) 

(defparameter *debug-NSP-connection-gc* nil "Don't signal an error if this Var is either nil or :notify.")

(defun dormant-NSP-connection-gc-top-level ()
  (loop
    (process-sleep dormant-nsp-connection-gc-wait-interval "DNA GC Sleep")
    (condition-case-if  (not *debug-NSP-connection-gc*)
			(condition) 
	(reset-dormant-nsp-port)
      (error
       (when (eq *debug-NSP-connection-gc* :notify)
	 (tv:notify nil "Error from NSP connection GC: ~s" (send condition :report-string)))))))

(defun reset-dormant-nsp-port ()
  (let ((dormant-port-list 
	  (loop with dormant-list = ()
		for used-port first (send dna:*nsp-object* :allocated-port-list)
		then (send used-port :next)	   
		until (null used-port)
		do (when (and (eql (send used-port :state) :running)
			      (member (send used-port :who-line-state) '("Directory" "Close")
				      :test #'string-equal)
			      (> (time-difference (time)
						 (send used-port :last-activity-time))
				 host-unit-lifetime ))
		     (push used-port dormant-list))
		finally (return dormant-list))))
    (dolist (used-port dormant-port-list)
      (send used-port :close))
    ))

(defun init-dormant-nsp-gc-process (&optional (enable-p t))
  (cond ((and enable-p dormant-nsp-gc-process)
	 (process-reset dormant-nsp-gc-process))
	
	(enable-p
	 (setf dormant-nsp-gc-process
	       (process-run-function '(:name "Dormant DNA connection GC"
					     :restart-after-reset t
					     :restart-after-boot t
					     :priority 0)
				     'dormant-NSP-connection-gc-top-level)))
	
	(dormant-NSP-gc-process
	 (send dormant-NSP-gc-process :kill)
	 (setf dormant-NSP-gc-process nil))))


(init-dormant-nsp-gc-process)
  

))

#!C
; From file NSP_PORTS.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; NSP_PORTS.#"


(defflavor NSP-port
	   (NEXT			   ;Pointer to next linked port
	    STATE		           ;State of the port
	    NODEaddr			   ;Remote node address
	    ADDRloc			   ;Local link address
	    ADDRrem			   ;Remote link address
	    Port-packets-retransmitted	   ;Number of packet retransmissions
	    NUMdat 			   ;Number of next DATA SEGMENT message to transmit.
	    NUMdat-remote		   ;Number of DATA SEGMENT message acknowledged by remote host.
	    NUMoth 			   ;Number of next INTERRUPT or LINK SERVICE message to transmit.
	    NUMoth-remote		   ;Number of highest OTHER DATA message acknowledged by remote host.
	    NUMdata-received		   ;Highest numbered DATA SEGMENT message received from remote NSP.
	    NUMoth-received                ;Highest numbered OHTER DATA segment received from remote NSPl
	    ACKxmt-dat		           ;Number of last DATA SEGMENT message ack sent by local NSP.
	    ACKxmt-oth		           ;Number of last Interrupt or Link service message ack sent by local NSP
	    ACKrcv-dat	        	   ;Number of highest DATA SEGMENT message received from remote NSP
	    FLOWrem-data		   ;The cumulative normal data request count received from remote NSP
	    FLOWrem-int 		   ;Cumulative interrupt data request count received from the remote NSP
	    FLOWrem-sw 		           ;Normal data on/off switch most recently received from remote NSP :SEND :NO-SEND
	    FLOWrem-type 	           ;Flow control type of the remote receiver :NONE :SEGMENT :SESSION-CONTROL
	    BUFxmt 		           ;Data to be transmitted with CONNECT CONFIRM, DISCONNECT INITIATE or INTERRUPT
	    BUFrcv 		           ;Buffer to contain data for received INTERRUPT msgs.
	    BUFdis			   ;Buffer to contain DISCONNECT INITIATE message.
	    BUFacc 		           ;Contains accept data for received CONNECT CONFIRM msg.
	    BUFcon 		           ;Buffer to contain data from a transmitted or received CONNECT INITIATE msg.
	    Ack-Window-size		   ;Round trip time-of-day value for start of round trip time estimation.
	    BOM-packet			   ;If this message does NOT contain an EOM.
	    Who-line-state		   ;Who line message.
	    SIZEseg 			   ;Transmit segment size
	    RETRANSMISSION-QUEUE 	   ;Queue for holding normal packets for retransmission
	    RECEIVE-PACKET-QUEUE           ;Queue for received data packets
	    QUEUED-OTHER-DATA-MESSAGES	   ;Queue for interrupt data messages
	    REASON-FOR-USE		   ;If host or server.
	    LOCAL-WINDOW-SIZE		   ;Number of received messages to queue before sending a "halt" message
	    remote-window-size		   ;Number of messages to send before waiting for an ack.
	    average-time-between-packets   ;Raw average between ANY packet arrivals.
	    last-packet-received-time	   ;time (in 60ths) when last packet arrived.
	    halt-time			   ;If we have sent a HALT link service request. (cleared by sending a RESUME)
	    try-sending 		   ;Set T if we receive a LS message - used if we "miss" a data ack.
	    preallocated-ack-pkt
	    last-activity-time		   ;A timestamp of activity except link services DAB 02-15-89
	    )
	    ()
           (:ordered-instance-variables)
           (:settable-instance-variables)
           (:documentation "NSP Port flavor"))


(defmethod (nsp-port :init) (&rest ignore)
  (declare (special *nsp-window-size*))
  (setf NEXT nil)
  (setf STATE :CLOSED)
  (setf NODEaddr nil)	
  (setf ADDRloc 0)	
  (setf ADDRrem 0)
  (setf NUMdat 1)			   
  (setf NUMdat-remote 0)
  (setf NUMoth 1)			   
  (setf NUMoth-remote 0)
  (setf NUMoth-received  0)		   
  (setf NUMdata-received 0)
  (setf ACKxmt-dat 0)		
  (setf ACKxmt-oth 0)		
  (setf ACKrcv-dat 0)		
  (setf FLOWrem-data 0)	
  (setf FLOWrem-int 1)	
  (setf FLOWrem-sw :SEND)	
  (setf FLOWrem-type :NONE)	
  (setf BUFxmt nil)		
  (setf BUFrcv nil)
  (setf BUFdis nil)
  (setf BUFacc nil)		
  (setf BUFcon nil)		
  (setf BOM-packet nil)	
  (setf SIZEseg 0)		
  (setf RETRANSMISSION-QUEUE nil)
  (setf RECEIVE-PACKET-QUEUE nil)
  (setf QUEUED-OTHER-DATA-MESSAGES nil)
  (setf Who-line-state "DNA Connect")
  (setf port-packets-retransmitted 0)
  (setf local-window-size *nsp-window-size*)
  (setf remote-window-size *nsp-remote-window-size*)
  (setf average-time-between-packets 0.)
  (setf last-packet-received-time 0.)
  (setf halt-time nil)
  (setf try-sending nil)
  (setf preallocated-ack-pkt nil)
  (setf last-activity-time  (time:time)) ; DAB 02-15-89
  )

))

#!C
; From file NSP_TRANSMIT.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; NSP_TRANSMIT.#"


(defmethod (NSP-port :transmit-buffer) (pkt &optional (node-id NODEaddr)
					(queue-for-retransmit-p t) (return-pkt t)
					(update-last-activity-time T) ; DAB 02-15-89
					)
  (declare (special *routing-handler* *trace-window*))
  ;;; DAB 02-15-89 last-activity-time is used by reset-dormant-nsp-port to determined idle connections.
  (when update-last-activity-time ; DAB 02-15-89 
    (setf last-activity-time (time:time)))
  (send *routing-handler* :transmit-pkt pkt node-id)
  (if queue-for-retransmit-p
      (send self :queue-for-retransmission pkt 1)
      (when return-pkt
	(nsp-return-pkt pkt))))




(defmethod (NSP-port :send-other-data-ack-message) (&optional (ack-p t))
  (let* ((pkt (nsp-get-pkt))
	 (header (get-nsp-header
		  pkt
		  :NSP
		  (if ack-p nsp-other-data-ack-overhead-size 7.))))
    (set-msgflag-send-other-data-ack header)
    (setf (dstaddr header) ADDRrem)
    (setf (srcaddr header) ADDRloc)
    (setf (ACKNUM-number header) NUMoth-received)
    (setf (ACKNUM-qual header) qual-ack)
    (set-acknum-presence header t)
    (when ack-p
      (setf (ACKOTH-number header) NUMdata-received)
      (setf (ACKOTH-qual header) cross-subchannel-ack)
      (set-ackoth-presence header t))
    (nsp-display "~%TRANSMIT other data ack message. Ack number= ~d" NUMoth-received)
    (send self :transmit-buffer pkt NODEaddr nil T nil)	;Do not queue for retransmission ; DAB 02-15-89
    (incf Other-data-Acknowledgment-message-transmitted)))

(defmethod (nsp-port :send-fast-other-data-ack-message) ()
  (let ((header (get-nsp-header
		  (if preallocated-ack-pkt preallocated-ack-pkt
		      (setf preallocated-ack-pkt (nsp-get-pkt)))
		  :NSP
		  7)))
    (set-msgflag-send-other-data-ack header)
    (setf (dstaddr header) ADDRrem)
    (setf (srcaddr header) ADDRloc)
    (setf (ACKNUM-number header)  NUMoth-received)
    (setf (ACKNUM-qual header) qual-ack)
    (set-acknum-presence header t)
    (nsp-display "~%TRANSMIT FAST other data ack message. Ack num = ~d." NUMdata-received)
    (send self :transmit-buffer preallocated-ack-pkt NODEaddr nil nil
	  nil)	   ;Do not queue for retransmission  ; DAB 02-15-89
    (incf Other-Data-Acknowledgment-message-transmitted)))


(defmethod (NSP-port :send-ls-message) (pkt ack-p FCMOD-value FCVAL-INT-value
					FCVAL-value &optional (queue-p nil))
  (let ((header (get-nsp-header
		  pkt
		  :NSP
		  (if ack-p nsp-link-service-overhead-size nsp-link-service-less-optional-overhead-size))))
    (set-msgflag-link-service header)
    (setf (dstaddr header) ADDRrem)
    (setf (srcaddr header) ADDRloc)
    (when ACK-p 
      (setf (acknum-number header) NUMoth-received)
      (setf (acknum-qual header) qual-ack))
    (set-acknum-presence header ACK-p)
    (when ACK-p
      (setf (ackoth-number header) NUMdata-received)
      (setf (ackoth-qual header) cross-subchannel-ack)
      (set-ackoth-presence header ACK-p))
    (setf (segnum header) (+segnum 1 NUMoth-remote))
    (setf (FC-MOD header) FCMOD-value)
    (setf (FCVAL-INT header) FCVAL-INT-value)
    (setf (FCVAL header) FCVAL-value)
    (nsp-display "~%TRANSMIT ~[data request ~;interrupt-request~] ~
                             ~[NO-CHANGE ~;data HALT ~;data REQUEST~] pkt. Segnum = ~d."
		              FCVAL-value FCMOD-value (1+ NUMoth-remote))
    (send self :transmit-buffer pkt NODEaddr queue-p T nil) ; DAB 02-15-89
    (incf Link-service-message-transmitted)))

(DNA:RESET T)


))
