;;; -*- Mode: Common-Lisp; Package: DNA; Base: 10.; Patch-File: T -*-

;;; Reason: DAP-connect would not update the correct host object if a bad userid and password was entered when there are more host object for a given decent node.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for DECNET version 1.52
;;; Written 02/15/89 14:13:17 by BERGER,
;;; while running on ARIES from band LOD1
;;; With Experimental SYSTEM 6.0, Experimental VIRTUAL-MEMORY 6.0, Experimental EH 6.0,
;;;  Experimental MAKE-SYSTEM 6.0, Experimental MICRONET 6.0, Experimental LOCAL-FILE 6.0,
;;;  Experimental BASIC-PATHNAME 6.0, Experimental NETWORK-SUPPORT-COLD 6.0, Experimental BASIC-NAMESPACE 6.0,
;;;  Experimental NETWORK-NAMESPACE 6.0, Experimental DISK-IO 6.0, Experimental DISK-LABEL 6.0,
;;;  Experimental BASIC-FILE 6.0, Experimental MAC-PATHNAME 6.0, Experimental NETWORK-PATHNAME 6.0,
;;;  Experimental COMPILER 6.0, Experimental TV 6.0, Experimental DATALINK 6.0, Experimental CHAOSNET 6.0,
;;;  Experimental GC 6.0, Experimental MEMORY-AUX 6.0, Experimental NVRAM 6.0, Experimental SYSLOG 6.0,
;;;  Experimental STREAMER-TAPE 6.0, Experimental CLEH 1.0, Experimental UCL 6.0,
;;;  Experimental INPUT-EDITOR 6.0, Experimental METER 6.0, Experimental ZWEI 6.0,
;;;  Experimental DEBUG-TOOLS 6.0, Experimental NETWORK-SUPPORT 6.0, Experimental NETWORK-SERVICE 6.0,
;;;  DATALINK-DISPLAYS 6.0, Experimental FONT-EDITOR 6.0, Experimental SERIAL 6.0,
;;;  Experimental PRINTER 6.0, Experimental MAC-PRINTER-TYPES 6.0, Experimental PRINTER-TYPES 6.0,
;;;  Experimental IMAGEN 6.0, Experimental SUGGESTIONS 6.0, Experimental MAIL-DAEMON 6.0,
;;;  Experimental MAIL-READER 6.0, Experimental TELNET 6.0, Experimental VT100 6.0,
;;;  Experimental NAMESPACE-EDITOR 6.0, Experimental PROFILE 6.0, VISIDOC 6.0, Experimental BUG 10.6,
;;;  IP 3.40, Experimental TI-CLOS 14.16, Experimental CLX 3.0, Experimental X11M 1.0,
;;;  Experimental CLUE 15.0, RPC 4.19, NFS 3.4, DECNET 1.49,  microcode 414, Band Name: REL 6.0 2/2 +DNA

#!C
; From file SC_HOST.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; SC_HOST.#"


(defun sc-establish-connection
	   (&key dstnode nbytes format name usrcode
	    grpcode descrpt userid password accounting objtype
	    dstnode-host-object) ; DAB 02-13-89
  (declare (special *nsp-object*))
  (let ((port nil)
	(retry-p t)) ; first try has to be t
    (loop      
      (if retry-p
	  (let* ((reject-pkt nil)(reject-reason nil)
		 (pkt (allocate-special-pkt
			    :kind :CONNECT :len nbytes :format format :name name
			    :usrcode usrcode :grpcode grpcode :descrpt descrpt
			    :rqstrid userid :password password
			    :account accounting :objtyp objtype)))
	    
	    (setf port (send *NSP-object* :connect-xmt dstnode pkt))
	    
	    (if (send port :wait-for-connect-accept)
		(setf retry-p nil)	                ;don't need to retry - have succeeded
		(setf reject-pkt (send port :get-reject-data)))
	    
	    ; handle the case of no response from the other end
	    ; give the caller a second chance
	    (when (and retry-p (null reject-pkt))       ;failed on first try, but no reject packet was received
	      (if (send port :wait-for-connect-accept)
		  (setf retry-p nil)	                ; succeeded on second try
		  (progn
		    (setf reject-pkt (send port :get-reject-data))
		    (send port :close)
		    (when (null reject-pkt)               ; host did not respond
		      (ferror 'dna:host-not-responding-during-connection "Host ~A not responding."
			       dstnode)))))
	    (when reject-pkt
	      (progn
		    (send port :close)
		    (setf reject-reason (get-two-byte-field (get-session-header reject-pkt :DNA-SESSION 2) 0))
		    (return-pkt reject-pkt)
		    (when reject-reason
		      (cond
			((zerop reject-reason))
			((= reject-reason inv-dst-name)
			 (cerror  "Tries the connection again with a different end user name."
				  "Connection rejected because ~s is an invalid end user name."
				  dstnode)
			 (setf dstnode (prompt-and-read :string-trim " Enter a new value:")))
			
			((= reject-reason inv-pwd-rqstid)
			 (multiple-value-setq (userid password)
			   (fs:file-get-password-etc
			     (or dstnode-host-object (sc-get-DECnet-name dstnode)) ; DAB 02-13-89
			     userid nil t)))
			
			((= reject-reason inv-acct)
			 (cerror "Tries the connection again with a different account number."
				 "Connection rejected because of an invalid account number or balance.")
			 (setf accounting (prompt-and-read :string-trim " Enter a new account number:")))
			
			(t (multiple-value-bind (signal-name signal-string)(map-reject-to-signal reject-reason)
			     (sc-error signal-name signal-string))))))	    
		))     
	  (return nil)))
    port))

))



#!C
; From file DAP_INTERFACE.LISP#> DECNET; sys:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "sys: DECNET; DAP_INTERFACE.#"


(defun dap-connect (dstnode &key userid password accounting dstnode-host-object ) ; DAB 02-13-89
  (let ((nbytes 3)(src-format 2)
	(port (reuse-port dstnode)) (send-config-msg nil))
    (unless port
      (loop 
	(setf port (sc-establish-connection :dstnode dstnode :nbytes nbytes
			 :format src-format :usrcode 0 :grpcode 0 :descrpt userid
			 :userid userid :password password :accounting accounting
			 :objtype dap-connect-objtype :dstnode-host-object dstnode-host-object)) ; DAB 02-13-89
	
	(if (or (not port) (not (eql (send port :STATE) :RUNNING)))
	    (progn
	      (when port (send port :close))  ;10-27-88 DAB If sc-establish-connection fails port will be nil.
	      (setf port nil)
	      (decnet-signal-error 'dna-connect-error "failed to connect"
				   (if (typep dstnode 'net:host)
				       (send dstnode :name)
				       (send (or dstnode-host-object
						 (si:get-host-from-address dstnode :dna))
					     :name))
				   t t))
	    (progn
	      (setf send-config-msg t)
	      (setf (send port :who-line-state) "DAP Connect")
	      (return)))))
    (values port send-config-msg)))

(defun directory-decnet (host dir-pathname error-p)
  (multiple-value-bind (user-id password)
      (find-password dir-pathname host)
    (fs:file-operation-retry
      (let* ((host-address (second (send host :network-typep :dna)))
	     (dir-list nil)(error-condition nil)
	     (file-string (send dir-pathname :string-for-host)))
	(multiple-value-bind (port reused)
	    (dap-connect host-address
			 :userid user-id
			 :password password
			 :dstnode-host-object host) ; DAB 02-13-89
	  (when port
	    (setf (send port :who-line-state) "Directory")
	    (unwind-protect
		(let ((dap-stream (make-instance 'dap-basic-stream
						 :pathname dir-pathname
						 :port port
						 :send-config? reused
						 :signal-error error-p)))
		  (multiple-value-setq (dir-list error-condition)
		    (send dap-stream :pathname-directory-list file-string))
		  (send dap-stream :clear-receive-queue))
	      (if dir-list
		  (add-port-to-released-list port)
		  (send port :close)))
	    (if dir-list dir-list (signal error-condition))))))))

(defun rename-decnet (host old-pathname new-pathname error-p)
  (multiple-value-bind (user-id password)
      (find-password old-pathname host)

    (setf new-pathname (fs:merge-pathnames new-pathname old-pathname))
    (fs:file-operation-retry
      (let* ((host-address (second (send host :network-typep :dna)))
	     truename-object error-object
	     (old-file-string (send old-pathname :string-for-host))
	     (new-file-string (send new-pathname :string-for-host)))
	(multiple-value-bind (port reused)
	    (dap-connect host-address
			 :userid user-id
			 :password password
			 :dstnode-host-object host) ; DAB 02-13-89
	  (when port
	    (setf (send port :who-line-state) "Rename File")
	    (unwind-protect
		(let ((dap-stream (make-instance 'dap-basic-stream
						 :pathname old-pathname
						 :port port
						 :send-config? reused
						 :signal-error error-p)))
		  (multiple-value-setq (truename-object error-object)
				       (send dap-stream :pathname-rename old-file-string new-file-string))
		  (send dap-stream :clear-receive-queue))
	      (if truename-object
		  (add-port-to-released-list port)
		  (send port :close)))
	    (if truename-object
		truename-object
		error-object)))))))

(defun delete-decnet (host pathname error-p)
  (multiple-value-bind (user-id password)
      (find-password pathname host)
    (fs:file-operation-retry
      (let* ((host-address (second (send host :network-typep :dna)))
	     truename-object error-condition
	     (file-string (send pathname :string-for-host)))
	(multiple-value-bind (port reused)
	    (dap-connect host-address
			 :userid user-id
			 :password password
			 :dstnode-host-object host) ; DAB 02-13-89
	  (when port
	    (setf (send port :who-line-state) "Delete File")
	    (unwind-protect
		(let ((dap-stream (make-instance 'dap-basic-stream
						 :pathname pathname
						 :port port
						 :send-config? reused
						 :signal-error error-p)))
		  (multiple-value-setq (truename-object error-condition)
				       (send dap-stream :pathname-delete file-string))
		  (send dap-stream :clear-receive-queue))
	      (add-port-to-released-list port))
	    (if truename-object
		truename-object
		error-condition)))))))

(defun pathname-truename-decnet (pathname host error-p)
  (multiple-value-bind (user-id password)
      (find-password pathname host)
    (fs:file-operation-retry
      (let* ((host-address (second (send  host :network-typep :dna)))
	     error-condition
	     truename-object)
	(multiple-value-bind (port reused)
	    (dap-connect host-address
			 :userid user-id
			 :password password
			 :dstnode-host-object host) ; DAB 02-13-89
	  (when port
	    (setf (send port :who-line-state) "Truename")
	    (unwind-protect
		(let ((dap-stream (make-instance 'dap-basic-stream
						 :pathname pathname
						 :port port
						 :send-config? reused
						 :signal-error error-p)))
		  (multiple-value-setq (truename-object error-condition)
				       (send dap-stream :pathname-truename pathname))
		  (send dap-stream :clear-receive-queue))	      
	      (add-port-to-released-list port))
	  (if truename-object
	      truename-object
	      error-condition)))))))

(DEFUN open-decnet (host pathname &rest
		    &key (direction (if batch-file :output :input))
		    (characters :default)	
		    (element-type 'string-char element-type-p)
		    (byte-size :default)
		    (if-exists
		      (when (eq direction :output)
			(if (member (pathname-version pathname) '(:newest :unspecific))
			    :new-version 
			    :error)))
		    (if-does-not-exist
		      (case direction
			((:input nil) :error)	   ;here is where nil and :probe differ
			(:output (if (or (eql if-exists :append)
					 (eql if-exists :overwrite)) :error :create))
			(:otherwise nil)))
		    (error (not (member direction '(:probe :probe-directory))))	   ;nil, probe differ here
		    estimated-length
;; new keys:		    
		    ascii-stream
		    batch-file
		    character-format	   ; :ascii / :ebcdic / (:ascii/:ebcdic :non-print/:telnet/:carriage-control)
		    maximum-record-length  ; must also give estimated-length to use this parm
;; Keys ignored:
		    preserve-dates deleted ;zwei uses these
		    temporary            
		    &aux phony-characters sign-extend-bytes
		    (file-format nil) truename properties)
  (declare (ignore preserve-dates deleted temporary))


  (si:ccase direction			                         ;prompts user if not one of listed values
    ((:input :output :probe-directory nil :probe))
    ((:io :probe-link) 
     (ferror nil "Option ':DIRECTION :~A' not supported." direction)))
  
  (si:ccase if-does-not-exist
    (:create 
      (when (eq direction :input) (ferror nil "Option ':IF-DOES-NOT-EXIST :Create' not legal for input")))
    (:error nil)
    (nil))				                         ;legal option
  
  (when (and if-exists (eq direction :input))
    (ferror nil "Option ':IF-EXISTS' legal only for output"))
  
  (si:ccase if-exists
    ((:rename :rename-and-delete)	                         ;zmacs uses on replace of existing file
     (setq if-exists :rename))
    (:append 
      (when (neq direction :output) (ferror nil "Option ':IF-EXISTS :Append' illegal on input")))
    (:overwrite
      (when (neq direction :output) (ferror nil "Option ':IF-EXISTS :Overwrite' illegal on input"))
      (ferror nil "Option ':IF-EXISTS :Overwrite' not supported"))
    ((:error :truncate :supersede :new-version nil))) ;other legal options
  
  ;; This is a QFILE routine used as-is
  (when element-type-p 
    (setf (values characters byte-size phony-characters sign-extend-bytes)
	  (fs:decode-element-type element-type byte-size)))
  
  (si:ccase characters
    (:default				                         ;used by fs-utilities (e.g. dired)
      (IF character-format
	  ;; THEN assume  characters if character-format specified
	  (setq characters t)
	  ;; ELSE check file type
	  (setq characters (not (member (send pathname :canonical-type) fs:*copy-file-known-binary-types*)))))
    ((t nil)))
  
  (when (eq byte-size :default)			
    (setq byte-size 
          (or (get (locf properties) :byte-size)
              (if characters 8 16.))))
  
  (if character-format
      (progn
	(if (not characters)
	    (ferror nil "Option ':CHARACTER-FORMAT ~A' not valid for binary file" character-format))
	
	(when  (listp character-format)
	  (setq file-format (second character-format))
	  (setq character-format (car character-format)))
	
	(si:ccase character-format	   ; (changed above)
          ((:ascii :ebcdic)) 
          (nil (setq character-format :ascii)))
	
	(si:ccase file-format				
	  ((:non-print :telnet :carriage-control))
	  (nil (setq file-format :non-print)))
	)
      
      (if characters
	  (progn
	    (setq character-format :ascii)
	    (setq file-format :non-print))
	  (setq character-format 
		(cond
		  ((eql byte-size 16.) (if (send pathname :send-if-handles :omit-byte-size)
					   :image
					   :byte-size))
		  ((eql byte-size 8.) :image)
		  (:otherwise (ferror nil "Byte sizes other than 8 and 16 are not supported"))))) 
      ) 
  
  
  (when (and  characters (neq byte-size 8.)) 
    (ferror nil "Byte sizes other than 8 are not supported for character files"))  
  
  (when maximum-record-length 
    (unless estimated-length 
      (ferror nil "MAXIMUM-RECORD-LENGTH not allowed unless ESTIMATED-LENGTH is also specified")))
  
;; ---------------Port connection and Stream Construction:
  
  (setq properties			   
	(nconc 
	  (list 
	    :truename truename
	    :byte-size byte-size	   
	    :length 0			   ;set this to 0 even if we know :length-in-bytes (this is the remote length
					   ;and due to ascii-translation it can be off (see :who-line-information)
	    :characters characters
	    :character-format character-format
	    :file-format file-format
	    ;; don't set qfaslp 
	    )
	  (if (eq direction :output)
	      (progn (send pathname :set-property-list nil) nil)
	      properties)	
	  ))
  (multiple-value-bind (user-id password)
      (find-password pathname host)
   (let*  (truename-object error-object (signaled-error t)
	   basic-stream
	   stream)
      (multiple-value-bind (port reused)
	  (dap-connect host  
		       :userid user-id
		       :password password
		       :dstnode-host-object host) ; DAB 02-13-89
	
	(when port
	  (unwind-protect
	      (progn
		(setf (send port :who-line-state)
		      (if (or (eql direction :probe)
			      (eql direction :probe-directory)
			      (eql direction nil)) "Probe" "Open"))
		(setq basic-stream (make-instance 'dap-basic-stream
					    :pathname pathname
					    :port port
					    :property-list properties
					    :send-config? reused
					    :signal-error error
					    :current-direction direction))
		(setq stream (make-instance (decnet-stream-flavor-type direction byte-size
								       characters ascii-stream)
		                            :basic-stream basic-stream
					    :pathname pathname))
		(multiple-value-setq (truename-object error-object)
				     (send stream :open pathname
					   direction
					   (or (eql if-exists :new-version)	         ;New version?
					       (eql if-does-not-exist :create))
					   (eql if-does-not-exist :error)                ;Error if not there
					   (if characters ascii-data image-data)         ;Binary or text
					   (eql if-exists :overwrite)	                 ;Overwrite existing
					   (eql if-exists :append)                       ;Append to end of file?
					   (eql if-exists :supersede)	                 ;Supersede existing?  
					   byte-size
					   ascii-stream
					   batch-file
					   estimated-length))
		(setf signaled-error nil))   ;this means we made it through open without errors.
	    (if (or error-object signaled-error)
		(progn
		  (send stream :close)
		  (send basic-stream :clear-receive-queue)
		  (add-port-to-released-list port))
		(setf (send port :who-line-state) (if (eql direction :output) "File Output" "File Input"))))
							
	  (cond
	    (truename-object stream)
	    ((eql direction :PROBE) nil)
	    ((and (eql direction :input)
		  (not if-does-not-exist))
	     nil)
	    (error (signal error-object))
 	    (t error-object)))))))
))
