;;; -*- Mode:Common-Lisp; Package:DNA; Base:10; Patch-file:T -*-
;;; Patch file for DECNET version 1.60
;;; Reason: Modify reset-routing so that it will not try to dynamically change the Ethernet address on a microExplorer. Doing so
;;; hangs AppleTalk. Instead we write out an eadr resource on the EtherTalk driver and disable decnet until after the next 
;;; Macintosh boot (when etherTalk will start up with the correct ethernet address).
;;; Written 03/28/89 22:35:25 by jones,
;;; while running on HOBBS from band N327
;;; With SYSTEM 5.28, GC 5.4, VIRTUAL-MEMORY 5.5, MICRONET 5.6, MICRONET-COMM 5.14,
;;;  DISK-IO 5.10, BASIC-PATHNAME 5.4, MAC-PATHNAME 5.1, NETWORK-SUPPORT-COLD 5.2,
;;;  BASIC-NAMESPACE 5.8, BASIC-FILE 5.5, RPC 5.6, NFS 5.18, EH 5.4, MAKE-SYSTEM 5.4,
;;;  MEMORY-AUX 5.1, MACTOOLBOX 1.35, COMPILER 5.3, TV 5.27, NVRAM 5.1, UCL 5.0, INPUT-EDITOR 5.1,
;;;  METER 5.1, ZWEI 5.14, DEBUG-TOOLS 5.1, WINDOW-MX 5.38, PRINTER 5.12, MAC-PRINTER-TYPES 5.4,
;;;  Experimental NETWORK-PATHNAME 5.1, NETWORK-NAMESPACE 5.0, DATALINK 5.7, CHAOSNET 5.6,
;;;  Experimental NETWORK-SUPPORT 5.1, Experimental NETWORK-SERVICE 5.1, DATALINK-DISPLAYS 5.0,
;;;  NAMESPACE-EDITOR 5.2, IP 3.42, NFS-SERVER 5.3, PRINTER-TYPES 5.5, IMAGEN 5.3,
;;;  MAIL-DAEMON 5.7, MAIL-READER 5.6, TELNET 5.3, VT100 5.1, STREAMER-TAPE 5.7, DECNET 1.59,
;;;  VISIDOC 5.6, PROFILE 5.1, DISK-LABEL 5.1, Experimental MX-SERIAL 1.1,  microcode 126,
;;;  Band Name: n327+decnet-fix

#!C
; From file ROUTING.LISP#> decnet; sys:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "Sys: decnet; ROUTING.#"

(Defun SETUP-ETHERNET-ADDRESS (controller)
  "Reads the Config ROM Ethernet Address from CONTROLLER, manufactures a new Decnet
   Ethernet Address, then issues an IA-Setup command to CONTROLLER to load the
   new address."  
  (Let ((Net-Addr (Send Controller :Read-Rom-Id)))
    ;; Only change the ethernet address if we have a decnet address and a controller that is new enough
    ;; to give us the correct ethernet address (some very old ones that were not released will give us all 0's or all F's).
    (when (and my-address
	       (not (zerop my-address))
	       (not (zerop Net-addr))
	       (not (= Net-Addr #xFFFFFFFFFFFF)))
    (Send Controller :IA-Setup (+ hiord_16 (net:swap-bytes my-address)))))
  )

(defvar *routing-reset-inhibit* nil "If t Decnet cannot be enabled because the ethernet address is wrong.")
(add-initialization "*routing-reset-inhibit*" '(setf *routing-reset-inhibit* nil) :before-cold)

(defun ROUTING-RESET (&key (enable t) (hard t))
  "Put the DECnet module into a (hopefully) runnable state.  Note that this host's ETHERNET
    address may change if this is the first time that the machine was reset with DECnet installed."
  (when net:controller-list
    (dna:disable)
    (dna:reset-on-ethernet-cache)
    (dna:reset-node-counters)
    (dna:setup-my-address)
    (unless (or (null my-address) (zerop my-address))
      (if hard
	  (progn
	    ;; Don't keep trying to set the Ethernet address on a microExplorer. *BJ*
	    (when *routing-reset-inhibit*
	      (tv:notify nil "Decnet is disabled until after the next Macintosh boot.")
	      (return-from routing-reset nil))
	    
	    (ethernet:add-protocol dna:*ROUTE-ETHERNET-TYPE* 'Routing-Receiver)
	    (Ethernet:Add-Net-Packet-Allocator
	      dna:*ROUTE-ETHERNET-TYPE* 'CHAOS::ALLOCATE-INT-PKT 'CHAOS::FREE-INT-PKT)
	    (net:reset-packets)
	    (setf *ROUTING-HANDLER* (make-instance 'routing-handler))
	    (net:send-all-controllers :disable)	       ;need to permanently flush old controllers using...
	    (Dolist (Cont Net:Controller-List)
	      (when (and (not (Setup-Ethernet-Address Cont)) (si:mx-p))
		;; If the address was not set on an MX then disable decnet.
		(setf my-address 0)
		(setf *routing-reset-inhibit* t)
		(return-from routing-reset)))
	    (ethernet:reset-address-translations)      ;force rebroadcasts of our (new) ethernet address
	    (net:send-all-controllers :reset)
	    (net:send-all-controllers :enable))	
	  (send *routing-handler* :reset))
      (reset-hello-xmitter)
      (set-host-addresses)
      (when enable
	(dna:enable))
      (setup-multicast-addresses))
    ))

))