;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Written 03/30/89 09:30:33 by REINER,
;;; Reason: Wrote a new function, simple-universal-time which returns a fixnum timestamp only consisting of hours, minutes, seconds.
;;; This will help reduce the amount of consing caused by calling get-universal-time.
;;;     Code Readers:  DAB, LER   03/30/89
;;; while running on MX30 from band NB22
;;; With SYSTEM 5.19, GC 5.3, VIRTUAL-MEMORY 5.5, MICRONET 5.5, MICRONET-COMM 5.13,
;;;  DISK-IO 5.9, BASIC-PATHNAME 5.2, MAC-PATHNAME 5.0, NETWORK-SUPPORT-COLD 5.1,
;;;  BASIC-NAMESPACE 5.6, BASIC-FILE 5.3, RPC 5.4, NFS 5.10, EH 5.3, MAKE-SYSTEM 5.2,
;;;  MEMORY-AUX 5.1, MACTOOLBOX 1.25, COMPILER 5.1, TV 5.21, NVRAM 5.1, UCL 5.0, INPUT-EDITOR 5.0,
;;;  METER 5.0, ZWEI 5.9, DEBUG-TOOLS 5.1, WINDOW-MX 5.28, PRINTER 5.11, MAC-PRINTER-TYPES 5.4,
;;;  NETWORK-PATHNAME 5.0, NETWORK-NAMESPACE 5.0, DATALINK 5.7, CHAOSNET 5.6, NETWORK-SUPPORT 5.0,
;;;  NETWORK-SERVICE 5.0, DATALINK-DISPLAYS 5.0, NAMESPACE-EDITOR 5.1, IP 3.33, NFS-SERVER 5.3,
;;;  PRINTER-TYPES 5.2, IMAGEN 5.1, MAIL-DAEMON 5.1, MAIL-READER 5.3, TELNET 5.1,
;;;  VT100 5.0, STREAMER-TAPE 5.6, DECNET 1.45, VISIDOC 5.4, PROFILE 5.1, DISK-LABEL 5.1,
;;;   microcode 128, Band Name: microExplorer Network (11/22)

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(defun simple-universal-time ()
 "Returns the encoded hour, minute, and second - the date is not encoded. 
  This is an alternative to get-universal-time which causes bignum consing." 
 (time:update-timebase)
 (+ time:*last-time-seconds* 
   (* 60. time:*last-time-minutes*) (* 3600. time:*last-time-hours*) (* 3600. time:*timezone*))) 

))

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(defun ROUTING-RECEIVER (circuit data-array nbytes type src-addr dst-addr
			 &optional (free-buffer t))
  "Receive a DECnet packet from the datalink handler.  This function must never block, go into the error
    handler, or take excessive CPU time.  It is executed in the receiver process for the datalink."
  (declare (special *nsp-object*))
  (unwind-protect
      (when (and *receive-enabled*
		 (not (eql 0 nbytes))
		 (not (null dna:my-addresses)))
	(let* ((INHIBIT-SCHEDULING-FLAG T)
	       (start-time  (sys:%fixnum-microsecond-time))
	       (default-cons-area *decnet-area*)
	       (pkt (send *routing-handler* :allocate-pkt :input))
	       (header-preamble
		 (progn (WITHOUT-INTERRUPTS
			  (SYS:%BLT (SYS:%MAKE-POINTER-OFFSET
				  SYS:DTP-FIX data-array
				  (SI:ARRAY-DATA-OFFSET data-array))
				(SYS:%MAKE-POINTER-OFFSET
				  SYS:DTP-FIX (send pkt :base-array)
				  (SI:ARRAY-DATA-OFFSET
				    (send pkt :base-array)))
				(min (CEILING NBYTES 4)
				     (FLOOR (length (send pkt :base-array)) 4))
				1))
			(unless (<= (* 4 (CEILING NBYTES 4))	   ;rwf 10/31/86
				    (length (send pkt :base-array)))
			  (when *packet-trace-enable*
			    (record-unusual-packet :too-big
						   (send pkt :base-array) :receive nbytes
						   type src-addr dst-addr circuit)))
			(setf (net:packet-time-received     pkt) (time))
			(setf (net:packet-datalink-src-addr pkt) src-addr) 
			(setf (net:packet-datalink-dst-addr pkt) dst-addr) 
			(setf (net:packet-datalink-nbytes   pkt)
			      (min nbytes (+ 2 (FLAGS-PACKET-SIZE (send pkt :base-array)))))
			(setf (net:packet-datalink-type     pkt) type)
			(setf (net:packet-datalink-circuit  pkt) circuit)
			(net:get-datalink-header
			  pkt :pre-routing
			  (DNA:ROUTE-HEADER-OFFSET (send pkt :base-array)))))
	       (datalink-data (net:get-datalink-data
				pkt (min (- nbytes (length header-preamble))
					 (flags-packet-size header-preamble))
				)
			      )
	       (header        (net:get-network-header
				pkt
				(DNA:PARSE-ROUTE-MESSAGE-TYPE  datalink-data)
				(DNA:PARSE-ROUTE-HEADER-LENGTH datalink-data)))
	       (data          (net:get-network-data
				pkt (min (- nbytes
					    (length header-preamble)
					    (length header))
					 (- (flags-packet-size header-preamble)
					    (length header)))
				)
			      )
	       )

	      (if (prog1 (or (null header)(null data)(zerop (length header))) ;crw 5-24-88
			 (when *packet-trace-enable*
			   (record-packet (send pkt :base-array) :receive (net:packet-datalink-nbytes pkt)
					  type src-addr dst-addr circuit)))
		  (incf (routing-handler-ill-formed-packets-received *routing-handler*)) 
		  (progn
		    (unless (member (area (src-node header)) dna:my-areas)
		      (incf (routing-handler-packets-from-other-area *routing-handler*))
		      (when *packet-trace-enable*
			(record-unusual-packet :foreign-area
					       (send pkt :base-array) :receive nbytes
					       type src-addr dst-addr circuit))
		      )
		    (incf (routing-handler-packets-received        *routing-handler*))
		    
		    
		    (when (< 2 (length header-preamble))
		      (incf (routing-handler-padded-packets-received *routing-handler*)))
		    (incf (routing-handler-bytes-received   *routing-handler*) nbytes)
		    (let* ((source (src-node header)))
		      (when source
			(let* ((entry (LOOKUP-NODE-COUNTERS source)))
			  (when entry 
			    (case (net:get-network-header-kind pkt)
				  (:ENDNODE-HELLO-MESSAGE
				   (incf (node-counters-ENDNODE-HELLO-packets-received   entry)))
				  (:ROUTER-HELLO-MESSAGE
				   (incf (node-counters-ROUTER-HELLO-packets-received    entry)))   
				  (:LEVEL-2-ROUTING-MESSAGE
				   (incf (node-counters-level-2-routing-packets-received entry)))
				  (:LEVEL-1-ROUTING-MESSAGE
				   (incf (node-counters-level-1-routing-packets-received entry)))
				  (:HELLO-AND-TEST-MESSAGE
				   (incf (node-counters-hello-and-test-packets-received  entry)))
				  (:ROUTING
		                   (setf (node-counters-last-receive-time entry) (simple-universal-time))
				   (incf (node-counters-long-route-packets-received      entry))
				   (incf (node-counters-packets-received entry))
				   (incf (node-counters-bytes-received   entry) nbytes)
				   )
				  (otherwise
				   (setf (node-counters-last-receive-time entry) (simple-universal-time))
				   (incf (node-counters-unknown-type-packets-received    entry))
				   (incf (node-counters-packets-received entry))
				   (incf (node-counters-bytes-received   entry) nbytes)
				   )
				  )	   ;case
			    )		   ;when
			  )		   ;let*
			)		   ;when
		      )			   ;let*
		    (case (net:get-network-header-kind pkt)
			  (:ENDNODE-HELLO-MESSAGE (process-endnode-hello pkt dst-addr))
			  (:ROUTER-HELLO-MESSAGE  (process-router-hello  pkt dst-addr))   
			  (:LEVEL-2-ROUTING-MESSAGE
			   (incf (routing-handler-level-2-routing-packets-received *routing-handler*))
			   (send *routing-handler* :deallocate-pkt pkt))
			  (:LEVEL-1-ROUTING-MESSAGE
			   (incf (routing-handler-level-1-routing-packets-received *routing-handler*))
			   (send *routing-handler* :deallocate-pkt pkt))
			  (:HELLO-AND-TEST-MESSAGE
			   (incf (routing-handler-hello-and-test-packets-received  *routing-handler*))
			   (send *routing-handler* :deallocate-pkt pkt))
			  (:ROUTING
			   (incf (routing-handler-long-route-packets-received      *routing-handler*))
			   (when (and (not *passive-nsp*)
				      (variable-boundp *nsp-object*)
				      *nsp-object*)
			     (send *nsp-object* :receive-message pkt (src-node header))))
			  (otherwise
			   (incf (routing-handler-unknown-type-packets-received    *routing-handler*))
			   (when *packet-trace-enable*
			     (record-unusual-packet :unknown-packet-type
						    (send pkt :base-array) :receive nbytes
						    type src-addr dst-addr circuit))
			   (send *nsp-object* :receive-odd-message pkt
				 (net:swap-bytes (logxor src-addr hiord_16))))			   
			  )		   ;case
		    )			   ;else (progn)
		  )			   ;if
	  (incf (routing-handler-time-receiving *routing-handler*)        
		(round (time-difference (sys:%fixnum-microsecond-time) start-time) 1000.)
		)))
	;; clean up resources.
	(when free-buffer
	  (ethernet::deallocate-net-packet type data-array)))
    )

))

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(defun ROUTING-TRANSMIT (output-array nbytes destination &optional (self-receive nil))
  "The transmit routine for DNA Routing."
  (when (and *transmit-enabled*
	     (not (null dna:my-addresses))
	     net:controller-list)
    (let* ((pkt-nbytes  nbytes)
	   (controller  (first net:controller-list))	   ;TEMPORARY!!!!!
	   (Destination-Ether-Address (if (< destination #x10000)
					  (logior HIORD_16 (net:swap-bytes destination))
					  destination))
	   (Source-Ether-Address      (send controller :ethernet-address))
	   (output-word-array nil))
      (when (eql (flags-padded byte-pkt) 1)
	(INCF (routing-handler-padded-packets-sent *ROUTING-HANDLER*)))
      (when *packet-trace-enable*
	(record-packet output-array :transmit pkt-nbytes
		       *ROUTE-ETHERNET-TYPE*
		       Source-Ether-Address
		       Destination-Ether-Address
		       Controller))
      
      (INCF (routing-handler-packets-sent *ROUTING-HANDLER*))
      (INCF (routing-handler-bytes-sent   *ROUTING-HANDLER*) pkt-nbytes)
      
      (when (< destination #x10000)
	(let* ((entry (LOOKUP-NODE-COUNTERS destination)))
	  (when entry 
	    (case (DNA:ROUTE-MESSAGE-TYPE  byte-pkt)
	      (:ENDNODE-HELLO-MESSAGE
	       (incf (node-counters-ENDNODE-HELLO-packets-sent   entry)))
	      (:ROUTER-HELLO-MESSAGE
	       (incf (node-counters-ROUTER-HELLO-packets-sent    entry)))   
	      (:LEVEL-2-ROUTING-MESSAGE
	       (incf (node-counters-level-2-routing-packets-sent entry)))
	      (:LEVEL-1-ROUTING-MESSAGE
	       (incf (node-counters-level-1-routing-packets-sent entry)))
	      (:HELLO-AND-TEST-MESSAGE
	       (incf (node-counters-hello-and-test-packets-sent  entry)))
	      (:ROUTING
	       (setf (node-counters-last-send-time entry) (simple-universal-time))
	       (incf (node-counters-long-route-packets-sent      entry))
	       (incf (node-counters-packets-sent entry))
	       (incf (node-counters-bytes-sent   entry) pkt-nbytes)
	       )
	      (otherwise
	       (setf (node-counters-last-send-time entry) (simple-universal-time))
	       (incf (node-counters-packets-sent entry))
	       (incf (node-counters-bytes-sent   entry) pkt-nbytes)
	       )
	      )
	    )
	  )
	)
      (without-interrupts
	(setq output-word-array (if (= 16. (si:array-element-size output-array))
				    output-array       
				    (when (SI:ARRAY-INDIRECT-P output-array)
				      (ferror nil "~& Illegal (indirect) input to net."))
				    (assert (<= (* 4 (CEILING pkt-nbytes 4))	   ;rwf 10/31/86
						(* 2 (length real-word-pkt))))
				    
				    (SYS:%BLT (SYS:%MAKE-POINTER-OFFSET
						SYS:DTP-FIX
						output-array
						(SI:ARRAY-DATA-OFFSET output-array))
					      (SYS:%MAKE-POINTER-OFFSET
						SYS:DTP-FIX
						real-word-pkt
						(SI:ARRAY-DATA-OFFSET real-word-pkt))
					      (CEILING pkt-nbytes 4)
					      1)
				    real-word-pkt))
	
	(unless (MEMBER destination dna:my-addresses)
	  (SEND controller :TRANSMIT-FAST
		*ROUTE-ETHERNET-TYPE*
		Destination-Ether-Address
		output-word-array
		pkt-nbytes)
	  )
	
	(when (or (MEMBER destination dna:my-addresses) self-receive)
	  (ROUTING-RECEIVER
	    controller
	    output-word-array
	    pkt-nbytes
	    *ROUTE-ETHERNET-TYPE*
	    Source-Ether-Address      
	    Destination-Ether-Address 
	    nil)
	  (process-allow-schedule))
	)
      )					   ;when
    )					   ;let
  )					   ;defun

))

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(Defun CLEAR-NODE-COUNTERS (host)
  "Clear all counters entries for one host or all hosts on the local net."
  (if (null host)
      (without-interrupts
	(dolist (entry-cons *node-counters*)
	  (let* ((entry (cdr entry-cons)))
	    (when entry
	      (array-initialize entry 0)
	      (setf (node-counters-time entry) (simple-universal-time))
	      ))))
      (let ((entry (lookup-node-counters (parse-dna-address-spec host) nil)))
	(when entry
	  (array-initialize entry 0)
	  (setf (node-counters-time entry) (simple-universal-time))
	  )))
  )

))

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(Defun LOOKUP-NODE-COUNTERS (node-id &optional (create t))
  "Given a node it, return the counters entry, if any, for it"
  (let ((entry (cdr (assoc node-id *node-counters*))))
    (when (and create (null entry))
      (setq entry (make-node-counters))
      (setf (node-counters-time entry) (simple-universal-time))
      (STORE-NODE-COUNTERS node-id entry))
    entry))

))

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(defun UPDATE-ON-ETHERNET-CACHE (pkt dst-addr)
  "Given a just-received hello packet, modify the cache to reflect its contents"
  (let* ((header (net:get-network-header pkt))
	 (node-id (src-node header))
	 (entry (LOOKUP-on-ethernet-cache-element node-id)))
    (unless entry
      (setq entry (ALLOCATE-ON-ETHERNET-CACHE-ELEMENT))
      (setf (on-ethernet-cache-header entry) (ALLOCATE-ON-ETHERNET-CACHE-HEADER))
      (setf (on-ethernet-cache-multicast-address entry) dst-addr)
      (STORE-ON-ETHERNET-CACHE-ELEMENT node-id entry)
      )
    (copy-array-portion header 0 (length header)
			(on-ethernet-cache-header entry) 0 (length header)
			)
    (setf (on-ethernet-cache-time             entry) (simple-universal-time))
    (incf (on-ethernet-cache-packets-received entry))
    (incf (on-ethernet-cache-bytes-received   entry) (net:get-physical-data-nbytes pkt))
    entry))

))

#!C
; From file ROUTING.LISP#> DECNET; SYS:
#10R DNA#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "DNA"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DECNET; ROUTING.#"


(defun CHECK-CACHE-STATUS ()
  "Check the hello cache for old entries."
  (dolist (entry-cons *on-ethernet-cache*)
    (let* ((entry (cdr entry-cons)))
      (when (> (- (simple-universal-time) (on-ethernet-cache-time entry))
	       (max CACHETIMEOUT (* bct3mult (router-hello-timer (on-ethernet-cache-header entry)))))
	(without-interrupts
	  (when (eql (car entry-cons) *designated-router*)
	    (setf *designated-router* nil))
	  (setf *on-ethernet-cache* (delete entry-cons *on-ethernet-cache*))
	  (FREE-ON-ETHERNET-CACHE-HEADER (on-ethernet-cache-header entry))
	  (FREE-ON-ETHERNET-CACHE-ELEMENT entry))
	)
      )
    )
  )

))
