;;; -*- cold-load:t; Mode:Common-Lisp; Package:EH; Fonts:(CPTFONT HL12B HL12BI); Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
1;;
;; Copyright (c) 1981 Massachusetts Institute of Technology 
;; Copyright (c) 1984-1989 Texas Instruments Incorporated.  All Rights Reserved.
;;

;; New error handler.*

;;;
;;; Edit History
;;;
;;;                   Patch
;;;   Date    Author  Number   Description
;;;--------------------------------------------------------------------
;;;   07-86     pf             Original, from parts of Rel2 EH files.
;;;  03-04-87   clm   *L.3     Fix sg-eval-in-frame to prevent crashing
;;;                            when trying to Meta-(ctrl)-Abort out of a
;;;                            Resume.
;;;  03-09-87   ab    *L.11    Fix push-new-frame to initialize local
;;;                            block of frame to NILs so we don't pick
;;;                            up un-scavenged garbage (oldspace pointers).
;;;  04-09-87   clm   *M       In SG-SAVE-STATE, if microstack pushed, set 
;;;                            bit in Foothold call-info word just in case 
;;;                            there's a call-out in a microinstruction. In
;;;                            certain cases the pdl's where getting out of sync. 
;;;  03/20/89   clm            Integrated CLOS changes into EH.


;;  pdl before PUSH-NEW-FRAME:			 pdl after PUSH-NEW-FRAME:
;;
;;						|                     | <--- new-frame
;;						|---------------------|
;;			 			|     new frame's     |
;;						|  5 call state words |
;;						|---------------------|
;;						|                     | <--- local-offset
;;						|---------------------|
;;						|     next frame's    |
;;						|  5 call state words |
;;  |                      | <--- next-frame	|---------------------|
;;  |----------------------|			|  new frame's arg n  |
;;  |     next frame's     |		 	|      .......        |
;;  |  5 call state words  |			|  new frame's arg 0  | <--- arg-offset
;;  |----------------------|			|---------------------|
;;  |     .......          |			|      .......        |


;; 03/10/87 CLM - Set %%call-info-lexpr-funcall-flag bit in call-info-word to 0.
;;                This wasn't being done and was causing the arguments to be handled
;;                incorrectly.
;; 05/20/88 CLM - Added CLOS support.
;; 03/20/89 clm - Integrated CLOS changes into EH.
(defun push-new-frame (sg frame function args
		       &optional (call-info-word 0) (arg-offset (- frame %call-state-length)))
  "Using FUNCTION and ARGS, builds a new frame on top of FRAME in regular pdl of SG"
  (let* ((reg-pdl (sg-regular-pdl sg))
	 (next-frame-arg-offset (rp-argument-offset sg reg-pdl frame))
	 (next-frame-local-offset (rp-local-offset sg reg-pdl frame))
	 (num-args (length args))
	 (local-offset (+ arg-offset num-args %call-state-length))
	 (new-frame-state (+ local-offset
			     (IF (TYPEP function 'compiled-function)
				 (fef-number-of-locals function)
				 0)))			; The only important non-fef case so far is microcode entries.  
	 (new-frame (+ new-frame-state %call-state-length)))
    ;; Move top frame's call-state words to just below new frame's local index
    ;; and just above the args for the new frame.
    ;; This does not have to be done when there are no args for the new frame
    ;; (i.e., when the frame index is the same as the local-offset index)
    (unless (= frame local-offset)
      (if (< frame local-offset)
	  (loop for i from 1 to %call-state-length
		doing (setf (aref reg-pdl (- local-offset i)) (aref reg-pdl (- frame i))))
	  (loop for i from %call-state-length downto 1
		doing (setf (aref reg-pdl (- local-offset i)) (aref reg-pdl (- frame i)))))
      ;; Need to convert arg and local pointer of frame state we just moved up
      (setf (sg-top-frame sg) local-offset)
      (setf (rp-local-pointer reg-pdl local-offset) (pdl-buffer-index sg next-frame-local-offset))
      (setf (rp-argument-pointer reg-pdl local-offset) (pdl-buffer-index sg next-frame-arg-offset)))

    ;; Set new top frame to this new frame for FUNCTION
    (setf (sg-top-frame sg) new-frame)
    ;; Set all 5 call-state words for new frame:
    (setf (rp-location-counter-offset reg-pdl new-frame) (if (typep function 'compiled-function)
							     (fef-initial-pc function)
							     0))
    (setf (rp-function-word reg-pdl new-frame) function)
    ;; NOTE: PDL-BUFFER-INDEX converts a pdl array index to a pdl buffer index
    (setf (rp-local-pointer reg-pdl new-frame) (pdl-buffer-index sg local-offset))
    (setf (rp-argument-pointer reg-pdl new-frame) (pdl-buffer-index sg arg-offset))
    (setf (rp-call-info-word reg-pdl new-frame)
	  (dpb 0 %%call-info-lexpr-funcall-flag    ;set the lexpr-funcall bit to 0***
	       (dpb (sg-restore-microstack sg)	    ; Preserve the bit.  - pf, Apr 18, 1986
		    %%call-info-microstack-pushed
		    (dpb (length args)
			 %%call-info-number-of-arguments
			 call-info-word))))
    ;;if this is a clos method the args contain the mapping-tables and the
    ;;the continuation, these should be included in the call-info args field
    (setf (rp-number-of-arguments reg-pdl new-frame)
	  (if (zerop (ldb si:%%call-info-clos-info-provided call-info-word))
	      num-args
	      (let* ((debug-info (extract-debug-info-struct-from-fef function ))
		     (map-slots (get-debug-info-field debug-info :map-slots))
		     (continuation-slot 1))   ;;there will always be a continuation
		(- num-args map-slots continuation-slot ))))
    

    ;; Move args onto stack
    (do ((arg args (cdr arg))
	 (i arg-offset (1+ i)))
	((null arg))
      (setf (aref reg-pdl i) (car arg))
      (%p-store-cdr-code (locf (aref reg-pdl i)) cdr-next))
     ;; Store NILs in local slots.
    (loop for i from local-offset below new-frame-state
	  do (setf (aref reg-pdl i) nil))
    ;; No microstack saved for this new frame
    (setf (sg-restore-microstack sg) 0)
    new-frame))


1;; FOOTHOLD 
;; This function isn't normally called, it just exists to name state-save frames.
;; If this function is ever returned to (see (:METHOD PROCESS :INTERRUPT))
;; then it will restore the saved state and resume it.
;; Do not trace nor redefine this function!*

(defprop foothold t :error-reporter)

(defun foothold ()
  (funcall %error-handler-stack-group '(resume-foothold)))

1;; SG-SAVE-STATE 
;; Save a stack group's state on its stack so we can use it and then restore the state.
;; The information goes on the pdl in a fake frame belonging to the function FOOTHOLD.
;; Each Q is saved as 2 words (pointer and tag) to avoid data type problems.
;; You must call this before pushing a call block, even if calling SG-RUN-GOODBYE,
;; in order to clean up the QBBFL and the U-STACK Q's.*

(defun sg-save-state (sg &optional suppress-pdl-growing)
  "Save the state (accumulators, etc) of stack group SG on its pdl.
SUPPRESS-PDL-GROWING inhibits making the stack bigger heuristically.
Once this function has been called, you can then push and invoke a call block
to cause the stack group to execute things without clobbering the saved
accumulators from the error."
  ;; 04/09/87 CLM - Save the microstack-pushed bit in the call-info word of foothold
  ;;                in order to restore correctly (problem arose in a call-out in
  ;;                a micro-instruction).
  ;; 04/11/88 CLM - Part of the fix for SPR 7417; do not try to signal an error - pass T
  ;;                for inhibit-error arg to SG-MAYBE-GROW-PDLS.
   (let* (sg-ptr
	  (reg-pdl (sg-regular-pdl sg))
	  (old-top-frame (sg-top-frame sg))
	  ;; the new frame's args will start on top of the top frame's call state words.
	  (new-arg-offset (- old-top-frame %call-state-length))
	  ;; the stack group items will be saved as args for the new frame
	  (length-sg (array-leader-length sg))
	  ;; local offset is just above the new frame's args and old top frame's call state words
	  (new-local-offset (+ new-arg-offset (* 2 length-sg) %call-state-length))
	  ;; the top of new frame will include its own call state words
	  (new-top-frame (+ new-local-offset (fef-number-of-locals #'foothold) %call-state-length)))

     ;; Make sure there is room to save SG state
     ;; DON'T signal an error, if pdl had to grow
     (unless suppress-pdl-growing (sg-maybe-grow-pdls sg nil nil nil t))
     ;; Move the previous top most frame's call state words to just before the new frame's local offset
     (do ((i 1 (1+ i)))
	 ((> i %call-state-length))
       (setf (aref reg-pdl (- new-local-offset i)) (aref reg-pdl (- old-top-frame i))))
     ;; Save SG items in new frame's arg area on regular pdl;
     ;; Each Q is saved as 2 words (pointer and tag) to avoid data type problems.
     (do ((i 0 (1+ i))
	  (j new-arg-offset (+ 2 j)))
	 ((>= i length-sg))
       (setq sg-ptr (ap-leader sg i))
       (setf (aref reg-pdl j)
	     (if (%p-pointerp sg-ptr) (%p-contents-as-locative sg-ptr) (%p-pointer sg-ptr)))
       (setf (aref reg-pdl (1+ j))
	     (%p-ldb %%q-all-but-pointer sg-ptr)))
     ;; Need to convert arg and local pointer of frame state we just moved up
     ;;  (I don't think we need to do this.  - pf, May 8, 1986)
;     (let ((prev-top-frame new-local-offset))
;       (setf (sg-top-frame sg) prev-top-frame)
;       (setf (rp-local-pointer reg-pdl prev-top-frame) (pdl-buffer-index sg prev-frame-local-offset))
;       (setf (rp-argument-pointer reg-pdl prev-top-frame) (pdl-buffer-index sg prev-frame-arg-offset)))
     ;; Set new top frame to this new "FOOTHOLD" frame
     (setf (sg-top-frame sg) new-top-frame)
     ;; Set all 5 call-state words for new frame
     ;; FOOTHOLD will be the FEF for this frame, but it should never be called.
     (setf (rp-location-counter-offset reg-pdl new-top-frame) (fef-initial-pc #'foothold))
     (setf (rp-function-word reg-pdl new-top-frame) #'foothold)
     ;; NOTE: PDL-BUFFER-INDEX converts a pdl array index to a pdl buffer index
     (setf (rp-local-pointer reg-pdl new-top-frame) (pdl-buffer-index sg new-local-offset))
     (setf (rp-argument-pointer reg-pdl new-top-frame) (pdl-buffer-index sg new-arg-offset))
     (setf (rp-call-info-word reg-pdl new-top-frame) 0)
     ;;need to save the microstack-pushed bit in the call-info word for exceptional cases
     (when (plusp (sg-restore-microstack sg))
       (setf (rp-call-info-word reg-pdl new-top-frame)
	     (dpb 1 %%call-info-microstack-pushed
			 (rp-call-info-word reg-pdl new-top-frame)))
       (setf (sg-restore-microstack sg) 0))
;;  I don't believe we need to do this anymore.  - pf, July 18, 1986
;     ;;Clear QBBFL left over from previous frame
;     (setf (sg-flags-qbbfl sg) 0)
     ))

1;;; SG-RESTORE-STATE
;;; Undo SG-SAVE-STATE*

(defun sg-restore-state (sg &optional (n-frames-back 1))1       ; Skip the frame for the FH- function.*
  2"Undo SG-SAVE-STATE on SG.  Pop saved state from pdl back into saved accumulators.
N-FRAMES-BACK is the number of stack frames on the pdl to discard
before finding the data to be popped."*
   (let ((reg-pdl (sg-regular-pdl sg))
	 (frame (sg-top-frame sg))
	 (length-sg (array-leader-length sg)))
     (dotimes (i n-frames-back)
       (setq frame (sg-next-frame sg frame)))
     (when (null frame)
       (ferror nil "~S state not saved." sg))
     (when (neq (rp-function-word reg-pdl frame) #'foothold)
       (ferror nil "No saved state (no foothold found) for ~S at ~S[~A]." 
	       sg reg-pdl frame))
     1;; Get starting index to saved SG state*
     (let ((arg-offset (rp-argument-offset sg reg-pdl frame))
	   (local-offset (rp-local-offset sg reg-pdl frame)))
       1;; Store back into SG its saved state found in reg pdl;*
       1;; Each Q was saved as 2 words (pointer and tag) to avoid data type problems.*
       1;; This updates sg-top-frame to point to frame just after foothold*
       (do ((i 0 (1+ i))
	    (j arg-offset (+ 2 j)))
	   ((>= i length-sg))
	 (%p-store-tag-and-pointer (ap-leader sg i)
				   (aref reg-pdl (1+ j))
				   (aref reg-pdl j)))
       1;;  Move the state words back to their old position atop the stack.  Sg-top-frame is set in the loop above.*
       (do ((i 1 (1+ i))
	    (top-frame (sg-top-frame sg)))
	   ((> i %call-state-length))
	 (setf (aref reg-pdl (- top-frame i)) (aref reg-pdl (- local-offset i)))))))


1;; Running things in the other stack group.*

(defun run-sg (sg)
  2"Activate a call block in stack group SG.
The FUNCTION with which the call block was made should resume the error handler
stack group with one argument.  That argument is returned from RUN-SG.
If the value is EH:LOSE, we throw to EH:QUIT.
Should provide this (error handler) stack group as an argument to the function
to be run so it knows who to resume back."*
  (declare (special *error-handler-running* *last-second-level-error-handler-sg*
		    *error-handler-reprint-error* *current-stack-group* *error-object*))  
  (assure-free-space)
  1;; This is used to determine if the "Back to ..." message should be printed*
  (when *error-handler-running*
    (setq *last-second-level-error-handler-sg* *current-stack-group*))
  1;; Turn off the MAR (why??)*
  (setf (sg-flags-mar-mode sg) 0)
  1;; Resume erring stack group, SG.  Save result if resume back to here*
  (setf (sg-current-state sg) sg-state-invoke-call-on-return)
  (let ((result (stack-group-resume sg nil))
	(inner-trap-on-call (sg-flags-trap-on-call sg)))
    1;; Restore old state of SG*
    (sg-restore-state sg)
    1;; If trap-on-call set before returning, leave it on.*
    (unless (zerop inner-trap-on-call) (setf (sg-flags-trap-on-call sg) 1))
    1;; Print "Back to .." message if returning from another EH stack group*
    (when (and *error-handler-running* *error-handler-reprint-error*
	       (neq *last-second-level-error-handler-sg* *current-stack-group*))
      (format t "~%Back to ")
      (print-carefully "error message"
	(send *error-object* :print-error-message sg t *standard-output*))
      (warn-about-special-variables sg)
      (setq *last-second-level-error-handler-sg* *current-stack-group*))
    1;; If the result is EH:LOSE, throw to EH:QUIT*
    (when (and *error-handler-running*
	       (eq result 'lose))
      (throw 'quit nil))
    result))

(defun sg-run-goodbye (sg)
  2"Restart stack group SG, expecting it not to return.
If done in a second-level error handler stack group,
the error handler stack group is marked as free."*
  (declare (special *error-handler-running* *current-stack-group*))  
  (assure-free-space)
1  ;; Resume erring stack group, SG, not expecting to return.*
  (setf (sg-current-state sg) sg-state-invoke-call-on-return)
  (without-interrupts
1    ;; Free this current error handler stack group before resuming erring stack group*
    (when *error-handler-running*
      (free-second-level-error-handler-sg *current-stack-group*))
    (stack-group-resume sg nil)))

(defun sg-run-goodbye-initial-call (sg)
  "Restart stack group SG with the state 'awaiting-initial-call', expecting
it not to return. If done in a second-level error handler stack group,
the error handler stack group is marked as free."
  (declare (special *error-handler-running* *current-stack-group*))  
  (assure-free-space)
  ;; Resume erring stack group, SG, not expecting to return.
  (setf (sg-current-state sg) sg-state-awaiting-initial-call)
  (without-interrupts
    ;; Free this current error handler stack group before resuming erring stack group
    (when *error-handler-running*
      (free-second-level-error-handler-sg *current-stack-group*))
    (stack-group-resume sg nil)))


(defun sg-eval (sg form &optional rebind-streams saved-regpdl-segment saved-specpdl-segment)
  2"Evaluate FORM in stack group SG and return a list of its values.
This is a high-level function, in that SG's state is preserved.
If REBIND-STREAMS is NON-NIL, FORM is evaluated with streams such as
*STANDARD-INPUT* and *STANDARD-OUTPUT* set to the current value
of *TERMINAL-IO* in the stack group that calls SG-EVAL
instead of whatever they are in stack group SG.
SAVED-REGPDL-SEGMENT and SAVED-SPECPDL-SEGMENT are only for use by SG-EVAL-IN-FRAME."*
  (declare (special *current-frame* *error-object* *current-stack-group* 
		    *error-handler-running* *error-depth*))
  (let ((prev-fh (sg-foothold-data sg)))
    1;; Save the current state of SG in a new frame pushed on regpdl*
    (sg-save-state sg)
    (setf (sg-foothold-data sg) (sg-top-frame sg))
    1;; Push a new frame on reg pdl of SG*;
    1;; This new frame's fef will evaluate form*
    1;; Note: SG-FRAME-SINGLE-VALUE knows the position of *error-object* in the list of arguments*
    1;; *Current-frame* is offset by the arguments to the foothold frame, made when we did sg-save-state above,*
    1;; since the state words of the frame are between the arguments and locals of the frame above.*
    (let ((default-args (list form + * / ++ ** // +++ *** /// *values*
			      sg (if (numberp *current-frame*)
				     (+ *current-frame* (* 2 (array-leader-length sg)))
				     nil)
			      *error-object* *current-stack-group* *error-handler-running* prev-fh))
	  (extra-args (list saved-regpdl-segment saved-specpdl-segment *error-depth* *terminal-io*))
	  (function (if rebind-streams #'fh-stream-binding-evaler #'fh-evaler)))
      (push-new-frame sg (sg-top-frame sg) function 
		      (if rebind-streams (nconc default-args extra-args) default-args)))
    1;; Now resume SG and hopefully return*
    (run-sg sg)))

(defun sg-eval-no-trap (sg form &optional rebind-streams)
  2"Like SG-EVAL but suppress trap-on-call."*
  (let ((saved-trap-on-call (sg-flags-trap-on-call sg)))
    (setf (sg-flags-trap-on-call sg) 0)
    (unwind-protect
	(sg-eval sg form rebind-streams)
      (setf (sg-flags-trap-on-call sg) saved-trap-on-call))))

(defun sg-eval-in-frame (sg form frame &optional rebind-streams)
  ;;  3/04/87 CLM - Fixed to prevent crashing when trying to Meta-(Ctrl)-Abort
  ;;                out of a Resume situation.
  "Evaluate FORM in stack group SG in the environment within frame FRAME.
In other respects, like SG-EVAL."
  (declare (special *innermost-visible-frame*))
  (let ((prev-frame (sg-previous-frame sg frame *innermost-visible-frame*)))
    (if (null prev-frame)
	;; If no previous frame just evaluate FORM in SG
	(sg-eval sg form rebind-streams)
      ;; To evaluate FORM in environment of FRAME...
      (let* ((reg-pdl (sg-regular-pdl sg))
	     (regpdl-pointer (sg-regular-pdl-pointer sg))
	     (specpdl-pointer (sg-special-pdl-pointer sg))
;	     (saved-qbbfl (sg-flags-qbbfl sg))
	     (saved-restore-microstack (sg-restore-microstack sg))
	     (saved-top-frame (sg-top-frame sg))
	     (prev-frame-arg-offset (rp-argument-offset sg reg-pdl prev-frame))
	     (regpdl-segment (allocate-saved-pdl-segment
			       (- (1+ regpdl-pointer) prev-frame-arg-offset)))
	     (prev-frame-specpdl-index
	       (1+ (sg-frame-special-pdl-index sg prev-frame)))
	     (specpdl-segment (allocate-saved-pdl-segment
				(- (1+ specpdl-pointer)
				   prev-frame-specpdl-index)))
	     (saved-catch-pointer (sg-catch-pointer sg)))  
	;;reset the microstack bit if necessary
	(when (typep (aref (sg-special-pdl sg)  (1- prev-frame-specpdl-index)) :locative)
	  (setf (sg-restore-microstack sg) 0))
	;; Set up SG regpdl so that FRAME is the innermost frame left.
	(without-interrupts
	  ;; Copy out the data of the frames beyond that one.
	  (%blt-typed (locf (aref (sg-regular-pdl sg) prev-frame-arg-offset)) 
		      (locf (aref regpdl-segment 0)) 
		      (- (1+ regpdl-pointer) prev-frame-arg-offset)
		      1)
	  (%blt-typed (locf (aref (sg-special-pdl sg) prev-frame-specpdl-index)) 
		      (locf (aref specpdl-segment 0)) 
		      (- (1+ specpdl-pointer) prev-frame-specpdl-index)
		      1)
;	  ;; Reset M-FLAGS QBBFL to its state in that FRAME.
;	  (setf (sg-flags-qbbfl sg) (rp-binding-block-pushed (sg-regular-pdl sg) frame))
	 ;;adjust sg-catch-pointer so that in case of abort we know where to go    
	  (do ((x (locf (aref reg-pdl frame)))
	       (y (sg-catch-pointer sg) (%p-contents-offset y %catch-block-saved-catch-pointer))) ;;access catch-block-field
	      ((plusp (%pointer-difference x y))   
	       (setf (sg-catch-pointer sg) y) ) )
	  ;; Mark the inner frames as not there.
	  (setf (sg-top-frame sg) frame)
	  (setf (sg-special-pdl-pointer sg) (1- prev-frame-specpdl-index)))
	(unwind-protect
	    ;; Do the evaluation.
	    (sg-eval sg form rebind-streams regpdl-segment specpdl-segment)
	  ;; Copy back the frames we removed.
	  (without-interrupts
	    (%blt-typed (locf (aref regpdl-segment 0)) 
			(locf (aref (sg-regular-pdl sg) prev-frame-arg-offset)) 
			(- (1+ regpdl-pointer) prev-frame-arg-offset)
			1)
	    (%blt-typed (locf (aref specpdl-segment 0)) 
			(locf (aref (sg-special-pdl sg) prev-frame-specpdl-index)) 
			(- (1+ specpdl-pointer) prev-frame-specpdl-index)
			1)
	    ;; Reset saved QBBFL and RESTORE-MICROSTACK state
;	    (setf (sg-flags-qbbfl sg) saved-qbbfl)
	    (setf (sg-restore-microstack sg) saved-restore-microstack)
	    ;; Point back to inner frames
	    ;;return the catch-pointer
	    (setf (sg-catch-pointer sg) saved-catch-pointer)
	    (setf (sg-top-frame sg) saved-top-frame)
	    (setf (sg-special-pdl-pointer sg) specpdl-pointer)
	    ;; Return the pdl save arrays for re-use.
	    (free-saved-pdl-segment regpdl-segment)
	    (free-saved-pdl-segment specpdl-segment)))))))


(defun sg-apply (sg function arguments &optional no-error-restart goodbye)
  2"Apply FUNCTION to ARGUMENTS in stack group SG.
If NO-ERROR-RETART NIL, CATCH-ERROR-RESTART is done around the APPLY.
If GOODBYE is NON-NIL we expect not to return after apply function.
This is a high-level function, in that SG's state is preserved."*
  (declare (special *current-frame* *error-object* *current-stack-group* *error-handler-running*))
  (let ((prev-fh (sg-foothold-data sg)))
    1;; Save the current state of SG in a new frame pushed on regpdl*
    (sg-save-state sg)
    (setf (sg-foothold-data sg) (sg-top-frame sg))
    1;; Push a new frame on reg pdl of SG;*
    1;; This new frame's fef will apply function and args.*
    1;; *Current-frame* is offset by the arguments to the foothold frame, made when we did sg-save-state above,*
    1;; since the state words of the frame are between the arguments and locals of the frame above.*
    (let ((args (list function arguments + * / ++ ** // +++ *** /// *values*
		      sg (if (numberp *current-frame*)
			     (+ *current-frame* (* 2 (array-leader-length sg)))
			     nil)
		      *error-object* *current-stack-group* *error-handler-running* prev-fh))
	  (function (if no-error-restart #'fh-applier-no-restart #'fh-applier)))
      (push-new-frame sg (sg-top-frame sg) function args))
    1;; Now resume SG (may or may not return)*
    (if goodbye
	(sg-run-goodbye sg)
	(run-sg sg))))

(defun sg-apply-no-trap (sg function arguments &optional no-error-restart goodbye)
  2"Apply FUNCTION to ARGUMENTS in stack group SG, suppressing trap-on-call.
If NO-ERROR-RETART is NIL, CATCH-ERROR-RESTART is done around the APPLY.
If GOODBYE is NON-NIL we expect not to return after apply function.
This is a high-level function, in that SG's state is preserved."*
  (let ((saved-trap-on-call (sg-flags-trap-on-call sg)))
    (setf (sg-flags-trap-on-call sg) 0)
    (unwind-protect
	(sg-apply sg function arguments no-error-restart goodbye)
      (setf (sg-flags-trap-on-call sg) saved-trap-on-call))))

(defun sg-funcall (sg function &rest arguments)
  2"Apply FUNCTION to ARGUMENTS in stack group SG.
This is a high-level function, in that SG's state is preserved."*
  (sg-apply sg function arguments))

(defun sg-funcall-no-restart (sg function &rest arguments)
  2"Apply FUNCTION to ARGUMENTS in stack group SG.
Does not do a CATCH-ERROR-RESTART around the apply.
This is a high-level function, in that SG's state is preserved."*
  (sg-apply sg function arguments t))

(defun sg-funcall-no-trap (sg function &rest arguments)
  2"Apply FUNCTION to ARGUMENTS in stack group SG, suppressing trap-on-call.
This is a high-level function, in that SG's state is preserved."*
  (let ((old-trap-on-call (sg-flags-trap-on-call sg)))
    (setf (sg-flags-trap-on-call sg) 0)
    (unwind-protect
	(sg-apply sg function arguments)
      (setf (sg-flags-trap-on-call sg) old-trap-on-call))))

(defun sg-funcall-in-frame (sg frame function &rest arguments)
  2"Apply ARGUMENTS to FUNCTION in stack group SG in the environment within frame FRAME.
In other respects, like SG-APPLY."*
  ;; 05/12/87 CLM - Two fixes: 1) after we have copied out the top of the reg and spec pdl,
  ;;                we need to check the top of the specpdl; if it's a fixnum we need to 
  ;;                set the sg-restore-microstack to 1 so that if we abort we know we
  ;;                have to flush this upcs off the specpdl; 2) we also have to save the
  ;;                sg-catch-pointer, again in case of aborting we know where to throw back
  ;;                to.
  (declare (special *innermost-visible-frame*))
  (let ((prev-frame (sg-previous-frame sg frame *innermost-visible-frame*)))
    (if (null prev-frame)
	1;; If no previous frame just apply FORM in SG.*
	(sg-apply sg function arguments)
1      ;; To apply FORM in environment of FRAME...*
      (let* ((reg-pdl (sg-regular-pdl sg))
	     (regpdl-pointer (sg-regular-pdl-pointer sg))
	     (specpdl-pointer (sg-special-pdl-pointer sg))
;	     (saved-qbbfl (sg-flags-qbbfl sg))
	     (saved-restore-microstack (sg-restore-microstack sg))
	     (saved-top-frame (sg-top-frame sg))
	     (prev-frame-arg-offset (rp-argument-offset sg reg-pdl prev-frame))
	     (regpdl-segment (allocate-saved-pdl-segment
			       (- (1+ regpdl-pointer) prev-frame-arg-offset)))
	     (prev-frame-specpdl-index
	       (1+ (sg-frame-special-pdl-index sg prev-frame)))
	     (specpdl-segment (allocate-saved-pdl-segment
				(- (1+ specpdl-pointer)
				   prev-frame-specpdl-index)))
	     (saved-catch-pointer (sg-catch-pointer sg)))
	1;; Set up SG regpdl so that FRAME is the innermost frame left.*
	(without-interrupts
	1  ;; Copy out the data of the frames beyond that one.*
	  (%blt-typed (locf (aref (sg-regular-pdl sg) prev-frame-arg-offset)) 
		      (locf (aref regpdl-segment 0)) 
		      (- (1+ regpdl-pointer) prev-frame-arg-offset)
		      1)
	  (%blt-typed (locf (aref (sg-special-pdl sg) prev-frame-specpdl-index)) 
		      (locf (aref specpdl-segment 0)) 
		      (- (1+ specpdl-pointer) prev-frame-specpdl-index)
		      1)
	 ;;adjust sg-catch-pointer so that in case of abort we know where to go    
	  (do ((x (locf (aref reg-pdl frame)))
	       (y (sg-catch-pointer sg) (%p-contents-offset y %catch-block-saved-catch-pointer))) ;;access catch-block-field
	      ((plusp (%pointer-difference x y))   
	       (setf (sg-catch-pointer sg) y) ) )

;	  (setf (sg-flags-qbbfl sg) saved-qbbfl)
	1  ;; Mark the inner frames as not there.*
	  (setf (sg-top-frame sg) frame)
	  (setf (sg-special-pdl-pointer sg) (1- prev-frame-specpdl-index)))
	  ;1; Keep track of sg-restore-microstack*
	  (if (typep (aref (sg-special-pdl sg) (sg-special-pdl-pointer sg)) 'fixnum)
	      (setf (sg-restore-microstack sg) 1)
	    (setf (sg-restore-microstack sg) 0))

	(unwind-protect
	1    ;; Do the apply of function and args*
	    (sg-apply sg function arguments)
	1  ;; Copy back the frames we removed.*
	  (without-interrupts
	    (%blt-typed (locf (aref regpdl-segment 0)) 
			(locf (aref (sg-regular-pdl sg) prev-frame-arg-offset)) 
			(- (1+ regpdl-pointer) prev-frame-arg-offset)
			1)
	    (%blt-typed (locf (aref specpdl-segment 0)) 
			(locf (aref (sg-special-pdl sg) prev-frame-specpdl-index)) 
			(- (1+ specpdl-pointer) prev-frame-specpdl-index)
			1)
	1    ;; Reset saved QBBFL and RESTORE-MICROSTACK state*
;	    (setf (sg-flags-qbbfl sg) saved-qbbfl)
	    (setf (sg-restore-microstack sg) saved-restore-microstack)
	1    ;; Point back to inner frames *
	    (setf (sg-top-frame sg) saved-top-frame)
	    (setf (sg-special-pdl-pointer sg) specpdl-pointer)
	    (setf (sg-catch-pointer sg) saved-catch-pointer)
	1    ;; Return the pdl save arrays for re-use.*
	    (free-saved-pdl-segment regpdl-segment)
	    (free-saved-pdl-segment specpdl-segment)))))))


(defun sg-throw (sg label value &optional ignore)
   2"Throw VALUE to LABEL in stack group SG."*
   (declare (special *allow-pdl-grow-message*))
   (let ((*pdl-grow-ratio* 1))
     (declare (special *pdl-grow-ratio*))
     1;; Don't make the stack much bigger; we just want room to pop it;*
     1;; and don't signal a pdl-overflow error if stack did grow*
     (sg-maybe-grow-pdls sg *allow-pdl-grow-message* 1024. 0 t))
   1;; Save the current state of SG in a new frame pushed on regpdl*
   (sg-save-state sg t)
   1;; Push a new frame on reg pdl of SG;*
   1;; This new frame's fef will do the throw*
   (let ((args (list label value (sg-flags-trap-on-call sg))))
     (push-new-frame sg (sg-top-frame sg) #'fh-thrower args))
   1;; Don't allow trap on this frame*
   (setf (sg-flags-trap-on-call sg) 0)
   1;; Resume sg and expect not to return*
   (sg-run-goodbye sg))


1;; UNWINDING STACK *

(defun sg-unwind (sg label value count action disposal)
  2"Unwind, or prepare to unwind, stack group SG.
LABEL, VALUE, COUNT and ACTION are like the args to *UNWIND-STACK.
DISPOSAL is either EH:SETUP to just set up the call, EH:CALL to make the call but not free the EH,
or EH:FREE to make the call and free the EH"*
  (declare (special *error-handler-running* *current-stack-group*))  
  1;; Save the current state of SG in a new frame pushed on regpdl*
  (sg-save-state sg)
  1;;* 1Need to make up for the frame pushed by SG-SAVE-STATE.*
  (when count (incf count))
  1;; Push a new frame on reg pdl of SG;*
  1;; This new frame's fef will do the unwind*
  (let ((args (list label value count action (sg-flags-trap-on-call sg))))
    (push-new-frame sg (sg-top-frame sg) #'fh-unwinder args))
  1;; Don't allow trap on this frame*
  (setf (sg-flags-trap-on-call sg) 0)
  1;; Setup current state of SG,*
  1;; free current EH stack group if FREE state, and*
  1;; resume SG if not SETUP only state*
  (setf (sg-current-state sg) sg-state-invoke-call-on-return)
  (without-interrupts
    (if (and *error-handler-running* (eq disposal 'free))
	(free-second-level-error-handler-sg *current-stack-group*))
    (unless (eq disposal 'setup) 
      (stack-group-resume sg nil))))

(defun sg-unwind-to-frame (sg dest-frame goodbye-p &optional value (label t))
  2"Unwind the stack group, SG, until DEST-FRAME is the innermost frame left.
If GOODBYE-P is NON-NIL, free the error handler stack group and unwind SG to DEST-FRAME.
If GOODBYE-P is NIL, unwind SG to DEST-FRAME and resume back to the error handler stack group.
VALUE is the value to return from DEST-FRAME.
LABEL is a catch tag.  If LABEL is specified, we throw VALUE to tag LABEL in SG."*
  ;; 03/10/87 CLM - Call sg-frame-special-pdl-index with FRAME rather than *CURRENT-FRAME* to
  ;;                get correct special-pdl index.
  (declare (special *innermost-visible-frame* *error-locus-frame* *current-frame*
		    *current-stack-group*))
  (assure-free-space)
  (when (> *innermost-visible-frame* (sg-top-frame sg))
    (setf (sg-top-frame sg) *innermost-visible-frame*))
  1;; Get number of frames till at DEST-FRAME*
  (let ((num-frames (do ((frame (sg-top-frame sg) (sg-next-frame sg frame))
			 (n 1 (1+ n)))
			((= frame dest-frame) n))))
    1;; UNWIND to DEST-FRAME*
    (sg-unwind sg label value num-frames 
	       1;; The action passed to *unwind-sg (whether or not to resume back to error handler SG)*
	       (if goodbye-p nil *current-stack-group*) 
	       1;; FREE -> free EH stack group and unwind SG (don't expect to come back to EH);*
	       1;; CALL -> unwind SG *
	       (if goodbye-p 'free 'call))
    (when (null goodbye-p)
      1;; Flush the call back to this SG *
      1;; (ie, the frame for the funcall of the current EH stack group should be skipped)*
      (let ((rp (sg-regular-pdl sg)) 
	    (frame (sg-top-frame sg)))
	(unless (and (eq (rp-function-word rp frame) #'si:call-stack-group)
		     (> (rp-number-of-arguments rp frame) 0)
		     (eq (aref rp (rp-argument-offset sg rp frame)) *current-stack-group*))
	  (format t "~&Warning:  While returning from SG-UNWIND:~@
		       Top frame should be call to ~S, but was call of ~S.~%"
		  *current-stack-group*
		  (rp-function-word rp frame)))
	(setq *current-frame* (sg-next-frame sg frame))	1; Clean up the regular pdl.*
	(setf (sg-special-pdl-pointer sg) (sg-frame-special-pdl-index sg frame))
	1;;  Copy the state words over the arguments to si:call-stack-group, which never has zero arguments.*
	(let ((starting-address (locf (rp-call-info-word rp *current-frame*))))
	  (%blt-typed starting-address
		      (%make-pointer dtp-locative (- (%pointer starting-address)
						     (rp-number-of-arguments rp frame)))
		      %call-state-length
		      1))
	(decf *current-frame* (rp-number-of-arguments rp frame))	1; Adjust after the copy.*
	(setf (sg-restore-microstack sg) (rp-micro-stack-pushed rp frame))
	(setf (sg-top-frame sg) *current-frame*)
	(setq *error-locus-frame* (min *error-locus-frame* *current-frame*))
1;;  I don't believe we need to do this anymore.  - pf, July 18, 1986*
;	1;; QBBFL must correspond to current frame to work!*
;	(setf (sg-flags-qbbfl sg) (rp-binding-block-pushed rp *current-frame*))
	))))

(defun sg-unwind-to-frame-and-reinvoke (sg dest-frame
					&optional (form (get-frame-function-and-args sg dest-frame)))
  "Unwind stack group SG until DEST-FRAME is the innermost frame, and prepare to reinvoke it.
If FORM is specified, it is what to execute instead of what DEST-FRAME was executing previously.
Uses STACK-GROUP-RESUME to make SG take off and run."
  ;; 03/25/87 CLM - Fix to set the binding-block-pushed bit in the call-info word to zero when we
  ;;                remove all special bindings (previously it was setting qbbfl flag to zero).
  ;; 05/05/87 CLM - After unwinding the stack, check the sg-state word to see if there is a saved
  ;;                microstack that needs to be popped.  Previously we checked the call-info-word 
  ;;                bit for this, but because of another correction this was no longer valid.  This 
  ;;                fixes the problems encountered in trying to resume from a too-few/many args error.
  ;;                In theses cases where we don't want to resume the microcode we must make sure we
  ;;                flush any saved microstacks on the specpdl.  Also add test to make sure second element
  ;;                of a binding pair is a locative.
  ;; 05/20/88 CLM - Added CLOS support.
  ;; 02/21/89 clm - ran into a situation where we want to return values from a CLOS method
  ;;                via ctrl-r.  In this case, the method is at dest-frame and has mapping tables
  ;;                etc., but the function we are going to invoke is VALUES.  Before we assume we
  ;;                are going to reinvoke a CLOS method should make sure that's the case, and if not,
  ;;                modify the original call-info-word to a non-clos form.
  ;; 03/20/89 clm - Integrated CLOS changes into EH.
  (declare (special *innermost-visible-frame* ))
  (let ((rp (sg-regular-pdl sg)))
    (assure-free-space)
    ;; Unwind back to point where frame to be retried is about to return.
    ;; This gets rid of its unwind-protects but not its special bindings
    (let ((old-trap-on-call (sg-flags-trap-on-call sg)))
      (setf (sg-flags-trap-on-call sg) 0)
      (sg-unwind-to-frame sg dest-frame nil)
      (setf (sg-flags-trap-on-call sg) old-trap-on-call))
    ;;  Unwinding throws away the args of the last frame unwound.  If the frame immediately above
    ;;  dest-frame was for foothold, the "args" are the state of the stack-group saved by sg-save-state
    ;;  which were stored as two words for each element of the leader of the stack-group.  Thus, we
    ;;  check against the greater of call-arguments-limit, for normal functions, and the size of the foothold
    ;;  state, for the case of unwinding back to the top frame after a trap.  - pf, July 11, 1986
    (unless (<= 0 (- dest-frame (sg-top-frame sg)) (max call-arguments-limit (* 2 (array-leader-length sg))))
      (format t "~&Warning:  While returning from sg-unwind-to-frame, ~@
		   DEST-FRAME: ~A, ~A  and TOP-FRAME: ~A, ~A " 
	      dest-frame (rp-function-word rp dest-frame)
	      (sg-top-frame sg) (rp-function-word rp (sg-top-frame sg))))
    ;;  Unwinding throws away the args of the unwound frame immediately above the destination frame,
    ;;  so dest-frame will always be greater than sg-top-frame when unwinding to a non-top frame.  Since
    ;;  this is known behavior of *unwind-stack (and different from vm1), we'll just adjust dest-frame here.
    ;;     - pf, May 8, 1986
    (setq dest-frame (sg-top-frame sg))
    ;; Next line prevents total disaster if error in the code below
    (setq *innermost-visible-frame* (sg-top-frame sg))
    ;; Get rid of the saved microstack in special pdl for DEST-FRAME.
    ;; There will at least be an entry for XUWR1+1.
   ;; Changed the test from checking the call-info-word to checking the sg-state word to see if
   ;; there is a saved microstack. This is the result of a previous fix.
    (let ((start-idx (sg-frame-special-pdl-range sg dest-frame)))
      (unless (zerop (sg-restore-microstack sg) )
	(do ((sp (sg-special-pdl sg))
	     (spp (sg-special-pdl-pointer sg) (1- spp))
	     (loc))
	    (nil)
	  (setq loc (locf (aref sp spp)))
	  (when (neq (%p-data-type loc) dtp-fix)
	    (ferror nil "Where's my saved microstack?"))
	  (when (not (zerop (%p-ldb %%specpdl-block-start-flag loc)))
	    (return (setf (sg-special-pdl-pointer sg) (1- spp)))))
	;; No longer any saved microstack in special pdl
	(setf (sg-restore-microstack sg) 0))
      ;; Now flush any special bindings made within DEST-FRAME,
      ;; but keep any that were made by closures, flavor instances, etc.
      ;; on entry to DEST-FRAME.
      (when start-idx
	(do ((sp (sg-special-pdl sg))
	     ;; SPP is the index of the first word of a binding pair.
	     ;; So is START-IDX.
	     (spp (1- (sg-special-pdl-pointer sg)) (- spp 2))
	     (loc))
	    ((or (< spp start-idx)
		 (not (zerop (%p-ldb %%specpdl-closure-binding (locf (aref sp spp))))))
	     ;; SPP points to a pair that doesn't belong to DEST-FRAME
	     ;; or else should not be flushed.
	     (setf (sg-special-pdl-pointer sg) (1+ spp))
	     ;; If there was a binding block and we flushed the whole thing,
	     ;; clear the flag saying DEST-FRAME has made a binding block.
	     (when (< spp start-idx)
	       (setf (rp-binding-block-pushed rp dest-frame) 0))
	     ;;set the bit in the call-info word instead of the qbbfl flag
	     ;;(when (< spp start-idx)
	     ;;  (setf (sg-flags-qbbfl sg) 0))
	     )
	  (setq loc (locf (aref sp (1+ spp))))
	  ;;make sure that the second element of the binding pair is a locative and
	  ;;if not, error out. 
	  (when (neq (%p-data-type loc) dtp-locative)
	    (ferror nil "The last word of a binding pair wasn't a locative")))))
    ;; Make sure exit pc is within bounds
    (let ((function (rp-function-word rp dest-frame)))
      (when (and (typep function 'compiled-function)
		 (> (rp-location-counter-offset rp dest-frame) (* 2 (fef-length function))))
	(setf (rp-location-counter-offset rp dest-frame) 0)))
    ;; Now rebuild DEST-FRAME as if it was about to be called
    ;; Redo DEST-FRAME with function and args in form
    (let ((func-to-reinvoke (if (si:validate-function-spec (car form))
				(fdefinition (car form))
				(car form)))
	  (args (cdr form))
	  (next-frame (sg-next-frame sg dest-frame))
	  (call-info-word (rp-call-info-word rp dest-frame)))
      ;;check for clos methods; must repush the mapping-tables and the continuation if present;
      ;;the number of mapping-tables and the local offset for the continuation are
      ;;kept in the functions debug-info
      (unless (zerop (ldb si:%%call-info-clos-info-provided call-info-word))
	(let* ((debug-info (extract-debug-info-struct-from-fef func-to-reinvoke ))
	       (map-slots (get-debug-info-field debug-info :map-slots))
	       (continuation-slot (get-debug-info-field debug-info :continuation-slot)))
	  ;;if there is no continuation (as for a combined method) we must still push a
	  ;;NIL in that slot
	  (if map-slots
	      (progn
		(if continuation-slot
		    (push-end (sg-frame-local-value sg dest-frame continuation-slot) args)
		    (push-end nil args))
		(unless (zerop map-slots)
		  (push-end (sg-frame-local-value sg
						  dest-frame
						  sys:local-for-first-mapping-table) args)
		  (when (> map-slots 1)
		    (dotimes (local (1- map-slots))
		      (push-end (sg-frame-local-value sg
						      dest-frame
						      (+ local sys:locals-for-mapping-table-base))
				args)))))
	      ;; clm 02/21/89
	      (setq call-info-word (si:%logdpb 0 si:%%call-info-clos-info-provided call-info-word))
	      )
	  
	    ))
      (push-new-frame sg next-frame func-to-reinvoke args
		      call-info-word (rp-argument-offset sg rp dest-frame)))
    ;; Now send the SG on its way
    (setf (sg-current-state sg) sg-state-invoke-call-on-return)))

(defun unwind-sg (sg continuation argument process-p)
  2"The CONTINUATION is a function called with one argument, ARGUMENT, 
in the newly reset stack group, SG. 
If PROCESS-P is NON-NIL, rather than calling the function now, in this process,
we simply leave SG in such a state that the next time it is called,
e.g. by the scheduler, it will call it.
If the current state of SG is awaiting initial call, unwind SG
with value ARGUMENT and action CONTIONUATION."*
  (setf (sg-inst-disp sg) 0)1  ;sg-main-dispatch*
  (let ((current-state (sg-current-state sg)))
    (cond ((not (or (= current-state sg-state-awaiting-initial-call)
		    (= current-state 0)))
	1   ;; SG has been run, unwind it*
	   (sg-unwind sg t argument nil continuation (if process-p 'setup 'call)))
	  (t
	   1;; SG has not been run, don't unwind, but do leave in same state*
	   (stack-group-preset sg continuation argument)
	   (unless process-p (stack-group-resume sg nil))))
    (unless process-p
      (setf (sg-current-state sg) sg-state-exhausted))))


1;;
;;  Macro for stack-consing used below.*
(defmacro condition-resume-multiple-if (cond-form handlers &body body)
   (let ((tag (gensym)))
     `(with-stack-list* (,tag ,@handlers *condition-resume-handlers*)
	(let-if ,cond-form
		((*condition-resume-handlers* ,tag))
	  ,@body))))

1;; The FH- functions are those intended to run in the other stack group.
;; Those that come back should be started up with RUN-SG.
;; Those that do not come back should be started up with SG-RUN-GOODBYE.
;; They must be given the error handler stack group as an argument
;; so that they can call it back.  This they must do without making any other
;; intervening active call blocks on the stack, so that the foothold data
;; can be found from the TOP frame when it returns.
;; They must also be given the value of ERROR-HANDLER-RUNNING as an argument, 
;; so that if it is T they can do an unwind protect that does a
;; FREE-SECOND-LEVEL-ERROR-HANDLER-SG on the stack group that they 
;; aren't going to return to in that case. 
;; They must also be given the previous foothold's offset so that
;; SG-FOOTHOLD-DATA can be reset in case of a throw.*

(defun fh-evaler (form x+ x* x/ x++ x** x// x+++ x*** x/// x-values
		  x-erring-sg x-current-frame x-error-object sg eh-running-p prev-fh)
  (declare (special *condition-handlers* *condition-default-handlers* *condition-resume-handlers*
		    *error-depth* *current-stack-group*))
  (unwind-protect
      (let ((tag (list "Return to debugger level ~D." *error-depth*)))
	(catch tag
	  1;; Resume back to SG after evaluating form*
	  (stack-group-resume
	    sg
	    1;; Note: no special variables should be bound*
	    1;; outside this point (when we return to the error handler SG).*
	    (let-if eh-running-p
		    1;; Only bind these if error handler is running*
		    ((*condition-handlers* nil)
		     (*condition-default-handlers* nil))
	      1;;  Have to stack-cons the link to *condition-resume-handlers* to avoid copying*
	      1;;  all the earlier stack-consed handlers into the heap, which is certainly slow*
	      1;;  and might also damage eq-ness.  - pf, Sept 17, 1986*
	      (condition-resume-multiple-if eh-running-p
					    (`((sys:abort error) ,tag t ,tag
					       si:catch-error-restart-throw ,tag)
					     t)	1; Note the T barrier.*
		(let ((+ x+) (++ x++) (+++ x+++) (* x*) (** x**) (*** x***)
		      (/ x/) (// x//) (/// x///) (- form)
		      (*values* x-values)
		      (*erring-sg* x-erring-sg) (*erring-frame* x-current-frame) (*error-object* x-error-object)
		      (*evalhook* nil) (*applyhook* nil) (*errset-status* nil))
		  (declare (special *erring-sg* *erring-frame* *error-object* *errset-status*))
		  (let-if (area-temporary-p default-cons-area)
			  ((default-cons-area sys:background-cons-area))
		    1;; EVAL FORM* 1!!*
		    (with-self-variables-bound
		      (multiple-value-list (eval form)))))))))
	1;; This is in case the CATCH catches.  The VALUES is here to make sure it compiles into a misc-op.*
	(values (stack-group-resume sg 'lose)))
    1;; Restore foothold data and free EH stack group*
    (setf (sg-foothold-data *current-stack-group*) prev-fh)
    (when eh-running-p
      (free-second-level-error-handler-sg sg))))

(defun fh-stream-binding-evaler (form x+ x* x/ x++ x** x// x+++ x*** x/// x-values
				 x-erring-sg x-current-frame x-error-object
				 sg eh-running-p prev-fh
				 regpdl-segment specpdl-segment *error-depth*	1; Must be *error-depth*.  - pf*
				 eh-terminal-io)
  (declare (special *condition-handlers* *condition-default-handlers* *condition-resume-handlers* 
		     *current-stack-group*))
  (unwind-protect
      (let* ((tag (list "Return to debugger level ~D." *error-depth*))
	     win-p result)
	(catch tag
	  1;; Note: no special variables should be bound*
	  1;; outside this point (when we return to the error handler sg).*
	  (let-if eh-running-p
		  1;; Only bind these if error handler is running*
		  ((*condition-handlers* nil)
		   (*condition-default-handlers* nil))
	    1;;  Have to stack-cons the link to *condition-resume-handlers* to avoid copying*
	    1;;  all the earlier stack-consed handlers into the heap, which is certainly slow*
	    1;;  and might also damage eq-ness.  - pf, Sept 17, 1986*
	    (condition-resume-multiple-if eh-running-p
					  (`((sys:abort error) ,tag t ,tag
					     si:catch-error-restart-throw ,tag)
					   t)
	      (let ((+ x+) (++ x++) (+++ x+++) (* x*) (** x**) (*** x***)
		    (/ x/) (// x//) (/// x///) (- form)
		    (*values* x-values)
		    (*erring-sg* x-erring-sg) (*erring-frame* x-current-frame) (*error-object* x-error-object)
		    (*terminal-io* eh-terminal-io)
		    (*standard-input* si:syn-terminal-io)
		    (*standard-output* si:syn-terminal-io)
		    (*query-io* si:syn-terminal-io)
		    (*error-output* si:syn-terminal-io)
		    (*evalhook* nil) (*applyhook* nil) (*errset-status* nil)
		    (rubout-handler nil))
		(declare (special *erring-sg* *erring-frame* *error-object* *errset-status*)) 
		(let-if (area-temporary-p default-cons-area)
			((default-cons-area sys:background-cons-area))
		  1;; Eval FORM !!*
		  (with-self-variables-bound
		    (setq result (multiple-value-list (si:eval-abort-trivial-errors form))
			  win-p t)))))))
	1;; Resume back to SG.*
	(if win-p
	    (values (stack-group-resume sg result))	;1 The VALUES is here to make sure it compiles into a misc-op.*
	    1;; This is in case the CATCH catches.*
	    (values (stack-group-resume sg 'lose))))	;1 The VALUES is here to make sure it compiles into a misc-op.*
    1;; Need to make sure certain things are freed for reuse*
    (when regpdl-segment (free-saved-pdl-segment regpdl-segment))
    (when specpdl-segment (free-saved-pdl-segment specpdl-segment))
    (setf (sg-foothold-data *current-stack-group*) prev-fh)
    (when eh-running-p (free-second-level-error-handler-sg sg))))

(defun fh-applier (fn args x+ x* x/ x++ x** x// x+++ x*** x/// x-values
		   x-erring-sg x-current-frame x-error-object sg eh-running-p prev-fh)
  (declare (special *condition-handlers* *condition-default-handlers* *condition-resume-handlers*
		    *current-stack-group* *error-depth*))
  (unwind-protect
      (let* ((tag (list "Return to debugger level ~D." *error-depth*)))
	(catch tag
	  1;; Resume back to SG after evaluating form*
	  (stack-group-resume
	    sg
	    1;; Note: no special variables should be bound*
	    1;; outside this point (when we return to the error handler sg).*
	    (let-if eh-running-p
		    1;; Only bind these if error handler is running*
		    ((*condition-handlers* nil)
		     (*condition-default-handlers* nil))
	      1;;  Have to stack-cons the link to *condition-resume-handlers* to avoid copying*
	      1;;  all the earlier stack-consed handlers into the heap, which is certainly slow*
	      1;;  and might also damage eq-ness.  - pf, Sept 17, 1986*
	      (condition-resume-multiple-if eh-running-p
					    (`((sys:abort error) ,tag t ,tag
					       si:catch-error-restart-throw ,tag)
					     t)
		(let ((+ x+) (++ x++) (+++ x+++) (* x*) (** x**) (*** x***)
		      (/ x/) (// x//) (/// x///)
		      (*values* x-values)
		      (*erring-sg* x-erring-sg) (*erring-frame* x-current-frame) (*error-object* x-error-object)
		      (*evalhook* nil) (*applyhook* nil) (*errset-status* nil))
		  (declare (special *erring-sg* *erring-frame* *error-object* *errset-status*)) 
		  (let-if (area-temporary-p default-cons-area)
			  ((default-cons-area sys:background-cons-area))
		    1;; Apply function !!*
		    (multiple-value-list (apply fn args))))))))
	1;; This is in case the CATCH catches.  The VALUES is here to make sure it compiles into a misc-op.*
	(values (stack-group-resume sg 'lose)))
    1;; This is reached only if we throw through this frame.*
    1;; Need to restore foothold data and free EH stack group*
    (setf (sg-foothold-data *current-stack-group*) prev-fh)
    (when eh-running-p (free-second-level-error-handler-sg sg))))

(defprop fh-applier-no-restart t :error-reporter)

(defun fh-applier-no-restart (fn args ignore ignore ignore ignore ignore ignore ignore
			      ignore ignore ignore ignore ignore error-object
			      sg eh-running-p prev-fh)
  (declare (special *current-stack-group*)
	   (ignore error-object))
  (unwind-protect
      1;; Resume back to SG after applying function.  The VALUES is here to make sure it compiles into a misc-op.*
      (values (stack-group-resume sg (multiple-value-list (apply fn args))))
    1;; This is reached only if we throw through this frame.*
    1;; Need to restore foothold data and free EH stack group*
    (setf (sg-foothold-data *current-stack-group*) prev-fh)
    (when eh-running-p (free-second-level-error-handler-sg sg))))

(defvar *fh-aux-sg* (make-stack-group "Foothold Aux SG"))

(defun set-trap-on-call (sg)
1  ;; Used by FH-THROWER and FH-UNWINDER
  ;; This is run in stack group FH-AUX-SG, to
  ;; set the trap on call flag of the stack group
  ;; which FH-THROWER or FH-UNWINDER was running in.*
  (setf (sg-flags-trap-on-call sg) 1)
  (stack-group-return nil))

(defun fh-thrower (label value trap-on-call-flag)
1  ;; Set trap-on-call flag for current sg, 
  ;; if it was set before this function was called*
  (declare (special *fh-aux-sg* *current-stack-group*))
  (unless (zerop trap-on-call-flag)
    (without-interrupts
      (stack-group-preset *fh-aux-sg* 'set-trap-on-call *current-stack-group*)
      (funcall *fh-aux-sg* nil)))
  (throw label value))

(defun fh-unwinder (label value count action trap-on-call-flag)
1  ;; Set trap-on-call flag for current sg, 
  ;; if it was set before this function was called*
  (declare (special *fh-aux-sg* *current-stack-group*))
  (unless (zerop trap-on-call-flag)
    (without-interrupts
      (stack-group-preset *fh-aux-sg* 'set-trap-on-call *current-stack-group*)
      (funcall *fh-aux-sg* nil)))
  (*unwind-stack label value count action))
