;; -*- Mode:Common-Lisp; Package:TV; Base:8; Fonts:(CPTFONT HL12B TR10BI) -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;; Copyright (c) 1981 Massachusetts Institute of Technology 
;; Copyright (c) 1984-1989 Texas Instruments Incorporated.  All Rights Reserved.

;; Alternate Input editor (Rubout Handler) Command Loop

;; This function gets called whenever a :TYI message is sent to the stream and we are
;; inside the rubout handler.  If a normal character is typed, it is echoed, put in the
;; buffer, and returned.  If a rubout or any other editing character is typed, any number
;; of editing commands are processed by modifying the buffer, then, when the first
;; non-editing character is typed, a throw is done back to the top level of the read
;; function and the buffered input is re-scanned.  The character must be typed at the end
;; of the line in order for the throw to take place.  ** The kludges labelled below may be
;; flushed if this rubout handler is more carefully integrated with the rest of the
;; system. **

1;; Several modifications for UCL features were made to this command loop:
;; 1. Rh command look-up now uses a UCL rubout handler command table.  This table may be
;;    editted by the user to rebind keystrokes.
;; 2. Rubout handler preempts on UCL application keystrokes.  While
;;    typing a form, the user can execute an application command without clearing input.*

(DEFUN-RH alternate-rubout-handler ()
  (LET ((CH) (CH-CHAR) (CH-CONTROL-META) (COMMAND) (keys-read)
	(FILL-POINTER (RH-FILL-POINTER))
	(TYPEIN-POINTER (RH-TYPEIN-POINTER))
	(INITIAL-ENTRY (RHB-INITIAL-ENTRY))
	(RUBBED-OUT-SOME? NIL)
	(INITIAL-INPUT (CADR (ASSOC :INITIAL-INPUT RUBOUT-HANDLER-OPTIONS :TEST #'EQ)))
	(INITIAL-INPUT-POINTER (CADR (ASSOC :INITIAL-INPUT-POINTER RUBOUT-HANDLER-OPTIONS :TEST #'EQ)))
	(NUMERIC-ARG NIL)
	(NUMERIC-ARG-NEGATIVE NIL))
    (SETF (RHB-INITIAL-ENTRY) NIL)
    ;; Kludge #1.  If this is the first time this rubout handler has been invoked
    ;; in this stream, then we must create the input history.
    (OR (RH-INPUT-RING)
	(WHEN (FBOUNDP 'ZWEI:MAKE-HISTORY)
	  (SETF (RH-INPUT-RING) (RH-MAKE-INPUT-RING))))
    (WHEN INITIAL-ENTRY
      ;; save the previous input on the input history,
      ;; unless the previous read said not to save it.
      (COND ((AND (NOT (RH-DONT-SAVE-FLAG))
		  TYPEIN-POINTER
		  (NEQ INITIAL-ENTRY :RESTORED)
		  (NOT (ZEROP TYPEIN-POINTER)))
	     ;; only add the contents if it is different from the last entry, and
	     ;; the entry is at least 2 characters long.
	     (SETF (FILL-POINTER RUBOUT-HANDLER-BUFFER) TYPEIN-POINTER)
	     (WHEN (AND (> TYPEIN-POINTER 1)
			(FBOUNDP 'ZWEI:HISTORY-LATEST-ELEMENT)
			(MISMATCH RUBOUT-HANDLER-BUFFER
				  (ZWEI:HISTORY-LATEST-ELEMENT (RH-INPUT-RING))))
	       (ZWEI:PUSH-ON-HISTORY (SUBSEQ RUBOUT-HANDLER-BUFFER 0 TYPEIN-POINTER)
				     (RH-INPUT-RING)))
	     (SETF (FILL-POINTER RUBOUT-HANDLER-BUFFER) FILL-POINTER)))
      ;; Then initialize the typein pointer.
      (SETF (RH-TYPEIN-POINTER) FILL-POINTER)
      (SETQ TYPEIN-POINTER FILL-POINTER)
      ;; Gobble the initial input if any.
      (WHEN (AND INITIAL-INPUT (NEQ INITIAL-ENTRY :RESTORED))
	(RH-INSERT-STRING INITIAL-INPUT 0 NIL NIL NIL)
	(SETQ RUBBED-OUT-SOME? T FILL-POINTER (RH-FILL-POINTER))
        (SETQ INITIAL-INPUT-POINTER
	      (MAX (MIN (OR INITIAL-INPUT-POINTER TYPEIN-POINTER)
			(LENGTH RUBOUT-HANDLER-BUFFER)) 0))
	(SETF TYPEIN-POINTER INITIAL-INPUT-POINTER
	      (RH-TYPEIN-POINTER) INITIAL-INPUT-POINTER)
 	(RH-SET-POSITION TYPEIN-POINTER))
      ;; Record whether this unit of input should be saved on the history.
      (SETF (RH-DONT-SAVE-FLAG)
	    (OR (CADR (ASSOC :DONT-SAVE RUBOUT-HANDLER-OPTIONS :TEST #'EQ))
		(CADR (ASSOC :NO-INPUT-SAVE RUBOUT-HANDLER-OPTIONS :TEST #'EQ))))
      )
    ;;; Can this ever go off? :pass-though now only allows non-bucky. -- mly ;;;
    ;; Kludge #5.  We can't echo or rub out a bucky char or a bli,
    ;; so if the last char inserted was a either of those
    ;; and it did not terminate the input, flush it.
    (AND (NOT (ZEROP TYPEIN-POINTER))
	 (OR (CONSP (AREF RUBOUT-HANDLER-BUFFER (1- TYPEIN-POINTER)))
	     (Not (Zerop (CHAR-BITS (Aref RUBOUT-HANDLER-BUFFER (1- Typein-Pointer))))))
	 (SETQ TYPEIN-POINTER (SETF (RH-TYPEIN-POINTER) (DECF (RH-FILL-POINTER)))))
    ;; Kludge #4.  After resuming a Break, the stream's cursorpos is wrong.
    ;; In fact, the cursor is at the end of the string in that case.
    ;; So, if it is supposed to be elsewhere, move it.
    ;; This condition also avoids wasting time when we are reading typein
    ;; at the end of the string.
    (OR (= FILL-POINTER TYPEIN-POINTER)
	(RH-CURSOR-MOTION TYPEIN-POINTER))
    ;; In case we had to return to the caller with a EDITING-COMMAND char
    ;; while RUBBED-OUT-SOME? was T, make things consistent again
    ;; by causing a rescan now.
    (WHEN (AND (NOT INITIAL-ENTRY)
	       (= (RH-SCAN-POINTER) MOST-POSITIVE-FIXNUM))
      (SETF (RH-SCAN-POINTER) 0)
      (THROW 'RUBOUT-HANDLER T)) 
    (CATCH 'RETURN-CHARACTER
      (WHEN RUBOUT-HANDLER-ACTIVATION-CHARACTER
	(THROW 'RETURN-CHARACTER
		(PROG1 RUBOUT-HANDLER-ACTIVATION-CHARACTER
		       (SETQ RUBOUT-HANDLER-ACTIVATION-CHARACTER NIL))))
      ;; Read characters.  If an ordinary character is typed and nothing has been rubbed out,
      ;; return immediately.  Otherwise, let all editing operations complete
      ;; before returning. 
      (DO (*LAST-COMMAND-TYPE*
	   *CURRENT-COMMAND-TYPE*
	   *RUBOUT-HANDLER-MARK*
	   (EDITING-COMMAND    (CDR (ASSOC :EDITING-COMMAND RUBOUT-HANDLER-OPTIONS :TEST #'EQ)))
	   (DO-NOT-ECHO        (CDR (ASSOC :DO-NOT-ECHO     RUBOUT-HANDLER-OPTIONS :TEST #'EQ)))
	   (PASS-THROUGH       (CDR (ASSOC :PASS-THROUGH    RUBOUT-HANDLER-OPTIONS :TEST #'EQ)))
	   (COMMAND-HANDLER    (ASSOC :COMMAND     RUBOUT-HANDLER-OPTIONS :TEST #'EQ))
	   (PREEMPTABLE        (ASSOC :PREEMPTABLE RUBOUT-HANDLER-OPTIONS :TEST #'EQ))
	   (ACTIVATION-HANDLER (ASSOC :ACTIVATION  RUBOUT-HANDLER-OPTIONS :TEST #'EQ)))
	  (NIL)
	;; Read a character from the stream after bypassing ourself.
	(SETQ CH
	      (LET ((RUBOUT-HANDLER NIL)
		    (TV:*DEFAULT-READ-WHOSTATE* (IF NUMERIC-ARG
						    (FORMAT NIL "Arg = ~d" numeric-arg)
						  "Keyboard")))
		(READ-ANY SELF)))
	(IF (LISTP CH)
	    (COND ((EQ (CAR CH) 'REDISPLAY-RUBOUT-HANDLER)
		   (SEND SELF :SET-CURSORPOS
			 PROMPT-STARTING-X PROMPT-STARTING-Y)
		   (SEND SELF :CLEAR-EOL)
		   (RH-REPRINT-INPUT NIL T))
		  (PREEMPTABLE
		   (MULTIPLE-VALUE-BIND (STRING INDEX)
		       (SEND SELF :SAVE-RUBOUT-HANDLER-BUFFER)
		     (SETQ OLD-TYPEAHEAD (LIST STRING INDEX)))
		   ;; Save the text, rub it all out, and unread the blip.
		   ;; The :FULL-RUBOUT option will cause the RH to return
		   ;; to the caller who will then read the blip.
		   (UNREAD-ANY CH SELF)         ;Should this be UNREAD-CHAR??  It was :UNTYI.
		   (RH-COM-CLEAR-INPUT NIL)
		   (SETF (RH-SCAN-POINTER) 0)
		   (THROW 'RETURN-FROM-RUBOUT-HANDLER
			   (VALUES CH (CADR PREEMPTABLE))))
		  ((AND (EQ (CAR CH) :MOUSE-BUTTON)
			(EQL (CADR CH) #\MOUSE-3-1))
		   (MOUSE-CALL-SYSTEM-MENU)))
	  (PROGN                                ;CH is a character, not a blip.
            (WHEN (AND (EQL CH #\)) *RH-MATCHING-PAREN-OPTIONS*
                       (< (RH-TYPEIN-POINTER) (CADDR *RH-MATCHING-PAREN-OPTIONS*)))
              (RH-BLINK-RIGHT-PAREN))             ;Cause matching parens to blink.
            ;;This code handles auto-completion on UCL command names and special typein modes.
            ;;Auto-completion used to be implemented as an Input Edtor command, but
            ;;this was inefficient on large typed expressions because each SPACE would cause 
            ;;re-reading of the input.
            (WHEN (AND (EQL CH #\SPACE)
                       ucl:command-loop-typein?  ;;Auto-complete only during ucl:READ-FOR-UCL.
                       user:auto-complete?  ;;Users can turn off auto-completion.
                       (>= (RH-TYPEIN-POINTER) (RH-FILL-POINTER))
                       (PLUSP (RH-TYPEIN-POINTER))
                       (NOT (CHAR= #\SPACE (AREF RUBOUT-HANDLER-BUFFER (1- (RH-TYPEIN-POINTER))))))
              (FUNCALL (complete-word-function rh-completion-handler)
                       :RECOGNITION (LIST 'complete-lisp-atom :auto-complete-p) T)
              (SETQ RUBBED-OUT-SOME? T))  ;;Force rescan of completed input.
            
            (SETQ CH-CHAR (CHAR-CODE CH))       ;WARNING - CH-CHAR is a Fixnum !!
            (SETQ CH-CONTROL-META (CHAR-BITS CH))
            (IF (AND (ZEROP ch-control-meta) (MEMBER CH PASS-THROUGH))
                (SETQ command NIL)
                ;;Look up the key in the RH command table, fetching more keys as necessary (in
                ;;case the user has bound RH commands to keystroke sequences greater than 1).
                ;;If the look-up fails, we have only read one key, and we are doing type-in for a
                ;;UCL aplication, try looking up the key in the UCL's tables.  If a command is found
                ;;or there are several commands found starting with the key, let the UCL
                ;;application preempt us to process the command.
                (PROGN
                  (MULTIPLE-VALUE-SETQ (command keys-read)
                    (ucl-rh-lookup-character ch))
                  (WHEN (AND (NULL command) ucl:command-loop-typein? preemptable (NULL (CDR keys-read)))
                    (WHEN (AND (SETQ command (SEND ucl:command-loop-typein? :lookup-keys keys-read))
                               (NEQ command :FETCH-MORE-KEYS)
                               (SEND command :GET :CANT-PREEMPT))
                      ;;If this command can't preempt typed expressions, don't try to.
                      (SETQ command NIL)))))
            (COND
              ((AND COMMAND-HANDLER
                    (APPLY (CADR COMMAND-HANDLER) CH (CDDR COMMAND-HANDLER)))
               (SETF (RH-SCAN-POINTER) 0)
               (THROW 'RETURN-FROM-RUBOUT-HANDLER
                      (VALUES
                        `(:COMMAND ,CH ,(* (OR NUMERIC-ARG 1)
                                           (IF NUMERIC-ARG-NEGATIVE -1 1)))
                        :COMMAND)))
              ((OR (MEMBER CH DO-NOT-ECHO)
                   (AND ACTIVATION-HANDLER
                        (APPLY (CADR ACTIVATION-HANDLER) CH (CDDR ACTIVATION-HANDLER))))
               (RH-SET-POSITION (RH-FILL-POINTER))
               (LET ((VALUE
                       (IF (MEMBER CH DO-NOT-ECHO) CH
                           `(:ACTIVATION ,CH ,(* (OR NUMERIC-ARG 1)
                                                 (IF NUMERIC-ARG-NEGATIVE -1 1))))))
                 (COND (RUBBED-OUT-SOME?
                        ;; Why isn't this done in the :RUBOUT-HANDLER method loop?
                        (SETQ RUBOUT-HANDLER-ACTIVATION-CHARACTER VALUE)
                        (SETF (RH-SCAN-POINTER) 0)
                        (THROW 'RUBOUT-HANDLER T))
                       (T (THROW 'RETURN-CHARACTER VALUE)))))
              ;; Don't touch this character, just return it to caller.
              ((OR (MEMBER CH EDITING-COMMAND)
                   (SI:ASSQ-CAREFUL CH EDITING-COMMAND))
               ;; Cause rubout handler rescan next time the user does :TYI.
               (If RUBBED-OUT-SOME? (SETF (RH-SCAN-POINTER) MOST-POSITIVE-FIXNUM))
               (RETURN CH))
              ;;Handle Control-number and Control-U specially.  This code was moved forward in
              ;;order to catch numeric args before the UCL application's num.  arg command claimes
              ;;the keystroke.  Note that CH-CHAR is a fixnum, CH is a Character.
              ((AND (NOT (ZEROP CH-CONTROL-META)) (<= (Char-Code #\0) CH-CHAR (Char-Code #\9)))
               (SETQ NUMERIC-ARG (+ (* (OR NUMERIC-ARG 0) 10.) (- CH-CHAR (Char-Code #\0)))))
              ((CHAR= CH #\CONTROL-U)
               (SETQ NUMERIC-ARG (* (OR NUMERIC-ARG 1) 4)))
              ((AND (NOT (ZEROP CH-CONTROL-META)) (= CH-CHAR (Char-Code #\-)))
               (IF NUMERIC-ARG
                   (SEND SELF :BEEP)
                   (SETQ NUMERIC-ARG-NEGATIVE (NOT NUMERIC-ARG-NEGATIVE))))
              (command
               (IF (NOT (TYPEP command 'rh-command))
                   ;;Keystroke was found in a UCL application's command-table.  Have the read
                   ;;function return the values '(:KEYSTROKE <ch>) :MOUSE-CHAR to indicate that
                   ;;we are interrupting typein to execute another command.  <ch> will be
                   ;;processed as if it were typed at top level.
                   (IO-BUFFER-PUSH IO-BUFFER (LIST :KEYSTROKE (CHAR-INT ch)))
                   ;; A standard rh editing command of some sort. The RUBBED-OUT-SOME bit can only
                   ;; be cleared by entering this function again.  The function is passed the
                   ;; numeric argument, and returns T if we are going to need to throw out (like
                   ;; DIS-ALL in the editor).
                   (PROGN
                     (SETQ *LAST-COMMAND-TYPE* *CURRENT-COMMAND-TYPE*
                           *CURRENT-COMMAND-TYPE* NIL)
                     (SETQ RUBBED-OUT-SOME?
                           (OR (SEND command :execute self
                                     (* (OR numeric-arg 1) (IF numeric-arg-negative -1 1))
                                     NIL)
                               RUBBED-OUT-SOME?))
                     (SETQ NUMERIC-ARG NIL NUMERIC-ARG-NEGATIVE NIL)
                     ;; If the buffer is empty and the :FULL-RUBOUT option is active, then throw now
                     ;; This will throw if the user types Rubout or ClearScreen immediately after
                     ;; entering the read function.  It is important that we check for this here
                     ;; and not in RH-DELETE-STRING since some commands, such as Yank-Pop, may
                     ;; temporarily empty the buffer.  It wouldn't be the right thing to throw
                     ;; if the buffer only contained whitespace since it is the responsibility
                     ;; of the caller to discard whitespace when looking for special characters.
                     (COND ((AND (ZEROP (RH-FILL-POINTER))
                                 (ASSOC :FULL-RUBOUT RUBOUT-HANDLER-OPTIONS :TEST #'EQ))
                            ;; This SETF should be done in the :RUBOUT-HANDLER method loop.
                            (SETF (RH-SCAN-POINTER) 0)
                            (THROW 'RUBOUT-HANDLER T))))))
              ;;Beep and ignore.
              ((NOT (ZEROP ch-control-meta))
               (SEND SELF :beep)
               (SETQ numeric-arg nil numeric-arg-negative nil))
              
              ;; Self-inserting character.  Set RUBBED-OUT-SOME since if we return,
              ;; we were typing in the middle of the line.  Typing at the end of the
              ;; line throws to RETURN-CHARACTER.
              (T (UNLESS NUMERIC-ARG-NEGATIVE
                   (RH-INSERT-CHAR CH (OR NUMERIC-ARG 1) RUBBED-OUT-SOME?)
                   (SETQ RUBBED-OUT-SOME? T))  
                 (SETQ *LAST-COMMAND-TYPE* *CURRENT-COMMAND-TYPE*
                       *CURRENT-COMMAND-TYPE* NIL
                       *RUBOUT-HANDLER-MARK* NIL)
                 (SETQ NUMERIC-ARG NIL NUMERIC-ARG-NEGATIVE NIL)))
	    ;; on some occasions we just want to know that something changed
	    ;;  PMH 3/25
	    (when (and RUBBED-OUT-SOME? rh-rescan-any-change)
	      (SETF (RH-SCAN-POINTER) 0)
	      (THROW 'RUBOUT-HANDLER T))
	    ))))))


(DEFUN-RH RH-BLINK-RIGHT-PAREN ()
  (LET ((LEFT-PAREN-POS (RH-SEARCH-FOR-MATCHING-LEFT-PAREN (RH-TYPEIN-POINTER))))
    (WHEN (AND LEFT-PAREN-POS
	       (CHAR= #\( (AREF RUBOUT-HANDLER-BUFFER LEFT-PAREN-POS))
	       (> LEFT-PAREN-POS 0))
      (RH-CURSOR-MOTION LEFT-PAREN-POS)
      (let ((tv:more-processing-global-enable nil))
	1;;If tv:more-processing-global-enable isn't turned off here, :clear-char can cause a more break*
	1;;while a user is in the middle of typing a form.*
	(declare (special tv:more-processing-global-enable))
	(DOTIMES (I (CAR *RH-MATCHING-PAREN-OPTIONS*))
	  (SEND SELF :CLEAR-CHAR (Char-Int #\())	;1Char-Int just to be safe.*
	  (DOTIMES (X (CADR *RH-MATCHING-PAREN-OPTIONS*)) 'IGNORE)  
	  (SEND SELF :TYO #\()                  ;1:TYO Seems to take Char's OK.*
	  (SEND SELF :BACKWARD-CHAR (Char-Int #\()))) 
      (RH-CURSOR-MOTION (RH-TYPEIN-POINTER))
      (WHEN (SEND SELF :GET-HANDLER-FOR :refresh-help)
	1;;When we close off an expression, see if the expression is an argument*
	1;;imbedded in another expression.  If so, have the current window show arglist info*
	1;;on the function in the who-line (if it is a who-line help window).  This way,*
	1;;as the user completes typing each argument expression, he gets back the function's*
	1;;arglist to help with the next arg.*
	(LET ((start-pos (rh-search-for-function (1- left-paren-pos))))
	  (WHEN start-pos
	    (SETF (completion-word rh-completion-handler)
		  (rh-substring-of-buffer start-pos (rh-word-end start-pos)))
	    (SEND SELF :refresh-help 'ARGLIST)))))))

(DEFUN-RH RH-SEARCH-FOR-MATCHING-LEFT-PAREN (POS)
  "2Search for a left paren matching a hypothetical right paren at POS.*  2TV:RH-BLINK-RIGHT-PAREN
uses this function BEFORE inserting right parens* 2in order to blink the matching left paren.*"
  (LOOP FOR new-pos = (CATCH 'FAILED (RH-SEARCH-BACKWARD-PARENS-HELP pos))
	WHILE new-pos
	IF (= new-pos pos)
	RETURN (MULTIPLE-VALUE-BIND (found which-char)
		   (rh-search-backward-for-chars '(#\( #\) #\") 0 POS)
		 (WHEN (AND FOUND (CHAR= which-char #\())
		   found))
	ELSE DO
	(SETQ pos new-pos)))

(DEFUN-RH RH-SEARCH-FOR-FUNCTION (POS)
  "2Return the position in RUBOUT-HANDLER-BUFFER which is the start of the function name* 2at POS's
nesting. * 2For example, if *\2"(CONS (LIST a) foo)*\2" represents the contents of the* 2buffer and POS
points to the F in FOO, the location of C in CONS is returned.*  2Note that NIL is returned when POS is
the left paren before LIST.*  2Returns NIL if search fails (such as on empty list).*"
  (LOOP UNTIL (ZEROP POS)  1;;No outer function nesting.  Example:  (null nil)*
	IF (CHAR= (AREF RUBOUT-HANDLER-BUFFER (1- POS)) #\()
	RETURN POS 1;;Found the position*
	1;;Keep searching*
	FOR search = (RH-SEARCH-BACKWARD-PARENS pos)
	IF (= search pos) RETURN NIL
	ELSE DO (SETQ pos search)))


(DEFUN ucl-rh-lookup-character (ch)
  "2This loop was derived from the (UCL:BASIC-COMMAND-LOOP :LOOKUP-KEY)* 2method. * 2Given a
character, it returns the corresponding rubout handler command.*  2However, the hair involved assumes that
the user might modify an rh command, using* 2the UCL Command Editor, to be invoked by more than one
key. * 2This is actually reasonable* 2since the number of RH commands has grown considerably.*"
  (DECLARE (VALUES rubout-handler-command-instance-or-nil))
  (LOOP WITH keys-so-far = (LIST ch)
	FOR command-or-flag = (SEND rh-command-table :lookup-keys keys-so-far)
	IF (EQ command-or-flag :FETCH-MORE-KEYS)
	;;Fetch a keystroke or mouse button from the user, add it to the key
	;;sequence we are looking up. 
	DO (PUSH-END (LET ((rubout-handler nil))
		        (ucl:read-next-key keys-so-far))
		      keys-so-far)
	ELSE
	;;The result was a TV:RH-COMMAND instance.
	RETURN (VALUES command-or-flag keys-so-far)))
