;;; -*- Mode: Common-Lisp; Package: IP; Base: 10.; Patch-File: T -*-

;;; Reason: (tftp-connection :tftp-complete-transfer) was not handling :XLD mode properly.
;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;; Written 04/27/88 09:07:26 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 4.59, VIRTUAL-MEMORY 4.4, EH 4.5, MAKE-SYSTEM 4.5, MICRONET 4.4, LOCAL-FILE 4.1,
;;;  BASIC-PATHNAME 4.12, NETWORK-SUPPORT-COLD 4.1, NAMESPACE 4.19, NETWORK-NAMESPACE 4.2,
;;;  DISK-IO 4.10, DISK-LABEL 4.0, BASIC-FILE 4.6, MAC-PATHNAME 4.3, NETWORK-PATHNAME 4.0,
;;;  COMPILER 4.10, TV 4.79, DATALINK 4.14, CHAOSNET 4.18, GC 4.3, MEMORY-AUX 4.0,
;;;  NVRAM 4.6, SYSLOG 4.3, STREAMER-TAPE 4.4, UCL 4.1, INPUT-EDITOR 4.0, METER 4.3,
;;;  ZWEI 4.18, DEBUG-TOOLS 4.2, NETWORK-SUPPORT 4.5, NETWORK-SERVICE 4.0, DATALINK-DISPLAYS 4.0,
;;;  FONT-EDITOR 4.0, SERIAL 4.0, PRINTER 4.7, PRINTER-TYPES 4.2, IMAGEN 4.0, SUGGESTIONS 4.0,
;;;  MAIL-DAEMON 4.7, MAIL-READER 4.6, TELNET 4.1, VT100 4.4, NAMESPACE-EDITOR 4.5,
;;;  PROFILE 4.4, VISIDOC 4.4, Inconsistent IP 3.16, Experimental BUG 10.0,  microcode 306,
;;;  Band Name: pre 4.1 + IP 4/25

#!C
; From file TFTP-F-N-M.LISP#> IP; SYS:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; TFTP-F-N-M.#"


(DEFMETHOD (tftp-connection :tftp-complete-transfer) ()
  "A stream is created which can process the appropriate file io--determined by the
    direction of the transfer. The direction should be eithor :READ or :WRITE. 
    If the stream is created with no errors, then the transmission is processed.
    The success or failure of the transmission is returned to the caller."
  (WHEN *tftp-debug-mode*
    (tv:notify () "COMPLETE TRANSFER: MODE = ~A " mode))
  ;; SETUP the fat array if doing an XFASL or XLD transfer
  (WHEN  (or (EQUAL mode :xfasl)
	     (EQUAL mode :xld))  ;04-27-88 DAB
    (SETF xfasl-packet
	  (allocate-xfasl-packet (IF (EQUAL operation :read)
				     current-packet
				     network-packet))))
  ;; Open the local file
  (WITH-OPEN-FILE-CASE
    (local-file-stream local-filename
		       :direction (IF (EQUAL operation :read)
				      :input	   ; read data from a file
				      :output	   ; write data to a file
				      )
		       :characters (EQUAL mode :netascii)
		       :byte-size (IF (or (EQUAL mode :xfasl) (EQUAL mode :xld)) ;04-27-88 DAB
				      16
				      8))  ;End of options
    ;; <TFTP ERROR CODE 0.>
    (fs:device-not-found
     (SEND self :send-error-packet tftp-error-text "The device was not found.")
     local-file-stream)
    (fs:file-open-for-output
     (SEND self :send-error-packet tftp-error-text "The file was open for output.")
     local-file-stream)
    ;; < TFTP ERROR CODE 1. >
    (fs:file-not-found
     (SEND self :send-error-packet tftp-file-not-found "")
     local-file-stream)
    ;; < TFTP ERROR CODE 6. >
    (fs:file-already-exists
     (SEND self :send-error-packet tftp-file-exists "")
     local-file-stream)
    ;; < TFTP ERROR CODE 2. >
    (fs:directory-not-found
     (SEND self :send-error-packet tftp-access-violation "The directory was not found.")
     local-file-stream)
    (fs:access-error
     (SEND self :send-error-packet tftp-access-violation "")
     local-file-stream)
    (fs:not-available
     (SEND self :send-error-packet tftp-access-violation "File is not available.")
     local-file-stream)
    (fs:file-lookup-error
     (SEND self :send-error-packet tftp-access-violation "File lookup error.")
     local-file-stream)
    (fs:creation-failure
     (SEND self :send-error-packet tftp-access-violation "Could not create file. ")
     local-file-stream)
    ;; < TFTP ERROR CODE 3. >
    (fs:no-more-room
     (SEND self :send-error-packet tftp-allocation-exceeded "")
     local-file-stream)
    (fs:file-error
     (SEND self :send-error-packet tftp-error-text "Error occurred while opening the file")
     local-file-stream)
    ;; < SUCCESSFUL OPEN >
    ;; Complete the transmission.
    (:no-error (WHEN *tftp-debug-mode*
		 (tv:notify () "SUCCESSFUL OPEN"))
	       (SETF stream local-file-stream)
	       (IF (EQUAL (SEND stream :direction) :input)
		   (SEND self :read)	   ; local file is the source file.
		   (PROGN
		     (WHEN (EQUAL (SEND current-process :name) tftp-process-name)
		       (SEND stream :change-properties () :author tftp-process-name))
		     (SEND self :write))))))

(ip:reset-tftp-service)
))
