;;; -*- Mode: Common-Lisp; Package: FS; Base: 10.; Patch-File: T -*-

;;; Reason:  Lispm-directory-line-parse was truncating the first digit of the hour from the creation time.
;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;; Written 04/28/88 08:38:17 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 4.59, VIRTUAL-MEMORY 4.4, EH 4.5, MAKE-SYSTEM 4.5, MICRONET 4.5, LOCAL-FILE 4.1,
;;;  BASIC-PATHNAME 4.12, NETWORK-SUPPORT-COLD 4.1, Inconsistent NAMESPACE 4.20, NETWORK-NAMESPACE 4.2,
;;;  DISK-IO 4.10, DISK-LABEL 4.0, BASIC-FILE 4.6, MAC-PATHNAME 4.3, NETWORK-PATHNAME 4.0,
;;;  COMPILER 4.10, Inconsistent TV 4.82, DATALINK 4.14, CHAOSNET 4.18, GC 4.3, MEMORY-AUX 4.0,
;;;  NVRAM 4.6, SYSLOG 4.3, STREAMER-TAPE 4.4, UCL 4.1, INPUT-EDITOR 4.0, METER 4.3,
;;;  ZWEI 4.18, DEBUG-TOOLS 4.2, NETWORK-SUPPORT 4.5, NETWORK-SERVICE 4.0, DATALINK-DISPLAYS 4.0,
;;;  FONT-EDITOR 4.0, SERIAL 4.0, PRINTER 4.7, PRINTER-TYPES 4.2, IMAGEN 4.0, SUGGESTIONS 4.0,
;;;  MAIL-DAEMON 4.7, MAIL-READER 4.6, TELNET 4.1, VT100 4.4, NAMESPACE-EDITOR 4.5,
;;;  PROFILE 4.4, VISIDOC 4.4, Inconsistent IP 3.17, Experimental BUG 10.0,  microcode 306,
;;;  Band Name: pre 4.1 + IP 4/25

#!C
; From file FTPINT.LISP#> IP; SYS:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; FTPINT.#"


(DEFUN lispm-directory-line-parse (dirobj line fast? &aux pathobj cursor cursor2 result month day year hours minutes seconds
				   length-in-blocks length-in-bytes byte-size)
  ;;; FORMAT:
  ;;; {d}  <pathname>  #-blocks  #-bytes(byte-size)  {!}{@}{$}{@} mm/dd//yy hh:mm:ss {author}

  ;;;                           ""    DIRECTORY         ""
  (WHEN (CHAR-EQUAL (AREF line 0) #\d)
    (PUSH t result)
    (PUSH :deleted result))
  ;; Careful - there may be multiple blanks embedded in the pathname and pathname parsing doesn't handle :Junk-allowed very well
  (SETQ cursor (POSITION #\Space (THE string (STRING line)) :start 1 :test-not 'CHAR-EQUAL))
  (SETQ cursor2
	(POSITION #\Space (THE string (STRING line)) :start
		  (POSITION #\# (THE string (STRING line)) :start cursor :test 'CHAR-EQUAL)
		  :test 'CHAR-EQUAL))
;; Make the pathname to merge with instead of parsing it. *BJ*
  (SETQ pathobj (MERGE-PATHNAMES
		  (parse-pathname (SUBSEQ line cursor cursor2) nil dirobj)
		  (fs:make-PATHNAME :host (send dirobj :host)
				    :type :unspecific)))
  (UNLESS fast?
    (MULTIPLE-VALUE-SETQ (length-in-blocks cursor)
			 (PARSE-INTEGER line :start cursor2 :junk-allowed t))
    (PUSH length-in-blocks result)
    (PUSH :length-in-blocks result)
    (IF (MULTIPLE-VALUE-SETQ (length-in-bytes cursor)
			     (PARSE-INTEGER line :start cursor :junk-allowed t))
	(PROGN
	  (MULTIPLE-VALUE-SETQ (BYTE-SIZE cursor)
			       (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
	  (PUSH byte-size result)
	  (PUSH :byte-size result)
	  (PUSH length-in-bytes result)
	  (PUSH :length-in-bytes result))
	;; ELSE
	(PROGN
	  (PUSH t result)
	  (PUSH :directory result)
	  (INCF cursor (LENGTH "DIRECTOR"))))	;leave at "Y" (will be incremented)
    (push (DETERMINE-CHARACTERS-AND-BYTE-SIZE result pathobj pathobj nil nil nil) result)   ;04-12-88 DAB
    (push :characters result)                                                               ;04-12-88 DAB
    (LOOP
      (SETQ cursor
	    (POSITION #\Space (THE string (STRING line)) :start (1+ cursor) :test-not
		      'CHAR-EQUAL))
      (SELECTOR (AREF line cursor) char-equal (#\! (PUSH t result) (PUSH :not-backed-up result))
		(#\# (PUSH t result) (PUSH :dont-supersede result))
		(#\$ (PUSH t result) (PUSH :dont-reap result))
		(#\@ (PUSH t result) (PUSH :dont-delete result)) (:otherwise (RETURN))))
    (MULTIPLE-VALUE-SETQ (month cursor)
			 (PARSE-INTEGER line :start cursor :junk-allowed t))
    (MULTIPLE-VALUE-SETQ (day cursor)
			 (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
    (MULTIPLE-VALUE-SETQ (year cursor)
			 (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
    (MULTIPLE-VALUE-SETQ (hours cursor)
			 (PARSE-INTEGER line :start cursor :junk-allowed t)) ;04-28-88 DAB parse-integer ate the space
    (MULTIPLE-VALUE-SETQ (minutes cursor)
			 (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
    (MULTIPLE-VALUE-SETQ (seconds cursor)
			 (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
    (PUSH (ENCODE-UNIVERSAL-TIME seconds minutes hours day month year) result)
    (PUSH :creation-date result)
    (IF (SETQ cursor
	      (POSITION #\Space (THE string (STRING line)) :start cursor :test-not 'CHAR-EQUAL))
	(PROGN
	  (SETQ cursor2
		(POSITION #\Space (THE string (STRING line)) :start cursor :test 'CHAR-EQUAL))
	  (PUSH (SUBSEQ line cursor cursor2) result))
	;; else
	(PUSH (CAR (SEND dirobj :directory)) result))
    (PUSH :author result))
  (PUSH pathobj result))
))
