;;; -*- Mode: Common-Lisp; Package: IP; Base: 10.; Patch-File: T -*-

;;; Reason: Added some error catching in ftp-server-stor and ftp-server-retr to catch bad host names and pathname syntaxs.
;;; 	
;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;; Written 05/04/88 10:17:15 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 4.60, VIRTUAL-MEMORY 4.4, EH 4.5, MAKE-SYSTEM 4.5, MICRONET 4.5, LOCAL-FILE 4.1,
;;;  BASIC-PATHNAME 4.12, NETWORK-SUPPORT-COLD 4.1, NAMESPACE 4.22, NETWORK-NAMESPACE 4.2,
;;;  DISK-IO 4.10, DISK-LABEL 4.0, BASIC-FILE 4.7, MAC-PATHNAME 4.5, NETWORK-PATHNAME 4.1,
;;;  COMPILER 4.11, TV 4.83, DATALINK 4.14, CHAOSNET 4.18, GC 4.3, MEMORY-AUX 4.0,
;;;  NVRAM 4.6, SYSLOG 4.0, STREAMER-TAPE 4.4, UCL 4.1, INPUT-EDITOR 4.0, METER 4.3,
;;;  ZWEI 4.18, DEBUG-TOOLS 4.2, NETWORK-SUPPORT 4.5, NETWORK-SERVICE 4.0, DATALINK-DISPLAYS 4.0,
;;;  FONT-EDITOR 4.0, SERIAL 4.0, PRINTER 4.8, PRINTER-TYPES 4.2, IMAGEN 4.0, SUGGESTIONS 4.0,
;;;  MAIL-DAEMON 4.7, MAIL-READER 4.6, TELNET 4.1, VT100 4.4, NAMESPACE-EDITOR 4.5,
;;;  PROFILE 4.4, VISIDOC 4.4, IP 3.18, Experimental BUG 10.0,  microcode 306, Band Name: Rel 4.1 + IP 5/2

#!C
; From file FTP-SERVER.LISP#> IP; SYS:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; FTP-SERVER.#"

(DEFUN ftp-server-retr (&aux mode path done)
  "FTP Server's handler for 'RETR<SP><pathname><CRLF>'"
  (DECLARE (ATOM mode))
  (DECLARE (ATOM done))
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL data-connection))
  (DECLARE (SPECIAL type-code))
  (DECLARE (SPECIAL byte-size))
  (IF (EQUAL type-code :ascii)
      (SETQ mode t)
      (SETQ mode ()))
  (condition-case (condition)		   ;05-04-88 DAB 
      (progn
       (SETQ path (FTP-PARSE-FOR-SERVER arg default-directory))
       (WITH-OPEN-FILE (fromfile path :if-does-not-exist :error :characters mode
				 :byte-size byte-size :error nil)
	 (IF (OR (NULL fromfile) (TYPEP fromfile 'condition))
	     (FORMAT control-connection "550 Cannot access ~a: ~a~%" arg fromfile)
	     (IF (ftp-server-data-conn path)
		 (LET ((file-stream fromfile))
		   (DECLARE (SPECIAL file-stream))
		   (UNWIND-PROTECT
		       (BLOCK retrieve
			 (SETQ done ())
			 (DO ((buf)
			      (offset)
			      (limit))
			     (nil)
			   (MULTIPLE-VALUE-SETQ (buf offset limit)
			     (SEND fromfile :read-input-buffer))
			   (IF (NULL buf)
			       (RETURN ()))
			   (WHEN (SECOND
				   (MULTIPLE-VALUE-LIST
				     (IGNORE-ERRORS (SEND data-connection :string-out buf offset limit))))
			     (RETURN-FROM retrieve))
			   (SEND fromfile :advance-input-buffer)
			   (IF (ftp-server-abort-check)
			       (RETURN-FROM retrieve)))
			 (SETQ done t))
		     (IF done
			 (IGNORE-ERRORS (CLOSE data-connection))
			 (IGNORE-ERRORS (CLOSE data-connection :abort t)))
		     (SETQ data-connection ())
		     (IF done
			 (FORMAT control-connection "226 ~a retrieved~%" arg)
			 (FORMAT control-connection "226 retrieving ~a aborted~%" arg))))))))
    (error (FORMAT control-connection "550 Cannot access ~a: ~a~%" arg
		   (send condition :report-string))))	   ;05-04-88 DAB
  )


(DEFUN ftp-server-store (creation-mode &aux mode path done)
  "FTP Server's internal routine used by STOR, APPE and STOU"
  (DECLARE (keyword creation-mode))
  (DECLARE (ATOM mode))
  (DECLARE (ATOM done))
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (DECLARE (SPECIAL data-connection))
  (DECLARE (SPECIAL type-code))
  (DECLARE (SPECIAL byte-size))
  (IF (EQUAL type-code :ascii)
      (SETQ mode t)
      (SETQ mode ()))
  (condition-case (condition)   ;05-04-88 DAB 
      (progn
       (SETQ path (FTP-PARSE-FOR-SERVER arg default-directory))
       (IF (AND (EQUAL creation-mode :truncate) (EQUAL (SEND path :version) :newest))
	   (SETQ creation-mode :new-version))
       (WITH-OPEN-FILE (tofile path :direction :output :if-exists creation-mode :if-does-not-exist :create
			       :characters mode :byte-size byte-size :error nil)
	 (IF (OR (NULL tofile) (TYPEP tofile 'condition))
	     (FORMAT control-connection "553 Cannot access ~a: ~a~%" arg tofile)
	     (IF (ftp-server-data-conn path)
		 (LET ((file-stream tofile))
		   (DECLARE (SPECIAL file-stream))
		   (UNWIND-PROTECT
		       (BLOCK store
			 (SETQ done ())
			 (IF user-name
			     (SEND tofile :change-properties () :author user-name))
			 (DO ((buf)
			      (offset)
			      (limit))
			     (nil)
			   (MULTIPLE-VALUE-SETQ (buf offset limit)
			     (SEND data-connection :read-input-buffer))
			   (IF (NULL buf)
			       (RETURN ()))			   (SEND tofile :string-out buf offset limit)
			   (SEND data-connection :advance-input-buffer)
			   (IF (ftp-server-abort-check)
			       (RETURN-FROM store)))
			 (SETQ done t))
		     (IF done
			 (IGNORE-ERRORS (CLOSE data-connection))
			 (IGNORE-ERRORS (CLOSE data-connection :abort t)))
		     (SETQ data-connection ())
		     (IF done
			 (FORMAT control-connection "226 ~a ~a~%" arg
				 (IF (EQUAL creation-mode :append)
				     "appended"
				     "stored"))
			 (FORMAT control-connection "226 ~a ~a aborted~%"
				 (IF (EQUAL creation-mode :append)
				     "appending"
				     "storing")
				 arg)))))))
       )
    (error (FORMAT control-connection "553 Cannot access ~a: ~a~%" arg
		   (send condition :report-string))))  ;05-04-88 DAB
  )


(DEFUN ftp-server-list (&aux path dir done)
  "FTP Server's handler for 'LIST[<SP><pathname>]<CRLF>' and 'NLST[<SP><pathname>]<CRLF>'"
  (DECLARE (ATOM done))
  (DECLARE (SPECIAL cmd))
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL data-connection))
  (SETQ path arg)
  (IF (EQUAL path "")
      (SETQ path default-directory))
  (IF *ftp-server-debug*
      (tv:notify () "FTP-Server-list: path ~a" path))
  (condition-case (condition)		   ;05-04-88 DAB 
      (progn
       (SETQ dir (fs:directory-list (FTP-PARSE-FOR-SERVER
				      path
				      (SEND default-directory :new-pathname :version :wild))
				    :noerror))
       (IF (TYPEP dir 'condition)
	   (FORMAT control-connection "450 Cannot list ~a: ~a~%" path dir)
	   (PROGN
	     (IF (ftp-server-data-conn path)
		 (UNWIND-PROTECT
		     (BLOCK list
		       (SETQ done ())
		       (IF (STRING-EQUAL cmd "list")
			   (DOLIST (file (CDR dir))
			     (WHEN (SECOND
				     (MULTIPLE-VALUE-LIST
				       (IGNORE-ERRORS
					 (zwei::default-list-one-file file data-connection
					   :string-for-host))))
			       (RETURN-FROM list))
			     (IF (ftp-server-abort-check)
				 (RETURN-FROM list)))
			   (DOLIST (file (CDR dir))
			     (UNLESS (IGNORE-ERRORS
				       (WRITE-LINE (SEND (CAR file) :string-for-host) data-connection))
			       (RETURN-FROM list))
			     (IF (ftp-server-abort-check)
				 (RETURN-FROM list))))
		       (SETQ done t))
		   (IF done
		       (IGNORE-ERRORS (CLOSE data-connection))
		       (IGNORE-ERRORS (CLOSE data-connection :abort t)))
		   (SETQ data-connection ())
		   (IF done
		       (FORMAT control-connection "226 ~a listed~%" path)
		       (FORMAT control-connection "226 listing ~a aborted~%" path))))))
       )				   ;progn
    (error (FORMAT control-connection "550 Cannot access ~a: ~a~%" arg
		   (send condition :report-string))))	   ;05-04-88 DAB
  )

(DEFUN ftp-server-cwd (&aux path result)
  "FTP Server's handler for 'CWD<SP><pathname><CRLF>'"
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (IF (NULL user-name)
      (WRITE-LINE "530 Please login first" control-connection)
      (condition-case (condition)	   ;05-04-88 DAB 
	  (progn
	   (IF (EQUAL arg "")
	       (SETQ path
		     (SEND default-directory :new-pathname :host 'lm :directory user-name :name :wild
			   :type :unspecific :version :newest))
	       (SETQ path (FTP-PARSE-FOR-SERVER arg default-directory)))
	   (IF (TYPEP (SETQ result (fs:directory-list path :noerror)) 'condition)
	       (FORMAT control-connection "550 Cannot change to ~a: ~a~%" path result)
	       (PROGN
		 (SETQ default-directory path)
		 (FORMAT control-connection "250 Directory changed to ~a~%" path))))
	(error (FORMAT control-connection "550 Cannot access ~a: ~a~%" path
		       (send condition :report-string))))  ;05-04-88 DAB
      
      
      ))
))
