;;; -*- Mode:Common-Lisp; Package:FILE-SYSTEM; Base:10; Patch-file:T -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;; Written 02/12/88 15:31:30 by jones,
;;; Reason: Most of the fixes needed for FTP on the microExplorer. Will not handle pathnames with blanks as a character.
;;; Set the FTP-IMPLEMENTATION-TYPE to :MicroExplorer in the namespace.
;;; while running on HEYERDAHL from band LOD1
;;; With Experimental SYSTEM 4.17, Experimental COMPILER 4.0, Experimental NAMESPACE 4.1,
;;;  Experimental TV 4.9, Experimental DATALINK 4.6, CHAOSNET 4.2, Experimental GC 4.0,
;;;  Experimental NVRAM 4.0, Experimental SYSLOG 4.0, Experimental IO 4.7, Experimental STREAMER-TAPE 4.0,
;;;  Experimental UCL 4.0, Experimental INPUT-EDITOR 4.0, Experimental PATHNAME-COLD 4.11,
;;;  Experimental FILE-COLD 4.0, Experimental METER 4.0, Experimental ZWEI 4.3, Experimental DEBUG-TOOLS 4.0,
;;;  Experimental NETWORK-SUPPORT 4.0, Experimental NETWORK-SERVICE 4.0, Experimental DATALINK-DISPLAYS 4.0,
;;;  Experimental FONT-EDITOR 4.0, Experimental SERIAL 4.0, Experimental PRINTER 4.1,
;;;  Experimental IMAGEN 4.0, Experimental SUGGESTIONS 4.0, Experimental MAIL-DAEMON 4.0,
;;;  Experimental MAIL-READER 4.2, Experimental TELNET 4.0, Experimental VT100 4.1,
;;;  Experimental NAMESPACE-EDITOR 4.1, Experimental PROFILE 4.0, Experimental GLOSSARY 4.0,
;;;  VISIDOC 1.0, Experimental IP 4.0, Experimental BUG 10.0, Experimental MICRONET 4.3,
;;;  Experimental MICRONET-COMM 4.0, Experimental RPC 1.0, Experimental NFS 4.16,
;;;   microcode 14, Band Name: pre 4.0+IP 1-27

#!C
; From file FTP-SERVER.LISP#> JONES; HEYERDAHL:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :Common-lisp)
                          (*READTABLE* Sys:Common-lisp-readtable)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* Sys::*common-lisp-symbol-substitutions*))
  (COMPILER#:PATCH-SOURCE-FILE "ip: ip; FTP-SERVER.#"


(DEFUN ftp-server-user (&aux path)
  "FTP Server's handler for 'USER<SP><username><CRLF>'"
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (IF (EQUAL arg "")
      (WRITE-LINE "501 Missing user name" control-connection)
      (PROGN
	(SETQ user-name arg)
	(SETQ path
	      (MAKE-PATHNAME :host "boot|lm"
			     :directory user-name
			     :name :wild
			     :type :unspecific
			     :version :newest))
	(IF (NOT (TYPEP (fs:directory-list path :noerror) 'condition))
	    (SETQ default-directory path))
	(FORMAT control-connection "230 User ~a logged in; default directory: ~a~%" user-name
		default-directory))))

(defun ftp-parse-for-server
       (string &optional (default-pathname (fs:sample-pathname "boot|lm")) &aux pathname)
  (condition-case (result)		
      (if (typep default-pathname 'fs:lispm-pathname)
	  (fs:merge-pathname-defaults string default-pathname :unspecific :newest)
	  (multiple-value-bind (device directory name type version ignore quoted-string)
	      (send default-pathname :parse-namestring nil string)
	    (setf pathname
		  (fs:make-pathname-internal
		    quoted-string (pathname-host default-pathname)
		    device directory name type version)))
	  (fs:merge-pathname-defaults pathname default-pathname :unspecific :newest))
    (pathname-error (send result :report-string))))

(DEFUN ftp-server-cwd (&aux path result)
  "FTP Server's handler for 'CWD<SP><pathname><CRLF>'"
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (IF (NULL user-name)
      (WRITE-LINE "530 Please login first" control-connection)
      (PROGN
	(IF (EQUAL arg "")
	    (SETQ path
		  (SEND default-directory :new-pathname :host 'lm :directory user-name :name :wild
			:type :unspecific :version :newest))
	    (SETQ path (FTP-PARSE-FOR-SERVER arg default-directory)))
	(IF (TYPEP (SETQ result (fs:directory-list path :noerror)) 'condition)
	    (FORMAT control-connection "550 Cannot change to ~a: ~a~%" path result)
	    (PROGN
	      (SETQ default-directory path)
	      (FORMAT control-connection "250 Directory changed to ~a~%" path))))))

(DEFUN ftp-server-dele (&aux path result)
  "FTP Server's handler for 'DELE<SP><pathname><CRLF>'"
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (IF (NULL user-name)
      (WRITE-LINE "530 Please login first" control-connection)
      (PROGN
	(IF (NULL (SETQ path (PROBE-FILE (FTP-PARSE-FOR-SERVER arg default-directory))))
	    (FORMAT control-connection "550 ~a does not exist.~%" arg)
	    (PROGN
	      (SETQ result (SEND path :delete-and-expunge ()))
	      (IF (TYPEP result 'condition)
		  (FORMAT control-connection "550 Cannot delete ~a: ~a~%" path result)
		  (FORMAT control-connection "250 ~a deleted~%" path)))))))

(DEFUN ftp-server-list (&aux path dir done)
  "FTP Server's handler for 'LIST[<SP><pathname>]<CRLF>' and 'NLST[<SP><pathname>]<CRLF>'"
  (DECLARE (ATOM done))
  (DECLARE (SPECIAL cmd))
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL data-connection))
  (SETQ path arg)
  (IF (EQUAL path "")
      (SETQ path default-directory))
  (IF *ftp-server-debug*
      (tv:notify () "FTP-Server-list: path ~a" path))
  (SETQ dir (fs:directory-list (FTP-PARSE-FOR-SERVER path (SEND default-directory :new-pathname :version :wild))
						:noerror))
  (IF (TYPEP dir 'condition)
      (FORMAT control-connection "450 Cannot list ~a: ~a~%" path dir)
      (PROGN
	(IF (ftp-server-data-conn path)
	    (UNWIND-PROTECT
		(BLOCK list
		  (SETQ done ())
		  (IF (STRING-EQUAL cmd "list")
		      (DOLIST (file (CDR dir))
			(WHEN (SECOND
				(MULTIPLE-VALUE-LIST
				  (IGNORE-ERRORS
				    (zwei::default-list-one-file file data-connection
				      :string-for-host))))
			  (RETURN-FROM list))
			(IF (ftp-server-abort-check)
			    (RETURN-FROM list)))
		      (DOLIST (file (CDR dir))
			(UNLESS (IGNORE-ERRORS
				  (WRITE-LINE (SEND (CAR file) :string-for-host) data-connection))
			  (RETURN-FROM list))
			(IF (ftp-server-abort-check)
			    (RETURN-FROM list))))
		  (SETQ done t))
	      (IF done
		  (IGNORE-ERRORS (CLOSE data-connection))
		  (IGNORE-ERRORS (CLOSE data-connection :abort t)))
	      (SETQ data-connection ())
	      (IF done
		  (FORMAT control-connection "226 ~a listed~%" path)
		  (FORMAT control-connection "226 listing ~a aborted~%" path)))))))


(DEFUN ftp-server-rnfr (&aux path)
  "FTP Server's handler for 'RNFR<SP><pathname><CRLF>'"
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (IF (NULL user-name)
      (WRITE-LINE "530 Please login first" control-connection)
      (PROGN
	(IF (PROBE-FILE (SETQ path (FTP-PARSE-FOR-SERVER arg default-directory)))
	    (FORMAT control-connection "350 Rename ~a to:~%" path)
	    (FORMAT control-connection "550 ~a does not exist.~%" path)))))

(DEFUN ftp-server-rnto (&aux old-path new-path)
  "FTP Server's handler for 'RNTO<SP><pathname><CRLF>'"
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL previous-cmd))
  (DECLARE (SPECIAL previous-arg))
  (WHEN (NOT (STRING-EQUAL previous-cmd "rnfr"))
    (WRITE-LINE "503 Missed RNFR" control-connection)
    (RETURN-FROM ftp-server-rnto))
  (RENAME-FILE (SETQ old-path (FTP-PARSE-FOR-SERVER previous-arg default-directory))
	       (SETQ new-path (FTP-PARSE-FOR-SERVER arg default-directory)))
  (IF (PROBE-FILE new-path)
      (FORMAT control-connection "250 File ~a renamed to ~a~%" old-path new-path)
      (FORMAT control-connection "553 File ~a not renamed~%" old-path)))

(DEFUN ftp-server-retr (&aux mode path done)
  "FTP Server's handler for 'RETR<SP><pathname><CRLF>'"
  (DECLARE (ATOM mode))
  (DECLARE (ATOM done))
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL data-connection))
  (DECLARE (SPECIAL type-code))
  (DECLARE (SPECIAL byte-size))
  (IF (EQUAL type-code :ascii)
      (SETQ mode t)
      (SETQ mode ()))
  (SETQ path (FTP-PARSE-FOR-SERVER arg default-directory))
  (WITH-OPEN-FILE (fromfile path :if-does-not-exist :error :characters mode
			    :byte-size byte-size :error nil)
    (IF (OR (NULL fromfile) (TYPEP fromfile 'condition))
	(FORMAT control-connection "550 Cannot access ~a: ~a~%" arg fromfile)
	(IF (ftp-server-data-conn path)
	    (LET ((file-stream fromfile))
	      (DECLARE (SPECIAL file-stream))
	      (UNWIND-PROTECT
		  (BLOCK retrieve
		    (SETQ done ())
		    (DO ((buf)
			 (offset)
			 (limit))
			(nil)
		      (MULTIPLE-VALUE-SETQ (buf offset limit)
			(SEND fromfile :read-input-buffer))
		      (IF (NULL buf)
			  (RETURN ()))
		      (WHEN (SECOND
			      (MULTIPLE-VALUE-LIST
				(IGNORE-ERRORS (SEND data-connection :string-out buf offset limit))))
			(RETURN-FROM retrieve))
		      (SEND fromfile :advance-input-buffer)
		      (IF (ftp-server-abort-check)
			  (RETURN-FROM retrieve)))
		    (SETQ done t))
		(IF done
		    (IGNORE-ERRORS (CLOSE data-connection))
		    (IGNORE-ERRORS (CLOSE data-connection :abort t)))
		(SETQ data-connection ())
		(IF done
		    (FORMAT control-connection "226 ~a retrieved~%" arg)
		    (FORMAT control-connection "226 retrieving ~a aborted~%" arg))))))))

(DEFUN ftp-server-stat (&aux path dir)
  "FTP Server's handler for 'STAT[<SP><pathname>]<CRLF>'"
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL data-connection))
  (DECLARE (SPECIAL data-port-number))
  (DECLARE (SPECIAL data-port-address))
  (DECLARE (SPECIAL type-code))
  (DECLARE (SPECIAL byte-size))
  (DECLARE (SPECIAL format))
  (DECLARE (SPECIAL mode))
  (DECLARE (SPECIAL structure))
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL previous-cmd))
  (DECLARE (SPECIAL user-name))
  (IF (EQUAL arg "")
      (PROGN
	(FORMAT control-connection "211-User: ~a, Directory: ~a~%" (IF user-name
								       user-name
								       "none")
		default-directory)
	(FORMAT control-connection " Data-port ~d, Data-address ~d~%" data-port-number
		data-port-address)
	(FORMAT control-connection " Type: ~a, Byte-size: ~d, Previous cmd: ~a~%" type-code
		byte-size previous-cmd)
	(FORMAT control-connection " Format: ~a, Mode: ~a, Structure: ~a~%" format mode structure)
	(FORMAT control-connection "211 Control status: ~a, Data status: ~a~%"
		(SEND control-connection :status)
		(IF data-connection
		    (SEND data-connection :status)
		    "No connection")))
      (IF (NULL user-name)
	  (WRITE-LINE "530 Please login first" control-connection)
	  (PROGN
	    (SETQ path (FTP-PARSE-FOR-SERVER arg default-directory))
	    (SETQ dir (fs:directory-list path :noerror))
	    (IF (TYPEP dir 'condition)
		(FORMAT control-connection "450 Cannot list ~a: ~a~%" arg dir)
		(CASE (LENGTH dir)
		      (1 (WRITE-LINE "212-Directory listing:" control-connection)
			 (zwei::default-list-one-file (CAR dir) control-connection :string-for-printing)
			 (FORMAT control-connection "212 ~a listed~%" arg))
		      (2 (WRITE-LINE "213-File listing:" control-connection)
			 (zwei::default-list-one-file (CADR dir) control-connection :string-for-printing)
			 (FORMAT control-connection "213 ~a listed~%" path))
		      (t (WRITE-LINE "212-Directory listing:" control-connection)
			 (DOLIST (file (CDR dir))
			   (zwei::default-list-one-file file control-connection :string-for-printing))
			 (FORMAT control-connection "212 ~a listed~%" path))))))))

(DEFUN ftp-server-store (creation-mode &aux mode path done)
  "FTP Server's internal routine used by STOR, APPE and STOU"
  (DECLARE (keyword creation-mode))
  (DECLARE (ATOM mode))
  (DECLARE (ATOM done))
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (DECLARE (SPECIAL data-connection))
  (DECLARE (SPECIAL type-code))
  (DECLARE (SPECIAL byte-size))
  (IF (EQUAL type-code :ascii)
      (SETQ mode t)
      (SETQ mode ()))
  (SETQ path (FTP-PARSE-FOR-SERVER arg default-directory))
  (IF (AND (EQUAL creation-mode :truncate) (EQUAL (SEND path :version) :newest))
      (SETQ creation-mode :new-version))
  (WITH-OPEN-FILE (tofile path :direction :output :if-exists creation-mode :if-does-not-exist :create
			  :characters mode :byte-size byte-size :error nil)
    (IF (OR (NULL tofile) (TYPEP tofile 'condition))
	(FORMAT control-connection "553 Cannot access ~a: ~a~%" arg tofile)
	(IF (ftp-server-data-conn path)
	    (LET ((file-stream tofile))
	      (DECLARE (SPECIAL file-stream))
	      (UNWIND-PROTECT
		  (BLOCK store
		    (SETQ done ())
		    (IF user-name
			(SEND tofile :change-properties () :author user-name))
		    (DO ((buf)
			 (offset)
			 (limit))
			(nil)
		      (MULTIPLE-VALUE-SETQ (buf offset limit)
			(SEND data-connection :read-input-buffer))
		      (IF (NULL buf)
			  (RETURN ()))
		      (SEND tofile :string-out buf offset limit)
		      (SEND data-connection :advance-input-buffer)
		      (IF (ftp-server-abort-check)
			  (RETURN-FROM store)))
		    (SETQ done t))
		(IF done
		    (IGNORE-ERRORS (CLOSE data-connection))
		    (IGNORE-ERRORS (CLOSE data-connection :abort t)))
		(SETQ data-connection ())
		(IF done
		    (FORMAT control-connection "226 ~a ~a~%" arg
			    (IF (EQUAL creation-mode :append)
				"appended"
				"stored"))
		    (FORMAT control-connection "226 ~a ~a aborted~%"
			    (IF (EQUAL creation-mode :append)
				"appending"
				"storing")
			    arg))))))))

(DEFUN ftp-server-xmkd (&aux path result)
  "FTP Server's handler for 'XMKD<SP><pathname><CRLF>' and 'MKD<SP><pathname><CRLF>'"
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (IF (NULL user-name)
      (WRITE-LINE "530 Please login first" control-connection)
      (PROGN
	(SETQ path (FTP-PARSE-FOR-SERVER arg default-directory))
	(IF (PROBE-FILE (SEND path :directory-pathname-as-file))
	    (FORMAT control-connection "550 Directory ~a already exists.~%" path)
	    (IF (EQUAL (SETQ result (fs:create-directory path :error ())) t)
		(PROGN
		  (SETQ path (SEND path :directory-pathname-as-file))
		  (fs:change-file-properties path () :author user-name)
		  (FORMAT control-connection "257 #\"~a#\" directory created~%" path))
		(FORMAT control-connection "550 Cannot create ~a: ~a~%" path result))))))

(DEFUN ftp-server-xrmd (&aux path result)
  "FTP Server's handler for 'XRMD<SP><pathname><CRLF>' and 'RMD<SP><pathname><CRLF>'"
  (DECLARE (SPECIAL arg))
  (DECLARE (SPECIAL default-directory))
  (DECLARE (SPECIAL control-connection))
  (DECLARE (SPECIAL user-name))
  (IF (NULL user-name)
      (WRITE-LINE "530 Please login first" control-connection)
      (PROGN
	(SETQ path (SEND (FTP-PARSE-FOR-SERVER arg default-directory) :directory-pathname-as-file))
	(IF (PROBE-FILE path)
	    (PROGN
	      (SETQ result (SEND path :delete-and-expunge))
	      (IF (TYPEP result 'condition)
		  (FORMAT control-connection "550 Cannot delete directory ~a: ~a~%" path result)
		  (FORMAT control-connection "250 Directory ~a deleted~%" path)))
	    (FORMAT control-connection "550 Directory ~a does not exist.~%" path)))))



))

#!C
; From file FTPINT.LISP#> JONES; HEYERDAHL:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :Common-lisp)
                          (*READTABLE* Common-lisp-readtable)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* Sys::*common-lisp-symbol-substitutions*))
  (COMPILER#:PATCH-SOURCE-FILE "IP: IP; FTPINT.#"

(DEFVAR *ftp-implementation-types* '(:basic :minimal :explorer :MicroExplorer :tops20 :bsd4.2 :wollongong :vms-excelan))

(DEFMETHOD (FTP :case :get-ftp-implementation-type-internal :microExplorer) (IGNORE) :microExplorer)

(DEFMETHOD (FTP :case :login-user-internal :MicroExplorer) (IGNORE)
  (UNLESS user-id
    (force-user-to-login))
  (VALUES user-id ""))

(DEFMETHOD (FTP :case :directory-list-from-ftp :MicroExplorer) (PATHNAME ftp stream fast? &aux data-stream entry)
  (UNLESS (ftp-unfavorable (SEND ftp :open (dir-pathstring-for-ftp pathname) :list))
	 (add-directory-stream stream (SETQ data-stream (SEND ftp :data-connection)))
	 (LOOP for line = (READ-LINE data-stream nil) while line when
	       (SETQ entry (microExplorer-directory-line-parse pathname line fast?)) collecting entry)))
  
(DEFUN MicroExplorer-directory-line-parse (dirobj line fast? &aux cursor cursor2 result creation-date-end
					   month day year hours minutes seconds directoryp)
  (declare (ignore fast?)) 
  ;;; FORMAT:
  ;;;    <pathname>  #-bytes(byte-size)  {!}{@}{$}{@} mm/dd//yy hh:mm:ss {author} DIRECTORY
  ;;;     
  (setf cursor (position #\space line :test #'char-equal :from-end t))
  
  ;; Try to find the creation date. This test may be overkill but might as well be sure.
  (loop
    (setf cursor (position #\/ line :test #'char-equal :end cursor :from-end t))
    (when (and
	    
	    ;; Look for the year and closing paren.
	    (digit-char-p (char line (+ cursor 1)))
	    (digit-char-p (char line (+ cursor 2)))
	    (char-equal (char line (+ cursor 3)) #\))
	    
	    ;; Look for the day and /
	    (digit-char-p (char line (- cursor 1)))
	    (digit-char-p (char line (- cursor 2)))
	    (char-equal (char line (- cursor 3)) #\/)
	    
	    ;; Look for the month and )
	    (digit-char-p (char line (- cursor 4)))
	    (digit-char-p (char line (- cursor 5)))
	    (char-equal (char line (- cursor 6)) #\())
      
      (setf creation-date-end (+ cursor 4))
      (return)))
  
  ;; Now get the author.
  (setf cursor (+ cursor 4))
  
  (cond ((SETQ cursor
	       (POSITION #\Space line :start cursor :test-not 'CHAR-EQUAL))
	 (SETQ cursor2
	       (POSITION #\Space line :start cursor :test 'CHAR-EQUAL))
	 (PUSH (SUBSEQ line cursor cursor2) result))
	;; else
	(t
	 (PUSH (CAR (SEND dirobj :directory)) result)))
  (PUSH :author result)
  
  ;; Now find the directory property
  (when (SETQ cursor
	      (POSITION #\Space (THE string (STRING line)) :start cursor2 :test-not 'CHAR-EQUAL))
    (SETQ cursor2
	  (POSITION #\Space (THE string (STRING line)) :start cursor :test 'CHAR-EQUAL))
    
    (when (string-equal "DIRECTORY" line :start2 cursor :end2 cursor2)
      (setf directoryp t)
      (push t result)
      (PUSH :directory result)))
  
  ;; Start going backwards through the line. Creation date.
  (setf year (PARSE-INTEGER line :start (- creation-date-end 3) :junk-allowed t))
  (setf day (PARSE-INTEGER line :start (- creation-date-end 6) :junk-allowed t))
  (setf month (PARSE-INTEGER line :start (- creation-date-end 9) :junk-allowed t))
  (push (encode-universal-time 0 0 0 day month year) result)
  (push :creation-date result)

  ;; Last modification date.
  (setf cursor (position #\space line
			 :from-end t
			 :test-not #'char-equal
			 :end (position #\space line :from-end t :end creation-date-end :test #'char-equal)))
  
  (setf seconds (parse-integer line :start (- cursor 1) :junk-allowed t))
  (setf minutes (parse-integer line :start (- cursor 4) :junk-allowed t))
  (setf hours   (parse-integer line :start (- cursor 7) :junk-allowed t))

  (setf cursor (position #\space line
			 :from-end t
			 :test-not #'char-equal
			 :end (- cursor 8)))
  
  (setf year    (parse-integer line :start (- cursor 1) :junk-allowed t))
  (setf day     (parse-integer line :start (- cursor 4) :junk-allowed t))
  (setf month   (parse-integer line :start (- cursor 7) :junk-allowed t))

  (push (encode-universal-time seconds minutes hours day month year) result)
  (push :modification-date result)

  ;; get byte size if it is there.
  (setf cursor (position #\space line
			 :from-end t
			 :test-not #'char-equal
			 :end (position #\space line :from-end t :end (- cursor 8) :test #'char-equal)))
  
  (setf cursor2 cursor)
  (when (char-equal (char line cursor) #\))
    (setf cursor2 (position #\( line :from-end t :end cursor :test #'char-equal))
    (unless directoryp
      (push (parse-integer line :start (1- cursor) :junk-allowed t) result)
      (push :byte-size result)))

  (setf cursor (position #\space line :from-end t :test #'char-equal :end cursor2))
  (push (parse-integer line :start cursor :junk-allowed t) result)
  (cond (directoryp
	 (push :length-in-blocks result))
	(t
	 (push :length-in-bytes result)))
  
  ;; Last of all put the pathname on the entry.
  (setf cursor2 (position #\space line :test-not #'char-equal :from-end t :end cursor))
  (setf cursor  (position #\space line :test-not #'char-equal))
  
  (push (MERGE-PATHNAMES (ip:ftp-parse-for-server (SUBSEQ line cursor (+ cursor2 1)) dirobj) (PATHNAME ".")) result))

))
