;;; -*- Mode: Common-Lisp; Package: IP; Base: 10.; Patch-File: T -*-

;;; Reason: fixed display-recent-packet-headers to handle :protocol and :display-data properly.
;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;; Patch file for IP version 3.20
;;; Written 06/01/88 07:30:28 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 4.70, VIRTUAL-MEMORY 4.4, EH 4.5, MAKE-SYSTEM 4.5, MICRONET 4.5, LOCAL-FILE 4.2,
;;;  BASIC-PATHNAME 4.13, NETWORK-SUPPORT-COLD 4.1, NAMESPACE 4.22, NETWORK-NAMESPACE 4.3,
;;;  DISK-IO 4.14, DISK-LABEL 4.0, BASIC-FILE 4.11, MAC-PATHNAME 4.7, NETWORK-PATHNAME 4.1,
;;;  COMPILER 4.12, TV 4.92, DATALINK 4.15, CHAOSNET 4.19, GC 4.3, MEMORY-AUX 4.1,
;;;  NVRAM 4.6, SYSLOG 4.0, STREAMER-TAPE 4.5, UCL 4.1, INPUT-EDITOR 4.0, METER 4.3,
;;;  ZWEI 4.21, DEBUG-TOOLS 4.2, NETWORK-SUPPORT 4.6, NETWORK-SERVICE 4.0, DATALINK-DISPLAYS 4.0,
;;;  FONT-EDITOR 4.0, SERIAL 4.0, PRINTER 4.9, PRINTER-TYPES 4.4, IMAGEN 4.0, SUGGESTIONS 4.0,
;;;  MAIL-DAEMON 4.9, MAIL-READER 4.9, TELNET 4.1, VT100 4.5, NAMESPACE-EDITOR 4.5,
;;;  PROFILE 4.4, VISIDOC 4.5, IP 3.19, Experimental BUG 10.1,  microcode 308, Band Name: Rel 4.1 + IP 5/9

#!C
; From file IP-DEBUG.LISP#> IP; SYS:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; IP-DEBUG.#"


(DEFUN DISPLAY-RECENT-PACKET-HEADERS (&KEY (STREAM *STANDARD-OUTPUT*) (LIMIT 16)
					   ;max number of packets to display
				      (IP-HEADER T)	   ;display the ip header part?
				      (TRANSPORT-HEADER T) ;display the udp, tcp, icmp part?
				      (DISPLAY-DATA 0)	   ;how much data?
				      (PROTOCOLS '(:TCP :UDP :ICMP))	   ;show which protocol pkts?
				      (PORT NIL)   ;show if src OR dst matches, udp or tcp port
				      (SRC-PORT NIL)	   ;show if src port matches
				      (DST-PORT NIL)	   ;show if dst port matches
				      (ADDRESS NIL)	   ;show if src or dst matches, ip address    
				      (SRC-ADDR NIL)	   ;show if src address matches
				      (DST-ADDR NIL)	   ;show if dst address matches
				      (RECEIVE T)  ;show received packets
				      (TRANSMIT T))	   ;show transmitted packets
  
  "Outputs a formatted summary of most information in packet copies in the history buffer.
   The packets to be displayed can be chosen via a variety of criteria.  See the source
   for details.  To initiate trace, set *record-packet-mode* to T."
  (CATCH 'DISPLAY-DONE
    (LET* (I
	   (DISPLAYED 0)
	   (DISPLAY-DATA-LENGTH 0)
	   (RECENT-PACKET-INDEX NEXT-RECENT-PACKET))
      (DOTIMES (X *MAX-RECENT-PACKETS* NIL)
	(WHEN (> DISPLAYED LIMIT)
	  (THROW 'DISPLAY-DONE
		 ()))
	(SETQ I (MOD (- RECENT-PACKET-INDEX X 1) *MAX-RECENT-PACKETS*))
	(UNLESS (OR (NULL (AREF RECENT-PACKETS I)) (NULL (RCNT-PKT-COPY (AREF RECENT-PACKETS I))))
	  (LET* ((SLOT (AREF RECENT-PACKETS I))
		 (PKT (RCNT-PKT-COPY SLOT))
		 (UDP-PKT (MAKE-IP-DATA PKT))
		 (TCP-PKT UDP-PKT)
		 (PKT-PROTOCOL
		   (CAR (RASSOC (IP-PROTOCOL PKT) *IP-PROTOCOL-MAPPING* :TEST #'EQUAL))))
	    (WHEN (AND (MEMBER PKT-PROTOCOL PROTOCOLS :TEST #'EQUAL)
		       (OR
			 (AND RECEIVE	   ;include sends and receives
			      (MEMBER (RCNT-PKT-MODE (AREF RECENT-PACKETS I)) '(:RCV :RECEIVE) :TEST
				      #'EQUAL))
			 (AND TRANSMIT
			      (MEMBER (RCNT-PKT-MODE (AREF RECENT-PACKETS I)) '(:XMT :XMIT :TRANSMIT)
				      :TEST #'EQUAL)))
		       (OR (AND (NULL ADDRESS)	   ;include all addresses
				(NULL SRC-ADDR) (NULL DST-ADDR))
			   (AND (NUMBERP ADDRESS)
				(OR (EQL ADDRESS (IP-SRC-ADDR PKT)) (EQL ADDRESS (IP-DST-ADDR PKT))))
			   (AND (NUMBERP SRC-ADDR) (EQL SRC-ADDR (IP-SRC-ADDR PKT)))
			   (AND (NUMBERP DST-ADDR) (EQL DST-ADDR (IP-DST-ADDR PKT))))
		       (OR (AND (NULL PORT)	   ;include all ports
				(NULL SRC-PORT) (NULL DST-PORT))
			   (SELECT PKT-PROTOCOL
			     (:UDP
			      (OR
				(AND (NUMBERP PORT)
				     (OR (EQL PORT (UDP-HEADER-SOURCE-PORT UDP-PKT))
					 (EQL PORT (UDP-HEADER-DESTINATION-PORT UDP-PKT))))
				(AND (NUMBERP SRC-PORT) (EQL SRC-PORT (UDP-HEADER-SOURCE-PORT UDP-PKT)))
				(AND (NUMBERP DST-PORT)
				     (EQL DST-PORT (UDP-HEADER-DESTINATION-PORT UDP-PKT)))))
			     (:TCP
			      (OR
				(AND (NUMBERP PORT)
				     (OR (EQL PORT (TCP-HEADER-SOURCE-PORT TCP-PKT))
					 (EQL PORT (TCP-HEADER-DESTINATION-PORT TCP-PKT))))
				(AND (NUMBERP SRC-PORT) (EQL SRC-PORT (TCP-HEADER-SOURCE-PORT TCP-PKT)))
				(AND (NUMBERP DST-PORT)
				     (EQL DST-PORT (TCP-HEADER-DESTINATION-PORT TCP-PKT)))))
			     (:OTHERWISE T))))
	      (DISPLAY-RECORDED-PACKET-HEADER I STREAM IP-HEADER TRANSPORT-HEADER)
	      (SETF DISPLAY-DATA-LENGTH (MIN DISPLAY-DATA (* 2 (IP-TOTAL-LENGTH PKT))))
	      (WHEN (PLUSP DISPLAY-DATA-LENGTH)
		(PRINT-PACKET TCP-PKT :STREAM STREAM :SIZE DISPLAY-DATA-LENGTH)) 
	      (INCF DISPLAYED))))))))
))
