;;; -*- Mode: Common-Lisp; Package: IP; Base: 10.; Patch-File: T -*-

;;; Reason: Kludge to work around TCP PC implementations that try to read the window field as a signed number. [8209]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for IP version 3.23
;;; Written 07/11/88 08:00:55 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 4.74, VIRTUAL-MEMORY 4.4, EH 4.5, MAKE-SYSTEM 4.6, MICRONET 4.5, Experimental LOCAL-FILE 4.2,
;;;  BASIC-PATHNAME 4.13, NETWORK-SUPPORT-COLD 4.1, NAMESPACE 4.23, NETWORK-NAMESPACE 4.3,
;;;  DISK-IO 4.16, DISK-LABEL 4.0, BASIC-FILE 4.13, MAC-PATHNAME 4.10, NETWORK-PATHNAME 4.1,
;;;  COMPILER 4.13, TV 4.93, DATALINK 4.16, CHAOSNET 4.19, GC 4.3, MEMORY-AUX 4.1,
;;;  NVRAM 4.6, SYSLOG 4.0, STREAMER-TAPE 4.8, UCL 4.1, INPUT-EDITOR 4.0, METER 4.3,
;;;  ZWEI 4.21, DEBUG-TOOLS 4.2, NETWORK-SUPPORT 4.6, NETWORK-SERVICE 4.0, DATALINK-DISPLAYS 4.0,
;;;  FONT-EDITOR 4.0, SERIAL 4.0, PRINTER 4.10, PRINTER-TYPES 4.5, IMAGEN 4.1, SUGGESTIONS 4.0,
;;;  MAIL-DAEMON 4.9, MAIL-READER 4.9, TELNET 4.1, VT100 4.6, NAMESPACE-EDITOR 4.5,
;;;  PROFILE 4.4, VISIDOC 4.5, IP 3.21, Experimental BUG 10.0,  microcode 308, Band Name: Rel 4.1.1 + IP 6/15

#!C
; From file TCPCONN.LISP#> IP; SYS:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; TCPCONN.#"


(defvar *max-syn-window-size-bit-length* 16.
  "This variable determines the number of bits that will be used when determining the window size.
   This variable should normally be set to 16.
   To communicate to IP implementations that treat this field as a signed number, set it to 15.")

(DEFUN prescan-rd-q ()
  "This method is used in syn-received state to determine the size of the receive-window to advertise,
prior to the actual registering of the receive descriptors."
  (DECLARE (:self-flavor tcp-connection))
  (DO ((rd rd-q (rd-link rd))
       (syn-window 0))
     ((OR (NULL rd)
	   (NOT (EQ (rd-state rd) :pending))
	   ;; don't consider more than the window can hold
	   (> (+ syn-window (rd-length rd)) (1- (EXPT 2 *max-syn-window-size-bit-length* ))))  ;07-11-88 DAB
       ;; return value
       syn-window)
    (WHEN (EQ (rd-state rd) :pending)
      (INCF syn-window (rd-length rd)))))

(DEFUN scan-rd-q ()
  "Scan rd-q register all possible receive descriptors."
  (DECLARE (:self-flavor tcp-connection))
  (DO ((rd rd-q (rd-link rd))
       (old-receive-window receive-window))
      ((OR (NULL rd)
	   ;; don't register more than the window can hold
	   (AND (EQ (rd-state rd) :pending)
		(> (+ rd-registered-length (rd-length rd))
		   (1- (EXPT 2 *max-syn-window-size-bit-length*)))))   ;07-11-88 DAB

       ;; update receive-window as per RFC813
       (WHEN (<= (* 2 receive-window) (- rd-registered-length (-seq receive-next return-sequence)))
	 (SETF receive-window (- rd-registered-length (-seq receive-next return-sequence))))
       ;; send delayed acknowledgement immediately when window opens
       (WHEN (AND delayed-ack-timestamp (NOT (EQUAL old-receive-window receive-window)))
	 (SETF ack-due-p t))
       ;; when receive-window opens, process queued remote-probe-segment
       (WHEN (AND (ZEROP old-receive-window) (PLUSP receive-window))
	 ;; according to the spec this new window information will be rejected, unless the seq or ack numbers increase
	 (WHEN *window-kludge-switch* (SETF ack-due-p t))
	 (WHEN remote-probe-segment
	   (queue remote-probe-segment receive-q receive-q-end)
	   (SETF remote-probe-segment ()))))
    (WHEN (EQ (rd-state rd) :pending)
      (INCF rd-registered-length (rd-length rd))
      (SETF (rd-state rd) :registered))))


))
