;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Patch file for IP version 3.27
;;; Reason: Change IP to use new checksum miscop.  8/25/88.
;;; Written 08/25/88 04:27:31 by ab,
;;; while running on MX6 from band N817
;;; With SYSTEM 5.8, GC 5.0, VIRTUAL-MEMORY 5.1, MICRONET 5.2, MICRONET-COMM 5.2,
;;;  DISK-IO 5.0, BASIC-PATHNAME 5.1, MAC-PATHNAME 5.0, Experimental NETWORK-SUPPORT-COLD 5.0,
;;;  BASIC-NAMESPACE 5.0, BASIC-FILE 5.3, Inconsistent RPC 5.3, Experimental NFS 5.5,
;;;  EH 5.0, MAKE-SYSTEM 5.0, COMPILER 5.0, TV 5.0, NVRAM 5.0, UCL 5.0, INPUT-EDITOR 5.0,
;;;  METER 5.0, ZWEI 5.3, DEBUG-TOOLS 5.0, WINDOW-MX 5.5, Experimental NETWORK-PATHNAME 5.0,
;;;  Experimental NETWORK-NAMESPACE 5.0, DATALINK 5.2, CHAOSNET 5.2, Experimental NETWORK-SUPPORT 5.0,
;;;  Experimental NETWORK-SERVICE 5.0, DATALINK-DISPLAYS 5.0, Experimental NAMESPACE-EDITOR 5.0,
;;;  IP 3.26, Experimental NFS-SERVER 5.0, PRINTER 5.0, MAIL-DAEMON 5.0, MAIL-READER 5.0,
;;;  TELNET 5.0, VT100 5.0, VISIDOC 5.0, PROFILE 5.0,  microcode 82, Band Name: mX Network (8-17) trained

;;; This patch should be a no-op unless you're running a REL 5 system.  The %SUM-ARRAY miscop
;;; is not defined before rel 5.

#!C
; From file IP.LISP#> IP; MR-X:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "sys: IP; IP.#"

(WHEN (>= (si:get-system-version 'system) 5)
  
  (DEFUN ip-ones-complement-checksum  (array &optional (ARRAY-LENGTH (1- (LENGTH array))) &rest parm-list)
    (let ((elem-size (array-element-size array))
	  (offset 0)
	  (sum nil))
      (loop while (array-indirect-p array)
	    as new-elem-size = (array-element-size array)
	    as index-offset = (sys:array-index-offset array)
	    when index-offset
	    do (cond ((= new-elem-size elem-size)
		      (incf offset index-offset))
		     ((< new-elem-size elem-size)      ; Going to assume only 8 or 16 bit elements, so new is 8 here.
		      (setq elem-size new-elem-size
			    offset    (ash offset 1))
		      (incf offset index-offset))
		     ((> new-elem-size elem-size)      ; Going to assume only 8 or 16 bit elements, so new is 16 here.
		      (when (oddp offset)	       ; May be a premature check.  The old form handles these.
			(error "Can't checksum 16b array if length not on byte boundary."))
		      (setq elem-size new-elem-size
			    offset    (ash offset -1))
		      (incf offset index-offset)))
	    doing (setq array (sys:array-indirect-to array)))
      
      (setq sum (si:%sum-array array offset array-length))     ;ksct 8/25/88  use new miscop.
      
      (setq sum (+ (ash sum -8) (ash (logand #xff sum) 8)))
      (loop for parm in parm-list
	    doing (incf sum parm))
      (loop until (< sum #x10000)
	    doing (setq sum (+ (logand #xFFFF sum) (ash sum -16))))
      
      ;; Return the ones COMPLEMENT of the sixteen bit number.  If the checksum is 0 (#XFFFF
      ;; before the LOGXOR), return >FFFF. 
      (if (= #xffff sum)
	  (values #xffff)
	  (values (logxor #xffff sum)))))
  
  (DEFUN ip-ones-complement-checksum-16 (sixteen-bit-array length &rest parm-list)
    "Calculates the ones complement checksum for an art-16b array."
    (let (sum)
      (setq sum (si:%sum-array sixteen-bit-array 0 length))
      (loop for parm in parm-list
	    doing (incf sum parm))
      (loop until (< sum #x10000)
	    doing (setq sum (+ (logand #xFFFF sum) (ash sum -16))))
      (if (= #xffff sum)
	  (values #xffff)
	  (values (logxor #xffff sum)))))
  
  (DEFUN ip-ones-complement-checksum-8 (eight-bit-array length &rest parm-list)
    "Calculates the ones complement checksum for an art-8b array."
    (let (sum)
      (SETQ sum (si:%sum-array eight-bit-array 0 length))
      (loop for parm in parm-list
	    doing (incf sum parm))
      (loop until (< sum #x10000)
	    doing (setq sum (+ (logand #xFFFF sum) (ash sum -16))))
      (if (= #xffff sum)
	  (values #xffff)
	  (values (logxor #xffff sum)))))
  )

))




