;;; -*- Mode: Common-Lisp; Package: FS; Base: 10.; Patch-File: T -*-

;;; Reason: Bsd4.2-directory-line-parse would barf when doing a directory-list of a unix directory that contains a file with a name of " ". Now we will ignore this file. [8757].

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 10/03/88 07:27:57 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 5.8, VIRTUAL-MEMORY 5.4, EH 5.1, MICRONET 5.2, LOCAL-FILE 5.0, BASIC-PATHNAME 5.1,
;;;  NETWORK-SUPPORT-COLD 5.0, BASIC-NAMESPACE 5.1, NETWORK-NAMESPACE 5.0, DISK-IO 5.1,
;;;  DISK-LABEL 5.0, BASIC-FILE 5.3, MAC-PATHNAME 5.0, NETWORK-PATHNAME 5.0, COMPILER 5.0,
;;;  MAKE-SYSTEM 5.0, TV 5.1, DATALINK 5.4, CHAOSNET 5.3, GC 5.1, MEMORY-AUX 5.0,
;;;  NVRAM 5.0, SYSLOG 5.0, STREAMER-TAPE 5.0, UCL 5.0, INPUT-EDITOR 5.0, METER 5.0,
;;;  ZWEI 5.4, DEBUG-TOOLS 5.0, NETWORK-SUPPORT 5.0, NETWORK-SERVICE 5.0, DATALINK-DISPLAYS 5.0,
;;;  FONT-EDITOR 5.0, SERIAL 5.0, PRINTER 5.3, PRINTER-TYPES 5.2, IMAGEN 5.1, SUGGESTIONS 5.0,
;;;  MAIL-DAEMON 5.0, MAIL-READER 5.0, TELNET 5.0, VT100 5.0, Experimental NAMESPACE-EDITOR 5.1,
;;;  PROFILE 5.0, VISIDOC 5.2, Experimental BUG 10.2, IP 3.28, VISIDOC-SERVER 5.2,
;;;   microcode 397, Band Name: Pre 5.0 +1p 9/23 +visidoc

#!C
; From file FTPINT.LISP#> IP; SYS:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; FTPINT.#"


(DEFUN bsd4.2-directory-line-parse (dirobj directory-name open-name line fast? in-subdirectory
				    &aux pathobj result start cursor cursor2 year day month
				    (minutes 0) (hours 0) (seconds 0) time-or-year link-ix device)
  "Open-name is equivalent to parameter pathname considering directory link translation.
Directory-name is the pathname actually passed across the control connection."
  ;; FORMATS:
  ;;   total n
  ;;   {n}  [d|-|]rwxrwxr-- n kari   1152 Apr 10 1985  chess
  ;;          "                            Jun 7  11:48  foo.bar 
  ;;   {n}  l              "                                       -> link translation
  ;;   {n}  [b|c]rw-rw-rw- n xxxx   m,n   "
  ;;   directory-pathname:
  (SETQ start (POSITION #\Space (THE string (STRING line)) :test-not 'CHAR-EQUAL))
  (COND
    ;; throw out ignored lines
    ((NULL start)  ;; ignore blank lines
     (VALUES nil in-subdirectory))
    ((AND (> (LENGTH line) 5)  ;; the totals line is ignored
	  (STRING-EQUAL (SUBSEQ line start (+ start 6)) "total "))
     (VALUES nil in-subdirectory t))
    ;; top-level directory name (followed by subordinate files)
    ((CHAR-EQUAL (AREF line (SETQ cursor2 (1- (LENGTH line)))) #\:)
     (SETF (AREF line cursor2) #\Space) (SETQ in-subdirectory t)
     (LET ((p1 (IGNORE-ERRORS (parse-pathname line nil dirobj))))
       (IF (PATHNAMEP p1)
	   (PROGN
	     (OR (PATHNAME-NAME p1) (SETQ p1 (SEND p1 :new-name :unspecific)))
	     (OR (PATHNAME-TYPE p1) (SETQ p1 (SEND p1 :new-type :unspecific)))
	     (SETQ pathobj (IGNORE-ERRORS (merge-pathname-defaults p1 dirobj))))
	   (SETQ pathobj nil)))
     (VALUES (WHEN pathobj
	       (IF fast?
		   (LIST pathobj)
		   (LIST pathobj :directory t :dired-echo "")))
	     in-subdirectory))
    ;; don't include subordinate files (they don't make any sense to dired at this point)
    (in-subdirectory (VALUES nil t))
    ;; regular lines 
    (:otherwise
     (WHEN (AND (SETF cursor (STRING-SEARCH-SET "dbcp-l" line))	; Get start of mode
		(< (+ cursor 3) (LENGTH line))	; don't run off the end
		(OR (EQUAL (AREF line (+ cursor 1)) #\r)	; next character is r or
		    (EQUAL (AREF line (+ cursor 1)) #\-))	; a dash.
		(OR (EQUAL (AREF line (+ cursor 2)) #\w)	; next character is w or
		    (EQUAL (AREF line (+ cursor 2)) #\-))	; a dash.
		(OR (EQUAL (AREF line (+ cursor 3)) #\x)	; next character is x or
		    (EQUAL (AREF line (+ cursor 3)) #\-)))	; a dash.
       (SELECTOR (AREF line cursor) char-equal
	 (#\d (SETF (GETF result :directory) t))
	 (#\l
	  ;; This link really needs to be chased - do it when the control connection is free in modify-dir-list.
	  ;; When the link is chased, we will insert the :directory property, if necessary.  Note that the
	  ;; :directory and :link-to properties are the only ones that refer to the end of the link chain.
	  ;; All other properties refer to the link itself.
	  (SETF cursor2 (SEARCH " ->" (STRING line) :from-end t :test 'CHAR-EQUAL))
	  (SETF link-ix (STRING-SEARCH-NOT-SET #\SPACE line (+ 3 cursor2))))
	 ((#\b #\c) (SETQ device t))))
     (MULTIPLE-VALUE-SETQ (cursor cursor2)
       (backup-token line (WHEN link-ix cursor2)))
     (if
       ;;Some unix system allow a file to be name " " [space]. In this case backup-token will return the hour:minute field
       ;;as the pathname object. 
       ;;We can not handle this because pathname merging
       ;;will alway replace the " " with a name from the defaults. Therefore, ignore it.
       (and (= (- cursor2 cursor) 5)
	    (eql (aref (SUBSEQ line cursor cursor2) 2) #\:))
       (setf pathobj nil)
       ;else
       (LET ((p1 (parse-pathname (SUBSEQ line cursor cursor2) nil dirobj)))
	 (OR (PATHNAME-NAME p1) (SETQ p1 (SEND p1 :new-name :unspecific)))
	 (OR (PATHNAME-TYPE p1) (SETQ p1 (SEND p1 :new-type :unspecific)))
	 (SETQ pathobj (merge-pathname-defaults p1 dirobj)))
       ;; we cannot determine the link-name until pathobj has been determined
       ;; merge-pathnames is necessary here in case link-name has a :relative or :up directory component
       (WHEN (AND pathobj link-ix)
	 (SETF (GETF result :link-to)
	       (MERGE-PATHNAMES
		 (parse-pathname (SUBSEQ line link-ix (STRING-SEARCH-SET #\SPACE line link-ix)) nil dirobj)
		 (IF (AND (NOT (pathname-equal directory-name open-name))
			  (pathname-equal pathobj directory-name))
		     directory-name
		     open-name)))))
     (COND
       ((NULL pathobj) (VALUES nil in-subdirectory))
       (fast? (VALUES (LIST pathobj) in-subdirectory))
       (:otherwise
	;; CREATION DATE
	(SETQ cursor (backup-token line cursor))
	(MULTIPLE-VALUE-SETQ (time-or-year cursor2)
	  (PARSE-INTEGER line :start cursor :junk-allowed t))
	(IF (CHAR-EQUAL #\: (AREF line cursor2))
	    (PROGN
	      (SETQ hours time-or-year)
	      (SETQ minutes (PARSE-INTEGER line :start (1+ cursor2) :junk-allowed t)))
	    ;; else
	    (SETQ year time-or-year))
	(SETQ cursor (backup-token line cursor))
	(SETQ day (PARSE-INTEGER line :start cursor :junk-allowed t))
	(SETQ cursor (backup-token line cursor))
	(SETQ month (translate-alphabetic-month (SUBSEQ line cursor (+ cursor 3))))
	(WHEN (AND (NUMBERP seconds) (NUMBERP minutes) (NUMBERP hours) (NUMBERP day) (NUMBERP month))
	  (MULTIPLE-VALUE-BIND (this-sec this-min this-hour this-day this-month this-year) (time:get-decoded-time)
	    (IGNORE this-sec this-min this-hour this-day)
	    ;; Year appears only if date is > 180 days old.  Otherwise decide whether assumed year is last year or this year.
	    ;; Note that July 1st is either day 181 (non-leap year) or day 182 (leap year).
	    (UNLESS year
	      (SETF year (IF (>= month (+ this-month 6))
			     (1- this-year)
			     this-year)))
	    (SETF (GETF result :creation-date) (time:encode-universal-time seconds minutes hours day month year)))) 
	;; LENGTH-IN-BYTES 
	(UNLESS device
	  (SETQ cursor (backup-token line cursor))
	  (SETF (GETF result :length-in-bytes) (PARSE-INTEGER line :start cursor :junk-allowed t)))
	;; DIRED-ECHO
	(SETF (GETF result :dired-echo) (SUBSEQ line 0 cursor))
	(PUSH pathobj result))))))
))
