;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason:  Fixed (FTP :case :directory-list-from-ftp :bsd4.2) to handle "not found" errors from SUN os 4.0.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 10/20/88 14:50:23 by BERGER,
;;; while running on ARIES from band LODX
;;; With NIL SYSTEM 5.8, VIRTUAL-MEMORY 5.4, EH 5.1, MICRONET 5.2, LOCAL-FILE 5.0,
;;;  BASIC-PATHNAME 5.2, NETWORK-SUPPORT-COLD 5.0, BASIC-NAMESPACE 5.3, NETWORK-NAMESPACE 5.0,
;;;  DISK-IO 5.5, DISK-LABEL 5.1, BASIC-FILE 5.3, MAC-PATHNAME 5.0, NETWORK-PATHNAME 5.0,
;;;  COMPILER 5.1, MAKE-SYSTEM 5.1, TV 5.7, DATALINK 5.4, CHAOSNET 5.3, GC 5.2, MEMORY-AUX 5.0,
;;;  NVRAM 5.0, SYSLOG 5.0, STREAMER-TAPE 5.6, UCL 5.0, INPUT-EDITOR 5.0, METER 5.0,
;;;  ZWEI 5.5, DEBUG-TOOLS 5.0, NETWORK-SUPPORT 5.0, NETWORK-SERVICE 5.0, DATALINK-DISPLAYS 5.0,
;;;  FONT-EDITOR 5.0, SERIAL 5.0, PRINTER 5.10, PRINTER-TYPES 5.2, IMAGEN 5.1, SUGGESTIONS 5.0,
;;;  MAIL-DAEMON 5.0, MAIL-READER 5.0, TELNET 5.0, VT100 5.0, Experimental NAMESPACE-EDITOR 5.1,
;;;  PROFILE 5.0, VISIDOC 5.3, Experimental BUG 10.2, IP 3.29, VISIDOC-SERVER 5.2,
;;;   microcode 398, Band Name: Pre 5.0 +1p 9/23 +visidoc

#!C
; From file FTPINT.LISP#> IP; sys:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "sys: IP; FTPINT.#"


(DEFMETHOD (FTP :case :directory-list-from-ftp :bsd4.2) (pathname ftp stream fast?
							 &aux data-stream line entry entry-list
							 in-subdirectory total-line-p open-name directory-name)
  (let ((sys:*new-line-delimiter* '(#\newline #\linefeed))) ;01-27-88 DAB
    (declare (special sys:*new-line-delimiter* ))
    (BLOCK bsd4.2
    (LOOP
      ;; open-name is equivalent to parameter pathname considering directory link translation
      ;; directory-name is the pathname actually passed across the control connection
      (SEND ftp :open
	    (PROG1 (SETF directory-name (dir-pathstring-for-ftp (SETF open-name (OR *ftp-directory-link* pathname))))
		   (SETF directory-name (parse-pathname directory-name nil pathname)))
	    ;; Unix nlist returns a null listing for both nonexistant and empty directories
	    ;; Don't use nlist to speed up the listing on a recursive entry, or empty directories will appear as nonexistant
	    :list)
      ;; workaround a Unix 425 bug which seems to go away on the second open attempt
      (UNLESS (AND (EQ 425 (SEND ftp :reply-code))
		   (SEARCH "Can't create data socket" (SEND ftp :reply-string))
		   (SEARCH "Address already in use" (SEND ftp :reply-string)))
	(RETURN)))
    (UNLESS (or (ftp-unfavorable (SEND ftp :reply-code))
		(not (SEND ftp :data-connection)))   ;03-03-88 DAB
      (SETF data-stream (SEND ftp :data-connection))
      ;; work around Unix bug of giving favorable status for a nonexistant directory
      (WHEN (NULL (SETF line (READ-LINE data-stream nil)))
	(COND
	  ;; if FTP list command on parent directory, then directory not found
	  ((OR *ftp-recursive-directory-list-from-ftp*
	       (NOT (pathname-equal directory-name open-name))
	       (search " not found" line))   ;10-20-88 DAB
	   (SETF (SEND ftp :reply-code) 450)
	   (SETF (SEND ftp :reply-string) (FORMAT nil "Directory not found for ~a" pathname))
	   (RETURN-FROM bsd4.2 nil))
	  ;; if FTP list command on root directory, then file not found 
	  ((EQL :root (SEND pathname :directory))
	   (RETURN-FROM bsd4.2 nil))
	  ;; must check existence of parent directory to make decision
	  (t (LET ((*ftp-recursive-directory-list-from-ftp* t))
	       (SEND ftp :close)
	       (SEND self :directory-list-from-ftp :bsd4.2
		     (SEND pathname :directory-pathname-as-file) ftp stream fast?)
	       ;; reply code is set by the recursive call
	       (RETURN-FROM bsd4.2 nil)))))
      ;; work around Unix bug of giving favorable status for a protected directory
      (WHEN (SEARCH (STRING-APPEND (SEND directory-name :string-for-host) " unreadable") line)
	(SETF (SEND ftp :reply-code) 450)
	(SETF (SEND ftp :reply-string) (FORMAT nil "~a: Permission denied" pathname))
	(RETURN-FROM bsd4.2 nil))
      ;; when recursively entered, we were only checking the existance of the parent directory, so exit.
      ;; (Note also that Unix will chase links with the NLST comand, so we don't have to worry about links.)
      (IF *ftp-recursive-directory-list-from-ftp*
	  (RETURN-FROM bsd4.2 nil)
	  (add-directory-stream stream data-stream))
      (MULTIPLE-VALUE-SETQ (entry in-subdirectory total-line-p)
	(bsd4.2-directory-line-parse pathname directory-name open-name line fast? in-subdirectory))
      ;; They gave us /d1/d2.  Unix lists d2 if it is a directory.
      ;; We considered d2 as the file spec and thus don't match anything if d2 is a directory.
      ;; Assumption is based on the fact that we saw the total line first (e.g. total n) when only a single file was requested.
      (WHEN (AND total-line-p (pathname-equal directory-name open-name))
	(RETURN-FROM bsd4.2 (LIST (LIST pathname :directory t))))
      ;; Workaround Unix feature of not chasing the link path with the LIST command (and returning the link information instead),
      ;; when we were listing the parent directory.  We throw out with a single entry, which needs its link chased.
      (WHEN (AND (GETF (CDR entry) :link-to)
		 (NOT (pathname-equal directory-name open-name))
		 (pathname-equal (FIRST entry) directory-name))
	(THROW :directory-link-error (LIST entry)))
      (LOOP
	(WHEN entry (PUSH entry entry-list))
	(UNLESS (SETF line (READ-LINE data-stream nil))
	  (RETURN-FROM bsd4.2 entry-list))
	(MULTIPLE-VALUE-SETQ (entry in-subdirectory total-line-p)
	  (bsd4.2-directory-line-parse pathname directory-name open-name line fast? in-subdirectory)))))))
))
