;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Patch file for IP version 3.32
;;; Reason: Give the ip background processes an attribute of restart-after-boot. This prevents the process object from being
;;; given to another process before we are through with it.
;;; Written 11/04/88 12:39:39 by jones,
;;; while running on HOBBS from band LOD1
;;; With SYSTEM 5.17, GC 5.3, VIRTUAL-MEMORY 5.5, MICRONET 5.5, Inconsistent MICRONET-COMM 5.12,
;;;  DISK-IO 5.8, BASIC-PATHNAME 5.2, MAC-PATHNAME 5.0, NETWORK-SUPPORT-COLD 5.1,
;;;  BASIC-NAMESPACE 5.5, BASIC-FILE 5.3, RPC 5.4, NFS 5.8, EH 5.2, MAKE-SYSTEM 5.1,
;;;  MEMORY-AUX 5.1, MACTOOLBOX 1.20, COMPILER 5.1, TV 5.15, NVRAM 5.1, UCL 5.0, INPUT-EDITOR 5.0,
;;;  METER 5.0, ZWEI 5.8, DEBUG-TOOLS 5.0, WINDOW-MX 5.24, PRINTER 5.10, MAC-PRINTER-TYPES 5.2,
;;;  NETWORK-PATHNAME 5.0, NETWORK-NAMESPACE 5.0, DATALINK 5.5, CHAOSNET 5.4, NETWORK-SUPPORT 5.0,
;;;  NETWORK-SERVICE 5.0, DATALINK-DISPLAYS 5.0, NAMESPACE-EDITOR 5.1, IP 3.31, NFS-SERVER 5.3,
;;;  PRINTER-TYPES 5.2, IMAGEN 5.1, MAIL-DAEMON 5.0, MAIL-READER 5.3, TELNET 5.0,
;;;  VT100 5.0, STREAMER-TAPE 5.6, DECNET 1.39, VISIDOC 5.4, PROFILE 5.1, DISK-LABEL 5.1,
;;;   microcode #<DTP-STACK-GROUP ...error printing #<Stack-group 1140111>..., Band Name: nb04

#!C
; From file UDPHAND.LISP#> IP; MR-X:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :Common-lisp)
                          (*READTABLE* Sys:Common-lisp-readtable)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* Sys::*common-lisp-symbol-substitutions*))
  (COMPILER#:PATCH-SOURCE-FILE "Sys: IP; UDPHAND.#"


(DEFUN reset-udp-service (&optional (enable-p nil) (debug-p nil) &aux background-process)
  "Reset all UDP data structures.
ENABLE-P (Optional) Enable UDP server when t (nil default).
DEBUG-P (Optional) Enable UDP debugging when t (nil default)."
  (WHEN (TYPEP *udp-handler* 'udp-handler)
    (WITHOUT-INTERRUPTS
      (MAPC #'(lambda (entry)
		(SEND *udp-handler* :return-port (CDR entry)))
	    *udp-receive-list*)
      (SETF *udp-receive-list* ()))
    (WHEN (SETF background-process (SEND *udp-handler* :background-process))
      (SEND background-process :kill t)))
  (CLEAR-RESOURCE 'udp-packet nil nil)
  (CLEAR-RESOURCE 'udp-ports nil nil)
  (COND ((NOT enable-p)
	 (SETF *udp-handler* #'(lambda (operation &rest ignore)
				 (CASE operation
				   (:receive-data nil)
				   (otherwise
				    (FERROR 'udp-error "UDP Service is not enabled."))))))
	(t (SETF *udp-handler* (MAKE-INSTANCE (IF (TYPEP *udp-handler* 'udp-handler)
						  *udp-handler*
						  'udp-handler)
					      :background-process nil
					      :background-event nil
					      :background-requests nil
					      :packets-received 0
					      :checksum-errors 0
					      :server-packet-list nil
					      :server-packet-timeouts 0
					      :server-packet-claims 0
					      :nonexistant-ports 0))
	   (SETF (SEND *udp-handler* :background-process)
		 (PROCESS-RUN-FUNCTION '(:name "UDP Background" :priority 15 :restart-after-reset t :restart-after-boot t)
				       *udp-handler* :run-function))
	   (SETF *udp-debug-mode* debug-p)))
  "UDP Reset Complete") 

))

#!C
; From file TCPHAND.LISP#> IP; MR-X:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :Common-lisp)
                          (*READTABLE* Sys:Common-lisp-readtable)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* Sys::*common-lisp-symbol-substitutions*))
  (COMPILER#:PATCH-SOURCE-FILE "Sys: IP; TCPHAND.#"


(DEFUN reset-tcp-service (&optional (enable-p nil) &aux background-process)
  "Reset all TCP data structures.
ENABLE-P (Optional) Enable TCP service when t (nil default)."
  (DECLARE (SPECIAL *ip-area*))
  ;; disable tcp
  (WHEN (TYPEP *tcp-handler* 'tcp-handler)
    (MAPHASH
      #'(lambda (IGNORE tcb)
	  (DO ((p tcb q)
	       q)
	      ((NULL p))
	    (SETF q (SEND p :port-link))
	    (SEND *tcp-handler* :delete-tcb p :dont-deregister)))
      (SEND *tcp-handler* :tcb-table))
    ;; nail the background process and make sure its unwinds complete
    (WHEN (SETF background-process (SEND *tcp-handler* :background-process))
      (SEND background-process :kill t)))
  (CLEAR-RESOURCE 'tcbs nil nil)
  (CLEAR-RESOURCE 'tcp-segment nil nil)
  (CLEAR-RESOURCE 'tcp-stream-binary-segment nil nil)
  (CLEAR-RESOURCE 'tcp-stream-ascii-buffer nil nil)
  (COND ((NOT enable-p)
	 (SETF *tcp-handler* #'(lambda (operation &rest ignore)
				 (CASE operation
				   (:receive-data nil)
				   (otherwise
				    (FERROR 'tcp-error "TCP Service is not enabled."))))))
	(t (SETF *tcp-handler*
		 (MAKE-INSTANCE (IF (TYPEP *tcp-handler* 'tcp-handler)
				    *tcp-handler*
				    'tcp-handler)
				:background-process ()
				:background-event ()
				:probe-event (WHEN  *tcp-dormant-probe-interval*
					       (TIME-INCREMENT (TIME) *tcp-dormant-probe-interval*))
				:retransmit-list-event () :retransmit-list ()
				:delayed-ack-list-event () :delayed-ack-list ()
				:time-wait-list-event () :time-wait-list ()
				:syn-list ()
				:background-requests ()
				:tcb-table
				(MAKE-HASH-TABLE :compare-function #'EQ
						 :hash-function ()
						 :size 123)
				:port-table
				(MAKE-ARRAY *max-tcp-ports*
					    :element-type 'integer
					    :initial-element 0
					    :area *ip-area*)
				:segments-received 0
				:checksum-errors 0 :nonexistant-connections 0
				:ports-currently-reserved 0 :retransmissions 0 :delayed-acks 0
				:idle-probes 0 :syn-listings 0 :syn-timeouts 0 :syn-claims 0))
	   (DOLIST (server-entry *tcp-server-alist*)
	     (SEND *tcp-handler* :reserve-port (FIRST server-entry)))
	   (SETF (SEND *tcp-handler* :background-process)
		 (PROCESS-RUN-FUNCTION '(:name "TCP Background" :priority 15 :restart-after-reset t :restart-after-boot t)
				       *tcp-handler* :run-function))))
  "TCP Reset Complete")

))

#!C
; From file FTPINT.LISP#> IP; MR-X:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :Common-lisp)
                          (*READTABLE* Common-lisp-readtable)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* Sys::*common-lisp-symbol-substitutions*))
  (COMPILER#:PATCH-SOURCE-FILE "sys: IP; FTPINT.#"


(DEFUN reset-ftp-service (&optional enable-p)
  (WHEN (TYPEP *ftp-service* 'FTP)
    (WHEN *ftp-gc-process*
      (WITHOUT-INTERRUPTS
	(SEND *ftp-gc-process* :kill t)
	(SETF *ftp-gc-process* nil)))
    (DOLIST (h host:*host-list*)
      (WHEN (ftp-host-p h)
	(SEND *ftp-service* :reset :ftp h))))
  (COND ((NOT enable-p)
	 (SETF *ftp-service* #'(lambda (&rest ignore)
				 (FERROR 'ftp-error "FTP Service is not enabled."))))
	(t
	 (SETF *ftp-gc-process* (PROCESS-RUN-FUNCTION
				  '(:name "Dormant FTP Connection GC" :priority 0 :restart-after-boot t :restart-after-reset t) 'ftp-gc))
	 (SETF *ftp-service* (host:find-service-implementation :ftp))))
  "FTP Reset Complete")

))

#!C
; From file IP.LISP#> IP; MR-X:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :Common-lisp)
                          (*READTABLE* Sys:Common-lisp-readtable)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* Sys::*common-lisp-symbol-substitutions*))
  (COMPILER#:PATCH-SOURCE-FILE "Sys: IP; IP.#"


(DEFUN enable-ip-service (&optional (local-only t))
  "Sets up the protocol handler list for use by IP."
  (DECLARE (SPECIAL *udp-protocol-ip-id* *udp-handler* *tcp-protocol* *tcp-handler* icmp-protocol))
  (LET (network-broadcast-addr route-entry)
    (SETF *ip-handler* (MAKE-INSTANCE (IF (TYPEP *ip-handler* 'ip-handler)
					  *ip-handler*
					  'ip-handler)))
    (SETF *ip-pkt-sent-list* ())
    (SETF *ip-pkt-rvcd-list* ())
    (SETF *ip-id-counter* (MOD (time:microsecond-time) 65536))
    (WHEN (VARIABLE-BOUNDP *udp-handler*)
      (PUSH (CONS :udp '*udp-handler*) *ip-protocol-handler-alist*)
      (PUSH (CONS *udp-protocol-ip-id* '*udp-handler*) *ip-protocol-handler-alist*))
    (WHEN (VARIABLE-BOUNDP *tcp-handler*)
      (PUSH (CONS :tcp '*tcp-handler*) *ip-protocol-handler-alist*)
      (PUSH (CONS *tcp-protocol* '*tcp-handler*) *ip-protocol-handler-alist*))
    (WHEN (VARIABLE-BOUNDP *icmp-handler*)
      (PUSH (CONS icmp-protocol '*icmp-handler*) *ip-protocol-handler-alist*)
      (PUSH (CONS :icmp '*icmp-handler*) *ip-protocol-handler-alist*))
    (reset-other-machines)
    (SETF ethernet::ip-ether-address-translations ())
    (SETF ethernet::*ip-arp-requests-received* 0)
    (SETF ethernet::*ip-arp-replies-sent* 0)
    (SETF ethernet::*ip-arp-requests-sent* 0)
    (SETF ethernet::*ip-arp-replies-received* 0)
    (ethernet::add-protocol *ip-ethernet-type*
			    #'(lambda (controller ip-pkt nbytes type src-addr dest-addr)
				(SEND *ip-handler* :receive-data
				      controller ip-pkt nbytes type src-addr dest-addr)))
    (reset-ip-routing-table local-only)
    (SETF (SEND *ip-handler* :timer-process)
	  (PROCESS-RUN-FUNCTION
	    '(:name "IP Packet Fragment Timer" :priority -15 :restart-after-boot t :restart-after-reset t)
	    'ip-fragment-time-keeper))
    (enable)
    (DOLIST (addr my-addresses)
      (SETF route-entry (get-routing-entry addr))
      (SETF network-broadcast-addr
	    (LOGIOR (LOGAND addr (ip-routing-mask route-entry))
		    (LOGXOR #xFFFFFFFF (ip-routing-mask route-entry))))
      (DOTIMES (dummy 5)
	(IF *gateway-host-p*
	    (icmp-address-mask-transmit address-mask-reply 0 1 network-broadcast-addr)
	    (icmp-address-mask-transmit address-mask-request 0 1 network-broadcast-addr))
	(sleep .5 "Setup IP Routing")))))

))
