;;; -*- Mode: Common-Lisp; Package: FS; Base: 10.; Patch-File: T -*-

;;; Reason:  Fixed (FTP :case :directory-list-from-ftp :bsd4.2) to handle "not found" errors from SUN os 4.0.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 01/31/89 15:22:13 by BERGER,
;;; while running on ARIES from band LODX
;;; With Experimental SYSTEM 6.0, Experimental VIRTUAL-MEMORY 6.0, Experimental EH 6.0,
;;;  Experimental MAKE-SYSTEM 6.0, Experimental MICRONET 6.0, Experimental LOCAL-FILE 6.0,
;;;  Experimental BASIC-PATHNAME 6.0, Experimental NETWORK-SUPPORT-COLD 6.0, Experimental BASIC-NAMESPACE 6.0,
;;;  Experimental NETWORK-NAMESPACE 6.0, Experimental DISK-IO 6.0, Experimental DISK-LABEL 6.0,
;;;  Experimental BASIC-FILE 6.0, Experimental MAC-PATHNAME 6.0, Experimental NETWORK-PATHNAME 6.0,
;;;  Experimental COMPILER 6.0, Experimental TV 6.0, Experimental DATALINK 6.0, Experimental CHAOSNET 6.0,
;;;  Experimental GC 6.0, Experimental MEMORY-AUX 6.0, Experimental NVRAM 6.0, Experimental SYSLOG 6.0,
;;;  Experimental STREAMER-TAPE 6.0, Experimental CLEH 1.0, Experimental UCL 6.0,
;;;  Experimental INPUT-EDITOR 6.0, Experimental METER 6.0, Experimental ZWEI 6.0,
;;;  Experimental DEBUG-TOOLS 6.0, Experimental NETWORK-SUPPORT 6.0, Experimental NETWORK-SERVICE 6.0,
;;;  DATALINK-DISPLAYS 6.0, Experimental FONT-EDITOR 6.0, Experimental SERIAL 6.0,
;;;  Experimental PRINTER 6.0, Experimental MAC-PRINTER-TYPES 6.0, Experimental PRINTER-TYPES 6.0,
;;;  Experimental IMAGEN 6.0, Experimental SUGGESTIONS 6.0, Experimental MAIL-DAEMON 6.0,
;;;  Experimental MAIL-READER 6.0, Experimental TELNET 6.0, Experimental VT100 6.0,
;;;  Experimental NAMESPACE-EDITOR 6.0, Experimental PROFILE 6.0, VISIDOC 6.0, Experimental TI-CLOS 14.4,
;;;  IP 3.38, Experimental BUG 10.0,  microcode 413, Band Name: Rel 6.0 + IP 1/10

#!C
; From file FTPINT.LISP#> IP; SYS:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; FTPINT.#"


(DEFMETHOD (FTP :case :directory-list-from-ftp :bsd4.2) (pathname ftp stream fast?)
  (let (data-stream line entry entry-list in-subdirectory total-line-p open-name directory-name
	(sys:*new-line-delimiter* '(#\newline #\linefeed)) ;01-27-88 DAB
	 (error-parse-line-list ()) ;01-10-89 DAB Added to catch errors in parsing line from host.
	 ) 
    (declare (special sys:*new-line-delimiter* ))
    (unwind-protect
    (BLOCK bsd4.2
    (LOOP
      ;; open-name is equivalent to parameter pathname considering directory link translation
      ;; directory-name is the pathname actually passed across the control connection
      (SEND ftp :open
	    (PROG1 (SETF directory-name (dir-pathstring-for-ftp (SETF open-name (OR *ftp-directory-link* pathname))))
		   (SETF directory-name (parse-pathname directory-name nil pathname)))
	    ;; Unix nlist returns a null listing for both nonexistant and empty directories
	    ;; Don't use nlist to speed up the listing on a recursive entry, or empty directories will appear as nonexistant
	    :list)
      ;; workaround a Unix 425 bug which seems to go away on the second open attempt
      (UNLESS (AND (EQ 425 (SEND ftp :reply-code))
		   (SEARCH "Can't create data socket" (SEND ftp :reply-string))
		   (SEARCH "Address already in use" (SEND ftp :reply-string)))
	(RETURN)))
    (UNLESS (or (ftp-unfavorable (SEND ftp :reply-code))
		(not (SEND ftp :data-connection)))   ;03-03-88 DAB
      (SETF data-stream (SEND ftp :data-connection))
      ;; work around Unix bug of giving favorable status for a nonexistant directory
      (WHEN (or (NULL (SETF line (READ-LINE data-stream nil)))
		(search " not found" line))  ; DAB 01-31-89
	(COND
	  ;; if FTP list command on parent directory, then directory not found
	  ((OR *ftp-recursive-directory-list-from-ftp*
	       (NOT (pathname-equal directory-name open-name))
	       )  
	   (SETF (SEND ftp :reply-code) 450)
	   (SETF (SEND ftp :reply-string) (FORMAT nil "Directory not found for ~a" pathname))
	   (RETURN-FROM bsd4.2 nil))
	  ;; if FTP list command on root directory, then file not found 
	  ((EQL :root (SEND pathname :directory))
	   (RETURN-FROM bsd4.2 nil))
	  ;; must check existence of parent directory to make decision
	  (t (LET ((*ftp-recursive-directory-list-from-ftp* t))
	       (SEND ftp :close)
	       (SEND self :directory-list-from-ftp :bsd4.2
		     (SEND pathname :directory-pathname-as-file) ftp stream fast?)
	       ;; reply code is set by the recursive call
	       (RETURN-FROM bsd4.2 nil)))))
      ;; work around Unix bug of giving favorable status for a protected directory
      (WHEN (SEARCH (STRING-APPEND (SEND directory-name :string-for-host) " unreadable") line)
	(SETF (SEND ftp :reply-code) 450)
	(SETF (SEND ftp :reply-string) (FORMAT nil "~a: Permission denied" pathname))
	(RETURN-FROM bsd4.2 nil))
      ;; when recursively entered, we were only checking the existance of the parent directory, so exit.
      ;; (Note also that Unix will chase links with the NLST comand, so we don't have to worry about links.)
      (IF *ftp-recursive-directory-list-from-ftp*
	  (RETURN-FROM bsd4.2 nil)
	  (add-directory-stream stream data-stream))
      (condition-case (condition) ;01-10-89 DAB Bad characters in pathname, bad dates,etc , catch them, but keep parsing.
	  (MULTIPLE-VALUE-SETQ (entry in-subdirectory total-line-p) 
	    (bsd4.2-directory-line-parse pathname directory-name open-name line fast? in-subdirectory))
	(error (push line error-parse-line-list)))
      ;; They gave us /d1/d2.  Unix lists d2 if it is a directory.
      ;; We considered d2 as the file spec and thus don't match anything if d2 is a directory.
      ;; Assumption is based on the fact that we saw the total line first (e.g. total n) when only a single file was requested.
      (WHEN (AND total-line-p (pathname-equal directory-name open-name))
	(RETURN-FROM bsd4.2 (LIST (LIST pathname :directory t))))
      ;; Workaround Unix feature of not chasing the link path with the LIST command (and returning the link information instead),
      ;; when we were listing the parent directory.  We throw out with a single entry, which needs its link chased.
      (WHEN (AND (GETF (CDR entry) :link-to)
		 (NOT (pathname-equal directory-name open-name))
		 (pathname-equal (FIRST entry) directory-name))
	(THROW :directory-link-error (LIST entry)))
      (LOOP
	(WHEN entry (PUSH entry entry-list))
	(UNLESS (SETF line (READ-LINE data-stream nil))
	  (RETURN-FROM bsd4.2 entry-list))
	 (condition-case (condition) ;01-10-89 DAB Bad characters in pathname, bad dates,etc , catch them, but keep parsing.
	     (MULTIPLE-VALUE-SETQ (entry in-subdirectory total-line-p)
	       (bsd4.2-directory-line-parse pathname directory-name open-name line fast? in-subdirectory))
	   (error (push line error-parse-line-list)
		  )))))
  (when error-parse-line-list
    (tv:notify nil "Error parsing the following lines in directory ~a:~%~{ ~a~%~}~%Lines discarded."
	       pathname error-parse-line-list)))
  ))
))


