;;; -*- Mode: Common-Lisp; Package: IP; Base: 10.; Patch-File: T -*-

;;; Reason: Changed :next-input-buffer for character and binary streams to send-ack if (zerop pending-octet-count). Also modified open-stream to use a default number-of-input-buffers of 8.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 03/22/89 08:21:14 by BERGER,
;;; while running on ARIES from band LODX
;;; With Experimental REL6D 1.0, Experimental SYSTEM 6.0, Experimental VIRTUAL-MEMORY 6.0,
;;;  Experimental EH 6.0, Experimental MAKE-SYSTEM 6.0, Experimental MICRONET 6.0,
;;;  Experimental LOCAL-FILE 6.0, Experimental BASIC-PATHNAME 6.0, Experimental NETWORK-SUPPORT-COLD 6.0,
;;;  Experimental BASIC-NAMESPACE 6.0, Experimental NETWORK-NAMESPACE 6.0, Experimental DISK-IO 6.0,
;;;  Experimental DISK-LABEL 6.0, Experimental BASIC-FILE 6.0, Experimental MAC-PATHNAME 6.0,
;;;  Experimental NETWORK-PATHNAME 6.0, Experimental COMPILER 6.0, Experimental TV 6.0,
;;;  Experimental DATALINK 6.0, Experimental CHAOSNET 6.0, Experimental GC 6.0, Experimental MEMORY-AUX 6.0,
;;;  Experimental NVRAM 6.0, Experimental SYSLOG 6.0, Experimental STREAMER-TAPE 6.0,
;;;  Experimental CLEH 1.0, Experimental UCL 6.0, Experimental INPUT-EDITOR 6.0, Experimental METER 6.0,
;;;  Experimental ZWEI 6.0, Experimental DEBUG-TOOLS 6.0, Experimental NETWORK-SUPPORT 6.0,
;;;  Experimental NETWORK-SERVICE 6.0, Experimental DATALINK-DISPLAYS 6.0, Experimental FONT-EDITOR 6.0,
;;;  Experimental SERIAL 6.0, Experimental PRINTER 6.0, Experimental MAC-PRINTER-TYPES 6.0,
;;;  Experimental PRINTER-TYPES 6.0, Experimental IMAGEN 6.0, Experimental SUGGESTIONS 6.0,
;;;  Experimental MAIL-DAEMON 6.0, Experimental MAIL-READER 6.0, Experimental TELNET 6.0,
;;;  Experimental VT100 6.0, Experimental NAMESPACE-EDITOR 6.0, Experimental PROFILE 6.0,
;;;  Experimental VISIDOC 6.0, Experimental BUG 11.3, IP 3.41, Experimental TI-CLOS 14.32,
;;;  Experimental CLX 3.0, Experimental X11M 1.42, Experimental CLUE 16.0, Experimental RPC 6.0,
;;;  NFS 3.5,  microcode 418, Band Name: Rel 6.0 4S 3/13

#!C
; From file TCP-STREAM.LISP#> IP; SYS:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; TCP-STREAM.#"


(DEFMETHOD (character-input-stream-mixin :next-input-buffer) (no-hang-p
							      &aux segment beg end text-space octet-deficit first-rd)
  (BLOCK next-input-buffer
    (IF (AND fin-received-p (ZEROP pending-octet-count))
	(RETURN-FROM next-input-buffer ())
	(SEND self :load-receives))
    (WHEN (AND (ZEROP pending-octet-count) (NOT no-hang-p))
      (PROCESS-WAIT-WITH-TIMEOUT (OR *tcp-stream-whostate* "TCP Input")
				 (WHEN timeout (* timeout 60))
				 #'(lambda (stream)
				     (OR (EQ :closed (SEND stream :status))
					 (SEND stream :fin-received-p)
					 (PLUSP (SEND stream :pending-octet-count))))
				 self))
    (COND
      ;; force error that occurred in background to appear now
      (condition-handler-entered-p (SEND connection :abort))
      ((ZEROP pending-octet-count)
       (UNLESS (OR no-hang-p fin-received-p (EQ :closed (SEND self :status)))
	 (SEND self :timeout-handler)))
      (t
       (SETF segment (SEND connection :return-q))
       (SETF beg (segment-index segment))
       (SETF end (segment-size segment))
       ;; insure segment is covered by receives
       (WHEN (< pending-octet-count (SETF text-space (- end beg)))
	 (SEND self :load-receives (1- text-space)))
       (with-connection-lock
	 ;; insure segment is fully returned
	 (WHEN (PLUSP (SETF octet-deficit (- text-space pending-octet-count)))
	   (INCF pending-octet-count octet-deficit)
	   (DECF (rd-length (SETF first-rd (SEND connection :rd-q))) octet-deficit)
	   (WHEN (EQ :registered (rd-state first-rd))
	     (DECF (SEND connection :rd-registered-length) octet-deficit))
	   (SETF (SEND connection :return-sequence) (+seq (SEND connection :return-sequence) octet-deficit)))
	 (IF (SETF (SEND connection :return-q) (segment-link segment))
	     (SETF (segment-back-link (SEND connection :return-q)) ())
	     (SETF (SEND connection :return-q-end) ()))
	 (DECF pending-octet-count text-space)

	 ;; ; DAB 03-22-89 For PJ.
	 ;;  Idea: if there isn't any data ready, and we've already queued receives,
	 ;;  try sending an ack to prompt the other end to send us stuff (theory is
	 ;;  that the receive-window closed and we've since pulled off enough data to
	 ;;  open it up).
	 (when (and (eq (send connection :state) :established)
		    (zerop pending-octet-count)
;		    (> pending-receives 0)
		    )
	   (send connection :funcall-inside-yourself #'send-ack))

	 (WHEN (EQL (SETF (SEND connection :out-sequence) (+seq (SEND connection :out-sequence) text-space))
		    (SEND connection :return-sequence))
	   (CASE (SEND connection :state)
	     (:closed
	      (SEND *tcp-handler* :delete-tcb connection)
	      (SEND self :close-complete-handler))
	     (:time-wait 
	      (SEND *tcp-handler* :link-tcb-to-time-wait-list connection)
	      (SEND self :close-complete-handler))))
	 (SETF urgent-input-index (WHEN (tcp-header-urgent-p segment)
				    (MAX end (+ beg (tcp-header-urgent-pointer segment)))))
	 (VALUES segment beg end))))))

(DEFMETHOD (binary-input-stream-mixin :next-input-buffer) (no-hang-p
							   &aux segment beg end text-space octet-deficit first-rd)
  (BLOCK next-input-buffer
    (IF (AND fin-received-p (ZEROP pending-octet-count))
	(RETURN-FROM next-input-buffer ())
	(SEND self :load-receives))
    (WHEN (AND (ZEROP pending-octet-count) (NOT no-hang-p))
      (PROCESS-WAIT-WITH-TIMEOUT (OR *tcp-stream-whostate* "TCP Input")
				 (WHEN timeout (* timeout 60))
				 #'(lambda (stream)
				     (OR (EQ :closed (SEND stream :status))
					 (SEND stream :fin-received-p)
					 (PLUSP (SEND stream :pending-octet-count))))
				 self))
    (COND
      ;; force error that occurred in background to appear now
      (condition-handler-entered-p (SEND connection :abort))
      ((ZEROP pending-octet-count)
       (UNLESS (OR no-hang-p fin-received-p (EQ :closed (SEND self :status)))
	 (SEND self :timeout-handler)))
      (t
       (SETF segment (SEND connection :return-q))
       ;; insure segment is covered by receives
       (WHEN (< pending-octet-count (SETF text-space (- (segment-size segment) (segment-index segment))))
	 (SEND self :load-receives (1- text-space)))
       (with-connection-lock
	 ;; insure segment is fully returned
	 (WHEN (PLUSP (SETF octet-deficit (- text-space pending-octet-count)))
	   (INCF pending-octet-count octet-deficit)
	   (DECF (rd-length (SETF first-rd (SEND connection :rd-q))) octet-deficit)
	   (WHEN (EQ :registered (rd-state first-rd))
	     (DECF (SEND connection :rd-registered-length) octet-deficit))
	   (SETF (SEND connection :return-sequence) (+seq (SEND connection :return-sequence) octet-deficit)))
	 (IF (SETF (SEND connection :return-q) (segment-link segment))
	     (SETF (segment-back-link (SEND connection :return-q)) ())
	     (SETF (SEND connection :return-q-end) ()))
	 ;; permute segment into even number of text octets, on even octet boundary (inserting saved-octet)
	 (WHEN saved-octet
	   (COPY-ARRAY-PORTION segment
			       (segment-index segment)
			       (segment-size segment)
			       segment
			       (1- (segment-index segment))
			       (1- (segment-size segment)))
	   (DECF (segment-size segment))
	   (SETF (segment-sequence-# segment) (+seq (segment-sequence-# segment) -1))
	   (DECF (segment-index segment) 2)
	   (SETF (AREF segment (segment-index segment)) saved-octet)
	   (SETF saved-octet ()))
	 (SETF beg (segment-index segment))
	 (SETF end (segment-size segment))
	 (WHEN (ODDP (- end beg))
	   (DECF end)
	   (DECF (segment-size segment))
	   (SETF saved-octet (AREF segment (segment-size segment))))
	 (DECF pending-octet-count text-space)
	 ;; ; DAB 03-22-89 for PF.
	 ;;  Idea: if there isn't any data ready, and we've already queued receives,
	 ;;  try sending an ack to prompt the other end to send us stuff (theory is
	 ;;  that the receive-window closed and we've since pulled off enough data to
	 ;;  open it up).
	 (when (and (eq (send connection :state) :established)
		    (zerop pending-octet-count)
;		    (> pending-receives 0)
		    )
	   (send connection :funcall-inside-yourself #'send-ack))

	 (WHEN (EQL (SETF (SEND connection :out-sequence) (+seq (SEND connection :out-sequence) text-space))
		    (SEND connection :return-sequence))
	   (CASE (SEND connection :state)
	     (:closed
	      (SEND *tcp-handler* :delete-tcb connection)
	      (SEND self :close-complete-handler))
	     (:time-wait
	      (SEND *tcp-handler* :link-tcb-to-time-wait-list connection)
	      (SEND self :close-complete-handler))))
	 (SETF urgent-input-index (WHEN (tcp-header-urgent-p segment)
				    (TRUNCATE (MAX end (+ beg (tcp-header-urgent-pointer segment)))
					      2)))
	 (VALUES
	   (allocate-tcp-stream-binary-segment segment beg end)
	   0
	   (TRUNCATE (- end beg) 2)))))))

(DEFUN open-stream (host &key &optional (local-port 0) (remote-port 0) (direction :bidirectional)
		    (characters t) (wait-for-establishment t) (timeout 10) (timeout-after-open (WHEN timeout 300)) 
		    (error t) (input-buffer-size *default-tcp-stream-input-buffer-size*) (number-of-input-buffers 8)
		    &aux (host-name "Unknown") connection mode destination-address stream normal-exit
		    default-timeout-handler timeout-occurred-p)
  "Establish a TCP connection and return a stream object for accessing that connection.
HOST The host to connect to.  May be a name, IP address, or host object.  If nil, the open is passive (listen).
:LOCAL-PORT (Optional) The local TCP port.  If 0 (default), a non-well known port will be generated and used.
:REMOTE-PORT (Optional) The remote TCP port on the specified host.  May be 0 (default) for passive opens.
:DIRECTION (Optional) :Input, :output, or :bidirectional (default).
:CHARACTERS (Optional) Either t (characters, the default), nil (binary), or :ascii (ascii characters).
:WAIT-FOR-ESTABLISHMENT (Optional) If nil, don't wait for connection establishment (default t).
:TIMEOUT  (Optional) The timeout (in seconds) used in open-stream when waiting for establishment.  
  Nil indicates no timeout.  Default is 10 seconds.
:TIMEOUT-AFTER-OPEN (Optional) The timeout used after the connection has been established.  Nil indicates no timeout.  
  The default for user-timeout is 300 seconds when :TIMEOUT is non-nil, and nil when :TIMEOUT is nil.  
  This timeout is used for the TCP 'user timeout' (the timeout on acknowledgement of data sent), and also used as 
  the timeout on :next-input-buffer.  This timeout can be changed after open-stream by sending the stream a :set-timeout message.
:ERROR (Optional) If t (default), error conditions will be signaled in open-stream.  Otherwise, the error condition will be returned.
:INPUT-BUFFER-SIZE (Optional) The size of the receive window in characters (default 2048).  For binary streams
  each character occupies two octets.
:NUMBER-OF-INPUT-BUFFERS (Optional) The number of input buffers supplied to TCP (default 8)."
  (FLET
    ;; internal function
    ((make-stream (connection direction characters timeout input-buffer-size number-of-input-buffers
			      &aux stream)
		  ;; ftp provides its own stream instantiator
		  (LET ((*tcp-stream-instantiator*
			  (OR *tcp-stream-instantiator*
			      #'(lambda (connection timeout input-buffer-size number-of-input-buffers)
				  (CASE direction
				    (:input
				     (MAKE-INSTANCE
				       (COND
					 ((EQ characters :ascii) 'ascii-translating-input-character-stream)
					 (characters 'input-character-stream)
					 (t 'input-binary-stream))
				       :connection connection :timeout timeout :input-buffer-size input-buffer-size
				       :number-of-input-buffers number-of-input-buffers))
				    (:output
				     (MAKE-INSTANCE
				       (COND
					 ((EQ characters :ascii) 'ascii-translating-output-character-stream)
					 (characters 'output-character-stream)
					 (t 'output-binary-stream))
				       :connection connection :timeout timeout))
				    (:bidirectional
				     (MAKE-INSTANCE
				       (COND
					 ((EQ characters :ascii) 'ascii-translating-character-stream)
					 (characters 'character-stream)
					 (t 'binary-stream))
				       :connection connection :timeout timeout :input-buffer-size input-buffer-size
				       :number-of-input-buffers number-of-input-buffers)))))))
		    (SETF stream
			  (FUNCALL *tcp-stream-instantiator* connection timeout input-buffer-size number-of-input-buffers)))
		  ;; insert actual handlers now that stream is known
		  (SETF (SEND connection :buffer-handler) #'(lambda (LENGTH push-p urgent-p)
							      (SEND stream :buffer-handler length push-p urgent-p)))
		  (SETF (SEND connection :urgent-handler) #'(lambda ()
							      (SEND stream :urgent-handler)))
		  (SETF (SEND connection :process-fin-handler) #'(lambda ()
								   (SEND stream :process-fin-handler)))
		  (SETF (SEND connection :receive-fin-handler) #'(lambda ()
								   (SEND stream :receive-fin-handler)))
		  (SETF (SEND connection :close-complete-handler) #'(lambda ()
								      (SEND stream :close-complete-handler)))
		  (SETF (SEND connection :user-timeout-handler) #'(lambda ()
								    (SEND stream :timeout-handler :user-timeout)))
		  (SETF (SEND connection :condition-handler) #'(lambda (condition-object)
								 (SEND stream :condition-handler condition-object)))
		  stream))
    ;; body of open-stream
    (CONDITION-CASE-IF (NOT error) (error-object)
	(PROGN
	 (COND (host
		(SETF mode :active)
		(SETF host (parse-ip-host-spec host))
		(SETF host-name (SEND host :short-name))	      
		(SETF destination-address (FIRST (closest-addresses-to-network (get-ip-addresses host)))))
	       (t (SETF mode :passive)
		  (SETF destination-address 0)))
	 (SETF connection (SEND *tcp-handler* :make-connection nil nil nil nil nil nil nil))
	 (SETF stream (make-stream connection direction characters timeout input-buffer-size number-of-input-buffers))
	 ;; use special handler until connection is established
	 (SETF default-timeout-handler (SEND connection :user-timeout-handler))
	 (SETF (SEND connection :user-timeout-handler) #'(lambda ()
							   (SETF timeout-occurred-p t)))
	 (UNWIND-PROTECT
	     (PROGN
	       (SEND connection :open local-port remote-port destination-address mode timeout)
	       (SEND stream :load-receives)
	       (IF (NOT wait-for-establishment)
		   (SETF normal-exit t)
		   (DO ()
		       ((CATCH-ERROR-RESTART-EXPLICIT-IF error
							 (host-not-responding-during-connection
							   :retry-connection "Try the connection again.")
			  (SETF timeout-occurred-p nil)
			  (WHEN (NOT (PROCESS-WAIT-WITH-TIMEOUT
				       (COND
					 (*tcp-stream-whostate*)
					 ((NULL host) "TCP Listen")
					 (t
					  (FORMAT () "TCP Connect: ~A" host-name)))
				       ;; when waiting for active establishment we depend on the tcp user timeout
				       (IF (OR (EQ mode :active) (NULL timeout))
					   ()
					   (* timeout 60))
				       #'(lambda (connection)
					   (CASE (SEND connection :state)
					     ((:listen :syn-sent :syn-received) timeout-occurred-p)
					     (otherwise t)))
				       connection))
			    ;; when timing out passive establishment we must explicitly call the user-timeout-handler
			    (FUNCALL (SEND connection :user-timeout-handler)))
			  (IF timeout-occurred-p
			      (FERROR 'host-not-responding-during-connection "Host ~*~A not responding."
				      connection host-name)
			      t))
			(SETF normal-exit t))
		     (WHEN (AND timeout (<= timeout 60))
		       (SETF timeout 60)
		       (SETF (SEND connection :user-timeout) timeout)))))
	   (WHEN (NOT (OR normal-exit (EQ (SEND connection :status) :closed)))
	     (SEND connection :abort)))
	 ;; restore user timeout to be used with established connection
	 (SETF (SEND connection :user-timeout-handler) default-timeout-handler)
	 (SETF (SEND stream :timeout) timeout-after-open))
      (ERROR error-object)
      (:no-error stream))))

))
