;;; -*- Mode: Common-Lisp; Package: FS; Base: 10.; Patch-File: T -*-

;;; Reason:  Fixed lispm-directory-line-parse. It was not determining the byte size properly for XLD files. This fixes patch 3-38, which case XLD files to be copied as 8 byte character files. 

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 04/21/89 10:15:11 by BERGER,
;;; while running on ARIES from band LOD1
;;; With Experimental REL6G 6.0, Experimental SYSTEM 6.0, Experimental VIRTUAL-MEMORY 6.0,
;;;  Experimental EH 6.0, Experimental MAKE-SYSTEM 6.0, Experimental MICRONET 6.0,
;;;  Experimental LOCAL-FILE 6.0, Experimental BASIC-PATHNAME 6.0, Experimental NETWORK-SUPPORT-COLD 6.0,
;;;  Experimental BASIC-NAMESPACE 6.0, Experimental NETWORK-NAMESPACE 6.0, Experimental DISK-IO 6.0,
;;;  Experimental DISK-LABEL 6.0, Experimental BASIC-FILE 6.0, Experimental MAC-PATHNAME 6.0,
;;;  Experimental NETWORK-PATHNAME 6.0, Experimental COMPILER 6.0, Experimental TV 6.0,
;;;  Experimental DATALINK 6.0, Experimental CHAOSNET 6.0, Experimental GC 6.0, Experimental MEMORY-AUX 6.0,
;;;  Experimental NVRAM 6.0, Experimental SYSLOG 6.0, Experimental STREAMER-TAPE 6.0,
;;;  Experimental CLEH 1.0, Experimental UCL 6.0, Experimental INPUT-EDITOR 6.0, Experimental METER 6.0,
;;;  Experimental ZWEI 6.0, Experimental DEBUG-TOOLS 6.0, Experimental NETWORK-SUPPORT 6.0,
;;;  Experimental NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0, Experimental FONT-EDITOR 6.0,
;;;  Experimental SERIAL 6.0, Experimental PRINTER 6.0, Experimental MAC-PRINTER-TYPES 6.0,
;;;  Experimental PRINTER-TYPES 6.0, Experimental IMAGEN 6.0, Experimental SUGGESTIONS 6.0,
;;;  Experimental MAIL-DAEMON 6.0, Experimental MAIL-READER 6.0, Experimental TELNET 6.0,
;;;  Experimental VT100 6.0, Experimental NAMESPACE-EDITOR 6.0, Experimental PROFILE 6.0,
;;;  VISIDOC 6.0, Experimental TI-CLOS 16.4, Experimental CLX 4.0, Experimental X11M 2.0,
;;;  Experimental CLUE 19.0, Experimental RPC 6.0, NFS 3.5, Experimental BUG 11.4,
;;;  IP 3.42, DECNET 1.64,  microcode 423, Band Name: 6.0 4/19 SLE,DNA,NP2

#!C
; From file FTPINT.LISP#> IP; SYS:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; FTPINT.#"


(DEFUN lispm-directory-line-parse (dirobj line fast? &aux pathobj cursor cursor2 result month day year hours minutes seconds
				   length-in-blocks length-in-bytes byte-size)
  ;;; FORMAT:
  ;;; {d}  <pathname>  #-blocks  #-bytes(byte-size)  {!}{@}{$}{@} mm/dd//yy hh:mm:ss {author}

  ;;;                           ""    DIRECTORY         ""
  (WHEN (CHAR-EQUAL (AREF line 0) #\d)
    (PUSH t result)
    (PUSH :deleted result))
  ;; Careful - there may be multiple blanks embedded in the pathname and pathname parsing doesn't handle :Junk-allowed very well
  (SETQ cursor (POSITION #\Space (THE string (STRING line)) :start 1 :test-not 'CHAR-EQUAL))
  (SETQ cursor2
	(POSITION #\Space (THE string (STRING line)) :start
		  (POSITION #\# (THE string (STRING line)) :start cursor :test 'CHAR-EQUAL)
		  :test 'CHAR-EQUAL))
;; Make the pathname to merge with instead of parsing it. *BJ*
  (SETQ pathobj (MERGE-PATHNAMES
		  (parse-pathname (SUBSEQ line cursor cursor2) nil dirobj)
		  (fs:make-PATHNAME :host (send dirobj :host)
				    :type :unspecific)))
  (UNLESS fast?
    (MULTIPLE-VALUE-SETQ (length-in-blocks cursor)
			 (PARSE-INTEGER line :start cursor2 :junk-allowed t))
    (PUSH length-in-blocks result)
    (PUSH :length-in-blocks result)
    (IF (MULTIPLE-VALUE-SETQ (length-in-bytes cursor)
			     (PARSE-INTEGER line :start cursor :junk-allowed t))
	(PROGN
	  (MULTIPLE-VALUE-SETQ (BYTE-SIZE cursor)
			       (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
	  (PUSH byte-size result)
	  (PUSH :byte-size result)
	  (PUSH length-in-bytes result)
	  (PUSH :length-in-bytes result)
	  (push (DETERMINE-CHARACTERS-AND-BYTE-SIZE result pathobj pathobj nil 8 nil) result)   ; DAB 04-21-89
	  (push :characters result)                                                               ;04-12-88 DAB
	  )
	;; ELSE
	(PROGN
	  (PUSH t result)
	  (PUSH :directory result)
	  (INCF cursor (LENGTH "DIRECTOR"))))	;leave at "Y" (will be incremented)
    (LOOP
      (SETQ cursor
	    (POSITION #\Space (THE string (STRING line)) :start (1+ cursor) :test-not
		      'CHAR-EQUAL))
      (SELECTOR (AREF line cursor) char-equal (#\! (PUSH t result) (PUSH :not-backed-up result))
		(#\# (PUSH t result) (PUSH :dont-supersede result))
		(#\$ (PUSH t result) (PUSH :dont-reap result))
		(#\@ (PUSH t result) (PUSH :dont-delete result)) (:otherwise (RETURN))))
    (MULTIPLE-VALUE-SETQ (month cursor)
			 (PARSE-INTEGER line :start cursor :junk-allowed t))
    (MULTIPLE-VALUE-SETQ (day cursor)
			 (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
    (MULTIPLE-VALUE-SETQ (year cursor)
			 (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
    (MULTIPLE-VALUE-SETQ (hours cursor)
			 (PARSE-INTEGER line :start cursor :junk-allowed t)) ;04-28-88 DAB parse-integer ate the space
    (MULTIPLE-VALUE-SETQ (minutes cursor)
			 (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
    (MULTIPLE-VALUE-SETQ (seconds cursor)
			 (PARSE-INTEGER line :start (1+ cursor) :junk-allowed t))
    (PUSH (ENCODE-UNIVERSAL-TIME seconds minutes hours day month year) result)
    (PUSH :creation-date result)
    (IF (SETQ cursor
	      (POSITION #\Space (THE string (STRING line)) :start cursor :test-not 'CHAR-EQUAL))
	(PROGN
	  (SETQ cursor2
		(POSITION #\Space (THE string (STRING line)) :start cursor :test 'CHAR-EQUAL))
	  (PUSH (SUBSEQ line cursor cursor2) result))
	;; else
	(PUSH (CAR (SEND dirobj :directory)) result))
    (PUSH :author result))
  (PUSH pathobj result))
))
