;;; -*- Mode: Common-Lisp; Package: Ip; Base: 10.; Patch-File: T -*-

;;; Reason:  Added locking to :get-port so that the same port number is not allocated to two different processes.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 04/24/89 07:50:03 by BERGER,
;;; while running on ARIES from band LOD1
;;; With Experimental REL6G 6.0, Experimental SYSTEM 6.0, Experimental VIRTUAL-MEMORY 6.0,
;;;  Experimental EH 6.0, Experimental MAKE-SYSTEM 6.0, Experimental MICRONET 6.0,
;;;  Experimental LOCAL-FILE 6.0, Experimental BASIC-PATHNAME 6.0, Experimental NETWORK-SUPPORT-COLD 6.0,
;;;  Experimental BASIC-NAMESPACE 6.0, Experimental NETWORK-NAMESPACE 6.0, Experimental DISK-IO 6.0,
;;;  Experimental DISK-LABEL 6.0, Experimental BASIC-FILE 6.0, Experimental MAC-PATHNAME 6.0,
;;;  Experimental NETWORK-PATHNAME 6.0, Experimental COMPILER 6.0, Experimental TV 6.0,
;;;  Experimental DATALINK 6.0, Experimental CHAOSNET 6.0, Experimental GC 6.0, Experimental MEMORY-AUX 6.0,
;;;  Experimental NVRAM 6.0, Experimental SYSLOG 6.0, Experimental STREAMER-TAPE 6.0,
;;;  Experimental CLEH 1.0, Experimental UCL 6.0, Experimental INPUT-EDITOR 6.0, Experimental METER 6.0,
;;;  Experimental ZWEI 6.0, Experimental DEBUG-TOOLS 6.0, Experimental NETWORK-SUPPORT 6.0,
;;;  Experimental NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0, Experimental FONT-EDITOR 6.0,
;;;  Experimental SERIAL 6.0, Experimental PRINTER 6.0, Experimental MAC-PRINTER-TYPES 6.0,
;;;  Experimental PRINTER-TYPES 6.0, Experimental IMAGEN 6.0, Experimental SUGGESTIONS 6.0,
;;;  Experimental MAIL-DAEMON 6.0, Experimental MAIL-READER 6.0, Experimental TELNET 6.0,
;;;  Experimental VT100 6.0, Experimental NAMESPACE-EDITOR 6.0, Experimental PROFILE 6.0,
;;;  VISIDOC 6.0, Experimental TI-CLOS 16.5, Experimental CLX 4.0, Experimental X11M 2.0,
;;;  Experimental CLUE 19.0, Experimental RPC 6.0, NFS 3.5, Experimental BUG 11.4,
;;;  IP 3.43, DECNET 1.64,  microcode 423, Band Name: 6.0 4/19 SLE,DNA,NP2

#!C
; From file UDPDEFS.LISP#> IP; SYS:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; UDPDEFS.#"


(defvar *udp-get-port-lock* nil
     "Lock used in :GET-PORT to make sure the same port number isn't allocated
      to more than one process.")
))

#!C
; From file UDPHAND.LISP#> IP; SYS:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; UDPHAND.#"
(DEFMETHOD (udp-handler :get-port) (&optional (port-number 0)
				    (source-address 0)
				    (destination-port 0)
				    (destination-address 0)
				    &aux port)
  (LET* ((default-cons-area *ip-area*))
    ;; If the port number is not specified, generate one.  Otherwise, be sure the specified
    ;; one is unique and will fit in two bytes (less than #X10000).
    ;;
    ;;Add lock to make sure we don't allocate a port to more   ; DAB 04-25-89
    ;;than one process
    (with-lock (*udp-get-port-lock*)
      (IF (ZEROP port-number)
	  (SETF port-number (udp-get-port-number))
	  (IF (valid-udp-port-p port-number)
	      (UNLESS (udp-unique-port-number-p port-number)
		(FERROR 'port-in-use "The specified source port, ~a, is in use" port-number))
	      (FERROR 'port-is-invalid "The specified source port, ~a, is not a valid UDP port"
		      port-number)))
      (SETF port (ALLOCATE-RESOURCE 'udp-ports))
      (SETF (SEND port :port-number) port-number
	    (SEND port :who-state) "UDP Input"
	    (SEND port :sender-port) destination-port
	    (SEND port :sender-address) destination-address
	    (SEND port :receiver-address) (IF (ZEROP source-address)
					      (closest-local-address destination-address)
					      source-address))
      (WITHOUT-INTERRUPTS
	;; claim any waiting packets for server
	(DO ((p server-packet-list (CDR p))
	     (q nil p))
	    ((NULL p))
	  (WHEN (EQL (udp-header-destination-port (FIRST p)) port-number)
	    ;; Keep the first incoming packet, not the last  ; DAB 04-25-89
	    (if (send port :received-packet)
		;; Not the first packet, so drop it correctly  ; DAB 04-25-89
		(progn
		  (deallocate-resource 'udp-packet (first p))
		  ;;Increment the overrun count
		  (incf (udp-port-receive-overruns port))
		  )
		;; This is the first packet, so queue it   ; DAB 04-25-89
		(progn
		  (SETF (SEND port :received-packet) (FIRST p))
		  (SETF (SEND port :last-buffered-time) (time:get-universal-time))
		  (SETF (SEND port :sender-port) (udp-header-source-port (FIRST p)))
		  (SETF (SEND port :sender-address) (udp-packet-source-address (FIRST p)))
		  )
		)
	    (IF q
		(RPLACD q (CDR p))
		(SETF server-packet-list (CDR p)))
	    (INCF server-packet-claims)))
	(PUSH (CONS port-number port) *udp-receive-list*))
      )
    (VALUES port)))



))
