;;; -*- Mode: Common-Lisp; Package: IP; Base: 10.; Patch-File: T -*-

;;; Reason: Modified update-send-unack to retransmit segments that a remote host returned a partial ack of the segment. This problem was causing very slow file transfer rates to some SUN machines.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 09/26/89 12:25:37 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 6.15, VIRTUAL-MEMORY 6.1, EH 6.4, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.2, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.10, TV 6.15, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.5,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.4, TI-CLOS 6.20, CLEH 6.5, IP 3.48,
;;;  Experimental CLX 6.2, CLUE 6.10, X11M 6.13, Experimental BUG 11.15, VISIDOC-SERVER 6.1,
;;;   microcode 429, Band Name: Rel 6.0 + SLE 8/30

#!C
; From file TCPCONN.LISP#> IP; SYS:
#10R IP#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "IP"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; TCPCONN.#"


(DEFUN update-send-unack (ack-number &aux rtt)
  "Update send-unack, remove acked segments, and update the retransmission timeout (RFC793, p41)."
  (DECLARE (:self-flavor tcp-connection))
  (SETF send-unack ack-number)
  (DO ((p retransmit-q retransmit-q))
      ((OR (NULL p)
	   (<seq send-unack (+seq (segment-sequence-# p) (segment-sequence-length p))))
       ;;On exit check for partial segment ack. That is, The remote host only accept a portion of the 
       ;;max-segment-size. IN this case we should retransmit this segment right away.  DAB 09-26-89
       (when (and p (>seq send-unack (segment-sequence-# p)))	   ;partial segment ack
	 ;;the retransmission-timeout is the timestamp used by process-retransmission-queue to 
         ;;determine if the segments needs retransmitting.
	 (setf (segment-retransmission-timeout p) (time))
	 (WHEN (NULL rtt)
	   (SETF rtt (TIME-DIFFERENCE (TIME) (segment-time p))))
	 ))
    (WHEN (NULL rtt)
      (SETF rtt (TIME-DIFFERENCE (TIME) (segment-time p))))
    (DEALLOCATE-RESOURCE 'tcp-segment (dequeue retransmit-q retransmit-q-end))
    )
  ;; a non-nil rtt indicates that at least one segment was removed from the retransmit-q
  (WHEN rtt
    (SEND *tcp-handler* :link-tcb-to-retransmit-list self
	  (WHEN retransmit-q (segment-retransmission-timeout retransmit-q)))
    (SETF smoothed-round-trip-time
	  (IF (NULL smoothed-round-trip-time)
	      rtt
	      (ROUND (+ (* *alpha* smoothed-round-trip-time) (* (- 1 *alpha*) rtt)))))
    (SETF retransmission-timeout
	  (MIN *ubound* (MAX *lbound* (* *beta* smoothed-round-trip-time))))))
))
