;;; -*- Mode: Common-Lisp; Package: FS; Base: 10.; Patch-File: T -*-
;;; Patch file for IP version 3.8
;;; Reason: Corrected several mispelling of PRIVILEGE. Therefore access to secure file would generate a
;;; FILE NOT FOUND error instead of a file security access error.
;;; Written 04/07/88 07:44:12 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 4.48, Experimental VIRTUAL-MEMORY 4.0, Experimental EH 4.0, Experimental MAKE-SYSTEM 4.4,
;;;  Experimental MICRONET 4.3, Experimental LOCAL-FILE 4.0, BASIC-PATHNAME 4.8, Experimental DISK-IO 4.2,
;;;  Experimental DISK-LABEL 4.0, BASIC-FILE 4.5, Experimental MAC-PATHNAME 4.3, Experimental NETWORK-PATHNAME 4.0,
;;;  Experimental COMPILER 4.1, NETWORK-NAMESPACE 4.3, TV 4.60, DATALINK 4.13, Experimental CHAOSNET 4.5,
;;;  Experimental GC 4.3, Experimental MEMORY-AUX 4.0, Experimental NVRAM 4.3, Experimental SYSLOG 4.0,
;;;  Experimental STREAMER-TAPE 4.1, Experimental UCL 4.1, Experimental INPUT-EDITOR 4.0,
;;;  Experimental METER 4.3, Experimental ZWEI 4.16, Experimental DEBUG-TOOLS 4.2,
;;;  Experimental NETWORK-SUPPORT 4.2, Experimental NETWORK-SERVICE 4.0, Experimental DATALINK-DISPLAYS 4.0,
;;;  Experimental FONT-EDITOR 4.0, Experimental SERIAL 4.0, Experimental PRINTER 4.6,
;;;  Experimental PRINTER-TYPES 4.0, Experimental IMAGEN 4.0, Experimental SUGGESTIONS 4.0,
;;;  Experimental MAIL-DAEMON 4.3, Experimental MAIL-READER 4.5, Experimental TELNET 4.0,
;;;  Experimental VT100 4.4, Experimental NAMESPACE-EDITOR 4.4, PROFILE 4.4, VISIDOC 4.4,
;;;  NAMESPACE 4.14, Experimental NETWORK-SUPPORT-COLD 4.1, Experimental IP 3.6, Experimental BUG 10.0,
;;;   microcode 522, Band Name: pre 4.0 + IP 3/24

#!C
; From file FTPINT.LISP#> IP; SYS:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: IP; FTPINT.#"


(DEFMETHOD (FTP :rename) (medium pathname new-pathname &optional (error-p t) &aux ftp finished-p ftp-result)
  (ip:with-stream-whostate
    "Rename"
    (file-operation-retry
      (UNWIND-PROTECT
	  (PROGN
	    ;; insure the destination directory exists (allow directory creation with condition handlers)
	    (SETF ftp-result (SEND self :directory-list medium new-pathname))
	    ;; with ftp the rename operation takes two steps
	    (UNLESS (ERRORP ftp-result)
	      (SETQ ftp (reserve-ftp-connection (SEND pathname :host) :rename)) ;; last arg is nicety for peek
	      (SEND ftp :quote (FORMAT () "RNFR ~a" (pathstring-for-ftp pathname)))
	      (SETF ftp-result (COND ((OR (EQL 450 (SEND ftp :reply-code))
					  (EQL 550 (SEND ftp :reply-code)))
				      (LET ((reply-string (SEND ftp :reply-string))
					    directory-list)
					(free-ftp-connection (PROG1 ftp (SETF ftp nil)) nil :abort-data)
					(COND ((OR (SEARCH "ermission" reply-string)
						   (SEARCH "ERMISSION" reply-string)
						   (SEARCH "rivilege" reply-string)   ;03-22-88 DAB mispelled privilege
						   (SEARCH "RIVILEGE" reply-string))  ;03-22-88 DAB mispelled privilege
					       (MAKE-CONDITION 'incorrect-access-to-file reply-string pathname :rename))
					      (t
					       ;; directory-list will supply the directory-not-found condition object
					       (SETF directory-list (SEND *ftp-service* :directory-list
									  :ftp pathname '(:noerror)))
					       (IF (ERRORP directory-list)
						   directory-list
						   (MAKE-CONDITION 'fs:file-not-found "File not found for ~a"
								   pathname :rename))))))
				     ((EQL 350 (SEND ftp :reply-code)) 350)
				     (t (check-ftp-status pathname ftp nil))))
	      (UNLESS (ERRORP ftp-result)
		(SEND ftp :quote (FORMAT () "RNTO ~a" (pathstring-for-ftp new-pathname)))
		;; the spec does not allow 450 or 550 but Unix may return them
		(SETF ftp-result
		      (COND ((OR (EQL 450 (SEND ftp :reply-code))
				 (EQL 550 (SEND ftp :reply-code)))
			     (LET ((reply-string (SEND ftp :reply-string))
				   directory-list)
			       (free-ftp-connection (PROG1 ftp (SETF ftp nil)) nil :abort-data)
			       (COND ((OR (SEARCH "ermission" reply-string) (SEARCH "ERMISSION" reply-string)
					  (SEARCH "rivilege" reply-string)    ;03-22-88 DAB mispelled privilege
					  (SEARCH "RIVILEGE" reply-string))   ;03-22-88 DAB mispelled privilege
				      (MAKE-CONDITION 'incorrect-access-to-file reply-string new-pathname :rename))
				     (t
				      ;; directory-list will supply the directory-not-found condition object
				      (SETF directory-list (SEND *ftp-service* :directory-list
								 :ftp new-pathname '(:noerror)))
				      (IF (ERRORP directory-list)
					  directory-list
					  (MAKE-CONDITION 'ftp-open-error reply-string new-pathname :rename))))))
			    (t (check-ftp-status pathname ftp nil)))))
	      (SETF finished-p t)))
	;; cleanup
	(WHEN ftp (free-ftp-connection ftp (OR (TYPEP ftp-result 'condition) (NOT finished-p)))))
      (IF (ERRORP ftp-result)
	  ;; THEN return error condition 
	  (IF error-p (SIGNAL-CONDITION ftp-result) ftp-result)
	  ;; ELSE return new-truename, T (or some number for chaos version) (second value is ignored I think)
	  (VALUES (SEND self :truename medium new-pathname) t)))))

(DEFMETHOD (FTP :delete-and-expunge) (medium pathname &optional (error-p t) &aux ftp finished-p ftp-result)
  (IGNORE medium)
  (ip:with-stream-whostate
    "Delete-and-expunge"
    (file-operation-retry
      (UNWIND-PROTECT
	  (PROGN
	    (SETQ ftp (reserve-ftp-connection (SEND pathname :host) :delete-and-expunge)) ;; last arg is nicety for peek
	    (SEND ftp :delete (pathstring-for-ftp pathname))
	    (SETF ftp-result (COND ((OR (EQL 450 (SEND ftp :reply-code))
					(EQL 550 (SEND ftp :reply-code)))
				    (LET ((reply-string (SEND ftp :reply-string))
					  directory-list)
				      (free-ftp-connection (PROG1 ftp (SETF ftp nil)) nil :abort-data)
				      (COND ((OR (SEARCH "ermission" reply-string) (SEARCH "ERMISSION" reply-string)
						 (SEARCH "rivilege" reply-string)  ;03-22-88 Mispelled privilege
						 (SEARCH "RIVILEGE" reply-string)) ;03-22-88 Mispelled privilege
					     (MAKE-CONDITION 'incorrect-access-to-file
							     reply-string pathname :delete-and-expunge))
					    (t
					     ;; directory-list will supply the directory-not-found condition object
					     (SETF directory-list (SEND *ftp-service* :directory-list
									:ftp pathname '(:noerror)))
					     (IF (ERRORP directory-list)
						 directory-list
						 (MAKE-CONDITION 'fs:file-not-found "File not found for ~a"
								 pathname :delete-and-expunge))))))
				   (t (check-ftp-status pathname ftp nil))))
	    (SETF finished-p t))
	;; cleanup
	(WHEN ftp (free-ftp-connection ftp (OR (TYPEP ftp-result 'condition) (NOT finished-p)))))
      (IF (ERRORP ftp-result)
	  ;; THEN return error condition 
	  (IF error-p (SIGNAL-CONDITION ftp-result) ftp-result)
	  ;; ELSE return blocks freed, second value?
	  1))))

(DEFMETHOD (FTP :create-directory) (ignore pathname &optional (error-p t) &aux ftp finished-p result)
  (ip:with-stream-whostate
    "Create-directory"
    (file-operation-retry
      (UNWIND-PROTECT
	  (PROGN
	    (SETQ ftp (reserve-ftp-connection (SEND pathname :host) :create-directory)) ;; last arg is nicety for peek
	    (SETF result
		  (IF (ftp-unfavorable (SEND ftp :create-directory
					     (dir-pathstring-for-ftp (SEND pathname :new-pathname
									   :name :wild :type nil :version nil))))
		      (IF (NOT (EQL 550 (SEND ftp :reply-code)))
			  (MAKE-CONDITION 'ftp-error (SEND ftp :reply-string) pathname :create-directory)
			  (COND ((OR (SEARCH "ermission" (SEND ftp :reply-string))
				     (SEARCH "ERMISSION" (SEND ftp :reply-string))
				     (SEARCH "rivilege" (SEND ftp :reply-string))    ;03-22-88 DAB mispelled privilege
				     (SEARCH "RIVILEGE" (SEND ftp :reply-string)))   ;03-22-88 DAB mispelled privilege
				 (MAKE-CONDITION 'incorrect-access-to-directory (SEND ftp :reply-string)
						 pathname :create-directory))
				((OR (SEARCH "xists" (SEND ftp :reply-string))
				     (SEARCH "XISTS" (SEND ftp :reply-string))
				     ;; workaround Unix returning two differing strings here
				     (SEARCH "Is a directory" (SEND ftp :reply-string)))
				 (MAKE-CONDITION 'file-already-exists (SEND ftp :reply-string)
						 pathname :create-directory))
				(t (MAKE-CONDITION 'directory-not-found "Directory not found for ~a"
						   pathname :create-directory))))
		      t))
	    (SETF finished-p t))
	;; cleanup
	(free-ftp-connection ftp (OR (TYPEP result 'condition) (NOT finished-p))))
      (IF (AND error-p (TYPEP result 'condition))
	  (SIGNAL-CONDITION result)
	  result))))

(DEFMETHOD (FTP :directory-stream) (medium pathname &optional options
				    &aux (error-p t) ftp ftp-list stream result fast? finished-p reply-code link-name)
  ;; Process options:
  ;;   only supported options are :noerror, :no-extra-info (probably not exactly the same), :sorted
  ;;   other options of directory-list =  :deleted
  ;;   other options of chaos-pathname :directory-stream = :fast, :directories-only
  (IGNORE medium)
  (ip:with-stream-whostate
    "Directory"
    (file-operation-retry
      (WHEN (MEMBER :noerror options :test 'EQ)
	(SETQ error-p nil))
      (SETQ fast? (OR *ftp-fast-dir-list* (MEMBER :no-extra-info options :test 'EQ)))
      (UNWIND-PROTECT
	  (PROGN
	    (SETF stream (MAKE-INSTANCE 'ftp-directory-stream :state :newly-opened
					:pathname pathname
					:string-for-printing (SEND pathname :string-for-printing)))
	    ;; Loop to retry when we open a link and the operating system (e.g. Unix) did not resolve the link.
	    ;; Whenever a throw is done to :link-error the value thrown is the name of the link.
	    (LOOP
	      (SETF ftp-list
		    (PROG1
		      (CATCH :directory-link-error
			;; unless a throw occurs, this return will exit the loop
			(RETURN
			  (LET ((*ftp-directory-link* link-name))
			    (SETQ ftp (reserve-ftp-connection (PATHNAME-HOST pathname) stream)) ;; last arg is nicety for peek
			    (SETQ ftp-list (SEND self :directory-list-from-ftp
						 (SEND self :get-ftp-implementation-type (PATHNAME-HOST pathname))
						 pathname ftp stream fast?))
			    (SEND stream :set-done t))))
		      ;; Exit the loop when we have already tried what we thought was the truename.
		      ;; Aborting this second retry attempt may occur with a circular link, invalid link, etc.
		      (WHEN ftp-list
			(SETF (SEND ftp :reply-code) 450)
			(SETF (SEND ftp :reply-string) (FORMAT nil "Directory not found for ~a" pathname))
			(RETURN))))
	      ;; reset to try again with directory link resolved
	      (UNWIND-PROTECT
		  (free-ftp-connection ftp nil :abort-data-conn)
		(SETF ftp nil))
	      (SEND stream :close)
	      ;; resolve the directory link portion of the pathname by chasing links
	      (SETF ftp-list (modify-dir-list (FIRST (FIRST ftp-list)) ftp-list nil))
	      (SETF link-name (GETF (CDR (FIRST ftp-list)) :link-to))
	      (SETF link-name (SEND pathname :new-directory
				    (APPEND (SEND link-name :directory)
					    (LIST (STRING-APPEND (IF (STRINGP (SEND link-name :name))
								     (SEND link-name :name)
								     "")
								 (IF (STRINGP (SEND link-name :type))
								     (STRING-APPEND "." (SEND link-name :type))
								     "")))))))
	    ;; NOTE - this assumes that the final data connection hasn't been closed so that the control connection
	    ;;          still contains the status
	    (SETF result
		  (COND
		    ;; some Unix-like systems will violate the spec and return a 550 here
		    ((OR (EQL 450 (SETF reply-code (SEND ftp :reply-code)))
			 (EQL 550 (SETF reply-code (SEND ftp :reply-code))))
		     (IF (OR (SEARCH "ermission" (SEND ftp :reply-string)) (SEARCH "ERMISSION" (SEND ftp :reply-string))
			     (SEARCH "rivilege" (SEND ftp :reply-string))      ;03-22-88 DAB mispelled privilege
			     (SEARCH "RIVILEGE" (SEND ftp :reply-string)))     ;03-22-88 DAB mispelled privilege
			 (MAKE-CONDITION 'incorrect-access-to-directory (SEND ftp :reply-string)
					 pathname :directory-stream)
			 (MAKE-CONDITION 'directory-not-found "Directory not found for ~a"
					 pathname :directory-stream))) 
		    ((NOT (ftp-unfavorable reply-code))
		     ;; release the control connection so that modify-dir-list can use it
		     (UNWIND-PROTECT
			 (free-ftp-connection ftp)
		       (SETF ftp nil))
		     (SETQ ftp-list (modify-dir-list pathname ftp-list (MEMBER :sorted options :test 'EQ)))
		     (SETF (SEND stream :dir-list) (PUSH (LIST nil :pathname pathname) ftp-list))
		     stream)
		    (t (check-ftp-status pathname ftp nil))))
	    (SETF finished-p t)
	    )			;; end progn
	(WHEN ftp
	  ;; close data-connection with :abort in case all data hasn't been read
	  (free-ftp-connection ftp (OR (TYPEP result 'condition) (NOT finished-p)) :abort-data-conn)) 
	(SEND tv:who-line-file-state-sheet :delete-stream stream))	;; end unwind protect
      (IF (AND error-p (TYPEP result 'condition))
	  (SIGNAL-CONDITION result)
	  result))))


))
