;; -*- Mode: COMMON-LISP; Package: File-System; cold-load:t; Base: 10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated.  All rights reserved.
; Copyright (C) 1980, Massachusetts Institute of Technology

;;; These Macros were in FSGUTS to start with.  -------R.N.B.

(DEFSUBST REQUIRE-DISK-CONFIGURATION ()
  (IF (NOT (BOUNDP 'DISK-CONFIGURATION))
    (LM-SIGNAL-ERROR 'NO-FILE-SYSTEM)))

(DEFSUBST MARK-BLOCK-SPACE (LOC NPAGES NEW-STATUS)
  "Marks NPAGES starting at LOC in the PUT as NEW-STATUS.  
   Called only during file system boot."
  (LET ((LIM (+ LOC NPAGES)))
    (DO ((I LOC (1+ I)))
	((>= I LIM))
      (UNLESS (ZEROP (AREF PAGE-USAGE-TABLE I));if already allocated, throw out.
	(NOTIFY "Possible dual allocation in file band.")
	(THROW 'MUST-SALVAGE
	       T))
      (SETF (AREF PAGE-USAGE-TABLE I) NEW-STATUS))
    (DECF (AREF PUT-USAGE-ARRAY PUT-FREE) NPAGES)
    (INCF (AREF PUT-USAGE-ARRAY NEW-STATUS) NPAGES)))


 
(DEFSUBST SKIP-BYTES (STREAM NBYTES)
  "Skips the next NBYTES in STREAM."
  (SEND STREAM :INCREMENT-POINTER NBYTES))  
  
(DEFSUBST SKIP-STRING (STREAM)
  "Skips the next string on STREAM"
  (SEND STREAM :INCREMENT-POINTER (SEND STREAM :TYI)))  

(DEFSUBST SKIP-LINE (STREAM)
  "Skips the next line on STREAM. 
   Returns T if an EOF was encountered."
  (SEND STREAM :SKIP-UNTIL #\NEWLINE))  

(DEFSUBST SKIP-UNTIL (STREAM CHAR)
  "Reads characters from STREAM until a non-slashified CHAR is found.
   Returns T if an EOF was encountered."
  (SEND STREAM :SKIP-UNTIL CHAR T))  

(defun skip-bignum (stream)
  (let ((length-bignum (get-bytes stream 1.)))             ;first byte is the length, not counting sign byte.
    (skip-bytes stream (1+ length-bignum))))

(DEFSUBST SKIP-ALL-NON-SPACES (STREAM)
  (SKIP-UNTIL STREAM #\SPACE)
  (SEND STREAM :UNREAD-CHAR #\SPACE))

(DEFSUBST LMFS-DELETED-FILE? (FILE)
  (AND (FILE-DELETED? FILE) (NOT (FILE-CLOSED? FILE))))  



(DEFSUBST LMFS-CLOSED-FILE? (FILE)
  (AND (FILE-CLOSED? FILE) (NOT (FILE-DELETED? FILE))))  



(DEFSUBST LMFS-FILE-BEING-WRITTEN-OR-SUPERSEDED? (FILE)
  (NOT
   (AND (FILE-CLOSED? FILE) (NULL (FILE-OVERWRITE-FILE FILE)) (NULL (FILE-DONT-SUPERSEDE? FILE)))))  



(DEFSUBST REQUIRE-CLOSED-FILE (FILE)
  (IF (OR (NOT (FILE-CLOSED? FILE)) (MINUSP (FILE-OPEN-COUNT FILE)))
    (LM-SIGNAL-ERROR 'OPEN-OUTPUT-FILE)))  



(DEFSUBST REQUIRE-CLOSED-FILE-NOT-BEING-SUPERSEDED (FILE)
  (IF (LMFS-FILE-BEING-WRITTEN-OR-SUPERSEDED? FILE)
    (LM-SIGNAL-ERROR 'OPEN-OUTPUT-FILE)))  



(DEFSUBST REQUIRE-DELETABLE-FILE (FILE)
  (IF (FILE-ATTRIBUTE FILE :DONT-DELETE)
    (LM-SIGNAL-ERROR 'DONT-DELETE-FLAG-SET)))  



(DEFSUBST REQUIRE-ZERO-OPEN-COUNT (FILE)
  (IF (NOT (ZEROP (FILE-OPEN-COUNT FILE)))
    (FERROR () "Expected zero open count for ~S." FILE)))  



(DEFSUBST REQUIRE-FILE-NOT-BEING-SUPERSEDED (FOBJ NAME TYPE)
  (AND (TYPEP FOBJ 'FILE) (STRING-EQUAL NAME (FILE-NAME FOBJ))
     (STRING-EQUAL TYPE (FILE-TYPE FOBJ))
     (IF (FILE-DONT-SUPERSEDE? FOBJ)
       (LM-SIGNAL-ERROR 'OUTPUT-FILE-BEING-SUPERSEDED))))  

