;;; -*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.
;;;

;;;----------------------------------------------------------------------------------------------
;;;                            BUG REPORT GENERATION SOFTWARE

;;; This file contains all the support to generate a bug report.  The function (bug), 
;;; M-X bug in ZMACS, or C-m in the error handler, will involk this code.  If not in the
;;; cold-load-stream, it creates a bug report template in a ZMACS mail buffer which can be
;;; edited and mailed to the appropriate location.  When in the cold-load-stream, it prompts
;;; for input to the necessary fields and mails the identically formatted bug template as if
;;; submitted directly from ZMACS.

;;; Some fields are pre-loaded from site/host/user data in the namespace.  This data can exist
;;; at any of these levels, but the lowest level found is the data that is used (see si:get-site-option).
;;; The attributes used are:
;;;    :mail-address-for-bug-reports   - for To: address to mail the bug report (normally a SITE attribute)
;;;    :mail-address                   - for From: and CC: (normally a USER attribute) 
;;;    :customer-id-for-bug-reports    - for Customer-ID field (normally a SITE attribute)
;;;    :personal-name                  - for Name field (normally a USER attribute)
;;;    :location                       - for Location field (could be SITE/HOST/USER)
;;;    :postal-address                 - for Address field  (typically a SITE attribute)
;;;    :work-phone                     - for Phone field (typically a USER attribute)

;;; The bug template always includes software and hardware configuration information and also
;;; will include a backtrace when submitting a bug from the error handler.

;;; In case the network or mailer is totally hosed, a postal address is on the template where
;;; the bug report can be mailed after getting a hard copy.
;;;----------------------------------------------------------------------------------------------

(defflavor BUG-REPORT-OBJECT
	   ((to "") (cc "") (subject "")  (system-that-errored "")
	    (name "") (location "") (postal-address "") (phone "") (mail-address "") (tracking-id "")
	    (date-time "") (priority "") (type "") (machine-type "")
	    (user-description "") (work-around "")
	    (backtrace "") (software-configuration "") (hardware-configuration ""))
	   ()
  :gettable-instance-variables
  :settable-instance-variables
  :inittable-instance-variables)


;;; Add on the name of the host that submitted the bug onto the tail of the software configuration string.
;;; Done in MAIL-READER 3.1 for release 3-1. -dkm 6/87

;;; Added an initialization of the tracking-id field (customer-id) from the namespace.  
;;; Done in MAIL-READER 3.2 for release 3-1. -dkm 6/87

(defmethod (BUG-REPORT-OBJECT :INIT-FROM-ENVIRONMENT) (&optional eh-backtrace eh-system-that-errored)
  ;; This will initialize a bug object for the standard information that is contained
  ;; in a bug report except the data the user must fill in.  The backtrace and system
  ;; that errored come from the error handler and must be passed in.
  (let ((attribute-alist (get-bug-report-attribute-alist)))
    (setq date-time              (time:print-current-time nil))
    (setq to                     (or (cadr (assoc :mail-address-for-bug-reports attribute-alist)) ""))
    (setq cc                     (or (cadr (assoc :mail-address attribute-alist)) ""))
    (setq subject                "Explorer Bug Report")
    (setq priority               "?")
    (setq type                   "B")
    (setq machine-type           (if (si:addin-p) "M" "E"))  ;; LAS 5-17-88 default to microExplorer if you are on one
    (setq tracking-id            (or (cadr (assoc :customer-id-for-bug-reports attribute-alist)) "")) ;new dkm 6/87
    (setq name                   (or (cadr (assoc :personal-name attribute-alist)) ""))
    (setq location               (or (cadr (assoc :location attribute-alist)) ""))
    (setq postal-address         (or (cadr (assoc :postal-address attribute-alist)) ""))
    (setq phone                  (or (cadr (assoc :work-phone attribute-alist)) ""))
    (setq mail-address           (or (cadr (assoc :mail-address attribute-alist)) ""))
    (setq system-that-errored    (or eh-system-that-errored ""))
    (setq backtrace              (or (pad-string eh-backtrace) ""))
    (setq software-configuration (or (pad-string (pretty-description-string  ;append host name on config info  dkm 5/87
						   (string-append (si:system-version-info)
								  ", Host: " si:local-host-name))) ""))
    (setq hardware-configuration (or (pad-string (si:get-hardware-info)) ""))))


;;; Add FROM: field onto the bug report template.  Done in MAIL-READER 3.1  -dkm 6/87

(defmethod (BUG-REPORT-OBJECT :FULL-BUG-REPORT) ()
  ;; Returns a string with the entire bug report, including headers.  Suitable
  ;; for stashing in an editor buffer, for example.

  (string-append
    "From:    " cc #\return
    "To:      " to #\return
    "CC:      " cc #\return
    "Subject: " subject #\return #\return
    (send self :bug-report-body)))


(defmethod (BUG-REPORT-OBJECT :BUG-REPORT-BODY) ()
  ;; Returns a string with the bug report body (no headers like to: cc: subject:)

  (string-append
    "      EXPLORER (TM) BUG REPORT" #\return #\return
    "DATE-TIME   : " date-time #\return
    "PRIORITY    : " priority     "     (H)igh (M)edium (L)ow" #\return
    "TYPE        : " type         "     (B)ug  (D)esign (M)anual" #\return
    "MACHINE-TYPE: " machine-type "     (E)xplorer (M)icroExplorer" #\return #\return
    "DESCRIPTION-OF-PROBLEM:" #\return "   " user-description #\return #\return
    "WORK-AROUND:" #\return "   " work-around #\return #\return
    "CUSTOMER-ID: " tracking-id #\return #\return
    "NAME       : " name #\return
    "LOCATION   : " location #\return
    "ADDRESS    : " postal-address #\return
    "PHONE      : " phone #\return
    "NET-ADDRESS: " mail-address #\return #\return
    (if (and backtrace
	     (plusp (length backtrace)))
	(string-append "BACKTRACE:" #\return #\return backtrace #\return)
	"")
    "SOFTWARE-CONFIGURATION:" #\return #\return software-configuration #\return #\return
    "HARDWARE-CONFIGURATION:" #\return hardware-configuration #\return
    "
		    ********************************
		    *         Return to:           *
		    *                              *
		    *  EXPLORER BUG REPORTS        *
		    *  c/o Explorer Project Manager*
		    *  TEXAS INSTRUMENTS, MS 2201  *
		    *  P.O. Box 2909               *
		    *  Austin, Texas   78769       *
		    ********************************" #\return))

(compile-flavor-methods bug-report-object)


(defun BUG ()
  "Create A bug report editor frame and go to it from Lisp"
  (mail-bug))

(defcom COM-BUG
	"Mail a bug report." ()
  (mail-bug)
  dis-none)

;;; Add "BUG" to the zmacs command table, therefore allowing meta-x BUG.
(eval-when (load)
  (set-comtab
    *zmacs-comtab*
    nil
    '(("BUG" . com-bug))))

;; The error handler passes the backtrace string and a string for the system which errored.
;; This needs to get passed into the mail template -- but mail templates don't support arguments.
;; So, unfortunately, I need to pass them in as globals to the template -- I can't make local
;; bindings either because of the process switch.  So, we need the following globals:

(defvar *BACKTRACE* nil)
(defvar *SYSTEM-THAT-ERRORED* nil)
(defvar *WINDOW-SUBMITTING-BUG* nil)


;;; Changed to call find-or-create-idle-zmacs-window with new optional arg
;;; that says it's OK to select the window if it's busy running M-X com-bug
;;; Also added a global variable so that the correct window (where bug originated)
;;; is reselected after sending the bug report if it came from the error handler.
;;; done in patch MAIL-READER 3.1 for release 3-1.

(defun MAIL-BUG (&optional backtrace (system-that-errored "Explorer"))
  "General routine that can be called from anywhere, or by the error handler to
generate a bug report.  This will work when in the cold-load-stream, getting
the input directly from the keyboard instead of a pre-formatted mail buffer.
BACKTRACE -  the string that the error handler passes in showing the stack,etc.
SYSTEM-THAT-ERRORED - a string indicating what system failed, also passed by
                      the error handler (however, this seems to always be NIL)."
    
  (cond ((and (eq *terminal-io* sys:cold-load-stream)
	      (fboundp 'mail:submit-mail)
	      (fboundp 'qsend-get-message))
	 (format *query-io* "~%Mail a Bug Report from the Cold Load Stream ... ~2%")
	 (mail-cold-bug backtrace system-that-errored))
	
	((and (fboundp 'mail:submit-mail)
	      (fboundp 'find-or-create-idle-zmacs-window))
	 (let ((fcn (get-mail-template-function 'Bug-Report)))
	   (if fcn
	       (progn
		 (setq *window-submitting-bug* (and backtrace tv:selected-window))
		 (format *query-io* "~&Mail a bug report from Mail Buffer.  Entering the editor...")
		 (let ((zmacs-window (find-or-create-idle-zmacs-window 'com-bug)))
		   (w:await-window-exposure)
		   ;; These get set here, and get cleared out by the Bug Template
		   ;; You can't reset them in this function, because it would occur
		   ;; before the bug template had a chance to use them.
		   (setq *backtrace* backtrace
			 *system-that-errored* system-that-errored)
		   (send zmacs-window :force-kbd-input
			 `(:execute  ,fcn))
		   (send zmacs-window :select)))
	       
	       ;;else
	       (format *query-io* "~&Bug Report Template not found.  Attempting simple Bug Report ...")
	       (mail-cold-bug backtrace system-that-errored))))
	
	(t (format *query-io* "~&Cannot file a bug report because the necessary systems are not installed."))))


;;; Indicate the window to switch to as indicated in *window-submitting-bug*.  This will be NIL
;;; unless we came from the error handler.  Done in patch for MAIL-READER 3.1 -dkm 6/87

(define-mail-template BUG-REPORT "Bug Report" :mail
  "File a Bug Report"

  ;; Put the window that called the bug template on the :return-to-window-after-send 
  ;; property, so that we go back to that window after the bug report is sent out. -dkm 6/87
  (and *window-submitting-bug*
       (putprop *interval* *window-submitting-bug* :return-to-window-after-send))

  ;; Create an instance of a bug-report object and initialize it with the
  ;; data in the current environment.  This includes everything except what
  ;; the user needs to type in to describe the problem.  Note *backtrace*
  ;; and *system-that-errored* are globals which are normally nil but are bound
  ;; to the appropriate values just before this template is called.  Then they
  ;; are set back to nil, to be sure that this data isn't used accidently later.
  ;; This could be eliminated if there was a way to pass arguments into the mail
  ;; template ...

  (let ((bug-object (make-instance 'bug-report-object)))
    (send bug-object :init-from-environment *backtrace* *system-that-errored*)

    ;; Insert the bug info into the buffer
    (insert (point) (send bug-object :full-bug-report)))

  (setq *backtrace* nil
	*system-that-errored* nil
	*window-submitting-bug* nil)

  ;; position point at the description field. This fails ofcourse, if the
  ;; text is changed for this line.
  (do ((line (bp-line (point)) (line-next line)))
      ((or (null line)
	   (string-equal line "DESCRIPTION-OF-PROBLEM:"))
       (and line
	    (move-bp (point) (end-line (create-bp (line-next line) 0)))))))


;;;  Fix prompt when submitting a bug from the cold load stream to use "Customer-ID" instead of
;;; Tracking ID, like the mail format has. Patched in MAIL-READER 3-1.  -dkm 6/87

(defun MAIL-COLD-BUG (&optional backtrace system-that-errored)
  "Collects user information for a bug that landed you in the cold load stream,
and mails it to address specified"

  (let ((bug-object (make-instance 'bug-report-object))

	;; This input list indicates which bug fields will be edited.  It must contain 3 elements
	;; and can have 4 elements.
	;;    ele 1 - instance variable of bug object to edit
	;;    ele 2 - string, without args, to be passed to format for the prompt
	;;    ele 3 - t if RETURN (or END) should end the field.  nil if END must end the field
	;;    ele 4 - optional list of acceptable strings this field can take
	(input-list `((:to                "~&~%Mail to: " t)
		      (:cc                "~&CC: " t)
		      (:subject           "~&Subject: " t)
		      (:priority          "~&Priority - (H)igh (M)edium (L)ow : " t ("h" "m" "l"))
		      (:type              "~&Type of Problem - (B)ug (D)esign (M)anual : " t ("b" "d" "m"))
                      (:machine-type      "~&Type of Machine - (E)xplorer (M)icroExplorer " t ("e" "m"))
		      (:user-description  "~&~%Description of Problem: (press END when finished)~%" nil)
		      (:work-around       "~&~%Work Around: (press END when finished)~%" nil)
		      (:tracking-id       "~&~%Customer-ID (optional): " t))))   ;changed from Tracking-ID dkm 6/87

    (send bug-object :init-from-environment backtrace system-that-errored)

    ;; since we're not in an editor buffer, let's not default these two fields.
    (send bug-object :set-cc "")
    (send bug-object :set-subject "")
    (send bug-object :set-priority "")
    
    ;; repeat  until bug report sucessfully sent or they quit
    (do (done) (done)

      ;; let them input/edit some stuff for the bug report
      (dolist (input input-list)
	(do () (nil)
	  (let ((old-value (send bug-object (first input))))
	    (format t (second input))
	    (if (not (equal old-value ""))
		(format t " (default-> ~a) " old-value))
	    (let ((values-list (fourth input))
		  (input-text (qsend-get-message *standard-input* nil (third input))))
	      (when (not (equal input-text ""))
		    (setf (send bug-object (first input)) input-text))
	      (if (and values-list
		       (not (member (send bug-object (first input)) values-list :test #'string-equal)))
		  (beep)
		  ;;else
		  (return))))))
      
      (when (y-or-n-p "~&~%Mail Bug Report? ")

	;; build up the actual bug report and (attempt to) mail it.
	 (setq done (mail:submit-mail (send bug-object :bug-report-body)
				       :to (list (send bug-object :to))
				       :other-headers (list (string-append "CC:    " (send bug-object :cc)))
				       :subject (send bug-object :subject))))
      (unless done
	(and (y-or-n-p "~%Bug Report not successfully handled.  Quit? ")
	     (setq done t))))))


;;; Use :string-for-message instead of :string-for-printing. Patched in  MAIL-READER 3.1 -dkm 6/87

;;; Added another attribute to extract from the namespace.  This is really only intended to be
;;; used locally to identify locally submitted bug reports.  Customers shouldn't have a need to
;;; use this attribute in the namespace, but they could.  Patched in MAIL-READER 3.2 -dkm 6/87

(defun GET-BUG-REPORT-ATTRIBUTE-ALIST (&optional local-only-p)
;;; Returns an alist of the attributes used in the bug report.  It does this
;;; safely, so in case the network/namespace is broken, we don't go into the error
;;; handler again.  A default alist is returned in the case the namespace in
;;; available at all.  This also will replace #\return with a ";" so that things
;;; like :POSTAL-ADDRESS will come out on a single line.

  (let ((bug-attributes '(:mail-address-for-bug-reports
			  :customer-id-for-bug-reports  ;added 6/87 -dkm
			  :location
			  :postal-address
			  :personal-name
			  :work-phone))
	(default-bug-report-address "EXPBUG@CSC.TI.COM")
	(default-mail-address (string-append user-id #\@ si:local-host-name))
	alist)
    (condition-case nil
	(progn
	 (dolist (option bug-attributes alist)
	   (push (list option
		       (or (substitute #\; #\return (get-site-option option local-only-p))
			   (cond ((eq option :mail-address-for-bug-reports)
				  default-bug-report-address)
				 ((eq option :personal-name)
				  user-id)
				 ((eq option :mail-address)
				  default-mail-address)
				 (t ""))))
		 alist))
	 ;; This object holds the correct mail address.  It looks for mail:*user-mail-address*
	 ;; or :mail-address on USER object in the namespace, or USER-ID@SI:LOCAL-HOST-NAME.
	 (push (list :mail-address (send (mail:default-from-address) :string-for-message)) alist))
      
      ;; If we run into some error hitting the namespace,  try again just getting
      ;; LOCAL cached namespace information (if this wasn't a local access).  Otherwise
      ;; just return some defaulted values.
      (error
       (if local-only-p
	   `((:mail-address-for-bug-reports ,default-bug-report-address)
	     (:personal-name ,user-id)
	     (:mail-address ,default-mail-address))
	   ;;else
	   (get-bug-report-attribute-alist t))))))


(DEFUN PRETTY-DESCRIPTION-STRING (description)
  ;; Insert CR after every 72 characters in description string.
  (LET ((NEW-DESCRIPTION DESCRIPTION))
    (LOOP WITH LINE-START = 0
	  FOR START = 0 THEN (+ COMMA-POS 2)
	  AS PREV-COMMA-POS = NIL THEN COMMA-POS
	  AS COMMA-POS = (LISP:SEARCH ", " (THE STRING NEW-DESCRIPTION)
				      :START2 START
				      :TEST #'CHAR-EQUAL)
	  WHEN (> (- (OR COMMA-POS (LENGTH NEW-DESCRIPTION )) LINE-START) 72.)
	  UNLESS (NULL PREV-COMMA-POS)
	  DO (SETF (AREF NEW-DESCRIPTION (1+ PREV-COMMA-POS)) #\NEWLINE)
	     (SETQ LINE-START (+ PREV-COMMA-POS 2))
	     (SETQ COMMA-POS PREV-COMMA-POS)
	  UNTIL (NULL COMMA-POS))
    NEW-DESCRIPTION))

(defun PAD-STRING (str &optional (pad 3))
  "Put PAD number of blanks before every line in STR"
  (and str
       (stringp str)
       (let ((new-string (make-array 0 :type :art-string :fill-pointer 0 :initial-element #\space))
	     (pad-string (make-string pad :initial-element #\space))
	     (str-len (length str)))
	 (do* ((start 0 (1+ end))
	       (end (position #\return str) (position #\return str :start (1+ end))))
	      ((eq start str-len)
	       (copy-seq new-string))
	   (if (null end)
	       (setq end (1- str-len)))
	   (string-nconc new-string pad-string (subseq str start (1+ end)))))))
