;;; -*- Mode:Common-Lisp; Package:ZWEI; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985- 1989 Texas Instruments Incorporated. All rights reserved.

(defvar *FILTER-SUMMARY-FONTS* '(:cptfont :cptfont :cptfontb))


(defvar *EMPTY-MESSAGE-SEQUENCE* (make-instance 'message-sequence :name "*Empty-Message-Sequence*"))


(defmethod (filter-summary-buffer :after :init) (&rest ignore)
  (set-buffer-fonts self *filter-summary-fonts*)
  (push (list 'filter-summary-mode) saved-mode-list))


(defminor COM-FILTER-SUMMARY-MODE FILTER-SUMMARY-MODE "Filter-Summary" 1
	  "Minor mode for editing a DIRED style summary of mail filters." ()
    (set-comtab *read-mail-comtab*
	      '(#\S com-expand-filter
		#\N com-mail-down-real-line
		#\P com-mail-up-real-line)
	      '(("Delete Filters" . com-delete-filters)
		("Create Filter" . com-filter-messages)
		("Create Filter From Keywords" . com-create-filter-from-keywords)
		("Filtered Mode" . com-toggle-filter-summary-mode))))


(defcom COM-EXPAND-FILTER
	"Expand or Unexpand the filter at point."
	()
  (let* ((line (bp-line (point)))
	 (line-message (getf (line-plist line) :summary-message))
	 (line-filter (getf (line-plist line) :filter-summary-filter))
	 filter expand-p)
    (cond ((not (filter-summary-p *interval*))
	   (barf "Not within a filter summary buffer."))
	  (line-message
	   ;; On a message line (of an expanded filter).  Unexpand it.
	   (let ((sequence (getf (line-plist line) :summary-sequence)))
	     (setq filter (get sequence :filter-summary-filter))))
	  ((and line-filter
		(get line-filter :expanded-p))
	   (setq filter line-filter))
	  (line-filter
	   ;; On a line with a filter name and it's unexpanded.  Expand it.
	   (setq filter line-filter)
	   (setq expand-p t))
	  (t
	   (barf "Line does not describe message or filter.")))
    (let ((sequence (get filter :message-sequence))
	  (filter-line (get filter :filter-summary-line)))
      (when (and sequence filter filter-line)
	(setf (get filter :expanded-p) expand-p)
	(with-read-only-suppressed (*interval*)
	  (change-font-interval (create-bp filter-line 0)
				(create-bp filter-line (length filter-line))
				t (if expand-p 2 1)))
	(update-filter-summary *interval* sequence)
	(when (and (eq (get filter :expanded-p) t)
		   (eql (total-messages sequence) 0))
	  (format *query-io* "No messages pass this filter.")))))
  dis-text)

(defcom COM-DELETE-FILTERS
	"Choose filters to delete from this filter summary."
	()
  
  (in-mail-context (:require-buffer t)
    (let* ((mail-file-buffer (mail-file-buffer-of *mail-buffer*))
	   (summary (mail-summary-of mail-file-buffer))
	   item-list)
      (loop
	for filter in (buffer-filter-list summary)
	for name-string = (send filter :string-for-filter-summary)
	unless (equalp (send filter :name) "ALL MESSAGES")
	do
	(push (cons name-string filter) item-list))
      (setq item-list (nreverse item-list))
      (multiple-value-bind (nuke-list do-it-p)
	  (w:multiple-menu-choose item-list :label "Choose Filters to Delete")
	(when do-it-p
	  (dolist (filter nuke-list)
	    (setf (get filter :expanded-p) nil))
	  (update-filter-summary summary)
	  (with-read-only-suppressed (summary)
	    (loop
	      for filter in nuke-list
	      for line = (get filter :filter-summary-line)
	      do
	      (delete-interval (create-bp line 0) (create-bp (line-next line) 0))
	      (setf (buffer-filter-list summary) (deleq filter (buffer-filter-list summary)))
	      (setf (get-mail-option mail-file-buffer :filter-alist)
		    (delete (send filter :string-for-filter-summary)
			    (get-mail-option mail-file-buffer :filter-alist)
			    :key #'car :test #'equalp))))))
      (mung-node mail-file-buffer))
    (values dis-text dis-text)))


(defcom COM-TOGGLE-FILTER-SUMMARY-MODE
	"Toggle Filter Summary mode."
	()
  (in-mail-context (:require-buffer t)
    (let* ((summary (mail-summary-of *mail-buffer*))
	   (current-mode (if summary
			     (if (filter-summary-p summary)
				 :filtered :basic)
			   *mail-summary-mode*)))
      (setq *mail-summary-mode* (if (eq current-mode :filtered)
				    :basic :filtered))
      (com-revert-mail-summary))))


;;; Add this command - left out in 3.1, spr6335 - nn
(defcom COM-CREATE-FILTER-FROM-KEYWORDS
	"Create filter from Currently Existing Keywords."
	()
  (let ((*mail-mouse-command* t))
    (in-mail-context (:require-buffer t)
      (let ((filter (get-mail-filter 'keyword)))
	(when filter
	    (add-new-filter-summary-filter filter (mail-file-buffer-of *mail-buffer*))
	    (return-from com-create-filter-from-keywords dis-text))
	(values dis-text dis-text)))))  


(defun MAKE-FILTER-SUMMARY (mail-file-buffer name)
  
  (let ((summary (make-instance 'filter-summary-buffer :name name :sequence-buffer mail-file-buffer)))
    
    (with-read-only-suppressed (summary)
      (insert-summary-heading (interval-last-bp summary))
      (insert-line-with-leader (make-diagram-line 'mail-summary-black-line-diagram)
			       (bp-line (interval-last-bp summary))))
    
    (let ((filter (make-instance 'system-mail-filter
				 :name "ALL MESSAGES"
				 :function 'messagep
				 :string-for-filter-summary "ALL MESSAGES")))
      (add-filter-summary-filter filter summary))
    (dolist (filter-item (get-mail-option mail-file-buffer :filter-alist))
      (let ((filter nil))
	(when (fboundp (second filter-item))
	  (setq filter (make-instance 'system-mail-filter
				      :name (first filter-item)
				      :string-for-filter-summary (first filter-item)
				      :function (second filter-item)
				      :args (cddr filter-item)))
	  (when filter
	    (add-filter-summary-filter filter summary)))))
    summary))


(defun ADD-FILTER-SUMMARY-FILTER (filter filter-summary)
  
  (let ((mail-file-buffer (mail-file-buffer-of filter-summary)))
    (cond ((equalp (send filter :name) "ALL MESSAGES")
	   (setf (get mail-file-buffer :filter-summary-filter) filter)
	   (setf (get filter :message-sequence) mail-file-buffer)
	   (push filter (buffer-filter-list filter-summary)))
	  (t
	   (let ((seq (make-mail-buffer-from-filter filter mail-file-buffer nil)))
	     (setf (get seq :filter-summary-filter) filter)
	     (setf (get filter :message-sequence) seq)
	     (setf (buffer-summary-buffer seq) filter-summary))
	   (push-end filter (buffer-filter-list filter-summary))))
    
    (with-read-only-suppressed (filter-summary)
      (let ((filter-string (send filter :string-for-filter-summary))
	    line)
	(when filter-string
	  (setq line (create-line 'art-string (+ (length filter-string) 4) filter-summary))
	  (setf (line-length line) 0)
	  (append-to-line line "  ")
	  (append-to-line line filter-string)
	  (splice-line-into-interval line (bp-line (interval-last-bp filter-summary)))
	  (mung-line line)
	  (setf (get filter :filter-summary-line) line)
	  (setf (getf (line-plist line) :filter-summary-filter) filter))))))


(defun UPDATE-FILTER-SUMMARY (summary &optional sequence)
  
  (cond (sequence
	 (let* ((filter (get sequence :filter-summary-filter))
		(summary-line (get filter :filter-summary-line)))
	   (when (and filter summary-line)
	     (cond ((get filter :expanded-p)
		    (unless (mail-file-buffer-p sequence)
		      (revert-filter-buffer sequence nil))
		    (update-summary summary sequence (create-bp (line-next summary-line) 0)))
		   (t
		    (update-summary summary *empty-message-sequence* (create-bp (line-next summary-line) 0)))))))
	;; Update all expanded filters
	(t
	 (loop
	   for filter in (buffer-filter-list summary)
	   for filter-buffer = (get filter :message-sequence)
	   for summary-line = (get filter :filter-summary-line)
	   when (and filter-buffer summary-line)
	   do
	   (cond ((get filter :expanded-p)
		  (unless (mail-file-buffer-p filter-buffer)
		    (revert-filter-buffer filter-buffer nil))
		  (update-summary summary filter-buffer (create-bp (line-next summary-line) 0)))
		 (t 
		  (update-summary summary *empty-message-sequence* (create-bp (line-next summary-line) 0))))))))


(defun REMINDER-MESSAGES-FILTER (msg)
  (msg-reminder-p msg))
  
(defun UNSEEN-MESSAGES-FILTER (msg)
  (message-attribute-p :unseen msg))

(defun RECENT-MESSAGES-FILTER (msg)
  (message-attribute-p :recent msg))

(defun DELETED-MESSAGES-FILTER (msg) 
  (message-attribute-p :deleted msg))

(defun NOT-ANSWERED-MESSAGES-FILTER (msg) 
  (not (message-attribute-p :answered msg)))

(defun KEYWORD-FILTER-FUNCTION (msg keywords)
  (keyword-and-filter msg keywords))

