;;; -*- Mode:Common-Lisp; Package:ZWEI; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985- 1989 Texas Instruments Incorporated. All rights reserved.


(defvar *DEFAULT-HEADER-SEARCH-STRING* "")

(defvar *DEFAULT-MAIL-SEARCH-STRING* "")

(defvar *DEFAULT-MAIL-DATE-STRING* "Yesterday")

(defvar *DEFAULT-MAIL-ATTRIBUTE-STRING* "")

(defvar *DEFAULT-KEYWORDS-FOR-FILTER* '())

(defflavor MAIL-FILTER-MENU
	   ((tv:item-list-pointer '*mail-filter-item-list*)
	    )
	   (tv:dynamic-momentary-menu
	    tv:menu-highlighting-mixin
	    )
  (:method-combination (:daemon-with-override :base-flavor-last :execute)))



;;;(defflavor MAIL-FILTER-MENU
;;;	   ()
;;;	   (w:menu)
;;;  (:default-init-plist 
;;;    :dynamic t
;;;    :item-list-pointer '*mail-filter-item-list*
;;;    :pop-up t
;;;    :abort-on-deexpose t
;;;    :permanent nil)
;;;  (:method-combination (:daemon-with-override :base-flavor-last :execute)))



(defmethod (mail-filter-menu :override :EXECUTE) (item &rest ignore)
  (and (consp item)
       (eq (third item) :not)
       :not))

	   
(w:defwindow-resource MAIL-FILTER-MENU ()
	:make-window (mail-filter-menu)
	:initial-copies 1
	:reusable-when :deexposed)


(defun CHOOSE-MAIL-FILTER ()
  
  (using-resource (menu mail-filter-menu *window*)
    (send menu :update-item-list)
    (send menu :set-highlighted-items nil)
    (send menu :set-label "Select Filter")
    ;; adding color conversion options - nn.
    (send menu :set-label-color      tv:*default-menu-label-foreground*)
    (send menu :set-label-background tv:*default-menu-label-background*)
    (send menu :set-geometry t t nil)
    ;; increase hysteresis because of many nags about the mouse accidently "slipping" out of pop up windows
    (when (< (send menu :hysteresis) 50)
      (send menu :set-hysteresis 50))
    (tv:expose-window-near menu '(:mouse))
    (let (invert-p choice)
      (loop
	(setq choice (send menu :choose))
	(cond ((eq choice :not)
	       (setq invert-p (not invert-p))
	       (if invert-p
		   (send menu :add-highlighted-value :not)
		   (send menu :remove-highlighted-value :not)))
	      ((consp choice)
	       (return (get-mail-filter (car choice) invert-p) t))
	      ((and choice (symbolp choice))
	       (return (get-mail-filter choice invert-p)))
	      (t
	       (return nil)))))))


(defun READ-MAIL-FILTER-FROM-MINI-BUFFER ()
  
  (let* ((*mini-buffer-default-string* *default-mail-filter*)
	 (*mini-buffer-dont-record* t)
	 (filter
	   (completing-read-from-mini-buffer
	     (format nil "Enter filter type: (Default: ~A) - Press Help for list of types."
		     *default-mail-filter*)
	     *mail-filter-completion-alist* nil)))
    (cond ((and (stringp filter) (zerop (length filter)))
	   (setq filter (cdr (assoc *default-mail-filter*
				    *mail-filter-completion-alist* :test #'string-equal))))
	  ((consp filter)
	   (setq *default-mail-filter* (car filter))
	   (setq filter (cdr filter))))
    (values (get-mail-filter filter) *numeric-arg-p*)))



(defun GET-MAIL-FILTER (id &optional inverse)
  
  (let ((filter (or (get id :user-mail-filter)
		    (get id :system-mail-filter))))
    (unless filter
      (ferror 'mail-reader-error
	      "~A has no filter definition (no :system-mail-filter or :user-mail-filter property found)"
	      id))
    (when inverse
      (let* ((inverse-id (send filter :inverse))
	     (inverse-filter (and inverse-id (or (get inverse-id :user-mail-filter)
						 (get inverse-id :system-mail-filter)))))
	(unless inverse-id
	  (barf "Filter ~A is not invertible." id))
	(unless inverse-filter
	  (ferror 'mail-reader-error "Filter ~A specified its inverse as ~A which does not exist." id inverse-id))
	(setq filter inverse-filter)))
    (if (send filter :dynamic-p)
	(apply (send filter :function) (send filter :args))
	filter)))


(defun MAKE-MAIL-BUFFER-FROM-FILTER (filter using-buffer &optional (reuse-query-p t))
  
  (let* ((name (format nil "(~A) ~A" (send filter :name) (buffer-name using-buffer)))
	 (buffer (find-buffer-named name))
	 (function (send filter :function))
	 (args (send filter :args))
	 (file-buffer (buffer-mail-file-buffer using-buffer)))
    (when (and buffer (not (message-sequence-p buffer)))
      (barf "There is already a buffer named ~A~% and it is not a mail buffer."))
    (when (and buffer (mail-file-buffer-p buffer))
      (barf "There is already a mail buffer named ~A~% which is associated with a file."))
    
    (when buffer
      (cond ((or (null reuse-query-p)
		 (y-or-n-p "Buffer already exists for ~A~%Reuse it?" buffer))
	     (clear-message-sequence buffer))
	    (t
	     (setq name (loop 
			  for i from 2
			  for bufnam = (format nil "(~A)~D ~A" (send filter :name) i (buffer-name using-buffer))
			  unless (find-buffer-named bufnam)
			  return bufnam))
	     (setq buffer nil))))
    
    (unless buffer
      (setq buffer (make-instance 'message-sequence :name name))
      (setf (buffer-mail-file-buffer buffer) file-buffer)
      (push-end buffer (buffer-subsequences file-buffer))
      (setf (get buffer :filters-used) (append (get using-buffer :filters-used) (list filter))))
    (let ((abort t))
      (unwind-protect
	  (progn
	    (domsgs (msg using-buffer)
	      (when (apply function msg args)
		(add-message-to-buffer msg buffer t)))
	    (setq abort nil))
	(when abort
	  (send buffer :kill))))
    buffer))


(defun REVERT-FILTER-BUFFER (buffer &optional (update-summary t))
  
  (let ((filter-list (get buffer :filters-used)))
    (when (and filter-list
	       (message-sequence-p buffer)
	       (not (mail-file-buffer-p buffer)))
      (clear-message-sequence buffer)
      (domsgs (msg (mail-file-buffer-of buffer))
	(when 
	  (dolist (filter filter-list t)
	    (unless (apply (send filter :function) msg (send filter :args))
	      (return nil)))
	  (add-message-to-buffer msg buffer t)))
      (when (and update-summary
		 (mail-summary-of buffer))
	(update-summary (mail-summary-of buffer)))))
  buffer)

;;; Filters/Functions for dealing with attributes

;; This exists only to reorder the arguments for the convenience of filters

(defsubst MSG-HAS-ATTRIBUTE-P (msg attribute)
  (message-attribute-p attribute msg))

(defsubst MSG-DOES-NOT-HAVE-ATTRIBUTE-P (msg attribute)
  (not (message-attribute-p attribute msg)))

(define-system-mail-filter UNSEEN
			   (:name "Unseen"
			    :function msg-has-attribute-p
			    :args (:unseen) 
			    :string-for-menu "Unseen"
			    :string-for-filter-summary "UNSEEN MESSAGES"
			    :filter-summary-default-p t
			    :inverse seen)
  "Select messages that are unseen.")
				  

(define-system-mail-filter SEEN
			   (:name "Seen"
			    :function msg-does-not-have-attribute-p
			    :args (:unseen)
			    :string-for-filter-summary "SEEN MESSAGES"))

(define-system-mail-filter DELETED
			   (:name "Deleted"
			    :function msg-has-attribute-p
			    :args (:deleted)
			    :string-for-menu "Deleted"
			    :string-for-filter-summary "DELETED MESSAGES"
			    :filter-summary-default-p t
			    :inverse not-deleted)
  "Select message that are deleted.")

(define-system-mail-filter NOT-DELETED
			   (:name "Not Deleted"
			    :function msg-does-not-have-attribute-p
			    :args (:deleted)
			    :string-for-filter-summary "NOT DELETED MESSAGES"))

(define-system-mail-filter RECENT
			   (:name "Recent"
			    :function msg-has-attribute-p
			    :args (:recent)
			    :string-for-menu "Recent"
			    :string-for-filter-summary "RECENT MESSAGES"
			    :filter-summary-default-p t
			    :inverse not-recent)
  "Select messages received since buffer last saved or expunged.")

(define-system-mail-filter NOT-RECENT
			   (:name "Not Recent"
			    :function msg-does-not-have-attribute-p
			    :args (:recent)
			    :string-for-filter-summary "NOT RECENT MESSAGES"))
			    

(define-system-mail-filter ANSWERED
			   (:name "Answered"
			    :function msg-has-attribute-p
			    :args (:answered)
			    :string-for-menu "Answered"
			    :string-for-filter-summary "ANSWERED MESSAGES"
			    :filter-summary-default-p t
			    :inverse not-answered)
  "Select messages that have been answered.")


(define-system-mail-filter NOT-ANSWERED
			   (:name "Unanswered"
			    :function  msg-does-not-have-attribute-p
			    :args (:answered)
			    :string-for-filter-summary "NOT ANSWERED MESSAGES"))

(define-system-mail-filter REMINDER
			   (:name "Reminder"
			    :function msg-reminder-p
			    :args ()
			    :string-for-menu "Reminder"
			    :string-for-filter-summary "REMINDER MESSAGES"
			    :filter-summary-default-p t
			    :inverse not-reminder)
  "Select messages that are reminders.")


(define-system-mail-filter NOT-REMINDER
			   (:name "Not Reminder"
			    :function  msg-not-reminder-p
			    :args ()
			    :string-for-filter-summary "NOT REMINDER MESSAGES"))


(defun msg-reminder-p (msg)
  ;; Try to cover various ways different systems do reminders
  (or (message-attribute-p :remind msg)
      (message-attribute-p :reminder msg)
      (message-attribute-p :remindme msg)
      (message-keyword-p :remind msg)
      (message-keyword-p :reminder msg)
      (message-keyword-p :remindme msg)
      (message-keyword-p :remindnow msg)))

(defun msg-not-reminder-p (msg)
  (not (msg-reminder-p msg)))

(define-system-mail-filter MARKED
			   (:name "Marked"
			    :function msg-has-attribute-p
			    :args (:apply)
			    :string-for-menu "Marked"
			    :string-for-filter-summary "MARKED MESSAGES"
			    :filter-summary-default-p t
			    :inverse not-marked)
  "Select messages that are marked for apply.")

(define-system-mail-filter NOT-MARKED
			   (:name "Not Marked"
			    :function  msg-does-not-have-attribute-p
			    :args (:apply)
			    :string-for-filter-summary "NOT MARKED MESSAGES"))

(define-system-mail-filter FILED
			   (:name "Filed"
			    :function msg-has-attribute-p
			    :args (:filed)
			    :string-for-menu "Filed"
			    :string-for-filter-summary "FILED MESSAGES"
			    :filter-summary-default-p t
			    :inverse not-filed)
  "Select messages that have been copied to another file.")

(define-system-mail-filter NOT-FILED
			   (:name "Not Filed"
			    :function  msg-does-not-have-attribute-p
			    :args (:filed)
			    :string-for-filter-summary "NOT FILED MESSAGES"))

;;; Filters/Functions for dealing with keywords


(define-system-mail-filter KEYWORD
			   (:name "Keywords"
			    :function make-keyword-filter
			    :dynamic-p t
			    :string-for-menu "Keywords"
			    :string-for-filter-summary "KEYWORDS"
			    :filter-summary-default-p t
			    :inverse not-keyword)
  "Select messages containing certain keywords.")


(defun MAKE-KEYWORD-FILTER ()
  
  (let ((choices (get-mail-option (mail-file-buffer-of *mail-buffer*) :labels)))
    (unless choices
      (barf "There are no keywords."))
;    (multiple-value-bind (keywords choice)
;	(choose-mail-keywords choices "Choose Keywords for Filter" *default-keywords-for-filter* '(:any :all))

;;  This part is for keyboard entry of keywords when doing keyboard commands (like other
;;  places that ask for keywords).  The only problem is that the user can pick nonexistent
;;  keywords if he presses return without waiting for completion.  However, the filtering
;;  stuff naturally won't find any messages that match that keyword, and there won't be
;;  any side-effects.

    (let (keywords choice)
      (cond (*mail-mouse-command*
	     (multiple-value-setq (keywords choice)
	       (choose-mail-keywords (copy-list choices) "Choose Keywords for Filter"
				     *default-keywords-for-filter* '(:any :all))))
	    (:else
	     (setq keywords (read-mail-keywords "Enter keywords for filter (separated by commas):"
						*default-keywords-for-filter*))
	     (barf-if-nonexistent-keywords keywords *mail-buffer*)
	     (setq choice (if (> (length keywords) 1)
			      (fquery '(:type :tyi :list-choices t :choices (((:any "Any") #\1 #\Space)
									     ((:all "All") #\2 #\Rubout)))
				      "Choose messages containing ANY of these keywords (1) or ALL of them (2) ")
			      :any))))
      (when choice
	(setq *default-keywords-for-filter* keywords)
	(let (name fsummary)
	  (cond ((null keywords)
		 (setq name (if (eq choice :all) "No Keywords" "Any Keyword"))
		 (setq fsummary (if (eq choice :all) "MESSAGES WITH NO KEYWORDS" "MESSAGES WITH ANY KEYWORD")))
		((= (length keywords) 1)
		 (setq name (format nil "Keyword: ~A" (car keywords)))
		 (setq fsummary (format nil "MESSAGES WITH KEYWORD: ~A" (car keywords))))
		(t
		 (setq name (format nil "~AKeywords: ~{~A~^ ~}" (if (eq choice :all) "" "Any Of ") keywords))
		 (setq fsummary (format nil "MESSAGES WITH ~AKEYWORDS: ~{~A~^ ~}"
					(if (eq choice :all) "" "ANY OF ") keywords))))
	  (make-instance 'system-mail-filter
			 :name name
			 :function (if (eq choice :all) 'keyword-and-filter 'keyword-or-filter)
			 :args (list keywords)
			 :string-for-filter-summary fsummary))))))


(defun BARF-IF-NONEXISTENT-KEYWORDS (keywords buffer)
   (let ((bad-keywords (set-difference keywords
				       (if *choose-from-all-mail-keywords-p*
					   *all-mail-keywords*
					   (get-mail-option (mail-file-buffer-of buffer) :labels)))))
     (when bad-keywords
       (barf "The keyword~P ~{~A~^, ~} ~:[are~;is~] not available in this buffer."
	     (length bad-keywords)
	     bad-keywords
	     (= (length bad-keywords) 1)))))


(defun KEYWORD-AND-FILTER (msg keywords)
  
  ;; If KEYWORDS is nil and MSG contains no keywords, T is returned
  (if (null keywords)
      (null (message-keywords msg))
      (dolist (key keywords t)
	(unless (message-keyword-p key msg)
	  (return nil)))))


(defun KEYWORD-OR-FILTER (msg keywords)
  
  ;; If KEYWORDS is nil and MSG contains no keywords, nil is returned
  (if (null keywords)
      (not (null (message-keywords msg)))
      (dolist (key keywords nil)
	(when (message-keyword-p key msg)
	  (return t)))))


(define-system-mail-filter NOT-KEYWORD
			   (:name "Not Keywords"
			    :function make-inverse-keyword-filter
			    :dynamic-p t))



(defun MAKE-INVERSE-KEYWORD-FILTER ()
  
  (let ((choices (get-mail-option (mail-file-buffer-of *mail-buffer*) :labels)))
    (unless choices
      (barf "There are no keywords in this buffer."))
;    (multiple-value-bind (keywords choice)
;	(choose-mail-keywords choices "Choose Keywords for Filter" *default-keywords-for-filter* '(:do-it))
    (let (keywords choice)
      (cond (*mail-mouse-command*
	     (multiple-value-setq (keywords choice)
	       (choose-mail-keywords choices "Choose Keywords for Filter" *default-keywords-for-filter* '(:do-it))))
	    (:else
	     (setq keywords (read-mail-keywords "Enter keywords for filter (separated by commas):"
						*default-keywords-for-filter*))
	     (barf-if-nonexistent-keywords keywords *mail-buffer*)
	     (setq choice t)))

      (when choice
	(setq *default-keywords-for-filter* keywords)
	(let (name fsummary)
	  (cond ((null keywords)
		 (setq name "No Keywords")
		 (setq fsummary "MESSAGES WITH NO KEYWORDS"))
		((= (length keywords) 1)
		 (setq name (format nil "Without Keyword: ~A" (car keywords)))
		 (setq fsummary (format nil "MESSAGES WITHOUT KEYWORD: ~A" (car keywords))))
		(t
		 (setq name (format nil "None of Keywords: ~{~A~^ ~}" keywords))
		 (setq fsummary (format nil "MESSAGES WITH NONE OF KEYWORDS: ~{~A~^ ~}" keywords))))
	  (make-instance 'system-mail-filter
			 :name name
			 :function 'keyword-not-filter
			 :args (list keywords)
			 :string-for-filter-summary fsummary))))))


(defun KEYWORD-NOT-FILTER (msg keywords)
  
  (if (null keywords)
      (null (message-keywords msg))
    (dolist (key keywords t)
      (when (message-keyword-p key msg)
	(return nil)))))



;;; Filters/Functions for dealing with text search


(define-system-mail-filter SEARCH
			   (:name "Search"
			    :function make-search-filter
			    :args (nil)
			    :dynamic-p t
			    :string-for-menu "Search"
			    :string-for-filter-summary "SEARCH"
			    :filter-summary-default-p t
			    :inverse not-search)
  "Select messages containing a certain string.")

(define-system-mail-filter NOT-SEARCH
			   (:name "Not Search"
			    :function make-search-filter
			    :args (t)
			    :dynamic-p t))

(defun MAKE-SEARCH-FILTER (&optional invert)
  
  (multiple-value-bind (search-function search-key)
      (get-extended-string-search-strings nil
					  (if invert
					      "Select Messages Not Containing Text:"
					      "Select Messages Containing Text:")
					  *search-mini-buffer-comtab*
					  *default-mail-search-string*)
    (setq *default-mail-search-string* *extended-string-search-last-string*)
    
    (make-instance 'system-mail-filter
		   :name (format nil "~@[~*Not ~]Search: ~A" invert search-key)
		   :string-for-filter-summary (format nil "MESSAGES ~@[~*NOT ~]CONTAINING TEXT: ~A" invert search-key)
		   :function (if invert 'msg-does-not-have-search-string 'msg-has-search-string)
		   :args (list search-function search-key))))
		   
(defun MSG-HAS-SEARCH-STRING (msg search-function search-key)
  (and (funcall search-function (interval-first-bp msg) search-key nil nil nil (interval-last-bp msg))
       t))

(defun MSG-DOES-NOT-HAVE-SEARCH-STRING (msg search-function search-key)
  (not (funcall search-function (interval-first-bp msg) search-key nil nil nil (interval-last-bp msg))))



;;; Filters/Functions for dealing with message headers

(defun MAKE-HEADER-SEARCH-FILTER (&optional name header-types invert)
  
  (unless header-types
    (let* ((*mini-buffer-default-string* *default-header-search-string*)
	   (header
	     (typein-line-readline "Search within header field: (Default is ~S)" *default-header-search-string*)))
      (if (zerop (length header))
	  (setq header *default-header-search-string*)
	  (setq *default-header-search-string* header))
      (unless (position #\: header)
	(setq header (string-append header ":")))
      (setq header-types (list (or (mail:header-line-type header) (barf "~A is not a valid header."))))))
  
  (unless name
    (setq name (string-capitalize (format nil "~@[~*Not ~]~{~A~^/~}:" invert header-types))))
  
  (multiple-value-bind (search-function search-key)
      (get-extended-string-search-strings nil
					  (format nil "Select messages ~@[~*not ~]containing string in ~{~A~^/~}:"
						  invert header-types)
					  *search-mini-buffer-comtab*
					  *default-mail-search-string*)
    (setq *default-mail-search-string* *extended-string-search-last-string*)
    
    (make-instance 'system-mail-filter
		   :name (format nil "~A ~S" name search-key)
		   :function (if invert 'msg-header-does-not-have-search-string 'msg-header-has-search-string)
		   :args (list header-types search-function search-key)
		   :string-for-filter-summary (format nil "MESSAGES ~A ~A"
						      (string-upcase name) search-key))))


(define-system-mail-filter SUBJECT
			   (:name "Subject"
			    :function make-header-search-filter
			    :args ("Subject:" (:SUBJECT))
			    :dynamic-p t
			    :string-for-filter-summary "MESSAGES WITH SUBJECT"
			    :filter-summary-default-p t
			    :string-for-menu "Subject"
			    :inverse not-subject)
  "Select messages containing a certain string in their Subject: field.")

(define-system-mail-filter NOT-SUBJECT
			   (:name "Not Subject"
			    :function make-header-search-filter
			    :args ("Not Subject:" (:SUBJECT) t)
			    :dynamic-p t))


(define-system-mail-filter FROM
			   (:name "From"
			    :function make-header-search-filter
			    :args ("From:" (:FROM))
			    :dynamic-p t
			    :string-for-filter-summary "MESSAGES FROM"
			    :filter-summary-default-p t
			    :string-for-menu "From"
			    :inverse not-from)
  "Select messages containing a certain string in their From: field.")


(define-system-mail-filter NOT-FROM
			   (:name "Not From"
			    :function make-header-search-filter
			    :args ("Not From:" (:FROM) t)
			    :dynamic-p t))

(define-system-mail-filter TO/CC
			   (:name "To/CC"
			    :function make-header-search-filter
			    :args ("To/CC:" (:TO :CC))
			    :dynamic-p t
			    :string-for-filter-summary "MESSAGES TO/CC"
			    :filter-summary-default-p t
			    :string-for-menu "To/CC"
			    :inverse not-to/cc)
  "Select messages containing a certain string in their To: or CC: field.")


(define-system-mail-filter NOT-TO/CC
			   (:name "Not To/CC"
			    :function make-header-search-filter
			    :args ("Not To/cc:" (:TO :CC) t)
			    :dynamic-p t))


(define-system-mail-filter HEADER
			   (:name "Header"
			    :function make-header-search-filter
			    :args nil
			    :dynamic-p t
			    :string-for-menu "Header"
			    :inverse not-header)
  "Select messages containing a certain string in a certain header field.")


(define-system-mail-filter NOT-HEADER
			   (:name "Not Header"
			    :function make-header-search-filter
			    :args (nil nil t)
			    :dynamic-p t))


(defun MSG-HEADER-HAS-SEARCH-STRING (msg header-types search-function search-key)
  
  (assure-message-parsed msg)
  (dolist (type header-types)
    (dolist (header-status (get-message-header-all msg type))
      (let ((*interval* (getf header-status :interval)))
	(when *interval*
	  (when (funcall search-function (interval-first-bp *interval*) search-key)
	    (return-from msg-header-has-search-string t))))))
  
  nil)


(defun MSG-HEADER-DOES-NOT-HAVE-SEARCH-STRING (msg header-types search-function search-key)
  
  (assure-message-parsed msg)
  (dolist (type header-types)
    (dolist (header-status (get-message-header-all msg type))
      (let ((*interval* (getf header-status :interval)))
	(when *interval*
	  (when (funcall search-function (interval-first-bp *interval*) search-key)
	    (return-from msg-header-does-not-have-search-string nil))))))
  
  t)


;;;
;;; Filters for dealing with dates
;;;
	  
;;; This needs some smarts to handle reletive dates in a relative way.

;;; The smallest resolution is forced to one day by making the time for
;;; all dates considered be midnight.  This seems good because when you
;;; ask for messages before yesterday, the current time (which is
;;; supplied by default from parse-universal-time when no time is
;;; specified) will not cause unexpected results


(define-system-mail-filter AFTER-DATE
			   (:name "After"
			    :function make-after-date-filter
			    :dynamic-p t
			    :string-for-menu "After"
			    :string-for-filter-summary "MESSAGES AFTER DATE"
			    :filter-summary-default-p t)
  "Select message dated after a certain date.")

(defun MAKE-AFTER-DATE-FILTER ()

  (multiple-value-bind (date-string utime)
      (read-date-from-minibuffer "Select messages dated after:")
    (make-instance 'system-mail-filter
		   :name (string-append "After: " date-string)
		   :string-for-filter-summary (string-append "MESSAGES DATED AFTER: " date-string)
		   :function 'msg-date-greaterp
		   :args (list utime))))


(defun MSG-DATE-GREATERP (msg target-utime)
   (assure-message-parsed msg)
   (let ((msg-utime (get-message-header msg :date :universal-time)))
     (when (numberp msg-utime)
       (> msg-utime target-utime))))


(define-system-mail-filter BEFORE-DATE
			   (:name "Before"
			    :function make-before-date-filter
			    :dynamic-p t
			    :string-for-menu "Before"
			    :string-for-filter-summary "MESSAGES BEFORE DATE"
			    :filter-summary-default-p t)

  "Select message dated before a certain date.")

(defun MAKE-BEFORE-DATE-FILTER ()

  (multiple-value-bind (date-string utime)
      (read-date-from-minibuffer "Select messages dated before:")
    (make-instance 'system-mail-filter
		   :name (string-append "Before: " date-string)
		   :string-for-filter-summary (string-append "MESSAGES DATED BEFORE: " date-string)
		   :function 'msg-date-lessp
		   :args (list utime))))


(defun MSG-DATE-LESSP (msg target-utime)
   (assure-message-parsed msg)
   (let ((msg-utime (get-message-header msg :date :universal-time)))
     (when (numberp msg-utime)
       (< msg-utime target-utime))))


(define-system-mail-filter ON-DATE
			   (:name "On"
			    :function make-on-date-filter
			    :dynamic-p t
			    :string-for-menu "On"
			    :string-for-filter-summary "MESSAGES ON DATE"
			    :filter-summary-default-p t)
  "Select message dated on a certain date.")


(defun MAKE-ON-DATE-FILTER ()

  (multiple-value-bind (date-string utime)
      (read-date-from-minibuffer "Select messages dated:")
    (make-instance 'system-mail-filter
		   :name (string-append "On: " date-string)
		   :string-for-filter-summary (string-append "MESSAGES DATED ON: " date-string)
		   :function 'msg-date-samep
		   :args (list utime))))


(defun MSG-DATE-SAMEP (msg target-utime)
   (assure-message-parsed msg)
   (let ((msg-utime (get-message-header msg :date :universal-time)))
     (when (numberp msg-utime)
       ;; SPR#6170:  compare times after setting msg's hours and minutes to zero   \/\/,nn 09/87
       (= (set-utime-to-zero-hour msg-utime) target-utime))))


(defun SET-UTIME-TO-ZERO-HOUR (utime)
  (multiple-value-bind (nil nil nil date month year nil nil nil)
      (time:decode-universal-time utime)
    (time:encode-universal-time 0 0 0 date month year)))


(defun READ-DATE-FROM-MINIBUFFER (prompt)
  
  (let* ((*mini-buffer-default-string* *default-mail-date-string*)
	 (string (typein-line-readline "~A (Default is ~S)" prompt *default-mail-date-string*)))
    (if (zerop (length string))
	(setq string *default-mail-date-string*)
	(setq *default-mail-date-string* string))
    (condition-case (condition)
	(values string (time:parse-universal-time string 0 nil nil (set-utime-to-zero-hour (get-universal-time))))
      (time:parse-error 
       (barf (send condition :report-string))))))


(define-system-mail-filter LAST-N-DAYS
			   (:name "Last-N-Days"
			    :function make-last-n-days-filter
			    :dynamic-p t
			    :string-for-menu "Last-N-Days"
			    :string-for-filter-summary ""
			    :filter-summary-default-p t)
  "Select all messages received in the last N days.")

(defun MAKE-LAST-N-DAYS-FILTER ()
  (multiple-value-bind (n-days)
      (read-number-from-minibuffer "Select all messages in the last N days?" 1)
    (make-instance 'system-mail-filter
		   :name (string-append "Last " n-days " days")
		   :string-for-filter-summary (format nil "LAST ~:[~d DAYS~;DAY~] OF MESSAGES" 
                                                          (EQL n-days 1) n-days)
		   :function 'msg-date-n-days
		   :args (list n-days))))

(defun MSG-DATE-N-DAYS (msg n-days)
   (assure-message-parsed msg)
   (let ((msg-utime (get-message-header msg :date :universal-time)))
     (when (numberp msg-utime)
       (multiple-value-bind (ignore ignore ignore cur-date cur-mo cur-yr)
	     (decode-universal-time (get-universal-time))
	 (let ((msg-zero-hour-utime (set-utime-to-zero-hour msg-utime))
	       (n-days-ago-utime (encode-universal-time 0 0 0 (- cur-date n-days) cur-mo cur-yr)))
	   (> msg-zero-hour-utime n-days-ago-utime))))))

(defun READ-NUMBER-FROM-MINIBUFFER (prompt default)
  (let ((n-days (typein-line-read "~A (Default is ~d)" prompt default)))
    (if (or (not (integerp n-days))
	    (eq n-days '*eof*))
	default
	n-days)))
