;;; -*- Mode:Common-Lisp; Package:addin; Base:10.; Cold-Load: T; -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1987- 1989 Texas Instruments Incorporated. All rights reserved.

;;; This file contains definitions for the ADDIN COMMUNICATION BLOCK (ACB) data structure

;;;
;;; Edit History
;;;
;;;                   Patch
;;;   Date    Author  Number   Description
;;;--------------------------------------------------------------------
;;; 09-87     bj/ab     --     o Original
;;; 10-07-87    ab             o GET-PARM-STRING support.
;;;                              Made PARM-nB accessors INLINEs and made SETF-able.
;;;                              WITH-UNLOADED-PARMS-nB macros.
;;; 11-15-87    ab     *5.3  o New ACB allocation queue support.
;;; 11-27-87    ab     *6.2  o Fix SET-RECEIVER-COMPLETE & friends to do only
;;;                            byte accesses so the other byte of status isn't written
;;;                            at the same time
;;; 12-14-87    ab      --   o Support having the first physical memory chunk 
;;;                            allocated by the microcode.  Call INIT-SPECIAL-ACBS from 
;;;                            the ACB-Q init routine.
;;; 12-16-87    ab      --   o Clean up compile warnings.
;;; 12-22-87    ab     *8.2  o Set boot status to indicate ACB allocation has occurred.
;;; 01-07-87    bj    *10.1  o GET-ACB-FAST fixes for device driver implementation.
;;; 02-08-88    ab     *4.1  o Remove hard-coded slot check in Q-LINK.  Make Q-MEMBER
;;;                            check conditional on new *q-debug* variable.
;;; 02-08-88    ab     *4.2  o Use leader slot for storing ACB physical address to avoid
;;;                            excessive bignum consing.
;;; 03-02-88    ab     *5.2  o Set *open-acbs* list to NIL before re-initializing ACB-RESOURCE.
;;;                            This fixes occasional error after disk-save about not being
;;;                            able to return an ACB to the ACB-RESOURCE.
;;; 07-26-88	jjp	      o Change initialize-acb-qs to reset the qs during
;;;				warm boot and stop the ethernet controller board
;;;				from writing to it's cache of buffers from befor
;;;				the warm boot. (Those acbs are invalid)
;;; 08-03-88    ab     --    o Moved some ACB init fns to ACB-FUNCTIONS.  These did not
;;;                            get moved in the new-queueing integration.
;;; 08-05-88    ab           o Added a 32K allocation queue.
;;; 04-28-89   jlm           o Changed name of TYPE to ACB-TYPE to remove function def from LISP::TYPE
;;; 05/10/89   jlm           o Changed SET-TYPE to be SET-ACB-TYPE and fix DEFSETF for ACB-TYPE 

;;; 
;;; Variables
;;;

(proclaim '(special si:%words-per-addin-queue si:%addin-max-allocation-queues
		    si:addin-initial-chunk-pages si:micronet-reserved-pages
		    si:%driver-data-start
		    ))



(defvar *micronet-debug* nil)

(defvar *acb-leader-type-identifier* :%%%%acb%%%% "A keyword that is put on the array leader to use when type checking.")

(DEFPARAMETER *addin-acb-good-size-list*
	      '(244. 256. 500. 512. 2048. 8192. 16384. 104448.)
  "List of possible sizes for ACBs.  Sorted in ascending order.")

(DEFCONSTANT *addin-largest-acb* (FIRST (LAST *addin-acb-good-size-list*)))

(defconstant *ACB-OVERHEAD-BYTES* si:%acb-parm-0-byte)

;;(defconstant *END-OF-QUEUE-MARK* 0)

;; Default ACB status:
;;   Block-Deallocatable = NIL
;;   Requestor-Complete = T
;;   Input-Complete = T
;;   Command-Complete = T
(defconstant *ACB-STATUS-READY*
	     (dpb 1 si:%%acb-bd-bit-in-stat-hw
		  (dpb 0 si:%%acb-rc-bit-in-stat-hw
		       (dpb 0 si:%%acb-ic-bit-in-stat-hw
			    (dpb 0 si:%%acb-sc-bit-in-stat-hw 0)))))



;;;
;;; ACB Cmd Block Accessors
;;;

(proclaim '(inline opcode))
(defun OPCODE (acb)
  (ldb si:%%acb-opcode-in-op-hw
       (aref acb si:%acb-opcode-halfword)))

(proclaim '(inline set-opcode))
(defun SET-OPCODE (acb value)
  (setf (aref acb si:%acb-opcode-halfword)
	(dpb value
	     si:%%acb-opcode-in-op-hw
	     (aref acb si:%acb-opcode-halfword))))

(defsetf OPCODE set-opcode)


(proclaim '(inline subopcode))
(defun SUBOPCODE (acb)
  (ldb si:%%acb-subopcode-in-subop-hw
       (aref acb si:%acb-subopcode-halfword)))

(proclaim '(inline set-subopcode))
(defun SET-SUBOPCODE (acb value)
  (setf (aref acb si:%acb-subopcode-halfword)
	(dpb value
	     si:%%acb-subopcode-in-subop-hw
	     (aref acb si:%acb-subopcode-halfword))))

(defsetf SUBOPCODE set-subopcode)


(proclaim '(inline error-code))
(defun ERROR-CODE (acb)
  (ldb si:%%acb-error-code-in-err-hw 
       (aref acb si:%acb-error-halfword)))

(proclaim '(inline set-error-code))
(defun SET-ERROR-CODE (acb value)
  (setf (aref acb si:%acb-error-halfword)
	(dpb value
	     si:%%acb-error-code-in-err-hw 
	     (aref acb si:%acb-error-halfword))))

(defsetf ERROR-CODE set-error-code)


(proclaim '(inline status))
(defun STATUS (acb)
  (aref acb si:%acb-status-halfword))

(proclaim '(inline set-status))
(defun SET-STATUS (acb status)
  (setf (aref acb si:%acb-status-halfword) status))

(defsetf STATUS set-status)


(proclaim '(inline address))
(defun ADDRESS (acb)
  "Returns the physical address of ACB."
  (si:array-data-buffer-address acb))


;;ab 2/10/88.  Store physical address in ACB leader to avoid bignum consing.
(proclaim '(inline physical-address))
(defun PHYSICAL-ADDRESS (acb &aux adr)
  (cond ((setq adr (array-leader acb si:%acb-leader-physical-address))
	 adr)
	(t
	 (setf (array-leader acb si:%acb-leader-physical-address)
	       (address acb)) 
	 (array-leader acb si:%acb-leader-physical-address))))

(proclaim '(inline set-physical-address))
(defun SET-PHYSICAL-ADDRESS (acb phys-addr)
  (setf (array-leader acb si:%acb-leader-physical-address) phys-addr))

(defsetf PHYSICAL-ADDRESS set-physical-address)


(proclaim '(inline requestor-complete))
(defun REQUESTOR-COMPLETE (acb)
  (ldb-test si:%%acb-rc-bit-in-stat-hw
	    (aref acb si:%acb-status-halfword)))

(proclaim '(inline set-requestor-complete))
(defun SET-REQUESTOR-COMPLETE (acb t-or-nil)
  (let* ((phys-adr (physical-address acb))
	 (slot (ldb si:%%nubus-f-and-slot-bits phys-adr))
	 (offset (+ (ldb si:%%nubus-all-but-f-and-slot-bits phys-adr)
		    si:%acb-status-rc-byte)))
    (si:%nubus-write-8b
      slot offset (dpb (if t-or-nil 1 0)
		       si:%%acb-rc-bit-in-rc-byte
		       (si:%nubus-read-8b slot offset))))
  )

(defsetf REQUESTOR-COMPLETE set-requestor-complete)


(proclaim '(inline servicer-complete))
(defun SERVICER-COMPLETE (acb)
  (ldb-test si:%%acb-sc-bit-in-stat-hw
	    (aref acb si:%acb-status-halfword)))

(proclaim '(inline set-servicer-complete))
(defun SET-SERVICER-COMPLETE (acb t-or-nil)
  (let* ((phys-adr (physical-address acb))
	 (slot (ldb si:%%nubus-f-and-slot-bits phys-adr))
	 (offset (+ (ldb si:%%nubus-all-but-f-and-slot-bits phys-adr)
		    si:%acb-status-sc-byte)))
    (si:%nubus-write-8b
      slot offset (dpb (if t-or-nil 1 0)
		       si:%%acb-sc-bit-in-sc-byte
		       (si:%nubus-read-8b slot offset))))
  )


(defmacro COMMAND-COMPLETE (acb)
  `(servicer-complete ,acb))

(defsetf SERVICER-COMPLETE set-servicer-complete)


(proclaim '(inline input-complete))
(defun INPUT-COMPLETE (acb)
  (ldb-test si:%%acb-ic-bit-in-stat-hw
	    (aref acb si:%acb-status-halfword)))

(proclaim '(inline set-input-complete))
(defun SET-INPUT-COMPLETE (acb t-or-nil)
  (let* ((phys-adr (physical-address acb))
	 (slot (ldb si:%%nubus-f-and-slot-bits phys-adr))
	 (offset (+ (ldb si:%%nubus-all-but-f-and-slot-bits phys-adr)
		    si:%acb-status-sc-byte)))
    (si:%nubus-write-8b
      slot offset (dpb (if t-or-nil 1 0)
		       si:%%acb-ic-bit-in-sc-byte
		       (si:%nubus-read-8b slot offset))))
  )

(defsetf INPUT-COMPLETE set-input-complete)


(proclaim '(inline deallocatable))
(defun DEALLOCATABLE (acb)
  (ldb-test si:%%acb-bd-bit-in-stat-hw
	    (aref acb si:%acb-status-halfword)))

(proclaim '(inline set-deallocatable))
(defun SET-DEALLOCATABLE (acb t-or-nil)
  (let* ((phys-adr (physical-address acb))
	 (slot (ldb si:%%nubus-f-and-slot-bits phys-adr))
	 (offset (+ (ldb si:%%nubus-all-but-f-and-slot-bits phys-adr)
		    si:%acb-status-rc-byte)))
    (si:%nubus-write-8b
      slot offset (dpb (if t-or-nil 1 0)
		       si:%%acb-bd-bit-in-rc-byte
		       (si:%nubus-read-8b slot offset))))
  )


(defsetf DEALLOCATABLE set-deallocatable)


(proclaim '(inline parm-bytes))
(defun PARM-BYTES (acb)
  (dpb (aref acb si:%acb-parm-length-halfword-hi)
       (byte 16 16)
       (aref acb si:%acb-parm-length-halfword-lo)))

(proclaim '(inline set-parm-bytes))
(defun SET-PARM-BYTES (acb num-bytes)
  (setf (aref acb si:%acb-parm-length-halfword-lo)
	(ldb (byte 16 0) num-bytes))
  (setf (aref acb si:%acb-parm-length-halfword-hi)
	(ldb (byte 16 16) num-bytes))
  num-bytes)

(defsetf PARM-BYTES set-parm-bytes)


(proclaim '(inline data-bytes))
(defun DATA-BYTES (acb)
  (dpb (aref acb si:%acb-data-bytes-halfword-hi)
       (byte 16 16)
       (aref acb si:%acb-data-bytes-halfword-lo)))

(proclaim '(inline set-data-bytes))
(defun SET-DATA-BYTES (acb num-bytes)
  (setf (aref acb si:%acb-data-bytes-halfword-lo)
	(ldb (byte 16 0) num-bytes))
  (setf (aref acb si:%acb-data-bytes-halfword-hi)
	(ldb (byte 16 16) num-bytes))
  num-bytes)

(defsetf DATA-BYTES set-data-bytes)


(proclaim '(inline queue-link))
(defun QUEUE-LINK (acb)
  (let* ((phys-adr (physical-address acb))
	 (slot (ldb si:%%nubus-f-and-slot-bits phys-adr))
	 (offset (+ (ldb si:%%nubus-all-but-f-and-slot-bits phys-adr) %ACB-link-Word)))
    (%nubus-read slot offset)))

(proclaim '(inline set-queue-link))
(defun SET-QUEUE-LINK (acb obj)
  (let ((tem (si:%pointer obj)))
    (setf (aref acb si:%acb-link-halfword-lo)
	  (ldb si:%%q-low-half tem))
    (setf (aref acb si:%acb-link-halfword-hi)
	  (si:%logldb
	    (byte (- (byte-size si:%%q-pointer) (byte-size si:%%q-low-half))
		  (byte-position si:%%q-high-half))
	    tem))
    tem))

(defsetf QUEUE-LINK set-queue-link)


(proclaim '(inline backlink))
(defun BACKLINK (acb)
  (si:%logdpb
    (ldb (byte (- (byte-size si:%%q-pointer) (byte-size si:%%q-low-half)) 0)
	 (aref acb si:%acb-array-backlink-halfword-hi))
    (byte (- (byte-size si:%%q-pointer) (byte-size si:%%q-low-half))
	  (byte-position si:%%q-high-half))
    (aref acb si:%acb-array-backlink-halfword-lo))
  )

(proclaim '(inline set-backlink))
(defun SET-BACKLINK (acb obj)
  (let ((tem (si:%pointer obj)))
    (setf (aref acb si:%acb-array-backlink-halfword-lo)
	  (ldb si:%%q-low-half tem))
    (setf (aref acb si:%acb-array-backlink-halfword-hi)
	  (si:%logldb
	    (byte (- (byte-size si:%%q-pointer) (byte-size si:%%q-low-half))
		  (byte-position si:%%q-high-half))
	    tem))
    obj))

(defsetf BACKLINK set-backlink)

(proclaim '(inline source-port))
(defun source-port (acb)
  (aref acb si:%ACB-Source-Port-Halfword))

(proclaim '(inline set-source-port))
(defun set-source-port (acb value)
  (setf (aref acb si:%ACB-Source-Port-Halfword) value))

(defsetf source-port set-source-port)

(proclaim '(inline source-slot))
(defun source-slot (acb)
  (aref acb si:%ACB-Source-slot-Halfword))

(proclaim '(inline set-source-slot))
(defun set-source-slot (acb value)
  (setf (aref acb si:%ACB-Source-slot-Halfword) value))

(defsetf source-slot set-source-slot)

(proclaim '(inline destination-port))
(defun destination-port (acb)
  (aref acb si:%ACB-destination-Port-Halfword))

(proclaim '(inline set-destination-port))
(defun set-destination-port (acb value)
  (setf (aref acb si:%ACB-destination-Port-Halfword) value))

(defsetf destination-port set-destination-port)

(proclaim '(inline destination-slot))
(defun destination-slot (acb)
  (aref acb si:%ACB-destination-slot-Halfword))

(proclaim '(inline set-destination-slot))
(defun set-destination-slot (acb value)
  (setf (aref acb si:%ACB-destination-slot-Halfword) value))

(defsetf destination-slot set-destination-slot)


;;;
;;; ACB Array Leader Info Accessors
;;;

(proclaim '(inline info))
(defun INFO (acb)
  (array-leader acb si:%acb-leader-info))

(proclaim '(inline set-info))
(defun SET-INFO (acb value)
  (setf (array-leader acb si:%acb-leader-info) value))

(defsetf INFO set-info)


(proclaim '(inline acb-type))	; jlm 4/28/89
(defun acb-TYPE (acb)		; jlm 4/28/89
  (array-leader acb si:%acb-leader-type))

(proclaim '(inline set-acb-type))		; jlm 5/10/89
(defun SET-ACB-TYPE (acb type)
  (setf (array-leader acb si:%acb-leader-type) type))

(defsetf ACB-TYPE set-acb-type)			; jlm 5/10/89


(proclaim '(inline owner))
(defun OWNER (acb)
  (array-leader acb si:%acb-leader-owner))

(proclaim '(inline set-owner))
(defun SET-OWNER (acb obj)
  (setf (array-leader acb si:%acb-leader-owner) obj))

(defsetf OWNER set-owner)


(proclaim '(inline bytes-requested))
(defun BYTES-REQUESTED (acb)
  (array-leader acb si:%acb-leader-data-bytes-requested))

(proclaim '(inline set-bytes-requested))
(defun SET-BYTES-REQUESTED (acb num-bytes)
  (setf (array-leader acb si:%acb-leader-data-bytes-requested) num-bytes))

(defsetf BYTES-REQUESTED set-bytes-requested)


(proclaim '(inline parm-buf-string))
(defun PARM-BUF-STRING (acb)
  (array-leader acb si:%acb-leader-parm-array-string))

(proclaim '(inline set-parm-buf-string))
(defun SET-PARM-BUF-STRING (acb string)
  (setf (array-leader acb si:%acb-leader-parm-array-string) string))

(defsetf PARM-BUF-STRING set-parm-buf-string)


(proclaim '(inline parm-buf-8b))
(defun PARM-BUF-8B (acb)
  (array-leader acb si:%acb-leader-parm-array-8b))

(proclaim '(inline set-parm-buf-8b))
(defun SET-PARM-BUF-8B (acb ary)
  (setf (array-leader acb si:%acb-leader-parm-array-8b) ary))

(defsetf PARM-BUF-8B set-parm-buf-8b)


(proclaim '(inline parm-buf-16b))
(defun PARM-BUF-16B (acb)
  (array-leader acb si:%acb-leader-parm-array-16b))

(proclaim '(inline set-parm-buf-16b))
(defun SET-PARM-BUF-16B (acb ary)
  (setf (array-leader acb si:%acb-leader-parm-array-16b) ary))

(defsetf PARM-BUF-16B set-parm-buf-16b)


(proclaim '(inline parm-buf-32b))
(defun PARM-BUF-32B (acb)
  (array-leader acb si:%acb-leader-parm-array-32b))

(proclaim '(inline set-parm-buf-32b))
(defun SET-PARM-BUF-32B (acb ary)
  (setf (array-leader acb si:%acb-leader-parm-array-32b) ary))

(defsetf PARM-BUF-32B set-parm-buf-32b)


;;; 
;;; ACB Parm Accessors
;;;

(proclaim '(inline 16b-parm-number))
(defun 16B-PARM-NUMBER (byte-number &optional (byte-offset 0))
  "Given an 8b parm number (BYTE-NUMBER), returns the corresponding 16b parm number."
  (floor (+ byte-offset byte-number) 2))

(proclaim '(inline 32b-parm-number))
(defun 32B-PARM-NUMBER (byte-number &optional (byte-offset 0))
  "Given an 8b parm number (BYTE-NUMBER), returns the corresponding 32b parm number."
  (floor (+ byte-offset byte-number) 4))


(proclaim '(inline num-parm-bytes))
(defun NUM-PARM-BYTES (acb)
  "Returns the number of bytes of parameters ACB is sized to hold."
  (array-total-size (parm-buf-8b acb)))

(proclaim '(inline num-parm-halfwords))
(defun NUM-PARM-HALFWORDS (acb)
  "Returns the number of 16-bit halfwords of parameters ACB is sized to hold."
  (array-total-size (parm-buf-16b acb)))

(proclaim '(inline num-parm-words))
(defun NUM-PARM-WORDS (acb)
  "Returns the number of 32-bit words of parameters ACB is sized to hold."
  (array-total-size (parm-buf-32b acb)))


;;; General PARM accessor macros

(defmacro PARM-BLOCK-ACCESSOR (acb element-size)
  "Return the displaced array from the array leader of an ACB that accesses elements of size element-size."
  `(array-leader ,acb
		 ,(cond ((integerp element-size)
			 (case element-size
			   (8.  si:%acb-leader-parm-array-8b)
			   (16. si:%acb-leader-parm-array-16b)
			   (32. si:%acb-leader-parm-array-32b)
			   (t (ferror nil "~d is not a legal element-size." element-size))))
			((eq element-size :string)
			 si:%acb-leader-parm-array-string)
			(t (ferror nil "~d is not a legal element-size." element-size)))))


(defmacro READ-PARM (acb n element-size)
  "Return the nth parameter of size element-size from acb."
  `(aref (parm-block-accessor ,acb ,element-size) ,n))

(defmacro SET-PARM (acb n value element-size)
  "Set the nth parameter of size element-size in acb to value."
  `(setf (aref (parm-block-accessor ,acb ,element-size) ,n) ,value))

(defmacro COPY-PARMS (acb array acb-start array-start number-of-parms direction element-size)
  "Copy NUMBER-OF-PARMS elements of size ELEMENT-SIZE between acb and array.
   Values for direction are :TO-ACB and :TO-ARRAY.  ACB-START and ARRAY-START are the starting
elements of ACB and ARRAY for the copy."
  (once-only (acb-start array-start number-of-parms)
    (case direction
      (:to-array
       `(copy-array-portion (parm-block-accessor ,acb ,element-size) ,acb-start (+ ,acb-start ,number-of-parms)
			    ,array ,array-start (+ ,array-start ,number-of-parms)))    
      (:to-acb
       `(copy-array-portion ,array ,array-start (+ ,array-start ,number-of-parms)
			    (parm-block-accessor ,acb ,element-size) ,acb-start (+ ,acb-start ,number-of-parms)))))
  )

(defmacro LOAD-PARMS (acb element-size &rest parameters)
  "Fills in the ACB command block with the parameters provided."
  (do* ((buf (gensym))
	(start-index (case element-size
		       ((8. 32. :string) 0)
		       (16. (floor *acb-overhead-bytes* 2.))
		       (t (ferror nil "~d is not a legal element-size." element-size))))
	(parms parameters (cdr parms))
	(parm (car parameters) (car parms)) 
	(parm-index start-index (1+ parm-index))
	(accum (when parameters
		 (cons `(setf (aref ,buf ,parm-index) ,parm) nil))
	       (cons `(setf (aref ,buf ,parm-index) ,parm) accum)))
       ((null (cdr parms))
	`(let ((,buf ,(case element-size
			(8. `(parm-buf-8b ,acb))
			(16. acb)
			(32. `(parm-buf-32b ,acb))
			(:string `(parm-buf-string ,acb)))))
	   ,@(reverse accum)))))

(defmacro WITH-UNLOADED-PARMS (acb element-size parameter-var-name-list &body body)
  "Executes BODY with variables in PARAMETER-VAR-NAME-LIST bound to consecutive 
ACB parameters (sized by ELEMENT-SIZE)."
  (do* ((buf (gensym))
	(start-index (case element-size
		       ((8. 32. :string) 0)
		       (16. (floor *acb-overhead-bytes* 2.))
		       (t (ferror nil "~d is not a legal element-size." element-size))))
	(parm-vars parameter-var-name-list (cdr parm-vars))
	(var (car parameter-var-name-list) (car parm-vars)) 
	(parm-index start-index (1+ parm-index))
	(accum (when parameter-var-name-list
		 (cons `(,var (aref ,buf ,parm-index)) nil))
	       (cons `(,var (aref ,buf ,parm-index)) accum)))
       ((null (cdr parm-vars))
	`(let* ((,buf ,(case element-size
			(8. `(parm-buf-8b ,acb))
			(16. acb)
			(32. `(parm-buf-32b ,acb))
			(:string `(parm-buf-string ,acb))))
		,@(reverse accum))
	   ,@body))))


;;; 8-bit parms

(proclaim '(inline parm-8b))
(defun PARM-8B (acb n)
  "Return the Nth 8 bit parameter in ACB."
  (read-parm acb n 8))

(proclaim '(inline set-parm-8b))
(defun SET-PARM-8B (acb n value)
  "Set the Nth 8 bit parameter in ACB to value."
  (set-parm acb n value 8))

(defsetf PARM-8B set-parm-8b)

(defmacro COPY-PARMS-8B (acb array &optional (direction :to-acb)
			     number-of-parms (acb-start 0) (array-start 0))
  "Copy NUMBER-OF-PARMS 8-bit elements between ACB and ARRAY.  When NUMBER-OF-PARMS is
unsupplied or NIL the active length of the source array is used.
  Values for direction are :TO-ACB and :TO-ARRAY."
  `(copy-parms ,acb ,array ,acb-start ,array-start
	       ,(if number-of-parms
		    number-of-parms
		    (if (eq direction :to-acb)
			`(array-active-length ,array)
			`(array-active-length ,acb)))
	       ,direction 8.)
  )

(defmacro LOAD-PARMS-8B (acb &rest parameters)
  "Loads a number of 8-bit parameters into consecutive parm bytes of ACB."
  `(load-parms ,acb 8. ,@parameters))

(defmacro WITH-UNLOADED-PARMS-8B (acb parameter-var-name-list &body body)
  "Executes BODY with variables in PARAMETER-VAR-NAME-LIST bound to consecutive 
8-bit ACB parameters."
  `(with-unloaded-parms ,acb 8. ,parameter-var-name-list ,@body))


;;; 16-bit parms.
;;; Note that this case is optimized not to use the 16-bit displaced array in the header.

(defmacro PARM-16B (acb n)
  "Return the Nth 16 bit parameter in ACB."
  `(aref ,acb (+ ,n (floor *acb-overhead-bytes* 2))))

(defmacro SET-PARM-16B (acb n value)
  "Set the Nth 16 bit parameter in ACB to value."
  `(setf (aref ,acb (+ ,n (floor *acb-overhead-bytes* 2))) ,value))

(defsetf PARM-16B set-parm-16b)

(defmacro COPY-PARMS-16B (acb array &optional (direction :to-acb)
			     number-of-parms (acb-start 0) (array-start 0))
  "Copy NUMBER-OF-PARMS 16-bit elements between ACB and ARRAY.  When NUMBER-OF-PARMS is
unsupplied or NIL the active length of the source array is used.
  Values for direction are :TO-ACB and :TO-ARRAY."
  `(copy-parms ,acb ,array ,acb-start ,array-start
	       ,(if number-of-parms
		    number-of-parms
		    (if (eq direction :to-acb)
			`(array-active-length ,array)
			`(array-active-length ,acb)))
	       ,direction 16.)
  )

(defmacro LOAD-PARMS-16B (acb &rest parameters)
  `(load-parms ,acb 16. ,@parameters))

(defmacro WITH-UNLOADED-PARMS-16B (acb parameter-var-name-list &body body)
  "Executes BODY with variables in PARAMETER-VAR-NAME-LIST bound to consecutive 
16-bit ACB parameters."
  `(with-unloaded-parms ,acb 16. ,parameter-var-name-list ,@body))


;;; 32-bit parms

(defmacro PARM-32B (acb n)
  "Return the Nth 32 bit parameter in ACB."
    `(read-parm ,acb ,n 32))

(defmacro SET-PARM-32B (acb n value)
  "Set the Nth 32 bit parameter in ACB to value."
  `(set-parm ,acb ,n ,value 32))

(defsetf PARM-32B set-parm-32b)

(defmacro COPY-PARMS-32B (acb array &optional (direction :to-acb)
			     number-of-parms (acb-start 0) (array-start 0))
  "Copy NUMBER-OF-PARMS 32-bit elements between ACB and ARRAY.  When NUMBER-OF-PARMS is
unsupplied or NIL the active length of the source array is used.
  Values for direction are :TO-ACB and :TO-ARRAY."
  `(copy-parms ,acb ,array ,acb-start ,array-start
	       ,(if number-of-parms
		    number-of-parms
		    (if (eq direction :to-acb)
			`(array-active-length ,array)
			`(array-active-length ,acb)))
	       ,direction 32.)
  )

(defmacro LOAD-PARMS-32B (acb &rest parameters)
  `(load-parms ,acb 32. ,@parameters))

(defmacro WITH-UNLOADED-PARMS-32B (acb parameter-var-name-list &body body)
  "Executes BODY with variables in PARAMETER-VAR-NAME-LIST bound to consecutive 
32-bit ACB parameters."
  `(with-unloaded-parms ,acb 32. ,parameter-var-name-list ,@body))


;;; String Parms
;;; Note that a length byte is prefixed to strings in ACBs.

(defmacro PARM-STRING (acb n)
  "Return the Nth string element parameter in ACB."
  `(read-parm ,acb ,n :string))

(defmacro SET-PARM-STRING (acb n char)
  "Set the Nth string element parameter in ACB to CHAR."
  `(set-parm ,acb ,n ,char :string))

(defmacro COPY-PARMS-STRING (acb array &optional (direction :to-acb)
			     number-of-parms (acb-start 0) (array-start 0))
  "Copy NUMBER-OF-PARMS string elements between ACB and ARRAY.  When NUMBER-OF-PARMS is
unsupplied or NIL the active length of the source array is used.
  Values for direction are :TO-ACB and :TO-ARRAY."
  `(copy-parms ,acb ,array ,acb-start ,array-start
	       ,(if number-of-parms
		    number-of-parms
		    (if (eq direction :to-acb)
			`(array-active-length ,array)
			`(array-active-length ,acb)))
	       ,direction :string)
  )

(defconstant MAX-GETTABLE-ACB-STRING-SIZE 255.)

(proclaim '(inline get-acb-string))
(defun GET-ACB-STRING (acb start-byte)
  (let ((str (make-array max-gettable-acb-string-size 
			 :element-type '(string-char)
			 :fill-pointer 0))
	(len (add:parm-8b acb start-byte)))
    (copy-parms-string acb str :to-array (min len max-gettable-acb-string-size)
			   (1+ start-byte) 0)
    (setf (fill-pointer str) (min len max-gettable-acb-string-size))
    str))

(proclaim '(inline put-acb-string))
(defun PUT-ACB-STRING (acb start-byte string)
  (let ((len (length string)))
    (add:copy-parms-string acb string :to-acb (min len max-gettable-acb-string-size)
			   (1+ start-byte) 0)
    (setf (parm-8b acb start-byte) (min len max-gettable-acb-string-size))
    string))

(defsetf GET-ACB-STRING put-acb-string)



;;;
;;; Misc ACB Support
;;;

(defconstant *NO-ERROR-CODE* 0)

(proclaim '(inline init-acb))
(defun INIT-ACB (acb opcode subopcode)
  (setf (opcode acb) opcode)
  (setf (subopcode acb) subopcode)
  (setf (error-code acb) *no-error-code*))


(proclaim '(inline check-error))		;needs work
(defun CHECK-ERROR (acb)
  (let ((err (error-code acb)))
    (unless (eql err *no-error-code*)
      (cerror "Continue ignoring error" "ACB error code ~a" err))))


;;;
;;; ACB Construction
;;;

(proclaim '(inline acb-phys-memory-allocated))
(defun ACB-PHYS-MEMORY-ALLOCATED (acb)
  (not (zerop (si:array-data-buffer-address acb))))

(proclaim '(inline byte-length))
(defun BYTE-LENGTH (acb)
  "Returns total amount of storage in bytes of ACB."
  (* 2 (array-total-size acb)))

(proclaim '(inline elements))
(defun ELEMENTS (acb)
  (array-total-size acb))



;;;
;;; ACB Resource
;;;

(defun GET-ACB (nbytes &optional (record-for-deallocator t) &aux acb)
  "Allocate an Addin Command Block (ACB) suitable to hold NBYTES of data."
  (setf acb (get-acb-fast nbytes))
  (let ((array (parm-block-accessor acb 32)))
    (si:fill-array array (array-total-size array) 0))
  (setf (deallocatable acb) record-for-deallocator)
  acb)

(defun RETURN-ACB (acb &optional force-p)
  "Return Addin Command Block ACB to the free ACB pool."
  (return-acb-fast acb force-p))

(proclaim '(inline clear-acb))
(defun CLEAR-ACB (acb)
  (si:fill-array acb (elements acb) 0))



;;;
;;; ACB Queue Setup
;;;

(proclaim '(inline q-type))
(defun Q-TYPE (q)
  (array-leader q 1))

(proclaim '(inline set-q-type))
(defun SET-Q-TYPE (q q-type)
  (setf (array-leader q 1) q-type))

(defsetf Q-TYPE set-q-type)

(defparameter *NUMBER-ACB-QUEUES* si:%addin-max-allocation-queues)	
(defparameter *ACB-Q-SIZES* '(32. 100. 512. 2200.  10000. 32768.))	;must be divisible by 4. ab 8/5/88 
(defparameter *ACB-Q-INITIAL-ELEMENTS* '(50. 25. 10. 10. 4. 4.))	;ab 8/5/88
(defvar *ALLOC-ACBS* nil)
(defvar *ACB-QS* nil)

(defparameter *ACB-QUEUE-BASE* si:%alloc-q-base)

(defvar *Q-DEBUG* nil)

(proclaim '(inline q-acb))
(defun Q-ACB (q acb)
  (q-element q (physical-address acb)))

(proclaim '(inline deq-acb))
(defun DEQ-ACB (q)
  (let ((acb-addr (deq-element q)))
    (if (not (= 0 acb-addr))
	(find-acb acb-addr))))

(defun PRINT-ACB-QS ()
  (format t "~%*ACB-Qs* : ~a" *acb-qs*)
  (dolist (q *acb-qs*)
    (dump-q q)
    (print-q q)))

(proclaim '(inline find-q))
(defun FIND-Q (size)
  (loop for good-size in *acb-q-sizes*
	for el-num = 0 then (1+ el-num) do
	(when (<= size good-size)
	  (return (nth el-num *acb-qs*)))
	finally (ferror nil "~a bytes is too large for an ACB." size)))


;;;
;;; Fast ACB allocation
;;;

(proclaim '(inline return-to-alloc-q))
(defun RETURN-TO-ALLOC-Q (acb)
  (let ((q (find-q (data-bytes acb))))
    (cond ((not *q-debug*)
	   (q-acb q acb))
	  ((q-member (physical-address acb) q)
	   (ferror nil "~%ACB ~a is already on queue ~a" acb q))
	  (t (q-acb q acb)))))

(defun REPLENISH-Q (q)
  (let ((num-elements (nth (position q *acb-qs*) *acb-q-initial-elements*))
	(size (nth (position q *acb-qs*) *acb-q-sizes*))
	acb)
    (dotimes (i num-elements)
      (setq acb (make-allocation-acb size))
      (push-end acb *alloc-acbs*)
      (q-acb q acb))))

(defun GET-ACB-FAST (nbytes &aux acb)
  "Allocate an Addin Command Block (ACB) suitable to hold NBYTES of data."
  (setq acb (or (deq-acb (find-q nbytes))
		(progn (replenish-q (find-q nbytes)) (deq-acb (find-q nbytes)))
		(ferror nil "ACB of size ~a not currently available." nbytes)))
  (setf (status acb) *acb-status-ready*)
  (if (and *micronet-debug*
	   (not (= (ldb (byte 4 24) (physical-address acb)) (ldb (byte 4 0) si:*addin-memory-slot*))))
      (ferror nil "get-acb-fast was passed an acb from another slot."))
  acb)

(defun acb-p (acb)
  (and (eql (array-leader-length acb) si:%ACB-Num-Leader-Words)
       (eq (array-leader acb si:%ACB-Leader-Type-Id) *acb-leader-type-identifier*)))

(defun RETURN-ACB-FAST (acb &optional force-p)
  "Return Addin Command Block ACB to the free ACB pool."
  (when *micronet-debug*
    (when (not (= (ldb (byte 4 24) (physical-address acb)) (ldb (byte 4 0) si:*addin-memory-slot*)))      
      (ferror nil "return-acb-fast getting ready to return an acb from another slot."))

    (when (not (acb-p acb))
      (ferror nil "Attempt to return something that was not an acb.")))
  
  (cond ((and (deallocatable acb)
	      (or force-p
		  (and (servicer-complete acb)
		       (requestor-complete acb))))
	 (return-to-alloc-q acb))))

(defun return-acb-servicer (acb)
  "Set the servicer-complete and input-complete bits of acb and return it if requestor-complete is set."
  (let ((return-p (requestor-complete acb)))
    (setf (servicer-complete acb) t)
    (when return-p
      (return-acb-fast acb))))

(defun FORCE-FREE-ACBS ()
  (dolist (acb *alloc-acbs*)
    (setf (deallocatable acb) t)
    (setf (command-complete acb) t)
    (setf (input-complete acb) t)
    (setf (requestor-complete acb) t)
    (return-acb-fast acb)))



(defun find-lost-acbs (&optional verbose)
  "A debug function for finding lost acbs. This function will not try to return lost acbs. Note that
   some of the lost acbs may have been in-process on the Mac side when we locked the queues or may be 
   allocated as reuseable acbs somewhere."
  
  (let* (all-elements-found
	 lost-elements)

    (when verbose
      (format t "~&Finding all acbs..."))

    (setf all-elements-found
	  (without-interrupts
	    (unwind-protect
		(progn
		  
		  ;; First lock all of the queues
		  (loop for q in *acb-qs*
			do (get-q-lock q))
		  
		  (loop for channel in *channel-list*
			nconc (loop for q in (list (send channel :host-queue) (send channel :mx-queue))
				    do (get-q-lock q)))
		  
		  ;; Now collect all of thier elements
		  (nconc
		    
		    
		    (loop for q in *acb-qs*
			  nconc (loop for el = (q-head q) then (q-link el)
				      until (zerop el)
				      collect (find-acb el)))
		    (loop for channel in *channel-list*
			  nconc (loop for q in (list (send channel :host-queue) (send channel :mx-queue))
				      nconc (loop for el = (q-head q) then (q-link el)
						  until (zerop el)
						  collect (find-acb el))))
		    
		    (loop for controller in net:controller-list
			  nconc (loop for prot in (send controller :receive-acb-list)
				      append (second prot))
			  append (send controller :transmit-acb))
		    
		    ))
	      
	      ;; Now unlock all of the queues.
	      (loop for q in *acb-qs*
		    do (unlock-q q))
	      
	      (loop for channel in *channel-list*
		    nconc (loop for q in (list (send channel :host-queue) (send channel :mx-queue))
				do (unlock-q q))))))

    (when verbose
      (format t "~&Finding lost elements ..."))
    
    (setf lost-elements (set-difference *alloc-acbs* all-elements-found :test #'eq))
    
    (when verbose
      (format t "~&There were ~d lost elements from a total of ~d acbs." (length lost-elements) (length *alloc-acbs*)))
    
    (when (and verbose (y-or-n-p "Print the elements?"))
      (loop for el in lost-elements
	    do (print-acb el)))

    lost-elements
    
    ))
