;;; -*-Mode:Lisp; Package:System-Internals; Base:8.-*-
;;;
;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              DALLAS, TEXAS
;;;
;;; Copyright (C) 1987-1989 Texas Instruments Incorporated. All rights reserved.

;;;
;;; Queues
;;;

;;
;;       +-----------------------------------------------+  
;;   0   |           Lock Word (bit 7 = lock bit)        |  %Queue-Lock-Word
;;       +-----------------------------------------------+                  
;;   1   |           Lock Collision Count                |  %Queue-Lock-Collisions
;;       +-----------------------------------------------+                  
;;   2   |           Queue Head Pointer                  |  %Queue-Head
;;       +-----------------------------------------------+                  
;;   3   |           Queue Tail Pointer                  |  %Queue-Tail
;;       +-----------------------------------------------+                  
;;   4   |           Queue Length                        |  %Queue-Length
;;       +-----------------------------------------------+
;;   5   |           Spare                               |  %Queue-Spare-1
;;       +-----------------------------------------------+                  
;;

(defsysconst %Addin-Max-Allocation-Queues 6.)	;ab 8/5/88 add 32K allocation q
(defsysconst %Words-Per-Addin-Queue 6.)

(DefAlternate Addin-Queue-Words (Q-corresponding-variable-lists System-Constant-Lists)
  (%Queue-Lock-Word 0.
   %Queue-Lock-Collisions 1.
   %Queue-Head 2.
   %Queue-Tail 3.
   %Queue-Length 4.
   %Queue-Spare-1 5.)
  )

(DefAlternate Addin-Queue-Misc-Fields (Q-corresponding-variable-lists
					System-Constant-Lists)
  (%Queue-Lock-Byte 0
    %%Queue-Lock-Bit-in-Lock-Byte (byte 1 7.))
  )


;;;
;;; Channels
;;;

;; A channel is a set of queues used for two-way communication
;; between the Addin and the MAC.  A separate channel exists for
;; each type of IO device, and for different types of communications
;; protocols.


;;       +-----------------------------------------------+                  
;;   0   |     Attribute Bits    |     Status  Bits      |  %Channel-Status-Register
;;       +-----------------------------------------------+
;;   1   |                Max Queue Length               |  %Channel-max-queue-length
;;       +-----------------------------------------------+  
;;   2   |                   Queue host                  |  %Channel-Queue-host
;;       -                                               -
;;       +-----------------------------------------------+  
;;   8   |                    Queue mx                   |  %Channel-Queue-mx
;;       -                                               -
;;       +-----------------------------------------------+                  
;;   14  |                     Spare                     |  %Channel-Spare
;;       +-----------------------------------------------+  
;;



;; Initial configuration constants.
;;(defsysconst %Addin-Initial-Channel-Base-Offset #xFFFF0)

(defsysconst %Addin-Max-Number-of-Channels 16.)
(defsysconst %Words-Per-Addin-Channel 15.)


(DefAlternate Addin-Channel-Words (Q-corresponding-variable-lists
				    System-Constant-Lists)
  (%Channel-Status-Register 0.

   %Channel-max-queue-length 1.

   %Channel-queue-host 2.
		      
   %Channel-queue-mx 8.

   %Channel-Spare 14.

   )
  )

;; 
;; Channel Status Register:
;;
;;        31                    16 15                   0
;;       +-----------------------------------------------+                  
;;       |            | I | R | Q |          |DE|ER|EN|IN|  %Channel-Status-Register 
;;       +-----------------------------------------------+
;;
;;    STATUS BITS <00:15>:
;;
;;       IN <00:01> :  Initialized.   0 = initialized  1 = uninitialized
;;       EN <01:01> :  Enabled.       0 = disabled     1 = enabled
;;       ER <02:01> :  Error.         0 = no error     1 = error
;;       DE <03:01> :  Debug Enable.  0 = disabled     1 = debug enabled
;;                  
;;    ATTRIBUTE BITS <16:31>:
;;
;;       Q  <16:01> :  Queued.        0 = no queue     1 = requests queued
;;       R  <17:01> :  Requestor.     0 = addin        1 = host
;;       I  <18:01> :  Interrupt.     0 = polled       1 = interrupt
;;                  


(DefAlternate Channel-Status-Fields (Q-corresponding-variable-lists
				      System-Constant-Lists)
  (
   ;; Status bits
   %%Channel-Status-Status      (byte 16. 0)
   %%Chan-Status-Initialized    (byte 1 0)
   %%Chan-Status-Enabled        (byte 1 1)
   %%Chan-Status-Error          (byte 1 2)
   %%Chan-Status-Debug-Enabled  (byte 1 3)

   ;; Attribute bits
   %%Channel-Status-Attributes (byte 16. 16.)
   %%Chan-Status-Queued        (byte 1 16.)
   %%Chan-Status-Requestor     (byte 1 17.)
   %%Chan-Status-Interrupt     (byte 1 18.)

   ;; Same as above, but relative to high halfword.
   %%Chan-Attr-Queued          (byte 1 0.)
   %%Chan-Attr-Requestor       (byte 1 1.)
   %%Chan-Attr-Interrupt       (byte 1 2.)
   )
  )


(DefAlternate Channel-Status-Halfwords (Q-corresponding-variable-lists
				        System-Constant-Lists)
  (
   %Channel-Status-Status 0
   %Channel-Status-Attributes 1)
  )


;;;
;;; Channel Ownership Definitions
;;;

(DefAlternate Addin-Channel-Types (Q-corresponding-variable-lists
				    System-Constant-Lists)
  (%Chan-Type-Error 0.
   %Chan-Type-Display-IO 1.
   %Chan-Type-enet 2.
   %Chan-Type-Spare-3 3.
   %Chan-Type-Misc 4.
   %Chan-Type-Disk-IO 5.
   %Chan-Type-MicroNet-Ports 6.
   %Chan-Type-Spare-7 7.
   %Chan-Type-toolbox-driver 8.
   %Chan-Type-toolbox-0 9.
   %Chan-Type-Toolbox-1 10.
   %Chan-Type-Toolbox-2 11.
   %Chan-Type-Toolbox-3 12. 
   %Chan-Type-Toolbox-4 13. 
   %Chan-Type-Toolbox-5 14.
   %Chan-Type-Toolbox-6 15.

   %Chan-Type-Error-Host 0.
   %Chan-Type-Error-Addin 0.

   %Chan-Type-Display-IO-Addin 1.

   %Chan-Type-Block-Allocation-Host 2.
   %Chan-Type-Block-Allocation-Addin 2.
   
   %Chan-Type-Misc-Host 4.
   %Chan-Type-Misc-Addin 4.

   %Chan-Type-Disk-IO-Addin 5.

   %Chan-Type-RPC-Host 6.
   %Chan-Type-RPC-Addin 6.

   )
  )

(DefAlternate Addin-Channel-Attributes (Q-corresponding-variable-lists System-Constant-Lists)

  (%Chan-Error-Attributes             '(si:queued nil si:interrupt nil)   
   %Chan-Display-IO-Attributes        '(si:queued t si:interrupt nil si:max-queue-length 10.)   
   %Chan-Misc-Attributes              '(si:queued t   si:interrupt nil)   
   %Chan-Disk-IO-Attributes           '(si:queued nil si:interrupt nil)
   %Chan-MicroNet-Ports-Attributes    '(si:queued t si:interrupt nil)
   %Chan-Toolbox-driver-attributes    '(si:queued t si:interrupt nil si:max-queue-length 10.)
   %Chan-Toolbox-0-attributes         '(si:queued t si:interrupt nil si:max-queue-length 10.)
   %Chan-Toolbox-1-attributes         '(si:queued t si:interrupt nil si:max-queue-length 10.)
   %Chan-Toolbox-2-attributes         '(si:queued t si:interrupt nil si:max-queue-length 10.)
   %Chan-Toolbox-3-attributes         '(si:queued t si:interrupt nil si:max-queue-length 10.)
   %Chan-Toolbox-4-attributes         '(si:queued t si:interrupt nil si:max-queue-length 10.)
   %Chan-Toolbox-5-attributes         '(si:queued t si:interrupt nil si:max-queue-length 10.)
   %Chan-Toolbox-6-attributes         '(si:queued t si:interrupt nil si:max-queue-length 10.)
  
   )
  )



;;;
;;; ACB arrays and cmd blocks.
;;;

;;; An ACB array an ART-8B (byte) physical array.  The physical address
;;; space contains a command block with the following format
;;
;;                          OVERHEAD
;; 
;;           Byte 3     Byte 2      Byte 1      Byte 0      
;;       +-----------------------------------------------+                  
;;   3   |             Link word for queueing            |  %ACB-Link
;;       +-----------------------------------------------+
;;       |                 Return Queue                  |  %ACB-Return-Queue
;;       +-----------------------------------------------+  
;;       |     Source Slot       |      Source Port      |  %ACB-Source-Address
;;       +-----------------------------------------------+  
;;       |   Destination Slot    |   Destination Port    |  %ACB-Destination-Address
;;       +-----------------------------------------------+  
;;       | Opcode    |           | SubOpcode |           |  %ACB-Command-word
;;       +-----------------------------------------------+  
;;       | Return act|           |           |           |  %ACB-Return-Action-Word
;;       +-----------------------------------------------+  
;;   1   |SC|IC|     |RC|BD|     | Error     |           |  %ACB-Status-word
;;       +-----------------------------------------------+                  
;;   2   |                Total Data Bytes               |  %ACB-Data-Bytes
;;       +-----------------------------------------------+  
;;       |               Number Bytes Parms              |  %ACB-Parm-Bytes
;;       +-----------------------------------------------+  
;;   4   |    Address of ACB array pointing to us        |  %ACB-Array-Backlink 
;;       +-----------------------------------------------+  
;;
;;                    
;;                         PARAMETERS
;;       +-----------------------------------------------+  
;;   5   |                                               |  %ACB-Parm-Word-0
;;       +-----------------------------------------------+                  
;;       .                                               .
;;       .                                               .
;;       .                                               .
;;       +-----------------------------------------------+                  
;;   n   |                                               |  (nth parm word)
;;       +-----------------------------------------------+ 
;;

(DefAlternate ACB-Data-Words (Q-corresponding-variable-lists
				 System-Constant-Lists)
    (%ACB-link-Word 0.
     %ACB-Return-Queue 1.
     %ACB-Source-Address 2.
      %%ACB-Source-Slot (byte 16. 0)
      %%ACB-Source-Port (byte 16. 16.)
     %ACB-Destination-Address 3
      %%ACB-Destination-Slot (byte 16. 0)
      %%ACB-Destination-Port (byte 16. 16.)
     %ACB-Command-Word 4.
;       %%ACB-Opcode (byte 8. 24.)
;       %%ACB-Sub-Opcode (byte 8. 8.)
       %%ACB-Command-Opcode (byte 8. 24.)
       %%ACB-Command-Subop (byte 8. 8.)
     %ACB-Return-Action-Word 5.
       %%ACB-Return-Action (byte 8. 24.)

     %ACB-Status-Word 6.
      %%ACB-Status-Input-Complete      (byte 1. 30.)
      %%ACB-Status-Servicer-Complete   (byte 1. 31.)
      %%ACB-Status-Block-Deallocatable (byte 1. 22.)
      %%ACB-Status-Requestor-Complete  (byte 1. 23.)
;      %%ACB-Error (byte 8. 8.)
      %%ACB-Status-Error (byte 8. 8.)

     %ACB-Data-Bytes 7.
     %ACB-Parm-Bytes 8.
     %ACB-Array-Backlink-Word 9.
     %ACB-Parm-Word-0 10.
     )
  )

(DefAlternate ACB-Data-Halfwords (Q-corresponding-variable-lists
				  System-Constant-Lists)
    (%ACB-Link-Halfword-Lo 0.
     %ACB-Link-Halfword-Hi 1.
     %ACB-Return-Queue-Halfword-Lo 2.
     %ACB-Return-Queue-Halfword-Hi 3.
     %ACB-Source-Port-Halfword 4.
     %ACB-Source-Slot-Halfword 5.
     %ACB-Destination-Port-Halfword 6.
     %ACB-Destination-Slot-Halfword 7.
     %ACB-Subopcode-Halfword 8. 
       %%ACB-Subopcode-in-Subop-HW (byte 8. 8.)
     %ACB-Opcode-Halfword 9.
       %%ACB-Opcode-in-Op-HW (byte 8. 8.)
     %ACB-Error-Halfword 10.
       %%ACB-Error-Code-in-Err-HW (byte 8. 8.)
     %ACB-Return-Action-Halfword 11.
       %%ACB-Return-Action-in-Return-Action-Halfword (byte 8. 8.)

     %ACB-Status-Halfword-Unused 12.
       
     %ACB-Status-Halfword 13.
       %%ACB-BD-Bit-In-Stat-HW (byte 1. 6.)
       %%ACB-RC-Bit-In-Stat-HW (byte 1. 7.)
       %%ACB-IC-Bit-In-Stat-HW (byte 1. 14.)
       %%ACB-SC-Bit-In-Stat-HW (byte 1. 15.)

     %ACB-Data-Bytes-Halfword-lo 14.
     %ACB-Data-Bytes-Halfword-HI 15.
     %ACB-Parm-Length-Halfword-Lo 16.
     %ACB-Parm-Length-Halfword-Hi 17.
     %ACB-Array-Backlink-Halfword-Lo 18.
     %ACB-Array-Backlink-Halfword-Hi 19.
     %ACB-Parm-Word-0-Halfword 20.
     )
  )

(DefAlternate ACB-Data-Bytes (Q-corresponding-variable-lists
				 System-Constant-Lists)
    (%ACB-Link-Byte-1 0
     %ACB-Link-Byte-2 1.
     %ACB-Link-Byte-3 2.
     %ACB-Link-Byte-4 3.

     %ACB-Return-Queue-Byte-1 4.
     %ACB-Return-Queue-Byte-2 5.
     %ACB-Return-Queue-Byte-3 6.
     %ACB-Return-Queue-Byte-4 7.

     %ACB-Source-Port-Byte-1 8.
     %ACB-Source-Port-Byte-2 9.
     %ACB-Source-Slot-Byte-3 10.
     %ACB-Source-Slot-Byte-4 11.

     %ACB-Destination-Port-Byte-1 12.
     %ACB-Destination-Port-Byte-2 13.
     %ACB-Destination-Slot-Byte-3 14.
     %ACB-Destination-Slot-Byte-4 15.

     %ACB-Unused-Byte-1 16.
     %ACB-Command-Subop-Byte 17.
     %ACB-Unused-byte-2 18.
     %ACB-Command-Opcode-Byte 19.

     %ACB-Unused-Byte-3  20.
     %ACB-Status-Error-Byte 21.
     %ACB-Unused-Byte-4 22.
     %ACB-Return-Action-Byte 23.

     %ACB-Ststus-byte-unused-1 24.

     %ACB-Ststus-byte-unused-2 25.

     %ACB-Status-RC-Byte 26.
       %%ACB-BD-Bit-In-RC-Byte (byte 1. 6.)
       %%ACB-RC-Bit-In-RC-Byte (byte 1. 7.)

     %ACB-Status-SC-Byte 27.
       %%ACB-IC-Bit-In-SC-Byte (byte 1. 6.)
       %%ACB-SC-Bit-In-SC-Byte (byte 1. 7.)

     %ACB-Data-Bytes-1 28.
     %ACB-Data-Bytes-2 29.
     %ACB-Data-Bytes-3 30.
     %ACB-Data-Bytes-4 31.

     %ACB-Parm-Bytes-1 32.
     %ACB-Parm-Bytes-2 33.
     %ACB-Parm-Bytes-3 34.
     %ACB-Parm-Bytes-4 35.

     %ACB-Backlink-Byte-0 36.
     %ACB-Backlink-Byte-1 37.
     %ACB-Backlink-Byte-2 38.
     %ACB-Backlink-Byte-3 39.

     %ACB-Parm-0-Byte 40.
     )
  )

;;; An ACB array an ART-8B (byte) physical array with the following leader words:

;;
;;       +-----------------------------------------------+                  
;;   9   |        %ACB-Leader-Spare-2                    |   Spare
;;       +-----------------------------------------------+                 
;;   8   |        %ACB-Leader-Spare-1                    |   Spare
;;       +-----------------------------------------------+                  
;;   7   |        %ACB-Leader-Parm-Array-32B             |   32-bit array overlaying parameters
;;       +-----------------------------------------------+                 
;;   6   |        %ACB-Leader-Parm-Array-16B             |   16-bit array overlaying parameters
;;       +-----------------------------------------------+                  
;;   5   |        %ACB-Leader-Parm-Array-8B              |   8-bit array overlaying parameters
;;       +-----------------------------------------------+                  
;;   4   |        %ACB-Leader-Parm-Array-String          |   String array overlaying parameters
;;       +-----------------------------------------------+                  
;;   3   |        %ACB-Leader-Data-Bytes-Requested       |   Data size originally requested
;;       +-----------------------------------------------+                  
;;   2   |        %ACB-Leader-Type                       |   Special ACB type (optional)
;;       +-----------------------------------------------+                  
;;   1   |        %ACB-Leader-Info                       |   Misc. info
;;       +-----------------------------------------------+ 
;;   0   |        %ACB-Leader-Link                       |   Link word for queueing
;;       +-----------------------------------------------+ 
;;

(DefAlternate ACB-Leader-Words (Q-corresponding-variable-lists
				 System-Constant-Lists)
    (%ACB-Num-Leader-Words 10.
     %ACB-Leader-Info 0.
     %ACB-Leader-Type 1.
     %ACB-Leader-Owner 2.
     %ACB-Leader-Data-Bytes-Requested 3.
     %ACB-Leader-Parm-Array-String 4.
     %ACB-Leader-Parm-Array-8B 5.
     %ACB-Leader-Parm-Array-16B 6.
     %ACB-Leader-Parm-Array-32B 7.
     %ACB-Leader-Physical-Address 8.
     %ACB-Leader-Type-Id 9.
;     %ACB-Leader-Spare-2 9.
     )
  )


;;;
;;; Misc Global Data
;;;

(DefSysConst Addin-Initial-Chunk-Pages 64.)
(DefSysConst Micronet-Reserved-Pages 4.)	;Lowest reserved area address = #xFE000

(DefSysConst %Channel-Register-Base #xFF000)	;Start of 3rd free page, for #x800 bytes, #xc0 words #xff000 *BJ*
(DefSysConst %Alloc-Q-Base          #xFF400)	;for #x100 bytes (#x40 words)
(DefSysConst %Channel-Q-Base        #xFF010)	;for #x300 bytes (#xC0 words)

(DefSysConst %Driver-Data-Start     #xFF800)    ;Start of last free page, for #x800 bytes, #x200 words

(DefSysConst %addin-nupi-register 0)
(defsysconst %addin-crash-register 8.)

(DefAlternate Driver-Data-Offsets (Q-corresponding-variable-lists
				     System-Constant-Lists)
  (%DD-Init-Flags    20.			;ab 3-25-88
   %DD-Kbd-Valid     22.
   %DD-Kbd-Buffer    24.
   %DD-Mouse-Valid   28.
   %DD-Mouse-Y       30.
   %DD-Mouse-X       32.
   %DD-Mouse-Buffer  34.
   %DD-Mouse-Screen  36.
;;  offsets for: run bar (all longs) / who state (21 chars including overhead)
   %DD-total-elapsed-time	40.
   %DD-cpu-time			44.
   %DD-page-time	 	48.
   %DD-scavenge-time 		52.
   %DD-total-interval		56.
   %DD-cpu-interval 		60.
   %DD-who-state 		64.
   %DD-file-state		86.
   %DD-mouse-doc-1		168.
   %DD-mouse-doc-2		298.
   %DD-mouse-doc-3		428.
   %DD-mouse-doc-4		558.
   %DD-mouse-doc-5		688.
   %DD-rb-enabled               818.
   %DD-rb-onp                   819.
   %DD-rb-start-y               820.
   %DD-rb-start-x               822.
   %DD-rb-end-y                 824.
   %DD-rb-end-x                 826.
   %DD-rb-draw-pattern          828.
   %DD-rb-direction             830.
   %DD-rb-in-window-p           832.
   %DD-rb-clip-y1               834.
   %DD-rb-clip-x1               836.
   %DD-rb-clip-y2               838.
   %DD-rb-clip-x2               840.
   %DD-next-available-spot	840.
   
   %DD-Boot-Type               2044.		;last word on page
   )
  )

;;; The length of shared memory buffers including trailing null.
;;; The byte following the null signals new data written.
(DefSysConst %MX-Who-State-String-Max 21.)

(DefSysConst %MX-Who-File-String-Max 81.)

(DefSysConst %MX-Mouse-Doc-Max 129.)		;5 lines x 129 chars/line

(DefSysConst %MX-Max-mouse-lines 5.)

;;;
;;; Misc Port & Misc Channel
;;;

(DefAlternate Misc-Channel-Opcodes (Q-corresponding-variable-lists
				     System-Constant-Lists)
  (%MC-Disk-Cmd    #x01
   %MC-readcrash   #x02
   %MC-tvcalls     #x03
   %MC-Printer     #x04
   )   
  )

;; ab 8/29/88.  Moved this to DISK-IO; DEFINITIONS
;;(DefAlternate Disk-SubOps (Q-corresponding-variable-lists
;;			    System-Constant-Lists)
;;  (%DC-Get-Partition-List 1
;;   %DC-Get-Number-Partition-List-Entries 2
;;   %DC-Get-Volume-Name 3
;;   %DC-Get-Booted-Load-Band-Info 4
;;   %DC-Get-Booted-Mcr-Band-Info 5
;;   %DC-Get-Startup-Default-Device-Info 6		;1.18.88 MBC
;;   %DC-Get-Startup-Host-Name-Info 7
;;   )
;;  )

(DefAlternate readcrash-SubOps (Q-corresponding-variable-lists
			    System-Constant-Lists)
  (%RC-Read-Crash-Record 2
   %RC-Number-of-Crecs 3
   )
  )
(DefAlternate tvcalls-SubOps (Q-corresponding-variable-lists
			    System-Constant-Lists)
  (%TC-Beep     2.
   %TC-Get-Time 3.
   %TC-Set-Time 4.
   %TC-TYO      5.
   %TC-TYI      6.
   %TC-LISTEN   7.
   %TC-KEYSTATE 8.
   %TC-SETPOS   9.
   %TC-CLREOL   10.
   %TC-CLRSCR   11.
   %TC-SCREEN-INFO 12.
   %TC-CURSOR   13.
   %TC-STRING-OUT 14.
   %TC-DIRECT-DRAW 15.
   )
  )


;;;
;;; Error Port Handling
;;;

(DefSysConst %Error-Command-Block-Length 100.)       ;Longest we should need

(DefAlternate Error-Channel-Opcodes (Q-corresponding-variable-lists
				      System-Constant-Lists)
    (%EC-Crash-Op      #x01
     %EC-Misc-Ucode-Op #x02
     )
  )


(DefAlternate Crash-SubOps (Q-corresponding-variable-lists
			     System-Constant-Lists)
    (%CO-Write-Crash-Record 1
     )
  )

(DefAlternate Misc-Ucode-SubOps (Q-corresponding-variable-lists
				  System-Constant-Lists)
    (%MU-Boot-Status 1
     %MU-Reboot-Request 2
     )
  )


