;;; -*- Mode:Common-Lisp; Package:Micronet; Base:10 -*-
;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1987- 1989 Texas Instruments Incorporated. All rights reserved.


;;;
;;; Edit History
;;;
;;;                   Patch
;;;   Date    Author  Number   Description
;;;--------------------------------------------------------------------
;;;  4-18-88  BJ      ----     Moved code from acb-defs.lisp.



;;;
;;; External functions.
;;;
;;; Q-Link:
;;; Set-Q-Link:
;;;
;;; Q-Head:
;;; Set-Q-Head:
;;;
;;; Q-Tail:
;;; Set-Q-Tail:
;;;
;;; Q-Lock-Collisions:
;;; Set-Q-Lock-Collisions:
;;;
;;; Make-Q:
;;; Clr-Q:
;;; Unlock-Q:
;;; Get-Q-Lock:
;;; Q-Member:
;;; Q-Length:
;;; Set-Q-Length:
;;; Q-Element:
;;; Deq-Element:

;; This is to guarantee that various variable values will be compiled to constants.
(eval-when (eval compile load)
(defvar *Q-CONSTANTS*
	'(si:%%nubus-all-but-f-and-slot-bits si:%%nubus-all-but-f-and-slot-bits
	  si:%acb-link-word si:%queue-head si:%queue-tail si:%queue-lock-collisions si:%queue-lock-word))
  
(defun FIX-Q-CONSTANTS ()
  (dolist (var *q-constants*)
    (setf (get var 'compiler:system-constant) t)))
(fix-q-constants))

(proclaim '(inline q-link))
(defun Q-LINK (acb-phys-adr)
  (let* ((slot (ldb si:%%nubus-f-and-slot-bits acb-phys-adr))
	 (offset (ldb si:%%nubus-all-but-f-and-slot-bits acb-phys-adr)))
    (logand (si:%nubus-read slot (+ offset (* si:%acb-link-word 4))) #xffffffff)))

(proclaim '(inline set-q-link))
(defun SET-Q-LINK (acb-phys-adr val)
  (let* ((slot (ldb si:%%nubus-f-and-slot-bits acb-phys-adr))
	 (offset (ldb si:%%nubus-all-but-f-and-slot-bits acb-phys-adr)))
    (si:%nubus-write slot (+ offset (* si:%acb-link-word 4)) val)
    val))

(defsetf Q-LINK set-q-link)

(proclaim '(inline q-lock))
(defun Q-LOCK (q)
  (array-leader q 0))

(proclaim '(inline q-head))
(defun Q-HEAD (q)
  (aref q si:%queue-head))

(proclaim '(inline set-q-head))
(defun SET-Q-HEAD (q val)
  (setf (aref q si:%queue-head) val))

(defsetf Q-HEAD set-q-head)

(proclaim '(inline q-tail))
(defun Q-TAIL (q)
  (aref q si:%queue-tail))

(proclaim '(inline set-q-tail))
(defun SET-Q-TAIL (q val)
  (setf (aref q si:%queue-tail) val))

(defsetf Q-TAIL set-q-tail)

(proclaim '(inline q-lock-collisions))
(defun Q-LOCK-COLLISIONS (q)
  (aref q si:%queue-lock-collisions))

(proclaim '(inline set-q-lock-collisions))
(defun SET-Q-LOCK-COLLISIONS (q val)
  (setf (aref q si:%queue-lock-collisions) val))

(defsetf Q-LOCK-COLLISIONS set-q-lock-collisions)

(PROCLAIM '(inline q-length))
(defun Q-LENGTH (q)
  (aref q si:%queue-length))

(proclaim '(inline set-q-length))
(defun SET-Q-LENGTH (q val)
  (setf (aref q si:%queue-length) val))

(defsetf Q-LENGTH set-q-length)

(proclaim '(notinline make-q))
(defun MAKE-Q (byte-offset &optional (q-type :allocation))
  (let ((phys-adr (dpb si:*addin-memory-slot* si:%%nubus-f-and-slot-bits byte-offset))
	q)
    (setq q (make-array si:%words-per-addin-queue
			:element-type '(unsigned-byte 32)
			:displaced-to-physical-address phys-adr
			:leader-list (list (+ phys-adr
					      (* 4. %queue-lock-word)
					      %queue-lock-byte)
					   q-type)))
    q))

(proclaim '(notinline clr-q))
(defun CLR-Q (q)
  (dotimes (i si:%words-per-addin-queue)
    (setf (aref q i) 0))
  (setf (array-leader q 0)
	(+ (si:array-data-buffer-address q)
	   (* 4. %queue-lock-word)
	   %queue-lock-byte)))

(proclaim '(inline q-end))
(defun Q-END (val)
  (eql val *q-end*))

(proclaim '(inline q-empty))
(defun Q-EMPTY (q)
  (eql (q-length q) 0))

(proclaim '(inline q-locked-p))
(defun Q-LOCKED-P (q)
  (let* ((lock-adr (q-lock q))
	 (slot (ldb si:%%nubus-f-and-slot-bits lock-adr))
	 (offset (ldb si:%%nubus-all-but-f-and-slot-bits lock-adr))
	 (lock-val (si:%nubus-read-8b-careful slot offset)))
    (values (and (numberp lock-val) (plusp lock-val))
	    lock-val)))

(proclaim '(inline unlock-q))
(defun UNLOCK-Q (q)
  (setf (aref q si:%queue-lock-word) 0))

(defvar *Q-LOCK-RESETS* 0)
(defvar *Q-END* 0)
(defvar *error-on-reset* t)


;; Make get-q-lock respect add:*no-interrupt*. *BJ*.
(proclaim '(inline get-q-lock))
(defun GET-Q-LOCK (q)
  (declare (special *no-interrupt*))
  (let* ((lock-adr (q-lock q))
	 (slot (ldb si:%%nubus-f-and-slot-bits lock-adr))
	 (offset (ldb si:%%nubus-all-but-f-and-slot-bits lock-adr))
	 (collision-count 0))
    (loop 	  
      when (si:%test&set-68k slot offset)
      return t
      
      when (> collision-count 10000)
      return (q-lock-reset-error q)
      
      when (and (>  collision-count 500) (null add:*no-interrupt*))
      return (process-wait "Wait Q Lock"
			   #'(lambda (slot offset failure-count)
			       (cond ((si:%test&set-68k slot offset)
				      t)
				     ((> (contents failure-count) 10000)
				      (q-lock-reset-error q)
				      t)
				     (t
				      (incf (contents failure-count))
				      nil)))
			   slot offset (locf collision-count)))
    (if (< 0 collision-count)
	(incf (q-lock-collisions q) collision-count))))

(defun q-lock-reset-error (q)
  (declare (ignore q))
  (incf *q-lock-resets*)
  (when *error-on-reset*
    (let ((si:cold-load-stream-owns-keyboard t)
	  (*terminal-io* si:cold-load-stream))
      (eh:save-screen-for-cold-load-stream t)
      (cerror "Proceed" "q lock stuck")
      (eh:restore-screen-for-cold-load-stream))))

(defun Q-MEMBER (acb-address q)
  (loop for q-el = (q-head q) then (q-link q-el)
	until (q-end q-el) do
	(when (eql acb-address q-el)
	  (return t))
	finally (return nil)))

(defun Q-ELEMENT (q addr)
  (when (and *q-debug* (q-member addr q))
    (ferror nil "Element ~a is already on Q ~a" addr q))
  (unless (si:wired-p #'q-element)
    (si:wire-structure #'q-element))
  (setf (q-link addr) *q-end*)
  (without-interrupts
    (get-q-lock q) 
    (cond ((q-empty q)
	   (setf (q-head q) addr)
	   (setf (q-tail q) addr))
	  (t
	   (setf (q-link (q-tail q)) addr)
	   (setf (q-tail q) addr)))
    (incf (q-length q))
    (unlock-q q)))

(defun DEQ-ELEMENT (q &aux hd new-hd)
  (unless (si:wired-p #'deq-element)
    (si:wire-structure #'deq-element))
  (without-interrupts
    (get-q-lock q)
    (cond ((not (q-end (setq hd (q-head q))))
	   (setf (q-head q)
		 (setq new-hd (q-link hd)))
	   (setf (q-link hd) *q-end*)
	   (cond ((q-end new-hd)
		  (setf (q-head q) *q-end*)
		  (setf (q-tail q) *q-end*)))
	   hd)
	  (t nil))
    (cond ((q-end hd) nil)
	  (t
	   (decf (q-length q))))
    (unlock-q q))
  hd)



(defun DUMP-Q (q &optional (strm *standard-output*))
  (format strm "~2%Queue:         ~a~
                 ~%Type:          ~a~
                 ~%Length:        ~a~
                 ~%Displaced to:  #x+~8,'0,,x"
	  q
	  (q-type q)
	  (q-length q)
	  (si:array-data-buffer-address q))
  
  (format strm "~%Contents:")
  (dolist (el si:addin-queue-words)
    (format strm "~%   #x+~8,'0,,x  ~a"
	    (aref q (symbol-value el)) el)))


(defun PRINT-Q (q)
  (format t "~%   Q: ~a (~a)" q (coerce q 'list))
  (loop for el = (q-head q) then (q-link el)
	until (q-end el) do
	(format t "~%       #x~x,  ACB: ~@28a,  ACB q-link: #x~x"
		el (if (not (= 0 el)) (find-acb el)) (q-link el)))
  )