;; -*- Mode: Common-Lisp; Package: HOST; Base: 10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.

;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151

;;; Copyright (C) 1985-1989 Texas Instruments Incorporated. All rights reserved.

;; This file contains information necessary to enable the function HOST:HOST-STATUS
;; File History
;; ======================
;; Created by LS 11/18/86
;; ======================


;;; 04-24-89 DAB Added all function, variables, flavors and instances that are documented to export list.

;;; 05-11-88 DAB Print out unknown host in host-status.

(defvar *poll-each-status-p* t
  "If T, then poll each status during HOST:HOST-STATUS.  If NIL, then poll only the first status found.")

(export '(net:host-status   ;function  ; DAB 04-24-89
	   net:*poll-each-status-p*    ;variable
	   net:reset        ;function
	   )
	'net)

;;=====================
;; CHAOS STATUS SERVICE

(defflavor CHAOS-STATUS-SERVICE
	   ((name :CHAOS-STATUS)
	    (desirability .85))
	   (service-implementation-mixin)
  :gettable-instance-variables
  :settable-instance-variables
  :initable-instance-variables)

(define-service-implementation 'CHAOS-STATUS-SERVICE)

(defmethod (CHAOS-STATUS-SERVICE :STATUS) (host)
  (list :CHAOS (chaos:host-up-p host)))



;;--------------------------------------------------------------------------------
;;--------------------------------------------------------------------------------
;; NEW HOST:HOST-STATUS FUNCTION

(defun HOST-STATUS (&rest hosts)
  "Poll host(s) for the status of each network that it supports.
When called with no arguments, all hosts on the HOST:*HOST-LIST* will be polled."
  (format *standard-output* "~2%ADDR~14THOST~35TSTATUS~2%")
  (let ((status-list))
    (unless hosts
      (setf hosts host:*host-list*)
      (setf hosts (sort hosts 'string-lessp :KEY #'(lambda (host) (send host :fully-qualified-name)))))
    (loop for host in hosts
	  for host2 = host ;05-11-88 DAB
	  do
	  (setf host (si:parse-host host t))
	  ;; DON'T GIVE STATUS INFO ON HOSTS IN THE DISTRIBUTION-NAMESPACE
	  (when (not host) (format *standard-output* "~&*Unknown host ~14T~s"  host2))  ;05-11-88 DAB
		 
	  (when (and host
		     (not (name:key-compare name:*distribution-name*
					    (send host :domain))))
	    (setf status-list
		  (send host :status (null *poll-each-status-p*)))	  
	    (cond ((null status-list)
		   (format *standard-output* "~&~14T~s~35Thas no defined status function."
			   (send host :fully-qualified-name)))
		  (t
		   (unless (null status-list)
		     (dolist (status status-list)
		       (format *standard-output* "~&~s~14T~s~35T~a~a."
			       (send host :network-address (first status))
			       (send host :fully-qualified-name)
			       (if (second status)
				   "is responding on medium "
				   ;;else
				   "is NOT responding on medium ")
			       (first status))))))))))



;; ================================================================================
;; ================================================================================
;;  NEW METHOD ON BASIC-HOST

(defmethod (BASIC-HOST :STATUS) (&OPTIONAL FIRST-ONLY-P &rest ignore)
  (let ((result-list))
    (loop for service in (send self :sorted-service-list)
	  do
	  (when (and (eq (first service) :status)
		     (find-service-implementation (third service)))
		(push-end 
		  (send (find-service-implementation (third service)) :status self)
		  result-list)
		(when first-only-p
		  (return))))
    (setf result-list (delete-duplicates result-list :test #'equal))
    result-list))

(compile-flavor-methods basic-host)




