;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Patch file for NFS version 3.2
;;; Reason: Modified nfs-server-create to use :truncate option to fix a file length
;;; problem when copying to an existing file
;;; Written 11/16/87 09:15:49 by HMCCREARY,
;;; while running on BOOTES from band LOD1
;;; With IO 3.46, PATHNAME 3.17, FILE 3.16, SYSTEM 3.114, METER 3.7, SERIAL 3.2, CHAOSNET 3.25,
;;;  ETHERNET 3.4, GC 3.24, IMAGEN 3.0, MAIL-DAEMON 3.8, NETWORK-SUPPORT 3.21, PROFILE 3.6,
;;;  SUGGESTIONS 3.7, UCL 3.4, ZWEI 3.34, STREAMER-TAPE 3.17, DEBUG-TOOLS 3.6, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.1, INPUT-EDITOR 3.0, MAIL-READER 3.23, NAMESPACE-EDITOR 3.7, NVRAM 3.6,
;;;  TELNET 3.7, TV 3.33, NAMESPACE 3.15, COMPILER 3.35, PRINTER 3.10, SYSLOG 3.1,
;;;  VT100 3.14, VISIDOC 1.0, IP 2.76, Experimental BUG 8.0, RPC 3.1, NFS 3.0,  microcode 230,
;;;  Band Name: Release 3.2 - 10/22 +IP

#!C
; From file NFS_SERVER.LISP#> NFS; MR-X:
#10R NETWORK-FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NETWORK-FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NFS; NFS_SERVER.#"


(DEFUN-METHOD NFS-DISPATCHER NFS-SERVER (request xdr-stream)
   (declare (values ignore))

   (catch 'DISPATCHER-EXIT
     ;; enforce NULL procedure convention even if file system is dismounted
     (when (= NFSPROC-NULL (rpc:svc-req-procedure request))
       (send self :sendreply :xdr-void nil)
       (throw 'DISPATCHER-EXIT nil))

       ;; detect special case of recently dismounted file system
     (when (not (local-file-system-mounted-p))
       (send self :sendreply :xdr-enum NFSERR-NXIO)
       (throw 'DISPATCHER-EXIT nil))

     ;; if the same request as the previous one. send NIL
     (when (= *previous-request-id* (rpc:rpc-msg-xid rpc:request))
       (send self :sendreply :xdr-void nil)
       (throw 'DISPATCHER-EXIT nil))

     (flet ((GETARGS (xdr-in locf-of-in)
	     ;; common routine for reading procedure argument LOCF-OF-IN filtered by XDR-IN
	     (declare (values value-of-in))
	     (if (null (send self :getargs xdr-in locf-of-in))
		 (progn (send self :svcerr-decode)
			(throw 'DISPATCHER-EXIT nil))
		 (contents locf-of-in)))


	    (FILE-ERROR-HANDLER (condition status)
	     ;; handle FS:FILE-ERROR by returning an NFSERR status unless SIGNAL-FILE-ERRORS-P is true
	     (declare (ignore condition))
	     (when (not (send xdr-stream :signal-file-errors-p)) (throw-nfserr-exit status xdr-stream)))
	      

	    (SETATTR-INTERNAL (sattr pathname &optional open-stream)
              ;; Attaches SATTR attributes to PATHNAME.  If OPEN-STREAM is non-NIL, it is
             ;; a stream already opened to PATHNAME and will be closed on exit"
             (declare (values ignore))
	     (let ((properties () )
		   (characters nil)
		   (byte-size  nil)
		   (mode       nil))
	       ;; handle special case of SIZE
	       (when (/= (sattr-size sattr) SATTR-IGNORED)
		 (if (streamp open-stream)
		     (setf characters (send open-stream :characters)
			   byte-size  (send open-stream :byte-size))
		     (setf characters (getf (rest (send pathname :properties)) :characters)
			   byte-size  (getf (rest (send pathname :properties)) :byte-size)))
		 (with-open-stream (file-stream (or open-stream
						    (open pathname :direction :output
							  :characters characters
							  :byte-size  byte-size  
							  :if-exists :OVERWRITE
							  :if-does-not-exist :ERROR)))
		   (send file-stream :set-pointer
			 (truncate (* 8 (sattr-size sattr)) (send file-stream :byte-size)))))
	       (when (streamp open-stream) (close open-stream))	   ;make sure its closed even if no SATTR-SIZE
	       
	         ;; handle general file properties
	       (when (/= (sattr-uid sattr) SATTR-IGNORED)
		 (setf properties (list* :UID (sattr-uid sattr) properties)))
	       (when (/= (sattr-gid sattr) SATTR-IGNORED)
		 (setf properties (list* :GID (sattr-gid sattr) properties)))
	       (when (/= (sattr-mtime-sec sattr) SATTR-IGNORED)
		 (setf properties (list* :CREATION-DATE (rpc:unix-lisp-ut (sattr-mtime-sec sattr))
					 properties)))
			 
	       (when (and  (/= (sattr-mode sattr) SATTR-IGNORED)
			   (/= (sattr-mode sattr) #xffff))
		 (if (directory-file-p pathname)
		   (setf mode (logior S-IFDIR (sattr-mode sattr)))
		   (setf mode (logior S-IFREG (sattr-mode sattr))))
		 (setf properties (list* :MODE mode properties)))

	       (apply #'fs:change-file-properties pathname t properties)))


	    (SENDREPLY-DIROPRES (attributes fhandle)
	     ;; collects ATTRIBUTES and FHANDLE into a DIROPRES structure and sends it as an NFS-OK reply
	     (setf (diropres-status     diropres-struct) NFS-OK
		   (diropres-file       diropres-struct) fhandle
		   (diropres-attributes diropres-struct) attributes)
	     (send self :sendreply #'xdr-diropres diropres-struct))
	      );flet bindings


       ;; IMPLEMENTATION NOTE:  In general, any FS:FILE-ERRORs signaled during exection of the Dispatcher
         ;; should be reported as an NFS status other then NFS-OK.  Which NFS status you return for which
         ;; error does not seems to be context sensitive.  Thererfore, no signals are handled in the Dispatcher's
         ;; action routines.  Instead, they are all intercepted on exit from the Dispatcher, override any partially
         ;; completed output the action routine may have created, and an appropriate NFS error status is 
         ;; returned instead.
       (condition-bind ((fs:access-error            #'file-error-handler NFSERR-ACCES)
			(fs:directory-not-empty     #'file-error-handler NFSERR-NOTEMPTY)
			(fs:dont-delete-flag-set    #'file-error-handler NFSERR-ACCES)
			(fs:file-already-exists     #'file-error-handler NFSERR-EXIST)
			(fs:file-locked             #'file-error-handler NFSERR-ACCES)
			(fs:file-lookup-error       #'file-error-handler NFSERR-NOENT)
			(fs:file-open-for-output    #'file-error-handler NFSERR-ACCES)
			(fs:filepos-out-of-range    #'file-error-handler NFSERR-NODEV)
			(fs:invalid-operation-for-directory
	                                            #'file-error-handler NFSERR-ISDIR)
			(fs:invalid-pathname-syntax #'file-error-handler NFSERR-NOENT)
			(fs:login-problems          #'file-error-handler NFSERR-ACCES)
			(fs:no-file-system          #'file-error-handler NFSERR-NODEV)
			(fs:no-more-room            #'file-error-handler NFSERR-NOSPC)
			(fs:not-available           #'file-error-handler NFSERR-NXIO)
			(fs:rename-to-existing-file #'file-error-handler NFSERR-EXIST)
			(fs:superior-not-directory  #'file-error-handler NFSERR-NOTDIR)
			(fs:file-error              #'file-error-handler NFSERR-IO))
	 ;; make sure we're logged in properly
	 (when (and fs:server-login	   ; we are supposed to automatically login to this host
		    (or (null user-id)	   ; no one'e logged in now
			(zerop (length user-id))))
	    ;; then no one's logged in and someone needs to be
	   (login fs:server-login-id si:local-host t)  ; login
	   (fs:print-server-login-exegesis))           ; tell them we're only file serving
	 
	 ;; set lup previous request id 
	 (setf *previous-request-id* (rpc:rpc-msg-xid rpc:request))

	 ;; execute processor based on procedure number
	 (case (rpc:svc-req-procedure request)
	       
	   (#.NFSPROC-CREATE		   ; case of create file
	    (getargs #'xdr-create-arg (locf create-arg-struct))
	    (let* ((diroparg         (create-arg-where create-arg-struct))
		   (fh-pathname      (or (fhandle-pathname (diropargs-dir diroparg) t)
					 (throw-nfserr-exit NFSERR-STALE)))
		   (name             (diropargs-name diroparg))
		   (pathname         nil)
		   (canonical-type   nil)
		   (charactersp      nil)  
		   (binary-16-type-p nil)  ; TRUE=> type *KNOWN* to be (unsigned-byte 16), non-character
		   (binary-8-type-p  nil)) ; TRUE=> type *KNOWN* to be (unsigned-byte 8), non-character
	      (validate-fh-pathname-name fh-pathname name) 
	      (setf pathname (make-plausible-fhandle-name-file fh-pathname name nil))
	      (when (directory-file-p pathname) (throw-nfserr-exit NFSERR-ISDIR))

	      ;; otherwise, this does not appear to be a retry
	      (setf canonical-type (send pathname :canonical-type)
	            binary-16-type-p (member canonical-type fs:*copy-file-known-binary-types*)
	            binary-8-type-p  (member canonical-type fs:*copy-file-known-short-binary-types*)
		    charactersp      (member canonical-type fs:*copy-file-known-text-types*))
	      (with-open-file (new-file pathname :direction :output
					:byte-size  (if binary-16-type-p 16 8)
					:characters (if charactersp t nil)
					:if-does-not-exist :CREATE :if-exists :TRUNCATE)
		(setattr-internal (create-arg-attributes create-arg-struct) pathname new-file))
	      (sendreply-diropres (send pathname :properties t) (pathname-fhandle pathname))
	      ))

	   
	   (#.NFSPROC-GETATTR		   ; case of get file attributes
	     (getargs #'xdr-fhandle (locf fhandle-struct))
	     (let ((fh-pathname  (or (fhandle-pathname fhandle-struct t)
				     (throw-nfserr-exit NFSERR-STALE))))

	       (close-stream-if-open fh-pathname)
	       (if (not (root-directory-p fh-pathname))
		   (send self :sendreply #'xdr-attrstat (send fh-pathname :properties t))
		   (send self :sendreply #'xdr-attrstat `(,*local-root-pathname*
                                                          :DIRECTORY t
							  :BYTE-SIZE 8
							  :LENGTH-IN-BLOCKS 1
							  :LENGTH-IN-BYTES 512
							  :AUTHOR "NFS"
							  :CHARACTERS nil
							  :CREATION-DATE 2745945450
							  :NOT-BACKED-UP t)))))

	   (#.NFSPROC-LOOKUP		   ; case of look up another node of file handle, return new fhandle
	    (getargs #'xdr-diropargs (locf diropargs-struct))
	    (let* ((fh-pathname  (or (fhandle-pathname (diropargs-dir diropargs-struct) t)
				     (throw-nfserr-exit NFSERR-STALE)))
		   (name         (diropargs-name diropargs-struct))
		   (pathname     nil))
	      (validate-fh-pathname-name fh-pathname name)
	      (setf pathname (or (make-fhandle-name-subdirectory   fh-pathname name t)
				 (make-plausible-fhandle-name-file fh-pathname name t)))
	      (when (not (pathnamep pathname)) (throw-nfserr-exit NFSERR-NOENT))   ; couldn't derive a pathname

	      (close-stream-if-open pathname)
	      (if (not (root-directory-p pathname))
		  (sendreply-diropres (send pathname :properties t) (pathname-fhandle pathname))
		  (sendreply-diropres `(,*local-root-pathname*
					:DIRECTORY t
					:BYTE-SIZE 8
					:LENGTH-IN-BLOCKS 1
					:LENGTH-IN-BYTES 512
					:AUTHOR "NFS"
					:CHARACTERS nil
					:CREATION-DATE 2745945450
					:NOT-BACKED-UP t) (pathname-fhandle pathname)))))
	   

	   (#.NFSPROC-MKDIR		   ; case of create directory
	    (getargs #'xdr-create-arg (locf create-arg-struct))
	    (let* ((diroparg    (create-arg-where create-arg-struct))
		   (fh-pathname (or (fhandle-pathname (diropargs-dir diroparg) t)
				    (throw-nfserr-exit NFSERR-STALE)))
		   (name        (diropargs-name diroparg))
		   (pathname    nil))
	      (validate-fh-pathname-name fh-pathname name)
	      (setf pathname (make-fhandle-name-subdirectory fh-pathname name nil))
	      (when (not (directory-file-p pathname))    (throw-nfserr-exit NFSERR-NOTDIR))
	      (when (file-exists-p pathname)             (throw-nfserr-exit NFSERR-EXIST))
	      (fs:create-directory (send pathname :pathname-as-directory) :error t :recursive nil)
	      (setattr-internal (create-arg-attributes create-arg-struct) pathname nil)
	      (sendreply-diropres (send pathname :properties t) (pathname-fhandle pathname))))


	   (#.NFSPROC-READ		   ; case of read file
	    (getargs #'xdr-read-arg (locf read-arg-struct))
	    (send self :sendreply :PREAMBLE-ONLY nil)
	    (nfsproc-read-processor read-arg-struct) ; this will filter its own outputs
	    (send self :sendreply :POSTAMBLE-ONLY nil))


	   (#.NFSPROC-RMDIR		   ; case of delete directory
	    (getargs #'xdr-diropargs (locf diropargs-struct))
	    (let ((fh-pathname (or (fhandle-pathname (diropargs-dir diropargs-struct) t)
				   (throw-nfserr-exit NFSERR-STALE)))
		  (name        (diropargs-name diropargs-struct))
		  (pathname    nil))
	      (validate-fh-pathname-name fh-pathname name)
	      (setf pathname (make-fhandle-name-subdirectory fh-pathname name t))
	      (when (null pathname)                      (throw-nfserr-exit NFSERR-NOENT))
	      (when (not (directory-file-p pathname))    (throw-nfserr-exit NFSERR-NOTDIR))
	      (send pathname :delete-and-expunge t)
	      (delete-fhandle pathname)
	      (send self :sendreply :xdr-enum NFS-OK)))


	   (#.NFSPROC-ROOT		   ;case of get root (obsolete)
	    (send self :sendreply :xdr-void nil))
	   

	   (#.NFSPROC-SETATTR		   ; case of set file attributes
	    (getargs #'xdr-setattr-arg (locf setattr-arg-struct))
	    (let* ((fh-pathname (or (fhandle-pathname (setattr-arg-file setattr-arg-struct) t)
				    (throw-nfserr-exit NFSERR-STALE)))
		   (sattr       (setattr-arg-attributes setattr-arg-struct)))
	      
	      (close-stream-if-open fh-pathname)
	      (setattr-internal sattr fh-pathname nil)
	      (send self :sendreply #'xdr-attrstat (send fh-pathname :properties t))))

	   
	   (#.NFSPROC-READDIR		   ; case of directory read
	    (getargs #'xdr-readdir-arg (locf readdir-arg-struct))
	    (send self :sendreply :PREAMBLE-ONLY nil)
	    (nfsproc-readdir-processor readdir-arg-struct)
	    (send self :sendreply :POSTAMBLE-ONLY nil))
	   

	   (#.NFSPROC-REMOVE		   ; case of delete a file
	    (getargs #'xdr-diropargs (locf diropargs-struct))
	    (let ((fh-pathname (or (fhandle-pathname (diropargs-dir diropargs-struct) t)
				   (throw-nfserr-exit NFSERR-STALE)))
		  (name        (diropargs-name diropargs-struct))
		  (pathname    nil))

	      (validate-fh-pathname-name fh-pathname name)
	      (setf pathname (make-plausible-fhandle-name-file fh-pathname name t))
	      (when (null pathname)             (throw-nfserr-exit NFSERR-NOENT))
	      (when (directory-file-p pathname) (throw-nfserr-exit NFSERR-ISDIR))
	      (close-stream-if-open pathname)
	      (send pathname :delete-and-expunge t)
	      (delete-fhandle pathname)
	      (send self :sendreply :xdr-enum NFS-OK)))
	   

	   (#.NFSPROC-RENAME		   ; case of rename a file
	    (send self :sendreply :xdr-enum
		  (nfsproc-rename-processor (getargs #'xdr-rename-arg (locf rename-arg-struct)))))
	   

	   (#.NFSPROC-STATFS		   ; case of file system status procedure, return file system attributes
	    (setf (statfs-reply-status statfs-reply-struct) NFS-OK
		  ;; other slots are defaulted to constant values
	          (statfs-reply-bfree  statfs-reply-struct) (aref fs::put-usage-array fs:put-free)
	          (statfs-reply-bavail statfs-reply-struct) (aref fs::put-usage-array fs:put-free))
	    (send self :sendreply #'xdr-statfs-reply statfs-reply-struct))


	   (#.NFSPROC-WRITE		   ;case of write to a file
	    (send self :sendreply #'xdr-attrstat (nfsproc-write-processor)))
	    

	   (#.NFSPROC-WRITECACHE	   ;case of write to cache (obsolete)
	    (send self :sendreply :xdr-void nil))
	   

	   (otherwise			   ; case of unsupported procedure, return error
	    ;;, NFSPROC-LINK, NFSPROC-SYMLINK, & NFSPROC-READLINK
	    (send self :svcerr-noproc))
	   );case
	 );condition-bind
         );flet
     );catch
    )
))
