;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: This patch duplicates for Explorer NFS 3.0 the fix in microExplorer NFS
;;; PATCH-4-37.  If an Explorer is browsing a remote NFS file system and comes
;;; upon a symbolic link which points back to the Explorer, it now signals an
;;; NFSERR-NOLINK error rather than trying to chase the link on the Explorer.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for NFS version 3.3
;;; Written 01/09/89 15:19:55 by CORNISH,
;;; while running on Bravo from band LOD1
;;; With SYSTEM 5.9, VIRTUAL-MEMORY 5.5, EH 5.3, MAKE-SYSTEM 5.3, MICRONET 5.5, LOCAL-FILE 5.0,
;;;  BASIC-PATHNAME 5.3, NETWORK-SUPPORT-COLD 5.1, BASIC-NAMESPACE 5.8, NETWORK-NAMESPACE 5.0,
;;;  DISK-IO 5.9, DISK-LABEL 5.1, BASIC-FILE 5.4, MAC-PATHNAME 5.0, NETWORK-PATHNAME 5.0,
;;;  COMPILER 5.2, TV 5.24, DATALINK 5.7, CHAOSNET 5.6, GC 5.4, MEMORY-AUX 5.1, NVRAM 5.1,
;;;  SYSLOG 5.2, STREAMER-TAPE 5.7, UCL 5.0, INPUT-EDITOR 5.0, METER 5.0, ZWEI 5.11,
;;;  DEBUG-TOOLS 5.1, Experimental NETWORK-SUPPORT 5.1, Experimental NETWORK-SERVICE 5.1,
;;;  DATALINK-DISPLAYS 5.0, FONT-EDITOR 5.0, SERIAL 5.0, PRINTER 5.12, MAC-PRINTER-TYPES 5.4,
;;;  PRINTER-TYPES 5.5, IMAGEN 5.1, SUGGESTIONS 5.0, MAIL-DAEMON 5.5, MAIL-READER 5.5,
;;;  TELNET 5.2, VT100 5.1, NAMESPACE-EDITOR 5.1, PROFILE 5.1, VISIDOC 5.4, IP 3.37,
;;;  Experimental BUG 10.2, RPC 6.0, NFS 3.2,  microcode 398, Band Name: Rel 5.0 + IP 10/7

#!C
; From file SIGNALS.LISP NFS; NFS:
#10R NETWORK-FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NETWORK-FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "NFS: NFS; SIGNALS.#"


;; Added this for a link bug -- 6/17/88 
(DEFSIGNAL NFSERR-NOLINK
	     (fs:file-lookup-error fs:file-not-found fs:directory-not-found nfserr-nolink nfs-error
				   nfserr-noent)
	     (pathname operation linkname)
   "symbolic link file or directory not found"
   :format-string "The file or a directory in \"~2@*~A\" does not exist.")

))

#!C
; From file SERVICE_FUNCTIONS.LISP NFS; NFS:
#10R NETWORK-FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NETWORK-FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "NFS: NFS; SERVICE_FUNCTIONS.#"


(DEFUN NFS-TRUENAME (nfs-client pathname &optional credentials)
   (declare (values truename))

   (let ((rpc:*call-who-state* (or rpc:*call-who-state* 
				   *probe-open-who-state*))
	 (ancestors-list       ()))	   ; recursion history to detect circular links
     
     (labels
       ((FILTER-./ (namestring)
	  ;; If NAMESTRING is in the form "./foo", then return "foo", else return NAMESTRING
	  ;; PROGRAM MAINTENANCE NOTE:  This function can be removed once we can parse "./foo"
	  ;; without signalling an error.
	  (if (and (char= #\. (char namestring 0))
		   (char= #\/ (char namestring 1)))
	      (subseq namestring 2)
	      namestring))
	      
	(DEFAULT-REPLY-PATH (linkname reply-path)
	   ;; return tenative-truename of REPLY-PATH defaulted by LINKNAME with Apollo filters 
	   (declare (values tenative-truename))
	   (block :default-reply-path
	     (let ((pathname (parse-namestring (filter-./ reply-path) (pathname-host linkname))))
	       (if (and (char= #\/ (char reply-path 0))
			*resolve-remote-symbolic-link-on-lm-p*)
		    ;; then pretend the root or this reply-path is the local machine as NFS says
		  (signal 'nfserr-nolink nil pathname :truename
				 (send pathname :new-pathname :host sys:local-host))

		   ;; else resolve it with respect to where we found it
		   (merge-pathnames pathname linkname)))))

		 
	      (FOLLOW-LINK (linkname cred fhandle)
		;; return the truename of LINKNAME
		(declare (values truename))
		(let ((rpc:*call-who-state* *readlink-who-state*)
		      (reply                nil))
		   ;; read the symbolic link and check reply
		  (send (send nfs-client :new-xid) :call-with-retry NFSPROC-READLINK
			#'xdr-fhandle        fhandle
			#'xdr-readlink-reply (locf reply) cred)
		  (when (/= NFS-OK (nfs-status reply))
		    (signal-nfserr (nfs-status reply) linkname :truename
				   :proceed-types '(:retry-file-operation :new-pathname)))
		  (setf linkname
			(default-reply-path linkname (readlink-reply-path reply))) 
		  (when (member linkname ancestors-list :test #'eq)
		    (signal 'FS:CIRCULAR-LINK *circular-link-msg*
			    (first ancestors-list) :TRUENAME (copy-list ancestors-list)))
		  (truename-internal linkname cred)))
	      
			 
	      (TRUENAME-INTERNAL (tenative-truename cred)
		 ;; returns the truename of TENATIVE-TRUENAME
		 (validate-pathname pathname :truename)
		 (push tenative-truename ancestors-list)
		 (with-fhandle (fhandle nfs-client tenative-truename cred)
		   (if (/= S-IFLNK (ifmt-mode (get-file-attributes nfs-client tenative-truename cred)))
		       ;; then this is not a link, so return tenative-truename as THE truename
		       tenative-truename
		      
		       ;; else this is a link, so read from link to find its truename
		       (follow-link tenative-truename cred fhandle))))
	        );labels bindings

       
       (catch 'no-link
	 (condition-resume `((nfserr-noent fs:circular-link relative-pathnames-not-allowed
					   rpc:proc-unavail)
			     :return-last-link
			     t (,*use-last-link-proceed-msg*)
			     (lambda (condition)
			       (throw 'no-link (send condition :pathname))))
	   (with-credentials (cred pathname credentials)
	     (truename-internal pathname cred))))
         );labels
       );let
   );nfs-truename

))
