;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Changed GET-USER-NAMES to correctly calculate the size read data it could
;;; accept.  It previously was not allowing space in the return buffer for the
;;; RPC READ header.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/03/89 17:17:52 by CORNISH,
;;; while running on Bravo from band LOD1
;;; With SYSTEM 5.9, VIRTUAL-MEMORY 5.5, EH 5.3, MAKE-SYSTEM 5.3, MICRONET 5.5, LOCAL-FILE 5.0,
;;;  BASIC-PATHNAME 5.4, NETWORK-SUPPORT-COLD 5.1, BASIC-NAMESPACE 5.8, NETWORK-NAMESPACE 5.0,
;;;  DISK-IO 5.9, DISK-LABEL 5.1, BASIC-FILE 5.4, MAC-PATHNAME 5.1, Experimental NETWORK-PATHNAME 5.1,
;;;  COMPILER 5.2, TV 5.24, DATALINK 5.7, CHAOSNET 5.6, GC 5.4, MEMORY-AUX 5.1, NVRAM 5.1,
;;;  SYSLOG 5.2, STREAMER-TAPE 5.7, UCL 5.0, INPUT-EDITOR 5.0, METER 5.1, ZWEI 5.12,
;;;  DEBUG-TOOLS 5.1, Experimental NETWORK-SUPPORT 5.1, Experimental NETWORK-SERVICE 5.1,
;;;  DATALINK-DISPLAYS 5.0, FONT-EDITOR 5.0, SERIAL 5.0, PRINTER 5.12, MAC-PRINTER-TYPES 5.4,
;;;  PRINTER-TYPES 5.5, IMAGEN 5.2, SUGGESTIONS 5.0, MAIL-DAEMON 5.6, MAIL-READER 5.6,
;;;  TELNET 5.2, VT100 5.1, NAMESPACE-EDITOR 5.1, PROFILE 5.1, VISIDOC 5.4, IP 3.41,
;;;  Experimental BUG 10.6, Experimental RPC 6.0, NFS 3.4,  microcode 398, Band Name: Rel 5.0 + IP 10/7

#!C
; From file MISC_FUNCTIONS.LISP NFS; SYS:
#10R NETWORK-FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NETWORK-FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NFS; MISC_FUNCTIONS.#"


(DEFUN GET-USER-NAMES (host-or-file-spec credentials)
   "Returns an alist of keys which are a two element list of UID and GID numbers dotted with a two element list
of user name string and home directory namestring that represents the contents of the /etc/passwd file on the host.
HOST-OR-FILE-SPEC is a host object,a pathname object, or a namestring.  If HOST-OR-FILE-SPEC is a pathname
or namestring, it is read in place of /etc/passwd.  Returns NIL if password file cannot be read."
   (declare (values uid-gid-user-names-alist))
   
   (let ((pathname                nil)
	 (host                    nil)
	 (uid-gid-user-name-alist () )
	 (passwd-file-size        nil)
	 (client-buffer-limit     nil)	   ; size in bytes of buffer allocated in client instance (must
					   ;   hold RPC packet header and all variable arguments).
	 (read-count-limit        nil))	   ; size in bytes of maximum space in client buffer which is - 03 Feb 89 Mac
					   ;   available for read reply data
     
     ;; default arguments, as required
     (if (typep host-or-file-spec 'net:host)
	 (setf host     host-or-file-spec
	       pathname (parse-namestring "/etc/passwd" host nil))
	 (setf host     (pathname-host pathname)
	       pathname (translated-pathname host-or-file-spec)))
     
     (condition-case ()
	 ;; find out how big it is and allocate it a buffer
	 (rpc:with-client (nfs-client host NFSPROG NFSVERS (rpc:get-host-protocol host))
	   (setf client-buffer-limit (array-total-size (send nfs-client :memory-buffer)))
	   (setf read-count-limit	                         ; 03 Feb 89 Mac
		 (- (min client-buffer-limit	                 ; total can't be bigger than CLIENT's buffer
			 (rpc:protocol-max-msg-size              ; total can't be bigger than protocol's maximum
			   (send nfs-client :protocol-keyword)))
		    (+ RPC:MAX-REPLY-HEADER READ-REPLY-HEADER))) ; must leave room for READ reply header
	   (with-credentials (cred pathname credentials)
	     ;; try to access the password file
	     (setf passwd-file-size (fattr-size (get-file-attributes nfs-client pathname cred)))
	 
	     (using-resource (data-buffer NFS-DATA-BUFFER (min passwd-file-size
							       client-buffer-limit))
	       (with-fhandle (fhandle nfs-client pathname cred)
		 (let* ((buffer       (make-array (length data-buffer) :element-type 'string-char
						  :displaced-to data-buffer))
			(read-arg     (make-read-arg   fhandle
						       0           ; OFFSET updated each iteration
						       0))         ; COUNT updated each iteration
			(read-reply   (make-read-reply 0           ; STATUS updated by each read
						         nil	   ; ATTRIBUTES updated by each successful read
						         buffer))	   ; DATA updated by each successful read
			(rpc:*call-who-state*          *read-who-state*)
			(delimiter-indices
			  (make-array 6 :element-type 'fixnum))
			(line-start      0)		   ; substring indices delimiting a line within a buffer
			(line-end        nil))
	     
		   (loop			   ; loop for each buffer of read data
		       ;; initialize this iteration and return if finished
		     (setf (read-arg-count read-arg)    ; 03 Feb 89 Mac
			   (rpc:round-to-block	       ; count must be multiple of 512
			     (min read-count-limit     ; can't be larger than largest legal read
				  (- passwd-file-size (read-arg-offset read-arg))))) ;not larger than remainder
		     (when (zerop (read-arg-count read-arg)) (return))	   ; return when there's nothing left to read
		 
		       ;; read the block and return NIL for function if any error
		     (assert (or (> 512 (read-arg-count read-arg))
				 (zerop (logand #x1FF (read-arg-count read-arg)))) ()
			     "READ-ARG-COUNT=~D. which is not a positive multiple of 512."
			     (read-arg-count read-arg))
		     (send (send nfs-client :new-xid) :call NFSPROC-READ
			   #'xdr-read-arg   read-arg
			   #'xdr-read-reply (locf read-reply) cred)
		     (when (/= NFS-OK (nfs-status read-reply)) (return-from get-user-names nil))
		     (setf line-start 0)

		       (loop			   ; loop for all lines in a buffer
		          ;; initialize this iteration and return if finished
		        (setf line-end (position #\
 buffer :start line-start
						 :end (read-arg-count read-arg))); #\
 = UNIX NL character
			(when (null line-end) (return))

			(when (>= (count #\: buffer :start line-start :end line-end) 6)
			   ;; then this line has enough of a known format to be safely parsed
			  (do ((i line-start (1+ i))
			       (n 0))
			      ((>= n 6))
			    (when (char= #\: (char buffer i))
			      (setf (aref delimiter-indices n) i)
			      (incf n))
			      );do

			  (setf uid-gid-user-name-alist
				(acons (list (read-from-string
					       buffer nil nil	   ; UID
					       :start (1+ (aref delimiter-indices 1))
					       :end (aref delimiter-indices 2))
					     (read-from-string
					       buffer nil nil	   ; GID
					       :start (1+ (aref delimiter-indices 2))
					       :end (aref delimiter-indices 3)))
				       (list (subseq buffer
						     line-start (aref delimiter-indices 0))
					     (subseq buffer (1+
							      (aref delimiter-indices 4))
						     (aref
						       delimiter-indices 5)))
				       uid-gid-user-name-alist))
			   );when

			;; finish up the iteration for this line
			(setf line-start (1+ line-end))
			);loop
		       ;; finish up the iteration for this buffer
		      (incf (read-arg-offset read-arg) line-start)
		        );loop
		   (setf (file-attributes nfs-client pathname) (read-reply-attributes read-reply))))
	         );using-resources
	       );with-credentials
	    );with-client
         ;; condition-case cases
       ((fs:directory-not-found fs:file-not-found)
	(when (typep host-or-file-spec 'net:host)
	   ;; then we were using the default and it wasn't there, make sure we don't bother to try again
	  (setf (user-names host)  '(nil)))
	(return-from get-user-names nil))
       ((lisp:error sys:abort) (return-from get-user-names nil))
       );condition-case

     (setf (user-names host)  (nreverse uid-gid-user-name-alist))
     );;let
    )
))
