;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: UNMOUNT-FHANDLE modified to that it returns quietly if it is asked to unmount a
;;; file handle which is not currently in the Mount Table.  It previously signaled an
;;; error which the server ignored but which was reported back to the client and an
;;; unspecific system error.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 03/09/89 13:59:06 by CORNISH,
;;; while running on Bravo from band LOD1
;;; With SYSTEM 5.9, VIRTUAL-MEMORY 5.5, EH 5.3, MAKE-SYSTEM 5.3, MICRONET 5.6, LOCAL-FILE 5.0,
;;;  BASIC-PATHNAME 5.4, NETWORK-SUPPORT-COLD 5.1, BASIC-NAMESPACE 5.8, NETWORK-NAMESPACE 5.0,
;;;  DISK-IO 5.9, DISK-LABEL 5.1, BASIC-FILE 5.5, MAC-PATHNAME 5.1, Experimental NETWORK-PATHNAME 5.1,
;;;  COMPILER 5.3, TV 5.27, DATALINK 5.7, CHAOSNET 5.6, GC 5.4, MEMORY-AUX 5.1, NVRAM 5.1,
;;;  SYSLOG 5.2, STREAMER-TAPE 5.7, UCL 5.0, INPUT-EDITOR 5.0, METER 5.1, ZWEI 5.13,
;;;  DEBUG-TOOLS 5.1, Experimental NETWORK-SUPPORT 5.1, Experimental NETWORK-SERVICE 5.1,
;;;  DATALINK-DISPLAYS 5.0, FONT-EDITOR 5.0, SERIAL 5.0, PRINTER 5.12, MAC-PRINTER-TYPES 5.4,
;;;  PRINTER-TYPES 5.5, IMAGEN 5.3, SUGGESTIONS 5.0, MAIL-DAEMON 5.6, MAIL-READER 5.6,
;;;  TELNET 5.2, VT100 5.1, NAMESPACE-EDITOR 5.1, PROFILE 5.1, VISIDOC 5.4, IP 3.41,
;;;  Experimental BUG 10.7, Experimental RPC 6.0, NFS 3.5,  microcode 398, Band Name: Rel 5.0 + IP 10/7

#!C
; From file MOUNT_SERVER.LISP NFS; SYS:
#10R NETWORK-FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NETWORK-FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS:NFS;MOUNT_SERVER.."


(DEFUN UNMOUNT-FHANDLE (fhandle client)
   "Remove CLIENT from FHANDLE's mount list entry and then remove the entire entry if clients remain.
 - If FHANDLE is :WILD, then all of this clients FHANDLEs are unmounted.
 - If CLIENT is :WILD, then the FHANDLE's entry is removed.
 - If they are both :WILD, the whole *MOUNT-TABLE* is emptied.
 - If FHANDLE is not in the mount list or if CLIENT is not in FHANDLE's client host list, then does nothing.
NOTE: Unmounting does not effect the file handles interned in *FHANDLE-TABLE*.
Once made, those entires are never deleted or changed."
   (declare (values ignore))
   (without-interrupts
     (if (eq client :WILD)
	  ;; then remove one or more FHANDLE entries
	 (if (eq fhandle :WILD)
	       ;; then remove all entries
	     (setf *mount-table* nil)

	       ;; else just remove this FHANDLE's entry
	     (setf *mount-table* (delete fhandle *mount-table* :test #'fhandle= :key #'first)))
       
	  ;; else remove one or client entries
	 (if (eq :WILD fhandle)
	       ;; then remove all FHANDLEs for this CLIENT
	     (dolist (mount-entry *mount-table*)
	       (setf (rest mount-entry) (delete client (rest mount-entry) :test #'eq)))

	       ;; else remove this FHANDLE for this CLIENT
	     (let ((fhandle-clients (assoc fhandle *mount-table* :test #'fhandle=)))
	       (when (not (null fhandle-clients))  ; 9 Mar 89 - Mac
		 ;; then we found something we can delete
		 (setf (rest fhandle-clients) (delete client (rest fhandle-clients) :test #'eq)))))
	  );if
       ;; purge the Mount List 
     (setf *mount-table* (delete nil *mount-table* :key #'rest :test #'eq))
     );without-interrupts
   );unmount-fhandle

))
