;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Modified MAKE-PLAUSIBLE-FHANDLE-NAME-FILE to default the type component
;;; of the parsed name argument to be :UNSPECIFIC rather than NIL.  This should
;;; fix the problem of parsing MS-DOS pathnames which may or may not have a
;;; file extension.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for NFS version 3.7
;;; Written 03/14/89 14:10:55 by CORNISH,
;;; while running on Bravo from band LOD1
;;; With SYSTEM 5.9, VIRTUAL-MEMORY 5.5, EH 5.3, MAKE-SYSTEM 5.3, MICRONET 5.6, LOCAL-FILE 5.0,
;;;  BASIC-PATHNAME 5.4, NETWORK-SUPPORT-COLD 5.1, BASIC-NAMESPACE 5.8, NETWORK-NAMESPACE 5.0,
;;;  DISK-IO 5.9, DISK-LABEL 5.1, BASIC-FILE 5.5, MAC-PATHNAME 5.1, Experimental NETWORK-PATHNAME 5.1,
;;;  COMPILER 5.3, TV 5.27, DATALINK 5.7, CHAOSNET 5.6, GC 5.4, MEMORY-AUX 5.1, NVRAM 5.1,
;;;  SYSLOG 5.2, STREAMER-TAPE 5.7, UCL 5.0, INPUT-EDITOR 5.0, METER 5.1, ZWEI 5.13,
;;;  DEBUG-TOOLS 5.1, Experimental NETWORK-SUPPORT 5.1, Experimental NETWORK-SERVICE 5.1,
;;;  DATALINK-DISPLAYS 5.0, FONT-EDITOR 5.0, SERIAL 5.0, PRINTER 5.12, MAC-PRINTER-TYPES 5.4,
;;;  PRINTER-TYPES 5.5, IMAGEN 5.3, SUGGESTIONS 5.0, MAIL-DAEMON 5.6, MAIL-READER 5.6,
;;;  TELNET 5.2, VT100 5.1, NAMESPACE-EDITOR 5.1, PROFILE 5.1, VISIDOC 5.4, IP 3.41,
;;;  Experimental BUG 10.7, Experimental RPC 6.0, NFS 3.5,  microcode 398, Band Name: Rel 5.0 + IP 10/7

#!C
; From file NFS_SERVER.LISP NFS; SYS:
#10R NETWORK-FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NETWORK-FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NFS; NFS_SERVER.#"


(DEFUN MAKE-PLAUSIBLE-FHANDLE-NAME-FILE (fh-pathname name &optional verify-existence-p)
   "FH-PATHNAME is a pathname object of a directory file and NAME is a string identifying a file in
FH-PATHNAME.  NAME is either a UNIX-like name component or an Explorer-like name-type-version
namestring fragment.  Returns a UNIX-like pathname with type=:UNSPECIFIC and version=1 or returns an
Explorer-like pathname with type and version extracted from NAME depending upon the appearance of NAME.
If VERIFY-EXISTENCE-P is true, return NIL if the file does not exist."
   (declare (values file-pathname-or-nil)
	    (string name))
   (when (not (root-directory-p fh-pathname))
     (setf fh-pathname (send fh-pathname :pathname-as-directory)))

   ;;; IMPLEMENTATION NOTE:  The purpose of parsing the name component with respect to the caller
   ;;; is that :GETCALLER should return us the true host object which will tell us if the remote host is
   ;;; really a non-UNIX host hiding behind an NFS-UNIX facade.  For example, if the foreign host is really
   ;;; an Explorer, name will be parsed with # representing a version number introducer.  However, if the 
   ;;; remote host is truly a UNIX machine, then # is just another character.

   (let* ((parsed-name (fs:parse-pathname name (send self :getcaller)))
	  (pathname    (make-pathname :host si:local-host
					:directory (pathname-directory fh-pathname)
					:name      (pathname-name parsed-name)
					:type      (or (pathname-type parsed-name)
						       :UNSPECFIC) ; 14 Mar 89
					:version :NEWEST)))
          
     ;; now, verify existence, if needed
     (if verify-existence-p
       (when (file-exists-p pathname) pathname)
       pathname)
     );;let
   );;make-plausible-fhandle-name-file 

))
