;;; -*- Mode: Common-Lisp; Package: NFS; Base: 10.; Patch-File: T -*-

;;; Reason: Changed NFS-DIRECTORY-LIST to not be case sentitive with pathnames from Explorer and microExplorers.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 07/13/89 07:18:33 by BERGER,
;;; while running on ARIES from band LOD1
;;; With SYSTEM 6.10, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.7, TV 6.12, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.3,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.1, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, TI-CLOS 6.11, CLEH 6.4, IP 3.47,
;;;  Experimental BUG 11.10, Experimental CLX 6.2, CLUE 6.9, X11M 6.1, RPC 6.1, NFS 3.8,
;;;   microcode 429, Band Name: 6.0 + patches 6/26

#!C
; From file SERVICE_FUNCTIONS.LISP NFS; SYS:
#10R NETWORK-FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NETWORK-FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NFS; SERVICE_FUNCTIONS.#"


(DEFUN NFS-DIRECTORY-LIST (nfs-client pathname options &optional credentials)
   (declare (values directory-list))
   
   (when (null (pathname-name pathname))
     (setf pathname (send pathname :new-pathname :name :wild)))
;   (when (eq (pathname-type pathname) :UNSPECIFIC)
;     (setf pathname (send pathname :new-pathname :type :wild)))
   (when (member :DIRECTORIES-ONLY options)
     (setf options (remove :DIRECTORIES-ONLY options))	   ; avoid infinite loop with :ALL-DIRECTORIES
     (return-from nfs-directory-list (nfs-all-directories nfs-client pathname options credentials)))
     
   (let* ((rpc:*call-who-state* (or rpc:*call-who-state* *read-directory-who-state*))
	  (superior-path        (superior-pathname pathname))
	  (client-buffer-limit  (array-total-size (send nfs-client :memory-buffer))))
     
     (with-credentials (cred superior-path credentials)
       (when (= S-IFLNK (ifmt-mode (get-file-attributes nfs-client superior-path cred)))
	 (setf (file-handle nfs-client superior-path cred) nil))
       (with-mounted-dir (superior-fhandle superior-path cred)
         (let* ((dir-list       () )
	       (entry-pathname nil)
	       (fast-p         (member :fast options))
	       (long-dir        (and (null fast-p)
				     (boundp 'si:local-lx-host)
				     (eq (send nfs-client :host) si:local-lx-host)))
	       (readdir-arg    (make-readdir-arg superior-fhandle
						 0 ; updated each iteration
						 (rpc:round-to-block
						   (- (min client-buffer-limit
							   (rpc:protocol-max-msg-size
							     (send nfs-client :protocol-keyword)))
						      RPC:MAX-REPLY-HEADER
						      READDIR-REPLY-HEADER))))  
	       (readdir-reply  () ))
	   (assert (and (<= 512 (readdir-arg-count readdir-arg))
			(zerop (logand #x1FF (readdir-arg-count readdir-arg)))) ()
		   "READDIR-ARG-COUNT=~D, which is not a positive multiple of 512."
		   (readdir-arg-count readdir-arg))
	   (loop
	       ;; read the next block of directory contents and check reply
	     (if long-dir
		 (send (send nfs-client :new-xid) :call-with-retry NFSPROC-LONGDIR
		       #'xdr-readdir-arg   readdir-arg
		       #'xdr-longdir-reply-fast-decode (locf readdir-reply) cred)
		 (send (send nfs-client :new-xid) :call-with-retry NFSPROC-READDIR
		       #'xdr-readdir-arg   readdir-arg
		       #'xdr-readdir-reply-fast-decode (locf readdir-reply) cred))
	     (when (integerp readdir-reply)
	       (signal-nfserr readdir-reply superior-path :DIRECTORY-LIST :pathname-type :DIRECTORY
			      :proceed-types '(:retry-file-operation :new-pathname)))

	     ;; create the basic directory list
	     (let ((rpc:*call-who-state* nil)
		   (entry-string nil)
		   ;;DAB 07-13-89 Case sentitivity is important for UNIX machines.
		   (string=-function (if (member (type-of pathname) '(FS:mac-pathname FS:lispm-pathname))
					 'string-equal 'string= ))
		     )
	       (dolist (entry (rest readdir-reply))
		 (setf entry-string (if long-dir (car entry) entry))
		 ;;Do not call merge-pathnames unconditionally
		 (if (or (eq (send pathname :name) :WILD)
			 (eq (send pathname :type) :WILD)
			 (apply string=-function entry-string (file-namestring pathname) ())) ; DAB 07-13-89
		   (setf entry-pathname
			 (merge-pathnames entry-string
					  ;;Make the default type :unspecific, not :wild.
					  (if (typep pathname 'fs:lispm-pathname)
					      pathname
					      (send pathname :new-type :unspecific)) ; DAB 07-13-89
					  ))
		   (setf entry-pathname nil))

		 (when (and (send pathname :pathname-match entry-pathname)
			    (string/= "."          entry-string)
			    (string/= ".."         entry-string)
			    (string/= "`node_data" entry-string))

		    ;; then this is one of the pathnames implied by the original (possibly wildcarded) pathname
		   (push (if fast-p
			     (list entry-pathname)   ; just the name, don't bother with properties
			     (if long-dir
				 (nfs-properties nfs-client entry-pathname cred (cdr entry))
				 (condition-case ()
				     (nfs-properties nfs-client entry-pathname cred)
				   ((fs:file-not-found fs:directory-not-found nfserr-acces)
				      ;; don't complain if we can't get the properties of some files...it happens
				    (list entry-pathname)))))
			 dir-list)
		    );when
		  );dolist
	       );let
	     (when (null (first readdir-reply)) (return))
	     
	       ;; first entry was really the last one sent, so remember its COOKIE for the next iteration
	     (setf (readdir-arg-cookie readdir-arg) (first readdir-reply))
	       );loop
	 
	    ;; sort directory list, if necessary
	   (when (member :sorted options)
	     (setf dir-list (sort dir-list #'fs:pathname-lessp :key #'first)))
	 
	   ;; prepend directory info and return
	   (cons (if (or fast-p (member :DIRECTORIES-ONLY options))
		     nil
		     (let* ((rpc:*call-who-state* nil)
			    (statfs-reply (get-filesystem-attributes nfs-client superior-path cred)))
		       (list nil
			     :DISK-SPACE-DESCRIPTION
			     (format nil "Total=~D., Free=~D., Available to Users=~D. (~D. bytes/block)"
				     (statfs-reply-blocks statfs-reply)
				     (statfs-reply-bfree  statfs-reply)
				     (statfs-reply-bavail statfs-reply)
				     (statfs-reply-bsize  statfs-reply))
			     :SETTABLE-PROPERTIES SETTABLE-PROPERTIES
			     :PATHNAME (condition-bind    ; don't die chasing circular or non-existent links 
					 (((nfserr-noent fs:circular-link rpc:proc-unavail
							 relative-pathnames-not-allowed)
					    #'(lambda (c) ; return last known link instead
					        (declare (ignore c))
						:return-last-link)))
					 (nfs-truename nfs-client superior-path cred))
			     :NFS-FILESYSTEM-ATTRIBUTES statfs-reply))
		       );if
		 dir-list)
	    );let
	  );with-mounted-dir
         );witn-credentials
       );let
   )

))

