;;; -*- Mode: Common-Lisp; Package: fs; Base: 10.; Patch-File: T -*-

;;; Reason: Increased timeout waiting for next packet in rfile-server before giving up.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 04/03/90 15:05:43 by BERGER,
;;; while running on Pasteur from band LOD2
;;; With SYSTEM 6.31, VIRTUAL-MEMORY 6.3, EH 6.6, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.4, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.10, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.1,
;;;  COMPILER 6.14, TV 6.24, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.13,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.42, CLEH 6.5, IP 3.60,
;;;  Experimental CLX 6.9, CLUE 6.86, X11M 6.21, Experimental BUG 11.18,  microcode 648,
;;;  Band Name: rel6.0 1/23

#!C
; From file QFILE-SERVER.LISP#> CHAOSNET; sys:
#10R FILE-SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "FILE-SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CHAOSNET; QFILE-SERVER.#"


(defun rfile-server ()
  
  (if (and server-login (or (null user-id)
			    (zerop (length user-id))))
      (trap-lossage (error "Server Login")
		    (progn (login server-login-id si:local-host)
			   (print-server-login-exegesis))))
  (let ((*print-case* :upcase)
	(*print-length* nil)
	(*print-level* nil)
	(*package* *system-package*)   ;11-16-87 DAB
	(time:*default-date-print-mode* :mm/dd/yy)    
	tid conn-stream
	conn
	alldatas
	server-openings
	(server-instance (gensym))
	(user-id server-login-id)
	(server-protocol-version server-protocol-version)
	(*local-server-via-net* nil)
	(*mac-nfs-copy-fhandle* nil)  ; DAB 11-01-89 Special NFS-MX variables
        (*mac-nfs-copy-creation-date* nil) ; DAB 11-01-89
	)
    (declare (special *mac-nfs-copy-fhandle*  *mac-nfs-copy-creation-date* ))   ; DAB 11-01-89
    (unwind-protect
	(trap-lossage
	  (error "Server Top Level")
	  (catch
	    'server-chaos-disappear
	    (setq conn (chaos:listen "FILE"))
	    (when *lmfs-server-dont-answer-logins*
	      (chaos:reject (prog1 conn
				   (setq conn nil))
			    *lmfs-server-dont-answer-logins*)
	      (ferror nil *lmfs-server-dont-answer-logins*))
	    (let* ((pkt (chaos:read-pkts conn))
		   (result (server-parse-rfc pkt)))
	      (cond ((integerp result)
		     (setq server-protocol-version result))
		    (t
		     (chaos:reject (prog1 conn
					  (setq conn nil))
				   result)
		     (ferror nil result))))
	    (chaos:accept conn)
	    (funcall tv:who-line-file-state-sheet
		     :add-server
		     conn
		     "FILE"
		     current-process
		     'lmfs-peek-server
		     (process-stack-group current-process))
	    (setq conn-stream (chaos:make-stream conn))
	    (if *server-shutdown-message* (send-single-shutdown-message conn))
	    (error-restart-loop
	      ((system:abort error) "Return to server command-reading loop.")
	      (let
		(pkt op)
		(setq
		  pkt
		  (trap-lossage
		    (error "Server Reading packets")
		    (condition-bind
		      ((system:host-stopped-responding
			 (function (lambda (&rest ignore)
				     (throw 'server-chaos-disappear nil))))
		       (system:connection-lost (function (lambda (&rest ignore)
							   (throw 'server-chaos-disappear
								  nil)))))
		      (with-timeout ((* 24 60 60 60)  ; DAB 10-23-89 Timeout, then disappear. 48 HRs
				      (throw 'server-chaos-disappear nil))
			(chaos:get-next-pkt conn))
		      )
		    (ferror 'server-control-conn-network-lossage "Control connection lost")))
		(setq op (chaos:pkt-opcode pkt))
		(cond ((or (= op chaos:eof-op)
			   (= op chaos:cls-op))
		       (funcall conn-stream :force-output)
		       (chaos:return-pkt pkt)
		       (throw 'server-chaos-disappear nil))
		      ((not (= op chaos:dat-op))
		       (ferror nil "Unrecognized packet opcode: ~S" op)))
		(let*
		  ((string (chaos:pkt-string pkt))
		   (strings (get-strings-from-pktstring string)))
		  (if trace-server-enabled
		      (without-interrupts (push (string-append string) server-traces)))
		  (destructuring-bind
		    (tid fh cmd . rest)
		    (parse-cmd-string string)
		    (if *lmfs-server-dont-answer-logins*
			(format conn-stream
				"~A ~A ERROR HNA F Host not available - ~A "
				tid
				(or fh "")
				*lmfs-server-dont-answer-logins*)
			(case cmd
			  (:login  (setq user-id (file-server-login rest)))
			  (:open  (file-server-open fh rest (first (foreign-pname-string conn strings))))
			  (:open-for-lispm  (apply
					      'file-server-open-for-lispm
					      fh
					      (first strings)
					      (let
						((*print-base* 10)
						 (*read-base* 10)
						 (*package* si:pkg-user-package)
						 (*readtable* si:initial-readtable))
						(read-from-string
						  string
						  nil
						  nil
						  :start
						  (+ (position #\Newline
							       (the string (string string))
							       :test
							       (function char=))
						     (length (first strings))
						     2)))))
			  (:direct-output (apply 'file-server-direct-output   ;new commands to support OPEN :IO
						 fh                           ; 09-02-87 DAB IO Support
						 rest
						 ))	;direct write
			  (:read  (apply 'file-server-direct-read              ;new commands to support OPEN :IO
					 fh                                    ; 09-02-87 DAB IO Support
					 rest
					 ))	;direct read
			  (:extended-command  (apply
						'file-server-extended-command
						fh
						(first rest)
						(first strings)
						(let
						  ((*print-base* 10)
						   (*read-base* 10)
						   (*package* si:pkg-user-package)
						   (*readtable* si:initial-readtable))
						  (read-from-string
						    string
						    nil
						    nil
						    :start
						    (+ (position #\Newline
								 (the string (string string))
								 :test
								 (function char=))
						       (length (first strings))
						       2)))))
			  (:data-connection  (file-server-data-connection fh rest))
			  (:undata-connection  (file-server-undata-connection fh))	
			  (:close  (file-server-close-connection fh rest))
			  (:filepos  (file-server-filepos fh rest))
			  (:delete  (file-server-delete fh (foreign-pname-string conn strings)))
			  (:rename  (file-server-rename fh (list
							     (first (foreign-pname-string conn (list (first strings))))
							     (first (foreign-pname-string conn (rest strings)))
							     )))
			  (:expunge  (file-server-expunge fh (foreign-pname-string conn strings)))
			  (:complete  (file-server-complete fh rest (foreign-pname-string conn strings)))
			  (:continue  (file-server-continue fh))
			  (:directory  (file-server-directory fh rest (foreign-pname-string conn strings)))
			  (:change-properties  (file-server-change-props fh (foreign-pname-string conn strings)))
			  (:create-directory  (file-server-create-directory fh
									    (foreign-pname-string conn strings)))
			  (:create-link  (file-server-create-link fh (foreign-pname-string conn strings)))
			  (:delete-and-expunge  (file-server-delete-and-expunge fh
										(foreign-pname-string conn strings)))
			  (otherwise  (format conn-stream
					      "~A ~A ERROR UKC F Unknown command: ~A"
					      tid
					      (or fh "")
					      cmd)))))
		  (funcall conn-stream :force-output)
		  (chaos:return-pkt pkt))))))
      (cond (conn
	     (funcall tv:who-line-file-state-sheet :delete-server conn)
	     (trap-lossage (error "Server Top Level close")
			   (chaos::close-conn conn
					      (or *lmfs-server-dont-answer-logins*
						  "Server error")))
	     (chaos:remove-conn conn)))
      (if server-openings
	  (trap-lossage (error "Server finish closing remaining openings")
			(dolist (opening  server-openings)
			  (funcall opening :close :abort))))
      (trap-lossage
	(error "Closeout undata")
	(dolist
	  (data alldatas)
	  (rplaca
	    (server-dataproc-comm-cell (get (server-dataproc-comm-sibling data)
					    server-instance))
	    'undata)
	  (rplaca (server-dataproc-comm-cell data) 'undata)
	  (chaos:remove-conn (server-dataproc-comm-conn data)))))))
))
