;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10; Patch-file:T -*-

;;; Reason: Fix bug in class precedence list of CLOS slot descriptions which was 
;;; preventing them from being printed.  [SPR 9736]
;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.

;;; Written 05/23/89 13:19:00 by GRAY,
;;; while running on Kelvin from band LODA
;;; With SYSTEM 6.2, VIRTUAL-MEMORY 6.0, EH 6.0, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.0, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.0, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.1, TV 6.5, DATALINK 6.0, CHAOSNET 6.0, GC 6.0, MEMORY-AUX 6.0, NVRAM 6.0,
;;;  SYSLOG 6.0, STREAMER-TAPE 6.0, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.0,
;;;  DEBUG-TOOLS 6.0, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.0, SERIAL 6.0, PRINTER 6.0, MAC-PRINTER-TYPES 6.0, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.0, TI-CLOS 6.0, CLEH 6.0, IP 3.45,
;;;  Experimental BUG 11.5, CLX 6.0, CLUE 6.1, X11M 6.0, Experimental DOCUMENTER 619.0,
;;;   microcode 429, Band Name: Release 6.0 + SLE 5/22

;;; SPR Number 9736
;;; 	
;;; Problem
;;;          In the 6.0 bands of 5/12 and 5/22,
;;;          (SUBTYPEP (FIND-CLASS 'SLOT-DEFINITION) (FIND-CLASS 'T))
;;;          returns NIL, although it is obvious that it should be
;;;          T.  The problem is that the class-precedence list for SLOT-DESCRIPTION
;;;          was patched to include class METAOBJECT, but somehow the classes
;;;          STRUCTURE-OBJECT and T were lost.  This may not appear too serious by
;;;          itself, but it has the consequence that slot objects cannot be printed
;;;          -- get "no applicable method" error in PRINT-OBJECT.
;;; 
;;; Solution
;;; 	Fix file "SYS:CLOS;STRUCTURE-METACLASS.LISP" to compose the
;;; 	class before trying to insert METAOBJECT into the precedence list.

#!C
; From file STRUCTURE-METACLASS.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; STRUCTURE-METACLASS.#"

;;  5/7/89 DNG - Patch up slot description class to look like a subclass of METAOBJECT.
;; 5/23/89 DNG - Fix to call COMPOSE-CLASS before looking at the precedence list.
(let* ((class (find-class 'slot-description))
       (cd (class-description class))
       (mo (find-class 'metaobject))
       (cpl (progn (compose-class class)
		   (class-description-class-precedence-list cd))))
  (unless (member mo cpl :test #'eq)
    (setf (class-description-class-precedence-list cd)
	  (list* class mo (rest cpl)))
    (add-direct-subclass mo class)))
))

;; The following is not needed in the source; this is just in case someone 
;; tried to print a slot description before loading the patch.  This will fix 
;; up the method hash table so it doesn't dispatch to the "no applicable method"
;; handler.
(prepare-generic-function #'print-object (make-instance 'slot-definition))