;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10; Patch-file:T -*-

;;; Reason: Force re-generation of combined methods when redefining a method to use 
;;; CALL-NEXT-METHOD when it was not used by the previous definition.  [SPR 9630]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for TI-CLOS version 6.10
;;; Written 06/23/89 17:26:30 by GRAY,
;;; while running on Kelvin from band LOD2
;;; With Inconsistent SYSTEM 6.9, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0,
;;;  LOCAL-FILE 6.0, BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1,
;;;  NETWORK-NAMESPACE 6.0, DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0,
;;;  NETWORK-PATHNAME 6.0, Inconsistent COMPILER 6.6, TV 6.11, DATALINK 6.0, CHAOSNET 6.0,
;;;  GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1, SYSLOG 6.1, STREAMER-TAPE 6.0, UCL 6.0, INPUT-EDITOR 6.0,
;;;  METER 6.0, ZWEI 6.3, Inconsistent DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0,
;;;  DATALINK-DISPLAYS 6.0, FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.1, MAC-PRINTER-TYPES 6.1,
;;;  PRINTER-TYPES 6.0, IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0,
;;;  TELNET 6.0, VT100 6.0, NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, Inconsistent TI-CLOS 6.9,
;;;  CLEH 6.4, IP 3.46, Experimental BUG 11.10, Experimental CLX 6.1, CLUE 6.5, X11M 6.1,
;;;  Experimental DOCUMENTER 619.0, Experimental GENASYS 3.0,  microcode 429, Band Name: 6.0 SLE 6/5 + u429 6/8

;;; BUG REPORT NUMBER:  9630
;;;
;;; PROBLEM:  CLOS error "no next method" in a user-defined method
;;;	on PRINT-OBJECT that uses CALL-NEXT-METHOD.  The error only seems
;;;	to happen if the method was previously defined and used without
;;;	using CALL-NEXT-METHOD.
;;;
;;; SOLUTION:  When adding a method, if we decide that the combined method 
;;;	needs to be re-generated (in this case because of the change in the 
;;;	next-methods-not-used status), call RECOMBINE-METHODS with an additional 
;;;	argument which tells it and the functions it calls to not re-use any 
;;;	existing combine methods.

#!C
; From file GENERIC.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; GENERIC.#"


;;  6/23/89 DNG - Pass new dont-reuse-old-combined-methods-p flag to 
;;		recombine-methods to fix SPR 9630.
(defmethod (standard-generic-function :add-method) ( method)
  (let ((gfunc self))
	 ;;We need to check the lambda-list congruence with the generic-function.
    (unless (= (length (method-parameter-specializers method)) (length argument-precedence-order))
      (error "The number of specializers of method ~s does not match the 
number expected by the generic function ~s." method gfunc))
    (unless (combined-method-p  method)
      (check-lambda-list-for-method method  gfunc))
    (let ((keys (reorder-parameter-specializers gfunc (method-parameter-specializers method)))
	  (existing-method (*find-method gfunc
				       (method-qualifiers method )
				       (method-parameter-specializers method) nil))
	  (method-name (function-name (*method-function method ))))
      ;;at this point, we should go from the class-name to the real class object.
      ;;[first class things]
      (setf method-list (insert-in-method-list keys method-list method (method-qualifiers method)
					       method-name))
      ;;Side effect the method so it points to its generic function.
      (setf (symeval-in-instance method 'generic-function) gfunc)
      (unless (or (not (eq (car-safe method-name) 'method))
		  (eq gfunc (second method-name))
		  (and (variable-boundp discriminator-code)
		       (or (eq discriminator-code (second method-name))
			   (eq discriminator-code
			       (sys:fdefinition-safe (second method-name) t)))))
	(setf (second method-name) gfunc))
       (when (generic-function-method-hash-table gfunc)
	 (unless (handlerp method)
	   (when (and existing-method
		      (incompatible-arg-desc-p
			(method-arglist-description method)
			(method-arglist-description existing-method)))
	     ;; Here when replacing a method that was previously defined with the 
	     ;; wrong number of optional arguments.  Need to make sure that the 
	     ;; combined method is re-generated.
	     (remove-method gfunc existing-method)
	     (setq existing-method nil))
	   (if (and existing-method (same-arglist-signature-p existing-method method)
			(not (and (getf (method-plist existing-method ) 'next-methods-not-used )
				  (null (getf (method-plist method) 'next-methods-not-used )))))
	       (let ((forward (getf (method-plist existing-method) 'forwarded-from)))
		  (when (and forward
			      (eq existing-method (method-forwarded-method forward)))
		    ;; Update combined method that pointed to the method being replaced.
		    (setf (method-forwarded-method forward) method)
		    (setf (getf (method-plist method) 'forwarded-from) forward)))
	     ;;Don't recombine if is a redefinition of the same method.
	     ;;This can be refined if we do inlining of methods inside combined methods.
	     ;;With the new congruence rule, if the new method does not have the same arglist
	     ;; the old one has, then we must redo the combined method.
	     (recombine-methods gfunc keys nil t))
	   ))
       (unless (or existing-method (combined-effective-method-p method))
	 (dolist (k keys )
	   (unless (consp k) 
	     (add-method-on-specializer k  method))))
      gfunc)))
))

#!C
; From file METHOD-COMBINATION.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; METHOD-COMBINATION.#"


;;  6/23/89 DNG - Add dont-reuse-old-combined-methods-p argument as part of fix for SPR 9630.
(defun recombine-methods (generic-function arg-classes &optional initial-points
			  dont-reuse-old-combined-methods-p)
  ;;Starting at arg-classes, refigures out the handler for subclasses
  ;;Combine-method is the expensive call.  However, combining method for
  ;;only cached classes will tend to make combined methods too specific.
  ;;Arg-classes is supposed to be ordered by argument-precedence-order.
  ;;Take the method-alist, 
  ;;then combine each node of the tree.
  (let* ((ht (generic-function-method-hash-table generic-function))
	 (points (mapcar #'list arg-classes))
;;	 (n-args (length arg-classes))
	 )
    (declare (special *t-class*))
    (labels ((cm1 (tree points args-c)
		  (if (null points)
		      nil 
		      (dolist (ael tree)
			(when (subclassp (car ael) (car args-c))
			  (pushnew (car ael) (car points) :test #'class-eql))
			(cm1 (cdr ael)  (cdr points) (cdr args-c)))))
	     (ncm (rargs tree points seen)
		(if (null points)
		    (ignore-method-combination-error
		      (combine-method generic-function (reverse rargs)
				      dont-reuse-old-combined-methods-p))
		    (dolist (c (car points))
		      (unless (member c (car seen) :test #'class-eql)
			(push c  (car seen))
			(let ((subtree (check-tree c tree points)))
			  (and subtree
			       (ncm (cons c rargs) subtree (cdr points) (list nil))))
			(unless (or (individual-typep c) (eql *t-class* c) (= (length tree) 1))
			  (with-stack-list* (new-points (class-direct-subclasses c) (cdr points))
			    (ncm rargs tree new-points seen))))
		      )))
	     
	     )
      ;;First combine the method for arg-classes, recombine-method is called
      ;;when a new method is added.
      ;;This will prevent a more specific set of arguments that would inherit this
      ;;combined method to make a combined method of their own, just because they were
      ;;in the hashtable
      (unless initial-points
	(ignore-method-combination-error (combine-method generic-function arg-classes)))
      (let* ((ml (generic-function-method-list  generic-function))
	     (tree ml ))
	(when tree
	  (unless initial-points (cm1 tree  (mapcar #'nreverse points) arg-classes))
	  (ncm nil tree (or initial-points points) (list nil))
	  (when (and ht (null initial-points))
	    (fix-cached-methods generic-function  arg-classes))
	  )))))
))

#!C
; From file METHOD-COMBINATION.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; METHOD-COMBINATION.#"


;;  6/23/89 DNG - Add dont-reuse-old-combined-methods-p argument as part of fix for SPR 9630.
(defun combine-method  (gen-function arg-types &optional dont-reuse-old-combined-methods-p)
  ;;Arg-types is supposed to be ordered as arguments-precedence-order
  (let ((default-cons-area sys:background-cons-area))
    (multiple-value-bind (handler code derivation methods)
	(find-handler gen-function  arg-types #'get-sorted-alists dont-reuse-old-combined-methods-p)
      (when (and handler (generic-function-method-hash-table gen-function))
	(put-method-in-hash-table  gen-function  handler arg-types))
      (or handler
	  (if methods
	      (make-combined-method gen-function
				    (order-parameter-specializers gen-function arg-types)
				    code derivation methods)
	    (let ((m (*find-method gen-function '(:combined) (REORDER-PARAMETER-SPECIALIZERS gen-function arg-types) nil)))
	      (when m (remove-method gen-function m))
	      (when (generic-function-method-hash-table gen-function)
		(remove-method-in-hash-table gen-function arg-types))
	      nil))))))
))

#!C
; From file METHOD-COMBINATION.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; METHOD-COMBINATION.#"


;;  6/23/89 DNG - Add dont-reuse-old-combined-methods-p argument as part of fix for SPR 9630.
(defun find-handler (gen-function arg-types &optional (sort-function #'get-sorted-alists)
		     dont-reuse-old-combined-methods-p)
  ;;Arg-types is supposed to be ordered as arguments-precedence-order
  ;;Returns one to four values:
  ;;First value is the handler, if found;
  ;;Second value is the code of the combined method (if the handler is a combined method)
  ;;Third value is the combined method derivation.
  ;;Fourth value is the applicable methods, so make-combined-method can compute its arglist.
  (declare (values handler code derivation applicable-methods))
  (let* ((generic-function gen-function)
	 (ml (generic-function-method-list  generic-function))
	 (combination (generic-function-method-combination generic-function))
	 (derivation))
    (declare (special generic-function derivation))
    (multiple-value-bind (methods combined-methods)
	(compute-method-precedence-list ml arg-types sort-function)
      (if methods
	  (multiple-value-bind (ignore keylist)
	      (compute-combined-lambda-list gen-function methods)
	    (when (trivial-method-combination-p (car combination))
	      (setf derivation (list  combination
				      (mapcar #'(lambda (m) (function-name (*method-function m))) methods)
				      keylist))
	      (unless dont-reuse-old-combined-methods-p
		 (let* ((handler (have-combined-method combined-methods derivation)))
		   (when handler (return-from find-handler handler nil derivation)))))
	    (let ((code `(macrolet ((call-method (method next-method-list)
						 ,(call-next-method-code gen-function)))
			   ,(compute-effective-method gen-function combination methods))))
	      (if (trivial-method-combination-p (car combination))
		  (values nil code derivation methods)
		(values  (have-combined-method combined-methods derivation)
			 code
			 (list combination (copy-tree code) keylist)
			 methods))))))))
))
