;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10; Patch-file:T -*-

;;; Reason: Fix defining of hybrid classes that don't include any flavors.  [SPR 10193]


;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for TI-CLOS version 6.16
;;; Written 07/06/89 20:15:53 by GRAY,
;;; while running on Kelvin from band LOD2
;;; With SYSTEM 6.10, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  Inconsistent COMPILER 6.7, TV 6.12, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0,
;;;  NVRAM 6.1, SYSLOG 6.1, STREAMER-TAPE 6.0, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1,
;;;  ZWEI 6.3, DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.2, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, Inconsistent TI-CLOS 6.15, CLEH 6.4,
;;;  IP 3.47, Experimental BUG 11.10, Experimental CLX 6.2, CLUE 6.8, X11M 6.1, Experimental DOCUMENTER 619.0,
;;;   microcode 429, Band Name: 6.0 SLE 6/5 + u429 6/8

;;; BUG REPORT NUMBER:  10193
;;;
;;; PROBLEM:  Got error "instance variable TICLOS::FLAVOR is unbound" when 
;;;	trying to use a hybrid class that doesn't actually include any flavors.  
;;;
;;; SOLUTION:  In (METHOD ADD-NAMED-CLASS (HYBRID-CLASS)), remove the test on
;;;	(UNLESS (NULL SUPER-FLAVORS) ...) so that an associated flavor is always 
;;;	created.

#!C
; From file CLASS-REDEFINITION.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; CLASS-REDEFINITION.#"



;;  7/06/89 DNG - Create an associated flavor even if SUPER-FLAVORS is null so 
;;		that later methods won't trap on the FLAVOR slot being unbound. [SPR 10193]
(defmethod clos:add-named-class ((class-prototype hybrid-class) &key name 
				 superclass-names environment &allow-other-keys)
  (let* ((class-name name)
	 (super-flavors nil)
	 (class-object (call-next-method)))
    (declare (list super-flavors) (type hybrid-class class-object))
    (dolist (super-name superclass-names)
      (when (sys:get-flavor super-name environment)
	(push super-name super-flavors)))
    (let* ((flavor-name class-name)
	   (compiler:*compile-file-environment* environment)
	   (sys::*just-compiling* (and 
				    (compiler:environment-remote-p environment))))
      (sys::defflavor1 flavor-name
		       (collect-body
			 (dolist (sd (class-direct-slots class-object))
			   (declare (type slot-description sd))
			   (when (eq (slot-allocation sd) :instance)
			     (collect
			       (if (non-empty-p (slot-initform sd))
				   (list (slot-name sd) (slot-definition-initform sd))
				 (slot-name sd))))))
	(nreverse super-flavors)
	(if (typep class-prototype 'hybrid-class-with-flavor-instances)
	    '()
	  '((:instantiation-flavor-function dont-make-hybrid-flavor)) ))
      (let ((fl (sys:get-flavor flavor-name environment)))
	(setf (slot-value class-object 'flavor) fl)
	(setf (class-description-class-object fl) class-object)
	;; shouldn't need this now that defflavor1 knows about this
	;;(setf (class-named class-name t environment) class-object)
	))
    class-object))

))
