;;; -*- Mode:Common-Lisp; Package:System; Base:10; Patch-file:T -*-

;;; Reason: Fix CLOS:DESCRIBE to be a generic function on microExplorer.  [SPR 10258]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for TI-CLOS version 6.18
;;; Written 07/14/89 11:22:30 by GRAY,
;;; while running on Kelvin from band LOD2
;;; With Inconsistent SYSTEM 6.12, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0,
;;;  LOCAL-FILE 6.0, BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1,
;;;  NETWORK-NAMESPACE 6.0, DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0,
;;;  NETWORK-PATHNAME 6.0, COMPILER 6.9, TV 6.12, DATALINK 6.0, CHAOSNET 6.0, GC 6.3,
;;;  MEMORY-AUX 6.0, NVRAM 6.1, SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0,
;;;  METER 6.1, ZWEI 6.4, DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1,
;;;  DATALINK-DISPLAYS 6.0, FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1,
;;;  PRINTER-TYPES 6.1, IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2,
;;;  TELNET 6.0, VT100 6.0, NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, Inconsistent TI-CLOS 6.17,
;;;  CLEH 6.4, IP 3.47, Experimental BUG 11.10, Experimental CLX 6.2, CLUE 6.9, X11M 6.1,
;;;  Experimental DOCUMENTER 619.0,  microcode 429, Band Name: 6.0 SLE 6/5 + u429 6/8

;;; BUG REPORT NUMBER:  10258
;;;
;;; PROBLEM:  On the microExplorer, the symbols LISP:DESCRIBE and CL:DESCRIBE 
;;;	are the same and have the function definition for CL:DESCRIBE.  On the
;;;	Explorer these are correctly two separate symbols.
;;;
;;; SOLUTION:  No source change is needed, just need to re-execute the 
;;;	installation of these two functions.

#!C
; From file DESCRIBE.LISP#> CLOS; MR-X:
#10R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; DESCRIBE.#"



;;;03/20/89 clm - Integrated CLOS DESCRIBE into the Kernel
;;; 05/08/89 DNG - Updated DESCRIBE to use the new generic function DESCRIBE-OBJECT.

(eval-when (eval compile load)
  (import '(ticlos:describe-object))
  (when (and (not (eq (find-package "LISP") (find-package "CL")))
	     (eq (find-symbol "DESCRIBE" "LISP") (find-symbol "DESCRIBE" "CL")))
    (unintern (find-symbol "DESCRIBE" "CL") "CL")
    (export (intern "DESCRIBE" "CL") "CL")))
))

(unless (ticlos:generic-function-p 'lisp:describe)
  #!C
  ; From file DESCRIBE.LISP#> CLOS; MR-X:
  #10R SYSTEM#:
  (COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
		 (SI:*LISP-MODE* :COMMON-LISP)
		 (*READTABLE* COMMON-LISP-READTABLE)
		 (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
    (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; DESCRIBE.#"
				 
				 
 ;;; 03/20/89 clm - for CLOS, remove old DESCRIBE.
 (eval-when (eval compile load)
   (fmakunbound 'describe))
 ))
  
  
  #!C
  ; From file DESCRIBE.LISP#> CLOS; MR-X:
  #10R SYSTEM#:
  (COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
		 (SI:*LISP-MODE* :COMMON-LISP)
		 (*READTABLE* COMMON-LISP-READTABLE)
		 (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
    (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; DESCRIBE.#"
				 
				 
	 (ticlos:defgeneric describe (object)	; DESCRIBE according to 88-002R
	   (:documentation "Describe the value or components of any Lisp object.
This is a good way to find out more than the printed representation says."))
				 
	 (defmethod describe ((object t))
	   (cl:describe object *standard-output*))
	 ))
  
  ) ; end unless generic

#!C
; From file DESCRIBE.LISP#> CLOS; MR-X:
#10R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; DESCRIBE.#"


(defun cl:describe (object &optional output-stream) ; DESCRIBE as adopted by X3J13 March 1989
  "Describe the value or components of any Lisp object.
This is a good way to find out more than the printed representation says.
The behavior can be customized by adding methods to generic function DESCRIBE-OBJECT."
  (let* ((stream (decode-print-arg output-stream))
	 (stream2 stream))
    (when (> *describe-depth* 0)
      (when (or (null object)			;Don't recursively describe boring things
		(numberp object) (symbolp object) (stringp object) (consp object)
		(and *print-level* (> *describe-depth* *print-level*)))
	(return-from cl:describe (values)))
      (fresh-line stream)
      (setq stream2				;Arrange for indentation by 5 spaces
	    #'(lambda (&rest args)
		(when (eql (send stream :send-if-handles :read-cursorpos) 0)
		  (write-string "     " stream))
		(lexpr-send stream args))))
    (let ((*describe-depth* (1+ *describe-depth*)))
      (describe-object object stream2)))
  (values))
))


;; re-install this so it calls the right DESCRIBE
#!C
; From file DESCRIBE.LISP#> CLOS; MR-X:
#10R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; DESCRIBE.#"



;; 12/05/88 DNG
(defmethod describe-object :after ((dobj ticlos:class) stream)
  (ticlos:with-slots (ticlos:class-description) dobj
    (when (and (ticlos:slot-boundp dobj 'ticlos:class-description)
	       (typep ticlos:class-description 'ticlos:class-description))
      (cl:describe ticlos:class-description stream))))

))
