;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10; Patch-file:T -*-

;;; Reason: Fix a bug in which SHARED-INITIALIZE sometimes stored the default values for 
;;; slots of hybrid classes in the wrong slot.  [SPR 9714]

;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.

;;; Patch file for TI-CLOS version 6.2
;;; Written 05/23/89 15:32:29 by GRAY,
;;; while running on Kelvin from band LODA
;;; With SYSTEM 6.2, VIRTUAL-MEMORY 6.0, EH 6.0, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.0, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.0, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.1, TV 6.5, DATALINK 6.0, CHAOSNET 6.0, GC 6.0, MEMORY-AUX 6.0, NVRAM 6.0,
;;;  SYSLOG 6.0, STREAMER-TAPE 6.0, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.0,
;;;  DEBUG-TOOLS 6.0, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.0, SERIAL 6.0, PRINTER 6.0, MAC-PRINTER-TYPES 6.0, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.0, Inconsistent TI-CLOS 6.1, CLEH 6.0,
;;;  IP 3.45, Experimental BUG 11.5, CLX 6.0, CLUE 6.1, X11M 6.0, Experimental DOCUMENTER 619.0,
;;;   microcode 429, Band Name: Release 6.0 + SLE 5/22

;;; SPR Number 	9714
;;; 	
;;; Problem 
;;; 	CLOS hybrid classes that use 
;;; 	(METHOD SHARED-INITIALIZE (STANDARD-OBJECT) T) rather than their
;;; 	own generated method [e.g.  if there are too many initargs] do
;;; 	not get initialized correctly.  The problem is that the
;;; 	SHARED-INITIALIZE method assumes that the slots are allocated in
;;; 	the same order that they appear in the CLASS-ALL-SLOTS list,
;;; 	which was not necessarily true for slots inherited from a
;;; 	flavor.
;;; 
;;; 	One consequence of this is an error when trying to make an
;;; 	instance of CLUE:INTERACTIVE-STREAM.
;;; 
;;; Solution
;;; 	Update (:METHOD HYBRID-CLASS :COMPUTE-INSTANCE-SIZE) to put the
;;; 	all-slots list in the same order as the instance variable names.


#!C
; From file CLASS-DEFS.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; CLASS-DEFS.#"

(defmethod (hybrid-class :compute-instance-size) ()
  ;; 12/02/88 DNG - Original.
  ;;  1/05/89 DNG - Fix to permit composing flavor defined in compile-time environment only.
  ;;  4/29/89 DNG - Add binding of *JUST-COMPILING*.
  ;;  5/08/89 DNG - Ensure that the instance slot names list is cdr-coded.
  ;;  5/23/89 DNG - Put the all-slots list in the same order as the instance 
  ;;		variable names because (METHOD SHARED-INITIALIZE (STANDARD-OBJECT) T) 
  ;;		assumes that they will be. [SPR 9714]

  ;; Let Flavors figure out the order of the slots so that any ordered 
  ;; instance variables will be handled correctly, but need to add any slots 
  ;; from non-flavor super-classes.
  (when (null (sys:flavor-depends-on-all flavor))
    (let ((si:*just-compiling* (si:just-compiling)) ; needed by SYS:MAP-OVER-COMPONENT-FLAVORS
	  (loc (locf (get (sys:flavor-name flavor) 'sys:flavor))))
      (if (eq flavor (contents loc))
	  (sys:compose-flavor-combination flavor)
	(without-interrupts
	  ;; Temporarily install the flavor definition globally so that it can be 
	  ;; composed.  (Much of the flavor system code assumes that
	  ;; compile-time flavors will never be instantiated.)
	  (sys:%bind loc flavor)
	  (sys:compose-flavor-combination flavor)))))
  (let ((ivars (sys:flavor-all-instance-variables flavor))
	(slots (class-description-instance-slot-names class-description)))
    (if (equal ivars slots)
	(sys:flavor-instance-size flavor)
      (progn (dolist (x slots)
	       (unless (member x ivars :test #'eq)
		 ;; must have been inherited from a non-flavor super-class.
		 (push-end x ivars)))
	     (setf ivars (copy-list ivars)) ; the microcode requires that this be cdr-coded.
	     (setf (sys:flavor-all-instance-variables flavor) ivars)
	     (setf (class-description-instance-slot-names class-description) ivars)
	     ;; Need to re-order the all-slots list because
	     ;; (METHOD SHARED-INITIALIZE (STANDARD-OBJECT T)) assumes the slots are 
	     ;; allocated in that order.
	     (let ((new (copy-list ivars))
		    (all-slots (class-all-slots self)))
	        (declare (list new all-slots))
	        (do ((tail new (cdr tail)))
		     ((null tail))
		   (setf (car tail)
			  (or (find (car tail) all-slots :key #'slot-description-name :test #'eq)
			       (error "Internal CLOS error - lost slot ~S." (car tail)))))
		(setf (class-all-slots self) new))
	     (setf (sys:flavor-instance-size flavor) (1+ (length ivars)))))))

))
