;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10; Patch-file:T -*-

;;; Reason: Fix bug in DEFCLASS slot reader methods when there are :BEFORE or :AFTER 
;;; methods attached.  [SPR 10582]

;;;			      RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED	  
;;;   P.O. BOX 149149, M/S 2151		    
;;;   AUSTIN, TEXAS 78714		  
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for TI-CLOS version 6.22
;;; Written 09/21/89 15:37:31 by GRAY,
;;; while running on Kelvin from band LOD2
;;; With SYSTEM 6.15, VIRTUAL-MEMORY 6.1, EH 6.4, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.2, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.10, TV 6.15, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.5,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.4, Inconsistent TI-CLOS 6.21, CLEH 6.5,
;;;  IP 3.48, Experimental CLX 6.2, CLUE 6.10, X11M 6.13, Experimental BUG 11.15,
;;;  Experimental DOCUMENTER 701.0,  microcode 430, Band Name: 6.0+Scribe,&c,u430 9/6

;;; BUG REPORT NUMBER:  10582
;;;
;;; PROBLEM:
;;;    I can't seem to get a before method to work.  It claims I am calling
;;;    with too many args.  This method was a before method to a slot accessor
;;;    whose primary method was system defined.
;;;    Backtrace follows:
;;;    >>Trap #o15104 (TOO-MANY-ARGS)
;;;	  Function (METHOD SLOT1 (QWERTY)) called with too many arguments (1).
;;;	  Backtrace from the debugger:
;;;	  Values to be collected for MULTIPLE-VALUE-LIST.
;;;	  Lexpr-funcalled.
;;;	     
;;;	  (METHOD SLOT1 (QWERTY)) (P.C. = 0)
;;;	     
;;;	     --Extraneous args:--
;;;	   Arg 0: #<QWERTY 42530417>
;;;	     
;;;	  Disassembled code:
;;;		6 (AUX) RETURN-NIL 
;;;	...
;;;
;;; SOLUTION:  Fix a typographical error in function MAKE-READER -- the lambda 
;;;	expression for a slot reader was being constructed as
;;;	 `(SYS:NAMED-LAMBDA
;;;	    (,METHOD-SPEC (INSTANCE) (SLOT-VALUE INSTANCE ',SLOT-NAME)))
;;;	which takes no arguments and always returns NIL.  It should be:
;;;	 `(SYS:NAMED-LAMBDA (,METHOD-SPEC) (INSTANCE)
;;;	    (SLOT-VALUE INSTANCE ',SLOT-NAME))
;;;	Usually it doesn't matter since this code is not used anyway 
;;;	[ (:METHOD STANDARD-READER-METHOD :GET-HASH-ENTRIES-FROM-HANDLER) causes 
;;;	#'READ-INSTANCE-SLOT to be placed in the method hash table in its place];
;;;	the error only appeared when the method needed to be referenced by a 
;;;	combined method.
;;;
;;; DEPENDENCIES:  [none]
;;;
;;; CODEREAD:

#!C
; From file CLASS.LISP#> SYS6.CLOS; Kelvin:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; CLASS.#"


;;; DNG 9/21/89 Fixed mismatched parentheses in the lambda expression. [SPR 10582]
(DEFUN make-reader (name slot-name class-name)
  (ensure-generic-fn name '(instance))
  (LET ((method-spec `(method ,name (,class-name))))
    (WHEN (SYS:RECORD-SOURCE-FILE-NAME method-spec)
      (add-method (FDEFINITION name)
		  (make-instance
		    'standard-reader-method
		    :slot-name slot-name
		    :qualifiers nil
		    :parameter-specializers
		      (list (class-named class-name))
		    :function
		      (compile nil
			       `(sys:named-lambda
				  (,method-spec) (instance)
				  (slot-value instance ',slot-name)))
		    :derivation
		      `((standard) (,method-spec) nil))))))
))
