;;; -*- Mode: Common-Lisp; Package: TICLOS; Base: 10.; Patch-File: T -*-

;;; Reason: Modified modify-generic-function to return a generic function object. [10842]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 01/16/90 07:56:35 by BERGER,
;;; while running on ARIES from band LODA
;;; With SYSTEM 6.28, VIRTUAL-MEMORY 6.3, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.7, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.22, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.12,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.4, MAIL-READER 6.6, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.28, CLEH 6.5, IP 3.57,
;;;  Experimental CLX 6.8, CLUE 6.46, X11M 6.19, Experimental BUG 11.17,  microcode 9,
;;;  Band Name: REl 6.0 SLE 12/28

#!C
; From file GENERIC.LISP#> CLOS; SYS:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; GENERIC.#"


(defun modify-generic-function (old-fef &rest options
				&key (generic-function-class nil generic-function-class-p)
				(lambda-list nil llp) name environment &allow-other-keys)
  (let* ((ins (get-generic-function-object old-fef))
	 gfun-class)
    (when (and generic-function-class-p
	       (not (class-eql (setq gfun-class (class-named generic-function-class nil environment))
			       (class-of ins))))
      (let ((new-ins (make-instance gfun-class
				    :method-list (generic-function-method-list ins)
				    :name (generic-function-name ins)
				    :method-class (generic-function-method-class ins)
				    :lambda-list (generic-function-lambda-list  ins)
				    :argument-precedence-order (generic-function-argument-precedence-order  ins)
				    :declare (generic-function-declare  ins)
				    :documentation (generic-function-documentation  ins)
				    :method-combination (generic-function-method-combination  ins)
				    :arglist-description (generic-function-arglist-description ins)
				    :reorder-specs (generic-function-reorder-specs ins))))
	(sys:structure-forward ins new-ins)
	(setf ins new-ins)))
    (apply ins  :modify-generic-function options)
    (let* ((declare (generic-function-declare ins))
	   (documentation (generic-function-documentation ins))
	   (arg-desc (generic-function-arglist-description ins))
	   (rest-p (or (plusp (arglist-description-n-optional arg-desc))
		       (arglist-description-&rest-p arg-desc)
		       (arglist-description-&key-p arg-desc)))
	   (new-fef (compile name
			 `(named-lambda (,name ( :generic-function ,ins
						    :descriptive-arglist
						    ,(if llp lambda-list
						       (generic-function-lambda-list ins))))
					;;The lambda-list needs to be changed. We must be sure to accept any keywords
					;;At this point. The checking is done by the lambda-list of the handler.
					;;The original lambda-list can be put as a declared arglist.
					(,@(specializable-arguments (generic-function-lambda-list ins))
					 ,@(and rest-p '(&rest .rest.)))
					,.(if documentation (cons documentation nil) nil)
					,@(if declare `((declare ,@declare)) nil)
					(declare (unspecial ,@(specializable-arguments (generic-function-lambda-list ins)) ))
					,@(and rest-p '(.rest.))
					. ,(compute-main-body ins))))
	   )
      ;; This is now done in COMPILER:LAP-MFEF
      ;;(sys:%p-dpb-offset 0 sys:%%FEF-Storage-Length-Generic-Function-Flag new-fef 1)
      (setf (generic-function-discriminator-code ins) new-fef)
      (if (symbolp new-fef)  ; DAB 01-16-90 Need to return a generic function object. See CLOS spec. p2-47 [10842]
	  (symbol-function new-fef)
	  new-fef)
      )))


(DEFMACRO WITH-ADDED-METHODS ((function-specifier lambda-list &rest options) &body body)
  "Syntax:  (WITH-ADDED-METHODS (function-spec lambda-list {option}*) {form}*)
Like GENERIC-LABELS except that the local function inherits methods from any 
previous definition of the function."
  ;; 12/08/88 DNG - Initial crude version - not completely correct.
  ;; 12/19/88 DNG - Fixed for (SETF ...) functions, but still not as complete as the compiler's version.
  (let ((required-args (loop for x in lambda-list
			     until (member x '(&rest &optional &key &aux))
			     unless (member x lambda-list-keywords :test #'eq)
			     collect (if (atom x) x (car x))))
	(defined (if (and (compiler:external-symbol-p function-specifier)
		       (fboundp function-specifier)
		       (member (symbol-package function-specifier)
			       '#,(mapcar #'find-package '("LISP" "TICL" "TICLOS" "SYS"))
			       :test #'eq))
		  't
		(if (symbolp function-specifier)
		    `(fboundp ',function-specifier)
		  `(fdefinedp ',function-specifier))))
	(outer-function (if (symbolp function-specifier)
			    `(quote ,function-specifier)
			  `(cdr (si:eval-at-load-time  ; DAB 01-16-90 Added package SI [10250]
				  (fdefinition-location ',function-specifier))))))
    `(ticlos:generic-labels ((,function-specifier ,lambda-list
			      ,(if (= (length required-args) (length lambda-list))
				   `(:method ,required-args
					     (if ,defined
						 (funcall ,outer-function . ,required-args)
					       (ticlos:no-applicable-method
						 #',function-specifier . ,required-args)))
				 `(:method (,@required-args &rest ticlos::.rest.)
					   (if ,defined
					       (apply ,outer-function
						      ,@required-args ticlos::.rest.)
					     (apply #'ticlos:no-applicable-method
						    #',function-specifier ,@required-args ticlos::.rest.))))
			      . ,options))
       . ,body)))
))
