;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10; Patch-file:T -*-

;;; Reason: Fixed standard-class' :build-class-slots method to prevent adjust-array from choking 
;;; when the class-slot-structure contains DTP-NULL pointers (representing unbound slots) 
;;; and the structure is being lengthened.  [SPR 10922]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/02/90 16:32:25 by BAF,
;;; while running on PCAI from band LOD6
;;; With SYSTEM 6.28, VIRTUAL-MEMORY 6.3, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.7, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.23, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.12,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.4, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.34, CLEH 6.5, IP 3.57,
;;;  Experimental BUG 11.18, Experimental CLX 6.8, CLUE 6.46, X11M 6.1,  microcode 604,
;;;  Band Name: Release 6.0 patched 1-19-90

#!C
; From file COMPOSE-CLASS.LISP#> CLOS; HOTEL:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; COMPOSE-CLASS.#"


(defmethod (standard-class :build-class-slots) ()
  (let* ((old-alist (class-class-slot-alist self))
	 (all-slots (class-all-slots self))
	 (new-class-slots (collect-body
			    (dolist (s all-slots)
			      (when (and (eq :class (car-safe (slot-allocation s)))
					 (eq self (cadr (slot-allocation s))))
				(collect (slot-name s))))))
	 (array (class-description-class-variable-structure class-description))
	 (size (or (and (arrayp array)
			(length array))  0)))
    (dolist (s old-alist)
      (if (member (car s) new-class-slots :test #'eq)
	  (setf new-class-slots (delete (car s) new-class-slots :test #'eq :count 1))
	  (setf (cdr s) nil)))
    (setf (class-class-slot-alist self)
	  (append old-alist
		  (collect-body
		    (dolist (s new-class-slots )
		      (collect (cons s size))
		      (incf size)))))
    (when (plusp size)
      (setf array (if array
		      (adjust-class-slot-structure array size)	; BAF 2/1/90
		      (make-class-slot-structure
			:make-array (:length size)
			:class-description  class-description)))
      (without-interrupts
	(dolist (s (class-class-slot-alist self))
	  (unless (assoc (car s) old-alist :test #'eq)
	    (sys:%P-STORE-DATA-TYPE-AND-POINTER
	      (locf (aref array (cdr s))) sys:DTP-Null 0))))
      (setf (class-description-class-variable-structure class-description)
	    array))
    (setf (class-description-class-slot-alist  class-description)
	  (copy-list
	    (mapcan #'(lambda (s)
			(and (eq :class (car-safe (slot-allocation s)))
			     (let ((slot-class (cadr (slot-allocation s))))
			       ;; The slot-class must be composed, so the class-slot
			       ;; can exist
			       (unless (or (eql slot-class self)
					   (class-composed-p slot-class))
				 (compose-class slot-class))
			       (cons (cons (slot-name s)
					   (class-slot-location
					     slot-class (slot-name s)))
				     nil))))
		    (class-all-slots self))))))

))

#!C
; From file COMPOSE-CLASS.LISP#> CLOS; HOTEL:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; COMPOSE-CLASS.#"


(defun adjust-class-slot-structure (array size)
  (without-interrupts
    (let ((unbound-indices (collect-body
			     (dotimes (i (length array))
			       (unless (location-boundp (locf (aref array i)))
				 (setf (aref array i) '#:unbound)
				 (collect i)) )) ))
      (setf array (adjust-array array size))
      (dolist (i unbound-indices)
	(SYS::%P-STORE-DATA-TYPE-AND-POINTER (locf (aref array i)) SYS:DTP-NULL 0))
      ))
  array )
))
