;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Fixed TICLOS::CACHE-HANDLER-FOR-ARGUMENTS to always call TICLOS::PUT-METHOD-IN-HASH-TABLE, regardless of
;;; whether an EQL specializer is used by any methods of the generic function. [SPR 11105]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 05/24/90 09:38:31 by FLORMAN,
;;; while running on PCAI from band LOD6
;;; With SYSTEM 6.32, VIRTUAL-MEMORY 6.3, EH 6.7, MAKE-SYSTEM 6.3, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.4, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-FILE 6.10, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.2,
;;;  COMPILER 6.14, TV 6.25, DATALINK 6.0, CHAOSNET 6.6, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.16,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.6, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.8, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.3, VISIDOC 6.7, TI-CLOS 6.46, CLEH 6.5, IP 3.62,
;;;  Experimental BUG 11.19, Experimental CLX 6.11, CLUE 6.104, X11M 6.1,  microcode 429,
;;;  Band Name: Rel6 patched 4/2/90

#!C
; From file METHOD-HASH-TABLE.LISP#> CLOS; Hotel:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; METHOD-HASH-TABLE.#"


(defun CACHE-HANDLER-FOR-ARGUMENTS (gfunc method &optional args)
  (let* ((method-specializers (reorder-parameter-specializers
				gfunc (method-parameter-specializers method)))
	 (arg-specializers (if (null args)
			       method-specializers
			       (mapcar #'(lambda (arg  specializer)
					   (if (individual-typep specializer)
					       specializer
					       (class-of arg)))
				       args method-specializers))))
    ;; If we need to create a new hash table to accommodate an inherited method, 
    ;; need to check the original dispatch path of the inherited method to see 
    ;; if if passes through any individual hash tables at or after the 
    ;; argument for which we make a new hash table.  If so, call 
    ;; COMBINE-METHODS to install all of the necessary individual entries in 
    ;; the new dispatch path.
    ;; Note: it would be more efficient to have add-method maintain in the generic
    ;; function object a list of specializer lists for which there are eql methods
    ;; defined, so that we don't have to scan the method list every time here.
    (labels ((ANY-EQL (mspecs aspecs mlist)
		      (when (or (null mspecs) (null mlist))
			(return-from any-eql nil))
		      (dolist (x mlist)
			(when (and (individual-typep (car x))
				   (not (equal (car x) (first aspecs)))
				   (eq (class-of (individual-type (car x)))
				       (first aspecs)))
			  (return-from any-eql t) ))
		      ;; Follow down the most likely path first.
		      (or (any-eql (rest mspecs) (rest aspecs)
				   (cdr (assoc (first mspecs) mlist :test #'equal)))
			  ;; Use SUBCLASSP as the last resort since it takes time.
			  (dolist (x mlist nil)
			    (when (and (not (individual-typep (car x)))
				       (not (eq (car x) (first mspecs)))  ; already checked that path
				       (subclassp (first aspecs) (car x))
				       (any-eql (rest mspecs) (rest aspecs) (cdr x)))
			      (return t))))
		      ))
      (if (any-eql method-specializers arg-specializers
		   (generic-function-method-list gfunc))
	  ;; Update this path and all others that may be affected.
	  (combine-methods gfunc arg-specializers) )
      (put-method-in-hash-table gfunc method arg-specializers) )
    (values)))

))
