;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Modified TICLOS::MAKE-COMBINED-METHOD, TICLOS::COMPUTE-SHARED-INITIALIZE-METHOD and friends to incorporate
;;; the class names properly into their function specs.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 05/30/90 16:57:25 by FLORMAN,
;;; while running on PCAI from band LOD6
;;; With SYSTEM 6.35, VIRTUAL-MEMORY 6.3, EH 6.8, MAKE-SYSTEM 6.3, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.5, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.8, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-FILE 6.13, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.2,
;;;  COMPILER 6.14, TV 6.25, DATALINK 6.0, CHAOSNET 6.8, GC 6.4, MEMORY-AUX 6.0, NVRAM 6.3,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.6, UCL 6.0, INPUT-EDITOR 6.0, METER 6.2, ZWEI 6.18,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.3, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.7, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.8, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.5, PROFILE 6.3, VISIDOC 6.7, TI-CLOS 6.47, CLEH 6.5, IP 3.65,
;;;  Experimental BUG 11.19, Experimental CLX 6.11, CLUE 6.104, X11M 6.1,  microcode 429,
;;;  Band Name: Rel6 patched 4/2/90

#!C
; From file MAKE-INSTANCE.LISP#> CLOS; Hotel:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; MAKE-INSTANCE.#"


(defun compute-shared-initialize-method (class)
  (let* ((slots (class-all-slots class))
	 (init-alist (class-all-slot-initargs class))
	 (init-keylist (mapcar #'(lambda (x) `((,(car x) ,(gensym)) nil ,(gensym))) init-alist))
	 (instance-slot-names (class-description-instance-slot-names (class-description class)))
	 (init-code (mapcar #'(lambda (x y)
				 (let ((offs (slot-offset-or-locative (second y) slots instance-slot-names)))
				   `(when ,(third x)
				      (setf ,(if (numberp offs)
						 `(sys:%instance-ref instance ,(1+ offs))
						 `(contents ,offs))
					    ,(second (first x))))))
			     init-keylist init-alist))
	 (loc-sym (gensym))
	 (default-code
	   (collect-body
	     (dolist (s slots)
	       (when (slot-initform s)
		 (let ((offs (slot-offset-or-locative (slot-name s) slots instance-slot-names)))
		   (collect
		     `(let ((,loc-sym ,(if (numberp offs)
					   `(locf (sys:%instance-ref  instance ',(1+ offs)))
					   offs)))
			(unless (slot-location-boundp ,loc-sym)
			  (setf (contents ,loc-sym) 
				,(if (and (slot-initclosure s)
					  (not (constantp (car (slot-initform s)))))
				     `(funcall ',(slot-initclosure s))
				     (car (slot-initform s)))))))))))))

    
    (multiple-value-bind (qualifiers specializers function lambda-list)
	(parse-method
	  `(((instance ,(or (class-name class) class)) slots-for-initform  &rest initargs )	; BAF 5/30/90
	    (declare (sys:function-parent ,(class-name class) defclass))
	    (apply #'(lambda (&key ,@init-keylist &allow-other-keys)
		       ,@init-code
		       (if (eq slots-for-initform t)
			   (progn
			     ,@default-code)
			   ;;the following code could be put in a function, shared by shared-initialize and this
			   ;;specialized method.
			   (shared-initialize-default instance slots-for-initform )))

		   initargs))
		   'shared-initialize nil)
      (let ((method (make-method qualifiers specializers
				 (compile nil function)
				 lambda-list)))
	(setf (getf (method-plist method ) 'computed-method) t)
	(add-method #'shared-initialize
		    method )))))

))

#!C
; From file METHOD-COMBINATION.LISP#> CLOS; Hotel:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; METHOD-COMBINATION.#"


(defun MAKE-COMBINED-METHOD (gen-function arg-types code derivation methods)
  (multiple-value-bind (lambda-list keylist keyp)
      (compute-combined-lambda-list gen-function methods)
    (let ((exp (sys:macroexpand-all code )) 
	  method m)
      (if (and (eq (car exp)
		   'sys:%apply-method)
	       (eq (car-safe (cadr exp))
		   'function)
	       (eq (nth-safe 4 exp )
		   nil)
	       (or (null keyp) (not (listp keylist)))
	       (setf m (if (consp (cadr exp))
			   (method-named (cadr (cadr exp)) t gen-function)
			   (sys:get-debug-info-field (sys:get-debug-info-struct (cadr exp)) :method)))
	       ;;Can't dispatch to interpreted functions.
	       (typep (*method-function m) 'compiled-function)
	       (let ((arg-desc (ticlos:method-arglist-description m)))
		 (= 0 (arglist-description-n-optional arg-desc))))
	  (progn (setf method (sys:make-flavor-instance 'null-combined-method
							:function nil
							:forwarded-method m
							:parameter-specializers arg-types
							:qualifiers '(:combined)
							:lambda-list lambda-list
							:derivation derivation))
		 (setf (getf (method-plist m) 'forwarded-from) method)
		 method)
	  
	  (let* ((function
		   (compile nil 
			    `(named-lambda ((method ,(generic-function-name gen-function)
						    :combined
						    ,(mapcar #'(lambda (type)	; BAF 05/30/90
								 (if (typep type 'class)
								     (class-name type)
								     type))
							     arg-types) )
					    (:descriptive-arglist ,(generic-function-lambda-list gen-function)))
					   ( ,@lambda-list)
					   (declare (unspecial ,@(generic-function-argument-precedence-order gen-function)))
					   ,@(and keyp (listp keylist)
						  `((check-keyargs (nthcdr ,(arglist-description-n-optional
									      (generic-function-arglist-description 
										gen-function ))
									   .rest. )
								   ',keylist)))
					   ,code)))
		 (sys:%inhibit-read-only t)
		 (ds (sys:get-debug-info-struct function t)))
	    (setf method (sys:make-flavor-instance 'combined-method
						   :parameter-specializers arg-types
						   :qualifiers '(:combined)
						   :lambda-list lambda-list
						   :function function
						   :derivation derivation))
	    (setf (sys:get-debug-info-field ds :method) method))
	  )
      (add-method gen-function  method)
      (when (generic-function-method-hash-table gen-function)
	(put-method-in-hash-table gen-function  method)
	;; remember that this handler is cached in the hash-table
	;; so we can remove it when the class gets redefined.
	;;	(dolist (c arg-types)
	;;	  (handler-cached (if (individual-typep c) (class-of (individual-type c)) c) method)))
	)
      method)))

))

#!C
; From file MAKE-INSTANCE.LISP#> CLOS; Hotel:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; MAKE-INSTANCE.#"


(defmethod (init-generic-function :after :add-method) (method)
  (LET((class (first (method-parameter-specializers method))))
    (when (AND (not (combined-method-p method))
	       (not (getf (method-plist method ) 'computed-method))
	       (typep class 'standard-class)
	       (not (eq class '#,(find-class 'standard-object))))
      (definalize-class class))))		; BAF 05/30/90

))

#!C
; From file MAKE-INSTANCE.LISP#> CLOS; Hotel:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; MAKE-INSTANCE.#"


(defun DEFINALIZE-CLASS (class)
  "   Mark the class and all of its subclasses as *NOT* finalized.  Remove any
   computed MAKE-INSTANCE methods so that FINALIZE-INHERITANCE will be called
   when appropriate."
  (when (internal-class-finalized-p class)
    (let* ((gfunc (get-generic-function-object #'make-instance))
	   (method (find-method gfunc nil (list (list 'eql class)) nil))
	   )
      (if (and method (getf (method-plist method) 'computed-method))
	  (remove-method gfunc method) ))
    (setf (class-description-finalize-status (class-description class)) 0) )
  (dolist (c (class-direct-subclasses class))
    (definalize-class c))
  (values))

))
