;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Written 05/31/89 13:33:51 by MCCREARY,
;;; Reason: Fix two problems introduced in CLOS patch 6.4.  Fixes
;;; PUT-DEFAULT-VALUE-IN-HASH-TABLE and PUT-METHOD-IN-HASH-TABLE
;;; so that methods are entered into the correct hash table.
;;; while running on Jules-Verne from band LODA
;;; With SYSTEM 6.0, VIRTUAL-MEMORY 6.0, EH 6.0, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.0, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.0, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.0, TV 6.0, DATALINK 6.0, CHAOSNET 6.0, GC 6.0, MEMORY-AUX 6.0, NVRAM 6.0,
;;;  SYSLOG 6.0, STREAMER-TAPE 6.0, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.0,
;;;  DEBUG-TOOLS 6.0, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.0, SERIAL 6.0, PRINTER 6.0, MAC-PRINTER-TYPES 6.0, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.0, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.0, VISIDOC 6.0, TI-CLOS 6.4, CLEH 6.0, IP 3.45,
;;;  Experimental BUG 11.4, CLX 6.0, CLUE 6.0, X11M 6.0,  microcode 429, Band Name: Release 6.0 + SLE 5/22

#!C
; From file METHOD-HASH-TABLE.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; METHOD-HASH-TABLE.#"


(defun put-default-value-in-hash-table (hash-table &optional value-list)
  ;; HASH-TABLE is the table we intend to fill with default values.
  ;; VALUE-LIST contains the entries to be put into HASH-TABLE.  The first item
  ;; in the list is either a locative to the default method or a hash table
  ;; containing the default method.  The second item is the mapping table list.
  ;; The third item, if supplied, is the list of arguments passed to put-method-
  ;; in-hash-table.  This is needed so that we can traverse down the levels of
  ;; hash tables and put the default method in its proper place.
  ;; 05/30/89 clm - put the default value into the proper hash table.
  ;; 05/31/89 clm - fix problem in previous patch; was destructively modifying
  ;;                needed info within the DO.
  (with-lock ((sys:hash-table-lock hash-table) :whostate "Hash Table Lock")
    ;; Set all of hash table to NIL with cdr-next.
    (without-interrupts
      (multiple-value-bind (subh mtl args)
	  (values-list value-list)
	(setf hash-table (sys:follow-structure hash-table))
	(if value-list
	    (setf (sys:hash-table-instance hash-table )
		  (if (third value-list) (list subh mtl) value-list) )
	    (setf value-list (sys:hash-table-instance hash-table )))
	;;Do a do loop in order to preserve the existing entries.
	(let ((hlen (array-total-size hash-table)))
	  (do ((p (locf (aref hash-table 0)) (sys:%make-pointer-offset sys:dtp-locative p 3)))
	      ((> (sys:%pointer-difference p hash-table) hlen))
	    (if (= (sys:%p-data-type p) sys:dtp-null)
		(progn
		  ;;fill in all the empty slots in the hash table with the default value
		  (sys:%p-store-contents (sys:%make-pointer-offset sys:dtp-locative p 1) (car value-list))
		  (sys:%p-store-contents (sys:%make-pointer-offset sys:dtp-locative p 2) (cadr value-list)))
		;;if the slot is not empty and the entry is another hash table, go to next level to
		;;enter the default value
		(let ((new-hash (gethash (car p) hash-table))
		      new-hash2)
		  
		  (when (and new-hash
			     (hash-table-p new-hash)
			     args
			     )
		    (when (hash-table-p new-hash)
		      (setq new-hash2 (gethash (class-hash-key (car (third value-list))) new-hash)))
		    (if (individual-hash-table-p new-hash2)
			;;when recursing, if this is the last level of recursion (indicated by
			;;the next arg being nil, we cannot use a hash table value to fill in the default
			;;slots in the next level individual-hash-table, this would cause a "ran out of
			;;args to dispatch on before finding method" error.  the next level iht must
			;;be filled in with pointers to the default method itself.
			(let ((key (car args))
			      subh-t
			      mtl-t
			      (argl (cdr args)))
			  (when (null argl)
			    ;;this is the last level of recursion, get value from hash-table in value-list
			    ;;and pass that
			    (multiple-value-bind (loc ignore entry)
				(gethash (class-hash-key key) subh)
			      (setf subh-t loc)
			      (setf mtl-t (third entry))) )
			  (put-default-value-in-hash-table new-hash2 (list subh-t mtl-t argl)))
			(multiple-value-bind (nil nil entry)
			    (gethash (class-hash-key (car (third value-list))) (car value-list))
			  ;;the value (gethash (class-hash-key (third value-list)) (car value-list))
			  ;;is the hash table in which we have stuffed the default method that we are
			  ;;filling up the iht with - after this each entry in the hash table will be
			  ;;set to this
			  (puthash (class-hash-key (car (third value-list)))
				   (cadr entry)
				   new-hash
				   (caddr entry))) )  ) )
		)))
	)				   ;multiple-value-bind
      value-list
      )))
))

#!C
; From file METHOD-HASH-TABLE.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; METHOD-HASH-TABLE.#"


(defun put-method-in-hash-table (gfunc method &optional
				 (arg-specializers  nil))
  ;;arg-specializers are already ordered by arguments-precedence-order
  ;; 05/30/89 clm - made two fixes so that methods will be put into the
  ;;                correct hash tables.
  ;; 05/31/89 clm - fix for last fix; when updating the hash-table of a
  ;;                subclass, make sure we don't overwrite a more specific
  ;;                method for that class.
  (let*((ht (generic-function-method-hash-table gfunc))
	m
	(specializers (reorder-parameter-specializers
			gfunc
			(method-parameter-specializers method))))
    (multiple-value-bind (meth-loc mtl)
	(get-hash-entries-from-handler method (method-parameter-specializers method)
				       (order-parameter-specializers gfunc
								     (or arg-specializers specializers )))
      (labels
	((put-arg-in-hash-table (args specs ht &optional parent)
	   (declare (list args specs))
	   (if (null args)
	       (values meth-loc mtl)
	       (progn
		 (when (or (null ht) (not (hash-table-p ht)))
		   (if (and parent
			    (every #'(lambda (c) (class-eql c *t-class*)) specs)
			    (locativep (gethash (class-hash-key *t-class*) parent)))
		       ;; This is a default handler, we don't need to cons up new hash tables.
		       (return-from put-arg-in-hash-table meth-loc mtl)
		       (progn
			 ;; ;;There is no right place, since it was a default.
			 ;;(unless (null ht)
			 ;; ;;handler is in there. record it so it can be put back in the right place.
			 ;;(setf m (sys:get-debug-info-field
			 ;;               (sys:get-debug-info-struct (contents ht)) :method))
			 ;; (unless m (ferror nil "unable to get the method object from the function ~s" ht))) 
			 (setf ht (sys:make-hash-array  :number-of-values 2 :funcallable-p t :test #'eq)))))
		 (if (individual-typep (car args))
		     (let ((dht (multiple-value-bind (ov ignore vals)
				    (gethash (class-hash-key (class-of (individual-type (car args)))) ht)
				  (if (individual-hash-table-p ov)
				      ov
				      (make-individual-dispatch-hash-table 
					:default-value (or (copy-list (cdr vals)) (list nil nil)))))))
		       (puthash (class-hash-key (class-of (individual-type (car args)))) dht ht t)
		       (multiple-value-bind (subh mtl)
			   (put-arg-in-hash-table (cdr args) (cdr specs)
						  (gethash (individual-type (car args))
							   (individual-dispatch-hash-table-ht dht)))
			 (puthash (individual-type (car args))
				  subh
				  (individual-dispatch-hash-table-ht dht)
				  mtl))
		       ht)
		     (multiple-value-bind (ov ignore )
			 (gethash (class-hash-key (car args)) ht)
		       (if (individual-hash-table-p ov)
			   (multiple-value-bind  (subh mtl)
			       (put-arg-in-hash-table (cdr args) (cdr specs)
						      (car (individual-dispatch-hash-table-default-value ov)))
			     (let ((def-arg-list (cdr args)))
			       (dolist (arg def-arg-list)
				 (when (and (consp arg) (eq (car arg) 'eql))
				   (setq def-arg-list nil)
				   (return nil)))
			       ;;pass the args so that we can put default value
			       ;;into the correct hash-table (done at run time).
			       (setf (individual-dispatch-hash-table-default-value ov)
				     (list subh mtl def-arg-list) )))
			   (multiple-value-bind (subh mtl)
			       (put-arg-in-hash-table (cdr args) (cdr specs)
						      (gethash (class-hash-key (car args)) ht)
						      ht)
			     (prog1
			       (puthash (class-hash-key (car args))
					subh
					ht
					mtl)
			       (when (classp (car args))
				   (maphash
				     #'(lambda (key val ignore)
					 ;;if (car args) is a super-class of other classes
					 ;;in this hash-table level, then we also need to
					 ;;add the method to subclasses of the super.
					 (when
					   (and (subclassp (class-description-class-object key)
							   (car args))
						(not (eq (class-description-class-object key) (car args)))) 
					   (if (individual-hash-table-p val)
					       (setf (individual-dispatch-hash-table-default-value val)
						     (list subh mtl )) 
					       ;;must check for the _last_ node to make sure we are
					       ;;not overwriting an already existing, more specific
					       ;;method for the subclass
					       (when (and (hash-table-p val)
							  (or (cddr args)    ;not the last node
							      (not (gethash (class-hash-key (cadr args))
									    val))))  ;method already exists
						 
						 (put-arg-in-hash-table (cdr args) (cdr specs)
								    val
								    ht)) )))
				     ht) ) )))
		       ht))))))
	(put-arg-in-hash-table (or arg-specializers specializers ) specializers  ht)
	(when m
	  (put-method-in-hash-table gfunc m)))))
  (values))
))
