;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10; Patch-file:T; Fonts:(CPTFONT CPTFONTB) -*-

;;; Reason: Fix handling of EQL specializers in methods using CALL-NEXT-METHOD.  [SPR 9914]

;;; Copyright (C) 1989 Texas Instruments Incorporated. All rights reserved.

;;; Patch file for TI-CLOS version 6.8
;;; Written 06/16/89 14:38:34 by GRAY,
;;; while running on Kelvin from band LOD2
;;; With SYSTEM 6.6, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  Inconsistent COMPILER 6.4, TV 6.10, DATALINK 6.0, CHAOSNET 6.0, GC 6.2, MEMORY-AUX 6.0,
;;;  NVRAM 6.0, SYSLOG 6.0, STREAMER-TAPE 6.2, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0,
;;;  ZWEI 6.2, Inconsistent DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0,
;;;  DATALINK-DISPLAYS 6.0, FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.1, MAC-PRINTER-TYPES 6.1,
;;;  PRINTER-TYPES 6.0, IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0,
;;;  TELNET 6.0, VT100 6.0, NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, Inconsistent TI-CLOS 6.7,
;;;  CLEH 6.4, IP 3.46, Experimental BUG 11.9, Experimental CLX 6.1, CLUE 6.5, X11M 6.1,
;;;  Experimental DOCUMENTER 6.0,  microcode 429, Band Name: 6.0 SLE 6/5 + u429 6/8

;;; BUG REPORT NUMBER:  9914
;;;
;;; PROBLEM:  An invalid warning message was issued by the compiler on an 
;;;	in-memory compilation of a CLOS method that uses an EQL specializer and 
;;;	CALL-NEXT-METHOD.  For example, CTRL-SHIFT-C on:
;;;	  (defmethod fvv ((b (eql 'foo)))
;;;	    (call-next-method))
;;;	says:  "B is declared to be of type CONS but its initial value is a 
;;;	#<BUILT-IN-CLASS SYMBOL>."
;;;
;;; DIAGNOSIS:  This is actually the fault of TICLOS::PARSE-METHOD, which is not
;;;	correctly handling EQL specializers when generating a DEFAULT-TYPE 
;;;	declaration.
;;;
;;; SOLUTION:  Call TYPE-NAME to convert the specializer name to a 
;;;	valid type name.

#!C
; From file GENERIC.LISP#> CLOS; MR-X:
#10R TICLOS#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TICLOS"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLOS; GENERIC.#"


1;;  6/16/89 DNG - When constructing a COMPILER:DEFAULT-TYPE declaration, call 
;;*		1TYPE-NAME so that EQL specializers are handled correctly.  [SPR 9914]*

;;;Should check the arglist better. A nil will be accepted here!
(defun parse-method (things &optional name environment)
  (declare (values method-combination-identifiers specializers function
		   specialized-lambda-list fspec call-next-method-used additional-forms))
  (let* ((additional-forms '())
	 (method-combination-identifiers
	   (collect-body
	     (loop (when(listp (car things ))
		     (return))
		   (collect (pop things)))))
	 (specialized-lambda-list (pop things))
	 (body things)
	 (parameter-and-specializers (lambda-list-specializer  specialized-lambda-list))
	 (lambda-list (extract-lambda-list  specialized-lambda-list))
	 (specializers
	   (mapcar #'(lambda (sp)
		       (unless (symbolp (first sp))
			 (error "Non-symbol ~S found where a method parameter name was expected."
				(first sp)))
		       (let ((spec (second sp)))
			 (cond ((symbolp spec) spec)
			       ((and (consp spec)
				     (eq (car spec) 'eql)
				     (null (cddr spec)))
				(let ((form (second spec)))
				  (when (and compiler:qc-file-in-progress
					     (fboundp 'compiler:macro-cons-area))
				    (setq form
					  (let ((compiler:p1value 'compiler:single-value))
					    (compiler:pre-optimize form t))))
				  (cond ((constantp form)
					 ;; Evaluate it now in current lexical environment.
					 (let ((value (compiler:eval-for-target form environment)))
					   (if (null value)
					       'null
					     `(eql ,value))))
					((or (and compiler:qc-file-in-progress compiler:vars)
					     (compiler:env-vars environment)
					     (compiler:env-functions environment))
					 ;; Generate code to evaluate in proper lexical environment.
					 (setq spec (list 'eql form))
					 ;; Use load-time-value to prevent constant folding or write-protection.
					 (push-end `(rplaca (cdr (compiler:load-time-value ',spec nil)) ,form)
						   additional-forms)
					 spec)
					((or (not compiler:qc-file-in-progress)
					     compiler:qc-file-load-flag)
					 ;; Evaluate it now in current lexical environment.
					 (let ((value (compiler:eval-for-target form environment)))
					   (if (null value)
					       'null
					     `(eql ,value))))
					(t	; Have the loader evaluate it.
					 `(eql ,(cons compiler:eval-at-load-time-marker form))))))
			       ((sys:classp spec) spec)
			       (t (error "Invalid specializer syntax in method parameter ~S." sp)))))
		   parameter-and-specializers))
	 (fspec `(method ,name ,@method-combination-identifiers
			 ,specializers))
	 (argl (gensym))
	 (spec-args  (specializable-arguments lambda-list)))
    (multiple-value-bind (undeclared-body declarations documentation)
	(sys:parse-body body environment t)
      ;; Find and remove any user-supplied ARGLIST declaration which is not 
      ;; congruent with the actual lambda list, because we need to use the 
      ;; :descriptive-arglist to remember the original DEFMETHOD args, and an 
      ;; ARGLIST declaration is not very useful here anyway.  -- DNG 12/14/88
      (do ((decls declarations (rest decls)))
	  ((atom decls))
	(let ((temp (assoc 'arglist (rest (first decls)))))
	  (when (and temp
		     (not (equal (multiple-value-list (check-lambda-list (rest temp) t))
				 (multiple-value-list (check-lambda-list lambda-list t)))))
	    (setf (first decls) (remove temp (the list (first decls)) :test #'eq)))))
      (when (and documentation (record-documentation-p))
	(push documentation declarations))
      (let* ((call-next-method-used
	       (catch 'walk
		 (ignore-errors
		   (compiler:code-walk
		     `(progn . ,undeclared-body)
		     #'(lambda (form)
			 (when (member (car form) '(call-next-method next-method-p) :test #'eq)
			   (throw 'walk t))
			 form)
		     #'identity
		     nil
		     environment
		     :function #'(lambda (form)
				   (when (member (cadr form) '(call-next-method next-method-p) :test #'eq)
				     (throw 'walk t))
				   form) )
		   ;; successfully walked the tree without finding anything.
		   (throw 'walk nil))
		 ;; if there was an error, assume the worst.
		 t))
	     (block-form (cond ((and (symbolp name) (not (null name)))
				`(block ,name . ,undeclared-body))
			       ((and (consp name)
				     (eq (first name) 'setf)
				     (symbolp (second name)))
				`(block ,(second name) . ,undeclared-body))
			       (t `(progn . ,undeclared-body))))
	     (function (if call-next-method-used
			   (let ((subargl (compute-lambda-list-for-method lambda-list nil))
				 (reqargs (mapcar #'(lambda (x) (if (consp x) (car x) x)) spec-args)))
			     `(,(if name
				    'named-lambda
				  'lambda)
			       ,@(and name `((,fspec  (:method nil :DESCRIPTIVE-ARGLIST ,lambda-list))))
			       (,@spec-args ,@(and subargl `(&rest ,argl)))
			       (DECLARE (TICLOS::SPECIALIZERS ,@specializers))
			       ;; Because the same declarations are being put in two places, the magic 
			       ;; flags .ARG. and .AUX. tell the compiler to suppress warnings on 
			       ;; declarations referring to variables not present.
			       ,@(and declarations `((declare (compiler::.ARG.)) . ,declarations))
			       (flet (,@(generate-call-next-method '(SYS:NEXT-METHOD-LIST)
								   `(,@spec-args ,(and subargl  argl))))
				 (declare (inline call-next-method next-method-p))
				 #'call-next-method
				 #'next-method-p
				 (apply #'(lambda ,(append reqargs subargl)
					    ,@(and declarations `((declare (compiler::.AUX.)) . ,declarations))
					    (declare . ,(loop for x in parameter-and-specializers
							      unless (eq (second x) 't)
							      collect `(compiler:default-type
									 ,1(type-name* (second x)1)* ,(first x))))
					    ,@reqargs	; to suppress compiler warnings on unused variables
					    ,block-form)
					,@reqargs ,(and subargl  argl)))))
			 `(,(if name
				'named-lambda
			      'lambda) ,@(and name `((,fspec  (:method nil :DESCRIPTIVE-ARGLIST ,lambda-list
								       :next-methods-not-used t	    ))))
			   ( ,@(compute-lambda-list-for-method lambda-list))
			   (DECLARE (TICLOS::SPECIALIZERS ,@specializers))
			   ,@declarations
			   ,@(specialized-arguments specialized-lambda-list )
			   ,block-form ))))
	(values method-combination-identifiers specializers function
		specialized-lambda-list fspec call-next-method-used additional-forms)))))
))
