;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Written 05/23/89 14:51:44 by KK,
;;; Reason: dispatch-event: remove compression of :enter/leave-notify events.
;;; while running on Sierra from band LOD1
;;; With SYSTEM 6.2, VIRTUAL-MEMORY 6.0, EH 6.0, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.0, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.0, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.1, TV 6.5, DATALINK 6.0, CHAOSNET 6.0, GC 6.0, MEMORY-AUX 6.0, NVRAM 6.0,
;;;  SYSLOG 6.0, STREAMER-TAPE 6.0, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.0,
;;;  DEBUG-TOOLS 6.0, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.0, SERIAL 6.0, PRINTER 6.0, MAC-PRINTER-TYPES 6.0, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.0, TI-CLOS 6.0, CLEH 6.0, IP 3.45,
;;;  Experimental BUG 11.5, CLX 6.0, CLUE 6.1, X11M 6.0, Experimental SC 1.3,  microcode 429,
;;;  Band Name: Release 6.0 + SLE 5/22

#!C
; From file EVENTS.LISP#> CLUE; MR-X:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "CLUE: CLUE; EVENTS.#"


(defun dispatch-event (event event-key send-event-p sequence contact)
  ;; Called from PROCESS-NEXT-EVENT to filter events and call event handlers.
  (declare (inline sensitive-p))
  (with-slots ((event_key key)
	       (event-sequence sequence)
	       (event-send-event-p send-event-p)
	       (event-contact contact)) (the event event)
    (setf event_key event-key
	  event-send-event-p send-event-p
	  event-sequence sequence
	  event-contact contact))
  ;;
  ;; Check for destroyed or invalid event-windows
  ;;
  (if (not (typep contact 'basic-contact))
      ;; Not a contact!!  Let the default root contact handle the event
      (handle-event (display-root (drawable-display contact)) event)
      
      (if (destroyed-p contact)
	  
	  ;; Destroyed-contact!
	  (when (eq event-key :destroy-notify)
	    (destroy-finish contact))

	  ;; Bind event for reference within with-event forms
	  (let (($event$ event))
	    (declare (special $event$))
	    
	    ;;
	    ;; Do key translation
	    ;;
	    (when (or (eq event-key :key-press)
		      (eq event-key :key-release))
	      (with-slots (keysym character display code state) (the event event)
		(let ((keysym-index (default-keysym-index display code state)))
		  (setf keysym (keycode->keysym display code keysym-index)
			character (keycode->character display code state :keysym-index keysym-index)))))
	    ;;
	    ;; Call the before event handlers
	    ;;
	    (let ((actions (before-actions (contact-display contact))))
	      (when actions
		(let ((class (class-name-of contact)))	      
		  (dolist (before-action actions)
		    (when (subtypep class (car before-action))
		      (call-action-internal contact (cdr before-action)))))))
	    ;;
	    ;; Handle insensitive contacts
	    ;;
	    (when (and (member event-key *sensitive-events* :test #'EQ)
		       (not (sensitive-p contact)))
	      (return-from dispatch-event nil))
	    
	    ;;
	    ;; Handle modes 
	    ;;
	    (let ((modes (display-mode-stack (contact-display contact))))
	      (when (and modes (not (contact-mode contact)))
		(when
		  (or (member event-key *restrict-events* :test #'eq)
		      (and (member event-key *remap-events* :test #'eq)
			   (dolist (mode modes t) ;; Search for first :spring-loaded mode
			     (when (eq (second mode) :spring-loaded)
			       (format t "~%Remapping ~s from ~s to ~s" event-key contact (first mode)) ;; *** DEBUG ***
			       (setq contact (first mode)) ;; Remap contact
			       (return nil)))))
		  ;; Call mode action on for first :exclusive or :spring-loaded mode
		  (dolist (mode modes)
		    (unless (eq (second mode) :non-exclusive)
		      (call-action-internal (first mode) (cddr mode))
		      ;; quit
		      (return-from dispatch-event nil))))))
	    
	    ;; 
	    ;; Handle event compression
	    ;;
	    (with-slots ((contact-compress-motion compress-motion)
			 (contact-compress-exposures compress-exposures))
			(the contact contact)
	      
	      (case event-key
		(:exposure			; Check for exposure compression
		 (when (and (eq contact-compress-exposures :on)
			    (plusp (slot-value (the event event) 'count)))
		   (return-from dispatch-event nil)))
		
		(:motion-notify			; Check for motion compression
		 (when (eq contact-compress-motion :on)
		   (let ((count 0))
		     
		     ;; Count consecutive :motion-notify's currently in queue
		     (event-case ((contact-display contact) :peek-p t :timeout 0)
		       (:motion-notify (window)
				       (not (and (eq window contact) (incf count))))
		       (otherwise ()   t))
		     
		     (when (plusp count) 
		       ;; Remove all but last and quit immediately
		       (do () ((zerop (decf count)))
			 (event-case ((contact-display contact) :timeout 0)
			   (otherwise ()   t)))
		       (return-from dispatch-event nil)))))))
	    ;;
	    ;; Handle event translations
	    ;;
	    (handle-event contact event)))))

))
