;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: find-gcontext: Ensure that when last cache element is modified, that it is correctly deleted
;;; from cache.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for CLUE version 6.20
;;; Written 10/10/89 15:56:43 by KK,
;;; while running on Quintillian from band LOD2
;;; With SYSTEM 6.17, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.1, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.2, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.3, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.12, TV 6.15, DATALINK 6.0, CHAOSNET 6.1, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.5,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.4, TI-CLOS 6.24, CLEH 6.5, IP 3.50,
;;;  Experimental CLX 6.4, CLUE 6.19, X11M 6.14, Experimental BUG 11.15, Experimental SC 2.0,
;;;  Experimental CLIO 6.0, Experimental CLIO-TEST 5.0,  microcode 429, Band Name: Rel 6.0 + SLE 9/8

#!C
; From file GC-CACHE.LISP#> CLUE; Hotel:
#10R XLIB#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "XLIB"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; GC-CACHE.#"


(defun find-gcontext (gcontext drawable)
  (let ((display     (drawable-display drawable))
	(local-state (gcontext-local-state gcontext))
	(cache (getf (drawable-plist drawable) 'gcontext-cache)))
    (declare (type list cache))
    (declare-array gcontext-state local-state)
    
    (with-display (display)			; Get a lock, just in case
      ;; The cache starts with a null cons.  This enables us to
      ;; update the cache without searching the plist again.
      (unless cache
	(setf (getf (drawable-plist drawable) 'gcontext-cache)
	      (setq cache (list nil))))
      
      ;; Ensure the timestamp and font match
      (setf (gcontext-internal-timestamp local-state) 0)
      (setf (gcontext-internal-font local-state) nil)
      
      ;; Search for a matching gcontext
      (do* ((previous cache entry)
	    (entry (cdr cache) (cdr entry))
	    (n 0 (index+ n 1)))
	   
	   ((endp entry)
	    (if (< n *gcontext-cache-length*)
		;; Create a new gcontext if our limit isn't up
		(create-temp-gcontext gcontext drawable)
		
		;; Else modify the last cache entry
		(let* ((gc                   (first previous))
		       (updated-local-state  (gcontext-local-state gc))
		       (updated-server-state (gcontext-server-state gc)))
		  (declare (type gcontext gc))
		  (declare-array gcontext-state updated-local-state updated-server-state)
		  
		  ;; Modify cache gcontext with requested state
		  (replace updated-local-state local-state)

		  ;; Ensure that force-gcontext-changes will update server state
		  (fill updated-server-state nil)
		  (setf (gcontext-internal-timestamp updated-server-state) -1)

		  ;; Remove from last entry from cache
		  (rplacd (nthcdr (index- n 1) cache) nil)
		  (deallocate-gcontext gcontext)
		  gc)))
	
	(declare (type list previous entry))
	
	(let* ((gc         (car entry))
	       (test-state (gcontext-local-state gc)))
	  (declare (type gcontext gc))
	  (declare-array gcontext-state test-state)
	  
	  (when (equalp test-state local-state)	; Found a match!
	    (setf (cdr previous) (cdr entry))
	    (deallocate-gcontext gcontext)
	    (return gc)))))))

))
