;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Patch file for CLUE version 6.22
;;; Reason: root gmgmt: Process any pending :configure-notify before reconfiguring.
;;; Written 10/11/89 18:24:30 by KK,
;;; while running on RIGI from band n817
;;; With SYSTEM 6.19, GC 6.3, VIRTUAL-MEMORY 6.2, MICRONET 6.0, MICRONET-COMM 6.1,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-PATHNAME 6.2, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.1,
;;;  BASIC-NAMESPACE 6.4, BASIC-FILE 6.4, RPC 6.1, NFS-MX 6.3, EH 6.5, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.12, TV 6.15, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.7, METER 6.1, ZWEI 6.5, DEBUG-TOOLS 6.3, WINDOW-MX 6.8, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.24, CLEH 6.5, NETWORK-PATHNAME 6.0,
;;;  NETWORK-NAMESPACE 6.0, DATALINK 6.0, CHAOSNET 6.1, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.2, IP 3.50, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.3, MAIL-READER 6.5,
;;;  TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.4, DECNET 1.70, VISIDOC 6.5, PROFILE 6.2,
;;;  Experimental SC 2.0, Experimental CLX 6.4, CLUE 6.20,  microcode 138, Band Name: microExplorer Network (8/17)

#!C
; From file ROOT-GMGMT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; ROOT-GMGMT.#"


(defmethod manage-geometry ((parent root) (shell wm-shell) x y width height border-width &key) 
  (declare (type contact shell)
	   (type (or null int16) x y)
	   (type (or null card16) width height border-width)
	   (values success-p x y width height border-width))

  (with-slots ((contact-x x)
	       (contact-y y)
	       (contact-width width)
	       (contact-height height)
	       (contact-border-width border-width)) (the contact shell)
    
    (if (realized-p shell)

	(progn 
	  ;; Wait for any previous :configure-notify events to be handled
	  (process-all-events (contact-display parent))
	  
	  ;; Negotiate top-level geometry with window mgr
	  (with-event-mode (shell '(:configure-notify (throw-action :configure-notify)))      
	    ;; Reconfigure top-level window, if necessary
	    (let (changed-p)
	      (with-state (shell)
		(when (and x (/= x contact-x)
			   (setf (drawable-x shell) x))
		  (setf changed-p t))
		(when (and y (/= y contact-y)
			   (setf (drawable-y shell) y))
		  (setf changed-p t))
		(when (and width (/= width contact-width)
			   (setf (drawable-width shell) width))
		  (setf changed-p t))
		(when (and height (/= height contact-height)
			   (setf (drawable-height shell) height))
		  (setf changed-p t))
		(when (and border-width (/= border-width contact-border-width)
			   (setf (drawable-border-width shell) border-width))
		  (setf changed-p t)))
	      
	      ;; Return approved geometry
	      (values
		(or
		  ;; Null changed approved immediately
		  (not changed-p)
		  
		  ;; Actual change approved if it is not modified by window mgr.
		  (progn
		    ;; Wait for :configure-notify to report actual new window geometry.
		    ;; Top-level shell's handle-event will update geometry slots in
		    ;; response to :configure-notify.	      
		    (catch :configure-notify
		      (loop (process-next-event (contact-display parent))))
		    
		    ;; Assert: shell slots now contain actual (wm-approved) geometry
		    ;; Return approval of original geometry request
		    (and
		      (not (and x (/= x contact-x)))
		      (not (and y (/= y contact-y)))
		      (not (and width (/= width contact-width)))
		      (not (and height (/= height contact-height)))
		      (not (and border-width (/= border-width contact-border-width))))))
		
		contact-x contact-y contact-width contact-height contact-border-width))))

	;; Else approve change to unrealized shell immediately
	(values t
		(or x contact-x) (or y contact-y)
		(or width contact-width) (or height contact-height)
		(or border-width contact-border-width)))))


(defmethod manage-priority ((parent root) (shell wm-shell) priority sibling &key)  
  (declare (type (member :above :below :top-if :bottom-if :opposite) priority)
	   (type (or null contact) sibling)
	   (values success-p priority sibling))

  ;; Wait for any previous :configure-notify events to be handled
  (process-all-events (contact-display parent))
  
  (with-event-mode (shell '(:configure-notify return-above-sibling))
    
    ;; Reconfigure top-level window
    (setf (window-priority shell sibling) priority)
    
    ;; Wait for :configure-notify to report actual new window priority.
    (let ((above-sibling
	    (catch :configure-notify
	      (loop (process-next-event (contact-display parent))))))
      
      ;; Return approval for original priority request
      (values
	(and (eq sibling above-sibling) (eq priority :above))
	:above))))

))
