;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Patch file for CLUE version 6.25
;;; Reason: define-contact-resources: Signal error if different slot and resource :initform's are given for
;;; a contact resource. This prevents a bug which caused steps #4 and #5 in resource initialization
;;; (see manual, sec 18.2.1) to be reversed.
;;; Written 10/13/89 11:49:58 by KK,
;;; while running on RIGI from band n817
;;; With SYSTEM 6.19, GC 6.3, VIRTUAL-MEMORY 6.2, MICRONET 6.0, MICRONET-COMM 6.1,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-PATHNAME 6.2, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.1,
;;;  BASIC-NAMESPACE 6.4, BASIC-FILE 6.4, RPC 6.1, NFS-MX 6.3, EH 6.5, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.12, TV 6.15, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.7, METER 6.1, ZWEI 6.5, DEBUG-TOOLS 6.3, WINDOW-MX 6.8, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.24, CLEH 6.5, NETWORK-PATHNAME 6.0,
;;;  NETWORK-NAMESPACE 6.0, DATALINK 6.0, CHAOSNET 6.1, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.2, IP 3.50, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.3, MAIL-READER 6.5,
;;;  TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.4, DECNET 1.70, VISIDOC 6.5, PROFILE 6.2,
;;;  Experimental SC 2.0, Experimental CLX 6.5, CLUE 6.24,  microcode 138, Band Name: microExplorer Network (8/17)

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(defun define-contact-resources (class superclass-names variables resource-list)
  "Construct and validate the resource list for CLASS.  A :slot option is added to
each resource specification containing the resource slot name, if any."
  (do* (name
	(resources resource-list   (cdr resources))
	(resource  (car resources) (car resources))
	
	;; Initialize result with all inherited resources
	(result    (mapcan #'(lambda (parent) (copy-list (clue-resources parent)))
			   superclass-names)))
       ((endp resources)
	(reconcile-contact-resource-with-slots variables result))    
    
    ;; Make name and class keywords
    (let ((class (getf (cdr resource) :class)))
      (setf name (intern (string (car resource)) 'keyword)
	    (car resource) name)
      (when class
	(setf (getf (cdr resource) :class) (intern (string class) 'keyword))))

    ;; Error checking
    (do ((option (cdr resource) (cddr option)))
	((endp option))      
      (let ((key (car option)))
	
	(assert (member key '(:initform :type :class :documentation :slot :remove)) ()
		"~s is an unknown option for resource ~s in ~s." key (car resource) class)
	
	(when (eq key :initform)
	  (assert
	    (let ((slot-initform (getf (cdr (assoc (symbol-name name) variables
						   :key #'symbol-name
						   :test #'equal))
				       :initform)))
	      (or (not slot-initform) (equal slot-initform (second option))))
	    ()
	    "Different slot and resource :initform's specified for resource ~s."
	    (car resource)))))
    
    ;; Merge resource with parent's resource
    (do* ((inherited-resource (cdr (assoc name result :test #'eq)))
	  (old                inherited-resource   (cddr old)))
	 ((endp old))
      (unless (getf (cdr resource) (car old))
	(setq resource (append resource `(,(car old) ,(cadr old))))))
    
    ;; Add (updated) resource spec to final list (unless the :remove option was given)
    (setq result (delete name result :key #'car :test #'eq))
    (unless (getf (cdr resource) :remove) (push resource result))))

))
