;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Patch file for CLUE version 6.34
;;; Reason: Make the following functions generic: add-event, delete-event, event-actions, initial-state-transition.
;;; Make change-priority call a new generic function (setf contact-priority) to actually reset window priority.
;;; Written 11/20/89 17:55:54 by KK,
;;; while running on RIGI from band n817
;;; With SYSTEM 6.23, GC 6.3, VIRTUAL-MEMORY 6.2, MICRONET 6.0, MICRONET-COMM 6.1,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-PATHNAME 6.2, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.6, BASIC-FILE 6.6, RPC 6.2, NFS-MX 6.4, EH 6.5, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.19, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.7, METER 6.1, ZWEI 6.8, DEBUG-TOOLS 6.3, WINDOW-MX 6.9, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.26, CLEH 6.5, NETWORK-PATHNAME 6.0,
;;;  NETWORK-NAMESPACE 6.0, DATALINK 6.0, CHAOSNET 6.5, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.56, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.3, MAIL-READER 6.6,
;;;  TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.4, DECNET 1.70, VISIDOC 6.5, PROFILE 6.2,
;;;  Experimental SC 2.0, Experimental CLX 6.7, CLUE 6.32,  microcode 138, Band Name: microExplorer Network (8/17)

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"

(fmakunbound 'add-event)

(defmethod add-event ((contact basic-contact) event-spec &rest actions)
  "Add EVENT-SPEC and ACTIONS to the event-translations for CONTACT."
  (let ((event-binding  (parse-event-spec (list* event-spec (copy-list actions)))))
    (with-slots (event-mask event-translations) (the contact contact)
      (let ((previous (assoc (car event-binding) event-translations :test #'equal)))
	(if previous
	    (setf (cdr previous) (cdr event-binding))
	  (push event-binding event-translations)))
      (when (realized-p contact)
	(let ((new-mask (update-event-mask event-binding event-mask)))
	  (unless (= new-mask event-mask) ;; When modified
	    (setf event-mask new-mask)
	    (setf (window-event-mask contact) event-mask))))))
  (values))

(fmakunbound 'delete-event)

(defmethod delete-event ((contact basic-contact) event-spec)
  "Remove EVENT-SPEC from the event-translations for CONTACT."
  (let ((event-binding (parse-event-spec (list event-spec))))
    (with-slots (event-mask event-translations) (the contact contact) 
      ;; Compute event mask without current event translations	
      (setf event-translations
	    (delete (car event-binding) event-translations
		    :key #'car :count 1
		    :test #'equal))
      (when (realized-p contact)	
	;; Update the contact event mask
	(let ((new-mask (compute-contact-event-mask contact))
	      (old-bit (update-event-mask event-binding 0)))
	  (when (zerop (logand new-mask old-bit)) ;; When modified
	    ;; Only modify the event-mask bit for the event being deleted
	    (setf new-mask (logandc2 event-mask old-bit)
		  event-mask new-mask
		  (window-event-mask contact) new-mask))))))
  (values))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"

(fmakunbound 'event-actions)

(defmethod event-actions ((contact basic-contact) event-spec)
  "Return the list of actions for event-spec."
  (declare (values action-list)) 

  ;; Check instance translations
  (let ((event-binding (car (parse-event-spec (list event-spec)))))

    (cdr
      (or
	;; Instance translation?
	(assoc event-binding
	       (slot-value (the contact contact) 'event-translations)
	       :test #'equal)

	;; Class translation?
	(assoc event-binding
	       (get (class-name-of contact) 'event-translations)
	       :test #'equal)

	;; Superclass class translation?
	(dolist (class (class-all-superclasses (class-name-of contact)))
	  (let ((actions (assoc event-binding
				(get class 'event-translations)
				:test #'equal)))
	    (when actions
	      (return actions))))))))

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defmethod initial-state-transition ((contact basic-contact))
  "Return the old-state/new-state for the initial (setf contact-state) after CONTACT
   is realized. Return nil if (setf contact-state) need not be called, i.e. no
   initial state transition is necessary."
  (with-slots (state) contact
    (when (eq :mapped state)
      (values :managed :mapped))))

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defmethod (setf contact-priority) (new-priority (contact contact) &optional new-sibling)
  (setf (window-priority contact new-sibling) new-priority))

(defun change-priority (contact priority &key sibling accept-p)
  "Change CONTACT's stacking order"
  (declare (type contact contact)
	   (type (member :above :below :top-if :bottom-if :opposite) priority)
	   (type (or null contact) sibling)
	   (values success-p priority sibling))
  
  ;; Refuse request, with no compromise, if unrealized
  (when (realized-p contact)
    (multiple-value-bind (success-p new-priority new-sibling)
	(if (managed-p contact)
	    (manage-priority (contact-parent contact) contact priority sibling)
	    (values t priority sibling))
      
      (when (and (or success-p accept-p)
		 (not (top-level-p contact)))	       ; Already done, if top-level
	(setf (contact-priority contact new-sibling) new-priority))
      
      (values success-p new-priority new-sibling))))
))
