;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Class precedence list cache used by event translation changed to avoid crashing on unfinalized class
;;; and to eliminate classes with no defevent's. defcontact and all base contact classes recompiled/reloaded.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for CLUE version 6.42
;;; Written 12/31/89 15:22:59 by KK,
;;; while running on RIGI from band PB28
;;; With SYSTEM 6.27, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-PATHNAME 6.2, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.7, RPC 6.2, NFS-MX 6.4, EH 6.5, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.21, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.7, METER 6.1, ZWEI 6.11, DEBUG-TOOLS 6.4, WINDOW-MX 6.9, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.26, CLEH 6.5, NETWORK-PATHNAME 6.0,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.56, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.4, MAIL-READER 6.6,
;;;  TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.70, VISIDOC 6.7, PROFILE 6.2,
;;;  Experimental CLX 6.7, CLUE 6.41, Experimental BUG 11.17, Experimental SC 2.0,
;;;  Experimental CONFORMANCE-CHECKER 2.0,  microcode 139, Band Name: microExplorer Network + SLE (11/28)

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(defun class-name-direct-superclasses (class-name)
  #+(and ti clos)
  (mapcar #'class-name (ticlos:class-direct-superclasses (find-class class-name)))
  #-(and ti clos)
  (get class-name 'direct-superclasses))

(defsetf class-name-direct-superclasses (class-name) (classes)
  ;; Should only be called by defcontact.
  #+(and ti clos)
  `(and ,class-name ,classes)
  #-(and ti clos)
  `(setf (get ,class-name 'direct-superclasses) ,classes))


(defun class-name-precedence-list (class-name)
  "Return a list of class name symbols for the class precedence list of CLASS-NAME."
  (or
    #+(and ti clos)
    (let ((class (find-class class-name)))      
      (when (ticlos:class-finalized-p class)
	(mapcar #'class-name (ticlos:class-precedence-list class))))

    #-(and ti clos)
    nil
    
    (delete-duplicates
      (apply #'append
	     (list class-name)
	     (mapcar 'class-name-precedence-list
		     (class-name-direct-superclasses class-name))))))


(proclaim '(inline class-name-event-translations))
(defun class-name-event-translations (class-name)
  (get class-name 'event-translations))

(defsetf class-name-event-translations (class-name) (translations)
  `(setf (get ,class-name 'event-translations) ,translations))


(defsetf class-name-event-precedence-list (class-name) (classes)
  `(setf (get ,class-name 'event-precedence-list) ,classes))
  
(defun class-name-event-precedence-list (class-name)
  (or
    (get class-name 'event-precedence-list)
    (setf (class-name-event-precedence-list class-name)
	  (delete nil (class-name-precedence-list class-name)
		  :key #'class-name-event-translations))))

))

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(defmacro defcontact (class superclass-names variables &rest options)
  "Defines a contact CLASS."

  (let (resources documentation constraints)
    ;; Normalize slot variables
    (setq variables
	  (mapcar #'(lambda (r) (if (consp r) r (list r))) variables))
    
    ;; Collect options
    (dolist (option options)
      (ecase (car option)
	(:constraints
	 (setf constraints (mapcar #'(lambda (r) (if (consp r) r (list r))) (cdr option))))	
	(:resources
	 (setf resources   (mapcar #'(lambda (r) (if (consp r) r (list r))) (cdr option))))	 
	(:documentation
	 (setf documentation (second option)))))

    ;; Fill in options
    (setf constraints
	  (define-constraints class superclass-names constraints))
    (setf resources
	  (define-contact-resources class superclass-names variables resources))
    (setf variables 
	  (mapcar
	    #'(lambda (var)
		;; Fill in :initform and :initarg for slot resources
		(let* ((name     (intern (symbol-name (car var)) 'keyword))		       
		       (initarg  (getf (cdr var) :initarg))
		       (type     (getf (cdr var) :type))		       
		       (resource (assoc name resources :test #'eq)))
		  
		  (when resource
		    (unless initarg		      
		      ;; Variables that show up in the resources list need :initarg's
		      (setf var `(,@var :initarg ,(or initarg name))))
		    
		    ;; Save slot :initform in resource spec, if necessary.
		    (unless (getf (cdr resource) :initform)
		      (setf (getf (cdr resource) :initform)
			    (getf (cdr var) :initform)))
		    
		    ;; Set the slot :initform to nil so that resource values from the
		    ;; database will supersede (see define-initialize-resource-slots)		    
		    (setf (getf (cdr var) :initform) nil)
		    
		    (when type
		      ;; Adjust type so that nil :initform is valid
		      (setf (getf (cdr var) :type) `(or null ,type)))))
		var)
	    variables))

    ;; Expand definition
    `(progn
       (eval-when (compile load eval)
	 ;; Update contact class properties
	 (setf
	   (class-name-event-precedence-list ',class) nil
	   (class-name-direct-superclasses ',class)   ',superclass-names 
	   (clue-resources ',class)                   ',resources)
	 ,@(when constraints
	     `(setf (clue-constraints ',class)                 ',constraints)))

       ;; Define contact class
       (defclass ,class ,superclass-names
	 ,variables
	 #+comment ;; this code when meta-classes are used
	 (:metaclass clue)
	 #+comment ;; this code when meta-classes are used
	 (:resources ,@(def-contact-resources class superclass-names variables resources))
	 ,@(when documentation `((:documentation ,documentation))))

       ;; Define initialization methods
       ,(define-clue-default-options class resources)
       ,(define-initialize-resource-slots class resources)
       ,(define-initialize-constraints class constraints)

       ;; Return class symbol
       ',class)))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defun set-event-bindings (class event-spec &rest actions)
  ;; Internal function used by DEFEVENT
  (setf (class-name-event-translations class)
	(cons (cons event-spec (copy-list actions))
	      (delete event-spec
		      (class-name-event-translations class)
		      :key #'car :test #'equal :count 1))))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defun delete-event-bindings (class event-spec)
  ;; Internal function used by UNDEFEVENT
  (setf (class-name-event-translations class)
	(delete event-spec
		(class-name-event-translations class)
		:key #'car :count 1 :test #'equal)))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defmethod event-actions ((contact basic-contact) event-spec)
  "Return the list of actions for event-spec."
  (declare (values action-list)) 

  ;; Check instance translations
  (let ((event-binding (car (parse-event-spec (list event-spec)))))

    (cdr
      (or
	;; Instance translation?
	(assoc event-binding
	       (slot-value (the basic-contact contact) 'event-translations)
	       :test #'equal)

	;; Class translation?
	(dolist (class (class-name-event-precedence-list (class-name-of contact)))
	  (let ((actions (assoc event-binding
				(class-name-event-translations class)
				:test #'equal)))
	    (when actions
	      (return actions))))))))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defun compute-contact-event-mask (contact)
  "Compute an event-mask from the event translations (class and instance) for CONTACT."
  (declare (values mask32))
  (let ((mask  0))
    
    ;; Instance translations
    (dolist (event (slot-value (the basic-contact contact) 'event-translations))
      (setq mask (update-event-mask event mask)))
    
    ;; Class translations
    (dolist (class (class-name-event-precedence-list (class-name-of contact)))
      (dolist (event (class-name-event-translations class))
	(setq mask (update-event-mask event mask))))
    
    mask))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defun translate-event (contact event)
  "Returns the actions for the first event-translation matching EVENT"
  (declare (type basic-contact contact)
	   (type event         event))
  (labels ((find-translation (event event-key translations)
	     (dolist (event-binding translations)
	       (let ((event-spec (car event-binding)))
		 (when (if (atom event-spec)
			   ;; Simple EQ test when event spec is an atom
			   (eq event-key event-spec)
			 ;; When event spec is a list, and
			 ;; the car of the list is EQ to the event, and
			 ;; the matcher function returns T
			 (and (eq event-key (car event-spec))
			      (apply (cadr event-spec) event (cddr event-spec))))
		   (return event-binding))))))

    (let ((key (slot-value (the event event) 'key)))
      (cdr
	(or
	  ;; Instance translations
	  (find-translation event key (slot-value (the basic-contact contact) 'event-translations))

	  ;; Class translations 
	  (dolist (class (class-name-event-precedence-list (class-name-of contact)))
	    (let ((translation (find-translation event key (class-name-event-translations class))))
	      (when translation (return translation)))))))))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defun describe-event-translations (contact &optional (stream *standard-output*))
  "Print the event translations for CONTACT
   If contact is a contact class name, print the event translations
   for that contact class."
  (flet ((print-event (class event stream)
	   (format stream "~%From ~20a ~s" class (car event))
	   (dolist (action (cdr event))
	     (write-char #\space stream)
	     (prin1 action stream))))
    
    (let ((translations (when (typep contact 'basic-contact)
			  (slot-value (the basic-contact contact) 'event-translations))))
      
      ;; Print instance event translations for the contact
      (dolist (event translations)
	(print-event contact event stream))
      
      ;; Print event-translations for the contact's superclasses
      (dolist (class (class-name-event-precedence-list
		       (if (symbolp contact) contact (class-name-of contact))))
	(dolist (event (class-name-event-translations class))
	  (unless (assoc (car event) translations :test #'equal)
	    (print-event class event stream)
	    (push event translations)))))))

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defcontact basic-contact (xlib:window)
  ((display            :initarg :display
		       :reader contact-display)
   
   (parent             :initarg :parent
		       :reader contact-parent)
   
   (name               :type symbol
		       :initarg :name
		       :initform :unnamed
		       :reader contact-name)
   
   (callbacks          :type list
		       :reader contact-callbacks
		       :initform nil)
   
   (event-translations :type list
		       :initform nil)
   
   (event-mask         :type (or null event-mask) ;; Converted to mask32 after realization. 
		       :initform #.(make-event-mask :exposure)
		       :accessor contact-event-mask)
   
   (state              :initform :mapped
		       :type (member :withdrawn :managed :mapped)
		       :accessor contact-state)
   
   (sensitive          :initform :on
		       :type (member :off :on)
		       :accessor contact-sensitive)
   
   (x                  :type int16
		       :initform 0
		       :reader contact-x)
   
   (y                  :type int16
		       :initform 0
		       :reader contact-y)
   
   (width              :type card16
		       :initform 0
		       :reader contact-width)
   
   (height             :type card16
		       :initform 0
		       :reader contact-height)
   
   (border-width       :type card16
		       :initform 1
		       :reader contact-border-width)
   
   ;; Class allocated slots
   (compress-motion    :initform :on :type (member :off :on)
		       :reader contact-compress-motion
		       :allocation :class)
   
   (compress-exposures :initform :off :type (member :off :on)
		       :reader contact-compress-exposures
		       :allocation :class
		       ))
  (:documentation "Basic contact using parent's window")
  (:resources
    (screen :type (or null card8))			;Selects screen when parent is a display
    ;; Slots
    name
    callbacks
    event-translations
    event-mask
    state
    sensitive
    x y width height border-width
    ))

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defcontact contact (basic-contact)
  ((background :type (or (member :none :parent-relative) pixel pixmap)
	       :initform :parent-relative :accessor contact-background)
   (depth :type card16 :initform 0 :reader contact-depth)
   (initialization :type (or (member :destroy) list))	; Internal slot for window initialization and destruction
   )
  (:documentation "Basic contact")
  (:resources
    (documentation :type (or list string))

    ;; Slots
    (background :type (or (member :none :parent-relative) pixel pixmap))
    (depth :type card16)
    
    ;; Create-window options
;;    (backing-pixel :type (or null pixel))
;;    (backing-planes :type (or null pixel))
    (backing-store :type (or null (member :not-useful :when-mapped :always)))
;;    (bit-gravity :type (or null bit-gravity))
    (border :type (or null (member :copy) pixel pixmap))
;;    (class :type (member :copy :input-output :input-only) :initform :copy)
;;    (colormap :type (or null (member :copy) colormap))
    (cursor :type (or null (member :none) cursor))
;;    (do-not-propagate-mask :type (or null device-event-mask))
;;    (gravity :type (or null win-gravity))
    (override-redirect :type (or null (member :on :off)))
    (save-under :type (or null (member :on :off)))
;;    (visual :type (or (member :copy) card29) :initform :copy)
    )
  (:documentation "A basic CLUE window which all CLUE contacts use"))

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defcontact composite (contact)
  ((children :initform nil
	     :type list
	     :reader composite-children)
   (focus    :initform nil
	     :type (or null contact)
	     :reader composite-focus)
   (shells   :type list
	     :initform nil
	     :reader composite-shells))
  (:resources
    (event-mask :type (or null event-mask) :initform #.(make-event-mask))
    (focus-name :type symbol))
  (:documentation "A basic CLUE contact with children"))

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defcontact root (composite)
  ((screen       :type     screen
		 :initarg  :screen)
   
   (pixmap-cache :type     list
		 :initform nil)
   
   (cursor-cache :type     list
		 :initform nil)
   
   (x            :initform 0)
   (y            :initform 0)
   (width        :initform 0)			; actual value filled in by initialize-instance
   (height       :initform 0)			; actual value filled in by initialize-instance
   (border-width :initform 0)
   (depth        :initform 0)
   (background   :initform :none))
  
  (:resources
    ;; Remove all inherited resources that cannot actually be changed by user
    (background            :remove t)
    (backing-store         :remove t)
    (border                :remove t)
    (border-width          :remove t)
    (depth                 :remove t)
    (documentation         :remove t)
    (focus-name            :remove t)
    (height                :remove t)
    (name                  :remove t)
    (override-redirect     :remove t)
    (save-under            :remove t)
    (sensitive             :remove t)
    (screen                :remove t)
    (state                 :remove t)
    (width                 :remove t)
    (x                     :remove t)
    (y                     :remove t)))

))

#!C
; From file SHELLS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; SHELLS.#"


(defcontact shell (composite)
  ((state        :type     (member :withdrawn :iconic :mapped)
		 :accessor contact-state)
   
   (owner        :type     composite
		 :reader   shell-owner))
  (:resources
    (state       :type     (member :withdrawn :iconic :mapped)
		 :initform (shell-default-state instance)))
  
  (:documentation
    "Base class for all shell contacts."))

))

#!C
; From file SHELLS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; SHELLS.#"


(defcontact override-shell (shell)
  ()
  (:resources
    (override-redirect :initform :on))
  (:documentation
    "Base class for shells which override the window manager."))

))

#!C
; From file SHELLS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; SHELLS.#"


(defcontact wm-shell (shell)
  ((hints                :type     (OR null wm-hints)
			 :initform nil 
			 :initarg  :hints
			 :accessor shell-hints)
   
   (normal-hints         :type     (OR null wm-size-hints)
			 :initform nil 
			 :initarg  :normal-hints
			 :accessor shell-normal-hints)
   
   (protocols            :type     (OR null list) 
			 :initform nil 
			 :initarg  :protocols
			 :accessor wm-protocols)
   
   (title                :type     (OR null stringable)
			 :initform nil 
			 :initarg  :wm-title
			 :accessor wm-title))

  (:resources
    (event-mask :initform #.(make-event-mask :structure-notify)))
    
  (:documentation
    "Base class for shells which interact with the window manager."))

))

#!C
; From file SHELLS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; SHELLS.#"


(defcontact transient-shell (wm-shell)
  ()  
  (:documentation
    "Base class for shells which are never iconified."))

))

#!C
; From file SHELLS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; SHELLS.#"


(defcontact top-level-shell (wm-shell)
  ((colormap-owners      :type     list
			 :initform nil 
			 :initarg  :colormap-owners
			 :accessor wm-colormap-owners)
   
   (icon                 :type     (or null drawable)
			 :initform nil 
			 :initarg  :icon
			 :accessor wm-icon)
   
   (icon-mask            :type     (or null pixmap)
			 :initform nil 
			 :initarg  :icon-mask
			 :accessor wm-icon-mask)
   
   (icon-title           :type     (or null stringable)
			 :initform nil 
			 :initarg  :icon-title
			 :accessor wm-icon-title)
   
   (icon-x               :type     (or null int16)
			 :initform nil 
			 :initarg  :icon-x
			 :accessor wm-icon-x)
   
   (icon-y               :type     (or null int16)
			 :initform nil 
			 :initarg  :icon-y
			 :accessor wm-icon-y))
  
  (:documentation
    "Base class for normal top-level shells."))

))

#!C
; From file SHELLS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; SHELLS.#"


(defcontact top-level-session (top-level-shell)
  ((client-host          :type     (or null stringable)
			 :initform nil 
			 :initarg  :client-host
			 :accessor sm-client-host)

   (command              :type     (or null string)
			 :initform nil 
			 :initarg  :command
			 :accessor sm-command))  
  (:documentation
    "Base class for top-level shells that communicate with a session manager."))

))

#!C
; From file STREAM.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: PATCH.CLUE; CLUE-6-7.#"


(defcontact interactive-stream
	    (contact #+(and Explorer CLOS) TICLOS:FUNDAMENTAL-CHARACTER-OUTPUT-STREAM
		     #+(and Explorer CLOS) TICLOS:FUNDAMENTAL-CHARACTER-INPUT-STREAM)
  ((gcontext
     :type     (or null gcontext)
     :initform nil
     :reader   stream-gcontext)
   
   (font
     :type     font
     :reader   stream-font
     :initform 'fixed)
   
   (cursor-x
     :initform 0
     :type     integer
     :reader   stream-cursor-x)
   
   (cursor-y
     :initform 0
     :type     integer
     :reader   stream-cursor-y)
   
   (line-height
     :initform 0
     :type     integer
     :accessor stream-line-height) ; Pixels per character line.

   (tab-width
     :initform 0
     :type     integer
     :accessor stream-tab-width)   ; Number of pixels in a tab

   (lozenge-font
     :type     font
     :accessor stream-lozenge-font
     :initform 'micro)

   (unreadp
     :type     boolean
     :initform nil)		   ; True if a character was unread (already echoed)
   
   ;; more-height:
   ;; When non-nil, then every time a new line is output, this is decremented by
   ;; LINE-HEIGHT.  When this is less than zero, MORE-PROCESSING is called.
   ;; MORE-HEIGHT gets set to the value of (RESET-MORE-HEIGHT stream) in
   ;; STREAM-READ-CHAR, STREAM-CLEAR-OUTPUT and :after REALIZE.
   (more-height
     :initform t
     :type (or boolean card16)
     :accessor stream-more-height)
   
   (rubout-handler-function
     :initform 'simple-rubout-handler	
     :accessor stream-rubout-handler-function)
   
   (output-history-top
     :initform nil)		   ; Output history line at top of window

   (output-history
     :initform nil)		   ; Circular list of strings

   (output-history-size
     :initform 100))
  
  (:resources
    (background :initform :black)
    cursor-x
    cursor-y
    font
    gcontext
    line-height
    lozenge-font
    more-height
    output-history-size
    rubout-handler-function
    tab-width))

))
