;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: defcontact: fix bug interpreting :constraints specification.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 01/03/90 16:35:56 by ARELLANO,
;;; while running on Quintillian from band LOD2
;;; With SYSTEM 6.28, VIRTUAL-MEMORY 6.3, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.7, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.22, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.11,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.4, MAIL-READER 6.6, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.26, CLEH 6.5, IP 3.57,
;;;  Experimental CLX 6.8, CLUE 6.42, X11M 6.19, Experimental BUG 11.17, Experimental SC 2.0,
;;;  Experimental CLIO 12.0, Experimental CLIO-DEMO 1.0,  microcode 429, Band Name: REL 6.0 + SLE 11/28

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(defmacro defcontact (class superclass-names variables &rest options)
  "Defines a contact CLASS."

  (let (resources documentation constraints)
    ;; Normalize slot variables
    (setq variables
	  (mapcar #'(lambda (r) (if (consp r) r (list r))) variables))
    
    ;; Collect options
    (dolist (option options)
      (ecase (car option)
	(:constraints
	 (setf constraints (mapcar #'(lambda (r) (if (consp r) r (list r))) (cdr option))))	
	(:resources
	 (setf resources   (mapcar #'(lambda (r) (if (consp r) r (list r))) (cdr option))))	 
	(:documentation
	 (setf documentation (second option)))))

    ;; Fill in options
    (setf constraints
	  (define-constraints class superclass-names constraints))
    (setf resources
	  (define-contact-resources class superclass-names variables resources))
    (setf variables 
	  (mapcar
	    #'(lambda (var)
		;; Fill in :initform and :initarg for slot resources
		(let* ((name     (intern (symbol-name (car var)) 'keyword))		       
		       (initarg  (getf (cdr var) :initarg))
		       (type     (getf (cdr var) :type))		       
		       (resource (assoc name resources :test #'eq)))
		  
		  (when resource
		    (unless initarg		      
		      ;; Variables that show up in the resources list need :initarg's
		      (setf var `(,@var :initarg ,(or initarg name))))
		    
		    ;; Save slot :initform in resource spec, if necessary.
		    (unless (getf (cdr resource) :initform)
		      (setf (getf (cdr resource) :initform)
			    (getf (cdr var) :initform)))
		    
		    ;; Set the slot :initform to nil so that resource values from the
		    ;; database will supersede (see define-initialize-resource-slots)		    
		    (setf (getf (cdr var) :initform) nil)
		    
		    (when type
		      ;; Adjust type so that nil :initform is valid
		      (setf (getf (cdr var) :type) `(or null ,type)))))
		var)
	    variables))

    ;; Expand definition
    `(progn
       (eval-when (compile load eval)
	 ;; Update contact class properties
	 (setf
	   (class-name-event-precedence-list ',class) nil
	   (class-name-direct-superclasses ',class)   ',superclass-names 
	   (clue-resources ',class)                   ',resources)
	 ,(when constraints
	     `(setf (clue-constraints ',class)        ',constraints)))

       ;; Define contact class
       (defclass ,class ,superclass-names
	 ,variables
	 #+comment ;; this code when meta-classes are used
	 (:metaclass clue)
	 #+comment ;; this code when meta-classes are used
	 (:resources ,@(def-contact-resources class superclass-names variables resources))
	 ,@(when documentation `((:documentation ,documentation))))

       ;; Define initialization methods
       ,(define-clue-default-options class resources)
       ,(define-initialize-resource-slots class resources)
       ,(define-initialize-constraints class constraints)

       ;; Return class symbol
       ',class)))

))
