;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Miscellaneous fixes to INTERACTIVE-STREAM's rubout-handler.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/12/90 12:39:38 by MARKY,
;;; while running on LIBRA from band LODB
;;; With SYSTEM 6.23, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.6, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.6, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.23, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.12,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.3, MAIL-READER 6.6, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.5, TI-CLOS 6.26, CLEH 6.5, IP 3.56,
;;;  Experimental CLX 6.8, CLUE 6.46, X11M 6.16, Experimental BUG 11.18, MMON 6.2,
;;;  Experimental SC 2.0, Experimental CLIO 15.0, Experimental CONFORMANCE-CHECKER 2.0,
;;;  Experimental CLIO-TEST 7.0, MACTOOLBOX 2.0,  microcode 429, Band Name: REL6-mmon-11/15/89

#!C
; From file STREAM.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; STREAM.#"


(defmethod stream-read-char ((self interactive-stream))
  "Read a character from the keyboard.
 Echoing is handled by the rubout handler (see rh-read-char)"
  (declare (special *rubout-handler*))
  (reset-more-height self)
  (let (char)
    (cond ((slot-value (the interactive-stream self) 'unreadp)
	   (setf (slot-value (the interactive-stream self) 'unreadp) nil)
	   (setq char (read-character (contact-display self))))
	  ((eq *rubout-handler* 'tyi)
	   (setq char (read-character (contact-display self))))
	  (*rubout-handler*
	   (setq char (rh-read-char self)))
	  (t (setq char (read-character (contact-display self)))
	     ;; Note: characters are not echoed without a rubout handler
	     ))
    #+ti
    ;; may 12/18/89 Ignore asynchronous chars unless we are inside
    ;; the rubout-handler *AND* in rh-read-char which binds *rubout-handler*
    ;; to 'tyi. This fixes the abort key inside the debugger which should return
    ;; to whence it came just like super-a. Similiar to use of eh:*reading-command*
    ;; in (:method ucl-debugger :fetch-input) .
    (when (eq *rubout-handler* 'tyi)
      (handle-asynchronous-characters char self))
    char))

(defun stream-rubout-handler (contact options function &rest args)
  ;; Rubout handling in the zetalisp tradition
  ;; may 12/18/89 Remove any options that are not value pairs such as
  ;; are passed in by eh:command-loop-read i.e., (:activation member (#\end #\return))
  ;; This fixes missing " Eval: " prompt when in debugger.
  (setq options (remove-if-not #'(lambda (pair) (= 2 (length pair))) options))

  (if (typep contact 'interactive-stream)
      (let ((option-plist nil))
	(dolist (option options)
	  (setq option-plist (append option-plist option)))
	(if args
	    (funcall (stream-rubout-handler-function contact) contact option-plist
		     #'(lambda () (apply function args)))
	  (funcall (stream-rubout-handler-function contact) contact option-plist function)))
    (apply function args)))


(DEFUN prompt-if-any (STREAM &optional prompt)
  "Print prompt on STREAM. *rhb* must be bound.
PROMPT defaults to the :reprompt or :prompt option
in *rhb* options - whichever is found first, if any."
  (SETQ prompt (or prompt
		   (getf (rhb-options *rhb*) :reprompt)
		   (getf (rhb-options *rhb*) :prompt)))
  (cond ((null prompt) nil)	;; may 12/18/89 Return nil not T
	((stringp prompt)
	 (IF stream
	     (stream-write-string stream prompt)
	     prompt)) ;; just return string if stream is nil, as format would do
	#+explorer ;; explorer error-handler hack
	;; old zetalisp required 2 arguments to prompt and read.
	((eq prompt 'sys:prompt-and-read-prompt-function)
	 (funcall prompt stream nil))
	(t
	 (funcall prompt stream))))


;; may 12/18/89 Fix, on read errors, to make rubout code in
;; simple-rubout-handler-edit (that expects the prompt to
;; be displayed on the line ...) to work.
(defun simple-rubout-handler (contact options function)
  ;; A rubout handler in the common-windows tradition
  ;; Options include:
  ;; :full-rubout flag    If the user erases all of the characters then presses
  ;;                      the rubout character once more, control is returned
  ;;			  from the input editor immediately.  Two values are
  ;;			  returned: NIL and FLAG.  In the absence of this option,
  ;;			  the input editor simply waits for more characters.
  ;;
  ;; :prompt string	  string to display or function of one argument (the contact)
  ;; :reprompt string
  ;; :initial-input string
  ;; :initial-input-pointer card16
  (let ((*rhb* (allocate-rhb)))
    (unwind-protect
	(progn
	  (setf (rhb-options *rhb*) options)
	  (setf (rhb-fill-pointer *rhb*) 0)	   ; number of characters in the buffer
	  (setf (rhb-scan-pointer *rhb*) 0)	   ; number of characters sent to application

	  ;; PROMPT option
	  (prompt-if-any contact (getf options :prompt))	;; may 12/14/89 

	  ;; INITIAL-INPUT option
	  (let ((initial-input (getf options :initial-input)))
	    (when initial-input
	      (let* ((initial-input-pointer (getf options :initial-input-pointer 0))
		     (length (- (length initial-input) initial-input-pointer))
		     (size (array-total-size (rhb-buffer *rhb*))))
		(when (> length size)
		  (setf (rhb-buffer *rhb*) (adjust-array (rhb-buffer *rhb*) (+ length size))))
		(replace (rhb-buffer *rhb*) initial-input :start1 initial-input-pointer)
		(setf (rhb-fill-pointer *rhb*) length)
		(stream-write-string contact initial-input initial-input-pointer))))

	  (when (slot-value (the interactive-stream contact) 'unreadp) ; Make sure type ahead is processed
	    (simple-rubout-handler-edit contact)		    ; by rubout handler, not just by TYI
	    (setf (rhb-scan-pointer *rhb*) 0))
	  (do ((*rubout-handler* 'read)	   ;Establish rubout handler
	       #+ti (si:rubout-handler t)  ;Needed for explorer compatability
	       )
	      (nil)
	    (catch 'rubout-handler		; Throw here when rubbing out
	      (progn
		#+cleh				; Hopefully, someday everyone will use this
		(conditions:handler-case
		    (return (funcall function))
		  (error (condition) (princ condition)))
		#+(and lispm (not cleh))
		(si:catch-error			; If a read-error occurs, print a message and loop back
		  (return			; Exit rubbout handler when read function returns
		    (funcall function)))	; Call read function
		#+(and kcl (not cleh))
		(multiple-value-bind (tag value)
		    (si:error-set `(funcall ',function))
		  (unless tag (return value)))
		#+(and excl (not cleh))
		(multiple-value-bind (tag value)
		    (excl:error-set (funcall function) :announce)
		  (unless tag (return value)))
		#-(or cleh lispm kcl excl)
		(return (funcall function))
		;; We come here after read errors (catch-error caught)
		(STREAM-fresh-line contact)		; Echo the rubout handler buffer	;; may 12/14/89 (for sn)
		;; may 12/18/89 Reprint the prompt as spaces since #\rubout expect the line to have printed the prompt
		(let ((prompt (prompt-if-any nil)))
		  (dotimes (i (length prompt)) (princ #\space)))
		(stream-write-string contact (rhb-buffer *rhb*) 0 (rhb-fill-pointer *rhb*))
		(loop (stream-read-char contact))))	; and force user to edit it
	    ;; Come here on throw to 'rubout-handler
	    ;; Maybe return when user rubs all the way back
	    (and (zerop (rhb-fill-pointer *rhb*))
		 (let ((full-rubout-option (getf options :full-rubout)))
		   (and full-rubout-option (return (values nil full-rubout-option)))))))
      (deallocate-rhb *rhb*))))


;; may 01/03/90 Added ctrl-y code. Fixed rubout on continuation line.
;; 
(defun simple-rubout-handler-edit (contact)
  ;; This is the "guts" of the rubout handler, where the editing occurs
  ;; This needs LOTS more editing commands!
  (do ((rubbed-out-some nil)
       (*rubout-handler* 'tyi)
       (ch))
      (nil) ;; forever
    (setq ch (stream-read-char contact))
    (case ch
      ((#\control-u #\control-U)				;CLEAR-INPUT flushes all buffered input
       (setf (rhb-fill-pointer *rhb*) 0)
       (setq rubbed-out-some t)			;Will need to throw out
       (stream-write-char contact ch)		;Echo and advance to new line
       (stream-write-char contact #\Newline))
      ((#\control-l #\control-L)				;Retype ALL visible area
	 (display contact))	;; may 12/14/89
      ;; may 12/14/89 Added
      #+ti
      ((#\clear-screen)						;Retype buffered input
       (stream-clear-output contact)
       (prompt-if-any contact)
       (stream-write-string contact (rhb-buffer *rhb*) 0 (rhb-fill-pointer *rhb*)))
      #+ti	;; may 01/03/90 Added ctrl-y code from tv:RH-COM-YANK-INPUT-RING & tv:RH-YANK-FROM-HISTORY
      ((#\control-y #\control-Y)		;yank from kill history
       (LET* ((COPY (zwei:string-interval (ZWEI:HISTORY-ELEMENT-SET-YANK-POINTER ZWEI:*KILL-HISTORY* 1)))
	      (STRING (tv:STRING-REMOVE-FONTS COPY))
	      (LEN (LENGTH STRING))
	      (LAST-CHAR (AREF STRING (1- LEN)))
	      (fill-pointer (rhb-fill-pointer *rhb*))
	      (END (IF ;;(AND CHOP
		     (OR (EQL LAST-CHAR #\RETURN) 
			 (EQL LAST-CHAR #\SPACE)
			 (EQL LAST-CHAR #\TAB)
			 (EQL LAST-CHAR #\)))
		     (1- LEN)
		     len)))
	 ;;(RH-INSERT-STRING STRING 0 END T NIL)
	 (stream-write-string contact string 0 end)	;; update display
	 (si:copy-array-portion string 0 end  ;; cram in rhb buffer
				(rhb-buffer *rhb*) fill-pointer (+ len fill-pointer))
	 (setf (rhb-fill-pointer *rhb*) (+ len fill-pointer))	;; update fill-pointer
	 (setf (rhb-scan-pointer *rhb*) 0)  	;; update # of chars sent to application
	 (throw 'rubout-handler t)
	 ))
      (#\Rubout
       (let ((len (rhb-fill-pointer *rhb*)))
	 (unless (zerop len)
	   (setf (rhb-fill-pointer *rhb*) (setq len (1- len)))
	   ;; may 12/14/89 Change output-history, too, for control-l
	   (LET* ((THIS-LINE (CAR (SLOT-VALUE CONTACT 'OUTPUT-HISTORY)))
		  (THIS-LINE-LEN (AND THIS-LINE (LENGTH THIS-LINE))))
	     (WHEN (PLUSP THIS-LINE-LEN)
	       (SETF (FILL-POINTER THIS-LINE) (1- THIS-LINE-LEN))))
	   (LET* ((rub-ch (aref (rhb-buffer *rhb*) len))
		  (new-y (stream-cursor-y contact))
		  (new-x (cond ((OR (char= rub-ch #\newline)
				    (AND (PLUSP (RHB-FILL-POINTER *RHB*)) 	;; Not at start of typein ..
					 (ZEROP (STREAM-CURSOR-X CONTACT))))	;; ... but at start of a wrapped line.       
				(SETQ new-y (MAX 0 (- new-y (stream-line-height contact))))
				;; return new-x at end of previous line
				(+ (text-width (stream-font contact)
					       (rhb-buffer *rhb*) 
					       :start 0 :end (rhb-fill-pointer *rhb*))
				   (text-width (stream-font contact) (prompt-if-any nil))))
			       ((graphic-char-p rub-ch)	;; may 12/14/89 
				(- (stream-cursor-x contact)
				   (char-width (stream-font contact)
					       (char-int rub-ch))))
			       ((char= rub-ch #\tab)
				(- (stream-cursor-x contact)
				   (stream-tab-width contact)))
			       (t (- (stream-cursor-x contact)
				     (lozenge-width rub-ch (stream-lozenge-font contact)))))))	;; may 12/14/89 
	     (set-cursorpos contact :x new-x :y new-y)
	     (clear-eol contact)
	     (setq rubbed-out-some t))
	   (when (zerop len) ;; when all rubbed out
	     (setf (rhb-scan-pointer *rhb*) 0)
	     (throw 'rubout-handler t)))))
      (otherwise
       (if (plusp (char-bits ch))
	   (bell (contact-display contact)) ;; unknown command
	 ;; Echo character
	 #+ti
	 (when (handle-asynchronous-characters-p ch)
	   ;; Make the reader closure re-read all input from the beginning
	   (setf (rhb-scan-pointer *rhb*) 0)
	   (stream-fresh-line contact)
	   (prompt-if-any contact (GET (rhb-options *rhb*) :reprompt))
	   (THROW 'rubout-handler t))
	 (let ((fill-pointer (rhb-fill-pointer *rhb*)))
	   (stream-write-char contact ch)
	   ;; Put character in buffer, after first ensuring its big enough
	   (when (> (setf (rhb-fill-pointer *rhb*) (1+ fill-pointer))
		    (array-total-size (rhb-buffer *rhb*)))
	     (setf (rhb-buffer *rhb*)
		   (adjust-array (rhb-buffer *rhb*) (* 2 fill-pointer))))
	   (setf (aref (rhb-buffer *rhb*) fill-pointer) ch)
	   (cond (rubbed-out-some
		  ;; Make the reader closure re-read all input from the beginning
		  (setf (rhb-scan-pointer *rhb*) 0)
		  (throw 'rubout-handler t))
		 (t
		  ;; New character at the end of the buffer, just return it.
		  (setf (rhb-scan-pointer *rhb*) (rhb-fill-pointer *rhb*))
		  (return ch)))))))))
))
