;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: change-geometry: call after-effect function returned by manage-geometry.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/16/90 13:50:20 by KK,
;;; while running on RIGI from band PB28
;;; With SYSTEM 6.29, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.2,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-PATHNAME 6.3, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.7, RPC 6.2, NFS-MX 6.4, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.23, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.13, METER 6.1, ZWEI 6.12, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.37, CLEH 6.5, NETWORK-PATHNAME 6.0,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.57, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.5, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.71, VISIDOC 6.7, PROFILE 6.2,
;;;  Experimental CLX 6.8, CLUE 6.53, Experimental BUG 11.18, Experimental SC 3.0,
;;;  Experimental CONFORMANCE-CHECKER 2.0,  microcode 139, Band Name: microExplorer Network + SLE (11/28)

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defun change-geometry (contact &key x y width height border-width accept-p)
  "Request geometry change for CONTACT. The first value returned indicates
   if the request was approved and performed. If nil, then the request was not
   approved, and the remaining values specify an acceptable compromise change. 
   If ACCEPT-P, then any compromise request is performed immediately."

  (declare (type contact          contact)
	   (type (or null int16)  x y)
	   (type (or null card16) width height border-width)
	   (type boolean          accept-p)
	   (values success-p x y width height border-width))

  (unless (destroyed-p contact)
    (let ((accept-p (or accept-p (top-level-p contact))))
      (with-slots
	((contact-x            x)
	 (contact-y            y)
	 (contact-width        width)
	 (contact-height       height)
	 (contact-border-width border-width)
	 (contact-parent       parent))
	(the basic-contact contact)

	(multiple-value-bind (success-p approved-x approved-y approved-width approved-height approved-border-width)
	    (if (and (managed-p contact) (realized-p contact-parent))
		;; Ask parent for approval.
		(manage-geometry contact-parent contact x y width height border-width)

		;; Else approve immediately.
		(values t
			(or x contact-x)
			(or y contact-y)
			(or width contact-width)
			(or height contact-height)
			(or border-width contact-border-width)))
	  
	  (when (or success-p (and accept-p approved-x))
	    ;; Perform approved change
	    (with-state (contact)
	      ;; Get after-effect function, if any.
	      (let ((after-effect (or success-p
				      (manage-geometry contact-parent contact
						       approved-x approved-y
						       approved-width approved-height
						       approved-border-width))))
		(assert after-effect ()
			"MANAGE-GEOMETRY for ~a failed to accept its own compromise geometry." contact-parent)

	        ;; Change contact geometry
		(when (or (not (= contact-x approved-x))
			  (not (= contact-y approved-y)))
		  (move contact approved-x approved-y))
		(when (or (not (= contact-width approved-width))
			  (not (= contact-height approved-height))
			  (not (= contact-border-width approved-border-width)))
		  (resize contact approved-width approved-height approved-border-width))

		;; Perform after-effect
		(when (and (functionp after-effect) (not (composite-changing-layout-p contact-parent)))
		  (funcall after-effect)))))

	  ;; Return result of geometry mgmt.
	  (values success-p approved-x approved-y approved-width approved-height approved-border-width))))))

))
