;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: change-geometry: fix error in changing unrealized shell.
;;; manage-geometry: new :around method for composite ensures immediate approval if parent or child is unrealized.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/19/90 11:32:38 by KK,
;;; while running on GOLF from band LODA
;;; With SYSTEM 6.30, VIRTUAL-MEMORY 6.3, EH 6.6, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.3, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.7, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.1,
;;;  COMPILER 6.14, TV 6.23, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.12,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.5, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.38, CLEH 6.5, IP 3.57,
;;;  Experimental CLX 6.8, CLUE 6.58, X11M 6.20, Experimental BUG 11.18, Experimental SC 3.0,
;;;  Experimental CLIO 15.0, Experimental SCIFI 1.0,  microcode 429, Band Name: REL 6.0 + SLE 11/28

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defun change-geometry (contact &key x y width height border-width accept-p)
  "Request geometry change for CONTACT. The first value returned indicates
   if the request was approved and performed. If nil, then the request was not
   approved, and the remaining values specify an acceptable compromise change. 
   If ACCEPT-P, then any compromise request is performed immediately."

  (declare (type contact          contact)
	   (type (or null int16)  x y)
	   (type (or null card16) width height border-width)
	   (type boolean          accept-p)
	   (values success-p x y width height border-width))

  (unless (destroyed-p contact)
    (with-slots
      ((contact-x            x)
       (contact-y            y)
       (contact-width        width)
       (contact-height       height)
       (contact-border-width border-width)
       (contact-parent       parent))
      (the basic-contact contact)

      (assert contact-parent () "Cannot change the geometry of a root.")
      
      (multiple-value-bind
	(success-p approved-x approved-y approved-width approved-height approved-border-width)
	  (manage-geometry contact-parent contact x y width height border-width)
	
	(when (and (or success-p			 ; Approved or...
		       (and accept-p approved-x))	 ; Compromise exists and is acceptable, and ... 
		   (or					 ; Not already done by window mgr...
		     (slot-value contact-parent 'parent) ;   (i.e. either non-top-level or unrealized.)
		     (not (realized-p contact))))	 ;   (See manage-geometry for root)
	  
	  ;; Perform approved change
	  (with-state (contact)
	    ;; Get after-effect function, if any.
	    (let ((after-effect (or success-p
				    (manage-geometry contact-parent contact
						     approved-x approved-y
						     approved-width approved-height
						     approved-border-width))))
	      (assert after-effect ()
		      "MANAGE-GEOMETRY for ~a failed to accept its own compromise geometry." contact-parent)
	      
	      ;; Change contact geometry
	      (when (or (not (= contact-x approved-x))
			(not (= contact-y approved-y)))
		(move contact approved-x approved-y))
	      (when (or (not (= contact-width approved-width))
			(not (= contact-height approved-height))
			(not (= contact-border-width approved-border-width)))
		(resize contact approved-width approved-height approved-border-width))
	      
	      ;; Perform after-effect
	      (when (and (functionp after-effect) (not (composite-changing-layout-p contact-parent)))
		(funcall after-effect contact-parent)))))
	
	;; Return result of geometry mgmt.
	(values success-p approved-x approved-y approved-width approved-height approved-border-width)))))

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defmethod manage-geometry :around ((parent composite) (contact basic-contact)
				    x y width height border-width &key)
  (declare (type contact          contact)
	   (type (or null int16)  x y)
	   (type (or null card16) width height border-width)
	   (values success-p x y width height border-width))
  (if (and (realized-p parent) (realized-p contact) (managed-p contact))
      (call-next-method)
      
      ;; Else approve immediately.
      (with-slots ((contact-x x)
		   (contact-y y)
		   (contact-width width)
		   (contact-height height)
		   (contact-border-width border-width))
		  contact
	(values t
		(or x contact-x)
		(or y contact-y)
		(or width contact-width)
		(or height contact-height)
		(or border-width contact-border-width)))))

))

#!C
; From file ROOT-GMGMT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; ROOT-GMGMT.#"


(defmethod manage-geometry ((parent root) (shell wm-shell) x y width height border-width &key) 
  (declare (type contact shell)
	   (type (or null int16) x y)
	   (type (or null card16) width height border-width)
	   (values success-p x y width height border-width))

  (with-slots ((contact-x x)
	       (contact-y y)
	       (contact-width width)
	       (contact-height height)
	       (contact-border-width border-width)) (the contact shell)
    
    
    ;; Wait for any previous :configure-notify events to be handled
    (process-all-events (contact-display parent))
    
    ;; Negotiate top-level geometry with window mgr
    (with-event-mode (shell '(:configure-notify (throw-action :configure-notify)))      
      ;; Reconfigure top-level window, if necessary
      (let (changed-p)
	(with-state (shell)
	  (when (and x (/= x contact-x)
		     (setf (drawable-x shell) x))
	    (setf changed-p t))
	  (when (and y (/= y contact-y)
		     (setf (drawable-y shell) y))
	    (setf changed-p t))
	  (when (and width (/= width contact-width)
		     (setf (drawable-width shell) width))
	    (setf changed-p t))
	  (when (and height (/= height contact-height)
		     (setf (drawable-height shell) height))
	    (setf changed-p t))
	  (when (and border-width (/= border-width contact-border-width)
		     (setf (drawable-border-width shell) border-width))
	    (setf changed-p t)))
	
	;; Return approved geometry
	(values
	  (or
	    ;; Null changed approved immediately
	    (not changed-p)
	    
	    ;; Actual change approved if it is not modified by window mgr.
	    (progn
	      ;; Wait for :configure-notify to report actual new window geometry.
	      ;; Top-level shell's handle-event will update geometry slots in
	      ;; response to :configure-notify.	      
	      (catch :configure-notify
		(loop (process-next-event (contact-display parent))))
	      
	      ;; Assert: shell slots now contain actual (wm-approved) geometry
	      ;; Return approval of original geometry request
	      (and
		(not (and x (/= x contact-x)))
		(not (and y (/= y contact-y)))
		(not (and width (/= width contact-width)))
		(not (and height (/= height contact-height)))
		(not (and border-width (/= border-width contact-border-width))))))
	  
	  contact-x contact-y contact-width contact-height contact-border-width)))))

))
