;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: change-geometry:  Perform after-effect of manage-geometry only if geometry really is changed.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/21/90 18:20:21 by KK,
;;; while running on GOLF from band LODA
;;; With SYSTEM 6.30, VIRTUAL-MEMORY 6.3, EH 6.6, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.3, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.7, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.1,
;;;  COMPILER 6.14, TV 6.24, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.12,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.5, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.38, CLEH 6.5, IP 3.57,
;;;  Experimental CLX 6.8, CLUE 6.59, X11M 6.20, Experimental BUG 11.18, Experimental SC 3.0,
;;;  Experimental CLIO 15.0,  microcode 429, Band Name: REL 6.0 + SLE 11/28

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defun change-geometry (contact &key x y width height border-width accept-p)
  "Request geometry change for CONTACT. The first value returned indicates
   if the request was approved and performed. If nil, then the request was not
   approved, and the remaining values specify an acceptable compromise change. 
   If ACCEPT-P, then any compromise request is performed immediately."

  (declare (type contact          contact)
	   (type (or null int16)  x y)
	   (type (or null card16) width height border-width)
	   (type boolean          accept-p)
	   (values success-p x y width height border-width))

  (unless (destroyed-p contact)
    (with-slots
      ((contact-x            x)
       (contact-y            y)
       (contact-width        width)
       (contact-height       height)
       (contact-border-width border-width)
       (contact-parent       parent))
      (the basic-contact contact)

      (assert contact-parent () "Cannot change the geometry of a root.")
      
      (multiple-value-bind
	(success-p approved-x approved-y approved-width approved-height approved-border-width)
	  (manage-geometry contact-parent contact x y width height border-width)
	
	(when (and (or success-p			 ; Approved or...
		       (and accept-p approved-x))	 ; Compromise exists and is acceptable, and ... 
		   (or					 ; Not already done by window mgr...
		     (slot-value contact-parent 'parent) ;   (i.e. either non-top-level or unrealized.)
		     (not (realized-p contact))))	 ;   (See manage-geometry for root)
	  
	  ;; Perform approved change
	  (with-state (contact)
	    ;; Get after-effect function, if any.
	    (let ((after-effect (or success-p
				    (manage-geometry contact-parent contact
						     approved-x approved-y
						     approved-width approved-height
						     approved-border-width))))
	      (assert
		after-effect ()
		"MANAGE-GEOMETRY for ~a failed to accept its own compromise geometry."
		contact-parent)
	      
	      ;; Change contact geometry, if necessary
	      (let
		((position-changed-p (or (not (= contact-x approved-x))
					 (not (= contact-y approved-y))))
		 (size-changed-p     (or (not (= contact-width approved-width))
					 (not (= contact-height approved-height))
					 (not (= contact-border-width approved-border-width)))))
		(when position-changed-p
		  (move contact approved-x approved-y))
		(when size-changed-p
		  (resize contact approved-width approved-height approved-border-width))
		
		;; Perform after-effect if...
		(when (and
			;; Something actually changed...
			(or size-changed-p position-changed-p)

			;; After-effect function returned...
			(functionp after-effect)

			;; Not in the middle of a batch of layout changes...
			(not (composite-changing-layout-p contact-parent)))
		  (funcall after-effect contact-parent)))))
	
	;; Return result of geometry mgmt.
	(values (when success-p t)
		approved-x approved-y
		approved-width approved-height
		approved-border-width))))))

))
