;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: defcontact: Ensure that class initialization methods are not defined if unnecessary.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for CLUE version 6.64
;;; Written 03/05/90 16:28:07 by KK,
;;; while running on RIGI from band PB28
;;; With SYSTEM 6.30, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.2,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-PATHNAME 6.3, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.8, RPC 6.2, NFS-MX 6.5, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.24, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.14, METER 6.1, ZWEI 6.12, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.39, CLEH 6.5, NETWORK-PATHNAME 6.1,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.57, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.5, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.72, VISIDOC 6.7, PROFILE 6.2,
;;;  Experimental CLX 6.8, CLUE 6.60, Experimental BUG 11.18, Experimental SC 3.0,
;;;  Experimental CONFORMANCE-CHECKER 2.0, Experimental CLIO 18.0,  microcode 139,
;;;  Band Name: microExplorer Network + SLE (11/28)

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(defun define-initialize-resource-slots (contact-class resources)
  "Define the initialize-resource-slots method for CONTACT-CLASS."
  (let (code slot)
    (when (dolist (resource resources code)
	    (when (setq slot (getf (cdr resource) :slot))
	      (push
		(set-resource-slot (car resource) resources slot)
		code)))
      
    `(defmethod initialize-resource-slots ((instance ,contact-class) resource-table app-defaults)
       
       ;; Check resource types and fill in defaults.
       ;; Assumes the :initform for all resource slots NIL (the true initform is evaluated here).       
       (let (options (parent (slot-value (the ,contact-class instance) 'parent)))
	 
	 ;; NOTE: PARENT is null when contact-class is ROOT.
	 ;; This may lose for some root resources requiring conversion.
	 ,@code
	 options)))))

))

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(defun define-initialize-constraints (composite-class constraints)
  "Define the initialize-constraints method for COMPOSITE-CLASS."
  (when constraints
    (let (code)
      (when (dolist (constraint constraints code)
	      (push
		(push-constraint (car constraint) constraints)
		code))

	`(defmethod initialize-constraints ((parent ,composite-class) initargs resource-table)
	   (let (options (app-defaults (getf initargs :defaults)))
	     
	     ;; Generate code to find constraint values, convert to representation type,
	     ;; and add to option list
	     ,@code
	     options))))))

))

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(defun define-clue-default-options (contact-class resources)
  "Defines the default-options method that defaults non-slot resources for CONTACT-CLASS."
  (let (code)
    (when (dolist (resource resources code)
	    (unless (getf (cdr resource) :slot)
	      (push
		(push-resource (car resource) resources)
		code)))
      `(defmethod default-options ((class (eql ',contact-class)) initargs)       
	 (let ((options      (copy-list initargs))
	       (app-defaults (getf initargs :defaults))
	       (parent       (getf initargs :parent)))
	   
	   ;; "Use" app-defaults to avoid compiler warning when no non-slot resources exist.
	   app-defaults
	   
	   ;; Convert parent arg, if necessary
	   (when (display-p parent)	  
	     (setf options (list* :parent
				  (setf parent (display-root parent (getf initargs :screen)))
				  options)))
	   
	   ;; Build a resource table for fast resource lookup
	   (let ((resource-table (get-contact-resource-table class parent initargs)))
	     
	     ;; Generate code to find resource values, convert to representation type,
	     ;; and add to option list
	     ,@code
	     (list* :resource-table resource-table options)))))))

))
