;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Handle :configure-notify events correctly for the root by
;;; fixing how the with-event-mode macro remembers the event actions
;;; it temporarily overrides.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 03/16/90 10:06:16 by GRENINGER,
;;; while running on Salk from band LODX
;;; With SYSTEM 6.31, VIRTUAL-MEMORY 6.3, EH 6.6, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.4, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.8, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.1,
;;;  COMPILER 6.14, TV 6.24, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.13,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.5, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.40, CLEH 6.5, IP 3.58,
;;;  Experimental CLX 6.8, CLUE 6.72, X11M 6.20, Experimental BUG 11.18, Experimental SC 3.0,
;;;  Experimental CLIO 18.0, Experimental SCIFI 1.0, Experimental CONFORMANCE-CHECKER 2.0,
;;;   microcode 431, Band Name: R6 11/28 patched to 2/12

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defmacro with-event-mode ((contact &rest translations) &body body)
  "The given event TRANSLATIONS are defined for the CONTACT only within
the dynamic extent of the BODY. The TRANSLATIONS are processed before any
other previously-defined instance or class translations for CONTACT."
  (let ((previous-actions (gensym))
	(new-translations (gensym))
	(translation      (gensym))
	(previous         (gensym)))

  `(let* ((,new-translations (list ,@translations))
	  (,previous-actions
	   ;; Save any actions from previous instance translations for these event specs
	   (let (pa)
	     (dolist (,translation ,new-translations (nreverse pa))
	       (push (rest (assoc (car (parse-event-spec ,translation))
			    (slot-value (the contact ,contact) 'event-translations)
			    :test #'equal))
		     pa)))))

     ;; Add modal translations
     (dolist (,translation ,new-translations)
       (apply #'add-event ,contact ,translation))
     

     (unwind-protect
	 (progn
	   ,@body)
       
       ;; Delete modal translations and restore any previous ones
       (dolist (,translation ,new-translations)
	 (let ((,previous (pop ,previous-actions)))
	   (if ,previous
	       (apply #'add-event ,contact (first ,translation) ,previous)
	       (delete-event ,contact (first ,translation)))))))))

))

#!C
; From file ROOT-GMGMT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: PATCH.CLUE; CLUE-6-22.#"

;;;
;;;  Just recompile this to use the corrected macro definition...
;;;

(defmethod manage-geometry ((parent root) (shell wm-shell) x y width height border-width &key) 
  (declare (type contact shell)
	   (type (or null int16) x y)
	   (type (or null card16) width height border-width)
	   (values success-p x y width height border-width))

  (with-slots ((contact-x x)
	       (contact-y y)
	       (contact-width width)
	       (contact-height height)
	       (contact-border-width border-width)) (the contact shell)
    
    
    ;; Wait for any previous :configure-notify events to be handled
    (process-all-events (contact-display parent))
    
    ;; Negotiate top-level geometry with window mgr
    (with-event-mode (shell '(:configure-notify (throw-action :configure-notify)))      
      ;; Reconfigure top-level window, if necessary
      (let (changed-p)
	(with-state (shell)
	  (when (and x (/= x contact-x)
		     (setf (drawable-x shell) x))
	    (setf changed-p t))
	  (when (and y (/= y contact-y)
		     (setf (drawable-y shell) y))
	    (setf changed-p t))
	  (when (and width (/= width contact-width)
		     (setf (drawable-width shell) width))
	    (setf changed-p t))
	  (when (and height (/= height contact-height)
		     (setf (drawable-height shell) height))
	    (setf changed-p t))
	  (when (and border-width (/= border-width contact-border-width)
		     (setf (drawable-border-width shell) border-width))
	    (setf changed-p t)))
	
	;; Return approved geometry
	(values
	  (or
	    ;; Null changed approved immediately
	    (not changed-p)
	    
	    ;; Actual change approved if it is not modified by window mgr.
	    (progn
	      ;; Wait for :configure-notify to report actual new window geometry.
	      ;; Top-level shell's handle-event will update geometry slots in
	      ;; response to :configure-notify.	      
	      (catch :configure-notify
		(loop (process-next-event (contact-display parent))))
	      
	      ;; Assert: shell slots now contain actual (wm-approved) geometry
	      ;; Return approval of original geometry request
	      (and
		(not (and x (/= x contact-x)))
		(not (and y (/= y contact-y)))
		(not (and width (/= width contact-width)))
		(not (and height (/= height contact-height)))
		(not (and border-width (/= border-width contact-border-width))))))
	  
	  contact-x contact-y contact-width contact-height contact-border-width)))))

))
