;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Use new method for computing an event mask from contact event translations. delete-event will
;;; no longer remove event mask bits set in a resource initform. Affects: add-event, delete-event,
;;; defevent, undefevent, defcontact.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 03/20/90 18:02:08 by KK,
;;; while running on GOLF from band LODA
;;; With SYSTEM 6.31, VIRTUAL-MEMORY 6.3, EH 6.6, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.4, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.8, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.1,
;;;  COMPILER 6.14, TV 6.24, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.13,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.5, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.41, CLEH 6.5, IP 3.58,
;;;  Experimental CLX 6.8, CLUE 6.81, X11M 6.20, Experimental BUG 11.18, Experimental SC 3.0,
;;;  Experimental CLIO 19.0,  microcode 429, Band Name: REL 6.0 + SLE 11/28

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(proclaim '(inline event-translations-mask))
(defun event-translations-mask (translations &optional (initial-mask 0))
  "Return the bitwise-or of the INITIAL-MASK and the event mask 
specified by the given event TRANSLATIONS."
  (reduce #'event-translation-mask translations :initial-value initial-mask))

(defun event-translation-mask (initial-mask translation)
  "Return the bitwise-or of the INITIAL-MASK and the event mask 
specified by the given event TRANSLATION."
  (logior
    initial-mask
    
    ;; Compute mask needed for this translation.
    (multiple-value-bind (event-key event-spec-args)
	(let ((event-spec  (first translation)))
	  (if (consp event-spec)
	      (values (first event-spec) (cddr event-spec))
	      event-spec))
      
      ;; Get event mask and filter function for the event keyword.
      (let* ((mask-spec   (rest (assoc event-key *event-mask-alist* :test #'eq)))
	     (mask        (first mask-spec))
	     (mask-filter (second mask-spec)))
	(cond	   
	  ((null mask) 0)				 ; Return empty mask
	  
	  ((not (and mask-filter event-spec-args))	 ; Return mask for event-key
	   mask)
	  
	  (t (apply mask-filter event-spec-args))	 ; Return mask for event-spec
	  )))))

(defsetf class-name-event-mask (class-name) (mask)
  `(setf (get ,class-name 'event-mask) ,mask))
  
(defun class-name-event-mask (class-name)
  "Return the event mask shared by all instances of CLASS-NAME."
  (or
    (get class-name 'event-mask)
    
    (setf
      (class-name-event-mask class-name)

      (labels
	((class-event-translations-mask
	   (classes mask)
	   ;; Return the bitwise-or of the initial MASK and
	   ;; the event mask specified by the class event translations of
	   ;; all the CLASSES.
	   (if classes
	       
	       (class-event-translations-mask
		 (rest classes)
		 (event-translations-mask
		   (class-name-event-translations (first classes))
		   mask))
	       
	       mask)))

	(class-event-translations-mask
	  (class-name-event-precedence-list class-name)

	  ;; Initial event mask is most specific class resource initform.	  
	  (dolist (class (class-name-precedence-list class-name) 0)
	    (let ((init-mask (getf (rest (assoc :event-mask (clue-resources class))) :initform)))
	      (when init-mask
		(return init-mask)))))))))

))

#!C
; From file DEFCONTACT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; DEFCONTACT.#"


(defmacro defcontact (class superclass-names variables &rest options)
  "Defines a contact CLASS."

  (let (resources documentation constraints)
    ;; Normalize slot variables
    (setq variables
	  (mapcar #'(lambda (r) (if (consp r) r (list r))) variables))
    
    ;; Collect options
    (dolist (option options)
      (ecase (car option)
	(:constraints
	 (setf constraints (mapcar #'(lambda (r) (if (consp r) r (list r))) (cdr option))))	
	(:resources
	 (setf resources   (mapcar #'(lambda (r) (if (consp r) r (list r))) (cdr option))))	 
	(:documentation
	 (setf documentation (second option)))))

    ;; Fill in options
    (setf constraints
	  (define-constraints class superclass-names constraints))
    (setf resources
	  (define-contact-resources class superclass-names variables resources))
    (setf variables 
	  (mapcar
	    #'(lambda (var)
		;; Fill in :initform and :initarg for slot resources
		(let* ((name     (intern (symbol-name (car var)) 'keyword))		       
		       (initarg  (getf (cdr var) :initarg))
		       (type     (getf (cdr var) :type))		       
		       (resource (assoc name resources :test #'eq)))
		  
		  (when resource
		    (unless initarg		      
		      ;; Variables that show up in the resources list need :initarg's
		      (setf var `(,@var :initarg ,(or initarg name))))
		    
		    ;; Save slot :initform in resource spec, if necessary.
		    (unless (getf (cdr resource) :initform)
		      (setf (getf (cdr resource) :initform)
			    (getf (cdr var) :initform)))
		    
		    ;; Set the slot :initform to nil so that resource values from the
		    ;; database will supersede (see define-initialize-resource-slots)		    
		    (setf (getf (cdr var) :initform) nil)
		    
		    (when type
		      ;; Adjust type so that nil :initform is valid
		      (setf (getf (cdr var) :type) `(or null ,type)))))
		var)
	    variables))

    ;; Expand definition
    `(progn
       (eval-when (compile load eval)
	 ;; Update contact class properties
	 (setf
	   (class-name-event-precedence-list ',class) nil
	   (class-name-event-mask ',class)            nil
	   (class-name-direct-superclasses ',class)   ',superclass-names 
	   (clue-resources ',class)                   ',resources)
	 ,(when constraints
	     `(setf (clue-constraints ',class)        ',constraints)))

       ;; Define contact class
       (defclass ,class ,superclass-names
	 ,variables
	 #+comment ;; this code when meta-classes are used
	 (:metaclass clue)
	 #+comment ;; this code when meta-classes are used
	 (:resources ,@(def-contact-resources class superclass-names variables resources))
	 ,@(when documentation `((:documentation ,documentation))))

       ;; Define initialization methods
       ,(define-clue-default-options class resources)
       ,(define-initialize-resource-slots class resources)
       ,(define-initialize-constraints class constraints)

       ;; Return class symbol
       ',class)))

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(proclaim '(inline contact-event-translations-mask))
(defun contact-event-translations-mask (contact)
  "Return the event mask from the event translations (class and instance) for CONTACT."
  (declare (values mask32))
  (with-slots (event-translations) (the basic-contact contact)
    (logior
      ;; Instance translations    
      (event-translations-mask event-translations)
      
      ;; Class translations
      (class-name-event-mask (class-name-of contact)))))


(defmethod realize ((contact contact))
  "Create the window for CONTACT. This function should not be called
   by application programs."
  
  (with-slots
    (    
     background border-width depth event-mask
     height initialization parent width x y
     )
    contact

    ;; Ensure the parent is realized
    (assert
      (realized-p parent) ()
      "Parent of ~s is not realized." contact)

    ;; Ensure width/height initialized
    (assert
      (and (plusp width) (plusp height)) ()
      "Width and height have not been initialized for ~s" contact)

    ;; Calculate event-mask
    (setf event-mask (contact-event-translations-mask contact))

    (let ((input-only-p (eq :input-only (getf initialization :class))))
      ;; Create the contact window
      (apply
	#'xlib:create-window
	:window           contact
	:parent           parent
	:x                x
	:y                y
	:width            width
	:height           height
	:border-width     border-width
	:event-mask       event-mask
	:background       (unless input-only-p background)
	:depth            depth
	:allow-other-keys t
	initialization)
      
      
      ;; Record depth, if inherited from parent
      (unless (or (plusp depth) input-only-p)
	(setf depth (contact-depth parent))))

    (let* ((documentation (getf initialization :documentation)))
      (when documentation
	(setf (window-documentation contact) documentation)))

;; Keep initialiation around for awhile, it's useful for debugging
;;    (setf initialization nil) ;; Give initialization list to the garbage collector
    ))
))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defmacro defevent (class event-spec &rest actions)
  "Add an event binding to the EVENT-TRANSLATIONS property of CLASS,
   where it can be shared by all instances of CLASS."
  (let ((event-parse          (parse-event-spec (list* event-spec actions)))
	(canonical-event-spec (gensym)))
    `(progn
       ;; Generate compiler warnings for missing actions
       #+ti ,@(mapcar #'(lambda (action)
			  (when (consp action) (setq action (first action)))
			  `(eval-when (compile)
			     (compiler:function-referenced
			       ',action ',(intern (format nil "DEFEVENT ~s ~s" class event-spec)))))
		      (rest event-parse))
       (let ((,canonical-event-spec ',(first event-parse)))
	
	 (setf
	   ;; Update class event translations
	   (class-name-event-translations ',class)
	   (cons
	     (cons ,canonical-event-spec ',(rest event-parse))
	     (delete ,canonical-event-spec
		     (class-name-event-translations ',class)
		     :key #'first :test #'equal :count 1))

	   ;; Flush cached class event mask
	   (class-name-event-mask ',class)
	   nil
	   )))))


(defmacro undefevent (class event-spec &rest actions)
  "Remove an event binding from the EVENT-TRANSLATIONS property of CLASS."
  (declare (ignore actions))
  (let ((event-spec (parse-event-spec (list event-spec))))
    `(setf
       ;; Update class event translations
       (class-name-event-translations ',class)
       (delete ',(first event-spec)
	       (class-name-event-translations ',class)
	       :key #'first :count 1 :test #'equal)
       
       ;; Flush cached class event mask
       (class-name-event-mask ',class)
       nil)))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defmethod add-event ((contact basic-contact) event-spec &rest actions)
  "Add EVENT-SPEC and ACTIONS to the event translations for CONTACT."
  
  ;; Compute canonical event translation.
  (let ((translation  (parse-event-spec (list* event-spec (copy-list actions)))))
    (with-slots (event-mask event-translations) (the contact contact)
      
      ;; Translation for this event spec already exists?
      (let ((previous (assoc (first translation) event-translations :test #'equal)))	
	(if previous
	    ;; Yes, modify it with the new actions.
	    (setf (rest previous) (rest translation))
	    
	    ;; No, add new translation.
	    (push translation event-translations)))
      
      ;; Update window event mask, if necessary
      (when (realized-p contact)
	(let ((new-mask (event-translation-mask event-mask translation)))
	  (unless (= new-mask event-mask) 
	    (setf (window-event-mask contact) (setf event-mask new-mask)))))))
  (values))

(defmethod delete-event ((contact basic-contact) event-spec)
  "Remove any translation for EVENT-SPEC from the event translations for CONTACT."

    ;; Compute a canonical event translation for the event spec
  (let ((translation (parse-event-spec (list event-spec))))
    (with-slots (event-mask event-translations) (the contact contact) 

      ;; Remove any matching translation.
      (setf event-translations
	    (delete (first translation) event-translations
		    :key #'first :count 1 :test #'equal))

      ;; Update window event mask, if necessary
      (when (realized-p contact)		
	(let ((old-bit  (event-translation-mask 0 translation)))
	  
	  ;; Don't change event mask if some other translation sets this bit
	  (when (zerop (logand (contact-event-translations-mask contact) old-bit))
	    
	    ;; Only modify the event-mask bit for the event being deleted
	    (setf (window-event-mask contact)
		  (setf event-mask (logandc2 event-mask old-bit))))))))
  (values))

))
