;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: 1. Add new exported function: owns-focus-p
;;; 
;;; 2. Add new refresh function (moved from CLIO)
;;; 
;;; 3. (setf contact-sensitive): Uses refresh to update display. Uses owns-focus-p to give
;;;    up focus when becoming insensitive.
;;; 
;;; 4. accept-focus-p: Don't accept if insensitive.
;;; 
;;; 5. *sensitive-events*: Allow :focus-out to be handled when insensitive.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for CLUE version 6.87
;;; Written 04/04/90 18:02:02 by KK,
;;; while running on RIGI from band PB28
;;; With SYSTEM 6.31, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.2,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-PATHNAME 6.4, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.8, RPC 6.2, NFS-MX 6.6, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.24, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.14, METER 6.1, ZWEI 6.13, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.42, CLEH 6.5, NETWORK-PATHNAME 6.1,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.58, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.5, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.72, VISIDOC 6.7, PROFILE 6.2,
;;;  Experimental CLX 6.8, CLUE 6.86, Experimental BUG 11.18, Experimental SC 3.0,
;;;  Experimental CLIO 20.0,  microcode 139, Band Name: microExplorer Network + SLE (11/28)

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defparameter *sensitive-events*
	      '(:key-press :key-release :button-press :button-release
		:motion-notify :enter-notify :leave-notify
		:focus-in)
  "These 'user' events are ignored by an insensitive contact.")

))

#!C
; From file INTRINSICS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"

(when (find-package 'clio-open)  
  (let ((clio-open-core (find-symbol "CORE" 'clio-open)))
    (when (ticl:fdefinedp `(method (setf contact-sensitive) (t ,clio-open-core)))
      (eval `(ticl:undefmethod (setf contact-sensitive) (value (self ,clio-open-core)))))))

(defmethod (setf contact-sensitive) (value (self contact))
  (declare (type (member :off :on) value))  
  (check-type value (member :off :on) ":ON or :OFF")
  (with-slots (x y width height sensitive parent display) self
    (let ((old sensitive))
      (setf sensitive value)

      ;; Redisplay when changing sensitive
      (when (and (not (eq old value)) (visible-p self))
	(refresh self)

	;; Give up focus if insensitive
	(when (and (eq :off value) (owns-focus-p self))
	  ;; Send focus to parent.
	  (set-input-focus display parent :parent)))))
  value)

(when (find-package 'clio-open) 
  (unintern 'refresh 'clio-open))

(export 'refresh 'cluei)
(export 'refresh 'clue)

(defun refresh (window &key (x 0) (y 0) width height)
  "Generate :exposure events for the given region of the WINDOW and for any descendant
   within this region. By default, WIDTH/HEIGHT is the distance from X/Y to the right/bottom
   edge of the WINDOW."
  (let ((transient-window
	  (xlib:create-window
	    :parent window :override-redirect :on
	    :x x
	    :y y
	    :width (or width (- (contact-width window) x))
	    :height (or height (- (contact-height window) y)))))
    (map-window transient-window)    
    (destroy-window transient-window)))



(export 'owns-focus-p 'cluei)
(export 'owns-focus-p 'clue)

(defmethod owns-focus-p ((contact contact))
  (with-slots (display) contact
    (eq contact (input-focus display))))

(defmethod owns-focus-p ((composite composite))
  (with-slots (display) composite
    (let ((focus (input-focus display)))
      (or (eq focus composite)
	  (ancestor-p focus composite)))))

(defmethod accept-focus-p ((contact contact))
  "Returns non-nil when CONTACT is willing to become the keyboard input focus"
  (declare (values boolean))
  (and (visible-p contact)
       (plusp (logand (contact-event-mask contact)
		      #.(make-event-mask :key-press :key-release)))
       (sensitive-p contact)))

))

