;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Recompile convert methods with CLOS 6.11 or greater in order to avoid "convert bug".

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 07/07/89 17:11:28 by KK,
;;; while running on Sierra from band LOD1
;;; With SYSTEM 6.10, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.6, TV 6.12, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.3,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.2, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, Inconsistent TI-CLOS 6.16, CLEH 6.4,
;;;  IP 3.47, Experimental BUG 11.10, Experimental CLX 6.2, CLUE 6.8, X11M 6.6, Experimental SC 1.3,
;;;  Experimental CLUSO 10.0,  microcode 429, Band Name: Release 6.0 + SLE  6/26

#!C
; From file RESOURCE.LISP#> CLUE; MR-X:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "CLUE: CLUE; RESOURCE.#"

(multiple-value-bind (major minor) (sys:get-system-version 'clos)
  (declare (ignore major) )
  (unless (>= minor 11)
    (user:load-patches 'clos :noconfirm :unreleased)))

(fmakunbound 'convert)

(defmethod convert (contact value (type t))
  "Convert VALUE to TYPE"
  (cond ((and (consp type) (eq (car type) 'or))	; OR type
	 (dolist (typ (cdr type))		; Use the first conversion that works
	   (if (eq typ 'null)
	       (when (null value) (return nil))
	     (let ((result (convert contact value typ)))
	       (when result
		 (return result))))))

	((and (consp type) (eq (car type) 'member))	; MEMBER type
	 (unless (keywordp value)
	   (setq value (convert contact value 'keyword)))
	 (and (member value (cdr type) :test #'eq) value))

	((typep value type) value)		; If type works, use it!

	((or (stringp value)			; Last resort, try read-from-string
	     (symbolp value))
	 (let ((value (string value))
	       (*read-base* 10.)
	       (eof '#.(gensym)))
	   (multiple-value-bind (result index)
	       (read-from-string value nil eof)
	     (and (= index (length value))
		  (not (eq result eof))
		  (typep result type)
		  result))))
	(t nil)))

(defmethod convert (contact value (type (eql 'keyword)))
  (declare (ignore contact))
  (typecase value
    (keyword value)
    (symbol (intern (symbol-name value) 'keyword))
    (string
     (unless (position #\space (the string value))
       (intern (string-upcase value) 'keyword)))
    (otherwise nil)))

(defmethod convert (contact value (type (eql 'pixel)))
  (typecase value
    (stringable					; a color name
     (when (symbolp value) (setq value (symbol-name value)))
     (let ((screen (contact-screen contact)))
       (cond ((equalp value "WHITE") (screen-white-pixel screen))
	     ((equalp value "BLACK") (screen-black-pixel screen))
	     (t (let* ((colormap (screen-default-colormap screen))
		       (cache (getf (screen-plist screen) :color-cache))
		       (pixel (cdr (assoc value cache :test #'equalp))))
		  (or pixel
		      (progn
			(when (setf pixel (#+lispm si:ignore-errors ;; Color may not be found
					   #-lispm progn 
					    (alloc-color colormap value)))
			  (setf (getf (screen-plist screen) :color-cache)
				(cons (cons value pixel) cache)))
			pixel)))))))
    (pixel value)
    (otherwise nil)))

(defmethod convert (contact value (type (eql 'color)))
  (typecase value
    (stringable				   ; a color name
     (lookup-color (screen-default-colormap (contact-screen contact)) value)
     ;; Don't use window-colormap, because that causes a server round-trip.
     ;; (lookup-color (window-colormap contact) value)
     )
    (color value)
    (otherwise nil)))

(defmethod convert (contact value (type (eql 'font)))
  (typecase value
    (stringable				   ; a color name
     (open-font (contact-display contact) value))
    (font value)
    (otherwise nil)))
     
(defmethod convert (contact value (type (eql 'pixmap)))
  (typecase value
    (stringable
     (let ((image (convert contact value 'image)))
       (and image (get-pixmap contact image))))
    ((or (rational 0 1)
	 (float 0.0 1.0))
     (let ((gray (svref '#(0%gray 6%gray 12%gray 25%gray 37%gray 50%gray 62%gray 75%gray 88%gray 93%gray 100%gray)
			(round (* value 10)))))
       (and gray (boundp gray) (get-pixmap contact (symbol-value gray)))))
    (image (get-pixmap contact value))
    (pixmap value)
    (otherwise nil)))

(defmethod convert (contact value (type (eql 'image)))
  (declare (ignore contact))
  (declare (special cluei::*bitmap-images*))
  (typecase value
    (stringable
     (let ((symbol (find value cluei::*bitmap-images* :test #'xlib::stringable-equal)))
       (and symbol
	    (boundp symbol)
	    (typep (symbol-value symbol) 'image)
	    (symbol-value symbol))))
    (image value)
    (otherwise nil)))

(defmethod convert (contact value (type (eql 'cursor)))
  (declare (special *cursor-names*))
  (typecase value
    (card8 (get-cursor contact value))
    (stringable
     (let ((entry (assoc (string value) *cursor-names* :test #'string-equal)))
       (cond (entry
	      (get-cursor contact (second entry)))
	     ((and (stringp value)
		   (setq entry (parse-integer value :junk-allowed t)))
	      (get-cursor contact entry)))))
    (cursor value)
    (otherwise nil)))

(defmethod convert (contact value (type (eql 'boolean)))
  (declare (ignore contact))
  (typecase value
    ((member t nil) value)
    (stringable
     (member value '(:yes :on :true t) :test #'xlib::stringable-equal))
    (otherwise nil)))

(defmethod convert (contact value (type (eql 'stringable)))
  (declare (ignore contact))
  (typecase value
    (stringable value)
    (otherwise
     (princ-to-string value))))

(defmethod convert (contact value (type (eql 'string)))
  (declare (ignore contact))
  (typecase value
    (string value)
    (symbol (symbol-name value))
    (otherwise
     (princ-to-string value))))

))
