;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: dispatch-event: Speed up exposure compression; ensure count is zero in final exposure event.
;;; delete-before-action: Make this work!

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 04/12/90 17:13:28 by KK,
;;; while running on GOLF from band LODA
;;; With SYSTEM 6.31, VIRTUAL-MEMORY 6.3, EH 6.6, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.4, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-FILE 6.10, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.1,
;;;  COMPILER 6.14, TV 6.24, DATALINK 6.0, CHAOSNET 6.6, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.13,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.4, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.3, VISIDOC 6.7, TI-CLOS 6.43, CLEH 6.5, IP 3.60,
;;;  Experimental CLX 6.10, CLUE 6.89, X11M 6.21, Experimental BUG 11.19, Experimental CLIO 20.0,
;;;  Experimental SC 3.0,  microcode 429, Band Name: REL 6.0 + SLE 11/28


#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defun add-before-action (display class action-name &rest args)
    "Call the action named ACTION-NAME with ARGUMENTS before every event
   on DISPLAY directed to a contact whose class is the same as
   or superclass of the action class."
    (setf (before-actions display)
	  (cons
	    (list* class action-name (copy-list args))
	    (delete-if #'(lambda (entry)
			   (and (eq class (first entry))
				(eq action-name (second entry))))
		       (before-actions display)
		       :count 1)))
    action-name)
  
(defun delete-before-action (display class action-name)
  "Remove a before event-handler from display"
  (setf (before-actions display)
	(delete-if #'(lambda (entry)
		       (and (eq class (first entry))
			    (eq action-name (second entry))))
		   (before-actions display)
		   :count 1))
  action-name)  

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defun dispatch-event (event event-key send-event-p sequence contact)
  ;; Called from PROCESS-NEXT-EVENT to filter events and call event handlers.
  (declare (inline sensitive-p))
  (with-slots ((event_key key)
	       (event-sequence sequence)
	       (event-send-event-p send-event-p)
	       (event-contact contact)) (the event event)
    (setf event_key event-key
	  event-send-event-p send-event-p
	  event-sequence sequence
	  event-contact contact))

  (let ((class (class-name-of contact)))
    ;;
    ;; Check for non-contact event drawables.
    ;;
    (if (or (eq class 'window) (eq class 'pixmap))
      
      ;; Not a contact!!  Let the default root contact handle the event
      (handle-event (display-root (drawable-display contact)) event)
      
      (if (destroyed-p contact)
	  
	  ;; Destroyed-contact!
	  (when (eq event-key :destroy-notify)
	    (destroy-finish contact))

	  ;; Bind event for reference within with-event forms
	  (let ((display (slot-value contact 'display))
		($event$ event))
	    (declare (special $event$))
	    
	    ;;
	    ;; Do key translation
	    ;;
	    (when (or (eq event-key :key-press)
		      (eq event-key :key-release))
	      (with-slots (keysym character code state) (the event event)
		(let ((keysym-index (default-keysym-index display code state)))
		  (setf keysym (keycode->keysym display code keysym-index)
			character (keycode->character display code state :keysym-index keysym-index)))))
	    ;;
	    ;; Call the before event handlers
	    ;;
	    (dolist (before-action (before-actions display))
	      (when (subtypep class (first before-action))
		(call-action-internal contact (rest before-action))))
	    ;;
	    ;; Handle insensitive contacts
	    ;;
	    (when (and (member event-key *sensitive-events* :test #'EQ)
		       (not (sensitive-p contact)))
	      (return-from dispatch-event nil))
	    
	    ;;
	    ;; Handle modes 
	    ;;
	    (let ((modes (display-mode-stack display)))
	      (when (and modes (not (contact-mode contact)))
		(when
		  (or (member event-key *restrict-events* :test #'eq)
		      (and (member event-key *remap-events* :test #'eq)
			   (dolist (mode modes t) ;; Search for first :spring-loaded mode
			     (when (eq (second mode) :spring-loaded)
			       (format t "~%Remapping ~s from ~s to ~s" event-key contact (first mode)) ;; *** DEBUG ***
			       (setq contact (first mode)) ;; Remap contact
			       (return nil)))))
		  ;; Call mode action on for first :exclusive or :spring-loaded mode
		  (dolist (mode modes)
		    (unless (eq (second mode) :non-exclusive)
		      (call-action-internal (first mode) (cddr mode))
		      ;; quit
		      (return-from dispatch-event nil))))))
	    
	    ;; 
	    ;; Handle event compression
	    ;;
	    (with-slots ((contact-compress-motion compress-motion)
			 (contact-compress-exposures compress-exposures))
			(the contact contact)
	      
	      (case event-key
		(:exposure			; Check for exposure compression
		 (when (and (eq contact-compress-exposures :on)
			    (plusp (slot-value event 'count)))
		   ;; Accumulate total exposed area into one event
		   (let* ((exposed-min-x (slot-value event 'x))
			  (exposed-min-y (slot-value event 'y))
			  (exposed-max-x (+ exposed-min-x (slot-value event 'width)))
			  (exposed-max-y (+ exposed-min-y (slot-value event 'height))))

		     (event-case (display :force-output-p nil :discard-p t)
		       ;; Assert: We can discard all events up to 0-count :exposure
		       ;; because the protocol says that no non-exposure events can intervene.
		       (:exposure (x y width height count)
				  (setf exposed-min-x (min x exposed-min-x) 
					exposed-min-y (min y exposed-min-y)
					exposed-max-x (max (+ x width)  exposed-max-x)
					exposed-max-y (max (+ y height) exposed-max-y)) 
				  (zerop count)))
		     
		     (setf (slot-value event 'x)      exposed-min-x
			   (slot-value event 'y)      exposed-min-y
			   (slot-value event 'width)  (- exposed-max-x exposed-min-x)
			   (slot-value event 'height) (- exposed-max-y exposed-min-y)
			   (slot-value event 'count)  0))))
		
		(:motion-notify			; Check for motion compression
		 (when (eq contact-compress-motion :on)
		   (let ((count 0))
		     
		     ;; Count consecutive :motion-notify's currently in queue
		     (event-case (display :force-output-p nil :peek-p t :timeout 0)
		       (:motion-notify (window)
				       (not (and (eq window contact) (incf count))))
		       (otherwise ()   t))
		     
		     (when (plusp count) 
		       ;; Remove all but last and quit immediately
		       (do () ((zerop (decf count)))
			 (event-case (display :timeout 0)
			   (otherwise ()   t)))
		       (return-from dispatch-event nil)))))))
	    ;;
	    ;; Handle event translations
	    ;;
	    (handle-event contact event))))))

))
