;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: manage-geometry (root wm-shell): Replace process-all-events that was causing infinite recursion on update-state.
;;; Use internal-process-event instead.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for CLUE version 6.92
;;; Written 04/17/90 19:31:49 by KK,
;;; while running on RIGI from band PB28
;;; With SYSTEM 6.31, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.3,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-PATHNAME 6.4, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.10, RPC 6.2, NFS-MX 6.6, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.24, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.17, METER 6.1, ZWEI 6.13, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.4,
;;;  MAC-PRINTER-TYPES 6.2, CLIPBOARD 6.1, TI-CLOS 6.43, CLEH 6.5, NETWORK-PATHNAME 6.1,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.6, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.60, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.72, VISIDOC 6.7, PROFILE 6.3,
;;;  Experimental CLX 6.10, CLUE 6.91, Experimental BUG 11.19, Experimental SC 3.0,
;;;  Experimental CLIO 21.0,  microcode 139, Band Name: microExplorer Network + SLE (11/28)

#!C
; From file ROOT-GMGMT.LISP#> CLUE; Hotel:
#10R CLUEI#:
(LISP:COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; ROOT-GMGMT.#"


(defmethod manage-geometry ((parent root) (shell wm-shell) x y width height border-width &key) 
  (declare (type contact          shell)
	   (type (or null int16)  x y)
	   (type (or null card16) width height border-width)
	   (values success-p x y width height border-width))
  
  (with-slots ((contact-x x)
	       (contact-y y)
	       (contact-width width)
	       (contact-height height)
	       (contact-border-width border-width)) shell
    
    
    ;; Reconfigure top-level window, if necessary
    (let (changed-p (display (contact-display parent)))
      
      ;; Wait for any previous :configure-notify events to be handled.
      (loop
	;; Flush output buffer and wait for resulting events
	(display-finish-output display)
	
	;; Any events left to process?
	(if (event-listen display)
	    
	    ;; Yes, process remaining event queue
	    (loop
	      (unless (internal-process-next-event display 0) (return)))
	    
	    ;; No
	    (return)))
      
      (with-state (shell)
	(when (and x (/= x contact-x)
		   (setf (drawable-x shell) x))
	  (setf changed-p t))
	(when (and y (/= y contact-y)
		   (setf (drawable-y shell) y))
	  (setf changed-p t))
	(when (and width (/= width contact-width)
		   (setf (drawable-width shell) width))
	  (setf changed-p t))
	(when (and height (/= height contact-height)
		   (setf (drawable-height shell) height))
	  (setf changed-p t))
	(when (and border-width (/= border-width contact-border-width)
		   (setf (drawable-border-width shell) border-width))
	  (setf changed-p t)))
      
      ;; Return approved geometry
      (values
	(or
	  ;; Null changed approved immediately
	  (not changed-p)
	  
	  ;; Actual change approved if it is not modified by window mgr.
	  (progn	    
	    ;; Wait for :configure-notify to report actual new window geometry.
	    ;; wm-shell's handle-event will update geometry slots in
	    ;; response to :configure-notify.	    
	    (with-event-mode (shell '(:configure-notify (throw-action :configure-notify)))
	      (catch :configure-notify
		(loop (process-next-event display))))
	    
	    ;; Assert: shell slots now contain actual (wm-approved) geometry
	    ;; Return approval of original geometry request
	    (and
	      (not (and x (/= x contact-x)))
	      (not (and y (/= y contact-y)))
	      (not (and width (/= width contact-width)))
	      (not (and height (/= height contact-height)))
	      (not (and border-width (/= border-width contact-border-width))))))
	
	contact-x contact-y contact-width contact-height contact-border-width))))

))
