;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: (un)defevent: Clear cached event precedence list, so it will be recomputed on demand
;;; to include new class event translations.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for CLUE version 6.98
;;; Written 05/07/90 17:24:07 by KK,
;;; while running on RIGI from band PB28
;;; With SYSTEM 6.31, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.3,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-PATHNAME 6.4, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.10, RPC 6.2, NFS-MX 6.6, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.24, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.17, METER 6.1, ZWEI 6.13, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.4,
;;;  MAC-PRINTER-TYPES 6.2, CLIPBOARD 6.1, TI-CLOS 6.44, CLEH 6.5, NETWORK-PATHNAME 6.1,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.6, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.60, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.72, VISIDOC 6.7, PROFILE 6.3,
;;;  Experimental CLX 6.10, CLUE 6.97, Experimental BUG 11.19, Experimental SC 3.0,
;;;  Experimental CLIO 23.0,  microcode 139, Band Name: microExplorer Network + SLE (11/28)

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defmacro defevent (class event-spec &rest actions)
  "Add an event binding to the EVENT-TRANSLATIONS property of CLASS,
   where it can be shared by all instances of CLASS."
  (let ((event-parse          (parse-event-spec (list* event-spec actions)))
	(canonical-event-spec (gensym)))
    `(progn
       ;; Generate compiler warnings for missing actions
       #+ti ,@(mapcar #'(lambda (action)
			  (when (consp action) (setq action (first action)))
			  `(eval-when (compile)
			     (compiler:function-referenced
			       ',action ',(intern (format nil "DEFEVENT ~s ~s" class event-spec)))))
		      (rest event-parse))
       (let ((,canonical-event-spec ',(first event-parse)))
	
	 (setf
	   ;; Update class event translations
	   (class-name-event-translations ',class)
	   (cons
	     (cons ,canonical-event-spec ',(rest event-parse))
	     (delete ,canonical-event-spec
		     (class-name-event-translations ',class)
		     :key #'first :test #'equal :count 1))

	   ;; Flush cached class event mask, event precedence list
	   (class-name-event-mask ',class)
	   nil

	   (class-name-event-precedence-list ',class)
	   nil
	   )))))

))

#!C
; From file EVENTS.LISP#> CLUE; Hotel:
#10R CLUEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defmacro undefevent (class event-spec &rest actions)
  "Remove an event binding from the EVENT-TRANSLATIONS property of CLASS."
  (declare (ignore actions))
  (let ((event-spec (parse-event-spec (list event-spec))))
    `(setf
       ;; Update class event translations
       (class-name-event-translations ',class)
       (delete ',(first event-spec)
	       (class-name-event-translations ',class)
	       :key #'first :count 1 :test #'equal)
       
       ;; Flush cached class event mask, event precedence list
       (class-name-event-mask ',class)
       nil

       (class-name-event-precedence-list ',class)
       nil
       )))

))
