;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Change default-gcontext so the cache is kept on the root rather than the display.
;;; Also, include the latest improvements to find-gcontext in a patch.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/27/90 12:04:52 by pf,
;;; while running on Islington-Terrace from band LOD2
;;; With SYSTEM 6.41, VIRTUAL-MEMORY 6.3, EH 6.8, MAKE-SYSTEM 6.5, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.5, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.8, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.4, DISK-LABEL 6.1, BASIC-FILE 6.14, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.2,
;;;  COMPILER 6.18, TV 6.32, DATALINK 6.0, CHAOSNET 6.8, GC 6.4, MEMORY-AUX 6.0, NVRAM 6.4,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.6, UCL 6.1, INPUT-EDITOR 6.1, METER 6.2, ZWEI 6.27,
;;;  DEBUG-TOOLS 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.3, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.8, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.9, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.6, PROFILE 6.3, VISIDOC 6.7, TI-CLOS 6.53, CLEH 6.5, IP 3.65,
;;;  Experimental BUG 11.19, Experimental CLX 7.1, Experimental CLUE 7.9, X11M 6.32,
;;;  RPC 5.9, NFS 5.25, Experimental CONFORMANCE-CHECKER 2.0, Experimental CLIO 31.0,
;;;  Experimental SC 4.0,  microcode 483, Band Name: 6.1B+SLE(6-18)+NFS,custom,6-19

#!C
; From file INTRINSICS.LISP#> CLUE; MR-X:
#10R CLUEI#:
(lisp:COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


;;  Not in the source, this is here because contact-root is changing from
;;  a normal function to a generic function.
(fmakunbound 'contact-root)

(defmethod contact-root ((contact contact))
  ;; Return the root contact associated with CONTACT
  (declare (values root))
  (do* ((root contact parent)
	(parent (contact-parent contact) (contact-parent root)))
       ((null parent) root)))

))

#!C
; From file CACHES.LISP#> CLUE; MR-X:
#10R CLUEI#:
(lisp:COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; CACHES.#"


(defmethod contact-root ((drawable drawable))
  (or (getf (drawable-plist drawable) 'root)
      (setf (getf (drawable-plist drawable) 'root)
	    (drawable-root drawable))))

))

#!C
; From file CACHES.LISP#> CLUE; MR-X:
#10R CLUEI#:
(lisp:COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; CACHES.#"


(let ((gcontext (xlib::make-gcontext)))

  (defun default-gcontext (drawable default)
    "Return a gcontext with the same attributes as DEFAULT which can be used
to search the gcontext cache for the DRAWABLE. The second return value 
is the cache to be searched."
    (declare
      (type drawable           drawable)
      (type (or null gcontext) default)
      (values gcontext list))
    (let*
      ((display  (drawable-display drawable))
       (root	 (contact-root	   drawable))
       (depth	 (contact-depth	   drawable))

       ;; Look up root gcontext cache for this depth.
       (cache    (rest
		   (or
		     (assoc depth (getf (drawable-plist root) 'gcontexts))
		     
		     ;; Initialize cache for this depth.
		     (let ((entry (list 
				    depth
				    
				    ;; Also save default gcontext state vector here.
				    ;; Actually create/delete an unnecessary default
				    ;; gcontext to avoid knowledge of state vector
				    ;; representation.
				    (let ((default (create-gcontext :drawable drawable)))
				      (prog1
					(copy-seq (xlib::gcontext-local-state default))
					(free-gcontext default))))))
		       (push entry (getf (drawable-plist root) 'gcontexts))
		       entry)))))

      ;; Return a gcontext (with given default state) and the cache to search.
      (setf (gcontext-display gcontext) display)
      (replace
	(xlib::gcontext-local-state gcontext)
	(if default (xlib::gcontext-local-state default) (first cache)))
      
      (values gcontext cache))))

))

#!C
; From file CACHES.LISP#> CLUE; MR-X:
#10R CLUEI#:
(lisp:COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; CACHES.#"


(defconstant *gcontext-test-sequence*
	     (mapcar
	       #'(lambda (key) (position key xlib::*gcontext-components*))
	       '(
		 :foreground
		 :fill-style
		 :background
		 :stipple
		 :function
		 :font
		 :clip-mask
		 :subwindow-mode
		 :line-width
		 :ts-y
		 :ts-x
		 :tile
		 :line-style
		 :arc-mode
		 :plane-mask
		 :exposures
		 :join-style
		 :cap-style
		 :dashes
		 :dash-offset
		 :clip-y
		 :clip-x				 
		 :fill-rule
		 ))
  "A permutation of gcontext state indexes, in order of `most-likely-to-differ-first'. 
Use this order to test if two gcontext states are equal, in order to detect inequality
quickly.")

(defun find-gcontext (drawable match cache &rest initargs)
  "Return a gcontext for the DRAWABLE that matches the MATCH gcontext.
Return an existing element of the gcontext CACHE or create and 
return a new CACHE element. The INITARGS, if given, specify the
attributes used to create a matching gcontext."
  (let ((desired (xlib::gcontext-local-state match)))
    (declare (type xlib::gcontext-state desired))
    (declare (optimize speed (safety 0)))
    
    (or
      ;; Look up matching gcontext in cache.
      (do*
	((prev     cache         (rest prev))
	 (gcontext (second prev) (second prev)))
	
	((or
	   ;; End of cache?
	   (not gcontext)
 
	   (and
	     ;; Next gcontext matches?...
	     (let ((test-state (xlib::gcontext-local-state gcontext)))
	       (declare (type xlib::gcontext-state test-state))
	       (dolist (i *gcontext-test-sequence* t)
		 (unless (equalp  (svref test-state i) (svref desired i))
		   (return nil))))

	     ;; ...and matching gcontext is at head of cache?
	     (or (eq prev cache)

		 ;; No, promote to head of cache.
		 (progn
		   (setf (cdr prev) (cddr prev))
		   (push gcontext (rest cache))))))

	 ;; Return matching gcontext (or nil).
	 gcontext))
    
      
      
      ;; Create new gcontext.
      (let ((gcontext (apply #'create-gcontext :drawable drawable initargs)))
	
	;; If initargs are given, then they specify the desired attributes
	;; for the new gcontext. Otherwise, update local state in anticipation
	;; of next force-gcontext-changes.
	(unless initargs
	  (let ((local-state (xlib::gcontext-local-state gcontext)))
	    (declare (type xlib::gcontext-state local-state))
	    (replace local-state desired)
	    (setf (xlib::gcontext-internal-timestamp local-state) -1)))
	
	;; Add new gcontext to head of cache.
	(push gcontext (rest cache))
	gcontext))))

))
