;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: with-event-mode: use original (not canonical) form when restoring previous translations.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/28/90 10:41:13 by DELMAR,
;;; while running on Uniform from band LOD1
;;; With SYSTEM 6.41, VIRTUAL-MEMORY 6.3, EH 6.8, MAKE-SYSTEM 6.5, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.5, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.8, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.4, DISK-LABEL 6.1, BASIC-FILE 6.14, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.2,
;;;  COMPILER 6.18, TV 6.32, DATALINK 6.0, CHAOSNET 6.8, GC 6.7, MEMORY-AUX 6.0, NVRAM 6.4,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.6, UCL 6.1, INPUT-EDITOR 6.1, METER 6.2, ZWEI 6.27,
;;;  DEBUG-TOOLS 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.3, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.8, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.9, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.7, PROFILE 6.3, VISIDOC 6.7, TI-CLOS 6.53, CLEH 6.5, IP 3.65,
;;;  Experimental BUG 11.19, Experimental CLX 7.0, Experimental CLUE 7.13, X11M 6.33,
;;;  PICTURES 5.0,  microcode 483, Band Name: Rel 6.1 *b +SLE 6/18

#!C
; From file EVENTS.LISP#> CLUE; MR-X:
#10R CLUEI#:
(lisp:COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; EVENTS.#"


(defmacro with-event-mode ((contact &rest translations) &body body)
  "The given event TRANSLATIONS are defined for the CONTACT only within
the dynamic extent of the BODY. The TRANSLATIONS are processed before any
other previously-defined instance or class translations for CONTACT."
  (let ((previous-translations (gensym))
	(new-translations (gensym))
	(translation      (gensym))
	(previous         (gensym))
	(slot             (gensym)))

  `(let* ((,new-translations (list ,@translations))
	  (,slot             (slot-value (the contact ,contact) 'event-translations))
	  (,previous-translations
	   ;; Save any actions from previous instance translations for these event specs
	   (let (,previous-translations)
	     (dolist (,translation ,new-translations (nreverse ,previous-translations))
	       (when (assoc (first (parse-event-translation (first ,translation) (rest ,translation)))
			    ,slot
			    :test #'equal)
		 (push ,translation ,previous-translations))))))

     (unwind-protect
	 (progn
	   ;; Add modal translations
	   (dolist (,translation ,new-translations)
	     (apply #'add-event ,contact ,translation))
	   
	   ,@body)
       
       ;; Delete modal translations and restore any previous ones
       (dolist (,translation ,new-translations)
	 (let ((,previous (pop ,previous-translations)))
	   (if ,previous
	       (apply #'add-event ,contact ,previous)
	       (delete-event ,contact (first ,translation)))))))))



(defmethod manage-geometry ((parent root) (shell wm-shell) x y width height border-width &key) 
  (declare (type contact          shell)
	   (type (or null int16)  x y)
	   (type (or null card16) width height border-width)
	   (values success-p x y width height border-width))
  
  (with-slots ((contact-x x)
	       (contact-y y)
	       (contact-width width)
	       (contact-height height)
	       (contact-border-width border-width)) shell
    
    
    ;; Reconfigure top-level window, if necessary
    (let (changed-p (display (contact-display parent)))
      
      ;; Wait for any previous :configure-notify events to be handled.
      ;; Don't update-state to avoid infinite recursion during realization.
      (process-all-events display nil)
      
      (with-state (shell)
	(when (and x (/= x contact-x)
		   (setf (drawable-x shell) x))
	  (setf changed-p t))
	(when (and y (/= y contact-y)
		   (setf (drawable-y shell) y))
	  (setf changed-p t))
	(when (and width (/= width contact-width)
		   (setf (drawable-width shell) width))
	  (setf changed-p t))
	(when (and height (/= height contact-height)
		   (setf (drawable-height shell) height))
	  (setf changed-p t))
	(when (and border-width (/= border-width contact-border-width)
		   (setf (drawable-border-width shell) border-width))
	  (setf changed-p t)))
      
      ;; Return approved geometry
      (values
	(or
	  ;; Null changed approved immediately
	  (not changed-p)
	  
	  ;; Actual change approved if it is not modified by window mgr.
	  (progn	    
	    ;; Wait for :configure-notify to report actual new window geometry.
	    ;; wm-shell's handle-event will update geometry slots in
	    ;; response to :configure-notify.	    
	    (with-event-mode (shell '(:configure-notify (throw-action :configure-notify)))
	      (catch :configure-notify
		;; Don't update-state to avoid infinite recursion during realization.
		(loop (process-next-event display nil nil))))
	    
	    ;; Assert: shell slots now contain actual (wm-approved) geometry
	    ;; Return approval of original geometry request
	    (and
	      (not (and x (/= x contact-x)))
	      (not (and y (/= y contact-y)))
	      (not (and width (/= width contact-width)))
	      (not (and height (/= height contact-height)))
	      (not (and border-width (/= border-width contact-border-width))))))
	
	contact-x contact-y contact-width contact-height contact-border-width))))

(defmethod (setf contact-state) (new-state (shell wm-shell))
  (check-type new-state (member :withdrawn :iconic :mapped))    
  
  (with-slots (parent display state) shell
    (unless (eq state new-state)
      (let ((old-state state))
	
	(setf state new-state)
	
	(if
	  (realized-p shell)
	  
	  ;; Change state now -- but don't send side-effect requests if inside
	  ;; without-requests wrapper (i.e. eq *contact-notified*) ---
	  ;; that is, if responding to notification of state change from window mgr.
	  (when
	    ;; Was a (un)map request actually sent?
	    (case new-state
	      (:mapped    (shell-mapped shell)
			  (unless (eq *contact-notified* shell)
			    (map-window shell)
			    t))				 ; Request sent
	      
	      (:iconic    (if (eq old-state :withdrawn)
			      
			      (unless (eq *contact-notified* shell)
				(unless (eq :iconic (wm-initial-state shell))
				  (setf (wm-initial-state shell) :iconic))
				(map-window shell)
				t)			 ; Request sent
			      
			      (progn
				(unless (eq *contact-notified* shell)
				  (send-event parent
					      :client-message
					      #.(make-event-mask)
					      :event-window parent
					      :type :wm_change_state
					      :format 32
					      :data '(3) ; Crock: this should be an xlib defconstant
					      ))	
				(shell-unmapped shell)
				nil)))			 ; No request sent
	      
	      (:withdrawn (prog1
			    (unless (eq *contact-notified* shell)
			      (unmap-window shell)
			      (send-event parent
					  :unmap-notify
					  #.(make-event-mask :substructure-redirect :substructure-notify)
					  :event-window parent
					  :window shell
					  :configure-p nil)
			      t)			 ; Request sent
			    (shell-unmapped shell))))
	    
	    ;; Wait until resulting :(un)map-notify event has been received.
	    (let ((*ignore-map-notify* t))
	      (declare (special *ignore-map-notify*))
	      (with-event-mode (shell '(:map-notify   (throw-action :map-notify))
				      '(:unmap-notify (throw-action :map-notify)))
		(catch :map-notify
		  ;; Don't update-state to avoid infinite recursion during realization.
		  (loop (process-next-event display nil nil))))))
	  
	  ;; Not realized, let UPDATE-STATE do the work
	  (setf (display-update-flag display) t)))))
  new-state)

))
