;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: shell: Change state :initform to compute default value properly.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/19/90 13:52:11 by KK,
;;; while running on RIGI from band P614
;;; With SYSTEM 6.41, GC 6.4, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.4,
;;;  DISK-IO 6.4, DISK-LABEL 6.1, BASIC-PATHNAME 6.5, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.8, BASIC-FILE 6.13, RPC 6.2, NFS-MX 6.9, EH 6.8, MAKE-SYSTEM 6.5,
;;;  MEMORY-AUX 6.0, COMPILER 6.18, TV 6.30, NVRAM 6.4, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.25, METER 6.2, ZWEI 6.22, DEBUG-TOOLS 6.5, WINDOW-MX 6.12, PRINTER 6.8,
;;;  MAC-PRINTER-TYPES 6.2, CLIPBOARD 6.1, TI-CLOS 6.53, CLEH 6.5, NETWORK-PATHNAME 6.2,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.8, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.3,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.6, IP 3.65, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.8,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.6, DECNET 1.72, VISIDOC 6.7, PROFILE 6.3,
;;;  Experimental BUG 11.19, Experimental CLX 7.1, Experimental CLUE 7.1, Experimental CLIO 1.0,
;;;  Experimental SC 3.0,  microcode 191, Band Name: Rel 6.2 internal

#!C
; From file SHELLS.LISP#> CLUE; MR-X:
#10R CLUEI#:
(lisp:COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; SHELLS.#"


(defun shell-default-state ()
  (declare (special *new-shell*))		       ; Bound to shell by initialize-instance :around
  
  ;; WARNING: We assume this function is called during (initialize-instance :after (basic-contact)).
  ;; Therefore, the parent slot still contains the owner and has not yet been reset to the root!
  
  ;; Is this a root shell?
  (if (eq (contact-root *new-shell*) (contact-parent *new-shell*))
      ;; Yes...
      :mapped
      
      ;; No...
      :withdrawn))

(defcontact shell (composite)
  ((state        :type     (member :withdrawn :iconic :mapped)
		 :accessor contact-state)
   
   (owner        :type     composite
		 :reader   shell-owner))
  (:resources
    (state       :type     (member :withdrawn :iconic :mapped)
		 :initform (shell-default-state)))
  
  (:documentation
    "Base class for all shell contacts."))



(defmethod initialize-instance :around ((shell shell) &rest initargs)
  (declare (ignore initargs))
  ;; Must bind new shell in order to evaluate initform for state slot.
  (let ((*new-shell* shell))
    (declare (special *new-shell*))
    (call-next-method)))
    
(defmethod initialize-instance :after ((shell shell) &rest initargs)
  (declare (ignore initargs))
  (with-slots (background owner) shell
    ;; Use initial background, if not :parent-relative.
    ;; Else look for background inherited from owner.
    (setf background
	  (do*
	    ((shell     shell ancestor)
	     (ancestor  owner (if (typep ancestor 'shell)
				  (shell-owner ancestor)
				  (contact-parent ancestor)))
	     bg)
	    ((not (and ancestor (eq (setf bg (contact-background shell))
				    :parent-relative)))
	     bg)))))

(defcontact override-shell (shell)
  ()
  (:resources
    (override-redirect :initform :on))
  (:documentation
    "Base class for shells which override the window manager."))

(defcontact wm-shell (shell)
  ((hints                :type     (or null wm-hints)
			 :initform nil 
			 :initarg  :wm-hints
			 :accessor shell-hints)
   
   (normal-hints         :type     (or null wm-size-hints)
			 :initform nil 
			 :initarg  :wm-normal-hints
			 :accessor shell-normal-hints)
   
   (protocols-used       :type     (or null list) 
			 :initform nil 
			 :initarg  :wm-protocols-used
			 :accessor wm-protocols-used)
   
   (title                :type     (or null string)
			 :initform nil 
			 :initarg  :wm-title
			 :accessor wm-title))

  (:resources
    (event-mask       :initform #.(make-event-mask :structure-notify))
    
    (wm-base-height   :type (or null card16))
    (wm-base-width    :type (or null card16))             
    (wm-delta-height  :type (or null card16))                 
    (wm-delta-width   :type (or null card16))              
    (wm-gravity       :type (or null (member :north-west :north  :north-east
					      :west       :center :east
					      :south-west :south  :south-east)))          
    (wm-initial-state :type (or null (member :normal :iconic)))
    (wm-max-aspect    :type (or null number))             
    (wm-max-height    :type (or null card16))             
    (wm-max-width     :type (or null card16))            
    (wm-min-aspect    :type (or null number))             
    (wm-min-height    :type (or null card16))             
    (wm-min-width     :type (or null card16))            
     wm-title
    (wm-user-specified-position-p :type boolean)
    (wm-user-specified-size-p     :type boolean))
    
  (:documentation
    "Base class for shells which interact with the window manager."))


(defcontact transient-shell (wm-shell)
  ()  
  (:documentation
    "Base class for shells which are never iconified."))


(defcontact top-level-shell (wm-shell)
  ((colormap-owners      :type     list
			 :initform nil 
			 :accessor wm-colormap-owners)
   
   (icon-title           :type     (or null string)
			 :initform nil 
			 :initarg  :wm-icon-title
			 :accessor wm-icon-title))
  
  (:resources
    (wm-icon        :type (or null drawable))           
    (wm-icon-mask   :type (or null pixmap))                
     wm-icon-title
    (wm-icon-x      :type (or null int16))             
    (wm-icon-y      :type (or null int16)))

  (:documentation
    "Base class for normal top-level shells."))


(defcontact top-level-session (top-level-shell)
  ((client-host          :type     (or null string)
			 :initform nil 
			 :initarg  :sm-client-host
			 :accessor sm-client-host)

   (command              :type     (or null string)
			 :initform nil 
			 :initarg  :sm-command
			 :accessor sm-command))

  (:resources
     sm-command)
  
  (:documentation
    "Base class for top-level shells that communicate with a session manager."))



))
