;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: change-geometry: ensure both move and resize are called if necessary.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/22/90 10:30:23 by KK,
;;; while running on RIGI from band P614
;;; With SYSTEM 6.41, GC 6.4, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.4,
;;;  DISK-IO 6.4, DISK-LABEL 6.1, BASIC-PATHNAME 6.5, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.8, BASIC-FILE 6.13, RPC 6.2, NFS-MX 6.9, EH 6.8, MAKE-SYSTEM 6.5,
;;;  MEMORY-AUX 6.0, COMPILER 6.18, TV 6.31, NVRAM 6.4, UCL 6.1, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.25, METER 6.2, ZWEI 6.24, DEBUG-TOOLS 6.5, WINDOW-MX 6.12, PRINTER 6.8,
;;;  MAC-PRINTER-TYPES 6.2, CLIPBOARD 6.1, TI-CLOS 6.53, CLEH 6.5, NETWORK-PATHNAME 6.2,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.8, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.3,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.6, IP 3.65, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.2, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.8,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.6, DECNET 1.72, VISIDOC 6.7, PROFILE 6.3,
;;;  Experimental BUG 11.19, Experimental CLX 7.1, Experimental CLUE 7.3, Experimental SC 4.0,
;;;  Experimental CLIO 30.0,  microcode 191, Band Name: Rel 6.2 internal

#!C
; From file INTRINSICS.LISP#> CLUE; MR-X:
#10R CLUEI#:
(lisp:COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "CLUEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLUE; INTRINSICS.#"


(defun change-geometry (contact &key x y width height border-width accept-p)
  "Request geometry change for CONTACT. The first value returned indicates
   if the request was approved and performed. If nil, then the request was not
   approved, and the remaining values specify an acceptable compromise change. 
   If ACCEPT-P, then any compromise request is performed immediately."
  
  (declare (type contact          contact)
	   (type (or null int16)  x y)
	   (type (or null card16) width height border-width)
	   (type boolean          accept-p)
	   (values success-p x y width height border-width))
  
  (unless (destroyed-p contact)
    (with-slots
      ((contact-x            x)
       (contact-y            y)
       (contact-width        width)
       (contact-height       height)
       (contact-border-width border-width)
       (contact-parent       parent))
      (the basic-contact contact)
      
      (assert contact-parent () "Cannot change the geometry of a root.")
      
      (multiple-value-bind
	(success-p approved-x approved-y approved-width approved-height approved-border-width)
	  (manage-geometry contact-parent contact x y width height border-width)
	
	(when (and
		(or success-p			; Approved or...
		    (and accept-p approved-x))	; Compromise exists and is acceptable, and ... 
		
		(or				; Not already done by window mgr...
		  (slot-value contact-parent 'parent)	;   (i.e. either non-top-level or unrealized.)
		  (not (realized-p contact))))	;   (See manage-geometry for root)
	  
	  ;; Perform approved change 
	  (let
	    ;; Get after-effect function, if any.
	    ((after-effect (or success-p
			       (manage-geometry contact-parent contact
						approved-x approved-y
						approved-width approved-height
						approved-border-width))))
	    (assert
	      after-effect ()
	      "MANAGE-GEOMETRY for ~a failed to accept its own compromise geometry."
	      contact-parent)
	    
	    ;; Change contact geometry.
	    (when
	      ;; Perform after-effect if...
	      (and
		;; Something actually changed...
		(let ((moved-p (move contact approved-x approved-y))
		      (sized-p (resize contact approved-width approved-height approved-border-width)))
		  (or moved-p sized-p)) 
		
		;; and after-effect function returned...
		(functionp after-effect)
		
		;; and not in the middle of a batch of layout changes...
		(not (composite-changing-layout-p contact-parent)))
	      
	      (funcall after-effect contact-parent))))
	
	;; Return result of geometry mgmt.
	(values (when success-p t)
		approved-x approved-y
		approved-width approved-height
		approved-border-width)))))

))
