;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: encode-wm-size-hints: Set user-specified-size/position-p even if size-hints does not include x/y, width/height.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for CLX version 6.11
;;; Written 05/07/90 16:14:35 by KK,
;;; while running on RIGI from band PB28
;;; With SYSTEM 6.31, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.3,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-PATHNAME 6.4, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.10, RPC 6.2, NFS-MX 6.6, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.24, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.17, METER 6.1, ZWEI 6.13, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.4,
;;;  MAC-PRINTER-TYPES 6.2, CLIPBOARD 6.1, TI-CLOS 6.44, CLEH 6.5, NETWORK-PATHNAME 6.1,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.6, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.60, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.72, VISIDOC 6.7, PROFILE 6.3,
;;;  Experimental CLX 6.10, CLUE 6.97, Experimental BUG 11.19, Experimental SC 3.0,
;;;  Experimental CLIO 23.0,  microcode 139, Band Name: microExplorer Network + SLE (11/28)

#!C
; From file MANAGER.LISP#> CLX; Hotel:
#10R XLIB#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "XLIB"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: CLX; MANAGER.#"


(defun encode-wm-size-hints (hints)
  (declare (type wm-size-hints hints))
  (declare-values simple-vector)
  (let ((usposition #b1)			;User Specified position
	(ussize     #b10)			;User Specified size
	(pposition  #b100)			;Program specified position
	(psize      #b1000)			;Program specified size
	(pminsize   #b10000)			;Program specified minimum size
	(pmaxsize   #b100000)			;Program specified maximum size
	(presizeinc #b1000000)			;Program specified resize increments
	(paspect    #b10000000)			;Program specfied min and max aspect ratios
	)
    (let ((vector (make-array 15 :initial-element 0))
	  (flags 0))
      (declare (type (simple-vector 15) vector)
	       (type card16 flags))
      (setf flags (if (wm-size-hints-user-specified-position-p hints) usposition pposition))
      (when (and (wm-size-hints-x hints) (wm-size-hints-y hints)) 
	(setf (aref vector 1) (wm-size-hints-x hints)
	      (aref vector 2) (wm-size-hints-y hints)))
      (setf flags (logior flags (if (wm-size-hints-user-specified-position-p hints) ussize psize)))
      (when (and (wm-size-hints-width hints) (wm-size-hints-height hints))
	(setf (aref vector 3) (wm-size-hints-width hints)
	      (aref vector 4) (wm-size-hints-height hints)))
      
      (when (and (wm-size-hints-min-width hints) (wm-size-hints-min-height hints))
	(setf flags (logior flags pminsize)
	      (aref vector 5) (wm-size-hints-min-width hints)
	      (aref vector 6) (wm-size-hints-min-height hints)))
      (when (and (wm-size-hints-max-width hints) (wm-size-hints-max-height hints))
	(setf flags (logior flags pmaxsize)
	      (aref vector 7) (wm-size-hints-max-width hints)
	      (aref vector 8) (wm-size-hints-max-height hints)))
      (when (and (wm-size-hints-width-inc hints) (wm-size-hints-height-inc hints))
	(setf flags (logior flags presizeinc)
	      (aref vector 9) (wm-size-hints-width-inc hints)
	      (aref vector 10) (wm-size-hints-height-inc hints)))
      (let ((min-aspect (wm-size-hints-min-aspect hints))
	    (max-aspect (wm-size-hints-max-aspect hints)))
	(when (and min-aspect max-aspect)
	  (setf flags (logior flags paspect)
		min-aspect (rationalize min-aspect)
		max-aspect (rationalize max-aspect)
		(aref vector 11) (numerator min-aspect)
		(aref vector 12) (denominator min-aspect)
		(aref vector 13) (numerator max-aspect)
		(aref vector 14) (denominator max-aspect))))
      (setf (aref vector 0) flags)
      vector)))

))
