;;; -*- Mode: Common-Lisp; Package: SYSTEM; Base: 10.; Patch-File: T -*-
;;; Written 04/10/90 07:32:38 by BERGER,
;;; Reason: Modified Add-partition to probe directory MICRONET before attempting to add partition. [9903]
;;; while running on MX29 from band n817
;;; With SYSTEM 6.31, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.2,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-PATHNAME 6.4, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.10, RPC 6.2, NFS-MX 6.6, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.24, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.17, METER 6.1, ZWEI 6.13, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.4,
;;;  MAC-PRINTER-TYPES 6.2, CLIPBOARD 6.1, TI-CLOS 6.43, CLEH 6.5, NETWORK-PATHNAME 6.1,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.6, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.60, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.72, VISIDOC 6.7, PROFILE 6.2,
;;;   microcode 138, Band Name: microExplorer Network (8/17)

#!C
; From file DISK-LABEL-INTERMEDIATES.LISP#> DISK-IO; SYS:
#10R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: DISK-IO; DISK-LABEL-INTERMEDIATES.#"

(define-unless :disk				       ;mx-only

(DEFUN add-partition (partition-name unit size &key (partition-type :load))
  "Create a new partition named PARTITION-NAME on UNIT SIZE number of blocks of the
type specified by the PARTITION-TYPE keyword value.
   PARTITION-NAME can be a symbol or a string.  Up to 31 characters are used in the
new partition-file's name, but the first 4 characters must be unique on this volume.
   UNIT should be the logical unit number or name of an online unit (not a floppy disk).
   SIZE should be a positive integer.  An error will be signalled if there is not enough
free disk space to create a partition that large.
   PARTITION-TYPE should be a valid partition type keyword as in *MX-PARTITION-TYPES*."
  (LET (type-num namestring filestring volume
	real-unit short-name free part-list probe-directory)
    (SETQ unit (mx-decode-unit unit (format nil "Adding partition ~a" partition-name)))
    (if (closurep unit)	;10-21-88 DAB Remote-disk?			;ab 10/8/88
	(let (result)
	  (setf result (funcall unit :add-partition  partition-name unit size partition-type))
	  (format T "~a" result))
	;else
	(CHECK-ARG partition-name (OR (STRINGP partition-name) (SYMBOLP partition-name))
		   "a string or symbol")
	(CHECK-ARG size (AND (NUMBERP size) (PLUSP size)) "a positive number")
	
	(SETQ partition-name (STRING partition-name))
	
	(SETQ type-num (type-number-from-type-symbol partition-type)
	      volume (get-volume-name-internal unit)
	      real-unit (get-real-unit unit)
	      partition-name (get-valid-partition-name partition-name (- 31. 5))	;31 char max including .TYPE
	      short-name (get-valid-partition-name partition-name)
	      filestring (STRING-APPEND partition-name "."
					(SYMBOL-NAME partition-type))
	      namestring (STRING-APPEND volume ":" *mx-directory-name* ":" filestring)
	      probe-directory (send (send (pathname (string-append "lm:" namestring))  ; DAB 04-10-90
					   :directory-pathname-as-file) :new-version 1)

	      )
	(unless (fs:probe-file probe-directory)  ; DAB 04-10-90 Make sure MICRONET directory before going on. [9903]
	  (ferror () "The directory ~a does not exist." probe-directory) )
	
	(WHEN (> (LENGTH (THE string filestring)) 31.)
	  (FERROR nil "File name string ~s longer than 31. characters." filestring))
	
	(SETQ part-list (partition-exists-p short-name unit type-num))
	(WHEN part-list
	  (FERROR nil "Partition named ~s already exists on unit ~s." short-name unit))
	
	(COND ((<= (+ size 100.) (SETQ free (get-volume-space-info unit)))
	       ;;There is enough space on disk
	       (let ((acb (add:get-acb Medium-Disk-Command-Size t))
		     (ch (add:find-channel Disk-Channel)))
		 (unwind-protect
		     (PROGN
		       ;; Fill in command overhead
		       (add:init-acb acb %MC-Disk-Cmd %DC-Add-Partition)
		       
		       ;; Input parameters
		       (add:set-parm-16b acb (add:16b-parm-number %MP-Physical-Unit)
					 real-unit)
		       (add:set-parm-16b acb (add:16b-parm-number %MP-Partition-Type)
					 type-num)
		       (add:set-parm-32b acb (add:32b-parm-number %MP-Partition-Length) size)
		       (add:put-acb-string acb %MP-Part-Name (pad-name-field short-name 4.))
		       (add:put-acb-string acb %MP-File-Name filestring)
		       (add:put-acb-string acb %MP-Whole-File-Name namestring)
		       
		       ;; Execute command
		       (add:transmit-packet-and-wait acb ch)
		       (add:check-error acb))
		   (setf (add:requestor-complete acb) t)
		   (add:return-acb-fast acb)))
	       (flush-volume unit)
	       t)
	      
	      (t (FERROR nil "Not enough free space (only ~s blocks) on volume ~s"
			 free volume)))))
  )
 )

))
