;;; -*- Mode: Common-Lisp; Package: EH; Base: 10.; Patch-File: T -*-
;;; Written 06/09/89 09:45:07 by MCCREARY,
;;; Reason: Fixed SYMEVAL-IN-STACK-GROUP so that if looking for a symbol's global value,
;;; first make sure that the symbol is bound [spr 9979].
;;; while running on Jules-Verne from band LOD9
;;; With SYSTEM 6.5, VIRTUAL-MEMORY 6.1, EH 6.0, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.0, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.1, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.2, TV 6.6, DATALINK 6.0, CHAOSNET 6.0, GC 6.2, MEMORY-AUX 6.0, NVRAM 6.0,
;;;  SYSLOG 6.0, STREAMER-TAPE 6.0, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.0,
;;;  DEBUG-TOOLS 6.0, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.1, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.0, TI-CLOS 6.5, CLEH 6.0, IP 3.46,
;;;  Experimental BUG 11.7, Experimental CLX 6.1, CLUE 6.4, X11M 6.1,  microcode 429,
;;;  Band Name: Release 6.0 + SLE 6/5

#!C
; From file SYMEVAL-IN-STACK-GROUP.LISP#> EH; MR-X:
#10R EH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "EH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: EH; SYMEVAL-IN-STACK-GROUP.#"


(defun symeval-in-stack-group (sym sg &optional frame as-if-current)
  "Returns the value and location of variable, SYM, in the binding environment
of FRAME for SG.  If FRAME is NIL, the stack group's current frame is used.
FRAME can be 0 to mean the global environment.
If the variable is unbound, the first value returned is NIL and so is the second.
If the variable is bound, the first value returned is the variable's value and the 
second is non-NIL  (in fact, it is a copy of the third value).
The third value is the location of the binding, or NIL if there is no location.
If the variable is closure-bound, this location will contain an EVCP.
If AS-IF-CURRENT is non-NIL,  the second value returned is the location where
the value would be when that SG is running. The first value returned, however,
is the current value -- not what is now stored in that location.
**WARNING** Don't call this if the stack group could be running 
in another process and thus changing its state."
  ;; 06/09/89 clm - fixed so that if looking for a symbol's global value, first make
  ;;                sure that the symbol is bound.
  (declare (special *current-stack-group*)
	   (values value location bound-p))
  (cond ((or (eq frame 0)
	     (and (neq sg *current-stack-group*)
		  (not as-if-current)))
	 (block result
	   (let* (prev-frame
		  (vcl-sym (follow-cell-forwarding (value-cell-location sym) nil))
		  (sp (sg-special-pdl sg))
		  (spp (or (and (eq frame 0) 0)
			   ;; either use index to special pdl for FRAME ...
			   (and frame (setq prev-frame (sg-previous-interesting-frame sg frame))
				(sg-frame-special-pdl-index sg prev-frame))
			   ;; or top special pdl index
			   (sg-special-pdl-pointer sg))))
	     ;; Search for variable in spec pdl
	     (unless (or (zerop (sg-in-swapped-state sg))
			 (<= spp 0))
	       ;; If bindings are swapped out, then search through them
	       (do ((i spp (1- i))
		    (sp-loc))
		   ((<= i 0))
		 (setq sp-loc (locf (aref sp i)))
		 (cond ((eq (%p-data-type sp-loc) dtp-locative)
			;; If this is a binding pair ...
			(setq sp-loc (%make-pointer-offset dtp-locative sp-loc -1))
			(if (eq (aref sp i) vcl-sym)
			    ;; and is for the variable, SYM,
			    ;; then return the saved value, invz'ing if neccesary
			    (if (location-boundp (follow-cell-forwarding sp-loc t))
				(return-from result (car sp-loc) sp-loc sp-loc)
				(return-from result nil nil sp-loc))
			    ;; else, space over second Q of binding pair
			    ;; and continue search
			    (setq i (1- i))))
		       ;; Ignore non-binding blocks, and continue search
		       (t NIL))))
	     ;; The variable isn't bound in SG or FRAME is 0, so we want its global value.
	     ;; But, must ignore bindings in our current stack group.
	     (setq sp (sg-special-pdl *current-stack-group*)
		   spp (get-own-special-pdl-pointer sp))
	     (let ((sym-loc (locf (symbol-value sym))))
	       (do ((i spp (1- i))
		    (sp-loc))
		   ((<= i 0) (if (location-boundp (follow-cell-forwarding sym-loc t))
				 (return-from result (contents sym-loc) sym-loc sym-loc)
				 (return-from result nil nil sym-loc)))
		 (setq sp-loc (locf (aref sp i)))
		 (when (eq (%p-data-type sp-loc) dtp-locative)
		   ;; If this is a binding pair ...
		   (setq sp-loc (%make-pointer-offset dtp-locative sp-loc -1))
		   ;; and is for the variable, SYM,
		   ;; then use this as SYM's loc
		   (if (eq (aref sp i) vcl-sym)
		       (setq sym-loc sp-loc))
		   ;; Space over second Q of binding pair
		   ;; and continue search
		   (setq i (1- i))))))))
	;; If current stack group and no frame specified, 
	;; just give global value
	((and (eq sg *current-stack-group*)
	      (null frame))
	 (if (boundp sym) ;; clm 06/09/89 
	     (values (symbol-value sym) (value-cell-location sym) (value-cell-location sym))
	     (values nil nil (value-cell-location sym))))
	;; Use FRAME in our current stack group,
	;; or in another stack group as if it were current.
	(t
	 (let* (prev-frame
		(vcl-sym (follow-cell-forwarding (value-cell-location sym) nil))
	        (sp (sg-special-pdl sg))
		(innermost-binding-loc nil)
		(spp (if (eq sg *current-stack-group*)
			 (get-own-special-pdl-pointer sp)
		       (sg-special-pdl-pointer sg)))
		(sp-limit
		  ;; either use index to special pdl for FRAME
		  ;; or top special pdl pointer
		  (or (let ((*innermost-visible-frame* (sg-top-frame sg)))
			(declare (special *innermost-visible-frame*))
			(and (setq prev-frame (sg-previous-frame sg frame))
			     (sg-frame-special-pdl-index sg prev-frame)))
		      spp)))
	   ;; Search thru special pdl from the special pdl pointer out index for FRAME
	   ;; and remember the LAST (outermost) binding we find.
	   ;; That is where the binding in FRAME was saved.
	   (do ((i spp (1- i))
		(sp-loc))
	       ((<= i sp-limit))
	     (setq sp-loc (locf (aref sp i)))
	     (when (eq (%p-data-type sp-loc) dtp-locative)
	       ;; If this is a binding pair ...
	       (setq sp-loc (%make-pointer-offset dtp-locative sp-loc -1))
	       ;; and is for variable, SYM,
	       ;; then save pointer to the value-slot.
	       (if (eq (aref sp i) vcl-sym)
		   (setq innermost-binding-loc sp-loc))
	       ;; Space over second Q of binding pair, and continue search
	       (setq i (1- i))))
	   ;; No binding found -- use current one.
	   (unless innermost-binding-loc
	       (setq innermost-binding-loc (value-cell-location sym)))
	   ;; Now we have the right location for SYM.
	   ;; So return its value and location
	   (if (eq sg *current-stack-group*)
	       (if (location-boundp (follow-cell-forwarding innermost-binding-loc t))
		   (values (contents innermost-binding-loc) innermost-binding-loc innermost-binding-loc)
		 (values nil nil innermost-binding-loc))
	     ;; If not really current stack group, just pretending,
	     ;; then the current contents of that location is not the correct value.
	     (multiple-value-bind (value ignore location)
		 ;; So get the correct value.
		 (symeval-in-stack-group sym sg frame nil)
	       (values value (if location innermost-binding-loc) innermost-binding-loc)))))))
))
