;;; -*- Mode: Common-Lisp; Package: EH; Base: 10.; Patch-File: T -*-
;;; Patch file for EH version 6.3
;;; Reason: Add modifications to Error Handler to enhance the 
;;; bactrace in bug reports by showing the disassembly
;;; of a few instructions for the top two functions.
;;; Written 06/16/89 08:22:23 by MCCREARY,
;;; while running on Jules-Verne from band LOD9
;;; With SYSTEM 6.6, VIRTUAL-MEMORY 6.1, EH 6.2, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.2, TV 6.8, DATALINK 6.0, CHAOSNET 6.0, GC 6.2, MEMORY-AUX 6.0, NVRAM 6.0,
;;;  SYSLOG 6.0, STREAMER-TAPE 6.2, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.2,
;;;  Experimental DEBUG-TOOLS 6.2, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.1, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, TI-CLOS 6.5, CLEH 6.4, IP 3.46,
;;;  Experimental BUG 11.8, Experimental CLX 6.1, CLUE 6.5, X11M 6.1,  microcode 429,
;;;  Band Name: Release 6.0 + SLE 6/5

#!C
; From file DEBUGGER-COMMANDS.LISP#> EH; MR-X:
#10R EH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "EH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: EH; DEBUGGER-COMMANDS.#"


(defun show-frame-for-bug-message (sg frame &optional disassemble)
       ;;  06/17/89 DNG - Add DISASSEMBLE option instead always telling SHOW-ALL-OF-IT
       ;;                 to not disassemble.
  (let* ((rp (sg-regular-pdl sg))
	 (function (rp-function-word rp frame)))
    ;; show return values if any
    (show-additional-info sg rp frame)
    (terpri)
    ;; show function, args and if fef, the disassembled code
    (if (typep function 'compiled-function)
	(let ((*disassemble-instruction-count* 6))
	  (show-all-of-it sg frame (not (and disassemble
					     (not (eq (fef-name function) 'BREAKON-THIS-TIME))))))
	(print-function-and-args sg frame))))
))

#!C
; From file DEBUGGER-COMMANDS.LISP#> EH; MR-X:
#10R EH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "EH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: EH; DEBUGGER-COMMANDS.#"


(defun show-frame-briefly-for-bug-message (sg frame)
  ;; Print function name and if fef, the last pc
  ;; 06/16/89 DNG - Show arguments of important functions.
  (let* ((rp (sg-regular-pdl sg))
	 (function (rp-function-word rp frame))
	 (name (fef-name function)))
    (send *standard-output* :fresh-line)
    (if (function-spec-get name 'important-function)
	(print-function-and-args sg frame)
      (if (typep function 'compiled-function)
	  (format t "~S (P.C. = ~A)" name (rp-location-counter-offset rp frame))
	(prin1 (function-name function))))
    (terpri)))

))

#!C
; From file ERROR-CONDITION-DEFINITIONS.LISP#> EH; MR-X:
#10R EH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "EH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: EH; ERROR-CONDITION-DEFINITIONS.#"


(defmethod (condition :bug-report-description) (stream &optional n-frames)
  ;; bug report describption contains the error message,
  ;; and a backtrace of frames
  ;; 06/16/89 DNG - Disassemble the top two frames.
  (declare (special *error-sg* *error-locus-frame* *default-bug-report-frames*))
  (send self :print-error-message *error-sg* nil stream)
  (format stream "Backtrace from the debugger:")
  (let ((*standard-output* stream)
	total-frames)
    ;; get total number of frames possible for this sg
    (do ((frame *error-locus-frame* (sg-next-frame *error-sg* frame))
	 (i 0 (1+ i)))
	((null frame) (setq total-frames i)))
    ;; show frame details for up to so many frames;
    ;; after so many frames, show brief detail until get to 10 frames;
    ;; after 10 frames, show a ...
    (do ((frame *error-locus-frame* (sg-next-frame *error-sg* frame))
	 (i 0 (1+ i)))
	((null frame))
      (cond ((< i (or n-frames *default-bug-report-frames*))
	     (show-frame-for-bug-message *error-sg* frame (< i 2))
	     (when (< i 3)
	       (let ((*current-frame* frame))
		 (com-print-frame-bindings *error-sg* self))))
	    ((= i (or n-frames *default-bug-report-frames*))
	     (format t "~%~%Remainder of stack:~%~%")
	     (show-frame-briefly-for-bug-message *error-sg* frame))
	    ((< i (+ 10. (or n-frames *default-bug-report-frames*)))
	     (show-frame-briefly-for-bug-message *error-sg* frame))
	    ((> i (- total-frames 10.))
	     (show-frame-briefly-for-bug-message *error-sg* frame))
	    ((= i (+ 10. (or n-frames *default-bug-report-frames*)))
	     (format t "~&..."))))))
))

#!C
; From file ERROR-CONDITION-DEFINITIONS.LISP#> EH; MR-X:
#10R EH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "EH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: EH; ERROR-CONDITION-DEFINITIONS.#"


(defprop sys:fasload-internal t important-function)
(defprop sys:readfile-internal t important-function)
(defprop compiler:compile-stream t important-function)
(defprop compiler:qc-translate-function t important-function)
(defprop compiler:compile-top-level-form t important-function)
(defprop make-system t important-function)

))
