;;; -*- Mode: Common-Lisp; Package: EH; Base: 10.; Patch-File: T -*-

;;; Reason: Changed eh:SG-save-state to save registers in a without-interrupts, in case we make a 
;;; locative to words with data type "trap"

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for EH version 6.5
;;; Written 10/02/89 12:21:36 by FISH,
;;; while running on DaVinci from band LOD2
;;; With SYSTEM 6.14, VIRTUAL-MEMORY 6.1, EH 6.4, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.2, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.10, TV 6.15, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.5,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, TI-CLOS 6.20, CLEH 6.5, IP 3.47,
;;;  Experimental CLX 6.2, CLUE 6.10, X11M 6.13, Experimental BUG 11.15, Experimental SEYMOUR3 2.0,
;;;  Experimental SLAP 3.15,  microcode 465, Band Name: rel6 7/25 seymour

#!C
; From file RUN-ERRING-SG.LISP#> EH; Hotel:
#10R EH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "EH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: EH; RUN-ERRING-SG.#"


(defun sg-save-state (sg &optional suppress-pdl-growing)
  "Save the state (accumulators, etc) of stack group SG on its pdl.
SUPPRESS-PDL-GROWING inhibits making the stack bigger heuristically.
Once this function has been called, you can then push and invoke a call block
to cause the stack group to execute things without clobbering the saved
accumulators from the error."
  ;; 04/09/87 CLM - Save the microstack-pushed bit in the call-info word of foothold
  ;;                in order to restore correctly (problem arose in a call-out in
  ;;                a micro-instruction).
  ;; 04/11/88 CLM - Part of the fix for SPR 7417; do not try to signal an error - pass T
  ;;                for inhibit-error arg to SG-MAYBE-GROW-PDLS.
   (let* (sg-ptr
	  (reg-pdl (sg-regular-pdl sg))
	  (old-top-frame (sg-top-frame sg))
	  ;; the new frame's args will start on top of the top frame's call state words.
	  (new-arg-offset (- old-top-frame %call-state-length))
	  ;; the stack group items will be saved as args for the new frame
	  (length-sg (array-leader-length sg))
	  ;; local offset is just above the new frame's args and old top frame's call state words
	  (new-local-offset (+ new-arg-offset (* 2 length-sg) %call-state-length))
	  ;; the top of new frame will include its own call state words
	  (new-top-frame (+ new-local-offset (fef-number-of-locals #'foothold) %call-state-length)))

     ;; Make sure there is room to save SG state
     ;; DON'T signal an error, if pdl had to grow
     (unless suppress-pdl-growing (sg-maybe-grow-pdls sg nil nil nil t))
     ;; Move the previous top most frame's call state words to just before the new frame's local offset
     (do ((i 1 (1+ i)))
	 ((> i %call-state-length))
       (setf (aref reg-pdl (- new-local-offset i)) (aref reg-pdl (- old-top-frame i))))
     ;; Save SG items in new frame's arg area on regular pdl;
     ;; Each Q is saved as 2 words (pointer and tag) to avoid data type problems.
     ;; Do without-interrupts in case we make a locative to words with data type "trap"
     (without-interrupts
       (do ((i 0 (1+ i))
	    (j new-arg-offset (+ 2 j)))
	   ((>= i length-sg))
	 (setq sg-ptr (ap-leader sg i))
	 (setf (aref reg-pdl j)
	       (if (%p-pointerp sg-ptr) (%p-contents-as-locative sg-ptr) (%p-pointer sg-ptr)))
	 (setf (aref reg-pdl (1+ j))
	       (%p-ldb %%q-all-but-pointer sg-ptr))))
     ;; Need to convert arg and local pointer of frame state we just moved up
     ;;  (I don't think we need to do this.  - pf, May 8, 1986)
;     (let ((prev-top-frame new-local-offset))
;       (setf (sg-top-frame sg) prev-top-frame)
;       (setf (rp-local-pointer reg-pdl prev-top-frame) (pdl-buffer-index sg prev-frame-local-offset))
;       (setf (rp-argument-pointer reg-pdl prev-top-frame) (pdl-buffer-index sg prev-frame-arg-offset)))
     ;; Set new top frame to this new "FOOTHOLD" frame
     (setf (sg-top-frame sg) new-top-frame)
     ;; Set all 5 call-state words for new frame
     ;; FOOTHOLD will be the FEF for this frame, but it should never be called.
     (setf (rp-location-counter-offset reg-pdl new-top-frame) (fef-initial-pc #'foothold))
     (setf (rp-function-word reg-pdl new-top-frame) #'foothold)
     ;; NOTE: PDL-BUFFER-INDEX converts a pdl array index to a pdl buffer index
     (setf (rp-local-pointer reg-pdl new-top-frame) (pdl-buffer-index sg new-local-offset))
     (setf (rp-argument-pointer reg-pdl new-top-frame) (pdl-buffer-index sg new-arg-offset))
     (setf (rp-call-info-word reg-pdl new-top-frame) 0)
     ;;need to save the microstack-pushed bit in the call-info word for exceptional cases
     (when (plusp (sg-restore-microstack sg))
       (setf (rp-call-info-word reg-pdl new-top-frame)
	     (dpb 1 %%call-info-microstack-pushed
			 (rp-call-info-word reg-pdl new-top-frame)))
       (setf (sg-restore-microstack sg) 0))
;;  I don't believe we need to do this anymore.  - pf, July 18, 1986
;     ;;Clear QBBFL left over from previous frame
;     (setf (sg-flags-qbbfl sg) 0)
     ))
))
