;;; -*- Mode: Common-Lisp; Package: SI; Base: 8.; Patch-File: T -*-

;;; Reason: Changed dump-memory default base to base-16. Changed print-base for dump-objects
;;; and dump-objects-in-region to default to base 16.  Made the region number consistently
;;; print in base 10. - RJF

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for GC version 6.4
;;; Written 05/23/90 11:07:07 by FISH,
;;; while running on DaVinci from band LOD2
;;; With SYSTEM 6.31, VIRTUAL-MEMORY 6.3, EH 6.7, MAKE-SYSTEM 6.3, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.4, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-FILE 6.10, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.2,
;;;  COMPILER 6.14, TV 6.24, DATALINK 6.0, CHAOSNET 6.6, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.3,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.15,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.5, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.8, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.3, VISIDOC 6.7, TI-CLOS 6.46, CLEH 6.5, IP 3.62,
;;;  Experimental CLX 6.11, CLUE 6.104, X11M 6.24, Experimental BUG 11.19,  microcode 483,
;;;  Band Name: Rel 6.0 + SLE 5/21

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-typed-q (q &key (base 16.))
  (DECLARE (UNSPECIAL base))
  (LET* ((*print-base* base)
	 (*read-base* base)
	 (cc (LDB %%Q-Cdr-Code q))
	 (dtp (LDB %%Q-Data-Type q))
	 (ptr (LDB %%Q-Pointer q)))
    
    (UNLESS *dtp-list-for-printing* (make-short-type-names))
    (FORMAT nil "<~[~;CDR-ERROR ~;CDR-NIL ~;CDR-NEXT ~]~a ~a>"
	    cc
	    (OR (NTH dtp q-data-types) 'dtp-unused)
	    ptr)
    ))
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-memory (address-or-object &key (length 16.)
	                                   (base 16.)
					   (addr-or-obj)
				           (stream *standard-output*)
	                                   (bignum-is-dump-object nil))
  "Dumps virtual memory in raw (numeric) format starting at ADDRESS-OR-OBJECT.
  ADDRESS-OR-OBJECT can be an integer or an arbitrary object.  If it is an
integer (a bignum or a fixnum) it is used as a virtual memory address.  
If it is an object, the object's pointer field is used as the address.
  A check is done prior to dumping to assure that the resulting address
is valid virtual memory.
  The memory display will be in the base supplied as the BASE keyword.  This
must be 8, 10 or 16.
  The LENGTH keyword specifies how many words to dump.  If this amount
would run into unallocated virtual memory it will be truncated to a safe
amount and a warning will be issued.
  When ADDRESS-OR-OBJECT is nil, the value of the ADDR-OR-OBJ keyword is
used instead.   This allows you to be free of positional parameters.
  Usually if ADDRESS-OR-OBJECT is a bignum it will be used as an address.
However, you may actually want the memory containing the bignum dumped instead.
In that case supply a non-NIL value for the BIGNUM-IS-DUMP-OBJECT keyword."
  (DECLARE (UNSPECIAL base length))
  (CHECK-ARG base (AND (INTEGERP base)
			       (OR (= base 8.) (= base 10.) (= base 16.)))
	     "A integer base number 8. 10. or 16.")

  (LET (addr ptr reg page-addr area
	last-allocated-addr free-ptr
	last-addr-on-allocated-page
	last-addr-to-dump
	dump-type)

    ;; Housekeeping
    (WHEN (NULL *dtp-list-for-printing*) (make-short-type-names))
    (WHEN (NULL address-or-object) (SETQ address-or-object addr-or-obj))
    (TERPRI)
    
    ;; Perform concordance.
    (COND ((INTEGERP address-or-object)
	   (COND (bignum-is-dump-object 
		  (SETQ ptr (%POINTER address-or-object))
		  (SETQ addr (convert-to-unsigned ptr)))
		 (t
		  (COND ((> (INTEGER-LENGTH address-or-object)
			    (BYTE-SIZE %%Q-POINTER))
			 (FORMAT stream "~%#o~o does not make sense as ~d-bit address."
				 address-or-object (BYTE-SIZE %%Q-POINTER))
			 (RETURN-FROM dump-memory nil))
			(t
			 (SETQ ptr (convert-to-signed address-or-object))
			 (SETQ addr (convert-to-unsigned address-or-object)))))))
	  (t
	   (SETQ ptr (%POINTER address-or-object))
	   (SETQ addr (convert-to-unsigned ptr))))

    ;; Error checking.
    ;; If the first word on PTR's page is valid (in allocated portion of some region)
    ;; we can safely look at Qs in the whole page.  Otherwise, the address is illegal.
    (SETQ reg (%REGION-NUMBER ptr)
	  page-addr (LOGAND ptr (- page-size)))
    (IF (NULL reg)
	(SETQ free-ptr (1+ (convert-to-unsigned -1))
	      last-allocated-addr (convert-to-unsigned -1))
	(SETQ free-ptr (convert-to-unsigned
			 (%POINTER-PLUS (AREF #'region-origin reg) (AREF #'region-free-pointer reg)))
	      last-allocated-addr (1- free-ptr)))
    (COND ((a-memory-address-p ptr)
	   (SETQ dump-type "A-Memory"))
	  ((io-space-address-p ptr)
	   (SETQ dump-type "IO-Space"))
	  ((NULL reg)
	   (FORMAT stream "~%Address #o~o is not currently assigned to any region." addr)
	   (RETURN-FROM dump-memory nil))
	  ;; Can dump any address in a fixed-wired area
	  ((perm-wired-address-p ptr)
	   t)
	  ((NOT (pointer-valid-p page-addr))
	   (FORMAT stream "~%Address #o~o is assigned to region ~d. in area ~d. (~a),~
                           ~%but no part of the page which contains it is before the region free pointer~
                           ~%at #o~o."
		   addr reg (SETQ area (%AREA-NUMBER ptr)) (AREF #'AREA-NAME area) free-ptr)
	   (RETURN-FROM dump-memory nil))
	  ((NOT (pointer-valid-p ptr))
	   (FORMAT stream "~%Warning:  Address #o~o is assigned to region ~d. in area ~d. (~a)~
                           ~%          but is beyond the region free pointer at #o~o.  However it is~
                           ~%          on an assigned page so still can be be dumped."
		   addr reg (SETQ area (%AREA-NUMBER ptr)) (AREF #'AREA-NAME area) free-ptr)))

    (SETQ last-addr-to-dump (convert-to-unsigned (%POINTER-PLUS ptr (1- length)))
	  last-addr-on-allocated-page (convert-to-unsigned
					(%POINTER-PLUS
					  (LOGAND (convert-to-signed last-allocated-addr)
						  (- page-size))
					  (1- page-size))))
    (WHEN (AND (> last-addr-to-dump last-addr-on-allocated-page)
	       (NOT (perm-wired-address-p last-addr-to-dump)))
      (FORMAT stream "~2%Warning:  Dump length of ~d. would go past end of the last allocated page in~
                       ~%          region ~d. (origin #o~o, free-pointer #o~o, last-page-addr #o~o).~
                       ~%          Truncating length to ~d."
	      length reg (convert-to-unsigned (AREF #'region-origin reg)) free-ptr last-addr-on-allocated-page
	      (SETQ length (- length (- last-addr-to-dump last-addr-on-allocated-page)))))

    ;; OK, can now do dumping.
    (UNLESS dump-type
      (SETQ dump-type
	    (FORMAT nil "REGION ~d. in AREA ~d. (~a)"
		    reg (SETQ area (%AREA-NUMBER ptr)) (AREF #'AREA-NAME area))))
    (FORMAT stream "~&~2%  ~a dump of ~a." 
	    (CASE base
	      (8.         "OCTAL")
	      (10.        "DECIMAL")
	      (16.        "HEX"))
	    dump-type)

   (FORMAT
     stream
     "~2%     ADDRESS          VALUE            CDR-CODE                DATA-TYPE               POINTER     REVERSED ASCII~
      ~2%  -------------  ----------------  -----------------  ----------------------------  -------------  --------------")
   
   (DO ((target ptr (%POINTER-PLUS target 1))
	(count 0 (1+ count)))
       ((>= count length))
     (LET* ((target-addr (convert-to-unsigned target))
	    (cdr-code (%P-LDB %%Q-CDR-CODE target))
	    (cc-name (NTH cdr-code *cdr-code-list-for-printing*))
	    (dtp (%P-LDB %%Q-DATA-TYPE target))
	    (dtp-name (NTH dtp *dtp-list-for-printing*))
	    (pointer (convert-to-unsigned (%P-LDB %%Q-POINTER target)))
	    (word (DPB cdr-code %%Q-CDR-CODE
		       (DPB dtp %%Q-DATA-TYPE pointer)))
            (char0   (INT-CHAR (LDB #o0010 word)))
            (char1   (INT-CHAR (LDB #o1010 word)))
            (char2   (INT-CHAR (LDB #o2010 word)))
            (char3   (INT-CHAR (LDB #o3010 word)))
	    (splat   (INT-CHAR 0)))
       (CASE base
	 (8.  (FORMAT stream "~%   ~:11,,o    ~:14,,o    ~1o  ~12a    ~2o  ~22a    ~:11,,o   "
		      target-addr word cdr-code cc-name dtp dtp-name pointer))
	 (10. (FORMAT stream "~%   ~:11,,d    ~:14,,d    ~1d  ~12a    ~2d  ~22a    ~:11,,d   "
		      target-addr word cdr-code cc-name dtp dtp-name pointer))
	 (16. (FORMAT stream "~%     ~8x        ~8x       ~1x  ~12a    ~2x   ~22a     ~8x    "
		      target-addr word cdr-code cc-name dtp dtp-name pointer)))
       (FORMAT stream "   ~c ~c ~c ~c"
	       (IF (GRAPHIC-CHAR-P char0) char0 splat) (IF (GRAPHIC-CHAR-P char1) char1 splat)
	       (IF (GRAPHIC-CHAR-P char2) char2 splat) (IF (GRAPHIC-CHAR-P char3) char3 splat))))
   ))
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-objects-hdr-function (region strm)
  (FORMAT strm "~2%  REGION ~d. (~a)" region (region-space-type region))
  (FORMAT strm "~2%    Address      Size    Object ~
                 ~%  -----------  --------  --------")
  )
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-objects (start-addr &key num-objects
		                     (max-print-length 3.)
				     (max-print-level 3.) 
				     (max-string-length 50.)
				     (base 16.)
				     (stream *standard-output*))
  "Start dumping objects at address START-ADDR.  START-ADDR should be
a fixnum.  If it is an invalid virtual memory address, nothing will be dumped.  
If START-ADDR is in the middle of a structure (even an unboxed structure), the 
dump will begin with the object containing START-ADDR.
  The :NUM-OBJECTS keyword specifies the maximum number of objects to dump.  
It defaults to all the objects in the region containing START-ADDR.
  The :MAX-PRINT-LENGTH and :MAX-PRINT-LEVEL keywords can be used to
control the settings of the *PRINT-LENGTH* and *PRINT-LEVEL* variables
respectively.  :MAX-STRING-LENGTH says how much of strings to print.
  The address and size of each object is printed.   The :BASE keyword
says what base to do this in.
  :STREAM specifies the stream for the output."
  (DECLARE (UNSPECIAL base ignore))
  (LET* ((*print-level* max-print-level)
	 (*print-length* max-print-length)
	 (*print-base* base)
	 (*read-base* base)
	 (strm (OR stream 'null-stream))
	 (start-fn #'dump-objects-hdr-function)
	 (obj-fn #'dump-objects-obj-function))
    
    (map-objects start-addr :num-objects num-objects
		 :analysis-function obj-fn
		 :analysis-function-args `(,strm ,max-string-length)
		 :start-function start-fn
		 :safe-for-oldspace-objects t))
  )
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-objects-in-region (region &key (start-offset 0)
			                   (num-objects :all)
			                   (max-print-length 3.)
					   (max-print-level 3.) 
					   (max-string-length 50.)
					   (base 16.)
					   (stream *standard-output*))
  "Start dumping objects at offset START-OFFSET in region REGION.
If region is an invalid (free) region, nothing will be dumped.  If 
START-OFFSET is in the middle of a structure (even an unboxed structure), 
the dump will begin with the object containing START-OFFSET.
  The :NUM-OBJECTS keyword specifies the maximum number of objects to dump.  
It defaults to all the objects in the region containing START-OFFSET.
  The :MAX-PRINT-LENGTH and :MAX-PRINT-LEVEL keywords can be used to
control the settings of the *PRINT-LENGTH* and *PRINT-LEVEL* variables
respectively.  :MAX-STRING-LENGTH says how much of strings to print.
  The address and size of each object is printed.   The :BASE keyword
says what base to do this in.
  :STREAM specifies the stream for the output."
  (DECLARE (UNSPECIAL base))
  (LET* ((*print-level* max-print-level)
	 (*print-length* max-print-length)
	 (*print-base* base)
	 (*read-base* base)
	 (strm (OR stream 'null-stream))
	 (reg-start-fn #'dump-objects-hdr-function)
	 (obj-fn #'dump-objects-obj-function))
    
    (map-objects-in-region region :start-offset start-offset
			   :num-objects num-objects
			   :analysis-function obj-fn
			   :analysis-function-args `(,strm ,max-string-length)
			   :region-start-function reg-start-fn
			   :safe-for-oldspace-objects t))
  )
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-a-memory (&key (start-offset 0)
		           (length 32.)
			   (all nil)
			   (base 16.))
  "Dump the contents of A-Memory starting at :START-OFFSET for :LENGTH words.
If the :ALL keyword is non-nil, the :LENGTH keyword is ignored and all of A-Memory
is dumped."
  (DECLARE (UNSPECIAL base))
  (WHEN all (SETQ length (%POINTER-DIFFERENCE -1 a-memory-virtual-address)))
  (FORMAT t "~2%     Address          Value        Q-Representation~
              ~%  -------------  ----------------  -------------------------------")
  (DO* ((offset start-offset (1+ offset))
	(addr (convert-to-unsigned (%POINTER-PLUS a-memory-virtual-address offset))
	      (1+ addr))
	(val)
	(end-addr (convert-to-unsigned
		    (%POINTER-PLUS (MOD length (1+ (%POINTER-DIFFERENCE -1 a-memory-virtual-address)))
				   a-memory-virtual-address))))
       ((>= addr end-addr))
    (TERPRI)
    (CASE base
      (8.  (FORMAT t "   ~:11,,o    ~14,,o   ~a" 
		   addr (SETQ val (%read-a-memory offset)) (dump-typed-q val :base base)))
      (10. (FORMAT t "   ~:11,,d    ~:14,,d   ~a"
		   addr (SETQ val (%read-a-memory offset)) (dump-typed-q val :base base)))
      (16. (FORMAT t "     ~7,,x         ~8,,x      ~a"
		   addr (SETQ val (%read-a-memory offset)) (dump-typed-q val :base base)))
      (t (FORMAT t "   ~@11,,a    ~@14,,a   ~a" 
		 addr (SETQ val (%read-a-memory offset)) (dump-typed-q val :base base))))
    ))
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-m-memory-q-storage (&key (base 16.))
  "Prints out names and values for all M-Memory-Location-Names"
  (DECLARE (UNSPECIAL base))
  (LET ((*print-base* base)
	(*read-base* base)
	(MAX 0) tem)
    (DOLIST (m-q m-memory-location-names max)
      (WHEN (> (SETQ tem (LENGTH (THE string (SYMBOL-NAME m-q))))
	       max)
	(SETQ max tem)))
    (DOLIST (m-q m-memory-location-names)
      (FORMAT t "~%  ~vs  ~s" max m-q (SYMBOL-VALUE m-q)))
    (VALUES)
    ))
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-a-memory-q-storage (&key (base 16.))
  "Prints out names and values for all A-Memory-Location-Names"
  (DECLARE (UNSPECIAL base))
  (LET ((*print-base* base)
	(*read-base* base)
	(max 0) tem)
    (DOLIST (a-q a-memory-location-names max)
      (WHEN (> (SETQ tem (LENGTH (THE string (SYMBOL-NAME a-q))))
	       max)
	(SETQ max tem)))
    (DOLIST (a-q a-memory-location-names)
      (FORMAT t "~%  ~vs  ~s" max a-q (SYMBOL-VALUE a-q)))
    (VALUES)
    ))
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-a-memory-counters (&key (base 16.))
  "Prints out names and values for all A-Memory-Counter-Block-Names"
  (DECLARE (UNSPECIAL base))
  (LET ((*print-base* base)
	(*read-base* base)
	(max 0) tem)
    (DOLIST (a-counter a-memory-counter-block-names max)
      (WHEN (> (SETQ tem (LENGTH (THE string (SYMBOL-NAME a-counter))))
	       max)
	(SETQ max tem)))
    (DOLIST (a-counter a-memory-counter-block-names)
      (FORMAT t "~%  ~vs  ~@12a" max a-counter
	      (%read-a-memory (+ (SYMBOL-VALUE a-counter) %counter-block-a-mem-address))))
    (VALUES)
    ))
))

#!C
; From file MEMORY-DEBUG.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; MEMORY-DEBUG.#"


(DEFUN dump-sca-words (&key (base 16.))
  "Prints out names and values for all System-Communication-Area words"
  (DECLARE (UNSPECIAL base))
  (LET ((*print-base* base)
	(*read-base* base)
	(max 0) tem)
    (DOLIST (sca-symbol System-Communication-Area-Qs)
      (WHEN (> (SETQ tem (LENGTH (THE string (SYMBOL-NAME sca-symbol))))
	       max)
	(SETQ max tem)))
    (DOLIST (sca-symbol System-Communication-Area-Qs)
      (FORMAT t "~%  ~vs  (~2a):  ~a" max sca-symbol (SYMBOL-VALUE sca-symbol)
	      (LET* ((base-adr (AREF #'region-origin system-communication-area))
		     (ptr (%POINTER-PLUS base-adr (SYMBOL-VALUE sca-symbol)))
		     (dtp (%P-LDB %%Q-Data-Type ptr))
		     (dtp-symb (Q-DATA-TYPES dtp)))
		(IF (OR (MEMBER dtp-symb *lisp-pointer-types* :test #'EQ)
			(MEMBER dtp-symb *lisp-immediate-types* :test #'EQ))
		    (AREF #'system-communication-area (SYMBOL-VALUE sca-symbol))
		    (dump-typed-q (%p-ldb-word ptr))))))
    (VALUES)
    ))
))
