;;; -*- Mode: Common-Lisp; Package: SI; Base: 8.; Patch-File: T -*-
;;; Patch file for GC version 6.7
;;; Reason: GC-PROCESS changes to keep gen 1 and gen 2 sizes as large as possible and to
;;; avoid cons-driven scavenging of higher-level generations.  Also small display
;;; changes to GC-STATUS.
;;; Written 06/27/90 16:05:01 by COURTS,
;;; while running on jomama from band LOD1
;;; With SYSTEM 6.28, VIRTUAL-MEMORY 6.3, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.8, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.22, DATALINK 6.0, CHAOSNET 6.5, GC 6.6, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.11,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.4, MAIL-READER 6.6, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.26, CLEH 6.5, IP 3.57,
;;;  Experimental SNRL 4.0, Experimental SST-WINDOWS 1.0, Experimental SNRL-ADD-ONS 1.0,
;;;  Experimental CONFLICT-RESOLUTION 46.0, Experimental QUERY 1.0, MMON 6.2, Experimental GRAPHICS-WINDOW 1.0,
;;;  Experimental GRAPHICAL-SORT 3.0, Experimental VIEWS 34.4, Inconsistent SPREADSHEET 26.8,
;;;  Experimental W-SPREADSHEET 17.4, Experimental SPREADSHEET-VIEW 10.0,  microcode 432,
;;;  Band Name: R6.0++SST+MMON 6/27/90

#!C
; From file GC-DEFS.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; GC-DEFS.#"

(eval-when (load)
  (si:gc-off-temporarily))

;; HRC 6/27/90.  Use NIL console delay so idle scavenging will keep us ahead of the game.
(DEFVAR *gc-console-delay-interval* nil
	"Number of seconds the console may be unused before idle scavenging may kick in.
NIL means no delay; T means disable idle scavenging.")


;; HRC 6/27/90.  Make default NIL.
(DEFVAR GC-PROMOTE-GEN-2 nil
  "Controls whether a incremental non-EAS generation 2 collection is promoting")


;; HRC 6/27/90.  New, for GC-PROCESS mods
(DEFVAR *GC-CONS-GUARANTEE* 10.)	;; AMOUNT OF SCAV FREE CONSING WE ATTEMPT TO PROVIDE DURING A GEN1 COLLECTION, MB.
(DEFVAR *GC-ADDRESS-SPACE-SAFETY-PAD* 8.) ;; SUBTRACTIVE JUDGEMENT APPLIED TO ADDRESS SPACE RESERVE, MB.

))

#!C
; From file GC.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; GC.#"

;; HRC 6/27/90.  Changes to GC-PROCESS flip threshold logic  for generations 1, 2,
;; and (EAS)  3 to  make the  flip thrshold  as large as safety will allow.  Also
;; changes to avoid cons-driven scavenging in higher-generation collections.

(DEFUN gc-process ()
  (LET* (THRESHOLD USED-VIRTUAL-ADDRESS-SPACE RESERVE C BIAS S
	 (LAG-PREVIOUS-GENERATION-HIGH NIL)
	 (PREVIOUS-GENERATION-HIGH 1)
	 (B
	   (ROUND (* GC-FRACTION-OF-RAM-FOR-GENERATION-ZERO
		     (SYSTEM:SYSTEM-COMMUNICATION-AREA
		       SYSTEM:%SYS-COM-MEMORY-SIZE)))))
    (SETF *GC-MAX-BUCKET-SIZE* 2500000.)
    
    (SETF (AREF FLIP-SIZE 0) B
	  (AREF FLIP-SIZE 1) 0
	  (AREF FLIP-SIZE 2) 0
	  (AREF FLIP-SIZE 3) 0)
    (DO ()
	(NIL)					;Do forever
      (OR %GC-FLIP-READY (PROCESS-WAIT "Await scavenge" 'SYMBOL-VALUE '%GC-FLIP-READY))
      (GC-RECLAIM-OLDSPACE)
      (SETF B (ROUND (* GC-FRACTION-OF-RAM-FOR-GENERATION-ZERO
			(SYSTEM:SYSTEM-COMMUNICATION-AREA
			  SYSTEM:%SYS-COM-MEMORY-SIZE))))
      (SETF (AREF FLIP-SIZE 0) B)
      (IF EXTENDED-ADDRESS-SPACE
	  (SETF *GC-MAX-INCREMENTAL-GENERATION* 3)
	  (IF (> *GC-MAX-INCREMENTAL-GENERATION* 2)
	      (SETF *GC-MAX-INCREMENTAL-GENERATION* 2)))
      ;; CONVERT GC-CONS-GUARANTEE AND GC-ADDRESS-SPACE-SAFETY-PAD TO Q UNIT OF MEASURE AND STORE IN
      ;; LOCAL VARIABLES WITH SHORT NAMES, C AND S.
      (SETF C (FLOOR (* *GC-CONS-GUARANTEE* 1024. 1024.)
		     4.))
      (SETF S (FLOOR (* *GC-ADDRESS-SPACE-SAFETY-PAD* 1024. 1024.)
		     4.))
      (MULTIPLE-VALUE-BIND (W D used-active PDL)
	  (GET-GENERATION-SIZES)
	(SETF USED-VIRTUAL-ADDRESS-SPACE 0)
	(DOTIMES (I 8.) (INCF USED-VIRTUAL-ADDRESS-SPACE (AREF D I)))
	(MULTIPLE-VALUE-BIND (USABLE-SWAP-SPACE SWAP-SPACE-FREE)
	    (SWAP-SPACE-INFO)
	  (SETF SWAP-SPACE-FREE	
		(* (- SWAP-SPACE-FREE
		      (COUNT-PAGES-NEEDING-WRITABLE-PAGING-STORE (NOT EXTENDED-ADDRESS-SPACE)))
		   PAGE-SIZE))			; Also check phys mem.
	  (SETF RESERVE (-  (MIN (- *max-virtual-address* USED-VIRTUAL-ADDRESS-SPACE 2097152.)
				SWAP-SPACE-FREE)
			   S))
	  (COND ((= *GC-MAX-INCREMENTAL-GENERATION* 0)
		 (SETF THRESHOLD 0))
		((= *GC-MAX-INCREMENTAL-GENERATION* 1)
		 (SETF THRESHOLD (FLOOR (- (* 16. (- (+ RESERVE
							(AREF D 1))
						     B
						     C))
					   (* 22. (AREF D 0)))
					35.))
		 (SETF BIAS (- RESERVE
			       (FLOOR (* 19. (AREF D 1))
				      16.))))
		((= *GC-MAX-INCREMENTAL-GENERATION* 2)
		 (SETF THRESHOLD (FLOOR (- (* 16. (- (+ RESERVE
							(AREF D 1))
						     B
						     C))
					   (* 25. (AREF D 0))
					   (* 19. (AREF D 2)))
					38.))
		 (SETF BIAS (- RESERVE
			       (FLOOR (+ (* 22. (AREF D 1))
					 (* 19. (AREF D 2)))
				      16.)))
		 ;; CHECK FOR NEED TO SHUT DOWN GENERATION 2.
		 (IF (AND LAG-PREVIOUS-GENERATION-HIGH
			  (= LAG-PREVIOUS-GENERATION-HIGH 2.)
			  (NOT GC-PROMOTE-GEN-2)
			  (> (AREF D 1) THRESHOLD))
		     ;; SHUT DOWN GENERATION 2
		     (SETF *GC-MAX-INCREMENTAL-GENERATION* 1)))
		(T
		 (SETF THRESHOLD (FLOOR (- (* 16. (- (+ RESERVE
							(AREF D 1))
						     B
						     C))
					   (* 28. (AREF D 0))
					   (* 22. (AREF D 2))
					   (* 19. (AREF USED-ACTIVE 3 3))
					   (AREF USED-ACTIVE 3 0)
					   (AREF USED-ACTIVE 3 1)
					   (AREF USED-ACTIVE 3 2)
					   (AREF D 4.))
					41.))
		 (SETF BIAS (- RESERVE
			       (FLOOR (+ (* 25. (AREF D 1))
					 (* 22. (AREF D 2))
					 (* 19. (AREF USED-ACTIVE 3 3))
					 (AREF USED-ACTIVE 3 0)
					 (AREF USED-ACTIVE 3 1)
					 (AREF USED-ACTIVE 3 2)
					 (AREF D 4.))
				      16.)))))
	  (SETF (AREF FLIP-SIZE 1) THRESHOLD)
	  (SETF LAG-PREVIOUS-GENERATION-HIGH PREVIOUS-GENERATION-HIGH)
	  (COND ((OR PREVIOUS-GENERATION-HIGH
		     (= 0 *GC-MAX-INCREMENTAL-GENERATION*)
		     (AND (< (AREF D 1) THRESHOLD)
			  (NOT (AND (ACTIVE-OVERSIZED-P (AREF USED-ACTIVE 3. 0))
				    (> (- (+ (AREF USED-ACTIVE 3. 2.) (AREF USED-ACTIVE 3. 3.)) PDL)
				       (FLOOR (* 3. *GC-MAX-BUCKET-SIZE*)
					      2.)))))) 
		 ;; THIS LEG WILL FLIP GENERATION 0.
		 (WRITE-METER '%GC-SAFETY-PAD 0.)
		 (SETF PREVIOUS-GENERATION-HIGH NIL)
		 (WHEN (AND (ACTIVE-OVERSIZED-P (AREF USED-ACTIVE 3. 0))
			    (<= (- (+ (AREF USED-ACTIVE 3. 2.) (AREF USED-ACTIVE 3. 3.)) PDL)
				(FLOOR (* 3. *GC-MAX-BUCKET-SIZE*)
				       2.)))
		   (SHIFT-GEN-THREE))
		 (IF (< (AREF D 0) (* 2. B))
		     ;; GENERATION 0 SIZE IS UNDER CONTROL SO USE A NORMAL DELAYED SCAVENGE FLIP.
		     (GC-FLIP-NOW 1 (* 16. B))
		     ;; GENERATION 0 SIZE IS OUT OF CONTROL. THIS MUST BE THE FIRST GEN 0 COLLECTION
		     ;; AFTER A HIGHER GENERATION COLLECTION. IN THIS CASE WE WANT TO FIRE THE SCAVENGER
		     ;; IMMEDIATELY AND TRY TO AVOID CONS DRIVEN SCAVENGING IN THE GEN 0 COLLECTION.
		     ;; FIRST WE MUST CALCULATE THE MAX SAFE BIAS.
		     (COND ((= *GC-MAX-INCREMENTAL-GENERATION* 0)
			    (SETF BIAS (- RESERVE
					  (FLOOR (* 19. (AREF D 0))
						 16.))))
			   ((= *GC-MAX-INCREMENTAL-GENERATION* 1)
			    (SETF BIAS (- RESERVE
					  (FLOOR (+ (* 22. (AREF D 0))
						    (* 19. (AREF D 1)))
						 16.))))
			   ((= *GC-MAX-INCREMENTAL-GENERATION* 2)
			    (SETF BIAS (- RESERVE
					  (FLOOR (+ (* 25. (AREF D 0))
						    (* 22. (AREF D 1))
						    (* 19. (AREF D 2)))
						 16.))))
			   (T
			    (SETF BIAS (- RESERVE
					  (FLOOR (+ (* 28. (AREF D 0))
						    (* 25. (AREF D 1))
						    (* 22. (AREF D 2))
						    (* 19. (AREF USED-ACTIVE 3 3))
						    (AREF USED-ACTIVE 3 0)
						    (AREF USED-ACTIVE 3 1)
						    (AREF USED-ACTIVE 3 2)
						    (AREF D 4.))
						 16.)))))
		     (GC-FLIP-NOW 1 (MAX (* 16. BIAS) 1))
		     ;; WE MUST DIRECTLY SET THE BIAS TO BE ABLE TO GET A REALLY BIG ONE.
		     (WRITE-METER '%COUNT-SCAVENGER-WORK (MAX (* 16. BIAS) 0))
		     ;; FIRE OFF THE IDLE SCAVENGER IMMEDIATELY
		     (SETF %scavenger-ws-enable NIL)))
		(T
		 ;; THIS LEG WILL FLIP GENERATION 1 AND POSSIBLY GENERATIONS 2 AND 3.
		 (SETF PREVIOUS-GENERATION-HIGH 1)
		 ;; FLIP GENERATION 1 WITH PROMOTION.
		 (WRITE-METER '%GC-SAFETY-PAD 0)
		 (GC-FLIP-NOW 3 (MAX (* 16. BIAS) 1))
		 ;; WE MUST DIRECTLY SET THE BIAS TO BE ABLE TO GET A REALLY BIG ONE.
		 (WRITE-METER '%COUNT-SCAVENGER-WORK (MAX (* 16. BIAS) 0))
		 ;; FIRE OFF THE IDLE SCAVENGER IMMEDIATELY
		 (SETF %scavenger-ws-enable NIL)
		 (WHEN (> *GC-MAX-INCREMENTAL-GENERATION* 1.)
		   ;; WAIT IN THIS LEG SO WE CAN LOOK AT THE RESULTS
		   (PROCESS-WAIT "Await scavenge" 'SYMBOL-VALUE '%GC-FLIP-READY)
		   (GC-RECLAIM-OLDSPACE)
		   ;; GET UPDATED SIZE INFORMATION
		   (MULTIPLE-VALUE-SETQ (W D USED-ACTIVE PDL) (GET-GENERATION-SIZES))
		   (SETF USED-VIRTUAL-ADDRESS-SPACE 0)
		   (DOTIMES (I 8.) (INCF USED-VIRTUAL-ADDRESS-SPACE (AREF D I)))
		   (MULTIPLE-VALUE-SETQ (USABLE-SWAP-SPACE SWAP-SPACE-FREE) (SWAP-SPACE-INFO))
		   (SETF SWAP-SPACE-FREE	
			 (* (- SWAP-SPACE-FREE
			       (COUNT-PAGES-NEEDING-WRITABLE-PAGING-STORE (NOT EXTENDED-ADDRESS-SPACE)))
			    PAGE-SIZE))
		   (SETF RESERVE (- (MIN (- *max-virtual-address* USED-VIRTUAL-ADDRESS-SPACE 2097152.)
					 SWAP-SPACE-FREE)
				    S))
		   (COND ((= *GC-MAX-INCREMENTAL-GENERATION* 2)
			  (SETF THRESHOLD (FLOOR (- (* 16. (- (+ RESERVE
								 (AREF D 2))
							      B
							      C))
						    (* 25. (AREF D 0))
						    (* 22. (AREF D 1)))
						 35.))
			  (SETF BIAS (- RESERVE
					(FLOOR (* 19. (AREF D 2))
					       16.)))
			  (SETF (AREF FLIP-SIZE 2.) THRESHOLD) 
			  (WHEN (> (AREF D 2) THRESHOLD)
			    ;; FLIP GENERATION 2 (PROMOTION CONTROLLED BY GC-PROMOTE-GEN-2)
			    (SETF PREVIOUS-GENERATION-HIGH 2)
			    (IF GC-PROMOTE-GEN-2
				(GC-FLIP-NOW 5. (* 16. (MAX BIAS 1)))
				(GC-FLIP-NOW 4. (* 16. (MAX BIAS 1))))
			    ;; WE MUST DIRECTLY SET THE BIAS TO BE ABLE TO GET A REALLY BIG ONE.
			    (WRITE-METER '%COUNT-SCAVENGER-WORK (* 16. (MAX BIAS 0)))
			    ;; FIRE OFF THE IDLE SCAVENGER IMMEDIATELY
			    (SETF %scavenger-ws-enable NIL)))
			 (T
			  ;; THIS LEG IS EAS
			  (SETF THRESHOLD (FLOOR (- (* 16. (- (+ RESERVE
								 (AREF D 2))
							      B
							      C))
						    (* 28. (AREF D 0))
						    (* 25. (AREF D 1))
						    (* 19. (AREF USED-ACTIVE 3 3))
						    (AREF USED-ACTIVE 3 0)
						    (AREF USED-ACTIVE 3 1)
						    (AREF USED-ACTIVE 3 2)
						    (AREF D 4.))
						 38.))
			  (SETF BIAS (- RESERVE
					(FLOOR (+ (* 22. (AREF D 2))
						  (* 19. (AREF USED-ACTIVE 3 3))
						  (AREF USED-ACTIVE 3 0)
						  (AREF USED-ACTIVE 3 1)
						  (AREF USED-ACTIVE 3 2)
						  (AREF D 4.))
					       16.)))
			  (SETF (AREF FLIP-SIZE 2.) THRESHOLD)
			  (WHEN (OR (> (AREF D 2) THRESHOLD)
				    (AND (ACTIVE-OVERSIZED-P (AREF USED-ACTIVE 3. 0))
					 (> (- (+ (AREF USED-ACTIVE 3. 2.) (AREF USED-ACTIVE 3. 3.)) PDL)
					    (FLOOR (* 3. *GC-MAX-BUCKET-SIZE*)
						   2.))))
			    ;; FLIP GENERATION 2 WITH PROMOTION.
			    (WRITE-METER '%GC-SAFETY-PAD 0)
			    (GC-FLIP-NOW 5. (* 16. (MAX BIAS 1)))
			    ;; WE MUST DIRECTLY SET THE BIAS TO BE ABLE TO GET A REALLY BIG ONE.
			    (WRITE-METER '%COUNT-SCAVENGER-WORK (* 16. (MAX BIAS 0)))
			    ;; FIRE OFF THE IDLE SCAVENGER IMMEDIATELY
			    (SETF %scavenger-ws-enable NIL)
			    ;; WAIT IN THIS LEG SO WE CAN LOOK AT THE RESULTS
			    (PROCESS-WAIT "Await scavenge" 'SYMBOL-VALUE '%GC-FLIP-READY)
			    (GC-RECLAIM-OLDSPACE)
			    ;; GET UPDATED SIZE INFORMATION
			    (MULTIPLE-VALUE-SETQ (W D USED-ACTIVE PDL) (GET-GENERATION-SIZES))
			    (SETF USED-VIRTUAL-ADDRESS-SPACE 0)
			    (DOTIMES (I 8.) (INCF USED-VIRTUAL-ADDRESS-SPACE (AREF D I)))
			    (MULTIPLE-VALUE-SETQ (USABLE-SWAP-SPACE SWAP-SPACE-FREE) (SWAP-SPACE-INFO))
			    (SETF SWAP-SPACE-FREE	
				  (* (- SWAP-SPACE-FREE
					(COUNT-PAGES-NEEDING-WRITABLE-PAGING-STORE (NOT EXTENDED-ADDRESS-SPACE)))
				     PAGE-SIZE))
			    (SETF RESERVE (- (MIN (- *max-virtual-address* USED-VIRTUAL-ADDRESS-SPACE 2097152.)
						  SWAP-SPACE-FREE)
					     S))
			    (SETF THRESHOLD (FLOOR (- (* 16. (- (+ RESERVE
								   (AREF USED-ACTIVE 3 3))
								B
								C))
						      (* 28. (AREF D 0))
						      (* 25. (AREF D 1))
						      (* 22. (AREF D 2))
						      (AREF USED-ACTIVE 3 0)
						      (AREF USED-ACTIVE 3 1)
						      (AREF USED-ACTIVE 3 2)
						      (AREF D 4.))
						   35.))
			    (SETF BIAS (- RESERVE
					  (FLOOR (+ (* 19. (AREF USED-ACTIVE 3 3))
						    (AREF USED-ACTIVE 3 0)
						    (AREF USED-ACTIVE 3 1)
						    (AREF USED-ACTIVE 3 2)
						    (AREF D 4.))
						 16.)))
			    (SETF (AREF FLIP-SIZE 3.) THRESHOLD)
			    (WHEN (OR (> (MAX (AREF USED-ACTIVE 3. 3.)
					      (AREF USED-ACTIVE 3. 2.)
					      (AREF USED-ACTIVE 3. 1.)
					      (AREF USED-ACTIVE 3. 0.))
					 THRESHOLD)
				      (AND (ACTIVE-OVERSIZED-P (AREF USED-ACTIVE 3. 0))
					   (> (- (+ (AREF USED-ACTIVE 3. 2.) (AREF USED-ACTIVE 3. 3.)) PDL)
					      (FLOOR (* 3. *GC-MAX-BUCKET-SIZE*)
						     2.))))
			      (WRITE-METER '%GC-SAFETY-PAD 0)
			      ;; FLIP GENERATION 3 WITH EXPORT
			      (GC-FLIP-NOW 7. (* 16. (MAX BIAS 1)))
			      ;; WE MUST DIRECTLY SET THE BIAS TO BE ABLE TO GET A REALLY BIG ONE.
			      (WRITE-METER '%COUNT-SCAVENGER-WORK (* 16. (MAX BIAS 0)))
			      ;; FIRE OFF THE IDLE SCAVENGER IMMEDIATELY
			      (SETF %scavenger-ws-enable NIL)
			      ;; SHIFT GENERATION 3 
			      (SHIFT-GEN-THREE)))))))))))))
))


#!C
; From file GC.LISP#> MEMORY-MANAGEMENT; Hotel:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MEMORY-MANAGEMENT; GC.#"

;;HRC 6/27/90.  Add new GC-PROCESS vars.  Remove *adaptive-training-enabled* (redundant).
(DEFVAR gc-status-var-list
	'(*gc-report-stream*
	   *gc-notifications*
	   *gc-max-incremental-generation* 
	   *gc-console-delay-interval*
	   *GC-CONS-GUARANTEE*
	   *GC-ADDRESS-SPACE-SAFETY-PAD*))

(DEFUN gc-status-auto-gc (&key (STREAM *standard-output*) (verbose nil)
			  &aux tem)
  (DECLARE (IGNORE verbose))
  (FORMAT stream "~%AUTOMATIC GC STATUS~
                   ~%-------------------")
  (COND ((gc-active-p)
	 (FORMAT stream "~%Automatic GC is ON.")
	 (FORMAT stream "~%Garbage collection of generation ~d (~[without~;with~] promotion) now in progress"
		 (FLOOR gc-type-of-flip 2) (REM gc-type-of-flip 2))
	 (IF (SETQ tem (gc-arrest-reasons))
	     (FORMAT stream ",~%but has been arrested for reason~p: ~a."
		     (LENGTH tem) tem)
	     (FORMAT stream ".")))
	(t (FORMAT stream "~%Automatic GC is OFF.")))
  (FORMAT stream "~%Adaptive Training is ~a"
		 (IF *adaptive-training-enabled* "ENABLED" "DISABLED"))
  (IF (training-active)
      (LET ()
	(IF *adaptive-training-enabled*
	    (FORMAT stream " and ACTIVE.")
	    (FORMAT stream " but is currently still ACTIVE.")))
      (IF *adaptive-training-enabled*
	  (FORMAT stream " but is currently INACTIVE (until the next garbage collection begins).")
	  (FORMAT stream " and INACTIVE.")))
  (WHEN EXTENDED-ADDRESS-SPACE
    (LET ((NUMBER-OF-WORLDS (IF (LISTP EXTENDED-ADDRESS-SPACE)
				(LENGTH EXTENDED-ADDRESS-SPACE)
				0)))
      (FORMAT stream "~%Extended Address Space is active with ~d. external worlds."
	      NUMBER-OF-WORLDS)
      (FORMAT stream "~%There have been ~d. shifts of the active generation 3 domains." *gc-active-shift-count*))))

(DEFUN gc-status-batch-gc (&key (STREAM *standard-output*) (verbose nil))
  (LET ((space-size-struct (get-space-size-info *tem-space-size-info*))
	space-needed space-free distance)
    (FORMAT stream "~2%BATCH GC STATUS~
                     ~%---------------")
    (COND
      ((AND (gc-in-progress-p)
	    gc-batch-mode)
       (FORMAT stream "~%A ~a is now in progress."
	       (COND ((EQ gc-batch-mode :full) "FULL-GC")
		     ((EQ gc-batch-mode :immediate) "GC-IMMEDIATELY")
		     (t "GC-IMMEDIATELY")))
       (FORMAT stream "~%Currently collecting generation ~a ~[without~;with~] promotion."
	       (FLOOR gc-type-of-flip 2) (REM gc-type-of-flip 2))) 
      ;; Batch GC not in progress
      (t
       ;; FULL-GC :promote t
       (MULTIPLE-VALUE-SETQ (space-needed space-free)
	 (get-space-needed-for-gc :max-gen 3 :promote t :space-size-struct space-size-struct))
       (SETQ distance (- space-free space-needed))
       (FORMAT stream "~%(FULL-GC :max-gen 3 :promote t)           ~a. ~a."
	       (IF (PLUSP distance)
		   (FORMAT nil "OK for another ~a" (gc-marked-size-rj distance))
		   (FORMAT nil "~a too late" (gc-marked-size-smallest (- distance))))  
	       (FORMAT nil "~a needed" (gc-marked-size-rj space-needed)))
       ;; FULL-GC :promote nil
       (MULTIPLE-VALUE-SETQ (space-needed space-free)
	 (get-space-needed-for-gc :max-gen 3 :promote nil :space-size-struct space-size-struct))
       (SETQ distance (- space-free space-needed))
       (FORMAT stream "~%(GC-IMMEDIATELY :max-gen 3 :promote nil)  ~a. ~a."
	       (IF (PLUSP distance)
		   (FORMAT nil "OK for another ~a" (gc-marked-size-rj distance))
		   (FORMAT nil "~a too late" (gc-marked-size-smallest (- distance))))  
	       (FORMAT nil "~a needed" (gc-marked-size-rj space-needed)))
       ;; GC-IMMEDIATELY :max-gen 2 :promote nil
       (MULTIPLE-VALUE-SETQ (space-needed space-free)
	 (get-space-needed-for-gc :max-gen 2 :promote nil :space-size-struct space-size-struct))
       (SETQ distance (- space-free space-needed))
       (FORMAT stream "~%(GC-IMMEDIATELY :max-gen 2 :promote nil)  ~a. ~a."
	       (IF (PLUSP distance)
		   (FORMAT nil "OK for another ~a" (gc-marked-size-rj distance))
		   (FORMAT nil "~a too late" (gc-marked-size-smallest (- distance))))  
	       (FORMAT nil "~a needed" (gc-marked-size-rj space-needed)))
       (WHEN verbose
	 ;; Training session.  All gen 3 static regions will become dynamic.
	 (MULTIPLE-VALUE-SETQ (space-needed space-free)
	   (get-space-needed-for-gc :max-gen 3  :promote nil :static-regions-are-dynamic t
				    :space-size-struct space-size-struct))
	 (SETQ distance (- space-free space-needed))
	 (FORMAT stream "~%(SYS:START-TRAINING-SESSION T)             ~a. ~a."
	       (IF (PLUSP distance)
		   (FORMAT nil "OK for another ~a" (gc-marked-size-rj distance))
		   (FORMAT nil "~a too late" (gc-marked-size-smallest (- distance))))  
	       (FORMAT nil "~a needed" (gc-marked-size-rj space-needed))))))))
))

(eval-when (load)
  (si:gc-off-temporarily-back-on))
