;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Set up printer-graphics-window based on mono or color default-screen.
;;; Also, create one of these on the microExplorer, since previous change
;;; caused one not to be created.  KJF, PMH.
;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;; Written 05/01/88 10:38:33 by FISCHER,
;;; while running on F-ASTAIRE from band LODA
;;; With SYSTEM 4.59, VIRTUAL-MEMORY 4.4, EH 4.5, MAKE-SYSTEM 4.5, MICRONET 4.4, LOCAL-FILE 4.1,
;;;  BASIC-PATHNAME 4.12, NETWORK-SUPPORT-COLD 4.1, NAMESPACE 4.19, NETWORK-NAMESPACE 4.2,
;;;  DISK-IO 4.10, DISK-LABEL 4.0, BASIC-FILE 4.6, MAC-PATHNAME 4.3, NETWORK-PATHNAME 4.0,
;;;  COMPILER 4.10, TV 4.83, DATALINK 4.14, CHAOSNET 4.18, GC 4.3, MEMORY-AUX 4.0,
;;;  NVRAM 4.6, SYSLOG 4.3, STREAMER-TAPE 4.4, UCL 4.1, INPUT-EDITOR 4.0, METER 4.3,
;;;  ZWEI 4.18, DEBUG-TOOLS 4.2, NETWORK-SUPPORT 4.5, NETWORK-SERVICE 4.0, DATALINK-DISPLAYS 4.0,
;;;  FONT-EDITOR 4.0, SERIAL 4.0, PRINTER 4.7, PRINTER-TYPES 4.2, IMAGEN 4.0, SUGGESTIONS 4.0,
;;;  MAIL-DAEMON 4.7, MAIL-READER 4.6, TELNET 4.1, VT100 4.4, NAMESPACE-EDITOR 4.5,
;;;  PROFILE 4.4, VISIDOC 4.4, IP 3.14, Experimental BUG 10.0, GRAPHICS-WINDOW 4.0,
;;;   microcode 306, Band Name: pre 4.1 + IP 4/25

#!C
; From file PRINT.LISP#> GWIN; SYS:
#10R GWIN#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "GWIN"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: GWIN; PRINT.#"



(DEFVAR color-printer-graphics-window nil)
(DEFVAR mono-printer-graphics-window nil)

(DEFUN setup-printer-graphics-window ()
  (IF (color-sheet-p printer-graphics-window)
      (SETQ color-printer-graphics-window printer-graphics-window)
      (SETQ mono-printer-graphics-window printer-graphics-window))
  (IF (color-sheet-p default-screen)
      (UNLESS (color-sheet-p printer-graphics-window)
	(SETQ printer-graphics-window
	      (OR color-printer-graphics-window
		  (SETQ color-printer-graphics-window
			(make-instance 'gwin-cache-window
				       :bit-array
				       (MAKE-ARRAY '(1024 1024)
						   :element-type (tv:sheet-array-type-cl default-screen))
				       :width 1024
				       :height 1024)))))
      ;; else formS
      (WHEN (color-sheet-p printer-graphics-window)
	(SETQ printer-graphics-window
	      (OR mono-printer-graphics-window
		  (SETQ mono-printer-graphics-window
			(make-instance 'gwin-cache-window
				       :bit-array
				       (MAKE-ARRAY '(3008 3008)
						   :element-type (tv:sheet-array-type-cl default-screen))
				       :width 3008
				       :height 3008)))))))

))

#!C
; From file PRINT.LISP#> GWIN; SYS:
#10R GWIN#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "GWIN"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: GWIN; PRINT.#"


(defun bitmap-of-picture-file (picture-file
			       &optional printer-dpi)
  "Given a picture file, returns its bitmap array, width and height.  If dpi of printer
 is given, scaling is done"
  
  (declare (values picture-array width height))

  ;; Set up window based on color or mono default-screen.  04/29/88 KJF
  ;; Added during addition of multiple screen support (Rel 4.1).
  (setup-printer-graphics-window)

  (let ((world (send printer-graphics-window :world))
	transform
	x-transform
	y-transform
	(x-scale (if printer-dpi (quotient printer-dpi lm-x-pixels-per-inch) 1))
	(y-scale (if printer-dpi (quotient printer-dpi lm-y-pixels-per-inch) 1))
	width
	height
	left
	top
	right
	bottom
	left-edge
	top-edge
	right-edge
	bottom-edge)
    
    (send world :read-display-list picture-file)
    (send printer-graphics-window :clear-screen)
    
    ;; Make sure the initial transform is 1
    
    (send printer-graphics-window :set-scales 1 1)
    
    ;; Get the current extents of the picture in world coordinates
    
    (multiple-value-setq (bottom left right top)
      (send world :calculate-extents))
    
    ;; Get the edges of the window in world coordinates
    
    (multiple-value-setq (left-edge top-edge right-edge bottom-edge)
      (send printer-graphics-window :world-edges))
    
    ;; Adjust the scale of the window so that the whole picture will fit in the
    ;; standard window.
    
    (if (or (>= right right-edge)
	    (>= bottom bottom-edge)
	    (<= left left-edge)
	    (<= top top-edge))
      (send printer-graphics-window :world-extents-window))
    
    ;; Set the transform according to the printer scale
    
    (setq transform (send printer-graphics-window :transform))
    (setq x-transform (aref transform 0 0))
    (setq y-transform (aref transform 1 1))
    (send printer-graphics-window :set-scales (* x-transform x-scale)
	                                      (* y-transform y-scale))
    
    ;; Now that the transform has been set, draw the picture
    
    (send printer-graphics-window :draw-picture-list
	                          (send world :display-list))
    
    ;; Get the new extents of the picture in world coordinates
    
    (setq right  (send world :right-limit)
	  bottom (send world :bottom-limit))
    
    ;; The height and width are returned in normal window coordinate values
    
    (multiple-value-setq (width height)
      (send printer-graphics-window :transform-point right bottom))
    
    (values (TV:sheet-screen-array printer-graphics-window) width height)))
))

;; This changes previous change for microExplorer which did not make a
;; printer-graphics-window.  Since the bitmap-of-picture-file function
;; can be used even if not printing picture files on a printer, and the
;; function uses this window, then we really ought to create one.
(WHEN (tv:mac-system-p)
  (SETQ gwin:*printer-graphics-window-dimensions* '(1024 1024))
  (SETQ gwin:printer-graphics-window
	(make-instance 'gwin:gwin-cache-window
		       :bit-array (make-array gwin:*printer-graphics-window-dimensions*
					      :element-type (tv:sheet-array-type-cl tv:default-screen))
		       :width (FIRST gwin:*printer-graphics-window-dimensions*)
		       :height (SECOND gwin:*printer-graphics-window-dimensions*))))