;;; -*- Mode: Common-Lisp; Package: MAIL; Base: 10.; Patch-File: T -*-
;;; Patch file for MAIL-DAEMON version 6.2
;;; Reason: Added ignore-errors and with-timeout around :Close connection in smtp-mail-server to handle hang and error trap problems with host that close connections first.
;;; Written 05/22/89 08:14:30 by BERGER,
;;; while running on ARIES from band LOD1
;;; With SYSTEM 6.1, VIRTUAL-MEMORY 6.0, EH 6.0, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.0, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.0, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.0, TV 6.3, DATALINK 6.0, CHAOSNET 6.0, GC 6.0, MEMORY-AUX 6.0, NVRAM 6.0,
;;;  SYSLOG 6.0, STREAMER-TAPE 6.0, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.0,
;;;  DEBUG-TOOLS 6.0, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.0, SERIAL 6.0, PRINTER 6.0, MAC-PRINTER-TYPES 6.0, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.0, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.0, TI-CLOS 6.0, CLEH 6.0, IP 3.45,
;;;  Experimental BUG 11.4, Experimental CLX 6.0, Experimental CLUE 21.0, Experimental X11M 4.0,
;;;  DECNET 1.65, VISIDOC-SERVER 5.0,  microcode 427, Band Name: Rel 6.0 + SLE 5/12

#!C
; From file SERVER.LISP#> MAIL-DAEMON; SYS:
#10R MAIL#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MAIL"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MAIL-DAEMON; SERVER.#"


(defun SMTP-MAIL-SERVER (&optional stream)
  
  (let ((state :send-reply)
	(cleanup nil)
	(error nil)
	(*read-base* 10)
	(*print-base* 10)
	(*log-session* (incf *log-session*))
	(user-id "Mailer")
	(command-string (allocate-xstring 4))
	command line message server-object foreign-host conn)
    (declare (special sys:*new-line-delimiter* )) ;05-06-88 nn
    
    (setf (fill-pointer command-string) 4)
    
    (unwind-protect
	(condition-call-if (not *debug-mailer*) (condition)
	    (block session
	      (unless stream
		(setq stream
		      (host:listen-for-connection-on-medium
			:byte-stream "SMTP"
			:stream-type :ascii-translating-character-stream)))
	      ;; SPR#4511 -- sometimes we get NIL instead of a stream  \/\/
	      (unless stream
		(return-from session))
	      (setq cleanup t)
	      (send si:current-process :set-priority -7)
	      (let ((reject-reason (reject-mail-connection-p)))
		(when reject-reason
		  (smtp-reply 421 stream reject-reason)
		  (setq state :quit)
		  (log-debug :smtp-server "Rejected connection: ~A" reject-reason)
		  (return-from session)))
	      
	      (smtp-reply 220 stream)
	      (setq foreign-host (or (send stream :send-if-handles :foreign-host) "UNKNOWN"))
	      (setq conn (send stream :send-if-handles :connection))
	      (setq server-object (make-instance 'smtp-server-object :conn conn :foreign-host foreign-host))
	      (log-debug :smtp-server "SMTP serving ~A" foreign-host)
	      ;; May want to conditionalize the who line for Chaos vs. TCP serving
	      (send w:who-line-file-state-sheet :add-server server-object "SMTP")
	      (let-if (member (send foreign-host :system-type) '(:unix :unix-ucb))  ;05-06-88 nn
		      ((sys:*new-line-delimiter* '(#\newline #\linefeed)))
		(loop
		  (setq state :command-wait)
		  (setq line (read-line-with-timeout stream))
		  (setq state :send-reply)
		  (setq command (and (> (length line) 3)
				     (copy-array-portion line 0 4 command-string 0 4)
				     (intern (nstring-upcase command-string) *utility-package*)))
		  (log-debug :smtp-server "<- ~A" line)
		  (case command
		    (:HELO
		     (when message
		       ;; Must start with a clean message
		       (deallocate-message message))
		     (setq message (allocate-message))
		     (when (> (length line) 5)
		       (setq foreign-host (subseq line 5 (position #\Space line :start 5))))
		     (smtp-reply 250 stream))
		    (:MAIL
		     (let* ((address-start (position #\: (the string line)))
			    (address (and address-start (parse-address line (1+ address-start) nil nil :route-address))))
		       ;;? should verify return address?
		       (cond (address
			      (when message
				;; Must start with a clean message
				(deallocate-message message))
			      (setq message (allocate-message))
			      (send message :set-return-path (send address :add-local-host))
			      (smtp-reply 250 stream))
			     (t
			      (smtp-reply 501 stream)))))
		    
		    (:RCPT
		     (cond (message
			    (let* ((address-start (position #\: (the string line)))
				   (address (and address-start
						 (parse-address line (1+ address-start) nil nil :route-address))))
			      (when address
				(setq address (send address :remove-local-host)))
			      (cond ((send address :accept-p)
				     (send message :add-recipient address)
				     (smtp-reply 250 stream))
				    (address
				     (smtp-reply 550 stream (send address :rejection-report-string)))
				    (t
				     (smtp-reply 501 stream "No address found in RCPT command.")))))
			   (t
			    (smtp-reply 503 stream "SMTP RCPT command must be preceeded by a MAIL command"))))
		    
		    (:DATA
		     (log-debug :smtp-server "<- [Receive text]")
		     (cond (message
			    (send message :append-received-line foreign-host
				  (or (send stream :send-if-handles :network-type) "UNKNOWN")
				  "SMTP")
			    (smtp-reply 354 stream)
			    (setq state :data)
			    (loop
			      with headers-end and header-type
			      with hop-count = 0
			      for line = (read-line-with-timeout stream nil nil)
			      while line
			      do
			      (cond ((or (= (length line) 0)
					 (not (eql (char line 0) #\.)))
				     (send message :line-out line))
				    ;; Line starts with a dot -- if a "quoted dot" drop it
				    ((> (length line) 1)
				     (send message :line-out line 1))
				    ;; End of data
				    (t
				     (log-debug :smtp-server "<- [Received <CRLF>.<CRLF>]")
				     (return)))
			      (setq headers-end (or headers-end (string-blank-p line)))
			      (when (not headers-end)
				(setq header-type (header-line-type line))
				(cond ((eq header-type :received)
				       (when (> (incf hop-count) *max-message-network-hops*)
					 (send message :set-loop-detected-p t)))
				      ((eq header-type :subject)
				       (send message :set-subject (header-line-body-string line))))))
			    (setq state :send-reply)
			    (log-debug :smtp-server "Enqueue message.")
			    ;; Check if any recipients (client might be stupid and send data
			    ;;  even if no recipients were valid)
			    (let ((result (and (send message :address-list)
					       (enqueue-message message))))
			      (cond  ((errorp result)
				      (log-debug :smtp-server "Queue error: ~A" result)
				      ;; Assume errors at this point are transient
				      (smtp-reply 421 stream (sanitize-string (send result :report-string)))
				      ;;(setq state :quit) (setq message nil) (return-from session)
				      )
				     (t
				      (smtp-reply 250 stream)))
			      (setq message nil)))
			   (t
			    (smtp-reply 503 stream "SMTP DATA command must be preceeded by MAIL and RCPT commands"))))
		    
		    (:RSET
		     (when message
		       (deallocate-message message))
		     (setq message nil)
		     (smtp-reply  250 stream))
		    
		    (:VRFY
		     (cond ((> (length line) 5)
			    (let* ((address (parse-address line 5 nil nil :route-address))
				   (ok (and address (send address :accept-p))))
			      (if ok
				  (smtp-reply 250 stream (send address :string-for-message))
				  (smtp-reply 550 stream (send address :rejection-report-string)))))
			   (t
			    (smtp-reply 501 stream "No address supplied for VRFY command"))))
		    
		    (:NOOP
		     (smtp-reply 250 stream))
		    
		    (:QUIT
		     (when message
		       (deallocate-message message))
		     (setq message nil)
		     (smtp-reply 221 stream)
		     (setq state :complete)
		     (return-from session))
		    
		    (:HELP
		     (smtp-reply 214 stream))
		    
		    ((:SEND :SOML :SAML :EXPN :TURN)
		     (smtp-reply 502 stream))
		    
		    (:OTHERWISE
		     (smtp-reply 500 stream))
		    ))))
	  
	  ;; condition-call clauses
	  ((handle-condition-p condition)
	   (when message
	     (deallocate-message message))
	   (cond ((ignored-network-condition-p condition)
		  (log-debug :smtp-server "Error caught and ignored: ~A" condition)
		  (setq cleanup nil))
		 (t
		  (setq cleanup t)
		  (setq error t)
		  (when (condition-typep condition 'error)
		    (mailer-error :smtp-server "SMTP server error while serving host ~A~%~A"
				  foreign-host (send condition :report-string)))))))
      
      ;; unwind-protect cleanup forms
      (when (and cleanup (neq state :complete) (streamp stream))
	(case state
 	  (:data
	   ;; Drain the data until we can send a reply
	   (loop until (null line)
		 until (and (= (length line) 1)
			    (eql (char line 0) #\.))
		 do
		 (setq line (read-line-with-timeout stream nil nil)))
	   (setq state :send-reply))
	  ((:command-wait :quit)
	   (setq line (read-line-with-timeout stream nil nil))))
	;; At this point must be in a state to send a reply
	(loop
	  with count = 1
	  while line
	  do
	  (when (or (eq state :complete) (string-equal line "QUIT" :start1 0 :end1 4))
	    (smtp-reply 221 stream)
	    (return))
	  (if error
	      (smtp-reply 421 stream "~A closing transmission channel.  Local error.")
	    (smtp-reply 421 stream "~A closing transmission channel.  Need QUIT command." si:local-host))
	  ;; Give up after 10 attempts to quit
	  (when (>= (incf count) 10)
	    (return))
	  (setq line (read-line-with-timeout stream nil nil))))
      (when (and (streamp stream)
		 (TYPEP (SEND stream :connection) 'ip:tcp-connection)) ;11-11-88 DAB make sure it is a valid connection.
	  ;; DAB 05-19-89 Some hosts may shutdown connection before receiving a connection close from us. In this 
          ;; case attempt to close it , after 30 seconds give up, close connection aborted.
	(ignore-errors
	  (with-timeout ((* 60 30) (unless (not cleanup) (send stream :close T)))
	    (send stream :close (not cleanup))))
	)
      
      (when server-object
	(send w:who-line-file-state-sheet :delete-server server-object)))
    (force-log-output)))
))
