;;; -*- Mode: Common-Lisp; Package: MAIl; Base: 10.; Patch-File: T -*-

;;; Reason: Added locking macro in log-output and :after :force-output to keep mail servers from stomping over each other.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for MAIL-DAEMON version 6.3
;;; Written 10/04/89 08:03:32 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 6.17, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.1, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.2, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.3, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.12, TV 6.15, DATALINK 6.0, CHAOSNET 6.1, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.5,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.4, TI-CLOS 6.24, CLEH 6.5, IP 3.50,
;;;  Experimental CLX 6.3, CLUE 6.17, X11M 6.14, Experimental BUG 11.15, VISIDOC-SERVER 6.1,
;;;   microcode 429, Band Name: Rel 6.0 + SLE 8/30

#!C
; From file DEFINITIONS.LISP#> MAIL-DAEMON; SYS:
#10R MAIL#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MAIL"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MAIL-DAEMON; DEFINITIONS.#"



(defvar log-output-lock nil   ; DAB 10-04-89
  "Locking variable used in LOG-OUTPUT to make sure buffers do not get changed out from under it.")

(defun LOG-OUTPUT (catagory format-string &rest format-args)
  (when *log-output*
    (with-lock (log-output-lock )  ; DAB 10-04-89
      (send *log-output* :send-if-handles :set-more-p nil)
      (fresh-line *log-output*)
      (time:print-current-time *log-output*)
      (format *log-output* " ~3D~@[ ~S~]  " *log-session* catagory)
      (apply #'format *log-output* format-string format-args))))



))

#!C
; From file MAILER.LISP#> MAIL-DAEMON; SYS:
#10R MAIL#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MAIL"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MAIL-DAEMON; MAILER.#"


(defmethod (LOG-STREAM :AFTER :FORCE-OUTPUT) ()
  
  ;;May want to defer writing when buffer is small because each append to a file uses one block minimum.  
  ;;Appending lots of small buffers really eats up disk space.
  
  (when (> buffer-index 0)
    (let ((unlock-on-exit nil)
	  (if-exists-action (if (> (get-universal-time) next-new-version-time) :new-version :append)))
      (unwind-protect
	  (condition-call-if (not *debug-mailer*) (condition)
	      (block do-output
		;;Check/set lock and swap buffers without interruption
		(without-interrupts
		  (when output-lock
		    (return-from do-output))
		  (setq output-lock t)
		  (setq unlock-on-exit t)
		  (setq old-buffer
			(prog1 buffer (setq buffer old-buffer)))
		  (setq old-buffer-index
			(prog1 buffer-index (setq buffer-index 0))))
		(when (and (file-system-usable-p)
			   user-id (not (equal user-id ""))
			   (send (send pathname :host) :local-host-p))
		  (with-lock (log-output-lock)   ; DAB 10-04-89
		    (with-open-file (out pathname :direction :output :if-exists if-exists-action
					 :if-does-not-exist :create :error nil)
		      (unless (errorp out)
			(send out :string-out old-buffer 0 old-buffer-index)
			(when overflow-p
			  (format out "~2%>>>> Log buffer overflew.  ~D bytes of logging information lost.~2%" bytes-lost)
			  (setq overflow-p nil)))))
		  ;; Delete excess version if just created a new one.
		  (when (eq if-exists-action :new-version)
		    (incf next-new-version-time (* 24 60 60))
		    (let ((dirlist (fs:directory-list (send pathname :new-version :wild) :sorted :deleted :noerror)))	
		      (when (listp dirlist)
			;; Elimitate the directory entry from list
			(setq dirlist (delete (assoc nil dirlist) dirlist))
			(dotimes (x (- (length dirlist) versions-kept))
			  (let ((path (car (nth x dirlist))))
			    (send path :delete nil)
			    ;; Expunge if it is allowed.
			    (when (send path :undeletable-p)
			      (send path :expunge :error nil))))))))
		(setq overflow-p nil))
	    ((handle-condition-p condition)
	     condition))
	;; Unwind protect forms
	(when unlock-on-exit
	  (setq output-lock nil))))))
))
