;;; -*- Mode: Common-Lisp; Package: mail; Base: 10.; Patch-File: T -*-

;;; Reason: Modified :accept-p to forward incoming mail to primary-mail-server when local host can not develop the mail, unless local-host is a primary-mail-server. [10940]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for MAIL-DAEMON version 6.5
;;; Written 02/12/90 07:44:02 by BERGER,
;;; while running on Pasteur from band LOD2
;;; With SYSTEM 6.28, VIRTUAL-MEMORY 6.3, EH 6.6, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.3, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-FILE 6.7, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.23, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.12,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.4, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.7, TI-CLOS 6.37, CLEH 6.5, IP 3.57,
;;;  Experimental CLX 6.8, CLUE 6.48, X11M 6.20, Experimental BUG 11.18,  microcode 648,
;;;  Band Name: rel6.0 1/23

#!C
; From file ADDRESS.LISP#> MAIL-DAEMON; SYS:
#10R MAIL#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MAIL"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MAIL-DAEMON; ADDRESS.#"


(defmethod (ADDRESS :ACCEPT-P) ()

  ;; For servers to use in deciding whether to accept an address.
  ;; Acts just like :verify except obeys the :non-local-addresses value
  ;; of the reject-mail host option.
  ;; This does *not* heed the use-primary-mail-servers
  ;; option, which is intended only for decisions about mail
  ;; submitted by a user, not by a server.

  (cond ((and (eq (reject-mail) :non-local-addresses)
	      (not (send self :local-p)))
	 (setf (get self :rejection-report-string) "This machine only accepts local mail.  Cannot forward.")
	 nil)
	((not (send self :verify))
	 ;; ; DAB 02-09-90
         ;;When mail is delivered to an Explorer-resident mailing list from outside, the
         ;;behavior is often different from when the mail originates on the Explorer.
         ;;Most notably, there are certain message-routings which will *succeed* if sent
         ;;while sitting at the Explorer, but which will *fail* if sent through the
         ;;Explorer from another machine.
         ;; The fix below will attempt to forward this to it primary mail server, unless of course he is one. [10940]
	 (if (and (not (member (send si:local-host :name) (get-site-option :primary-mail-servers) :test 'string-equal))
		  (if *forward-unknown-addresses* 
		      (neq (use-primary-mail-servers) :never)
		      (eq (use-primary-mail-servers) :unknown-addresses)))
	     ;; Keep address, make note of forwarding
	     (progn (send self :set-delivery-status :forward-unknown)
		    (string-append (send self :verification-report-string)
			    (format () " Mail being forwarded to primary mail servers ~a"
				    (get-site-option :primary-mail-servers)))
		    T)	 
	     (setf (get self :rejection-report-string) (send self :verification-report-string))
	     nil))
	(t t)))
))
