;;; -*- Mode: Common-Lisp; Package: mail; Base: 10.; Patch-File: T -*-
;;; Written 03/27/90 12:33:16 by BERGER,
;;; Reason: Mailer-top-level was merging in a bad file type causing the WORK files not to be found. [11042]
;;; while running on MX29 from band n817
;;; With SYSTEM 6.31, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.2,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-PATHNAME 6.4, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.10, RPC 6.2, NFS-MX 6.6, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.24, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.14, METER 6.1, ZWEI 6.13, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.42, CLEH 6.5, NETWORK-PATHNAME 6.1,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.60, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.5, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.72, VISIDOC 6.7, PROFILE 6.2,
;;;   microcode 138, Band Name: microExplorer Network (8/17)

#!C
; From file MAILER.LISP#> MAIL-DAEMON; sys:
#10R MAIL#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MAIL"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MAIL-DAEMON; MAILER.#"


(defun MAILER-TOP-LEVEL (&optional (mode :boot) (read-queue t) check-routing log-pathname)
  
  (when (eq mode :boot)
    ;; After boot, wait 10 minutes for the world to settle down
    (process-sleep (* 60 60 10) "Pause"))
  
  (let ((user-id "Mailer")
	(*log-session* (incf *log-session*))
	(deliver t)
	(probe t)
	(new-items-only nil)
	(route-check check-routing)
	(last-route-check (get-universal-time))
	(last-delivery (get-universal-time))
	(last-probe (get-universal-time))
	utime)
    (wait-till-delivery-ok)
    (condition-call-if (not *debug-mailer*) (condition)     ; DAB 05-08-89 Catch error here too.
	(let ((dir (mailer-directory)))
	  (unless (probe-file (send dir :directory-pathname-as-file))
	    (fs:create-directory dir)))
      ((handle-condition-p condition)
       (mailer-error :mail-queue "Mailer error in top level loop.~%~A" condition)))
    (condition-call (condition)
	(when (eq mode :boot)
	  (when (directory "LM:MAILER;*.QUEUE")
	    (notify-nicely nil "~&The MAILER directory on this machine contains queued messages in Release 2 format.~
~%Use the function MAIL:CONVERT-REL2-MAILER-FILES to convert these files to Release 3 format.~
~%Otherwise, the queued messages cannot be delivered while running a Release 3 band.")))
      ((handle-condition-p condition)
       ))
    (condition-call-if (not *debug-mailer*) (condition)
	(progn 
	 (when (and log-pathname *log-enabled*)
	   (force-log-output)
	   (setq *log-output* (make-instance 'log-stream :pathname log-pathname))
	   (log-event :mail-queue ">>>>Mail Daemon started.  Mode: ~A, Read queue: ~A." mode read-queue))
	 (when read-queue
	   (send *mail-queue* :reuse)
	   ;; Initialize with queue for local host at the front to speed local delivery.
	   (send *mail-queue* :set-host-queue-list (list (make-host-queue :contact si:local-host)))
	   (let ((dir (mailer-directory)))
	     (unless (probe-file (send dir :directory-pathname-as-file))
	       (fs:create-directory dir))
	     (send *mail-queue* :restore (fs:directory-list (send dir :new-type "work") ; DAB 03-27-90 Make type a string.
							    :sorted)))
	   (setq check-routing nil)))
      ((handle-condition-p condition)
       (mailer-error :mail-queue "Error in MAILER-TOP-LEVEL initialization: ~A" condition)))
    (loop
      (condition-call-if (not *debug-mailer*) (condition)
	  (block top-loop
	    (when deliver
	      (when (send (mailer-directory) :undeletable-p)
		(send (mailer-directory) :expunge))
	      (when route-check
		(send *mail-queue* :check-routing)
		(setq last-route-check (get-universal-time)))
	      (log-debug :delivery "Begin queue delivery.")
	      (send *mail-queue* :deliver new-items-only)
	      (log-debug :delivery "Queue delivery complete.")
	      (force-log-output)
	      (unless new-items-only
		(setq last-delivery (get-universal-time))))
	    (when (and probe
		       ;;  check on consol idle time and limit probe interval. spr5989 - dc/nn
		       (time-lessp (time) (+ tv:kbd-last-activity-time (* 30 60 60))))
	      ;; Probe inboxes.  Don't get stuck here; give up after 2 minutes.
	      (with-timeout ((* 2 60 60))
		(probe-inboxes))
	      (setq last-probe (get-universal-time)))
	    (force-log-output))
	((handle-condition-p condition)
	 (mailer-error :mail-queue "Mailer error in top level loop.~%~A" condition)))
      (process-wait "Mailer Sleep" #'(lambda (wakeup-time)
				       (or *wakeup-mailer*
					   (>= (get-universal-time) wakeup-time)))
		    (max (+ (get-universal-time) (* 5 60))	;Never loop faster than 5 minutes
			 (min (+ last-delivery *mail-background-sleep-time*)
			      (+ last-probe  *inbox-probe-sleep-time*))))
      (setq utime (get-universal-time))
      (wait-till-delivery-ok)
      (setq deliver (or *wakeup-mailer*
			(>= utime (+ last-delivery *mail-background-sleep-time*))))
      (setq new-items-only (eq *wakeup-mailer* :new-item))
      (setq probe (or *wakeup-mailer*
		      (>= utime (+ last-probe  *inbox-probe-sleep-time*))))
      (setq route-check (>= utime (+ last-route-check *check-routing-interval*)))
      (setq *wakeup-mailer* nil))))
))
