;;; -*- Mode: Common-Lisp; Package: sys; Base: 8.; Patch-File: T -*-

;;; Reason: Modified OTHER-FILES-NEWER-THAN-FILE-P to handle all legitimate module specifications. Also modified MAKE-SYSTEM to load patches for all component systems.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for MAKE-SYSTEM version 6.3
;;; Written 05/07/90 14:45:21 by DAB,
;;; while running on Pasteur from band LOD2
;;; With SYSTEM 6.31, VIRTUAL-MEMORY 6.3, EH 6.7, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.4, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-FILE 6.10, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.1,
;;;  COMPILER 6.14, TV 6.24, DATALINK 6.0, CHAOSNET 6.6, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.14,
;;;  DEBUG-TOOLS 6.4, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.4, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.7, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.3, VISIDOC 6.7, TI-CLOS 6.44, CLEH 6.5, IP 3.60,
;;;  Experimental CLX 6.10, CLUE 6.97, X11M 6.22, Experimental BUG 11.19, Experimental FOO 2.0,
;;;  Experimental BAR 2.0, Experimental BAZ 2.0, Experimental FOOBAR 3.0,  microcode 648,
;;;  Band Name: rel6.0 1/23

#!C
; From file DEFSYS.LISP#> MAKE-SYSTEM; sys:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MAKE-SYSTEM; DEFSYS.#"


(DEFUN OTHER-FILES-NEWER-THAN-FILE-P (MODULES FILE
				      &AUX CREATION-DATE TRANSFORMATION-INPUT-TYPE
				      INPUT-FILE-TYPE OUTPUT-FILE-TYPE expand-modules temp) 
  (SETQ CREATION-DATE (SYSTEM-GET-FILE-PROPERTY FILE :CREATION-DATE))
  (dolist (module modules)
  ;; DAB 05-07-90 Legitimate modules specs include a module structure, list of modules or (system-bar sb-module )
  ;; I added the change below to handle the (system-bar sb-module) form used in :compile-load-init transformation.
    (cond ((SYMBOLP module) (push module expand-modules))               ; DAB 05-07-90
	  ((typep module 'sys:module) (push module expand-modules))
	  ((not (listp module))
	   (FERROR nil "~S is not a recognized module component specification" module))
	  
	  ((SYMBOLP (CAR module))
	   (COND ((setf temp (FIND-MODULE-NAMED (CAR module) *SYSTEM-being-made* T))  ;support list of modules  -dkm 6/87
		  (push  temp expand-modules))
		 (T (VALIDATE-EXTERNAL-MODULE module)   ;must be an external spec  -dkm 6/87
		    (let ((sys (find-system-named (car module))))
		       (dolist (ext-mod (cdr module))
			 (push (find-module-named (second module) sys) expand-modules))))))))
  (DOLIST (MODULE (reverse expand-MODULES))  ; DAB 05-07-90
    (OR (DOLIST (TRANSFORMATION (SYSTEM-TRANSFORMATIONS (MODULE-SYSTEM MODULE)))
	  (WHEN (EQ (TRANSFORMATION-INPUT TRANSFORMATION) MODULE)
	    (LET ((XFORM-TYPE (TRANSFORMATION-TRANSFORMATION-TYPE TRANSFORMATION)))
	      (SETQ INPUT-FILE-TYPE (CAR (TRANSFORMATION-TYPE-INPUT-FILE-TYPES XFORM-TYPE)))
	      (SETQ OUTPUT-FILE-TYPE (CAR (TRANSFORMATION-TYPE-OUTPUT-FILE-TYPES XFORM-TYPE))) ;could be nil
	      (SETQ TRANSFORMATION-INPUT-TYPE (TRANSFORMATION-TYPE-TRANSFORMATION-INPUT-TYPE XFORM-TYPE))
	      (RETURN T))))
	(FERROR nil "Module ~S not found in any transformation" MODULE))
    ;; check source and binary pathnames to see if they are newer than this binary.
    (WHEN
      (DOLIST (PATHNAMES (GET-MODULE-PATHNAMES MODULE T TRANSFORMATION-INPUT-TYPE))
	;; if either the source or binary of any pathnames in MODULES is newer than FILE, return T  02/88
	(LET ((INPUT-PATHNAME (MERGE-PATHNAME-TYPE
				 (SECOND PATHNAMES)	;first element is package. Second is pathname
				 (EVAL INPUT-FILE-TYPE)))
	      (OUTPUT-PATHNAME (AND OUTPUT-FILE-TYPE
				    (MERGE-PATHNAME-TYPE
				      (SECOND PATHNAMES)	;first element is package. Second is pathname
				      (EVAL OUTPUT-FILE-TYPE)))))
	  (AND (OR (> (SYSTEM-GET-FILE-PROPERTY input-pathname
						:CREATION-DATE)
		      CREATION-DATE)
		   (AND OUTPUT-PATHNAME        ;output-pathname may not exist  dkm 02/88
			(LET ((OUTPUT-PATHNAME-DATE (SYSTEM-GET-FILE-PROPERTY OUTPUT-PATHNAME :CREATION-DATE T)))
			  (AND OUTPUT-PATHNAME-DATE
			       (> OUTPUT-PATHNAME-DATE CREATION-DATE))))
		   ;; See if this file is about to be an output.  Makes :compile-load-init transitive  dkm 02/88
		   (DOLIST (FILE-XFORM SI:*FILE-TRANSFORMATION-LIST*)
		     (AND (OR (EQ input-PATHNAME  (CAR (FILE-TRANSFORMATION-OUTPUTS FILE-XFORM)))
			      (EQ output-PATHNAME (CAR (FILE-TRANSFORMATION-OUTPUTS FILE-XFORM))))
			  (MEMBER (FILE-TRANSFORMATION-STATE FILE-XFORM) '(:PENDING :PROBABLY))
			  (RETURN T))))
	       
	       (RETURN T))))
	(RETURN T))))
))



#!C
; From file MAKSYS.LISP#> MAKE-SYSTEM; sys:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MAKE-SYSTEM; MAKSYS.#"



(DEFUN MAKE-SYSTEM (SYSTEM &REST KEYWORDS &AUX *SOMETHING-LOADED*)
  "Operate on the files of the system SYSTEM.
Most commonly used to compile or load those files which need it.
Keywords are not followed by values.
Commonly used keywords include:

 :COMPILE            - recompile source files that have been changed since last make-system.
 '(:COMPILE :compiler-keyword1 value1 :compiler-keyword2 value2 ...)
                     - pass these options to the compiler.
 :RECOMPILE          - recompile and reload all files for this system.
 '(:RECOMPILE :compiler-keyword1 value1 :compiler-keyword2 value2 ...)
                     - pass these options to the compiler.
 :NOLOAD             - don't load compiled files.
 :RELOAD             - load even files already loaded.
 :SELECTIVE          - ask user about each file individually.
 :NOCONFIRM          - do not ask for confirmation of make-system files.
 :NOWARN             - do not prompt for ANY confirmations, including loader redefinition warnings.
 :SILENT             - don't print lists of files or loader warnings on the terminal at all.
 :NO-INCREMENT-PATCH - don't increment the patch version number of a patchable system.
 :INCREMENT-PATCH    - do increment the patch version number.
 :NO-LOAD-PATCHES    - do not load patches for patchable system being loaded.
 :UNRELEASED         - load unreleased patches as well. Ignored if :no-load-patches is specified.
 :FORCE-UNFINISHED   - load unfinished patches as well. Ignored if :no-load-patches is specified.
 :NO-RELOAD-SYSTEM-DECLARATION - don't reload the file that contains the DEFSYSTEM.
 :PRINT-ONLY         - don't load or compile anything, just say what needs to be done.
 :DESCRIBE           - say when files were compiled or loaded, etc.
 :BATCH              - write a file containing any warnings produced by compilation.
                       Just load the file, as lisp code, to reload the warnings.
 :SERIAL             - do transformations exactly in the order specifed in the defsystem
 :DEFAULTED-BATCH    - like :BATCH except warnings file is defaulted instead of asked for.
 :DO-NOT-DO-COMPONENTS - do not include systems defined by :component-systems.
 :RECORD             - record the file version numbers for the current system
 (:VERSION [num])    - remake an old major version of a system if that previous 
                       system was recorded via the :RECORD option.
 :SAFE               - in determining source later than object, go by the creation date.
                       The default depends on :OUTPUT-VERSION from the DEFSYSTEM
 :NOOP               - this option is ignored."
;**********************
;; note:  *somethig-loaded* is a special variable defined by a defvar in the file DEFS.
;; It realy should be included in the list *make-system-special-variables* and removed
;; from the defun line. It is not really an aux variable but a klude way to insure
;; *something-loaded* is reset on each entry.

  (PROGW *MAKE-SYSTEM-SPECIAL-VARIABLES*
    (UNWIND-PROTECT
      (CATCH 'EXIT-MAKE-SYSTEM
	(SETQ KEYWORDS (COPY-LIST KEYWORDS))				    ;get copy of &rest arg to be safe
	(FIND-SYSTEM-NAMED SYSTEM NIL NIL KEYWORDS)			    ;be sure the defsystem is loaded
	(MAYBE-RELOAD-SYSTEM-DECLARATION SYSTEM KEYWORDS)		    ; and that it is current

	;;initialize some make-system-special-variables
	(SETQ *SYSTEM-BEING-MADE* (FIND-SYSTEM-NAMED SYSTEM t t KEYWORDS)) ;get the real system, in case new one loaded
	(SETQ *SYSTEM-DEFAULT-BINARY-FILE-TYPE*
	      (OR (GETF (SYSTEM-PLIST *SYSTEM-BEING-MADE*) 'DEFAULT-BINARY-FILE-TYPE)
		  (LOCAL-BINARY-FILE-TYPE)))
	(SETQ *TOP-LEVEL-TRANSFORMATIONS*
	      `(,@*LOAD-TYPE-TRANSFORMATIONS* DO-COMPONENTS-INTERNAL))
	(SETQ *MAKE-SYSTEM-SERIALLY* (SERIAL-P *SYSTEM-BEING-MADE*))   ;get the :serial attribute for the system. dkm 02/88 DKM

	;; Do all the keywords			   
	(DO-THE-KEYWORDS KEYWORDS)
	(SETUP-FOR-OUTPUT-VERSION)
	;;If we are doing an old version (via :VERSION keyword), get all that setup
	(AND *USE-OLD-VERSION*
	     (DO-VERSION-KEYWORD))
	;; Make :NO-INCREMENT-PATCH override :COMPILE even if :COMPILE comes later.			   
	(WHEN *NO-INCREMENT-PATCH*
	  (SETQ *TOP-LEVEL-TRANSFORMATIONS*
		(DELETE-IF
		  #'(LAMBDA (X)
		      (MEMBER X '(INCREMENT-COMPILED-VERSION) :TEST #'EQ))
		  *TOP-LEVEL-TRANSFORMATIONS*)))

	;; If this is a patchable system, let's be sure the patch files are 
	;; around now instead of waiting for all the other transformations to
	;; finish before finding this out.  This isn't necessary, but it is a
	;; convience for the user to know of this situation early.

;        (AND (SYSTEM-PATCHABLE-P *SYSTEM-BEING-MADE*) ;all we care about is the side effect of
;             (PATCH-VERSION-NEWER-THAN-LOADED))       ;insuring the patch directories are out there

	;; Process forms with compiler context			   
	(DOLIST (FORM *MAKE-SYSTEM-FORMS-TO-BE-EVALED-BEFORE*)
	  (eval form))

	(IF (FBOUNDP 'COMPILER:COMPILER-WARNINGS-CONTEXT-BIND)
	    (COMPILER:COMPILER-WARNINGS-CONTEXT-BIND
	       (PERFORM-TRANSFORMATIONS
		 (COLLECT-TOP-LEVEL-TRANSFORMATIONS *SYSTEM-BEING-MADE*)))
	    ;;Compiler isn't around.  Go without it.
	    (PERFORM-TRANSFORMATIONS
	      (COLLECT-TOP-LEVEL-TRANSFORMATIONS *SYSTEM-BEING-MADE*)))

	;; Finally process any forms queued by the keywords with compiler context			   
	(DOLIST (FORM *MAKE-SYSTEM-FORMS-TO-BE-EVALED-AFTER*)
	  (eval form))
	;; See if any patches need to be loaded for this system.
	(WHEN (AND *LOAD-PATCHES*
		   (GET-PATCH-SYSTEM-NAMED *SYSTEM-BEING-MADE* T T)
		   (SYSTEM-PATCHABLE-P *SYSTEM-BEING-MADE*))
	  (LET ((LOAD-PATCHES-ARGS NIL))
	    (AND *SILENT-P* (PUSH :SILENT LOAD-PATCHES-ARGS))
	    (AND (EQ *QUERY-TYPE* :NOCONFIRM) (PUSH :NOCONFIRM LOAD-PATCHES-ARGS))
	    (AND *UNRELEASED*         (PUSH :UNRELEASED LOAD-PATCHES-ARGS))  ; DAB 01-12-89
	    (AND *FORCE-UNFINISHED*   (PUSH :FORCE-UNFINISHED LOAD-PATCHES-ARGS))  ; DAB 01-12-89
	    (APPLY #'LOAD-PATCHES :SYSTEMS
		   (cons (SYSTEM-SYMBOLIC-NAME *SYSTEM-BEING-MADE*)
			 ;;Lisp Ref. P 24-2. All operation form on all components.
			 (SYSTEM-COMPONENT-SYSTEMS *SYSTEM-BEING-MADE*))
		   LOAD-PATCHES-ARGS)))
	;;If :RECORD option was specified, do it.
	(AND *RECORD-VERSION-NUMBERS* (RECORD-SYSTEM-IN-LOG))
	;; System has been made!  Set :MADE-P flag to T                      ;;;;; BMK -- 8/26/87
	(if (and *SOMETHING-LOADED* (not (member :print-only keywords))) ;01-17-89 DAB
	    (progn
	      (SETF (GETF (SYSTEM-PLIST *SYSTEM-BEING-MADE*) :MADE-P)        ;;;;;
		    T)				                       ;;;;;
	      (PUSHNEW (STRING (SYSTEM-SYMBOLIC-NAME *SYSTEM-BEING-MADE*))   ;;;;;
		       *MODULES*)))
	(remember-generic-pathnames SYSTEM)  ; DAB 04-14-89
	)                                           ;;;;;

      ;; Now forms outside of compiler context
      ;; These are done even if there was an error.			  
      (DOLIST (FORM *MAKE-SYSTEM-FORMS-TO-BE-EVALED-FINALLY*)
	 (eval form)))
  *SOMETHING-LOADED*))

))