;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Changes to Monitor Configurator to prevent locking inside
;;; without-interrupts when invoking via system menu.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 10/12/89 10:20:16 by MARKY,
;;; while running on LIBRA from band LOD1
;;; With SYSTEM 6.20, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.4, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.4, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.12, TV 6.15, DATALINK 6.0, CHAOSNET 6.1, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.7,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.0, MAIL-DAEMON 6.3, MAIL-READER 6.5, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.3, PROFILE 6.2, VISIDOC 6.5, TI-CLOS 6.25, CLEH 6.5, IP 3.52,
;;;  Experimental CLX 6.4, CLUE 6.20, X11M 6.14, Experimental BUG 11.15, Experimental GRAPHICS-WINDOW 6.1,
;;;  MMON 6.0,  microcode 429, Band Name: Rel 6.0 + SLE 9/8

;;; by may, code-read cjj 

#!C
; From file MONITOR-CONFIGURATOR.LISP#> MMON; SYS:
#10R TV#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TV"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: MMON; MONITOR-CONFIGURATOR.#"


(DEFMETHOD (monitor-configurator :update-anchor-monitor)
	   ()
  "Call whenever SUPERIOR is established or changes."
  (WITH-LOCK (*monitor-configurator-lock* :who-state "update anchor monitor")
;;    (WITH-LOCK (configurator-lock :who-state "update anchor monitor")	;; may 10/12/89 
      (LET ((monitors (SEND (w:sheet-get-screen self) :visible-on-monitors)))
	(SETF anchor-monitor
	      ;; If this gets called before monitor-initialization, a "monitor" will not be a monitor instance...
	      (UNLESS (CONSP (CAR monitors))
		(OR
		  ;; Normal case: just one, so use it.
		  (UNLESS (CDR monitors)
		    (CAR monitors))
		  ;; Use first on-line monitor...
		  (LET ((on-line-monitors (REMOVE-IF 'monitor-off-line-p monitors)))
		    (WHEN on-line-monitors
		      (CAR on-line-monitors)))
		  ;; Use first off-line monitor...
		  (CAR monitors)))))
      (SEND self :maybe-refresh)))

(DEFMETHOD (monitor-configurator :draw-monitor)
	   (monitor-left
	    monitor-top
	    monitor-width
	    monitor-height
	    multi-line-description
	    &key
	    erase
	    highlight
	    no-text)
  (without-interrupts	;; may 10/12/89 Must make multiple drawing steps atomic
  (SEND self
	:draw-inside-rounded-rectangle
	monitor-left
	monitor-top
	monitor-width
	monitor-height
	(IF highlight window-highlight-thickness window-normal-thickness)
	(IF (color-sheet-p self) color-difference w:black)
	(IF (color-sheet-p self) (IF erase w:alu-sub w:alu-add) w:alu-xor))
  (UNLESS no-text
    (SEND self
	  :draw-string
	  description-font
	  multi-line-description
	  (+ monitor-left (FLOOR monitor-width 16.))	   ;x
	  (+ monitor-top (FLOOR monitor-height 16.))	   ;y
	  (IF (color-sheet-p self) color-difference w:black)
	  (+ monitor-left (FLOOR monitor-width 16.))	   ;left
	  8				   ;tab -- not used if no tabs in MULTI-LINE-DESCRIPTION
	  ;; inverse scale
	  (LOOP WITH max-x-this-line = 0.
		AND characters = (SEND description-font :characters)
		AND horz-spacing = (SEND description-font :horz-spacing)
		FOR character BEING EACH ARRAY-ELEMENT OF multi-line-description
		IF (CHAR= #\newline character)
		MAXIMIZE max-x-this-line INTO max-x
		AND SUMMING 1. INTO line-count
		AND DOING (SETF max-x-this-line 0.)
		ELSE
		DOING (INCF max-x-this-line
			    (OR (SEND (AREF characters (CHAR-INT character)) :horizontal-spacing)
				horz-spacing))
		FINALLY
		;; Account for last line...
		(UNLESS (ZEROP max-x-this-line)
		  (WHEN (> max-x-this-line max-x)
		    (SETF max-x max-x-this-line))
		  (INCF line-count))
		(RETURN
		  (MAX (/ max-x
			  (* monitor-width 7/8))
		       (/ (* line-count (SEND description-font :vert-spacing))
			  (* monitor-height 7/8)))))
	  (IF (color-sheet-p self) (IF erase w:alu-sub w:alu-add) w:alu-xor)))))

(DEFMETHOD (monitor-configurator :monitor-under)
	   (point-x point-y)
  "Returns the monitor instance, if any, whose icon is displayed at the point \(POINT-X, POINT-Y\).
POINT-X and POINT-Y should be in window coordinates relative to the outside upper-left corner of SELF."
  (MULTIPLE-VALUE-SETQ
    (point-x point-y)
    (SEND self :untransform-and-round-point point-x point-y))
  (without-interrupts ;;WITH-LOCK (configurator-lock :who-state "monitor at point")	;; may 10/12/89 
    (WHEN anchor-monitor
      (SEND anchor-monitor :monitor-at-point point-x point-y))))


(DEFMETHOD (monitor-configurator :who-line-documentation-string)
	   (&rest ignore)
  ;; This is called from mouse-process, so don't let monitor-configurator process change anything until we're done...
  (without-interrupts ;WITH-LOCK (configurator-lock)	;; may 10/12/89 
    (IF moving-p
	(LIST :mouse-L-1 (FORMAT nil
				 "Put ~A at currently shown position"
				 current-monitor-description)
	      :mouse-M-1 (FORMAT nil
				 "Abort moving ~A and put it back where it was"
				 current-monitor-description)
	      :mouse-R-1 "System menu")
	;; ELSE...
	(LIST :mouse-L-1 (IF current-monitor
			     (FORMAT nil
				     "Move ~A to a new position"
				     current-monitor-description)
			     ;; ELSE...
			     (IF (AND anchor-monitor
				      (SEND anchor-monitor :other-monitors-and-relative-positions))
				 "Not active until mouse is positioned over a monitor"
				 ;; ELSE...
				 "Not active until another monitor is added."))
	      :mouse-R-1 "System menu"))))


(DEFMETHOD (monitor-configurator :begin-monitor-move)
	   ()
  (IF (AND current-monitor
	   (NOT moving-p))
      ;; Don't let :mouse-moves in mouse process mess us up...
      (without-interrupts; WITH-LOCK (configurator-lock)	;; may 10/12/89 
	;; If current-monitor is the only one, there will be no acceptable places to move...
	(WHEN (SETF acceptable-places-to-move (translate-edge-list
						(SEND current-monitor :acceptable-places-to-move)
						current-monitor-x
						current-monitor-y))
	  (SETF moving-p t
		original-current-monitor-x current-monitor-x
		original-current-monitor-y current-monitor-y))
	(w:mouse-wakeup)
	(SEND superior :update-context))
      ;; ELSE...
      (BEEP)))


(DEFMETHOD (monitor-configurator :make-current-monitor)
	   (new-monitor new-x new-y)
  ;; Don't highlight a monitor to move if it's the only one...
  (WHEN (AND new-monitor
	     (NOT (SEND new-monitor :other-monitors-and-relative-positions)))
    (SETF new-monitor nil))
  ;; may 10/12/89  drawing monitors in steps - don't let :mouse-moves from mouse-process interfere
  (without-interrupts ;; may 10/12/89 
    (UNLESS (EQ new-monitor current-monitor)
      (WHEN current-monitor ;; make a monitor with no-text - just a border
	(SEND self
	      :draw-monitor
	      current-monitor-x
	      current-monitor-y
	      current-monitor-width
	      current-monitor-height
	      current-monitor-multi-line-description
	      :highlight t
	      :erase t
	      :no-text t)
	(SEND self
	      :draw-monitor
	      current-monitor-x
	      current-monitor-y
	      current-monitor-width
	      current-monitor-height
	      current-monitor-multi-line-description
	      :no-text t))
      (SETF current-monitor new-monitor
	    current-monitor-x new-x
	    current-monitor-y new-y
	    current-monitor-width (AND current-monitor
				       (SEND current-monitor :width))
	    current-monitor-height (AND current-monitor
					(SEND current-monitor :height))
	    current-monitor-description (AND current-monitor
					     (SEND current-monitor :description))
	    current-monitor-multi-line-description (AND current-monitor
							(SEND current-monitor :multi-line-description)))
      (WHEN current-monitor
	(SEND self
	      :draw-monitor
	      current-monitor-x
	      current-monitor-y
	      current-monitor-width
	      current-monitor-height
	      current-monitor-multi-line-description
	      :erase t
	      :no-text t)
	(SEND self
	      :draw-monitor
	      current-monitor-x
	      current-monitor-y
	      current-monitor-width
	      current-monitor-height
	      current-monitor-multi-line-description
	      :highlight t
	      :no-text t)))))


(DEFMETHOD (monitor-configurator :after :mouse-moves)
	   (x y)
  (MULTIPLE-VALUE-SETQ
    (x y)
    (SEND self :untransform-and-round-point x y))
  ;; Don't change modes in the middle of erasing and drawing, or images will get left on the screen...
  (without-interrupts ;WITH-LOCK (configurator-lock)	;; may 10/12/89 
    ;; may 10/12/89  Ignore mouse moves that occur before we have drawn anything!!!
    (unless needs-refreshing-p ;; may 10/12/89 Added
    (IF moving-p
	(PROGN
	  (MULTIPLE-VALUE-SETQ
	    (x y)
	    (SEND self :closest-acceptable-place-to x y))
	  (WHEN (OR (/= current-monitor-x x)
		    (/= current-monitor-y y))
	    (SEND self :set-cursorpos 0 0)
	    (SEND self :clear-eol)
	    (FORMAT self "~8D ~8D" x y)
	    (SEND self
		  :draw-monitor
		  current-monitor-x
		  current-monitor-y
		  current-monitor-width
		  current-monitor-height
		  current-monitor-multi-line-description
		  :highlight t
		  :erase t)
	    (SETF current-monitor-x x
		  current-monitor-y y)
	    (SEND self
		  :draw-monitor
		  current-monitor-x
		  current-monitor-y
		  current-monitor-width
		  current-monitor-height
		  current-monitor-multi-line-description
		  :highlight t)))
	;; ELSE...
	(WHEN anchor-monitor
	  (MULTIPLE-VALUE-BIND
	    (new-monitor new-x new-y)
	      (SEND anchor-monitor :monitor-at-point x y)
	    (SEND self :make-current-monitor new-monitor new-x new-y)))))))


(DEFMETHOD (monitor-configurator :around :clear-screen)
	   (continuation mapping-table original-argument-list &rest ignore)
  "Ensure CLEAR-SCREEN is not interrupted by mouse-process."
  (without-interrupts ;;WITH-LOCK (configurator-lock) ;; may 10/12/89 
    (LEXPR-FUNCALL-WITH-MAPPING-TABLE
      continuation mapping-table original-argument-list)))


(DEFMETHOD (monitor-configurator :after :clear-screen)
	   (&rest ignore)
  (SETF moving-p nil
	current-monitor nil)
	;;needs-refreshing-p nil)	;; may 10/12/89 moved AFTER :draw-self-and-other-monitors-on
  (WHEN anchor-monitor
    (SEND anchor-monitor :draw-self-and-other-monitors-on self))
  (setq needs-refreshing-p nil) ;; may 10/12/89 moved here so :mouse-moves could test it.
  (SEND superior :update-context))


(DEFMETHOD (monitor-configurator :add-monitor)
	   ()
  (UNLESS moving-p
    (LET ((local-*unconfigured-monitors*
	    (WITH-LOCK (*monitor-configurator-lock* :who-state "adding monitor")
	      (COPY-LIST *unconfigured-monitors*))))
      (IF local-*unconfigured-monitors*
	  (LET ((new-monitor (select-monitor local-*unconfigured-monitors*)))
	    (WHEN new-monitor
	      (WITH-LOCK (*monitor-configurator-lock* :who-state "adding monitor")
		;; Another monitor configurator may have already added it since we selected it, so check...
		(WHEN (MEMBER new-monitor *unconfigured-monitors* :test #'EQ)
		  (progn; WITH-LOCK (configurator-lock :who-state "adding monitor")	;; may 10/12/89 
		    ;; Another monitor configurator may have deleted our anchor-monitor, so check...
		    (WHEN anchor-monitor
		      ;; Figure out where new monitor can be added...
		      (SETF acceptable-places-to-move
			    (SEND anchor-monitor
				  :acceptable-places-to-move
				  (SEND new-monitor :width)
				  (SEND new-monitor :height)))
		      (MULTIPLE-VALUE-BIND
			(x y)
			  (SEND self :closest-acceptable-place-to 0 0)
			;; If something interrupts us before we get NEW-MONITOR on *ALL-THE-MONITORS*, things could get bad...
			(WITHOUT-INTERRUPTS
			  (SEND new-monitor :add-to anchor-monitor x y))
			(SEND new-monitor :send-if-handles :expose)
			(LOOP FOR screen IN all-the-screens
			      WHEN (AND (MEMBER new-monitor (SEND screen :visible-on-monitors) :test #'EQ)
					(NOT (EQ screen default-screen))
					(SEND screen :operation-handled-p :make-default-screen)
					(screen-exposable-p screen)
					(NOT (FIND screen *previously-selected-screens* :test #'EQ)))
			      DOING (add-to-previously-selected-screens screen t))
			;; New monitor may be ANCHOR-MONITOR for other monitor configurators...
			;; Abort all moves in progress.  Recenter and repaint all monitor icons...
			(LOOP FOR monitor-configurator IN *all-the-monitor-configurators*
			      ;; Don't change own ANCHOR-MONITOR, because this will translate X and Y...
			      UNLESS (EQ self monitor-configurator)
			      DOING
			      (SEND monitor-configurator :update-anchor-monitor))
			(SEND self :maybe-refresh)
			;; Allow user to position the new monitor...
			(without-interrupts	;; may 10/13/89 :mouse-moves could ALSO :make-current-monitor
			  (SEND self :make-current-monitor new-monitor x y)
			  (SEND self :begin-monitor-move))
			;; Exposing the added monitor may have deselected us and moved the mouse sheet...
			(LET ((screen (w:sheet-get-screen self)))
			  (UNLESS (EQ screen w:mouse-sheet)
			    (w:mouse-set-sheet screen)))
			(UNLESS (SEND superior :self-or-substitute-selected-p)
			  (SEND superior :select)))))))))
	  ;; ELSE...
	  (FORMAT *standard-output*
		  "~&All attached monitors have already been configured.  No more can be added.")))))


(DEFMETHOD (monitor-configurator :delete-anchor-monitor-if-eq-to)
	   (monitor-to-delete)
  (without-interrupts; WITH-LOCK (configurator-lock :who-state "deleting monitor")	;; may 10/12/89 
    (WHEN (EQ anchor-monitor monitor-to-delete)
      (SETF anchor-monitor nil))))

(DEFMETHOD (monitor-configurator :delete-monitor)
	   ()
  (UNLESS moving-p
    (LET ((deletable-monitors
	    (WITH-LOCK (*monitor-configurator-lock* :who-state "deleting monitor")
	      ;; Don't allow deletion of own monitor...
	      ;; Don't allow deletion of a monitor which bridges the gap between other monitors...
	      (COPY-LIST
		(REMOVE-IF 'connecting-monitor-p
			   (REMOVE anchor-monitor *all-the-monitors* :test #'EQ))))))
      (IF deletable-monitors
	  (LET ((monitor-to-delete (select-monitor deletable-monitors)))
	    (WHEN monitor-to-delete
	      (WHEN (w:mouse-confirm
		      (FORMAT nil
			      "Are you sure you want to delete~%~A?"
			      (SEND monitor-to-delete :description)))
		(WITH-LOCK (*monitor-configurator-lock* :who-state "deleting monitor")
		  ;; Another monitor configurator may have already deleted it since we selected it, so check...
		  (WHEN (MEMBER monitor-to-delete
				(REMOVE-IF 'connecting-monitor-p
					   (REMOVE anchor-monitor *all-the-monitors* :test #'EQ))
				:test #'EQ)
		    (progn ;WITH-LOCK (configurator-lock :who-state "deleting monitor")	;; may 10/12/89 
		      ;; Another monitor configurator may have deleted our anchor-monitor, so check...
		      (WHEN anchor-monitor
			(SEND monitor-to-delete :send-if-handles :deexpose)
			(WITHOUT-INTERRUPTS
			  ;; This removes all cross-pointers to this monitor from other monitors,
			  ;;  and adjusts the global variables...
			  (SEND monitor-to-delete :delete)
			  (LOOP WITH index = 0
				AND end = (ARRAY-TOTAL-SIZE *previously-selected-screens*)
				FOR screen = (AREF *previously-selected-screens* index)
				WHILE screen
				UNTIL (>= index end)
				IF (screen-exposable-p screen)
				DOING (remove-from-previously-selected-screens screen)
				ELSE
				DOING (INCF index)))
			(LOOP FOR monitor-configurator IN *all-the-monitor-configurators*
			      DOING
			      (SEND monitor-configurator :delete-anchor-monitor-if-eq-to monitor-to-delete))
			;; Abort all moves in progress.  Recenter and repaint all monitor icons...
			(clear-all-monitor-configurators)
			;; Deexposing the deleted monitor may have deselected us and moved the mouse sheet...
			(LET ((screen (w:sheet-get-screen self)))
			  (UNLESS (EQ screen w:mouse-sheet)
			    (w:mouse-set-sheet screen)))
			(UNLESS (SEND superior :self-or-substitute-selected-p)
			  (SEND superior :select)))))))))
	  ;; ELSE...
	  (FORMAT
	    *standard-output*
	    "~&No more monitors may be deleted.  The monitor you are currently using may not be deleted.~%~
Monitors which provide the only physical \"bridge\" between other monitors may not be deleted.")))))

))
