;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Put the eadr resource into the system file instead of the EtherTalk driver. Also allow more proceed options
;;; when resetting the Ethernet address and look for the case where we are booting up with someone else's
;;; Ethernet address (because they copied a system file).

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/03/89 04:33:54 by JONES,
;;; while running on HOBBS from band P526
;;; With SYSTEM 6.2, GC 6.0, VIRTUAL-MEMORY 6.0, MICRONET 6.0, MICRONET-COMM 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-PATHNAME 6.0, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0,
;;;  BASIC-NAMESPACE 6.0, BASIC-FILE 6.0, RPC 6.0, NFS 6.0, EH 6.0, MAKE-SYSTEM 6.0,
;;;  MEMORY-AUX 6.0, COMPILER 6.1, TV 6.6, NVRAM 6.0, UCL 6.0, INPUT-EDITOR 6.0, MACTOOLBOX 2.0,
;;;  METER 6.0, ZWEI 6.0, DEBUG-TOOLS 6.0, WINDOW-MX 6.0, PRINTER 6.0, MAC-PRINTER-TYPES 6.0,
;;;  CLIPBOARD 6.0, NETWORK-PATHNAME 6.0, NETWORK-NAMESPACE 6.0, DATALINK 6.0, CHAOSNET 6.0,
;;;  NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0, MX-DATALINK 6.0,
;;;  NAMESPACE-EDITOR 6.0, IP 3.42, NFS-SERVER 6.0, MX-SERIAL 6.0, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.0,
;;;  DECNET 1.65, VISIDOC 6.0, PROFILE 6.1, TI-CLOS 6.3, CLEH 6.0, CLX 6.0, CLUE 6.0,
;;;  Experimental BUG 11.5, Experimental ACTION 2.0,  microcode 135, Band Name: Release 6.0 mx 5/26


;;
;; 1) Put the eadr resource into the system file instead of the ethernet driver file.
;;
;; 2) Allow a proceed option that re-opens the ethernet driver with the new address (breaking all AppleTalk connections).
;;
;; 3) Check when initializing the controller to see if we are using someone else's ethernet address and provide proceed options.

#!C
; From file MAC-ENC.LISP#> JONES; HEYERDAHL:
#10R ETHERNET#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "ETHERNET"))
                          (SI:*LISP-MODE* :Common-lisp)
                          (*READTABLE* Common-lisp-readtable)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* Sys::*common-lisp-symbol-substitutions*))
  (COMPILER#:PATCH-SOURCE-FILE "sys: Datalink; MAC-ENC.#"

(defflavor wrong-ethernet-address
	   (foo)
	   (eh:error))

(defmethod (wrong-ethernet-address :case :proceed-asking-user :wait-for-next-boot) (continuation read-object)
  (declare (ignore read-object))
  (tv:notify nil "~5&***                Ethernet address changed for Decnet                    ***~&~
                    *** Decnet will be disabled until after the next Macintosh boot.           ***~5&")
  (send continuation :wait-for-next-boot nil))

(defmethod (wrong-ethernet-address :case :document-proceed-type :wait-for-next-boot) (stream ignore)
  (format stream "Wait until the next Macintosh boot leaving Decnet disabled until then."))

(defmethod (wrong-ethernet-address :case :proceed-asking-user :reset-now) (continuation read-object)
  (declare (ignore read-object))
  (let* ((c (get self :controller))
	 (enable-p (send c :enabled)))

    (format *query-io* "Please close any open AppleTalk connections and then press any key to continue ... ")
    (beep)
    (read-char *query-io*)
    (send c :disable)
    (send c :close)
    (send c :initialize)
    (when enable-p
      (send c :enable))
  (tv:notify nil "~5&***                Ethernet address changed for Decnet                    ***~&~
                     ***                This broke any open AppleTalk connections.             ***~5&")
  
  (send continuation :reset-now t)))

(defmethod (wrong-ethernet-address :case :document-proceed-type :reset-now) (stream ignore)
  (format stream "Reset the Ethernet address and enable Decnet now. This will break open AppleTalk connections."))

(defmethod (wrong-ethernet-address :case :proceed-asking-user :delete-eadr-resource) (continuation read-object)
  (declare (ignore read-object))
  (let* ((c (get self :controller))
	 (enable-p (send c :enabled)))

    (format *query-io* "Please close any open AppleTalk connections and then press any key to continue ... ")
    (beep)
    (read-char *query-io*)
    (send c :disable)
    (send c :delete-eadr-resource)
    (send c :close)
    (send c :initialize)
    (when enable-p
      (send c :enable))
  (tv:notify nil "~5&***                Ethernet address restored.                             ***~&~
                     ***                This broke any open AppleTalk connections.             ***~5&")
  
  (send continuation :delete-eadr-resource t)))

(defmethod (wrong-ethernet-address :case :document-proceed-type :delete-eadr-resource) (stream ignore)
  (format stream "Restore the correct (non Decnet) Ethernet address. This will break open AppleTalk connections."))

(defmethod (wrong-ethernet-address :case :proceed-asking-user :ignore-wrong-ethernet-address) (continuation read-object)
  (declare (ignore read-object))
  (send continuation :ignore-wrong-ethernet-address nil))

(defmethod (wrong-ethernet-address :case :document-proceed-type :ignore-wrong-ethernet-address) (stream ignore)
  (format stream "Continue using the possibly incorrect Ethernet address."))

(compile-flavor-methods wrong-ethernet-address)

(defsignal wrong-ethernet-address wrong-ethernet-address (controller) "Signaled when changing the ethernet address.")


(Defmethod (MAC-ENC :INITIALIZE) (&rest ignore &aux (decnet-hiord #xAA000400))
  "Performs the setup of the Ethernet controller and resets it."
  
  (Unless Open
    (Setf Channel
	  (Add:Find-Channel LAPCHANNEL))	; get channel
    (Send Self :Open)                           ; open driver
    (Send Self :General-Mode))                  ; and put it into general mode
  (Setf Ethernet-Address
	(Send Self :Read-Rom-Id))		; Get Ethernet id from board

  ;; Check for the case where you have gotten someone elses ethernet address. This will happen
  ;; when someone copies a system file that includes an eadr resource from one machine to another. 
  (when (and 
	  (not (string-equal (send si:local-host :domain) name:*distribution-name*))
	  (= (ldb (byte 32 16) ethernet-address) decnet-hiord))
    (let ((decnet-address (send si:local-host :network-address :dna)))
      (cond ((and decnet-address
		  (not (= (ldb (byte 16 0) ethernet-address) (net:swap-bytes decnet-address))))
	     (signal 'wrong-ethernet-address
		     (format nil "This machine appears to have the another machines Ethernet address ~
                                ~&because the current Decnet address does not match the Ethernet address. ") self
		     :proceed-types '(:delete-eadr-resource :ignore-wrong-ethernet-address))
	     )
	    ((null decnet-address)
	     (signal 'wrong-ethernet-address "This machine appears to have an Ethernet address for Decnet but does not have a Decnet address." self
		     :proceed-types '(:ignore-wrong-ethernet-address :delete-eadr-resource))
	     ))))
  
  (Send Self :Screen-Setup)			; Attach Protocols, setup receive acbs
  (setf Transmit-Acb nil)			; allocated from :transmit 8/88 JP
  (Send Self :Enable-Reception)			; Begin receiving frames

  (Send Self :MC-Setup)
  )


(DefMethod (MAC-ENC :SET-EADDR) (address)
  (COND ((= (SEND self :read-eadr) address)
	 ;; Our address is already ADDRESS.
	 (SETQ ethernet-address address)
	 t)
	(t
	 ;; Use the toolbox interface to put an EADR resource on the ethernet driver. *BJ*
	 (let* ((tb:*application-channel* 8)
		(inhibit-scheduling-flag t)
		eaddr-resource
	       (cur-enet-addr 0)
	       (ether-res-file 0)
	       old-res-file-attrs
	       (current-res-file (tb:!CurResFile)))
	   (unwind-protect
	       (progn
		 
		 (tb:!UseResFile ether-res-file)
		 (setf old-res-file-attrs (tb:!GetResFileAttrs ether-res-file))
		 (tb:!SetResFileAttrs ether-res-file (logior old-res-file-attrs tb:!mapChanged tb:!mapCompact))
		 (tb:!UpdateResFile ether-res-file)
		 
		 (tb:suppress-some-oserrs (tb:!resNotFound) (setf eaddr-resource (tb:Get1Resource "eadr" net:slot)))
		 
		 (cond ((or (not (typep eaddr-resource 'tb:mac-handle))
				 (zerop (send eaddr-resource :handle)))
			(setf eaddr-resource (tb:!NewHandle 6))
			(when (zerop (send eaddr-resource :handle))
			  (ferror 'cant-set-enet-addr "Cannot set the ethernet address because of a resource manager error."))
			(tb:!addResource eaddr-resource "eadr" net:slot "ethernet address"))
		       (t		  
			(dotimes (x 3)
			  (setf cur-enet-addr (dpb (tb:fetchWordHandle eaddr-resource (* 2 (- 2 x))) (byte 16 (* x 16)) cur-enet-addr)))))
		 
		 ;; If the ethernet address has changed write it out.
		 (unless (= cur-enet-addr address)
		   (dotimes (x 3)
		     (tb:StowWordHandle eaddr-resource (* 2 (- 2 x)) (ldb (byte 16 (* x 16)) address)))
		   (tb:!changedResource eaddr-resource)
		   (tb:!WriteResource eaddr-resource))
		 (tb:!UpdateResFile ether-res-file)
		 (tb:!UseResFile current-res-file)
		 )
	     (when (and ether-res-file (not (zerop ether-res-file)))
	       (tb:!CloseResFile ether-res-file)))

	   (cond ((eq :reset-now (signal 'wrong-ethernet-address "The Ethernet address needs to be reset for this machine to match its Decnet address. "
					 self :proceed-types '(:wait-for-next-boot)))
		  (setf ethernet-address address)
		  t)
		 (t nil))
	   ))))

(defmethod (mac-enc :delete-eadr-resource) ()
  (let* ((tb:*application-channel* 8)
	 (inhibit-scheduling-flag t)
	 eaddr-resource
	 (ether-res-file 0)
	 old-res-file-attrs
	 (current-res-file (tb:!CurResFile)))
    (unwind-protect
	(progn
	  
	  (tb:!UseResFile ether-res-file)
	  (setf old-res-file-attrs (tb:!GetResFileAttrs ether-res-file))
	  (tb:!SetResFileAttrs ether-res-file (logior old-res-file-attrs tb:!mapChanged tb:!mapCompact))
	  (tb:!UpdateResFile ether-res-file)
	  
	  (tb:suppress-some-oserrs (tb:!resNotFound) (setf eaddr-resource (tb:Get1Resource "eadr" net:slot)))

	  (when (and (typep eaddr-resource 'tb:mac-handle)
		     (not (zerop (send eaddr-resource :handle))))
	    (tb:!RmveResource eaddr-resource)
	    (tb:!UpdateResFile ether-res-file))
	  (tb:!UseResFile current-res-file)
	  )
      (when (and ether-res-file (not (zerop ether-res-file)))
	(tb:!CloseResFile ether-res-file)))
    ))

(DefMethod (MAC-ENC :IA-SETUP) (address)
  "Set up our Ethernet address to be ADDRESS."
  (cond ((= ethernet-address address))
	(t
	 (send self :set-eaddr address))))

(Defmethod (MAC-ENC :CLOSE) ()
  "Closes the Ethernet driver on the MAC side and puts the MAC-ENC Ethernet 
controller in a closed state."
    
  (let (Acb)
    (unwind-protect
	(progn
	  (setf acb (Add:Get-acb-fast STATUS-SIZE))
	  (Add:Init-Acb Acb ECLOSE ECLOSE)
	  (Setf (Lap-Refnum Acb) Refnum)       
	  (Add:Transmit-packet-And-wait Acb Channel)
	  (Send Self :Signal-Error Acb))
      (Add:Return-Acb Acb t)))
  (WITHOUT-INTERRUPTS 
    (Setf Open Nil)
    (DOLIST (el transmit-acb)
      (add:return-acb-fast el t))
    (LOOP for (prot acb-lst) in receive-acb-list do
	  (LOOP for el in acb-lst do
		(add:return-acb-fast el t)))
    (SETF receive-acb-list nil
	  xmit-pending nil
	  transmit-acb nil
	  valid-pkt-types nil
	  attached-protocols nil
	  transmit-acb-lock nil
	  )))

(compile-flavor-methods mac-enc)

))
