;;; -*- Mode: Common-Lisp; Package: NET; Base: 10.; Patch-File: T -*-

;;; Reason: Host-time uses the variable time-server-hosts which is not initialized until the site-option-initialization-list. Therefore, the time is not updated from the network server. 

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/27/89 14:50:00 by BERGER,
;;; while running on ARIES from band LOD1
;;; With SYSTEM 6.9, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.6, TV 6.11, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.3, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.3,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.1, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, TI-CLOS 6.9, CLEH 6.4, IP 3.47,
;;;  Experimental BUG 11.10, Experimental CLX 6.1, CLUE 6.5, X11M 6.1,  microcode 429,
;;;  Band Name: 6.0 + patches 6/26

#!C
; From file TIME.LISP#> NETWORK-SERVICE; SYS:
#10R NET#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NET"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NETWORK-SERVICE; TIME.#"


(defun host-time (&optional hosts &aux connections last-host)
  (declare (special time-server-hosts ))
  (unless hosts  ; DAB 06-27-89
    (setf hosts (or time-server-hosts
		    (get-site-option :primary-time-servers)))) ; DAB 06-27-89
  (when (and (atom hosts)
	     (not (null hosts)))
    (setq hosts (list hosts)))
  (setf hosts (mapcar #'si:parse-host hosts))
  (setq last-host (car (last hosts)))
  (unwind-protect
      (loop named host-time
	    for host in hosts
	    with connection 
	    if (not (errorp (setf connection (condition-case (condition)
						 (send host :time :connect)
					       (sys:network-error condition)))))
	    do (push connection connections)
	    (when (process-wait-with-timeout "Ask the Time"
					     (if (eq host last-host) 300 60)
					     #'(lambda (connections)
						 (loop for connection in connections
						       when (send connection :packet-present-p)
						       return t))
					     connections)
	      (loop for connection in connections
		    with pkt and pkt-string
		    do (unwind-protect
			   (when (multiple-value-setq (pkt pkt-string) (send connection :get-next-pkt t))
			     (return-from host-time
			       (decode-canonical-time-packet connection pkt-string)))
			 (when pkt (send connection :return-input-pkt pkt)))))
	    else
	    do (format *error-output* "~&Connection to host ~a failed with ~a" host connection)
	    finally (return (format nil "~:[No hosts responded.~;Host not responding.~]"
				    (equalp 1 (length hosts)))))
    (mapc #'(lambda (conn) (send conn :return-connection)) connections)))
))
