;;; -*- Mode: Common-Lisp; Package: NSe; Base: 10.; Patch-File: T -*-

;;; Reason: Modified verify-host-name to check for host "MT", which should not be assigned because the Backup Utility uses it.[10267]
;;; Also modified verify-addresses to make sure addresses do not conflict with name:*network-dummy-addresses.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 10/04/89 08:56:47 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 6.17, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.1, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.2, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.3, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.12, TV 6.15, DATALINK 6.0, CHAOSNET 6.1, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.5,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.4, TI-CLOS 6.24, CLEH 6.5, IP 3.50,
;;;  Experimental CLX 6.3, CLUE 6.17, X11M 6.14, Experimental BUG 11.15, VISIDOC-SERVER 6.1,
;;;   microcode 429, Band Name: Rel 6.0 + SLE 8/30

#!C
; From file NETWORK-MODE-EXPERT-EDITORS.LISP#> NAMESPACE-EDITOR; Hotel:
#10R NSE#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NSE"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NAMESPACE-EDITOR; NETWORK-MODE-EXPERT-EDITORS.#"


(defun VERIFY-HOST-NAME ()
  (let ((host-name (object-name *object*)))
    (if (and (stringp host-name)
		 (null (position #\: host-name)))
	(when (string-equal  host-name "MT")  ; DAB 08-02-89
	  (verify-err ; DAB 08-02-89
	    "The host name \"MT\" is reserved by the Backup Tape Utility and therefore, should not be assigned.")
	   )
	(verify-err
	  "The host name ~s is invalid.  Host names should be strings with no colons in them."
	  host-name))
    ))

(defun VERIFY-ADDRESSES ()
  "Verify that the value of :ADDRESSES is an alist.
   Verify validity of addresses by calling the :NETWORK-ADDRESS-PARSER found on
   the address-type symbol \(:CHAOS, :IP\).  Verify that addresses are unique."
   ;; VERIFY FORMAT OF ADDRESSES
  (unless (verify-list-of-lists)
    (return-from verify-addresses))
  ;; VERIFY THAT :ADDRESSES IS AN ALIST
  (dolist (sublist *value*)
    (unless (eq (length sublist) 2)
      (verify-err "Invalid format of ~s for ~s.  ~s must be an alist."
		  (safe-car *attribute*) (object-name *object*)
		  (safe-car *attribute*))
      (return)))
  ;; VERIFY ONE ADDRESS AT A TIME FOR VALIDITY AND UNIQUENESS
  (let ((addrs *value*))
    (dolist (addr-item addrs)
      (let* ((addr-type (first addr-item))
	     (addr (second addr-item))
	     (parsed-addr)
	     (ERROR-OCCURRED-P)
	     (duplicates)
	     (test `test-addr-unique)
	     (DUMMY-NETWORK-ADDRESSES (second (assoc addr-type Name:*DUMMY-NETWORK-ADDRESSES*))); DAB 10-04-89
	     )
	;; CALL ADDRESS-PARSER.
	;;Check to make sure the address does not conflict with the Name:*DUMMY-NETWORK-ADDRESSES*, which
        ;;is used during namespace initializations. This could cause this host to crash when other
        ;;machines boot, espically other MXs.
	(if (and DUMMY-NETWORK-ADDRESSES  ; DAB 10-04-89
	          (eq addr DUMMY-NETWORK-ADDRESSES))
	    (setf ERROR-OCCURRED-P :DUMMY-NETWORK-ADDRESSES)
	    ;else
	    (setf (VALUES parsed-addr ERROR-OCCURRED-P)
		  (net:parse-network-address addr addr-type  nil)))
	(cond
	  ((EQ ERROR-OCCURRED-P :NO-PARSER-EXISTS)
	   (verify-wrn
	     "No network address parser exists for ~s." addr-type))
	  ((EQ ERROR-OCCURRED-P  :DUMMY-NETWORK-ADDRESSES)  ; DAB 10-04-89
	   (verify-wrn
	     "The host ~s specified ~a Address ~d, which is being assigned an address reserved by the Namespace Initializations. This will cause unexpected crashes."
	     (object-name *object*) addr-type addr))
	  (ERROR-OCCURRED-P
	   (verify-err
	     "The host ~s specified ~a Address ~d, which is not a valid ~a address."
	     (object-name *object*) addr-type addr addr-type))
	  (t
	   ;; IF ADDRESS IS VALID, CHECK TO SEE THAT IT IS UNIQUE
	   (when (setf duplicates
		       (name:list-objects-from-properties
			 :namespace (send *buffer* :namespace-object)
			 :property-list (list parsed-addr addr-type)
			 :brief t
			 :test test))
	     ;; REMOVE THE ORIGINAL OBJECT FROM THE LIST OF DUPLICATES
	     (setf duplicates
		   (remove
		     (assoc (object-name *object*) duplicates :test 'name:key-compare)
		     duplicates :test 'name:key-compare))
	     (when duplicates
	       (verify-wrn "Duplicate ~a address ~s exists on Hosts ~s and ~s."
			   addr-type (or parsed-addr addr) (object-name *object*)
			   (object-name (first duplicates))))
	     )))))))
))
