;;; -*- Mode: Common-Lisp; Package: NSE; Base: 10.; Patch-File: T -*-

;;; Reason: Rewrote edit-namespace to make sure we are in a zmacs buffer, otherwise zwei:*window* is unbounded and the user is thrown into the error handle. [7053]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/22/90 08:58:54 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 6.41, VIRTUAL-MEMORY 6.3, EH 6.8, MAKE-SYSTEM 6.5, MICRONET 6.0, LOCAL-FILE 6.2,
;;;  BASIC-PATHNAME 6.5, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.8, NETWORK-NAMESPACE 6.1,
;;;  DISK-IO 6.4, DISK-LABEL 6.1, BASIC-FILE 6.13, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.2,
;;;  COMPILER 6.18, TV 6.31, DATALINK 6.0, CHAOSNET 6.8, GC 6.4, MEMORY-AUX 6.0, NVRAM 6.4,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.6, UCL 6.1, INPUT-EDITOR 6.0, METER 6.2, ZWEI 6.25,
;;;  DEBUG-TOOLS 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.3, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.8, MAC-PRINTER-TYPES 6.2, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.8, TELNET 6.1, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.6, PROFILE 6.3, VISIDOC 6.7, TI-CLOS 6.53, CLEH 6.5, IP 3.65,
;;;  Experimental BUG 11.19, Experimental CLX 7.0, Experimental CLUE 7.3, X11M 6.31,
;;;   microcode 483, Band Name: Rel 6.1 *b +SLE 6/18

#!C
; From file COMMANDS.LISP#> NAMESPACE-EDITOR; SYS:
#10R NSE#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "NSE"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NAMESPACE-EDITOR; COMMANDS.#"


(defun EDIT-NAMESPACE (&key (select t) name
		       namespace-object namespace-type (search-list-loc :beginning)
		       (menu-label "Namespace to Copy to"))
  "Main entry point for the Namespace Editor -- It searches for an appropriate Namespace
Edit Buffer and selects it.  If it does not find one, it creates a new one,
sets up some initial data, and calls the :TRUE-BUFFER-REVERT method to fill it up with
data (the list of Classes for that Namespace).  When select is NIL, do not
select the new namespace buffer."
  (if select
      (let* ((sheet (zwei:find-or-create-idle-zmacs-window)))  ; 08-09-88 DAB
	   (funcall sheet :force-kbd-input
		    `(:execute EDIT-NAMESPACE-1 :select ,select :name ,name
			       :namespace-object ,namespace-object :namespace-type ,namespace-type
			       :search-list-loc ,search-list-loc :menu-label ,menu-label))
	   (send sheet :select) 
	   (tv:await-window-exposure)
	   sheet)
      (EDIT-NAMESPACE-1 :select select :name name
			:namespace-object namespace-object :namespace-type namespace-type
			       :search-list-loc search-list-loc :menu-label menu-label)
      ))

(defun EDIT-NAMESPACE-1 (&key (select t) name
		       namespace-object namespace-type (search-list-loc :beginning)
		       (menu-label "Namespace to Copy to"))
  "Main entry point for the Namespace Editor -- It searches for an appropriate Namespace
Edit Buffer and selects it.  If it does not find one, it creates a new one,
sets up some initial data, and calls the :TRUE-BUFFER-REVERT method to fill it up with
data (the list of Classes for that Namespace).  When select is NIL, do not
select the new namespace buffer."
  (let* ((namespace-name)
	 (usage)
	 (namespace-edit-buffer))
    (setf *package* (si:find-package 'user))  ;force package to USER.
    (setf namespace-name
	  (or name
	      (get-namespace-name-from-user menu-label)))

    ;; LOOK FOR AN EXISTING NAMESPACE-EDIT-BUFFER IN THE *ZMACS-BUFFER-LIST*
    (setf namespace-edit-buffer
	  (dolist (buffer *zmacs-buffer-list*)
	    (when (name:key-compare (format nil "Namespace Edit Buffer for ~a" namespace-name)
				(buffer-name buffer))
	      (return buffer))))
    ;; IF AN EXISTING BUFFER WAS FOUND, SELECT THE BUFFER (IF APPROPRIATE)
    ;;       AND RETURN FROM THIS FUNCTION
    (when namespace-edit-buffer
      (when select
	(send namespace-edit-buffer :select))
      (return-from edit-namespace-1 namespace-edit-buffer))

    ;; SET UP NAMESPACE-OBJECT
    (unless namespace-object
      ;; If the namespace already exists, NAME:CONFIGURE-NAMESPACE returns NIL for the SEARCH-LIST-LOC.
      ;;  Therefore, we should only call configure-namespace if the namespace doesn't already exist.
      ;; The CONVERT-SITEINFO utility is dependent on this ordering of operations.
      (or
	(setf namespace-object (name:find-namespace namespace-name :if-does-not-exist nil))
	(multiple-value-setq (namespace-object usage search-list-loc)
	  (name::configure-namespace namespace-name))))

    ;; Make sure the usage gets set up correctly.
    (setf usage (send namespace-object :usage))

    ;; MAKE SURE NAMESPACE-NAME IS THE SAME AS THE ONE
    ;; THE USER SPECIFIED WHEN CONFIGURING A NAMESPACE DOMAIN OBJECT
    (setf namespace-name (send namespace-object :domain-name))
    (condition-call (condition)
	;; CREATE A NAMESPACE-EDIT BUFFER FOR THE NAMESPACE
	(unless namespace-edit-buffer
	  (setf namespace-edit-buffer
		(make-instance 'namespace-edit-buffer
			       :namespace-name namespace-name
			       :namespace-object namespace-object
			       :namespace-type namespace-type ;; only used for INTERNAL namespaces
			       :name
			       (format nil "Namespace Edit Buffer for ~a" namespace-name)
			       :search-list-loc search-list-loc))
	  ;;; SET UP MODES AND FONTS ON THE BUFFER
	  (setf (buffer-saved-major-mode namespace-edit-buffer) 'namespace-edit-mode)
	  (setf (node-undo-status namespace-edit-buffer) :dont)
	  (set-buffer-fonts namespace-edit-buffer *nse-fonts*)
	  (setf (send namespace-edit-buffer :NEEDS-SAVING-P) nil)
	  ;; PUT NSE BUFFER ON BUFFER-HISTORY
	  (send namespace-edit-buffer :activate)
	  ;; SELECT NSE BUFFER IF APPROPRIATE
	  (when select
	    (send namespace-edit-buffer :select))
	  ;; REVERT THE NSE BUFFER (WRITES THE CLASS-LINES AND HEADERS IN THE BUFFER)
	  (send namespace-edit-buffer :true-buffer-revert)
	  (make-buffer-read-only namespace-edit-buffer))
      (t
       (send namespace-edit-buffer :kill)
       (FERROR 'NAMESPACE-ERROR
	       "Error while setting up namespace:  ~%~a"
	       (SEND condition :report-string))))
    namespace-edit-buffer)) 

))
