;;; -*- Mode: Common-Lisp; Package: SI; Base: 8.; Patch-File: T -*-

;;; Reason: Changed load-crash-table to check for the new naming convention for crash files of a microExplorer. Lm:main:microexp:expsys:ubin:Mddd.crash.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/23/89 10:25:04 by BERGER,
;;; while running on ARIES from band LOD1
;;; With SYSTEM 6.9, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.4, TV 6.11, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.0,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.3, UCL 6.0, INPUT-EDITOR 6.0, METER 6.0, ZWEI 6.3,
;;;  Experimental DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.1, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, TI-CLOS 6.8, CLEH 6.4, IP 3.46,
;;;  Experimental BUG 11.10, Experimental CLX 6.1, CLUE 6.5, X11M 6.1, RPC 6.1, NFS 3.8,
;;;   microcode 429, Band Name: Release 6.0 6/16

#!C
; From file CRASH.LISP#> NVRAM; SYS:
#8R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NVRAM; CRASH.#"


(Defun Load-Crash-Table (Ucode-Version record)
  ;; Returns crash table for version UCODE-VERSION of the microcode.
  ;;  4/19/88 CLM - Changed to handle new mcr naming convention (e.g. exp1-ucode-540.crash#1).
  (With-Sys-Host-Accessible
   (let* ((ucode-name (OR (get-ucode-name-from-crec record) "EXP2-UCODE"))
	  (probe (probe-file (send (fs:parse-pathname *Default-Crash-Table-Pathname*) :new-pathname   ; DAB 06-21-89 Very old method
				  :name (concatenate 'string ucode-name (format nil "-~s"  ucode-version))
				  :type "crash")))
          (Pathname
	    (or probe
		;; Current Explorer conventions,MX convention when stored on Explorer.
		(probe-file (send (send (fs:parse-pathname *Default-Crash-Table-Pathname*) :new-name ucode-name)
				  :new-type-and-version "crash" ucode-version))
		(if (eq (send si:local-host :pathname-flavor)  'fs:mac-pathname)   ; DAB 04-18-89
		    ;;Old method of stroing on microExplorer.
		    ;;; *Default-Crash-Table-Pathname* is a logical pathname.
		    ;;  If on the MX sys-host must have a device specific in it translations. ; DAB 06-23-89
		    (or (probe-file (send (send (fs:parse-pathname *Default-Crash-Table-Pathname*) ; DAB 04-18-89
					   :new-name (format nil "MX~s"  ucode-version)) ; DAB 04-18-89
				     :new-type "crash"))
			;;Latest way on microExplorer. We have to try each because it could fit in any case.
			(send (send (fs:parse-pathname *Default-Crash-Table-Pathname*) ; DAB 04-18-89
				    :new-name (format nil "M~s"  ucode-version))  ; DAB 06-12-89 Yet a new convention
			      :new-type "crash"))
		    (send (send (fs:parse-pathname *Default-Crash-Table-Pathname*) :new-name ucode-name)
				  :new-type-and-version "crash" ucode-version)
		    ))))
     ;;;(Funcall (Fs:Parse-Pathname *Default-Crash-Table-Pathname*) :New-Type-And-Version
     ;;;	    "CRASH" Ucode-Version)));loading control.crash will set Microcode-Crash-Table
     ;var to be our crash table database.
     (Catch-Error-Restart (Error "Continue display without Shutdown Description")
	(Do ()
	    ((Load Pathname :Package "SI" :If-Does-Not-Exist ()))
	  (Multiple-Value-Setq (Nil Pathname)
	    (Signal
	     (Make-Condition 'Fs:File-Not-Found "Crash record file ~s not found." Pathname)
	     :Proceed-Types :New-Pathname))

          ; The following 3 lines of codes are only needed until the pathname bug is fixed - RJF
	  (setq pathname (send pathname :new-type "crash"))
	  ;;in the new scheme of things, there is only one version number: #1. - CLM
	  (when (and (null probe)
		     (equal (send pathname :version) :newest))
		(setq pathname (send pathname :new-version ucode-version)))
	                                    ))
     (Setq Microcode-Crash-Tables
	   (Cons (Cons Microcode-Crash-Table-Version-Number Microcode-Crash-Table)
		 Microcode-Crash-Tables))
     Microcode-Crash-Table)))
))
