;;; -*- Mode: Common-Lisp; Package: sys; Base: 10.; Patch-File: T -*-

;;; Reason: Modified system-shutdown to close-all-files before the network gets reset.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for NVRAM version 6.2
;;; Written 10/05/89 09:39:06 by BERGER,
;;; while running on MX29 from band P815
;;; With SYSTEM 6.17, GC 6.3, VIRTUAL-MEMORY 6.2, MICRONET 6.0, MICRONET-COMM 6.1,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-PATHNAME 6.1, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0,
;;;  BASIC-NAMESPACE 6.2, BASIC-FILE 6.3, RPC 6.1, NFS-MX 6.1, EH 6.5, MAKE-SYSTEM 6.0,
;;;  MEMORY-AUX 6.0, COMPILER 6.12, TV 6.15, NVRAM 6.1, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.7, METER 6.1, ZWEI 6.5, DEBUG-TOOLS 6.3, WINDOW-MX 6.8, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.20, CLEH 6.5, NETWORK-PATHNAME 6.0,
;;;  NETWORK-NAMESPACE 6.0, DATALINK 6.0, CHAOSNET 6.1, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.0, IP 3.50, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.1, IMAGEN 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2,
;;;  TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.4, DECNET 1.70, VISIDOC 6.4, PROFILE 6.1,
;;;  Experimental CLX 6.3, CLUE 6.17, Experimental BUG 11.15,  microcode 138, Band Name: esmicroExplorer Network + SLE (8/15)

#!C
; From file SHUTDOWN.LISP#> NVRAM; SYS:
#10R SYSTEM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SYSTEM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* *COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: NVRAM; SHUTDOWN.#"


(Defun System-Shutdown (&Key (Type :System) Reason-String (Stream *Standard-Output*) (Return ()))
  "This function is meant to be called only from system-level code to shut down the system
in an orderly fashion.  It does general cleanup work, such as calling LOGOUT, before halting
the processor.  DO NOT use this function if you just want the machine to drop dead because
you have encountered an unrecoverable error.  In that case, use si:%crash instead, and be
sure to provide a crash code.
  If the keyword RETURN is non-nil, this function WILL return (that is, it will shut
down all activity but not call si%crash).  If you use this option PLEASE look at the
code to make sure it does all the quieting you want done.  Do not expect the network
or the file system to be around when the function returns.  If the RETURN keyword is NIL
\(the default) the function will end by calling si:%crash with a code of 0, indicating
a normal shutdown.
  If the optional STREAM keyword is non-nil (it defaults to *STANDARD-OUTPUT*) some status
messages will be displayed to stream STREAM as the system shuts down.
  The TYPE keyword controls the shutdown message that will be sent to any active servers
and written to the who-line documentation window.  The message used is picked up
from the SI:SHUTDOWN-REASON-STRING property of the symbol used as TYPE.  You may take
advantage of this feature to send your own customized message to the who line when
calling Shutdown.  The default package for the TYPE symbol is SI.  Optionally, you may
provide a REASON-STRING that will be used as the message text."
  (If (Null Reason-String)
    (Let ((Prop-Reason (Get Type 'Shutdown-Reason-String)))
      (Setq Reason-String
	    (If (Not (Null Prop-Reason))
	      Prop-Reason
	      (Format () "Shutdown for ~A" Type)))))
  ;; Do miscellaneous cleanup.
  (when (fboundp 'chaos:notify)
    (If Stream
	(Format Stream "~%Notifying network servers"))
    (Setq Chaos::Chaos-Servers-Enabled ())	;Allow no new servers.
    (Notify-All-Servers Reason-String);Notify any friends we have left we're going down.
    )
  (when (fboundp 'Fs::Lmfs-Close-All-Files)
    (If Stream
	(Format Stream "~%Shutting down local file system"))
    ;;Close any open streams our local file system
    ;; is serving.  Not done by file sys dismount,
    ;; but may be redundant considering what comes next.
    (Fs::Lmfs-Close-All-Files)
    )
  (If Stream
      (Format Stream "~%Processing Close-all-files"))
  
  ;; DAB 10-05-89 Close the files before the network gets reset.
  (Close-All-Files);Close any open file streams we've left open.
  (If Stream
      (Format Stream "~%Processing logout"))
  (Logout);Undo user inits and get username off screen.
 
 (when (fboundp 'Chaos:Reset)
    ;; Blow off any network connections
    ;; that are left.  Stand alone after this.
    (Chaos:Reset))
  (when (and (find-package "LX")
	     (funcall (or (find-symbol "LX-ACTIVE-P" "LX")
			  #'false)))
    (If Stream
	(Format Stream "~%sync Unix file system"))
    (ignore-errors
      (with-timeout (300) ; 5 seconds
	(let ((fn (find-symbol "ISSUE-SHELL-COMMAND" "LX")))
	  (funcall fn "sync")
	  (sleep 1/3)
	  (funcall fn "sync")))))
  (when (fboundp 'Fs::Dismount-File-System)
    (Fs::Dismount-File-System);Help keep file system nice.
    )
  ;; Tell user going down now.
  (If Stream
    (Format Stream "~%Shutting down system..."))
  (Without-Interrupts;Don't allow who-line updates etc.
    (Setq tv::Who-Line-Run-State "Shutdown")
    (COND  ((AND (addin-p) (FBOUNDP 'tv:display-mac-mouse-documentation))
	    (tv:display-mac-mouse-documentation
	      t "To restart: Select Warm Boot from the Special menu. To quit: Select Quit from the File menu.")
	    (copy-array-portion tv:who-line-run-state 0 (LENGTH tv:who-line-run-state)
			   tv:mac-who-string 0 (- si:%MX-Who-State-String-Max 1)))
	   ((addin-p) nil)
	   (t  ;;Don't do any of this for addin.
	    (%Open-Mouse-Cursor);Hide the mouse
	    (when (boundp 'W::Who-Line-Documentation-Window)
	      (Catch 'W::Page-Overflow			;;Dont let documentation window overflow
		(progn					;; stop shutdown
		  ;; Make mouse doc window empty.
		  (Send W::Who-Line-Documentation-Window :Clear-Screen)
		  (Send W::Who-Line-Documentation-Window :Line-Out Reason-String)
		  ;; Display message about restarting.
		  (Send W::Who-Line-Documentation-Window :Line-Out
			"To restart:  Hold both control keys and both meta keys and press RUBOUT, M, or ABORT")))
	      (W:Who-Line-Update T)
	      ))))
  ;; If no return requested, really go down by calling %crash.
  ;; First arg of 0 is normal shutdown code; object = nil.
  ;; PAWS-UP-P is t so screen will invert.
  (If (Not Return)
      (%Crash #o0 () T)))
))
