;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-
;;; Written 02/24/89 14:09:24 by GRENINGER,
;;; Reason: Add rubber-banded mouse cursor support.  Add fast dashed-line support.
;;; while running on MX23 from band NB22
;;; With SYSTEM 5.27, GC 5.4, VIRTUAL-MEMORY 5.5, MICRONET 5.6, MICRONET-COMM 5.14,
;;;  DISK-IO 5.9, BASIC-PATHNAME 5.4, MAC-PATHNAME 5.1, NETWORK-SUPPORT-COLD 5.1,
;;;  BASIC-NAMESPACE 5.8, BASIC-FILE 5.5, RPC 5.6, NFS 5.16, EH 5.3, MAKE-SYSTEM 5.3,
;;;  MEMORY-AUX 5.1, MACTOOLBOX 1.35, COMPILER 5.3, TV 5.27, NVRAM 5.1, UCL 5.0, INPUT-EDITOR 5.0,
;;;  METER 5.1, ZWEI 5.13, DEBUG-TOOLS 5.1, WINDOW-MX 5.37, PRINTER 5.12, MAC-PRINTER-TYPES 5.4,
;;;  Experimental NETWORK-PATHNAME 5.1, NETWORK-NAMESPACE 5.0, DATALINK 5.7, CHAOSNET 5.6,
;;;  Experimental NETWORK-SUPPORT 5.1, Experimental NETWORK-SERVICE 5.1, DATALINK-DISPLAYS 5.0,
;;;  NAMESPACE-EDITOR 5.2, IP 3.41, NFS-SERVER 5.3, PRINTER-TYPES 5.5, IMAGEN 5.3,
;;;  MAIL-DAEMON 5.6, MAIL-READER 5.6, TELNET 5.2, VT100 5.1, STREAMER-TAPE 5.7, DECNET 1.53,
;;;  VISIDOC 5.5, PROFILE 5.1, DISK-LABEL 5.1, Experimental MX-SERIAL 1.1,  microcode 96,
;;;  Band Name: microExplorer Network (11/22)





#!C
; From file VARIABLES.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; VARIABLES.#"


(DEFVAR *fast-rb-q-limit* 3
  "Maximum  from-host queue length before RB mouse cursor drawing gets inhibited.")
))


#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"


(Defmacro SWAP-16b (Word)
  (Once-Only (Word)
    `(Dpb (Ldb (Byte 8 0) ,Word) (Byte 8 8) (Ldb (Byte 8 8) ,Word))))

(Defmacro SWAP-32b (Word)
  (Once-Only (Word)
    `(let* ((n (etypecase ,word
		(integer ,word)
		(mac-pointer (send ,word :pointer))
		(mac-handle (send ,word :handle))))
	    (swapped-n (dpb (swap-16b (ldb (byte 16 0) n)) (byte 16 16) (swap-16b (ldb (byte 16 16) n)))))
       (typecase ,word
	 (integer swapped-n)
	 (mac-pointer (make-instance 'mac-pointer :pointer swapped-n))
	 (mac-handle (make-instance 'mac-handle :handle swapped-n))))))




(defmethod (MAC-FLAVOR :OPEN-RUBBER-BAND) (anchor-x anchor-y &optional (direction :from) (pattern :solid) (window tv:selected-window))  
  (let ((acb (add:get-acb-fast 36))
	clip-x1 clip-x2 clip-y1 clip-y2)

    (setf (values clip-x1 clip-y1) (tv:sheet-calculate-offsets window nil))
    (setf clip-x2 (+ clip-x1 (tv:sheet-inside-right window)))
    (setf clip-y2 (+ clip-y1 (tv:sheet-inside-bottom window)))

    (setf (add:opcode acb) 23)
    (setf (add:requestor-complete acb) t)
    (add:load-parms acb 16
		    anchor-x
		    anchor-y
		    *fast-rb-q-limit*
		    (+ (* 256 (ecase pattern (:solid 0) (:dashed 1)))
		       (ecase direction (:none 0) (:from 1) (:to 2)))
		    clip-x1
		    clip-y1
		    clip-x2
		    clip-y2)
    (add:transmit-packet acb mac:channel)))



(defmethod (mac:MAC-FLAVOR :CLOSE-RUBBER-BAND) (&optional leave-line-drawn)
  "Turns off rubber-banding. If leave-line-drawn is non nil the final line is left drawn and its location returned. 
   Otherwise the return value has no meaning."
  (let ((acb (add:get-acb-fast 20)))
    (setf (add:opcode acb) 24)

    (cond (leave-line-drawn
	   (add:load-parms acb 16 1)
	   (setf (add:requestor-complete acb) 1)
	   (add:transmit-packet-and-wait acb mac:channel))
	  (t
	   (add:load-parms acb 16 0)
	   (setf (add:requestor-complete acb) 0)
	   (add:transmit-packet acb mac:channel)))
	   

    (values (add:parm-16b acb 1) (add:parm-16b acb 2) (add:parm-16b acb 3) (add:parm-16b acb 4))

    ))

(DEFMETHOD (mac:mac-flavor :draw-rubber-band) (x1 y1 x2 y2 &optional  (direction :from) (pattern :solid))
  (LET ((acb (add:get-acb-fast 24)))
    (SETF (add:opcode acb) 25)
    (add:load-parms acb 16
		    x1 y1 x2 y2
		    (ECASE direction (:none 0) (:from 1) (:to 2))
		    (ECASE pattern (:solid 0) (:dashed 1)))
    (add:transmit-packet acb mac:channel)))
))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"


(DEFMETHOD (mac-flavor :DrawDashedLine)
	   (window x0 y0 x1 y1 alu dash-spacing space-literally-p offset dash-length color
	    &optional (thickness 1))
  (LET ((pattern (adjust-source-pattern alu color)))
    (MULTIPLE-VALUE-BIND (p1 p2 p3 p4)
        (mac-pattern (IF pattern pattern color))
      (let ((acb (add:get-acb-fast 40)))
	(setf (add:opcode acb) #x6)
	(setf (add:requestor-complete acb) t)
	(add:load-parms acb 16
			p1 p2 p3 p4
			x0 y0 x1 y1
			thickness thickness 
			(transfer-mode alu)
			(Mac-window-id window)
			dash-spacing (IF space-literally-p 1 0)
			offset dash-length
			w:clipping-rectangle-left-edge
			w:clipping-rectangle-top-edge
			(MIN #X3FFF W:CLIPPING-RECTANGLE-RIGHT-EDGE)
			(MIN #X3FFF W:CLIPPING-RECTANGLE-BOTTOM-EDGE))
	(add:transmit-packet acb channel)))))
))

#!C
; From file MAC-MESSAGES.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; MAC-MESSAGES.#"


(DEFUN send-drawdashedline (mac-window x0 y0 x1 y1 alu dash-spacing space-literally-p offset dash-length color
			       &optional (thickness 1))
  (WHEN (AND (tv:sheet-screen-array Mac-window) (NOT *ignore-commands-for-the-Mac*))
    (LET (window-id (add:*no-interrupt* inhibit-scheduling-flag))
      (dump-draw-char-cache)
      (SETF window-id (make-bit-array-Mac-resident Mac-window))
      (remember-call :drawingx window-id)
      (IF  *dont-clip-at-the-margins*
	  (SEND *mac* :DrawDashedLine self x0 y0 x1 y1 alu dash-spacing
	    space-literally-p offset
	    dash-length color thickness)
	;; else...
	(with-clipping-rectangle
	  ((w:sheet-left-margin-size Mac-window)
	   (w:sheet-top-margin-size Mac-window)
	   (- (w:sheet-width Mac-window) (w:sheet-right-margin-size Mac-window))
	   (- (w:sheet-height Mac-window) (w:sheet-bottom-margin-size Mac-window)))
	  (SEND *mac* :DrawDashedLine self x0 y0 x1 y1 alu dash-spacing
	    space-literally-p offset
	    dash-length color thickness))))
      t
      ))

))