;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Changes to DISPLAY-MAC-MOUSE-DOCUMENTATION and WHO-LINE-DOCUMENTATION-FUNCTION-MAC
;;; to handle strings in write-protected memory as is found inside a compiled function.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/06/89 16:31:59 by MARKY,
;;; while running on MX5 from band p526
;;; With SYSTEM 6.4, GC 6.1, VIRTUAL-MEMORY 6.1, MICRONET 6.0, MICRONET-COMM 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-PATHNAME 6.0, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0,
;;;  BASIC-NAMESPACE 6.0, BASIC-FILE 6.0, RPC 6.1, NFS 6.0, EH 6.0, MAKE-SYSTEM 6.0,
;;;  MEMORY-AUX 6.0, COMPILER 6.2, TV 6.6, NVRAM 6.0, UCL 6.0, INPUT-EDITOR 6.0, MACTOOLBOX 2.0,
;;;  METER 6.0, ZWEI 6.0, DEBUG-TOOLS 6.0, WINDOW-MX 6.0, PRINTER 6.1, MAC-PRINTER-TYPES 6.1,
;;;  CLIPBOARD 6.0, NETWORK-PATHNAME 6.0, NETWORK-NAMESPACE 6.0, DATALINK 6.0, CHAOSNET 6.0,
;;;  NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0, MX-DATALINK 6.0,
;;;  NAMESPACE-EDITOR 6.0, IP 3.45, NFS-SERVER 6.0, MX-SERIAL 6.0, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.0,
;;;  DECNET 1.67, VISIDOC 6.0, PROFILE 6.1, TI-CLOS 6.5, CLEH 6.0, CLX 6.0, CLUE 6.0,
;;;  Experimental BUG 11.6,  microcode 137, Band Name: Release 6.0 mx 5/26

;;; SPR 9899 Handle write-protected strings with newlines
;;; such as doc strings in cvv menu of name:CONFIGURE-NAMESPACE 

#!C
; From file MAC-WHOLIN.LISP#> WINDOW-MX; SYS:
#10R TV#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "TV"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; MAC-WHOLIN.#"


(DEFUN WHO-LINE-DOCUMENTATION-FUNCTION-MAC ()
  "This displays the who line documentation for the currently selected
window.  If the selected window is locked an error message is blinked
in the who line area.  Who line documentation may be either a string
of text which is to be displayed or a list of keyword value pairs.  To
see what keywords are accepted see the documentation for the
tv:process-who-line-documentation-list function."
  (LET* ((W MOUSE-WINDOW)
	 (MSG  "*** Error with window locked; try TERMINAL C-CLEAR-INPUT or TERMINAL C-C ***")
	 (MSG1 "    Error with window locked; try TERMINAL C-CLEAR-INPUT or TERMINAL C-C")
	 (NEW-STATE (COND (LOCKED-ERROR-WINDOWS
			    ;; To attract attention, make this message blink.
			    ;; The following EQ test should remain as EQ in spite of what the
			    ;; compiler says.  In this case EQ is both faster and correct.
			    (IF (compiler-let ((inhibit-style-warnings-switch t))
				  (EQ last-mac-mouse-doc-1 MSG))	;Compiler-let added by PMH 7/2/87
				MSG1 MSG))
                          ((SYMBOLP W)
                           (AND W WHO-LINE-MOUSE-GRABBED-DOCUMENTATION))
                          (T (MULTIPLE-VALUE-BIND (DOC ERROR)
                                 (CATCH-ERROR
                                   (FUNCALL W :WHO-LINE-DOCUMENTATION-STRING)
                                   NIL)
                               (IF ERROR
                                   "Error getting documentation string" DOC))))))
    (COND ((AND (NEQ last-mac-mouse-doc-1  NEW-STATE)
                (NOT (EQUAL last-mac-mouse-doc-1  NEW-STATE)))
           (SETQ last-mac-mouse-doc-1  NEW-STATE)
	   (COND ((TYPEP NEW-STATE 'STRING)
		  ;; may 06/06/89 Moved loop that replaced newlines with spaces from here
		  ;; to DISPLAY-MAC-MOUSE-DOCUMENTATION and added copy for write-protected strings. 
		  (display-mac-mouse-documentation t new-state))
		 ((CONSP NEW-STATE)

		  (PROCESS-WHO-LINE-DOCUMENTATION-LIST-MAC NEW-STATE)
		  )
		 (T (CATCH 'PAGE-OVERFLOW
		      (PROCESS-WHO-LINE-DOCUMENTATION-LIST-MAC
			'(:MOUSE-R-1 "Bring up the System Menu."))
		      ))))))
  (unless (ZEROP (LENGTH current-doc-line))
    (display-mac-mouse-documentation t))
  (SETF mac-line 1)
  )

(DEFUN display-mac-mouse-documentation (only-one-line
					&optional (mouse-doc-string current-doc-line))
  ;;Display the current mouse documentation on line mac-line of the mouse documentation display
  ;; may 06/06/89 Start patch ...
  ;; Replace newline chars with spaces. This requires that the string be copied
  ;; into NON-write-protected memory - in case the string is compiled inside a function
  ;; where we CANNOT change it!
  (let ((newline-search-loc -1))
    (loop with modified-string-p = nil do
      (if (setf newline-search-loc (search newline-seq (the string mouse-doc-string) :start2 (incf newline-search-loc)))
	  (progn
	    (unless modified-string-p
	      (setq mouse-doc-string (copy mouse-doc-string)) ; get string OUT of possibly write-protected memory!
	      (setq modified-string-p t))
	    (setf (aref mouse-doc-string newline-search-loc) #\space))
	  (return))))
  ;; may 06/06/89 ... End patch
  (COND
    ((< mac-line 2)
     (copy-array-portion mouse-doc-string 0 (LENGTH mouse-doc-string)
			 mac-who-mouse-doc-string-1 0 (- si:%MX-Mouse-Doc-Max 1))
       (SETF (AREF  mac-who-mouse-doc-string-1  si:%MX-Mouse-Doc-Max ) 1))
    ((= mac-line 2)
     (copy-array-portion mouse-doc-string 0 (LENGTH mouse-doc-string)
			 mac-who-mouse-doc-string-2 0 (- si:%MX-Mouse-Doc-Max 1))
      (SETF (AREF  mac-who-mouse-doc-string-2  si:%MX-Mouse-Doc-Max ) 1))
    (t nil))
  (when (and (> (length mouse-doc-string) si:%MX-Mouse-Doc-Max)
	     (< mac-line si:%MX-Max-mouse-lines))
    (setf current-doc-line (nsubstring mouse-doc-string si:%MX-Mouse-Doc-Max))
    (incf mac-line)
    (display-mac-mouse-documentation only-one-line)
    )
  (setf current-doc-line "")
  (IF only-one-line
      (SETF mac-line 1)  ; also blank out remaining lines.
      (INCF mac-line)))

))
