;;; -*- Mode: Common-Lisp; Package: MACINTOSH; Base: 10.; Patch-File: T -*-

;;; Reason: Fixed acquire-nbytes-of-cache-space to reject any window that's an inferior of one of the wids-not-to-touch. [10954]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 02/06/90 13:56:51 by BERGER,
;;; while running on MX29 from band N102
;;; With SYSTEM 6.28, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.2,
;;;  DISK-IO 6.2, DISK-LABEL 6.0, BASIC-PATHNAME 6.3, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.7, RPC 6.2, NFS-MX 6.4, EH 6.5, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.23, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.12, METER 6.1, ZWEI 6.12, DEBUG-TOOLS 6.4, WINDOW-MX 6.9, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.35, CLEH 6.5, NETWORK-PATHNAME 6.0,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.5, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.57, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.4, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.71, VISIDOC 6.7, PROFILE 6.2,
;;;  Experimental CLX 6.8, CLUE 6.46, Experimental BUG 11.18,  microcode 138, Band Name: esmicroExplorer Network + SLE (8/15)

#!C
; From file MAC-MESSAGES.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; MAC-MESSAGES.#"


(DEFUN acquire-Nbytes-of-cache-space (nbytes-to-acquire wids-not-to-touch)
  (LET* (nbytes-free-space acb-image16)
    (SETF nbytes-free-space (get-modified-rectangles))
    (SETF acb-image16 *modified-rectangles*)
    (LOOP when (>= *nbytes-free-space* nbytes-to-acquire)
	       do
	       ;;  Compact the Mac's bit-array cache...
	       (remember-call :rects *nbytes-free-space*)
	       (get-modified-rectangles)
	       (RETURN t)

	  for oldest-dbase =
	  (LOOP with oldest-age = #x10000000 and oldest-dbase
		for i from 1 to *n-descriptors*
		for dbase from 22 by 8
		for age = (+ (ASH (AREF acb-image16 (+ dbase 3)) 16)
			     (AREF acb-image16 (+ dbase 2)))
		for wid = (LSH (AREF acb-image16 (+ dbase 1)) -8)
		for window = (AREF *all-windows-and-screens* wid)
		when (AND (< age oldest-age)
			  window
			  (NOT (TYPEP window 'mac-screen))
			  (NOT (MEMBER wid wids-not-to-touch))
			  (NOT (LOOP for sup first tv:selected-window then
				     (tv:sheet-superior sup)
				     until (NULL sup)
				     when (EQ window tv:selected-window)
				     do; (FORMAT t "~%~s rejected #1." window)
				     (RETURN t)))
			  (NOT (LOOP for sup first window then
				     (tv:sheet-superior sup)
				     until (NULL sup)
				     when (or (EQ sup tv:selected-window)
					      (member (tv:sheet-window-id sup) wids-not-to-touch))  ; DAB 02-06-90 FOR LG
				     do; (FORMAT t "~%~s rejected #2." window)
				     (RETURN t)))
			  )
		do
		(SETF oldest-age age)
		(SETF oldest-dbase dbase)
		finally (RETURN oldest-dbase))
	  when (NULL oldest-dbase) return NIL
	  for oldest-wid = (LSH (AREF acb-image16 (+ oldest-dbase 1)) -8)
	  for oldest-window = (AREF *all-windows-and-screens* oldest-wid)
	  when (NULL oldest-window)
	  do
	  (LET ((si:cold-load-stream-owns-keyboard t)
		  (*terminal-io* si:cold-load-stream))
	      (eh:save-screen-for-cold-load-stream t)
	      (CERROR "Proceed, throwing away recent updates for window-id ~d"
		      "Mac has bitarray cached for window-id ~d not currently assigned to a window"
		      oldest-wid)
	      (eh:restore-screen-for-cold-load-stream))
	  (SEND *mac* :deallocatebitarray oldest-wid)
	  (SETF (AREF acb-image16 (+ oldest-dbase 2)) 0
		(AREF acb-image16 (+ oldest-dbase 3)) #x1000)
	  (emulate-cache-deallocation oldest-wid)
	  else do

#|
	  (WITH-STACK-LIST (modified-area-descriptor (AREF acb-image16 (+ oldest-dbase 5))
						     (AREF acb-image16 (+ oldest-dbase 4))
						     (AREF acb-image16 (+ oldest-dbase 7))
						     (AREF acb-image16 (+ oldest-dbase 6)))
|#
          (WITH-STACK-LIST (modified-area-descriptor 0 0 (tv:sheet-width oldest-window)
						     (tv:sheet-height oldest-window))
	    (retrieve-a-bit-array-from-the-Mac oldest-window modified-area-descriptor))
	  (remember-call :rects oldest-window *nbytes-free-space*)
	  ;;  Remove this window's entry from further consideration by giving it
	  ;;  a very large time-of-last-update...
	  (SETF (AREF acb-image16 (+ oldest-dbase 2)) 0
		(AREF acb-image16 (+ oldest-dbase 3)) #x1000))))
))
