;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Clean up deactivated windows whenever fewer than 64 window
;;; ids exist in order to minimize likelyhood of going to cold load
;;; stream and minimize number of Mac grafports open simultaneously.
;;; [SPR 10093]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/16/89 14:44:42 by GRENINGER,
;;; while running on MX23 from band P608
;;; With SYSTEM 6.6, GC 6.2, VIRTUAL-MEMORY 6.1, MICRONET 6.0, MICRONET-COMM 6.1,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-PATHNAME 6.0, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0,
;;;  BASIC-NAMESPACE 6.0, BASIC-FILE 6.1, RPC 6.1, NFS 6.0, EH 6.0, MAKE-SYSTEM 6.0,
;;;  MEMORY-AUX 6.0, COMPILER 6.2, TV 6.6, NVRAM 6.0, UCL 6.0, INPUT-EDITOR 6.0, MACTOOLBOX 2.0,
;;;  METER 6.0, ZWEI 6.0, DEBUG-TOOLS 6.0, WINDOW-MX 6.1, PRINTER 6.1, MAC-PRINTER-TYPES 6.1,
;;;  CLIPBOARD 6.0, NETWORK-PATHNAME 6.0, NETWORK-NAMESPACE 6.0, DATALINK 6.0, CHAOSNET 6.0,
;;;  NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.0, DATALINK-DISPLAYS 6.0, MX-DATALINK 6.0,
;;;  NAMESPACE-EDITOR 6.0, IP 3.46, NFS-SERVER 6.0, MX-SERIAL 6.0, PRINTER-TYPES 6.0,
;;;  IMAGEN 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.0,
;;;  DECNET 1.67, VISIDOC 6.0, PROFILE 6.1, TI-CLOS 6.5, CLEH 6.0, Experimental CLX 6.1,
;;;  CLUE 6.4, Experimental BUG 11.7,  microcode 138, Band Name: Rel 6.0 mxnet 6/8

#!C
; From file MAC-WINDOWS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; MAC-WINDOWS.#"


(DEFUN allocate-Mac-Explorer-window-id ()
  "Returns a positive integer to be used as a Mac-Explorer window id."
  (DECLARE (VALUES Mac-Explorer-window-id))
  (WITH-LOCK (*window-id-lock*)
    (LET (id n-wids-left)
      ;;  Notify user if we're about to run out of window-ids...
      (SETF n-wids-left (- *maximum-legal-mac-explorer-window-id* *largest-window-id-ever-assigned*))
      (WHEN (<= n-wids-left 64)
	(INCF n-wids-left (LENGTH *list-of-free-window-ids*))
	(WHEN (<= n-wids-left 64)
	  (clean-up-deactivated-windows)
	  (INCF n-wids-left (LENGTH *list-of-free-window-ids*)))
	(WHEN (MEMBER n-wids-left '(10 6 3  1))
	  (let ((si:cold-load-stream-owns-keyboard t)
		(*terminal-io* si:cold-load-stream))
	    (eh:save-screen-for-cold-load-stream t)
	    (CERROR "Proceed" "Only ~d free window-id~:P left." n-wids-left)
	    (eh:restore-screen-for-cold-load-stream))))
      (SETF id
	    ;;  First try to allocate a previously used window id...
	    (IF *list-of-free-window-ids*
		(POP *list-of-free-window-ids*)
	      ;; else try to allocate an unused window id...
	      (IF (< *largest-window-id-ever-assigned*
		     *maximum-legal-mac-explorer-window-id*)
		    (INCF *largest-window-id-ever-assigned*)
		;; else flush all deactivated windows, try for one of their window ids...
		(clean-up-deactivated-windows)
		(IF *list-of-free-window-ids*
		    (POP *list-of-free-window-ids*)
		   ;; else we're out of window ids (and luck)...
		  (IF (si:mx-p)
		      (PROG1 nil
			     (let ((si:cold-load-stream-owns-keyboard t)
				   (*terminal-io* si:cold-load-stream))
			       (eh:save-screen-for-cold-load-stream t)
			       (ferror nil "You are trying to create 257 active windows, 256 is the maximum.")
			       (eh:restore-screen-for-cold-load-stream)))
		    ;; else...
		    (reset-Mac-explorer-connection
		      '(tv:kbd-terminal-arrest-all nil)))))))
      (remember-call :id-assignment id self)
      id)))

))
