;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Load new compilations of all window-mx functions who bind 
;;; add:*no-interrupt* and call add:transmit-packet to fix the build 
;;; problem that omitted all bindings of add:*no-interrupt* from these 
;;; functions' executables. [SPR 10068]

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 06/28/89 14:34:03 by GRENINGER,
;;; while running on MX23 from band P608
;;; With SYSTEM 6.9, GC 6.3, VIRTUAL-MEMORY 6.1, MICRONET 6.0, MICRONET-COMM 6.1,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-PATHNAME 6.0, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0,
;;;  BASIC-NAMESPACE 6.1, BASIC-FILE 6.2, RPC 6.1, NFS 6.0, EH 6.3, MAKE-SYSTEM 6.0,
;;;  MEMORY-AUX 6.0, COMPILER 6.6, TV 6.11, NVRAM 6.1, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.3, METER 6.0, ZWEI 6.3, DEBUG-TOOLS 6.3, WINDOW-MX 6.3, PRINTER 6.1,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.0, NETWORK-PATHNAME 6.0, NETWORK-NAMESPACE 6.0,
;;;  DATALINK 6.0, CHAOSNET 6.0, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  MX-DATALINK 6.1, NAMESPACE-EDITOR 6.0, IP 3.47, NFS-SERVER 6.0, MX-SERIAL 6.0,
;;;  PRINTER-TYPES 6.0, IMAGEN 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.0, TELNET 6.0,
;;;  VT100 6.0, STREAMER-TAPE 6.3, DECNET 1.69, VISIDOC 6.2, PROFILE 6.1, TI-CLOS 6.9,
;;;  CLEH 6.4, Experimental CLX 6.1, CLUE 6.5, Experimental BUG 11.10,  microcode 138,
;;;  Band Name: Rel 6.0 mxnet 6/8

;;;  Guarantee that just in case this file gets compiled on an Explorer
;;;  that the variable add:*no-interrupt* is seen as special...
(PROCLAIM '(SPECIAL add:*no-interrupt*))

#!C
; From file MAC-MESSAGES.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; MAC-MESSAGES.#"


(DEFUN send-create-screen ()
  "Called when creating the Mac screen to (1) tell the Mac to create a window to receive
our commands and (2) get the size of this Mac window -- our screen."
  (DECLARE (VALUES Mac-screen-height Mac-screen-width))
  (LET* ((inhibit-scheduling-flag t)
	 (add:*no-interrupt* inhibit-scheduling-flag))
    (UNLESS *ignore-commands-for-the-Mac*
      (dump-draw-char-cache)
      (SEND *mac* :are-you-there?)
      (MULTIPLE-VALUE-BIND (width height)
	  (SEND *mac* :GetScreenSize)
	(remember-call :selection height width)
	(VALUES height width)))))

))

#!C
; From file MAC-WINDOWS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; MAC-WINDOWS.#"


(DEFUN set-direct-drawing-mode (new-mode)
  (CHECK-ARG new-mode (MEMBER new-mode '(:on :off))
	     ":ON or :OFF")
  (LET* ((inhibit-scheduling-flag t)
	 (add:*no-interrupt* inhibit-scheduling-flag)
	 (acb (add:get-acb 4))
	 (channel (add:find-channel si:%Chan-Type-Misc)))
    (add:init-acb acb
		  si:%MC-tvcalls
		  si:%TC-DIRECT-DRAW)
    (add:set-parm-8b acb 0 (IF (EQ new-mode :ON) 1 0))
    (SETF (add:requestor-complete acb) t)
    (add:transmit-packet acb channel)))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :set-mouse-blinker)
	   (CHARACTER &optional glyph mask x-offset y-offset
	    &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Sends a message to the Mac to use a different cursor."
  (LET ((acb (add:get-acb-fast 144)))
    (COND ((NOT (ZEROP character))		       ; Send just the resource number if specified.
	   (add:load-parms acb 16 (+ mac-mouse-cursor-offset character) 0))
	  
	  (t
	   (WITHOUT-INTERRUPTS
	     ;; else send a zero resource number, the glyph, the mask, and the hot spot...
	     (ADJUST-ARRAY *art-16b-glyph* 16 :displaced-to glyph)
	     (ADJUST-ARRAY *art-16b-mask* 16 :displaced-to mask)
	     (add:load-parms acb 16
			     0
			     0
			     (AREF *art-16b-glyph* 0)  ; The glyph...
			     (AREF *art-16b-glyph* 1)
			     (AREF *art-16b-glyph* 2)
			     (AREF *art-16b-glyph* 3)
			     (AREF *art-16b-glyph* 4)
			     (AREF *art-16b-glyph* 5)
			     (AREF *art-16b-glyph* 6)
			     (AREF *art-16b-glyph* 7)
			     (AREF *art-16b-glyph* 8)
			     (AREF *art-16b-glyph* 9)
			     (AREF *art-16b-glyph* 10)
			     (AREF *art-16b-glyph* 11)
			     (AREF *art-16b-glyph* 12)
			     (AREF *art-16b-glyph* 13)
			     (AREF *art-16b-glyph* 14)
			     (AREF *art-16b-glyph* 15)
			     (AREF *art-16b-mask* 0)   ; The mask...
			     (AREF *art-16b-mask* 1)
			     (AREF *art-16b-mask* 2)
			     (AREF *art-16b-mask* 3)
			     (AREF *art-16b-mask* 4)
			     (AREF *art-16b-mask* 5)
			     (AREF *art-16b-mask* 6)
			     (AREF *art-16b-mask* 7)
			     (AREF *art-16b-mask* 8)
			     (AREF *art-16b-mask* 9)
			     (AREF *art-16b-mask* 10)
			     (AREF *art-16b-mask* 11)
			     (AREF *art-16b-mask* 12)
			     (AREF *art-16b-mask* 13)
			     (AREF *art-16b-mask* 14)
			     (AREF *art-16b-mask* 15)
			     x-offset		       ; The hot spot...
			     y-offset
			     ))))

    (SETF (add:opcode acb) #x7)
    (SETF (add:requestor-complete acb) t)
    (add::transmit-packet acb channel)))

;----------------------------------------------------------------------------------------

;;;
;;; Window Commands
;;;


))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :GetScreenSize)
	   (&aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Sets the screen-width and screen-height instance variables to the current size
of the LISPM window on the Mac screen. Returns the width and height."
						       ;(BREAK " :GetScreenSize")
  (LET (screen-resizing-at-boot-p
	(acb (add:get-acb-fast 8)))
    (SETF (add:opcode acb) #x0f)
    (add::transmit-packet-and-wait acb channel)
    (SETF screen-width (MAX 0 (- (add:parm-16b acb 2)
				 (add:parm-16b acb 0)
				 ))
	  screen-height (MAX 0 (- (add:parm-16b acb 3)
				  (add:parm-16b acb 1)
				  ))
	  screen-resizing-at-boot-p (NOT (ZEROP (add:parm-16b acb 4))))
    (add:return-acb acb t)
    (VALUES screen-width screen-height screen-resizing-at-boot-p)))


))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :AdjustBitArray)
	   (window width height contents-matter &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Causes the Mac to create a graphport to hold the bit array of window.
If it already exists then the Mac will resize it to width,height. If width and
height are zero then the graphport will have no bit array."
						       ;(BREAK " :AdjustBitArray")
  (LET (return-value
	(acb (add:get-acb-fast 8)))
    
    (add:load-parms acb 16
		    (Mac-window-id window)
		    width
		    height
		    (IF contents-matter 1 0))
    
    (SETF (add:opcode acb) #x10)
    (add::transmit-packet-and-wait acb channel)
    
    (SETF return-value (NOT (ZEROP (ADD:PARM-16B acb 0))))
    (add:return-acb acb t)
    return-value))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"


		
(DEFMETHOD (MAC-flavor :RedirectDrawing)	       ;
	   (window-being-redirected bitmap-redirected-to x y &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Updates the drawing indirection table maintained on the Mac. Drawing
done to window-being-redirected will actually take place to bitmap-redirected-to.
If (Mac-window-id window-directed-to) is zero then drawing goes to the screen."
						       ;(BREAK " :RedirectDrawing")
  (LET ((acb (add:get-acb-fast 8)))
    (SETF (add:opcode acb) #x11)
    (add:load-parms
      acb
      16
      (Mac-window-id window-being-redirected)
      (Mac-window-id bitmap-redirected-to)
      x
      y)
    (SETF (add:requestor-complete acb) t)
    (add::transmit-packet acb channel)))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DeallocateBitArray)
	   (window &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Allows the Mac to destroy the graphport and reclaim the memory space
associated with this window."
  (LET ((acb (add:get-acb-fast 2)))
    (remember-call :selection self)
    (SETF (add:opcode acb) #x12) 
    (add:load-parms acb 16 (Mac-window-id window))
    (SETF (add:requestor-complete acb) t)
    (add::transmit-packet acb channel)))


))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :TotalCacheSize) ()
  "Returns the number of bytes in the Mac's bit-array cache.  Uses a
strange option of the AdjustScreenSize command to avoid allocating a new
ACB/command opcode for this minor command."
  (LET ((add:*no-interrupt* inhibit-scheduling-flag)
	(acb (add:get-acb-fast 10))
	return-value)
    
    (SETF (add:opcode acb) #x13)    
    (add:load-parms acb 16 0 0 0 0 4)
    (add::transmit-packet-and-wait acb channel)
    (SETF return-value (ADD:PARM-32B acb 0))
    (add:return-acb acb t)
    return-value))


))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :AdjustScreenSize)
	   (Explorer-screen-ID Explorer-window-ID width height visibility)
  "Creates or changes the size of a Mac-resident Explorer screen."
  (LET ((add:*no-interrupt* inhibit-scheduling-flag)
	(acb (add:get-acb-fast 10))
	return-value)
    (SETF (add:opcode acb) #x13)
    ;;(BREAK " :AdjustScreenSize")
    (add:load-parms acb 16
		    Explorer-screen-ID Explorer-window-ID width height 
		    (TYPECASE visibility
		      (symbol (IF visibility 1 0))
		      (fixnum visibility)))

    (add::transmit-packet-and-wait acb channel)
    (remember-call :adjscr (ADD:PARM-16B acb 0))
    (SETF return-value (NOT (= #xFFFF (ADD:PARM-16B acb 0))))
    (add:return-acb acb t)
    return-value))


))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :SelectWindow)
	   (window hide-instead-of-select-p)
  "Make the Mac window containing the specified Explorer window the
top-most Mac window so the user can see it."
  (LET ((add:*no-interrupt* inhibit-scheduling-flag)
	(acb (add:get-acb-fast 4)))

    (SETF (add:opcode acb) #x14)
    (SETF (add:requestor-complete acb) t)
    (add:load-parms acb 16 (Mac-window-id window) (IF hide-instead-of-select-p 1 0))
    (add::transmit-packet acb channel)))


))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :GetModifiedRectangles)
	   (&aux (add:*no-interrupt* inhibit-scheduling-flag))
  (LET (acb n-windows-with-mac-resident-bitmaps acb-image)
    (SETF n-windows-with-mac-resident-bitmaps
	  (LOOP for (ARRAY . window) in *undisplaced-mac-window-arrays*
		for wid = (tv:sheet-window-id window)
		count (AND (INTEGERP wid)
			   (NOT (AREF *has-the-Mac-forgotten-my-redirection?* wid)))))
    (SETF acb (add:get-acb-fast (+ 4 (* 16 n-windows-with-mac-resident-bitmaps) 128)))
    (SETF (add:opcode acb) #x08)
    (add::transmit-packet-and-wait acb channel)
    (SETF acb-image (MAKE-ARRAY (ARRAY-TOTAL-SIZE acb) :element-type '(unsigned-byte 16)))
    (COPY-ARRAY-CONTENTS acb acb-image)
    (add:return-acb acb t)
    acb-image))
;----------------------------------------------------------------------------------------

;;;;
;;;; Mac commands
;;;;

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawLine)
	   (x1 y1 x2 y2 alu draw-end-point window
	    &optional (line-width 1) (color w:black)
	    &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Draws a line on the Mac screen from x1,y1 to x2,y2 with LISPM alu."
  (DECLARE (IGNORE draw-end-point))
						       ;(BREAK " :DrawLine")
  ;; If line-width <> 1 then a complex calculation of pen-width and pen-height needs to go here...
  ;;   It would calc pen width and height from the geometric relationship of line angle and line width.
  ;;   Line end point will also have to be adjusted if pen width and height are not 1
  ;;   For speed should look for special cases of vertical and horizontal line.
  (LET ((pattern (adjust-source-pattern alu color)))
    (MULTIPLE-VALUE-BIND (p1 p2 p3 p4)
        (mac-pattern (IF pattern pattern color))
      (LET ((acb (add:get-acb-fast 32)))
	(SETF (add:opcode acb) #x4)
	(SETF (add:requestor-complete acb) t)
	(add:load-parms acb 16
			p1 p2 p3 p4
			x1 y1 x2 y2
			line-width line-width	       ;simple way out for now
			(transfer-mode alu)
			(Mac-window-id window)
			w:clipping-rectangle-left-edge
			w:clipping-rectangle-top-edge
			(MIN #X3FFF W:CLIPPING-RECTANGLE-RIGHT-EDGE)
			(MIN #X3FFF W:CLIPPING-RECTANGLE-BOTTOM-EDGE))
	(add::transmit-packet acb channel)))))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"


    
(DEFMETHOD (mac-flavor :Nop)
	   (&aux (add:*no-interrupt* inhibit-scheduling-flag))
  "This sends a nil command to the Mac. Mostly for timing purposes. This requires
330us. It represents the fastest we will be able to deliver commands to the Mac."
  ;(BREAK " :Nop")
  (LET ((acb (add:get-acb-fast 2)))
    (SETF (add:opcode acb) #xff)
    (SETF (add:requestor-complete acb) t)
    (add::transmit-packet acb channel)))


;;; Text output routines

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawCharacter)
	   (font char x y alu window &optional size)
  "Draw LISPM CHAR on the Mac screen at x,y in LISPM FONT with LISPM ALU.
If size is specified then it it used instead of the standard font size and the
font baseline is assumed to be size."
  (LET ((add:*no-interrupt* inhibit-scheduling-flag))
    (font-cache font)
    (LET ((acb (add:get-acb-fast 24)))
      (SETF (add:opcode acb) #x1)
      (SETF (add:requestor-complete acb) t)
      (add:load-parms acb 16
		      (AREF (Character-Mapping-Table font)
			    (CHAR-CODE char))
		      current-mac-font
		      current-mac-font-style
		      (OR size current-mac-font-size)
		      x (+ y (OR size current-mac-font-baseline))
		      (transfer-mode alu)
		      (Mac-window-id window)
		      w:clipping-rectangle-left-edge
		      w:clipping-rectangle-top-edge
		      (MIN #X3FFF W:CLIPPING-RECTANGLE-RIGHT-EDGE)
		      (MIN #X3FFF W:CLIPPING-RECTANGLE-BOTTOM-EDGE))
      (add::transmit-packet acb channel))))


))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawString)
	   (font string index end xpos ypos xlim alu window &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Draws string from index to end on the Mac."
  (DECLARE (inline graphic-char-p))
  (LET ((char-mapping-table (Character-Mapping-Table font))
	(fwt (w:font-char-width-table font))
	ch
	(index-of-first-char-to-not-draw index)
	(xpos-to-stop-at xpos)
	(limit (OR xlim w:clipping-rectangle-right-edge))
	(acb (add:get-acb-fast (+ (* 2 DrawString-parms) DrawString-max-size))))
    
    (SETF (add:opcode acb) #x3)
    
    (font-cache font)
    
    ;;Calculate length of string that fits in the clipping rectangle and stuff translated characters into the ACB.
    (DOTIMES (j (1- drawstring-max-size))
      (UNLESS (< index-of-first-char-to-not-draw end)
	(RETURN))			   ; reached the end
      (UNLESS (GRAPHIC-CHAR-P
		(SETQ ch (CHAR-CODE (AREF string index-of-first-char-to-not-draw))))
	(RETURN))			   ; non printing char
      (IF (< limit (INCF xpos-to-stop-at (AREF fwt ch)))
	  (PROGN (DECF xpos-to-stop-at (AREF fwt ch))	   ; this character goes too far
		 (RETURN))	   
	  ;; else
	  (SETF (add:parm-8b acb (+ (* 2 Drawstring-Parms) j))
		(AREF char-mapping-table ch)) 
	  (INCF index-of-first-char-to-not-draw)))
    
    (remember-call :drawing window xpos-to-stop-at index-of-first-char-to-not-draw)
    
    (add:load-parms-16b
      acb
      current-mac-font
      current-mac-font-style 
      current-mac-font-size 
      (transfer-mode alu)
      0
      (- index-of-first-char-to-not-draw index)
      xpos (+ ypos current-mac-font-baseline)
      xpos-to-stop-at 
      (Mac-window-id window)
      w:clipping-rectangle-left-edge
      w:clipping-rectangle-top-edge
      (MIN #x3FFF w:clipping-rectangle-right-edge)
      (MIN #x3FFF w:clipping-rectangle-bottom-edge))
    (SETF (add:requestor-complete acb) t)
    (add::transmit-packet acb channel)
    
    (DPB (- index-of-first-char-to-not-draw index)
	 (BYTE 12. 12.)
	 xpos-to-stop-at)))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :FontMetrics)
	   (font &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Gets the global width table the Macs version of LISPM-font. The returned
table is a 256 entry array indexed by the ASCII code. Although the Mac returns
fixed point values for the width the routine rounds off the fractional
part and returns an array of integers. Note that kerning information is
not returned."
  (font-cache font)
  
  (LET ((acb (add:get-acb-fast 1050))
	ascent
	descent
	leading
	widMax
	width-table-length
	vScale
	hScale
	width-table)
    
    (add:load-parms acb 16
		    current-mac-font
		    current-mac-font-style
		    current-mac-font-size)
    (SETF (add:opcode acb) #xd)
    (add::transmit-packet-and-wait acb channel)
    
    (SETF ascent (read-fix-32 (add:parm-16b acb 0)     ;Integer part
			      (add:parm-16b acb 1)))   ;Fractional part 
    (SETF descent (read-fix-32 (add:parm-16b acb 2)  
			       (add:parm-16b acb 3)))
    (SETF leading (read-fix-32 (add:parm-16b acb 4)  
			       (add:parm-16b acb 5)))
    (SETF widMax (read-fix-32 (add:parm-16b acb 6)  
			      (add:parm-16b acb 7)))
    (SETF width-table-length 256)
    (SETF vScale (* (read-fix-16 (add:parm-16b acb
					       (+ 8  (* 2 width-table-length))))
		    (read-fix-16 (add:parm-16b acb
					       (+ 10 (* 2 width-table-length))))))
    (SETF hScale (* (read-fix-16 (add:parm-16b acb
					       (+ 9  (* 2 width-table-length))))
		    (read-fix-16 (add:parm-16b acb
					       (+ 11 (* 2 width-table-length))))))
    
    (SETF vScale 1.0)				       ;Mac IV-41 says VFactor is not used. KED 3/1/88
    (SETF width-table (MAKE-ARRAY width-table-length  
				  :fill-pointer #xFF
				  :leader-length 6
				  :leader-list	   
				  `(0		       ;Space for fill-pointer (unused) 
				     ,(ROUND (* widMax hScale))	       ; See below.
				     ,(ROUND (* (+ ascent descent) vScale)) ;;height
				     ,(ROUND (* ascent vScale))
				     ,(ROUND (* descent vScale))
				     ,(ROUND (* leading vScale))
				     )))
    (LOOP with char-mapping-table = (Character-Mapping-Table font)
	  for char from 0 to 255
	  for translated-char = (AREF char-mapping-table char)
	  do (SETF (AREF width-table char)
		   (ROUND (* hScale
			     (read-fix-32 (add:parm-16b
					    acb
					    (+ 8 (* 2 translated-char)))
					  (add:parm-16b
					    acb
					    (+ 9 (* 2 translated-char))))))))
    (SETF (ARRAY-LEADER width-table 1)
	  (AREF width-table (CHAR-CODE #\m))); AB for GG  ;Use size of space char for tv:font-char-width
    (add:return-acb acb t)
    width-table))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawTriangle)
	   (x0 y0 x1 y1 x2 y2
	    alu
	    draw-first-edge draw-second-edge draw-third-edge
	    pattern-or-nil
	    window &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Like si:%draw-shaded-triangle but draws it on the Mac screen.
For now edges are always drawn."
  (DECLARE (IGNORE draw-first-edge draw-second-edge draw-third-edge))
  (LET ((pattern (adjust-source-pattern alu pattern-or-nil)))
    (MULTIPLE-VALUE-BIND (p1 p2 p3 p4)
	(mac-pattern (IF pattern pattern w:black))
						       ;(BREAK " :DrawTriangle")
      (LET ((acb (add:get-acb-fast 32)))
	(SETF (add:opcode acb) #x5)
	(SETF (add:requestor-complete acb) t)
	(add:load-parms acb 16
			p1 p2 p3 p4
			x0 y0 x1 y1 x2 y2
			(transfer-mode alu)
			(Mac-window-id window)
			w:clipping-rectangle-left-edge
			w:clipping-rectangle-top-edge
			(MIN #x3FFF w:clipping-rectangle-right-edge)
			(MIN #x3FFF w:clipping-rectangle-bottom-edge))
	(add::transmit-packet acb channel)))))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawFilledPolygon)
	   (vertices alu fill-pattern-or-nil window
	    &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Like si:%draw-shaded-triangle but draws an arbitrary filled polygon on the Mac
screen. For now edges are always drawn."
  (LET ((pattern (adjust-source-pattern alu fill-pattern-or-nil)))
    (MULTIPLE-VALUE-BIND (p1 p2 p3 p4)
	(mac-pattern (IF pattern pattern w:black))
      (LET* ((n-vertices (/ (LENGTH vertices) 2))
	     (acb (add:get-acb-fast (* 2 (+ 12 n-vertices n-vertices)))))
	(SETF (add:opcode acb) 14)
	(SETF (add:requestor-complete acb) t)
	(add:load-parms acb 16
			p1 p2 p3 p4
			(transfer-mode alu)
			(Mac-window-id window)
			n-vertices 0
			w:clipping-rectangle-left-edge
			w:clipping-rectangle-top-edge
			(MIN #x3FFF w:clipping-rectangle-right-edge)
			(MIN #x3FFF w:clipping-rectangle-bottom-edge))
	(LOOP for i from 32
	      for x-y-coord in vertices
	      do (SETF (AREF acb i) x-y-coord))
	(add::transmit-packet acb channel)))))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawRectangle)
	   (x y width height color alu window &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Like si:%draw-rectangle but draws it on the Mac screen." 
  (MULTIPLE-VALUE-BIND (p1 p2 p3 p4)
      (mac-pattern (adjust-source-pattern alu color))
						       ;(BREAK " :DrawRectangle")
    (LET ((acb (add:get-acb-fast 28)))
      (SETF (add:opcode acb) #x9)
      (SETF (add:requestor-complete acb) t)
      (add:load-parms acb 16
		      p1 p2 p3 p4
		      x y
		      (+ x width) (+ y height)
		      (transfer-mode alu)
		      (Mac-window-id window)
		      w:clipping-rectangle-left-edge
		      w:clipping-rectangle-top-edge
		      (MIN #x3FFF w:clipping-rectangle-right-edge)
		      (MIN #x3FFF w:clipping-rectangle-bottom-edge))
      (add::transmit-packet acb channel))))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawHollowRectangle)
	   (x y width height thickness color alu window &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "Like si:%draw-rectangle but draws a hollow rectangle on the Mac screen." 
  (MULTIPLE-VALUE-BIND (p1 p2 p3 p4)
      (mac-pattern (adjust-source-pattern alu color))
						       ;(BREAK " :DrawHollowRectangle")
    (LET ((acb (add:get-acb-fast 32)))
      (SETF (add:opcode acb) #xa)
      (SETF (add:requestor-complete acb) t)
      
      (add:load-parms acb 16
		      p1 p2 p3 p4
		      x y
		      (+ x width) (+ y height)
		      (transfer-mode alu)
		      thickness thickness
		      (Mac-window-id window)
		      w:clipping-rectangle-left-edge
		      w:clipping-rectangle-top-edge
		      (MIN #x3FFF w:clipping-rectangle-right-edge)
		      (MIN #x3FFF w:clipping-rectangle-bottom-edge))
      (add::transmit-packet acb channel))))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawCircle)
	   (x y radius
	    &optional
	    (color w:black)
	    (alu w:alu-seta)
	    (num-points 29)
	    (draw-edge t) window &aux (add:*no-interrupt* inhibit-scheduling-flag))
  "This draws a filled oval just inside the rectangle specified by left, top, right, and bottom.
Like the graphics mixin :draw-filled-circle."
  (DECLARE (IGNORE num-points))
  (LET ((r (MAX 0 (- radius (IF draw-edge 0 1)))))
    (MULTIPLE-VALUE-BIND (p1 p2 p3 p4)
	(mac-pattern (adjust-source-pattern alu color))
						       ;(BREAK " :DrawCircle")
      (LET ((acb (add:get-acb-fast 28)))
	(SETF (add:opcode acb) #xb)
	(SETF (add:requestor-complete acb) t)
	
	(add:load-parms acb 16
			p1 p2 p3 p4
			(- x r) (- y r)
			(+ x r) (+ y r)
			(transfer-mode alu)
			(Mac-window-id window)
			w:clipping-rectangle-left-edge
			w:clipping-rectangle-top-edge
			(MIN #X3FFF W:CLIPPING-RECTANGLE-RIGHT-EDGE)
			(MIN #X3FFF W:CLIPPING-RECTANGLE-BOTTOM-EDGE))
	(add::transmit-packet acb channel)))))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :DrawHollowCircle)
	   (x y radius
	    &optional
	    (thickness 1)
	    (color w:black)
	    (alu w:alu-seta)
	    (num-points 29) window)
  "This draws a filled oval just inside the rectangle specified by left, top, right, and bottom.
Like the graphics mixin :draw-circle."
  (DECLARE (IGNORE num-points)) 
  (LET ((add:*no-interrupt* inhibit-scheduling-flag))
    (MULTIPLE-VALUE-BIND (p1 p2 p3 p4)
	(mac-pattern (adjust-source-pattern alu color))
      ;;(BREAK " :DrawHollowCircle")
      
      (LET ((acb (add:get-acb-fast 32)))
	(SETF (add:opcode acb) #xc)
	(SETF (add:requestor-complete acb) t)
	
	(add:load-parms acb 16
			p1 p2 p3 p4
			(- x radius) (- y radius)
			(+ x radius) (+ y radius)
			(transfer-mode alu)
			thickness thickness
			(Mac-window-id window)
			w:clipping-rectangle-left-edge
			w:clipping-rectangle-top-edge
			(MIN #X3FFF W:CLIPPING-RECTANGLE-RIGHT-EDGE)
			(MIN #X3FFF W:CLIPPING-RECTANGLE-BOTTOM-EDGE))
	
	(add::transmit-packet acb channel)))))



;;  This is rather dirty but since we know that *bitblt-indirect-array*
;;  is always a displaced array, a physical array, and has 2 dimensions,
;;  and all we ever want to do is change the dimensions of this array in a
;;  way that will never violate the physical storage underlying the array,
;;  this is the fastest way to do it...

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac-flavor :CopyBits)
	   (alu
	    width height
	    from-array from-x from-y
	    to-array to-x to-y
	    &optional (source-width width) (source-height height) replication delay-updates
	    restoration-p
	    &aux (add:*no-interrupt* inhibit-scheduling-flag) (add:*wait-error* nil))
  "Copies the from array to an intermediate physical array and then to the to array. 
The purpose of the intermediate array is to word align it (adjust for source bitwise misalignment) and to
linearize it in physical memory.
    The source and destination are controlled by the type of from-array and to-array:
    IF from-array is number and to-array is a number then the transfer takes place between
two Mac bit maps with the number being the bitmap ID. A zero ID indicates
the source or destination is the screen.
    IF from-array is an array and to array is a number then the transfer takes place
from the LISPM array to a bitmap on the Mac.
    If from-array is a number and to-array is an array then the transfer takes place
from a Mac bitmap to a LISPM array.
    If both from-array and to-array are arrays the transfer is between arrays on the
LISPM."
  (LET (word-width)
    (SETF width (ABS width)
          height (ABS height))
    
    (UNLESS source-width
      (SETF source-width width))

    (UNLESS source-height
      (SETF source-height height))
    
    (COND
      ((AND (NUMBERP from-array) (NUMBERP to-array))	   ;Mac to Mac case	     
       ;(BREAK " :CopyBits Mac-to-Mac")
       (LET ((acb (add:get-acb-fast (* 2 (force-even CopyBits-parms)))))
	 (SETF (add:opcode acb) #x0)
	 (SETF (add:requestor-complete acb) t)
	 (add:load-parms acb 16
			 1
			 (transfer-mode alu)	       ;Mac sets srcMode
			 width height		   
			 from-array
			 to-array
			 from-x from-y
			 to-x to-y
			 source-width
			 source-height
			 (IF delay-updates 1 0)
			 (IF restoration-p 1 0))
	 (add::transmit-packet acb channel)))
      
      ((AND (ARRAYP from-array)	(NUMBERP to-array))	   ;LISPM to Mac case
       (WHEN replication
         (IF (AND (<= source-width 32) (< (* source-width 4) width))
	 (SETF source-width (* source-width 4)))
         (IF (AND (<= source-height 32) (< (* source-height 4) height))
	 (SETF source-height (* source-height 4))))
       (SETF word-width
	     (TRUNCATE 32. (ARRAY-ELEMENT-SIZE from-array))) 
       (quick-change-of-two-dimension-indirect-arrays-dimensions 
	 *bitblt-intermediate-array*
	 source-height (* word-width (CEILING source-width word-width)))
       
       ;(BREAK " :CopyBits Exp-to-Mac")
       (LET ((acb (add:get-acb-fast (+ (* (CEILING source-width 32) 4 source-height)
				       (* 2 (force-even CopyBits-parms))))))
	 
	 (si:change-physical-array *bitblt-intermediate-array*
				   (+ (add:address acb)
				      add:*acb-overhead-bytes*
				      (* 2 (force-even CopyBits-parms))))
	 (SETF (add:requestor-complete acb) t)
	 (SETF (add:opcode acb) #x0)
	 (si-BITBLT			   ;Move image to intermediate array
	      w:alu-seta
	      source-width source-height
	      from-array from-x from-y
	      *bitblt-intermediate-array* 0 0)
	 
	 (add:load-parms acb 16
	   (ARRAY-ELEMENT-SIZE from-array)	       ;type of the array   
	   (transfer-mode alu)			       ;Mac sets srcMode
	   width height
	   -1					       ;negative tells Mac blt is comming from LISPM
	   to-array				       ;zero here means Mac screen else deexposed graphport
	   0 0					       ;from offset is meaningless in this case
	   to-x to-y
	   source-width
	   source-height
	   (IF delay-updates 1 0)
	   (IF restoration-p 1 0))
	 (add::transmit-packet acb channel)))

      ((AND (NUMBERP from-array) (ARRAYP to-array))	   ;Mac to LISPM
       (SETF word-width
	     (TRUNCATE 32. (ARRAY-ELEMENT-SIZE to-array)))
       (quick-change-of-two-dimension-indirect-arrays-dimensions 
	 *bitblt-intermediate-array*
	 source-height (* word-width (CEILING source-width word-width)))

       (LET ((acb (add:get-acb-fast (+ (* (CEILING source-width 32) 4 source-height)
				       (* 2 (force-even CopyBits-parms))))))
	 (si:change-physical-array *bitblt-intermediate-array*
				   (+ (add:address acb)
				      add:*acb-overhead-bytes*
				      (* 2 (force-even CopyBits-parms))))
	 (SETF (add:requestor-complete acb) nil)
	 (SETF (add:opcode acb) #x0)

	 (add:load-parms acb 16
	   1
	   (transfer-mode w:alu-seta)	   ;Mac sets srcMode
	   source-width source-height
	   from-array			   ;zero here means Mac screen else deexposed graphport
	   -1				   ;negative tells Mac blt should go to LISPM
	   (IF replication 0 from-x) (IF replication 0 from-y)
	   0 0				   ;to offset is meaningless in this case
	   source-width
	   source-height
	   (IF delay-updates 1 0)
	   (IF restoration-p 1 0))
	 (add::transmit-packet-and-wait acb channel)
					   ;Bit image from Mac should be ready now.
	 (si-BITBLT alu			   ;Move it from the intermediate array to LISPM virtual memory
		    width height 
		    *bitblt-intermediate-array*
		    (IF replication from-x 0) (IF replication from-y 0)
		    to-array to-x to-y)
	 (SETF (add:requestor-complete acb) t)
	 (add:return-acb-fast acb)
	 ))
      
      (t (si-BITBLT alu			   ;LISPM to LISPM case - Mac not involved
		    width height
		    from-array from-x from-y
		    to-array to-x to-y)))))


;;; ----------------------------------------------------------------------------------------------------------------------
;;;
;;;	ICAD's fast rubberbanded mouse cursor support...
;;;

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (MAC-FLAVOR :OPEN-RUBBER-BAND) (anchor-x anchor-y &optional (direction :from) (pattern :solid) (window tv:selected-window))  
  (LET ((acb (add:get-acb-fast 36))
	clip-x1 clip-x2 clip-y1 clip-y2)

    (SETF (VALUES clip-x1 clip-y1) (tv:sheet-calculate-offsets window nil))
    (SETF clip-x2 (+ clip-x1 (tv:sheet-inside-right window)))
    (SETF clip-y2 (+ clip-y1 (tv:sheet-inside-bottom window)))

    (SETF (add:opcode acb) 23)
    (SETF (add:requestor-complete acb) t)
    (add:load-parms acb 16
		    anchor-x
		    anchor-y
		    *fast-rb-q-limit*
		    (+ (* 256 (ECASE pattern (:solid 0) (:dashed 1)))
		       (ECASE direction (:none 0) (:from 1) (:to 2)))
		    clip-x1
		    clip-y1
		    clip-x2
		    clip-y2)
    (add::transmit-packet acb mac:channel)))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac:MAC-FLAVOR :CLOSE-RUBBER-BAND) (&optional leave-line-drawn)
  "Turns off rubber-banding. If leave-line-drawn is non nil the final line is left drawn and its location returned. 
   Otherwise the return value has no meaning."
  (LET ((acb (add:get-acb-fast 20)))
    (SETF (add:opcode acb) 24)

    (COND (leave-line-drawn
	   (add:load-parms acb 16 1)
	   (SETF (add:requestor-complete acb) 1)
	   (add::transmit-packet-and-wait acb mac:channel))
	  (t
	   (add:load-parms acb 16 0)
	   (SETF (add:requestor-complete acb) 0)
	   (add::transmit-packet acb mac:channel)))
	   

    (VALUES (add:parm-16b acb 1) (add:parm-16b acb 2) (add:parm-16b acb 3) (add:parm-16b acb 4))

    ))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"



(DEFMETHOD (mac:mac-flavor :draw-rubber-band) (x1 y1 x2 y2 &optional  (direction :from) (pattern :solid))
  (LET ((acb (add:get-acb-fast 24)))
    (SETF (add:opcode acb) 25)
    (add:load-parms acb 16
		    x1 y1 x2 y2
		    (ECASE direction (:none 0) (:from 1) (:to 2))
		    (ECASE pattern (:solid 0) (:dashed 1)))
    (add::transmit-packet acb mac:channel)))



;;; ----------------------------------------------------------------------------------------------------------------------
;;;
;;;	Fast dashed-line support...
;;;
;;; ----------------------------------------------------------------------------------------------------------------------
;;;  Although all the argument of the :draw-dashed-line methods appear to be supported, in reality the Mac-side code
;;;  always draws a dashed line using a pattern of
;;;
;;;	00001111                        	00000000
;;;	00001111                        	00000000
;;;	00001111                        	00000000
;;;	00001111            or    	00000000
;;;	00001111            	11111111
;;;	00001111              	11111111
;;;	00001111              	11111111
;;;	00001111              	11111111
;;;
;;;  depending on whether the line is more horizontal or vertical.  This results in approximately 4-pixel dashes spaced at 8
;;;  pixels.  This is MUCH faster than drawing lots of short lines and looks pretty good.

))

#!C
; From file BASIC-COMM-PROTOCOLS.LISP#> WINDOW-MX; SYS:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-PROTOCOLS.#"


(DEFMETHOD (mac-flavor :are-you-there?)
	   (&optional (timeout nil))
  "Sends Nop command to MAC and waits for MAC to acknowledge the
command, set input complete, and set command complete.  Returns
3 values: ACKNOWLEDGED-COMMAND INPUT-COMPLETE and COMMAND-COMPLETE
status.  Optional TIMEOUT is in seconds."
;;  (add:wait-port-ready Nop-Command channel)
  (LET* ((inhibit-scheduling-flag t)
	 (add:*no-interrupt* inhibit-scheduling-flag)
	 (acb (add:get-acb 2 t)))
    (setf (add:opcode acb) #xff)
    (add:transmit-packet acb channel)
    (LET (input-complete command-complete)
      (SETQ input-complete
	    (OR (LOOP repeat 1000
		      when (add:input-complete acb)
		      do (RETURN t))
		(PROCESS-WAIT-WITH-TIMEOUT
		  "Wait Input Complete"
		  timeout
		  #'(lambda () (add:input-complete acb)))))
    (WHEN input-complete
      (SETQ command-complete
	    (OR (LOOP repeat 1000
		      when (add:command-complete acb)
		      do (RETURN t))
		(PROCESS-WAIT-WITH-TIMEOUT
		  "Wait Command Complete"
		  timeout
		  #'(lambda () (add:command-complete acb))))))
    
    (setf (add:requestor-complete acb) t)
    (add:return-acb acb)
    (SEND (tv-channel) :reset t)
    (VALUES t input-complete command-complete))))

))
