;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Correct remember-call's handling of a list.  Perform all screen resizing under
;;; without-interrupts protection.  Correct the calculation of the number of Mac-
;;; resident bit arrays in :GetModifiedRectangles.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 149149, M/S 2151             
;;;   AUSTIN, TEXAS 78714-9149                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 09/14/89 10:29:46 by GRENINGER,
;;; while running on MX23 from band P817
;;; With SYSTEM 6.14, GC 6.3, VIRTUAL-MEMORY 6.1, MICRONET 6.0, MICRONET-COMM 6.1,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-PATHNAME 6.1, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.0,
;;;  BASIC-NAMESPACE 6.2, BASIC-FILE 6.2, RPC 6.1, NFS-MX 6.1, EH 6.4, MAKE-SYSTEM 6.0,
;;;  MEMORY-AUX 6.0, COMPILER 6.10, TV 6.15, NVRAM 6.1, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.7, METER 6.1, ZWEI 6.5, DEBUG-TOOLS 6.3, WINDOW-MX 6.7, PRINTER 6.3,
;;;  MAC-PRINTER-TYPES 6.1, CLIPBOARD 6.1, TI-CLOS 6.20, CLEH 6.5, NETWORK-PATHNAME 6.0,
;;;  NETWORK-NAMESPACE 6.0, DATALINK 6.0, CHAOSNET 6.0, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.0, IP 3.47, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.1, IMAGEN 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2,
;;;  TELNET 6.0, VT100 6.0, STREAMER-TAPE 6.4, DECNET 1.69, VISIDOC 6.2, PROFILE 6.1,
;;;  Experimental CLX 6.2, CLUE 6.10, Experimental BUG 11.14,  microcode 138, Band Name: microExplorer Network + SLE (8/17)

#!C
; From file DEFINITIONS.LISP#> WINDOW-MX; Sys:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; DEFINITIONS.#"


(DEFMACRO remember-call (&rest args)
  "Maintains the list *calls* as a circular history of the stack and the values of
ARGS as of its calls.  ARGS must all be symbols; their names and values will be
remembered.  If the first of ARGS is an integer, then it is used as a limit on the stack
depth remembered, otherwise all levels are remembered."
  (DECLARE (SPECIAL *default-stack-depth-to-record* *all-debugging-classes*))
  (LET ((to-a-depth-of *default-stack-depth-to-record*) (debugging-class t)
	(default-cons-area sys:background-cons-area) arg-names)
    ;;  If 1st arg is a number use it as the depth to which the stack is to be reported...
    (WHEN (AND (> (LENGTH args) 0) (INTEGERP (FIRST args)))
      (SETF to-a-depth-of (FIRST args))
      (SETF args (REST args)))
    ;;  If 1st arg is a keyword then it specifies this call's debugging class.  Add it to the list of legal
    ;;  debugging classes.
    (WHEN (AND (> (LENGTH args) 0)
	       (SYMBOLP (FIRST args))
	       (KEYWORDP (FIRST args)))
      (SETF debugging-class (FIRST args))
      (SETF args (REST args))
      (UNLESS (MEMBER debugging-class *all-debugging-classes*)
	(PUSH debugging-class *all-debugging-classes*)))
    (IF args
	(PROGN
	  (SETF arg-names (LOOP for arg in args
				collecting (FORMAT nil "~s=" arg)
				collecting arg))
	  `(WITHOUT-INTERRUPTS
	     (UNLESS (OR *tracing-off* (suppress-tracing-p ,debugging-class))
	       (LET ((%v% (LIST ,@arg-names)))
		 (LOOP FOR arg-tail first (CDR %v%)
		       then (CDDR arg-tail)
		       while arg-tail
		       do (SETF (FIRST arg-tail)
				(TYPECASE (FIRST arg-tail)
				  (symbol (SYMBOL-VALUE (FIRST arg-tail)))
				  (CONS (IF (FUNCTIONP (FIRST (FIRST arg-tail)))
					    (EVAL (FIRST arg-tail))
					  (FIRST arg-tail)))
				  (t (FIRST arg-tail))))
		       (when (consp (first arg-tail))
			 (setf (first arg-tail) (copy-tree (first arg-tail)))))
		 (syz2 %v% ,to-a-depth-of)))))
      ;; else...
      `(WITHOUT-INTERRUPTS
	 (UNLESS (OR *tracing-off* (suppress-tracing-p ,debugging-class))
	   (syz2 nil ,to-a-depth-of))))))

))

#!C
; From file MAC-WINDOWS.LISP#> WINDOW-MX; Sys:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; MAC-WINDOWS.#"


(DEFUN emulate-cache-allocation (window-id new-width new-height)
  "Returns T if able to allocate, else returns nil."
  (LET ((size-needed (* (CEILING new-width 16) 2 new-height))
	(old-free-space *nbytes-free-space*))
    (remember-call :rects *emulated-cache*)
    (LOOP with end-of-previous-entry = 0 and success-p
	  and old-entry-entries and hole-entries
	  for last-entries first *emulated-cache* then entries
	  for entries = (REST last-entries)
	  for entry = (FIRST entries)
	  until (NULL entry)
	  do
	  
	  (WHEN (= (cache-entry-wid entry) window-id)
	    (IF (= (cache-entry-nbytes entry) size-needed)
		;;  If ENTRY is window-id's existing entry and is exactly the size
		;;  needed, do nothing, just return T...
		(RETURN t)
	      
	      ;;  ...else ENTRY is not the right size, so remember it for later.
	      ;;  We'll delete it if we can successfully allocate a new entry...
	      (SETF old-entry-entries last-entries)))
	  
	  ;;  Check the hole in front of ENTRY.  If big enough, and we
	  ;;  haven't allocated space for WINDOW-ID yet, remember it for doing
	  ;;  so...
	  (WHEN (AND (NULL hole-entries)
		     (>= (- (cache-entry-start-adr entry) end-of-previous-entry)
			 size-needed))
	    (SETF hole-entries last-entries))
	  
	  ;;  Always remember where the start of the next hole is...
	  (SETF end-of-previous-entry (+ (cache-entry-start-adr entry)
					 (cache-entry-nbytes entry)))
	  
	  finally
	  
	  (remember-call :rects success-p old-free-space *nbytes-free-space* size-needed)
	  (WHEN hole-entries
	    ;;  Now allocate an entry of size-needed
	    (SETF entry (FIRST hole-entries))
	    (RPLACD hole-entries
		    (CONS (make-cache-entry
			    :start-adr (+ (cache-entry-start-adr entry)
					  (cache-entry-nbytes entry))
			    :nbytes size-needed
			    :wid window-id)
			  (REST hole-entries)))
	    (DECF *nbytes-free-space* size-needed)

	    ;;  If successful in allocating a new entry and found an old entry for
	    ;;  WINDOW-ID, delete it...
	    (WHEN old-entry-entries
	      ;;  In case we just added an entry in front of the one we planned to
	      ;;  delete, make sure we delete the right entry...
	      (WHEN (EQ old-entry-entries hole-entries)
		(SETF old-entry-entries (REST old-entry-entries)))
	      (INCF *nbytes-free-space*
		    (cache-entry-nbytes (SECOND old-entry-entries)))
	      (RPLACD old-entry-entries (CDDR old-entry-entries)))
	    (RETURN t))
	  
	  ;;  We couldn't allocate an entry, yet total free space is sufficient,
	  ;;  so compact the cache and try again...
	  (WHEN (>= *nbytes-free-space* size-needed)
	    (emulate-cache-compaction)
	    (SETF success-p (emulate-cache-allocation window-id new-width new-height)))
	  
	  (RETURN success-p))))

))

#!C
; From file MAC-INITIATED-COMMANDS.LISP#> WINDOW-MX; Sys:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; MAC-INITIATED-COMMANDS.#"


(DEFUN resize-screen (screen-id new-screen-height new-screen-width
		    &optional debug-p old-screen-height old-screen-width)
  (LET* (screen inferiors blinkability-states true-mouse-cursor
	 old-right old-bottom left top right bottom windows-to-resize
	 h-expansion-factor w-expansion-factor error (errors 0))

    (WHEN (SETF screen (the-screen (AREF *mac-resident-explorer-screens* screen-id)))
      (SETF inferiors (SEND screen :inferiors))
      (UNLESS old-screen-height
	(SETF old-screen-height (SEND screen :height)))
      (UNLESS old-screen-width
	(SETF old-screen-width (SEND screen :width)))
      (SETF new-screen-width (* 32 (CEILING new-screen-width 32)))
      
      (SETF h-expansion-factor (/ new-screen-height old-screen-height)
	    w-expansion-factor (/ new-screen-width old-screen-width))

      ;;  Ask the Mac whether the new screen size will fit.  Don't try if it won't...
      (IF (NOT (can-this-screen-be-resized
		 screen-id new-screen-width new-screen-height))
	  (pop-up-format-at-origin "~
Insufficient space in Mac's screen/window image
cache to accomodate the requested resizing.")
	;; else...
              ;;  Change the Mac's mouse cursor to the watch while we thrash bitmaps...
	(SETF true-mouse-cursor tv:mouse-blinker)
	(SEND *mac* :set-mouse-blinker (- 4 mac-mouse-cursor-offset))

	(tv:without-screen-management
	  (SETF blinkability-states
		(collect-and-turn-off-sheets-blinker-states screen))
	  (resize-a-Mac-resident-Explorer-screen
	    screen-id screen new-screen-height new-screen-width nil)
	  (LOOP for window in inferiors
		do
		(MULTIPLE-VALUE-SETQ (left top right bottom)
		  (SEND window :edges))
		(SETF old-right right old-bottom bottom)
		
		;;  If the screen is growing, expand the window proportionally...
		(WHEN (> w-expansion-factor 1)
		  (SETF right
			(+ left (ROUND (* w-expansion-factor (- right left))))))
		(WHEN (> h-expansion-factor 1)
		  (SETF bottom
			(+ top (ROUND (* h-expansion-factor (- bottom top))))))
		
		;;  Shrink the window to fit if it won't after the screen changes size...
		(WHEN (> right new-screen-width)
		  (SETF right new-screen-width))
		(WHEN (> bottom new-screen-height)
		  (SETF bottom new-screen-height))
		
		;; If the window needs to be resized and can be resized to this new
		;; size, then add it to the list of windows to be resized.  If it needs
		;; to be resized but cannot be, stop trying and tell the Mac we failed...
		(WHEN (OR (/= right old-right) (/= bottom old-bottom))
		  (IF (MULTIPLE-VALUE-SETQ (nil error)
			(SEND window :set-edges left top right bottom :verify))
		      (PUSH (LIST window left top right bottom) windows-to-resize)
		      ;; else...
		      ;; just tell user, truck on, if unresized window fits on resized screen...
		      (IF (AND (> new-screen-width right) (> new-screen-height bottom))
			  (pop-up-format-at-origin "~
Window ~S
could not resize because ~A
but still can fit on the resized ~dx~d screen."
						   window error
						   new-screen-height new-screen-width)
			  ;; else...
			  ;; count as a serious resize error, window won't fit on resized screen..
			  (INCF errors)
			  (pop-up-format-at-origin "~
Window ~S
prevented resizing this screen from ~dx~d to ~dx~d
because ~A."
						   window
						   old-screen-height old-screen-width
						   new-screen-height new-screen-width
						   error)))))
	  
	  ;;  Really change the windows-to-resize of the windows that needed it...
	  (IF (NOT (OR (ZEROP errors) debug-p))
	      (PROGN
		(resize-a-Mac-resident-Explorer-screen
		  screen-id screen old-screen-height old-screen-width nil)
		(SEND screen :refresh))
	    ;; else...
	    (resize-a-Mac-resident-Explorer-screen
	      screen-id screen new-screen-height new-screen-width t)
	    (LOOP for (window left top right bottom) in
		  windows-to-resize
		  do
		  (WITHOUT-INTERRUPTS
		    (let ((add:*no-interrupt* t))
		      (SEND window :set-edges left top right bottom))))
	    ;;  Make sure all  windows get rebuilt correctly...
	    (DOLIST (i (w:sheet-inferiors screen))
	      (PROCESS-RUN-FUNCTION "Refresh"
				    #'(lambda (sheet)
					(WITHOUT-INTERRUPTS
					  (let ((add:*no-interrupt* t))
					    (AND (w:sheet-can-get-lock sheet)
						 (SEND sheet :refresh))))) i))
	    ))
	(reestablish-sheets-blinker-states blinkability-states)
	;;  Having done everything with screen management disabled, fix things up...
	(SEND screen :screen-manage)
	(w:who-line-clobbered)

	;;  Restore the Mac's mouse cursor to what the Explorer thinks it is...
	(SEND true-mouse-cursor :create-Mac-image-of-Explorer-mouse-cursor nil nil)))))

))

#!C
; From file BASIC-COMM-METHODS.LISP#> WINDOW-MX; Sys:
#10R MACINTOSH#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "MACINTOSH"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: WINDOW-MX; BASIC-COMM-METHODS.#"


(DEFMETHOD (mac-flavor :GetModifiedRectangles)
	   (&aux (add:*no-interrupt* inhibit-scheduling-flag))
  (LET (acb n-windows-with-mac-resident-bitmaps acb-image)
    (setf  n-windows-with-mac-resident-bitmaps (- (length *emulated-cache*) 2))
    (SETF acb (add:get-acb (+ 4 (* 16 n-windows-with-mac-resident-bitmaps) 128)))
    (SETF (add:opcode acb) #x08)
    (add:transmit-packet-and-wait acb channel)
    (SETF acb-image (MAKE-ARRAY (ARRAY-TOTAL-SIZE acb) :element-type '(unsigned-byte 16)))
    (COPY-ARRAY-CONTENTS acb acb-image)
    (add:return-acb acb t)
    acb-image))

))
