;;; -*- Mode: Common-Lisp; Package: X11; Base: 10.; Patch-File: T -*-

;;; Reason: Modify SERVER-TOP-LEVEL to properly handle connection shutdown. The function was not waiting until the connection completed emptying its queue. This could cause the client to hang in a CLOSE state.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 09/28/89 07:22:29 by BERGER,
;;; while running on ARIES from band LODX
;;; With SYSTEM 6.15, VIRTUAL-MEMORY 6.1, EH 6.4, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.2, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.10, TV 6.15, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.5,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.2, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.4, TI-CLOS 6.20, CLEH 6.5, IP 3.49,
;;;  Experimental CLX 6.2, CLUE 6.10, X11M 6.13, Experimental BUG 11.15, VISIDOC-SERVER 6.1,
;;;   microcode 429, Band Name: Rel 6.0 + SLE 8/30

#!C
; From file SERVER.LISP#> X11M.SERVER; SYS:
#10R X11#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "X11"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: X11M.SERVER; SERVER.#"


(DEFUN SERVER-TOP-LEVEL ()
  (LET (CONNECTION)
    (UNWIND-PROTECT
        (PROGN
          (SETQ CONNECTION
                (NET:LISTEN-FOR-CONNECTION-ON-MEDIUM
		  :BYTE-STREAM
		  "X11"
		  :STREAM-TYPE :CHARACTER-STREAM
		  :DIRECTION :BIDIRECTIONAL
		  :TIMEOUT (* 60 30) ;; 30 seconds
		  :TIMEOUT-AFTER-OPEN NIL
		  :INPUT-BUFFER-SIZE *DEFAULT-BUFFER-SIZE*))
          (WHEN CONNECTION
            ;; Connection didn't time out.
            (ESTABLISH-CONNECTION CONNECTION)))
      ;; Indicate transmission complete and await close completion.
      (WHEN CONNECTION  ;(AND CONNECTION (SERVER-UP-P CONNECTION)) ; DAB 09-28-89
        (SI:SEND CONNECTION :CLOSE)
	;;DAB 09-28-89 The close will send the proper FINs and ACKs but the we must 
        ;;give it time to empty its queue. Otherwise the this process will shutdown and
        ;;the FIN packets will not go out. This will cause the client to hang in TCP-CLOSE.
	(process-wait-with-timeout "Shutting Down X Server Connection"
				   (* 60 30)
				   #'(lambda (stream) (eq :CLOSED
							  (send stream :status)))
				   connection))
      (SETQ CONNECTION NIL))
    ;; When thrown out, abort the connection.
    (WHEN (AND CONNECTION (SERVER-UP-P CONNECTION))
      (SI:SEND CONNECTION :ABORT))
    (SERVER-TRACE "~%Connection shut down.~%"))
  (WHEN (NULL (GLOBAL-STATE.STATES *GLOBALS*))
    (SERVER-RESET)))
))
