;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Fixes problem where lines wider than one pixel were not drawn
;;; correctly. For example, a wide horizontal line had the top few
;;; pixels shifted slightly to the left. SPR 9977.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 12/01/89 15:19:29 by HAGY,
;;; while running on Zwingli from band LOD1
;;; With SYSTEM 6.25, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.6, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.19, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.8,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.3, MAIL-READER 6.6, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.5, TI-CLOS 6.26, CLEH 6.5, IP 3.56,
;;;  Experimental CLX 6.7, CLUE 6.34, X11M 6.17, Experimental BUG 11.17,  microcode 430,
;;;  Band Name: REL 6.0 + SLE 11/28

#!C
; From file POLY-LINE.LISP#> X11M.SERVER; SYS:
#10R X11#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "X11"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: X11M.SERVER; POLY-LINE.#"


(DEFUN CALCULATE-BOUNDING-BOX (X0 Y0 X1 Y1 HALF-WIDTH)
  "Return the corners of the polygon formed for a line with endpoints xy/y0, xy/y1 and half-width"
  ;; If width is >1, then the bounding box, defined
  ;; as (A0,B0) (A1,B1) (A2,B2) & (A3,B3) is returned.
  ;;
  ;;                     (X1,Y1)
  ;;            A1,B1 +-----X-----+ A2,B2
  ;;                  |    /|\    |
  ;;                  |  /  |  \  |
  ;;                  |     |     |
  ;;                  |     |     |
  ;;                  |     |     |
  ;;                  |     |     |
  ;;                  |     |     |
  ;;                  |     |     |
  ;;	              |     |     |
  ;;     	      |     |     |
  ;;	        A0,B0 +-----X-----+ A3,B3
  ;;                     (X0,Y0)
  (let* ((diffx (- x0 x1))
	 (diffy (- y0 y1))
	 (hypot (short-float (sqrt (+ (expt diffx 2) (expt diffy 2))))))
    (if (zerop hypot) ;; zero length line
	(values
	  (floor (+ x0 half-width)) y1
	  (floor (+ x0 half-width)) y1
	  (floor (- x0 half-width)) y1
	  (floor (- x0 half-width)) y1)
      (let ((dx (quotient (* diffy half-width) hypot))
	    (dy (quotient (* diffx half-width) hypot)))
	(values (round (- x0 dx)) (round (+ y0 dy))
		(round (- x1 dx)) (round (+ y1 dy))
		(round (+ x1 dx)) (round (- y1 dy))
		(round (+ x0 dx)) (round (- y0 dy)))))))
))
