;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: Fixes 'vanishing icon' problem associated with
;;; R4 TWM. SPRs 11050 & 11079. Event masks were not being
;;; handled properly when client was not the window owner.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 04/19/90 17:31:42 by HAGY,
;;; while running on Zwingli from band LOD1
;;; With SYSTEM 6.25, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.7, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.6, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.19, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.5, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.8,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.3, MAIL-READER 6.6, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.5, TI-CLOS 6.26, CLEH 6.5, IP 3.56,
;;;  Experimental CLX 6.7, CLUE 6.34, X11M 6.21, Experimental BUG 11.17,  microcode 430,
;;;  Band Name: REL 6.0 + SLE 11/28

#!C
; From file EVENTS.LISP#> X11M.SERVER; Hotel:
#10R X11#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "X11"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: X11M.SERVER; EVENTS.#"


(DEFUN EVENT-SELECT-FOR-WINDOW (WINDOW CLIENT MASK)
  (DECLARE (TYPE WINDOW WINDOW)
           (TYPE STATE CLIENT)
           (TYPE INTEGER MASK)
           (VALUES INTEGER))
  (EVENT-TRACE-ENTERING "EVENT-SELECT-FOR-WINDOW")
  (EVENT-TRACE ", WINDOW=x~16R, mask=x~16R, Same client=~A"
               (WINDOW.ID WINDOW) MASK (EQ (WINDOW.CLIENT WINDOW) CLIENT))
  (PROG1
    (block get-out
      (LET ((CHECK (LOGAND MASK AT-MOST-ONE-CLIENT)))
	(block mask-set
          (WHEN (NOT (ZEROP (LOGAND CHECK (WINDOW.ALL-EVENT-MASKS WINDOW))))
            ;; It is illegal for two different clients to select on any of the events
            ;; for AT-MOST-ONE-CLIENT.  However, it is OK, for some client to continue
            ;; selecting on one of those events.
            (LOOP FOR OTHER IN (WINDOW.EVENT-MASKS WINDOW)
                  DO (WHEN (AND (NEQ (OTHER-CLIENT.CLIENT OTHER) CLIENT)
                                (NOT (ZEROP (LOGAND CHECK (OTHER-CLIENT.MASK OTHER)))))
                       (RETURN-FROM GET-OUT BAD-ACCESS))))
          (IF (EQ (WINDOW.CLIENT WINDOW) CLIENT)
              (PROGN
                (SETQ CHECK (WINDOW-EVENT-MASKS WINDOW))
                (SETF (WINDOW-EVENT-MASKS WINDOW) MASK))
              ;;ELSE
              (PROGN
                (LOOP FOR OTHER IN (WINDOW.EVENT-MASKS WINDOW)
                      WHEN (EQ (OTHER-CLIENT.CLIENT OTHER) CLIENT)
                      DO (PROGN
                           (SETQ CHECK (OTHER-CLIENT.MASK OTHER))
                           (WHEN (ZEROP MASK)
                             (return-from GET-OUT STATUS-SUCCESS))
                           (SETF (OTHER-CLIENT.MASK OTHER) MASK)
                           (return-from mask-set nil)))
                (LET ((*state* client))
		  (setf (window-event-masks window)
			(logior (window-event-masks window) mask)))
		(SETQ CHECK 0))))
	;; Mask-set
	(WHEN (AND (EQ MOTION-HINT-WINDOW WINDOW)
		   (NOT (ZEROP (LOGAND MASK POINTER-MOTION-HINT-MASK)))
		   (ZEROP (LOGAND CHECK POINTER-MOTION-HINT-MASK))
		   (NULL (DEVICE.GRAB (INPUT-INFO.POINTER INPUT-INFO))))
	  (SETQ MOTION-HINT-WINDOW NIL))
	(RECALCULATE-DELIVERABLE-EVENTS WINDOW)))
    (EVENT-TRACE-LEAVING "EVENT-SELECT-FOR-WINDOW")))
))

#!C
; From file EVENTS.LISP#> X11M.SERVER; Hotel:
#10R X11#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "X11"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: X11M.SERVER; EVENTS.#"


(DEFUN EVENT-MASK-FOR-CLIENT (WINDOW CLIENT)
  (DECLARE (TYPE WINDOW  WINDOW)
           (TYPE STATE   CLIENT)
           (VALUES INTEGER INTEGER))
  (EVENT-TRACE-ENTERING "EVENT-MASK-FOR-CLIENT")
  (LET ((HER 0)
        (ALL-MASK 0))
    (LOOP FOR OTHER IN (WINDOW.EVENT-MASKS WINDOW)
          DO (WHEN (EQ (OTHER-CLIENT.CLIENT OTHER) CLIENT)
               (SETQ HER (OTHER-CLIENT.MASK OTHER)))
          (SETQ ALL-MASK (LOGIOR ALL-MASK (OTHER-CLIENT.MASK OTHER))))
    (EVENT-TRACE-LEAVING "EVENT-MASK-FOR-CLIENT")
    (VALUES HER ALL-MASK)))
))
