;;; -*- Mode: Common-Lisp; Package: User; Base: 10.; Patch-File: T -*-

;;; Reason: When a client establishes a connection, ensure the X server window
;;; is instantiated and initialized.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for X11M version 6.9
;;; Written 07/07/89 19:33:45 by buehring,
;;; while running on Spud from band LOD4
;;; With SYSTEM 6.10, VIRTUAL-MEMORY 6.1, EH 6.3, MAKE-SYSTEM 6.0, MICRONET 6.0, LOCAL-FILE 6.0,
;;;  BASIC-PATHNAME 6.1, NETWORK-SUPPORT-COLD 6.0, BASIC-NAMESPACE 6.1, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.0, DISK-LABEL 6.0, BASIC-FILE 6.2, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.7, TV 6.12, DATALINK 6.0, CHAOSNET 6.0, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.1,
;;;  SYSLOG 6.1, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.3,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.1, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.1,
;;;  IMAGEN 6.0, SUGGESTIONS 6.0, MAIL-DAEMON 6.2, MAIL-READER 6.1, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.0, PROFILE 6.1, VISIDOC 6.2, TI-CLOS 6.11, CLEH 6.4, IP 3.47,
;;;  Experimental BUG 11.10, Experimental CLX 6.2, CLUE 6.5, X11M 6.7,  microcode 429,
;;;  Band Name: Release 6.0 + SLE  6/26

#!C
; From file SERVER.LISP#> X11M.SERVER; MR-X:
#10R X11#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "X11"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: X11M.SERVER; SERVER.#"


(defun establish-connection (stream)
  (unless (find-free-x-server-window)
    (initialize-monochrome-server))
  (let* (
         ;; Read the `bytes-swapped' mode from the client.
         (MODE (INT-CHAR (SERVER-STREAM-TYI STREAM)))
         STATE
         ;; The B test is for bytes being swapped.
         ;; This tells us which order the client wants bytes sent.
         (swapped (eql (char-code mode) (char-code #\B))))
    (let ((*resources* (global-state.resources *globals*))
	  (*delayed-tree-busies* 0)
	  (*trace-output* *monochrome-server-trace*) ;; Bind *TRACE-OUTPUT* for debug
	  (*package* (find-package 'x11)))     ;; bind *package* for easier debugging
      (unwind-protect
          (PROGN
            (setq state (cons-state stream swapped))
            ;; I'm not sure that doing the next SETF is correct, but I think so.  (TWE)
            ;; I think it is because the event code needs a hook into *resources*.
            (setf (state.resources state) *resources*)
            (global-state-add-server state)
            (process-client-connection state)
            (condition-case-if (not *debug-client-connections-p*) (condition)
                (server state)
              (sys:network-error
               (server-trace "~%IN ESTABLISH-CONNECTION, Got a network error:~%~a"
			     condition))
              (:NO-ERROR
               (SERVER-TRACE "~%IN ESTABLISH-CONNECTION, Exited SERVER normally.")))
            )
	(finish-tree-delays)
	(case (state.server-state state)
	  (:activist (deactivate-server state))
	  (:soloist  (deactivate-solo   state))
	  (:grabist  (deactivate-solo   state :grabist)))
	(state-shutdown state)))))

))
