;;; -*- Mode: Common-Lisp; Package: User; Base: 8.; Patch-File: T -*-

;;; Reason: Changed COMPLETE-STRING to prevent aref'inf string at invalid index.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Written 03/12/90 09:13:55 by MARKY,
;;; while running on LIBRA from band LODB
;;; With SYSTEM 6.30, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.6, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.6, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.24, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.12,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.3, MAIL-READER 6.6, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.5, TI-CLOS 6.40, CLEH 6.5, IP 3.56,
;;;  Experimental CLX 6.8, CLUE 6.66, X11M 6.20, Experimental BUG 11.18, MMON 6.2,
;;;  Experimental SC 3.0, Experimental CLIO 18.0, Experimental CLIO-TEST 7.0, Experimental CONFORMANCE-CHECKER 2.0,
;;;   microcode 429, Band Name: REL6-mmon-11/15/89

;;; SPR 11027 Swappped sequence of AND forms to prevent aref'ing string at invalid index.

#!C
; From file COMD.LISP#> ZMACS; Hotel:
#8R ZWEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "ZWEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: ZMACS; COMD.#"


(DEFUN COMPLETE-STRING (STRING ALIST DELIMS
                        &OPTIONAL DONT-NEED-LIST CHAR-POS TRUNC IGNORE-TRAILING-SPACE
                        &AUX NCHUNKS CHUNKS CHUNK-DELIMS FILLS
                        CHAMB TEMS RETS RCHUNKS TEM LEN COMPLETED-P CHAR-CHUNK
                        CHAR-OFFSET MAGIC-POS TAIL ONE-BEFORE-TAIL TEMSTRING)
  "Complete a given STRING from an ALIST of strings.
DELIMS is a list of delimiter characters that delimit chunks.
 Each chunk is matched against the chunks of strings in the ALIST.
DONT-NEED-LIST says we really don't want all the possibilities, just not NIL.
CHAR-POS is position in string to be relocated with new things inserted.
\(The CHAR-POS value is the position of the same in the completed string).
TRUNC says dont complete more than one chunk at end.
IGNORE-TRAILING-SPACE non-NIL says ignore a trailing space character if any.
Returns new STRING, matching subset of ALIST, COMPLETED-P if some completion was done,
new CHAR-POS, and MAGIC-POS location of first point of ambiguity.
COMPLETED-P is 'NOSPACE if proper delimiter is already at end of string.
For efficiency, if ALIST is an ART-Q-LIST array, it is assumed to be alphabetically
sorted."
  (SETQ CHUNKS (MAKE-ARRAY 24 :LEADER-LIST '(0))
	CHUNK-DELIMS (MAKE-ARRAY 24 :ELEMENT-TYPE 'T :LEADER-LIST '(0)))
  (SETQ LEN (LENGTH STRING))
  (AND IGNORE-TRAILING-SPACE
       (> LEN 0)
       (CHAR= (AREF STRING (1- LEN)) #\SPACE)
       (DECF LEN))
  (DO ((I 0 (1+ I))
       (J 0))
      ((> I LEN))
    (COND ((COND ((= I LEN)
		  (SETQ TEM -1))                ;Last character delimits a chunk unless it is empty.
		 (T
		  (MEMBER (SETQ TEM (AREF STRING I)) DELIMS :TEST #'EQ)))
	   (AND CHAR-POS
		(> CHAR-POS J)                  ;Keep track of relative position
		(SETQ CHAR-CHUNK (ARRAY-LEADER CHUNKS 0)
		      CHAR-OFFSET (- CHAR-POS J)))
           ;; Modified to disallow multiple blanks between chunks of extended commands,
           ;; since they would cause completion to fail. Done by rpm (from ap) on 9-29-86.
           (UNLESS (AND (STRING-EQUAL (NSUBSTRING STRING J I) " ")
			(NOT (= I LEN))
			(string-equal (aref string i) " ")	;; may 03/12/90 Moved AFTER check for i = index above
			)
             (VECTOR-PUSH-EXTEND (NSUBSTRING STRING J I) CHUNKS)
             (VECTOR-PUSH-EXTEND TEM CHUNK-DELIMS))
	   (SETQ J I))))
  (SETQ NCHUNKS (ARRAY-ACTIVE-LENGTH CHUNKS)
	FILLS (MAKE-ARRAY NCHUNKS)
	TEMS (MAKE-ARRAY NCHUNKS)
	RCHUNKS (MAKE-ARRAY NCHUNKS)
	CHAMB (MAKE-ARRAY NCHUNKS :ELEMENT-TYPE 'BIT))
  (AND (ARRAYP ALIST)
       (MULTIPLE-VALUE-SETQ (ALIST TAIL ONE-BEFORE-TAIL)
	 (COMPLETE-STRING-BOUNDS ALIST DELIMS NCHUNKS CHUNKS CHUNK-DELIMS)))
  (AND ONE-BEFORE-TAIL
       (N-CHUNKS-MATCH-P (CAAR ALIST) (CAAR ONE-BEFORE-TAIL) NCHUNKS DELIMS)
       ;; The first and last possibilities are the same, for as many chunks as we need,
       ;; so all in between must also be the same.
       DONT-NEED-LIST
       ;; So if we don't need all the possibilities,
       ;; keep just the first one and the last one.
       (SETQ ALIST (LIST (CAR ALIST) (CAR ONE-BEFORE-TAIL))
	     TAIL NIL))
  (DO ((L ALIST (CDR L))
       (ALL-AMBIG))
      ((EQ L TAIL))
    (COND ((ATOM L))				;Indirect through multiple alists
          ((NULL (COMPLETE-CHUNK-COMPARE (CAAR L) DELIMS NCHUNKS CHUNKS CHUNK-DELIMS TEMS
					 (AND (NULL RETS) RCHUNKS)))
	   (OR RETS (SETQ CHUNKS RCHUNKS))	;First winner determines case of result
	   (PUSH (CAR L) RETS)			;add to list of partial matches
	   (SETQ ALL-AMBIG DONT-NEED-LIST)
	   (DO ((I 0 (1+ I))
		(FILL))
	       ((>= I NCHUNKS))
	     (SETQ TEM (AREF TEMS I)
		   FILL (AREF FILLS I))
	     (COND ((NULL FILL)			;First one to complete a chunk
		    (SETF (AREF FILLS I) TEM)	;save for later ones
		    (AND (PLUSP (LENGTH TEM))
			 (SETQ ALL-AMBIG NIL)))	;This chunk not ambiguous yet
		   ((AND (EQUAL FILL "")
			 (ZEROP (AREF CHAMB I))
			 (NOT (EQUAL TEM "")))
		    ;; If there was an exact match found for this chunk,
		    ;; ignore everything that is NOT an exact match in this chunk.
		    (SETQ ALL-AMBIG NIL)
		    (RETURN NIL))
		   ((AND (EQUAL TEM "")
	                 (NOT (AND (EQUAL FILL "")
	                           (ZEROP (AREF CHAMB I)))))
		    ;; The first time we find an exact match for this chunk,
		    ;; from now on consider only exact matches for it,
		    ;; and forget anything we concluded about later chunks
		    ;; from completions that were inexact in this chunk.
		    (SETF (AREF FILLS I) "")
		    (SETF (AREF CHAMB I) 0)
		    (DO ((I (1+ I) (1+ I)))
			((= I NCHUNKS))
		      (SETF (AREF FILLS I) NIL)
		      (SETF (AREF CHAMB I) 0))
		    (SETQ ALL-AMBIG NIL))
		   (T
		    (SETQ LEN (LENGTH FILL))
		    (DO ((J 0 (1+ J))
			 (LEN1 (LENGTH TEM)))
			((>= J LEN)
			 (OR (ZEROP LEN)
			     (AND (= I (1- NCHUNKS))
				  (= LEN 1)
				  (MEMBER (AREF FILL 0) DELIMS :TEST #'EQ))
			     (SETQ ALL-AMBIG NIL)))
		      (COND ((OR (>= J LEN1)
				 (NOT (CHAR-EQUAL (AREF FILL J) (AREF TEM J))))
			     ;;Not the same completion, shorten final version
			     (SETF (AREF FILLS I) (NSUBSTRING FILL 0 J))
			     (SETF (AREF CHAMB I) 1)	;Remember this was ambiguous
			     (OR (ZEROP J)
				 (SETQ ALL-AMBIG NIL))
			     (RETURN NIL)))))))
	   ;;If not going to complete and dont need actual list, finish up now.
	   (AND ALL-AMBIG
		(NULL (AREF FILLS (1- NCHUNKS)))
		(RETURN NIL)))))
  (COND ((AND TRUNC (SETQ TEMSTRING (AREF FILLS (1- NCHUNKS))))
	 (SETQ LEN (LENGTH TEMSTRING))
	 (AND (ZEROP (AREF CHAMB (1- NCHUNKS)))	;If last chunk wasn't ambiguous,
	      (SETQ TRUNC 'NOSPACE))		;shouldn't have delimiter there
	 (DO ((I 0 (1+ I)))
	     ((>= I LEN)
	      NIL)
	   (COND ((MEMBER (AREF TEMSTRING I) DELIMS :TEST #'EQ)
		  (SETF (AREF FILLS (1- NCHUNKS)) (NSUBSTRING TEMSTRING 0 (1+ I)))
		  (SETQ TRUNC 'NOSPACE)		;Already gave a delimiter
		  (RETURN NIL))))))
  (SETQ TEMSTRING "")
  (DO ((I 0 (1+ I)))
      ((>= I NCHUNKS)
       NIL)
    (AND CHAR-POS
	 CHAR-CHUNK
	 (= I CHAR-CHUNK)                                         ;In case inside chunk not completed,
	 (SETQ CHAR-POS (+ (LENGTH TEMSTRING) CHAR-OFFSET)))      ;relocate
    (SETQ TEMSTRING (STRING-APPEND TEMSTRING (AREF CHUNKS I)))
    (COND ((AND (SETQ TEM (AREF FILLS I)) (> (LENGTH TEM) 0))
	   (SETQ TEMSTRING (STRING-APPEND TEMSTRING TEM) COMPLETED-P T)
	   (AND CHAR-POS
		CHAR-CHUNK
		(= I CHAR-CHUNK)	                ;If inside completed chunk,
		(SETQ CHAR-POS (LENGTH TEMSTRING)))))	;move to end of it
    (OR MAGIC-POS
	(ZEROP (AREF CHAMB I))             ;Remember end of leftmost ambiguous chunk
	(SETQ MAGIC-POS (LENGTH TEMSTRING))))
  (AND COMPLETED-P
       (EQ TRUNC 'NOSPACE)
       (SETQ COMPLETED-P 'NOSPACE))
  (WHEN (OR (AND (ARRAY-HAS-LEADER-P TEMSTRING)
		 (MINUSP (FILL-POINTER TEMSTRING)))
	    (AND CHAR-POS (MINUSP CHAR-POS))
	    (AND MAGIC-POS (MINUSP MAGIC-POS)))
    (FERROR "Internal error in completion.  Report a bug."))
  (VALUES TEMSTRING (NREVERSE RETS) COMPLETED-P CHAR-POS MAGIC-POS))
))
