;;; -*- Mode: Common-Lisp; Package: SRCCOM; Base: 8.; Patch-File: T -*-
;;; Written 04/26/90 10:25:01 by BERGER,
;;; Reason: Added error checking to  Source-Compare and create-file.
;;; while running on MX29 from band n817
;;; With SYSTEM 6.31, GC 6.3, VIRTUAL-MEMORY 6.3, MICRONET 6.0, MICRONET-COMM 6.3,
;;;  DISK-IO 6.3, DISK-LABEL 6.0, BASIC-PATHNAME 6.4, MAC-PATHNAME 6.0, NETWORK-SUPPORT-COLD 6.2,
;;;  BASIC-NAMESPACE 6.7, BASIC-FILE 6.10, RPC 6.2, NFS-MX 6.6, EH 6.6, MAKE-SYSTEM 6.2,
;;;  MEMORY-AUX 6.0, COMPILER 6.14, TV 6.24, NVRAM 6.2, UCL 6.0, INPUT-EDITOR 6.0,
;;;  MACTOOLBOX 2.17, METER 6.1, ZWEI 6.13, DEBUG-TOOLS 6.4, WINDOW-MX 6.10, PRINTER 6.4,
;;;  MAC-PRINTER-TYPES 6.2, CLIPBOARD 6.1, TI-CLOS 6.44, CLEH 6.5, NETWORK-PATHNAME 6.1,
;;;  NETWORK-NAMESPACE 6.1, DATALINK 6.0, CHAOSNET 6.6, NETWORK-SUPPORT 6.1, NETWORK-SERVICE 6.2,
;;;  DATALINK-DISPLAYS 6.0, MX-DATALINK 6.1, NAMESPACE-EDITOR 6.4, IP 3.60, NFS-MX-SERVER 6.0,
;;;  MX-SERIAL 6.1, PRINTER-TYPES 6.2, IMAGEN 6.1, MAIL-DAEMON 6.6, MAIL-READER 6.7,
;;;  TELNET 6.1, VT100 6.0, STREAMER-TAPE 6.5, DECNET 1.72, VISIDOC 6.7, PROFILE 6.3,
;;;   microcode 138, Band Name: microExplorer Network (8/17)

#!C
; From file SRCCOM.LISP#> ZMACS; SYS:
#8R SRCCOM#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "SRCCOM"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: ZMACS; SRCCOM.#"

(DEFUN CREATE-FILE (FILENAME)
  "Make a SRCCOM:FILE object for a file to be source compared.
FILENAME is opened and the SRCCOM:FILE object contains a stream for it."
  (let (stream generic-pathname)
    (condition-case (condition)  ; DAB 04-26-90 Added error checking.
	(progn
	 (setf STREAM (OPEN FILENAME :CHARACTERS T :DIRECTION :INPUT))
	 (setf GENERIC-PATHNAME (SEND FILENAME :GENERIC-PATHNAME))
	 (FS:READ-ATTRIBUTE-LIST GENERIC-PATHNAME STREAM)
	 (MULTIPLE-VALUE-BIND (VARS VALS)
	     (FS:FILE-ATTRIBUTE-BINDINGS GENERIC-PATHNAME)
	   (MAKE-FILE :FILE-STREAM STREAM
		      :FILE-NAME (SEND STREAM :TRUENAME)
		      :FILE-VARS VARS
		      :FILE-VALUES VALS
		      :FILE-MAJOR-MODE (OR (SEND GENERIC-PATHNAME :GET :MODE) :LISP))))
      (error (when stream (close stream)) condition))))

(DEFUN SOURCE-COMPARE (FILENAME-1 FILENAME-2
		       &OPTIONAL (OUTPUT-STREAM *STANDARD-OUTPUT*) (TYPE :TEXT)
		       &AUX FILE-1 FILE-2)
  "Source compare files FILENAME-1 and FILENAME-2, output to OUTPUT-STREAM.
See SOURCE-COMPARE-FILES for some additional information."
  (ASSERT (MEMBER TYPE '(:TEXT :FORM) :TEST #'EQ)
	  (TYPE)
	  "Only the values :TEXT or :FORM may be supplied for argument TYPE")
  (SETQ FILENAME-1 (FS:MERGE-AND-SET-PATHNAME-DEFAULTS FILENAME-1 *PATHNAME-DEFAULTS*)
	FILENAME-2 (FS:MERGE-PATHNAME-DEFAULTS FILENAME-2 FILENAME-1))
  (UNWIND-PROTECT
      (PROGN
	(SETQ FILE-1 (CREATE-FILE FILENAME-1)
	      FILE-2 (CREATE-FILE FILENAME-2))
	(if (and file-1 file-2 (not (errorp file-1)) (not (errorp file-2)))  ; DAB 04-26-90 Added error checking. Print the error.
	    (progn 
	      (DESCRIBE-SRCCOM-SOURCES FILE-1 FILE-2 OUTPUT-STREAM)
	      (SOURCE-COMPARE-FILES FILE-1 FILE-2 OUTPUT-STREAM TYPE))
	    (format output-stream "~%~%Source Compare for ~a and ~a aborted because ~a."
                    filename-1 filename-2
		    (cond ((errorp file-1) (send file-1 :report-string))
			  ((errorp file-2) (send file-2 :report-string))
			  (t "of unknown reasons")))
	    ()))
	    
    (AND FILE-1 (not (errorp file-1)) (SEND (FILE-STREAM FILE-1) :CLOSE))
    (AND FILE-2 (not (errorp file-2)) (SEND (FILE-STREAM FILE-2) :CLOSE))))
))
