;;; -*- Mode: Common-Lisp; Package: User; Base: 8.; Patch-File: T -*-

;;; Reason: Changes to meta-. to allow editing of keywords to be interpreted as
;;; flavor method or defsystem names, etc. Also presents a menu of possible
;;; defs when an ambiguity exits. Also changes to understand CLOS method specs.

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;   TEXAS INSTRUMENTS INCORPORATED      
;;;   P.O. BOX 2909, M/S 2151             
;;;   AUSTIN, TEXAS 78769                 
;;;
;;; Copyright (C) 1989 Texas Instruments Incorporated.
;;; All rights reserved.

;;; Patch file for ZWEI version 6.16
;;; Written 05/21/90 14:29:53 by MARKY,
;;; while running on LIBRA from band LODB
;;; With SYSTEM 6.23, VIRTUAL-MEMORY 6.2, EH 6.5, MAKE-SYSTEM 6.2, MICRONET 6.0, LOCAL-FILE 6.1,
;;;  BASIC-PATHNAME 6.2, NETWORK-SUPPORT-COLD 6.2, BASIC-NAMESPACE 6.6, NETWORK-NAMESPACE 6.0,
;;;  DISK-IO 6.1, DISK-LABEL 6.0, BASIC-FILE 6.6, MAC-PATHNAME 6.0, NETWORK-PATHNAME 6.0,
;;;  COMPILER 6.14, TV 6.19, DATALINK 6.0, CHAOSNET 6.5, GC 6.3, MEMORY-AUX 6.0, NVRAM 6.2,
;;;  SYSLOG 6.2, STREAMER-TAPE 6.4, UCL 6.0, INPUT-EDITOR 6.0, METER 6.1, ZWEI 6.14,
;;;  DEBUG-TOOLS 6.3, NETWORK-SUPPORT 6.0, NETWORK-SERVICE 6.2, DATALINK-DISPLAYS 6.0,
;;;  FONT-EDITOR 6.1, SERIAL 6.0, PRINTER 6.3, MAC-PRINTER-TYPES 6.1, PRINTER-TYPES 6.2,
;;;  IMAGEN 6.1, SUGGESTIONS 6.1, MAIL-DAEMON 6.3, MAIL-READER 6.6, TELNET 6.0, VT100 6.0,
;;;  NAMESPACE-EDITOR 6.4, PROFILE 6.2, VISIDOC 6.5, TI-CLOS 6.26, CLEH 6.5, IP 3.56,
;;;  Experimental CLX 6.7, CLUE 6.32, X11M 6.16, Experimental BUG 11.17, MMON 6.2,
;;;   microcode 429, Band Name: REL6-mmon-11/15/89

;;; Major changes for meta-. to handle keywords, CLOS methods, etc.

#!C
; From file SECTIO.LISP#> ZMACS; Hotel:
#8R ZWEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "ZWEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: ZMACS; SECTIO.#"


(DEFUN list-methods-thru-mouse (spec)
  (declare (values spec-list string))
  (LET ((choice-methods (sort (MAPCAR #'THIRD 
                                (list-methods-internal spec))
			      #'(lambda (x y) (and (eq (symbol-package (second x)) *package*)
						   (not (eq (symbol-package (second y)) *package*)))))))
    ;; may 05/21/90 Be able to edit keywords with source-file-properties such as in DEFSYSTEM <foo>
    (when (SOURCE-FILE-NAMES spec nil nil)      ;; may 05/21/90 
      (push spec choice-methods))		;; may 05/21/90

    (COND ((NULL choice-methods)
           (barf "~s is not a method keyword for any flavor." spec))
          ((= (LENGTH choice-methods) 1)
           (values choice-methods (FORMAT nil "~s" (CAR choice-methods))))
          (t (LET ((meth (w:menu-choose (MAPCAR #'(lambda (x) (LIST (prin1-to-string x) :value x))
						choice-methods)
                                        :label "Select a Method to Edit:")))
               
               (IF meth
                   (VALUES (LIST meth) (prin1-to-string meth))
                   (barf "No method chosen. Aborting.")))))))

;;  05/21/90 MAY - Added converison of 'method to 'ticlos:method for user typed
;;                 strings to READ-function-name .
;; 03/28/89 MAY - Added changes to source code.
;;  1/06/89 DNG - Include fix for Scheme DEFINE.  Fix for SETF and LOCF 
;;		function specs.  Fix for DEFWHOPPER.
;;  2/11/89 DNG - Fix for :COND specs.
;;  3/23/89 DNG - Replace test (MEMBER TYPE '(SETF LOCF :COND) :TEST #'EQ) with 
;;		(NOT (MEMBER TYPE '(:METHOD :HANDLER :PROPERTY :INTERNAL) :TEST #'EQ)) so 
;;		that TICLOS:METHOD specs are handled correctly by META-. and we don't have 
;;		to keep patching this for every new kind of function spec.
(DEFUN SYMBOL-FROM-STRING (STR &OPTIONAL LINE OK-TO-ASK SYM &AUX (EOF '(NIL)) ERROR-P)
  "Given a string STR as found after DEF..., return the name of the object being defined.
LINE is the line that the string was found in.  It is used for
finding the particular defining construct used; this affects the result
since (DEFUN (FOO BAR) defines (:PROPERTY FOO BAR)
while (DEFMETHOD (FOO BAR) defines (:METHOD FOO BAR).
OK-TO-ASK means in certain circumstances
where things are not clear, ask the user.  Otherwise we guess.

The arg can also be an object; then its printed representation is used as the string.

The second value is a canonicalized string for the object
 (maybe the same string specified, maybe not).
The third is T if there was a problem
 in parsing the string (such as unbalanced parens).

You can pass the read-in form of the object as the fourth arg
if you already know it."
  (DECLARE (VALUES SYM STR ERROR-P))
  (IF (ARRAYP STR)
      (UNLESS SYM
	(MULTIPLE-VALUE-SETQ (SYM ERROR-P)
			     (CATCH-ERROR (READ-FROM-STRING STR NIL EOF) NIL))
	(SETQ ERROR-P (AND (NULL SYM) ERROR-P)))	; chk 4 error on multi-val funct
							; patch 94.68 ddd/gsl 2/28/84.
      (SETQ SYM STR
	    STR (FORMAT NIL "~S" STR)))
  (COND (ERROR-P
	 (VALUES NIL NIL ERROR-P))
	((SYMBOLP SYM)
	 (VALUES SYM (SYMBOL-NAME SYM)))
	((OR (ATOM SYM) (EQUAL SYM EOF))
	 (VALUES NIL NIL T))
	(T
	 ;; Here SYM is a list.  Certain types of function specs have two ways to
	 ;; type them, with and without the leading type keyword.  Also certain types
	 ;; of functions and other definitions do not follow the standard form
	 ;; of (DEFxxx name options...).  What we do here is to recognize and
	 ;; standardize those cases.  The variables are:
	 ;;	TYPE - the type of function spec or non-function definition
	 ;;	SYM - the function spec or definition name
	 ;;	SPEC - the variant of SYM which appears in the source code
	 ;;	STR - SPEC converted to a string
	 ;; :HANDLER doesn't appear in source files, but gets translated into
	 ;; an appropriate :METHOD here, by analyzing the combined method.
	 ;; :INTERNAL doesn't appear in source files, but might be given as the argument
	 ;; to M-X Disassemble.  The code here just tries not to destory it.
	 (LET ((TYPE (CAR SYM))
	       DELIM-IDX SPEC)
	   (IF (AND (SYMBOLP TYPE)		; Beware of pathological cases.
		    ;;  may 05/21/90 Handle ticlos methods, too - autoconverting method to ticlos:method
		    ;; Return-from SYMBOL-FROM-STRING below for CLOS method specs
		    ;; because (case type ...) below does NOT understand CLOS methods, nor
		    ;; is there any modification or converison that needs to be done
		    ;; that is not handled by read-function-name SYMBOL-FROM-STRING returns to.
		    (prog1 t (when (and (neq :method type)
					(string= "METHOD" (symbol-name type)))
			       (setf (car sym) (setq type 'ticlos:method))))
		    (GET TYPE 'SYSTEM:FUNCTION-SPEC-HANDLER))
	       (IF (AND (NOT (MEMBER TYPE '(:METHOD :HANDLER :PROPERTY :INTERNAL) :TEST #'EQ)) ; handled specially below
			(SYSTEM:VALIDATE-FUNCTION-SPEC SYM))
		   (RETURN-FROM SYMBOL-FROM-STRING (VALUES SYM STR))
		 (SETQ SPEC (CDR SYM)
		       STR (DEFINITION-NAME-AS-STRING TYPE SPEC)))
	       (SETQ SPEC SYM
		     DELIM-IDX (AND LINE (STRING-SEARCH-SET "( " LINE 1))
		     TYPE (COND ((NULL LINE)
				 :MAYBE-METHOD)
				((AND (EQL DELIM-IDX 12)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFMETHOD" 0 12))
				 :ALWAYS-METHOD)
				((AND (EQL DELIM-IDX 13)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFWRAPPER" 0 13))
				 (SETQ SPEC (LIST (CAR SPEC) :WRAPPER (SECOND SPEC)))
				 :ALWAYS-METHOD)
				((AND (EQL DELIM-IDX 12)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFSTRUCT" 0 12))
				 :DEFSTRUCT)
				((AND (EQL DELIM-IDX 12)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFSELECT" 0 12))
				 :DEFSELECT)
				((AND (EQL DELIM-IDX 7)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFINE" 0 7)) ; Scheme DEFINE
				 :DEFSTRUCT) ; use (CAR SPEC)
				((AND (EQL DELIM-IDX 13)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFWHOPPER" 0 13))
				 :ALWAYS-METHOD)
				(T
				 :PROPERTY))))
	   (OR (CASE TYPE
		 (:ALWAYS-METHOD
		  (SETQ SYM (CONS :METHOD SPEC)))
		 ((:METHOD :HANDLER :MAYBE-METHOD)
		  (LET ((FLAVOR (CAR SPEC))
			(MESSAGE (IF (CDDR SPEC)
				     (CADDR SPEC)
				     (CADR SPEC)))
			FL)
		    (COND ((SETQ FL (GET FLAVOR 'SI:FLAVOR)))
			  (OK-TO-ASK
			   (DOLIST (SYMBOL (PACKAGE-LOOKALIKE-SYMBOLS FLAVOR NIL '(SI:FLAVOR)))
			     (IF (FQUERY '(:SELECT T) "Do you mean ~S? "
					 `(:METHOD ,SYMBOL ,@(CDR SPEC)))
				 (RETURN (SETQ FLAVOR SYMBOL
					       SPEC (CONS FLAVOR (CDR SPEC))
					       FL (GET FLAVOR 'SI:FLAVOR)))))))
		    (COND ((SYMBOLP FL)			;T or NIL
			   (AND (EQ TYPE :MAYBE-METHOD)
				(VALIDATE-2-LONG-LIST SPEC)
				(SETQ SYM (CONS :PROPERTY SPEC))))
			  ((FDEFINEDP `(:METHOD . ,SPEC))
			   (SETQ SYM `(:METHOD . ,SPEC)))
			  (OK-TO-ASK
			   (DOLIST (SYMBOL (OR (FIND-COMBINED-METHODS FLAVOR MESSAGE NIL)
					       (SI:FLAVOR-ALL-INHERITABLE-METHODS FLAVOR MESSAGE)))
			     (IF (FQUERY '(:SELECT T) "Do you mean ~S? " SYMBOL)
				 (RETURN (SETQ SYM SYMBOL))))))))
		 (:DEFSTRUCT
		  (SETQ SYM (CAR SPEC))
		  (LOOP WHILE (CONSP SYM) ; this can happen in Scheme DEFINE
			DO (SETQ SYM (CAR SYM)))
		  (SETQ STR (SYMBOL-NAME SYM)))
		 (:DEFSELECT
		  (SETQ SYM (CAR SPEC))
		   (IF (SYMBOLP SYM)
		       (SETQ STR (SYMBOL-NAME SYM))
		       (MULTIPLE-VALUE-SETQ (SYM STR)
			 (SYMBOL-FROM-STRING SYM))))
		 (:PROPERTY
		  (AND (VALIDATE-2-LONG-LIST SPEC)
		       (SETQ SYM (CONS TYPE SPEC))))
		 (:INTERNAL
		  (SETQ SYM (CONS TYPE SPEC))
		  (SETQ STR (DEFINITION-NAME-AS-STRING NIL (CAR SPEC)))))
	       ;; Something we don't understand, make a bogus symbol to use as a property
	       ;; list to remember the location of this definition
	       (SETQ SYM (INTERN STR *UTILITY-PACKAGE*))))
	 (IF (NOT (SYSTEM:VALIDATE-FUNCTION-SPEC SYM))
	     (VALUES NIL NIL T)
	     (VALUES SYM STR)))))

))


#!C
; From file POSS.LISP#> ZMACS; Hotel:
#8R ZWEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "ZWEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: ZMACS; POSS.#"


(DEFCOM COM-EDIT-DEFINITION "Go to the definition of a specified function.
The name of the function is read from the mini-buffer.
Relevant Zmacs variables:
  Edit Definition Is Read Only
  Find Patch Definitions Too" ()
  (IF *NUMERIC-ARG-P*
      (EDIT-NEXT-DEFINITION)
      (LET (SPEC STRING EXPLICIT-PACKAGE-P)
	(SETF (VALUES SPEC STRING EXPLICIT-PACKAGE-P)
	      (READ-FUNCTION-NAME "Edit definition" (RELEVANT-FUNCTION-NAME (POINT))
				  'AARRAY-OK 'MULTIPLE-OK))
        (if (keywordp spec) 			;; may/kb 05/08/89 
	    (multiple-value-setq (spec string) 	;; may/kb 05/08/89 
	      (list-methods-thru-mouse spec)) 	;; may/kb 05/08/89 
            (progn 				;; may/kb 05/08/89
	      ;; may 12/13/89 READ-FUNCTION-NAME does NOT always return spec inside a list, even though
	      ;; 'MULTIPLE-OK was passed in. Insure spec at this point is ALWAYS a list of specs.
              (OR (AND (CONSP SPEC)
		       ;; may 12/13/89
		       ;; (:property :common-lisp fs:file-attribute-bindings) and
		       ;; (METHOD (SETF CLUEI:CONTACT-CALLBACKS) (T CLUEI:BASIC-CONTACT)) both
		       ;; are not caught by form below when default in read-function-name is correct.
		       ;(NEQ (CAR SPEC) :METHOD) 	;; may 05/21/90 Replaced this narrow test with ...
		       (not (si:function-spec-p spec))	;; may 05/21/90 ... this more encompassing test. 
		       )
                  (SETQ SPEC (LIST SPEC)))
              ;; If there's only one entry in the aarray, and its for a different package,
              ;; but the symbol in the current package has some sort of definition in a file,
              ;; include them both.
              (IF (AND (NOT EXPLICIT-PACKAGE-P) (SYMBOLP (CAR SPEC)))
                  (MULTIPLE-VALUE-BIND (THIS-PKG-SYMBOL FOUNDP)
                      (FIND-SYMBOL (STRING-UPCASE (STRING (CAR SPEC))))
                    (IF (AND FOUNDP (NOT (EQUAL THIS-PKG-SYMBOL (CAR SPEC)))	;neq no good for methods, gsl
                             (GET THIS-PKG-SYMBOL :SOURCE-FILE-NAME))
                        (PUSH THIS-PKG-SYMBOL SPEC))))
	      ;;;--------------------------------------------------------
	      ;;;
	      ;;; sort list so that all defs in this package appear first    
	      ;;;
	      ;;;--------------------------------------------------------
	      (if (and (consp (car spec))
		       (equal :method (caar spec)))
		  (setq spec (sort spec #'(lambda (x y)							;; may/kb 05/08/89 
					    (and (eq (symbol-package (second x)) *package*)		;; may/kb 05/08/89 
						 (not (eq (symbol-package (second y)) *package*)))))))	;; may/kb 05/08/89
	      ))
	;; may 12/13/89 Call EDIT-DEFINITION-1-choose since we may have to make a choice for generic-functions
	(EDIT-DEFINITION-1-choose (CAR SPEC)
			   (IF EXPLICIT-PACKAGE-P T SPEC)
			   STRING)))
  (IF *EDIT-DEFINITION-IS-READ-ONLY*
      (MAKE-BUFFER-READ-ONLY *INTERVAL*))
  DIS-TEXT)

(DEFUN EDIT-DEFINITION-1-choose (OBJECT &optional ignore ignore)
  "Handle multiple definitions by poping up a menu of choices to choose from."
  (let ((definitions '()))
    (flet ((add-def (kind key &optional (object object))	;; may 12/13/89 Added optional arg
		    (push (list (format nil "~A ~S" kind object) key) definitions)))
      (when (symbolp object)
	(let ((system (sys:find-system-named object t t)))
	  (unless (null system)
	    (add-def "system" 'defsystem (si:force-to-keyword-symbol object))))	;; may 12/13/89 
	(when (or (compiler:specialp object)
		  (si:get-source-file-name object 'defvar))
	  (add-def (if (constantp object) "constant" "variable") 'defvar))
	(let ((class (clos:find-class object nil)))
	  (if class
	      (typecase class
		(ticlos:flavor-class (add-def "flavor" 'defflavor))
		(ticlos:structure-class (add-def "structure" 'defstruct))
		(t (add-def "class" 'clos:defclass)))
	      (if (get object 'si:defstruct-description)
		  (add-def "structure" 'defstruct))))
	)
      (when (sys:validate-function-spec object)
	(let ((names (si:function-spec-get object :source-file-name)))
	  (when (consp names)
	    (dolist (x names)
	      (unless (or (eq (car x) 'defun)
			  (member (car x) definitions :key #'second))
		(add-def (case (car x)
			   (sys:encapsulation "encapsulation on")
			   (defpackage "package")
			   (otherwise (car x)))
			 (car x))))))
	(when (sys:fdefinedp object)
	  (cond ((ticlos:generic-function-p object)
		 (dolist (method (clos:generic-function-methods
				   (sys:fdefinition-safe object t)))
		   (let ((method-spec (function-name (clos:method-function method))))
		     (unless (member method-spec definitions :test #'equal :key #'second)
		       (push (LIST (prin1-to-string method-spec) method-spec)
			     definitions))))
		 ;; Since there may be a lot of methods, sort in alphabetical order to make 
		 ;; it easier to find the desired one.
		 (setf definitions (sort definitions #'string< :key #'first))
		 (when (si:get-source-file-name object 'defun)
		   (add-def 'clos:defgeneric 'defun)))
		((not (symbolp object)) (add-def "function" 'defun))
		((special-form-p object) (add-def "special form" 'defun))
		((macro-function object) (add-def "macro" 'defun))
		(t (add-def "function" 'defun))
		)))
      )	; end flet

    (let (choice)
      (case (length definitions)
	(0 (let ((new (unless (keywordp object)	;; may 12/13/89
			;; may 12/13/89 
			;; we may pass in a list of symbols here (list 'tv:ged 'ged) but no harm seems to be done ...
			(SYSTEM:DWIMIFY-PACKAGE-0 OBJECT 'ZMACS-DEFINEDP))))
	     (if (and new (not (equal new object)))
		 (EDIT-DEFINITION-1-choose new)
		 (EDIT-DEFINITION-1 object (if (keywordp object) t (list object))))))	;; may 12/13/89 
	(1 (edit-definition-1 object t))
	(t (setq choice (w:menu-choose definitions :label
				       "Choose definition to edit"))))
      (when choice
	(if (symbolp choice)
	    (edit-definition-1 object t nil nil choice)
	      (edit-definition-1 choice t nil nil 'defun)))
      )))

(DEFUN EDIT-DEFINITION-1 (OBJECT &OPTIONAL (OBJECTS (LIST OBJECT))
			  STRING DEFINITION-GENERIC-PATHNAME DEFINITION-TYPE
			  &AUX DEF)
  "Visit the definition(s) of OBJECT, or OBJECTS.
STRING will be used eventually to look for look-alike objects,
so it should be a printed representation of OBJECT.
If OBJECTS is T, only OBJECT precisely is used, no matter how desperate the user gets.
Then STRING is not needed.
DEFINITION-GENERIC-PATHNAME restricts to definitions in that file,
and DEFINITION-TYPE restricts to that type of definition.
DEFINITION-TYPE should be something like DEFUN, DEFVAR, etc., or NIL."
  ;; may 05/21/90  SPR 10617
  ;; Note: meta-. for symbols such as 'tv:window define a defsystem.
  ;; we must change the OBJECT to the keyword version of the symbol since THAT
  ;; is where the source-file-name property is kept! 
  ;;;(DEFSYSTEM-1
  ;;;     ',(force-to-keyword-symbol name) ...)
  ;;;(DEFUN DEFSYSTEM-1 (NAME OPTIONS &OPTIONAL (ADD-P T) DONT-RECORD-IN-PATHNAMES &AUX PLIST)
  ;;;  (AND (RECORD-SOURCE-FILE-NAME NAME 'DEFSYSTEM) ...))
  (when (eq t objects)
    (when (or (keywordp object)
	      (and (eq DEFINITION-TYPE 'defsystem) (setq object (si:force-to-keyword-symbol object))))
      ;; must also convert to the correct name that was passed to defsystem - not a nickname
      (let ((system (si:find-system-named object t t)))
	(when system (setq object (si:system-symbolic-name system))))))

  (SETQ OBJECT (FLUSH-INTERNAL-SPEC OBJECT))
  (IF (CONSP OBJECTS)
      (SETQ OBJECTS (MAPCAR 'FLUSH-INTERNAL-SPEC OBJECTS)))
  (AND (OR (EQ OBJECTS T)
	   (AND (NULL (CDR OBJECTS))
		(EQUAL (CAR OBJECTS) OBJECT)))
       (SETQ DEF (ONLY-DEFINITION-OR-NIL OBJECT DEFINITION-TYPE)))
  (IF DEF
      ;; If there is only one definition of this object that could possibly be meant,
      ;; and it is in a buffer and still real, just go there.
      ;; Don't do any hacking with the possibilities buffer.
      ;; This is probably the most common case, so it should be fast.
      ;; If there is only one definition but it is in a file,
      ;; that is going to be slow enough anyway so no need to special case it.
      (PROGN
	(SOURCE-FILE-NAMES OBJECT T DEFINITION-TYPE)	;Print names of any patch files.
	(POINT-PDL-PUSH (POINT) *WINDOW* T)
	(MAKE-BUFFER-CURRENT (CAR DEF))
	(MOVE-BP (POINT) (CDR DEF) 0)
	(RECENTER-WINDOW *WINDOW* :START
			 (BACKWARD-OVER-COMMENT-LINES (OR (BACKWARD-OVER-PACKAGE-PREFIX (POINT))
							  (POINT))
						      NIL)))
      (INSERT-EDIT-DEFINITION-POSSIBILITY OBJECT OBJECTS STRING
					  DEFINITION-GENERIC-PATHNAME
					  DEFINITION-TYPE)))


;; may 05/21/90 for DNG - Original.
(DEFUN DEFINITION-TYPE (SYMBOL)
   (CASE SYMBOL
      (( DEFMACRO DEFSUBST CLOS:DEFGENERIC DEFMETHOD DEF DEFF DEFF-MACRO 
	 DEFSELECT DEFUN-METHOD ZLC:DEFUNP DEFWHOPPER DEFWRAPPER)
        'DEFUN)
      (( DEFPARAMETER DEFCONSTANT ZLC:DEFCONST) 'DEFVAR)
      (T (GET SYMBOL 'DEFINITION-TYPE SYMBOL))))

(DEFPROP SYS:DEF-MISC-OP DEFUN DEFINITION-TYPE)
(DEFPROP SYS:DEF-UCODE-ENTRY DEFUN DEFINITION-TYPE)
(DEFPROP COMPILER:DEF-MISC-OP DEFUN DEFINITION-TYPE)
(DEFPROP COMPILER:DEF-UCODE-ENTRY DEFUN DEFINITION-TYPE)

(DEFUN FIND-DEFINITIONS-BUFFERS-AND-FILES (SPECS DEFINITION-TYPE
					   &AUX BUFFERS-TO-RETURN FILES
					   SPEC-SOURCE-FILES SPEC-PATCHES)	;gsl 10-28-83
  "Find all ZMACS buffers, and all non-visited files, that have definitions of SPECS.
SPECS is a list of function specs, and definitions of any of them count.
DEFINITION-TYPE is the type of definition (DEFUN, DEFVAR, etc). to look for,
 unless it is NIL.
The first value is a list of ZMACS buffers; the second, a list of pathnames
of files not visited in ZMACS."
  (DOLIST (SPEC SPECS)
    (LET ((BUFFERS (SI:FUNCTION-SPEC-GET SPEC 'ZMACS-BUFFERS)))
      (DOLIST (ELT BUFFERS)
	(LET ((BUFFER (CAR ELT)))
	  (IF (MEMBER BUFFER *ZMACS-BUFFER-LIST* :TEST #'EQ)
	      (OR (MEMBER BUFFER BUFFERS-TO-RETURN :TEST #'EQ)
		  ;; may 05/21/90 (for DNG) Start patch ...
		  (AND DEFINITION-TYPE
		        (LET ((FN (IGNORE-ERRORS (READ-FROM-STRING (CDR ELT) NIL NIL :START 1))))
			   (AND FN (SYMBOLP FN) (NOT (EQ DEFINITION-TYPE (DEFINITION-TYPE FN))))))
		  ;; may 05/21/90 .. end patch
		  (PUSH BUFFER BUFFERS-TO-RETURN))
	      (IF (TYPEP BUFFER 'FILE-BUFFER)
		  (PUSH (OR (BUFFER-PATHNAME BUFFER)
			    (BUFFER-NAME BUFFER))
			FILES))))))
    (MULTIPLE-VALUE-SETQ (SPEC-SOURCE-FILES SPEC-PATCHES)		;gsl 10-28-83
      (SOURCE-FILE-NAMES SPEC T DEFINITION-TYPE))
    (SETQ FILES (IF *FIND-PATCH-DEFINITIONS-TOO*			;gsl. 3-25-85
		    (NCONC SPEC-SOURCE-FILES SPEC-PATCHES FILES)	;gsl 10-28-83
		    (NCONC SPEC-SOURCE-FILES FILES))))			;gsl. 3-25-85
  (VALUES BUFFERS-TO-RETURN
	  (REMOVE-IF-NOT #'(LAMBDA (FILE &AUX (GENERIC-PATHNAME (SEND FILE ':GENERIC-PATHNAME)))
			     (NOT (DOLIST (BUF BUFFERS-TO-RETURN)
				    (AND (EQ (BUFFER-GENERIC-PATHNAME BUF)
					     GENERIC-PATHNAME)
					 (RETURN T)))))
			 FILES)))

(DEFUN SOURCE-FILE-NAMES (FUNCTION-SPEC &OPTIONAL PRINT-P TYPE &AUX LIST)
  "Return a list of source file pathnames containing definitions of FUNCTION-SPEC.
PRINT-P says also print the names of any patch files that have redefined it.
TYPE is the type of definition (DEFUN, DEFVAR, etc). to look for,
 unless it is NIL."
  ;; may 05/21/90 Start patch ..
  (LOOP WHILE (EQ (CAR-SAFE FUNCTION-SPEC) ':INTERNAL) ; in case not defined anymore, don't want to error
	 DO (SETQ FUNCTION-SPEC (SECOND FUNCTION-SPEC)))

  ;; may 05/21/90 Form below existed previously ..
  (SETQ LIST (SI:GET-ALL-SOURCE-FILE-NAMES FUNCTION-SPEC))

  (WHEN (AND (NULL LIST)
	       (SYMBOLP FUNCTION-SPEC)
	       (EQ (GET FUNCTION-SPEC 'SPECIAL) 'COMPILER:INIT-SYSTEM-VAR-PROPERTIES))
     (SETQ LIST '((DEFVAR #,(SEND (PATHNAME "SYS:UCODE;LROY-QCOM.LISP") :GENERIC-PATHNAME)))))
  ;; may 05/21/90  .. End patch

  (SETQ LIST (MAPCAN #'(LAMBDA (ELT)
			 (AND (OR (NULL TYPE)
				  (EQ (CAR ELT) TYPE))
			      (COPY-LIST (CDR ELT))))
		     LIST))
  ;; All patch files that have defined the function since the real file should be noted.
  (LOOP FOR FILE IN LIST
	AS PATCH-FILE-P = (SEND FILE :GET :PATCH-FILE)
	AS SOURCE-FILE = (SEND (SEND FILE :SOURCE-PATHNAME) :TRANSLATED-PATHNAME)
	WHEN (NOT PATCH-FILE-P)
	COLLECT SOURCE-FILE INTO REAL-FILES
	WHEN (AND PATCH-FILE-P (NULL REAL-FILES))
	COLLECT SOURCE-FILE INTO PATCH-FILES
	FINALLY (AND PRINT-P PATCH-FILES
		     (FORMAT *QUERY-IO* "~&~S ~:[also~;only~] defined by patch file~P ~{~A~^, ~}"
			     FUNCTION-SPEC (NOT REAL-FILES) (LENGTH PATCH-FILES) PATCH-FILES))
	(RETURN (VALUES (OR REAL-FILES PATCH-FILES)
			(IF REAL-FILES PATCH-FILES)))))
))
