;;-*-Mode:Common-LISP;Package:FILE-SYSTEM; Base:10.; COLD-LOAD:T -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985- 1989 Texas Instruments Incorporated. All rights reserved.
;;;  ** (c) Copyright 1980 Massachusetts Institute of Technology **

;;;    
;;; REVSIONS:			PUT LATEST AT TOP
;;;     03-17-88 DAB    Added :root to (PATHNAME :PARSE-DIRECTORY-SPEC).
;;; 	03.08.88 MBC	Add :NEW-NAME-AND-TYPE method to support NFS server.
;;;			Fix make-pathname-1 to disallow string directory component unless
;;;			in proper list form.
;;;	03.07.88 MBC	Rosbustize :PATHNAME-AS-DIRECTORY for unexpected name components.
;;;	02.20.88 DNG	Major overhaul of code for GC'ing unused pathnames.
;;;     02.10.88 MBC	New (temporary) method, :MERGE-DEVICE, for MAC until it really uses :DEVICE.
;;;	02.03.88 MBC	:STRING-FOR-NFS added for all pathnames.
;;;	02.02.88 MBC	Reintern-Pathnames comments added.
;;;	04.30.87 MBC	Allow restore of files to lm:~; without Length of :ROOT error.
;;;	04.13.87 MBC	Repair :NEW-SUGGESTED-DIRECTORY to work properly.  1.) For Explorer 
;;;			pathnames allow list form for directory.  2.)  For normally lowercase
;;;			pathnames, coerce non-mixed types to lowercase.
;;;	04.01.87 MBC	Fix :WILDCARD-MAP to not change case of name or type.
;;;	03.27.87 MBC    Fix :WILDCARD-MAP to traverse subs even when type field is not :WILD.
;;;	03.24.87 MBC    Tweak to :CANONICAL-TYPE so :WILD isn't lost. (Fixes dired getting non-wild type.)
;;;	03.18.87 MBC	Fix :CANONICAL-TYPE to return second value properly and make
;;;			:NEW-CANONICAL-TYPE handle original-type arg for PATHNAME-NORMALLY-LOWERCASE-MIXIN.
;;;	03.12.87 MBC	Plug a hole in :generic-pathname's handling of device.
;;;	03.12.87 MBC	Fix :wildcard-map to force name & type to wild IF we need to include sub-directories.
;;;	03.09.87 MBC	Two SPEEDUPS: Use VERIFY-LIST-OF-DIRECTORIES-1 instead of GET-LIST-OF-DIRECTORIES
;;;			in :WILDCARD-MAP.  Also don't force name & type to wild for directory-list.
;;;	03.09.87 MBC	Add :PROBE-DIRECTORY support.
;;;     02.11.87 HE     Changed :patch-file-pathname to always use :patch-directory for type.
;;;	02.05.87 MBC	:GENERIC-PATHNAME should NEVER translate a logical
;;;			pathname, nor try to backtranslate it.
;;;  	01.28.87 MBC	Make sure copy-file do things alphabetically
;;;  			[call directory-list with :SORTED keyword]
;;;	01.11.87 MBC	Make generic pathname handle logical pathnames whose
;;;			directories do not have translations.
;;;     06.16.87 MBC    Add new method, :NEW-PREFERRED-CANONICAL-TYPE.
;;; 	06.17.87 MBC 	Make :PARSE-TRUENAME convert no type to :UNSPECIFIC.  Added new optional arg, host.
;;;	06.24.87 MBC    Get rid of VERIFY-LIST-OF-DIRECTORIES and put in Report stream handling to
;;;			notify user of non-existent & empty directories.
;;;	06.25.87 MBC	Clean up PATHNAME-AREA support.  Invoked during full-gc.
;;;	07.16.87 MBC	Duplicate definition, :PARSE-TYPE-SPEC, deleted.
;;;	07.22.87 MBC 	Remove duplicate definition, (:METHOD PATHNAME :HOST)
;;;     12-17-87 DAB    Added :oldest as a valid version to (PATHNAME :PARSE-VERSION-SPEC).

;;; 
;;; THESE are the actual base flavors for pathnames.
;;; They used to live in Io;Pathnm.  9-6-86
;;; 

(DEFFLAVOR PATHNAME
	   (HOST
	     (DEVICE NIL)
	     (DIRECTORY ())
	     (NAME NIL)
	     (TYPE NIL)
	     (VERSION NIL)
	     (SI:PROPERTY-LIST NIL)		;Where files properties are stored
	     (STRING-FOR-PRINTING NIL)
	     (SHORT-STRING-FOR-PRINTING NIL))
	   (SI:PROPERTY-LIST-MIXIN)
  :ORDERED-INSTANCE-VARIABLES
  (:ACCESSOR-PREFIX PATHNAME-RAW-)
  (:OUTSIDE-ACCESSIBLE-INSTANCE-VARIABLES HOST DEVICE DIRECTORY NAME TYPE VERSION  SI:PROPERTY-LIST)
  (:GETTABLE-INSTANCE-VARIABLES HOST DEVICE DIRECTORY NAME TYPE VERSION SI:PROPERTY-LIST)
  (:INITABLE-INSTANCE-VARIABLES HOST DEVICE DIRECTORY NAME TYPE VERSION SI:PROPERTY-LIST)
  (:METHOD-COMBINATION (:OR :BASE-FLAVOR-LAST :WILD-P))
  (:REQUIRED-METHODS :PARSE-NAMESTRING))


;;; PRINC of a pathname is just like PRINC of the :STRING-FOR-PRINTING
;;; PRIN1 prints inside a # ...  so it can be read back.

(DEFMETHOD (PATHNAME :PRINT-SELF) (STREAM IGNORE SLASHIFY-P)
  (IF SLASHIFY-P
    (LET ((*PACKAGE* SI:PKG-USER-PACKAGE))
      (PRINC "#" STREAM)
      (PRIN1 (TYPE-OF SELF) STREAM)
      (WRITE-CHAR #\SPACE STREAM)
      (PRIN1 (SEND SELF :STRING-FOR-PRINTING) STREAM)
      (WRITE-CHAR #\ STREAM))
    (FUNCALL STREAM :STRING-OUT (SEND SELF :SHORT-STRING-FOR-PRINTING))))	;SHORT



(DEFMETHOD (PATHNAME :DESCRIBE) ()
  (FORMAT T "~&~S, an object of flavor ~S,~% has instance variable values:~%" SELF
	  (TYPE-OF SELF))
  (DO ((IVARS (SI:FLAVOR-ALL-INSTANCE-VARIABLES (SI:INSTANCE-FLAVOR SELF)) (CDR IVARS))
       (I 1. (1+ I)))
      ((NULL IVARS))
    (UNLESS (EQUAL (CAR IVARS) 'QUOTED-STRING)
      (FORMAT T "	~S:~27T " (CAR IVARS))
      (COND
	((= (si:%P-DATA-TYPE-OFFSET SELF I) DTP-NULL) (FORMAT T "unbound~%"))
	(T (FORMAT T "~S~%" (%INSTANCE-REF SELF I)))))))  

;And this is what is called to read back a PRIN1'd pathname.


(DEFMETHOD (PATHNAME :READ-INSTANCE) (IGNORE STREAM)
  (PARSE-PATHNAME (READ STREAM T)))  



(DEFMETHOD (PATHNAME :PLIST) ()
  SI:PROPERTY-LIST)

;;  This is a duplicate of an instance variable already available 
;;  from the flavor, PATHNAME.  7.22.87 MBC
;;(DEFMETHOD (PATHNAME :HOST) ()
;;  HOST)

;For bootstrapping.

;;; Duplicate definition deleted.  7.16.87 MBC
;;;(DEFMETHOD (PATHNAME :PARSE-TYPE-SPEC) (SPEC) 
;;;	   (SEND SELF :PARSE-COMPONENT-SPEC SPEC))  



(DEFMETHOD (PATHNAME :CHARACTER-TRANSLATIONS) ()
  SELF)  



(DEFMETHOD (PATHNAME :CHARACTER-BACK-TRANSLATIONS) (PATH)
  PATH)

;;; Moved here 9-5-86 MBC from OPEN
(DEFMETHOD (PATHNAME :DIRECTORY-LIST) (OPTIONS)
  (LET ((ERROR
	 (MAKE-CONDITION 'UNKNOWN-OPERATION "Can't list directory on file system of ~A." SELF
			 :DIRECTORY-LIST)))
    (IF (MEMBER :NOERROR OPTIONS :TEST #'EQ)
	ERROR
	(SIGNAL ERROR)))) 

(DEFUN DEFAULT-DIRECTORY-LIST-STREAM (OP &REST ignore)
  (DECLARE (SPECIAL  DIRECTORY-LIST))
  (CASE OP
    (:ENTRY (POP DIRECTORY-LIST))
    (:CLOSE )))

(DEFMETHOD (PATHNAME :DIRECTORY-LIST-STREAM) (OPTIONS &AUX &SPECIAL DIRECTORY-LIST)
  (IF (ERRORP (SETQ DIRECTORY-LIST (SEND SELF :DIRECTORY-LIST OPTIONS)))
    DIRECTORY-LIST
    (CLOSURE '(DIRECTORY-LIST) #'DEFAULT-DIRECTORY-LIST-STREAM))) 
  

(DEFMETHOD (PATHNAME :ALL-DIRECTORIES) (OPTIONS)
  (LOOP FOR FILE IN (CDR
		      (APPLY #'DIRECTORY-LIST
			     (SEND SELF :NEW-PATHNAME :DIRECTORY :ROOT
				        :NAME :WILD :TYPE :WILD :VERSION :WILD)
			     OPTIONS))
     WHEN (GET FILE :DIRECTORY)
     COLLECT (CONS (SEND (CAR FILE) :PATHNAME-AS-DIRECTORY) ())))

(DEFMETHOD (PATHNAME :PROBE-DIRECTORY) ()	;3.9.87 MBC
  :MAYBE)					;default for everybody is we're not sure ==> try it.

(DEFMETHOD (PATHNAME :MEANINGFUL-ROOT) () T)	;4.30.87


;;(DEFMETHOD (PATHNAME :ALL-DIRECTORIES) (OPTIONS)
;;  (LET ((ERROR
;;	 (MAKE-CONDITION 'UNKNOWN-OPERATION "Can't list all directories on file system of ~A."
;;			 SELF :ALL-DIRECTORIES)))
;;    (IF (MEMBER :NOERROR OPTIONS :TEST #'EQ)
;;	ERROR
;;	(SIGNAL ERROR)))) 

;;(DEFFLAVOR MEANINGFUL-ROOT-MIXIN () () (:REQUIRED-FLAVORS PATHNAME)
;;   (:REQUIRED-METHODS :DIRECTORY-PATHNAME-AS-FILE)
;;   (:METHOD-COMBINATION (:DAEMON-WITH-OR :BASE-FLAVOR-LAST :VALID-DIRECTORY-COMPONENT-P))
;;   (:DOCUMENTATION :MIXIN
;;    "For use with file systems where the root directory is treated
;;as an ordinary directory."))  

;;; Get rid of MEANINGFUL-ROOT-MIXIN.  Make its one and only method :all-directories become
;;; the default by putting it in Pathname.  Local-file-access has its own.
;;; Of course rip MEANINGFUL-ROOT-MIXIN out of all the other deafflavors.  MBC 10-24-86

(DEFMETHOD (PATHNAME :EXPUNGE) (&REST ARGS)
  (LET ((ERROR
	 (MAKE-CONDITION 'UNKNOWN-OPERATION "Can't expunge or undelete on file system of ~A."
			 SELF :EXPUNGE))
	(ERRORP (COND
		  ((NULL ARGS) T)
		  ((NULL (CDR ARGS)) (CAR ARGS))
		  (T (GETF ARGS :ERROR)))))
    (IF ERRORP
	(SIGNAL ERROR)
	ERROR))) 


(DEFMETHOD (PATHNAME :REMOTE-CONNECT) (&KEY &OPTIONAL ERROR ACCESS)
  ACCESS
  (LET ((CONDITION
	 (MAKE-CONDITION 'UNKNOWN-OPERATION
			 "Can't do remote connect or access on file system of ~A." SELF
			 :REMOTE-CONNECT)))
    (IF ERROR
	(SIGNAL CONDITION)
	CONDITION))) 

(DEFMETHOD (PATHNAME :CREATE-LINK) (LINK-TO &KEY (ERROR T))
  LINK-TO
  (LET ((CONDITION
	 (MAKE-CONDITION 'UNKNOWN-OPERATION "Can't create links on file system of ~A." SELF
			 :CREATE-LINK)))
    (IF ERROR
	(SIGNAL CONDITION)
	CONDITION))) 

(DEFMETHOD (PATHNAME :CREATE-DIRECTORY) (&KEY &OPTIONAL (ERROR T))
  (LET ((CONDITION
	 (MAKE-CONDITION 'UNKNOWN-OPERATION "Can't create directory on file system of ~A." SELF
			 :CREATE-DIRECTORY)))
    (IF ERROR
	(SIGNAL CONDITION)
	CONDITION))) 

;;; To here. 9-6-86 MBC

;Here for bootstrapping.
(DEFMETHOD (PATHNAME :CANONICAL-TYPE) ()
  (LET ((-TYPE- (SEND SELF :TYPE))
	(SYSTEM-TYPE (SEND HOST :SYSTEM-TYPE))
	(ALPHABETIC-CASE-AFFECTS-STRING-COMPARISON T))
    (IF (NOT (MEMBER -TYPE- '(NIL :UNSPECIFIC :WILD) :TEST #'EQUAL))
	(LOOP FOR (CANONICAL-TYPE DEFN) ON CANONICAL-TYPES BY 'CDDR
	      AS PER-SYSTEM = (OR (ASSOC SYSTEM-TYPE DEFN :TEST #'EQ) (ASSOC () DEFN :TEST #'EQ))
	      DO
	      (IF (MEMBER -TYPE- (CDR PER-SYSTEM) :TEST #'EQUAL)
		  (RETURN CANONICAL-TYPE (UNLESS (EQUAL -TYPE- (CADR PER-SYSTEM))
					   -TYPE-)))
	      FINALLY (RETURN -TYPE- NIL))	;3.18.87 Return second value as NIL  [SPR 4014]
	(values -TYPE- NIL))))			;3.24.87 Return :wild etc too.


(DEFMETHOD (PATHNAME :RAW-DEVICE) ()
  DEVICE)  

(DEFMETHOD (PATHNAME :RAW-DIRECTORY) ()
  DIRECTORY)  

(DEFMETHOD (PATHNAME :RAW-NAME) ()
  NAME)  

(DEFMETHOD (PATHNAME :RAW-TYPE) ()
  TYPE)  

(DEFMETHOD (PATHNAME :RAW-QUOTED-STRING) ())  

(DEFUN PATHNAME-HOST (P)
  (PATHNAME-RAW-HOST (IF (TYPEP P 'PATHNAME)
		       P
		       (PARSE-PATHNAME P))))  


(DEFMETHOD (PATHNAME :INIT) (IGNORE)
  (OR (VARIABLE-BOUNDP HOST) (FERROR () "Host must be specified when initializing a pathname")))   

;;; Caching of strings


(DEFWRAPPER (fs:PATHNAME :STRING-FOR-PRINTING) (IGNORE . BODY)
   `(CACHE-IN-VARIABLE STRING-FOR-PRINTING ,@BODY))

(DEFWRAPPER (PATHNAME :SHORT-STRING-FOR-PRINTING) (IGNORE . BODY)
   `(CACHE-IN-VARIABLE SHORT-STRING-FOR-PRINTING ,@BODY))

(DEFMETHOD (PATHNAME :STRING-FOR-WHOLINE) (&OPTIONAL (LENGTH ()))
  (LET ((WHOLINE-STRING (SEND SELF :SHORT-STRING-FOR-PRINTING)))
    (IF (AND LENGTH (> LENGTH (LENGTH WHOLINE-STRING)))
	WHOLINE-STRING
	(SUBSEQ WHOLINE-STRING 0 LENGTH))))

(DEFMETHOD (PATHNAME :STRING-FOR-DIRED) ()
  (SEND SELF :STRING-FOR-PRINTING))  

(DEFMETHOD (PATHNAME :STRING-FOR-MINI) ()
  (SEND SELF :STRING-FOR-PRINTING))  

(DEFMETHOD (PATHNAME :STRING-FOR-DIRECTORY) ()
  (SEND SELF :STRING-FOR-PRINTING))  


(DEFMETHOD (PATHNAME :WILD-P) ()
  "Checks if the pathname has a wild character anywhere in the pathname string"
  (IF (OR (SEND SELF :DEVICE-WILD-P) (SEND SELF :DIRECTORY-WILD-P) (SEND SELF :NAME-WILD-P)
      (SEND SELF :TYPE-WILD-P) (SEND SELF :VERSION-WILD-P))
    T))  


(DEFMETHOD (PATHNAME :DEVICE-WILD-P) ()
  (SEND SELF :COMPONENT-CONTAINS-WILD DEVICE))  


(DEFMETHOD (PATHNAME :DIRECTORY-WILD-P) ()
  (OR (SEND SELF :COMPONENT-CONTAINS-WILD DIRECTORY) (EQ DIRECTORY :ROOT)))  

(DEFMETHOD (PATHNAME :NAME-WILD-P) ()
  (SEND SELF :COMPONENT-CONTAINS-WILD NAME))

(DEFMETHOD (PATHNAME :TYPE-WILD-P) ()
  (SEND SELF :COMPONENT-CONTAINS-WILD TYPE))  

(DEFMETHOD (PATHNAME :VERSION-WILD-P) ()
  (SEND SELF :COMPONENT-CONTAINS-WILD VERSION))  

(DEFMETHOD (PATHNAME :PARSE-DIRECTORY-WILD-SPECS) (DIR-SPECS)
  (LET ((NATURAL-SPECS)
	(WILD-SPECS (IF (ATOM DIR-SPECS)
		      (CONS DIR-SPECS ())
		      DIR-SPECS)))
    (SETQ DIR-SPECS WILD-SPECS)
    (DOLIST (DIR-COMPONENT DIR-SPECS)
      (IF (SEND SELF :COMPONENT-CONTAINS-WILD DIR-COMPONENT)
	(RETURN (VALUES (OR NATURAL-SPECS :ROOT) WILD-SPECS))
	(SETQ WILD-SPECS (CDR WILD-SPECS)
	      NATURAL-SPECS (APPEND NATURAL-SPECS (CONS DIR-COMPONENT ())))))
    (VALUES (OR NATURAL-SPECS :ROOT) WILD-SPECS)))  

(SI:MAKE-OBSOLETE FS:VERIFY-LIST-OF-DIRECTORIES "use FS:DIRECTORY-LIST")

;; push :short for NFS copy-file -- Tung
(DEFMETHOD (PATHNAME :WILDCARD-MAP) (OPERATION PLISTP DIR-LIST-OPTIONS &REST OPTIONS
				     &KEY PASS-DIRECTORIES-ON VERBOSE REPORT-STREAM
				     &ALLOW-OTHER-KEYS)
  "maps wildcarded pathname components:directory, name, type and version numbers are wildcarded
  as * only."
  (LET ((r-stream (or report-stream (get options :report-stream)))	;6.24.87
	(verbose1 (or verbose (get options :verbose)))
	RESULTS)
    (UNLESS (MEMBER :SORTED DIR-LIST-OPTIONS :TEST #'EQ)
      (SETQ DIR-LIST-OPTIONS (APPEND DIR-LIST-OPTIONS (CONS :SORTED ()))))
    (UNLESS (MEMBER :NOERROR DIR-LIST-OPTIONS :TEST #'EQ)
      (SETQ DIR-LIST-OPTIONS (APPEND DIR-LIST-OPTIONS (CONS :NOERROR ()))))
    (LET ((DLIST
	    (APPLY #'DIRECTORY-LIST
		   (SEND self :NEW-PATHNAME
			 :RAW-NAME (if PASS-DIRECTORIES-ON :wild (or name :WILD))	;Use RAW so case
			 :RAW-TYPE (if PASS-DIRECTORIES-ON :wild (or type :WILD))	;doesn't change 4.1.87
			 :VERSION (IF (MEMBER VERSION '(:NEWEST :OLDEST) :TEST #'EQ)
				      VERSION
				      :WILD))
		   (IF (MEMBER :short options)
		       (PUSH :short DIR-LIST-OPTIONS)	;ab
		       DIR-LIST-OPTIONS))))
      
      (WHEN VERBOSE1 (UNLESS R-STREAM (SETQ R-STREAM *STANDARD-OUTPUT*)))	;6.24.87
      
      ;;; When a directory does not exist, Directory-List is inconsistent:
      ;;; If its an Explorer an error object is returned.  Otherwise the 
      ;;; returned list has a NIL Cdr, same as a list of an empty directory.
      ;;; So we never know if a directory is empty or doesn't exist except
      ;;; when we get an error object back.
      (if (ERRORP DLIST)
	  (when r-stream (format r-stream "~&[~a]" (send dlist :report-string)))	;6.24.87
	  (if (cdr dlist)
	      (DOLIST (DLIST-ENTRY (CDR DLIST))
		(WHEN (OR (NOT (GET DLIST-ENTRY :DIRECTORY)) PASS-DIRECTORIES-ON)
		  (IF (OR (GET DLIST-ENTRY :DIRECTORY)
			  (SEND SELF :PATHNAME-MATCH-NO-DEVICE (CAR DLIST-ENTRY) ()))
		      (when (NOT PLISTP)
			(SETQ DLIST-ENTRY (CAR DLIST-ENTRY)))
		      (SETQ DLIST-ENTRY ()))
		  (OR (ERRORP DLIST-ENTRY) (NULL DLIST-ENTRY)
		      (LET ((RETURNED-FILE (APPLY OPERATION DLIST-ENTRY OPTIONS)))
			(OR (ERRORP RETURNED-FILE) (NULL RETURNED-FILE)
			    (PUSH RETURNED-FILE RESULTS))))))
	      (when r-stream (format r-stream "~&[Directory not found or empty for ~a]"  self)))))	;6.24.87
    (NREVERSE RESULTS)))


(DEFMETHOD (PATHNAME :SUPPRESSED-DEVICE-NAMES) ()
  '(NIL :UNSPECIFIC))  

(DEFMETHOD (PATHNAME :PRIMARY-DEVICE) ()
  (FUNCALL HOST :PRIMARY-DEVICE))  

;T if files of this host are can be undeleted.


(DEFMETHOD (PATHNAME :UNDELETABLE-P) ()
  ())  

;T if a namestring with no file type, on this system, reads in as :UNSPECIFIC.
;This implies that LOAD should treat :UNSPECIFIC in the type
;as possibly a request for a default, or possibly a specification
;of a file whose name includes no type.


(DEFMETHOD (PATHNAME :UNSPECIFIC-TYPE-IS-DEFAULT) ()
  ())  

(DEFMETHOD (PATHNAME :PATHNAME-AS-DIRECTORY) ()
  (if (member NAME '(:UNSPECIFIC :WILD NIL))	;3.07.88 MBC
      (send self :new-pathname :NAME :UNSPECIFIC :TYPE :UNSPECIFIC :VERSION :UNSPECIFIC)
      (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY
	    (IF (EQ DIRECTORY :ROOT)
		(CONS NAME ())
		(APPEND (IF (CONSP DIRECTORY)
			    DIRECTORY
			    (CONS DIRECTORY ()))
			(CONS NAME ())))
	    :NAME :UNSPECIFIC :TYPE :UNSPECIFIC :VERSION :UNSPECIFIC)))  


(defvar *DONT-BACKTRANSLATE-GENERIC-PATHNAME* t "Dont waste time backtranslating for generic pathnames.")

(DEFMETHOD (PATHNAME :GENERIC-PATHNAME) ()
  (LET* ((TYP (SEND SELF :CANONICAL-TYPE))
	 (NEW-TYPE (FUNCALL HOST :GENERIC-BASE-TYPE TYP))
	 (DEV (SEND SELF :DEVICE))
	 (DEV1 dev)				;3.12.87
	 NEW-HOST NEW-DEVICE NEW-DIRECTORY PATH)
    (AND (MEMBER DEV (SEND SELF :SUPPRESSED-DEVICE-NAMES) :TEST #'EQ)
	 (SETQ DEV1 (FUNCALL HOST :PRIMARY-DEVICE)))
    (if (or (typep host 'logical-host) *dont-backtranslate-generic-pathname*)
	;Don't EVER backtranslate logical selves 2.5.87 MBC
	(setf new-host host
	      new-device device
	      new-directory directory)
	(MULTIPLE-VALUE-SETQ (NEW-HOST NEW-DEVICE NEW-DIRECTORY)
	  (BACKTRANSLATE-HOST-DIRECTORY
	    HOST DEV1 DIRECTORY)))		;2.5.87
    (UNLESS NEW-HOST
      (SETQ NEW-HOST HOST))
    (OR NEW-DEVICE (EQUAL DEV1 DEV) (SETQ NEW-DEVICE DEV1))
    (COND
      ((AND (EQ NEW-TYPE :UNSPECIFIC) (NEQ HOST NEW-HOST))
       (SETQ NEW-TYPE (FUNCALL NEW-HOST :GENERIC-BASE-TYPE TYP))))
    (COND
      ((AND (NEQ DEV :UNSPECIFIC)
	    (eq (send host :system-type) :logical))
       (OR (EQUAL DEV "DSK")			;Temporary for when default devices start changing.
	   (EQUAL DEV (FUNCALL HOST :DEFAULT-DEVICE)) (NULL DEV)
	   (FERROR () "Unable to generate generic-pathname for pathname ~s" SELF))
       (SETQ NEW-DEVICE :UNSPECIFIC)))		;device of logical pathname always unspecific
    (SETQ PATH
	  (SEND SELF :NEW-PATHNAME
		:HOST
		;; Even if not actually changing these components,
		;; convert them to standard case and back.
		;; This is important on systems which allow case in pathnames
		;; but ignore case on lookup.
		NEW-HOST :DEVICE (OR NEW-DEVICE (SEND SELF :DEVICE)) :RAW-DIRECTORY	;4.3.87
		(OR NEW-DIRECTORY (SEND SELF :RAW-DIRECTORY)) :NAME (SEND SELF :NAME) :TYPE NEW-TYPE
		:VERSION :UNSPECIFIC))
    (IF (TYPEP NEW-HOST 'LOGICAL-HOST)
	(SEND (SAMPLE-PATHNAME HOST) :CHARACTER-BACK-TRANSLATIONS PATH)
	PATH)))
  

(DEFMETHOD (PATHNAME :SOURCE-PATHNAME) ()
  (GENERIC-PATHNAME-SOURCE-PATHNAME (SEND SELF :GENERIC-PATHNAME)))

(DEFUN GENERIC-PATHNAME-SOURCE-PATHNAME (PATHNAME)
  "Given the generic pathname PATHNAME, return a pathname for the source file.
We use the actual source file name as recorded, if possible."
  (LET ((QFASL-SOURCE (SEND PATHNAME :GET :QFASL-SOURCE-FILE-UNIQUE-ID))
	(LOADED-FILE (CAADAR (SEND PATHNAME :GET :FILE-ID-PACKAGE-ALIST))))
    (MULTIPLE-VALUE-BIND (CTYPE OTYPE)
	(COND
	  (QFASL-SOURCE
;;; This type of pathname list should no longer exist with genasys.  MBC 10.28.86 
;	    (IF (CONSP QFASL-SOURCE)		;dont bomb if list frobs somehow 
;						;left from cold load.
;		(SETQ QFASL-SOURCE (PATHNAME-FROM-COLD-LOAD-PATHLIST QFASL-SOURCE)))
	    (SEND QFASL-SOURCE :CANONICAL-TYPE))
	  ((AND LOADED-FILE
		(NOT (EQ (SEND LOADED-FILE :CANONICAL-TYPE) (SI:LOCAL-BINARY-FILE-TYPE))))
	   (SEND LOADED-FILE :CANONICAL-TYPE))
	  (T (SEND (SEND PATHNAME :HOST) :GENERIC-SOURCE-TYPE (SEND PATHNAME :CANONICAL-TYPE))))
      (SEND PATHNAME :NEW-PATHNAME :VERSION
	    :NEWEST
	    ;; The replacement of :UNSPECIFIC by :LISP
	    ;; is for files that were last compiled on ITS or FC.
	    :CANONICAL-TYPE (IF (EQ CTYPE :UNSPECIFIC)
				:LISP
				CTYPE)
	    :ORIGINAL-TYPE OTYPE))))
  

(DEFMETHOD (PATHNAME :TRANSLATED-PATHNAME) ()
  SELF)  

(DEFMETHOD (PATHNAME :BACK-TRANSLATED-PATHNAME) (PATHNAME)
  PATHNAME)  

;;; This is used to parse a string which may not have the host in it
;;; CAUTION:  This is patched in CHAOSNET-3-9.  MBC 6.17.87
(DEFMETHOD (PATHNAME :PARSE-TRUENAME) (STRING &optional HOST1)	;6.17.87
"Just like parse-namestring except if string specifies no type, 
the resulting pathname will have a type of :UNSPECIFIC."
  (let* ((with-respect-to (if host1 host1 host))
	 (truename (PARSE-PATHNAME STRING with-respect-to (send with-respect-to :sample-pathname))))
    (if (send truename :type) truename (send truename :new-type :unspecific))))


(DEFMETHOD (PATHNAME :SYSTEM-TYPE) ()
  (SEND HOST :SYSTEM-TYPE))  

(DEFMETHOD (PATHNAME :TYPES-FOR-CANONICAL-TYPE) (CANONICAL-TYPE)
  (MULTIPLE-VALUE-BIND (NIL TEM)
    (DECODE-CANONICAL-TYPE CANONICAL-TYPE (SEND HOST :SYSTEM-TYPE))
    TEM))  

(DEFMETHOD (PATHNAME :NEW-PREFERRED-CANONICAL-TYPE) (Input-Truename)	;6.16.87
  (if (member type '(:WILD NIL))		;Only change canonical type when desired...
      (let (input-canonical-type input-original-type)
	(multiple-value-setq (input-canonical-type input-original-type)
	  (send input-truename :canonical-type))
	(if (stringp input-canonical-type)
	    self
	    (if (member input-original-type 
		    (send self :TYPES-FOR-CANONICAL-TYPE input-canonical-type) :TEST #'equal)
	    (send self :new-canonical-type input-original-type)
	    (send self :new-canonical-type input-canonical-type))))
      self))

(DEFMETHOD (PATHNAME :OPEN-CANONICAL-DEFAULT-TYPE) (CANONICAL-TYPE &REST ARGS)
  (IF TYPE
    (LEXPR-SEND SELF :OPEN SELF ARGS)
    (LEXPR-SEND SELF :OPEN-CANONICAL-TYPE CANONICAL-TYPE SELF ARGS)))  

(DEFMETHOD (PATHNAME :OPEN-CANONICAL-TYPE) (CANONICAL-TYPE PRETRANSLATED-PATHNAME &REST ARGS &KEY &OPTIONAL (ERROR T) &ALLOW-OTHER-KEYS)
  (LET ((SURFACE-TYPES (SEND SELF :TYPES-FOR-CANONICAL-TYPE CANONICAL-TYPE)))
    (DO ((TYPES SURFACE-TYPES (CDR TYPES)))
	((NULL TYPES)
	 (LET ((CONDITION
		;; If no file found, signal with the preferred surface type.
		(MAKE-CONDITION 'FILE-NOT-FOUND "File not found for ~A."
				(SEND SELF :NEW-TYPE CANONICAL-TYPE) :OPEN)))
	   (IF ERROR
	     (SIGNAL-CONDITION CONDITION)
	     CONDITION)))
      (CONDITION-CASE (RESULT)
	 (LEXPR-SEND (SEND (OR PRETRANSLATED-PATHNAME SELF)
			   :NEW-PATHNAME :TYPE (CAR TYPES)) :OPEN
	    (SEND (OR PRETRANSLATED-PATHNAME SELF) :NEW-PATHNAME :TYPE
	       (IF (EQ TYPES SURFACE-TYPES)
		 CANONICAL-TYPE
		 (CAR TYPES)))
	    :ERROR T ARGS)
	 (FILE-NOT-FOUND NIL) (:NO-ERROR (RETURN RESULT))))))  

;;; Creation of pathnames, normal user interface.

;;These two methods are right for file systems where upper case is normally used.

(DEFMETHOD (PATHNAME :NEW-SUGGESTED-NAME) (NEW-NAME)
  (SEND SELF :NEW-PATHNAME :NAME (STRING-UPCASE NEW-NAME)))  

(DEFMETHOD (PATHNAME :NEW-SUGGESTED-DIRECTORY) (NEW-DIRECTORY)
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY (CONVERT-UPPER-CASE NEW-DIRECTORY)))	;4.13.87

(DEFMETHOD (PATHNAME :NEW-DEVICE) (NEW-DEVICE)
  (SEND SELF :NEW-PATHNAME :DEVICE NEW-DEVICE))  

(DEFMETHOD (PATHNAME :NEW-DIRECTORY) (NEW-DIRECTORY)
  (SEND SELF :NEW-PATHNAME :DIRECTORY NEW-DIRECTORY))  

(DEFMETHOD (PATHNAME :NEW-NAME) (NEW-NAME)
  (SEND SELF :NEW-PATHNAME :NAME NEW-NAME))  

(DEFMETHOD (PATHNAME :NEW-TYPE) (NEW-TYPE)
  (SEND SELF :NEW-PATHNAME :TYPE NEW-TYPE))  

(DEFMETHOD (PATHNAME :NEW-VERSION) (NEW-VERSION)
  (SEND SELF :NEW-PATHNAME :VERSION NEW-VERSION))  

(DEFMETHOD (PATHNAME :NEW-RAW-DEVICE) (NEW-DEVICE)
  (SEND SELF :NEW-PATHNAME :RAW-DEVICE NEW-DEVICE))  

(DEFMETHOD (PATHNAME :NEW-RAW-DIRECTORY) (NEW-DIRECTORY)
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY NEW-DIRECTORY))  

(DEFMETHOD (PATHNAME :NEW-RAW-NAME) (NEW-NAME)
  (SEND SELF :NEW-PATHNAME :RAW-NAME NEW-NAME))  

(DEFMETHOD (PATHNAME :NEW-RAW-TYPE) (NEW-TYPE)
  (SEND SELF :NEW-PATHNAME :RAW-TYPE NEW-TYPE))  

(DEFMETHOD (PATHNAME :NEW-CANONICAL-TYPE) (CANONICAL-TYPE &OPTIONAL ORIGINAL-TYPE)
  (SEND SELF ':NEW-PATHNAME ':CANONICAL-TYPE CANONICAL-TYPE ':ORIGINAL-TYPE ORIGINAL-TYPE))

;;; For NFS server.  3.8.88
(DEFMETHOD (PATHNAME :NEW-NAME-AND-TYPE) (SPEC)
  (multiple-value-bind (ignore ignore nam typ)
      (SEND (send (send SELF :host) :sample-pathname) :PARSE-NAMESTRING nil spec)
    (send self :new-pathname :type typ :name nam)))

;;; These exist for the sake of ITS
(DEFMETHOD (PATHNAME :NEW-TYPE-AND-VERSION) (NEW-TYPE NEW-VERSION)
  (SEND SELF :NEW-PATHNAME :TYPE NEW-TYPE :VERSION NEW-VERSION))  

(DEFMETHOD (PATHNAME :TYPE-AND-VERSION) ()
  (VALUES TYPE VERSION))  

(DEFMETHOD (PATHNAME :PATCH-FILE-PATHNAME) (NAM SAME-DIRECTORY-P PATOM TYP &REST ARGS)
  (LET ((PATOM (STRING-UPCASE (IF SAME-DIRECTORY-P
				  PATOM
				  NAM))))
    (CASE TYP
      (:SYSTEM-DIRECTORY
       (SEND SELF :NEW-PATHNAME :NAME PATOM
	     :TYPE :PATCH-DIRECTORY
	     :VERSION :NEWEST))
      
      (:VERSION-DIRECTORY
       (SEND SELF :NEW-PATHNAME :NAME (FORMAT () "~A-~D" PATOM (CAR ARGS))
	     :TYPE :PATCH-DIRECTORY
	     :VERSION :NEWEST))
      
      (:PATCH-FILE
       (SEND SELF :NEW-PATHNAME :NAME (FORMAT () "~A-~D-~D" PATOM (CAR ARGS) (CADR ARGS)) :TYPE
	     (CADDR ARGS) :VERSION :NEWEST)))))

(DEFMETHOD (PATHNAME :QUIET-HOMEDIR) (&OPTIONAL (USER USER-ID))
  (SEND (MAKE-PATHNAME :HOST HOST :DEVICE (SEND SELF :PRIMARY-DEVICE))
	;;; Default is for users to have their own directories.
     :NEW-SUGGESTED-DIRECTORY (COMPUTE-HOMEDIR-FROM-USER-ID USER HOST)))

(DEFMETHOD (PATHNAME :NEW-MAIL-PATHNAME) ()
  (SEND SELF :NEW-PATHNAME :NAME "MAIL" :TYPE "TEXT" :VERSION :NEWEST))  


;;; Make sure that a :NEW-PATHNAME which specifies a new host
;;; is processed by the flavor of pathname for that host.
;;; Make the end test check names for host, not if hosts are eq.

(DEFWRAPPER (PATHNAME :NEW-PATHNAME) (OPTIONS . BODY)
  `(LET ((NEW-HOST (GETF OPTIONS :HOST))
	 NEW-PATHNAME-HOST)
     
     (IF (AND NEW-HOST
	      (SETQ NEW-PATHNAME-HOST (GET-PATHNAME-HOST NEW-HOST))
	      (NOT (EQ (if (eq (send host :system-type) :logical)
			   (get-pathname-host HOST) host)
		       NEW-PATHNAME-HOST)))
	 
	 (LEXPR-SEND (SAMPLE-PATHNAME NEW-PATHNAME-HOST) :NEW-PATHNAME
			:STARTING-PATHNAME (OR (GET (LOCF OPTIONS) ':STARTING-PATHNAME) SELF)
			OPTIONS)
	 ,@BODY)))

;;; This is the fundamental way of altering some of the components of a pathname.
;;; Specify an alternating list of components and values.
;;; Components allowed are :HOST, :DEVICE, :DIRECTORY, :NAME, :TYPE and :VERSION;
;;; All the :NEW-x operations call this one (by default),
;;; and all the :PARSE-x operations (except :PARSE-NAMESTRING)
;;; are called only by MAKE-PATHNAME-1, which is called only from here.

;STARTING-PATHNAME is specified if we are doing the work on behalf of that pathname
;because it was told to change hosts, and we are a pathname of the correct new host.

(DEFMETHOD (PATHNAME :NEW-PATHNAME) (&REST OPTIONS &KEY &OPTIONAL STARTING-PATHNAME &ALLOW-OTHER-KEYS)
  (APPLY #'MAKE-PATHNAME-1 :STARTING-PATHNAME (OR STARTING-PATHNAME SELF) :PARSING-PATHNAME SELF
	 OPTIONS))  

;MAKE-PATHNAME is equivalent to this if the standard :NEW-PATHNAME method is in use.
;MAKE-PATHNAME could do this directly, except that would take away
;the ability for some pathname flavor to replace this entirely.

(DEFUN MAKE-PATHNAME-1 (&REST OPTIONS &KEY &ALLOW-OTHER-KEYS STARTING-PATHNAME
			(PARSING-PATHNAME STARTING-PATHNAME)     ;12-17-87 DAB Removed &optional - not needed.
			(HOST NIL HOST-P) (VERSION NIL VERSION-P)
			(ORIGINAL-TYPE NIL ORIGINAL-TYPE-P)
			&AUX DEVICE-P DIRECTORY-P NAME-P TYPE-P DEVICE DIRECTORY
  NAME TYPE CANONICAL-TYPE)
  (LOOP FOR (KEYWORD VALUE) ON OPTIONS BY 'CDDR DO
     (CASE KEYWORD
       (:NAME (UNLESS NAME-P
		(SETQ NAME VALUE
		      NAME-P T)))
       (:RAW-NAME (UNLESS NAME-P
		    (SETQ NAME VALUE
			  NAME-P :RAW)))
       (:DIRECTORY (UNLESS DIRECTORY-P
		     (SETQ DIRECTORY VALUE
			   DIRECTORY-P T)))
       (:RAW-DIRECTORY (UNLESS DIRECTORY-P
			 (SETQ DIRECTORY VALUE
			       DIRECTORY-P :RAW)))
       (:DEVICE (UNLESS DEVICE-P
		  (SETQ DEVICE VALUE
			DEVICE-P T)))
       (:RAW-DEVICE (UNLESS DEVICE-P
		      (SETQ DEVICE VALUE
			    DEVICE-P :RAW)))
       (:TYPE
	(UNLESS TYPE-P
	  (IF (AND (SYMBOLP VALUE) (NOT (MEMBER VALUE '(NIL :UNSPECIFIC) :TEST #'EQ)))
	    (SETQ CANONICAL-TYPE VALUE
		  TYPE-P :CANONICAL)
	    (SETQ TYPE VALUE
		  TYPE-P T))))
       (:RAW-TYPE (UNLESS TYPE-P
		    (SETQ TYPE VALUE
			  TYPE-P :RAW)))
       (:CANONICAL-TYPE (UNLESS TYPE-P
			  (SETQ CANONICAL-TYPE VALUE
				TYPE-P :CANONICAL)))
       ;; All keywords that do NOT require special decoding must go here.
       ;; All keywords that do NOT require special decoding must go here.
       ((:HOST :VERSION :STARTING-PATHNAME :PARSING-PATHNAME :ORIGINAL-TYPE :DEFAULTS NIL) NIL)
       (T (FERROR () "Unknown keyword ~S to MAKE-PATHNAME or :NEW-PATHNAME." KEYWORD))))
  (OR HOST-P (SETQ HOST (PATHNAME-HOST STARTING-PATHNAME)))
  (SETQ HOST (GET-PATHNAME-HOST HOST))
  ;; Turn a specified canonical type into a string (in standard case).
  (IF (EQ TYPE-P :CANONICAL)
    (MULTIPLE-VALUE-BIND (PREFERRED ALL)
      (DECODE-CANONICAL-TYPE CANONICAL-TYPE (SEND HOST :SYSTEM-TYPE))
      (SETQ TYPE
	      (LET ((case-sensitive (send host :case-sensitive-when-hashing)))	;3.18.87
		(UNLESS ORIGINAL-TYPE-P
		  (SETQ ORIGINAL-TYPE (PATHNAME-TYPE STARTING-PATHNAME)))
		(IF (MEMBER (convert-solid-case ORIGINAL-TYPE) ALL
			    :TEST (if case-sensitive #'STRING= #'STRING-EQUAL))
		    (convert-solid-case ORIGINAL-TYPE)	;3.18.87
		    PREFERRED)))))
  (IF (EQ (PATHNAME-HOST STARTING-PATHNAME) HOST)
    (PROGN
      (OR DEVICE-P (SETQ DEVICE (PATHNAME-RAW-DEVICE STARTING-PATHNAME)
			 DEVICE-P :RAW))
      (OR DIRECTORY-P
	 (SETQ DIRECTORY (PATHNAME-RAW-DIRECTORY STARTING-PATHNAME)
	       DIRECTORY-P :RAW))
      (OR NAME-P (SETQ NAME (PATHNAME-RAW-NAME STARTING-PATHNAME)
		       NAME-P :RAW))
      (OR TYPE-P (SETQ TYPE (PATHNAME-RAW-TYPE STARTING-PATHNAME)
		       TYPE-P :RAW)))
    ;; Hosts don't match; must convert to standard syntax and reparse.
    (PROGN
      (OR DEVICE-P (SETQ DEVICE (PATHNAME-DEVICE STARTING-PATHNAME)))
      (OR DIRECTORY-P (SETQ DIRECTORY (PATHNAME-DIRECTORY STARTING-PATHNAME)))
      (OR NAME-P (SETQ NAME (PATHNAME-NAME STARTING-PATHNAME)))
      (OR TYPE-P (SETQ TYPE (PATHNAME-TYPE STARTING-PATHNAME)))))
  (OR VERSION-P (SETQ VERSION (PATHNAME-RAW-VERSION STARTING-PATHNAME)))
  ;; The new fields are parsed only once to save time, consing, and possible errors
  ;; due to incompatible fields in different types of pathnames.
    (AND (NEQ DEVICE-P :RAW) (SETQ DEVICE (FUNCALL PARSING-PATHNAME :PARSE-DEVICE-SPEC DEVICE)))
    (AND (NEQ DIRECTORY-P :RAW)
       (SETQ DIRECTORY (FUNCALL PARSING-PATHNAME :PARSE-DIRECTORY-SPEC DIRECTORY)))
    (AND (NEQ NAME-P :RAW) (SETQ NAME (FUNCALL PARSING-PATHNAME :PARSE-NAME-SPEC NAME)))
    (AND (NEQ TYPE-P :RAW) (SETQ TYPE (FUNCALL PARSING-PATHNAME :PARSE-TYPE-SPEC TYPE)))
    (SETQ VERSION (FUNCALL PARSING-PATHNAME :PARSE-VERSION-SPEC VERSION))
    (if (stringp directory) (setf directory (list directory)))	;3.08.88
  (MAKE-PATHNAME-INTERNAL (SEND STARTING-PATHNAME :RAW-QUOTED-STRING) HOST DEVICE DIRECTORY NAME
			  TYPE VERSION))


;;; Default is to leave the string alone

(DEFMETHOD (PATHNAME :PARSE-COMPONENT-SPEC) (SPEC)
  SPEC)  

;;; These operations should in general convert an interchange component	to a raw one
;;; and also turn any invalid component into something valid.

(DEFMETHOD (PATHNAME :PARSE-DEVICE-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (SEND SELF :PARSE-COMPONENT-SPEC SPEC))
    ((AND (CONSP SPEC) (STRINGP (CAR SPEC)) (NULL (CDR SPEC)))
     (SEND SELF :PARSE-COMPONENT-SPEC (CAR SPEC)))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD) :TEST #'EQ) SPEC)
    (T (PATHNAME-DEVICE (QUIET-USER-HOMEDIR HOST)))))  

(DEFMETHOD (PATHNAME :PARSE-DIRECTORY-SPEC) (SPEC)
  (COND
    ((OR (STRINGP SPEC) (CONSP SPEC)) (SEND SELF :PARSE-COMPONENT-SPEC SPEC))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD :root) :TEST #'EQ) SPEC)  ;03-17-88 DAB Added :root
    (T (PATHNAME-DIRECTORY (QUIET-USER-HOMEDIR HOST)))))  

(DEFMETHOD (PATHNAME :PARSE-NAME-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (SEND SELF :PARSE-COMPONENT-SPEC SPEC))
    ((AND (CONSP SPEC) (STRINGP (CAR SPEC)) (NULL (CDR SPEC)))
     (SEND SELF :PARSE-COMPONENT-SPEC (CAR SPEC)))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD) :TEST #'EQ) SPEC)
    (T "FOO")))

(DEFMETHOD (PATHNAME :PARSE-TYPE-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (SEND SELF :PARSE-COMPONENT-SPEC SPEC))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD) :TEST #'EQ) SPEC)
    ((KEYWORDP spec) (string (DECODE-CANONICAL-TYPE spec (SEND HOST :SYSTEM-TYPE))))	;allow unknown :TYPEs
    (T (DECODE-CANONICAL-TYPE :LISP (SEND HOST :SYSTEM-TYPE)))))

;; Since there are no "interchange" versions,
;; this is only to convert invalid ones to valid ones.


(DEFMETHOD (PATHNAME :PARSE-VERSION-SPEC) (SPEC)
  (IF (OR (AND (FIXNUMP SPEC) (> SPEC 0.))
      (MEMBER SPEC '(NIL :UNSPECIFIC :WILD :NEWEST :OLDEST) :TEST #'EQ))   ;12-17-87 DAB Added :oldest
    SPEC
    :NEWEST))  

(DEFMETHOD (PATHNAME :FASD-FORM) ()
  `(MAKE-FASLOAD-PATHNAME ',(FUNCALL HOST :NAME-AS-FILE-COMPUTER) ',DEVICE ',DIRECTORY ',NAME
			  ',TYPE ',VERSION))  

(DEFMETHOD (PATHNAME :EQUAL) (OTHER-PATHNAME)
  (EQ OTHER-PATHNAME SELF))  

(DEFMETHOD (PATHNAME :SORT-COMPONENTS) ()
  (VALUES (PATHNAME-HOST SELF) (PATHNAME-DEVICE SELF) (PATHNAME-DIRECTORY SELF)
	  (PATHNAME-NAME SELF) (PATHNAME-TYPE SELF) (PATHNAME-VERSION SELF)))  

;Redefine this too, to use your standard components.

(DEFMETHOD (PATHNAME :SORT-LESSP) (OTHER-PATHNAME &AUX TEM)
  (MULTIPLE-VALUE-BIND (OHOST ODEVICE ODIRECTORY ONAME OTYPE OVERSION)
    (FUNCALL OTHER-PATHNAME :SORT-COMPONENTS)
    (AND (ZEROP (SETQ TEM (PATHNAME-COMPONENT-COMPARE (PATHNAME-HOST SELF) OHOST)))
       (ZEROP (SETQ TEM (PATHNAME-COMPONENT-COMPARE (PATHNAME-DEVICE SELF) ODEVICE)))
       (ZEROP (SETQ TEM (PATHNAME-COMPONENT-COMPARE (PATHNAME-DIRECTORY SELF) ODIRECTORY)))
       (ZEROP (SETQ TEM (PATHNAME-COMPONENT-COMPARE (PATHNAME-NAME SELF) ONAME)))
       (ZEROP (SETQ TEM (PATHNAME-COMPONENT-COMPARE (PATHNAME-TYPE SELF) OTYPE)))
       (SETQ TEM (PATHNAME-COMPONENT-COMPARE (PATHNAME-VERSION SELF) OVERSION)))
    (MINUSP TEM))) 


(DEFUN PATHNAME-COMPONENT-COMPARE (X Y)
  (COND
    ((EQUAL X Y) 0.)
    ((OR (SYMBOLP X) (SYMBOLP Y));rla 11/15/85 (use keyword compare if EITHER is a symbol)
     (PATHNAME-KEYWORD-COMPARE X Y))
    ((STRINGP X) (IF (STRINGP Y)
		   (STRING-COMPARE X Y)
		   -1.))
    ((STRINGP Y) 1.)
    ((NUMBERP X) (IF (NUMBERP Y)
		   (- X Y)
		   -1.))
    ((NUMBERP Y) 1.)
    ((ALPHALESSP X Y) -1.)
    (T 1.)))  

(DEFUN PATHNAME-KEYWORD-COMPARE (X Y)
 ;; rla 11/15/85 
 ;; Now implements these rules:
 ;;  1. :NEWEST is always greater
 ;;  2. other symbols (nil :Unspecific) are always less than non-symbols (strings, numbers, lists)
 ;;  3. two symbols are compared using string-compare
  (COND
    ((EQ X :NEWEST) (COND
		      ((EQ Y :NEWEST) 0.)
		      (T 1.)))
    ((EQ Y :NEWEST) (COND
		      ((EQ X :NEWEST) 0.)
		      (T -1.)))
    ((AND (SYMBOLP X) (SYMBOLP Y)) (STRING-COMPARE X Y))
    ((SYMBOLP X) -1.);x is a symbol, y is not
    (T 1.)))  					;y is a symbol, x is not

(DEFMETHOD (PATHNAME :NEW-DEFAULT-PATHNAME) (&REST OPTIONS)
  (APPLY SELF :NEW-PATHNAME OPTIONS))  


(DEFMETHOD (PATHNAME :DEFAULT-NAMESTRING) (NAMESTRING &OPTIONAL (DEFAULT-TYPE ':UNSPECIFIC) (DEFAULT-VERSION ':NEWEST))
  (MERGE-PATHNAME-DEFAULTS NAMESTRING SELF DEFAULT-TYPE DEFAULT-VERSION))  

(DEFMETHOD (PATHNAME :MERGE-DEVICE) ()		;MAC hack.  The mac version of this does something....
  self)						;...until volume is a real DEVICE not a directory node.  2.10.88

;;; Operations that refer to the file.

;;; The default is not to have completion at all
(DEFMETHOD (PATHNAME :COMPLETE-STRING) (STRING IGNORE)
  (VALUES (STRING-APPEND (FUNCALL HOST :NAME-AS-FILE-COMPUTER) ": " STRING) ()))  

(DEFMETHOD (PATHNAME :MULTIPLE-FILE-PLISTS) (FILES OPTIONS &AUX (CHARACTERS T))
  (LOOP FOR (IND OPT) ON OPTIONS BY 'CDDR DO
     (CASE IND
       (:CHARACTERS (SETQ CHARACTERS OPT))
       (OTHERWISE (FERROR () "~S is not a known MULTIPLE-FILE-PLISTS option" IND))))
  (LOOP FOR FILE IN FILES AS STREAM = (OPEN FILE :DIRECTION () :ERROR () :CHARACTERS CHARACTERS)
     COLLECT
     (CONS FILE
	   (AND (NOT (ERRORP STREAM))
	      (LET* ((LIST (FUNCALL STREAM :PLIST))
		     (PLIST (LOCF LIST)))
		(OR (GET PLIST :TRUENAME)
		   (SETF (GET PLIST :TRUENAME) (FUNCALL STREAM :TRUENAME)))
		LIST)))))  

(DEFMETHOD (PATHNAME :UNDELETE) (&OPTIONAL (ERROR-P T))
  (CHANGE-FILE-PROPERTIES SELF ERROR-P :DELETED ()))  

(DEFMETHOD (PATHNAME :TRUENAME) (&OPTIONAL (ERROR-P T))
  (WITH-OPEN-FILE (STREAM SELF :ERROR ERROR-P :DIRECTION ())
    (IF (ERRORP STREAM)
      STREAM
      (FUNCALL STREAM :TRUENAME))))  

;;; This isn't implemented as a separate file subprotocol, just use the directory
;; This is now part of file-access-mixin. *BJ*
;(DEFMETHOD (PATHNAME :PROPERTIES) (&OPTIONAL (ERROR-P T))
;  (FILE-OPERATION-RETRY
;   (LET ((DIR (SEND SELF :DIRECTORY-LIST (IF ERROR-P
;					   '(:DELETED)
;					   '(:NOERROR :DELETED)))))
;     (COND
;       ((CONSP DIR)
;	(IF (CADR DIR)
;	  (VALUES (CADR DIR) (GET (CAR DIR) :SETTABLE-PROPERTIES))
;	  ;; It is possible for a nonexistent file to give no error
;	  ;; but just return an empty directory.
;	  (FILE-PROCESS-ERROR-NEW " ERROR FNF F File not found" SELF () (NOT ERROR-P)
;				  :PROPERTIES)))
;       (T DIR)))))


;;; This is a pathname that corresponds to a real (as opposed to logical) host
(DEFFLAVOR HOST-PATHNAME
   ((STRING-FOR-EDITOR NIL) (STRING-FOR-DIRED NIL) (STRING-FOR-DIRECTORY NIL)) (PATHNAME))  

(DEFMETHOD (HOST-PATHNAME :INIT-FILE) (PROGRAM-NAME)
  (SEND SELF :NEW-PATHNAME :NAME (STRING-APPEND PROGRAM-NAME "-INIT") :TYPE () :VERSION :NEWEST))

;(DEFWRAPPER (HOST-PATHNAME :STRING-FOR-WHOLINE) (IGNORE . BODY)
;  `(CACHE-IN-VARIABLE STRING-FOR-WHOLINE . ,BODY))

(DEFWRAPPER (HOST-PATHNAME :STRING-FOR-EDITOR) (IGNORE . BODY)
   `(CACHE-IN-VARIABLE STRING-FOR-EDITOR ,@BODY))  

(DEFWRAPPER (HOST-PATHNAME :STRING-FOR-DIRED) (IGNORE . BODY)
   `(CACHE-IN-VARIABLE STRING-FOR-DIRED ,@BODY))

(DEFWRAPPER (HOST-PATHNAME :STRING-FOR-DIRECTORY) (IGNORE . BODY)
   `(CACHE-IN-VARIABLE STRING-FOR-DIRECTORY ,@BODY))  

;;; This is a pathname connected by some remote medium.  It might even be a local file
;;; system if that interfaces through a :STRING-FOR-HOST.


(DEFFLAVOR REMOTE-PATHNAME ((STRING-FOR-HOST NIL);Caches
			    ) (HOST-PATHNAME)
   (:REQUIRED-METHODS :STRING-FOR-HOST))  

(DEFMETHOD (PATHNAME :SHORT-STRING-FOR-PRINTING) ()
  (STRING-APPEND (FUNCALL HOST :SHORT-NAME) ": " (SEND SELF :STRING-FOR-HOST)))

(DEFMETHOD (PATHNAME :STRING-FOR-PRINTING) ()
  (STRING-APPEND (FUNCALL HOST :NAME-AS-FILE-COMPUTER) ": " (SEND SELF :STRING-FOR-HOST)))

(DEFMETHOD (PATHNAME :STRING-FOR-EDITOR) ()
  (STRING-APPEND (SEND SELF :STRING-FOR-HOST) #\SPACE (FUNCALL HOST :SHORT-NAME) #\:))

(DEFWRAPPER (REMOTE-PATHNAME :STRING-FOR-HOST) (IGNORE . BODY)
   `(CACHE-IN-VARIABLE STRING-FOR-HOST ,@BODY))  

(DEFMETHOD (REMOTE-PATHNAME :STRING-FOR-MINI) ()
  (SEND SELF :STRING-FOR-HOST))  

(DEFMETHOD (REMOTE-PATHNAME :STRING-FOR-DIRECTORY) ()
  (SEND SELF :STRING-FOR-HOST))


;;; New :STRING-FOR-NFS support.  Nfs file servers can use this.  2.03.88 MBC
(DEFUN NFS-FILENAME (NAME TYPE &AUX (NEW-TYPE TYPE))
  (when (consp name) (setf name (car name)))
  (select name
    (:UNSPECIFIC (SETQ NAME ""))
    (:WILD (IF (MEMBER TYPE '(:WILD :UNSPECIFIC NIL) :TEST #'EQ)
	       "*"				
	       (STRING-APPEND "*." NEW-TYPE)))
    (T
     (IF (AND (NULL NAME) (MEMBER TYPE '(NIL :UNSPECIFIC) :TEST #'EQ))
	 ""
	 (PROGN
	   (OR NAME (SETQ NAME ""))
	   (COND
	     ((EQ TYPE :WILD)
	      (if (zerop (length name)) "*"
		  (IF (AND
			(CHAR= #\* (AREF NAME 0.))
			(EVERY #'(LAMBDA (X)
				   (CHAR= #\. X))
			       (NSUBSTRING NAME 1.)))
		      (STRING-APPEND NAME "*")
		      (STRING-APPEND NAME ".*"))))
	     ((eq TYPE :UNSPECIFIC) NAME)
	     ((eq TYPE NIL) NAME)
	     (T (STRING-APPEND NAME "." NEW-TYPE))))))))

(DEFUN NFS-DIRECTORY-COMPONENT (STRING)
  (CASE STRING
    (:WILD "*")
    ((:UP :RELATIVE) "")			;delimiter inserted by caller
    (OTHERWISE (STRING STRING))))

(DEFUN NFS-DIRECTORY-STRING ()
  (DECLARE (:SELF-FLAVOR REMOTE-PATHNAME))
  (let ((DELIM #\/)
	(DIRECT DIRECTORY))
    (cond ((MEMBER DIRECT '(NIL :UNSPECIFIC :ROOT) :TEST #'EQ) "")
	  ((not (consp direct)) direct)
	  ((NULL (CDR DIRECT))
	   (STRING-APPEND (NFS-DIRECTORY-COMPONENT (CAR DIRECT)) DELIM))
	  ((stringp DIRECT)
	   (NFS-DIRECTORY-COMPONENT DIRECT))
	  ((CONSP DIRECT)
	   (LOOP FOR SUBDIR IN DIRECT
		 WITH STRING = (MAKE-ARRAY 20. :ELEMENT-TYPE 'STRING-CHAR :LEADER-LIST '(0.))
		 DO
		 (SETQ STRING (STRING-NCONC STRING (NFS-DIRECTORY-COMPONENT SUBDIR)))
		 (VECTOR-PUSH-EXTEND DELIM STRING)
		 FINALLY (RETURN STRING))))))

(DEFMETHOD (REMOTE-PATHNAME :STRING-FOR-NFS) ()
  (FORMAT () "/~@[~A/~]~@[~A~]~A"	 
	  (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ) NIL DEVICE)	;1.27.88 MBC
	  (NFS-DIRECTORY-STRING)
	  (NFS-FILENAME NAME TYPE)))  


;; This method is used by the host code when a user wants to reset the print name of a host.
(defmethod (REMOTE-PATHNAME :RESET-PRINT-NAMES) ()
  (setf short-string-for-printing nil)
  (setf fs:string-for-printing nil)
  (setf fs:string-for-host nil)
  (setf fs:string-for-dired nil)
  (setf fs:string-for-editor nil)
  (setf fs:string-for-directory nil))

;;;
;;;  Clean up Non-Garbage-Garbage in pathname area before FULL-GC.
;;;  NGG was generated during pathname parsing.
;;;  This is done by clearing the hash table, then combing though 
;;;  PATHNAME-AREAS's regions, re-interning any pathnames found there.

;;;  Add this method to clear out print strings from the band
;;;
(defmethod (PATHNAME :RESET-PRINT-NAMES) ()
  (setf fs:short-string-for-printing nil)
  (setf fs:string-for-printing nil))

;;;
;;;  This method will change the strings in the pathname to use only a 
;;;  single copy of the string
;;;
(defmethod (pathname :collapse-strings) ()
  (declare (special fs:*string-hash-table*))
  (macrolet ((replace-component (component)
		 `(when (stringp ,component)   
		      (setf ,component
			  (find-or-add-hash ,component fs:*string-hash-table*)))))
    ;; (replace-component fs:version) ; version is always a fixnum or symbol
    (replace-component fs:type)
    (replace-component fs:name)
    (replace-component fs:device)
    (unless (symbolp fs:directory)
      (setf fs:directory (find-or-add-hash (list-find-or-add-hash fs:directory)
					   fs:*string-hash-table*)))
    (values)
    ))

(defun find-or-add-hash (value hash)
  "This function looks to see if VALUE is in HASH.  If it is, it returns the value in 
     the hash table.  If VALUE is not in HASH, it is added."
  (let ((v (gethash value hash)))
    (if v
	v
	(puthash value value hash))))

(defun list-find-or-add-hash (object)
  ;; If object is a list, collapse any elements which are strings or lists.
  (declare (special fs:*string-hash-table*))
  (do ((tail object (cdr tail)))
      ((atom tail))
    (unless (or (symbolp (car tail))
		(fixnump (car tail)))
      (setf (car tail)
	    (find-or-add-hash (list-find-or-add-hash (car tail))
			      fs:*string-hash-table*))))
  object)


(DEFUN restore-pathnames-in-region (region)
  "This function goes through the REGION and puts any pathname objects
   found into the pathname hash table."
  (declare (special fs:*pathname-hash-table*))
  (flet ((obj-fn (object &rest ignore)
	    (when (pathnamep object)
	      (let ((key (send object :get :hash-key)))
		(unless (null key)
		  (puthash key object fs:*pathname-hash-table*)
		  (send object :remprop :hash-key))))))
    (sys::map-objects-in-region region :analysis-function #'obj-fn))
  t)

(defun restore-pathnames-after-gc ()
  "Restores all the pathnames in the pathname area to the pathname hash table."
  ;; The table ought to be empty now; warn if not.
  (unless (zerop (hash-table-count fs:*pathname-hash-table*))
    (format t "~&Warning: the following pathnames which were created during FULL-GC may lose
their properties or may have two non-EQ instances:")
    (maphash #'(lambda (key value)
		 (declare (ignore key))
		 (print value))
	     fs:*pathname-hash-table*)
    (terpri))
  (let ((gc (si:gc-active-p)))
    (when gc (GC-OFF))
    (DO ((region (AREF #'area-region-list fs:pathname-area)
		 (AREF #'region-list-thread region))
	 (object-list nil))
	((MINUSP region)
	 (RETURN object-list))
      (restore-pathnames-in-region region))
    (when gc (gc-on))))

(defun clear-pathnames-before-gc ()
  ;; Clear out pathname hash table before full GC.
  ;; Also merges duplicate strings and lists.
  (let* ((count (hash-table-count fs:*pathname-hash-table*))
	 (fs:*string-hash-table* (make-hash-table :test #'equal
						  :size (truncate count 2)
						  :rehash-size 1.5s0)))
    (declare (special fs:*string-hash-table*))
    (maphash #'(lambda (key value)	   ;remember the keys
		 (send value :putprop (list-find-or-add-hash key) :hash-key)
		 (send value :collapse-strings)
		 (send value :reset-print-names)
		 )
	     fs:*pathname-hash-table*)
    ;; re-make the hash table
    (PATHNAME-INITIALIZE (max 1200. (truncate count (hash-table-rehash-threshold
						      fs:*pathname-hash-table*))))
    )
  (values))


(DEFUN PATHNAME-PASS-THROUGH-SPEC (IGNORE SPEC)
  (SEND SELF :PARSE-COMPONENT-SPEC SPEC))  

(DEFFLAVOR PATHNAME-NORMALLY-LOWERCASE-MIXIN () () (:REQUIRED-FLAVORS PATHNAME))  

(DEFUN CONVERT-SOLID-CASE (OBJECT)
  (COND
    ((STRINGP OBJECT)
     (LET (SOME-LC
	   SOME-UC)
       (DOTIMES (I (LENGTH OBJECT))
	 (LET ((CH (AREF OBJECT I)))
	   (IF (<= #\A CH #\Z)
	     (SETQ SOME-UC T))
	   (IF (<= #\a CH #\z)
	     (SETQ SOME-LC T))))
       (IF (NEQ SOME-UC SOME-LC)
	 (IF SOME-UC
	   (STRING-DOWNCASE OBJECT)
	   (STRING-UPCASE OBJECT))
	 OBJECT)))
    ((CONSP OBJECT) (MAPCAR 'CONVERT-SOLID-CASE OBJECT))
    (T OBJECT)))  


(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :TRANSLATION-CASE-CONVERTER) ()
  'CONVERT-SOLID-CASE)  

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :PARSE-COMPONENT-SPEC) (SPEC)
  (CONVERT-SOLID-CASE SPEC))  

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :NAME) ()
  (CONVERT-SOLID-CASE NAME))  

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :DEVICE) ()
  (CONVERT-SOLID-CASE DEVICE))  

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :DIRECTORY) ()
  (CONVERT-SOLID-CASE DIRECTORY))  

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :TYPE) ()
  (CONVERT-SOLID-CASE TYPE))  

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :NEW-DIRECTORY) (SPEC)
  (SEND SELF :NEW-RAW-DIRECTORY (CONVERT-SOLID-CASE SPEC)))  

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :NEW-DEVICE) (SPEC)
  (SEND SELF :NEW-RAW-DEVICE (CONVERT-SOLID-CASE SPEC)))  

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :NEW-NAME) (SPEC)
  (SEND SELF :NEW-RAW-NAME (CONVERT-SOLID-CASE SPEC)))

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :NEW-CANONICAL-TYPE) (CANONICAL-TYPE &OPTIONAL ORIGINAL-TYPE)
  (SEND SELF ':NEW-PATHNAME ':CANONICAL-TYPE CANONICAL-TYPE ':ORIGINAL-TYPE ORIGINAL-TYPE))	;3.18.87

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :NEW-TYPE) (SPEC)
  (IF (SYMBOLP SPEC)
    (SEND SELF :NEW-RAW-TYPE
       (CONVERT-SOLID-CASE (DECODE-CANONICAL-TYPE SPEC (SEND SELF :SYSTEM-TYPE))))
    (SEND SELF :NEW-RAW-TYPE (CONVERT-SOLID-CASE SPEC)))) 
 

;; These two are right for systems where mixed case is normally used.


(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :NEW-SUGGESTED-NAME) (NEW-NAME)
  (SEND SELF :NEW-PATHNAME :NAME NEW-NAME))  


(DEFUN CONVERT-SOLID-CASE-TO-LOWER (OBJECT)	;4.13.87
  (COND
    ((STRINGP OBJECT)
     (LET (SOME-LC
	   SOME-UC)
       (DOTIMES (I (LENGTH OBJECT))
	 (LET ((CH (AREF OBJECT I)))
	   (IF (<= #\A CH #\Z)
	     (SETQ SOME-UC T))
	   (IF (<= #\a CH #\z)
	     (SETQ SOME-LC T))))
       (IF (NEQ SOME-UC SOME-LC)
	   (STRING-DOWNCASE OBJECT)
	 OBJECT)))
    ((CONSP OBJECT) (MAPCAR 'CONVERT-SOLID-CASE-TO-LOWER OBJECT))
    (T OBJECT)))

(DEFMETHOD (PATHNAME-NORMALLY-LOWERCASE-MIXIN :NEW-SUGGESTED-DIRECTORY) (NEW-DIRECTORY)	;4.13.87
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY (CONVERT-SOLID-CASE-TO-LOWER NEW-DIRECTORY)))


(DEFMETHOD (PATHNAME :CASE-COMPONENT) (SPEC)
  SPEC)  

(DEFFLAVOR RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN () () (:REQUIRED-FLAVORS PATHNAME))  

(DEFMETHOD (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN :CASE-COMPONENT) (SPEC)
  (CONVERT-UPPER-CASE SPEC))

(DEFMETHOD (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN :AROUND :PARSE-COMPONENT-SPEC) (CONT MT &REST ORG-ARG-LIST)
  (CONVERT-UPPER-CASE (LEXPR-FUNCALL-WITH-MAPPING-TABLE CONT MT ORG-ARG-LIST)))  

(DEFMETHOD (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN :AROUND :PARSE-DEVICE-SPEC) (CONT MT &REST ORG-ARG-LIST)
  (CONVERT-UPPER-CASE (LEXPR-FUNCALL-WITH-MAPPING-TABLE CONT MT ORG-ARG-LIST)))  

(DEFMETHOD (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN :AROUND :PARSE-DIRECTORY-SPEC) (CONT MT &REST ORG-ARG-LIST)
  (CONVERT-UPPER-CASE (LEXPR-FUNCALL-WITH-MAPPING-TABLE CONT MT ORG-ARG-LIST)))  

(DEFMETHOD (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN :AROUND :PARSE-NAME-SPEC) (CONT MT &REST ORG-ARG-LIST)
  (CONVERT-UPPER-CASE (LEXPR-FUNCALL-WITH-MAPPING-TABLE CONT MT ORG-ARG-LIST)))  

(DEFMETHOD (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN :AROUND :PARSE-TYPE-SPEC) (CONT MT &REST ORG-ARG-LIST)
  (CONVERT-UPPER-CASE (LEXPR-FUNCALL-WITH-MAPPING-TABLE CONT MT ORG-ARG-LIST)))  



(DEFFLAVOR PATHNAME-ALWAYS-UPPER-CASE-MIXIN () () (:REQUIRED-FLAVORS PATHNAME))  

(DEFUN CONVERT-UPPER-CASE (OBJECT)
  (COND
    ((STRINGP OBJECT) (STRING-UPCASE OBJECT))
    ((CONSP OBJECT) (MAPCAR 'CONVERT-UPPER-CASE OBJECT))
    (T OBJECT)))  

(DEFMETHOD (PATHNAME-ALWAYS-UPPER-CASE-MIXIN :TRANSLATION-CASE-CONVERTER) ()
  'CONVERT-UPPER-CASE)  

(DEFMETHOD (PATHNAME-ALWAYS-UPPER-CASE-MIXIN :NAME) ()
  (CONVERT-UPPER-CASE NAME))  

(DEFMETHOD (PATHNAME-ALWAYS-UPPER-CASE-MIXIN :DEVICE) ()
  (CONVERT-UPPER-CASE DEVICE))  

(DEFMETHOD (PATHNAME-ALWAYS-UPPER-CASE-MIXIN :DIRECTORY) ()
  (CONVERT-UPPER-CASE DIRECTORY))

(DEFMETHOD (PATHNAME-ALWAYS-UPPER-CASE-MIXIN :TYPE) ()
  (CONVERT-UPPER-CASE TYPE))  

;; These two are right for systems where mixed case is normally used.


(DEFMETHOD (PATHNAME-ALWAYS-UPPER-CASE-MIXIN :NEW-SUGGESTED-NAME) (NEW-NAME)
  (SEND SELF :NEW-PATHNAME :NAME NEW-NAME))  


(DEFMETHOD (PATHNAME-ALWAYS-UPPER-CASE-MIXIN :NEW-SUGGESTED-DIRECTORY) (NEW-DIRECTORY)	;4.13.87
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY (CONVERT-UPPER-CASE NEW-DIRECTORY)))

(DEFMETHOD (PATHNAME :DEFAULT :DEFAULT-DIRECTORY-STREAM-PARSER) ()
  #'READ-FROM-SUBSTRING) 


(DEFFLAVOR DIRECTORY-STREAM-MIXIN
	   ()
	   ()
  (:REQUIRED-FLAVORS PATHNAME)) 


(DEFMETHOD (DIRECTORY-STREAM-MIXIN :DIRECTORY-LIST) (OPTIONS &AUX DIR-LIST)
  (WITH-OPEN-STREAM (STREAM (SEND SELF :DIRECTORY-STREAM
				  (REMOVE :SORTED (THE LIST OPTIONS) :TEST #'EQ)))
    (IF (ERRORP STREAM)
	STREAM
	(PROGN
	  (SETQ DIR-LIST
		(LET ((PATHNAME (SEND STREAM :PATHNAME)))
		  (LOOP AS ENTRY = (SEND PATHNAME :READ-DIRECTORY-STREAM-ENTRY STREAM)
			UNTIL (NULL ENTRY)
			COLLECTING ENTRY)))
	  (IF (MEMBER :SORTED OPTIONS :TEST #'EQ)
	      (LET ((NULL-ELEM (ASSOC () DIR-LIST :TEST #'EQ)))
		(AND NULL-ELEM (SETQ DIR-LIST (DELETE NULL-ELEM (THE LIST DIR-LIST) :TEST #'EQ)))
		(SETQ DIR-LIST (SORT DIR-LIST #'PATHNAME-LESSP :KEY #'CAR))
		(AND NULL-ELEM (PUSH NULL-ELEM DIR-LIST))))
	  DIR-LIST)))) 


(DEFMETHOD (DIRECTORY-STREAM-MIXIN :DIRECTORY-LIST-STREAM) (&OPTIONAL OPTIONS)
  (LET ((STREAM (SEND SELF :DIRECTORY-STREAM OPTIONS)))
    (IF (ERRORP STREAM)
      STREAM
      (LET-CLOSED ((INTERNAL-STREAM STREAM)
		   (DEFAULTING-PATHNAME (SEND STREAM :PATHNAME)))
	 #'DIRECTORY-STREAM-DIRECTORY-LIST-STREAM)))) 

(DEFUN  DIRECTORY-STREAM-DIRECTORY-LIST-STREAM (OP &REST ARGS)
  (DECLARE (SPECIAL INTERNAL-STREAM DEFAULTING-PATHNAME))
  (CASE OP
    (:ENTRY (SEND DEFAULTING-PATHNAME :READ-DIRECTORY-STREAM-ENTRY INTERNAL-STREAM))
    (:CLOSE (apply INTERNAL-STREAM :CLOSE args))))

;(LOCALLY (DECLARE (SPECIAL INTERNAL-STREAM DEFAULTING-PATHNAME))
;   (DEFSELECT DIRECTORY-STREAM-DIRECTORY-LIST-STREAM
;     (:ENTRY NIL (SEND DEFAULTING-PATHNAME :READ-DIRECTORY-STREAM-ENTRY INTERNAL-STREAM))
;     (:CLOSE (&OPTIONAL MODE) (SEND INTERNAL-STREAM :CLOSE MODE)))) 

;This is a pathname operation so that pathname flavors can define their own
;way of parsing an entry from the directory stream.

(DEFMETHOD (DIRECTORY-STREAM-MIXIN :READ-DIRECTORY-STREAM-ENTRY) (STREAM)
  (READ-DIRECTORY-STREAM-ENTRY STREAM SELF))



;;;
;;; Pathname error and it continuation
;;;
(DEFFLAVOR PATHNAME-ERROR () (FERROR))  

(DEFMETHOD (PATHNAME-ERROR :CASE :PROCEED-ASKING-USER :NEW-PATHNAME) (CONTINUATION READ-OBJECT-FUNCTION)
  (FUNCALL CONTINUATION :NEW-PATHNAME
	   (FUNCALL READ-OBJECT-FUNCTION '(:PATHNAME :DEFAULTS (NIL))
		    "Pathname to use instead: ")))  

(DEFSIGNAL PATHNAME-PARSE-ERROR (PATHNAME-ERROR PATHNAME-PARSE-ERROR)
   (NOT-USED ALSO-NOT-USED PARSE-END-INDEX)
   "Any error that makes it impossible to parse a string into a pathname.")  

(DEFVAR PARSE-PATHNAME-FLAG ())  

(DEFUN PATHNAME-ERROR (INDEX FORMAT-STRING &REST ARGS)
  (IF PARSE-PATHNAME-FLAG
    (THROW 'PARSE-PATHNAME
	   INDEX)
    (APPLY 'FERROR 'PATHNAME-PARSE-ERROR FORMAT-STRING ARGS)))

(COMPILE-FLAVOR-METHODS PATHNAME-ERROR)  


(DEFMACRO FILE-OPERATION-RETRY (&BODY BODY)
  "Like ERROR-RESTART for condition FILE-ERROR, but uses proceed-type :RETRY-FILE-OPERATION."
  (LET* ((TAG (GENSYM)))
    `(BLOCK FILE-OPERATION-RETRY
       (TAGBODY
	RETRY
	   (RETURN-FROM FILE-OPERATION-RETRY
	     (CATCH-CONTINUATION-IF T ',TAG #'(LAMBDA (IGNORE) (GO RETRY)) NIL
	       (WITH-STACK-LIST* (EH:CONDITION-RESUME-HANDLERS
				   '(FILE-ERROR :RETRY-FILE-OPERATION T
						("Retry the operation on the same file.")
						SI:CATCH-ERROR-RESTART-THROW ,TAG)
				   EH:CONDITION-RESUME-HANDLERS)
	         . ,BODY)))))))
