;; -*- Mode:COMMON-LISP; Package:FS; Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985- 1989 Texas Instruments Incorporated. All rights reserved.
;;;  ** (c) Copyright 1980 Massachusetts Institute of Technology **

;;;
;;; REVISED:  
;;;	05.06.87 MBC 	Fix TOPS20 :read-instance to use :TOPS20, not :TOPS-20.
;;;	Added :READ-INSTANCE methods for the four host flavor types below.  11-12-85 MBC
;;;	Fixed :PATCH-FILE-PATHNAME for aref on PATOM.  11-26-85 MBC

;;                 delimeter, EOS,     previous delimeter, parsing-function




(defvar Tops20-table '(((#\:)     nil        (#\space)       tenex-parse-device)           ; Device
		       ((#\<)     nil      (#\space #\:)     nil                nil  t)
		       ((#\>)     nil          (#\<)         lm-parse-directory     )      ; Directory
		       ((#\. )     t     (#\space #\: #\>)   lm-parse-name-component)      ; Name
		       ((#\. )     t           (#\.)         Tenex-Parse-Type nil nil 1)   ; Type
		       ( nil       t           (#\.)         Tenex-parse-version)))        ; Version




(defvar Tenex-table  '(((#\:)     nil        (#\space)       tenex-parse-device)           ; Device
		       ((#\<)     nil      (#\space #\:)     nil                nil  t)
		       ((#\>)     nil          (#\<)         lm-parse-directory     )      ; Directory
		       ((#\. #\;)  t     (#\space #\: #\>)   lm-parse-name-component)      ; Name
		       ((#\;)      t           (#\.)         tenex-parse-type nil nil 1)   ; Type
		       ( nil       t           (#\;)         Tenex-parse-version)))        ; Version


(defvar ITS-table  '(((#\:)     nil        (#\space)         tenex-parse-device)           ; Device
                     ((#\;)     nil      (#\space #\:)       ITS-PARSE-DIRECTORY)          ; Directory
		     ((#\space)   t     (#\space #\: #\;)    LM-parse-name-COMPONENT)      ; Name
		     ( nil       t          (#\space)        ITS-parse-Type-AND-Version))) ; Type or Version


(Defvar Multics-Table
	'(((#\> #\<) Nil    (#\SPACE)     Unix-Parse-Directory   T)      ; Directory 
	  (Nil        T (#\> #\< #\SPACE)  Unix-Parse-Name-And-Type)))   ; Name and Type


;;;----------------        Pathname Dependant Parsing Routines       ----------------




(DEFUN TENEX-PARSE-TYPE (STRING SELF &OPTIONAL JUNK-ALLOWED UNQUOTED-IS-OKAY)
  (COND ((ZEROP (LENGTH STRING)) :UNSPECIFIC)
	((= (LENGTH STRING) 1.)
	 (LET ((CHR (AREF STRING 0.)))
	   (COND ((CHAR= CHR #\*) :WILD)
		 ((CHAR= CHR #\) NIL)
		 ((CHAR= CHR #\) :UNSPECIFIC)
		 (T
		  (NAME-VALIDATION STRING (SEND SELF :NAME-LENGTH) () SELF JUNK-ALLOWED UNQUOTED-IS-OKAY)))))
	(T (NAME-VALIDATION STRING (SEND SELF :NAME-LENGTH) () SELF JUNK-ALLOWED UNQUOTED-IS-OKAY))))  



(DEFUN TENEX-PARSE-DEVICE (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (COND
    ((ZEROP (LENGTH STRING)) NIL)
    (T (NAME-VALIDATION STRING (SEND SELF :NAME-LENGTH) () JUNK-ALLOWED))))  



(DEFUN TENEX-PARSE-VERSION (STRING IG-NORE &OPTIONAL JUNK-ALLOWED)
  ;;;  (declare (optimize (safety 0) (space 2) (compilation-speed 0)))   ;7.16.87
  ig-nore					;Depending on OPTIMIZE options ig-nore is/is not referenced. 7.16.87
  (LET (NUM
	(S-LEN (LENGTH STRING)))
    (COND
      ((null string) :NEWEST)
      ((ZEROP S-LEN) NIL)
      ((AND (= S-LEN 1.) (CHAR= (AREF STRING 0.) #\*)) :WILD)
      ((AND (= S-LEN 1.) (CHAR= (AREF STRING 0.) #\)) :UNSPECIFIC)
      ((AND (= S-LEN 1.) (CHAR= (AREF STRING 0.) #\)) NIL)
      ((SETQ NUM (NUMERIC-P STRING () T))
       (cond ((= NUM -2) :OLDEST)
	     ((= NUM -3) :WILD)
	     (T NUM)))
      (JUNK-ALLOWED
       (TENEX-PARSE-VERSION (NSUBSTRING STRING 0. (- S-LEN 1.)) JUNK-ALLOWED))
      (T (FERROR 'PATHNAME-PARSE-ERROR "Invalid version spec ~S" STRING)))))  


(DEFUN ITS-PARSE-DIRECTORY (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (LET ((DIR (LM-PARSE-NAME-COMPONENT STRING SELF JUNK-ALLOWED)))
    (IF DIR
	(LIST DIR)
	DIR)))  

(DEFUN ITS-PARSE-TYPE-AND-VERSION (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (LET (TEM
	TYP
	VERS)
    (COND
      ((NULL STRING))
      ((SETQ TEM (NUMERIC-P STRING)) (SETQ VERS TEM
					   TYP :UNSPECIFIC))
      ((= (LENGTH STRING) 1.)
       (LET ((CHR (AREF STRING 0.)))
	 (COND
	   ((CHAR= CHR #\>) (SETQ VERS :NEWEST
				  TYP :UNSPECIFIC))
	   ((CHAR= CHR #\<) (SETQ VERS :OLDEST
				  TYP :UNSPECIFIC))
	   ((CHAR= CHR #\*) (SETQ TYP :WILD
				  VERS :WILD))
	   (T
	    (SETQ TYP (NAME-VALIDATION STRING (SEND SELF :NAME-LENGTH) () JUNK-ALLOWED)
		  VERS :UNSPECIFIC)))))
      (T
       (SETQ TYP (NAME-VALIDATION STRING (SEND SELF :NAME-LENGTH) () JUNK-ALLOWED)
	     VERS :UNSPECIFIC)))
    (VALUES TYP VERS)))


;;;; Multics Support

(DEFFLAVOR NEW-MULTICS-PATHNAME-MIXIN () () (:REQUIRED-FLAVORS PATHNAME))  


(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING &OPTIONAL (START 0.) END)
  (MULTIPLE-VALUE-BIND (DIR NAM TYP LEN)
    (LM-PARSE-NAMESTRING NAMESTRING SELF MULTICS-TABLE START END)
    (VALUES :UNSPECIFIC DIR NAM TYP :UNSPECIFIC LEN)))  

(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :PARSE-RELATIVE-DIRECTORY) (STRING IGNORE)
  (LMFS-PARSE-RELATIVE-PATHNAME STRING))  

(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :NAME-LENGTH) ()
  256.)  

(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :TYPE-LENGTH) ()
  256.)  

(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :STRING-FOR-HOST) ()
  (FORMAT () "~@[~A~]~A" (LMFS-DIRECTORY DIRECTORY) (UNIX-FILENAME NAME TYPE)))
  
(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :STRING-FOR-EDITOR) ()
  (let ((DIR (LMFS-DIRECTORY DIRECTORY)))
    (IF (ZEROP (LENGTH DIR))
	(SETQ DIR ()))
    (FORMAT () "~A~@[ ~A~] ~A:"
	    (UNIX-FILENAME NAME TYPE)
	    DIR
	    (SEND HOST :SHORT-NAME))))

(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :STRING-FOR-DIRECTORY) ()
  (IF DIRECTORY
      (LMFS-DIRECTORY DIRECTORY)
      ""))


(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :DIRECTORY-DELIMITER-CHARACTER) ()
  #\>)

(DEFMETHOD (NEW-MULTICS-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :MULTICS))

;;;; TOPS-20 (TWENEX) support



(DEFFLAVOR NEW-TOPS20-PATHNAME-MIXIN () (PATHNAME-ALWAYS-UPPER-CASE-MIXIN)
   (:INCLUDED-FLAVORS REMOTE-PATHNAME))  



(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING &OPTIONAL (START 0.) END)
  (MULTIPLE-VALUE-BIND (DEV DIR NAM TYP VER LEN)
    (LM-PARSE-NAMESTRING NAMESTRING SELF TOPS20-TABLE START END)
    (VALUES DEV DIR NAM (IF (AND (EQ TYP :UNSPECIFIC) VER)	;Back in.. 11.15.86 MBC
			  ()
			  TYP)
	    VER LEN)))


(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :PRIMARY-DEVICE) ()
  "PS")  


(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :NAME-LENGTH) ()
  ())  
	   

(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :SUPPRESSED-DEVICE-NAMES) ()
  '(NIL :UNSPECIFIC "PS"))  


(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :QUOTE-CHARACTER) ()
  #\)  


(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :DELIMITERS) ()
  '(#\: #\< #\> #\.))  


(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :CHARACTER-NEEDS-QUOTING-P) (CH)
  (MEMBER CH '(#\: #\< #\> #\.) :TEST #'EQ))  



(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :TOPS20))	;5.6.87



(DEFMETHOD (NEW-TOPS20-PATHNAME-MIXIN :NEW-MAIL-PATHNAME) ()
  (SEND SELF :NEW-PATHNAME :NAME "MAIL" :TYPE "TXT" :VERSION 1.))  

;;;; TENEX support



(DEFFLAVOR NEW-TENEX-PATHNAME-MIXIN () (PATHNAME-ALWAYS-UPPER-CASE-MIXIN)
   (:INCLUDED-FLAVORS REMOTE-PATHNAME))  




(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING &OPTIONAL (START 0.) END)
  (MULTIPLE-VALUE-BIND (DEV DIR NAM TYP VER LEN)
    (LM-PARSE-NAMESTRING NAMESTRING SELF TENEX-TABLE START END)
    (VALUES DEV DIR NAM (IF (AND (EQ TYP :UNSPECIFIC) VER)
			  ()
			  TYP)
	    VER LEN)))  



(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :PRIMARY-DEVICE) ()
  "DSK")  



(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :NAME-LENGTH) ()
  ())  



(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :VERSION-DELIMITER) ()
  #\;)  



(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :SUPPRESSED-DEVICE-NAMES) ()
  '(NIL :UNSPECIFIC "DSK"))  



(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :QUOTE-CHARACTER) ()
  #\)  



(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :DELIMITERS) ()
  '(#\: #\< #\> #\. #\;))  



(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :CHARACTER-NEEDS-QUOTING-P) (CH)
  (MEMBER CH '(#\: #\< #\> #\. #\;) :TEST #'EQ))  



(DEFMETHOD (NEW-TENEX-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :TENEX))  

;;;; ITS support

(DEFFLAVOR NEW-ITS-PATHNAME-MIXIN () (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN)
   (:REQUIRED-FLAVORS PATHNAME))  

;;; STRING-OR-WILD is also used by Tenex-Family and Vms.

(DEFUN STRING-OR-WILD (FIELD SELF &OPTIONAL NO-QUOTE-P SPECIALS REPLACED-BY)
  "Convert FIELD, a pathname component, to a string to appear in a printed representation.
NO-QUOTE-P inhibits insertion of quoting characters;
otherwise, quote characters are inserted and some characters translated:
SPECIALS is a list of characters to be translated,
and REPLACED-BY is an equally-long list of characters to translate them to."
  ;; Taken from MIT-patch 31 by BRG.
  (COND
    ((EQ FIELD :WILD) "*")
    ((MEMBER FIELD '(NIL :UNSPECIFIC) :TEST #'EQ) NIL)
    (NO-QUOTE-P (STRING FIELD))
    ((QUOTE-COMPONENT-STRING FIELD SELF SPECIALS REPLACED-BY))))

(DEFUN QUOTE-COMPONENT-STRING (STRING SELF &OPTIONAL SPECIALS REPLACED-BY &AUX LENGTH)
  "Put a quote character before each character of STRING that needs one.
Copies STRING if it has to be changed.  Any instance of a char in SPECIALS
is replaced by the corresponding char in REPLACED-BY, not quoted."
  (SETQ STRING (STRING STRING)
	LENGTH (LENGTH STRING))
  (DO ((NSTRING NIL)
       (QUOTE-IDX 0. NQUOTE-IDX)
       (NQUOTE-IDX -1.))
      (NIL)
    (SETQ NQUOTE-IDX
	  (DO ((I (1+ NQUOTE-IDX) (1+ I)))
	      ((>= I LENGTH))
	    (AND (NOT (MEMBER (AREF STRING I) SPECIALS :TEST #'EQ))
		 (SEND SELF :CHARACTER-NEEDS-QUOTING-P (AREF STRING I)) (RETURN I))))
    (AND (SEND SELF :QUOTE-CHARACTER) (OR NQUOTE-IDX NSTRING)
	 (SETQ NSTRING
	       (IF NSTRING
		   (STRING-APPEND NSTRING (SEND SELF :QUOTE-CHARACTER)
				  (SUBSEQ STRING QUOTE-IDX NQUOTE-IDX))
		   (SUBSEQ STRING QUOTE-IDX NQUOTE-IDX))))
    (OR NQUOTE-IDX (IF (INTERSECTION SPECIALS REPLACED-BY :TEST #'EQUAL)	;avoid endless loop 11-27-85
		       (RETURN STRING))
	(PROGN
	  (LOOP
	    (LET ((I (STRING-SEARCH-SET SPECIALS (OR NSTRING STRING))))
	      (OR I (RETURN))
	      (OR NSTRING (SETQ NSTRING (STRING-APPEND STRING)))
	      (SETF (AREF NSTRING I)
		    (NTH (POSITION (AREF NSTRING I) SPECIALS :TEST #'EQ) REPLACED-BY))))
	  (RETURN (OR NSTRING STRING))))))  


;;; If the device is DSK, avoid printing it.
(DEFUN ITS-DEVICE-STRING (SELF &OPTIONAL OMIT-HOST &AUX HOSTS PRINT-HOSTS)
  (LET ((DEVICE (PATHNAME-DEVICE SELF))
	(HOST (PATHNAME-HOST SELF)))
    (SETQ HOSTS (SEND HOST :NAME-AS-FILE-COMPUTER))
    (SETQ PRINT-HOSTS (IF OMIT-HOST
			  (IF (MEMBER DEVICE '("DSK" NIL :UNSPECIFIC) :TEST #'EQUAL)
			      ()
			      DEVICE)
			  HOSTS))
    (IF (OR (MEMBER DEVICE '("DSK" NIL :UNSPECIFIC) :TEST #'EQUAL) (EQUAL DEVICE HOSTS) OMIT-HOST)
	PRINT-HOSTS
	(STRING-APPEND PRINT-HOSTS ": " DEVICE))))  

;;; If name is a list, its first component is the FN1 and second is FN2
;;; If only FN2 is present, FN1 is placeholder ""


(DEFUN ITS-FN1-STRING (SELF &OPTIONAL NO-QUOTE-P NO-PLACEHOLDER Use-Up-Horseshoe-p)
  (LET ((NAME (PATHNAME-NAME SELF)))
    (COND
      ((NULL NAME) (IF NO-PLACEHOLDER
		     ()
		     ""))
      ((EQ NAME :UNSPECIFIC)
       (if Use-Up-Horseshoe-p "" nil))
      ((CONSP NAME) (IF NO-QUOTE-P
		      (CAR NAME)
		      (QUOTE-COMPONENT-STRING (CAR NAME) SELF)))
      (T (STRING-OR-WILD NAME SELF NO-QUOTE-P)))))  



(DEFUN ITS-FN2-STRING (SELF &OPTIONAL NO-QUOTE-P NO-PLACEHOLDER)
  (LET ((NAME (PATHNAME-NAME SELF))
	(TYPE (PATHNAME-TYPE SELF))
	(VERSION (PATHNAME-VERSION SELF)))
    (COND
      ((AND (CONSP NAME) (CDR NAME))
       (IF NO-QUOTE-P
	 (CADR NAME)
	 (QUOTE-COMPONENT-STRING (CADR NAME) SELF)))
      ((AND (NULL TYPE) (NULL VERSION)) (IF NO-PLACEHOLDER
					  ()
					  ""))
      ((AND (MEMBER TYPE '(NIL :UNSPECIFIC) :TEST #'EQ)
	  (MEMBER VERSION '(NIL :UNSPECIFIC) :TEST #'EQ))
       NIL)
      ((EQ VERSION :OLDEST) "<")
      ((EQ VERSION :WILD) "*")
      ((OR (EQ VERSION :UNSPECIFIC) (NOT (MEMBER TYPE *ITS-UNINTERESTING-TYPES* :TEST #'EQUAL)))
       (STRING-OR-WILD TYPE SELF NO-QUOTE-P))
      ((NOT (MEMBER VERSION '(NIL :UNSPECIFIC :NEWEST) :TEST #'EQ)) (FORMAT () "~D" VERSION))
      (T ">"))))  


;;; Truncate to six characters


(DEFUN SIX-SIXBIT-CHARACTERS (STRING &OPTIONAL QUOTE-P (START 0.) (END (LENGTH STRING)))
  "Truncate STRING to six characters and make sure they are all SIXBIT characters.
START and END specify the part of STRING to use (default is all).
QUOTE-P non-NIL says insert quoting characters where appropriate."
  (DO ((I START (1+ I))
       (NCH 0.)
       (CH)
       (NEED-COPY NIL))
      ((OR (>= I END) (= NCH 6.))
       (COND
	 ((AND (= START 0.) (= I (LENGTH STRING)) (NOT NEED-COPY)) STRING)	;To avoid consing
	 ((NOT NEED-COPY) (SUBSEQ STRING START I))
	 (T
	  (DO ((NSTRING (MAKE-ARRAY NCH :ELEMENT-TYPE 'STRING-CHAR))
	       (J 0.)
	       (K START (1+ K))
	       (CH))
	      ((>= K I)
	       NSTRING)
	    (SETQ CH (AREF STRING K))
	    (COND
	      ((NOT (AND QUOTE-P (MEMBER CH '(#\ #\) :TEST #'EQ)))
	       (SETQ CH (COND
			  ((< CH 32.) (+ CH 32.))
			  ((< CH 96.) CH)
			  (T (- CH 32.))))
	       (SETF (AREF NSTRING J) CH) (SETQ J (1+ J))))))))
    (SETQ CH (AREF STRING I))
    (IF (AND QUOTE-P (MEMBER CH '(#\ #\) :TEST #'EQ))
	(SETQ NEED-COPY T)
	(SETQ NCH (1+ NCH)))
    (OR (AND (>= CH #\SPACE) (<= CH #\_))	;Already legal SIXBIT
	(SETQ NEED-COPY T))))

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING &OPTIONAL (START 0.) END)
  (MULTIPLE-VALUE-BIND (DEV DIR NAM TYP VERS LEN)
    (LM-PARSE-NAMESTRING NAMESTRING SELF ITS-TABLE START END)
    (VALUES DEV DIR NAM TYP VERS LEN)))  


(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :NAME-LENGTH) ()
  6.)  

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :QUOTE-CHARACTER) ()
  #\)  

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :CHARACTER-NEEDS-QUOTING-P) (CH)
  (MEMBER CH '(#\; #\: #\SPACE) :TEST #'EQ))  

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :DELIMITERS) ()
  '(#\; #\: #\SPACE))  

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :STRING-FOR-HOST) ()
  (LET ((DIR (STRING-OR-WILD (IF (CONSP DIRECTORY)
			       (CAR DIRECTORY)
			       DIRECTORY)
			     SELF))
	(XDEV (ITS-DEVICE-STRING SELF T))
	(FN1 (ITS-FN1-STRING SELF nil nil t))	;T for up-horseshoes
	(FN2 (ITS-FN2-STRING SELF () T)))
    (IF (AND (NULL FN2) (NULL NAME))
      (SETQ FN1 ()))
    (FORMAT () "~@[~A: ~]~@[~A;~]~@[ ~A~]~@[ ~A~]" XDEV DIR FN1 FN2)))  

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :STRING-FOR-PRINTING) ()
  (LET ((DEV (ITS-DEVICE-STRING SELF))
	(DIR (STRING-OR-WILD (IF (CONSP DIRECTORY)
				 (CAR DIRECTORY)
				 DIRECTORY)
			     SELF))
	(FN1 (ITS-FN1-STRING SELF nil nil t))	;T for up-horseshoes
	(FN2 (ITS-FN2-STRING SELF () T)))
    (IF (AND (NULL FN2) (NULL NAME))
	(SETQ FN1 ()))
    (FORMAT () "~A:~:[ ~A;~:[ ~]~;~2*~]~@[~A~]~@[ ~A~]" DEV
	    (MEMBER DIRECTORY '(NIL :UNSPECIFIC) :TEST #'EQ) DIR
	    (AND (MEMBER NAME '(NIL :UNSPECIFIC) :TEST #'EQ)	;name could be :UNSPECIFIC  11.15.86 MBC
		 (MEMBER TYPE '(NIL :UNSPECIFIC) :TEST #'EQ)
		 (MEMBER VERSION '(NIL :UNSPECIFIC) :TEST #'EQ))
	    FN1 FN2)))


(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :STRING-FOR-EDITOR) ()
  (LET ((FN1 (ITS-FN1-STRING SELF))
	(FN2 (ITS-FN2-STRING SELF () T))
	(DIR (STRING-OR-WILD (IF (CONSP DIRECTORY)
				 (CAR DIRECTORY)
				 DIRECTORY)
			     SELF)))
    (IF (AND (NULL FN2) (NULL NAME))
	(SETQ FN1 ()))
    (FORMAT () "~@[~A ~]~:[~A ~;~*~]~:[~A: ~;~*~]~@[~A; ~]~A:"
	    FN1
	    ;; Do we really have a significant type or version worth printing?
	    (NOT (OR (AND TYPE (NEQ TYPE :UNSPECIFIC)) (AND VERSION (NEQ VERSION :UNSPECIFIC))))
	    FN2
	    (OR (MEMBER DEVICE '("DSK" NIL :UNSPECIFIC) :TEST #'EQUAL)
		(EQUAL DEVICE (SEND HOST :NAME-AS-FILE-COMPUTER)))
	    (SEND SELF :DEVICE) DIR (SEND HOST :SHORT-NAME))))

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :STRING-FOR-DIRECTORY) ()
  (LET ((DIR (STRING-OR-WILD (IF (CONSP DIRECTORY)
			       (CAR DIRECTORY)
			       DIRECTORY)
			     SELF)))
    (FORMAT () "~:[~A: ~;~*~]~@[~A;~]"
	    (MEMBER DEVICE '("DSK" NIL :UNSPECIFIC) :TEST #'EQUAL)
	    (SEND SELF :DEVICE) DIR)))

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :STRING-FOR-DIRED) ()
  (LET ((FN1 (ITS-FN1-STRING SELF))
	(FN2 (ITS-FN2-STRING SELF () T)))
    (IF (AND (NULL FN2) (NULL NAME))
      (SETQ FN1 ()))
    (FORMAT () "~@[~6A~]~@[ ~A~]" FN1 FN2)))

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :FN1) ()
  (ITS-FN1-STRING SELF T T))

(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :FN2) ()
  (ITS-FN2-STRING SELF T T))

(DEFVAR *ITS-UNINTERESTING-TYPES* '("LISP" "TEXT" NIL :UNSPECIFIC))

;; Differs from the default method in that if both type and version are specified
;; we clobber one of them to :UNSPECIFIC.
;; If only one is specified, we clobber the other to :UNSPECIFIC.
;; Exception: if either of them is :WILD, both of them are :WILD.


(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :NEW-PATHNAME) (&REST OPTIONS &KEY &OPTIONAL STARTING-PATHNAME ((:VERSION -VERSION-) NIL VERSION-P)
  &ALLOW-OTHER-KEYS &AUX -TYPE- TYPE-P)
  (SETQ -TYPE- (CADR (GETL (LOCF OPTIONS) '(:TYPE :CANONICAL-TYPE))))
  (COND
    ((EQ -TYPE- :WILD) (SETQ -VERSION- :WILD
			     VERSION-P T))
    ((EQ -VERSION- :WILD) (SETQ -TYPE- :WILD
				TYPE-P T))
    ((AND (NOT (MEMBER -TYPE- '(NIL :UNSPECIFIC) :TEST #'EQ))
	(NOT (MEMBER -VERSION- '(NIL :UNSPECIFIC) :TEST #'EQ)))
     (IF (MEMBER -TYPE- *ITS-UNINTERESTING-TYPES* :TEST 'STRING-EQUAL)
       (SETQ -TYPE- :UNSPECIFIC
	     TYPE-P T)
       (SETQ -VERSION- :UNSPECIFIC
	     VERSION-P T)))
    ((NOT (MEMBER -TYPE- '(NIL :UNSPECIFIC) :TEST #'EQ))
     (SETQ -VERSION- :UNSPECIFIC
	   VERSION-P T))
    ((NOT (MEMBER -VERSION- '(NIL :UNSPECIFIC) :TEST #'EQ)) (SETQ -TYPE- :UNSPECIFIC
								  TYPE-P T)))
  (APPLY #'MAKE-PATHNAME-1 (IF TYPE-P
			     :TYPE)
	 -TYPE- (IF VERSION-P
		  :VERSION)
	 -VERSION- :STARTING-PATHNAME (OR STARTING-PATHNAME SELF) :PARSING-PATHNAME SELF OPTIONS))  



(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :AFTER :INIT) (IGNORE)
  (AND (NOT (MEMBER TYPE '(NIL :UNSPECIFIC) :TEST #'EQ))
     (NOT (MEMBER VERSION '(NIL :UNSPECIFIC) :TEST #'EQ))
     (NOT (AND (EQ TYPE :WILD) (EQ VERSION :WILD)))
     (FERROR () "ITS-PATHNAME created with type ~S and version ~S." TYPE VERSION)))  

;;; For most components, just upcase the string


(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :PARSE-COMPONENT-SPEC) (SPEC)
  (COND
    ((CONSP SPEC) (MAPCAR #'(LAMBDA (X)
			      (SEND SELF :PARSE-COMPONENT-SPEC X))
			  SPEC))
    ((STRINGP SPEC) (SIX-SIXBIT-CHARACTERS SPEC))
    (T SPEC)))  



(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :PARSE-NAME-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (SEND SELF :PARSE-COMPONENT-SPEC SPEC))
    ((AND (CONSP SPEC) (STRINGP (CAR SPEC)) (CONSP (CDR SPEC)) (STRINGP (CADR SPEC))
	(NULL (CDDR SPEC)))
     ;; Allow list of length two as name component.
     (SEND SELF :PARSE-COMPONENT-SPEC SPEC))
    ((AND (CONSP SPEC) (STRINGP (CAR SPEC)) (NULL (CDR SPEC)))
     (SEND SELF :PARSE-COMPONENT-SPEC (CAR SPEC)))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD) :TEST #'EQ) SPEC)
    (T "FOO")))  




(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :PARSE-DIRECTORY-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (ITS-PARSE-DIRECTORY SPEC SELF T))
    ((AND (CONSP SPEC) (= (LENGTH SPEC) 1.)) (ITS-PARSE-DIRECTORY (CAR SPEC) SELF T))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :ROOT :WILD) :TEST #'EQ) SPEC)
    (T (PATHNAME-DIRECTORY (QUIET-USER-HOMEDIR HOST)))))  




(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :PARSE-VERSION-SPEC) (SPEC)
  (IF (OR (AND (FIXNUMP SPEC) (> SPEC 0.))
      (MEMBER SPEC '(NIL :UNSPECIFIC :WILD :NEWEST :OLDEST) :TEST #'EQ))
    SPEC
    :NEWEST))  



(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :INIT-FILE) (PROGRAM-NAME)
  (SEND SELF :NEW-PATHNAME :NAME USER-ID :TYPE PROGRAM-NAME))  

;;; These are for things like the microcode files that need to compact both a name and a type
;;; into one pathname.


(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :NEW-TYPE-AND-VERSION) (NEW-TYPE NEW-VERSION)
  (SEND SELF :NEW-PATHNAME :TYPE
     (SIX-SIXBIT-CHARACTERS (FORMAT () "~D~A" (GCD NEW-VERSION 1000.) NEW-TYPE))))  



(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :TYPE-AND-VERSION) (&AUX TYP VERS I)
  (COND
    ((STRINGP TYPE) (MULTIPLE-VALUE-SETQ (VERS I)
					 (NUMERIC-P TYPE T))
		    (AND I (SETQ TYP (SUBSEQ TYPE I))) (VALUES TYP VERS))
    (T (VALUES TYPE TYPE))))  

;;; Patch system interface, more kludges for only six character filenames


(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :PATCH-FILE-PATHNAME) (NAM SAME-DIRECTORY-P PATOM TYP &REST ARGS)
  (CASE TYP
    (:SYSTEM-DIRECTORY
     (SEND SELF :NEW-PATHNAME :NAME
	(IF SAME-DIRECTORY-P
	  PATOM
	  (IF SAME-DIRECTORY-P
	    PATOM
	    (IF (< (LENGTH NAM) 7.)
	      NAM
	      (SI:SYSTEM-SHORT-NAME NAM))))
	:TYPE :PATCH-DIRECTORY))
    (:VERSION-DIRECTORY
     (SEND SELF :NEW-PATHNAME :NAME
	(WITH-OUTPUT-TO-STRING (STREAM)
	  (LET ((SNAME (IF SAME-DIRECTORY-P
			 (STRING PATOM)
			 (SI:SYSTEM-SHORT-NAME NAM))))
	    (DOTIMES (I (MIN (LENGTH SNAME) 3.))
	      (SEND STREAM :TYO (AREF SNAME I))))
	  (FORMAT STREAM "~D" (GCD (CAR ARGS) 1000.)))
	:TYPE :PATCH-DIRECTORY))
    (:PATCH-FILE
     (SEND SELF :NEW-PATHNAME :NAME
	(FORMAT () "~:[~*~;~C~]~D.~D" SAME-DIRECTORY-P PATOM (GCD (CAR ARGS) 100.)
		(GCD (CADR ARGS) (IF SAME-DIRECTORY-P
				   100.
				   1000.)))
	:TYPE (CADDR ARGS)))))  



(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :ITS))  



(DEFMETHOD (NEW-ITS-PATHNAME-MIXIN :NEW-MAIL-PATHNAME) ()
  (SEND SELF :NEW-PATHNAME :NAME (OR NAME USER-ID) :TYPE "MAIL"))  

  


(DEFFLAVOR ITS-PATHNAME () (NEW-ITS-PATHNAME-MIXIN FILE-ACCESS-MIXIN REMOTE-PATHNAME))

(DEFFLAVOR TOPS20-PATHNAME () (NEW-TOPS20-PATHNAME-MIXIN 
			       TENEX-FAMILY-PATHNAME-MIXIN FILE-ACCESS-MIXIN REMOTE-PATHNAME))

(DEFFLAVOR TENEX-PATHNAME () (NEW-TENEX-PATHNAME-MIXIN 
			      TENEX-FAMILY-PATHNAME-MIXIN FILE-ACCESS-MIXIN REMOTE-PATHNAME))

(DEFFLAVOR MULTICS-PATHNAME () (NEW-MULTICS-PATHNAME-MIXIN UNIX-PATHNAME FILE-ACCESS-MIXIN REMOTE-PATHNAME))
