;; -*- Mode:COMMON-LISP; Package:FILE-SYSTEM; cold-load:t; Base:10.  -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985- 1989 Texas Instruments Incorporated. All rights reserved.

;; Pathname parsing support for that special LM system, Symbolics.

;;;
;;; REVISED:  
;;;	03.23.87 MBC 	Don't put horseshoes in :STRING-FOR-HOST.
;;;	03.09.87 MBC	Add :PROBE-DIRECTORY support.
;;;   Added :READ-INSTANCE method for the host flavor, LMFS.  11-12-85 MBC


;;                 delimeter, EOS,     previous delimeter, parsing-function



(DEFVAR LMFS-TABLE
   '(((#\> #\<) NIL (#\SPACE) LMFS-PARSE-DIRECTORY T); Directory 
     (NIL T (#\> #\< #\SPACE) LMFS-PARSE-NAME-TYPE-VERSION)))   ; Name Type and Version

;;;-------------     Lmfs Pathname Dependant Parsing Routines    -------------



(DEFUN LMFS-PARSE-NAME-COMPONENT (STRING &OPTIONAL IGNORE)
  (COND
    ((ZEROP (LENGTH STRING)) NIL)
    ((= (LENGTH STRING) 1.)
     (LET ((CHR (AREF STRING 0.)))
       (COND
	 ((CHAR= CHR #\*) :WILD)
	 ((CHAR= CHR #\) NIL)
	 ((CHAR= CHR #\) :UNSPECIFIC)
	 (T STRING))))
    (T STRING)))  



(DEFUN LMFS-PARSE-TYPE-COMPONENT (STRING &OPTIONAL IGNORE)
  (LET ((TYP (LMFS-PARSE-NAME-COMPONENT STRING)))
    (IF (AND (STRINGP TYP) (> (LENGTH TYP) 14.))
      (NSUBSTRING TYP 0. 14.)
      TYP)))  



(DEFUN LMFS-PARSE-VERSION (STRING &OPTIONAL JUNK-ALLOWED)
  (LET ((ALPHABETIC-CASE-AFFECTS-STRING-COMPARISON NIL))
    (COND
      ((NUMERIC-P STRING () T))
      ((ZEROP (LENGTH STRING)) NIL)
      ((AND (= (LENGTH STRING) 1.)
	    (LET ((CHR (AREF STRING 0.)))
	      (COND
		((CHAR= CHR #\*) :WILD)
		((CHAR= CHR #\) NIL)
		((CHAR= CHR #\) :UNSPECIFIC)
		(T nil)))))
      ((STRING-EQUAL STRING "NEWEST") :NEWEST)
      ((STRING-EQUAL STRING "OLDEST") :OLDEST)
      (JUNK-ALLOWED
       (LMFS-PARSE-VERSION (SUBSEQ STRING 0. (- (LENGTH STRING) 1.)) JUNK-ALLOWED))
      (T (FERROR 'PATHNAME-PARSE-ERROR "Invalid version spec ~S" STRING)))))  



(DEFUN MORE-THAN-ONE-PERIOD (STRING)
  (LET ((END (LENGTH STRING))
	(START 0.)
	(COUNT 0.)
	POSITION)
    (LOOP
     (IF (> START END)
       (RETURN (> COUNT 1.))
       (IF (SETQ POSITION     ; bump up the count if we find a period
	      (POSITION #\. (THE STRING (STRING STRING)) :START START :END END :TEST #'CHAR=))
	 (SETQ COUNT (1+ COUNT)
	       START (1+ POSITION))
	 (RETURN (> COUNT 1.)))))))  
    


(DEFUN LMFS-PARSE-NAME-TYPE-AND-VERSION (STRING &OPTIONAL JUNK-ALLOWED)
  (LET (START
	VER
	NAM
	TYP)
    (SETQ START					; reverse search for periods
	  (POSITION #\. (THE STRING (STRING STRING)) :FROM-END T :TEST #'CHAR=))
    (WHEN START
      (SETQ VER (LMFS-PARSE-VERSION (NSUBSTRING STRING (1+ START)) JUNK-ALLOWED)))
    (MULTIPLE-VALUE-SETQ (NAM TYP)
			 (LMFS-PARSE-NAME-AND-TYPE (NSUBSTRING STRING 0. START) JUNK-ALLOWED))
    (VALUES NAM TYP VER)))  



(DEFUN LMFS-PARSE-NAME-AND-TYPE (STRING &OPTIONAL JUNK-ALLOWED)
  (LET (START
	TYP
	NAM)
    (SETQ START					; reverse search for periods
	  (POSITION #\. (THE STRING (STRING STRING)) :FROM-END T :TEST #'CHAR=))
    (WHEN START
      (SETQ TYP (LMFS-PARSE-TYPE-COMPONENT (NSUBSTRING STRING (1+ START)) JUNK-ALLOWED)))
    (SETQ NAM (LMFS-PARSE-NAME-COMPONENT (NSUBSTRING STRING 0. START) JUNK-ALLOWED))
    (VALUES NAM TYP ())))  
      


(DEFUN LMFS-PARSE-NAME-TYPE-VERSION (STRING IGNORE &OPTIONAL JUNK-ALLOWED)
  (IF STRING
    (IF (MORE-THAN-ONE-PERIOD STRING)
      (LMFS-PARSE-NAME-TYPE-AND-VERSION STRING JUNK-ALLOWED)
      (LMFS-PARSE-NAME-AND-TYPE STRING JUNK-ALLOWED))
    (VALUES () () ())))  



(DEFUN LMFS-PARSE-DIRECTORY-COMPONENT (STRING &OPTIONAL JUNK-ALLOWED)
  (COND
    ((STRING= STRING "*") :WILD)
    ((STRING= STRING "**") :WILD-INFERIORS)
    ((ZEROP (LENGTH STRING))
     (IF JUNK-ALLOWED
       ()
       (FERROR 'PATHNAME-PARSE-ERROR "Null directory components not allowed ~S" STRING)))
    (T STRING)))  



(DEFUN LMFS-PARSE-DIRECTORY (STRING IGNORE &OPTIONAL JUNK-ALLOWED)
  (LET ((START-INDEX 0.)
	(END-INDEX 0.)
	RESULT)
    (MULTIPLE-VALUE-SETQ (RESULT STRING)
      (LMFS-PARSE-RELATIVE-PATHNAME STRING))
    (WHEN STRING
      (DOTIMES (I (LENGTH STRING))
	(IF (NOT (CHAR= (AREF STRING I) #\>))
	  (SETQ END-INDEX (1+ END-INDEX))
	  (LET ((COMP
		 (LMFS-PARSE-DIRECTORY-COMPONENT (NSUBSTRING STRING START-INDEX END-INDEX)
						 JUNK-ALLOWED)))
	    (WHEN COMP
	      (PUSH COMP RESULT))
	    (SETQ START-INDEX (1+ END-INDEX)
		  END-INDEX (1+ END-INDEX)))))
      (IF (NEQ START-INDEX END-INDEX)
	(LET ((COMP
	       (LMFS-PARSE-DIRECTORY-COMPONENT (NSUBSTRING STRING START-INDEX END-INDEX)
					       JUNK-ALLOWED)))
	  (WHEN COMP
	    (PUSH COMP RESULT)))))
    (IF (CONSP RESULT)
      (NREVERSE RESULT)
      RESULT)))  



(DEFUN LMFS-PARSE-RELATIVE-PATHNAME (STRING); USED BY LMFS AND MULTICS PATHNAMES !!
  (IF (NULL STRING)
    (VALUES () ())
    (LET (RESULT
	  (FIRST-RELATIVE T))
      (DOTIMES (I (LENGTH STRING))
	(IF (CHAR= (AREF STRING I) #\<)
	  (PROGN
	    (WHEN FIRST-RELATIVE
	      (PUSH :RELATIVE RESULT)
	      (SETQ FIRST-RELATIVE ()))
	    (PUSH :UP RESULT))
	  (IF (NOT (CHAR= (AREF STRING I) #\>))
	    (PROGN
	      (WHEN FIRST-RELATIVE
		(PUSH :RELATIVE RESULT))
	      (RETURN (VALUES RESULT (NSUBSTRING STRING I))))
	    (IF (EQ (LENGTH STRING) 1.)
	      (RETURN (VALUES :ROOT ()))
	      (RETURN (VALUES RESULT (NSUBSTRING STRING (1+ I)))))))
	(VALUES RESULT ())))))  
	      
;;;-------------     End of Lmfs Pathname Dependant Parsing Routines    -------------







(DEFFLAVOR NEW-LMFS-PATHNAME-MIXIN ()
   (PATHNAME-ALWAYS-UPPER-CASE-MIXIN PATHNAME-NORMALLY-LOWERCASE-MIXIN)
   (:INCLUDED-FLAVORS PATHNAME))  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING &OPTIONAL (START 0.) END)
  (MULTIPLE-VALUE-BIND (DIR NAM TYP VERS LEN)
    (LM-PARSE-NAMESTRING NAMESTRING SELF LMFS-TABLE START END)
    (VALUES :UNSPECIFIC DIR NAM TYP VERS LEN)))  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :STRING-FOR-PRINTING) ()
  (FORMAT () "~A:~@[~A~]~A" (FUNCALL HOST :NAME-AS-FILE-COMPUTER) (LMFS-DIRECTORY DIRECTORY)
	  (LMFS-FILENAME NAME TYPE VERSION T)))

(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :STRING-FOR-EDITOR) ()
  (FORMAT () "~A ~@[~A~] ~A:" (LMFS-FILENAME NAME TYPE VERSION) (LMFS-DIRECTORY DIRECTORY)
	  (FUNCALL HOST :SHORT-NAME)))

(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :SHORT-STRING-FOR-PRINTING) ()
  (FORMAT () "~A:~@[~A~]~A" (FUNCALL HOST :SHORT-NAME) (LMFS-DIRECTORY DIRECTORY)
	  (LMFS-FILENAME NAME TYPE VERSION)))

(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :STRING-FOR-HOST) ()
  (FORMAT () "~@[~A~]~A" (LMFS-DIRECTORY DIRECTORY) (LMFS-FILENAME NAME TYPE VERSION)))	;3.23.87



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :STRING-FOR-DIRED) ()
  (LMFS-FILENAME NAME TYPE VERSION))  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :STRING-FOR-DIRECTORY) ()
  (FORMAT () "~@[~a~]" (LMFS-DIRECTORY DIRECTORY)))



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :LMFS :LMFS))  	;SYMBOLICS: indicate file-system-type



(DEFUN LMFS-FILENAME (NAME TYPE VERSION &optional print-horseshoe)
  (LET ((NAME-S (LMFS-NAME-COMPONENT NAME print-horseshoe))
	(TYPE-S (LMFS-NAME-COMPONENT TYPE print-horseshoe))
	(VERSION-S (LMFS-VERSION-COMPONENT VERSION print-horseshoe)))
    (UNLESS (OR (AND VERSION VERSION-S) (AND TYPE TYPE-S))
      (SETQ TYPE-S ()))
    (UNLESS (OR TYPE-S (AND NAME NAME-S))
      (SETQ NAME-S ()))
    (FORMAT () "~@[~A~]~@[.~A~]~@[~:*~:[.~;~].~D~]" NAME-S TYPE-S VERSION-S)))  



(DEFUN LMFS-NAME-COMPONENT (NAME &optional print-horseshoe)
  (SELECT NAME
    ((NIL) NIL)
    (:UNSPECIFIC (and print-horseshoe #\))
    (:WILD "*")
    (OTHERWISE NAME)))  


(DEFUN LMFS-VERSION-COMPONENT (VERSION &optional print-horseshoe)
  (IF (NUMBERP VERSION)
      VERSION
      (CASE VERSION
	    (:UNSPECIFIC (and print-horseshoe #\))
	    (:WILD "*")
	    (OTHERWISE VERSION))))



(DEFUN LMFS-DIRECTORY (DIRECTORY)
  (LET ((ABSOLUTE-PATHNAME T))
    (WITH-OUTPUT-TO-STRING (S)
      (DOLIST (ONE-COMPONENT DIRECTORY)
	(CASE ONE-COMPONENT
	  ((:RELATIVE :ROOT) (SETQ ABSOLUTE-PATHNAME ()))
	  (:UP (SEND S :TYO #\<))
	  (:WILD (WHEN ABSOLUTE-PATHNAME
		   (SEND S :TYO #\>))
	   (SEND S :TYO #\*) (SETQ ABSOLUTE-PATHNAME T))
	  (:WILD-INFERIORS (WHEN ABSOLUTE-PATHNAME
			     (SEND S :TYO #\>))
	   (SEND S :STRING-OUT "**") (SETQ ABSOLUTE-PATHNAME T))
	  (OTHERWISE (WHEN ABSOLUTE-PATHNAME
		       (SEND S :TYO #\>))
	   (SEND S :STRING-OUT ONE-COMPONENT) (SETQ ABSOLUTE-PATHNAME T))));FAKE THINGS OUT.
      (WHEN DIRECTORY
	(SEND S :TYO #\>)))))  


(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :INTERNAL-WILD-CHARACTERS) ()
  (VALUES #\* -1.))  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :UNDELETABLE-P) ()
  T)  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :DIRECTORY-PATHNAME-AS-FILE) LMFS-DIRECTORY-PATHNAME-AS-FILE)  

(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :PROBE-DIRECTORY) ()	;3.9.87 MBC
  (Probe-file (send self :directory-pathname-as-file)))

(DEFUN LMFS-DIRECTORY-PATHNAME-AS-FILE (IGNORE &AUX DIR NAM)
  (DECLARE (:SELF-FLAVOR PATHNAME))
  (COND
    ((EQ DIRECTORY :ROOT)
     (FERROR 'PATHNAME-PARSE-ERROR "There is no pathname for the root as a file"))
    ((OR (ATOM DIRECTORY) (NULL (CDR DIRECTORY)))
     (SETQ DIR :ROOT
	   NAM (IF (CONSP DIRECTORY)
		 (CAR DIRECTORY)
		 DIRECTORY)))
    (T (LET ((LAST (LAST DIRECTORY)))
	 (SETQ DIR (LDIFF DIRECTORY LAST)
	       NAM (CAR LAST)))))
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY DIR :RAW-NAME NAM :TYPE
     (SEND SELF :DIRECTORY-FILE-TYPE) :VERSION (SEND SELF :DIRECTORY-VERSION)))  

;;; By default, a directory is stored as a file in the superior directory whose name gives
;;; the name of the component at this level.


(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :PATHNAME-AS-DIRECTORY) ()
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY
	(IF (EQ DIRECTORY :ROOT)
	    (CONS NAME ())
	    (APPEND (IF (CONSP DIRECTORY)
			DIRECTORY
			(CONS DIRECTORY ()))
		    (CONS NAME ())))
	:NAME :UNSPECIFIC :TYPE :UNSPECIFIC :VERSION :UNSPECIFIC))  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :DIRECTORY-FILE-TYPE) ()
  "DIRECTORY")  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :DIRECTORY-VERSION) ()
  1.)  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :PARSE-COMPONENT-SPEC) (SPEC)
  (COND
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD) :TEST #'EQ) SPEC)
    ((STRINGP SPEC) SPEC)
    (T "FOO")))  


(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :PARSE-TYPE-SPEC) (SPEC)
  (COND
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD) :TEST #'EQ) SPEC)
    ((STRINGP SPEC) (LMFS-PARSE-TYPE-COMPONENT SPEC T))
    ((KEYWORDP spec) (string (DECODE-CANONICAL-TYPE spec :LMFS)))	;allow unknown :TYPEs
    (T (DECODE-CANONICAL-TYPE :LISP (SEND HOST :SYSTEM-TYPE)))))


(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :PARSE-DIRECTORY-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (LIST SPEC))
    ((AND (CONSP SPEC)
	(LOOP FOR ELT IN SPEC ALWAYS
	   (OR (MEMBER ELT '(:RELATIVE :UP :WILD-INFERIORS :WILD) :TEST #'EQ) (STRINGP ELT)))
	(NOT (MEMBER :RELATIVE (CDR SPEC) :TEST #'EQ)))
     (LOOP FOR ELT IN SPEC COLLECT (IF (SYMBOLP ELT)
				     ELT
				     (SEND SELF :PARSE-COMPONENT-SPEC ELT))))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :ROOT) :TEST #'EQ) SPEC)
    ((MEMBER SPEC '(:WILD :WILD-INFERIORS) :TEST #'EQ) (LIST SPEC))
    (T (PATHNAME-DIRECTORY (QUIET-USER-HOMEDIR HOST)))))  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :PARSE-VERSION-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (LMFS-PARSE-VERSION SPEC T))
    ((OR (FIXNUMP SPEC) (MEMBER SPEC '(NIL :UNSPECIFIC :WILD :NEWEST :OLDEST) :TEST #'EQ)) SPEC)
    (T :NEWEST)))  


;; Not really a parsing function, but...
;(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN  :HOMEDIR) (&OPTIONAL (USER USER-ID))
;  (FUNCALL-SELF ':NEW-DIRECTORY USER))



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :PRIMARY-DEVICE) ()
  :UNSPECIFIC)  



(DEFMETHOD (NEW-LMFS-PATHNAME-MIXIN :DIRECTORY-WILD-P) ()
  (IF (OR (SEND SELF :COMPONENT-CONTAINS-WILD DIRECTORY) (EQ DIRECTORY :ROOT)
      (AND (CONSP DIRECTORY) (MEMBER :WILD-INFERIORS DIRECTORY :TEST #'EQ)))
    T))


(DEFFLAVOR LMFS-PATHNAME () (NEW-LMFS-PATHNAME-MIXIN FILE-ACCESS-MIXIN REMOTE-PATHNAME))
