;; -*- Mode:Common-Lisp; Package:FILE-SYSTEM; Base:10.; COLD-LOAD:T  -*-
;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985- 1989 Texas Instruments Incorporated. All rights reserved.

;; Pathname parsing support for LM systems.

;;; 
;;; REVISED:
;;;     01-30-89 DAB    Modified lm-print-name to print "." only when you need to.
;;;     07-11-88 DAB    Modify lm-table to end version search with SPACE.
;;;	03.23.88 MBC	Fix DEFAULT-DIRECTORY-PATHNAME-AS-FILE & :PROBE-DIRECTORY to
;;;			handle :ROOT properly.
;;;     02.19.88 MBC	Move the rest of the lm-parse-... & NAME-VALIDATION to "PARSE".
;;;	02.17.88 MBC	Moved lm-parse-device out of this file and to "parse".
;;;	02.15.88 MBC	Moved lispm-pathname compile-flavor-method here.  Obsoletes
;;			the file "SYS: PATHNAME; COMPILE-FLAVOR-METHODS.LISP".
;;;     09.29.87 DNG	Moved variables LOGICAL-TABLE, LM-QUOTE-CHARACTER, and 
;;;			LM-DELIMITERS and functions LM-NAMESTRING and 
;;;			LM-NAMESTRING-FOR-DIRECTORY from here to file "LOGICAL".
;;;	07.16.87 MBC 	Fudge around compiler warnings.
;;;	04.30.87 MBC	Allow restore of files to lm:~; without Length of :ROOT error.
;;;	03.23.87 MBC 	Don't put horseshoes in :STRING-FOR-HOST.
;;;	03.09.87 MBC	Add :PROBE-DIRECTORY support.
;;; 10-30-86 MBC	Make :STRING-FOR-PRINTING & :STRING-FOR-HOST print Up-Horseshoe
;;;			for :UNSPECIFIC components.

;;               delimiter    EOS     previous delimiter       parsing function


(DEFVAR LM-TABLE
   '(((#\:) NIL (#\SPACE) LM-PARSE-DEVICE); Device
     ((#\<) NIL (#\SPACE #\:) NIL NIL T) ((#\>) NIL (#\<) LM-PARSE-DIRECTORY); Twenex style directory
     ((#\;) NIL (#\SPACE #\: #\<) LM-PARSE-DIRECTORY); Directory
     ((#\. #\#) T (#\SPACE #\: #\; #\>) LM-PARSE-NAME-COMPONENT); Name
     ((#\. #\#) T (#\.) LM-PARSE-NAME-COMPONENT); Type
     ((#\SPACE) T (#\. #\#) LM-PARSE-VERSION)))  ;07-11-88 DAB       ; Version



;;;; LM FILE support

(DEFFLAVOR NEW-LM-PARSING-MIXIN () (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN)
   (:INCLUDED-FLAVORS REMOTE-PATHNAME))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING &OPTIONAL (START 0.) END)
  (MULTIPLE-VALUE-BIND (DEV DIR1 DIR2 NAM TYP VER LEN)
    (LM-PARSE-NAMESTRING NAMESTRING SELF LM-TABLE START END)
    (VALUES DEV (OR DIR1 DIR2) NAM TYP VER LEN)))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :QUOTE-CHARACTER) () #\/)  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :DELIMITERS) () '(#\: #\; #\< #\> #\. #\#))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :PARSE-DEVICE-SPEC) (IGNORE) ())  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :NAME-LENGTH) () ())  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :short-STRING-FOR-PRINTING) ()
  (LM-NAMESTRING HOST (SEND SELF :DEVICE) (SEND SELF :DIRECTORY) (SEND SELF :NAME)
		 (SEND SELF :TYPE) (SEND SELF :VERSION) :short-p))

(DEFMETHOD (NEW-LM-PARSING-MIXIN :STRING-FOR-PRINTING) ()
  (LM-NAMESTRING HOST (SEND SELF :DEVICE) (SEND SELF :DIRECTORY) (SEND SELF :NAME)
		 (SEND SELF :TYPE) (SEND SELF :VERSION)))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :STRING-FOR-HOST) ()
  (LM-NAMESTRING () (SEND SELF :DEVICE) (SEND SELF :DIRECTORY) (SEND SELF :NAME)
		 (SEND SELF :TYPE) (SEND SELF :VERSION) :short-p))	;3.23.87



(DEFMETHOD (NEW-LM-PARSING-MIXIN :STRING-FOR-DIRED) ()
  (LM-NAMESTRING-FOR-DIRED (SEND SELF :NAME) (SEND SELF :TYPE) (SEND SELF :VERSION)))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :STRING-FOR-EDITOR) ()
  (LM-NAMESTRING-FOR-EDITOR HOST (SEND SELF :DEVICE) (SEND SELF :DIRECTORY) (SEND SELF :NAME)
			    (SEND SELF :TYPE) (SEND SELF :VERSION)))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :STRING-FOR-DIRECTORY) ()
  (LM-NAMESTRING-FOR-DIRECTORY (SEND SELF :DEVICE) (SEND SELF :DIRECTORY)))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :DIRECTORY-PATHNAME-AS-FILE) DEFAULT-DIRECTORY-PATHNAME-AS-FILE)  

(DEFUN DEFAULT-DIRECTORY-PATHNAME-AS-FILE (IGNORE &AUX DIR NAM)
  (DECLARE (:SELF-FLAVOR PATHNAME))
  (COND
    ((EQ DIRECTORY :ROOT)
     (return-from DEFAULT-DIRECTORY-PATHNAME-AS-FILE
       (send self :new-pathname :type nil :name nil :version nil)))	;3.23.88  removed ferror
    ((OR (ATOM DIRECTORY) (NULL (CDR DIRECTORY)))
     (SETQ DIR :ROOT
	   NAM (IF (CONSP DIRECTORY)
		 (CAR DIRECTORY)
		 DIRECTORY)))
    (T (LET ((LAST (LAST DIRECTORY)))
	 (SETQ DIR (LDIFF DIRECTORY LAST)
	       NAM (CAR LAST)))))
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY DIR :RAW-NAME NAM :TYPE
     (SEND SELF :DIRECTORY-FILE-TYPE) :VERSION :NEWEST))  


(DEFMETHOD (NEW-LM-PARSING-MIXIN :DIRECTORY-FILE-TYPE) ()
  "DIRECTORY")  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :PARSE-COMPONENT-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (LM-PARSE-NAME-COMPONENT SPEC SELF T T))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD) :TEST #'EQ) SPEC)
    (T "FOO")))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :PARSE-DIRECTORY-COMPONENT-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) SPEC)
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD :ROOT) :TEST #'EQ) SPEC)
    (T "FOO")))  




(DEFMETHOD (NEW-LM-PARSING-MIXIN :PARSE-DIRECTORY-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (LIST SPEC))
    ((CONSP SPEC) (MAPCAR SELF (CIRCULAR-LIST :PARSE-DIRECTORY-COMPONENT-SPEC) SPEC))
    ((EQ SPEC :WILD) :ROOT)
    ((MEMBER SPEC '(NIL :UNSPECIFIC :ROOT :WILD) :TEST #'EQ) SPEC)
    (T (PATHNAME-DIRECTORY (QUIET-USER-HOMEDIR HOST)))))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :PARSE-VERSION-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (LM-PARSE-VERSION SPEC SELF T))
    ((OR (FIXNUMP SPEC) (MEMBER SPEC '(NIL :UNSPECIFIC :WILD :NEWEST :OLDEST) :TEST #'EQ)) SPEC)
    (T :NEWEST)))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :UNDELETABLE-P) ()
  T)  

(DEFMETHOD (new-lm-parsing-mixin :MEANINGFUL-ROOT) () NIL)	;4.30.87


;; Not really a parsing function, but...


(DEFMETHOD (NEW-LM-PARSING-MIXIN :PROBE-DIRECTORY) ()	;3.23.88 MBC
  (if (eq directory :ROOT)			;Root exist 
      self
      (probe-file (send self :directory-pathname-as-file))))


(DEFMETHOD (NEW-LM-PARSING-MIXIN :HOMEDIR) (&OPTIONAL (USER USER-ID))
  (SEND SELF :NEW-DIRECTORY USER))  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :PRIMARY-DEVICE) ()
  ())  



(DEFMETHOD (NEW-LM-PARSING-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :LISPM))






(DEFVAR *LMFS-USE-TWENEX-SYNTAX* () "When non-NIL, print out LMFS pathnames in Twenex syntax.")  


(DEFUN LM-PRINT-DIRECTORY (IGNORE DIRECTORY S SPACE)
  (COND ((NOT (MEMBER DIRECTORY '(NIL :UNSPECIFIC) :TEST #'EQ))
	 (IF *LMFS-USE-TWENEX-SYNTAX*
	     (WRITE-CHAR #\< S)
	     (IF (NULL SPACE)
		 (SETQ SPACE T)
		 (WRITE-CHAR #\SPACE S)))
	 (COND ((OR (MEMBER DIRECTORY '(NIL :ROOT) :TEST #'EQ)
		    (AND (CONSP DIRECTORY)
			 (MEMBER (CAR DIRECTORY) '(NIL :ROOT) :TEST #'EQ)
			 (NULL (CDR DIRECTORY))))
		(IF (NOT *LMFS-USE-TWENEX-SYNTAX*)
		    (WRITE-CHAR #\~ S)))
	       ((CONSP DIRECTORY)
		(DO ((D DIRECTORY (CDR D)))
		    ((NULL D))
		  (LM-PRINT-COMPONENT (CAR D) S)
		  (WHEN (CDR D)
		    (WRITE-CHAR #\. S))))
	       (T (LM-PRINT-COMPONENT DIRECTORY S)))
	 (WRITE-CHAR (IF *LMFS-USE-TWENEX-SYNTAX* #\> #\;) S))
	(T (SETF SPACE NIL)))
  SPACE)  


(DEFUN LM-PRINT-NAME (NAME TYPE VERSION S SPACE &optional (print-horseshoe nil))
  (when name
    (when SPACE
      (IF (NOT *LMFS-USE-TWENEX-SYNTAX*)
	  (FUNCALL S :TYO #\SPACE))
      (SETQ SPACE ()))
    (if (EQ NAME :UNSPECIFIC)
	(and print-horseshoe (SEND S :TYO #\))
	(LM-PRINT-COMPONENT NAME S)))
      
  (when type
    (if space
	(progn
	(IF (NOT *LMFS-USE-TWENEX-SYNTAX*)
	    (FUNCALL S :TYO #\SPACE))
	(SETQ SPACE ())))
    (select TYPE ; DAB 01-30-89 Don't print "." unless type.
      (:UNSPECIFIC (when print-horseshoe (FUNCALL S :TYO #\.) (SEND S :TYO #\)))
      (nil)
      (t (FUNCALL S :TYO #\.)
	 (LM-PRINT-COMPONENT TYPE S))))

  (IF *LMFS-USE-TWENEX-SYNTAX*
      (COND
	((NOT (MEMBER VERSION '(NIL :NEWEST) :TEST #'EQ))
	 (FUNCALL S :TYO #\.)
	 (IF (EQ VERSION :UNSPECIFIC)
	     (and print-horseshoe (SEND S :TYO #\))
	     (LM-PRINT-COMPONENT VERSION S T))))
      (when version    ; DAB 01-30-89 Don't print "#" unless version.
	(if SPACE (FUNCALL S :TYO #\SPACE) (SETQ SPACE ()))
	(SELECT VERSION 
	  (:unspecific (and print-horseshoe (FUNCALL S :TYO #\#) (SEND S :TYO #\)))
	  (nil)
	  (t (FUNCALL S :TYO #\#) (LM-PRINT-COMPONENT VERSION S T)))))
  SPACE)



(DEFUN LM-PRINT-HOST (HOST S &optional short-p)
  (COND
    ((NOT (NULL HOST)) (princ (if short-p (send host :name) (send HOST :NAME-AS-FILE-COMPUTER)) S)
     (FUNCALL S :TYO #\:) T)
    (T NIL)))



(DEFUN LM-NAMESTRING-FOR-EDITOR (HOST DEVICE DIRECTORY NAME TYPE VERSION)
  (WITH-OUTPUT-TO-STRING (S)
    (LM-PRINT-NAME NAME TYPE VERSION S ())
    (FUNCALL S :TYO #\SPACE)
    (LM-PRINT-DIRECTORY DEVICE DIRECTORY S ())
    (FUNCALL S :TYO #\SPACE)
    (LM-PRINT-HOST HOST S t)))



(DEFUN LM-NAMESTRING-FOR-DIRED (NAME TYPE VERSION)
  (WITH-OUTPUT-TO-STRING (S)
    (COND
      ((NOT (MEMBER NAME '(NIL :UNSPECIFIC) :TEST #'EQ)) (LM-PRINT-COMPONENT NAME S)))
    (COND
      (TYPE (FUNCALL S :TYO #\.) (OR (EQ TYPE :UNSPECIFIC) (LM-PRINT-COMPONENT TYPE S))))
    (IF (NOT *LMFS-USE-TWENEX-SYNTAX*)
      (FORMAT S "~15T"))
    (COND
      (VERSION (FUNCALL S :TYO (IF *LMFS-USE-TWENEX-SYNTAX*
				 #\.
				 #\#))
       (OR (EQ VERSION :UNSPECIFIC) (LM-PRINT-COMPONENT VERSION S T))))))  



(DEFUN LM-PRINT-COMPONENT (SPEC STREAM &OPTIONAL VERSION-P &AUX TEM)
  (COND
    ((EQ SPEC :WILD) (FUNCALL STREAM :TYO #\*))
    ((NUMBERP SPEC) (FORMAT STREAM "~D" SPEC))
    (VERSION-P
     (COND
       ((SETQ TEM (CDR (ASSOC SPEC '((:NEWEST . #\>) (:OLDEST . #\<)) :TEST #'EQ)))
	(FUNCALL STREAM :TYO TEM))
       (T (FERROR "Attempt to print ~S, which is not a valid version." SPEC))))
    ((STRINGP SPEC)
     (DOTIMES (I (ARRAY-ACTIVE-LENGTH SPEC))
       (AND (MEMBER (SETQ TEM (AREF SPEC I)) LM-DELIMITERS :TEST #'EQ)
	  (FUNCALL STREAM :TYO LM-QUOTE-CHARACTER))
       (FUNCALL STREAM :TYO TEM)))
    (T (FERROR () "Attempt to print ~S, which is not a valid component." SPEC)))) 





;; This is called explicitly by BOOT-FILE-SYSTEM, No one should call this anymore. *BJ*
;(DEFUN MAYBE-ADD-LOCAL-PATHNAME-HOST ()
;  (ADD-CHAOSNET-FILE-COMPUTER SI:LOCAL-HOST))


;; When bound to T, access local file system via the server (mainly for debugging).
;; When bound to NIL, access local file system directly.


;; (REMPROP '*LOCAL-SERVER-VIA-NET* :SOURCE-FILE-NAME)  


(DEFVAR *LOCAL-SERVER-VIA-NET* ())

(DEFFLAVOR LISPM-PATHNAME () (NEW-LM-PARSING-MIXIN FILE-ACCESS-MIXIN REMOTE-PATHNAME))

(DEFMETHOD (LISPM-PATHNAME :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :LISPM))
  
(COMPILE-FLAVOR-METHODS LISPM-PATHNAME)		;2.15.88
