;; -*- Mode:COMMON-LISP; Package:FS; Base:10.; patch-file:t; -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1987- 1989 Texas Instruments Incorporated. All rights reserved.

;;;
;;; ORIGINAL:	8.20.87 MBC	Macintosh pathname support.  Compatible
;;;				with the 128K ROM version, with respect to
;;;				pathname component lengths.

;;; REVISED:  			REAL DEVICE VERSION !!!

;;;             02-01-89 DAB    Changed (MAC-PATHNAME-MIXIN :NEW-TYPE) to use canonical-type ,except for nil and :unspecific.
;;;             06-20-88 DAB     Fixed :mac-pathname :new-pathname to handle :directory '(:root) , '("abc" :wild) and
;;;                              '(:root "abc").
;;;             06-16-88 DAB    Remove parse-pathname-find-colon in (MAC-PATHNAME-MIXIN :PARSE-NAMESTRING). It was 
;;;                             removing devices that had names the same as hosts.
;;;             06-08-88 DAB    Fix to (MAC-PATHNAME-MIXIN :NEW-PATHNAME) to truncate the name based on the length
;;;                             of the canonical type if supplied.
;;;             06-01-88 DAB    Fixes (MAC-PATHNAME-MIXIN :PARSE-NAMESTRING) to not parse out the host component
;;;                             when START or END is supplied.
;;;             05-19-88 DAB    Fixes case in :new-pathname :directory "abc:def".
;;;             05-23-88 DAB    Numerous correction to mac name and type trunications.           
;;;		03.21.88 MBC	Logical pathname translated to MAC passed type as :CANONICAL-TYPE,
;;;				fooling mac's :NEW-PATHNAME into returning name + type field > 31 chars.
;;;             03-18-88 DAB    Added :directory-wild-p to mac-pathname-mixin.
;;;             03-17-88 DAB    Added some changes to mac-pathname to handle ROOT.
;;;		03.11.88 MBC	Fix :pathname-as-directory to disallow NIL in directory component list.
;;;		03.09.88 MBC	:MERGE-DEVICE and :QUIET-HOMEDIR methods fixed for :DEVICE.
;;;		03.02.88 MBC	Use MAC-FILENAME in :STRING-FOR-NFS, not UNIX-FILENAME.
;;;		02.25.88 MBC	Introduce MAC-PARSE-DEVICE-AND-DIRECTORY to correctly handle device
;;;				parsing and relative directory parsing.
;;;	        02.15.88 MBC	Clean up:  Remove MAC-UNIX-PATHNAME and its mixin. 
;;;				Remove large piece of kruft from tail end of this source file.
;;;     	02.10.88 MBC	New (temporary) method, :MERGE-DEVICE, for MAC until it really uses :DEVICE.
;;;		02.07.88 MBC	Make :PATHNAME-AS-DIRECTORY treat :UNSPECIFIC name correctly.
;;;		02.02.88 MBC	:STRING-FOR-HOST uses native MAC directory delimiters and a new
;;;				method, :STRING-FOR-NFS, just for NFS.  All other servers, protocols
;;;				should use :STRING-FOR-HOST, which they already do.
;;;		01.27.88 MBC	Hack to put back DEVICE back.
;;;		01.11.88 MBC	Moved several functions out of this file and into their proper files:
;;;				- IO;OPEN, PATHNAME;PARSE, PATHNAME;CANONICAL-TYPES & PATHNAME;DRIVER.
;;;		01.08.88 MBC	:NEW-PATHNAME method created to prevent creation of mac pathnames
;;;				with name and type fields longer than legal limit.
;;;  		01.04.88 MBC	Keep :NEW-NAME-AND-TYPE from needlessly truncating name when 
;;;				object rcving msg had a long type field.
;;;		12.18.87 MBC	:HOMEDIR calls SI:GET-ALL-VOLUME-NAMES and hardcodes login to the first 
;;;				volume of the MAC.
;;;				NOTE: When NOT booting off of load band on MAC this gives
;;;				Explorer volume names instead of MAC's. 
;;;		11.05.87 MBC 	Tweaks to MAC-PARSE-DIRECTORY to handle "::kdkkd".
;;;				Add :NEW-NAME-AND-TYPE for NFS use as optimization.
;;;				Complete-String from IO;OPEN now has mac-pathname dependency.
;;;		10.29.87 MBC	Parse of ":::" and character validation range is to D8 not 8D.
;;;		10.26.87 MBC	:PATHNAME-AS-DIRECTORY handle type, too.
;;;		10.23.87 MBC	Handle (parse-pathname "X:" <MAC-HOST>).

;;; Implementation notes:

;;; Syntax is simply all components seperated by #\: (colons), with the exception
;;;   of the name and type which are seperated by a PERIOD.
;;; Lengths are limited to 27 for device (volume) and 31 for the rest.  
;;; Directory nodes limited to 31 each. ???
;;; Characters allowed are all "printable characters" which is the Ascii
;;;   range of #x20 to #xD8, excluding colon, (#xA3).
;;;   There are two blanks, #x20 and #xCA, the "nonbreaking space".
;;;   Since the character sets differ so much for the range of #x80 to #xD8
;;;   it might be wise to just disallow them.  
;;;   #x7F is the DEL character code, which IS NOT printable.
;;;   General observation:  Mac directory nodes often have embedded #\SPACEs.
;;;
;;; Pathnames on the MAC do have the concept of a relative pathname which they
;;;   call a "partial pathname".  An absolute pathname is called a "full pathname".
;;;   We may want to support relative pathnames, but on the MAC it is relative to
;;;   the parent's directory ID, an integer.  It would require some kind of 
;;;   working directory default to be kept around, since passing MAC directory ID's 
;;;   around on the Explorer seems horribly ungraceful.
;;; Total string length on the MAC is limited to 255, so one document states that
;;;   "it may not be possible to identify every node in a catalog with a full pathname."
;;;   This probably doesn't affect pathname support directly, but it could be a problem
;;;   for the MAC side of NFS.  In fact "partial pathnames" WITH directory IDs may 
;;;   Become a neccessity to be able access files whose full pathnames exceed 255
;;;   characters.  Perhaps a full pathname can be passed to the MAC NFS side and it
;;;   could arbitrarily break it into a directory ID and a partial pathname.

;;; In the MAC there are two types of file systems.
;;;   MFS, Macintosh File System, which is always flat and unsorted.
;;;   HFS, Hierarchical File System, is the upward compatible replacement
;;;   and supports directories, sorted catalogs, as well as a better 
;;;   disk allocation scheme.


;;; For NFS there is a :STRING-FOR-NFS method that returns a string, that is not cached,
;;; that uses #\/ for the directory delimiter, instead of #\:.
;;; 

;;; SLASHIFICATION PROBLEM:				11.10.87 MBC 
;;;	Mac pathnames that contain an actual forward-slash (#\/) in either the 
;;;	name component or in a directory component confuse NFS.  For now we should
;;;     document that we don't handle the / in a MAC pathname, but the correct
;;;	solution would be as follows.
;;;	On explorer originated pathnames we should protect the / with slashification
;;; 	  in the :string-for-host.  The MAC NFS server side would need recognize
;;;	  this slashification and treat the pathname accordingly.
;;;     The converse, a MAC originated pathname that contains the / would need to be
;;;	  to be slashified on the MAC side before being sent accross RPC, otherwise
;;;	  information is lost.   The Explorer side that converts this string to
;;;	  a pathname should also recognize this slashification and do the right thing.


;;;  Delimiters,  EOS, Prev-Delimiters, Parsing-Fnc, Reverse-Srch, Reset-Ptr, Number-Vals, Edge-Spaces-P


(DEFVAR MAC-TABLE
	'(
	  ((#\:) NIL (#\SPACE #\:) MAC-PARSE-DEVICE-AND-DIRECTORY T NIL 2 T)	;Device 1.27.88 MBC 
	  (NIL T (#\: #\SPACE) MAC-PARSE-NAME-AND-TYPE NIL NIL T)))       ; Name

(defvar *MERGE-MAC-TYPES* T)			;Make it True 10.8.87 MBC
						;controls value of non-specified type.
						;When NIL then it'll be :UNSPECFIC, no type merging.
						;When T then type will be NIL, allowing merging.

;;;----------------        Mac Pathname Dependant Parsing Routines       ----------------

(DEFUN MAC-PARSE-NAME (STRING SELF &OPTIONAL JUNK-ALLOWED with-respect-to-type)
  (MAC-NAME-VALIDATION (MAC-PARSE-NAME-COMPONENT STRING)
		       (SEND SELF :NAME-LENGTH  with-respect-to-type)
		       JUNK-ALLOWED))

(DEFUN MAC-PARSE-TYPE (STRING SELF &OPTIONAL JUNK-ALLOWED  with-respect-to-name)
  (MAC-NAME-VALIDATION (MAC-PARSE-NAME-COMPONENT STRING)
		       (SEND SELF :TYPE-LENGTH  with-respect-to-name)
		       JUNK-ALLOWED))

(DEFUN MAC-PARSE-NAME-AND-TYPE (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (COND
    ((OR (NOT (STRINGP STRING)) (ZEROP (LENGTH STRING)))
     (VALUES NIL (IF *MERGE-MAC-TYPES*	NIL :UNSPECIFIC)))	;NIL case still use global flag 1.23.87 MBC
    (T
     (LET ((PERIOD-POSITION			;Find right most period position
	     (POSITION #\. (THE STRING (STRING STRING)) :FROM-END T :TEST #'CHAR=))
	   TYP NAM NON-PERIOD-POSITION)
       (COND
	 ((OR (NULL PERIOD-POSITION) (ZEROP PERIOD-POSITION))	;No period or right most period is in position 0
	  (SETQ TYP (IF *MERGE-MAC-TYPES* NIL :UNSPECIFIC)
		NAM (MAC-PARSE-NAME STRING SELF JUNK-ALLOWED :unspecific)));05-23-88 DAB
	 ((= (1+ PERIOD-POSITION) (LENGTH STRING))	;Period is at right most position.
	  (SETQ TYP (IF *MERGE-MAC-TYPES* NIL :UNSPECIFIC)
		NAM (MAC-PARSE-NAME (NSUBSTRING STRING 0 (1+ PERIOD-POSITION)) SELF JUNK-ALLOWED :unspecific))) ;05-23-88 DAB
	 (T
	  (SETQ NON-PERIOD-POSITION		;6.12.87 MBC
		(POSITION #\. (THE STRING (STRING STRING)) :FROM-END T :TEST-NOT #'CHAR= :END PERIOD-POSITION))
	  (if non-period-position		;if there isn't another non-period character, type is NIL/:UNSPECIFIC
	      (SETQ TYP (MAC-PARSE-TYPE
			  (NSUBSTRING STRING (1+ PERIOD-POSITION))
				       SELF JUNK-ALLOWED :unspecific);05-23-88 DAB
		    NAM (MAC-PARSE-NAME
			  (NSUBSTRING STRING 0 PERIOD-POSITION)
			  SELF JUNK-ALLOWED :unspecific))
	      (SETQ TYP (IF *MERGE-MAC-TYPES* NIL :UNSPECIFIC)
		    NAM (MAC-PARSE-NAME STRING SELF JUNK-ALLOWED  :unspecific)))));05-23-88 DAB
       (when (and (stringp nam) (stringp typ))
	 ;;; together name and type are limited to the name-length... But we don't want to
	 ;;; truncate either one too severely, so lets keep 9 letters minimum for name and
	 ;;; type will always have a minimum of four letters.   So... we're sometimes forced
	 ;;; to truncate BOTH name and type, but the distribution is kind of arbitrary.		9.21.87
	 (setf nam (mac-name-validation               ;05-23-88 DAB
		     nam
		     (max 9.
			  (- (send self :NAME-LENGTH :unspecific)
			     (LENGTH typ) 1))))
	 (when (> (+ (length nam) (length typ) 1)
		  (send self :name-length :unspecific))
	   (setf typ  (mac-name-validation typ (send self :type-LENGTH nam)))))
       (VALUES NAM TYP)))))


(DEFUN MAC-PARSE-NAME-COMPONENT (STRING)
  (COND
    ((ZEROP (LENGTH STRING)) NIL)
    ((= (LENGTH STRING) 1.)
     (LET ((CHR (AREF STRING 0.)))
       (COND
	 ((CHAR= CHR #\*) :WILD)
	 ((CHAR= CHR #\) NIL)
	 ((CHAR= CHR #\) :UNSPECIFIC)
	 (T STRING))))
    (T STRING))) 


;(DEFUN MAC-NAME-VALIDATION (STRING VALID-LENGTH &OPTIONAL IGNORE)
;  (when (STRINGP STRING)
;    (when (> (LENGTH STRING) VALID-LENGTH)
;      (SETQ STRING (NSUBSTRING STRING 0. VALID-LENGTH)))
;    (WHEN
;      (let (char result
;	    (end (1- (length string))))
;	(do ((i 0 (incf i)))
;	    ((or result (> i end)) result)
;	  (setf char (char-int (aref string i)))
;	  (setf result
;		(cond ((= char #x7F) T)		;PAD
;		      ((< char #x20) T)		;before #\SPACE
;		      ((> char #xD8) T)		;10.29.87  After last printable
;		      (T NIL)))))
;      (FERROR 'PATHNAME-PARSE-ERROR "Invalid character in spec ~s" STRING)))
;  STRING)

(DEFUN MAC-NAME-VALIDATION (STRING VALID-LENGTH &OPTIONAL IGNORE)
  (WHEN (PLUSP valid-length)  ;05-23-88 DAB
    (when (STRINGP STRING) 
      (when (> (LENGTH STRING) VALID-LENGTH)  
	(SETQ STRING (NSUBSTRING STRING 0. VALID-LENGTH)))
      (WHEN
	(let (char result
	      (end (1- (length string))))
	  (do ((i 0 (incf i)))
	      ((or result (> i end)) result)
	    (setf char (char-int (aref string i)))
	    (setf result (or (= char #x80) (= char #x3A)))     ;only #\NULL and #\: are illegal characters in MAC pathname -- Tung (4/14/88)
;		(cond ((= char #x7F) T)		;PAD
;		      ((< char #x20) T)		;before #\SPACE
;		      ((> char #xD8) T)		;10.29.87  After last printable
;		      (T NIL))
	    ))					       ;let
	(FERROR 'PATHNAME-PARSE-ERROR "Invalid character in spec ~s" STRING)))
    STRING))

(DEFFLAVOR MAC-PATHNAME-MIXIN ()
	   ()
  (:REQUIRED-FLAVORS PATHNAME))


(DEFMETHOD (MAC-PATHNAME-MIXIN :NAME-LENGTH) (&optional intype)  ;05-24-88 DAB
  (UNLESS intype (SETQ intype type))
  (IF (MEMBER inTYPE '(NIL :UNSPECIFIC :WILD) :TEST #'EQ)
      31.
      (- 30. (LENGTH (STRING inTYPE)))))			;9.28.87 MBC

(DEFMETHOD (MAC-PATHNAME-MIXIN :TYPE-LENGTH) (&optional inname)  ;05-24-88 DAB
  (UNLESS inname (SETQ inname name))
  (IF (MEMBER inNAME '(NIL :UNSPECIFIC :WILD) :TEST #'EQ)
    31.
    (- 30. (LENGTH (STRING inNAME)))))			; count the PERIOD ! - 9.28.87 MBC

(DEFMETHOD (MAC-PATHNAME-MIXIN :DIRECTORY-LENGTH) ()	   ;directory nodes 9.28.87
  31.)					   

(DEFMETHOD (MAC-PATHNAME-MIXIN :DEVICE-LENGTH) ()
  27.)  

(DEFMETHOD (MAC-PATHNAME-MIXIN :DIRECTORY-DELIMITER-CHARACTER) ()
  #\:)  
    

(DEFUN MAC-PARSE-DEVICE-AND-DIRECTORY (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (LET ((START-INDEX 0)
	(END-INDEX 0)
	DEV-RESULT DIR-RESULT Dir-RELATIVE-P
	(DELIM (SEND SELF :DIRECTORY-DELIMITER-CHARACTER)))
    (if (null string)
	(values NIL nil)
	(if (string-equal string "*")
	    (setf dir-result '(:WILD))		;2.17.88 MBC
	    (MULTIPLE-VALUE-SETQ (DIR-RESULT STRING)
	      (SEND SELF :PARSE-RELATIVE-DIRECTORY STRING))
	    (if dir-result (setf dir-relative-p t))
	    (WHEN STRING			
	      ;;(if (zerop (length string))		;;; don't forget relative results!  11.5.87 MBC
	      ;;  (setf result nil)			
	      (SETQ START-INDEX (MAC-SKIP-ALL-DIRECTORY-DELIMITERS STRING DELIM 0)
		    END-INDEX START-INDEX)
	      (UNLESS (NULL START-INDEX)	;if string was all colons, don't lose :relative. 10.29.87
		(LOOP WITH I = START-INDEX DO
		      (IF (NOT (CHAR= (AREF STRING I) DELIM))
			  (SETQ END-INDEX (1+ END-INDEX)
				I (1+ I))
			  (LET ((COMP
				  (MAC-PARSE-DIRECTORY-COMPONENT (Nsubstring STRING START-INDEX END-INDEX) SELF
								 JUNK-ALLOWED)))
			    (WHEN COMP
			      (PUSH COMP DIR-RESULT))
			    (SETQ I (MAC-SKIP-ALL-DIRECTORY-DELIMITERS STRING DELIM (1+ END-INDEX)))
			    (WHEN I
			      (SETQ END-INDEX I
				    START-INDEX I))))
		      WHILE I)))
	    (when (listp dir-result)
	      
	      (setf dir-result (NREVERSE DIR-RESULT))
	      
	      (unless dir-relative-p (setf dev-result
					   (MAC-NAME-VALIDATION
					     (car dir-result)
					     (SEND SELF :DEVICE-LENGTH))
					   dir-result
					   (or (cdr dir-result) :root)))))
	    (VALUES DEV-RESULT (or DIR-RESULT :ROOT)))))

(DEFMETHOD (MAC-PATHNAME-MIXIN :PARSE-RELATIVE-DIRECTORY) (STRING)
  (MAC-PARSE-RELATIVE-PATHNAME STRING self))

;;; no leading :  => FULL (absolute) directory
;;; One leading : => Partial :RELATIVE   ""
;;; Two or more : => Partial :RELATIVE :UP   ""
(defun mac-parse-relative-pathname (string self)	
  (cond ((zerop (length string))
	 (values nil nil))
	((string-equal (subseq string 0 2) (send self :DIRECTORY-UP-DELIMITER))
	 (values (list :up :relative) (subseq string 2)))
	((char-equal (aref string 0) #\:)
	 (values (list :relative) (subseq string 1)))
	(t (values nil string))))
	     
(DEFUN MAC-PARSE-DIRECTORY-COMPONENT (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (COND
    ((STRING= STRING "*") :WILD)
    ((STRING= STRING "") :ROOT)  ;03-17-88 DAB
    ((ZEROP (LENGTH STRING))
     (IF JUNK-ALLOWED
	 ()
	 (FERROR 'PATHNAME-PARSE-ERROR "Null directory components not allowed ~S" STRING)))
    (T (MAC-NAME-VALIDATION STRING (SEND SELF :DIRECTORY-LENGTH) JUNK-ALLOWED))))  ;9.28.87


(Defun Mac-Skip-All-Directory-Delimiters (String Delim Start)
  (Loop For I From Start To (1- (length String))
	When (Char/= (Aref String I) Delim)
	Do (Return I)))


 

;;;----------------        End Of MAC Pathname Parsing Routines        ------------------

(defparameter mac-char-upcase-vector	;;  Tung (4/11/88)
    (loop with vector = (make-array char-code-limit :element-type 'bit)
	  for char below char-code-limit
	  do (setf (aref vector char) (if (and (both-case-p char)
					       (< char #xE0)	;; Tung (4/11/88)
					       (lower-case-p char)) 1 0))
	  finally (return vector)))

(DEFMETHOD (MAC-PATHNAME-MIXIN :NEW-DIRECTORY) (SPEC)
  (let ((sys:char-upcase-vector fs:mac-char-upcase-vector)) ;; Tung - 4/14/88
    (unless (listp spec)				;2.10.88
      (setf spec (list spec)))
    (SEND SELF :NEW-RAW-DIRECTORY SPEC)) )


(DEFMETHOD (MAC-PATHNAME-MIXIN :NEW-DEVICE) (SPEC)
  (let ((sys:char-upcase-vector fs:mac-char-upcase-vector)) ;; Tung - 4/14/88
    (SEND SELF :NEW-RAW-DEVICE SPEC)) )


(DEFMETHOD (MAC-PATHNAME-MIXIN :NEW-NAME) (SPEC)
  (let ((sys:char-upcase-vector fs:mac-char-upcase-vector)) ;; Tung - 4/14/88
    (SEND SELF :NEW-RAW-NAME (MAC-NAME-VALIDATION SPEC (send self :name-length)))) )

(DEFMETHOD (MAC-PATHNAME-MIXIN :NEW-TYPE) (SPEC)
     (let ((sys:char-upcase-vector fs:mac-char-upcase-vector)) ;; Tung - 4/14/88
       ;;DAB for RDA: Added this (cribbed from PATHNAME-NORMALLY-LOWERCASE-MIXIN) ; DAB 02-01-89
       (IF (SYMBOLP SPEC) 
	   (SEND SELF :NEW-RAW-TYPE
		 (MAC-NAME-VALIDATION
		   (DECODE-CANONICAL-TYPE SPEC (SEND SELF :SYSTEM-TYPE))
		   (send self :type-length)))
	   (SEND SELF :NEW-RAW-TYPE (MAC-NAME-VALIDATION SPEC (send self :type-length)))
	   )
       )
     )


(DEFMETHOD (MAC-PATHNAME-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING &OPTIONAL (START 0) END)
  (let ((fs:*defaults-are-per-host* t)   ;03-17-88 DAB
	(sys:char-upcase-vector fs:mac-char-upcase-vector)) ;; Tung - 4/14/88
    ;;;I removed parse-pathname-find-colon on 06-16-88 DAB It was causing more trouble than it was worth.
    (MULTIPLE-VALUE-BIND (DEV DIR NAM TYP LEN)
	(LM-PARSE-NAMESTRING NAMESTRING SELF MAC-TABLE START END)
      (VALUES DEV DIR NAM TYP :UNSPECIFIC LEN ))))	;1.27.88 MBC


;;; 11.5.87 
(DEFMETHOD (MAC-PATHNAME-MIXIN :NEW-NAME-AND-TYPE) (SPEC)
  (multiple-value-bind (ignore ignore nam typ)
      (SEND (send (send SELF :host) :sample-pathname) :PARSE-NAMESTRING nil spec)
    (send (send
	    (send self :new-type :unspecific)	;Keep previous type length from causing truncation of name. 1.4.88
	    :new-name nam) :new-type typ)))


;;; 1.8.88  When given :name and/or :type that are too long, we get an illegal length in a Mac Pathname.
(Defmethod (MAC-PATHNAME-MIXIN :NEW-PATHNAME) (&REST OPTIONS &KEY &OPTIONAL STARTING-PATHNAME &ALLOW-OTHER-KEYS)
  (LET ((fs:*defaults-are-per-host* t)
	(new-device (OR (GETF options :device) (GETF options :raw-device)))   ;05-13-88 DAB
	(new-directory (OR (GETF options :directory) (GETF options :raw-directory)))   ;05-13-88 DAB
	(new-name (OR (GETF options :name) (GETF options :raw-name)))
	(new-type (OR (GETF options :type) (GETF options :raw-type)
		      (and (GETF options :canonical-type) ;truncation should be based on the canonical type 06-08-88 DAB
			   (DECODE-CANONICAL-TYPE (GETF options :canonical-type)
						  (SEND self :SYSTEM-TYPE)))
		      )))	;3.21.88 MBC for :TRANSLATED-PATHNAME length problems.
    (when (STRINGP new-device)   ;05-13-88 DAB
      (SETF new-device (MAC-NAME-VALIDATION new-device (send self :device-length))))
    (when new-directory   ;05-13-88 DAB
      (setf new-directory
	    (cond ((MEMBER new-directory '(:root :unspecific :wild)) new-directory)
		  ((or (stringp new-directory)
		       (symbolp new-directory))
		   (setq new-directory (string new-directory))
                   (do ((start 0)
			(end 0)
			(new-list()))
		       ((null end) (reverse (remove (string "") new-list :test #'string-equal)))
		     (setq end (position #\: new-directory :test #'char= :start start))
		     (push (MAC-NAME-VALIDATION
			     (nsubstring new-directory start end)
			     (send self :directory-length)) new-list)
		     (when end (setf start (1+ end)))))
		
		  ((consp new-directory)    ;handle case '("dir1" "dir2" ...)
		   (if (and (eq (car new-directory) :root) (eq (length new-directory) 1))   ;case '(:root) 06-20-88 DAB
		       :root
		       (let ((new-list ()))
			 (dolist (old-dir new-directory (reverse new-list))
			   (case old-dir                                   ;06-20-88 DAB
			     (:root ())                                    ;case '(:root "abc" ...)
			     ((:unspecific :wild) (push old-dir new-list)) ;case '("abc" :wild)
			     (T (push (MAC-NAME-VALIDATION (string old-dir) (send self :directory-length)) new-list)))
			   )
		       ))))))

      (COND ((AND (STRINGP new-name) (STRINGP new-type))       ;
	     ;;; periods in the TYPE component throw this scheme off. 
	     (LET ((name-and-type (STRING-APPEND new-name #\. new-type)))
	       (MULTIPLE-VALUE-SETQ (new-name new-type)
		 (MAC-PARSE-NAME-AND-TYPE name-and-type self))))
	    ((stringp new-name)
	     (SETF new-name
		   (MAC-NAME-VALIDATION new-name
					(send self :name-length new-type))))
	    ((stringp new-type)
	     (SETF new-type (MAC-NAME-VALIDATION
			      new-type
			      (send self :type-length new-name)))))
    (WHEN new-device (SETF (GETF options :device) new-device))   ;05-13-88 DAB
    (WHEN new-directory (SETF (GETF options :directory) new-directory))   ;05-13-88 DAB
    (WHEN (GETF options :name) (SETF (GETF options :name) new-name))
    (WHEN (GETF options :type) (SETF (GETF options :type) new-type))
    (when (getf options :version)
      (setf (getf options :version) :unspecific))  ;03-18-88 DAB for AB
    )	;3.11.88
  (APPLY #'MAKE-PATHNAME-1 :STARTING-PATHNAME (OR STARTING-PATHNAME SELF)
	 :PARSING-PATHNAME SELF OPTIONS))



(DEFUN MAC-FILENAME (NAME TYPE &AUX (NEW-TYPE TYPE))
  (when (consp name) (setf name (car name)))
;;; cond would be better?
  (select name
    (:UNSPECIFIC (SETQ NAME ""))
    (:WILD (IF (MEMBER TYPE '(:WILD :UNSPECIFIC NIL) :TEST #'EQ)
	       ;;; Is this neccessary for MAC?? 8.25
	       "*"				
	       (STRING-APPEND "*." NEW-TYPE)))
    (T
     (IF (AND (NULL NAME) (MEMBER TYPE '(NIL :UNSPECIFIC) :TEST #'EQ))
	 ""
	 (PROGN
	   (OR NAME (SETQ NAME ""))
	   (COND
	     ((EQ TYPE :WILD)
	      (if (zerop (length name)) "*"	;5.18.87
		  (IF (AND
			(CHAR= #\* (AREF NAME 0.))
			(EVERY #'(LAMBDA (X)
				   (CHAR= #\. X))
			       (NSUBSTRING NAME 1.)))
		      (STRING-APPEND NAME "*")
		      (STRING-APPEND NAME ".*"))))
	     ((eq TYPE :UNSPECIFIC) NAME)
	     ((eq TYPE NIL) NAME)
	     (T (STRING-APPEND NAME "." NEW-TYPE))))))))

;;; simplify this and just krank thru all dire components pumping out
;;;; the right thing... 8.21.87n 
(DEFUN MAC-DIRECTORY-STRING (&optional NFS)
  (DECLARE (:SELF-FLAVOR MAC-PATHNAME-MIXIN))
  (let ((DELIM (if NFS #\/ (SEND SELF :DIRECTORY-DELIMITER-CHARACTER)))
	(DIRECT DIRECTORY))
    (cond ((MEMBER DIRECT '(NIL :UNSPECIFIC :ROOT) :TEST #'EQ) "")	;9.29.87
	  ((equal direct '(:wild))		;2.17.88 MBC
	   (string-append "*" delim))		;3.4.88 MBC
	  ((not (consp direct)) direct)		;To handle malformed dir components. 10.6.87
	  ((NULL (CDR DIRECT))
	   (STRING-APPEND (MAC-DIRECTORY-COMPONENT (CAR DIRECT)) DELIM))
	  ((stringp DIRECT)
	   (MAC-DIRECTORY-COMPONENT DIRECT))
	  ((CONSP DIRECT)
	   (LOOP FOR SUBDIR IN DIRECT
		 WITH STRING = (MAKE-ARRAY 20. :ELEMENT-TYPE 'STRING-CHAR :LEADER-LIST '(0.))
		 DO
		 (SETQ STRING (STRING-NCONC STRING (MAC-DIRECTORY-COMPONENT SUBDIR)))
		 (VECTOR-PUSH-EXTEND DELIM STRING)
		 FINALLY (RETURN STRING))))))


(DEFMETHOD (MAC-PATHNAME-MIXIN :DIRECTORY-UP-DELIMITER) () "::")  


(DEFUN MAC-DIRECTORY-COMPONENT (STRING)
  (CASE STRING
    (:WILD "*")
    (:root :root) ;03-17-88 DAB
    ((:UP :RELATIVE) "")			;delimiter inserted by caller
    (OTHERWISE (STRING STRING))))


(DEFMETHOD (MAC-PATHNAME-MIXIN :STRING-FOR-HOST) ()	
  (FORMAT () "~@[~A:~]~@[~A~]~A"		;2.02.88 MBC  2.04.88
	  (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ) NIL DEVICE)	;1.27.88 MBC
	  (MAC-DIRECTORY-STRING)
	  (MAC-FILENAME NAME TYPE)))


(DEFMETHOD (MAC-PATHNAME-MIXIN :STRING-FOR-NFS) ()
  (FORMAT () "/~@[~A/~]~@[~A~]~A"		;2.02.88 MBC
	  (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ) NIL DEVICE)	;1.27.88 MBC
	  (MAC-DIRECTORY-STRING T)
	  (MAC-FILENAME NAME TYPE)))



(DEFMETHOD (MAC-PATHNAME-MIXIN :STRING-FOR-HOST-1) ()
  (LET ((DIR-STRING (MAC-DIRECTORY-STRING)))
    (FORMAT () "~@[~A:~]~A~A"			;1.27.88 MBC
	    (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ) NIL DEVICE)
	    (IF (STRINGP DIR-STRING)
		(NSTRING-UPCASE DIR-STRING)
		DIR-STRING)
	    (MAC-FILENAME (SEND SELF :NAME) (SEND SELF :TYPE)))))

(DEFMETHOD (MAC-PATHNAME-MIXIN :SHORT-STRING-FOR-PRINTING) ()
  (STRING-APPEND (FUNCALL HOST :SHORT-NAME) ": " (SEND SELF :STRING-FOR-HOST-1)))

(DEFMETHOD (MAC-PATHNAME-MIXIN :STRING-FOR-PRINTING) ()
  (STRING-APPEND (FUNCALL HOST :NAME-AS-FILE-COMPUTER) ": " (SEND SELF :STRING-FOR-HOST-1)))


(DEFMETHOD (MAC-PATHNAME-MIXIN :STRING-FOR-EDITOR) ()
  (LET ((DIR-STRING (MAC-DIRECTORY-STRING)))
    (FORMAT () "~A ~@[~A:~]~@[~A~] ~A:"
	    (MAC-FILENAME (SEND SELF :NAME) (SEND SELF :TYPE))	;4.2.87
	    (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ)
	      ()
	      (SEND SELF :DEVICE))
	    (IF (STRINGP DIR-STRING)
	      (NSTRING-UPCASE DIR-STRING)
	      DIR-STRING)
	    (SEND HOST :SHORT-NAME))))  


(DEFMETHOD (MAC-PATHNAME-MIXIN :STRING-FOR-DIRECTORY) ()
  (LET ((DIR-STRING (MAC-DIRECTORY-STRING)))
    (FORMAT () "~@[~A:~]~@[~A~]"
	    (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ)
	      ()
	      (SEND SELF :DEVICE))
	    (IF (STRINGP DIR-STRING)
	      (NSTRING-UPCASE DIR-STRING)
	      DIR-STRING))))  

(DEFMETHOD (MAC-PATHNAME-MIXIN :STRING-FOR-DIRED) ()
  (MAC-FILENAME (SEND SELF :NAME) (SEND SELF :TYPE)))

(DEFMETHOD (MAC-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :MAC))


;;; This is a re-written version of NEW-UNIX-pathname-mixin.  It could
;;; be retrofited, or even isolated in a mixin for sharing.  10.26.87 MBC
;;; Fix minor problem of getting NIL for directory component. 3.21.88 MBC
(DEFMETHOD (MAC-PATHNAME-MIXIN :PATHNAME-AS-DIRECTORY) ()
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY
	(let ((name-part (cond ((TYPEP name 'keyword) NIL)	;2.07.88 MBC
			       ((null name) NIL)	;3.21.88
			       ((or (null type) (typep type 'keyword))
				(cons Name Nil))
			       (T (cons (string-append NAME (string #\.) type) nil))))
	      (dir-part  (IF (listp DIRECTORY)	;3.11.88
				    DIRECTORY
				    (CONS DIRECTORY ()))))
	  (IF (EQ DIRECTORY :ROOT)
	      (or name-part :ROOT)		;3.21.88
	      (if (eq (car name-part) nil)
		  dir-part
		  (append dir-part name-part))))
	:NAME :UNSPECIFIC :TYPE :UNSPECIFIC :VERSION :UNSPECIFIC))

(DEFMETHOD (MAC-PATHNAME-MIXIN :DIRECTORY-PATHNAME-AS-FILE) (&optional IGNORE)
  (let (dir nam)
    (COND
      ((EQ DIRECTORY :ROOT)
       (FERROR 'PATHNAME-PARSE-ERROR "There is no pathname for the root as a file"))
      ((OR (ATOM DIRECTORY) (NULL (CDR DIRECTORY)))
       (SETQ DIR :ROOT
	     NAM (IF (CONSP DIRECTORY)
		     (CAR DIRECTORY)
		     DIRECTORY)))
      (T (LET ((LAST (LAST DIRECTORY)))
	   (SETQ DIR (LDIFF DIRECTORY LAST)
		 NAM (CAR LAST)))))
    (values
      (SEND SELF :NEW-PATHNAME
	    :RAW-DIRECTORY DIR :RAW-NAME NAM :TYPE :unspecific :VERSION :NEWEST))))

(DEFMETHOD (MAC-PATHNAME-MIXIN :DIRECTORY-FILE-TYPE) (&optional IGNORE) :UNSPECIFIC)	;3.11.88

;;;  Add special handling of TYPE to :PATHNAME-MATCH :PATHNAME-MATCH-NO-DEVICE.
;;;  Use default methods for :PATHNAME-MATCH-NO-VERSION & :PATHNAME-MATCH-NO-VERSION-OR-DEVICE
;;;; If type is :UNSPECIFIC allow the match.  This makes things work like a Unix user expects,
;;;  i.e. Copy of /dir/* should get all names and types.  Avoids /dir/*.* work around syntax.
;;;  - MBC 9.29.87
(DEFMETHOD (MAC-PATHNAME-MIXIN :PATHNAME-MATCH) (PATHNAME &OPTIONAL (MATCH-HOST T))
  (MULTIPLE-VALUE-BIND (W* W1)
    (SEND SELF :INTERNAL-WILD-CHARACTERS)
    (AND (OR (NOT MATCH-HOST) (EQ HOST (PATHNAME-HOST PATHNAME)))
       (PATHNAME-COMPONENT-MATCH DEVICE (PATHNAME-DEVICE PATHNAME) W* W1)
       (PATHNAME-COMPONENT-MATCH DIRECTORY (PATHNAME-DIRECTORY PATHNAME) W* W1)
       (PATHNAME-COMPONENT-MATCH NAME (PATHNAME-NAME PATHNAME) W* W1)
       (or (eq type :unspecific)
	   (pathname-component-match type (pathname-type pathname) w* w1)))))


(DEFMETHOD (MAC-PATHNAME-MIXIN :PATHNAME-MATCH-NO-DEVICE) (PATHNAME &OPTIONAL (MATCH-HOST T))
  (MULTIPLE-VALUE-BIND (W* W1)
    (SEND SELF :INTERNAL-WILD-CHARACTERS)
    (AND (OR (NOT MATCH-HOST) (EQ HOST (PATHNAME-HOST PATHNAME)))
       (PATHNAME-COMPONENT-MATCH DIRECTORY (PATHNAME-DIRECTORY PATHNAME) W* W1)
       (PATHNAME-COMPONENT-MATCH NAME (PATHNAME-NAME PATHNAME) W* W1)
       (OR (eq type :unspecific)
	   (pathname-component-match type (pathname-type pathname) w* w1)))))

;;; This could go away and if regular merging would put the device in the proper place.   3.08.88
(DEFMETHOD (MAC-PATHNAME-MIXIN :MERGE-DEVICE) ()
  (let ((default-device (send host :default-device)))
    (if (stringp device) self (send self :NEW-DEVICE default-device))))	;3.09.88

(DEFMETHOD (MAC-PATHNAME-MIXIN  :DIRECTORY-WILD-P) ()  ;03-18-88 DAB
  (SEND SELF :COMPONENT-CONTAINS-WILD DIRECTORY))


(DEFMETHOD (MAC-PATHNAME-MIXIN :QUIET-HOMEDIR) (&OPTIONAL (USER USER-ID))
  (SEND (MAKE-PATHNAME :HOST HOST :DEVICE (SEND host :DEFAULT-DEVICE))	;3.09.88
	;;; Default is for users to have their own directories.
     :NEW-SUGGESTED-DIRECTORY (COMPUTE-HOMEDIR-FROM-USER-ID USER HOST)))

(DEFFLAVOR MAC-PATHNAME () (MAC-PATHNAME-MIXIN RAW-COMPONENTS-ALWAYS-UP