;; -*- Mode:COMMON-LISP; Package:FS; Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985- 1989 Texas Instruments Incorporated. All rights reserved.
;;;  ** (c) Copyright 1980 Massachusetts Institute of Technology **

;;;
;;; REVISED: 
;;;     02-01-89 DAB    Fixed (NEW-UNIX-PATHNAME-MIXIN :NEW-TYPE) and (NEW-MSDOS-PATHNAME-MIXIN :NEW-TYPE)
;;;                     to use canonical-type in the input type is a symbol.
;;;      3/02/88 LAS    Added new-name-and-type method (moved from nfs;service_functions) 
;;;      1/28/88 MBC    Moved COMPILE-FLAVOR-METHODS forms to new COMPILE-OTHER-FLAVOR-METHODS file.
;;;     10-22-87 DAB    Added support for forward slash in ms-dos pathnames.
;;;	09.25.87 MBC	Name length validation incorrectly allowed name and type lengths
;;;			to exceed maximum length for Unix and Unix-Ucb pathnames.
;;;	06.16.87 MBC	APOLLO leading slashes support in :PARSE-NAMESTRING.
;;;	06.16.87 MBC	Clean up :DIRECTORY-LIST.
;;;	06.12.87 MBC	Parse "....foo" to have name = "....foo" not "...".
;;;	05.18.87 MBC	Don't aref "" string in unix-filename when type is :WILD.
;;;	05.04.87 MBC	DWIM fix for handling merging types, when given "*" instead of "*.*".
;;;	04.16.87 MBC	Make Probe-directory use direction :PROBE now that TCP-IP is
;;;			fixed and protect it from non-local exits.
;;; 			Force :DIRECTORY-LIST on UNIX pathname to make type wild, if
;;; 			pathname object's type is :UNSPECIFIC or NIL.
;;;	04.02.87 MBC	Use MSDOS-FILENAME instead of UNIX-FILENAME for msdos string-for-...s
;;;	03.13.87 MBC	Add :PROBE-DIRECTORY support.
;;;	02.27.87 MBC 	:DIRECTORY-PATHNAME-AS-FILE now returns second value, NOT-VALID, to
;;;			tell GET-LIST-OF-DIRECTORIES-1 that he can't expect to probe UNIX directory.
;;;     02.24.87 MBC    Make *MERGE-UNIX-TYPES* default to NIL.
;;;	01.28.87 MBC    Make Unix methods :PATHNAME-AS-DIRECTORY & :DIRECTORY-PATHNAME-AS-FILE
;;;			unique in handling Name+Type --> Directory-Node --> Name+Type conversions.
;;;	01.23.87 MBC	Print  for NIL types.  To make "T:name."'s namestring parse to its self.
;;;			Make parse of unix type use *MERGE-UNIX-TYPES* global consistently.
;;;			Add all Release 2 pathname flavor ALIAS-FLAVORs.
;;;	01.22.87 MBC	Fix fatal flaw in MSDOS defflavor.
;;;     11-12-85 MBC    Added :READ-INSTANCE methods for three host flavor types: UNIX, UNIX-UCB, & MSDOS.  
;;;	12.19.86 MBC	Fix parse of "T: " blank to give proper type.
;;;     12/19/86 HW     Allow UNIX name component to end in a dot
;;;     12/19/86 HW     Put compile-flavor-methods for non-cold loaded pathnames in here.

;;                 delimeter, EOS, previous delimeter, parsing-function


(DEFVAR UNIX-TABLE
   '(((#\/) NIL (#\SPACE) UNIX-PARSE-DIRECTORY T); Directory 
     (NIL T (#\/ #\SPACE) UNIX-PARSE-NAME-AND-TYPE)))       ; Name


(DEFVAR MSDOS-TABLE
   '(((#\:) NIL (#\SPACE) MSDOS-PARSE-DEVICE); Device
     ((#\\) NIL (#\: #\SPACE) UNIX-PARSE-DIRECTORY T); Directory 
     ((#\.) T (#\: #\\ #\SPACE) UNIX-PARSE-NAME); Name
     (NIL T (#\.) UNIX-PARSE-TYPE)))                ; Type, no Version


(DEFVAR MSDOS-TABLE-forward-slash  ;used when msdos switchchar is not /
   '(((#\:) NIL (#\SPACE) MSDOS-PARSE-DEVICE); Device
     ((#\/) NIL (#\: #\SPACE) UNIX-PARSE-DIRECTORY T)
     ((#\.) T (#\: #\/ #\SPACE) UNIX-PARSE-NAME); Name
     (NIL T (#\.) UNIX-PARSE-TYPE)))


;;;----------------        Unix Pathname Dependant Parsing Routines       ----------------

(DEFVAR *MERGE-UNIX-TYPES* NIL)  ;; non-nil means let merging do it's thing.
                              ;; nil means force types to :Unspecific if not provided

(DEFUN UNIX-PARSE-NAME-COMPONENT (STRING IGNORE &OPTIONAL IGNORE)
  (COND
    ((ZEROP (LENGTH STRING)) NIL)
    ((= (LENGTH STRING) 1.)
     (LET ((CHR (AREF STRING 0.)))
       (COND
	 ((CHAR= CHR #\*) :WILD)
	 ((CHAR= CHR #\) NIL)
	 ((CHAR= CHR #\) :UNSPECIFIC)
	 (T STRING))))
    (T STRING))) 


(DEFUN UNIX-NAME-VALIDATION (STRING VALID-LENGTH &OPTIONAL IGNORE)
  (WHEN (AND (STRINGP STRING) (> (LENGTH STRING) VALID-LENGTH))
    (SETQ STRING (NSUBSTRING STRING 0. VALID-LENGTH)))
  STRING)
    

(DEFUN MSDOS-PARSE-DEVICE (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (UNIX-NAME-VALIDATION (UNIX-PARSE-NAME-COMPONENT STRING SELF JUNK-ALLOWED)
			(SEND SELF :DEVICE-LENGTH) JUNK-ALLOWED))  

(DEFUN UNIX-PARSE-NAME-AND-TYPE (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (COND
    ((OR (NOT (STRINGP STRING)) (ZEROP (LENGTH STRING)))
     (VALUES NIL (IF *MERGE-UNIX-TYPES*	NIL :UNSPECIFIC)))	;NIL case still use global flag 1.23.87 MBC
    (T
     (LET ((PERIOD-POSITION			;Find right most period position
	     (POSITION #\. (THE STRING (STRING STRING)) :FROM-END T :TEST #'CHAR=))
	   TYP NAM NON-PERIOD-POSITION)
       (COND
	 ((OR (NULL PERIOD-POSITION) (ZEROP PERIOD-POSITION))	;No period or right most period is in position 0
	  (SETQ TYP (IF *MERGE-UNIX-TYPES* NIL :UNSPECIFIC)
		NAM (UNIX-PARSE-NAME STRING SELF JUNK-ALLOWED)))
	 ((= (1+ PERIOD-POSITION) (LENGTH STRING))	;Period is at right most position.
	  (SETQ TYP (IF *MERGE-UNIX-TYPES* NIL :UNSPECIFIC)
		NAM (UNIX-PARSE-NAME (NSUBSTRING STRING 0 (1+ PERIOD-POSITION)) SELF JUNK-ALLOWED)))
	 (T
	  (SETQ NON-PERIOD-POSITION		;6.12.87 MBC
		(POSITION #\. (THE STRING (STRING STRING)) :FROM-END T :TEST-NOT #'CHAR= :END PERIOD-POSITION))
	  (if non-period-position		;if there isn't another non-period character, type is NIL/:UNSPECIFIC
	      (SETQ TYP (UNIX-PARSE-TYPE (NSUBSTRING STRING (1+ PERIOD-POSITION)) SELF JUNK-ALLOWED)
		    NAM (UNIX-PARSE-NAME (NSUBSTRING STRING 0 PERIOD-POSITION) SELF JUNK-ALLOWED))
	      (SETQ TYP (IF *MERGE-UNIX-TYPES* NIL :UNSPECIFIC)
		    NAM (UNIX-PARSE-NAME STRING SELF JUNK-ALLOWED)))))
       (when (and (stringp nam) (stringp typ))
	 ;;; together name and type are limited to the name-length... But we don't want to
	 ;;; truncate either one too severely, so lets keep 9 letters minimum for name and
	 ;;; type will always have a minimum of four letters.   So... we're sometimes forced
	 ;;; to truncate BOTH name and type, but the distribution is kind of arbitrary.		9.21.87
	 (setf nam (unix-name-validation nam (max 9. (- (send self :NAME-LENGTH) (length typ) 1))))
	 (when (> (+ (length nam) (length typ)) (send self :name-length))
	   (setf typ  (unix-name-validation typ (- (send self :NAME-LENGTH) (length nam) 1)))))
       (VALUES NAM TYP)))))


(DEFUN UNIX-PARSE-NAME (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (LET ((NAM
	 (UNIX-NAME-VALIDATION (UNIX-PARSE-NAME-COMPONENT STRING SELF JUNK-ALLOWED)
			       (SEND SELF :NAME-LENGTH) JUNK-ALLOWED)))
    (IF (AND (STRINGP NAM) (<= (LENGTH NAM) 2.) (EQ (SEND SELF :SYSTEM-TYPE) :UNIX)
	(OR (STRING-EQUAL NAM ".") (STRING-EQUAL NAM "..")))
      ()
      NAM))) 

(DEFUN UNIX-PARSE-TYPE (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (UNIX-NAME-VALIDATION (UNIX-PARSE-NAME-COMPONENT STRING SELF JUNK-ALLOWED)
			(SEND SELF :TYPE-LENGTH) JUNK-ALLOWED))  


(DEFUN UNIX-PARSE-DIRECTORY-COMPONENT (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (WHEN (STRINGP STRING)
    (SETQ STRING (STRING-TRIM '(#\SPACE) STRING)))
  (COND
    ((STRING= STRING "*") :WILD)
    ((STRING= (subseq STRING 0 2) "..") :UP) ;02-08-88 DAB 
    ((ZEROP (LENGTH STRING))
     (IF JUNK-ALLOWED
       ()
       (FERROR 'PATHNAME-PARSE-ERROR "Null directory components not allowed ~S" STRING)))
    (T (UNIX-NAME-VALIDATION STRING (SEND SELF :NAME-LENGTH) JUNK-ALLOWED))))  


(Defun Unix-Skip-All-Directory-Delimiters (String Delim Start)
  (Loop For I From Start To (1- (length String))
	When (Char/= (Aref String I) Delim)
	Do (Return I)))


(DEFUN UNIX-PARSE-DIRECTORY (STRING SELF &OPTIONAL JUNK-ALLOWED)
  (LET ((START-INDEX 0)
	(END-INDEX 0)
	(RESULT (QUOTE NIL))
	(DELIM (SEND SELF :DIRECTORY-DELIMITER-CHARACTER)))
    (MULTIPLE-VALUE-SETQ (RESULT STRING)
			 (SEND SELF :PARSE-RELATIVE-DIRECTORY STRING DELIM))
    (WHEN STRING			
      (if (zerop (length string))		;(:relative) is equivalent to dir = NIL
	  (setf result nil)			;MBC 10-23-86
	  (SETQ START-INDEX (UNIX-SKIP-ALL-DIRECTORY-DELIMITERS STRING DELIM 0)
		END-INDEX START-INDEX)
	  (LOOP WITH I = START-INDEX DO
		(IF (NOT (CHAR= (AREF STRING I) DELIM))
		    (SETQ END-INDEX (1+ END-INDEX)
			  I (1+ I))
		    (LET ((COMP
			    (UNIX-PARSE-DIRECTORY-COMPONENT (NSUBSTRING STRING START-INDEX END-INDEX) SELF
							    JUNK-ALLOWED)))
		      (WHEN COMP
			(PUSH COMP RESULT))
		      (SETQ I (UNIX-SKIP-ALL-DIRECTORY-DELIMITERS STRING DELIM (1+ END-INDEX)))
		      (WHEN I
			(SETQ END-INDEX I
			      START-INDEX I))))
		WHILE I)))
    (IF (LISTP RESULT)
	(NREVERSE RESULT)
	RESULT))) 

(defun Unix-Parse-Relative-Pathname (string delim)
  (if (zerop (length string))
      (values NIL nil)
      (let ((first-char (aref string 0))
	    (length (length string)))
	(if (char= first-char delim)
	    (if (null (Unix-Skip-All-Directory-Delimiters string delim 1))
		(values :ROOT nil)
		(values nil (nsubstring string 1.)))
	    (if (>= length 2)
		(let ((second-char (aref string 1)))
		  (cond ((char= first-char #\.)
			 (cond ((char= second-char delim)
				(values (list :RELATIVE) (nsubstring string 2)))
			       ((char= second-char #\.)
				(if (>= length 3.)
				    
				    (cond ((char= (aref string 2) delim)
					   (values (list :RELATIVE) string))
					  ((char= (aref string 2) #\.)    ;02-08-88 DAB
					   (values (list :RELATIVE) string))
					  ( t (values nil string)))

				    
				    (values (list :RELATIVE) string)))
			       (t (values nil string))))
			(t (values (list :RELATIVE) string))))
		(values nil string))))))
  


;;;----------------        End Of UNIX Pathname Dependant Parsing Routines        ------------------


;;;; MS-DOS Support

;;;MS-DOS supports a switch char. Normally this is the forward slash, but it can be modified by the user.
;;;If the switch char is something other than the / the directory delimiter( normally \)  is automatically changed to 
;;; a forward slash. 

;;The variable below controls wheither the ms-dos host is using the backslash or the forward slash as the directory
;; delimiter. This variable must be set to the proper character before  pathname parsing  occurrs to the ms-dos 
;; host. If you try to parse a pathname using the opposite delimiter, you will get unexpected results ( truncated 
;; directory names). Once a pathname is parsed for a given host, the resulting pathname from future parsing will 
;; alway return the directory using that delimiter, no matter what msdos-directory-delimiter-character is set to.


(defvar msdos-directory-delimiter-character #\\
  "Valid entries are the backslash (#\\\\) and forward slash (#\\/) characters.")  ;10-22-87 DAB

(DEFFLAVOR NEW-MSDOS-PATHNAME-MIXIN ()
   (RAW-COMPONENTS-ALWAYS-UPPER-CASE-MIXIN PATHNAME-ALWAYS-UPPER-CASE-MIXIN)
   (:REQUIRED-FLAVORS PATHNAME))  


(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :NEW-DIRECTORY) (SPEC)
  (SEND SELF :NEW-RAW-DIRECTORY SPEC))  


(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :NEW-DEVICE) (SPEC)
  (SEND SELF :NEW-RAW-DEVICE SPEC))  


(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :NEW-NAME) (SPEC)
  (SEND SELF :NEW-RAW-NAME SPEC))  

(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :NEW-TYPE) (SPEC)
  (IF (SYMBOLP SPEC) 
      (SEND SELF :NEW-RAW-TYPE  ; DAB 02-01-89
	    (DECODE-CANONICAL-TYPE SPEC (SEND SELF :SYSTEM-TYPE)))
      (SEND SELF :NEW-RAW-TYPE  SPEC)
      )
  )


(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :DIRECTORY-DELIMITER-CHARACTER) ()  ;10-22-87 DAB
  (if (member msdos-directory-delimiter-character '(#\\ #\/))
      msdos-directory-delimiter-character
      #\\))


(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING &OPTIONAL (START 0.) END)
  (MULTIPLE-VALUE-BIND (DEV DIR NAM TYP LEN)
    (LM-PARSE-NAMESTRING NAMESTRING SELF
			 (if (eql #\\ msdos-directory-delimiter-character )
			     MSDOS-TABLE  MSDOS-TABLE-forward-slash) ;10-22-87 DAB
			 START END)
    (VALUES DEV DIR NAM TYP :UNSPECIFIC LEN)))

(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :PARSE-RELATIVE-DIRECTORY) (STRING DELIM)
  (UNIX-PARSE-RELATIVE-PATHNAME STRING DELIM))  


(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :NAME-LENGTH) ()
  8.)  

(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :TYPE-LENGTH) ()
  3.)  

(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :DEVICE-LENGTH) ()
  1.)  


;;; New fnc so MSDOS won't be affected by Unix weird TYPE particulars.  4.2.87
;;; i.e.  (send (pathname "ppc163:\\USR\\*.*") :string-for-host)
;;; should return "\\USR\\*.*"  NOT  "\\USR\\*".
(DEFUN MSDOS-FILENAME (NAME TYPE)
  (IF (EQ NAME :UNSPECIFIC) (SETQ NAME ""))
  (IF (EQ NAME :WILD)
      (IF (eq TYPE :WILD)
	  "*.*"
	  (if (MEMBER TYPE '(NIL :UNSPECIFIC))
	      "*"
	      (STRING-APPEND "*." TYPE)))
      (IF (AND (NULL NAME) (MEMBER TYPE '(NIL :UNSPECIFIC)))
	  ""
	  (OR NAME (SETQ NAME ""))
	  (COND ((EQ TYPE :WILD)
		 (FORMAT NIL "~A.*" NAME))
		((MEMBER TYPE '(NIL :UNSPECIFIC))
		 NAME)
		(T
		 (STRING-APPEND NAME "." TYPE))))))

(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :STRING-FOR-HOST) ()
  (LET ((DIR-STRING (UNIX-DIRECTORY-STRING)))
    (FORMAT () "~@[~A:~]~@[~A~]~A"
	    (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ)
	      ()
	      (SEND SELF :DEVICE))
	    (IF (STRINGP DIR-STRING)
	      (NSTRING-UPCASE DIR-STRING)
	      DIR-STRING)
	    (MSDOS-FILENAME (SEND SELF :NAME) (SEND SELF :TYPE)))))	;4.2.87


(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :STRING-FOR-EDITOR) ()
  (LET ((DIR-STRING (UNIX-DIRECTORY-STRING)))
    (FORMAT () "~A ~@[~A:~]~@[~A~] ~A:"
	    (MSDOS-FILENAME (SEND SELF :NAME) (SEND SELF :TYPE))	;4.2.87
	    (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ)
	      ()
	      (SEND SELF :DEVICE))
	    (IF (STRINGP DIR-STRING)
	      (NSTRING-UPCASE DIR-STRING)
	      DIR-STRING)
	    (SEND HOST :SHORT-NAME))))  


(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :STRING-FOR-DIRECTORY) ()
  (LET ((DIR-STRING (UNIX-DIRECTORY-STRING)))
    (FORMAT () "~@[~A:~]~@[~A~]"
	    (IF (MEMBER DEVICE '(:UNSPECIFIC NIL) :TEST #'EQ)
	      ()
	      (SEND SELF :DEVICE))
	    (IF (STRINGP DIR-STRING)
	      (NSTRING-UPCASE DIR-STRING)
	      DIR-STRING))))  

(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :STRING-FOR-DIRED) ()
  (MSDOS-FILENAME (SEND SELF :NAME) (SEND SELF :TYPE)))	;4.2.87

(DEFMETHOD (NEW-MSDOS-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :MSDOS))  

;;;; UNIX 4.2 support

(DEFFLAVOR NEW-UNIX-UCB-PATHNAME-MIXIN ((link-to nil)) ()
  (:REQUIRED-FLAVORS PATHNAME)
  (:settable-instance-variables)
  (:gettable-instance-variables))


(DEFMETHOD (NEW-UNIX-UCB-PATHNAME-MIXIN :NAME-LENGTH) ()
  (IF (MEMBER TYPE '(NIL :UNSPECIFIC :WILD) :TEST #'EQ)
      256.
    (- 255. (LENGTH TYPE))))			;9.25.87 MBC

(DEFMETHOD (NEW-UNIX-UCB-PATHNAME-MIXIN :TYPE-LENGTH) ()
  (IF (MEMBER NAME '(NIL :UNSPECIFIC :WILD) :TEST #'EQ)
    256.
    (- 255. (LENGTH NAME))))			; count the PERIOD ! - 9.25.87 MBC


(DEFMETHOD (NEW-UNIX-UCB-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :UNIX-UCB))  

;;;; UNIX support

(DEFFLAVOR NEW-UNIX-PATHNAME-MIXIN ((link-to nil)) (PATHNAME-NORMALLY-LOWERCASE-MIXIN)
   (:REQUIRED-FLAVORS PATHNAME)
   (:settable-instance-variables)
   (:gettable-instance-variables))

 

(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :NEW-NAME) (SPEC)	;9.25.87 MBC
  (SEND SELF :NEW-RAW-NAME (UNIX-NAME-VALIDATION SPEC (SEND SELF :NAME-LENGTH) NIL)))

(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :NEW-TYPE) (SPEC)   ;9.25.87 MBC
  (IF (SYMBOLP SPEC) 
      (SEND SELF :NEW-RAW-TYPE   ; DAB 02-01-89
	    (unix-NAME-VALIDATION (DECODE-CANONICAL-TYPE SPEC (SEND SELF :SYSTEM-TYPE))
				 (send self :type-length)))
      
      (SEND SELF :NEW-RAW-TYPE (UNIX-NAME-VALIDATION SPEC (SEND SELF :TYPE-LENGTH) NIL)))
  )

;;; Special APOLLO support to preserve extra leading slashes.  6.16.87
(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PARSE-NAMESTRING) (IGNORE NAMESTRING
							&OPTIONAL (START 0) END)
  (MULTIPLE-VALUE-BIND (DIR NAM TYP LEN)
      (LM-PARSE-NAMESTRING NAMESTRING SELF UNIX-TABLE START END)
    (when (and (eq :APOLLO (net:get-host-attribute host :machine-type))	;6.16.87
	       (> (or end (length namestring)) (+ 3 start))
	       (char= #\/ (char namestring start))
	       (char= #\/ (char namestring (1+ start))))
      (typecase dir
	(keyword (setf dir (list (concatenate 'string "/" nam))
		       nam nil))
	(cons    (setf (first dir) (concatenate 'string "/" (first dir))))
	(string  (setf dir (concatenate 'string "/" dir)))))
    (VALUES :UNSPECIFIC DIR NAM TYP :UNSPECIFIC LEN )))



(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PARSE-RELATIVE-DIRECTORY) (STRING DELIM)
  (UNIX-PARSE-RELATIVE-PATHNAME STRING DELIM))


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :CHARACTER-NEEDS-QUOTING-P) (IGNORE)
  ())  

(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :DIRECTORY-WILD-P) ()
  (SEND SELF :COMPONENT-CONTAINS-WILD DIRECTORY))

(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :UNSPECIFIC-TYPE-IS-DEFAULT) ()
  T)


;;; Given /dir/foo.bar return /dir ==> best way to probe a unix directory.
;;; Could lose on empty directories....  3.13.87 MBC
(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :DIRECTORY-PATHNAME-AS-FILE) (&optional IGNORE)
  (let (dir nam)
    (COND
      ((EQ DIRECTORY :ROOT)
       (FERROR 'PATHNAME-PARSE-ERROR "There is no pathname for the root as a file"))
      ((OR (ATOM DIRECTORY) (NULL (CDR DIRECTORY)))
       (SETQ DIR :ROOT
	     NAM (IF (CONSP DIRECTORY)
		     (CAR DIRECTORY)
		     DIRECTORY)))
      (T (LET ((LAST (LAST DIRECTORY)))
	   (SETQ DIR (LDIFF DIRECTORY LAST)
		 NAM (CAR LAST)))))
    (values
      (SEND SELF :NEW-PATHNAME
	    :RAW-DIRECTORY DIR :RAW-NAME NAM :TYPE :unspecific :VERSION :NEWEST))))


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :DIRECTORY-LIST) (OPTIONS)	;6.16.87
  (send host :file :directory-list self options))


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PROBE-DIRECTORY) ()	;4.16.87  
  (with-open-file (s (send self :directory-pathname-as-file) :direction :probe) s))


;;; Now pay attention to NAME/TYPE games that are unique to UNIX.  1.28.87 MBC
;;; When Type exist and is not a keyword use it to form the last directory node.
(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PATHNAME-AS-DIRECTORY) ()
  (SEND SELF :NEW-PATHNAME :RAW-DIRECTORY
	(IF (EQ DIRECTORY :ROOT)
	    (cond ((or (null type) (typep type 'keyword))
		   (cons Name Nil))
		  (T (cons (string-append NAME (string #\.) type) nil)))
	    (APPEND (IF (CONSP DIRECTORY)
			DIRECTORY
			(CONS DIRECTORY ()))
		    (cond ((or (null type) (typep type 'keyword))
			   (cons Name Nil))
			  (T (cons (string-append NAME (string #\.) type) nil)))))
	:NAME :UNSPECIFIC :TYPE :UNSPECIFIC :VERSION :UNSPECIFIC))


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :DIRECTORY-FILE-TYPE) ()
  :UNSPECIFIC)  


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :DIRECTORY-DELIMITER-CHARACTER) ()
  #\/)  


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :DIRECTORY-UP-DELIMITER) ()
  "..")  


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :STRING-FOR-HOST) ()
  (FORMAT () "~@[~A~]~A" (UNIX-DIRECTORY-STRING) (UNIX-FILENAME NAME TYPE)))


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :STRING-FOR-EDITOR) ()
  (FORMAT () "~A ~@[~A ~]~A:" (UNIX-FILENAME NAME TYPE) (UNIX-DIRECTORY-STRING)
	  (SEND HOST :NAME)))


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :STRING-FOR-DIRED) ()
  (UNIX-FILENAME NAME TYPE))  


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :STRING-FOR-DIRECTORY) ()
  (IF DIRECTORY
    (UNIX-DIRECTORY-STRING)
    ""))  


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :READ-INSTANCE) (IGNORE STREAM)
  (READ-PATHNAME-INSTANCE STREAM :UNIX))  


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :NEW-NAME-AND-TYPE) (spec)
  (multiple-value-bind (ignore ignore nam typ)
      (send self :parse-namestring nil spec)
    (send (send self :new-raw-name nam) :new-raw-type typ)))


(DEFUN UNIX-FILENAME (NAME TYPE &AUX (NEW-TYPE TYPE))
  (IF (EQ NAME :UNSPECIFIC)
      (SETQ NAME ""))
  (IF (EQ NAME :WILD)
      (IF (MEMBER TYPE '(:WILD :UNSPECIFIC NIL) :TEST #'EQ)
	  "*"					;Both wild, just *
	  (STRING-APPEND "*." NEW-TYPE))
      (IF (AND (NULL NAME) (MEMBER TYPE '(NIL :UNSPECIFIC) :TEST #'EQ))
	  ""
	  (PROGN
	    (OR NAME (SETQ NAME ""))
	    (COND
	      ((EQ TYPE :WILD)
	       (if (zerop (length name)) "*"	;5.18.87
		   (IF (AND 
			 (CHAR= #\* (AREF NAME 0.))
			 (EVERY #'(LAMBDA (X)
				    (CHAR= #\. X))
				(NSUBSTRING NAME 1.)))
		       (STRING-APPEND NAME "*")
		       (STRING-APPEND NAME ".*"))))
;	      ((MEMBER TYPE '(NIL :UNSPECIFIC) :TEST #'EQ) NAME)	;1.23.87 MBC (3 lines)
	      ((eq TYPE :UNSPECIFIC) NAME)
	      ((eq TYPE NIL) (string-append NAME "."))	
	      (T (STRING-APPEND NAME "." NEW-TYPE)))))))


(DEFUN UNIX-DIRECTORY-STRING ()
  (DECLARE (:SELF-FLAVOR NEW-UNIX-PATHNAME-MIXIN))
  (IF (MEMBER DIRECTORY '(NIL :UNSPECIFIC) :TEST #'EQ)
      ()
      (LET ((DIRECT DIRECTORY)
	    (SUPPRESS-DELIM NIL))
	(STRING-APPEND
	  (COND
	    ((EQ DIRECT :ROOT) "")
	    ((AND (CONSP DIRECT) (EQ (CAR DIRECT) :RELATIVE)) (POP DIRECT) "")
	    (T (SEND SELF :DIRECTORY-DELIMITER-CHARACTER)))
	  (COND
	    ((EQ DIRECT :ROOT) "")
	    ((ATOM DIRECT) (UNIX-DIRECTORY-COMPONENT DIRECT))
	    ((NULL (CDR DIRECT))
	     (LET (STRING)
	       (MULTIPLE-VALUE-SETQ (STRING SUPPRESS-DELIM)
				    (UNIX-DIRECTORY-COMPONENT (CAR DIRECT)))
	       STRING))
	    (T
	     (LOOP FOR SUBDIR IN DIRECT WITH STRING =
		   (MAKE-ARRAY 20. :ELEMENT-TYPE 'STRING-CHAR :LEADER-LIST '(0.)) AS DELIM-P = () THEN T DO
		   (AND DELIM-P (NOT SUPPRESS-DELIM)
			(VECTOR-PUSH-EXTEND (SEND SELF :DIRECTORY-DELIMITER-CHARACTER) STRING))
		   (LET (SUBSTR)
		     (MULTIPLE-VALUE-SETQ (SUBSTR SUPPRESS-DELIM)
					  (UNIX-DIRECTORY-COMPONENT SUBDIR))
		     (SETQ STRING (STRING-NCONC STRING SUBSTR)))
		   FINALLY (RETURN STRING))))
	  (COND
	    (SUPPRESS-DELIM "")
	    (T (SEND SELF :DIRECTORY-DELIMITER-CHARACTER)))))))  


(DEFUN UNIX-DIRECTORY-COMPONENT (STRING)
  (CASE STRING
    (:WILD "*")
    (:UP
     (LET ((DELIM (SEND SELF :DIRECTORY-UP-DELIMITER)))
       (IF (STRINGP DELIM)
	   DELIM
	   (VALUES (STRING DELIM) T))))
    (OTHERWISE (STRING STRING))))  


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PARSE-DIRECTORY-SPEC) (SPEC)
  (COND
    ((STRINGP SPEC) (LIST (SEND SELF :PARSE-COMPONENT-SPEC SPEC)))
    ((AND (CONSP SPEC)
	  (LOOP FOR ELT IN SPEC ALWAYS
		(OR (MEMBER ELT '(:UP :WILD :RELATIVE) :TEST #'EQ) (STRINGP ELT)))
	  (NOT (MEMBER :RELATIVE (CDR SPEC) :TEST #'EQ)))
     (LOOP FOR ELT IN SPEC COLLECT
	   (IF (SYMBOLP ELT)
	       ELT
	       (SEND SELF :PARSE-COMPONENT-SPEC ELT))))
    ((MEMBER SPEC '(NIL :UNSPECIFIC :WILD :ROOT) :TEST #'EQ) SPEC)
    (T (PATHNAME-DIRECTORY (QUIET-USER-HOMEDIR HOST)))))  

;; This completely prevents any Unix pathname from having a non-:UNSPECIFIC version.
(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PARSE-VERSION-SPEC) (IGNORE)
  :UNSPECIFIC)  


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :NAME-LENGTH) ()
  (IF (MEMBER TYPE '(NIL :UNSPECIFIC :WILD) :TEST #'EQ)
    14.
    (- 13. (LENGTH TYPE))))			;9.25.87 MBC

(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :TYPE-LENGTH) ()
  (IF (MEMBER NAME '(NIL :UNSPECIFIC :WILD) :TEST #'EQ)
    14.
    (- 13. (LENGTH NAME))))  ; count the PERIOD ! - 9.25.87 MBC

;;; Patch system interface, more kludges for only 14. character UNIX filenames

(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PATCH-FILE-PATHNAME) (NAM SAME-DIRECTORY-P PATOM TYP &REST ARGS)
  (LET ((SNAME (IF SAME-DIRECTORY-P
		 PATOM
		 (SI:SYSTEM-SHORT-NAME NAM))))
    (IF (>= (LENGTH SNAME) (SEND SELF :NAME-LENGTH))
      (SETQ SNAME (SUBSEQ SNAME 0. (SEND SELF :NAME-LENGTH))))
    (CASE TYP
      (:SYSTEM-DIRECTORY
       (SEND SELF :NEW-PATHNAME :NAME SNAME :TYPE :PATCH-DIRECTORY :VERSION :NEWEST))
      (:VERSION-DIRECTORY
       (SEND SELF :NEW-PATHNAME :NAME (FORMAT () "~A-~D" SNAME (GCD (CAR ARGS) 10000.)) :TYPE
	  :PATCH-DIRECTORY :VERSION :NEWEST))
      (:PATCH-FILE
       (SEND SELF :NEW-PATHNAME :NAME
	  (FORMAT () "~A-~D-~D" (SUBSEQ SNAME 0. (MIN 3. (LENGTH SNAME)))
		  (GCD (CAR ARGS) 1000.) (GCD (CADR ARGS) 1000.))
	  :TYPE (CADDR ARGS) :VERSION :NEWEST)))))  

;;;
;;;  Unix matching now checks for type, but still not for version.
;;;  Add type to :PATHNAME-MATCH :PATHNAME-MATCH-NO-DEVICE.
;;;  Use default methods for :PATHNAME-MATCH-NO-VERSION & :PATHNAME-MATCH-NO-VERSION-OR-DEVICE
;;;  - MBC 10-20-86
;;;; If type is :UNSPECIFIC allow the match.  This makes things work like a Unix user expects,
;;;  i.e. Copy of /dir/* should get all names and types.  Avoids /dir/*.* work around syntax.
;;;  - MBC 5.4.87
(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PATHNAME-MATCH) (PATHNAME &OPTIONAL (MATCH-HOST T))
  (MULTIPLE-VALUE-BIND (W* W1)
    (SEND SELF :INTERNAL-WILD-CHARACTERS)
    (AND (OR (NOT MATCH-HOST) (EQ HOST (PATHNAME-HOST PATHNAME)))
       (PATHNAME-COMPONENT-MATCH DEVICE (PATHNAME-DEVICE PATHNAME) W* W1)
       (PATHNAME-COMPONENT-MATCH DIRECTORY (PATHNAME-DIRECTORY PATHNAME) W* W1)
       (PATHNAME-COMPONENT-MATCH NAME (PATHNAME-NAME PATHNAME) W* W1)
       (if (eq type :unspecific)
	   T					;5.4.87
	   (pathname-component-match type (pathname-type pathname) w* w1)))))


(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :PATHNAME-MATCH-NO-DEVICE) (PATHNAME &OPTIONAL (MATCH-HOST T))
  (MULTIPLE-VALUE-BIND (W* W1)
    (SEND SELF :INTERNAL-WILD-CHARACTERS)
    (AND (OR (NOT MATCH-HOST) (EQ HOST (PATHNAME-HOST PATHNAME)))
       (PATHNAME-COMPONENT-MATCH DIRECTORY (PATHNAME-DIRECTORY PATHNAME) W* W1)
       (PATHNAME-COMPONENT-MATCH NAME (PATHNAME-NAME PATHNAME) W* W1)
       (if (eq type :unspecific)
	   T					;5.4.87
	   (pathname-component-match type (pathname-type pathname) w* w1)))))

(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :DWIM-ON-TARGET-TRANSLATION) () T)	;5.4.87

(DEFMETHOD (NEW-UNIX-PATHNAME-MIXIN :NEW-MAIL-PATHNAME) ()
  (SEND SELF :NEW-PATHNAME :DIRECTORY '("USR" "SPOOL" "MAIL") :NAME USER-ID :TYPE :UNSPECIFIC
     :VERSION :UNSPECIFIC))  

;;;
;;;
;;;
(DEFFLAVOR UNIX-PATHNAME () (NEW-UNIX-PATHNAME-MIXIN FILE-ACCESS-MIXIN REMOTE-PATHNAME))

(DEFFLAVOR UNIX-UCB-PATHNAME () (NEW-UNIX-UCB-PATHNAME-MIXIN NEW-UNIX-PATHNAME-MIXIN FILE-ACCESS-MIXIN REMOTE-PATHNAME))

(DEFFLAVOR MSDOS-PATHNAME () (NEW-MSDOS-PATHNAME-MIXIN UNIX-PATHNAME FILE-ACCESS-MIXIN REMOTE-PATHNAME))

;;; Alias flavors to handle Rel2 generated flavors occuring in text.  1.23.87 MBC
;;; There is a test, under Test.Io.Pathname; called ALIAS-FLAVORS.

(DEFFLAVOR REMOTE-LM-PATHNAME 	() 	(LISPM-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR LM-PATHNAME 		() 	(LISPM-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR REMOTE-LM-PATHNAME-FTP ()	(LISPM-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR UNIX-PATHNAME-FTP 	() 	(UNIX-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR UCB-PATHNAME-FTP 	() 	(UNIX-UCB-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR ITS-PATHNAME-FTP 	() 	(ITS-PATHNAME)  	:ALIAS-FLAVOR)
(DEFFLAVOR TOPS20-PATHNAME-FTP 	() 	(TOPS20-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR TENEX-PATHNAME-FTP 	() 	(TENEX-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR VMS-PATHNAME-FTP 	() 	(VMS-PATHNAME)  	:ALIAS-FLAVOR)
(DEFFLAVOR MULTICS-PATHNAME-FTP () 	(MULTICS-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR LMFS-PATHNAME-FTP 	() 	(LMFS-PATHNAME) 	:ALIAS-FLAVOR)
(DEFFLAVOR MSDOS-PATHNAME-FTP 	() 	(MSDOS-PATHNAME) 	:ALIAS-FLAVOR)
