;;; -*- Mode: common-LISP; Package: PRINTER; Base: 10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; This file defines print device:  Mac-printer.


;;; CHANGE HISTORY:
;;; 
;;;  Date      Author   Description
;;; -------------------------------------------------------------------------------------
;;; 04-05-89 DAB        Modified :print-text-file to accept font-list.
;;; 04-05-89 DAB        Fixed :start-new-page. It was losing the last line of every page.
;;; 01-24-89    DAB     Added support for controlling tab-width thru arge option or mode-line.
;;; 11-17-88 DAB        Fixed (basic-mac-printer :after :init). It was trying to set a window instance
;;;                     variable. If the request is for a foreign host, print-stream will be a NFS-character-buffer-stream.
;;; 10-17-88     ab     Mac-Printer-Types 5-2.
;;;                     Moved a bunch of stuff here from PRINTER; UTILITIES.
;;;                     Re-wrote mx-start-document, mx-finish-document to be composed of simpler commands.
;;;                     Wrote new command MX-GET-PR-INFO to get printer page rectangle dimension (= "screen size")
;;;                     Fixed line-wrap problems in :TYO-CHAR.  Changed several specialty methods in small ways to
;;;                     support this change (:TYO, :START-NEW-LINE, etc).
;;;                     Changed :PRINT-BITMAP to use smaller of array size or screen size.
;;; 11-16-88 DAB Added *MAC-unwind-protect-forms* to mx-open-document,mx-close-document, mx-open-printer and 
;;;              mx-close-printer to make sure mx-opens and mx-closes are balanced.
;;; 10-17-88 AB  Moved some of these here from PRINTER; UTILITIES and added some new ones.

(DEFPARAMETER *mx-printer-error-alist*
	      '((1 PRINTER-NOT-INITIALIZED)
                (2 MEMORY-NOT-AVAILABLE
		   (open-printer "Can't get memory for print-record during OPEN-PRINTER"))
		(3 SPOOLING-FAILED)
		(4 OPEN-FAILED
		   (open-printer "Call to PrOpen failed."))    ; (PrError 0 PrOpen)
		(5 CLOSE-FAILED)
		(6 PR-DEFAULT-FAILED)
		(7 PR-JOB-FAILED)
		(8 OPEN-DOC-FAILED)
		(9 CLOSE-DOC-FAILED)
		(10 OPEN-PAGE-FAILED)
		(11 CLOSE-PAGE-FAILED)
		(12 DRAFTBITS-FAILED)
		(13 PR-VALIDATE-FAILED)
		(14 PRINTER-OPEN
		    (open-printer "Printer already open during OPEN-PRINTER"))
		(15 PR-STYLE-FAILED)
		(16 OPEN-DOCUMENT-FAILED)
		))

(DEFUN check-printer-error (acb)
  (WHEN (ZEROP (add:error-code acb))
    (RETURN-FROM check-printer-error nil))
  ;; An error occurred.
  (LET ((error-entry (ASSOC (add:error-code acb) *mx-printer-error-alist*)))
    (COND ((NULL error-entry)
	   (FERROR nil "Unknown printer error: ~d during command ~d." (add:error-code acb)))
	  ((NULL (THIRD error-entry))
	   ;; Just use symbol.
	   (FERROR nil "Printer error ~d:  ~a"
		   (add:error-code acb) (STRING (SECOND error-entry))))
	  ((OR (SYMBOLP (THIRD error-entry)) (STRINGP (THIRD error-entry)))
	   (FERROR nil "Printer error ~d: ~a" (STRING (THIRD error-entry))))
	  (t
	   ;; Should be a list of opcode symbols followed by format string and
	   ;; ACB parms numbers to be used as format args.
	   (LOOP for el in (CDDR error-entry) do 
		 (WHEN (= (add:subopcode acb) (SYMBOL-VALUE (FIRST el)))
		   (FERROR nil (APPLY 'FORMAT nil (SECOND el) nil)))))))
  )

(defvar *printer-channel* ())
(defvar *printer-opcode* 22.)			       ;on TV channel.
(DEFVAR *printer-cmd-size* 10.)			       ;# 16-b parms

(si:DefAlternate Printer-Subops (si:Q-corresponding-variable-lists
				  si:System-Constant-Lists)
  (Start-Document 1
   Page-Eject 2
   Finish-Document 3
   Open-Printer 4
   Close-Printer 5
   Open-Document 6
   Close-Document 7
   Get-Printer-Info 8
   Style-Dialog-Needed 9
   Job-Dialog-Needed 10
   Do-Style-Dialog 11
   Do-Job-Dialog 12
   )
  )

(DEFUN mx-do-style  ()
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) Do-Style-Dialog)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb))
      (add:return-acb-fast acb t))))

(DEFUN mx-do-job  ()
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) Do-Job-Dialog)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb))
      (add:return-acb-fast acb t))))

(DEFUN mx-style-needed-p  ()
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) Style-Dialog-Needed)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb)
	  (NOT (ZEROP (add:parm-16b acb 0))))
      (add:return-acb-fast acb t))))

(DEFUN mx-job-needed-p  ()
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) job-dialog-needed)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb)
	  (NOT (ZEROP (add:parm-16b acb 0))))
      (add:return-acb-fast acb t))))

;;; *MAC-unwind-protect-forms* will be used to make sure mx-opens and mx-closes are balanced. IN another words
;;; if mx-open-printer was done a mx-close-printer must be done before anothe open, otherwise the mac really
;;; get screwed up. Unwind protects have been added to print-file-and-wait, print-bitmap-and-wait, :print-file and
;;; :print-array to make sure this happen even when we get a error during printing.

(DEFUN mx-open-printer  ()
  (declare (special *MAC-unwind-protect-forms*))
  (when  *MAC-unwind-protect-forms* ;is the mac printer window is use, if so,wait.
    (process-wait "Waiting for MAC printer"
		  #'(lambda () (not *MAC-unwind-protect-forms* ))))
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) open-printer)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb)
	  (push 'mx-close-printer *MAC-unwind-protect-forms*))  ;11-16-88 DAB
      (add:return-acb-fast acb t))))

(DEFUN mx-close-printer  ()
  (declare (special *MAC-unwind-protect-forms*))
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) close-printer)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb)
	  (setf *MAC-unwind-protect-forms* (remove 'mx-close-printer *MAC-unwind-protect-forms* )) ;11-16-88 DAB
	  )
      (add:return-acb-fast acb t))))

(DEFUN mx-open-document  ()
  (declare (special *MAC-unwind-protect-forms*))
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) open-document)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb)
	  (push 'mx-close-document *MAC-unwind-protect-forms*))  ;11-16-88 DAB
      (add:return-acb-fast acb t))))

(DEFUN mx-close-document  ()
  (declare (special *MAC-unwind-protect-forms*))
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) close-document)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb)
	  (setf *MAC-unwind-protect-forms* (remove 'mx-close-document *MAC-unwind-protect-forms* )) ;11-14-88 DAB
	  )
      (add:return-acb-fast acb t))))


(DEFUN mx-get-pr-info  ()
  (DECLARE (values vertical-dots-per-inch horizontal-dots-per-inch page-width-in-pixels page-height-in-pixels))
  (setf *printer-channel* (add:find-channel 'display-io))
  (let (acb);
    (unwind-protect
	(progn
 	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) get-printer-info)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb)
	  (VALUES (add:parm-16b acb 0) (add:parm-16b acb 1) (add:parm-16b acb 2) (add:parm-16b acb 3)))
      (add:return-acb-fast acb t))))

(defun mx-page-eject ()
  (tv:dump-draw-char-cache)
  (let (acb)
    (unwind-protect
	(progn
	  (setf acb (add:get-acb-fast *printer-cmd-size*))
	  (setf (add:opcode acb) *printer-opcode*)
	  (setf (add:subopcode acb) page-eject)
	  (setf (add:requestor-complete acb) nil)
	  (add:transmit-packet-and-wait acb *printer-channel*)
	  (check-printer-error acb))
      (add:return-acb-fast acb t))))


(DEFUN new-mx-start-document ()
  (mx-open-printer)
  (WHEN (mx-style-needed-p)
    (mx-do-style))
  (WHEN (mx-job-needed-p)
    (mx-do-job))
  (mx-open-document)
  )

(DEFUN new-mx-finish-document ()
  (mx-close-document)
  (mx-close-printer))

;;ab 10-17-88.  End of stuff moved here from PRINTER; UTILITIES.


(DEFFLAVOR BASIC-mac-printer
	   (file-font-list
	    (LINE-BUFFER (MAKE-ARRAY 1440 ':LEADER-LIST (LIST 0) ':TYPE 'ART-STRING))  ;used in :print-bitmap to remote host.
	    )
	   (basic-printer)
  :initTABLE-INSTANCE-VARIABLES
  :setTABLE-INSTANCE-VARIABLES
  :getTABLE-INSTANCE-VARIABLES)
 
(DEFMETHOD (BASIC-mac-printer :CURRENT-PRINTER-STATE) (STREAM-TYPE)
  (declare (ignore stream-type))
  '(:ONLINE T :FAULT NIL :PAPER-OUT NIL)
  )


(defmethod (basic-mac-printer :before :start-document) (&rest ignore)
  (declare (special *mac-printer-window* ))
  (setf (send printer-stream :more-p ) NIL)
  (array-initialize (send *mac-printer-window* :buffer) 0)
  (send *mac-printer-window* :home-cursor)
  (SETQ CHARS-PER-LINE (* 7.5 CHAR-PITCH))
  )

(DEFMETHOD (basic-mac-printer :after :init) (&rest ignore)
  (declare (special *mac-printer-window* ))
  (when (and *mac-printer-window*   ;09-21-88 DAB
	     (typep printer-stream 'tv:mac-screen-for-printer)) ;11-17-88 DAB Could be a nfs-buffer-stream if remote printing.
    (SETF (tv:sheet-erase-aluf printer-stream) tv:alu-setz))
  )

(defmethod (basic-mac-printer :after :end-document) (&rest ignore)
  )

(DEFMETHOD (basic-mac-printer :PRINT-TEXT-FILE) (FILE-STREAM &OPTIONAL font-list)
  "Print an ascii text file."
  (when (and font-list   ; DAB 04-05-89 Hanlde font-list, it may be an atom or a list.
	     (atom font-list)) (setf font-list (list font-list)))
  (setq file-font-list (or font-list (GET-FONT-ATTRIBUTE-LIST FILE-STREAM)))  ; DAB 04-05-89
  (when file-font-list
    (send printer-stream :set-font-map file-font-list)
    (send printer-stream :set-current-font (car file-font-list)))
  (let ((PREVIOUS-FONT (send printer-stream :current-font)))
    (LOOP FOR IN-CHAR = (TYI FILE-STREAM ())
	  UNTIL (NULL IN-CHAR)
	  DO (cond ((= in-CHAR #\)
		    (setq in-char (tyi file-STREAM ()))
		    (COND
		      ((<= #\0 in-CHAR #\9)
		       (setq previous-font (send printer-stream :current-font))
		       (send printer-stream :set-current-font
			     (or (nth (read-from-string (string (int-char in-char))) file-font-list)
				 (car file-font-list))))
		      ((= IN-CHAR #\*)
		       (let ((temp-font (send printer-stream :current-font)))
			 (send printer-stream :set-current-font previous-font)
			 (setq previous-font temp-font))) 
		      (T
		       ;; just print <epsilon> and character following it
		       (SEND SELF :TYO-CHAR #\)
		       (SEND SELF :TYO-CHAR in-CHAR))))
		   (t (SEND SELF :TYO-CHAR IN-CHAR)))))

  )



;;ab 10-17-88.  Start changes.

(DEFUN make-mac-printer-window (height rounded-width)
  (DECLARE (SPECIAL  *mac-printer-window* tv:*undisplaced-Mac-window-arrays*
		     tv:*printer-id* tv:*Mac-resident-Explorer-screens* ))
  (setf *mac-printer-window*
	(tv:define-screen
	  'tv:Mac-screen-for-printer
	  (INTERN (FORMAT nil "MAC-SCREEN-for-printer-~3,'0D"
			  tv:*printer-id*)
		  'Mac)
	  :buffer (MAKE-ARRAY `(,height ,rounded-width)
			      :element-type '(MOD 2))
	  :bits-per-pixel 1.
	  :height height
	  :width rounded-width))
  (PUSH (CONS (SEND   *mac-printer-window* :buffer)
	      *mac-printer-window*)
	tv:*undisplaced-Mac-window-arrays*)
  (SETF (AREF tv:*Mac-resident-Explorer-screens* tv:*printer-id*)
	(tv:make-explorer-screen :the-screen   *mac-printer-window*)))

(DEFUN make-mac-printer-stream ()
  (declare (special *mac-printer-window*))
  (new-mx-start-document)
  (let (height width buf
	(tv:mac-explorer-screen-ID-for-screen-being-created 255.)
	(tv:not-visible-p T))
    (declare (special tv:mac-explorer-screen-ID-for-screen-being-created
		      tv:not-visible-p))
    
    ;; Find out dimensions of print this time.
    (MULTIPLE-VALUE-SETQ (nil nil height width)
      (mx-get-pr-info))
    
    (COND ((null *mac-printer-window*)
	   (make-mac-printer-window height width))
	  (t					       ;already exists.  Adjust size
	   (UNLESS (AND (= width (tv:sheet-width *mac-printer-window*))
			(= height (tv:sheet-height *mac-printer-window*)))
	     (SETF (si:symeval-in-instance *mac-printer-window* 'tv:width) width)
	     (SETF (si:symeval-in-instance *mac-printer-window* 'tv:height) height)
	     (SETF buf (MAKE-ARRAY (LIST height width) :element-type 'bit))
	     (SETF (si:symeval-in-instance *mac-printer-window* 'tv:screen-array) buf
		   (si:symeval-in-instance *mac-printer-window* 'tv:old-screen-array) buf)
	     (tv:remember-bit-array *mac-printer-window* buf)
	     (SETF buf (MAKE-ARRAY (LIST height width) :element-type 'bit))
	     (SETF (si:symeval-in-instance *mac-printer-window* 'tv:buffer) buf)
	     (tv:remember-bit-array *mac-printer-window* buf))
	   ))
    
    (setf (send *mac-printer-window* :more-p ) NIL)
    (array-initialize (send *mac-printer-window* :buffer) 0)
    (send *mac-printer-window* :home-cursor)
    *mac-printer-window*))


(DEFMETHOD (BASIC-mac-printer :before :START-NEW-PAGE) 
	   (&rest ignore)
  nil)

(DEFMETHOD (BASIC-mac-PRINTER :TYO-RAW) (CHAR)
  (SEND PRINTER-STREAM :TYO CHAR)) 

(DEFMETHOD (BASIC-mac-PRINTER :STRING-OUT-RAW) (STRING)
  (LOOP FOR X FROM 0 BELOW (LENGTH STRING) DO (SEND PRINTER-STREAM :TYO (AREF STRING X)))) 

(DEFMETHOD (basic-mac-printer :FORM) ()
  (SEND self :start-new-page)
  )

(DEFMETHOD (basic-mac-printer :CR) ()
  (SEND SELF :start-new-line))

(DEFMETHOD (basic-mac-printer :newline) ()
  (SEND self :start-new-line))

(DEFMETHOD (basic-mac-printer :linefeed) ()
  (SEND self :start-new-line))

(DEFMETHOD (basic-mac-printer :TAB) ()
  (unless (si:integerp tab-width) (setf tab-width *default-tab-width*)) ; DAB 01-24-89
  (DOTIMES (i tab-width) (SEND self :tyo-char #\space))  ; DAB 01-24-89
  )

(DEFMETHOD (basic-mac-printer :OVERSTRIKE) (&REST chars-list)  
  (LOOP INITIALLY (SEND SELF :TYO-char (CAR CHARS-LIST)) FOR CHAR IN (CDR CHARS-LIST) DO
      (SEND SELF :TYO-char BACKSPACE) (SEND SELF :TYO-char CHAR))
  )

(DEFMETHOD (basic-mac-printer :start-new-page) ()
  (mx-page-eject)
  (send printer-stream :home-cursor)
  )

(DEFMETHOD (basic-mac-printer :START-NEW-LINE) ()
  (tv:dump-draw-char-cache)
  (SETF (tv:SHEET-CURSOR-X printer-stream) 0)
  (INCF (tv:sheet-cursor-y printer-stream) (tv:sheet-line-height printer-stream))
  (WHEN (> (+ (tv:sheet-cursor-y printer-stream)
	      (tv:sheet-line-height printer-stream))  ; DAB 04-05-89 Adjust 1 one height down, it was losing the last line.
	   (tv:sheet-height printer-stream))
    (SEND self :start-new-page))
  )


(DEFMETHOD (basic-mac-printer :string-out) (STRING)
  (DOTIMES (i (LENGTH (THE string string)))
    (SEND self :tyo-char (AREF string i))))
			      
(DEFMETHOD (basic-mac-printer :TYO-CHAR) (CHAR)
  "Before sending out a character, look for special characters"
  (COND
    ((= CHAR #\TAB) (SEND SELF :TAB))
    ((= CHAR #\PAGE) (SEND SELF :START-NEW-PAGE))
    ;; on Lisp Machine CR (RETURN) really means CR-LF
    ((= CHAR #\NEWLINE) (SEND SELF :START-NEW-LINE))
    ((= CHAR #\LINEFEED) (SEND SELF :START-NEW-LINE))
    ;; ^M (which means CR ONLY on 2060/VAX) gets interpreted on Lisp Machines as octal 15,
    ;; but #o15 is the same as #\circle-plus on Lisp Machines; can't have both
;;    ((= CHAR #o15) (SEND SELF :CR))
    ((= CHAR #\BACKSPACE) (SEND SELF :TYO-RAW BACKSPACE))
    ((AND (>= CHAR #o40) (< CHAR #o200)); chars space, 0-9, a-z, A-Z, etc.
     (LET* ((font (tv:sheet-current-font printer-stream))
	    (CHAR (CHAR-CODE CHAR))
	    (WIDTH (SEND printer-stream :character-width char font))
	    (XPOS (tv:SHEET-CURSOR-X printer-stream))
	    (RIGHT-LIM (tv:SHEET-width printer-stream)))
       (WHEN (> (+ XPOS WIDTH) RIGHT-LIM) (SEND self :start-new-line))
       (SEND self :tyo-raw char)))
    (T
     ;; chars like down-arrow, epsilon, <ABORT>, etc.
     (LET* ((font (tv:sheet-current-font printer-stream))
	    (CHAR (CHAR-CODE CHAR))
	    (WIDTH (SEND printer-stream :character-width char font))
	    (XPOS (tv:SHEET-CURSOR-X printer-stream))
	    (RIGHT-LIM (tv:SHEET-width printer-stream)))
       (WHEN (> (+ XPOS WIDTH) RIGHT-LIM) (SEND self :start-new-line))
       (SEND self :tyo-raw char))))
  )


(DEFMETHOD (basic-mac-printer :END-DOCUMENT) (&REST IGNORE)
  (new-mx-finish-document))

;;ab 10-17-88 End changes.

(DEFMETHOD (basic-mac-printer :PRINT-HEADER-PAGE) (FILENAME USERNAME)
  "Prints filename, username, and current date on one page."
  (SEND SELF :STRING-OUT-CHARS
     (FORMAT () "~A~%~A~%~%    ~A~@[~%    created: ~A~]~%" USERNAME CURRENT-TIME FILENAME
	     (IF (TYPEP FILENAME 'PATHNAME)
	       (TIME:PRINT-UNIVERSAL-TIME (GET (SEND FILENAME :PROPERTIES) :CREATION-DATE) ()))))
  )

(DEFMETHOD (basic-mac-printer :SELECT-FONT-MODULE) (&rest ignore)
  ())

(DEFMETHOD (basic-mac-printer :SET-CHAR-PITCH) (&rest ignore)
  ())


(DEFMETHOD (BASIC-mac-printer :PRINT-RAW-FILE) (FILE-STREAM)
  "Files already in raw form, rebuilds the screen image array. Then it uses BITBLT to paint the
   screen to the printer."
  ;;; File-stream contains 8 bit bytes. Each byte is built from 8 consecutive bits of *full-screen-array*.
  ;The first 4 bytes of the file is the heigth, the next four is the width (in bits).
  (let ((width  (make-string 4 :initial-element #\space))
	(height (make-string 4 :initial-element #\space))
	screen-array hp wp)
    (send file-stream :string-in nil height)   
    (setf height  (read-from-string  height))  ; returns the height in bits. 
    (setf hp (ceiling height 8))  ;Number of bytes
    (send file-stream :string-in nil width)
    (setf width  (read-from-string width))  ;return the width in bits.
    (setf wp (ceiling width 8))   ;number of bytes
    (setf screen-array
	  (make-array (list (* 8 hp)
			    (* 32 (ceiling width 32)))  ;2nd dimension must by in multiple of 32 for BITBLT
		      :element-type '(mod 2)))
   (do ((h 0 (1+ h))) ;rebuild array in height - width order.
       ((= h height))
     (do ((w 0 (1+ w))
	  (bit-w 0 (+ 8 bit-w))
	  (in-char (send file-stream :tyi) (send file-stream :tyi)))
	  ((or (= w wp)
 	       (null in-char)))
       (dotimes (bit 8) (setf (aref screen-array h (+ bit bit-w))
			      (ldb (byte 1 (- 7 bit)) in-char))))
     )
   (let (mac-screen-array)
     (declare (special *mac-printer-window* ))
     (SETQ mac-screen-array (send *mac-printer-window* :buffer))
     (BITBLT W:ALU-SETA  WIDTH HEIGHT
	     screen-array 0 0
	     mac-screen-array  0 0))
   ))


(DEFMETHOD (BASIC-mac-printer :PRINT-BITMAP) (BITMAP-ARRAY &OPTIONAL WIDTH HEIGHT
					      (START-X 0) (START-Y 0)
					      ORIENTATION DOTS-PER-INCH &AUX LANDSCAPEP)
  "Copy the bitmap array to a print window if local printing.
   Otherwise copy the bitmap to a file stream. It assumes 8 bit data transfer"
  (DECLARE (IGNORE ORIENTATION DOTS-PER-INCH )(SPECIAL *mac-printer-window* ))
  (UNLESS WIDTH
    (SETQ WIDTH (array-dimension BITMAP-ARRAY 1)))
  (UNLESS HEIGHT
    (SETQ HEIGHT (array-dimension BITMAP-ARRAY 0)))
  ;;*mac-printer-window is a special window created on the mac to handle printing to a local printer. Its ID is 255.
  (if  (and *mac-printer-window*  ;is this local?
	    (typep printer-stream 'tv:mac-screen-for-printer)) ;11-17-88 DAB
      (let (mac-screen-array)  
	(SETQ mac-screen-array (send *mac-printer-window* :buffer))    
	(BITBLT W:ALU-SETA (MIN WIDTH (SEND *mac-printer-window* :width))      ;ab 10-17-88
		(MIN HEIGHT (SEND *mac-printer-window* :height))
		bitmap-array 0 0
		mac-screen-array  0 0))
      ;else printing is remote
      (send self :copy-array-to-xld-file BITMAP-ARRAY WIDTH HEIGHT START-X START-Y  LANDSCAPEP)
      ) ;if 
  T)

(DEFMETHOD (BASIC-mac-printer :copy-array-to-xld-file) (BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP)
  "Copy the bitmap to a file. The bits are packed in height - width order(i.e (0 1) (0 2) ..)."
  ;;We will need the height and width to recreate the array in :print-raw-file.
  (send printer-stream :string-out (format nil "~4d~4d" HEIGHT WIDTH))
  (let ((real-string-lenght (1+ (ceiling width 8))))
    (LOOP FOR Y FROM START-Y BELOW HEIGHT BY 1 DO
	  ;; print a row of chars,
	  ;; each char is 8 bits: <y>, <y>+1, ... <y>+7
	  (SEND SELF :GET-LINE BITMAP-ARRAY WIDTH HEIGHT START-X Y LANDSCAPEP)
	  (SEND PRINTER-STREAM :STRING-OUT LINE-BUFFER 0 real-string-lenght))))



(DEFMETHOD (BASIC-mac-printer :GET-LINE) (BITMAP-ARRAY WIDTH HEIGHT START-X FROM-Y LANDSCAPEP)
  "Get a line image of bitmap array from horizontal bit positions, START-X to WIDTH, starting at FROM-Y by 1 bit.
 The result of this function is characters being pushed on line-buffer. Returns NIL."
    (declare (ignore  LANDSCAPEP height))
    (STORE-ARRAY-LEADER 0 LINE-BUFFER 0) ;clear line-buffer
    (LOOP FOR X FROM START-X BELOW WIDTH BY 8 AS CHAR = 0 DO    ;pack 8 bits at a time.
       (LOOP FOR BIT-POS-OF-CHAR FROM
	     7 DOWNTO 0                                         ; always pack full bytes.
	  DO
	  ;; sets each bit of CHAR to an array entry's value
	  ;; which is always either 0 or 1.
	  ;; recall (dpb <byte> <ppss> <num>) puts the
	  ;; <ss> bit(s) of <byte> in <num> starting at bit <pp>
	  (SETQ CHAR
		(DPB
		 (COND
		   ((>= (+ x (- 7 bit-pos-of-char)) width) 0)  ;are we pass the left margin?
		   (T (AREF BITMAP-ARRAY from-y (+ x (- 7 bit-pos-of-char)))))
		 (BYTE 1 BIT-POS-OF-CHAR)    
		 CHAR)))
       DO (VECTOR-PUSH CHAR LINE-BUFFER)))

(DEFMETHOD (BASIC-mac-printer :SCREEN-IMAGE-FILE-P) (FILENAME)
  (LET ((FILE-TYPE (SEND FILENAME :TYPE)))
    (EQUALP "MAC-PRINTER" FILE-TYPE)) )


(DEFFLAVOR mac-printer
	   ()
	   (BASIC-mac-printer)
  :SETTABLE-INSTANCE-VARIABLES)


(COMPILE-FLAVOR-METHODS  BASIC-mac-printer mac-printer)

