;;; -*- Mode: common-LISP; Package: PRINTER; Fonts:(Cptfont Hl12b hl12i hl12i medfnb); Base: 10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1987-1989 Texas Instruments Incorporated. All rights reserved.

;;;Overview:
;;;PostScript printers are implemented differently than other Explorer printers. The basic outline is this:
;;;The ExplorerPS font mappings are contained in the variable *explorer-postscript-font-map*. Users may specify
;;; a different mapping for all spooled print requests by editting this variable. A user may specify a different mapping for 
;;; all print requests from his machine by defining a different font mapping in the variable *default-ps-font-map*. 
;;; A different mapping for a specific print request can be made by supplying a :ps-font-map keyword argument
;;; to print-file. If some or all of those variable are defined and there are duplications, the precedence (in decreasing order)
;;; of the mappings to use are (1) :ps-font-map keyword, (2) *default-ps-font-map* variable, (3) *explorer-postscript-font-map*.
;;;
;;;If the first 2 characters of the file to be printed are "%!", that file is assumed to already contain postscript page
;;; descriptions and it is simply dumped to the printer without further ado. Otherwise, it is assumed to just be a generic text file.
;;; In that case, a postscript source code driver must be downloaded to the printer. If this driver file does not currenly exist on the 
;;; print server, it is copied to the print server from sys:printer; .  Once the download is complete then the
;;; text data is dumped to the printer.
;;;
;;;Debugging help: PS printers are harder to debug than others because you must look at both the Lisp code and the 
;;; PostScript source code to figure out what's going on. Debugging the Lisp code is similar to any other. Info on debugging 
;;; PostScript source code is in the variable *postscript-driver*. It is done this way to add in patching. The 
;;; function build-postscript-source-variable uses this variable to create a memory image of the driver. This
;;; image array is downloaded to the printer before each print request.

;;; Revision history:
;;; 12-13-88  DAB Fix make-current-list. It was not handling variables in the font list correctly.
;;; 12-13-88 DAB  Handle the error that occurs when we send to many character that we overflow the 255 readline buffer in the PS driver. [8528,7548,9002]. Also fixes Check-printer-state for postscript. [8470]
;;; *********** Rel 5.0 12-12-88 DAB[
;;; 01-24-89 DAB  Added TAB-width support. Tab-width can be specified by including :tab-width as an argument in print-file
;;;               ,or include tab-width in the mode line of the file or use default *default-tab-width*.
;;; 9-29-88   INT  Support of international, and fix for the epsilon character that was printed twice.
;;; 07-18-88 DAb Fixed (postscript-printer :print-bitmap) to handle term-q on a microExplorer.
;;; 05-19-88 DAB Fixed postscript driver file to handle pathnames greater than 80 characters
;;; 4-7-88   DAB Fixed make-current-list to  handle cases :font "bigfnt" and :font-list '("bigfnt").
;;; 1-5-88   DAB Converted the file postscript-driver.text to a defvar to allow patching.
;;; 10-22-87 DAB Changed build-postscript-source-variable to blank suppress and remove tab and comments to the 
;;;              *postscript-program-in-memory. Also changed :write-postscript-program to use new length.
;;; 09-23-87 DAB Fixed Print-wide option and changed the way postscript-driver is loaded.
;;; 08/15/87  HW  Original.


(DEFVAR *Explorer-PostScript-font-map*
	'(
	  ("5X5" ("Courier" 5))
	  ("BIGFNT" ("Helvetica" 15 ))
	  ("CMB8" ("Times-Bold" 22))
	  ("CMDUNH" ("Times-Roman" 28))
	  ("CMOLD" ("Times-Bold" 28))
	  ("CMR10" ("Times-Roman" 16))
	  ("CMR18" ("Times-Bold" 32))
	  ("COURIER" ("Courier" 11))
	  ("CPTFONT" ("Courier" 10))
	  ("CPTFONTB" ("Courier-Bold" 10))
	  ("CPTFONTBI" ("Courier-BoldOblique" 10))
	  ("CPTFONTI" ("Courier-Oblique" 10))
	  ("DEFAULT" ("Courier" 10))
	  ("HIGHER-MEDFNB" ("Courier-Bold" (10 14)))
	  ("HIGHER-TR8" ("Times-Roman" (8 10)))
	  ("HL10" ("Helvetica" 9))
	  ("HL10B" ("Helvetica-Bold" 9))
	  ("HL12" ("Helvetica" 10))
	  ("HL12B" ("Helvetica-Bold" 10))
	  ("HL12BI" ("Helvetica-BoldOblique" 10))
	  ("HL12I" ("Helvetica-Oblique" 10))
	  ("HL6" ("Helvetica" 6))
	  ("HL7" ("Helvetica" 7))
	  ("MEDFNB" ("Courier-Bold" (11 10)))
	  ("MEDFNT" ("Courier" (11 10)))
	  ("MEDFNTB" ("Courier-Bold" (11 10)))
	  ("MEDFNTBI" ("Courier-BoldOblique" (11 10)))
	  ("MEDFNTI" ("Courier-Oblique" (11 10)))
	  ("METS" ("Times-Roman" 16))
	  ("METSB" ("Times-Bold" 16))
	  ("METSBI" ("Times-BoldItalic" 16))
	  ("METSI" ("Times-Italic" 16))
	  ("TINY" ("Helvetica" 5))
	  ("TR10" ("Times-Roman" 9))
	  ("TR10B" ("Times-Bold" 9))
	  ("TR10BI" ("Times-BoldItalic" 9))
	  ("TR10I" ("Times-Italic" 9))
	  ("TR12" ("Times-Roman" 10))
	  ("TR12B" ("Times-Bold" 10))
	  ("TR12BI" ("Times-BoldItalic" 10))
	  ("TR12I" ("Times-Italic" 10))
	  ("TR18" ("Times-Roman" 18))
	  ("TR18B" ("Times-Bold" 18))
	  ("TR8" ("Times-Roman" 8))
	  ("TR8B" ("Times-Bold" 8))
	  ("TR8I" ("Times-Italic" 8))
	  ("TVFONT" ("Courier" 7))
	  ("WIDER-FONT" ("Helvetica" (18 16)))
	  ("WIDER-MEDFNT" ("Courier" (14 10)))
	  ))



(DEFFLAVOR PostScript-Printer
	   ((attribute-line)           ;9-29-88 INT:  t if the file has an attribute line DAB
	    (font-list nil)	       ;List with Explorer to Postscript mappings
	    (File-font-list nil)       ;list with the fonts from the file being printed
	    (fonts-used-list nil)      ;map the list from file with Explorer list and form list to be used
	    (previous-font '("Courier" 10))
	    (current-font '("Courier" (10 10)))
	    (max-font-select nil)      ;character code of the maximum font that can be used
	    (title-name nil)		       ;if a title page is requested, name of the file being printed
	    (user-machine-id nil)	       ;machine user id from the print daemon
	    (file-object nil)	       ;printer stream
	    (eof 0)		       ;inform of a end of file
	    output-string	       ;stream to write file input into
	    input-string	       ;stream to read printer input from
	    )
	   (printer:basic-printer)
  :settable-instance-variables)

(defprop PostScript-Printer  (0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5 0.55 0.6 0.65 0.7
				   0.75 0.8 0.85 0.9 0.95 1. 2. 3. 4. 5. 6. 7. 8. 9. 10. ) dots-per-inch)


(DEFMETHOD (PostScript-Printer :get-fonts) ()
  "get fonts from the attribute list"

  (LET ((attri-line (fs:extract-attribute-list file-object)))
    (IF (NOT attri-line)
	(SETF file-font-list '(:DEFAULT))	;then use default font
						;else if no fonts use default font
	(IF (NOT (SETF file-font-list (GETF attri-line :fonts))) 
	    (SETF file-font-list '(:DEFAULT))
	    file-font-list))))
			     
(DEFMETHOD (PostScript-Printer :before :start-document) (&rest ignore)
   "Set up Explorer to PostScript font mappings. User defined mapping is appended to standard mapping
      and takes precedence."
  (SEND self :set-font-list (APPEND ps-font-map *explorer-postscript-font-map*))
  )

(DEFMETHOD (PostScript-Printer :make-font-list) ()
  "font list to be used by current application"
  (setf attribute-line (fs:extract-attribute-list file-object))     ;9-29-88 INT : now, attribute-line is an instance variable DAB
  (IF (EQUAL file-font-list nil) (SEND self :get-fonts))
  (SETF fonts-used-list (make-current-list font-list file-font-list) )
  (SETF max-font-select (+ (length fonts-used-list ) -1 (CHAR-INT #\0)))
  (SETF current-font (get-font-list #\0 fonts-used-list ))
  (setf previous-font (list current-font)))  ;10-19-87 DAB

(DEFMETHOD (PostScript-Printer :after :make-font-list) ()            ;9-9-88   INT (new method)
  "Modify the encoding vector of the Postscript fonts needed to print the file"
  (FORMAT (SEND self :printer-stream) "IntFont")	;to execute the Postscript procedure "IntFont"
  (SEND self :tyo-raw 10)			;send a newline character to end the string
  (let ((already-done nil))                     ; already-done will contain the name of the fonts already modified
    (dolist (my-current-font fonts-used-list)   ;loop to send the name of all Postscript fonts to use
      (let ((current-font-name (caar my-current-font)))
	(unless (member current-font-name already-done :test #'string-equal) ; do not send the font name if it is already sent
	    (FORMAT (SEND self :printer-stream) current-font-name)	     ;send the font name to postscript procedure
	    (SEND self :tyo-raw 10)			                     ;send a newline character to end the string
	  (push current-font-name already-done))))
    (FORMAT (SEND self :printer-stream) "no-more-font-to-map")               ;send "no-more-font-to-map" when finished
    (SEND self :tyo-raw 10)			                             ;send a newline character to end the string
    )
  )


(DEFUN make-current-list (alist file-font-list)
  "Move through the file font list and get the PostScript equivilant from the Explorer font list mapping.
    If the explorer font is not in the mapping list, use the default."
  (UNLESS (LISTP file-font-list)
    (SETF file-font-list (LIST file-font-list)))
  (LET ((used-ps-fonts nil)
	real-font)
    (DOLIST (file-font-item file-font-list)
      (setf real-font    ;handle the case where the fonts can be font objects, symbols or keywords 01-28-88 DAB
	    (do ((temp-font file-font-item)
		 (level 0))
		(nil)
	      (COND ((TYPEP temp-font 'font) (return (w:font-name temp-font )))
		    ((stringp temp-font)   ;04-07-88 case of :font "bigfnt" and :font-list '("bigfnt") DAB
		     (return (INTERN-local temp-font)))
		    ((eq (SYMBOL-PACKAGE temp-font) *keyword-package*)
		     (setq temp-font (INTERN temp-font 'fonts)))      ;10-06-88 DAB
		    ((> level 0) (return  temp-font))
		    ((AND  (SYMBOLP temp-font)  (not (BOUNDP temp-font))) (return temp-font))
		    ((AND  (SYMBOLP temp-font)  (BOUNDP temp-font))
		     (setq temp-font (EVAL temp-font)) (setq level (+ 1 level))))))

      (LET ((postscript-font (ASSOC (STRING real-font) alist :test #'string-equal)))
	(WHEN (NOT postscript-font)
	  (SETF  postscript-font (ASSOC "Default" alist :test #'STRING-EQUAL)))
	(PUSH (CDR postscript-font) used-ps-fonts )))
    (VALUES (REVERSE used-ps-fonts))))


			     
(DEFMETHOD (PostScript-Printer :print-text-file) (stream-name &optional fonts-read )
  (LET (first-char second-char)
    (SETF first-char (SEND stream-name :tyi))	   ;read first char in file
    (SETF second-char (SEND stream-name :tyipeek)) ;peek at second char in file
    (SEND stream-name :untyi first-char)	;put back first char
    (IF (AND (EQUAL first-char #x25)		;if the first two chars are %! then do a raw dump of the file
	     (EQUAL second-char #x21))		; instead of downloading our postscript driver. This means the file 
						; already contains postscript commands and we don't want to mess it up.
	
	(SEND self :dump-file stream-name)	;do raw dump
						;else do our stuff
	
	(SETF file-font-list fonts-read)	;use the font list from the user
	(SEND self :set-file-object stream-name)	;get the object for the file stream
	(SEND self  :write-postscript-program)	;write the postsript code to the printer
	(SEND self :set-output-string (MAKE-STRING-OUTPUT-STREAM))	;make a string stream because of VAX
	(SEND self :set-lfpad 0) (SEND self :set-ffpad 0)	;set defaults to 0 for padding
	(SEND self :set-lines-per-page 9999)(SEND self :set-chars-per-line 9999)
	(SEND self :make-font-list )		;make up the file-font-list from the file
	                                        ;9-29-88   INT : the following is moved after :make-font-list to separate
                                                ;            the 2 postscript procedures INTFONT and PRINFILE
	(FORMAT (SEND self :printer-stream) "PrintFile" )	;sent the start postsript command
	(SEND self :tyo-raw 10)			;send a newline character to end the string
	(SEND self :print)			;start the printing of the file
	)
    ))

(DEFMETHOD (PostScript-Printer :dump-file) (file-stream)
  "do raw file dump to printer"
  (LET (char-out)
    (SEND self :tyo-raw 04)			;reset printer
    (LOOP
      (SETF char-out (SEND file-stream :tyi ))
      (COND ((EQUAL char-out nil) (RETURN))
	    ((EQUAL (int-char char-out) #\return )
	     (SEND self :tyo-raw 10))
	    (t (SEND self :tyo-raw char-out))
	    )
      )  
    (SEND self :tyo-raw 04)			;inform printer of eof
    ))
  
(DEFMETHOD (PostScript-Printer :end-document) (&rest ignore)
  nil)

(DEFMETHOD (PostScript-Printer :start-document) (header-name user-name ignore)
  "send the command to the printer to start"
  (SETF eof 0)					;reset the flag for the end of file
  (SEND self :set-title-name header-name)	;set the file-name string
  (SEND self :set-user-machine-id user-name)
  )


(DEFMETHOD (PostScript-Printer :set-line-height) ()
  "set the font size to the largest font in the line This allows for variable line spacing"
  (LET ( next-char font-select line-height (char-out 0) new-point temp-font)
    (SETF line-height (CADR current-font))     ;get the font point
    (IF (ATOM line-height )	       ;check for a list and if it is a list get the second element
	()
	(SETF line-height (CADR line-height ))
	)
    (LOOP
      (SETF char-out (SEND file-object :tyi))
      ;add an extra return if return not at end of file
      (IF (EQUAL char-out nil)
	  (PROGN (SETF char-out #\return)
		 (SETF eof -1) ))
      (SEND output-string :tyo char-out)
      (COND 
	((EQUAL (INT-CHAR char-out) #\return) (RETURN))
	((EQUAL char-out 6) (PROGN
			      (SETF next-char (SEND file-object :tyipeek))
			      (SEND output-string :tyo next-char)
			      (IF (OR (<= #\0 (INT-CHAR next-char) #\I)	;allow up to 26 fonts
				      (EQUAL (INT-CHAR next-char) #\*))	
				  (PROGN
				    (SETF font-select (SEND file-object :tyi))	;character after Epsilon
				    (SETF temp-font (SEND self :font-select font-select))
				    (SETF new-point (CADR temp-font))	;set point size of new font
				    (IF (ATOM new-point) () (SETF new-point (CADR new-point)))
				    (IF (> new-point line-height) (SETF line-height new-point))
				    )
				  (if attribute-line               ;9-9-88 INT epsilon was printed twice
				      (SEND file-object :tyi)) 
				  )
			      ))
	)
      )
    (FORMAT (SEND self :printer-stream)
	    "~d" (if (integerp line-height) line-height 0))  ;10-19-87 DAB Change to printline to get out of ascii control.
    (SEND self :tyo-raw 10)
    (SETF input-string (MAKE-STRING-INPUT-STREAM (GET-OUTPUT-STREAM-STRING output-string)))
    )
  )  

    
(DEFMETHOD (Postscript-printer :header-stuff) (page-heading-p) 
 (IF (NOT page-heading-p)
      (FORMAT (SEND self :printer-stream) "~a" "")
      (IF (NOT (SECOND page-heading))
	  (FORMAT (SEND self :printer-stream) "~a" " ")
	  (FORMAT (SEND self :printer-stream) "~a" (SECOND page-heading))))
  (SEND self :tyo-raw 10) 
  (FORMAT (SEND self :printer-stream) "~a" (TIME:print-current-time nil ))
  (SEND self :tyo-raw 10)
  (IF file-object
      (LET ((NAMESTRING (FIRST page-heading)))
	(PRINC (IF (> (LENGTH namestring)  60)
		   (SUBSEQ namestring 0 60)
		   namestring)
	     (SEND self :printer-stream))  
	)
      (FORMAT (SEND self :printer-stream) "~a" ""))
  (SEND self :tyo-raw 10)
  (FORMAT (SEND self :printer-stream) "~a" (or user-machine-id user-id))
  (SEND self :tyo-raw 10)
  
  (IF title-name
      (PROGN					;print header page
	(FORMAT (SEND self :printer-stream) "~a"
		(cond ((PATHNAMEP title-name)
		       (SEND title-name :string-for-printing))
		      ((LISTP title-name)
		       (FIRST title-name))
		      ((STRINGP title-name)
		       title-name)
		      (t "")))
	(SEND self :tyo-raw 10))
      (FORMAT (SEND self :printer-stream) "~a" "")
      (SEND self :tyo-raw 10))) 


(DEFMETHOD (PostScript-Printer :print) ()
  "print the file stream"
  
  (LET (next-char char-out font-select new-font prev-char ;???? Check for default header values ????
	font-size (end-document 0)
	(maxed-ps-buffer 0) ; The postscript driver has a finite buffer length of 255 character. 11-14-88 DAB
	)

    (SEND self :header-stuff page-heading)	  ;send header information to the postscript printer
    ;; Swapped statements 09-22-87 DAB
    (IF print-wide				  ;if wide print, rotate 90 degrees
	(FORMAT (SEND self :printer-stream) "~d" 90)
	(FORMAT (SEND self :printer-stream) "~d" 0))
    (SEND self :tyo-raw 10)

    (SETF font-size (CADR current-font))	  ;set point size of  font
    (SETF new-font (CAR current-font))		  ;set new font to first atom of list string with name
    (SEND self :set-line-height)
    ;; Swapped statements 09-22-87 DAB
    (SEND self :send-font-size font-size)	  ;send  for initialization PS-SetPointSize
    (FORMAT (SEND self :printer-stream) "~a" new-font)
    (SEND self :tyo-raw 10)
						  ;header information download complete
    (FORMAT (SEND self :printer-stream) "~d" (or tab-width   ; DAB 01-24-89
						 (getf attribute-line :tab-width)
						 *default-tab-width*))  ; DAB 01-24-89
    (send self :tyo-raw 10)  ; DAB 01-24-89

    (LOOP
      (IF (EQUAL end-document -1)(RETURN) )
      (SETF prev-char char-out)
      (SETF char-out (SEND input-string :tyi))
      (IF  (EQUAL char-out nil) (RETURN))
      (COND
	((EQUAL (INT-CHAR char-out) #\return) (PROGN
						(SEND self :tyo-raw 10)
						(SEND self :tyo-raw 10)
						(IF (EQUAL eof -1)
						    (PROGN
						      (SETF end-document -1)
						      (FORMAT (SEND self :printer-stream) "~d" 0) ;10-19-87 DAB
						      (SEND self :tyo-raw 10)
						      (RETURN)))
						(SEND self :set-line-height)
						(setf maxed-ps-buffer 0) ;11-14-88 DAB
						))
	((EQUAL (INT-CHAR char-out) #\tab) (PROGN
					     (SEND self :tyo-raw 10)	  ;end of string
					     (SEND self :tyo-raw 9)	  ;do a tab function
					     (setf maxed-ps-buffer 0) ;11-14-88 DAB
					     )
					   )
	((EQUAL (INT-CHAR char-out) #\epsilon)
	 
	 (PROGN
	   (SETF next-char  (SEND input-string :tyi))
	   (IF (OR (<= #\0 (INT-CHAR next-char) #\I)  (EQUAL (INT-CHAR next-char) #\*))
	       (PROGN 
		 (SEND self :tyo-raw 10)	  ;send a return so they are expecting the font change
		 (SEND self :tyo-raw char-out)	  ;inform the printer of a impending font change
		 (SETF font-select next-char)	  ;character after Epsilon
		 (SETF current-font (SEND self :font-select font-select))
		 (SETF new-font (CAR current-font))	  ;set new font to first atom of list string with name	       
		 (SETF font-size (CADR current-font))	  ;set point size of new font
		 (SEND self :send-font-size font-size)
		 (FORMAT (SEND self :printer-stream) "~a" new-font)
		 (SEND self :tyo-raw 10)
		 (setf maxed-ps-buffer 0) ;11-14-88 DAB
		 
		 )
	       
	       (SEND self :tyo-raw 10)		  ;end the string for postscript
	       (SEND self :tyo-raw 8)		  ;inform postscript of a symbol
	       (SEND self :tyo-raw #o316)	  ;print the epsilon character
	       (SEND self :tyo-raw 10)		  ;send the epsilon character
	       (setf maxed-ps-buffer 0) ;11-14-88 DAB
	       )
	   )
	 )
	
	((OR (<= 0 char-out 31)			  ;math symbol
	     (CHAR-EQUAL (INT-CHAR char-out) #\))
	 (SEND self :convert-char (INT-CHAR char-out))
	 (setf maxed-ps-buffer 0) ;11-14-88 DAB
	 )
	
	((EQUAL (INT-CHAR char-out) #\page)
	 (PROGN
	   (SEND self :tyo-raw 10)		  ;send a return so they are expecting page change
	   (SEND self :tyo-raw 12)		  ;send the formfeed character
	   (setf maxed-ps-buffer 0) ;11-14-88 DAB
	   )
	 )
	
	
	(t
	 (if (>= maxed-ps-buffer 255.)  ;check to make sure we are not over filling the PS buffer.
	     (progn (SEND self :tyo-raw 10)  ;terminate readline
		    (SEND self :tyo-raw 0))  ;this is a dummy statement but it gets the PS driver back to readline.
	     (setf maxed-ps-buffer (1+ maxed-ps-buffer))) ;11-14-88 DAB
 	 (SEND self :tyo-char char-out)	  ;print the standard character
   
	   )
	)
      )
    (SEND self :tyo-raw 4)			  ;inform %stdin of Postscript the end of a file

    )						  ;end of let
  )

(DEFVAR *postscript-symbol-map*             ;9-9-88   INT : new mapping for international characters
;;Format of this list is <explorer char code> <postscript char code> ....
;; if <postscript char code> is a list then that charcode is from postscript's SYMBOL font
;; if <postscript char code> is NIL then there is no mapping	
	    '((#o0  (#o267)) (#o1  (#o257)) (#o2  (#o141)) (#o3  (#o142))     ;math symbols
	      (#o4  (#o331)) (#o5  (#o330)) (#o6  (#o145)) (#o7  (#o160))
	      (#o10 (#o154)) (#o11 (#o147)) (#o12 (#o144)) (#o13 (#o255))
	      (#o14 (#o261)) (#o15 (#o305)) (#o16 (#o245)) (#o17 (#o266))
	      (#o20 (#o314)) (#o21 (#o311)) (#o22 (#o307)) (#o23 (#o310))
	      (#o24 (#o042)) (#o25 (#o044)) (#o26 (#o304)) (#o27 (#o253))
	      (#o30 (#o254)) (#o31 (#o256)) (#o32 (#o271)) (#o33 (#o340))
	      (#o34 (#o243)) (#o35 (#o263)) (#o36 (#o272)) (#o37 (#o332))
	      (#o177 (#o362)) 

	                                                                     ;ISO chars
	      (#o240 #o40)    (#o246 #o174) 
              (#o251 (#o343)) (#o254 (#o330)) (#o255 #o55)    (#o256 (#o342))
              (#o260 (#o260)) (#o261 (#o261)) (#o262 #o62)    (#o263 #o63)
              (#o265 (#o155))
              (#o271 #o263)   (#o274 nil)     (#o275 nil)     (#o276 nil)
              (#o320 #o104)   (#o327 (#o264))
	      (#o335 #o131)   (#o336 #o120)  
              (#o360 #o144)   (#o367 (#o270))
              (#o375 #o171)   (#o376 #o160) 
	      ))

;(DEFVAR *postscript-symbol-map*
;;;Format of this list is <explorer char code> <postscript char code> ....
;;; if <postscript char code> is a list then that charcode is from postscript's SYMBOL font
;;; if <postscript char code> is NIL then there is no mapping	
;	    '((#o0  (#o267)) (#o1  (#o257)) (#o2  (#o141)) (#o3  (#o142))     ;math symbols
;	      (#o4  (#o331)) (#o5  (#o330)) (#o6  (#o145)) (#o7  (#o160))
;	      (#o10 (#o154)) (#o11 (#o147)) (#o12 (#o144)) (#o13 (#o255))
;	      (#o14 (#o261)) (#o15 (#o305)) (#o16 (#o245)) (#o17 (#o266))
;	      (#o20 (#o314)) (#o21 (#o311)) (#o22 (#o307)) (#o23 (#o310))
;	      (#o24 (#o042)) (#o25 (#o044)) (#o26 (#o304)) (#o27 (#o253))
;	      (#o30 (#o254)) (#o31 (#o256)) (#o32 (#o271)) (#o33 (#o340))
;	      (#o34 (#o243)) (#o35 (#o263)) (#o36 (#o272)) (#o37 (#o332))
;	      (#o177 (#o362))

;	      (#o240 #o040)   (#o241 #o241)   (#o242 #o242)   (#o243 #o243)
;	      (#o244 #o250)   (#o245 #o245)   (#o246 #o174)   (#o247 #o247)
;	      (#o250 #o310)   (#o251 (#o343)) (#o252 #o343)   (#o253 #o253)
;	      (#o254 (#o330)) (#o255 #o261)   (#o256 (#o342)) (#o257 (#o140))
;	      (#o260 (#o260)) (#o261 (#o261)) (#o262 #o62)    (#o263 #o63)
;	      (#o264 #o302)   (#o265 (#o155)) (#o266 #o266)   (#o267 (#o267))
;	      (#o270 #o313)   (#o271 #o365)   (#o272 #o353)   (#o273 #o273)
;	      (#o274 nil)     (#o274 nil)     (#o276 nil)     (#o277 #o277)
;	      (#o300 #o101)   (#o301 #o101)   (#o302 #o101)   (#o303 #o101)
;	      (#o304 #o101)   (#o305 #o101)   (#o306 #o341)   (#o307 #o103)
;	      (#o310 #o105)   (#o311 #o105)   (#o312 #o105)   (#o313 #o105)
;	      (#o314 #o111)   (#o315 #o111)   (#o316 #o111)   (#o317 #o111)
;	      (#o320 #o104)   (#o321 #o116)   (#o322 #o117)   (#o323 #o117)
;	      (#o324 #o117)   (#o325 #o117)   (#o326 #o117)   (#o327 (#o264))
;	      (#o330 #o351)   (#o331 #o125)   (#o332 #o125)   (#o333 #o125)
;	      (#o334 #o125)   (#o335 #o131)   (#o336 #o120)   (#o337 #o373)
;	      (#o340 #o141)   (#o341 #o141)   (#o342 #o141)   (#o343 #o141)
;	      (#o344 #o141)   (#o345 #o141)   (#o346 #o361)   (#o347 #o143)
;	      (#o350 #o145)   (#o351 #o145)   (#o352 #o145)   (#o353 #o145)
;	      (#o354 #o151)   (#o355 #o151)   (#o356 #o151)   (#o357 #o151)
;	      (#o360 #o144)   (#o361 #o156)   (#o362 #o157)   (#o363 #o157)
;	      (#o364 #o157)   (#o365 #o157)   (#o366 #o157)   (#o367 (#o270))
;	      (#o370 #o371)   (#o371 #o165)   (#o372 #o165)   (#o373 #o165)
;	      (#o374 #o165)   (#o375 #o171)   (#o376 #o160)   (#o377 #o171)
;	      )
;  )

(DEFMETHOD (PostScript-Printer :write-symbol) (char)
  (SEND self :tyo-raw 10)			;end the string for postscript
  (SEND self :tyo-raw 8)			;info postscript of a symbol
  (SEND self :tyo-raw char)
  (SEND self :tyo-raw 10))

(DEFMETHOD (PostScript-Printer :convert-char) (char)
  (LET ((converted-char (ASSOC (CHAR-INT char) *postscript-symbol-map*)))
    (if converted-char
	(IF (CADR converted-char)                                 ;if there a mapping?
	    (IF (LISTP (CADR converted-char))		      ;  then if it is a list
		(SEND self :write-symbol (CAADR converted-char))  ;    then it's from the PS SYMBOL font
		(SEND self :tyo-raw (CADR converted-char)))       ;    else it's just from the regular PS font
	    (progn                          
	      (SEND self :tyo-raw 10)                               ;  else write a lozenge with the character code
	      (SEND self :tyo-raw 7)
	      (FORMAT (SEND self :printer-stream) "#o~o" (CHAR-INT char))
	      (SEND self :tyo-raw 10)))
	(SEND self :tyo-raw char))                            ;9-9-88   INT : print the char if no mapping
    )
  )

(DEFMETHOD (PostScript-Printer :simulate-lispm-char) (CHAR)
  (IF (< #x7f (CHAR-CODE char) #xA0)
      (PROGN					;then
	(SEND self :tyo-raw 10)
	(SEND self :tyo-raw 7)
	(LET (lozenge-name)
	  (SETF lozenge-name
		(CASE (int-char char)   ;09-23-87 DAB
		  (#\NULL "NULL")
		  (#\BREAK "BREAK")
		  (#\CLEAR-INPUT "CLEAR-INPUT")
		  (#\CALL "CALL")
		  (#\TERM "TERM")
		  (#\MACRO "MACRO")
		  (#\HELP "HELP")
		  (#\RUBOUT "RUBOUT")
		  (#\OVERSTRIKE "OVERSTRIKE")
		  (#\BACKSPACE "BACKSPACE")
	;	  (#LINEFEED "LINEFEED")
		  (#\DELETE "DELETE")
		  (#\QUOTE "QUOTE")
		  (#\HOLD-OUTPUT "HOLD-OUTPUT")
		  (#\STOP-OUTPUT "STOP-OUTPUT")
		  (#\ABORT "ABORT")
		  (#\RESUME "RESUME")
		  (#\STATUS "STATUS")
		  (#\END "END")
		  (#\F1 "F1")
		  (#\F2 "F2")
		  (#\F3 "F3")
		  (#\F4 "F4")
		  (#\LEFT "LEFT")
		  (#\MIDDLE "MIDDLE")
		  (#\RIGHT "RIGHT")
		  (#\CENTER "CENTER")
		  (#\SYSTEM "SYSTEM")
		  (#\NETWORK "NETWORK")
		  (#\UNDO "UNDO")
		  (otherwise (FORMAT nil "#o~o" (CHAR-INT char))) 
		  )
		)
	  (FORMAT (SEND self :printer-stream) "~a" lozenge-name)
	  (SEND self :tyo-raw 10)
	  )
	)
      (SEND self :convert-char char)		;else
      )
  )



(DEFMETHOD (PostScript-Printer :send-font-size) (font-size)
  "sent the fontsize to the postscript printer"
  (LET (font-height font-width)
  	       (IF (ATOM font-size) (PROGN
				      (SETQ font-height font-size)
				      (SETQ font-width font-size))
		   (SETF font-height (CADR font-size)) (SETF font-width (CAR font-size))
		   ) ;allow for fonts to be set with one variable to two variables
	       (FORMAT (SEND self :printer-stream) "~d"  font-width)     ;10-19-87 DAB
 	       (SEND self :tyo-raw 10)  
	       (FORMAT (SEND self :printer-stream) "~d"  font-height)    ;10-19-87 DAB
	       (SEND self :tyo-raw 10)
	       )
  )

(DEFMETHOD (PostScript-Printer :font-select) (font-select)
  "Choose a font for PostScript depending on the character after the epsilon"
  (COND ((EQUAL (INT-CHAR font-select) #\* )
	   (SETF current-font (if previous-font (pop previous-font) current-font))   ;20-19-87 DAB
	   )
	((<= (CHAR-INT #\0) font-select max-font-select)
	 (push current-font previous-font)  ;20-19-87 DAB
	 (SETF current-font (get-font-list font-select fonts-used-list))
	 current-font 
	 )
	((<= (INT-CHAR (+ max-font-select 1)) font-select #\I)	;allow up to 26 fonts
	 (push current-font previous-font)   ;20-19-87 DAB
	 (SETF current-font (get-font-list (INT-CHAR max-font-select) fonts-used-list))
	 current-font )
  ))


(DEFUN get-font-list (font-select font-list)
  (CAR (ELT font-list (MOD (CHAR-INT font-select) #x30)))
  )

(DEFVAR *postscript-program-in-buffer*  nil
  "If nil, then the PostScript driver file is not resident in memory; otherwise points to the buffer containing it.")
;(DEFVAR *postscript-driver-file* "sys:printer;postscript-driver.text"
;  "This file is downloaded to the printer. It is handles the printer's interface with the Explorer.")
;(DEFVAR *postscript-driver-source* "sys:printer;postscript-driver.text"
;  "Latets version of *postscript-driver-file*.")

;;; The postscript-driver is now defined by the variable  *postscript-driver*. The function build-postscript-source-variable
;;; will accept an optional file name. This will allow the user the define an alternate driver file.
(defun build-postscript-source-variable (&optional (source *postscript-driver*))
  "Build a condense version of the file that is download to the Postscript. The source file 
   is defined SOURCE. SOURCE can be a file or an array in the format as *postscript-driver*."
  (declare (special *postscript-driver*))
  (case (typep source)
    (:ARRAY
     (UNLESS  *postscript-driver*		;if no local copy
       (FERROR () "Could not find PostScript driver variable *postscript-driver*")	;   else error
       )
     (SETF *postscript-program-in-buffer* (MAKE-ARRAY (* (array-dimension *postscript-driver* 0) 132.) :leader-length 1))
     (LET ((s *postscript-driver*))
       (setf (array-leader *postscript-program-in-buffer*  0)	;1-5-88 DAB Change postscript-driver.text to defvar
	     (do* ((x 0) (continue-read nil) (paren-count 0) (move-byte-to-buffer nil nil)
		   (previous-byte (char-int #\Space) current-byte)
		   (line-counter 0)
		   (char-counter 0 (1+ char-counter))
		   (current-string (aref s line-counter))
		   (current-string-length (length current-string) (length current-string))
		   (current-byte (aref current-string char-counter)
				 (if (= char-counter current-string-length)
				     141
				     (aref current-string char-counter)))
		   )
		  ((or (= current-string-length 0)
		       (= line-counter (array-dimension s 0)))
		   x)
	       
	       (cond ((= current-byte 137) ())	;tab
		     ((= current-byte 141)	;newline
		      (setq continue-read nil)
		      (setf (aref *postscript-program-in-buffer* x) 32)	;force a space because next line start in col 0
		      (setq x (1+ x))
		      (setq line-counter (1+ line-counter ))
		      (setq char-counter -1)
		      (setq current-string (if (<= line-counter (array-dimension s 0))
					       (aref s line-counter)
					       ""))
		      )
		     (continue-read ())		;continue to read from input file, must be in a comment.
		     ((= current-byte 32)	;let's do some blank suppression
		      (unless (= previous-byte 32)
			(setf move-byte-to-buffer t)))
		     ((= current-byte 37)	;%
		      (if (= 0 paren-count)	;must be the start of a comment
			  (setf continue-read t)
			  (setf move-byte-to-buffer t)))	;must be part of a string
		     ((= current-byte 40)	;(
		      (unless (= Previous-byte 92) (setf paren-count (1+ paren-count)))	;check for back slash
		      (setf move-byte-to-buffer t))
		     ((= current-byte 41)	;)
		      (unless (= Previous-byte 92) (setf paren-count (1- paren-count)))	;check for back slash
		      (setf move-byte-to-buffer t))
		     (t (setf move-byte-to-buffer t)))
	       (when move-byte-to-buffer
		 (setf (aref *postscript-program-in-buffer* x) current-byte)
		 (setf x (1+ x)))
	       ))))
    (T
     (UNLESS (PROBE-FILE SOURCE)		;if no local copy
       (FERROR () "Could not find PostScript driver file ~a"  SOURCE)	;   else error
       )
     (SETF *postscript-program-in-buffer* (MAKE-ARRAY (FILE-LENGTH   SOURCE) :leader-length 1))
     (WITH-OPEN-FILE (s source :direction :input :if-does-not-exist :error)
       (setf (array-leader *postscript-program-in-buffer*  0)	;10-22-87 DAB
	     (do* ((x 0) (continue-read nil) (paren-count 0) (move-byte-to-buffer nil nil)
		   (previous-byte (char-int #\Space) current-byte)
		   (current-byte (send s :tyi) (send s :tyi)))
		  ((null current-byte ) x)
	       
	       (cond ((= current-byte 137) ())	;tab
		     ((= current-byte 141)	;newline
		      (setq continue-read nil)
		      (setf (aref *postscript-program-in-buffer* x) 32)	;force a space because next line start in col 0
		      (setf x (1+ x)))
		     (continue-read ())		;continue to read from input file, must be in a comment.
		     ((= current-byte 32)	;let's do some blank suppression
		      (unless (= previous-byte 32)
			(setf move-byte-to-buffer t)))
		     ((= current-byte 37)	;%
		      (if (= 0 paren-count)	;must be the start of a comment
			  (setf continue-read t)
			  (setf move-byte-to-buffer t)))	;must be part of a string
		     ((= current-byte 40)	;(
		      (unless (= Previous-byte 92) (setf paren-count (1+ paren-count)))	;check for back slash
		      (setf move-byte-to-buffer t))
		     ((= current-byte 41)	;)
		      (unless (= Previous-byte 92) (setf paren-count (1- paren-count)))	;check for back slash
		      (setf move-byte-to-buffer t))
		     (t (setf move-byte-to-buffer t)))
	       (when move-byte-to-buffer
		 (setf (aref *postscript-program-in-buffer* x) current-byte)
		 (setf x (1+ x)))
	       )))))
  )


(DEFMETHOD (PostScript-Printer :write-postscript-program) ()
  "Write the standard postcript driver file. This file can be changed by sending a new file name"
  (declare (special  *postscript-driver* *postscript-program-in-buffer*))
  (UNLESS *postscript-program-in-buffer*	
    (FERROR () "Could not find PostScript driver file ~a" *postscript-driver*))
  (SEND self :tyo-raw 4)			;reset the printer
  
  (DOTIMES (i (array-leader *postscript-program-in-buffer* 0))  ;10-22-87 DAB       ;now write the buffer to the printer
    (IF (EQUAL (AREF *postscript-program-in-buffer* i)	        ;if carriage return
	       (CHAR-INT #\return))			                
	(SEND self :tyo-raw 10)                                       ; then send CR to PS printer
	(SEND self :tyo-raw (AREF *postscript-program-in-buffer* i))  ; else send the character
	  ))
  )
   

(DEFMETHOD (postscript-printer :bitmap-setup) ()
  "Setup to print bitmap"
  (SEND self :write-postscript-program)
  (FORMAT (SEND self :printer-stream) "~a" "PrintScreenImage")
  (SEND self :tyo-raw 10)
  (SEND self :header-stuff nil)
 )
 
(DEFMETHOD (postscript-printer :print-bitmap) 
	   (bitmap-array
	    &OPTIONAL width height (start-x 0) (start-y 0) orientation dots-per-inch
	    &AUX picstr-width (image-hex 0) (image-bit 0) pic-width (default-scale .24)
	         user-scale bit-add (top-margin 700.0) (rotate-angle 0) (scale-multiplier 2)
	         (x-fudge 100) (y-fudge 50)
	     )
  "Copy the bitmap array to a print stream;
printer-stream could be a file stream or serial stream. It assumes 8 bit data transfer"

						  ;if I am printing to a remote host, don't do :bitmap-setup
						  ; because it will be done by :print-raw file on the server
  (UNLESS  (member (intern-local (TYPE-OF (SEND self :printer-stream)) 'printer)  ;08-19-88 DAB
		       '(lm-character-output-stream
			  nfs-buffered-output-character-stream)	;07-18-88 DAB Added for screen prints on a MX.
		   :test #'equal)
    (SEND self :bitmap-setup))

  (IF (EQUAL orientation :landscape )
      (PROGN (SETQ rotate-angle 90)
	     (SETQ top-margin 10)
	     (SETF x-fudge 50)
	     (SETF y-fudge 50)
	     (SETQ scale-multiplier 3)))
  (IF (NUMBERP dots-per-inch)
      (SETQ scale-multiplier dots-per-inch))
  (SETF user-scale (* default-scale scale-multiplier))
  (unless width
    (setq width (ARRAY-DIMENSION bitmap-array 1)))
  (unless height
    (setq height (ARRAY-DIMENSION bitmap-array 0)))

  (IF (EQUAL   width  (* (quotient  width  8) 8))
      (SETF pic-width  width )
      (PROGN
	(SETF pic-width (* 8 (+ (quotient width 8) 1)) )
	(SETF width pic-width)))
  (SETF picstr-width (quotient pic-width 8))
 
  (FORMAT (SEND self :printer-stream) "~d" rotate-angle )
  (SEND self :tyo-raw 10)
  (FORMAT (SEND self :printer-stream) "~d" user-scale)
  (SEND self :tyo-raw 10)
  (FORMAT (SEND self :printer-stream) "~d" width)
  (SEND self :tyo-raw 10)  
  (FORMAT (SEND self :printer-stream) "~d" height)
  (SEND self :tyo-raw 10)
;  (FORMAT (SEND self :printer-stream) "~d" x-fudge)
;  (SEND self :tyo-raw 10)
  (FORMAT (SEND self :printer-stream) "~d" (- (- (quotient top-margin user-scale)  height) y-fudge))
  (SEND self :tyo-raw 10)
  (FORMAT (SEND self :printer-stream) "~d" picstr-width)
  (SEND self :tyo-raw 10)
  (FORMAT (SEND self :printer-stream) "~d" pic-width)
  (SEND self :tyo-raw 10)
  (LET ((*print-base* 16))

    (DO (( y  start-y  (+ y 1))) ((EQUAL y    height  )  ()) 
      (DO (( x  start-x  (+ x 1))) ((= x    width ) ())
	(SETF bit-add (AREF bitmap-array y x))
	(IF (= bit-add 0)(SETF bit-add 1)(SETF bit-add 0))
	(SETF image-hex (+ image-hex  bit-add ))
	(SETF image-bit (+ image-bit 1))
	(IF (EQUAL image-bit 8) (PROGN
				  (IF (< image-hex 16)(PRINC "0" (SEND self :printer-stream)))
				  (SETF image-bit 0)
				  (princ image-hex (SEND self :printer-stream ))
				  (princ " " (SEND self :printer-stream ))
				  (SEND self :tyo-raw 10)
				  (SETF image-hex 0)
				  )
	    )
	(SETF image-hex (LSH image-hex 1))
	)
      ))
  (FORMAT (SEND self :printer-stream) "showpage")(SEND self :tyo-raw 10)
  (SEND self :tyo-raw 04)

  
  )

(DEFMETHOD (PostScript-Printer :screen-image-file-p) (filename)
  "Return T if this is a PostScript screen dump file"
  (LET ((FILE-TYPE (SEND FILENAME :TYPE)))
    (EQUALP "POSTSCRIPT-PRINTER" FILE-TYPE)))


(DEFMETHOD (PostScript-Printer :print-raw-file) (STREAM)
  "Write a screen dump file generated by the requestor to the printer"
  (SEND self :bitmap-setup)
  (LOOP FOR IN-CHAR = (TYI STREAM ()) UNTIL (NULL IN-CHAR) DO
	    (SEND self :tyo-raw in-char))	 
  )

(DEFMETHOD (PostScript-Printer :print-page-heading) (&rest ignore)  ;10-19-87 DAB
  )

(DEFMETHOD (postscript-printer :CURRENT-PRINTER-STATE) (STREAM-TYPE)
  (IF (OR (NULL STREAM-TYPE); determine whether null or serial,
                            ; or else stream is parallel
          (EQ :SERIAL STREAM-TYPE))
    (LIST :ONLINE (LDB-TEST SI::%%SERIAL-B-STATUS-REGISTER-DSR
                            (SI::SERIAL-B-STATUS)))
    ;ti855 p1 is dtr go to
    (LIST :FAULT (SEND PRINTER-STREAM :SEND-IF-HANDLES
                       :GET :FAULT);dcdb on z8530 chip Channel B
          :ONLINE (SEND PRINTER-STREAM :SEND-IF-HANDLES :GET :ONLINE))))

(DEFUN validate-postscript-driver-version (&optional reload?)
  "Make sure  the postscript driver file is loaded in memory. The file will be laoded and binded to 

This function should be called from the LM:SITE;BOOT-INIT.LISP file on print servers that have a PostScript printer."
  (when
    (or (not *postscript-program-in-buffer*)
	reload?
	)
    (build-postscript-source-variable)
    )
  )
(COMPILE-FLAVOR-METHODS postscript-printer)

;;;Overview:
;;;PostScript printers are implemented differently than other Explorer printers. The basic outline is this:
;;;The ExplorerPS font mappings are contained in the variable *explorer-postscript-font-map*. Users may specify
;;; a different mapping for all spooled print requests by editting this variable. A user may specify a different mapping for 
;;; all print requests from his machine by defining a different font mapping in the variable *default-ps-font-map*. 
;;; A different mapping for a specific print request can be made by supplying a :ps-font-map keyword argument
;;; to print-file. If some or all of those variable are defined and there are duplications, the precedence (in decreasing order)
;;; of the mappings to use are (1) :ps-font-map keyword, (2) *default-ps-font-map* variable, (3) *explorer-postscript-font-map*.
;;;
;;;If the first 2 characters of the file to be printed are "%!", that file is assumed to already contain postscript page
;;; descriptions and it is simply dumped to the printer without further ado. Otherwise, it is assumed to just be a generic text file.
;;; In that case, a postscript source code driver must be downloaded to the printer. If this driver file does not currenly exist on the 
;;; print server, it is copied to the print server from sys:printer; .  Once the download is complete then the
;;; text data is dumped to the printer.

;;; Source code that is downloaded to the postscript-printer.
;;; build-postscript-source-variable will use this variable to build a condense version without
;;; tab, newlines or comments. Do not insert blank line. All lines will be truncated after 132 characters.
;;; The array can be extended if necessary. Remember to backslash backslashes and double quotes and do 
;;; not insert any newlines.
;;; All patches to this variable must also include (build-postscript-source-variable) at the end. 
(defvar *postscript-driver*
	(make-array '(500)   ;current lenght is 470
	:initial-contents
	'("%Explorer PostScript driver file"
	  "%!PS-Adobe-1.0"
	  "%%%                           RESTRICTED RIGHTS LEGEND "
          "%%%Use, duplication, or disclosure by the Government is subject to"
	  "%%%restrictions as set forth in subdivision (b)(3)(ii) of the Rights in"
	  "%%%Technical Data and Computer Software clause at 52.227-7013."
	  "%%%"
	  "%%%                     TEXAS INSTRUMENTS INCORPORATED."
	  "%%%                              P.O. BOX 2909"
	  "%%%                           AUSTIN, TEXAS 78769"
	  "%%%                                 MS 2151"
	  "%%%"
	  "%%% Copyright (C) 1987, Texas Instruments Incorporated. All rights"
	  "%%% reserved."
	  "%%% NOTE: DO NOT ALLOW ANY TAB CHARACTERS IN THIS FILE. POSTSCRIPT WILL"
	  "%%%  SIMPLY STOP PROCESSING WHEN IT SEES A TAB."
	  "%%%This file contains the PS program used by the Explorer PS printer"
	  "%%% driver. This file is downloaded to the printer for each print"
	  "%%% request. The entry point for text files is PrintFile. The entry"
	  "%%% point for screen dumps is PrintScreenImage. This code is tough to"
	  "%%% debug. Usually if there is an error, the printer just quits. The"
	  "%%% best way to figure out what went wrong is to use the subroutine"
	  "%%% called Test. Insert a call to Test at about where you suspect the"
	  "%%% problem. If the printer gets that far, then a page will be printed "
	  "%%% displaying the top two elements of the PS stack. Sometimes you just"
	  "%%% have to do a binary search placing Test in various location until"
	  "%%% you narrow down the problem. Once you know which part of this code"
	  "%%% is failing, you can usually figure out what's going on by executing"
	  "%%% it in Interactive mode. Kermit (in the public directory) is a good "
	  "%%% terminal emulator to use. The code in public:public;vt100-yank.xld"
	  "%%% allows you to mark text in Zmacs and then yank it into the VT100"
	  "%%% emulator instance running Kermit. Happy hunting"
	  "%%%"                    ;9-9-88   INT : start of international characters handling
	  "/reencsmalldict 12 dict def      % This procedure is exactly the one "
	  "                                 % described in the program 18 of the"
	  "                                 % Postscript Tutorial and Cookbook,"
	  "                                 % except the new encoding vector"
	  "/ReEncodeSmall"
	  "{reencsmalldict begin"
	  "/newcodesandnames exch def"
	  "/newfontname exch def"
	  "/basefontname exch def"
	  "/basefontdict basefontname findfont def"
	  "/newfont basefontdict maxlength dict def"
	  "basefontdict"
	  "{exch dup /FID ne"
	  "{dup /Encoding eq"
	  "{exch dup length array copy"
	  "newfont 3 1 roll put}"
	  "{exch newfont 3 1 roll put}"
	  "ifelse"
	  "}"
	  "{pop pop}"
	  "ifelse"
	  "} forall"
	  "newfont /FontName newfontname put"
	  "newcodesandnames aload pop"
	  "newcodesandnames length 2 idiv"
	  "{newfont /Encoding get 3 1 roll put}"
	  "repeat"
	  "newfontname newfont definefont pop"
	  "end"
	  "} def"
	  "/patvec [          % new encoding vector for international (8859-1)"
	  "8#244 /currency"
	  "%"
	  "8#250 /dieresis                  8#252 /ordfeminine"
	  "8#257 /macron"
	  "%"
	  "8#262 /tilde          8#263 /dotlessi"
	  "8#264 /acute"
	  "%"
	  "8#270 /cedilla                   8#272 /ordmasculine"
	  "8#274 /grave        8#275 /ring  8#276 /circumflex"
	  "%"
	  "8#300 /Agrave       8#301 /Aacute 8#302 /Acircumflex   8#303 /Atilde"
	  "8#304 /Adieresis    8#305 /Aring  8#306 /AE            8#307 /Ccedilla"
	  "%"
	  "8#310 /Egrave       8#311 /Eacute 8#312 /Ecircumflex   8#313 /Edieresis"
	  "8#314 /Igrave       8#315 /Iacute 8#316 /Icircumflex   8#317 /Idieresis"
	  "%"
	  "8#321 /Ntilde 8#322 /Ograve        8#323 /Oacute"
	  "8#324 /Ocircumflex  8#325 /Otilde 8#326 /Odieresis"
	  "%"
	  "8#330 /Oslash       8#331 /Ugrave 8#332 /Uacute        8#333 /Ucircumflex"
	  "8#334 /Udieresis                                       8#337 /germandbls"
	  "%"
	  "8#340 /agrave       8#341 /aacute 8#342 /acircumflex   8#343 /atilde"
	  "8#344 /adieresis    8#345 /aring  8#346 /ae            8#347 /ccedilla"
	  "%"
	  "8#350 /egrave      8#351 /eacute 8#352 /ecircumflex   8#353 /edieresis"
	  "8#354 /igrave      8#355 /iacute 8#356 /icircumflex   8#357 /idieresis"
	  "%"
	  "8#361 /ntilde 8#362 /ograve        8#363 /oacute"
	  "8#364 /ocircumflex 8#365 /otilde 8#366 /odieresis"
	  "%"
	  "8#370 /oslash      8#371 /ugrave 8#372 /uacute        8#373 /ucircumflex"
	  "8#374 /udieresis                                      8#377 /ydieresis" 
	  "] def"
	  "/IntFont                % loop to remap all fonts needed, until we"
	  "                        % read the string no-more-font-to-map"
	  "{"
	  "/my-in-int (%stdin) (r) file def"
	  "/my-font-name 40 string def"
	  "{"
	  "my-in-int my-font-name readline pop dup (no-more-font-to-map) ne"
	  "     % read the fontname of the font to remap"
	  "     % to get ride of the boolean"
	  "     % dupplicate the font name"
	  "     % no-more-font-to-map ?"
	  "{"
	  "dup patvec ReEncodeSmall       % dupplicate the font name"
	  "                               % remap the font by overwriting itself"
	  "}"
	  "{pop exit} ifelse              % to get ride of the dupplicated font name, and exit"
	  "} loop"
	  "} def"
	  "%%%"             ;9-9-88   INT : end of international characters handling
	  "/Test"
	 "{"
	 "/bitstr 80 string def"
	 "/Helvetica findfont 10 scalefont setfont"
	 "200 200 moveto"
	 "(TEST: stack top 2 = ) show dup bitstr cvs show exch ( ) show "
	 "dup bitstr cvs show exch"
	 "showpage"
	 "} def"
	 "/PrinterCheck"
	 "{(*) print"
	 "}def"
	 "/my-in (%stdin) (r) file def"
	 "/Orientation"
	 "%Determine if wide printing"
	 "{"
	 "/my-in (%stdin) (r) file def"
	 "/rstr 10 string def"
	 "/leftmargin 40 def"
	 "/ybot 25 def"
	 "my-in rstr readline pop"
	 "rstr cvi 0 eq "
	 "{"
	 "/topmargin 770 def"
	 "/rightmargin 576 def"
	 "}"
	 "{"
	 "rstr cvi rotate"
	 "0 -612 translate      %09-22-87 DAB"
	 "/topmargin 576 def"
	 "/rightmargin 770 def"
	 "} ifelse "
	 "} bind def"
	 "/MyShow"
	 "{show} bind def"
	 "/PostScriptProgram"
	 "%Handles PS program embedded in file being printed"
	 "{ gsave"
	 "{"
	 "my-in nstr readline "
	 "pop"
	 "%   Test"
	 "cvx"
	 "exec "
	 "my-in read pop"
	 "%   TesT"
	 "1 eq {exit} if"
	 "} loop"
	 "grestore"
	 "} bind def"
	 "/SetPointSize"
	 "%Sets the current point size of the font"
	 "{/pointstring 4 string def"
	 "my-in pointstring readline                               %point size"
	 "pop                                      %Get rid of true"
	 "cvi                                      %10-19-87 DAB To get value out of the control char range."
	 "/PointSize exch def                      %Exchange top of stack to put"
	 "} bind def"
         "/SetTabWidth"
	 "%Sets the current number of space of the TAB"
	 "{/tabstring 4 string def"
	 "my-in tabstring readline"
	 "pop                                      %Get rid of true"
	 "cvi                                      %10-19-87 DAB To get value out of the control char range."
	 "(w) stringwidth pop mul"
	 "/TabWidth exch def                       %Exchange top of stack to put"
	 "} bind def"
	 "/TitlePage"
	 "%Print the header page"
	 "{initgraphics                              %reset graphics state"
	 "25 setlinewidth   %Draw box on page"
	 "0 setlinejoin"
	 "2 setlinecap"
	 ".5 setgray"
	 "36 760 moveto"
	 "537 0 rlineto"
	 "0  -735 rlineto"
	 "-537 0 rlineto"
	 "0 735 rlineto"
	 "stroke"
	 "0 setgray"
	 "/Times-Roman findfont 12 scalefont setfont"
	 "leftmargin 15 add 725 moveto"
	 "(Owner: ) show leftmargin 60 add 725 moveto counttomark 1 add index show"
	 "leftmargin 15 add 700 moveto"
	 "(File:) show leftmargin 60 add 700 moveto pstr show   %counttomark 2 add index show"
	 "leftmargin 15 add 675 moveto"
	 "(Date:) show leftmargin 60 add 675 moveto counttomark 3 add index show"
	 "leftmargin 15 add 100 moveto"
	 "[11 0 0 13 0 0]"
	 "/Times-Roman findfont exch makefont setfont 100 100 moveto (T) show"
	 "[10 0 0 9 0 0]"
	 "/Times-Roman findfont exch makefont setfont 106 100 moveto (EXAS) show"
	 "[11 0 0 13 0 0]"
	 "/Times-Roman findfont exch makefont setfont 135 100 moveto (I) show"
	 "[10 0 0 9 0 0]"
	 "/Times-Roman findfont exch makefont setfont 139 100 moveto"
	 "(NSTRUMENTS) show"
	 "420 55 moveto (PostScript) show"
	 "/Symbol findfont 6 scalefont setfont"
	 "currentpoint 3 add moveto"
	 "(\\322 ) show"
	 "currentpoint 3 sub moveto"
	 "[10 0 0 9 0 0]"
	 "/Times-Roman findfont exch makefont setfont"
	 "(Version ) show version show"
	 "/Helvetica-Oblique findfont 12 scalefont setfont"
	 "100 85 moveto (Explorer) show"
	 "/Symbol findfont 8 scalefont setfont"
	 "currentpoint exch 2 add exch 3 add moveto"
	 "(\\344) show"
	 "showpage"
	 "} bind def"
	 "/Startup"
	 "%Sets up initial strings and constants. Reads header page and page"
	 "%heading information. Determines if header page is needed, etc. "
	 "{clear"
	 "/leftmargin 40 def"
	 "/topmargin 770 def"
	 "/ybot 25 def                            %Offset from bottom margin"
	 "/nstr 255 string def                    %Various strings"
	 "/astr 80 string def"
	 "/bstr 255 string def"
	 "/cstr 80 string def"
	 "/hstr 255 string def"
	 "/tstr 20 string def"
	 "/fstr 255 string def"
	 "/pstr 255 string def"                  
	 "1                       %push current page count on stack"
	 "my-in hstr readline pop                 %Send the the heading second"
	 "% line. If its not nil print"
	 "% header."
	 "my-in astr readline pop                 %Read current time"
	 "my-in bstr readline pop                 %Read file name"
	 "my-in cstr readline pop                 %Read name of current user"
	 "mark                                    %Mark start of page header"
	 "% info on stack"
	 "my-in pstr readline pop                 %Get title page info"
	 "length 0 gt {TitlePage} if              %Print title page?"
	 "}  bind def"
	 "/NewHeadingLine"
	 "%Newline for the page header."
	 "{"
	 "currentpoint 12 sub "
	 "exch pop"
	 "leftmargin exch moveto"
	 "} bind def"
	 "/PrintHeading"
	 "%Prints the page heading on top of each page if needed."
	 "{ counttomark 4 add index length 0 gt    %Heading requested? "
	 "{currentfont "
	 "/Courier findfont 10 scalefont setfont"
	 "counttomark 2 add index show            %Print file name"
	 "currentpoint exch pop 400 exch moveto"
	 "( ) show counttomark 3 add index show   %Print current time"
	 "( Page:) show counttomark 5 add index tstr cvs show  %Print page number"
	 "counttomark 6 add -1 roll               %Get page number on top of stack"
	 "1 add                                   %Increment page number"
	 "counttomark 5 add 1 roll                %Put stack back the way it was"
	 "NewHeadingLine"
	 "counttomark 4 add index show            %Print 2nd line of header"
	 "NewHeadingLine"
	 "setfont                                 %Reset to old font"
	 "%Draw solid line"
	 "gsave currentpoint exch rightmargin add exch lineto "
	 "stroke grestore "
	 "NewLine} if "
	 "} bind def "
	 "/NewPage "
	 "%Handle page feed "
	 "{gsave showpage grestore"
	 "leftmargin topmargin PointSize sub moveto  %Set the top of the page"
	 "PrintHeading"
	 "} bind def"
	 "/moveDown"
	 "%Calculates how much to drop line: Pointsize + 10%"
	 "{ PointSize PointSize 10 div add sub} bind def"
	 "/ChangeFont"
	 "{/pointstring 4 string def"
	 "%10-19-87 DAB Get the chars values out of the range of ascii control char range"
	 "[                                        %make an array for the font"
	 "my-in pointstring readline                %read point width"
	 "pop cvi                                   %get rid of the true"
	 "0"
	 "0"
	 "my-in pointstring readline                %read point height"
	 "pop cvi                                    %get rid of the true"
	 "0"
	 "0"
	 "]                                         %end of fontmatrix array"
	 "my-in fstr readline                       %read font"
	 "pop                                       %get rid of true"
	 "cvn"
	 "findfont                                  %get off of top of stack"
	 "exch  makefont setfont"
	 "} bind def"
	 "/NewLine"
	 "{"
	 "currentpoint exch pop                     %Get the y value by it self"
	 "%If too close to bottom do"
	 "% page feed, else new line"
	 "moveDown dup ybot lt {NewPage pop} {leftmargin  exch  moveto } ifelse "
	 "} bind def"
	 "/LineWrap"
	 "%Handle line wrap condition"
	 "{"
	 "{"
	 "{dup dup length 1 sub                   %String length minus CR"
	 "0 exch getinterval                     %String without CR"
	 "dup stringwidth pop"
	 "dup 0 eq {pop exit} if "
	 "currentpoint pop add rightmargin gt {} {exit} ifelse"
	 "} loop"
	 "dup"
	 "MyShow                                    %Print first line"
	 "(!) MyShow"
	 "length /StringIndex exch def "
	 "cleartomark"
	 "NewLine"
	 "dup length                               %Put remainder of line on stack"
	 "StringIndex sub StringIndex exch getinterval"
	 "dup stringwidth pop leftmargin add rightmargin lt {exit}"
         "{dup mark exch} ifelse %10-28-87 DAB Fixed mulitiple line wraps."
         "} loop"
	 "} bind def"
	 "/TabSpace"
	 "%Print tab"
	 "{"
	 "currentpoint "
	 "exch                                      %get x-coordinate"
	 "( ) stringwidth pop add                   %position past next char spot"
	 "exch moveto"
	 "leftmargin"
	 "{TabWidth add                             %cumulatively add TabWidth to itself"
	 "dup currentpoint pop gt                 % until we hit a valid tab stop"
	 "{exit} if                               % leaving tab stop on top of stack "
	 "} loop"
	 "currentpoint exch pop"
	 "exch dup rightmargin gt                   %10-28-87 DAB %if this would put us past right margin"
	 "{currentpoint exch pop rightmargin 1 add exch moveto (!) show NewLine }  %10-28-87 DAB % then do a newline"
	 "{exch moveto} ifelse                      % else go to tab stop"
	 "} bind def"
	 "/StartText"
	 "%Set things up for text files."
	 "{"
	 "SetPointSize"
	 "ChangeFont"
	 "/TabWidth (w) stringwidth pop 8.0 mul def"
	 "leftmargin topmargin PointSize sub moveto  %Initial starting spot"
	 "PrintHeading"
	 "} bind def"
	 "/PrintSymbol"
	 "%Print character from the Symbol font"
	 "{currentfont                              %Get the font matrix for the"
	 "% current font."
	 "dup"
	 "begin                                    %Put dictionary on stack"
	 "FontMatrix"
	 "end                                      %Pop dictionaly"
	 "[1000 0 0 1000 0 0]                      %Matrix to multiply to get"
	 "% ready for makefont."
	 "[1 0 0 1 0 0]                            %Matrix to store the changes in"
	 "concatmatrix"
	 "/Symbol findfont exch makefont setfont"
	 "my-in nstr readline"
	 "pop"
	 "dup stringwidth pop currentpoint pop add rightmargin gt"
	 "{currentpoint exch pop rightmargin 1 add exch moveto (!) show NewLine} if"
	 "show"
	 "setfont"
	 "} bind def"
	 "/PrintLozenge"
	 "%Print a lozenge"
	 "{ "
	 "/lstr 20 string def"
	 "currentfont "
	 "/Helvetica findfont 5 scalefont setfont"
	 "my-in lstr readline pop stringwidth pop   %Read lozenge contents and calculate size"
	 "dup 2 add currentpoint pop add rightmargin gt %test if enough room in the line "
	 "{1 index currentfont"
	 "exch setfont             %reset the previous font"
	 "currentpoint exch pop rightmargin 1 add exch moveto (!) show"
	 "setfont                  %reset Helvetica"
	 "NewLine"
	 "} if"
	 "currentpoint"
	 "3 -1 roll                %put the string on top"
	 "gsave"
	 "currentpoint 4 add exch 2 add exch moveto %left vertex"
	 "5 4 rlineto                               %left /"
	 "dup 0 rlineto                             %top"
	 "5 -4 rlineto                              %right \\"
	 "-5 -4 rlineto                             %right /"
	 "dup neg 0 rlineto                         %bottom"
	 "-5 4 rlineto                              %left \\"
	 "closepath"
	 "5 -2 rmoveto                              %start name of lozenge"
	 "pop"
	 "lstr show "
	 ".3 setlinewidth stroke "
	 "grestore"
	 "exch lstr stringwidth pop add "
	 "exch"
	 "moveto"
	 "setfont                                   %Set point for next character"
	 "} bind def"
	 "/PrintFile"
	 "{"
	 "Startup"
	 "Orientation   %Swapped statements 09-22-87 DAB"
	 "StartText"
	 "SetTabWidth"
	 "{"
	 "my-in nstr readline true eq           %Read line from file"
	 "%Check for linewrap "
	 "{dup stringwidth pop currentpoint pop add  rightmargin gt { dup mark exch LineWrap} if"
	 "MyShow"
	 "my-in read pop                        %character for end of line "
	 "dup 1 eq {PostScriptProgram} if       %Embedded PS program (NOT SUPPORTED)"
	 "dup 6 eq {ChangeFont} if"
	 "dup 10 eq {SetPointSize NewLine} if"
	 "dup 12 eq {NewPage} if "
	 "dup 9 eq {TabSpace} if"
	 "dup 8 eq {PrintSymbol} if"
	 "dup 7 eq {PrintLozenge} if"
	 "dup 4 eq {exit} if"
	 "cleartomark"
	 "mark"
	 "}"
	 "{exit} ifelse"
	 "}loop"
	 "showpage"
	 "} bind def"
	 "/PrintScreenImage"
	 "%Do screen dump"
	 "{Startup"
	 "/astr 10 string def             %rotation angle"
	 "/ustr 10 string def             %user scale factor"
	 "/wstr 10 string def             %width"
	 "/hstr 10 string def             %height"
	 "/pcstr 10 string def             %picstr-width"
	 "/shstr 10 string def            %scaled height"
	 "/swstr 10 string def            %scaled width"
	 "/pwstr 10 string def            %pic-width"
	 "/my-in (%stdin) (r) file def"
	 "my-in astr readline pop cvi rotate  %get rotation (0=portrait;90=landscape)"
	 "my-in ustr readline pop "
	 "dup cvr dup cvr scale           %set scaling (magnification) factor"	
	 "my-in wstr readline pop"
	 "my-in hstr readline pop"
	 "% my-in swstr readline pop"
	 "my-in shstr readline pop"
	 "50 shstr cvr translate wstr cvr hstr cvr scale %scaling and translation"
	 "my-in pcstr readline pop"
	 "/picstr pcstr cvi string def"
	 "my-in pwstr readline pop"
	 "/screenimage {                  %make screen image array"
	 "pwstr cvi hstr cvi 1 [pwstr cvi 0 0 hstr cvi neg 0 hstr cvi]"
	 "{ my-in picstr readhexstring pop}"
	 "image } def"
	 "screenimage"
	 "} bind def"
	 )))

;;;Load the Postscript variable into *postscript-program-in-buffer*
;;;THis is only necessary at load time. This will allow you to patch SYS: Printer; Postscript-driver.text.
;;; Patch Procedure: edit postscript-driver.text, save to SYS; printer; , create a printer patch with the following 
;;; function call.
(build-postscript-source-variable)
