;;; -*- Mode: common-LISP; Package: PRINTER; Base: 10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; This file defines print devices:  TI2015

;;; CHANGE HISTORY:
;;; 
;;;  Date      Author   Description
;;; -------------------------------------------------------------------------------------

;;;  01-24-89   DAB     Added support for controlling tab-width thru args option or mode-line.
;;;  05-31-88   DAB     Doc string was reversed in *explorer 2015-font-map*.
;;;  04-07-88   DAB     Added support for font selection using the global font numbers available on ROM 4.2.
;;;  2/12/87    LCO     Removed null call of :print-page-heading in ti2015-graphics-mixin flavor 
;;;                     which shadowed printing of page-headings for text. Incorporated the
;;;                     following code-read suggestions:
;;;                     1. Explicitly calculate Defparameter *Increments-per-pixel*;
;;;                     2. Definition of Line-buffer removed from Ti2015-graphics-mixin defflavor, and
;;;                        placed in :before :init method.
;;;  2/27/87    LCO     Added method :set-char-pitch to shadow 855 and basic-printer flavors.
;;;  3/03/87    LCO     Added new methods to enable control over left and top margins: 
;;;                        :set-text-margins
;;;                        :tab-to-left-margin,
;;;                        :set-top-margin.
;;;                     Also modified existing methods (shadowing basic-printer's):
;;;                        :tyo-char,       :tab,   
;;;                        :start-new-line, :setup-normal-mode,
;;;                        :start-new-page.
;;;  3/25/87    LCO     Completed fixed margin control changes. 
;;;

(DEFPROP TI2015 (75 100 150 300) DOTS-PER-INCH) 


(DEFPARAMETER *PRINTABLE-PIXELS-ACROSS-A-PAGE* 2400.0 "# of printable pixels across a page - 8 inches.")
(DEFPARAMETER *PRINTABLE-PIXELS-DOWN-A-PAGE* 3150.0 "# of printable pixels down a page - 10.5 inches.") 
(DEFPARAMETER *INCREMENTS-IN-A-LINE* 5760.0 "# of 1/720ths of an inch increments in an 8.0 inch wide line.") 
(DEFPARAMETER *INCREMENTS-DOWN-A-PAGE* 7560.0 "# of 1/720ths of an inch increments in a 10.5 inch long page.")
(DEFPARAMETER *VERTICAL-INCREMENTS* 495.0 "# of 1/720ths of an inch increments to adjust margin vertically.")
(DEFPARAMETER *HORIZONTAL-INCREMENTS* 135.0 "# of 1/720ths of an inch increments to adjust margin horizontally.")
(DEFPARAMETER *INCREMENTS-PER-PIXEL* (/ *INCREMENTS-IN-A-LINE* *PRINTABLE-PIXELS-ACROSS-A-PAGE*))


;;; Available Font/SYMBOL SETS
;;;Courier 10 pitch,PC                  - 0031   Courier 10 Pitch, R8    - 0034
;;;Courier 12 pitch,PC                  - 0032   Courier 12 Pitch, R8    - 0035
;;;Courier 16.7 pitch,PC                - 0033   Courier 16.7 Pitch, R8  - 0036
			       
;;;COURIER 10 Pitch,IBM256           - 0080
;;;COURIER 10 Pitch,USACII           - 0081
;;;COURIER 10 Pitch,HP-Legal         - 0082 
;;;COURIER 10 Pitch,HP-ROMAN8        - 0083 
;;;COURIER 10 Pitch Bold,IBM256      - 0084 COURIER 12 Pitch Bold,IBM256      - 008C
;;;COURIER 10 Pitch Bold,USACII      - 0085 COURIER 12 Pitch Bold,USACII      - 008D
;;;COURIER 10 Pitch Bold,HP-Legal    - 0086 COURIER 12 Pitch Bold,HP-Legal    - 008E
;;;COURIER 10 Pitch Bold,HP-ROMAN8   - 0087 COURIER 12 Pitch Bold,HP-ROMAN8   - 008F  
;;;COURIER 10 Pitch Italics,IBM256   - 0088 COURIER 12 Pitch Italics,IBM256   - 0090 
;;;COURIER 10 Pitch Italics,USACII   - 0089 COURIER 12 Pitch Italics,USACII   - 0091  
;;;COURIER 10 Pitch Italics,HP-Legal - 008A COURIER 12 Pitch Italics,HP-Legal - 0092 
;;;COURIER 10 Pitch Italics,HP-ROMAN8- 008B COURIER 12 Pitch Italics,HP-ROMAN8- 0093 

;;;Omni-Vetica 6 point,USACII        - 0400   Omni-Vetica 8 point,USACII        - 0401
;;;Omni-Vetica 8 point Bold,USACII   - 0402   Omni-Vetica 10 point,USACII       - 0403
;;;Omni-Vetica 10 point,HP-Legal     - 040C   Omni-Vetica 10 point Bold,USACII  - 0404
;;;Omni-Vetica 10 point Bold,HP-Legal- 040D   Omni-Vetica 10 point Italics,USACII-0405
;;;Omni-Vetica 10 point Italics,HP-leagl-040E Omni-Vetica 12 point Bold,USACII  - 0406
;;;Omni-Vetica 12 point Bold,HP-Leagl- 040F   Omni-Vetica 14 point Bold,USACII  - 0407
;;;Omni-Vetica 14 point Bold,USACII  - 0408   Omni-Vetica 14 point Bold ,USASCII HPCrtB - 0364
;;;Omni-Vetica 14 point Bold, HP-Roman8 HPCrtF - 0354 
;;;Omni-Vetica 6 point, USASCII HPCrtT - 0370 Omni-Vetica 8 point, USASCII HPCrtT - 0371
;;;Omni-Vetica 8 point Bold, USASCII HPCrtT - 0372 Omni-Vetica 10 point Bold, USASCII HPCrtT - 0373
;;;Omni-Vetica 12 point Bold, USASCII HPCrtT- 0374 Omni-Vetica 14 point Bold, USASCII HPCrtT - 0375

;;;Letter Gothic 10 pitch,IBM256     - 0100  Letter Gothic 15 pitch,IBM256           - 0110
;;;Letter Gothic 10 pitch,USASCII    - 0101  Letter Gothic 15 pitch,IBM256           - 0111
;;;Letter Gothic 10 pitch,HP-legal   - 0102  Letter Gothic 15 pitch,IBM256           - 0112
;;;Letter Gothic 10 pitch,HP-ROMAN8  - 0103  Letter Gothic 15 pitch,IBM256           - 0113
;;;Letter Gothic 12 Pitch,IBM256     - 0104  Letter Gothic 12 Pitch Bold,IBM256      - 0108
;;;Letter Gothic 12 Pitch,USACII     - 0105  Letter Gothic 12 Pitch Bold,USACII      - 0109
;;;Letter Gothic 12 Pitch,HP-Legal   - 0106  Letter Gothic 12 Pitch Bold,HP-Legal    - 010A
;;;Letter Gothic 12 Pitch,HP-ROMAN8  - 0107  Letter Gothic 12 Pitch Bold,HP-ROMAN8   - 010B  
;;;Letter Gothic 12 Pitch Italics,IBM256   - 010C Letter Gothic 12 Pitch Italics,IBM256   - 0090 
;;;Letter Gothic 12 Pitch Italics,USACII   - 010D Letter Gothic 12 Pitch Italics,USACII   - 0091  
;;;Letter Gothic 12 Pitch Italics,HP-Legal - 010E Letter Gothic 12 Pitch Italics,HP-Legal - 0092 
;;;Letter Gothic 12 Pitch Italics,HP-ROMAN8- 010F Letter Gothic 12 Pitch Italics,HP-ROMAN8- 0093 

;;;Prestige Elite 12 pitch,IBM256    - 0200  Prestige Elite 12 pitch Bold,IBM256    - 0204
;;;Prestige Elite 12 pitch,USASCII   - 0201  Prestige Elite 12 pitch Bold,USASCII   - 0205
;;;Prestige Elite 12 pitch,HP-Legal  - 0202  Prestige Elite 12 pitch Bold,HP-Legal  - 0206
;;;Prestige Elite 12 pitch,HP-Roman8 - 0203  Prestige Elite 12 pitch Bold,HP-Roman8 - 0207
;;;Prestige Elite 12 pitch Italics,IBM256  - 0208  Prestige Elite 16.7 pitch,IBM256 - 020C
;;;Prestige Elite 12 pitch Italics,USASCII - 0209  Prestige Elite 16.7 pitch,USASCII- 020D  
;;;Prestige Elite 12 pitch Italics,HP-Legal -020A  Prestige Elite 16.7 pitch,HP-Legal-020E 
;;;Prestige Elite 12 pitch Italics,HP-Roman8-020B  Prestige Elite 16.7 pitch,HP-Roman8-020F

;;;Omni-Roman 8 point,USACII          - 0500  Omni-Roman 10 point,USACII            - 0503
;;;Omni-Roman 8 point,HP-Legal        - 0501  Omni-Roman 10 point,HP-Legal          - 0504
;;;Omni-Roman 8 point,HP-Roman        - 0502 
;;;Omni-Roman 10 point Bold,USACII    - 0506  Omni-Roman 10 point Bold,HP-Legal     - 0507
;;;Omni-Roman 10 point Italics,USACII - 0509  Omni-Roman 10 point Italics,HP-Legal  - 050A
;;;Omni-Roman 10 point, USACII HPCrtB - 0360  Omni-Roman 10 point Bold, USACII HPCrtB-0361
;;;Omni-Roman 10 point Italics, USACII HPCrtB - 0362 Omni-Roman 8 point, USACII HPCrtB - 0363
;;;Omni-Roman 10 point, HP-Roman8 HPCrtF - 0350 Omni-Roman 10 point Bold, HP-Roman8 HPCrtF - 0351
;;;Omni-Roman 10 point Italics, HP-Roman8 HPCrtF - 0352 Omni-Roman 8 point, HP-Roman8 HPCrtF - 0353

;;;Greek/math 12 pitch 10 point (Math 7) - 0300  Greek/math 10 pitch 10 point (Math 7) - 0301
;;;OCR-A 10 pitch 12 point (USASCII)   - 0302 OCR-B 10 pitch 12 point (USASCII)     - 0303
;;;H-P Line Drawing 10 pitch 14 point  - 0305 H-P Line Drawing 10 pitch 12 point    - 0306
;;;H-P Line Drawing 12 pitch 12 point  - 0307 H-P Tax Line Drawing 10 pitch 12 point- 0308

;;;Bar Code 3 of 9 4.6 pitch           -0309  Bar Code 3 of 9 9.3 pitch           -030A
;;;Bar Code 3 EAN/UPC 10 mil           -030B  Bar Code 3 EAN/UPC 13 mil           -030C
;;;			       


(DEFVAR *Explorer-2015-font-map*    ;05-31-88 DAB Doc string was reversed.
	'(
	  ("5X5" ("Courier 10" "0031"))
	  ("BIGFNT" ("Courier 16.7" "0036"))
	  ("CMB8" ("Courier 10B" "0084"))
	  ("CMDUNH" ("Courier 16.7" "0036"))
	  ("CMOLD" ("Courier 10B" "0084"))
	  ("CMR10" ("Courier 16.7" "0036"))
	  ("CMR18" ("Courier 10B" "0084"))
	  ("COURIER" ("Courier 10" "0031"))
	  ("CPTFONT" ("Courier 10" "0031"))
	  ("CPTFONTB" ("Courier 10B" "0084"))
	  ("CPTFONTBI" ("Courier 10B" "0084"))
	  ("CPTFONTI" ("Courier 10I" "0089"))
	  ("DEFAULT" ("Courier 10" "0031"))
	  ("HIGHER-MEDFNB" ("Courier 10B" "0084"))
	  ("HIGHER-TR8" ("Courier 10" "0031"))
	  ("HL10" ("Vetica 14B" "0407"))
	  ("HL10B" ("Vetica 14B" "0407"))
	  ("HL12" ("Vetica 14B" "0407"))
	  ("HL12B" ("Vetica 14B" "0407"))
	  ("HL12BI" ("Vetica 14B" "0407"))
	  ("HL12I" ("Vetica 14B" "0407"))
	  ("HL6" ("Vetica 14B" "0407"))
	  ("HL7" ("Vetica 14B" "0407"))
	  ("MEDFNB" ("Courier 10B" "0084"))
	  ("MEDFNT" ("Courier 10B" "0084"))
	  ("MEDFNTB" ("Courier 10B" "0084"))
	  ("MEDFNTBI" ("Courier 10B" "0084"))
	  ("MEDFNTI" ("Courier 10I" "0089"))
	  ("METS" ("Courier 16.7" "0033"))
	  ("METSB" ("Courier 10B" "0084"))
	  ("METSBI" ("Courier 10B" "0084"))
	  ("METSI" ("Courier 10I" "0089"))
	  ("TINY" ("vetica 14B" "0407"))
	  ("TR10" ("Courier 10" "0031"))
	  ("TR10B" ("Courier 10B" "0084"))
	  ("TR10BI" ("Courier 10B" "0084"))
	  ("TR10I" ("Courier 10I" "0089"))
	  ("TR12" ("Courier 10" "0031"))
	  ("TR12B" ("Courier 10B" "0084"))
	  ("TR12BI" ("Courier 10B" "0084"))
	  ("TR12I" ("Courier 10I" "0089"))
	  ("TR18" ("Courier 16.7" "0033"))
	  ("TR18B" ("Courier 10B" "0084"))
	  ("TR8" ("Courier 10" "0031"))
	  ("TR8B" ("Courier 10B" "0084"))
	  ("TR8I" ("Courier 10I" "0089"))
	  ("TVFONT" ("Courier 10" "0031"))
	  ("WIDER-FONT" ("vetica 24B" "0408"))
	  ("WIDER-MEDFNT" ("Courier 12" "0032")))
  "This variable maps the Explorer fonts to fonts available on the TI2015 printer. The user may edit this
         variable to customizes the map based on his/her available font cartridges. The format of this list is as follows:
         ( (<Explorer font> (<font description> <global number>)) .. ) .All entry must be strings. The global number
         is assigned by TI. To get the global number place the cartridge in a slot, press on-line, then press test."
)


(defvar ti2015-socket-mode-font-number-list ())

(defvar *TI2015-ROM-VERSION* 4.0)   ;Font selection is only available for ROM version 4.2 or greater.



(DEFVAR *LEFT-MARGIN-VAR* 0)
(DEFVAR *TOP-MARGIN-VAR* 0)


(DEFFLAVOR BASIC-TI2015
	   ((HTAB 9)
	    (LF 10)
	    (FF 12)
	    (CR 13)
	    (DLE 16); used in some escape sequences
	    (ESC 27)
            (NCOPIES 1)
	    (LEFT-MARGIN 5)
	    (TOP-MARGIN 1)
	    (LAST-CHAR #\PAGE)
	    (PRINT-WIDE NIL)
 	    )
	   (TI855
	    BASIC-PRINTER)
  :SETTABLE-INSTANCE-VARIABLES)

(DEFMETHOD (BASIC-TI2015 :around :build-font-tables) (cont mt &rest arg)  ;new method to support font selection on rom 4.2
  
  (MULTIPLE-VALUE-bind (FONT-TO-SOCKET-ALIST SOCKET-AND-MODE-LIST)
    (apply 'funcall-with-mapping-table cont mt arg)
    (when (and *TI2015-ROM-VERSION* (> *TI2015-ROM-VERSION* 4.19))
      (dolist (font-pair  SOCKET-AND-MODE-LIST)
	(push (list (car font-pair)
		    (or (cadr (assoc (third font-pair)  *Explorer-2015-font-map* :test #'string-equal))
			'("Courier 10" "0031")))
	      ti2015-socket-mode-font-number-list)))
    (values FONT-TO-SOCKET-ALIST  SOCKET-AND-MODE-LIST)
    ))

(DEFMETHOD (BASIC-TI2015 :around :Send-font-sequence) (cont mt oper SOCKET-NUMBER &rest arg)
   (apply 'funcall-with-mapping-table cont mt oper SOCKET-NUMBER arg)
   (when (and *TI2015-ROM-VERSION* (> *TI2015-ROM-VERSION* 4.19))
    (SEND SELF :TYO-RAW #x1B)
    (SEND SELF :TYO-RAW #x10)
    (SEND SELF :TYO-RAW #x47)
    (send self :string-out-raw (or (second
				     (cadr
				       (assoc (car SOCKET-NUMBER) ti2015-socket-mode-font-number-list :test #'EQ)))
				   "0031"))
    ))

;Note: currently we force values for the top and left margins, depending upon the cpi.
;A possible later fix to the print-functions (and requests) to pass the options list
;or create new :left-margin & :top-margin keywords, will enable user control of these
;--with some minimal modification of the 2015's methods.
(DEFMETHOD (BASIC-TI2015 :SET-TEXT-MARGINS) (LEFT-MARGIN-COLUMN-NUM TOP-MARGIN-NUM-OF-LINES-DOWN-PAGE)
  "Set values for left and top margins, and the number of characters per line."
  (IF (PLUSP LEFT-MARGIN-COLUMN-NUM) (SETQ *LEFT-MARGIN-VAR* LEFT-MARGIN-COLUMN-NUM))
  (IF (PLUSP TOP-MARGIN-NUM-OF-LINES-DOWN-PAGE) (SETQ *TOP-MARGIN-VAR* TOP-MARGIN-NUM-OF-LINES-DOWN-PAGE))
  (WHEN (>= *TOP-MARGIN-VAR* LINES-PER-PAGE) 
    (SETQ *TOP-MARGIN-VAR* 1))  ;if specified top-margin exceeds text length, reset to default top-margin.
  (WHEN (OR (< *LEFT-MARGIN-VAR* 1) (>= *LEFT-MARGIN-VAR* (* 8 CHAR-PITCH)))
    (SETQ *LEFT-MARGIN-VAR* 5)) ;set default left margin.
  (SELECTOR CHAR-PITCH =
    (12 (SETQ *LEFT-MARGIN-VAR* 6))
    (16 (SETQ *LEFT-MARGIN-VAR* 8))
    (T  (SETQ *LEFT-MARGIN-VAR* 5)))		
  (SETQ PRINTER::CHARS-PER-LINE (- (* 8 CHAR-PITCH) *LEFT-MARGIN-VAR*)));the logical right margin is same as chars-per-line.


(DEFMETHOD (BASIC-TI2015 :TAB-TO-LEFT-MARGIN) (LEFT-MARGIN-COLUMN-NUM)
  "Space to specified left margin column number."
  (LOOP FOR I FROM 1 TO LEFT-MARGIN-COLUMN-NUM DO (SEND SELF :TYO-RAW #\SPACE)))


(DEFMETHOD (BASIC-TI2015 :SET-TOP-MARGIN) (NUM-OF-LINES-DOWN-PAGE)
  "Set top margin."
  (WHEN (PLUSP NUM-OF-LINES-DOWN-PAGE)
    (DOTIMES (I NUM-OF-LINES-DOWN-PAGE)
      (SEND SELF :TYO-RAW LF))
    (SETQ LINE-COUNT (+ LINE-COUNT NUM-OF-LINES-DOWN-PAGE))))


;;;--------------------------------------------------------
;;;
;;; METHODS SHADOWING TI855 & BASIC-PRINTER FLAVORS
;;;
;;;--------------------------------------------------------


(DEFMETHOD (BASIC-TI2015 :START-DOCUMENT) (HEADER-NAME USER-NAME N-COPIES-TO-PRINT)
  "Sets up printer before printing file//bitmap.  Also prints header page."
  (SETQ NCOPIES N-COPIES-TO-PRINT)
  (SEND SELF :SETUP-INITIAL-PROPERTIES)
  (SEND SELF :SETUP-NORMAL-MODE)
  (WHEN HEADER-NAME
    (SEND SELF :PRINT-HEADER-PAGE HEADER-NAME USER-NAME)))
 

(DEFMETHOD (BASIC-TI2015 :END-DOCUMENT) (&REST IGNORE)
  "Sets up printer after printing filebitmap: shadows basic-printer's method."
  (SEND SELF :FORM)
  (SEND SELF :RESET-TI855-DP-INTERFACE))


(DEFMETHOD (BASIC-TI2015 :SETUP-NORMAL-MODE) ()
  "Sets up the printer's cpi, lpi, form-length, form-width."
  (SEND SELF :CANCEL-BOLD-PRINT)
  (SEND SELF :CANCEL-EXPANDED-PRINT)
  (SEND SELF :CANCEL-SHADOW-PRINT)
  (SEND SELF :CANCEL-COMPRESSED-PRINT)
  (SEND SELF :SET-CHAR-PITCH CHAR-PITCH)
  (SEND SELF :SET-TEXT-MARGINS LEFT-MARGIN TOP-MARGIN)
  (SEND SELF :SET-FORM-LENGTH LINE-SPACING)
  (SEND SELF :SET-LINE-SPACING LINE-SPACING)
  )


(DEFMETHOD (BASIC-TI2015 :SET-FORM-LENGTH)(LINES-PER-INCH)
  "Sets form length."
  (SELECTOR LINES-PER-INCH =
    (8 (SETQ LINES-PER-PAGE 80))
    (10 (SETQ LINES-PER-PAGE 100))
    (T (SETQ LINES-PER-PAGE 60))))


(DEFMETHOD (BASIC-TI2015 :SET-LINE-SPACING) (LINES-PER-INCH)
  "Sets lines-per-inch to either 6, 8, or 10."
  (SEND SELF :ESC-COMMAND (SELECTOR LINES-PER-INCH = (6 #\2) (8 #\0) (10 #\1) (OTHERWISE #\2))))


(DEFMETHOD (BASIC-TI2015 :PRINT-TEXT-FILE) (FILE-STREAM &OPTIONAL FONT-LIST)
  "Print a text file withwithout fonts
Font-To-Socket-Alist has entries (<font-name> . <socket-number>),
Socket-And-Mode-List has entries (<socket-number> <mode>),
where <mode> is either NORMAL, BOLD, ITALIC or BOLD-ITALIC"
  (LET (FONT-TO-SOCKET-ALIST
	SOCKET-AND-MODE-LIST
	CURRENT-FONT
	PREVIOUS-FONT
	(ti2015-socket-mode-font-number-list ()))
    (UNLESS FONT-LIST
      (SETQ FONT-LIST (GET-FONT-ATTRIBUTE-LIST FILE-STREAM)))
    ;; if font-list is still nil, force a font 
    (UNLESS FONT-LIST (SETQ FONT-LIST 'STARTFONT))
    ;; get socket-and-mode-list from font-list
    (MULTIPLE-VALUE-SETQ (FONT-TO-SOCKET-ALIST SOCKET-AND-MODE-LIST)
      (SEND SELF :BUILD-FONT-TABLES FONT-LIST))
    ;; setup to use the first font in list, if it exists
    (WHEN SOCKET-AND-MODE-LIST
      (SETQ CURRENT-FONT 0)
      (SEND SELF :SEND-FONT-SEQUENCE (FIRST SOCKET-AND-MODE-LIST) ()))
    (SETQ CURRENT-FONT (IF SOCKET-AND-MODE-LIST
			 0
			 ()))
    (DO ((IN-CHAR)
	 (FONT-STACK (MAKE-ARRAY 50 :LEADER-LIST '(0)))
	 (SML-LENGTH (LENGTH SOCKET-AND-MODE-LIST)))
	((NULL (SETQ IN-CHAR (SEND FILE-STREAM :TYI))))  ;return a numuber.
      (COND
	((AND CURRENT-FONT (= IN-CHAR (char-int #\)))  ;02-18-88 DAB  char-init around epilison.
	 (SETQ IN-CHAR (SEND FILE-STREAM :TYI))
	 (COND
	   ((<= (char-int #\0) IN-CHAR (char-int #\9))
	    ;; <epsilon> (ie, 6) is followed by a digit <n>, 
	    ;; switch to font <n>
	    (UNLESS (VECTOR-PUSH CURRENT-FONT FONT-STACK)
	      (COPY-ARRAY-PORTION FONT-STACK 20 (ARRAY-TOTAL-SIZE FONT-STACK) FONT-STACK 0
				  (- (ARRAY-TOTAL-SIZE FONT-STACK) 20))
	      (DECF (FILL-POINTER FONT-STACK) 20)
	      (VECTOR-PUSH CURRENT-FONT FONT-STACK))
	    (SETQ PREVIOUS-FONT CURRENT-FONT)
	    (LET ((POS (- IN-CHAR (char-int #\0))))
	      (IF (< POS SML-LENGTH)
		(SETQ CURRENT-FONT POS)))
	    (UNLESS (= CURRENT-FONT PREVIOUS-FONT)
	      (SEND SELF :SEND-FONT-SEQUENCE (NTH CURRENT-FONT SOCKET-AND-MODE-LIST)
		 (NTH PREVIOUS-FONT SOCKET-AND-MODE-LIST))))
	   ((= IN-CHAR (char-int #\*))
	    ;; <epsilon> is followed by a *,
	    ;; switch to previously used font
	    (SETQ PREVIOUS-FONT CURRENT-FONT)
	    (OR (ZEROP (FILL-POINTER FONT-STACK)) (SETQ CURRENT-FONT (VECTOR-POP FONT-STACK)))
	    (UNLESS (= CURRENT-FONT PREVIOUS-FONT)
	      (SEND SELF :SEND-FONT-SEQUENCE (NTH CURRENT-FONT SOCKET-AND-MODE-LIST)
		 (NTH PREVIOUS-FONT SOCKET-AND-MODE-LIST))))
	   (T
	    ;; just print <epsilon> and character following it
	    (SEND SELF :TYO-CHAR #\) (SEND SELF :TYO-CHAR IN-CHAR))))
	(T (SEND SELF :TYO-CHAR IN-CHAR))))
    ;; since we're done, setup to use the default font socket module and mode NORMAL
    (IF CURRENT-FONT
      (SEND SELF :SEND-FONT-SEQUENCE (LIST FONT-MODULE 'NORMAL)
	 (NTH CURRENT-FONT SOCKET-AND-MODE-LIST)))))



(DEFMETHOD (BASIC-TI2015 :SET-CHAR-PITCH) (CHARS-PER-INCH)
  "Sets char-pitch to 10, 12, or 16.  Also sets chars-per-line."
  (COND
    ((< CHARS-PER-INCH 12) (SEND SELF :ESC-COMMAND #\y) (SETQ CHAR-PITCH 10))
    ((< CHARS-PER-INCH 16) (SEND SELF :ESC-COMMAND #\z) (SETQ CHAR-PITCH 12))
    (T (SEND SELF :SELECT-COMPRESSED-PRINT) (SETQ CHAR-PITCH 16)))
  (SETQ CHARS-PER-LINE (- (* 8 CHAR-PITCH) *LEFT-MARGIN-VAR*)))



(DEFMETHOD (BASIC-TI2015 :TYO-CHAR) (CHAR &AUX PRINTABLE)
  "Before sending out a character, look for special characters. Also, implement end-of-line wraparound."
  (WHEN (OR (AND (SETQ PRINTABLE (AND (>= CHAR #o40) (< CHAR #o200)))    ; If printable, i.e., space, 0-9, a-z, A-Z, etc.   
		 (> (SETQ CHAR-COUNT (1+ CHAR-COUNT)) CHARS-PER-LINE))   ;   and there is no more room on the line,
	    (AND (EQL #\TAB (INT-CHAR CHAR))                             ; or a tab character and the next tab
		 (> (+ (REM CHAR-COUNT 8) CHAR-COUNT) CHARS-PER-LINE)))  ;   stop is beyond the end of the line,
    (SEND SELF :START-NEW-LINE))                                         ; then start a new line. 
  (COND
    (PRINTABLE                              ; Handle the easy stuff first.
     (SEND SELF :TYO-RAW CHAR))             ; 
    ((EQL (INT-CHAR CHAR) #\TAB)            ; TAB is the next most frequently used character,
     (SEND SELF :TAB))                      ;    so handle it next.
    ((OR (EQL (INT-CHAR CHAR) #\NEWLINE)    ; On Lisp Machine CR really means CR-LF.
	 (EQL (INT-CHAR CHAR) #\LINEFEED)   ;    Make LF mean the same thing.
	 (EQL CHAR 13))                     ;    ^M (which means CR ONLY on 2060/VAX) gets
                                            ;    interpreted on Lisp Machines as octal 15, but
                                            ;    #o15 is the same as #\circle-plus on Lisp
                                            ;    Machines; can't have both...
     (SEND SELF :START-NEW-LINE))
    ((eql (int-char CHAR) #\PAGE) (SEND SELF :START-NEW-PAGE))
    ((eql (int-char CHAR) #\BACKSPACE) (SEND SELF :TYO-RAW PRINTER::BACKSPACE))
    (T; chars like <PAGE>, <ABORT>, etc. 
     (SEND SELF :SIMULATE-LISPM-CHAR CHAR)))
  (SETQ LAST-CHAR CHAR))


(DEFMETHOD (BASIC-TI2015 :START-NEW-LINE)()
  "Start a new line."
    (IF (> (SETQ LINE-COUNT (1+ LINE-COUNT)) LINES-PER-PAGE)
	(SEND SELF :START-NEW-PAGE)     ; start a new page
        (PROGN
	  (SETQ CHAR-COUNT 0)           ; otherwise, start a new line
	  (SEND SELF :TYO-RAW LF)
	  (SEND SELF :TAB-TO-LEFT-MARGIN *LEFT-MARGIN-VAR*)))); simulate a left margin


(DEFMETHOD (BASIC-TI2015 :START-NEW-PAGE) (&OPTIONAL OVERRIDE-PAGE-HEADING)
  "Start a new page."
  (SEND SELF :FORM)
  (SETQ LINE-COUNT 0)
  (SETQ CHAR-COUNT 0)
  (SEND SELF :SET-TOP-MARGIN *TOP-MARGIN-VAR*)
  (SEND SELF :TAB-TO-LEFT-MARGIN *LEFT-MARGIN-VAR*)
  (UNLESS OVERRIDE-PAGE-HEADING
    (SEND SELF :DO-PAGE-HEADING)))


(DEFMETHOD (BASIC-TI2015 :PRINT-HEADER-PAGE) (FILENAME USERNAME &AUX (TEXT-CHAR-PITCH CHAR-PITCH) 
					       (TEXT-LINE-SPACING LINE-SPACING) (TEXT-LEFT-MARGIN *LEFT-MARGIN-VAR*)
					       (TEXT-CURRENT-FONT CURRENT-FONT-SOCKET-AND-MODE)
					       (HEADER-FONT (LIST FONT-MODULE 'NORMAL)))
  "Sets pitch for header, then prints filename, username, and current date on one page."
  (SEND SELF :SET-CHAR-PITCH 10); always select pitch of 10 cpi for header.
  (SEND SELF :SET-LINE-SPACING 6); always select 6 lpi for header.
  (SETQ *LEFT-MARGIN-VAR* 4); always select left-margin = 4 for header.
  (SEND SELF :SEND-FONT-SEQUENCE HEADER-FONT TEXT-CURRENT-FONT); switch to header font.
  (SEND SELF :SELECT-EXPANDED-PRINT)
  ;; output header strings
  (DOTIMES (I 5)  
    (SEND SELF :STRING-OUT-CHARS
	  (FORMAT () "~%~15A~17A" USERNAME PRINTER:CURRENT-TIME)))
  (SEND SELF :CANCEL-EXPANDED-PRINT)
  (DOTIMES (I 2)
    (SEND SELF :STRING-OUT-CHARS 
	  (FORMAT () "~%~70,,,'_A" '_)))
  (SEND SELF :STRING-OUT-CHARS
	(FORMAT () "~2%~5Towner~20T~A~%~5Tprinted-on~20T~A~%~5Tname~20T~A~%~5Tcreated-on~@[~20T~A~]~2%" USERNAME PRINTER:CURRENT-TIME FILENAME (IF (TYPEP FILENAME 'PATHNAME)
					      (TIME:PRINT-UNIVERSAL-TIME 
						(GET (SEND FILENAME :PROPERTIES) :CREATION-DATE) ()))))
  (DOTIMES (I 2)
    (SEND SELF :STRING-OUT-CHARS 
	  (FORMAT () "~70,,,'_A~%" '_)))
  (SEND SELF :SELECT-EXPANDED-PRINT)
  (DOTIMES (I 5)  
    (SEND SELF :STRING-OUT-CHARS                                                                
	  (FORMAT () "~%~15A~17A" USERNAME PRINTER:CURRENT-TIME)))
  ;;
  (SEND SELF :CANCEL-EXPANDED-PRINT)
  ;;reset values
  (SETQ *LEFT-MARGIN-VAR* TEXT-LEFT-MARGIN)
  (SEND SELF :SEND-FONT-SEQUENCE TEXT-CURRENT-FONT HEADER-FONT)
  (SEND SELF :SET-CHAR-PITCH TEXT-CHAR-PITCH)
  (SEND SELF :SET-LINE-SPACING TEXT-LINE-SPACING)
  )


(DEFMETHOD (BASIC-TI2015 :PRINT-PAGE-HEADING) (&AUX FIRST-LINE SECOND-LINE 
					       (TEXT-CURRENT-FONT CURRENT-FONT-SOCKET-AND-MODE)
					       (PAGE-HEADING-FONT (LIST FONT-MODULE 'NORMAL)) 
					       FORMAT-STRING FORMAT-LINE)
  "Print page heading after formatting page-heading string based upon text's character pitch.  
   Increment page count. Switch to page-heading's font and then back to text's."
  (SETQ PAGE-COUNT (1+ PAGE-COUNT))
  (IF (CONSP PRINTER::PAGE-HEADING)
    (SETQ FIRST-LINE (FIRST PRINTER::PAGE-HEADING)
	  SECOND-LINE (SECOND PRINTER::PAGE-HEADING))
    (SETQ FIRST-LINE PRINTER::PAGE-HEADING))
  ;; select and set up page-heading string based upon character pitch
  (SELECTOR PRINTER::CHAR-PITCH =
    (12 (AND (SETQ FORMAT-STRING "~%~54A~18A Page -~D-~@[~%~A~]")
	     (SETQ FORMAT-LINE "~%~85,,,'_A~2%")))
    (16 (AND (SETQ FORMAT-STRING "~%~72A~18A Page -~D-~@[~%~A~]")
	     (SETQ FORMAT-LINE "~%~110,,,'_A~2%")))
    (T (AND (SETQ FORMAT-STRING "~%~45A~18A Page -~D-~@[~%~A~]")
	    (SETQ FORMAT-LINE "~%~72,,,'_A~2%"))))
  (IF (NOT TEXT-CURRENT-FONT)
      (PROGN (SEND SELF :STRING-OUT-CHARS
		   (FORMAT () FORMAT-STRING FIRST-LINE PRINTER::CURRENT-TIME PRINTER::PAGE-COUNT SECOND-LINE))
	     (SEND SELF :STRING-OUT-CHARS (FORMAT () FORMAT-LINE '_)))
      (PROGN 
	(SEND SELF :SEND-FONT-SEQUENCE PAGE-HEADING-FONT TEXT-CURRENT-FONT); set to page-heading font
	(SEND SELF :STRING-OUT-CHARS
	      (FORMAT () FORMAT-STRING FIRST-LINE PRINTER::CURRENT-TIME PRINTER::PAGE-COUNT SECOND-LINE))
	(SEND SELF :STRING-OUT-CHARS (FORMAT () FORMAT-LINE '_))
	(SEND SELF :SEND-FONT-SEQUENCE TEXT-CURRENT-FONT PAGE-HEADING-FONT))); reset to text's font
  )


(DEFMETHOD (BASIC-TI2015 :TAB) (&AUX TAB-COUNT)
  "Space to tab mark."
  (unless (si:integerp tab-width) (setf tab-width *default-tab-width*)) ; DAB 01-24-89
  (SETQ TAB-COUNT (-  tab-width (REM CHAR-COUNT  tab-width))) ; DAB 01-24-89
  (LOOP FOR I FROM 1 TO TAB-COUNT DO (SEND SELF :TYO-RAW #\SPACE))
  (SETQ CHAR-COUNT (+ CHAR-COUNT TAB-COUNT)))


(DEFMETHOD (BASIC-TI2015 :ESC-COMMAND) (COMMAND-CHAR &OPTIONAL N1 N2)
  "Send <ESC> <command-char> <n1> <n2>, where <n1> and <n2> are optional"
  (SEND SELF :TYO-RAW #x1B)   ;02-18-88 DAB
  (SEND SELF :TYO-RAW  COMMAND-CHAR)
  (IF N1
    (SEND SELF :TYO-RAW N1))
  (IF N2
    (SEND SELF :TYO-RAW N2))) 


(DEFMETHOD (BASIC-TI2015 :EXTENDED-COMMAND) (COMMAND-CHAR &REST CHAR-LIST)
  "Send <ESC> <command-char> <char-list>"
  (SEND SELF :TYO-RAW ESC)
  (SEND SELF :TYO-RAW COMMAND-CHAR)
  (LOOP FOR CHAR IN CHAR-LIST DO (SEND SELF :TYO-RAW CHAR)))

;;;
;;;--------------------------------------------------------
;;;


(DEFMETHOD (BASIC-TI2015 :RESET-HP-LASERJET-INTERFACE) ()
  "Resets HP-LaserJet software interface."
  (SEND SELF :ESC-COMMAND #\E))


(DEFMETHOD (BASIC-TI2015 :RESET-TI855-DP-INTERFACE) ()
  "Resets TI855-DP software interface."
  (SEND SELF :ESC-COMMAND #\@))


(DEFMETHOD (BASIC-TI2015 :RESET-TI855-WP-INTERFACE) ()
  "Resets TI855-DP software interface."
  (SEND SELF :ESC-COMMAND CR #\P))


(DEFMETHOD (BASIC-TI2015 :SETUP-INITIAL-PROPERTIES) ()
  "Sets up the printer's initial global properties."
  (SEND SELF :ESC-COMMAND DLE #\T #\0); select limitless paper mode to automatically switch trays
;  (SEND SELF :ESC-COMMAND DLE #\E #\3); select alternating paper-eject-position
  (SEND SELF :SELECT-TI855-DP-INTERFACE)); select TI855-DP software interface & defaults


(DEFMETHOD (BASIC-TI2015 :SELECT-TI855-DP-INTERFACE) ()
  "Select TI855-DP software interface & defaults."
  (SEND SELF :ESC-COMMAND DLE #\3))


(DEFMETHOD (BASIC-TI2015 :SELECT-HP-LASERJET-INTERFACE) ()
  "Select HP-LaserJet software interface & defaults."
   ; Defaults are:
   ;    primary & secondary symbol sets   = Roman-8;
   ;     ---        ---     font weight   = medium;
   ;     ---        ---     font pitch    = 10 cpi;
   ;     ---        ---     font style    = upright;
   ;     ---        ---     font typeface = Courier;
   ;     ---        ---     font height   = 12 point;
   ;    left margin       = 0 column;
   ;    line spacing      = 6 lpi;
   ;    raster resolution = 75 dpi
   ;    orientation       = portrait 
  (SEND SELF :ESC-COMMAND DLE #\6))


(DEFMETHOD (BASIC-TI2015 :SET-ABSOLUTE-VTAB-INCREMENT) (&OPTIONAL (INCREMENT 0))
  "Set absolute vertical tab using 1/120th inch increments."
  (SEND SELF :EXTENDED-COMMAND #\& #\a)
  (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" INCREMENT))
  (SEND SELF :TYO-RAW #\V))


(DEFMETHOD (BASIC-TI2015 :SET-ABSOLUTE-HTAB-INCREMENT) (&OPTIONAL (INCREMENT 0))
  "Set absolute horizontal tab using 1/120th inch increments."
  (SEND SELF :EXTENDED-COMMAND #\& #\a)
  (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" INCREMENT))
  (SEND SELF :TYO-RAW #\H))


(DEFMETHOD (BASIC-TI2015 :SET-RELATIVE-VTAB-INCREMENT) (&OPTIONAL (INCREMENT 0))
 "Set relative vertical tab using 1/120th inch increments. If increment is negative, tabs up the 
   page from current logical position; if positive, tabs down."
  (COND ((< INCREMENT 0)
	 (SEND SELF :EXTENDED-COMMAND #\& #\a #\-)
	 (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" INCREMENT))
	 (SEND SELF :TYO-RAW #\V))
	(T
	 (SEND SELF :EXTENDED-COMMAND #\& #\a #\+)
	 (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" INCREMENT))
	 (SEND SELF :TYO-RAW #\V))))


(DEFMETHOD (BASIC-TI2015 :SET-RELATIVE-HTAB-INCREMENT) (&OPTIONAL (INCREMENT 0))
  "Set relative horizontal tab using 1/120th inch increments. If increment is negative, tabs left 
   from current logical position; if positive, tabs right."
  (COND ((< INCREMENT 0)
	 (SEND SELF :EXTENDED-COMMAND #\& #\a #\-)
         (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" INCREMENT))
         (SEND SELF :TWO-RAW #\C))
	(T
	 (SEND SELF :EXTENDED-COMMAND #\& #\a #\+)
         (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" INCREMENT))
         (SEND SELF :TWO-RAW #\C))))


(DEFMETHOD (BASIC-TI2015 :PRINT-CONTROL-CHARS) (&OPTIONAL (NUMBER-OF-BYTES 0))
  (SEND SELF :EXTENDED-COMMAND #\& #\p)
  (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" NUMBER-OF-BYTES))
  (SEND SELF :TYO-RAW #\X))


;Note: 2015 can print multiple copies of each page (up to 99), but then output is not collated;
;so we use the print function's global copy instead.
(DEFMETHOD (BASIC-TI2015 :PRINTS-MULTIPLE-COPIES-P) ()
  "Cannot print multiple copies itself"
  ())


;Note: the printer could be enabled to make its own copies (rather than have the print-daemon do it)
;by calling the following method either immediately after :setup-normal-mode -- to enable the 
;the printing of copies of both text and bitmaps, or before the :print-xxx methods (where xxx is from
;the set {raw-file, text-file, bitmap} -- to enable the printing of copies of either text or bitmaps or
;both.  Unfortunately, we must then set the above method (i.e., :prints-multiple-copies-p) to T,
;and we are forced to handle copies of everything, meaning that for text, we print out uncollated
;copies of each page.
(DEFMETHOD (BASIC-TI2015 :PRINT-MULTIPLE-PAGE-COPIES) (NUMBER-OF-COPIES)
  "Print copies of current and following pages."
  (COND ((<= 1 NUMBER-OF-COPIES 99)
	 (SEND SELF :EXTENDED-COMMAND #\& #\l)
	 (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" NUMBER-OF-COPIES))
	 (SEND SELF :TYO-RAW #\X))
	(T
	 (SEND SELF :EXTENDED-COMMAND #\& #\l)
	 (SEND SELF :STRING-OUT-RAW (FORMAT () "1"))
	 (SEND SELF :TYO-RAW #\X))))





(DEFFLAVOR TI2015-GRAPHICS-MIXIN
	   ((MAX-WIDTH-75-DPI    600) ; i.e., (8 x 75) pixels
     	    (MAX-HEIGHT-75-DPI   787) ; i.e., (floor (10.5 x 75)) 
	    (MAX-WIDTH-100-DPI   800) 
	    (MAX-HEIGHT-100-DPI 1075)
	    (MAX-WIDTH-150-DPI  1200)
	    (MAX-HEIGHT-150-DPI 1575)
	    (MAX-WIDTH-300-DPI  2400) ; i.e., (8 x 300) pixels
	    (MAX-HEIGHT-300-DPI 1200)
	    (LINE-BUFFER))
	   ()
  (:REQUIRED-FLAVORS BASIC-TI2015)
  :SETTABLE-INSTANCE-VARIABLES)


(DEFMETHOD (TI2015-GRAPHICS-MIXIN :BEFORE :INIT) (&REST IGNORE)
  "Create the line-buffer array."
  (SETF LINE-BUFFER (MAKE-ARRAY 200 :ELEMENT-TYPE 'ART-STRING :INITIAL-ELEMENT 0 :FILL-POINTER 0)))


(DEFMETHOD (TI2015-GRAPHICS-MIXIN :SET-RESOLUTION) (DOTS-PER-INCH)
  "Set graphics resolution."
  (SEND SELF :EXTENDED-COMMAND #\* #\t)
  (SEND SELF :STRING-OUT-RAW (FORMAT () "~D" DOTS-PER-INCH))
  (SEND SELF :TYO-RAW #\R))


(DEFMETHOD (TI2015-GRAPHICS-MIXIN :START-RASTER-GRAPHICS) (&OPTIONAL POSITION)
  "Start raster graphics at either current logical position or leftmost absolute print position
   (default)."
  (IF (NULL POSITION) (SEND SELF :EXTENDED-COMMAND #\* #\r #\0 #\A); start at leftmost position 
      (SEND SELF :EXTENDED-COMMAND #\* #\r #\1 #\A))); otherwise start at current logical position


(DEFMETHOD (TI2015-GRAPHICS-MIXIN :END-RASTER-GRAPHICS) ()
  "Signal end of transmission of raster data."
  (SEND SELF :EXTENDED-COMMAND #\* #\r #\B))


(DEFMETHOD (TI2015-GRAPHICS-MIXIN :OUTPUT-MAGNIFIED-BITMAP-TO-PRINTER)
       (BITMAP-ARRAY HEIGHT WIDTH START-X START-Y MAGNIFICATION LANDSCAPEP
	&AUX SOURCE-BIT (MAGNIFIED-HEIGHT (* HEIGHT MAGNIFICATION))
                        (MAGNIFIED-WIDTH (* WIDTH MAGNIFICATION))
			MAGNIFIED-BITMAP-HEIGHT MAGNIFIED-BITMAP-WIDTH)
  "If the bitmap is small enough to warrant a resolution greater than 75 dpi,
   magnify the bitmap."
  (SETQ MAGNIFIED-BITMAP-HEIGHT (* 32 (CEILING MAGNIFIED-HEIGHT 32)))
  (SETQ MAGNIFIED-BITMAP-WIDTH (* 32 (CEILING MAGNIFIED-WIDTH 32)))
   (USING-RESOURCE
    (MAGNIFIED-BITMAP-ARRAY PRINTER:SCREEN-IMAGE-BIT-ARRAY
			    MAGNIFIED-BITMAP-WIDTH MAGNIFIED-BITMAP-HEIGHT)
    ;; first, zero out the entire magnified bitmap array...
    (BITBLT tv:alu-setz
	    (ARRAY-DIMENSION MAGNIFIED-BITMAP-ARRAY 1)
	    (ARRAY-DIMENSION MAGNIFIED-BITMAP-ARRAY 0)
	    MAGNIFIED-BITMAP-ARRAY 0 0 MAGNIFIED-BITMAP-ARRAY 0 0)
    ;; Then build a magnified version of the given array...
    (LOOP FOR I FROM 0 TO (1- HEIGHT)
	  AND III FROM 0 TO (1- MAGNIFIED-HEIGHT) BY MAGNIFICATION DO
	  (LOOP FOR J FROM 0 TO (1- WIDTH)
		AND JJJ FROM 0 TO (1- MAGNIFIED-WIDTH) BY MAGNIFICATION DO
		(SETQ SOURCE-BIT (AREF BITMAP-ARRAY
				       (+ START-Y I)
				       (+ START-X J)))
		(DOTIMES (II MAGNIFICATION)
		  (DOTIMES (JJ MAGNIFICATION)
		    (SETF (AREF MAGNIFIED-BITMAP-ARRAY (+ III II) (+ JJJ JJ)) SOURCE-BIT)))))

    ;; finally, actually output the bitmap to the printer...
    (IF LANDSCAPEP
	(SEND SELF :OUTPUT-LANDSCAPE-BITMAP-TO-PRINTER
	      MAGNIFIED-BITMAP-ARRAY MAGNIFIED-BITMAP-WIDTH MAGNIFIED-BITMAP-HEIGHT START-X START-Y)
	(SEND SELF :OUTPUT-PORTRAIT-BITMAP-TO-PRINTER
	      MAGNIFIED-BITMAP-ARRAY MAGNIFIED-BITMAP-WIDTH MAGNIFIED-BITMAP-HEIGHT START-X START-Y))))
    


(DEFMETHOD (TI2015-GRAPHICS-MIXIN :PRINT-BITMAP) (BITMAP-ARRAY
						  &OPTIONAL WIDTH HEIGHT (START-X 0) (START-Y 0)
						  ORIENTATION DOTS-PER-INCH &AUX LANDSCAPEP
                                                  magnification)
  "Copy the bitmap array to a print stream; printer-stream could be a file stream or serial stream.
   Use HP-LaserJet s/w interface, which requires 8 bit data transfer."
  (SEND SELF :SELECT-HP-LASERJET-INTERFACE); select HP-LaserJet interface
  (IF (> NCOPIES 1) (SEND SELF :PRINT-MULTIPLE-PAGE-COPIES NCOPIES)) 
  (IF (NULL WIDTH) (SETQ WIDTH (ARRAY-DIMENSION BITMAP-ARRAY 1)))
  (IF (NULL HEIGHT) (SETQ HEIGHT (ARRAY-DIMENSION BITMAP-ARRAY 0)))
  (UNLESS (MEMBER DOTS-PER-INCH '(75 100 150 300) :TEST #'EQ)   ; determine best dpi unless
    (SETQ DOTS-PER-INCH (SEND SELF :GET-BEST-DPI WIDTH HEIGHT)));  a specific value is requested
  ;verify that requested dpi does not exceed printable dimensions
  (IF (AND (EQL DOTS-PER-INCH 75) 
	      (OR (> WIDTH MAX-WIDTH-75-DPI) (> HEIGHT MAX-HEIGHT-75-DPI)))
	 (SETQ DOTS-PER-INCH 100)) ; even a full screen with who-line at 100 dpi will fit in both orientations
  (SEND SELF :SET-RESOLUTION DOTS-PER-INCH)
  (SETQ LANDSCAPEP (SEND SELF :COMPUTE-ORIENTATION ORIENTATION DOTS-PER-INCH WIDTH))
  (COND (LANDSCAPEP
         ; determine if increased magnification beyond 75 dots-per-inch is desirable,
         ; where the dpi values and their corresponding magnification factors are the following:
         ; 300 dpi = 1 mag, 150 dpi = 2 mag (actually 4x the area at 300 dpi),
         ; 100 dpi = 3 mag (9x area at 300 dpi) , 75 dpi = 4 mag (16x area at 300 dpi)
         (SETQ MAGNIFICATION (SEND SELF :DETERMINE-MAGNIFICATION DOTS-PER-INCH HEIGHT WIDTH))
	 ; set margins
         (SEND SELF :SET-MARGINS MAGNIFICATION HEIGHT WIDTH)
         ;we only magnify if the computed magnification > 4, since 75 dpi = 4 mag,
         ;and the printer itself can handle down to 75 dpi.
         (COND ((> MAGNIFICATION 4) 
		(SEND SELF :SET-RESOLUTION 300); reset resolution to 300 as base for magnification
		(SEND SELF :OUTPUT-MAGNIFIED-BITMAP-TO-PRINTER
		      BITMAP-ARRAY HEIGHT WIDTH START-X START-Y MAGNIFICATION LANDSCAPEP))
	       (T ; else print unmagnified landscape
		(SEND SELF :OUTPUT-LANDSCAPE-BITMAP-TO-PRINTER
		      BITMAP-ARRAY WIDTH HEIGHT START-X START-Y))))
	(T ; else portrait print
         ; determine if increased magnification beyond 75 dots-per-inch is desirable,
         ; where the dpi values and their corresponding magnification factors are the following:
         ; 300 dpi = 1 mag, 150 dpi = 2 mag, 100 dpi = 3 mag, 75 dpi = 4 mag
         (SETQ MAGNIFICATION (SEND SELF :DETERMINE-MAGNIFICATION DOTS-PER-INCH WIDTH HEIGHT))
	 ; set margins
         (SEND SELF :SET-MARGINS MAGNIFICATION WIDTH HEIGHT)
         ;we only magnify if the computed magnification > 4, since 75 dpi = 4 mag,
         ;and the printer itself can handle down to 75 dpi.
         (COND((> MAGNIFICATION 4)
	       (SEND SELF :SET-RESOLUTION 300); reset resolution to 300 as base for magnification
	       (SEND SELF :OUTPUT-MAGNIFIED-BITMAP-TO-PRINTER
		     BITMAP-ARRAY HEIGHT WIDTH START-X START-Y MAGNIFICATION LANDSCAPEP))
              (T ; else print unmagnified portrait
	       (SEND SELF :OUTPUT-PORTRAIT-BITMAP-TO-PRINTER
		     BITMAP-ARRAY WIDTH HEIGHT START-X START-Y))))))



(DEFMETHOD (TI2015-GRAPHICS-MIXIN :OUTPUT-LANDSCAPE-BITMAP-TO-PRINTER) (BITMAP-ARRAY WIDTH HEIGHT START-X START-Y)
  "Output a landscape-oriented bitmap."
  (SEND SELF :START-RASTER-GRAPHICS 1); start at current logical position
  (LOOP FOR J FROM (1- WIDTH) DOWNTO START-X DO                 ;get the bitmap
	(SETF (FILL-POINTER LINE-BUFFER) 0)                     ;zero out the line-buffer
	(LOOP FOR JJ FROM START-Y TO (1- HEIGHT) BY 8 DO        ;get a line of bytes
	      (VECTOR-PUSH 
		(LOOP FOR JJJ FROM 0 TO 7 SUMMING               ;get 8 bits (pixel values) and put 
		      (LSH                                      ; into element of line-buffer
			(IF (AND (>= J START-X) (< (+ JJ JJJ) HEIGHT))
			    (AREF BITMAP-ARRAY (+ JJ JJJ) J)
			    0)
			(- 7 JJJ)))
		LINE-BUFFER))
	(SEND SELF :ESC-COMMAND #\* #\b)                   ;output one line 
	(SEND SELF :STRING-OUT-RAW (FORMAT () "~D" (LENGTH LINE-BUFFER)))
	(SEND SELF :TYO-RAW #\W)
	(SEND SELF :STRING-OUT-RAW LINE-BUFFER))
  (SEND SELF :END-RASTER-GRAPHICS))



(DEFMETHOD (TI2015-GRAPHICS-MIXIN :OUTPUT-PORTRAIT-BITMAP-TO-PRINTER) (BITMAP-ARRAY WIDTH HEIGHT START-X START-Y)
  "Output a portrait-oriented bitmap."
  (SEND SELF :START-RASTER-GRAPHICS 1); start at current logical position
  (LOOP FOR J FROM START-Y to (1- HEIGHT) DO                    ;get the bitmap
	(SETF (FILL-POINTER LINE-BUFFER) 0)                     ;zero out the line-buffer
	(LOOP FOR JJ FROM START-X to (1- WIDTH) BY 8 DO         ;get a line of bytes
	      (VECTOR-PUSH 
		(LOOP FOR JJJ FROM 0 TO 7 SUMMING               ;get 8 bits (pixel values) and put 
		      (LSH                                      ; into element of line-buffer
			(IF (AND (< J  HEIGHT) (< (+ JJ JJJ) WIDTH))
			    (AREF BITMAP-ARRAY J (+ JJ JJJ))
			    0)
			(- 7 JJJ)))
		LINE-BUFFER))
	(SEND SELF :ESC-COMMAND #\* #\b)                   ;output one line 
	(SEND SELF :STRING-OUT-RAW (FORMAT () "~D" (LENGTH LINE-BUFFER)))
	(SEND SELF :TYO-RAW #\W)
	(SEND SELF :STRING-OUT-RAW LINE-BUFFER))
  (SEND SELF :END-RASTER-GRAPHICS))



(DEFMETHOD (TI2015-GRAPHICS-MIXIN :COMPUTE-ORIENTATION) (ORIENTATION DOTS-PER-INCH WIDTH)
  "Returns whether or not landscape orientation is desired"
  (CASE ORIENTATION
    (:LANDSCAPE T)
    (:PORTRAIT (IF (AND (<= DOTS-PER-INCH 100)         ; if 75 or 100 dpi,
		        (> WIDTH (* 8 DOTS-PER-INCH))) ; and width is greater than max,
		   t                                   ; then orientation must be landscape
		   nil)) ;else portrait  
    (:BEST
     ;; Note -- will return NIL if dpi is NIL
     (AND (MEMBER DOTS-PER-INCH '(75 100 150 300) :TEST #'EQ)
	  (> WIDTH (* 8 DOTS-PER-INCH))))))



(DEFMETHOD (TI2015-GRAPHICS-MIXIN :GET-BEST-DPI) (WIDTH HEIGHT)
  "Returns best dpi given the width and height"
  (COND
    ((AND (<= WIDTH MAX-WIDTH-75-DPI)  (<= HEIGHT MAX-HEIGHT-75-DPI)) 75)
    ((AND (<= WIDTH MAX-WIDTH-100-DPI) (<= HEIGHT MAX-HEIGHT-100-DPI)) 100)
    ((AND (<= WIDTH MAX-WIDTH-150-DPI) (<= HEIGHT MAX-HEIGHT-150-DPI)) 150)
    ((AND (<= WIDTH MAX-WIDTH-300-DPI) (<= HEIGHT MAX-HEIGHT-300-DPI)) 300)
    (T 150))); default is 150 dpi.


(DEFMETHOD (TI2015-GRAPHICS-MIXIN :SET-MARGINS) (MAGNIFICATION WIDTH HEIGHT)
  "Compute and set margins."
  (SEND SELF :SET-ABSOLUTE-HTAB-INCREMENT 
	(FLOOR (- (+ *INCREMENTS-IN-A-LINE* *HORIZONTAL-INCREMENTS*)
		  (*  *INCREMENTS-PER-PIXEL*
		      MAGNIFICATION WIDTH)) 
	       2))
  (SEND SELF :SET-ABSOLUTE-VTAB-INCREMENT 
	(FLOOR (- (- *INCREMENTS-DOWN-A-PAGE* *VERTICAL-INCREMENTS*)
		  (*  *INCREMENTS-PER-PIXEL*
		      MAGNIFICATION HEIGHT))
	       2)))


 (DEFMETHOD (TI2015-GRAPHICS-MIXIN :DETERMINE-MAGNIFICATION)
	   (DOTS-PER-INCH WIDTH HEIGHT &AUX (MAXIMUM-MAGNIFICATION 6))
  "Determines if increased magnification beyond 75 dots-per-inch is desirable,
   where the dpi values and their corresponding magnification factors are the following:
   300 dpi = 1 mag, 150 dpi = 2 mag, 100 dpi = 3 mag, 75 dpi = 4 mag."
  (SELECTOR DOTS-PER-INCH EQL
    (300 1) (150 2) (100 3)
    (T (MIN MAXIMUM-MAGNIFICATION (FLOOR *PRINTABLE-PIXELS-DOWN-A-PAGE* HEIGHT)
	    (FLOOR *PRINTABLE-PIXELS-ACROSS-A-PAGE* WIDTH)))))


(DEFFLAVOR TI2015
	   ()
	   (TI2015-GRAPHICS-MIXIN
	    BASIC-TI2015))


(DEFMETHOD (TI2015 :SCREEN-IMAGE-FILE-P) (FILENAME)
  (LET ((FILE-TYPE (SEND FILENAME :TYPE)))
    (EQUALP "TI2015" FILE-TYPE)))


(COMPILE-FLAVOR-METHODS TI2015)


(DEFMETHOD (BASIC-TI2015 :CURRENT-PRINTER-STATE) (STREAM-TYPE)
  (IF (OR (NULL STREAM-TYPE); determine whether null or serial, or else stream is parallel
	  (EQ :SERIAL STREAM-TYPE))
    (LIST :ONLINE (LDB-TEST SI::%%SERIAL-B-STATUS-REGISTER-DSR (SI::SERIAL-B-STATUS)))
    ;ti2015 p1 is dtr go to
    (LIST :FAULT (SEND PRINTER-STREAM :SEND-IF-HANDLES :GET :FAULT);dcdb on z8530 chip Channel B 
	  :ONLINE (SEND PRINTER-STREAM :SEND-IF-HANDLES :GET :ONLINE)))) 

