;;; -*- Mode: common-LISP; Package: PRINTER; Base: 10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; This file defines print devices:  TI855 and TI850


;;; CHANGE HISTORY:
;;; 
;;;  Date      Author   Description
;;; -------------------------------------------------------------------------------------
;;;  12-13-88   dab     Changed :set-line-spacing to handle LPI of 3 and 4.
;;;  ****************** Rel 5.0 12-12-88
;;;  2/12/87    LCO     Deleted :tab method, allowing basic-printer's instead (which works
;;;                     correctly).


(DEFPROP TI855 (60 72 120 144) DOTS-PER-INCH) 

(DEFPROP TI850  (60 72 120 144) DOTS-PER-INCH) 


(DEFFLAVOR BASIC-TI855
	   ((SO #o16); selects expanded printing
	    (DC4 #o24); cancels expanded printing
	    (SI #o17); selects compresed printing
	    (DC2 #o22); cancels compressed printing
	    (DC1 #o21); enables the printer
	    (DC3 #o23); disables the printer
	    (ETX #o3); ask printer if ready
	    (ACK #o4); response from printer if ready
)
	   (BASIC-PRINTER)
  :SETTABLE-INSTANCE-VARIABLES) 


(DEFMETHOD (BASIC-TI855 :ESC-COMMAND) (COMMAND-CHAR &OPTIONAL N1 N2)
  "Send <ESC> <command-char> <n1> <n2>, where <n1> and <n2> are optional"
  (SEND SELF :TYO-RAW ESC)
  (SEND SELF :TYO-RAW COMMAND-CHAR)
  (IF N1
    (SEND SELF :TYO-RAW N1))
  (IF N2
    (SEND SELF :TYO-RAW N2))) 


(DEFMETHOD (BASIC-TI855 :EXTENDED-COMMAND) (COMMAND-CHAR &REST CHAR-LIST)
  "Send <ESC> <command-char> <char-list> ... <NUL>"
  (SEND SELF :TYO-RAW ESC)
  (SEND SELF :TYO-RAW COMMAND-CHAR)
  (LOOP FOR CHAR IN CHAR-LIST DO (SEND SELF :TYO-RAW CHAR) FINALLY (SEND SELF :TYO-RAW NUL))) 


(DEFMETHOD (BASIC-TI855 :SET-VTABS) (&REST POSITIONS)
  (LEXPR-SEND SELF :EXTENTED-COMMAND #\B POSITIONS)) 


(DEFMETHOD (BASIC-TI855 :SET-HTABS) (&REST POSITIONS)
  (LEXPR-SEND SELF :EXTENDED-COMMAND #\D POSITIONS)) 


(DEFMETHOD (BASIC-TI855 :SET-FORM-LENGTH) (NO-OF-LINES)
  (SEND SELF :ESC-COMMAND #\C NO-OF-LINES)) 


(DEFMETHOD (basic-ti855 :set-line-spacing) (lines-per-inch)
  (SELECTOR
    lines-per-inch =
    (8 (SEND self :esc-command #\0))
    (3 (SEND self :set-n-over-72-inches-per-line (/ 72 3)))   ;11-14-88 DAB Handle lpi of 3 and 4.
    (4 (SEND self :set-n-over-72-inches-per-line (/ 72 4)))
    (otherwise (SEND self :esc-command #\2)))) 


(DEFMETHOD (BASIC-TI855 :SET-N-OVER-72-INCHES-PER-LINE) (N)
  "Set to n72 inches per line"
  (COND
    ((= N 7) (SEND SELF :ESC-COMMAND #\1)); 7/72 in per line
    ((<= 1 N 127) (SEND SELF :ESC-COMMAND #\A N)); n/72 in per line
    (T (SEND SELF :ESC-COMMAND #\1)))) 


(DEFMETHOD (BASIC-TI855 :SET-N-OVER-216-INCHES-PER-LINE) (N)
  "Set to n216 inches per line"
  (COND
    ((<= 1 N 255) (SEND SELF :ESC-COMMAND #\3 N)); n/216 in per line
    (T (SEND SELF :ESC-COMMAND #\1)))) 


(DEFMETHOD (BASIC-TI855 :SET-N-OVER-144-INCHES-PER-LINE) (N)
  "Set to n144 inches per line;
For TI855, need to convert n144 to the nearest n216,
i.e., solve equation x216 = n144, x = n * 216144 = n * 32 = n * 1.5"
  (SEND SELF :SET-N-OVER-216-INCHES-PER-LINE (VALUES (FLOOR (+ (* 1.5 N) 0.5))))) 


(DEFMETHOD (BASIC-TI855 :SELECT-COMPRESSED-PRINT) ()
  (SEND SELF :ESC-COMMAND #\P)) 


(DEFMETHOD (BASIC-TI855 :CANCEL-COMPRESSED-PRINT) ()
  (SEND SELF :ESC-COMMAND #\Q)) 


(DEFMETHOD (BASIC-TI855 :SELECT-EXPANDED-PRINT) (); double-wide print mode
  (SEND SELF :ESC-COMMAND #\S)) 


(DEFMETHOD (BASIC-TI855 :CANCEL-EXPANDED-PRINT) ()
  (SEND SELF :ESC-COMMAND #\T)) 


(DEFMETHOD (BASIC-TI855 :SELECT-SHADOW-PRINT) (); emphasized print mode
  (SEND SELF :ESC-COMMAND #\E))                       ; (double-strike at a horizontal offset)


(DEFMETHOD (BASIC-TI855 :CANCEL-SHADOW-PRINT) ()
  (SEND SELF :ESC-COMMAND #\F)) 


(DEFMETHOD (BASIC-TI855 :SELECT-BOLD-PRINT) (); double-strike print mode
  (SEND SELF :ESC-COMMAND #\G)) 


(DEFMETHOD (BASIC-TI855 :CANCEL-BOLD-PRINT) ()
  (SEND SELF :ESC-COMMAND #\H)) 


(DEFMETHOD (BASIC-TI855 :SELECT-QUALITY-PRINT) (); enhanced print mode
  (SEND SELF :ESC-COMMAND #\e)) 


(DEFMETHOD (BASIC-TI855 :SELECT-DRAFT-PRINT) (); standard print mode
  (SEND SELF :ESC-COMMAND #\d)) 


(DEFMETHOD (BASIC-TI855 :SUPERSCRIPT) (STRING)
  (SEND SELF :ESC-COMMAND #\^)
  (SEND SELF :STRING-OUT-RAW STRING)
  (SEND SELF :ESC-COMMAND #\v)) 


(DEFMETHOD (BASIC-TI855 :SUBSCRIPT) (STRING)
  (SEND SELF :ESC-COMMAND #\v)
  (SEND SELF :STRING-OUT-RAW STRING)
  (SEND SELF :ESC-COMMAND #\^)) 


(DEFMETHOD (BASIC-TI855 :START-NEW-LINE) ()
  (SETQ CHAR-COUNT 0)
  (SEND SELF :TYO-RAW LF)
  (DOTIMES (I LFPAD)
    (SEND SELF :TYO-RAW NUL)))

(defmethod (basic-ti855 :set-tabs) () ; DAB 01-24-89
  (unless (integerp tab-width) (setf tab-width *default-tab-width*))
  (do* ((tab-position tab-width (+ tab-position tab-width))  ; DAB 01-24-89 Set tab spaces based on tab-width.
	(tabs '(1)))                                         ;Default is 8.
       ((> tab-position 120.)
	(apply SELF :SET-HTABS (reverse tabs)))
    (push tab-position tabs)))

(DEFMETHOD (BASIC-TI855 :SETUP-NORMAL-MODE) ()
  "Sets up the printer's cpi, lpi, and tabs."
  (SEND SELF :CANCEL-BOLD-PRINT)
  (SEND SELF :CANCEL-EXPANDED-PRINT)
  (SEND SELF :CANCEL-SHADOW-PRINT)
  (SEND SELF :CANCEL-COMPRESSED-PRINT)
  (SEND SELF :SET-CHAR-PITCH CHAR-PITCH)
  (SEND SELF :SET-LINE-SPACING LINE-SPACING)
  ;(send self :set-tabs tab-width)   ; DAB 01-24-89  
  )


(DEFMETHOD (BASIC-TI855 :PRINT-HEADER-PAGE) (FILENAME USERNAME)
  "Prints filename, username, and current date on one page."
  (SEND SELF :SELECT-EXPANDED-PRINT)
  (SEND SELF :STRING-OUT-CHARS
     (FORMAT () "~A~%~A~%~%    ~A~@[~%    created: ~A~]" USERNAME CURRENT-TIME FILENAME
	     (IF (TYPEP FILENAME 'PATHNAME)
	       (TIME:PRINT-UNIVERSAL-TIME (GET (SEND FILENAME :PROPERTIES) :CREATION-DATE) ()))))
  (SEND SELF :CANCEL-EXPANDED-PRINT);2.1 fix removed start-new-page. This will be done in :before :print-text-file. This was
  ;causing two page 1s.
) 


(DEFMETHOD (BASIC-TI855 :PRINT-TEXT-FILE) (FILE-STREAM &OPTIONAL IGNORE)
  "Print an ascii text file."
  (LOOP FOR IN-CHAR = (TYI FILE-STREAM ()) UNTIL (NULL IN-CHAR) DO (SEND SELF :TYO-CHAR IN-CHAR))) 


(DEFMETHOD (BASIC-TI855 :PRINT-RAW-FILE) (FILE-STREAM)
  "Files already in raw form, just send out characters to printer"
  (LOOP FOR IN-CHAR = (TYI FILE-STREAM ()) UNTIL (NULL IN-CHAR) DO (SEND SELF :TYO-RAW IN-CHAR))) 




(DEFFLAVOR TI855-FONT-MIXIN
	   ((FONT-MODULE 1); default <socket number>
	    (CURRENT-FONT-SOCKET-AND-MODE)); format: (<socket-number> <mode>) 
	   ()
  (:REQUIRED-FLAVORS BASIC-TI855)
  :SETTABLE-INSTANCE-VARIABLES) 


(DEFMETHOD (TI855-FONT-MIXIN :SELECT-FONT-MODULE) (SOCKET-NUMBER)
  "Socket number 0 selects the internal limited character set;
1, 2, or 3 selects the font module in socket 1, 2, or 3, respectively;
4 selects the downloaded character set as the active font (not available yet)"
  (IF (<= 0 SOCKET-NUMBER 3)
    (SEND SELF :ESC-COMMAND #\f SOCKET-NUMBER)
    (SEND SELF :ESC-COMMAND #\f 1))) 


(DEFMETHOD (TI855-FONT-MIXIN :SET-CHAR-PITCH) (CHARS-PER-INCH)
  (COND
    ((< CHARS-PER-INCH 12) (SEND SELF :ESC-COMMAND #\y) (SETQ CHAR-PITCH 10))
    ((< CHARS-PER-INCH 16) (SEND SELF :ESC-COMMAND #\z) (SETQ CHAR-PITCH 12))
    (T (SEND SELF :SELECT-COMPRESSED-PRINT) (SETQ CHAR-PITCH 16)))
  (SETQ CHARS-PER-LINE (* 8 CHAR-PITCH))) 


(DEFMETHOD (TI855-FONT-MIXIN :AFTER :SETUP-NORMAL-MODE) ()
  "Selects a font module."
  (SEND SELF :SELECT-FONT-MODULE FONT-MODULE)) 


(DEFMETHOD (TI855-FONT-MIXIN :PRINT-PAGE-HEADING) (&AUX FIRST-LINE SECOND-LINE (TEXT-CURRENT-FONT CURRENT-FONT-SOCKET-AND-MODE)
  (PAGE-HEADING-FONT (LIST FONT-MODULE 'NORMAL)))
  "Print page heading.  Switch to default font module if using different fonts."
  (SETQ PAGE-COUNT (1+ PAGE-COUNT))
  (IF (CONSP PAGE-HEADING)
    (SETQ FIRST-LINE (FIRST PAGE-HEADING)
	  SECOND-LINE (SECOND PAGE-HEADING))
    (SETQ SECOND-LINE PAGE-HEADING))
  (IF (NOT TEXT-CURRENT-FONT)
    (SEND SELF :STRING-OUT-CHARS
       (PAGE-HEADING-STRING FIRST-LINE SECOND-LINE CURRENT-TIME PAGE-COUNT))
    ;; switch between text font and page heading font
    (PROGN
      (SEND SELF :SEND-FONT-SEQUENCE PAGE-HEADING-FONT TEXT-CURRENT-FONT)
      (SEND SELF :STRING-OUT-CHARS
	 (PAGE-HEADING-STRING FIRST-LINE SECOND-LINE CURRENT-TIME PAGE-COUNT))
      (SEND SELF :SEND-FONT-SEQUENCE TEXT-CURRENT-FONT PAGE-HEADING-FONT)))) 


(DEFMETHOD (TI855-FONT-MIXIN :PRINT-TEXT-FILE) (FILE-STREAM &OPTIONAL FONT-LIST)
  "Print a text file with or without fonts
Font-To-Socket-Alist has entries (<font-name> . <socket-number>),
Socket-And-Mode-List has entries (<socket-number> <mode>),
where <mode> is either NORMAL, BOLD, ITALIC or BOLD-ITALIC"
  (LET (FONT-TO-SOCKET-ALIST
	SOCKET-AND-MODE-LIST
	CURRENT-FONT
	PREVIOUS-FONT)
    (UNLESS FONT-LIST
      (SETQ FONT-LIST (GET-FONT-ATTRIBUTE-LIST FILE-STREAM)))
    ;; get socket-and-mode-list from font-list
    (MULTIPLE-VALUE-SETQ (FONT-TO-SOCKET-ALIST SOCKET-AND-MODE-LIST)
      (SEND SELF :BUILD-FONT-TABLES FONT-LIST))
    ;; setup to use the first font in list, if it exists
    (WHEN SOCKET-AND-MODE-LIST
      (SETQ CURRENT-FONT 0)
      (SEND SELF :SEND-FONT-SEQUENCE (FIRST SOCKET-AND-MODE-LIST) ()))
    (SETQ CURRENT-FONT (IF SOCKET-AND-MODE-LIST
			 0
			 ()))
    (unless tab-width ; DAB 01-24-89
      (setf tab-width (or (GET-tab-width-ATTRIBUTE-LIST FILE-STREAM)
			  *default-tab-width*)))
    (send self :set-tabs)  ; DAB 01-24-89
      
    (DO ((IN-CHAR)
	 (FONT-STACK (MAKE-ARRAY 50 :LEADER-LIST '(0)))
	 (SML-LENGTH (LENGTH SOCKET-AND-MODE-LIST)))
	((NULL (SETQ IN-CHAR (SEND FILE-STREAM :TYI))))
      (COND
	((AND CURRENT-FONT (= IN-CHAR #\)) (SETQ IN-CHAR (SEND FILE-STREAM :TYI))
	 (COND
	   ((<= #\0 IN-CHAR #\9)
	    ;; <epsilon> (ie, 6) is followed by a digit <n>, 
	    ;; switch to font <n>
	    (UNLESS (VECTOR-PUSH CURRENT-FONT FONT-STACK)
	      (COPY-ARRAY-PORTION FONT-STACK 20 (ARRAY-TOTAL-SIZE FONT-STACK) FONT-STACK 0
				  (- (ARRAY-TOTAL-SIZE FONT-STACK) 20))
	      (DECF (FILL-POINTER FONT-STACK) 20)
	      (VECTOR-PUSH CURRENT-FONT FONT-STACK))
	    (SETQ PREVIOUS-FONT CURRENT-FONT)
	    (LET ((POS (- IN-CHAR #\0)))
	      (IF (< POS SML-LENGTH)
		(SETQ CURRENT-FONT POS)))
	    (UNLESS (= CURRENT-FONT PREVIOUS-FONT)
	      (SEND SELF :SEND-FONT-SEQUENCE (NTH CURRENT-FONT SOCKET-AND-MODE-LIST)
		 (NTH PREVIOUS-FONT SOCKET-AND-MODE-LIST))))
	   ((= IN-CHAR #\*)
	    ;; <epsilon> is followed by a *,
	    ;; switch to previously used font
	    (SETQ PREVIOUS-FONT CURRENT-FONT)
	    (OR (ZEROP (FILL-POINTER FONT-STACK)) (SETQ CURRENT-FONT (VECTOR-POP FONT-STACK)))
	    (UNLESS (= CURRENT-FONT PREVIOUS-FONT)
	      (SEND SELF :SEND-FONT-SEQUENCE (NTH CURRENT-FONT SOCKET-AND-MODE-LIST)
		 (NTH PREVIOUS-FONT SOCKET-AND-MODE-LIST))))
	   (T
	    ;; just print <epsilon> and character following it
	    (SEND SELF :TYO-CHAR #\) (SEND SELF :TYO-CHAR IN-CHAR))))
	(T (SEND SELF :TYO-CHAR IN-CHAR))))
    ;; since we're done, setup to use the default font socket module and mode NORMAL
    (IF CURRENT-FONT
      (SEND SELF :SEND-FONT-SEQUENCE (LIST FONT-MODULE 'NORMAL)
	 (NTH CURRENT-FONT SOCKET-AND-MODE-LIST))))) 


(DEFMETHOD (TI855-FONT-MIXIN :SEND-FONT-SEQUENCE) (CURRENT-FONT PREVIOUS-FONT)
  "Both Current-Font and Previous-Font should be either a list with format
of (<socket-number> <mode>) or nil;
When Previous-Font is NON-NIL, cancel its mode (ie, bold, italic or both);
When Current-Font is NON-NIL, select its module socket (if not already) and set its mode"
  (LET (CURR-SOCKET
	PREV-SOCKET
	CURR-MODE
	PREV-MODE)
    (IF CURRENT-FONT
      (SETQ CURRENT-FONT-SOCKET-AND-MODE CURRENT-FONT
	    CURR-SOCKET (CAR CURRENT-FONT)
	    CURR-MODE (CADR CURRENT-FONT)))
    (IF PREVIOUS-FONT
      (SETQ PREV-SOCKET (CAR PREVIOUS-FONT)
	    PREV-MODE (CADR PREVIOUS-FONT)))
    (WHEN (AND CURR-SOCKET (NEQ CURR-SOCKET PREV-SOCKET))
      (SEND SELF :SELECT-FONT-MODULE CURR-SOCKET)
      (SEND SELF :SET-CHAR-PITCH CHAR-PITCH))
    (CASE PREV-MODE
      (BOLD (SEND SELF :CANCEL-BOLD-PRINT))
      (ITALIC (SEND SELF :CANCEL-SHADOW-PRINT))
      (BOLD-ITALIC (SEND SELF :CANCEL-BOLD-PRINT) (SEND SELF :CANCEL-SHADOW-PRINT)))
    (CASE CURR-MODE
      (BOLD (SEND SELF :SELECT-BOLD-PRINT))
      (ITALIC (SEND SELF :SELECT-SHADOW-PRINT));for italic use shadow print on TI855
      (BOLD-ITALIC (SEND SELF :SELECT-BOLD-PRINT) (SEND SELF :SELECT-SHADOW-PRINT))))) 

(DEFMETHOD (TI855-FONT-MIXIN :BUILD-FONT-TABLES) (FONT-LIST)
  "From FONT-LIST return FONT-TO-SOCKET-ALIST and SOCKET-AND-MODE-LIST
FONT-TO-SOCKET-ALIST has entries (<font-name> . <socket-number>),
SOCKET-AND-MODE-LIST has entries (<socket-number> <mode>),
where <mode> is either NORMAL, BOLD, ITALIC or BOLD-ITALIC"
  (IF (AND FONT-LIST (ATOM FONT-LIST))
    (SETQ FONT-LIST (LIST FONT-LIST)))
  (LOOP FOR FONT IN FONT-LIST WITH FONT-STRING and font-long-string AND MODE AND SOCKET AND FONT-TO-SOCKET-ALIST WITH
     NEXT-SOCKET = 0 DO
     (MULTIPLE-VALUE-SETQ (FONT-STRING MODE font-long-string)
       (SEND SELF :GET-FONTSTR-AND-MODE FONT))
     (SETQ SOCKET
	   (OR (CDR (ASSOC FONT-STRING FONT-TO-SOCKET-ALIST :TEST 'STRING-EQUAL))
	      (PROG1
		(SETQ NEXT-SOCKET
		      (SELECTOR FONT-STRING EQUAL ("MODULE1" 1) ("MODULE2" 2) ("MODULE3" 3)
			 (T (1+ NEXT-SOCKET))))
		(PUSH (CONS FONT-STRING NEXT-SOCKET) FONT-TO-SOCKET-ALIST))))
     COLLECT (LIST SOCKET MODE  font-long-string) INTO SOCKET-AND-MODE-LIST FINALLY
     (RETURN (VALUES FONT-TO-SOCKET-ALIST SOCKET-AND-MODE-LIST))))

(DEFMETHOD (TI855-FONT-MIXIN :GET-FONTSTR-AND-MODE) (FONT)
  "Given a FONT, return list (<font-name> <mode>)
For example, for font HL12BI, return (HL12 BOLD-ITALIC)"
  (LET (FONT-PNAME
	B-OR-I-POS
	POSB
	POSI
	POS
	MODE)
    (SETQ FONT-PNAME (IF (SYMBOLP FONT)
		       (SYMBOL-NAME FONT)
		       (FORMAT () "~A" FONT)))
    (SETQ B-OR-I-POS (IF (< (LENGTH FONT-PNAME) 2)
		       0
		       (- (LENGTH FONT-PNAME) 2)))
    (SETQ POSB
	  (POSITION #\b (THE STRING (STRING FONT-PNAME)) :START B-OR-I-POS :TEST #'CHAR-EQUAL))
    (SETQ POSI
	  (POSITION #\i (THE STRING (STRING FONT-PNAME)) :START B-OR-I-POS :TEST #'CHAR-EQUAL))
    (COND
      ((AND POSB POSI) (SETQ POS (MIN POSB POSI)
			     MODE 'BOLD-ITALIC))
      (POSB (SETQ POS POSB
		  MODE 'BOLD))
      (POSI (SETQ POS POSI
		  MODE 'ITALIC))
      (T (SETQ POS (LENGTH FONT-PNAME)
	       MODE 'NORMAL)))
    (VALUES (STRING-UPCASE (SUBSeq FONT-PNAME 0 POS)) MODE FONT-PNAME))) 


 




(DEFFLAVOR TI855-GRAPHICS-MIXIN
	   ((MAX-WIDTH-60-DPI (* 8 60))
	    (MAX-HEIGHT-60-DPI (* 10 60))
	    (MAX-WIDTH-72-DPI (* 8 72))
	    (MAX-HEIGHT-72-DPI (* 10 72))
	    (MAX-WIDTH-120-DPI (* 8 120))
	    (MAX-HEIGHT-120-DPI (* 10 120))
	    (MAX-WIDTH-144-DPI (* 8 144))
	    (MAX-HEIGHT-144-DPI (* 10 144))
	    (LINE-BUFFER (MAKE-ARRAY 1440 ':LEADER-LIST (LIST 0) ':TYPE 'ART-STRING)))
	   ()
  (:REQUIRED-FLAVORS BASIC-TI855)) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :SELECT-RASTER-GRAPHICS) (NO-OF-CHARS-TO-PRINT &OPTIONAL (DOTS-PER-INCH 144))
  "Before sending the graphics characters, must send the number of chars to print
and the desired resolution"
  (UNLESS (ZEROP NO-OF-CHARS-TO-PRINT)
    (MULTIPLE-VALUE-BIND (QUOT REM)
      (TRUNCATE NO-OF-CHARS-TO-PRINT 256)
      (SEND SELF :ESC-COMMAND
	 (SELECTOR DOTS-PER-INCH = (60 #\K) (72 #\N) (120 #\L) (144 #\O) (OTHERWISE #\O)) REM
	 QUOT)))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :PRINT-BITMAP) (BITMAP-ARRAY &OPTIONAL WIDTH HEIGHT (START-X 0) (START-Y 0) ORIENTATION DOTS-PER-INCH &AUX
  LANDSCAPEP)
  "Copy the bitmap array to a print stream;
printer-stream could be a file stream or serial stream. It assumes 8 bit data transfer"
  (UNLESS WIDTH
    (SETQ WIDTH (array-dimension BITMAP-ARRAY 1)))
  (UNLESS HEIGHT
    (SETQ HEIGHT (array-dimension BITMAP-ARRAY 0)))
  (SETQ LANDSCAPEP (SEND SELF :COMPUTE-ORIENTATION ORIENTATION DOTS-PER-INCH WIDTH HEIGHT))
  (IF LANDSCAPEP
    (PSETQ WIDTH HEIGHT HEIGHT WIDTH START-X START-Y START-Y START-X))
  (UNLESS (MEMBER DOTS-PER-INCH '(60 72 120 144) :TEST #'EQ)
    (SETQ DOTS-PER-INCH (SEND SELF :GET-BEST-DPI WIDTH HEIGHT)))
  (SEND SELF :START-NEW-LINE)
  (SELECTOR DOTS-PER-INCH =
     (60 (SEND SELF :PRINT-AT-60-DPI BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP))
     (72 (SEND SELF :PRINT-AT-72-DPI BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP))
     (120 (SEND SELF :PRINT-AT-120-DPI BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP))
     (144 (SEND SELF :PRINT-AT-144-DPI BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP)))
  T) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :COMPUTE-ORIENTATION) (ORIENTATION DOTS-PER-INCH WIDTH HEIGHT)
  "Returns whether or not landscape orientation is desired"
  HEIGHT
  (CASE ORIENTATION
    (:LANDSCAPE T)
    (:PORTRAIT NIL)
    (:BEST
     ;; Note -- will return NIL if dpi is NIL
     (AND (MEMBER DOTS-PER-INCH '(60 72 120 144) :TEST #'EQ) (> WIDTH (* 8 DOTS-PER-INCH)))))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :GET-BEST-DPI) (WIDTH HEIGHT)
  "Returns best dpi given the width and height"
  (COND
    ((AND (<= WIDTH MAX-WIDTH-60-DPI) (<= HEIGHT MAX-HEIGHT-60-DPI)) 60)
    ((AND (<= WIDTH MAX-WIDTH-72-DPI) (<= HEIGHT MAX-HEIGHT-72-DPI)) 72)
    ((AND (<= WIDTH MAX-WIDTH-120-DPI) (<= HEIGHT MAX-HEIGHT-120-DPI)) 120)
    (T 144))) 

;; Compute one line of characters at a time,
;; Each character represents 8 bits of data from the bitmap array - 
;; ie, one horizontal position and 8 vertical positions 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :PRINT-AT-72-DPI) (BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP)
  (SEND SELF :SET-N-OVER-72-INCHES-PER-LINE 10)
  ;; 8/72 in. per line
  (LOOP FOR Y FROM START-Y BELOW HEIGHT BY 8 DO
					      ;; print a row of chars,
					      ;; each char is 8 bits: <y>, <y>+1, ... <y>+7
     (SEND SELF :GET-LINE BITMAP-ARRAY WIDTH HEIGHT START-X Y LANDSCAPEP 1)
     (SEND SELF :SELECT-RASTER-GRAPHICS (ARRAY-ACTIVE-LENGTH LINE-BUFFER) 72)
     (SEND PRINTER-STREAM :STRING-OUT LINE-BUFFER) (SEND SELF :START-NEW-LINE))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :PRINT-AT-144-DPI) (BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP)
  (LOOP FOR Y FROM START-Y BELOW HEIGHT BY 14 DO
					      ;; Print a row of chars,
					      ;; each char is 8 bits: <y>, <y>+2, ... , <y>+14
     (SEND SELF :GET-LINE BITMAP-ARRAY WIDTH HEIGHT START-X Y LANDSCAPEP 2)
     (SEND SELF :SELECT-RASTER-GRAPHICS (ARRAY-ACTIVE-LENGTH LINE-BUFFER) 144)
     (SEND PRINTER-STREAM :STRING-OUT LINE-BUFFER)
     ;; move down 1/144 in.
     (SEND SELF :SET-N-OVER-144-INCHES-PER-LINE 1)
     (SEND SELF :START-NEW-LINE)
     ;; print another row of chars
     ;; each char is 8 bits: <y>+1, <y>+3, ... , <y>+15
     (SEND SELF :GET-LINE BITMAP-ARRAY WIDTH HEIGHT START-X (1+ Y) LANDSCAPEP 2)
     (SEND SELF :SELECT-RASTER-GRAPHICS (ARRAY-ACTIVE-LENGTH LINE-BUFFER) 144)
     (SEND PRINTER-STREAM :STRING-OUT LINE-BUFFER)
     ;; move down 15/144 in.
     (SEND SELF :SET-N-OVER-144-INCHES-PER-LINE 13) (SEND SELF :START-NEW-LINE))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :GET-LINE) (BITMAP-ARRAY WIDTH HEIGHT START-X FROM-Y LANDSCAPEP INCREMENT)
  "Get a line image of bitmap array.
from horizontal bit positions, START-X to WIDTH  
and for 8 vertical bit positions, starting at FROM-Y by INCREMENT"
  (LET ((NO-OF-BITS (* 8 INCREMENT)))
    (STORE-ARRAY-LEADER 0 LINE-BUFFER 0)
    (LOOP FOR X FROM START-X BELOW WIDTH AS CHAR = 0 DO
       (LOOP FOR Y FROM FROM-Y BELOW (+ FROM-Y NO-OF-BITS) BY INCREMENT FOR BIT-POS-OF-CHAR FROM
	  7 DOWNTO 0
	  DO
	  ;; sets each bit of CHAR to an array entry's value
	  ;; which is always either 0 or 1.
	  ;; recall (dpb <byte> <ppss> <num>) puts the
	  ;; <ss> bit(s) of <byte> in <num> starting at bit <pp>
	  (SETQ CHAR
		(DPB
		 (COND
		   ((>= Y HEIGHT) 0)
		   (LANDSCAPEP (ARRAY-2-REF BITMAP-ARRAY Y (- WIDTH X 1)))
		   (T (ARRAY-2-REF BITMAP-ARRAY X Y)))
		 (DPB BIT-POS-OF-CHAR (BYTE 6 6) 1) CHAR)))
       WHEN (ZEROP CHAR) DO
			 ;; If all zeros to end of line, no need to send them
       (LET ((NON-ZERO-XPOS
	      (SEND SELF :NON-ZERO-POSITION BITMAP-ARRAY WIDTH HEIGHT X FROM-Y LANDSCAPEP
		 INCREMENT)))
	 (IF (= NON-ZERO-XPOS WIDTH)
	   (RETURN)
	   (PROGN
	     (DOTIMES (COUNT (- NON-ZERO-XPOS X))
	       (VECTOR-PUSH 0 LINE-BUFFER))
	     (SETQ X (1- NON-ZERO-XPOS)))))
       ELSE DO (VECTOR-PUSH CHAR LINE-BUFFER)))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :NON-ZERO-POSITION) (BITMAP-ARRAY WIDTH HEIGHT FROM-X FROM-Y LANDSCAPEP INCREMENT)
  "Return the position of the first non-zero character"
  (LET ((NO-OF-BITS (* 8 INCREMENT)))
    (LOOP FOR X FROM FROM-X BELOW WIDTH AS CHAR = 0 DO
       (LOOP FOR Y FROM FROM-Y BELOW (+ FROM-Y NO-OF-BITS) BY INCREMENT FOR BIT-POS-OF-CHAR FROM
	  7 DOWNTO 0 DO
	  (SETQ CHAR
		(DPB
		 (COND
		   ((>= Y HEIGHT) 0)
		   (LANDSCAPEP (ARRAY-2-REF BITMAP-ARRAY Y (- WIDTH X 1)))
		   (T (ARRAY-2-REF BITMAP-ARRAY X Y)))
		 (DPB BIT-POS-OF-CHAR (BYTE 6 6) 1) CHAR)))
       UNLESS (ZEROP CHAR) RETURN X FINALLY (RETURN WIDTH)))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :PRINT-AT-120-DPI) (BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP)
  (LOOP FOR Y FROM START-Y BELOW HEIGHT BY 10
     DO
     ;; Print a row of chars,
     ;; each char is from bits at <y>+<n> where <n> = 0, 2, 4, 5, 7, 9
     (SEND SELF :GET-LINE-FROM-OFFSET-LIST BITMAP-ARRAY WIDTH HEIGHT START-X Y LANDSCAPEP
	'(0 2 4 5 7 9))
     (SEND SELF :SELECT-RASTER-GRAPHICS (ARRAY-ACTIVE-LENGTH LINE-BUFFER) 120)
     (SEND PRINTER-STREAM :STRING-OUT LINE-BUFFER)
     ;; move down 1/144 in.
     (SEND SELF :SET-N-OVER-144-INCHES-PER-LINE 1)
     (SEND SELF :START-NEW-LINE)
     ;; print another line of chars
     ;; each char is from bits at <y>+<n> where <n> = 1, 3, 6, & 8
     ;; <n> of -1 means skip over one bit of char
     (SEND SELF :GET-LINE-FROM-OFFSET-LIST BITMAP-ARRAY WIDTH HEIGHT START-X Y LANDSCAPEP
	'(1 3 -1 6 8 -1))
     (SEND SELF :SELECT-RASTER-GRAPHICS (ARRAY-ACTIVE-LENGTH LINE-BUFFER) 120)
     (SEND PRINTER-STREAM :STRING-OUT LINE-BUFFER)
     ;; move down 11/144 in. 
     (SEND SELF :SET-N-OVER-144-INCHES-PER-LINE 11) (SEND SELF :START-NEW-LINE))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :PRINT-AT-60-DPI) (BITMAP-ARRAY WIDTH HEIGHT START-X START-Y LANDSCAPEP)
  (LOOP FOR Y FROM START-Y BELOW HEIGHT BY 6
     DO
     ;; Print a row of chars,
     ;; each char is from bits at <y>+<n> where <n> = 0, 1, 4, 5 and
     ;; and lots of skips
     ;; <n> of -1 means skip over one bit of char
     (SEND SELF :GET-LINE-FROM-OFFSET-LIST BITMAP-ARRAY WIDTH HEIGHT START-X Y LANDSCAPEP
	'(0 1 -1 -1 -1 4 5 -1))
     (SEND SELF :SELECT-RASTER-GRAPHICS (ARRAY-ACTIVE-LENGTH LINE-BUFFER) 60)
     (SEND PRINTER-STREAM :STRING-OUT LINE-BUFFER)
     ;; move down 1/144 in.
     (SEND SELF :SET-N-OVER-144-INCHES-PER-LINE 1)
     (SEND SELF :START-NEW-LINE)
     ;; print another line of chars
     ;; each char is from bits at <y>+<n> where <n> = 2, 3, and
     ;; lots of skips
     ;; <n> of -1 means skip over one bit of char
     (SEND SELF :GET-LINE-FROM-OFFSET-LIST BITMAP-ARRAY WIDTH HEIGHT START-X Y LANDSCAPEP
	'(-1 -1 2 3 -1 -1 -1))
     (SEND SELF :SELECT-RASTER-GRAPHICS (ARRAY-ACTIVE-LENGTH LINE-BUFFER) 60)
     (SEND PRINTER-STREAM :STRING-OUT LINE-BUFFER)
     ;; move down 14/144 or 7/72 in. 
     (SEND SELF :SET-N-OVER-72-INCHES-PER-LINE 7) (SEND SELF :START-NEW-LINE))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :GET-LINE-FROM-OFFSET-LIST) (BITMAP-ARRAY WIDTH HEIGHT START-X FROM-Y LANDSCAPEP OFFSET-LIST)
  "Get a line image of bitmap array
from horizontal bit positions, START-X to WIDTH
and for up to 8 vertical bit positions, starting at FROM-Y with offsets in offset-list"
  (STORE-ARRAY-LEADER 0 LINE-BUFFER 0)
  (LOOP FOR X FROM START-X BELOW WIDTH AS CHAR = 0 DO
     (LOOP FOR Y-OFFSET IN OFFSET-LIST FOR BIT-POS-OF-CHAR FROM 7 DOWNTO 0 AS Y =
	(IF (MINUSP Y-OFFSET)
	  Y-OFFSET
	  (+ FROM-Y Y-OFFSET))
	DO
	;; sets each bit of CHAR to an array entry's value
	;; which is always either 0 or 1.
	;; recall (dpb <byte> <ppss> <num>) puts the
	;; <ss> bit(s) of <byte> in <num> starting at bit <pp>
	(SETQ CHAR
	      (DPB
	       (COND
		 ((MINUSP Y) 0)
		 ((>= Y HEIGHT) 0)
		 (LANDSCAPEP (ARRAY-2-REF BITMAP-ARRAY Y (- WIDTH X 1)))
		 (T (ARRAY-2-REF BITMAP-ARRAY X Y)))
	       (DPB BIT-POS-OF-CHAR (BYTE 6 6) 1) CHAR)))
     WHEN (ZEROP CHAR) DO
		       ;; If all zeros to end of line, no need to send them
     (LET ((NON-ZERO-XPOS
	    (SEND SELF :NON-ZERO-POSITION-FROM-OFFSET-LIST BITMAP-ARRAY WIDTH HEIGHT X FROM-Y
	       LANDSCAPEP OFFSET-LIST)))
       (IF (= NON-ZERO-XPOS WIDTH)
	 (RETURN)
	 (PROGN
	   (DOTIMES (COUNT (- NON-ZERO-XPOS X))
	     (VECTOR-PUSH 0 LINE-BUFFER))
	   (SETQ X (1- NON-ZERO-XPOS)))))
     ELSE DO (VECTOR-PUSH CHAR LINE-BUFFER))) 


(DEFMETHOD (TI855-GRAPHICS-MIXIN :NON-ZERO-POSITION-FROM-OFFSET-LIST) (BITMAP-ARRAY WIDTH HEIGHT FROM-X FROM-Y LANDSCAPEP OFFSET-LIST)
  "Return the position of the first non-zero character built from bitmap array
at positions in offset list"
  (LOOP FOR X FROM FROM-X BELOW WIDTH AS CHAR = 0 DO
     (LOOP FOR Y-OFFSET IN OFFSET-LIST FOR BIT-POS-OF-CHAR FROM 7 DOWNTO 0 AS Y =
	(IF (MINUSP Y-OFFSET)
	  Y-OFFSET
	  (+ FROM-Y Y-OFFSET))
	DO
	(SETQ CHAR
	      (DPB
	       (COND
		 ((MINUSP Y) 0)
		 ((>= Y HEIGHT) 0)
		 (LANDSCAPEP (ARRAY-2-REF BITMAP-ARRAY Y (- WIDTH X 1)))
		 (T (ARRAY-2-REF BITMAP-ARRAY X Y)))
	       (DPB BIT-POS-OF-CHAR (BYTE 6 6) 1) CHAR)))
     UNLESS (ZEROP CHAR) RETURN X FINALLY (RETURN WIDTH))) 




(DEFFLAVOR TI855
	   ()
	   (TI855-FONT-MIXIN
	    TI855-GRAPHICS-MIXIN
	    BASIC-TI855)) 


(DEFMETHOD (TI855 :SCREEN-IMAGE-FILE-P) (FILENAME)
  (LET ((FILE-TYPE (SEND FILENAME :TYPE)))
    (OR (EQUALP "TI855" FILE-TYPE) (EQUALP "855" FILE-TYPE)))) 




(DEFFLAVOR TI850
	   ()
	   (TI855-GRAPHICS-MIXIN
	    BASIC-TI855)) 


(DEFMETHOD (TI850 :SET-CHAR-PITCH) (CHARS-PER-INCH)
  (COND
    ((< CHARS-PER-INCH 16) (SEND SELF :CANCEL-COMPRESSED-PRINT) (SETQ CHAR-PITCH 10))
    (T (SEND SELF :SELECT-COMPRESSED-PRINT) (SETQ CHAR-PITCH 16)))
  (SETQ CHARS-PER-LINE (* 8 CHAR-PITCH))) 


(DEFMETHOD (TI850 :SCREEN-IMAGE-FILE-P) (FILENAME)
  (LET ((FILE-TYPE (SEND FILENAME :TYPE)))
    (OR (EQUALP "TI850" FILE-TYPE) (EQUALP "850" FILE-TYPE))))

;;;This method was added for 2.1. This method returns the current state of the printer. A check
;;;will be done before printing begins to make sure printer is online. The serial interface was
;;;not checking this and resulted in loss printout is offline.
 
(DEFMETHOD (BASIC-TI855 :CURRENT-PRINTER-STATE) (STREAM-TYPE)
  (IF (OR (NULL STREAM-TYPE); determine whether null or serial, or else stream is parallel
	  (EQ :SERIAL STREAM-TYPE))
    (LIST :ONLINE (LDB-TEST SI::%%SERIAL-B-STATUS-REGISTER-DSR (SI::SERIAL-B-STATUS)))
    ;ti855 p1 is dtr go to
    (LIST :FAULT (SEND PRINTER-STREAM :SEND-IF-HANDLES :GET :FAULT);dcdb on z8530 chip Channel B 
	  :ONLINE (SEND PRINTER-STREAM :SEND-IF-HANDLES :GET :ONLINE))))

 



(COMPILE-FLAVOR-METHODS TI855 TI850) 

