;;; -*- Mode:Common-Lisp; Package:user; Base:10; Fonts:(CPTFONT HL12B HL12I MEDFNB) -*-


;;;   The following hack prints an array of any dimension, with optional
;;;labeled indices.  Formatting could be improved, but the following does
;;;the job:
      
(defun ARRAY-CONTENTS (array &key limit (stream *standard-output*) label-p
one-column-p)
  "Display the contents of ARRAY on STREAM, optionally limited by LIMIT. 
LABEL-P non-nil 
will include the indices for each slot.  ONE-COLUMNP non-nil force each value
on a  
separate line."
	     
  (let* ((dim-array (array-dimensions array))
	 (ndims (length dim-array))
	 (indices (make-list ndims :initial-element 0))
	 (elements (if (= ndims 1)
		       (length array)
		     (array-total-size array)))
	 (times (if (null limit)
		    elements
		  (min limit elements)))
	 (line-feeds 1))
      
    (dotimes (x times)
      (when (plusp line-feeds)
	(format stream "~V%" line-feeds)
	(setq line-feeds (if one-column-p 1 0)))
      
      (when label-p
      
        ;; write out the indice label
	(format stream "[~d~{ ~d~}]: " (car indices) (cdr indices))
      
        ;; update the label 
	(do ((s (1- ndims) (1- s)))
	    ((minusp s))
	  (let ((inc (1+ (nth s indices))))
	    (cond ((>= inc (nth s dim-array))
		   (setf (nth s indices) 0)
		   (incf line-feeds))
		  (t (setf (nth s indices) inc)
		     (return))))))
      ;; write out the array element
      (format stream "~8s " (ar-1-force array x))))
      
  ;; don't return anything
  (values))
