;;  -*- Mode:Lisp; Package:USER; Base:10; Lowercase:YES; Syntax: Common-lisp -*-


;;;  Arithmetic Benchmarks
(defvar *arithmetic* ())
(defvar *fixnum-arithmetic* ())
(defvar *flonum-arithmetic* ())
(defvar *bignum-arithmetic* ())

(defvar val1 1231232)
(defvar val2 6353462)
(defvar val3 5424)
(defvar val4 645)
(defvar val5 35242345347567234658726347563248756872346589324659)
(defvar val6 6487246358746385762394569872346587923465897263478956238794658792346589)

(define-benchmark fixnum-addition "Fixnum addition" 3000
  :classes (*zippel-benchmarks* *fixnum-arithmetic* *arithmetic*)
  :allow-page-faults nil
  :normalization (+ (normalization push-local) (normalization store-local))
  :bindings ((v1 val1) (v2 val2) result)
  :body (setq result (+ v1 v2)))

(define-benchmark fixnum-subtraction "Fixnum subtraction" 3000
  :classes (*zippel-benchmarks* *fixnum-arithmetic* *arithmetic*)
  :allow-page-faults nil
  :normalization (+ (normalization push-local) (normalization store-local))
  :bindings ((v1 val1) (v2 val2) result)
  :body (setq result (- v1 v2)))

(define-benchmark fixnum-multiplication "Fixnum multiplication" 3000
  :classes (*zippel-benchmarks* *fixnum-arithmetic* *arithmetic*)
  :allow-page-faults nil
  :normalization #+3600(+ (* 2 (normalization push-local)) (normalization store-local))
                 #-3600(+ (normalization push-local) (normalization store-local))
  :bindings ((v3 val3) (v4 val4) result)
  :body (setq result (* v3 v4)))

(define-benchmark fixnum-division "Fixnum division" 3000
  :classes (*zippel-benchmarks* *fixnum-arithmetic* *arithmetic*)
  :allow-page-faults nil
  :normalization #+3600(+ (* 2 (normalization push-local)) (normalization store-local))
                 #-3600(+ (normalization push-local) (normalization store-local))
  :bindings ((v3 val3) (v4 val4) result)
  :body (setq result (floor v4 v3)))

(define-benchmark float-addition "Floating point addition" 3000
  :classes (*zippel-benchmarks* *flonum-arithmetic* *arithmetic*)
  :allow-page-faults nil			; I uncommented this line. - pf
  :normalization (+ (* 1 (normalization push-local)) (normalization store-local))
  :bindings ((v1 float-val1) (v2 float-val2) result)
  :body (setq result (+ v1 v2)))

(define-benchmark float-subtraction "Floating point subtraction" 3000
  :classes (*zippel-benchmarks* *flonum-arithmetic* *arithmetic*)
  :allow-page-faults nil			; I uncommented this line. - pf
  :normalization (+ (* 1 (normalization push-local)) (normalization store-local))
  :bindings ((v1 float-val1) (v2 float-val2) result)
  :body (setq result (- v1 v2)))

(define-benchmark float-multiplication "Floating point multiplication" 3000
  :classes (*zippel-benchmarks* *flonum-arithmetic* *arithmetic*)
  :allow-page-faults nil			; I uncommented this line. - pf
  :normalization #+3600(+ (* 2 (normalization push-local)) (normalization store-local))
                 #-3600(+ (* 1 (normalization push-local)) (normalization store-local))
  :bindings ((v1 float-val1) (v2 float-val2) result)
  :body (setq result (* v1 v2)))

(define-benchmark float-division "Floating point division" 3000
  :classes (*zippel-benchmarks* *flonum-arithmetic* *arithmetic*)
  :allow-page-faults nil			; I uncommented this line. - pf
  :normalization #+3600(+ (* 2 (normalization push-local)) (normalization store-local))
                 #-3600(+ (* 1 (normalization push-local)) (normalization store-local))
  :bindings ((v1 float-val1) (v2 float-val2) result)
  :body (setq result (/ v1 v2)))

(define-benchmark bignum-addition "Bignum addition" 3000
  :classes (*zippel-benchmarks* *bignum-arithmetic* *arithmetic*)
  :allow-page-faults nil			; I uncommented this line. - pf
  :normalization (+ (* 1 (normalization push-local)) (normalization store-local))
  :bindings ((v5 val5) (v6 val6) result)
  :body (setq result (+ v5 v6)))

(define-benchmark bignum-subtraction "Bignum subtraction" 3000
  :classes (*zippel-benchmarks* *bignum-arithmetic* *arithmetic*)
  :allow-page-faults nil			; I uncommented this line. - pf
  :normalization (+ (* 1 (normalization push-local)) (normalization store-local))
  :bindings ((v5 val5) (v6 val6) result)
  :body (setq result (- v5 v6)))

(define-benchmark bignum-multiplication "Bignum multiplication" 3000
  :classes (*zippel-benchmarks* *bignum-arithmetic* *arithmetic*)
  :allow-page-faults nil			; I uncommented this line. - pf
  :normalization #+3600(+ (* 2 (normalization push-local)) (normalization store-local))
                 #-3600(+ (* 1 (normalization push-local)) (normalization store-local))
  :bindings ((v5 val5) (v6 val6) result)
  :body (setq result (* v5 v6)))

(define-benchmark bignum-division "Bignum division" 3000
  :classes (*zippel-benchmarks* *bignum-arithmetic* *arithmetic*)
  :allow-page-faults nil			; I uncommented this line. - pf
  :normalization #+3600(+ (* 2 (normalization push-local)) (normalization store-local))
                 #-3600(+ (* 1 (normalization push-local)) (normalization store-local))
  :bindings ((v5 val5) (v6 val6) result)
  :body (setq result (floor v6 v5)))

