;;;  -*- Mode:Lisp; Package:USER; Base:10; Lowercase:YES; Syntax: Common-lisp -*-

;;; These are the normalization benchmarks


(defvar *instructions* ()
  "Class of benchmarks used to factor setups of other operations")

(defvar *a* ())
(defvar *b* ())
(defvar *c* ())
(defvar *d* ())
(defvar *e* ())


;; These normalizations include the time required to fetch the Push instructions

(defvar *push-immediate-normalization* 0.0
  "Number of microseconds required to push an immediate constant")

(defmacro normalization (type)
  `(benchmark-average-time
     (get ',(intern (format nil "~A-NORMALIZATION" type)) my-benchmark-property)))

(define-benchmark push-immediate-normalization "Push an immediate constant" 100
  :classes (*instructions*)
  :allow-page-faults nil
  :real-body
  (benchmark-body (1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)
    (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1) (si:%push 1)))

(define-benchmark push-local-normalization "Push a local variable" 100
  :classes (*instructions*)
  :allow-page-faults nil
  :bindings ((temp (microsecond-clock)))
  :real-body
  (benchmark-body (1)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)
    (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp) (si:%push temp)))

(define-benchmark store-local-normalization "Store into local variable" 100
  :classes (*instructions*)
  :allow-page-faults nil
  :normalization #-3600(normalization push-immediate) #+3600 0
  :real-body
  (let (*a0* *a1* *a2* *a3* *a4* *a5* *a6* *a7* *a8* *a9*
	*a10* *a11* *a12* *a13* *a14* *a15* *a16* *a17* *a18* *a19*
	*a20* *a21* *a22* *a23* *a24* *a25* *a26* *a27* *a28* *a29*
	*a30* *a31* *a32* *a33* *a34* *a35* *a36* *a37* *a38* *a39*
	*a40* *a41* *a42* *a43* *a44* *a45* *a46* *a47* *a48* *a49*
	)
  (benchmark-body (1)
	(setq *a0* 1) (setq  *a1* 1) (setq  *a2* 1) (setq *a3* 1) (setq *a4* 1)
	(setq *a5* 1) (setq *a6* 1) (setq *a7* 1) (setq *a8* 1) (setq *a9* 1)
	(setq *a10* 1) (setq *a11* 1) (setq *a12* 1) (setq *a13* 1) (setq *a14* 1)
	(setq *a15* 1) (setq *a16* 1) (setq *a17* 1) (setq *a18* 1) (setq *a19* 1)
	(setq *a20* 1) (setq *a21* 1) (setq *a22* 1) (setq *a23* 1) (setq *a24* 1)
	(setq *a25* 1) (setq *a26* 1) (setq *a27* 1) (setq *a28* 1) (setq *a29* 1)
	(setq *a30* 1) (setq *a31* 1) (setq *a32* 1) (setq *a33* 1) (setq *a34* 1)
	(setq *a35* 1) (setq *a36* 1) (setq *a37* 1) (setq *a38* 1) (setq *a39* 1)
	(setq *a40* 1) (setq *a41* 1) (setq *a42* 1) (setq *a43* 1) (setq *a44* 1)
	(setq *a45* 1) (setq *a46* 1) (setq *a47* 1) (setq *a48* 1) (setq *a49* 1)
	(setq *a0* 2) (setq  *a2* 2) (setq  *a2* 2) (setq *a3* 2) (setq *a4* 2)
	(setq *a5* 2) (setq *a6* 2) (setq *a7* 2) (setq *a8* 2) (setq *a9* 2)
	(setq *a20* 2) (setq *a22* 2) (setq *a22* 2) (setq *a23* 2) (setq *a24* 2)
	(setq *a25* 2) (setq *a26* 2) (setq *a27* 2) (setq *a28* 2) (setq *a29* 2)
	(setq *a20* 2) (setq *a22* 2) (setq *a22* 2) (setq *a23* 2) (setq *a24* 2)
	(setq *a25* 2) (setq *a26* 2) (setq *a27* 2) (setq *a28* 2) (setq *a29* 2)
	(setq *a30* 2) (setq *a32* 2) (setq *a32* 2) (setq *a33* 2) (setq *a34* 2)
	(setq *a35* 2) (setq *a36* 2) (setq *a37* 2) (setq *a38* 2) (setq *a39* 2)
	(setq *a40* 2) (setq *a42* 2) (setq *a42* 2) (setq *a43* 2) (setq *a44* 2)
	(setq *a45* 2) (setq *a46* 2) (setq *a47* 2) (setq *a48* 2) (setq *a49* 2)
	)))

(clear-benchmarks :recompute-normalizations t :report-p nil)
