;;; -*- Mode:Common-Lisp; Package:USER; Base:10 -*-

(DEFVAR *bench-repeat-count* 1. "Repeat count to use in averaging benchmark times")

(defmacro timer (name form . args)
  `(progn'compile
     (defun ,name ,args ,form)))

#+comment
(defmacro defarray (name type sizes)
  `(progn 'compile
	  (defmacro ,name (&rest l) `(aref ,',name ,@l))
	  (defconst ,name (MAKE-ARRAY ,sizes ':TYPE
				      #-3600 ,(IF (EQ TYPE 'FLONUM) 'ART-FLOAT 'ART-Q)
				      #+3600 'ART-Q
				      ':INITIAL-VALUE ,(CASE TYPE
							 (FIXNUM 0)
							 (FLONUM 0.0)
							 (OTHERWISE NIL))))))
;;
;; Create-the-packages
;;

; This is a list of packages to declare under the GLOBAL package.
; When an item of the list is another list, the car is the package name, and the cdr is
; a list of symbols to shadow
(DEFVAR *rpg-benchmarks*)
(setq *rpg-benchmarks*
      `( TAK 
	(TRAVERSE "SELECT" "REMOVE" "RANDOM")))

(DOLIST (x *rpg-benchmarks*)
  (let* ((pkg-name (string-append "CL-" (if (atom x) x (CAR x))))
	 (pkg (pkg-find-package pkg-name :FIND)))
    (unless pkg
      (MAKE-PACKAGE PKG-NAME
		    #+3600 :use #+3600 '(symbolics-common-lisp common-lisp-global)
		    :import '(#+(and explorer (not elroy)) nil #+3600 QUOTIENT timer defarray)
		    :shadow (AND (LISTP x) (CDR x))))))


; These are all the bench-mark programs.
; Comment out the ones you don't want to run.
; A number sets the repeat count
; An atom resets the temporary area that consing is done in
; This results in a very fast (and dirty) garbage collection

(DEFVAR *gabriel-cl-benchmarks* nil "The list of gabriel benchmarks")

(define-gabriel-benchmarks '*gabriel-cl-benchmarks* 1
      '(2
	(gabriel-traverse-init ignore)	   ;NO GC HERE! save for timit below
	(gabriel-traverse multiple-bench
		  (gabriel-traverse-init cl-traverse:init-timit)	;Cdr =
		  (gabriel-traverse cl-traverse:timit)) gc
	6
	(gabriel-tak cl-tak:timit)))
