;;; -*- Mode:Common-Lisp; Package:USER; Base:10 -*-
;===============================================================================
;
;   (c) Unpublished Copyright 1986 by Texas Instruments.  All rights reserved.
;
;===============================================================================

;;; Created 10/22/85 12:59:02 by LaMott G. OREN

;;;
;;; Code to report benchmarks
;;;

;;; Correction history:
;;;
;;; 11/11/86 LaMott Oren	Converted to Common-Lisp
;;; 11/03/86 LaMott Oren	Fixed bug in MEAN of MEAN, MEAN of MAX, MEAN of MIN reporting.


;;; The following are report formats.  Each report is a list of symbols found in the REPORT-TYPES list (below),
;;; or a list whose car is a symbol in REPORT-TYPES and whose cdr is parameters to be passed to the report-type.
;;; Each entry in the list defines a column of report output, with the exception of numbers.
;;; If a number appears in a report list, it is used to switch machines being reported on.  The report-benchmark
;;; function takes a list of machines to report on.  1 signifies the first machine in the machine-list, 2 the
;;; second machine, and so on.

(DEFVAR *all-reports* nil "A list of the known report lists (there may be others)")

(DEFVAR *real-max-min* nil "When nil, get max/min's associated with max/min real-time.
 Else, search for the REAL max/min.")

(eval-when (compile load eval)
(DEFMACRO defreport (name &body report-types)
  `(PROGN (UNLESS (MEMBER ',name *all-reports* :test #'eq) (PUSH ',name *all-reports*))
	  (DEFPARAMETER ,name ',report-types)))
)

(defreport std-report long-pretty-name count time-mean time-median time-dev fault-mean disk-mean cons-mean)
(defreport plot-report long-pretty-name count time-mean time-median time-dev fault-mean disk-mean cons-mean cpu-plot)
(defreport gabriel-report med-pretty-name count cpu-mean cpu-dev cpu-min cpu-max disk-max disk-min fault-max
					   max-response (fix :extra-pdl mean))
(defreport gabriel-report-tgc med-pretty-name count cpu-mean cpu-dev cpu-min cpu-max disk-max disk-min fault-max
					   (fix :tgc-read mean) (fix :tgc-write mean))
(defreport gabriel-compare
  med-pretty-name 1 count blank-2
  1 cpu-min 2 cpu-min (compare-1-2 cpu-min)
  1 time-mean 2 time-mean (compare-1-2 time-mean)
  1 disk-mean 2 disk-mean 1 cons-mean 2 cons-mean)
(defreport gabriel-compare-x
  med-pretty-name 1 real-count blank-2 filter-1-1
  1 cpu-min 2 cpu-min (compare-1-2 cpu-min)
  1 time-mean 2 time-mean (compare-1-2 time-mean)
  1 disk-mean 2 disk-mean)
(defreport gabriel-compare3
  short-pretty-name 1 count cpu-min 2 cpu-min 3 cpu-min
  (compare-1-2 cpu-min) (compare-1-3 cpu-min) (compare-2-3 cpu-min)
  blank-3 1 time-min 2 time-min 3 time-min)

(defreport min-max-report  name count cpu-mean cpu-min cpu-max time-min time-max disk-min disk-max)

(defreport file-compare
  med-pretty-name 1 count blank-2
  1 cpu-min 2 cpu-min (compare-1-2 cpu-min)
  1 time-mean 2 time-mean (compare-1-2 time-mean)
  1 time-dev 2 time-dev
  1 cons-mean 2 cons-mean)

(defreport instruction-report long-pretty-name long-count cpu-mean cpu-median cpu-dev
	   fault-mean disk-mean cons-mean un-normalized-time)
(defreport instruction-compare
  long-pretty-name  1 long-count 2 long-count blank-2
  1 cpu-mean 2 cpu-mean  (compare-1-2 cpu-mean)
  1 cpu-dev 2 cpu-dev)
(defreport instruction-compare-x
  long-pretty-name  filter-1-1 1 real-count 2 real-count
  1 cpu-mean 2 cpu-mean  (compare-1-2 cpu-mean)
  1 cpu-dev 2 cpu-dev)
(defreport instruction-compare-u
  long-pretty-name  filter-1-1 1 real-count 2 real-count
  1 cpu-mean 2 cpu-mean  (compare-1-2 cpu-mean)
  1 un-normalized-time 2 un-normalized-time (compare-1-2 un-normalized-time))

(defreport interactive-report
  med-pretty-name count time-mean time-dev time-min time-max
  disk-max disk-min fault-min fault-max)

(defreport interactive-compare
  pretty-name count
  1 time-mean 2 time-mean (compare-1-2 time-mean) blank-2
  1 cpu-mean 2 cpu-mean (compare-1-2 cpu-mean) blank-2
  1 cons-max 2 cons-max (compare-1-2 cons-max)
  )

(defreport realtime-compare
  pretty-name count
  1 time-min 2 time-min (compare-1-2 time-min)
  1 time-max 2 time-max (compare-1-2 time-max)
  1 time-dev 2 time-dev
  )
  
(defreport cpu-compare
  pretty-name count
  1 cpu-min 2 cpu-min (compare-1-2 cpu-min)
  1 cpu-max 2 cpu-max (compare-1-2 cpu-max)
  1 cpu-dev 2 cpu-dev
  )

(defreport paging-compare
  pretty-name count
  1 disk-min 2 disk-min (compare-1-2 disk-min)
  1 disk-max 2 disk-max (compare-1-2 disk-max) (compare-1-2 disk-mean)
  1 cons-max 2 cons-max (compare-1-2 cons-max)
  )


(defreport interactive-compare3 ;; same as realtime-compare3
  pretty-name  count 1 time-mean 2 time-mean 3 time-mean
  (compare-1-2 time-mean)  (compare-1-3 time-mean)  (compare-2-3 time-mean)
  blank-3 1 time-dev 2 time-dev 3 time-dev
  )

(defreport realtime-compare3 ;; Same as interactive-compare3
  pretty-name  count 1 time-mean 2 time-mean 3 time-mean
  (compare-1-2 time-mean)  (compare-1-3 time-mean)  (compare-2-3 time-mean)
  blank-3 1 time-dev 2 time-dev 3 time-dev
  )

(defreport instruction-compare3 ;; Almost same as cpu-compare3
  long-pretty-name  1 cpu-mean 2 cpu-mean 3 cpu-mean
  (compare-1-2 cpu-mean)  (compare-1-3 cpu-mean)  (compare-2-3 cpu-mean)
  1 cpu-dev 2 cpu-dev 3 cpu-dev
  )

(defreport cpu-compare3
  pretty-name  count 1 cpu-mean 2 cpu-mean 3 cpu-mean
  (compare-1-2 cpu-mean)  (compare-1-3 cpu-mean)  (compare-2-3 cpu-mean)
  blank-3 1 cpu-dev 2 cpu-dev 3 cpu-dev
  )

(defreport min-cpu-compare3
  pretty-name  count 1 cpu-min 2 cpu-min 3 cpu-min
  (compare-1-2 cpu-min)  (compare-1-3 cpu-min)  (compare-2-3 cpu-min)
  blank-3 1 cpu-dev 2 cpu-dev 3 cpu-dev
  )

(defreport paging-compare3
  pretty-name  count 1 disk-mean 2 disk-mean 3 disk-mean
  (compare-1-2 disk-mean)  (compare-1-3 disk-mean)  (compare-2-3 disk-mean)
  blank-3 1 disk-dev 2 disk-dev 3 disk-dev
  )

(defreport system-summary
  pretty-name count
  time-mean time-dev cpu-mean cpu-dev
  disk-mean fault-min fault-max cons-min cons-max
  )
 

;;; The following alist defines the report-type symbols used above.
;;; The list is of the form: (name format-string top-header bottom-header function . args)
;;; Where NAME 		is the report-type name used in a report-list
;;;       FORMAT-STRING is a format control string used for printing the headers and results.
;;;       TOP-HEADER	is a string used as the top line of the two line label
;;;	  BOTTOM-HEADER is a string used as the bottom line of the two line label
;;;	  FUNCTION	is the function whose result is printed with FORMAT-STRING as the report value.
;;;	  ARGS		FUNCTION is called with arguments BENCHMARK . ARGS, where BENCHMARK is the
;;;			benchmark structure object being reported.  If the report is called within
;;;			a list [for example (compare-1-3 disk-mean)] then the function is called
;;;			with arguments `(,benchmark ,@args . (cdr call))

(DEFPARAMETER report-types
	  '((name "~40a" "" "Benchmark Name" benchmark-name)
	    (long-pretty-name "~40a" "" "Benchmark Name" benchmark-short-pretty-name 40)
	    (pretty-name "~32a" "" "Benchmark Name" benchmark-short-pretty-name 32)
	    (med-pretty-name "~25a" "" "Benchmark Name" benchmark-short-pretty-name 25)
	    (short-pretty-name "~16a" "" "Benchmark Name" benchmark-short-pretty-name 16)
	    (count " ~3d" "Num" "Run" benchmark-count)
	    (long-count " ~6d" "Number" "  Run " benchmark-count)
	    (real-count " ~4d" "Num" "Rep" analyze-history time count) ;; takes filter into account
	    (size " ~3d" "Size" "Samples" analyze-history time size)
	    (samples " ~3d" "Samples" "Run" analyze-history time samples)
	    (real-count " ~7a" " size* " "samples" analyze-history time real-count)

	    (un-normalized-time " ~13a" "un-normalized" " RealTime" average-time-string benchmark-un-normalized-time)

	    (mean-response " ~10a" "Mean" "Resp" average-disk-response mean)
	    (max-response " ~10a" " Max" "Resp" average-disk-response max)
	    (min-response " ~10a" " Min" "Resp" average-disk-response min)

	    (time-mean " ~10a" "  Mean  " "RealTime" analyze-time TIME mean)
	    (time-median " ~10a" " Median " "RealTime" analyze-time TIME median)
	    (time-dev " ~10a" "Std.Dev." "RealTime" analyze-time TIME std-dev)
	    (time-min " ~10a" "  Min  " "RealTime" analyze-time TIME Min)
	    (time-max " ~10a" "  Max  " "RealTime" analyze-time TIME max)
	    (time-total " ~8d" "total" "" analyze-history time total)
	    (time-plot"" "" "" plot-history time)
 
	    (cpu-mean " ~10a" "  Mean  " "  Cpu   " analyze-time CPU mean) 
	    (cpu-median " ~10a" " Median " "  Cpu   " analyze-time CPU median)
	    (cpu-dev " ~10a" "Std.Dev." "  Cpu   " analyze-time CPU std-dev)
	    (cpu-min " ~10a" "  Min  " "  Cpu   " analyze-time CPU Min)
	    (cpu-max " ~10a" "  Max  " "  Cpu   " analyze-time CPU max)
	    (cpu-plot "" "" "" plot-history cpu)

	    (disk-mean " ~10a" "  Mean  " "PageTime" analyze-time DISK-time mean) 
	    (disk-median " ~10a" " Median " "PageTime" analyze-time DISK-time median)
	    (disk-dev " ~10a" "Std.Dev." "PageTime" analyze-time DISK-time std-dev)
	    (disk-min " ~10a" "  Min  " "PageTime" analyze-time DISK-time Min)
	    (disk-max " ~10a" "  Max  " "PageTime" analyze-time DISK-time max)
	    (disk-plot"" "" "" plot-history disk-time)

	    (paging-mean " ~10a" "  Mean  " "PagingTime" analyze-time PAGING-time mean) 
	    (paging-median " ~10a" " Median " "PagingTime" analyze-time PAGING-time median)
	    (paging-dev " ~10a" "Std.Dev." "PagingTime" analyze-time PAGING-time std-dev)
	    (paging-min " ~10a" "  Min  " "PagingTime" analyze-time PAGING-time Min)
	    (paging-max " ~10a" "  Max  " "PagingTime" analyze-time PAGING-time max)
	    (paging-plot"" "" "" plot-history paging-time)

	    (fault-mean " ~6d" " Mean " "Faults" analyze-history-fix FAULTS mean) 
	    (fault-median " ~6d" "Median" "Faults" analyze-history-fix FAULTS median)
	    (fault-dev " ~6d" "StdDev" "Faults" analyze-history-fix FAULTS std-dev)
	    (fault-min " ~6d" " Min " "Faults" analyze-history-fix FAULTS Min)
	    (fault-max " ~6d" " Max " "Faults" analyze-history-fix FAULTS max)
	    (fault-plot"" "" "" plot-history faults)

	    (cons-mean " ~6d" " Mean " " Cons " analyze-history-fix CONS mean) 
	    (cons-median " ~6d" "Median" " Cons " analyze-history-fix CONS median)
	    (cons-dev " ~6d" "StdDev" " Cons " analyze-history-fix CONS std-dev)
	    (cons-min " ~6d" " Min " " Cons " analyze-history-fix CONS Min)
	    (cons-max " ~6d" " Max " " Cons " analyze-history-fix CONS max)
	    (cons-plot"" "" "" plot-history cons)

	    (compare-1-2 " ~5,2f" "2 rel" "to 1" compare-machines 1 2)
	    (compare-1-3 " ~5,2f" "3 rel" "to 1" compare-machines 1 3)
	    (compare-2-3 " ~5,2f" "3 rel" "to 2" compare-machines 2 3)

	    (compare-2-1 " ~5,2f" "1 rel" "to 2" compare-machines 2 1)
	    (compare-3-1 " ~5,2f" "1 rel" "to 3" compare-machines 3 1)
	    (compare-2-3 " ~5,2f" "2 rel" "to 3" compare-machines 3 2)

	    (delta-1-2 " ~5,2f" "2 rel" "to 1" compare-percent 1 2)
	    (delta-1-3 " ~5,2f" "3 rel" "to 1" compare-percent 1 3)
	    (delta-2-3 " ~5,2f" "3 rel" "to 2" compare-percent 2 3)

	    (delta-2-1 " ~5,2f" "1 rel" "to 2" compare-percent 2 1)
	    (delta-3-1 " ~5,2f" "1 rel" "to 3" compare-percent 3 1)
	    (delta-2-3 " ~5,2f" "2 rel" "to 3" compare-percent 3 2)

	    (filter " ~8a" "Filter" "" set-analyze-filter)
	    (filter-0-0 ~8a "Filter" " Off " set-analyze-filter (0 0))
	    (filter-1-0 ~8a "Ignore" "Smallest" set-analyze-filter (1 0))
	    (filter-0-1 ~8a "Ignore" "Largest" set-analyze-filter (0 1))
	    (filter-1-1 ~8a "Ignore" "Extreme" set-analyze-filter (1 1))
	    (filter-2-0 ~8a "Ignore" "Small 2" set-analyze-filter (2 0))
	    (filter-0-2 ~8a "Ignore" "Large 2" set-analyze-filter (0 2))
	    (filter-2-2 ~8a "Ignore" "Extreme 2" set-analyze-filter (2 2))

	    (GET ~8f header2 header1 analyze-history)	;; Example of use: (get cpu mean)
	    (time ~10a header2 header1 analyze-time)
	    (FIX ~8d header2 header1 analyze-history-fix)

	    (FUNCALL " ~8a" " User " "Function" report-funcall)  ;; Use as (funcall function . args)
							         ;; Calls FUNCTION with args benchmark . ARGS
	    							 ;; and prints the results

	    (EVAL " ~10a" " User " "Expression" report-eval)     ;; Use as (eval (expression))
	    ;; Warning: on the explorer this is global:eval	 ;; Evals expression with BENCHMARK bound to the benchmark
	    ;;          when in common-lisp mode.		 ;; and prints the results
	    
	    (blank-1 " " "" "" ignore)
	    (blank-2 "  " "" "" ignore)
	    (blank-3 "   " "" "" ignore)
	    (blank-4 "    " "" "" ignore)
	    ))


(DEFUN report-benchmark (name &optional (machines *all-machines*) &aux bench)
  (UNLESS (CONSP machines) (SETQ machines (LIST machines)))
  (DOLIST (machine machines)
    (WHEN (SETQ bench (get-benchmark name machine t))
      (FORMAT t "~2%~a" machine)
      (FORMAT t "~%~8a ~8a ~8a ~8a ~6d ~12d" "RealTime" "DiskTime" "CPU Time" "PageTime" "Faults" "Consing")
      (LOOP for (real-time disk-time page-faults consing cpu-time paging-time . rest) in (benchmark-history bench)
	    do (FORMAT t "~%~8a ~8a ~8a ~8a ~6d ~12d ~{ ~s ~d~}"
		       (pretty-time real-time)
		       (pretty-time disk-time)
		       (pretty-time cpu-time)
		       (IF paging-time (pretty-time paging-time) "")
		       page-faults consing rest)))))

(DEFVAR benchmark-pagenum nil "Page number for report-benchmarks")

(DEFVAR analyze-filter '(0 0)
 "Analyze filter specs.  Ignore the shortest (first filter) and ignore the longest (second filter)")

(defun report-benchmarks (&optional benchmarks (report-list std-report)
			  machine-types (stream *standard-output*) &aux reports)
  "Print a line for each benchmark in BENCHMARKS and one column for each entry in REPORT-LIST.
MACHINE-TYPES is a list of machines to report on.
See *all-benchmark-classes* for a list of all known benchmark lists.
See *all-reports* for a list of all known report lists.
See *all-machines* for a list of all loaded machine-types (see restore-results)."
  (DECLARE (SPECIAL machine-types))
  (UNLESS benchmarks (setq benchmarks *all-benchmarks*))
  (UNLESS (CONSP machine-types) (SETQ machine-types (LIST machine-types)))
  (WHEN (SYMBOLP report-list)
    (IF (MEMBER REPORT-LIST *ALL-REPORTS* :TEST #'EQ)
	(SETQ report-list (SYMBOL-VALUE report-list))
      (FERROR "~s isn't a report list or name of one - see *all-reports*" report-list)))
  ;; Report-list is a list of report names.
  ;; Transform it into a list of report descriptors from the report-types alist.
  (SETQ reports
	(LOOP for report in report-list
	      with type do
	      (COND ((CONSP report)
		     (UNLESS (SETQ type (ASSOC (CAR REPORT) REPORT-TYPES :TEST #'EQ))
		       (FERROR "~s isn't a known report-type" (CAR report)))
		     (SETQ type (APPEND type (CDR report))))
		    ((NUMBERP report) (SETQ type report))
		    (t (UNLESS (setq type (ASSOC REPORT REPORT-TYPES :TEST #'EQ))
			 (FERROR "~s isn't a known report-type" report))))
	      collect type))
  ;; Print out a line for every benchmark
  (loop for name in benchmarks
	with i = 0 and value 
	with analyze-filter = analyze-filter ;; Bind analyze-filter to its default value
	do
	(WHEN (ZEROP (MOD i 48.)) ;; Print header every 48 lines
	  (print-report-header reports machine-types stream)
	  (INCF i 3))
	;case 1: benchmark ran on all machines so print results
	;case 2: put benchmark name on property list of machines not ran
	(COND ((LOOP for machine in machine-types	;If benchmark found for all machine types
		  for benchmark = (get-benchmark name machine :no-error)
		  always (AND benchmark (benchmark-history benchmark)))
	       (TERPRI stream)
	       (LOOP for report in reports
		     with machine = (FIRST machine-types)
		     with benchmark = (get-benchmark name machine)
		     when (FIXNUMP report)
		     do (SETQ machine (NTH (1- report) machine-types)
			      benchmark (get-benchmark name machine))
		     else do
		     (LET* ((benchmark-name name)
			    (analyze-function (GET benchmark-name 'analyze-function))
			    (fmt (SECOND report))
			    (FUNCTION (FIFTH report))
			    (args (NTHCDR 5 report)))
		       (DECLARE (SPECIAL benchmark-name analyze-function))
		       (SETQ value (IF (benchmark-history benchmark)
				       (APPLY function benchmark args)
				     "Not Run"))
		       (report stream fmt value)))
	       (INCF i))
	      (t (DOLIST (machine machine-types)	
		   (LET ((benchmark (get-benchmark name machine :no-error)))
		     (UNLESS (AND benchmark (benchmark-history benchmark))
		       (PUTPROP machine (CONS name (GET machine :notrun)) :notrun)))))
	)
	finally
	(DOLIST (mach machine-types)
	  (WHEN (get mach :notrun)
	    (report stream "~2%The following benchmarks were not run on machine: ~a" mach)
	    (report stream "~%~{~%~40a~}" (GET mach :notrun))
	    (REMPROP mach :notrun)))
  )
		    
  (VALUES))

(DEFF report-results 'report-benchmarks) ;; synonym

(DEFUN print-report-header (reports machine-types stream &aux machine temp time)
  (FORMAT stream "~%~|~%")
  (WHEN benchmark-pagenum
    (INCF benchmark-pagenum)
    (SETQ time (WITH-OUTPUT-TO-STRING (stream) (time:print-universal-time (time:get-universal-time) stream)))
    (FORMAT stream "~20tBenchmark Results~60tTI internal data~:[~100t~;~80t*~100t~]~a  Page ~3d~2%"
	    *real-max-min* time Benchmark-Pagenum))
  (IF (NULL (cdr machine-types)) ;; When only one machine type
      (PRINC (CAR machine-types) stream)
    (SETQ machine (FIRST machine-types))
    (LOOP for report in reports
	  when (FIXNUMP report) do (SETQ machine (NTH (1- report) machine-types))
	  else do (REPORT stream (SECOND report)
			  (COND ((MEMBER (FIFTH REPORT) '(COMPARE-MACHINES IGNORE) :TEST #'EQ) "")
				((<= (LENGTH (string machine))
				     (SETQ temp (MAX (LENGTH (string (THIRD report)))
						     (LENGTH (string (FOURTH report)))))) machine)
				(t (NSUBSTRING machine 0 temp))))))
  (TERPRI stream)
  (LOOP for report in reports
	unless (FIXNUMP report) do
	(REPORT stream (second report) (print-header report (THIRD report))))
  (TERPRI stream)
  (LOOP for report in reports
	unless (FIXNUMP report) do
	(REPORT stream (second report) (print-header report (FOURTH report)))))

(DEFUN print-header (report header)
  (IF (STRINGP header) header
    (APPLY header (NTHCDR 5 report))))

(DEFUN header1 (ARG &rest ignore)
  arg)

(DEFUN header2 (IGNORE arg &rest ignore)
  arg)

(DEFVAR compare-with-percent nil "Set non-nil to see comparisions in percent.")

(DEFUN compare-percent (benchmark m1 m2 report)
  (LET ((compare-with-percent t))
    (compare-machines benchmark m1 m2 report)))

(DEFUN compare-machines (benchmark m1 m2 report)
  (DECLARE (SPECIAL machine-types benchmark-name))
  benchmark					;not used
  (WHEN (FIXNUMP m1) (SETQ m1 (NTH (1- m1) machine-types)))
  (WHEN (FIXNUMP m2) (SETQ m2 (NTH (1- m2) machine-types)))
  (LET* ((type (OR (ASSOC REPORT REPORT-TYPES :TEST #'EQ)
		   (FERROR "~a isn't a known report-type" report)))
	 (args (NTHCDR 5 type))
	 (b1 (get-benchmark benchmark-name m1))
	 (b2 (get-benchmark benchmark-name m2))
	 (s1 (GET (benchmark-plist b1) 'summary-list))
	 (s2 (GET (benchmark-plist b2) 'summary-list)))
    (IF (OR s1 s2)
	(compare-machine-summary m1 m2 report (INTERSECTION s1 s2) (GET (benchmark-plist b1) 'summary))
      (compare-benchmarks b1 b2 (FIRST args) (SECOND args)))))

(DEFUN compare-benchmarks (benchmark1 benchmark2 type need)
  "Return the comparision of (/ benchmark1 benchmark2) for TYPE and NEED.
 NEED is one of (mean mode median std-dev min max total size samples count real-count)
 TYPE is one of (time disk-time paging-time faults cons cpu)"
  (LET ((result1 (analyze-history benchmark1 type need))
	(result2 (analyze-history benchmark2 type need)))
    (cond ((NOT (AND (NUMBERP result1) (NUMBERP result2))) "")
	  ((ZEROP result2) "")
	  ((OR (MINUSP result1) (MINUSP result2)) "")
	  (compare-with-percent
	   (FORMAT nil "~d%" (ROUND (* (/ (- result1 result2) (FLOAT result1)) -100.))))
	  (t (/ result1 (FLOAT result2))))))

(DEFUN compare-machine-summary (m1 m2 report benchmark-list need)
  (LET* ((report-type (OR (ASSOC REPORT REPORT-TYPES :TEST #'EQ)
			  (FERROR "~a isn't a known report-type" report)))
	 (type (CAR (NTHCDR 5 report-type))))
    (LOOP for name in benchmark-list
	  with n = 0 and total = 0 and result
	  do (SETQ result (compare-benchmarks (get-benchmark name m1) (get-benchmark name m2) type need))
	  (WHEN (NUMBERP result)
	    (INCF n)
	    (INCF total result))
	  finally (RETURN (IF (PLUSP n) (/ total n) "")))))

(DEFUN set-analyze-filter (ignore filter &optional result)
  (SETQ analyze-filter filter)
  (OR result analyze-filter))

(DEFUN report-funcall (benchmark function &rest args)
  (APPLY function benchmark args))

(DEFUN report-eval (benchmark expression)
  (DECLARE (SPECIAL benchmark))
  (EVAL expression))

(DEFUN average-disk-response (benchmark type)
  (LET ((faults (analyze-history benchmark 'faults type))
	(disk-time (analyze-history benchmark 'disk-time type)))
    (IF (AND (NUMBERP faults) (PLUSP faults)
	     (NUMBERP disk-time))
	(pretty-time (/ disk-time faults))
      "")))

(DEFUN average-time-string (benchmark function)
  (let ((count (benchmark-count benchmark)))
    (if (plusp count)
	(pretty-time (/ (FLOAT (FUNCALL function benchmark)) count))
	(pretty-time 0))))

(DEFUN analyze-time (benchmark type need)
  (pretty-time (analyze-history benchmark type need)))

(DEFUN analyze-history-fix (benchmark type need &aux result)
  (SETQ result (analyze-history benchmark type need))
  (IF (NUMBERP result)
      (ROUND result)
    result))

(EVAL-WHEN (COMPILE load eval)
(defmacro square (x)
   (once-only (x)
     `(* ,x ,x)))

(defmacro half (x)
   `(truncate ,x 2))
) ;end eval-when

(DEFSUBST get-history-item (item history)
  (IF (SYMBOLP item)
      (GET (NTHCDR 5 history) item)
    (NTH item history)))

(DEFUN type-item-num (type)
  (LET ((item-num (SECOND (ASSOC TYPE '((TIME 0) (DISK-TIME 1) (FAULTS 2) (CONS 3) (CPU 4) (PAGING-TIME 5)) :TEST #'EQ))))
    (UNLESS item-num
      (IF (AND (SYMBOLP type) (EQ (SYMBOL-PACKAGE type) (FIND-PACKAGE 'keyword)))
	  (SETQ item-num type)
	(FERROR "~s isn't a known item name (time disk-time paging-time faults cons cpu) or a keyword" type)))
    item-num))

(defun analyze-history (benchmark type need)
  "Return the NEED statistic from the TYPE numbers in BENCHMARK
 NEED is one of (mean mode median std-dev min max total size samples count real-count un-normalized-time)
 TYPE is one of (time disk-time paging-time faults cons cpu) or a keyword"
  (LET* ((item-num (type-item-num type))
	 (history (benchmark-history benchmark))
	 (sorted-history (benchmark-sorted-history benchmark))
	 (ignore-small (FIRST analyze-filter))
	 (ignore-large (SECOND analyze-filter))
	 number-of-times total mean)
    (WHEN (AND (CONSP history) (NOT (get-history-item item-num (FIRST history))))
      (SETQ history nil)) ;; Don't blow up when history doesn't the item we need
    (IF (ATOM history)
	(IF history
	    (FUNCALL history benchmark type need)
	  "")
      (UNLESS sorted-history
	(SETF (benchmark-sorted-history benchmark)
	      (SETQ sorted-history
		    (sort (COPY-LIST (benchmark-history benchmark))
			  #'(lambda (a b) (< (CAR a) (CAR b)))))))
      (SETQ number-of-times (LENGTH sorted-history))
      (WHEN (> number-of-times (+ ignore-small ignore-large))
	(SETQ number-of-times (- number-of-times ignore-small ignore-large))
	(SETQ sorted-history (NTHCDR ignore-small sorted-history)))
      (SETQ total (LOOP for entry in sorted-history
			for i below number-of-times
			summing (GET-HISTORY-ITEM item-num entry)))
      (SETQ mean (/ total (FLOAT number-of-times)))
      (case need
	(mean mean)
	(mode (half (+ (get-history-item item-num (first sorted-history))
		       (get-history-item item-num (NTH (1- number-of-times) sorted-history)))))
	(median (if (oddp number-of-times)
		    (GET-HISTORY-ITEM item-num (nth (half (1- number-of-times)) sorted-history))
		  (half (+ (GET-HISTORY-ITEM item-num (nth (half number-of-times) sorted-history))
			   (GET-HISTORY-ITEM item-num (nth (half (1- number-of-times)) sorted-history))))))
	(std-dev (loop for item in sorted-history
		       for i below number-of-times
		       for time = (GET-HISTORY-ITEM item-num item)
		       summing (square (- time mean)) into variation
		       finally (return (sqrt (/ (float variation)
						 (1+ number-of-times))))))
	(MIN (IF *real-max-min*
		 (LOOP for item in history
		       minimize (GET-HISTORY-ITEM item-num item))
	       (GET-HISTORY-ITEM item-num (FIRST sorted-history))))
	(MAX (IF *real-max-min*
		 (LOOP for item in history
		       maximize (GET-HISTORY-ITEM item-num item))
	       (GET-HISTORY-ITEM item-num (NTH (1- number-of-times) sorted-history))))
	(total total)
	(size (/ (benchmark-count benchmark) (LENGTH (benchmark-history benchmark))))
	(samples number-of-times)
	(COUNT (* number-of-times (/ (benchmark-count benchmark)
				      (LENGTH (benchmark-history benchmark)))))
	(real-count (FORMAT nil "~d*~d" number-of-times (/ (benchmark-count benchmark)
							    (LENGTH (benchmark-history benchmark)))))
	(un-normalized-time (/ (FLOAT (benchmark-un-normalized-time benchmark)) (benchmark-count benchmark)))
	(otherwise (FERROR "~s isn't a known statistic type (mean mode median std-dev min max)" need))))))

(DEFUN plot-history (benchmark type)
  (let ((item-num (type-item-num type))
	min max)
    (WHEN (get-history-item item-num (FIRST (benchmark-history benchmark)))
      (LOOP for item in (benchmark-history benchmark)
	    for value = (ROUND (GET-HISTORY-ITEM item-num item))
	    maximize value into maxv
	    minimize value into minv
	    finally (SETQ min minv max maxv))
      (REPORT t "~%~12a~2%~12a~2%~12a" (pretty-time max) type (pretty-time min))
      (MULTIPLE-VALUE-BIND (x-base y-base) (SEND *terminal-io* :read-cursorpos)
	(INCF y-base (half (tv:sheet-line-height *terminal-io*)))
	(LOOP with plot-width = (* 4 (tv:sheet-line-height *terminal-io*))
	      for item in (benchmark-history benchmark)
	      for p-x first nil then x
	      for x upfrom x-base by 4
	      for p-y first nil then y
	      for y = (- y-base (interpolate (GET-HISTORY-ITEM item-num item) plot-width max min))
	      for i upfrom 1
	      do (SEND *terminal-io* :draw-point x (+ y-base 2))
	      WHEN (ZEROP (MOD i 10))
	      do (SEND *terminal-io* :draw-point x (+ y-base 3))
	      when p-y do
	      (SEND *terminal-io* :draw-line p-x p-y x y))))))

(DEFUN interpolate (value size max min)
  (IF (= MAX MIN) 0
    (ROUND (* (- value min) (/ size (FLOAT (- max min)))))))

(defun pretty-time (time)
  (cond ((STRINGP time) time)
	((MINUSP time) "????")
	((< time 0.001) "   0")
	((< time 1)
	 (report nil "  ~3F ns" (* time 1000.0)))
	((< time 1000)
	 (report nil " ~4F us" time))
	((< time 1000000.)
	 (report nil " ~4F ms" (/ time 1000.0)))
	(t
	 (report nil "~5F s " (/ time 1e6)))))

;; SUM  of MEAN for *list* is the total of all the mean times for all the benchmarks in *list*
;; SUM  of MAX  for *list* is the total of all the max  times for all the benchmarks in *list*
;; SUM  of MIN  for *list* is the total of all the min  times for all the benchmarks in *list*

;; MEAN of MEAN for *list* is the average of the RATIOS between mean benchmark results for the benchmarks in *list*
;; MEAN of MAX  for *list* is the average of the RATIOS between max  benchmark results for the benchmarks in *list*
;; MEAN of MIN  for *list* is the average of the RATIOS between max  benchmark results for the benchmarks in *list*

(DEFMACRO sum-benchmarks (benchmark-list machine &optional type)
  "Append Sum and average pseudo benchmarks to the end of benchmarks in BENCHMARK-LIST
TYPE is one of (mean mode median std-dev min max)
TYPE and MACHINE may be lists."
  #-3600 (DECLARE (ARGLIST benchmark-list machine &optional (type '(mean max min))))
  `(PROGN
     (SETQ ,benchmark-list (sum-benchmarks-internal ,benchmark-list ',benchmark-list ,machine ,type))
     (VALUES)))

;; Loop over lists of machines and types
(DEFUN sum-benchmarks-internal (benchmark-list list-name machine-list type-list &aux result)
  (UNLESS type-list (SETQ type-list '(mean max min)))
  (UNLESS (CONSP machine-list) (SETQ machine-list (LIST machine-list)))
  (UNLESS (CONSP type-list) (SETQ type-list (LIST type-list)))
  (LOOP for machine in machine-list do
	(LOOP for type in type-list do
	      (SETQ result (sum-benchmarks-internal1 benchmark-list list-name machine type machine-list))))
  result)

;; Do the real work
;;modified 5/13/86 SBW. added an additional predicate to ensure
;;summaries include only benchmarks that were run on all machines.

(DEFUN sum-benchmarks-internal1 (benchmark-list list-name machine type machine-list)
  (CHECK-TYPE benchmark-list list)
  (CHECK-TYPE list-name symbol)
  (LET ((sum-name (INTERN (STRING-APPEND "SUM-" type "-" list-name)))
	(mean-name (INTERN (STRING-APPEND "MEAN-" type "-" list-name))))
    ;; Remove old sum and mean, if any
    (SETQ benchmark-list (DELETE SUM-NAME (THE LIST (DELETE MEAN-NAME (THE LIST BENCHMARK-LIST) :COUNT 1 :TEST #'EQ))
				 :COUNT 1 :TEST #'EQ))
    (LOOP for name in benchmark-list
	  for benchmark = (get-benchmark name machine :no-error)
	  with list-of-benchmarks
	  with time and  disk and paging and faults and cons and cpu
	  with time-total = 0 and disk-total = 0 and paging-total = 0 and fault-total = 0
	  with cons-total = 0 and cpu-total = 0 and n = 0.0 do
	  (WHEN (AND (LOOP for mach in machine-list                            ;;Don't summarise unless..
			   for bmark = (get-benchmark name mach :no-error)
			   always (AND bmark (benchmark-history bmark)))       ;;Benchmark was ran on all machines
		     (NOT (GET (benchmark-plist benchmark) 'summary)))         ;;and benchmark is not a summary benchmark
	    (SETQ time (analyze-history benchmark 'TIME type)
		  disk (analyze-history benchmark 'disk-time type)
		  paging (analyze-history benchmark 'paging-time type)
		  faults (analyze-history benchmark 'faults type)
		  cons (analyze-history benchmark 'cons type)
		  cpu (analyze-history benchmark 'cpu type))
	    (INCF time-total time)
	    (INCF disk-total disk)
	    (WHEN (NUMBERP paging) ;; Because not all results have paging
	      (INCF paging-total paging))
	    (INCF fault-total faults)
	    (INCF cons-total cons)
	    (INCF cpu-total cpu)
	    (INCF n)
	    (PUSH name list-of-benchmarks))
	  finally
	  (LET ((sum-history
		  (make-benchmark-history
		    real-time time-total disk-time disk-total paging-time paging-total
		    page-faults fault-total consing cons-total cpu-time cpu-total)))
	    (LET ((sum (make-benchmark name sum-name pretty-name (FORMAT nil "Sum of ~a for ~a" type list-name)
				       history (LIST sum-history) count (FLOOR n)
				       plist (LIST nil 'summary type 'classes (LIST list-name))))
		  (mean (make-benchmark name mean-name pretty-name (FORMAT nil "Mean of ~a for ~a" type list-name)
					history 'summing-analyzer count (FLOOR n) 
					plist (LIST nil 'summary type 'summary-list (NREVERSE list-of-benchmarks)))))
	      (PUTPROP sum-name sum machine)
	      (PUTPROP mean-name mean machine)
	      )))
    (NCONC benchmark-list (LIST sum-name mean-name))))

(DEFUN summing-analyzer (&rest ignore) "")

;; This function doesn't do anything with paging time - update it someday
(DEFUN filter-bad-cputime (&optional (machines my-machine-type) &aux (COUNT 0))
  "Delete history items with negative cpu-times for all benchmarks on MACHINES.
Negative cpu-times occur when the disk-time is greater than the real-time.
This happens because the microcode is careless about microsecond clock wrap-around
when updating the disk-time meter."
  (UNLESS (CONSP machines) (SETQ machines (LIST machines)))
  (DOLIST (machine machines)
    (LOOP for name in *all-benchmarks*
	  for benchmark = (get-benchmark name machine :no-error)
	  for fixed = 0
	  with history and history-length and new-length
	  when (AND benchmark (SETQ history (benchmark-history benchmark)))
	  do (SETQ history (benchmark-history benchmark)
		   history-length (LENGTH history))
	     (SETQ history
		   (DELETE nil (THE list history) :test
			   #'(lambda (IGNORE element)
			       (WHEN (numberp (benchmark-history-paging-time element))
				 (SETF (benchmark-history-cpu-time element) (- (benchmark-history-real-time element)
									       (benchmark-history-paging-time element))))
			       (WHEN (MINUSP (benchmark-history-cpu-time element))
				 (SETF (benchmark-history-cpu-time element) (- (benchmark-history-real-time element)
									       (benchmark-history-disk-time element)))
				 (UNLESS (MINUSP (benchmark-history-cpu-time element))
				   (INCF fixed)))
			       (MINUSP (benchmark-history-cpu-time element))))
		   new-length (LENGTH history))
	     (WHEN (PLUSP fixed) ;; These should never happen
	       (FORMAT t "~%Fixed ~d bad events from ~a" fixed (benchmark-name benchmark))
	       (SETF (benchmark-sorted-history benchmark) nil))
	     (WHEN (NEQ history-length new-length)
	       (FORMAT t "~%Deleted ~d bad events from ~a" (- history-length new-length) (benchmark-name benchmark))
	       (INCF count)
	       (SETF (benchmark-count benchmark) (* (/ (benchmark-count benchmark) history-length) new-length))
	       (SETF (benchmark-history benchmark) history)
	       (SETF (benchmark-sorted-history benchmark) nil))))
  count)

(DEFUN merge-results (machine-list new-machine &optional (benchmark-list *all-benchmarks*))
  "For every benchmark in BENCHMARK-LIST, take the results for each of the machines in MACHINE-LIST
and combine them under the name NEW-MACHINE.  NEW-MACHINE may be one of the machines in machine-list,
or a new name.  This can also be used to copy benchmarks from one machine name to another.
Benchmark pretty-name and property list come from the benchmarks for the first machine in machine-list."
  (UNLESS (CONSP machine-list) (SETQ machine-list (LIST machine-list)))
  (LOOP for machine in machine-list
	for herald = (GET machine 'herald)
	do (COND ((NULL herald) (FERROR "~s isn't the name of a benchmark."))
		 ((STRINGP herald) (SETQ herald (LIST herald))))
	append herald into merged-herald
	finally (PUTPROP new-machine merged-herald 'herald))
  (LOOP for name in benchmark-list
	with not-run
	do (LOOP for machine in machine-list
		 for benchmark = (get-benchmark name machine t)
		 with first = t and incomplete and new-benchmark do
		 (IF (NULL benchmark) (SETQ incomplete t)
		   (COND (FIRST (SETQ first nil)
				(SETQ new-benchmark (1make-benchmark))
				(COPY-ARRAY-CONTENTS benchmark new-benchmark)
				(SETF (benchmark-sorted-history new-benchmark) nil))
			 (t (INCF (benchmark-count new-benchmark) (benchmark-count benchmark))
			    (INCF (benchmark-un-normalized-time new-benchmark) (benchmark-un-normalized-time benchmark))
			    (SETF (benchmark-history new-benchmark) (APPEND (benchmark-history new-benchmark)
									    (benchmark-history benchmark))))))
		 finally (UNLESS first
			   (WHEN incomplete (PUSH name not-run))
			   (PUTPROP name new-benchmark new-machine)))
	finally (WHEN not-run
		  (REPORT t "~2%The following benchmarks were not run for all machines:~
     			     ~{~%~40a~}" (NREVERSE not-run)))))

(PUTPROP 'test (make-benchmark name 'test pretty-name "Analyze Test benchmark"
			        count 5 unnormalized-time 1500 history
			       (LIST (make-benchmark-history real-time 100 cpu-time 100)
				     (make-benchmark-history real-time 200 cpu-time 200)
				     (make-benchmark-history real-time 300 cpu-time 300)
				     (make-benchmark-history real-time 400 cpu-time 400)
				     (make-benchmark-history real-time 500 cpu-time 500)))
	 my-benchmark-property)

(DEFUN sort-reports (benchmarks &special machines &optional (type 'TIME) (need 'mean))
  "Sort BENCHMARKS by TYPE and NEED for machine.  If machine is a list, sort by
 the ratio of (first machines) (second machines).
 NEED is one of (mean mode median std-dev min max total size samples count real-count)
 TYPE is one of (time disk-time paging-time faults cons cpu)"
  (COND ((CONSP machines)
	 (LET ((m1 (FIRST machines))
	       (m2 (SECOND machines)))
	   (DECLARE (SPECIAL m1 m2))
	   (SORT (COPY-LIST benchmarks)
		 #'(lambda (n1 n2 &aux b1m1 b1m2 b2m1 b2m2 r1 r2)
		     (WHEN (AND (SETQ b1m1 (get-benchmark n1 m1 t))
				(SETQ b1m2 (get-benchmark n1 m2 t))
				(SETQ b2m1 (get-benchmark n2 m1 t))
				(SETQ b2m2 (get-benchmark n2 m2 t)))
		       (SETQ r1 (compare-benchmarks b1m1 b1m2 type need))
		       (SETQ r2 (compare-benchmarks b2m1 b2m2 type need))
		       (COND ((AND (NUMBERP r1) (NUMBERP r2)) (< r1 r2))
			     ((NUMBERP r1) t)
			     (t nil)))))))
	(t (SORT (COPY-LIST benchmarks)
		 #'(lambda (n1 n2 &aux b1 b2 r1 r2)
		     (WHEN (AND (SETQ b1 (get-benchmark n1 machines t))
				(SETQ b2 (get-benchmark n2 machines t)))
		       (SETQ r1 (analyze-history b1 type need))
		       (SETQ r2 (analyze-history b2 type need))
		       (COND ((AND (NUMBERP r1) (NUMBERP r2)) (< r1 r2))
			     ((NUMBERP r1) t)
			     (t nil))))))))

;------------------------------------------------------------------


(DEFUN report-instruction (machines &optional (reports '(instruction-compare
						     cpu-compare)))
  (DECLARE (SPECIAL *zippel-benchmarks* *string-benchmarks* *storage-allocators* *list-access* *function-call*
		    *message-passing* *array-reference* *arithmetic* *fixnum-arithmetic* *flonum-arithmetic*
		    *bignum-arithmetic* *miscellaneous*
		    *instructions* *pathname-benchmarks* *extended-zippels* *hash-benchmarks* *graphics-benchmarks*
		    *array-push-benchmarks* *eval-benchmarks* *branch-benchmarks*))
  (load-reports machines 'zippel 'push-local-normalization)
  (WHEN (VARIABLE-BOUNDP *miscellaneous*) ;; Don't consider miscellaneous tests to be arithmetic
    (setq *arithmetic* (my-SET-DIFFERENCE *arithmetic* *miscellaneous*)))
  (DOLIST (x '(wallas-test fibinacci-1000 factorial-1000)) 
    (SETQ *arithmetic* (DELETE X (THE LIST *ARITHMETIC*) :COUNT 1 :TEST #'EQ)))
  (sum-benchmarks *instructions* machines)
  (sum-benchmarks *function-call* machines)
  (sum-benchmarks *message-passing* machines)
  (sum-benchmarks *array-reference* machines)
  (sum-benchmarks *arithmetic* machines)
  (sum-benchmarks *fixnum-arithmetic* machines)
  (sum-benchmarks *flonum-arithmetic* machines)
  (sum-benchmarks *bignum-arithmetic* machines)
  (setq *zippel-benchmarks* (union *function-call*	;Union list to Include benchmark sums
				   *message-passing*
				   *array-reference*
				   *arithmetic*
				   *fixnum-arithmetic*
				   *flonum-arithmetic*
				   *bignum-arithmetic*
				   ))
  (LET* ((sym-exp (LIST (FIRST machines) (CAR (LAST machines))))
	 (instruction-cpu (sort-reports (append *instructions* *zippel-benchmarks*) sym-exp 'cpu 'mean)))
    (sum-benchmarks *zippel-benchmarks* machines)
    (sum-benchmarks instruction-cpu machines)
    
    (run-reports (append *instructions* *zippel-benchmarks*) machines '(instruction-report) reports '(instruction-compare3))
    (run-reports instruction-cpu machines '(instruction-report) reports '(instruction-compare3)))
  (print-heralds machines)
  )

(DEFUN report-extensions (machines &optional (reports '(instruction-compare
							 cpu-compare)))
  (DECLARE (SPECIAL *string-benchmarks* *storage-allocators* *list-access* *pathname-benchmarks*
		    *extended-zippels* *hash-benchmarks* *graphics-benchmarks* *miscellaneous*
		    *array-push-benchmarks* *eval-benchmarks* *branch-benchmarks*
		    *flavor-normalizations* *flavor-benchmarks* *interesting-flavor-benchmarks*
		    *method-combination* *instance-variables* *defflavor-benchmarks*
		    *instantiate-flavor* *defmethod-benchmarks* *new-message-passing*
		    *print-benchmarks* *read-benchmarks* *resource-benchmarks* *binding-benchmarks*))
  (load-reports machines 'extensions 'bitblt)
  (sum-benchmarks *string-benchmarks* machines)
  (sum-benchmarks *storage-allocators* machines)
  (sum-benchmarks *list-access* machines)
  (sum-benchmarks *pathname-benchmarks* machines)
  (sum-benchmarks *hash-benchmarks* machines)
  (sum-benchmarks *graphics-benchmarks* machines)
  (sum-benchmarks *array-push-benchmarks* machines)
  (sum-benchmarks *eval-benchmarks* machines)
  (sum-benchmarks *branch-benchmarks* machines)
  (sum-benchmarks *flavor-benchmarks* machines)
  (sum-benchmarks *resource-benchmarks* machines)
  (sum-benchmarks *binding-benchmarks* machines)
  (sum-benchmarks *print-benchmarks* machines)
  (sum-benchmarks *read-benchmarks* machines)
  (SETQ *extended-zippels* (UNION *storage-allocators* *list-access* *miscellaneous*
				  *string-benchmarks* *read-benchmarks* *print-benchmarks* *eval-benchmarks*
				  *pathname-benchmarks* *hash-benchmarks*
				  *resource-benchmarks* *graphics-benchmarks* *array-push-benchmarks*
				  *branch-benchmarks* *binding-benchmarks* *flavor-benchmarks*))
  (LET* ((sym-exp (LIST (FIRST machines) (CAR (LAST machines))))
	 (instruction-cpu (sort-reports *extended-zippels* sym-exp 'cpu 'mean)))
    (sum-benchmarks *extended-zippels* machines)
    (sum-benchmarks instruction-cpu machines)
    (run-reports *extended-zippels* machines '(instruction-report) reports '(instruction-compare3 realtime-compare3))
    (run-reports instruction-cpu machines '(instruction-report) reports '(instruction-compare3 realtime-compare3)))
  (print-heralds machines)
  )

(DEFUN report-gabriel (machines &optional (reports '(gabriel-compare
						      realtime-compare
						      cpu-compare
						      paging-compare)))
  (DECLARE (SPECIAL *gabriel-benchmarks* *gabriel-print*))
  (load-reports machines 'gabriel 'gabriel-fft)
  (SETQ *gabriel-benchmarks* (UNION *gabriel-benchmarks* *gabriel-print*))
  (LET* ((sym-exp (LIST (FIRST machines) (CAR (LAST machines))))
	 (gabriel-cpu (sort-reports *gabriel-benchmarks* sym-exp 'cpu 'min)))
    (sum-benchmarks *gabriel-benchmarks* machines)
    (sum-benchmarks gabriel-cpu machines)
    (run-reports *gabriel-benchmarks* machines '(gabriel-report)
					       reports
					       '(gabriel-compare3 realtime-compare3 paging-compare3))	
    (run-reports gabriel-cpu machines '(gabriel-report)
				      reports
				      '(gabriel-compare3 realtime-compare3 paging-compare3)))
  (print-heralds machines)
  )

(DEFUN report-cl-gabriel (machines &optional (reports '(gabriel-compare
						      realtime-compare
						      cpu-compare
						      paging-compare)))
  (DECLARE (SPECIAL *gabriel-cl-benchmarks*))
  (load-reports machines 'gabriel 'gabriel-fft)
;;(SETQ *gabriel-cl-benchmarks* (UNION *gabriel-cl-benchmarks* *gabriel-print*))
  (LET* ((sym-exp (LIST (FIRST machines) (CAR (LAST machines))))
	 (gabriel-cpu (sort-reports *gabriel-cl-benchmarks* sym-exp 'cpu 'min)))
    (sum-benchmarks *gabriel-cl-benchmarks* machines)
    (sum-benchmarks gabriel-cpu machines)
    (run-reports *gabriel-cl-benchmarks* machines '(gabriel-report)
						  reports
						  '(gabriel-compare3 realtime-compare3 paging-compare3))
    (run-reports gabriel-cpu machines '(gabriel-report)
				      reports
				      '(gabriel-compare3 realtime-compare3 paging-compare3)))
  (print-heralds machines)
  )
(DEFVAR *lisp-listener-general* '(ll-error ll-print-disk-label system-menu ll-ferror ll-ed-nofind ll-break
				  ll-hostat ll-eval-function ll-enter-function ll-time-function ll-disassemble
				  ll-envoke-dired ll-apropos ll-print-herald))
				  
(DEFVAR *zmacs-general* '(edit-meta-point-load-file edit-select-buffer edit-dired edit-unset-minor-mode
			  edit-new-file edit-dired2 edit-set-major-mode edit-set-minor-mode edit-kill-buffer
			  editor-break edit-list-methods edit-meta-point-change-buffer edit-help-apropos
			  edit-scroll-forward edit-small-modification edit-meta-point-same-buffer
			  edit-source-compare edit-mark-region edit-kill-region edit-enter-text
			  edit-yank-region edit-list-buffers edit-select-previous-buffer edit-scroll-backward
			  edit-source-compare-merge-text)) 
			  
(DEFVAR *create-window* '(select-lisp-create select-converse-create select-peek-create select-inspector-create
			  select-file-create select-vt100-create select-telnet-create select-editor-create)) 
					     
(DEFVAR *select-window* '(select-peek select-file select-vt100 select-inspector select-telnet select-lisp
			  select-converse select-editor)) 
			  
(DEFVAR *file-system* '(edit-local-completion edit-remote-completion edit-delete-remote-file edit-expunge-remote
			edit-copy-file-remote-to-local edit-find-remote-file edit-expunge-local
			edit-save-remote-file edit-copy-file-local edit-find-local-file edit-save-local-file
			edit-copy-file-remote edit-delete-local-file)) 
			
(DEFVAR *compiler* '(edit-compile-buffer edit-compile-remote-file edit-compile-local-file ll-compile-function))
  

(DEFUN report-interactive (machines &optional)
  (DECLARE (SPECIAL script *ia-benchmarks*))
  (load-reports machines 'script 'select-lisp)
    (sum-benchmarks *lisp-listener-general* machines)
    (sum-benchmarks *zmacs-general* machines)
    (sum-benchmarks *create-window* machines)
    (sum-benchmarks *select-window* machines)
    (sum-benchmarks *file-system* machines)
    (sum-benchmarks *compiler* machines)
    (setq *ia-benchmarks* (union *lisp-listener-general*	;Union list to Include benchmark sums
				   *zmacs-general*
				   *create-window*
				   *select-window*
				   *file-system*
				   *compiler*
				   ))
    (sum-benchmarks *ia-benchmarks* machines)
     (run-reports *ia-benchmarks* machines '(system-summary)
		 '(interactive-compare realtime-compare paging-compare)
		 '(interactive-compare3 paging-compare3))
     (run-reports *ia-benchmarks* machines '() '(cpu-compare) '(min-cpu-compare3))
  (print-heralds machines)
  )

(DEFUN old-report-interactive (machines &optional)
  (DECLARE (SPECIAL script))
  (load-reports machines 'script 'select-lisp)
  (LET* ((sym-exp (LIST (FIRST machines) (CAR (LAST machines))))
	 (sorted-time (sort-reports script sym-exp 'time 'mean))
	 (sorted-cpu (sort-reports script sym-exp 'cpu 'min)))
    (sum-benchmarks script machines)
    (sum-benchmarks sorted-time machines)
    (sum-benchmarks sorted-cpu machines)
    (run-reports sorted-time machines '(system-summary)
		 '(interactive-compare realtime-compare paging-compare)
		 '(interactive-compare3 paging-compare3))
    (run-reports sorted-cpu machines '() '(cpu-compare) '(min-cpu-compare3)))
  (print-heralds machines)
  )

(DEFUN report-all (machines &optional (PATHNAME "results:results;print.del"))
  (WITH-OPEN-FILE (*standard-output* pathname :direction :output)
    (FORMAT *terminal-io* "~%Printing benchmark results to ~a" (SEND *standard-output* :truename))
    (report-gabriel machines) (PRINC "." *terminal-io*)
    (report-instruction machines)  (PRINC "." *terminal-io*)
    (report-extensions machines) (PRINC "." *terminal-io*)
    (report-interactive machines)
    (print-heralds machines)
    ))

(DEFUN load-reports (machines name &optional benchmark)
  (UNLESS (CONSP machines) (SETQ machines (LIST machines)))
  (DOLIST (machine machines)
    (WHEN (OR (NOT (GET machine 'herald)) (NULL benchmark) (NOT (get-benchmark benchmark machine t)))
      (restore-results (STRING-APPEND "results:results;" machine "-" name ".bench") machine))))

(DEFVAR *short-report* t)

(DEFUN run-reports (benchmarks machines report1 &optional reports2 reports3)
  (LET ((benchmark-pagenum 0.))
    (UNLESS (CONSP machines) (SETQ machines (LIST machines)))
    ;; Compare all three machines
    (WHEN (= (LENGTH machines) 3)
      (dolist (report reports3)
	(report-benchmarks benchmarks (SYMBOL-VALUE report) machines))
      ;; Compare two machines at a time
      (UNLESS *short-report*
	(dolist (report reports2)
	  (report-benchmarks benchmarks (SYMBOL-VALUE report) (list (first machines) (SECOND machines))))
	(dolist (report reports2)
	  (report-benchmarks benchmarks (SYMBOL-VALUE report) (list (first machines) (third machines))))
	(dolist (report reports2)
	  (report-benchmarks benchmarks (SYMBOL-VALUE report) (list (second machines) (third machines))))
	))
    ;; Only two macines - compare them
    (WHEN (= (LENGTH machines) 2)
      (dolist (report reports2)
	(report-benchmarks benchmarks (SYMBOL-VALUE report) machines)))

    ;; Report on each machine
    (WHEN (= (LENGTH machines) 1)
      (dolist (report report1)
	(report-benchmarks benchmarks (symbol-value report) machines)))
    ))


(DEFUN print-heralds (machines)
  "print the herald for each machine in machines"
  
    (DOLIST (machine machines)
	(TERPRI)
	(Print (benchmark-herald machine 'user))
	(PRINT (benchmark-herald machine 'time))
	(PRINT (benchmark-herald machine 'herald)))
    )





