;;; -*- Mode:Common-Lisp; Package:User; Base:10; Fonts:(CPTFONT CPTFONTB) -*-
;
;

(defvar *pre-darkened-color*  :unbound
  "Remember the screen color before we darkened it.")
(defvar *idle-screen-blanker* nil
  "This is the process that automatically darkens the screen.")


(defun darken-screen ()
  "Darken the display screen.  No meaningful return."
  (setq *pre-darkened-color* tv:*current-screen-color*)
  (si:%Nubus-Write
    TV:Tv-Slot-Number
    SI:%Sib-Tv-Video-Attribute
    1))
 
(defun lighten-screen ()
  "Relights the display screen.  No meaningful return."
  (si:%Nubus-Write
    tv:Tv-Slot-Number
    SI:%Sib-Tv-Video-Attribute
    0)
  (when *pre-darkened-color*
    (tv:complement-bow-mode))
  (setq tv:*current-screen-color* *pre-darkened-color*))
 
(defun start-idle-screen-blanker (&optional (minutes 30))
  "Starts a process that will blank the screen if the keyboard is idle
   for MINUTES minutes.  This process will turn the screen back on 
   if any key is pressed. No meaningful return."
  (when *idle-screen-blanker*			;kill any existing blanker
    (send *idle-screen-blanker* :kill))
    (setq *idle-screen-blanker*			;create new blanker
	(process-run-function
	  '(:name "Idle Screen Blanker" :priority 10)	;priority is high to lighten screen fast
	  'blank-screen minutes )))

(defun kbd-idle-p (delay)
  "Returns t if keyboard has been idle for DELAY 60ths of a second."
  (> (time) (+  delay tv:kbd-last-activity-time)))

(defun blank-screen (minutes)
  "Main loop for automatic screen blanker.  Will darken the screen 
   if the keyboard is idle for more than MINUTES.  Will relight the 
   screen when any key is pressed."
  (do-forever
    (process-wait "Wait to Blank Screen" 'kbd-idle-p (* 3600 minutes))
    (darken-screen)
    (catch-all (send tv:selected-window :any-tyi))
    (lighten-screen)))

(tv:add-terminal-key #\Hyper-D '(user:darken-screen) "Darken Screen")
(tv:add-terminal-key #\Hyper-L '(user:lighten-screen) "Lighten Screen")
(tv:add-terminal-key #\Hyper-B '(user:start-idle-screen-blanker 30) "Blank Screen if Idle 30 Minutes")



