;;; -*- Mode: Common-Lisp; Package: User; Base: 10. -*-

;;  Copyright (C) 1988, Texas Instruments Incorporated. All rights reserved.

;;;   *-----------------------------------------------------------*
;;;   |			   COPY-NEWER-FILES			  |
;;;   *-----------------------------------------------------------*

;;  12/04/87 DNG - Original.
;;  12/22/87 DNG - Rewritten to avoid using unreliable FS:BALANCE-DIRECTORIES.
;;   1/09/88 DNG - Add support for copying sub-directories.
;;   5/26/88 CLM - Fixed problem in call to fs:directory-list (options to function
;;                 are keywords without following value).
;;   6.6.88  ab  - Moved this function to BASIC-FILE; DIRECTORY-SUPPORT.
;;                 make any further changes there!   File-Support patch 13.
;;                 (But put the version below in TICL package just in case anyone
;;                 loads it).


(defun ticl:copy-newer-files (from-directory to-directory &optional (query-p t))
  "For each file in FROM-DIRECTORY that is more recent than the corresponding 
file in TO-DIRECTORY, copy the file from FROM-DIRECTORY to TO-DIRECTORY.
This is particularly useful for updating copies on Unix or Macintosh where 
there aren't any version numbers."
  (let* ((from-path (pathname from-directory))
	 (to-path (pathname to-directory))
	 ;; Carfully avoid using pathname merging because it is too hard to
	 ;; get it to do what I want.
	 (from-dir (send from-path :new-pathname
			  :name (or (send from-path :name) :wild)
			  :type (or (send from-path :type) :wild)
			  :version :newest))
	 (to-dir (send (send to-path :new-pathname
			     :directory (or (send to-path :directory)
					    (send from-path :directory))
			     :name nil
			     :type nil
			     :version nil)
		       :translated-pathname))
	 (from-list (fs:directory-list from-dir))
	 (to-list (fs:directory-list (send to-dir :new-pathname
					   :name (or (send from-path :name) :wild)
					   :type (or (send from-path :canonical-type) :wild) ;ab 3/18/88
					   :version :newest)
				     :noerror ))	;ab 3/17/88
	 (files-to-copy '())
	 (directories-to-copy '())
	 (source-versions '?)
	 (destination-versions '?))
    (declare (list files-to-copy))
    (flet ((newer-pathname-p (from-file from-plist to-file to-plist)
	      ;; is from-file newer than to-file?
	      (let ((from-version (send from-file :version))
		    (to-version (send to-file :version)))
		(if (and (numberp from-version)
			 (numberp to-version))
		    (> from-version to-version)
		  (> (getf from-plist :creation-date 0)
		     (getf to-plist :creation-date 0))))))
      (dolist (from-info from-list)
	(let ((from-file (car from-info))
	      (from-plist (cdr from-info)))
	  (cond ((null from-file)) ; information about current directory, ignore.
		((getf from-plist :directory nil)	; a sub-directory
		 (let* ((from-name (send from-file :name))
			(sub-dir (send from-file :new-pathname
				       :directory (append (send from-file :directory)
							  (list from-name))
				       :name nil :type nil :version nil)))
		   (when (or (not query-p)
			     (y-or-n-p "Copy sub-directory \"~A\"?" sub-dir))
		     (push (list sub-dir
				 (send to-dir :new-pathname
				       :directory (append (send to-dir :directory)
							  (list from-name))
				       ))
			   directories-to-copy)) ))
		(t				; else a file.
		 (when (eq source-versions '?)
		   (setq source-versions (numberp (send from-file :version))))
		 (let* ((from-name (send from-file :name))
			(from-type (send from-file :canonical-type))
			(to-pathname (send to-dir :new-pathname
					   :name from-name
					   :canonical-type (send from-file :canonical-type)
					   :version (if (and source-versions destination-versions)
							(send from-file :version)
						      :newest)))
			(to-file nil)
			(to-plist nil)
			)
		   (dolist (to-info to-list)
		     (unless (null (car to-info))
		       (when (eq destination-versions '?)
			 (setq destination-versions (numberp (send (car to-info) :version))))
		       (when (and (string-equal from-name
						(send (car to-info) :name))
				  (equal from-type (send (car to-info) :canonical-type)))
			 (setq to-file (car to-info))
			 (setq to-plist (cdr to-info))
			 (return))))
		   (when (and (or (null to-file)
				  (newer-pathname-p from-file from-plist to-file to-plist))
			      (or (not query-p)
				  (y-or-n-p "Copy \"~A\" to \"~A\"?" from-file to-pathname)))
		     (push (list from-file to-pathname)
			   files-to-copy))))))))
    (setq directories-to-copy (nreverse directories-to-copy))
    (dolist (elt directories-to-copy)
      (CATCH-ERROR-RESTART (ERROR "Give up copying sub-directory \"~A\"" (first elt))
	(copy-newer-files (first elt) (second elt) query-p)
	))
    (setq files-to-copy (nreverse files-to-copy))
    (dolist (elt files-to-copy)
      (CATCH-ERROR-RESTART (ERROR "Give up copying \"~A\" and continue with the next file."
				  (first elt))
	(copy-file (first elt) (second elt) :report-stream t :create-directories t)	;ab 3/17/88
	))
    (values)))