;;; -*- Mode:Common-Lisp; Package:ZWEI; Base:10; Patch-file:T -*-
;;; Written 05/15/87 02:06:16 by buehring,
;;;
;;; Reason: Provide Filter Summary mode.  See SYS:PUBLIC;FILTER-SUMMARY.TEXT for documentation.
;;;
;;; while running on Spud from band LOD9
;;; With IO 3.0, PATHNAME 3.0, FILE 3.0, SYSTEM 3.0, METER 3.0, SERIAL 3.0, CHAOSNET 3.0, ETHERNET 3.0, GC 3.0, IMAGEN 3.0, MAIL-DAEMON 3.0, NETWORK-SUPPORT 3.0, PROFILE 3.0, SUGGESTIONS 3.0, UCL 3.0, ZWEI 3.0, STREAMER-TAPE 3.0, DEBUG-TOOLS 3.0, FONT-EDITOR 3.0, GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.0, NAMESPACE-EDITOR 3.0, NVRAM 3.0, TELNET 3.0, TV 3.0, NAMESPACE 3.0, COMPILER 3.0, PRINTER 3.0, SYSLOG 3.0, VT100 3.0, Experimental IP 2.1, Experimental BUG 4.0,  microcode 207, Band Name: REL3.0+IP+BUG,U207.



#!C
; From file FILTER-SUMMARY.LISP V4UD:[BUEHRING.BANDY] Vaxen:
#10R ZWEI#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "ZWEI"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "sys: MAIL-READER; FILTER-SUMMARY.#"


(defvar *FILTER-SUMMARY-FONTS* '(:cptfont :cptfont :cptfontb))


(defvar *EMPTY-MESSAGE-SEQUENCE* (make-instance 'message-sequence :name "*Empty-Message-Sequence*"))


(defmethod (filter-summary-buffer :after :init) (&rest ignore)
  (set-buffer-fonts self *filter-summary-fonts*)
  (push (list 'filter-summary-mode) saved-mode-list))


(defminor COM-FILTER-SUMMARY-MODE FILTER-SUMMARY-MODE "Filter-Summary" 1
	  "Minor mode for editing a DIRED style summary of mail filters." ()
    (set-comtab *read-mail-comtab*
	      '(#\S com-expand-filter
		#\N com-mail-down-real-line
		#\P com-mail-up-real-line)
	      '(("Delete Filters" . com-delete-filters)
		("Create Filter" . com-filter-messages)
		("Create Filter From Keywords" . com-create-filter-from-keywords)
		("Filtered Mode" . com-filter-summary-mode))))

(set-comtab *read-mail-comtab* nil '(("Filtered Mode" . com-filter-summary-mode)))
  

(defcom COM-EXPAND-FILTER
	"Expand or Unexpand the filter at point."
	()
  (let* ((line (bp-line (point)))
	 (line-message (getf (line-plist line) :summary-message))
	 (line-filter (getf (line-plist line) :filter-summary-filter))
	 filter expand-p)
    (cond ((not (filter-summary-p *interval*))
	   (barf "Not within a filter summary buffer."))
	  (line-message
	   ;; On a message line (of an expanded filter).  Unexpand it.
	   (let ((sequence (getf (line-plist line) :summary-sequence)))
	     (setq filter (get sequence :filter-summary-filter))))
	  ((and line-filter
		(get line-filter :expanded-p))
	   (setq filter line-filter))
	  (line-filter
	   ;; On a line with a filter name and it's unexpanded.  Expand it.
	   (setq filter line-filter)
	   (setq expand-p t))
	  (t
	   (barf "Line does not describe message or filter.")))
    (let ((sequence (get filter :message-sequence))
	  (filter-line (get filter :filter-summary-line)))
      (when (and sequence filter filter-line)
	(setf (get filter :expanded-p) expand-p)
	(with-read-only-suppressed (*interval*)
	  (change-font-interval (create-bp filter-line 0)
				(create-bp filter-line (length filter-line))
				t (if expand-p 2 1)))
	(update-filter-summary *interval* sequence)
	(when (and (eq (get filter :expanded-p) t)
		   (eql (total-messages sequence) 0))
	  (format *query-io* "No messages pass this filter.")))))
  dis-text)

(defun MAKE-FILTER-SUMMARY (mail-file-buffer name)
  
  (let ((summary (make-instance 'filter-summary-buffer :name name :sequence-buffer mail-file-buffer)))
    
    (with-read-only-suppressed (summary)
      (insert-summary-heading (interval-last-bp summary))
      (insert-line-with-leader (make-diagram-line 'mail-summary-black-line-diagram)
			       (bp-line (interval-last-bp summary))))
    
    (let ((filter (make-instance 'system-mail-filter
				 :name "ALL MESSAGES"
				 :function 'messagep
				 :string-for-filter-summary "ALL MESSAGES")))
      (add-filter-summary-filter filter summary))
    (dolist (filter-item (get-mail-option mail-file-buffer :filter-alist))
      (let ((filter nil))
	(when (fboundp (second filter-item))
	  (setq filter (make-instance 'system-mail-filter
				      :name (first filter-item)
				      :string-for-filter-summary (first filter-item)
				      :function (second filter-item)
				      :args (cddr filter-item)))
	  (when filter
	    (add-filter-summary-filter filter summary)))))
    summary))


(defun ADD-FILTER-SUMMARY-FILTER (filter filter-summary)
  
  (let ((mail-file-buffer (mail-file-buffer-of filter-summary)))
    (cond ((equalp (send filter :name) "ALL MESSAGES")
	   (setf (get mail-file-buffer :filter-summary-filter) filter)
	   (setf (get filter :message-sequence) mail-file-buffer)
	   (push filter (buffer-filter-list filter-summary)))
	  (t
	   (let ((seq (make-mail-buffer-from-filter filter mail-file-buffer nil)))
	     (setf (get seq :filter-summary-filter) filter)
	     (setf (get filter :message-sequence) seq)
	     (setf (buffer-summary-buffer seq) filter-summary))
	   (push-end filter (buffer-filter-list filter-summary))))
    
    (with-read-only-suppressed (filter-summary)
      (let ((filter-string (send filter :string-for-filter-summary))
	    line)
	(when filter-string
	  (setq line (create-line 'art-string (+ (length filter-string) 4) filter-summary))
	  (setf (line-length line) 0)
	  (append-to-line line "  ")
	  (append-to-line line filter-string)
	  (splice-line-into-interval line (bp-line (interval-last-bp filter-summary)))
	  (change-font-interval (create-bp line 0)
				(create-bp line (length line))
				t 1)
	  (mung-line line)
	  (setf (get filter :filter-summary-line) line)
	  (setf (getf (line-plist line) :filter-summary-filter) filter))))))



(defun UPDATE-FILTER-SUMMARY (summary &optional sequence)
  
  (cond (sequence
	 (let* ((filter (get sequence :filter-summary-filter))
		(summary-line (get filter :filter-summary-line)))
	   (when (and filter summary-line)
	     (cond ((get filter :expanded-p)
		    (unless (mail-file-buffer-p sequence)
		      (revert-filter-buffer sequence nil))
		    (update-summary summary sequence (create-bp (line-next summary-line) 0)))
		   (t
		    (update-summary summary *empty-message-sequence* (create-bp (line-next summary-line) 0)))))))
	;; Update all expanded filters
	(t
	 (loop
	   for filter in (buffer-filter-list summary)
	   for filter-buffer = (get filter :message-sequence)
	   for summary-line = (get filter :filter-summary-line)
	   when (and filter-buffer summary-line)
	   do
	   (cond ((get filter :expanded-p)
		  (unless (mail-file-buffer-p filter-buffer)
		    (revert-filter-buffer filter-buffer nil))
		  (update-summary summary filter-buffer (create-bp (line-next summary-line) 0)))
		 (t 
		  (update-summary summary *empty-message-sequence* (create-bp (line-next summary-line) 0))))))))


(defun REMINDER-MESSAGES-FILTER (msg)
  (msg-reminder-p msg))
  
(defun UNSEEN-MESSAGES-FILTER (msg)
  (message-attribute-p :unseen msg))

(defun RECENT-MESSAGES-FILTER (msg)
  (message-attribute-p :recent msg))

(defun DELETED-MESSAGES-FILTER (msg) 
  (message-attribute-p :deleted msg))

(defun NOT-ANSWERED-MESSAGES-FILTER (msg) 
  (not (message-attribute-p :answered msg)))

(defun KEYWORD-FILTER-FUNCTION (msg keywords)
  (keyword-and-filter msg keywords))


;;; From MAIL-READER;SUMMARY

(defun MAKE-SUMMARY (sequence)
  
  (when (and (boundp '*window*) *window* (send *window* :exposed-p))
    (format *query-io* "~&Creating summary."))
  (let* ((name (string-append (buffer-name sequence) " S"))
	 (summary (find-buffer-named name))
	 (new-p (not summary))
	 line)
    (cond (summary
	   (update-summary summary sequence)
	   summary)
	  ;; new cond clause for filter summary
	  ((and (eq *mail-summary-mode* :filtered)
		(mail-file-buffer-p sequence))
	   (setq summary (make-filter-summary sequence name)))
	  (t
	   (setq summary (make-instance 'mail-summary-buffer :name (string-append (buffer-name sequence) " S")))
	   (with-read-only-suppressed (summary)
	     (insert-summary-heading (interval-last-bp summary))
	     (insert-line-with-leader (make-diagram-line 'mail-summary-black-line-diagram)
				      (bp-line (interval-last-bp summary)))
	     (domsgs (msg sequence index)
	       (setq line (make-summary-line msg sequence))
	       (setf (line-node line) summary)
	       (insert-line-with-leader line (bp-line (interval-last-bp summary)))))))
    (when new-p
      (send summary :activate)
      (setf (buffer-sequence-buffer summary) sequence)
      (setf (buffer-summary-buffer sequence) summary)
      (setf (buffer-pathname summary) (buffer-pathname sequence))
      (setf (buffer-file-id summary) nil)
      (setf (buffer-version-string summary) (buffer-version-string sequence)))
    summary))


;;; From MAIL-READER;SUMMARY

(defun UPDATE-SUMMARY (summary &optional sequence start-bp stop-bp)
  "Update SUMMARY to reflect the current contents of SEQUENCE.
 If START-BP and STOP-BP supplied, only update summary lines within the
defined interval, otherwise do the whole buffer.  Update message numbers
and insert/delete summary lines for messages that have been
added/removed.  Other summary info must already be correct."
  
  ;; New check for filter summary
  (when (and (filter-summary-p summary)
	     (null start-bp))
    (return-from update-summary (update-filter-summary summary)))

  (unless sequence
    (setq sequence (message-sequence-of summary)))

  (let ((*interval* summary)			;needed because delete-interval and beg-line may reference *interval*
	(*undo-save-small-changes* nil))
    (with-read-only-suppressed (summary)
      (loop
	with stop-line = (if stop-bp (bp-line stop-bp) (bp-line (interval-last-bp summary)))
	with line = (if start-bp (bp-line start-bp) (first-message-summary-line summary))
	with bp = (create-bp line 0)
	with index = 0
	with len = (length (buffer-message-array sequence))
	for msg = (and (< index len) (aref (buffer-message-array sequence) index))
	for line-msg = (getf (line-plist line) :summary-message)
	;; Check for hitting next filter in filter summary
	unless line-msg do (setq stop-line line)
	while (or (< index len) (neq line stop-line))
	do
	(cond
	  ((and (neq line stop-line) msg line-msg (eq msg line-msg))
	   ;; Summary line's message matches message in sequence... ensure message number is correct.
	   (unless (eql (1+ index) (getf (line-plist line) :summary-message-number))
	     (make-summary-line msg sequence line))	;update existing line to reflect correct message number
	   ;; Advance both summary line and sequence
	   (setq line (or (line-next line) stop-line))
	   (incf index))
	  ((and msg
		(or (lisp:find line-msg (buffer-message-array sequence) :test #'eq :start index)
		    (eq line stop-line)))
	   ;; This summary line's message exists further down or we've exausted the summary lines but not the sequence...
	   ;; Insert new summary for this message
	   (let ((new-line (allocate-summary-line t)))
	     (splice-line-into-interval (make-summary-line msg sequence new-line) line)
	     (setf (line-node new-line) summary))
	   (incf index))
	  ((neq line stop-line)
	   ;; This summary line's message is gone from sequence or we've exausted the sequence but not the summary lines...
	   ;; remove the summary line.
	   (move-bp bp line 0)			;position bp at line to delete
	   (setq line (or (line-next line) stop-line))	;advance to next line before deleting
	   (remove-summary-line bp))
	  (t
	   ;; Already covered all the bases... shouldn't get here.
	   (return))))))
  (must-redisplay-buffer summary dis-text))


;;; From MAIL-READER;COMMAND




(defcom COM-FILTER-MESSAGES
	"Filter messages with specified characteristics into a separate buffer.
With a prefix argument filter only the messages in the current buffer (e.g.  to apply a
filter to a filter buffer), otherwise filter all messages in current mail file."
	()
  (in-mail-context (:require-buffer t)
    (multiple-value-bind (filter current-only)
	(if (or *mail-mouse-command*
		;; Check if invoked from suggestions
		(invoked-from-suggestions-p))
	    (choose-mail-filter)
	  (read-mail-filter-from-mini-buffer))
      (when filter
	(cond ((filter-summary-p (mail-summary-of (mail-file-buffer-of *mail-buffer*)))
	       (add-new-filter-summary-filter filter (mail-file-buffer-of *mail-buffer*))
	       ;; Prevent sync-ing mail buffers so cursor remains on line with new filter.
	       (return-from com-filter-messages dis-text))
	      (t
	       (let ((buffer (make-mail-buffer-from-filter
			       filter
			       (if current-only
				   *mail-buffer*
				 (mail-file-buffer-of *mail-buffer*)))))
		 (cond ((or (> (total-messages buffer) 0)
			    (filter-summary-p (mail-summary-of buffer)))
			(send buffer :activate)
			(when (mail-summary-of buffer)
			  (update-summary (mail-summary-of buffer)))
			(make-mail-buffer-current buffer)
			(format *query-io* "~&Press ~A to return to all messages."
				(key-for-command 'com-exit-mail-reader *comtab* nil nil #\Q)))
		       (t
			(send buffer :kill)
			(barf "No messages pass this filter."))))))))
    (values dis-text dis-text)))


(defun ADD-NEW-FILTER-SUMMARY-FILTER (filter mail-file-buffer)
  
  (let ((filter-summary (mail-summary-of mail-file-buffer))
	(filter-name (send filter :string-for-filter-summary)))
    
    (when (and (eq *mail-summary-mode* :filtered)
	       (filter-summary-p filter-summary)
	       (not (member filter-name (get-mail-option mail-file-buffer :filter-alist)
			    :test #'equalp :key #'car)))
      (when (y-or-n-p "Do you want this filter stored in the mail file?")
	(let* ((*mini-buffer-default-string* filter-name)
	       (string (typein-line-readline "Enter name of filter (Default is ~S)" filter-name)))
	  (when (not (zerop (length string)))
	    (setq filter-name string)))
	
	(push-end (append (list filter-name (send filter :function)) (send filter :args))
		  (get-mail-option mail-file-buffer :filter-alist))
	(mung-node mail-file-buffer))
      
      ;; Use a distinct filter to stick add to the filter summary
      (let ((new-filter (make-instance 'system-mail-filter
				       :name (send filter :name)
				       :function (send filter :function)
				       :args (send filter :args)
				       :string-for-filter-summary filter-name)))
	(add-filter-summary-filter new-filter filter-summary)
      (when (eq *interval* filter-summary)
	(move-bp (point) (create-bp (get new-filter :filter-summary-line) 0)))))))


(defun SUMMARY-MAKE-MESSAGE-CURRENT (message summary-buffer &optional sequence-buffer)
  "Update SUMMARY-BUFFER to point at MESSAGE and return the updated bp
or NIL if MESSAGE not found.  If SEQUENCE-BUFFER is supplied, ensure
that the summary line selected is from that sequence (for Dired style
summaries)."
  
  (when (eq summary-buffer *interval*)
    (move-bp (buffer-saved-point summary-buffer) (point)))
  (let ((bp (buffer-saved-point summary-buffer)))
    
    (cond ((summary-line-message-equal (bp-line bp) message sequence-buffer)
	   ;; no need to move, just return current bp
	   bp)
	  (t
	   (do ((line (bp-line (interval-first-bp summary-buffer)) (line-next line)))
	       ((null line) nil)
	     
	     (when (summary-line-message-equal line message sequence-buffer)
	       (move-bp bp line
			(min (bp-index bp) (line-length line)))	;Stay in same column if possible
	       (return bp)))))))



(defcom COM-MAIL-DOWN-REAL-LINE
	"Move down vertically to next real line, keeping summary and message buffers in sync.
Moves as far as possible horizontally toward the goal column for successive
commands.  The goal column is normally the column you start at,
but the command C-X C-N sets a semipermanent goal column."
	(km r)
  (in-mail-context ()
    (down-real-line *numeric-arg*)
    (when (mail-summary-p *interval*)
      ;; Inside summary -- sync message sequence to new summary point.  
      (multiple-value-setq (*mail-buffer* *msg*) (sync-mail-buffers nil))
      (when (not (getf (line-plist (bp-line (point))) :summary-message))
	(return-from com-mail-down-real-line dis-bps)))
    (values dis-bps dis-bps)))


(defcom COM-MAIL-UP-REAL-LINE
	"Move up vertically to previous real line, keeping summary and message buffers in sync.
Moves as far as possible horizontally toward the goal column for successive
commands.  The goal column is normally the column you start at,
but the command C-X C-N sets a semipermanent goal column."
	(km -r)
  (down-real-line (- *numeric-arg*))
  (when (mail-summary-p *interval*)
    (when (getf (line-plist (bp-line (point))) :summary-message)
      (multiple-value-setq (*mail-buffer* *msg*) (sync-mail-buffers nil))
      (must-redisplay-mail-buffer *mail-buffer* dis-text dis-bps)))
  dis-bps)



(defcom COM-MAIL-DOWN-REAL-LINE
	"Move down vertically to next real line, keeping summary and message buffers in sync.
Moves as far as possible horizontally toward the goal column for successive
commands.  The goal column is normally the column you start at,
but the command C-X C-N sets a semipermanent goal column."
	(km r)
  (down-real-line *numeric-arg*)
  (when (mail-summary-p *interval*)
    (when (getf (line-plist (bp-line (point))) :summary-message)
      (multiple-value-setq (*mail-buffer* *msg*) (sync-mail-buffers nil))
      (must-redisplay-mail-buffer *mail-buffer* dis-text dis-bps)))
  dis-bps)



(defcom COM-MOUSE-SELECT-OR-MARK
	"Select message under mouse or do mouse region marking."
	()
  (com-mouse-mark-region)
  (when (mail-summary-p *interval*)
    (when (getf (line-plist (bp-line (point))) :summary-message)
      (multiple-value-setq (*mail-buffer* *msg*) (sync-mail-buffers nil))
      (must-redisplay-mail-buffer *mail-buffer* dis-text dis-bps)))
  dis-bps)


(defcom COM-DELETE-FILTERS
	"Choose filters to delete from this filter summary."
	()
  
  (in-mail-context (:require-buffer t)
    (let* ((mail-file-buffer (mail-file-buffer-of *mail-buffer*))
	   (summary (mail-summary-of mail-file-buffer))
	   item-list)
      (loop
	for filter in (buffer-filter-list summary)
	for name-string = (send filter :string-for-filter-summary)
	unless (equalp (send filter :name) "ALL MESSAGES")
	do
	(push (cons name-string filter) item-list))
      (setq item-list (nreverse item-list))
      (multiple-value-bind (nuke-list do-it-p)
	  (w:multiple-menu-choose item-list :label "Choose Filters to Delete")
	(when do-it-p
	  (dolist (filter nuke-list)
	    (setf (get filter :expanded-p) nil))
	  (update-filter-summary summary)
	  (with-read-only-suppressed (summary)
	    (loop
	      for filter in nuke-list
	      for line = (get filter :filter-summary-line)
	      do
	      (delete-interval (create-bp line 0) (create-bp (line-next line) 0))
	      (setf (buffer-filter-list summary) (deleq filter (buffer-filter-list summary)))
	      (setf (get-mail-option mail-file-buffer :filter-alist)
		    (delete (send filter :string-for-filter-summary)
			    (get-mail-option mail-file-buffer :filter-alist)
			    :key #'car :test #'equalp))))))
      (mung-node mail-file-buffer))
    (values dis-text dis-text)))


(defcom COM-FILTER-SUMMARY-MODE
	"Toggle Filter Summary mode."
	()
  (in-mail-context (:require-buffer t)
    (let* ((summary (mail-summary-of *mail-buffer*))
	   (current-mode (if summary
			     (if (filter-summary-p summary)
				 :filtered :basic)
			   *mail-summary-mode*)))
      (setq *mail-summary-mode* (if (eq current-mode :filtered)
				    :basic :filtered))
      (com-revert-mail-summary))))


(defcom COM-REVERT-MAIL-SUMMARY
        "Recreate the summary of the current mail buffer from scratch.
Use after modifying option variables which affect the summary display or if the 
summary becomes mangled."
	()
  (in-mail-context (:require-buffer t)
    (let ((old-summary (mail-summary-of *mail-buffer*)))
      (when old-summary
	(send old-summary :kill))
      (make-mail-buffer-current *mail-buffer* :summary))
    (values dis-none dis-text)))


(defcom COM-EXIT-MAIL-READER
	"Return to mail file buffer or default viewing mode or completely exit mail reader."
	()
  (in-mail-context (:require-buffer t)
    (when (and *msg* (mail-buffer-window *mail-buffer*))
      (delete-message-attribute :unseen *msg*))
    (cond ((two-mail-reader-windows-p *mail-buffer*)
	   (cond ((mail-file-buffer-p *mail-buffer*)
		  (setf (get *mail-buffer* :prev-mail-window-configuration) :both)
		  (exit-mail-reader t))
		 (t
		  (make-mail-buffer-current (mail-file-buffer-of *mail-buffer*) :both))))
	  ((and (mail-summary-p *interval*)
		(eq (default-mail-window-configuration) :message))
	   (make-mail-buffer-current *mail-buffer* :message))
	  ((and (message-sequence-p *interval*)
		(eq (default-mail-window-configuration) :summary))
	   (make-mail-buffer-current *mail-buffer* :summary))
	  ((and (not (mail-file-buffer-p *mail-buffer*))
		(not (filter-summary-p *interval*)))
	   (make-mail-buffer-current (mail-file-buffer-of *mail-buffer*)))
	  (t
	   (setf (get *mail-buffer* :prev-mail-window-configuration)
		 (if (message-sequence-p *interval*)
		     :message :summary))
	   (exit-mail-reader)))
    
    (values dis-text dis-text)))



(defun INVOKED-FROM-SUGGESTIONS-P ()
  (and (type-specifier-p 'sugg:suggestions-command-menu-mixin)
       (typep (w:window-owning-mouse) 'sugg:suggestions-command-menu-mixin)))
))


