;;;                           RESTRICTED RIGHTS LEGEND

;;;     Use, duplication, or disclosure by the Government is subject to
;;;   restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;   Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                        TEXAS INSTRUMENTS INCORPORATED.
;;;                                 P.O. BOX 2909
;;;                              AUSTIN, TEXAS 78769-2909
;;;                                    MS 2223
;;;
;;;   Copyright (C) 1987, Texas Instruments Incorporated. All rights reserved.

;-------------------------------------------------------------------------------------------------------------------------
;
; WARNING: This is not a Common Lisp program.  It is a PCScheme program meant to be used by a Personal Consultant Plus
;          Explorer version user who is porting his knowledge base to a personal computer.  It is to be moved to a
;          personal computer and run under PCScheme.
;
;-------------------------------------------------------------------------------------------------------------------------

; Function name: CNVERT
;
; Function call: (CNVERT kb-name)
;
;   Where kb-name is a string or symbol representing the name of the knowledge
;   base.  This name should not contain a file extension or file pathname;
;
; Language:    Scheme running on a Personal Computer.
;
; Description: Replaces the symbols T and NIL in all PCPlus Explorer kb files ported to a personal computer,
;              with #!TRUE and ().	The files must be in the current directory when CNVERT
;	       is run.	This takes the place of doing the same thing using
;	       the ATOM QUERY REPLACE extended command in Zmacs.  See section
;	       C.5.4 of the Personal Consultant Plus Explorer "Users Guide".
;
;              New files are created with unique extensions .FB, .F1, etc. these
;              files must be renamed to *.K*, before trying to load them into PCPlus
;              on the PC.
;
;              This function expects all kb files to be present.  It will stop if no .KB file is present or with the
;              first .Kn file not found. 
;
;              Use the function FILTER-T-NIL to convert a single kb file. 
;
; Warning:     If you have user defined functions developed on the Explorer, load
;	       the kb  into Personal Consultant Plus on the PC, and
;	       use the ERASE command to delete them.  At this point, the new kb
;	       should probably be saved off, before continuing development on
;	       the Scheme user defined functions.
;
;------------------------------------------------------------------------------

(define cnvert
  (lambda (kb)
    (let* ((inext ".k")
           (outext ".f")
           (kbname
             (if (symbol? kb)
                 (symbol->string kb)
                 kb))
           (infile (string-append kbname inext "b"))
           (outfile (string-append kbname outext "b")))
      (if (not (file-exists? infile))
          (print (string-append "Knowledge base " kbname " does not exist."))
          (begin
            (print
              (string-append
                "Converting knowledge base " kbname "."))
            (print (string-append infile " -> " outfile))
            (filter-t-nil infile outfile)
            (do ((frame 2 (1+ frame))
                 (infile (string-append kbname ".k1") (string-append kbname inext (integer->string frame 10)))
                 (outfile (string-append kbname ".f1") (string-append kbname outext (integer->string frame 10))))
                ((not (file-exists? infile))
                 (print "Knowledge base conversion complete.")             
                 (newline)
                 (print "Remember that .F? files must be copied to .K? files before the knowledge base can be run."))
                (print (string-append infile " -> " outfile))
                (filter-t-nil infile outfile)))))))


(define filter-t-nil
;
;  Replaces the symbols T and NIL in a file, creating a new file.
;
;  Call: (FILTER-T-NIL "input file" "output file")
;
;  Where "input file" is a .KB, .K1, .K2, file etc., developed on the Explorer;
;	 "output file" is a corresponding kb file to be run on the PC.
;
; The input and output files must be specified with extensions.  Remember that the knowledge base remembers the
; file name it was created in, and so the output file name must be the same and the output file extension
; must be different.  You might use a .F? extension.  This file will have to be rename to a .K* file before
; it can be loaded into PCPlus on the PC.
;
  (lambda (input-file output-file)
    (let ((input-port (open-input-file  input-file))
          (output-port (open-output-file output-file))
          (object nil))
      (set! object (read input-port))
      (do () ((eof-object? object))
            (write (f-t-nil object) output-port)
            (set! object (read input-port))
            )
      (close-input-port input-port)
      (close-output-port output-port)
)))

(define f-t-nil
;
;  Replaces the symbols T and NIL in an expression.
;
;  Call: (F-T-NIL expression)
;
  (lambda (object)
    (append (if (pair? object)
                          (mapcar f-t-nil object)
                          (cond ((eq? object 't) #!true)
                                ((eq? object 'nil) (list))
                                (t object))))))
