;;; -*- Mode:Common-Lisp; Package:DEMO; Base:10 -*-

(defvar maxvel 5000)
(defvar elastic 800)				; elasticity coefficient
(defvar ball-array (make-array 100))            ; The universe of balls.
(defvar number-balls 0)                         ; current number of balls.
(defvar ball-image)				; bitblt array
(defvar ball-temp-image)			; temporary bitblt array
(proclaim '(special MinX MaxX MinY MaxY))	; window dimensions, actual
(proclaim '(special VMaxX VMaxY))		; window dimensions, virtual

;;; Macros to translate between internal (virtual) coordinates and
;;;   screen coordinates:

(defconstant virt-coordinates-scale-factor 256)
(defmacro virt-to-screen(q) `(round ,q virt-coordinates-scale-factor))
(defmacro screen-to-virt(q) `(round (* ,q virt-coordinates-scale-factor)))

(defconstant *ball-radius* 16)			; radius of ball
(defconstant *ball-diameter* (* 2 *ball-radius*)); diameter
(defconstant *ball-v-diameter* (screen-to-virt (* 2 *ball-radius*))); diameter
(defconstant *ball-dsq* (* (screen-to-virt *ball-diameter*)
			   (screen-to-virt *ball-diameter*)))
(defconstant *mini-raster-size* 48)		; size of miniraster
(defconstant *mini-raster-offset*		; distance to miniraster center
	     (round *mini-raster-size* 2))
(defconstant *mini-raster-slack*
	     (- *mini-raster-offset* *ball-radius*))

;;; Initialize window, various globals.

(defun SetupCrapolla()				; in honor of you-know-who
  
  (setq ball-array (make-array 100))
  (setq number-balls 0)
  (multiple-value-setq (MinX MinY MaxX MaxY) (send tv:selected-window ':edges))
  (setq MaxY (- MaxY 16))			; FUDGE: account for window label.
  (setq MaxX (- MaxX 8))			; FUDGE: who knows why.
  
  ;;; Virtual coordinates q range:  0 <= q < VMaxQ...
  (setq VMaxX (screen-to-virt (- MaxX MinX *mini-raster-size*)))
  (setq VMaxY (screen-to-virt (- MaxY MinY *mini-raster-size*)))
  
  (send tv:selected-window :clear-screen)
  
  ;;;  Build the ball image mini-raster directly on the screen:
  (setq ball-image (tv:make-sheet-bit-array tv:selected-window
					    (* 2 *mini-raster-offset*) (* 2 *mini-raster-offset*)))
  (setq ball-temp-image (tv:make-sheet-bit-array tv:selected-window
						 (* 2 *mini-raster-offset*) (* 2 *mini-raster-offset*)))
  (send tv:selected-window :draw-filled-circle *mini-raster-offset* *mini-raster-offset* *Ball-radius* tv:alu-xor)
  (send tv:selected-window :bitblt-from-sheet tv:alu-seta
	(* 2 *mini-raster-offset*) (* 2 *mini-raster-offset*)
	0 0 ball-image 0 0)
  
  ;;; Now paint the initial screen:
  (send tv:selected-window :clear-screen)
  (do ((i 0 (+ 1 i)))
      ((= i 5))
    (let ((dd (- *mini-raster-offset* *ball-radius* i i)))
      (send tv:selected-window :draw-polyline	;tv:alu-xor
	    (coerce (list (+ MinX dd) 
			  (+ MinX dd) 
			  (- MaxX dd) 
			  (- MaxX dd) 
			  (+ MinX dd)) '(array integer 1)) 
	    (coerce (list (+ MinY dd)
			  (- MaxY dd)
			  (- MaxY dd)
			  (+ MinY dd)
			  (+ MinY dd)) '(array integer 1)))))
  )

(defflavor Ball ((x-position 100)		; Positions of CENTER of ball...
		 (y-position 100)		;   ... in virt coordinates.
		 (x-velocity 0)			; 1st differences, virt coords.
		 (y-velocity 0)
		 ball-number
		 (OnScreen nil))		; T iff image on screen.
	   ()
  :gettable-instance-variables
  :settable-instance-variables
  #+3600
  :initable-instance-variables
  #+Explorer
  :inittable-instance-variables
  )

;;; XOR a ball on the screen at center position x, y (screen)
(defun XOR-ball(x y)
  (send tv:selected-window :bitblt tv:alu-xor
		*mini-raster-size* *mini-raster-size* ball-image 0 0 x y))

(defun MOVE-ball(x y dx dy)
  (bitblt tv:alu-seta *mini-raster-size* *mini-raster-size*
	  ball-image 0 0
	  ball-temp-image 0 0)
  (bitblt tv:alu-xor *ball-diameter* *ball-diameter*
	  ball-image *mini-raster-slack* *mini-raster-slack*
	  ball-temp-image
	  (+ *mini-raster-slack* dx)
	  (+ *mini-raster-slack* dy) )

  (send tv:selected-window :bitblt tv:alu-xor
		*mini-raster-size* *mini-raster-size* ball-temp-image 0 0
		(virt-to-screen x)
		(virt-to-screen y)))


(defun dist(b1 b2)
     (let ((dx (- (send b1 :x-position)
		  (send b2 :x-position)))
	   (dy (- (send b1 :y-position)
		  (send b2 :y-position))))
       (isqrt (+ (* dx dx) (* dy dy)))))

(defmethod (ball :move-to)(vx vy)
  (let ((dx (- vx x-position))
	(dy (- vy y-position))
	(x (virt-to-screen vx))
	(y (virt-to-screen vy)))
    (if OnScreen
	(if (and (<= (abs dx) *mini-raster-slack*)
		 (<= (abs dy) *mini-raster-slack*))
	    (move-ball x y dx dy)
	    (progn (xor-ball (virt-to-screen x-position) (virt-to-screen y-position))
		   (xor-ball x y)))
	(xor-ball x y))
    (setq x-position vx)
    (setq y-position vy))
  (setq OnScreen t))

;;; Apply a force vector to a ball:
(defmethod (ball :force)(dx dy)				; The force vector.
 	   (setq x-velocity (+ (round (* elastic dx) 1024) x-velocity))
	   (setq y-velocity (+ (round (* elastic dy) 1024) y-velocity)))

;;; Generate a force (DDQ) for bouncing:
(defun bounce-force(displacement) (* displacement 1))

;;; Update a coordinate:
;;; Adds in velocity after checking for wall bounce.

(defmacro update-ball-coord(Q DQ QMAX)`(progn
	   (if (< ,Q 0)(setq ,DQ (+ (bounce-force (- ,Q)) ,DQ)))
	   (if (>= ,Q ,QMAX)(setq ,DQ (- ,DQ (bounce-force (- ,Q ,QMAX)))))
	   (if (> ,DQ maxvel)(setq ,DQ maxvel))
	   (if (< ,DQ (- maxvel))(setq ,DQ (- maxvel)))
	   (setq ,Q (+ ,Q ,DQ)) ))

;;; Update the position of a ball:
(defmethod (ball :update)
	   (&aux (oldx x-position)(oldy y-position))
;  (send self :dump)
	   (update-ball-coord oldx x-velocity VMaxX)
	   (update-ball-coord oldy y-velocity VMaxY)
	   (send self :move-to oldx oldy))

(defmethod (ball :dump) ()
;  (send tv:selected-window :set-cursorpos 0 (* 16 (send self :ball-number)))
  (format tv:selected-window "~& ~d: ~s: xy = (~d, ~d);  vxy = (~d, ~d)    "
	  (send self :ball-number)
	  self
	  (send self :x-position)
	  (send self :y-position)
	  (send self :x-velocity)
	  (send self :y-velocity)))

;;; Handle ball-ball collisions:
(defun collide(b1 b2)
  (let ((dx (- (send b1 :x-position)
	       (send b2 :x-position)))
	(dy (- (send b1 :y-position)
	       (send b2 :y-position)))
	dsq ddsq)
    
    (setq dsq (+ (* dx dx) (* dy dy)))
    
    (if (< dsq *ball-dsq*)
	(progn (setq dsq (isqrt dsq))
	       (setq ddsq (- *ball-v-diameter* dsq))
	       (setq dx (round (* ddsq dx) dsq))
	       (setq dy (round (* ddsq dy) dsq))
;	       (print (list dx dy dsq))
	       (send b1 :force dx dy)
	       (send b2 :force (- dx)(- dy))
	       ))))

;;; Update the world:
(defun update-step()
	   (do* ((i 0 (+ 1 i))
		 (b1 nil))
		((>= i number-balls))
		(send (setq b1 (aref ball-array i)) :update)
		(do ((j (+ 1 i) (+ 1 j)))
		    ((>= j number-balls))
		    (collide b1 (aref ball-array j)) )))

;; Make a new ball:
(defun new-ball(x y &optional (vx 0) (vy 0))
  (setf (aref ball-array number-balls)
        (make-instance 'Ball :x-position x :y-position y :x-velocity vx :y-velocity vy))
  (send (aref ball-array number-balls) :set-ball-number number-balls)
  (setq number-balls (+ 1 number-balls)))

;;; Make a random ball:
(defun rand-ball(&aux (x (random VMaxX))
		      (y (random VMaxY))
		      (vx (srandom (/ maxvel 2)))
		      (vy (srandom (/ maxvel 2))))
  (new-ball x y vx vy))

;;; generate a signed, random fixnum:
(defun srandom(mag)
  (- (random (+ 1 mag mag)) 1 mag))

;;; Try some stuff:

(proclaim '(special b1 b2))

(defun try(&optional (n 10))
  (SetupCrapolla)
  (do ((i 0 (+ i 1)))
      ((= i n))
    (rand-ball))

  (setq b1 (aref ball-array 0))
  (setq b2 (aref ball-array 1))

  (do ()
      (nil)
    (update-step))
  )


; (DEFDEMO "Balls" "Many bouncing balls" (try))
