;;; -*- Mode:Common-Lisp; Package:DEMO; Fonts:(MEDFNT HL12B HL12I MEDFNB); Base:10 -*-

;;; Copyright (C) 1986, Texas Instruments Incorporated. All rights reserved.

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985, Texas Instruments Incorporated. All rights reserved.


(PROCLAIM '(SPECIAL BITS ORDER WORM-TURNS WORM-X WORM-Y FITP X-WORM Y-WORM WORM-SG))

(PROCLAIM '(SPECIAL WORM-ALU-FUNCTION ORDER WORM-X WORM-Y BITS DIR))

(DEFVAR WORM-WINDOW nil
  "The window where the worm is demoed.")

(DEFVAR *CHAR*)

;These are the "character codes" in the worm font
;for the various kinds of spots.
(DEFPARAMETER WORM-BIG-CHAR 5)
(DEFPARAMETER WORM-STRIPE-CHAR 4)
(DEFPARAMETER WORM-GRAY-CHAR 3)
(DEFPARAMETER WORM-BLACK-CHAR 2)

;This is a 6-long array of stack groups,
;each of which runs one worm.
(DEFVAR WORMS
	(LET ((ARRAY (MAKE-ARRAY 6 :TYPE 'ART-Q-LIST)))
	  (DO ((I 0 (1+ I))) ((>= I 6))
	      (setf (aref ARRAY I) (MAKE-STACK-GROUP (FORMAT NIL "WORM-~D" I))))
	  ARRAY))

(DEFUN PRESET (SG CHAR ALU-FN N)
       (STACK-GROUP-PRESET SG
			   (FUNCTION FLOP)
			   (SYMBOL-VALUE CHAR)
			   (SYMBOL-VALUE ALU-FN)
			   ORDER
			   WORM-X
			   WORM-Y
			   BITS
			   (* N (EXPT 3 (1- ORDER)))
			   *TERMINAL-IO*
			   SYS:%CURRENT-STACK-GROUP))

(DEFUN WORM (&OPTIONAL (BITS 0) (ORDER 7) (WORM-X 146.) (WORM-Y 511.) 
	     &AUX LENGTH *TERMINAL-IO*)
  (setf worm-window nil)   ;; to temporarily fix VERY strange bug about function NIL being undefined. -- dilip.
  (When (OR (NULL WORM-WINDOW)
	    (NEQ TV:DEFAULT-SCREEN
		 (SEND WORM-WINDOW :SUPERIOR)))
    (SETF WORM-WINDOW (TV:MAKE-WINDOW 'HOF-WINDOW :BLINKER-P NIL))
    (SETF *TERMINAL-IO* WORM-WINDOW))
  (TV:WINDOW-CALL (WORM-WINDOW :DEACTIVATE)
    (CATCH-ERROR-RESTART ((SYS:ABORT ERROR) "Exit WORM.")
      (SEND WORM-WINDOW :SET-LABEL "Worm")
      (SETF LENGTH (EXPT 3 (1+ ORDER)))
      (PRESET (AREF WORMS 0) 'WORM-BIG-CHAR 'TV:ALU-IOR 0)
      (SETF FITP NIL)
      (DO ((I 0 (1+ I)))
	  ((OR FITP (>= I 2)))		; Paint blackness over whole worm
	  (SETF FITP T)
	  (SEND *TERMINAL-IO* :CLEAR-SCREEN)
	  (DOTIMES (I (1+ LENGTH))
	    (IF (FUNCALL (AREF WORMS 0))
		(signal eh:abort-object)))) ;;quit
      (SETF WORM-X X-WORM WORM-Y Y-WORM)
      (MAPC (FUNCTION PRESET)			; Preset wormlets
	    (G-L-P WORMS)
	    '(WORM-GRAY-CHAR   WORM-BLACK-CHAR
	      WORM-STRIPE-CHAR WORM-BLACK-CHAR
	      WORM-BLACK-CHAR  WORM-BLACK-CHAR)
	    '(TV:ALU-XOR TV:ALU-ANDCA TV:ALU-IOR TV:ALU-IOR TV:ALU-ANDCA TV:ALU-IOR)
	    '(0 1 2 3 5 6))			;4 is intentionally missing!
      (WORM-COMMAND-LOOP))))

(DEFUN WORM-COMMAND-LOOP (&AUX (YPOS (- (TV:SHEET-INSIDE-HEIGHT *TERMINAL-IO*)
					(* 3 (TV:SHEET-LINE-HEIGHT *TERMINAL-IO*)))))
  (DO ((I 0 (1+ I))
       RUN
       (STOP-VAL 0))
      (NIL)
    (COND ((OR RUN (< I STOP-VAL))
	   (AND (READ-CHAR-NO-HANG *TERMINAL-IO*)
		(SETF RUN NIL STOP-VAL 0)))
	  (T (SEND WORM-WINDOW :SET-LABEL
		      (LET ((*PRINT-BASE* 9.))
			(FORMAT NIL "Worm    Generation ~S (base 9)" I)))
	     (SEND *TERMINAL-IO* :SET-CURSORPOS 0 YPOS)
	     (SEND *TERMINAL-IO* :CLEAR-EOL)
	     (SEND *TERMINAL-IO* :STRING-OUT
		      "    P: Run  nR: Run until n  nN: Run n steps  nS: Run till nth order  Abort: exit
    ")
	     (PROG (CH VAL)
		   (SETF VAL 0)
		LOOP (SETF CH (CHAR-UPCASE (READ-CHAR)))
		   (COND ((<= #\0 (CHAR-CODE CH) #\9)
			  (SETF VAL (+ (* VAL 9.) (- (CHAR-CODE CH) (CHAR-CODE #\0))))
			  (GO LOOP))
			 ((EQL CH #\N) (SETF STOP-VAL (+ VAL I)))
			 ((EQL CH #\R) (SETF STOP-VAL VAL))
			 ((EQL CH #\S) (SETF VAL (EXPT 3 VAL)
					    STOP-VAL (* VAL (1+ (TRUNCATE I VAL)))))
			 ((EQL CH #\P) (SETF RUN T))))
	     (IF (OR RUN (< I STOP-VAL))
		 (SEND WORM-WINDOW :SET-LABEL "Worm    Type any character to stop"))))
    (DOTIMES (I 6)
      (IF (FUNCALL (AREF WORMS I))
	  (SIGNAL EH:ABORT-OBJECT)))))

(OR (BOUNDP 'WORM-TURNS) (SETF WORM-TURNS (MAKE-ARRAY 12. :TYPE 'ART-Q-LIST)))

(zlc:FillArray WORM-TURNS '( 6  0
			 3  5
			-3  5
			-6  0
			-3 -5
			 3 -5))

;This is the coroutining function for each worm.
;Returns to the calling stack group with NIL normally,
;T if the user typed Abort.
(DEFUN FLOP (*CHAR* WORM-ALU-FUNCTION ORDER WORM-X WORM-Y BITS SNOOZE *TERMINAL-IO* WORM-SG)
  (CATCH-ERROR-RESTART ((SYS:ABORT ERROR) "Return to Worm command level.")
    (DO ((I 0 (1+ I)))
	((>= I SNOOZE))
      (STACK-GROUP-RETURN NIL))
    (DO ((I 0 (1+ I))
	 (DIR (BOOLE 4 (- ORDER) 1)))
	(NIL)
      (TERD ORDER BITS)
      (WORM-STEP)
      (SETF X-WORM WORM-X Y-WORM WORM-Y)
      (SETF DIR (+ DIR 4))))
  (DO () (()) (STACK-GROUP-RETURN T)))

(DEFUN TERD (N BITS)
  (WHEN (PLUSP N)
    (COND ((zlc:Bit-Test BITS 1)
	   (TERD (1- N) (LSH BITS -1))
	   (SETF DIR (- DIR -2))
	   (WORM-STEP)
	   (SETF DIR (- DIR 4))
	   (TERD (1- N) (LSH BITS -1))
	   (WORM-STEP)
	   (SETF DIR (+ DIR 2))
	   (TERD (1- N) (LSH BITS -1)))
	  (T
	   (TERD (1- N) (LSH BITS -1))
	   (WORM-STEP)
	   (SETF DIR (+ DIR 4))
	   (TERD (1- N) (LSH BITS -1))
	   (SETF DIR (- DIR 2))
	   (WORM-STEP)
	   (SETF DIR (- DIR 2))
	   (TERD (1- N) (LSH BITS -1))))))
	    
(DEFSUBST CLIP (Z N D)
       (CASE (TRUNCATE (+ N (SET Z (+ (SYMBOL-VALUE Z) (AREF WORM-TURNS D))))
			  N)
            (0 (SET Z 1) (SETF FITP NIL))
            (1)
            (2 (SET Z (1- N)) (SETF FITP NIL))))

(DEFUN WORM-STEP ()
   (CLIP 'WORM-X (TV:SHEET-INSIDE-WIDTH *TERMINAL-IO*) (SETF DIR (REM (+ 12. DIR) 12.)))
   (CLIP 'WORM-Y (- (TV:SHEET-INSIDE-HEIGHT *TERMINAL-IO*) 55) (1+ DIR))
   (TV:PREPARE-SHEET (*TERMINAL-IO*)
      (TV:%DRAW-CHARACTER FONTS:WORM-FONT *CHAR* NIL WORM-X WORM-Y WORM-ALU-FUNCTION *TERMINAL-IO*))
   (STACK-GROUP-RETURN NIL))

(DEFDEMO "Worm" "Pretty fractal patters, by Gosper and Holloway." (WORM))
