;;; -*- Mode:Common-lisp; Package:user; Base:10.; Fonts: MEDFNT,HL12B -*-

1;;; After loading HELP, load HELP-LIST and ZHELP-LIST.
;;;
;;; HELP provides three areas of assistance.  The first is an action/object interface
;;; to top level Explorer commands.  If you want to do something, such as look
;;; at a file, but do not know the command, type (HELP).  A set of actions (the
;;; items in capital letters) will be displayed.  Click on one that you want to do, such
;;; as SHOW.  Another window appears with a set of items that apply to the action
;;; you choose, such as file.  Finally the actual LISP command is displayed along with 
;;; the documentation and arguments for the command.
;;;
;;; The second area is all of the COMMON Lisp constants, variables, macros, and functions
;;; are in the data base.  These are the lower case items that appear on the first window.
;;; They are broken into essentially the same categories that Steel's Common Lisp book has
;;; chapters and sections.  Some of the items are a subset of other items.  Such as "string
;;; access" is a subset of "strings".
;;;
;;; The third are contains most of the DEFFLAVOR, DEFUN, DEFSUBST, DEFMACRO, DEFVAR,
;;; DEFCONST, DEFSIGNAL, DEFPARAMETER, DEFCOM,  and DEFCOMMAND names within the 
;;; system.  You have to execute (LOAD-PACKAGE-TABLES) to set up this data base.  Then
;;; execute (HELP T).  The first window is the subsystem you want to examine, the second
;;; the type, and the third the list of items.  Clicking on the third repeats the second window.
;;;
;;; ZHELP provides a quick reference to the ZMACs commands explained in the Reference 
;;; Manual.  Type (ZHELP) to execute the command.
;;;*

(DEFMACRO THEN (&BODY BODY)
  `(PROGN ,@BODY))

(DEFMACRO ELSE (&BODY BODY)
  `(PROGN ,@BODY))

(DEFCONSTANT HELP-CHOICE-LIST '(("Do it" (*THROW 'DOIT NIL))
				"Repeat first screen"
				("Repeat second screen" (*THROW '2ND NIL))
				("     Abort []" (*THROW 'ABORT NIL))
				))
(DEFVAR VERB NIL)

(DEFVAR VAR-LIST (LIST 'USER:VAR1 'USER:VAR2 'USER:VAR3 'USER:VAR4 'USER:VAR5 'USER:VAR6 'USER:VAR7
		       'USER:VAR8 'USER:VAR9 'USER:VAR10 'USER:VAR11 'USER:VAR12 'USER:VAR13 'USER:VAR14
		       'USER:VAR15 'USER:VAR16 'USER:VAR17 'USER:VAR18 'USER:VAR19 'USER:VAR20 'USER:VAR21
		       'USER:VAR22 'USER:VAR23 'USER:VAR24 'USER:VAR25 'USER:VAR26 'USER:VAR27 'USER:VAR28
		       'USER:VAR29 'USER:VAR30 'USER:VAR31 'USER:VAR32 'USER:VAR33 'USER:VAR34 'USER:VAR35))

(DEFVAR VERB-TABLE (MAKE-HASH-TABLE :TEST #'EQUAL))

(DEFUN HELP  (&OPTIONAL (DIRECTORY NIL)
              &AUX FUNCTION PACK CHOOSEN-PACK V-TABLE
	      FUNCTION-ARGUMENT ARGUMENTS ARGUMENT-LIST
	      DOC DOC2 DISPLAY VERB NOUN (OPTIONAL NIL) (REST NIL)
	      (KEY NIL) (ALLOW-OTHER-KEYS NIL)
	      REMOVE KEY-VARIABLE DONE (CHOOSEN-VERB NIL) (CHOOSEN-NOUN NIL))
  "2Display possible functions to perform desired tasks.  If the verb and/or noun is
    known, it may be entered.  The arguments are entered with double quote marks
    around them.  I would give an example if I could remember how to put a double
    quote within a documentation string.  Such is life.
    If DIRECTORY is T the system functions, variables, macros and flavors will be 
    displayed*"
  
  (DECLARE (SPECIAL VERB-TABLE HELP-CHOICE-LIST VAR-LIST PACKAGE-TABLE))
  
  (*CATCH 'ABORT 
    (WHEN DIRECTORY
      (SETF PACK (PACKAGE-LIST))
      (SETF CHOOSEN-PACK
	    (HELP-MENU-CHOOSE PACK "Desired subsystem")))      
    
    (IF (NOT VERB)
	(THEN
	  (SETF VERB (VERB-LIST (SETF V-TABLE (IF DIRECTORY
						  (THEN (GETHASH CHOOSEN-PACK PACKAGE-TABLE))
						  (ELSE VERB-TABLE)))))))
    
    (*CATCH 'DOIT
      
      (DO-FOREVER
	(*CATCH '1ST
	  (IF (NOT CHOOSEN-VERB)
	      (THEN
		(IF DIRECTORY
		      (THEN
			(SETF CHOOSEN-VERB
			      (HELP-MENU-CHOOSE VERB "zzz-quit-help to terminate")))
		      (ELSE
			(SETF CHOOSEN-VERB
			      (HELP-MENU-CHOOSE VERB "Initial concept  -  lower case items are Common Lisp
                 -  zzz-quit-help to terminate"))))))
	  (WHEN (EQUAL "zzz-quit-help" choosen-verb)
	    (THROW 'abort t))
	  
	  (WHEN choosen-verb
	    
	    (SETF NOUN (NOUN-LIST CHOOSEN-VERB V-TABLE))
	    
	    (*CATCH '2ND   
	      
	      (IF (NOT CHOOSEN-NOUN)
		  (THEN
		    (SETF CHOOSEN-NOUN
			  (HELP-MENU-CHOOSE NOUN CHOOSEN-VERB))))
	      (WHEN DIRECTORY
		(SETF CHOOSEN-VERB NIL
		      CHOOSEN-NOUN NIL)
		(*THROW '1ST NIL))
	      
;1 GET THE FUNCTION FOR THE SPECIFIED VERB/NOUN PAIR*
	      (SETF FUNCTION (GETHASH CHOOSEN-NOUN
				      (GETHASH CHOOSEN-VERB V-TABLE)))
	      
	      
;1 DISPLAY THE FUNCTION DOCUMENTATION AND PROMPTS*
	      
	      (SETF DOC (IF (OR (FUNCTIONP FUNCTION)
				(SPECIAL-FORM-P FUNCTION)
				(MACRO-FUNCTION FUNCTION))
			    (THEN
			      (SETF ARGUMENTS (ARGLIST FUNCTION))
			      (DOCUMENTATION FUNCTION 'FUNCTION))
			    (ELSE
			      (SETF ARGUMENTS NIL)
			      (DOCUMENTATION FUNCTION 'VARIABLE))))
	      
	      (SETF DOC2 NIL)
	      (DO* ((START 0 LINE-LENGTH)
		    (LINE-LENGTH 80 80))
		   ((> (+ START 80) (LENGTH DOC))
		    (SETF DOC2 (NCONC DOC2 (LIST (SUBSTRING DOC START)))))
		(SETF LINE-LENGTH (POSITION #\SPACE (THE STRING DOC)
					    :FROM-END T :END (+ START 80.) :TEST #'CHAR=))
		(SETF DOC2 (NCONC DOC2 (LIST (SUBSTRING DOC START LINE-LENGTH)))))
	      
	      
	      (SETF ARGUMENT-LIST NIL)
	      
	      (SETF DISPLAY NIL)
	      
	      (DO* ((INDEX 0 (1+ INDEX))
		    (LOCAL-VARIABLE (NTH INDEX VAR-LIST) (NTH INDEX VAR-LIST)))
		   
		   ((OR (NOT (LISTP ARGUMENTS))
			(= INDEX (LENGTH ARGUMENTS)))
		    (IF (ZEROP INDEX)
			(THEN
			  (SETF (SYMBOL-VALUE LOCAL-VARIABLE) " ")
			  (IF (OR (FUNCTIONP FUNCTION)
				  (MACRO-FUNCTION FUNCTION))
			      (THEN
				(SETF DISPLAY 
				      (NCONC
					DISPLAY
					(LIST `(,LOCAL-VARIABLE
						"There are no arguments" :DOCUMENTATION "" :SEXP)))))
			      (ELSE
				(IF (CONSTANTP FUNCTION)
				    (THEN
				      (SETF DISPLAY 
					    (NCONC
					      DISPLAY
					      (LIST `(,LOCAL-VARIABLE
						      "This is a constant" :DOCUMENTATION "" :SEXP)))))
				    (ELSE
				      (SETF DISPLAY 
					    (NCONC
					      DISPLAY
					      (LIST `(,LOCAL-VARIABLE
						      "This is a variable" :DOCUMENTATION "" :SEXP)))))))))))
		
		(IF (ATOM (NTH INDEX ARGUMENTS))
		    (THEN
		      (SETF (SYMBOL-VALUE LOCAL-VARIABLE) " ")
		      (SETF FUNCTION-ARGUMENT (STRING (NTH INDEX ARGUMENTS))))
		    (ELSE
		      (SETF FUNCTION-ARGUMENT (STRING (CAR (NTH INDEX ARGUMENTS))))
		      (SETF (SYMBOL-VALUE LOCAL-VARIABLE)
			    (IF (SYMBOLP (CADR (NTH INDEX ARGUMENTS)))
				(THEN
				  (SYMBOL-VALUE (CADR (NTH INDEX ARGUMENTS))))
				(ELSE
				  (CADR (NTH INDEX ARGUMENTS)))))))
		(IF (SEARCH "&" FUNCTION-ARGUMENT) ;
		    (THEN
		      (COND
			((SEARCH "OPTIONAL" FUNCTION-ARGUMENT)
			 (SETF OPTIONAL INDEX))
			((SEARCH "REST" FUNCTION-ARGUMENT)
			 (SETF REST INDEX))
			((SEARCH "ALLOW-OTHER-KEYS" FUNCTION-ARGUMENT)
			 (SETF ALLOW-OTHER-KEYS INDEX))
			((SEARCH "KEY" FUNCTION-ARGUMENT)
			 (SETF KEY INDEX))
			(T
			 (SETF REST INDEX)))
		      (SETF DISPLAY (NCONC DISPLAY (LIST " " FUNCTION-ARGUMENT))))
		    (ELSE
		      (IF KEY
			  (THEN
					   ;(SETF FUNCTION-ARGUMENT (STRING-APPEND ":" FUNCTION-ARGUMENT))
			    (SETF ARGUMENT-LIST (NCONC
						  ARGUMENT-LIST
						  (LIST FUNCTION-ARGUMENT LOCAL-VARIABLE))))
			  (ELSE
			    (SETF ARGUMENT-LIST (NCONC ARGUMENT-LIST (LIST LOCAL-VARIABLE)))))
		      (SETF DISPLAY 
			    (NCONC
			      DISPLAY
			      (LIST `(,LOCAL-VARIABLE ,FUNCTION-ARGUMENT :DOCUMENTATION "" :SEXP)))))))
	      (SETF DISPLAY
		    (NCONC
		      DOC2 (LIST " ") DISPLAY))
					   ;(FORMAT t "~%~A ~a" function (NULL function))
	      (WHEN function
		(TV:CHOOSE-VARIABLE-VALUES
		  DISPLAY
		  :LABEL (STRING FUNCTION)
		  :MARGIN-CHOICES HELP-CHOICE-LIST
		  :WIDTH 90.
		  :HEIGHT 500.)
		)
	      
	      (SETF CHOOSEN-VERB NIL))
	    (SETF CHOOSEN-NOUN NIL)
	    )))
      )
;1;  NOTE THIS HAS PROBLEMS WHEN KEY  OR OPTIONAL ARGUMENTS HAVE A VALUE OF NIL*
;1;  REST NOT TAKEN CARE OF.*
    (IF KEY				   ;1 If there are key word args*
	(THEN				   ;1 ...remove all keywords from the list that the user did not provide*
	  (SETF REMOVE NIL)
	  (SETF KEY-VARIABLE (1+ (IF OPTIONAL
				     (THEN (MAX OPTIONAL KEY))
				     (ELSE KEY))))
	  
	  (DOLIST (ITEM (NLEFT (- (LENGTH ARGUMENT-LIST)
				  (IF OPTIONAL
				      (THEN (MIN KEY OPTIONAL))
				      (ELSE KEY)))
			       ARGUMENT-LIST))
	    (IF REMOVE
		(THEN
		  (IF (EQUAL " " (SYMBOL-VALUE (NTH KEY-VARIABLE VAR-LIST)))
		      (THEN
			(SETF ARGUMENT-LIST (DELETE ITEM (THE LIST ARGUMENT-LIST) :TEST #'EQ))
			(SETF ARGUMENT-LIST (DELETE REMOVE (THE LIST ARGUMENT-LIST) :TEST #'EQ))))
		  (INCF KEY-VARIABLE)
		  (SETF REMOVE NIL))
		(ELSE
		  (SETF REMOVE ITEM))))	   ;1 Save keyword*
	  (IF (AND OPTIONAL		   ;1 If there are optional args before the keys*
		   (< OPTIONAL (1- KEY)))
	      (THEN			   ;1 ...remove the ones that were not set*
		(DO ((INDEX (1+ OPTIONAL) (1+ INDEX)))
		    ((= INDEX KEY))
		  (IF (EQUAL " " (SYMBOL-VALUE (NTH INDEX VAR-LIST)))
		      (THEN
			(SETF ARGUMENT-LIST (DELETE
					      (NTH INDEX VAR-LIST)
					      (THE LIST ARGUMENT-LIST) :TEST #'EQ))))))))
	(ELSE
	  (IF OPTIONAL
	      (THEN
		(SETF DONE NIL)
		(DO ((INDEX (1+ OPTIONAL) (1+ INDEX)))
		    ((OR DONE
			 (> INDEX (LENGTH ARGUMENT-LIST))))
		  (IF (EQUAL " " (SYMBOL-VALUE (NTH INDEX VAR-LIST)))
		      (THEN
			(SETF DONE T)
			(SETF ARGUMENT-LIST (FIRSTN (1- INDEX) ARGUMENT-LIST)))))))))
    
;1;  Evaluate the variables in the argument list.  Remember that some items in the list*
;1;  may be keywords.*
    (DOLIST (ITEM ARGUMENT-LIST)
      (IF (MEMBER ITEM VAR-LIST)
	  (THEN
	    (NSUBST (IF (AND (SYMBOLP (SYMBOL-VALUE ITEM))
			     (BOUNDP (SYMBOL-VALUE ITEM)))
			(THEN
			  (SYMBOL-VALUE (SYMBOL-VALUE ITEM)))
			(ELSE
			  (SYMBOL-VALUE ITEM)))
		    ITEM ARGUMENT-LIST))
	  (ELSE
	    (NSUBST (INTERN ITEM 'KEYWORD) ITEM ARGUMENT-LIST))))
    
    (IF (FUNCTIONP FUNCTION)
	(THEN
	  (APPLY FUNCTION ARGUMENT-LIST))
	(ELSE
	  (EVAL FUNCTION)))
    ))


(DEFUN VERB-LIST (TABLE
		  &AUX (VERBS NIL))
  1"Return the list of defined verbs."*

  (MAPHASH #'(LAMBDA (VERB IGNORE)
	   (SETF VERBS (APPEND VERBS (LIST VERB))))
	   TABLE)

  (SORT VERBS #'(LAMBDA (X Y) (STRING< X Y))))


(DEFUN NOUN-LIST (CHOOSEN-VERB TABLE
		  &AUX (NOUNS NIL))
  "2Return the list of nouns associated with the specified verb*"

  (MAPHASH #'(LAMBDA (NOUN IGNORE)
	       (SETF NOUNS (APPEND NOUNS (LIST NOUN))))
	   (GETHASH CHOOSEN-VERB TABLE))

  (SORT NOUNS #'(LAMBDA (X Y) (STRING< X Y))))

(DEFUN PACKAGE-LIST (
		  &AUX (PACKAGES NIL))
  1"Return the list of defined 2packages.*"*

  (DECLARE (SPECIAL PACKAGE-TABLE))

  (MAPHASH #'(LAMBDA (PACK IGNORE)
	   (SETF PACKAGES (APPEND PACKAGES (LIST PACK))))
	   PACKAGE-TABLE)

  (SORT PACKAGES #'(LAMBDA (X Y) (STRING< X Y))))


(defun HELP-MENU-CHOOSE (mlist &optional (label nil)
		       (near-mode '(:mouse))
		       default-item
		       (superior tv:mouse-sheet)
		       &aux (choosen-item nil)
		       (last-item    nil))
  2"Brings up a choose menu that is set to have the correct geometry."*
  ;1 This would be unnecessary if the tv:momentary-menu resource were set up to*
  ;1 reintialize each window's characteristics to a specified default before allocation.*
  (unless (null mlist)
    (when (eq (car near-mode) ':window)
      (setq superior (tv:sheet-superior (cadr near-mode))))
    (using-resource (menu tv:momentary-menu superior)
      (funcall menu ':set-label label)
      (funcall menu ':set-item-list mlist)
      (funcall menu ':set-geometry  3 nil nil nil nil nil)
      (funcall menu ':set-last-item default-item)
      (tv:expose-window-near menu near-mode)
      (when (and (not (null default-item))
		 (not (memq (car near-mode) '(:mouse :point))))
	(multiple-value-bind (x y)
	    (funcall menu ':item-cursorpos default-item)
	  (unless (or (null x) (null y))
	    (funcall menu ':set-mouse-position
		     (+ x (tv:sheet-inside-left menu))
		     (+ y (tv:sheet-inside-top  menu))))))
      (setq choosen-item (send menu ':choose))
      (setq last-item    (send menu ':last-item))
      (funcall menu ':set-geometry nil nil nil nil nil nil) ;1set default next time menu resourced*
      ))
  (values choosen-item last-item))


(DEFVAR ZVERB-TABLE (MAKE-HASH-TABLE :TEST #'EQUAL))
(DEFCONSTANT ZHELP-CHOICE-LIST '("Repeat first screen"
				 ("Repeat second screen" (*THROW '2ND NIL))
				 ("Abort []" (*THROW 'ABORT NIL))
				))
(DEFVAR USER:ZHELP-VARIABLE NIL)
(DEFUN ZHELP (
              &AUX CHOOSEN-VERB CHOOSEN-NOUN NOUN DOC DOC2 DISPLAY)
  "2HELP on finding the commands to perform a desired operation in ZMACS*"
  
  (DECLARE (SPECIAL ZVERB-TABLE ZHELP-CHOICE-LIST USER:ZHELP-VARIABLE))
  
  (*CATCH 'ABORT 
    
    (SETF VERB (VERB-LIST ZVERB-TABLE))
    
    (DO-FOREVER
      (SETF DISPLAY NIL
	    CHOOSEN-NOUN "")
      
      (UNLESS CHOOSEN-VERB
	(SETF CHOOSEN-VERB
	      (HELP-MENU-CHOOSE VERB "Initial concept  -  zzz-quit-help to terminate"))
	(WHEN (EQUAL "zzz-quit-help" choosen-verb)
	  (THROW 'abort t)))
      
      (*CATCH '2ND
	(SETF DOC (IF (HASH-TABLE-P (GETHASH CHOOSEN-VERB ZVERB-TABLE))
		      (THEN
			(SETF NOUN (NOUN-LIST CHOOSEN-VERB ZVERB-TABLE))
			(SETF CHOOSEN-NOUN
			      (HELP-MENU-CHOOSE NOUN CHOOSEN-VERB))
			(GETHASH CHOOSEN-NOUN
				 (GETHASH CHOOSEN-VERB ZVERB-TABLE)))
		      (ELSE
			(GETHASH CHOOSEN-VERB ZVERB-TABLE))))
	
;1 DISPLAY THE FUNCTION DOCUMENTATION AND PROMPTS*
	
	(SETF DOC2 NIL)
	(DO* ((START 0 LINE-LENGTH)
	      (LINE-LENGTH 80 80))
	     ((> (+ START 80) (LENGTH DOC))
	      (SETF DOC2 (NCONC DOC2 (LIST (SUBSTRING DOC START)))))
	  (SETF LINE-LENGTH (POSITION #\SPACE (THE STRING DOC)
				      :FROM-END T :END (+ START 80.) :TEST #'CHAR=))
	  (SETF DOC2 (NCONC DOC2 (LIST (SUBSTRING DOC START LINE-LENGTH)))))
	
	
	(SETF DISPLAY 
	      (NCONC
		DISPLAY (LIST '(USER:ZHELP-VARIABLE
				 "" :DOCUMENTATION "" :SEXP))))
	(SETF DISPLAY
	      (NCONC
		DOC2 (LIST " ") DISPLAY))
	
	(TV:CHOOSE-VARIABLE-VALUES
	  DISPLAY
	  :LABEL (STRING-APPEND CHOOSEN-VERB " " CHOOSEN-NOUN)
	  :MARGIN-CHOICES ZHELP-CHOICE-LIST
	  :WIDTH 90.
	  :HEIGHT 200.)
	(SETF CHOOSEN-VERB NIL)
	)				   ;1 end of '2ND*
      )					   ;1 end of DO-FOREVER*
    )					   ;1 end of ABORT*
  )					   ;1 end of ZHELP*

(DEFUN ADD-ZHELP-FUNCTION (VERBS NOUNS DOC
			   &AUX THIS-VERB-TABLE)
  "2Accept a verb and noun that describe the use of a function.  Both VERB and
    NOUN may be a list.*"
  
  (DECLARE (SPECIAL ZVERB-TABLE VERB))
  
  (SETF VERB NIL)			   ;1 Force rebuilt of the verb list*
  (IF (ATOM VERBS)
      (THEN
	(SETF VERBS (LIST VERBS))))
  (IF (ATOM NOUNS)
      (THEN
	(SETF NOUNS (LIST NOUNS))))
  
  (DO* ((VERB-INDEX 0 (1+ VERB-INDEX))
	(THIS-VERB (CAR VERBS) (NTH VERB-INDEX VERBS)))
       
       ((NOT THIS-VERB))
    
    (WHEN (NOT (SETF THIS-VERB-TABLE (GETHASH THIS-VERB ZVERB-TABLE)))
      (IF (CAR NOUNS)
	  (THEN
	    (SETF THIS-VERB-TABLE
		  (SETF (GETHASH THIS-VERB ZVERB-TABLE) (MAKE-HASH-TABLE :TEST #'EQUAL))))
	  (ELSE
	    (SETF (GETHASH THIS-VERB ZVERB-TABLE) DOC))))
    
    (DO* ((NOUN-INDEX 0 (1+ NOUN-INDEX))
	  (THIS-NOUN (CAR NOUNS) (NTH NOUN-INDEX NOUNS)))
	 
	 ((NOT THIS-NOUN))
      (SETF (GETHASH THIS-NOUN THIS-VERB-TABLE) DOC))))


(DEFUN ADD-HELP-FUNCTION (VERBS NOUNS FUNCTION
		    &AUX THIS-VERB-TABLE)
  "2Accept a verb and noun that describe the use of a function.  Both VERB and
    NOUN may be a list.*"

  (DECLARE (SPECIAL VERB-TABLE VERB))

  (SETF VERB NIL)    ;1 Force rebuilt of the verb list*
  (IF (ATOM VERBS)
      (THEN
	(SETF VERBS (LIST VERBS))))
  (IF (ATOM NOUNS)
      (THEN
	(SETF NOUNS (LIST NOUNS))))

  (DO* ((VERB-INDEX 0 (1+ VERB-INDEX))
	(THIS-VERB (CAR VERBS) (NTH VERB-INDEX VERBS)))

       ((NOT THIS-VERB))

    (IF (NOT (SETF THIS-VERB-TABLE (GETHASH THIS-VERB VERB-TABLE)))
	(THEN
	  (SETF THIS-VERB-TABLE
		(SETF (GETHASH THIS-VERB VERB-TABLE) (MAKE-HASH-TABLE :TEST #'EQUAL)))))

    (DO* ((NOUN-INDEX 0 (1+ NOUN-INDEX))
	  (THIS-NOUN (CAR NOUNS) (NTH NOUN-INDEX NOUNS)))

	 ((NOT THIS-NOUN))
      (SETF (GETHASH THIS-NOUN THIS-VERB-TABLE) FUNCTION))))



;1;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*
;1;;*
;1;; CODE TO SETUP PACKAGE DATA BASE*
;1;;*
(DEFCONST FUNCTION-CALL "(USER:ADD-HELP-FUNCTION-2 ")

(DEFUN SETUP-PACKAGE-DATA-BASE-DIRECTORY (DIRECTORY OUTPUT-PATHNAME 
					  &AUX ENTRIES NUMBER-OF-ENTRIES)
  
  (SETF ENTRIES (DIRECTORY DIRECTORY))	   ;1 List of files/directories in the directory*
  (SETF NUMBER-OF-ENTRIES (LENGTH ENTRIES))
  
  (DOTIMES (INDEX          
	     NUMBER-OF-ENTRIES)
    
    (LET ((ENTRY (NTH INDEX ENTRIES)))	   ;1 get the next entry*
;1; Do not process a directory or a file that is not the newest version or not a LISP file.*
      (WHEN (AND (NOT (GETF (CDR (FS:FILE-PROPERTIES ENTRY)) :DIRECTORY))
		 (IF (= (1+ INDEX) NUMBER-OF-ENTRIES)	   
		     T 		;1 last entry is newest version*
		     (NOT (EQUAL ;1 compare file names *
			   (STRING-RIGHT-TRIM "0 1 2 3 4 5 6 7 8 9" ENTRY)
			   (STRING-RIGHT-TRIM "0 1 2 3 4 5 6 7 8 9" (NTH (1+ INDEX) ENTRIES)))))
		 (EQUAL (STRING-UPCASE (PATHNAME-TYPE ENTRY)) "LISP"))
	(SETUP-PACKAGE-DATA-BASE-FILE ENTRY OUTPUT-PATHNAME)))))


(DEFUN SETUP-PACKAGE-DATA-BASE-FILE (INPUT-PATHNAME OUTPUT-PATHNAME
				     &AUX LINE EOF DEFINITION ITEM START TEMP FILE-PACKAGE)
  "2Read a lisp file and pull out all of the DEFFLAVOR, DEFUN, DEFSUBST, DEFMACRO,
    DEFVAR, DEFCONST, and DEFSIGNAL forms.  Put them in a file to be used as input
    for ADD-HELP-2.*"
  
  (WITH-OPEN-FILE (OUTPUT OUTPUT-PATHNAME
			  :DIRECTION :OUTPUT
			  :IF-DOES-NOT-EXIST :CREATE
			  :IF-EXISTS :APPEND)
    
    (WITH-OPEN-FILE (INPUT INPUT-PATHNAME)
      (MULTIPLE-VALUE-SETQ (LINE EOF) (READ-LINE INPUT NIL))
      (WHEN (NOT EOF)
	(SETF TEMP (search "PACKAGE:" (string-upcase line)))
	(IF TEMP
	    (THEN
	      (SETF TEMP (+ 8 TEMP))
	      (SETF FILE-PACKAGE (STRING-TRIM
				   #\SPACE
				   (SUBSTRING LINE
					      TEMP
					      (MIN (OR (POSITION #\; LINE :START TEMP) (LENGTH LINE))
						   (OR (POSITION #\SPACE LINE :START (+ 2 TEMP)) (LENGTH LINE))
						   )))))
	    (ELSE
	      (SETF EOF T)))
	(DO ()
	    (EOF)
	  (MULTIPLE-VALUE-SETQ (LINE EOF) (READ-LINE INPUT NIL))
	  (WHEN (AND (NOT EOF) (EQUAL (CHAR LINE 0) #\())
	    (SETF DEFINITION (STRING-UPCASE
			       (SUBSTRING LINE 1 (MIN (OR (POSITION #\SPACE LINE) (LENGTH LINE))
						      (OR (POSITION #\) LINE) (LENGTH LINE))))))
	    (WHEN (OR (EQUAL DEFINITION "DEFFLAVOR")
		      (EQUAL DEFINITION "DEFUN")
		      (EQUAL DEFINITION "DEFSUBST")
		      (EQUAL DEFINITION "DEFMACRO")
		      (EQUAL DEFINITION "DEFVAR")
		      (EQUAL DEFINITION "DEFCONST")
		      (EQUAL DEFINITION "DEFSIGNAL")
		      (EQUAL DEFINITION "DEFPARAMETER")
		      (EQUAL DEFINITION "DEFCOM")
		      (EQUAL DEFINITION "DEFCOMMAND"))
	      
	      (SETF START (+ 2 (LENGTH DEFINITION)))
	      (WHEN (< START (LENGTH LINE))
		(SETF ITEM
		      (STRING-UPCASE
			(STRING-TRIM
			  #\SPACE
			  (SUBSTRING LINE
				     START
				     (MIN (OR (POSITION #\SPACE LINE :START START) (LENGTH LINE))
					  (OR (POSITION #\) LINE :START START) (LENGTH LINE))
					  (OR (POSITION #\( LINE :START START) (LENGTH LINE)))))))
		
                (WHEN (AND (> (LENGTH ITEM) 0)
			   (STRING/= "(" ITEM :END2 1))
		  (WRITE-STRING FUNCTION-CALL OUTPUT)
		  (WRITE-CHAR #\" OUTPUT)
		  (WRITE-STRING (STRING-UPCASE FILE-PACKAGE) OUTPUT)
		  (WRITE-CHAR #\" OUTPUT)
		  (TERPRI OUTPUT)
		  
		  (WRITE-STRING "                       " OUTPUT)
		  (WRITE-CHAR #\" OUTPUT)	    
		  (WRITE-STRING (STRING-UPCASE DEFINITION) OUTPUT)
		  (WRITE-CHAR #\" OUTPUT)
		  (TERPRI OUTPUT)
		  
		  (WRITE-STRING "                       " OUTPUT)
		  (WRITE-CHAR #\" OUTPUT)	    
		  (WRITE-STRING ITEM OUTPUT)
		  (WRITE-CHAR #\" OUTPUT)
		  (TERPRI OUTPUT)
		  
		  (WRITE-STRING "                       " OUTPUT)
		  (WRITE-CHAR #\' OUTPUT)	    
		  (WRITE-STRING ITEM OUTPUT)
		  (WRITE-CHAR #\) OUTPUT)
		  (TERPRI OUTPUT)
		  )))))))))
  
  
(DEFVAR PACKAGE-TABLE NIL)
(DEFVAR CB (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;1COLD-BOOT*
(DEFVAR DU (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;1DEBUG-UTILITIES*
(DEFVAR ER (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;1ERROR-HANDLER*
(DEFVAR FI (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;1FILE*
(DEFVAR FE (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;1FONT-EDITOR*
(DEFVAR FO (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;FORMATTER
(DEFVAR GE (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;1GED*
(DEFVAR GL (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;GLOSSARY
(DEFVAR GR (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;GRASPER
(DEFVAR GW (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;GWIN
(DEFVAR IM (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;IMAGEN
(DEFVAR IE (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;INPUT-EDITOR
(DEFVAR IO (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;1IO AND IO1*
(DEFVAR IP (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;1 IP*
(DEFVAR MA (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;MAIL
(DEFVAR MM (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;MEMORY-MANAGEMENT
(DEFVAR NE (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;NET
(DEFVAR NC (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;NET-CONFIG
(DEFVAR NL (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;NLMENU
(DEFVAR PA (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;PATHNAME
(DEFVAR LP (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;PRINTER
(DEFVAR PR (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;PROFILE
(DEFVAR RA (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;RASTER-TECH
(DEFVAR RT (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;RTMS
(DEFVAR SE (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;SERIAL
(DEFVAR ST (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;STREAMER-TAPE
(DEFVAR SU (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;SUGGESTIONS
(DEFVAR SY (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;SYS1 AND SYS2*
(DEFVAR SL (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;SYSLOG
(DEFVAR TE (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;TELNET
(DEFVAR UC (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;UCL
(DEFVAR VT (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;VT100
(DEFVAR WI (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;WINDOW
(DEFVAR ZW (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 17))  ;ZWEI
(DEFVAR MAIN-TABLE (MAKE-HASH-TABLE :TEST #'EQUAL :SIZE 60))

(DEFUN LOAD-PACKAGE-TABLES ()
  (DECLARE (SPECIAL CB DU ER FI FE FO GE GL GR GW IM IE IO IP
		    MA MM NE NC NL PA LP PR RA RT SE ST SU SY
		    SL TE UC VT WI ZW))
  (SETF PACKAGE-TABLE CB)
  (LOAD "INDUS:GLEN.HELP.REL2;COLD-BAND.XFASL#>")
  (SETF PACKAGE-TABLE DU)
  (LOAD "INDUS:GLEN.HELP.REL2;DEBUG-UTILITIES.XFASL#>")
  (SETF PACKAGE-TABLE ER)
  (LOAD "INDUS:GLEN.HELP.REL2;EH.XFASL#>")
  (SETF PACKAGE-TABLE FI)
  (LOAD "INDUS:GLEN.HELP.REL2;FILE.XFASL#>")
  (SETF PACKAGE-TABLE FE)
  (LOAD "INDUS:GLEN.HELP.REL2;FONT-EDITOR.XFASL#>")
  (SETF PACKAGE-TABLE FO)
  (LOAD "INDUS:GLEN.HELP.REL2;FORMATTER.XFASL#>")
  (SETF PACKAGE-TABLE GE)
  (LOAD "INDUS:GLEN.HELP.REL2;GED.XFASL#>")
  (SETF PACKAGE-TABLE GL)
  (LOAD "INDUS:GLEN.HELP.REL2;GLOSSARY.XFASL#>")
  (SETF PACKAGE-TABLE GR)
  (LOAD "INDUS:GLEN.HELP.REL2;GRASPER.XFASL#>")
  (SETF PACKAGE-TABLE GW)
  (LOAD "INDUS:GLEN.HELP.REL2;GWIN.XFASL#>")
  (SETF PACKAGE-TABLE IM)
  (LOAD "INDUS:GLEN.HELP.REL2;IMAGEN.XFASL#>")
  (SETF PACKAGE-TABLE IE)
  (LOAD "INDUS:GLEN.HELP.REL2;INPUT-EDITOR.XFASL#>")
  (SETF PACKAGE-TABLE IO)
  (LOAD "INDUS:GLEN.HELP.REL2;IO.XFASL#>")
;  (LOAD "INDUS:GLEN.HELP.REL2;IO1.XFASL#>")
  (SETF PACKAGE-TABLE IP)
  (LOAD "INDUS:GLEN.HELP.REL2;IP.XFASL#>")
  (SETF PACKAGE-TABLE MA)
  (LOAD "INDUS:GLEN.HELP.REL2;MAIL.XFASL#>")
  (SETF PACKAGE-TABLE MM)
  (LOAD "INDUS:GLEN.HELP.REL2;MEMORY-MANAGEMENT.XFASL#>")
  (SETF PACKAGE-TABLE NE)
  (LOAD "INDUS:GLEN.HELP.REL2;NET.XFASL#>")
  (SETF PACKAGE-TABLE NC)
  (LOAD "INDUS:GLEN.HELP.REL2;NET-CONFIG.XFASL#>")
  (SETF PACKAGE-TABLE NL)
  (LOAD "INDUS:GLEN.HELP.REL2;NLMENU.XFASL#>")
  (SETF PACKAGE-TABLE PA)
  (LOAD "INDUS:GLEN.HELP.REL2;PATHNAME.XFASL#>")
  (SETF PACKAGE-TABLE LP)
  (LOAD "INDUS:GLEN.HELP.REL2;PRINTER.XFASL#>")
  (SETF PACKAGE-TABLE PR)
  (LOAD "INDUS:GLEN.HELP.REL2;PROFILE.XFASL#>")
  (SETF PACKAGE-TABLE RA)
  (LOAD "INDUS:GLEN.HELP.REL2;RASTER-TECH.XFASL#>")
  (SETF PACKAGE-TABLE RT)
  (LOAD "INDUS:GLEN.HELP.REL2;RTMS.XFASL#>")
  (SETF PACKAGE-TABLE SE)
  (LOAD "INDUS:GLEN.HELP.REL2;SERIAL.XFASL#>")
  (SETF PACKAGE-TABLE ST)
  (LOAD "INDUS:GLEN.HELP.REL2;STREAMER-TAPE.XFASL#>")
  (SETF PACKAGE-TABLE SU)
  (LOAD "INDUS:GLEN.HELP.REL2;SUGGESTIONS.XFASL#>")
  (SETF PACKAGE-TABLE SY)
  (LOAD "INDUS:GLEN.HELP.REL2;SYS.XFASL#>")
  (LOAD "INDUS:GLEN.HELP.REL2;SYS2.XFASL#>")
  (SETF PACKAGE-TABLE SL)
  (LOAD "INDUS:GLEN.HELP.REL2;SYSLOG.XFASL#>")
  (SETF PACKAGE-TABLE TE)
  (LOAD "INDUS:GLEN.HELP.REL2;TELNET.XFASL#>")
  (SETF PACKAGE-TABLE UC)
  (LOAD "INDUS:GLEN.HELP.REL2;UCL.XFASL#>")
  (SETF PACKAGE-TABLE VT)
  (LOAD "INDUS:GLEN.HELP.REL2;VT100.XFASL#>")
  (SETF PACKAGE-TABLE WI)
  (LOAD "INDUS:GLEN.HELP.REL2;WINDOW.XFASL#>")
  (SETF PACKAGE-TABLE ZW)
  (LOAD "INDUS:GLEN.HELP.REL2;ZWEI.XFASL#>")
  (SETF PACKAGE-TABLE MAIN-TABLE)
  (SETF (GETHASH "COLD-BOOT" MAIN-TABLE) CB)
  (SETF (GETHASH "DEBUG-UTILITIES" MAIN-TABLE) DU)
  (SETF (GETHASH "ERROR-HANDLER" MAIN-TABLE) ER)
  (SETF (GETHASH "FILE" MAIN-TABLE) FI)
  (SETF (GETHASH "FONT-EDITOR" MAIN-TABLE) FE)
  (SETF (GETHASH "FORMATTER" MAIN-TABLE) FO)
  (SETF (GETHASH "GED" MAIN-TABLE) GE)
  (SETF (GETHASH "GLOSSARY" MAIN-TABLE) GL)
  (SETF (GETHASH "GRASPER" MAIN-TABLE) GR)
  (SETF (GETHASH "GWIN" MAIN-TABLE) GW)
  (SETF (GETHASH "IMAGEN" MAIN-TABLE) IM)
  (SETF (GETHASH "INPUT-EDITOR" MAIN-TABLE) IE)
  (SETF (GETHASH "IO AND IO1" MAIN-TABLE) IO)
  (SETF (GETHASH "IP" MAIN-TABLE) IP)
  (SETF (GETHASH "MAIL" MAIN-TABLE) MA)
  (SETF (GETHASH "MEMORY-MANAGEMENT" MAIN-TABLE) MM)
  (SETF (GETHASH "NET" MAIN-TABLE) NE)
  (SETF (GETHASH "NET-CONFIG" MAIN-TABLE) NC)
  (SETF (GETHASH "NLMENU" MAIN-TABLE) NL)
  (SETF (GETHASH "PATHNAME" MAIN-TABLE) PA)
  (SETF (GETHASH "PRINTER" MAIN-TABLE) LP)
  (SETF (GETHASH "PROFILE" MAIN-TABLE) PR)
  (SETF (GETHASH "RASTER-TECH" MAIN-TABLE) RA)
  (SETF (GETHASH "RTMS" MAIN-TABLE) RT)
  (SETF (GETHASH "SERIAL" MAIN-TABLE) SE)
  (SETF (GETHASH "STREAMER-TAPE" MAIN-TABLE) ST)
  (SETF (GETHASH "SUGGESTIONS" MAIN-TABLE) SU)
  (SETF (GETHASH "SYS AND SYS2" MAIN-TABLE) SY)
  (SETF (GETHASH "SYSLOG" MAIN-TABLE) SL)
  (SETF (GETHASH "TELNET" MAIN-TABLE) TE)
  (SETF (GETHASH "UCL" MAIN-TABLE) UC)
  (SETF (GETHASH "VT100" MAIN-TABLE) VT)
  (SETF (GETHASH "WINDOW" MAIN-TABLE) WI)
  (SETF (GETHASH "ZWEI" MAIN-TABLE) ZW)
 )

  
(DEFUN ADD-HELP-FUNCTION-2 (PACK DEF NAME FUNCTION
			    &AUX THIS-DEF-TABLE)
  "2Accept a package, definition form (DEFUN, DEFVAR, ...) a function name, and
    the function and set up hash tables that (HELP T) will access.*"
  
  (DECLARE (SPECIAL PACKAGE-TABLE))
  
  (IF (NOT (SETF THIS-DEF-TABLE (GETHASH DEF PACKAGE-TABLE)))
      (THEN
	(SETF THIS-DEF-TABLE
	      (SETF (GETHASH DEF PACKAGE-TABLE) (MAKE-HASH-TABLE :TEST #'EQUAL)))))
  
  (SETF NAME (STRING-APPEND PACK ":" NAME))
  (SETF (GETHASH NAME THIS-DEF-TABLE) FUNCTION))
  
  
  
