;;; -*- Mode:Common-Lisp; Package:METER; Base:10; Patch-file:T -*-
;;; Written 07/14/87 09:19:35 by DMEYER,
;;; while running on BAZ from band LOD3
;;; With IO 3.23, PATHNAME 3.16, FILE 3.8, SYSTEM 3.44, METER 3.1, SERIAL 3.0, CHAOSNET 3.10,
;;;  ETHERNET 3.0, GC 3.9, IMAGEN 3.0, MAIL-DAEMON 3.1, NETWORK-SUPPORT 3.7, PROFILE 3.0,
;;;  SUGGESTIONS 3.3, UCL 3.0, ZWEI 3.12, STREAMER-TAPE 3.5, DEBUG-TOOLS 3.1, FONT-EDITOR 3.1,
;;;  GLOSSARY 3.0, INPUT-EDITOR 3.0, MAIL-READER 3.2, NAMESPACE-EDITOR 3.4, NVRAM 3.1,
;;;  TELNET 3.3, TV 3.9, NAMESPACE 3.7, COMPILER 3.19, PRINTER 3.2, SYSLOG 3.0, VT100 3.5,
;;;  IP 2.44, Experimental BUG 7.0,  microcode 400, Band Name: Release 3.0 - 7/9+IP

;;;                N E W    METER-ANALYZE menuing
;;; This patch makes using meter-analyze much easier, as it provides a primary menu of the
;;; primary things you want to do (SUMMARY, RAW EVENTS, FIND-CALLERS, INTERACTIVE-WINDOW)
;;; and then offers options specific to each, instead of having a single menu with all
;;; options mixed together.  NOTE: This requires the manual to change quite a bit in the
;;; METERING section, as much of it has the old menu and it's options.  -dkm 7/87


#!C
; From file METER.LISP#> METER; MR-X:
#10R METER#:
(COMPILER-LET ((*PACKAGE* (FIND-PACKAGE "METER"))
                          (SI:*LISP-MODE* :COMMON-LISP)
                          (*READTABLE* SYS:COMMON-LISP-READTABLE)
                          (SI:*READER-SYMBOL-SUBSTITUTIONS* SYS::*COMMON-LISP-SYMBOL-SUBSTITUTIONS*))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: METER; METER.#"

(defvar last-meter-analyzer nil) ;analyzer used on last analyze run

(DEFUN meter-analyze ()
  "Menu interface to meter:analyze"
  (DECLARE
   (SPECIAL meter-analyzer meter-sort-function meter-file meter-buffer meter-find-callers
      meter-summarize meter-inclusive-p meter-info-p meter-output))

  (UNLESS (VARIABLE-BOUNDP meter-analyzer)
   ;;Initialize defaults first time around, and remember user options for next time
    (SETQ meter-analyzer :tree
	  meter-sort-function 'max-run-time
	  meter-inclusive-p ()
	  meter-info-p ()
	  meter-find-callers ()
	  meter-summarize ()
	  meter-file ()
	  meter-buffer ()
	  meter-output 'summarize-tree))

  (DO ((loop :repeat))             ;this loop is just in case they abort from an option
      ((neq loop :repeat))         ; so the initial menu comes right back up.
    
    (SETQ meter-info-p (AND meter-info
			    (EQ meter-analyzer last-meter-analyzer))
	  last-meter-analyzer meter-analyzer)

    (SETQ loop (CATCH 'abort-meter-analyze
	        (let ((selection
		       (w:menu-choose
			 '(("Meter Summary"  :value Summary
			      :documentation "Summarize and format Metered data in TREE fashion. (Normally what you want)")
			   ("Interactive Window" :value Inspect
			      :documentation "View meter data dynamically in an interactive Window")
			   ("List Events"     :value Events
			      :documentation "List each meter EVENT as it occured")
			   ("Find Callers"   :value Callers
			      :documentation "Show the callers of specified functions"))
			 :label "Select Meter Analysis Method"
			 :item-alignment :center)))

		  ;; Put up the appropriate CVV window based on what they selected
		  (CASE selection

		   (SUMMARY

		    (WHEN (EQUAL meter-output 'tree-inspect) ;we don't want tree-inspect as an option here 
		      (SETQ meter-output 'summarize-tree))

		    (SETQ meter-analyzer :tree
			  meter-find-callers nil)

		    (w:choose-variable-values
		      `((meter-output
			  "Output type"
			  :DOCUMENTATION ,(FORMAT nil "SUMMARIZE-TREE (what you usually want), ~
                                            TREE-PRINT prints a line for every call, ~
                                            CALL-TREE is TREE-PRINT with indentation.")
			  :MENU ',(LOOP for (FUNCTION . documentation) in *tree-output-functions*
					when (not (member function '(tree-inspect tree-null)))
					collect (LIST function :value function :documentation documentation)))
			(meter-sort-function
			  "Sort function"
			  :DOCUMENTATION "Select how the output is to be sorted (SUMMARIZE-TREE only)"
			  :MENU '(max-real-time max-run-time max-calls
						max-run-time-per-call max-page-faults max-words-consed))
			(meter-inclusive-p
			  "inclusive?"
			  :DOCUMENTATION ,(format nil "If YES, the times for each function include time ~
                                            spent executing subroutines called from the function ~
                                            (SUMMARIZE-TREE only)")
			  :boolean)
			(meter-summarize
			  "Function List"
			  :DOCUMENTATION ,(format nil "If specified, the list of Functions to output ~
                                                       statistics on (SUMMARIZE-TREE only)")
			  :SEXP :EDIT)
			(meter-info-p
			  "Use previous meter-info?"
			  :documentation ,(format nil "Say YES when you haven't metered anything since ~
                                            the last analyze, and you're using the same analyzer")
			  :boolean)
			(meter-file
			  "Output file"
			  :documentation "If specified, write output to this pathname."
			  :pathname-or-nil)
			(meter-buffer
			  "Edit buffer"
			  :documentation "If specified, write output to this zmacs edit buffer name."
			  :string-or-nil))
		      :width 85.
		      :label "Options for METER SUMMARIZE"
		      :margin-choices '(("Abort" (THROW 'abort-meter-analyze :repeat)) "Do It")))
		    
		   (EVENTS

		    (SETQ meter-analyzer :list-events
			  meter-info-p   nil )      ;Temporary. See comment for meter-info-p below.  -dkm 7/87

		    (w:choose-variable-values
		      `(
			;; For some reason, List Events cannot reuse it's meter-info.  So for now
			;; don't offer to reuse previous meter-info, and always set meter-info-p to nil -dkm 7/87
;			(meter-info-p
;			  "Use previous meter-info?"
;			  :documentation ,(format nil "Say YES when you haven't metered anything ~
;                                           since the last analyze, and you're using the same analyzer")
;			  :boolean)

			(meter-file
			  "Output file"
			  :documentation "If specified, write output to this pathname."
			  :pathname-or-nil)
			(meter-buffer
			  "Edit buffer"
			  :documentation "If specified, write output to this zmacs edit buffer name."
			  :string-or-nil))
		      :label "Options for LIST-EVENTS"
		      :margin-choices '(("Abort" (THROW 'abort-meter-analyze :repeat)) "Do It")))
		    
		   
		   (INSPECT
		    (SETQ meter-analyzer :tree
			  meter-output   'tree-inspect))
		   
		   (CALLERS

		    (SETQ meter-analyzer :tree
			  meter-output   'summarize-tree
			  meter-summarize nil)

		    (w:choose-variable-values
		      `(""
			(meter-find-callers
			  "Find Callers of" :DOCUMENTATION "list of functions to find callers of" :list)
			""
			(meter-info-p
			  "Use previous meter-info?"
			  :documentation ,(FORMAT nil "Say YES when you haven't metered anything ~
                                           since the last analyze, and you're using the same analyzer")
			  :BOOLEAN)
			(meter-file
			  "Output file" :documentation "If specified, write output to this pathname."
			  :pathname-or-nil)
			(meter-buffer
			  "Edit buffer" :documentation "If specified, write output to this zmacs edit buffer name."
			  :string-or-nil))
		      :width 85.
		      :label "FIND CALLERS for metered functions"
		      :margin-choices '(("Abort" (THROW 'abort-meter-analyze :repeat))  "Do It"))

		    (UNLESS meter-find-callers
		      (THROW 'abort-meter-analyze :repeat)))
		    
		   (NIL (RETURN-FROM meter-analyze NIL))
		   
		   (OTHERWISE
		    (FORMAT t "~&Cannot handle selected menu item")
		    (RETURN-from meter-analyze nil))))

	       ;; Now ANALYZE based on what was selected.  Lot's of checking is done here to be
	       ;; insure that options which don't makes sense are ignored aren't passed thru.

	       (WHEN (AND meter-summarize meter-find-callers)
		 (BEEP)
		 (FORMAT t "~%Can't both SUMMARIZE and FIND CALLERS on the same analyze.")
		 (RETURN-FROM meter-analyze))

	       (WHEN (AND meter-info (NOT meter-info-p))
		 (analyze-free meter-info))
	       
	       (SETQ meter-info
		     (APPLY #'analyze
			    `(,@(AND meter-file `(:file ,meter-file))
			      ,@(AND meter-buffer `(:buffer ,meter-buffer))
			      ,@(AND meter-info meter-info-p `(:info ,meter-info)) :return t :analyzer
			      ,meter-analyzer
			      ,@(WHEN (EQ meter-analyzer :tree)
				  `(,@(WHEN (NEQ meter-output 'summarize-tree)
					`(:output ,meter-output))
				    ,@(WHEN (AND (NULL meter-summarize)
						 (NULL meter-find-callers)
						 (EQ meter-output 'summarize-tree)
						 (NEQ meter-sort-function 'max-run-time))
					`(:sort-function ,meter-sort-function))
				    ,@(WHEN (AND meter-inclusive-p
						 (EQ meter-output 'summarize-tree)    ;only valid in summarize-tree
						 (NULL meter-summarize)
						 (NULL meter-find-callers))
					'(:inclusive t))
				    ,@(WHEN (AND meter-summarize
						 (EQ meter-output 'summarize-tree))   ;summarize-tree only -dkm 7/87
					`(:summarize ,meter-summarize))
				    ,@(WHEN (AND meter-find-callers
						 (EQ meter-output 'summarize-tree))   ;summarize-tree only -dkm 7/87
					`(:find-callers ,meter-find-callers))
				    
				    )))))
	       

	       (values)))))
    

))
