;;; -*- Mode:Common-Lisp; Package:NET; Base:10 -*-


;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1987, Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1987 Unisys Corporation
;;; All Rights Reserved

(DEFINE-MEDIUM :SERIAL-STREAM
	       :CONNECTION-STEPS '((:NETWORK :SERIAL))
               :MEDIUM-DESIRABILITY .9
               :IMPLEMENTATION-DESIRABILITY .9  
               :CONNECT-FUNCTION 'SERIAL-CONNECTION-FUNCTION
               :CONNECTION-POSSIBLE-P-FUNCTION 'SERIAL-CONNECTION-POSSIBLE-P-FUNCTION
               :SUPERIOR-MEDIUM-LIST '(:BYTE-STREAM))

(DEFINE-STREAM-TYPE :ASCII-TRANSLATING-INPUT-CHARACTER-STREAM 'SYS:SERIAL-ASCII-STREAM :SERIAL-STREAM)

(DEFINE-STREAM-TYPE :ASCII-TRANSLATING-OUTPUT-CHARACTER-STREAM 'SYS:SERIAL-ASCII-STREAM :SERIAL-STREAM)

(DEFINE-STREAM-TYPE :ASCII-TRANSLATING-CHARACTER-STREAM 'SYS:SERIAL-ASCII-STREAM :SERIAL-STREAM)

(DEFINE-STREAM-TYPE :INPUT-CHARACTER-STREAM 'SYS:SERIAL-STREAM :SERIAL-STREAM)

(DEFINE-STREAM-TYPE :OUTPUT-CHARACTER-STREAM 'SYS:SERIAL-STREAM :SERIAL-STREAM)

(DEFINE-STREAM-TYPE :CHARACTER-STREAM 'SYS:SERIAL-STREAM :SERIAL-STREAM)

(DEFINE-STREAM-TYPE :INPUT-BINARY-STREAM 'SYS:SERIAL-STREAM :SERIAL-STREAM)

(DEFINE-STREAM-TYPE :OUTPUT-BINARY-STREAM 'SYS:SERIAL-STREAM :SERIAL-STREAM)

(DEFINE-STREAM-TYPE :BINARY-STREAM 'SYS:SERIAL-STREAM :SERIAL-STREAM)

(DEFVAR *SERIAL-HOST* NIL "The currently connected serial host.")

(DEFVAR *SCRIPT-WINDOW*
	(MAKE-INSTANCE 'TV:WINDOW
			:EDGES  '(94 110 930 480)      ; left,top,right,bottom
			:SAVE-BITS T
			:BORDERS 2
			:LABEL '(:TOP
				  :CENTERED
				  :STRING "Script Status Window"
				  :FONT FONTS:HIGHER-MEDFNB)
			:DEEXPOSED-TYPEOUT-ACTION :PERMIT
			:MORE-P NIL))

(DEFUN SERIAL-CONNECTION-FUNCTION (host logical-contact-name connection &KEY
				   stream-type
				   window-size
				   (timeout (* 10 60))
				   timeout-after-open
				   &allow-other-keys &AUX stream)
  (DECLARE (IGNORE connection window-size timeout timeout-after-open))
  (WHEN (TRANSLATE-LOGICAL-CONTACT-NAME logical-contact-name :SERIAL-STREAM)
    (LET ((vt100-frame (SEND *TERMINAL-IO* :SUPERIOR)))
      (SETQ stream 
	    (MAKE-SERIAL-STREAM :FLAVOR (IF (GET-HOST-ATTRIBUTE host :ASCII-CHARACTERS)
					    'SI:SERIAL-ASCII-STREAM
					    'SI:SERIAL-STREAM)
				:XON-XOFF-PROTOCOL (GET-HOST-ATTRIBUTE host :XON-XOFF-PROTOCOL)
				:EOT-ENABLE (GET-HOST-ATTRIBUTE host :EOT-ENABLE)
				:INPUT-BUFFER-SIZE (GET-HOST-ATTRIBUTE host :INPUT-BUFFER-SIZE)
				:OUTPUT-BUFFER-SIZE (GET-HOST-ATTRIBUTE host :OUTPUT-BUFFER-SIZE)
				:FORCE-OUTPUT (GET-HOST-ATTRIBUTE HOST host :FORCE-OUTPUT)
				:NUMBER-OF-DATA-BITS (GET-HOST-ATTRIBUTE host :NUMBER-OF-DATA-BITS)
				:NUMBER-OF-STOP-BITS (GET-HOST-ATTRIBUTE host :NUMBER-OF-STOP-BITS)
				:PARITY (GET-HOST-ATTRIBUTE host :PARITY)
				:BAUD (GET-HOST-ATTRIBUTE host :BAUD)))
      (SETQ *SERIAL-HOST* host) 
      (SET-IN-INSTANCE vt100-frame 'TELNET:ECHO-FLAG T)
      (SEND vt100-frame :PUTPROP (GET-HOST-ATTRIBUTE host :PROTOCOL-CONVERTER) :PROTOCOL-CONVERTER)
      (IF (GET-HOST-ATTRIBUTE host :SCREEN-BUFFER?)
	  (LET*
	    ((screen-buffer (SEND
			      vt100-frame
			      :PUTPROP 
			      (MAKE-ARRAY (SYMEVAL-IN-INSTANCE vt100-frame 'TELNET:BOTTOM-OF-SCROLL)
					  :LEADER-LENGTH 3
					  :LEADER-LIST (LIST 0 
							     (IF (SYMEVAL-IN-INSTANCE vt100-frame 'TELNET:COLUMN-MODE-80-P)
								 79
								 131)
							     (1- (SYMEVAL-IN-INSTANCE vt100-frame 'TELNET:BOTTOM-OF-SCROLL)))
					  :FILL-POINTER 0
					  :INITIAL-ELEMENT NIL)
			      :SCREEN-BUFFER))
	     (size (ARRAY-TOTAL-SIZE screen-buffer)))
	    (DOTIMES (line size)
	      ;; Create a line array
	      (SETF (AREF screen-buffer line) (MAKE-ARRAY 132.
							  :TYPE :ART-STRING
							  :FILL-POINTER 0
							  :INITIAL-ELEMENT #\SPACE)))
	    (SETF (FILL-POINTER screen-buffer) size))
	  (SEND vt100-frame :remprop :screen-buffer))
      (WHEN (GET-HOST-ATTRIBUTE host :AUTODIAL?)
	(UNLESS (TELNET:AUTODIAL :MENU NIL) 
	  (RETURN-FROM SERIAL-CONNECTION-FUNCTION stream)))
      (WHEN (GET-HOST-ATTRIBUTE host :RUN-SCRIPT?)
	(LET* ((vt100-screen *TERMINAL-IO*)
	       (process (SYMEVAL-IN-INSTANCE vt100-frame 'TELNET:TYPEOUT-PROCESS)))
	  (UNWIND-PROTECT
	      (PROGN
		(SEND *SCRIPT-WINDOW* :CLEAR-SCREEN)
		(SEND process :ARREST-REASON :SCRIPT)
		(SEND *SCRIPT-WINDOW* :SET-SUPERIOR vt100-frame)
		(SEND vt100-frame :SET-SELECTION-SUBSTITUTE *SCRIPT-WINDOW*)
		(TELNET:RUN-SCRIPT-FROM-FILE
		  (GET-HOST-ATTRIBUTE host :SCRIPT)
		  :STREAM SYS:*SERIAL-PORT-OWNER*
		  :DEBUG-STREAM *SCRIPT-WINDOW*
		  :ASCII-TRANSLATE (EQ (FIND-STREAM-TYPE stream-type :SERIAL-STREAM) 'SYS:SERIAL-STREAM))
		(PRINT "****** PRESS A KEY TO CONTINUE ******" *SCRIPT-WINDOW*)
		(READ-CHAR *SCRIPT-WINDOW*))
	    (SEND *SCRIPT-WINDOW* :SET-SUPERIOR W:MAIN-SCREEN)
	    (SEND vt100-frame :SET-SELECTION-SUBSTITUTE vt100-screen)
	    (SEND process :REVOKE-ARREST-REASON :SCRIPT)))))
    stream))

(DEFUN SERIAL-CONNECTION-POSSIBLE-P-FUNCTION (host-a &OPTIONAL (host-b local-host) logical-contact-name)
  (declare (ignore logical-contact-name))	; The rules have changed, so need this argument.
  (AND (EQ host-b LOCAL-HOST)
       (EQ (GET-HOST-ATTRIBUTE host-a :SYSTEM-TYPE) :SERIAL-PORT)))

(DEFVAR *DEFAULT-SERIAL-HOST-NAME*                  "PORT")
(DEFVAR *DEFAULT-SERIAL-HOST-NAMESPACE*             "BOOT")
(DEFVAR *DEFAULT-SERIAL-HOST-NAMESPACE-ACTION*      T)
(DEFVAR *DEFAULT-SERIAL-HOST-BAUD*                  1200)
(DEFVAR *DEFAULT-SERIAL-HOST-FORCE-OUTPUT*          T)
(DEFVAR *DEFAULT-SERIAL-HOST-NUMBER-OF-DATA-BITS*   8)
(DEFVAR *DEFAULT-SERIAL-HOST-NUMBER-OF-STOP-BITS*   2)
(DEFVAR *DEFAULT-SERIAL-HOST-PARITY*                :NONE)
(DEFVAR *DEFAULT-SERIAL-HOST-XON-XOFF-PROTOCOL*     NIL)
(DEFVAR *DEFAULT-SERIAL-HOST-ASCII-CHARACTERS*      NIL)
(DEFVAR *DEFAULT-SERIAL-HOST-INPUT-BUFFER-SIZE*     256) 
(DEFVAR *DEFAULT-SERIAL-HOST-OUTPUT-BUFFER-SIZE*    256)
(DEFVAR *DEFAULT-SERIAL-HOST-AUTODIAL-PREFIX*       "ATDT")
(DEFVAR *DEFAULT-SERIAL-HOST-AUTODIAL-NUMBER*       "9,18005551212")
(DEFVAR *DEFAULT-SERIAL-HOST-AUTODIAL?*             NIL)
(DEFVAR *DEFAULT-SERIAL-HOST-SCRIPT*                NIL)
(DEFVAR *DEFAULT-SERIAL-HOST-RUN-SCRIPT?*           NIL)
(DEFVAR *DEFAULT-SERIAL-HOST-PROTOCOL-CONVERTER*    :NONE)
(DEFVAR *DEFAULT-SERIAL-HOST-SCREEN-BUFFER?*        NIL)

(DEFUN DEFINE-SERIAL-HOST (&KEY (name *DEFAULT-SERIAL-HOST-NAME*)
			        (namespace *DEFAULT-SERIAL-HOST-NAMESPACE*)
				(namespace-action *DEFAULT-SERIAL-HOST-NAMESPACE-ACTION*)
				(baud *DEFAULT-SERIAL-HOST-BAUD*)
				(force-output *DEFAULT-SERIAL-HOST-FORCE-OUTPUT*)
				(number-of-data-bits *DEFAULT-SERIAL-HOST-NUMBER-OF-DATA-BITS*)
				(number-of-stop-bits *DEFAULT-SERIAL-HOST-NUMBER-OF-STOP-BITS*)
				(parity *DEFAULT-SERIAL-HOST-PARITY*)
				(xon-xoff-protocol *DEFAULT-SERIAL-HOST-XON-XOFF-PROTOCOL*)
				(ascii-characters *DEFAULT-SERIAL-HOST-ASCII-CHARACTERS*)
				(input-buffer-size *DEFAULT-SERIAL-HOST-INPUT-BUFFER-SIZE*)
				(output-buffer-size *DEFAULT-SERIAL-HOST-OUTPUT-BUFFER-SIZE*)
				(autodial-prefix *DEFAULT-SERIAL-HOST-AUTODIAL-PREFIX*)
				(autodial-number *DEFAULT-SERIAL-HOST-AUTODIAL-NUMBER*)
				(autodial? *DEFAULT-SERIAL-HOST-AUTODIAL?*)
				(script *DEFAULT-SERIAL-HOST-SCRIPT*)
				(run-script? *DEFAULT-SERIAL-HOST-RUN-SCRIPT?*)
				(protocol-converter *DEFAULT-SERIAL-HOST-PROTOCOL-CONVERTER*)
				(screen-buffer? *DEFAULT-SERIAL-HOST-SCREEN-BUFFER?*))
  "Define a named serial host in a specified namespace.  Returns the new host object.
:NAME  The name of the host being defined as requiring this particular configuration of a serial port.
:NAMESPACE  The name of the namespace in which to define this host.
:NAMESPACE-ACTION  T when the host is to be stored only in the local cache of the specified namespace;
                   NIL when the host is to be stored in the local cache and the Name Server of the namespace.
:BAUD  A serial line speed of 300, 1200, 2400, 4800, 9600, or 19200.
:FORCE-OUTPUT  T to transmit characters immediately;  NIL to transmit characters only when the buffer is full.
:NUMBER-OF-DATA-BITS  5, 6, 7, or 8 data bits per character.
:NUMBER-OF-STOP-BITS  1 or 2 stop bits per character.
:PARITY  :ODD, :EVEN, or :NONE.
:XON-XOFF-PROTOCOL  T to use XON-XOFF flow control.
:ASCII-CHARACTERS  T to automatically translate between ASCII and Explorer characters.
:INPUT-BUFFER-SIZE  Size (in words) to allocate for the input buffer.
:OUTPUT-BUFFER-SIZE  Size (in words) to allocate for the output buffer.
:AUTODIAL-PREFIX  Character sequence that activates autodial capabilities of autodial modem.
:AUTODIAL-NUMBER  Telephone number to dial.  Use a comma to insert a 1 second delay.
:Autodial?  T to automatically autodial upon connection.
:SCRIPT  The pathname of a file containing a script defined per TELNET:RUN-SCRIPT.
:RUN-SCRIPT?  T to automatically run the script upon connection.
:PROTOCOL-CONVERTER  Type of Protocol Converter connected to serial port (for use by SNA-COMM package).
                     Enter :NONE, :COMMTEX-CX-80, or :IBM-3708.
:SCREEN-BUFFER  T maintains a local character buffer for VT100 screen (required with SNA-COMM)."

  (NAME:ADD-OBJECT name
		   :host
		   :attributes (list :SYSTEM-TYPE :SERIAL-PORT
				     :SERVICES '((:LOGIN :SERIAL-STREAM :TELNET))
				     :BAUD baud
				     :FORCE-OUTPUT force-output
				     :NUMBER-OF-DATA-BITS number-of-data-bits
				     :NUMBER-OF-STOP-BITS number-of-stop-bits
				     :PARITY parity
				     :XON-XOFF-PROTOCOL xon-xoff-protocol
				     :ASCII-CHARACTERS ascii-characters
				     :INPUT-BUFFER-SIZE input-buffer-size
				     :OUTPUT-BUFFER-SIZE output-buffer-size
				     :AUTODIAL-PREFIX autodial-prefix
				     :AUTODIAL-NUMBER autodial-number
				     :AUTODIAL? autodial?
				     :SCRIPT script
				     :RUN-SCRIPT? run-script?
				     :PROTOCOL-CONVERTER protocol-converter
				     :SCREEN-BUFFER? screen-buffer?)
		   :NAMESPACE namespace
		   :LOCAL namespace-action)
  (PARSE-HOST name))

(DEFUN DEFINE-SERIAL-HOST-FROM-CVV ()
  "Produces a CVV to select serial host attributes, then creates a serial
host object using DEFINE-HOST.  Returns the new host object."
  (LET ((name *DEFAULT-SERIAL-HOST-NAME*)
	(namespace *DEFAULT-SERIAL-HOST-NAMESPACE*)
	(namespace-action *DEFAULT-SERIAL-HOST-NAMESPACE-ACTION*)
	(baud *DEFAULT-SERIAL-HOST-BAUD*)
	(force-output *DEFAULT-SERIAL-HOST-FORCE-OUTPUT*)
	(number-of-data-bits *DEFAULT-SERIAL-HOST-NUMBER-OF-DATA-BITS*)
	(number-of-stop-bits *DEFAULT-SERIAL-HOST-NUMBER-OF-STOP-BITS*)
	(parity *DEFAULT-SERIAL-HOST-PARITY*)
	(xon-xoff-protocol *DEFAULT-SERIAL-HOST-XON-XOFF-PROTOCOL*)
	(ascii-characters *DEFAULT-SERIAL-HOST-ASCII-CHARACTERS*)
	(input-buffer-size *DEFAULT-SERIAL-HOST-INPUT-BUFFER-SIZE*)
	(output-buffer-size *DEFAULT-SERIAL-HOST-OUTPUT-BUFFER-SIZE*)
	(autodial-prefix *DEFAULT-SERIAL-HOST-AUTODIAL-PREFIX*)
	(autodial-number *DEFAULT-SERIAL-HOST-AUTODIAL-NUMBER*)
	(autodial? *DEFAULT-SERIAL-HOST-AUTODIAL?*)
	(script *DEFAULT-SERIAL-HOST-SCRIPT*)
	(run-script? *DEFAULT-SERIAL-HOST-RUN-SCRIPT?*)
	(protocol-converter *DEFAULT-SERIAL-HOST-PROTOCOL-CONVERTER*)
	(screen-buffer? *DEFAULT-SERIAL-HOST-SCREEN-BUFFER?*))
    
    (DECLARE (SPECIAL name namespace namespace-action baud force-output
                      number-of-data-bits number-of-stop-bits parity
                      xon-xoff-protocol ascii-characters input-buffer-size
                      output-buffer-size autodial-prefix autodial-number
                      autodial? script run-script? protocol-converter
		      screen-buffer?))

    (CATCH 'ABORT-CVV
      (W:CHOOSE-VARIABLE-VALUES
	'((name
	    "Host Name"
	    :DOCUMENTATION "The name of the host being defined as having this particular configuration of a serial port."
	    :STRING)
	  (namespace
	    "Namespace"
	    :DOCUMENTATION "The namespace in which to define this host."
	    :MENU (NAME:LIST-KNOWN-NAMESPACES))
	  (namespace-action
	    "Namespace Action"
	    :DOCUMENTATION "Choose LOCAL to update only the local cache of the specified namespace.
Choose GLOBAL to update the local cache and the Name Server of the specified namespace."
	    :MENU-ALIST (("LOCAL" T) ("GLOBAL" NIL)))
	  (baud
	    "Baud rate"
	    :DOCUMENTATION "Line speed.  (Most asynchronous modems use 1200 or 300)"
	    :CHOOSE (300 1200 2400 4800 9600 19200))
	  (force-output
	    "Force output"
	    :DOCUMENTATION "YES: send characters immediately.  NO: send characters when buffer is full."
	    :BOOLEAN)
	  (number-of-data-bits
	    "Data Bits"
	    :DOCUMENTATION "Number of data bits."
	    :CHOOSE (5 6 7 8))
	  (number-of-stop-bits
	    "Stop Bits"
	    :DOCUMENTATION "Number of stop bits."
	    :CHOOSE (1 2))
	  (parity
	    "Parity"
	    :DOCUMENTATION "Type of parity to use."
	    :CHOOSE (:NONE :EVEN :ODD))
	  (xon-xoff-protocol
	    "XON-XOFF"
	    :DOCUMENTATION "YES: use XON-XOFF characters.  NO: don't implement XON-XOFF characters."
	    :BOOLEAN)
	  (ascii-characters
	    "Translate ASCII"
	    :DOCUMENTATION "YES: Automatically translate between ASCII and Explorer characters.  NO: don't translate."
	    :BOOLEAN)
	  (input-buffer-size
	    "Input Buffer size"
	    :DOCUMENTATION "Size (in words) to allocate for the input buffers."
	    :NUMBER)
	  (output-buffer-size
	    "Output Buffer size"
	    :DOCUMENTATION "Size (in words) to allocate for the output buffers."
	    :NUMBER)
	  (autodial-prefix
	    "Autodial Prefix"
	    :DOCUMENTATION "Character sequence that activates autodial capabilities of autodial modem."
	    :STRING)
	  (autodial-number
	    "Phone Number"
	    :DOCUMENTATION "Telephone number to dial."
	    :STRING)
	  (autodial?
	    "Autodial?"
	    :DOCUMENTATION "YES: Automatically autodial upon connection.  NO: don't autodial."
	    :BOOLEAN)
	  (script
	    "Script"
	    :DOCUMENTATION "The pathname of a file containing a script defined per TELNET:RUN-SCRIPT."
	    :TYPEP STRING)
	  (run-script?
	    "Run Script?"
	    :DOCUMENTATION "YES: Automatically run the script upon connection.  NO: don't run the script."
	    :BOOLEAN)
	  (protocol-converter
	    "Protocol Converter"
	    :DOCUMENTATION "Type of Protocol Converter connected to serial port."
	    :CHOOSE (:NONE :IBM-3708 :COMMTEX-CX-80))
	  (screen-buffer?
	    "Screen Buffer?"
	    :DOCUMENTATION "Select YES to maintain a local character buffer for VT100 screen (required with SNA)."
	    :BOOLEAN))
	:NEAR-MODE '(:POINT 500 400)
	:LABEL "Define Serial Host"
	:MARGIN-CHOICES '(("Abort" (THROW 'ABORT-CVV NIL))
			  "Do It")))
      
      (DEFINE-SERIAL-HOST :NAME name
	                  :NAMESPACE namespace
			  :NAMESPACE-ACTION namespace-action
	                  :BAUD baud
			  :FORCE-OUTPUT force-output
			  :NUMBER-OF-DATA-BITS number-of-data-bits
			  :NUMBER-OF-STOP-BITS number-of-stop-bits
			  :PARITY parity
			  :XON-XOFF-PROTOCOL xon-xoff-protocol
			  :ASCII-CHARACTERS ascii-characters
			  :INPUT-BUFFER-SIZE input-buffer-size
			  :OUTPUT-BUFFER-SIZE output-buffer-size
			  :AUTODIAL-PREFIX autodial-prefix
			  :AUTODIAL-NUMBER autodial-number
			  :AUTODIAL? autodial?
			  :SCRIPT script
			  :RUN-SCRIPT? run-script?
			  :PROTOCOL-CONVERTER protocol-converter
			  :SCREEN-BUFFER? screen-buffer?)))


