;;; -*- Mode:Common-Lisp; Package:TELNET; Base:10; Patch-file:T -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1987, Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1987 Unisys Corporation
;;; All Rights Reserved

(DEFMETHOD (VT100-ESCAPE-SEQUENCE-MIXIN :AUXILIARY-UP) ()
  (SEND SELF :NET-OUTPUT #\ESCAPE)
  (SEND SELF :NET-OUTPUT #\[)
  (SEND SELF :NET-OUTPUT #\A))

(DEFMETHOD (VT100-ESCAPE-SEQUENCE-MIXIN :AUXILIARY-DOWN) ()
  (SEND SELF :NET-OUTPUT #\ESCAPE)
  (SEND SELF :NET-OUTPUT #\[)
  (SEND SELF :NET-OUTPUT #\B))

(DEFMETHOD (VT100-ESCAPE-SEQUENCE-MIXIN :AUXILIARY-RIGHT) ()
  (SEND SELF :NET-OUTPUT #\ESCAPE)
  (SEND SELF :NET-OUTPUT #\[)
  (SEND SELF :NET-OUTPUT #\C))

(DEFMETHOD (VT100-ESCAPE-SEQUENCE-MIXIN :AUXILIARY-LEFT) ()
  (SEND SELF :NET-OUTPUT #\ESCAPE)
  (SEND SELF :NET-OUTPUT #\[)
  (SEND SELF :NET-OUTPUT #\D))

(DEFMETHOD (VT100-ESCAPE-SEQUENCE-MIXIN :PROCESS-ESCAPE) (CH)
  (COND
    (PROCESS-CH?
     (IF (= CH NVT-IAC)
	 (SEND SELF :HANDLE-IAC)
	 (SEND SELF :TEST-FOR-ESCAPE-SEQUENCE CH)))
    (T (SEND SELF :BUFFERED-TYO CH))))

(DEFMETHOD (VT100-ESCAPE-SEQUENCE-MIXIN :TEST-FOR-ESCAPE-SEQUENCE) (CH)
  (COND
    (LARGE-CHARACTERS			   ;Test for the large size character set
     (COND
       ((OR (= CH ESC) (= CH CR) (= CH LF))
	(SEND VT100-PANE :SET-CURRENT-FONT SAVE-FONT)
	(SETQ LARGE-CHARACTERS ())))))
  (MULTIPLE-VALUE-BIND (CURRENT-X CURRENT-Y)
      (SEND VT100-PANE :READ-CURSORPOS :CHARACTER)
      ;   (cond ((and (>= current-y bottom-of-scroll)	;Is this the bottom of scroll region?
      ;		(<= current-y (+ bottom-of-scroll 1))   ; Done just once per line
      ;		(= current-x 0.))	   
      ;	   (setf (tv:sheet-end-page-flag vt100-pane) 0.)
      ;	   (sheet-scroll-up vt100-pane new-line)))
    (COND
      ((OR (< CH ENQ) (= CH VT)) NIL)
      (ESCAPE-FLAG			                ; Yes, we got an esc
       (SETF ch (LDB #o0007 ch))                        ;Strip off the parity bit if it is set. 
       (SEND SELF :PROCESS-ESCAPE-SEQUENCE CH))
      ((= CH ESC)			                ; Is it 'ESC'
       (SETQ ESCAPE-FLAG T))
      (T				                ; The char isn't in an escape sequence
       (SEND SELF :PROCESS-CHARACTER CH CURRENT-X CURRENT-Y)))))

(DEFMETHOD (VT100-FRAME :AFTER :INIT) (&REST IGNORE)
  (SETQ VT100-PANE (SEND SELF :GET-PANE 'VT100-PANE))
  (WINDOW-BACKGROUND-COLOR VT100-PANE VT100-DEFAULT-SCREEN-COLOR)
  ;(SETQ LED-PANE (SEND SELF :GET-PANE 'LED-PANE))
  (SETQ AUX-KEYPAD-PANE (SEND SELF :GET-PANE 'KEYPAD))
  (SETQ TELNET-PANE VT100-PANE)
  (SEND SELF :SELECT-PANE VT100-PANE)
  (SETQ VT100-WINDOW SELF)
  (SETQ VT100-SCREEN VT100-PANE)
  ;(SETQ LED-SCREEN LED-PANE)
  (SETQ TERMINAL-STREAM VT100-PANE)
  (SETQ OUTPUT-BUFFER ())
  ;; we don't want buffering
  (SEND SELF :INITIAL-SETUP)
  (SETF (TV:SHEET-TRUNCATE-LINE-OUT-FLAG VT100-PANE) 1)   ;TURN ON TRUNCATION
  (SEND typeout-process :SET-PRIORITY 10.))

(DEFMETHOD (VT100-FRAME :NET-OUTPUT) (CH)
  (LOCK-OUTPUT
    (COND ((AND (NOT (LDB-TEST TELNET-KEYS-BYTE-SPEC CH))
		(LDB-TEST (BYTE 1 7.) CH)
		(NOT (TYPEP STREAM 'SYS:SERIAL-STREAM)))    ;;don't do this for a serial stream
	   ;; reverse cases since this is a meta character.
	   (SETQ CH (MOD CH 128.))
	   (COND ((< CH #.(CHAR-CODE #\a)) ; #. (char-code #\a) is same as #o141
		  (SETQ CH (CHAR-DOWNCASE CH)))
		 
		 (T
		   (SETQ CH (CHAR-UPCASE CH)))))
	  
	  ((LDB-TEST TELNET-KEYS-BYTE-SPEC CH)	   ;An NVT char from TELNET-KEYS table
	   (IF NEW-TELNET-P
	       (SEND STREAM :TYO NVT-IAC))
	   (SETQ CH (LDB (BYTE 8. 0) CH))))
    (SEND STREAM :TYO CH)
    (COND ((AND (= CH CR) (NOT (TYPEP STREAM 'SYS:SERIAL-STREAM)))  ;;just send the CR to a serial stream
	   (SEND STREAM :TYO LF))	   ;Carrage Return is two chars, CR LF
	  ((AND (= CH NVT-IAC)
		NEW-TELNET-P)
	   (SEND STREAM :TYO NVT-IAC)))))

(DEFCOMMAND (VT100-FRAME :DEFINE-SERIAL-HOST) ()
  '(:DESCRIPTION "Display a CVV menu to add a serial host to a namespace."
    :NAMES ("Define Serial Host"))
  (NET:DEFINE-SERIAL-HOST-FROM-CVV))

(DEFCOMMAND (VT100-FRAME :AUTODIAL) ()
  '(:DESCRIPTION "Display a pop-up menu to dial a phone number using an autodialer."
		 :NAMES ("Autodial"))
  (COND (STREAM
	 (FUNCALL 'AUTODIAL :STREAM STREAM :MENU T))
	(T
	 (FORMAT T "~&Not connected.  Can't Autodial.")
	 (WHEN (NOT UCL:PREEMPTING?)
	  (SEND SELF ':HANDLE-PROMPT)))))


(DEFCOMMAND (VT100-FRAME :KERMIT) ()
  '(:DESCRIPTION "Display a pop-up menu of KERMIT file-transfer commands."
		 :Names ("Kermit"))
  (DECLARE (SPECIAL KERMIT:*KERMIT-FRAME*))
  (COND (STREAM
	 (LET
	   ((VT100-SUBSTITUTE (SEND SELF :SELECTION-SUBSTITUTE))
	    (KERMIT-SUPERIOR (SEND KERMIT:*KERMIT-FRAME* :SUPERIOR))
	    ;(MENU-PANE (SEND SELF :GET-PANE 'MENU-TELNET))
	    )
	   (UNWIND-PROTECT
	       (LET
		 ((FORM NIL))
		 (SEND TYPEOUT-PROCESS :ARREST-REASON 'KERMIT)	; Stop the vt100 process from using serial stream
		 (SETQ FORM (KERMIT:INTERACTIVE-KERMIT STREAM NIL)) ; Get the Kermit arguments
		 (WHEN FORM
		   ;(SETF (SEND MENU-PANE :INVISIBLE-TO-MOUSE-P) T)	; Make the vt100 menu items non-mousable
		   (SEND KERMIT:*KERMIT-FRAME* :SET-SUPERIOR SELF)
		   (SEND SELF :SET-SELECTION-SUBSTITUTE KERMIT:*KERMIT-FRAME*)	; Attach the kermit frame to vt100
		   (EVAL FORM)))		; Call Kermit
	     (SEND TYPEOUT-PROCESS :REVOKE-ARREST-REASON 'KERMIT)	; Reallow vt100 to use serial
	     ;(SETF (SEND MENU-PANE :INVISIBLE-TO-MOUSE-P) NIL)	; Make menu items mousable
	     (SEND SELF :SET-SELECTION-SUBSTITUTE VT100-SUBSTITUTE)
	     (SEND KERMIT:*KERMIT-FRAME* :SET-SUPERIOR KERMIT-SUPERIOR))))
	(T
	 (FORMAT T "~&Not connected.  Can't run KERMIT.")
	 (WHEN (NOT UCL:PREEMPTING?)
	   (SEND SELF ':HANDLE-PROMPT)))))

(DEFCOMMAND (vt100-frame :LOCAL-ECHO-COMMAND) ()
  '(:DESCRIPTION "Toggle local echo mode of Vt100 screen pane."
		 :NAMES ("Local Echo"))
  (SETF ECHO-FLAG (IF ECHO-FLAG NIL T))
  (FORMAT T "~&Local echo now ~A.~%"
	  (IF ECHO-FLAG "off" "on"))                  ; echo-flag=T means local echo is off!
  (WHEN (AND (NULL STREAM) (NOT UCL:PREEMPTING?))
      (SEND SELF ':HANDLE-PROMPT)))

(DEFCOMMAND (VT100-FRAME :LONG-BREAK) ()
  '(:DESCRIPTION "Send a short break to stream."
    :NAMES ("Long Break")
    :KEYS ((#\NETWORK #\CTRL-BREAK)))
  (IF STREAM
      (SEND STREAM :SEND-LONG-BREAK)
      (FORMAT T "~&Not connected. Can't send Long Break.")
      (WHEN (NOT UCL:PREEMPTING?)
	(SEND SELF :HANDLE-PROMPT))))

(DEFCOMMAND (VT100-FRAME :SHORT-BREAK) ()
  '(:DESCRIPTION "Send a short break to stream."
    :NAMES ("Short Break")
    :KEYS ((#\NETWORK #\BREAK)))
  (IF STREAM
      (SEND STREAM :SEND-SHORT-BREAK)
      (FORMAT T "~&Not connected. Can't send Short Break.")
      (WHEN (NOT UCL:PREEMPTING?)
	(SEND SELF :HANDLE-PROMPT))))


;;; Redefine the VT100 layout and menu
;;;

(DEFFLAVOR VT100-TELNET-MENU
	   (TV:INVISIBLE-TO-MOUSE-P)
	   (TV:DYNAMIC-ITEM-LIST-MIXIN TV:COMMAND-MENU)
  (:SETTABLE-INSTANCE-VARIABLES TV:INVISIBLE-TO-MOUSE-P)
  (:DEFAULT-INIT-PLIST 
    :LABEL (LIST :TOP :FONT FONTS:HL12B :STRING "VT100 & Telnet Commands")
    :ROWS 4
    :COLUMNS 5
    :VSP 8.
    :FONT-MAP (LIST FONTS:MEDFNT)
    :LABEL-BOX-P NIL
    :ITEM-LIST NIL)
  (:DOCUMENTATION :COMBINATION
    "Command menu needs dynamic-item-list-mixin for UCL."))

; moves the commands into a pop-up menu 
(DEFCOMMAND VT100-TELNET-COMMANDS
	    nil
	    '(:keys #\mouse-M-1
		    :documentation "Bring up the menu of Telnet & VT100 Commands"
		    :names ("VT100 & Telnet Command Menu"))
  (pop-up-vt100-telnet-command-menu 'TELNET:UCL-VT100-TELNET-MENU))

(BUILD-COMMAND-TABLE 'VT100-TELNET-CMD-TABLE 'VT100-FRAME
  '((:method telnet-frame :exit-command)
    vt100-telnet-commands                    ;CRW 3-21-88
    (:method telnet-frame :disconnect-command)
    (:method telnet-frame :interrupt-process-command)
    :send-answerback-command
    :reverse-video-command
    :reset-command
    :escape-processing-command
    (:method telnet-frame :quit-and-disconnect-command)
    (:method telnet-frame :status-command)
    (:method telnet-frame :abort-output-command)
    :column-command
    :truncate-command
    :set-vt100-lines
    :network-help-command
    (:method telnet-frame :clear-input-command)
    :define-serial-host
    :autodial
    :kermit
    :local-echo-command
    :short-break
    :long-break
    )
  :INIT-OPTIONS
  '(:NAME "Vt100 & Telnet Commands"
	  :DOCUMENTATION "The Vt100 & Telnet commands."))


(BUILD-MENU 'UCL-VT100-TELNET-MENU 'VT100-FRAME
  :DEFAULT-ITEM-OPTIONS '(:label " VT100 & Telnet Commands "
				 :item-alignment :center					  
				 :FONT FONTS:MEDFNT)	    
  
  :ITEM-LIST-ORDER  
    '((:method telnet-frame :exit-command)		
      (:method telnet-frame :disconnect-command)
      (:method telnet-frame :interrupt-process-command)
      :send-answerback-command
      :reverse-video-command
      :reset-command
      :escape-processing-command
      (:method telnet-frame :quit-and-disconnect-command)
      (:method telnet-frame :status-command)
      (:method telnet-frame :abort-output-command)
      :column-command
      :truncate-command
      :set-vt100-lines
      :network-help-command
      :define-serial-host
      :autodial
      :kermit
      :local-echo-command
      :short-break
      :long-break
      ))


; moves the commands into a pop-up menu 
(DEFCOMMAND VT100-TELNET-COMMANDS
	    nil
	    '(:keys #\mouse-M-1
		    :documentation "Bring up the menu of Telnet & VT100 Commands"
		    :names ("VT100 & Telnet Command Menu"))
  (pop-up-vt100-telnet-command-menu 'TELNET:UCL-VT100-TELNET-MENU))


(BUILD-COMMAND-TABLE 'VT100-TELNET-CMD-TABLE 'VT100-FRAME
  '((:method telnet-frame :exit-command)
    vt100-telnet-commands                       ;CRW 3-21-88
    (:method telnet-frame :disconnect-command)
    (:method telnet-frame :interrupt-process-command)
    :send-answerback-command
    :reverse-video-command
    :reset-command
    :escape-processing-command
    (:method telnet-frame :quit-and-disconnect-command)
    (:method telnet-frame :status-command)
    (:method telnet-frame :abort-output-command)
    :column-command
    :truncate-command
    :set-vt100-lines
    :network-help-command
    (:method telnet-frame :clear-input-command)
    :define-serial-host
    :autodial
    :kermit
    :local-echo-command
    :short-break
    :long-break
    )
  :INIT-OPTIONS
  '(:NAME "Vt100 & Telnet Commands"
	  :DOCUMENTATION "The Vt100 & Telnet commands."))









