;;; -*- Mode:Common-Lisp; Package:DEMO; Fonts:(MEDFNT HL12B HL12BI); Base:10 -*-
;;; -*- Mode:Common-Lisp; Package:(DEMO :use (lisp ticl)); Fonts:(MEDFNT HL12B HL12BI); Base:10 -*-

;1;; Old low-level sound access routines from Austin*

(eval-when (compile load eval)
  (IMPORT tv:'(def-beep-function defsound do-sound delay reset-sound tone volume noise wait with-sound-enabled))
  )

;;; The following constants relate to the sound register address field.
(DEFCONSTANT TONE-1            0)                ;  set up tone register 1
(DEFCONSTANT ATTENUATION-1     1)                ;  2 db Attenuation
(DEFCONSTANT TONE-2            2)                ;  set up tone register 2
(DEFCONSTANT ATTENUATION-2     3)                ;  6 db Attenuation
(DEFCONSTANT TONE-3            4)                ;  set up tone register 3
(DEFCONSTANT ATTENUATION-3     5)                ; 10 db Attenuation
(DEFCONSTANT NOISE-CONTROL     6)
(DEFCONSTANT NOISE-ATTENUATION 7)

(DEFCONSTANT ATTENUATE-SILENT  #o17)             ; turn it off. Man!
(DEFCONSTANT ATTENUATE-LOUD    0)                ; Full Blast !!!

(DEFUN WRITE-SOUND-REG (REG VAL)
  (do-sound (DPB REG #o0403		   ; set up selected register
		 (DPB 1 #o0701		   ; set sound first word data bit 
		      (LDB #o0004 VAL))))) ; write first word

(DEFUN WRITE-TONE (REG VAL)
  (WRITE-SOUND-REG REG VAL)
  (do-sound (LDB #o0406 VAL)))

(DEFUN ATTENUATION-OFF (&OPTIONAL (REG1 ATTENUATE-LOUD)
                           (REG2 ATTENUATE-LOUD)
                           (REG3 ATTENUATE-LOUD))

  (WRITE-SOUND-REG ATTENUATION-1 REG1)
  (WRITE-SOUND-REG ATTENUATION-2 REG2)
  (WRITE-SOUND-REG ATTENUATION-3 REG3)
 )

(DEFUN ATTENUATION-ON (&OPTIONAL (REG1 ATTENUATE-SILENT)
                          (REG2 ATTENUATE-SILENT)
                          (REG3 ATTENUATE-SILENT))

  (WRITE-SOUND-REG ATTENUATION-1 REG1)
  (WRITE-SOUND-REG ATTENUATION-2 REG2)
  (WRITE-SOUND-REG ATTENUATION-3 REG3)
 )


;;; Enter:
;;; The Home Audio Entertainment Center.

;;; Notes are HZ representations of the the sound.
;;; These are convertable to half-wavelength by this macros.

(Defmacro Convert-Note (note)
  `(floor (quotient 2048000. (float (* 32. (max 4. (min 16000. ,note)))))))

(DefMacro Attenuate-Reg-Of-Voice (Voice)
  `(1- (* ,Voice 2)))

(DEFCONSTANT Twelfth-Root (expt 2.0 (/ 1.0 12.0)))

(Defun Octave (note)
  (* note 2.0))

(Defun MAJOR-SECOND (note)
  (floor (* note Twelfth-Root Twelfth-Root)))

(Defun MAJOR-THIRD (note)
  (floor (* note (expt Twelfth-Root 4.0))))

(Defun MAJOR-FOURTH (note)
  (floor (* note (expt Twelfth-Root 5.0))))

(Defun MAJOR-FIFTH (note)
  (floor (* note (expt Twelfth-Root 7.0))))

(Defun MAJOR-SIXTH (note)
  (floor (* note (expt Twelfth-Root 9.0))))

(Defun MAJOR-SEVENTH (note)
  (floor (* note (expt Twelfth-Root 11.0))))

;;; Minor Scale
(Defun MINOR-SECOND (Note)
  (* note Twelfth-Root))

(Defun MINOR-THIRD (note)
  (* note (expt Twelfth-Root 3.0)))

(Deff MINOR-FOURTH #'MAJOR-FOURTH)

(Defun MINOR-FIFTH (note)
  (* note (expt Twelfth-Root 6.0)))

(Defun MINOR-SIXTH (note)
  (* note (expt Twelfth-Root 8.0)))

(Defun MINOR-SEVENTH (note)
  (* note (expt Twelfth-Root 10.0)))


(Defun Major-Chord (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (Major-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Major-Fifth Tonic))))

(Defun Major-Chord2 (Tonic)
  (Write-Tone TONE-1 (Convert-Note (Octave Tonic)))
  (Write-Tone TONE-2 (Convert-Note (Major-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Major-Fifth Tonic))))

(Defun Minor-Chord (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (Minor-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Major-Fifth Tonic))))

(Defun Minor-Chord2 (Tonic)
  (Write-Tone TONE-1 (Convert-Note (Octave Tonic)))
  (Write-Tone TONE-2 (Convert-Note (Minor-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Major-Fifth Tonic))))

(Defun Major-Seventh-Chord (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (Major-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Major-Seventh  Tonic))))

(Defun Major-Seventh-Chord2 (Tonic)
  (Write-Tone TONE-1 (Convert-Note (Octave Tonic)))
  (Write-Tone TONE-2 (Convert-Note (Major-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Major-Seventh  Tonic))))

(Defun Minor-Seventh-Chord (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (minor-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (minor-Seventh  Tonic))))

(Defun Minor-Seventh-Chord2 (Tonic)
  (Write-Tone TONE-1 (Convert-Note (Octave Tonic)))
  (Write-Tone TONE-2 (Convert-Note (minor-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (minor-Seventh  Tonic))))

(Defun Dominant-Seventh (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (Major-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Minor-Seventh Tonic))))

(Defun Dominant-Seventh2 (Tonic)
  (Write-Tone TONE-1 (Convert-Note (Octave Tonic)))
  (Write-Tone TONE-2 (Convert-Note (Major-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Minor-Seventh Tonic))))

(Defun Diminished-Chord (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (Minor-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Minor-Fifth Tonic))))

(Defun Diminished-Chord2 (Tonic)
  (Write-Tone TONE-1 (Convert-Note (Octave Tonic)))
  (Write-Tone TONE-2 (Convert-Note (Minor-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Minor-Fifth Tonic))))

(Defun Diminished-Seventh (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (Minor-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (major-sixth Tonic))))

(Defun Diminished-Seventh2 (Tonic)
  (Write-Tone TONE-1 (Convert-Note (octave Tonic)))
  (Write-Tone TONE-2 (Convert-Note (Minor-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (major-sixth Tonic))))

(Defun Augmented-Chord  (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (Major-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Minor-Sixth Tonic))))

(Defun Augmented-Chord2 (Tonic)
  (Write-Tone TONE-1 (Convert-Note (Octave Tonic)))
  (Write-Tone TONE-2 (Convert-Note (Major-Third Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Minor-Sixth Tonic))))

(Defun Dual-1-4 (Tonic)
  (Write-Tone TONE-1 (Convert-Note Tonic))
  (Write-Tone TONE-2 (Convert-Note (Major-Fourth Tonic)))
  (Write-Tone TONE-3 (Convert-Note (Octave Tonic))))  


#| ;; Nobody uses this stuff
;;; Note Profiles

;;; These are control the envelope shape of the notes
;;; Since these are made to be combined each is represented by a list
;;; of sound modification steps, consistings of a function to run.
;;; Each function returns the time it took to run (in microseconds) and
;;; the length of time (in microseconds) until the next function should
;;; run and the arguments that should be passed on.

;;; Simple on-off sound -- Nothing Fancy

(DefProp Solid Note-Profile (Note-Start Note-End))
(DefProp Solid Note-Profile-Args
         (Sustain-Length Sustain-Attenuation Space-Length))

(Defun Note-Start (Voice Args)
  (Let ((Length (Car Args))
        (Attenuation (Cadr Args)))
    (Write-Sound-Reg (Attenuate-Reg-Of-Voice Voice) Attenuation)
    (Values 10. Length (Cddr Args))))

(Defun Note-End (Voice Args)
  (Let ((Length (Car Args)))
    (Write-Sound-Reg (Attenuate-Reg-Of-Voice Voice) ATTENUATE-SILENT)
    (Values 6. Length (Cdr Args))))

;;; Like SOLID but does not end so can continue into the next group of notes
(DefProp Solid-Glue (Note-Start) Note-Profile)
(DefProp Solid-Glue (Sustain-Length Sustain-Attenuation) Note-Profile-Args)

(DefProp Solid-Glue-Continue (Note-Continue)   Note-Profile)
(DefProp Solid-Glue-Continue (Continue-Length) Note-Profile-Args)

;;; For finishing a note that is continued from a -Glue start
(DefProp Solid-Glue-End (Note-Continue Note-End) Note-Profile)
(DefProp Solid-Glue-End (Continue-Length Space-Length) Note-Profile-Args)

(Defun Note-Continue (Voice Args)
  (Values 10. (Car Voice) (Cdr Args)))

;;; Profiles for Simple Attack envelope.  This note starts at some attenuation,
;;; then goes to a different attenuation.

(DefProp Simple-Attack (Note-Start Note-Attenuate Note-End) Note-Profile)
(DefProp Simple-Attack (Attack-Length Attack-Attenuation Sustain-Length Sustain-Attenuation
                        Space-Length)         Note-Profile-Args)

(Defun Note-Attenuate (Voice Args)
  (Let ((Length (Car Args))
        (Attenuation (Cadr Args))
        )
    (Write-Sound-Reg (Attenuate-Reg-Of-Voice Voice) Attenuation)
    (Values 6. Length (Cddr Args))))

(DefProp Simple-Attack-Glue (Note-Start Note-Attenuate) Note-Profile)
(DefProp Simple-Attack-Glue (Attack-Length Attack-Attenuation
                                           Sustain-Length Sustain-Attenuation) Note-Profile)

(DefProp Simple-Attack-Glue-Continue (Note-Continue) Note-Profile)
(DefProp Simple-Attack-Continue (Continue-Length)    Note-Profile-Args)

(DefProp Simple-Attack-Glue-End (Note-Continue Note-End)  Note-Profile)
(DefProp Simple-Attack-End (Continue-Length Space-Length) Note-Profile-Args)


(DefProp Quiet (Note-Silent) Note-Profile)
(DefProp Quite nil Note-Profile-Args)

(Defun Note-Silent (Voice Args)
  (Write-Sound-Reg (Attenuate-Reg-Of-Voice Voice) ATTENUATE-SILENT)
  (Values 2 nil Args))

;;; Note Scheduler

;;;
;;; This function accepts the Note Profiles for all four voices.  It
;;; returns when all profiles have completed.  The TONE registers should
;;; have been set up before calling this.

(Defun Note-Scheduler (Profile-1 Args-1 &Optional
                       Profile-2 Args-2
                       Profile-3 Args-3
                       Profile-Noise Args-Noise)
  (Let ((Next-Event-Time 0)
        (Rel-Time 0)
        (P-Step-1 (Get Profile-1 'Note-Profile))
        (P-Step-2 (If Profile-2
                      (Get Profile-2 'Note-Profile)
                      nil))
        (P-Step-3 (If Profile-3
                      (Get Profile-3 'Note-Profile)
                      nil))
        (P-Step-N (If Profile-Noise
                      (Get Profile-Noise 'Note-Profile)
                      nil))
        (Next-Event-1 0)
        (Next-Event-2 0)
        (Next-Event-3 0)
        (Next-Event-N 0)
        (Int-Time)
        (Til-Next)
        )
    (Do ((Total-Int-Time 0 0))
        ((And (Null P-Step-1)
              (Null P-Step-2)
              (Null P-Step-3)
              (Null P-Step-N)))
      (Do ((Time Rel-Time (1+ Time)))
          ((>= Time Next-Event-Time) (Setq Rel-Time (1+ Time))))
      (When (And (Not (Null P-Step-1)) (>= Rel-Time Next-Event-1))
        (Multiple-value-setq (Int-Time Til-Next Args-1)
          (Funcall (Car P-Step-1) 1 Args-1))
        (Setq Total-Int-Time (+ Total-Int-Time Int-Time 10.))
        (Setq Next-Event-1 (+ Rel-Time Til-Next))
        (Setq P-Step-1 (Cdr P-Step-1)))
      (When (And (Not (Null P-Step-2)) (>= Rel-Time Next-Event-2))
        (Multiple-value-setq (Int-Time Til-Next Args-2)
          (Funcall (Car P-Step-2) 2 Args-2))
        (Setq Total-Int-Time (+ Total-Int-Time Int-Time 10.))
        (Setq Next-Event-2 (+ Rel-Time Til-Next))
        (Setq P-Step-2 (Cdr P-Step-2)))
      (When (And (Not (Null P-Step-3)) (>= Rel-Time Next-Event-3))
        (Multiple-value-setq (Int-Time Til-Next Args-3)
          (Funcall (Car P-Step-3) 3 Args-3))
        (Setq Total-Int-Time (+ Total-Int-Time Int-Time 10.))
        (Setq Next-Event-3 (+ Rel-Time Til-Next))
        (Setq P-Step-3 (Cdr P-Step-3)))
      (When (And (Not (Null P-Step-N)) (>= Rel-Time Next-Event-N))
        (Multiple-value-setq (Int-Time Til-Next Args-Noise)
          (Funcall (Car P-Step-N) 4 Args-Noise))
        (Setq Total-Int-Time (+ Total-Int-Time Int-Time 10.))
        (Setq Next-Event-N (+ Rel-Time Til-Next))
        (Setq P-Step-N (Cdr P-Step-N)))
      (Setq Rel-Time (+ Rel-Time Total-Int-Time))
      (Setq Next-Event-Time (Min Next-Event-1 Next-Event-2 Next-Event-3 Next-Event-N))
      )
    ))
|#


(defvar Beeps*chords
	(list 'Major-Chord                      ;;  0
	      'Major-Chord2			;;  1
	      'Minor-chord			;;  2
	      'Minor-chord2			;;  3
	      'Major-Seventh-Chord		;;  4
	      'Major-Seventh-Chord2		;;  5
	      'Minor-Seventh-Chord		;;  6
	      'Minor-Seventh-Chord2		;;  7
	      'Dominant-Seventh			;;  8
	      'Dominant-Seventh2		;;  9
	      'Diminished-Chord			;; 10
	      'Diminished-Chord2		;; 11
	      'Diminished-Seventh		;; 12
	      'Diminished-Seventh2		;; 13
	      'Augmented-Chord			;; 14
	      'Augmented-Chord2			;; 15
	      'Dual-1-4 ))			;; 16

(defun beeps*notes (note &optional (tonic 320.))
  ;;       c    c#   d    eb   e    f    f#   g    ab   a    bb   b
  ;;       0    1    2    3    4    5    6    7    8    9    10   11
  ;;      200. 189. 178. 168. 159. 150. 141. 133. 126. 119. 112. 106.
  ;;	  100. 94.  89.  84.  79.  75.  71.  66.  63.  59.  56.  53.
  (CASE note
	(0 Tonic)                           ;;; 0
	(1 (minor-second  Tonic))           ;;; 1
	(2 (major-second  Tonic))           ;;; 2
	(3 (minor-third   Tonic))           ;;; 3
	(4 (major-third   Tonic))           ;;; 4
	(5 (major-fourth  Tonic))           ;;; 5
	(6 (minor-fifth   Tonic))           ;;; 6
	(7 (major-fifth   Tonic))           ;;; 7
	(8 (minor-sixth   Tonic))           ;;; 8
	(9 (major-sixth   Tonic))           ;;; 9
	(10 (minor-seventh Tonic))           ;;;10
	(11 (major-seventh Tonic))           ;;;11
	(12 (octave Tonic))                  ;;;12
	(13 (minor-second  (octave Tonic)))  ;;;13
	(14 (major-second  (octave Tonic)))  ;;;14
	(15 (minor-third   (octave Tonic)))  ;;;15
	(16 (major-third   (octave Tonic)))  ;;;16
	(17 (major-fourth  (octave Tonic)))  ;;;17
	(18 (minor-fifth   (octave Tonic)))  ;;;18
	(19 (major-fifth   (octave Tonic)))  ;;;19
	(20 (minor-sixth   (octave Tonic)))  ;;;20
	(21 (major-sixth   (octave Tonic)))  ;;;21
	(22 (minor-seventh (octave Tonic)))  ;;;22
	(23 (major-seventh (octave Tonic)))  ;;;23
	(24 (octave (octave Tonic)))))       ;;;24

(defun PLAY-notes (notes lengths &optional Chords (master-duration 120.) (tone-scale 320.))
  "2NOTES is a list of note numbers - see BEEPS*NOTES.
LENGTHS is a list of times - one element corresponding to each element of NOTES.
  The time is the length element times master-duration milliseconds.
CHORDS is an optional list with elements corresponding to NOTES. 
  These are chord numbers - see BEEPS*CHORDS.*"
  (reset-sound)
  (with-sound-enabled
    (tv:WITH-real-time
      (DO ((a-note notes (CDR a-note))
	   (a-length lengths (CDR a-length))
	   (a-chord chords (CDR a-chord)))
	  ((NULL a-note))
	(do-sound nil (truncate master-duration 10))	   ;pause a bit
	(play-note (CAR a-note) (* (OR (CAR a-length) 1) master-duration) (car a-chord) tone-scale)))))

(DEFUN play-note (note-number &optional (duration 500) (play-Chords Nil) (tone-scale 320.)
		  &aux tone chord)
  (IF (minusp note-number)
      (do-sound nil duration)       ;;;; change this to call the attenuation scheduler
    (setq tone (beeps*notes (REM note-number 25.) tone-scale))
    (COND (play-chords
	   (setq chord (NTH (REM play-chords 17.) beeps*chords))
	   (funcall chord tone)                               
	   (attenuation-off)
	   (do-sound nil duration)
	   (attenuation-on))
	  (T           
	   (Write-Tone Tone-1 (convert-note tone))
	   (write-sound-reg attenuation-1 attenuate-loud)
	   (do-sound nil duration)
	   (write-sound-reg attenuation-1 Attenuate-Silent)))))

#| ;; All these have been translated to use the music-interpreter

(DEFMACRO def-beep-notes (name &body notes)
  `(def-beep-function ,name play-notes ,@notes))


(def-beep-notes :cscale-up 
  (0 1 2 3 4 5 6 7 8. 9. 10. 11. 12.)
  (1 1 1 1 1 1 1 1 1  1   1   1   2))

(def-beep-notes :cscale-down
  (12. 11. 10. 9. 8. 7 6 5 4 3 2 1 0)
  ( 1   1   1  1  1  1 1 1 1 1 1 1 2))

(def-beep-notes :cscale-both
  (0 1 2 3 4 5 6 7 8. 9. 10. 11. 12. 12. 11. 10. 9. 8. 7 6 5 4 3 2 1 0)
  (1 1 1 1 1 1 1 1 1  1   1   1   2   1   1   1  1 1 1 1 1 1 1 1 1 2))

(def-beep-notes :scale-up
  (0 2 4 5 7 9. 11. 12.)
  (1 1 1 1 1 1   1   4))

(def-beep-notes :scale-down
  (12. 11. 9. 7 5 4 2 0)
  ( 1   1  1  1 1 1 1 4))

(def-beep-notes :scale-both
  (0 2 4 5 7 9. 11. 12. 12. 11. 9. 7 5 4 2 0)
  (1 1 1 1 1 1   1   4   1   1  1  1 1 1 1 4))

(def-beep-notes :joy
  (12. 11. 9. 7 -1 5 4 2 0 -1 7 9. -1 9. 11. -1 11. 12.)
  ( 3   2  1  3  1 1 2 2 4  2 2 4   1 2   4   1  2   4))

(def-beep-notes :nuts
  ( 6 18. 17. 15. 13. 11. 10. 10. 10. -1  3 15. 13. 11. 10.  8.  6)
  ( 2  3   1   1   1   1   1   1   4   2  2  3   1   1   1   1  2))

(def-beep-notes :kiss
  (0 2 4 7 9. 12. 11. 7 4 -1)
  (2 1 1 1 1   1   2  1 2  2))

(def-beep-notes :tree
  (0 5 5 5 -1 7 9. 9. 9. -1 9. 7 9. 10. 4 7 5 5)
  (2 1 1 2  0 2 1  1  2   2 2  1 1   2  2 2 1 3))

(def-beep-notes :deck
  (14. 12. 11. 9. 7 9. 11. 7  9. 11. 12. 9. 11. 9. 7 6 7)
  ( 2   1   2  2  2 2   2  3  1   1   1  1   3  1  2 2 4))

(def-beep-notes :jingle
  (4 4 4 -1 4 4 4 -1 4 7 0 2 4 -1 5 5 5 5 5 4 4 4 7 7 5 2 0)
  (2 2 2  2 2 2 2  2 2 2 3 1 4  2 2 2 3 1 2 2 3 1 2 2 2 2 4))

(def-beep-notes :zowie
  (5  5  5  4  5  7  5  7  9.  7  5  4  2  4  5)
  (1  1  2  1  1  2  1  1  2   1  1  2  1  1  2))

(def-beep-notes :races
  (0 5 9. 12. 12. 12. -1 9. 9. 9. -1 5 9. 5 0 -1 0 5 9. 12. 12. 12. -1 9. 9. 9. -1 0 0 0 5)
  (2 2 2   1   1   1   1 1  1  1   1 1 1  1 3  2 2 2 2   1   1   1   1 1  1  1   1 1 1 1 5))

(def-beep-notes :swars
  (0  7 -1  5  4  2 12.  7 -1  5  4  2 12.  7 -1  5  4  5  2) 
  (2  3  0  1  1  1  2   2  1  1  1  1  2   2  1  1  1  1  4.))

(def-beep-notes :et
  (12. 19. 17. 16. 14. 16. 12. 7 -1  14. 21. 19. 18. 16. 18. 14.)
  ( 2   6   2   2   2   2   4  6  2   2   6   2   2   2   2   6))

(def-beep-notes :tzone
  (4 5 4 1 4 5 4 1 4 5 4 1 -1)
  (1 1 1 1 1 1 1 1 1 1 1 1 2))

(def-beep-notes :bat5
  (6 6 6 3 -1 5 5 5 2)
  (1 1 1 8. 2 1 1 1 8.))
|#
;1-----------------------------------------------------------------------------*

(def-beep-function :beep-rise beep-rise)
(def-beep-function :beep-drop beep-drop)
(def-beep-function :boop boop 2)
(def-beep-function :poob poob 2)
(def-beep-function :oopb oopb 15 750 10 15000 7000)
(def-beep-function :ship oopb 8 800 100 7000 7000)
(def-beep-function :diddle oopb 8 133 58 15210 6948)

(def-beep-function :broop broop 20 750 10 15000 7000 1)
(def-beep-function :spoob spoob)
(def-beep-function :dreep dreep)
(def-beep-function :breep breep)
(def-beep-function :freep freep)
(def-beep-function :greep greep)
(def-beep-function :srieep srieep)
(def-beep-function :wuca srieep 284 2806 3 225558 6598)
(def-beep-function :fsrieep fsrieep)

(defvar boop-increment 50.)
(defvar boop-min 100.)
(defvar boop-max 510.)

(defun beep-rise (&optional (pitch 250.) (rate 4000.))
  (diminished-chord pitch)
  (attenuation-off)
  (do ((i 0 (1+ i))
       (freq  pitch (+ freq 5.)))
      ((or (> i 100.) (> freq 1020.)))
    (diminished-chord freq)
    (delay rate))
  (attenuation-on))

(defun beep-drop (&optional (pitch 950.) (rate 4000.))
  (diminished-chord pitch)
  (attenuation-off)
  (do ((i 0 (1+ i))
       (freq  pitch (- freq 5.)))
      ((or (> i 100.) (< freq 65.)))
    (diminished-chord freq)
    (delay rate))
  (attenuation-on))

;1; Whis is a lot like whoop*
(defun boop (&optional (incr (random 5) boop-increment)
	     (pitch 250.)
	     (rate 4000.))
  (write-tone tone-1 pitch)
  (write-sound-reg attenuation-1 attenuate-loud)
  (do ((i 0 (1+ i))
       (freq  pitch (- freq (random 5) incr)))
      ((or (> i 75.) (< freq 65.)))
    (write-tone tone-1 freq)
    (delay rate))
  (write-sound-reg attenuation-1 attenuate-silent))

;1; This is a lot like shoop*
(defun poob (&optional (incr (random 5) boop-increment)
	     (pitch 75.)
	     (rate 2000.))
  (write-tone tone-1 pitch)
  (write-sound-reg attenuation-1 attenuate-loud)
  (do ((i 0 (1+ i))
       (freq  pitch (+ freq (random 5) incr)))
      ((or (> i 75.) (> freq 1000.)))
    (write-tone tone-1 freq)
    (delay rate))
  (write-sound-reg attenuation-1 attenuate-silent))

(defun oopb (&optional
	     (incr boop-increment)
	     (max-val (ROUND (* boop-max (+ 1 (* .75 (RANDOM-U2))))))
	     (min-val (ROUND (* boop-min (+ 1 (* .75 (RANDOM-U2))))))
	     (dir-1 (RANDOM 20000))
	     (dir-2 (RANDOM 20000)))
  (write-tone tone-1 max-val)
  (write-sound-reg attenuation-1 attenuate-loud)
  
  (do ((tone max-val (- tone (+ (random 5) incr))))
      ((or (< tone min-val) (< tone 65.) (> tone 1020.)))
    (write-tone tone-1 tone)
    (delay dir-1)
    (write-tone tone-1 (+ min-val (- max-val tone)))
    (delay dir-2))
  (write-sound-reg attenuation-1 attenuate-silent))

(defun broop (&optional
	      (incr boop-increment)
	      (max-val boop-max)
	      (min-val boop-min)
	      (dir-1 (RANDOM 20000))
	      (dir-2 (RANDOM 20000))
	      (dT-1 (RANDOM 2))
	      (dT-2 (IF (zerop dT-1) 1 0)))
  (write-tone tone-1 (* max-val dT-1))
  (write-sound-reg attenuation-1 attenuate-loud)
  
  (do ((tone max-val (- tone (+ (random 5) incr))))
      ((< tone min-val))
    (write-tone tone-1 (* tone dT-1))
    (delay dir-1)
    (write-tone tone-1 (* (+ min-val (- max-val tone)) dT-2))
    (delay dir-2)
    )
  (write-sound-reg attenuation-1 attenuate-silent))

(defun spoob (&optional
	      (val-1 (+ boop-min (RANDOM (- boop-max boop-min))))
	      (val-2 (+ boop-min (RANDOM (- boop-max boop-min))))
	      (dir-max 5000))
  (write-tone tone-1 val-1)
  (write-sound-reg attenuation-1 attenuate-loud)
  (delay 100000)
  (do ((dir 0 (+ dir 400 (random 400))))
      ((> dir dir-max))
    (write-tone tone-1 val-2)
    (delay (* dir 10))
    (write-tone tone-1 val-1)
    (delay (* (- dir-max dir) 10)))
  (write-tone tone-1 val-2)
  (delay 200000)
  (write-sound-reg attenuation-1 attenuate-silent))

(defun spoop (&optional
	      (incr boop-increment)
	      (val (+ boop-min (RANDOM (- boop-max boop-min))))
	      (max-val boop-max)
	      (min-val boop-min)
	      (dir-1 10000)
	      (dir-2 20000))
  
  (write-tone tone-1 val)
  (write-sound-reg attenuation-1 attenuate-loud)
  (do ((tone max-val (- tone (+ (random 5) incr))))
      ((< tone min-val))
    (write-tone tone-1 val)
    (delay dir-1)
    (write-tone tone-1 dir-2)
    (delay dir-2))
  (write-sound-reg attenuation-1 attenuate-silent))

(defun breep (&optional (incr 50) (tone-start 200))
  (write-tone tone-1 tone-start)
  (write-sound-reg attenuation-1 attenuate-loud)
  
  (do ((i 6 (1+ i))
       (tone1 tone-start (if (> i 10) (+ tone1 incr) tone1))
       (tone2 tone-start (if (> i 20) (+ tone2 incr) tone2))
       (tone3 tone-start (if (> i 30) (+ tone3 incr) tone3))
       (tone4 tone-start (if (> i 40) (+ tone4 incr) tone4))
       (tone5 tone-start (if (> i 50) (+ tone5 incr) tone5)))
      ((OR (> i 30) (> tone1 1020)))
    (Write-Tone Tone-1 tone1)
    (delay 10000)
    (Write-Tone Tone-1 tone2)
    (delay 10000)
    (Write-Tone Tone-1 tone3)
    (delay 10000)
    (Write-Tone Tone-1 tone4)
    (delay 10000)
    (Write-Tone Tone-1 tone5)
    (delay 10000))
  (write-sound-reg attenuation-1 attenuate-silent))

(defun dreep (&optional (incr 75) (tone-start 900) (seperation 10))
  (write-tone tone-1 tone-start)
  (write-sound-reg attenuation-1 attenuate-loud)
  
  (do ((i 0 (1+ i))
       (tone1 tone-start (if (> i (* 0 seperation)) (- tone1 incr) tone1))
       (tone2 tone-start (if (> i (* 1 seperation)) (- tone2 incr) tone2))
       (tone3 tone-start (if (> i (* 2 seperation)) (- tone3 incr) tone3))
       (tone4 tone-start (if (> i (* 3 seperation)) (- tone4 incr) tone4))
       (tone5 tone-start (if (> i (* 4 seperation)) (- tone5 incr) tone5)))
      ((OR (> i 250) (< tone5 100)))
    (Write-Tone Tone-1 tone5)
    (delay 10000)
    (Write-Tone Tone-1 tone4)
    (delay 10000)
    (Write-Tone Tone-1 tone3)
    (delay 10000)
    (Write-Tone Tone-1 tone2)
    (delay 10000)
    (Write-Tone Tone-1 tone1)
    (delay 10000))
  
  (write-sound-reg attenuation-1 attenuate-silent))


(defun freep (&optional (sstep 25) (nstep 5) (stime 50) (fmin 400) (fmax 2500)  )
  (do ((i 0 (1+ i))
       (fi (+ fmin (RANDOM (- fmax fmin)))
	   (MAX fmin (MIN fmax (+ fi (* sstep (- nstep (RANDOM (+ 1 nstep nstep)))))))))
      ((> i (truncate 1000 stime)))
    (IF (= fi fmin) (setq fi (+ fmin (* sstep (RANDOM 5)))))
    (Write-Tone Tone-1 fi)
    (write-sound-reg attenuation-1 attenuate-loud)
    (delay (* (1+ (RANDOM stime)) 10000)))
  (attenuation-on))

(defun greep (&optional (sstep 25) (nstep 5) (stime 25) (fmin 400) (fmax 2500)  )
  (do ((i 0 (1+ i))
       (fi (+ fmin (RANDOM (- fmax fmin)))
	   (MAX fmin (MIN fmax (+ fi (* sstep (ROUND (* nstep (RANDOM-U2)))))))))
      ((> i (truncate 1000 stime)))
    (IF (= fi fmin) (setq fi (+ fmin (* sstep (RANDOM 5)))))
    (Write-Tone Tone-1 fi)
    (write-sound-reg attenuation-1 attenuate-loud)
    (delay (* (1+ (RANDOM stime)) 10000)))
  (attenuation-on))

(defun nreep (&optional (sstep 10) (nstep 25) (stime 10) (fmin 400) (fmax 2500)  )
  (do ((i 0 (1+ i))
       (fi1 (+ fmin (RANDOM (- fmax fmin)))
	    (MAX fmin (MIN fmax (+ fi1 (* sstep (ROUND (* nstep (RANDOM-U2))))))))
       (fi2 (+ fmin (RANDOM (- fmax fmin)))
	    (MAX fmin (MIN fmax (+ fi2 (* sstep (ROUND (* nstep (RANDOM-U2))))))))
       (fi3 (+ fmin (RANDOM (- fmax fmin)))
	    (MAX fmin (MIN fmax (+ fi3 (* sstep (ROUND (* nstep (RANDOM-U2))))))))
       (fi4 (+ fmin (RANDOM (- fmax fmin)))
	    (MAX fmin (MIN fmax (+ fi4 (* sstep (ROUND (* nstep (RANDOM-U2)))))))))
      ((> i (truncate 75 stime)))
    (Write-Tone Tone-1 fi1)
    (write-sound-reg attenuation-1 attenuate-loud)
    (delay (* (1+ (RANDOM stime)) 10000))
    (Write-Tone Tone-1 fi2)
    (delay (* (1+ (RANDOM stime)) 10000))
    (Write-Tone Tone-1 fi3)
    (delay (* (1+ (RANDOM stime)) 10000))
    (Write-Tone Tone-1 fi4)
    (delay (* (1+ (RANDOM stime)) 10000)))
  (attenuation-on))

(defun srieep (&optional
	       (top-f      (+ 100       (RANDOM 500)))
	       (bot-f      (+ top-f 100 (RANDOM 4000)))
	       (num-cycles (+ 5         (RANDOM 15)))
	       (period     (+ 50000     (RANDOM 200000)))
	       (beep-time  (+ 2000      (RANDOM 8000)))
	       &aux ff w-increment)
  (setq w-increment (/ (* 2.0 3.14159265 beep-time) period))
  (do ((c 0 (1+ c)))
      ((= c num-cycles))
    (do ((k 0 (+ k beep-time))
	 (w 0 (+ w w-increment)))
	((>= k period))
      (setq ff (+ top-f (ROUND (* .5 (- bot-f top-f) (+ 1 (COS w))))))
      (Write-Tone Tone-1 ff)
      (write-sound-reg attenuation-1 attenuate-loud)
      (delay (*  (truncate beep-time ff) ff))))
  (attenuation-on))

(defun fsrieep (&optional
		(top-f        (+ 100          (RANDOM 2000)))
		(bot-f        (+ top-f 100    (RANDOM 2000)))
		(decay-factor (+ .8 (* .4 (/ (RANDOM 100) 100.0))))
		(num-cycles   (+ 5            (RANDOM 15)))
		(period       (+ 100000       (RANDOM 600000)))
		(beep-time    (+ 2000         (RANDOM 8000)))
		&aux ff w-increment)
  
  (setq w-increment (/ (* 2.0 3.14159265 beep-time)  period))
  (do ((c 1 (1+ c))
       (tf top-f (* tf decay-factor))
       (bf bot-f (* bf decay-factor)))
      ((>= c num-cycles))
    (do ((k 0 (+ k beep-time))
	 (w 0 (+ w w-increment)))
	((>= k period))
      (setq ff (ROUND (+ tf (* .5 (- bf tf) (+ 1 (COS w))))))
      (Write-Tone Tone-1 ff)
      (write-sound-reg attenuation-1 attenuate-loud)
      (delay (*  (truncate beep-time ff) ff))))
  (attenuation-on))

(DEFVAR random-u2-random-array NIL)

(defun RANDOM-U2 (&optional (rand-seed nil)
		 &aux (fineness 32768.))

" This routine will generate a uniformly distributed random number between -1 and 1.

  See the documentation Random-N for more complete documentation.

    The two differences between that routine and this one is that one generates
    normally distributed random numbers and this one generates linearly distributed
    random numbers, and that routine uses random-n-random-array as it's random array,
    while this routine uses random-u2-random-array for the same purpose."

  (COND ((NOT (NULL rand-seed))			; If rand-seed is non-zero, generate array
	 (setq random-u2-random-array (si:random-create-array 71 35 rand-seed))))

  (COND ((NOT (arrayp random-u2-random-array))   ; random array not set up, use default
	 (1- (/ (FLOAT (RANDOM (* 2 fineness))) fineness)))
	(t
	 (1- (/ (FLOAT (RANDOM (* 2 fineness) random-u2-random-array)) fineness)))))
