;;; -*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.



;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;;  3/21/89    TWE	When the user generates splines a comment is now generated into
;;;			the .tex file telling the user that non-standard LaTeX commands
;;;			are present in the file and what the \documentstyle command
;;;			should look like to generate the picture.
;;;  3/20/89    TWE	Added support for splines.
;;; 10/24/88    TWE	Added the variable *DEFAULT-LANDSCAPE-LATEX-FONT* to allow one to
;;;			select a LaTeX font other than \tiny for a picture in a landscape
;;;			orientation.
;;; 10/121*/88    TWE	1Changed  *READ-PICTURE-FILE 1by adding an optional argument for the stream.*
;;; 10/19/88    TWE	Added SYS:PRINT-READABLY-MIXIN as a component flavor to
;;;			BASIC-OBJECT so that picture objects can be dumped to a file.
;;;			Added READ-ONLY-PICTURE to the list of settable instance
;;;			variables for picture-window.
;;; 10/17/88    TWE	Added a function to allow one to toggle read-only mode via the
;;;			keyboard.
;;; 10/14/88    TWE	Put support in place for read-only mode.
;;; 10/13/88    TWE	Updated to be able to handle both the Explorer and microExplorer
;;;			screen sizes.  Recreated the old version of :WRITE-OBJECT-STATUS
;;;			for the microExplorer which has a different implementation for
;;;			the cold-load-stream.
;;; 10/12/88    TWE	Updated WRITE-TEX-FILE to allow the user to have text generated
;;;			as is, instead of being texified.  If the orientation is
;;;			landscape then \tiny
;;; 10/10/88    TWE	Fixed LATEX-TEXT-MAPPINGS to put backslash in lower case.
;;; 10/04/88    TWE	Fixed TEXIFY to handle the >< characters too.
;;;  9/22/88    TWE	Fixed WRITE-TEX-FILE to output line-thickness and to wrap the
;;;			figure environment in a {\tiny } to make the fonts print
;;;			reasonably small instead of being too large.
;;;  7/28/88    TWE	Fixed VALID-POSITION-ITEM-LIST to not have the duplicate
;;;			positions and to contain documentation strings.  Fixed
;;;			GET-POSITION-DATA to remember the position the user selected and
;;;			to use that as a default for future position selections.
;;;  6/14/88    TWE	Added a documentation string to the POPUP-FONT-MENU to describe
;;;			what `default' means.
;;;  6/13/88    TWE	Changed POPUP-FONT-MENU so that it doesn't show `default' when
;;;			the user is changing the default font.
;;;  6/13/88    TWE	Fixed READ-PICTURE-FILE to handle old pictures which have objects
;;;			that have a thickness.
;;;  6/08/88    TWE	Changed PICTURE-ERROR to output the CR after the message instead
;;;			of before.  This seems to be the normal convention.
;;;  5/25/88    TWE	Fixed MOUSE-Y-OR-N-P to return the correct value.  Fixed the font
;;;			menu usage to look at the proper value returned by W:MENU-CHOOSE.
;;;  5/16/88    TWE	Added support for a default font.
;;;  5/04/88    TWE	Changed the coordinate display to not cause a flicker of the
;;;			mouse.  Previously, the mouse would flicker, making it difficult
;;;			to determine where it is.
;;;  4/20/88    TWE	Fixed up MOUSE-Y-OR-N-P to move the mouse to the `yes' item.
;;; 10/12/87    TWE	Fixed mouse moves to change the blinker position too.
;;; 10/01/87    TWE	Fixed the bug where one has a box with text in it and then edits
;;;			the box to remove the text.  Modified ENTER-TEXT-LINES.
;;;  9/29/87    TWE	Fixed up to translate object being read in that have negative
;;;			dimensions to make them positive.  Fixed up text entry and TEXIFY
;;;			to remove trailing returns.  LaTeX doesn't allow these either.
;;;  9/10/87    TWE	Added support for landscape mode.
;;;  7/20/87    TWE	Added the DESCRIPTIVE-FONT-NAME function for objects with a font.
;;;  7/17/87    TWE	Implement the code to display information about the point the
;;;			mouse is on when specifying an initial position for an object.
;;;  7/06/87    TWE	Edited MOUSE-GLYPH-HOTSPOT-MAPPING to make it more readable.
;;;  6/29/87    TWE	Added the support for MINIMUM-X/Y to make the picture only as big
;;;			as the user makes it.
;;;  6/23/87    TWE	Added support for fonts.
;;;  6/22/87    TWE	Added more hotspot definitions.
;;;  6/18/87    TWE	Fixed up pathname defaulting to work this time.  Also fixed a
;;;			double refresh problem when reading in a file.  To get see this
;;;			work properly, a file which has this problem will have to be
;;;			rewritten out to disk.
;;;  6/02/87    TWE	Moved the edit code to a separate file.  Added more mouse glyphs.
;;;  6/01/87    TWE	Implemented the :WRITE-OBJECT-STATUS method so that duplicate
;;;			code could be removed from the object drawing routines.  Also
;;;			added code to properly support text which contains empty lines.
;;;  5/29/87    TWE	Fixed up GET-OBJECT-DATA to support sequencing through the items.
;;;  5/27/87    TWE	Added a :SET-POSITION method to fix the make-system change made
;;;			on 5/26.  Moved the status description to the AAA-STATUS file.
;;;  5/26/87    TWE	Fixed the file I/O functions to fix the following problems.  The
;;;			line thickness wasn't getting written out to the .PIC file.  The
;;;			current margin wasn't getting erased when reading in a .PIC file
;;;			that defines a new margin.  Made the make-system cleaner by
;;;			removing the :SETTABLE-INSTANCE-VARIABLES option from the
;;;			position-mixin defflavor.  Put error checking into the picture
;;;			reading function to handle the case of not being able to find
;;;			the file.  Fixed up pathname defaulting for the write function so
;;;			that it works better for .TEX files.
;;;  5/19/87    TWE	Finished the editing code.
;;;  5/08/87    TWE	Started writing the editing code.
;;;  5/07/87    TWE	Updated the default font to be tr8.  This font is very close to
;;;			the actual size relative to the screen dimensions.
;;;  4/30/87    TWE	Fixed the position-mixin to be allow for instance variables to be
;;;			settable.
;;;  4/16/87    TWE	Changed TV package prefixes to W.
;;;  4/13/87    TWE	Fixed problems with positioning within the Y axis.
;;;  3/30/87	TWE	Converted to release 3.  Removed the :settable-instance-variables
;;;			for the position-mixin to remove a compiler warning message.
;;;			Added an X/Y origin for the picture flavor.  This will help in 
;;;			getting the Y coordinate moving in the positive upward direction
;;;			(instead of positive downward direction).  Fixed up the two mouse
;;;			input functions to use these new instance variables.
;;;  1/15/87	TWE	Fixed up positions to be lower case instead of upper case (i.e.
;;;			lb instead of LB).
;;; 12/29/86	TWE	Restructures by splitting up into several files and creating a
;;;			defsystem.


;;; This file defines a program which allows one to interactively specify a
;;; single picture for subsequent generation by the LaTeX phototypesetting
;;; system.  The user specifies certain parameters, the graphic objects to
;;; be drawn and their positions.  This program will then generate the
;;; appropriate commands to have LaTeX reproduce those same graphic objects.

;;; Every object flavor needs to have at least the following methods defined.
;;;
;;; The two :WRITE-... methods are used to save the picture on disk for later
;;; use.  The :WRITE-LATEX method writes out the commands which LaTeX will
;;; understand to have it generate the picture.  The :WRITE-PICTURE method
;;; writes out forms which can be used to reconstruct the internal
;;; representation of the picture.  This is useful when one wants to modify a
;;; picture when it isn't in the current environment.  The stream arguments
;;; are where the output is written to.
;;;(DEFMETHOD (OBJECT-FLAVOR-NAME :WRITE-LATEX) (LATEX-STREAM))
;;;(DEFMETHOD (OBJECT-FLAVOR-NAME :WRITE-PICTURE) (PICTURE-STREAM))
;;; The :TRANSLATE-UNITS method is used to translate the units of the object
;;; to a new unit.  For example, if one is changing from millimeters to
;;; inches, this method is called for each object to have the object adjust
;;; to the new unit of measure.  Most of the time, one can simply take an
;;; existing method and modify it slightly for the new object flavor.
;;;(DEFMETHOD (OBJECT-FLAVOR-NAME :TRANSLATE-UNITS) (NEW-OVER-OLD X-ORIGIN Y-ORIGIN))
;;; This method will draw the object on the WINDOW.
;;;(DEFMETHOD (OBJECT-FLAVOR-NAME :DRAW) (WINDOW))
;;; The following three methods are used to allow one to edit an object.
;;; The :DISTANCE-FROM-POINT method returns the smallest distance that this
;;; object is from the point (X,Y).  X and Y are measured in terms of the
;;; current unit.  If the point (X,Y) is inside of the object then 0.0 is
;;; returned.  The WINDOW object may or may not be needed.
;;;(DEFMETHOD (OBJECT-FLAVOR-NAME :DISTANCE-FROM-POINT) (WINDOW X Y))
;;; The :NAME method returns a string which identifies the object.  This is
;;; used displayed in the interaction pane to give feedback to the user on
;;; the object which was selected.
;;;(DEFMETHOD (OBJECT-FLAVOR-NAME :NAME) ())
;;; The :EDIT-FUNCTION method returns the symbol which is the function to do
;;; editing for this object.  Most of the time this will be the same function
;;; which is invoked via the command menu.  This function has one optional
;;; argument which is the object being edited.
;;;(DEFMETHOD (OBJECT-FLAVOR-NAME :EDIT-FUNCTION) ())


;;; Implementation notes:
;;; Generally, when creating a new object, it is a good idea to first
;;; implement the low-level function which will draw the object on a window
;;; using window system functions (or methods).  Next, implement the function
;;; which will draw the object given the object itself and an ALU.  This
;;; second function will be called when a `rubber-banding' operation is being
;;; performed.  It will translate all object units into pixel values and then
;;; call the low-level function to do the real work.  Three other functions
;;; which should be written are:
;;;
;;;   (1) a `MOVE' function to perform the actual moving of the object to a
;;;   new position on the screen.
;;;   (2) a `RUBBER-BAND' function to implement the rubber-banding.  This is
;;;   where one specifies a corner of the object and then moves the mouse to
;;;   the other corner.  While one is moving the mouse, the correct object
;;;   size for that mouse position is drawn.
;;;   (3) a function to be called from the command menu which will handle the
;;;   user-interface of object creation.
;;;
;;; Be sure to look at the implementation of existing objects to see how they
;;; handle rubber banding and obtaining user input.  These mechanisms are
;;; intended to be general enough to allow them to be used for any object.
;;;
;;; The method :SETUP-INSTANCE-ENVIRONMENT is helpful when one wants to invoke
;;; a function which expects the instance variables of the PICTURE-WINDOW to
;;; be bound.
;;;
;;; If you want to define a new mouse character for your object, be sure to
;;; update the constant MOUSE-GLYPH-HOTSPOT-MAPPING to recognize your
;;; character.  This constant contains an array of (X,Y) positions of the
;;; `hotspot' of that character.  This is the pixel offset from the upper left
;;; corner of the character box which is where the user's focus of attention
;;; will be.  Use the font editor to help identify the correct (X,Y) position.
;;;
;;; The :TURN-ON-RUBBERBAND method is the implementation of the generalized
;;; rubber banding mechanism.  Its primary argument is a function which will
;;; implement the redrawing and updating of the object.  After this method is
;;; called, rubberbanding of an object is enabled.  The lexical environment of
;;; the function assumes that the object is accessible.  See the TEXT file for
;;; a simple example use of this method.
;;;
;;; The READ-MOUSE-XY-INPUT function is used to handle user interaction after
;;; rubber banding has been turned on.  In this way the user interface can be
;;; isolated to one function.  It returns when the user presses the ABORT key
;;; or clicks on a mouse button.  After obtaining user input, the
;;; :TURN-OFF-RUBBEBAND is called.
;;;
;;; There is a lot of support for text positioning.  This is where one can
;;; specify where within an object text is to be placed.  Normally it is
;;; placed in the center of an object, but the user is allowed to change this.
;;; The support which is present only supports placement within rectangular
;;; objects, so triangular objects would need additional support.
;;;
;;; The function GET-OBJECT-DATA is another generalized user interface
;;; function which is used by all object functions.  This function is called
;;; by the function identified in the command menu (for example) when one is
;;; creating a new object.  The GET-OBJECT-DATA function takes an object, a
;;; menu item list and a menu label as its arguments.  It will bring up the
;;; menu and look at the value returned.  If the value is :DONE or :ABORT then
;;; it will return T or NIL respectively.  Otherwise it will use that value as
;;; a function and pass the function as arguments, the object and any
;;; additional arguments, if any.
;;;
;;; The function MOUSE-SPECIFY-POINT is used to do just that.  It returns an
;;; (X,Y) pixel position in the first quadrant (remember your high school
;;; geometry class) instead of the fourth quadrant normally used by the window
;;; system's coordinate system.  The MOUSE-SPECIFY-POINT function takes uses
;;; the mouse character and the who line documentation string specified by the
;;; user to aid the user in the selection.
;;;
;;; Also located in this file are utility functions.  Currently only two
;;; function are present SIGN and FORCE-SIGN to handle signs of numbers.


;;; Make sure the mouse glpyh names are loaded.
(EVAL-WHEN (COMPILE LOAD EVAL)
  (WHEN (NOT (BOUNDP 'W:MOUSE-GLYPH-NORTH-WEST-ARROW))
    (FORMAT T "~%Starting to load DLS:WINDOW;DEFINE-MOUSE-MAPPING")
    (LOAD "DLS:WINDOW;DEFINE-MOUSE-MAPPING")))

(DEFCONSTANT LATEX-EXTENSION "TEX"
  "2Filename extension for LaTeX output files.*")

(DEFCONSTANT PICTURE-EXTENSION "PIC"
  "2Filename extension for Picture files.*")

;;; This variable exists because of font problems when printing landscape pictures.
;;; Very often landscape pictures are run through SliTeX, instead of LaTeX, which
;;; chooses a font which is quite large when contrasted with LaTeX.  In an effort
;;; to choose a font more consistent with LaTeX, yet one which isn't too small for
;;; a foil, we shift into another font while writing out the .tex file.  Normally
;;; this font will be either tiny or footnotesize.  We default to tiny here because
;;; that is what it was originally.  There are some sites which do not support the
;;; tiny font.  For those sites, this variable can be changed to a font which is
;;; supported.
(DEFPARAMETER *DEFAULT-LANDSCAPE-LATEX-FONT* "tiny"
  "2The font LaTeX (or most likely SliTeX) should use when in landscape orientation.*")

(DEFPARAMETER *DEFAULT-DISPLAY-FONT* FONTS:TR8
  "2Default font used to display text strings.*")

(DEFPARAMETER *BOLDFACE-DISPLAY-FONT* FONTS:TR8B)

(DEFPARAMETER *EMPHATIC-DISPLAY-FONT* FONTS:TR8I)

(DEFPARAMETER *FONT-NAME-TO-FONT-MAPPING* `((:DEFAULT-FONT ,*DEFAULT-DISPLAY-FONT*)
                                            ("rm"          ,*DEFAULT-DISPLAY-FONT*)
                                            ("bf"          ,*BOLDFACE-DISPLAY-FONT*)
                                            ("em"          ,*EMPHATIC-DISPLAY-FONT*))
  "2Maps LaTeX font names to Explorer fonts.*")

(DEFPARAMETER
  *DESCRIPTIVE-FONT-NAMES*
  '((:DEFAULT-FONT "Default" "Reverts to whatever the current default font is.  Look at the `Change Global State' menu item named `Default Font'.~%The current default font is ~A.")
    ("rm" "Roman" "A roman font.  This uses the LaTeX \\rm command.")
    ("bf" "Boldface" "A boldface font.  This uses the LaTeX \\bf command.")
    ("em" "Emphatic" "An emphatic font.  This uses the LaTeX \\em command.")))

(DEFPARAMETER *ALL-FONTS* (REMOVE-DUPLICATES
                            (LOOP FOR (NIL FONT) IN *FONT-NAME-TO-FONT-MAPPING*
                                COLLECT FONT))
  "2List of fonts suitable for use as a font-map for a window.*")

(DEFUN LOOKUP-EXPLORER-FONT (FONT-NAME)
  (WHEN (NULL FONT-NAME)
    (SETQ FONT-NAME :DEFAULT-FONT))
  (CADR (ASSOC FONT-NAME *FONT-NAME-TO-FONT-MAPPING* :TEST #'(LAMBDA (X Y)
                                                               (OR (EQ X Y)
                                                                   (STRING-EQUAL X Y))))))

(DEFUN DESCRIPTIVE-FONT-NAME (FONT-NAME)
  "Convert a LaTeX font name or an Explorer font object into a description."
  (WHEN (NULL FONT-NAME)
    (SETQ FONT-NAME :DEFAULT-FONT))
  (IF (OR (STRINGP FONT-NAME) (EQ FONT-NAME :DEFAULT-FONT))
      (CADAR (MEMBER FONT-NAME *DESCRIPTIVE-FONT-NAMES*  :KEY #'(LAMBDA (X) (CAR X))
                     :TEST #'(LAMBDA (X Y)
                               (OR (EQ X Y)
                                   (STRING-EQUAL X Y)))))
      ;;ELSE
      ;; This is an Explorer font.
      (DESCRIPTIVE-FONT-NAME (CAAR (MEMBER FONT-NAME (CDR *FONT-NAME-TO-FONT-MAPPING*)
                                         :KEY 'CADR)))))

(DEFPARAMETER SCREEN-HEIGHT 400)
(DEFPARAMETER SCREEN-WIDTH  608)
(DEFPARAMETER SCREEN-HEIGHT (SEND W:MAIN-SCREEN :HEIGHT))
(DEFPARAMETER SCREEN-WIDTH  (SEND W:MAIN-SCREEN :WIDTH))

(DEFPARAMETER PAPER-RATIO (/ 11 8.5))
(DEFPARAMETER EXPLORER-P (= SCREEN-WIDTH 1024))

;;; Determine the dimensions of the Picture Window based upon the screen
;;; dimensions.  This was done so that this program could be used on the
;;; microExplorer with its smaller screen.
(DEFPARAMETER LABEL-WINDOW-HEIGHT (IF EXPLORER-P
				      13
				      ;;ELSE the fonts are a bit higher onthe microExplorer
				      15))

;;; The 2 is for the borders.
(DEFPARAMETER PORTRAIT-HEIGHT (- SCREEN-HEIGHT LABEL-WINDOW-HEIGHT 2))
(DEFPARAMETER PORTRAIT-WIDTH  (TRUNCATE PORTRAIT-HEIGHT PAPER-RATIO))

(DEFPARAMETER LANDSCAPE-WIDTH  (- SCREEN-WIDTH 195 2))
(DEFPARAMETER LANDSCAPE-HEIGHT (TRUNCATE LANDSCAPE-WIDTH PAPER-RATIO))

(DEFCONSTANT *THICK-LINES* 4
  "2Line thickness for thick lines.*")

(DEFCONSTANT *THIN-LINES* 2
  "2Line thickness for thin lines.*")

(DEFPARAMETER *CURRENT-ORIENTATION* :PORTRAIT
  "2Orientation of the picture pane.  Can be :PORTRAIT or :LANDSCAPE.*")

;;; The following two function return the current number of inches in
;;; a particular direction of the page.
(DEFUN MAX-PAGE-WIDTH ()
  (IF (EQ *CURRENT-ORIENTATION* :PORTRAIT)
      8.5 11.0))

(DEFUN MAX-PAGE-HEIGHT ()
  (IF (EQ *CURRENT-ORIENTATION* :PORTRAIT)
      11.0 8.5))

(DEFCONSTANT *MIN-PAGE-WIDTH* 0.0
  "2Minimum page width in inches.*")

(DEFCONSTANT *MIN-PAGE-HEIGHT* 0.0
  "2Minimum page height in inches.*")

(DEFUN VALIDATE-LEFT-MARGIN (VALUE DIMENSION)
  (IF (NOT (NUMBERP VALUE))
      "The value must be a number."
      ;;1ELSE*
      (LET ((VALUE-IN-INCHES (/ VALUE (UNITS-PER-INCH DIMENSION)))
            (MINIMUM-IN-UNITS (* *MIN-PAGE-WIDTH* (UNITS-PER-INCH DIMENSION))))
        (COND ((< VALUE-IN-INCHES *MIN-PAGE-WIDTH*)
               (FORMAT NIL "Must not be less than ~D." MINIMUM-IN-UNITS))
              (T
               NIL)))))

(DEFUN VALIDATE-TOP-MARGIN (VALUE DIMENSION)
  (IF (NOT (NUMBERP VALUE))
      "The value must be a number."
      ;;1ELSE*
      (LET ((VALUE-IN-INCHES (/ VALUE (UNITS-PER-INCH DIMENSION)))
            (MINIMUM-IN-UNITS (* *MIN-PAGE-HEIGHT* (UNITS-PER-INCH DIMENSION))))
        (COND ((< VALUE-IN-INCHES *MIN-PAGE-HEIGHT*)
               (FORMAT NIL "Must not be less than ~D." MINIMUM-IN-UNITS))
              (T
               NIL)))))

(DEFUN VALIDATE-RIGHT-MARGIN (VALUE DIMENSION)
  (IF (NOT (NUMBERP VALUE))
      "The value must be a number."
      ;;1ELSE*
      (LET ((VALUE-IN-INCHES (/ VALUE (UNITS-PER-INCH DIMENSION)))
            (MAXIMUM-IN-UNITS (* (MAX-PAGE-WIDTH) (UNITS-PER-INCH DIMENSION))))
        (COND ((> VALUE-IN-INCHES (MAX-PAGE-WIDTH))
               (FORMAT NIL "Must not be greater than ~D." MAXIMUM-IN-UNITS))
              (T
               NIL)))))

(DEFUN VALIDATE-BOTTOM-MARGIN (VALUE DIMENSION)
  (IF (NOT (NUMBERP VALUE))
      "The value must be a number."
      ;;1ELSE*
      (LET ((VALUE-IN-INCHES (/ VALUE (UNITS-PER-INCH DIMENSION)))
            (MAXIMUM-IN-UNITS (* (MAX-PAGE-HEIGHT) (UNITS-PER-INCH DIMENSION))))
        (COND ((> VALUE-IN-INCHES (MAX-PAGE-HEIGHT))
               (FORMAT NIL "Must not be greater than ~D." MAXIMUM-IN-UNITS))
              (T
               NIL)))))

(DEFUN VALIDATE-PAGE-WIDTH (VALUE DIMENSION)
  (OR (VALIDATE-LEFT-MARGIN  VALUE DIMENSION)
      (VALIDATE-RIGHT-MARGIN VALUE DIMENSION)))

(DEFUN VALIDATE-PAGE-HEIGHT (VALUE DIMENSION)
  (OR (VALIDATE-TOP-MARGIN    VALUE DIMENSION)
      (VALIDATE-BOTTOM-MARGIN VALUE DIMENSION)))

(DEFFLAVOR PICTURE-WINDOW
   ((DISPLAY-LIST NIL)

    (STATUS-PANE NIL)

    (CURRENT-MOUSE-BLINKER NIL)
    (DEFAULT-MOUSE-BLINKER W:MOUSE-GLYPH-NORTH-WEST-ARROW)
    (CURRENT-MOUSE-DOCUMENTATION NIL)

    ;; Implements rubberbanding and point specification.
    (MOUSING-STATUS NIL)
    (OLD-MOUSE-X NIL)
    (OLD-MOUSE-Y NIL)
    (RUBBERBAND-FUNCTION NIL)
    (MOUSE-OBJECT NIL)

    (READ-ONLY-PICTURE NIL)

    ;; Grid parameters.  STATUS says whether or not it is on or off.  X and Y
    ;; size indicate how big a grid cell is in their respective dimensions.
    ;; The unit of these numbers is base-unit.  Since the default for base-unit
    ;; is :millimeter, these are 5.0 millimeter squares.
    (GRID-STATUS-P T)
    (GRID-X-SIZE 10.0)
    (GRID-Y-SIZE 10.0)
    ;; Thickness of lines.  This will be 2 (thin) or 4 (thick) or a larger
    ;; number if the user wants a different default.
    (LINE-THICKNESS *THIN-LINES*)

    (MAXIMUM-X-UNIT 0.0)
    (MAXIMUM-Y-UNIT 0.0)
    (MINIMUM-X-UNIT MOST-POSITIVE-SHORT-FLOAT)
    (MINIMUM-Y-UNIT MOST-POSITIVE-SHORT-FLOAT)

    ;; Note that we are restricting the units to be whole numbers
    ;; of BASE-UNITs.  For example, a base unit can't be measured
    ;; in half-inches.
    (BASE-UNIT :MILLIMETER)

    ;; Origin of the picture.
    (X-ORIGIN 0.0)
    (Y-ORIGIN 0.0)
    ;; Equivalent of x/y-origin except that these are the pixel equivalents.
    ;; These are the pixel positions in the window that correspond to the
    ;; x/y unit origins.
    (X-PIXEL-ORIGIN (TRANSLATE-UNIT-TO-PIXEL-INTERNAL  0.0 :INCH))
    (Y-PIXEL-ORIGIN (TRANSLATE-UNIT-TO-PIXEL-INTERNAL (MAX-PAGE-HEIGHT) :INCH))

    (MARGIN-OBJECT NIL)                ; Object which will draw the margin.
    (MARGIN-UNIT :INCH)
    (LEFT-MARGIN 0.0)
    (TOP-MARGIN 0.0)
    (RIGHT-MARGIN  (MAX-PAGE-WIDTH))
    (BOTTOM-MARGIN (MAX-PAGE-HEIGHT)))
   (W:LIST-MOUSE-BUTTONS-MIXIN W:WINDOW)
  (:SETTABLE-INSTANCE-VARIABLES DISPLAY-LIST STATUS-PANE MOUSING-STATUS
                                MARGIN-OBJECT MARGIN-UNIT BASE-UNIT
                                LEFT-MARGIN TOP-MARGIN RIGHT-MARGIN BOTTOM-MARGIN
                                X-PIXEL-ORIGIN Y-PIXEL-ORIGIN
                                MAXIMUM-X-UNIT MAXIMUM-Y-UNIT MINIMUM-X-UNIT MINIMUM-Y-UNIT
                                READ-ONLY-PICTURE)
  (:GETTABLE-INSTANCE-VARIABLES LINE-THICKNESS)
  (:OUTSIDE-ACCESSIBLE-INSTANCE-VARIABLES STATUS-PANE MOUSING-STATUS Y-PIXEL-ORIGIN))

(DEFMETHOD (PICTURE-WINDOW :SETUP-INSTANCE-ENVIRONMENT) (LIST-OR-FUNCTION)
  "2This method simply sets up the instance variable environment so we can call FUNCTION.*"
  (LET ((FUNCTION LIST-OR-FUNCTION)
        (ARGS NIL))
    (WHEN (LISTP LIST-OR-FUNCTION)
        (SETQ FUNCTION (CAR LIST-OR-FUNCTION)
              ARGS     (CDR LIST-OR-FUNCTION)))
    (APPLY (SYMBOL-FUNCTION FUNCTION) ARGS)))

(DEFMETHOD (PICTURE-WINDOW :SET-LINE-THICKNESS) (NEW-THICKNESS &OPTIONAL (REFRESH-PICTURE T))
  (COND ((EQ NEW-THICKNESS :THICK)
         (SETQ LINE-THICKNESS *THICK-LINES*))
        ((EQ NEW-THICKNESS :THIN)
         (SETQ LINE-THICKNESS *THIN-LINES*))
        ((EQ NEW-THICKNESS :TOGGLE)
         (SETQ LINE-THICKNESS (IF (= LINE-THICKNESS *THICK-LINES*) *THIN-LINES* *THICK-LINES*))))
  (WHEN REFRESH-PICTURE
    (SEND SELF :REFRESH)))

(DEFUN CHANGE-LINE-THICKNESS (NEW-THICKNESS)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (SEND SELF :SET-LINE-THICKNESS NEW-THICKNESS)))

;;; The following information should probably be in the define-mouse-mapping file.
;;; It defines the location of the hotspot for each mouse glyph (or at least the ones
;;; being used here).
(DEFCONSTANT MOUSE-GLYPH-HOTSPOT-MAPPING
             (LET ((MAPPER (MAKE-ARRAY `(2 ,(LENGTH FONTS:MOUSE)) :ELEMENT-TYPE :FIXNUM
                                       :INITIAL-ELEMENT 0)))
               ;; Only care about a few of them so far.
               (LOOP FOR (NAME X Y) IN `((,W:MOUSE-GLYPH-3-12-ARC-POINTER 0 4)
                                         (,W:MOUSE-GLYPH-BLOCK-LETTER-T 5 2)
                                         (,W:MOUSE-GLYPH-CIRCLE-X 6 8.)
                                         (,W:MOUSE-GLYPH-DOUBLE-UP-ARROW 6 0)
                                         (,W:MOUSE-GLYPH-DOUBLE-UP-ARROW-LETTERED-D 6 0)
                                         (,W:MOUSE-GLYPH-EYE-GLASSES 8. 4)
                                         (,W:MOUSE-GLYPH-HOLLOW-BOX-POINTER 7 7)
                                         (,W:MOUSE-GLYPH-NORTH-WEST-ARROW 0 0)
                                         (,W:MOUSE-GLYPH-SCISSOR 5 8.)
                                         (,W:MOUSE-GLYPH-SPLINE-POINTER 7 7)
                                         (,W:MOUSE-GLYPH-THICK-HOLLOW-CROSS 7 7)
                                         (,W:MOUSE-GLYPH-THICK-LINE-POINTER 7 7)
                                         (,W:MOUSE-GLYPH-THICK-UP-ARROW-LETTERED-D 6 0))
                     DO (PROGN (SETF (AREF MAPPER 0 NAME) X)
                               (SETF (AREF MAPPER 1 NAME) Y)))
               MAPPER)
  "2Defines the location of the `hot spot' for mouse glyphs.*")

(DEFMETHOD (PICTURE-WINDOW :MOUSE-STANDARD-BLINKER) ()
  "2Change the mouse blinker to indicate the current operation, or to a default value.*"
  (LET* ((NEW-BLINKER (OR CURRENT-MOUSE-BLINKER DEFAULT-MOUSE-BLINKER))
         (HOTSPOT-X (AREF MOUSE-GLYPH-HOTSPOT-MAPPING 0 NEW-BLINKER))
         (HOTSPOT-Y (AREF MOUSE-GLYPH-HOTSPOT-MAPPING 1 NEW-BLINKER)))
    (W:MOUSE-SET-BLINKER-DEFINITION :CHARACTER HOTSPOT-X HOTSPOT-Y :ON
                                     :SET-CHARACTER NEW-BLINKER
                                     'FONTS:MOUSE)))

(DEFCONSTANT DEFAULT-PICTURE-WINDOW-DOCUMENTATION "Click on a menu item.")

(DEFMETHOD (PICTURE-WINDOW :WHO-LINE-DOCUMENTATION-STRING) ()
  (OR CURRENT-MOUSE-DOCUMENTATION
      DEFAULT-PICTURE-WINDOW-DOCUMENTATION))

(DEFMETHOD (PICTURE-WINDOW :MOUSE-MOVES) (X Y)
  (WHEN MOUSING-STATUS
    (LET ((INSIDE-X (- X (W:SHEET-INSIDE-LEFT)))
          (INSIDE-Y (- Y-PIXEL-ORIGIN (- Y (W:SHEET-INSIDE-TOP)))))
      ;; Don't bother doing anything if the mouse hasn't moved.
      (WHEN (OR (NOT (EQ INSIDE-X OLD-MOUSE-X))
                (NOT (EQ INSIDE-Y OLD-MOUSE-Y)))
        (MULTIPLE-VALUE-SETQ (INSIDE-X INSIDE-Y)
          (SEND SELF :GRIDIFY-POINT INSIDE-X INSIDE-Y))
        ;; Don't bother doing anything more if the mouse hasn't moved enough.
        (WHEN (OR (NOT (EQ INSIDE-X OLD-MOUSE-X))
                  (NOT (EQ INSIDE-Y OLD-MOUSE-Y)))

          (WHEN (EQ MOUSING-STATUS :ENABLE-RUBBERBAND)
            (WHEN OLD-MOUSE-X
              ;; If there was an old value then we need to erase the old image
              ;; before we draw the new one.
              (FUNCALL RUBBERBAND-FUNCTION OLD-MOUSE-X OLD-MOUSE-Y))
            (FUNCALL RUBBERBAND-FUNCTION INSIDE-X INSIDE-Y))
          ;; Remember these values in case the user moves the mouse again.
          (SETQ OLD-MOUSE-X INSIDE-X
                OLD-MOUSE-Y INSIDE-Y)))))
  (W:MOUSE-SET-BLINKER-CURSORPOS X Y))


(DEFMETHOD (PICTURE-WINDOW :TURN-ON-RUBBERBAND) (RB-FUNCTION &OPTIONAL (START-X NIL) (START-Y NIL))
  "2Initialize instance variables to start rubber banding an object.  The function must
be defined such that the object is in its environment.*"
  (SETQ OLD-MOUSE-X             START-X
        OLD-MOUSE-Y             (IF START-Y (- Y-PIXEL-ORIGIN START-Y) NIL)
        RUBBERBAND-FUNCTION     RB-FUNCTION)
  (SETQ MOUSING-STATUS          :ENABLE-RUBBERBAND)
  (WHEN (AND START-X START-Y)
    (SEND SELF :SET-MOUSE-CURSORPOS START-X START-Y)))

(DEFMETHOD (PICTURE-WINDOW :TURN-OFF-RUBBERBAND) ()
  (SETQ MOUSING-STATUS NIL))

(DEFMETHOD (PICTURE-WINDOW :TURN-ON-MOUSE-POINTING) (OBJECT)
  "2Initialize instance variables to start specifying an initial position for an object.
Object is an instance of the object whose point is to be specified.*"
  (SETQ OLD-MOUSE-X    NIL
        OLD-MOUSE-Y    NIL
        MOUSE-OBJECT   OBJECT
        MOUSING-STATUS :SPECIFYING-POINT)
  (PROCESS-RUN-FUNCTION '(:NAME "Mouse Display" :PRIORITY -1)
                        'DISPLAY-MOUSE-PIXEL-POSITION SELF OBJECT))

(DEFMETHOD (PICTURE-WINDOW :TURN-OFF-MOUSE-POINTING) ()
  (SETQ MOUSING-STATUS NIL))

(DEFUN DISPLAY-MOUSE-PIXEL-POSITION (WINDOW OBJECT)
  (LOOP WITH OLD-INSIDE-X = MOST-POSITIVE-FIXNUM
        WITH OLD-INSIDE-Y = MOST-POSITIVE-FIXNUM
        FOR INSIDE-X = (- W:MOUSE-X (W:SHEET-INSIDE-LEFT WINDOW))
        FOR INSIDE-Y = (- (PICTURE-WINDOW-Y-PIXEL-ORIGIN WINDOW)
                          (- W:MOUSE-Y (W:SHEET-INSIDE-TOP WINDOW)))
        WHILE (EQ (PICTURE-WINDOW-MOUSING-STATUS WINDOW) :SPECIFYING-POINT)
        DO (PROGN
             (PROCESS-SLEEP 5 "Move Da Mouse")
             (WHEN (SEND WINDOW :SELF-OR-SUBSTITUTE-SELECTED-P)
               (WHEN (OR (NOT (= INSIDE-X OLD-INSIDE-X))
                         (NOT (= INSIDE-Y OLD-INSIDE-Y)))
                 (SETQ OLD-INSIDE-X INSIDE-X
                       OLD-INSIDE-Y INSIDE-Y)
                 (SEND WINDOW :WRITE-OBJECT-STATUS OBJECT
                       (MULTIPLE-VALUE-BIND (NEW-UNIT-X NEW-UNIT-Y)
                           (SEND WINDOW :GRIDIFY-MEASURE
                                 (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                                       `(TRANSLATE-PIXEL-TO-UNIT ,INSIDE-X))
                                 (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                                       `(TRANSLATE-PIXEL-TO-UNIT ,INSIDE-Y)))
                         (FORMAT NIL "at (~,2F,~,2F)"
                                 NEW-UNIT-X NEW-UNIT-Y))))))))

(WHEN (NOT EXPLORER-P)
  ;; Old version which haD a lot of mouse flicker.  This is the version to
  ;; use on a microExplorer since the mouse doesn't flicker there anyway.
  (DEFMETHOD (PICTURE-WINDOW :WRITE-OBJECT-STATUS) (OBJECT &OPTIONAL STRING)
    (LET* ((THE-STATUS-PANE (PICTURE-WINDOW-STATUS-PANE SELF))
	   (OUTPUT-WINDOW   THE-STATUS-PANE))
      (SEND OUTPUT-WINDOW :SET-CURSORPOS 0 0)
      (SEND OUTPUT-WINDOW :CLEAR-EOF)
      (WHEN OBJECT
	(LET ((OBJECT-SHORT-NAME (SEND OBJECT :SHORT-NAME)))
	  (SEND OUTPUT-WINDOW :STRING-OUT OBJECT-SHORT-NAME)))
      (WHEN STRING
	(SEND OUTPUT-WINDOW :STRING-OUT STRING)))))

(WHEN EXPLORER-P
  ;; Implementation note:  The :WRITE-OBJECT-STATUS method is more complex than it
  ;; previously was in an attempt to remove mouse flicker.  The previous version
  ;; used normal window system output functions, which would do the following:
  ;;   Turn off the mouse blinker,
  ;;   perform the output operation, and then
  ;;   turn the mouse blinker back on.
  ;; When the mouse is moved quickly, the blinker would not be turned back on, making
  ;; it difficult to determine the mouse position.  This implementation performs the
  ;; same function, but does it in a different way.  The cold load stream is a stream
  ;; normally only used during a system build, or when the window system gets wedged.
  ;; It is useful here because it totally ignores the window system and will generate
  ;; its output no matter what the window system is doing.  What this means is that
  ;; we can perform output without having to turn the mouse blinker off and on.  We
  ;; can get away with this since we can assume that the mouse will never be moved
  ;; into the status pane.  If the user does move the mouse there, then it will only
  ;; be visible while it is being moved, but will return to its normal behavior when
  ;; moved outside of the status pane.
  ;;
  ;; The tricky part about the cold load stream is that it assumes that it has control
  ;; over the entire screen.  To overcome this, we use SYS:%DRAW-RECTANGLE (instead of
  ;; :CLEAR-EOF or :CLEAR-EOL) to only clear the interior part of the status pane.
  ;; Special care is taken to handle output which may contain RETURN characters.  See
  ;; the comment below for why this is necessary.
  (DEFMETHOD (PICTURE-WINDOW :WRITE-OBJECT-STATUS) (OBJECT &OPTIONAL STRING)
    (LET* ((THE-STATUS-PANE (PICTURE-WINDOW-STATUS-PANE SELF))
	   (OUTSIDE-TOP     (+ (W:SHEET-Y-OFFSET    THE-STATUS-PANE)
			       (W:SHEET-INSIDE-TOP  THE-STATUS-PANE)))
	   (OUTSIDE-LEFT    (+ (W:SHEET-X-OFFSET    THE-STATUS-PANE)
			       (W:SHEET-INSIDE-LEFT THE-STATUS-PANE)))
	   (STATUS-LINE-HEIGHT (W:SHEET-LINE-HEIGHT THE-STATUS-PANE)))
      (SEND SYS:COLD-LOAD-STREAM :SET-CURSORPOS OUTSIDE-LEFT OUTSIDE-TOP)
      (SYS:%DRAW-RECTANGLE (W:SHEET-INSIDE-WIDTH  THE-STATUS-PANE)
			   (W:SHEET-INSIDE-HEIGHT THE-STATUS-PANE)
			   (W:SHEET-INSIDE-LEFT   THE-STATUS-PANE)
			   (W:SHEET-INSIDE-TOP    THE-STATUS-PANE)
			   (W:SHEET-ERASE-ALUF    THE-STATUS-PANE)
			   (W:SHEET-SCREEN-ARRAY  THE-STATUS-PANE))
      (WHEN OBJECT
	(LET ((OBJECT-SHORT-NAME (SEND OBJECT :SHORT-NAME))
	      PART-TO-OUTPUT
	      LOC-NEWLINE)
	  ;; Output the object's short name.  This code is more complex than one would
	  ;; expect because it handles the RETURN character by moving the cold-load-stream's
	  ;; cursor to the next line in the status-pane.  Normally, when one outputs a
	  ;; RETURN character to the cold-load-stream, it will position to next line, but will
	  ;; put its cursor at the left-most pixel of the screen (and not the status-pane).
	  (LOOP
	    (WHEN (ZEROP (LENGTH OBJECT-SHORT-NAME))
	      ;; Have outputted everything.
	      (RETURN NIL))
	    (IF (SETQ LOC-NEWLINE (POSITION #\NEWLINE OBJECT-SHORT-NAME))
		(PROGN
		  ;; Split the line up into the part to output, and everything after that.
		  (SETQ PART-TO-OUTPUT (SUBSEQ OBJECT-SHORT-NAME 0 LOC-NEWLINE))
		  (SETQ OBJECT-SHORT-NAME (SUBSEQ OBJECT-SHORT-NAME (1+ LOC-NEWLINE)))
		  (SEND SYS:COLD-LOAD-STREAM :STRING-OUT PART-TO-OUTPUT)
		  ;; Position the cold-load-stream's cursor to the next line.
		  (INCF OUTSIDE-TOP STATUS-LINE-HEIGHT)
		  (SEND SYS:COLD-LOAD-STREAM :SET-CURSORPOS OUTSIDE-LEFT OUTSIDE-TOP))
		;;ELSE
		(PROGN
		  (SETQ PART-TO-OUTPUT OBJECT-SHORT-NAME)
		  (SETQ OBJECT-SHORT-NAME "")
		  (SEND SYS:COLD-LOAD-STREAM :STRING-OUT PART-TO-OUTPUT))))))
      (WHEN STRING
	(SEND SYS:COLD-LOAD-STREAM :STRING-OUT STRING)))))

(DEFFLAVOR INTERACTION-WINDOW ()
   (W:LIST-MOUSE-BUTTONS-MIXIN W:WINDOW))


;;; Define a type which is useful for optional parameters.
(DEFTYPE FLOAT-OR-NIL ()
  "2A new type which is either NIL or a float number.*"
  '(OR NULL (SATISFIES FLOATP)))

(DEFTYPE INTEGER-OR-NIL ()
  "2A new type which is either NIL or a INTEGER number.*"
  '(OR NULL (SATISFIES INTEGERP)))

(DEFUN PICTURE-ERROR (FORMAT-STRING &REST ARGS)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((INTERACTION-PANE (GET-INTERACTION-PANE SELF)))
    (APPLY #'FORMAT INTERACTION-PANE FORMAT-STRING ARGS)
    (TERPRI INTERACTION-PANE)))

(DEFFLAVOR BASIC-OBJECT
           ((X-POSITION NIL :TYPE FLOAT-OR-NIL)
            (Y-POSITION NIL :TYPE FLOAT-OR-NIL)
            ;; T if this object is a standard LaTeX object.
            ;; NIL if this object must be drawn using EEPIC.
            (PORTABLE-OBJECT T :TYPE ATOM))
           (SYS:PRINT-READABLY-MIXIN)
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES
  :INITTABLE-INSTANCE-VARIABLES
  (:DOCUMENTATION "Lowest level flavor for pictures."))

(DEFMETHOD (BASIC-OBJECT :WRITE) (LATEX-STREAM PICTURE-STREAM TEXIFY)
  "2Write the description of this object to a LaTeX file and to an auxilary file.*"
  (WHEN LATEX-STREAM
    (SEND SELF :WRITE-LATEX LATEX-STREAM TEXIFY))
  (WHEN PICTURE-STREAM
    (SEND SELF :WRITE-PICTURE PICTURE-STREAM)))


(DEFMETHOD (BASIC-OBJECT :INSIDE-BOX-P) (LEFT TOP RIGHT BOTTOM WINDOW)
  "Returns non-NIL if this object is within the box defined by LEFT, TOP, RIGHT and BOTTOM.
Returns NIL otherwise."
  (MULTIPLE-VALUE-BIND (OBJECT-LEFT OBJECT-TOP OBJECT-RIGHT OBJECT-BOTTOM)
      (SEND SELF :OBJECT-LIMITS WINDOW)
    (AND (<= LEFT   OBJECT-LEFT)
         (>= RIGHT  OBJECT-RIGHT)
         (>= TOP    OBJECT-TOP)
         (<= BOTTOM OBJECT-BOTTOM))))
  
(DEFFLAVOR POSITION-MIXIN ((POSITION "")) ()
  :INITTABLE-INSTANCE-VARIABLES
  :GETTABLE-INSTANCE-VARIABLES
  (:DOCUMENTATION "Defines a position string to specify the location of text
or the part of an oval"))

(DEFCONSTANT VALID-POSITION-LIST
             '("" "l" "r" "t" "b" "lt" "rt" "lb" "rb" "tl" "tr" "bl" "br")
  "2specifies all valid positions.*")

(DEFCONSTANT VALID-NON-DUPLICATED-POSITION-LIST
             (LOOP WITH POSITION-LIST = (COPY-LIST VALID-POSITION-LIST)
                   FOR DUPLICATE IN '("tl" "tr" "bl" "br")
                   DO (DELETE DUPLICATE POSITION-LIST :TEST 'EQUAL)
                   FINALLY (RETURN POSITION-LIST))
  "2specifies all valid positions.*")

(DEFCONSTANT
  VALID-POSITION-ITEM-LIST
  (LOOP FOR POSITION IN VALID-NON-DUPLICATED-POSITION-LIST
        COLLECTING (IF (ZEROP (LENGTH POSITION))
                       '("Centered" :VALUE ""
                         :DOCUMENTATION "Center the text in the middle of the object")
                       ;;1ELSE*
                       (APPEND
                         (LOOP FOR INDEX FROM 0 BELOW (LENGTH POSITION)
                               FOR CHAR = (AREF POSITION INDEX)
                               WITH STRING-RESULT = ""
                               DO (PROGN (WHEN (NOT (ZEROP INDEX))
                                           (SETQ STRING-RESULT (CONCATENATE
                                                                 'SIMPLE-STRING
                                                                 STRING-RESULT " ")))
                                         (SETQ STRING-RESULT (CONCATENATE
                                                               'SIMPLE-STRING
                                                               STRING-RESULT
                                                               (CASE CHAR
                                                                 (#\l "Left")
                                                                 (#\t "Top")
                                                                 (#\r "Right")
                                                                 (#\b "Bottom")))))
                               FINALLY (RETURN (LIST STRING-RESULT :VALUE POSITION)))
                         `(:DOCUMENTATION
                            ,(CONCATENATE 'SIMPLE-STRING
                                         "Position the text so that it is "
                                         (COND
                                           ((EQUAL POSITION "l")
                                            "left justified, but centered vertically")
                                           ((EQUAL POSITION "r")
                                            "right justified and centered vertically")
                                           ((EQUAL POSITION "t")
                                            "at the top and centered horizontally")
                                           ((EQUAL POSITION "b")
                                            "at the bottom and centered horizontally")
                                           ((EQUAL POSITION "lt")
                                            "left justified in the upper left corner")
                                           ((EQUAL POSITION "rt")
                                            "right justified in the upper right corner")
                                           ((EQUAL POSITION "lb")
                                            "left justified in the lower left corner")
                                           ((EQUAL POSITION "rb")
                                            "right justified in the lower right corner")))))))
  "2Same as VALID-NON-DUPLICATED-POSITION-LIST but suitable for use as a menu item-list.*")

(DEFMETHOD (POSITION-MIXIN :SET-POSITION) (NEW-POSITION)
  (SEND SELF :SET :POSITION NEW-POSITION))

(DEFMETHOD (POSITION-MIXIN :CASE :SET :POSITION) (NEW-POSITION)
  ;; Validate the position.  The only valid possibilities are:
  ;; empty l r t b lt rt lb rb tl tr bl and br.
  (WHEN (MEMBER NEW-POSITION VALID-POSITION-LIST :TEST #'STRING-EQUAL)
    (SETQ POSITION NEW-POSITION)))


(DEFFLAVOR TEXT-MIXIN
           ((TEXT "" :TYPE SIMPLE-STRING)
            ;; This is either NIL or the two letter LaTeX designation for a font.
            ;; For example, bf for a boldface font.
            (TEXT-FONT NIL))
           ()
  :SETTABLE-INSTANCE-VARIABLES
  :INITTABLE-INSTANCE-VARIABLES
  :GETTABLE-INSTANCE-VARIABLES)

(DEFPARAMETER LAST-CHOSEN-FONT NIL
  "Used to remember what font the user selected last.  The user will most
likely select the same font every time, so this will reduce mouse movement.")

(DEFUN POPUP-FONT-MENU (&OPTIONAL (DEFAULT-TOO T) (LABEL  "Specify font for this object"))
  (DECLARE (VALUES FONT-NAME ITEM))
  (LET ((OLD-X W:MOUSE-X)
        (OLD-Y W:MOUSE-Y)
        (ITEMS (LOOP FOR (FONT-NAME NIL) IN *FONT-NAME-TO-FONT-MAPPING*
                    FOR DESCRIPTIVE-NAME = (CAR (MEMBER FONT-NAME
                                                        *DESCRIPTIVE-FONT-NAMES*
                                                        :KEY #'(LAMBDA (X) (CAR X))
                                                        :TEST #'STRING-EQUAL))
                    FOR ITEM = `(,(SECOND DESCRIPTIVE-NAME) :VALUE ,FONT-NAME
                                 :FONT ,(LOOKUP-EXPLORER-FONT FONT-NAME)
                                 :DOCUMENTATION ,(FORMAT NIL (THIRD DESCRIPTIVE-NAME)
                                                         (DESCRIPTIVE-FONT-NAME
                                                           *DEFAULT-DISPLAY-FONT*)))
                    COLLECT ITEM))
        LAST-CHOSEN-ITEM)
    ;; Determine which item was chosen last time.
    (SETQ LAST-CHOSEN-ITEM (LOOP FOR ITEM IN ITEMS
                            WHEN (STRING-EQUAL (CAR LAST-CHOSEN-FONT) (CAR ITEM))
                            RETURN ITEM
                            FINALLY (RETURN NIL)))
    (MULTIPLE-VALUE-BIND (FONT-NAME ITEM)
        (W:MENU-CHOOSE (IF DEFAULT-TOO
                           ITEMS
                           ;;ELSE
                           (CDR ITEMS))
                       :LABEL LABEL
                       :DEFAULT-ITEM LAST-CHOSEN-ITEM)
      (SETQ W:MOUSE-X OLD-X
            W:MOUSE-Y OLD-Y)
      (WHEN FONT-NAME
        (SETQ LAST-CHOSEN-FONT ITEM))
      (VALUES FONT-NAME ITEM))))

(DEFUN GET-FONT-DATA (OBJECT)
  "2Obtain font data from the user.
OBJECT is the object instance."*
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (MULTIPLE-VALUE-BIND (FONT-NAME IGNORE)
      (POPUP-FONT-MENU)
    (WHEN FONT-NAME
      (WHEN (EQ FONT-NAME :DEFAULT-FONT)
        (SETQ FONT-NAME NIL))
      (WHEN (PLUSP (LENGTH (SEND OBJECT :TEXT)))
        ;; Erase the text in the old font.
        (SEND OBJECT :DRAW-TEXT SELF W:ALU-XOR))
      (SEND OBJECT :SET-TEXT-FONT FONT-NAME)
      ;; Draw the text in the newly selected font.
      (WHEN (PLUSP (LENGTH (SEND OBJECT :TEXT)))
        (SEND OBJECT :DRAW-TEXT SELF)))))

(DEFUN NEW-DEFAULT-FONT ()
  "Change the default font for new text objects."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (MULTIPLE-VALUE-BIND (FONT-NAME ITEM)
          (POPUP-FONT-MENU NIL "Specify default font for objects")
        (WHEN FONT-NAME
          (LET ((NEW-DISPLAY-FONT (GETF (CDR ITEM) :FONT)))
            ;; Change the lookup table and its associated constant.
            (SETF (CADR (ASSOC :DEFAULT-FONT *FONT-NAME-TO-FONT-MAPPING*)) NEW-DISPLAY-FONT)
            (SETQ *DEFAULT-DISPLAY-FONT* NEW-DISPLAY-FONT)))
        ;; Show the text that defaulted in the new font'
        (SEND SELF :REFRESH))))

;;; The GENERATE-ACTION-ITEM-SEQUENCE function is used to sequence through menu
;;; items.  The ITEM-LIST argument specifies an item list for a menu.  The
;;; ACTION-SEQUENCE specifies sequencing information relative to that item list.
;;; Both parts of each element of the action sequence (i.e. the CAAR and CADAR
;;; of the action sequence) specify an item in the menu item list.  If we are on
;;; the item specified by the first part then the next menu item to become the
;;; default item will be the item specified by the second part.  For example (1 2)
;;; as an element of the action sequence says that if we just selected the first
;;; item in the menu, then the default for the next call to menu-choose will be the
;;; second item.  NIL as an element is for the case where menu-choose is called for
;;; the first time.  The reason for all of this hair is to make it obvious which
;;; items need to be selected first and to mimimize mouse movement.  For example,
;;; for all objects, the initial position needs to be specified before anything
;;; else can be chosen.
(DEFUN GENERATE-ACTION-ITEM-SEQUENCE (ITEM-LIST ACTION-SEQUENCE)
  "2Generate a sequence showing what happens when one presses a menu item.*"
  (LOOP FOR (FROM-ITEM-NUMBER TO-ITEM-NUMBER) IN ACTION-SEQUENCE
        COLLECTING (LIST (IF FROM-ITEM-NUMBER
                             (NTH (1- FROM-ITEM-NUMBER) ITEM-LIST)
                             ;;1ELSE*
                             NIL)
                         (NTH (1- TO-ITEM-NUMBER) ITEM-LIST))))

(DEFUN GET-OBJECT-DATA (OBJECT MENU-ITEM-LIST MENU-LABEL ITEM-SEQUENCING)
  "2Obtain object data from the user.
OBJECT is the object instance.
MENU-ITEM-LIST is an item list for MENU-CHOOSE specifying what can be done with the object.
MENU-LABEL is the label for MENU-CHOOSE
ITEM-SEQUENCING is a list specifying the sequencing through the menu items."*
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LOOP WITH ACTION-ITEM-LIST = NIL
            FOR DEFAULT-ITEM FIRST (CADR (ASSOC NIL ITEM-SEQUENCING :TEST #'EQ)) THEN
            (CADR (ASSOC ACTION-ITEM-LIST ITEM-SEQUENCING :TEST #'EQ))
            FOR ACTION = (MULTIPLE-VALUE-BIND (ACTION ITEM)
                             (W:MENU-CHOOSE MENU-ITEM-LIST :LABEL MENU-LABEL
                                            :DEFAULT-ITEM DEFAULT-ITEM)
                           (SETQ ACTION-ITEM-LIST ITEM)
                           ACTION)
            FOR ARGS   = NIL
            DO (WHEN (NOT (NULL ACTION))
                 (WHEN (EQ ACTION :ABORT)
                   (RETURN NIL))
                 (WHEN (EQ ACTION :DONE)
                   (RETURN T))
                 (WHEN (LISTP ACTION)
                   (PSETQ ACTION (CAR ACTION)
                          ARGS   (CDR ACTION)))
                 (APPLY (SYMBOL-FUNCTION ACTION) OBJECT ARGS)))))


(DEFUN MOUSE-SPECIFY-POINT (OBJECT NEW-BLINKER MOUSE-DOCUMENTATION &OPTIONAL (IGNORE-CHARACTERS T))
  "2Get an (X,Y) coordinate from the user, return it as two values.
These are relative to the (x,y) pixel origin, and Y increases in the upward direction.
NEW-BLINKER is the blinker associated with this `mode'.
MOUSE-DOCUMENTATION is a mouse documentation `string' for this mode.
IGNORE-CHARACTERS is an optional hack.  When T (the default) only the ABORT character is handled,*
	2and all other characters to be ignored.*
	2When NIL this will cause the ABORT character to be handled, and all other characters*
	2to be put back into the I/O buffer.  This hack allows the `sticky' mode to work in the*
	2edit command and yet allow one to enter characters for other commands.*"
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (SETQ CURRENT-MOUSE-BLINKER       NEW-BLINKER
        CURRENT-MOUSE-DOCUMENTATION MOUSE-DOCUMENTATION)
  (MULTIPLE-VALUE-PROG1
    (LOOP FOR INPUT = (CONDITION-CASE ()
                          (PROGN
                           (SEND SELF :TURN-ON-MOUSE-POINTING OBJECT)
                           (PROG1
                             (SEND SELF :ANY-TYI)
                             (SEND SELF :TURN-OFF-MOUSE-POINTING)))
                        (EH:ABORT #\ABORT))
          DO (PROGN
               (WHEN (OR (INTEGERP INPUT) (CHARACTERP INPUT))
                 (WHEN (CHAR= INPUT #\ABORT)
                   (RETURN NIL)2)*
                 ;; Unknown character.  Put it back into the I/O buffer so it
                 ;; will be handled later.  Hopefully this is the right thing
                 ;; to do.
                 (WHEN (NULL IGNORE-CHARACTERS)
                   (SEND SELF :FORCE-KBD-INPUT INPUT)
                   (RETURN NIL)))
               (WHEN (LISTP INPUT)
                 (COND ((EQ (CAR INPUT) :MOUSE-BUTTON)
                        (WHEN (CHAR= (SECOND INPUT) #\MOUSE-M-1)
                          ;; User aborted out.
                          (RETURN NIL))
                        (LET (NEW-X NEW-Y)
                          (RETURN
                            ;; We have to do this weird progn because we only
                            ;; want to return 2 values.
                            (PROGN
                              (MULTIPLE-VALUE-SETQ (NEW-X NEW-Y)
                                (SEND SELF :GRIDIFY-POINT
                                      (- (FOURTH INPUT) (W:SHEET-INSIDE-LEFT SELF) X-PIXEL-ORIGIN)
                                      (- Y-PIXEL-ORIGIN (- (FIFTH  INPUT)
                                                           (W:SHEET-INSIDE-TOP SELF)))))
                              NEW-X)
                            NEW-Y)))
                       ((EQ (CAR INPUT) :MENU)
                        ;; We have a menu item.  We need put the blip back into the I/O
                        ;; buffer and then get out of this function, returning NIL.
                        (SEND SELF :FORCE-KBD-INPUT INPUT)
                        (RETURN NIL))
                       (T ;; Unknown input.  Ignore it.
                        )))))
    (SETQ CURRENT-MOUSE-BLINKER       NIL
          CURRENT-MOUSE-DOCUMENTATION NIL)
    ;; Make sure that the mouse changes too.
    (SEND SELF :MOUSE-STANDARD-BLINKER)))


(DEFUN TOGGLE-READ-ONLY-PICTURE ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (SETQ READ-ONLY-PICTURE (NOT READ-ONLY-PICTURE)))

(DEFCONSTANT RETURN-STRING (FORMAT NIL "~%"))

(DEFCONSTANT LATEX-TEXT-MAPPINGS
             '((#\\ "$\\backslash$")
               (#\{ "$\\{$")
               (#\} "$\\}$")
               (#\> "$>$")
               (#\< "$<$")
               (#\^ "\\^{}")
               (#\~ "\\~{}")
               (#\$ "\\$")
               (#\& "\\&")
               (#\# "\\#")
               (#\_ "\\_")
               (#\% "\\%"))
  "2Used to map reserved LaTe*X2 characters into printable characters.*")

(DEFUN TEXIFY (TEXT FONT-NAME TRANSLATE)
  "2Convert TEXT to a form suitable for LaTeX.*"
  ;; Remove trailing returns.  LaTeX doesn't handle them well.  Even though
  ;; the text entry function removes them, we do this here for old files
  ;; that did this too.
  (SETQ TEXT (STRING-RIGHT-TRIM #\RETURN TEXT))
  (WHEN TRANSLATE
    ;; Perform a simple mapping of single characters to their LaTeX equivalent.
    (SETQ TEXT (LOOP FOR INDEX FROM 0 BELOW (LENGTH TEXT)
                     FOR CHAR = (AREF TEXT INDEX)
                     WITH NEW-TEXT = ""
                     WITH MATCH-LOCATION = NIL
                     DO (PROGN
                          (SETQ MATCH-LOCATION (ASSOC CHAR LATEX-TEXT-MAPPINGS :TEST #'CHAR=))
                          (SETQ NEW-TEXT (CONCATENATE 'SIMPLE-STRING
                                                      NEW-TEXT
                                                      (IF MATCH-LOCATION
                                                          (CADR MATCH-LOCATION)
                                                          ;;1ELSE*
                                                          (STRING CHAR)))))
                     FINALLY (RETURN NEW-TEXT))))
  ;; Go through another loop putting the \\ between every line.  Also, if a font is
  ;; specified, it can only be present on a single line.  If there are multiple lines
  ;; then we need to do it for each line separately.
  (LOOP FOR START-INDEX = 0 THEN (+ LINE-BREAK-INDEX 1)
          FOR LINE-BREAK-INDEX = (POSITION #\RETURN TEXT :TEST #'CHAR= :START START-INDEX)
          WITH CURLY-FONT = (CONCATENATE 'SIMPLE-STRING "{\\" FONT-NAME " ")
          WITH NEW-LINES = ""
          DO (PROGN
               (IF LINE-BREAK-INDEX
                   (SETQ NEW-LINES (CONCATENATE
                                     'SIMPLE-STRING
                                     NEW-LINES
                                     (WHEN FONT-NAME
                                       CURLY-FONT)
                                     (SUBSEQ TEXT START-INDEX LINE-BREAK-INDEX)
                                     (WHEN FONT-NAME
                                       "}")
                                     " \\\\" RETURN-STRING))
                   ;;1ELSE*
                   (PROGN
                     (SETQ NEW-LINES (CONCATENATE
                                       'SIMPLE-STRING
                                       NEW-LINES
                                     (WHEN FONT-NAME
                                       CURLY-FONT)
                                       (SUBSEQ TEXT START-INDEX LINE-BREAK-INDEX)
                                     (WHEN FONT-NAME
                                       "}")))
                     (RETURN NEW-LINES))))
          FINALLY (RETURN NEW-LINES)))

(DEFUN READ-MOUSE-XY-INPUT ()
  "2Read input from the window and return mouse information when obtained.
Also handles the abort key and aborting via the mouse with mouse middle.
These are relative to the (x,y) pixel origin, and Y increases in the upward direction.*"
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LOOP FOR INPUT = (CONDITION-CASE ()
                        (SEND SELF :ANY-TYI)
                      (EH:ABORT #\ABORT))
        DO (PROGN
             (WHEN (AND (OR (INTEGERP INPUT) (CHARACTERP INPUT))
                        (CHAR= INPUT #\ABORT))
               (RETURN NIL))
             (WHEN (LISTP INPUT)
               (COND ((EQ (CAR INPUT) :MOUSE-BUTTON)
                      INPUT (CAR INPUT) (EQ (CAR INPUT) :MOUSE-BUTTON)
                      (WHEN (CHAR= (SECOND INPUT) #\MOUSE-M-1)
                        ;; User aborted out.
                        (RETURN NIL))
                      (LET (NEW-X NEW-Y)
                        (RETURN
                          ;; We have to do this weird progn because we only
                          ;; want to return 2 values.
                          (PROGN
                            (MULTIPLE-VALUE-SETQ (NEW-X NEW-Y)
                              (SEND SELF :GRIDIFY-POINT
                                    (- (FOURTH INPUT) (W:SHEET-INSIDE-LEFT SELF) X-PIXEL-ORIGIN)
                                    (- Y-PIXEL-ORIGIN (- (FIFTH  INPUT)
                                                         (W:SHEET-INSIDE-TOP SELF)))))
                            NEW-X)
                          NEW-Y)))
                     (T ;; Unknown input.  Ignore it.
                      ))))))

;;; Fixed up version of MOUSE-Y-OR-N-P that puts the mouse on the Yes item, and to
;;; also save/restore the mouse position.  In this way, we can minimize the amount
;;; of user interaction and confusion at the same time.
(DEFUN MOUSE-Y-OR-N-P (MESSAGE)
  (LET ((OLD-MOUSE-X TV:MOUSE-X)
        (OLD-MOUSE-Y TV:MOUSE-Y)
        (YES-ITEM '(:YES :VALUE T :DOCUMENTATION		
                         "Click mouse or press <END> to select YES.    Pressing <ABORT> or moving off menu will also select NO.")))
    (PROG1
      (W:MENU-CHOOSE
        `((:NO :VALUE NIL :DOCUMENTATION
               "Click mouse or press <END> to select NO.   Pressing <ABORT> or moving off menu will also select NO.")
          ,YES-ITEM)
        :LABEL MESSAGE :SCROLLING-P NIL
        :DEFAULT-ITEM YES-ITEM
        :COLUMNS 2)
      (SETQ TV:MOUSE-X OLD-MOUSE-X
            TV:MOUSE-Y OLD-MOUSE-Y))))

(DEFUN ENTER-TEXT-LINES (OBJECT LABEL)
  "Get lines of text from the user."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((LINES (GET-TEXT-LINES LABEL (SEND OBJECT :TEXT))))
    (WHEN (AND (NOT (NULL LINES)) (PLUSP (LENGTH LINES)))
      (WHEN (EQUAL (AREF (SUBSEQ LINES (1- (LENGTH LINES))) 0) #\RETURN)
        (PICTURE-ERROR "Warning: LaTeX does not handle trailing RETURNs.~
                      ~%Trailing RETURNs removed.~%")
        (SETQ LINES (STRING-RIGHT-TRIM #\RETURN LINES)))
      (SETF (SEND OBJECT :TEXT) LINES))
    LINES))

(DEFPARAMETER LAST-CHOSEN-POSITION NIL
  "Remember the last position the user selected to try an reduce the amount of mouse movement.")

(DEFUN GET-POSITION-DATA (OBJECT)
  "Get position data from the user."
  (LET ((OLD-X W:MOUSE-X)
        (OLD-Y W:MOUSE-Y))
    (MULTIPLE-VALUE-BIND (POSITION ITEM)
        (W:MENU-CHOOSE VALID-POSITION-ITEM-LIST
                       :LABEL "Specify Position of Text"
                       :DEFAULT-ITEM LAST-CHOSEN-POSITION)
      (SETQ W:MOUSE-X OLD-X
            W:MOUSE-Y OLD-Y)
      (WHEN (NOT (NULL POSITION))
        (SETQ LAST-CHOSEN-POSITION ITEM)
        (SETF (SEND OBJECT :POSITION) POSITION))
      POSITION)))

;;; The following function is called when something changes in the display list.
;;; For example, an object which is deleted may change these bounds.
(DEFUN RECALCULATE-BOUNDS ()
  "Recalculate the max/min x/y bounds of the picture."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (SEND SELF :SET-MAXIMUM-X-UNIT 0.0)
  (SEND SELF :SET-MAXIMUM-Y-UNIT 0.0)
  (SEND SELF :SET-MINIMUM-X-UNIT MOST-POSITIVE-SHORT-FLOAT)
  (SEND SELF :SET-MINIMUM-Y-UNIT MOST-POSITIVE-SHORT-FLOAT)
  ;; Calculate the max/max-mum x/y units based upon the updated
  ;; display list.  The easiest way to make sure that we have the
  ;; right values for these is to recalculate them now.
  (LOOP FOR OBJECT IN DISPLAY-LIST
        DO (WHEN (NOT (EQ OBJECT MARGIN-OBJECT))
             (MULTIPLE-VALUE-BIND (LEFT TOP RIGHT BOTTOM)
                 (SEND OBJECT :OBJECT-LIMITS SELF)
               (SETQ MAXIMUM-X-UNIT (MAX MAXIMUM-X-UNIT RIGHT )
                     MINIMUM-X-UNIT (MIN MINIMUM-X-UNIT LEFT  )
                     MAXIMUM-Y-UNIT (MAX MAXIMUM-Y-UNIT TOP   )
                     MINIMUM-Y-UNIT (MIN MINIMUM-Y-UNIT BOTTOM))))))

(DEFPARAMETER PICTURE-PATHNAME-DEFAULTS NIL)

(DEFUN MAKE-PATHNAME-FORCE-TYPE (PATHNAME TYPE)
  "Construct a pathname but force its type."
  (LET* ((SIMPLE-PATHNAME (MERGE-PATHNAMES PATHNAME (OR PICTURE-PATHNAME-DEFAULTS
                                                       *DEFAULT-PATHNAME-DEFAULTS*)))
         (REAL-PATHNAME (MAKE-PATHNAME :HOST      (PATHNAME-HOST      SIMPLE-PATHNAME)
                                       :DEVICE    (PATHNAME-DEVICE    SIMPLE-PATHNAME)
                                       :DIRECTORY (PATHNAME-DIRECTORY SIMPLE-PATHNAME)
                                       :NAME      (PATHNAME-NAME      SIMPLE-PATHNAME)
                                       :TYPE      TYPE
                                       :VERSION   (PATHNAME-VERSION   SIMPLE-PATHNAME))))
    (SETQ PICTURE-PATHNAME-DEFAULTS REAL-PATHNAME)
    REAL-PATHNAME))

(DEFUN MAKE-PATHNAME-DEFAULT (TYPE)
  (MAKE-PATHNAME-FORCE-TYPE (OR PICTURE-PATHNAME-DEFAULTS
                                (FS:DEFAULT-PATHNAME))
                            TYPE))

(DEFUN WRITE-TEX-FILE (&OPTIONAL (TEXIFY T))
  "2Write the entire picture out to disk for LaTeX.*"
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((WINDOW (GET-INTERACTION-PANE SELF))
        LATEX-PATHNAME
        PICTURE-PATHNAME)
    (FORMAT WINDOW "Enter pathname (~A)"
            (MAKE-PATHNAME-DEFAULT LATEX-EXTENSION))
    (SETQ LATEX-PATHNAME   (MAKE-PATHNAME-FORCE-TYPE (READ-LINE WINDOW)
                                                       LATEX-EXTENSION)
          PICTURE-PATHNAME (MAKE-PATHNAME-FORCE-TYPE LATEX-PATHNAME
                                                       PICTURE-EXTENSION))
  (WITH-OPEN-FILE (LATEX-STREAM LATEX-PATHNAME :DIRECTION :OUTPUT)
    (WITH-OPEN-FILE (PICTURE-STREAM PICTURE-PATHNAME :DIRECTION :OUTPUT)
      ;; Output a leading parenthesis so that we can later read in the
      ;; whole list in one whack.
      (SEND PICTURE-STREAM :LINE-OUT "(")
      (FORMAT LATEX-STREAM "% -*- Mode:TeX -*-~%")
      (WHEN (LOOP FOR OBJECT IN DISPLAY-LIST
                  FOR PORTABLE = (SEND OBJECT :PORTABLE-OBJECT)
                  DO (WHEN (NOT PORTABLE)
                       (RETURN T))
                  FINALLY (RETURN NIL))
        (FORMAT LATEX-STREAM "%--------------------------------------------------------~%~
                              % This file contains non-standard LaTeX picture commands.~%~
                              % The LaTeX eepic `style' macros must be used.  The~%~
                              % documentstyle LaTeX command should look something like:~%~
                              % \\documentstyle[epic,eepic]{report}~%~
                              %--------------------------------------------------------~%"))
      ;; If we are in landscape mode then assume that we are also using SliTeX.
      ;; The fonts used by SliTeX are rather large for a figure.  Shifting into
      ;; a smaller font makes them appear in a better size.
      (WHEN (EQ *CURRENT-ORIENTATION* :LANDSCAPE)
        (FORMAT LATEX-STREAM "\\begin{~A}~%" *DEFAULT-LANDSCAPE-LATEX-FONT*))
      (FORMAT LATEX-STREAM "\\setlength{\\unitlength}{~Fin}~%~
              \\begin{picture}(~F,~F)(~F,~F)~%~
              {\\~:[thick~;thin~]lines~%"
              (/ 1 (UNITS-PER-INCH BASE-UNIT))
              ;; Allow for compatibility for those pictures which didn't have
              ;; a MINIMUM-X/Y-UNIT.
              (IF (>= MINIMUM-X-UNIT MOST-POSITIVE-SHORT-FLOAT)
                  MAXIMUM-X-UNIT
                  ;;1ELSE*
                  (- MAXIMUM-X-UNIT MINIMUM-X-UNIT))
              (IF (>= MINIMUM-Y-UNIT MOST-POSITIVE-SHORT-FLOAT)
                  MAXIMUM-Y-UNIT
                  ;;1ELSE*
                  (- MAXIMUM-Y-UNIT MINIMUM-Y-UNIT))
              (IF (>= MINIMUM-X-UNIT MOST-POSITIVE-SHORT-FLOAT)
                  0
                  ;;1ELSE*
                  MINIMUM-X-UNIT)
              (IF (>= MINIMUM-Y-UNIT MOST-POSITIVE-SHORT-FLOAT)
                  0
                  ;;1ELSE*
                  MINIMUM-Y-UNIT)
              (= LINE-THICKNESS *THIN-LINES*))
      ;; Generate the descriptions of every object except for the margin
      ;; object.  That one isn't really supposed to be something that
      ;; we want LaTeX to see.
      (LOOP FOR OBJECT IN DISPLAY-LIST
            DO (IF (NOT (EQ OBJECT MARGIN-OBJECT))
                 (SEND OBJECT :WRITE LATEX-STREAM PICTURE-STREAM TEXIFY)
                 ;;1ELSE*
                 (SEND OBJECT :WRITE-PICTURE PICTURE-STREAM)))
      (SEND PICTURE-STREAM :LINE-OUT ")")
      (SEND PICTURE-STREAM :LINE-OUT
            (FORMAT NIL "(PROGN (SETQ PICTURE:*CURRENT-ORIENTATION* ~S)~% ~
                                (PICTURE:SET-ORIENTATION-~A (SEND W:SELECTED-WINDOW :SUPERIOR))"
                    *CURRENT-ORIENTATION* *CURRENT-ORIENTATION*))
      (SEND PICTURE-STREAM :LINE-OUT
            (FORMAT NIL "(SEND SELF :SET-LINE-THICKNESS ~D NIL)" LINE-THICKNESS))
      (SEND PICTURE-STREAM :LINE-OUT
            (FORMAT NIL "(SEND SELF :SET-BASE-UNIT :~A)" BASE-UNIT))
      (LOOP FOR OBJECT IN DISPLAY-LIST
            FOR INDEX FROM 0 BY 1
            WHEN (EQ OBJECT MARGIN-OBJECT)
            DO (PROGN
                 (SEND PICTURE-STREAM :LINE-OUT
                       (FORMAT NIL "(SEND SELF :SET-MARGIN-OBJECT (NTH ~D DISPLAY-LIST))"
                               INDEX))
                 (SEND PICTURE-STREAM :LINE-OUT
                       (FORMAT NIL "(SEND SELF :SET-MARGIN-UNIT :~A)" MARGIN-UNIT))
                 (SEND PICTURE-STREAM :LINE-OUT
                       (FORMAT NIL "(SEND SELF :SET-LEFT-MARGIN ~D)" LEFT-MARGIN))
                 (SEND PICTURE-STREAM :LINE-OUT
                       (FORMAT NIL "(SEND SELF :SET-TOP-MARGIN ~D)" TOP-MARGIN))
                 (SEND PICTURE-STREAM :LINE-OUT
                       (FORMAT NIL "(SEND SELF :SET-RIGHT-MARGIN ~D)" RIGHT-MARGIN))
                 (SEND PICTURE-STREAM :LINE-OUT
                       (FORMAT NIL "(SEND SELF :SET-BOTTOM-MARGIN ~D)" BOTTOM-MARGIN))
                 (RETURN NIL)))
      (SEND PICTURE-STREAM :LINE-OUT ")")
      (FORMAT LATEX-STREAM "}~%~
                            \\end{picture}~%")
      (WHEN (EQ *CURRENT-ORIENTATION* :LANDSCAPE)
        ;; Close off the landscape font scope.
        (FORMAT LATEX-STREAM "\\end{~A}~%" *DEFAULT-LANDSCAPE-LATEX-FONT*))))))

;;; The following function is a hack which translates any negative
;;; dimension of an object to be a positive dimension with the related
;;; position value adjusted appropriately.  The reason this is done
;;; is because LaTeX doesn't handle negative dimensions properly.
(DEFUN NORMALIZE-OBJECT (LIST)
  "2Make sure that the object is defined properly.*"
  (LET ((OBJECT-HEIGHT (GETF LIST :HEIGHT)))
    (WHEN (AND OBJECT-HEIGHT (MINUSP OBJECT-HEIGHT))
      (SETF (GETF LIST :HEIGHT) (- OBJECT-HEIGHT))
      (DECF (GETF LIST :Y-POSITION) OBJECT-HEIGHT)))
  (LET ((OBJECT-WIDTH (GETF LIST :WIDTH)))
    (WHEN (AND OBJECT-WIDTH (MINUSP OBJECT-WIDTH))
      (SETF (GETF LIST :WIDTH) (- OBJECT-WIDTH))
      (INCF (GETF LIST :X-POSITION) OBJECT-WIDTH)))
  LIST)

(DEFUN READ-PICTURE-FILE (&OPTIONAL (INPUT-STREAM NIL))
  "2Read the entire picture from disk from the PICture file.*"
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LET* ((OUTPUT-TOO (NOT INPUT-STREAM))
	     (WINDOW (OR INPUT-STREAM (GET-INTERACTION-PANE SELF)))
	     PICTURE-PATHNAME)
	(WHEN OUTPUT-TOO
	  (FORMAT WINDOW "Enter pathname (~A)"
		  (MAKE-PATHNAME-DEFAULT PICTURE-EXTENSION)))
        (SETQ PICTURE-PATHNAME (MAKE-PATHNAME-FORCE-TYPE (READ-LINE WINDOW) PICTURE-EXTENSION))
        (IF (NULL (PROBE-FILE PICTURE-PATHNAME))
            (WHEN OUTPUT-TOO
	      (PICTURE-ERROR "The file ~a was not found." PICTURE-PATHNAME))
            ;;1ELSE*
            (WITH-OPEN-FILE (PICTURE-STREAM PICTURE-PATHNAME)
              ;; We are going to calculate these, so set them to `zero'.
              (SEND SELF :SET-MAXIMUM-X-UNIT 0.0)
              (SEND SELF :SET-MAXIMUM-Y-UNIT 0.0)
              (SEND SELF :SET-MINIMUM-X-UNIT MOST-POSITIVE-SHORT-FLOAT)
              (SEND SELF :SET-MINIMUM-Y-UNIT MOST-POSITIVE-SHORT-FLOAT)
              ;; The first form in the picture file is a list of MAKE-INSTANCEs to create the
              ;; objects as they were earlier.
              (SETQ DISPLAY-LIST (LOOP FOR ELEMENT IN (READ PICTURE-STREAM)
                                       FOR OBJECT = (EVAL (NORMALIZE-OBJECT ELEMENT))
                                       COLLECTING OBJECT))
              ;; The last form in the picture file is simply a form to evaluate.
              (LET ((*PACKAGE* (FIND-PACKAGE 'PICTURE)))
                (SEND SELF :EVAL-INSIDE-YOURSELF (READ PICTURE-STREAM)))
              ;; Calculate the max/max-mum x/y units based upon the display list.
              ;; The maximums used to be stored in the .PIC file but the minimums
              ;; didn't make it (a bug).  The easiest way to make sure that we
              ;; have the right values for these is to calculate them now.
              (LOOP FOR OBJECT IN DISPLAY-LIST
                    DO (WHEN (NOT (EQ OBJECT MARGIN-OBJECT))
                         (MULTIPLE-VALUE-BIND (LEFT TOP RIGHT BOTTOM)
                             (SEND OBJECT :OBJECT-LIMITS SELF)
                           (SETQ MAXIMUM-X-UNIT (MAX MAXIMUM-X-UNIT RIGHT )
                                 MINIMUM-X-UNIT (MIN MINIMUM-X-UNIT LEFT  )
                                 MAXIMUM-Y-UNIT (MAX MAXIMUM-Y-UNIT TOP   )
                                 MINIMUM-Y-UNIT (MIN MINIMUM-Y-UNIT BOTTOM)))))
              ;; Some old pictures have objects that don't have a thickness.  Make sure
              ;; that the objects which have a thickness get the default.
              ;; We have to wait to do the refresh now because we may have
              ;; changed units in the file.
              (SEND SELF :REFRESH))))))

;;; Simple utility functions
(DEFUN SIGN (NUMBER)
  (IF (PLUSP NUMBER)
      1
      ;;1ELSE*
      (IF (MINUSP NUMBER)
          -1
          ;;ELSE
          0)))
(DEFUN FORCE-SIGN (NUMBER SIGN)
  (IF (NOT (= (SIGN NUMBER) SIGN))
    (IF (ZEROP SIGN)
        0
        ;;1ELSE*
        (- NUMBER))
    ;;1ELSE*
    NUMBER))


;;; Debug functions.

(DEFUN GET-INSTANCE ()
  "Return the instance of a picture that is currently being used."
  (LOOP FOR OBJECT-INDEX FROM 0 BELOW (LENGTH W:PREVIOUSLY-SELECTED-WINDOWS)
        FOR OBJECT = (AREF W:PREVIOUSLY-SELECTED-WINDOWS OBJECT-INDEX)
        WHEN (TYPEP OBJECT 'PICTURE-FRAME)
        DO (RETURN OBJECT)))
;;;  (LOOP WITH RESOURCE-OBJECT = (GET 'PICTURE-FRAME 'DEFRESOURCE)
;;;        WITH N-OBJECTS = (SI:RESOURCE-N-ENTRIES RESOURCE-OBJECT)
;;;        FOR OBJECT-INDEX FROM 0 BY 3 BELOW (* 3 N-OBJECTS)
;;;        FOR STATUS   = (AREF RESOURCE-OBJECT    OBJECT-INDEX)
;;;        FOR OBJECT   = (AREF RESOURCE-OBJECT (+ OBJECT-INDEX 1))
;;;        FOR SUPERIOR = (AREF RESOURCE-OBJECT (+ OBJECT-INDEX 2))
;;;        WHILE (NULL STATUS)
;;;        FINALLY (RETURN OBJECT)))

(DEFUN GET-DISPLAY-LIST ()
  (SEND (SEND (GET-INSTANCE) :GET-PANE 'PICTURE-PANE)
        :DISPLAY-LIST))

(DEFUN DESCRIBE-DISPLAY-LIST ()
  (LOOP FOR OBJECT IN (GET-DISPLAY-LIST)
        DO (DESCRIBE OBJECT)))
