;;; -*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;;  1/03/89    TWE	Changed move-object and copy-object to pop up a menu when the
;;;			object being picked is not unique.
;;; 10/14/88    TWE	Put support in place for read-only mode.
;;;  7/17/87    TWE	Fixed up the copy and move code so that the min/max x/y picture
;;;			gets calculated.  Previously these didn't get updated.  Also
;;;			changed so that the user could see the position the mouse is on
;;;			when mouse-specify-point gets called.
;;;  6/02/87    TWE	Changed copy and move to use the new macro WITH-GRID-DISABLED to
;;;			make object selection work better.  Changed the mouse glyphs to
;;;			be more like GED.
;;;  6/01/87    TWE	Fixed up move-object to handle the abort case better.
;;;  5/26/87    TWE	Added code to move an object.
;;;  5/21/87	TWE	Initial creation.

#|

This file defines the commands that perform copying and moving of objects.

|#

(DEFCONSTANT COPY-NEW-OBJECT-MOUSE-DOCUMENTATION
             '(:MOUSE-L-1 "Copy the object to this position."
               :MOUSE-M-1 "Abort out of this operation"))
             

(DEFCONSTANT COPY-OBJECT-MOUSE-DOCUMENTATION
             '(:MOUSE-L-1 "Pick an object to copy"
               :MOUSE-M-1 "Abort out of this operation"))

(DEFUN COPY-OBJECT ()
  "Handles all interaction for copying objects."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LOOP WITH NEW-OBJECT = NIL
            WITH X = NIL
            WITH Y = NIL
            DO (PROGN
                 (WITH-GRID-DISABLED
                   (MULTIPLE-VALUE-SETQ (X Y)
                     (MOUSE-SPECIFY-POINT NIL W:MOUSE-GLYPH-DOUBLE-UP-ARROW-LETTERED-D
                                          COPY-OBJECT-MOUSE-DOCUMENTATION NIL)))
                 (WHEN (NULL X)
                   (RETURN NIL))
                 (SETQ X (TRANSLATE-PIXEL-TO-UNIT X)
                       Y (TRANSLATE-PIXEL-TO-UNIT Y))
                 (LET ((OBJECT-TO-COPY (LOOP FOR OBJECT IN DISPLAY-LIST
                                             WITH OBJECT-DISTANCE = NIL
                                             WITH NEAREST-OBJECTS = NIL
                                             WITH NEAREST-DISTANCE = MOST-POSITIVE-SHORT-FLOAT
                                             WHEN (NOT (EQ OBJECT MARGIN-OBJECT))
                                             DO (PROGN
                                                  (SETQ OBJECT-DISTANCE (SEND OBJECT
                                                                              :DISTANCE-FROM-POINT
                                                                              SELF X Y))
                                                  (COND ((< OBJECT-DISTANCE NEAREST-DISTANCE)
                                                         (SETQ NEAREST-DISTANCE OBJECT-DISTANCE
                                                               NEAREST-OBJECTS `(,OBJECT)))
                                                        ((= OBJECT-DISTANCE NEAREST-DISTANCE)
                                                         (PUSH OBJECT NEAREST-OBJECTS))))
                                             FINALLY (RETURN NEAREST-OBJECTS)))
                       (INTERACTION-PANE (GET-INTERACTION-PANE SELF)))
                   ;; Identify the object being copied.
                   (IF (NULL OBJECT-TO-COPY)
                       ;; There aren't any object.  Nice try.
                       (SEND INTERACTION-PANE :LINE-OUT "There are no objects to copy.")
                       ;;1ELSE*
                       (PROGN
                         (SETQ OBJECT-TO-COPY (IF (> (LENGTH OBJECT-TO-COPY) 1)
                                                  (W:MENU-CHOOSE
                                                    (LOOP FOR OBJECT IN OBJECT-TO-COPY
                                                          COLLECT `(,(REMOVE-RETURNS
                                                                       (SEND OBJECT :NAME)
                                                                       #\SPACE)
                                                                    :VALUE ,OBJECT)))
                                                  (CAR OBJECT-TO-COPY)))
                         (WHEN OBJECT-TO-COPY
                           (SEND INTERACTION-PANE :STRING-OUT "Copying ")
                           (SEND INTERACTION-PANE :LINE-OUT (SEND OBJECT-TO-COPY :NAME))
                           (SETQ NEW-OBJECT (SEND OBJECT-TO-COPY :REPLICATE))
                           (SEND NEW-OBJECT :DRAW SELF W:ALU-XOR)
                           (SETQ CURRENT-MOUSE-BLINKER     W:MOUSE-GLYPH-DOUBLE-UP-ARROW-LETTERED-D
                                 CURRENT-MOUSE-DOCUMENTATION COPY-NEW-OBJECT-MOUSE-DOCUMENTATION)
                           (IF (FUNCALL (SYMBOL-FUNCTION (SEND NEW-OBJECT :PICK-CORNER-FUNCTION))
                                        NEW-OBJECT)
                               (PROGN
                                 (PUSH NEW-OBJECT DISPLAY-LIST)
                                 (RECALCULATE-BOUNDS))
                               ;;1ELSE*
                               ;; Make sure that the old object doesn't get zapped from the screen.
                               (SEND OBJECT-TO-COPY :DRAW SELF W:ALU-XOR))))))
                 (SEND SELF :MOUSE-STANDARD-BLINKER))))
  (SETQ CURRENT-MOUSE-BLINKER       NIL
        CURRENT-MOUSE-DOCUMENTATION NIL))

(DEFCONSTANT MOVE-NEW-OBJECT-MOUSE-DOCUMENTATION
             '(:MOUSE-L-1 "Move the object to this position."
               :MOUSE-M-1 "Abort out of this operation"))
             

(DEFCONSTANT MOVE-OBJECT-MOUSE-DOCUMENTATION
             '(:MOUSE-L-1 "Pick an object to move"
               :MOUSE-M-1 "Abort out of this operation"))

(DEFUN MOVE-OBJECT ()
  "Handles all interaction for moving objects."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LOOP WITH X = NIL
            WITH Y = NIL
            DO (PROGN
                 (WITH-GRID-DISABLED
                   (MULTIPLE-VALUE-SETQ (X Y)
                     (MOUSE-SPECIFY-POINT NIL W:MOUSE-GLYPH-THICK-UP-ARROW-LETTERED-D
                                          MOVE-OBJECT-MOUSE-DOCUMENTATION NIL)))
                 (WHEN (NULL X)
                   (RETURN NIL))
                 (SETQ X (TRANSLATE-PIXEL-TO-UNIT X)
                       Y (TRANSLATE-PIXEL-TO-UNIT Y))
                 (LET ((OBJECT-TO-MOVE (LOOP FOR OBJECT IN DISPLAY-LIST
                                             WITH OBJECT-DISTANCE = NIL
                                             WITH NEAREST-OBJECTS = NIL
                                             WITH NEAREST-DISTANCE = MOST-POSITIVE-SHORT-FLOAT
                                             WHEN (NOT (EQ OBJECT MARGIN-OBJECT))
                                             DO (PROGN
                                                  (SETQ OBJECT-DISTANCE (SEND OBJECT
                                                                              :DISTANCE-FROM-POINT
                                                                              SELF X Y))
                                                  (COND ((< OBJECT-DISTANCE NEAREST-DISTANCE)
                                                         (SETQ NEAREST-DISTANCE OBJECT-DISTANCE
                                                               NEAREST-OBJECTS `(,OBJECT)))
                                                        ((= OBJECT-DISTANCE NEAREST-DISTANCE)
                                                         (PUSH OBJECT NEAREST-OBJECTS))))
                                             FINALLY (RETURN NEAREST-OBJECTS)))
                       (INTERACTION-PANE (GET-INTERACTION-PANE SELF)))
                   ;; Identify the object being copied.
                   (IF (NULL OBJECT-TO-MOVE)
                       ;; There aren't any object.  Nice try.
                       (SEND INTERACTION-PANE :LINE-OUT "There are no objects to move.")
                       ;;1ELSE*
                       (PROGN
                         (SETQ OBJECT-TO-MOVE (IF (> (LENGTH OBJECT-TO-MOVE) 1)
                                                  (W:MENU-CHOOSE
                                                    (LOOP FOR OBJECT IN OBJECT-TO-MOVE
                                                          COLLECT `(,(REMOVE-RETURNS
                                                                       (SEND OBJECT :NAME)
                                                                       #\SPACE)
                                                                    :VALUE ,OBJECT)))
                                                  (CAR OBJECT-TO-MOVE)))
                         (WHEN OBJECT-TO-MOVE
                           (SEND INTERACTION-PANE :STRING-OUT "Moving ")
                           (SEND INTERACTION-PANE :LINE-OUT (SEND OBJECT-TO-MOVE :NAME))
                           (SETQ CURRENT-MOUSE-BLINKER W:MOUSE-GLYPH-THICK-UP-ARROW-LETTERED-D
                                 CURRENT-MOUSE-DOCUMENTATION MOVE-NEW-OBJECT-MOUSE-DOCUMENTATION)
                           (FUNCALL (SYMBOL-FUNCTION (SEND OBJECT-TO-MOVE :PICK-CORNER-FUNCTION))
                                    OBJECT-TO-MOVE)
                           (RECALCULATE-BOUNDS)))))
                 (SEND SELF :MOUSE-STANDARD-BLINKER))))
  (SETQ CURRENT-MOUSE-BLINKER       NIL
        CURRENT-MOUSE-DOCUMENTATION NIL))

