;;;-*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 10/11/88    TWE	Added extra argument to :WRITE-LATEX. 
;;;  6/15/88    TWE	Fixed drawing to copy the object to the picture window when the
;;;			user has specified enough parameters (instead of XORing the
;;;			object and interfering with other objects).  This keeps the user
;;;			from having to refresh the screen so often.
;;; 10/15/87    TWE	Fixed the error checking to allow for a width with a zero value
;;;			for the height, and flag that as an error.
;;; 10/07/87    TWE	Fixed to work with the edit operation.
;;;  9/29/87    TWE	Fixed corners to handle negative dimensions.
;;;  9/11/87    TWE	Fixed replicate to include the orientation of the corner.
;;;  8/31/87    TWE	Fixed :DISTANCE-FROM-POINT to more correctly compute the distance
;;;			using the limits of the object, instead of assuming that the
;;;			description of the object gives the limits.  Fixed the drawing
;;;			primitive to calculate the radius properly.  Previously, it used
;;;			MAX-HOLLOW-CIRCLE-DIAMETER as a radius.  Now it divides it by 2.
;;;  8/07/87    TWE	Fixed up the default position for a corner object to be a lower
;;;			case `l' instead of an upper case `L'.  LaTeX only understands
;;;			the lower case letter.
;;;  7/20/87    TWE	Update the :NAME method to reflect the shorter interaction pane.
;;;  7/17/87    TWE	Fixed up so that the user can see the coordinate being specified
;;;			when a single point is specified.
;;;  6/29/87    TWE	Fixed up the argument VARIABLE to be CVV-VARIABLE to get around
;;;			a compiler warning message.  Also added the support for
;;;			MINIMUM-X/Y to make the picture only as big as the user makes it.
;;;  6/22/87    TWE	Fixed it up so that it would work.
;;;  6/21/86	TWE	Initial creation.


#|

This file defines code to handle rounded corners.

|#

(DEFFLAVOR ROUNDED-CORNER-OBJECT
  ()
  (POSITION-MIXIN OVAL-OBJECT)
  :INITTABLE-INSTANCE-VARIABLES
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES
  (:DEFAULT-INIT-PLIST :POSITION "l"))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :PICK-CORNER-FUNCTION) ()
  'PICK-ROUNDED-CORNER-POINT)

(DEFMETHOD (ROUNDED-CORNER-OBJECT :TRANSLATE-UNITS) (NEW-OVER-OLD X-ORIGIN Y-ORIGIN)
  (SETQ WIDTH      (* NEW-OVER-OLD WIDTH)
	HEIGHT     (* NEW-OVER-OLD HEIGHT)
        X-POSITION (+ (* NEW-OVER-OLD (- X-POSITION X-ORIGIN)) X-ORIGIN)
        Y-POSITION (+ (* NEW-OVER-OLD (- Y-POSITION Y-ORIGIN)) Y-ORIGIN)))

(DEFCONSTANT ROUNDED-CORNER-MOUSE-DOCUMENTATION
             '(:MOUSE-L-1 "Pick the bottom left corner of the rounded corner"
               :MOUSE-M-1 "Abort out of this operation"))

(DEFUN PICK-ROUNDED-CORNER-POINT (OBJECT)
  "Pick an (X,Y) point for an rounded corner via the mouse."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF (AND (SEND OBJECT :X-POSITION)
	   (SEND OBJECT :WIDTH))
      ;; The user has already specified the position.  All
      ;; that is being done now is to change the position.
      (MOVE-ROUNDED-CORNER OBJECT)
      ;;1ELSE*
      (MULTIPLE-VALUE-BIND (NEW-X NEW-Y)
            (MOUSE-SPECIFY-POINT OBJECT W:MOUSE-GLYPH-3-12-ARC-POINTER
                                 ROUNDED-CORNER-MOUSE-DOCUMENTATION)
	    (IF NEW-X
		(PROGN
		 (MULTIPLE-VALUE-BIND (X-UNIT Y-UNIT)
		       (SEND SELF :GRIDIFY-MEASURE
			     (TRANSLATE-PIXEL-TO-UNIT NEW-X)
			     (TRANSLATE-PIXEL-TO-UNIT NEW-Y))
		    (SETF (SEND OBJECT :X-POSITION) X-UNIT)
		    (SETF (SEND OBJECT :Y-POSITION) Y-UNIT))
		 ;; Only draw the rounded corner when the size has been specified.
		 (WHEN (SEND OBJECT :WIDTH)
		    (SEND OBJECT :DRAW SELF))
		 T)
		;;1ELSE*
		NIL))))

(DEFUN MOVE-ROUNDED-CORNER (OBJECT)
  "Use the mouse to move an rounded corner to a new location."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (COND ((NULL (SEND OBJECT :X-POSITION))
         (PICTURE-ERROR "A corner position is not specified yet.")
         T)
        ((NULL (SEND OBJECT :WIDTH))
         (PICTURE-ERROR "A size is not specified yet.")
         T)
        (T
         (LET ((OLD-X-POSITION (SEND OBJECT :X-POSITION))
               (OLD-Y-POSITION (SEND OBJECT :Y-POSITION))
               USER-INPUT)
           (SEND SELF :TURN-ON-RUBBERBAND
		 #'(LAMBDA (X Y)
		     (MULTIPLE-VALUE-BIND (X-UNIT Y-UNIT)
                         (SEND SELF :GRIDIFY-MEASURE
			       (TRANSLATE-PIXEL-TO-UNIT X)
			       (TRANSLATE-PIXEL-TO-UNIT Y))
			 (SETF (SEND OBJECT :X-POSITION) X-UNIT)
			 (SETF (SEND OBJECT :Y-POSITION) Y-UNIT))
		     (SEND SELF :WRITE-OBJECT-STATUS OBJECT)
		     (DRAW-ROUNDED-CORNER-OBJECT OBJECT W:ALU-XOR))
                 (TRANSLATE-X-UNIT-TO-PIXEL OLD-X-POSITION)
                 (TRANSLATE-Y-UNIT-TO-PIXEL OLD-Y-POSITION))
           (SETQ USER-INPUT (READ-MOUSE-XY-INPUT))
           ;; Turn off rubberbanding.
           (SEND SELF :TURN-OFF-RUBBERBAND)
           (IF USER-INPUT
               (PROGN
                 ;; Draw the object with a IOR alu after we have finished moving it.
                 (SEND OBJECT :DRAW SELF)
                 T)
               ;;ELSE
               (PROGN
                 ;; Set the rounded corner back the way it was.
                 (SEND OBJECT :DRAW SELF W:ALU-XOR)
                 (SETF (SEND OBJECT :X-POSITION) OLD-X-POSITION)
                 (SETF (SEND OBJECT :Y-POSITION) OLD-Y-POSITION)
                 (SEND OBJECT :DRAW SELF W:ALU-XOR)
                 NIL))))))

(DEFUN RUBBER-BAND-ROUNDED-CORNER (OBJECT)
  "Use the mouse to specify the size of an rounded corner."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF (NULL (SEND OBJECT :X-POSITION))
      (PICTURE-ERROR "A corner position is not specified yet.")
      ;;ELSE
      (LET* ((OLD-WIDTH    (SEND OBJECT :WIDTH))
	     (OLD-HEIGHT   (SEND OBJECT :HEIGHT))
             (X-POSITION   (SEND OBJECT :X-POSITION))
             (Y-POSITION   (SEND OBJECT :Y-POSITION))
             (OLD-X-PIXEL-POSITION (TRANSLATE-X-UNIT-TO-PIXEL X-POSITION))
             (OLD-Y-PIXEL-POSITION (TRANSLATE-Y-UNIT-TO-PIXEL Y-POSITION))
             USER-INPUT)
        (SEND SELF :TURN-ON-RUBBERBAND
	      #'(LAMBDA (X Y)
                  (MULTIPLE-VALUE-BIND (NEW-WIDTH NEW-HEIGHT)
                      (SEND SELF :GRIDIFY-MEASURE
			    (- (TRANSLATE-PIXEL-TO-UNIT X) X-POSITION)
			    (- Y-POSITION (TRANSLATE-PIXEL-TO-UNIT Y)))
                    (SETF (SEND OBJECT :WIDTH ) NEW-WIDTH)
                    (SETF (SEND OBJECT :HEIGHT) NEW-HEIGHT)
                    (SEND SELF :WRITE-OBJECT-STATUS OBJECT)
                    (DRAW-ROUNDED-CORNER-OBJECT OBJECT W:ALU-XOR)))
              (IF OLD-WIDTH
                  (+ (TRANSLATE-UNIT-TO-PIXEL OLD-WIDTH ) OLD-X-PIXEL-POSITION)
                  ;;1ELSE*
                  NIL)
              (IF OLD-HEIGHT
                  (+ (TRANSLATE-UNIT-TO-PIXEL OLD-HEIGHT) OLD-Y-PIXEL-POSITION)
                  ;;1ELSE*
                  NIL))
        (SETQ USER-INPUT (READ-MOUSE-XY-INPUT))
        ;; Turn off rubberbanding.
        (SEND SELF :TURN-OFF-RUBBERBAND)
        (IF USER-INPUT
            (PROGN
              ;; Draw the object with a IOR alu after we have finished moving it.
              (SEND OBJECT :DRAW SELF)
              T)
            ;;ELSE
            (PROGN
              ;; Erase the object from the screen.
              (SEND OBJECT :DRAW SELF W:ALU-XOR)
              (WHEN OLD-WIDTH
                ;; Set the object back the way it was.
                (SETF (SEND OBJECT :WIDTH ) OLD-WIDTH)
                (SETF (SEND OBJECT :HEIGHT) OLD-HEIGHT)
                (SEND OBJECT :DRAW SELF))
              NIL)))))

(DEFUN GET-ROUNDED-CORNER-SIZE-DATA (OBJECT)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  "Get the exact dimensions of the rounded corner from the user."
  (LET ((OBJECT-WIDTH  (IF (ZEROP (SEND OBJECT :WIDTH )) 2.0 (SEND OBJECT :WIDTH )))
	(OBJECT-HEIGHT (IF (ZEROP (SEND OBJECT :HEIGHT)) 2.0 (SEND OBJECT :HEIGHT)))
        (PICTURE-BASE-UNIT BASE-UNIT))
    (FLET ((VALIDATE-WIDTH (WINDOW CVV-VARIABLE OLD-VALUE NEW-VALUE)
                           (DECLARE (IGNORE WINDOW CVV-VARIABLE OLD-VALUE))
                           (VALIDATE-PAGE-WIDTH NEW-VALUE PICTURE-BASE-UNIT))
	   (VALIDATE-HEIGHT (WINDOW CVV-VARIABLE OLD-VALUE NEW-VALUE)
                            (DECLARE (IGNORE WINDOW CVV-VARIABLE OLD-VALUE))
                            (VALIDATE-PAGE-HEIGHT NEW-VALUE PICTURE-BASE-UNIT)))
      (CONDITION-CASE ()
          (W:CHOOSE-VARIABLE-VALUES
           `((,(LOCF OBJECT-WIDTH)  "Width"  :CONSTRAINT ,#'VALIDATE-WIDTH
              :DOCUMENTATION "Width of the rounded corner.")
             (,(LOCF OBJECT-HEIGHT) "Height" :CONSTRAINT ,#'VALIDATE-HEIGHT
              :DOCUMENTATION "Height of the rounded corner."))
           :MARGIN-CHOICES '("Do It"
                             ("Abort" (SIGNAL-CONDITION EH:ABORT-OBJECT)))
           :LABEL (FORMAT NIL "Enter size for an rounded corner in ~a"
                          (GET 'UNIT-NAME-STRING PICTURE-BASE-UNIT)))
        (EH:ABORT (SETQ OBJECT-WIDTH NIL))))
    (IF OBJECT-WIDTH
        (PROGN
          (WHEN (SEND OBJECT :X-POSITION)
            ;; Erase the rounded corner only if it is there.
            (SEND OBJECT :DRAW SELF W:ALU-XOR))
          (SETF (SEND OBJECT :WIDTH ) OBJECT-WIDTH)
          (SETF (SEND OBJECT :HEIGHT) OBJECT-HEIGHT)
          ;; If the corner position hasn't been specified then don't draw the rounded corner.
          (WHEN (SEND OBJECT :X-POSITION)
            (SEND OBJECT :DRAW SELF))
          T)
      ;;ELSE
      NIL)))

;;; Primitive which just draws an rounded corner on a window.  Actually this isn't
;;; so simple since there are so many possibilities to take into account.
(DEFUN SIMPLE-DRAW-ROUNDED-CORNER (WINDOW X Y WIDTH HEIGHT POSITION
                                   &OPTIONAL (ALU W:ALU-IOR) (THICKNESS 1))
  (WHEN (AND (NOT (ZEROP WIDTH)) (NOT (ZEROP HEIGHT)))
    ;; Handle negative oval dimensions.
    (WHEN (MINUSP WIDTH)
      (SETQ WIDTH (- WIDTH))
      (DECF X WIDTH))
    (WHEN (MINUSP HEIGHT)
      (SETQ HEIGHT (- HEIGHT))
      (DECF Y HEIGHT))
    (LET ((RADIUS (MIN (TRUNCATE WIDTH 2) (TRUNCATE HEIGHT 2)
                       (TRUNCATE MAX-HOLLOW-CIRCLE-DIAMETER 2))))
      (SETQ THICKNESS (TRUNCATE THICKNESS 2))
      ;; Need to calculate the end points first.  The following oval shows
      ;; what each of the things being computed in the following LET
      ;; represents.  Also shown are the compass points used when drawing
      ;; the oval's corners.
      ;;
      ;;  NW                 NE
      ;;      
      ;;          X1.5
      ;;            |
      ;;        X1  | X2   OUTER-X
      ;;        v   v  v   v
      ;;       .---------.
      ;;      '           `
      ;;     '             `
      ;;    |               |<Y1
      ;;    |               |
      ;;    |               |<---Y1.5
      ;;    |               |
      ;;    |               |<Y2
      ;;    `.             .'
      ;;      `._________.'  < OUTER-Y
      ;;
      ;;  SW                 SE
      (LET ((X1 (+ X RADIUS))
            (Y1 (+ Y RADIUS))
            (X1.5 (+ X (TRUNCATE WIDTH  2)))
            (Y1.5 (+ Y (TRUNCATE HEIGHT 2)))
            (X2 (- (+ X WIDTH ) RADIUS))
            (Y2 (- (+ Y HEIGHT) RADIUS))
            (OUTER-X  (+ X WIDTH))
            (OUTER-Y  (+ Y HEIGHT)))
      (COND ((OR (NULL POSITION)
                 (STRING-EQUAL POSITION "")
                 (STRING-EQUAL POSITION "L"))
             (SEND WINDOW :DRAW-LINE X1      Y       X1.5    Y       THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X1 Y1 RADIUS :NW ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X       Y1      X       Y2      THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X1 Y2 RADIUS :SW ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X1      OUTER-Y X1.5    OUTER-Y THICKNESS W:BLACK ALU))
            ((STRING-EQUAL POSITION "R")
             (SEND WINDOW :DRAW-LINE X1.5    Y       X2      Y       THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X2 Y2 RADIUS :SE ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE OUTER-X Y1      OUTER-X Y2      THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X2 Y1 RADIUS :NE ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X1.5    OUTER-Y X2      OUTER-Y THICKNESS W:BLACK ALU))
            ((STRING-EQUAL POSITION "T")
             (SEND WINDOW :DRAW-LINE OUTER-X Y1      OUTER-X Y1.5    THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X2 Y1 RADIUS :NE ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X1      Y       X2      Y       THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X1 Y1 RADIUS :NW ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X       Y1      X       Y1.5    THICKNESS W:BLACK ALU))
            ((STRING-EQUAL POSITION "B")
             (SEND WINDOW :DRAW-LINE OUTER-X Y1.5    OUTER-X Y2      THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X1 Y2 RADIUS :SW ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X1      OUTER-Y X2      OUTER-Y THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X2 Y2 RADIUS :SE ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X       Y1.5    X       Y2      THICKNESS W:BLACK ALU))
            ((MEMBER POSITION '("LT" "TL") :TEST #'STRING-EQUAL)
             (SEND WINDOW :DRAW-LINE X1      Y       X1.5    Y       THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X1 Y1 RADIUS :NW ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X       Y1      X       Y1.5    THICKNESS W:BLACK ALU))
            ((MEMBER POSITION '("RT" "TR") :TEST #'STRING-EQUAL)
             (SEND WINDOW :DRAW-LINE X1.5    Y       X2      Y       THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X2 Y1 RADIUS :NE ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE OUTER-X Y1      OUTER-X Y1.5    THICKNESS W:BLACK ALU))
            ((MEMBER POSITION '("LB" "BL") :TEST #'STRING-EQUAL)
             (SEND WINDOW :DRAW-LINE X  Y1.5         X       Y2      THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X1 Y2 RADIUS :SW ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X1   OUTER-Y    X1.5    OUTER-Y THICKNESS W:BLACK ALU))
            ((MEMBER POSITION '("RB" "BR") :TEST #'STRING-EQUAL)
             (SEND WINDOW :DRAW-LINE OUTER-X Y1.5    OUTER-X Y2      THICKNESS W:BLACK ALU)
             (SIMPLE-DRAW-CORNER WINDOW X2 Y2 RADIUS :SE ALU         THICKNESS)
             (SEND WINDOW :DRAW-LINE X1.5 OUTER-Y    X2      OUTER-Y THICKNESS W:BLACK ALU)))))))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :WRITE-LATEX) (LATEX-STREAM TEXIFY)
  "2Write the description of this object to a LaTeX file.*"
  (DECLARE (IGNORE TEXIFY))
  (FORMAT LATEX-STREAM "\\put(~F,~F){\\oval(~F,~F)[~A]}~%"
          ;; For LaTeX, the (X,Y) position is in the center of the oval.
          (+ X-POSITION (/ WIDTH  2.0))
          (- Y-POSITION (/ HEIGHT 2.0)) WIDTH HEIGHT POSITION))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :WRITE-PICTURE) (PICTURE-STREAM)
  "2Write the description of this object to an auxilary file.*"
  ;; We write a string which, when read using the READ function, will create this object.
  (FORMAT PICTURE-STREAM "(MAKE-INSTANCE 'PICTURE:ROUNDED-CORNER-OBJECT ~
                                         :X-POSITION ~D :Y-POSITION ~D ~
                                         :WIDTH ~D :HEIGHT ~D ~
                                         :POSITION ~S)~%"
          X-POSITION Y-POSITION WIDTH HEIGHT POSITION))

(DEFUN DRAW-ROUNDED-CORNER-OBJECT (OBJECT &OPTIONAL (ALU W:CHAR-ALUF))
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((PIXEL-WIDTH  (TRANSLATE-UNIT-TO-PIXEL   (SEND OBJECT :WIDTH)))
	(PIXEL-HEIGHT (TRANSLATE-UNIT-TO-PIXEL   (SEND OBJECT :HEIGHT)))
        (PIXEL-X      (TRANSLATE-X-UNIT-TO-PIXEL (SEND OBJECT :X-POSITION)))
        (PIXEL-Y      (TRANSLATE-Y-UNIT-TO-PIXEL (SEND OBJECT :Y-POSITION)))
        (POSITION     (SEND OBJECT :POSITION)))
    ;; Note that we do not translate the line-thickness since this is not in terms of
    ;; a specific dimension; it can only be thick or thin.
    (SIMPLE-DRAW-ROUNDED-CORNER SELF PIXEL-X PIXEL-Y PIXEL-WIDTH PIXEL-HEIGHT
                                POSITION ALU LINE-THICKNESS)))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :DRAW) (WINDOW &OPTIONAL (ALU NIL))
  (IF ALU
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-ROUNDED-CORNER-OBJECT SELF ALU))
      ;;1ELSE*
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-ROUNDED-CORNER-OBJECT SELF))))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :DISTANCE-FROM-POINT) (WINDOW X Y)
  "Returns the smallest distance that this object is from the point (X,Y).
X and Y are measured in terms of the current unit.
If the point (X,Y) is inside of the object then 0.0 is returned."
  ;; This code incorrectly assumes that a rounded corner is a box.  For now,
  ;; this is OK, but someday it needs to be fixed.  The limits of the box are
  ;; the object limits of the rounded corner.
  (MULTIPLE-VALUE-BIND (LEFT TOP RIGHT BOTTOM)
      (SEND SELF :OBJECT-LIMITS WINDOW)
    (DISTANCE-FROM-BOX X Y LEFT TOP RIGHT BOTTOM)))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :OBJECT-LIMITS) (WINDOW)
  "Returns the minimum and maximum limits of this object.
The values returned are the left top right and bottom unit values."
  (DECLARE (IGNORE WINDOW))
  (LET* ((OBJECT-WIDTH  WIDTH)
         (OBJECT-HEIGHT HEIGHT)
         (HALF-WIDTH  (/ WIDTH  2.0))
         (HALF-HEIGHT (/ HEIGHT 2.0))
         (OBJECT-X-POSITION X-POSITION)
         (OBJECT-Y-POSITION Y-POSITION))
    ;; Adjust the size and X/Y-POSITION to reflect the shape of the corner.
    (WHEN (MEMBER POSITION '("l" "r" "lt" "rt" "lb" "rb" "tl" "tr" "bl" "br") :TEST #'STRING-EQUAL)
      (SETQ OBJECT-WIDTH HALF-WIDTH))
    (WHEN (MEMBER POSITION '("t" "b" "lt" "rt" "lb" "rb" "tl" "tr" "bl" "br") :TEST #'STRING-EQUAL)
      (SETQ OBJECT-HEIGHT HALF-HEIGHT))
    (WHEN (MEMBER POSITION '("r" "rt" "rb" "tr" "br") :TEST #'STRING-EQUAL)
      (INCF OBJECT-X-POSITION HALF-WIDTH))
    (WHEN (MEMBER POSITION '("b" "lb" "rb" "bl" "br") :TEST #'STRING-EQUAL)
      (DECF OBJECT-Y-POSITION HALF-HEIGHT))
    ;; Note that the (X,Y) position of the box is the upper left corner.
    ;; This makes the maximum Y value of the box the Y-POSITION.
    (VALUES OBJECT-X-POSITION
            OBJECT-Y-POSITION
            (+ OBJECT-X-POSITION OBJECT-WIDTH)
            (- OBJECT-Y-POSITION OBJECT-HEIGHT))))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :NAME) ()
  "Return a string which identifies the object."
  (FORMAT NIL "Rounded Corner at (~D,~D)~%with width ~D and height ~D"
          X-POSITION Y-POSITION WIDTH HEIGHT))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :SHORT-NAME) ()
  "Return a narrow string which identifies the object."
  (FORMAT NIL "Rounded Corner ~
              ~:[~2*~;at ~%(~,2F,~,2F)~]~
              ~:[~*~;~%Width=~D~]~
              ~:[~*~;~%Height=~D~]"
          X-POSITION X-POSITION Y-POSITION (NOT (ZEROP WIDTH)) WIDTH
          (NOT (ZEROP HEIGHT)) HEIGHT))

(DEFMETHOD (ROUNDED-CORNER-OBJECT :EDIT-FUNCTION) ()
  "Returns the symbol which is the function to do editing for this object."
  'DRAW-ROUNDED-CORNER)

(DEFMETHOD (ROUNDED-CORNER-OBJECT :REPLICATE) ()
  "2Make a copy of this object.*"
  (MAKE-INSTANCE 'ROUNDED-CORNER-OBJECT
                 :X-POSITION X-POSITION
                 :Y-POSITION Y-POSITION
                 :WIDTH WIDTH :HEIGHT HEIGHT
                 :POSITION POSITION))

(DEFUN GET-CORNER-POSITION-DATA (OBJECT)
  "2Obtain position data from the user and redisplay the text, if any exists.*"
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((NEW-POSITION (W:MENU-CHOOSE 
                        ;; Get rid of the centered item as it doesn't make sense for corners.
                        (REMOVE '("CENTERED") VALID-POSITION-ITEM-LIST
                                :TEST #'(LAMBDA (X Y) (STRING-EQUAL (CAR X) (CAR Y))))
                        :LABEL "Specify Position of oval")))
    ;; If the user aborts out of GET-POSITION-DATA then the position will be nil.
    (WHEN NEW-POSITION
      ;; Erase the oval before we display it at the new position.
      (SEND OBJECT :DRAW SELF W:ALU-XOR)
      (SETF (SEND OBJECT :POSITION) NEW-POSITION)
      (SEND OBJECT :DRAW SELF))))


(DEFCONSTANT ROUNDED-CORNER-ACTION-ITEM-LIST
             '(("Pick a corner" :VALUE PICK-ROUNDED-CORNER-POINT
                :DOCUMENTATION "Pick the bottom left corner of the rounded corner.")
               ("Corner Position" :VALUE GET-CORNER-POSITION-DATA
                :DOCUMENTATION "Specify how the part of the oval to show.")
               ("Rounded Corner size" :BUTTONS ((NIL :VALUE RUBBER-BAND-ROUNDED-CORNER)
                                     ()
                                     (NIL :VALUE GET-ROUNDED-CORNER-SIZE-DATA))
                :DOCUMENTATION (:MOUSE-L-1 "Specify rounded corner size using the mouse"
                                :MOUSE-R-1 "Enter the rounded corner size using the keyboard"))
               ("Abort" :VALUE :ABORT
                :DOCUMENTATION "Quit without specifying an rounded corner.")
               ("Done" :VALUE :DONE
                :DOCUMENTATION "Click here when you are done specifying an rounded corner."))
  "Menu item-list through which a user specifies an rounded corner.")

(DEFCONSTANT ROUNDED-CORNER-ITEM-SEQUENCING
             (GENERATE-ACTION-ITEM-SEQUENCE ROUNDED-CORNER-ACTION-ITEM-LIST
                                            '((NIL 1) (1 2) (2 3) (3 5))))

(DEFUN DRAW-ROUNDED-CORNER (&OPTIONAL (OLD-OBJECT NIL))
  "Handles all interaction for drawing an rounded corner."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((OBJECT NIL))
    (IF OLD-OBJECT
        (SETQ OBJECT OLD-OBJECT)
        ;;1ELSE*
        (SETQ OBJECT (MAKE-INSTANCE 'ROUNDED-CORNER-OBJECT)))
    (IF (GET-OBJECT-DATA OBJECT ROUNDED-CORNER-ACTION-ITEM-LIST "Specify an rounded corner"
                         ROUNDED-CORNER-ITEM-SEQUENCING)
        (IF (OR (ZEROP (SEND OBJECT :WIDTH))
                (ZEROP (SEND OBJECT :HEIGHT))
                (NULL  (SEND OBJECT :X-POSITION))
                (NULL  (SEND OBJECT :Y-POSITION)))
            ;; The OBJECT needs to have a size and a position.
            (PROGN
              (PICTURE-ERROR "The ~a component of the corner must be specified."
                             (COND ((ZEROP (SEND OBJECT :WIDTH  )) "width")
				   ((ZEROP (SEND OBJECT :HEIGHT )) "height")
                                   ((NULL  (SEND OBJECT :X-POSITION)) "corner"))))
            ;;1ELSE*
            (PROGN
              (LET ((OBJECT-HEIGHT (SEND OBJECT :HEIGHT)))
                (WHEN (MINUSP OBJECT-HEIGHT)
                  (SETF (SEND OBJECT :HEIGHT) (- OBJECT-HEIGHT))
                  (DECF (SEND OBJECT :Y-POSITION) OBJECT-HEIGHT)))
              (LET ((OBJECT-WIDTH (SEND OBJECT :WIDTH)))
                (WHEN (MINUSP OBJECT-WIDTH)
                  (SETF (SEND OBJECT :WIDTH) (- OBJECT-WIDTH))
                  (INCF (SEND OBJECT :X-POSITION) OBJECT-WIDTH)))
              (MULTIPLE-VALUE-BIND (LEFT TOP RIGHT BOTTOM)
                  (SEND OBJECT :OBJECT-LIMITS SELF)
                (SETQ MAXIMUM-X-UNIT (MAX MAXIMUM-X-UNIT RIGHT )
                      MINIMUM-X-UNIT (MIN MINIMUM-X-UNIT LEFT  )
                      MAXIMUM-Y-UNIT (MAX MAXIMUM-Y-UNIT TOP   )
                      MINIMUM-Y-UNIT (MIN MINIMUM-Y-UNIT BOTTOM)))
	      (WHEN (NULL OLD-OBJECT)
		(PUSH OBJECT DISPLAY-LIST))))
        ;;1ELSE*
        ;; See if enough of the object's parameters were entered for it to be displayed.
        (WHEN (AND (NOT (ZEROP (SEND OBJECT :WIDTH)))
                   (SEND OBJECT :X-POSITION)
                   (SEND OBJECT :Y-POSITION))
          ;; Don't XOR the object if the user aborted out of the menu
          ;; for an existing object.
          (WHEN (NULL OLD-OBJECT)
            (SEND OBJECT :DRAW SELF W:ALU-XOR))))))
