;;; -*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 11/03/88    TWE	Encapsulated much of the hypermedia for delete-object as a
;;;			separate function.
;;; 10/27/88    TWE	Cleaned up support for hypermedia links during deletion.
;;; 10/19/88    TWE	Added support for hypermedia links during deletion.
;;; 10/14/88    TWE	Put support in place for read-only mode.
;;;  4/20/88    TWE	Fixed up verification prompt to use an improved MOUSE-Y-OR-N-P.
;;;  7/17/87    TWE	Fixed up delete-object to also do what was done on 7/14.
;;;  7/14/87    TWE	Fixed up delete-all-objects to set up the min/maximum-x/y-units.
;;;  5/21/87	TWE	Initial creation.

#|

This file defines the commands that perform deletion.

|#

(DEFCONSTANT DELETE-OBJECT-MOUSE-DOCUMENTATION
             '(:MOUSE-L-1 "Pick an object to delete"
               :MOUSE-M-1 "Abort out of this operation"))


(DEFUN DELETE-OBJECT ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LOOP WITH X = NIL
            WITH Y = NIL
            DO (PROGN
                 (WITH-GRID-DISABLED
                   (MULTIPLE-VALUE-SETQ (X Y)
                     (MOUSE-SPECIFY-POINT NIL W:MOUSE-GLYPH-SCISSOR
                                          DELETE-OBJECT-MOUSE-DOCUMENTATION
                                          NIL)))
                 (WHEN (NULL X)
                   (RETURN NIL))
                 (SETQ X (TRANSLATE-PIXEL-TO-UNIT X)
                       Y (TRANSLATE-PIXEL-TO-UNIT Y))
                 (LET ((OBJECT-TO-DELETE (LOOP
                                           FOR OBJECT IN DISPLAY-LIST
                                           WITH OBJECT-DISTANCE = NIL
                                           WITH NEAREST-OBJECT = NIL
                                           WITH NEAREST-DISTANCE = MOST-POSITIVE-SHORT-FLOAT
                                           WHEN (NOT (EQ OBJECT MARGIN-OBJECT))
                                           DO (PROGN
                                                (SETQ OBJECT-DISTANCE (SEND OBJECT
                                                                            :DISTANCE-FROM-POINT
                                                                            SELF X Y))
                                                (WHEN (< OBJECT-DISTANCE NEAREST-DISTANCE)
                                                  (SETQ NEAREST-DISTANCE OBJECT-DISTANCE
                                                        NEAREST-OBJECT   OBJECT)
                                                  (WHEN (ZEROP OBJECT-DISTANCE)
                                                    (RETURN OBJECT))))
                                           FINALLY (RETURN NEAREST-OBJECT)))
                       (INTERACTION-PANE (GET-INTERACTION-PANE SELF)))
                   ;; Identify the object being deleted.
                   (IF (NULL OBJECT-TO-DELETE)
                       ;; There aren't any object.  Nice try.
                       (SEND INTERACTION-PANE :LINE-OUT "There are no objects to delete.")
                       ;;1ELSE*
                       (SEND INTERACTION-PANE :STRING-OUT "Deleting ")
                       (SEND INTERACTION-PANE :LINE-OUT (SEND OBJECT-TO-DELETE :NAME))
                       (WHEN (MOUSE-Y-OR-N-P (FORMAT NIL "Delete ~A?"
                                                     (SEND OBJECT-TO-DELETE :NAME)))
                         ;; Hack for hypermedia.  We have links which are offsets from the
                         ;; end of the display list.  If the object we are trying to delete
                         ;; affects the positioning of a link then we need to update its offset.
                         (WHEN (FBOUNDP 'USER:UPDATE-LATEX-OBJECT-POSITION)
                           (FUNCALL 'USER:UPDATE-LATEX-OBJECT-POSITION DISPLAY-LIST
                                    OBJECT-TO-DELETE))
                         (SEND OBJECT-TO-DELETE :DRAW SELF W:ERASE-ALUF)
                         (SETQ DISPLAY-LIST (REMOVE OBJECT-TO-DELETE DISPLAY-LIST :TEST #'EQ))
                         (RECALCULATE-BOUNDS))))))))

(DEFUN DELETE-ALL-OBJECTS ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (WHEN (MOUSE-Y-OR-N-P "Delete all objects?")
        (IF MARGIN-OBJECT
            (SETQ DISPLAY-LIST (LIST MARGIN-OBJECT))
            ;;1ELSE*
            (SETQ DISPLAY-LIST NIL))
        (SEND SELF :SET-MAXIMUM-X-UNIT 0.0)
        (SEND SELF :SET-MAXIMUM-Y-UNIT 0.0)
        (SEND SELF :SET-MINIMUM-X-UNIT MOST-POSITIVE-SHORT-FLOAT)
        (SEND SELF :SET-MINIMUM-Y-UNIT MOST-POSITIVE-SHORT-FLOAT)
        (SEND SELF :REFRESH))))

