;;;-*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 10/21/88    TWE	Changed the empty box to not shade the entire box but to only
;;;			shade the borders.
;;; 10/19/88    TWE	Fixed up SIMPLE-SHADED-BOX to pass the shade in as an argument.
;;;			This allows this function to be used elsewhere where a shade
;;;			different from that needed here is specified.
;;; 10/11/88	TWE	Initial creation.  Copied code from BOX.


#|

This file defines code to handle boxes with no borders.

|#


(DEFFLAVOR EMPTY-BOX-OBJECT
  ()
  (BASIC-BOX-OBJECT)
  :INITTABLE-INSTANCE-VARIABLES
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES)

(DEFPARAMETER EMPTY-SHADE W:25%-GRAY)

(DEFPARAMETER SHADED-BOX-OUTLINE-THICKNESS 4
  "Thickness in pixels of the empty box.")

;;; The following is a low-level function which draws the outline of the empty box.
(DEFUN DRAW-SHADED-BOX-OUTLINE (WINDOW LEFT TOP RIGHT BOTTOM &OPTIONAL (ALU W:ALU-IOR))
  ;; Special care is taken to make sure that the box can be drawn.
  
  ;; Also, the coordinates specified are the pixels of the exterior part of the box.
  ;; Large thicknesses must be scaled again to make them look correct on the screen.
  ;; The problem here is that thin lines are drawn with only one pixel of thickness, but
  ;; their comparative thickness on the page is much smaller.  Thick lines are also much
  ;; thicker on the screen than they appear on paper.  What we are doing here is to
  ;; scale lines thicker than `thick lines' to make them appear the same on the screen
  ;; as they appear on paper.
  (LET ((THICKNESS SHADED-BOX-OUTLINE-THICKNESS))
    (WHEN (AND (NOT (= LEFT RIGHT)) (NOT (= TOP BOTTOM)))
      ;; Handle the cases where one makes one of the edges out of order.
      (WHEN (> TOP BOTTOM)
        (PSETQ TOP    BOTTOM
               BOTTOM TOP))
      (WHEN (> LEFT RIGHT)
        (PSETQ LEFT  RIGHT
               RIGHT LEFT))
      (WHEN (PLUSP (- BOTTOM TOP THICKNESS THICKNESS))
        (LET ((INSIDE-LEFT   (W:SHEET-INSIDE-LEFT   WINDOW))
              (INSIDE-TOP    (W:SHEET-INSIDE-TOP    WINDOW))
              (INSIDE-RIGHT  (W:SHEET-INSIDE-RIGHT  WINDOW))
              (INSIDE-BOTTOM (W:SHEET-INSIDE-BOTTOM WINDOW))
              MINIMUM-SIZE)
          ;; Adjust to the edges of the window.
          (SETQ LEFT   (+ LEFT   INSIDE-LEFT)
                TOP    (+ TOP    INSIDE-TOP)
                RIGHT  (+ RIGHT  INSIDE-LEFT)
                BOTTOM (+ BOTTOM INSIDE-TOP))
          ;; Clip to the inside edges of the window.
          (WHEN (< LEFT   INSIDE-LEFT  ) (SETQ LEFT   INSIDE-LEFT))
          (WHEN (< TOP    INSIDE-TOP   ) (SETQ TOP    INSIDE-TOP))
          (WHEN (> RIGHT  INSIDE-RIGHT ) (SETQ RIGHT  INSIDE-RIGHT))
          (WHEN (> BOTTOM INSIDE-BOTTOM) (SETQ BOTTOM INSIDE-BOTTOM))
          (WHEN (> LEFT   INSIDE-RIGHT ) (SETQ LEFT   INSIDE-RIGHT))
          (WHEN (> TOP    INSIDE-BOTTOM) (SETQ TOP    INSIDE-BOTTOM))
          (WHEN (< (SETQ MINIMUM-SIZE (MIN (- RIGHT LEFT) (- BOTTOM TOP)))
                   (* 2 THICKNESS))
            (SETQ THICKNESS (TRUNCATE MINIMUM-SIZE 2)))
          (IF (= (* 2 THICKNESS) MINIMUM-SIZE)
              ;; Simply draw a shaded rectangle.
              (SIMPLE-SHADED-BOX WINDOW LEFT TOP (- RIGHT LEFT) (- BOTTOM TOP) ALU EMPTY-SHADE)
              ;;ELSE
              ;; AAAAAAAAAAAAAAAAAAAAAAAAAAA
              ;; C                         D
              ;; C                         D
              ;; C                         D
              ;; C                         D
              ;; C                         D
              ;; C                         D
              ;; C                         D
              ;; BBBBBBBBBBBBBBBBBBBBBBBBBBB
              ;; 
              ;; Part A
              (SIMPLE-SHADED-BOX WINDOW LEFT TOP (+ (- RIGHT LEFT) 1) THICKNESS ALU EMPTY-SHADE)
              ;; Part B
              (SIMPLE-SHADED-BOX WINDOW LEFT (+ (- BOTTOM THICKNESS) 1)
                                 (+ (- RIGHT LEFT) 1) THICKNESS ALU EMPTY-SHADE)
              ;; Part C
              (SIMPLE-SHADED-BOX WINDOW LEFT (+ TOP THICKNESS)
                                 THICKNESS (+ (- BOTTOM TOP THICKNESS THICKNESS) 1)
                                 ALU EMPTY-SHADE)
              ;; Part D
              (SIMPLE-SHADED-BOX WINDOW (- RIGHT (1- THICKNESS)) (+ TOP THICKNESS)
                                 THICKNESS (+ (- BOTTOM TOP THICKNESS THICKNESS) 1)
                                 ALU EMPTY-SHADE)))))))

;;; This is a helper function which draws a box filled in with the indicated shade.
(DEFUN SIMPLE-SHADED-BOX (WINDOW X Y WIDTH HEIGHT &OPTIONAL (ALU W:ALU-IOR) SHADE)
  (W:PREPARE-SHEET (WINDOW)
    (SYS:%DRAW-SHADED-TRIANGLE X Y (+ X WIDTH -1) Y (+ X WIDTH -1) (+ Y HEIGHT -1) ALU
                               T T T SHADE WINDOW)
    (SYS:%DRAW-SHADED-TRIANGLE X Y X (+ Y HEIGHT -1) (+ X WIDTH -1) (+ Y HEIGHT -1) ALU
                               T T NIL SHADE WINDOW)))

(DEFUN SIMPLE-DRAW-EMPTY-BOX (WINDOW LEFT TOP RIGHT BOTTOM &OPTIONAL (ALU W:ALU-IOR))
  (WHEN (AND (NOT (= LEFT RIGHT)) (NOT (= TOP BOTTOM)))
    ;; Handle the cases where one makes one of the edges out of order.
    (WHEN (> TOP BOTTOM)
      (PSETQ TOP    BOTTOM
             BOTTOM TOP))
    (WHEN (> LEFT RIGHT)
      (PSETQ LEFT  RIGHT
             RIGHT LEFT))
    (LET ((INSIDE-LEFT   (W:SHEET-INSIDE-LEFT   WINDOW))
          (INSIDE-TOP    (W:SHEET-INSIDE-TOP    WINDOW))
          (INSIDE-RIGHT  (W:SHEET-INSIDE-RIGHT  WINDOW))
          (INSIDE-BOTTOM (W:SHEET-INSIDE-BOTTOM WINDOW)))
      ;; Adjust to the edges of the window.
      (SETQ LEFT   (+ LEFT   INSIDE-LEFT)
            TOP    (+ TOP    INSIDE-TOP)
            RIGHT  (+ RIGHT  INSIDE-LEFT)
            BOTTOM (+ BOTTOM INSIDE-TOP))
      ;; Clip to the inside edges of the window.
      (WHEN (< LEFT   INSIDE-LEFT  ) (SETQ LEFT   INSIDE-LEFT))
      (WHEN (< TOP    INSIDE-TOP   ) (SETQ TOP    INSIDE-TOP))
      (WHEN (> RIGHT  INSIDE-RIGHT ) (SETQ RIGHT  INSIDE-RIGHT))
      (WHEN (> BOTTOM INSIDE-BOTTOM) (SETQ BOTTOM INSIDE-BOTTOM))
      (WHEN (> LEFT   INSIDE-RIGHT ) (SETQ LEFT   INSIDE-RIGHT))
      (WHEN (> TOP    INSIDE-BOTTOM) (SETQ TOP    INSIDE-BOTTOM))
      (W:PREPARE-SHEET (WINDOW)
        (DRAW-SHADED-BOX-OUTLINE WINDOW LEFT TOP RIGHT BOTTOM ALU)))))


(DEFMETHOD (EMPTY-BOX-OBJECT :WRITE-LATEX) (LATEX-STREAM TEXIFY)
  "2Write the description of this object to a LaTeX file.*"
  (FORMAT LATEX-STREAM "\\put(~F,~F){\\makebox(~F,~F)~:[~*~;[~A]~]~
                        ~2:*{\\shortstack~:[~*~;[~A]~]{~A}}}"
          ;; The X and Y positions point to the upper left corner of the box.
          ;; For LaTeX, these point to the lower left corner of the box.
          X-POSITION (- Y-POSITION HEIGHT) WIDTH HEIGHT
          (NOT (ZEROP (LENGTH POSITION))) POSITION
          (TEXIFY TEXT TEXT-FONT TEXIFY))
  (TERPRI LATEX-STREAM))


(DEFMETHOD (EMPTY-BOX-OBJECT :WRITE-PICTURE) (PICTURE-STREAM)
  "2Write the description of this object to an auxilary file.*"
  ;; We write a string which, when read using the READ function, will create this object.
  (FORMAT PICTURE-STREAM "(MAKE-INSTANCE 'PICTURE:EMPTY-BOX-OBJECT ~
                                         :X-POSITION ~F :Y-POSITION ~F ~
                                         :WIDTH ~F :HEIGHT ~F ~
                                         :POSITION ~S :TEXT ~S :TEXT-FONT ~S)~%"
          X-POSITION Y-POSITION WIDTH HEIGHT POSITION TEXT TEXT-FONT))


(DEFMETHOD (EMPTY-BOX-OBJECT :TRANSLATE-UNITS) (NEW-OVER-OLD X-ORIGIN Y-ORIGIN)
  (SETQ WIDTH      (* NEW-OVER-OLD WIDTH)
        HEIGHT     (* NEW-OVER-OLD HEIGHT)
        X-POSITION (+ (* NEW-OVER-OLD (- X-POSITION X-ORIGIN)) X-ORIGIN)
        Y-POSITION (+ (* NEW-OVER-OLD (- Y-POSITION Y-ORIGIN)) Y-ORIGIN)))

(DEFUN DRAW-EMPTY-BOX-OBJECT (OBJECT &OPTIONAL (ALU W:CHAR-ALUF))
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((PIXEL-WIDTH          (TRANSLATE-UNIT-TO-PIXEL   (SEND OBJECT :WIDTH)))
        (PIXEL-HEIGHT         (TRANSLATE-UNIT-TO-PIXEL   (SEND OBJECT :HEIGHT)))
        (PIXEL-X              (TRANSLATE-X-UNIT-TO-PIXEL (SEND OBJECT :X-POSITION)))
        (PIXEL-Y              (TRANSLATE-Y-UNIT-TO-PIXEL (SEND OBJECT :Y-POSITION))))
    ;; Only try to draw a box when we have something to draw.
    (WHEN PIXEL-X
      (SIMPLE-DRAW-EMPTY-BOX SELF PIXEL-X PIXEL-Y
                             (+ PIXEL-X PIXEL-WIDTH) (+ PIXEL-Y PIXEL-HEIGHT) ALU)
      ;; Be sure to draw the text if any is specified.
      (WHEN (NOT (ZEROP (LENGTH (SEND OBJECT :TEXT))))
        (DISPLAY-TEXT OBJECT ALU)))))

(DEFMETHOD (EMPTY-BOX-OBJECT :DRAW) (WINDOW &OPTIONAL (ALU NIL))
  (IF ALU
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-EMPTY-BOX-OBJECT SELF ALU))
      ;;1ELSE*
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-EMPTY-BOX-OBJECT SELF))))

(DEFMETHOD (EMPTY-BOX-OBJECT :NAME) ()
  "Return a string which identifies the object."
  (FORMAT NIL "Non-bordered box at (~D,~D) that is ~D by ~D~
               ~:[~;,~%text is \"~A\", font is ~A~]."
          X-POSITION Y-POSITION WIDTH HEIGHT
          (PLUSP (LENGTH TEXT)) (REMOVE-RETURNS TEXT) (DESCRIPTIVE-FONT-NAME TEXT-FONT)))

(DEFMETHOD (EMPTY-BOX-OBJECT :SHORT-NAME) ()
  "Return a narrow string which identifies the object."
  (FORMAT NIL "Non-bordered box ~
              ~:[~2*~;at ~%(~,2F,~,2F)~]~
              ~:[~*~;~%Width=~D~]~
              ~:[~*~;~%Height=~D~]"
          X-POSITION X-POSITION Y-POSITION (NOT (ZEROP WIDTH)) WIDTH (NOT (ZEROP HEIGHT)) HEIGHT))

(DEFMETHOD (EMPTY-BOX-OBJECT :EDIT-FUNCTION) ()
  "Returns the symbol which is the function to do editing for this object."
  'DRAW-EMPTY-BOX)

(DEFMETHOD (EMPTY-BOX-OBJECT :REPLICATE) ()
  "2Make a copy of this object.*"
  (MAKE-INSTANCE 'EMPTY-BOX-OBJECT
                 :X-POSITION X-POSITION
                 :Y-POSITION Y-POSITION
                 :WIDTH WIDTH :HEIGHT HEIGHT
                 :POSITION POSITION :TEXT TEXT
                 :TEXT-FONT TEXT-FONT))

(DEFUN RUBBER-BAND-EMPTY-BOX (OBJECT)
  "Use the mouse to specify the size of a box."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF (NULL (SEND OBJECT :X-POSITION))
      (PICTURE-ERROR "A corner position is not specified yet.")
      ;;ELSE
      (LET* ((OLD-WIDTH  (SEND OBJECT :WIDTH))
             (OLD-HEIGHT (SEND OBJECT :HEIGHT))
             (X-POSITION (SEND OBJECT :X-POSITION))
             (Y-POSITION (SEND OBJECT :Y-POSITION))
             (OLD-X-PIXEL-POSITION (TRANSLATE-X-UNIT-TO-PIXEL X-POSITION))
             (OLD-Y-PIXEL-POSITION (TRANSLATE-Y-UNIT-TO-PIXEL Y-POSITION))
             USER-INPUT)
        (SEND SELF :TURN-ON-RUBBERBAND
              #'(LAMBDA (X Y)
                  (MULTIPLE-VALUE-BIND (NEW-WIDTH NEW-HEIGHT)
                      (SEND SELF :GRIDIFY-MEASURE
                            (- (TRANSLATE-PIXEL-TO-UNIT X) X-POSITION)
                            (- Y-POSITION (TRANSLATE-PIXEL-TO-UNIT Y)))
                    (SETF (SEND OBJECT :WIDTH ) NEW-WIDTH)
                    (SETF (SEND OBJECT :HEIGHT) NEW-HEIGHT)
                    (SEND SELF :WRITE-OBJECT-STATUS OBJECT)
                    (DRAW-EMPTY-BOX-OBJECT OBJECT W:ALU-XOR)))
              (IF OLD-WIDTH
                  (+ (TRANSLATE-UNIT-TO-PIXEL OLD-WIDTH) OLD-X-PIXEL-POSITION)
                  ;;1ELSE*
                  NIL)
              (IF OLD-HEIGHT
                  (+ (TRANSLATE-UNIT-TO-PIXEL OLD-HEIGHT) OLD-Y-PIXEL-POSITION)
                  ;;1ELSE*
                  NIL))
        (SETQ USER-INPUT (READ-MOUSE-XY-INPUT))
        ;; Turn off rubberbanding.
        (SEND SELF :TURN-OFF-RUBBERBAND)
        (IF (NULL USER-INPUT)
            (PROGN
              ;; Erase the object from the screen.
              (SEND OBJECT :DRAW SELF W:ALU-XOR)
              (WHEN OLD-WIDTH
                ;; Set the object back the way it was.
                (SETF (SEND OBJECT :WIDTH ) OLD-WIDTH)
                (SETF (SEND OBJECT :HEIGHT) OLD-HEIGHT)
                (SEND OBJECT :DRAW SELF))
              NIL)
            ;;ELSE
            T))))

(DEFUN ENTER-EMPTY-BOX-TEXT-LINES (OBJECT)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (ENTER-BASIC-BOX-TEXT-LINES OBJECT "Text Lines for an non-bordered Box"))

(DEFCONSTANT EMPTY-BOX-ACTION-ITEM-LIST
#| 1 |#      '(("Pick a corner" :VALUE PICK-BOX-POINT :DOCUMENTATION "Pick a corner of the box.")
#| 2 |#        ("Box size" :BUTTONS ((NIL :VALUE RUBBER-BAND-EMPTY-BOX)
                                     ()
                                     (NIL :VALUE GET-BOX-SIZE-DATA))
                :DOCUMENTATION (:MOUSE-L-1 "Specify box size using the mouse"
                                :MOUSE-R-1 "Enter the box dimensions using the keyboard"))
#| 3 |#        ("Lines of text" :VALUE ENTER-BOX-TEXT-LINES
                :DOCUMENTATION "Enter the lines of text to be placed inside the box.")
#| 4 |#        ("Text Position" :VALUE GET-BOX-POSITION-DATA
                :DOCUMENTATION "Specify how the text is to be positioned in the box.")
#| 5 |#        ("Text font" :VALUE GET-FONT-DATA
                :DOCUMENTATION "Specify a font for the text")
#| 6 |#        ("Abort" :VALUE :ABORT
                :DOCUMENTATION "Quit without specifying a box.")
#| 7 |#        ("Done" :VALUE :DONE
                :DOCUMENTATION "Click here when you are done specifying a box."))
  "Menu item-list through which a user specifies a non-bordered  box.")

(DEFCONSTANT EMPTY-BOX-ITEM-SEQUENCING
             (GENERATE-ACTION-ITEM-SEQUENCE EMPTY-BOX-ACTION-ITEM-LIST
                                            '((NIL 1) (1 2) (2 3) (3 4) (4 5) (5 7))))

(DEFUN DRAW-EMPTY-BOX (&OPTIONAL (OLD-OBJECT NIL))
  "Handles all interaction for drawing an empty box."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((OBJECT NIL))
    (IF OLD-OBJECT
        (SETQ OBJECT OLD-OBJECT)
        ;;1ELSE*
        (SETQ OBJECT (MAKE-INSTANCE 'EMPTY-BOX-OBJECT)))
    (IF (GET-OBJECT-DATA OBJECT EMPTY-BOX-ACTION-ITEM-LIST "Specify a non-bordered Box"
                         EMPTY-BOX-ITEM-SEQUENCING)
        (PROGN
          (IF (OR (ZEROP (SEND OBJECT :WIDTH))
                  (ZEROP (SEND OBJECT :HEIGHT))
                  (NULL  (SEND OBJECT :X-POSITION))
                  (NULL  (SEND OBJECT :Y-POSITION))
                  (ZEROP (LENGTH (SEND OBJECT :TEXT))))
              ;; The box needs to have a size and a position.
              (PICTURE-ERROR "The ~a component of the box must be specified."
                             (COND ((ZEROP (SEND OBJECT :WIDTH     )) "width")
				   ((ZEROP (SEND OBJECT :HEIGHT    )) "height")
                                   ((NULL  (SEND OBJECT :X-POSITION)) "initial corner")
                                   ((ZEROP (LENGTH (SEND OBJECT :TEXT))) "text")))
              ;;ELSE
              (PROGN
                (LET ((OBJECT-HEIGHT (SEND OBJECT :HEIGHT)))
                  (WHEN (MINUSP OBJECT-HEIGHT)
                    (SETF (SEND OBJECT :HEIGHT) (- OBJECT-HEIGHT))
                    (DECF (SEND OBJECT :Y-POSITION) OBJECT-HEIGHT)))
                (LET ((OBJECT-WIDTH (SEND OBJECT :WIDTH)))
                  (WHEN (MINUSP OBJECT-WIDTH)
                    (SETF (SEND OBJECT :WIDTH) (- OBJECT-WIDTH))
                    (INCF (SEND OBJECT :X-POSITION) OBJECT-WIDTH)))
                (MULTIPLE-VALUE-BIND (LEFT TOP RIGHT BOTTOM)
                    (SEND OBJECT :OBJECT-LIMITS SELF)
                  (SETQ MAXIMUM-X-UNIT (MAX MAXIMUM-X-UNIT RIGHT )
                        MINIMUM-X-UNIT (MIN MINIMUM-X-UNIT LEFT  )
                        MAXIMUM-Y-UNIT (MAX MAXIMUM-Y-UNIT TOP   )
                        MINIMUM-Y-UNIT (MIN MINIMUM-Y-UNIT BOTTOM)))
                (WHEN (NULL OLD-OBJECT)
                  (PUSH OBJECT DISPLAY-LIST)))))
        ;;1ELSE*
        ;; See if enough of the object's parameters were entered for it to be displayed.
        (WHEN (AND (NOT (ZEROP (SEND OBJECT :WIDTH)))
                   (NOT (ZEROP (SEND OBJECT :HEIGHT)))
                   (SEND OBJECT :X-POSITION)
                   (SEND OBJECT :Y-POSITION))
          ;; Don't XOR the object if the user aborted out of the menu
          ;; for an existing object.
          (WHEN (NULL OLD-OBJECT)
            (SEND OBJECT :DRAW SELF W:ALU-XOR))))))
