;;; -*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 11/03/88    TWE	Changed DISPLAY-HELP-INFORMATION to call COMMAND-CHARACTER-LIST
;;;			instead of accessing the list directly.
;;; 10/12/88    TWE	Updated the help to describe the new writing options, orientation
;;;			and scaling.
;;;  9/20/88    TWE	Changed the help window to an editor to allow the user to scan
;;;			the help text easier.
;;; 10/01/87    TWE	Updated for the scale menu item.
;;;  9/14/87    TWE	Updated for landscape mode.
;;;  5/21/87	TWE	Initial creation.

#|

This file defines the command that displays helpful information for the user.

We use an editor window instead of the standard typeout window to make it
easier on the user.  The typeout window approach is useful when there is one
screenful of information for the user to read.  Here, we have several
screenfuls so the typeout window approach is marginal at best.  Instead, we pop
up an editor window with the help string and make the buffer read-only.  This
allows the user to move backward as well as forward, and allows the user to
utilize other Zmacs commands like incremental search.

|#


(DEFUN POP-UP-HELP-STRING (STRING &KEY MODE-LINE-LIST)
  "Pop up an editor window with the help string in it."
  (USING-RESOURCE (WINDOW ZWEI:POP-UP-STANDALONE-EDITOR-FRAME)
    (UNWIND-PROTECT
        (PROGN
          (SEND WINDOW :SET-COMTAB ZWEI:*STANDALONE-COMTAB*)
          (SEND WINDOW :SET-SIZE-FROM-STRING STRING NIL NIL)
          (SEND WINDOW :SET-INTERVAL-STRING STRING)
          (SEND WINDOW :SET-MODE-LINE-LIST (OR MODE-LINE-LIST '("Standalone editor frame")))
          ;; Don't allow the user to alter the help string.
          (SEND (SYMEVAL-IN-CLOSURE (SEND WINDOW :EDITOR-CLOSURE) 'ZWEI:*INTERVAL*)
                :SET-READ-ONLY-P T)
          (W:EXPOSE-WINDOW-NEAR WINDOW '(:MOUSE))
          (SEND WINDOW :TYPEIN-LINE
                "This is a Zmacs window.  Use Zmacs commands to scan this help text.")
          (W:WINDOW-CALL (WINDOW :DEACTIVATE)
            (CATCH 'ZWEI:ABORT-STANDALONE-EDIT
              (SEND WINDOW :EDIT)
              (SEND WINDOW :INTERVAL-STRING))))
      ;; Set the resource object back to read/write.
      (SEND (SYMEVAL-IN-CLOSURE (SEND WINDOW :EDITOR-CLOSURE) 'ZWEI:*INTERVAL*)
            :SET-READ-ONLY-P NIL))))

(DEFUN DISPLAY-HELP-INFORMATION ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET* ((OLD-X W:MOUSE-X)
         (OLD-Y W:MOUSE-Y)
         (LABEL  (FORMAT NIL "  Press ~C when done.  (RO)"
                         #\END)))
    (POP-UP-HELP-STRING
      (CONCATENATE
        'SIMPLE-STRING
        (FORMAT NIL "                        LaTeX Picture program


CONTENTS
  INTRODUCTION
  WINDOW LAYOUT
  CREATING OBJECTS
  ALTERING OLD OBJECTS
  SAVING THE PICTURE
  GLOBAL ENVIRONMENT
  KEYBOARD

INTRODUCTION

This program was written to enable one to enter pictures into a LaTeX
document without having so specify each object manually.  The author
of this program has defined pictures for LaTeX by hand and discovered
that it is a very tedious process.  That discovery eventually led to
the implementation of the LaTeX Picture program, which allows one to
specify LaTeX picture objects in a WYSIWYG (What You See Is What You
Get) manner.  Great care has been taken to attempt to simulate on the
screen what one would see LaTeX generate.

Since its initial creation, this program has been enhanced to support
landscape mode.  This allows one to use the LaTeX Picture program to
generate foils via SliTeX which contain graphics.

WINDOW LAYOUT

The layout of the LaTeX Picture program consists of a set of 4 types
of windows: Picture Window, Interaction Window, Status Window and
menu window(s).  The Picture Window is where all of the graphics are
drawn which represent the graphics one would see on a page generated
by LaTeX.  When the LaTeX Picture program is first brought up one will
see tiny dots in this window.  These dots form a grid which allows one
to connect objects like lines and boxes in a regular fashion.  This
grid can be altered via the `Toggle Grid State' menu item (search
forward for details).  The Interaction Window is where all messages of
general interest appear.  The following messages appear in this
window: command identification which indicate which menu item was
clicked on, error messages, login prompts, filename prompts, and
object identification (when the user clicks the mouse near an object
in the Picture Window).  All keyboard input is entered in the
Interaction Window.  When the user uses the mouse to specify some
component of an object (for example the initial position of a line)
information is displayed in the Status Window which indicates the
numerical value of this component.  Normally this will only be useful
when one is concerned about exact placement of objects.


CREATING OBJECTS

To use the LaTeX Picture program, specify the type of object to be
drawn by clicking on menu items, and specify the other non-textual
information via the mouse.  After all objects have been specified,
click on the `Write TeX file' menu item, and then enter the filename
when prompted.  This file contains all of the information by LaTeX to
draw the picture.  It is recommended that one use the \\input command
from another file rather than physically placing the contents of this
file into another one.  A picture, once written, can be reloaded by
clicking on the `Read Picture file' menu item.  This actually reads in
a file which makes reconstruction of the picture easy; it does not
read in the .TEX file.  This means that if one edits the .TEX file
that those changes will not be recognized if one clicks on the 'Read
Picture file' menu item.

After clicking on a menu item to draw a specific type of object,
another menu will appear which shows the attributes of that object.
If the menu appears in a poor position, just move the mouse away and
it will reappear an a different location.  For all objects, one must
specify the initial position before specifying anything else.  After
specifying a value for an attribute, one may change it to something
else at any time.  For example, one can specify an initial position
for a hollow box, then specify the size of the box and then go back to
specify a different initial position.  The placement of the mouse
cursor within menu items it not done like it is in other applications.
This placement attempts to guess what item the user will position the
mouse on next.  The initial mouse cursor placement on all object
creation menus is on the `initial position' menu item.  Often one will
not need to move the mouse at all on an object creation menu.


ALTERING OLD OBJECTS

In addition to the object creation commands, there are commands to
manipulate existing objects - edit, copy and delete.  Each of these
commands is `sticky', in that once it is invoked, it will remain as
the current mode until another command is entered.  This allows, for
example, one to edit a collection of objects.  The `Scale' menu item
gives one another way to alter old objects.  See GLOBAL ENVIRONMENT
for more details.

SAVING THE PICTURE

The file operation commands are the interface between the picture on
the screen and the file system.  When one writes out a TeX file, two
files are actually written - a .TEX file and a .PIC file.  The .TEX
file is suitable for input directly into a LaTeX document - no changes
should be necessary.  The .PIC file is used by the LaTeX Picture
program to reconstruct the image that corresponds to the file.
Neither file should be edited directly unless one is certain that the
Picture program will not need to read them.

The LaTeX Picture program goes to great length to try and maintain the
WYSIWYG theme.  At times, however, it is useful to insert LaTeX
commands into text to output special characters.  For example, one may
have a mathematical formula that is to be inside a box.  Normally the
LaTeX Picture program will take that formula, say $x^{2}$, and output
characters in the .TEX file that will cause LaTeX to print out just
those characters.  For this example it will actually output
\\$x\\^{}$\\{$2$\\}$\\$ which is the way to tell LaTeX to print
$x^{2}$.  To print out the formula `x squared', instead of the text
$x^{2}$ the `Write TeX file' menu item is used, but click the right
mouse button (look at the mouse documentation window) instead of
clicking left.  Clicking right on this menu item will cause the LaTeX
Picture program to not translate special characters and just output
characters `as is'.

GLOBAL ENVIRONMENT

The next set of commands deal with altering the global environment of
a picture.

LaTeX supports two thicknesses for line segments: thick and thin.  The
`Change Line Thickness' menu item allows one to alter the current
setting.  One can also numerically specify thicknesses other than
those standard ones via this same menu item.

The LaTeX Picture program also supports all of the linear dimensions
that are supported by LaTeX, such as inches and millimeters.  The
`Change Units' menu item allows one to change the current unit of
measure.  After specifying a unit, all objects are automatically
scaled to reflect the new unit so that they do not change their
position or size.  Most users will not care anything about units or
where things are placed.  The 'Change Units' menu item allows those
other users to fine tune their picture.

The `Specify margins' menu item allows one to draw the margins which
LaTeX will be using to format your picture.  The author has found this
helpful since LaTeX will have trouble drawing pictures that are too
large horizontally to fit on a page.  Keep in mind that the margin
graphic is only an aid for the user.  It has no relation to the actual
page layout used by LaTeX.  It is also not displayed on the page; one
need not turn it off when saving the picture.

The `Toggle Grid State' menu item deals with the grid.  If you have
used the graphics editor (GED) you are probably already familiar with
grids.  A grid makes it easy for one to have the boundaries of objects
touch one another.  If the grid is turned on, the default, this will
happen automatically.  If the grid is turned off then one has greater
freedom (down to the pixel level) in placing objects on the screen.
One can also change the dimensions of the grid box in the X or Y
dimensions so it is more in tune with the objects being drawn.  All
objects will have at least the initial position be on a grid point.
Some objects, such as hollow boxes and dashed boxes will have all
edges be on grid points.  Other objects, such as lines and vectors,
will have the second endpoint be on a grid point if it is possible.

The `Orientation' menu item allows one to change from a portrait
orientation for paper to a landscape orientation for slides.  The
screen layout changes to reflect the new orientation of the Picture
Window.  If objects are already present in the Picture Window before
changing orientation they may not be visible when changing to the
other orientation.

The `Scale' menu item allows one to scale the entire picture by a
specified amount.  The user enters a number and all objects are scaled
by that amount.  For example, a scale factor of .5 would make all
objects half their previous size.  There are some objects which may
not be effected by this command.  Text in general does not scale.
Also, circles have upper bounds for their size which may not be
exceeded.  The thickness of line segments also does not change.
One can specify the origin for scaling depending upon which mouse
button one presses for this menu item.  In this way one can control
the placement of the scaled objects.

KEYBOARD

If you don't like using the mouse so much, you can use the keyboard
for some of the commands.  The following command characters are
defined:
~
    ")
        (LOOP WITH LINES = ""
              FOR ELEMENT IN (COMMAND-CHARACTER-LIST)
              DO (SETQ LINES (CONCATENATE
                               'SIMPLE-STRING
                               LINES
                               (FORMAT NIL "~%~30<~@:C~>  ~A" (FIRST ELEMENT) (THIRD ELEMENT))))
              FINALLY (RETURN LINES)))
      :MODE-LINE-LIST `(,LABEL))
    ;; Put the mouse back where it was.
    (SETQ W:MOUSE-X OLD-X
          W:MOUSE-Y OLD-Y)))
