;;;-*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 10/11/88    TWE	Added extra argument to :WRITE-LATEX. 
;;; 10/15/87    TWE	Fixed up the first menu item to say `pick the center' instead of
;;;			`pick a corner'.
;;; 10/07/87    TWE	Fixed to work with the edit operation.
;;;  7/20/87    TWE	Update the :NAME method to reflect the shorter interaction pane.
;;;  7/17/87    TWE	Fixed up draw-circle to handle the menu abort case better.
;;;  6/29/87    TWE	Added the support for MINIMUM-X/Y to make the picture only as big
;;;			as the user makes it.
;;;  6/19/86	TWE	Initial creation.


#|

This file defines code to handle hollow circles.

|#

(DEFFLAVOR HOLLOW-CIRCLE-OBJECT () (CIRCLE-OBJECT))

;;; Primitive which just draws a hollow circle on a window.
(DEFUN SIMPLE-DRAW-CIRCLE (WINDOW X Y DIAMETER &OPTIONAL (ALU W:ALU-IOR) (THICKNESS 1))
  (WHEN (> DIAMETER MAX-HOLLOW-CIRCLE-DIAMETER)
    (SETQ DIAMETER MAX-HOLLOW-CIRCLE-DIAMETER))
  (SEND WINDOW :DRAW-CIRCLE X Y (/ DIAMETER 2.0) THICKNESS W:BLACK ALU 50.))

(DEFMETHOD (HOLLOW-CIRCLE-OBJECT :WRITE-LATEX) (LATEX-STREAM TEXIFY)
  "2Write the description of this object to a LaTeX file.*"
  (DECLARE (IGNORE TEXIFY))
  (FORMAT LATEX-STREAM "\\put(~F,~F){\\circle{~F}}~%"
          X-POSITION Y-POSITION DIAMETER))

(DEFMETHOD (HOLLOW-CIRCLE-OBJECT :WRITE-PICTURE) (PICTURE-STREAM)
  "2Write the description of this object to an auxilary file.*"
  ;; We write a string which, when read using the READ function, will create this object.
  (FORMAT PICTURE-STREAM "(MAKE-INSTANCE 'PICTURE:HOLLOW-CIRCLE-OBJECT ~
                                         :X-POSITION ~D :Y-POSITION ~D ~
                                         :DIAMETER ~D)~%"
          X-POSITION Y-POSITION DIAMETER))

(DEFUN DRAW-HOLLOW-CIRCLE-OBJECT (OBJECT &OPTIONAL (ALU W:CHAR-ALUF))
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((PIXEL-DIAMETER       (TRANSLATE-UNIT-TO-PIXEL   (SEND OBJECT :DIAMETER)))
        (PIXEL-X              (TRANSLATE-X-UNIT-TO-PIXEL (SEND OBJECT :X-POSITION)))
        (PIXEL-Y              (TRANSLATE-Y-UNIT-TO-PIXEL (SEND OBJECT :Y-POSITION))))
    ;; Note that we do not translate the line-thickness since this is not in terms of
    ;; a specific dimension; it can only be thick or thin.
    (SIMPLE-DRAW-CIRCLE SELF PIXEL-X PIXEL-Y PIXEL-DIAMETER ALU LINE-THICKNESS)))

(DEFMETHOD (HOLLOW-CIRCLE-OBJECT :DRAW) (WINDOW &OPTIONAL (ALU NIL))
  (IF ALU
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-HOLLOW-CIRCLE-OBJECT SELF ALU))
      ;;1ELSE*
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-HOLLOW-CIRCLE-OBJECT SELF))))

(DEFMETHOD (HOLLOW-CIRCLE-OBJECT :NAME) ()
  "Return a string which identifies the object."
  (FORMAT NIL "Hollow circle with center at (~D,~D)~%with diameter ~D"
          X-POSITION Y-POSITION DIAMETER))

(DEFMETHOD (HOLLOW-CIRCLE-OBJECT :SHORT-NAME) ()
  "Return a narrow string which identifies the object."
  (FORMAT NIL "Hollow circle ~
              ~:[~2*~;at~%(~,2F,~,2F)~]~
              ~@[~%diameter=~D~]"
          X-POSITION X-POSITION Y-POSITION DIAMETER))

(DEFMETHOD (HOLLOW-CIRCLE-OBJECT :EDIT-FUNCTION) ()
  "Returns the symbol which is the function to do editing for this object."
  'DRAW-CIRCLE)

(DEFMETHOD (HOLLOW-CIRCLE-OBJECT :REPLICATE) ()
  "2Make a copy of this object.*"
  (MAKE-INSTANCE 'HOLLOW-CIRCLE-OBJECT
                 :X-POSITION X-POSITION
                 :Y-POSITION Y-POSITION
                 :DIAMETER DIAMETER))

(DEFCONSTANT HOLLOW-CIRCLE-ACTION-ITEM-LIST
             '(("Pick the center" :VALUE PICK-CIRCLE-POINT
                :DOCUMENTATION "Pick the center of a hollow circle.")
               ("Circle size" :BUTTONS ((NIL :VALUE RUBBER-BAND-CIRCLE)
                                     ()
                                     (NIL :VALUE GET-CIRCLE-SIZE-DATA))
                :DOCUMENTATION (:MOUSE-L-1 "Specify circle size using the mouse"
                                :MOUSE-R-1 "Enter the circle diameter using the keyboard"))
               ("Abort" :VALUE :ABORT
                :DOCUMENTATION "Quit without specifying a hollow circle.")
               ("Done" :VALUE :DONE
                :DOCUMENTATION "Click here when you are done specifying a hollow circle."))
  "Menu item-list through which a user specifies a hollow circle.")

(DEFCONSTANT HOLLOW-CIRCLE-ITEM-SEQUENCING
             (GENERATE-ACTION-ITEM-SEQUENCE HOLLOW-CIRCLE-ACTION-ITEM-LIST
                                            '((NIL 1) (1 2) (2 4))))

(DEFUN DRAW-CIRCLE (&OPTIONAL (OLD-OBJECT NIL))
  "Handles all interaction for drawing a hollow circle."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((OBJECT NIL))
    (IF OLD-OBJECT
        (SETQ OBJECT OLD-OBJECT)
        ;;1ELSE*
        (SETQ OBJECT (MAKE-INSTANCE 'HOLLOW-CIRCLE-OBJECT)))
    (IF (GET-OBJECT-DATA OBJECT HOLLOW-CIRCLE-ACTION-ITEM-LIST "Specify a hollow Circle"
                         HOLLOW-CIRCLE-ITEM-SEQUENCING)
        (IF (OR (ZEROP (SEND OBJECT :DIAMETER))
                (NULL  (SEND OBJECT :X-POSITION))
                (NULL  (SEND OBJECT :Y-POSITION)))
            ;; The OBJECT needs to have some width and a position.
            (PROGN
              (PICTURE-ERROR "The ~a component of the object must be specified."
                             (COND ((ZEROP (SEND OBJECT :DIAMETER  )) "diameter")
                                   ((NULL  (SEND OBJECT :X-POSITION)) "center"))))
            ;;ELSE
            (PROGN
              (MULTIPLE-VALUE-BIND (LEFT TOP RIGHT BOTTOM)
                  (SEND OBJECT :OBJECT-LIMITS SELF)
                (SETQ MAXIMUM-X-UNIT (MAX MAXIMUM-X-UNIT RIGHT )
                      MINIMUM-X-UNIT (MIN MINIMUM-X-UNIT LEFT  )
                      MAXIMUM-Y-UNIT (MAX MAXIMUM-Y-UNIT TOP   )
                      MINIMUM-Y-UNIT (MIN MINIMUM-Y-UNIT BOTTOM)))
	      (WHEN (NULL OLD-OBJECT)
		(PUSH OBJECT DISPLAY-LIST))))
        ;;1ELSE*
        ;; See if enough of the object's parameters were entered for it to be displayed.
        (WHEN (AND (SEND OBJECT :DIAMETER)
                   (NOT (ZEROP (SEND OBJECT :DIAMETER)))
                   (SEND OBJECT :X-POSITION)
                   (SEND OBJECT :Y-POSITION))
          ;; Don't XOR the object if the user aborted out of the menu
          ;; for an existing object.
          (WHEN (NULL OLD-OBJECT)
            (SEND OBJECT :DRAW SELF W:ALU-XOR))))))
