;;; -*- Mode:Zetalisp; Package:COMPILER; Fonts:(CPTFONT HL12B HL12BI); Base:10 -*-

;;; Copyright (C) 1984, 1987 Texas Instruments Incorporated. All rights reserved.

;;; UCL-window interface to the completion probes utility.
;;; See TOOLS;PROBES.TEXT for user documentation.
;;; Revision:
;;;  11/19/84 DNG - Original public version.
;;;  12/05/86 DNG - Modify PROBER function so that it selects the current
;;;		instantiation insteading of always making a new one.  Fix
;;;		report display to start at the top of the screen.  Update
;;;		the HELP screen to refer to the PROBES.TEXT file.
;;;   1/30/87 CLM - Modified for Explorer release 3.0.
;;;   5/21/87 DNG - Added (:METHOD PROBER-FRAME :BEFORE :KILL); use new release 
;;;		3 scroll bar; fix to start function definition display at the
;;;		top of the window.
;;;   3/02/89 DNG - Add binding of a SYSTEM key.
;;;   5/02/89 DNG - Updated pathname of PROBES file.

(eval-when ( load )
  (unless (fboundp 'probe-report-internal)
    (load "SYS:PUBLIC.PROBE;PROBES") ) )

(DEFCONST  *text-font* fonts:cptfont "The normal font for text.")
(DEFCONST  *headline-font* fonts:medfnt "The font for headings")
(DEFCONST  *light-font* fonts:cptfont "The font for executed code.")
(DEFCONST  *heavy-font* fonts:hl12bi "The font for emphasizing unexecuted code.")
(DEFCONST  *italic-font* fonts:hl12bi "Italics")
(DEFCONST  *mousable-font* fonts:cptfont "The font for mouse sensitive items.")

(DEFVAR *SAVED-DISPLAY* NIL)
(DEFVAR *DEFAULT-MOUSE-DOC* "  R2: System menu.")
(DEFVAR *MOUSE-DOC* *DEFAULT-MOUSE-DOC*)

#+IEEE-FLOATING-POINT ; Explorer release 3
(DEFFLAVOR prober-main-pane ()
           (TV:TOP-LABEL-MIXIN
	    TV:BORDERS-MIXIN
            TV:TEXT-SCROLL-WINDOW-EMPTY-GRAY-HACK   
            TV:FUNCTION-TEXT-SCROLL-WINDOW
            W:MOUSE-SENSITIVE-TEXT-SCROLL-WINDOW 
	    W:SCROLL-BAR-MIXIN
	    TV:WINDOW)
  :SETTABLE-INSTANCE-VARIABLES
  (:DOCUMENTATION "The main pane for Prober."))

#-IEEE-FLOATING-POINT ; Explorer release 1 or 2
(DEFFLAVOR prober-main-pane ()
           (TV:TOP-LABEL-MIXIN
	    TV:BORDERS-MIXIN
            TV:TEXT-SCROLL-WINDOW-EMPTY-GRAY-HACK   
            TV:FUNCTION-TEXT-SCROLL-WINDOW        
            TV:MOUSE-SENSITIVE-TEXT-SCROLL-WINDOW 
            TV:FLASHY-MARGIN-SCROLLING-MIXIN   
	    TV:BASIC-SCROLL-BAR
	    TV:WINDOW)
  :SETTABLE-INSTANCE-VARIABLES
  (:DOCUMENTATION "The main pane for Prober."))


(DEFMETHOD (prober-main-pane :WHO-LINE-DOCUMENTATION-STRING) ()
 *MOUSE-DOC* )

(DEFUN prober-print-function (item IGNORE window &optional ignore)
 1 "This function is called for each item in the scroll window."*
  (LET ((previous-font (SEND window :current-font)))
    (SELECTQ (CAR item)
      (:text  (SEND window :set-current-font *text-font*)
	      (FORMAT window "~a" (CADR item)))
      (:MIXED (IF (NULL (REST (SECOND ITEM))) ; one font used for whole line
		  (PROGN  
		    (UNLESS (NULL (SECOND ITEM))
		      (SEND window :set-current-font
			    (CDR (FIRST (SECOND ITEM)))) )
		    (FORMAT window "~a" (THIRD item)))
		;; else more than one font is used
		(LET (( FONTS (SECOND ITEM) ) ( LINE (THIRD ITEM) ))
		  (DOTIMES ( COLUMN (LENGTH LINE) )
		    (LOOP WHILE (AND (CONSP FONTS)
				     (<= (CAR (FIRST FONTS)) COLUMN) )
			  DO (PROGN
			       (SEND window :set-current-font (CDR (FIRST FONTS)))
			       (SETQ FONTS (REST FONTS)) ) )
		    (SEND WINDOW :TYO (AREF LINE COLUMN)) ) ) ) )
      (t  (SEND window :item1 item (CAR item) 'prober-print-m-s-item)))
    (SEND window :set-current-font previous-font)))


(DEFUN prober-print-m-s-item (item window &optional &rest ignore)
 1 "This function is called by the window system :item1 method to print a mouse sensitive item."*
  (LET ((previous-font (SEND window :current-font)))
    (SEND window :set-current-font *mousable-font*)
    (FORMAT window "~a" (CADR item))
    (SEND window :set-current-font previous-font)))


(DEFFLAVOR prober-menu-pane ()
	   (TV:DYNAMIC-ITEM-LIST-MIXIN
	    TV:MENU-HIGHLIGHTING-MIXIN
	    TV:COMMAND-MENU)
  (:DOCUMENTATION "The menu for Prober."))


(DEFFLAVOR prober-interaction-pane ()
	   (UCL:COMMAND-AND-LISP-TYPEIN-WINDOW
	    TV:PREEMPTABLE-READ-ANY-TYI-MIXIN
	    TV:NOTIFICATION-MIXIN
	    TV:AUTOEXPOSING-MORE-MIXIN
	    TV:AUTO-SCROLLING-MIXIN
	    TV:WINDOW)
  (:DOCUMENTATION :COMBINATION "The interaction window for command and lisp typein."))


(DEFMETHOD (prober-interaction-pane :WHO-LINE-DOCUMENTATION-STRING) ()
  "Type in a command name or Lisp form.  R2: System Menu.")


(DEFFLAVOR prober-frame
	   ((appl-frame nil)			;1Pointers to the various frames which are bound in the *
	    (main-pane nil)			;1 after init method and can be used by various functions.*
	    (menu-pane nil)
	    (interactor-pane nil)
	    (item-list-alist nil))		;1Use this to save the various item lists for the scroll window.*
						;1The format of this list should be.*
						;1 ( (:tag1  (item-list for this tag))*
						;1   (:tag2  (item-list for this tag))*
						;1 ... )*
						;1If the user wants to see something else in the scroll window,*
						;1first check here before rebuilding the list.*
	   (UCL:COMMAND-LOOP-MIXIN
	    TV:STREAM-MIXIN
	    TV:FRAME-DONT-SELECT-INFERIORS-WITH-MOUSE-MIXIN
	    TV:BORDERED-CONSTRAINT-FRAME-WITH-SHARED-IO-BUFFER
	    TV:LABEL-MIXIN)
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES
  :INITTABLE-INSTANCE-VARIABLES
  :SPECIAL-INSTANCE-VARIABLES
  (:DEFAULT-INIT-PLIST
    :active-command-tables '(prober-cmd-table)
    :all-command-tables '(prober-cmd-table)
    :menu-panes '((menu prober-menu-ucl))
    :save-bits :delayed
    :typein-handler :handle-typein-input
    :basic-help '(prober-doc)
    :prompt "Lisp> "
    :label nil)
  (:DOCUMENTATION :COMBINATION "The frame for Prober."))
			   

(DEFMETHOD (prober-frame :BEFORE :INIT) (IGNORE)
  1"Define the panes and constraints for this application."*
  (SETQ tv:PANES  `((main prober-main-pane
                      :LABEL (:string "Prober Report Display Pane"
				      :top :font fonts:hl12bi)
		      :FONT-MAP ,(DELETE-DUPLICATES
				   (LIST *text-font* *light-font* *heavy-font*
					 *headline-font* 
					 *italic-font* *mousable-font*)
				   :TEST #'EQ)
		      :MARGIN-SCROLL-REGIONS ((:TOP "Top" "More Above" FONTS:hl10b)
					      (:BOTTOM "Bottom"
						       "More Below" FONTS:hl10b))
		      :SAVE-BITS T
		      :PRINT-FUNCTION-ARG nil
		      :PRINT-FUNCTION prober-print-function
		      :SENSITIVE-ITEM-TYPES '(:text)
		      :MORE-P t
		      :BORDER-MARGIN-WIDTH 3)
	            (menu prober-menu-pane
                      :LABEL (:string "Prober commands" :font fonts:hl12bi)
                      :FONT-MAP ,(LIST fonts:cptfont))
		    (interactor prober-interaction-pane
		      :LABEL nil)))
  (SETQ tv:CONSTRAINTS `((main . ((main menu interactor)
			          ((menu 5 :lines))
			          ((interactor 8 :lines))
			          ((main :even)))))))


(DEFMETHOD (prober-frame :AFTER :INIT) (IGNORE)
  "1Set up the pane variables so that they can be used in the commands etc.*"
  (SETQ appl-frame self)
  (SETQ main-pane (SEND self :get-pane 'main))
  (SETQ menu-pane (SEND self :get-pane 'menu))
  (SETQ interactor-pane (SEND self :get-pane 'interactor))
  (SEND interactor-pane :select))


(DEFMETHOD (prober-frame :DESIGNATE-IO-STREAMS) ()
  1"Redefine a ucl method to bind io to the appropriate panes."*
  (SETQ terminal-io interactor-pane)
  (SETQ standard-output interactor-pane)
  (SETQ standard-input interactor-pane))
      

(DEFMETHOD (prober-frame :around :HANDLE-UNKNOWN-INPUT)
	   (cont mt ignore)
  "1Intercept this UCL method to handle the mouse sensitive item blips.*"
  (UNLESS (handle-mouse-sensitive-item-blip ucl:kbd-input)
    (FUNCALL-WITH-MAPPING-TABLE cont mt :handle-unknown-input)))

(DEFFLAVOR SAVED-WINDOW-STATE ( ITEM-LIST SCROLL-POS MOUSE-DOC
			       MOUSE-X-POSITION MOUSE-Y-POSITION )
	   ( ) )
(DEFMETHOD (SAVED-WINDOW-STATE :SAVE) ( WINDOW )
  (SETQ ITEM-LIST (SEND WINDOW :ITEMS))
  (SETQ MOUSE-DOC *MOUSE-DOC*)
  (SETQ SCROLL-POS (MULTIPLE-VALUE-LIST (SEND WINDOW :SCROLL-POSITION)))
  (SETQ MOUSE-X-POSITION TV:MOUSE-X)
  (SETQ MOUSE-Y-POSITION TV:MOUSE-Y) )
(DEFMETHOD (SAVED-WINDOW-STATE :RESTORE) ( WINDOW )
  (TV:DELAYING-SCREEN-MANAGEMENT
    (SEND WINDOW :SET-ITEMS ITEM-LIST)
    (SEND WINDOW :SCROLL-TO (FIRST SCROLL-POS) :ABSOLUTE)
    (SETQ *MOUSE-DOC* MOUSE-DOC)
    (SETQ TV:MOUSE-X MOUSE-X-POSITION)
    (SETQ TV:MOUSE-Y MOUSE-Y-POSITION) ) )
(COMPILE-FLAVOR-METHODS SAVED-WINDOW-STATE)

(DEFUN SAVE-DISPLAY ()
  (DECLARE (SPECIAL MAIN-PANE))
  (WHEN (NULL *SAVED-DISPLAY*)
    (SETQ *SAVED-DISPLAY* (MAKE-INSTANCE 'SAVED-WINDOW-STATE)))
  (SEND *SAVED-DISPLAY* :SAVE MAIN-PANE) )

(DEFUN handle-mouse-sensitive-item-blip (blip)
  1"If blip is a mouse sensitive item handle it and return t, else nil."*
  ;;  5/21/87 DNG - Fix to position first item at top of screen.
  (DECLARE (SPECIAL interactor-pane appl-frame main-pane))
  (COND ((ATOM blip) NIL)
	((AND (EQ (FIRST BLIP) ':MOUSE-BUTTON)
	      (= (SECOND BLIP) #\MOUSE-M-1)
	      (NOT (NULL *SAVED-DISPLAY*)) )
	 (SEND *SAVED-DISPLAY* :RESTORE MAIN-PANE)
	 T )
	((MEMBER (CAR blip) (SEND main-pane :sensitive-item-types))
	 (LET* (( STRING (SECOND (SECOND BLIP)))
		( NAME (READ-FROM-STRING STRING)) )
	   (COND ((= (FOURTH BLIP) #\MOUSE-L-1)
		  (LET ( REPORT-ITEM-LIST )
		    (SAVE-DISPLAY)
		    (LET (( *STANDARD-OUTPUT* (MAKE-INSTANCE 'SCROLL-OUTPUT-STREAM) ))
		      (SEND *STANDARD-OUTPUT* :SET-CURRENT-FONT *headline-font*)
		      (PROBE-REPORT-1 NAME T)
		      (SETQ REPORT-ITEM-LIST (SEND *STANDARD-OUTPUT* :GET-ITEM-LIST)) )
		    (SEND MAIN-PANE :SET-ITEMS REPORT-ITEM-LIST)
		    ;; position first item at top of screen
		    (SEND main-pane :put-item-in-window (SEND main-pane :item-of-number 0))
		    (SETQ *MOUSE-DOC*
                      " M: Return to summary listing, R2: System menu.")) )
		 ((= (FOURTH BLIP) #\MOUSE-M-1)
		  (ED NAME) )
		 )
	   t))
	) )


(DEFCOMMAND prober-doc-cmd ()
  '(:DESCRIPTION "Display specific helpful documentation about this application."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Help")
    :KEYS (#\meta-help))
  (prober-doc))


(DEFCOMMAND exit-cmd ()
  '(:DESCRIPTION "Exit this application."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Exit")
    :KEYS (#\end))
  (DECLARE (SPECIAL appl-frame))
  (SEND appl-frame :quit))

(DEFCOMMAND edit-cmd ()
  '(:DESCRIPTION "Go to the ZMACS editor process; use  to return."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Edit")
    :KEYS (#\super-E))
  (ed) )

(DEFCOMMAND top-cmd ()
  '(:DESCRIPTION "Display top page of scroll window."
    :NAMES ("Top")
    :KEYS (#\meta-<))
  (DECLARE (SPECIAL main-pane))
  (SEND main-pane :put-item-in-window (SEND main-pane :item-of-number 0)))


(DEFCOMMAND bottom-cmd ()
  '(:DESCRIPTION "Display bottom page of scroll window."
    :NAMES ("Bottom")
    :KEYS (#\meta->))
  (DECLARE (SPECIAL main-pane))
  (SEND main-pane :put-last-item-in-window))


(DEFCOMMAND page-up-cmd ()
  '(:DESCRIPTION "Display the next page up in the scroll window."
    :NAMES ("Page up")
    :KEYS (#\meta-v))
  (DECLARE (SPECIAL main-pane))
  (SEND main-pane :scroll-relative :top :bottom))


(DEFCOMMAND page-down-cmd ()
  '(:DESCRIPTION "Display the next page down in the scroll window."
    :NAMES ("Page down")
    :KEYS (#\control-v))
  (DECLARE (SPECIAL main-pane))
  (SEND main-pane :scroll-relative :bottom :top))


(DEFCOMMAND probe-functions-cmd ()
  '(:DESCRIPTION "Set up probes in some specified functions.  (Functions must have been compiled or evaled in Zmacs.)"
    :MENUS ((prober-menu-ucl))
    :NAMES ("Probe functions")
    :KEYS (#\super-f))
  (DECLARE (SPECIAL interactor-pane appl-frame))
  (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (FORMAT interactor-pane "~a" "; (COMPILER:PROBE-FUNCTION function-spec ...)")
  (SEND appl-frame :handle-prompt)
  (SEND interactor-pane :force-kbd-input "(compiler:probe-function "))


(DEFCOMMAND probing-on-cmd ()
  '(:DESCRIPTION "Cause insertion of probes in anything compiled after this."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Probing on")
    :KEYS ())
  (DECLARE (SPECIAL interactor-pane appl-frame))
  (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (SEND interactor-pane :force-kbd-input "(compiler:probing-on)"))

(DEFCOMMAND probing-off-cmd ()
  '(:DESCRIPTION "Don't insert probes in anything compiled after this."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Probing off")
    :KEYS ())
  (DECLARE (SPECIAL interactor-pane appl-frame))
  (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (SEND interactor-pane :force-kbd-input "(compiler:probing-off)"))

#|
(DEFCOMMAND probe-file-cmd ()
  '(:DESCRIPTION "Compile a Lisp file into memory with probes in all of the functions."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Probe file")
    :KEYS (#\super-l))
  (DECLARE (SPECIAL interactor-pane appl-frame))
  (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (FORMAT interactor-pane "~a" "; (COMPILER:PROBE-FUNCTION pathname ...)")
  (SEND appl-frame :handle-prompt)
  (SEND interactor-pane :force-kbd-input "(compiler:probe-function "))


(DEFCOMMAND save-probes-cmd ()
  '(:DESCRIPTION "Compiles a Lisp file into object with probes inserted."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Compile file")
    :KEYS (#\super-s))
  (DECLARE (SPECIAL interactor-pane appl-frame))
  (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (SEND interactor-pane :force-kbd-input "(compiler:compile-file-with-probes  "))

(DEFCOMMAND sample-report-cmd ()
  '(:DESCRIPTION "Display the report.  (with mouse sensitive items for displaying individual functions)"
    :MENUS ((prober-menu-ucl))
    :NAMES ("Report")
    :KEYS (#\super-r))
  (DECLARE (SPECIAL main-pane item-list-alist))
  (LET ((report-item-list (OR (get-item-list :report))))
    (UNLESS report-item-list
      (SEND main-pane :set-sensitive-item-types
	    (CONS :fct1 (CONS :fct2 (CONS :fct3 (SEND main-pane :sensitive-item-types)))))
      (SETQ report-item-list
	    '((:text "some text ")
	      (:fct1 "function 1")
	      (:fct2 "funciton 2")
	      (:fct3 "function 3")
	      (:text "more text")))
      (SETQ item-list-alist (CONS (CONS :report (LIST report-item-list)) item-list-alist)))
    (SEND main-pane :set-items report-item-list)))
|#

(DEFUN get-item-list (tag)
  1"This function looks through item-list-alist for the alist with tag and returns the corresponding item list."*
  (DECLARE (SPECIAL item-list-alist))
  (DOLIST (tag-and-item-list item-list-alist)
    (WHEN (EQ tag (CAR tag-and-item-list)) (RETURN (CADR tag-and-item-list)))))


(DEFCOMMAND report-summary-cmd ()
  '(:DESCRIPTION "Display summary report of probed functions."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Report summary")
    :KEYS (#\super-r))
  (DECLARE (SPECIAL main-pane interactor-pane appl-frame))
  (SEND main-pane :set-sensitive-item-types
	    (CONS :fname (SEND main-pane :sensitive-item-types)))
  (SAVE-DISPLAY)
 (LET ( REPORT-ITEM-LIST )
   (LET (( *STANDARD-OUTPUT* (MAKE-INSTANCE 'SCROLL-OUTPUT-STREAM) ))
     (SEND *STANDARD-OUTPUT* :SET-CURRENT-FONT *text-font*)
     (PROBE-REPORT-INTERNAL NIL)
     (SETQ REPORT-ITEM-LIST (SEND *STANDARD-OUTPUT* :GET-ITEM-LIST)) )
   (DOLIST ( ITEM REPORT-ITEM-LIST )
     (WHEN (EQ (FIRST ITEM) ':TEXT)
       (SETF (FIRST ITEM) ':FNAME)
       (SETQ *MOUSE-DOC*
 "L:Display definition highlighting unused code, M:Edit this function, R2:System menu."
     ) ) )
   (SEND MAIN-PANE :SET-ITEMS REPORT-ITEM-LIST)
   ;; position first item at top of screen
   (SEND main-pane :put-item-in-window (SEND main-pane :item-of-number 0)) )
 (VALUES) )

(DEFCOMMAND report-frequency-cmd ()
  '(:DESCRIPTION "List functions in order of number of times called."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Report frequency")
    :KEYS (#\super-c))
  (DECLARE (SPECIAL main-pane interactor-pane appl-frame))
  (SAVE-DISPLAY)
  (SEND main-pane :set-sensitive-item-types
	    (CONS :fname (SEND main-pane :sensitive-item-types)))
 (LET ( REPORT-ITEM-LIST )
   (LET (( *STANDARD-OUTPUT* (MAKE-INSTANCE 'SCROLL-OUTPUT-STREAM) ))
     (SEND *STANDARD-OUTPUT* :SET-CURRENT-FONT *text-font*)
     (COMPILER:PROBE-REPORT-FREQUENCY)
     (SETQ REPORT-ITEM-LIST (SEND *STANDARD-OUTPUT* :GET-ITEM-LIST)) )
   (DOLIST ( ITEM REPORT-ITEM-LIST )
     (WHEN (EQ (FIRST ITEM) ':TEXT)
       (SETF (FIRST ITEM) ':FNAME)
       (SETQ *MOUSE-DOC*
 "L:Display definition highlighting unused code, M:Edit this function, R2:System menu."
     ) ) )
   (SEND MAIN-PANE :SET-ITEMS REPORT-ITEM-LIST)
   ;; position first item at top of screen
   (SEND main-pane :put-item-in-window (SEND main-pane :item-of-number 0)) )
 (VALUES)
 )
(DEFCOMMAND backup-cmd ()
  '(:DESCRIPTION "Go back to the previous display"
    :MENUS ((prober-menu-ucl))
    :NAMES ("Previous display")
    :KEYS (#\c-m-L #\rubout))
  (DECLARE (SPECIAL main-pane interactor-pane appl-frame))
  (UNLESS (NULL *SAVED-DISPLAY*)
    (SEND *SAVED-DISPLAY* :RESTORE MAIN-PANE))
 )

(DEFCOMMAND report-cmd ()
	    '(:DESCRIPTION "Display information for a single function."
			   :MENUS ((prober-menu-ucl))
			   :NAMES ("Report one")
			   :KEYS (#\super-q))
  (DECLARE (SPECIAL main-pane interactor-pane appl-frame))
  (SAVE-DISPLAY)
  (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (FORMAT interactor-pane "~a"
	  "; (PROBE-REPORT-WINDOW 'function-name ...)")
  (SEND appl-frame :handle-prompt)
  (SEND interactor-pane :force-kbd-input
	"(compiler:probe-report-window ")
  )
(DEFUN PROBE-REPORT-WINDOW ( &REST NAMES )
  (LET ( REPORT-ITEM-LIST )
    (DECLARE (SPECIAL MAIN-PANE))
    (LET (( *STANDARD-OUTPUT* (MAKE-INSTANCE 'SCROLL-OUTPUT-STREAM) ))
      (SEND *STANDARD-OUTPUT* :SET-CURRENT-FONT *headline-font*)
      (PROBE-REPORT-INTERNAL NAMES)
      (SETQ REPORT-ITEM-LIST (SEND *STANDARD-OUTPUT* :GET-ITEM-LIST))
      (SETQ *MOUSE-DOC* *DEFAULT-MOUSE-DOC*) )
    (SEND MAIN-PANE :SET-ITEMS REPORT-ITEM-LIST)
    (SETQ *MOUSE-DOC* *DEFAULT-MOUSE-DOC*) )
  (VALUES) )

(DEFCOMMAND report-all-cmd ()
  '(:DESCRIPTION "Display the report followed by a display of all of the paths executed by the probed functions."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Report all")
    :KEYS (#\super-a))
  (DECLARE (SPECIAL main-pane interactor-pane appl-frame))
 #| (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (FORMAT interactor-pane "~a" "The report all command is currently unimplemented.")
  (SEND appl-frame :handle-prompt)
 |#
 ; (let ((*standard-output* main-pane))
 ;   (probe-report t) )
 (LET ( REPORT-ITEM-LIST )
   (LET (( *STANDARD-OUTPUT* (MAKE-INSTANCE 'SCROLL-OUTPUT-STREAM) ))
     (SEND *STANDARD-OUTPUT* :SET-CURRENT-FONT *headline-font*)
     (PROBE-REPORT-INTERNAL '(T))
     (SETQ REPORT-ITEM-LIST (SEND *STANDARD-OUTPUT* :GET-ITEM-LIST)) )
   (SEND MAIN-PANE :SET-ITEMS REPORT-ITEM-LIST)
   (SETQ *MOUSE-DOC* *DEFAULT-MOUSE-DOC*) )
 )

(DEFCOMMAND report-zero-cmd ()
  '(:DESCRIPTION "List all probed functions which were never called."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Report never called")
    :KEYS (#\super-N))
  (DECLARE (SPECIAL main-pane interactor-pane appl-frame))
  (SAVE-DISPLAY)
 (LET ( REPORT-ITEM-LIST )
   (LET (( *STANDARD-OUTPUT* (MAKE-INSTANCE 'SCROLL-OUTPUT-STREAM) ))
     (SEND *STANDARD-OUTPUT* :SET-CURRENT-FONT *text-font*)
     (PROBE-REPORT-INTERNAL '(0))
     (SETQ REPORT-ITEM-LIST (SEND *STANDARD-OUTPUT* :GET-ITEM-LIST)) )
   (SEND MAIN-PANE :SET-ITEMS REPORT-ITEM-LIST)
   (SETQ *MOUSE-DOC* *DEFAULT-MOUSE-DOC*))
 )

(DEFCOMMAND clear-data-cmd ()
  '(:DESCRIPTION "Erase all of the current probe data."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Clear data")
    :KEYS ())
  (DECLARE (SPECIAL main-pane interactor-pane appl-frame))
  (SEND main-pane :set-items nil)
  (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (SEND interactor-pane :force-kbd-input "(compiler:clear-probe-data)"))


(DEFCOMMAND write-data-cmd ()
  '(:DESCRIPTION "Write the probe data out to an XLD file which can be used again later."
    :MENUS ((prober-menu-ucl))
    :NAMES ("Write data")
    :KEYS (#\super-w))
  (DECLARE (SPECIAL interactor-pane appl-frame))
  (SEND interactor-pane :clear-screen)
  (SEND appl-frame :handle-prompt)
  (FORMAT interactor-pane "~a" "; (COMPILER:FASD-PROBE-DATA pathname)")
  (SEND appl-frame :handle-prompt)
  (SEND interactor-pane :force-kbd-input "(compiler:fasd-probe-data "))



(DEFUN prober-doc ()
  (DECLARE (SPECIAL appl-frame))
  (SI:WITH-HELP-STREAM (WINDOW :LABEL "Documentation for Prober"
			       :SUPERIOR appl-frame)
    ;; Note: multiple FORMAT calls are used below because of
    ;;  microcode bug on copying strings longer than 512.
    (FORMAT WINDOW "
  This is the completion probes utility.  It aids software testing
  by telling you which sections of code have been executed and 
  which haven't.  This then enables you to improve your test
  programs to exercise the missed sections.

  A typical use of this utility would be to do the following steps:
   1. Click on `Probing on' to enable probing.
   2. Compile the functions you want to test.  Probes will be
      automatically inserted into each function to monitor execution.
   3. Run your tests.")
	    (FORMAT WINDOW "
   4. Click on `Report summary' to get a list of all of the probed
      functions showing how many times each was called and what 
      percentage of the function was actually executed.
   5. Click on any individual function name in the display to see
      the function's definition with the un-executed portions of 
      the code shown in boldface.
   6. Click on `Previous display' to return to the summary listing,
      and repeat steps 5 and 6 as desired.")
	    (FORMAT WINDOW "
   7. When finished, click on `Probing off' to return the 
      compiler to its normal state.

  Other options are available which you can find by perusing the menu.
  For batch mode use, or to direct reports to a file, see the additional
  information in the file /"TOOLS;PROBES.TEXT/"."
 )))

(BUILD-COMMAND-TABLE 'prober-cmd-table 'prober-frame
  '(prober-doc-cmd
    exit-cmd
    edit-cmd
    top-cmd
    bottom-cmd
    page-up-cmd
    page-down-cmd
    probing-on-cmd probing-off-cmd
    probe-functions-cmd
;;    probe-file-cmd
;;    save-probes-cmd
    report-cmd
    report-summary-cmd
    report-frequency-cmd
    report-all-cmd
    report-zero-cmd
    backup-cmd
    clear-data-cmd
    write-data-cmd)
  :INIT-OPTIONS
  '(:NAME "Prober command table"
    :DOCUMENTATION "The Prober commands."))


(BUILD-MENU 'prober-menu-ucl 'prober-frame
  :default-item-options
    '(:font fonts:cptfont)
  :item-list-order  
    '(prober-doc-cmd
      backup-cmd
      exit-cmd
      edit-cmd
      probing-on-cmd probing-off-cmd
      probe-functions-cmd
;;      probe-file-cmd
;;      save-probes-cmd
      report-summary-cmd
      report-frequency-cmd
      report-cmd
      report-all-cmd
      report-zero-cmd
      clear-data-cmd
      write-data-cmd))

(DEFVAR prober nil "A var to hold the prober instance.")

(DEFUN user:prober (&optional new-instantiation-p)
  "This function activates the window interface of the completness probe utility."
  (WHEN (OR (NULL prober)
	     new-instantiation-p)
    (SETQ prober (MAKE-INSTANCE 'prober-frame)))
  (SEND prober :select))

(defmethod (prober-frame :before :kill) (&rest ignore)
  (when (eq self prober) (setq prober nil)) ; make a new instance next time
  )

(COMPILE-FLAVOR-METHODS prober-frame prober-main-pane
			prober-interaction-pane  prober-menu-pane)

;;; --------------

(DEFFLAVOR SCROLL-OUTPUT-STREAM (
	     ( ITEM-LIST NIL )
	     ( CURRENT-LINE (MAKE-ARRAY 150 :ELEMENT-TYPE 'STRING-CHAR
					    :FILL-POINTER 0) )
	     ( CURRENT-FONT NIL )
	     ( FONT-LIST NIL )
	     ( FIRST-NON-BLANK-COLUMN NIL )
	    ) ( SI:OUTPUT-STREAM )
  :INITTABLE-INSTANCE-VARIABLES
  (:DOCUMENTATION "Transform output into an item list for a scrolling window.") )

(DEFMETHOD ( SCROLL-OUTPUT-STREAM :TYO ) ( CHAR )
  (IF (= CHAR #\NEWLINE)
      (SEND SELF :FRESH-LINE)
    (PROGN
      (WHEN (AND (NULL FIRST-NON-BLANK-COLUMN)
		 (NOT (= CHAR #\SPACE)) )
	(SETQ FIRST-NON-BLANK-COLUMN (FILL-POINTER CURRENT-LINE))
	(UNLESS (OR (EQ CURRENT-FONT *TEXT-FONT*)
		    (= FIRST-NON-BLANK-COLUMN 0))
	  ;; Always use the same font for leading blanks so that
	  ;; indentation is uniform.
	  (SETQ FONT-LIST (LIST (CONS 0 *TEXT-FONT*)
				(CONS FIRST-NON-BLANK-COLUMN CURRENT-FONT)))) )
      (VECTOR-PUSH CHAR CURRENT-LINE)
      (WHEN (AND (NULL FONT-LIST)
		 (NOT (NULL CURRENT-FONT)) )
	(SETQ FONT-LIST (LIST (CONS 0 CURRENT-FONT))) )
       ) )
  CHAR )

(DEFMETHOD ( SCROLL-OUTPUT-STREAM :FRESH-LINE ) ()
  (LET (( LINE (SUBSEQ CURRENT-LINE 0 (FILL-POINTER CURRENT-LINE)) ))
    (PUSH (IF (EQUAL FONT-LIST `((0 . ,*TEXT-FONT*)))
	      (LIST :TEXT LINE)
	    (LIST :MIXED FONT-LIST LINE) )
	  ITEM-LIST) )
  (SETF (FILL-POINTER CURRENT-LINE) 0)
  (SETQ FONT-LIST NIL)
  (SETQ FIRST-NON-BLANK-COLUMN NIL))
	   
(DEFMETHOD ( SCROLL-OUTPUT-STREAM :SET-CURRENT-FONT ) ( NEW-FONT )
  (UNLESS (EQ NEW-FONT CURRENT-FONT)
    (LET* (( COLUMN (FILL-POINTER CURRENT-LINE) )
	   ( PREVIOUS-COLUMN (- COLUMN 1)))
      (WHEN (AND (NOT (NULL FIRST-NON-BLANK-COLUMN))
		 (= PREVIOUS-COLUMN FIRST-NON-BLANK-COLUMN)
		 (= (AREF CURRENT-LINE PREVIOUS-COLUMN) #/( ) )
	;; Avoid leading left parenthesis in different font.
	(SETQ COLUMN PREVIOUS-COLUMN) )
      (SI:PUSH-END (CONS COLUMN NEW-FONT) FONT-LIST)
      (SETQ CURRENT-FONT NEW-FONT) ) ) )

(DEFMETHOD ( SCROLL-OUTPUT-STREAM :READ-CURSORPOS ) ( UNITS )
  ;; This is needed for FORMAT with ~T to work.
  (ASSERT (EQ UNITS :CHARACTER))
  (VALUES (FILL-POINTER CURRENT-LINE) 0) )
(DEFMETHOD ( SCROLL-OUTPUT-STREAM :INCREMENT-CURSORPOS ) ( X Y UNITS )
  ;; This is needed for FORMAT with ~T to work.
  (ASSERT (EQ UNITS :CHARACTER))
  (IGNORE Y)
  (DOTIMES ( J X )
    (VECTOR-PUSH #\SPACE CURRENT-LINE) ) )

(DEFMETHOD ( SCROLL-OUTPUT-STREAM :GET-ITEM-LIST ) ()
  (UNLESS (<= (FILL-POINTER CURRENT-LINE) 0)
    (SEND SELF :FRESH-LINE) )
  (LET (( RETURN-VALUE (NREVERSE ITEM-LIST) ))
    (SETQ ITEM-LIST NIL)
    RETURN-VALUE ) )

(COMPILE-FLAVOR-METHODS SCROLL-OUTPUT-STREAM)

;;; =====================

(eval-when ( load )
  (format t "~&; Execute (USER:PROBER) to start the probe utility.") 
  
  (dolist (key '( #\R #\Q #\Z #\SUPER-P #\HYPER-P #\SUPER-R)) ; find an available system key
    (unless (assoc key W:*system-keys*)
      (when (ignore-errors (w:add-system-key key '(or prober 'prober-frame)
					     "Report execution data for probed code."
					     t nil "Prober")
			   t)
	(format t "  Or you can press SYSTEM ~:@C." key)
	(return))))
  (format t "~&"))
