;;; -*- Mode:Common-Lisp; Package:system; Base:10; Fonts:(CPTFONT HL12B HL12I MEDFNB) -*-


;;; This is a useful function  in time-critical applications
;;; It's a modified version of process-enqueue, but with a timeout added.
;;; It returns T if the lock is held, nil if not.

(defun process-enqueue-with-timeout (queue timeout &optional queue-value (whostate "Lock")
				     &aux (starting-time (time:time)))
  (unless queue-value
    (setq queue-value sys:current-process))
  (when (si:have-queue-lock queue queue-value)
    (error "~S is already locked by this process" queue))
  ;; Try and lock or enqueue us, 
  (loop until (time:time-elapsed-p timeout starting-time)
	for sys:inhibit-scheduling-flag = t
	for queue-length = (length queue)
	and current-pointer = (si:process-queue-current-pointer queue)
	;; Lock if queue is empty
	when (null (aref queue current-pointer))
	do (setf (aref queue current-pointer) queue-value)
	   (setf (si:process-queue-insert-pointer queue)
		 (mod (1+ current-pointer) queue-length))
	   (return)
	for insert-pointer = (si:process-queue-insert-pointer queue)
	;; If queue is not full, then enqueue us
	when (neq insert-pointer current-pointer)
	do (setf (aref queue insert-pointer) queue-value)
	   (setf (si:process-queue-insert-pointer queue)
		 (mod (1+ insert-pointer) queue-length))
	   (let ((aborted t))
	     (unwind-protect
	       (progn
		 (process-wait whostate
		   #'(lambda (queue queue-value timeout starting-time)
		       (or (si:have-queue-lock queue queue-value)
			   (time:time-elapsed-p timeout starting-time)))
		   queue queue-value timeout starting-time)
		 (setq aborted (time:time-elapsed-p timeout starting-time)))
	       (when aborted
		 (without-interrupts
		   (cond ((si:have-queue-lock queue queue-value)
			  (si:process-dequeue queue queue-value))
			 (t (setf (aref queue insert-pointer) nil))))))
	     (return))
	do (process-wait whostate
	     #'(lambda (queue timeout starting-time)
		 (let ((current (si:process-queue-current-pointer queue)))
		   (or (null (aref queue current))
		       (neq current (si:process-queue-insert-pointer queue))
		       (time:time-elapsed-p timeout starting-time))))
	     queue timeout starting-time))
  (si:have-queue-lock queue queue-value))