;;; -*- Mode:Common-Lisp; Package:FILE-SYSTEM; Base:8; Patch-file:T; Fonts:(CPTFONT) -*-


;;; These patches are to be loaded on a RELEASE 2 system only. They provide compatibilty 
;;;  for backing up data from a Release 3 system to a Release 2 system

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1986, Texas Instruments Incorporated. All rights reserved.
;;; Patch file for File System version 2.2
;;; Reason       : Convert pathname string on the magtape header to pathname object so that Rel2 system can read Rel3 generated tapes.
;;;-
;;; Bug Report   : none
;;;-
;;; Problem      : Rel2 software can not read the tape generated by rel3 software.
;;;-
;;; Solution     : Convert pathname string on the magtape header to pathname object.
;;;-
;;; Dependencies : none
;;;-
;;; Code Readers : mp mbc
;;;-
;;; Written 1/16/87 14:58:28 by MARYAM,
;;; while running on Hotel from band LOD1
;;; with System 2.78, Compiler 2.7, File System 2.1, Universal Command Loop 2.0, Window System 2.10, Input Editor 2.0, ZMACS 2.11, Error Handler 2.2, Suggestions 2.23, Debug Utilities 2.12, Explorer-Net 2.7, Telnet 2.2, Vt100 2.1, File Server 2.0, Net-Config 2.4, Font Editor 2.8, Mailer 2.7, Mail-Reader 2.5, Streamer-Tape 2.25, Local-File 2.37, System-Log 2.3, Serial-Parallel 2.0, Printer 2.6, Glossary 2.0, IMAGEN 2.3, NVRAM 2.3, User Profile Utility 2.2, UCODE-DEPENDENT 2.18, IP 1.28, Experimental Code Management Interface 2.31, Experimental TGC 1.113, Experimental CMI 1.0, Experimental Explorer Bug System 22.1, microcode 61, FAN36-TGC61-AUS.



#!Z
; From file PATHNM.LISP#> TAPE-TEST; Hotel:
#8R FILE-SYSTEM#:
(COMPILER-LET ((PACKAGE (PKG-FIND-PACKAGE "FILE-SYSTEM"))
                         (SI:LISP-MODE :ZETALISP)
                         (*READTABLE* SI:STANDARD-READTABLE)
                         (SI:*READER-SYMBOL-SUBSTITUTIONS* NIL))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: TAPE-TEST; PATHNM.#"


(defun pathname-string-to-object (string)	;so everyone can read the new "MORE GENERIC" versions
  (let ((value (with-lisp-mode :COMMON-LISP
		 (READ-FROM-STRING string))))
    (if (consp value)
	(APPLY #'MAKE-PATHNAME VALUE)		;We're expecting (:name "nnn" :directory ("d1" "d2") ...)
	value)))
))




;;; Patch file for File System version 2.5
;;; Reason: Modified write-property-list, read-property-list and skip-property-list to handle BIGNUm and Negative Numbers.
;;; Written 3/24/87 06:59:16 by BERGER,
;;; while running on ARIES from band LOD2
;;; with System 2.78, Compiler 2.7, File System 2.1, Universal Command Loop 2.0, Window System 2.10, Input Editor 2.0, ZMACS 2.10, Error Handler 2.2, Suggestions 2.22, Debug Utilities 2.12, Explorer-Net 2.7, Telnet 2.2, Vt100 2.1, File Server 2.0, Net-Config 2.4, Font Editor 2.2, Mailer 2.7, Mail-Reader 2.5, Streamer-Tape 2.20, Local-File 2.31, System-Log 2.3, Serial-Parallel 2.0, Printer 2.6, Glossary 2.0, IMAGEN 2.3, NVRAM 2.3, User Profile Utility 2.1, UCODE-DEPENDENT 2.17, microcode 315, Rel 2.2 Minprod.



#!Z
; From file FSDEFS.LISP file; sys:
#8R FILE-SYSTEM#:
(COMPILER-LET ((PACKAGE (PKG-FIND-PACKAGE "FILE-SYSTEM"))
                         (SI:LISP-MODE :ZETALISP)
                         (*READTABLE* SI:STANDARD-READTABLE)
                         (SI:*READER-SYMBOL-SUBSTITUTIONS* NIL))
  (COMPILER#:PATCH-SOURCE-FILE "sys: file; FSDEFS.;"

(DEFUN GET-BYTES (STREAM NBYTES)
     (LET ((VALUE 0))
       (DOTIMES (I NBYTES VALUE)
         (SETF VALUE (+ (* VALUE 256.)
   		     (SEND STREAM :TYI))))))
			       
(DEFMACRO PUT-BYTES (STREAM NBYTES VALUE)
  (if (fixnump nbytes)
      `(PROGN .  ,(LOOP FOR PPSS FROM (+ (* (1- nbytes) 512.) 8.) DOWNTO 8. BY 512.
		   COLLECTING `(FUNCALL ,STREAM ':TYO (LDB ,PPSS ,VALUE))))
      `(LOOP FOR PPSS FROM (+ (* (1- ,nbytes) 512.) 8.) DOWNTO 8. BY 512.
		   do (FUNCALL ,STREAM ':TYO (LDB PPSS ,VALUE)))))


(DEFUN WRITE-PROPERTY-LIST (STREAM PLIST &AUX (X (FLOOR (LENGTH PLIST) 2)))
  (PUT-BYTES STREAM 1 X)
  (DO ((I X (1- I))
       (L PLIST (CDDR L))
       (PROP))
      ((ZEROP I))
    (PUT-STRING (format nil "~s" (car l)) STREAM)
    (COND ((NULL (SETQ PROP (CADR L)))
	   (FUNCALL STREAM ':TYO 0))
	  ((EQ PROP T)
	   (FUNCALL STREAM ':TYO 1))
	  ((SYMBOLP PROP)
	   (COND ((MEMQ (SYMBOL-PACKAGE PROP) '#,(LIST si:pkg-user-package
						       si:pkg-global-package
						       NIL))	;Uninterned (interned later)
		  (FUNCALL STREAM ':TYO 2)
		  (PUT-STRING (GET-PNAME PROP) STREAM))
		 (T (FUNCALL STREAM ':TYO 3)
		    ;; Use this odd order for ease in interning later.
		    (PUT-STRING (GET-PNAME PROP) STREAM)
		    (PUT-STRING (SI:PKG-NAME (SYMBOL-PACKAGE PROP)) STREAM))))
	  ((AND (NUMBERP PROP)			;modified to handle neagtive numbers and bignums in property list
		(integerp PROP))
	   (cond ((eq (type-of prop) 'BIGNUM)
		  (let* ((prop-value (abs prop))
			 (length-prop (length (format nil "~a" prop-value))))   
		    (funcall stream :tyo 9.)
		    (funcall stream :tyo length-prop)
		    (if (plusp prop)
			(funcall stream :tyo 0)	;positive bignum
			(funcall stream :tyo 1)	;negative bignum
			)
		    (put-bytes stream length-prop  prop-value)
		    ))
		 (t (if (or (plusp prop)
			    (zerop prop))
			(progn
			  (FUNCALL STREAM ':TYO 4.)	;positive integers
			  (PUT-BYTES STREAM 3 PROP))	;negative integers
			(FUNCALL STREAM ':TYO 8.)
			(PUT-BYTES STREAM 3 PROP)))))
	  ((STRINGP PROP)
	   (FUNCALL STREAM ':TYO 5)
	   (PUT-STRING PROP STREAM))
	  ((LISTP PROP)
	   (FUNCALL STREAM ':TYO 6)
	   (LET ((*PACKAGE* SI:PKG-USER-PACKAGE)
		 (IBASE 10.)
		 (*NOPOINT T)
		 (SI:PRINT-READABLY T))
	     (PRIN1 PROP STREAM)))
	  ((TYPEP PROP 'PATHNAME)
	   (FUNCALL STREAM ':TYO 7)
	   (LET ((*PACKAGE* SI:PKG-USER-PACKAGE)
		 (BASE 10.)
		 (*NOPOINT T)
		 (SI:PRINT-READABLY T))
	     ;; These are the arguments to MAKE-FASLOAD-PATHNAME.
	     (FUNCALL STREAM ':TYO #/()
	     (PRIN1-THEN-SPACE (FUNCALL (PATHNAME-HOST PROP) ':NAME-AS-FILE-COMPUTER) STREAM)
	     (PRIN1-THEN-SPACE (PATHNAME-DEVICE PROP) STREAM)
	     (PRIN1-THEN-SPACE (PATHNAME-DIRECTORY PROP) STREAM)
	     (PRIN1-THEN-SPACE (PATHNAME-NAME PROP) STREAM)
	     (PRIN1-THEN-SPACE (PATHNAME-TYPE PROP) STREAM)
	     (PRIN1 (PATHNAME-VERSION PROP) STREAM)
	     (FUNCALL STREAM ':TYO #/))))
	  (T (FERROR NIL "I don't know how to write ~S as a property." PROP)))))

(DEFUN READ-PROPERTY-LIST (STREAM &AUX LIST (PAK SI:PKG-keyword-PACKAGE))
  (SETQ LIST (MAKE-LIST (* (FUNCALL STREAM ':TYI) 2)))
  (DO ((L LIST (CDDR L)))
      ((NULL L))
    (RPLACA L (LET ((*PACKAGE* PAK)) (IGNORE-ERRORS (READ-FROM-STRING (GET-STRING STREAM)))))
    (SELECTQ (FUNCALL STREAM ':TYI)
      (0)
      (1 (SETF (CADR L) T))
      (2 (SETF (CADR L) (INTERN (GET-STRING STREAM) PAK)))
      (3 (SETF (CADR L) (INTERN (READ-FROM-STRING (GET-STRING STREAM))
				(PKG-FIND-PACKAGE (GET-STRING STREAM) ':find))))
      (4 (SETF (CADR L) (GET-BYTES STREAM 3)))
      (5 (SETF (CADR L) (GET-STRING STREAM)))
      (6 (SETF (CADR L)
	       (PKG-BIND SI:PKG-USER-PACKAGE
		 (LET ((IBASE 10.))
		   ;; This can lose pretty badly with #<'s, etc.  -- DLA
		   (READ STREAM)))))
      (7 (PKG-BIND ""
	   (LET* ((IBASE 10.)
		  (FORM (READ STREAM))
		  (DEFAULT-CONS-AREA WORKING-STORAGE-AREA))	;<-*
	     (SETF (CADR L)
		   (IF (= (LENGTH FORM) 6)
		       (APPLY #'FS:MAKE-FASLOAD-PATHNAME FORM)
		       (EVAL FORM))))))		;Obsolete form for pathnames to be written.
      (8. (SETF (CADR L) (%logdpb #x01 #o3001 (GET-BYTES STREAM 3.))))    ;negative number
      (9. (let* ((length-prop (get-bytes stream 1))                       ;BIGNUMs
		 (sign-prop (get-bytes stream 1))
		 (prop-value (get-bytes stream length-prop)))
	         (setf (cadr l) (if (eq sign-prop 0) prop-value (* -1 prop-value)))))
      (OTHERWISE (FERROR NIL "Invalid Plist property designator."))))
  LIST)

))

#!Z
; From file FSGUTS.LISP file; sys:
#8R FILE-SYSTEM#:
(COMPILER-LET ((PACKAGE (PKG-FIND-PACKAGE "FILE-SYSTEM"))
                         (SI:LISP-MODE :ZETALISP)
                         (*READTABLE* SI:STANDARD-READTABLE)
                         (SI:*READER-SYMBOL-SUBSTITUTIONS* NIL))
  (COMPILER#:PATCH-SOURCE-FILE "SYS: FILE; FSGUTS.#"


(defun skip-bignum (stream)
  (let ((length-bignum (get-bytes stream 1.)))             ;first byte is the length, not counting sign byte.
    (skip-bytes stream (1+ length-bignum))))

(DEFUN SKIP-PROPERTY-LIST (STREAM)
  "Skips over the property list of the directory entry on STREAM."
  (LET ((IBASE 10.))
    (DOTIMES (I (FUNCALL STREAM :TYI))
	
      (SKIP-STRING STREAM)
      (SELECTQ (FUNCALL STREAM :TYI)
	(0)
	(1 NIL)
	(2 (SKIP-STRING STREAM))
	(3 (PROGN (SKIP-STRING STREAM) (SKIP-STRING STREAM)))
	(4 (SKIP-BYTES STREAM 3))
	(5 (SKIP-STRING STREAM))
	(6 (READ STREAM))
	(7 (SKIP-PATHNAME STREAM))
	(8. (skip-bytes stream 3.))
	(9. (skip-bignum stream))
	(OTHERWISE (NOTIFY "Invalid file property list designator.")
		   (*THROW 'MUST-SALVAGE T))))))
))
