;;; -*- Mode:Common-Lisp; Package:USER; Base:10; Fonts:(CPTFONT) -*-

;;; Change History
;;;
;;;  Date	Author	Description
;;; -------------------------------------------------------------------------------------
;;; 09/06/88    MAY     1Deleted (listen) from BALLS.*
;;; 05/18/88    RSW     TWE's IF to push on hacks list or else directly
;;;                     as only hack.
;;; 10/30/87	LGO	Modified to allow setting ball radius as a parameter.
;;; 10/20/87    TWE	Updated to handle different sized balls in a more generic manner.
;;;			Also fixed up the calculation for *mini-raster-offset* to make
;;;			sure that the size of the ball image array is calculated
;;;			correctly.  Previously it was `off by one'.  Made the drawing of
;;;			the balls appear smoother by displaying the results of a movement
;;;			instead of doing two XORs.  This makes large balls appear much
;;;			better by removing much of the flicker.  Fixed up the top-level
;;;			function (i.e. balls) to use its ball count argument properly.

(DEFPARAMETER maxvel 1500)
(DEFPARAMETER elastic 800)             ; elasticity coefficient
(defparameter *number-of-balls* 5)
(DEFVAR ball-array (make-array *number-of-balls*))       ; The universe of balls.
(defvar number-balls 0)          ; current number of balls.
(defvar ball-image nil)          ; bitblt array
(defvar ball-temp-image nil)     ; temporary bitblt array
(defvar combined-ball-image nil) ; combination of old and new balls
(proclaim '(special MinX MaxX MinY MaxY))	; window dimensions, actual
(proclaim '(special VMaxX VMaxY))		; window dimensions, virtual

;;; Macros to translate between internal (virtual) coordinates and
;;;   screen coordinates:

(defconstant virt-coordinates-scale-factor 64)
(defmacro virt-to-screen(q) `(floor ,q virt-coordinates-scale-factor))
(defmacro screen-to-virt(q) `(floor (* ,q virt-coordinates-scale-factor)))

(defparameter *ball-radius* 32)			; radius of ball
(defvar *ball-diameter*); diameter
(defvar *ball-v-diameter*); diameter
(defvar *ball-dsq*)
(defvar *mini-raster-size*)      ; size of miniraster
(defvar *mini-raster-offset*)    ; distance to miniraster center
(defvar *mini-raster-slack*)

;;; Initialize window, various globals.

(defun SetupCrapolla(number-of-balls radius)  ; in honor of you-know-who

  (SETQ *ball-radius* radius)
  (SETQ *ball-diameter* (* 2 *ball-radius*)); diameter
  (SETQ *ball-v-diameter* (screen-to-virt (* 2 *ball-radius*))); diameter
  (SETQ *ball-dsq* (* (screen-to-virt *ball-diameter*)
			   (screen-to-virt *ball-diameter*)))
  (SETQ *mini-raster-size* (truncate (* *ball-radius* 2.5)))      ; size of miniraster
  (SETQ *mini-raster-offset*		; distance to miniraster center
	     (ceiling *mini-raster-size* 2))
  (SETQ *mini-raster-slack* (- *mini-raster-offset* *ball-radius*))

  (setq ball-array (make-array number-of-balls))
  (setq number-balls 0)
  (multiple-value-setq (MinX MinY MaxX MaxY) (send *terminal-io* ':inside-edges))
  (incf MinX *ball-radius*)
  (incf MinY *ball-radius*)

  ;;; Virtual coordinates q range:  0 <= q < VMaxQ...
  (setq VMaxX (screen-to-virt (- MaxX MinX *mini-raster-size*)))
  (setq VMaxY (screen-to-virt (- MaxY MinY *mini-raster-size*)))

  (send *terminal-io* :clear-screen)

  ;;;  Build the ball image mini-raster directly on the screen:
  (setq ball-image (tv:make-sheet-bit-array *terminal-io*
		   (* 2 *mini-raster-offset*) (* 2 *mini-raster-offset*)))
  (setq ball-temp-image (tv:make-sheet-bit-array *terminal-io*
		   (* 2 *mini-raster-offset*) (* 2 *mini-raster-offset*)))
  (setq combined-ball-image (tv:make-sheet-bit-array *terminal-io*
		   (* 6 *mini-raster-offset*) (* 6 *mini-raster-offset*)))
  (CREATE-BALL-IMAGE)

  ;;; Now paint the initial screen:
  (send *terminal-io* :clear-screen)
  #+comment
  (do ((i 0 (+ 1 i)))
      ((= i 5))
    (let ((dd (- *mini-raster-offset* *ball-radius* i i)))
      (send *terminal-io* :draw-lines tv:alu-xor
	  (+ MinX dd) (+ MinY dd)
	  (+ MinX dd) (- MaxY dd)
	  (- MaxX dd) (- MaxY dd)
	  (- MaxX dd) (+ MinY dd)
	  (+ MinX dd) (+ MinY dd) )))
  )

(DEFUN CREATE-BALL-IMAGE ()
;;;  The old code to create the ball is commented out.  It left a tiny opening in the
;;;  center of the ball.  The new code does not have this problem.
;;;  (if (get-handler-for *terminal-io* :draw-filled-in-circle)
;;;      (send *terminal-io* :draw-filled-in-circle *mini-raster-offset* *mini-raster-offset*
;;;            *Ball-radius* tv:alu-xor)
;;;      (send *terminal-io* :draw-filled-circle *mini-raster-offset* *mini-raster-offset*
;;;            *Ball-radius* w:black w:alu-xor))
;;;  (send *terminal-io* :bitblt-from-sheet tv:alu-seta
;;;		   (* 2 *mini-raster-offset*) (* 2 *mini-raster-offset*)
;;;		   0 0 ball-image 0 0)
  (CREATE-BALL *mini-raster-offset* *mini-raster-offset* BALL-IMAGE))

(defun CREATE-BALL (center-x center-y array)
  (LET ((ALU W:ALU-XOR)
        (radius *ball-radius*))
    (DO ((Y 0)
         (F 0)                         ; F is just Y squared without any multiplies
         (X RADIUS)
         (last-x radius))
        (NIL)
      (si:%draw-shaded-raster-line (- CENTER-X X) (+ CENTER-X X) (- CENTER-Y Y) alu t nil array)
      ;; Draw the middle line only once.
      (when (not (zerop y))
        (si:%draw-shaded-raster-line (- CENTER-X X) (+ CENTER-X X) (+ CENTER-Y Y) alu t nil array))
      ;; Handle pixel errors by only drawing this upper/lower part once.
      (when (not (= x last-x))
        (si:%draw-shaded-raster-line (- CENTER-X y) (+ CENTER-X y) (- CENTER-Y x) alu t nil array)
        (si:%draw-shaded-raster-line (- CENTER-X y) (+ CENTER-X y) (+ CENTER-Y x) alu t nil array))
      (setq last-x x)
      (SETQ F (+ F Y Y 1)
            Y (1+ Y))
      (COND ((>= F X) (SETQ F (- F X X -1)
                            X (- X 1))))
      (COND ((> Y X) (RETURN)))
      (COND ((= Y X) (RETURN))))))

(defflavor Ball ((x-position 100)		; Positions of CENTER of ball...
		 (y-position 100)		;   ... in virt coordinates.
		 (x-velocity 0)			; 1st differences, virt coords.
		 (y-velocity 0)
		 ball-number
		 (OnScreen nil))		; T iff image on screen.
	   ()
  :gettable-instance-variables
  :settable-instance-variables
  #+3600
  :initable-instance-variables
  #+Explorer
  :inittable-instance-variables
  (:outside-accessible-instance-variables x-position y-position)
  )

;;; XOR a ball on the screen at center position x, y (screen)
(defun XOR-ball(x y)
  (bitblt tv:alu-xor *mini-raster-size* *mini-raster-size*
          ball-image 0 0
          (w:sheet-screen-array *terminal-io*) (max 0 x) (max y 0)))

(defun MOVE-ball(x y dx dy)
  (bitblt tv:alu-seta *mini-raster-size* *mini-raster-size*
	  ball-image 0 0
	  ball-temp-image 0 0)
  (bitblt tv:alu-xor *ball-diameter* *ball-diameter*
	  ball-image *mini-raster-slack* *mini-raster-slack*
	  ball-temp-image
	  (+ *mini-raster-slack* dx)
	  (+ *mini-raster-slack* dy) )

  (send *terminal-io* :bitblt tv:alu-xor
		*mini-raster-size* *mini-raster-size* ball-temp-image 0 0
		(virt-to-screen x)
		(virt-to-screen y)))

;;; Combine the two balls together and display the change.
;;; In this way we reduce the flicker due to XOR when the
;;; size of a ball is large.  Care is taken to only bitblt
;;; what we need to and no more.
(defun move-balls (old-x old-y new-x new-y)
  (let ((delta-x (- new-x old-x))
        (delta-y (- new-y old-y))
        (inside-left   (w:sheet-inside-left   *terminal-io*))
        (inside-top    (w:sheet-inside-top    *terminal-io*))
        (inside-right  (w:sheet-inside-right  *terminal-io*))
        (inside-bottom (w:sheet-inside-bottom *terminal-io*))
        (dest-x (- old-x *mini-raster-size*))
        (dest-y (- old-y *mini-raster-size*))
        (source-x 0)
        (source-y 0)
        (combined-image-width  (* 3 *mini-raster-size*))
        (combined-image-height (* 3 *mini-raster-size*)))
    ;; Erase the combined image.
    (bitblt tv:alu-setz combined-image-width combined-image-height
            combined-ball-image 0 0
            combined-ball-image 0 0)
    ;; Copy a ball to the center.
    (bitblt tv:alu-seta  *mini-raster-size*  *mini-raster-size*
            ball-image 0 0
            combined-ball-image  *mini-raster-size* *mini-raster-size*)
    ;; Combine with the new ball.
    (bitblt tv:alu-xor *mini-raster-size* *mini-raster-size*
            ball-image 0 0
            combined-ball-image (+ delta-x *mini-raster-size*) (+ delta-y *mini-raster-size*))

    ;; Make the image width/height and source-x/y be just big enough for the image.
    (setq combined-image-width (+ *mini-raster-size* (abs delta-x)))
    (if (plusp delta-x)
        (setq source-x *mini-raster-size*
              dest-x (+ dest-x *mini-raster-size*))
        ;;1ELSE*
        (setq source-x (+ *mini-raster-size* delta-x)
              dest-x (+ dest-x *mini-raster-size* delta-x)))
    (setq combined-image-height (+ *mini-raster-size* (abs delta-y)))
    (if (plusp delta-y)
        (setq source-y *mini-raster-size*
              dest-y (+ dest-y *mini-raster-size*))
        ;;1ELSE*
        (setq source-y (+ *mini-raster-size* delta-y)
              dest-y (+ dest-y *mini-raster-size* delta-y)))

    ;; Copy the result to the screen.
    ;;  Clip on the left and top.
    (when (< dest-x inside-left)
      (incf combined-image-width (- dest-x inside-left))
      (setq source-x (+ source-x (- dest-x) inside-left))
      (setq dest-x inside-left))
    (when (< dest-y inside-top)
      (incf combined-image-height (- dest-y inside-top))
      (setq source-y (+ source-y (- dest-y) inside-top))
      (setq dest-y inside-top))
    ;; Clip on the right and bottom.
    (when (> (+ combined-image-width dest-x) inside-right)
      (setq combined-image-width (- inside-right dest-x)))
    (when (> (+ combined-image-height dest-y) inside-bottom)
      (setq combined-image-height (- inside-bottom dest-y)))
    (bitblt tv:alu-xor combined-image-width combined-image-height
            combined-ball-image source-x source-y
            (w:sheet-screen-array *terminal-io*) dest-x dest-y)))

(defmethod (ball :move-to)(vx vy)
  (let ((dx (- vx x-position))
	(dy (- vy y-position))
	(x (virt-to-screen vx))
	(y (virt-to-screen vy)))
    (if OnScreen
	(if (and nil (<= (abs dx) *mini-raster-slack*)
		 (<= (abs dy) *mini-raster-slack*))
	    (move-ball x y dx dy)
	    (progn
;;              (xor-ball (virt-to-screen x-position) (virt-to-screen y-position))
;;              (xor-ball x y)
              (move-balls x y (virt-to-screen x-position) (virt-to-screen y-position))
                   ))
        (xor-ball x y))
;;        (xor-ball (+ x *mini-raster-offset*) (+ y *mini-raster-offset*)))
    (setq x-position vx)
    (setq y-position vy))
  (setq OnScreen t))

;;; Apply a force vector to a ball:
(defmethod (ball :force)(dx dy)				; The force vector.
 	   (setq x-velocity (+ (floor (* elastic dx) 1024) x-velocity))
	   (setq y-velocity (+ (floor (* elastic dy) 1024) y-velocity)))

;;; Generate a force (DDQ) for bouncing:
(defun bounce-force(displacement) (* displacement 1))

;;; Update a coordinate:
;;; Adds in velocity after checking for wall bounce.

(defmacro update-ball-coord(Q DQ QMAX)`(progn
	   (if (< ,Q 0)(setq ,DQ (+ (bounce-force (- ,Q)) ,DQ)))
	   (if (>= ,Q ,QMAX)(setq ,DQ (- ,DQ (bounce-force (- ,Q ,QMAX)))))
	   (if (> ,DQ maxvel)(setq ,DQ maxvel))
	   (if (< ,DQ (- maxvel))(setq ,DQ (- maxvel)))
	   (setq ,Q (+ ,Q ,DQ)) ))

;;; Update the position of a ball:
(defmethod (ball :update)
	   (&aux (oldx x-position)(oldy y-position))
	   (update-ball-coord oldx x-velocity VMaxX)
	   (update-ball-coord oldy y-velocity VMaxY)
	   (send self :move-to oldx oldy))

#+comment
(defmethod (ball :dump) ()
;  (send *terminal-io* :set-cursorpos 0 (* 16 (send self :ball-number)))
  (format *terminal-io* "~& ~d: ~s: xy = (~d, ~d);  vxy = (~d, ~d)    "
	  (send self :ball-number)
	  self
	  (send self :x-position)
	  (send self :y-position)
	  (send self :x-velocity)
	  (send self :y-velocity)))

;;; Handle ball-ball collisions:
(defun collide(b1 b2)
  (let ((dx (- (ball-x-position b1)
	       (ball-x-position b2)))
	(dy (- (ball-y-position b1)
	       (ball-y-position b2)))
	dsq ddsq)
    
    (setq dsq (+ (* dx dx) (* dy dy)))
    
    (if (< dsq *ball-dsq*)
	(progn (setq dsq (isqrt dsq))
	       (setq ddsq (- *ball-v-diameter* dsq))
	       (setq dx (floor (* ddsq dx) dsq))
	       (setq dy (floor (* ddsq dy) dsq))
;	       (print (list dx dy dsq))
	       (send b1 :force dx dy)
	       (send b2 :force (- dx)(- dy))
	       ))))

;;; Update the world:
(defun update-step()
	   (do* ((i 0 (+ 1 i))
		 (b1 nil))
		((>= i number-balls))
		(send (setq b1 (aref ball-array i)) :update)
		(do ((j (+ 1 i) (+ 1 j)))
		    ((>= j number-balls))
		    (collide b1 (aref ball-array j)) )))

;; Make a new ball:
(defun new-ball(x y &optional (vx 0) (vy 0))
  (setf (aref ball-array number-balls)
        (make-instance 'Ball :x-position x :y-position y :x-velocity vx :y-velocity vy))
  (send (aref ball-array number-balls) :set-ball-number number-balls)
  (setq number-balls (+ 1 number-balls)))

;;; Make a random ball:
(defun rand-ball(&aux (x (random VMaxX))
		      (y (random VMaxY))
		      (vx (srandom (FLOOR maxvel 2)))
		      (vy (srandom (FLOOR maxvel 2))))
  (new-ball x y vx vy))

;;; generate a signed, random fixnum:
(defun srandom(mag)
  (- (random (+ 1 mag mag)) 1 mag))

;;; Try some stuff:

(defun balls (IGNORE &optional (n *number-of-balls*) (radius *ball-radius*))
  (IF (and (= number-balls n) (= n (length ball-array)) (= (* 2 radius) *ball-diameter*))
      (PROGN
	(send *terminal-io* :clear-screen)
	(DOTIMES (i n)
	  (SEND (AREF ball-array i) :set-onscreen nil)))
      ;;ELSE
    (SetupCrapolla n radius)
    (DOTIMES (i n)
      (rand-ball)))
  
  (do ()
      (nil ;(LISTEN)1 ;; may-9-1-88*
	)
;;    (PROCESS-ALLOW-SCHEDULE) ;; Do process-allow-schedule without updating the who-line-run-state
    (setf (si:process-quantum-remaining current-process) -1)
    (funcall si:scheduler-stack-group)
    
    (update-step))
  )

(IF (BOUNDP 'TV:*SCREEN-SAVER-HACKS-LIST*)
  (PUSHNEW 'BALLS TV:*SCREEN-SAVER-HACKS-LIST*)
  ;;ELSE
  (SETQ TV:*SCREEN-SAVER-HACKS* 'BALLS))



