;;; -*- Mode:Common-Lisp; Package:USER; Base:10 -*-

;; This file defines some more sounds for use by the BEEP function.
;;
;;  5/22/89 D.N.G.

(eval-when (eval compile)

(defun ding (&optional (frequency 614) (duration 50) (decay-rate 4) (decay-deceleration 10))
  (w:with-sound-enabled
    (w:do-sound (w:tone 1 frequency))
    (w:do-sound (w:volume 1 :on) (* 3 duration))
    (do ((level (+ w:on-volume 2) (1+ level))
	 (step (truncate duration decay-deceleration))
	 (time (truncate duration decay-rate) (+ time step)))
	((> level 12.))
      (w:do-sound (w:volume 1 level) time)))
  (values))

(defun chime (&optional (frequency 614) (duration 50) (decay-rate 4) (decay-deceleration 10))
  (w:with-sound-enabled
    (w:do-sound (w:tone 0 (* frequency 1)))
    (w:do-sound (w:tone 1 (* frequency 8.93s0)))
    (w:do-sound (w:tone 2 (* frequency 13.34s0)))
    (w:do-sound (w:volume 1 2))
    (w:do-sound (w:volume 2 4))
    (w:do-sound (w:volume 0 2) (* 3 duration))
    (do ((level (+ w:on-volume 2) (1+ level))
	 (step (truncate duration decay-deceleration))
	 (time (truncate duration decay-rate) (+ time step)))
	((> level 12.))
      (w:do-sound (w:volume 1 (min 15. (+ level 2))))
      (w:do-sound (w:volume 2 (min 15. (+ level 3))))
      (w:do-sound (w:volume 0 level) time)
      ))
  (values))

(defun bell (&optional (frequency 614) (duration 50) (decay-rate 4) (decay-deceleration 10))
  (w:with-sound-enabled
    (w:do-sound (w:tone 0 (* frequency 1.2s0)))
    (w:do-sound (w:tone 1 (* frequency 2.0s0)))
    (w:do-sound (w:tone 2 (* frequency 2.5s0)))
    (w:do-sound (w:volume 1 4))
    (w:do-sound (w:volume 2 6))
    (w:do-sound (w:volume 0 2) (* 3 duration))
    (do ((level (+ w:on-volume 2) (1+ level))
	 (step (truncate duration decay-deceleration))
	 (time (truncate duration decay-rate) (+ time step)))
	((> level 12.))
      (w:do-sound (w:volume 1 (min 15. (+ level 2))))
      (w:do-sound (w:volume 2 (min 15. (+ level 3))))
      (w:do-sound (w:volume 0 level) time)
      ))
  (values))

(defun play-note (reg &optional (freq 614) (volume 2))
  (w:with-sound-enabled
    (w:do-sound (w:tone reg freq))
    (w:do-sound (w:volume reg volume) 200)
    )
  (values))

(defun make-noise ( &optional (type :white) (hiss 1) (volume w:on-volume) (duration 200))
  (w:with-sound-enabled
    (w:do-sound (w:noise type hiss))
    (w:do-sound (w:volume :noise volume) duration)
    )
  (values))

(defun make-noise-f ( &optional (frequency 500) (type :white) (volume w:on-volume) (duration 200))
  (w:with-sound-enabled
    (w:do-sound (w:tone 2 (* frequency 14.3s0)))
    (w:do-sound (w:noise type :override))
    (w:do-sound (w:volume :noise volume) duration)
    )
  (values))

(defun scratch ()
  (make-noise-f 30 :white 1 150))

(defun dit-dit-dit ()
  (make-noise-f 14 :periodic 2 400))

(defun buzz ()
  (make-noise :periodic 2 0 160))

(defun swish ()
  (w:with-sound-enabled
    (w:do-sound (w:noise :white 0))
    (do ((level w:on-volume (1+ level)))
	((> level 12.))
      (w:do-sound (w:volume :noise level) 15))
    )
  (values))

(defun ouch (&optional (frequency 1800.) (increment -120.) (interval-time 8.))
  (w:with-sound-enabled
    (dotimes (i 8)
      (w:do-sound (w:tone 0 frequency))
      ;;(w:do-sound (w:tone 1 frequency))
      ;;(w:do-sound (w:tone 2 frequency))
      ;;(w:do-sound (w:volume 1 :off))
      ;;(w:do-sound (w:volume 2 :off))
      (w:do-sound (w:volume 0 (+ i w:on-volume)) interval-time)
      (INCF frequency increment)))
  (values))

) ; end of eval-when

(w:defsound ouch (ouch))

(w:defsound ding (ding))

(w:defsound bell (bell))

(w:defsound buzz (buzz))

(w:defsound swish (swish))