;; -*- Mode: Common-Lisp; Package: TV; Patch-File: T; Base: 8; Fonts: CPTFONT,HL12B,HL12BI -*-


;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985, Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1986, Douglas Johnson. All rights reserved.
#|
This file redefines the function which displays the user-id to display the current 
consing activity instead.  It also displays the lisp mode (common or zeta) and the
current base.  The consing is displayed as the number of words consed in the last
second (or since the last update of the who line if that was longer than a second). 
It will display a negative number occasionally when a temporary area (usually the
extra PDL) is reset.  This code is provided without warranty or support of any kind. 
|#


si:(defun area-length (area)
     (multiple-value-bind (nil used nil)
	 (room-get-area-length-used area)
       used))

(defvar *last-cons-count* 0)
(defvar *who-cons-string* (make-array 10 :type :art-string :leader-length 1))
(setf (fill-pointer *who-cons-string*) 0)
(defvar *last-cons-time* 0)

(DEFUN WHO-LINE-USER-OR-PROCESS
       (WHO-SHEET &AUX CURRENT-LANGUAGE CURRENT-PRINT-BASE SG)
  (DECLARE (:SELF-FLAVOR WHO-LINE-SHEET))
  (WHEN (SETQ LAST-WHO-LINE-PROCESS (OR WHO-LINE-PROCESS
                                        (AND
                                          SELECTED-IO-BUFFER
                                          (IO-BUFFER-LAST-OUTPUT-PROCESS
                                            SELECTED-IO-BUFFER))))
    (when (> (time-difference (time) *last-cons-time*) 60)
      (SETQ SG (PROCESS-STACK-GROUP LAST-WHO-LINE-PROCESS))
      (COND ((EQ SG %CURRENT-STACK-GROUP)
	     (SETQ CURRENT-LANGUAGE   SI:*LISP-MODE*
		   CURRENT-PRINT-BASE *READ-BASE*))
	    ((TYPEP SG ':STACK-GROUP)
	     (SETQ CURRENT-LANGUAGE   (SYMEVAL-IN-STACK-GROUP 'SI:*LISP-MODE* SG)
		   CURRENT-PRINT-BASE (SYMEVAL-IN-STACK-GROUP '*READ-BASE*    SG)))
	    (T (SETQ CURRENT-LANGUAGE   SI:*LISP-MODE*
		     CURRENT-PRINT-BASE *READ-BASE*)))
      (setf (fill-pointer *who-cons-string*) 0)
      (let ((current-consing 0))
	(dolist (area area-list)
	  (incf current-consing (si:area-length
#+elroy				  (symbol-value area)
#-elroy 			  (symeval area)
				  )))
	(decf current-consing (si:area-length extra-pdl-area))
	
	(setf *last-cons-time* (time))
	(format *who-cons-string* "~5d " (let ((cons  (- current-consing
							 *last-cons-count*))) 
					   (setq tv:*last-cons-count* current-consing)
					   cons))
	(string-nconc *who-cons-string* (if (EQ CURRENT-LANGUAGE :ZETALISP) "ZL" "CL"))
	(FORMAT *who-cons-string* " ~d" CURRENT-PRINT-BASE))
  
      (WHO-LINE-STRING WHO-SHEET  *who-cons-string* t))))

(DEFUN WHO-LINE-STRING (WHO-SHEET NEW-STRING &optional force-display)
  "Output NEW-STRING on WHO-SHEET, a part of the who line if it has changed or
   FORCE-DISPLAY is not nil.  The last value is remembered in the
   WHO-LINE-ITEM-STATE instance variable."
  (DECLARE (:SELF-FLAVOR WHO-LINE-SHEET))
  (SETQ NEW-STRING (STRING NEW-STRING))
  (COND ((or force-display (NEQ WHO-LINE-ITEM-STATE NEW-STRING))
	 (PREPARE-SHEET (WHO-SHEET)
	   (SHEET-CLEAR WHO-SHEET)
	   (SHEET-STRING-OUT WHO-SHEET NEW-STRING
			     0 (MIN (ARRAY-ACTIVE-LENGTH NEW-STRING)
				    (TRUNCATE (SHEET-INSIDE-WIDTH WHO-SHEET)
					      (SHEET-CHAR-WIDTH   WHO-SHEET)))))
	 (SETQ WHO-LINE-ITEM-STATE NEW-STRING))))
                                               
