;;; -*- Mode:common-Lisp; Package:System-Internals; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; COPYRIGHT (C) 1984-1989 Texas Instruments INCORPORATED.  ALL RIGHTS RESERVED.

;;; 07-14-88 DAB Modify (SERIAL-STREAM-MIXIN :TYO) to move the second process wait inside the WHEN.

;;; Stream support for the explorer RS232.
;;; This code uses the following system variables:
;;;   *Serial-Port*          - The device descriptor for the serial port
;;;   *Serial-Input-Buffer*  - the input buffer where microcode places incomming characters
;;;   *Serial-Output-Buffer* - the output buffer where the microcode gets characters to go out


(DEFVAR *SERIAL-PORT-OWNER* ()) 		;normally the stream that is currently active on the port

(DEFPARAMETER EOT-CHARACTER #o4)     ;need a good place for these constants !

(DEFPARAMETER XON-CHARACTER #o21) 



(DEFUN MAKE-SERIAL-STREAM (&REST KEYWORDS)
  "Make a stream that talks over the explorer RS232 port.
Keywords are:
:FLAVOR                 -- flavor name to make this stream an instance of
:ASCII-CHARACTERS       -- T enables translation between Ascii and Lisp machine character sets
:XON-XOFF-PROTOCOL      -- T enables XONXOFF (^S^Q) control of output
:EOT-ENABLE             -- T enables special processing of the EOT (04) character (NIL by default)
:INPUT-BUFFER-SIZE      -- default buffer size is 256. words (1 page)
:OUTPUT-BUFFER-SIZE
:FORCE-OUTPUT           -- NIL means you must :FORCE-OUTPUT to cause transmission,
                           T (the default) means an automatic :FORCE-OUTPUT occurs after :TYO
                           It is recommended that :FORCE-OUTPUT be T with the TI855 printer.
:NUMBER-OF-DATA-BITS    -- 5, 7, 6, or 8 (default 8)
:NUMBER-OF-RX-DATA-BITS -- the Z8530 chip allows you to do these seperately
:NUMBER-OF-TX-DATA-BITS --  this is not recommended
:NUMBER-OF-STOP-BITS    -- default 2
:PARITY                 -- init keywords are ':none ':even ':odd (default is ':none)
:BAUD                   -- baud rate (default 4800.)

:DATA-TERMINAL-READY    -- NIL disables DTR output - default is true
:REQUEST-TO-SEND        -- NIL disables RTS output - default is true

:CHECK-FRAMING-ERRORS   -- NIL disables error if break received
:CHECK-OVER-RUN-ERRORS  -- NIL disables error if input characters lost in UART
:CHECK-PARITY-ERRORS    -- NIL disables error if wrong parity

"
  (DECLARE
   (ARGLIST &KEY &OPTIONAL FLAVOR ASCII-CHARACTERS XON-XOFF-PROTOCOL FORCE-OUTPUT EOT-ENABLE
	    NUMBER-OF-DATA-BITS NUMBER-OF-RX-DATA-BITS NUMBER-OF-TX-DATA-BITS
	    NUMBER-OF-STOP-BITS PARITY BAUD DATA-TERMINAL-READY REQUEST-TO-SEND
	    INPUT-BUFFER-SIZE OUTPUT-BUFFER-SIZE CHECK-FRAMING-ERRORS CHECK-OVER-RUN-ERRORS
	    CHECK-PARITY-ERRORS))
  (IF *SERIAL-PORT-OWNER*;somebody already has the port ?
    (PROGN
      (CERROR "Close Existing Stream and Open New Stream"
	      "Attempt to open a Serial Stream on Port already in use by: ~a "
	      *SERIAL-PORT-OWNER*)
      ;; user wants to proceed - close the old stream and start the new one
      (SEND *SERIAL-PORT-OWNER* :CLOSE)))
  (INSTANTIATE-FLAVOR
   (OR (GETF KEYWORDS :FLAVOR);user specified flavor or ...
      (IF (GETF KEYWORDS :ASCII-CHARACTERS)
	'SERIAL-ASCII-STREAM
	'SERIAL-STREAM))
   (LOCF KEYWORDS) T)) 

;;; Flavor definitions
;;; This used to use the basic-buffered-output-stream flavor but it did a 
;;; lot of stuff we don't really need so I took it out. -- JSL


(DEFFLAVOR SERIAL-STREAM-MIXIN
	   ((WR3-CONTENTS #o301); rx enable, 8 bits
	    (WR4-CONTENTS #o114); x16 clock, 2 stop bits, no parity  (4c for 16x 0c for 1x)
	    (WR5-CONTENTS #o352); DTR, RTS, Tx enable, 8 bits
	    (WR12-CONTENTS #o16); low baud constant, 4800. (set for x16 clock)
	    (WR13-CONTENTS 0); high baud constant
	    (FORCE-OUTPUT T)
	    (EOT-ENABLE NIL)
	    (INPUT-BUFFER-SIZE PAGE-SIZE)
	    (OUTPUT-BUFFER-SIZE PAGE-SIZE)
	    (SERIAL-ERROR-MASK 7);error checking on as default
	    (SERIAL-UNTYI-CH NIL)
	    (CLOSED-P T));signal stream is not open for I/O
	   (LINE-OUTPUT-STREAM-MIXIN)
  (:REQUIRED-FLAVORS BIDIRECTIONAL-STREAM CHARACTER-STREAM UNBUFFERED-LINE-INPUT-STREAM)
  (:INITABLE-INSTANCE-VARIABLES SERIAL-UNTYI-CH FORCE-OUTPUT EOT-ENABLE INPUT-BUFFER-SIZE
   OUTPUT-BUFFER-SIZE WR3-CONTENTS WR4-CONTENTS WR5-CONTENTS WR12-CONTENTS WR13-CONTENTS)
  (:INIT-KEYWORDS :XON-XOFF-PROTOCOL :ASCII-CHARACTERS :FORCE-OUTPUT :EOT-ENABLE
   :NUMBER-OF-DATA-BITS :NUMBER-OF-TX-DATA-BITS :NUMBER-OF-RX-DATA-BITS :NUMBER-OF-STOP-BITS
   :PARITY :BAUD :REQUEST-TO-SEND :DATA-TERMINAL-READY :RECEIVE-ENABLE :TRANSMIT-ENABLE
   :INPUT-BUFFER-SIZE :OUTPUT-BUFFER-SIZE :CHECK-FRAMING-ERRORS :CHECK-OVER-RUN-ERRORS
   :CHECK-PARITY-ERRORS :FLAVOR)
  (:DOCUMENTATION :MIXIN "The guts of RS232 Serial I/O Streams")) 


(DEFFLAVOR SERIAL-STREAM
	   ()
	   (SERIAL-STREAM-MIXIN
	    BIDIRECTIONAL-STREAM
	    CHARACTER-STREAM
	    UNBUFFERED-LINE-INPUT-STREAM)
  (:DOCUMENTATION :COMBINATION "RS232 Serial I/O Stream, no character-set translation")) 


(DEFFLAVOR SERIAL-CHARACTER-STREAM
	   ()
	   (SERIAL-ASCII-STREAM)) 	;compatibility


(DEFFLAVOR SERIAL-ASCII-STREAM
	   ()
	   (ASCII-TRANSLATING-OUTPUT-STREAM-MIXIN
	    ASCII-TRANSLATING-INPUT-STREAM-MIXIN
	    SERIAL-STREAM-MIXIN
	    BIDIRECTIONAL-STREAM
	    CHARACTER-STREAM
	    UNBUFFERED-LINE-INPUT-STREAM)
  (:DOCUMENTATION :COMBINATION "RS232 Serial I/O Stream, Ascii character translation")) 



;;; Housekeeping methods


(DEFMETHOD (SERIAL-STREAM-MIXIN :INIT) (INIT-PLIST)
  ;; first make sure the *serial-port* structure has been set up
  ;; this should have been part of the cold initialization list
  ;; NOTE: this must only be done once!
  (IF (EQ *SERIAL-PORT* ())
      (INITIALIZE-SERIAL-INTERRUPT))
  (IF (AND *SERIAL-PORT-OWNER*			;somebody already has the port ?
	   (Not (eq *SERIAL-PORT-OWNER* SELF)))	;and its not us
      (PROGN
	(CERROR "Close Existing Stream and Initialize New Stream"
		"Attempt to Initialize Serial Stream: ~a on Serial Port already in use by: ~a "
		SELF *SERIAL-PORT-OWNER*)
	;; user wants to proceed - close the old stream and start the new one
	(SEND *SERIAL-PORT-OWNER* :CLOSE)))
  (SEND SELF :RESET-HARDWARE)			;reset Z8530 using current WR values
  (CLEAR-SERIAL-PORT-STATUS-BITS)		;clear device descriptor status bits
  (SERIAL-PORT-DISABLE-XON-XOFF)		;disable xon/xoff as default condition
  ;; now set chip parameters from the keyword list 
  (LOOP FOR (KWD VAL) ON (CDR INIT-PLIST) BY 'CDDR UNLESS
	(MEMBER KWD '(:FLAVOR :ASCII-CHARACTERS) :TEST #'EQ) DO
	(ERRSET (SEND SELF :PUT KWD VAL) ()))
  (SEND SELF :RESET)) 


(DEFMETHOD (SERIAL-STREAM-MIXIN :RESET) (&OPTIONAL (ABORT-OUTPUT ':ABORT))
  ;; can only reset the current holder of the port
  (IF (AND *SERIAL-PORT-OWNER*			;somebody already has the port ?
	   (Not (EQ *SERIAL-PORT-OWNER* SELF)))	;and its not us
      (PROGN
	(CERROR "Reset (Open) Stream ~a and Close Stream ~a "
		"Attempt to reset Serial Stream ~a on Port already in use by: ~a " SELF
		*SERIAL-PORT-OWNER*)
	;; user wants to proceed - close the old stream and start the new one
	(SEND *SERIAL-PORT-OWNER* :CLOSE)))
  (SEND SELF :CLOSE ABORT-OUTPUT)		;Forget about any remaining output characters
  (SEND SELF :RESET-HARDWARE)			;reset the serial controller chip
  (CLEAR-SERIAL-PORT-STATUS-BITS)		;clear serial device descriptor status bits
  ;; allocate the serial input & output buffers
  (SETQ *SERIAL-INPUT-BUFFER* (GET-SERIAL-BUFFER INPUT-BUFFER-SIZE))
  (SETQ *SERIAL-OUTPUT-BUFFER* (GET-SERIAL-BUFFER OUTPUT-BUFFER-SIZE))
  (ARRAY-DPB-OFFSET
    (+ (%POINTER *SERIAL-INPUT-BUFFER*)
       (1+ (%P-LDB %%ARRAY-LONG-LENGTH-FLAG *SERIAL-INPUT-BUFFER*)))
    %%Q-POINTER *SERIAL-PORT* %SERIAL-RECEIVE-BUFFER)
  (ARRAY-DPB-OFFSET
    (+ (%POINTER *SERIAL-OUTPUT-BUFFER*)
       (1+ (%P-LDB %%ARRAY-LONG-LENGTH-FLAG *SERIAL-OUTPUT-BUFFER*)))
    %%Q-POINTER *SERIAL-PORT* %SERIAL-TRANSMIT-BUFFER)
  (SERIAL-PORT-SET-RECEIVE-BITS-PER-CHARACTER
    (NTH (LDB (BYTE 2 6) WR3-CONTENTS) '(31 127 63 255)))	;reset bits-per-char
  (WRITE-Z-REG #o11 #o10)			;enable serial interrupts (set master interrupt enable)
  (ENABLE-SERIAL-EVENT)				;enable SIB serial event posting
  (SETQ CLOSED-P ()				;stream is open now
	
	*SERIAL-PORT-OWNER* SELF)) 


(DEFMETHOD (SERIAL-STREAM-MIXIN :RESET-HARDWARE) ()	;stage 1 - modes and constants
  (WRITE-Z-REG #o11 #o300)			;full reset
  (WRITE-Z-REG 4 WR4-CONTENTS)		;stop bits, parity
  (WRITE-Z-REG 2 0)				;no interrupt vector
  (WRITE-Z-REG 3 (DPB 0 (BYTE 1 0) WR3-CONTENTS))	;Rx disabled
  (WRITE-Z-REG 5 (DPB 0 (BYTE 1 3) WR5-CONTENTS))	;Tx disabled
  (WRITE-Z-REG 6 0)				;no sync chars
  (WRITE-Z-REG 7 0)				;no sync chars
  (WRITE-Z-REG #o11 0)				;no interrupts enabled
  (WRITE-Z-REG #o12 0)				;no misc control
  (WRITE-Z-REG #o13 #o122)			;clock with baud rate generator
  (WRITE-Z-REG #o14 WR12-CONTENTS)		;baud rate low
  (WRITE-Z-REG #o15 WR13-CONTENTS)		;baud rate high
  (WRITE-Z-REG #o16 #o142)			;BR source select, disable PLL
						;stage 2 - enables
  (WRITE-Z-REG #o16 #o143)			;BR generator enable
  (WRITE-Z-REG 3 (DPB 1 (BYTE 1 0) WR3-CONTENTS))	;Rx enabled
  (WRITE-Z-REG 5 (DPB 1 (BYTE 1 3) WR5-CONTENTS))	;Tx enabled
  (WRITE-Z-REG 0 #o200)				;reset TxCRC
						;stage 3 - interrupt enables
  (WRITE-Z-REG #o17 0)				;disable external interrupts
  (WRITE-Z-REG 0 #o20)				;reset EXT/STATUS
  (WRITE-Z-REG 0 #o20)				;reset EXT/STATUS - should do this twice
  (WRITE-Z-REG 1 #o22)			;Tx, Rx interrupts enabled
  ) 



(DEFMETHOD (SERIAL-STREAM-MIXIN :CLOSE) (&OPTIONAL ABORT-P)
  ;; deallocate the serial buffers
  (WHEN (EQUAL SELF *SERIAL-PORT-OWNER*)	;can only close the real owner of the port
    (IF (NOT ABORT-P)
	(SEND SELF :FINISH))
    (WRITE-Z-REG #o11 0)			;disable serial interrupts (clear master interrupt enable)
    (DISABLE-SERIAL-EVENT)			;disable SIB serial event posting
    (WHEN *SERIAL-OUTPUT-BUFFER*
      (RETURN-SERIAL-BUFFER *SERIAL-OUTPUT-BUFFER*)
      (SETQ *SERIAL-OUTPUT-BUFFER* ()))
    (WHEN *SERIAL-INPUT-BUFFER*
      (RETURN-SERIAL-BUFFER *SERIAL-INPUT-BUFFER*)
      (SETQ *SERIAL-INPUT-BUFFER* ()))
    (CLEAR-RESOURCE 'SERIAL-BUFFER () ())	;2.1 fix ,see if this stabilizes the system
    (ARRAY-DPB-OFFSET 0 %%Q-POINTER *SERIAL-PORT* %SERIAL-RECEIVE-BUFFER)
    (ARRAY-DPB-OFFSET 0 %%Q-POINTER *SERIAL-PORT* %SERIAL-TRANSMIT-BUFFER)
    (SETQ CLOSED-P T				;stream is closed -- any I/O method use could crash the system
	  
	  *SERIAL-PORT-OWNER* ()))) 



(DEFMETHOD (SERIAL-STREAM-MIXIN :FINISH) ()
 ;; get things started if they aren't already
  (SEND SELF :FORCE-OUTPUT)
  (PROCESS-WAIT "Serial Finish" #'(LAMBDA (BUFFER)
				    (NOT (si:SERIAL-BUFFER-NOT-EMPTY-P BUFFER)))
		si:*SERIAL-OUTPUT-BUFFER*)) 



(DEFMETHOD (SERIAL-STREAM-MIXIN :CLEAR-INPUT) ()
  (IF (NULL CLOSED-P)
    (FLUSH-SERIAL-BUFFER *SERIAL-INPUT-BUFFER*))
  (SETQ SERIAL-UNTYI-CH ())) 


(DEFMETHOD (SERIAL-STREAM-MIXIN :CLEAR-OUTPUT) ()
  (IF (NULL CLOSED-P)
    (FLUSH-SERIAL-BUFFER *SERIAL-OUTPUT-BUFFER*))) 



(DEFMETHOD (SERIAL-STREAM-MIXIN :STATUS) ()
  "Print out the status of the serial i/o interface."
  (FORMAT T "~%Current read registers.")
  (FORMAT T "~%RR0  = ~16R" (SERIAL-A-STATUS))
  (FORMAT T "~%RR1  = ~16R" (READ-A-REG 1))
  (FORMAT T "~%RR2  = ~16R" (READ-B-REG 2))
  (FORMAT T "~%RR3  = ~16R" (READ-A-REG 3))
  (FORMAT T "~%RR10 = ~16R" (READ-A-REG #o12))
  (FORMAT T "~%RR12 = ~16R" (READ-A-REG #o14))
  (FORMAT T "~%RR13 = ~16R" (READ-A-REG #o15))
  (FORMAT T "~%RR15 = ~16R" (READ-A-REG #o17))
  (FORMAT T "~%Current write registers.")
  (FORMAT T "~%WR3  = ~16R" WR3-CONTENTS)
  (FORMAT T "~%WR4  = ~16R" WR4-CONTENTS)
  (FORMAT T "~%WR5  = ~16R" WR5-CONTENTS)
  (FORMAT T "~%WR12 = ~16R" WR12-CONTENTS)
  (FORMAT T "~%WR13 = ~16R" WR13-CONTENTS)
  (FORMAT T "~%Stream Status:  ~a " (IF CLOSED-P
				      :CLOSED
				      :OPEN))) 
	
;;; NOTE: the following table was taken from the SIB spec 
;;;       IT IS INCORRECT
;;;       The clock seems to be running twice as fast as documented
;;;       so the baud rate constants must half as fast as documented.

;;;(defvar serial-baud-table '((#x+2FFE . 50.) (#x+1FFE . 75.) (#x+11E7 . 110.)
;;;			    (#x+11D6 . 134.5) (#x+0FFE . 150.) (#x+0BFE . 200.)
;;;			    (#x+07FE . 300.) (#x+03FE . 600.) (#x+01FE . 1200.)
;;;			    (#x+0153 . 1800.) (#x+00FE . 2400.) (#x+00A9 . 3600.)
;;;			    (#x+007E . 4800.) (#x+0053 . 7200.) (#x+003E . 9600.)
;;;			    (#x+001E . 19200.)))

;;; ANOTHER NOTE: the chip won't run in the normal 1x clock mode so we must
;;; program it for 16x in register 4.  This means that the baud rate constants
;;; must be modified again.

#|
;;; 1x clock table

(DEFVAR SERIAL-BAUD-TABLE
   '((4094 . 300) (1022 . 1200) (510 . 2400) (254 . 4800) (126 . 9600) (62 . 19200)))   |#
  
 ;;16x clock table

(DEFVAR SERIAL-BAUD-TABLE
   '((254 . 300) (62 . 1200) (30 . 2400) (14 . 4800) (6 . 9600) (2 . 19200)))   
#|
;;32x clock table

(DEFVAR SERIAL-BAUD-TABLE
   '((126 . 300) (30 . 1200) (14 . 2400) (6 . 4800) (2 . 9600) (0 . 19200)))   |#

;;; Get the current RS232 parameters

(DEFMETHOD (SERIAL-STREAM-MIXIN :GET) (PROP)
  (LET ()
    (CASE PROP
      (:REQUEST-TO-SEND (LOGTEST 2 WR5-CONTENTS))
      (:DATA-TERMINAL-READY (LOGTEST #o200 WR5-CONTENTS))
      (:DATA-CARRIER-DETECT (LOGTEST #o10 (SERIAL-A-STATUS)))
      (:CLEAR-TO-SEND (LOGTEST #o40 (SERIAL-A-STATUS)))
      (:TX-BUFFER-EMPTY (LOGTEST #o4 (SERIAL-A-STATUS)))
      (:RX-CHAR-AVAILABLE (LOGTEST 1 (SERIAL-A-STATUS)))
      (:NUMBER-OF-STOP-BITS (NTH (LDB (BYTE 2 2) WR4-CONTENTS) '(0 1 1.5 2)))
      (:PARITY (NTH (LDB (BYTE 2 0) WR4-CONTENTS) '(:NONE :ODD NIL :EVEN)))
      (:NUMBER-OF-DATA-BITS
       (LET ((RX (NTH (LDB (BYTE 2 6) WR3-CONTENTS) '(5 7 6 8)))
	     (TX (NTH (LDB (BYTE 2 5) WR5-CONTENTS) '(5 7 6 8))))
	 (VALUES RX TX)))
      (:NUMBER-OF-RX-DATA-BITS (NTH (LDB (BYTE 2 6) WR3-CONTENTS) '(5 7 6 8)))
      (:NUMBER-OF-TX-DATA-BITS (NTH (LDB (BYTE 2 5) WR5-CONTENTS) '(5 7 6 8)))
      (:BAUD
       (CDR (ASSOC (DPB WR13-CONTENTS (BYTE 8 8) WR12-CONTENTS) SERIAL-BAUD-TABLE :TEST #'EQ)))
      (:RECEIVE-ENABLE (LOGTEST 1 WR3-CONTENTS))
      (:TRANSMIT-ENABLE (LOGTEST #o10 WR5-CONTENTS))
      (:WR3 WR3-CONTENTS)
      (:WR4 WR4-CONTENTS)
      (:WR5 WR5-CONTENTS)
      (:WR12 WR12-CONTENTS)
      (:WR13 WR13-CONTENTS)
      (:INPUT-BUFFER-SIZE INPUT-BUFFER-SIZE)
      (:OUTPUT-BUFFER-SIZE OUTPUT-BUFFER-SIZE)
      (:FORCE-OUTPUT FORCE-OUTPUT)
      (:EOT-ENABLE EOT-ENABLE)
      (:XON-XOFF-PROTOCOL (SERIAL-PORT-XON-XOFF-ENABLE-P))
      (:OTHERWISE (FERROR () "~S not a valid property name" PROP))))) 

;;; Modify RS232 parameters

(DEFMETHOD (SERIAL-STREAM-MIXIN :PUT) (PROP VAL)
  "Change various properties of the stream.  The serial stream should be :reset for certain new values to 
take effect."
  (CASE PROP
    (:NUMBER-OF-DATA-BITS
     (SETQ VAL
	   (IF (AND (> VAL 4) (< VAL 9))
	     (NTH (- VAL 5) '(0 2 1 3))
	     (FERROR () ":NUMBER-OF-DATA-BITS is ~D, not 5, 6, 7, or 8." VAL)))
     (SETQ WR3-CONTENTS (DPB VAL (BYTE 2 6) WR3-CONTENTS))
     (WRITE-Z-REG 3 WR3-CONTENTS)
     ;; set number of receiver data bits into %serial-information word for xon-xoff microcode
     (SERIAL-PORT-SET-RECEIVE-BITS-PER-CHARACTER
      (NTH (LDB (BYTE 2 6) WR3-CONTENTS) '(31 127 63 255)))
     (SETQ WR5-CONTENTS (DPB VAL (BYTE 2 5) WR5-CONTENTS)) (WRITE-Z-REG 5 WR5-CONTENTS))
    (:NUMBER-OF-RX-DATA-BITS
     (SETQ VAL
	   (IF (AND (> VAL 4) (< VAL 9))
	     (NTH (- VAL 5) '(0 2 1 3))
	     (FERROR () ":NUMBER-OF-RX-DATA-BITS is ~D, not 5, 6, 7, or 8." VAL)))
     (SETQ WR3-CONTENTS (DPB VAL (BYTE 2 6) WR3-CONTENTS)) (WRITE-Z-REG 3 WR3-CONTENTS)
							   ;; set number of receiver data bits into %serial-information word for xon-xoff microcode
     (SERIAL-PORT-SET-RECEIVE-BITS-PER-CHARACTER
      (NTH (LDB (BYTE 2 6) WR3-CONTENTS) '(31 127 63 255))))
    (:NUMBER-OF-TX-DATA-BITS
     (SETQ VAL
	   (IF (AND (> VAL 4) (< VAL 9))
	     (NTH (- VAL 5) '(0 2 1 3))
	     (FERROR () ":NUMBER-OF-TX-DATA-BITS is ~D, not 5, 6, 7, or 8." VAL)))
     (SETQ WR5-CONTENTS (DPB VAL (BYTE 2 5) WR5-CONTENTS)) (WRITE-Z-REG 5 WR5-CONTENTS))
    (:PARITY
     (SETQ VAL
	   (CASE VAL
	     (NIL 0)
	     (:NONE 0)
	     (:ODD 1)
	     (:EVEN 3)
	     (OTHERWISE (FERROR () ":PARITY is ~S, not NIL, :ODD, or :EVEN." VAL))))
     (SETQ WR4-CONTENTS (DPB VAL (BYTE 2 0) WR4-CONTENTS)) (WRITE-Z-REG 4 WR4-CONTENTS))
    (:NUMBER-OF-STOP-BITS
     (SETQ VAL
	   (case VAL (1 1) (1.5 2) (2 3)
	      (OTHERWISE (FERROR () ":NUMBER-OF-STOP-BITS is ~S, not 1, 1.5, or 2." VAL))))
     (SETQ WR4-CONTENTS (DPB VAL (BYTE 2 2) WR4-CONTENTS)) (WRITE-Z-REG 4 WR4-CONTENTS))
    (:BAUD (SETQ VAL (CAR (RASSOC VAL SERIAL-BAUD-TABLE :TEST #'EQ)))
     (IF (NOT VAL)
       (FERROR () "Invalid serial stream baud rate, ~D." VAL))
     (SETQ WR12-CONTENTS (LDB (BYTE 8 0) VAL)) (SETQ WR13-CONTENTS (LDB (BYTE 8 8) VAL))
     (WRITE-Z-REG #o14 WR12-CONTENTS) (WRITE-Z-REG #o15 WR13-CONTENTS))
    (:RECEIVE-ENABLE (SETQ WR3-CONTENTS (DPB (IF VAL
					       1
					       0)
					     (BYTE 1 0) WR3-CONTENTS))
     (WRITE-Z-REG 3 WR3-CONTENTS))
    (:TRANSMIT-ENABLE (SETQ WR5-CONTENTS (DPB (IF VAL
						1
						0)
					      (BYTE 1 3) WR5-CONTENTS))
     (WRITE-Z-REG 5 WR5-CONTENTS))
    (:REQUEST-TO-SEND (SETQ WR5-CONTENTS (DPB (IF VAL
						1
						0)
					      (BYTE 1 1) WR5-CONTENTS))
     (WRITE-Z-REG 5 WR5-CONTENTS))
    (:DATA-TERMINAL-READY (SETQ WR5-CONTENTS (DPB (IF VAL
						    1
						    0)
						  (BYTE 1 7) WR5-CONTENTS))
     (WRITE-Z-REG 5 WR5-CONTENTS))
    (:INPUT-BUFFER-SIZE (SETQ INPUT-BUFFER-SIZE VAL))
    (:OUTPUT-BUFFER-SIZE (SETQ OUTPUT-BUFFER-SIZE VAL))
    (:FORCE-OUTPUT (SETQ FORCE-OUTPUT (IF VAL
					T
					())))
    (:EOT-ENABLE (SETQ EOT-ENABLE (IF VAL
				    T
				    ())))
    (:XON-XOFF-PROTOCOL (IF VAL
			  (SERIAL-PORT-ENABLE-XON-XOFF)
			  (SERIAL-PORT-DISABLE-XON-XOFF)))
    (:CHECK-PARITY-ERRORS
     (SETQ SERIAL-ERROR-MASK (DPB (IF VAL
				    1
				    0)
				  (BYTE 1 0) SERIAL-ERROR-MASK)))
    (:CHECK-OVER-RUN-ERRORS
     (SETQ SERIAL-ERROR-MASK (DPB (IF VAL
				    1
				    0)
				  (BYTE 1 1) SERIAL-ERROR-MASK)))
    (:CHECK-FRAMING-ERRORS
     (SETQ SERIAL-ERROR-MASK (DPB (IF VAL
				    1
				    0)
				  (BYTE 1 2) SERIAL-ERROR-MASK)))
    (:OTHERWISE (FERROR () "~S not a valid property name" PROP)))) 



;;; Stream Input Protocol

;;; This method looks to see if the buffer has any characters in it.  If xon-xoff-protocol is enabled,
;;; we may have to issue an xon to start filling the buffer up again.


(DEFMETHOD (SERIAL-STREAM-MIXIN :INPUT-CHARS-AVAILABLE-P) ()
  (COND
    (CLOSED-P NIL);don't do anything & return nil
    (SERIAL-UNTYI-CH T)
    (T
     (IF (AND (SERIAL-PORT-XON-XOFF-ENABLE-P);d1/d3 enabled ?
	 (SERIAL-PORT-XOFF-SENT-P);xoff sent ?
	 (NOT (SERIAL-BUFFER-NOT-EMPTY-P *SERIAL-INPUT-BUFFER*)));empty buffer ?
       (PROGN
	 (%IO 1 *SERIAL-PORT*);force out XON special I/O op.
	 (SERIAL-PORT-CLEAR-XOFF-SENT);clear xoff sent flag
))
     ;; now return the results of this operation
     (SERIAL-BUFFER-NOT-EMPTY-P *SERIAL-INPUT-BUFFER*)))) 



(DEFMETHOD (SERIAL-STREAM-MIXIN :TYI) (&OPTIONAL IGNORE)
  (IF CLOSED-P
    (PROGN
      (CERROR "Reset (Open) Serial Stream ~a and Retry"
	      "Attempt to initiate IO on closed Serial Stream: ~a " SELF)
      (SEND SELF :RESET)))
  ;; stream should be open at this point
  (COND
    ((NULL SERIAL-UNTYI-CH)
     (IF (AND (NOT (SEND SELF :INPUT-CHARS-AVAILABLE-P)) (NOT (SERIAL-PORT-EOT-RECEIVED-P)))
       (PROCESS-WAIT "Serial In" #'(LAMBDA (BUFFER)
				     (si:SERIAL-BUFFER-NOT-EMPTY-P BUFFER))
		     si:*SERIAL-INPUT-BUFFER*))
     (MULTIPLE-VALUE-BIND (CH STATUS)
       (READ-SERIAL-BUFFER *SERIAL-INPUT-BUFFER*)
       (IF (LOGTEST SERIAL-ERROR-MASK STATUS)
	 (SETQ CH
	       (CERROR T () () "Serial input ~:[framing ~]~:[over-run ~]~:[parity ~]error: ~O"
		       (ZEROP (LOGAND 4 STATUS SERIAL-ERROR-MASK))
		       (ZEROP (LOGAND 2 STATUS SERIAL-ERROR-MASK))
		       (ZEROP (LOGAND 1 STATUS SERIAL-ERROR-MASK)) CH)))
       (IF EOT-ENABLE
	 (COND
	   ((EQ CH EOT-CHARACTER) (SERIAL-PORT-SET-EOT) NIL);eot in data returns nil when enabled
	   (T (IF CH
		(SERIAL-PORT-CLEAR-EOT))
	    CH));return character if not eot
	 CH)));return character if eot not enabled
    (T (PROG1
	 SERIAL-UNTYI-CH
	 (SETQ SERIAL-UNTYI-CH ()))))) 



(DEFMETHOD (SERIAL-STREAM-MIXIN :UNTYI) (CHAR)
  (SETQ SERIAL-UNTYI-CH CHAR)) 


(DEFMETHOD (SERIAL-STREAM-MIXIN :LISTEN) ()
  (OR (NOT (NULL SERIAL-UNTYI-CH)) (SEND SELF :INPUT-CHARS-AVAILABLE-P))) 


(DEFMETHOD (SERIAL-STREAM-MIXIN :TYI-NO-HANG) (&OPTIONAL IGNORE)
  (AND (SEND SELF :INPUT-CHARS-AVAILABLE-P) (SEND SELF :TYI))) 


;;; Serial stream output protocoll
;;; If the buffer is full, we want to do an automatic :FORCE-OUTPUT


(DEFMETHOD (SERIAL-STREAM-MIXIN :TYO) (CH)
  (IF CLOSED-P
    (PROGN
      (CERROR "Reset Stream ~a and Retry" "Attempt to Type Out on closed Serial Stream: ~a "
	      SELF)
      (SEND SELF :RESET)))
  (IF (SERIAL-PORT-XON-XOFF-ENABLE-P)
    (PROCESS-WAIT "Serial Output Hold" #'(LAMBDA ()
					   (NOT (si:SERIAL-PORT-XOFF-RECEIVED-P)))))
  (when (NOT (SERIAL-BUFFER-NOT-FULL-P *SERIAL-OUTPUT-BUFFER*))
    (SEND SELF :FORCE-OUTPUT)
    (PROCESS-WAIT "Serial Out" #'SERIAL-BUFFER-NOT-FULL-P *SERIAL-OUTPUT-BUFFER*)) ;07-14-88 DAB
  (WRITE-SERIAL-BUFFER *SERIAL-OUTPUT-BUFFER* CH)
  (IF FORCE-OUTPUT
    (SEND SELF :FORCE-OUTPUT))) 


;;; The UART may be in a dormant state (i.e. not generating interrupts).
;;; To start a series of transmit interrupts, we must first force one
;;; character out the port.  This is done by calling a special microcode 
;;; io function that will force the next character in the output buffer
;;; through the serial port.  It must be microcoded so the SIB can't
;;; interrupt us with a real serial interrupt while we're doing it.
;;; Yes Robin, some would call this a kludge . . . 


(DEFMETHOD (SERIAL-STREAM-MIXIN :FORCE-OUTPUT) ()
  (IF (SERIAL-BUFFER-NOT-EMPTY-P *SERIAL-OUTPUT-BUFFER*)
    (%IO 0 *SERIAL-PORT*))) 

  


(COMPILE-FLAVOR-METHODS SERIAL-STREAM SERIAL-ASCII-STREAM) 



