;;; -*- Mode: common-LISP; Package: MT; Base: 10; Fonts: COURIER,HL12B,HL12BI -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; Doc strings and/or items of command menu lists should be consolidated for Rel 3.2.
;; 
;;
;;  REVISION:
;;;     10-10-88 DAB    Changed Default-directory from Lm:~;*.* to GUESS-directory to make defaults come
;;;                     out properly on microExplorer.
;;;     10-10-88 DAB    Fixes guess-directory. Was not defaulting the proper pathname for a microExplorer.
;;;     10-06-88 DAB    Fixed remote-tape's load,unload and logout functions. Also added the defparameter
;;;                     *backup-frame-font* to control the current-font of the backup-text-pane. It is
;;;                     defaulted to CPTFONT for the Explorer and Halvetica 10 for the microExplorer. Changed
;;;                     (tape-mixin :help) to turn on more processing to help screen wrap problems.
;;;     10-04-88 DAB    Change *command-menu-font* and *label-fonts* for the microExplorer.
;;;     08-15-88 DAB    Bury frame in command-loop when #/END was pressed.
;;;     06-07-88 DAB    Added some changes to (BACKUP-FRAME :BEFORE :INIT) to display remote menu on
;;;                     non-Explorers.
;;;     04-25-88 DAB      Changed :make-tar-tape to accept the keyword :unspecifies-type-action to handle 
;;;                        unknown file types.
;;;     04-07-88 DAB    :verify-partition should call execute-verify-partition if remote.
;;;     02.09.88 LAS    Changed add-to-system-menu  call to modify-system-access-spec
;;;     10.30.87 PMH    Colorize the Backup WINDOW in an :AFTER :INIT method
;;;	08.11.87 MBC	Add TAR interface.  Re-structured :Prepare-Tape.
;;;	07.16.87 MBC	Remove duplicate method, (BACKUP-FRAME :WHO-LINE-DOCUMENTATION-STRING).
;;;	06.19.87 MBC 	Correct verbage for backup-directory mouse line.  INDENT file.
;1;;        03/18/87 HW      Changed call of prepare-tape to prepare-tape-drive.*
;;;	03.02.87 MBC	Remove global variable, *BACKUP-WINDOW* used for telling CVVs
;;;			who :SUPERIOR is.  Now use local declared special.  Avoids
;;;			multiple backup window conflicts and pointers at leftovers.
;;;     03-07-88 DAB    Fixed  03-07-88 changed = to eq  in tape-mixin :backup-partition.
;;;
;;;
;;;
;;;     Add documentation and commands for making, restoring and listing Carry tapes.
;;;       1-9-87 MRR
;;;     Changed a doc string for 2K page size. 1-8-87, -ab
;;;     Change BACKUP-FILE, RESTORE-FILE VERIFY-FILE and VERIFY-DIRECTORY to provide
;;;       specific default for pathname defaulting and use :directory-pathname in
;;;       their CVV windows.  5-2-86 JEB
;;;     Reword question and help line for incremental directory backup.  4-30-86 JEB
;;;     Removed "Click right to edit this field." from comments as it didn't apply to
;;;       the left handed mouse.  4-30-86 JEB
;;;     Force CPTFONT using (BACKUP-FRAME :after :init) for the text-pane. 4-21-86
;;;     Add Use-Original-Host feature to support logical pathnames on Verifys.  4-24-86
;;;           ALSO ----> SINCE TIME fix is  :STRING-OR-NIL -->NEEDS A PATCH 4-23-86 MBC
;;;     Also init more-p to nil, keeps more off when window is first instantiated. 4-23
;;;     Add Use-Original-Host feature to support logical pathnames on restores.  4-10-86
;;;	Fix :backup-directory :method to correctly handle :sub-directories keyword. 3-24-86
;;;	Add :create-directories keyword to :restore-directory :method.
;;;	 	Includes new var, create-directories, and new doc string,
;;;		doc-rd-create-directories. 3-24-86
;;;	Fix up doc strings for partition operations to not mention "name of host".
;;;		Source fix only.  12-11-85 MBC
;;;	Add SELECTIVE-SOURCE-PATTERN option to make restore and verify able
;;;	  to handle subdirectories and various wildcard situations.  10-21-85 MBC 
;;;	Add :DIRECTORY-PATHNAME keyword to CVV window calls for "multiple" functions.
;;;       i.e. B-M-F, R-M-F, & V-M-F  9-15-85 MBC
;;;	Allow Top-Level to catch EOF and END-OF-RECORDED-MEDIA.  Also made
;;;	  default names in backup window the same for all directory commands
;;; 	  and the same for all file commands.  12-3-85 MBC
;;;	Made all CVVs that prompt for pathnames 80 wide instead of 50.
;;;		(Source change only; not in any patch.)  12-9-85 MBC
;;;     Changed :prepare-tape method so that the user is prompted for block-mode
;;;     (:fixed or :variable) and density (single or double) and block-size. 2-23-87 mp
;;;     Partition name can be entered either as lower case or upper case.      2-23-87 mp
;;;     A few help messages have been added.    2-23-87 mp
;;;     The following functions: mt:backup-partition-half-inch-tape
;;;                              mt:restore-partition-half-inch-tape
;;;                              mt:verify-partition-half-inch-tape
;;;     are being called for 1/2" tape operations. 2/23/87 mp
;;;     Put back create-directories option for Restore Directory command. 2-23-87 mp
;;
;;   ORIGINAL:  Backup system window - 10/18/84 MBC
;;
;;;


(DEFPARAMETER *COMMAND-MENU-FONT* (if (si:addin-p)  FONTS:HL7 FONTS:HL12BI)  ;10-04-88 DAB added font for mac
  "Used for the standard entries in the command menu.")


(DEFPARAMETER *LABEL-FONT* (if (si:addin-p) FONTS:TR8I FONTS:TR12I) ;10-04-88 DAB added font for mac
  "Used for labels for all of the panes.")

(DEFPARAMETER *backup-frame-FONT* (if (si:addin-p) FONTS:HL10 FONTS:CPTFONT) ;10-06-88 DAB added font for mac
  "Used for all printouts in the backup text pane, including the HELP menus.")


(defvar *local-host-machine-type* (if (eq (send si:local-host  :system-type) :lispm)
				      :explorer :microexplorer))   ;08-11-88 DAB

(DEFPARAMETER *LANDSCAPE-MONITOR-P*
	      (MULTIPLE-VALUE-BIND (WIDTH HEIGHT)
		  (SEND TV:MAIN-SCREEN :SIZE)
		(> WIDTH HEIGHT))
  "2Boolean which indicates the orientation of the current console display.
T is landscape and NIL is Portrait.*")

(defconstant *partition-num-list* '("0" "1" "2" "3" "4" "5" "6" "7" "8" "9")
  "The number x for a partition name corresponds to lodx")

(defvar block-size nil)

(DEFVAR DOIT ()) 
(defparameter quarter-inch-command-menu-list 
; Miscellaneous operations
	      '(("Prepare Tape" :VALUE :prepare-tape :DOCUMENTATION "Prepare the tape device by looking for an online tape and making it the current tape unit.  Selection of tape format (density) is required.")
		("1Prepare* 1Remote Tape*" :VALUE 1:prepare-remote-tape* :DOCUMENTATION "Prepare 1a remote* tape device by 1specifing* 1a remote host name*.  Selection of tape format (density) is required.") ;1; Kenny Tung for remote*
		("Unload Tape" :VALUE :unload :DOCUMENTATION "Unload the tape cartridge.  This leaves the tape offline until a Load Tape is performed.")
		("Load Tape" :VALUE :load-tape :DOCUMENTATION "Load the tape cartridge, making it available and leaving it rewound to its physical beginning.  This resets the tape formatter and drive.")
		("" :no-select nil)
		("Erase Entire Tape" :VALUE :erase :DOCUMENTATION "Erase the entire tape, after confirmation.")
		("Re-Tension" :VALUE :tension :DOCUMENTATION "Tension tape by unwinding and rewinding.")
		("" :no-select nil)
		("Rewind" :VALUE :rewind :DOCUMENTATION
		 "Position the tape to its physical beginning, ready to read or write.")
		("Prepare to Append" :VALUE :position-to-append :DOCUMENTATION
		 "Prepare to append to tape by positioning the tape past the end of recorded media (EORM), after the last thing recorded on it.")
		("Position Past File (EOF)" :VALUE :space-to-eof :DOCUMENTATION 
		 "Position the tape past the next end of file (moves tape past next file) or multiple files or partitions.")
		("Position Past Blocks" :VALUE :space-blocks :DOCUMENTATION 
		 "Position the tape forward a number of blocks.  Generally, this is not recommended.")
		("" :no-select nil)	   
		("List Contents" :VALUE :list-contents :DOCUMENTATION
		 "Read the rest of the entire tape, displaying all file and partition headers until the end of recorded media is encountered.")
		("List Directory" :VALUE :list-directory :DOCUMENTATION
		 "Read the next directory, displaying headers from tape until a double EOF is encountered.  Files and partitions backed up without an extra EOF
will be treated as a directory.  If there are sub-directories written by a single Backup Directory command, they will also be listed.")
; Data movement operations
   ("" :no-select nil)
   ("Backup File" :VALUE :backup-file :DOCUMENTATION "Copy a file to tape.  Wildcarding is not allowed.")
   ("Backup Directory" :VALUE :backup-directory :DOCUMENTATION
    "Copy a directory, its sub directories, and all their files to tape.") ;6.19.87
   ("Backup Partition" :VALUE :backup-partition :DOCUMENTATION "Copy a partition to tape.")
   ("Make Bootable Tape" :VALUE :make-bootable-tape :DOCUMENTATION
    "Make a bootable tape by creating a label and copying the label and one or more partitions to a tape.")
   ("" :no-select nil)
   ("Restore File" :VALUE :restore-file :DOCUMENTATION "Restore the next file on tape to some filename.")
   ("Restore Directory" :VALUE :restore-directory :DOCUMENTATION
    "Restore the next directory from the tape until its delimiting extra EOF is encountered.")
   ("Restore Partition" :VALUE :restore-partition :DOCUMENTATION "Restore the next partition on the tape to disk.")
   ("Restore Bootable Tape" :VALUE :restore-bootable-tape :DOCUMENTATION
    "Restore one or more partitions from a bootable tape.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("Verify File" :VALUE :verify-file :DOCUMENTATION "Verify the next file on the tape with some filename.")
   ("Verify Directory" :VALUE :verify-directory :DOCUMENTATION
    "Verify the next directory from the tape until its delimiting extra EOF is encountered.")
   ("Verify Partition" :VALUE :verify-partition :DOCUMENTATION "Verify the next partition on the tape.")
   ("Verify Bootable Tape" :VALUE :verify-bootable-tape :DOCUMENTATION
    "Verify one or more partitions on a bootable tape.")
   ("" :no-select nil)
   ("Load Distribution Tape" :VALUE :load-distribution-tape :DOCUMENTATION
    "Load software from a distribution tape on to the local file system.")
   ("" :no-select nil)	   
   ("Write EOF" :VALUE :write-eof :DOCUMENTATION "Write an EOF marker on tape. This should only be done to seperate groups of backups for convenience.  It does not mark the end of tape.")
   ("" :no-select nil)
   ("Help" :VALUE :help :DOCUMENTATION "Menu of help subtopics. Subjects include Read-Write Mode, Streaming, Errors, and Partitions.")
   ("Exit" :VALUE :exit :DOCUMENTATION "Exit this window and return to the previous window.")))


(defparameter carry-tape-command-menu-list 
; Miscellaneous operations
 '(("Prepare Tape" :VALUE :prepare-tape :DOCUMENTATION "Prepare the tape device by looking for an online tape and making it the current tape unit.  Selection of tape format (density) is required.")
   ("Unload Tape" :VALUE :unload :DOCUMENTATION "Unload the tape cartridge.  This leaves the tape offline until a Load Tape is performed.")
   ("Load Tape" :VALUE :load-tape :DOCUMENTATION "Load the tape cartridge, making it available and leaving it rewound to its physical beginning.  This resets the tape formatter and drive.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("Erase Entire Tape" :VALUE :erase :DOCUMENTATION "Erase the entire tape, after confirmation.")
   ("Re-Tension" :VALUE :tension :DOCUMENTATION "Tension tape by unwinding and rewinding.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("Rewind" :VALUE :rewind :DOCUMENTATION
    "Position the tape to its physical beginning, ready to read or write.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("" :no-select nil)
   ("List Carry Tape Contents" :VALUE :list-carry-tape :DOCUMENTATION
    "List files on a Carry tape.")
   ("" :no-select nil)
   ("Make a Carry Tape" :VALUE :make-carry-tape :DOCUMENTATION
    "Backup files to a cartridge tape using carry format. Files on a carry tape can be read on a Symbolics Lisp machine.")
   ("" :no-select nil)
   ("Restore a Carry Tape" :VALUE :restore-carry-tape :DOCUMENTATION
    "Restore files from a carry tape.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("" :no-select nil)
   ("" :no-select nil)
   ("" :no-select nil)
   ("Help" :VALUE :help :DOCUMENTATION "Menu of help subtopics. Subjects include Read-Write Mode, Streaming, Errors, and Partitions.")
   ("" :no-select nil)
   ("Exit" :VALUE :exit :DOCUMENTATION "Exit this window and return to the previous window.")))

(defparameter half-inch-command-menu-list 
; Miscellaneous operations
 '(("Prepare Tape" :VALUE :prepare-tape :DOCUMENTATION "Prepare the tape device by looking for an online tape and making it the current tape unit.  Selection of tape format (density) is required.")
   ("Unload Tape" :VALUE :unload :DOCUMENTATION "Unload the tape cartridge.  This leaves the tape offline until a Load Tape is performed.")
   
   ("Erase Tape" :VALUE :erase :DOCUMENTATION "Erase the entire tape, after confirmation.")
   
   ("" :no-select nil)
   ("Rewind" :VALUE :rewind :DOCUMENTATION
    "Position the tape to its physical beginning, ready to read or write.")
   
   ("Position Past File (EOF)" :VALUE :space-to-eof :DOCUMENTATION 
    "Position the tape past the next end of file (moves tape past next file) or multiple files or partitions.")
   ;("Position Past Blocks" :VALUE :space-blocks :DOCUMENTATION 
   ; "Position the tape forward a number of blocks.  Generally, this is not recommended.")
   ("Space Reverse (EOF)" :value :space-reverse-by-filemarks :documentation
    "Moves the tape header back by one or multiple filemarks.")
   ;("Space Reverse Blocks" :value :space-reverse-blocks :documentation
   ; "Position the tape backward a number of blocks.  Generally, this is not recommended.")
   ("" :no-select nil)	   
   ("List Contents" :VALUE :list-reel-tape :DOCUMENTATION
	  "List files on a Reel tape.")
   ;("List directory" :VALUE :list-directory :DOCUMENTATION
	  ;"Read the next directory, displaying headers from tape until a double EOF is encountered.  Files and partitions backed up without an extra EOF
;will be treated as a directory.  If there are sub-directories written by a single Backup Directory command, th;ey will also be listed.")
; Data movement operations
   ("" :no-select nil)
   ("Backup File" :VALUE :backup-file :DOCUMENTATION "Copy a file to tape.  Wildcarding is not allowed.")
   ("Backup Directory" :VALUE :backup-directory :DOCUMENTATION
    "Copy a directory, its sub directories, and its all their files to tape.")
   ("Backup Partition" :VALUE :backup-partition :DOCUMENTATION "Copy a partition to tape.")
   ("" :no-select nil)
   
   ("Restore File" :VALUE :restore-file :DOCUMENTATION "Restore the next file on tape to some filename.")
   ("Restore Directory" :VALUE :restore-directory :DOCUMENTATION
    "Restore the next directory from the tape until its delimiting extra EOF is encountered.")
   ("Restore Partition" :VALUE :restore-partition :DOCUMENTATION "Restore the next partition on the tape to disk.")
   ("" :no-select nil)
   
   ("Verify File" :VALUE :verify-file :DOCUMENTATION "Verify the next file on the tape with some filename.")
   ("Verify Directory" :VALUE :verify-directory :DOCUMENTATION
    "Verify the next directory from the tape until its delimiting extra EOF is encountered.")
   ("Verify Partition" :VALUE :verify-partition :DOCUMENTATION "Verify the next partition on the tape.")
   ("Write EOF" :VALUE :write-eof :DOCUMENTATION "Write an EOF marker on tape. This should only be done to seperate groups of backups for convenience.  It does not mark the end of tape.")
   ("" :no-select nil)
   ("Help" :VALUE :help :DOCUMENTATION "Menu of help subtopics. Subjects include Read-Write Mode, Streaming, Errors, and Partitions.")
   ("Exit" :VALUE :exit :DOCUMENTATION "Exit this window and return to the previous window.")))

(defparameter Tar-tape-command-menu-list   ;8.11.87
 '(("Prepare Tape" :VALUE :prepare-tape :DOCUMENTATION "Prepare the tape device by looking for an online tape and making it the current tape unit.  Selection of tape format (density) is required.")
   ("1Prepare* 1Remote Tape*" :VALUE 1:prepare-remote-tape* :DOCUMENTATION "Prepare 1a remote* tape device by 1specifing* 1a remote host name*.  Selection of tape format (density) is required.")
   ("Unload Tape" :VALUE :unload :DOCUMENTATION "Unload the tape cartridge.  This leaves the tape offline until a Load Tape is performed.")
   ("Load Tape" :VALUE :load-tape :DOCUMENTATION "Load the tape, making it available and leaving it rewound to its physical beginning.  This resets the tape formatter and drive.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("Erase Entire Tape" :VALUE :erase :DOCUMENTATION "Erase the entire tape, after confirmation.")
   ("Re-Tension" :VALUE :tension :DOCUMENTATION "Tension tape by unwinding and rewinding.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("Rewind" :VALUE :rewind :DOCUMENTATION
    "Position the tape to its physical beginning, ready to read or write.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("" :no-select nil)
   ("List Tar Tape" :VALUE :list-tar-tape :DOCUMENTATION
    "List files on a Tar tape.")
   ("" :no-select nil)
   ("Make a Tar Tape" :VALUE :make-tar-tape :DOCUMENTATION
    "Copy files to a tape using standard UNIX Tar format.")
   ("" :no-select nil)
   ("Restore a Tar Tape" :VALUE :restore-tar-tape :DOCUMENTATION
    "Restore files from a standard UNIX Tar format tape.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("" :no-select nil)
   ("Write EOF" :VALUE :write-eof :DOCUMENTATION "Write an EOF marker on tape.")
   ("" :no-select nil)
   ("" :no-select nil)
   ("Help" :VALUE :help :DOCUMENTATION "Menu of help subtopics. Subjects include Read-Write Mode, Streaming, Errors, Tar, Carry, and Partitions.")
   ("" :no-select nil)
   ("Exit" :VALUE :exit :DOCUMENTATION "Exit this window and return to the previous window.")))


;;;
;;; Choose Variables
;;;

(DEFVAR RESTORE-SELECTIVE-NAME ()) 

(DEFVAR VERIFY-SELECTIVE-NAME "")      ;this isn't used, but I didn't want to pull it for a patch JEB

(DEFVAR BACKUP-DIRECTORY-NAME ()) 

(DEFVAR BACKUP-FILE-NAME ()) 


(DEFVAR FORWARD-BLOCKS 1) 

(DEFVAR FORWARD-EOFS 1) 

(DEFVAR STARTING-OFFSET 0) 

(DEFVAR PARTITION-UNIT 0) 

(DEFVAR PARTITION-BAND ()) 

(DEFVAR VERIFY-HOST ()) 

(DEFVAR WRITE-EOF-P T) 

(DEFVAR QUERY-P ()) 

(DEFVAR SET-BACKED-UP-FLAG ()) 

(DEFVAR ONLY-LATEST ()) 

(DEFVAR SINCE ()) 

(DEFVAR NOT-BACKED-UP-ONLY ()) 

(DEFVAR PREPARE-TAPE-UNIT ()) 

(DEFVAR STREAM-MODE T) 

(DEFVAR NORMAL-ENCODING T) 

(DEFVAR SUBS T) 

(DEFVAR CREATE-DIRECTORIES T) 

(DEFVAR *USE-ORIGINAL-HOST* ()) 

;;;
;;; Documentation strings for CVV entries
;;;

(DEFVAR DOC-RMF-ORIGINAL-HOST
   "Choose Yes to restore to the host that was specified during Backup.  This will merge the host name read from tape with the rest of the 
   Destination Pathname.  Choose No to restore to the host specified in the Destination Pathname.") 


(DEFVAR DOC-BP-UNIT "Enter the number of the unit where the partition to backup resides.") 

(DEFVAR DOC-BP-PART "Enter the name of the partition to backup to tape.") 

(DEFVAR DOC-BP-OFFSET
   "Enter zero to start at the beginning or a number specifying how many 100-block units into the partition to start at.")

(DEFVAR DOC-VP-UNIT
   "Enter the number of the unit where the partition to verify against resides.") 

(DEFVAR DOC-VP-PART "Enter the name of the partition to verify against the partition on tape.") 

(DEFVAR DOC-VP-STREAM "To verify the partition in stream mode (hardware permitting) choose Yes.") 

(DEFVAR DOC-RP-UNIT "Enter the number of the unit to copy the partition to.") 

(DEFVAR DOC-RP-PART "Enter the name of the disk partition to copy the tape partition to.") 

(defvar doc-space-reverse-by-filemarks "Move the tape backward past one or more EOFs (filemarks).")

(defvar doc-space-reverse-blocks "Move the tape backward some number of 1K byte blocks.")

(DEFVAR DOC-SPACE-EOF "Move the tape forward past how many EOFs (file marks).") 

(DEFVAR DOC-SPACE-BLOCKS "Move the tape forward some number of 1K byte blocks.") 


;;; changed .. 8-8-85

(DEFVAR DOC-RMF-DIR
   "Specify the destination pathname.  Wildcards are allowed.  For example, to restore to the original directory, name, type, and version, but on
the local machine enter LM:*		   ;*.*#*.") 

;;; changed .. 10-22-85

(DEFVAR DOC-VF-PATH
	"Compare the file(s) on tape with the file(s) specified by this pathname.  Wildcard characters are allowed.") 

(DEFVAR DOC-BMF-EOF
	"Choose Yes to write an EOF, marking this backup as a directory (logical backup).  By not writing an EOF afterwards, directory backups can be concantenated together, allowing a single Restore Directory or Verify Directory to operate on them later.") 

(DEFVAR DOC-RMF-SEL
	"Specify a selective source pattern pathname.  Files on tape whose original names match this pattern will be be operated on.  Files that do not
match will be skipped over.  Wildcards are allowed.") 

(DEFVAR DOC-BMF-DIR
	"Enter a pathname specifying the file(s) to be backed up. Wildcard characters are allowed.") 

(DEFVAR DOC-BF-PATH
	"Enter the name of the file to be written to tape.  Wildcard characters are not allowed.") 



(DEFVAR DOC-BMF-SUBS "Choose Yes to traverse sub-directories.") 

(DEFVAR DOC-RMF-QUERY "Choose Yes to be queried on each file whether or not to restore it.") 

(DEFVAR DOC-BMF-QUERY "Choose Yes to be asked on each file whether or not to back it up.") 

(DEFVAR DOC-BMF-SINCE
	"Enter a universal time.  Only files created since this time will be backed up.") 

(DEFVAR DOC-BMF-INCREMENTAL
	"Choose Yes to only backup files that are not already marked as backed up.") 

(DEFVAR DOC-BMF-SET-FLAG
	"Choose Yes to mark files as backed up after they are all copied to tape.") 

(DEFVAR DOC-BMF-ONLY-LATEST "Choose Yes to backup only the latest version of each file.") 


(DEFVAR DOC-RD-CREATE-DIRECTORIES
	"Choose Yes to create directories without further intervention.  Choose No to be prompted whether to create new directories.") 

(DEFVAR DOC-VMF-ORIGINAL-HOST
	"Choose Yes to verify against the host that was specified during Backup.  This will merge the host name read from tape with the rest of the   
Destination Pathname.  Choose No to verify against the host specified in the Destination Pathname.")

(DEFFLAVOR BACKUP-WINDOW-WITH-WHOLINE-DOC
	   ()
	   (TV:WINDOW))

(DEFMETHOD (BACKUP-WINDOW-WITH-WHOLINE-DOC :WHO-LINE-DOCUMENTATION-STRING) ()
  "R2: Bring up the System Menu.")



(DEFFLAVOR BACKUP-FRAME
	   (OUT-STREAM)
	   (TV:INFERIORS-NOT-IN-SELECT-MENU-MIXIN
	    ;;tv:full-screen-hack-mixin
	    TV:PROCESS-MIXIN		   ;Process-Mixin should come before Select-Mixin (See doc for P-M)
	    TV:SELECT-MIXIN
	    TV:CENTERED-LABEL-MIXIN
	    TV:TOP-LABEL-MIXIN
	    TV:LABEL-MIXIN
	    TV::ESSENTIAL-MOUSE
	    TV:WINDOW-WITH-TYPEOUT-MIXIN
	    TV:STREAM-MIXIN
	    TAPE-MIXIN
	    TV:BORDERED-CONSTRAINT-FRAME-WITH-SHARED-IO-BUFFER)
  (:SETTABLE-INSTANCE-VARIABLES OUT-STREAM)
  (:GETTABLE-INSTANCE-VARIABLES OUT-STREAM)
  (:DEFAULT-INIT-PLIST :PROCESS '(TOP-LEVEL	   ;Function to be run by the process
				   :REGULAR-PDL-SIZE 3000) ;Ensure the PDL stack isn't overflown
    :SAVE-BITS :DELAYED			   ;Don't create a bit-save array until we need it
    :LABEL (LIST ':STRING "Backup System"))
  (:DOCUMENTATION "Window interface to the magtape Backup System.")) 

(DEFMETHOD (BACKUP-FRAME :BEFORE :INIT) (IGNORE)
  "Set panes and constraints."
  ;;Only display the remote menu on non Explorers.
  (declare (special remote-half-inch-command-menu-list))
  (if (eq *local-host-machine-type* :explorer)  ;06-07-88 DAB
      (SETQ TV:PANES
	    `((TEXT-PANE MT::BACKUP-WINDOW-WITH-WHOLINE-DOC
			 :LABEL (:FONT ,*LABEL-FONT* :STRING "Backup Typeout Window"))
	      (COMMAND-PANE BACKUP-MENU-PANE
			    :LABEL (:FONT ,*LABEL-FONT* :STRING "Backup Command Menu")
			    :DEFAULT-FONT ,*COMMAND-MENU-FONT*
			    :ITEM-LIST ,quarter-inch-command-menu-list))
	    TV:CONSTRAINTS
	    '((LANDSCAPE-1 (WHOLE-THING)
			   ((WHOLE-THING :HORIZONTAL (:EVEN) (TEXT-PANE COMMAND-PANE) ((TEXT-PANE 0.8))
					 ((COMMAND-PANE :EVEN))))))
	    TV::CONFIGURATION (IF *LANDSCAPE-MONITOR-P*
				  'LANDSCAPE-1
				  'PORTRAIT-1))
      (SETQ TV:PANES
	    `((TEXT-PANE MT::BACKUP-WINDOW-WITH-WHOLINE-DOC
			 :LABEL (:FONT ,*LABEL-FONT* :STRING "Backup Typeout Window"))
	      (COMMAND-PANE BACKUP-MENU-PANE
			    :LABEL (:FONT ,*LABEL-FONT* :STRING "Remote Half Inch Tape Menu")
			    :DEFAULT-FONT ,*COMMAND-MENU-FONT*
			    :ITEM-LIST ,remote-half-inch-command-menu-list))
	    TV:CONSTRAINTS
	    '((LANDSCAPE-1 (WHOLE-THING)
			   ((WHOLE-THING :HORIZONTAL (:EVEN) (TEXT-PANE COMMAND-PANE) ((TEXT-PANE 0.8))
					 ((COMMAND-PANE :EVEN))))))
	    TV::CONFIGURATION (IF *LANDSCAPE-MONITOR-P*
				  'LANDSCAPE-1
				  'PORTRAIT-1))))
 


(DEFMETHOD (BACKUP-FRAME :AFTER :INIT) (IGNORE)
  (LET ((TEXT-PANE (SEND SELF :GET-PANE 'TEXT-PANE)))
    (SEND TEXT-PANE :SET-MORE-P ())
    (send text-pane :set-current-font *backup-frame-FONT* t)   ;10-06-88 DAB
    )) 


(DEFMETHOD (BACKUP-FRAME :BEFORE :SET-MORE-P) (ARG)
  (SEND (SEND SELF :GET-PANE 'TEXT-PANE) :SET-MORE-P ARG)) 


(DEFFLAVOR BACKUP-MENU-PANE
	   ()
	   (TV:COMMAND-MENU-PANE)
  (:DOCUMENTATION :COMBINATION "2Command menu for the *Backup window2.*")
  (:DEFAULT-INIT-PLIST :MORE-P NIL)) 

;;; This will set up basic color for the backup window
;;; NOTE: it also modifies the superior  PMH
(defmethod (backup-menu-pane :after :init)(&rest ignore)
  (send self :set-background-color w:33%-gray-color)
  (setf (tv:label-background tv:label) w:66%-gray-color)
  (setf (tv:label-background (send tv:superior :label)) w:75%-gray-color)
  (setf (tv:label-color (send tv:superior :label)) w:black)
  )

(DEFUN TOP-LEVEL (MT-WINDOW)
  (SEND MT-WINDOW :TOP-LEVEL)) 


(DEFMETHOD (BACKUP-FRAME :top-level) ()
  (loop
    (condition-case (error)
	(command-loop self)	
      (eh:unit-bad-number-error
       (format (send self ':out-stream)
	       "~&Tape unit not initialized.  Mouse Prepare Tape with a tape loaded in the drive before attempting any tape operations.  This must be done once after each boot."))
      (eh:unit-offline-error
       (format (send self ':out-stream) (send error :report-string)))
      ((system:tape-error :EOF :END-OF-RECORDED-MEDIA)
       (let ((report (send error :report-string)))
	 (if (equal report "")
	     (format (send self ':out-stream) "~&Unknown Tape Error: ~%~a" error)     
	     (format (send self ':out-stream) "~&Tape Error: ~a" report)))))))


(DEFMETHOD (BACKUP-FRAME :WHO-LINE-DOCUMENTATION-STRING) ()
  "R2: Bring up the System Menu.")

(DEFUN COMMAND-LOOP (FRAME)
  (LET* ((TEXT-PANE (SEND FRAME :GET-PANE 'TEXT-PANE))
	 (BLINKY (CAR (SEND TEXT-PANE :BLINKER-LIST)))
	 (*TERMINAL-IO* TEXT-PANE)
	 (LOCAL-OUT-STREAM TEXT-PANE))
    (SEND FRAME :SET-OUT-STREAM LOCAL-OUT-STREAM)
    (SEND BLINKY :SET-VISIBILITY :BLINK)
    (FORMAT LOCAL-OUT-STREAM "~&")	   ;this gives a new line after ABORTing out of error
    (CATCH :EXIT
      (LOOP FOR INPUT = (SEND FRAME :ANY-TYI) DO
	    (COND
	      ((AND (CONSP INPUT) (NOT (EQUALP (CAR INPUT) :MOUSE-BUTTON)))
	       (LET ((VAL (AND (CONSP INPUT) (SECOND INPUT)))
		     (*print-BASE* 10)
		     (*read-BASE* 10))
		 (SETQ VAL (GET VAL :VALUE))
		 (if (or *remote-tape-p*		;1; added for remote -- Kenny Tung*
			 (null *TAPE-UNIT*)		;1; the very first time*
			 ;1; Next will allow server host to use another server.*
			 (member val '(:prepare-remote-tape :help :exit :unload) )
			 (and *TAPE-UNIT*		;1; make sure* 1tape unit is used by the same host (person)*
			      (string= *TAPE-UNIT* si:local-host-name)) )
		     (progn
		       ;(format t "~%VAL = ~a." VAL)
		       (SEND BLINKY :SET-VISIBILITY :BLINK)
		       (CASE (SEND FRAME :SEND-IF-HANDLES VAL)
			 (:ABORT (FORMAT LOCAL-OUT-STREAM "~&~A aborted.~%" VAL))
			 (:COMPLETE (FORMAT LOCAL-OUT-STREAM " ~A complete.~%" VAL))
			 (:MUTE NIL)
			 (NIL (FORMAT LOCAL-OUT-STREAM "~&Command ~A not yet handled.~%" VAL)))) ;1; progn*
		     (FORMAT (SEND SELF :OUT-STREAM) "~&This tape unit in use by remote user ~s. Try later.~%"
			     *tape-unit*) )		;1; if*
		     ) )
	      ((EQl INPUT (char-code #\END))
	       (send frame :bury)   ;08-15-88 DAB
	       (THROW :EXIT
		      :EXIT-FROM-KEYBOARD))
	      ((EQl INPUT (char-code #\HELP))	   ;need any others ???
	       (SEND FRAME :HELP))
	      ((EQl INPUT (char-code #\clear-screen))	   ;need any others ???
	       (SEND TEXT-PANE :CLEAR-SCREEN))
	      (T T))
	    (SETQ INPUT ()) UNTIL ()))))


;;;
;;;  This tape stuff probably belongs in a separate file ??? 
;;;

(DEFFLAVOR TAPE-MIXIN
	   ()				   ;no instance variables
	   ()				   ;no other flavors
  ) 


(DEFMETHOD (TAPE-MIXIN :DISPLAY-MSG) (MSG)
  "Write a msg string to out-stream"
  (FORMAT (SEND SELF :OUT-STREAM) "~&~A ..." MSG)) 


;;;   Selections of format type and corresponding initialize function.
(DEFPARAMETER INITIALIZE-FORMAT-ALIST
	      (LIST '("1/4 inch Cartridge Tape" SETUP-QUARTER-INCH-TAPE)
		    '("1/2 inch Reel Tape" SETUP-HALF-INCH-TAPE)
		    '("1/4 inch Cartridge Tape - Carry format" Carry-tape)
		    '("1/4 inch Cartridge Tape - TAR format" QUARTER-INCH-TAR-TAPE)
		    '("1/2 inch Reel Tape - TAR format" HALF-INCH-TAR-TAPE)))	   ;8.19.87


;;; The following defvars were added for 1/2" tape support.- 2/23/87 mp
(defvar *temp-block-mode* :variable)
(defvar *temp-density* :double)
(defvar *temp-block-size* 512.)



(defvar *tape-mode* 'qic-24)
(defconstant get-quarter-inch-mode-doc-1 "Setup the cartridge tape drive to read in QIC-11 (4-track), or QIC-24 (9-track) mode, or 6250 mode which is QIC-11 with 9-tracks.")

(defun get-quarter-inch-mode ()
  (declare (Values tape-mode))
  (unless (member *tape-mode* '('qic-11 'qic-24 6250))
    (setf *tape-mode* 'qic-24))	   
  (SETQ DOIT ())
  (catch 'gqi-abort
    (EVAL
      `(tv:choose-variable-values '((*tape-mode*
				      "Mode (Density)" :documentation ,get-quarter-inch-mode-doc-1 
				      :choose (QIC-11 QIC-24 6250)))
				  :MARGIN-CHOICES
				  '(("Do it" (AND (SETQ DOIT T) (THROW 'gqi-abort ()))) "Abort")
				  :LABEL "Select mode" :EXTRA-WIDTH 25)))
  (if DOIT *tape-mode* NIL))

(defconstant ghip-density "Select SINGLE to use 1600 bpi or select DOUBLE to use 3200 bpi.")

;;; 8.11.87 MBC
(defun get-half-inch-parameters ()
  (declare (Values BLOCK-SIZE BLOCK-MODE DENSITY))
  (unless (member *temp-density* '(:single :double))
    (setf *temp-density* :double))
  (unless (member *temp-block-mode* '(:Variable :Fixed))
    (setf *temp-block-mode* :Variable))
  (let (DENSITY)
    (SETQ DOIT ())
    (catch 'ghi-abort
      (eval
	`(tv:choose-variable-values '((*temp-block-mode* "Block Mode" :choose (:Variable :Fixed))
				      (*temp-density* "Density"
						      :documentation ,ghip-density
						      :choose (:single :double)))
				    :MARGIN-CHOICES
				    '(("Do it" (AND (SETQ DOIT T) (THROW 'ghi-abort ()))) "Abort")
				    :LABEL "Select block mode and density" :EXTRA-WIDTH 25)))
    (when doit				   ;End pressed or doit clicked
      (if (eq  *temp-density* :single)
	  (setq density 2.)
	  (setq density 6.))
      (if (eq *temp-block-mode* :variable)
	  (setq *temp-block-size* 0.)
	  (progn
	    (setf doit nil)		   ;reuse doit flag
	    (UNLESS (MEMBER *temp-block-size* '(512. 1024.))
	      (SETF *temp-block-size* 1024.))
	    (catch 'ghi1-abort
	      (tv:choose-variable-values '((*temp-block-size* "Block size" :choose (512. 1024.))) 
					 :MARGIN-CHOICES
					 '(("Do it" (AND (SETQ DOIT T) (THROW 'ghi1-abort ()))) "Abort")
					 :LABEL "Enter a value for the block size"))
	    )))
    (if DOIT				   ;detect ABORTs
	(values *TEMP-BLOCK-SIZE* *TEMP-BLOCK-MODE* DENSITY)
	NIL)))
(DEFMETHOD (tape-mixin :prepare-tape) () ;1; tung done*
  "Prepare tape."
  (let ((fini :ABORTED) doit
	(unit (pick-drive))
	(tape-type :quarter-inch))
    (when unit
      (let ((fun (w:menu-choose initialize-format-alist :label "Choose format"
				:default-item (first initialize-format-alist))))
	(setf *current-unit* unit)
	(selector fun equal
	  (('setup-half-inch-tape )
	   (setq tape-type :half-inch)
	   (multiple-value-bind (block-size block-mode density)	   ;8.11.87 MBC
	       (get-half-inch-parameters)
	     (when block-size
	       (prepare-tape-drive unit :block-size block-size :block-mode block-mode
				   :tape-type tape-type :density density)
	       (setq fini :COMPLETE)
	       (send (send self :get-pane 'command-pane)
		     :set-label `(:FONT ,*LABEL-FONT*
					:STRING "Half Inch Tape Menu"
					:background ,w:66%-gray-color))
	       (send (send self :get-pane 'command-pane)
		     :set-item-list  half-inch-command-menu-list)))
	   )
	  (('setup-quarter-inch-tape)
	   (prepare-tape-drive unit)	   ;hw
	   (setq fini :COMPLETE)
	   (send (send self :get-pane 'command-pane)
		 :set-label `(:FONT ,*LABEL-FONT*
				    :STRING "Quarter Inch Tape Menu"
				    :background ,w:66%-gray-color))
	   (send (send self :get-pane 'command-pane) :set-item-list
		 quarter-inch-command-menu-list)
	   )
	  (('carry-tape)
	   (prepare-tape-drive unit)	   ;hw
	   (setq fini :COMPLETE)
	   (send (send self :get-pane 'command-pane)
		 :set-label `(:FONT ,*LABEL-FONT*
				    :STRING "Carry Tape Menu"
				    :background ,w:66%-gray-color))
	   (send (send self :get-pane 'command-pane) :set-item-list carry-tape-command-menu-list)
	   )
	  (('QUARTER-INCH-TAR-TAPE)
	   (multiple-value-bind (tape-mode)	   ;8.11.87 MBC
	       (get-quarter-inch-mode)
	     (when tape-mode
	       (INITIALIZE-TAPE-DRIVE :UNIT *CURRENT-UNIT* :BUFFERED T :SPEED 0 :BLOCK-SIZE 1024.
				      :DENSITY 
				      (CASE tape-mode
					(QIC-11 4)
					(QIC-24 5)
					(6250 132.)))
	       
	       (setq fini :COMPLETE)
	       (send (send self :get-pane 'command-pane)
		     :set-label `(:FONT ,*LABEL-FONT*
					:STRING "TAR Tape Menu"
					:background ,w:66%-gray-color))
	       (send (send self :get-pane 'command-pane)
		     :set-item-list tar-tape-command-menu-list)))
	   )
	  (('HALF-INCH-TAR-TAPE)
	   (setq tape-type :half-inch)
	   (multiple-value-bind (block-size block-mode density)	   ;8.11.87 MBC
	       (get-half-inch-parameters)
	     (when block-size
	       (prepare-tape-drive unit :block-size block-size :block-mode block-mode
				   :tape-type tape-type :density density)
	       (setq fini :COMPLETE)
	       (send (send self :get-pane 'command-pane)
		     :set-label `(:FONT ,*LABEL-FONT* :STRING "TAR Tape Menu"))
	       (send (send self :get-pane 'command-pane)
		     :set-item-list Tar-tape-command-menu-list))))
	  )) ) ;1; when unit*
    (when (eq fini :COMPLETE)
      (setf *tape-unit* SI:LOCAL-HOST-NAME)	;1; this is server side command -- TUNG*
      (setf (get 'mt:*current-unit* 'tape-type) tape-type))   ;8.11.87 Remember type
    fini))






(DEFMETHOD (TAPE-MIXIN :ERASE) () ;1; tung done*
  "Erase the entire tape."
  (LET ((FINI :ABORTED))
    (IF (YES-OR-NO-P "~%Do you really want to erase everything on the tape?")
	(PROGN
	  ()
	  (SEND SELF :DISPLAY-MSG "Erasing")
	  (cond (*remote-tape-p*
		 ;(format (send self ':out-stream) " Remote working.") )
		 (and (tape-service-req *remote-hostname* $ERASE-TAPE)
		      (SETQ FINI :COMPLETE)) )
		(t (ERASE)
		   (SETQ FINI :COMPLETE))) ;1; cond*
	  ))	;1; if*
    FINI))

(DEFMETHOD (TAPE-MIXIN :REWIND) () ;1; tung done*
  "Rewind tape method"
  (LET ((FINI :ABORTED))
    (SEND SELF :DISPLAY-MSG "Rewinding")
    (cond (*remote-tape-p*
	   ;(format (send self ':out-stream) " Remote working.")
	   (and (tape-service-req *remote-hostname* $REWIND-TAPE)
		(SETQ FINI :COMPLETE)) )
	  (t (REWIND)
	     (setf (get '*current-unit* :AT-BOT-P) T)	   ;8.11.87 MBC
	     (SETQ FINI :COMPLETE) )) ;1; cond*
    FINI))
 


(DEFMETHOD (TAPE-MIXIN :LOAD-TAPE) () ;1; tung done*
  "Load tape method"
  (LET ((FINI :ABORTED))
    (SEND SELF :DISPLAY-MSG "Loading")
    (cond (*remote-tape-p*
	   (and (tape-service-req *remote-hostname* $LOAD-TAPE)
		(SETQ FINI :COMPLETE)) )
	  (t (LOAD-TAPE)
	     (setf *tape-unit* SI:LOCAL-HOST-NAME)
	     (SETQ FINI :COMPLETE)) )
    FINI)) 

(DEFMETHOD (TAPE-MIXIN :UNLOAD) () ;; tung done
  "Unload tape method"
  (LET ((FINI :ABORTED)
	(continue-p T) )
    (SEND SELF :DISPLAY-MSG "Unloading")
    (cond (*remote-tape-p*
	   (when (tape-service-req *remote-hostname* $UNLOAD-TAPE)
	     (SETQ FINI :COMPLETE
		   ;*remote-tape-p*  NIL   ;10-06-88 DAB Do this at logout.
		   )
	     (when (eq *local-host-machine-type* :explorer)   ;06-07-88 DAB
	       (send (send self :get-pane 'command-pane)
		     :set-label `(:FONT ,*LABEL-FONT*
					:STRING "Quarter Inch Tape Menu"
					:background ,w:66%-gray-color))
	       (send (send self :get-pane 'command-pane) :set-item-list
		     quarter-inch-command-menu-list ))))
	  (t (when (and *TAPE-UNIT*		;; make sure tape unit is used by the same host (person)
			(string/= *TAPE-UNIT* si:local-host-name))
	       (unless (yes-or-no-p "~%Tape unit in use by ~a, Unload anyway?" *TAPE-UNIT*)
		   (setq continue-p NIL)) )	;; No, do not unload
	     (when continue-p
	       (ignore-errors (UNLOAD))
	       (SETQ FINI :COMPLETE		;; client should never set *tape-unit* except this one
		     *TAPE-UNIT* NIL
		     ))  ) )	;; cond, Tung for remote tape
    FINI))

 


(DEFMETHOD (TAPE-MIXIN :TENSION) () ;1; tung done*
  "Tension tape method"
  (LET ((FINI :ABORTED))
    (SEND SELF :DISPLAY-MSG "Tensioning")
    (cond (*remote-tape-p*
	   (and (tape-service-req *remote-hostname* $RE-TENSION)
		(SETQ FINI :COMPLETE)) )
	  (t (TENSION)
	     (SETQ FINI :COMPLETE) ))
    FINI)) 


(DEFMETHOD (TAPE-MIXIN :POSITION-TO-APPEND) () ;1; tung done*
  "Position tape for appending."
  (LET ((FINI :ABORTED))
    (SEND SELF :DISPLAY-MSG "Preparing to append")
    (cond (*remote-tape-p*
	   (and (tape-service-req *remote-hostname* $PREPARE-APPEND)
		(SETQ FINI :COMPLETE)) )
	  (t (SPACE-TO-APPEND)
	     (SETQ FINI :COMPLETE) ))
    FINI)) 



(DEFMETHOD (TAPE-MIXIN :SPACE-TO-EOF) () ;1; tung done*
  "Space to EOF method"
  (LET ((FINI :ABORTED)
	(backup-window self))
    (declare (special backup-window))
    (SETQ DOIT ())
    (CATCH 'SPACE-DOIT
      (EVAL
	`(TV:CHOOSE-VARIABLE-VALUES
	   '((FORWARD-EOFS "Number of EOFs" :DOCUMENTATION ,DOC-SPACE-EOF :NUMBER))
	   :MARGIN-CHOICES '(("Do it" (AND (SETQ DOIT T) (THROW 'SPACE-DOIT ()))) "Abort")
	   :LABEL "Space over file marks" :SUPERIOR BACKUP-WINDOW)))
    (WHEN DOIT
      (FORMAT (SEND SELF :OUT-STREAM) "~&Positioning forward ~D EOFs..." FORWARD-EOFS)
      (cond (*remote-tape-p*
	     (and (tape-service-req *remote-hostname* $PAST-EOF)
		  (SETQ FINI :COMPLETE)) )
	    (t (SPACE-TO-EOF *CURRENT-UNIT* FORWARD-EOFS)
	       (SETQ FINI :COMPLETE)) )
    FINI)) ) 


(DEFMETHOD (TAPE-MIXIN :SPACE-BLOCKS) () ;1; tung not done*
  "Space forward blocks method"
  (LET ((FINI :ABORTED)
	(backup-window self))
    (declare (special backup-window))
    (SETQ DOIT ())
    (CATCH 'BLOCKS-DOIT
      (EVAL
	`(TV:CHOOSE-VARIABLE-VALUES
	   '((FORWARD-BLOCKS "Number of blocks" :DOCUMENTATION ,DOC-SPACE-BLOCKS :NUMBER))
	   :MARGIN-CHOICES '(("Do it" (AND (SETQ DOIT T) (THROW 'BLOCKS-DOIT ()))) "Abort")
	   :LABEL "Space Forward 1K Blocks" :SUPERIOR BACKUP-WINDOW)))
    (WHEN DOIT
      (FORMAT (SEND SELF :OUT-STREAM) "~&Positioning forward ~D blocks..." FORWARD-BLOCKS)
      (SPACE-FORWARD *CURRENT-UNIT* FORWARD-BLOCKS)
      (SETQ FINI :COMPLETE))
    FINI)) 

(DEFMETHOD (TAPE-MIXIN :space-reverse-by-filemarks) ()
  "Space to EOF method"
  (LET ((FINI :ABORTED)
	(backup-window self))
    (declare (special backup-window))
    (setq doit nil)
    (catch 'space-doit
      (eval `(tv:choose-variable-values 
	       '((forward-eofs "Number of EOFs"  :documentation ,doc-space-eof :number))
	       :margin-choices '(("Do it" (and (setq doit t) (throw 'space-doit nil))) "Abort")
	       :label "Space over file marks"
	       :SUPERIOR BACKUP-WINDOW)))
    (when doit
      (format (send self ':out-stream) "~&Positioning backward ~D EOFs..." forward-eofs)
      (cond (*remote-tape-p*
	   (when (tape-service-req *remote-hostname* $REVERSE-EOF)
	     (SETQ FINI :COMPLETE)) )
	  (t (mt:space-reverse-by-filemarks *current-unit* forward-eofs)
	     (setq fini :COMPLETE)) ))
    fini))

(DEFMETHOD (TAPE-MIXIN :space-reverse-blocks) ()
  "Space backward blocks method"
  (LET ((FINI :ABORTED)
	(backup-window self))
    (declare (special backup-window))
    (setq doit nil)
    (catch 'blocks-doit
      (eval `(tv:choose-variable-values 
	       '((forward-blocks "Number of blocks" :documentation ,doc-space-blocks :number))
	       :margin-choices '(("Do it" (and (setq doit t) (throw 'blocks-doit nil))) "Abort")
	       :label "Space Backward 1K Blocks"
	       :SUPERIOR BACKUP-WINDOW)))
    (when doit      
      (format (send self ':out-stream) "~&Positioning backward ~D blocks..." forward-blocks)
      (mt:space-reverse-blocks *current-unit* forward-blocks)
      (setq fini :COMPLETE))
    fini))
;;;
(DEFMETHOD (TAPE-MIXIN :BACKUP-FILE) () ;1; tung done*
  "Backup a file to tape."
  (LET ((FINI :ABORTED)
	(backup-window self))
    (declare (special backup-window))
    (SETQ DOIT ())
    (IF BACKUP-FILE-NAME
	(LET ((BACKUP-PATHNAME (PATHNAME BACKUP-FILE-NAME)))
	  (IF (SEND BACKUP-PATHNAME :WILD-P)
	      (SETF BACKUP-FILE-NAME
		    (SEND BACKUP-PATHNAME :TRANSLATE-WILD-PATHNAME BACKUP-PATHNAME
			  (SEND (USER-HOMEDIR-PATHNAME) :NEW-PATHNAME :NAME "FOO" :TYPE :LISP :VERSION
				:NEWEST)))))
	(SETF BACKUP-FILE-NAME
	      (SEND (USER-HOMEDIR-PATHNAME) :NEW-PATHNAME :NAME "FOO" :TYPE :LISP :VERSION :NEWEST)))
    (CATCH 'B-F-DOIT
      (TV:CHOOSE-VARIABLE-VALUES
	`((BACKUP-FILE-NAME "Pathname" :DOCUMENTATION ,DOC-BF-PATH :directory-PATHNAME BACKUP-FILE-NAME))
	:MARGIN-CHOICES '(("Do it" (AND (SETQ DOIT T) (THROW 'B-F-DOIT ()))) "Abort") :WIDTH 80
	:LABEL "Backup File" :SUPERIOR BACKUP-WINDOW))
    (WHEN DOIT
      (FORMAT (SEND SELF :OUT-STREAM) "~&Backing up file ~A ..." BACKUP-FILE-NAME)
      (cond (*remote-tape-p*
	     (and (tape-service-req *remote-hostname* $BACKUP-FILE)
		  (SETQ FINI :COMPLETE)) )
	    (t (when (BACKUP-FILE BACKUP-FILE-NAME)
		   (SETQ FINI :COMPLETE)
		   )) ) ;1; cond*
      (when (eq FINI :ABORTED)
	(format (send self :out-stream) "~%->[File not backed-up] ~%") )
      ) ;1; when doit*
    FINI))



(DEFUN GUESS-DIRECTORY (&OPTIONAL (VERSION :NEWEST))
  (send (MAKE-PATHNAME :HOST "LM" :DIRECTORY :ROOT :NAME :WILD :TYPE :WILD :VERSION VERSION) :merge-device)) 




(DEFMETHOD (TAPE-MIXIN :BACKUP-DIRECTORY) () ;1; tung done*
  "Backup Directory to tape."
  (LET ((FINI :ABORTED)
	(backup-window self))
    (declare (special backup-window))
    (SETQ DOIT ())
    (UNLESS BACKUP-DIRECTORY-NAME
      (SETF BACKUP-DIRECTORY-NAME (GUESS-DIRECTORY)))
    (setf Tv::*Default-Directory-Pathname* Backup-Directory-Name)
    (CATCH 'BMF-DOIT
      (TV:CHOOSE-VARIABLE-VALUES
	`((BACKUP-DIRECTORY-NAME "Pathname" :DOCUMENTATION ,DOC-BMF-DIR
				 :directory-PATHNAME BACKUP-DIRECTORY-NAME)
	  (WRITE-EOF-P "Write an EOF at the end?" :DOCUMENTATION ,DOC-BMF-EOF :BOOLEAN)
	  (SUBS "Backup sub-directories too?" :DOCUMENTATION ,DOC-BMF-SUBS :BOOLEAN)
	  (SET-BACKED-UP-FLAG "Set backed up flag?" :DOCUMENTATION ,DOC-BMF-SET-FLAG :BOOLEAN)
	  (SINCE "Since time?" :DOCUMENTATION ,DOC-BMF-SINCE :STRING-OR-NIL)
	  (NOT-BACKED-UP-ONLY "Only files that are not backed up (incremental)?" :DOCUMENTATION
			      ,DOC-BMF-INCREMENTAL :BOOLEAN)
	  (QUERY-P "Query?" :DOCUMENTATION ,DOC-BMF-QUERY :BOOLEAN))
	:MARGIN-CHOICES '(("Do it" (AND (SETQ DOIT T) (THROW 'BMF-DOIT ()))) "Abort") :WIDTH 80
	:LABEL "Backup Directory" :SUPERIOR BACKUP-WINDOW))
    (WHEN DOIT
      (FORMAT (SEND SELF :OUT-STREAM) "~&Backing up Directory ~A ..." BACKUP-DIRECTORY-NAME)
      (cond (*remote-tape-p*
	     (and (tape-service-req *remote-hostname* $BACKUP-directory)
		  (SETQ FINI :COMPLETE)
		  (format (send self :out-stream) "~%")) )
	    (t (if (BACKUP BACKUP-DIRECTORY-NAME :SET-BACKED-UP-FLAG SET-BACKED-UP-FLAG 
			   :AFTER SINCE :NOT-BACKED-UP-ONLY NOT-BACKED-UP-ONLY :SELECTIVE QUERY-P
			   :SUB-DIRECTORIES SUBS :MARK-AS-DIRECTORY WRITE-EOF-P)
		   (SETQ FINI :COMPLETE))) )
      (when (eq FINI :ABORTED)
	(format (send self :out-stream) "~%->[Directory not backed-up] ~%")) ) ;1; when doit*
    FINI))




(DEFMETHOD (tape-mixin :backup-partition) () ;1; tung done*
  "Backup a partition to tape."
  (LET ((FINI :ABORTED)
	(part-name)
	(backup-window self))
    (declare (special backup-window))
    (setq doit nil)
    (catch 'b-p-doit
      (if (or (eq (get 'mt:*current-unit* 'tape-type) :half-inch)
	      (eq *CUR-TAPE-FORMAT* $HALF-INCH-TAPE-NUM)) ;03-07-88 changed = to eq
	  (eval `(tv:choose-variable-values
		   '((partition-unit "Unit number" :documentation ,doc-bp-unit)
		     (partition-band "Partition number or name" :string :documentation ,doc-bp-part))
		   :width 50.
		   :margin-choices '(("Do it" (and (setq doit t) (throw 'b-p-doit nil))) "Abort")
		   :label "Backup Partition"
		   :SUPERIOR BACKUP-WINDOW))
	  (eval `(tv:choose-variable-values
		   '((partition-unit "Unit number" :documentation ,doc-bp-unit)
		     (partition-band "Partition number or name" :string :documentation ,doc-bp-part)
		     (starting-offset "Starting offset" :documentation ,doc-bp-offset))
		   :width 50.
		   :margin-choices '(("Do it" (and (setq doit t) (throw 'b-p-doit nil))) "Abort")
		   :label "Backup Partition"
		   :SUPERIOR BACKUP-WINDOW))
	  ))
;    (unless (disk-unit-verified partition-unit)
;      (setq doit ()))
    (when doit
      (SETQ part-name PARTITION-BAND)
      (IF (AND (EQ (LENGTH part-name) 1) (MEMBER part-name *partition-num-list* :test #'string=))
	  (SETQ PARTITION-BAND (FORMAT () "~a~d" "LOD" part-name)))
      (format (send self :out-stream) "~&Backing up partition ~A on unit ~A ..." partition-band partition-unit)
      (cond (*remote-tape-p*
	     (and (tape-service-req *remote-hostname* $BACKUP-PARTITION)
		  (SETQ FINI :COMPLETE)) )
	    (t (if (eq (get 'mt:*current-unit* 'tape-type) :half-inch)
		   (backup-partition-half-inch-tape partition-unit partition-band)
		   (execute-backup-partition partition-unit partition-band starting-offset))
	       (setq fini :COMPLETE))) ) ;1; when doit*
    ;(and (eq FINI :COMPLETE) (WRITE-EOF))	1;; can't write-eof because client may have no tape unit*
    fini))


(DEFMETHOD (TAPE-MIXIN :RESTORE-FILE) ()
  "Restore a file to tape."
  (LET ((FINI :ABORTED)
	(backup-window self)
	result)
    (declare (special backup-window))
    (SETQ DOIT ())
    (UNLESS RESTORE-SELECTIVE-NAME
      (SETF RESTORE-SELECTIVE-NAME (GUESS-DIRECTORY)))
    (UNLESS BACKUP-FILE-NAME
      (SETF BACKUP-FILE-NAME (GUESS-DIRECTORY)))
    (CATCH 'R-F-DOIT
      (TV:CHOOSE-VARIABLE-VALUES
	`((RESTORE-SELECTIVE-NAME "Source Pathname" :DOCUMENTATION ,DOC-RMF-SEL
				  :directory-PATHNAME RESTORE-SELECTIVE-NAME)
	  (BACKUP-FILE-NAME "Destination Pathname" :DOCUMENTATION ,DOC-RMF-DIR
			    :directory-PATHNAME BACKUP-FILE-NAME)
	  (*USE-ORIGINAL-HOST* "Use host name from tape?" :DOCUMENTATION ,DOC-RMF-ORIGINAL-HOST
			       :BOOLEAN))
	:WIDTH 80 :MARGIN-CHOICES '(("Do it" (AND (SETQ DOIT T) (THROW 'R-F-DOIT ()))) "Abort")
	:LABEL "Restore File" :SUPERIOR BACKUP-WINDOW))
    (WHEN DOIT
      (SEND SELF :DISPLAY-MSG "Restoring file")
      (cond (*remote-tape-p*
	     (setf result (tape-service-req *remote-hostname* $RESTORE-FILE))
	     (and  result (SETQ FINI :COMPLETE)) )
	    (t (SETF RESULT (RESTORE-1 BACKUP-FILE-NAME :SELECTIVE-SOURCE-PATTERN RESTORE-SELECTIVE-NAME
				       :USE-ORIGINAL-HOST *USE-ORIGINAL-HOST*))
	       (IF (ERRORP RESULT)
		   (PROGN ()
			  (FORMAT (SEND SELF :OUT-STREAM) "~&File not restored because ~A~%"
				  (SEND RESULT :REPORT-STRING))
			  (SETF FINI :MUTE))
		   (SETQ FINI :COMPLETE))) ) ;1; cond*
      (UNLESS RESULT	;1;* 1fixed* erroneous "restore-file complete" message.
	(FORMAT (SEND SELF :OUT-STREAM)  "~%->[File not restored] ~%")
	(SETF FINI :MUTE)) )	;1; cond, when doit*
    FINI))


(DEFMETHOD (tape-mixin :list-carry-tape) ()
  "2List* files 2on* a Carry tape."
  (let ((fini :ABORTED))
    (send self :display-msg "Listing contents of Carry tape ")
    (IF (mt:list-carry-tape)	;1; ADDED THIS CHECK*
	(setq fini :COMPLETE)
	(TERPRI) )
    fini))

(DEFMETHOD (tape-mixin :restore-carry-tape) ()
  "Restore files from a Carry tape."
  (let ((fini :ABORTED))
    (send self :display-msg "Restoring Carry tape")
    (if (mt:restore-carry-tape)
	(setq fini :COMPLETE)
	(TERPRI) )
    fini))

(DEFMETHOD (tape-mixin :make-carry-tape) ()
  "Backup files to tape using carry-format."
  (let ((fini :ABORTED))
    (send self :display-msg "Building a Carry tape")
    (if (eq (mt:make-carry-tape) :ABORTED)
	(TERPRI)
	(setq fini :COMPLETE))
    fini))


;;; Begin Tar interface; Make, Restore, List.  8.11.87
;;;
(DEFVAR restore-tar-pathname ())
(DEFVAR unspecified-type-action ())
(DEFVAR query-mode ())
(DEFVAR source-pattern ())
(DEFVAR create-directories ())
(DEFCONSTANT doc-restore-tar-pathname "Specify the destination pathname.  Wildcards are allowed and will be merged with the name(s) read from tape.")
(DEFCONSTANT doc-restore-source-pattern
	     "Enter NIL to restore all files on tape.  To restore only certain files enter the name and type, seperated by a period.  Only files on tape that
match will be restored.  Wildcards are allowed.  For example a*.c would cause all files that started with a and have a type of c to be restored.")
(DEFCONSTANT doc-restore-query-mode "Select YES to be queried about restoring each file.")
(DEFCONSTANT doc-restore-unspecified-type-action "Select YES to restore any files with an unspecified type as character files.  Select NO to restore them as binary files.
Select QUERY to be asked on restoring each such file as either a binary or a character file.")
(DEFCONSTANT doc-restore-create-directories "Select YES to automatically create directories if they do not exist.  Select QUERY to be asked about creating non-existent directories.")
(DEFVAR tar-verbose T)				;8.13.87
(DEFCONSTANT doc-tar-verbose "Select YES to list additional information about each file.")
(DEFVAR unix-listing ())
(DEFCONSTANT doc-list-tar-unix "Select YES to list files in a UNIX format.  Select No to list them in an Explorer format.")
(DEFCONSTANT doc-make-tar-unspecified-type-action "Select YES to backup any files with an unspecified type as character files.  Select NO to backup them as binary files with a byte-size of 8.
Select QUERY to be asked on backing up each such file as either a binary or a character file.")

(DEFMETHOD (tape-mixin :list-tar-tape) ()  ;8.11.87
  "List files on a Tar tape."
  (let (doit (fini :ABORTED) (backup-window self))
    (declare (special backup-window))
    (Catch 'L-T-T-DOIT
      (EVAL
	`(Tv:Choose-Variable-Values
	   '((tar-verbose "Verbose?" :documentation ,doc-tar-verbose :boolean)
	     (unix-listing "Unix Style Listing?" :documentation ,doc-list-tar-unix :boolean)
	     )
	   :width 80. :margin-choices
	   '(("Do it" (And (Setq Doit T) (Throw 'L-T-T-DOIT ())))
	     "Abort")
	   :label "List Tar Tape" :Superior Backup-window)))
    (WHEN doit
      (send self :display-msg "Listing contents of Tar tape ")
      (cond (*remote-tape-p*
	     (and (tape-service-req *remote-hostname* $LIST-TAR-TAPE)
		  (SETQ FINI :COMPLETE)) )
	    (t (list-tar-tape :verbose tar-verbose :unix-listing unix-listing)
	       (setq fini :COMPLETE)))
      (FORMAT (SEND SELF :OUT-STREAM) "~&"))
    fini))

(DEFMETHOD (tape-mixin :restore-tar-tape) ()	   ;8.11.87
  "Restore files from a Tar tape."
  (unless restore-tar-pathname
    (SETF restore-tar-pathname (guess-directory)))   ;(PATHNAME "lm:~;*.*#>") 10-10-88 DAB
  (UNLESS (member unspecified-type-action '(:yes :no :query))
    (SETF unspecified-type-action :no))
  (UNLESS (member create-directories '(:yes :no :query))
    (SETF create-directories :yes))
  (LET (doit (fini :ABORTED)  (backup-window self))
    (declare (special backup-window))
    (setf Tv::*Default-Directory-Pathname* (guess-directory))    ;(PATHNAME "lm:~;*.*#>") 10-10-88 DAB
    (Catch 'R-T-T-DOIT
      (EVAL
	`(Tv:Choose-Variable-Values
	   '((restore-tar-pathname "Destination Pathname" :documentation ,doc-restore-tar-pathname
				   :directory-pathname)
	     (source-pattern "Source Pattern to Match" :documentation ,doc-restore-source-pattern  :string) 
	     (query-mode "Query on restoring each file?" :documentation ,doc-restore-query-mode :boolean)
	     (create-directories "Create directories automatically?"
				 :documentation ,doc-restore-create-directories
				 :choose (:yes :no :query))    
	     (unspecified-type-action "Characters?" :documentation ,doc-restore-unspecified-type-action
				      :choose (:yes :no :query)))
	   :width 80.
	   :MARGIN-CHOICES
	   '(("Do it" (And (Setq Doit T) (Throw 'R-T-T-DOIT ())))
	     "Abort")
	   :label "Restore Tar Tape" :Superior Backup-window)))
    (WHEN doit
      (send self :display-msg "Restoring Tar tape")
      (cond (*remote-tape-p*
	     (and (tape-service-req *remote-hostname* $RESTORE-TAR-TAPE)
		  (SETQ FINI :COMPLETE)) )
	    (t (mt:restore-tar-tape restore-tar-pathname
				    :unspecified-type-action
				    (case unspecified-type-action
				      (:yes t)
				      (:no nil)
				      (:query :query))
				    :query-mode query-mode
				    :source-pattern source-pattern
				    :create-directories
				    (CASE create-directories
				      (:yes t)
				      (:no nil)
				      (:query :query)))
	       (setq fini :COMPLETE)) )
      (FORMAT (SEND SELF :OUT-STREAM) "~&"))	; when doit
    fini))

(DEFVAR source-pathname ())
(DEFCONSTANT doc-make-tar-pathname
	     "Enter the pathname of the files to be copied to tape.  Wildcards are acceptable.")
(DEFVAR sub-directories ())
(DEFCONSTANT doc-make-tar-sub-dir "Select YES to copy subdirectories as well.")
(DEFVAR working-directory :RELATIVE) ;10-08-87 DAB
(DEFCONSTANT doc-make-tar-working-dir ;10-08-87 DAB
	     "Enter ABSOLUTE to make the name put on tape be an absolute directory.
              Enter RELATIVE to make the name put on tape be a relative directory.")  


(defvar tar-blocking-factor 20.)		;8.13.87
(defconstant doc-tar-blocking-factor
	     "Blocking-Factor specifies the maximum number of 1024 byte blocks to write when using 1/2 inch tape in variable block mode.")
(DEFVAR write-eof t)                            ;8.17.87
(DEFCONSTANT doc-tar-write-eof "Write an EOF on the tape.  Some Unix machines require an EOF at the end of 1/4 inch Tar tapes.")

(DEFMETHOD (tape-mixin :make-tar-tape) ()  ;8.11.87
  "Backup files to tape using Tar format."
  (declare (special MAX-BLOCKS-PER-RQB ))
  (unless source-pathname
    (SETF source-pathname (guess-directory)))   ;(PATHNAME "lm:~;*.*#>") 10-10-88 DAB
  (let ((fini :ABORTED)
	(backup-window self)
	doit)
    (declare (special backup-window)) 
    (setf Tv::*Default-Directory-pathname* (guess-directory))     ;(PATHNAME "lm:~;*.*#>") 10-10-88 DAB 
    (unless (member working-directory '(:absolute :relative))	   ;09-29-87 DAB
      (SETF working-directory :relative))
    (UNLESS (member unspecified-type-action '(:yes :no :query))
      (SETF unspecified-type-action :no))
    
    (Catch 'M-T-T-DOIT
      (EVAL
	`(Tv:Choose-Variable-Values
	   '((SOURCE-PATHNAME "Source Pathname" :documentation ,doc-make-tar-pathname :directory-pathname)
	     (sub-directories "Subdirectories also?" :documentation ,doc-make-tar-sub-dir :boolean)
	     (tar-verbose "Verbose?" :documentation ,doc-tar-verbose :boolean)
	     (tar-blocking-factor "Blocking Factor" :documentation ,doc-tar-blocking-factor :Fixnum)	   ;8.13.87
	     (working-directory "Working Directory Format" :documentation ,doc-make-tar-working-dir
				:choose (:ABSOLUTE :RELATIVE))	   ;10-08-87  DAB
	     (unspecified-type-action "Characters?" :documentation ,doc-make-tar-unspecified-type-action
				      :choose (:yes :query :no ))   ;04-25-88 DAB
	     (write-eof "Write an EOF?" :documentation ,doc-tar-write-eof :boolean))	   ; 8.17.87
	   :width 80. :margin-choices
	   '(("Do it" (And (Setq Doit T) (Throw 'M-T-T-DOIT ())))
	     "Abort")
	   :label "Make Tar Tape" :Superior Backup-window)))
    (UNLESS (tar-check-working-directory (SEND (PATHNAME working-directory) :directory)
					 (SEND (PATHNAME source-pathname) :directory))	   ;8.20.87
      (SETQ doit nil)
      (FORMAT t
	      "~&The Working Directory, ~a, is not within the hierarchy of the Source~&Pathname directory, ~a"
	      (PATHNAME working-directory) (PATHNAME source-pathname))
      (FORMAT t "~&Make Tar Tape Aborted~%"))
    (WHEN doit
      (send self :display-msg "Building a Tar tape")
      (cond (*remote-tape-p*
	     (and (tape-service-req *remote-hostname* $MAKE-TAR-TAPE)
		  (SETQ FINI :COMPLETE)) )
	    (t (mt:make-tar-tape source-pathname :sub-directories sub-directories :verbose tar-verbose
				 :working-directory working-directory :write-eof write-eof
				 :blocking-factor
				 (IF (eq (get 'mt:*current-unit* 'block-mode) :variable)
				     tar-blocking-factor
				     MAX-blocks-PER-RQB)
				 :unspecified-type-action  ;04-25-88
				 (case unspecified-type-action
				   (:yes t)
				   (:no nil)
				   (:query :query))
				 ) 
	       (setq fini :COMPLETE)) )
      (FORMAT (SEND SELF :OUT-STREAM) "~&"))	   ;when doit
    fini))




(Defmethod (Tape-Mixin :Restore-Directory) ()
  "Restore Directory to tape."
  (LET ((FINI :ABORTED)
	(backup-window self)
	result n-files)
    (declare (special backup-window))
    (Setq Doit ())
    (Unless Backup-Directory-Name
      (Setf Backup-Directory-Name (Guess-Directory)))
    (Unless restore-selective-name
      (Setf restore-selective-name (Guess-Directory)))
    (setf Tv::*Default-Directory-Pathname* Backup-Directory-Name)
    (Catch 'R-D-Doit
      (Eval
	`(Tv:Choose-Variable-Values
	   '((Restore-Selective-Name "Source Pathname" :Documentation ,Doc-Rmf-Sel
				     :directory-Pathname)
	     (Backup-Directory-Name "Destination Pathname" :Documentation ,Doc-Rmf-Dir
				    :directory-Pathname)
	     (*Use-Original-Host* "Use host name from tape?" :Documentation ,Doc-Rmf-original-host :Boolean)
	     (Query-P "Query?" :Documentation ,Doc-Rmf-Query :Boolean)
	     (create-directories "Create directories automatically?" :documentation
				 ,doc-rd-create-directories :boolean))
	   :Width 80. :Margin-Choices
	   '(("Do it" (And (Setq Doit T) (Throw 'R-D-Doit ())))
	     "Abort")
	   :Label "Restore Directory" :Superior Backup-window)))
    (When Doit
      (SEND SELF :DISPLAY-MSG "Restoring Directory")
      (cond (*remote-tape-p*
	     (setf n-files (or (tape-service-req *remote-hostname* $RESTORE-DIRE)
			       0)) )	;; NIL means zero file
	    (t (setq result (Restore Backup-Directory-Name
				       :Selective-Source-Pattern Restore-Selective-Name
				       :Selective Query-P
				       :create-directories (if create-directories T :query)
				       :use-original-host *Use-Original-Host*))
	       (setq n-files (length result))) ) ;1; cond*
      (Format (Send Self ':Out-Stream) "~&~:[~d~;Zero~*~] file~:1*~P restored.~%"
	      (zerop n-files)  n-files)
      (setf fini (if (zerop n-files) :mute :Complete)) ) ;1; when doit*
    Fini)) 



(DEFMETHOD (tape-mixin :restore-partition) ()
  "Restore a partition to tape."
  (LET ((FINI :ABORTED)
	(backup-window self)
	part-name )
    (declare (special backup-window))
    (setq doit nil)
    (catch 'r-p-doit
      (eval `(tv:choose-variable-values 
	       '((partition-unit "Unit number"  :documentation ,doc-rp-unit)
		 (partition-band "Partition number or name"  :string :documentation ,doc-rp-part))
	       :width 50
	       :margin-choices '(("Do it" (and (setq doit t) (throw 'r-p-doit nil))) "Abort")
	       :label "Restore Partition"
	       :SUPERIOR BACKUP-WINDOW)))
    (when doit
      (SETQ part-name PARTITION-BAND)
      (AND (EQ (LENGTH part-name) 1)
	   (MEMBER part-name *partition-num-list* :test #'string=)
	   (SETQ PARTITION-BAND (FORMAT () "~a~d" "LOD" part-name)))
      (format (send self :out-stream) "~&Restoring partition to ~S on unit ~A~a..."
	      partition-band partition-unit (if *remote-tape-p* " from remote tape " " "))
      (unless (local-partition partition-unit)
	(beep)
	(format (send self :out-stream) "~2&** Warning: you are clobbering third party's disk on ~a !"
		partition-unit) ) ;1; unless*
      (cond (*remote-tape-p*
	     (multiple-value-bind (TO-PART-BASE PART-SIZE)
		 (if (local-partition partition-unit)
		     (SYS:FIND-DISK-PARTITION-FOR-WRITE PARTITION-BAND () PARTITION-UNIT)
		     (sys:FIND-DISK-PARTITION PARTITION-BAND () PARTITION-UNIT () T))
	       (and TO-PART-BASE	;1; yes, do it !*
		    (setq TO-PART-BASE (tape-service-req *remote-hostname* $RESTORE-PART PART-SIZE)) )
	       (if TO-PART-BASE
		   (SETQ FINI :COMPLETE)
		   (format (send self :out-stream) "~%->[Partition not restored] ~%") ) ))
	    (t (condition-case (condition)
		   (if (eq (get 'mt:*current-unit* 'tape-type) :half-inch)
		       (mt:restore-partition-half-inch-tape partition-unit partition-band)
		       (execute-restore-partition partition-unit partition-band))
		 (NOT-PARTITION
		  (PROGN
		    (FORMAT (SEND SELF :OUT-STREAM) (SEND CONDITION :REPORT-STRING))
		    (IF (EQ
			  (FQUERY
			    '(:BEEP T :CLEAR-INPUT T :FRESH-LINE NIL :CONDITION NIL :LIST-CHOICES NIL
				    :CHOICES ((:REWIND #\R) (:SPACE-TO-EOF #\S)))
			    "~&~
                                                  ~% Press R   To rewind the tape~
                                                  ~%       S   To space-to-eof~
                                                  ~%")
			  :REWIND)
			(FORMAT (SEND SELF :OUT-STREAM) "rewinding ..." (REWIND))
			(FORMAT (SEND SELF :OUT-STREAM) "space-to-eof ..." (SPACE-TO-EOF)))
		    (FORMAT (SEND SELF :OUT-STREAM) "Done~%")
		    (setq fini :mute))))
	       (setq fini :complete)) ))  ;1; cond, when doit*
    fini))

(DEFMETHOD (TAPE-MIXIN :VERIFY-FILE) ()
  "Verify a file on tape."
  (LET ((FINI :ABORTED)
	(backup-window self)
	result)
    (declare (special backup-window))
    (SETQ DOIT ())
    (UNLESS RESTORE-SELECTIVE-NAME
      (SETF RESTORE-SELECTIVE-NAME (GUESS-DIRECTORY)))
    (UNLESS BACKUP-FILE-NAME
      (SETF BACKUP-FILE-NAME (GUESS-DIRECTORY)))
    (CATCH 'V-F-DOIT
      (TV:CHOOSE-VARIABLE-VALUES
	`((RESTORE-SELECTIVE-NAME "Source Pathname" :DOCUMENTATION ,DOC-RMF-SEL
				  :directory-PATHNAME RESTORE-SELECTIVE-NAME)
	  (BACKUP-FILE-NAME "Destination Pathname" :DOCUMENTATION ,DOC-VF-PATH
			    :directory-PATHNAME BACKUP-FILE-NAME)
	  (*USE-ORIGINAL-HOST* "Use host name from tape?" :DOCUMENTATION ,DOC-VMF-ORIGINAL-HOST
			       :BOOLEAN))
	:WIDTH 80 :MARGIN-CHOICES '(("Do it" (AND (SETQ DOIT T) (THROW 'V-F-DOIT ()))) "Abort")
	:LABEL "Verify File" :SUPERIOR BACKUP-WINDOW))
    (WHEN DOIT
      (FORMAT (SEND SELF :OUT-STREAM) "~&Verifying file with ~A" BACKUP-FILE-NAME)
      (cond (*remote-tape-p*
	     (setq result (tape-service-req *remote-hostname* $VERIFY-FILE)) )
	    (t (setq result (VERIFY-1 BACKUP-FILE-NAME () :SELECTIVE-SOURCE-PATTERN RESTORE-SELECTIVE-NAME
				      :USE-ORIGINAL-HOST *USE-ORIGINAL-HOST*))) )
      (IF result
	  (SETQ FINI :COMPLETE)
	  (FORMAT (SEND SELF :OUT-STREAM)
		  ;mp Display the following message when tape is positioned incorrectly.
		  "~%->[File not verified] ~%")) ) ;1; when doit*
    FINI)) 


(DEFMETHOD (TAPE-MIXIN :VERIFY-DIRECTORY) ()
  "Verify Directory on tape until extra EOF is encountered."
  (LET ((FINI :ABORTED)
	(backup-window self)
	result)
    (declare (special backup-window))
    (SETQ DOIT ())
    (UNLESS RESTORE-SELECTIVE-NAME
      (SETF RESTORE-SELECTIVE-NAME (GUESS-DIRECTORY :WILD)))
    (UNLESS BACKUP-DIRECTORY-NAME
      (SETF BACKUP-DIRECTORY-NAME (GUESS-DIRECTORY :WILD)))
    (CATCH 'VMF-DOIT
      (TV:CHOOSE-VARIABLE-VALUES
	`((RESTORE-SELECTIVE-NAME "Source Pathname" :DOCUMENTATION ,DOC-RMF-SEL
				  :directory-PATHNAME RESTORE-SELECTIVE-NAME)
	  (BACKUP-DIRECTORY-NAME "Destination Pathname" :DOCUMENTATION ,DOC-VF-PATH
				 :directory-PATHNAME BACKUP-DIRECTORY-NAME)
	  (*USE-ORIGINAL-HOST* "Use host name from tape?" :DOCUMENTATION ,DOC-VMF-ORIGINAL-HOST
			       :BOOLEAN))
	:WIDTH 80 :MARGIN-CHOICES '(("Do it" (AND (SETQ DOIT T) (THROW 'VMF-DOIT ()))) "Abort")
	:LABEL "Verify Directory" :SUPERIOR BACKUP-WINDOW))
    (WHEN DOIT
      (FORMAT (SEND SELF :OUT-STREAM) "~&Verifying Directory with ~A" BACKUP-DIRECTORY-NAME)
      (cond (*remote-tape-p*
	     (if (probe-directory BACKUP-FILE-NAME)
		 (setq result (tape-service-req *remote-hostname*  $VERIFY-dire))
		 (FORMAT *standard-output* "~%  Directory ~s -> [Does not exist] " BACKUP-DIRECTORY-NAME)) )
	    (t (setq result (VERIFY BACKUP-DIRECTORY-NAME () :SELECTIVE-SOURCE-PATTERN RESTORE-SELECTIVE-NAME
			   :USE-ORIGINAL-HOST *USE-ORIGINAL-HOST*))  ) ) ;1; cond*
      (if result
	  (FORMAT (SEND SELF :OUT-STREAM) "~&Verify Directory completed SUCCESSFULLY.~%")
	  (FORMAT (SEND SELF :OUT-STREAM) "~&Verify Directory FAILED.~%"))
      (SETQ FINI :MUTE))
    FINI))
 


(DEFMETHOD (tape-mixin :verify-partition) ()
  "Verify a partition on tape."
  (LET ((FINI :ABORTED)
	(backup-window self)
	part-name)
    (declare (special backup-window))
    (setq doit nil)
    (catch 'v-p-doit
      (eval `(tv:choose-variable-values 
	       '((partition-unit "Unit number" :documentation ,doc-vp-unit)
		 (partition-band "Partition number or name" :string :documentation ,doc-vp-part))
	       :width 50
	       :margin-choices '(("Do it" (and (setq doit t) (throw 'v-p-doit nil))) "Abort" )
	       :label "Verify Partition"
	       :SUPERIOR BACKUP-WINDOW)))
    (unless partition-band
      (format (send self ':out-stream) "~&No partition named ~A on unit ~A.~%" partition-band partition-unit)
      (setq doit ()) )
    (when doit
      (SETQ part-name PARTITION-BAND)
      (IF (AND (EQ (LENGTH part-name) 1) (MEMBER part-name *partition-num-list* :test #'string=))
	  (SETQ PARTITION-BAND (FORMAT () "~a~d" "LOD" part-name)))
      (format (send self ':out-stream) "~&Verifying partition ~A on unit ~A ..." partition-band partition-unit)
      (cond (*remote-tape-p*
	     (if (tape-service-req *remote-hostname*  $VERIFY-part)
		 (SETQ FINI :COMPLETE)
		 (print-msg "Verify partition failed" t)) )
	    (t (condition-case (condition)
		   (if (eq (get 'mt:*current-unit* 'tape-type) :half-inch)
		       (mt:verify-partition-half-inch-tape partition-unit partition-band)
		       (MT:execute-verify-partition partition-unit partition-band))   ;04-06-88 DAB
		 (NOT-PARTITION
		  (PROGN
		    (FORMAT (SEND SELF :OUT-STREAM) (SEND CONDITION :REPORT-STRING))
		    (IF (EQ
			  (FQUERY
			    '(:BEEP T :CLEAR-INPUT T :FRESH-LINE NIL :CONDITION NIL :LIST-CHOICES NIL
				    :CHOICES ((:REWIND #\R) (:SPACE-TO-EOF #\S)))
			    "~&~
                                                  ~% Press R   To rewind the tape~
                                                  ~%       S   To space-to-eof~
                                                  ~%")
			  :REWIND)
			(FORMAT (SEND SELF :OUT-STREAM) "rewinding ..." (REWIND))
			(FORMAT (SEND SELF :OUT-STREAM) "space-to-eof ..." (SPACE-TO-EOF)))
		    (FORMAT (SEND SELF :OUT-STREAM) "Done~%")
		    (setq fini :mute))))
	       (setq fini :complete)) ))
    fini))

(DEFMETHOD (tape-mixin :list-contents) () ;1; tung done*
  "List contents of tape until end of recorded media is encountered."
  (let ((fini :ABORTED))
    (send self :display-msg "Listing contents")
    (cond (*remote-tape-p*
	   (and (tape-service-req *remote-hostname* $LIST-CONTENTS)
		(SETQ FINI :COMPLETE)) )
	  (t (condition-case (condition)
		 (MT:list-contents nil :query t)
	       (mt:unknown-tape-header (format (send self :out-stream) (send condition :report-string))) )
	     (setq fini :COMPLETE) ))  ;1; cond*
    (format (send self ':out-stream) "~&")
    fini))

(DEFMETHOD (tape-mixin :list-reel-tape) ()
  "2List* files 2on* a Reel tape."
  (let ((fini :ABORTED))
    (send self :display-msg "Listing contents of Reel tape ")
    (cond (*remote-tape-p*
	   (when (tape-service-req *remote-hostname*  $LIST-DIRECTORY)
	     (terpri)
	     (SETQ FINI :COMPLETE)) )
	  (t (condition-case (condition)
		 (mt:list-contents nil :stop-at-double-eof t)
	       (mt:tape-media-error
		(format (send self ':out-stream)
			"~&End of data or media error on tape unit ~d." mt:*current-unit*)))
					   ;(FORMAT (SEND SELF :OUT-STREAM) "~&")
	     (setq fini :COMPLETE) ))
    fini))

(DEFMETHOD (TAPE-MIXIN :LIST-DIRECTORY) ()
  "List contents of tape until double EOF is encountered."
  (LET ((FINI :ABORTED))
    (SEND SELF :DISPLAY-MSG "Listing directory")
    (cond (*remote-tape-p*
	   (and (tape-service-req *remote-hostname* $LIST-DIRECTORY)
		(SETQ FINI :COMPLETE)) )
	  (t (LIST-CONTENTS () :STOP-AT-DOUBLE-EOF T)
	     (SETQ FINI :COMPLETE)) ) ;1; cond*
    (FORMAT (SEND SELF :OUT-STREAM) "~&")
    FINI)) 


;;;
;;; Access the BOOTABLE support.
;;;


(DEFMETHOD (TAPE-MIXIN :RESTORE-BOOTABLE-TAPE) ()
  "Restore partitions from a boot tape."
  (IF (RESTORE-BOOTABLE-TAPE (SEND SELF :OUT-STREAM))
      :COMPLETE
      :MUTE))


(DEFMETHOD (tape-mixin :make-bootable-tape) ()
  "Restore partitions from a boot tape."
  (if (make-bootable-tape (send self ':out-stream))
      :COMPLETE
      :fini))
;;;
;;; The next couple methods support operations not yet implemented. 3/19/85
;;;

(DEFMETHOD (TAPE-MIXIN :LOAD-DISTRIBUTION-TAPE) ()
  "Loads distribution tape on to local file system."
  (LET ((FINI :ABORTED))		   ;    (send self :display-msg "Not yet implemented.")
    (LOAD-DISTRIBUTION-TAPE (SEND SELF :OUT-STREAM))
    (SETQ FINI :COMPLETE)
    FINI)) 



(DEFMETHOD (TAPE-MIXIN :WRITE-EOF) () ;1; tung done*
  "Write EOF method"
  (LET ((FINI :ABORTED))
    (SEND SELF :DISPLAY-MSG "Writing EOF")
    (cond (*remote-tape-p*
	   (and (tape-service-req *remote-hostname* $WRITE-EOF)
		(SETQ FINI :COMPLETE)) )
	  (t (WRITE-EOF)) )
    (SETQ FINI :COMPLETE)
    FINI))
 


(DEFMETHOD (TAPE-MIXIN :EXIT) ()
  "Exit program method"
  (TV:DESELECT-AND-MAYBE-BURY-WINDOW SELF)
  :MUTE)


(defun show-help-basic-help () 
  "basic help"
  (LET ((OUT-STREAM (SEND SELF :OUT-STREAM)))
    (format out-stream
	    "~%
                                 BACKUP SYSTEM HELP
                                    -BASIC HELP-


Several steps must be taken before using any tape operation:

     * INSERT TAPE--The tape to be used must be inserted into the tape
            drive and the drive lever must then be closed.

     * PREPARE TAPE--The Prepare Tape command from the Backup Menu must then be selected
            to choose the appropriate tape mode from the pop-up menu.
   

After performing these operations, you can select the appropriate type of backup or 
restore operation that you want.



WORDS OF CAUTION concerning tape usage:

     * Tape can be appended only at the end of recorded media.

     * Data on tape can be overwritten only by writting on the beginning of the tape. 

     * You cannot backup a file or logical backup over multiple tapes.  Each logical
       backup must fit on only one tape.

     * Before verifying or restoring data, the tape must be explicitly positioned
       before the data to be verified or restored.

     * You should rewind the tape before removing it from the drive.  This positions
       blank leader tape under the access door of the catridge and helps prevent damage 
       to recorded media.~%")))   


(DEFUN SHOW-HELP-PARTITIONS ()
  (LET ((OUT-STREAM (SEND SELF :OUT-STREAM)))
    (FORMAT OUT-STREAM
	    "~%
                                 BACKUP SYSTEM HELP
                                    -PARTITIONS-


BACKING UP and RESTORING partitions are accomplished by using the partition
commands displayed on the right of this screen.

   BACKING UP PARTITIONS -- Select the BACKUP PARTITION command, then fill in the 
     appropriate information in the pop-up menu.  Verify the backup by first
     positioning the tape to the partition.  The correct position can be
     determined when the LIST CONTENTS command lists the partition name as the
     first entry.  After positioning the tape correctly, use VERIFY PARTITION.

   RESTORING PARTITIONS -- Position the tape immediately before the partition;
     then select RESTORE-PARTITION.  To verify the partition, use the steps
     prescribed above.~%")))


(DEFUN SHOW-HELP-STREAMING ()
  (LET ((OUT-STREAM (SEND SELF :OUT-STREAM)))
    (FORMAT OUT-STREAM
	    "~%
                                 BACKUP SYSTEM HELP
                                    -STREAMING-


Streaming is an efficient mode of writing and reading the tape.  It is
characterized by small intertape gaps (more capacity per tape)
and by a lack of stopping and starting while reading or writing the tape.~%")))



(DEFUN SHOW-HELP-READ-WRITE ()
  
  (LET ((OUT-STREAM (SEND SELF :OUT-STREAM)))
    (FORMAT OUT-STREAM
	    "~%
                                 BACKUP SYSTEM HELP
                                    -READ-WRITE-


Reads and writes cannot be intermixed on the tape drive.

To change from WRITE TO READ:
    * Select REWIND

To change from READ TO WRITE either:

    * Start tape at the beginning
       -Select REWIND

    * Add the new data to the end of existing data on the tape
       -Select PREPARE TO APPEND

The tape hardware recognizes the end of recorded media by a large erased area on the
tape.  This erased area is automatically provided during write   operations.  

The only way to overwrite data on a tape is to rewind the tape  and  start  overwriting at its 
beginning.  All data previously  recorded  on the  tape  will become inaccessible, no
matter how little is written.~%"))) 

(DEFUN SHOW-HELP-BOOTABLE-TAPE ()
  
  (LET ((OUT-STREAM (SEND SELF :OUT-STREAM)))
    (FORMAT OUT-STREAM
	    "~%

                                 BACKUP SYSTEM HELP
                                  -BOOTABLE-TAPE-


A bootable tape can be made by copying a bootable partition from disk
and copying it to the tape.

The tape must be positioned at the beginning before a partition can be
copied.  Select the MAKE BOOTABLE TAPE command and then select the partition
to be copied from the pop-up menu by clicking Mouse-Left.

A bootable tape is formatted so that you can boot from the tape the
same way you can boot from a disk.  Note that you cannot boot Lisp from
the tape.  If your load band or microcode band had been destroyed and you
cannot boot Lisp, you should use the diagnostics utility to rebuild and/or
restore your disk.~%")))

(defun show-help-carry-tape ()		   ;mrr
  (LET ((OUT-STREAM (SEND SELF :OUT-STREAM)))
    (FORMAT OUT-STREAM
	    "~%

                                 BACKUP SYSTEM HELP
                                    -CARRY-TAPE-


The Carry format is intended as a means to move character files between various 
vendor's Lisp machines.  Carry tapes written on the Explorer can be restored on 
the Symbolics 3640 or 3670 and newer models, but not the 3600 due to incompatibilities
in the tape drive hardware.

Carry tapes can be in one of three formats: QIC-11 (4 track),QIC-24, and 6250.  
The 6250 format is also known as QIC-11 9-track.

The Explorer can restore files from tapes written in QIC-24, QIC-11, or 6250 formats,
but the Explorer can only backup files using either the QIC-24 or 6250 formats.


   Backing up a Carry tape -- Explorer character files can be backed up in Carry tape
format by using the Make Carry Tape command.  Binary files cannot be backed up to a 
Carry tape.  The prompts are similar to those of the Backup File command.


   Restoring a Carry tape -- Symbolics character files can be restored to an Explorer
disk by using the Restore Carry Tape command.  Binary files cannot be restored from a
Carry tape.  The prompts are similar to those of the Restore File command.


   Listing contents of a Carry tape -- The contents of a Carry tape can be listed by
using the List Carry Tape Contents command.~%")))


(DEFUN SHOW-HELP-LOAD-DISTRIBUTION-TAPE ()
  
  (LET ((OUT-STREAM (SEND SELF :OUT-STREAM)))
    (FORMAT OUT-STREAM
	    "~%

                                 BACKUP SYSTEM HELP
                              -LOAD-DISTRIBUTION-TAPE-



The Load Distribution Tape command loads a utility from tape.  This command is the first
step required to install a distribution tape.  After this command is successfully completed,
you should provide a newly initialized Lisp environment and then execute the Install-Program
function.~%")))


(DEFMETHOD (TAPE-MIXIN :ERRORS) ()
  "Errors help method"
  (FORMAT (SEND SELF :OUT-STREAM)
	  "~%
 ERRORS HELP:    The following errors may require specific actions.  Usually just press
  ABORT and take the appropriate action.  
   
  [Unit NIL is an invalid value occurs] if a Prepare Tape has not been performed 
  before any other tape operation.  Mouse the Prepare Tape command.  

  [Unknown tape header format] indicates that data was read when a header was
  expected.  Position the tape to a valid header by Position To Eof or with a Rewind.

  [Unknown error] will occur if a command is issued just after the tape has been
  inserted in the drive.  During the few seconds it takes for the drive  to position
  the tape, commands are returned with this error.  Press  ABORT  and  re-issue  the
  command.

  [Illegal command sequence or Invalid parameter] means that an illegal  command 
  sequence was issued.  Switching  from  reads to  writes  can  only  be done at the 
  beginning of tape (do a rewind), or at the end of recorded media (prepare to append).

  [Formatter Not Responding] may indicate  that  the  mass storage cabling is not
  correct, that the tape mass storage enclosure's power is not on, or that the tape
  drive failed to  pass self test.  To correct  the  last,  rebooting  the  Explorer
  is recommended.~%")
  :MUTE)  

(DEFMETHOD (TAPE-MIXIN :help-tar-tape) ()	;8.12.87
  "Tar tape help method"
  (FORMAT (SEND SELF :OUT-STREAM)
	  "~%

                                 BACKUP SYSTEM HELP
                                     -TAR TAPE-


  The Tar tape commands may be accessed by clicking on Prepare Tape and
selecting one of the Tar Format choices.
  
  The Tar format is designed to allow the interchange of files between any
system that supports Tar tapes and an Explorer.  Quarter inch cartridge is
supported in QIC-24, QIC-11, and 6250 formats.  Half inch is supported in 
single (1600 bpi) and double (3200 bpi) density.   Half inch can be used 
in variable or fixed block modes. 

  Making a Tar tape -- Files can be put on a tape in Tar format by using the 
Make Tar Tape command.  Optionally, a working directory may be specified.
This will make the file names on the tape be relative to the working directory.
To seperate multiple Tars on a single tape you should write a EOF between each.

  Restoring a Tar tape -- Unix files can be restored on an Explorer by using the
Restore Tar Tape command.  All files or just files matching a specified pattern
may be restored.  The pattern matches on the name and type of the file only.
Restore Tar Tape will stop if an EOF is encountered or if an extended block of
nulls is read after any file.

  Listing contents of a Tar tape -- The contents of a Tar tape can be listed by
using the List Tar Tape command.~%")
  :MUTE)



(DEFVAR HELP-CHOOSE-MENU ()) 

(DEFMETHOD (TAPE-MIXIN :HELP) ()
  "Help method"
  (send self :set-more-p t)
  (FORMAT (SEND SELF :OUT-STREAM) "~%")
  (CASE (SEND SELF :HELP-MENU)
    (:help-basic-help (show-help-basic-help))
    (:HELP-READ-WRITE (SHOW-HELP-READ-WRITE))
    (:HELP-STREAMING (SHOW-HELP-STREAMING))
    (:HELP-PARTITIONS (SHOW-HELP-PARTITIONS))
    (:HELP-ERRORS (SEND SELF :ERRORS))
    (:HELP-TAR-TAPE (SEND SELF :HELP-TAR-TAPE))
    (:HELP-BOOTABLE-TAPE (SHOW-HELP-BOOTABLE-TAPE))
    (:help-carry-tape (show-help-carry-tape))
    (:HELP-LOAD-DISTRIBUTION-TAPE (SHOW-HELP-LOAD-DISTRIBUTION-TAPE))
    (OTHERWISE NIL))
  (send self :set-more-p ())
  :MUTE)


(DEFMETHOD (TAPE-MIXIN :HELP-MENU) ()
  (SETQ HELP-CHOOSE-MENU
	(TV:MAKE-WINDOW 'TV:MOMENTARY-MENU :LABEL "Help Topics" :BORDERS 3 :ITEM-LIST
			'(("Basic-Help" :value :help-basic-help)
			  ("Read-Write Mode" :VALUE :HELP-READ-WRITE)
			  ("Streaming" :VALUE :HELP-STREAMING)
			  ("Errors" :VALUE :HELP-ERRORS)
			  ("Partitions" :VALUE :HELP-PARTITIONS)
			  ("Bootable tape" :VALUE :HELP-BOOTABLE-TAPE)
			  ("Carry Tape" :value :help-carry-tape)
			  ("Tar Tape" :value :help-tar-tape)
			  ("Load Distribution Tape" :VALUE :HELP-LOAD-DISTRIBUTION-TAPE))))  
  (SEND HELP-CHOOSE-MENU :CHOOSE))

(DEFMETHOD (TAPE-MIXIN :VERIFY-BOOTABLE-TAPE) ()
  "Verify partitions from a boot tape."
  (IF (VERIFY-BOOTABLE-TAPE (SEND SELF :OUT-STREAM))
      :COMPLETE
      :MUTE))

(defun disk-unit-error (From-PART unit)
  "return t if disk-read error"
;  (setq unit (sys:DECODE-UNIT-ARGUMENT unit "receive-band"))
;  (COND ((NUMBERP UNIT)
;	 (NUPI-DISKP UNIT) )
;	(t (FUNCALL UNIT :READ () 0 0))) )
  (let ((Remote-UNIT (sys:DECODE-UNIT-ARGUMENT unit "receive-band"))
	ret error-p)
  (multiple-value-setq (ret error-p)
    (catch-error (sys:FIND-DISK-PARTITION From-PART () Remote-UNIT) ()) )
  (format t "~%ret = ~s, error = ~s." ret error-p)
  (condition-case (condi)
      (sys:FIND-DISK-PARTITION From-PART () Remote-UNIT)
    (error ;(print condi)
	   (format t "~2%Report Message = ")
	   (send condi :report *debug-io*)
	   ;(signal-condition condi () t))
	   ))
  (and (null ret)
       error-p) ))

;;;  INITIALIZATIONS
;;;

;;; Put M in the system key list.
(defun add-backup-to-system-menu ()
  "Adds backup utility to the system menu"
  ;;Initialize remote tape varibles
  (setq *tape-unit* ())    ;05-23-88 DAB
  (setq *REMOTE-TAPE-P* ())
  (setq *REMOTE-HOSTNAME* ())
  (W:MODIFY-SYSTEM-ACCESS-SPEC 'MAGTAPE-BACKUP	   ;changed to call new function - las 
			     :ASSIGN-DEFAULTS)
  (SETF (SYS:SYSTEM-MADE-P 'MAGTAPE-BACKUP) T))

(defun initialize-system-menu-for-backup ()
  "It initializes the system menu to display the backup menu"
  (add-initialization "Backup" '(add-backup-to-system-menu) '(first))
  )

;;;(TV:ADD-SYSTEM-KEY #\B 'BACKUP-FRAME
;;;		   "Magnetic tape - Backup\\restore files and partitions with magnetic tape media."
;;;		   T) 


(COMPILE-FLAVOR-METHODS TAPE-MIXIN BACKUP-MENU-PANE BACKUP-FRAME) 

(initialize-system-menu-for-backup)

