;;; -*- Mode: common-Lisp; Package: MT; Base: 10.; Fonts: MEDFNT,HL12B,HL12BI -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; Copyright (C) 1980, Massachusetts Institute of Technology


;;; REVISIONS:
;;;
;;;	08.03.87 MBC	Correct Unit default of RAW-DUMP.
;;;     07.13.87 AB     Remove references to PROCESSOR-TYPE-CODE (old LAMBDA support).
;;;  Changed some hardcoded 1024's to sys:disk-block-byte-size.  1-8-87, -ab
;;;  Undo Display-File-Header mod.  Instead make change using *typeout-window-font*. 5-6-86
;;;  Modify Display-File-Header to accomodate MEDFNB.  Remove size-in-blocks and 
;;;      time portion of time-and-date-stamp.  4-23-86 
;;;  Add Use-Original-Host feature to support logical pathnames on restore.  4-10-86
;;;  Fix Backup-File to not do a CAR on something not a cons.  4-1-86 MBC
;;;  Restore of files that contained multiple subdirectories, used to arbitrarily replace first
;;;      directory node with  node or nodes of user supplied pathname.  Now the originally specified 
;;;      pathname is recorded in the tape header.  Also add SELECTIVE-SOURCE-PATTERN argument that
;;;      is used as matching pattern to provide name selective restoration from tape.
;;;      Save original backup specification as a string in each file's tape header.  Recover it via
;;;      :SOURCE-PATTERN method after open of tape file for read, and use it to effect desired
;;;      pathname merging.  This stuff depends on the :OPEN-PARAMETERS option.  Without this
;;;      support restores will default to ...
;;;      Changes to are interdependent with MTSTR changes, primarily :WILDCARD-MAP.  10-17-85 MBC
;;;
;;; Remove unused code; Parse-Directory, Restore-Magtape, & misc. cruft.  9-26-85 MBC
;;; New function, Signal-Invalid-File-Type, uses FERROR correctly.  9-18-85 MBC
;;; For dispose unit of band handler, do a rewind instead of space-to-eof.
;;; Have list-contents trigger on "large" (over 2 Mbyte) partitions and query user to continue. 8/13/85 MBC
;;; Fix dispose of unit to space-to-eof.  Keeps user from seeing  , not a partition message. 5/13/85 MBC
;;; Entry functions which deal only with magtape
;;; MAGTAPE-LIST-FILES -- List files on a magtape.
;;; RESTORE-MAGTAPE -- Restore whatever's on the mounted magtape.
;;;
;;; Old junk follows:
;;; MT-WRITE-FILES -- Write files to magtape.
;;; MT-WRITE-DIRECTORY -- Write an entire directory to magtape.
;;; MT-WRITE-PARTITION -- Write a partition onto a magtape.
;;;
;;; Other functions:
;;; PRINT-MAGTAPE -- Directly print contents of the magtape (for debugging).
;;;
;;; Modification history:
;;;   Make display-magtape-header use :STRING-FOR-PRINTING method; fix tab stops better. 7.30.85 MBC
;;;   New versions of BACKUP, RESTORE,  BACKUP-1, RESTORE-1, and BACKUP-FILE (does ONLY one file)
;;;     per the :WILDCARD-MAP changes.  10-24-85 MBC
;;;   Make LIST-CONTENTS keep going to END-OF-RECORDED-MEDIA.  Some option keywords provided to 
;;;    give it ability to act as old version did, stopping at double EOF (a directory).  10-25-85 MBC
;;;   Fix List-Contents to not infinite loop when large partition lets user opt to not continue. 11-14
;;;    
	
;; Allow for partitions also.

(DEFUN MAGTAPE-LIST-FILES (&OPTIONAL IGNORE &KEY STOP-AT-DOUBLE-EOF QUERY)
  "List contents of tape.  :STOP-AT-DOUBLE-EOF causes listing to stop at marked
 end of directories, otherwise :QUERY determines if listing should continue.
 Listing may be redirected by binding *STANDARD-OUTPUT*."
  (LET ((N 0)
	DONE)
    (LOOP;loop till :END-OF-RECORDED-MEDIA or STOP-AT-DOUBLE-EOF 
     (DO ((STREAM));loop on a logical directory
	 ((ERRORP
	   (SETQ STREAM (MAKE-MT-FILE-STREAM :DIRECTION :INPUT :ERROR () :UNIT *CURRENT-UNIT*)))
	  (FORMAT *STANDARD-OUTPUT* "~&")
	  (SEND STREAM :REPORT *STANDARD-OUTPUT*)
	  (IF (MEMBER :END-OF-RECORDED-MEDIA (SEND STREAM :CONDITION-NAMES) :TEST #'EQUAL)
	    (SETF DONE T)))
       (SETF N (1+ N))
       (IF (FUNCALL STREAM :GET :PARTITION)
	 (LET ((SIZE (FUNCALL STREAM :GET :SIZE)))
	   (FORMAT *STANDARD-OUTPUT* "~%~4d Partition ~A: ~S, Size ~D" N
		   (FUNCALL STREAM :GET :NAME) (FUNCALL STREAM :GET :COMMENT) SIZE)
	   (IF (AND QUERY (> SIZE 2000));Arbitrary partition size to wait for
	     (UNLESS (Y-OR-N-P "Continue listing contents")
	       (FORMAT *STANDARD-OUTPUT* "   Rewinding... ")
	       (REWIND)
	       (SETF DONE T)
	       (RETURN))))
	 (DISPLAY-FILE-HEADER STREAM *STANDARD-OUTPUT* N))
       (FUNCALL STREAM :CLOSE))
     (IF (OR STOP-AT-DOUBLE-EOF DONE)
       (RETURN))
     (IF (AND QUERY (NOT (Y-OR-N-P "Double EOF (end of directory) reached.  Continue? ")))
       (RETURN))))) 



(DEFUN DISPLAY-FILE-HEADER (STREAM OUT-STREAM N)
  "Extracts info from a mt-filehandle object and displays it on OUT-STREAM. 
   Does not move the tape."
  (FORMAT OUT-STREAM "~%~4d ~A" N (SEND STREAM :STRING-FOR-PRINTING))
  (LET* ((MT-PLIST (SEND STREAM :PLIST))
	 (LENGTH-IN-BYTES (GETF MT-PLIST :LENGTH-IN-BYTES))
	 (LENGTH-IN-BLOCKS (GETF MT-PLIST :LENGTH-IN-BLOCKS))
	 (CREATION (FUNCALL STREAM :CREATION-DATE)))
    (UNLESS LENGTH-IN-BYTES
      (SETQ LENGTH-IN-BYTES (GETF MT-PLIST :LENGTH)))
    (UNLESS LENGTH-IN-BLOCKS
      (SETQ LENGTH-IN-BLOCKS
	    (CEILING (* LENGTH-IN-BYTES (IF (GETF MT-PLIST :CHARACTERS)
					  1
					  2)) sys:disk-block-byte-size)))
    (FORMAT OUT-STREAM "~58,4T~3D ~8D (~D)~60,2T" LENGTH-IN-BLOCKS LENGTH-IN-BYTES
	    (FUNCALL STREAM :BYTE-SIZE))
    (IF (NUMBERP CREATION)
      (TIME:PRINT-UNIVERSAL-TIME CREATION OUT-STREAM)
      (FORMAT OUT-STREAM "~A" CREATION)))
  (LET ((AUTHOR (SEND STREAM :SEND-IF-HANDLES :AUTHOR)))
    (IF AUTHOR
      (FORMAT OUT-STREAM "~80,2T~A" AUTHOR)))) 

;;; Attempt to make display smaller for larger fonts.  
;;;  Just control backup window's font instead.  5-6-86
;(defun DISPLAY-FILE-HEADER (stream OUT-STREAM N)
;  "Extracts info from a mt-filehandle object and displays it on OUT-STREAM. 
;   Does not move the tape."
;  (FORMAT OUT-STREAM "~%~4d ~A" N (send stream :string-for-printing))
;  (let* ((mt-plist (send stream :plist))
;	 (length-in-bytes (get (locf mt-plist) :length-in-bytes))
;	 (creation (FUNCALL STREAM ':CREATION-DATE)))
;    (unless length-in-bytes
;      (setq length-in-bytes (get (locf mt-plist) :length)))
;    (format OUT-STREAM "~%~45,2T~8D (~D)~4,2T"
;	    length-in-bytes
;	    (FUNCALL STREAM ':BYTE-SIZE))
;    (let* ((date-string
;	    (if (numberp creation)
;		(with-output-to-string (s)
;		  (TIME:PRINT-UNIVERSAL-TIME creation s))
;		creation))
;	   (End (Position #/space date-string :test #'CHAR=)))
;      (format out-stream "~A" (subseq date-string 0 end))))
;  (let ((author (send stream ':send-if-handles ':AUTHOR)))
;    (if author (format OUT-STREAM " ~A" author))))


;;; Stuff carried over from now-defunct file, New-Names.  10-22-85 MBC

(DEFF SPACE-BLOCKS 'SPACE-FORWARD) 

(DEFF LIST-CONTENTS 'MAGTAPE-LIST-FILES) 

(DEFF RETENSION 'TENSION) 		   ;so it can be found!

;; Tape creation functions which hack EOT.


(DEFUN MT-EOT-HANDLER (&REST IGNORE)
  (THROW 'EOT
	 ())) 


(DEFUN MT-WRITE-FILES (&REST FILES)
  (LOOP FOR FILE IN FILES DOING
     (LOOP UNTIL (CATCH 'EOT
		   (COPY-FILE FILE "MT")
		   T)
	DO (SPACE-REV-TO-BOF) (WRITE-EOF) (REWIND) (BEEP)
	(FORMAT *ERROR-OUTPUT* "~&>>> MagTape reached end of tape <<<") (OFFLINE)
	(LOOP DOING (FORMAT *ERROR-OUTPUT* "~%Type [Resume] when new tape is mounted:") UNTIL
	   (FUNCALL *STANDARD-INPUT* :CLEAR-INPUT)
	   (EQl (FUNCALL *STANDARD-INPUT* :TYI) #\RESUME)
	   DOING (BEEP)))
     FINALLY (WRITE-EOF) (SPACE-REV))) 

;; MagTape Band transfer handler.

(PROCLAIM '(SPECIAL *BAND-WRITE* *BAND-PLIST* *BAND-STREAM* *abort-flag*))  ;*abort-flag* was added
                                                                            ;mp 2/17/87

(DEFUN FS::MAKE-BAND-MAGTAPE-HANDLER (*BAND-WRITE*)
  (MAKE-BAND-MAGTAPE-HANDLER *BAND-WRITE*)) 

;;AB 7-13-87.  Remove reference to PROCESSOR-TYPE-CODE (old LAMBDA support).
(DEFUN MAKE-BAND-MAGTAPE-HANDLER (*BAND-WRITE* &OPTIONAL PLIST)
  (LET ((*BAND-PLIST* `(:PARTITION T :BYTE-SIZE 16 :AUTHOR ,USER-ID))
	(*BAND-STREAM* NIL)
	(*abort-flag* nil))      ;mp 2/17/87
    (COND
      ((NULL *BAND-WRITE*)
       (SETQ *BAND-STREAM*
	     (MAKE-MT-FILE-STREAM :DIRECTION :INPUT :RECORD-SIZE
				  (* 85 sys:disk-block-byte-size)   ;how many blocks? may lose if -->
				  ;; copy-disk-partition invoked with N-BLOCKS-AT-A-TIME not equal to 85.
				  :CHARACTERS () :UNIT *CURRENT-UNIT* :PLIST PLIST))
       (when (ERRORP *BAND-STREAM*)
	 (signal-condition *band-stream*))
       (UNLESS (SEND *BAND-STREAM* :GET :PARTITION)
	 (FERROR 'NOT-PARTITION "~%The tape header is not pointing to a partition.~%"))
       (SETQ *BAND-PLIST* (FUNCALL *BAND-STREAM* :PLIST))
       (SETQ *BAND-STREAM* ())))
    (CLOSURE '(*BAND-WRITE* *BAND-PLIST* *BAND-STREAM* *abort-flag*) 'BAND-MAGTAPE-HANDLER)))  ; mp 2/17/87

;;; The following two functions were added to fix the problem we had due to aborting out of a restore or verify
;;; partition.

(defvar *inhibit-partition-space-to-eof* :query
  "It can have the following values:
    NIL -- Meaning space-to-eof when aborting out of a  restore or verify partition.
    T   -- Meaning rewind the tape.
    :query -- Meaning ask the user for one of the above actions.") 

(defun ask-user-for-space-to-eof-on-read ()
  (case *inhibit-partition-space-to-eof*
    (nil (space-to-eof))
    (:query (if (yes-or-no-p "Do you want to rewind? (No means  space-to-eof)")
	      (mt:rewind)
	      (space-to-eof)))
    ('t (mt:rewind))
    ))

;;AB 7-13-87.  Remove reference to PROCESSOR-TYPE-CODE (old LAMBDA support).
(DEFUN BAND-MAGTAPE-HANDLER (oper &optional arg1 arg2 &rest ignore)
  (declare (special  *BAND-STREAM* *CURRENT-UNIT*  *BAND-PLIST*))
  (declare (special *abort-flag*))
  (case oper
    (:READ  
      (OR *BAND-STREAM*
	  (SETQ *BAND-STREAM*
		(MAKE-MT-FILE-STREAM :DIRECTION :INPUT	   ;can only happen on chaparral
				     :RECORD-SIZE (* (SYSTEM:RQB-Nblocks arg1) sys:disk-block-byte-size) :UNIT
				     ;;supplying plist, indicates header has been passed
				     *CURRENT-UNIT* :CHARACTERS () :PLIST *BAND-PLIST*)))
      (FUNCALL *BAND-STREAM* :STRING-IN "unexpected EOF"
	       (SYSTEM:RQB-BUFFER arg1) 0
	       (* (SYSTEM:RQB-Nblocks arg1) 512)))
    (:WRITE
      (let* ((N-BLOCKS (SYSTEM:RQB-Nblocks arg1))
	    (N-HWDS (* N-BLOCKS 512))
            (BUF (SYSTEM:RQB-BUFFER arg1)))
	(OR *BAND-STREAM*
	    (SETQ *BAND-STREAM*
		  (MAKE-MT-FILE-STREAM :DIRECTION :OUTPUT :PLIST *BAND-PLIST* :RECORD-SIZE
					  (* N-BLOCKS sys:disk-block-byte-size)
					  :UNIT *CURRENT-UNIT* :CHARACTERS ())))
	(FUNCALL *BAND-STREAM* :STRING-OUT BUF 0 N-HWDS)))
    (:FORCE-LABEL-OUT
      (LET ((STYLE (GETF *BAND-PLIST* :STYLE)))
	(SETQ *BAND-STREAM*
	      (MAKE-MT-FILE-STREAM :DIRECTION :OUTPUT :PLIST *BAND-PLIST* :UNIT *CURRENT-UNIT*
				      :HEADER-STYLE (IF STYLE
							STYLE
							:MIT)
				      :CHARACTERS ()))))
    (:DISPOSE
      (COND
	(*BAND-STREAM* (FUNCALL *BAND-STREAM* :CLOSE) (SETQ *BAND-STREAM* ()))
	(*BAND-PLIST* (if *abort-flag* (ask-user-for-space-to-eof-on-read) (space-to-eof))))) ;mp 2/17/87

    (:HANDLES-LABEL T)
    (:GET (GETF *BAND-PLIST* arg1))
    (:SET-PLIST (SETQ *BAND-PLIST* arg1))
    (:PUT (SETF (GETF *BAND-PLIST* arg2) arg1))
    (:FIND-DISK-PARTITION
      (let ((TEM))
	(IF (SETQ TEM (GETF *BAND-PLIST* :NAME))
	    (SETF (GETF *BAND-PLIST* :NAME) arg1))
	(VALUES 0 (OR (GETF *BAND-PLIST* :SIZE) 1048575) ())))
     (:PARTITION-COMMENT (GETF *BAND-PLIST* :COMMENT)))) 

;;;          
;;; RESTORE - Revised versions from July '85 that'll eventually (Rel 2.0) replace old ones.
;;;    

(DEFUN RESTORE (&OPTIONAL PATHNAME &REST COPY-OPTIONS &KEY SELECTIVE-SOURCE-PATTERN (VERBOSE T)
  (USE-ORIGINAL-HOST NIL) &ALLOW-OTHER-KEYS)
  "Restore files from tape to PATHNAME until double EOF is encountered.
 Components of PATHNAME not specified are supplied from the tape.
 :SELECTIVE-SOURCE-PATTERN - restore only the tape files whose names match this pattern.
 :VERBOSE - causes listing information to be directed to *STANDARD-OUTPUT*.
 :USE-ORIGINAL-HOST - supercede Pathname's host component with the host from tape.
 Returns a list of the file objects that were restored."
  (LET* ((MT-OBJECT (PATHNAME "MT:"))
	 RETURN-VALUE)
    (UNLESS PATHNAME
      (SETF PATHNAME ""))
    (HANDLE-SELECTIVE-SOURCE-PATTERN SELECTIVE-SOURCE-PATTERN MT-OBJECT)
    (SEND MT-OBJECT :SET-USE-ORIGINAL-HOST USE-ORIGINAL-HOST)
    (CONDITION-CASE (ERROR)
       (SETF RETURN-VALUE
	     (APPLY #'FS::COPY-DIRECTORY MT-OBJECT PATHNAME :VERBOSE VERBOSE COPY-OPTIONS))
       ((:EOF :END-OF-RECORDED-MEDIA) (SETF RETURN-VALUE (GET ERROR :RETURN-VALUES))
	(IF VERBOSE
	  (FORMAT *STANDARD-OUTPUT* "~&~a" (SEND ERROR :REPORT-STRING)))))
    RETURN-VALUE)) 


(DEFUN RESTORE-1 (&OPTIONAL PATHNAME &REST COPY-OPTIONS &KEY SELECTIVE-SOURCE-PATTERN (VERBOSE T)
  (USE-ORIGINAL-HOST NIL) &ALLOW-OTHER-KEYS)
  "Copy one file from tape to PATHNAME.  Tape always traverses exactly one file.
 Components of PATHNAME not specified are supplied from the tape.
 :SELECTIVE-SOURCE-PATTERN - restore the tape file only if its name matches this pattern.
 :VERBOSE - causes listing information to be directed to *STANDARD-OUTPUT*.
 :USE-ORIGINAL-HOST - supercede Pathname's host component with the host from tape.
 Returns the file object restored."
  (LET* ((MT-OBJECT (PATHNAME "MT:"))
	 R1
	 R2
	 R3
	 R4)
    (UNLESS PATHNAME
      (SETF PATHNAME ""))
    (HANDLE-SELECTIVE-SOURCE-PATTERN SELECTIVE-SOURCE-PATTERN MT-OBJECT)
    (SEND MT-OBJECT :SET-USE-ORIGINAL-HOST USE-ORIGINAL-HOST)
    (CONDITION-CASE (ERROR)
       (MULTIPLE-VALUE-SETQ (R1 R2 R3 R4)
	 (APPLY #'COPY-FILE MT-OBJECT PATHNAME () () :VERBOSE VERBOSE COPY-OPTIONS))
       ((:EOF :END-OF-RECORDED-MEDIA)
	(IF VERBOSE
	  (FORMAT *STANDARD-OUTPUT* "~&~a" (SEND ERROR :REPORT-STRING)))))
    (IF (CONSP R3)
      (CAR R3)
      R3))) 


(DEFUN HANDLE-SELECTIVE-SOURCE-PATTERN (SELECTIVE-SOURCE-PATTERN MT-OBJECT)
  (LET* (PARSED-PATHNAME
	 MERGED-PATHNAME)
    (WHEN SELECTIVE-SOURCE-PATTERN
      (SETQ PARSED-PATHNAME (PATHNAME SELECTIVE-SOURCE-PATTERN)
	    MERGED-PATHNAME
	    (SEND PARSED-PATHNAME :NEW-PATHNAME :DEVICE
	       (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :DEVICE) :WILD) :DIRECTORY
	       (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :DIRECTORY) :WILD) :NAME
	       (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :NAME) :WILD) :TYPE
	       (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :TYPE) :WILD) :VERSION
	       (OR (SEND PARSED-PATHNAME :SEND-IF-HANDLES :VERSION) :NEWEST))))
    (SEND MT-OBJECT :SET-SELECTIVE-SOURCE-PATTERN MERGED-PATHNAME))) 


(DEFUN SIGNAL-INVALID-FILE-TYPE (STREAM)
  (FERROR 'INVALID-FILE-TYPE "~%Tape partition ~S is not a file.  Use Restore-Partition."
	  (SEND STREAM :GET :COMMENT))) 

;;;          
;;; BACKUP - Revised versions from July '85 that'll eventually (Rel 2.0) replace old ones.
;;; 


(DEFUN BACKUP-FILE (PATHNAME &REST COPY-OPTIONS &KEY (VERBOSE T) &ALLOW-OTHER-KEYS)
  "Copy one file specified by PATHNAME to tape.  PATHNAME may be a string or a pathname object.
 If it is a string it must represent a fully qualified pathname.  Wildcards are not allowed.
 :VERBOSE controls whether to log the copy operation to *STANDARD-OUTPUT*.
 Other keyword options allowed are the same as COPY-FILE's.
 Returned value is the file object copied."
  (LET ((PARSED-NAME (PATHNAME PATHNAME))
	RESULT)
    (IF (SEND PARSED-NAME :WILD-P)
      (IF VERBOSE
	(FORMAT *STANDARD-OUTPUT* "~%Wilds not allowed in pathname, ~a" PARSED-NAME))
      (PROGN
	()
	(SETF RESULT (APPLY #'BACKUP-1 PARSED-NAME COPY-OPTIONS))
	(IF (CONSP RESULT)
	  (CAR RESULT)
	  RESULT))))) 


(DEFUN BACKUP-1 (PATHNAME &REST COPY-OPTIONS &KEY (VERBOSE T) &ALLOW-OTHER-KEYS)
  "Copy a file(s) specified by PATHNAME to tape.  PATHNAME may be a string or a pathname object.
 Wildcarding may cause copying of more than one file.
 :VERBOSE controls whether to log the the copy operation to *STANDARD-OUTPUT*.
 Other keyword options allowed are the same as COPY-FILE's.
 Returned value is a list of file object(s) copied."
  (THIRD
   (MULTIPLE-VALUE-LIST
    (APPLY #'COPY-FILE PATHNAME "MT:" :VERBOSE VERBOSE :OPEN-PARAMETERS
	   (LIST (QUOTE NIL);input parameters
		 (LIST :SOURCE-PATTERN;output parameters
		       (FS:PARSE-PATHNAME PATHNAME))) COPY-OPTIONS)))) 


(DEFUN BACKUP (PATHNAME &REST COPY-OPTIONS &KEY (VERBOSE T) (SET-BACKED-UP-FLAG NIL) (NOT-BACKED-UP-ONLY NIL)
	   (MARK-AS-DIRECTORY T) (SUB-DIRECTORIES T) &ALLOW-OTHER-KEYS)
  "Copy files specified by PATHNAME to tape.
 :VERBOSE            - controls whether to log the the operation to *STANDARD-OUTPUT*.
 :SET-BACKED-UP-FLAG - the file system's backed up bit will be set for files backed up.
 :NOT-BACKED-UP-ONLY - only files without the backed up bit will be backed up.
 :MARK-AS-DIRECTORY  - marks the file(s) backed up as a directory by writing an EOF.
 :SUB-DIRECTORIES    - traverse sub-directories.
 Other keyword options allowed are the same as COPY-FILE's."
  (LET (RETURN-VALUE)
    (SETF RETURN-VALUE
	  (IF SUB-DIRECTORIES
	    (APPLY #'FS::COPY-DIRECTORY PATHNAME "MT:" :VERBOSE VERBOSE :NOT-BACKED-UP-ONLY
		   NOT-BACKED-UP-ONLY :SET-BACKED-UP-FLAG SET-BACKED-UP-FLAG :OPEN-PARAMETERS
		   (LIST (QUOTE NIL) (LIST :SOURCE-PATTERN (FS:PARSE-PATHNAME PATHNAME)))
		   COPY-OPTIONS)
	    (APPLY #'BACKUP-1 PATHNAME :VERBOSE VERBOSE :NOT-BACKED-UP-ONLY NOT-BACKED-UP-ONLY
		   :SET-BACKED-UP-FLAG SET-BACKED-UP-FLAG COPY-OPTIONS)))
    (IF (AND RETURN-VALUE MARK-AS-DIRECTORY)
      (WRITE-EOF))
    RETURN-VALUE)) 


(DEFF RESTORE-FILE 'RESTORE-1) 

(DEFF RESTORE-DIRECTORY 'RESTORE) 

(DEFF BACKUP-DIRECTORY 'BACKUP) 

;;;
;;;  This old hack to avoid flattening directories of restored files, is now used
;;;  by :WILDCARD-MAP to accomplish the same thing when an "old" tape is read.
;;;  "Old" just means that the property :SOURCE-PATTERN is not in the header.
;;;    10-24-85 MBC
;;;

(DEFUN BUILD-NAME-AND-MATCH (PATHNAME STREAM)
  "Returns pathname merged with stream and match predicate, based on wildcards matching.
Directorys may be merged in an arbitrary way."
  (LET ((MATCH-P T)
	NAM
	TYP
	DIR
	DEVICE)
    (COND
      ((NULL PATHNAME);get entire destination from tape stream
       (VALUES (SEND STREAM :PATHNAME) T));match flag always T for this case.
      (T (IF (STRINGP PATHNAME)
	   (SETF PATHNAME (FS:PARSE-PATHNAME PATHNAME)))
       (LET ((DIR-WILD-OR-NIL?
	      (OR (SEND PATHNAME :DIRECTORY-WILD-P) (NULL (SEND PATHNAME :DIRECTORY)))))
	 (SETF PATHNAME (FS:MERGE-PATHNAME-DEFAULTS PATHNAME STREAM));merge NIL components
	 (WHEN (SEND PATHNAME :WILD-P)
	   (SETF PATHNAME (SEND PATHNAME :TRANSLATE-WILD-PATHNAME PATHNAME STREAM));WILD things
	   (MULTIPLE-VALUE-SETQ (DEVICE DIR NAM TYP)
	     (SEND PATHNAME :PATHNAME-MATCH-SPECS STREAM))
	   (SETF MATCH-P (AND (OR (NULL DIR-WILD-OR-NIL?) DIR) NAM TYP)))
	 (VALUES
	  (IF DIR-WILD-OR-NIL?
	    PATHNAME
	    (MAYBE-COMBINE-DIRECTORIES PATHNAME (SEND STREAM :TRUENAME)))
	  MATCH-P)))))) 


(DEFUN MAYBE-COMBINE-DIRECTORIES (PATHNAME STREAM)
  "This should combine directory names from the tape STREAM with PATHNAME, the destination."
  (LET* ((S-DIR (SEND STREAM :DIRECTORY))
	 (P-DIR (SEND PATHNAME :DIRECTORY)))
    (IF (EQ P-DIR :ROOT)
      (SEND PATHNAME :NEW-PATHNAME :DIRECTORY S-DIR)
      (PROGN
	(WHEN (ATOM P-DIR)
	  (SETQ P-DIR (CONS P-DIR ())))
	(IF (ATOM S-DIR)
	  (SETQ S-DIR ())
	  (SETQ S-DIR (CDR S-DIR)))
	(SEND PATHNAME :NEW-PATHNAME :DIRECTORY (APPEND P-DIR S-DIR)))))) 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Hack a RAW magtape stream. 9-7-85 MBC
;;;

(DEFUN RAW-DUMP (&OPTIONAL (OUT-STREAM *STANDARD-OUTPUT*) (RECORD-SIZE sys:disk-block-byte-size))
  "Dump tape contents in RAW mode until EOF or EORM is reached."
  (LET (R-STREAM)
    (UNWIND-PROTECT (PROGN
		     (SETF R-STREAM
			   (APPLY 'MAKE-MT-STREAM :CHARACTERS T :BYTE-SIZE 8
				  :UNIT *current-unit*	   ;8.3.87
				  :DENSITY ()))
		     (SEND R-STREAM :SET-RECORD-SIZE RECORD-SIZE)
		     (FORMAT OUT-STREAM "~%***DUMPING TAPE IN RAW MODE***")
		     (LOOP
		      (LET ((BUF (RAW-READ-BUFFER R-STREAM)))
			(IF BUF
			  (FORMAT OUT-STREAM "~&~a" BUF)
			  (RETURN (FORMAT OUT-STREAM "~%***RAW DUMP DONE***"))))))
      (SEND R-STREAM :CLOSE)))) 


(DEFUN RAW-READ-BUFFER (STREAM)
  (LET (BUF
	LOW
	COUNT)
    (MULTIPLE-VALUE-SETQ (BUF LOW COUNT)
      (FUNCALL STREAM :NEXT-INPUT-BUFFER))
    (IF BUF
      (SUBSeq BUF LOW COUNT)))) 
