;;; -*- Mode:Common-Lisp; Package:MT; Base:10; Fonts:(MEDFNT HL12B HL12BI) -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; Copyright (C) 1980, Massachusetts Institute of Technology


;;; REVISIONS:
;;;     04-22-88 DAB    Fixed (MT-FILEHANDLE :WILDCARD-MAP) to use :use-original-host when set to T.
;;; 	08.10.87 MBC	Make REEL :NEXT-INPUT-BUFFER handle all non-error status properly.
;;;	08.04.87 MBC	Adjust TO-INDEX in :SEND-OUTPUT-BUFFER based on BYTE-SIZE.
;;;	08.03.87 MBC	Add residue handling to (MT-INPUT-MIXIN :NEXT-INPUT-BUFFER)
;;;			and corrected handling of illegal-length-indicator (ILI)
;;;			residue in the method (REEL-MT-INPUT-MIXIN :NEXT-INPUT-BUFFER)
;;;			to support Tar tapes properly.
;;; 	07.27.87 MBC	Re-order to put method ahead of its defflavor.
;;;	07.14.87 MBC 	Add PADDED [Ascii-Backup] support.  
;;;     07.13.87 AB     Remove references to PROCESSOR-TYPE-CODE (old LAMBDA support).
;;;	06.22.87 MBC 	Pass new arguments to FS:DETERMINE-IF-TYPE-KNOWN.  [SPR 5552]
;;;			IO PATCH 3-17 REQUIRED.   Don't error off with "NIL" is an
;;;			unknown host when List-Contents encounters a partition.
;;;			Output :CHARACTERS keyword and value when writing headers.
;;;	06.16.87 MBC 	Fix two problems with backing up files that have no type. 
;;;			First, don't stuff "UNKNOWN" out to tape as the type, [SPR 5693].
;;;			Second, don't assume that its a character file a mung the data
;;;			via ascii translating,  [SPR 5521, ..not a valid binary object file].
;;;	05.22.87 MBC	Fix "LM" default host to use :SYSTEM-TYPE :LISPM.
;;;	05.08.87 MBC	Fix lost subdirectories when restore logical backups, by NOT 
;;;			translating source-pattern.  Also make sure that backtranslated 
;;;			directory names don't get on tape.  REQUIRES CHANGE in PNMAP.
;;;	04.22.87 MBC	Remove special directory handling in :WILDCARD-MAP, since this is
;;;			now handled by :TARGET-TRANSLATE-WILD-PATHNAMES.
;;;	04.13.87 MBC	Keep domain out of host component when creating header info.
;;;			Also :System-Type method needed to extract info from property list.
;;;     04.03.87 MP,MBC Imrpove handling of logical pathname on tape.  Translate the original
;;;                     Source-pattern from the tape inside the wildcard-map method of mt-filehandle
;;;                     before trying to match the pathnames.
;;;                     Put the system-type info on the file header on the tape.
;;;     
;;; Fixes for 2K page-size.  1-8-87, -ab
;;; Modifications to support SELECTIVE-SOURCE-PATTERN,  a new argument that
;;;      is used as a matching pattern to provide name selective restoration from tape.
;;;      Recover it via :SOURCE-PATTERN method after open of tape file for read, and use it for
;;;      pathname merging.  Also supports the :OPEN-PARAMETERS option in MAKE-MT-FILE-STREAM.
;;;      Changes to are interdependent with MTAUX changes, primarily RESTORE.  10-17-85 MBC
;;;
;;; Support for WILDCARD-MAP.  :TARGET-TRANSLATE-WILD-PATHNAME, SELECTIVE-SOURCE-PATTERN 
;;;     and SOURCE-PATTERN, added; DEFFLAVOR of MT-FILEHANDLE redefined.  10-16-85 MBC
;;; Remove CHAPARRAL cpu test, deletes lambda support.   Make magtape file stream
;;;     support use one large RQB and wire different number of pages
;;;     to control physical I-O size.                         - 9-5-85 MBC
;;; Add PLIST-LOCF method so all properties can be gotten.  - 11/08/84 MBC
;;; Add HOST & BACKUP-DATE to headers output; add flags for header displaying;
;;;     add HOST, BACKUP-DATE, PROPERTIES methods to mt-filehandle. - 11/05/84 MBC
;;; Use With-Zetalisp-On in Read-Magtape-Header and With-common-lisp-on in
;;;     Write-Magtape-Header;  also write headers with #!C. *MT-READTABLE* goes away.  11/6/85
;;; Read-Magtape-Header looks for LABL and signals error, "bootable tape".  12-6-85
;;; Fix small hole in :Wildcard-Map when Selective-Source-Pattern has fewer directory
;;;	nodes than the Tape-Pathname-Object.  Fix two problems when backing up files
;;;	whose byte-sizes are less than 8.  1: Divide by zero. 2: Page-Size-In-Bits. 5-6-86
;;; Version matching on source pathname in wildcard-map.  Commented only. 5-13-86
;;; Fix Wildcard-map to not try to hack null source pattern.  5-21-86

;;; Methods required to support stream that are defined here and used externally:
;;;
;;;  :NEXT-INPUT-BUFFER
;;;  :DISCARD-INPUT-BUFFER
;;;  :SEND-OUTPUT-BUFFER
;;;  :NEW-OUTPUT-BUFFER
;;;  :DISCARD-OUTPUT-BUFFER


;;; This file contains stream definitions for the magtape.

;;; Entrypoints are:
;;; MAKE-MT-STREAM &rest options
;;; MAKE-MT-FILE-STREAM &rest options
;;; MAKE-MAGTAPE-FILE-PROBE-STREAM &rest options
;;;
;;; Two flags to allow easy display of headers going to & from tape:

(DEFVAR SHOW-IN-HEADER ()) 

(DEFVAR SHOW-OUT-HEADER ()) 

(DEFVAR *LOGICAL-PATHNAMES-ON* T) 

(defvar   variable-block-max-blocks-per-rqb 32)

;;; Lossages:

;;;  The method of allocating RQBs used by the Explorer Streamer Tape in handling files is not
;;;  compatible with the philosophy of the system's RQB allocation scheme.  When the system creates
;;;  an RQB of a given size, that RQB will exist until the next boot, regardless of it being 
;;;  returned via si:return-disk-rqb.  Get-disk-rqb and return-disk-rqb simply keep track of
;;;  a pool of existing RQBs.  When a request is made for an RQB of N blocks, the pool is searched
;;;  for an unused RQB exactly N blocks long.  If none is found, an attempt is made to allocate
;;;  an rqb from the DISK-BUFFER-AREA.  The Explorer Streamer Tape tailors its RQB requests sizes
;;;  to fit the size of the current file.  After running a while all possible RQB sizes from 
;;;  1 to max-blocks-per-rqb may be requested.  If max-blocks-per-rqb is too large, eventually 
;;;  DISK-BUFFER-AREA (currently four megabytes) is all used up.
;;;
;;;  The tape system SHOULD get one or two large RQBs and wire only the number of blocks needed.
;;;  Then when it tailors I/O size to the file's size, new RQBs won't continually be allocated.
;;;

(DEFPARAMETER block-size-in-bits (* sys:disk-block-word-size 32.))

(DEFPARAMETER MAX-blocks-PER-RQB 100) 

(DEFPARAMETER *DEFAULT-RECORD-SIZE* 4096) 

(DEFPARAMETER *NUPI-DEFAULT-RECORD-SIZE* (* 4 1024)) 

(DEFPARAMETER *NUPI-DEFAULT-UNIT* 6) 	     


(DEFVAR TAPE-HEADER-LENGTH) 

(DEFPARAMETER *NUPI-TAPE-HEADER-LENGTH* 1024) 

(DEFPARAMETER *LAMBDA-TAPE-HEADER-LENGTH* 4096) 

;; This is also in FSDEFS

(REMPROP 'QUOTIENT-CEILING 'SOURCE-FILE-NAME) 

(DEFUN QUOTIENT-CEILING (Y X)
  "The smallest integer greater than or equal to Y/X"
  (QUOTIENT (+ Y (1- X)) X)) 

;; Get a buffer of the correct byte size

(DEFUN MT-GET-RQB-ARRAY (RQB BYTE-SIZE &AUX TYPE)
  (COND
    ((= BYTE-SIZE 16) (SYSTEM:RQB-BUFFER RQB))
    ((= BYTE-SIZE 8) (SYSTEM:RQB-8-BIT-BUFFER RQB))
    ((SETQ TYPE (CDR (ASSOC BYTE-SIZE '((4 . ART-4B) (2 . ART-2B) (1 . ART-1B)) :TEST #'EQ)))
     (MAKE-ARRAY
       (QUOTIENT (* (SYSTEM:RQB-nblocks RQB) block-size-in-bits) BYTE-SIZE)
       :AREA FS::LOCAL-FILE-SYSTEM-AREA :TYPE TYPE :DISPLACED-TO RQB
       ;; This is a system bug.  Don't try to figure it out.
       :DISPLACED-INDEX-OFFSET
       (* (si:%P-CONTENTS-OFFSET (SYSTEM:RQB-BUFFER RQB) 3)
	  (QUOTIENT 16 BYTE-SIZE))))
    (T (FERROR () "~D is an invalid byte size." BYTE-SIZE)))) 


(DEFFLAVOR MT-MIXIN
	   ((STATUS ':OPEN)
	    (UNIT *NUPI-DEFAULT-UNIT*)
	    (RECORD-SIZE *NUPI-DEFAULT-RECORD-SIZE*)
	    (DENSITY 0)
	    (IBM-MODE NIL)
	    (BYTE-SIZE 8)
	    (RQB NIL))
	   ()
  (:INCLUDED-FLAVORS STREAM)
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES
  :INITABLE-INSTANCE-VARIABLES)


(DEFMETHOD (MT-MIXIN :ADVANCE-FILE) ()
  ()) 


(DEFMETHOD (MT-MIXIN :CLOSE) (&OPTIONAL ABORTP)
  (OR (EQ ABORTP :RAW) (SEND SELF :ADVANCE-FILE))
  (COND
    ((NEQ STATUS :CLOSED) (SETQ STATUS :CLOSED) (AND RQB (SYSTEM:RETURN-DISK-RQB RQB))
			  (SETQ RQB ()) T))) 


(DEFMETHOD (MT-MIXIN :RETURN-RQB) ()
  (COND
    (RQB (SYSTEM:RETURN-DISK-RQB RQB) (SETQ RQB ())))) 


(DEFFLAVOR MT-INPUT-MIXIN
	   ()
	   (MT-MIXIN)
  (:INCLUDED-FLAVORS SI:BUFFERED-INPUT-STREAM)) 


(DEFFLAVOR MT-OUTPUT-MIXIN
	   ()
	   (MT-MIXIN)
  (:INCLUDED-FLAVORS SI:BUFFERED-OUTPUT-STREAM)) 

;; CAUTION:  There is an :AROUND on this method.

(DEFVAR DEBUG ()) 

;;; Add residue handling.  8.3.87
(DEFMETHOD (MT-INPUT-MIXIN :NEXT-INPUT-BUFFER)
	   (&OPTIONAL IGNORE)
  (let ((request-size-in-blocks (QUOTIENT-CEILING RECORD-SIZE si:disk-block-byte-size)))
    (COND ((EQ STATUS :CLOSED) (FERROR NIL "Attempt to get input from ~S, which is closed." SELF))
	  ((EQ STATUS :EOF) NIL)
	  ((> request-size-in-blocks max-blocks-per-rqb)   ;8.3.87
	   (FERROR NIL ":Record-Size in blocks, ~d. exceeds MAX-BLOCKS-PER-RQB, ~d."
		   record-size max-blocks-per-rqb))
	  (T (OR RQB (SETQ RQB (SI:GET-DISK-RQB max-blocks-per-rqb)))
	     ;;; Record size is in bytes, but mt-run takes blocks when in fixed block mode.
	     (let ((BUFFER (MT-GET-RQB-ARRAY RQB BYTE-SIZE)))
	       (Mt-Run RQB %MT-COMMAND-READ request-size-in-blocks UNIT DENSITY)
	       (cond ((MT-STATUS-EOF)
		      (let ((count-from-residue (- record-size (SI:RQ-TRANSFER-LENGTH rqb))))
			(SETQ STATUS :EOF)
			(if (zerop count-from-residue)	   ;8.3.87
			    (values NIL)
			    (VALUES BUFFER 0 (QUOTIENT (* 8 count-from-residue) BYTE-SIZE)))))
		     (T (VALUES BUFFER 0 (QUOTIENT (* 8 RECORD-SIZE) BYTE-SIZE)))))))))


(DEFMETHOD (MT-INPUT-MIXIN :DISCARD-INPUT-BUFFER) IGNORE) 


(DEFMETHOD (MT-INPUT-MIXIN :NEXT-FILE) ()
  (IF (EQ STATUS :EOF)
      (SETQ STATUS :OPEN)
      (PROGN
	(SPACE-TO-EOF UNIT)
	(SEND SELF :CLEAR-INPUT)))) 


(DEFMETHOD (MT-INPUT-MIXIN :ADVANCE-FILE) ()
  (COND
    ((NOT (MEMBER STATUS '(:EOF :CLOSED) :TEST #'EQ)) (SPACE-TO-EOF UNIT) (SETQ STATUS :EOF)))) 


(DEFMETHOD (MT-OUTPUT-MIXIN :NEW-OUTPUT-BUFFER) ()
  (OR RQB (SETQ RQB (SYSTEM:GET-DISK-RQB MAX-blocks-PER-RQB)))
  (VALUES (MT-GET-RQB-ARRAY RQB BYTE-SIZE) 0
	  (QUOTIENT (* RECORD-SIZE 8) BYTE-SIZE))) 

;;; Remove debug flag and use TO-INDEX.  7.14.87
(DEFMETHOD (MT-OUTPUT-MIXIN :SEND-OUTPUT-BUFFER) (BUFFER TO-INDEX)
  (UNLESS
    (AND (ARRAY-INDIRECT-P BUFFER)
	 (EQ
	   (si:%P-CONTENTS-OFFSET BUFFER (1+ (si:%P-LDB-OFFSET SYSTEM:%%ARRAY-LONG-LENGTH-FLAG BUFFER 0)))
	   RQB))
    (FERROR () "Attempt to :SEND-OUTPUT-BUFFER ~S, which is not indirected into ~S" BUFFER RQB))
  (MT-RUN RQB
	  %MT-COMMAND-WRITE
	  (MIN (CEILING RECORD-SIZE SYS:DISK-BLOCK-BYTE-SIZE)	   ;7.14.87
	       (CEILING (quotient (* byte-size to-index) 8.) SYS:DISK-BLOCK-BYTE-SIZE))	   ;8.4.87
	  UNIT DENSITY IBM-MODE))


(DEFMETHOD (MT-OUTPUT-MIXIN :DISCARD-OUTPUT-BUFFER) IGNORE) 


(DEFMETHOD (MT-OUTPUT-MIXIN :ADVANCE-FILE) ()
  (OR (EQ STATUS :CLOSED) (WRITE-EOF UNIT))) 


(DEFFLAVOR MT-INPUT-STREAM
	   ()
	   (MT-INPUT-MIXIN
	    SI:BUFFERED-INPUT-STREAM)) 


(DEFFLAVOR MT-OUTPUT-STREAM
	   ()
	   (MT-OUTPUT-MIXIN
	    SI:BUFFERED-OUTPUT-STREAM)) 


(DEFFLAVOR MT-CHARACTER-INPUT-STREAM
	   ()
	   (MT-INPUT-MIXIN
	    SI::BUFFERED-LINE-INPUT-STREAM)) 


(DEFFLAVOR MT-CHARACTER-OUTPUT-STREAM
	   ()
	   (MT-OUTPUT-MIXIN
	    SI:BUFFERED-OUTPUT-CHARACTER-STREAM)) 


(COMPILE-FLAVOR-METHODS MT-INPUT-STREAM MT-OUTPUT-STREAM MT-CHARACTER-INPUT-STREAM
			MT-CHARACTER-OUTPUT-STREAM) 

;; This function makes a simple stream to the magtape.
;; It can be used by programs which want to read arbitrarily formated magtapes.



(DEFUN MAKE-MT-STREAM (&OPTIONAL &KEY (DIRECTION :INPUT) (BYTE-SIZE 8)
		       (CHARACTERS T) (UNIT *NUPI-DEFAULT-UNIT*)
		       (RECORD-SIZE *DEFAULT-RECORD-SIZE*) (DENSITY 0) (IBM-MODE NIL) &ALLOW-OTHER-KEYS)
  
  (CHECK-ARG DIRECTION (MEMBER DIRECTION '(:INPUT :OUTPUT) :TEST #'EQ) ":INPUT or :OUTPUT")
  (CHECK-TYPE BYTE-SIZE :NUMBER)
  (CHECK-ARG CHARACTERS (MEMBER CHARACTERS '(T NIL) :TEST #'EQ) "T or NIL")
  (MAKE-INSTANCE
    (IF (EQ DIRECTION :INPUT)
	(IF CHARACTERS
	    (if (eq (get 'mt:*current-unit* 'block-mode) :variable)
		'reel-mt-character-input-stream
		'MT-CHARACTER-INPUT-STREAM)
	    
	    (if (eq (get 'mt:*current-unit* 'block-mode) :variable)
		'reel-mt-input-stream
		'MT-INPUT-STREAM))
	
	(IF CHARACTERS
	    (if (eq (get 'mt:*current-unit* 'block-mode) :variable)
		'reel-mt-character-output-stream
		'MT-CHARACTER-OUTPUT-STREAM)
	    
	    (if (eq (get 'mt:*current-unit* 'block-mode) :variable)
		'reel-mt-output-stream
		'MT-OUTPUT-STREAM)))
    
    :BYTE-SIZE BYTE-SIZE :UNIT UNIT :RECORD-SIZE RECORD-SIZE :DENSITY DENSITY :IBM-MODE IBM-MODE)) 

;;; Magtapes with headers.

;;; The following function reads a block from the magtape and tries to
;;; determine the header plist.  If it reads an EOF, it returns NIL.
;;; The second value returned is the style of header.
;;; Currently known about are :MIT, :MIT, and :MIT.

;;; this code returns a pure PLIST and does not make any assumptions about
;;; name formats or anything.


(DEFUN READ-MAGTAPE-HEADER (STREAM &AUX BUF LOW COUNT PLIST)
  (declare (special  FILE-SYSTEM-PACKAGE
		     *remote-tape-error-stream*
		     *format-error*))
  (LET ((CURRENT-RECORD-SIZE (SEND STREAM :RECORD-SIZE)))  ;nupi
    (UNWIND-PROTECT (PROGN
		      (SEND STREAM :SET-RECORD-SIZE TAPE-HEADER-LENGTH)
		      (MULTIPLE-VALUE-SETQ (BUF LOW COUNT)
			(FUNCALL STREAM :GET-INPUT-BUFFER))
		      (COND
			((NULL BUF) NIL)   ;EOF
			((AND (>= COUNT 4) (STRING-EQUAL BUF "LMFL" LOW 0 (+ LOW 4) 4))
			 ;; MIT format.
			 (LOOP REPEAT 4 DO (FUNCALL STREAM :TYI))
			 (LET ((*PRINT-BASE* 10.)
			       (*PRINT-LENGTH* NIL)
			       (*PRINT-LEVEL* NIL)
			       (*PRINT-RADIX* T)
			       (*PRINT-PRETTY* NIL)
			       (*PACKAGE* FILE-SYSTEM-PACKAGE)
			       (SI:PRINT-READABLY T))
			   (WITH-ZETALISP-ON (SETQ PLIST (READ STREAM T))))
					   ;always read in zeta, but expect #!C from 2.0 tapes
			 (IF SHOW-IN-HEADER
			     (PRINT PLIST))
			 (VALUES PLIST :MIT))
			((AND (>= COUNT 4) (STRING-EQUAL BUF "LABL" LOW 0 (+ LOW 4) 4))
			 (CERROR () () 'SYSTEM:TAPE-ERROR "Bad Header - Tape is a Bootable Tape."))
			(T1 *(if *remote-tape-error-stream*	;1; if remote don't ferror*
			       (progn
				 (setq *format-error*		;1; save error message, later used by list-contents*
				       (Format () "format error: Unknown tape header format"))
				 ())
			       (FERROR 'mt:unknown-tape-header "Unknown tape header format.  Header is:~%~A"
				       (SUBSeq BUF LOW (+ LOW (MIN 256 COUNT))))))) ) ;1progn*
      (SEND STREAM :SET-RECORD-SIZE CURRENT-RECORD-SIZE)
      (FUNCALL STREAM :DISCARD-CURRENT-INPUT-BUFFER))))




(DEFUN WRITE-MAGTAPE-HEADER (STREAM PLIST &OPTIONAL (STYLE :MIT))
  (declare (special  FILE-SYSTEM-PACKAGE))
  (LET ((CURRENT-RECORD-SIZE (SEND STREAM :RECORD-SIZE)))  ;nupi
    (UNWIND-PROTECT (SEND
		      STREAM
		      :SET-RECORD-SIZE
		      TAPE-HEADER-LENGTH)
      (CASE STYLE
	(:MIT
	 (LET* ((*PRINT-BASE* 10.)
		(*PRINT-LENGTH* NIL)
		(*PRINT-LEVEL* NIL)
		(*PRINT-RADIX* T)
		(*PRINT-PRETTY* NIL)
		(*PACKAGE* FILE-SYSTEM-PACKAGE)
		(SI:PRINT-READABLY T))
	   (WITH-COMMON-LISP-ON (SEND STREAM :STRING-OUT "LMFL") (FORMAT STREAM "#!C~s" PLIST)
				(IF SHOW-OUT-HEADER
				    (PRINT PLIST)))
	   (FUNCALL STREAM :PAD-AND-SEND-BUFFER #\SPACE)))
	(:BOOT
	 (LET* ((*PRINT-BASE* 10.)
		(*PRINT-LENGTH* NIL)
		(*PRINT-LEVEL* NIL)
		(*PRINT-RADIX* T)
		(*PRINT-PRETTY* NIL)
		(*PACKAGE* FILE-SYSTEM-PACKAGE)
		(SI:PRINT-READABLY T))
	   (SI::WRITE-BOOT-HEADER STREAM)
	   (FUNCALL STREAM :PAD-AND-SEND-BUFFER 0)))
	(OTHERWISE (FERROR () "~S is an unknown tape header format." STYLE)))
      (SEND STREAM :SET-RECORD-SIZE CURRENT-RECORD-SIZE)
      (FUNCALL STREAM :FINISH)))) 


(DEFFLAVOR HEADER-MIXIN
	   (HEADER-FORMAT
	    (SI:PROPERTY-LIST NIL))
	   (SI:PROPERTY-LIST-MIXIN)
  (:INCLUDED-FLAVORS MT-MIXIN)
  :INITABLE-INSTANCE-VARIABLES) 


(DEFFLAVOR HEADER-INPUT-MIXIN
	   ()
	   (HEADER-MIXIN)) 


(DEFFLAVOR HEADER-OUTPUT-MIXIN
	   ()
	   (HEADER-MIXIN)) 

;; Write the header at the last possible minute.

(DEFMETHOD (HEADER-OUTPUT-MIXIN :BEFORE :NEXT-INPUT-BUFFER) ()
  (AND (EQ STATUS :OPEN) (NULL RQB) (WRITE-MAGTAPE-HEADER SELF SI:PROPERTY-LIST HEADER-FORMAT))) 

;this tries to responds to pathcomponent stream messages, but may not make any
; assumptions about the exact flavor of path involved (which file system, etc).

(DEFFLAVOR MT-FILE-MIXIN
	   ()
	   ()
  (:INCLUDED-FLAVORS HEADER-MIXIN MT-MIXIN)) 


(DEFFLAVOR MT-FILE-INPUT-MIXIN		   ;nupi - this ok for lambda? extra junk.
	   ((EXPECTED-LENGTH NIL)
	    (RUNNING-TOTAL))
	   (MT-FILE-MIXIN)
  (:GETTABLE-INSTANCE-VARIABLES EXPECTED-LENGTH RUNNING-TOTAL)
  (:SETTABLE-INSTANCE-VARIABLES EXPECTED-LENGTH RUNNING-TOTAL)
  (:INITABLE-INSTANCE-VARIABLES EXPECTED-LENGTH RUNNING-TOTAL)) 


(DEFFLAVOR MT-FILE-OUTPUT-MIXIN
	   ()
	   (MT-FILE-MIXIN)) 


(DEFFLAVOR MT-FILE-INPUT-STREAM
	   ()
	   (MT-FILE-INPUT-MIXIN
	    HEADER-INPUT-MIXIN
	    MT-INPUT-STREAM)) 


(DEFFLAVOR MT-FILE-CHARACTER-INPUT-STREAM
	   ()
	   (MT-FILE-INPUT-MIXIN
	    HEADER-INPUT-MIXIN
	    MT-CHARACTER-INPUT-STREAM)) 


(DEFFLAVOR MT-FILE-OUTPUT-STREAM
	   ()
	   (MT-FILE-OUTPUT-MIXIN
	    HEADER-OUTPUT-MIXIN
	    MT-OUTPUT-STREAM)) 


(DEFFLAVOR MT-FILE-CHARACTER-OUTPUT-STREAM
	   ()
	   (MT-FILE-OUTPUT-MIXIN
	    HEADER-OUTPUT-MIXIN
	    MT-CHARACTER-OUTPUT-STREAM)) 


;;; Begin new PADDED support code.  7.14.87
(defvar PAD-CHAR #xFF)

;;; Pad all character output streams.  7.14.87   7.27.87
(DEFMETHOD (MT-FILE-CHARACTER-OUTPUT-STREAM :SEND-OUTPUT-BUFFER) (BUFFER TO-INDEX &aux n-pages)
  (UNLESS
    (and (ARRAY-INDIRECT-P BUFFER)
	 (EQ
	   (si:%P-CONTENTS-OFFSET BUFFER (1+ (si:%P-LDB-OFFSET SYSTEM:%%ARRAY-LONG-LENGTH-FLAG BUFFER 0)))
	   RQB))
    (FERROR () "Attempt to :SEND-OUTPUT-BUFFER ~S, which is not indirected into ~S" BUFFER RQB))
  (if (>= to-index record-size)
      (setf n-pages (QUOTIENT-CEILING record-size SI:DISK-BLOCK-BYTE-SIZE))
      (progn
	(setf n-pages (QUOTIENT-CEILING
			(quotient (* byte-size to-index) 8.)	   ;8.4.87
			SI:DISK-BLOCK-BYTE-SIZE))  
	(si:array-initialize buffer PAD-CHAR To-Index (* N-PAGES SI:DISK-BLOCK-BYTE-SIZE))))	   ;7.14.87
  (MT-RUN RQB %MT-COMMAND-WRITE n-pages UNIT DENSITY IBM-MODE))

(defvar *ASCII-PADDED-ENABLED* T
  "Enables reading and writing of character files on tape using a padded mode.
This is required to back up files from hosts that don't provide accurate file
lengths.  This implies that the length shown by List-Contents may be inaccurate.
If restoring files that contain data byte values of #xFF (255.) turn this flag off.")


(DEFFLAVOR MT-FILE-CHARACTER-INPUT-STREAM-PADDED
	   ()
	   (MT-FILE-INPUT-MIXIN-PADDED HEADER-INPUT-MIXIN MT-CHARACTER-INPUT-STREAM))

(DEFFLAVOR MT-FILE-INPUT-MIXIN-PADDED
	   ((expected-length nil)
	    running-total)
	   (MT-FILE-MIXIN)
  ;;; we don't use these, but lets not error off cause of a lack of having them declared.
  (:GETTABLE-INSTANCE-VARIABLES expected-length running-total)
  (:SETTABLE-INSTANCE-VARIABLES expected-length running-total)
  (:INITABLE-INSTANCE-VARIABLES expected-length running-total))

(defun FIND-PAD (buffer actual-read-count)
  (let* ((start-index-of-last-block
	   (* SI:DISK-BLOCK-BYTE-SIZE
	      (floor (1- actual-read-count) SI:DISK-BLOCK-BYTE-SIZE)))
	 pad-index)
    (when (= PAD-CHAR (aref buffer (1- actual-read-count)))	   ;Only padded if last byte is PAD
      (setf pad-index
	    (si:string-reverse-search-not-set
	      PAD-CHAR buffer (1- actual-read-count) start-index-of-last-block)))
    (or (when pad-index (1+ pad-index))
	actual-read-count)))

;;; Guarranteed to have a BYTE-FACTOR = 1.
;;; Record-size is not used here, since we don't know what to expect.
;;; Running-total is kept simply for completeness.
;;; So.. just read maximum size each physical I/O, then
;;; if EOF figure residue and scan last block for pads
;;; Else scan last block of maximum for pads [in case the file 
;;; ends in the maximum block and EOF is just ahead.
(DEFMETHOD (MT-FILE-INPUT-MIXIN-PADDED :NEXT-INPUT-BUFFER)
	   (&OPTIONAL IGNORE)
  (COND ((EQ STATUS :CLOSED)
	 (FERROR NIL "Attempt to get input from ~S, which is closed." SELF))
	((EQ STATUS :EOF) NIL)
	(T (OR RQB (SETQ RQB (SI:GET-DISK-RQB max-blocks-per-rqb)))
	   (let ((request-count (* max-blocks-per-rqb SI:DISK-BLOCK-BYTE-SIZE))
		 (BUFFER (MT-GET-RQB-ARRAY RQB BYTE-SIZE)))
	     (MT-RUN RQB %MT-COMMAND-READ request-count UNIT DENSITY)
	     (cond ((MT-STATUS-EOF)
		    (let ((count-from-residue (- request-count (SI:RQ-TRANSFER-LENGTH rqb))))
		      (SETQ STATUS :EOF)
		      (if (zerop count-from-residue)
			  (values NIL)
			  (let ((actual-count (FIND-PAD buffer count-from-residue)))
			    (SETQ RUNNING-TOTAL (+ RUNNING-TOTAL actual-count))
			    (VALUES BUFFER 0 actual-count)))))
		   (T (let ((actual-count (FIND-PAD buffer request-count)))
			(SETQ RUNNING-TOTAL (+ RUNNING-TOTAL actual-count))
			(VALUES BUFFER 0 actual-count))))))))
;;; End of new independent PADDED support code.  7.14.87

(DEFMETHOD (MT-FILE-MIXIN :DEVICE) ()
  (GETF SI:PROPERTY-LIST :DEVICE)) 


(DEFMETHOD (MT-FILE-MIXIN :DIRECTORY) ()
  (OR (GETF SI:PROPERTY-LIST :DIRECTORY) (GETF SI:PROPERTY-LIST 'DIRECTORY))) 


(DEFMETHOD (MT-FILE-MIXIN :NAME) ()
  (GETF SI:PROPERTY-LIST :NAME)) 


(DEFMETHOD (MT-FILE-MIXIN :TYPE) ()
  (GETF SI:PROPERTY-LIST :TYPE)) 


(DEFMETHOD (MT-FILE-MIXIN :CANONICAL-TYPE) ()
  (GETF SI:PROPERTY-LIST :TYPE)) 


(DEFMETHOD (MT-FILE-MIXIN :HOST) ()
  (LET ((HOST (GETF SI:PROPERTY-LIST :HOST)))	   ;try not to break logical hosts
    (IF (FS:GET-PATHNAME-HOST HOST T)	   ;but handle unknown host
	HOST				   ;just assume local makes sense
	(setf (GETF SI:PROPERTY-LIST :SYSTEM-TYPE) :lispm) ;5.22.87
	"LM")))

(DEFMETHOD (MT-FILE-MIXIN :SYSTEM-TYPE) () ;4.13.87
  (GETF SI:PROPERTY-LIST :SYSTEM-TYPE))

(DEFMETHOD (MT-FILE-MIXIN :SOURCE-PATTERN) ()
  (LET ((SP (GETF SI:PROPERTY-LIST :SOURCE-PATTERN)))
    (IF SP
	(FS::PATHNAME-STRING-TO-OBJECT SP (send self :host)	   ;3.10.87
				       (send self :system-type)))))	   ;3.17.87


(DEFMETHOD (MT-FILE-MIXIN :AUTHOR) ()
  (GETF SI:PROPERTY-LIST :AUTHOR)) 


(DEFMETHOD (MT-FILE-MIXIN :BACKUP-DATE) ()
  (GETF SI:PROPERTY-LIST :BACKUP-DATE)) 


(DEFMETHOD (MT-FILE-MIXIN :VERSION) ()
  (GETF SI:PROPERTY-LIST :VERSION)) 


(DEFMETHOD (MT-FILE-MIXIN :PLIST-LOCF) ()
  (LOCF SI:PROPERTY-LIST)) 


(DEFMETHOD (MT-FILE-MIXIN :CREATION-DATE) ()
  (GETF SI:PROPERTY-LIST :CREATION-DATE)) 


(DEFMETHOD (MT-FILE-MIXIN :QFASLP) ()
  (GETF SI:PROPERTY-LIST :QFASLP)) 


(DEFMETHOD (MT-FILE-OUTPUT-MIXIN :CHANGE-PROPERTIES) (IGNORE &REST PROPERTIES)
  (LOOP FOR (IND PROP) ON PROPERTIES BY #'CDDR DO (SETF (GETF SI:PROPERTY-LIST IND) PROP))) 

;;;
;;; New methods for Rel 1.0.1 - 7-25-85 MBC
;;;
(import 'fs:mt-filehandle)

(DEFFLAVOR MT-FILEHANDLE
	   (NAMESTRING
	    (SELECTIVE-SOURCE-PATTERN NIL)
	    (USE-ORIGINAL-HOST NIL))
	   ()
  (:GETTABLE-INSTANCE-VARIABLES SELECTIVE-SOURCE-PATTERN USE-ORIGINAL-HOST)
  (:SETTABLE-INSTANCE-VARIABLES SELECTIVE-SOURCE-PATTERN USE-ORIGINAL-HOST)
  (:INITABLE-INSTANCE-VARIABLES SELECTIVE-SOURCE-PATTERN USE-ORIGINAL-HOST NAMESTRING))


;;; Remove calls to HACK-TARGET-PATTERN and HACK-LOGICAL-SUBS since
;;; :TARGET-TRANSLATE-WILD-PATHNAME now correctly supports directories.  4.22.87 
(DEFMETHOD (MT-FILEHANDLE :WILDCARD-MAP) (OPERATION IGNORE IGNORE SOURCE-PATTERN TARGET-PATTERN IGNORE IGNORE
					  &REST OPTIONS &KEY  PASS-DIRECTORIES-ON &ALLOW-OTHER-KEYS)
  ;1; used by copy/restore directory*
  (DECLARE (special *REMOTE-TAPE-STREAM*)
	   (ARGLIST OPERATION IGNORE IGNORE SOURCE-PATTERN TARGET-PATTERN SOURCE-INSTANCE &REST OPTIONS
		    &KEY PASS-DIRECTORIES-ON))
;  (and  *REMOTE-TAPE-STREAM*
;	(format t "wildcard-map called in mt-patch, subdir = ~s" PASS-DIRECTORIES-ON) (sleep 3) )
  (LET (RESULTS  TEMP-TARGET-PATTERN)
    (LOOP
      (WITH-OPEN-STREAM (TAPE-STREAM (SEND SELF :OPEN :DIRECTION :INPUT))
	(COND
	  ((ERRORP TAPE-STREAM)
	   (if *REMOTE-TAPE-STREAM*
	       (progn
		 (when PASS-DIRECTORIES-ON 	;1; restoring directory*
		   (setq PASS-DIRECTORIES-ON ())			;1; to stop when restore directory*
		   (reject-command *REMOTE-TAPE-STREAM* $PREPARE-TO-READ-MORE)) ;1; when*
		 (reject-command *REMOTE-TAPE-STREAM* TAPE-STREAM))	;1; progn -- remote*

	       (SETF (GET TAPE-STREAM :RETURN-VALUES) (NREVERSE RESULTS))
	       (FERROR TAPE-STREAM)) )	;1; errorp*
	  ((SEND TAPE-STREAM :GET :PARTITION)
	   ;1; *(princ "before SIGNAL-INVALID-FILE-TYPE")
	   (if *REMOTE-TAPE-STREAM*
	       (progn
		 (and PASS-DIRECTORIES-ON 	;; restoring directory
		      (reject-command *REMOTE-TAPE-STREAM* $PREPARE-TO-READ-MORE)) ;1; and*
		 (reject-command *REMOTE-TAPE-STREAM*
				 (string-append "Tape partition [" (SEND TAPE-STREAM :GET :COMMENT)
						"] is not a " (if PASS-DIRECTORIES-ON "directory" "file")) )
		 (setq PASS-DIRECTORIES-ON ()) )	;1; remote *;1; to stop when restore directory*
	       (SIGNAL-INVALID-FILE-TYPE TAPE-STREAM)) )
	  (T (LET* ((TAPE-PATHNAME-OBJECT  (SEND TAPE-STREAM :TRUENAME))
		    (TRANSLATED-TAPE-PATHNAME-OBJECT (translated-pathname (SEND TAPE-STREAM :TRUENAME)))
		    (TRUE-SOURCE-PATTERN (SEND TAPE-STREAM :SOURCE-PATTERN))	   ;5.8.87
		    HACKED-SELECTIVE-SOURCE-PATTERN)
	       (WHEN SELECTIVE-SOURCE-PATTERN
		 (SETF HACKED-SELECTIVE-SOURCE-PATTERN
		       (send (HACK-SOURCE-PATTERN
			       (translated-pathname SELECTIVE-SOURCE-PATTERN)
			       TRANSLATED-TAPE-PATHNAME-OBJECT)
			     :translated-pathname))
		 (IF (EQ (PATHNAME-VERSION HACKED-SELECTIVE-SOURCE-PATTERN) :NEWEST)
		     (SETF HACKED-SELECTIVE-SOURCE-PATTERN
			   (SEND HACKED-SELECTIVE-SOURCE-PATTERN :NEW-VERSION :WILD))))
	       (WHEN (OR (NOT SELECTIVE-SOURCE-PATTERN)
			 (SEND
			   TRANSLATED-TAPE-PATHNAME-OBJECT
			   :PATHNAME-MATCH-NO-DEVICE
			   (SEND HACKED-SELECTIVE-SOURCE-PATTERN :TRANSLATE-WILD-PATHNAME
				 HACKED-SELECTIVE-SOURCE-PATTERN
				 TRANSLATED-TAPE-PATHNAME-OBJECT)
			   ()))		   ;Operate-On-This-File predicate.
		 (IF TRUE-SOURCE-PATTERN   ;Tape made under 2.0 Release or later.
		     (SETF SOURCE-PATTERN  ;4.22.87 No more HACKing, wildcarding fixed in PNMAP.
			   (HACK-SOURCE-PATTERN TRUE-SOURCE-PATTERN TAPE-PATHNAME-OBJECT)
			   TEMP-TARGET-PATTERN TARGET-PATTERN)
		     ;; HACK-PATTERN hacks :WILD-INFERIORS till we have REAL support.
		     (PROGN		   ;Old tape, so do directory node surgery as in 1.0.1 Release.
		       (SETF TEMP-TARGET-PATTERN (BUILD-NAME-AND-MATCH TARGET-PATTERN TAPE-STREAM))
		       (SETF SOURCE-PATTERN TAPE-PATHNAME-OBJECT)))
		 (when use-original-host    ;04-22-88 DAB Added support for :use-original-host
		   (setf temp-target-pattern (send temp-target-pattern :new-pathname 
						   :host (send SOURCE-PATTERN :host))))
		 (LET ((RETURNED-FILE
			 (APPLY OPERATION () SOURCE-PATTERN TEMP-TARGET-PATTERN TAPE-STREAM ()
				OPTIONS)))
		   (IF RETURNED-FILE
		       (PUSH RETURNED-FILE RESULTS))))))))
      (UNLESS PASS-DIRECTORIES-ON
	(RETURN)))	 ;1; loop*
    (NREVERSE RESULTS)))



(DEFUN HACK-SOURCE-PATTERN (SP SI)
  (DECLARE (VALUES SOURCE-PATTERN))
  (LET (NODE-DIFFERENCE
	(SP-DIRS (SEND SP :DIRECTORY))
	(SI-DIRS (SEND SI :DIRECTORY)))
    (IF (AND (CONSP SP-DIRS) (CONSP SI-DIRS)
	     (SETQ NODE-DIFFERENCE (- (LENGTH SI-DIRS) (LENGTH SP-DIRS))))
	(IF (PLUSP NODE-DIFFERENCE)
	    (PROGN
	      (DOTIMES (I NODE-DIFFERENCE)
		(SETQ SP-DIRS (APPEND SP-DIRS (CONS :WILD ()))))
	      (SEND SP :NEW-DIRECTORY SP-DIRS))
	    SP)
	(IF (AND (EQ SP-DIRS :ROOT) (CONSP SI-DIRS))
	    (SEND SP :NEW-DIRECTORY SI-DIRS)
	    SP)))) 


(DEFMETHOD (MT-FILEHANDLE :TARGET-TRANSLATE-WILD-PATHNAME) (&REST IGNORE)
  SELF)					   ;for output, target is simple


(DEFMETHOD (MT-FILEHANDLE :DIRECTORY) (&REST IGNORE)
  ()) 


(DEFMETHOD (MT-FILEHANDLE :NEW-PATHNAME) (&REST IGNORE)
  SELF) 


(DEFMETHOD (MT-FILE-MIXIN :TRUENAME) (&REST IGNORE)
  (LET* ((SP (SEND SELF :SEND-IF-HANDLES :SOURCE-PATTERN))
	 (DEFAULT-DEVICE (AND SP (SEND SP :DEVICE))))
    (MAKE-PATHNAME :HOST (SEND SELF :HOST) :DEVICE
		   (OR DEFAULT-DEVICE (GETF SI:PROPERTY-LIST :DEVICE)) :DIRECTORY
		   (GETF SI:PROPERTY-LIST :DIRECTORY) :NAME (GETF SI:PROPERTY-LIST :NAME)
		   :TYPE (GETF SI:PROPERTY-LIST :TYPE) :VERSION
		   (GETF SI:PROPERTY-LIST :VERSION))))

(DEFMETHOD (MT-FILE-MIXIN :STRING-FOR-PRINTING) ()
  (LET ((SP (SEND SELF :SOURCE-PATTERN)))
    (IF SP
	(SEND
	  (SEND SP :NEW-PATHNAME :DIRECTORY (SEND SELF :DIRECTORY) :NAME (SEND SELF :NAME) :TYPE
		(SEND SELF :TYPE) :VERSION (SEND SELF :VERSION))
	  :STRING-FOR-PRINTING)
	(FORMAT () "~a:~a"		   ;For tapes made under Rel 1.0 which,
		(SEND SELF :HOST)	   ;didn't have :Source-Pattern,
		(SEND
		  (MAKE-PATHNAME :HOST SI:LOCAL-HOST	   ;we kludge a print name.
				 :DIRECTORY (SEND SELF :DIRECTORY) :NAME (SEND SELF :NAME) :TYPE
				 (SEND SELF :TYPE) :VERSION (SEND SELF :VERSION))
		  :STRING-FOR-HOST))))) 


(DEFMETHOD (MT-FILE-MIXIN :PATHNAME) ()	   ;SHOULD BE IN MTSTR
  (SEND SELF :TRUENAME)) 



(DEFUN ROUND-UP-TO-block (X)
  (LET ((b-SIZE sys:disk-block-byte-size))
    (* (CEILING X b-SIZE) b-SIZE))) 
;;
;; Heres an extra nupi file input method.  Should do nothing on Lambda.
;; Use the input file's plist (expected-length) to determine how much to trim
;; from the last block read for a file.  Trim is done by getting a new rqb.
;; Bytes-read counts either 8 or 16 bit bytes to satisfy flavor stream.
;; Expected-length, running-total, record-size etc. are always counting 8-bit bytes.   
;;

(DEFMETHOD (MT-FILE-INPUT-MIXIN :AROUND :NEXT-INPUT-BUFFER) (CONTINUATION MAPPING-TABLE &REST ARG-LIST
							     &AUX SAVE-RECORD-SIZE)
  "Input file's plist (expected-length) determines how much to trim last read."
  (COND
    ((= RUNNING-TOTAL EXPECTED-LENGTH)	   ;NUPI/EXPLORER - fake an eof
     (VALUES () () ()))
    (T					   ;NUPI/EXPLORER - check if trimming is needed
     (LET (BUF
	   LOW
	   BYTES-READ
	   (TRIM-SIZE (ROUND-UP-TO-block (- EXPECTED-LENGTH RUNNING-TOTAL))))
       (IF (< TRIM-SIZE RECORD-SIZE)
	   (IF (< TRIM-SIZE (* sys:disk-block-byte-size MAX-blocks-PER-RQB))
	       (PROGN
		 ()
		 (SETF SAVE-RECORD-SIZE RECORD-SIZE)
		 (SETF RECORD-SIZE TRIM-SIZE))
	       (FERROR () "~&Attempt to make too big of an rqb, size= ~d. " TRIM-SIZE)))
       (MULTIPLE-VALUE-SETQ (BUF LOW BYTES-READ)
	 (LEXPR-FUNCALL-WITH-MAPPING-TABLE CONTINUATION MAPPING-TABLE ARG-LIST))
       (IF BYTES-READ
	   (LET ((BYTE-FACTOR (IF (> BYTE-SIZE 8)
				  2
				  1)))	   ;used to be (// byte-size 8.)
	     (IF SAVE-RECORD-SIZE
		 (SETF RECORD-SIZE SAVE-RECORD-SIZE
		       SAVE-RECORD-SIZE ()))
	     (IF (<= TRIM-SIZE RECORD-SIZE)	   ; NOTE: if we trimmed or not 
		 (PROGN
		   ()
		   (SETQ BYTES-READ
			 (QUOTIENT (- EXPECTED-LENGTH RUNNING-TOTAL) BYTE-FACTOR))
		   (SETQ RUNNING-TOTAL EXPECTED-LENGTH))
		 (PROGN
		   (SETQ RUNNING-TOTAL (+ RUNNING-TOTAL (* BYTES-READ BYTE-FACTOR)))
		   (IF (> RUNNING-TOTAL EXPECTED-LENGTH)
		       (SETQ BYTES-READ
			     (+ BYTES-READ
				(QUOTIENT (- EXPECTED-LENGTH RUNNING-TOTAL) BYTE-FACTOR))))))))
       (PROGN
	 ()
	 (VALUES BUF LOW BYTES-READ)))))) 


;; This takes keywords compatibly with OPEN
;; Probe openings are not implemented.
;; This function expects the magtape to be at the beginning of a file header.
;; Errors returned unique to magtape are:
;;  PNA F Probe openings not allowed
;;  EOT F End of Tape
(DEFUN MAKE-MT-FILE-STREAM (&OPTIONAL
			    &REST OPTIONS
			    &KEY
			    (DIRECTION :INPUT)
			    (CHARACTERS :DEFAULT)
			    (BYTE-SIZE :DEFAULT)
			    (UNIT *NUPI-DEFAULT-UNIT*)
			    (RECORD-SIZE *default-record-size*)
			    (DENSITY 0)
			    (IBM-MODE NIL)
			    (PLIST NIL)
			    (HEADER-STYLE :MIT)	   ;Default for write
			    (PADDED T)
			    &ALLOW-OTHER-KEYS
			    &AUX STREAM TEM)
  "On input if called with PLIST, then assume header has been read and tape is positioned after header."
  (setq tape-header-length *NUPI-TAPE-HEADER-LENGTH*)
  (COND
    ((EQ DIRECTION :INPUT)
     (if plist 
	 nil				   ;don't read a new plist if header has been passed on tape
	 (unwind-protect
	     (MULTIPLE-VALUE-SETQ (PLIST HEADER-STYLE)
	       (READ-MAGTAPE-HEADER
		 (SETQ STREAM (APPLY 'MAKE-MT-STREAM :CHARACTERS T :BYTE-SIZE 8
				     :UNIT UNIT :DENSITY DENSITY OPTIONS))))
	   (if (streamp stream)
	       (FUNCALL STREAM :CLOSE :RAW))))
     (IF (NULL PLIST)
	 (MAKE-EXPLORER-OPEN-ERROR unit)
	 (progn
	   (unless (GETF PLIST :PARTITION) ;Its a partition 6.22.87
	     (progn
	       (AND (EQ BYTE-SIZE :DEFAULT) (SETQ BYTE-SIZE (GETF PLIST :BYTE-SIZE)))	   ;Moved 6.16.87
	       (AND (EQ CHARACTERS :DEFAULT)
		    (IF (SETQ TEM (GETL (LOCF PLIST) '(:CHARACTERS)))
			(SETQ CHARACTERS (CADR TEM))
			(let* ((type (getf plist :type))
			       (truename   ;Don't guess, ask user.  6.16.87
				 (fs:make-pathname :host (or (si:parse-host (getf plist :host) t)
							     si:local-host)	   ;Fix 
						   :directory (getf plist :directory)
						   :name (getf plist :name)
						   :type type
						   :device (getf plist :device)
						   :version (getf plist :version)))
			       (canonical-type (send truename :canonical-type)))
			  (MULTIPLE-VALUE-SETQ (CHARACTERS BYTE-SIZE)
			    (FS:DETERMINE-IF-TYPE-KNOWN canonical-type type characters))   ;6.22.87
			  (unless byte-size	   ;nothing known, ask user  6.16.87
			    (MULTIPLE-VALUE-SETQ (CHARACTERS BYTE-SIZE)
			      (FS:ASK-USER-FOR-CHARACTERS-AND-BYTE-SIZE TRUENAME CHARACTERS BYTE-SIZE))
			    (IF (NOT (MEMBER TYPE '(NIL :UNSPECIFIC) :TEST #'EQUAL))
				(SETQ FS:*USER-DEFINED-TYPES*	   ;and remember user's answer.
				      (ACONS TYPE (LIST CHARACTERS BYTE-SIZE) FS:*USER-DEFINED-TYPES*)))))))
	       
	       (AND (NULL BYTE-SIZE) (SETQ BYTE-SIZE (IF CHARACTERS 8 16)))))
	   
	   (setq stream
		 (MAKE-INSTANCE
		   (IF CHARACTERS
		       (if (eq (get 'mt:*current-unit* 'block-mode) :variable)
			   'reel-mt-file-character-input-stream
			   (if (and *ASCII-PADDED-ENABLED* (getf plist :padded))
			       'MT-FILE-CHARACTER-INPUT-STREAM-PADDED
			       'MT-FILE-CHARACTER-INPUT-STREAM))		       
		       (if (eq (get 'mt:*current-unit* 'block-mode) :variable)
			   'reel-mt-file-input-stream
			   'MT-FILE-INPUT-STREAM))
		   :BYTE-SIZE BYTE-SIZE :UNIT UNIT :RECORD-SIZE RECORD-SIZE :DENSITY DENSITY
		   :PROPERTY-LIST PLIST :HEADER-FORMAT HEADER-STYLE :IBM-MODE IBM-MODE))
	   
	   (IF (GETF PLIST :PARTITION)	   ;its a partition
	       (send stream :set-expected-length   ;record-size, bytes are 8-bit
		     (* (GET (LOCF PLIST) :size) si:disk-block-byte-size)) ;size is in pages
	       (PROGN
		 (send stream :set-expected-length ;Not a partition
		       (COND
			 ((= BYTE-SIZE 16) ;expected is 8-bit bytes always
			  (* 2 (OR (GETF PLIST :LENGTH-IN-BYTES)
				   (GETF PLIST :LENGTH))))
			 (T (OR (GETF PLIST :LENGTH-IN-BYTES)	   ;which ever one
				(GETF PLIST :LENGTH)))))   ;we're using today !
		 (SEND STREAM :SET-RECORD-SIZE
		       (min  (* (if (eq (get 'mt:*CURRENT-UNIT* 'BLOCK-MODE) :VARIABLE)
				    VARIABLE-BLOCK-MAX-BLOCKS-PER-RQB
				    MAX-BLOCKS-PER-RQB)
				SYS:DISK-BLOCK-BYTE-SIZE)
			     (round-up-to-block (send stream :EXPECTED-LENGTH))))))
	   (send STREAM :SET-RUNNING-TOTAL 0)
	   STREAM)))
    
    ((EQ DIRECTION :OUTPUT)
     
     (AND (EQ BYTE-SIZE :DEFAULT) (SETQ BYTE-SIZE (GETF PLIST :BYTE-SIZE)))
     (AND (EQ CHARACTERS :DEFAULT) (SETQ CHARACTERS (GETF PLIST :CHARACTERS)))	   ;6.16.87
     (AND (NULL BYTE-SIZE) (SETQ BYTE-SIZE (IF CHARACTERS 8 16)))
     
     (let ((real-characters (if (getf plist :characters) t nil)))  ;6.22.87
       (setf (getf plist :characters) real-characters))	   ;Remember as much as we can
     
     (when (and padded *ASCII-PADDED-ENABLED* (getf plist :characters))	   ;7.14.87
       (setf (getf plist :PADDED) T))	   ;Declare PADDED in header written to tape.
     
     (unwind-protect
	 (WRITE-MAGTAPE-HEADER
	   (SETQ STREAM (APPLY 'MAKE-MT-STREAM
			       :DIRECTION :OUTPUT
			       :CHARACTERS T
			       :BYTE-SIZE 8
			       :UNIT UNIT
			       :RECORD-SIZE *NUPI-TAPE-HEADER-LENGTH* 
			       :DENSITY DENSITY
			       OPTIONS))
	   PLIST HEADER-STYLE)
       (if (STREAMP STREAM) (FUNCALL STREAM :CLOSE :RAW)))
     
     (SETQ STREAM (MAKE-INSTANCE
		    (IF CHARACTERS
			(if (eq (get 'mt:*current-unit* 'block-mode) :variable)
			    'reel-mt-file-character-output-stream
			    'MT-FILE-CHARACTER-OUTPUT-STREAM)
			(if (eq (get 'mt:*current-unit* 'block-mode) :variable)
			    'reel-mt-file-output-stream
			    'MT-FILE-OUTPUT-STREAM))
		    :BYTE-SIZE BYTE-SIZE :UNIT UNIT :RECORD-SIZE RECORD-SIZE :DENSITY DENSITY
		    :PROPERTY-LIST PLIST :HEADER-FORMAT HEADER-STYLE :IBM-MODE IBM-MODE))
     (UNLESS (GETF PLIST :PARTITION)
       (SEND STREAM :SET-RECORD-SIZE
	     (MIN (* (if (eq (get 'mt:*current-unit* 'block-mode) :variable)
			 variable-block-max-blocks-per-rqb
			 max-blocks-per-rqb)
		     sys:disk-block-byte-size)
		  
		  (ROUND-UP-TO-block
		    (COND
		      ((= BYTE-SIZE 16)
		       (* 2 (OR (GETF PLIST :LENGTH-IN-BYTES) (GETF PLIST :LENGTH))))
		      (T (OR (GETF PLIST :LENGTH-IN-BYTES) ;which ever one
			     (GETF PLIST :LENGTH))))))))
     STREAM)
    (T (FERROR NIL "Probe opens not allowed on magtape."))))



(DEFUN MAKE-EXPLORER-OPEN-ERROR (UNIT)
  (LET* ((ERR (SI::GET-LAST-ERROR *CURRENT-UNIT*))
	 (ERR-STRING
	   (SELECT ERR (:EOF "End of file on unit ~D.")
		   (:END-OF-RECORDED-MEDIA "End of recorded media on unit ~D.")
		   (T "Unknown error on unit ~D."))))
    (MAKE-CONDITION ERR ERR-STRING UNIT))) 


(DEFUN MT-OPEN-ERROR (CONDITION ERROR-P)
  (IF ERROR-P
      (SIGNAL-CONDITION CONDITION)
      CONDITION)) 


(COMPILE-FLAVOR-METHODS MT-FILE-INPUT-STREAM MT-FILE-CHARACTER-INPUT-STREAM
			MT-FILE-OUTPUT-STREAM MT-FILE-CHARACTER-OUTPUT-STREAM
			MT-FILE-CHARACTER-INPUT-STREAM-PADDED)	   ;7.14.87

;;; RESTORE-MAGTAPE

;; This is the driver for RESTORE-MAGTAPE and friends.
;; The first value returned describes what it found on the magtape.
;; Possible values are:
;;  :EOF		An end of file block was found.
;;  :STREAM		A header block was found and sucessfully decoded.
;;			Second value is an input stream, from which can be found
;;			properties, truenames, partition-p, etc.
;;  :TAPE-HEADER	A block was found which was a tape header.
;;			The plist from this header is the second value.
;;  :HEADER		This is some sort of header which is not a tape header.
;;			The plist from this header is the second value.
;;  :ERROR		A tape header couldn't be decoded.
;;			Reason (a string) is the second value.

;This tries to be just enuf to allow opening MT: for input or output.
; It specifially tries to make no assumptions about pathnames.

(DEFFLAVOR MT-HOST
	   (DEVICE-NAME)
	   (SI:BASIC-HOST)
  (:GETTABLE-INSTANCE-VARIABLES DEVICE-NAME)
  (:INITABLE-INSTANCE-VARIABLES DEVICE-NAME)) 


(DEFMETHOD (MT-HOST :NAME) ()
  DEVICE-NAME) 

(DEFMETHOD (mt-host :SYSTEM-TYPE) ()
  :MT)
(DEFMETHOD (mt-host :host-version) ()
  host:*current-host-version*)

(DEFMETHOD (mt-host :cache-timestamp) ()
  0)

(DEFMETHOD (mt-host :local-host-p) ()
  nil)

(DEFMETHOD (mt-host :initialize) (&rest ignore)
  nil)

(DEFMETHOD (MT-HOST :NAME-AS-FILE-COMPUTER) ()
  DEVICE-NAME) 

(DEFMETHOD (MT-HOST :PATHNAME-HOST-NAMEP) (NAME)
  (STRING-EQUAL NAME DEVICE-NAME)) 


(DEFMETHOD (MT-HOST :PATHNAME-FLAVOR) ()
  (VALUES 'MT-PATHNAME ())) 


(DEFFLAVOR MT-PATHNAME
	   ()
	   (PATHNAME)) 


(DEFMETHOD (MT-PATHNAME :PARSE-NAMESTRING) (HOST-SPECIFIED STRING &OPTIONAL (START 0) END)
  HOST-SPECIFIED
  (VALUES :NO-INTERN (MAKE-INSTANCE 'MT-FILEHANDLE :NAMESTRING (SUBSeq STRING START END)))) 


;This is a kludge to make the copy-patch-files-of-system work.

(DEFMETHOD (MT-FILEHANDLE :PATCH-FILE-PATHNAME) (&REST IGNORE)
  "MT:") 


(DEFUN PUTPROP-MAYBE (PLIST VALUE PROP)
  (WHEN (AND VALUE (NULL (GET PLIST PROP)))
    (SETF (GET PLIST PROP) VALUE))) 




(DEFMETHOD (MT-FILEHANDLE :PROPERTIES) (&REST IGNORE)	   ;hack
  (LET* ((IN-STREAM (SEND SELF :OPEN () :DEFAULTS-FROM-STREAM))
	 (PROP-LIST (SEND IN-STREAM :PROPERTY-LIST))
	 (PATHNAME
	   (MAKE-PATHNAME :DIRECTORY (SEND IN-STREAM :DIRECTORY) :NAME (SEND IN-STREAM :NAME)
			  :TYPE (SEND IN-STREAM :TYPE) :VERSION (SEND IN-STREAM :VERSION))))
    (VALUES (CONS PATHNAME PROP-LIST) ())))	   ;plist + settable flag


(Defmethod (Mt-Filehandle :Open) (Ignore &Rest Keyword-Args
				  &Key &Optional
				  (Direction :Input) Defaults-From-Stream Byte-Size
				  Author Host
				  Source-Pattern &Allow-Other-Keys)
  
  (Cond
    ((Eq Direction :Input) (Apply 'Make-Mt-File-Stream Keyword-Args))
    ((Eq Direction :Output)
     (Cond (Defaults-From-Stream
	    (Let* ((source-pattern-string
		     (Fs::Pathname-Object-To-String
		       (If *Logical-Pathnames-On* Source-Pattern
			   (Translated-Pathname Source-Pattern))))
		   (Truename (Funcall Defaults-From-Stream :Truename))	   ;5.15.87
		   (best-host (if *Logical-Pathnames-On*
				  (send (send source-pattern :host) :short-name)   ;4.13.87 Skip Domain
				  (Or (if (stringp host)
					  (name:split-name host)   ;split simple string 4.13.87
					  (and host (send (si:parse-host host) :short-name)))	   ;be sure its a host 4.13.87
				      Si:Local-Host-Name)))
		   (Plist
		     (Filter-Plist (Funcall Defaults-From-Stream :Plist) NIL))	   ;2.24.87 MBC
		   (Real-Plist (Locf Plist)))
	      (Putprop-Maybe Real-Plist Byte-Size :Byte-Size)
	      (If (Getf Plist :Characters) (Putprop-Maybe Real-Plist 8. :Byte-Size))
	      (Putprop Real-Plist source-pattern-string :Source-Pattern)
	      (Putprop-Maybe Real-Plist Author :Author)
	      (Putprop Real-Plist 
		       (Send (FS:TRUENAME-AS-LOGICAL Truename source-pattern) :Directory)  ;5.18.87
		       :Directory)
	      (Putprop Real-Plist (Send Truename :Name) :Name)
	      (Putprop Real-Plist (Send Truename :Type) :Type)
	      (Putprop Real-Plist (Send Truename :Version) :Version)
	      (Putprop Real-Plist (Send (si:parse-host best-host) :SYSTEM-TYPE) :SYSTEM-TYPE)	   ;3.17.87
	      (Setf (Get Real-Plist :Backup-Date) (Get-Universal-Time))	   
	      (Putprop Real-Plist best-host :Host) ;MBC 7-30-86
	      ;;;Don't mess with :UNSPECIFIC types... just pass 'em along.  6.16.87	      
	      (Apply 'Make-Mt-File-Stream :Plist Plist Keyword-Args)))
	   (T (Ferror () "MT: wins only for copying with stream default"))))
    (T (Ferror () "mt-filehandle :open had no direction" T))))



;TRY TO GET RID OF ANY UNPRINTABLE THINGS FROM PLIST-TO-FILTER..

(DEFUN FILTER-PLIST (PLIST-TO-FILTER PLIST-SO-FAR)
  (COND
    ((NULL PLIST-TO-FILTER) PLIST-SO-FAR)
    ((OR (NULL (CADR PLIST-TO-FILTER)) (NUMBERP (CADR PLIST-TO-FILTER))
	 (STRINGP (CADR PLIST-TO-FILTER)) (CONSP (CADR PLIST-TO-FILTER))
	 (SYMBOLP (CADR PLIST-TO-FILTER)))
     (FILTER-PLIST (CDDR PLIST-TO-FILTER)
		   (CONS (CAR PLIST-TO-FILTER) (CONS (CADR PLIST-TO-FILTER) PLIST-SO-FAR))))
    (T (FILTER-PLIST (CDDR PLIST-TO-FILTER) PLIST-SO-FAR)))) 


(DEFUN ADD-MT-HOST (&OPTIONAL (NAME "MT"))
  (COND
    ((NULL (si:parse-host name t nil))	   ;use si:parse-host instead of fs:get-pathname-host. 3/6/87 mp. 
     (LET ((HOST (MAKE-INSTANCE 'MT-HOST :DEVICE-NAME NAME)))
       (host:add-foreign-host-to-namespace  host)))
    ))

;(DEFUN DELETE-MT-HOST (&OPTIONAL (PKG "FS") (NAME "MT"))
;  (COND
;    ((FS:GET-PATHNAME-HOST NAME T)
;     (LET ((*PACKAGE* (FIND-PACKAGE PKG)))
;       (LET ((HOST (FS:GET-PATHNAME-HOST NAME T)))
;	 (SETQ FS:*PATHNAME-HOST-LIST*
;	       (DELETE HOST (THE LIST FS:*PATHNAME-HOST-LIST*) :TEST #'EQ))))))) 


(COMPILE-FLAVOR-METHODS MT-HOST MT-PATHNAME MT-FILEHANDLE) 

;(DELETE-MT-HOST "FS") 				;be sure we're not using old FS mt-host


;(ADD-MT-HOST)  3/25/87 mp

(ADD-INITIALIZATION "Add MT Host"
		    '(mt:add-mt-host)
		    '(:now) 'net:*network-warm-initialization-list*)	   ;3/25/87 mp



;;;  REVISED:
;;;   	First released in patch release 2.1.   5-6-86 by MBC.
;;;
;;;  Support for REEL half inch start-stop kinds of tapes.
;;;  Specifically the CIPHER MT3201 Tape Drive.
;;;
;;;  Source (commented out) is provided for some test drivers,
;;;  like Simple-Drive-Test-Control, in this file.
;;;

;;;  This code provides variable block support.  It uses the
;;;  Nupi command option bit, as defined in Tape Patch 2.14,
;;;  called Si:%Nupi-Cmd-Option-Variable.Or.Device-Block,
;;;  to enable variable block mode in the Nupi subsystem and
;;;  in the tape drive's formatter.  When this bit is on, the
;;;  record-size of the stream defines the block size.  On 
;;;  reads this is an "expected" block size and the total I/O
;;;  length may be longer than the actual request length.
;;;  An error "SCSI command completed w/o data transfer" is
;;;  a likely outcome of the Nupi attempting to transfer more
;;;  data than the Lisp system has room for.

;;;  Use MAKE-REEL-MT-STREAM to create this kind of stream.


(DEFFLAVOR REEL-MT-MIXIN
	   (RQB
	    EXTENDED-STATUS)
	   (MT-MIXIN)
  :SETTABLE-INSTANCE-VARIABLES
  :GettABLE-INSTANCE-VARIABLES
  :iniTABLE-INSTANCE-VARIABLES) 



(defmethod (reel-mt-mixin :AFTER :INIT) (&rest ignore)
  (declare (special variable-block-max-blocks-per-rqb))
  (setf record-size (* variable-block-max-blocks-per-rqb sys:disk-block-byte-size)))

(DEFMETHOD (reel-MT-MIXIN :CLOSE) (&OPTIONAL ABORTP)
  (OR (EQ ABORTP :RAW) (send SELF :ADVANCE-FILE))
  (COND ((NEQ STATUS :CLOSED)
	 (SETQ STATUS :CLOSED)
	 (AND RQB (SI:RETURN-DISK-RQB RQB))
	 (SETQ RQB NIL)
	 T))) 


(DEFMETHOD (REEL-MT-MIXIN :GET-EXTENDED-STATUS) ()
  (VALUES-LIST EXTENDED-STATUS)) 


(DEFMETHOD (reel-MT-MIXIN :RETURN-RQB) ()
  (COND (RQB (SI:RETURN-DISK-RQB RQB)
	     (SETQ RQB NIL))))

(DEFFLAVOR REEL-MT-INPUT-MIXIN
	   ()
	   (REEL-MT-MIXIN)) 


(DEFMETHOD (REEL-MT-INPUT-MIXIN :EOF-STATUS) ()	   ;True if end-of-file reached on nupi tape
  (LET ((LAST-ERROR (SI::GET-LAST-ERROR UNIT)))
    (OR (EQ LAST-ERROR :EOF) (EQ LAST-ERROR :END-OF-RECORDED-MEDIA)))) 



(DEFPARAMETER SI::TAPE-ILI-ERROR-CODE 73) 

;;; 49 	ILI
;;; 4A  EOM
;;; 4B	ILI, EOM
;;; 4C  FM
;;; 4D  ILI, FM
;;; 4E  EOM, FM
;;; 4F  ILI, EOM, FM



(defvar reel-ili-error-codes '(#x49 #x4B #x4D #x4F))	
(defvar reel-eorm-error-codes '(#x4A #x4B #x4E #x4F))	
(defvar reel-eof-error-codes '(#x4C #x4D #x4E #x4F))


;;; Add ILI recognition & simplify Extended-Status contruction.  8.10.87 MBC
(defmethod (reel-mt-input-mixin :NEXT-INPUT-BUFFER)
	   (&optional ignore)
  (declare (special variable-block-max-blocks-per-rqb))
  (let (buffer bytes-transferred error-code residue real-block-size)
    (COND
      ((EQ status :closed)
       (ferror 'si:tape-error "attempt to get input from ~s, which is closed." self)) ;8.18.87
      ((eq status :eof) nil)
      (t
       (or rqb
	   (setq rqb (si:get-disk-rqb  variable-block-max-blocks-per-rqb)))
       (condition-case (error)
	   (mt-run rqb %mt-command-read (- record-size) unit 0 nil :variable-block-size t)
	 (system:tape-error
	  (unless (= si:tape-ili-error-code
		     (ldb si:%%nupi-status-low-dev-error (aref rqb si:%io-rq-status)))
	    (ferror 'mt:tape-media-error "~a" (send error :report-string)))))	   ;residue
       (setq buffer (mt-get-rqb-array rqb byte-size))
       (SETQ error-code (ldb si:%%nupi-status-low-dev-error (aref rqb si:%io-rq-status)))
       (COND
	 ((send self :eof-status)
	  (setq status :eof
		bytes-transferred 0))
	 (t
	  (SETQ residue (aref rqb si:%io-rq-transfer-length))	   ;assumes halfword accuracy is ok.
	  (SETQ bytes-transferred residue)
	  (if (member error-code reel-ili-error-codes :test #'=)
	      (cond ((ldb-test (byte 1 15) residue)	   ;if residue is neg.
		     ;;; Then residue tells what was lost from reading with too short a block size
		     ;;; That is there was more data than expected.
		     (setq bytes-transferred record-size   ;8.3.87
			   real-block-size ;8.3.87
			   (- #x10000 residue)))
		    (t (setq bytes-transferred
			     (- record-size residue))))))) ;else xfer size was less than expected.
       (setq extended-status
	     (list bytes-transferred
		   (if (member error-code reel-eof-error-codes :test #'=) :EOF nil)
		   (if (member error-code reel-eorm-error-codes :test #'=) :EOT nil)
		   (if (member error-code reel-ILI-error-codes :test #'=) :ILI nil)
		   real-block-size))	   ;8.3.87
       (when debug (format t "~&*** REEL :NEXT-INPUT-BUFFER, Extended-status = ~a, Length=~d.~&~a"
			   extended-status (quotient (* 8 bytes-transferred) byte-size)
			   (subseq (SEND self :buffer) 0  (quotient (* 8 bytes-transferred) byte-size))))
       (values buffer 0  (quotient (* 8 bytes-transferred) byte-size))))))




(DEFMETHOD (reel-MT-INPUT-MIXIN :NEXT-FILE) ()
  (IF (EQ STATUS :EOF)
      (SETQ STATUS :OPEN)
      (SPACE-TO-EOF UNIT)
      (send SELF :CLEAR-INPUT)))

(DEFMETHOD (reel-MT-INPUT-MIXIN :ADVANCE-FILE) ()
  (COND ((NOT (MEMber STATUS '(:EOF :CLOSED) :test #'eq))
	 (SPACE-TO-EOF UNIT)
	 (SETQ STATUS :EOF))))

(DEFFLAVOR REEL-MT-OUTPUT-MIXIN
	   ()
	   (REEL-MT-MIXIN)) 



(defmethod (reel-mt-output-mixin :NEW-OUTPUT-BUFFER) ()
  (declare (special variable-block-max-blocks-per-rqb))
  (or rqb
      (setq rqb (si:get-disk-rqb    variable-block-max-blocks-per-rqb)))
  (let ((new-array (mt-get-rqb-array rqb byte-size)))
    (values new-array 0 (length new-array)))) 


(DEFMETHOD (REEL-MT-OUTPUT-MIXIN :SEND-OUTPUT-BUFFER) (BUFFER TO-INDEX)
  (OR
    (AND (ARRAY-INDIRECT-P BUFFER)
	 (EQ
	   (si:%P-CONTENTS-OFFSET BUFFER (1+ (si:%P-LDB-OFFSET SYSTEM:%%ARRAY-LONG-LENGTH-FLAG BUFFER 0)))
	   RQB))
    (FERROR () "attempt to :send-output-buffer ~s, which is not indirected into ~s" BUFFER RQB))
  (LET ((COUNT
	  (QUOTIENT (* (min to-index record-size) BYTE-SIZE) 8)))
    
    (COND
      ((PLUSP COUNT)
       (CONDITION-CASE (ERROR)
	   (MT-RUN RQB %MT-COMMAND-WRITE (- COUNT) UNIT 0 () :VARIABLE-BLOCK-SIZE T)
	 (SYSTEM:TAPE-ERROR		   ;just catch ILI condition so it doesn't trap
	  (UNLESS (= SI::TAPE-ILI-ERROR-CODE
		     (LDB SYSTEM:%%NUPI-STATUS-LOW-DEV-ERROR (AREF RQB SYSTEM:%IO-RQ-STATUS)))
	    (SIGNAL ERROR))))		   ;if not ILI throw again.
       ))))     



(DEFMETHOD (REEL-MT-OUTPUT-MIXIN :AFTER :SEND-OUTPUT-BUFFER) (&REST IGNORE)
  (LET* ((ERROR-CODE (LDB SYSTEM:%%NUPI-STATUS-LOW-DEV-ERROR (AREF RQB SYSTEM:%IO-RQ-STATUS)))
	 (RESIDUE (AREF RQB SYSTEM:%IO-RQ-TRANSFER-LENGTH))
	 (BYTES-TRANSFERRED RESIDUE))
    (IF (= ERROR-CODE SI::TAPE-ILI-ERROR-CODE)	   ;bytes-transferred-mismatch
	(COND
	  ((LDB-TEST (BYTE 1 15) RESIDUE)  ;if residue is neg. 
	   (SETQ BYTES-TRANSFERRED (+ (- 65536 RESIDUE) RECORD-SIZE)))	   ;then xfer size was greater than expected.
	  (T (SETQ BYTES-TRANSFERRED (- RECORD-SIZE RESIDUE)))))   ;else xfer size was less than expected.
    (SETQ EXTENDED-STATUS
	  (APPEND (LIST BYTES-TRANSFERRED)
		  (CASE ERROR-CODE
		    (SI::NUPI-EOF '(:EOF NIL))
		    (SI::NUPI-EORM '(NIL :EOT))
		    (SI::NUPI-EORM+EOF '(:EOF :EOT))
		    (OTHERWISE '(NIL NIL))))))) 



(DEFFLAVOR REEL-MT-CHARACTER-INPUT-STREAM
	   ()
	   (REEL-MT-INPUT-MIXIN
	    MT-CHARACTER-INPUT-STREAM)) 


(DEFFLAVOR REEL-MT-CHARACTER-OUTPUT-STREAM
	   ()
	   (REEL-MT-OUTPUT-MIXIN
	    MT-CHARACTER-OUTPUT-STREAM)) 


(DEFFLAVOR REEL-MT-INPUT-STREAM
	   ()
	   (REEL-MT-INPUT-MIXIN
	    MT-INPUT-STREAM)) 


(DEFFLAVOR REEL-MT-OUTPUT-STREAM
	   ()
	   (REEL-MT-OUTPUT-MIXIN
	    MT-OUTPUT-STREAM)) 

;;AB 7-13-87.  Remove reference to PROCESSOR-TYPE-CODE (old LAMBDA support).
(DEFUN MAKE-REEL-MT-STREAM (&OPTIONAL &KEY (DIRECTION :INPUT) (BYTE-SIZE 8) (CHARACTERS T)
			    (UNIT *NUPI-DEFAULT-UNIT*)
			    (RECORD-SIZE *DEFAULT-RECORD-SIZE*) (DENSITY 0)
			    (IBM-MODE NIL) &ALLOW-OTHER-KEYS)
  (declare (special *NUPI-DEFAULT-UNIT*  *DEFAULT-RECORD-SIZE*))
  (CHECK-ARG DIRECTION (MEMBER DIRECTION '(:INPUT :OUTPUT) :TEST #'EQ) ":input or :output")
  (CHECK-TYPE BYTE-SIZE :NUMBER)
  (CHECK-ARG CHARACTERS (MEMBER CHARACTERS '(T NIL) :TEST #'EQ) "t or nil")
  (MAKE-INSTANCE
    (IF (EQ DIRECTION :INPUT)
	(IF CHARACTERS
	    'REEL-MT-CHARACTER-INPUT-STREAM
	    'REEL-MT-INPUT-STREAM)
	(IF CHARACTERS
	    'REEL-MT-CHARACTER-OUTPUT-STREAM
	    'REEL-MT-OUTPUT-STREAM))
    :BYTE-SIZE BYTE-SIZE :UNIT UNIT :RECORD-SIZE RECORD-SIZE :DENSITY DENSITY :IBM-MODE IBM-MODE)) 


;;;
;;;  Status methods for REEL streams.
;;;



(DEFMETHOD (REEL-MT-INPUT-STREAM :SET-STATUS) (STAT)
  (SETQ STATUS STAT)) 



(DEFMETHOD (REEL-MT-OUTPUT-STREAM :TERMINATE-OUTPUT-STREAM) (&OPTIONAL (PAD 0))
  (IF PAD
      (SEND SELF :PAD-AND-SEND-BUFFER PAD))
  (SEND SELF :CLOSE)) 



(DEFMETHOD (REEL-MT-INPUT-STREAM :DESCRIBE-STREAM) ()
  (FORMAT T "~%                 RQB: ~A" RQB)
  (FORMAT T "~%     EXTENDED-STATUS: ~A"
	  (IF (VARIABLE-BOUNDP EXTENDED-STATUS)
	      EXTENDED-STATUS
	      'UNBOUND))
  (FORMAT T "~%              STATUS: ~A" (IF (VARIABLE-BOUNDP STATUS)
					     STATUS
					     'UNBOUND))
  (FORMAT T "~%  STREAM-INPUT-INDEX: ~A"
	  (IF (VARIABLE-BOUNDP SI::STREAM-INPUT-INDEX)
	      SI::STREAM-INPUT-INDEX
	      'UNBOUND))
  (FORMAT T "~%  STREAM-INPUT-LIMIT: ~A"
	  (IF (VARIABLE-BOUNDP SI::STREAM-INPUT-LIMIT)
	      SI::STREAM-INPUT-LIMIT
	      'UNBOUND))) 



(DEFMETHOD (REEL-MT-OUTPUT-STREAM :DESCRIBE-STREAM) ()
  (FORMAT T "~%                 RQB: ~A" RQB)
  (FORMAT T "~%     EXTENDED-STATUS: ~A"
	  (IF (VARIABLE-BOUNDP EXTENDED-STATUS)
	      EXTENDED-STATUS
	      'UNBOUND))
  (FORMAT T "~%              STATUS: ~A" (IF (VARIABLE-BOUNDP STATUS)
					     STATUS
					     'UNBOUND))
  (FORMAT T "~%  STREAM-OUTPUT-INDEX: ~A"
	  (IF (VARIABLE-BOUNDP SI::STREAM-OUTPUT-INDEX)
	      SI::STREAM-OUTPUT-INDEX
	      'UNBOUND))
  (FORMAT T "~%  STREAM-OUTPUT-LIMIT: ~A"
	  (IF (VARIABLE-BOUNDP SI::STREAM-OUTPUT-LIMIT)
	      SI::STREAM-OUTPUT-LIMIT
	      'UNBOUND))) 


(DEFMETHOD (reel-MT-OUTPUT-MIXIN :ADVANCE-FILE) ()
  (OR (EQ STATUS ':CLOSED)
      (WRITE-EOF UNIT)))



(DEFFLAVOR reel-HEADER-MIXIN
	   (HEADER-FORMAT
	    (SI:PROPERTY-LIST NIL))
	   (SI:PROPERTY-LIST-MIXIN)
  (:INCLUDED-FLAVORS reel-MT-MIXIN)
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES
  :INITABLE-INSTANCE-VARIABLES)

(DEFFLAVOR reel-HEADER-INPUT-MIXIN
	   ()
	   (reel-HEADER-MIXIN))

(DEFFLAVOR reel-HEADER-OUTPUT-MIXIN
	   ()
	   (reel-HEADER-MIXIN))

(DEFFLAVOR reel-MT-FILE-MIXIN
	   ()
	   ()
  (:INCLUDED-FLAVORS HEADER-MIXIN reel-MT-MIXIN mt-file-mixin))

(DEFFLAVOR reel-MT-FILE-INPUT-MIXIN	   ;nupi - this ok for lambda? extra junk.
	   ((expected-length nil)
	    (running-total))
	   (reel-MT-FILE-MIXIN)
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES
  :INITABLE-INSTANCE-VARIABLES)

(DEFFLAVOR reel-MT-FILE-OUTPUT-MIXIN
	   ()
	   (reel-MT-FILE-MIXIN))

(DEFFLAVOR reel-MT-FILE-INPUT-STREAM
	   ()
	   (reel-MT-FILE-INPUT-MIXIN reel-HEADER-INPUT-MIXIN reel-MT-INPUT-STREAM))


(DEFFLAVOR reel-MT-FILE-CHARACTER-INPUT-STREAM
	   ()
	   (reel-MT-FILE-INPUT-MIXIN reel-HEADER-INPUT-MIXIN reel-MT-CHARACTER-INPUT-STREAM))

(DEFFLAVOR reel-MT-FILE-OUTPUT-STREAM
	   ()
	   (reel-MT-FILE-OUTPUT-MIXIN reel-HEADER-OUTPUT-MIXIN reel-MT-OUTPUT-STREAM))

(DEFFLAVOR reel-MT-FILE-CHARACTER-OUTPUT-STREAM
	   ()
	   (reel-MT-FILE-OUTPUT-MIXIN reel-HEADER-OUTPUT-MIXIN reel-MT-CHARACTER-OUTPUT-STREAM))

;;;
;;;
;;;

;;; Some routines to test variable block mode:
;;;
;;;  Read-Variable-Test  
;;;  Write-Variable-Test
;;;     These two functions read and write a group of variable
;;;     block records to the half inch tape.  The block sizes used
;;;	are defined by *Variable-List*. 

;;; Simple-Drive-Test-Control 
;;;     This just calls, 
;;; Simple-Drive-Test
;;;     which allows selection of one block size and how many
;;;     records to read and write.
;;;
;;; Was hardcoded to unit 6.  Change to use MT:*CURRENT-UNIT*

(COMMENT
  (DEFUN SIMPLE-DRIVE-TEST-CONTROL ()
    (LET ((LOADED NIL)
	  (BKSZ NIL)
	  (NREC NIL)
	  (REPEAT NIL)
	  (BTYPE NIL))
      (FORMAT T "~%~% TAPE DRIVE TEST")
      (*CATCH 'GET-OUT
	      (*CATCH () (SETQ LOADED ())
		      (DO-FOREVER
			(FORMAT T "~%~% Scratch Loaded/On-Line/With-Write-Ring (y or n): ")
			(SETQ LOADED (READ-FROM-STRING (READLINE) 'N))
			(COND
			  ((EQ LOADED 'Y) (RETURN))
			  (T (FORMAT T "~% Invalid Drive Setup - Try Again Or 'ABORT'")
			     (SETQ LOADED ())))))
	      (IF (NULL LOADED)
		  (*THROW 'GET-OUT 'TAPE))
	      (DO-FOREVER
		(*CATCH () (SETQ BKSZ ())
			(DO-FOREVER (FORMAT T "~%~% Block Size In Bytes For Test (>= 800): ")
				    (SETQ BKSZ (READ-FROM-STRING (READLINE) 800))
				    (COND
				      ((AND (TYPEP BKSZ :FIXNUM) (>= BKSZ 800)) (RETURN))
				      (T (FORMAT T "~% Invalid Block Size - Try Again Or 'ABORT'")
					 (SETQ BKSZ ())))))
		(IF (NOT (TYPEP BKSZ :FIXNUM))
		    (*THROW 'GET-OUT 'BLOCK-SIZE))
		(*CATCH () (SETQ BTYPE ())
			(DO-FOREVER (FORMAT T "~%~% Output Block Type (f or v): ")
				    (SETQ BTYPE (READ-FROM-STRING (READLINE) 'F))
				    (COND
				      ((OR (EQ BTYPE 'F) (EQ BTYPE 'V)) (RETURN))
				      (T (FORMAT T "~% Invalid Block Type - Try Again Or 'ABORT'")
					 (SETQ BTYPE ())))))
		(IF (NULL BTYPE)
		    (*THROW 'GET-OUT 'BLOCK-TYPE))
		(*CATCH () (SETQ NREC ())
			(DO-FOREVER
			  (FORMAT T "~%~% Number Of Records To Write (800 Bytes) For Test: ")
			  (SETQ NREC (READ-FROM-STRING (READLINE) 10))
			  (COND
			    ((AND (TYPEP NREC :FIXNUM) (> NREC 0)) (RETURN))
			    (T
			     (FORMAT T "~% Invalid Minimum Number Of Records - Try Again Or 'ABORT'")
			     (SETQ NREC ())))))
		(IF (NOT (TYPEP NREC :FIXNUM))
		    (*THROW 'GET-OUT 'NREC))
		(SEND TERMINAL-IO :CLEAR-SCREEN)
		(SIMPLE-DRIVE-TEST :BLOCK-SIZE BKSZ :BLOCK-TYPE BTYPE :NUMBER-RECORDS NREC)
		(*CATCH () (SETQ REPEAT ()) (FORMAT T "~%~% Repeat Test (y or n): ")
			(SETQ REPEAT (READ-FROM-STRING (READLINE) 'N)))
		(IF (OR (NULL REPEAT) (EQ REPEAT 'N))
		    (*THROW 'GET-OUT 'NO-REPEAT))
		(SEND TERMINAL-IO :CLEAR-SCREEN)))))
  (DEFUN SIMPLE-DRIVE-TEST (&KEY (BLOCK-SIZE 800) (BLOCK-TYPE 'F) (NUMBER-RECORDS 1))
    (LET* ((STREM NIL)
	   (NB8 800)
	   (NB16 400)
	   (AR-16B (MAKE-ARRAY NB16 :TYPE 'ART-16B))
	   (AR-8B (MAKE-ARRAY NB8 :TYPE 'ART-8B :DISPLACED-TO AR-16B))
	   (BKSZ (IF (< BLOCK-SIZE 800)
		     800
		     BLOCK-SIZE))
	   (MREC (* 2 (CEILING BKSZ NB8)))
	   (NREC (IF (< NUMBER-RECORDS 0)
		     1
		     NUMBER-RECORDS))
	   (NBLOCK (CEILING (* NREC NB8) BKSZ))
	   (IND NIL)
	   (END NIL)
	   (PAD (IF (EQ BLOCK-TYPE 'F)
		    0
		    ())))
      (FORMAT T "~%~%***********************************************************************")
      (FORMAT T "~%~% SIMPLE HALF-INCH TAPE DRIVE TEST")
      (FORMAT T
	      "~%   RECORD SIZE = ~A  BLOCK SIZE = ~A  RECORDS = ~A  BLOCKS = ~A  BLOCK TYPE = ~A"
	      NB8 BLOCK-SIZE NREC NBLOCK BLOCK-TYPE)
      (SI::CONFIGURE-DISK-SYSTEM)
      ;;;     After init method does this.
      ;;;    (mt:setup-half-inch-tape :unit (or mt:*current-unit* (mt:guess-tape-unit)))
      (FORMAT T "~%~%***********************************************************************")
      (REWIND)
      (FORMAT T "~%~% WRITE TEST - STARTING ~%")
      (SETQ STREM
	    (MAKE-REEL-MT-STREAM :DIRECTION :OUTPUT :BYTE-SIZE 8 :CHARACTERS () :RECORD-SIZE BKSZ))
      (LOOP FOR REC FROM 1 TO NREC DO (ARRAY-INITIALIZE AR-16B REC) (SEND STREM :STRING-OUT AR-8B)
	    (FORMAT T "~% RECORD WRITTEN ~3,48D" REC))
      (SEND STREM :TERMINATE-OUTPUT-STREAM PAD)
      (WRITE-EOF)
      (WRITE-EOF)
      (WRITE-EOF)
      (FORMAT T "~%~% WRITE TEST - COMPLETED")
      (FORMAT T "~%~%***********************************************************************")
      (REWIND)
      (FORMAT T "~%~% READ TEST - STARTING ~%")
      (SETQ STREM
	    (MAKE-REEL-MT-STREAM :DIRECTION :INPUT :BYTE-SIZE 8 :CHARACTERS () :RECORD-SIZE BKSZ))
      (LOOP FOR REC FROM 1 TO (+ NREC MREC) FOR VALID = T DO
	    (MULTIPLE-VALUE (IND END) (SEND STREM :STRING-IN () AR-8B))
	    (WHEN END
	      (FORMAT T "~% *** EOF ENCOUNTERED ***")
	      (RETURN))
	    (FORMAT T "~% RECORD READ ~3,48D - " REC)
	    (LOOP FOR I FROM 0 BELOW NB16 WHEN (NOT (= REC (AREF AR-16B I))) DO (SETQ VALID ())
		  (RETURN))
	    (COND
	      (VALID (FORMAT T "GOOD"))
	      (T (FORMAT T "BAD"))))
      (SEND STREM :CLOSE)
      (FORMAT T "~%~% READ TEST - COMPLETED")
      (FORMAT T "~%~%***********************************************************************")))
  (DEFVAR *VARIABLE-LIST* '(800 800 10000 4000 800 800))   ;3rd block is 10000, 4th is 4000, etc.
  (DEFUN WRITE-VARIABLE-TEST ()
    (REWIND)
    (LET* ((STREM
	     (MAKE-REEL-MT-STREAM :DIRECTION :OUTPUT :BYTE-SIZE 8 :CHARACTERS () :RECORD-SIZE 800)))
      (DOLIST (BKSZ *VARIABLE-LIST*)
	(SEND STREM :SET-RECORD-SIZE BKSZ)
	(DOTIMES (N BKSZ)
	  (SEND STREM :TYO #\T))
	(FORMAT T "~% RECORD WRITTEN ~3,48D" BKSZ))
      (SEND STREM :CLOSE)
      (WRITE-EOF)
      (REWIND)))
  (DEFUN READ-VARIABLE-TEST ()
    (REWIND)
    (LET* ((VALID T)
	   (STREM
	     (MAKE-REEL-MT-STREAM :DIRECTION :INPUT :BYTE-SIZE 8 :CHARACTERS () :RECORD-SIZE 800)))
      (DOLIST (BKSZ *VARIABLE-LIST*)
	(SEND STREM :SET-RECORD-SIZE BKSZ)
	(DOTIMES (N BKSZ)
	  (LET ((CHAR (SEND STREM :TYI)))
	    (UNLESS (OR (EQUAL CHAR #\T) (NULL CHAR))
	      (SETF VALID ()))))
	(FORMAT T "~% RECORD READ ~3,48D" BKSZ)
	(COND
	  (VALID (FORMAT T "   GOOD"))
	  (T (FORMAT T "   BAD"))))
      (SEND STREM :CLOSE)
      (REWIND))))
