;;; -*- Mode: common-Lisp; Package: MT; Base: 10; Fonts: MEDFNT,HL12B,HL12BI -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;
;;       NUPI status commands - coats 10/11/84 
;; 
;; STATUS COMMAND CODES -- Length of returned status data
;; si:%nupi-command-rq-dev-status   -- si:%nupi-length-rq-dev-status
;; si:%nupi-command-rq-fmtr-status  -- si:%nupi-length-rq-fmtr-status
;; si:%nupi-command-rq-nupi-status  -- si:%nupi-length-rq-nupi-status
;;
;;


(DEFUN DEVICE-STATUS (&OPTIONAL (UNIT *NUPI-DEFAULT-UNIT*))
  "Issue a device status to a nupi UNIT."
  (declare (special *NUPI-DEFAULT-UNIT*))
  (ISSUE-STATUS UNIT SYSTEM:%NUPI-COMMAND-RQ-DEV-STATUS SYSTEM:%NUPI-LENGTH-RQ-DEV-STATUS)) 


(DEFUN FORMATTER-STATUS (&OPTIONAL (UNIT *NUPI-DEFAULT-UNIT*))
  "Issue a formatter status to a nupi UNIT."
  (declare (special *NUPI-DEFAULT-UNIT*))
  (ISSUE-STATUS UNIT SYSTEM:%NUPI-COMMAND-RQ-FMTR-STATUS SYSTEM:%NUPI-LENGTH-RQ-FMTR-STATUS)) 


(DEFUN NUPI-STATUS (&OPTIONAL (UNIT *NUPI-DEFAULT-UNIT*))
  "Issue a NUPI status; Use any UNIT on nupi."
  (declare (special *NUPI-DEFAULT-UNIT*))
  (ISSUE-STATUS UNIT SYSTEM:%NUPI-COMMAND-RQ-NUPI-STATUS SYSTEM:%NUPI-LENGTH-RQ-NUPI-STATUS)) 


(DEFUN ISSUE-STATUS (UNIT CMD LENGTH &AUX RQB RESULT);  "Issue a status and return the status data as an array."
  (UNWIND-PROTECT (PROGN
		   (SETQ RQB (SYSTEM:GET-DISK-RQB)
			 RESULT ())
		   (NUPI-CMD-SETUP RQB CMD 0 UNIT LENGTH)
		   (DOTIMES (I LENGTH)
		     (SETF (AREF RQB (+ I SYSTEM:%IO-RQ-PARAMETER-LIST)) 0))
		   (WIRE-MT-RQB RQB);wire, no ccw/scatter
		   (EXECUTE-MT-RQB RQB)
		   (SETQ RESULT
			 (MAKE-ARRAY LENGTH :element-TYPE '(unsigned-byte 16.)))
		   (COPY-ARRAY-PORTION RQB SYSTEM:%IO-RQ-PARAMETER-LIST
				       (+ LENGTH SYSTEM:%IO-RQ-PARAMETER-LIST) RESULT 0 LENGTH))
    (UNWIRE-MT-RQB RQB)
    (SYSTEM:RETURN-DISK-RQB RQB))
  RESULT) 




(COMMENT;the rest of this file...........
 (DEFUN SHOW-NUPI-STATUS (RQB)
   "Interpert the RQB's status word and print in English.
Return nil if unrecoverable error occured."
   (LET (STAT
	 ERR)
     (SETQ STAT (AREF RQB SYSTEM:%IO-RQ-STATUS-HIGH))
     (FORMAT T "~%Status = ~o (in octal)~%" STAT)
     (IF (= 1 (LDB SYSTEM:%%NUPI-STATUS-HIGH-BUSY STAT))
       (FORMAT T "busy, "))
     (IF (= 1 (LDB SYSTEM:%%NUPI-STATUS-HIGH-COMPLETE STAT))
       (FORMAT T "complete, "))
     (IF (= 1 (SETQ ERR (LDB SYSTEM:%%NUPI-STATUS-HIGH-ERROR STAT)))
       (FORMAT T "error, "))
     (IF (= 1 (LDB SYSTEM:%%NUPI-STATUS-HIGH-RETRY STAT))
       (FORMAT T "retry, "))
     (IF (= 1 (LDB SYSTEM:%%NUPI-STATUS-HIGH-AUX-STAT STAT))
       (FORMAT T "aux-status, "))
     (FORMAT T "~%Controller status is ")
     (SHOW-CLASS-CODE (LDB SYSTEM:%%NUPI-STATUS-HIGH-CNTL-ERROR STAT))
     (SETQ STAT (AREF RQB SYSTEM:%IO-RQ-STATUS))
     (FORMAT T "~%Device status is ")
     (SHOW-CLASS-CODE (LDB SYSTEM:%%NUPI-STATUS-LOW-DEV-ERROR STAT))
     (FORMAT T "~%Retrys = ~d" (LDB SYSTEM:%%NUPI-STATUS-LOW-RETRY STAT))
     (= 0 ERR)));true if no errror
 (DEFUN SHOW-CLASS-CODE (BITS)
   (FORMAT T "class = ~o, " (LDB 503 BITS))
   (FORMAT T "code = ~o" (LDB 5 BITS)))
 (DEFUN SHOW-NUPI-RQ-DEV-STATUS (RQB)
   "Interpert the status buffer returned by a NUPI device status request.
   Print it in plain English."
   (SHOW-6-13-WORD (AREF RQB SYSTEM:%IO-RQ-STATUS-HIGH)));need to pass high half-word
 (DEFUN SHOW-6-13-WORD (HALF-WORD)
   "Interpert the bits according to table 6-13 of the NUPI spec."
   (FORMAT T "~%Device type is ~A.~%"
	   (SELECT (LDB SI::%NUPI-TYPE-RQ-DEV-STATUS HALF-WORD) (1 "tape") (2 "disk")
	      (T "unknown")))
   (IF (= 1 (LDB SI::%NUPI-OFF-LINE-RQ-DEV-STATUS HALF-WORD))
     (FORMAT T "off-line, ")
     (FORMAT T "on-line, "))
   (IF (= 1 (LDB SI::%NUPI-ERROR-RQ-DEV-STATUS HALF-WORD))
     (FORMAT T "had error, "))
   (IF (= 1 (LDB SI::%NUPI-FAILED-RQ-DEV-STATUS HALF-WORD))
     (FORMAT T "failed self-test, "))
   (IF (= 1 (LDB SI::%NUPI-HOT-RQ-DEV-STATUS HALF-WORD))
     (FORMAT T "over temperature, "))
   (IF (= 1 (LDB SI::%NUPI-WRITE-PROTECTED-RQ-DEV-STATUS HALF-WORD))
     (FORMAT T "write-protected, ")
     (FORMAT T "write-enabled, "))
   (IF (= 1 (LDB SI::%NUPI-UNSAFE-RQ-DEV-STATUS HALF-WORD))
     (FORMAT T "HW unsafe, ")))
 |........................|) 
