
;;; -*- Mode:common-LISP; Package:MT; Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.


;;;      REVISIONS:
;;;     03-28-89  DAB  Added code to handle EOM during backup partition and restore partition.
;;;     06-01-88  DAB  Fixed the errors message in verify-boot-band to display correct partition name.
;;;     12-03-87 DAB    Fixed Verify-bootable-tape: Documentation problems and it was not detecting a non-bootable tape.
;;;	9.02.87 MBC	Correct base problem in verify-partition.  Verify of tape w/starting hundred
;;;			of 600 showed numbers starting w/864, and stopped verifying prematurely.
;;;	8.13.87 MBC 	Backup-partition to write-protected tape spaced-to-eof.   Fix it by implementing
;;;			*abort-flag* for backup-partition, just like restore- & verify-partition.
;;;      Fixes for 2K page size.  1-8-87, -ab
;;;      Fix b-p, r-p, & v-p to handle small (< 148 pages) partitions.  Ignore media-errors, too.
;;;        Renamed BUFFER-POOL-CHECK to be BUFFER-POOL-ADJUST.   3/23/85 MBC
;;;      Fix backup-partition, restore-partition & verify- to properly dispose magtape unit. 3/13/85 MBC
;;;      Add Check-For-Error-Or-Status to handle unexpected i/o errors 3/6/85 MBC
;;;      Streaming support for partition copying.  - 12/03/84 MBC
;;;      Major re-write: remove without-interrupts amoung other improvements. 2/27/85 MBC
;;;	Let Deallocate-Q return busy disk rqbs, since system code now can handle it.  Gets rid of
;;;		Clean-Up and *ABORTED-RQB-Q* variable. 12-6-85 MBC
;;;      Fixed restore-partition so that when the user enters no for "Do you really want to clobber your
;;;         partition", he has the option of rewinding instead of issuing a space-to-eof.  2/26/87 mp
;;;      Modified verify-bootable-tape so that the Cpu-Type of the source and destination partitions are
;;;         checked, if there is more than one partition with the same name.    3/11/86
;;;      Modified backup-partition to record the cpu-type of the partition on the header. 3/25/87 mp
;;;      Modified restore-partition to copy the cpu-type of the tape partition to the appropriate disk partition 3/25/87 mp
;;;      Modified verify-partition to compare the cpu-types as well.  A message is displayed when the cpu-types
;;;         do not match 3/25/87 mp
;;;      Put the partition-namestring on the partition header so that List-Contents would display the partition name.
;;;                                                                                                       4/3/87 mp 
;;;
;;;
;;;      ;;;
;;;      Entry points:
;;;          Backup-partition       - Top level hack that uses buffer pool & initiate i/o to stream to tape.
;;;          Restore-partition      -  Ditto, but to stream from tape.
;;;          Verify-partition       - Verify and streams too, (when on an Emulex/Cipher)
;;;          Backup-partition-half-inch-tape  - Backup a partition on a 1/2" tape.
;;;          Restore-partition-half-inch-tape - Restore a partition on a 1/2" tape.
;;;          Verify-partition-half-inch-tape  - Verify a partition on a 1/2" tape.
;;;      Misc routines:
;;;          Stream-Data-Copy       - Main copying help function for backup-partiton and restore-partition.
;;;          Maybe-Display-Hundred  - Bumps count of blocks xferred.  If greater than 100 multiple display number.
;;;          Clean-Up               - Attempt to throw away rqbs that got hung on *ABORTED-RQB-Q* queue.
;;;          Deallocate-Q           - Routine to toss rqbs; sticks 'em on *ABORTED-RQB-Q* if they don't complete.
;;;          Address-Check          
;;;          Buffer-pool-adjust      - Adjust buffer pool size so little partitions can be handled.
;;;          Assure-Partition-Or-Ferror - Checks the tape data set to be sure it is a partition.
;;;          Check-For-Error-Or-Status  - Looks errors/status & screens out harmless ones, signals error, or returns value.
;;;          
;;;

(DEFVAR *STREAMING-PRIORITY* 0)			;When trying to stream raise priority a little to this.

;;; New function  DAB 03-28-89. Handles multiple media backup Partition and  Restore partition. When
;;; EOT is encountered the user is prompted to continue with new tape.
(defun handle-eom-and-maybe-proceed (Message out-stream &optional (proceed? NIL))
  (select (if proceed? (FQUERY '(:CHOICES    ; DAB 04-20-89
				  (((:YES  "Load the next tape.") #\Y #\SPACE #\T #\)
				   ((nil   "Don't load the next tape and quit.") #\N)
				   ((:PROCEED "Proceed with this tape anyway.") #\P)
				   )) ;moved from N to S -dkm 6/87
			       Message)
	      (FQUERY '(:CHOICES    ; DAB 04-20-89
				  (((:YES  "Load the next tape.") #\Y #\SPACE #\T #\)
				   ((nil   "Don't load the next tape and quit.") #\N)
				   )) ;moved from N to S -dkm 6/87
			       Message))
	      
    (:yes 
     (let ((block-mode (get 'mt:*current-unit*  'block-mode))
	   (tape-type (get 'mt:*current-unit* 'tape-type))
	   (density (get 'mt:*current-unit*  'density))
	   (block-size (get 'mt:*current-unit*  'block-size))
	   (unit *current-unit*))
       (format out-stream "~%Rewinding ... ")
       (rewind)
       (format out-stream "Rewind completed.~%Unloading ... ")
       (unload)
       (format out-stream "Unload completed.~%Please load the next tape. Press any key to continue.")
       (send out-stream :tyi)
       (format out-stream "~%Preparing next volume ...")
       (loop 
	 (condition-case (condition) ; catch drive error occurring because tape unit is not ready.
	     (progn 
	      (format out-stream ".")
	      (prepare-tape-drive unit
				  :block-mode block-mode
				  :tape-type tape-type
				  :density density
				  :block-size block-size
				  )
	      (rewind unit)
	      (format out-stream "Preparation completed.~%")
	      (return T))
	   (error 
	    (sleep 2 "Waiting on Tape Unit") ;; DAB 03-21-89 Wait 2 seconds to allow drive to spend up,otherwise we get media change errors.
	    )))
       T))
    (:proceed :proceed)
    (otherwise ()))
  )

(DEFUN BACKUP-PARTITION (FROM-UNIT FROM-PART &OPTIONAL (STARTING-HUNDRED 0) (WHOLE-THING-P NIL) &KEY (BUF-SIZE 6)
			 (N-BUFFERS 5) (STYLE :MIT) (OUT-STREAM *STANDARD-OUTPUT*)
			 &AUX MT-CLOSURE (aborted-p T))
  "Stream copy a partition FROM-PART on FROM-UNIT to the tape.
Begin at offset, starting-hundred.  If whole-thing-p is T don't attempt to use measured size, but do
the whole thing.  Buf-size (in Kbytes) and n-buffers define the buffer pool."
  (UNWIND-PROTECT (LET
		    (FROM-PART-BASE FROM-PART-SIZE TO-PART-BASE TO-PART-SIZE TO-PART PART-COMMENT
		     TO-UNIT partition-namestring cpu-type)
		    (SETQ TO-UNIT *CURRENT-UNIT*)	;always a tape
		    (SETQ FROM-UNIT
			  (SI::DECODE-UNIT-ARGUMENT FROM-UNIT
						    (FORMAT () "reading ~A partition" FROM-PART))
			  MT-CLOSURE
			  (SI::DECODE-UNIT-ARGUMENT "MT:"
						    (FORMAT () "writing ~A partition" TO-PART) ()
						    T))
		    (MULTIPLE-VALUE-SETQ (FROM-PART-BASE FROM-PART-SIZE nil from-part nil partition-namestring)  ; 3/17/87 mp
		      (SYSTEM:FIND-DISK-PARTITION-FOR-READ FROM-PART () FROM-UNIT))
		    
		    (MULTIPLE-VALUE-SETQ (TO-PART-BASE TO-PART-SIZE NIL TO-PART)
		      (SYSTEM:FIND-DISK-PARTITION-FOR-WRITE "" () MT-CLOSURE))	;"" was TO-PART
		    (setq cpu-type (nth-value 1 (si:parse-partition-name partition-namestring))) ;get cpu 3/24/87 mp

		    (do ((Completed? nil)  ; DAB 03-21-89
			 (restart-on-block 0))   ; DAB 03-21-89
			((eq completed? :complete))   ; DAB 03-21-89
		      (IF TO-PART-BASE
			  (PROGN
			    (SETQ PART-COMMENT (SYSTEM:PARTITION-COMMENT partition-namestring from-unit))
			    (FORMAT OUT-STREAM "~&Copying ~S" PART-COMMENT)
			    (UNLESS (ZEROP STARTING-HUNDRED)
			      (FORMAT OUT-STREAM " starting at  ~D. hundred Kbyte offset."
				      STARTING-HUNDRED))
			    (AND (OR (NUMBERP FROM-PART) (STRING-EQUAL FROM-PART "LOD" :end1 3 :end2 3))
				 (NOT WHOLE-THING-P)
				 (LET ((RQB NIL)
				       )
				   (UNWIND-PROTECT (PROGN
						     (SETQ RQB (SYSTEM:GET-DISK-RQB sys:disk-blocks-per-page))
						     (LET ((SIZE
							     (si:lod-partition-info rqb from-unit from-part-base)))
						       (COND
							 ((AND (> SIZE 8) (<= SIZE FROM-PART-SIZE))
							  (SETQ FROM-PART-SIZE SIZE)
							  (FORMAT OUT-STREAM
								  "... using measured size of ~D. blocks."
								  SIZE)))))
				     (SYSTEM:RETURN-DISK-RQB RQB))))
			    (COND
			      ((FUNCALL MT-CLOSURE :HANDLES-LABEL)
			       (FUNCALL MT-CLOSURE :PUT PART-COMMENT :COMMENT)
			       (FUNCALL MT-CLOSURE :PUT FROM-PART-SIZE :SIZE)
			       (funcall mt-closure :put partition-namestring :name)  ; 4/3/87 mp
			       (FUNCALL MT-CLOSURE :PUT STARTING-HUNDRED :STARTING-OFFSET)
			       (FUNCALL MT-CLOSURE :PUT STYLE :STYLE)
			       (funcall mt-closure :put cpu-type :cpu-type)   ;copy cpu-type on the header. 3/24/87 mp
                               (unless (zerop  restart-on-block) (funcall mt-closure :put restart-on-block :MULTI-MEDIA))  ; DAB 04-20-89
			       (FUNCALL MT-CLOSURE :FORCE-LABEL-OUT)))))
		      (MULTIPLE-VALUE-SETQ (N-BUFFERS BUF-SIZE)
			(BUFFER-POOL-ADJUST FROM-PART-SIZE N-BUFFERS BUF-SIZE))
		      (setf (values completed? restart-on-block)
			    (STREAM-DATA-COPY FROM-UNIT TO-UNIT
					      (+ FROM-PART-BASE (* 100 STARTING-HUNDRED))	;start-up-adr
					      (+ FROM-PART-BASE FROM-PART-SIZE)	;from-high
					      N-BUFFERS BUF-SIZE OUT-STREAM STARTING-HUNDRED))
		       ;;; DAB 04-03-89 If we hit the end of tape lets give the user a chance to continue.
		      (if (eq completed? :end-of-tape)  ; DAB 04-03-89
			  (if (handle-eom-and-maybe-proceed 
				"The end of tape was encountered before completing the backup.~%Do you what to continue with a new tape?"
				 out-stream)
			      (setf  STARTING-HUNDRED restart-on-block) ;Yes, the new tape is insatlled, lets continue.
			      (return ())) ;no, then just return.
			  (return ()))
		      (setf aborted-p nil)
		      )	;do
		    ) ;let
    (when (closurep mt-closure)
      (si:set-in-closure mt-closure 'mt:*abort-flag*  aborted-p))	;8.13.87
    (SI::DISPOSE-OF-UNIT MT-CLOSURE)))
 


(DEFUN RESTORE-PARTITION (TO-UNIT TO-PART &OPTIONAL &KEY (BUF-SIZE 6) (N-BUFFERS 5) (OUT-STREAM *STANDARD-OUTPUT*) &AUX
			  MT-CLOSURE proceed-option)
  "Stream copy a partition from tape to TO-PART partition on TO-UNIT.
  Buf-size (in Kbytes) and n-buffers define the buffer pool."
  (let ((aborted-p T) cpu-type
    	TO-PART-BASE)				;use to see if user answered NO on clobber query..
    (UNWIND-PROTECT
	(LET
	  (FROM-PART-BASE FROM-PART-SIZE TO-PART-SIZE FROM-UNIT FROM-PART
	   PART-COMMENT STARTING-HUNDRED TO-END  partition-namestring)
	  (SETQ FROM-UNIT *CURRENT-UNIT*
		FROM-PART "")			;always a tape!  oops... how do we get unit number?
	  (SETQ TO-UNIT
		(SI::DECODE-UNIT-ARGUMENT TO-UNIT
					  (FORMAT () "writing ~A partition" TO-PART) ()
					  T))
	  (MULTIPLE-VALUE-SETQ (TO-PART-BASE TO-PART-SIZE NIL TO-PART nil partition-namestring)    ; 3/23/87 mp
	    (SYSTEM:FIND-DISK-PARTITION-FOR-WRITE TO-PART () TO-UNIT))

	  

	  (do ((Completed? nil)  ; DAB 03-21-89
	       (restart-on-block 0))   ; DAB 03-21-89
	      ((eq completed? :complete))   ; DAB 03-21-89
	    
	    (setq MT-CLOSURE
		  (SI::DECODE-UNIT-ARGUMENT "MT:"
					    (FORMAT () "reading ~A partition" FROM-PART)))
	    (MULTIPLE-VALUE-SETQ (FROM-PART-BASE FROM-PART-SIZE NIL FROM-PART) 
	      (SYSTEM:FIND-DISK-PARTITION-FOR-READ FROM-PART () MT-CLOSURE))
	    
	    (ASSURE-PARTITION-OR-FERROR MT-CLOSURE)
	    
	    (IF TO-PART-BASE
		(block restore-block
		  (SETQ PART-COMMENT (SYSTEM:PARTITION-COMMENT () MT-CLOSURE))
		  (FORMAT OUT-STREAM "~&Copying ~S" PART-COMMENT)
		  (SYSTEM:UPDATE-PARTITION-COMMENT partition-namestring "Incomplete Copy" TO-UNIT) ;it was to-part 3/23/87 mp
		  (SETQ STARTING-HUNDRED
			(OR (SEND MT-CLOSURE :GET :STARTING-OFFSET) 0))
		  (unless (zerop  restart-on-block)
		   (when (SEND MT-CLOSURE :GET :MULTI-MEDIA) 
		     (unless (= starting-hundred  (SEND MT-CLOSURE :GET :MULTI-MEDIA))
		       (if (setf proceed-option
				 (handle-eom-and-maybe-proceed 
				   (format nil 
					   "The tape seems to be out of order.~%The starting offset of this tape is ~a, we were expecting ~a.~%Do you what to continue with another tape?"

					   starting-offset restart-on-block)
				   out-stream T)) 
			  (unless (eql proceed-option :proceed)
			    (return-from restore-block)) ;YES, continue
			  (return (setf completed? :complete))))))
		       
		  (UNLESS (ZEROP STARTING-HUNDRED)
		    (FORMAT OUT-STREAM " starting at  ~D. hundred Kbyte offset."
			    STARTING-HUNDRED))
		  (ADDRESS-CHECK TO-PART-BASE TO-PART-SIZE partition-namestring  ;it was TO-PART 3/23/87 mp
				 (+ TO-PART-BASE (* 100 STARTING-HUNDRED))	;init
				 (- FROM-PART-SIZE (* 100 STARTING-HUNDRED)))	;total needed
		  (SETQ TO-END (+ TO-PART-BASE (MIN FROM-PART-SIZE TO-PART-SIZE)))
		  (SETQ FROM-PART-SIZE (- FROM-PART-SIZE (* STARTING-HUNDRED 100)))
		  (MULTIPLE-VALUE-SETQ (N-BUFFERS BUF-SIZE)
		    (BUFFER-POOL-ADJUST (MIN TO-PART-SIZE FROM-PART-SIZE) N-BUFFERS
					BUF-SIZE))
		  
		  (setf (values completed? restart-on-block)
			(STREAM-DATA-COPY FROM-UNIT TO-UNIT
					    (+ TO-PART-BASE (* 100 STARTING-HUNDRED))	;start-up-adr
					    TO-END	;end size
					    N-BUFFERS BUF-SIZE OUT-STREAM STARTING-HUNDRED))
		  ;;; DAB 04-03-89 If we hit the end of tape lets give the user a chance to continue.
		  (if (eq completed? :end-of-tape)
		      (if (handle-eom-and-maybe-proceed 
			    "The end of tape was encountered before completing the restore.~%Do you what to continue with another tape?"
			    out-stream)
			  () ;YES, continue
			  (return (setf completed? :complete))) ;No, then just return.
		      (WHEN (EQ completed? :COMPLETE)
			(SYSTEM:UPDATE-PARTITION-COMMENT partition-namestring PART-COMMENT TO-UNIT) ;; it was to-part
			(when (setq cpu-type (send mt-closure :get :cpu-type))   ;3/23/87 mp
			  (si:set-partition-cpu-type partition-namestring to-unit cpu-type)))
		      )))
	    	    )  ; DAB 03-21-89 DO
	  (setf aborted-p nil)
	  ) ;let unwind protects
      (when (closurep mt-closure)
	(unless aborted-p
	  (setq aborted-p (null to-part-base)))
	(si:set-in-closure mt-closure 'mt:*abort-flag*  aborted-p))
      (SI::DISPOSE-OF-UNIT MT-CLOSURE))))
 



(DEFUN STREAM-DATA-COPY (FROM-UNIT TO-UNIT START-UP-ADR FROM-END N-BUFFERS B-SIZE &OPTIONAL (OUT-STREAM *TERMINAL-IO*)
			 (N-HUNDRED 0))
  "This assumes that the to-unit address range has been checked !!!!!!"
  (LET ((FROM-REAL-UNIT (SI::GET-REAL-UNIT FROM-UNIT))
	(OLD-PRIORITY (SEND CURRENT-PROCESS :PRIORITY))
	(TO-REAL-UNIT (SI::GET-REAL-UNIT TO-UNIT))
	(N-blocks-WRITTEN (* 100 N-HUNDRED))
	(BUF-SIZE B-SIZE)
	(TO-ADR START-UP-ADR)
	FROM-Q
	TO-Q
	IDLE-Q
	completed?)

    (UNWIND-PROTECT
	(progn
	  (SEND CURRENT-PROCESS :SET-PRIORITY *STREAMING-PRIORITY*)  
	  (LET (RQB)
	    
	    (DOTIMES (I N-BUFFERS)
	      (SETQ RQB (SYSTEM:GET-DISK-RQB BUF-SIZE))
	      (SI::WIRE-NUPI-RQB RQB BUF-SIZE T T)
	      (SI:nupi-read-wired rqb from-real-unit from-unit START-UP-ADR t)
	      (PUSH-END RQB FROM-Q)
	      (SETQ START-UP-ADR (+ START-UP-ADR BUF-SIZE))))
	  (DO* ((FROM-ADR START-UP-ADR))
	       ((>= FROM-ADR FROM-END))		;end test
	    (condition-case (condition)   ; DAB 04-03-89 Trap for end-of-tape
		(DO ((RQB (CAR FROM-Q) (CAR FROM-Q)))
		    ((OR (NULL RQB)		;q empty --> exit
			 (NOT (SI::%IO-DONE RQB))))	;io not done --> exit
		  (IF (CHECK-FOR-ERROR-OR-STATUS RQB)
		      (CERROR T () 'SI::TAPE	;proceedable
			      "~&I/O error: ~A, on unit ~d." (SI::GET-LAST-ERROR FROM-UNIT)
			      FROM-UNIT))
		  (POP FROM-Q)
		  (SI:nupi-write-wired rqb to-real-unit to-unit to-adr t)	;turn it around
		  (SETQ TO-ADR (+ TO-ADR BUF-SIZE))
		  (PUSH-END RQB TO-Q))
	      (error
	       (if (search "END-OF-TAPE" (send condition :report-string))    ; DAB 04-03-89 Trap for end-of-tape
		   (setf completed?  :END-OF-TAPE)
		   (signal-condition condition)))
	      )					; condition-case
	    
	    (condition-case (condition)   ; DAB 04-03-89 Trap for end-of-tape
		(DO ((RQB (CAR TO-Q) (CAR TO-Q)))
		    ((OR (NULL RQB)		;q empty --> exit
			 (NOT (SI::%IO-DONE RQB))))	;io not done --> exit
		  (CHECK-FOR-ERROR-OR-STATUS RQB)	;A write completed
		  (SETQ N-blocks-WRITTEN (+ N-blocks-WRITTEN (Si:rqb-n-Blocks-Wired rqb)))	;count wired blocks
		  (SETQ N-HUNDRED
			(MAYBE-DISPLAY-HUNDRED N-blocks-WRITTEN N-HUNDRED OUT-STREAM))
		  (SETQ RQB (POP TO-Q))
		  (LET ((AMT (MIN BUF-SIZE (- FROM-END FROM-ADR))))	;Amount left to read
		    (IF (ZEROP AMT)		;If tape blocks complete quickly at end,...
			(SYSTEM:RETURN-DISK-RQB RQB)	;We're done... So deallocate
			(PROGN
			  (UNLESS (= AMT BUF-SIZE)	;ELSE keep going
			    (SI::UNWIRE-DISK-RQB RQB)	;Left-over is less than rqb size, so wire partial.
			    (SI::WIRE-NUPI-RQB RQB AMT T T))
			  (SI:nupi-read-wired rqb from-real-unit from-unit FROM-ADR t)
			  (SETQ FROM-ADR (+ FROM-ADR AMT))
			  (PUSH-END RQB FROM-Q)))))
	      (error   ; DAB 04-03-89 Trap for end-of-tape
	       (if (search "END-OF-TAPE" (send condition :report-string))
		   (setf completed?  :END-OF-TAPE)
		   (signal-condition condition))))
	    (when (eq completed? :end-of-tape) (return()))
	    )					;DO*	; FINISH - now write out all remaining read buffers as they finish
	  (unless (eq completed? :end-of-tape)  ; DAB 04-03-89
	    (condition-case (condition)
		(DO ((RQB (CAR FROM-Q) (CAR FROM-Q)))
		    ((NULL RQB))
		  (COND
		    ((SI::%IO-DONE RQB)
		     (IF (CHECK-FOR-ERROR-OR-STATUS RQB)
			 (CERROR T () 'SI::TAPE	;proceedable
				 "~&I/O error: ~A, on unit ~d."
				 (SI::GET-LAST-ERROR FROM-UNIT) FROM-UNIT))
		     (POP FROM-Q)
		     (SI:nupi-write-wired rqb to-real-unit to-unit to-adr t)
		     (SETQ TO-ADR (+ TO-ADR (Si:rqb-n-Blocks rqb))) (PUSH-END RQB TO-Q))
		    (T (PROCESS-ALLOW-SCHEDULE))))
	      (error
	       (if (search "END-OF-TAPE" (send condition :report-string))  ; DAB 04-03-89
		   (setf completed?  :END-OF-TAPE)
		   (signal-condition condition))))
	    (condition-case (condition)    ; DAB 04-03-89 Trap for end-of-tape
		(DO ((RQB (CAR TO-Q) (CAR TO-Q)))
		    ((NULL RQB)
		     :COMPLETE)
		  (COND
		    ((SI::%IO-DONE RQB) (CHECK-FOR-ERROR-OR-STATUS RQB)
					(SETQ N-blocks-WRITTEN (+ N-blocks-WRITTEN (Si:rqb-n-Blocks rqb)))
					(SETQ N-HUNDRED
					      (MAYBE-DISPLAY-HUNDRED N-blocks-WRITTEN N-HUNDRED OUT-STREAM))
					(POP TO-Q) (PUSH-END RQB IDLE-Q))
		    (T (PROCESS-ALLOW-SCHEDULE))))
	      (error
	       (if (search "END-OF-TAPE" (send condition :report-string))   ; DAB 04-03-89 Trap for end-of-tape
		   (setf completed?  :END-OF-TAPE)
		   (signal-condition condition))))
	    )
	  (if (eq completed? :end-of-tape)    ; DAB 04-03-89 Trap for end-of-tape
	      (values :END-OF-TAPE  N-HUNDRED) ;Yes, return :end-of-file
	      :complete)                       ;No, return :complete
	  )					;progn
      (SEND CURRENT-PROCESS :SET-PRIORITY OLD-PRIORITY)
      (DEALLOCATE-Q FROM-Q OUT-STREAM)
      (DEALLOCATE-Q TO-Q OUT-STREAM)
      (DEALLOCATE-Q IDLE-Q OUT-STREAM)
      (SI::DISPOSE-OF-UNIT FROM-UNIT)
      (SI::DISPOSE-OF-UNIT TO-UNIT)))) 



(DEFUN CHECK-FOR-ERROR-OR-STATUS (RQB)
  (LET ((NUPI-ERROR
	  (LDB SYSTEM:%%NUPI-STATUS-HIGH-CNTL-ERROR (AREF RQB SYSTEM:%IO-RQ-STATUS-HIGH)))
	(DEVICE-ERROR (LDB SYSTEM:%%NUPI-STATUS-LOW-DEV-ERROR (AREF RQB SYSTEM:%IO-RQ-STATUS)))
	(LOGICAL-UNIT (SI::GET-LOGICAL-UNIT (SI::RQB-REAL-UNIT RQB))))
    (COND
      ((NULL (ZEROP NUPI-ERROR)) (SI::SET-LAST-ERROR LOGICAL-UNIT NUPI-ERROR)
				 (CERROR () () 'SYSTEM:TAPE-ERROR	;never proceedable
					 "~&Nupi error: ~16R, on unit ~d." (SI::DECODE-NUPI-STATUS RQB) LOGICAL-UNIT))
      ((OR (= DEVICE-ERROR #o110) (= DEVICE-ERROR #o310)) NIL)	;ignore SCSI sense available & correctable data check error
      ((NULL (ZEROP DEVICE-ERROR))
       (SELECT DEVICE-ERROR
	 ((76. 77. 78. 79.) (SI::SET-LAST-ERROR LOGICAL-UNIT :EOF))	;FM, alias EOF seen
	 (75. (SI::SET-LAST-ERROR LOGICAL-UNIT :END-OF-RECORDED-MEDIA))
	 (74. (SI::SET-LAST-ERROR LOGICAL-UNIT :END-OF-TAPE)
	      (CERROR T () 'SYSTEM:TAPE-ERROR "~&Device error: ~16R, on unit ~d." :END-OF-TAPE
		      LOGICAL-UNIT))
	 (OTHERWISE (SI::SET-LAST-ERROR LOGICAL-UNIT DEVICE-ERROR)
		    (CERROR
		      (NULL
			(LDB-TEST SYSTEM:%%NUPI-STATUS-HIGH-ERROR (AREF RQB SYSTEM:%IO-RQ-STATUS-HIGH)))
		      () 'SYSTEM:TAPE-ERROR "~&Device error: ~16R, on unit ~d."
		      (SI::DECODE-NUPI-STATUS RQB) LOGICAL-UNIT)))
       (SI::GET-LAST-ERROR LOGICAL-UNIT))	;return non-nil value
      (T NIL)))) 


(DEFUN VERIFY-PARTITION (DISK-UNIT DISK-PART &OPTIONAL (WHOLE-THING-P NIL) &KEY (BUF-SIZE 6) (N-BUFFERS 5)
			 (OUT-STREAM *STANDARD-OUTPUT*) &AUX)
  "Compare partition DISK-PART on DISK-UNIT to partition on tape.  N-buffer specifies how many buffers
 to allocate for per device; tape and disk."
  (let ((aborted-p T))
    (proclaim '(special sys:%io-rq-leader-n-blocks))
    (LET (DISK-PART-BASE
	  DISK-PART-SIZE
	  TO-PART-BASE
	TO-PART-SIZE
	RQB
	NO-ERRORS
	TO-UNIT
	STARTING-HUNDRED
	TAPE-RQB
	DISK-RQB
	TAPE-BUF
	DISK-BUF
	READ-COUNT
	TOTAL
	DISK-ADR-INIT
	TO-ADR-INIT
	IDLE-Q
	DISK-Q
	TAPE-Q
	MT-CLOSURE
	TAPE-REAL-UNIT
	DISK-REAL-UNIT
	partition-namestring
	cpu-type
	(error-message-queue ()))
    (SETQ TO-UNIT *CURRENT-UNIT*)
    (SETQ DISK-UNIT
	  (SI::DECODE-UNIT-ARGUMENT DISK-UNIT (FORMAT () "reading ~A partition" DISK-PART)))
    (setq DISK-REAL-UNIT (SI::GET-REAL-UNIT DISK-UNIT))
    (UNWIND-PROTECT
	(do ()
	    ((not (eq :eom-error		; DAB 03-21-89 handle end-of-tape error when user whats to continue.
		      (catch 'eom-restart ; DAB 04-03-89
			(setf tape-q nil disk-q nil idle-q nil error-message-queue nil)
			(setq MT-CLOSURE
			      (SI::DECODE-UNIT-ARGUMENT "MT:"
							(FORMAT () "reading partition from tape unit ~d." TO-UNIT)))
			(SETQ TAPE-REAL-UNIT (SI::GET-REAL-UNIT TO-UNIT))
			
			(SETQ NO-ERRORS T)
			(MULTIPLE-VALUE-SETQ (DISK-PART-BASE DISK-PART-SIZE nil nil nil partition-namestring)	;3/23/87 mp
			  (SYSTEM:FIND-DISK-PARTITION-FOR-READ DISK-PART () DISK-UNIT))
			(MULTIPLE-VALUE-SETQ (TO-PART-BASE TO-PART-SIZE) 
			  (SYSTEM:FIND-DISK-PARTITION-FOR-READ "" () MT-CLOSURE))
			(ASSURE-PARTITION-OR-FERROR MT-CLOSURE)
			(SETQ STARTING-HUNDRED
			      (OR (SEND MT-CLOSURE :GET :STARTING-OFFSET) 0))
			(FORMAT OUT-STREAM "~&Comparing ~S and ~S"
				(SYSTEM:PARTITION-COMMENT partition-namestring DISK-UNIT)	;it was disk-part 3/23/87 mp
				(SYSTEM:PARTITION-COMMENT () MT-CLOSURE))
			(UNLESS (ZEROP STARTING-HUNDRED)
			  (FORMAT OUT-STREAM " starting at  ~D. hundred Kbyte offset."
				  STARTING-HUNDRED))
			(SETQ TOTAL (- (+ TO-PART-BASE TO-PART-SIZE) (* 100 STARTING-HUNDRED)))
			(IF (< TOTAL (* N-BUFFERS BUF-SIZE))
			    (SETQ BUF-SIZE (MIN 96. (FLOOR TOTAL N-BUFFERS))))
			(AND (STRING-EQUAL DISK-PART "LOD" :end1 3 :end2 3) (NOT WHOLE-THING-P)
			     (LET ((RQB NIL)
				   )
			       (UNWIND-PROTECT (PROGN
						 (SETQ RQB (SYSTEM:GET-DISK-RQB sys:disk-blocks-per-page))	
						 (LET ((SIZE
							 (si:lod-partition-info rqb disk-unit disk-part-base)))
						   (COND
						     ((AND (> SIZE 8) (<= SIZE DISK-PART-SIZE))
						      (SETQ DISK-PART-SIZE SIZE)
						      (FORMAT OUT-STREAM
							      "... using measured size of ~D. blocks."
							      SIZE)))))
				 (SYSTEM:RETURN-DISK-RQB RQB))))
			(SETQ TO-ADR-INIT (+ TO-PART-BASE (* 100 STARTING-HUNDRED)))
			(SETQ DISK-ADR-INIT (+ DISK-PART-BASE (* 100 STARTING-HUNDRED)))
			(ADDRESS-CHECK DISK-PART-BASE DISK-PART-SIZE
				       partition-namestring DISK-ADR-INIT TOTAL)	;it was disk-part
			(SETQ READ-COUNT 0)
			(MULTIPLE-VALUE-SETQ (N-BUFFERS BUF-SIZE)
			  (BUFFER-POOL-ADJUST DISK-PART-SIZE N-BUFFERS BUF-SIZE))
			(DOTIMES (I N-BUFFERS)	;read tape stuff first
			  (SETQ RQB (SYSTEM:GET-DISK-RQB BUF-SIZE))
			  (SI::WIRE-NUPI-RQB RQB BUF-SIZE T T)
			  (SI:nupi-read-wired rqb tape-real-unit to-unit 0 T)
			  (SETQ TO-ADR-INIT (+ TO-ADR-INIT BUF-SIZE))
			  (PUSH-END (LIST RQB (SYSTEM:RQB-8-BIT-BUFFER RQB)) TAPE-Q)	;now read disk stuff 
			  (SETQ RQB (SYSTEM:GET-DISK-RQB BUF-SIZE))
			  (SI::WIRE-NUPI-RQB RQB BUF-SIZE T T)
			  (SI:nupi-read-wired rqb disk-real-unit disk-unit disk-adr-init T)
			  (SETQ DISK-ADR-INIT (+ DISK-ADR-INIT BUF-SIZE))
			  (PUSH-END (LIST RQB (SYSTEM:RQB-8-BIT-BUFFER RQB)) DISK-Q)	;queues are list of '(rqb buffer)
			  (SETQ READ-COUNT (+ READ-COUNT BUF-SIZE)))
			(DO ((DISK-ADR DISK-ADR-INIT)
			     (TO-ADR TO-ADR-INIT)
			     (DISK-HIGH (+ DISK-PART-BASE DISK-PART-SIZE))
			     (TO-HIGH (+ TO-PART-BASE TO-PART-SIZE))
			     (N-HUNDRED STARTING-HUNDRED)
			     (L-hundred starting-hundred)  ; DAB 04-03-89
			     (N-blocks-COMPARED (* STARTING-HUNDRED 100.))	;9.2.87 MBC
			     (DISK-HIGH-ADR (+ TO-PART-BASE TO-PART-SIZE)))
			    ((>= N-blocks-COMPARED DISK-HIGH-ADR)
			     (dolist (error-msg error-message-queue (setf error-message-queue ()))
			       (format out-stream "~a" error-msg)))
			  ;;;;;; WAIT FOR READS TO COMPLETE AND COMPARE BUFFERS
			  (LET ((RQB-&-BUF (POP DISK-Q)))
			    (SETQ DISK-RQB (CAR RQB-&-BUF))	;get disk rqb
			    (DO ()		;no vars
				((SI::%IO-DONE DISK-RQB)	;end-test
				 (CHECK-FOR-ERROR-OR-STATUS DISK-RQB)
				 (SETQ DISK-BUF (CADR RQB-&-BUF)))
			      ()))


			  ;no body
			  (LET ((RQB-&-BUF (POP TAPE-Q)))
			    (SETQ TAPE-RQB (CAR RQB-&-BUF))	;get tape rqb
			    (condition-case (condition)  ; DAB 04-03-89 Handle end-of-tape errors.
				(DO ()		;no vars
				    ((SI::%IO-DONE TAPE-RQB)	;end-test
				     (IF (CHECK-FOR-ERROR-OR-STATUS TAPE-RQB)
					 (CERROR T () 'SI::TAPE	;proceedable
						 "~&Tape error: ~A, on unit ~d."
						 (SI::GET-LAST-ERROR TO-UNIT) TO-UNIT))
				     (SETQ TAPE-BUF (CADR RQB-&-BUF)))
				  ())
			      (error
			       (if (search "END-OF-TAPE" (send condition :report-string))
				   (if (handle-eom-and-maybe-proceed   ; DAB 04-03-89 Give them a change to continue.
					 "The end of tape was encountered before completing the verify.~%Do you what to continue with another tape?"
					 out-stream)
				       (throw 'eom-restart :eom-error)	;Yes,continue
				       ;;Otherwise show them the verification error in the last block. Then quit.
				       (dolist (error-msg error-message-queue (setf error-message-queue ()))
					 (format out-stream "~a" error-msg))
				       (throw 'eom-restart :complete)) )
			       (signal-condition condition)))
			    )
			  ;;; DAB 04-03-89 We are queue all error message for a given block because when the end of 
                          ;;; Tape occurs the last block is duplicate on the next tape, but because of the read /write
                          ;;; patterns we would get verification errors. We what these suppress if we continue with the next 
                          ;;; tape.
			  (when (and error-message-queue (not (= l-hundred  N-HUNDRED))) ; DAB 04-03-89
			    (dolist (error-msg error-message-queue (setf error-message-queue ()))
			      (format out-stream "~a" error-msg)))
			  (setf  L-HUNDRED  N-HUNDRED)
			  (LET ((AMT (Si:rqb-n-Blocks TAPE-RQB)))
			    (UNLESS (LET ((ALPHABETIC-CASE-AFFECTS-STRING-COMPARISON T))
				      (si:%STRING-EQUAL DISK-BUF 0 TAPE-BUF 0 (* 1024 AMT)))
			      (LET ((DISK-BUF-16-BIT (SYSTEM:RQB-BUFFER DISK-RQB))
				    (TAPE-BUF-16-BIT (SYSTEM:RQB-BUFFER TAPE-RQB)))
				(DO ((C 0 (1+ C))
				     (ERRS 0)
				     (LIM (* 512 AMT)))
				    ((OR (= C LIM) (= ERRS 3)))
				  (COND
				    ((NOT (= (AREF TAPE-BUF-16-BIT C) (AREF DISK-BUF-16-BIT C)))
				     (push-end  ; DAB 04-03-89
				       (FORMAT nil
					     "~%Compare error on block ~D,  halfword ~D, tape data(hex): ~16R   disk data: ~16R "
					     N-blocks-COMPARED (REM C 512)
					     (AREF TAPE-BUF-16-BIT C) (AREF DISK-BUF-16-BIT C))
				       error-message-queue)
				     (SETQ NO-ERRORS ()) (SETQ ERRS (1+ ERRS)))))))
			    (SETQ N-blocks-COMPARED (+ N-blocks-COMPARED AMT)))
			  (SETQ N-HUNDRED
				(MAYBE-DISPLAY-HUNDRED N-blocks-COMPARED N-HUNDRED OUT-STREAM))
			  ;;;;;;; IF THERE IS MORE TO READ RE-USE RQBS, ELSE QUEUE THEM FOR LATER CLEAN-UP
			  (COND
			    ((OR (= DISK-HIGH DISK-ADR) (= TO-HIGH TO-ADR))
			     (PUSH-END DISK-RQB IDLE-Q) (PUSH-END TAPE-RQB IDLE-Q))
			    (T
			     (LET ((AMT (MIN (- DISK-HIGH DISK-ADR) (- TO-HIGH TO-ADR) BUF-SIZE)))
			       (COND
				 ((NOT (= AMT BUF-SIZE)) (SI::UNWIRE-DISK-RQB TAPE-RQB)
							 (SI::UNWIRE-DISK-RQB DISK-RQB)
							 (SI::WIRE-NUPI-RQB TAPE-RQB AMT T T)
							 (SI::WIRE-NUPI-RQB DISK-RQB AMT T T)
							 (SETQ DISK-BUF (SYSTEM:RQB-8-BIT-BUFFER DISK-RQB)
							       TAPE-BUF (SYSTEM:RQB-8-BIT-BUFFER TAPE-RQB))))
			       (SI:nupi-read-wired tape-rqb tape-real-unit to-unit 0 T)
			       (SETQ TO-ADR (+ TO-ADR AMT))
			       (PUSH-END (LIST TAPE-RQB TAPE-BUF) TAPE-Q)	;queues are list of '(rqb buffer)
			       (SI:nupi-read-wired disk-rqb disk-real-unit disk-unit DISK-ADR T)
			       (SETQ DISK-ADR (+ DISK-ADR AMT))
			       (PUSH-END (LIST DISK-RQB DISK-BUF) DISK-Q)))))
			(setf aborted-p nil)
			(when (setq cpu-type (send mt-closure :get :cpu-type))	;Compare destination and source cpu, 3/19/87 mp
			  (if (not (eq (nth-value 1 (si:parse-partition-name partition-namestring))
				       cpu-type))	     
			      (format out-stream "~& [Cpu-Types did not match.]")))
			)			;catch
		      ) )			;do exit forms
	     NO-ERRORS)				;exit return
	  ())					;do
      ;;Unwind-protect forms
      (DEALLOCATE-Q DISK-Q OUT-STREAM)
      (DEALLOCATE-Q TAPE-Q OUT-STREAM)
      (DEALLOCATE-Q IDLE-Q OUT-STREAM)
      (SI::DISPOSE-OF-UNIT DISK-UNIT)
      (when (closurep mt-closure)
	(si:set-in-closure mt-closure 'mt:*abort-flag* aborted-p))	;set *abort-flag* on abort 3/23/87 mp
      (SI::DISPOSE-OF-UNIT MT-CLOSURE)		;NO-ERRORS returns a meaningful value
      )						;UNwind
    )))


(defun backup-partition-half-inch-tape (disk-unit disk-part) 
  "Backup a partition on a REEL tape.  Partial backup is not supported. "
  (unwind-protect
      (si:copy-disk-partition disk-unit disk-part "mt:" disk-part 32. nil)
    (mt:write-eof)))

(defun restore-partition-half-inch-tape (disk-unit disk-part)
  "Restore a partition from a tape to a partition on a disk.  Partial restoring is not supported."
  (si:copy-disk-partition "mt:" disk-part disk-unit disk-part 32. nil)
  )

(defun verify-partition-half-inch-tape (disk-unit disk-part)
  "Verify a partition from a tape against  a partition on a disk."
  (si:compare-disk-partition "mt:" disk-part disk-unit disk-part 32. nil)
  )



(DEFUN DEALLOCATE-Q (Q IGNORE)
  ;;; handle either kind of que entry: 1. its an rqb   2. its a list whose car is an rqb
  (LOOP
    (LET ((ENTRY (POP Q))
	  RQB)
      (IF (NULL ENTRY)
	  (RETURN)
	  (PROGN
	    (SETQ RQB (COND
			((CONSP ENTRY) (CAR ENTRY))
			(T ENTRY)))
	    (SI::UNWIRE-DISK-RQB RQB)
	    (SYSTEM:RETURN-DISK-RQB RQB)))))) 



(DEFUN MAYBE-DISPLAY-HUNDRED (BLOCKS COUNT-HUNDREDS OUT-STREAM)
  "Display the hundred counter if blocks has passed the next COUNT-HUNDREDS.  
  Alway return the count."
  (LET ((CURRENT (FLOOR BLOCKS 100)))
    (COND
      ((NOT (= CURRENT COUNT-HUNDREDS))		;adjust for halfword vs byte
       (SETQ COUNT-HUNDREDS CURRENT) (FORMAT OUT-STREAM "~D " COUNT-HUNDREDS) COUNT-HUNDREDS)
      (T COUNT-HUNDREDS)))) 



(DEFUN ADDRESS-CHECK (DISK-START DISK-PART-SIZE PART INIT TAPE-PART-SIZE)
  "Make sure INIT address and INIT + TAPE-PART-SIZE are within partition, PART, starting at START with length = SIZE."
  (IF (OR (< INIT DISK-START) (> INIT (+ DISK-START DISK-PART-SIZE)))
      (FERROR () "Disk address ~d is not within partition ~A. " INIT PART))
  (LET ((FINAL-ADR (+ TAPE-PART-SIZE INIT)))
    (IF (OR (< FINAL-ADR DISK-START) (> FINAL-ADR (+ DISK-START DISK-PART-SIZE)))
	(FERROR () "Partition ~a on disk is ~d blocks shorter than partition on tape." PART
		(- TAPE-PART-SIZE DISK-PART-SIZE))))) 



(DEFUN ASSURE-PARTITION-OR-FERROR (MAGTAPE-STREAM)
  (UNLESS (SEND MAGTAPE-STREAM :GET :PARTITION)
    (LET ((NAME (SEND MAGTAPE-STREAM :GET :NAME)))
      (IF (OR (NULL NAME) (EQUAL NAME ""))
	  (FERROR () "End of file or end of recorded media encountered.")
	  (FERROR () "Data set on tape, ~a, is not a partition." NAME))))) 



(DEFUN BUFFER-POOL-ADJUST (PART-SIZE N-BUFFERS BUF-SIZE)
  "Returns new n-buffers and buf-size if needed, to make sure N-BUFFERS * BUF-SIZE is less than partition size."
  (declare (special  MAX-blocks-PER-RQB))
  (IF (< PART-SIZE MAX-blocks-PER-RQB)
      (VALUES 1 PART-SIZE)
      (IF (>= PART-SIZE (* N-BUFFERS BUF-SIZE))
	  (VALUES N-BUFFERS BUF-SIZE)
	  (FLOOR PART-SIZE N-BUFFERS)))) 


;;;
;;;  Verify Bootable Tape Support
;;;



(DEFUN VERIFY-BOOTABLE-TAPE (OUT-STREAM)
  "This will attempt to read a boot tape and selectively verify its contents."
  (declare (special PARTITION-UNIT PARTITION-BAND ))
  (LET ((RQB (SYSTEM:GET-DISK-RQB 5)))  ;Get a big RQB and read label from tape
    (UNWIND-PROTECT (PROGN
		      (REWIND)
		      (SYSTEM:DISK-READ RQB *CURRENT-UNIT* 0 1) ;need to detect bad (LOGICAL) tape!!!
		      (UNLESS (STRING-EQUAL (NSUBSTRING (SYSTEM:RQB-8-BIT-BUFFER RQB) 0 4) "LABL")  ;12-03-87 DAB
			(CERROR () () 'SYSTEM:TAPE-ERROR
				"Verify-Bootable-Tape attempted to verify a tape that is not a bootable tape."))
		      (DOTIMES (I (1- (SYSTEM:GET-DISK-FIXNUM RQB SYSTEM:%DL-PARTITION-TABLE-START)))
			(SYSTEM:DISK-READ RQB *CURRENT-UNIT* 0 1 T 1 ()))
		      (SYSTEM:DISK-READ RQB *CURRENT-UNIT* 0 4 T 1 ())
		      (LOOP
			(LET ((PART-LIST (SI::PARTITION-LIST RQB () T))
			      (SELECTED-BAND)
			      (source-cpu-type))        ; 3/19/87 mp
			  
			  (IF (SETQ SELECTED-BAND
				    (SELECT-RESTORE-BAND (REVERSE part-list) "Select a band to verify:"))
			      (progn                                        
				(setq source-cpu-type   ;3/19/87 mp
				      (si:keyword-user-type (ldb si:%%cpu-type-code (sixth selected-band))))
				(IF (choose-destination-band out-stream)   
				    (unless (verify-BOOT-BAND (LIST *CURRENT-UNIT* SELECTED-BAND)
							      PARTITION-UNIT PARTITION-BAND
							      OUT-STREAM :source-cpu-type source-cpu-type)      ;3/19/87 mp
				      (RETURN ()))
				    (RETURN ())) ) 
			      (RETURN ()))
			  (UNLESS (TV::MOUSE-CONFIRM
				    (FORMAT () "Partition ~s verified with unit ~d partition ~s."
					    (FIRST SELECTED-BAND) PARTITION-UNIT PARTITION-BAND)
				    "Click on this window to verify another partition")
			    (RETURN ())))
			(REWIND)))
      (SYSTEM:RETURN-DISK-RQB RQB))))


(DEFUN SELECT-VERIFY-BAND (BAND-LIST DOC-STRING)
  (LET ((PART-LIST (w:MENU-CHOOSE BAND-LIST :label doc-string)))
    (VALUES PART-LIST)))

(DEFUN VERIFY-BOOT-BAND (FROM-LIST TO-UNIT TO-PART &OPTIONAL (OUT-STREAM *TERMINAL-IO*) &key source-cpu-type)
  
  "If sizes are ok, then verify tape partition described by FROM-LIST to TO-PART on TO-UNIT."
  (LET* (			        ;(FROM-UNIT (CAR FROM-LIST))
	 (FROM-PARMS (CADR FROM-LIST))  ;parms= (name start length comment desc-offset)
	 (FROM-COMMENT (STRING-RIGHT-TRIM (CHARACTER " ") (FOURTH FROM-PARMS)))
	 (RETURN-VALUE NIL)
	 LENGTH
	 destination-cpu-type)  	        ; 3/19/87 mp

    
    (MULTIPLE-VALUE-BIND (TO-START TO-LENGTH nil nil attributes partition-namestring)
	(SYSTEM:FIND-DISK-PARTITION TO-PART () TO-UNIT nil nil t)       ; set confirm-read to t 3/19/87 mp
      
      (when attributes (setq destination-cpu-type (si:keyword-user-type (ldb si:%%cpu-type-code attributes))))
      (IF TO-START
	  (PROGN
	    ()
	    (REWIND)
	    (FORMAT OUT-STREAM "~&Spacing to partition... ")
	    (SPACE-FORWARD *CURRENT-UNIT* (SECOND FROM-PARMS))
	    (IF (> (THIRD FROM-PARMS) TO-LENGTH)
		(FERROR () "Destination partition size is too small for source"))
	    (FORMAT OUT-STREAM "~&Verifying ~S on tape with partition ~a on unit ~a... "
		    FROM-COMMENT partition-namestring TO-UNIT)  ;it used to be to-part 3/23/87 mp
	    (SETQ LENGTH (MIN TO-LENGTH (THIRD FROM-PARMS)))
				        ;(MULTIPLE-VALUE-BIND (N-BUFFERS BUF-SIZE)
	    (BUFFER-POOL-ADJUST LENGTH 5 6)
	    (IF (SETQ RETURN-VALUE
		      (VERIFY-BAND-1 TO-UNIT partition-namestring () :OUT-STREAM OUT-STREAM :STYLE :BOOT)) ;it was to-part
		(progn
		  
		  (FORMAT OUT-STREAM "  Verify Successful")
		  (if (not (eq destination-cpu-type source-cpu-type))   ;Compare destination and source cpu, 3/19/87 mp
		      (format out-stream "~& [Cpu-Types did not match.]")))
		(FORMAT OUT-STREAM "  Verify Failed"))
	    
	    (SI::DISPOSE-OF-UNIT TO-UNIT))
	  (FORMAT OUT-STREAM "Partition ~a on unit ~a not found"
		  (or partition-namestring to-part)  ;06-01-88 DAB
		  TO-UNIT))) ;it was to-part mp
    (FORMAT OUT-STREAM "~& Rewinding...~&")
    (REWIND)
    (IF RETURN-VALUE
	T
	())))


;;;
;;; Later to make verify-partition use this code also, consider:
;;;   V-P reads tape label, handles STARTING-HUNDRED.
;;;

(DEFUN VERIFY-BAND-1 (DISK-UNIT DISK-PART &OPTIONAL (WHOLE-THING-P NIL)
		      &KEY (BUF-SIZE 6) (N-BUFFERS 2)
		      (OUT-STREAM *STANDARD-OUTPUT*) (STYLE :MIT))
  "Compare partition DISK-PART on DISK-UNIT to partition on tape.  N-buffer specifies how many buffers
 to allocate for per device; tape and disk."
  (declare (ignore style))
  (LET (DISK-PART-BASE
	DISK-PART-SIZE
	RQB
	(NO-ERRORS T)
	TO-UNIT
	DISK-ADR
	TAPE-RQB
	DISK-RQB
	TAPE-BUF
	DISK-BUF
	IDLE-Q
	DISK-Q
	TAPE-Q
	TOTAL
	TAPE-REAL-UNIT
	DISK-REAL-UNIT)
    (SETQ TO-UNIT *CURRENT-UNIT*)
    (SETQ TAPE-REAL-UNIT (SI::GET-REAL-UNIT TO-UNIT)
	  DISK-REAL-UNIT (SI::GET-REAL-UNIT DISK-UNIT))
    (UNWIND-PROTECT (PROGN			;	  (if (eq style :MIT)
						;	      (MULTIPLE-VALUE (TO-PART-BASE TO-PART-SIZE)
						;	    (si:FIND-DISK-PARTITION-FOR-READ "" NIL MT-CLOSURE)))
		      (MULTIPLE-VALUE-SETQ (DISK-PART-BASE DISK-PART-SIZE)
			(SYSTEM:FIND-DISK-PARTITION-FOR-READ DISK-PART () DISK-UNIT))
		      (SETQ TOTAL DISK-PART-SIZE)
		      (IF (< TOTAL (* N-BUFFERS BUF-SIZE))
			  (SETQ BUF-SIZE (MIN 96. (FLOOR TOTAL N-BUFFERS))))
		      (AND (STRING-EQUAL DISK-PART "LOD" :end1 3 :end2 3) (NOT WHOLE-THING-P)
			   (LET ((RQB NIL)
				 )
			     (UNWIND-PROTECT (PROGN
					       (SETQ RQB (SYSTEM:GET-DISK-RQB sys:disk-blocks-per-page))
					       (LET ((SIZE
						       (si:lod-partition-info rqb disk-unit disk-part-base)))
						 (COND
						   ((AND (> SIZE 8) (<= SIZE DISK-PART-SIZE))
						    (SETQ DISK-PART-SIZE SIZE)
						    (FORMAT OUT-STREAM
							    "... using measured size of ~D. blocks."
							    SIZE)))))
			       (SYSTEM:RETURN-DISK-RQB RQB))))
		      (MULTIPLE-VALUE-SETQ (N-BUFFERS BUF-SIZE)
			(BUFFER-POOL-ADJUST DISK-PART-SIZE N-BUFFERS BUF-SIZE))
		      (SETQ DISK-ADR DISK-PART-BASE)
		      (DOTIMES (I N-BUFFERS)	;read tape stuff first
			(SETQ RQB (SYSTEM:GET-DISK-RQB BUF-SIZE))
			(SI::WIRE-NUPI-RQB RQB BUF-SIZE T T)
			(SI:nupi-read-wired rqb tape-real-unit to-unit 0 T)	;no 'adr here
			(PUSH-END (LIST RQB (SYSTEM:RQB-8-BIT-BUFFER RQB)) TAPE-Q)	;now read disk stuff 
			(SETQ RQB (SYSTEM:GET-DISK-RQB BUF-SIZE))
			(SI::WIRE-NUPI-RQB RQB BUF-SIZE T T)
			(SI:nupi-read-wired rqb disk-real-unit disk-unit disk-adr T)
			(SETQ DISK-ADR (+ DISK-ADR BUF-SIZE))
			(PUSH-END (LIST RQB (SYSTEM:RQB-8-BIT-BUFFER RQB)) DISK-Q)	;queues are list of '(rqb buffer)
			)
		      (LET ((DISK-HIGH (+ DISK-PART-SIZE DISK-PART-BASE))
			    (N-HUNDRED 0)
			    (N-blocks-COMPARED 0))
			(MULTIPLE-VALUE-BIND (TIMES REMAINdER)
			    (FLOOR DISK-PART-SIZE BUF-SIZE)
			  (SETQ TIMES (+ TIMES (IF (> REMAINDER 0)
						   1
						   0)))
			  (DOTIMES (I TIMES)
			    ;;;;;; WAIT FOR READS TO COMPLETE AND COMPARE BUFFERS
			    (LET ((RQB-&-BUF (POP DISK-Q)))
			      (SETQ DISK-RQB (CAR RQB-&-BUF))	;get disk rqb
			      (DO ()		;no vars
				  ((SI::%IO-DONE DISK-RQB)	;end-test
				   (CHECK-FOR-ERROR-OR-STATUS DISK-RQB)
				   (SETQ DISK-BUF (CADR RQB-&-BUF)))
				()))		;no body
			    (LET ((RQB-&-BUF (POP TAPE-Q)))
			      (SETQ TAPE-RQB (CAR RQB-&-BUF))	;get tape rqb
			      (DO ()		;no vars
				  ((SI::%IO-DONE TAPE-RQB)	;end-test
				   (IF (CHECK-FOR-ERROR-OR-STATUS TAPE-RQB)
				       (CERROR T () 'SI::TAPE	;proceedable
					       "~&Tape error: ~A, on unit ~d."
					       (SI::GET-LAST-ERROR TO-UNIT) TO-UNIT))
				   (SETQ TAPE-BUF (CADR RQB-&-BUF)))
				()))
			    (LET ((AMT (Si:rqb-n-Blocks TAPE-RQB)))
			      (UNLESS (LET ((ALPHABETIC-CASE-AFFECTS-STRING-COMPARISON T))
					(si:%STRING-EQUAL DISK-BUF 0 TAPE-BUF 0 (* 1024 AMT)))
				(LET ((DISK-BUF-16-BIT (SYSTEM:RQB-BUFFER DISK-RQB))
				      (TAPE-BUF-16-BIT (SYSTEM:RQB-BUFFER TAPE-RQB)))
				  (DO ((C 0 (1+ C))
				       (ERRS 0)
				       (LIM (1- (* 512 AMT))))
				      ((OR (= C LIM) (= ERRS 3)))
				    (COND
				      ((NOT (= (AREF TAPE-BUF-16-BIT C) (AREF DISK-BUF-16-BIT C)))
				       (FORMAT OUT-STREAM
					       "~%Compare error on block ~D,  halfword ~D, tape data(hex): ~16R   disk data: ~16R "
					       N-blocks-COMPARED (REM C 512)
					       (AREF TAPE-BUF-16-BIT C) (AREF DISK-BUF-16-BIT C))
				       (SETQ NO-ERRORS ()) (SETQ ERRS (1+ ERRS)))))))
			      (SETQ N-blocks-COMPARED (+ N-blocks-COMPARED AMT)))
			    (SETQ N-HUNDRED
				  (MAYBE-DISPLAY-HUNDRED N-blocks-COMPARED N-HUNDRED OUT-STREAM))
			    ;;;;;;; IF THERE IS MORE TO READ RE-USE RQBS, ELSE QUEUE THEM FOR LATER CLEAN-UP
			    (COND
			      ((>= I (- TIMES N-BUFFERS)) (PUSH-END DISK-RQB IDLE-Q)
							  (PUSH-END TAPE-RQB IDLE-Q))
			      (T
			       (LET ((AMT (MIN (- DISK-HIGH DISK-ADR) BUF-SIZE)))
				 (COND
				   ((NOT (= AMT BUF-SIZE)) (SI::UNWIRE-DISK-RQB TAPE-RQB)
							   (SI::UNWIRE-DISK-RQB DISK-RQB)
							   (SI::WIRE-NUPI-RQB TAPE-RQB AMT T T)
							   (SI::WIRE-NUPI-RQB DISK-RQB AMT T T)
							   (SETQ DISK-BUF (SYSTEM:RQB-8-BIT-BUFFER DISK-RQB)
								 TAPE-BUF (SYSTEM:RQB-8-BIT-BUFFER TAPE-RQB))))
				 (SI:nupi-read-wired tape-rqb tape-real-unit to-unit 0 T)
				 (PUSH-END (LIST TAPE-RQB TAPE-BUF) TAPE-Q)	;queues are list of '(rqb buffer)
				 (SI:nupi-read-wired disk-rqb disk-real-unit disk-unit DISK-ADR T)
				 (SETQ DISK-ADR (+ DISK-ADR AMT))
				 (PUSH-END (LIST DISK-RQB DISK-BUF) DISK-Q))))))))
      ;;Unwind-protect forms
      (DEALLOCATE-Q DISK-Q OUT-STREAM)
      (DEALLOCATE-Q TAPE-Q OUT-STREAM)
      (DEALLOCATE-Q IDLE-Q OUT-STREAM)
      (SI::DISPOSE-OF-UNIT DISK-UNIT))
    NO-ERRORS)) 





