;;; -*- Mode:Common-Lisp; Package:User; Fonts:(Medfnt Hl12b Hl12bi); Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.

1;;This file defines the Sugg package and the Suggestions defsystem.

;;When the release 3 system build is actually done, the first two files, WITH-SUGGESTIONS-MENUS.LISP
;;and FUNCTION-DECLARATIONS-FUNCTIONS.LISP, should be loaded early in the build.  The next file,
;;ENCAPSULATED-FUNCTION-DEFINITIONS.LISP, is made up of system function definitions with the macro
;;WITH-SUGGESTIONS-MENUS-FOR added in.  this file should not be loaded at all.  Rather, the functions
;;in the file should be located in the various files they are defined in and the macro
;;WITH-SUGGESTIONS-MENUS-FOR should be added into them in their own source files.  The rest of the
;;files should be loaded AFTER all of the devtools have been loaded.*

(when (not (find-package 'SUGG)) (make-package "SUGG"))

(defsystem suggestions
  (:name "Suggestions")
  (:pathname-default "sys:suggestions;")
  (:patchable "sys:patch.suggestions;" patch)
  (:warnings-pathname-default "sys:cwarns;suggestions.lisp")
  (:DOCUMENTATION "Turn suggestions menus on or off")
  (:DEFAULT-MENU-COLUMN :USER-AIDS)
  (:DEFAULT-SYSTEM-KEY #\S)
  (:INSTANCE-TYPE :EVAL)
  (:INSTANCE-FINDER (SUGG:SYSTEM-MENU-TOGGLE-SUGGESTIONS-MENUS))
  (:INSTANCE-CREATOR NIL)
  (:module vars "global-variables")
  (:module defs "basic-suggestions-functions")
  (:module build-menu "suggestions-build-menu")
  (:module wrappers "function-wrappers")
  (:module advise "suggestions-advise")
  (:module resizing "window-resizing")
  (:module update "update-menus")
  (:module simulate-typing "simulate-typing")
  (:module pop-up-keys "pop-up-keystrokes-process")
  (:module process "suggestions-process")
  (:module off-frame "suggestions-off-frame")
  (:module frame "suggestions-frame")
  (:module documentation "display-documentation")
  (:module zmacs-support "zmacs-support")
  (:module menu-search "menu-search")
  (:module reorder "reorder-menu")
  (:module menu-tools "menu-tools")
  (:module set-up "set-up-suggestions")
  (:module break-menus "break-suggestions-menus")
  (:module converse-menus "converse-suggestions-menus")
  (:module debugger-menus "debugger-suggestions-menus")
  (:module exception-menus "exception-handling-menus")
  (:module input-editor-menus "input-editor-suggestions-menus")
  (:module lisp-menus "lisp-expressions-menus")
  (:module recursive-ie-menus "recursive-input-editor-menus")
  (:module stepper-menus "single-stepper-suggestions-menus")
  (:module zmacs-menus "zmacs-suggestions-menus")
  (:module inspector-menus "inspector-menus")
  (:module flavor-inspector "flavor-inspector-menus")
  (:compile-load vars)
  (:compile-load defs (:fasload vars) (:fasload vars))
  (:compile-load build-menu (:fasload defs) (:fasload defs))
  (:compile-load wrappers (:fasload build-menu) (:fasload build-menu))
  (:compile-load advise (:fasload wrappers) (:fasload wrappers))
  (:compile-load resizing (:fasload advise) (:fasload advise))
  (:compile-load update (:fasload resizing) (:fasload resizing))
  (:compile-load simulate-typing (:fasload update) (:fasload update))
  (:compile-load pop-up-keys (:fasload simulate-typing) (:fasload simulate-typing))
  (:compile-load process (:fasload pop-up-keys) (:fasload pop-up-keys))
  (:compile-load off-frame (:fasload process) (:fasload process))
  (:compile-load frame (:fasload off-frame) (:fasload off-frame))
  (:compile-load documentation (:fasload frame) (:fasload frame))
  (:compile-load zmacs-support (:fasload documentation) (:fasload documentation))
  (:compile-load menu-search (:fasload zmacs-support) (:fasload zmacs-support))
  (:compile-load reorder (:fasload menu-search) (:fasload menu-search))
  (:compile-load menu-tools (:fasload reorder) (:fasload reorder))
  (:compile-load set-up (:fasload menu-tools) (:fasload menu-tools))
  (:compile-load break-menus (:fasload set-up) (:fasload set-up))
  (:compile-load converse-menus (:fasload break-menus) (:fasload break-menus))
  (:compile-load debugger-menus (:fasload converse-menus) (:fasload converse-menus))
  (:compile-load exception-menus (:fasload debugger-menus) (:fasload debugger-menus))
  (:compile-load input-editor-menus
		 (:fasload exception-menus) (:fasload exception-menus))
  (:compile-load lisp-menus (:fasload input-editor-menus) (:fasload input-editor-menus))
  (:compile-load recursive-ie-menus (:fasload lisp-menus) (:fasload lisp-menus))
  (:compile-load stepper-menus (:fasload recursive-ie-menus)
		 (:fasload recursive-ie-menus))
  (:compile-load zmacs-menus (:fasload stepper-menus) (:fasload stepper-menus))
  (:compile-load inspector-menus (:fasload zmacs-menus) (:fasload zmacs-menus))
  (:compile-load flavor-inspector (:fasload inspector-menus) (:fasload inspector-menus)))
