;;; -*- Mode:Common-Lisp; Package:Sugg; Fonts:(Medfnt Hl12b Hl12bi); Base:10 -*- 

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (c) 1984-1989 Texas Instruments Incorporated. All rights reserved.

(defun height-needed-for-x-who-line-doc-lines (number-of-lines)
2  "Calculate the number of pixels needed to allow NUMBER-OF-LINES to be displayed in the
Mouse documentation window."*
  (* number-of-lines (w:sheet-line-height w:who-line-documentation-window)))

(defun calculate-label-pane-who-line-documentation-string (current-item &aux (doc ""))
  2"Returns the documentation string for CURRENT-ITEM, the item currently pointed to in the Suggestions frame label pane."*
  (declare (optimize (speed 3) (safety 0))
	   (string doc))
  (when (variable-boundp current-item)
    (cond ((eq current-item old-current-item) old-who-line-documentation-string)
	  (t (multiple-value-bind (ignore found? hash-table-values)
		 (gethash current-item *keystroke-and-doc-hash-table*)
	       (cond (found? (setq old-who-line-documentation-string (third hash-table-values)))
		     (t (setq doc (calculate-documentation current-item))
			(cond ((plusp (length doc))
			       (let ((cr? (position #\Newline doc :test #'char-equal))
				     (height-needed-for-2-line-format
				       (height-needed-for-x-who-line-doc-lines 2.)))
				 (cond ((>= (w:sheet-height w:who-line-documentation-window)
					    height-needed-for-2-line-format)
					(cond (cr? (rplaca (nleft 1 *label-pane-mouse-buttons*)
							   (if cr? (subseq doc 0 cr?) doc))
						   (rplaca (nleft 5 *label-pane-mouse-buttons*)
							   *right-click-more-doc-message*))
					      (t (rplaca (nleft 1 *label-pane-mouse-buttons*) doc)
						 (rplaca (nleft 5 *label-pane-mouse-buttons*)
							 *simple-right-click-doc-message*)))
					(setq old-who-line-documentation-string
					      (copy-list *label-pane-mouse-buttons*)))
				       (t (setq old-who-line-documentation-string
						(if cr? (subseq doc 0 cr?) doc))))))
			      (t (rplaca (nleft 1 *label-pane-mouse-buttons*) "")
				 (rplaca (nleft 5 *label-pane-mouse-buttons*)
					 *simple-right-click-doc-message*)
				 (setq old-who-line-documentation-string
				       (copy-list *label-pane-mouse-buttons*))))
			(puthash current-item nil *keystroke-and-doc-hash-table* old-who-line-documentation-string)))
	       (setq old-current-item current-item)
	       old-who-line-documentation-string)))))

(defmethod (my-menu-command-menu-mixin :who-line-documentation-string) (&aux (doc ""))
  2"Returns the documentation string for the item currently pointed to in the Suggestions frame."*
  (declare (optimize (speed 3) (safety 0))
	   (string doc))
  (when (variable-boundp w:current-item)
    (if (typep (w:window-owning-mouse) 'label-pane-manip-strip)
	(calculate-label-pane-who-line-documentation-string w:current-item)
	(cond ((eq w:current-item old-current-item) old-who-line-documentation-string)
	      (t (multiple-value-bind (keystroke found? hash-table-values)
		     (gethash w:current-item *keystroke-and-doc-hash-table*)
		   (cond (found? (setq old-who-line-documentation-string (third hash-table-values)))
			 (t (setq keystroke (convert-keystroke-to-string
					      (calculate-keystroke? w:current-item)))
			    (setq doc (calculate-documentation w:current-item))
			    (cond ((or (plusp (length doc)) keystroke)
				   (let ((cr? (position #\Newline doc :test #'char-equal))
					 (height-needed-for-2-line-format
					   (height-needed-for-x-who-line-doc-lines 2.)))
				     (cond ((>= (w:sheet-height w:who-line-documentation-window)
						height-needed-for-2-line-format)
					    (rplaca (nleft 1 *suggestions-mouse-buttons*)
						    (if cr? (subseq doc 0 cr?) doc))
					    (rplaca
					      (nleft 5 *suggestions-mouse-buttons*)
					      (let ((right-click-message
						      (if cr? *right-click-more-doc-message* 
							  *simple-right-click-doc-message*)))
						(if (and keystroke *keystroke-in-mouse-doc-line?*)
						    (string-append right-click-message
								   ", Keystroke: " keystroke)
						    right-click-message)))
					    (setq old-who-line-documentation-string
						  (copy-list *suggestions-mouse-buttons*)))
					   (t (setq old-who-line-documentation-string
						    (if cr? (subseq doc 0 cr?) doc))))))
				  (t (rplaca (nleft 1 *suggestions-mouse-buttons*) "")
				     (rplaca (nleft 5 *suggestions-mouse-buttons*)
					     *simple-right-click-doc-message*)
				     (setq old-who-line-documentation-string
					   (copy-list *suggestions-mouse-buttons*))))
			    (puthash w:current-item keystroke
				     *keystroke-and-doc-hash-table* old-who-line-documentation-string))))
		 (setq old-current-item w:current-item)
		 old-who-line-documentation-string)))))

(defun print-documentation (symbol)
  2"Display the documentation for SYMBOL in a typeout window."*
  (if w:selected-window
      (print (documentation symbol) (send w:selected-window :get-pane :typeout-pane))))

(defwindow-resource suggestions-pop-up-notification-window nil
  :make-window (w:pop-up-text-window)
  :reusable-when :deexposed
  :initial-copies 1)
