;;; -*- Mode:Common-Lisp; Package:Sugg; Fonts:(Medfnt Hl12b Hl12bi); Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (c) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; 04-24-89 DAB Added all function, variables, flavors and instances that are documented to export list.


1;;This file contains most of the global special variables used by Suggestions to keep track of various
;;aspects of the state of Suggestions, including the size of the Suggestions frame and application
;;windows, the fonts used to display menu items, pointers to the Suggestions frame instance and its
;;panes, and several miscellaneous other pieces of information.

;;The *default* variables should be treated as CONSTANTS that dictate the dimensions
;;of the Suggestions frame when it is initially built.  The other variables get their
;;values from these constants.*

(import '(sys:*currently-selected-application* sys:doing-apropos? sys:*mouse-in-suggestions-pane?*
	  sys:name-of-suggestions-menu sys:*previously-selected-application* sys:suggestions-frame
	  sys:suggestions-off-frame sys:*suggestions-menus-on?* sys:*suggestions-zmacs-modes*
	  sys:who-line-package sys:*windows-causing-suggestions-to-push-notification-menu*
	  sys:*windows-shrunk-for-suggestions* sys:*suggestions-loaded-p*
	  sys:*suggestions-help-functions-alist* sys:declare-suggestions-for
	  sys:with-suggestions-menus-for sys:make-default-around-form sys:undeclare-suggestions-for
	  sys:CLIMB-TO-RECOGNIZED-WINDOW sys:DEFAULT-POP-MENUS sys:DEFAULT-PUSH-MENUS
	  zwei:FIND-CALL-COMMAND-ON-KEYS zwei:FULL-COMTAB-NAME sys:HEIGHT-OF-SUGGESTIONS-FRAME
	  sys:INSTALL-MIDDLE-SUGGESTIONS-MENU sys:MAYBE-RESIZE-SELECTED-APPLICATION
	  sys:NAME-OF-SELECTED-MENU sys:NON-EVALING-BUT-INPUT-EDITING-WINDOW
	  sys:POP-MIDDLE-SUGGESTIONS-MENU sys:REPLACE-SUGGESTIONS-MENUS sys:SIGNAL-SUGGESTIONS-PROCESS
	  sys:SPECIAL-POP-MENUS sys:STACK-POSITION sys:SUGGESTIONS-DEFINED-FOR-THIS-WINDOW?
	  sys:SUGGESTIONS-ON-FOR-THIS-WINDOW? zwei:SUGGESTIONS-PROCESS-COMMAND-CHAR 
	  sys:SUGGESTIONS-SYMBOL-STRING sys:SUGGESTIONS-VISIBLE-FOR-APPLICATION? zwei:THE-COMTAB
	  sys:UPDATE-ALL-CONFIGURATIONS-OF-WINDOWS-SHRUNK-FOR-SUGGESTIONS zwei:WRAPPED-AROUND?
	  sys:GET-CURRENT-CONFIGURATION-FOR-WINDOW sys:SCROLL-MAIN-SUGGESTIONS-MENU
	  sys:DEFAULT-NOTIFICATION-MENU zwei:LISP-EXPRESSIONS tv:DEFAULT-MORE-MENU
	  sys:*keystroke-and-doc-hash-table* sys:HEIGHT-NEEDED-FOR-X-WHO-LINE-DOC-LINES))


(defparameter *default-portrait-height-of-suggestions-frame* 133
  "Initial height of the Suggestions frame when in the Portrait configuration.")  

(defparameter *default-portrait-width-of-suggestions-frame*
	      (send W:default-screen :width)
  "Initial width of the Suggestions frame when in the Portrait configuration.")

(defparameter *default-landscape-height-of-suggestions-frame*
	      (send W:default-screen :height)
  "Initial height of the Suggestions frame when in the Landscape configuration.")

(defparameter *default-landscape-width-of-suggestions-frame* 200
  "Initial width of the Suggestions frame when in the Landscape configuration.")

(defparameter *default-height-of-suggestions-off-frame* 20
  "Initial height of the one-line high strip that is used to turn Suggestions on.")

(defparameter *default-width-of-suggestions-off-frame*
	      (send W:default-screen :width)
  "Initial width of the one-line high strip that is used to turn Suggestions on.")

1;;The next four variables, *PORTRAIT-HEIGHT-OF-SUGGESTIONS-FRAME*,
;;*PORTRAIT-WIDTH-OF-SUGGESTIONS-FRAME*, *LANDSCAPE-WIDTH-OF-SUGGESTIONS-FRAME*
;;and *LANDSCAPE-HEIGHT-OF-SUGGESTIONS-FRAME*, should only change if the user changes
;;the dimensions of the Suggestions frame.  The values of these variables are the current
;;dimensions of the Suggestions frame for the configuration specified.*

(defparameter *portrait-height-of-suggestions-frame*
	      *default-portrait-height-of-suggestions-frame*
  "Current height of the Suggestions frame when in the Portrait configuration.")

(defparameter *portrait-width-of-suggestions-frame*
	      *default-portrait-width-of-suggestions-frame*
  "Current width of the Suggestions frame when in the Portrait configuration.")

(defparameter *landscape-width-of-suggestions-frame*
	      *default-landscape-width-of-suggestions-frame*
  "Current width of the Suggestions frame when in the Landscape configuration.")

(defparameter *landscape-height-of-suggestions-frame*
	      *default-landscape-height-of-suggestions-frame*
  "Current height of the Suggestions frame when in the Landscape configuration.")

1;;The next two variables, *WIDTH-OF-SUGGESTIONS-FRAME* and *HEIGHT-OF-SUGGESTIONS-FRAME*,
;;are bound to the current height and width of the Suggestions frame.  These change when the user
;;changes the configuration OR the dimensions of the Suggestions frame.*

(defvar *width-of-suggestions-frame* :unbound
  "Current width of the Suggestions frame.") 

(defvar *height-of-suggestions-frame* :unbound
  "Current height of the Suggestions frame.")

1;;*HEIGHT-OF-SUGGESTIONS-OFF-FRAME* contains the current height of the Suggestions Off Frame.*

(defvar *height-of-suggestions-off-frame* :unbound
  "The height of the one-line high strip that is used to turn Suggestions on.")

1;;The font variables point to fonts that are used for various purposes in the Suggestions frame
;;and who-line documentation.*

(defparameter *suggestions-default-font* fonts:cptfont)  

(defparameter *bold-suggestions-font* fonts:cptfontb)   ; fonts:metsi

(defparameter *italic-suggestions-font* fonts:tr10i)  

(defparameter suggestions-who-line-doc-font fonts:hl12bi)


;(defvar name-of-suggestions-menu :unbound "The main Suggestions menu pane.")
 
;(defvar suggestions-frame :unbound
;   "Bound to the most recently created instance of the flavor SUGGESTIONS-FRAME.")  

;(defvar suggestions-off-frame :unbound
;   "Bound to the one-line high strip that is used to turn Suggestions on.")

;(defvar *suggestions-menus-on?* nil
;   "This variable indicates whether the user has turned Suggestions on or off")
(setq *suggestions-menus-on?* nil)
 
1;;The list *windows-shrunk-for-suggestions* will have a form similar to:
;;  ((<window-instance-1> (:portrait (p1-left p1-top p1-right p1-bottom))
;;                          (:landscape (l1-left l1-top l1-right l1-bottom))
;;                          (:full-screen (fs1-left fs1-top fs1-right fs1-bottom))
;;                          (:suggestions-off-frame (sof1-left sof1-top sof1-right sof1-bottom)))
;;   (<window-instance-2> (:portrait (p2-left p2-top p2-right p2-bottom)))
;;   (<window-instance-3> (:landscape (l3-left l3-top l3-right l3-bottom)))
;;where the lists of items (LEFT TOP RIGHT BOTTOM) contain the dimensions for the window instance when
;;Suggestions is in that configuration.*

;(defvar *windows-shrunk-for-suggestions* nil
;  "Bound to an alist that records the dimensions of application windows that have been shrunk for Suggestions.")
(setq *windows-shrunk-for-suggestions* nil)

(defvar *all-windows-recognized-by-suggestions* nil
   "An alist associating the types of windows that can have Suggestions Menus with information about how to set up Suggestions when a window of the type is selected.")

(defvar *windows-recognized-by-suggestions* nil
   "An alist based on *ALL-WINDOWS-RECOGNIZED-BY-SUGGESTIONS* that contains only entries that the user has chosen via SELECT-SUGGESTIONS-APPLICATIONS.")

(defvar *menu-history-stack-symbols* nil
  "An alist of window types and symbols created to hold the menu history of that window type.")

;(defvar *previously-selected-application* :unbound
;   "Program we came from (from Suggestion's point of view)")

;(defvar *currently-selected-application* :unbound
;   "Current program (from Suggestion's point of view)")
(setq *currently-selected-application* nil)

;(defvar *windows-causing-suggestions-to-push-notification-menu* nil
;  "Used by the Suggestions functions called by W:NOTIFY to help control pushing and popping of the Notification Menu")

1;;Suggestions has its own variable for keeping track of the package in force for the selected window so that
;;Term <n> W doesn't mess things up for Lisp Expression menus.*

;(defvar who-line-package nil
;  "The package object in force in the selected window.
;   This will be the package showing in the Who Line unless the user has done Terminal <n> W.")

;(defvar doing-apropos? nil
;  "A predicate used by ZWEI:COM-FIND-COMMANDS, ZWEI:GET-EXTENDED-SEARCH-STRINGS and ZWEI:PRINT-DOC to determine when to make a menu out of commands collected by com-apropos.")

(setf *suggestions-zmacs-modes*
	'((zwei:dired-mode zwei:com-when-dired-menu)
	  (zwei:bdired-mode zwei:com-when-bdired-menu)
	  (zwei:edit-buffers-mode zwei:com-when-edit-buffers-menu)))

;(defvar *mouse-in-suggestions-pane?* #'(lambda (x) nil))

1;;; Maintenance Note:
;;; When these font names are formally defined in the TV package, these definitions need to be removed.
;;;
;;; All non-default fonts used in the menu entries below are symbolically defined here.  They are defined as
;;; DEFPARAMETERs rather than DEFCONSTANTs to allow user customization.  Notice, that substituting a larger
;;; font is almost certain to cause at least one menu entry to extend beyond the edge of the screen.*

(defparameter menu-subtitle-std-font fonts:tr8b
   "Font used for subtitles in Zmacs Suggestions Menus.  That is, the font to be used
for string type menu items.")  

(defparameter menu-subsubtitle-std-font fonts:tr8
   "Font used the subtitles to MENU-SUBTITLE-STD-FONT in Zmacs Suggestions Menus.")  

(defparameter mouse-comment-std-font fonts:hl10
   "Font used for unexecutable Mouse click comments in menus.")  

(defparameter menu-highlight-font fonts:cptfont
   "Font used to highlight entries for *BASIC* menus.")  

(defparameter first-time-user-highlight-font fonts:tr12b
   "Font used to highlight the \"First Time User MENU\" entry in the main menu.")  

(defparameter *suggestions-fonts* '(*suggestions-default-font* *bold-suggestions-font*))

(defparameter *suggestions-main-menu-fonts*
   '(*suggestions-default-font* *bold-suggestions-font* menu-highlight-font
     mouse-comment-std-font menu-subsubtitle-std-font menu-subtitle-std-font))

(defvar *simple-left-click-documentation-message* "Execute Command")

(defvar *simple-middle-click-documentation-message* "Command Keystroke")

(defvar *simple-right-click-doc-message* "Documentation")

(defvar *right-click-more-doc-message* "*More Documentation Available*")

(defparameter *suggestions-mouse-buttons-top-line*
   `(:mouse-l-1 ,*simple-left-click-documentation-message*
		:mouse-m-1 ,*simple-middle-click-documentation-message*
		:mouse-r-1 ,*simple-right-click-doc-message*
		:mouse-r-2 "System Menu"))  

(defparameter *suggestions-mouse-buttons*
   `(:mouse-l-1 ,*simple-left-click-documentation-message*
		:mouse-m-1 ,*simple-middle-click-documentation-message* 
		:mouse-r-1 ,*simple-right-click-doc-message*
		:font ,suggestions-who-line-doc-font :documentation ""))  

(defparameter *suggestions-off-mouse-buttons*
   `(:font ,suggestions-who-line-doc-font :mouse-l-1 ""))  

(defparameter *label-pane-mouse-buttons*
   `(:mouse-l-1 "Menu History" :mouse-m-1 "Menu Tools"
		:mouse-r-1 *simple-right-click-doc-message*
		:font ,suggestions-who-line-doc-font :documentation "")) 

(defparameter *menu-tools-menu-item*
	      '("Menu Tools" :process-eval
		(com-menu-tools-select menu-tools-display-menu)
		:documentation
		"Pops up a menu of commands that make it possible to build user-customized Suggestions menus."))

(defvar *menu-history-stack-length* 30)  

(add-initialization "Boot Suggestions" '(setq *windows-the-user-has-seen* nil)
		    '(:cold :warm))  

(defvar *keystroke-in-mouse-doc-line?* t)  

;;;based on menu-command-menu-mixin out of cadr3:zwei;screen

(defvar old-current-item nil)  

(defvar old-who-line-documentation-string "")  

(defvar *windows-the-user-has-seen* nil)  

;;;(defvar *windows-that-dont-change-for-suggestions* nil)  

(defvar name-of-top-menu :unbound)  

(defvar name-of-bottom-menu :unbound)  

(defvar name-of-label-pane :unbound)

(defvar menu-tools-menu-list nil)

(defvar menu-tools-display-menu nil)

(defvar *undefined-suggestions-menu-items* nil)

;; List of all functions which have been function-wrapped - for UN-FUNCTION-WRAP.

(defvar function-wrapped-functions nil
  "List of all function specs that have been function-wrapped.")

(defvar suggestions-advised-functions nil "Functions advised by Suggestions")

(EXPORT 'FUNCTION-WRAPPED-FUNCTIONS)
(EXPORT 'SUGGESTIONS-ADVISED-FUNCTIONS)
(export '(sugg:initialize-suggestions-for-application   ;macro    ; DAB 04-24-89
	   sugg:declare-suggestions-for                 ;macro
	   sugg:undeclare-suggestions-for               ;function
	   sugg:with-suggestions-menus-for              ;macro
	   sugg:advise-function-to-push-all-menus       ;macro
           sugg:advise-function-to-push-one-menu        ;macro
           sugg:set-active-command-tables
	   )
	'sugg)

(defvar *suggestions-expand-who-line* t
  "Should the mouse documentation window be expanded to 2 lines when Suggestions is turned on?")

(defvar menu-history-menu nil)

;;This variable controls whether or not to display keystrokes AUTOMATICALLY when the user clicks left or
;;right on a Suggestions menu item.  If the user clicks middle on an item, a keystroke should be displayed
;;even if *pop-up-key-msg?* is set to nil.
(defvar *pop-up-key-msg?* t)

(defvar *pop-up-window-reverse-video?* t)

(defvar pop-up-keystrokes-process
	(make-process "Pop-Up Keystrokes" :priority 0 :quantum 10)
   "The process that handles exposing and deexposing the windows that show the keystroke equivalents of Suggestions Menu items.")

(defvar pop-up-keystrokes-io-buffer (W:make-default-io-buffer))

(defvar *pop-up-test* nil)

;;(defvar zwei:*typeout-window-exposed* nil)

;;(defparameter *keystroke-and-doc-hash-table* (make-hash-table :size 200. :number-of-values 2. :test #'eq))

(defvar *default-bottom-menu-item-list*)
