;;; -*- Mode:Common-Lisp; Package:Sugg; Fonts:(Medfnt Hl12b Hl12bi); Base:10. -*- 

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (c) 1984-1989 Texas Instruments Incorporated. All rights reserved.

(defvar *suggestions-menu-search-function* :unbound)

(defvar *suggestions-menu-search-key* :unbound)

(defun get-function-and-key ()
  (multiple-value-setq
    (*suggestions-menu-search-function* *suggestions-menu-search-key*)
    (zwei:get-extended-search-strings "Suggestions Menu Search (Substring:)"))
  zwei:dis-none)

(defwindow-resource suggestions-pop-up-standalone-editor-frame nil
  :make-window (zwei:pop-up-standalone-editor-frame
		 :height (quotient (send w:selected-window :height) 2)
		 :width (send w:selected-window :width)
		 :number-of-mini-buffer-lines 2)
  :reusable-when :deexposed :initial-copies 0)  

(suggestions-build-menu
  'suggestions-menu-search-top-menu ""
  '(("Menu Search:" :no-select ignore :font *bold-suggestions-font*)
    ("" :no-select ignore)
    ("Continue Menu Search" :process-eval (setq continue-flag :continue)
     :font *bold-suggestions-font*
     :documentation "Search for next menu item that contains the specified string.")
    ("Exit Menu Search" :process-eval (setq continue-flag :exit)
     :font *bold-suggestions-font* :documentation "Exit Suggestions Menu Search."))
  nil "Menu Search Top Menu")

(suggestions-build-menu
  'suggestions-menu-search-bottom-menu ""
  '(("" :no-select ignore)
    ("Lisp Expressions" :suggestions-menu lisp-expressions)
    ("Menu Tools" :process-eval (com-menu-tools-select menu-tools-display-menu)
     :documentation
     "Pops up a menu of commands that make it possible to build user-customized Suggestions menus."))
  nil "Menu Search Bottom Menu")

(defun get-extended-suggestions-menu-search-strings ()
  (setq *suggestions-menu-search-function* nil)
  (setq *suggestions-menu-search-key* nil)
  (cond ((typep *currently-selected-application* 'zwei:zmacs-frame)
	 (w:io-buffer-put tv:selected-io-buffer
			  '(:suggestions-menu (ignore . get-function-and-key) 1)))
	(t (catch 'zwei:top-level
	     (let* ((zwei-mini-buffer-suggestions
		      (copy-list
			(cons 'zwei:zwei-mini-buffer
			      (cdr (assoc (type-of *currently-selected-application*)
					  *all-windows-recognized-by-suggestions*
					  :test #'eq)))))
		    (old-windows-recognized-by-suggestions
		      *windows-recognized-by-suggestions*))
	       (rplaca (nthcdr 2 zwei-mini-buffer-suggestions)
		       'zwei:*search-mini-buffer-comtab*)
	       (setq *windows-recognized-by-suggestions*
		     (cons zwei-mini-buffer-suggestions
			   *windows-recognized-by-suggestions*))
	       (unwind-protect
		   (using-resource
		     (pop-up-editor suggestions-pop-up-standalone-editor-frame
				    w:default-screen)
		     (send pop-up-editor :expose)
		     (let ((*terminal-io* pop-up-editor)
			   (editor-closure (send pop-up-editor :editor-closure)))
		       (w:window-call (pop-up-editor :bury)
				      (catch 'zwei:zwei-command-loop
					(funcall editor-closure pop-up-editor :eval-inside-yourself
						 '(get-function-and-key))
					1;;Clear out the editor's typeout window*
					1;;(otherwise we mess up the editor*
					1;;for the next user of the resource.)*
					(send *terminal-io* :clear-screen)
					(funcall editor-closure pop-up-editor :eval-inside-yourself
						 '(zwei:redisplay-all-windows))))))
		 (setq *windows-recognized-by-suggestions*
		       old-windows-recognized-by-suggestions)))))))

(defvar continue-flag nil)

(defun find-salient-part (item)
  (case (suggestions-menu-item-type-p item)
	(:symbol item)
	(:zmacs-command (cdr item))
	(:character (second item))
	(otherwise (third item))))

(defun item-already-in-list-p (item list)
  (let ((salient-part (find-salient-part item)))
    (do ((items list (cdr items)))
	((null items))
      (if (eq salient-part (find-salient-part (car items)))
	  (return t)))))

(defun get-part-to-test (item)
  (cond ((consp item) (car item))
	((symbolp item) (string item))))

(defun do-menu-looping (search-space)
  (let ((exit-flag nil)
	(collected-items nil))
    (do* ((menu-labels search-space (cdr menu-labels))
	  (current-menu-label (car menu-labels) (car menu-labels)))
	 ((or (null menu-labels) exit-flag) (values exit-flag collected-items))
      (when (eq (suggestions-menu-item-type-p current-menu-label) :suggestions-menu)
	(do* ((items (symbol-value (third current-menu-label)) (cdr items))
	      (current-item (car items) (car items))
	      (part-to-test (get-part-to-test current-item)
			    (get-part-to-test current-item)))
	     ((or (null items) exit-flag))
	  (when (funcall *suggestions-menu-search-function* *suggestions-menu-search-key*
			 part-to-test)
	    (let ((type-of-current-item (suggestions-menu-item-type-p current-item)))
	      (unless (or (null type-of-current-item)
			  (eq type-of-current-item :just-documentation)
			  (item-already-in-list-p current-item collected-items))
		(setq collected-items
		      (cons (rplaca (copy-list current-item)
				    (string-trim '(#\Space) (car current-item)))
			    collected-items))))
	    (when (neq current-menu-label
		       (car (send suggestions-frame :*suggestions-stack*)))
	      1;;(signal-suggestions-process `(:install :middle ,(third current-menu-label)))*
	      1;;WAS CHANGED TO:--Release 3.0 slm*
	      (sys:update-suggestions-menus	:install :middle (third current-menu-label))
	      (sugg:signal-suggestions-process :redisplay-middle))
	    (signal-suggestions-process
	      `(:mouse-box ,current-item ,(symbol-value (third current-menu-label))))
	    (process-wait "Menu Search" #'(lambda () continue-flag))
	    (case continue-flag
		  (:continue (setq continue-flag nil))
		  (:exit (setq continue-flag nil) (setq exit-flag t)))))))))

(defun loop-through-menus (search-space)
  (cond ((suggestions-visible-for-application?)
	 1;;(signal-suggestions-process '(:install :top suggestions-menu-search-top-menu))*
	 1;;(signal-suggestions-process '(:install :bottom suggestions-menu-search-bottom-menu))*
	 1;;WAS CHANGED TO:--Release 3.0 slm*
	 (sys:update-suggestions-menus :install :top 'suggestions-menu-search-top-menu)
	 (sys:update-suggestions-menus :install :bottom 'suggestions-menu-search-bottom-menu)
	 (sugg:signal-suggestions-process :redisplay-all)
	 (unwind-protect (do-menu-looping search-space)
	   (progn
	     1;;(signal-suggestions-process '(:pop :top suggestions-menu-search-top-menu))*
	     1;;(signal-suggestions-process '(:pop :bottom suggestions-menu-search-bottom-menu))*
	     1;;WAS CHANGED TO:--Release 3.0 slm*
	     (sys:update-suggestions-menus :pop :top 'suggestions-menu-search-top-menu)
	     (sys:update-suggestions-menus :pop :bottom 'suggestions-menu-search-bottom-menu)
	     (sugg:signal-suggestions-process :redisplay-all))))
	(t nil)))

(defun suggestions-menu-search ()
  (unwind-protect
      (let ((search-space (symbol-value (send suggestions-frame :*list-menus-symbol*)))
	    (correct-top-menu-contents (car (send suggestions-frame :*top-menu-stack*))))
	(get-extended-suggestions-menu-search-strings)
	(process-wait "Function and Key"
		      #'(lambda ()
			  (and (functionp *suggestions-menu-search-function*)
			       (stringp *suggestions-menu-search-key*)
			       1;;The following is a kludge to make sure that the*
			       1;;Suggestions Menu pushing and popping that may occur*
			       1;;during (GET-EXTENDED-SUGGESTIONS-MENU-SEARCH-STRINGS)*
			       1;;has stopped before LOOP-THROUGH-MENUS is called. *
			       (eq correct-top-menu-contents
				   (car (send suggestions-frame :*top-menu-stack*))))))
	(when (and (functionp *suggestions-menu-search-function*)
		   (stringp *suggestions-menu-search-key*)
		   (eq correct-top-menu-contents
		       (car (send suggestions-frame :*top-menu-stack*))))
	  1;;If (loop-through-menus search-space) returns t, the user exited the menu search intentionally.*
	  (process-run-function
	    "Loop"
	    #'(lambda ()
		(multiple-value-bind (exit-flag collected-commands)
		    (loop-through-menus search-space)
		  (cond (collected-commands
			 (let ((new-menu-name
				 (intern (string-append
					   "COM-MENU-SEARCH-"
					   (string-upcase *suggestions-menu-search-key*)
					   "-MENU") 'sugg))
			       (list-menus-symbol (send suggestions-frame :*list-menus-symbol*)))
			   (sugg:suggestions-build-menu
			     new-menu-name
			     (string-append "A menu of commands containing the string "
					    *suggestions-menu-search-key*)
			     (nreverse collected-commands)
			     (when list-menus-symbol (list list-menus-symbol)))
			   (process-wait
			     "Correct Menu"
			     #'(lambda ()
				 (eq correct-top-menu-contents
				     (car (send suggestions-frame :*top-menu-stack*)))))
			   (install-middle-suggestions-menu new-menu-name))
			 (unless exit-flag
			   (w:mouse-confirm
			     (string-append "No more occurrances of \""
					    *suggestions-menu-search-key* "\" found.")
			     "Move mouse away from window to continue.")))
			(t (w:mouse-confirm
			     (string-append "No occurrances of \""
					    *suggestions-menu-search-key* "\" found.")
			     "Move mouse away from window to continue."))))))))
    "Move mouse away from window to continue."))
