;;; -*- Mode:Common-Lisp; Package:SUGG; Fonts:(MEDFNT HL12B HL12BI); Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (c) 1984-1989 Texas Instruments Incorporated. All rights reserved.

;;; These are the menu tools
;;;     4/6/87 SLM
;;;  *  SELECT-SUGGESTIONS-APPLICATIONS
;;;     Modify arglist to take an optional arg containing the list of applications
;;;     that should have Suggestions turned on.  Also add return value of
;;;     :no-change when this function is entered, but the state of Suggestions for
;;;     any given application remains unchanged.  Without these two changes,
;;;     Profile is unable to turn Suggestions on/off for various applications, and
;;;     update its own SUGGESTIONS-MENUS-ON? variable.  Add helper function to
;;;     create the Suggestions association list for the list of applications.
;;;     Bug #4083.

;;;     Also add test for whether the variable SYS:SUGGESTIONS-FRAME is bound or
;;;     not, and if not, to make an instance of the Suggestions frame.  This delays
;;;     allocation of resources for the Suggestions frame AND delays creation of
;;;     the Suggestions Process until/unless it is needed.  This cuts down on the
;;;     work load of the scheduler.  (Process creation delay design change
;;;     requested by czar.  Also fix for bug #4319.)
;;; 7/18/88 clm - fixed REORDER-SUGGESTIONS-MENU to use the list of items in a menu
;;;               rather than the menu name as the list to reorder. (spr 8346)

(defun make-sure-suggestions-off ()
  "Turn off the non-window aspects of Suggestions"
  (setq *suggestions-menus-on?* nil)
  (setq *previously-selected-application* nil)
  (setq *currently-selected-application* nil)
  (setq *windows-the-user-has-seen* nil)
  (turn-off-pop-up-keystrokes)
  (when (variable-boundp sys:suggestions-frame)
    (let ((suggestions-process (suggestions-process)))
      (if suggestions-process (send suggestions-process :arrest-reason :suggestions-off)))
    (if (send suggestions-frame :exposed-p) (send suggestions-frame :bury))))

(defun turn-all-suggestions-menus-off ()
  "Turn Suggestions off for all applications."
  (make-sure-suggestions-off)
  (when (variable-boundp sys:suggestions-frame)
    (w:delaying-screen-management
      (grow-shrunk-screen-inferiors-back-to-full-screen)
      (send suggestions-frame :deexpose)
      (if (and (eq *currently-selected-application* w:initial-lisp-listener)
	       suggestions-off-frame
	       (not (w:sheet-overlaps-sheet-p
		      w:initial-lisp-listener suggestions-off-frame)))
	  (send suggestions-off-frame :expose))))
  ;;This function must return NIL so that Profile does the right thing.
  nil)

(defun change-suggestions-menus-to-values (suggestions-alist)
  "Change the keyword associated with the item in SUGGESTIONS-ALIST to :value"
  (do* ((items suggestions-alist (cdr items))
	(copied-item (copy-list (car items)) (copy-list (car items)))
	(return-list (list copied-item) (nconc return-list (list copied-item))))
       ((null (cdr items)) return-list)
    (rplaca (cdr copied-item) :value)))

(defun read-symbol-&-add-to-menu (&optional (in-zmacs? t))
  (declare (special lisp-expressions))
  (let* ((newsymbol (zwei:read-function-name
		      "Symbol to be added:"
		      (if in-zmacs? (zwei:relevant-function-name
				      (zwei:window-point w:selected-window)))
		      nil nil))
	 (menus (w:multiple-menu-choose
		  (change-suggestions-menus-to-values lisp-expressions)
		  :label (string-append "Lisp Expressions menus to add " newsymbol " to:"))))
    (when menus (loop for menu in menus
		      ;;;Does the symbol have to be added to the ITEMS property too?
		      do (set menu (cons newsymbol (symbol-value menu)))))
    ;;(signal-suggestions-process `(:install :middle ,(car menus)))
    ;;WAS CHANGED TO:--Release 3.0 slm
    (sys:update-suggestions-menus :install :middle (car menus))
    (signal-suggestions-process :redisplay-middle))
  zwei:dis-none)

(defun add-symbol-to-sugg-from-zmacs-buffer ()
  (cond ((typep w:selected-window 'zwei:zmacs-window-pane)
	 (w:io-buffer-put tv:selected-io-buffer
			   '(:suggestions-menu (ignore . sugg:read-symbol-&-add-to-menu) 1)))
	(t (catch 'zwei:top-level
	     (using-resource (pop-up-editor suggestions-pop-up-standalone-editor-frame
					    w:default-screen)
	       (send pop-up-editor :expose)
	       (let ((*terminal-io* pop-up-editor)
		     (editor-closure (send pop-up-editor :editor-closure)))
		 (w:window-call (pop-up-editor :deactivate)
		   (catch 'zwei:zwei-command-loop
		     (funcall editor-closure pop-up-editor :eval-inside-yourself
			      '(sugg:read-symbol-&-add-to-menu nil))
		     ;;Clear out the editor's typeout window (otherwise we mess up the
		     ;;editor for the next user of the resource).
		     (send *terminal-io* :clear-screen)
		     (funcall editor-closure pop-up-editor :eval-inside-yourself
			      '(zwei:redisplay-all-windows)))))))))) 

(defvar *top-blank-menu-symbols* nil)

(defun make-or-use-top-blank-menu (fn &aux symbol-name)
  (cond ((setq symbol-name (assoc fn *top-blank-menu-symbols* :test #'equal))
	 (second symbol-name))
	(t (setq symbol-name (intern (string-append (princ-to-string fn) "-TOP-BLANK-MENU") 'sugg))
;		 (intern-local (string-append (princ-to-string fn) "-TOP-BLANK-MENU") 'sugg))
	   (push (list fn symbol-name) *top-blank-menu-symbols*)
	   (eval `(defvar ,symbol-name nil))
	   (set symbol-name (cons `(,(string-append (prin1-to-string fn) ":")
				    :no-select ignore :font *bold-suggestions-font*)
				  '(("" :no-select ignore))))
	   symbol-name)))

(defun blank-out-suggestions-panes (function)
  (let ((alist-entry (copy-list (assoc function *windows-recognized-by-suggestions*
				       :test #'equal))))
    (rplaca (cdr alist-entry)
	    (make-or-use-top-blank-menu (or (seventh alist-entry) function)))
    (rplaca (cddr alist-entry) nil)
    (rplaca (cdddr alist-entry) nil)
    (pushnew alist-entry *windows-recognized-by-suggestions* :test #'eq)))

(defun make-suggestions-applications-alist (x &aux alist-entry)
  (let* ((x-recognized?
	   (and (setq alist-entry (assoc (car x) *windows-recognized-by-suggestions*
					 :test #'eq))
		(not (null (fourth alist-entry))))))
    (list (car x) (string-capitalize-words (princ-to-string (or (seventh x) (car x))))
	  `((:suggestions-on ,x-recognized?) (:suggestions-off ,(not x-recognized?))))))

(defun make-list-from-applications-list (applications &aux (alist-values nil))
  (dolist (entry *all-windows-recognized-by-suggestions* alist-values)
    (setq alist-values (append (if (member (car entry) applications)
				   `((,(car entry) :suggestions-on))
				   `((,(car entry) :suggestions-off)))
			       alist-values))))

(defun select-suggestions-applications (&optional applications-list)
  (let ((window-alist
	  (mapcar #'make-suggestions-applications-alist
		  (setq *all-windows-recognized-by-suggestions*
			(sort *all-windows-recognized-by-suggestions* #'alphalessp
			      :key #'(lambda (x) (or (seventh x) (car x))))))))
    (multiple-value-bind (choices exit-reason)
	(if applications-list
	    (values (make-list-from-applications-list applications-list) nil)
	    (w:multiple-choose
	      "Turn Suggestions On For These Applications" window-alist
	      '((:suggestions-on "On" nil (:suggestions-off) (:suggestions-off) nil)
		(:suggestions-off "Off" nil (:suggestions-on) (:suggestions-on) nil))))
      (unless exit-reason
	;;If exit-reason is non-nil, the user aborted out of the menu.
	(unless (boundp 'suggestions-frame)
	  (make-instance 'suggestions-frame))
	(let ((old-windows-recognized-by-suggestions *windows-recognized-by-suggestions*))
	  (do* ((choice choices (cdr choice))
		(window-or-function (caar choice) (caar choice))
		(status-keyword (cadar choice) (cadar choice)))
	       ((null choice))
	    ;;If the following predicate is NIL it means that the user has changed the status of
	    ;;WINDOW-OR-FUNCTION.
	    (unless (second (assoc status-keyword
				   (third (assoc window-or-function window-alist
						 :test #'eq)) :test #'eq))
	      (cond ((eq status-keyword :suggestions-on)
		     (pushnew (assoc window-or-function
				     *all-windows-recognized-by-suggestions*
				     :test #'eq)
			      *windows-recognized-by-suggestions* :test #'eq))
		    ((eq status-keyword :suggestions-off)
		     (cond ((member window-or-function suggestions-advised-functions
				    :test #'eq)
			    (blank-out-suggestions-panes window-or-function))
			   (t (setq *windows-recognized-by-suggestions*
				    (remove window-or-function
					    *windows-recognized-by-suggestions*
					    :test #'(lambda (x y) (eq x (car y)))))))))))
	  (unless *windows-recognized-by-suggestions* 
	    (w:mouse-confirm
	      "No applications have been selected.  Suggestions Menus will not be turned on."
	      "Click here or move mouse away to continue.")
	    (turn-all-suggestions-menus-off)
	    (setq exit-reason :abort))
	  (when (and *windows-recognized-by-suggestions* old-windows-recognized-by-suggestions
		     (equal *windows-recognized-by-suggestions* old-windows-recognized-by-suggestions))
	    (setq exit-reason :no-change)))
	(let ((current-application (or *currently-selected-application*
				       (climb-to-recognized-window w:selected-window))))
	  (when current-application
	    (maybe-resize-selected-application current-application))))
      exit-reason)))

(defun make-item-list-of-superior-menus (list-of-menus)
  (do ((menu list-of-menus (cdr menu)))
      ((null menu) list-of-menus)
    (rplaca menu (list (format nil "~a (~a)" (get (car menu) 'ucl:menu-name) (car menu))
			       :value (car menu)))))

(defun flush-menu (&optional (currently-installed-menu (name-of-selected-menu)))
  "Get rid of these suggestions."
  (let ((superior-menus (get currently-installed-menu 'ucl:superior-menu)))
    (cond (superior-menus
	   (multiple-value-bind (menus-to-remove-from not-aborted?)
	       (w:multiple-menu-choose
		 (make-item-list-of-superior-menus
		   (copy-list (if (symbolp superior-menus) (list superior-menus)
				  superior-menus)))
		 :label (string-append "Select superior menus to delete "
				(get currently-installed-menu 'ucl:menu-name) " from:"
				#\return
				"(Move mouse off of menu to abort menu removal)"))
	     (when not-aborted?
	       (do ((menus menus-to-remove-from (cdr menus)))
		   ((null menus))
		 (if (symbol-value (car menus))
		     (set (car menus)
			  (remove currently-installed-menu (symbol-value (car menus))
				  :test #'(lambda (x y)
					    (and (eq (suggestions-menu-item-type-p y)
						     :suggestions-menu)
						 (eq x (third y))))))))
	       (sys:update-suggestions-menus :remove :middle currently-installed-menu)
	       (install-middle-suggestions-menu
		 (third (car (symbol-value
			       (send suggestions-frame :*menu-history-symbol*))))))))
	  (t (sys:update-suggestions-menus :remove :middle currently-installed-menu)
	     (install-middle-suggestions-menu
	       (third (car (symbol-value
			     (send suggestions-frame :*menu-history-symbol*)))))))))

(defvar menu-for-dumping-into-buffer :unbound)

(defun dump-menu-code-into-buffer ()
  (setq menu-for-dumping-into-buffer (name-of-selected-menu))
  (cond ((typep *currently-selected-application* 'zwei:zmacs-frame)
	 (w:io-buffer-put
	   tv:selected-io-buffer
	   '(:suggestions-menu (ignore . add-menu-to-buffer-at-point) 1)))
	;;This is essentially (ed t), plus one line
	(t (let ((sheet (zwei:find-or-create-idle-zmacs-window)))
	     ;;And select something according to our argument.
	     (send sheet :force-kbd-input '(:execute ucl:edit-thing t))
	     (send sheet :force-kbd-input
		   '(:suggestions-menu (ignore . add-menu-to-buffer-at-point) 1))
	     (send sheet :select)
	     (w:await-window-exposure)))))

(defun last-menu-definition ()
  "Returns a suggestions-build-menu form for the last menu selected"
  (let ((current-menu menu-for-dumping-into-buffer))
    (if (boundp current-menu)
	(read-from-string (build-suggestions-menu-form current-menu) t))))

(defun add-menu-to-buffer-at-point ()
  "Adds code to re-create the current menu."
  (zwei:grind-into-bp (zwei:point) (last-menu-definition))
  (zwei:move-bp (zwei:point) (zwei:interval-last-bp zwei:*interval*))
  (zwei:insert-moving (zwei:point) #\Newline)
  zwei:dis-text)

(defun find-commands-for-ucl-application ()
  (multiple-value-bind (ignore commands search-string)
      (ucl:command-name-search)
    (when commands
      (let ((new-menu-name
	      (intern (string-append "APROPOS-" (string-upcase search-string) "-MENU")
		      'sugg))
	    (superior-menu (send suggestions-frame :*list-menus-symbol*)))
	(sugg:suggestions-build-menu new-menu-name
	 (string-append "A menu of commands containing the string " search-string)
	 ;;;Wrap a call to SORT around this list of items.
	 (loop for comm in commands collect (list (send comm :name) :value comm))
	 (list superior-menu))
	(install-middle-suggestions-menu new-menu-name)
	commands))))

(defun suggestions-command-name-search ()
  (cond ((typep *currently-selected-application* 'zwei:zmacs-frame)
	 (w:io-buffer-put tv:selected-io-buffer
			   '(:suggestions-menu (ignore . zwei:com-find-commands) 1)))
	((or (typep *currently-selected-application* 'ucl:basic-command-loop)
	     (typep *currently-selected-application* 'ucl:command-loop-mixin))
	 (send *currently-selected-application* :eval-inside-yourself
	       '(find-commands-for-ucl-application)))
	(t (w:mouse-confirm
	     (string-append "Find Commands is not supported for "
			    (type-of *currently-selected-application*) ".")
	     "Click here or move the mouse away to continue."))))

(defun toggle-suggestions-configuration ()
  (w:delaying-screen-management
  (case (send suggestions-frame :configuration)
	(:landscape (send suggestions-frame :set-configuration :portrait))
	(:portrait (send suggestions-frame :set-configuration :landscape)))))

(proclaim
  '(special new-menus-string-name
	    ;;;		  documentation 
	    superior-of-new-menu add-new-menu-commands? change-superior-menu?
	    new-menu-commands
	    new-menu-commands-pairs superior-of-new-menu-symbol))

(defun convert-spaces-in-string-to-dashes (string)
  (setq string (string string))
  (dotimes  (i (length string) string)
    (when (char= (aref string i) #\space)
      (setf (aref string i) #\-))))

(defun create-new-suggestions-menu-via-menu ()
  (catch 'out-of-create
    (let ((new-menus-string-name "")
	  (documentation "")
	  (superior-of-new-menu "List Menus")
	  (superior-of-new-menu-symbol (send suggestions-frame :*list-menus-symbol*))
	  (add-new-menu-commands? nil)
	  (change-superior-menu? nil)
	  (new-menu-commands nil)
	  (new-menu-commands-pairs nil)
	  (new-menu-name))
      (w:choose-variable-values
	'("" (new-menus-string-name " Menu name" :string)
	  (superior-of-new-menu " Superior Menu" :string)
	  (change-superior-menu? " Change Superior?" :boolean)
	  (documentation " Documentation" :string)
	  (new-menu-commands " Menu Commands" :princ)
	  (add-new-menu-commands? " Add Command?" :boolean) "")
	:label "New Menu Definition:" :function 'get-values-from-suggestions-menus
	:margin-choices '("Do-it" ("Abort" (*throw 'out-of-create nil)))
	:near-mode
	`(:window
	   ,(or (climb-to-recognized-window w:selected-window)
		w:selected-window))
	:superior w:default-screen)
      (setq new-menu-name
	    (intern (string-append "com-" (convert-spaces-in-string-to-dashes
					    new-menus-string-name)) 'sugg))
      ;;;figure out what the appropriate superior-menu is for this menu
      (sugg:suggestions-build-menu new-menu-name documentation new-menu-commands-pairs)
      (install-middle-suggestions-menu new-menu-name))))

(defun get-values-from-suggestions-menus (window variable old-value new-value)
  old-value				   ;ignore
  new-value
  (cond ((eq variable 'add-new-menu-commands?)
	 (send window :refresh)
	 (prog ()
	    look-for-command
	       (let ((mouse-blip (send *standard-input* :any-tyi)))
		 (push (car (second mouse-blip)) new-menu-commands)
		 (push (second mouse-blip) new-menu-commands-pairs)
		 (setq add-new-menu-commands? nil)
		 (send window :refresh))))
	((eq variable 'change-superior-menu?)
	 (let ((moused-item (send *standard-input* :any-tyi)))
	   (send window :refresh)
	   (setq superior-of-new-menu (car moused-item))
	   (setq superior-of-new-menu-symbol (cdr moused-item))
	   (send window :refresh)))
	(t nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;MENU RE-ORDERING STUFF
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar suggestions-menus-reorder-window :unbound)  

;;;7/18/88 clm - fixed to call symbol-value of command-name-of-menu (spr 8346)
(defun reorder-suggestions-menu (command-name-of-menu)
  (let* ((item-list (symbol-value command-name-of-menu))
	 (ordered-list (send suggestions-menus-reorder-window :reorder item-list)))
    (set command-name-of-menu ordered-list)))

;;;7/18/88 clm - added call to redisplay-suggestions-menus, goes with the above fix.
(defun reorder-current-suggestions-menu ()
  (cond ((not (variable-boundp suggestions-menus-reorder-window))
	 (setq suggestions-menus-reorder-window
	       (make-instance 'w:reorder-menu
			      ;;These would be nice, but don't work
			      ;;:near-mode `(:window suggestions-frame)
			      ;;:width (send name-of-suggestions-menu :width)
			      ))))
  (let ((current-menu (name-of-selected-menu)))
    (reorder-suggestions-menu current-menu)
    (install-middle-suggestions-menu current-menu)
    (redisplay-suggestions-menus)))

(defun symbol-apropos-internals ()
  "Since COM-SYMBOL-APROPOS doesn't return a useful value..."
  (let ((result))
    (zwei:with-typeout-font-map-of
      ((send zwei:*mode-line-window* :search-mini-buffer-window))
      (multiple-value-bind (pkg pkg-name)
	  (zwei:get-package-to-search)
	(multiple-value-bind (function key str)
	    (zwei:get-extended-search-strings
	      (string-append "List symbols in " pkg-name " containing substring:"))
	  (zwei:list-zmacs-callers-to-be-edited
	    "Symbols matching" str nil
	    (zwei:setup-zmacs-callers-to-be-edited
	      (let ((zwei:*function* function)
		    (zwei:*key* key)
		    (zwei:*list* nil))
		(declare (special zwei:*function* zwei:*key* zwei:*list*))
		(if (eq pkg pkg-global-package)
		    (do-all-symbols (sym (setq result (copy-list zwei:*list*)))
				    (and (or (fboundp sym) (boundp sym))
					 (funcall zwei:*function* zwei:*key* (string sym))
					 (push sym zwei:*list*)))
		    (do-local-symbols (sym pkg (setq result (copy-list zwei:*list*)))
				      (and (or (fboundp sym) (boundp sym))
					 (funcall zwei:*function* zwei:*key* (string sym))
					 (push sym zwei:*list*)))))))
;mapatoms and mapcar are obsolete--use do-all-symbols and do-local-symbols instead (see above)
;		(funcall (if (eq pkg pkg-global-package) #'mapatoms-all #'mapatoms)
;			 #'(lambda (sym)
;			     (when (and (or (fboundp sym) (boundp sym))
;					;;transposed these two remembering Lamott's
;					(funcall zwei:*function* zwei:*key*
;						 (string sym)))
;			       ;;claim that it is faster
;			       (push sym zwei:*list*)))
;			 pkg)
;		(setq result (copy-list zwei:*list*)))))
	  (values result str))))))

;;Use this to get around lower-case package problem in Find Symbols For Menu.
(defun make-package-alist ()
  (let ((package-alist (copy-list (sys:list-all-packages))))
    (do ((package-sublist package-alist (cdr package-sublist)))
	((null package-sublist) (sortcar package-alist #'alphalessp))
      (rplaca package-sublist
	      (cons (package-name (car package-sublist)) (car package-sublist))))))

(defun use-symbol-apropos-to-build-menu ()
  (multiple-value-bind (collected-symbols apropos-string)
      (let ((zwei:*numeric-arg*
	      (w:menu-choose
		'(("Look for Symbols in the Current Package" :value 0)
		  ("Look for Symbols in All Packages" :value 4)
		  ("Specify the Package to Look In" :value 16))
		:abort-on-deexpose t)))
	(when zwei:*numeric-arg*
	  (if (= zwei:*numeric-arg* 16)
	      (let ((*package*
		      (let ((pack
			      (w:menu-choose (make-package-alist) :abort-on-deexpose t)))
			(if pack pack *package*))))
		(setq zwei:*numeric-arg* 1)
		(symbol-apropos-internals))
	      (symbol-apropos-internals))))
    (if collected-symbols
	(let ((new-menu-name
		(intern (string-append "COM-" (string-upcase apropos-string)
				       "-SYMBOL-MENU") 'sugg)))
	  (sugg:suggestions-build-menu
	    new-menu-name
	    (string-append "A menu of symbols whose names contain the string "
			   apropos-string)
	    collected-symbols '(lisp-expressions)
	    (string-append (string-capitalize apropos-string) " Symbol Menu"))
	  (install-middle-suggestions-menu new-menu-name))))
  zwei:dis-none)

(defun find-symbols ()
  (cond ((typep *currently-selected-application* 'zwei:zmacs-frame)
	 (w:io-buffer-put
	   tv:selected-io-buffer
	   '(:suggestions-menu (ignore . use-symbol-apropos-to-build-menu) 1)))
	(t (catch 'zwei:top-level
	     (using-resource (pop-up-editor suggestions-pop-up-standalone-editor-frame
					    w:default-screen)
	       (send pop-up-editor :expose)
	       (let ((*terminal-io* pop-up-editor)
		     (editor-closure (send pop-up-editor :editor-closure)))
		 (w:window-call (pop-up-editor :deactivate)
		   (funcall editor-closure pop-up-editor :eval-inside-yourself
			    '(use-symbol-apropos-to-build-menu))
		   ;;Clear out the editor's typeout window (otherwise we mess up the editor
		   ;;for the next user of the resource).
		   (send *terminal-io* :clear-screen)
		   (funcall editor-closure pop-up-editor :eval-inside-yourself
			    '(zwei:redisplay-all-windows)))))))))

(defun back-to-initial-suggestions ()
  (find-suggestions-for-this-window
    (or *currently-selected-application* (climb-to-recognized-window w:selected-window))
    'forget-the-past))

(defun add-item-to-menu-tools-menu-list (item)
  (setq menu-tools-menu-list
	(sortcar (pushnew item menu-tools-menu-list
			  :test #'(lambda (x y) (string-equal (car x) (car y))))
		 #'alphalessp))
  (setq menu-tools-display-menu `(("" :no-select ignore)
				  ("MENU TOOLS" :no-select ignore :font fonts:tr12b)
				  ("" :no-select ignore) ,@menu-tools-menu-list
				  ("" :no-select ignore))))

;;; This is the menu tools select stuff ;;;

(add-item-to-menu-tools-menu-list
  '("Add menu to buffer"
    :eval (signal-suggestions-process '(:menu-tool (dump-menu-code-into-buffer)))
    :documentation
    "Put the code that can be loaded to reproduce the current menu in a Zmacs buffer at point."))

(add-item-to-menu-tools-menu-list
  '("Add symbol to a Lisp Expressions menu"
    :eval (signal-suggestions-process
	    '(:menu-tool (add-symbol-to-sugg-from-zmacs-buffer)))
    :documentation "Add a symbol to the currently installed Suggestions menu by typing it or mousing it in a Zmacs buffer."))

(add-item-to-menu-tools-menu-list
  '("Back to Initial Suggestions"
    :eval (signal-suggestions-process '(:menu-tool (back-to-initial-suggestions)))
    :documentation "Returns all of the Suggestions menus back to the contents they had when this window was selected."))

(add-item-to-menu-tools-menu-list
  '("Find Commands"
    :eval (signal-suggestions-process '(:menu-tool (suggestions-command-name-search)))
    :documentation "Find commands in the current application that are on the set of active command tables."))

(add-item-to-menu-tools-menu-list
  '("Find Symbols for Menu"
    :eval (signal-suggestions-process '(:menu-tool (find-symbols)))
    :documentation "Use Apropos to create a Lisp Expressions menu of all the symbols whose name contains some string."))

(add-item-to-menu-tools-menu-list
  '("Menu History"
    :eval (signal-suggestions-process '(:menu-tool (install-menu-history)))
    :documentation "Display menus that have been selected in this application."))

(add-item-to-menu-tools-menu-list
  '("Remove current menu" :eval (signal-suggestions-process '(:menu-tool (flush-menu)))
    :documentation "Eliminate the current menu."))

(add-item-to-menu-tools-menu-list
  '("Reorder menu items"
    :eval (signal-suggestions-process '(:menu-tool (reorder-current-suggestions-menu)))
    :documentation "Change the ordering of the items in the current menu."))

(add-item-to-menu-tools-menu-list
  '("Select Suggestions Applications"
    :eval (signal-suggestions-process '(:menu-tool (select-suggestions-applications)))
    :documentation "Select applications for which a Suggestions frame should be displayed."))

(add-item-to-menu-tools-menu-list
  '("Suggestions Menu Search"
    :eval (signal-suggestions-process '(:menu-tool (suggestions-menu-search)))
    :documentation "Search the menus of this application for items containing a specified substring."))

(add-item-to-menu-tools-menu-list
  '("Suggestions Menus Off"
    :eval (signal-suggestions-process '(:menu-tool (turn-all-suggestions-menus-off)))
    :documentation "Switch to a configuration without suggestions menus.  keystroke: TERM - SUPR-HELP"))

(add-item-to-menu-tools-menu-list
  '("Toggle Configuration"
    :eval (signal-suggestions-process '(:menu-tool (toggle-suggestions-configuration)))
    :documentation "Switch the Suggestions Frame between its landscape and portrait configurations."))

(add-item-to-menu-tools-menu-list
  '("Turn Off Pop-Up Keystrokes"
    :eval (signal-suggestions-process '(:menu-tool (turn-off-pop-up-keystrokes)))
    :documentation "No reverse video windows will pop-up with the keystroke equivalents of items clicked on in Suggestions Menus."))

(add-item-to-menu-tools-menu-list
  '("Turn On Pop-Up Keystrokes"
    :eval (signal-suggestions-process '(:menu-tool (turn-on-pop-up-keystrokes)))
    :documentation "Reverse video windows will pop-up with the keystroke equivalents of items clicked on in Suggestions Menus."))
