;;; -*- Mode:Common-Lisp; Package:Sugg; Fonts:(Medfnt Hl12b Hl12bi); Base:10. -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (c) 1984-1989 Texas Instruments Incorporated. All rights reserved.

(defun simulate-94-documentation (symbol &optional (long-form? nil) &aux temp doc)
  (cond (long-form?
	 (with-output-to-string (doc-string)
	   (cond-every
	     ((functionp symbol t)
	      (multiple-value-bind (args return macro-p)
		  (arglist symbol)
		(if (setq temp (get symbol 'si:defstruct-slot))
		    (princ (string-append symbol " is a defstruct accessor macro for "
					  (car temp) "." #\Newline) doc-string)
		    (princ (string-append
			     symbol " is a " (if macro-p "macro." "function.") #\Newline)
			   doc-string))
		(princ (string-append "Arguments  " (princ-to-string args) #\Newline)
		       doc-string)
		(when return (princ (string-append "Returns  " (princ-to-string return)
						   #\Newline) doc-string)))
	      (when (setq doc (documentation symbol))
		(princ (string-append doc #\Newline) doc-string)))
	     ((setq temp (get symbol 'si:flavor))
	      (princ (string-append symbol " is a flavor." #\Newline) doc-string)
	      (when (and (setq temp (si:flavor-plist temp))
			 (setq doc (second (member :documentation temp :test #'eq))))
		(princ (string-append (if (consp doc) (car (nleft 1 doc)) doc)
				      #\Newline) doc-string)))
	     ((setq temp (get symbol 'si:defstruct-description))
	      (princ (string-append symbol " is a structure." #\Newline) doc-string)
	      (when (and (setq temp (si:defstruct-description-property-alist temp))
			 (setq doc (rest (assoc :documentation temp :test #'eq))))
		(princ (string-append doc #\Newline) doc-string)))
	     ((and (setq temp (get symbol :documentation)) (neq doc temp))
	      (princ (string-append symbol " is a " (type-of symbol) "." #\Newline)
		     doc-string)
	      (princ (string-append temp #\Newline) doc-string))
	     ((and (symbolp symbol) (boundp symbol))
	      (princ (string-append symbol " is a variable." #\Newline) doc-string)
	      (if (setq doc (documentation symbol 'variable))
		  (princ (string-append doc #\Newline) doc-string))
	      (if (consp (setq temp (symbol-value symbol)))
		  (princ (string-append "Its value is a " (type-of temp) "." #\Newline)
			 doc-string)
		  (princ (string-append "Its value is " (prin1-to-string temp) "."
					#\Newline) doc-string))
	      (when (setq doc (get symbol :value-documentation))
		(princ (string-append doc "." #\Newline) doc-string)))
	     ((setq temp (get symbol 'eh:make-condition-function))
	      (princ (string-append (princ-to-string symbol) " is an "
				    (if (setq doc (documentation temp)) ""
					"undocumented ") "error condition." #\Newline)
		     doc-string)
	      (when doc (princ (string-append doc #\Newline) doc-string)))
	     (otherwise (princ (string-append #\Newline "Documentation for "
					      (prin1-to-string symbol) " was not found.")
			       doc-string)))
	   doc-string))
	(t (with-output-to-string (doc-string)
	     (cond-every ((stringp symbol)
			  (princ (string-append (or symbol "") #\Newline) doc-string))
			 ((and (symbolp symbol) (fboundp symbol))
			  (princ (string-append (or (documentation symbol) ""))
				 doc-string))
			 ((setq temp (get symbol 'si:flavor))
			  (if (and (setq temp (si:flavor-plist temp))
				   (setq doc (second (member :documentation temp
							     :test #'eq))))
			      (princ (string-append (or (second doc) "") #\Newline)
				     doc-string)))
			 ((and (symbolp symbol) (boundp symbol))
			  (princ (string-append (or (documentation symbol 'variable) "")
						#\Newline) doc-string))
			 (otherwise ""))))))

1;;;Thanks to John Hogge use (view-documentation (send (third chosen-item) :documentation))
;;;to handle middle or right clicks for commands.*

(defun compose-menu-blip (chosen-item bd self)
  2"Put the Suggestions menu item CHOSEN-ITEM, the location BD, and the window SELF into
the io-buffer of the currently selected io-buffer."*
  (if W:selected-window
      (W:io-buffer-put tv:selected-io-buffer `(:menu ,chosen-item ,bd ,self))))

(defun calculate-documentation (chosen-item &optional (long-form? nil))
  2"Return the documentation string associated with the Sugggestions menu item CHOSEN-ITEM.
If CHOSEN-ITEM is a renamed symbol, a non-nil LONG-FORM? returns full documentation, else the
short version is returned."*
  (let ((documentation
	  (case (suggestions-menu-item-type-p chosen-item)
		(:renamed-symbol (or (get chosen-item :documentation)
				     (simulate-94-documentation
				       (get chosen-item :symbol) long-form?) ""))
		(:symbol (or (get chosen-item :documentation)
			     (simulate-94-documentation chosen-item long-form?) ""))
		(:suggestions-menu
		 (or (second (member :documentation chosen-item :test #'eq))
		     (get (get chosen-item :suggestions-menu) 'documentation)
		     (get (third chosen-item) 'documentation) ""))
		(:zmacs-command (or (get (cdr chosen-item) :who-line-documentation)
				    (get (cdr chosen-item) 'documentation)
				    (documentation (cdr chosen-item)) ""))
		(:ucl-command (or (send (get chosen-item :value) :description) ""))
		(:character (or (get (cdr chosen-item) :documentation) ""))
		((:process-eval :form-to-eval) (or (get chosen-item :documentation) ""))
		(:just-documentation (or (get chosen-item :documentation) "")))))
    (if (stringp documentation) documentation)))

(defun remove-pop-up-documentation-message (pop-up-message-window)
  2"Enforce the typing of a character to remove the POP-UP-MESSAGE-WINDOW."*
  (cond ((suggestions-visible-for-application?)
	 (let ((saved-main-stack (send sugg:suggestions-frame :*suggestions-stack*))
	       (saved-selected-window
		 (sugg:climb-to-recognized-window W:selected-window))
	       (item-number (send name-of-suggestions-menu :top-row)))
	   (install-middle-suggestions-menu 'zwei:mouse-insensitive-typeout-menu)
	   (unwind-protect (read-char pop-up-message-window)
	     (cond ((eq saved-selected-window
			(climb-to-recognized-window W:selected-window))
		    (sys:update-suggestions-menus :replace :middle saved-main-stack)
		    1;;(sugg:signal-suggestions-process item-number)*
		    (sugg:scroll-main-suggestions-menu item-number))))))
	(t (read-char pop-up-message-window))))

1;;; Pop up a window with a message in it, require user to type a character to flush.*

(defun pop-up-documentation-message (prompt &optional (sup W:default-screen)
				     (W:pop-up-near '(:mouse)) &key (more-p? t))
  2"Set up the initializations and call the appropriate functions to get PROMPT displayed
in a pop-up window near screen location POP-UP-NEAR with SUP as the superior window.
The keyword MORE-P? is non-nil when more-processing should be allowed, and NIL when
more-processing should be turned off for this pop-up documentation display."*
  (using-resource (pop-up-message-window suggestions-pop-up-notification-window sup)
     (let ((message (string-append prompt #\Newline w:*remove-typeout-standard-message*)))
       (send pop-up-message-window :set-label nil)
       (send pop-up-message-window :set-more-p more-p?)
       (send pop-up-message-window :set-size-in-characters message
	  (string-append message #\Newline))
       (send pop-up-message-window :clear-input)
       (w:expose-window-near pop-up-message-window W:pop-up-near nil)
       (W:window-call (pop-up-message-window :deactivate)
	  (send pop-up-message-window :string-out message)
	  1;; back up the cursor by one.  this is easier than trying to make the window*
	  1;; come out wider, because of the interface to :set-size-in-characters.*
	  (multiple-value-bind (x-pos y-pos)
	    (send pop-up-message-window :read-cursorpos :character)
	    (send pop-up-message-window :set-cursorpos (1- x-pos) y-pos :character))
	  (remove-pop-up-documentation-message pop-up-message-window)))))

(defun document-button (chosen-item button menu-pointer)
  2"Handle the case where the second and third mouse buttons are used.  CHOSEN-ITEM is the
Suggestions menu item, BUTTON is an integer indicating which mouse button was used, and
MENU-POINTER is a pointer to the menu where CHOSEN-ITEM came from."*
  (let ((item-spec (suggestions-menu-item-type-p chosen-item)))
    1;;second button and third button handled here--look for 2 or 4 instead of 1*
    (case button
	  (2 (case item-spec
		   (:zmacs-command (compose-menu-blip chosen-item button menu-pointer))
		   (:ucl-command
		    (view-documentation
		      (send (third chosen-item) :text-for-keystroke-display)))
		   (:character (second chosen-item))
		   ((:renamed-symbol :symbol) nil)))
	  (4 (let ((documentation-to-display (calculate-documentation chosen-item t)))
	       (if documentation-to-display
		   (pop-up-documentation-message
		     (string-append documentation-to-display #\Newline) W:default-screen
		     '(:point 0 0))))))))

(defun scroll-main-suggestions-menu (number)
  (if (plusp number) (send name-of-suggestions-menu :scroll-to number :absolute)))

(defun find-who-line-run-state ()
  2"Updates the Suggestions menus on the basis of the run state in the status line."*
  (cond (*suggestions-menus-on?*
	 (let ((selected-window-run-state
		 (if (not W:who-line-process) tv:who-line-run-state
		     (calculate-process-state-of-selected-window))))
	   (cond
	     ((not (string-equal selected-window-run-state "Terminal-"))
	      (case
		(name-of-selected-menu)
		(arrested-process-menu
		 (maybe-pop-menu 'arrested-process-menu "Arrest"))
		(output-hold-menu (maybe-pop-menu 'output-hold-menu "Output Hold"))
		(no-selected-window-menu
		 (maybe-pop-menu 'no-selected-window-menu "No Selected Window"))
		(window-lock-menu (maybe-pop-menu 'window-lock-menu "Window Lock"))
		(lock-menu (maybe-pop-menu 'lock-menu "Lock"))
		(selected-menu (maybe-pop-menu 'selected-menu "Selected"))
		(default-notification-menu
		 (when
		   (not (eq *currently-selected-application*
			    (second
			      (assoc
				(tv:find-interesting-window)
				*windows-causing-suggestions-to-push-notification-menu*
				:test #'equal))))
		   1;;(signal-suggestions-process `(:remove :middle default-notification-menu))*
		   1;;WAS CHANGED TO:*
		   1;;(remove-middle-suggestions-menu 'default-notification-menu)*
		   1;;NOW CHANGED TO:*
		   (sys:update-suggestions-menus :remove :middle 'default-notification-menu)
		   (signal-suggestions-process :redisplay-middle))))
	      1;;Calculate SELECTED-WINDOW-RUN-STATE again because it changes so often.*
	      (setq selected-window-run-state
		    (if (not W:who-line-process) tv:who-line-run-state
			(calculate-process-state-of-selected-window)))
	      (cond ((string-equal selected-window-run-state "Arrest")
		     (push-and-tell 'arrested-process-menu))
		    ((string-equal selected-window-run-state "Output Hold")
		     (push-and-tell 'output-hold-menu))
		    ((string-equal selected-window-run-state "No Selected Window")
		     (push-and-tell 'no-selected-window-menu))
		    ((string-equal selected-window-run-state "Window Lock")
		     (push-and-tell 'window-lock-menu))
		    ((string-equal selected-window-run-state "Lock")
		     (push-and-tell 'lock-menu))
		    ((string-equal selected-window-run-state "Selected")
		     (push-and-tell 'selected-menu)))))))))

(defun suggestions-process ()
   2"The process that handles redisplay and scrolling of suggestions menus, simulated typein, etc."*
   (if (or (not (variable-boundp suggestions-frame)) (null suggestions-frame))
       (make-instance 'suggestions-frame))
   (send suggestions-frame :process))

(defun suggestions-io-buffer ()
   2"The io-buffer used to communicate with the Suggestions Process"*
   (if (or (not (variable-boundp suggestions-frame)) (null suggestions-frame))
       (make-instance 'suggestions-frame))
   (send suggestions-frame :io-buffer))

(defun signal-suggestions-process (blip)
  2"Puts BLIP into the io-buffer of the Suggestions process."*
  (let ((sugg-process (suggestions-process))
	(sugg-io-buffer (suggestions-io-buffer)))
    (and sugg-process (send sugg-process :runnable-p)
	 sugg-io-buffer (W:io-buffer-put sugg-io-buffer blip))))

1;;(W:who-line-process-change current-process) is what causes WHO-LINE-RUN-STATE-UPDATE 
;;to run*

(defun calculate-package-of-selected-window ()
  2"Returns the package of the currently selected window."*
  (let* ((process (if tv:selected-io-buffer
		      (W:io-buffer-last-output-process tv:selected-io-buffer)))
	 (sg (if process (process-stack-group process))))
    (or (if (and process (typep sg 'stack-group))
	    (symeval-in-stack-group '*package* sg)
	    *package*)
	si:pkg-user-package)))  

(defun calculate-who-line-package-for-suggestions ()
  2"Returns the package name displayed in the status line."*
  (if *suggestions-menus-on?*
      (setq who-line-package
	    (pkg-find-package
	      (si:pkg-shortest-name (calculate-package-of-selected-window)) :find))))


(defun suggestions-process-main-loop-internal (io-buffer)
2  "Calls the proper function to perform the task entered into the Suggestions IO-BUFFER."*
  1;;(ignore-errors*
  (let ((io-buffer-input (W:io-buffer-get io-buffer t)))
    (cond ((consp io-buffer-input)
	   (let ((operation-spec (car io-buffer-input)))
	     (cond ((eq (second io-buffer-input) :suggestions-menu)
		    (install-middle-suggestions-menu io-buffer-input))
		   (t (case operation-spec
			(:mouse-box (move-mouse-to-menu-item io-buffer-input))
			(:select-suggestions-menus
			 (select-suggestions-menus (second io-buffer-input)))
			(:menu-tool	   ;(eval (second io-buffer-input))
			 (process-run-function "Form to Eval" `(lambda () ,(second io-buffer-input)))
			 )
			(:install (install-suggestions-menus (cdr io-buffer-input)))
			(:remove (remove-suggestions-menus (cdr io-buffer-input)))
			(:replace (replace-suggestions-menus (cdr io-buffer-input)))
			(:pop (pop-suggestions-menus (cdr io-buffer-input)))
			(:put-io-buffer (let ((*package* (selected-windows-package)))
					  (figure-out-what-to-put-in-io-buffer
					    (second io-buffer-input)))))))))
	  ((numberp io-buffer-input) (scroll-main-suggestions-menu io-buffer-input))
	  (t (handle-miscellaneous-inputs io-buffer-input))))
  ;;)
  )

(defun suggestions-process-top-level (suggestions-frame)
2  "This is the top level loop of the Suggestions process.  It is in a waiting state until
there is work to do."*
  (let ((io-buffer (send suggestions-frame :io-buffer)))
    (do () (nil)
      (process-wait "Suggestions" #'(lambda (buffer)
				      (find-who-line-run-state)
				      (calculate-who-line-package-for-suggestions)
				      (not (W:io-buffer-empty-p buffer)))
		    io-buffer)
      (suggestions-process-main-loop-internal io-buffer))))
