;;; -*- Mode: Common-Lisp; Package: SYSLOG; Base: 8.; Patch-File: T -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985-1989 Texas Instruments Incorporated. All rights reserved.

;;; REVISIONS:
;;;	03.24.87 MBC 	Drop Buffer-Address macro.
;;;
;;;

(DEFVAR *START-BLOCK* () "Physical block number of the start of system log.") 

(DEFVAR *END-BLOCK* () "Physical block number of the end of system log.") 	

(DEFVAR *HEADER-RQB* () "One page rqb which usually contains the first page of the system log.") 

(DEFVAR *BUFFER-RQB* () "Rqb used for data buffering.") 

(DEFVAR *CURRENT-BUFFER-BLOCK* () "Rqb used for data buffering.") 

(DEFVAR *READ-RECORD* ()) 


(DEFVAR *LOG-ENABLED* () "Flag to enable and disable system logging.") 

(DEFVAR *LOG-BUSY* ()
  "Internal lock to prevent logging errors that occur in logging routines (racing).") 


(DEFVAR *LOG-UNIT* 1) 

(DEFVAR *LOG-NAME* "log") 


(DEFPARAMETER HEAD-INDEX 0)			;Use AREF-32b & ASET-32b, !!

(DEFPARAMETER TAIL-INDEX 1) 

(DEFPARAMETER LENGTH-INDEX 2)			;values are in bytes

(DEFPARAMETER INIT-TIME-INDEX 3) 

(DEFPARAMETER LAST-TIME-INDEX 4) 

(DEFPARAMETER INITED-KEY-INDEX 5)		;set to special value if written

(DEFPARAMETER BLOCK-LENGTH-INDEX 6)		;watch for length changes


(DEFPARAMETER INITED-VALUE 12333067547)		;special value "Slog"


(DEFVAR READ-LOG-STATE ()) 


(DEFMACRO READ-HEADER ()
  '(PROGN
     (UNLESS *HEADER-RQB*
       (SETF *HEADER-RQB* (SYSTEM:GET-DISK-RQB 1)))
     (SYSTEM:DISK-READ *HEADER-RQB* *LOG-UNIT* *START-BLOCK*))) 

(DEFMACRO WRITE-HEADER ()
  '(PROGN
     (SI::ASET-32B (GET-UNIVERSAL-TIME) (SYSTEM:RQB-BUFFER *HEADER-RQB*) LAST-TIME-INDEX)
     (SYSTEM:DISK-WRITE *HEADER-RQB* *LOG-UNIT* *START-BLOCK*))) 


(DEFMACRO WRITE-LOG ()
  '(PROGN
     (SYSTEM:DISK-WRITE *BUFFER-RQB* *LOG-UNIT* *CURRENT-BUFFER-BLOCK*)
     (IF (> *CURRENT-BUFFER-BLOCK* *END-BLOCK*)
	 (CERROR T () "writing past end of log")))) 


;;;
;;;  Fetch the pointers or offsets or length from the header block.
;;;

(DEFMACRO GET-HEAD ()
  '(SI::AREF-32B (SYSTEM:RQB-BUFFER *HEADER-RQB*) HEAD-INDEX)) 	


(DEFMACRO GET-TAIL ()
  '(SI::AREF-32B (SYSTEM:RQB-BUFFER *HEADER-RQB*) TAIL-INDEX)) 


(DEFMACRO GET-LENGTH ()
  '(SI::AREF-32B (SYSTEM:RQB-BUFFER *HEADER-RQB*) LENGTH-INDEX)) 


(DEFMACRO PUT-HEAD (VALUE)
  `(SI::ASET-32B ,VALUE (SYSTEM:RQB-BUFFER *HEADER-RQB*) HEAD-INDEX)) 


(DEFMACRO PUT-TAIL (VALUE)
  `(SI::ASET-32B ,VALUE (SYSTEM:RQB-BUFFER *HEADER-RQB*) TAIL-INDEX)) 


(DEFMACRO PUT-LENGTH (VALUE)
  `(SI::ASET-32B ,VALUE (SYSTEM:RQB-BUFFER *HEADER-RQB*) LENGTH-INDEX)) 


(DEFMACRO PUT-INIT-TIME ()
  '(SI::ASET-32B (GET-UNIVERSAL-TIME) (SYSTEM:RQB-BUFFER *HEADER-RQB*) INIT-TIME-INDEX)) 


(DEFMACRO GET-INIT-TIME ()
  '(SI::AREF-32B (SYSTEM:RQB-BUFFER *HEADER-RQB*) INIT-TIME-INDEX)) 


(DEFMACRO GET-LAST-TIME ()
  '(SI::AREF-32B (SYSTEM:RQB-BUFFER *HEADER-RQB*) LAST-TIME-INDEX)) 


(DEFMACRO START-DATA-BLOCK ()
  '(+ 1 *START-BLOCK*)) 
