;;; -*- Mode:Common-Lisp; Package:TELNET; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984- 1989 Texas Instruments Incorporated. All rights reserved.
;;;
;;; Copyright (c) 1980 Massachusetts Institute of Technology 

;;; This file defines: 
;;;	SUPDUP --  a Super-Duper TELNET Protocol.  It is a version of TELNET
;;;		   that incorporates the I.T.S. Intelligent Terminal Protocol.
;;;	SUPDUP-GRAPHICS-MIXIN  -- SUPDUP Graphics Protocol
;;;	GT40-BLINKER  -- GT40 Simulator which is used with the DEC simulator
;;;	                 on I.T.S. for running SUDS
;;;	ARDS SIMULATOR (for compatibility with PTV's)
;;; 03-02-88 DAB Added disconnect function to be called from logout.



;;; For information on SUPDUP Protocol, see RFC #734, 
;;; "Supdup Display Protocol",  Mark Crispin, SU-AI, Oct. 1977

;;; Translate Lisp Machine special character to Supdup command or ascii character


(DEFVAR SUPDUP-KEYS (MAKE-ARRAY 128 :ELEMENT-TYPE '(UNSIGNED-BYTE 16)))

(net:define-logical-contact-name "supdup" '((:chaos "SUPDUP") (:tcp 95)))

(REPLACE SUPDUP-KEYS
	 '(0 2114 2115 26			;null, break, clear, call
	     2113 31 2120 127			;esc, backnext, help, rubout
	     8 9 10 11				;bs, tab, lf, vt
	     12 13 2114 211			;form, cr, quote, hold-output
	     31 2115 200 0			;stop-output, abort, resume, status
	     155 0 0 0 0			;end, I, II, III, IV
	     0 0 0 0 0 2114))			;up, down, left, right, system, network

;; %TOCID - if non-nil, terminal capable of doing character insert/delete operations


(DEFVAR SUPDUP-%TOCID ())  



;;; Dispatch table for the %TD codes.


(DEFVAR SUPDUP-%TD-DISPATCH (MAKE-ARRAY 32))  



(REPLACE SUPDUP-%TD-DISPATCH
      '(SUPDUP-TDMOV SUPDUP-TDMV0 W:SHEET-CLEAR-EOF W:SHEET-CLEAR-EOL
		     W:SHEET-CLEAR-CHAR
		     ;;;  %TDMOV       %TDMV1       %TDEOF		  %TDEOL	     %TDDLF
		     ;;;  200	  201	       202		  203		     204
		     SUPDUP-NOTHING SUPDUP-GT40 W:SHEET-CRLF SUPDUP-NOTHING SUPDUP-TDBS
		     SUPDUP-TDLF
		     ;;;  %TDMTF	    %TDMTN      %TDCRL	      %TDNOP         %TDBS       %TDLF
		     ;;;  205	    206		207	      210	     211	 212
		     SUPDUP-TDCR SUPDUP-TDORS SUPDUP-TDQOT W:SHEET-SPACE SUPDUP-TDMV0
		     SUPDUP-CLEAR
		     ;;;  %TDCR	 %TDORS       %TDQOT       %TDFS          %TDMV0       %TDCLR
		     ;;;  213	 214	      215	   216		  217	       220
		     SUPDUP-BEEP SUPDUP-NOTHING SUPDUP-INSERT-LINE
		     SUPDUP-DELETE-LINE
		     ;;;  %TDBEL	 %TDINI		%TDILP		   %TDDLP
		     ;;;  221	 222		223		   224
		     SUPDUP-INSERT-CHAR SUPDUP-DELETE-CHAR SUPDUP-TDBOW SUPDUP-RESET
		     SUPDUP-GRAPHICS
		     ;;;  %TDICP	    	%TDDCP		   %TDBOW	 %TDRST	     %TDGRF
		     ;;;  225		226		   227		 230	     231
		     SUPDUP-REGION-UP
		     SUPDUP-REGION-DOWN
		     ;;;  %TDRSU	      %TDRSD
		     ;;;  232	      233
		     ;;;  PTV compatibility hacks (ARDS, etc.)
		     SUPDUP-NOTHING SUPDUP-ARDS-SET
		     ;;;  %TDGXT         %TDLNG
		     ;;;  234	    235
		     SUPDUP-ARDS-LONG SUPDUP-ARDS-SHORT
		     ;;;  %TDLV              %TDSV
		     ;;;  236		237
		     ))  

;;; Handle %TDMOV by ignoring two characters and then acting as if it were a %TDMV0.


(DEFUN SUPDUP-TDMOV (SHEET)
  (NVT-NETI)
  (NVT-NETI)
  (SUPDUP-TDMV0 SHEET))  

;;; Handle %TDMV0 or %TDMV1 by moving the cursor.  This is kludgey because
;;; ITS sends out positions as VPOS followed by HPOS.


(DEFUN SUPDUP-TDMV0 (SHEET &AUX YPOS)
  (DECLARE (:self-flavor W:WINDOW))
  (SETQ YPOS (* (NVT-NETI) W:LINE-HEIGHT))
  (W:SHEET-SET-CURSORPOS SHEET
			 (* (NVT-NETI) W:CHAR-WIDTH) YPOS))  

;;; This "null function" is used for codes which we should ignore.


(DEFUN SUPDUP-NOTHING (IGNORE)
  ())  

;;; %TDQOT means the next character should be quoted.


(DEFUN SUPDUP-TDQOT (SHEET)
  (W::SHEET-TYO SHEET (NVT-NETI)))  



(DEFUN SUPDUP-TDBS (SHEET)
  (SEND SHEET :BACKWARD-CHAR))  



(DEFUN SUPDUP-TDCR (SHEET)
  (W:SHEET-SET-CURSORPOS SHEET 0 (NTH-VALUE 1 (W:SHEET-READ-CURSORPOS SHEET))))  



(DEFUN SUPDUP-TDLF (SHEET)
  (DECLARE (:SELF-FLAVOR W:WINDOW))
  (SEND SHEET :INCREMENT-CURSORPOS 0 W:LINE-HEIGHT))  

;;; %TDBEL means ring the "bell".
;;; To avoid gross obnoxosity, we merge multiple consecutive beeps into one.


(DEFUN SUPDUP-BEEP (IGNORE)
  (DECLARE (:self-flavor BASIC-NVT))
  (SEND SELF :REMOTE-BEEP)
  (DO ((CH (FUNCALL STREAM :TYI-NO-HANG) (FUNCALL STREAM :TYI-NO-HANG)))
      ((OR (NULL CH) (/= CH 145))
       (AND CH (FUNCALL STREAM :UNTYI CH)))))  

;;; %TDCLR means erase the screen.


(DEFUN SUPDUP-CLEAR (SHEET)
  (DECLARE (SPECIAL GT40-DISPLAY-LIST))
  (W:SHEET-CLEAR SHEET)
  (REPLACE GT40-DISPLAY-LIST NIL))  
    
;;; %TDILP means to insert lines, takes one arg from stream which is number of lines to insert.
;;; Lines are inserted at current VPOS.  The current line is affected.


(DEFUN SUPDUP-INSERT-LINE (SHEET)
  (W:SHEET-INSERT-LINE SHEET (NVT-NETI)))  

;;; %TDDLP means to delete lines, takes one arg from stream which is the number of lines.
;;;Affects the current line.


(DEFUN SUPDUP-DELETE-LINE (SHEET)
  (W:SHEET-DELETE-LINE SHEET (NVT-NETI)))  

;;; %TDRSU followed by height, n-lines


(DEFUN SUPDUP-REGION-UP (SHEET &OPTIONAL (REGION-HEIGHT (NVT-NETI)) (SCROLL-AMOUNT (NVT-NETI)))
  (W:PREPARE-SHEET (SHEET)
    (LET ((ARRAY (W:SHEET-SCREEN-ARRAY SHEET))
	  (WIDTH (W:SHEET-INSIDE-WIDTH SHEET))
	  (LINE-HEIGHT (W:SHEET-LINE-HEIGHT SHEET))
	  REGION-BOTTOM
	  DELTA-HEIGHT)
      (SETQ REGION-HEIGHT (* REGION-HEIGHT LINE-HEIGHT)
	    REGION-BOTTOM
	    (MIN (+ (W:SHEET-CURSOR-Y SHEET) REGION-HEIGHT)
		 (* LINE-HEIGHT (TRUNCATE (W:SHEET-INSIDE-BOTTOM SHEET) LINE-HEIGHT)))
	    REGION-HEIGHT (- REGION-BOTTOM (W:SHEET-CURSOR-Y SHEET))
	    SCROLL-AMOUNT (MIN (* SCROLL-AMOUNT LINE-HEIGHT) REGION-HEIGHT))
      ;; Get size of region to BLT up
      (SETQ DELTA-HEIGHT (- REGION-HEIGHT SCROLL-AMOUNT))
      (OR (<= DELTA-HEIGHT 0)			;If some bits to move, move them
	  (BITBLT W:ALU-SETA WIDTH DELTA-HEIGHT ARRAY (W:SHEET-INSIDE-LEFT SHEET)
		  (+ (W:SHEET-CURSOR-Y SHEET) SCROLL-AMOUNT) ARRAY (W:SHEET-INSIDE-LEFT SHEET)
		  (W:SHEET-CURSOR-Y SHEET)))
      (SYSTEM:%DRAW-RECTANGLE WIDTH SCROLL-AMOUNT (W:SHEET-INSIDE-LEFT SHEET)
			      (- REGION-BOTTOM SCROLL-AMOUNT) (W:SHEET-ERASE-ALUF SHEET) SHEET))))  

;;; %TDRSD followed by height, n-lines


(DEFUN SUPDUP-REGION-DOWN (SHEET &OPTIONAL (REGION-HEIGHT (NVT-NETI)) (SCROLL-AMOUNT (NVT-NETI)))
  (W:PREPARE-SHEET (SHEET)
    (LET ((ARRAY (W:SHEET-SCREEN-ARRAY SHEET))
	  (WIDTH (W:SHEET-INSIDE-WIDTH SHEET))
	  (LINE-HEIGHT (W:SHEET-LINE-HEIGHT SHEET))
	  REGION-BOTTOM
	  DELTA-HEIGHT)
      (SETQ REGION-HEIGHT (* REGION-HEIGHT LINE-HEIGHT)
	    REGION-BOTTOM
	    (MIN (+ (W:SHEET-CURSOR-Y SHEET) REGION-HEIGHT)
		 (* LINE-HEIGHT (TRUNCATE (W:SHEET-INSIDE-BOTTOM SHEET) LINE-HEIGHT)))
	    REGION-HEIGHT (- REGION-BOTTOM (W:SHEET-CURSOR-Y SHEET))
	    SCROLL-AMOUNT (MIN (* SCROLL-AMOUNT LINE-HEIGHT) REGION-HEIGHT))
      ;; Get negative size of region to BLT down
      (SETQ DELTA-HEIGHT (- SCROLL-AMOUNT REGION-HEIGHT))
      (OR (>= DELTA-HEIGHT 0)			;If some bits to move, move them
	  (BITBLT W:ALU-SETA WIDTH DELTA-HEIGHT ARRAY (W:SHEET-INSIDE-LEFT SHEET)
		  (W:SHEET-CURSOR-Y SHEET) ARRAY (W:SHEET-INSIDE-LEFT SHEET)
		  (+ (W:SHEET-CURSOR-Y SHEET) SCROLL-AMOUNT)))
      (SYSTEM:%DRAW-RECTANGLE WIDTH SCROLL-AMOUNT (W:SHEET-INSIDE-LEFT SHEET)
			      (W:SHEET-CURSOR-Y SHEET) (W:SHEET-ERASE-ALUF SHEET) SHEET))))  

;;; %TDICP means insert character positions, takes an arg.


(DEFUN SUPDUP-INSERT-CHAR (SHEET)
  (W:SHEET-INSERT-CHAR SHEET (NVT-NETI)))  

;;; %TDDCP means delete character positions, takes an arg.


(DEFUN SUPDUP-DELETE-CHAR (SHEET)
  (W:SHEET-DELETE-CHAR SHEET (NVT-NETI)))  

;;; %TDBOW means display black characters on white screen.


(DEFUN SUPDUP-TDBOW (SHEET)
  (SEND SHEET :SET-BLACK-ON-WHITE T))  

;;; %TDRST means reset %TDBOW


(DEFUN SUPDUP-RESET (SHEET)
  (SEND SHEET :SET-BLACK-ON-WHITE ())
  (GRAPHICS-RESET SHEET))  



;;; SUPDUP Graphics Protocol



(DEFFLAVOR SUPDUP-GRAPHICS-MIXIN
	   (GRAPHICS-X-OFFSET
	    GRAPHICS-Y-OFFSET
	    GRAPHICS-VIRT-SCALE
	    GRAPHICS-XPOS
	    GRAPHICS-YPOS
	    GRAPHICS-LEFT
	    GRAPHICS-TOP
	    GRAPHICS-RIGHT
	    GRAPHICS-BOTTOM
	    GRAPHICS-XOR-MODE
	    GRAPHICS-VIRTUAL-MODE)
	   ()
  (:included-flavors w:sheet))

;Note that ALL four edge coordinates are INCLUSIVE:
;they are values corresponding to points which actually exist.
;This is in contrast to the window system,
;in which the lower limits are inclusive and the upper are exclusive.



(DEFMETHOD (SUPDUP-GRAPHICS-MIXIN :AFTER :INIT) (&REST IGNORE)
  (GRAPHICS-RESET SELF)
  (SETQ GRAPHICS-XPOS 0
	GRAPHICS-YPOS 0))  

;Initialize all the variables used for graphics commands.


(DEFUN GRAPHICS-RESET (WINDOW)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
    (LET ((CORRECTED-RIGHT
	    (+ (W:SHEET-INSIDE-LEFT WINDOW)
	       (* (W:SHEET-CHAR-WIDTH WINDOW)
		  (TRUNCATE (W:SHEET-INSIDE-WIDTH WINDOW) (W:SHEET-CHAR-WIDTH WINDOW)))))
	  (CORRECTED-BOTTOM
	    (+ (W:SHEET-INSIDE-TOP WINDOW)
	       (* (W:SHEET-LINE-HEIGHT WINDOW)
		  (TRUNCATE (W:SHEET-INSIDE-HEIGHT WINDOW) (W:SHEET-LINE-HEIGHT WINDOW))))))
      (SETQ GRAPHICS-X-OFFSET (TRUNCATE (+ (W:SHEET-INSIDE-LEFT WINDOW) CORRECTED-RIGHT) 2)
	    GRAPHICS-Y-OFFSET (TRUNCATE (+ (W:SHEET-INSIDE-TOP WINDOW) CORRECTED-BOTTOM) 2)
	    GRAPHICS-XPOS 0
	    GRAPHICS-YPOS 0
	    GRAPHICS-VIRT-SCALE
	    (/
	      (MIN (- CORRECTED-RIGHT (W:SHEET-INSIDE-LEFT WINDOW))
		   (- CORRECTED-BOTTOM (W:SHEET-INSIDE-TOP WINDOW)))
	      2.0s0 2048)
	    GRAPHICS-XOR-MODE ()
	    GRAPHICS-VIRTUAL-MODE ()
	    GRAPHICS-LEFT (- (W:SHEET-INSIDE-LEFT WINDOW) GRAPHICS-X-OFFSET)
	    GRAPHICS-RIGHT (- (W:SHEET-INSIDE-RIGHT WINDOW) GRAPHICS-X-OFFSET 1)
	    GRAPHICS-BOTTOM (- GRAPHICS-Y-OFFSET (1- (W:SHEET-INSIDE-BOTTOM WINDOW)))
	    GRAPHICS-TOP (- GRAPHICS-Y-OFFSET (W:SHEET-INSIDE-TOP WINDOW)))))  



(DEFVAR GRAPHICS-DISPATCH (MAKE-ARRAY 64))  



(REPLACE GRAPHICS-DISPATCH
      '(GRAPHICS-NOTHING GRAPHICS-MOVE GRAPHICS-XOR GRAPHICS-NOTHING GRAPHICS-NOTHING
			 GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-ERASE-SCREEN
			 GRAPHICS-PUSH GRAPHICS-VIRTUAL GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-LIMIT
			 GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-MOVE GRAPHICS-IOR
			 GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-NOTHING
			 GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-PHYSICAL
			 GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-NOTHING GRAPHICS-NOTHING
			 GRAPHICS-NOTHING))  



(DEFVAR DRAW-DISPATCH (MAKE-ARRAY 16))  



(REPLACE DRAW-DISPATCH
      '(GRAPHICS-NOTHING GRAPHICS-DRAW-LINE GRAPHICS-DRAW-POINT GRAPHICS-DRAW-RECT
			 GRAPHICS-DRAW-STRING GRAPHICS-DRAW-BITS GRAPHICS-DRAW-RUNS GRAPHICS-NOTHING))  

;;; %TDGRF #o231


(DEFUN SUPDUP-GRAPHICS (WINDOW)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (CATCH 'SUPDUP-GRAPHICS
    (DO (CH)
	(NIL)
      (SETQ CH (GRAPHICS-NETI))
      (COND
	((LOGTEST CH 64)
	 (FUNCALL (OR (AREF DRAW-DISPATCH (LOGAND CH 15)) 'GRAPHICS-NOTHING) WINDOW CH))
	(T (FUNCALL (OR (AREF GRAPHICS-DISPATCH CH) 'GRAPHICS-NOTHING) WINDOW CH))))))  

;;; Subroutines for graphics commands.



(DEFUN GRAPHICS-NETI ()
  (DECLARE (:self-flavor BASIC-NVT))
    (LET ((CH (NVT-NETI)))
      (AND (LOGTEST CH 128) (THROW 'SUPDUP-GRAPHICS
				  (FUNCALL
				    STREAM
				    :UNTYI
				    CH)))
      CH))  



(DEFUN GRAPHICS-READ-POINT (CH &AUX CH1 CH2 CH3 CH4)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
    (COND
      ((LOGTEST CH 16)
       (SETQ GRAPHICS-XPOS
	     (14-BIT-SIGN-EXTEND
	       (+ (SETQ CH1 (GRAPHICS-NETI)) (LSH (SETQ CH2 (GRAPHICS-NETI)) 7))))
       (SETQ GRAPHICS-YPOS
	     (14-BIT-SIGN-EXTEND
	       (+ (SETQ CH3 (GRAPHICS-NETI)) (LSH (SETQ CH4 (GRAPHICS-NETI)) 7)))))
      (T (INCF GRAPHICS-XPOS (SETQ CH1 (7-BIT-SIGN-EXTEND (GRAPHICS-NETI))))
	 (INCF GRAPHICS-YPOS (SETQ CH2 (7-BIT-SIGN-EXTEND (GRAPHICS-NETI)))))))  



(DEFUN 7-BIT-SIGN-EXTEND (NUMBER)
  (COND
    ((LOGTEST NUMBER 64) (- NUMBER 128))
    (T NUMBER)))  
  


(DEFUN 14-BIT-SIGN-EXTEND (NUMBER)
  (COND
    ((LOGTEST NUMBER 8192) (- NUMBER 16384))
    (T NUMBER)))  



(DEFUN GRAPHICS-ALU (CH)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
     (COND
       (GRAPHICS-XOR-MODE W:ALU-XOR)
       ((LOGTEST CH 32) W:ALU-ANDCA)
       (T W:ALU-IOR)))  



(DEFUN GRAPHICS-X-COORD (COORD)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
    (AND GRAPHICS-VIRTUAL-MODE (SETQ COORD (VALUES (ROUND (* COORD GRAPHICS-VIRT-SCALE)))))
    (+ GRAPHICS-X-OFFSET (MIN GRAPHICS-RIGHT (MAX GRAPHICS-LEFT COORD))))  



(DEFUN GRAPHICS-Y-COORD (COORD)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
    (AND GRAPHICS-VIRTUAL-MODE (SETQ COORD (VALUES (ROUND (* COORD GRAPHICS-VIRT-SCALE)))))
    (- GRAPHICS-Y-OFFSET (MIN GRAPHICS-TOP (MAX GRAPHICS-BOTTOM COORD))))  



(DEFUN GRAPHICS-Y-IN-RANGE (&AUX COORD)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (SETQ COORD GRAPHICS-YPOS)
  (AND GRAPHICS-VIRTUAL-MODE (SETQ COORD (VALUES (ROUND (* COORD GRAPHICS-VIRT-SCALE)))))
  (<= GRAPHICS-BOTTOM COORD GRAPHICS-TOP))  

;;; Graphics commands.



(DEFUN GRAPHICS-NOTHING (IGNORE IGNORE)
  ())  



(DEFUN GRAPHICS-XOR (IGNORE IGNORE)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (SETQ GRAPHICS-XOR-MODE T))  



(DEFUN GRAPHICS-IOR (IGNORE IGNORE)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (SETQ GRAPHICS-XOR-MODE ()))  



(DEFUN GRAPHICS-VIRTUAL (IGNORE IGNORE)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (SETQ GRAPHICS-VIRTUAL-MODE T))  



(DEFUN GRAPHICS-PHYSICAL (IGNORE IGNORE)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (SETQ GRAPHICS-VIRTUAL-MODE ()))  



(DEFUN GRAPHICS-MOVE (IGNORE CH)
  (GRAPHICS-READ-POINT CH))  



(DEFUN GRAPHICS-PUSH (WINDOW IGNORE)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (LET-GLOBALLY					;use to be LET before fast flavors.
    ((GRAPHICS-XPOS GRAPHICS-XPOS) (GRAPHICS-YPOS GRAPHICS-YPOS)
     (GRAPHICS-RIGHT GRAPHICS-RIGHT) (GRAPHICS-LEFT GRAPHICS-LEFT) (GRAPHICS-TOP GRAPHICS-TOP)
     (GRAPHICS-BOTTOM GRAPHICS-BOTTOM) (GRAPHICS-XOR-MODE GRAPHICS-XOR-MODE)
     (GRAPHICS-VIRTUAL-MODE GRAPHICS-VIRTUAL-MODE))
    (DO (CH)
	(NIL)
      (SETQ CH (GRAPHICS-NETI))
      (COND
	((LOGTEST CH 64)
	 (FUNCALL (OR (AREF DRAW-DISPATCH (LOGAND CH 15)) 'GRAPHICS-NOTHING) WINDOW CH))
	(T (FUNCALL (OR (AREF GRAPHICS-DISPATCH CH) 'GRAPHICS-NOTHING) WINDOW CH))))))  



(DEFUN GRAPHICS-LIMIT (IGNORE CH)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (GRAPHICS-READ-POINT CH)
  (LET ((OXPOS GRAPHICS-XPOS)
	(OYPOS GRAPHICS-YPOS))
    (GRAPHICS-READ-POINT CH)
    (SETQ GRAPHICS-LEFT (MIN GRAPHICS-XPOS OXPOS)
	  GRAPHICS-RIGHT (MAX GRAPHICS-XPOS OXPOS)
	  GRAPHICS-BOTTOM (MIN GRAPHICS-YPOS OYPOS)
	  GRAPHICS-TOP (MAX GRAPHICS-YPOS OYPOS))))  

;;; Drawing commands.



(DEFUN GRAPHICS-DRAW-LINE (WINDOW CH)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (W:PREPARE-SHEET (WINDOW)
    (LET ((OXPOS GRAPHICS-XPOS)
	  (OYPOS GRAPHICS-YPOS))
      (GRAPHICS-READ-POINT CH)
      (SYSTEM:%DRAW-LINE (GRAPHICS-X-COORD OXPOS) (GRAPHICS-Y-COORD OYPOS)
			 (GRAPHICS-X-COORD GRAPHICS-XPOS) (GRAPHICS-Y-COORD GRAPHICS-YPOS)
			 (GRAPHICS-ALU CH) T WINDOW))))  



(DEFUN GRAPHICS-DRAW-POINT (WINDOW CH)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (W:PREPARE-SHEET (WINDOW) (GRAPHICS-READ-POINT CH)
		   (SYSTEM:%DRAW-LINE (GRAPHICS-X-COORD GRAPHICS-XPOS) (GRAPHICS-Y-COORD GRAPHICS-YPOS)
				      (GRAPHICS-X-COORD GRAPHICS-XPOS) (GRAPHICS-Y-COORD GRAPHICS-YPOS)
				      (GRAPHICS-ALU CH) T WINDOW)))  



(DEFUN GRAPHICS-DRAW-RECT (WINDOW CH)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (W:PREPARE-SHEET (WINDOW)
    (LET ((OXPOS GRAPHICS-XPOS)
	  (OYPOS GRAPHICS-YPOS))
      (GRAPHICS-READ-POINT CH)
      (SYSTEM:%DRAW-RECTANGLE
	(ABS (1+ (- (GRAPHICS-X-COORD GRAPHICS-XPOS) (GRAPHICS-X-COORD OXPOS))))
	(ABS (1+ (- (GRAPHICS-Y-COORD GRAPHICS-YPOS) (GRAPHICS-Y-COORD OYPOS))))
	(MIN (GRAPHICS-X-COORD GRAPHICS-XPOS) (GRAPHICS-X-COORD OXPOS))
	(MIN (GRAPHICS-Y-COORD GRAPHICS-YPOS) (GRAPHICS-Y-COORD OYPOS)) (GRAPHICS-ALU CH)
	WINDOW))))  



(DEFUN GRAPHICS-ERASE-SCREEN (WINDOW IGNORE)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (W:PREPARE-SHEET (WINDOW)
    (SYSTEM:%DRAW-RECTANGLE
      (1+ (- (GRAPHICS-X-COORD GRAPHICS-RIGHT) (GRAPHICS-X-COORD GRAPHICS-LEFT)))
      (1+ (- (GRAPHICS-Y-COORD GRAPHICS-BOTTOM) (GRAPHICS-Y-COORD GRAPHICS-TOP)))
      (GRAPHICS-X-COORD GRAPHICS-LEFT) (GRAPHICS-Y-COORD GRAPHICS-TOP)
      (W:SHEET-ERASE-ALUF WINDOW) WINDOW)))  



(DEFUN GRAPHICS-DRAW-STRING (WINDOW CH)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (LET ((STRING (MAKE-ARRAY 64 :ELEMENT-TYPE 'STRING-CHAR :LEADER-LIST '(0))))
    (DO ()
	(NIL)
      (LET ((CH (GRAPHICS-NETI)))
	(AND (ZEROP CH) (RETURN (VALUES)))
	(VECTOR-PUSH-EXTEND CH STRING)))
    (SEND WINDOW :STRING-OUT-EXPLICIT STRING (GRAPHICS-X-COORD GRAPHICS-XPOS)
	  (GRAPHICS-Y-COORD GRAPHICS-YPOS) (GRAPHICS-X-COORD GRAPHICS-RIGHT) ()
	  (W:SHEET-CURRENT-FONT WINDOW) (GRAPHICS-ALU CH))
    (RETURN-ARRAY (PROG1
		    STRING
		    (SETF STRING ())))))  



(DEFUN GRAPHICS-DRAW-BITS (WINDOW CH &AUX BIT-CHANGE-FUNCTION)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (W:PREPARE-SHEET (WINDOW) (SETQ BIT-CHANGE-FUNCTION (GRAPHICS-ALU CH))
		   (DO ((I 1 (1+ I)))
		       (NIL)
		     (LET ((NBITS (COND
				    ((ZEROP (REM I 3)) 4)
				    (T 6)))
			   (X (GRAPHICS-X-COORD GRAPHICS-XPOS))
			   (Y (GRAPHICS-Y-COORD GRAPHICS-YPOS))
			   (CH (GRAPHICS-NETI)))
		       (AND (LOGTEST CH 64) (RETURN (VALUES)))
		       (AND (GRAPHICS-Y-IN-RANGE)
			    (DOTIMES (BIT NBITS)
			      (OR (>= GRAPHICS-XPOS GRAPHICS-RIGHT)
				  (AND (LOGTEST (LSH 1 (- NBITS 1 BIT)) CH)
				       (SETF (AREF W:SCREEN-ARRAY Y (+ BIT X))
					     (BOOLE BIT-CHANGE-FUNCTION (AREF W:SCREEN-ARRAY Y (+ BIT X)) 1))))))
		       (INCF GRAPHICS-XPOS NBITS)))))  



(DEFUN GRAPHICS-DRAW-RUNS (WINDOW CH &AUX BIT-CHANGE-FUNCTION)
  (DECLARE (:self-flavor SUPDUP-GRAPHICS-MIXIN))
  (W:PREPARE-SHEET (WINDOW) (SETQ BIT-CHANGE-FUNCTION (GRAPHICS-ALU CH))
		    (DO ()
			(NIL)
		      (LET ((CH (GRAPHICS-NETI))
			    (OLDX (GRAPHICS-X-COORD GRAPHICS-XPOS))
			    (Y (GRAPHICS-Y-COORD GRAPHICS-YPOS)))
			(AND (ZEROP CH) (RETURN (VALUES)))
			(INCF GRAPHICS-XPOS (LOGAND CH 63))
			(AND (LOGTEST CH 64) (GRAPHICS-Y-IN-RANGE)
			     (SYSTEM:%DRAW-LINE OLDX Y (MAX OLDX (1- (GRAPHICS-X-COORD GRAPHICS-XPOS))) Y
						BIT-CHANGE-FUNCTION T WINDOW))))))  


;;; BASIC-SUPDUP



(DEFFLAVOR BASIC-SUPDUP
	   ()
	   (BASIC-NVT
	    SUPDUP-GRAPHICS-MIXIN
	    W:FULL-SCREEN-HACK-MIXIN)
  (:DEFAULT-INIT-PLIST :PROGRAM-NAME "Supdup")
  (:DOCUMENTATION :SPECIAL-PURPOSE "A SUPDUP NVT"))  



(DEFMETHOD (BASIC-SUPDUP :BEFORE :INIT) (INIT-PLIST)
  (SETQ TV:LABEL "FOO")
  (SETF (GET INIT-PLIST :MORE-P) ()))  



(DEFMETHOD (BASIC-SUPDUP :VERIFY-NEW-EDGES)
	   (NEW-LEFT NEW-TOP NEW-WIDTH NEW-HEIGHT)
  NEW-LEFT
  NEW-TOP					;Unused
  (AND stream (OR (/= NEW-WIDTH W:WIDTH) (/= NEW-HEIGHT W:HEIGHT))
       "Attempt to change size while connected"))  

;;; If more processing is turned on, a disaster results!
;;; The output process gets stuck and you can't type input at it to unhang it.


(DEFMETHOD (BASIC-SUPDUP :SET-MORE-P) (IGNORE)
  ())  

(DEFMETHOD (BASIC-SUPDUP :GOBBLE-GREETING) ()
  (SEND-TTY-VARIABLES STREAM SELF NIL OVERPRINT)
  (SEND-FINGER-STRING STREAM)
  ;;Print out the greeting message ITS sends in ASCII.
  (DO ((CH #\NEWLINE (FUNCALL STREAM :TYI)))
      ((OR (NULL CH) (= CH 136)))		;The end is marked with a %TDNOP, NIL is eof
    (AND (< CH 32) (SETQ CH (+ 128 CH)))
    (OR (= CH 138)				;Don't type linefeeds (ITS sends CRLFs).
	(WRITE-CHAR (INT-CHAR CH) SELF))))


;;; Handle %TDORS.  Just tell ITS where the cursor position is, using the
;;;Intelligent Terminal Protocol's ^\ ^P command.

(DEFUN SUPDUP-TDORS (SHEET &AUX VPOS HPOS)
  (DECLARE (:self-flavor basic-supdup))
  (MULTIPLE-VALUE-SETQ (HPOS VPOS)
    (W:SHEET-READ-CURSORPOS SHEET))
  (LOCK-OUTPUT (FUNCALL STREAM :TYO 28)		;^\
	       (FUNCALL STREAM :TYO 16)		;^P
	       (FUNCALL STREAM :TYO (TRUNCATE VPOS TV:LINE-HEIGHT))
	       (FUNCALL STREAM :TYO (TRUNCATE HPOS TV:CHAR-WIDTH)) (FUNCALL STREAM :FORCE-OUTPUT)))

;;; Send the initial information describing the Lisp Machine as an
;;; intelligent terminal.  The TTYOPT word contains the following:
;;; %TOERS+%TOMVB+%TOSAI+%TOOVR+%TOMVU+%TOLWR+%TOFCI+%TOMOR+%TOLID,,%TPCBS+%TPORS+%TPRSC
;;; Furthermore, if SUPDUP-%TOCID is non-NIL, then %TOCID will be on as well.
;;; It is off by default, because the Lispm is so fast at outputting characters
;;; that EMACS is effectively faster for the user without CID capability.
;;; SUPDUPing to SAIL and using SUPDUP-OUTPUT are kludged to bind this to T.



(DEFUN SEND-TTY-VARIABLES (STREAM SHEET LOCAL-EDIT-FLAG OVERPRINT)
  (18BIT-OUT STREAM -6)				;First word LH has minus the count of following wds.
  (18BIT-OUT STREAM 0)
  (18BIT-OUT STREAM 0)				;TCTYP word must be %TNSFW: 0,,7
  (18BIT-OUT STREAM 7)
  (18BIT-OUT STREAM (+ 22938 (IF OVERPRINT
				512
				0)
		       (IF SUPDUP-%TOCID
			   1
			   0)))			;TTYOPT word explained above.
  (18BIT-OUT STREAM 44)
  (18BIT-OUT STREAM 0)				;TCMXV
  (18BIT-OUT STREAM (1- (TRUNCATE (W:SHEET-INSIDE-HEIGHT SHEET) (W:SHEET-LINE-HEIGHT SHEET))))
  (18BIT-OUT STREAM 0)				;TCMXH
  (18BIT-OUT STREAM (1- (TRUNCATE (W:SHEET-INSIDE-WIDTH SHEET) (W:SHEET-CHAR-WIDTH SHEET))))
  (18BIT-OUT STREAM 0)				;TTYROL
  (18BIT-OUT STREAM 0)				;No scrolling
  (18BIT-OUT STREAM
	     (+ (LSH (W:SHEET-LINE-HEIGHT SHEET) 8)	;TTYSMT
		(LSH (W:SHEET-CHAR-WIDTH SHEET) 6) 45))
  (18BIT-OUT STREAM
	     (+ 16384
		(IF LOCAL-EDIT-FLAG
		    (+ 32768 0			;(LSH 5 11.)
						;Don't turn on line-saving.  It slows TECO a lot to think about it. - RMS
		       )
		    0)))
  (FUNCALL STREAM :FORCE-OUTPUT))  



(DEFUN 18BIT-OUT (STREAM N)
  (FUNCALL STREAM :TYO (LDB (BYTE 6 12) N))
  (FUNCALL STREAM :TYO (LDB (BYTE 6 6) N))
  (FUNCALL STREAM :TYO (LDB (BYTE 6 0) N)))  

;;; Send the string to TELSER saying where we are, so that NAME can find it inside
;;; the TELSER and print it.  Boy, what a kludge.


(DEFUN SEND-FINGER-STRING (STREAM)
  (FUNCALL STREAM :TYO 192)			;SUPDUP escape string meaning that the FINGER
  (FUNCALL STREAM :TYO 194)			;identification string follows.
  (FUNCALL STREAM :STRING-OUT SI:LOCAL-FINGER-LOCATION)
  (FUNCALL STREAM :TYO 0)			; End with a 0.
  (FUNCALL STREAM :FORCE-OUTPUT))  

(DEFMETHOD (BASIC-SUPDUP :CONNECT) (&OPTIONAL PATH (NET-WINDOW 1024))
  (MULTIPLE-VALUE-BIND (HOST GATEWAY CONTACT CONTACT-P)
      (PARSE-PATH PATH "SUPDUP" 95.)
    ;; If the host runs the WAITS operating system, it will require char i/d.
    (LET ((SUPDUP-%TOCID
	    (COND
	      ((TYPEP HOST 'Net:HOST)
	       (MEMBER (FUNCALL HOST :SYSTEM-TYPE) '(:MULTICS :WAITS) :TEST #'EQ))
	      ((STRINGP HOST)			;No host name server.  flush this when
	       (MEMBER HOST '("SAIL" "S1-A") :TEST 'EQUALP))	;above thing wins.
	      (T SUPDUP-%TOCID))))
      (SEND SELF :NEW-CONNECTION HOST GATEWAY CONTACT CONTACT-P NET-WINDOW))))  



(DEFMETHOD (BASIC-SUPDUP :AFTER :DISCONNECT) (&OPTIONAL (ABORTP NIL))
  (IGNORE ABORTP)
  (SEND SELF :SET-LABEL (FORMAT () "~A -- not connected" W:NAME)))  


(DEFMETHOD (BASIC-SUPDUP :NET-OUTPUT-TRANSLATED) (CH)
  (UNLESS (CONSP CH)
    (LET ((char ch))				;((CHAR (LDB SI:%%KBD-CHAR CH)))
      (SEND SELF :NET-OUTPUT
	    (LOGIOR (LSH (char-bits CH) 7)
		    (COND
		      ((OR (CHAR-BIT char :control)
			   (CHAR-BIT char :meta)
			   (CHAR-BIT char :super)
			   (CHAR-BIT char :hyper))
		       char)
		      ((= CHAR 27) CHAR)	;(Special case)
		      ((< CHAR 32) (LOGIOR CHAR 2048))
		      ((< CHAR 128) CHAR)
		      (T (AREF SUPDUP-KEYS (- CHAR 128)))))))))  

;;; This sends a character of the ITS 12-bit character set to the network,
;;; using the ITS Intelligent Terminal Protocol to get the extra bits through.


(DEFMETHOD (BASIC-SUPDUP :NET-OUTPUT) (CH &AUX BITS)
  (SETQ BITS (LDB (BYTE 5 7) CH))
  (COND
    ((NOT (ZEROP BITS))
     (LOCK-OUTPUT (FUNCALL STREAM :TYO 28) (FUNCALL STREAM :TYO (LOGIOR 64 BITS))
		  (FUNCALL STREAM :TYO (LOGAND 127 CH))))
    ((= CH 28) (LOCK-OUTPUT (FUNCALL STREAM :TYO 28) (FUNCALL STREAM :TYO CH)))
    (T (LOCK-OUTPUT (FUNCALL STREAM :TYO CH)))))  



(DEFMETHOD (BASIC-SUPDUP :LOGOUT) ()
  (LOCK-OUTPUT (FUNCALL STREAM :TYO 192) (FUNCALL STREAM :TYO 193) (FUNCALL STREAM :FINISH)))

(DEFMETHOD (BASIC-SUPDUP :BUFFERED-TYO) (ch)
  ;(tv:pop-up-format "number and char is ~a ~a "ch (int-char ch))
  (COND
    ((< CH 128) (DO ()
		   ((VECTOR-PUSH CH OUTPUT-BUFFER))
		 (SEND SELF :FORCE-OUTPUT)))
    (T (SEND SELF :FORCE-OUTPUT)
     (OR (>= (SETQ CH (- CH 128)) (ARRAY-TOTAL-SIZE SUPDUP-%TD-DISPATCH))
	(FUNCALL (AREF SUPDUP-%TD-DISPATCH CH) SELF)))))  



;;; SUPDUP



(DEFWINDOW-RESOURCE SUPDUP-WINDOWS () :INITIAL-COPIES 0 :MAKE-WINDOW
   (SUPDUP :TYPEIN-PROCESS NIL :TYPEOUT-PROCESS NIL))  



(DEFVAR *SUPDUP-WINDOWS*)  


(DEFVAR *SUPDUP-DEFAULT-PATH*)  


(DEFVAR *SUPDUP-MODE*)  



(FORWARD-VALUE-CELL '*SUPDUP-WINDOWS* 'SUPDUP-WINDOWS)  


(FORWARD-VALUE-CELL '*SUPDUP-DEFAULT-PATH* 'SUPDUP-DEFAULT-PATH)  


(FORWARD-VALUE-CELL '*SUPDUP-MODE* 'SUPDUP-MODE)  



(DEFVAR SUPDUP-WINDOWS ())  


(DEFVAR SUPDUP-DEFAULT-PATH ())  	;NIL => use associated machine.


(DEFVAR SUPDUP-MODE T)  			;NIL => New window default

      
(DEFUN DISCONNECT-SUPDUP-CONNECTIONS ()    ;03-02-88 DAB for DC
  (DOLIST (WINDOW SUPDUP-WINDOWS)
    (WHEN (SEND WINDOW :CONNECTED-P)
      (SEND WINDOW :DISCONNECT)))) 
      
      
(ADD-INITIALIZATION "Close Supdup connections" '(DISCONNECT-SUPDUP-CONNECTIONS) :LOGOUT)     ;03-02-88 DAB for DC
  

(DEFFLAVOR SUPDUP
	   ()
	   (BASIC-SUPDUP
	    W:INITIALLY-INVISIBLE-MIXIN
	    W:WINDOW)
  (:DEFAULT-INIT-PLIST :SAVE-BITS T)
  (:DOCUMENTATION :COMBINATION))  



(DEFMETHOD (SUPDUP :BEFORE :SELECT) (&REST IGNORE);Move ourselves to the head of the list
  (WITHOUT-INTERRUPTS (SETQ SUPDUP-WINDOWS (DELETE SELF (THE LIST SUPDUP-WINDOWS) :TEST #'EQ))
     (PUSH SELF SUPDUP-WINDOWS)))  



(DEFMETHOD (SUPDUP :BEFORE :DEACTIVATE) (&REST IGNORE)
  (WITHOUT-INTERRUPTS (SETQ SUPDUP-WINDOWS (DELETE SELF (THE LIST SUPDUP-WINDOWS) :TEST #'EQ))))  



(DEFMETHOD (SUPDUP :AFTER :ACTIVATE) (&REST IGNORE)
  (WITHOUT-INTERRUPTS
    (OR (MEMBER SELF SUPDUP-WINDOWS :TEST #'EQ)
	(IF SUPDUP-WINDOWS
	    (RPLACD (LAST SUPDUP-WINDOWS) (CONS SELF ()))
	    (SETQ SUPDUP-WINDOWS (CONS SELF ()))))))  



(DEFVAR JOURNAL-STREAM ())  		;Stream for recording history of events for debugging.



(DEFUN SUPDUP (&OPTIONAL PATH (MODE SUPDUP-MODE))
  "Make a SUPDUP connection to machine specified by PATH.
PATH is a machine name or a string saying how to get to one, such as
<arpanet-gateway><host> or <host>/<contact-name or socket number>
or <gateway><host>/<socket-number>.
If MODE is NIL, SUPDUP runs in (a window substituting for) this window.
Otherwise a separate SUPDUP window is selected."
  (IF MODE
      (SUPDUP-SEPARATE PATH)
      (SUPDUP-BIND PATH)))  



(DEFVAR SUPDUP-FLAVOR 'SUPDUP)  



(DEFUN SUPDUP-SEPARATE (&OPTIONAL PATH &AUX SW)
  "Switch to a non-connected SUPDUP window and connect it to machine PATH.
If PATH is NIL, a connected SUPDUP window will be selected if there is one."
  (COND
    ((AND (NULL PATH) (SETQ SW (FIND-SELECTABLE-SUPDUP T NIL))) (FUNCALL SW :SELECT) NIL)
    (T (SETQ SW (OR (FIND-SELECTABLE-SUPDUP NIL) (W:MAKE-WINDOW SUPDUP-FLAVOR)))
       (FUNCALL SW :SET-CONNECT-TO (OR PATH SUPDUP-DEFAULT-PATH SI:ASSOCIATED-MACHINE))
       (FUNCALL SW :EXPOSE () :CLEAN)		;Don't come up with old garbage
       (FUNCALL SW :SELECT) T)))  



(DEFUN FIND-SELECTABLE-SUPDUP (CONNECTED-P &OPTIONAL (SUP W:MOUSE-SHEET))
  (DOLIST (W SUPDUP-WINDOWS)
    (AND (EQ (FUNCALL W :CONNECTED-P) CONNECTED-P)
       (OR (NULL SUP) (EQ SUP (W:SHEET-SUPERIOR W))) (RETURN W))))  



(DEFUN SUPDUP-BIND (&OPTIONAL (PATH SI:ASSOCIATED-MACHINE)
		    (WINDOW (SEND *TERMINAL-IO* :ALIAS-FOR-SELECTED-WINDOWS)))
  "Enter SUPDUP connection to machine specified by PATH, \"in\" this window.
The I/O is done in a window that overlies the one that is TERMINAL-IO.
Network Q exits SUPDUP and returns from this function."
  (USING-RESOURCE (SUPDUP-WINDOW SUPDUP-WINDOWS)
    (USING-RESOURCE (TP TYPEOUT-PROCESSES)
      (SEND SUPDUP-WINDOW :SETUP WINDOW CURRENT-PROCESS TP)
      (W:WITH-SELECTION-SUBSTITUTE (SUPDUP-WINDOW WINDOW)
	(FUNCALL SUPDUP-WINDOW :CONNECT PATH)
	(CATCH-ERROR-RESTART ((SYSTEM:ABORT ERROR) "Exit SUPDUP.")
	  (FUNCALL SUPDUP-WINDOW :TYPEIN-TOP-LEVEL ()))	;	(SETF (W:SHEET-BIT-ARRAY WINDOW) NIL)
	T))))  



(DEFMETHOD (SUPDUP :SETUP) (WINDOW IN-P OUT-P)
  (LEXPR-SEND SELF :SET-EDGES (MULTIPLE-VALUE-LIST (SEND WINDOW :EDGES)))
  (SETQ ALIAS-WINDOW WINDOW)
  (SETQ TYPEOUT-PROCESS OUT-P
	TYPEIN-PROCESS IN-P)
  (PROCESS-PRESET TYPEOUT-PROCESS SELF :TYPEOUT-TOP-LEVEL))  




;;; GT40 Simulator (used with the DEC simulator on I.T.S. for running SUDS)

;;; Display list array.


(DEFVAR GT40-DISPLAY-LIST (MAKE-ARRAY 8 :TYPE 'ART-Q-LIST))  


;;; This crock maintains a display list for writing, erasing, and moving display objects
;;; consisting of characters, vectors, and points.  This protocol is not documented
;;; anywhere except in the code for DECUUO.

;;; Dispatch table for the GT40 simulator.  These functions take one argument, the pc-ppr.


(DEFVAR GT40-DISPATCH (MAKE-ARRAY 15))  



(REPLACE GT40-DISPATCH
      '(GT40-INSERT-OR-DELETE GT40-INSERT
			      GT40-DELETE	;	     GT40-RESET
						;	     GT40-TURN-ON
						;	     GT40-TURN-OFF
						;	     GT40-COPY
						;	     GT40-MOVE
						;	     GT40-MODE
						;	     GT40-APPEND
						;	     GT40-SUBROUTINIZE
						;	     GT40-UNSUBROUTINIZE
			      SUPDUP-NOTHING))	;most are not used by DECUUO



(DEFVAR GT40-BLINKER ())  


(DEFVAR GT40-CURRENT-ITEM-NUMBER)  


(DEFVAR SUDS-KBD-TABLE				;allows thumb keys to be used
	(LET ((TBL (SI::KBD-MAKE-TI-TABLE)))
	  (DOLIST (L '((126 #\/ #\ #\/ #\) (70 #\\ #\| #\\ #\|) (79 #\[ #\{ #\[ #\{) (15 #\] #\} #\] #\})))
	    (LET ((NCH (FIRST L))
		  (LCH (NTHCDR 1 L)))
	      (DOTIMES (I 5)
		(SETF (AREF TBL I NCH) (CAR LCH))
		(IF (NTHCDR 1 LCH)
		    (SETQ LCH (NTHCDR 1 LCH))))))
	  TBL))  

;; %TDMTN is a crock for simulating GT-40's, used by DECUUO on ITS for Imlacs...



(DEFUN SUPDUP-GT40 (SHEET &AUX (BYTE (- (NVT-NETI) 64)))
  (IF (/= (AREF SUDS-KBD-TABLE 0 126)		;crock for thumb keys, only when
	  (AREF SI::KBD-TI-TABLE 0 126))		;doing GT40 simulation
      (SETQ SI::KBD-TI-TABLE SUDS-KBD-TABLE))
  (OR (< BYTE 0) (FUNCALL (AREF GT40-DISPATCH (LOGAND 15 BYTE)) SHEET)))  

;;; Macros used below to pack characters into words, decode vector formats, etc.

;;; Make a 16-bit "word" from 3 chars in 6-4-6 format


(DEFMACRO GT40-WORD ()
  '(DPB (NVT-NETI) 6 (DPB (NVT-NETI) 388 (DPB (NVT-NETI) 646 0))))  

;;; Get a word count


(DEFMACRO GT40-COUNT ()
  '(LSH (- (GT40-WORD) 5) -1))  

;;; Used in constructing display objects - used only in GT40-INSERT.


(DEFMACRO APUSH (DOB ITEM)
  `(vector-PUSH-EXTEND ,ITEM ,DOB 320))  

;;; Compute the index of the last thing pushed


(DEFMACRO GT40-LAST-INDEX (DOB)
  `(1- (ARRAY-ACTIVE-LENGTH ,DOB)))  

;;; Get the last item pushed onto a display object


(DEFMACRO GT40-LAST-ITEM (DOB)
  `(GLOBAL:AREF ,DOB (GT40-LAST-INDEX ,DOB)))  

;;; Short vector format


(DEFMACRO GT40-SHORT (DOB WORD)
  `(PROGN
     (APUSH ,DOB (* (LDB 454 ,WORD) (IF (LOGTEST 8192 ,WORD)
				      -1
				      1)))
     (APUSH ,DOB (* (LDB 6 ,WORD) (IF (LOGTEST 64 ,WORD)
				    -1
				    1)))
     (APUSH ,DOB (LOGTEST 16384 ,WORD))))  

;;; Long vector format


(DEFMACRO GT40-LONG (DOB WORD1 WORD2)
  `(LET ((WORD2 ,WORD2))
     (APUSH ,DOB (* (LOGAND 1023 ,WORD1) (IF (LOGTEST 8192 ,WORD1)
					  -1
					  1)))
     (APUSH ,DOB (* (LOGAND 1023 WORD2) (IF (LOGTEST 8192 WORD2)
					 -1
					 1)))
     (APUSH ,DOB (LOGTEST 16384 ,WORD1))))  

;;; Coordinate scaling macro


(DEFMACRO GT40-COORD (X)
  `(MAX 0 (TRUNCATE (* 7 ,X) 8)))  

;;; Draw a string.  Note special end of line hackery.  XPOS and YPOS must be symbols.


(DEFMACRO GT40-DRAW-STRING (STRING XPOS YPOS SHEET)
  `(LET ((MAX-Y 488))
     (SEND ,SHEET ':STRING-OUT-EXPLICIT ,STRING (GT40-COORD ,XPOS)
	   (- MAX-Y (GT40-COORD ,YPOS) 9) (W:SHEET-INSIDE-RIGHT ,SHEET) ()
	   (W:SHEET-CURRENT-FONT ,SHEET) W:ALU-XOR)))  

;;; Draw a vector.  XPOS and YPOS must be symbols


(DEFMACRO GT40-DRAW-VECTOR (XPOS YPOS X Y FLAG SHEET)
  `(LET ((MAX-Y 488)
	 (OXPOS ,XPOS)
	 (OYPOS ,YPOS))
     (SETQ ,XPOS (+ ,XPOS ,X)
	   ,YPOS (+ ,YPOS ,Y))
     (IF ,FLAG
	 (W:PREPARE-SHEET (,SHEET)
	   (SYSTEM:%DRAW-LINE (GT40-COORD OXPOS) (- MAX-Y (GT40-COORD OYPOS)) (GT40-COORD ,XPOS)
			      (- MAX-Y (GT40-COORD ,YPOS)) W:ALU-XOR () ,SHEET)))))  

;;; Read a vector out of the display list and draw it


(DEFMACRO GT40-VECTOR (DOB XPOS YPOS SHEET)
  `(LET ((I (GT40-LAST-INDEX ,DOB)))
     (GT40-DRAW-VECTOR ,XPOS ,YPOS (GLOBAL:AREF ,DOB (- I 2)) (GLOBAL:AREF ,DOB (- I 1))
		       (GLOBAL:AREF ,DOB I) ,SHEET)))	;visibility flag

;;; Display list format:  The display list is an ART-Q array of display objects, each of
;;; which is, in turn, an ART-Q array.  The format of display objects is a sequence of
;;; display items.  A display item is either a single string of characters or an in-line
;;; subsequence consisting of a symbol describing the item-type followed by 2 numbers (x,y)
;;; and a visibility flag.  Numbers and flags are repeated until a new symbol is encountered
;;; indicating a type change.

;;; GT40 Command 0 - Insert or delete display items


(DEFUN GT40-INSERT-OR-DELETE (SHEET)
  (CASE (LOGAND 3 (GT40-WORD));only 1 and 2 are recognized for now
    (1 (GT40-INSERT SHEET));insert a new display item
    (2 (GT40-DELETE SHEET (1+ (GT40-COUNT))))))  	;delete n items

;;; GT40 Command 1 - Insert a display item into the display list.


(DEFUN GT40-INSERT (SHEET &AUX (WORD-COUNT (GT40-COUNT)))
  (GT40-DELETE SHEET 1 NIL)			;Delete the item we are about to insert
  (DO ((I 0 (1+ I))				;Loop over words, contructing a display list
       (WORD) (MODE -1)				;Mode is initially undefined.
       (XPOS 0) (YPOS 0) (BLINK-THIS)
       (DOB					;Display OBject
	 (OR (AREF GT40-DISPLAY-LIST GT40-CURRENT-ITEM-NUMBER)	;Already an array or
						;cons an array with leader
	     (SETF (AREF GT40-DISPLAY-LIST GT40-CURRENT-ITEM-NUMBER)
		   (MAKE-ARRAY 128 :LEADER-LIST '(0 NIL))))))	;and install it
      ((>= I WORD-COUNT)
       (IF (= 0 MODE)				; was char mode, display the string
	   (GT40-DRAW-STRING (GT40-LAST-ITEM DOB) XPOS YPOS SHEET))
       (IF BLINK-THIS
	   (STORE-ARRAY-LEADER 'ON DOB 1)))
    (SETQ WORD (GT40-WORD))
    (COND
      ((LOGTEST 32768 WORD)			;If command, only look at blink bit and mode
       (IF (NOT (LOGTEST 16384 WORD))		;ignore words with the 40000 bit on
	   (LET ((NMODE (LDB (BYTE 3 11) WORD))
		 (BLINK-FLAG (AND (LOGTEST 16 WORD) (LOGTEST 8 WORD))))
	     (COND
	       ((NOT (= MODE NMODE))		;get the new datatype mode
		(IF (= 0 MODE)			; was char mode, display the string
		    (GT40-DRAW-STRING (GT40-LAST-ITEM DOB) XPOS YPOS SHEET))
		(SETQ MODE NMODE)
		(APUSH DOB
		       (CASE MODE		;initializings
			     (0 (MAKE-ARRAY 8 :ELEMENT-TYPE 'STRING-CHAR :LEADER-LIST '(0)))
			     (1 'VECTOR)
			     (2 'VECTOR)
			     (3 'POINT)
			     (6 'RPOINT)
			     ((4 5 7) 'UNKNOWN)))))
	     (COND
	       (BLINK-FLAG
		(OR (MEMBER GT40-BLINKER (W:SHEET-BLINKER-LIST SHEET) :TEST #'EQ)
		    (SETQ GT40-BLINKER (W:MAKE-BLINKER SHEET 'GT40-BLINKER)))
		(SETQ BLINK-THIS T))))))
      (T
       (CASE MODE
	     (0
	      (DO ((CHAR (LDB (BYTE 7 0) WORD) (LDB (BYTE 7 8) WORD))	;character mode
		   (STRING (GT40-LAST-ITEM DOB))
		   (I 0 (1+ I)))
		  ((= I 2))
		(OR (= 0 CHAR) (= 15 CHAR) (VECTOR-PUSH-EXTEND CHAR STRING))))
	     (1 (GT40-SHORT DOB WORD)		;short vector
		(GT40-VECTOR DOB XPOS YPOS SHEET))
	     (2 (SETQ I (1+ I))			;long vector
		(GT40-LONG DOB WORD (GT40-WORD)) (GT40-VECTOR DOB XPOS YPOS SHEET))
	     (3 (SETQ I (1+ I))			;point data
		(GT40-LONG DOB WORD (GT40-WORD))
		(LET ((I (GT40-LAST-INDEX DOB)))
		  (SETQ XPOS (AREF DOB (- I 2))
			YPOS (AREF DOB (- I 1)))
		  (GT40-DRAW-VECTOR XPOS YPOS 0 0 (AREF DOB I) SHEET)))
	     (4)				;graphplot x data (not used)
	     (5)				;graphplot y data (not used)
	     (6 (GT40-SHORT DOB WORD)		;relative point data
		(LET ((I (GT40-LAST-INDEX DOB)))
		  (SETQ XPOS (+ XPOS (AREF DOB (- I 2)))
			YPOS (+ YPOS (AREF DOB (- I 1))))
		  (GT40-DRAW-VECTOR XPOS YPOS 0 0 (AREF DOB I) SHEET)))
	     (7)))))				;not used
  (GT40-WORD))					;gobble the checksum

;;; GT40 Command 2 - Delete a display item from the display list


(DEFUN GT40-DELETE (SHEET &OPTIONAL (NITEMS 1) (CHECKSUM-FLAG T))
  (DO ((I 0 (1+ I))
       (DOB)
       (ITEM-NUMBER))
      ((>= I NITEMS))
    (SETQ ITEM-NUMBER (GT40-WORD)
	  GT40-CURRENT-ITEM-NUMBER ITEM-NUMBER	;record item # being hacked
	  
	  DOB (AREF GT40-DISPLAY-LIST ITEM-NUMBER))
    (IF DOB
	(PROGN
	  (OR (EQ 'OFF (ARRAY-LEADER DOB 1))	;don't erase if its already off
	      (GT40-DISPLAY-ITEM DOB SHEET))
	  (FILL DOB NIL)
	  (STORE-ARRAY-LEADER 0 DOB 0)		;zero the fill pointer
	  (STORE-ARRAY-LEADER () DOB 1))))	;blinking is off
  (IF CHECKSUM-FLAG
      (GT40-WORD)))				;gobble the checksum

;;; Display a display item.


(DEFUN GT40-DISPLAY-ITEM (DOB SHEET)
  (DO ((I 0 (1+ I))
       (END (ARRAY-ACTIVE-LENGTH DOB))
       (ITEM)
       (X)
       (Y)
       (FLAG)
       (XPOS 0)
       (YPOS 0))
      ((>= I END))
    (SETQ ITEM (AREF DOB I))
    (COND
      ((STRINGP ITEM) (GT40-DRAW-STRING ITEM XPOS YPOS SHEET))
      ((EQ 'UNKNOWN ITEM))			;ignore
      (T
       (DO ()
	   ((OR (<= (- END I) 3) (SYMBOLP (AREF DOB (1+ I))) (STRINGP (AREF DOB (1+ I)))))
	 (SETQ I (+ 3 I)
	       X (AREF DOB (- I 2))
	       Y (AREF DOB (- I 1))
	       FLAG (AREF DOB I))
	 (CASE ITEM
	       (VECTOR (GT40-DRAW-VECTOR XPOS YPOS X Y FLAG SHEET))
	       (POINT (SETQ XPOS X
			    YPOS Y) (GT40-DRAW-VECTOR XPOS YPOS 0 0 FLAG SHEET))
	       (RPOINT (SETQ XPOS (+ XPOS X)
			     YPOS (+ YPOS Y))
		       (GT40-DRAW-VECTOR XPOS YPOS 0 0 FLAG SHEET))))))))  

;;; GT40 BLINKER



(DEFFLAVOR GT40-BLINKER;Used by GT40-INSERT
	   ()
	   (W:BLINKER))  

;;; Blink a display item


(DEFMETHOD (GT40-BLINKER :BLINK) ()
  (LET-GLOBALLY ((PHASE ()))
    (DO ((ITEM (G-L-P GT40-DISPLAY-LIST) (CDR ITEM))
	 (BLINK-FLAG NIL NIL)
	 (DITEM))
	((NULL ITEM))
      (SETQ DITEM (CAR ITEM))
      (IF DITEM
	  (SETQ BLINK-FLAG (ARRAY-LEADER DITEM 1)))
      (IF (MEMBER BLINK-FLAG '(ON OFF) :TEST #'EQ)
	  (PROGN
	    (GT40-DISPLAY-ITEM DITEM W:SHEET)
	    (STORE-ARRAY-LEADER (CASE BLINK-FLAG
				      (ON 'OFF)
				      (OFF 'ON))
				DITEM 1))))))  



(DEFMETHOD (GT40-BLINKER :SIZE) ()
  (BLOCK ()
    (RETURN (VALUES (W:SHEET-INSIDE-WIDTH W:SHEET) (W:SHEET-INSIDE-HEIGHT W:SHEET)))
    ()))  



;;; ARDS simulator (for compatibility with PTV's)

;;; Todo: these variables should be instance variables
;;;       scaling and offset doesn't work right in this version...
;;;       SHOULD SEND LINE DRAWING MESSAGES RATHER THAN CALLING %DRAW-LINE



(DEFVAR ARDS-XPOS 0)  				;current pos in ARDS coordinates


(DEFVAR ARDS-YPOS 0)  


(DEFVAR ARDS-SCALE 1.0)  


(DEFVAR ARDS-SCR-XPOS 0)  			;current pos in screen coordinates


(DEFVAR ARDS-SCR-YPOS 0)  

;;; Setup scaling and offsets, then loop until exit condition


(DEFMACRO ARDS-LOOP (&REST BODY)
  `(LET* ((ARDS-MAX-X (+ W:X-OFFSET W:WIDTH))
	  (ARDS-MAX-Y (+ W:Y-OFFSET W:HEIGHT))
	  (ARDS-X-OFFSET W:X-OFFSET)
	  (ARDS-Y-OFFSET W:Y-OFFSET)
	  (ARDS-SCR-SCALE (* ARDS-SCALE (GLOBAL:/ (MIN W:WIDTH W:HEIGHT) 531.0)))
	  (ARDS-CENTER-OFFSET
	    (TRUNCATE (1+ (- (MAX W:WIDTH W:HEIGHT) (MIN W:WIDTH W:HEIGHT))) 2))
	  (ARDS-FLAG NIL))
     (IF (< W:WIDTH W:HEIGHT)
	 (SETQ ARDS-MAX-Y (- ARDS-MAX-Y ARDS-CENTER-OFFSET))
	 (SETQ ARDS-X-OFFSET (+ ARDS-X-OFFSET ARDS-CENTER-OFFSET)))
     (*CATCH 'ARDS-RETURN (DO ()
			      (NIL)
			    ,@BODY))))  

;;; Convert -512./511. to 0/1023. and scale if the user wants it.


(DEFMACRO ARDS-COORD (X)
  `(MAX 1 (FLOOR (+ 0.5 (* ARDS-SCR-SCALE (+ 330 ,X))))))  

;;; Get a character and punt out of graphics mode if it is a control char or %TD code


(DEFMACRO ARDS-GET ()
  '(LET ((X (NVT-NETI)))
     (IF (OR (< X 64) (> X 127))
       (THROW 'ARDS-RETURN
	      (PROGN
	       (FUNCALL STREAM ':UNTYI X)
	       (W:SHEET-SET-CURSORPOS SHEET ARDS-SCR-XPOS (- ARDS-SCR-YPOS 9)))))
     X))  

;;; Unpack long and short format coordinates


(DEFMACRO ARDS-LONG (F)
  `(LET ((A (ARDS-GET))
	 (B (ARDS-GET)))
     ,(IF F
	'(SETQ ARDS-FLAG (NOT (LOGTEST B 32))))
     (* (IF (LOGTEST A 1)
	  -1
	  1)
	(LOGIOR (LSH (LOGAND 63 A) -1) (LSH (LOGAND 31 B) 5)))))  



(DEFMACRO ARDS-SHORT ()
  '(LET ((A (ARDS-GET)))
     (SETQ ARDS-FLAG T)
     (* (IF (LOGTEST A 1)
	  -1
	  1)
	(LSH (LOGAND 63 A) -1))))  

;;; Draw a vector


(DEFMACRO ARDS-VECTOR (DX DY)
  `(LET ((X0 ARDS-XPOS)
	 (Y0 ARDS-YPOS))
     (SETQ ARDS-XPOS (+ ARDS-XPOS ,DX)
	   ARDS-YPOS (+ ARDS-YPOS ,DY)
	   ARDS-SCR-XPOS (MIN ARDS-MAX-X (+ ARDS-X-OFFSET (ARDS-COORD ARDS-XPOS)))
	   ARDS-SCR-YPOS (MAX ARDS-Y-OFFSET (- ARDS-MAX-Y (ARDS-COORD ARDS-YPOS))))
     (IF ARDS-FLAG
	 (W:PREPARE-SHEET (SHEET)
	   (SYSTEM:%DRAW-LINE (MIN ARDS-MAX-X (+ ARDS-X-OFFSET (ARDS-COORD X0)))
			      (MAX ARDS-Y-OFFSET (- ARDS-MAX-Y (ARDS-COORD Y0))) ARDS-SCR-XPOS
			      ARDS-SCR-YPOS W:ALU-IOR T SHEET)))))  



  (DEFUN SUPDUP-ARDS-SET (SHEET)
    (DECLARE (:self-flavor BASIC-SUPDUP))
    (ARDS-LOOP (SETQ ARDS-XPOS (ARDS-LONG T)
		     ARDS-YPOS (ARDS-LONG NIL)) (ARDS-VECTOR 0 0)))	;for plotting points

  (DEFUN SUPDUP-ARDS-LONG (SHEET)
    (DECLARE (:self-flavor BASIC-SUPDUP))
    (ARDS-LOOP (ARDS-VECTOR (ARDS-LONG T) (ARDS-LONG NIL))))

  (DEFUN SUPDUP-ARDS-SHORT (SHEET)
    (DECLARE (:self-flavor BASIC-SUPDUP))
    (ARDS-LOOP (ARDS-VECTOR (ARDS-SHORT) (ARDS-SHORT))))  





(COMPILE-FLAVOR-METHODS SUPDUP GT40-BLINKER)  
