;;; -*- Mode:Common-Lisp; Package:MACTOOLBOX; Base:10 -*-


(DEFUN build-trap-table (&optional
	 (outfil "lm:hd:microexp:macsys:toolbox-interface:traps.r")
		    &aux runner trap params index offset count)

  (COND  ((BOUNDP '*trap-table*)
	  (WITH-OPEN-FILE (outstrm outfil
				   :direction :output
				   :if-exists :overwrite
				   :if-does-not-exist :create)
;  Write out header stuff.

	    (FORMAT outstrm
"/* TBI CATAGORY B                                                        */
/*                                                                       */
/*                      RESTRICTED RIGHTS LEGEND                         */
/*                                                                       */
/* Use, duplication, or disclosure by the Government is subject to       */
/* restrictions as set forth in subdivision (b)(3)(ii) of the Rights in  */
/* Technical Data and Computer Software clause at 52.227-7013.           */
/*                                                                       */
/*                    EXPERTELLIGENCE INCORPORATED.                      */
/*                         5638 HOLLISTER AVE                            */
/*                         GOLETA, CA 93117                              */
/*                                                                       */
/*                                                                       */
/*  Copyright (C) 1988, 1987, 1986, ExperTelligence Incorporated.        */
/*  All rights reserved.                                                 */
/*                                                                       */


/*************************************************************************/
/*         Modification History                                          */
/*************************************************************************/
/* 8/22/88 rwr GetKeys is HndlPtrType. Put PasCalls in seq. order		 */
/*             added NewCDialog, UpdtDialog, HSetRBit, HClrRBit 		 */
/* 8/23/88 rwr added Palette, color + misc at end of file			     */
/* 8/25/88 rwr changed FontMetrics & FMSwapFont to HndlPtrType 			 */
/* 8/27/88 rwr changed rtype of !GetWVariant to rinttype                 */
/* 8/28/88 rwr OpenRFPerm had wrong trap number, now it doesn't          */
/*             HSetRBit & HClrRBit have correct trap number now          */
/*             Color2Index returns longint now, GetSubTable takes a hndl */
/* 8/30/88 rwr Fixed !SndNewChannel, added DelSearch, DelComp,           */
/*             SaveEntries, RestoreEntries, SetEntries, GetCTSeed        */
/* 9/1/88  rwr CMY, HSV, HSL colors changed to HndlPtrType               */
/* 9/13/88 rwr SetOSTrapAddress, GetOSTrapAddress, SetToolTrapAddress now*/
/*             have correct trapnumbers									 */


#define		PADDING		-1

#define		null		0
#define		selector	1		        /* set bit 0 */
#define		macPlus		2		        /* set bit 1 */
#define		mac2		4		        /* set bit 2 */
#define         PasCall         8			/* set bit 3 */
/*                              16		       ;unused */
#define		SaveResError	32		        /* set bit 5 */
#define		SaveQdError	64		        /* set bit 6 */

/*******************************************************************************************/
/*          Specific to ExperInterface for MicroExplorer 6/28/88  						   */
/*******************************************************************************************/



/* -------------------------------------------------------------------------------- */
/*		    register based traps pre processing routine								*/
/* -------------------------------------------------------------------------------- */

#define		preResStart	0

#define		PtrInA0		preResStart+1	  	/* pointer in a0	*/
#define		HndlInA0	PtrIna0+1		  	/* mac or lisp handle in a0	*/
#define		macHndlInA0	HndlIna0+1	  		/* mac handle in a0	*/
#define		ioParamInA0	macHndlIna0+1	  	/* ioParamBlock ptr in a0	*/
	
#define		PtrInA1		ioParamInA0+1	 	/* pointer in a1	*/
#define		HndlInA1	PtrInA1+1

#define		IntInD0		HndlInA1+1	  		/* Integer (fixNum) in d0	*/
#define		IntInMSW_D0	IntInD0+1		  	/* put int in msw of d0	*/
#define		IntInA0		IntInMSW_D0+1	  	/* Integer (fixNum) in d0	*/

#define		LongInD0	IntInA0+1		  	/* LongInt (float of fix) in d0	*/
#define		LongInA0	LongInD0+1	  		/* LongInt (float of fix) in a0	*/
/* #define		ioParamSel	LongInA0+1	*/  		/* ioParamBlock ptr in a0 /selector in d0	*/

/* #define		preRegEnd	ioParamSel */
#define		preRegEnd	LongInA0


/* --------------------------------------------------------------------------------	*/
/*		    stack based traps pre processing routine								*/
/* --------------------------------------------------------------------------------	*/

#define		preStackStart	preRegEnd

#define		inttype		preStackStart+1
#define		lngInttype	inttype+1
#define		Booltype	lngInttype+1
#define		ptrtype		Booltype+1

#define		hndltype	ptrtype+1
#define		Pttype		hndltype+1
#define		Recttype	Pttype+1
#define		Rgntype		Recttype+1

#define		Polytype	Rgntype+1
#define		Picttype	Polytype+1
#define		BitMaptype	Picttype+1
#define		EvtRectype	BitMaptype+1

#define		chartype	EvtRectype+1
#define		Strtype		chartype+1
#define		ResTypetype	Strtype+1
#define		paramtype	ResTypetype+1

#define		ioFileParamtype	paramtype+1
#define		ioVolParamtype	ioFileParamtype+1
#define		ioControltype	ioVolParamtype+1
#define		Patterntype		ioControltype+1
#define		FontInfotype	Patterntype+1

#define		Cursortype		FontInfotype+1
#define		PenStatetype	Cursortype+1
#define		QDProcstype		PenStatetype+1
#define		GrafPorttype	QDProcstype+1

#define		WindowRectype	GrafPorttype+1
#define		CntlHndltype	WindowRectype+1
#define		menuHndletype	CntlHndltype+1
#define		MenuInfotype	menuHndletype+1
#define		TERectype		MenuInfotype+1

#define		DialogRectype	TERectype+1
#define		DITLType		DialogRectype+1
#define		ScrapStufftype	DITLType+1
#define		Int64Bittype	ScrapStufftype+1
#define		vSFReplytype	Int64Bittype+1
#define     SFReplytype		vSFReplytype          /* These are the same */

#define		SFTypeListtype	SFReplytype+1
#define		AppFiletype		SFTypeListtype+1
#define		TPrPorttype		AppFiletype+1
#define		THPrinttype		TPrPorttype+1

#define		TPrStatustype	THPrinttype+1
#define		SysPararmtype	TPrStatustype+1
#define		DteTimetype		SysPararmtype+1

#define		Vinttype		DteTimetype+1
#define		VlngInttype		Vinttype+1
#define		Vptrtype		VlngInttype+1
#define		Vhndltype		Vptrtype+1

#define		VPttype			Vhndltype+1
#define		VRecttype		VPttype+1
#define		VStrtype		VRecttype+1
#define		VChartype		VStrtype+1

#define		VRestype		VChartype+1
#define		VWindPtrtype	VRestype+1
#define		VCntlHndltype	VWindPtrtype+1
#define		VInt64Bittype	VCntlHndltype+1
#define		VPenStatetype	VInt64Bittype+1
#define		VBitMaptype		VPenStatetype+1
#define		VFontInfoType	VBitMaptype+1
#define		VEvtRecType		VFontInfoType+1

#define		ListRecType		VEvtRecType+1	/* New 2/20/87 jmc */

#define		ioParamSel		ListRecType+1
#define		Colortype		ioParamSel+1
#define     HndlPtrtype		colortype+1
#define		preStckEnd		HndlPtrtype

/* -------------------------------------------------------------------------------- */
/*		    register based traps post processing routine							*/
/* -------------------------------------------------------------------------------- */

#define		postRegStart	0

#define		rPtrInA0		postRegStart+1	  	/* returned pointer in a0	*/
#define		rMacHndlInA0	rPtrInA0+1	  		/* returned mac handle in a0	*/
#define		rIntInD0		rMacHndlInA0+1	  	/* returned integer in d0	*/
#define		rLongInD0		rIntInD0+1	  		/* returned longint in d0	*/
#define		rResultCodeInD0	rLongInD0+1	  		/* returned resultCode in d0	*/
#define		rLongA0_LongD0	rResultCodeInD0+1	/* multi return (CHECK THIS ONE!!)	*/
#define		rPtrA0_resCD0	rLongA0_LongD0+1
#define		rHndlA0_resCD0	rPtrA0_resCD0+1

#define		postRegEnd		rHndlA0_resCD0

/* --------------------------------------------------------------------------------	*/
/*		    stack based traps post processing routine								*/
/* --------------------------------------------------------------------------------	*/

#define		postStackStart	postRegEnd

#define		rDontCare		postStackStart+1	 /* NEW!!!*/
#define		rinttype		rDontCare+1
#define		rMVInttype		rinttype+1
#define		rlngInttype		rMVInttype+1
#define		rBooltype		rlngInttype+1
#define		rptrtype		rBooltype+1
#define		rhndltype		rptrtype+1
#define		rRgntype		rhndltype+1
#define		rPolytype		rRgntype+1
#define		rPicttype		rPolytype+1
#define		rwindPtrtype	rPicttype+1
#define		rDialogRectype	rwindPtrtype+1
#define		rCntlHndltype	rDialogRectype+1

#define		postStckEnd		rCntlHndltype

type 'TRAP'
	{
	longint = $$CountOf(TrapResArray);
	array TrapResArray
		{
		unsigned hex integer;				/* the trap number */
		unsigned hex byte;					/* trap flags */
		unsigned hex byte;					/* selector value */
		unsigned hex byte;					/* return arg type */
		integer = $$CountOf(argInfoArray);	/* number of args */
		array argInfoArray	
			{
			unsigned hex byte;
			};
		pstring;							/* trap name */
		align long;							/* 2/27/87 word to long */
		};
	};
	
")

;  End of the header stuff.

	    (SETQ offset 4)
	    (SETQ index (LIST 4))
	    (FORMAT outstrm "resource 'TRAP' (1)~%   {~%      {~%")
	    (SETQ runner *trap-table*)
	    (while runner
		   (SETQ trap (CAR runner))
		   (SETQ runner (CDR runner))
		   (COND (trap
			  (FORMAT outstrm "        0x~x,~%" (FIRST trap))
			  (COND ((NULL (SECOND trap))
				 (FORMAT outstrm "           null,~%"))
				(t
				 (FORMAT outstrm "           ~a,~%"
					 (SECOND trap))))
			  (COND ((NULL (THIRD trap))
				 (FORMAT outstrm "           null,~%"))
				(t
				 (FORMAT outstrm "           ~a,~%"
					 (THIRD trap))))
			  (FORMAT outstrm "           ~a,~%" (FOURTH trap))
			  (FORMAT outstrm "           {~%")
			  (SETQ params (FIFTH trap))
			  (while (CONSP params)
				 (FORMAT outstrm "              ~a" (CAR params))
				 (SETQ params (CDR params))
				 (IF (CONSP params)
				     (FORMAT outstrm ";"))
				 (FORMAT outstrm "~%"))
			  (FORMAT outstrm "           },~%")
			  (if runner
			      (FORMAT outstrm "           ~s;~%~%" (SIXTH trap))
			      (FORMAT outstrm "           ~s~%~%" (SIXTH trap)))
			  (SETQ offset (+ offset
					  (* 4
					     (CEILING (/ (+ 8
							    (LENGTH (FIFTH trap))
							    (LENGTH (SIXTH trap)))
							 4)))))
			  (SETQ index (CONS offset index))
			  )
			 )
		   )
	    (FORMAT outstrm "      };~%   };~%~%~%")
	    (SETQ index (REVERSE index))
	    (FORMAT outstrm "data 'TRAP' (2, \"index\") {~%   $\"")
	    (SETQ count 1)
	    (while (CONSP index)
		   (FORMAT outstrm "~8,48x" (CAR index))
		   (SETQ index (CDR index))
		   (COND ((ZEROP (MOD count 4))
			  (COND ((CONSP index)
				 (FORMAT outstrm "\"~%   $\""))))
			 ((CONSP index)
			  (FORMAT outstrm " ")))
		   (SETQ count (1+ count)))
	    (FORMAT outstrm "\"~%")
	    (FORMAT outstrm "};")
	    )
	  )
	 (t
	  (FORMAT t "~%*trap-table* is not bound.~%"))
	 )
  )
