;;; -*- Mode:Common-Lisp; Package:USER; Base:10 -*-

;/*                                                                       */
;/*                      RESTRICTED RIGHTS LEGEND                         */
;/*                                                                       */
;/* Use, duplication, or disclosure by the Government is subject to       */
;/* restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  */
;/* Technical Data and Computer Software clause at 52.227-7013.           */
;/*                                                                       */
;/*                    TEXAS INSTRUMENTS INCORPORATED.                    */
;/*                            P.O. BOX 2909                              */
;/*                         AUSTIN, TEXAS 78769                           */
;/*                              MS 2151                                  */
;/*                                                                       */
;/*  Copyright (C) 1987-1989 Texas Instruments Incorporated.              */
;/*  All rights reserved.                                                 */
                     

;;; MODIFICATIONS
;;; 01/18/89 Mac  Added TOOLBOX-INIT file as last one to be loaded.  It
;;;               contains initialization forms which only need to be done at
;;;               final load time and are not needed by loads done for
;;;               compile-time dependencies.
;;; 01/18/89 Mac  Added TOOLBOX-CONSTANTS, ARGLIST-ALIST, DOCUMENTATION-ALIST,
;;;               and VALUES-ALIST to the DEFSYSTEM

(defpackage MacToolbox
  (:nicknames TB)
  (:use LISP TICL MICRONET))

(defsystem MacToolbox
  (:nicknames "toolbox-interface")
  (:pathname-default "sys:toolbox-interface;")
  (:warnings-pathname-default "sys:cwarns;MacToolbox.lisp")
  (:patchable "sys:patch.toolbox-interface;")
  (:module macros ("exper-tb-macros" "exper-tb-datatypes"))
  (:module table    "exper-toolbox-table")
  (:module errors      "toolbox-errors")
  (:module oserr       "oserr-alist")
  (:module source0     "toolbox-defs")
  (:module interface   "exper-toolbox-interface")
  (:module doc        ("arglist-alist" "documentation-alist" "values-alist"))
  (:module library     "generic-toolbox-library")
  (:module tb-const    "toolbox-constants")
  (:module application "mac-applications")
  (:module accessors   "exper-toolbox-accessors")
  (:module init        "toolbox-init")
  (:module last-file   "last-file")

  (:skip :compile-load macros)
  
  (:readfile          oserr)			; reading a defvar is just as good as compiling it
  (:skip :readfile    table)			; reading is as good as compiling, but we don't need
						;   to load unless we are compiling
  (:compile-load      errors)
  (:compile-load      tb-const)
  (:compile-load-init source0
		      (macros)
		      (:fasload tb-const macros)
		      nil)
  (:compile-load-init interface
		      (macros errors)
		      ((:readfile table)
		       (:fasload tb-const source0 macros errors))
		      (:fasload tb-const source0))
  (:readfile          doc
		      (:fasload interface))
  (:compile-load-init library
		      (macros)
		      (:fasload tb-const macros source0 interface)
		      (:fasload tb-const source0 interface))
  (:compile-load-init application
		      (macros errors)
		      (:fasload macros errors interface library)
		      (:fasload interface library))
  (:compile-load      accessors
		      (:fasload library)
		      (:fasload library))
  (:readfile          init
		      ((:readfile doc)
		       (:fasload interface library application accessors)))
  (:readfile          last-file
		      ((:readfile init doc)
		       (:fasload interface library application accessors))))

(defsystem MacToolbox-defs
  (:nicknames "toolbox-defs")
  (:pathname-default "sys:toolbox-interface;")
  (:warnings-pathname-default "sys:cwarns;MacToolbox.lisp")
  (:module macros ("exper-tb-macros" "exper-tb-datatypes"))
  (:module table    "exper-toolbox-table")
  (:module errors      "toolbox-errors")
  (:module oserr       "oserr-alist")
  (:module source0     "toolbox-defs")
  (:module interface   "exper-toolbox-interface")
  (:module doc        ("arglist-alist" "documentation-alist" "values-alist"))
  (:module library     "generic-toolbox-library")
  (:module tb-const    "toolbox-constants")
  (:module application "mac-applications")
  (:module accessors   "exper-toolbox-accessors")
;  (:module init        "toolbox-init")
  (:module last-file   "last-file")

  (:skip :compile-load macros)
  
  (:readfile          oserr)			; reading a defvar is just as good as compiling it
  (:skip :readfile    table)			; reading is as good as compiling, but we don't need
						;   to load unless we are compiling
  (:compile-load      errors)
  (:compile-load      tb-const)
  (:compile-load-init source0
		      (macros)
		      (:fasload tb-const macros)
		      nil)
  (:compile-load-init interface
		      (macros errors)
		      ((:readfile table)
		       (:fasload tb-const source0 macros errors))
		      (:fasload tb-const source0))
  (:readfile          doc
		      (:fasload interface))
  (:compile-load-init library
		      (macros)
		      (:fasload tb-const macros source0 interface)
		      (:fasload tb-const source0 interface))
  (:compile-load-init application
		      (macros errors)
		      (:fasload macros errors interface library)
		      (:fasload interface library))
  (:compile-load      accessors
		      (:fasload library)
		      (:fasload library))
;  (:readfile          init
;		      ((:readfile doc)
;		       (:fasload interface library application accessors)))

  (:readfile          last-file
		      ((:readfile doc)
		       (:fasload interface library application accessors))))



;; To create the Macintosh trap resource source file:
;; 1) make  build-trap-table.
;; 2) run the function build-trap-table.
;;

(defsystem build-trap-table
  (:pathname-default "sys:toolbox-interface;")
  (:warnings-pathname-default "sys:cwarns;MacToolbox-table.lisp")
  (:module macros (mactoolbox macros))
  (:module convert-table "convert-table")

  (:compile-load macros)
  (:compile-load-init convert-table
		      (macros)
		      (:fasload macros)
		      (:fasload macros)))