;;; -*- Mode:Common-Lisp; Package:MACTOOLBOX; Base:10 -*-


;;; For the puposes of the Texas Instruments/ ExperTelligence Contract, this
;;; software is in TBI Category "A".

;/*                                                                       */
;/*                      RESTRICTED RIGHTS LEGEND                         */
;/*                                                                       */
;/* Use, duplication, or disclosure by the Government is subject to       */
;/* restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  */
;/* Technical Data and Computer Software clause at 52.227-7013.           */
;/*                                                                       */
;/*                    TEXAS INSTRUMENTS INCORPORATED.                    */
;/*                            P.O. BOX 2909                              */
;/*                         AUSTIN, TEXAS 78769                           */
;/*                              MS 2151                                  */
;/*                                                                       */
;/*  Copyright (C) 1987-1989 Texas Instruments Incorporated.              */
;/*  All rights reserved.                                                 */
;/*                                                                       */

;;; ************************************************************************
;;; Modification History:
;;;   08/13/88 DWB  CHanged BitMap type to allow passing pointers
;;;   08/14/88 DWB  Changed ParamBlockREcord to ParamBlockRec
;;;   08/17/88 DWB  Changed VIntType to signed 16, instead of signed 25
;;;   08/17/88 DWB  Changed VLngInt to Unsigned from signed 25
;;;   08/17/88 DWB  Changed LinginD0 to positive integer instead of fixnum
;;;   08/17/88 DWB  Changed :fixnum to 'Fixnum for compiler warning
;;;   08/17/88 DWB  CHanged RRESULTCODEinD0 to signed 16
;;;   08/17/88 DWB  Changed VCharType to ldb 8 of data
;;;   08/17/88 DWB  Changed VResType & VStringType because of compiler over opt.
;;;   08/18/88 RR   Changed ListRec macro to do check-toolbox-type
;;;   08/22/88 RR   Changed RhndlA0_ResCD0 to extract :handle (not :pointer)
;;;   08/29/88 DWB&RR Slightly changed VIntType & VLngIntType
;;;   08/30/88 RWR  Changed Slightly VPtrType
;;;   08/30/88 DWB  Changed LongInD0 to >= 0, from plusp
;;;   09/14/88 RWR  Changed %p-store-contents to setf
;;;   01/16/89 MAC  Changed several datatype macros to return signed rather than
;;;                 unsigned values:  VRectType, VPtType, and rLngIntType.
;;; ************************************************************************

(DEFMACRO check-type-1 (arg-name type &optional type-string)
  (DECLARE (IGNORE type-string))
  `(progn (check-type ,arg-name ,type) t))

(defmacro check-toolbox-type (arg-name typ)
  `(or (and (typep ,arg-name 'Toolboxobject)
	    (typep ,arg-name ,typ))
       (and (not (typep ,arg-name 'Toolboxobject))
	    (typep ,arg-name 'mac-handle))))

(defmacro check-toolbox-type-or-pointer (arg-name typ)
  `(or (and (typep ,arg-name 'Toolboxobject)
	    (typep ,arg-name ,typ))
       (and (not (typep ,arg-name 'Toolboxobject))
	    (typep ,arg-name 'mac-pointer))))

;BITMAPTYPE !CopyBits, !CopyMask, !SetPortBits
;remember, it is ok to pass a PixMap on 256K ROM

(defmacro BitMapType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((or (check-toolbox-type ,arg-name 'BitMap)
		(check-toolbox-type ,arg-name 'PixMap))
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   ;;This is for passing pointers into Grafports
	   ((typep ,arg-name 'mac-pointer)
	    (setf (parm-32b ,acb ,index) 0)
	    (setf (parm-32b ,acb (1+ ,index)) (send ,arg-name :pointer)))
	   (t
	    (ferror nil "~s cannot be coerced into a Handle." ,arg-name)))))

(setf (get 'BitMapType :sizeof) 8)
(setf (get 'BitMapType :pretty-arg-name) "BitMap")

;BOOLTYPE

(defmacro BoolType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(setf (parm-32b ,acb ,index)
	  (if ,arg-name 1 0))))

(setf (get 'BoolType :sizeof) 4)
(setf (get 'BoolType :pretty-arg-name) "Boolean")

;CHARTYPE

(defmacro CharType (arg-name acb acb-index)
  (let ((index  acb-index))
    `(cond ((check-type-1 ,arg-name character)
	    (setf (parm-8b ,acb ,index) (char-code ,arg-name))))))

(setf (get 'CharType :sizeof) 4)
(setf (get 'CharType :pretty-arg-name) "Character")

;CNTLHNDLTYPE

(defmacro CntlHndlType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'ControlRecord) 
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s is not a ControlRecord." ,arg-name)))))

(setf (get 'CntlHndlType :sizeof) 4)
(setf (get 'CntlHndlType :pretty-arg-name) "ControlRecord-inst")

;COLORTYPE
(defmacro ColorType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'RGBColor)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s cannot be coerced into a RGB Color." ,arg-name)))))

(setf (get 'ColorType :sizeof) 4)
(setf (get 'ColorType :pretty-arg-name) "RGBColor-inst")

;CURSORTYPE

(defmacro CursorType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'Cursor)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s cannot be coerced into a Handle." ,arg-name)))))

(setf (get 'CursorType :sizeof) 4)
(setf (get 'CursorType :pretty-arg-name) "Cursor-inst")

;DIALOGRECTYPE

(defmacro DialogRecType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name DialogRecord)
	    (setf (parm-32b ,acb ,index)
		  (send ,arg-name :pointer))))))

(setf (get 'DialogRecType :sizeof) 4)
(setf (get 'DialogRecType :pretty-arg-name) "DialogRecord-inst")

;EVTRECTYPE

(defmacro EvtRecType (arg-name acb acb-index)
  (let ((index (floor acb-index 2)))
    `(cond ((check-type-1 ,arg-name EventRecord)
	(setf (parm-16b ,acb (+ ,index 0)) (send ,arg-name :what))
	(setf (parm-16b ,acb (+ ,index 2))
	      (ldb (byte 16 0) (send ,arg-name :message)))
	(setf (parm-16b ,acb (+ ,index 1))
	      (ldb (byte 16 16) (send ,arg-name :message)))
	(setf (parm-16b ,acb (+ ,index 4))
	      (ldb (byte 16 0) (send ,arg-name :when)))
	(setf (parm-16b ,acb (+ ,index 3))
	      (ldb (byte 16 16) (send ,arg-name :when)))
	(setf (parm-16b ,acb (+ ,index 5)) (send ,arg-name :v))
	(setf (parm-16b ,acb (+ ,index 6)) (send ,arg-name :h))
	(setf (parm-16b ,acb (+ ,index 7)) (send ,arg-name :modifiers)))))
  )

(setf (get 'EvtRecType :sizeof) 16)
(setf (get 'EvtRecType :pretty-arg-name) "EventRecord-inst")

;GRAFPORTTYPE

(defmacro GrafPortType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type-or-pointer ,arg-name 'aGrafPort)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :pointer)))
	   (t
	    (ferror nil "~s is not a GrafPort." ,arg-name)))))

(setf (get 'GrafPortType :sizeof) 4)
(setf (get 'GrafPortType :pretty-arg-name) "Grafport-inst")

;HNDLINA0

(defmacro HndlinA0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name mac-handle)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))))
  )

(setf (get 'HndlinA0 :sizeof) 4)
(setf (get 'HndlinA0 :pretty-arg-name) "Mac-Handle-inst")

;HNDLINA1

(defmacro HndlinA1 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name mac-handle)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle))))
    ))

(setf (get 'HndlinA1 :sizeof) 4)
(setf (get 'HndlinA1 :pretty-arg-name) "Mac-Handle-inst")

;HNDLTYPE

(defmacro HndlType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name mac-handle)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   )
    ))

(setf (get 'HndlType :sizeof) 4)
(setf (get 'HndlType :pretty-arg-name) "Mac-Handle-inst")

;HNDLPTRTYPE

(defmacro HndlPtrType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name mac-handle)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   ))
  )

(setf (get 'HndlPtrType :sizeof) 4)
(setf (get 'HndlPtrType :pretty-arg-name) "Mac-Handle-to-deref")

;INTINA0

(defmacro INTINA0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name fixnum)
	    (setf (parm-32b ,acb ,index) ,arg-name))
	   )
    ))

(setf (get 'INTINA0 :sizeof) 4)
(setf (get 'intina0 :pretty-arg-name) "Integer")

;INTIND0

(defmacro INTIND0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name fixnum)
	    (setf (parm-32b ,acb ,index) ,arg-name))
	   )
    ))

(setf (get 'INTIND0 :sizeof) 4)
(setf (get 'intind0 :pretty-arg-name) "Integer")

;INTINMSW_D0

(defmacro INTINMSW_D0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name fixnum)
	    (setf (parm-32b ,acb ,index) ,arg-name)))
    ))

(setf (get 'INTINMSW_D0 :sizeof) 4)
(setf (get 'intinMSW_D0 :pretty-arg-name) "Integer")

;INTTYPE - should we check for larger than 16 bit integers?
;Sign of number is coerced in TBServer by checking bit 24.
;Numbers > 32767 remain positive but are limited to 15 bits.
;Numbers < -32768 will have undefined results.

(defmacro IntType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name fixnum)
	    (setf (parm-32b ,acb ,index) ,arg-name)))))


(setf (get 'IntType :sizeof) 4)
(setf (get 'intType :pretty-arg-name) "Integer")

;ioParamInA0

(defmacro ioParamInA0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'ParamBlockRec)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s cannot be coerced into a Handle." ,arg-name)))))

(setf (get 'ioParamInA0 :sizeof) 4)
(setf (get 'ioParamInA0 :pretty-arg-name) "ParamBlockRec-inst")

;IOPARAMSEL
(defmacro ioParamSel (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'ParamBlockRec)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s cannot be coerced into a Handle." ,arg-name)))))

(setf (get 'ioParamSel :sizeof) 4)
(setf (get 'ioParamSel :pretty-arg-name) "ParamBlockRec-inst")

;LISTRECTYPE

(defmacro ListRecType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'ListRec)
	    (setf (parm-32b ,acb ,index)
		  (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s is not a ListRec." ,arg-name)))))

(setf (get 'ListRecType :sizeof) 4)
(setf (get 'ListRecType :pretty-arg-name) "ListRec-inst")

;LNGINTTYPE

(defmacro LngIntType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name integer)
	    (setf (parm-32b ,acb ,index) ,arg-name)))
    ))

(setf (get 'LngIntType :sizeof) 4)
(setf (get 'LngIntType :pretty-arg-name) "LongInteger")

;LONGIND0

(defmacro LONGIND0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((and (typep ,arg-name 'integer)(>= ,arg-name 0))
	    (setf (parm-32b ,acb ,index) ,arg-name))
	   (t
	    (ferror nil "~s is not a positive integer." ,arg-name)))))

(setf (get 'LONGIND0 :sizeof) 4)
(setf (get 'LONGIND0 :pretty-arg-name) "+LongInteger")

;LONGINA0
;Used only by !Delay
(defmacro LONGINA0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name fixnum)
	    (setf (parm-32b ,acb ,index) ,arg-name)))
    ))

(setf (get 'LONGINA0 :sizeof) 4)
(setf (get 'LONGINA0 :pretty-arg-name) "LongInteger")

;MACHNDLINA0

(defmacro MACHNDLINA0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name mac-handle)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle))))
    ))

(setf (get 'MACHNDLINA0 :sizeof) 4)
(setf (get 'MACHNDLINA0 :pretty-arg-name) "Mac-Handle-inst")

;MENUHNDLETYPE

(defmacro MenuHndleType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'MenuInfo)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s cannot be coerced into a Handle." ,arg-name)))))

(setf (get 'MenuHndleType :sizeof) 4)
(setf (get 'MenuHndleType :pretty-arg-name) "MenuInfo-inst")

;PATTERNTYPE

(defmacro PatternType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'Pattern)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s is not a pattern." ,arg-name)))))

(setf (get 'PatternType :sizeof) 4)
(setf (get 'PatternType :pretty-arg-name) "Pattern-inst")

;PENSTATETYPE

(defmacro PenStateType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name PenState)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle))))
    ))
	
(setf (get 'PenStateType :sizeof) 4)
(setf (get 'PenStateType :pretty-arg-name) "PenState-inst")

;PICTTYPE

(defmacro PictType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'Picture)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s is not a Picture." ,arg-name)))))

(setf (get 'PictType :sizeof) 4)
(setf (get 'PictType :pretty-arg-name) "Picture-inst")

;PolyTYPE

(defmacro PolyType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'Polygon)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s is not a Polygon." ,arg-name)))))

(setf (get 'PolyType :sizeof) 4)
(setf (get 'PolyType :pretty-arg-name) "Polygon-inst")

;PTRTYPE

(defmacro PtrType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name mac-pointer)
	    (setf (parm-32b ,acb ,index)
		  (send ,arg-name :pointer))))
    ))

(setf (get 'PtrType :sizeof) 4)
(setf (get 'PtrType :pretty-arg-name) "Mac-Pointer-inst")

;PTRinA0
(defmacro PtrInA0 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name mac-pointer)
	    (setf (parm-32b ,acb ,index)
		  (send ,arg-name :pointer))))
    ))

(setf (get 'PtrInA0 :sizeof) 4)
(setf (get 'PtrInA0 :pretty-arg-name) "Mac-Pointer-inst")

;PTRinA1
(defmacro PtrInA1 (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-type-1 ,arg-name mac-pointer)
	    (setf (parm-32b ,acb ,index)
		  (send ,arg-name :pointer))))
    ))

(setf (get 'PtrInA1 :sizeof) 4)
(setf (get 'PtrInA1 :pretty-arg-name) "Mac-Pointer-inst")

;PTTYPE

(defmacro PtType (arg-name acb acb-index)
  (let ((index (floor acb-index 2)))
    `(cond ((check-type-1 ,arg-name point)
	    (setf (parm-16b ,acb (+ ,index 0)) (send ,arg-name :v))
	    (setf (parm-16b ,acb (+ ,index 1)) (send ,arg-name :h))))
    ))

(setf (get 'PtType :sizeof) 4)
(setf (get 'PtType :pretty-arg-name) "Point")

;RECTTYPE

(defmacro RectType (arg-name acb acb-index)
  (let ((index (floor acb-index 2)))
    `(cond ((typep ,arg-name 'rect)
	    (setf (parm-16b ,acb (+ ,index 0)) (send ,arg-name :top))
	    (setf (parm-16b ,acb (+ ,index 1)) (send ,arg-name :left))
	    (setf (parm-16b ,acb (+ ,index 2)) (send ,arg-name :bottom))
	    (setf (parm-16b ,acb (+ ,index 3)) (send ,arg-name :right)))
	   ((listp ,arg-name)
	    (setf (parm-16b ,acb (+ ,index 0)) (first ,arg-name))
	    (setf (parm-16b ,acb (+ ,index 1)) (second ,arg-name))
	    (setf (parm-16b ,acb (+ ,index 2)) (third ,arg-name))
	    (setf (parm-16b ,acb (+ ,index 3)) (fourth ,arg-name)))
	   (t
	    (ferror nil "~s cannot be coerced into a Rect." ,arg-name)))))

(setf (get 'recttype :sizeof) 8)
(setf (get 'recttype :pretty-arg-name) "Rect")

;RESTYPETYPE

(defmacro RESTYPETYPE (arg-name acb acb-index)
  (let ((index acb-index))
    `(cond ((and (typep ,arg-name 'string) (= 4 (length ,arg-name)))
	    (dotimes (i 4)
	      (setf (parm-8b ,acb (+ ,index i)) (elt ,arg-name i))))
	   (t
	    (ferror nil "Resource types must be 4 character strings: ~s."
		    ,arg-name)))))

(setf (get 'RESTYPETYPE :sizeof) 4)
(setf (get 'RESTYPETYPE :pretty-arg-name) "ResType")

;RGNTYPE

(defmacro RgnType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'Region)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s needs a Region." ,arg-name)))))

(setf (get 'RgnType :sizeof) 4)
(setf (get 'RgnType :pretty-arg-name) "Region-inst")

;SFTYPELISTTYPE

(defmacro SFTypeListType (arg-name acb acb-index)
  `(cond ((and (listp ,arg-name)(<= (length ,arg-name) 4))
	  (let ((index ,acb-index))
	    (dolist (typ ,arg-name)
	      (let ((str (string typ)))
		(dotimes (i 4)
		  (setf (parm-8b ,acb (+ index i)) (elt str i)))
		(setq index (+ 4 index))))))
	 (t (ferror nil "~s must be a Lisp list of up to 4 ResType strings"
		    ,arg-name))))
	  
(setf (get 'SFTypeListType :sizeof) 16)
(setf (get 'SFTypeListType :pretty-arg-name) "LIST-of-ResTypes")

;STRTYPE

(defmacro StrType (arg-name acb acb-index)
  (let ((index acb-index))
    `(cond ((check-type-1 ,arg-name string)
	    (setf (parm-8b ,acb (+ ,index 0)) (length ,arg-name))
	    (dotimes (i (length ,arg-name))
	      (setf (parm-8b ,acb (+ ,index i 1)) (elt ,arg-name i)))))
    ))

(setf (get 'StrType :sizeof) 256)
(setf (get 'StrType :pretty-arg-name) "String256")

;TERECTYPE

(defmacro TERecType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'TERec)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s needs a TERec." ,arg-name)))))

(setf (get 'TERecType :sizeof) 4)
(setf (get 'TERecType :pretty-arg-name) "TERec-inst")

;THPRINTTYPE
(defmacro THPrintType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'THPrint)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s cannot be coerced into a THPrint." ,arg-name)))))

(setf (get 'THPrintType :sizeof) 4)
(setf (get 'THPrintType :pretty-arg-name) "THPrint-inst")

;TPRSTATUSTYPE
(defmacro TPrStatusType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type ,arg-name 'TPrStatus)
	    (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	   (t
	    (ferror nil "~s cannot be coerced into a TPrStatus." ,arg-name)))))

(setf (get 'TPrStatusType :sizeof) 4)
(setf (get 'TPrStatusType :pretty-arg-name) "TPrStatus-inst")

;****** VAR's *******

;(defmacro VAR (x)
;  `(locf ,x))

;VCharType (used only by !GetItemMark)
(defmacro VCharType (arg-name acb acb-index &optional ret)
    (let ((index (floor acb-index 4)))
      (cond (ret				       ;only need return
	     `(cond ((typep ,arg-name 'locative)
		     (setf (contents ,arg-name)
		       (character (ldb 8 (parm-32b ,acb ,index)))))
		    (t (ferror nil "Need a VAR arg for the character.")))))
      ))

(setf (get 'VCharType :sizeof) 4)
(setf (get 'VCharType :pretty-arg-name) "(VAR character)")

;This is used only by !FindControl
;Generates no garbage normally
(defmacro VCntlHndlType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 4)))
    (cond (ret		 ; When we have a VAR return value
	   `(let ((con (parm-32b ,acb ,index)))
	      (cond ((typep ,arg-name 'controlrecord)
		     (send ,arg-name :set-handle con))
		    ((typep ,arg-name 'locative)
		     (setf (contents ,arg-name)
		       (cond ((cdr (assoc con *mac-control-list*)))
			     (t (make-instance-no-init
				  'tb:controlrecord :handle con)))))
		    (t (ferror nil "Pass a (VAR ...) to !FindControl."))))
	   ))))

(setf (get 'VCntlHndlType :sizeof) 4)
(setf (get 'VCntlHndlType :pretty-arg-name) "(VAR ControlRecord-inst)")

;VEVTRECTYPE
;This is not a true VAR, since not immediate data

(defmacro VEvtRecType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 2)))
    (cond (ret
	  `(typecase ,arg-name
	     (EventRecord
		(send ,arg-name :set-what   (parm-16b ,acb (+ ,index 0)))
		(send ,arg-name :set-message (+  (parm-16b ,acb (+ ,index 2))
						 (ash (parm-16b ,acb (+ ,index 1))
						      16)))
		(send ,arg-name :set-when (+  (parm-16b ,acb (+ ,index 4))
						 (ash (parm-16b ,acb (+ ,index 3))
						      16)))
		(send ,arg-name :set-v  (parm-16b ,acb (+ ,index 5)))
		(send ,arg-name :set-h  (parm-16b ,acb (+ ,index 6)))
		(send ,arg-name :set-modifiers (parm-16b ,acb (+ ,index 7))))
	     (t (ferror nil "~s must be an EventRecord instance." ,arg-name)))) 
	  ))
  )

(setf (get 'VEvtRecType :sizeof) 16)
(setf (get 'VEvtRecType :pretty-arg-name) "EventRecord-inst")

(defmacro VHndlType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 4)))
    (cond ((null ret)
	   `(cond ((typep ,arg-name 'mac-handle)
		   (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
		  (t (ferror nil "~s must be a mac-handle instance."
			     ,arg-name))))
	(t ; When we have a VAR return value
	   `(cond ((typep ,arg-name 'mac-handle)
		   (send ,arg-name :set-handle (parm-32b ,acb ,index)))
		  )))
    ))

(setf (get 'VHndlType :sizeof) 4)
(setf (get 'VHndlType :pretty-arg-name) "Mac-Handle-inst")

;VINTTYPE
;Integers are immediate data types and need true VAR support.
;signs correct??
(defmacro VIntType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 4)))
    (cond ((null ret)
	   `(cond ((typep ,arg-name 'locative)
		   (setf (parm-32b ,acb ,index)
			 (if (boundp ',arg-name) (contents ,arg-name) 0)))
		  (t (ferror nil "~s must be a VAR arg -- e.g. (VAR id)." ,arg-name))))
	  (t ; When we have a VAR return value
	   `(setf (contents ,arg-name)
				    (signed-ldb 16 (parm-32b ,acb ,index)))))))
	

(setf (get 'VIntType :sizeof) 4)
(setf (get 'VIntType :pretty-arg-name) "(VAR integer)")

(defmacro VLngIntType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 4)))
    (cond ((null ret)
	   `(cond ((typep ,arg-name 'locative)
		   (setf (parm-32b ,acb ,index)
			 (if (boundp ',arg-name) (contents ,arg-name) 0)))
		  (t (ferror nil "~s must be a VAR arg -- e.g. (VAR id)." ,arg-name))))
	  (t ; When we have a VAR return value
	   `(setf (contents ,arg-name)
				    (parm-32b ,acb ,index))))))
	

(setf (get 'VLngIntType :sizeof) 4)
(setf (get 'VLngIntType :pretty-arg-name) "(VAR longint)")

;VPENSTATETYPE
;Used only by !GetPenState
(defmacro VPenStateType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 4)))
    (cond ((null ret)
	   `(typecase ,arg-name
	      (PenState (setf (parm-32b ,acb ,index) (send ,arg-name :handle)))
	      (t (ferror nil "Must pass a PenState instance to !GetPenState.")))))
    ))
	
(setf (get 'VPenStateType :sizeof) 4)
(setf (get 'VPenStateType :pretty-arg-name) "PenState-inst")

;VPTRTYPE
;Mac-Pointers are not immediate data types on the microExplorer
;True VAR support is maintained for compatability with the Mac toolbox.

(defmacro VPtrType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 4)))
    (cond ((null ret)
	   `(cond ((typep ,arg-name 'mac-pointer)
		   (setf (parm-32b ,acb ,index) (send ,arg-name :pointer)))
		  ((typep ,arg-name 'locative)
		   (setf (parm-32b ,acb ,index)
			 (send (contents ,arg-name) :pointer)))
		  (t (ferror nil "~s must be a mac-pointer or grafport instance."
			     ,arg-name))))
	(t ; When we have a VAR return value
	   `(cond ((typep ,arg-name 'mac-pointer)
		   (send ,arg-name :set-pointer (parm-32b ,acb ,index)))
		  ((typep (contents ,arg-name) 'mac-pointer)
		   (send (contents ,arg-name)
			 :set-pointer (parm-32b ,acb ,index)))
		  (t
		   (setf (contents ,arg-name)
		     (make-instance-no-init 'mac-pointer :pointer
				    (parm-32b ,acb ,index)))))))))
	

(setf (get 'VPtrType :sizeof) 4)
(setf (get 'VPtrType :pretty-arg-name) "Mac-Pointer-inst")

;VPTTYPE
;Points are not immediate data types, so are not true VARs
;However, points and rectangles are kept as uE data (unlike EvtRecs above)
;Since the data is changed on return, it must be recopied out of the ACB.

(defmacro VPtType (arg-name acb acb-index &optional ret)       ; 01/16/89 Mac
  (let ((index (floor acb-index 2)))
    (cond ((null ret)
	   `(cond ((typep ,arg-name 'point)
		   (setf (parm-16b ,acb (+ ,index 0)) (send ,arg-name :v))
		   (setf (parm-16b ,acb (+ ,index 1)) (send ,arg-name :h)))
		  (t  (ferror nil "~s needs to be a Point." ,arg-name))))
	  (t					       ;ACB has new data on return
	   `(progn (send ,arg-name :set-v
			 (signed-ldb 16 (parm-16b ,acb (+ ,index 0))))
		   (send ,arg-name :set-h
			 (signed-ldb 16 (parm-16b ,acb (+ ,index 1)))))))))

(setf (get 'VPtType :sizeof) 4)
(setf (get 'VPtType :pretty-arg-name) "Point")

;VFontInfoType
;Fonts are not immediate data types, so are not true VARs
;However, points, rectangles and Fonts are kept as uE data
;Data only is returned in this type

(defmacro VFontInfoType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 2)))
    (cond (ret
	  `(typecase ,arg-name
	     (FontInfo
		(send ,arg-name :set-ascent   (parm-16b ,acb (+ ,index 0)))
		(send ,arg-name :set-descent  (parm-16b ,acb (+ ,index 1)))
		(send ,arg-name :set-widmax   (parm-16b ,acb (+ ,index 2)))
		(send ,arg-name :set-leading  (parm-16b ,acb (+ ,index 3))))
	     (t (ferror nil "~s must be a FontInfo instance." ,arg-name)))) 
	  ))
  )

(setf (get 'VFontInfoType :sizeof) 8)
(setf (get 'VFontInfoType :pretty-arg-name) "FontInfo-inst")

;VRECTTYPE
;Rects are not immediate data types, so are not true VARs
;However, points and rectangles are kept as uE data
;Since the data is changed on return, it must be recopied out of the ACB.

(defmacro VRectType (arg-name acb acb-index &optional ret)     ; 01/16/89 Mac
  (let ((index (floor acb-index 2)))
    (cond ((null ret)
	   `(cond ((typep ,arg-name 'rect)
		   (setf (parm-16b ,acb (+ ,index 0)) (send ,arg-name :top))
		   (setf (parm-16b ,acb (+ ,index 1)) (send ,arg-name :left))
		   (setf (parm-16b ,acb (+ ,index 2)) (send ,arg-name :bottom))
		   (setf (parm-16b ,acb (+ ,index 3)) (send ,arg-name :right)))
		  (t
		   (ferror nil
	     "So called VAR Rect args must be 'Rect instances: ~s." ,arg-name))))
	  (t `(progn
		(send ,arg-name :set-top
		      (signed-ldb 16 (parm-16b ,acb (+ ,index 0))))
		(send ,arg-name :set-left
		      (signed-ldb 16 (parm-16b ,acb (+ ,index 1))))
		(send ,arg-name :set-bottom
		      (signed-ldb 16 (parm-16b ,acb (+ ,index 2))))
		(send ,arg-name :set-right
		      (signed-ldb 16 (parm-16b ,acb (+ ,index 3)))))))))

(setf (get 'VRectType :sizeof) 8)
(setf (get 'VRectType :pretty-arg-name) "Rect")

;VRESTYPE
(defmacro VResType (arg-name acb acb-index &optional ret)
  (let ((index acb-index))
    ;;No pre-processing since traps like !GetResInfo don't use the info
    (cond (ret					       ;post processing
	   `(cond ((typep ,arg-name 'locative)
		   (let ((s (make-string 4)))
		     (dotimes (i 4)
		       (setf (elt s i) (parm-8b ,acb (+ ,index  i))))
		     (setf (contents ,arg-name) s)))
		  (t (ferror nil "~&Use a (VAR vResType) for arg.")))
	   ))))

(setf (get 'VresType :sizeof) 4)
(setf (get 'VresType :pretty-arg-name) "(VAR ResType)")

;vSFReplyType
(defmacro vSFReplyType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 2)))
    (cond (ret
	   `(typecase ,arg-name
	      (SFReply
	       (send ,arg-name :set-good
		     (if (zerop (parm-8b ,acb ,acb-index)) nil t))
	       (let ((str (make-string 4)))
		 (send ,arg-name :set-ftype
		       (dotimes (i 4 str)
			 (setf (elt str i) (parm-8b ,acb (+ ,acb-index 2 i))))))
	       (send ,arg-name :set-vREfNum
		     (signed-ldb 16 (parm-16b ,acb (+ ,index 3))))
	       (send ,arg-name :set-version 0)
	       (let* ((len (min (parm-8b ,acb (+ ,acb-index 10)) 63))
		      (str (make-string len)))
		 (send ,arg-name :set-fname
		       (dotimes (i len str)
			 (setf (elt str i) (parm-8b ,acb (+ ,acb-index 11 i))))))
	       )
	      (t (ferror nil "SFReply instance needed instead of ~s." ,arg-name)))
	   ))))

(setf (get 'vSFReplyType :sizeof) 74)
(setf (get 'vSFReplyType :pretty-arg-name) "SFReply-inst")
	      
;VSTRTYPE

(defmacro VStrType (arg-name acb acb-index &optional ret)
  (let ((index acb-index))
    (cond ((null ret)
	   `(typecase ,arg-name
	      (locative
	       (let ((s (contents ,arg-name)))
		 (typecase s
		   (string 
		    (setf (parm-8b ,acb (+ ,index 0)) (length s))
		    (dotimes (i (length s))
		      (setf (parm-8b ,acb (+ ,index i 1)) (elt s i)))))))
	      (t
	       (ferror nil "~s is not a VAR String." ,arg-name))))
	  (t					       ;post processing
	   `(typecase ,arg-name
	      (locative
	       (let ((s (contents ,arg-name))(len (parm-8b ,acb (+ ,index 0))))
		 (setq s (make-string len))
		 (setf (contents ,arg-name) s)
		 (dotimes (i (parm-8b ,acb (+ ,index 0)))
		   (setf (elt s i) (parm-8b ,acb (+ ,index i 1))))))))
    )))

(setf (get 'VStrType :sizeof) 256)
(setf (get 'VStrType :pretty-arg-name) "(VAR string)")

;VWINDPTRTYPE -- for !FindWindow and !GetWMgrPort
;Generates no garbage normally
(defmacro VWindPtrType (arg-name acb acb-index &optional ret)
  (let ((index (floor acb-index 4)))
    (cond (ret		 ; When we have a VAR return value
	   `(let ((win (parm-32b ,acb ,index)))
	      (cond ((typep ,arg-name 'mac-pointer)
		     (send ,arg-name :set-pointer win))
		    ((typep ,arg-name 'locative)
		     (setf (contents ,arg-name)
		       (cond ((cdr (assoc win *mac-window-list*)))
			     (t (make-instance-no-init 'tb:window :pointer win)))))
		    (t (ferror nil "Pass a mac-pointer, or use a window inst."))))
	   ))))

(setf (get 'VWindPtrType :sizeof) 4)
(setf (get 'VWindPtrType :pretty-arg-name) "(VAR window)")

;WINDOWRECTYPE

(defmacro WindowRecType (arg-name acb acb-index)
  (let ((index (floor acb-index 4)))
    `(cond ((check-toolbox-type-or-pointer ,arg-name 'window)
	    (setf (parm-32b ,acb ,index)
		  (send ,arg-name :pointer)))
	   (t
	    (ferror nil "~s is not a Window." ,arg-name)))))

(setf (get 'WindowRecType :sizeof) 4)
(setf (get 'WindowRecType :pretty-arg-name) "Window-inst")

;************ Return Types *************************

;RBOOLTYPE
(defmacro rBoolType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(if (zerop (parm-32b ,acb ,index)) nil t)))

(setf (get 'rBoolType :sizeof) 4)
(setf (get 'rBoolType :pretty-arg-name) "Boolean")

;RCntlHndlType
(defmacro rCntlHndlType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(make-instance-no-init 'ControlRecord :handle (parm-32b ,acb ,index))))

(setf (get 'rCntlHndlType :sizeof) 4)
(setf (get 'rCntlHndlType :pretty-arg-name) "ControlRecord-inst")

;RDIALOGRECTYPE
(defmacro rDialogRecType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(make-instance-no-init 'DialogRecord :pointer (parm-32b ,acb ,index))))

(setf (get 'rDialogRecType :sizeof) 4)
(setf (get 'rDialogRecType :pretty-arg-name) "DialogRecord-inst")

;RDONTCARE
(defmacro rDontCare (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (declare (ignore acb acb-index))
  t)

(setf (get 'rDontCare :sizeof) 0)
(setf (get 'rDontCare :pretty-arg-name) "Nothing")

;RHNDLTYPE
(defmacro rHndlType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(make-instance-no-init 'mac-handle :handle (parm-32b ,acb ,index))))

(setf (get 'rHndlType :sizeof) 4)
(setf (get 'rHndlType :pretty-arg-name) "Mac-Handle-inst")

;RHNDLA0_RESCD0
(defmacro rHndlA0_ResCD0 (acb acb-index &optional trap-table-index)
  (let ((index (floor acb-index 4)))
    `(let ((err))
       (cond ((= 0 (setq err (signed-ldb 16 (parm-32b ,acb ,index))))
	      (values
		(make-instance-no-init 'mac-handle
				       :handle (parm-32b ,acb (+ ,index 1)))
		err))
	     (*signal-mac-oserr*
	      (values nil
		      (signal-oserr err ',(trap-symbol trap-table-index)
					  "Memory allocation error ~s~%" err)))
	     (t
	      (values nil err))))
    ))

(setf (get 'rHndlA0_ResCD0 :sizeof) 8)
(setf (get 'rHndlA0_ResCD0 :pretty-arg-name) "Mac-Handle-inst")

;RINTTYPE
;TBServer sign extends the result before returning it.
(defmacro rIntType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(signed-ldb 25 (parm-32b ,acb ,index))))

(setf (get 'rIntType :sizeof) 4)
(setf (get 'rIntType :pretty-arg-name) "Integer")

;RINTIND0
;Used only by !HGetState -- we treat as UNSIGNED byte, not signed as in IM
(defmacro rIntinD0 (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(parm-32b ,acb ,index)))

(setf (get 'rIntinD0 :sizeof) 4)
(setf (get 'rIntinD0 :pretty-arg-name) "Integer")

;RLNGINTTYPE
(defmacro rLngIntType (acb acb-index &optional trap-table-index)       ; 01/16/89 Mac
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(signed-ldb 32 (parm-32b ,acb ,index))))


(setf (get 'rLngIntType :sizeof) 4)
(setf (get 'rLngIntType :pretty-arg-name) "longint")

;RLONGA0_LONGD0 - Multiple Value Return of D0 & A0
(defmacro rLongA0_LongD0 (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(values (parm-32b ,acb ,index)
	     (parm-32b ,acb (+ ,index 1)))))

(setf (get 'rLongA0_LongD0 :sizeof) 8)
(setf (get 'rLongA0_LongD0 :pretty-arg-name) "longint longint")

;RLONGIND0

(defmacro rLonginD0 (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(parm-32b ,acb ,index)))

(setf (get 'rLonginD0 :sizeof) 4)
(setf (get 'rLonginD0 :pretty-arg-name) "longint")

;RMACHNDLINA0
(defmacro rMacHndlinA0 (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(make-instance-no-init 'mac-handle :handle (parm-32b ,acb ,index))))

(setf (get 'rMacHndlinA0 :sizeof) 4)
(setf (get 'rMacHndlinA0 :pretty-arg-name) "Mac-Handle-inst")

;RMVINTTYPE - Multiple Value Return of HiWord and LoWord
(defmacro rMVIntType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(values (signed-ldb 16 (parm-32b ,acb ,index))
	     (signed-ldb 16 (parm-32b ,acb (+ ,index 1))))))

(setf (get 'rMVIntType :sizeof) 8)
(setf (get 'rMVIntType :pretty-arg-name) "integer integer")

;RPICTTYPE
(defmacro rPictType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(make-instance-no-init 'picture :handle (parm-32b ,acb ,index))))

(setf (get 'rPictType :sizeof) 4)
(setf (get 'rPictType :pretty-arg-name) "Picture-inst")

;RPTRInA0
(defmacro rPtrInA0 (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(make-instance-no-init 'mac-pointer :pointer (parm-32b ,acb ,index))))

(setf (get 'rPtrInA0 :sizeof) 4)
(setf (get 'rPtrInA0 :pretty-arg-name) "Mac-pointer-inst")

;RPTRA0_RESCD0
(defmacro rPtrA0_ResCD0 (acb acb-index &optional trap-table-index)
  (let ((index (floor acb-index 4)))
    `(let ((err))
       (cond ((= 0 (setq err (signed-ldb 16 (parm-32b ,acb ,index))))
	      (values
		(make-instance-no-init 'mac-pointer
				       :pointer (parm-32b ,acb (+ ,index 1)))
		err))
	     (*signal-mac-oserr*
	      (values nil
		      (signal-oserr err ',(trap-symbol trap-table-index))))
	     (t
	      (values nil err))))
    ))

(setf (get 'rPtrA0_ResCD0 :sizeof) 8)
(setf (get 'rPtrA0_ResCD0 :pretty-arg-name) "Mac-pointer-inst")


(defmacro rTERecType (acb acb-index trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(make-instance-no-init 'terec :handle (parm-32b ,acb ,index))
    ))

(setf (get 'rTERecType :sizeof) 4)
(setf (get 'rTERecType :pretty-arg-name) "TERec-inst")

;RPTRTYPE
(defmacro rPtrType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(make-instance-no-init 'mac-pointer :pointer (parm-32b ,acb ,index))))

(setf (get 'rPtrType :sizeof) 4)
(setf (get 'rPtrType :pretty-arg-name) "Mac-pointer-inst")


;RRESULTCODEIND0
(defmacro rResultCodeInD0 (acb acb-index &optional trap-table-index)
  (declare (special *signal-mac-oserr*))
  (let ((index (floor acb-index 4)))
    `(cond (*signal-mac-oserr*
	    (signal-oserr (signed-ldb 16 (parm-32b ,acb ,index))
			  ',(intern (string-upcase
				     (trap-name trap-table-index)) 'tb)))
	   (t
	    (signed-ldb 16 (parm-32b ,acb ,index))))
       )
  )

(setf (get 'rResultCodeInD0 :sizeof) 4)
(setf (get 'rResultCodeInD0 :pretty-arg-name) "result-code")

;RWINDPTRTYPE
;The *mac-window-list* is the only way to associate an mX instance with a
; particular mac window.  Since mac window addr are unique, this even works in
; multiple mX process environment.
;The only calls that use the last clause are !NewWindows traps directly.
; In this case no assoc is needed.
;The window list is added to every time a new object is created, and deleted when
; disposed.
(defmacro rWindPtrType (acb acb-index &optional trap-table-index)
  (declare (ignore trap-table-index))
  (let ((index (floor acb-index 4)))
    `(let ((win (parm-32b ,acb ,index)))
       (cond ((cdr (assoc win *mac-window-list*)))
	     ((= 0 win) nil)
	     (t (make-instance-no-init 'window :pointer win))))
    ))

(setf (get 'rWindPtrType :sizeof) 4)
(setf (get 'rWindPtrType :pretty-arg-name) "Window-inst")


