;;; -*- Mode:Common-Lisp; Package:MACTOOLBOX; Base:10 -*-

;;; For the puposes of the Texas Instruments/ ExperTelligence Contract, this
;;; software is in TBI Category "A".

;/*                                                                       */
;/*                      RESTRICTED RIGHTS LEGEND                         */
;/*                                                                       */
;/* Use, duplication, or disclosure by the Government is subject to       */
;/* restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  */
;/* Technical Data and Computer Software clause at 52.227-7013.           */
;/*                                                                       */
;/*                    TEXAS INSTRUMENTS INCORPORATED.                    */
;/*                            P.O. BOX 2909                              */
;/*                         AUSTIN, TEXAS 78769                           */
;/*                              MS 2151                                  */
;/*                                                                       */
;/*  Copyright (C) 1987-1989 Texas Instruments Incorporated.              */
;/*  All rights reserved.                                                 */
;/*                                                                       */

;;; MODIFICATIONS
;;; 01/19/89 Mac  Definition of GENERATE-INSTACC macro moved to EXPER-TOOLBOX-ACCESSORS
;;;               which is the only place it is used.

(PROCLAIM '(SPECIAL *trap-table*))
(defvar *initial-trap-index* 0 "The index of the first trap in the trap table.")

(defvar *mac-trap-gen*)

;(defmacro send-if (inst msg &rest args)
;  `(let ((obj ,inst))
;     (if obj (send obj ,msg ,@args))))
  
(defun sizeof (sym)
  (declare (values length-in-bytes))
  (or (get sym :sizeof)
      (ferror "cannot take the sizeof ~s" sym)))

(defun acb-size-needed (trap-table-index &aux (byte-offset 0))
  (loop for i from 1 to (trap-nargs trap-table-index)
	do (setf byte-offset
		 (+ byte-offset
		    (sizeof (trap-arg-type trap-table-index i)))))
  (setq byte-offset (+ byte-offset
		       (sizeof (trap-return-type trap-table-index))))
  ;; Leave enough space for an error code at the end.
  (setq byte-offset (+ byte-offset 12))
  byte-offset)

(defmacro generate-trap (trap-table-index)
  (let ((byte-offset 4))
    `(defun ,(intern (string-upcase (trap-name trap-table-index)) 'tb)
	    ,(loop for i from 1 to (trap-nargs trap-table-index)
		   collect (trap-arg-name trap-table-index i))
            (declare (arglist
		       ,@(loop for i from 1 to (trap-nargs trap-table-index)
			      collect (trap-arg-pretty-name trap-table-index i)))
		     (values ,(trap-return-pretty-name trap-table-index)))
       (let ((acb (get-acb-fast ,(acb-size-needed trap-table-index))))
	 
	 ;; opcode 0 is a trap call, >0 are control calls.
	 (setf (add:opcode acb) 0)
	 
	 ;; Only set requestor complete if there is no return value or VAR arg.
	 ,(cond ((not (trap-wait-for-values trap-table-index))
		 `(setf (requestor-complete acb) t)))
	 
	 (setf (parm-32b acb 0) ,(+ *initial-trap-index* trap-table-index))
	 
	 ;; Each of the data type macros takes:
	 ;;  1) arg name
	 ;;  2) acb
	 ;;  3) byte offset into acb of where to put the arg

	 ,(cons 'progn
		(loop for i from 1 to (trap-nargs trap-table-index)
		      collect `(,(trap-arg-type trap-table-index i)
				,(trap-arg-name trap-table-index i)
				      acb
				      ,byte-offset)
		      do (setf byte-offset
			       (+ byte-offset
				  (sizeof (trap-arg-type trap-table-index i))))))

	 ,(cond ((not (trap-wait-for-values trap-table-index))
		 `(tb-transmit-packet acb))
		(t `(tb-transmit-packet-and-wait acb)))

	 ,(cond	((var-args? trap-table-index)
		 ;; Start over at the first arg.
		 (setf byte-offset 4)
		 (cons 'progn
			(loop for i from 1 to (trap-nargs trap-table-index)
			      when (var-arg-p trap-table-index i)
			      collect `(,(trap-arg-type trap-table-index i)
				       ,(trap-arg-name trap-table-index i)
				       acb
				       ,byte-offset
				       :return)
			     do (setf byte-offset
				      (+ byte-offset
					 (sizeof (trap-arg-type trap-table-index i)
						 )))))))
	 ,(cond ((trap-wait-for-values trap-table-index)
		 `(multiple-value-prog1
		    (,(trap-return-type trap-table-index)
		      acb
		      ,byte-offset
		      ,trap-table-index)		    
		    
		    ,(when (trap-signal-low-mem-error trap-table-index)
		       (setq byte-offset
			   (+ byte-offset
			      (sizeof (trap-return-type trap-table-index))))
		       
		       (let ((index (floor byte-offset 4)))
			 `(when *Signal-mac-oserr*
			    (signal-oserr (signed-ldb 16 (parm-32b acb ,index))
					  ',(trap-symbol trap-table-index)))))
		    (setf (requestor-complete acb) t)
		    (return-acb-fast acb)))
		((eq (trap-return-type trap-table-index) 'rDontCare)
		 T ;return just T for a dont care
		 ))

	 ))))

;;;


(defun trap-signal-low-mem-error  (trap-table-index)
  (member (trap-flags trap-table-index)
	  '(SaveResError MacPlus+SaveResError Mac2+SaveResError
			 SaveQdError Mac2+SaveQdError)))

(defun trap-flags (trap-table-index)
  (nth 1 (nth trap-table-index *trap-table*)))

(defun trap-name (trap-table-index)
  (declare (values trap-name-string))  
  (intern (string-upcase (nth 5 (nth trap-table-index *trap-table*))) 'tb))

(defun trap-symbol (trap-table-index)
  (declare (values trap-symbol))
  (intern (string-upcase (trap-name trap-table-index)) 'tb))

(defun trap-nargs (trap-table-index)
  (declare (values number-of-args))
  (length (nth 4 (nth trap-table-index *trap-table*))))

(defun trap-arg-name (trap-table-index n)
  (declare (values arg-name-string))
  (let ((name (nth (1- n) (nth 4 (nth trap-table-index *trap-table*)))))
    (intern (string-upcase (format nil "~a-~d" name n)) 'tb)))

(defun trap-arg-pretty-name (trap-table-index n)
  (declare (values arg-name-string))
  (get (nth (1- n) (nth 4 (nth trap-table-index *trap-table*))) :pretty-arg-name))

(defun trap-arg-type (trap-table-index n)
  (declare (values type-symbol))
  (let ((name (nth (1- n) (nth 4 (nth trap-table-index *trap-table*)))))
    (intern (string-upcase name) 'tb)))

(defun trap-return-type (trap-table-index)
  (declare (values return-type-symbol))
  (nth 3 (nth trap-table-index *trap-table*)))

(defun trap-return-pretty-name (trap-table-index)
  (declare (values return-type-symbol))
  (GET (nth 3 (nth trap-table-index *trap-table*)) :pretty-arg-name))

(defun var-arg-p (trap-table-index n)
  (declare (values t-or-nil))
  (char-equal (char (string (trap-arg-type trap-table-index n)) 0) #\v))

(defun var-args?  (trap-table-index)
  (declare (values t-or-nil))
    (loop for i from 1 to (trap-nargs trap-table-index)
	  when (var-arg-p trap-table-index i)
	  return t))


(defun trap-wait-for-values (trap-table-index)
  (declare (values t-or-nil))
  (or
    (loop for i from 1 to (trap-nargs trap-table-index)
	  when (var-arg-p trap-table-index i)
	  return t)
    (not (eq (trap-return-type trap-table-index) 'rDontCare))))


(defun find-trap-index (name)
  (position name *trap-table* :key #'sixth :test #'string-equal))


(defparameter *mac-traps* nil)

(defun generate-traps (&aux entry ok-to-build (cnt 0))
  (dotimes (i (length *trap-table*))
    (setq entry (nth i *trap-table*))
    (setq ok-to-build (good-tool-macro (fourth entry)))
    (dolist (j (fifth entry))
      (if (not (good-tool-macro j)) (setq ok-to-build nil)))
    (cond (ok-to-build
	   (incf cnt)
	   (cond (t ;(not (fboundp (intern (string-upcase (sixth entry)))))
		  (format t "Building ... ~s~%" (sixth entry))
		  (pushnew (eval `(generate-trap ,i)) *mac-traps*)
		  )))
	  (t (format t "Unable to build ... ~s~%" (sixth entry))))
    )
  (format t "~s traps generated Master." cnt))

(defun good-tool-macro (datatype)
  (cond ((or (and (fboundp datatype)
		  (typep (symbol-function datatype) 'cons)
		  (eq (car (symbol-function datatype)) 'macro))
	     (eq datatype 'rDontCare))
	 t)
	(t nil)))


(defmacro create-mac-traps (&aux entry ok-to-build (cnt 0))
  (setq *mac-trap-gen* nil *mac-traps* nil)
  (dotimes (i (length *trap-table*))
    (setq entry (nth i *trap-table*))
    (setq ok-to-build (good-tool-macro (fourth entry)))
    (dolist (j (fifth entry))
      (if (not (good-tool-macro j)) (setq ok-to-build nil)))
    (cond (ok-to-build
	   (push (intern (string-upcase (trap-name i))) *mac-traps*)
	   (incf cnt)
	   (pushnew `(generate-trap ,i) *mac-trap-gen*)
	   ))
    )
  (push `(setq *mac-traps* ',*mac-traps*) *mac-trap-gen*)
  (cons 'progn (nreverse *mac-trap-gen*))
  )




