;;; -*- Mode:Common-Lisp; Package:MACTOOLBOX; Base:10 -*-

;;; For the puposes of the Texas Instruments Contract, this software is in
;;; TBI Category "B". Copyright (C) 1988, 1987, ExperTelligence Incorporated.

;/*                                                                       */
;/*                      RESTRICTED RIGHTS LEGEND                         */
;/*                                                                       */
;/* Use, duplication, or disclosure by the Government is subject to       */
;/* restrictions as set forth in subdivision (b)(3)(ii) of the Rights in  */
;/* Technical Data and Computer Software clause at 52.227-7013.           */
;/*                                                                       */
;/*                           EXPERTELLIGENCE                             */
;/*                            5638 HOLLISTER                             */
;/*                          GOLETA, CA  93117                            */
;/*                                                                       */
;/*  Copyright (C) 1988, 1987, EXPERTELLIGENCE Incorporated.              */
;/*  All rights reserved.                                                 */
;/*                                                                       */


;;; ************************************************************************
;;; Modification History:
;;;   08/13/88 DWB  CHanged PixMap names to match IM V5
;;;   08/16/88 DWB  Changed ControlRecord "Action" slot to CntrlAction to
;;;                   avoid a name conflict.
;;;   08/23/88 rwr  Changed FontMetricRec to FMetricRec
;;;   08/28/88 rwr  Added nilhndl
;;;   08/29/88 rwr  deleted ScrapStuff
;;;   09/13/88 DWB&RWR Moved constants to ToolBox-Interface file. Deleted TERec
;;;   09/22/88 RJB  Removed accessors for CInfoPBRec, WDPBRec, and FCBPBRec
;;;                 since they are now defined in Generic Toolbox Library.  Also
;;;                 added accessors for directory variant of CInfoPBRec.
;;;   10/14/88 BJ   Add the accessor :ioDrDirId for cInfoPbRec.
;;;   01/18/89 MAC  Merged instance accessor changes from MACTOOLBOX-1-7
;;;   01/19/89 MAC  Definition of GENERATE-INSTACC macro moved here from
;;;                 EXPER-TB-MACROS since this is the only place it is ever used.
;;; ************************************************************************

; Return values

;# define	BooleanType		0x00
;# define	IntegerType		0x04
;# define	FloatType		0x06
;# define	ClassType		0x13
;# define	CharType		0x18
;# define	StringType		0x19
;# define	PtrType			0x1A
;# define	HandleType		0x1E
;# define	EscapeType		0x1F


;# define	normal			0x00
;# define	ClNormal		0x01
;# define	hPointer		0x02
;# define	pPointer		0x03
;# define	hInteger		0x04
;# define	pInteger		0x05
;# define	hByte			0x06
;# define	unused1			0x07
;# define	hFloat			0x08
;# define	pFloat			0x09
;# define	hResType		0x0A
;# define	hChar			0x0B
;# define	hString			0x0C
;# define	unused3			0x0D
;# define	hBoolean		0x0E
;# define	pBoolean		0x0F
;# define	unused4			0x10
;# define	unused5			0x11
;# define	hHandle			0x12
;# define	pHandle			0x13
;# define	unused6			0x14
;# define	unused7			0x15
;# define	unused8			0x16
;# define	unused9			0x17
;# define	unused10		0x18
;# define	unused11		0x19
;# define	unused12		0x1A
;# define	unused13		0x1B
;# define	unused14		0x1C
;# define	nothing			0x1D
;# define	readOnly		0x1E
;# define	writeOnly		0x1F


;******************* This macro generates instance-accessors *****************


(defmacro generate-instacc (class lst
			    &aux msg set-msg typchk fetchtyp methodlist
			    unknowndatatype coercetype)
  (setq class (intern (string-upcase class)))
  (cond ((get class 'sys::flavor)		       ;See if class exists
	; (format t  "~%Class: ~s" class)
	 (cons 'progn
	       (prog ()
	       top
		  (setq unknowndatatype nil)
	 (setq msg (intern (string-upcase (first lst)) 'keyword))
	 (setq set-msg (intern (string-upcase
				 (string-append "SET-" (first lst))) 'keyword))
	; (format t "~%    method ~s " msg)
	 (setq typchk (case (third lst)
			(IntegerType '(typep value 'Fixnum))
			(BooleanType '(or (eq value t) (eq value nil)))
			(FloatType   '(numberp value))
			(StringType  '(typep value 'string))
			(PtrType     '(typep value 'mac-pointer))
			(HandleType  '(typep value 'mac-handle))
			(pRectType    nil) ;no way to stow for now  
			(CharType    '(typep value 'character))
			(t  (setq UnknownDataType t))))

	 (setq fetchtyp
	       (case (fourth lst)
		 ((hPointer hHandle hFloat)
		  `(fetchhandle self ,(fifth lst)))
		 ((pPointer pHandle pFloat)
		  `(fetch self ,(fifth lst)))
		 (hInteger `(fetchwordhandle self ,(fifth lst)))
		 (pInteger `(fetchword self ,(fifth lst)))
		 ((hChar hByte hBoolean)
		  `(fetchbytehandle self ,(fifth lst)))
		 ((hRectType pRectType) 'value) ;for looks only-see below
		 (pBoolean `(fetchbyte self ,(fifth lst)))
		 (t (setq unknowndatatype t))))
	 (setq coercetype ;all stows must be typed number
	       (case (fourth lst)
		 ((hBoolean pBoolean) `(if value 1 0))
		 ((hPointer pPointer) `(send value :pointer))
		 ((hHandle  pHandle)  `(send value :handle))
		 (hChar `(char-code value))
		  ;no stowRect or StowRectHandle
		 (t 'value)
		 ))
	 (if (not unknowndatatype)
	  (setq methodlist (cons 
	 `(defmethod (,class ,msg)(&optional (value nil supp?))
	    (if supp?
		(cond (,typchk
		       (setf ,fetchtyp ,coercetype))
		      (t
		       (ferror nil
			       "Bad instance accessor value: ~s, ~s required.~%"
			       value ',(third lst))))
		,(case (third lst)
		   (IntegerType `(signed-ldb 16 ,fetchtyp))
		   (BooleanType `(if (zerop ,fetchtyp) nil t))
		   (FloatType `,fetchtyp)	       ;like integers
		   (StringType nil)		       ;skip for now
		   (PtrType `(make-instance 'mac-pointer :pointer ,fetchtyp))
		   (HandleType `(make-instance 'mac-handle :handle ,fetchtyp))
		   (pRectType `(let ((r (make-instance 'rect)))
				(tb:fetchrect self ,(fifth lst) r)
				r))
		   (hRectType `(let ((r (make-instance 'rect)))
				(tb:fetchrecthandle self ,(fifth lst) r)
				r))
		   (CharType `(character ,fetchtyp)))))
	 methodlist)))
	 (if (not unknowndatatype)
	     (setq methodlist
		   (cons 
		     `(defmethod (,class ,set-msg)(value)
			(cond (,typchk
			       (setf ,fetchtyp ,coercetype))
			      (t
			       (ferror nil
			       "Bad instance accessor value: ~s, ~s required.~%"
			       value ',(third lst))))
			)
		     methodlist)))

		  (setq lst (nthcdr 5 lst))
		  (if (null lst) (return methodlist))
		  (go top))))
	(t ;(format t "Class does not exist ~s~%" class)
	   )))


(generate-instacc "Cursor"
 (	"curdata1"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"curdata2"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"curdata3"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"curdata4"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"curdata5"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"curdata6"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"curdata7"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"curdata8"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"curdata9"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"curdata10"
		InstanceAcc
		IntegerType
		hInteger
		#x12
			
		"curdata11"
		InstanceAcc
		IntegerType
		hInteger
		#x14
			
		"curdata12"
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"curdata13"
		InstanceAcc
		IntegerType
		hInteger
		#x18
			
		"curdata14"
		InstanceAcc
		IntegerType
		hInteger
		#x1A
			
		"curdata15"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
			
		"curdata16"
		InstanceAcc
		IntegerType
		hInteger
		#x1E
			
		"curmask1"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"curmask2"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"curmask3"
		InstanceAcc
		IntegerType
		hInteger
		#x24
			
		"curmask4"
		InstanceAcc
		IntegerType
		hInteger
		#x26
			
		"curmask5"
		InstanceAcc
		IntegerType
		hInteger
		#x28
			
		"curmask6"
		InstanceAcc
		IntegerType
		hInteger
		#x2A
			
		"curmask7"
		InstanceAcc
		IntegerType
		hInteger
		#x2C
			
		"curmask8"
		InstanceAcc
		IntegerType
		hInteger
		#x2E
			
		"curmask9"
		InstanceAcc
		IntegerType
		hInteger
		#x30
			
		"curmask10"
		InstanceAcc
		IntegerType
		hInteger
		#x32
			
		"curmask11"
		InstanceAcc
		IntegerType
		hInteger
		#x34
			
		"curmask12"
		InstanceAcc
		IntegerType
		hInteger
		#x36
			
		"curmask13"
		InstanceAcc
		IntegerType
		hInteger
		#x38
			
		"curmask14"
		InstanceAcc
		IntegerType
		hInteger
		#x3A
			
		"curmask15"
		InstanceAcc
		IntegerType
		hInteger
		#x3C
			
		"curmask16"
		InstanceAcc
		IntegerType
		hInteger
		#x3E
			
		"hotspotv"
		InstanceAcc
		IntegerType
		hInteger
		#x40
			
		"hotspoth"
		InstanceAcc
		IntegerType
		hInteger
		#x42
	))




(generate-instacc "PenState"
 (	"pnLocv"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"pnLoch"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"pnSizev"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"pnSizeh"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"pnMode"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"pnPat1"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"pnPat2"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"pnPat3"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"pnPat4"
		InstanceAcc
		IntegerType
		hInteger
		#x10
	))





(generate-instacc "Picture"
 (	"PicSize"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"picFrametop"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"picFrameleft"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"picFramebottom"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"picFrameright"
		InstanceAcc
		IntegerType
		hInteger
		#x08
	))



(generate-instacc "Polygon"
 (	"PolySize"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"polyBBoxtop"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"polyBBoxleft"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"polyBBoxbottom"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"polyBBoxright"
		InstanceAcc
		IntegerType
		hInteger
		#x08
	))



(generate-instacc "QDProcs"
 (	"textproc"
		InstanceAcc
		PtrType
		hPointer
		#x00
			
		"lineproc"
		InstanceAcc
		PtrType
		hPointer
		#x04
			
		"rectproc"
		InstanceAcc
		PtrType
		hPointer
		#x08
			
		"rrectproc"
		InstanceAcc
		PtrType
		hPointer
		#x0C
			
		"ovalproc"
		InstanceAcc
		PtrType
		hPointer
		#x10
			
		"arcproc"
		InstanceAcc
		PtrType
		hPointer
		#x14
			
		"polyproc"
		InstanceAcc
		PtrType
		hPointer
		#x18
			
		"rgnproc"
		InstanceAcc
		PtrType
		hPointer
		#x1C
			
		"bitsproc"
		InstanceAcc
		PtrType
		hPointer
		#x20
			
		"commentproc"
		InstanceAcc
		PtrType
		hPointer
		#x24
			
		"txmeasproc"
		InstanceAcc
		PtrType
		hPointer
		#x28
			
		"getpicproc"
		InstanceAcc
		PtrType
		hPointer
		#x2C
			
		"putpicproc"
		InstanceAcc
		PtrType
		hPointer
		#x30
	))







(generate-instacc "ControlRecord"
 (	"next"
		InstanceAcc
		HandleType
		hHandle
		#x00
			
		"owner"
		InstanceAcc
		PtrType
		hPointer
		#x04
			
		"top"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"left"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"bottom"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"right"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"vis"
		InstanceAcc
		IntegerType
		hByte
		#x10
			
		"hilite"
		InstanceAcc
		IntegerType
		hByte
		#x11
			
;		"value"    ;duplicates generic methods
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x12
			
;		"min"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x14
			
;		"max"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x16
			
		"defproc"
		InstanceAcc
		HandleType
		hHandle
		#x18
			
		"data"
		InstanceAcc
		HandleType
		hHandle
		#x1C
			
		"Contrlaction"			       ;***---***db 8/16/88
		InstanceAcc
		PtrType
		hPointer
		#x20
			
		"refcon"
		InstanceAcc
		PtrType
		hPointer
		#x24
			
		"title"
		InstanceAcc
		StringType
		hString
		#x28
	))



(generate-instacc "MenuInfo"
 (	"menuID"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"menuWidth"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"menuHeight"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"menuProc"
		InstanceAcc
		HandleType
		hHandle
		#x06
			
		"menuEnableflags"
		InstanceAcc
		FloatType
		hFloat
		#x0A
			
		"menuData"
		InstanceAcc
		StringType
		hString
		#x0E
	))


(generate-instacc "DialogRecord"
 (	"items"
		InstanceAcc
		HandleType
		pHandle
		#x9C
			
		"textH"
		InstanceAcc
		HandleType
		pHandle
		#xA0
			
		"editField"
		InstanceAcc
		IntegerType
		pInteger
		#xA4
			
		"editOpen"
		InstanceAcc
		IntegerType
		pInteger
		#xA6
			
		"aDefItem"
		InstanceAcc
		IntegerType
		pInteger
		#xA8
	))



(generate-instacc "DialogTemplate"
 (	"BoundsTop"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"BoundsLeft"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"BoundsBottom"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"BoundsRight"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"procID"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"visible"
		InstanceAcc
		BooleanType
		hBoolean
		#x0A
			
		"goAwayFlag"
		InstanceAcc
		BooleanType
		hBoolean
		#x0C
			
		"refcon"
		InstanceAcc
		PtrType
		hPointer
		#x0E
			
		"itemsID"
		InstanceAcc
		IntegerType
		hInteger
		#x12
			
		"title"
		InstanceAcc
		StringType
		hString
		#x14
	))



(generate-instacc "AlertTemplate"
 (	"BoundsTop"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"BoundsLeft"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"BoundsBottom"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"BoundsRight"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"items"
		InstanceAcc
		IntegerType
		hInteger
		#x08
	))







(generate-instacc "Int64Bit"
 (	"hilong"
		InstanceAcc
		FloatType
		hFloat
		#x00
			
		"lolong"
		InstanceAcc
		FloatType
		hFloat
		#x04
	))




(generate-instacc "AppFile"
 (	"vrefnum"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"ftype"
		InstanceAcc
		StringType
		hResType
		#x02
			
		"versnum"
		InstanceAcc
		IntegerType
		hByte
		#x06
			
		"fname"
		InstanceAcc
		StringType
		hString
		#x08
	))


(generate-instacc "ControlParam"
 (	"ioCRefNum"
		InstanceAcc
		IntegerType
		hInteger
		#x18
			
		"csCode"
		InstanceAcc
		IntegerType
		hInteger
		#x1A
			
		"csParam"
		InstanceAcc
		PtrType
		hPointer
		#x1C
			
		"csParam+2"
		InstanceAcc
		PtrType
		hPointer
		#x1E
			
		"csParam+4"
		InstanceAcc
		PtrType
		hPointer
		#x20
			
		"csParam+6"
		InstanceAcc
		PtrType
		hPointer
		#x22
			
		"csParam+8"
		InstanceAcc
		PtrType
		hPointer
		#x24
			
		"track"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
			
		"writeProt"
		InstanceAcc
		IntegerType
		hByte
		#x1E
			
		"diskInPlace"
		InstanceAcc
		IntegerType
		hByte
		#x1F
			
		"Installed"
		InstanceAcc
		IntegerType
		hByte
		#x20
			
		"sides"
		InstanceAcc
		IntegerType
		hByte
		#x21
			
		"qLink"
		InstanceAcc
		PtrType
		hPointer
		#x22
			
		"qType"
		InstanceAcc
		IntegerType
		hInteger
		#x26
			
		"dQDrive"
		InstanceAcc
		IntegerType
		hInteger
		#x28
			
		"dQRefNum"
		InstanceAcc
		IntegerType
		hInteger
		#x2A
			
		"dQFsid"
		InstanceAcc
		IntegerType
		hInteger
		#x2C
			
		"twoSideFmt"
		InstanceAcc
		IntegerType
		hByte
		#x2E
			
		"needsFlush"
		InstanceAcc
		IntegerType
		hByte
		#x2F
			
		"diskErrs"
		InstanceAcc
		IntegerType
		hInteger
		#x30
			
			
		"fXOn"
		InstanceAcc
		IntegerType
		hByte
		#x1C
			
		"fCTS"
		InstanceAcc
		IntegerType
		hByte
		#x1E
			
		"XOn"
		InstanceAcc
		CharType
		hChar
		#x1F
			
		"XOff"
		InstanceAcc
		CharType
		hChar
		#x20
			
		"errs"
		InstanceAcc
		IntegerType
		hByte
		#x21
			
		"evts"
		InstanceAcc
		IntegerType
		hByte
		#x22
			
		"fInX"
		InstanceAcc
		IntegerType
		hByte
		#x23
			
		"null"
		InstanceAcc
		IntegerType
		hByte
		#x24
			
			
		"cumErrs"
		InstanceAcc
		IntegerType
		hByte
		#x1C
			
		"xOffSent"
		InstanceAcc
		IntegerType
		hByte
		#x1E
			
		"rdPend"
		InstanceAcc
		IntegerType
		hByte
		#x1F
			
		"wrPend"
		InstanceAcc
		IntegerType
		hByte
		#x20
			
		"ctsHold"
		InstanceAcc
		IntegerType
		hByte
		#x21
			
		"xOffHold"
		InstanceAcc
		IntegerType
		hByte
		#x22
	))




(generate-instacc "DrvSts"
 (	"track"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"writeprot"
		InstanceAcc
		IntegerType
		hByte
		#x02
			
		"diskinplace"
		InstanceAcc
		IntegerType
		hByte
		#x03
			
		"installed"
		InstanceAcc
		IntegerType
		hByte
		#x04
			
		"sides"
		InstanceAcc
		IntegerType
		hByte
		#x05
			
		"qlink"
		InstanceAcc
		PtrType
		hPointer
		#x06
			
		"qtype"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"dqdrive"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"dqrefnum"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"dqfsid"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"twosidefmt"
		InstanceAcc
		IntegerType
		hByte
		#x12
			
		"needsflush"
		InstanceAcc
		IntegerType
		hByte
		#x13
			
		"diskerrs"
		InstanceAcc
		IntegerType
		hInteger
		#x14
	))



(generate-instacc "SerShk"
 (	"fxon"
		InstanceAcc
		IntegerType
		hByte
		#x00
			
		"fcts"
		InstanceAcc
		IntegerType
		hByte
		#x01
			
		"xon"
		InstanceAcc
		CharType
		hChar
		#x02
			
		"xoff"
		InstanceAcc
		CharType
		hChar
		#x03
			
		"errs"
		InstanceAcc
		IntegerType
		hByte
		#x04
			
		"evts"
		InstanceAcc
		IntegerType
		hByte
		#x05
			
		"finx"
		InstanceAcc
		IntegerType
		hByte
		#x06
	))




(generate-instacc "SerStaRec"
 (	"cumerrs"
		InstanceAcc
		IntegerType
		hByte
		#x00
			
		"xoffsent"
		InstanceAcc
		IntegerType
		hByte
		#x01
			
		"rdpend"
		InstanceAcc
		IntegerType
		hByte
		#x02
			
		"wrpend"
		InstanceAcc
		IntegerType
		hByte
		#x03
			
		"ctshold"
		InstanceAcc
		IntegerType
		hByte
		#x04
			
		"xoffhold"
		InstanceAcc
		IntegerType
		hByte
		#x05
	))



(generate-instacc "SysParmType"
 (	"valid"
		InstanceAcc
		IntegerType
		hByte
		#x00
			
		"aTalkA"
		InstanceAcc
		IntegerType
		hByte
		#x01
			
		"aTalkB"
		InstanceAcc
		IntegerType
		hByte
		#x02
			
		"config"
		InstanceAcc
		IntegerType
		hByte
		#x03
			
		"portA"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"portB"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"alarm"
		InstanceAcc
		FloatType
		hFloat
		#x08
			
		"font"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"kbdPrint"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"volclick"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"misc"
		InstanceAcc
		IntegerType
		hInteger
		#x12
	))


(generate-instacc "DateTimeRec"
 (	"year"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"month"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"day"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"hour"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"minute"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"second"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"dayofweek"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
	))



(generate-instacc "DecForm"
 (	"style"
		InstanceAcc
		IntegerType
		hByte
		#x00
			
		"digits"
		InstanceAcc
		IntegerType
		hInteger
		#x02
	))



(generate-instacc "Decimal"
 (	"sgn"
		InstanceAcc
		IntegerType
		hByte
		#x00
			
		"exp"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"sig"
		InstanceAcc
		StringType
		hString
		#x04
	))




(generate-instacc "FMInput"
 (	"Family"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"Size"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"Face"
		InstanceAcc
		IntegerType
		hByte
		#x04
			
		"NeedBits"
		InstanceAcc
		BooleanType
		hBoolean
		#x05
			
		"Device"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"NumerV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"NumerH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"DenomV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"DenomH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x0E
	))




(generate-instacc "FMOutput"
 (	"ErrNum"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"FontHandle"
		InstanceAcc
		HandleType
		hHandle
		#x02
			
		"Bold"
		InstanceAcc
		IntegerType
		hByte
		#x06
			
		"Italic"
		InstanceAcc
		IntegerType
		hByte
		#x07
			
		"UlOffset"
		InstanceAcc
		IntegerType
		hByte
		#x08
			
		"UlShadow"
		InstanceAcc
		IntegerType
		hByte
		#x09
			
		"UlThick"
		InstanceAcc
		IntegerType
		hByte
		#x0A
			
		"Shadow"
		InstanceAcc
		IntegerType
		hByte
		#x0B
			
		"Extra"
		InstanceAcc
		IntegerType
		hByte
		#x0C
			
		"Ascent"
		InstanceAcc
		IntegerType
		hByte
		#x0D
			
		"Descent"
		InstanceAcc
		IntegerType
		hByte
		#x0E
			
		"WidMax"
		InstanceAcc
		IntegerType
		hByte
		#x0F
			
		"Leading"
		InstanceAcc
		IntegerType
		hByte
		#x10
			
		"Unused"
		InstanceAcc
		IntegerType
		hByte
			
		"NumerV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x12
			
		"NumerH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x14
		#x11
			
		"DenomV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"DenomH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x18
	))



(generate-instacc "FMetricRec"
 ( 	"AscentI"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"AscentF"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"DescentI"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"DescentF"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"LeadingI"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"LeadingF"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"WidMaxI"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"WidMaxF"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"WTabHandle"
		InstanceAcc
		HandleType
		hHandle
		#x10
	))



(generate-instacc "Intl0Rec"
 (	"decimalPt"
		InstanceAcc
		CharType
		hChar
		#x00
			
		"thousSep"
		InstanceAcc
		CharType
		hChar
		#x01
			
		"listSep"
		InstanceAcc
		CharType
		hChar
		#x02
			
		"currSym1"
		InstanceAcc
		CharType
		hChar
		#x03
			
		"currSym2"
		InstanceAcc
		CharType
		hChar
		#x04
			
		"currSym3"
		InstanceAcc
		CharType
		hChar
		#x05
			
		"currFmt"
		InstanceAcc
		IntegerType
		hByte
		#x06
			
		"dateOrder"
		InstanceAcc
		IntegerType
		hByte
		#x07
			
		"shrtDateFmt"
		InstanceAcc
		IntegerType
		hByte
		#x08
			
		"dateSep"
		InstanceAcc
		CharType
		hChar
		#x09
			
		"timeCycle"
		InstanceAcc
		IntegerType
		hByte
		#x0A
			
		"timeFmt"
		InstanceAcc
		IntegerType
		hByte
		#x0B
			
		"mornStr"
		InstanceAcc
		StringType
		hResType
		#x0C
			
		"eveStr"
		InstanceAcc
		StringType
		hResType
		#x10
			
		"timeSep"
		InstanceAcc
		CharType
		hChar
		#x14
			
		"time1Stuff"
		InstanceAcc
		CharType
		hChar
		#x15
			
		"time2Stuff"
		InstanceAcc
		CharType
		hChar
		#x16
			
		"time3Stuff"
		InstanceAcc
		CharType
		hChar
		#x17
			
		"time4Stuff"
		InstanceAcc
		CharType
		hChar
		#x18
			
		"time5Stuff"
		InstanceAcc
		CharType
		hChar
		#x19
			
		"time6Stuff"
		InstanceAcc
		CharType
		hChar
		#x1A
			
		"time7Stuff"
		InstanceAcc
		CharType
		hChar
		#x1B
			
		"time8Stuff"
		InstanceAcc
		CharType
		hChar
		#x1C
			
		"metricSys"
		InstanceAcc
		IntegerType
		hByte
		#x1D
			
		"intl0Vers"
		InstanceAcc
		IntegerType
		hInteger
		#x1E
	))




(generate-instacc "Intl1Rec"
 (	"day1"
		InstanceAcc
		StringType
		hString
		#x00
			
		"day2"
		InstanceAcc
		StringType
		hString
		#x10
			
		"day3"
		InstanceAcc
		StringType
		hString
		#x20
			
		"day4"
		InstanceAcc
		StringType
		hString
		#x30
			
		"day5"
		InstanceAcc
		StringType
		hString
		#x40
			
		"day6"
		InstanceAcc
		StringType
		hString
		#x50
			
		"day7"
		InstanceAcc
		StringType
		hString
		#x60
			
		"month1"
		InstanceAcc
		StringType
		hString
		#x70
			
		"month2"
		InstanceAcc
		StringType
		hString
		#x80
			
		"month3"
		InstanceAcc
		StringType
		hString
		#x90
			
		"month4"
		InstanceAcc
		StringType
		hString
		#xA0
			
		"month5"
		InstanceAcc
		StringType
		hString
		#xB0
			
		"month6"
		InstanceAcc
		StringType
		hString
		#xC0
			
		"month7"
		InstanceAcc
		StringType
		hString
		#xD0
			
		"month8"
		InstanceAcc
		StringType
		hString
		#xE0
			
		"month9"
		InstanceAcc
		StringType
		hString
		#xF0
			
		"month10"
		InstanceAcc
		StringType
		hString
		#x100
			
		"month11"
		InstanceAcc
		StringType
		hString
		#x110
			
		"month12"
		InstanceAcc
		StringType
		hString
		#x120
			
		"suppressDay"
		InstanceAcc
		IntegerType
		hByte
		#x130
			
		"lngDateFmt"
		InstanceAcc
		IntegerType
		hByte
		#x131
			
		"dayLeading0"
		InstanceAcc
		IntegerType
		hByte
		#x132
			
		"abbrLen"
		InstanceAcc
		IntegerType
		hByte
		#x133
			
		"st0"
		InstanceAcc
		StringType
		hResType
		#x134
			
		"st1"
		InstanceAcc
		StringType
		hResType
		#x138
			
		"st2"
		InstanceAcc
		StringType
		hResType
		#x13C
			
		"st3"
		InstanceAcc
		StringType
		hResType
		#x140
			
		"st4"
		InstanceAcc
		StringType
		hResType
		#x144
			
		"intl1Vers"
		InstanceAcc
		IntegerType
		hInteger
		#x148
			
		"localRtn"
		InstanceAcc
		IntegerType
		hInteger
		#x14A
	))



(generate-instacc "THPrint"
 (	"iPrVersion"
		InstanceAcc
		IntegerType
		hInteger
		#x00
		
		"iDev"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"iVRes"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"iHRes"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"rpagetop"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"rpageleft"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"rpagebottom"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"rpageright"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"rpapertop"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"rpaperleft"
		InstanceAcc
		IntegerType
		hInteger
		#x12
			
		"rpaperbottom"
		InstanceAcc
		IntegerType
		hInteger
		#x14
			
		"rpaperright"
		InstanceAcc
		IntegerType
		hInteger
		#x16
		
		
		
			
		"wDev"
		InstanceAcc
		IntegerType
		hInteger
		#x18
			
		"iPageV"
		InstanceAcc
		IntegerType
		hInteger
		#x1A
			
		"iPageH"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
			
		"bPort"
		InstanceAcc
		IntegerType
		hByte
		#x1E
			
		"feed"
		InstanceAcc
		IntegerType
		hByte
		#x1F
		
		
		
			
		"prInfoPT1"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"prInfoPT2"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"prInfoPT3"
		InstanceAcc
		IntegerType
		hInteger
		#x24
			
		"prInfoPT4"
		InstanceAcc
		IntegerType
		hInteger
		#x26
			
		"prInfoPT5"
		InstanceAcc
		IntegerType
		hInteger
		#x28
			
		"prInfoPT6"
		InstanceAcc
		IntegerType
		hInteger
		#x2A
			
		"prInfoPT7"
		InstanceAcc
		IntegerType
		hInteger
		#x2C
		
		
		
			
		"iRowBytes"
		InstanceAcc
		IntegerType
		hInteger
		#x2E
			
		"iBandV"
		InstanceAcc
		IntegerType
		hInteger
		#x30
			
		"iBandH"
		InstanceAcc
		IntegerType
		hInteger
		#x32
			
		"iDevBytes"
		InstanceAcc
		IntegerType
		hInteger
		#x34
			
		"iBands"
		InstanceAcc
		IntegerType
		hInteger
		#x36
			
		"bPatScale"
		InstanceAcc
		IntegerType
		hByte
		#x38
			
		"bULThick"
		InstanceAcc
		IntegerType
		hByte
		#x39
			
		"bULOffset"
		InstanceAcc
		IntegerType
		hByte
		#x3A
			
		"scan"
		InstanceAcc
		IntegerType
		hByte
		#x3B
			
		"bXInfoX"
		InstanceAcc
		IntegerType
		hByte
		#x3C
		
		
		
			
		"iFstPage"
		InstanceAcc
		IntegerType
		hInteger
		#x3E
			
		"iLstPage"
		InstanceAcc
		IntegerType
		hInteger
		#x40
			
		"iCopies"
		InstanceAcc
		IntegerType
		hInteger
		#x42
			
		"bJDocLoop"
		InstanceAcc
		IntegerType
		hByte
		#x44
			
		"fFromApp"
		InstanceAcc
		IntegerType
		hByte
		#x45
			
		"pIdleProc"
		InstanceAcc
		PtrType
		hPointer
		#x46
			
		"pFileName"
		InstanceAcc
		PtrType
		hPointer
		#x4A
			
		"iFileVol"
		InstanceAcc
		IntegerType
		hInteger
		#x4E
			
		"bFileVers"
		InstanceAcc
		IntegerType
		hByte
		#x50
	))

;***---*** DB 7/29/88 Check with new Apple Doc
(generate-instacc "TPrStatus"
		  ("iTotPages"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x00

		   "iCurPage"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x02
		   
		   "iTotCopies"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x04
		   
		   "iCurCopy"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x06
		   
		   "iTotBands"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x08
		   
		   "iCurBand"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x0A
		   
		"fPgDirty"
		InstanceAcc
		BooleanType
		hBoolean
		#x0C
			
		"fImaging"
		InstanceAcc
		BooleanType
		hBoolean
		#x0D
		))


(generate-instacc "ListRec"
 (	"rViewTop"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"rViewLeft"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"rViewBottom"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"rViewRight"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"port"
		InstanceAcc
		PtrType
		hPointer
		#x08
			
		"indentV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"indentH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"cellSizeV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"cellSizeH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x12
			
		"visibleTop"
		InstanceAcc
		IntegerType
		hInteger
		#x14
			
		"visibleLeft"
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"visibleBottom"
		InstanceAcc
		IntegerType
		hInteger
		#x18
			
		"visibleRight"
		InstanceAcc
		IntegerType
		hInteger
		#x1A
			
		"vScroll"
		InstanceAcc
		HandleType
		hHandle
		#x1C
			
		"hScroll"
		InstanceAcc
		HandleType
		hHandle
		#x20
			
		"selFlags"
		InstanceAcc
		IntegerType
		hByte
		#x24
			
		"lActive"
		InstanceAcc
		BooleanType
		hBoolean
		#x25
			
		"listFlags"
		InstanceAcc
		IntegerType
		hByte
		#x27
			
		"clickTime"
		InstanceAcc
		FloatType
		hFloat
		#x28
			
		"clikLocV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x2C
			
		"clikLocH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x2E
			
		"mouseLocV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x30
			
		"mouseLocH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x32
			
		"lClikLoop"
		InstanceAcc
		PtrType
		hPointer
		#x34
			
		"lastClickV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x38
			
		"lastClickH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x3A
			
		"refCon"
		InstanceAcc
		PtrType
		hPointer
		#x3C
			
		"listDefProc"
		InstanceAcc
		HandleType
		hHandle
		#x40
			
		"userHandle"
		InstanceAcc
		HandleType
		hHandle
		#x44
			
		"dataBoundsTop"
		InstanceAcc
		IntegerType
		hInteger
		#x48
			
		"dataBoundsLeft"
		InstanceAcc
		IntegerType
		hInteger
		#x4A
			
		"dataBoundsBottom"
		InstanceAcc
		IntegerType
		hInteger
		#x4C
			
		"dataBoundsRight"
		InstanceAcc
		IntegerType
		hInteger
		#x4E
			
		"cells"
		InstanceAcc
		HandleType
		hHandle
		#x50
			
		"maxIndex"
		InstanceAcc
		IntegerType
		hInteger
		#x54
		))







(generate-instacc "ABusRec"
 (	"abOpCode"
		
		InstanceAcc
		IntegerType
		hInteger
		#x00

		"abResult"
		
		InstanceAcc
		IntegerType
		hInteger
		#x02

		"abUserReference"
		
		InstanceAcc
		FloatType
		hFloat
		#x04
		))







(generate-instacc "lapRec"
 (	"lapAddress.dstNodeID"
		
		InstanceAcc
		IntegerType
		hByte
		#x08
		
		"lapsAddress.srcNodeID"
		
		InstanceAcc
		IntegerType
		hByte
		#x09
		
		"lapAddress.lapProtType"
		
		InstanceAcc
		IntegerType
		hByte
		#x0A

		"lapReqCount"
		
		InstanceAcc
		IntegerType
		hInteger
		#x0C

		"lapActCount"
		
		InstanceAcc
		IntegerType
		hInteger
		#x0E

		"lapDataPtr"
		
		InstanceAcc
		PtrType
		hPointer
		#x10
		))








(generate-instacc "ddpRec"
 (	"ddpType"
		
		InstanceAcc
		IntegerType
		hByte
		#x08

		"ddpSocket"
		
		InstanceAcc
		IntegerType
		hByte
		#x09

		"ddpAddress.aNet"
		
		InstanceAcc
		IntegerType
		hInteger
		#x0A

		"ddpAddress.aNode"
		
		InstanceAcc
		IntegerType
		hByte
		#x0C

		"ddpAddress.aSocket"
		
		InstanceAcc
		IntegerType
		hByte
		#x0D

		"ddpReqCount"
		
		InstanceAcc
		IntegerType
		hInteger
		#x0E

		"ddpActCount"
		
		InstanceAcc
		IntegerType
		hInteger
		#x10

		"ddpDataPtr"
		
		InstanceAcc
		PtrType
		hPointer
		#x12

		"ddpNodeID"
		
		InstanceAcc
		IntegerType
		hByte
		#x16
		))








(generate-instacc "nbpRec"
 (	"nbpEntityPtr"
		
		InstanceAcc
		PtrType
		hPointer
		#x08

		"nbpBufPtr"
		
		InstanceAcc
		PtrType
		hPointer
		#x0C

		"nbpBufSize"
		
		InstanceAcc
		IntegerType
		hInteger
		#x10

		"nbpDataField"
		
		InstanceAcc
		IntegerType
		hInteger
		#x12

		"nbpAddress.aNet"
		
		InstanceAcc
		IntegerType
		hInteger
		#x14

		"nbpAddress.aNode"
		
		InstanceAcc
		IntegerType
		hByte
		#x16

		"nbpAddress.aSocket"
		
		InstanceAcc
		IntegerType
		hByte
		#x17

		"nbpRetransmitInfo.Interval"
		
		InstanceAcc
		IntegerType
		hByte
		#x18

		"nbpRetransmitInfo.Count"
		
		InstanceAcc
		IntegerType
		hByte
		#x19
		))







(generate-instacc "atpRec"
 (	"atpSocket"
		
		InstanceAcc
		IntegerType
		hByte
		#x09

		"atpAddress.aNet"
		
		InstanceAcc
		IntegerType
		hInteger
		#x0A

		"atpAddress.aNode"
		
		InstanceAcc
		IntegerType
		hByte
		#x0C

		"atpAddress.aSocket"
		
		InstanceAcc
		IntegerType
		hByte
		#x0D

		"atpReqCount"
		
		InstanceAcc
		IntegerType
		hInteger
		#x0E

		"atpDataPtr"
		
		InstanceAcc
		PtrType
		hPointer
		#x10

		"atpRspBDSPtr"
		
		InstanceAcc
		PtrType
		hPointer
		#x14

		"atpBitMap"
		
		InstanceAcc
		BooleanType
		hBoolean
		#x18

		"atpTransID"
		
		InstanceAcc
		IntegerType
		hInteger
		#x1A

		"atpActCount"
		
		InstanceAcc
		IntegerType
		hInteger
		#x1C

		"atpUserData"
		
		InstanceAcc
		FloatType
		hFloat
		#x1E

		"atpXO"
		
		InstanceAcc
		BooleanType
		hBoolean
		#x22

		"atpEOM"
		
		InstanceAcc
		BooleanType
		hBoolean
		#x23

		"atpTimeOut"
		
		InstanceAcc
		IntegerType
		hByte
		#x25

		"atpRetries"
		
		InstanceAcc
		IntegerType
		hByte
		#x27

		"atpNumBufs"
		
		InstanceAcc
		IntegerType
		hByte
		#x29

		"atpNumRsp"
		
		InstanceAcc
		IntegerType
		hByte
		#x2B

		"atpBDSSize"
		
		InstanceAcc
		IntegerType
		hByte
		#x2D  

		"atpRspUData"
		
		InstanceAcc
		FloatType
		hFloat
		#x2E

		"atpRspBuf"
		
		InstanceAcc
		PtrType
		hPointer
		#x32

		"atpRspSize"
		
		InstanceAcc
		IntegerType
		hInteger
		#x36
		))






(generate-instacc "LAPAddrBlock"
 (	"dstNodeID"
		
		InstanceAcc
		IntegerType
		hByte
		#x00

		"srcNodeID"
		
		InstanceAcc
		IntegerType
		hByte
		#x01

		"lapProtType"
		
		InstanceAcc
		IntegerType
		hByte
		#x02
		))







(generate-instacc "AddrBloack"
 (	"aNet"
		
		InstanceAcc
		IntegerType
		hInteger
		#x00

		"aNode"
		
		InstanceAcc
		IntegerType
		hByte
		#x02

		"aSocket"
		
		InstanceAcc
		IntegerType
		hByte
		#x03
		))






(generate-instacc "BDSElement"
 (	"buffSize0"
			
		InstanceAcc
		IntegerType
		hInteger
		#x00

		"buffPtr0"
			
		InstanceAcc
		PtrType
		hPointer
		#x02

		"dataSize0"
			
		InstanceAcc
		IntegerType
		hInteger
		#x06

		"userBytes0"
			
		InstanceAcc
		FloatType
		hFloat
		#x08

		"buffSize1"
			
		InstanceAcc
		IntegerType
		hInteger
		#x0C

		"buffPtr1"
			
		InstanceAcc
		PtrType
		hPointer
		#x0E

		"dataSize1"
			
		InstanceAcc
		IntegerType
		hInteger
		#x10

		"userBytes1"
			
		InstanceAcc
		FloatType
		hFloat
		#x12

		"buffSize2"
			
		InstanceAcc
		IntegerType
		hInteger
		#x16

		"buffPtr2"
			
		InstanceAcc
		PtrType
		hPointer
		#x18

		"dataSize2"
			
		InstanceAcc
		IntegerType
		hInteger
		#x1C

		"userBytes2"
			
		InstanceAcc
		FloatType
		hFloat
		#x20

		"buffSize3"
			
		InstanceAcc
		IntegerType
		hInteger
		#x24

		"buffPtr3"
			
		InstanceAcc
		PtrType
		hPointer
		#x26

		"dataSize3"
			
		InstanceAcc
		IntegerType
		hInteger
		#x2A

		"userBytes3"
			
		InstanceAcc
		FloatType
		hFloat
		#x2C

		"buffSize4"
			
		InstanceAcc
		IntegerType
		hInteger
		#x30

		"buffPtr4"
			
		InstanceAcc
		PtrType
		hPointer
		#x32

		"dataSize4"
			
		InstanceAcc
		IntegerType
		hInteger
		#x36

		"userBytes4"
			
		InstanceAcc
		FloatType
		hFloat
		#x38

		"buffSize5"
			
		InstanceAcc
		IntegerType
		hInteger
		#x3C

		"buffPtr5"
			
		InstanceAcc
		PtrType
		hPointer
		#x3E

		"dataSize5"
			
		InstanceAcc
		IntegerType
		hInteger
		#x42

		"userBytes5"
			
		InstanceAcc
		FloatType
		hFloat
		#x44

		"buffSize6"
			
		InstanceAcc
		IntegerType
		hInteger
		#x48

		"buffPtr6"
			
		InstanceAcc
		PtrType
		hPointer
		#x4A

		"dataSize6"
			
		InstanceAcc
		IntegerType
		hInteger
		#x4E

		"userBytes6"
			
		InstanceAcc
		FloatType
		hFloat
		#x50

		"buffSize7"
			
		InstanceAcc
		IntegerType
		hInteger
		#x54

		"buffPtr7"
			
		InstanceAcc
		PtrType
		hPointer
		#x56

		"dataSize7"
			
		InstanceAcc
		IntegerType
		hInteger
		#x5A

		"userBytes7"
			
		InstanceAcc
		FloatType
		hFloat
		#x5C
		))





(generate-instacc "EntityName"
 (	"objStr"
			
		InstanceAcc
		StringType
		hString
		#x00

		"typeStr"
			
		InstanceAcc
		StringType
		hString
		#x22

		"zoneStr"
			
		InstanceAcc
		StringType
		hString
		#x44

		))





(generate-instacc "CInfoPBRec"
	  (	"ioDrUsrWdsfrRecttop"     ;  DInfo
		InstanceAcc
		IntegerType
		hInteger
		32
			
		"ioDrUsrWdsfrRectleft"
		InstanceAcc
		IntegerType
		hInteger
		34

		"ioDrUsrWdsfrRectbottom"
		InstanceAcc
		IntegerType
		hInteger
		36

		"ioDrUsrWdsfrRectright"
		InstanceAcc
		IntegerType
		hInteger
		38

		"ioDrUsrWdsfrFlags"
		InstanceAcc
		IntegerType
		hInteger
		40

		"ioDrUsrWdsfrLocationv"
		InstanceAcc
		IntegerType
		hInteger
		42

		"ioDrUsrWdsfrLocationH"
		InstanceAcc
		IntegerType
		hInteger
		44

		"ioDrUsrWdsfrView"
		InstanceAcc
		IntegerType
		hInteger
		46

		"ioDrDrDirId"
		InstanceAcc
		FloatType
		hFloat
		48

		"ioDrNmFls"
		InstanceAcc
		IntegerType
		hInteger
		52

		"ioDrCrDat"
		InstanceAcc
		FloatType
		hFloat
		72

		"ioDrMdDat"
		InstanceAcc
		FloatType
		hFloat
		76

		"ioDrBkDat"
		InstanceAcc
		FloatType
		hFloat
		80

		"ioDrFndrInfofrScrollv"       ;  (DXInfo)
		InstanceAcc
		IntegerType
		hInteger
		84

		"ioDrFndrInfofrScrollh"
		InstanceAcc
		IntegerType
		hInteger
		86

		"ioDrFndrInfofrOpenChain"
		InstanceAcc
		FloatType
		hFloat
		88

		"ioDrFndrInfofrUnused"
		InstanceAcc
		IntegerType
		hInteger
		92

		"ioDrFndrInfofrComment"
		InstanceAcc
		IntegerType
		hInteger
		94

		"ioDrFndrInfofrPutAway"
		InstanceAcc
		FloatType
		hFloat
		96

		"ioDrParID"
		InstanceAcc
		FloatType
		hFloat
		100

		))



(generate-instacc "XPPParam"
 (	"abortSCBPtr"
		InstanceAcc
		PtrType
		hPointer
		#x1C
			
		"aspMaxCmdSize"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
			
		"aspQuantumSize"
		InstanceAcc
		IntegerType
		hInteger
		#x1E
			
		"numSesss"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"sessRefnum"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
			
		"aspTimeout"
		InstanceAcc
		IntegerType
		hByte
		#x1E
			
		"aspRetry"
		InstanceAcc
		IntegerType
		hByte
		#x1F
			
		"serverAddr.aNet"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"serverAddr.aNode"
		InstanceAcc
		IntegerType
		hByte
		#x22
			
		"serverAddr.aSocket"
		InstanceAcc
		IntegerType
		hByte
		#x23
			
		"scbPointer"
		InstanceAcc
		PtrType
		hPointer
		#x24
			
		"attnRoutine"
		InstanceAcc
		PtrType
		hPointer
		#x28
			
		"cbSize"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"cbPtr"
		InstanceAcc
		PtrType
		hPointer
		#x22
			
		"rbSize"
		InstanceAcc
		IntegerType
		hInteger
		#x26
			
		"rbPtr"
		InstanceAcc
		PtrType
		hPointer
		#x28
			
		"afpAddrBlock.aNet"
		InstanceAcc
		IntegerType
		hInteger
		#x2C
			
		"afpAddrBlock.aNode"
		InstanceAcc
		IntegerType
		hByte
		#x2E
			
		"afpAddrBlock.aSocket"
		InstanceAcc
		IntegerType
		hByte
		#x2F
			
		"afpSCBPtr"
		InstanceAcc
		PtrType
		hPointer
		#x30
			
		"afpAttnRoutine"
		InstanceAcc
		PtrType
		hPointer
		#x34
			
		"wdSize"
		InstanceAcc
		IntegerType
		hInteger
		#x2C
			
		"wdPtr"
		InstanceAcc
		PtrType
		hPointer
		#x2E
			
		"ccbStart"
		InstanceAcc
		IntegerType
		hInteger
		#x32
	))






(generate-instacc "MPPParam"
 (	"wdsPointer"
		InstanceAcc
		PtrType
		hPointer
		#x1E
			
		"protType"
		InstanceAcc
		IntegerType
		hByte
		#x1C
			
		"handler"
		InstanceAcc
		PtrType
		hPointer
		#x1E
			
		"socket"
		InstanceAcc
		IntegerType
		hByte
		#x1C
			
		"checksumFlag"
		InstanceAcc
		IntegerType
		hByte
		#x1D
			
		"listener"
		InstanceAcc
		PtrType
		hPointer
		#x1E
			
		"interval"
		InstanceAcc
		IntegerType
		hByte
		#x1C
			
		"count"
		InstanceAcc
		IntegerType
		hByte
		#x1D
			
		"entityPtr"
		InstanceAcc
		PtrType
		hPointer
		#x1E
			
		"verifyFlag"
		InstanceAcc
		IntegerType
		hByte
		#x22
			
		"retBuffPtr"
		InstanceAcc
		PtrType
		hPointer
		#x22
			
		"retBuffSize"
		InstanceAcc
		IntegerType
		hInteger
		#x26
			
		"maxToGet"
		InstanceAcc
		IntegerType
		hInteger
		#x28
			
		"numGotten"
		InstanceAcc
		IntegerType
		hInteger
		#x2A
			
		"confirmAddr.aNet"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"confirmAddr.aNode"
		InstanceAcc
		IntegerType
		hByte
		#x24
			
		"confirmAddr.aSocket"
		InstanceAcc
		IntegerType
		hByte
		#x25
			
		"newSocket"
		InstanceAcc
		IntegerType
		hByte
		#x26
			
		"newSelfFlag"
		InstanceAcc
		IntegerType
		hByte
		#x1C
			
		"oldSelfFlag"
		InstanceAcc
		IntegerType
		hByte
		#x1D
			
		"nKillQEl"
		InstanceAcc
		PtrType
		hPointer
		#x1C
		))
	




(generate-instacc "ATPParam"
 (	"userData"
		InstanceAcc
		FloatType
		hFloat
		#x12
			
		"reqTID"
		InstanceAcc
		IntegerType
		hByte
		#x16
			
		"atpSocket"
		InstanceAcc
		IntegerType
		hByte
		#x1C
			
		"atpFlags"
		InstanceAcc
		IntegerType
		hByte
		#x1D
			
		"addrBlock.aNet"
		InstanceAcc
		IntegerType
		hInteger
		#x1E
			
		"addrBlock.aNode"
		InstanceAcc
		IntegerType
		hByte
		#x20
			
		"addrBlock.aSocket"
		InstanceAcc
		IntegerType
		hByte
		#x21
			
		"reqLength"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"reqPointer"
		InstanceAcc
		PtrType
		hPointer
		#x24
			
		"bdsPointer"
		InstanceAcc
		PtrType
		hPointer
		#x28
			
		"numOfBuffs"
		InstanceAcc
		IntegerType
		hByte
		#x2C
			
		"timeOutVal"
		InstanceAcc
		IntegerType
		hByte
		#x2D
			
		"numOfResps"
		InstanceAcc
		IntegerType
		hByte
		#x2E
			
		"retryCount"
		InstanceAcc
		IntegerType
		hByte
		#x2F
			
		"intBuff"
		InstanceAcc
		IntegerType
		hInteger
		#x30
			
		"bdsSize"
		InstanceAcc
		IntegerType
		hByte
		#x2D
			
		"transID"
		InstanceAcc
		IntegerType
		hInteger
		#x2E
			
		"bitMap"
		InstanceAcc
		IntegerType
		hByte
		#x2C
			
		"rspNum"
		InstanceAcc
		IntegerType
		hByte
		#x2C
			
		"aKillQEl"
		InstanceAcc
		PtrType
		hPointer
		#x2C
		))
		
		





(generate-instacc "AFPCommandBlock"
 (	"cmdByte"
		InstanceAcc
		IntegerType
		hByte
		#x00
			
		"startEndFlag"
		InstanceAcc
		IntegerType
		hByte
		#x01
			
		"forkRefNum"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"rwOffset"
		InstanceAcc
		FloatType
		hFloat
		#x04
			
		"reqCount"
		InstanceAcc
		FloatType
		hFloat
		#x08
			
		"newLineFlag"
		InstanceAcc
		IntegerType
		hByte
		#x0C
			
		"newLineChar"
		InstanceAcc
		CharType
		hChar
		#x0D
		))






(generate-instacc "WDSElement"
 (	"entryLength"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"entryPtr"
		InstanceAcc
		PtrType
		hPointer
		#x02
		))





(generate-instacc "NamesTableEntry"
 (	"qLink"
		InstanceAcc
		PtrType
		hPointer
		#x00
			
		"nteAddress.aNet"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"nteAddress.aNode"
		InstanceAcc
		IntegerType
		hByte
		#x06
			
		"nteAddress.aSocket"
		InstanceAcc
		IntegerType
		hByte
		#x07
		))






(generate-instacc "CCrsr"
 (	"crsrType"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"crsrMap"
		InstanceAcc
		HandleType
		hHandle
		#x02	
			
		"crsrData"
		InstanceAcc
		HandleType
		hHandle
		#x06
			
		"crsrXData"
		InstanceAcc
		HandleType
		hHandle
		#x0A
			
		"crsrXValid"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"crsrXHandle"
		InstanceAcc
		HandleType
		hHandle
		#x10
			
		"crsr1Data0"
		InstanceAcc
		IntegerType
		hInteger
		#x14
			
		"crsr1Data1"
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"crsr1Data2"
		InstanceAcc
		IntegerType
		hInteger
		#x18
			
		"crsr1Data3"
		InstanceAcc
		IntegerType
		hInteger
		#x1A
			
		"crsr1Data4"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
			
		"crsr1Data5"
		InstanceAcc
		IntegerType
		hInteger
		#x1E
			
		"crsr1Data6"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"crsr1Data7"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"crsr1Data8"
		InstanceAcc
		IntegerType
		hInteger
		#x24
			
		"crsr1Data9"
		InstanceAcc
		IntegerType
		hInteger
		#x26
			
		"crsr1Data10"
		InstanceAcc
		IntegerType
		hInteger
		#x28
			
		"crsr1Data11"
		InstanceAcc
		IntegerType
		hInteger
		#x2A
			
		"crsr1Data12"
		InstanceAcc
		IntegerType
		hInteger
		#x2C
			
		"crsr1Data13"
		InstanceAcc
		IntegerType
		hInteger
		#x2E
			
		"crsr1Data14"
		InstanceAcc
		IntegerType
		hInteger
		#x30
			
		"crsr1Data15"
		InstanceAcc
		IntegerType
		hInteger
		#x32
			
		"crsrMask0"
		InstanceAcc
		IntegerType
		hInteger
		#x34
			
		"crsrMask1"
		InstanceAcc
		IntegerType
		hInteger
		#x36
			
		"crsrMask2"
		InstanceAcc
		IntegerType
		hInteger
		#x38
			
		"crsrMask3"
		InstanceAcc
		IntegerType
		hInteger
		#x3A
			
		"crsrMask4"
		InstanceAcc
		IntegerType
		hInteger
		#x3C
			
		"crsrMask5"
		InstanceAcc
		IntegerType
		hInteger
		#x3E
			
		"crsrMask6"
		InstanceAcc
		IntegerType
		hInteger
		#x40
			
		"crsrMask7"
		InstanceAcc
		IntegerType
		hInteger
		#x42
			
		"crsrMask8"
		InstanceAcc
		IntegerType
		hInteger
		#x44
			
		"crsrMask9"
		InstanceAcc
		IntegerType
		hInteger
		#x46
			
		"crsrMask10"
		InstanceAcc
		IntegerType
		hInteger
		#x48
			
		"crsrMask11"
		InstanceAcc
		IntegerType
		hInteger
		#x4A
			
		"crsrMask12"
		InstanceAcc
		IntegerType
		hInteger
		#x4C
			
		"crsrMask13"
		InstanceAcc
		IntegerType
		hInteger
		#x4E
			
		"crsrMask14"
		InstanceAcc
		IntegerType
		hInteger
		#x50
			
		"crsrMask15"
		InstanceAcc
		IntegerType
		hInteger
		#x52
						
		"crsrHotSpotV"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x54
			
		"crsrHotSpotH"			       ; MACTOOLBOX-1-7
		InstanceAcc
		IntegerType
		hInteger
		#x56
			
		"crsrXTable"
		InstanceAcc
		FloatType
		hFloat
		#x58
			
		"crsrID"
		InstanceAcc
		FloatType
		hFloat
		#x5C
		))

			  




(generate-instacc "CIcon"
 (	"iconPMap.baseAddr"
		InstanceAcc
		PtrType
		hPointer
		#x00
			
		"iconPMap.rowBytes"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"iconPMap.boundsTop"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"iconPMap.boundsLeft"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"iconPMap.boundsBottom"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"iconPMap.boundsRight"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"iconPMap.pmVersion"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"iconPMap.packType"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"iconPMap.packSize"
		InstanceAcc
		FloatType
		hFloat
		#x12
			
		"iconPMap.hRes"
		InstanceAcc
		FloatType
		hFloat
		#x16
			
		"iconPMap.vRes"
		InstanceAcc
		FloatType
		hFloat
		#x1A
			
		"iconPMap.pixelType"
		InstanceAcc
		IntegerType
		hInteger
		#x1E
			
		"iconPMap.pixelSize"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"iconPMap.cmpCount"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"iconPMap.cmpSize"
		InstanceAcc
		IntegerType
		hInteger
		#x24
			
		"iconPMap.planeBytes"
		InstanceAcc
		FloatType
		hFloat
		#x26
			
		"iconPMap.pmTable"
		InstanceAcc
		HandleType
		hHandle
		#x2A
			
		"iconPMap.pmReserved"
		InstanceAcc
		FloatType
		hFloat
		#x2E
		
			
		"iconMask.baseAddr"
		InstanceAcc
		PtrType
		hPointer
		#x32
			
		"iconMask.rowBytes"
		InstanceAcc
		IntegerType
		hInteger
		#x36
			
		"iconMask.boundstop"
		InstanceAcc
		IntegerType
		hInteger
		#x38
			
		"iconMask.boundsleft"
		InstanceAcc
		IntegerType
		hInteger
		#x3A
			
		"iconMask.boundsbottom"
		InstanceAcc
		IntegerType
		hInteger
		#x3C
			
		"iconMask.boundsright"
		InstanceAcc
		IntegerType
		hInteger
		#x3E
		
			
		"iconBMap.baseAddr"
		InstanceAcc
		PtrType
		hPointer
		#x40
			
		"iconBMap.rowBytes"
		InstanceAcc
		IntegerType
		hInteger
		#x44
			
		"iconBMap.boundstop"
		InstanceAcc
		IntegerType
		hInteger
		#x46
			
		"iconBMap.boundsleft"
		InstanceAcc
		IntegerType
		hInteger
		#x48
			
		"iconBMap.boundsbottom"
		InstanceAcc
		IntegerType
		hInteger
		#x4A
			
		"iconBMap.boundsright"
		InstanceAcc
		IntegerType
		hInteger
		#x4C
			
		"iconData"
		InstanceAcc
		HandleType
		hHandle
		#x4E
			
		"iconMaskData"
		InstanceAcc
		IntegerType
		hInteger
		#x52
		))
			





(generate-instacc "GDevice"
 (	"gdRefNum"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"gdID"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"gdType"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"gdITable"
		InstanceAcc
		HandleType
		hHandle
		#x06
			
		"gdResPref"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"gdSearchProc"
		InstanceAcc
		HandleType
		hHandle
		#x0C
			
		"gdCompProc"
		InstanceAcc
		HandleType
		hHandle
		#x10
			
		"gdFlags"
		InstanceAcc
		IntegerType
		hInteger
		#x14
			
		"gdPMap"
		InstanceAcc
		HandleType
		hHandle
		#x16
			
		"gdRefCon"
		InstanceAcc
		FloatType
		hFloat
		#x1A
			
		"gdNextGD"
		InstanceAcc
		HandleType
		hHandle
		#x1E
			
		"gdRectTop"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"gdRectLeft"
		InstanceAcc
		IntegerType
		hInteger
		#x24
			
		"gdRectBottom"
		InstanceAcc
		IntegerType
		hInteger
		#x26
			
		"gdRectRight"
		InstanceAcc
		IntegerType
		hInteger
		#x28
			
		"gdMode"
		InstanceAcc
		FloatType
		hFloat
		#x2A
			
		"gdCCBytes"
		InstanceAcc
		IntegerType
		hInteger
		#x2E
			
		"gdCCDepth"
		InstanceAcc
		IntegerType
		hInteger
		#x30
			
		"gdCCXData"
		InstanceAcc
		HandleType
		hHandle
		#x32
			
		"gdCCXMask"
		InstanceAcc
		HandleType
		hHandle
		#x36
			
		"gdReserved"
		InstanceAcc
		FloatType
		hFloat
		#x3A
		))


;(generate-instacc "ColorSpec"
; (	"value"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x00
			
;		"Red"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x02
			
;		"Green"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x04
			
;		"Blue"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x06
;		))






;(generate-instacc "ColorTable"
; (	"ctSeed"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x00	
			
;		"transIndex"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x04	
			
;		"ctSize"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x06	
			
;		"ctTable"
;		InstanceAcc
;		IntegerType
;		hInteger
;		#x08	
;		))







(generate-instacc "MatchTable"
 (	"mtRed"
		InstanceAcc
		IntegerType
		hInteger
		#x00	
			
		"mtGreen"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"mtBlue"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"matchData"
		InstanceAcc
		FloatType
		hFloat
		#x06
		))

	





(generate-instacc "GammaTbl"
 (	"gVersion"
		InstanceAcc
		IntegerType
		hInteger
		#x00	
			
		"gType"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"gFormulaSize"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"gChanCnt"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"gDataCnt"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"gDataWidth"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"gFormulaData"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
		))








(generate-instacc "GrafVars"
 (	"rgbOpColorRed"
		InstanceAcc
		IntegerType
		hInteger
		#x00	
			
		"rgbOpColorBlue"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"rgbOpColorGreen"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"rgbHiliteColorRed"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"rgbHiliteColorBlue"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"rgbHiliteColorGreen"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"pmFgColor"
		InstanceAcc
		HandleType
		hHandle
		#x0C
			
		"pmFgIndex"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"pmBkColor"
		InstanceAcc
		HandleType
		hHandle
		#x12
			
		"pmBkIndex"
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"pmFlags"
		InstanceAcc
		IntegerType
		hInteger
		#x18
		))


(generate-instacc "ITab"
 (	"iTabSeed"
		InstanceAcc
		FloatType
		hFloat
		#x00	
			
		"iTabRes"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"iTTable"
		InstanceAcc
		IntegerType
		hInteger
		#x06
		))


(generate-instacc "AuxWinRec"
 (	"awNext"
		InstanceAcc
		HandleType
		hHandle
		#x00	
			
		"awOwner"
		InstanceAcc
		PtrType
		hPointer
		#x04	
			
		"awCTable"
		InstanceAcc
		HandleType
		hHandle
		#x08
			
		"dialogCItem"
		InstanceAcc
		HandleType
		hHandle
		#x0C
			
		"awFlags"
		InstanceAcc
		FloatType
		hFloat
		#x10	
			
		"awReserved"
		InstanceAcc
		FloatType
		hFloat
		#x14
			
		"awRefCon"
		InstanceAcc
		FloatType
		hFloat
		#x18
		))

(generate-instacc "WinCTab"
 (	"wCSeed"
		InstanceAcc
		FloatType
		hFloat
		#x00
			
		"wCReserved"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"ctSize"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"content.value"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"content.red"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"content.green"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"content.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"frame.value"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"frame.red"
		InstanceAcc
		IntegerType
		hInteger
		#x12
			
		"frame.green"
		InstanceAcc
		IntegerType
		hInteger
		#x14
			
		"frame.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"text.value"
		InstanceAcc
		IntegerType
		hInteger
		#x18
			
		"text.red"
		InstanceAcc
		IntegerType
		hInteger
		#x1A
			
		"text.green"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
			
		"text.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x1E
			
		"hilite.value"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"hilite.red"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"hilite.green"
		InstanceAcc
		IntegerType
		hInteger
		#x24
			
		"hilite.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x26
			
		"titlebar.value"
		InstanceAcc
		IntegerType
		hInteger
		#x28
			
		"titlebar.red"
		InstanceAcc
		IntegerType
		hInteger
		#x2A
			
		"titlebar.green"
		InstanceAcc
		IntegerType
		hInteger
		#x2C
			
		"titlebar.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x2E
		))






(generate-instacc "CtlCTab"
 (	"ccSeed"
		InstanceAcc
		FloatType
		hFloat
		#x00
			
		"ccRider"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"ctSize"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"ctTable0.value"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"ctTable0.rgb.red"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"ctTable0.rgb.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"ctTable0.rgb.green"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"ctTable1.value"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"ctTable1.rgb.red"
		InstanceAcc
		IntegerType
		hInteger
		#x12
			
		"ctTable1.rgb.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x14
			
		"ctTable1.rgb.green"
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"ctTable2.value"
		InstanceAcc
		IntegerType
		hInteger
		#x18
			
		"ctTable2.rgb.red"
		InstanceAcc
		IntegerType
		hInteger
		#x1A
			
		"ctTable2.rgb.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
			
		"ctTable2.rgb.green"
		InstanceAcc
		IntegerType
		hInteger
		#x1E
			
		"ctTable3.value"
		InstanceAcc
		IntegerType
		hInteger
		#x20
			
		"ctTable3.rgb.red"
		InstanceAcc
		IntegerType
		hInteger
		#x22
			
		"ctTable3.rgb.blue"
		InstanceAcc
		IntegerType
		hInteger
		#x24
			
		"ctTable3.rgb.green"
		InstanceAcc
		IntegerType
		hInteger
		#x26
		))







(generate-instacc "AuxCtlRec"
 (	"acNext"
		InstanceAcc
		HandleType
		hHandle
		#x00	
			
		"acOwner"
		InstanceAcc
		HandleType
		hHandle
		#x04
			
		"acCTable"
		InstanceAcc
		HandleType
		hHandle
		#x08
			
		"acFlags"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"acReserved"
		InstanceAcc
		FloatType
		hFloat
		#x0E
			
		"acRefCon"
		InstanceAcc
		FloatType
		hFloat
		#x12
		))

(generate-instacc "MCEntry"
       (	"mctID"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"mctItem"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"mctRGB1Red"
		InstanceAcc
		IntegerType
		hInteger
		#x04
			
		"mctRGB1Green"
		InstanceAcc
		IntegerType
		hInteger
		#x06
			
		"mctRGB1Blue"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"mctRGB2Red"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"mctRGB2Green"
		InstanceAcc
		IntegerType
		hInteger
		#x0C
			
		"mctRGB2Blue"
		InstanceAcc
		IntegerType
		hInteger
		#x0E
			
		"mctRGB3Red"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"mctRGB3Green"
		InstanceAcc
		IntegerType
		hInteger
		#x12
			
		"mctRGB3Blue"
		InstanceAcc
		IntegerType
		hInteger
		#x14
			
		"mctRGB4Red"
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"mctRGB4Green"
		InstanceAcc
		IntegerType
		hInteger
		#x18
			
		"mctRGB4Blue"
		InstanceAcc
		IntegerType
		hInteger
		#x1A
			
		"mctReserved"
		InstanceAcc
		IntegerType
		hInteger
		#x1C
		))






(generate-instacc "SndCommand"
 (	"cmd"
		InstanceAcc
		IntegerType
		hInteger
		#x00	
			
		"param1"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"param2"
		InstanceAcc
		FloatType
		hFloat
		#x04
		))

(generate-instacc "ModifierStub"
 (	"next"
		InstanceAcc
		PtrType
		hPointer
		#x00		
			
		"modifierCode"
		InstanceAcc
		PtrType
		hPointer
		#x04
			
		"userInfo"
		InstanceAcc
		FloatType
		hFloat
		#x08
			
		"count"
		InstanceAcc
		FloatType
		hFloat
		#x0C
			
		"every"
		InstanceAcc
		FloatType
		hFloat
		#x10
		))

(generate-instacc "SndChannel"
 (	"nextChan"
		InstanceAcc
		PtrType
		hPointer
		#x00
			
		"firstMOD"
		InstanceAcc
		PtrType
		hPointer
		#x04
			
		"callBack"
		InstanceAcc
		PtrType
		hPointer
		#x08
			
		"userInfo"
		InstanceAcc
		FloatType
		hFloat
		#x0C
		))


(generate-instacc "SoundHeader"
 (	"samplePtr"
		InstanceAcc
		PtrType
		hPointer
		#x00	
			
		"length"
		InstanceAcc
		FloatType
		hFloat
		#x04	
			
		"sampleRate"
		InstanceAcc
		FloatType
		hFloat
		#x08
			
		"loopStart"
		InstanceAcc
		FloatType
		hFloat
		#x0C
			
		"loopEnd"
		InstanceAcc
		FloatType
		hFloat
		#x10
			
		"baseNote"
		InstanceAcc
		IntegerType
		hInteger
		#x14
		))

(generate-instacc "Block0"
 (	"sbSig"
		InstanceAcc
		IntegerType
		hInteger
		#x00		
			
		"sbBlkSize"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"sbBlkCount"
		InstanceAcc
		FloatType
		hFloat
		#x04
			
		"sbDevType"
		InstanceAcc
		IntegerType
		hInteger
		#x08
			
		"sbDevID"
		InstanceAcc
		IntegerType
		hInteger
		#x0A
			
		"sbData"
		InstanceAcc
		FloatType
		hFloat
		#x0C
			
		"sbDrvrCount"
		InstanceAcc
		IntegerType
		hInteger
		#x10
			
		"ddBlock"
		InstanceAcc
		FloatType
		hFloat
		#x12	
			
		"ddSize"
		InstanceAcc
		IntegerType
		hInteger
		#x16
			
		"ddType"
		InstanceAcc
		IntegerType
		hInteger
		#x18
		))







(generate-instacc "Partition"
 (	"pmSig"
		InstanceAcc
		IntegerType
		hInteger
		#x00
			
		"pmSigPad"
		InstanceAcc
		IntegerType
		hInteger
		#x02
			
		"pmMapBlkCnt"
		InstanceAcc
		FloatType
		hFloat
		#x04
			
		"pmPyPartStart"
		InstanceAcc
		FloatType
		hFloat
		#x08
			
		"pmPartBlkCnt"
		InstanceAcc
		FloatType
		hFloat
		#x0C
			
		"pmPartName"
		InstanceAcc
		StringType
		hString
		#x10
			
		"pmParType"
		InstanceAcc
		StringType
		hString
		#x30
			
		"pmLgDataStart"
		InstanceAcc
		FloatType
		hFloat
		#x50
			
		"pmDataCnt"
		InstanceAcc
		FloatType
		hFloat
		#x54
			
		"pmPartStatus"
		InstanceAcc
		FloatType
		hFloat
		#x58
			
		"pmLgBootStart"
		InstanceAcc
		FloatType
		hFloat
		#x5C
			
		"pmBootSize"
		InstanceAcc
		FloatType
		hFloat
		#x60
			
		"pmBootAddr"
		InstanceAcc
		FloatType
		hFloat
		#x64
			
		"pmBootAddr2"
		InstanceAcc
		FloatType
		hFloat
		#x68
			
		"pmBootEntry"
		InstanceAcc
		FloatType
		hFloat
		#x6C
			
		"pmBootEntry2"
		InstanceAcc
		FloatType
		hFloat
		#x70
			
		"pmBootCksum"
		InstanceAcc
		FloatType
		hFloat
		#x74
			
		"pmProcessor"
		InstanceAcc
		StringType
		hString
		#x78		
		))

;;;
;;; Some specials
;;;

(generate-instacc "FinderInfo"
		  ( "message"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x00

		   "count"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x02

		    "trap-channel"
		   InstanceAcc
		   IntegerType
		   hInteger
		   #x04))



