;;; -*- Mode:Common-Lisp; Package:MACTOOLBOX; Base:10 -*-

;;; For the puposes of the Texas Instruments Contract, this software is in
;;; TBI Category "B". Copyright (C) 1988, 1987, ExperTelligence Incorporated.

;/*                                                                       */
;/*                      RESTRICTED RIGHTS LEGEND                         */
;/*                                                                       */
;/* Use, duplication, or disclosure by the Government is subject to       */
;/* restrictions as set forth in subdivision (b)(3)(ii) of the Rights in  */
;/* Technical Data and Computer Software clause at 52.227-7013.           */
;/*                                                                       */
;/*                           EXPERTELLIGENCE                             */
;/*                            5638 HOLLISTER                             */
;/*                          GOLETA, CA  93117                            */
;/*                                                                       */
;/*  Copyright (C) 1988, 1987, EXPERTELLIGENCE Incorporated.              */
;/*  All rights reserved.                                                 */
;/*                                                                       */

;;; Modification History
;;; DB May 88
;;; 08/29/88 RWR Moved NOT IN ROMS to end of file 
;;; 08/24/88 BB&RR Added NOT IN ROM Routines
;;; 08/28/88 BB&RR Changed !SerHShake
;;; 08/30/88 RWR Changed GetIndString, deref
;;; 09/13/88 DWB&RWR Added Initialize-Toolbox-Constants function
;;; 09/16/88 DWB&RWR Added High Level Var traps
;;; 09/20/88 RWR Added Generic BlockMove
;;; 10/06/88 MAC Modified !TEFromScrap to dispose of its handle in an
;;;              UNWIND-PROTECT, signal a NIL new handle, and signal its OSErrs
;;;          MAC Modified !TEToScrap, !GetIndPattern, !GetIndString, !SetTime,
;;;              mac-string-to-mx-string, and !GetTime to UNWIND-PROTECT
;;;              locking of their handles
;;;          MAC removed redundant assignment to local variable in !TopMem,
;;;              !GZSaveHnd, !GetApplLimit, !MemError, !ApplicZone, !GetSysPPtr,
;;;              !GetEvQHdr, !GetFSQHdr, !GetVCBQHdr, !GetDrvQHdr, !GetVBLQHdr, and
;;;              !SystemZone. 
;;;          MAC modified !DriveStatus, !SerStatus, and !SerGetBuf  signal a
;;;              MAC-OSErr rather an anonomous ferror
;;; 11/15/88 MAC replaced resNotFound in calls to signal-oserr with !resNotFound
;;; 01/12/89 MAC Correction to 10/06/88 modification to !TEFromScrap: the handle
;;;              created in the function can NOT be disposed because it is stored
;;;              globally and used later.  Also, function should return NoErr or
;;;              OSErr but was returning NIL or T.
;;; 01/17/89 MAC Added GETICON function from MACTOOLBOX-1-7
;;; 01/18/89 MAC moved initialization and MAKE-OBSOLETE forms to TOOLBOX-INIT
;---------- Macintosh constants -----------------------------------

(defvar *system-dir-wd* nil "If non nil the working directory reference number of the system folder.")
(defvar *doubleTime*)
(defvar *caretTime*)
(defvar *screen-rect*)
(defvar nilhndl :unbound "A Handle instance used with traps that can take NIL or a valid Handle")
(defvar nilptr :unbound "A Pointer instance with zero, used by !NewWindow, etc.")
(defvar oneptr :unbound "A Pointer with -1 , used by !NewWindow :behind etc.")
(defvar nilrgn :unbound "A Region with 0, used by several toolbox region calls")
(defvar black :unbound "A Macintosh toolbox pattern")
(defvar ltgray :unbound "A Macintosh toolbox pattern")
(defvar dkgray :unbound "A Macintosh toolbox pattern")
(defvar white :unbound "A Macintosh toolbox pattern")
(defvar !black :unbound "A Macintosh toolbox pattern")
(defvar !ltgray :unbound "A Macintosh toolbox pattern")
(defvar !dkgray :unbound "A Macintosh toolbox pattern")
(defvar !white :unbound "A Macintosh toolbox pattern")


(defvar *APPLICATION-CHANNEL* :unbound "The default mac application identifier.")

(defun tb-transmit-packet (acb &optional (channel *application-channel*))
  (add:transmit-packet acb channel))

(defun tb-transmit-packet-and-wait (acb &optional (channel *application-channel*))
  (add:transmit-packet-and-wait acb channel))

;;; the following form was moved to TOOLBOX-INIT - 01/18/89 Mac
;;; (setf (send (add:find-channel 8) :max-queue-length) 10)

;Generate all of the macintosh traps
(create-mac-traps)

;Users can call !calls directly if they (use-package 'tb)
(export *mac-traps*)


;;; ***---*** NOT IN ROM Routines
;;;  RJB 8/23/88

(export '(deref !GetGrayRgn !TEGetScrapLen !TESetScrapLen !TEScrapHandle !TEFromScrap !TEToScrap !SystemZone !ApplicZone !MemError !GetApplLimit !GZSaveHnd !TopMem !DlgCut !DlgCopy !DlgPaste !DlgDelete !GetAlrtStage !ResetAlrtStage !SetDAFont !GetSysPPtr !SetTagBuffer !DiskEject !DriveStatus !SetEventMask !GetEvQHdr !SetWordBreak !SetClickLoop !InfoScrap !GetFSQHdr !GetVCBQHdr !GetDrvQHdr !StartSound !StopSound !SetSoundVol !GetSoundVol !SerReset !SerSetBuf !SerHShake !SerSetBrk !SerClrBrk !SerGetBuf !SerStatus !GetVBLQHdr !GetTime !SetTime !GetIndPattern !ScreenRes !SetFractEnable !GetMMUMode !GetIndString !CountAppFiles !GetDblTime !GetCaretTime !GetDateTime))

;  If !Secs2Date is changed, !GetTime should be changed too.

(defun deref (hnd &aux ptr)
  (setq ptr (make-instance 'mac-pointer :pointer 
		 (logand #xFFFFFF (fetch (make-instance 'mac-pointer :pointer 
				       (send hnd :handle)) 0))
		 ))
  ptr)

(defun !GetDblTime ()
  (setq *DoubleTime* (fetch (make-instance 'mac-pointer :pointer #x2F0) 0)))

(defun !GetCaretTime ()
  (setq *CaretTime* (fetch (make-instance 'mac-pointer :pointer #x2F4) 0)))

(defun !GetDateTime (&optional x &aux y)
  (setq y (fetch (make-instance 'mac-pointer :pointer #x20C) 0))
  (cond ((typep x 'locative)
	 (sys:%p-store-contents x y))
	(t y)))

(defun !GetGrayRgn (&aux rgn)
  (setq rgn (make-instance-no-init 'tb:Region))
  (send rgn :set-handle (fetch (make-instance 'mac-pointer :pointer #x9EE) 0))
  rgn)

(defun !TEGetScrapLen ()
  (fetchword (make-instance 'mac-pointer :pointer #xAB0) 0))

(defun !TESetScrapLen (length)
  (stowword (make-instance 'mac-pointer :pointer #xAB0) 0 length))

(defun !TEScrapHandle (&aux x)
  (setq x (make-instance 'mac-handle :handle 
		 (fetch (make-instance 'mac-pointer :pointer #xAB4) 0)))
  x)

(defun !TEFromScrap ()
  (let ((TEScrapH nil)
	(len      nil))				  ;positive scrap length or OSErr
	  (setf TEScrapH (or (!NewHandle 10)
			     (ferror 'nil-handle-error "!NewHandle return NIL")))
	  (setq len (GetScrap TEScrapH "TEXT"))
	  (if (plusp len)
	      ;; then it got what we want
	      (progn
		(stow (make-instance 'mac-pointer :pointer #xAB4) 0
		      (send TEScrapH :handle))	       
		(!TeSetScrapLen len)
		0)				       ; Return NoErr if it worked
	      ;; else LEN is really an OSErr
	      (if *signal-mac-oserr*
		  ;; then we are supposed to signal it
		  (signal-oserr len 'getscrap)

		  ;; else just return len (which is really a negative OSErr in this case)
		  len)))
  ;;; IMPLEMENTATION NOTE:  We cannot dispose of the TEScrapH handle even though this
  ;;; function is finished with it because the handle is being saved elsewhere for use
  ;;; later.
  )


(defun !TEToScrap (&aux hnd result)
  (!ZeroScrap)
  (setq hnd (!TeScrapHandle))
  (unwind-protect				  ;make sure handle is unlocked
      (progn
	(!HLock hnd)
	(setq result (!PutScrap (!TeGetScrapLen) "TEXT" (deref hnd))))
    (!HUnlock hnd))
  result)

(defun !SystemZone ()
  (make-instance 'mac-pointer :pointer
		 (fetch (make-instance 'mac-pointer :pointer #x2A6) 0)))

(defun !ApplicZone ()
  (make-instance 'mac-pointer :pointer
		 (fetch (make-instance 'mac-pointer :pointer #x2AA) 0)))

(defun !MemError ()
  (fetchword (make-instance 'mac-pointer :pointer #x220) 0))

(defun !GetApplLimit ()
  (make-instance 'mac-pointer :pointer
		 (fetch (make-instance 'mac-pointer :pointer #x130) 0)))

(defun !GZSaveHnd ()
  (make-instance 'mac-handle :handle
			 (fetch (make-instance 'mac-pointer :pointer #x328) 0)))

(defun !TopMem ()
  (make-instance 'mac-pointer :pointer 
			 (fetch (make-instance 'mac-pointer :pointer #x108) 0)))

(defun !DlgCut (theDialog)
  (cond ((<= 0 (send theDialog :editField))
	 (!TECut (send theDialog :textH)))))

(defun !DlgCopy (theDialog)
  (cond ((<= 0 (send theDialog :editField))
	 (!TECopy (send theDialog :textH)))))

(defun !DlgPaste (theDialog)
  (cond ((<= 0 (send theDialog :editField))
	 (!TEPaste (send theDialog :textH)))))

(defun !DlgDelete (theDialog)
  (cond ((<= 0 (send theDialog :editField))
	 (!TEDelete (send theDialog :textH)))))

(defun !GetAlrtStage ()
  (fetchword (make-instance 'mac-pointer :pointer #xA9A) 0))

(defun !ResetAlrtStage ()
  (stowword (make-instance 'mac-pointer :pointer #xA9A) 0 -1))

(defun !SetDAFont (fontNum)
  (stowword (make-instance 'mac-pointer :pointer #xAFA) 0 fontNum))

(defun !GetSysPPtr ()
  (make-instance 'mac-pointer :pointer #x1F8))

(defun !SetTagBuffer (buffPtr &aux paramBlock)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum -5)
	(send paramBlock :csCode 8)
	(send paramBlock :csParam buffPtr)
	(!Control paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !DiskEject (drvNum &aux paramBlock)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum -5)
	(send paramBlock :ioVRefNum drvNum)
	(send paramBlock :csCode 7) 
	(!Control paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !DriveStatus (drvNum &aux driveStatus paramBlock result)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum -5)
	(send paramBlock :ioVRefNum drvNum)
	(send paramBlock :csCode 8)
	(setq result (!Status paramBlock))
	(cond ((zerop result)
	       (setq driveStatus (make-instance 'DrvSts))
	       (send driveStatus :track (send paramBlock :track))
	       (send driveStatus :writeProt (send paramBlock :writeProt))
	       (send driveStatus :diskInPlace (send paramBlock :diskInPlace))
	       (send driveStatus :installed (send paramBlock :installed))
	       (send driveStatus :sides (send paramBlock :sides))
	       (send driveStatus :qLink (send paramBlock :qLink))
	       (send driveStatus :qType (send paramBlock :qType))
	       (send driveStatus :dQDrive (send paramBlock :dQDrive))
	       (send driveStatus :dQRefNum (send paramBlock :dQRefNum))
	       (send driveStatus :dQFsid (send paramBlock :dQFsid))
	       (send driveStatus :twoSideFmt (send paramBlock :twoSideFmt))
	       (send driveStatus :needsFlush (send paramBlock :needsFlush))
	       (send driveStatus :diskErrs (send paramBlock :diskErrs))
	       driveStatus)
	      (*signal-mac-oserr*
	       (signal-oserr result '!DriveStatus))
	      (t result)))
    (when paramBlock (send paramBlock :dispose))))

(defun !SetEventMask (theMask)
  (stowword (make-instance 'mac-pointer :pointer #x144) 0 theMask))

(defun !GetEvQHdr ()
  (make-instance 'mac-pointer :pointer #x14A))

(defun !SetWordBreak (wBrkProc hTE)
  (cond ((instancep hTE)
	 (send hTE :wordbreak wBrkProc))))

(defun !SetClickLoop (clickProc hTE)
  (cond ((instancep hTE)
	 (send hTE :clickLoop clickProc))))


(defun !InfoScrap ()
  (make-instance 'tb:scrapstuff))

(defun !GetFSQHdr ()
  (make-instance 'mac-pointer :pointer #x360))

(defun !GetVCBQHdr ()
  (make-instance 'mac-pointer :pointer #x356))

(defun !GetDrvQHdr ()
  (make-instance 'mac-pointer :pointer #x308))

(defun !StartSound (synthRec numBytes completionRtn &aux paramBlock)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'ioParam))
	(send paramBlock :ioRefNum -4)
	(send paramBlock :ioBuffer synthRec)
	(send paramBlock :ioReqCount numBytes)
	(send paramBlock :ioCompletion completionRtn)
	(!Write paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !StopSound (&aux paramBlock)
  (unwind-protect
      (progn
	(stowword (make-instance 'mac-pointer :pointer #x280) 0 0)
	(setq paramBlock (make-instance 'ioParam))
	(send paramBlock :ioRefNum -4)
	(!KillIO paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !SetSoundVol (vol)
  (stowbyte (make-instance 'mac-pointer :pointer #x260) 0 (LogAnd vol 7)))

(defun !GetSoundVol ()
  (fetchbyte (make-instance 'mac-pointer :pointer #x260) 0))

(defun !SerReset (refNum serConfig &aux paramBlock)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum refNum)
	(send paramBlock :csCode 8)
	(stowwordhandle paramBlock #x1C serConfig)
	(!Control paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !SerSetBuf (refNum serBPtr serBLen &aux paramBlock)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum refNum)
	(send paramBlock :csCode 9)
	(send paramBlock :csParam serBPtr)
	(stowwordhandle paramBlock #x20 serBLen)
	(!Control paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !SerHShake (refNum flags &aux paramBlock)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum refNum)
	(send paramBlock :csCode 10)
	(send paramBlock :fXOn (send flags :fXOn))
	(send paramBlock :fCTS (send flags :fCTS))
	(send paramBlock :xOn (send flags :xOn))
	(send paramBlock :xOff (send flags :xOff))
	(send paramBlock :errs (send flags :errs))
	(send paramBlock :evts (send flags :evts))
	(send paramBlock :fInX (send flags :fInX))
	(!Control paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !SerSetBrk (refNum &aux paramBlock)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum refNum)
	(send paramBlock :csCode 12)
	(!Control paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !SerClrBrk (refNum &aux paramBlock)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum refNum)
	(send paramBlock :csCode 11)
	(!Control paramBlock))
    (when paramBlock (send paramBlock :dispose))))

(defun !SerGetBuf (refNum &aux paramBlock result)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum refNum)
	(send paramBlock :csCode 2)
	(setq result (!Status paramBlock))
	(cond ((zerop result)
	       (send (send paramBlock :csParam) :pointer))
	      (*signal-mac-oserr*
	       (signal-oserr result '!SerGetBuf))
	      (t result)))
    (when paramBlock (send paramBlock :dispose))))

(defun !SerStatus (refNum &aux paramBlock result statusRec)
  (unwind-protect
      (progn
	(setq paramBlock (make-instance 'controlParam))
	(send paramBlock :ioCRefNum refNum)
	(send paramBlock :csCode 8)
	(setq result (!Status paramBlock))
	(cond ((zerop result)
	       (setq statusRec (make-instance 'SerStaRec))
	       (send statusRec :cumErrs (send paramBlock :cumerrs))
	       (send statusRec :xOffSent (send paramBlock :xOffSent))
	       (send statusRec :rdPend (send paramBlock :rdPend))
	       (send statusRec :wrPend (send paramBlock :wrPend))
	       (send statusRec :ctsHold (send paramBlock :ctsHold))
	       (send statusRec :xOffHold (send paramBlock :xOffHold))
	       statusRec)
	      (*signal-mac-oserr*
	       (signal-oserr result '!SerStatus))
	      (t result)))
    (when paramBlock (send paramBlock :dispose))))

(defun !GetVBLQHdr ()
  (make-instance 'mac-pointer :pointer #x160))

(defun !GetTime (&aux date)
  (setq date (make-instance 'DateTimeRec))
  (unwind-protect				  ;make sure handle gets unlocked
      (progn
	(!HLock date)
	(!Secs2Date (!GetDateTime) (deref date)))
    (!HUnlock date))
  date)

(defun !SetTime (date &aux secs)
  (unwind-protect				  ;make sure handle gets unlocked
      (progn
	(!HLock date)
	(setq secs (!Date2Secs (deref date))))
    (!HUnlock date))
  (!SetDateTime secs))

(defun !GetIndPattern (patListID index &aux theRes theResPtr result startPtr)
  (setq result (make-instance 'pattern))
  (setq theRes (GetResource "PAT#" patListID))
  (unwind-protect				  ;make sure handle gets unlocked
      (progn
	(!HLock theRes)
	(setq theResPtr (deref theRes))
	(setq startPtr (make-instance 'mac-pointer :pointer 
				      (+ (send theResPtr :pointer) 
					 (* (- index 1) 8) 2)))
	(send result :one (fetchword startPtr 0))
	(send result :two (fetchword startPtr 2))
	(send result :three (fetchword startPtr 4))
	(send result :four (fetchword startPtr 6)))
    (!Hunlock theRes))
  result)

(defun !ScreenRes (&aux hRes vRes ptr)
  (setq ptr (make-instance 'mac-pointer :pointer #x102))
  (setq vRes (fetchword ptr 0))
  (setq hRes (fetchword ptr 2))
  (values vRes hRes))

(defun !SetFractEnable (fractEnable &aux f)
  (cond (fractEnable
	 (setq f 1))
	(t
	 (setq f 0)))
  (stowword (make-instance 'mac-pointer :pointer #xBF4) 0 f))

(defun !GetMMUMode ()
  (fetchword (make-instance 'mac-pointer :pointer #xCB2) 0))

(defun !GetIndString (strListID index &aux strres result offset len str)
  (setq strRes (GetResource "STR#" strListID))
  (cond ((or (> 1 index)
	     (> index (fetchwordhandle strRes 0)))
	 (setq result ""))
	(t
	 (unwind-protect			  ;make sure handle gets unlocked
	     (progn
	       (!HLock strRes)
	       (setq result (deref strRes))
	       (setq offset 2)
	       (while (/= 1 index)
		      (setq offset (+ offset (fetchbyte result offset) 1))
		      (setq index (1- index)))
	       (send result :set-pointer (+ (send result :pointer) offset))
	       (setq len (fetchbyte result 0))
	       (setq str (make-string len))
	       (dotimes (i len)
		 (setf (char str i) (character (fetchbyte result (1+ i))))))
	   (!HunLock strres))
	 (setq result str)
	 ))
  result)

(defun !CountAppFiles (&aux message count ptr1 ptr2 fndrInfo)
  (setq ptr1 (make-instance 'mac-pointer :pointer #x904))
  (setq ptr2 (make-instance 'mac-pointer :pointer (fetch ptr1 0)))
  (setq fndrInfo (make-instance 'mac-handle :handle (fetch ptr2 16)))
  (setq message (fetchwordhandle fndrInfo 0))
  (setq count (fetchwordhandle fndrInfo 2))
  (values message count))


;These are used a lot
(export '(nilptr oneptr nilrgn var nilhndl))

(defun Initialize-ToolBox-Constants ()
  (LET ((*application-channel* 8))
    (setf *system-dir-wd* (tb:FetchWord (make-instance 'tb:mac-pointer :pointer #x210)))
    (setq nilptr (make-instance 'mac-pointer :pointer 0))      ;defer init
    (setq oneptr (make-instance 'mac-pointer :pointer #xFFFFFF))       ;defer init
    (setq nilhndl (make-instance 'mac-handle :handle 0))
    (setq nilrgn (make-instance-no-init 'region))      ;defer init
    (setq black  (let ((i (make-instance 'pattern)))
		   (dolist (x '(:one :two :three :four) i)
		     (send i x -1 ))))		       ;defer init
    (setq !black black)
    (setq ltgray  (let ((i (make-instance 'pattern)))
		    (dolist (x '(:one :two :three :four) i)
		      (send i x -30686))))	       ;defer init
    (setq !ltgray ltgray)
    (setq dkgray  (let ((i (make-instance 'pattern)))
		    (dolist (x '(:one :two :three :four) i)
		      (send i x #x77DD))))	       ;defer init
    (setq !dkgray dkgray)
    (setq white  (make-instance 'pattern))
    (setq !white white)
    
    (setq *screen-rect* (let ((gp (make-instance 'grafport)))
			    (!GetWMgrPort gp)
			    (send gp :portRect)))
    
    (setq *DoubleTime* (fetch (make-instance 'mac-pointer :pointer #x2F0) 0))
    
    (setq *CaretTime* (fetch (make-instance 'mac-pointer :pointer #x2F4) 0))
    
    ))

(defun Clear-ToolBox-Constants ()
  (LET ((*application-channel* 8))
    (setf *system-dir-wd* nil)

    (send black :dispose)
    (setf black nil)

    (send ltgray :dispose)
    (setf ltgray nil)

    (send dkgray :dispose)
    (setf dkgray nil)

    (send white :dispose)
    (setf white nil)

    (setf *screen-rect* nil)
    
    (setq *DoubleTime* nil)
    
    (setq *CaretTime* nil)
    
    ))

;; A useful macro.
(defmacro with-system-directory (wdir &body body)
  "Excutes body with the working directory set to wdir. Wdir is a working directory reference number."
  `(let ((parm-block (make-instance 'tb:WDPbRec))
	 old-wd)
     (unwind-protect
	 (progn
	   (send parm-block :ioCompletion tb:nilptr)
	   (send parm-block :ioNamePtr tb:NilPtr)
	 
	   (tb:!GetVol parm-block)
	   (setf old-wd (send parm-block :IoVRefNum))
	   
	   (setf (send parm-block :ioVRefNum) ,wdir)

	   (tb:!SetVol parm-block)
	   
	   ,@body)
       (setf (send parm-block :ioVRefNum) old-wd)
       (tb:!SetVol parm-block)
       (send parm-block :dispose))))

;;; the following form was moved to TOOLBOX-INIT - 01/18/89 Mac
;;; (add-initialization "Clear toolbox constants" '(clear-toolbox-constants)
;;;		    :before-cold)

;;; Modified VAR Traps

(defun PTTOANGLE (r p &aux angle)
  (declare (arglist rect point) (values angle))
  (!PtToAngle r p (VAR angle))
  angle)

(defun GETFNUM (fontname &aux num)
  (declare (arglist fontname) (values fontnumber))
  (!GetFNum fontname (VAR num))
  num)

(export '(GetIcon))				       ; MACTOOLBOX-1-7
(defun GetIcon (iconID &aux res)
    "This version of GetIcon checks for nil handles.  It signals resNotFound
   instead of just returning a nil handle."
    (setq res (!GetIcon iconID))
  (if (zerop (send res :handle))
      (signal-oserr '!resNotFound 'GetIcon
		    "GetResource did not find icon ~s" iconID)) 
      res)


(defun GETITEMICON (menu item &aux icon)
  (declare (arglist menuInfo-inst item) (values icon-id))
  (!GetItemIcon menu item (VAR icon))
  icon)

(defun GETITEMSTYLE (menu item &aux style)
  (declare (arglist menuInfo-inst item) (values style))
  (!GetItemStyle menu item (VAR style))
  style)

(defun GETITEMMARK (menu item &aux chr)
  (declare (arglist menuInfo-inst item) (values character))
  (!GetItemMark menu item (VAR chr))
  chr)

(defun DIALOGSELECT (event &aux ptr item result)
  (declare (arglist eventrecord-inst)(values boolean dialog-pointer item-hit))
  (setq result (!DialogSelect event (VAR ptr) (VAR item)))
  (values result ptr item))

(defun GETDITEM (dialog item
		 &aux typ
		 (itemh (make-instance 'mac-handle))
		 (box (make-instance 'rect)))
  (declare (arglist DialogRecord-inst item)(values type item-handle box)) 
  (!GetDItem dialog item (VAR typ) itemh box)
  (values typ itemh box))

(defun MODALDIALOG (&optional (proc-ptr nilptr) &aux item)
  (declare (values item))
  (!ModalDialog proc-ptr (VAR item))
  item)

(defun GETRESINFO (handle &aux int res str)
  (declare (arglist handle)(values  integer-ID resType-string name-string))
  (!GetResinfo handle (VAR int) (VAR res) (VAR str))
  (values int res str)
  )

(defun GETSCRAP (handle restype &aux longint)
  (declare (values len-or-error-code))
  (!GetScrap handle restype (VAR longint)))

(defun SCSICOMPLETE (wait &aux status message oserr)
  (declare (values oserr status message))
  (setq oserr (!ScsiComplete (VAR status) (VAR message) wait))
  (values oserr status message))

(defun LGETCELL (ptr len cell han)
  (declare (arglist data-pointer data-length point List-handle)
	   (values data-length))
  (!LGetCell ptr (VAR len) cell han)
  len)

(defun LFIND (cell lhan &aux offset len)
  (declare (arglist cell listrec-inst) (values offset length))
  (!LFind (VAR offset) (VAR len) cell lhan)
  (values offset len))

(defun GETITEMCMD (menu item &aux chr)
  (declare (arglist menuinfo-inst item)(values character))
  (!GetItemCmd menu item (VAR chr))
  chr)

(defun GETENTRYUSAGE (palette srcentry &aux destusage desttol)
  (declare (arglist palette src-entry) (values dest-usage dest-tolerance))
  (!GetEntryUsage palette srcentry (VAR destusage) (VAR desttol))
  (values destusage desttol))

;;; These involve strings

(defun GETFONTNAME (number &aux str)
  (declare (arglist font-number)(values font-name))
  (!GetFontName number (VAR str))
  str)

(defun GETWTITLE (window &aux str)
  (declare (arglist window-inst)(values window-title))
  (!GetWTitle window (VAR str))
  str)

(defun GETITEM (menu item &aux str)
  (declare (arglist menuinfo-inst item)(values item-string))
  (!GetITem menu item (VAR str))
  str)

(defun GETCTITLE (control &aux str)
  (declare (arglist controlRecord-inst)(values control-title))
  (!GetCTitle control (VAR str))
  str)

(defun GETITEXT (item-handle &aux str)
  (declare (values string))
  (!GetIText item-handle (VAR str))
  str)

(defun GETINDTYPE (count &aux str)
  (declare (values restype))
  (!GetIndType (VAR str) count)
  str)

(defun GET1INDTYPE (count &aux str)
  (declare (values restype))
  (!Get1IndType (VAR str) count)
  str)

(defun IUDATESTRING (date-time form &aux str)
  (declare (values string))
  (!IUDateString date-time form (VAR str))
  str)

(defun IUDATEPSTRING (date-time form intl-param-handle &aux str)
  (declare (values string))
  (!IUDatePString date-time form (VAR str) intl-param-handle)
  str)

(defun IUTIMESTRING (date-time &optional want-seconds-boolean &aux str)
  (declare (values string))
  (!IUTimeString date-time want-seconds-boolean (VAR str))
  str)

(defun IUTIMEPSTRING (date-time want intl-param-handle &aux str)
  (declare (arglist want-seconds-boolean intl-param-handle) (values string))
  (!IUTimePString date-time want (VAR str) intl-param-handle)
  str)

(defun FindWindow (point &aux window result)
  (declare (arglist point)(values window-region-number window-inst))
  (setq result (!FindWindow point (VAR window)))
  (values result window))

(defun FindControl (point Window-inst &aux c-inst result)
  (declare (arglist point Window-inst) (values part-code controlRecord-inst))
  (setq result (!FindControl point window-inst (VAR c-inst)))
  (values result c-inst))

(defun GetWMgrPort (&aux (grafport (make-instance-no-init 'grafport)))
  (declare (values grafport))
  (!GetWMgrPort grafport)
  grafport)

(defun GetCWMgrPort (&aux (grafport (make-instance-no-init 'cgrafport)))
  (declare (values grafport))
  (!GetCWMgrPort grafport)
  grafport)


(defun Getport (&aux (cgrafport (make-instance-no-init 'cgrafport)))
  "This returns the current grafport, if it is a window it returns a window
   instance which is EQ to the one returned when it was created"
  (declare (values cgrafport-or-window-inst))
  (!GetPort cgrafport)
  (cond ((cdr (assoc (send cgrafport :pointer) *mac-window-list*)))
	(t cgrafport))
  )


(defun GetResource (typ num &aux res)
  "This version of !GetResource checks for nil handles.  It signals resNotFound
   instead of just returning a nil handle."
  (setq res (!GetResource typ num))
  (if (zerop (send res :handle))
      (signal-oserr '!resNotFound 'GetResource
		    "GetResource did not find ~s resource of ID= ~s" typ num)
      res))

(defun Get1Resource (typ num &aux res)
  "This version of !Get1Resource checks for nil handles.  It signals resNotFound
   instead of just returning a nil handle."
  (setq res (!Get1Resource typ num))
  (if (zerop (send res :handle))
      (signal-oserr '!resNotFound 'Get1Resource
		    "Get1Resource did not find ~s resource of ID= ~s" typ num)
      res))

(defun RGetResource (typ num &aux res)
  "This version of !RGetResource checks for nil handles.  It signals resNotFound
   instead of just returning a nil handle."
  (setq res (!RGetResource typ num))
  (if (zerop (send res :handle))
      (signal-oserr '!resNotFound 'RGetResource
		    "RGetResource did not find ~s resource of ID= ~s" typ num)
      res))



;;; Generic Blockmove
(export '(mac-pointer-or-handle BLOCK-MOVE makeloc address mx-string-to-mac-string
				mac-string-to-mx-string))
(defun ADDR-32-TO-24 (addr)
  (logior (ash (logand #xf000000 addr) -4) (logand #xFFFFF addr)))

(defun mac-pointer-or-handle (object)
  (or (typep object 'mac-pointer) (typep object 'mac-handle)))

(defun BLOCK-MOVE (src dest count)
  "Move count elements from src to destination. 
   Src and Dest can be a lisp array, a mac-pointer, or a mac-handle. 
    Note that count is in elements and not in bytes.
   Count is the number of elements to be transfered.

   The data that is moved is correctly byte swapped for its element size. The 
   element size is taken from the the lisp array type. In the Mac to Mac case 
   it is assumed to be 8.

   **Beware** There is no bounds checking on mac-pointers and mac-handles so 
    you are free to copy over anything you want."
  ;; Note that all Mac addresses are 24 bit addresses.
   
  (let (byte-count
	acb
	direction
	data-size)
    
    (setf direction (cond ((and (arrayp src)
				(arrayp dest))
			   :exp-to-exp)
			  ((and (mac-pointer-or-handle src)
				(mac-pointer-or-handle dest))
			   :mac-to-mac)
			  
			  ((and (mac-pointer-or-handle src)
				(arrayp dest))
			   (setf data-size (array-element-size dest))
			   :mac-to-exp)
			  ((and (mac-pointer-or-handle dest)
				(arrayp src))
			   (setf data-size (array-element-size src))
			   :exp-to-mac)
			  (t
			   (ferror 'invalid-datatypes
				   "Invalid data types were passed to block-move.")
			   )))
    (setf byte-count (ecase data-size
		       (8 count)
		       (16 (* count 2))
		       (32 (* count 4))))
    (case direction
      (:mac-to-mac
       (!BlockMove (send src :mac-pointer)
		   (send dest :mac-pointer)
		   count))
      (:exp-to-exp
       (si:copy-array-portion src 0 count dest 0 count))
      (:mac-to-exp
       
       (setf acb (get-acb-fast byte-count))

       (ecase data-size
	 (8
	  (!blockmove
	    (send src :mac-pointer) 
	    (make-instance 'mac-pointer
			   :pointer
			   (addr-32-to-24
			     (sys:array-data-buffer-address
			       (add:parm-block-accessor acb 8))))
	    count)
	  (copy-parms-8b acb dest :to-array count 0 0))
	 (16
	  (block-swap-16b
	    (send src :mac-pointer) 
	    (make-instance 'mac-pointer
			   :pointer
			   (addr-32-to-24
			     (sys:array-data-buffer-address
			       (add:parm-block-accessor acb 16))))
	    count)
	  (copy-parms-16b acb dest :to-array count 0 0))
	 (32
	  (block-swap-32b
	    src 
	    (make-instance 'mac-pointer
			   :pointer
			   (addr-32-to-24
			     (sys:array-data-buffer-address
			       (add:parm-block-accessor acb 32))))
	    count)
	  (copy-parms-32b acb dest :to-array count 0 0)))
       
       (add:return-acb-fast acb t))
      
      (:exp-to-mac
       
       (setf acb (get-acb-fast byte-count))
       
       (ecase data-size
	 (8
	  (copy-parms-8b acb src :to-acb count 0 0)
	  (!blockmove
	    (make-instance 'mac-pointer
			   :pointer
			   (addr-32-to-24
			     (sys:array-data-buffer-address
			       (add:parm-block-accessor acb 8)))) 
	    (send dest :mac-pointer)
	    count))
	 (16
	  (copy-parms-16b acb src :to-acb count 0 0)
	  (block-swap-16b
	    (make-instance 'mac-pointer
			   :pointer
			   (addr-32-to-24
			     (sys:array-data-buffer-address
			       (add:parm-block-accessor acb 16)))) 
	    (send dest :mac-pointer)
	    count))
	 (32
	  (copy-parms-32b acb src :to-acb count 0 0)
	  (block-swap-32b
	    (make-instance 'mac-pointer
			   :pointer
			   (addr-32-to-24
			     (sys:array-data-buffer-address
			       (add:parm-block-accessor acb 32)))) 
	    (send dest :mac-pointer)
	    count)))
       (add:return-acb-fast acb t)))
    dest))

(DEFF mx-string-to-mac-string '!NewString)

(defun mac-string-to-mx-string (mac-string-handle-or-pointer &optional mx-string
				&aux acb str len)
  (LET ((pointer nil))
    (unwind-protect				  ;make sure handle gets unlocked
	(progn
	  (setf pointer (ETYPECASE mac-string-handle-or-pointer
			  (mac-handle
			   (!HLock mac-string-handle-or-pointer)
			   (deref mac-string-handle-or-pointer))
			  (mac-pointer mac-string-handle-or-pointer)))
  
	  (setq len (FetchByte pointer 0))
	  (setq acb (add:get-acb-fast len))
	  (!BlockMove (SEND pointer :+ 1)
		      (make-instance 'mac-pointer
				     :pointer
				     (addr-32-to-24
				       (sys:array-data-buffer-address
					 (add:parm-block-accessor acb :string))))
		      len)
	  (setq str (if mx-string
			(adjust-array mx-string len)
			(make-string len)))
	  (add:copy-parms-string acb str :to-array len 0 0)
	  (add:return-acb-fast acb))
      (when (typep  mac-string-handle-or-pointer 'mac-handle)
	;; then we had to lock this to use it, so unlock it now
	(!HUnlock mac-string-handle-or-pointer)))
  str))


(defun makeloc (ptr offset)
  (send ptr :+ offset))

(defun address (ptr)
  (etypecase ptr
    (mac-pointer ptr)
    (integer (make-instance 'mac-pointer :pointer ptr))))



;;
;; Creating and using mx heap zones.
;;

(defvar *mx-heap-zone-list* nil)

(defun MAKE-MX-HEAP-ZONE (size &optional (number-of-masters 64))
  "Create a heap zone in the microExplorers memory for faster toolbox data structure access."
  (declare (values heap-zone-pointer))
  (let* ((acb (add:get-acb-fast (+ size 64 (* number-of-masters 4)))) ; The 64 refers to the header and trailer.
	 (zone-start (make-instance 'mac-pointer :pointer (ptr32to24 (add:address (add:parm-block-accessor acb 8)))))
	 (zone-end (send zone-start :+ (add:data-bytes acb)))
	 (parm-block (!NewPtr 14))
	 (current-zone (!GetZone)))

    (stow parm-block 0 (send zone-start :pointer))
    (stow parm-block 4 (send zone-end :pointer))
    (stowWord parm-block 8 number-of-masters)
    (stow parm-block 10 0)
    
    (!initZone parm-block)

    ;; Set the zone back to the original.
    (!SetZone current-zone)

    (!DisposPtr parm-block)
    (setf (getf *mx-heap-zone-list* (send zone-start :pointer)) acb)
    zone-start))

(defun REMOVE-MX-HEAP-ZONE (zone)
  "Returns the storage used for a MX heap zone. Warning: You should not reference any objects allocated in this zone after running this function."
  (let ((acb (getf *mx-heap-zone-list* (send zone :pointer))))
    (unless acb
      (ferror 'no-such-zone "~s is not a valid mx heap zone." zone))
    (add:return-acb acb t)))

(defmacro WITH-HEAP-ZONE ((zone) &body body)
  "Wrap this macro around the allocation of any mac toolbox objects so that they will end up in zone."
  `(let* ((current-zone (!GetZone))
	 (set-zone-p (not (= (send ,zone :pointer) (send current-zone :pointer)))))
     (when set-zone-p
       (!SetZone ,zone))
     (prog1
       (progn
	 ,@body)
       (when set-zone-p
	 (!SetZone current-zone)))))

(defun ptr24to32 (ptr)
  "Turn a 32 bit mac pointer or handle into a 24 bit pointer or handle."
  (let* ((addr (etypecase ptr (mac-pointer (send ptr :pointer)) (mac-handle (send ptr :handle)) (integer ptr)))
	 (value (dpb #xF (byte 4 28) (dpb (ldb (byte 4 20) addr) (byte 4 24) (ldb (byte 20 0) addr)))))
    (etypecase ptr
      ((or mac-pointer integer)
       (make-instance 'mac-pointer :pointer value))
      (mac-handle
       (make-instance 'mac-handle :handle value)))))

(defun ptr32to24 (ptr)
  "Turn a 24 bit mac pointer or handle into a 32 bit pointer or handle."
  (let* ((addr (etypecase ptr (mac-pointer (send ptr :pointer)) (mac-handle (send ptr :handle)) (integer ptr)))
	 (value (dpb (ldb (byte 4 24) addr) (byte 4 20) (ldb (byte 20 0) addr))))
    (etypecase ptr
      (integer value)
      (mac-pointer
       (make-instance 'mac-pointer :pointer value))
      (mac-handle
       (make-instance 'mac-handle :handle value)))))

(defun break-pointer-on-word-boundry (p)
  "Returns the number of bytes to the next word boundry."
  (let ((remainder (rem p 4)))
    (if (zerop remainder) remainder (- 4 (rem p 4)))))

(Defmacro SWAP-16b (Word)
  (Once-Only (Word)
    `(Dpb (Ldb (Byte 8 0) ,Word) (Byte 8 8) (Ldb (Byte 8 8) ,Word))))

(Defmacro SWAP-32b (Word)
  (Once-Only (Word)
    `(let* ((n (etypecase ,word
		(integer ,word)
		(mac-pointer (send ,word :pointer))
		(mac-handle (send ,word :handle))))
	    (swapped-n (dpb (swap-16b (ldb (byte 16 0) n)) (byte 16 16) (swap-16b (ldb (byte 16 16) n)))))
       (typecase ,word
	 (integer swapped-n)
	 (mac-pointer (make-instance 'mac-pointer :pointer swapped-n))
	 (mac-handle (make-instance 'mac-handle :handle swapped-n))))))

(defun split-address (addr)
  "Break addr into the values slot and offset. Addr can be a mac-pointer, mac-handle, or an integer."
  (declare (values slot offset))
  (let ((addr (etypecase addr (mac-pointer (send addr :pointer)) (mac-handle (send addr :handle)) (integer addr))))
    (values (ldb (byte 4 20) addr) (ldb (byte 20 0) addr))))

(defun addr-in-my-slot-p (addr)
  "Return t if addr points to memory in the MX's slot. Addr can be a mac-pointer, mac-handle, or an integer."
  (let ((addr (etypecase addr (mac-pointer (send addr :pointer)) (mac-handle (send addr :handle)) (integer addr))))
    (eql (split-address addr) (ldb (byte 4 0) si:*addin-memory-slot*))))

;;;
;;; 32 bit accessors.
;;;

(defun Fetch (pointer &optional (offset 0))
  (let (slot
	slot-offset)	   
    (assert (typep pointer 'tb:mac-pointer) (pointer) "~s was not a mac-pointer object." pointer)
    (cond ((addr-in-my-slot-p pointer)
	   (setf (values slot slot-offset) (split-address pointer))
	   (incf slot-offset offset)
	   
	   (case (break-pointer-on-word-boundry slot-offset)
	     (0     
	      (swap-32b (si:%nubus-read slot slot-offset)))
	     (1
	      (swap-32b (dpb (si:%nubus-read slot (+ slot-offset 1)) (byte 24 8)
			     (si:%nubus-read-8b slot slot-offset))))
	     (2
	      (swap-32b (dpb (si:%nubus-read-16b slot (+ slot-offset 2)) (byte 16 16)
			     (si:%nubus-read-16b slot slot-offset))))
	     (3 
	      (swap-32b (dpb (si:%nubus-read-8b slot (+ slot-offset 3)) (byte 8 24)
			     (ldb (byte 24 8) (si:%nubus-read slot (- slot-offset 1))))))))	   
	  (t
	   (fetch-internal pointer offset)))))

(defun stow (pointer &optional (offset 0) (value 0))
  (let (slot 
	slot-offset)
    (assert (typep pointer 'tb:mac-pointer) (pointer) "~s was not a mac-pointer object." pointer)
    (cond ((addr-in-my-slot-p pointer)
	   (setf (values slot slot-offset) (split-address pointer))
	   (incf slot-offset offset)

	   (setf value (swap-32b value))
	   
	   (case (break-pointer-on-word-boundry slot-offset)
	     (0     
	      (si:%nubus-write slot slot-offset value))
	     ((1 3)
	      (si:%nubus-write-8b slot slot-offset (ldb (byte 8 0) value))
	      (si:%nubus-write-16b slot (1+ slot-offset) (ldb (byte 16 8) value))
	      (si:%nubus-write-8b slot (+ 3 slot-offset) (ldb (byte 8 24) value)))
	     (2
	      (si:%nubus-write-16b slot slot-offset (ldb (byte 16  0) value))
	      (si:%nubus-write-16b slot (+ slot-offset 2) (ldb (byte 16 16) value)))))
	  (t
	   (stow-internal pointer offset value)))))

(defun FetchHandle (handle &optional (offset 0))
  (let (pointer
	slot
	slot-offset)
    (assert (typep handle 'tb:mac-handle) (handle) "~s was not a mac-handle object." handle)
    (cond ((and (addr-in-my-slot-p (send handle :handle)) (addr-in-my-slot-p (setf pointer (deref handle))))
	   (setf (values slot slot-offset) (split-address pointer))
	   
	   (incf slot-offset offset)
	   
	   (case (break-pointer-on-word-boundry slot-offset)
	     (0     
	      (swap-32b (si:%nubus-read slot slot-offset)))
	     (1
	      (swap-32b (dpb (si:%nubus-read slot (+ slot-offset 1)) (byte 24 8)
			     (si:%nubus-read-8b slot slot-offset))))
	     (2
	      (swap-32b (dpb (si:%nubus-read-16b slot (+ slot-offset 2)) (byte 16 16)
			     (si:%nubus-read-16b slot slot-offset))))
	     (3 
	      (swap-32b (dpb (si:%nubus-read-8b slot (+ slot-offset 3)) (byte 8 24)
			     (ldb (byte 24 8) (si:%nubus-read slot (- slot-offset 1))))))))
	  (t
	   (fetchHandle-internal handle offset)))))

(defun stowHandle (handle &optional (offset 0) (value 0))
  (let (pointer
	slot
	slot-offset)
    (assert (typep handle 'tb:mac-handle) (handle) "~s was not a mac-handle object." handle)
    (cond ((and (addr-in-my-slot-p handle) (addr-in-my-slot-p (setf pointer (deref handle))))
	   (setf (values slot slot-offset) (split-address pointer))
	   
	   (incf slot-offset offset)

	   (setf value (swap-32b value))
	   
	   (case (break-pointer-on-word-boundry slot-offset)
	     (0     
	      (si:%nubus-write slot slot-offset value))
	     ((1 3)
	      (si:%nubus-write-8b slot slot-offset (ldb (byte 8 0) value))
	      (si:%nubus-write-16b slot (1+ slot-offset) (ldb (byte 16 8) value))
	      (si:%nubus-write-8b slot (+ 3 slot-offset) (ldb (byte 8 24) value)))
	     (2
	      (si:%nubus-write-16b slot slot-offset (ldb (byte 16  0) value))
	      (si:%nubus-write-16b slot (+ slot-offset 2) (ldb (byte 16 16) value)))))
	  (t
	   (stowHandle-internal handle offset value)))))

;;;
;;; 16 bit accessors
;;;

(defun Fetchword (pointer &optional (offset 0))
  (let (slot
	slot-offset)
    (assert (typep pointer 'tb:mac-pointer) (pointer) "~s was not a mac-pointer object." pointer)
    (cond ((addr-in-my-slot-p pointer)
	   (setf (values slot slot-offset) (split-address pointer))
	   (setf slot-offset (+ slot-offset offset))
	   	   
	   (case (break-pointer-on-word-boundry slot-offset)

	     ((0 2)
	      (swap-16b (si:%nubus-read-16b slot slot-offset)))

	     ((1 3)
	      ;; Note the byte swapping here.
	      (dpb (si:%nubus-read-8b slot slot-offset) (byte 8 8) (si:%nubus-read-8b slot (1+ slot-offset))))))
	  (t
	   (FetchWord-internal pointer offset)))))

(defun StowWord (pointer &optional (offset 0) (value 0))
  (let (slot 
	slot-offset)
    (assert (typep pointer 'tb:mac-pointer) (pointer) "~s was not a mac-pointer object." pointer)
    (cond ((addr-in-my-slot-p pointer)
	   (setf (values slot slot-offset) (split-address pointer))
	   (setf slot-offset (+ slot-offset offset))
	   (ecase (break-pointer-on-word-boundry slot-offset)
	     ((0 2)	   
	      (si:%nubus-write-16b slot slot-offset (swap-16b value)))
	     ((1 3)
	      ;; Note the byte swapping here.
	      (si:%nubus-write-8b slot slot-offset (ldb (byte 8 8) value))
	      (si:%nubus-write-8b slot (1+ slot-offset) (ldb (byte 8 0) value)))

	     ))
	  (t
	   (StowWord-internal pointer offset value)))))

(defun FetchWordHandle (handle &optional (offset 0))
  (let (pointer
	slot
	slot-offset)
    (check-type handle mac-handle)
    (assert (typep handle 'tb:mac-handle) (handle) "~s was not a mac-handle object." handle)
    (cond ((and (addr-in-my-slot-p (send handle :handle)) (addr-in-my-slot-p (setf pointer (deref handle))))
	   (setf (values slot slot-offset) (split-address pointer))
	   (setf slot-offset (+ slot-offset offset))

	   (case (break-pointer-on-word-boundry slot-offset)

	     ((0 2)
	      (swap-16b (si:%nubus-read-16b slot slot-offset)))

	     ((1 3)
	      ;; Note the byte swapping here.
	      (dpb (si:%nubus-read-8b slot slot-offset) (byte 8 8) (si:%nubus-read-8b slot (1+ slot-offset))))))
	  (t
	   (FetchWordHandle-internal handle offset)))))

(defun StowWordHandle (handle &optional (offset 0) (value 0))
  (let (pointer
	slot
	slot-offset)
    (assert (typep handle 'tb:mac-handle) (handle) "~s was not a mac-handle object." handle)
    (cond ((and (addr-in-my-slot-p handle) (addr-in-my-slot-p (setf pointer (deref handle))))
	   (setf (values slot slot-offset) (split-address pointer))
	   (setf slot-offset (+ slot-offset offset))
	   (case (break-pointer-on-word-boundry slot-offset)
	     ((0 2)	   
	      (si:%nubus-write-16b slot slot-offset (swap-16b value)))
	     ((1 3)
	      ;; Note the byte swapping here.
	      (si:%nubus-write-8b slot slot-offset (ldb (byte 8 8) value))
	      (si:%nubus-write-8b slot (1+ slot-offset) (ldb (byte 8 0) value)))

	     ))
	  (t
	   (StowWordHandle-internal handle offset value)))))



;;;
;;; 8 bit accessors
;;;

(defun FetchByte (pointer &optional (offset 0))
  (let (slot
	slot-offset)	   
    (assert (typep pointer 'tb:mac-pointer) (pointer) "~s was not a mac-pointer object." pointer)
    (cond ((addr-in-my-slot-p pointer)
	   (setf (values slot slot-offset) (split-address pointer))
	   (si:%nubus-read-8b slot (+ offset slot-offset)))
	  (t
	   (FetchByte-internal pointer offset)))))

(defun StowByte (pointer &optional (offset 0) (value 0))
  (let (slot 
	slot-offset)
    (assert (typep pointer 'tb:mac-pointer) (pointer) "~s was not a mac-pointer object." pointer)
    (cond ((addr-in-my-slot-p pointer)
	   (setf (values slot slot-offset) (split-address pointer))
	   (si:%nubus-write-8b slot (+ offset slot-offset) value))
	  (t
	   (StowByte-internal pointer offset value)))))

(defun FetchByteHandle (handle &optional (offset 0))
  (let (pointer
	slot
	slot-offset)
    
    (assert (typep handle 'tb:mac-handle) (handle) "~s was not a mac-handle object." handle)
    (cond ((and (addr-in-my-slot-p (send handle :handle)) (addr-in-my-slot-p (setf pointer (deref handle))))
	   (setf (values slot slot-offset) (split-address pointer))
	   (si:%nubus-read-8b slot (+ offset slot-offset)))
	  (t
	   (FetchByteHandle-internal handle offset)))))

(defun StowByteHandle (handle &optional (offset 0) (value 0))
  (let (pointer
	slot
	slot-offset)
    (assert (typep handle 'tb:mac-handle) (handle) "~s was not a mac-handle object." handle)
    (cond ((and (addr-in-my-slot-p handle) (addr-in-my-slot-p (setf pointer (deref handle))))
	   (setf (values slot slot-offset) (split-address pointer))
	   (si:%nubus-write-8b slot (+ offset slot-offset) value))
	  (t
	   (StowByteHandle-internal handle offset value)))))


;;;
;;; Some primitive toolbox calls for efficency.
;;;


(defconstant *copybits-trap-index* 221.)

(defsubst %copyBits (srcbits-handle dstbits-handle srcleft srctop srcright srcbottom
		   dstleft dsttop dstright dstbottom
		   mode)
  (let ((acb (add:get-acb-fast 52))) ; Get the size from the macro acb-size-needed
    (setf (add:opcode acb) 0)
    (setf (add:requestor-complete acb) t)
    (setf (add:parm-32b acb 0) *copybits-trap-index*)

    ;; put srcbits in acb
    (setf (parm-32b acb 1) srcbits-handle)
    ; Slot 2 is ignored in this case.

    (setf (parm-32b acb 3) dstbits-handle)
    ; slot 4 is ignored in this case.


    ;; Put srcrect in acb.
    (setf (parm-16b acb 10) srctop)
    (setf (parm-16b acb 11) srcleft)
    (setf (parm-16b acb 12) srcbottom)
    (setf (parm-16b acb 13) srcright)


    ;; Put dstrect in acb.
    (setf (parm-16b acb 14) dsttop)
    (setf (parm-16b acb 15) dstleft)
    (setf (parm-16b acb 16) dstbottom)
    (setf (parm-16b acb 17) dstright)

    (setf (parm-32b acb 9) mode)
    (setf (parm-32b acb 10) 0)			       ; same as (send tb::!nilrgn :handle)

    (tb-transmit-packet acb)))

(defconstant *moveto-trap-index* 136.)

(defsubst %MoveTo (h v)
  (let ((acb (add:get-acb-fast 20)))
    (setf (add:requestor-complete acb) t)
    (setf (add:opcode acb) 0)
    (setf (add:parm-32b acb 0) *moveto-trap-index*)

    (setf (add:parm-32b acb 1) h)
    (setf (add:parm-32b acb 2) v)
    (tb-transmit-packet acb)))

(defconstant *lineto-trap-index* 134.)

(defsubst %LineTo (h v)
  (let ((acb (add:get-acb-fast 20)))
    (setf (add:requestor-complete acb) t)
    (setf (add:opcode acb) 0)
    (setf (add:parm-32b acb 0) *lineto-trap-index*)

    (setf (add:parm-32b acb 1) h)
    (setf (add:parm-32b acb 2) v)
    (tb-transmit-packet acb)))

(defconstant *PaintRect-trap-index* 149)

(defsubst %PaintRect (left top right bottom)
  (let ((acb (add:get-acb-fast 20)))
    (setf (add:requestor-complete acb) t)
    (setf (add:opcode acb) 0)
    (setf (add:parm-32b acb 0) *PaintRect-trap-index*)
    
    (setf (parm-16b acb 2) top)
    (setf (parm-16b acb 3) left)
    (setf (parm-16b acb 4) bottom)
    (setf (parm-16b acb 5) right)
    (tb-transmit-packet acb)))