;;; -*- Mode:Common-Lisp; Package:MACTOOLBOX; Base:10 -*-

;;; For the puposes of the Texas Instruments Contract, this software is in
;;; TBI Category "B". Copyright (C) 1988, 1987, ExperTelligence Incorporated.

;/*                                                                       */
;/*                      RESTRICTED RIGHTS LEGEND                         */
;/*                                                                       */
;/* Use, duplication, or disclosure by the Government is subject to       */
;/* restrictions as set forth in subdivision (b)(3)(ii) of the Rights in  */
;/* Technical Data and Computer Software clause at 52.227-7013.           */
;/*                                                                       */
;/*                           EXPERTELLIGENCE                             */
;/*                            5638 HOLLISTER                             */
;/*                          GOLETA, CA  93117                            */
;/*                                                                       */
;/*  Copyright (C) 1988, 1987, EXPERTELLIGENCE Incorporated.              */
;/*  All rights reserved.                                                 */
;/*                                                                       */

;;;  Written DWB Jan '88

;;; ************************************************************************
;;; Modification History:
;;;   08/16/88 DWB  Added KeyMap AHT class
;;;   08/18/88 RWR  Added keyword rView to ListRec defclass
;;;   08/22/88 RWR  Changes FMInput to size 16 (not 4)
;;;   08/23/88 RWR  Changed FontMetricRec to FMetricRec
;;;   08/28/88 RWR  added nilhndl global
;;;   08/29/88 RWR  Modified TEToScrap & TEFromScrap
;;;   08/29/88 RWR  Changed FMetricRec size to 20 (was incorrectly 16!) 
;;;   08/30/88 DWB  Added Picture methods :dispose & :close
;;;   09/01/88 RWR  Moved ScrapStuff here
;;;   09/09/88 DWB  Removed DisplayObject class
;;;   09/09/88 DWB  Fixed Rect 'ne :new to use -no-init
;;;   09/09/88 DWB  Added FontInfo 'new :new metamethods
;;;   09/13/88 DWB&RWR Added new TERec
;;;   09/14/88 DWB&RWR Added :set methods for CMY, HSV & HSL colors
;;;                    Added Dispose methods for handles & pointers
;;;   09/20/88 RWR Deleted dispose method for pointer
;;;   10/08/88 RWR&DWB Changed :SET to := on point,rect,RGB,CMY,HSL,HSV,region
;;;   10/10/88 RWR&DWB Uncommented SndChannel
;;;   10/13/88 RWR (TeRec :new) had its args in wrong order (now correct)
;;;   01/13/89a MAC from ET private patch MACTOOLBOX-PRIVATE-1-7
;;;              - (:METHOD RECT :CENTER) will now accept a rect object as its
;;;                argument as well as a point.
;;;              - (:METHOD RECT :CENTER-X) and (:METHOD RECT :CENTER-Y) now
;;;                accept an optional argument which is a displacement off true
;;;                center.
;;;   01/13/89b MAC from ET private patch MACTOOLBOX-PRIVATE-1-8
;;;                GET-CLASS-VARIABLE can now accept a variable name whose value
;;;                is NIL.
;;;   01/13/89c MAC from ET private patch MACTOLBOX-PRIVATE-1-9
;;;                A disposed handle or pointer is now set to 0.  Any attempt
;;;                to dispose of a handle or pointer which is already 0 will
;;;                signal a DISPOSE-NIL-HANDLE-ERROR or DISPOSE-NIL-POINTER-ERROR.
;;;   01/16/89 MAC Added documentation string to (:METHOD EVENTRECORD :MESSAGECHAR)
;;;                documenting that it returns an unmapped Macintosh character as
;;;                a Lisp character object.  Added functions EXPLORER-MESSAGECHAR
;;;                and MAC-MESSAGECHAR to convert between the values returned by
;;;                the :MESSAGECHAR method and the Explorer character set.
;;;   01/17/89 MAC Changed error message references to :SET methods to say := methods.
;;;   01/17/89 MAC Added UNDEFCLASS macro definition from MACTOOLBOX-1-7
;;; ************************************************************************



;Since these are used so much, do not require to be prefixed with Tb:
(export '(rect point))
(export '(make-instance-no-init while defclass describe-class send-if))

;;; the following PROCLAIM is needed so that this can be compiled on an Explorer
;;; without generating an undeclared, assumed special warning
(proclaim '(special TV:*Explorer-to-mac-character-code-map*
		    TV:*Mac-to-Explorer-character-code-map*))

;---------- Macintosh constants -----------------------------------

;These are initialized later, because classes are needed to init them.
(defconstant patCopy 8 "A Macintosh copybits transfer mode")
(defconstant patOr   9 "A Macintosh copybits transfer mode")
(defconstant patXor 10 "A Macintosh copybits transfer mode")
(defconstant patBic 11 "A Macintosh copybits transfer mode")
(defconstant notPatCopy 12 "A Macintosh copybits transfer mode")
(defconstant notPatOr 13 "A Macintosh copybits transfer mode")
(defconstant notPatXor 14 "A Macintosh copybits transfer mode")
(defconstant notPatBic 15 "A Macintosh copybits transfer mode")

(defconstant bold 1 "A Macintosh QuickDraw text mode")
(defconstant italic 2 "A Macintosh QuickDraw text mode")
(defconstant underline 4 "A Macintosh QuickDraw text mode")
(defconstant outline 8 "A Macintosh QuickDraw text mode")
;(defconstant shadow 16 "A Macintosh QuickDraw text mode") 
(defconstant condense 32 "A Macintosh QuickDraw text mode")
(defconstant extend 64 "A Macintosh QuickDraw text mode")

(proclaim '(special *screen-rect*))
;These are used by metamethod-dispatch
(proclaim '(special super))
(proclaim '(special self))

(DEFSETF fetch stow)
(DEFSETF fetchword stowword)
(DEFSETF fetchbyte stowbyte)
(DEFSETF fetchhandle stowhandle)
(DEFSETF fetchwordhandle stowwordhandle)
(DEFSETF fetchbytehandle stowbytehandle)

(defmacro check-tool-type (arg-name typ)
  `(or (and (typep ,arg-name 'Toolboxobject)
	    (typep ,arg-name ,typ))
       (and (not (typep ,arg-name 'Toolboxobject))
	    (typep ,arg-name 'mac-handle))))

;****** VAR's *******
(defmacro send-if (inst msg &rest args)
  `(let ((obj ,inst))
     (if obj (send obj ,msg ,@args))))


;DEFCLASS macro turns classes into flavors.
(defmacro defclass (clinfo &rest options &aux ivs optlist pflds allivs resultforms)
  (setq optlist (copy-seq options))
  (dolist (x options)
    (case (first x)
      (ivs (setq ivs (rest x))(setq optlist (delete x optlist)))
      (cvs (ferror "Flavors do not have 'Class' variables"))
      (pascal-fields (setf (get (first clinfo) :fields) (rest x))
		     (setq optlist (delete x optlist))
		     (setq pflds t))
     ; (t   (format t "~&Unknown defclass option: ~s -ignored" x))
      ))
  (COND (pflds
	 (setq _field_offset (cond ((get (second clinfo) 'obsize))
				   (t 0)))
	 (dolist (x (rest clinfo))
	   (sys:compose-flavor-combination (GET x 'sys:flavor))
	   (setq allivs (append allivs (sys:flavor-all-instance-variables
					 (get x 'sys:flavor)))))
	 (setq _class_kind
	       (cond ((member 'pointer allivs) :pointer)
		     ((member 'handle allivs)  :handle)
		     (t
		      (ferror nil
               "To use Pascal-fields option, ~s must be a subclass of a class with a pointer or handle instance variable."
		       (first clinfo)))))
	 ))
  (if (not (get (first clinfo) 'sys::flavor))
      (sys:defflavor2 (first clinfo) nil nil nil))
  (setq resultforms
	`(progn
    (defflavor ,(first clinfo)
		,ivs
		,(rest clinfo)
       :gettable-instance-variables
       :settable-instance-variables
       :inittable-instance-variables
       ,@optlist)
     
     ;;So can do (rect 'new)
     (defmacro ,(first clinfo) (metamsg &rest args)
       `(meta-send ',',(first clinfo) ,metamsg ,@args))
     
     ;;So can check for (super ...) in defmethod
     (setf (get ',(first clinfo) :ecl-class) t)
     
     ;;Classes evaluate to the flavor.
     (defparameter ,(first clinfo) (get ',(first clinfo) 'sys::flavor))
     ))
    (if pflds
	(setq resultforms
	      (append resultforms
	  `(
     (setf (get ',(first clinfo) :fields) ',(get (first clinfo) :fields))
     
     ;;Add any instance-accessors to the pascal record
     (setf (get ',(first clinfo) :accessors-generated)
	   (if (zerop ,_field_offset) nil
	       (reverse (get ',(second clinfo) :accessors-generated))))
     
     ,(macroexpand-all `(deffield ,(first clinfo) "" ,(first clinfo)))
     (setf (get ',(first clinfo) 'tb:obsize) ,_field_offset)
     
     (setf (get ',(first clinfo) :accessors-generated)
	   (reverse (get ',(first clinfo) :accessors-generated))) 

     ',(first clinfo)
     ))))
    resultforms
  )


(defmacro undefclass (class-name)		       ; MACTOOLBOX-1-7
  `(progn (makunbound ',class-name)
	  (fmakunbound ',class-name)
	  (undefflavor ',class-name)))


;;;This methods describes all fields in a class with a pascal record.
(defun describe-class (name &aux message type offset)
  (cond ((and (symbolp name) (get name :accessors-generated)))
	((typep name 'sys:flavor) (setq name (sys:flavor-name name))))
  (format t "~2&Mac Toolbox Class: ~s~2%" name)
  (dolist (x (get name :accessors-generated))
    (setq message (first x)
	  type    (second x)
	  offset  (third x))
    (format t "~&~s ~30,10T;~s ~36,10T~a"
	    message offset (string-downcase type)))
  )

;Used to call the metamethod.
(defun metamethod-dispatch (class msg &rest args &aux mm res)
  (cond ((prog (cl)
	       (setq cl class)
	       top
	       (setq mm (get cl msg))
	       (cond ((functionp mm)
		      (let ((super
			      (first (sys::flavor-depends-on
				       (get cl 'sys::flavor)))))
			(setq res (apply mm args)))
		      (return res)))
	       (setq cl (first (sys::flavor-depends-on (get cl 'sys::flavor))))
	       (if (null cl) (return nil))
	       (go top)))
	(t (ferror "Unknown meta message: ~s~%" msg))))	

(eval-when (compile load eval)
  (defmacro super (&rest args)
    `(metamethod-dispatch super ,(first args) ,@(rest args)))

  (defmacro defmetamethod (class-msg args &rest body)
    `(defun (:property ,(car class-msg) ,(cadr class-msg)) ,args
       ,@Body))
  
  (defmacro undefmetamethod (class-msg)
    `(if (remprop ',(car class-msg) ',(cadr class-msg)) t nil))
  )

;All Macintosh pointers and handles are encapsulated in a mac-pointer instance
;or a mac-handle-instance respectively
(EVAL-WHEN (COMPILE load eval) ;everything!!!
(defflavor mac-handle
	   ((handle 0))
	   ()  
  :gettable-instance-variables
  :settable-instance-variables
  :inittable-instance-variables)

(defsignal DISPOSE-NIL-HANDLE-ERROR (toolbox-error dispose-nil-error dispose-nil-handle-error)   ; 01/13/89c Mac
	   (function)
  "a dispose function or method was given a handle which was already 0"
  :format-string "~s was called to disposed of a MAC-HANDLE which was NIL.")

(defmethod (mac-handle :dispose)()
  (cond ((zerop handle)(signal 'DISPOSE-NIL-HANDLE-ERROR nil '(:METHOD MAC-HANDLE :DISPOSE)))  ; 01/13/89c Mac
	(t (!Disposhandle self)
	   (setq handle 0))))

(defmethod (mac-handle :print-self)(&optional strm &rest x)
  (declare (ignore x))
  (format strm "#<~a ~a ~6,48x >"
	  (sys::flavor-name (sys::instance-flavor self))
	  "Handle"
	  handle))

(defflavor mac-pointer
	   ((pointer 0))
	   ()  
  :gettable-instance-variables
  :settable-instance-variables
  :inittable-instance-variables)

(defsignal DISPOSE-NIL-POINTER-ERROR (toolbox-error dispose-nil-error dispose-nil-pointer-error)   ; 01/13/89c Mac
	   (function)
  "a dispose function or method was given a pointer which was already 0"
  :format-string "~s was called to disposed of a MAC-POINTER which was NIL.")

(defmethod (mac-pointer :print-self)(&optional strm &rest x)
  (declare (ignore x))
  (format strm "#<~a ~a ~6,48x >"
	  (sys::flavor-name (sys::instance-flavor self))
	  "Pointer"
	  pointer))

(defmethod (mac-handle :mac-pointer) ()
  (deref self))

(defmethod (mac-pointer :mac-pointer) ()
  self)


(DEFMETHOD (mac-pointer :+) (offset)
  (IF (ZEROP offset)
      self
      (MAKE-INSTANCE 'mac-pointer :pointer (+ pointer offset))))

;used by data type macros so won't recurse
;This kind of make-instance does NOT call the :init method.
(defun make-instance-no-init (flav &rest args)
  (instantiate-flavor flav (locf args) nil nil))

;*********This is the ROOT Macintosh flavor**********
(defflavor object
	   ()
	   ()  
  :gettable-instance-variables
  :settable-instance-variables
  :inittable-instance-variables)

;These must be BEFORE window 'new, etc so the macro (rect 'new) expands correctly.
(defclass (rect object)
	  (ivs (top 50)
	       (left 50)
	       (bottom 100)
	       (right 100)))

(defclass (point object)
	  (ivs (v 0)
	       (h 0)))

(defsignal UNKNOWN-CLASS-VARIABLE-ERROR (toolbox-error unknown-class-variable-error)   ; 01/13/89b Mac
	   (variable-name class)
  "TB:GET-CLASS-VARIABLE was given a VARIABLE-NAME argument of a variable not
defined in its CLASS argument"
  :format-string "\"~A\" is not a known variable name in class ~A.")

(defun get-class-variable (class variable-name &aux mm)	                               ; 01/13/89b Mac
  (prog (cl)
	(setq cl class)
     top
	(setq mm (get cl variable-name 'FOO-BAR-BAZ))
	(if (NEQ MM 'FOO-BAR-BAZ) (return mm))
	(setq cl (first (sys::flavor-depends-on (get cl 'sys::flavor))))
	(if (null cl) (signal 'unknown-class-variable-error nil variable-name class))
	(go top)))	

(defmetamethod (object New)(&rest args)
  (instantiate-flavor self (locf args) nil nil))

;Toolbox objects are all subclasses of this class.
(defclass (ToolboxObject object))

(defmetamethod (ToolboxObject New)(&optional contents &aux the-inst)
  (declare (ignore contents))
  (setq the-inst (super 'New))
  )

;;;This methods describes all fields in a pascal record.
(defmethod (toolboxobject :describe)(&aux name message type offset)
  (setq name (sys:flavor-name (sys:instance-flavor self)))
  (format t "~2&Mac Toolbox Object: ~s~2%" self)
  (dolist (x (get name :accessors-generated))
    (setq message (first x)
	  type    (second x)
	  offset  (third x))
    (format t "~&~s ~25,10T~s ~60,10T;~s ~66,10T~s"
	    message (send self message) offset type))
  )

; *********** SystemCreatedTo ****************************
; These objects are created on the Mac side by toolbox "newprocs" which are
;   kept on the property list of the class under the indicator 'newproc.
; These objects can be created three ways:
; 1)    (setq obj (make-instance 'MenuInfo {:menuID 15 :menuTitle "Test"} ))
; 2)    (setq obj (menuInfo 'new 15 "Test"))
; 3)    (setq obj (menuinfo :new {:menuID 15 :menuTitle "Test"} ))

(defclass (SystemCreatedTo ToolboxObject))

;ECL style metamethod.
(defmetamethod (SystemCreatedTo New)(&rest args &aux the-inst proc flav)
  (setq the-inst (super 'new))
  (setq flav (cond ((instancep self)(sys:flavor-name (sys::instance-flavor self)))
		   ((symbolp self) self)))
  (if (and (setq proc (get-class-variable flav 'newproc)) (neq proc 'ignore))
      (cond ((typep the-inst 'mac-handle)
	     (send the-inst :set-handle  (send (apply proc args) :handle)))
	    ((typep the-inst 'mac-pointer)
	     (send the-inst :set-pointer (send (apply proc args) :pointer)))))
  the-inst
  )

;For flavor style (make-instance 'flav), call a :new metamethod that uses keywords
;to give default values to all arguments.  Then the 'new SystemCreatedTo metamethod
;above calls the Toolbox newproc to fill in the pointer or handle.
; NOTE-- :init-keywords must be given for this to work.

(defmethod (SystemCreatedTo :init)(props)
  (apply #'metamethod-dispatch
	 (cons (sys:flavor-name (sys::instance-flavor self))
	       (cons :new (contents props))))) 

;******************* AutoHandleTo *************************
; These are Macintosh records inside Handles of size Obsize

(defclass (AutoHandleTo ToolboxObject mac-handle))

(defmetamethod (AutohandleTo new)(&aux the-inst size)
  (setq the-inst (super 'new))
  (setq size (get-class-variable self 'obsize))
  (if (and (numberp size) (> size 0))
      (send the-inst :set-handle (send (!newhandle size) :handle))
      (ferror "~s class needs to have a valid obsize" self))
  the-inst)

;for (make-instance 'cursor) ......
(defmethod (AutoHandleTo :init)(props &aux size flav)
  (declare (ignore props))
  (setq size
	(get-class-variable
	  (setq flav (sys::flavor-name (sys::instance-flavor self))) 'obsize))
  (if (and (numberp size) (> size 0))
      (setq handle (send (!NewHandle size) :handle))
      (ferror "~s class needs to have a valid obsize" flav))
  )

;---------------------------------------------------------------------------
;               SystemCreatedTo subclasses
;---------------------------------------------------------------------------

;**ControlRecord**  
(defclass (ControlRecord SystemCreatedTo mac-handle)
  (:init-keywords :theWindow :boundsrect :title :visible :value
		  :min :max :procID :refcon))

(setf (get 'controlrecord 'newproc) '!NewControl)


(defmetamethod (ControlRecord New)(&optional (thewindow (!FrontWindow))
				  (boundsrect (rect 'new))
				  (title "Control Title")
				  (visible t)
				  (value 0)
				  (min 0)
				  (max 10)
				  (procID 0)
				  (refcon 0) &aux inst)
  (setq inst (super 'new theWindow boundsRect title visible
		    value min max procID refcon))
  (push (cons (send inst :handle) inst) *mac-control-list*)
  inst 
  )

(defmetamethod (ControlRecord :New)(&key (thewindow (!FrontWindow))
				  (boundsrect (rect 'new))
				  (title "Control Title")
				  (visible t)
				  (value 0)
				  (min 0)
				  (max 10)
				  (procID 0)
				  (refcon 0) &aux inst)
  (setq inst (super 'new theWindow boundsRect title visible
		    value min max procID refcon))
  (push (cons (send inst :handle) inst) *mac-control-list*)
  inst 
  )

(defmethod (ControlRecord :dispose)()
  (cond ((/= handle 0)
	 (setq *mac-control-list*
	       (delete (assoc handle *mac-control-list*)
		       *mac-control-list*))
	 (!DisposeControl self)
	 (setq handle 0)
	 t)
	(t (ferror nil "Attempt to dispose of a control with a zero handle."))))

(defclass (Button ControlRecord))

(defmetamethod (Button :new)(&key (thewindow (!FrontWindow))
				  (boundsrect (rect 'new 100 20 200 40))
				  (title "Button")
				  (visible t)
				  (value 0)
				  (min 0)
				  (max 1)
				  (procID 0)
				  (refcon 0)
				  &allow-other-keys
				  &aux inst)
  (setq inst (super 'new thewindow boundsrect title visible
		    value min max procID refcon))
  inst
  )

(defclass (CheckBox ControlRecord)
	  (ivs (state nil)))

(defmetamethod (CheckBox :new)(&key (thewindow (!FrontWindow))
				  (boundsrect (rect 'new 100 50 200 70))
				  (title "CheckBox")
				  (visible t)
				  (value 0)
				  (min 0)
				  (max 1)
				  (procID 1)
				  (refcon 0)
				  &allow-other-keys
				  &aux inst state)
  (setq state (not (or (zerop value)(null value))))
  (setq inst (super 'new thewindow boundsrect title visible (if state 1 0)
		    min max procID refcon))
  (send inst :set-state state)
  inst
  )

;This method doesn't waste time sending known info across the bus
(defmethod (CheckBox :Value)(&optional (val 0 flag) &aux bool)
  (cond (flag
	 (setq bool (not (or (eql val 0)(null val))))
	 (cond ((neq state bool)
		;only do if need to change value
		(!SetCtlValue self (if bool 1 0))
		(setq state (not state)) ;toggle state
		val)))
	(t ;takes too long (if (= 0 (!GetCtlValue self)) nil t)
	   state))
  )

(defclass (RadioButton ControlRecord)
	  (ivs (state nil)))

(defmetamethod (RadioButton :new)(&key (thewindow (!FrontWindow))
				  (boundsrect (rect 'new 100 80 200 100))
				  (title "Radio")
				  (visible t)
				  (value 0)
				  (min 0)
				  (max 1)
				  (procID 2)
				  (refcon 0)
				  &allow-other-keys
				  &aux inst state)
  (setq state (not (or (eql value 0)(null value))))
  (setq inst (super 'new thewindow boundsrect title visible (if state 1 0)
		    min max procID refcon))
  (send inst :set-state state)
  inst
  )

;This method doesn't waste time sending known info across the bus
(defmethod (RadioButton :Value)(&optional (val 0 flag) &aux bool)
  (cond (flag
	 (setq bool (not (or (eql val 0)(null val))))
	 (cond ((neq state bool)
		;only do if need to change value
		(!SetCtlValue self (if bool 1 0))
		(setq state (not state)) ;toggle state
		val)))
	(t ;takes too long (if (= 0 (!GetCtlValue self)) nil t)
	   state))
  )

(defclass (ScrollBar ControlRecord))


(defmetamethod (ScrollBar :new)(&key (thewindow (!FrontWindow))
				  (boundsrect (rect 'new 100 110 200 126 ))
				  (title "")
				  (visible t)
				  (value 0)
				  (min 0)
				  (max 10)
				  (procID 16)
				  (refcon 0)
				  &allow-other-keys
				  &aux inst)
  (setq inst (super 'new thewindow boundsrect title visible
		    value min max procID refcon))
  inst
  )

(defmethod (ControlRecord :Title)(&optional title)
  (cond (title
	 (!SetCTitle self (string title)))
	(t (setq title (GetCTitle self ))))
  title
  )

(defmethod (ControlRecord :set-Title)(title)
  (!SetCTitle self (string title))
  title
  )

(defmethod (ControlRecord :Hide)()
  (!HideControl self)
  self)

(defmethod (ControlRecord :Show)()
  (!ShowControl self)
  self)


(defmethod (ControlRecord :Value)(&optional (val 0 flag))
  (cond (flag
	 (!SetCtlValue self val) val)
	(t (!GetCtlValue self)))
  )

(defmethod (ControlRecord :Min)(&optional (val 0 flag))
  (cond (flag
	 (!SetCtlMin self val) val)
	(t (!GetCtlMin self)))
  )

(defmethod (ControlRecord :Max)(&optional (val 10 flag))
  (cond (flag
	 (!SetCtlMax self val) val)
	(t (!GetCtlMax self)))
  )

;**BitMap**
(defclass (bitmap autoHandleTo)
	  (pascal-fields (baseAddr     pointer)
			 (rowBytes     integer)
			 (bounds       rect)))

;**Pattern**
(defclass (pattern autoHandleTo)
	  (pascal-fields (one          unsigned-integer)
			 (two          unsigned-integer)
			 (three        unsigned-integer)
			 (four         unsigned-integer)))

;**GrafPort**
(defclass (agrafport SystemCreatedTo mac-pointer))     ;common methods
(setf (get 'agrafport 'newproc) 'ignore)
(compile-flavor-methods agrafport)

(defclass (grafport agrafport)
	  (pascal-fields (device       integer)
			 (portBits     bitmap)
			 (portRect     rect)
			 (visRgn       RgnHandle)
			 (clipRgn      RgnHandle)
			 (bkPat        pattern)
			 (fillPat      pattern)
			 (pnLoc        Point)
			 (pnSize       Point)
			 (pnMode       integer)
			 (pnPat        pattern)
			 (pnVis        integer)
			 (txFont       integer)
			 (txFace       style)
			 (txMode       integer)
			 (txSize       integer)
			 (spExtra      fixed)
			 (fgColor      longint)
			 (bkColor      longint)
			 (colrBit      integer)
			 (patStretch   integer)
			 (picSave      Handle)
			 (rgnSave      Handle)
			 (polySave     Handle)
			 (grafProcs    QDProcsPtr)))

;**RGBColor**
(defclass (RGBColor autohandleto)
	  (pascal-fields (red   unsigned-integer)
			 (green unsigned-integer)
			 (blue  unsigned-integer))
  (:documentation "The representation of a color -see I.M. V-48"))

(defmethod (RGBColor :=)(&optional (r 65535)(g 0)(b 0))
  (send self :red r)
  (send self :green g)
  (send self :blue b)
  self)

(defclass (CMYColor autohandleto)
	  (pascal-fields (cyan       smallfract)
			 (magenta    smallfract)
			 (yellow     smallfract)))

(defmethod (CMYColor :=)(&optional (c 0)(m 0)(y 0))
  (send self :cyan c)
  (send self :magenta m)
  (send self :yellow y)
  self)

(defclass (HSLColor autohandleto)
	  (pascal-fields (hue        smallfract)
			 (saturation smallfract)
			 (lightness  smallfract)))

(defmethod (HSLColor :=)(&optional (h 0)(s 1)(l 1))
  (send self :hue h)
  (send self :saturation s)
  (send self :lightness l)
  self)

(defclass (HSVColor autohandleto)
	  (pascal-fields (hue        smallfract)
			 (saturation smallfract)
			 (value      smallfract)))

(defmethod (HSVColor :=)(&optional (h 0)(s 1)(v 1))
  (send self :hue h)
  (send self :saturation s)
  (send self :value v)
  self)


;**PixMap**
(defclass (PixMap SystemCreatedTo mac-handle)
	  (pascal-fields (baseAddr     pointer)
			 (rowbytes     integer)
			 (bounds       rect)
			 (PMVersion    integer)
			 (PackType     integer)
			 (PackSize     longint)
			 (HRes         fixed)
			 (VRes         fixed)
			 (PixelType    integer)
			 (PixelSize    integer)
			 (CMPCount     integer)
			 (CMPSize      integer)
			 (PlaneBytes   longint)
			 (PMTable      CTabHandle)
			 (PMReserved   longint)))

(setf (get 'PixMap 'newproc) '!NewPixMap)

(defmetamethod (PixMap :new)()
  (super 'new))

(defmethod (PixMap :dispose)()			       ; 01/13/89c Mac
  (cond ((zerop handle) (signal 'DISPOSE-NIL-HANDLE-ERROR nil '(:METHOD PIXMAP :DISPOSE)))
	(t
	 (!DisposPixMap self)
	 (setq handle 0))))

;**PixPat**
(defclass (PixPat SystemCreatedTo mac-handle)
	  (pascal-fields (patType      integer)
			 (patMap       pixMapHandle)
			 (patData      handle)
			 (patXData     handle)
			 (patXValid    integer)
			 (patXMap      handle)
			 (pat1Data     pattern)))

(setf (get 'PixPat 'newproc) '!NewPixPat)

(defmetamethod (PixPat :new)()
  (super 'new))

(defmethod (PixPat :dispose)()			       ; 01/13/89c Mac
  (cond ((zerop handle)(signal 'DISPOSE-NIL-HANDLE-ERROR nil '(:METHOD PIXPAT :DISPOSE)))
	(t
	 (!DisposPixPat self)
	 (setq handle 0))))

;;; **Color grafport**
(defclass (Cgrafport agrafport)
	  (pascal-fields (device       integer)
			 (portPixMap   PixMapHandle)
			 (portVersion  integer)
			 (grafvars     handle)
			 (chExtra      integer)
			 (penLocHFrac  integer)
			 (portRect     rect)
			 (visRgn       RgnHandle)
			 (clipRgn      RgnHandle)
			 (bkPixPat     PixPatHandle)
			 (rgbFgColor   RGBColor)
			 (rgbBkColor   RGBColor)
			 (pnLoc        Point)
			 (pnSize       Point)
			 (pnMode       integer)
			 (pnPixPat     PixPatHandle)
			 (fillPixPat   PixPatHandle)
			 (pnVis        integer)
			 (txFont       integer)
			 (txFace       style)
			 (txMode       integer)
			 (txSize       integer)
			 (spExtra      fixed)
			 (fgColor      longint)
			 (bkColor      longint)
			 (colrBit      integer)
			 (patStretch   integer)
			 (picSave      Handle)
			 (rgnSave      Handle)
			 (polySave     Handle)
			 (grafProcs    CQDProcsPtr)))


;:OPEN message allocates space and sets up grafport.
;This is done because no newproc exists for grafports.
;Note: To be safe, both grafports allocate room for a color gp.
(defmetamethod (agrafport :new)(&aux inst)
  (setq inst (super 'new))
;  (send inst :open) ;if init, then slower & gens garbage if want to !GetPort
  inst)

(defmethod (grafport :open)()
  (setq pointer (send (!NewPtr 108) :pointer))
  (!OpenPort self)
  self
  )

(defmethod (cgrafport :open)()
  (setq pointer (send (!NewPtr 108) :pointer))
  (!OpenCPort self)
  self
  )

(defmethod (grafport :close)()			       ; 01/13/89c Mac
  (cond ((zerop pointer)(signal 'DISPOSE-NIL-POINTER-ERROR nil '(:METHOD GRAFPORT :CLOSE)))
	(t
	 (!ClosePort self)
	 (!DisposPtr self)
	 (setq pointer 0)))
  t)

(defmethod (cgrafport :close)()			       ; 01/13/89c Mac
  (cond ((zerop pointer)(signal 'DISPOSE-NIL-POINTER-ERROR nil '(:METHOD CGRAFPORT :CLOSE)))
	(t
	 (!CloseCPort self)
	 (!DisposPtr self)
	 (setq pointer 0)))
  t)

;this does the same as :close
(defmethod (agrafport :dispose)()
  (send self :close)
  t)

(defmethod (agrafport :clip)(r)
  (typecase r
    (rect (!ClipRect r) r)
    (region (!SetClip r) r))
  )



;*****************************************************************
;Window  
;*****************************************************************


(defclass (window cgrafport)
	  (ivs (theApplication nil) (growlimits nil)(winrect nil))
	  (pascal-fields
	    (windowKind    integer)
	    (visible       boolean)
	    (hilited       boolean)
	    (goAwayFlag    boolean)
	    (zoomFlag      boolean)
	    (strucRgn      RgnHandle)
	    (contRgn       RgnHandle)
	    (updateRgn     RgnHandle)
	    (windowDefProc Handle)
	    (dataHandle    Handle)
	    (titleHandle   Handle)        ;  Actually a handle to a Pascal string.
	    (titleWidth    integer)
	    (controlList   ControlHandle)
	    (nextWindow    Pointer)       ;  Actually a WindowPeek.
	    (windowPic     PicHandle))
    (:init-keywords :wStorage :boundsRect :title :visible :procID :behind
		  :goawayflag :refcon)
  )

(defclass (cwindow window)) ;treat both as color windows
(setf (get 'window 'newproc) '!NewCWindow) ;this applies to both

(defmetamethod (Window :New)(&key (wStorage nilptr)
				      (boundsREct (good-window-size))
				      (title "New Window")
				      (visible t)
				      (procID 8)
				      (behind oneptr)
				      (goAwayFlag t)
				      (refcon 0) &aux inst)
  (setq inst (super 'new wStorage boundsrect title visible
		    procID behind goAwayFlag refcon))
  (push (cons (send inst :pointer) inst) *mac-window-list*)
  (send inst :set-winrect (send inst :portrect)) ;copy of boundsrect
  inst
  )

(defmetamethod (Window New)(&optional (wStorage nilptr)
				      (boundsREct (good-window-size))
				      (title "New Window")
				      (visible t)
				      (procID 8)
				      (behind oneptr)
				      (goAwayFlag t)
				      (refcon 0) &aux inst)
  (setq inst (super 'new wStorage boundsrect title visible
		    procID behind goAwayFlag refcon))
  (push (cons (send inst :pointer) inst) *mac-window-list*)
  (send inst :set-winrect (send inst :portrect)) ;copy of boundsrect
  inst
  )

(defun good-window-size (&aux r)
  (send (send (setq r (rect 'new tb:*screen-rect*)) :inset 10 50)
		 :set-bottom 200)
  r
  )

(defmethod (Window :SetPort)()
  (!SetPort self)
  self)

(defmethod (Window :Hide)()
  (!HideWindow self))

(defmethod (Window :show)()
  (!ShowWindow self))

(defmethod (Window :dispose)()
  (cond ((/= pointer 0)
	 (setq *mac-window-list*
	       (delete (assoc pointer *mac-window-list*)
		       *mac-window-list*))
	 (!DisposeWindow self)
	 (setq pointer 0)
	 t)
	(t (ferror nil "Attempt to dispose of a window with a zero pointer."))))

(defmethod (Window :title)(&optional str)
  (if str (!SetWTitle self (string str)) (setq str (GetWTitle self)))
  str)

(defmethod (Window :set-title)(str)
  (!SetWTitle self (string str))
  str)

(defmethod (Window :select)()
  (!SelectWindow self)
  (!SetPort self)
  self)

(defmethod (Window :move)(h v &optional (front t))
  (!MoveWindow self h v front))

;window calculations
(defmethod (Window :width)()
  (send (send self :portrect) :width))

(defmethod (Window :height)()
  (send (send self :portrect) :height))

(defmethod (Window :inside-p)(pt)
  (send winrect :inside-p pt))

(defmethod (Window :erase)()
  (send (send self :portrect) :erase))

(defmethod (Window :inval)()
  (send (send self :portrect) :inval))

(defmethod (Window :exposed-p)()
  (send self :visible))

(defmethod (Window :expose)()
  (!ShowWindow self)
  t)


;**DialogRecord**
(defclass (dialogrecord window)
  (:init-keywords :dStorage :boundsRect :title :visible :procID :behind
		  :goawayflag :refcon :items))
(setf (get 'dialogrecord 'newproc) '!NewDialog)

(defmethod (DialogRecord :new)(&key (dStorage nilptr)
				      (boundsREct (good-window-size))
				      (title "New Dialog")
				      (visible t)
				      (procID 1)
				      (behind oneptr)
				      (goAwayFlag nil)
				      (refcon 0)
				      (items (make-instance 'mac-handle)))
  (super 'new dStorage boundsrect title visible
	      procID behind goAwayFlag refcon items)
  )

(defmethod (DialogRecord :dispose)()
  (!DisposDialog self))

;**ListRec**
(defclass (ListRec SystemCreatedTo mac-handle)
  (:init-keywords :rView :databounds :csize :theProc :thewindow :drawIt
		  :hasGrow :scrollHoriz :scrollVert))
(setf (get 'ListREc 'newproc) '!LNew)

(defmetamethod (ListREc :new)(&key (rView (rect 'new))
			  (databounds (rect 'new 0 0 5 10))
			  (csize (point 'new))
			  (theProc 0)
			  (theWindow (!FrontWindow))
			  (drawIt t)(hasGrow t)(scrollHoriz t)(scrollVert t))
  (super 'new rview databounds csize theProc theWindow
	 drawit hasGrow scrollhoriz scrollvert))

(defmethod (ListRec :dispose)()
  (!LDispose self)
  (send self :set-handle 0)
  self)

;**MenuInfo**
(defclass (MenuInfo SystemCreatedTo mac-handle)
	  (ivs (menuApp nil)(menuName nil)(menuAction nil)(id 0)
	       (itemNames nil) (ItemActions nil)
	       (itemMarks nil)(itemEnables nil)
	       (itemColors nil)
	       (title-color nil)(bar-color nil)
	       (items-color nil)(background-color nil)
	       (type :normal))			       ;added this IV
  (:init-keywords :menuID :menuTitle :type))

(setf (get 'MenuInfo 'newproc) '!NewMenu)

(defmetamethod (MenuINfo :new)(&key (menuID 50) (menutitle "Menu")(menuApp nil)
				    (type :normal)
				    &aux m)
  (setq m (super 'new menuID menutitle))
  (send m :set-menuApp menuApp)
  (send m :set-ID menuID)
  (send m :set-type type)
  m)


(defmethod (MenuInfo :DoMenuAction)(itemno)
  (setq itemno (1- itemNo))
  (cond (menuAction
	 (send-if (send menuApp :appWindow) :setPort)
	 (send menuApp menuAction (nth itemno itemNames))
	 t))
  )

(defmethod (MenuInfo :DoItemAction)(itemno)
  (setq itemno (1- itemno))
  (send-if (send menuApp :appWindow) :setPort)
  (send menuApp (nth itemno itemActions) (nth itemno itemNames))
  )

(defmethod (MenuInfo :Dispose)()		       ; 01/13/89c Mac
  (cond ((zerop handle)
	 (signal 'DISPOSE-NIL-HANDLE-ERROR nil '(:METHOD MENUINFO :DISPOSE)))
	(t
	 (!DisposeMenu self)
	 (setq handle 0))))

(defmethod (MenuInfo :FindItemAction)(n)
  (cond ((numberp n)
	 (nth n ItemActions))
	((symbolp n)
	 (setq n (position n ItemNames))
	 (if (numberp n) (nth  n ItemActions)))))

(defmethod (MenuInfo :AppendItem)(str
				  &optional (action :noop)
				  (nam (intern str 'keyword)))
  (!AppendMenu self str)
  (setq ItemActions (append ItemActions (list action)))
  (setq ItemNames   (append ItemNames   (list nam   )))
  self
  )

(defmethod (MenuInfo :DeleteItem)(item)
  (!DelMenuItem self item)
  (setq ItemActions (delete (nth (1- item) ItemActions) ItemActions))
  (setq ItemNames   (delete (nth (1- item) ItemNames)   ItemNames))
  self)

;**Picture**
(defclass (Picture SystemCreatedTo mac-handle)
  (:init-keywords :picFrame))
(setf (get 'Picture 'newproc) '!OpenPicture)

(defmetamethod (picture :new)(&key (picFrame (rect 'new)))
  (super 'new picFrame))

(defmethod (picture :close)()
  (!ClosePicture))

(defmethod (picture :draw)(r)
  (!DrawPicture self r))

(defmethod (picture :dispose)()
  (!KillPicture self))

;**PolyGon**
(defclass (Polygon SystemCreatedTo mac-handle))
(setf (get 'Polygon 'newproc) '!OpenPoly)

(defmetamethod (Polygon :new)()
  (super 'new))

(defmethod (polygon :close)()
  (!ClosePoly))

(defmethod (polygon :dispose)()
  (!KillPoly self)
  (setq handle 0)
  t)

(defmethod (polygon :offset)(dx dy)
  (!OffsetPoly self dx dy)
  self)

;**Region**
(defclass (Region SystemCreatedTo mac-handle)
	  (pascal-fields (rgnsize      unsigned-integer)
			 (rgnbbox      rect)))

(setf (get 'Region 'newproc) '!NewRgn)

(defmetamethod (Region :new)()
  (super 'new))

;**TEREc**
(defclass (TERec SystemCreatedTo mac-handle)
	  (pascal-fields (destRect     rect)
			 (viewRect     rect)
			 (selRect      rect)
			 (lineHeight   integer)
			 (fontAscent   integer)
			 (selPoint     point)
			 (selStart     integer)
			 (selEnd       integer)
			 (active       integer)
			 (wordBreak    procPtr)
			 (clikLoop     procPtr)
			 (clickTime    longint)
			 (clickLoc     integer)
			 (caretTime    longint)
			 (caretState   integer)
			 (just         integer)
			 (teLength     integer)
			 (hText        handle)
			 (recalBack    integer)
			 (recalLines   integer)
			 (clikStuff    integer)
			 (crOnly       integer)
			 (txFont       integer)
			 (txFace       style)
			 (txMode       integer)
			 (txSize       integer)
			 (inPort       grafPtr)
			 (highHook     procPtr)
			 (caretHook    procPtr)
			 (nlines       integer))
  (:init-keywords :viewrect :destrect))

(setf (get 'TERec 'NewProc) '!TENew)

(defmetamethod (TERec :new)(&key (viewrect (rect 'new))(destrect (rect 'new)))
  (super 'new destrect viewrect))

(defmethod (terec :dispose)()
  (cond ((/= handle 0)
	 (!TEDispose self)
	 (setq handle 0)
	 t))
  )


;To make the following definitions easy.
(eval-when (compile load eval)
  (defmacro defAHTclass (name size)
    `(progn (setf (get ',name 'obsize) ,size)
	    (defclass (,name AutoHandleTo)))))

;**AlertTemplate
(defAHTclass AlertTemplate 12)




;**Cursor**
(defAHTclass Cursor 68)
(defAHTclass CCursor 96)

;**DialogTemplate**
(defAHTclass DialogTemplate 276)

;**DrvSts**
(defAHTclass DrvSts 22)

;**FontInfo** This record is kept on the mX side
;**Int64Bit** Makes no sense on a Lisp Machine that has BIGNUMS!

(defAHTclass KeyMap 32)

;**ParamBlockRecord**
(defclass (FInfo autohandleto)
	  (pascal-fields
	    (fdType     OSType)
	    (fdCreator  OSType)
	    (fdFlags    integer)
	    (fdLocation Point)
	    (fdFldr     integer))
     )

(defclass (paramBlockRec autohandleto)
	  (pascal-fields
	    (qLink        QElemPtr)
	    (qType        integer)
	    (ioTrap       integer)
	    (ioCmdAddr    pointer)
	    (ioCompletion ProcPtr)
	    (ioResult     OSErr)
	    (ioNamePtr    StringPtr)
	    (ioVRefNum    integer))
  )

(defclass (ioParam paramBlockRec)
	  (pascal-fields
	    (ioRefNum     integer)
	    (ioVersNum    signedbyte)
	    (ioPermssn    signedbyte)
	    (ioMisc       pointer)
	    (ioBuffer     pointer)
	    (ioReqCount   longint)
	    (ioActCount   longint)
	    (ioPosMode    integer)
	    (ioPosOffset  longint))
  )

(defclass (fileParam paramBlockRec)
	  (pascal-fields
	    (ioFRefNum    integer)
	    (ioFVersNum   signedbyte)
	    (filler1      signedbyte)
	    (ioFDirIndex  integer)
	    (ioFlAttrib   signedbyte)
	    (ioFlVersNum  signedbyte)
	    (ioFlFndrInfo FInfo)
	    (ioDirID      longint)
	    (ioFlStBlk    integer)
	    (ioFlLgLen    longint)
	    (ioFlPyLen    longint)
	    (ioFlRStBlk   integer)
	    (ioFlRLgLen   longint)
	    (ioFlRPyLen   longint)
	    (ioFlCrDat    longint)
	    (ioFlMdDat    longint))
  )

(defclass (volumeParam paramBlockRec)
	  (pascal-fields
	    (filler2      longint)
	    (ioVolIndex   integer)
	    (ioVCrDate    longint)
	    (ioVLsBkUp    longint)
	    (ioVAtrb      integer)
	    (ioVNmFls     integer)
	    (ioVDirSt     integer)
	    (ioVBlLn      integer)
	    (ioVNmAlBlks  integer)
	    (ioVAlBlkSiz  longint)
	    (ioVClpSiz    longint)
	    (ioAlBlSt     integer)
	    (ioVVNxtFNum  longint)
	    (ioVFrBlk     integer)

;  Extra fields for HParamBlockRec
	    (ioVSigWord   integer)
	    (ioVDrvInfo   integer)
	    (ioVDRefNum   integer)
	    (ioVFSID      integer)
	    (ioVBkUp      longint)
	    (ioVSeqNum    integer)
	    (ioVWrCnt     longint)
	    (ioVFilCnt    longint)
	    (ioVDirCnt    longint)
	    (ioVFndrInfo1 longint)
	    (ioVFndrInfo2 longint)
	    (ioVFndrInfo3 longint)
	    (ioVFndrInfo4 longint)
	    (ioVFndrInfo5 longint)
	    (ioVFndrInfo6 longint)
	    (ioVFndrInfo7 longint)
	    (ioVFndrInfo8 longint)
  ))

;  CInfoPBRec, CMovePBRec, WDPBRec, FCBPBRec added 9/22/88   RJB

;  CInfoPBRec defines larger of two variants.

(defclass (fxinfo autohandleto)
	  (pascal-fields
	    (fdiconid integer)
	    (fdunused1 integer)
	    (fdunused2 integer)
	    (fdunused3 integer)
	    (fdunused4 integer)
	    (fdcomment integer)
	    (fdputaway longint)
	    )
  )

(defclass (CInfoPBRec paramBlockRec)
	  (pascal-fields
	    (ioFRefNum integer)
	    (ioFVersNum signedbyte)
	    (filler1 signedbyte)
	    (ioFDirIndex integer)
	    (ioFlAttrib signedbyte)
	    (filler2 signedbyte)
	    (ioFlFndrInfo FInfo)
	    (ioDirID longint)
	    (ioFlStBlk integer)
	    (ioFlLgLen longint)
	    (ioFlPyLen longint)
	    (ioFlRStBlk integer)
	    (ioFlRLgLen longint)
	    (ioFlRPyLen longint)
	    (ioFlCrDat longint)
	    (ioFlMdDat longint)
	    (ioFlBkDat longint)
	    (ioFlXFndrInfo FXInfo)
	    (ioFlParID longint)
	    (ioFlClpSiz longint))
  )

(defclass (CMovePBRec paramBlockRec)
	  (pascal-fields
	    (filler1 longint)
	    (ioNewName pointer)    ;  Actually a StringPtr.
	    (filler2 longint)
	    (ioNewDirID longint)
	    (filler3-1 longint)    ;  Unused array of 2 longints.
	    (filler3-2 longint)
	    (ioDirID longint))
  )

(defclass (WDPBRec paramBlockRec)
	  (pascal-fields
	    (filler1 integer)
	    (ioWDIndex integer)
	    (ioWDProcID longint)
	    (ioWDVRefNum integer)
	    (filler2-1 integer)   ;  Unused array of 7 integers.
	    (filler2-2 integer)
	    (filler2-3 integer)
	    (filler2-4 integer)
	    (filler2-5 integer)
	    (filler2-6 integer)
	    (filler2-7 integer)
	    (ioWDDirID longint))
  )

(defclass (FCBPBRec paramBlockRec)
	  (pascal-fields
	    (ioRefNum integer)
	    (filler integer)
	    (ioFCBIndx longint)
	    (ioFCBFlNm longint)
	    (ioFCBFlags integer)
	    (ioFCBStBlk integer)
	    (ioFCBEOF longint)
	    (ioFCBPLen longint)
	    (ioFCBCrPs longint)
	    (ioFCBVRefNum integer)
	    (ioFCBClpSiz longint)
	    (ioFCBParID longint))
  )

(defclass (controlParam paramblockRec))		       ; 38 bytes
(setf (get 'controlParam 'obsize) 122)


;**PenState**
(defAHTclass PenState 18)

;**Point** This record is kept on the mX side

(defAHTclass THPrint 120)
(defAHTclass TPrStatus 100) ;26 bytes actual - on safe side till check

;**Rect** This record is kept on the mX side

;**MISC**
(defAHTclass SerShk 35)
(defAHTclass SerStaRec 6)
(defAHTclass SysParmType 20)
(defAHTclass DateTimeRec 14)
(defAHTclass DecForm 4)
(defAHTclass Decimal 260)
(defAHTclass FMInput 16)
(defAHTClass FMoutput 24)
(defAHTclass FMetricRec 20)
(defAHTclass Intl0Rec 32)
(defAHTclass Intl1Rec 348)
(defAHTclass WinCTab 48)
(defAHTclass AuxWinRec 28)
(defAHTclass sndCommand 8)
(defAHTclass sndChannel 1060)
(defAHTclass MCEntry 30)

;**SFReply** This record is kept on the mX side
;**SFTypeList** Not needed because its a list

);*****END OD HUGE (EVAL-WHEN ...)



;;; The following structures will be kept in mX memory

(defclass (EventRecord object point)
	  (ivs what
	       message
	       when
	      ; hwhere inherited from point
	      ; vwhere     "      "    "
	       where
	       modifiers))

(defmethod (EventRecord :MessageWindow)()
  (cond ((cdr (assoc message *mac-window-list*)))
	(t (make-instance-no-init 'tb:window :pointer message))))

(defmethod (EventRecord :MessageChar)()		       ; 01/16/89 Mac
  "Returns the unmapped Macintosh character code in the form of a Lisp character object.
To convert this character object to the Explorer character set, use EXPLORER-MESSAGECHAR."
  (character (ldb 8 message)))

(defun explorer-messagechar (messagechar)	       ; 01/16/89 Mac
  "MESSAGECHAR is an ASCII character code or a Lisp character object representing the
Macintosh character code such as is returned by (:METHOD EVENTRECORD :MESSAGECHAR).  This
function returns a character object which is MESSAGECHAR mapped to the Explorer character set.
See also MAC-MESSAGECHAR."
  (code-char (AREF TV:*Mac-to-Explorer-character-code-map* (char-code messagechar))))


(defun mac-messagechar (explorer-messagechar)	       ; 01/16/89 Mac
  "EXPLORER-MESSAGECHAR is an ASCII character code or a Lisp character object representing a
character in the Explorer character set.  This function returns a character object which is
EXPLORER-MESSAGECHAR mapped to the Macintosh character set such as the characters returned
by (:METHOD EVENTRECORD :MESSAGECHAR).  See also EXPLORER-MESSAGECHAR."
  (code-char (AREF TV:*Explorer-to-mac-character-code-map* (char-code explorer-messagechar))))

(defmethod (EventRecord :MessageKey)()
  (ldb (byte 8 8) message))

(defmethod (EventRecord :MessageDrNum)()
  (ldb (byte 16 0) message))

;for compatability with ECL
(defmethod (EventRecord :hwhere)()
  h)

(defmethod (EventRecord :vwhere)()
  v)

(defclass (SfReply object)
	  (ivs good
	       copy
	       ftype
	       vRefNum
	       version
	       fname))

;x,y
(defmethod (point :print-self)(&optional strm &rest x)
  (declare (ignore x))
  (format strm "#<~s x=~s y=~s>"
	  (sys:flavor-name (sys:instance-flavor self)) h v))

(defmethod (rect :offset)(dx dy)
  (setq top (+ top dy) left (+ left dx) bottom (+ bottom dy) right (+ right dx)))

;x,y
(defmethod (rect :print-self)(&optional strm &rest x)
  (declare (ignore x))
  (format strm "#<~s ~s,~s ~s,~s>"
	  (sys:flavor-name (sys:instance-flavor self)) left top right  bottom))

(defclass (FontInfo object)
	  (ivs ascent
	       descent
	       widmax
	       leading))

(defmetamethod (FontInfo New)(&aux inst)
  (setq inst (super 'new))
  (!GetFontInfo inst)
  inst)

(defmetamethod (FontInfo :New)(&key (ascent 12 a-s)
				    (descent 3 d-s)
				    (widmax 14 w-s)
				    (leading 1 l-s) &aux inst)
  (setq inst (super 'new))
  (!GetFontInfo inst)
  (if a-s (send inst :set-ascent ascent))
  (if d-s (send inst :set-descent descent))
  (if w-s (send inst :set-widmax widmax))
  (if l-s (send inst :set-leading leading))
  inst)

(defmethod (FontInfo :size)()
  (+ ascent descent leading))

(defmethod (FontInfo :init)(props)
  (apply #'metamethod-dispatch (list* 'fontinfo :new (contents props))))

(defmethod (FontInfo :print-self)(&optional strm &rest x)
  (declare (ignore x))
  (format strm "#<FontInfo ascent:~s descent:~s widmax:~s leading:~s>"
	  ascent descent widmax leading))



;*****************************************************************
;Points
;*****************************************************************

;**Point Creation**
(defmetamethod (point new)(&optional (h 0)(v 0) &aux p)
  (cond ((typep h 'point)
	 (setq p (make-instance self :v (send h :v) :h (send h :h))))
	((typep h 'rect)
	 (setq p (send h :center)))
	(t (setq p (make-instance self :v v :h h))))
  p)

(defmethod (point :=)(&optional (hor 0)(ver 0))
  (cond ((typep hor 'point)
	 (setq h (send hor :h)
	       v (send hor :v)))
	((typep hor 'eventrecord)
	 (setq h (send hor :h)
	       v (send hor :v)))
	(t (setq h hor
		 v ver)))
  self)

(defmethod (point :x)()
  h)

(defmethod (point :y)()
  v)

(defmethod (point :set-x)(x)
  (setq h x))

(defmethod (point :set-y)(y)
  (setq v y))

(defmethod (point :add)(&optional x y)
  (cond ((typep x 'point)
	 (setq h (+ h (send x :h)))
	 (setq v (+ v (send x :v))))
	((and (numberp x)(numberp y))
	 (setq h (+ h x)
	       v (+ v y))))
  self)

(defmethod (point :sub)(&optional x y)
  (cond ((typep x 'point)
	 (setq h (- h (send x :h)))
	 (setq v (- v (send x :v))))
	((and (numberp x)(numberp y))
	 (setq h (- h x)
	       v (- v y))))
  self)

(defmethod (point :equal)(pt)
  (and (= h (send pt :h)) (= v (send pt :v))))

;This replaces the slow trap !MapPt
(defmethod (point :map)(srcRect dstRect)
  (setq h (+ (truncate (* (send dstrect :width) (- h (send srcRect :left)))
		       (send srcRect :width))
	     (send dstRect :left)))
  (setq v (+ (truncate (* (send dstrect :height) (- v (send srcRect :top)))
		       (send srcRect :height))
	     (send dstRect :top)))
  self)

;This replaces the slow trap !ScalePt
(defmethod (point :scale)(srcRect dstRect)
  (setq h (max 1 (truncate (* (send dstrect :width)  h) (send srcRect :width))))
  (setq v (max 1 (truncate (* (send dstrect :height) v) (send srcRect :height))))
  self)

  
;This is !PinREct (as opposed to !PtInRect) - it pins a point in a rect
(defmethod (point :pin)(r)
  (setq h (min (max h (send r :left)) (1- (send r :right))))
  (setq v (min (max v (send r :top))  (1- (send r :bottom))))
  self)

(defmethod (point :MoveTo)()
  (!MoveTo h v)
  self)

(defmethod (point :LineTo)()
  (!LineTo h v)
  self)

;*****************************************************************
;Rectangles
;*****************************************************************
;Graphic Operations: Frame,Paint,Erase,Invert,Fill
;Calculations : Inset, intersection, union, difference, in, equal, empty
;Movement, Sizing, Copy : Copy, Set, Offset

;**Rectangle creation**
(defmetamethod (rect new)
  (&optional (left 50)(top 50)(right 100)(bottom 100) &aux r)
  ;ALSO CAN PASS A RECT OR POINT AS OPT ARG
  (cond ((typep left 'rect)
         (setq r (make-instance-no-init self
				:top    (send left :top)
				:left   (send left :left)
				:bottom (send left :bottom)
				:right  (send left :right))))
        ((and (typep left 'point) (typep top 'point))
	 (setq r (make-instance-no-init self
				:top    (send left :v)
				:left   (send left :h)
				:bottom (send top :v)
				:right  (send top :h))))
        (t (setq r (make-instance-no-init self
				  :top top :bottom bottom
				  :left left :right right
				  ))
	   ))
  r)

(defmetamethod (rect :new)
  (&key (left 50)(top 50)(right 100)(bottom 100) rect point1 point2 &aux r)
  ;ALSO CAN PASS A RECT OR POINT AS OPT ARG
  (cond ((typep rect 'rect)
         (setq r (make-instance-no-init self
				:top    (send rect :top)
				:left   (send rect :left)
				:bottom (send rect :bottom)
				:right  (send rect :right))))
        ((and (typep point1 'point) (typep point2 'point))
	 (setq r (make-instance-no-init self
				:top    (send point1 :v)
				:left   (send point1 :h)
				:bottom (send point2 :v)
				:right  (send point2 :h))))
        (t (setq r (make-instance-no-init self
				  :top top :bottom bottom
				  :left left :right right
				  ))
	   ))
  r)

(defmethod (rect :=)
  (&optional (arg1 left)           (arg2 top)
   (arg3 (if (numberp arg1)(+ arg1 (- right left))))
   (arg4 (if (numberp arg2)(+ arg2 (- bottom top)))))
  ;ALSO CAN PASS A RECT OR POINT AS OPT ARG
  (cond ((typep arg1 'rect)
         (setq top    (send arg1 :top)
	       left   (send arg1 :left)
	       bottom (send arg1 :bottom)
	       right  (send arg1 :right)))
        ((and (typep arg2 'point) (typep arg2 'point))
	 (setq top    (min (send arg1 :v) (send arg2 :v))
	       left   (min (send arg1 :h) (send arg2 :h))
	       bottom (max (send arg1 :v) (send arg2 :v))
	       right  (max (send arg1 :h) (send arg2 :h))))
        ((and (numberp arg1) (numberp arg2) (numberp arg3) (numberp arg4))
	 (setq left arg1 top arg2 right arg3 bottom arg4))
	((null arg1) (setq left 0 top 0 right 0 bottom 0)) ;make empty
	(t (ferror nil "Bad args to := method of a rect.")))
  self)

(defmethod (rect :equal)(r)
  (and (= top (send r :top))
       (= left (send r :left))
       (= bottom (send r :bottom))
       (= right (send r :right))))

(defmethod (rect :empty-p)()
  (or (<= bottom top) (<= right left)))

;This replaces the slow trap !MapRect
(defmethod (rect :map)(srcRect dstRect)
  (setq left (+ (truncate (* (send dstrect :width) (- left (send srcRect :left)))
		       (send srcRect :width))
	     (send dstRect :left)))
  (setq right (+ (truncate (* (send dstrect :width) (- right (send srcRect :left)))
		       (send srcRect :width))
	     (send dstRect :left)))
  (setq top (+ (truncate (* (send dstrect :height) (- top (send srcRect :top)))
		       (send srcRect :height))
	     (send dstRect :top)))
  (setq bottom (+
		 (truncate (* (send dstrect :height)(- bottom (send srcRect :top)))
		       (send srcRect :height))
	     (send dstRect :top)))
  self)

(defmethod (rect :scale)(srcRect destRect)
  (setq right (+ left (truncate (* (send self :width)(send destREct :width))
				(send srcRect :width))))
  (setq bottom (+ top (truncate (* (send self :height)(send destREct :height))
				(send srcRect :height))))
  )

;********** Graphic Operations: Frame,Paint,Erase,Invert,Fill ***********

(defmethod (rect :Frame)()
  (!FrameRect self))

(defmethod (rect :FrameOval)()
  (!FrameOval self))

(defmethod (rect :Paint)()
  (!PaintRect self))

(defmethod (rect :PaintOval)()
  (!PaintOval self))

(defmethod (rect :erase)()
  (!EraseRect self))

(defmethod (rect :EraseOval)()
  (!EraseOval self))

(defmethod (rect :invert)()
  (!Invertrect self))

(defmethod (rect :InvertOval)()
  (!InvertOval self))

(defmethod (rect :fill)(&optional (pat tb:ltgray))
  (!FillRect self pat))

(defmethod (rect :FillOval)(&optional (pat tb:ltgray))
  (!FillOval self pat))

(defmethod (rect :fillC)(&optional (pat (!GetPixPat 16)))
  (!FillCRect self pat))

(defmethod (rect :fillCOval)(&optional (pat (!GetPixPat 16)))
  (!FillCOval self pat))

;********* Calculations : Inset, intersection, union, difference, in, equal, empty, width, height
;Calculations do NOT call the toolbox to save time.

;This does the same thing as the toolbox, which doesn't do what Inside Mac says.
(defmethod (rect :inset)(dh dv)
  (setq top    (+ top dv)
	left   (+ left dh)
	bottom (- bottom dv)
	right  (- right dh))
  self)

(defmethod (rect :offset)(dh dv)
  (setq top    (+ top dv)
	left   (+ left dh)
	bottom (+ bottom dv)
	right  (+ right dh))
  self)

(defmethod (rect :intersection)(rect-1)
  (cond ((typep rect-1 'rect)
	 (setq top    (max top  (send rect-1 :top))
	       left   (max left (send rect-1 :left))
	       bottom (min bottom (send rect-1 :bottom))
	       right  (min right  (send rect-1 :right)))
	 (cond ((or (<= right left)
		    (<= bottom top))
		    nil)  ;return NIL if they don't intersect  
	       (t t)) ;return T   if they intersect
	     )
	(t (ferror "Intersection method of RECT requires a RECT."))))

(defmethod (rect :intersection-p)(rect-1)
  (not (or (<= (min right  (send rect-1 :right))
	       (max left (send rect-1 :left)))
	   (<= (min bottom (send rect-1 :bottom))
	       (max top  (send rect-1 :top)))))
  )

(defmethod (rect :union)(rect-1)
  (cond ((typep rect-1 'rect)
	 (setq top    (min top  (send rect-1 :top))
	       left   (min left (send rect-1 :left))
	       bottom (max bottom (send rect-1 :bottom))
	       right  (max right  (send rect-1 :right))))
	(t (ferror "Union method of RECT takes a RECT arg.")))
  self
)

;This is for !PtinRect
;due to order of send syntax; should this be named surrounds?, contains? ???
(defmethod (rect :inside-p)(pt-or-rect)
  (typecase pt-or-rect
    (point (and (< left (send pt-or-rect :h) right)
		(< top  (send pt-or-rect :v) bottom)))
    (rect (and (< left (send pt-or-rect :left) (send pt-or-rect :right) right)
	       (< top  (send pt-or-rect :top)  (send pt-or-rect :bottom) bottom)))
    (t (ferror "In method of RECT requires a point or a RECT."))))

(defmethod (rect :center-x)(&optional x &aux half-width)       ; 01/13/89a Mac for ET
  (cond (x
 	 (setq left (- x (setq half-width (truncate (- right left) 2))))
	 (setq right (+ x half-width))
	 self)
	(t
	 (+ left (truncate (- right left) 2)))))

(defmethod (rect :center-y)(&optional y &aux half-height)      ; 01/13/89a Mac for ET
  (cond (y
 	 (setq top (- y (setq half-height (truncate (- bottom top) 2))))
	 (setq bottom (+ y half-height))
	 self)
	 (t
	   (+ top  (truncate (- bottom top) 2))))
  )

(defmethod (rect :center)(&optional pt)		               ; 01/13/89a Mac for ET
  (declare (arglist pt-or-rect)(values self-or-pt))
  (cond ((typep pt 'point)
	 (send self :offset
	       (- (send pt :h) (truncate (+ left right) 2))
	       (- (send pt :v) (truncate (+ top bottom) 2))))
	((typep pt 'rect)
	 (send self :offset
	       (- (truncate (+ (send pt :left)(send pt :right)) 2)
		  (truncate (+ left right) 2))
	       (- (truncate (+ (send pt :top)(send pt :bottom)) 2)
		  (truncate (+ top bottom) 2))))
	(t
	 (make-instance 'point
			:v  (+ top (truncate  (- bottom top) 2))
			:h  (+ left (truncate (- right left) 2)))))
  )
  
(defmethod (rect :width)()
  (- right left))

(defmethod (rect :height)()
  (- bottom top))

(defmethod (rect :MoveCenter)()
  (!MoveTo (+ left (truncate (- right left) 2))
           (+ top  (truncate (- bottom top) 2))))

(defmethod (rect :moveTo)(h &optional (v 0))
  (cond ((typep h 'point)
	 (send self :offset (- (send h :h) left) (- (send h :v) top)))
	(t (send self :offset (- h left) (- v top))))
  self
  )

(defmethod (rect :size)(h v)
  (setq right (+ left h)
	bottom (+ top v))
  self)

(defmethod (rect :valid)()
  (!ValidRect self))

(defmethod (rect :inval)()
  (!InvalRect self))

(defmethod (rect :scroll)(dx dy rgn)
  (!Scrollrect self dx dy rgn)
  rgn)

;rects converted to uE 6/22/88 dwb

;*****************************************************************
;Ovals
;*****************************************************************
(defclass (oval rect))

(defmethod (oval :Frame)()
  (!FrameOval self))

(defmethod (oval :Paint)()
  (!PaintOval self))

(defmethod (oval :erase)()
  (!EraseOval self))

(defmethod (oval :invert)()
  (!InvertOval self))

(defmethod (oval :fill)(&optional (pat tb:ltgray))
  (!FillOval self pat))

;*****************************************************************
;RoundRect
;*****************************************************************
(defclass (RoundRect rect)
  (ivs (ovalwidth 10)
       (ovalheight 10)))

(defmethod (RoundRect :Frame)()
  (!FrameRoundRect self ovalwidth ovalheight))

(defmethod (RoundRect :Paint)()
  (!PaintRoundRect self ovalwidth ovalheight))

(defmethod (RoundRect :erase)()
  (!EraseRoundRect self ovalwidth ovalheight))

(defmethod (RoundRect :invert)()
  (!InvertRoundRect self ovalwidth ovalheight))

(defmethod (RoundRect :fill)(&optional (pat tb:ltgray))
  (!FillRoundRect self ovalwidth ovalheight pat))

;**************************************************
;Polygons
;**************************************************

(defmethod (polygon :map)(srcREct dstREct)
  (!MapPoly self srcRect dstRect)
  self)

(defmethod (polygon :Frame)()
  (!FramePoly self))

(defmethod (polygon :Paint)()
  (!PaintPoly self))

(defmethod (polygon :erase)()
  (!ErasePoly self))

(defmethod (polygon :invert)()
  (!InvertPoly self))

(defmethod (polygon :fill)(&optional (pat tb:ltgray))
  (!FillPoly self pat))

;**************************************************
;Regions
;**************************************************

(defmethod (region :map)(srcREct dstREct)
  (!MapRgn self srcRect dstRect)
  self)

(defmethod (region :open)()
  (!OpenRgn))

(defmethod (region :close)()
  (!CloseRgn self))

(defmethod (region :dispose)()			       ; 01/13/89c Mac
  (cond ((zerop handle)(signal 'DISPOSE-NIL-HANDLE-ERROR nil '(:METHOD REGION :DISPOSE)))
	(t
	 (!DisposeRgn self)
	 (setq handle 0)
	 t))
  )

(defmethod (region :=)(&optional x a b c)
  (cond ((null x) (!SetEmptyRgn self))
        ((typep x 'region) (!CopyRgn x self))
        ((typep x 'rect)
         (!SetRectRgn self (send x :left)(send x :top)
		      (send x :right)(send x :bottom)))
	((and (numberp x)(numberp a)(numberp b)(numberp c))
	 (!SetRectRgn self x a b c))
	((and (typep x 'point) (typep a 'point))
	 (!SetRectRgn self
		      (min (send x :h) (send a :h))
		      (min (send x :v) (send a :v))
		      (max (send x :h) (send a :h))
		      (max (send x :v) (send a :v))))
	(t (ferror nil "Bad arg types to (Region :=) ~a ~a ~a ~a." x a b c)))
  self
  )

(defmethod (region :offset)(dx dy)
  (!OffsetRgn self dx dy)
  self)

(defmethod (region :inset)(dx dy)
  (!InSetRgn self dx dy))

(defmethod (region :intersection)(r &aux rgn)
  (typecase r
    (region (!SectRgn self r self))
    (rect   (setq rgn (make-instance 'region))
	    (!SetRectRgn rgn (send r :left) (send r :top)
			 (send r :right) (send r :bottom))
            (!SectRgn self rgn self)
	    (send rgn :dispose))
    (t (ferror nil "Method :intersection of region requires a rect or a region"
		      )))
    self)

(defmethod (region :union)(r &aux region_dummy)
  (typecase r
    (region (!UnionRgn self r self))
    (rect   (setq region_dummy (make-instance 'region))
	    (!SetRectRgn region_dummy (send r :left) (send r :top)
			 (send r :right) (send r :bottom))
            (!UnionRgn self region_dummy self)
	    (send region_dummy :dispose))
    (t (ferror nil "Method :union of region requires a rect or a region.")))
  self)

;no interrupts
(defmethod (region :inside-p)(x &aux region_dummy)
  (typecase x
    (point  (!PtInRgn x self))
    (region (setq region_dummy (make-instance 'region))
	    (!SectRgn x self region_dummy)
	    (prog1 (not (!EmptyRgn region_dummy))
		   (send region_dummy :dispose)))
    (rect  (!RectInRgn x self))
    (t (ferror nil "Method :inside-p requires a point, region or rect.")))
  )

(defmethod (region :equal)(r)
  (!EqualRgn self r))

(defmethod (region :empty-p)()
  (!EmptyRgn self))

(defmethod (region :valid)()
  (!ValidRgn self)
  t)

(defmethod (region :inval)()
  (!InvalRgn self)
  t)

;Drawing Regions

(defmethod (region :frame)()
  (!FrameRgn self)
  self)

(defmethod (region :paint)()
  (!PaintRgn self)
  self)

(defmethod (region :erase)()
  (!EraseRgn self)
  self)

(defmethod (region :invert)()
  (!InvertRgn self)
  self)

(defmethod (region :fill)(&optional (pat tb:ltgray))
  (!FillRgn self pat)
  self)

(defmethod (region :dragGray)(thePoint &key
			      limitRect
			      (slopRect (send (!FrontWindow) :portRect))
			      (axis 0) ;no constraint
			      (actionProc nilPtr)
			      &aux rgnrect)
  (cond ((null limitrect)
	 (setq rgnrect (make-instance 'rect))
	 (fetchrecthandle self 2 rgnrect) ;regions bounding rect
	 (setq limitRect
	       (make-instance 'rect
			      :left
			      (+ (send sloprect :left) 1
				 (- (send thePoint :h)(send rgnrect :left)))
			      :top
			      (+ (send sloprect :top) 1
				 (- (send thePoint :v)(send rgnrect :top)))
			      :right
			      (- (send sloprect :right)
				 (- (send rgnrect :right) (send thePoint :h)))
			      :bottom
			      (- (send sloprect :bottom)
				 (- (send rgnrect :bottom)(send thePoint :v))))
	       )))			
  (!DragGrayRgn self thePoint limitrect sloprect axis actionProc)
  ;returns multiple values dy dx
  )




;;; ScrapStuff
(defclass (ScrapStuff Object)
	  (ivs scrapsize
	       scraphandle
	       scrapcount
	       scrapstate
	       scrapname))

(defmethod (ScrapStuff :init) (ignore &aux ptr)
  (setq ptr (!_InfoScrap))
  (send self :set-scrapsize (fetch ptr 0))
  (send self :set-scrapHandle (make-instance 'mac-handle :handle (fetch ptr 4)))
  (send self :set-scrapCount (fetchword ptr 8))
  (send self :set-scrapState (fetchword ptr 10))
  (send self :set-scrapName (make-instance 'mac-pointer :pointer (fetch ptr 12)))
  self)


;misc

(compile-flavor-methods object systemcreatedto autohandleto listrec terec menuinfo controlrecord agrafport grafport cgrafport window cwindow polygon picture region mac-handle mac-pointer intl1rec intl0rec fmetricrec fmoutput fminput decimal
decform datetimerec serstarec sershk drvsts paramblockrec controlparam volumeparam fileparam ioparam sfreply alerttemplate dialogtemplate eventrecord penstate cursor ccursor bitmap rect point fontinfo pattern sysparmtype winctab auxwinrec sndcommand mcentry checkbox button radiobutton scrollbar pixmap dialogrecord rgbcolor keymap fxinfo cinfopbrec cmovepbrec wdpbrec fcbpbrec pixpat tprstatus)
