;;; -*- Mode:Common-Lisp; Package:MACTOOLBOX; Base:10 -*-

;;; For the puposes of the Texas Instruments/ ExperTelligence Contract, this
;;; software is in TBI Category "A".

;/*                                                                       */
;/*                      RESTRICTED RIGHTS LEGEND                         */
;/*                                                                       */
;/* Use, duplication, or disclosure by the Government is subject to       */
;/* restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  */
;/* Technical Data and Computer Software clause at 52.227-7013.           */
;/*                                                                       */
;/*                    TEXAS INSTRUMENTS INCORPORATED.                    */
;/*                            P.O. BOX 2909                              */
;/*                         AUSTIN, TEXAS 78769                           */
;/*                              MS 2151                                  */
;/*                                                                       */
;/*  Copyright (C) 1987-1989 Texas Instruments Incorporated.              */
;/*  All rights reserved.                                                 */
;/*                                                                       */

;;; MODIFICATIONS
;;; 01/18/89 MAC Moved initialization code to TOOLBOX-INIT
;;; 01/18/89 MAC added UNWIND-PROTECT to SELECT-APPLICATION

(defvar *async-events* t "If non-nil, applications will be brought up in async event mode.")

(defvar *MAX-NUMBER-OF-MAC-APPLICATIONS* 7)

(defvar *APPLICATION-CHANNEL-OFFSET* 9 "The channel number of the first channel to use for toolbox servers.")

(defvar *MAC-APPLICATIONS* nil "The list of Macintosh applications.")
(defvar *mac-application* nil "The mac application that is currently active.")
(defvar *application-channels* nil "The list of all channel instances used for mac applications.")

(defvar *default-tb-server* nil)

;;; the following form was moved to TOOLBOX-INIT - 01/18/89 Mac
;;;(add-initialization "Initialize Application Channels"
;;;		    '(setf  *application-channels*
;;;			    (reverse (dotimes (x *max-number-of-mac-applications*
;;;						 *application-channels*)
;;;				       (push (add:find-channel
;;;					       (+ x *APPLICATION-CHANNEL-OFFSET*))
;;;					     *application-channels*))))
;;;		    '(:system :now))

(defflavor mac-application-internal
	   ((name)
	    (server-name)
	    (directory)
	    (lisp-function)
	    (lisp-function-args)
	    (process)
	    (process-options `(:name ,(string (type-of self))))
	    
	    (event-list)
	    (last-event-arrival-time 0)
	    (last-event-current-port 0)
	    (last-mouse-h 0)
	    (last-mouse-v 0)
	    (last-mouse-moved-time 0)
	    (last-mouse-moved-event-time 0)
	    (last-when 0)
	    (last-modifiers 0)
	    (async-events *async-events*)  ; If non nil then events are being transmitted from the toolbox server.
	    (last-event-return-time 0))
	   (si:property-list-mixin)
  (:init-keywords :no-launch-server)
  :gettable-instance-variables
  :settable-instance-variables
  :inittable-instance-variables
  :outside-accessible-instance-variables
  (:accessor-prefix "MAC-APPLICATION-")
  )

(defflavor MAC-APPLICATION
	   ((*application-channel*)
	    (*mac-application*))
	   (mac-application-internal)
  (:special-instance-variables *application-channel* *mac-application*)
  :gettable-instance-variables
  :settable-instance-variables
  :outside-accessible-instance-variables)

(defun clear-application-channel (&optional (ch *application-channel*))
  (let* ((ch (add:find-channel ch))
	 (q (send ch :mx-queue)))
    (dotimes (x (add:q-length q))
      (let ((acb (add:deq-acb q)))
	(when acb
	  (add:return-acb-fast acb t))))))

(defmethod (mac-application :after :init) (init-plist)
  (let ((no-launch-server (getf (contents init-plist) :no-launch-server)))
    
    (setf process  (apply #'process-run-function process-options
			  #'(lambda (application-inst &rest args) (lexpr-send application-inst :top-level args))
			  self no-launch-server lisp-function-args))))

(defmethod (mac-application :top-level) (no-launch-server &rest args)
  (unwind-protect
      (progn
	(cond (server-name
	       (send self :allocate-application-channel)
	       (unless no-launch-server
		   (send self :launch-server)))
	       (t
		(setf *application-channel* (send self :default-application-channel))))
	
	(setf *mac-application* self)
	(clear-application-channel)
	(send self :set-async-event-mode)
	(apply lisp-function args))
    (setf async-events nil)
    (send self :set-async-event-mode)
    (send self :kill-server)
    (send self :deallocate-mac-application)
    (setf *application-channel* nil)
    (setf *mac-application* nil))
  )

(defvar *set-async-event-mode-opcode* 3 "The micronet opcode used to set the toolbox servers async event mode.")

(defmethod (mac-application-internal :set-async-event-mode) (&optional (mode async-events))
  "Add or delete channel to the async-event handler's list of channels to service."
  (let ((acb (add:get-acb-fast 2)))
    (if mode
	(add-async-event-channel *application-channel* self)
	(delete-async-event-channel *application-channel*))
    (setf (add:opcode acb) *set-async-event-mode-opcode*)
    (setf (add:requestor-complete acb) t)
    (setf (add:parm-8b acb 0) (if mode 1 0))
    (tb:tb-transmit-packet acb)
    (setf async-events (not (null mode)))))

(defmethod (mac-application-internal :ALLOCATE-APPLICATION-channel) ()
  "Allocate a mac application number and optionally associate a process with it."
  (without-interrupts
    
    (let* ((in-use-channels
	     (loop for app in *mac-applications*
		   collect (send app :*application-channel*))
	     )
	   (avail-channels (set-difference *application-channels* in-use-channels)))
      
      (cond (avail-channels
	     (setf *application-channel* (first avail-channels))
	     (push self *mac-applications*)
	     (setf (send *application-channel* :max-queue-length) 10))
	    (t
	     (ferror 'no-mac-application-is-available "There are no free mac applications.")))
	     
    (setf process si:current-process)))
  *application-channel*)

(defmethod (mac-application-internal :DEALLOCATE-MAC-APPLICATION) ()
  "Free a mac application number."
  (when (not (= (add:channel-number *application-channel*) 8))
    (without-interrupts
      (dolist (evt event-list)
	(deallocate-event evt))
      (setf event-list nil)
      (setf *mac-applications* (delete self *mac-applications* :test #'eq)))))

(defmethod (mac-application-internal :default-APPLICATION-channel) ()
  "Allocate a mac application number and optionally associate a process with it."
  8)


(defAHTclass FinderInfo 6)

(compile-flavor-methods finderinfo)

(defun make-server-pathname (namestring)
  "Coerce namestring into a server pathname."
  (etypecase namestring
    (pathname namestring)
    (string
     (let (device
	   directory
	   name
	   type)
       
       (setf (values device directory name type)
	     (send (fs:sample-pathname si:local-host)
		   :parse-namestring nil namestring))
       
       (make-pathname :host si:local-host :device device
		      :directory directory :name name :type type)))))

fs:
(defmethod (mac-pathname :directory-string) () (mac-directory-string))

(defun set-working-directory (namestring)
  (let* ((pathname (fs:translated-pathname (make-server-pathname namestring)))
	 (dir-string (concatenate 'string ":" (send pathname :directory-string)))
	 (dev-string (concatenate 'string (send pathname :device) ":")))
    
    (when (< 0 (length dir-string))
      (let ((parm-block (make-instance 'WDPbRec))
	    (strptr (!NewPtr 256))
	    (slen 0)
	    (wdnum))
	(send parm-block :ioCompletion nilptr)
	(send parm-block :ioNamePtr strptr)
	(send parm-block :ioVRefNum 0)
	(send parm-block :ioWDProcId #x12345678)
	(send parm-block :ioWDDirId 0)

	(stowbyte strptr 0 (setq slen (min (length dev-string) 255)))
	(dotimes (i slen) (stowbyte strptr (1+ i) (char-code (char dev-string i)) ))


	(!SetVol parm-block)
	
	(stowbyte strptr 0 (setq slen (min (length dir-string) 255)))
	(dotimes (i slen) (stowbyte strptr (1+ i) (char-code (char dir-string i))))
	
	(!OpenWD parm-block)
	(setf wdnum (send parm-block :IoVRefNum))	

	(!SetVol parm-block)

	(send parm-block :dispose)
	(!DisposPtr strptr)
	wdnum))))

(defun generate-application-search-list (&optional search-list-directories)
  "Generate a search list for search-list-directories on all volumes."
  (declare (values search-list))
  (let* ((*application-channel* 8)
	 (device-list)
	 (vol-param (make-instance 'VolumeParam))
	 (cat-param (make-instance 'CInfoPbRec))
	 (name (!NewPtr 256))
	 (search-list))
    
    (setf (send vol-param :IoNamePtr) name)

    (setf device-list  (loop 
			 with error
			 for i from 1
			 do (setf (send vol-param :IoVolIndex) i)
			    (setf error
				  (suppress-some-oserrs (!NsVErr)
				    (!GetVolInfo vol-param)))
			 when (not (zerop error))
			 return dlist
			 collect (mac-string-to-mx-string name) into dlist))
    
    (setf search-list
	  (append
	    (loop for device in device-list
		  nconc (loop for dir in search-list-directories
				collect (string-append device ":" dir)))
	    (loop for device in device-list
		  collect (string-append device ":")
		  collect (string-append device ":" "microexp:"))))
    
    (setf search-list
	  (delete-duplicates search-list :Test #'string-equal))
    
    (setf (send cat-param :IoNamePtr) name)
    (loop for dir in search-list
	  with error
	  do (setf (FetchByte name 0) (length dir))
	     (block-move dir (send name :+ 1) (length dir))
	     (setf error (suppress-some-oserrs (!FnfErr !dirNFErr) (!GetCatInfo cat-param)))
	  unless (zerop error)
	  collect dir into dirs-not-found
	  finally (setf search-list (set-difference search-list dirs-not-found)))
    (!DisPosPtr name)
    (send cat-param :dispose)
    (send vol-param :dispose)
    search-list))

;;; the following form was moved to TOOLBOX-INIT - 01/18/89 Mac
;;;(add-initialization "Setup Toolbox constants"
;;;		    '(initialize-toolbox-constants)
;;;		    '(:warm :now))

(defvar *finder-info* nil)

(defmethod (mac-application-internal :LAUNCH-SERVER) ()
  "Launch a toolbox server to service cals for application-number. Does nothing if the server is already launched."  
  
  (block launch
    ;; The server for channel 8 is in the driver and always active, so don't mess with it.
    (when (not (eq (add:find-channel 8) *application-channel*))
      
      ;; Setup AppParmHandle with trap-channel    
      (let* ((channel-number (add:channel-number *application-channel*))
	     (*application-channel* (add:find-channel 8))
	     (directories (generate-application-search-list directory)))

	;; Setup finder-info so that application will know what channel to use.
	(setf *finder-info* (or *finder-info* (make-instance 'FinderInfo)))
	(send *finder-info* :message 3)
	(send *finder-info* :trap-channel channel-number) 
	(stow (make-instance 'mac-pointer :pointer #xAEC) 0 (send *finder-info* :handle))

	;; Now loop down search list trying to launch it.
	(loop
	  (dolist (dir directories)
	    
	    (set-working-directory dir)
	    
	    ;; Launch it here.
	    (when (>= (launch server-name) 0)
	      (return-from launch)))
	  
	  (cerror "Try Again"
		  "The server file ~a was not found in any of the directories on ~s or was already running."
		  server-name))))))

(defmethod (mac-application-internal :kill-server) ()
  "Kill the toolbox server that services calls for application-number."
  
  ;; The toolbox server for application 0 is in the driver and can never be killed.
  (when (not (eq (add:find-channel 8) *application-channel*))
    (!ExitToShell)))

(defmethod (mac-application-internal :ferror) (signal-name format-string &rest args)
  "Signal signal-name in mac applications process."
  (lexpr-send process :interrupt 'ferror signal-name format-string args))

(defmethod (mac-application-internal :kill) ()
  "This should only be used for error aborts."
  (when process
    (send process :kill t)))

(compile-flavor-methods mac-application mac-application-internal)


;;;
;;; Launch support
;;;

;; Launching steps:
;; 
;; 1) If the application is launched from the Mac
;;         A *mac-toolbox-launcher* command is processed from the Mac.
;;      otherwise
;;         the function process-run-mac-application is run.
;; 2) The application allocates a mac application number.
;; 3) If the application was started from the MX 
;;       it sends a cmd to the mac to launch a toolbox server for the application number.
;; 4) The application waits until the toolbox server is running.
;; 5) The application function is funcalled.
;; 6) The application sends a call to the mac to kill the toolbox server.
;; 7) The application waits until the server is dead.
;; 8) The application deallocates the application number.

(defvar *MISC-ERROR-WHILE-LAUNCHING-MAC-APPLICATION* -128 "When there is nothing better to say.")

(defvar *DEBUG-MAC-TOOLBOX-LAUNCHER* nil "Set to t causes toolbox launch errors to enter the error handler.")

;; Misc channel opcodes for the mac toolbox server.

(defconstant *MX-TOOLBOX-LAUNCHER* 15)
;(defconstant *MX-TOOLBOX-SERVER-REGISTER* ??????)
;(defconstant *MX-TOOLBOX-SERVER-DEREGISTER* ??????)

;;; The method (:METHOD ADD:MISC-HANDLER :CASE :HANDLE-OPCODE 15)  cannot
;;; be compiled on an Explorer because the ADD:MISC-HANDLER flavor does not exist
;;; there.  Therefore, the body of this method has been moved to the function
;;; MISC-HANDLER-OPCODE-15 below.  The DEFMETHOD itself has been moved to the file
;;; TOOLBOX-INIT which will be loaded only on the microExplorer.
(defun misc-handler-opcode-15 (cmd)
  ;; Args from the Mac
  ;; 1) Function name length. This can be nil in which case there will be no process associated with this application.
  ;; 2) Function name.
  ;; 3) Name or keyword length.
  ;; 4) Name or keyword value to pass to process-run-function.
  ;; 5) Args length.
  ;; 6) Args value to pass to process-run-function.
  ;;
  ;; Return values
  ;; 1) Application number.
  ;;
  ;; If application can not be started a code will be put in the error-code field of the acb and as much of 
  ;;  an explanation as possible in the body of the acb (length in parm 0, text in parm 1 - n).
  
  (condition-call (condition)
      
      (case (add:subopcode cmd)
	(0					       ; Application being launched.
	 (let* ((application-name-start 0)
		(application-string (add:get-acb-string cmd application-name-start))
		(application (read-from-string application-string nil))
;	     (process-options-start (1+ (add:parm-8b cmd 0)))
;	     (process-options-string (add:get-acb-string cmd process-options-start))	 
;	     (process-options (read-from-string process-options-string nil))
;	     (args-start (1+ (add:parm-8b cmd process-options-start)))
;	     (args-string (add:get-acb-string cmd args-start))
;	     (args (read-from-string args-string nil))
		(application-instance (make-instance application :no-launch-server t))
		(application-number (process-wait-with-timeout "Waiting For Application" nil
							       #'(lambda (appl)
								   (let ((ch (send  appl :*application-channel*)))
								     (when ch (add:channel-number ch))))
							       application-instance)))
	   
	   (setf (add:parm-8b cmd 0) application-number)
	   (setf (add:error-code cmd) 0)))
	(1					       ; Application being killed
	 (let* ((application-channel (add:parm-8b cmd 0))
		(application (find application-channel *mac-applications*
				   :key #'(lambda (app) (send (add:find-channel (mac-application-*application-channel* app)) :number))
				   :test #'eql)))
	   (when application (send application :kill)))
	 
	 ))
	
    ((null *debug-mac-toolbox-launcher*)
     (let* ((error-string (send condition :report-string))
	    (error-string-length (min (1- (add:data-bytes cmd)) (length error-string) 255))
	    (error-code *misc-error-while-launching-mac-application*))
       
       (setf (add:error-code cmd) error-code)
       (add:put-acb-string cmd 0 (subseq error-string 0 error-string-length))))))

;;;;
;;;; Sending acbs to the toolbox server.
;;;;



(defvar *server-launch* 1 "Opcode used to launch an application from lisp.")

(defun launch (name)
  "Launch name on the macintosh and return the error code from the toolbox."
  (declare (values OsErr))
  (let ((acb (add:get-acb-fast (min (1+ (length name)) 255)))
	return-value)
    (add:put-acb-string acb 0 name)
    (setf (add:opcode acb) *server-launch*)
    (setf (add:requestor-complete acb) nil)
    (let ((*application-channel* 8))
      (tb-transmit-packet-and-wait acb))
    (setf return-value (add:parm-16b acb 0))
    (setf (add:requestor-complete acb) t)
    (add:return-acb-fast acb)
    (signed-ldb (byte 16 0) return-value)))


;;
;; A default server 
;;

(defflavor default-tb-server
	   ((server-name "tbserver")
	    (channel))
	   (mac-application-internal))

(defmethod (default-tb-server :after :init) (init-plist &aux (launched-p))
  (let ((no-launch-server (getf (contents init-plist) :no-launch-server)))
    (cond (*default-tb-server*
	   (when no-launch-server
	     (ferror 'server-already-launched
		     "There was an attempt to re-launch the default tbServer.")))
	  (t
	   (setf *default-tb-server* self)
	   (unwind-protect
	       (progn
		 (send self :allocate-application-channel)
		 (setf channel *application-channel*)
		 (setf *mac-application* self)
		 (clear-application-channel)
		 (unless no-launch-server
		   (send self :launch-server))
		 (send self :set-async-event-mode)
		 (setf launched-p t))
	     (unless launched-p
	       (if *application-channel*
		   (send self :deallocate-mac-application))
	       (setf async-events nil)
	       (send self :set-async-event-mode)
	       (makunbound '*application-channel*)
	       (setf *mac-application* nil)
	       (setf *default-tb-server* nil)))))))

(defmethod (default-tb-server :kill) ()
  (setf async-events nil)
  (send self :set-async-event-mode)
  (send self :kill-server)
  (send self :deallocate-mac-application)
  (setf channel nil)
  (makunbound '*application-channel*)
  (setf *mac-application* nil)
  (setf *default-tb-server* nil)
  t)

(defmethod (default-tb-server :*application-channel*) ()
  channel)

(defmethod (default-tb-server :*mac-application*) ()
  self)

(compile-flavor-methods default-tb-server)

(defvar CurApName (make-instance 'mac-pointer :pointer #x910) "A pointer to the current application name in Mac memory.")

(defun select-application (&optional (application-name CurApName))
  "Select the MultiFinder application named application-name. Application-name can be a lisp string, Mac-pointer, or Mac-handle."
  (let ((*application-channel* 8)
	lisp-string-name)

    (etypecase application-name
      (:string
       (setf lisp-string-name application-name))
      (mac-pointer
       (setf lisp-string-name (mac-string-to-mx-string application-name)))
      (mac-handle
       (unwind-protect				       ; 01/18/89 Mac
	   (progn
	     (!Hlock application-name)
	     (setf lisp-string-name (mac-string-to-mx-string application-name)))
	 (!HUnlock application-name))))
    (!opendeskacc lisp-string-name)))
  
(defun launch-default-tb-server (&optional kill-and-relaunch-p)
  "If tb:*default-tb-server* is non-NIL, return it.  If tb:*default-tb-server* is
NIL, create and return a new TB:DEFAULT-TB-SERVER instance.  If KILL-AND-RELAUNCH-P
is true, first kill the current default TB Server and then unconditionally launch
it again."
  (declare (values *default-tb-server*))
  (when kill-and-relaunch-p
    (kill-default-tb-server))
  (when (not *default-tb-server*)
    (make-instance 'default-tb-server))

  ;; ReSelect the microexplorer application.
  (select-application)
  
  *default-tb-server*)

(defun kill-default-tb-server ()
  (when *default-tb-server*
    (send *default-tb-server* :kill)))

(defun mac-application-cleanup (&optional (reinitialize-p t))
  (kill-default-tb-server)
  (dolist (application *MAC-APPLICATIONS*)
    (let ((process (send application :send-if-handles :process)))
      (when process
	(send process :kill))))
  (setf *MAC-APPLICATIONS* nil)
  (when *finder-info*
    (let ((*application-channel* 8))
      (send *finder-info* :dispose)
      (setf *finder-info* nil)))
  (setf *application-channels* nil)
  (when reinitialize-p
    (setf  *application-channels*
			    (reverse (dotimes (x *max-number-of-mac-applications*
						 *application-channels*)
				       (push (add:find-channel
					       (+ x *APPLICATION-CHANNEL-OFFSET*))
					     *application-channels*))))))

;;; the following form was moved to TOOLBOX-INIT - 01/18/89 Mac
;;;(add-initialization "Reset mac application support" '(mac-application-cleanup nil)
;;;		    :before-cold)

(defun initialize-server (application-name server-name search-list)
  (without-interrupts
    (let* ((application-name (let ((*package*))
			       (format nil "~s" application-name)))
	   (*application-channel* 8)
	   id 
	   application-name-handle
	   (found-it nil))

      (loop
	(dolist (dir search-list)
	  (set-working-directory dir)
	  (setf id (suppress-some-oserrs (!fnferr !dirNFErr) (!OpenResFile server-name)))
	  (when (> id 0)
	    (setf found-it t)
	    
	    (format t "~&Initializing the server file ~a in the directory ~a" server-name dir)
	    
	    (setf application-name-handle (Get1Resource "NAME" 0))
	    
	    ;; If there is already a name resource then remove it.
	    (cond ((zerop (send application-name-handle :handle))
		   (setf application-name-handle (!NewString (string application-name)))
		   (!AddResource application-name-handle "NAME" 0 "Application-name"))
		  
		  (t
		   (let ((len (length application-name)))
		     (!SetHandleSize application-name-handle (1+ len))
		     (stowbytehandle application-name-handle 0 len)
		     (dotimes (x len)
		       (StowByteHandle application-name-handle (1+ x)
				       (char-int (char application-name x))))
		     (!ChangedResource application-name-handle))))
	    
	    (!WriteResource application-name-handle)
	    
	    (!CloseResFile id)
	    (!ReleaseResource application-name-handle)))
	
	(cond (found-it
	       (return))
	      (t
	       (cerror "Try Again"
		       "The server file ~a was not found in any of the directories on ~s"
		       server-name search-list)))))))
;;
;; User callable functions.
;;


(defmacro define-mac-application (name arglist options &body body &aux ivar-list)
  
  (cond ((listp (getf options :directory))
	 (setf (getf options :directory) `(quote ,(getf options :directory))))
	(t
	 (setf (getf options :directory) `(quote ,(list (getf options :directory))))))

  (setf ivar-list (loop for (keyword value) on options by #'cddr
			collect (list (intern (symbol-name keyword) 'tb) value)))
  
  (push-end `(tb:name ,(string name)) ivar-list)
  
  (unless (assoc 'tb:lisp-function ivar-list)
    (push `(tb:lisp-function #'(lambda ,arglist ,@body)) ivar-list))
  
  `(progn
     
     ;; This will place the name resource on the toolbox server file.
     (initialize-server ',name ',(second (assoc 'server-name ivar-list))
			(generate-application-search-list ,(getf options :directory)))
     (defflavor ,name
		(,@ivar-list)
		(tb:mac-application))))

(defun launch-mac-application (name &rest args)
  (apply 'make-instance name :lisp-function-args args))



;;;
;;; Async events
;;;

;;
;; Channels serviced by the async event handler.
;;

(defvar *event-channel-list* nil "The list of channels to poll for async toolbox events.")

(defun add-async-event-channel (channel &optional application)
  (without-interrupts (push (list (add:find-channel channel) application) *event-channel-list*)))

(defun delete-async-event-channel (channel)
  (without-interrupts (setf *event-channel-list* (delete channel *event-channel-list* :key #'first :test #'eq))))

;;
;; Resource support for event Records used by the async event handler.
;;

(defvar *event-resource* nil "A list of available events.")
(add-initialization "Clear *event-resource*" '(setf *event-resource* nil) '(:full-gc))

(defun allocate-event (&optional (what 0) (message 0) (when 0) (modifiers 0) (v 0) (h 0))
  "Allocate an EventRecord from the available pool."
  (let ((evt (or (without-interrupts (pop *event-resource*)) (make-instance 'EventRecord))))
    (setf (send evt :what)      what)
    (setf (send evt :message)   message)
    (setf (send evt :when)      when)
    (setf (send evt :modifiers) modifiers)
    (setf (send evt :h)         h)
    (setf (send evt :v)         v)
    evt))

(defun deallocate-event (event)
  "Return an EventRecord to the available pool."
  (without-interrupts (push event *event-resource*)))

(defconstant *mouse-moved-event-opcode* #xfa "The opcode in an app4Evt for a mouse event.")

(defmethod (mac-application-internal :handle-event) (acb)
  (block handle-event

    (when (or (not (= last-mouse-h (parm-16b acb 7)))
	      (not (= last-mouse-v (parm-16b acb 6))))
      (setf last-mouse-h   (parm-16b acb 7))
      (setf last-mouse-v   (parm-16b acb 6))
      (setf last-mouse-moved-time (time)))
    
    (setf last-event-arrival-time (time))
    (setf last-when      (parm-32b acb 2))
    (setf last-modifiers (parm-16b acb 8))
    (setf last-event-current-port (parm-32b acb 5))
    
    ;; Don't queue mouse moved or null events.
    (when (or (zerop (parm-16b acb 0))
	      (and (eql (parm-16b acb 0) !App4Evt)
		   (= (ldb (byte 8 24)  (parm-32b acb 1)) *mouse-moved-event-opcode*)))
      (return-from handle-event))
    
    (push-end	    
      (allocate-event
	(parm-16b acb 0)
	(parm-32b acb 1)
	(parm-32b acb 2)
	(parm-16b acb 8)
	(parm-16b acb 6)
	(parm-16b acb 7)
	)
      event-list)))

(defun async-toolbox-event-handler (&aux acb ch app)
  (without-interrupts
    
    (loop
      (catch-error-restart (t "Restart the Async event handler.")
	(process-wait "Event Wait"
		      #'(lambda (channel-var app-var acb-var &aux acb)
			  (loop for (ch application) in *event-channel-list* do
				(when (setf acb (add:receive-packet ch t))
				  (setf (contents channel-var) ch)
				  (setf (contents app-var) application)
				  (setf (contents acb-var) acb)
				  (return t))))
		      (locf ch) (locf app) (locf acb))
	(when acb
	  (when app
	    (send app :handle-event acb))
	  (add:return-acb-servicer acb))))))
  
(defvar *async-event-handler-process* nil)

(defun reset-async-event-handler (&optional enable)
  "Kill and optionally restart *async-event-handler-process*"
  (when *async-event-handler-process*
    (send *async-event-handler-process* :kill t)
    (setf *async-event-handler-process* nil))
  (when enable
    (setf *async-event-handler-process*
	  (process-run-function
	    '(:name "Async Event Handler" :priority 15 :restart-after-reset t :restart-after-boot t)
	    #'(lambda () (loop (async-toolbox-event-handler)))))))

(add-initialization "Async Event Handler" '(reset-async-event-handler t) :system)
(add-initialization "Async Event Handler" '(reset-async-event-handler nil) :before-cold)

(defvar *temp-point* (make-instance 'tb:point) "Used by !WaitNextEvent only under without-interrupts.")

(defmethod (mac-application-internal :dequeue-event) (event-mask user-event &optional leave-in-queue mouse-moved-region &aux evt)
  "Return an event from applications event queue for event-mask."
  (block dequeue-event

    (dolist (evttype nil) ;;'(#.!app4Evt #.!activateEvt #.!mouseDown #.!mouseUp #.!keyDown #.!keyUp #.!diskEvt
		       ;;#.!networkEvt #.!driverEvt #.!app1Evt #.!app2Evt #.!app3Evt #.!autoKey #.!updateEvt))

      ;; Return mouse moved events only if a region was specified and we have seen the mouse move.
      (when (and (= evttype !app4Evt)
		 mouse-moved-region
		 (not (zerop (send mouse-moved-region :handle)))
		 (not (= last-mouse-moved-time last-mouse-moved-event-time)))
	
	(without-interrupts
	  (send *temp-point* := last-mouse-v last-mouse-h)
	  (when (not (!PtInRgn *temp-point* mouse-moved-region))	  
	    
	    (setf last-mouse-moved-event-time  last-mouse-moved-time)
	    (setf (send user-event :what)      !app4Evt)
	    (setf (send user-event :message)   (dpb  *mouse-moved-event-opcode* (byte 8 24) 0))
	    (setf (send user-event :when)      last-mouse-moved-time)
	    (setf (send user-event :h)         last-mouse-h)
	    (setf (send user-event :v)         last-mouse-v)
	    (setf (send user-event :modifiers) last-modifiers)
	    (return-from dequeue-event t))))

      (when (and (logand (ash 1 evtType) event-mask)	     
		 (setf evt (find evttype event-list :test #'eq :key #'(lambda (el) (send el :what)))))
	
	
	(when (not leave-in-queue)
	  (without-interrupts (setf event-list (delete evt event-list :test #'eq :count 1))))
	
	(setf (send user-event :what)      (send evt :what))
	(setf (send user-event :message)   (send evt :message))
	(setf (send user-event :when)      (send evt :when))
	(setf (send user-event :modifiers) (send evt :modifiers))
	(setf (send user-event :h)         (send evt :h))
	(setf (send user-event :v)         (send evt :v))
	
	(deallocate-event evt)
	(return-from dequeue-event t)))
    
    ;; No events available. return a null event.    
    (setf (send user-event :what)      0)
    (setf (send user-event :message)   0)
    ;; A crude way to generate :when for null events.
    (setf (send user-event :when)      (+ (time-difference (time) last-event-arrival-time) last-when))
    (setf (send user-event :h)         last-mouse-h)
    (setf (send user-event :v)         last-mouse-v)
    (setf (send user-event :modifiers) last-modifiers)
    (return-from dequeue-event nil)))

(defmethod (mac-application :flush-events) (eventMask stopMask)
  (without-interrupts
    (let ((events-to-be-deleted))
      
      ;; Flush pending events on the mac side.
      (!FlushEventsInternal eventMask stopMask)
      
      ;; Give any events in transit time to be delivered.
      (process-sleep 60 "Event Wait")	     
      
      (dolist (evt event-list)
	(when (not (zerop (logand (ash 1 (send evt :what)) stopMask)))
	  (return))
	
	(when (not (zerop (logand (ash 1 (send evt :what)) eventMask)))
	  (pushnew evt events-to-be-deleted)))
      
      (dolist (evt events-to-be-deleted)
	(setf event-list (delete evt event-list :test #'eq :count 1)))
      )))

(defmethod (mac-application-internal :set-async-events) (mode)
  (let ((acb (add:get-acb-fast 2)))
    (if mode
	(add-async-event-channel *application-channel* *mac-application*)
	(delete-async-event-channel *application-channel*))
    (setf (add:opcode acb) 2)
    (setf (add:requestor-complete acb) t)
    (setf (add:parm-8b acb 0) (if mode 1 0))
    (tb:tb-transmit-packet acb)
    (setf async-events (not (null mode)))))

(defmethod (mac-application-internal :case :set :async-events) (mode)
  (send self :set-async-events mode))

;;
;; Patched traps.
;;


(defun !WaitNextEvent (eventmask AnEventRecordInstance sleep region)
  (cond ((mac-application-async-events *mac-application*)
	 (let (evt-p
	       timeout-interval
	       (exit-time (time-increment (time) (* (or sleep 0) 60))))
	   (loop
	     (setf evt-p (send *mac-application* :dequeue-event eventMask AnEventRecordInstance nil region))
	     
	     (cond ((zerop (send AnEventRecordInstance :what))

		    (cond
		      ;; If we have received an event since the last time we returned an event return a null
		      ;; event to show the mouse and modifiers.
		      ((time-lessp (mac-application-last-event-return-time *mac-application*)
				   (mac-application-last-mouse-moved-time *mac-application*))
		       
			   (setf (mac-application-last-event-return-time *mac-application*) (time))
			   (return-from !WaitNextEvent nil))
			  
			  ((setf timeout-interval (if (time-lessp (time) exit-time) (time-difference exit-time (time))))		    
			   
			   (process-wait-with-timeout "!WaitNextEvent"
						      timeout-interval
						      #'(lambda (time application) (or (not (send application :async-events))
										       (not (= time (send application :last-event-arrival-time)))))
						      (mac-application-last-event-arrival-time *mac-application*) *mac-application*))
			  (t
			   		      
			   (setf (mac-application-last-event-return-time *mac-application*) (time))
			   (return-from !WaitNextEvent nil)))
		    )
		   (t
		    (setf (mac-application-last-event-return-time *mac-application*) (time))
		    (return-from !WaitNextEvent evt-p))))))
	
	(t
	 (!WaitNextEventInternal eventmask AnEventRecordInstance sleep region))))

(defun !GetNextEvent (eventmask AnEventRecordInstance)
  (cond ((mac-application-async-events *mac-application*)
	 (setf (mac-application-last-event-return-time *mac-application*) (time))
	 (send *mac-application* :dequeue-event eventMask AnEventRecordInstance))
	(t
	 ;; Call WaitNextEvent here since GetNextEvent cannot be patched easily on the mac side.
	 (!WaitNextEventInternal eventmask AnEventRecordInstance 0 nilRgn)))
  )

(defun !EventAvail (eventmask AnEventRecordInstance)
  (cond ((mac-application-async-events *mac-application*)
	 (send *mac-application* :dequeue-event eventMask AnEventRecordInstance t))
	(t
	 (!EventAvailInternal eventmask AnEventRecordInstance)))
  )

(defun !FlushEvents (eventMask stopMask)
  (cond (