;;; -*- Mode:Common-Lisp; Package:MACTOOLBOX; Base:10 -*-

;;; For the puposes of the Texas Instruments/ ExperTelligence Contract, this
;;; software is in TBI Category "A".

;/*                                                                       */
;/*                      RESTRICTED RIGHTS LEGEND                         */
;/*                                                                       */
;/* Use, duplication, or disclosure by the Government is subject to       */
;/* restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  */
;/* Technical Data and Computer Software clause at 52.227-7013.           */
;/*                                                                       */
;/*                    TEXAS INSTRUMENTS INCORPORATED.                    */
;/*                            P.O. BOX 2909                              */
;/*                         AUSTIN, TEXAS 78769                           */
;/*                              MS 2151                                  */
;/*                                                                       */
;/*  Copyright (C) 1987-1989 Texas Instruments Incorporated.              */
;/*  All rights reserved.                                                 */
;/*                                                                       */

;;; **********************************************************************
;;; Modifications
;;; 09/14/88 RWR&DWB Changed smallfract type
;;; 09/28/88 BJ Added ! to constant names.
;;; 10/11/88 MAC Changed pseudo-negative DEFCONSTANTs (which presumed 16-bit
;;;          integers into equivalent positive hex notation:
;;;             !APP4MASK -32768 => #x8000
;;;             !STOP15   -32768 => #x8000
;;;             !STOP20   -16384 => #xC000
;;; 11/09/88 BJ   Make CMapHandle refer to .mac-handle instead of 'handle.
;;; 01/16/89 MAC Merged MACTOOLBOX-1-7 back into this source
;;; 01/18/89 MAC Commented out DEFCONSTANTS defining Macintosh constants.
;;;              These values will be provided by TOOLBOX-CONSTANTS

(defmacro meta-send* (class msg args)		       ; from MACTOOLBOX-1-7
  `(let ((argslst (list* ,class ,msg ,args))(self ,class))
     (apply #'metamethod-dispatch argslst)))

(defmacro meta-send (class msg &rest args)	       ; from MACTOOLBOX-1-7
  `(let* ((argslst (list ,class ,msg ,@args))(self ,class))   ;in case args are IV
     (apply #'metamethod-dispatch argslst)))

(defvar *signal-mac-oserr* t 
  "If t signal an error if a trap that returns an osErr returns a non zero 
   error code.")

;This macro and the datatype that follow are for instance accessors defined
;using the (pascal-fields ...) option of defclass.

(defvar _field_offset 0)
(defvar _class_kind :handle)
(defvar *mac-window-list* nil "An assoc list pairing mac-pointers with instances")
(defvar *mac-control-list* nil
  "An assoc list pairing controlHandles with instances")

(defmacro while (&rest m &aux (again (gensym)))
  `(prog nil
      ,again
	 (if ,(car m) 
	     (progn ,@(cond ((eq (second m) 'do) (cddr m))
			    (t (cdr m)))
		    (go ,again)))))


(defmacro defpascaltype (typename size checkexp fetch-han fetch-ptr
			 numext type-expr)
  `(progn
  (setf (get ',typename :size) ,size)
  (setf (get ',typename :type-check-expr) ',checkexp)
  (setf (get ',typename :fetch-handle) ',fetch-han)
  (setf (get ',typename :fetch-pointer) ',fetch-ptr)
  (setf (get ',typename :number-extract) ',numext)
  (setf (get ',typename :type-expr) ',type-expr)
  ',typename))

(defun round-up-even (n)
  (if (oddp n) (1+ n) n))

(defun notzerop (x)(not (zerop x)))

(defpascaltype boolean
	       1
  (or (eq value t) (eq value nil))
  fetchbytehandle fetchbyte
  (if value 1 0)
  (notzerop))

(defpascaltype integer
	       2
  (typep value 'fixnum)
  fetchwordhandle fetchword
  value
  (signed-ldb 16))

(defpascaltype OSErr
	       2
  (typep value 'fixnum)
  fetchwordhandle fetchword
  value
  (signed-ldb 16))

(defpascaltype unsigned-integer
	       2
  (typep value 'fixnum)
  fetchwordhandle fetchword
  value
  (ldb 16))

(defpascaltype signedByte
	       1
  (typep value 'fixnum)
  fetchbytehandle fetchbyte
  value
  (signed-ldb 8))

(defpascaltype style
 	       1
  (typep value 'fixnum)
  fetchbytehandle fetchbyte
  value
  (ldb 8))

(defpascaltype fixed
	       4
  (numberp value)
  fetchhandle fetch
  (cond ((integerp value) value)
	((floatp value) (round (* value 65536))))
  (funcall #'(lambda (x) (/ x 65536))))

(defpascaltype smallFract
	       2
  (numberp value)
  fetchwordhandle fetchword
  (cond ((integerp value) value)
	((AND (floatp value) (<= 0 value 1.0)) (round (* value 65535)))
	(t (FERROR nil "Number out of range for SmallFract type.")))
  (funcall #'(lambda (x) (/ (ldb 16 x) 65535))))

(defpascaltype longint
	       4
  (integerp value)
  fetchhandle fetch
  value
  (identity))

(defpascaltype OSType     ;4 byte string
	       4
  (integerp value)
  fetchhandle fetch
  value
  (identity))

(defpascaltype handle
	       4
  (typep value 'mac-handle)
  fetchhandle fetch
  (send value :handle)
  (make-instance-no-init 'mac-handle :handle))

(defpascaltype pointer
	       4
  (typep value 'mac-pointer)
  fetchhandle fetch
  (send value :pointer)
  (make-instance-no-init 'mac-pointer :pointer))

(defpascaltype QElemPtr
	       4
  (typep value 'mac-pointer)
  fetchhandle fetch
  (send value :pointer)
  (make-instance-no-init 'mac-pointer :pointer))

(defpascaltype ProcPtr
	       4
  (typep value 'mac-pointer)
  fetchhandle fetch
  (send value :pointer)
  (make-instance-no-init 'mac-pointer :pointer))

(defpascaltype QDProcsPtr
	       4
  (typep value 'mac-pointer)
  fetchhandle fetch
  (send value :pointer)
  (make-instance-no-init 'mac-pointer :pointer))

(defpascaltype CQDProcsPtr
	       4
  (typep value 'mac-pointer)
  fetchhandle fetch
  (send value :pointer)
  (make-instance-no-init 'mac-pointer :pointer))

(defpascaltype StringPtr
	       4
  (typep value 'mac-pointer)
  fetchhandle fetch
  (send value :pointer)
  (make-instance-no-init 'mac-pointer :pointer))

(defpascaltype OSType
	       4
  (stringp value)
  fetchhandle fetch
  (restolong value)
  (longtores))

(defpascaltype grafPtr
	       4
  (check-tool-type value 'agrafport)
  fetchhandle fetch
  (send value :pointer)
  (make-instance-no-init 'cgrafport :pointer))

(defun restolong (str &aux (lng 0))
  (dotimes (i 4 lng)
    (if (> i 0) (setq lng (ash lng 8)))
    (setq lng (logior (char str i) lng))
    ))

(defun longtores (lng &aux str)
  (setq str (make-string 4))
  (dotimes (i 4 str)
    (setf (char str (- 3 i)) (ldb (byte 8 (* 8 i)) lng))
    ))

;(defpascaltype rect ...)(defpascal point ...)
; Rectangles and Points are different because they exist on the mX side, except
; when they are part of another record type.  We do not define them as a defclass
; with pascal-fields because we don't want accessors generated for them.
; Other records containing rect and point need this info so it is provided here.

(setf (get 'rect :fields)
      '((top integer)(left integer)(bottom integer)(right integer)))
(setf (get 'rect :size) 8)

(setf (get 'rect :special-type) 'defrecttype)
(defmacro defrecttype (class name &aux fetchtype)
  (setq fetchtype (if (eq _class_kind :pointer) 'fetchrect 'fetchrecthandle))
  (progn `(defmethod (,class ,(intern (string-upcase name) 'keyword))(&aux r)
	    (setq r (make-instance-no-init 'rect))
	    (,fetchtype self ,_field_offset r)
	    r)))

(setf (get 'point :fields)
      '((v integer)(h integer)))
(setf (get 'point :size) 4)

(setf (get 'point :special-type) 'defpointtype)
(defmacro defpointtype (class name &aux fetchtype)
  (setq fetchtype (if (eq _class_kind :pointer) 'fetchword 'fetchwordhandle))
  (progn `(defmethod (,class ,(intern (string-upcase name) 'keyword))(&aux p)
	    (setq p (make-instance-no-init 'point))
	    (send p :set-v (,fetchtype self ,_field_offset))
	    (send p :set-h (,fetchtype self ,(+ _field_offset 2)))
	    p)))


(defpascaltype RgnHandle
	       4
  (check-tool-type value 'Region)		       ;class is "Region"
  fetchhandle fetch
  (send value :handle)
  (make-instance-no-init 'Region :handle))

(defpascaltype pixpathandle
	       4
  (check-tool-type value 'pixpat)
  fetchhandle fetch
  (send value :handle)
  (make-instance-no-init 'pixpat :handle))

(defpascaltype pixmaphandle
	       4
  (check-tool-type value 'pixmap)
  fetchhandle fetch
  (send value :handle)
  (make-instance-no-init 'pixmap :handle))

(defpascaltype controlhandle
	       4
  (check-tool-type value 'controlrecord)
  fetchhandle fetch
  (send value :handle)
  (make-instance-no-init 'controlrecord :handle))

(defpascaltype pichandle
	       4
  (check-tool-type value 'picture)
  fetchhandle fetch
  (send value :handle)
  (make-instance-no-init 'picture :handle))

(defpascaltype CTabHandle
	       4
  (typep value 'mac-handle)
  fetchhandle fetch
  (send value :handle)
  (make-instance-no-init 'mac-handle :handle))

;This macro is used internally by the (pascal-fields ...) option of defclass.
(defmacro deffield (class name type &aux flds res msg setmsg fname fetchtype
		    (separator "") spec-type)
  (if (setq spec-type (get type :special-type))
      (push `(,spec-type ,class ,name) res))
  (cond ((setq flds (get type :fields))
	 ;;If it expands into subfields
	 (cons 'progn
	       (reverse
		 (dolist (f flds res)
		   (setq fname
			 (if (plusp (length (string name)))
			     (concatenate 'string
					  (string name) separator
					  (string (first f)))
			     (string (first f))))
		   (push `(deffield ,class ,fname ,(second f))  res)
		   )
		 )))
	(t
	 (setq msg (intern (string-upcase name) 'keyword))
	 (setq setmsg (intern (concatenate
				'string "SET-" (string-upcase name)) 'keyword))
	 (if (null (get type :size))
	     (ferror nil
         "~&~s type needs to be a defined (defpascalrecord ...) or a class." type))
	 (if (not (oddp (get type :size)))
	     (setq _field_offset (round-up-even _field_offset)))
	 (setq fetchtype (if (eq _class_kind :pointer)
			    :fetch-pointer :fetch-handle))
	 (prog1
	   `(progn
	      (push (list ',msg ',type ,_field_offset)
		    (get ',class :accessors-generated)) 
	   (defmethod (,class ,msg)(&optional (value nil supp?))
	      ,(format nil "An instance accessor to the ~s at offset ~s."
		       type _field_offset)
	      (if supp?
		  (cond (,(get type :type-check-expr)
			 (setf (,(get type fetchtype) self ,_field_offset)
			       ,(get type :number-extract))
			 value)
			(t
			 (ferror
			   nil " Bad instance accessor value: ~s, ~s required"
			   value ',type)))
		  ,(append (get type :type-expr)
			   (list (list (get type fetchtype) 'self _field_offset)))))
	    (defmethod (,class ,setmsg)(value)
		  (cond (,(get type :type-check-expr)
			 (setf (,(get type fetchtype) self ,_field_offset)
			       ,(get type :number-extract))
			 value)
			(t
			 (ferror
			   nil " Bad instance accessor value: ~s, ~s required"
			   value ',type))))
	   )
	      
	   (incf _field_offset (get type :size))
	   )))
	)

(defmacro VAR (x)
  `(locf ,x))

#|  These constants are now defined by TOOLBOX-CONSTANTS - 10/18/89 MAC
;;; Macintosh constants

;  Constants from controls.h
(defvar old-defs `(
 (DEFCONSTANT !PUSHBUTPROC 0)
 (DEFCONSTANT !CHECKBOXPROC 1)
 (DEFCONSTANT !RADIOBUTPROC 2)
 (DEFCONSTANT !USEWFONT 8)
 (DEFCONSTANT !SCROLLBARPROC 16)
 (DEFCONSTANT !INBUTTON 10)
 (DEFCONSTANT !INCHECKBOX 11)
 (DEFCONSTANT !INUPBUTTON 20)
 (DEFCONSTANT !INDOWNBUTTON 21)
 (DEFCONSTANT !INPAGEUP 22)
 (DEFCONSTANT !INPAGEDOWN 23)
 (DEFCONSTANT !INTHUMB 129)
 (DEFCONSTANT !NOCONSTRAINT 0)
 (DEFCONSTANT !HAXISONLY 1)
 (DEFCONSTANT !VAXISONLY 2)
 (DEFCONSTANT !DRAWCNTL 0)
 (DEFCONSTANT !TESTCNTL 1)
 (DEFCONSTANT !CALCCRGNS 2)
 (DEFCONSTANT !INITCNTL 3)
 (DEFCONSTANT !DISPCNTL 4)
 (DEFCONSTANT !POSCNTL 5)
 (DEFCONSTANT !THUMBCNTL 6)
 (DEFCONSTANT !DRAGCNTL 7)
 (DEFCONSTANT !AUTOTRACK 8)
 (DEFCONSTANT !CFRAMECOLOR 0)
 (DEFCONSTANT !CBODYCOLOR 1)
 (DEFCONSTANT !CTEXTCOLOR 2)
 (DEFCONSTANT !CTHUMBCOLOR 3)

;  Constants from desk.h

 (DEFCONSTANT !ACCEVENT 64)
 (DEFCONSTANT !ACCRUN 65)
 (DEFCONSTANT !ACCCURSOR 66)
 (DEFCONSTANT !ACCMENU 67)
 (DEFCONSTANT !ACCUNDO 68)
 (DEFCONSTANT !ACCCUT 70)
 (DEFCONSTANT !ACCCOPY 71)
 (DEFCONSTANT !ACCPASTE 72)
 (DEFCONSTANT !ACCCLEAR 73)

;  Constants from dialogs.h

 (DEFCONSTANT !CTRLITEM 4)
 (DEFCONSTANT !BTNCTRL 0)
 (DEFCONSTANT !CHKCTRL 1)
 (DEFCONSTANT !RADCTRL 2)
 (DEFCONSTANT !RESCTRL 3)
 (DEFCONSTANT !STATTEXT 8)
 (DEFCONSTANT !EDITTEXT 16)
 (DEFCONSTANT !ICONITEM 32)
 (DEFCONSTANT !PICITEM 64)
 (DEFCONSTANT !USERITEM 0)
 (DEFCONSTANT !ITEMDISABLE 128)
 (DEFCONSTANT !OK 1)
 (DEFCONSTANT !CANCEL 2)
 (DEFCONSTANT !STOPICON 0)
 (DEFCONSTANT !NOTEICON 1)
 (DEFCONSTANT !CAUTIONICON 2)

;  Constants from errors.h

 (DEFCONSTANT !NOERR 0)
 (DEFCONSTANT !QERR -1)
 (DEFCONSTANT !VTYPERR -2)
 (DEFCONSTANT !CORERR -3)
 (DEFCONSTANT !UNIMPERR -4)
 (DEFCONSTANT !SENODB -8)
 (DEFCONSTANT !CONTROLERR -17)
 (DEFCONSTANT !STATUSERR -18)
 (DEFCONSTANT !READERR -19)
 (DEFCONSTANT !WRITERR -20)
 (DEFCONSTANT !BADUNITERR -21)
 (DEFCONSTANT !UNITEMPTYERR -22)
 (DEFCONSTANT !OPENERR -23)
 (DEFCONSTANT !CLOSERR -24)
 (DEFCONSTANT !DREMOVERR -25)
 (DEFCONSTANT !DINSTERR -26)
 (DEFCONSTANT !ABORTERR -27)
 (DEFCONSTANT !IIOABORTERR -27)
 (DEFCONSTANT !NOTOPENERR -28)
 (DEFCONSTANT !DIRFULERR -33)
 (DEFCONSTANT !DSKFULERR -34)
 (DEFCONSTANT !NSVERR -35)
 (DEFCONSTANT !IOERR -36)
 (DEFCONSTANT !BDNAMERR -37)
 (DEFCONSTANT !FNOPNERR -38)
 (DEFCONSTANT !EOFERR -39)
 (DEFCONSTANT !POSERR -40)
 (DEFCONSTANT !MFULERR -41)
 (DEFCONSTANT !TMFOERR -42)
 (DEFCONSTANT !FNFERR -43)
 (DEFCONSTANT !WPRERR -44)
 (DEFCONSTANT !FLCKDERR -45)
 (DEFCONSTANT !VLCKDERR -46)
 (DEFCONSTANT !FBSYERR -47)
 (DEFCONSTANT !DUPFNERR -48)
 (DEFCONSTANT !OPWRERR -49)
 (DEFCONSTANT !PARAMERR -50)
 (DEFCONSTANT !RFNUMERR -51)
 (DEFCONSTANT !GFPERR -52)
 (DEFCONSTANT !VOLOFFLINERR -53)
 (DEFCONSTANT !PERMERR -54)
 (DEFCONSTANT !VOLONLINERR -55)
 (DEFCONSTANT !NSDRVERR -56)
 (DEFCONSTANT !NOMACDSKERR -57)
 (DEFCONSTANT !EXTFSERR -58)
 (DEFCONSTANT !FSRNERR -59)
 (DEFCONSTANT !BADMDBERR -60)
 (DEFCONSTANT !WRPERMERR -61)
 (DEFCONSTANT !FONTDECERROR -64)
 (DEFCONSTANT !FONTNOTDECLARED -65)
 (DEFCONSTANT !FONTSUBERR -66)
 (DEFCONSTANT !FIRSTDSKERR -84)
 (DEFCONSTANT !LASTDSKERR -64)
 (DEFCONSTANT !NODRIVEERR -64)
 (DEFCONSTANT !OFFLINERR -65)
 (DEFCONSTANT !NONYBERR -66)
 (DEFCONSTANT !NOADRMKERR -67)
 (DEFCONSTANT !DATAVERERR -68)
 (DEFCONSTANT !BADCKSMERR -69)
 (DEFCONSTANT !BADBTSLPERR -70)
 (DEFCONSTANT !NODTAMKERR -71)
 (DEFCONSTANT !BADDCKSUM -72)
 (DEFCONSTANT !BADDBTSLP -73)
 (DEFCONSTANT !WRUNDERRUN -74)
 (DEFCONSTANT !CANTSTEPERR -75)
 (DEFCONSTANT !TK0BADERR -76)
 (DEFCONSTANT !INITIWMERR -77)
 (DEFCONSTANT !TWOSIDEERR -78)
 (DEFCONSTANT !SPDADJERR -79)
 (DEFCONSTANT !SEEKERR -80)
 (DEFCONSTANT !SECTNFERR -81)
 (DEFCONSTANT !FMT1ERR -82)
 (DEFCONSTANT !FMT2ERR -83)
 (DEFCONSTANT !VERERR -84)
 (DEFCONSTANT !CLKRDERR -85)
 (DEFCONSTANT !CLKWRERR -86)
 (DEFCONSTANT !PRWRERR -87)
 (DEFCONSTANT !PRINITERR -88)
 (DEFCONSTANT !RCVRERR -89)
 (DEFCONSTANT !BREAKRECD -90)
 (DEFCONSTANT !DDPSKTERR -91)
 (DEFCONSTANT !DDPLENERR -92)
 (DEFCONSTANT !NOBRIDGEERR -93)
 (DEFCONSTANT !LAPPROTERR -94)
 (DEFCONSTANT !EXCESSCOLLSNS -95)
 (DEFCONSTANT !PORTINUSE -97)
 (DEFCONSTANT !PORTNOTCF -98)
 (DEFCONSTANT !MEMROZERR -99)
 (DEFCONSTANT !MEMROZWARN -99)
 (DEFCONSTANT !NOSCRAPERR -100)
 (DEFCONSTANT !NOTYPEERR -102)
 (DEFCONSTANT !MEMFULLERR -108)
 (DEFCONSTANT !NILHANDLEERR -109)
 (DEFCONSTANT !MEMWZERR -111)
 (DEFCONSTANT !MEMPURERR -112)
 (DEFCONSTANT !MEMADRERR -110)
 (DEFCONSTANT !MEMAZERR -113)
 (DEFCONSTANT !MEMPCERR -114)
 (DEFCONSTANT !MEMBCERR -115)
 (DEFCONSTANT !MEMSCERR -116)
 (DEFCONSTANT !MEMLOCKEDERR -117)
 (DEFCONSTANT !DIRNFERR -120)
 (DEFCONSTANT !TMWDOERR -121)
 (DEFCONSTANT !BADMOVERR -122)
 (DEFCONSTANT !WRGVOLTYPERR -123)
 (DEFCONSTANT !VOLGONEERR -124)
 (DEFCONSTANT !RESNOTFOUND -192)
 (DEFCONSTANT !RESFNOTFOUND -193)
 (DEFCONSTANT !ADDRESFAILED -194)
 (DEFCONSTANT !ADDREFFAILED -195)
 (DEFCONSTANT !RMVRESFAILED -196)
 (DEFCONSTANT !RMVREFFAILED -197)
 (DEFCONSTANT !RESATTRERR -198)
 (DEFCONSTANT !MAPREADERR -199)
 (DEFCONSTANT !NBPBUFFOVR -1024)
 (DEFCONSTANT !NBPNOCONFIRM -1025)
 (DEFCONSTANT !NBPCONFDIFF -1026)
 (DEFCONSTANT !NBPDUPLICATE -1027)
 (DEFCONSTANT !NBPNOTFOUND -1028)
 (DEFCONSTANT !NBPNISERR -1029)
 (DEFCONSTANT !ASPBADVERSNUM -1066)
 (DEFCONSTANT !ASPBUFTOOSMALL -1067)
 (DEFCONSTANT !ASPNOMORESESS -1068)
 (DEFCONSTANT !ASPNOSERVERS -1069)
 (DEFCONSTANT !ASPPARAMERR -1070)
 (DEFCONSTANT !ASPSERVERBUSY -1071)
 (DEFCONSTANT !ASPSESSCLOSED -1072)
 (DEFCONSTANT !ASPSIZEERR -1073)
 (DEFCONSTANT !ASPTOOMANY -1074)
 (DEFCONSTANT !ASPNOACK -1075)
 (DEFCONSTANT !REQFAILED -1096)
 (DEFCONSTANT !TOOMANYREQS -1097)
 (DEFCONSTANT !TOOMANYSKTS -1098)
 (DEFCONSTANT !BADATPSKT -1099)
 (DEFCONSTANT !BADBUFFNUM -1100)
 (DEFCONSTANT !NORELERR -1101)
 (DEFCONSTANT !CBNOTFOUND -1102)
 (DEFCONSTANT !NOSENDRESP -1103)
 (DEFCONSTANT !NODATAAREA -1104)
 (DEFCONSTANT !REQABORTED -1105)
 (DEFCONSTANT !BUF2SMALLERR -3101)
 (DEFCONSTANT !NOMPPERR -3102)
 (DEFCONSTANT !CKSUMERR -3103)
 (DEFCONSTANT !EXTRACTERR -3104)
 (DEFCONSTANT !READQERR -3105)
 (DEFCONSTANT !ATPLENERR -3106)
 (DEFCONSTANT !ATPBADRSP -3107)
 (DEFCONSTANT !RECNOTFND -3108)
 (DEFCONSTANT !SKTCLOSEDERR -3109)
 (DEFCONSTANT !AFPACCESSDENIED -5000)
 (DEFCONSTANT !AFPAUTHCONTINUE -5001)
 (DEFCONSTANT !AFPBADUAM -5002)
 (DEFCONSTANT !AFPBADVERSNUM -5003)
 (DEFCONSTANT !AFPBITMAPERR -5004)
 (DEFCONSTANT !AFPCANTMOVE -5005)
 (DEFCONSTANT !AFPDENYCONFLICT -5006)
 (DEFCONSTANT !AFPDIRNOTEMPTY -5007)
 (DEFCONSTANT !AFPDISKFULL -5008)
 (DEFCONSTANT !AFPEOFERROR -5009)
 (DEFCONSTANT !AFPFILEBUSY -5010)
 (DEFCONSTANT !AFPFLATVOL -5011)
 (DEFCONSTANT !AFPITEMNOTFOUND -5012)
 (DEFCONSTANT !AFPLOCKERR -5013)
 (DEFCONSTANT !AFPMISCERR -5014)
 (DEFCONSTANT !AFPNOMORELOCKS -5015)
 (DEFCONSTANT !AFPNOSERVER -5016)
 (DEFCONSTANT !AFPOBJECTEXISTS -5017)
 (DEFCONSTANT !AFPOBJECTNOTFOUND -5018)
 (DEFCONSTANT !AFPPARMERR -5019)
 (DEFCONSTANT !AFPRANGENOTLOCKED -5020)
 (DEFCONSTANT !AFPRANGEOVERLAP -5021)
 (DEFCONSTANT !AFPSESSCLOSED -5022)
 (DEFCONSTANT !AFPUSERNOTAUTH -5023)
 (DEFCONSTANT !AFPCALLNOTSUPPORTED -5024)
 (DEFCONSTANT !AFPOBJECTTYPEERR -5025)
 (DEFCONSTANT !AFPTOOMANYFILESOPEN -5026)
 (DEFCONSTANT !AFPSERVERGOINGDOWN -5027)
 (DEFCONSTANT !AFPCANTRENAME -5028)
 (DEFCONSTANT !AFPDIRNOTFOUND -5029)
 (DEFCONSTANT !AFPICONTYPEERROR -5030)
 (DEFCONSTANT !ENVNOTPRESENT -5500)
 (DEFCONSTANT !ENVBADVERS -5501)
 (DEFCONSTANT !ENVVERSTOOBIG -5502)
 (DEFCONSTANT !EVTNOTENB 1)
 (DEFCONSTANT !DSSYSERR 32767)
 (DEFCONSTANT !DSBUSERROR 1)
 (DEFCONSTANT !DSADDRESSERR 2)
 (DEFCONSTANT !DSILLINSTERR 3)
 (DEFCONSTANT !DSZERODIVERR 4)
 (DEFCONSTANT !DSCHKERR 5)
 (DEFCONSTANT !DSOVFLOWERR 6)
 (DEFCONSTANT !DSPRIVERR 7)
 (DEFCONSTANT !DSTRACEERR 8)
 (DEFCONSTANT !DSLINEAERR 9)
 (DEFCONSTANT !DSLINEFERR 10)
 (DEFCONSTANT !DSMISCERR 11)
 (DEFCONSTANT !DSCOREERR 12)
 (DEFCONSTANT !DSIRQERR 13)
 (DEFCONSTANT !DSIOCOREERR 14)
 (DEFCONSTANT !DSLOADERR 15)
 (DEFCONSTANT !DSFPERR 16)
 (DEFCONSTANT !DSNOPACKERR 17)
 (DEFCONSTANT !DSNOPK1 18)
 (DEFCONSTANT !DSNOPK2 19)
 (DEFCONSTANT !DSNOPK3 20)
 (DEFCONSTANT !DSNOPK4 21)
 (DEFCONSTANT !DSNOPK5 22)
 (DEFCONSTANT !DSNOPK6 23)
 (DEFCONSTANT !DSNOPK7 24)
 (DEFCONSTANT !DSMEMFULLERR 25)
 (DEFCONSTANT !DSBADLAUNCH 26)
 (DEFCONSTANT !DSFSERR 27)
 (DEFCONSTANT !DSSTKNHEAP 28)
 (DEFCONSTANT !DSREINSERT 30)
 (DEFCONSTANT !DSNOTTHE1 31)
 (DEFCONSTANT !NEGZCBFREEERR 33)
 (DEFCONSTANT !DSGREETING 40)
 (DEFCONSTANT !DSFINDERERR 41)
 (DEFCONSTANT !SHUTDOWNALERT 42)
 (DEFCONSTANT !MENUPRGERR 84)
 (DEFCONSTANT !SWOVERRUNERR 1)
 (DEFCONSTANT !PARITYERR 16)
 (DEFCONSTANT !HWOVERRUNERR 32)
 (DEFCONSTANT !FRAMINGERR 64)
 (DEFCONSTANT !CMATCHERR -150)
 (DEFCONSTANT !CTEMPMEMERR -151)
 (DEFCONSTANT !CNOMEMERR -152)
 (DEFCONSTANT !CRANGEERR -153)
 (DEFCONSTANT !CPROTECTERR -154)
 (DEFCONSTANT !CDEVERR -155)
 (DEFCONSTANT !CRESERR -156)
 (DEFCONSTANT !ITABPURGERR -9)
 (DEFCONSTANT !NOCOLMATCH -10)
 (DEFCONSTANT !QALLOCERR -11)
 (DEFCONSTANT !TBLALLOCERR -12)
 (DEFCONSTANT !OVERRUN -13)
 (DEFCONSTANT !NOROOMERR -14)
 (DEFCONSTANT !SEOUTOFRANGE -15)
 (DEFCONSTANT !SEPROTERR -16)
 (DEFCONSTANT !I2CRANGEERR -17)
 (DEFCONSTANT !GDBADDEV -18)
 (DEFCONSTANT !RERANGEERR -19)
 (DEFCONSTANT !SEINVREQUEST -20)
 (DEFCONSTANT !SENOMEMERR -21)
 (DEFCONSTANT !UNITTBLFULLERR -29)
 (DEFCONSTANT !DCEEXTERR -30)
 (DEFCONSTANT !DSBADSLOTINT 51)
 (DEFCONSTANT !DSBADSANEOPCODE 81)
 (DEFCONSTANT !UPDPIXMEMERR -125)
 (DEFCONSTANT !MBARNFND -126)
 (DEFCONSTANT !HMENUFINDERR -127)
 (DEFCONSTANT !NOHARDWARE -200)
 (DEFCONSTANT !NOTENOUGHHARDWARE -201)
 (DEFCONSTANT !QUEUEFULL -203)
 (DEFCONSTANT !RESPROBLEM -204)
 (DEFCONSTANT !BADCHANNEL -205)
 (DEFCONSTANT !BADFORMAT -206)
 (DEFCONSTANT !SMSDMINITERR -290)
 (DEFCONSTANT !SMSRTINITERR -291)
 (DEFCONSTANT !SMPRAMINITERR -292)
 (DEFCONSTANT !SMPRIINITERR -293)
 (DEFCONSTANT !SMEMPTYSLOT -300)
 (DEFCONSTANT !SMCRCFAIL -301)
 (DEFCONSTANT !SMFORMATERR -302)
 (DEFCONSTANT !SMREVISIONERR -303)
 (DEFCONSTANT !SMNODIR -304)
 (DEFCONSTANT !SMLWTSTBAD -305)
 (DEFCONSTANT !SMNOSINFOARRAY -306)
 (DEFCONSTANT !SMRESRVERR -307)
 (DEFCONSTANT !SMUNEXBUSERR -308)
 (DEFCONSTANT !SMBLFIELDBAD -309)
 (DEFCONSTANT !SMFHBLOCKRDERR -310)
 (DEFCONSTANT !SMFHBLKDISPERR -311)
 (DEFCONSTANT !SMDISPOSEPERR -312)
 (DEFCONSTANT !SMNOBOARDSRSRC -313)
 (DEFCONSTANT !SMGETPRERR -314)
 (DEFCONSTANT !SMNOBOARDID -315)
 (DEFCONSTANT !SMINTSTATVERR -316)
 (DEFCONSTANT !SMINTTBLVERR -317)
 (DEFCONSTANT !SMNOJMPTBL -318)
 (DEFCONSTANT !SMBADBOARDID -319)
 (DEFCONSTANT !SMBUSERRTO -320)
 (DEFCONSTANT !SMBADREFID -330)
 (DEFCONSTANT !SMBADSLIST -331)
 (DEFCONSTANT !SMRESERVEDERR -332)
 (DEFCONSTANT !SMCODEREVERR -333)
 (DEFCONSTANT !SMCPUER -334)
 (DEFCONSTANT !SMSPOINTERNIL -335)
 (DEFCONSTANT !SMNILSBLOCKERR -336)
 (DEFCONSTANT !SMSLOTOOBERR -337)
 (DEFCONSTANT !SMSELOOBERR -338)
 (DEFCONSTANT !SMNEWPERR -339)
 (DEFCONSTANT !SMBLKMOVEERR -340)
 (DEFCONSTANT !SMCKSTATUSERR -341)
 (DEFCONSTANT !SMGETDRVRNAMERR -342)
 (DEFCONSTANT !SMDISDRVRNAMERR -343)
 (DEFCONSTANT !SMNOMORESRSRCS -344)
 (DEFCONSTANT !SMSGETDRVRERR -345)
 (DEFCONSTANT !SMBADSPTRERR -346)
 (DEFCONSTANT !SMBYTELANESERR -347)
 (DEFCONSTANT !SMOFFSETERR -348)
 (DEFCONSTANT !SMNOGOODOPENS -349)
 (DEFCONSTANT !SMSRTOVRFLERR -350)
 (DEFCONSTANT !SMRECNOTFND -351)
 (DEFCONSTANT !SLOTNUMERR -360)
 (DEFCONSTANT !SIINITSDTBLERR 1)
 (DEFCONSTANT !SIINITVBLQSERR 2)
 (DEFCONSTANT !SIINITSPTBLERR 3)
 (DEFCONSTANT !SDMJTINITERR 10)
 (DEFCONSTANT !SDMINITERR 11)
 (DEFCONSTANT !SDMSRTINITERR 12)
 (DEFCONSTANT !SDMPRAMINITERR 13)
 (DEFCONSTANT !SDMPRIINITERR 14)
 (DEFCONSTANT !DSMBARNFND 85)
 (DEFCONSTANT !DSHMENUFINDERR 86)

;  Constants from events.h

 (DEFCONSTANT !NULLEVENT 0)
 (DEFCONSTANT !MOUSEDOWN 1)
 (DEFCONSTANT !MOUSEUP 2)
 (DEFCONSTANT !KEYDOWN 3)
 (DEFCONSTANT !KEYUP 4)
 (DEFCONSTANT !AUTOKEY 5)
 (DEFCONSTANT !UPDATEEVT 6)
 (DEFCONSTANT !DISKEVT 7)
 (DEFCONSTANT !ACTIVATEEVT 8)
 (DEFCONSTANT !NETWORKEVT 10)
 (DEFCONSTANT !DRIVEREVT 11)
 (DEFCONSTANT !APP1EVT 12)
 (DEFCONSTANT !APP2EVT 13)
 (DEFCONSTANT !APP3EVT 14)
 (DEFCONSTANT !APP4EVT 15)
 (DEFCONSTANT !CHARCODEMASK 255)   ;0X000000FF)                     ***---***
 (DEFCONSTANT !KEYCODEMASK 65280)   ;0X0000FF00)                     ***---***
 (DEFCONSTANT !ADBADDRMASK 10158080)   ;0X00FF0000)                     ***---***
 (DEFCONSTANT !MDOWNMASK 2)
 (DEFCONSTANT !MUPMASK 4)
 (DEFCONSTANT !KEYDOWNMASK 8)
 (DEFCONSTANT !KEYUPMASK 16)
 (DEFCONSTANT !AUTOKEYMASK 32)
 (DEFCONSTANT !UPDATEMASK 64)
 (DEFCONSTANT !DISKMASK 128)
 (DEFCONSTANT !ACTIVMASK 256)
 (DEFCONSTANT !NETWORKMASK 1024)
 (DEFCONSTANT !DRIVERMASK 2048)
 (DEFCONSTANT !APP1MASK 4096)
 (DEFCONSTANT !APP2MASK 8192)
 (DEFCONSTANT !APP3MASK 16384)
 (DEFCONSTANT !APP4MASK #x8000)   ;(-32768))                     ***---***
 (DEFCONSTANT !EVERYEVENT -1)   ;(-1))                     ***---***
 (DEFCONSTANT !ACTIVEFLAG 1)
 (DEFCONSTANT !BTNSTATE 128)
 (DEFCONSTANT !CMDKEY 256)
 (DEFCONSTANT !SHIFTKEY 512)
 (DEFCONSTANT !ALPHALOCK 1024)
 (DEFCONSTANT !OPTIONKEY 2048)
 (DEFCONSTANT !CONTROLKEY 4096)

;  Constants from files.h

 (DEFCONSTANT !FONDESK 1)
 (DEFCONSTANT !FHASBUNDLE 8192)
 (DEFCONSTANT !FINVISIBLE 16384)
 (DEFCONSTANT !FTRASH -3)   ;(-3))                     ***---***
 (DEFCONSTANT !FDESKTOP -2)   ;(-2))                     ***---***
 (DEFCONSTANT !FDISK 0)
 (DEFCONSTANT !FSCURPERM 0)
 (DEFCONSTANT !FSRDPERM 1)
 (DEFCONSTANT !FSWRPERM 2)
 (DEFCONSTANT !FSRDWRPERM 3)
 (DEFCONSTANT !FSRDWRSHPERM 4)
 (DEFCONSTANT !FSATMARK 0)
 (DEFCONSTANT !FSFROMSTART 1)
 (DEFCONSTANT !FSFROMLEOF 2)
 (DEFCONSTANT !FSFROMMARK 3)
 (DEFCONSTANT !RDVERIFY 64)
 (DEFCONSTANT !IODIRFLG 3)
 (DEFCONSTANT !IODIRMASK 16)   ;(1 << 4))                     ***---***
 (DEFCONSTANT !FSRTPARID 1)
 (DEFCONSTANT !FSRTDIRID 2)

;  Constants from fonts.h

 (DEFCONSTANT !SYSTEMFONT 0)
 (DEFCONSTANT !APPLFONT 1)
 (DEFCONSTANT !NEWYORK 2)
 (DEFCONSTANT !GENEVA 3)
 (DEFCONSTANT !MONACO 4)
 (DEFCONSTANT !VENICE 5)
 (DEFCONSTANT !LONDON 6)
 (DEFCONSTANT !ATHENS 7)
 (DEFCONSTANT !SANFRAN 8)
 (DEFCONSTANT !TORONTO 9)
 (DEFCONSTANT !CAIRO 11)
 (DEFCONSTANT !LOSANGELES 12)
 (DEFCONSTANT !TIMES 20)
 (DEFCONSTANT !HELVETICA 21)
 (DEFCONSTANT !COURIER 22)
 (DEFCONSTANT !SYMBOL 23)
 (DEFCONSTANT !MOBILE 24)
 (DEFCONSTANT !COMMANDMARK (CODE-CHAR #o21))   ;(QUOTE 021))                     ***---***
 (DEFCONSTANT !CHECKMARK (CODE-CHAR #o22))   ;(QUOTE 022))                     ***---***
 (DEFCONSTANT !DIAMONDMARK (CODE-CHAR #o23))   ;(QUOTE 023))                     ***---***
 (DEFCONSTANT !APPLEMARK (CODE-CHAR #o24))   ;(QUOTE 024))                     ***---***
 (DEFCONSTANT !PROPFONT #x9000)   ;0X9000)                     ***---***
 (DEFCONSTANT !PRPFNTH #x9001)   ;0X9001)                     ***---***
 (DEFCONSTANT !PRPFNTW #x9002)   ;0X9002)                     ***---***
 (DEFCONSTANT !PRPFNTHW #x9003)   ;0X9003)                     ***---***
 (DEFCONSTANT !FIXEDFONT #xB000)   ;0XB000)                     ***---***
 (DEFCONSTANT !FXDFNTH #xB001)   ;0XB001)                     ***---***
 (DEFCONSTANT !FXDFNTW #xB002)   ;0XB002)                     ***---***
 (DEFCONSTANT !FXDFNTHW #xB003)   ;0XB003)                     ***---***
 (DEFCONSTANT !FONTWID #xACB0)   ;0XACB0)                     ***---***

;  Constants from memory.h

 (DEFCONSTANT !MAXSIZE #x800000)   ;0X800000)                     ***---***
 (DEFCONSTANT !OSDISPATCH #xA88F)   ;0XA88F)                     ***---***
 (DEFCONSTANT !MFMAXMEMSEL 21)
 (DEFCONSTANT !MFFREEMEMSEL 24)
 (DEFCONSTANT !MFTEMPNEWHANDLESEL 29)
 (DEFCONSTANT !MFTEMPHLOCKSEL 30)
 (DEFCONSTANT !MFTEMPHUNLOCKSEL 31)
 (DEFCONSTANT !MFTEMPDISPOSHANDLESEL 32)

;  Constants from packages.h

 (DEFCONSTANT !LISTMGR 0)
 (DEFCONSTANT !DSKINIT 2)
 (DEFCONSTANT !STDFILE 3)
 (DEFCONSTANT !FLPOINT 4)
 (DEFCONSTANT !TRFUNC 5)
 (DEFCONSTANT !INTUTIL 6)
 (DEFCONSTANT !BDCONV 7)
 (DEFCONSTANT !PUTDLGID -3999)   ;(-3999))                     ***---***
 (DEFCONSTANT !PUTSAVE 1)
 (DEFCONSTANT !PUTCANCEL 2)
 (DEFCONSTANT !PUTEJECT 5)
 (DEFCONSTANT !PUTDRIVE 6)
 (DEFCONSTANT !PUTNAME 7)
 (DEFCONSTANT !GETDLGID -4000)   ;(-4000))                     ***---***
 (DEFCONSTANT !GETOPEN 1)
 (DEFCONSTANT !GETCANCEL 3)
 (DEFCONSTANT !GETEJECT 5)
 (DEFCONSTANT !GETDRIVE 6)
 (DEFCONSTANT !GETNMLIST 7)
 (DEFCONSTANT !GETSCROLL 8)
 (DEFCONSTANT !CURRSYMLEAD 16)
 (DEFCONSTANT !CURRNEGSYM 32)
 (DEFCONSTANT !CURRTRAILINGZ 64)
 (DEFCONSTANT !CURRLEADINGZ 128)
 (DEFCONSTANT !MDY 0)
 (DEFCONSTANT !DMY 1)
 (DEFCONSTANT !YMD 2)
 (DEFCONSTANT !DAYLDINGZ 32)
 (DEFCONSTANT !MNTLDINGZ 64)
 (DEFCONSTANT !CENTURY 128)
 (DEFCONSTANT !SECLEADINGZ 32)
 (DEFCONSTANT !MINLEADINGZ 64)
 (DEFCONSTANT !HRLEADINGZ 128)
 (DEFCONSTANT !VERUS 0)
 (DEFCONSTANT !VERFRANCE 1)
 (DEFCONSTANT !VERBRITAIN 2)
 (DEFCONSTANT !VERGERMANY 3)
 (DEFCONSTANT !VERITALY 4)
 (DEFCONSTANT !VERNETHERLANDS 5)
 (DEFCONSTANT !VERBELGIUMLUX 6)
 (DEFCONSTANT !VERSWEDEN 7)
 (DEFCONSTANT !VERSPAIN 8)
 (DEFCONSTANT !VERDENMARK 9)
 (DEFCONSTANT !VERPORTUGAL 10)
 (DEFCONSTANT !VERFRCANADA 11)
 (DEFCONSTANT !VERNORWAY 12)
 (DEFCONSTANT !VERISRAEL 13)
 (DEFCONSTANT !VERJAPAN 14)
 (DEFCONSTANT !VERAUSTRALIA 15)
 (DEFCONSTANT !VERARABIA 16)
 (DEFCONSTANT !VERFINLAND 17)
 (DEFCONSTANT !VERFRSWISS 18)
 (DEFCONSTANT !VERGRSWISS 19)
 (DEFCONSTANT !VERGREECE 20)
 (DEFCONSTANT !VERICELAND 21)
 (DEFCONSTANT !VERMALTA 22)
 (DEFCONSTANT !VERCYPRUS 23)
 (DEFCONSTANT !VERTURKEY 24)
 (DEFCONSTANT !VERYUGOSLAVIA 25)

;  Constants from palette.h

 (DEFCONSTANT !PMCOURTEOUS 0)
 (DEFCONSTANT !PMTOLERANT 2)
 (DEFCONSTANT !PMANIMATED 4)
 (DEFCONSTANT !PMEXPLICIT 8)
 (DEFCONSTANT !PMALLCUPDATES #xC000)   ;0XC000)                     ***---***
 (DEFCONSTANT !PMBACKCUPDATES #x8000)   ;0X8000)                     ***---***
 (DEFCONSTANT !PMFRONTCUPDATES #x4000)   ;0X4000)                     ***---***

;  Constants from printing.h

 (DEFCONSTANT !BDRAFTLOOP 0)
 (DEFCONSTANT !BSPOOLLOOP 1)
 (DEFCONSTANT !IPFMAXPGS 128)
 (DEFCONSTANT !IPRPGFRACT 120)
 (DEFCONSTANT !IPRSAVPFIL -1)   ;(-1))                     ***---***
 (DEFCONSTANT !IPRABORT 128)
 (DEFCONSTANT !IPRDEVCTL 7)
 (DEFCONSTANT !LPRRESET #x10000)   ;0X00010000)                     ***---***
 (DEFCONSTANT !LPRLINEFEED #x30000)   ;0X00030000)                     ***---***
 (DEFCONSTANT !LPRLFSIXTH #x3FFFF)   ;0X0003FFFF)                     ***---***
 (DEFCONSTANT !LPRPAGEEND #x20000)   ;0X00020000)                     ***---***
 (DEFCONSTANT !LPRDOCOPEN #x10000)   ;0X00010000)                     ***---***
 (DEFCONSTANT !LPRPAGEOPEN #x40000)   ;0X00040000)                     ***---***
 (DEFCONSTANT !LPRPAGECLOSE #x20000)   ;0X00020000)                     ***---***
 (DEFCONSTANT !LPRDOCCLOSE #x50000)   ;0X00050000)                     ***---***
 (DEFCONSTANT !IPRBITSCTL 4)
 (DEFCONSTANT !LSCREENBITS 0)
 (DEFCONSTANT !LPAINTBITS 1)
 (DEFCONSTANT !IPRIOCTL 5)
 (DEFCONSTANT !SPRDRVR ".Print")   ;.Print)                     ***---***
 (DEFCONSTANT !IPRDRVRREF -3)   ;(-3))                     ***---***
 (DEFCONSTANT !GETRSLDATAOP 4)
 (DEFCONSTANT !SETRSLOP 5)
 (DEFCONSTANT !DRAFTBITSOP 6)
 (DEFCONSTANT !NODRAFTBITSOP 7)
 (DEFCONSTANT !GETROTNOP 8)
 (DEFCONSTANT !NOSUCHRSL 1)
 (DEFCONSTANT !RGTYPE1 1)

;  Constants from quickdraw.h

 (DEFCONSTANT !SRCCOPY 0)
 (DEFCONSTANT !SRCOR 1)
 (DEFCONSTANT !SRCXOR 2)
 (DEFCONSTANT !SRCBIC 3)
 (DEFCONSTANT !NOTSRCCOPY 4)
 (DEFCONSTANT !NOTSRCOR 5)
 (DEFCONSTANT !NOTSRCXOR 6)
 (DEFCONSTANT !NOTSRCBIC 7)
;(DEFCONSTANT !PATCOPY 8)
;(DEFCONSTANT !PATOR 9)
;(DEFCONSTANT !PATXOR 10)
;(DEFCONSTANT !PATBIC 11)
;(DEFCONSTANT !NOTPATCOPY 12)
;(DEFCONSTANT !NOTPATOR 13)
;(DEFCONSTANT !NOTPATXOR 14)
;(DEFCONSTANT !NOTPATBIC 15)
 (DEFCONSTANT !NORMALBIT 0)
 (DEFCONSTANT !INVERSEBIT 1)
 (DEFCONSTANT !REDBIT 4)
 (DEFCONSTANT !GREENBIT 3)
 (DEFCONSTANT !BLUEBIT 2)
 (DEFCONSTANT !CYANBIT 8)
 (DEFCONSTANT !MAGENTABIT 7)
 (DEFCONSTANT !YELLOWBIT 6)
 (DEFCONSTANT !BLACKBIT 5)
 (DEFCONSTANT !BLACKCOLOR 33)
 (DEFCONSTANT !WHITECOLOR 30)
 (DEFCONSTANT !REDCOLOR 205)
 (DEFCONSTANT !GREENCOLOR 341)
 (DEFCONSTANT !BLUECOLOR 409)
 (DEFCONSTANT !CYANCOLOR 273)
 (DEFCONSTANT !MAGENTACOLOR 137)
 (DEFCONSTANT !YELLOWCOLOR 69)
 (DEFCONSTANT !PICLPAREN 0)
 (DEFCONSTANT !PICRPAREN 1)
;(DEFCONSTANT !NORMAL #x0)   ;0X00)                     ***---***
;(DEFCONSTANT !BOLD #x1)   ;0X01)                     ***---***
;(DEFCONSTANT !ITALIC #x2)   ;0X02)                     ***---***
;(DEFCONSTANT !UNDERLINE #x4)   ;0X04)                     ***---***
;(DEFCONSTANT !OUTLINE #x8)   ;0X08)                     ***---***
;(DEFCONSTANT !SHADOW #x10)   ;0X10)                     ***---***
;(DEFCONSTANT !CONDENSE #x20)   ;0X20)                     ***---***
;(DEFCONSTANT !EXTEND #x40)   ;0X40)                     ***---***
 (DEFCONSTANT !INVALCOLREQ -1)
 (DEFCONSTANT !CLUTTYPE 0)
 (DEFCONSTANT !FIXEDTYPE 1)
 (DEFCONSTANT !DIRECTTYPE 2)
 (DEFCONSTANT !GDDEVTYPE 0)
 (DEFCONSTANT !RAMINIT 10)
 (DEFCONSTANT !MAINSCREEN 11)
 (DEFCONSTANT !ALLINIT 12)
 (DEFCONSTANT !SCREENDEVICE 13)
 (DEFCONSTANT !NODRIVER 14)
 (DEFCONSTANT !SCREENACTIVE 15)
 (DEFCONSTANT !HILITEBIT 7)
 (DEFCONSTANT !DEFQDCOLORS 127)

;  Constants from resources.h

 (DEFCONSTANT !RESSYSHEAP 64)
 (DEFCONSTANT !RESPURGEABLE 32)
 (DEFCONSTANT !RESLOCKED 16)
 (DEFCONSTANT !RESPROTECTED 8)
 (DEFCONSTANT !RESPRELOAD 4)
 (DEFCONSTANT !RESCHANGED 2)
 (DEFCONSTANT !MAPREADONLY 128)
 (DEFCONSTANT !MAPCOMPACT 64)
 (DEFCONSTANT !MAPCHANGED 32)

;  Constants from segload.h

 (DEFCONSTANT !APPOPEN 0)
 (DEFCONSTANT !APPPRINT 1)

;  Constants from serial.h

 (DEFCONSTANT !BAUD300 380)
 (DEFCONSTANT !BAUD600 189)
 (DEFCONSTANT !BAUD1200 94)
 (DEFCONSTANT !BAUD1800 62)
 (DEFCONSTANT !BAUD2400 46)
 (DEFCONSTANT !BAUD3600 30)
 (DEFCONSTANT !BAUD4800 22)
 (DEFCONSTANT !BAUD7200 14)
 (DEFCONSTANT !BAUD9600 10)
 (DEFCONSTANT !BAUD19200 4)
 (DEFCONSTANT !BAUD57600 0)
 (DEFCONSTANT !STOP10 16384)
 (DEFCONSTANT !STOP15 #x8000)   ;(-32768))                     ***---***
 (DEFCONSTANT !STOP20 #xC000)   ;(-16384))                     ***---***
 (DEFCONSTANT !NOPARITY 0)
 (DEFCONSTANT !ODDPARITY 4096)
 (DEFCONSTANT !EVENPARITY 12288)
 (DEFCONSTANT !DATA5 0)
 (DEFCONSTANT !DATA6 2048)
 (DEFCONSTANT !DATA7 1024)
 (DEFCONSTANT !DATA8 3072)
 (DEFCONSTANT !CTSEVENT 32)
 (DEFCONSTANT !BREAKEVENT 128)
 (DEFCONSTANT !XOFFWASSENT #x80)   ;0X80)                     ***---***
 (DEFCONSTANT !DTRNEGATED #x40)   ;0X40)                     ***---***

;  Constants from slots.h

 (DEFCONSTANT !FCARDISCHANGED 1)
 (DEFCONSTANT !FCKFORSAME 0)
 (DEFCONSTANT !FCKFORNEXT 1)
 (DEFCONSTANT !FWARMSTART 2)
 (DEFCONSTANT !STATENIL 0)
 (DEFCONSTANT !STATESDMINIT 1)
 (DEFCONSTANT !STATEPRAMINIT 2)
 (DEFCONSTANT !STATEPINIT 3)
 (DEFCONSTANT !STATESINIT 4)

;  Constants from sound.h

 (DEFCONSTANT !SWMODE -1)   ;(-1))                     ***---***
 (DEFCONSTANT !FTMODE 1)
 (DEFCONSTANT !FFMODE 0)
 (DEFCONSTANT !SYNTHCODERSRC "snth")   ;(QUOTE SNTH))                     ***---***
 (DEFCONSTANT !SOUNDLISTRSRC "snd ")   ;(QUOTE SND))                     ***---***
 (DEFCONSTANT !NOTESYNTH 2)
 (DEFCONSTANT !WAVETABLESYNTH 3)
 (DEFCONSTANT !MIDISYNTH 4)
 (DEFCONSTANT !SAMPLEDSYNTH 5)
 (DEFCONSTANT !TWELTHROOTTWO 1.0594631)
 (DEFCONSTANT !INFINITETIME #x7FFFFFFF)   ;0X7FFFFFFF)                     ***---***
 (DEFCONSTANT !NULLCMD 0)
 (DEFCONSTANT !INITCMD 1)
 (DEFCONSTANT !FREECMD 2)
 (DEFCONSTANT !QUIETCMD 3)
 (DEFCONSTANT !FLUSHCMD 4)
 (DEFCONSTANT !WAITCMD 10)
 (DEFCONSTANT !PAUSECMD 11)
 (DEFCONSTANT !RESUMECMD 12)
 (DEFCONSTANT !CALLBACKCMD 13)
 (DEFCONSTANT !SYNCCMD 14)
 (DEFCONSTANT !EMPTYCMD 15)
 (DEFCONSTANT !TICKLECMD 20)
 (DEFCONSTANT !REQUESTNEXTCMD 21)
 (DEFCONSTANT !HOWOFTENCMD 22)
 (DEFCONSTANT !WAKEUPCMD 23)
 (DEFCONSTANT !AVAILABLECMD 24)
 (DEFCONSTANT !NOTECMD 40)
 (DEFCONSTANT !RESTCMD 41)
 (DEFCONSTANT !FREQCMD 42)
 (DEFCONSTANT !AMPCMD 43)
 (DEFCONSTANT !TIMBRECMD 44)
 (DEFCONSTANT !WAVETABLECMD 60)
 (DEFCONSTANT !PHASECMD 61)
 (DEFCONSTANT !SOUNDCMD 80)
 (DEFCONSTANT !BUFFERCMD 81)
 (DEFCONSTANT !RATECMD 82)
 (DEFCONSTANT !MIDIDATACMD 100)
 (DEFCONSTANT !SETPTRBIT #x8000)   ;0X8000)                     ***---***
 (DEFCONSTANT !STDQLENGTH 128)
 (DEFCONSTANT !INITCHANLEFT #x2)   ;0X02)                     ***---***
 (DEFCONSTANT !INITCHANRIGHT #x3)   ;0X03)                     ***---***
 (DEFCONSTANT !INITCHAN0 #x4)   ;0X04)                     ***---***
 (DEFCONSTANT !INITCHAN1 #x5)   ;0X05)                     ***---***
 (DEFCONSTANT !INITCHAN2 #x6)   ;0X06)                     ***---***
 (DEFCONSTANT !INITCHAN3 #x7)   ;0X07)                     ***---***
 (DEFCONSTANT !INITSRATE22K #x20)   ;0X20)                     ***---***
 (DEFCONSTANT !INITSRATE44K #x30)   ;0X30)                     ***---***
 (DEFCONSTANT !INITMONO #x80)   ;0X80)                     ***---***
 (DEFCONSTANT !INITSTEREO #xC0)   ;0XC0)                     ***---***
 (DEFCONSTANT !FIRSTSOUNDFORMAT #x1)   ;0X0001)                     ***---***

;  Constants from textedit.h

 (DEFCONSTANT !TEJUSTLEFT 0)
 (DEFCONSTANT !TEJUSTCENTER 1)
 (DEFCONSTANT !TEJUSTRIGHT -1)   ;(-1))                     ***---***
 (DEFCONSTANT !TEFORCELEFT -2)   ;(-2))                     ***---***
 (DEFCONSTANT !DOFONT 1)
 (DEFCONSTANT !DOFACE 2)
 (DEFCONSTANT !DOSIZE 4)
 (DEFCONSTANT !DOCOLOR 8)
 (DEFCONSTANT !DOALL 15)
 (DEFCONSTANT !ADDSIZE 16)

;  Constants from toolutils.h

 (DEFCONSTANT !SYSPATLISTID 0)
 (DEFCONSTANT !IBEAMCURSOR 1)
 (DEFCONSTANT !CROSSCURSOR 2)
 (DEFCONSTANT !PLUSCURSOR 3)
 (DEFCONSTANT !WATCHCURSOR 4)

;  Constants from windows.h

 (DEFCONSTANT !DOCUMENTPROC 0)
 (DEFCONSTANT !DBOXPROC 1)
 (DEFCONSTANT !PLAINDBOX 2)
 (DEFCONSTANT !ALTDBOXPROC 3)
 (DEFCONSTANT !NOGROWDOCPROC 4)
 (DEFCONSTANT !ZOOMDOCPROC 8)
 (DEFCONSTANT !ZOOMNOGROW 12)
 (DEFCONSTANT !RDOCPROC 16)
 (DEFCONSTANT !DIALOGKIND 2)
 (DEFCONSTANT !USERKIND 8)
 (DEFCONSTANT !INDESK 0)
 (DEFCONSTANT !INMENUBAR 1)
 (DEFCONSTANT !INSYSWINDOW 2)
 (DEFCONSTANT !INCONTENT 3)
 (DEFCONSTANT !INDRAG 4)
 (DEFCONSTANT !INGROW 5)
 (DEFCONSTANT !INGOAWAY 6)
 (DEFCONSTANT !INZOOMIN 7)
 (DEFCONSTANT !INZOOMOUT 8)
 (DEFCONSTANT !WDRAW 0)
 (DEFCONSTANT !WHIT 1)
 (DEFCONSTANT !WCALCRGNS 2)
 (DEFCONSTANT !WNEW 3)
 (DEFCONSTANT !WDISPOSE 4)
 (DEFCONSTANT !WGROW 5)
 (DEFCONSTANT !WDRAWGICON 6)
 (DEFCONSTANT !WNOHIT 0)
 (DEFCONSTANT !WINCONTENT 1)
 (DEFCONSTANT !WINDRAG 2)
 (DEFCONSTANT !WINGROW 3)
 (DEFCONSTANT !WINGOAWAY 4)
 (DEFCONSTANT !WINZOOMIN 5)
 (DEFCONSTANT !WINZOOMOUT 6)
 (DEFCONSTANT !DESKPATID 16)
 (DEFCONSTANT !WCONTENTCOLOR 0)
 (DEFCONSTANT !WFRAMECOLOR 1)
 (DEFCONSTANT !WTEXTCOLOR 2)
 (DEFCONSTANT !WHILITECOLOR 3)
 (DEFCONSTANT !WTITLEBARCOLOR 4)
))
|#