;;;-*- Mode:Common-Lisp; Package:USER; Base:10 -*-
;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.


;;;
;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;;   5/09/89  clm      Moved `debug' module to the CLEH defsystem because of dependencies
;;;                     which require CLEH be built before the UCL-debugger.
;;;   3/19/87  DAN	Readded `debug' module.
;;;   3/12/87  DAN	Changed PATCHABLE pathname to new convention and added WARNINGS pathname.
;;;   3/06/87  DAN      Removed `debug' module. It still has problems.
;;;   2/27/87  TWE	Removed the SYS host prefix from the `debug' module since it is
;;;			specified in the pathname default.  This caused problems with
;;;			the change history analysis program.
;;;   2/26/87  DAN	Added UCL-debugger to the defsystem. It has to be loaded after UCL is built.


(DEFSYSTEM UCL
  (:NAME "Universal Command Loop")
  (:SHORT-NAME "UCL")
  (:NICKNAMES "UCL")
  (:PATHNAME-DEFAULT "SYS:UCL;")
  (:PATCHABLE "SYS:PATCH.UCL;" PATCH)
  (:warnings-pathname-default "sys:cwarns;ucl.lisp")

  (:MODULE break	"break-patch")
  (:MODULE comm		"command")		; UCL:COMMAND flavor definition.  Holds all of a command's data.
  (:MODULE comp		"completion")		; Generally useful functions for completion on Lisp and non-Lisp symbols.
  (:MODULE dcom		"defcom")		; Utilities for Programmers to set up the required UCL data structures.

  (:MODULE defs		"definitions")		; Macros and variables used by UCL
  (:MODULE disp		"display")		; Command definitions for the Command Display.

  (:MODULE dsup		"def-support")		; Functions used by UCL macros.
  (:MODULE edit		"edit")			; Command definitions for the Command Editor.
  (:MODULE ekeys	"edit-keys")		; Command definitions for the Keysroke Editor (a part of the Command Editor).
  (:MODULE ename	"edit-names")		; Command definitions for the Name Editor (a part of the Command Editor).
  (:MODULE loop		"command-loop")		; Flavor definitions and command loop methods of the UCL.
  (:MODULE menus	"menus")		; Code for building and maintaining UCL command menus.
  (:MODULE modes	"top-level-modes")	; Flavors for completing on and processing typed expressions.
  (:MODULE num		"numeric-arg")		; Flavor definition of a UCL loop which processes numeric arguments for UCL.
  (:MODULE select	"selective-features")	; Flavor definition to mix into UCL to selectively cut out features.
  (:MODULE table	"table")		; UCL:COMMAND-TABLE flavor definition. Groups commands for name & key lookup.
  (:MODULE temp		"temporary-commands")	; Utility functions for creating on-the-fly commands and command tables.
  (:MODULE univ		"universal-commands")   ; Universal commands given to every UCL application.
  (:MODULE util		"utility")
  (:MODULE wind		"window")		; Flavor definition of window used by UCL utilities such as Command Display,
						;  Command Editor, Name Editor, and Keystroke Editor.  There are several
						;  tricks done here.  Note that these utilities make use of the UCL.

;  (:MODULE debug        "EH;ucl-debugger")      ; UCL-debugger

  (:MODULE kit		("starter-kit"))	; the starter kit
  
  (:AUXILIARY kit)

  (:COMPILE-LOAD      dsup)
  (:COMPILE-LOAD-INIT defs  (dsup)	(:FASLOAD dsup) (:FASLOAD dsup))
  (:COMPILE-LOAD-INIT util  (defs)	(:FASLOAD defs))
  (:COMPILE-LOAD      comp		(:FASLOAD defs util))
  (:COMPILE-LOAD-INIT menus (defs)	(:FASLOAD defs util))
  (:COMPILE-LOAD-INIT comm  (defs)	(:FASLOAD defs util))
  (:COMPILE-LOAD-INIT table (defs)	(:FASLOAD defs util comm))
  (:COMPILE-LOAD-INIT dcom  (defs)	(:FASLOAD defs util comm table))
  (:COMPILE-LOAD-INIT modes (defs dcom) (:FASLOAD defs util dcom) (:FASLOAD defs util dcom))
  (:COMPILE-LOAD-INIT temp  (defs)	(:FASLOAD defs util comm table dcom))
  (:COMPILE-LOAD-INIT loop  (defs dcom)	(:FASLOAD defs util dcom comm table dcom)
		                        (:FASLOAD defs util dcom comm table dcom))
  (:COMPILE-LOAD      select		(:FASLOAD defs util loop ))
  (:COMPILE-LOAD-INIT wind  (comp modes comm dcom defs util loop menus select table temp)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select))
  (:COMPILE-LOAD-INIT disp  (comp modes comm dcom defs util loop menus select table temp)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select wind)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select wind))
  (:COMPILE-LOAD-INIT edit  (comp modes comm dcom defs util loop menus select table temp)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select wind)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select wind))
  (:COMPILE-LOAD-INIT ekeys (comp modes comm dcom defs util loop menus select table temp)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select wind)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select wind))
  (:COMPILE-LOAD-INIT ename (comp modes comm dcom defs util loop menus select table temp)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select wind disp edit ekeys)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select wind disp edit ekeys))
  (:COMPILE-LOAD-INIT univ  (comp modes comm dcom defs util loop menus select table temp)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select))
  (:COMPILE-LOAD-INIT num   (comp modes comm dcom defs util loop menus select table temp)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select))
  (:COMPILE-LOAD-INIT break (comp modes comm dcom defs util loop menus select table temp)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select)
			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select))
;  (:compile-load-init debug (comp modes comm dcom defs util loop menus select table temp)
;			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select)
;			    (:FASLOAD defs util comp menus comm table dcom modes temp loop select))
  )


(UNLESS (FIND-PACKAGE "UCL")
	(MAKE-PACKAGE "UCL"))
