;;; -*- Mode:Common-Lisp; Package:UCL; Fonts:(CPTFONT HL10B); Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (c) 1984- 1989 Texas Instruments Incorporated.  All rights reserved.


1;;These are the commands for the UCL Keystroke Editor, which is called from within the
;;Command Editor.*

(Defconstant EDIT-KEYS-DOCUMENTATION "
			   KEY EDITOR

You are  modifying the  keystrokes of a command.  Any key (or sequence of keys) 
displayed  in this  window may be typed to execute  the command.  The  who-line
at the bottom of the screen gives an indication of what the three mouse buttons
will do if they are clicked over one of the keys in  this  display.  A  menu of
options is popped up by a right mouse click."
  "1The general documentation for the UCL Key Editor.*") 


(DEFCOMMAND (command-display-window :edit-keys-add) ()
  '(:NAMES       "Add Key"
    :DESCRIPTION "Adds a key or key sequence to the list of keys for this command."
    :DOCUMENTATION
    "This command adds a new keystroke (or keystroke sequence) to the list of keys
for a command. The key (or sequence) is read from the keyboard, added to the list 
of keys for the command and then the scroll window display is updated to show the
new keys list.  Mouse buttons are treated like keystrokes, so you may be able to
assign a button to a command, depending on the application."
    :KEYS         #\CONTROL-O)
  (DECLARE (SPECIAL selected-command))
  (LET ((keys (COPY-LIST (SEND selected-command :keys)) )
	(new-key-sequence (get-keystrokes)))
    (WHEN new-key-sequence
      (SEND SELF :APPEND-ITEM (LIST (LENGTH keys) (parse-key-for-edit new-key-sequence)))
      (SEND selected-command :set-keys (NCONC keys (LIST new-key-sequence))))))


(DEFCOMMAND (command-display-window :edit-keys-delete) (selected-key-entry)
  '(:NAMES "Delete Key"
    :DESCRIPTION "Delete this key entry from the list of keys for this command."
    :DOCUMENTATION
    "This command deletes a keystroke (or keystroke sequence) from the list of keys
for a command.  The key entry is removed from the list of keys in the command and 
is deleted from the scroll window display."
    :ACTIVE-IN-DISPLAY? NIL
    :PROPERTY-LIST (need-arg T))
  (DECLARE (SPECIAL selected-command))
  (LET ((keys (COPY-LIST (SEND selected-command :keys)))
	(index (CAR selected-key-entry)))
    (SEND selected-command :set-keys (DELETE-IF #'(LAMBDA (IGNORE) T) keys :START index :END (1+ index)))
    (SEND SELF :SET-ITEMS (SEND selected-command :parse-keys-for-edit))))

(MAKE-COMMAND edit-keys-window-help
  `(:NAMES       "Help"
    :DESCRIPTION "Display help text for the Keystroke Editor."
    :KEYS        #\HELP
    :DEFINITION  (:METHOD command-display-window :help)
    :ARGUMENTS   (edit-keys-command-table)))

(MAKE-COMMAND list-key-edit-commands
  `(:NAMES       "List Commands"
    :DESCRIPTION "List the commands for the Keystroke Editor."
    :KEYS        #\CONTROL-HELP
    :DEFINITION  (:METHOD command-display-window :list-commands)
    :ARGUMENTS   ('edit-keys-command-table)))


(MAKE-COMMAND quit-keys-edit
  `(:NAMES       "Exit"
    :DESCRIPTION "Return from the Key Editor."
    :KEYS        #\END
    :DEFINITION  (:METHOD command-display-window :quit)))


(DEFCOMMAND (command-display-window :edit-keys-replace) (selected-key-entry)
  '(:NAMES "Replace Key"
    :DESCRIPTION "Replace this key entry with a new key."
    :DOCUMENTATION
    "This command replaces a selected keystroke (or keystroke sequence) for a 
command with a new keystroke (or keystroke sequence).  The new key entry is read 
from the keyboard, the old key entry is replaced, and then the scroll window 
display is updated to show the new entry."
    :ACTIVE-IN-DISPLAY? NIL
    :PROPERTY-LIST (need-arg T))
  (DECLARE (SPECIAL selected-command))
  (LET ((keys (COPY-LIST (SEND selected-command :keys)))
	(index (CAR selected-key-entry))
	(new-key-sequence (get-keystrokes)))
    1;;Change keystroke sequence, update the corresponding display item, and reprint.*
    (WHEN new-key-sequence
      (SEND selected-command :set-keys (FILL keys new-key-sequence :START index :END (1+ index)))
      (SETF (CADR selected-key-entry) (parse-key-for-edit1 *new-key-sequence))
      (SEND SELF :reprint-item selected-key-entry))))


(DEFCOMMAND (command-display-window :edit-keys-menu) (&OPTIONAL selected-key-entry)
  '(:DESCRIPTION "Select an key editing operation from a menu."
    :KEYS        #\MOUSE-3-1)
  (DECLARE (SPECIAL edit-keys-command-menu edit-keys-menu selected-command))
  (LET* ((index (CAR selected-key-entry))
	 (menu-list (IF index
			`((,(FORMAT NIL "( ~{~:@C ~})" (NTH index (SEND selected-command :keys)))
			   :NO-SELECT T :FONT *TITLE-FONT*)
			  ("" :NO-SELECT T)
			  . ,edit-keys-command-menu)
		      edit-keys-menu))
	 (key-editor-command (smart-menu-choose menu-list)))
    (WHEN key-editor-command  1;;Nil if user moved mouse off of menu* 
      (IF (SEND key-editor-command :GET 'need-arg)
	  1;;For those commands who need an index argument.*
	  (SEND key-editor-command :execute self :arguments-override `(',selected-key-entry))
	1;;For those commands who don't need an arg or who know their arg.*
	(send key-editor-command :execute self))))) 

(build-command-table 'edit-keys-command-table 'command-display-window
  '(:edit-keys-add
    :edit-keys-delete
    edit-keys-window-help
    list-key-edit-commands
    :edit-keys-menu
    quit-keys-edit
    :edit-keys-replace
    top-of-display
    bottom-of-display
    scroll-up
    scroll-down)
  :INIT-OPTIONS `(:NAME "Key Editing Commands"
		  :DOCUMENTATION ,edit-keys-documentation))

(build-menu 'edit-keys-command-menu 'command-display-window
  :item-list-order
  '(("Operations on existing key:" :font *label-font*)
    :edit-keys-delete
    :edit-keys-replace
    ""
    ("General:" :font *label-font*)
    :edit-keys-add    
    edit-keys-window-help
    list-key-edit-commands
    quit-keys-edit
    top-of-display
    bottom-of-display
    scroll-up
    scroll-down))


(build-menu 'edit-keys-menu 'command-display-window
  :item-list-order
  '(:edit-keys-add    
    edit-keys-window-help
    list-key-edit-commands
    quit-keys-edit
    top-of-display
    bottom-of-display
    scroll-up
    scroll-down))
