;;;-*- Mode:Common-Lisp; Package:UCL; Fonts:(CPTFONT HL10B); Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (c) 1984- 1989 Texas Instruments Incorporated.  All Rights Reserved.


(DEFFLAVOR selective-features-mixin () ()
  (:REQUIRED-FLAVORS basic-command-loop)
  (:INIT-KEYWORDS :REMOVE-FEATURES)
  (:DOCUMENTATION :MIXIN "
1This mixin is available for application flavors which incorporate either UCL:BASIC-COMMAND-LOOP
or UCL:COMMAND-LOOP-MIXIN; that is, applications which use the UCL.  This mixin gives programmers
the ability to turn off UCL features which aren't appropriate for their application.  For instance,
the UCL's command name type-in feature allows the user to type and evaluate lisp forms.  If
your application's users don't know Lisp or have no use for evaluating lisp, you may use this
mixin to get rid of lisp type-in.  Another example concerns the set of /"Universal Commands/"
which the UCL provides.  One universal command is the command editor, which allows users
to change command keystrokes, command names, etc.  If your users are too naive for this feature,
remove it using this mixin.

Not all of the UCL features can be turned off with this mixin; some may be controlled through
the instance variables of UCL:BASIC-COMMAND-LOOP (and UCL:COMMAND-LOOP-MIXIN).  Read
documentation on UCL:BASIC-COMMAND-LOOP for further information.

This mixin may be placed anywhere in your list of dependent flavors.  To turn off features,
you must supply the init keyword :REMOVE-FEATURES in your flavor's :DEFAULT-INIT-PLIST.
Example:
    (DEFFLAVOR foo () (ucl:basic-command-loop ucl:selective-features-mixin)
      (:DEFAULT-INIT-PLIST
         :REMOVE-FEATURES '(:LISP-TYPEIN ucl:EDIT-COMMAND-TABLES))

Note that while you CAN put :REMOVE-FEATURES in a call to MAKE-INSTANCE, such as
    (MAKE-INSTANCE 'foo :REMOVE-FEATURES '(:lisp-typein)),
the most recent instantiation of FOO dictates what features are removed from all other instantiations
of FOO.  So a given flavor always has the same set of features removed.  Because of this,
:REMOVE-FEATURES should be placed in the :DEFAULT-INIT-PLIST.

The :REMOVE-FEATURES init keyword takes a list of features (represented by symbols) to remove.
Note: part of the goal of the UCL is to provide uniform command interfaces accross all Explorer
software.  None of these features HAS to be removed, and before you remove any of them, keep
in mind that users may expect them to be there simply because other Explorer packages have them.
On the other hand, if you are developing a turn-key system for someone who will never use any
of the other Explorer packages, it is desirable to remove any of the features which don't help your
application, to aviod confusing the user.

These are the available features to remove:*
  :LISP-TYPEIN --
1      Gets rid of lisp expression processing.  The type-in handler will only process (and complete on)
      typed command names.*
  :INPUT-EDITOR-COMMANDS --
1      Gets rid of the Input Editor command table.  This means that at top level, the user won't
      be able to type CONTROL-C to yank the previously typed expression, CLEAR-SCREEN to clear
      the screen, etc.  Display-Commands won't show the Input Editor commands.  However,
      once he starts typing an expression, the Input Editor commands will be active.  There are
      probably only two reasons you'd turn off the feature; either you have designed your own
      Input Editor, or you have designed a typein-handler which is activated only by one
      keystroke (such as meta-x in Zmacs).*
  :ALL-UNIVERSAL-COMMANDS --
1      Removes the entire UNIVERSAL-COMMANDS command table from your application.  This is fairly
      drastic; it is the same as removing the rest of the features.  Since the rest of the features
      are universal commands, you may view detailed documentation or try them out by getting to
      a Lisp Listener (SYSTEM L) and hitting HYPER-CONTROL-HELP.  This brings up a display which
      includes the universal commands.*

  ucl:DISPLAY-COMMAND-TABLES --
1      Removes the command which displays all active command tables in a scroll window.
      You probably should never remove this feature.*
  ucl:DISPLAY-PREVIOUS-COMMANDS --
1      Removes the command which displays previously executed commands (the command history).*
  ucl:COMMAND-NAME-SEARCH --
1      Removes the command which accepts a search string from the user and displays all commands
      whose names match the search string.  It is probably advisable to remove this if your application
      doesn't have many commands.*
  ucl:KEYSTROKE-SEARCH --
1      Removes the command which accepts a keystroke from the user and finds which command (or
      commands) has the keystroke.  This feature is good for applications where it is dangerous
      to try out unknown keystrokes.*
  ucl:NUMERIC-ARGUMENT --
1      Removes the command which accepts a numeric argument from the user and repeats the next
      keystroke command numeric-argument number of times.  Note that you may set up commands
      to do something else with the numeric argument besides repetition; control-y in the Input Editor
      yanks the Nth kill, instead of yanking the previous kill N times.  See documentation on DEFCOMMAND
      under :ARGUMENTS*
  ucl:REDO-COMMAND --
1      Removes the command which executes the previous command with the same user-supplied 
      arguments (if any).  This command is strongly application-dependent.  In the graphics editor
      it is useless since commands never fetch arguments from the user using the :ARGUMENTS feature
      of DEFCOMMAND.  In the RTMS interface, MOST of the commands fetch arguments from the user,
      so it may be of some use.  Read documentation on DEFCOMAMND under :ARGUMENTS for information
      on argument fetching.*

  ucl:EDIT-COMMAND-TABLES --
1      Removes the command which allows the user to modify command data fields such as the
      names and keystrokes.  It is nice to let the user modify his command environment (especially
      by adding or modifying the command keystrokes), but if your users are very naive or your
      application only has a few commands (such as in Peek), you might remove the feature.
      To try out the Command Editor, go to a Lisp Listener (SYSTEM L) and  hit HYPER-CONTROL-STATUS.*
  ucl:BUILD-COMMAND-MACRO --
1      Removes the command which allows the user to build command macros which execute a sequence
      of other commands.  The usefulness of command macros is application-dependent.  Note that
      Build-Command-Macro is provided (in addition to Build-Keystroke-Macro) since not all commands
      are bound to keystrokes.*
  ucl:BUILD-KEYSTROKE-MACRO --
1      Removes the command which allows the user to build keystroke macros which force a series
      of keystrokes into the input buffer.  The usefulness of command macros is application-dependent.
      Note that the keystroke macro and command macro overlap in functionality, but that they each
      have their own use.  An example of a good keystroke macro for the Lisp Listener is a keystroke
      (such as HYPER-CONTROL-D) which forces the expression (DESCRIBE *) into the buffer.  Then
      after typing a form, you can quickly use the macro to DESCRIBE the value returned by the form.*
  ucl:LOAD-COMMANDS --
1      Removes the command which allows the user to load commands (and macro commands) which he
      tailored (using the Command Editor) in a previous session.  You should remove this feature IFF
      you remove the Command Editor (ucl:EDIT-COMMAND-TABLES).*
  ucl:SAVE-COMMANDS --
1      Removes the command which allows the user to save any modified commands (and macro 
      commands).  You should remove this IFF you remove ucl:LOAD-COMMANDS.*
  ucl:CONFIGURE-TYPEIN-MODES --
1      Removes the command which allows the user to rearrange the command and Lisp type-in modes.  To 
      see what this is about, hit HYPER-CONTROL-( in a Lisp Listener.  You should get rid of this feature
      if you set your :TYPEIN-HANDLER to NIL or you have removed :LISP-TYPEIN.  However, it
      won't hurt anything to leave it in, since it will just beep if your application doesn't process
      typed expressions.*
  ucl:TOP-LEVEL-CONFIGURER --
1      Removes the command which allows the user to configure the behavior of the command and
      lisp type-in feature (the Lisp Top Level).  For instance, the configurer allows him to turn off
      the package prompt.  Hit HYPER-CONTROL-T in a Lisp Listener to see the feature.  You should
      get rid of this IFF you get rid of UCL:CONFIGURE-TYPEIN-MODES.  However, it won't
      hurt anything to leave it in, since it will just beep if your application doesn't process
      typed expressions.*
  ucl:HELP-MENU --
1      Removes the command which executes when the user types the HELP key.  This command 
      gives the user access to all the UCL help commands, such as Display-Commands and 
      Command-Name-Search, plus the environment customization commands such as Edit-Commands.
      If your flavor sets the UCL:BASIC-HELP and UCL:TUTORIAL instance variables to non-nil, 
      entries for basic help and tutorial will appear in the menu.  (See documentation on flavor
      UCL:BASIC-COMMAND-LOOP.)  It is advisable to leave this feature in, since part of the idea 
      behind the UCL is to provide a uniform help interface.  If you remove the feature, be sure
      that your Help command lets the user get to Display-Commands, Edit-commands, etc. easily.
      Your help feature might do this by calling function UCL:HELP at some point.*
  ucl:SYSTEM-MENU --
1      Removes the command which brings up the system menu on single right mouse.  If you are
      doing something else with single right mouse, remove it.*"))


(DEFWRAPPER (selective-features-mixin :INIT) (init-plist . body)
  "1This wrapper removes all the features which aren't wanted by the application
flavor.  We do this after any :INIT demons have run, since UCL:BASIC-COMMAND-LOOP adds
features (some of which we are removing) in an :AFTER :INIT.*"
  `(LET ((features  (GET (CAR init-plist) :REMOVE-FEATURES)))
     (UNLESS features
       (FERROR NIL "Can not instantiate ~S since UCL:SELECTIVE-FEATURES-MIXIN requires the init keyword :REMOVE-FEATURES to be supplied a non-nil value." self))
     (PROGN . ,body)
     1;;Lisp type-in is removed by explicitly supplying the list of TYPEIN-MODES to be processed to include*
     1;;just UCL:COMMAND-NAMES. *
     (WHEN (MEMBER :LISP-TYPEIN features)
       (SETQ features (REMOVE :LISP-TYPEIN features :TEST #'EQUAL) 
	     typein-modes '(COMMAND-NAMES)))
     1;;Remove the Input Editor command table.*
     (WHEN (MEMBER :INPUT-EDITOR-COMMANDS features)
       (SETQ features (REMOVE :INPUT-EDITOR-COMMANDS features :TEST #'EQUAL)
	     system-command-tables (REMOVE 'tv:rh-command-table system-command-tables :TEST #'EQUAL)))
     1;;This keyword gets rid of ALL the universal commands.  We just remove the entire command.*
     1;;table. We also remove the macro command table, which isn't needed since the macro builders *
     1;;are universal commands.  We check to see that the programmer doesn't try to remove options*
     1;;which :ALL-UNIVERSAL-COMMANDS removes, for consistency.*
     (WHEN (MEMBER :ALL-UNIVERSAL-COMMANDS features)
       (SETQ features (REMOVE :ALL-UNIVERSAL-COMMANDS features :TEST #'EQUAL)
	     system-command-tables (REMOVE 'universal-commands system-command-tables :TEST #'EQUAL))
       (WHEN features
	 (FERROR NIL "
:ALL-UNIVERSAL-COMMANDS removes all universal commands from your application. 
The following features (or typos) were encountered in the :REMOVE-FEATURES list 
and should be taken out since :ALL-UNIVERSAL-COMMANDS removes them: ~%~S" 
		 features))
       (remove-macro-command-table))
     1;;If there are features left to remove, they are selected universal comamnds.  Remove them by*
     1;;replacing universal-commands with a command table not containing the selected features.*
     (WHEN features
       (replace-universal-commands features))
     1;;If both macro-building commands are removed, remove the macro command table.*
     (WHEN (AND (MEMBER 'build-command-macro features)
		(MEMBER 'build-keystroke-macro features)) 
       (remove-macro-command-table)))) 


(DEFUN remove-macro-command-table ()
  "1Removes the macro command table which the UCL gives the application.*"
  (DECLARE (:SELF-FLAVOR basic-command-loop))
  (LET (macro-table)
    (LOOPING-THROUGH-COMMAND-TABLES NIL
      (WHEN (SEND command-table :GET 'macro-table?)
	(SETQ macro-table (SEND command-table :symbol-name))
	(RETURN)))
    (SETQ system-command-tables
	  (REMOVE MACRO-TABLE SYSTEM-COMMAND-TABLES :TEST #'EQUAL))))

(DEFUN replace-universal-commands (removed-features)
  "1Takes ucl:universal-commands out of the application flavor and substitutes a similar command table
having all the universal commands except those specified by REMOVED-FEATURES.*"
  (DECLARE (:SELF-FLAVOR basic-command-loop)
	   (SPECIAL universal-commands))
  (LET (desired-features
	command
	(new-command-table-symbol
	  (make-up-name-for-selected-universal-commands self)))
    1;;Error-check for typos.*
    (DOLIST (feature removed-features)
      (UNLESS (AND (SETQ command (get-command feature))
		   (SEND universal-commands :LOOKUP-COMMAND command))
	(FERROR NIL "~S was encountered in the :REMOVE-FEATURES list but is not a recognized feature.
Read documentation on flavor UCL:SELECTIVE-FEATURES-MIXIN for a list of features you can turn off." feature)))
    1;;Subtract the unwanted features from the entire set of universal commands.*
    (SETQ desired-features
	  (LOOP FOR command BEING THE ARRAY-ELEMENTS OF
		            (SEND universal-commands :COMMANDS)
	        FOR definition = (SEND command :DEFNAME)
		UNLESS (MEMBER definition removed-features)
		COLLECT definition))
    1;;Build a new table holding the remaining features.*
    (BUILD-COMMAND-TABLE new-command-table-symbol NIL
			 desired-features
			 :INIT-OPTIONS '(:NAME "Universal Commands"
					 :KEY-LOOKUP-FUN univ-lookup-key
					 :DOCUMENTATION "This command table contains general help commands which may appear in any
application using the Universal Command Loop."))
    1;;Substitutes the subset command table for the universal commands.*
    (SETQ system-command-tables (REMOVE 'universal-commands system-command-tables :TEST #'EQUAL))
    (PUSH new-command-table-symbol system-command-tables))) 

(DEFUN make-up-name-for-selected-universal-commands (application-flavor)
  (INTERN (STRING-UPCASE (FORMAT NIL "~A SELECTED-UNIVERSAL-COMMANDS"
				 (TYPE-OF application-flavor)))
	  'UCL)) 
