;;; -*- Mode:Common-Lisp; Package:X11; Fonts:(MEDFNB HL12B HL12BI); Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

#|

This file contains only the keyboard-and-mouse event dispatcher code.
|#

;;;
;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 03/21/89	WJB	Patch 1.53; Add wholine support
;;; 12/21/88	LGO	1Set tv:kbd-last-activity-time in event-dispatcher to keep the screen-saver away*
;;; 12/06/88	DLS	Changed dispatcher process to be restartable so it can be reset
;;;			and so it will stay around when saved in a band.
;;;  5/13/88    TWE	Removed dependencies on I/O buffers.
;;;  5/06/88    TWE	Modified EVENT-DISPATCHER to set the dispatcher process variable
;;;			to NIL if the user kills it off.
;;;  5/04/88    TWE	Changed EVENT-DISPATCHER to clear out the event queues before
;;;			starting.
;;;  3/29/88    TWE	Changed DISPATCHER-PROCESS to a DEFVAR so that it wouldn't be set
;;;			to NIL when the system is remade.
;;;  3/09/88    TWE	Fixed up the initialization of the dispatcher process.
;;;  2/29/88    TWE	Initial creation.  Moved from MOUSE-HANDLER file.


(DEFVAR DISPATCHER-PROCESS NIL)

(DEFUN EVENT-DISPATCHER ()
  "2Check mouse and keyboard event queues for input to process.*"
  (CLEAR-EVENT-QUEUE KEYBOARD-EVENT-QUEUE)
  (CLEAR-EVENT-QUEUE MOUSE-EVENT-QUEUE)
  (SETQ DISPATCHER-PROCESS CURRENT-PROCESS)
  (X-WHOLINE-PROCESS DISPATCHER-PROCESS)
  (UNWIND-PROTECT
      (LOOP
        (PROCESS-WAIT "Key or Mouse" #'(LAMBDA (KEY MOUSE)
                                         (OR (NOT (EVENT-QUEUE-EMPTY-P KEY))
                                             (NOT (EVENT-QUEUE-EMPTY-P MOUSE))))
                      KEYBOARD-EVENT-QUEUE 
                      MOUSE-EVENT-QUEUE)
        (event-trace "~%IN EVENT-DISPATCHER: (EVENT-QUEUE-EMPTY-P MOUSE-EVENT-QUEUE)= ~s ~
			input-pointer= ~d output-pointer= ~d"
		      (EVENT-QUEUE-EMPTY-P MOUSE-EVENT-QUEUE)
		      (W:IO-BUFFER-INPUT-POINTER mouse-event-queue)
		      (W:IO-BUFFER-OUTPUT-POINTER mouse-event-queue))
        (event-trace "~%IN EVENT-DISPATCHER: (EVENT-QUEUE-EMPTY-P KEYBOARD-EVENT-QUEUE)= ~s ~
			input-pointer= ~d output-pointer= ~d"
		      (EVENT-QUEUE-EMPTY-P KEYBOARD-EVENT-QUEUE)
		      (W:IO-BUFFER-INPUT-POINTER keyboard-event-queue)
		      (W:IO-BUFFER-OUTPUT-POINTER keyboard-event-queue))
	(setq tv:kbd-last-activity-time (time)) ;1; keep the screen-saver out of our hair*
	(PROCESS-INPUT-EVENTS))
    ;; If the dispatcher is killed off we set this to nil so that when we initialize
    ;; again the dispatcher will be created again.
    (SETQ DISPATCHER-PROCESS NIL)))

(DEFUN INITIALIZE-DISPATCHER-PROCESS ()
  (if DISPATCHER-PROCESS
      (send dispatcher-process :reset)
      (PROCESS-RUN-RESTARTABLE-FUNCTION '(:NAME "X Dispatcher" :PRIORITY 5.)
					#'EVENT-DISPATCHER)))