;;; Zwei commands, see ZWEI;COMA for comments -*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1980, Massachusetts Institute of Technology

;;; Various Quantities.
;;; Change History :
;;; 02/24/89 	MAY	Modified  INITIALIZE-TAB-STOP-BUFFER & COM-TAB-TO-TAB-STOP - off by first char.
;;; 01/23/89	MAY	Modified DEFAULT-TAB-WIDTH to cause redisplay more efficiently.
;;; 01/18/89 	MAY	Changed INITIALIZE-TAB-STOP-BUFFER, COM-TAB-TO-TAB-STOP
;;;                     & added DEFAULT-TAB-WIDTH to make tabs in TEXT mode other
;;;                     than hardcoded 8 without editing tab stop buffer. Also
;;;                     allows default tab-width for ALL buffers to be changed
;;;                     if not specified in mode line ( or by m-x Set Tab Width )
;; Doc changed by patch 94.207.   ddd/gsl 3/5/84.

(DEFCOM COM-VARIOUS-QUANTITIES
   "Given characters with attribute bits or non-letters, inserts them.
  For example: C-Q C-G inserts a \"\";
               C-Q <CLEAR SCREEN> inserts a \"\";
               C-Q <RESUME> inserts a \"\".
  A question mark <?> displays this help message.
Given octal digits, inserts character with specified character code.
  For example: C-Q 101 inserts an \"A\";
               C-Q 141 inserts an \"a\";
               C-Q 201 inserts a \"\".  
  If you type fewer than three digits, type a space after the last digit.
  With a numeric argument, inserts the character arg times.

Otherwise, this command manipulates various quantities:
  The first character following Control-Q is the operation:
             @  mark region          R  rubout
             B  backward             S  save
             C  copy                 T  transpose 
             D  delete               U  uppercase
             F  forward              Z  reverse
             L  lowercase
  The second character following is the quantity type:
             A  atom                 S  sentence
             C  character            W  word
             D  defun                -  S-expression
             H  buffer               (  list 
             L  line                 )  list     
             P  paragraph            <CLEAR-SCREEN>  page
  Numeric arguments are obeyed." ()
   (LET (CH MODE-NAME MODE QUANTITY)
	 ;; Read next char turning off normal intercepted meanings of Abort and Break.
     (LET ((W:KBD-INTERCEPTED-CHARACTERS
	    (REMOVE-IF #'(LAMBDA (ELT)
			   (ZEROP (CHAR-BITS (CAR ELT))))
		       W:KBD-INTERCEPTED-CHARACTERS)))
       (SETQ CH (INPUT-WITH-PROMPTS *STANDARD-INPUT* :TYI)))
     (COND ((CHAR-BIT CH :CONTROL)
	    (DOTIMES (I *NUMERIC-ARG*)
	      (INSERT-MOVING (POINT) (INT-CHAR (LOGAND 37 (CHAR-CODE CH)))))
	    DIS-TEXT)
	   ((OR (CHAR< (SETQ CH (CHAR-UPCASE CH)) #\?) (CHAR> CH #\Z))
	    (COND ((AND (CHAR>= CH #\0) (CHAR<= CH #\7))
		   (DISCARD-LAST-PROMPT)
		   (PRINT-PROMPTS)
		   (WRITE-CHAR CH *QUERY-IO*)
		   (SETQ CH (CODE-CHAR (- (CHAR-CODE CH) (CHAR-CODE #\0))))
		   (DO ((I 2 (1- I))
			(CH1))
		       ((<= I 0))
		     (SETQ CH1 (READ-CHAR))
		     (COND ((AND (CHAR>= CH1 #\0) (CHAR<= CH1 #\7))
			    (WRITE-CHAR CH1 *QUERY-IO*)
			    (SETQ CH (CODE-CHAR (+ (* (CHAR-CODE CH) 10)
						   (- (CHAR-CODE CH1) (CHAR-CODE #\0))))))
			   (T
			    (OR (CHAR= CH1 #\SPACE)
				(UNREAD-CHAR CH1))
			    (RETURN NIL))))))
	    (LET ((*LAST-COMMAND-CHAR* CH))
	      (COM-SELF-INSERT)))
	   (T
	    (PROG NIL
	       GET-A-MODE
		  (CASE CH
			(#\? (PRINT-DOC :FULL 'COM-VARIOUS-QUANTITIES)
			     (FORMAT *QUERY-IO* "~&Type the OPERATION (F, B, D, R, T, @, U, L, S, C or Z)")
			     (DISCARD-LAST-PROMPT)
			     (PRINT-PROMPTS)
			     (SETQ CH (CHAR-UPCASE (INPUT-WITH-PROMPTS *STANDARD-INPUT* :TYI)))
			     (GO GET-A-MODE))
			(#\F (SETQ MODE-NAME "Forward" MODE 'COM-QUANTITY-FORWARD))
			(#\B (SETQ MODE-NAME "Backward" MODE 'COM-QUANTITY-BACKWARD))
			(#\D (SETQ MODE-NAME "Delete" MODE 'COM-QUANTITY-DELETE))
			(#\R (SETQ MODE-NAME "Rubout" MODE 'COM-QUANTITY-RUBOUT))
			(#\T (SETQ MODE-NAME "Twiddle" MODE 'COM-QUANTITY-TWIDDLE))
			(#\@ (SETQ MODE-NAME "Mark" MODE 'COM-QUANTITY-MARK))
			(#\U (SETQ MODE-NAME "Uppercase" MODE 'COM-QUANTITY-UPPERCASE))
			(#\L (SETQ MODE-NAME "Lowercase" MODE 'COM-QUANTITY-LOWERCASE))
			(#\S (SETQ MODE-NAME "Save" MODE 'COM-QUANTITY-SAVE))
			(#\C (SETQ MODE-NAME "Copy" MODE 'COM-QUANTITY-COPY))
			(#\Z (SETQ MODE-NAME "Reverse" MODE 'COM-QUANTITY-REVERSE))
			(OTHERWISE (BARF "Invalid quantity operation")))
	       GET-A-QUANTITY
		  (SETQ CH (CHAR-UPCASE (INPUT-WITH-PROMPTS *STANDARD-INPUT* :TYI)))
		  (CASE CH
			(#\? (PRINT-DOC :FULL 'COM-VARIOUS-QUANTITIES)
			     (FORMAT *QUERY-IO* "~&Type quantity name (C, W, S, P, A, L, -, ( or ), D, ~C, or H)"
				     #\PAGE)
			     (DISCARD-LAST-PROMPT) (PRINT-PROMPTS) (GO GET-A-QUANTITY))
			(#\C (SETQ MODE-NAME "Character" QUANTITY 'FORWARD-CHAR))
			(#\W (SETQ MODE-NAME "Word" QUANTITY 'FORWARD-WORD))
			(#\A (SETQ MODE-NAME "Atom" QUANTITY 'FORWARD-ATOM))
			(#\S (SETQ MODE-NAME "Sentence" QUANTITY 'FORWARD-SENTENCE))
			(#\P (SETQ MODE-NAME "Paragraph" QUANTITY 'FORWARD-PARAGRAPH-COMPLETE))
			(#\L (SETQ MODE-NAME "Line" QUANTITY 'FORWARD-LINE))
			(#\- (SETQ MODE-NAME "S-Expression" QUANTITY 'FORWARD-SEXP))
			((#\( #\)) (SETQ MODE-NAME "List" QUANTITY 'FORWARD-LIST))
			(#\D (SETQ MODE-NAME "Defun" QUANTITY 'FORWARD-DEFUN))
			(#\PAGE (SETQ MODE-NAME "Page" QUANTITY 'FORWARD-PAGE))
			(#\H (SETQ MODE-NAME "Buffer" QUANTITY 'FORWARD-BUFFER))
			(OTHERWISE (BARF "Invalid quantity type"))))
	    (LET ((*QUANTITY-MODE* QUANTITY))
	      (FUNCALL MODE)))))) 


(DEFCOM COM-QUANTITY-FORWARD "Move forward according to the current quantity mode." (KM)
   (CONTEXT-CHECK '*QUANTITY-MODE*)
   (MOVE-BP (POINT) (OR (FUNCALL *QUANTITY-MODE* (POINT) *NUMERIC-ARG*) (BARF))) DIS-BPS) 


(DEFCOM COM-QUANTITY-BACKWARD "Move backward according to the current quantity mode." (KM)
   (CONTEXT-CHECK '*QUANTITY-MODE*)
   (MOVE-BP (POINT) (OR (FUNCALL *QUANTITY-MODE* (POINT) (- *NUMERIC-ARG*)) (BARF))) DIS-BPS) 


(DEFCOM COM-QUANTITY-DELETE "Kill forward according to the current quantity mode." ()
   (CONTEXT-CHECK '*QUANTITY-MODE*) (SETQ *CURRENT-COMMAND-TYPE* 'KILL)
   (KILL-INTERVAL (POINT) (OR (FUNCALL *QUANTITY-MODE* (POINT) *NUMERIC-ARG*) (BARF)) () T)
   DIS-TEXT) 


(DEFCOM COM-QUANTITY-RUBOUT "Kill backward according to the current quantity mode." ()
   (CONTEXT-CHECK '*QUANTITY-MODE*) (SETQ *CURRENT-COMMAND-TYPE* 'KILL)
   (KILL-INTERVAL (OR (FUNCALL *QUANTITY-MODE* (POINT) (- *NUMERIC-ARG*)) (BARF)) (POINT) () ())
   DIS-TEXT) 


(DEFCOM COM-QUANTITY-TWIDDLE "Exchange things according to the current quantity mode." ()
   (CONTEXT-CHECK '*QUANTITY-MODE*) (EXCHANGE-SUBR *QUANTITY-MODE* *NUMERIC-ARG*) DIS-TEXT) 


(DEFCOM COM-QUANTITY-REVERSE "Reverse things according to the current quantity mode." ()
   (CONTEXT-CHECK '*QUANTITY-MODE*)
   (REVERSE-SUBR *QUANTITY-MODE* (1+ *NUMERIC-ARG*)) ;; patch 94.149.  ddd. 3/6/84.
   DIS-TEXT) 


(DEFCOM COM-QUANTITY-MARK "Mark according to the current quantity mode." (SM)
   (CONTEXT-CHECK '*QUANTITY-MODE*)
   (LET (BP1 BP2)
     (OR (SETQ BP1 (FUNCALL *QUANTITY-MODE* (POINT) *NUMERIC-ARG*)) (BARF))
     (OR (SETQ BP2 (FUNCALL *QUANTITY-MODE* BP1 (- *NUMERIC-ARG*))) (BARF))
     (AND (MINUSP *NUMERIC-ARG*) (SETQ BP2 (PROG1 BP1 (SETQ BP1 BP2))))
     (MOVE-BP (POINT) BP1)
     (MOVE-BP (MARK) BP2))
   DIS-BPS) 


(DEFCOM COM-QUANTITY-UPPERCASE "Uppercase according to the current quantity mode." ()
   (CONTEXT-CHECK '*QUANTITY-MODE*)
   (LET ((BP1 (OR (FUNCALL *QUANTITY-MODE* (POINT) *NUMERIC-ARG*) (BARF))))
     (LET ((BP2 (OR (FUNCALL *QUANTITY-MODE* BP1 (- *NUMERIC-ARG*)) (BARF))))
       (WITH-UNDO-SAVE ("Upcase" BP1 BP2 NIL) (UPCASE-INTERVAL BP1 BP2))
       (AND (PLUSP *NUMERIC-ARG*) (MOVE-BP (POINT) BP1))))
   DIS-TEXT) 


(DEFCOM COM-QUANTITY-LOWERCASE "Lowercase according to the current quantity mode." ()
   (CONTEXT-CHECK '*QUANTITY-MODE*)
   (LET ((BP1 (OR (FUNCALL *QUANTITY-MODE* (POINT) *NUMERIC-ARG*) (BARF))))
     (LET ((BP2 (OR (FUNCALL *QUANTITY-MODE* BP1 (- *NUMERIC-ARG*)) (BARF))))
       (WITH-UNDO-SAVE ("Downcase" BP1 BP2 NIL) (DOWNCASE-INTERVAL BP1 BP2))
       (AND (PLUSP *NUMERIC-ARG*) (MOVE-BP (POINT) BP1))))
   DIS-TEXT) 


(DEFCOM COM-QUANTITY-SAVE "Save on kill ring according to the current quantity mode." ()
   (CONTEXT-CHECK '*QUANTITY-MODE*)
   (LET ((BP1 (OR (FUNCALL *QUANTITY-MODE* (POINT) *NUMERIC-ARG*) (BARF))))
     (LET ((BP2 (OR (FUNCALL *QUANTITY-MODE* BP1 (- *NUMERIC-ARG*)) (BARF))))
       (KILL-RING-PUSH (COPY-INTERVAL BP1 BP2))
       (MOVE-BP (POINT) BP1)))
   DIS-TEXT) 


(DEFCOM COM-QUANTITY-COPY "Insert a copy according to the current quantity mode." ()
;; Fairly extensive change to this function by patch 94.149.   ddd.
   (CONTEXT-CHECK '*QUANTITY-MODE*)
   (LET ((BP1 (FUNCALL *QUANTITY-MODE* (POINT) *NUMERIC-ARG*)))
     ; check that it is not the end of the line 
     (COND (BP1
	    (LET ((BP2 (OR (FUNCALL *QUANTITY-MODE* BP1 (- *NUMERIC-ARG*)) (BARF))))
	      (MOVE-BP (POINT) (INSERT-INTERVAL BP2 (COPY-INTERVAL BP1 BP2)))))
     ; changes are made to this clause. 
	   (T
	    (LET ((BP3 (FUNCALL *QUANTITY-MODE* (POINT) *NUMERIC-ARG* T)))
	      (LET ((BP2 (FUNCALL *QUANTITY-MODE*
				  (FUNCALL *QUANTITY-MODE* BP3 (- *NUMERIC-ARG*)) *NUMERIC-ARG*)))
		(MOVE-BP (POINT) (INSERT-INTERVAL BP2 (COPY-INTERVAL BP1 BP2)))
		(COM-INSERT-CRS))))))
   DIS-TEXT) 


(DEFCOM COM-PREVIOUS-PAGE "Move to the previous page" (KM)
   (MOVE-BP (POINT) (FORWARD-PAGE (POINT) (- *NUMERIC-ARG*) T)) DIS-BPS) 


(DEFCOM COM-NEXT-PAGE "Move to the next page" (KM)
   (MOVE-BP (POINT) (FORWARD-PAGE (POINT) *NUMERIC-ARG* T)) DIS-BPS) 


(DEFCOM COM-MARK-WHOLE
   "Put mark at beginning of buffer and point end,
or with a numeric argument, vice versa" (SM)
   (LET ((BP1 (POINT))
	 (BP2 (MARK)))
     (AND *NUMERIC-ARG-P* (PSETQ BP1 BP2 BP2 BP1))
     (MOVE-BP BP1 (INTERVAL-LAST-BP *INTERVAL*))
     (MOVE-BP BP2 (INTERVAL-FIRST-BP *INTERVAL*)))
   DIS-BPS) 


(DEFCOM COM-MARK-DEFUN "Put point and mark around current defun." ()
   (LET ((INT (DEFUN-INTERVAL (POINT) *NUMERIC-ARG* () T T)))
     (OR INT (BARF))
     (SETF (WINDOW-MARK-P *WINDOW*) T)
     (SETQ *MARK-STAYS* T)
     (POINT-PDL-PUSH (POINT) *WINDOW* () ())
     (MOVE-BP (POINT) (INTERVAL-FIRST-BP INT))
     (MOVE-BP (MARK) (INTERVAL-LAST-BP INT)))
   DIS-BPS) 


(DEFCOM COM-REPOSITION-WINDOW
   "Try to get all of current defun in the window.
If function beginning is on the screen,
 scrolls down to bring the end onto the screen.
If function beginning is off screen, scrolls up so beginning appears,
 but will not push point off the bottom.
If function beginning is at the top of the screen,
 tries omitting or including the comments before it.
If function is entirely on screen, positions it at the top
 (or, with numeric arg, at the bottom) of the screen." (KM)
   (LET ((POINT (POINT))
	 (SHEET (WINDOW-SHEET *WINDOW*))
	 (N-PLINES (WINDOW-N-PLINES *WINDOW*))
	 (INT (DEFUN-INTERVAL (POINT) 1 T T))
	 START-BP
	 END-BP
	 RECENTER-BP)
     (COND ((NOT (NULL INT))
	    (SETQ START-BP (INTERVAL-FIRST-BP INT) END-BP (INTERVAL-LAST-BP INT))
	    ;; Don't include the blank line after the defun
	    (AND (ZEROP (BP-INDEX END-BP)) (SETQ END-BP (END-LINE END-BP -1 T)))
	    (COND ((AND (PLINE-OF-POINT T *WINDOW* START-BP)	;If start of defun on the screen
			(NULL (PLINE-OF-POINT T *WINDOW* END-BP))	;and not bottom
			(MULTIPLE-VALUE-BIND (LINE INDEX)
			    (PUT-POINT-AT-PLINE SHEET (BP-LINE END-BP)
						(BP-INDEX END-BP)
						(1- N-PLINES)
						(INTERVAL-FIRST-BP *INTERVAL*)
						(INTERVAL-LAST-BP *INTERVAL*))
			  (SETQ RECENTER-BP (CREATE-BP LINE INDEX))
			  ;; And can fit bottom of the defun on as well
			  ;; then start at the top of the function.
			  (NOT (BP-< START-BP RECENTER-BP)))))
		  ((BP-< START-BP
			 (SETQ RECENTER-BP (MULTIPLE-VALUE-BIND (LINE INDEX)
					       (PUT-POINT-AT-PLINE SHEET (BP-LINE POINT)
								   (BP-INDEX POINT)
								   (1- N-PLINES)
								   (INTERVAL-FIRST-BP *INTERVAL*)
								   (INTERVAL-LAST-BP *INTERVAL*))
					     (CREATE-BP LINE INDEX))))
		   ;; If displaying from the start of the defun would push point off
		   ;; the bottom, complain, and bring in as much as possible anyway.
		   (BEEP))
		  ;; Start of defun thru point fits on the screen.
		  ((AND *NUMERIC-ARG-P* (PLINE-OF-POINT T *WINDOW* END-BP))
		   ;; If numeric arg, and end of function is on screen,
		   ;; try putting end of function at bottom.
		   (MULTIPLE-VALUE-BIND (LINE INDEX) (PUT-POINT-AT-PLINE SHEET (BP-LINE END-BP)
									 (BP-INDEX END-BP) (1- N-PLINES)
									 (INTERVAL-FIRST-BP *INTERVAL*)
									 (INTERVAL-LAST-BP *INTERVAL*))
		     (SETQ RECENTER-BP (CREATE-BP LINE INDEX))))
		  (T
		   ;; If already at the default place, try omitting comments above the defun.
		   (AND (BP-= (WINDOW-START-BP *WINDOW*) START-BP)
			(SETQ START-BP (INTERVAL-FIRST-BP (DEFUN-INTERVAL (POINT) 1 T ()))))
		   (SETQ RECENTER-BP START-BP)))
	    (RECENTER-WINDOW *WINDOW* :START RECENTER-BP))
	   (T (BARF "no defun here")))
     DIS-NONE)) 



;Now that all windows record their input, this is a no-op (starting in system 87).

(DEFUN MAKE-RECORDING-STREAM (STREAM &REST IGNORE) STREAM) 

;;; rb - 10/12/84 - this is "com-what-lossage" renamed

(DEFCOM COM-LIST-KEYSTROKE-HISTORY
   "List the last sixty keystrokes typed inside the editor.
This is most useful when you've inadvertantly killed something by typing the wrong command." (KM)
   (COND ((NOT (MEMBER :PLAYBACK (SEND *STANDARD-INPUT* :WHICH-OPERATIONS) :TEST #'EQ))
	  (BARF "Your input was not being recorded; sorry."))
	 (T
	  (LET ((A (SEND *STANDARD-INPUT* :PLAYBACK))
		(WIDTH (OR (SEND *STANDARD-OUTPUT* :SEND-IF-HANDLES :SIZE-IN-CHARACTERS) 137)))
	    (LET ((P (ARRAY-LEADER A 1))
		  (L (ARRAY-LEADER A 0)))
	      (DO ((I (REM (1+ P) L) (REM (1+ I) L))
		   (J 0 (1+ J)))
		  ((>= J L))
		(LET ((CH (AREF A I)))
		  (AND CH
		       (NOT (CONSP CH))
		       (COND ((< (+ (SEND *STANDARD-OUTPUT* :READ-CURSORPOS :CHARACTER)
				    (LENGTH (FORMAT NIL "~:C" CH)))
				WIDTH)
			      (FORMAT T "~:C " CH))
			     (T
			       (TERPRI)
			       (FORMAT T "~:C " CH)
			       )))))))
	  (FRESH-LINE)))
   DIS-NONE) 


(DEFCOM COM-EXIT-CONTROL-R "Exits from a recursive edit" () (THROW 'EXIT-CONTROL-R ())) 


(DEFCOM COM-QUIT "Return from the top-level edit" () (THROW 'EXIT-TOP-LEVEL ())) 


(DEFCOM COM-ABORT-AT-TOP-LEVEL
   "Abort a command that is unfinished or reading arguments.
Aborts minibuffers and recursive edits, and things like C-X M and DIRED.

Actually, this particular definition is the one used at top level only,
and it does nothing except abort any keyboard macro being defined.
When you are actually typing the arguments to a command,
the Abort key has a different definition." ()
   (IF (WINDOW-MARK-P *WINDOW*)
       (SETQ *MARK-STAYS* ())
       (BARF
	 (IF (MEMBER (SEND *STANDARD-INPUT* :SEND-IF-HANDLES :MACRO-LEVEL) '(0 NIL) :TEST #'EQ)
	     "Already at top level."
	     "Aborting definition of keyboard macro.")))
   DIS-NONE) 


(DEFCOM COM-BREAK "Enter a lisp break loop" ()
  (LET ((*INSIDE-BREAK* T))
    (BREAK "ZMACS"))
  (SEND *STANDARD-OUTPUT* :MAKE-COMPLETE)
  DIS-NONE) 

; TAB TO TAB STOP stuff.
;This should really be something hairy with the mouse...
; A "tab stop buffer" has two lines: one to indicate characters to fill
;with, and the second to indicate where the tab stops are.  In the second line
;colons and periods are the only significant characters.  Everything else
;is ignored.  If there is a :, fill with spaces, else with contents of 
;the first line.  I dont think this can work reasonably with variable
;width fonts very well, so the initial version, at least, will assume
;that you are using only one fixed width font.

;; may 01/18/89 Reworked to allow optional arg. 
;; Called by DEFAULT-TAB-WIDTH and by INITIALIZE-ZWEI-GLOBALS.
(DEFUN INITIALIZE-TAB-STOP-BUFFER (&optional tab-width)
  (SETQ tab-width (MAX 1 (OR tab-width (GET :TAB-WIDTH 'DEFAULT-ATTRIBUTE-VALUE) 8.)))
  (LET ((*BATCH-UNDO-SAVE* T)
	(min-width 80.)
	(spaces (make-array tab-width :element-type 'string-char :initial-element #\space)))
    (SETQ *TAB-STOP-BUFFER* (CREATE-INTERVAL () () T))
    (INSERT (INTERVAL-FIRST-BP *TAB-STOP-BUFFER*)
	    (format nil "~a~%" (make-string min-width :initial-element #\space))) ;; first line of spaces/return
    ;; print colons every TAB-WIDTH spaces AFTER initial space
    (insert-chars (interval-LAST-bp *tab-stop-buffer*) #\space 1) ;; may 02/24/89 
    (setf (char spaces (1- tab-width)) #\:)
    (dotimes (i (floor min-width tab-width)) 
      (insert (interval-LAST-bp *tab-stop-buffer*) (format nil "~a" spaces))))) 

;; may 01/18/89 new function to be set by profile instead of
;; *tab-width* which is unsupported. Can be called by user.
(DEFUN default-tab-width (&optional tab-width)
  "Set the default tab-width for ZMACS buffer editing and for 
Tab To Tab Stop tabs. If no arg is given the current value is returned."
  ;; Return current-value if NIL argument
  (OR tab-width (RETURN-FROM default-tab-width (values (GET :TAB-WIDTH 'DEFAULT-ATTRIBUTE-VALUE))))
  
  ;; Save default tab width used by any buffer that does not have an explicit
  ;; tab-width in mode line or on SI:PROPERTY-LIST. Used by (zwei:FILE-BUFFER :AFTER :SET-ATTRIBUTE)
  (putprop :tab-width (MAX 1 tab-width) 'default-attribute-value)
  (initialize-tab-stop-buffer)  ; make tab-stop-buffer correlate
  
  ;; In EACH window if the current buffer is not :special
  ;; and is using the DEFAULT tab-width,
  ;; then those lines with TAB chars must be munged.
  (DOLIST (window *all-zmacs-windows*)
    (LET ((buffer (window-interval window)))
      (when (and buffer
		 (null (get buffer :special-type))	; don't do special buffers.
		 (null (send buffer :get-attribute :tab-width nil)))	; default tab width ( not stored )
	(redefine-window-tab-nchars window (send buffer :get-attribute :tab-width))
	(DOLIST (WINDOW *all-zmacs-windows*)
	  (SEND WINDOW :FORCE-KBD-INPUT '(REDISPLAY)))
	))))
;;;
;;; A creature feature which we may not want to keep around.
;;; This was a request from the guys in Austin, Hogan and/or Kirkpatrick. gsl.

(DEFUN LOAD-TAB-STOP-BUFFER (TEXT-LINE STOPS-LINE)
  "Sets up the *TAB-STOP-BUFFER*. Args should be the strings the buffer should contain.
The first is text to be inserted when tabbing to a . in the second line.
The second is . to mean copy text, or : to mean a normal tab stop."
  (LET ((*BATCH-UNDO-SAVE* T))
    (IF (BOUNDP '*TAB-STOP-BUFFER*)
	(DELETE-INTERVAL (INTERVAL-FIRST-BP *TAB-STOP-BUFFER*)
			 (INTERVAL-LAST-BP *TAB-STOP-BUFFER*) T)
	(SETQ *TAB-STOP-BUFFER* (CREATE-INTERVAL () () T)))
    (INSERT (INTERVAL-FIRST-BP *TAB-STOP-BUFFER*) (STRING-APPEND TEXT-LINE #\NEWLINE STOPS-LINE)))) 

;;;
;;; This fixes a problem when the text in the first line of the edit
;;; buffer ends befor the last tab stop.  The fix was to simply
;;; supply enough blanks to eh text line after the user exits the
;;; tab buffer.  PMH

(DEFCOM COM-EDIT-TAB-STOPS
   "Edit the tab-stop buffer.
In the second line of the tab-stop buffer,
a colon indicates a tab stop to move out to with spaces.
A period indicates a tab stop to space out to
by copying the text from the corresponding columns
in the first line of the tab stop buffer." ()
  (RECURSIVE-EDIT *TAB-STOP-BUFFER* "Edit tab stops (press  to return)")
  (LET ((DESIRED-LENGTH (LENGTH (LINE-NEXT (BP-LINE (INTERVAL-FIRST-BP *TAB-STOP-BUFFER*)))))
	(CURRENT-LENGTH (LENGTH (BP-LINE (INTERVAL-FIRST-BP *TAB-STOP-BUFFER*)))))
    (IF (< CURRENT-LENGTH DESIRED-LENGTH)
	(INSERT (END-LINE (INTERVAL-FIRST-BP *TAB-STOP-BUFFER*))
		(MAKE-ARRAY (- DESIRED-LENGTH CURRENT-LENGTH) :ELEMENT-TYPE 'STRING-CHAR
			    :INITIAL-VALUE #\SPACE))))
   DIS-ALL)  


;; may 01/18/89 changes to allow something besides hardcoded 8. for tab width.
(DEFCOM COM-TAB-TO-TAB-STOP
	"Tab to fixed column as specified by the tab-stop buffer.
Use the Edit Tab Stops command to edit the tab-stop settings." ()
  (LET* ((POINT (POINT))
	 (GOAL (BP-VIRTUAL-INDENTATION POINT))
	 (L2 (or (LINE-NEXT (BP-LINE (INTERVAL-FIRST-BP *TAB-STOP-BUFFER*)))
		 (progn (initialize-tab-stop-buffer)	;; may 01/18/89 happens if someone deletes first line
			(line-next (bp-line (interval-first-bp *tab-stop-buffer*)))))) ;; may 01/18/89 
	 (CHAR-POS))
    (MULTIPLE-VALUE-SETQ (NIL CHAR-POS)
      (W:SHEET-STRING-LENGTH (WINDOW-SHEET *WINDOW*) L2 0 () GOAL))
    (AND CHAR-POS
	 (SETQ GOAL
	       (DO ((I 0 (1+ I))
		    (CP CHAR-POS))
		   ((>= I *NUMERIC-ARG*)
		    CP)
		 (SETQ CP
		       (OR (STRING-SEARCH-SET '(#\: #\.) L2 (1+ CP))
			   (LET ((BP (END-OF-LINE L2))
				 (nspaces 					;; may 01/18/89
				   (MAX 1 (GET :TAB-WIDTH 'DEFAULT-ATTRIBUTE-VALUE))))
			     (insert-chars bp (in-current-font #\:) 1)		;; may 01/18/89
			     ;; first char in line is a space but really can be considered a ":"
			     (insert-chars bp (in-current-font #\space)
					   (IF (ZEROP (bp-index bp)) 		;; may 02/24/89 
					       nspaces				;; may 02/24/89 
					       (1- nspaces)))
			     ;; add nspaces to end of line ONE, too
			     (insert-chars (end-line bp -1) (in-current-font #\space) nspaces) ;; may 01/18/89
			     (SETQ I (1- I))
			     CP)))))
	 (IF (NOT (CHAR-EQUAL (AREF L2 GOAL) #\:))
	     (INSERT-MOVING POINT (NSUBSTRING (LINE-PREVIOUS L2) CHAR-POS GOAL))
	     (PROGN
	       (DELETE-AROUND *BLANKS* POINT)
	       (INDENT-TO POINT (BP-VIRTUAL-INDENTATION (CREATE-BP L2 GOAL)))))))
  DIS-TEXT) 


(DEFCOM COM-COMPILE-AND-EXIT "Compile the buffer and return from top-level" ()
   (SEND *STANDARD-OUTPUT* :MAKE-COMPLETE)
   (COM-COMPILE-BUFFER)
   (OR (AND (SEND *STANDARD-OUTPUT* :INCOMPLETE-P) ;If any compiler messages
	    (NOT (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		   (Y-OR-N-P "Exit anyway? "))))
       (THROW 'EXIT-TOP-LEVEL ()))
   DIS-NONE) 


(DEFCOM COM-EVALUATE-AND-EXIT "Evaluate the buffer and return from top-level" ()
   (COM-EVALUATE-BUFFER)
   (THROW 'EXIT-TOP-LEVEL ())) 


(DEFCOM COM-GRIND-DEFINITION
   "Grind the definition of a function into the buffer.
Reads the name of the function from the mini-buffer and inserts its ground definition
at point." ()
   (LET ((FUNSPEC (TYPEIN-LINE-READ "Name of function:")))
     (SETQ FUNSPEC (OR (SYSTEM:DWIMIFY-PACKAGE-0 FUNSPEC 'FDEFINEDP) FUNSPEC))
     (IF (NOT (FDEFINEDP FUNSPEC)) (BARF "~A is not a defined function spec." FUNSPEC))
     (PPRINT (FDEFINITION FUNSPEC) (INTERVAL-STREAM-INTO-BP (POINT))))
   DIS-TEXT) 


(DEFCOM COM-GRIND-EXPRESSION
   "Grind the evaluation of a form into the buffer.
Reads a form from the mini-buffer, evals it and inserts the result, ground, at
point." ()
   (LET ((TEM
	  (SI:EVAL-ABORT-TRIVIAL-ERRORS
	   (TYPEIN-LINE-MULTI-LINE-READ "Lisp form: (end with END)"))))
     (GRIND-INTO-BP (POINT) TEM))
   DIS-TEXT) 


(DEFCOM COM-DOWN-INDENTED-LINE "Move to the next line and past any indentation." (KM)
   (LET ((POINT (POINT))
	 (EOL))
     (COND
       ((AND (NOT *NUMERIC-ARG-P*)
	     (BP-= (SETQ EOL (END-LINE POINT))
		   (INTERVAL-LAST-BP *INTERVAL*)))
	(MOVE-BP POINT (INSERT-MOVING EOL #\NEWLINE)) DIS-TEXT)
       (T (MOVE-BP POINT (FORWARD-OVER *BLANKS* (FORWARD-LINE POINT *NUMERIC-ARG* T))) DIS-BPS)))) 


(DEFCOM COM-UP-INDENTED-LINE "Move to previous line and after any indentation." (KM)
   (MOVE-BP (POINT) (FORWARD-OVER *BLANKS* (FORWARD-LINE (POINT) (- *NUMERIC-ARG*) T))) DIS-BPS) 


(DEFUN FIND-FONT-CHANGE (BP LIMIT-BP REVERSE-P &AUX BP1 BP2)
  (COND ((SETQ BP1 (SEARCH BP #\ REVERSE-P () () LIMIT-BP))
	 (IF (NOT REVERSE-P)
	     (SETQ BP1 (DBP BP1)))
	 (SETQ BP2 (FORWARD-CHAR BP1 2 T))
	 (VALUES BP1 BP2 (STRING-INTERVAL BP1 BP2 T))))) 



(DEFUN SEARCH-STRING-SET-KLUDGE (BP STRING-LIST REVERSEP FIXUP-P LIMIT-BP)
  (LET (BP-LIST
	BP-FIRST-FOUND
	STRING)
    (SETQ BP-LIST
       (MAPCAR 'SEARCH (CIRCULAR-LIST BP) STRING-LIST (CIRCULAR-LIST REVERSEP)
	       (CIRCULAR-LIST FIXUP-P) (CIRCULAR-LIST ()) (CIRCULAR-LIST LIMIT-BP)))
    (DO ((BPS BP-LIST (CDR BPS))
	 (STRINGS STRING-LIST (CDR STRINGS)))
	((NULL BPS)
	 (VALUES BP-FIRST-FOUND STRING))
      (AND (CAR BPS)
	   (OR (NULL BP-FIRST-FOUND)
	       (IF REVERSEP
		   (BP-< BP-FIRST-FOUND (CAR BPS))
		   (BP-< (CAR BPS) BP-FIRST-FOUND)))
	   (SETQ BP-FIRST-FOUND (CAR BPS)
		 STRING (CAR STRINGS)))))) 

;;; COM-COPY-FROM-PREVIOUS-LINE:  
;;; This is vastly improved over the way it used to act. You can now copy a whole line
;;; (and only that line) by passing it a BIG numeric arg; it is now smart enough to stop 
;;; when it has run out of characters on the line being copied. It will also copy the first 
;;; line in the buffer (which it wouldn't), and does the right thing when finding the 
;;; previous non-blank line.
;;; TAB characters in the line you are copying into have a predictable (but perhaps unexpected)
;;; effect. Since we are merely copying by character position, WYSIWYG but what you see may not 
;;; be what you think you see. 
(DEFCOM COM-COPY-FROM-PREVIOUS-LINE "Copy characters from the last non-blank line. Use a numeric 
argument for the maximum number of characters to copy; thus, a BIG one gets the whole line." ()
  (LET* ((POINT (POINT))
	 (LEN 0)
	 (START-POS 0)
	 (BP (DO* ((BP (BEG-LINE POINT -1) (BEG-LINE BP -1))
		   (IDX (BP-INDEX POINT))
		   (LINE (BP-LINE BP) (BP-LINE BP)))
		  ((OR (NULL LINE)
		       (AND (NOT (LINE-BLANK-OR-DIAGRAM-P LINE))
			    (> (SETQ START-POS IDX)
			       (SETQ LEN (LINE-LENGTH LINE)))))
		   (BARF))
	       (AND (NOT (LINE-BLANK-OR-DIAGRAM-P LINE))
		    (RETURN (CREATE-BP LINE IDX)))))
	 (*NUMERIC-ARG* (MIN *NUMERIC-ARG* (ABS (- LEN START-POS)))))
    (INSERT-INTERVAL-MOVING POINT (COPY-INTERVAL BP (FORWARD-CHAR BP *NUMERIC-ARG* T))))
  DIS-TEXT)

;;;COM-COMPOSE-CHARACTER and COMPOSE added for ISO support by rpm (from hw) on 11-25-86.  
(DEFCOM COM-COMPOSE-CHARACTER "Compose a complex character
from a base character and a diacritic." (NM)
  (LET ((POINT (POINT)))
    (INSERT-MOVING POINT  (COMPOSE))
    (VALUES DIS-LINE (BP-LINE POINT) (BP-INDEX POINT))))

(DEFUN COMPOSE () "Allows composition of a complex character from a base character and
a diacritic (the little mark above the character.)  The composed character is returned."
  (LET ((base-character (PROMPT-AND-READ :character "~&Base character: "))
	(diacritic (PROMPT-AND-READ :character "~&Diacritic: ")))
    (CASE diacritic
      (#\~ (CASE base-character                 ;tilde	   
	     (#\N (MAKE-CHAR #xD1))
	     (#\A (MAKE-CHAR #xC3))
	     (#\O (MAKE-CHAR #xD5))
	     (#\a (MAKE-CHAR #xE3))
	     (#\o (MAKE-CHAR #xF5))
	     (#\n (MAKE-CHAR #xF1))
	     (otherwise (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))
      
      (#\^ (CASE base-character                 ;circumflex
	     (#\A (MAKE-CHAR #xC2))
	     (#\E (MAKE-CHAR #xCA))
	     (#\I (MAKE-CHAR #xCE))
	     (#\O (MAKE-CHAR #xD4))
	     (#\U (MAKE-CHAR #xDB))
	     (#\a (MAKE-CHAR #xE2))
	     (#\e (MAKE-CHAR #xEA))
	     (#\i (MAKE-CHAR #xEE))
	     (#\o (MAKE-CHAR #xF4))
	     (#\u (MAKE-CHAR #xFB))
	     (otherwise (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))

      (#\` (CASE base-character                 ;grave accent
	     (#\A (MAKE-CHAR #xC0))
	     (#\E (MAKE-CHAR #xC8))
	     (#\I (MAKE-CHAR #xCC))
	     (#\O (MAKE-CHAR #xD2))
	     (#\U (MAKE-CHAR #xD9))
	     (#\a (MAKE-CHAR #xE0))
	     (#\e (MAKE-CHAR #xE8))
	     (#\i (MAKE-CHAR #xEC))
	     (#\o (MAKE-CHAR #xF2))
	     (#\u (MAKE-CHAR #xF9))
	     (otherwise (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))

      (#\/ (CASE base-character                 ;slash
	     (#\A (MAKE-CHAR #xC6))
	     (#\O (MAKE-CHAR #xD8))
	     (#\a (MAKE-CHAR #xE6))
	     (#\o (MAKE-CHAR #xF8))
	     (otherwise (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))

      (#\- (CASE base-character                 ;hyphen
	     (#\D (MAKE-CHAR #xB0))
	     (#\P (MAKE-CHAR #xDE))
	     (#\d (MAKE-CHAR #xF0))
	     (#\p (MAKE-CHAR #xFE))
	     (otherwise (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))

      (#\ (CASE base-character           ;diaresis
	     (#\A (MAKE-CHAR #xC4))
	     (#\E (MAKE-CHAR #xCB))
	     (#\I (MAKE-CHAR #xCF))
	     (#\O (MAKE-CHAR #xD6))
	     (#\U (MAKE-CHAR #xDC))
	     (#\a (MAKE-CHAR #xE4))
	     (#\e (MAKE-CHAR #xEB))
	     (#\i (MAKE-CHAR #xEF))
	     (#\o (MAKE-CHAR #xF6))
	     (#\u (MAKE-CHAR #xFC))
	     (#\y (MAKE-CHAR #xFF))
	     (otherwise (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))

      (#\ (CASE base-character           ;ring
             (#\A (MAKE-CHAR #xC5))
	     (#\a (MAKE-CHAR #xe5))
	     (otherwise (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))
      
      ((#xB4 #x27) (CASE base-character         ;allow single quote (#x27) or acute accent (#xB4)
		         (#\A (MAKE-CHAR #xC1))
		 	 (#\E (MAKE-CHAR #xC9))
			 (#\I (MAKE-CHAR #xCD))
			 (#\O (MAKE-CHAR #xD3))
			 (#\U (MAKE-CHAR #xDA))
			 (#\Y (MAKE-CHAR #xDD))
			 (#\a (MAKE-CHAR #xE1))
			 (#\e (MAKE-CHAR #xE9))
			 (#\i (MAKE-CHAR #xED))
			 (#\o (MAKE-CHAR #xF3))
			 (#\u (MAKE-CHAR #xFA))
			 (#\y (MAKE-CHAR #xFD))
			 (otherwise
			  (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))

      ((#x2C #xB8)  (CASE base-character        ;allow comma (#x2C) or cedilla (#xB8) [they look alike]
			   (#\C (MAKE-CHAR #xC7))
			   (#\c (MAKE-CHAR #xE7))
			   (otherwise
			    (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic ))))

      (otherwise (BARF "~&Cannot compose character from ~C and ~C." base-character diacritic )))))