;;; Zwei commands, see ZWEI;COMA for comments -*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1980, Massachusetts Institute of Technology
;;;
;;; Change History :
;;;
;;; 04/20/89 	MAY	Changed COM-QUICK-DOCUMENTATION & COM-LONG-DOCUMENTATION 
;;;                     to handle special compiler functons like (ticlos:call-next-method). SPR 9579
;;; 01/20/89 	MAY	Modified documentation for com-long-documentation. SPR 9229.
;;; 07-11-88	MAY  	Added dotted pair test in FIND to handle forms like "(a . b)" (SPR 8272)
;;; 07-11-88    MAY     Changed COM-LONG-DOCUMENTATION to handle function-specifier forms,
;;;                     and provide for variable documentation, which was broke,
;;;                     and distinquish between constants and variables. Fixes SPR 8237 & 6077
;;; 12-21-88	MAY	Changed DO-DISASSEMBLE to allow closure spec. Spr 8949

(DEFUN FIND-CONTAINING-ATOM (BP SET)
  "Find the list around BP whose first element is a symbol STRNG-EQUAL to an elt of SET.
The first value is the BP at the front of the symbol;
the second is the element of SET.
The values are NIL if no containing list starts that way."
  (DO ((BP BP)
       (BP1)
       (BP2)
       (TEM))
      (NIL)
    (OR (SETQ BP (FORWARD-LIST BP -1 () 1))
	(RETURN ()))
    (SETQ BP1 (FORWARD-LIST BP 1 () -1 T)
	  BP2 (FORWARD-ATOM BP1))
    (AND (SETQ TEM (MEMBER (STRING-INTERVAL BP1 BP2 T) SET :TEST #'STRING-EQUAL))
	 (RETURN BP1 (CAR TEM))))) 


(DEFUN COUNT-LIST-ELEMENTS (BP &AUX START-BP END-BP)
  "Return the number of elements in the list whose text follows BP."
  (if (SETQ START-BP (FORWARD-LIST BP 1 () -1 T))
      (progn
	(SETQ END-BP (FORWARD-SEXP (DBP START-BP)))
	(DO ((BP (FORWARD-LIST BP 1 () -1 T) (FORWARD-SEXP BP))
	     (I -1 (1+ I)))
	    (NIL)
	  (AND (NULL BP) (RETURN ()))
	  (AND (BP-= BP END-BP) (RETURN I)))))) 

;;; This tries to find someplace that looks like it probably doesn't have enough parens
;;; It takes the first place that has a lesser indentation level than the given BP.

(DEFUN LIKELY-UNBALANCED-POINT (BP LIMIT-BP)
  "Return a bp to a point after BP that may not have enough parens.
The scan stops at LIMIT-BP.  Value can be LIMIT-BP."
  (DO ((IND (BP-INDENTATION BP))
       (LINE (LINE-NEXT (BP-LINE BP)) (LINE-NEXT LINE))
       (LIMIT-LINE (BP-LINE LIMIT-BP))
       (OLINE (BP-LINE BP)))
      ((EQ LINE LIMIT-LINE)
       LIMIT-BP)
    (COND ((NOT (MEMBER (LINE-TYPE LINE) '(:COMMENT :BLANK) :TEST #'EQ))
	   (AND (<= (LINE-INDENTATION LINE) IND)
		(RETURN (END-OF-LINE OLINE)))
	   (SETQ OLINE LINE))))) 






(DEFCOM COM-JUST-ONE-SPACE "Replace all whitespace around point with arg spaces" ()
   (LET ((SPACES-BEFORE (- (BP-INDEX (POINT)) (BP-INDEX (BACKWARD-OVER *BLANKS* (POINT))))))
     (DELETE-AROUND *BLANKS* (POINT))
     (DOTIMES (I *NUMERIC-ARG*)
       (INSERT (POINT) #\SPACE))
     (MOVE-BP (POINT) (FORWARD-CHAR (POINT) (MIN SPACES-BEFORE *NUMERIC-ARG*))))
   DIS-TEXT) 

(DEFCOM COM-FIND-UNBALANCED-PARENTHESES "Find parenthesis error in buffer" ()
   (LET ((BEG-BP (INTERVAL-FIRST-BP *INTERVAL*))
	 (END-BP (INTERVAL-LAST-BP *INTERVAL*))
	 (POINT (POINT))
	 (*BATCH-UNDO-SAVE* T)
	 (OLD-TICK (NODE-TICK *INTERVAL*))
	 BEG-BP-1
	 END-BP-1
	 BP)
     (MULTIPLE-VALUE-BIND (IGNORE IGNORE COMMENT-P)
	 (LISP-BP-SYNTACTIC-CONTEXT END-BP)
       (WHEN COMMENT-P 
	 (MOVE-BP END-BP (INSERT END-BP #\NEWLINE))))
     (UNWIND-PROTECT
	 (PROGN
	   (SETQ BEG-BP-1 (COPY-BP BEG-BP :MOVES)
		 END-BP-1 (COPY-BP END-BP :NORMAL))
	   (INSERT BEG-BP-1 "(")
	   (INSERT END-BP-1 ")")
	   (IF (SETQ BP (FORWARD-SEXP BEG-BP))
	       (IF (BP-= BP END-BP)		;All ok
		   (FORMAT *QUERY-IO* "~&All parens appear balanced.")
		   (PROGN
		     (POINT-PDL-PUSH POINT *WINDOW*)
		     (MOVE-BP POINT BP)
		     (FORMAT *QUERY-IO* "~&Probably extra right-paren here.")))
	       (PROGN
		 (OR (SETQ BP (FORWARD-SEXP END-BP -1))
		     (BARF "Cannot find unbalanced parenthesis"))
		 (POINT-PDL-PUSH POINT *WINDOW*)
		 (MOVE-BP POINT BP)
		 (FORMAT *QUERY-IO* "~&Probably no right-paren for this left-paren."))))
       (COND (BEG-BP-1
	      (DELETE-INTERVAL BEG-BP BEG-BP-1 T)
	      (FLUSH-BP BEG-BP-1)))
       (COND (END-BP-1
	      (DELETE-INTERVAL END-BP-1 END-BP T)
	      (FLUSH-BP END-BP-1)))
       (SETF (NODE-TICK *INTERVAL*) OLD-TICK)))
   DIS-BPS) 

;;; Pattern finding command

;;;
;;;  Change **-for-one-...-for-many to ?-for-one-??-for-many.  Unfortunately,
;;;  ... is illegal to the Common Lisp reader.  Although ... is legal to Zetalisp,
;;;  reading the pattern in Zetalisp all the time seems a bit arbitrary, so I
;;;  changed the special characters.  A prettier but harder solution would be
;;;  to maintain two special readtables (Zetalisp and Common Lisp) with ... legal.

(DEFCOM COM-FIND-PATTERN "Move to next occurrence of the given pattern.
The pattern must be a list.  ? matches any one thing, ?? any number of things, 
with even zero things being a match. A numeric argument repeats the last search." ()
   (LET (FORM RESTART BP)
     (COND (*NUMERIC-ARG-P*
	    (SETQ FORM (OR *LAST-PATTERN* (BARF "No previous pattern")))
	    (FORMAT *QUERY-IO* "~&Finding ~S ..." FORM)
	    (AND (BP-= (POINT) *LAST-PATTERN-BP*)
		 (SETQ RESTART *LAST-PATTERN-RESTART-LIST*)))
	   (T
	    (LET ((FORM-STRING (TYPEIN-LINE-READLINE "Pattern to search for:"))
		  (EOF '(NIL)))
	      (SETQ FORM (READ-FROM-STRING FORM-STRING NIL EOF))
	      (AND (EQUAL FORM EOF)
		   (BARF "Unbalanced parens"))
	      (OR (CONSP FORM)
		  (BARF "I only know how to search for lists"))
	      ;; This is sort of a kludge
	      (OR (EQ *PACKAGE* PKG-KEYWORD-PACKAGE)
		  (SETQ FORM (SUBLIS (LIST (CONS (INTERN "?") :?) (CONS (INTERN "??") :??)) FORM))))))
     (MULTIPLE-VALUE-SETQ (BP RESTART) (FIND-PATTERN (POINT) FORM RESTART))
     (OR BP (BARF "Can't find ~S." FORM))
     (MAYBE-PUSH-POINT BP)
     (MOVE-BP (POINT) BP)
     (SETQ *LAST-PATTERN* FORM
	   *LAST-PATTERN-BP* BP
	   *LAST-PATTERN-RESTART-LIST* RESTART))
   DIS-BPS) 

;;; Attempt to find an instance of THING after BP, return a new BP if successful
(DEFUN FIND-PATTERN (BP PATTERN &OPTIONAL RESTART-LIST)
  (BLOCK FIND-PATTERN
    (DO ((BP1 (FORWARD-DEFUN BP -1 T) (FORWARD-DEFUN BP2))
	 (BP2)
	 (STREAM (INTERVAL-STREAM *INTERVAL*))
	 (FORM)
	 (SI:XR-CORRESPONDENCE-FLAG T)
	 (SI:XR-CORRESPONDENCE NIL NIL)
	 (RESTART-LIST RESTART-LIST NIL)
	 (PLIST)
	 (TEM))
	((NULL BP1)
	 NIL)
      (DECLARE (SPECIAL SI:XR-CORRESPONDENCE-FLAG))
      (SETQ PLIST (LINE-CONTENTS-PLIST (BP-LINE BP1)))
      (SETQ BP2 (FORWARD-SEXP BP1))			;Find the end of this defun
      ;; Now get the form and correspondence for this defun, using previous if there
      (COND (BP2
	     (COND ((AND (SETQ TEM (GETF PLIST 'CORRESPONDENCE))
			 (COND ((> (CADR TEM) (INTERVAL-REAL-TICK BP1 BP2 T))
				(SETQ FORM (CAR TEM) SI:XR-CORRESPONDENCE (CADDR TEM))
				T)
			       (T
				(REMF PLIST 'CORRESPONDENCE)
				NIL))))
		   (T (SEND STREAM :SET-BP BP1)
		      (SETQ FORM (READ STREAM T))
		      (SETF (GETF PLIST 'CORRESPONDENCE) (LIST FORM (TICK) SI:XR-CORRESPONDENCE))))
	     (AND RESTART-LIST
		  (SETQ FORM (CAR RESTART-LIST) RESTART-LIST (CDR RESTART-LIST)))
	     (DO ((FORM FORM (CAR RESTART-LIST))
		  (RESTART-LIST RESTART-LIST (CDR RESTART-LIST))
		  (FOUND)
		  (BP3))
		 (NIL)
	       (MULTIPLE-VALUE-SETQ (FOUND RESTART-LIST)
				    (FIND PATTERN FORM RESTART-LIST))
	       (OR FOUND (RETURN ()))
	       (AND (SETQ BP3 (CADR (MEMBER FOUND SI:XR-CORRESPONDENCE :TEST #'EQ)))
		    (BP-< BP BP3)
		    (RETURN-FROM FIND-PATTERN BP3 RESTART-LIST))))
	    (T					;Look forward for next defun
	     (SETQ BP2 BP1))))))

;;; Attempt to find an instance of THING in LIST

(DEFUN FIND (THING LIST &OPTIONAL RESTART-LIST &AUX VAL)
  (COND ((AND RESTART-LIST
	      (MULTIPLE-VALUE-SETQ (VAL RESTART-LIST)
		(FIND THING (CAR RESTART-LIST) (CDR RESTART-LIST))))
	 (PUSH LIST RESTART-LIST))
	((MATCH THING LIST)
	 (SETQ VAL LIST RESTART-LIST ()))
	((ATOM LIST)
	 (SETQ VAL () RESTART-LIST ()))
	((NOT (consp (cdr list)))				;; may 7-11-88 added dotted pair test
	 (SETQ val () restart-list ()))				;; may 7-11-88
	(T
	 (DO ((LIST LIST (CDR LIST)))
	     ((NULL LIST)
	      (SETQ VAL () RESTART-LIST ()))
	   (MULTIPLE-VALUE-SETQ (VAL RESTART-LIST)
	     (FIND THING (CAR LIST)))
	   (COND (VAL
		  (PUSH (CDR LIST) RESTART-LIST)
		  (RETURN ()))))))
  (VALUES VAL RESTART-LIST)) 

;;; Simple minded pattern matcher
;;; ? matches an arbitrary frob, ?? an arbitrary number (possibly 0) of frobs

(DEFUN MATCH (A B)
  (DO ((A A (CDR A))
       (B B (CDR B))
       (VAL))
      (NIL)
    (COND ((EQ A B) (RETURN T))
	  ((EQ A :??) (RETURN 'CDR))
	  ((EQ A :?) (RETURN T))
	  ((NOT (= (SYS:%DATA-TYPE A) (SYS:%DATA-TYPE B))) (RETURN NIL))
	  ((CHARACTERP A) (RETURN (CHAR= A B)))
	  ((ARRAYP A) (RETURN (AND (STRINGP A) (STRINGP B) (STRING-EQUAL A B))))
	  ((NUMBERP A) (RETURN (= A B)))
	  ((ATOM A) (RETURN NIL))
	  ((NOT (SETQ VAL (MATCH (CAR A) (CAR B)))) (RETURN NIL))
	  ((NEQ VAL T) (RETURN (OR (NULL (SETQ A (CDR A)))
				   (DO ((B B (CDR B)))
				       ((NULL B)
					NIL)
				     (AND (MATCH A B) (RETURN T))))))))) 


(DEFCOM COM-EXECUTE-COMMAND-INTO-BUFFER
   "Execute following editor command, printing into the buffer.
Any output from the command which would ordinarily appear as type out
is inserted into the current buffer instead.
Trace and warning output are also inserted in the buffer." ()
   (LET* ((*TYPEOUT-WINDOW* (MAKE-INTERVAL-TYPEOUT-STREAM))
	  (*STANDARD-OUTPUT* *TYPEOUT-WINDOW*)
	  (*TRACE-OUTPUT* *TYPEOUT-WINDOW*)
	  (*ERROR-OUTPUT* *TYPEOUT-WINDOW*))
     (CLEAR-PROMPTS)
     (ADD-PROMPT "Key: ")
     (ALWAYS-DISPLAY-PROMPTS)
     (WITH-UNDO-SAVE ("Command output" (POINT) (POINT) T)
	(UNWIND-PROTECT (DO ()
			    ((NEQ :ARGUMENT
				  (PROCESS-COMMAND-CHAR (INPUT-WITH-PROMPTS *STANDARD-INPUT* :TYI)))))
			 ;; Redisplay properly if command is aborted.
	  (MUST-REDISPLAY *WINDOW* DIS-TEXT)))
     (MOVE-BP (MARK) (POINT))
     (MOVE-BP (POINT) (SEND *STANDARD-OUTPUT* :READ-BP))
     (SETQ *CURRENT-COMMAND-TYPE* 'YANK))
   DIS-TEXT) 


(DEFCOM COM-INSERT-DATE
   "Print the curent date into the buffer.
Calls TIME:PRINT-CURRENT-TIME, or if given an argument TIME:PRINT-CURRENT-DATE" ()
   (LET ((STREAM (INTERVAL-STREAM (POINT) (POINT) T)))
     (FUNCALL (IF *NUMERIC-ARG-P* #'TIME:PRINT-CURRENT-DATE #'TIME:PRINT-CURRENT-TIME) STREAM)
     (MOVE-BP (MARK) (POINT))
     (MOVE-BP (POINT) (SEND STREAM :READ-BP)))
   DIS-TEXT) 


(DEFCOM COM-COUNT-LINES-REGION "Print the number of lines in the region in the echo area." (KM)
   (REGION (BP1 BP2) (FORMAT *QUERY-IO* "~&~D line~:P.  " (1- (COUNT-LINES BP1 BP2 T)))) DIS-NONE) 


(DEFCOM COM-WHERE-AM-I
	"Print various things about where the point is.
Print the X and Y positions, the octal code for the following character,
the current line number and its percentage of the total file size.
If there is a region, the number of lines in it is printed.
Fast Where Am I prints a subset of this information faster." (KM)
  (REDISPLAY *WINDOW* :POINT NIL NIL T)
  (LET ((POINT (POINT))
	(FIRST-BP (INTERVAL-FIRST-BP *INTERVAL*))
	(LAST-BP (INTERVAL-LAST-BP *INTERVAL*)))
    (LET ((POINT-LINES (1- (COUNT-LINES FIRST-BP POINT)))
	  (INTERVAL-LINES (1- (COUNT-LINES FIRST-BP LAST-BP)))
	  (AT-END-P (BP-= (INTERVAL-LAST-BP *INTERVAL*) POINT))
	  (BP-IND (BP-INDENTATION POINT))
	  (SW (FONT-SPACE-WIDTH)))
      (FORMAT *QUERY-IO*
	      "~&X=[~D. chars|~D. pixels|~:[~S~;~D.~] columns] ~
			Y=~D.~@[ Char=~@C~] Line=~D.(~D%)"
	      (BP-INDEX POINT)
	      BP-IND
	      (ZEROP (REM BP-IND SW))
	      (IF (ZEROP (REM BP-IND SW))
		  (/ BP-IND SW)
		  (/ (FLOAT BP-IND) SW))
	      (FIND-BP-IN-WINDOW *WINDOW* POINT)
	      (AND (NOT AT-END-P) (BP-CHAR POINT))
	      POINT-LINES
	      (IF (ZEROP INTERVAL-LINES)
		  0
		  (TRUNCATE (* 144 POINT-LINES) INTERVAL-LINES)))))
  (AND (WINDOW-MARK-P *WINDOW*)
       (REGION (BP1 BP2)
	 (FORMAT *QUERY-IO* ", Region has ~D line~:P.  " (1- (COUNT-LINES BP1 BP2 T)))))
  DIS-NONE) 


(DEFCOM COM-ARGLIST "Prints the argument list of a function. Reads the function name from the minibuffer (the top
of the kill ring has the \"current\" function from the buffer) and prints the arglist in the echo area." ()
  (LET ((NAME (READ-FUNCTION-NAME "Arglist" (RELEVANT-FUNCTION-NAME (POINT)) T)))
    (PRINT-ARGLIST NAME))
  DIS-NONE) 


(DEFCOM COM-QUICK-ARGLIST "Print the argument list of the function to left of cursor." ()
   (QUICK-ARGLIST) DIS-NONE) 


(DEFUN QUICK-ARGLIST (&OPTIONAL (STREAM *QUERY-IO*))
  (IF *NUMERIC-ARG-P*
      (LET ((NAME (READ-FUNCTION-NAME "Arglist" (RELEVANT-FUNCTION-NAME (POINT)) T)))
	(PRINT-ARGLIST NAME STREAM))
      (LET ((SYMBOL (RELEVANT-FUNCTION-NAME (POINT))))
	(COND ((COND ((MEMBER SYMBOL '(FUNCALL FUNCALL-SELF SEND
	       			       LEXPR-FUNCALL LEXPR-FUNCALL-SELF LEXPR-SEND <-)
			      :TEST #'EQ)
		      (LET ((TEMP-SYMBOL (RELEVANT-METHOD-NAME
					   (POINT)
					   (IF (MEMBER SYMBOL '(FUNCALL-SELF
								 LEXPR-FUNCALL-SELF)
						       :TEST #'EQ)
					       1 2))))
			(AND TEMP-SYMBOL (SETQ SYMBOL TEMP-SYMBOL))))
		     ((EQ SYMBOL 'DEFMETHOD)
		      (LET ((METHOD-SYMBOL (RELEVANT-DEFMETHOD-METHOD-NAME (POINT))))
			(COND (METHOD-SYMBOL
			       (SETQ SYMBOL METHOD-SYMBOL)
			       T)))))
	       (MULTIPLE-VALUE-BIND (ARGLIST NAME RETLIST) (METHOD-ARGLIST SYMBOL)
		 (COND ((EQ STREAM *QUERY-IO*)
			(FORMAT *QUERY-IO* "~&")
			(SEND STREAM :SEND-IF-HANDLES :TYPEOUT-STAYS)))
		 (FORMAT STREAM "~S: ~:A~@[  ~:A~]" (OR NAME SYMBOL) ARGLIST RETLIST)))
	      ((AND SYMBOL
		    (OR (FDEFINEDP SYMBOL)
			(SI:MEMQ-ALTERNATED 'ARGLIST (SYMBOL-PLIST SYMBOL))))
	       (PRINT-ARGLIST SYMBOL STREAM))
	      ((BARF))))))			;Looked hard but couldn't find a defined function


(DEFUN PRINT-ARGLIST (FSPEC &OPTIONAL (STREAM *QUERY-IO*))
  (COND ((EQ STREAM *QUERY-IO*)
	 (FORMAT *QUERY-IO* "~&")
	 (SEND STREAM :SEND-IF-HANDLES :TYPEOUT-STAYS)))
  (MULTIPLE-VALUE-BIND (ARGLIST RETURNS TYPE) (ARGLIST FSPEC)
    (FORMAT STREAM "~S~@[ (~A)~]: " FSPEC TYPE)
    (IF (OR (CONSP ARGLIST) (NULL ARGLIST))
	(PRINT-ARGLIST-INTERNAL ARGLIST STREAM)
	(PRINC "??" STREAM))
    (AND RETURNS (FORMAT STREAM "  ~:A" RETURNS))
    (AND (NOT (FDEFINEDP FSPEC)) (SYMBOLP FSPEC)
       (SI:MEMQ-ALTERNATED 'ARGLIST (SYMBOL-PLIST FSPEC))
       (PRINC " (only in compiled code)" STREAM)))) 

;; This prints an arglist in a convenient form, ie:
;; (si:first &special si:second &local &optional (si:third (quote si:default)))
;; prints: (first &special si:second &local &optional (third 'si:default))

(DEFUN PRINT-ARGLIST-INTERNAL (LIST STREAM &AUX SPECIAL)
  (WRITE-CHAR #\( STREAM)
  (DO ((L LIST (CDR L)))
      ((NULL L)
       (WRITE-CHAR #\) STREAM))
    (COND ((SYMBOLP L)
	   (WRITE-STRING ". " STREAM)
	   (FUNCALL (IF SPECIAL #'PRIN1 #'PRINC) L STREAM)
	   (WRITE-CHAR #\) STREAM)
	   (RETURN ())))
    (CASE (CAR L)
      (&SPECIAL (SETQ SPECIAL T))
      (&LOCAL (SETQ SPECIAL ())))
    (COND ((OR (ATOM (CAR L))  ;If the element is a symbol
	       (ATOM (CDAR L)) ;Or if it's not a list with exactly two elmts.
	       (NOT (NULL (CDDAR L))))
	   (FUNCALL (IF SPECIAL #'PRIN1 #'PRINC) (CAR L) STREAM))	;Just print it.
	  (T
	   ;; This is the special case of an element with a default.
	   (WRITE-CHAR #\( STREAM)
	   (FUNCALL (IF SPECIAL #'PRIN1 #'PRINC) (CAAR L) STREAM)
	   (WRITE-CHAR #\SPACE STREAM)
	   (COND ((AND (CONSP (CADAR L))
		       (EQ (CAADAR L) 'QUOTE))
		  ;; If the default is quoted, print it nicely.
		  (WRITE-CHAR #\' STREAM)
		  (PRIN1 (CADR (CADAR L)) STREAM))
		 ((EQ (CADAR L) 'SI::*HAIRY*)
		  (WRITE-STRING "??" STREAM))
		 (T (PRIN1 (CADAR L) STREAM)))
	   (WRITE-CHAR #\) STREAM)))
    (AND (CDR L) (WRITE-CHAR #\SPACE STREAM)))) 


;; may 04/20/89 Allowed c-sh-d to get at special compiler arglist documentation.
(DEFCOM COM-QUICK-DOCUMENTATION
   "Prints documentation for the function point is at.
Prints the documentation string of the function which point is inside a call to.
With a numeric argument, reads the name of the function to document
from the mini buffer." ()
   (LET ((NAME (RELEVANT-FUNCTION-NAME (POINT))))
     (IF *NUMERIC-ARG-P*
	 (SETQ NAME (READ-FUNCTION-NAME "Brief Document" NAME T)))
     (LET ((DOC (DOCUMENTATION NAME 'FUNCTION)))
       (COND ((NULL DOC)
	      (IF (FDEFINEDP NAME)
		  (PRINT-ARGLIST NAME))
	      (FORMAT *QUERY-IO* "~&~S is not documented." NAME))
	     (T
	      (IF (or (FDEFINEDP NAME)  
		      (and (symbolp name) 							;; may 04/20/89
			   ;; for special compiler functions like: (ticlos:call-next-method) 	;; may 04/20/89 
			   (listp (get name 'arglist :default))))				;; may 04/20/89 
		  (PROGN
		    (FRESH-LINE)
		    (PRINT-ARGLIST NAME *STANDARD-OUTPUT*))
		  (FORMAT T "~S:" NAME))
	      (FORMAT T "~%~A" DOC)))))
   DIS-NONE) 


;; COM-LONG-DOCUMENTATION redefined by patch 94.215.
;; may 01/20/89 Changed documentation string and 'structure test.
;; may 04/20/89 Allowed meta-sh-d to get at special compiler arglist/function documentation.
(DEFCOM COM-LONG-DOCUMENTATION 
        "Print long documentation for almost anything. Reads the name 
from the mini-buffer or mouse (the default is the \"current\" 
function from the buffer) and displays the function's arguments,
whether it is a function, variable, flavor, structure, or condition, 
and its documentation" () ;;by OREN/STENGER. Patched in by gsl
  (LET ((NAME (READ-FUNCTION-NAME "Document" (RELEVANT-FUNCTION-NAME (POINT)) 'AARRAY-OK))
        (*PACKAGE* NIL)
        ;; Bind package to nil to force it to print out
        DOC
        TEMP)
    ;; (fdefinedp '(:method tv:sheet :init)) => t 
    ;; (functionp '(:method tv:sheet :init)) => nil
    (COND-EVERY (;; function
                 (or (FUNCTIONP NAME T) (fdefinedp name) 	;; may 7-11-88 added (fdefinedp name)
		     ;; for special compiler functions like: (ticlos:call-next-method) 	;; may 04/20/89 
		     (and (symbolp name) (documentation name))) 			;; may 04/20/89 
                 (MULTIPLE-VALUE-BIND (ARGS RETURN MACRO-P) (ARGLIST NAME)
                   (IF ;;(SETQ TEMP (GET NAME 'SI::DEFSTRUCT-SLOT)) 				;; may 01/25/89 obsolete
                       ;;(FORMAT T "~%~S is a defstruct accessor macro for ~s" NAME (CAR TEMP)) ;; may 01/25/89 sort of ...
                     (and (setq temp (si:function-parent name))					;; may 01/25/89 
                          (get temp 'si::defstruct-description)) 				;; may 01/25/89 
                     (FORMAT T "~%~S is a defstruct accessor function for ~s" NAME TEMP) 	;; may 01/25/89 
                     (FORMAT T "~%~s is a ~:[function~;~:*~a~]" NAME				;; may 01/25/89 
                             (and MACRO-P (string-downcase macro-p)))) 				;; may 01/25/89 macro/subst 
                   (FORMAT T "~%arguments  ~A~%" ARGS)
                   (WHEN RETURN
                     (FORMAT T "returns  ~{~A  ~}~%" RETURN)))
                 (WHEN (SETQ DOC (DOCUMENTATION NAME))
                   (FORMAT T "~A~%" DOC)))
                (;; flavor
                 (SETQ TEMP (GET NAME 'SI:FLAVOR))
                 (FORMAT T "~%~s is a flavor~%" NAME)
                 (WHEN (AND (SETQ TEMP (SI:FLAVOR-PLIST TEMP))
                            (SETQ DOC (SECOND (MEMBER :DOCUMENTATION TEMP :TEST #'EQ))))
                   (FORMAT T "~a~%" DOC)))
                (;; structure
                 (SETQ TEMP (GET NAME 'SI::DEFSTRUCT-DESCRIPTION))
                 (FORMAT T "~%~s is a structure~%" NAME)
                 (WHEN
                   ;;(AND (SETQ TEMP (SI::DEFSTRUCT-DESCRIPTION-PROPERTY-ALIST TEMP))	;; may 01/25/89 obsolete
                   ;;(SETQ DOC (REST (ASSOC :DOCUMENTATION TEMP :TEST #'EQ))))  	;; may 01/25/89 obsolete
                   (setq doc (documentation name 'structure)) 				;; may 01/25/89 
                   (FORMAT T "~a~%" DOC)))
                (;; other ...
                 (AND (SETQ TEMP (GET NAME :DOCUMENTATION)) (NEQ DOC TEMP))
                 (FORMAT T "~%~s is a ~a~%" NAME (TYPE-OF NAME))
                 (FORMAT T "~a~%" TEMP))
                (;; variable
                 (AND (SYMBOLP NAME) (BOUNDP NAME))
                 (if (get name 'compiler:system-constant) ;; may 7-11-88
                     (FORMAT T "~%~s is a constant~%" NAME);; may 7-11-88
                     (FORMAT T "~%~s is a variable~%" NAME))
                 (IF (CONSP (SETQ TEMP (SYMBOL-VALUE NAME)))
                     (FORMAT T "Value is a ~A~%" (TYPE-OF TEMP))
                     (FORMAT T "Value is ~s~%" TEMP))
                 (FORMAT T "~@[~a~%~]" (getf (GET NAME 'sys:documentation-property) 'variable))) ;; may 7-11-88
                (;; condition
                 (SETQ TEMP (GET NAME 'EH::MAKE-CONDITION-FUNCTION))
                 (FORMAT T "~%~S is an ~:[undocumented ~]error condition~%" NAME
                         (SETQ DOC (DOCUMENTATION TEMP)))
                 (WHEN DOC
                   (FORMAT T "~a~%" DOC)))
                (OTHERWISE (FORMAT *QUERY-IO* "~&~S was not found." NAME))))
  DIS-NONE) 


(DEFCOM COM-DESCRIBE-VARIABLE-AT-POINT
   "Prints information about variable at or before cursor.
The information printed is whether it is declared special, whether it has a value,
and whether it has documentation put on by DEFVAR.  If none of these are present,
looks for lookalike symbols in other packages." ()
   (LET* ((BP1 (FORWARD-ATOM (FORWARD-CHAR (POINT) 1 T) -1 T))
	  (BP2 (FORWARD-ATOM BP1)))
     (IF (NULL BP2) (BARF))
     (MULTIPLE-VALUE-BIND (VAR ERROR)
	 (CATCH-ERROR (WITH-INPUT-FROM-STRING (S (BP-LINE BP1) (BP-INDEX BP1) (BP-INDEX BP2))
			(READ S T)))
       (IF (OR ERROR (NOT (SYMBOLP VAR))) (BARF))
       (SEND *QUERY-IO* :CLEAR-SCREEN)  ;Clear the echo area
       (SEND *QUERY-IO* :SEND-IF-HANDLES :TYPEOUT-STAYS)
       (COND ((NOT (DESCRIBE-VARIABLE-INTERNAL VAR))
	      (FORMAT *QUERY-IO* "~&~S is not a declared variable." VAR)
	      (MAPC #'DESCRIBE-VARIABLE-INTERNAL
		    (PACKAGE-LOOKALIKE-SYMBOLS VAR () '(SPECIAL COMPILER::SYSTEM-CONSTANT)))))
       DIS-NONE))) 


(DEFUN DESCRIBE-VARIABLE-INTERNAL (VAR)
  (LET* ((DECL (GETL VAR '(SPECIAL COMPILER::SYSTEM-CONSTANT)))
	 (BOUND (BOUNDP VAR))
	 (DOC (DOCUMENTATION VAR 'VARIABLE))
	 (STREAM (IF DOC *STANDARD-OUTPUT* *QUERY-IO*)))
    (COND((OR DECL BOUND DOC)
	  (FORMAT STREAM "~&~S has ~:[no~;a~] value" VAR BOUND)
	  (IF (EQ (CAR DECL) 'SPECIAL)
	      (FORMAT STREAM " and is declared special ~:[by file ~A~]" (EQ (CADR DECL) T)
		      (CADR DECL)))
	  (IF (EQ (CAR DECL) 'COMPILER::SYSTEM-CONSTANT)
	      (FORMAT STREAM " and is a system-constant"))
	  (IF DOC (FORMAT STREAM "~%~A" DOC)) T)))) 


(DEFCOM COM-TRACE "Trace or untrace a function.
Reads the name of the function from the mini-buffer (the top of the kill
ring has the \"current\" function from the buffer) then pops up a menu
of trace options.  With an argument, omits menu step" ()
  (LET ((FCN (READ-FUNCTION-NAME "Trace" (RELEVANT-FUNCTION-NAME (POINT) NIL T T T) T)))
    (COND (*NUMERIC-ARG-P*
	   (SI:EVAL-ABORT-TRIVIAL-ERRORS
	     (IF (ATOM FCN)
		 `(TRACE (,FCN))
		 `(TRACE (:FUNCTION ,FCN)))))
	  ((FBOUNDP 'TV:TRACE-VIA-MENUS)
	   (TV:TRACE-VIA-MENUS FCN))
	  (T
	   (FORMAT T "~A will be traced. However, since trace menu support
is not currently loaded, using any of the trace options will
require executing the TRACE function from a Listener. To stop
the trace, execute the UNTRACE function from a Listener." FCN)
	   (SI:EVAL-ABORT-TRIVIAL-ERRORS (IF (ATOM FCN)
					     `(TRACE (,FCN))
					     `(TRACE (:FUNCTION ,FCN)))))))
  DIS-NONE) 


(DEFCOM COM-WHERE-IS-SYMBOL "Show which packages contain the specified symbol." ()
   (MULTIPLE-VALUE-BIND (SYMBOL NAME) (READ-FUNCTION-NAME "Where is symbol" NIL NIL T)
     (WHERE-IS (OR NAME SYMBOL)))
   DIS-NONE) 


(DEFCOM COM-COUNT-LINES-PAGE
   "Type number of lines on this page.
Also add, in parentheses, the number of lines on the page
before point, and the number of lines after point." (KM)
   (LET* ((POINT (POINT))
	  (PAGE-START (FORWARD-PAGE POINT -1 T))
	  (PAGE-END (FORWARD-PAGE PAGE-START 1 T)))
     (LET ((N1 (1- (COUNT-LINES PAGE-START POINT)))
	   (N2 (1- (COUNT-LINES POINT PAGE-END))))
       (FORMAT *QUERY-IO* "~&Page has ~D (~D + ~D) lines" (+ N1 N2) N1 N2)))
   DIS-NONE) 


(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST* FUNCTION-NAME "Disassemble" DO-DISASSEMBLE ()
   "Disassemble this function.") 


(DEFCOM COM-DISASSEMBLE "Disassemble the specified function." ()
   (DO-DISASSEMBLE (READ-FUNCTION-NAME "Disassemble" (RELEVANT-FUNCTION-NAME (POINT)) T))
   DIS-NONE) 


(DEFCOM COM-QUICK-DISASSEMBLE "Disassemble the function to the left of the cursor." ()
   (IF *NUMERIC-ARG-P*
       (DO-DISASSEMBLE (READ-FUNCTION-NAME "Disassemble" (RELEVANT-FUNCTION-NAME (POINT)) T))
       (DO-DISASSEMBLE (RELEVANT-FUNCTION-NAME (POINT))))
   DIS-NONE) 


(DEFUN DO-DISASSEMBLE (SYMBOL &AUX FSPEC)
  (COND ((NULL SYMBOL)
	 (BARF))
	((FDEFINEDP SYMBOL)
	 (SETQ FSPEC (FDEFINITION (SI:UNENCAPSULATE-FUNCTION-SPEC SYMBOL)))
	 (COND ((OR (TYPEP FSPEC 'COMPILED-FUNCTION)
		    (closurep fspec)		;; may 12-21-88
		    (AND (CONSP FSPEC)
			 (EQ (CAR FSPEC) 'MACRO)
			 (TYPEP (CDR FSPEC) 'COMPILED-FUNCTION)))
		(FORMAT T "~&~S:" SYMBOL)
		(DISASSEMBLE SYMBOL))
	       ((BARF "Can't find FEF for ~S" SYMBOL))))
	((BARF)))
  NIL) 


(DEFUN USER::TEACH-ZMACS ()
  "Run a Zmacs tutorial."
  (ED (FIND-FILE "sys:zwei;teach-zmacs.text"))) 



(DEFCOM COM-TEACH-ZMACS
   "Set up the Zmacs tutorial.
Makes private copies of the Zmacs tutorial files in the user's home directory
unless they're already there." ()
   (LOOP FOR FILE IN *TUTORIAL-FILES*
	 UNLESS (PROBE-FILE FILE)
	 DO (BARF "~&The Zmacs tutorial file ~A is not where it should be.~
                   ~&See your site administrator about this problem."
		  FILE)
	 AS USER-FILE = (MERGE-PATHNAMES (FS:USER-HOMEDIR) FILE)
	 UNLESS (PROBE-FILE USER-FILE)
	 DO (COPY-FILE FILE USER-FILE))
   (FORMAT *QUERY-IO* "~&The tutorial files have been copied to your directory on ~A."
	   SI:LOCAL-HOST)
   DIS-TEXT) 

