;;; -*- Mode:Common-Lisp; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1985-1989 Texas Instruments Incorporated. All rights reserved.

;;;
;;;  las 2/11/88 - Modified to load EH file as it has been evicted from our
;;;                cold bands and is only used by ZMACS
;;;

(UNLESS (FIND-PACKAGE "SRCCOM")
  (MAKE-PACKAGE "SRCCOM"))

(DEFSYSTEM ZWEI
  (:NAME "Zmacs Editor")
  (:SHORT-NAME "ZMACS")
  (:DOCUMENTATION "Editor for editing text or writing programs")
  (:DEFAULT-MENU-COLUMN :PROGRAMS)
  (:DEFAULT-SYSTEM-KEY #\E)
  (:PATHNAME-DEFAULT "SYS:ZMACS;")
  (:PATCHABLE "SYS:PATCH.ZMACS;" "PATCH")
  (:WARNINGS-PATHNAME-DEFAULT "SYS:CWARNS;ZMACS.LISP")
  (:INSTANCE-TYPE    :FLAVOR)
  (:INSTANCE-FINDER  ZWEI::ZMACS-FRAME)
  (:INSTANCE-CREATOR T)
  (:MODULE FONTS  ("SYS:FONTS;SEARCH") :PACKAGE "FONTS")
  (:MODULE EH     ("SYS:EH;INSERT-DELETE") :PACKAGE "EH")
  (:MODULE SETUP  ("SETUP"))	       ;Exporting, shadowing, suggestions, etc.
  (:MODULE DEFS   ("DEFS"	       ;Structure definitions and declarations.
	           "MACROS"))	       ;Lisp macros used in the ZWEI source.
  (:MODULE SCREEN ("SCREEN"))	       ;Interface to screen system.
  (:MODULE KBDMAC ("KBDMAC"))	       ;Keyboard macro stream.
  (:MODULE SRCCOM ("SRCCOM"))	       ;Source Compare two files or buffers. 
  (:MODULE MAIN   ("COMTAB"	       ;Functions regarding comtabs and command loop.
		   "DISPLA"	       ;Redisplay, and screen-related functions.
		   "FOR"	       ;Forward-this, forward-that functions.
		   "INDENT"	       ;Indention functions.
		   "INSERT"	       ;Insertion, deletion, and related functions.
		   "METH"	       ;Important methods for windows and buffers.
		   "PRIMIT"	       ;Random primitives and utilities.
		   "NPRIM"	       ;More recently written primitives.
		   "HISTORY"	       ;Kill history, minibuffer history, etc.
		   "FONT"	       ;Font hacking stuff.
		   "SEARCH"	       ;Searching functions.
		   
		   "COMA"	       ;Vanilla editing commands for movement, exchange, etc. 
		   "COMB"	       ;Editing commands for indentation and such.
		   "COMC"	       ;Commands for compiling and evaluating.
		   "COMD"	       ;Register, minibuffer, completion and apropos commands.
		   "COME"	       ;Includes quantity and tab commands.
		   "COMF"	       ;Includes pattern-matching, arglists, disassembly. 
		   "COMG"	       ;Commands for kbd-macro and describe-flavor operations.
		   "COMH"	       ;More assorted commands, featuring case-changing and commenting.
		   "COMS"	       ;Commands for searching and replacing.
		   
		   "DIRED"	       ;Directory editor.
		   "DOC"	       ;Self-documentation commands and functions.
		   "FASUPD"	       ;Update fasl file from core.
		   "FILES"	       ;File commands and utilities.
		   "HOST"	       ;Define ED:, ED-BUFFER:, ED-FILE: hosts.
		   "LPARSE"	       ;Parsing lisp code.
		   "MODES"	       ;Major and minor mode functions and commands.
		   "MOUSE"	       ;Mouse commands, sans screen interface.
		   "PATED"	       ;Patch commands.
		   "POSS"	       ;Visiting lists of possibilities.
		   "STREAM"	       ;Editor stream.
		   "SECTIO"	       ;Section-specific commands for ZMACS.
		   "ZMNEW"	       ;Buffer attribute-list commands.
		   "ZMACS"	       ;Multiple-buffer and file commands.
		   ))

  
  (:MODULE ZYMURG ("ZYMURG"))	       ;Combined methods.
  
  (:FASLOAD FONTS)
  (:FASLOAD EH)
  (:COMPILE-LOAD SETUP)
  (:COMPILE-LOAD DEFS   (:FASLOAD SETUP))
  (:COMPILE-LOAD SCREEN (:FASLOAD SETUP DEFS))
  (:COMPILE-LOAD KBDMAC (:FASLOAD SETUP DEFS SCREEN))
  (:COMPILE-LOAD SRCCOM (:FASLOAD SETUP DEFS))
  (:COMPILE-LOAD MAIN   (:FASLOAD SETUP DEFS SCREEN KBDMAC SRCCOM))
  (:COMPILE-LOAD ZYMURG (:FASLOAD SETUP DEFS SCREEN KBDMAC MAIN)))


;; Zmacs-dependent
(DEFSYSTEM Zmacs-Extras
  (:NAME "Zmacs Extras")
  (:PATHNAME-DEFAULT "SYS:ZMACS;")
  (:PATCHABLE "SYS:PATCH.ZMACS-Extras;" "PATCH")
  (:not-in-disk-label)
  (:WARNINGS-PATHNAME-DEFAULT "SYS:CWARNS;ZMACS-Extras.LISP")

  (:MODULE AUTOLOADING
	   ("LIBRARY-COMMANDS"	                  ;Tex Mode, Midas Mode.
	    "edit-disk-partitions"                ;meta-x Edit Disk Partitions
	    "SYS:PUBLIC.ZMACS;BDIRED"		  ;Directory Differences editor.
	    "SYS:PUBLIC.ZMACS;CHAR-SEARCH"	  ;Char Search and Reverse Char Search.
	    "SYS:PUBLIC.ZMACS;MISC-BROKEN"	  ;Miscellaneous broken commands.
	    "SYS:PUBLIC.ZMACS;MISC-UNSUPPORTED"	  ;Miscellaneous non-broken commands.
	    "SYS:PUBLIC.ZMACS;PL1MOD"		  ;PL/I mode commands.
	    "SYS:PUBLIC.ZMACS;UNSUPPORTED-MODES"  ;Teco, Bolio, Emacs.
	    ))
  
  (:MODULE TUTORIAL ("SYS:TI-EXAMPLES.ZMACS;GETTYSBURG-INSERT.TEXT"
		     "SYS:TI-EXAMPLES.ZMACS;C-CURVE.LISP"))
  
  (:AUXILIARY TUTORIAL)
  (:COMPILE AUTOLOADING)
  
  )

