;;; -*- Mode:Common-Lisp; Package:ZWEI; Fonts:(CPTFONT HL12B CPTFONT); Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;1;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.*

; Copyright (C) 1980, Massachusetts Institute of Technology

;;; This file contains utility functions for manipulating files, and various
;;; commands to do I/O to intervals.  It does not know about buffers and such,
;;; just intervals.
;;; 11-16-87 DAB Binded *merge-unix-type* around read-default-pathname.
;;; 12-01-87 DAB Removed *merge-unix-type*. It broke operations on file with unspecific types.


;;; Get a pathname from the user, return as a pathname actor.
(DEFUN READ-DEFAULTED-PATHNAME (PROMPT DEFAULTS
				&OPTIONAL SPECIAL-TYPE SPECIAL-VERSION (DIRECTION :READ)
				(MERGE-IN-SPECIAL-VERSION T) (MERGE-TYPE T)
				MINI-BUFFER-INITIAL-CONTENTS
				MINI-BUFFER-INITIAL-CHAR-POS)
  "Read and default a pathname, prompting with PROMPT, which should end with a colon.
DEFAULTS is a defaults-alist used as the defaults, and modified.
DIRECTION should be :READ or :WRITE; used for completion.
The remaining arguments are passed along to MAKE-DEFAULTED-PATHNAME (which see)
MINI-BUFFER-INITIAL-CONTENTS is a string to initialize the mini buffer with,
MINI-BUFFER-INITIAL-CHAR-POS an index in that string to start the cursor out at."
  (MULTIPLE-VALUE-BIND (STRING END-P)
      (READ-UNDEFAULTED-PATHNAME-STRING PROMPT
					DEFAULTS
					SPECIAL-TYPE SPECIAL-VERSION
					DIRECTION
					MERGE-IN-SPECIAL-VERSION
					MINI-BUFFER-INITIAL-CONTENTS
					MINI-BUFFER-INITIAL-CHAR-POS)
    (LET ((RESULT (MAKE-DEFAULTED-PATHNAME STRING
					   DEFAULTS
					   (OR SPECIAL-TYPE
					       (IF DEFAULTS
						   (SEND (FS:DEFAULT-PATHNAME DEFAULTS) :CANONICAL-TYPE)))
					   (OR SPECIAL-VERSION
					       (IF END-P
						   :NEWEST)
					       (IF DEFAULTS
						   (SEND (FS:DEFAULT-PATHNAME DEFAULTS) :VERSION)))
					   MERGE-IN-SPECIAL-VERSION
					   MERGE-TYPE)))
      (PROMPT-LINE "~A (~A)" PROMPT RESULT)
      RESULT))) 

(DEFUN READ-UNDEFAULTED-PATHNAME-STRING (PROMPT *READING-PATHNAME-DEFAULTS*
					 &OPTIONAL *READING-PATHNAME-SPECIAL-TYPE*
					 *READING-PATHNAME-SPECIAL-VERSION*
					 (*READING-PATHNAME-DIRECTION* :READ)
					 (MERGE-IN-SPECIAL-VERSION T)
					 MINI-BUFFER-INITIAL-CONTENTS
					 MINI-BUFFER-INITIAL-CHAR-POS
					 &AUX (*MINI-BUFFER-VALUE-HISTORY* *PATHNAME-ARGUMENT-HISTORY*))
  ;; MERGE-IN-SPECIAL-VERSION is for the case of wanting the default to have :OLDEST, but
  ;; not having pathnames typed in keeping to this.
  (IF (NOT MERGE-IN-SPECIAL-VERSION)
      (SETQ *READING-PATHNAME-SPECIAL-VERSION* NIL))	;Don't complete from this
  (MULTIPLE-VALUE-BIND (END-P NIL INTERVAL)	;Ok, now we can get the pathname.
      (EDIT-IN-MINI-BUFFER *PATHNAME-READING-COMTAB*
			   MINI-BUFFER-INITIAL-CONTENTS
			   MINI-BUFFER-INITIAL-CHAR-POS
			   (LIST (FORMAT NIL "~A (Default is ~A)"
					 PROMPT
					 (FS:DEFAULT-PATHNAME
					   *READING-PATHNAME-DEFAULTS*
					   NIL
					   *READING-PATHNAME-SPECIAL-TYPE*
					   *READING-PATHNAME-SPECIAL-VERSION*))
				 '(:RIGHT-FLUSH " (Completion)")))
    (LET ((STRING (STRING-INTERVAL INTERVAL)))
      (PUSH-ON-HISTORY STRING *PATHNAME-ARGUMENT-HISTORY*)
      (VALUES STRING END-P))))			;Return the correct pathname string

(DEFUN UNDO-ALL-MODES ()			;Undoes all the modes currently in effect.
  (DO ()					;It's like this in case the modes have hooks.
      ((NULL *MODE-NAME-LIST*))			;do it until they're all off.
    (TURN-OFF-MODE (CAR *MODE-NAME-LIST*)))) 

(DEFUN READ-DEFAULTED-AUX-PATHNAME (PROMPT &OPTIONAL SPECIAL-TYPE SPECIAL-VERSION (DIRECTION :READ))
  "Read and default a pathname, prompting with PROMPT, which should end with a colon.
SPECIAL-TYPE and SPECIAL-VERSION are passed along to MAKE-DEFAULTED-PATHNAME (which see).
DIRECTION should be :READ or :WRITE; used for completion.
*AUX-PATHNAME-DEFAULTS* are used as the defaults."
  (READ-DEFAULTED-PATHNAME PROMPT
			   *AUX-PATHNAME-DEFAULTS*
			   SPECIAL-TYPE
			   SPECIAL-VERSION
			   DIRECTION)) 

(DEFUN READ-TWO-DEFAULTED-PATHNAMES (PROMPT DEFAULTS &AUX FROM TO)
  "Read and return two pathnames using the minibuffer, defaulting to DEFAULTS.
Prompt with PROMPT, which should fit where a verb would fit."
  (SETQ FROM (READ-DEFAULTED-PATHNAME (FORMAT () "~A file:" PROMPT)
				      DEFAULTS)
	TO (READ-DEFAULTED-PATHNAME (FORMAT () "~A ~A to:" PROMPT FROM)
				    FROM NIL :NEWEST :WRITE))
  (PROMPT-LINE "~A ~A to ~A" PROMPT FROM TO)
  (VALUES FROM TO)) 

(DEFUN MAKE-DEFAULTED-PATHNAME (STRING DEFAULTS &OPTIONAL SPECIAL-TYPE SPECIAL-VERSION
				(MERGE-IN-SPECIAL-VERSION T)
				(MERGE-TYPE T))
  "Parse STRING into a pathname with defaults.  STRING is recorded for M-H-Y.
DEFAULTS is a defaults alist; the defaults are set to the specified pathname.
SPECIAL-TYPE is a filetype to use if STRING has no name or type.
If that is NIL, then FS:*NAME-SPECIFIED-DEFAULT-TYPE* is used
 unless MERGE-TYPE is NIL.
SPECIAL-VERSION is a version to use if STRING has no name or version.
If that is NIL, then :NEWEST is used, unless MERGE-TYPE is NIL.
If MERGE-IN-SPECIAL-VERSION is NIL, SPECIAL-VERSION is used
 only if STRING is the empty string."
  (IF (ZEROP (LENGTH STRING))
      ;; He didn't type anything, use the default.
      (FS:DEFAULT-PATHNAME DEFAULTS NIL SPECIAL-TYPE SPECIAL-VERSION)
      (PROGN
	(AND (NOT MERGE-IN-SPECIAL-VERSION)	;Was only for nullstring case
	     (SETQ SPECIAL-VERSION ()))
	(FS:MERGE-AND-SET-PATHNAME-DEFAULTS STRING
					    DEFAULTS
					    (OR SPECIAL-TYPE
						(AND MERGE-TYPE FS:*NAME-SPECIFIED-DEFAULT-TYPE*))
					    (OR SPECIAL-VERSION
						(AND MERGE-TYPE ':NEWEST)))))) 

(DEFUN EDITOR-FILE-NAME (FILE-NAME)
  "Given a string or pathname, return pathname and name for buffer.
The first value is the canonicalized pathname to put in the buffer;
the second value is a string to use as the buffer name."
  (AND (STRINGP FILE-NAME)
       (SETQ FILE-NAME (FS:MERGE-PATHNAME-DEFAULTS FILE-NAME *PATHNAME-DEFAULTS*)))
  (SETQ FILE-NAME (SEND FILE-NAME :TRANSLATED-PATHNAME))
  (VALUES FILE-NAME (SEND FILE-NAME :STRING-FOR-EDITOR))) 

;;; Special commands in the pathname mini-buffer

(DEFCOM COM-PATHNAME-COMPLETE "Try to complete the string so far as a pathname" ()
  (LET ((TEM (PATHNAME-COMPLETE)))
    (AND (NULL TEM) (BEEP)))
  DIS-TEXT) 

(DEFCOM COM-YANK-DEFAULT-PATHNAME "Insert the default pathname as text." ()
  (CONTEXT-CHECK '*READING-PATHNAME-DEFAULTS*)
  (INSERT-MOVING (POINT)
		 (SEND (FS:DEFAULT-PATHNAME *READING-PATHNAME-DEFAULTS*)
		       :STRING-FOR-PRINTING))
  DIS-TEXT) 

(DEFCOM COM-PATHNAME-COMPLETE-AND-EXIT-IF-UNIQUE
	"Try to complete the string so far as a pathname and return if unique" ()
  (LET ((TEM (PATHNAME-COMPLETE)))
    (COND ((NULL TEM)
	   (BEEP))
	  ((EQ TEM :OLD)
	   (COND ((WINDOW-READY-P *WINDOW*)
		  (MUST-REDISPLAY *WINDOW* DIS-TEXT)
		  (REDISPLAY *WINDOW* :NONE)))
	   (THROW 'RETURN-FROM-COMMAND-LOOP T))))
  DIS-TEXT) 

(DEFUN PATHNAME-COMPLETE (&AUX STRING VALUE)
  (SETQ STRING (STRING-APPEND (BP-LINE (POINT))))
  (MULTIPLE-VALUE-SETQ (STRING VALUE)
		       (FS:COMPLETE-PATHNAME *READING-PATHNAME-DEFAULTS* STRING *READING-PATHNAME-SPECIAL-TYPE*
					     *READING-PATHNAME-SPECIAL-VERSION* *READING-PATHNAME-DIRECTION*))
  (DELETE-INTERVAL *INTERVAL*)
  (INSERT-MOVING (POINT) STRING)
  VALUE) 

(DEFCOM COM-PATHNAME-LIST-COMPLETIONS "Show possible completions of string so far" ()
  (LET ((COMPLETION-LIST (PATHNAME-LIST-COMPLETIONS)))
    (IF (ERRORP COMPLETION-LIST)
	(BARF COMPLETION-LIST)
	(LET* ((POSS (LOOP FOR PATHNAME IN COMPLETION-LIST
			   COLLECT (CONS (SEND PATHNAME :STRING-FOR-PRINTING)
					 PATHNAME)))
	       (LEN (LENGTH POSS)))
	  (COND ((ZEROP LEN)
		 (FORMAT T "~&There are no possible completions of the text you have typed.~%"))
		((= LEN 1)
		 (FORMAT T "~&The only possible completion of the text you have typed is ")
		 (SEND *STANDARD-OUTPUT* :ITEM 'PATHNAME-COMPLETION (CAAR POSS)))
		((OR (< LEN 62)
		     (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		       (FQUERY NIL "There are ~D possibilities, do you really want to see them all? "
			       LEN)))
		 (FORMAT T "~&These are the possible completions of the text you have typed:~2%")
		 (SEND *STANDARD-OUTPUT* :ITEM-LIST 'PATHNAME-COMPLETION
		       (SORT (MAPCAR #'CAR POSS) #'STRING-LESSP))
		 (TERPRI *STANDARD-OUTPUT*))))))
  DIS-TEXT) 

(DEFUN PATHNAME-LIST-COMPLETIONS (&OPTIONAL (STRING (STRING-APPEND (BP-LINE (POINT)))))
  (FS::PATHNAME-COMPLETION-LIST *READING-PATHNAME-DEFAULTS*
				STRING
				*READING-PATHNAME-SPECIAL-TYPE*
				*READING-PATHNAME-SPECIAL-VERSION*
				:NOERROR)) 

(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST*
			  PATHNAME-COMPLETION
			  "Select"
			  SELECT-PATHNAME-COMPLETION
			  T
			  "Use this Pathname.") 

;;; Called if the user mouses one of the completions
(DEFUN SELECT-PATHNAME-COMPLETION (STRING)
  (DELETE-INTERVAL *INTERVAL*)
  (INSERT-MOVING (POINT) STRING)
  (MUST-REDISPLAY *WINDOW* DIS-TEXT)
  (REDISPLAY *WINDOW* :NONE)
  (THROW 'RETURN-FROM-COMMAND-LOOP T)) 


(DEFCOM COM-DOCUMENT-PATHNAME-READ "Help while getting a pathname" ()
   (FORMAT T "~%You are typing a pathname, in a minibuffer.~%
These special commands are available:

   Type ~A to complete the pathname as much as possible.
   ~A attempts completion and exits if the completion is unique.
   ~A exits without completion.
   ~A ~lists the possible completions of what you have typed.
The completions are mouse-sensitive;  click on one to choose it.~

   ~A inserts the defaults as text.
   ~A inserts the last pathname you typed, as text.

Here is what the pathname will be used for:

"
	   (KEY-FOR-COMMAND 'COM-PATHNAME-COMPLETE *COMTAB* NIL NIL #\ESCAPE)
	   (KEY-FOR-COMMAND 'COM-PATHNAME-COMPLETE-AND-EXIT-IF-UNIQUE *COMTAB* NIL NIL #\END)
	   (KEY-FOR-COMMAND 'COM-END-OF-MINIBUFFER *COMTAB* NIL NIL #\NEWLINE)
	   (KEY-FOR-COMMAND 'COM-PATHNAME-LIST-COMPLETIONS *COMTAB* NIL NIL #\c-?)
	   (KEY-FOR-COMMAND 'COM-YANK-DEFAULT-PATHNAME *COMTAB* NIL NIL #\c-\y)
	   (KEY-FOR-COMMAND 'COM-YANK-PREVIOUS-INPUT *COMTAB* NIL NIL #\m-\y))
   (AND *MINI-BUFFER-COMMAND-IN-PROGRESS*
	(COM-DOCUMENT-CONTAINING-COMMAND))
   DIS-NONE) 

;;; Various file-related commands on INTERVALs.

(DEFCOM COM-INSERT-FILE "Insert the contents of the specified file at point.
Reads a file name from the mini-buffer, and inserts the contents of that
file at point. Leaves mark at the end of inserted text, and point at the 
beginning, unless given an argument.  Acts like Yank (Control-Y) with respect
to the region." ()
  (POINT-PDL-PUSH (POINT) *WINDOW* NIL NIL)
  (MOVE-BP (MARK) (POINT))
  (SETQ *CURRENT-COMMAND-TYPE* :YANK)
  (LET ((PATHNAME (READ-DEFAULTED-AUX-PATHNAME "Insert file:")))
    (WITH-OPEN-FILE (STREAM PATHNAME :ERROR :RETRY)
      (LET* ((ATTRIBUTES (FS:EXTRACT-ATTRIBUTE-LIST STREAM))
	     (FONTS (GETF ATTRIBUTES :FONTS))
	     (START-POINT (COPY-BP (POINT))))
	(WITH-UNDO-SAVE ("Insert file" (POINT) (POINT) T)
	  (DBP START-POINT)
	  (MOVE-BP (POINT) (STREAM-INTO-BP STREAM
					   (POINT)
					   (OR (CDR FONTS)
					       (GETF ATTRIBUTES :DIAGRAM))))
	  (IBP START-POINT)
	  (DOLIST (BP (LINE-BP-LIST (BP-LINE (POINT))))
	    (WHEN (AND (BP-= (POINT) BP)
		       (NEQ BP (POINT))
		       (EQ (BP-STATUS BP) :NORMAL))
	      (MOVE-BP BP START-POINT)))
	  (FIXUP-FONTS-INTERVAL FONTS START-POINT (POINT)))))
    (MAYBE-DISPLAY-DIRECTORY :READ PATHNAME))
  (OR *NUMERIC-ARG-P*
      (SWAP-BPS (POINT) (MARK)))
  DIS-TEXT) 

(DEFCOM COM-INSERT-FILE-NO-FONTS "Insert the contents of the specified file at point.
Reads a file name from the mini-buffer, and inserts the contents of that
file at point. Leaves mark at the end of inserted text, and point at the 
beginning, unless given an argument.  Acts like Yank (Control-Y) with respect
to the region." ()
  (POINT-PDL-PUSH (POINT) *WINDOW* NIL NIL)
  (MOVE-BP (MARK) (POINT))
  (SETQ *CURRENT-COMMAND-TYPE* :YANK)
  (LET ((PATHNAME (READ-DEFAULTED-AUX-PATHNAME "Insert file:")))
    (WITH-OPEN-FILE (STREAM PATHNAME :ERROR :RETRY)
      (LET ((START-POINT (COPY-BP (POINT))))
	(WITH-UNDO-SAVE ("Insert file" (POINT) (POINT) T)
	  (DBP START-POINT)
	  (MOVE-BP (POINT) (STREAM-INTO-BP STREAM (POINT)))
	  (IBP START-POINT)
	  (DOLIST (BP (LINE-BP-LIST (BP-LINE (POINT))))
	    (WHEN (AND (BP-= (POINT) BP)
		       (NEQ BP (POINT))
		       (EQ (BP-STATUS BP) :NORMAL))
	      (MOVE-BP BP START-POINT))))))
    (MAYBE-DISPLAY-DIRECTORY :READ PATHNAME))
  (OR *NUMERIC-ARG-P*
      (SWAP-BPS (POINT) (MARK)))
  DIS-TEXT) 

(DEFCOM COM-WRITE-REGION "Write out the region to the specified file." ()
  (REGION (BP1 BP2)
    (LET ((PATHNAME (READ-DEFAULTED-AUX-PATHNAME "Write region to:"
						 NIL NIL :WRITE))
	  ;; Examine the buffer's current mode line.
	  ;; If the user has edited in a Fonts: property,
	  ;; save the font information, even if he has 
	  ;; failed to do Reparse Mode Line.
	  (PLIST (FS:FILE-EXTRACT-PROPERTY-LIST (INTERVAL-STREAM *INTERVAL*))))
      (WITH-OPEN-FILE (STREAM PATHNAME :DIRECTION :OUTPUT :ERROR :RETRY)
	(STREAM-OUT-INTERVAL STREAM BP1 BP2 T
			     (OR (GETL (LOCF PLIST) '(:FONTS :DIAGRAM))
				 (SEND *INTERVAL* :GET-ATTRIBUTE :FONTS)
				 (SEND *INTERVAL* :GET-ATTRIBUTE :DIAGRAM)))
	(CLOSE STREAM)
	(PRINT-FILE-WRITTEN STREAM))))
  DIS-NONE) 

(DEFUN PRINT-FILE-WRITTEN (STREAM &AUX (TRUENAME (SEND STREAM :TRUENAME)))
  (FORMAT *QUERY-IO* "~&Written: ~A" TRUENAME)
  (IF (SEND STREAM :OPERATION-HANDLED-P :READ-POINTER)
    (LET ((NCHARS (SEND STREAM :READ-POINTER)))
      (IF (< NCHARS 11610)
	(FORMAT *QUERY-IO* " -- ~D characters." NCHARS)
	(FORMAT *QUERY-IO* " -- ~DK characters." (ROUND NCHARS 2000)))))) 

(DEFCOM COM-APPEND-TO-FILE "Append region to the end of the specified file.
Creates the file if necessary." ()
  ;; 10/13/83 RAF - Reverse order of OPEN statements.
  (REGION (BP1 BP2)
    (LET ((PATHNAME (READ-DEFAULTED-AUX-PATHNAME "Append region to end of file:"
						 NIL NIL :READ))
	  (PLIST (FS:FILE-EXTRACT-PROPERTY-LIST (INTERVAL-STREAM *INTERVAL*))))
      ;; Open existing file to read in.
      (WITH-OPEN-FILE (ISTREAM PATHNAME :CHARACTERS T :DIRECTION :INPUT :ERROR NIL)
	(IF (ERRORP ISTREAM)
	    (IF (CONDITION-TYPEP ISTREAM 'FS:FILE-NOT-FOUND)
		(FORMAT *QUERY-IO* "~&(New File)")	;Tell user "New File"
		(BARF "Error: ~A" ISTREAM)))	;Bomb on any other error.
	;; Now, open a new version of the file for writing.
	(WITH-OPEN-FILE-RETRY (OSTREAM (PATHNAME FS:FILE-ERROR) :CHARACTERS T :DIRECTION :OUTPUT)
	  (UNLESS (ERRORP ISTREAM)		;Copy file if it exists.
	    (STREAM-COPY-UNTIL-EOF ISTREAM OSTREAM))
	  (STREAM-OUT-INTERVAL OSTREAM BP1 BP2 T
			       (OR (GETL (LOCF PLIST) '(:FONTS :DIAGRAM))
				   (SEND *INTERVAL* :GET-ATTRIBUTE :FONTS)
				   (SEND *INTERVAL* :GET-ATTRIBUTE :DIAGRAM)))
	  (CLOSE OSTREAM)
	  (PRINT-FILE-WRITTEN OSTREAM)))
      (MAYBE-DISPLAY-DIRECTORY :READ PATHNAME)))
  DIS-NONE) 

(DEFCOM COM-PREPEND-TO-FILE
	"Append region to the beginning of the specified file.
Creates the file if necessary." ()
  ;; 10/13/83 RAF - Reverse order of OPEN statements.
  (REGION (BP1 BP2)
    (LET ((PATHNAME (READ-DEFAULTED-AUX-PATHNAME "Append region to start of file:"
						 NIL NIL :READ))
	  (PLIST (FS:FILE-EXTRACT-PROPERTY-LIST (INTERVAL-STREAM *INTERVAL*))))
      ;; Open existing file to read in.
      (WITH-OPEN-FILE (ISTREAM PATHNAME :ERROR NIL)
	(IF (ERRORP ISTREAM)
	    (IF (CONDITION-TYPEP ISTREAM 'FS:FILE-NOT-FOUND)
		(FORMAT *QUERY-IO* "~&(New File)")	;Tell user "New File"
		(BARF "Error: ~A" ISTREAM)))	;Bomb on any other error.
	;; Now, open a new version of the file for writing.
	(WITH-OPEN-FILE-RETRY (OSTREAM (PATHNAME FS:FILE-ERROR) :CHARACTERS T :DIRECTION :OUTPUT)
	  (STREAM-OUT-INTERVAL OSTREAM BP1 BP2 T
			       (OR (GETL (LOCF PLIST) '(:FONTS :DIAGRAM))
				   (SEND *INTERVAL* :GET-ATTRIBUTE :FONTS)
				   (SEND *INTERVAL* :GET-ATTRIBUTE :DIAGRAM)))
	  (UNLESS (ERRORP ISTREAM)		;Copy file if it exists.
	    (STREAM-COPY-UNTIL-EOF ISTREAM OSTREAM))
	  (CLOSE OSTREAM)
	  (PRINT-FILE-WRITTEN OSTREAM)))
      (MAYBE-DISPLAY-DIRECTORY :READ PATHNAME)))
  DIS-NONE) 

(DEFCOM COM-VIEW-FILE "View contents of a file.
While viewing, you can use Space and Overstrike to scroll forward and backward.
Type Rubout to exit.  You cannot edit; but you can see the beginning
of the file without waiting for the whole file to be read in." ()
  (LET ((PATHNAME (READ-DEFAULTED-PATHNAME "View file:" (PATHNAME-DEFAULTS))))
    (VIEW-FILE PATHNAME))
  DIS-NONE) 

;;; Show the file in the "display window".
;;; The caller should set up a reasonable prompt.
;;; Macro for suggestions added by rpm 11-26-86.
(DEFUN VIEW-FILE (PATHNAME &OPTIONAL DELETED-P)
  "\"View\" the text of file PATHNAME.  DELETED-P means allow deleted files.
To scroll forward one screen: CTRL-, CTRL-V or SPACE
          backward one screen: CTRL- or META-V
          forward one line: CTRL-N or 
          backward one line: CTRL-P or .
A window overlying *WINDOW* is used."
  (SYS:WITH-SUGGESTIONS-MENUS-FOR ZWEI:VIEW-FILE
    (WITH-OPEN-FILE (STREAM PATHNAME :ERROR :RETRY :PRESERVE-DATES T :DELETED DELETED-P)
      (PROMPT-LINE "Viewing ~A" (SEND STREAM :TRUENAME))

1           ;; if stream is for NFS, then set the instance variable to () so that a view-file will not
           ;; call set-file-properties. This requires recompiling the nfs-stream flavor !! (view-file
           ;; does not change file properties -- Tung (6/27/88)*
      (when (SEND stream :operation-handled-p :client)
	(SEND stream :send-if-handles :set-preserve-dates-p nil) )

      (VIEW-STREAM STREAM)))) 

;; Modified to bind *interval* to the node created for the overlaying window. BJ 12-04-85.
(DEFUN VIEW-STREAM (STREAM &OPTIONAL (WINDOW (CREATE-OVERLYING-WINDOW *WINDOW*)))
  "\"View\" the text read from STREAM. 
To scroll forward one screen: CTRL-, CTRL-V or SPACE
          backward one screen: CTRL- or META-V
          forward one line: CTRL-N or 
          backward one line: CTRL-P or .
A window overlying *WINDOW* is used."
  (LET ((*INTERVAL* (MAKE-INSTANCE 'NODE)))
    (SEND (WINDOW-SHEET WINDOW) :SET-LABEL "")
    (SEND WINDOW :SET-INTERVAL *INTERVAL*)
    (SEND WINDOW :SET-SCROLL-BAR-MODE NIL)
    (TEMPORARY-WINDOW-SELECT (WINDOW)
      ;; Hack Alert!! Let scroll-bar process run so we can get it turned off.
      (PROCESS-ALLOW-SCHEDULE)
      (VIEW-WINDOW WINDOW STREAM)))) 

(DEFCOM COM-DELETE-FILE "Delete a file.
If wildcards are used, many files can be deleted." ()
  (LET ((PATHNAME (READ-DEFAULTED-PATHNAME "Delete file:" (PATHNAME-DEFAULTS))))
    (IF (SEND PATHNAME :WILD-P)
	(LET ((DIR (CDR (FS:DIRECTORY-LIST PATHNAME))))
	  (FORMAT T "~&Files to be deleted:~%")
	  (MAPC *DIRECTORY-SINGLE-FILE-LISTER* DIR)
	  (WHEN (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		  (Y-OR-N-P "Delete them? "))
	    (DOLIST (ELT DIR)
	      (CONDITION-CASE (ERROR)
		  (SEND (CAR ELT) :DELETE T)
		((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
		 (FORMAT T "~&Deletion failure: ~A" ERROR))))
	    (FORMAT T "~&Done.~%")))
	(CONDITION-CASE (VALUE)
	    (DELETE-FILE PATHNAME)
	  ((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
	   (BARF VALUE))
	  (:NO-ERROR
	   (FORMAT *QUERY-IO* "~&~A deleted." (CAAR VALUE))))))
  DIS-NONE) 

(DEFCOM COM-UNDELETE-FILE "Undelete a file.
If wildcards are used, many files can be undeleted." ()
  (LET ((PATHNAME (READ-DEFAULTED-PATHNAME "Undelete file:" (PATHNAME-DEFAULTS))))
    (IF (SEND PATHNAME :WILD-P)
	(LET ((DIR (CDR (FS:DIRECTORY-LIST PATHNAME :DELETED))))
	  (FORMAT T "~&Files to be undeleted:~%")
	  (MAPC *DIRECTORY-SINGLE-FILE-LISTER* DIR)
	  (WHEN (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		  (Y-OR-N-P "Undelete them? "))
	    (DOLIST (ELT DIR)
	      (CONDITION-CASE (ERROR)
		  (SEND (CAR ELT) :UNDELETE T)
		((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
		 (FORMAT T "~&Undeletion failure: ~A" ERROR))))
	    (FORMAT T "~&Done.~%")))
	(CONDITION-CASE (VALUE)
	    (UNDELETE-FILE PATHNAME)
	  ((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
	   (BARF VALUE))
	  (:NO-ERROR
	   (FORMAT *QUERY-IO* "~&~A undeleted." (CAAR VALUE))))))
  DIS-NONE) 

(DEFCOM COM-RENAME-FILE "Rename a file.
If wildcards are used, many files can be renamed." ()
  (LET* ((DEFAULT-PATHNAME (DEFAULT-PATHNAME))
	 (PATHNAME (READ-DEFAULTED-PATHNAME "Rename file:"
					    DEFAULT-PATHNAME NIL
					    (IF (EQ (SEND DEFAULT-PATHNAME :VERSION) :NEWEST)
						(IF (SYMBOLP (BUFFER-FILE-ID *INTERVAL*))
						    NIL
						    (SEND (CAR (BUFFER-FILE-ID *INTERVAL*)) :VERSION))
						NIL)))
	 (TO-SPEC (READ-UNDEFAULTED-PATHNAME-STRING
		    (FORMAT NIL "Rename file ~A to:" PATHNAME)
		    PATHNAME))
	 (TO-PATHNAME (MERGE-PATHNAMES TO-SPEC PATHNAME))
	 BUFFERS-CONSIDERED)
    (DECLARE (SPECIAL BUFFERS-CONSIDERED))
    (WHEN (NEQ (SEND (SEND PATHNAME :TRANSLATED-PATHNAME) :HOST)
	       (SEND (SEND TO-PATHNAME :TRANSLATED-PATHNAME) :HOST))
      (BARF "Can't rename across hosts."))
    (IF (SEND PATHNAME :WILD-P)
	(LET ((DIR (CDR (FS:DIRECTORY-LIST PATHNAME :SORTED))))
	  (FORMAT T "~&Files to be renamed:~%")
	  (MAPC *DIRECTORY-SINGLE-FILE-LISTER* DIR)
	  (WHEN (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		  (Y-OR-N-P "Rename them (to ~A)? "
			    (MERGE-PATHNAMES TO-SPEC PATHNAME (SEND PATHNAME :VERSION))))
	    (DOLIST (ELT DIR)
	      (CONDITION-CASE (ERROR)
		  (SEND (CAR ELT) :RENAME
			(SEND TO-PATHNAME :TARGET-TRANSLATE-WILD-PATHNAME PATHNAME (CAR ELT)) T)
		((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
		 (FORMAT T "~&Rename failure: ~A" ERROR))
		(:NO-ERROR
		 (RENAME-FILE-1 PATHNAME
				(SEND TO-PATHNAME :TARGET-TRANSLATE-WILD-PATHNAME PATHNAME (CAR ELT))))))
	    (FORMAT T "~&Done.~%")))
	(CONDITION-CASE (ERROR OLD-TRUENAME NEW-TRUENAME)
	    (RENAME-FILE PATHNAME TO-SPEC)
	  ((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
	   (BARF ERROR))
	  (:NO-ERROR
	   (FORMAT *QUERY-IO* "~&~A renamed to ~A." OLD-TRUENAME NEW-TRUENAME)
	   (RENAME-FILE-1 OLD-TRUENAME NEW-TRUENAME)))))
  DIS-NONE) 

(DEFUN RENAME-FILE-WITHIN-DIRECTORY (PATHNAME TO-SPEC)
  "This does the work of com-rename-file provided pathnames are within the same directory."
  (IF (SEND PATHNAME :WILD-P)
      (LET ((DIR (CDR (FS:DIRECTORY-LIST PATHNAME)))
	    (TO-PATHNAME (MERGE-PATHNAMES TO-SPEC PATHNAME)))
	(FORMAT T "~&Files to be renamed:~%")
	(MAPC *DIRECTORY-SINGLE-FILE-LISTER* DIR)
	(WHEN (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		(Y-OR-N-P "Rename them? "))
	  (DOLIST (ELT DIR)
	    (CONDITION-CASE (ERROR)
		(SEND (CAR ELT) :RENAME
		      (SEND TO-PATHNAME :TARGET-TRANSLATE-WILD-PATHNAME PATHNAME (CAR ELT)) T)
	      ((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
	       (FORMAT T "~&Rename failure: ~A" ERROR))
	      (:NO-ERROR
	       (RENAME-FILE-1 PATHNAME
			      (SEND TO-PATHNAME :TARGET-TRANSLATE-WILD-PATHNAME PATHNAME (CAR ELT))))))
	  (FORMAT T "~&Done.~%")))
      (CONDITION-CASE (ERROR OLD-TRUENAME NEW-TRUENAME)
	  (RENAME-FILE PATHNAME TO-SPEC)
	((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
	 (BARF ERROR))
	(:NO-ERROR
	 (FORMAT *QUERY-IO* "~&~A renamed~% to ~A." OLD-TRUENAME NEW-TRUENAME)
	 (RENAME-FILE-1 OLD-TRUENAME NEW-TRUENAME))))) 

(DEFUN RENAME-FILE-ACROSS-DIRECTORIES (PATHNAME TO-SPEC)
  ;;modified from copy-file-1 gsl.
  (LET* ((COPY-MODE :DEFAULT)
	 (COPY-CREATION-DATE? (< *NUMERIC-ARG* 4)))
    (IF (SEND PATHNAME :WILD-P)
	(LET ((DIR (CDR (FS:DIRECTORY-LIST PATHNAME)))
	      (TO-PATHNAME (MERGE-PATHNAMES TO-SPEC PATHNAME)))
	  (FORMAT T "~&Files to be renamed:~%")
	  (MAPC *DIRECTORY-SINGLE-FILE-LISTER* DIR)
	  (WHEN (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		  (Y-OR-N-P "Rename them? "))
	    (DOLIST (ELT DIR)
	      (CONDITION-CASE (ERROR)
		  (COPY-FILE (CAR ELT)
			     (SEND TO-PATHNAME :TARGET-TRANSLATE-WILD-PATHNAME PATHNAME (CAR ELT))
			     :COPY-CREATION-DATE COPY-CREATION-DATE?
			     :COPY-AUTHOR COPY-CREATION-DATE?
			     :CHARACTERS COPY-MODE
			     :REPORT-STREAM *STANDARD-OUTPUT*)
		(:NO-ERROR
		 (DELETE-AND-EXPUNGE (CAR ELT))
		 (RENAME-FILE-1 PATHNAME
				(SEND TO-PATHNAME :TARGET-TRANSLATE-WILD-PATHNAME PATHNAME (CAR ELT))))
		((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
		 (FORMAT T "~&Rename failure: ~A" ERROR))))
	    (FORMAT T "~&Done.~%")))
	(CONDITION-CASE (VALUE OLD-TRUENAME NEW-TRUENAME)
	    (COPY-FILE PATHNAME
		       TO-SPEC
		       :COPY-CREATION-DATE COPY-CREATION-DATE?
		       :COPY-AUTHOR COPY-CREATION-DATE?
		       :CHARACTERS COPY-MODE
		       :REPORT-STREAM *QUERY-IO*)
	  (:NO-ERROR
	   (DELETE-AND-EXPUNGE PATHNAME)
	   (RENAME-FILE-1 OLD-TRUENAME NEW-TRUENAME))
	  ((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
	   (BARF VALUE)))))) 

(DEFUN RENAME-FILE-1 (INPUT-PATHNAME OUTPUT-PATHNAME)
  (DECLARE (SPECIAL BUFFERS-CONSIDERED))
  ;; Offer to rename a buffer visiting this file, no specific version.
  ;; In order to avoid asking the same question for each file version renamed,
  ;; we record buffers that have been asked about and don't ask a second time.
  (LET ((BUF (FIND-FILE-BUFFER (SEND INPUT-PATHNAME :NEW-VERSION :NEWEST))))
    (WHEN (AND BUF (NOT (MEMBER BUF BUFFERS-CONSIDERED :TEST #'EQ)))
      (PUSH BUF BUFFERS-CONSIDERED)
      (IF (FQUERY NIL "~&Rename buffer ~A as well? " BUF)
	  (SET-BUFFER-PATHNAME (SEND OUTPUT-PATHNAME :NEW-VERSION :NEWEST)
			       BUF))))
  ;; Offer to rename a buffer visiting this version number specifically.
  (LET ((BUF (FIND-FILE-BUFFER INPUT-PATHNAME)))
    (WHEN (AND BUF (NOT (MEMBER BUF BUFFERS-CONSIDERED :TEST #'EQ)))
      (PUSH BUF BUFFERS-CONSIDERED)
      (WHEN (FQUERY NIL "~&Rename buffer ~A as well? " BUF)
	(SET-BUFFER-PATHNAME OUTPUT-PATHNAME BUF)
	(WHEN (CONSP (BUFFER-FILE-ID BUF))
	  (SETF (BUFFER-FILE-ID BUF)
		(CONS OUTPUT-PATHNAME (CDR (BUFFER-FILE-ID BUF))))))))) 

(DEFUN DELETE-AND-EXPUNGE (PATHNAME)
  "Deletes and expunges pathname in one operation if possible.
If a delete error the user is notified in *QUERY-IO*."
  (IF (SEND PATHNAME :OPERATION-HANDLED-P :DELETE-AND-EXPUNGE)
      (IF (ERRORP (SEND PATHNAME :DELETE-AND-EXPUNGE))
	  (FORMAT *QUERY-IO* "~&The copy to the new name succeeded, but ~a couldn't be deleted."
		  PATHNAME))
      (IF (ERRORP (SEND PATHNAME :DELETE NIL))
	  (FORMAT *QUERY-IO* "~&The copy to the new name succeeded, but ~a couldn't be deleted."
		  PATHNAME)
	  (SEND PATHNAME :EXPUNGE)))) 

(DEFCOM COM-COPY-BINARY-FILE
   "Copy one file to another, as a binary file.
With a numeric argument, don't copy creation date." ()
  (COPY-FILE-1 NIL *NUMERIC-ARG-P*)
  DIS-NONE) 

(DEFCOM COM-COPY-TEXT-FILE
   "Copy one file to another, as a text file.
With a numeric argument, don't copy creation date." ()
  (COPY-FILE-1 T *NUMERIC-ARG-P*)
  DIS-NONE) 

(DEFCOM COM-COPY-FILE "Copy one file to another.
If Arg=2, force copying in characters mode.
If Arg=3, force copying in binary mode.
If Arg=4, Don't copy creation date or author.
If Arg=5, Don't copy creation date or author, force characters.
If Arg=6, Don't copy creation date or author, force binary." ()
  (COPY-FILE-1 (CASE *NUMERIC-ARG*
		     ((2 5) 'T)
		     ((3 6) 'NIL)
		     (OTHERWISE :DEFAULT))
	       (< *NUMERIC-ARG* 4))
  DIS-NONE) 

(DEFUN COPY-FILE-1 (COPY-MODE COPY-CREATION-DATE?)
  (LET* ((FILE-TYPE-STRING (CASE COPY-MODE
				 ((T) "Copy text file")
				 ((NIL) "Copy binary file")
				 (OTHERWISE "Copy file")))
	 (DEFAULT-PATHNAME (DEFAULT-PATHNAME))
	 (PATHNAME (READ-DEFAULTED-PATHNAME (FORMAT NIL "~A:" FILE-TYPE-STRING)
					    DEFAULT-PATHNAME)) 
	 (TO-SPEC (READ-UNDEFAULTED-PATHNAME-STRING
		    (FORMAT NIL "Copy file ~A to:" PATHNAME)
		    PATHNAME))
	 )
    (IF (SEND PATHNAME :WILD-P)
	(LET ((DIR (CDR (FS:DIRECTORY-LIST PATHNAME)))
	      (TO-PATHNAME (MERGE-PATHNAMES TO-SPEC PATHNAME))
	      (COPY-FUNCTION #'COPY-FILE)
	      DIR-ENTRIES)
	  ;; See if there are any subdirectories in the list.
	  ;; All directory-handling stuff added by rpm (from dkm) on 11-25-86.
	  (DOLIST (ELT DIR)				    
	     (AND (GETF (CDR ELT) :DIRECTORY)
		  (PUSH ELT DIR-ENTRIES)))
	  ;; If so, ask if they want to copy them.
	  (WHEN DIR-ENTRIES
	    (COND ((LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		     (Y-OR-N-P "Copy Directories too? "))
		   (SETQ COPY-FUNCTION #'COPY-DIRECTORY))  
		  (T
		   ;;Remove subdirectories from dir list.
		   (DOLIST (ELT DIR-ENTRIES)		    
		      (SETQ DIR (DELETE ELT DIR)))
		   (SETQ DIR-ENTRIES NIL))))
	  (FORMAT T "~&Files ~:[~;and Directories ~]to be copied:~%" DIR-ENTRIES)
	  (MAPC *DIRECTORY-SINGLE-FILE-LISTER* DIR)
	  (WHEN (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
		  (Y-OR-N-P "Copy them? "))
	    (DOLIST (ELT DIR)
	      (CONDITION-CASE (ERROR)
		  (FUNCALL COPY-FUNCTION		   
			   (CAR ELT)
			   (SEND TO-PATHNAME :TARGET-TRANSLATE-WILD-PATHNAME PATHNAME (CAR ELT))
			   :COPY-CREATION-DATE COPY-CREATION-DATE?
			   :COPY-AUTHOR COPY-CREATION-DATE?
			   :CHARACTERS COPY-MODE
			   :REPORT-STREAM *STANDARD-OUTPUT*
			   :DIRECTORY-LIST ELT)	;Speeds it up a touch.
		((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
		 (FORMAT T "~&Copy failure: ~A" ERROR))))
	    (FORMAT T "~&Done.~%")))
	(CONDITION-CASE (VALUE)
	    (COPY-FILE PATHNAME
		       TO-SPEC
		       :COPY-CREATION-DATE COPY-CREATION-DATE?
		       :COPY-AUTHOR COPY-CREATION-DATE?
		       :CHARACTERS COPY-MODE
		       :REPORT-STREAM *QUERY-IO*)
	  ((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
	   (BARF VALUE)))))) 

(DEFCOM COM-PRINT-FILE "Print a file on the local hardcopy device." ()
  (IF (FBOUNDP 'PRINT-FILE)
      (LET ((PATHNAME (READ-DEFAULTED-PATHNAME "File to Print:" (PATHNAME-DEFAULTS)))
	    (*STANDARD-OUTPUT* *QUERY-IO*))
	(PRINT-FILE PATHNAME))
      (PROGN
	(BEEP)
	(FORMAT *QUERY-IO* "Unable to print a file - the printer functions are not loaded.")))
  DIS-NONE) 

;; Modified to support non-lisp (e.g. prolog) loading, by rpm on 12-16-86.
(DEFCOM COM-LOAD-FILE "Load a file.  With arg, compile if necessary." ()
  (IF *NUMERIC-ARG-P*
      (COM-COMPILE-AND-LOAD-FILE)
      (LET* ((PATHNAME (READ-DEFAULTED-PATHNAME "Load file:"
						(SEND (FS:DEFAULT-PATHNAME (PATHNAME-DEFAULTS))
						      :NEW-PATHNAME :TYPE NIL
						      :VERSION NIL)
						;; If user omits type, don't default it.
						NIL NIL :READ T NIL))
	     (BUFFER (FIND-FILE-BUFFER (SEND PATHNAME :SOURCE-PATHNAME)))
	     (LOAD-FUNCTION (OR (GET (SEND PATHNAME :CANONICAL-TYPE) 'LOAD-FILE-FUNCTION)
			       'LOAD)))
	(FORMAT *QUERY-IO* "~&~A loaded."
		(FUNCALL LOAD-FUNCTION PATHNAME
			 :PACKAGE (IF BUFFER (BUFFER-PACKAGE BUFFER))
			 :VERBOSE NIL))))
  DIS-NONE)

;; Modified to support non-lisp (e.g. prolog) compiling, by rpm on 12-16-86.
(DEFCOM COM-COMPILE-AND-LOAD-FILE "Load a file, after compiling it if necessary." ()
  (LET* ((PATHNAME (READ-DEFAULTED-PATHNAME "Compile and load file:"
					    (SEND (FS:DEFAULT-PATHNAME (PATHNAME-DEFAULTS))
						  :NEW-PATHNAME :TYPE NIL
						  :VERSION NIL)
					    ;; If user omits type, don't default it.
					    NIL NIL :READ T NIL))
	 (SOURCE-PATHNAME (SEND PATHNAME :SOURCE-PATHNAME))
	 (BUFFER (FIND-FILE-BUFFER SOURCE-PATHNAME))
	 (OUTPUT-FILE (EDITOR-COMPILE-FILE SOURCE-PATHNAME))
	 (LOAD-FUNCTION (OR (NULL OUTPUT-FILE)
			    (GET (SEND OUTPUT-FILE :CANONICAL-TYPE) 'LOAD-FILE-FUNCTION)
			    'LOAD)))
    (WHEN (NEQ 'T LOAD-FUNCTION)
      (FORMAT *QUERY-IO* "~&~A loaded."
	      (FUNCALL LOAD-FUNCTION PATHNAME
		       :PACKAGE (IF BUFFER (BUFFER-PACKAGE BUFFER))
		       :VERBOSE NIL))))
  DIS-NONE) 

(DEFCOM COM-SET-VERSION-LIMIT "Set version limits on files or directories.
The current version limit for the file or directory is displayed during the prompt.
This command only works properly for Explorer filesystems." ()
  (LET ((PATHNAME (READ-DEFAULTED-PATHNAME "Set version limit on filename: "
					   (DEFAULT-PATHNAME))))
    ;;fixed for release 3.0 to find Explorer host under any protocol - from patch 2.7 by drp.  
    (UNLESS (MEMBER (TYPE-OF PATHNAME) '(FS::LISPM-PATHNAME FS::REMOTE-LISPM-PATHNAME) :TEST #'EQ)
      (BARF "Set Version Limit only works properly on Explorer filesystems.")) 
    (LET ((LIMIT (TYPEIN-LINE-READ-WITH-DEFAULT
		   (FORMAT NIL "~S" (FS::LMFS-GET-FILE-VERSION-LIMIT PATHNAME))
		   "New version limit (any number < 2 gives unlimited versions)"
		   NIL)))
      (CONDITION-CASE (VALUE)
	  (FS::LMFS-SET-FILE-VERSION-LIMIT PATHNAME LIMIT)
	((FS:FILE-ERROR SYSTEM:REMOTE-NETWORK-ERROR)
	 (BARF VALUE))
	(:NO-ERROR
	 (FORMAT *QUERY-IO* "Version limit for ~A set to ~D." PATHNAME LIMIT)))))
  DIS-NONE) 


;;; Directory Listing stuff.

(DEFCOM COM-DISPLAY-DIRECTORY "Display current buffer's file's directory.
Use the directory listing function in the variable Directory Lister.
With an argument, accepts the name of a file to list." ()
  (LET ((PATHNAME (READ-DEFAULTED-WILD-PATHNAME "Display Directory:"
						(DEFAULT-PATHNAME)
						(NOT *NUMERIC-ARG-P*))))
    (FILE-RETRY-NEW-PATHNAME (PATHNAME FS:FILE-ERROR)
      (FUNCALL *DIRECTORY-LISTER* PATHNAME)))
  DIS-NONE) 

(DEFUN READ-DEFAULTED-WILD-PATHNAME (PROMPT &OPTIONAL (DEFAULT (DEFAULT-PATHNAME)) DONT-READ-P)
  "Read and return a pathname, defaulting type and version to :WILD.
DEFAULT is a pathname that provides defaults for the other things.
DONT-READ-P means just return a default, don't read anything."
  (SETQ DEFAULT (SEND DEFAULT :NEW-PATHNAME :TYPE :WILD :VERSION :WILD))
  (OR DONT-READ-P
      (SETQ DEFAULT (READ-DEFAULTED-PATHNAME PROMPT DEFAULT :WILD :WILD)))
  DEFAULT) 

(DEFUN MAYBE-DISPLAY-DIRECTORY (TYPE &OPTIONAL (PATHNAME (DEFAULT-PATHNAME)))
  "Do automatic directory display if the user wants it.
TYPE is :READ or :WRITE, saying what sort of file operation
has just been done.  PATHNAME specifies host, device, directory and name to list."
  (COND ((OR (AND (EQ TYPE :READ)
		  (MEMBER *AUTO-DIRECTORY-DISPLAY* '(:READ T) :TEST #'EQ))
	     (AND (EQ TYPE :WRITE)
		  (MEMBER *AUTO-DIRECTORY-DISPLAY* '(:WRITE T) :TEST #'EQ)))
	 (FUNCALL *DIRECTORY-LISTER* (SEND PATHNAME :NEW-PATHNAME :TYPE :WILD :VERSION :WILD))))) 

;;; This is the default directory listing routine
(DEFUN DEFAULT-DIRECTORY-LISTER (PATHNAME)
  "Print a directory listing of PATHNAME in the default manner.
Uses the value of *DIRECTORY-SINGLE-FILE-LISTER* on each element of the directory-list."
  (WITH-OPEN-STREAM (STREAM (FS:DIRECTORY-LIST-STREAM PATHNAME :SORTED :DELETED))
    (LET ((NILENTRY (SEND STREAM :ENTRY)))
      (IF (CAR NILENTRY)
	  (FERROR NIL "First entry returned by a directory-list stream is not for NIL"))
      ;; What directory did we actually read?
      (SETQ PATHNAME (OR (GET NILENTRY :PATHNAME) PATHNAME))
      (FORMAT T "~&~A~%" PATHNAME)
      (FUNCALL *DIRECTORY-SINGLE-FILE-LISTER* NILENTRY)
      (LOOP
	(LET ((ENTRY (SEND STREAM :ENTRY)))
	  (OR ENTRY (RETURN))
	  (FUNCALL *DIRECTORY-SINGLE-FILE-LISTER* ENTRY)))))
  (FORMAT T "Done.~%")) 

;Note that *DIRECTORY-SINGLE-FILE-LISTER* is expected to output lines.

(DEFRESOURCE STREAM-LIST-BUFFER-POOL (SIZE)	;1rla*
  :CONSTRUCTOR (MAKE-ARRAY SIZE :ELEMENT-TYPE 'STRING-CHAR :LEADER-LENGTH 1)
  :MATCHER (>= (ARRAY-TOTAL-SIZE OBJECT) SIZE))

;1;; 9/10/85 RLA - changed for FTP integration*
;1;; 11/13/85 RLA - additional changes for FTP*
;1;; 12/3/85 RLA - don't use width of selected window except for display to an interval stream*
;;; 1/17/86 RPM - reinstate mouse sensitivity and get width and char-width from *window*, not from W:selected-window
;;; 5/27/86 RPM - ensure proper spacing in MEDFNB after, e.g., a rename operation - from patch 2.9 by grj.
(DEFUN DEFAULT-LIST-ONE-FILE (FILE &OPTIONAL (STREAM *STANDARD-OUTPUT*)
			      (PRINT-MESSAGE :STRING-FOR-DIRED)	        ;rla (FTP server may call with :string-for-host)
			      (STREAM-WIDTH DEFAULT-STREAM-WIDTH)	;rla
			      &AUX PATHNAME)
  (declare (special *mac-namep*))
  (COND ((NULL (SETQ PATHNAME (CAR FILE)))
	 (COND ((GET FILE :DISK-SPACE-DESCRIPTION)
		(SEND STREAM :LINE-OUT (GET FILE :DISK-SPACE-DESCRIPTION)))
	       ((GET FILE :PHYSICAL-VOLUME-FREE-BLOCKS)
		(DO ((FREE (GET FILE :PHYSICAL-VOLUME-FREE-BLOCKS) (CDR FREE))
		     (FLAG T NIL))
		    ((NULL FREE) (WRITE-CHAR #\NEWLINE STREAM))
		  (FORMAT STREAM "~A #~A=~D" (IF FLAG "Free:" ",") (CAAR FREE) (CDAR FREE))))
	       (T
		(WRITE-CHAR #\NEWLINE STREAM))))
	;;rpm -- clause below modified on 1/17/86 and on 5/27/86 - see comments preceding this definition.
	((TYPEP STREAM 'INTERVAL-STREAM)   
	 (LET* ((NEW-STREAM-WIDTH (TRUNCATE (SEND *WINDOW* :WIDTH)	   ;don't use W:selected-window
					    (SEND *WINDOW* :CHAR-WIDTH)))  ;since it may be nil - rpm.
		(STRING (CREATE-LINE 'ART-STRING NEW-STREAM-WIDTH NIL)))
	   (DEFAULT-LIST-ONE-FILE FILE STRING PRINT-MESSAGE NEW-STREAM-WIDTH)
	   (SEND STREAM :LINE-OUT STRING)))
	;; rla -- quit using a completely duplicated body to write to any other type of stream.  The alignment is screwy
	;;       and having to duplicate every change is very error prone.  Just reuse a string and output it to the stream.
	;; rpm -- bring back mouse sensitivty with this STREAMP clause.
	;; grj -- add stream-width to the default-list-one-file call and use output-string
	;;       in the :ELSE instead of calling default-list-one-file again.
	((STREAMP STREAM)
	 (USING-RESOURCE (STREAM-LIST-BUFFER STREAM-LIST-BUFFER-POOL STREAM-WIDTH)	;grj - was default-stream-width.
	   (LET ((OUTPUT-STRING (DEFAULT-LIST-ONE-FILE
				  FILE STREAM-LIST-BUFFER PRINT-MESSAGE STREAM-WIDTH)))	;grj - stream-width added.
	     (COND ((SEND STREAM :OPERATION-HANDLED-P :ITEM)
		    (SEND STREAM :ITEM 'FILE PATHNAME "~A" OUTPUT-STRING)
		    (WRITE-CHAR #\NEWLINE STREAM))
		   (:ELSE
		    (SEND STREAM :LINE-OUT OUTPUT-STRING))))))	;grj - use output-string instead of another recursive call.
	((OR (NULL STREAM) (STRINGP STREAM))
	 (LET* ((STRING (OR STREAM (MAKE-ARRAY STREAM-WIDTH :ELEMENT-TYPE 'STRING-CHAR :LEADER-LENGTH 1)))
		(DIRED-ECHO (GET FILE :DIRED-ECHO))	;rla
		(SHOWN-DIRECTORY NIL))		        ;rla
	   (WHEN DIRED-ECHO			        ;rla - don't indent so far
	     (SETQ STREAM-WIDTH (MIN STREAM-WIDTH (+ 114 *MIN-FILE-WIDTH*))))    ;grj - for MEDFNB, etc.

	   (WHEN (OR *mac-namep*		;; TUNG
		     (typep pathname 'fs:mac-pathname))
	     (setq DIRED-ECHO ()))

	   (SETF (FILL-POINTER STRING) 0)
	   (ARRAY-INITIALIZE STRING #\SPACE 0 (ARRAY-TOTAL-SIZE STRING))
	   (VECTOR-PUSH (IF (GET FILE :DELETED) #\D #\SPACE) STRING)
	   (VECTOR-PUSH #\SPACE STRING)
	   (STRING-NCONC STRING (OR (GET FILE :PHYSICAL-VOLUME) ""))
	   (SETF (FILL-POINTER STRING) (1+ (MAX 5 (FILL-POINTER STRING))))
	   (STRING-NCONC STRING (SEND PATHNAME PRINT-MESSAGE))
	   (SETF (FILL-POINTER STRING)
		 (MAX (- STREAM-WIDTH 114) (1+ (FILL-POINTER STRING))))	         ; once 20. - pf
	   (LET ((LINK-TO (GET FILE :LINK-TO)))
	     (IF LINK-TO
		 (PROGN
		   (STRING-NCONC STRING "=> " LINK-TO " ")
		   (SETF (FILL-POINTER STRING)
			 (MAX (- STREAM-WIDTH 70) (FILL-POINTER STRING))))	 ; once 40. - pf
		 (PROGN
		   (LET ((LENGTH (GET FILE :LENGTH-IN-BLOCKS)))
		     (SETF (FILL-POINTER STRING)
			   (MAX (- STREAM-WIDTH 111) (FILL-POINTER STRING)))	 ; once 23. - pf
		     (COND ((NULL LENGTH)
			    (STRING-NCONC STRING "     "))
			   ((> LENGTH 1747)
			    (SETF (FILL-POINTER STRING)
				  (NUMBER-INTO-ARRAY STRING LENGTH 12 (FILL-POINTER STRING) 4))
			    (VECTOR-PUSH #\SPACE STRING))
			   (T
			    (SETF (FILL-POINTER STRING)
				  (MAX (- STREAM-WIDTH 110) (FILL-POINTER STRING)))	; once 24. - pf
			    (SETF (FILL-POINTER STRING)
				  (NUMBER-INTO-ARRAY STRING LENGTH 12 (FILL-POINTER STRING) 3))
			    (VECTOR-PUSH #\SPACE STRING))))
		   (LET ((LENGTH (GET FILE :LENGTH-IN-BYTES)))
		     ;; RLA - If length-in-bytes is known, but not :length-in-blocks, show length here instead of "DIRECTORY"
		     ;; (show that at the end of the line -- see below)
		     (IF (AND (GET FILE :DIRECTORY)
			      (OR (GET FILE :LENGTH-IN-BLOCKS) (NOT LENGTH)))
			 (SETQ SHOWN-DIRECTORY (STRING-NCONC STRING "  DIRECTORY"))
			 (WHEN LENGTH
			   (SETF (FILL-POINTER STRING)
				 (NUMBER-INTO-ARRAY STRING LENGTH 12 (FILL-POINTER STRING) 6))
			   ;; rla - don't print NIL or :DEFAULT byte-size
			   (WHEN (INTEGERP (SETQ LENGTH (GET FILE :BYTE-SIZE)))
			     (VECTOR-PUSH #\( STRING)
			     (SETF (FILL-POINTER STRING)
				   (NUMBER-INTO-ARRAY STRING LENGTH 12 (FILL-POINTER STRING)))
			     (VECTOR-PUSH #\) STRING)))))
		   (SETF (FILL-POINTER STRING)
			 (MAX (- STREAM-WIDTH 71) (FILL-POINTER STRING)))	 ; once 39. - pf
		   (VECTOR-PUSH #\SPACE STRING)
		   (VECTOR-PUSH (COND ((GET FILE :OFFLINE) #\O)
				      ((GET FILE :NOT-BACKED-UP) #\!)
				      (T #\SPACE))
				STRING))))       
	   (VECTOR-PUSH (IF (GET FILE :DONT-DELETE) #\@ #\SPACE) STRING)       
	   (VECTOR-PUSH (IF (GET FILE :DONT-SUPERSEDE) #\# #\SPACE) STRING)       
	   (VECTOR-PUSH (IF (GET FILE :DONT-REAP) #\$ #\SPACE) STRING)
	   (VECTOR-PUSH #\SPACE STRING)		;rla
	   (TIME-INTO-ARRAY STRING (GET FILE :CREATION-DATE))
	   (LET* ((DATE-LAST-EXPUNGE (OR (GET FILE :DATE-LAST-EXPUNGED)
					 (GET FILE :DATE-LAST-EXPUNGE)))
		  (REFERENCE-DATE (OR DATE-LAST-EXPUNGE
				      (GET FILE :REFERENCE-DATE))))
	     (WHEN REFERENCE-DATE
	       (STRING-NCONC STRING (IF DATE-LAST-EXPUNGE " X=" " ("))
	       (TIME-INTO-ARRAY STRING REFERENCE-DATE NIL)
	       (OR DATE-LAST-EXPUNGE (STRING-NCONC STRING ")"))))
	   (LET ((AUTHOR (GET FILE :AUTHOR))
		 (DIRECTORY (SEND PATHNAME :DIRECTORY)))
	     (WHEN (AND AUTHOR (NOT (IF (CONSP DIRECTORY)
					(STRING-EQUAL AUTHOR (CAR (LAST DIRECTORY)))
					(STRING-EQUAL AUTHOR DIRECTORY))))
	       (SETF (FILL-POINTER STRING)
		     (MAX (- STREAM-WIDTH 26) (FILL-POINTER STRING)))	         ; once 74. - pf
	       (VECTOR-PUSH #\SPACE STRING)
	       (STRING-NCONC STRING AUTHOR)))
	   (LET ((READER (GET FILE :READER))
		 (DIRECTORY (SEND PATHNAME :DIRECTORY)))
	     (WHEN (AND READER (NOT (IF (CONSP DIRECTORY)
					(STRING-EQUAL READER (CAR (LAST DIRECTORY)))
					(STRING-EQUAL READER DIRECTORY))))
	       (SETF (FILL-POINTER STRING)
		     (MAX (- STREAM-WIDTH 14) (FILL-POINTER STRING)))	         ; once 84. - pf
	       (STRING-NCONC STRING READER)))
	   ;; RLA - if length is known for a directory go ahead and show length in bytes instead of DIRECTORY (see above)
	   ;; and to preserve alignment, show DIRECTORY here instead
	   (IF (AND (GET FILE :DIRECTORY) (NOT SHOWN-DIRECTORY))
	       (STRING-NCONC STRING " DIRECTORY")
	       (STRING-NCONC STRING "          "))
	   (WHEN DIRED-ECHO
	     (VECTOR-PUSH #\SPACE STRING)
	     (STRING-NCONC STRING DIRED-ECHO))	;rla
	   ;; get rid of trailing blanks (from missing properties) - rla
	   (LOOP FOR I = (FILL-POINTER STRING)
		 THEN (1- I)
		 WHILE (OR (> I (ARRAY-TOTAL-SIZE STRING))
			   (EQUAL (AREF STRING (1- I)) #\SPACE))
		 DO (SETF (FILL-POINTER STRING) I))
	   STRING))))
 

(DEFSUBST 2-DIGIT-INTO-ARRAY (ARRAY NUMBER)
  (VECTOR-PUSH-EXTEND (CODE-CHAR (+ 60 (FLOOR NUMBER 12))) ARRAY)	;convert tens digit to character
  (VECTOR-PUSH-EXTEND (CODE-CHAR (+ 60 (REM NUMBER 12))) ARRAY))	;convert units digit to character

(DEFUN TIME-INTO-ARRAY (ARRAY TIME &OPTIONAL (INCLUDE-TIME T))
  (IF TIME
      (MULTIPLE-VALUE-BIND (SECONDS MINUTES HOURS DAY MONTH YEAR)
	  (DECODE-UNIVERSAL-TIME TIME)
	(2-DIGIT-INTO-ARRAY ARRAY MONTH)
	(VECTOR-PUSH-EXTEND #\/ ARRAY)
	(2-DIGIT-INTO-ARRAY ARRAY DAY)
	(VECTOR-PUSH-EXTEND #\/ ARRAY)
	(2-DIGIT-INTO-ARRAY ARRAY (MOD YEAR 144))
	(WHEN INCLUDE-TIME
	  (VECTOR-PUSH-EXTEND #\SPACE ARRAY)
	  (2-DIGIT-INTO-ARRAY ARRAY HOURS)
	  (VECTOR-PUSH-EXTEND #\: ARRAY)
	  (2-DIGIT-INTO-ARRAY ARRAY MINUTES)
	  (VECTOR-PUSH-EXTEND #\: ARRAY)
	  (2-DIGIT-INTO-ARRAY ARRAY SECONDS)))
      (STRING-NCONC ARRAY (IF INCLUDE-TIME
			      "                 "
			      "        ")))) 

(DEFUN READ-DIRECTORY-NAME (PROMPT PATHNAME &OPTIONAL (WILDP :WILD)
			    &AUX TYPEIN PROMPT-WITH-DEFAULT)
  "Read a pathname to pass to FS:DIRECTORY-LIST.
Prompt with PROMPT, a string probably ending in a colon.
PATHNAME gives the defaults for host, device, directory.
WILDP gives the default used for the other components;
 normally :WILD, but could be NIL."
  (SETQ PATHNAME (SEND PATHNAME :NEW-PATHNAME :NAME WILDP
		                              :TYPE WILDP
		                              :VERSION WILDP)
	PROMPT-WITH-DEFAULT (FORMAT NIL "~A (Default is ~A)" PROMPT PATHNAME))
  (LET ((*READING-PATHNAME-DEFAULTS* PATHNAME)
	(*READING-PATHNAME-SPECIAL-TYPE* :WILD)
	(*READING-PATHNAME-SPECIAL-VERSION* :WILD)
	(*READING-PATHNAME-DIRECTION* :READ)
	(*MINI-BUFFER-VALUE-HISTORY* *PATHNAME-ARGUMENT-HISTORY*))
    (MULTIPLE-VALUE-BIND (NIL NIL INTERVAL)
	(EDIT-IN-MINI-BUFFER *PATHNAME-READING-COMTAB* NIL NIL
			     (LIST PROMPT-WITH-DEFAULT '(:RIGHT-FLUSH " (Completion)")))
      (SETQ TYPEIN (STRING-INTERVAL INTERVAL))))
  (COND ((EQUAL TYPEIN "")
	 (PUSH-ON-HISTORY PATHNAME *PATHNAME-ARGUMENT-HISTORY*)
	 (PROMPT-LINE "~A ~A" PROMPT PATHNAME)
	 PATHNAME)
	(T
	 (LET ((PATHNAME (FS:MERGE-PATHNAME-DEFAULTS TYPEIN PATHNAME :WILD :WILD)))
	   (PUSH-ON-HISTORY PATHNAME *PATHNAME-ARGUMENT-HISTORY*)
	   (PROMPT-LINE "~A ~A" PROMPT PATHNAME)
	   PATHNAME)))) 

(DEFUN READ-UNDEFAULTED-DIRECTORY-STRING (PROMPT PATHNAME &OPTIONAL (WILDP :WILD))
  "Read a string specifying a pathname to pass to FS:DIRECTORY-LIST.
Prompt with PROMPT, a string probably ending in a colon.
PATHNAME gives the defaults for host, device, directory.
WILDP gives the default used for the other components;
 normally :WILD, but could be NIL.
These defaults are used only for completion."
  (SETQ PATHNAME (SEND PATHNAME :NEW-PATHNAME :NAME WILDP
		                              :TYPE WILDP
		                              :VERSION WILDP)
	PROMPT (FORMAT () "~A (Default is ~A)" PROMPT PATHNAME))
  (LET ((*READING-PATHNAME-DEFAULTS* PATHNAME)
	(*READING-PATHNAME-SPECIAL-TYPE* :WILD)
	(*READING-PATHNAME-SPECIAL-VERSION* :WILD)
	(*READING-PATHNAME-DIRECTION* :READ)
	(*MINI-BUFFER-VALUE-HISTORY* *PATHNAME-ARGUMENT-HISTORY*))
    (MULTIPLE-VALUE-BIND (NIL NIL INTERVAL)
      (EDIT-IN-MINI-BUFFER *PATHNAME-READING-COMTAB* NIL NIL
			   (LIST PROMPT '(:RIGHT-FLUSH " (Completion)")))
      (LET ((STRING (STRING-INTERVAL INTERVAL)))
	(PUSH-ON-HISTORY STRING *PATHNAME-ARGUMENT-HISTORY*)
	STRING)))) 

(DEFCOM COM-LIST-FILES "Brief directory listing.
Lists several files per line" ()
  (LET ((PATHNAME (READ-DIRECTORY-NAME "List Files:" (DEFAULT-PATHNAME))))
    (FORMAT T "~&~A~%" PATHNAME)
    (LET ((LIST (FS:DIRECTORY-LIST PATHNAME :FAST)))
      (SETQ LIST (DELETE (ASSOC NIL LIST :TEST #'EQ)
			 (THE LIST LIST) :TEST #'EQ))		;Don't care about system info
      (DO ((L LIST (CDR L)))
	  ((NULL L) NIL)
	(SETF (CAR L) (CONS (SEND (CAAR L) :STRING-FOR-DIRED) (CAAR L))))
      (SEND *STANDARD-OUTPUT* :ITEM-LIST 'FILE LIST)))
  DIS-NONE) 

(DEFUN VIEW-DIRECTORY (VIEWED-DIRECTORY)
  "\"View\" the text of the directory listing of VIEWED-DIRECTORY.
To scroll forward one screen: CTRL-, CTRL-V or SPACE
          backward one screen: CTRL- or META-V
          forward one line: CTRL-N or 
          backward one line: CTRL-P or ."
  (SETQ VIEWED-DIRECTORY (FS:MERGE-PATHNAME-DEFAULTS VIEWED-DIRECTORY *PATHNAME-DEFAULTS*))
  (FILE-RETRY-NEW-PATHNAME (VIEWED-DIRECTORY FS:FILE-ERROR)
     (WITH-OPEN-STREAM (STREAM (DIRECTORY-INPUT-STREAM VIEWED-DIRECTORY))
       (SETQ VIEWED-DIRECTORY (SEND STREAM :DIRECTORY-PATHNAME))
       (PROMPT-LINE "Viewing directory ~A" VIEWED-DIRECTORY)
       (VIEW-STREAM STREAM)))
  DIS-NONE) 

(DEFUN DIRECTORY-INPUT-STREAM (DIRECTORY &AUX (DIR DIRECTORY))
  "Return a stream that reads a directory listing of pathname DIRECTORY.
The stream supports only the :LINE-IN operation."
  (LET-CLOSED ((DIRECTORY DIR)
	       (DIRECTORY-LIST-STREAM NIL)
	       (REREAD-ENTRY))
    (COND ((ERRORP (SETQ DIRECTORY-LIST-STREAM
			 (FS:DIRECTORY-LIST-STREAM DIRECTORY :DELETED :SORTED)))
	   (BARF "Error: ~A" DIRECTORY-LIST-STREAM)))
    (SETQ REREAD-ENTRY (SEND DIRECTORY-LIST-STREAM :ENTRY))
    (AND REREAD-ENTRY
	 (NULL (CAR REREAD-ENTRY))
	 (SETQ DIRECTORY (GET REREAD-ENTRY :PATHNAME)))
    #'DIRECTORY-INPUT-STREAM-IO)) 

;REREAD-ENTRY, if non-NIL, is an entry from the directory-list-stream
;that was peeked-ahead at.  The analogue of an UNTYI'd character.
(DEFUN DIRECTORY-INPUT-STREAM-IO (OP &REST ARGS)
  ;;this was implemented previously as a defselect - rpm 7/18/86
  (DECLARE (SPECIAL DIRECTORY-LIST-STREAM DIRECTORY REREAD-ENTRY))
  (CASE OP
    (:DIRECTORY-PATHNAME			;ARGS = ()
     DIRECTORY)
    (:CHARACTERS				;ARGS = ()
     T)
    (:LINE-IN					;ARGS = (&OPTIONAL (LEADER 1) &AUX STRING TEM)
     (LET ((LEADER (CAR ARGS))
	   STRING TEM)
       (COND (DIRECTORY-LIST-STREAM
	      (COND ((OR DIRECTORY
			 (SETQ TEM (OR (PROG1 REREAD-ENTRY (SETQ REREAD-ENTRY NIL))
				       (SEND DIRECTORY-LIST-STREAM :ENTRY))))
		     (SETQ STRING (MAKE-ARRAY 120
					      :ELEMENT-TYPE 'STRING-CHAR
					      :LEADER-LENGTH (IF (NUMBERP LEADER) LEADER 1)
					      :LEADER-LIST '(0)))
		     (WITH-OUTPUT-TO-STRING (S STRING)
		       (IF TEM
			   (FUNCALL *DIRECTORY-SINGLE-FILE-LISTER* TEM S)
			   (PROGN
			     (WRITE-STRING (SEND DIRECTORY :STRING-FOR-PRINTING) S)
			     (SETQ DIRECTORY NIL))))
		     ;; Chop off any CRs
		     (COND ((CHAR= (AREF STRING (SETQ TEM (1- (ARRAY-LEADER STRING 0))))
				   #\NEWLINE)
			    (SETF (ARRAY-LEADER STRING 0) TEM)))
		     STRING)
		    (T (SEND DIRECTORY-LIST-STREAM :CLOSE)
		       (SETQ DIRECTORY-LIST-STREAM NIL)
		       (VALUES NIL T))))
	     (T (VALUES NIL T)))))
    (:CLOSE					;ARGS = (&OPTIONAL MODE)
     (LET ((MODE (CAR ARGS)))
       (COND ((AND DIRECTORY-LIST-STREAM
		   (NOT (TYPEP DIRECTORY-LIST-STREAM 'PATHNAME)))
	      (SEND DIRECTORY-LIST-STREAM :CLOSE MODE)
	      (SETQ DIRECTORY-LIST-STREAM NIL)
	      T))))
    (:WHICH-OPERATIONS
     '(:DIRECTORY-PATHNAME :CHARACTERS :LINE-IN :CLOSE))
    (:OPERATION-HANDLED-P
     (AND (MEMBER (CAR ARGS) '(:DIRECTORY-PATHNAME :CHARACTERS :LINE-IN :CLOSE) :TEST #'EQ)
	  T))
    (:SEND-IF-HANDLES
     (IF (MEMBER (CAR ARGS) '(:DIRECTORY-PATHNAME :CHARACTERS :LINE-IN :CLOSE) :TEST #'EQ)
	 (DIRECTORY-INPUT-STREAM-IO (CAR ARGS) (CADR ARGS)))))) 
       
       
       
(DEFUN SUBSET-DIRECTORY-LISTING (PATHNAME)
  (LET ((FN1 (SEND PATHNAME :NAME))
	(FN2 (SEND PATHNAME :FN2)))
    (FORMAT T "~&~A~%" PATHNAME)
    (LET ((LINE NIL)
	  (FREE-ARRAY (MAKE-ARRAY 10))
	  (USED-ARRAY (MAKE-ARRAY 10)))
      (WITH-OPEN-FILE (STREAM (SEND PATHNAME :NEW-STRUCTURED-NAME '(".FILE." "(DIR)"))
			      :CHARACTERS T :DIRECTION :INPUT)
	;; First find out how much space is free.
	(SETQ LINE (SEND STREAM :LINE-IN))
	(SETQ LINE (SEND STREAM :LINE-IN))
	(DIRECTORY-FREE-SPACE LINE FREE-ARRAY)
	;; Make any pack that exists show up in the "used" display even if used=0
	(DOTIMES (IDX 10)
	  (AND (AREF FREE-ARRAY IDX)
	       (SETF (AREF USED-ARRAY IDX) 0)))
	;; Next, go through lines of dir, counting USED and printing some lines.
	(DO ((KEY (STRING-APPEND " "
				 (IF (STRING-EQUAL FN1 "TS") FN2 FN1)
				 " "))
	     (LINE) (EOF))
	    (NIL)
	  (MULTIPLE-VALUE-SETQ (LINE EOF)
			       (SEND STREAM :LINE-IN))
	  (AND (OR EOF (ZEROP (LENGTH LINE)))
	       (RETURN NIL))
	  (AND (LISP:SEARCH (STRING KEY) (STRING LINE) :TEST #'CHAR-EQUAL)
	       (WRITE-LINE LINE))
	  (OR (CHAR= (AREF LINE 2) #\L)
	      (LET ((USED (PARSE-NUMBER LINE 24))
		    (PACK (PARSE-NUMBER LINE 2)))
		(LET ((IDX (IF (OR (< PACK 12) (> PACK 20))
			       0
			       (- PACK 11))))
		  (SETF (AREF USED-ARRAY IDX)
			(+ (OR (AREF USED-ARRAY IDX) 0) USED))))))
	(FORMAT-DISK-BLOCKS-ARRAY T "Free: " FREE-ARRAY)
	(FORMAT-DISK-BLOCKS-ARRAY T ", Used: " USED-ARRAY))))) 
       
;Element 0 of FREE-ARRAY is for packs other than 10.-16.
(DEFUN DIRECTORY-FREE-SPACE (LINE FREE-ARRAY)
  (DO ((I (POSITION #\# (THE STRING (STRING LINE)) :TEST #'CHAR-EQUAL)
	  (POSITION #\# (THE STRING (STRING LINE)) :START I :TEST #'CHAR-EQUAL))
       (NUM) (IDX) (BLKS))
      ((NULL I))
    (MULTIPLE-VALUE-SETQ (NUM I)
      (PARSE-NUMBER LINE (1+ I)))
    (MULTIPLE-VALUE-SETQ (BLKS I)
      (PARSE-NUMBER LINE (1+ I)))
    (SETQ IDX (IF (OR (< NUM 12) (> NUM 20))
		  0
		  (- NUM 11)))
    (SETF (AREF FREE-ARRAY IDX)
	  (+ (OR (AREF FREE-ARRAY IDX) 0) BLKS)))) 

(DEFUN FORMAT-DISK-BLOCKS-ARRAY (STREAM TITLE ARRAY)
  (FORMAT STREAM TITLE)
  (DO ((IDX 0 (1+ IDX))
       (LIM (ARRAY-TOTAL-SIZE ARRAY))
       (FIRSTP T)
       (BLKS))
      ((= IDX LIM))
    (COND ((SETQ BLKS (AREF ARRAY IDX))
	   (FORMAT STREAM "~:[+~]~D" FIRSTP BLKS)
	   (SETQ FIRSTP ()))))) 


(DEFUN ROTATED-DIRECTORY-LISTING (PATHNAME)
  (CATCH 'ABORT
    (LET ((DEV (SEND PATHNAME :DEVICE))
	  (DIR (SEND PATHNAME :DIRECTORY))
	  (FN1 (SEND PATHNAME :NAME))
	  (FN NIL))
      (SETQ FN (SEND PATHNAME :NEW-STRUCTURED-NAME '(".FILE." "(DIR)")))
      (PROMPT-LINE "Directory Listing")
      (FORMAT T "~&~A  ~A    --   ~A~%" DEV DIR PATHNAME)
      (LET ((LINE NIL)
	    (X 0)
	    (Y 0))
	(WITH-OPEN-FILE (STREAM FN :CHARACTERS T :DIRECTION :INPUT)
	  (SETQ LINE (SEND STREAM :LINE-IN))
	  (FORMAT T "~A~%" (SEND STREAM :LINE-IN))
	  (DO ((LINE (SETQ LINE (SEND STREAM :LINE-IN)) (SETQ LINE (SEND STREAM :LINE-IN)))
	       (LFN1 (LENGTH FN1))
	       (LFN16 (+ (LENGTH FN1) 6)))
	      ((NULL LINE)
	       (FORMAT T "There is no file named ~A in the directory.~%" FN1))
	    (COND ((STRING-EQUAL LINE FN1 :START1 6 :START2 0 :END1 LFN16 :END2 LFN1)
		   ;; Found one.
		   (LET ((FIRST LINE))
		     (SETQ LINE
			   (DO ((LINE LINE (SEND STREAM :LINE-IN)))
			       ((OR (CHAR= (AREF LINE 0) #\PAGE)
				    (NOT (STRING-EQUAL LINE FN1
						       :START1 6 :START2 0
						       :END1 LFN16 :END2 LFN1)))
				LINE)
			     (FORMAT T "~A~%" LINE)))
		     (FORMAT T "==MORE==")
		     (OR (CHAR= (READ-CHAR) #\SPACE)
			 (THROW 'ABORT NIL))
		     (MULTIPLE-VALUE-SETQ (X Y)
		       (SEND *STANDARD-OUTPUT* :READ-CURSORPOS))
		     (SEND *STANDARD-OUTPUT* :SET-CURSORPOS 0 Y)
		     (SEND *STANDARD-OUTPUT* :CLEAR-EOL)
		     (DO ((LINE LINE (SEND STREAM :LINE-IN)))
			 ((EQUAL LINE FIRST))
		       (COND ((ZEROP (LENGTH LINE))
			      (FORMAT T "------------------------------------------------~%")
			      (CLOSE STREAM)
			      (SETQ STREAM (OPEN FN :CHARACTERS T :DIRECTION :INPUT))
			      (SEND STREAM :LINE-IN)
			      (SEND STREAM :LINE-IN)
			      (SETQ LINE (SEND STREAM :LINE-IN))))
		       (FORMAT T "~A~%" LINE)))
		   (RETURN ()))))))))) 


(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST* DIRECTORY "Edit" DIRECTORY-EDIT-1
			  T "Run DIRED on this directory.") 

(DEFUN DIRECTORY-EDIT-1 (DIRECTORY)
  (DIRECTORY-EDIT DIRECTORY)
  NIL) 

(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST* DIRECTORY "View" VIEW-DIRECTORY
			  NIL "View this directory") 

(DEFCOM COM-LIST-ALL-DIRECTORY-NAMES "List names of all disk directories." ()
  (LET* ((DEFAULT (SEND (DEFAULT-PATHNAME)
			:NEW-PATHNAME
			:DIRECTORY :WILD
			:NAME :UNSPECIFIC
			:TYPE :UNSPECIFIC
			:VERSION :UNSPECIFIC))
	 (PATHNAME (READ-DEFAULTED-PATHNAME "List directories:" DEFAULT
					    :UNSPECIFIC :UNSPECIFIC))
	 (DIRS (FS:ALL-DIRECTORIES PATHNAME :NOERROR)))
    (IF (ERRORP DIRS)
	(BARF "Error: ~A" DIRS)
	(PROGN
	  (SETQ DIRS (SORT DIRS #'FS:PATHNAME-LESSP :KEY #'CAR))
	  (SEND *STANDARD-OUTPUT* :ITEM-LIST 'DIRECTORY
		(LOOP FOR (PATHNAME) IN DIRS
		      COLLECT `(,(SEND PATHNAME :STRING-FOR-DIRECTORY)
				. ,(SEND PATHNAME
					 :NEW-PATHNAME
					 :NAME :WILD
					 :TYPE :WILD
					 :VERSION :WILD)))))))
  DIS-NONE) 

(DEFCOM COM-EXPUNGE-DIRECTORY "Expunge deleted files from a directory" ()
  (LET* ((DIRECTORY (READ-DIRECTORY-NAME "Expunge directory" (DEFAULT-PATHNAME)))
	 (RESULT (FS:EXPUNGE-DIRECTORY DIRECTORY :ERROR :RETRY)))
    (IF (ERRORP RESULT)
	(BARF "Cannot expunge ~A: ~A" DIRECTORY RESULT)
	(FORMAT *QUERY-IO* "~&~A: ~D block~:P freed" DIRECTORY RESULT)))
  DIS-NONE) 

;;;  Directory creation:  if error, barf;  if new, create;  if exists, acknowledge.
(DEFCOM COM-CREATE-DIRECTORY "Create a directory" ()
  (LET* ((DIRECTORY (READ-DIRECTORY-NAME "Create directory" (DEFAULT-PATHNAME)))
	 (DIRNAME (SEND DIRECTORY :STRING-FOR-DIRECTORY))
	 ;; +++  We have to make sure that the directory's pathname is version ourselves,
	 ;;      because the LOSING pathname code tends to pass the original version.
	 (PATHNAME (SEND (SEND DIRECTORY :DIRECTORY-PATHNAME-AS-FILE) :NEW-VERSION 1))
	 (DIR-PROBE (PROBE-FILE PATHNAME)))
    (IF (NULL DIR-PROBE)
	(LET ((RESULT (FS:CREATE-DIRECTORY DIRECTORY :ERROR NIL)))
	  (IF (ERRORP RESULT)
	      (BARF "Cannot create directory ~A: ~A" DIRNAME RESULT)
	      (FORMAT *QUERY-IO* "~&Directory ~A created." DIRNAME)))
	(FORMAT *QUERY-IO* "~&Directory ~A already exists." DIRNAME)))
  DIS-NONE) 

