;;; Font hacking function and commands -*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1980, Massachusetts Institute of Technology


(DEFUN CHANGE-FONT-INTERVAL (START-BP &OPTIONAL END-BP IN-ORDER-P (FONT-NUM *FONT*))
  "Set the font number of all characters in the specified interval to FONT-NUM."
  (GET-INTERVAL START-BP END-BP IN-ORDER-P)
  (WITH-UNDO-SAVE-IF (NOT *BATCH-UNDO-SAVE*) ("Font change" START-BP END-BP T)
    (MUNG-BP-INTERVAL START-BP)
    (DO ((LINE (BP-LINE START-BP) (LINE-NEXT LINE))
	 (LIMIT-LINE (BP-LINE END-BP))
	 (START-INDEX (BP-INDEX START-BP) 0)
	 (LAST-LINE-P))
	(NIL)
      (SETQ LAST-LINE-P (EQ LINE LIMIT-LINE))
      (OR (ZEROP FONT-NUM)
	  (EQ (ARRAY-TYPE LINE) 'ART-FAT-STRING)
	  (SETQ LINE (SET-LINE-ARRAY-TYPE LINE 'ART-FAT-STRING)))
	(MUNG-LINE LINE)
      (DO ((INDEX START-INDEX (1+ INDEX))
	   (LIMIT-INDEX (IF LAST-LINE-P
			    (BP-INDEX END-BP)
			    (LINE-LENGTH LINE))))
	  ((>= INDEX LIMIT-INDEX))
	(if (not (and (>= (char-code (AREF LINE INDEX)) 137.) (<= (char-code (AREF LINE INDEX)) 211.)))
	    ;;; do not change the font on control characters.   3/3/88 
	    (SETF (AREF LINE INDEX) (MAKE-CHAR (AREF LINE INDEX)
					       (CHAR-BITS (AREF LINE INDEX))	;preserve the bits attribute
					       FONT-NUM))))
      (AND LAST-LINE-P
	   (RETURN NIL))))
  DIS-TEXT)

(DEFUN CHANGE-ONE-FONT-INTERVAL (OFONT NFONT FROM-BP &OPTIONAL TO-BP IN-ORDER-P)
  "For each character in specified interval, if it's in font OFONT, change to NFONT.
Both OFONT and NFONT are font numbers."
  (GET-INTERVAL FROM-BP TO-BP IN-ORDER-P)
  (WITH-UNDO-SAVE ("Font change" FROM-BP TO-BP T)
    (MUNG-BP-INTERVAL FROM-BP)
    (DO ((LINE (BP-LINE FROM-BP) (LINE-NEXT LINE))
	 (LIMIT-LINE (BP-LINE TO-BP))
	 (START-INDEX (BP-INDEX FROM-BP) 0)
	 (LAST-LINE-P))
	(NIL)
      (SETQ LAST-LINE-P (EQ LINE LIMIT-LINE))
      (OR (ZEROP NFONT)
	  (EQ (ARRAY-TYPE LINE) 'ART-FAT-STRING)
	  (SETQ LINE (SET-LINE-ARRAY-TYPE LINE 'ART-FAT-STRING)))
      (MUNG-LINE LINE)
      (DO ((INDEX START-INDEX (1+ INDEX))
	   (LIMIT-INDEX (IF LAST-LINE-P
			    (BP-INDEX TO-BP)
			    (LINE-LENGTH LINE))))
	  ((>= INDEX LIMIT-INDEX))
	(WHEN (= (CHAR-FONT (AREF LINE INDEX)) OFONT)	   
	  (SETF (AREF LINE INDEX) (MAKE-CHAR (AREF LINE INDEX)
					     (CHAR-BITS (AREF LINE INDEX))	;preserve the bits attribute
					     NFONT))))
      (AND LAST-LINE-P
	   (RETURN ())))))

;; New function by rpm 2-4-87. Called by REDEFINE-FONTS.
(DEFUN CLEANUP-FONTED-INTERVAL (FONT-ALIST-LENGTH FROM-BP &OPTIONAL TO-BP IN-ORDER-P)
  "For each character in specified interval, if it's in a font that's outside
the range of fonts in the font alist, unfont the character."
  (GET-INTERVAL FROM-BP TO-BP IN-ORDER-P)
  (WITH-UNDO-SAVE ("FONT CHANGE" FROM-BP TO-BP T)
    (MUNG-BP-INTERVAL FROM-BP)
    (DO ((LINE (BP-LINE FROM-BP) (LINE-NEXT LINE))
	 (LIMIT-LINE (BP-LINE TO-BP))
	 (START-INDEX (BP-INDEX FROM-BP) 0)
	 (LAST-LINE-P))
	(NIL)
      (SETQ LAST-LINE-P (EQ LINE LIMIT-LINE))
      (MUNG-LINE LINE)
      (DO ((INDEX START-INDEX (1+ INDEX))
	   (LIMIT-INDEX (IF LAST-LINE-P
			    (BP-INDEX TO-BP)
			    (LINE-LENGTH LINE))))
	  ((>= INDEX LIMIT-INDEX))
	(WHEN (>= (CHAR-FONT (AREF LINE INDEX)) FONT-ALIST-LENGTH)	   
	  (SETF (AREF LINE INDEX) (MAKE-CHAR (AREF LINE INDEX)
					     (CHAR-BITS (AREF LINE INDEX))))))
      (AND LAST-LINE-P
	   (RETURN ()))))) 

(DEFUN MAP-FONTS-INTERVAL (PERMUTATION FROM-BP &OPTIONAL TO-BP IN-ORDER-P)
  "Permute fonts, for all characters in the specified interval.
Each char's font is used as an index in PERMUTATION (a sequence)
to get the new font number for that character."
  (GET-INTERVAL FROM-BP TO-BP IN-ORDER-P)
  (WITH-UNDO-SAVE-IF (NOT *BATCH-UNDO-SAVE*) ("Font change" FROM-BP TO-BP T)
    (MUNG-BP-INTERVAL FROM-BP)
    (DO ((LINE (BP-LINE FROM-BP) (LINE-NEXT LINE))
	 (LIMIT-LINE (BP-LINE TO-BP))
	 (START-INDEX (BP-INDEX FROM-BP) 0)
	 (LAST-LINE-P))
	(NIL)
      (SETQ LAST-LINE-P (EQ LINE LIMIT-LINE))
      (OR (EQ (ARRAY-TYPE LINE) 'ART-FAT-STRING)
	  (SETQ LINE (SET-LINE-ARRAY-TYPE LINE 'ART-FAT-STRING)))
      (MUNG-LINE LINE)
      (DO ((INDEX START-INDEX (1+ INDEX))
	   (LIMIT-INDEX (IF LAST-LINE-P
			    (BP-INDEX TO-BP)
			    (LINE-LENGTH LINE))))
	  ((>= INDEX LIMIT-INDEX))	
	(SETF (AREF LINE INDEX) (MAKE-CHAR (AREF LINE INDEX)
					   (CHAR-BITS (AREF LINE INDEX))	;preserve the bits attribute
					   (ELT PERMUTATION (CHAR-FONT (AREF LINE INDEX))))))
      (AND LAST-LINE-P
	   (RETURN ()))))) 

(DEFUN FIXUP-FONTS-INTERVAL (FONTS FROM-BP &OPTIONAL TO-BP IN-ORDER-P)
  "Permute the fonts in some inserted text, to show up properly in this interval.
FROM-BP and TO-BP should specify a subinterval of *INTERVAL*.
The text in that subinterval is assumed to have been copied out
of another interval (or a file) in which FONTS was the font list.
Update *INTERVAL*'s font list, and the font numbers of the characters,
so that each character appears in the same font as it did where it came from."
  (GET-INTERVAL FROM-BP TO-BP IN-ORDER-P)
  (IF (NOT (AND FONTS (BUFFER-SAVED-FONT-ALIST *INTERVAL*)))
      (CHANGE-FONT-INTERVAL FROM-BP TO-BP T 0)
      (PROGN
	(ADD-FONTS-TO-BUFFER FONTS)
	(LET ((PERMUTATION (MAKE-ARRAY (LENGTH FONTS))))
	  (DOTIMES (I (LENGTH FONTS))
	    (DO ((J 0 (1+ J))
		 (LIST (WINDOW-FONT-ALIST *WINDOW*) (CDR LIST)))
		((EQUAL (CAAR LIST) (SYMBOL-NAME (NTH I FONTS)))
		 (SETF (AREF PERMUTATION I) J))))
	  (MAP-FONTS-INTERVAL PERMUTATION FROM-BP TO-BP T))))) 


(DEFUN INPUT-FONT-NAME (USE-PREVIOUS-P &OPTIONAL (PROMPT "Font ID: ") &AUX NUM)
  "Read a font number from the user.
The user can type a letter (A signifies font 0), ESCAPE and a font name,
mouse left on a character, or mouse right and get a menu of fonts.
USE-PREVIOUS-P means if previous command called this function,
just return the same number we returned for that command."
  (SETQ *CURRENT-COMMAND-TYPE* 'FONT-CHANGE)
  (COND ((AND USE-PREVIOUS-P (EQ *LAST-COMMAND-TYPE* 'FONT-CHANGE))
	 *SAVE-FONT-NUM*)
	(T
	 (FORMAT *QUERY-IO* PROMPT)
	 (DO (CH
	      (ALIST (WINDOW-FONT-ALIST *WINDOW*)))
	     (NIL)
	   (LET-GLOBALLY
	     ((*GLOBAL-MOUSE-CHAR-BLINKER-DOCUMENTATION-STRING*
		(IF (NULL ALIST)
		    "L: Use font of character pointed at."
		    "L: Use font of character pointed at, R: click right for menu of fonts.")))
	     (UNLESS CH
		     (TYPEIN-LINE-ACTIVATE
		       (SETQ CH (CHAR-UPCASE (W:READ-MOUSE-OR-KBD))))))
	   (COND ((CHAR= CH #\c-G)
		  (BARF))
		 ((CHAR= CH #\ESCAPE)
		  (SETQ NUM (INPUT-FONT-NAME-FROM-MINI-BUFFER))
		  (RETURN NIL))
		 ((CHAR= CH #\MOUSE-L)
		  (COND ((SETQ CH (MOUSE-CHAR *WINDOW*)) 
			 (SETQ NUM (CHAR-FONT CH))
			 (RETURN NIL))))
		 ((CHAR= CH #\MOUSE-R)
		  (COND ((NULL ALIST)
			 (BEEP)
			 (TYPEIN-LINE
			   "~&A font menu is only available when the buffer has a FONTS attribute.")
			 (TYPEIN-LINE PROMPT)
			 (SETQ CH NIL))
			((SETQ CH (W:MENU-CHOOSE ALIST))
			 (DO ((I 0 (1+ I))	;Have the font itself, but want the number
			      (L (WINDOW-FONT-ALIST *WINDOW*) (CDR L)))
			     ((EQ (CDAR L) CH)
			      (SETQ NUM I)))
			 (RETURN NIL))))
		 ((CHAR<= #\A CH #\Z)
		  (SETQ NUM (- (CHAR-CODE CH) (CHAR-CODE #\A)))
		  (RETURN ()))
		 ((OR (CHAR= CH #\HELP) (CHAR= CH #\?))
		  (FORMAT *QUERY-IO*
			  "~:|Type a font letter, ~
				  or ESCAPE to enter a new font in a mini-buffer, ~@
				  or mouse a character left for its font~:[~;, ~
				  or mouse-right for a menu~].~%"
			  ALIST)
		  (TYPEIN-LINE PROMPT)
		  (SETQ CH NIL))
		 (T
		  (BEEP)
		  (TYPEIN-LINE "~&~:@C is not a proper choice.  Try pressing HELP." CH)
		  (TYPEIN-LINE PROMPT)
		  (SETQ CH NIL))))
	 (FORMAT *QUERY-IO* "~C (~A)" (CODE-CHAR (+ NUM (CHAR-CODE #\A)))
		 (CAR (NTH NUM (WINDOW-FONT-ALIST *WINDOW*))))
	 (SETQ *SAVE-FONT-NUM* NUM)))) 

(DEFUN INPUT-FONT-NAME-FROM-MINI-BUFFER (&AUX FONT NEW-P)
  (SETQ FONT (COMPLETING-READ-FROM-MINI-BUFFER "Font ID:" (WINDOW-FONT-ALIST *WINDOW*) T))
  (COND ((EQUAL FONT "")			; Empty, take the default.
	 (SETQ FONT *FONT*))
	((STRINGP FONT)				; String, means a new font, add to list.
	 (SETQ NEW-P 'T)
	 (ADD-FONTS-TO-BUFFER (LIST FONT))
	 (SETQ FONT (POSITION (ASSOC (STRING-UPCASE FONT) (WINDOW-FONT-ALIST *WINDOW*) :TEST #'EQUAL)
			      (THE LIST (WINDOW-FONT-ALIST *WINDOW*))
			      :TEST #'EQ)))
	(:ELSE					; Known font, get its number.
	 (SETQ FONT (POSITION FONT (THE LIST (WINDOW-FONT-ALIST *WINDOW*)) :TEST #'EQ))))
  (FORMAT *QUERY-IO* (IF NEW-P
			 "Added as font "
			 "Font "))		;Caller will append letter & name of font
  FONT) 

(DEFUN ADD-FONTS-TO-BUFFER (FONTS)
  "Make sure that all the fonts in the list FONTS are available in the current buffer."
  (PKG-BIND "FONTS"
    (SETQ FONTS
	  (LOOP FOR FONT IN FONTS
		COLLECT (W:FONT-NAME (SEND (W:SHEET-GET-SCREEN (WINDOW-SHEET *WINDOW*))
					   :PARSE-FONT-DESCRIPTOR
					   (READ-FROM-STRING (STRING FONT) T))))))
  (LET ((OLD-LIST (WINDOW-FONT-ALIST *WINDOW*)))
    (AND (>= (LENGTH OLD-LIST) 32)
	 (BARF "The maximum number of fonts is 26."))
    (OR OLD-LIST
	(SETQ OLD-LIST (LET ((FONT0 (CURRENT-FONT *WINDOW* 0)))
			 (LIST (CONS (SYMBOL-NAME (W:FONT-NAME FONT0)) FONT0)))))
    (SETQ FONTS (REMOVE-IF #'(LAMBDA (FONT)
			       (ASSOC (SYMBOL-NAME FONT) OLD-LIST :TEST #'EQUAL))
			   FONTS))
    (WHEN FONTS
      (FORMAT *QUERY-IO* "~&Font~P added to buffer's font list.~%" (LENGTH FONTS))
      (SEND *INTERVAL* :SET-ATTRIBUTE :FONTS
	    (NCONC (MAPCAR #'(LAMBDA (AELT) (W:FONT-NAME (CDR AELT))) OLD-LIST) FONTS)
	    :QUERY)
      (REDEFINE-FONTS *WINDOW* (APPEND OLD-LIST
				       (MAPCAR #'(LAMBDA (FONT)
						   (CONS (SYMBOL-NAME FONT) (SYMBOL-VALUE FONT)))
					       FONTS))))
    FONTS)) 


(DEFUN REDEFINE-FONTS (WINDOW FONT-ALIST &OPTIONAL VSP VIEWING-ONLY &AUX NEW-LENGTH)
  "Set the font alist of WINDOW (a window defstruct) to FONT-ALIST."
  ;; If the list of fonts in this buffer has been shortened, unfont any characters 
  ;; in fonts that are no longer around. This way, what we now see in the buffer is
  ;; really what we will get when we save it. rpm 2-5-87 
  (AND *INTERVAL*
       (NOT VIEWING-ONLY)
       (NOT (BUFFER-READ-ONLY-P *INTERVAL*))
       (EQ (WINDOW-INTERVAL WINDOW) *INTERVAL*)
       (< (SETQ NEW-LENGTH (LENGTH FONT-ALIST))
	  (LENGTH (WINDOW-FONT-ALIST WINDOW)))
       (CLEANUP-FONTED-INTERVAL NEW-LENGTH (WINDOW-INTERVAL WINDOW)))
  (COND ((NOT (EQUAL FONT-ALIST (WINDOW-FONT-ALIST WINDOW)))
	 (LET ((SHEET (WINDOW-SHEET WINDOW)))
	   (SEND SHEET :SET-FONT-MAP (MAPCAR #'CDR FONT-ALIST))
	   (SEND SHEET :SET-CURRENT-FONT (AREF (W:SHEET-FONT-MAP SHEET) 0)))
	 (MUST-REDISPLAY WINDOW DIS-ALL)
	 (SETF (WINDOW-FONT-ALIST WINDOW) FONT-ALIST)
	 (WHEN (EQ WINDOW *WINDOW*)
	   (IF (NULL FONT-ALIST)
	       (SETQ *FONT* 0)
	       (IF (>= *FONT* (LENGTH FONT-ALIST))
		   (SETQ *FONT* 0)))
	   (UPDATE-FONT-NAME))))
  (SEND (WINDOW-INTERVAL WINDOW) :SET-SAVED-FONT-ALIST FONT-ALIST)
  (COND ((AND VSP (/= VSP (SEND (WINDOW-SHEET WINDOW) :VSP)))
	 (MUST-REDISPLAY WINDOW DIS-ALL)
	 (SEND (WINDOW-SHEET WINDOW) :SET-VSP VSP)))) 

(DEFUN UPDATE-FONT-NAME ()
  (SETQ *FONT-NAME*
	(AND (WINDOW-FONT-ALIST *WINDOW*)
	     (FORMAT NIL "~C (~A)" (CODE-CHAR (+ *FONT* (CHAR-CODE #\A)))
		     (CAR (NTH *FONT* (WINDOW-FONT-ALIST *WINDOW*))))))) 

;;; This is used by ZMACS
(DEFUN SET-BUFFER-FONTS (BUFFER &OPTIONAL (FONTS (SEND BUFFER :GET-ATTRIBUTE :FONTS)))
  "Set the fonts of BUFFER according to its attribute list.
We assume that the buffer's attribute list has been read in and stored."
  (COND ((AND FONTS (SYMBOLP FONTS))
	 (SETQ FONTS (INTERN (SYMBOL-NAME FONTS) "FONTS"))
	 (SETQ FONTS (AND (BOUNDP FONTS) (LIST (CONS (SYMBOL-NAME FONTS) (SYMBOL-VALUE FONTS))))))
	(T
	 (DO ((FL FONTS (CDR FL))
	      (L NIL)
	      (F))
	     ((NULL FL)
	      (SETQ FONTS (NREVERSE L)))
	   (SETQ F (INTERN (SYMBOL-NAME (CAR FL)) "FONTS"))
	   (COND ((NOT (BOUNDP F))
		  (send (send (send *window* :superior) :superior) :parse-font-descriptor f)
		  (OR (BOUNDP F)
		      ;;If font not loaded, substitute default to keep font numbers ok
		      (SETQ F (W:FONT-NAME (W:SCREEN-DEFAULT-FONT W:DEFAULT-SCREEN))))))
	   (PUSH (CONS (SYMBOL-NAME F) (SYMBOL-VALUE F)) L))))
  (SEND BUFFER :SET-SAVED-FONT-ALIST FONTS)
  FONTS) 
      

(DEFCOM COM-CHANGE-FONT-CHAR "Change the font of one or more characters forward.
Reads in the echo area a letter signifying the new font
\(A for the first font, etc.).
Or type ESCAPE and a font name, click mouse left on a char in the buffer
to use its font, or mouse right and get a menu of fonts.
If the previous command was a font-change command,
the same font is used, without reading any argument." ()
  (LET ((BP1 (FORWARD-CHAR (POINT) *NUMERIC-ARG* T)))
    (CHANGE-FONT-INTERVAL (POINT) BP1 NIL
			  (INPUT-FONT-NAME T "Change font of characters to (Font ID): "))
    (IF (PLUSP *NUMERIC-ARG*)
	(MOVE-BP (POINT) BP1))
    DIS-TEXT)) 
  
(DEFCOM COM-CHANGE-FONT-WORD "Change the font of one or more words forward.
Reads in the echo area a letter signifying the new font
\(A for the first font, etc.).
Or type ESCAPE and a font name, click mouse left on a char in the buffer
to use its font, or mouse right and get a menu of fonts.
If the previous command was a font-change command,
the same font is used, without reading any argument." ()
  (LET ((BP1 (FORWARD-WORD (POINT) *NUMERIC-ARG* T)))
    (CHANGE-FONT-INTERVAL (POINT) BP1 NIL
			  (INPUT-FONT-NAME T "Change font of words to (Font ID): "))
    (AND (PLUSP *NUMERIC-ARG*)
	 (MOVE-BP (POINT) BP1))
    DIS-TEXT)) 

(DEFCOM COM-CHANGE-FONT-REGION "Change the font between point and the mark.
Reads in the echo area a letter signifying the new font (A for the first font, etc.).
Or type ESCAPE and a font name, click mouse left on a char in the buffer to use its font,
or mouse right and get a menu of fonts." ()
  (REGION (BP1 BP2)
    (CHANGE-FONT-INTERVAL BP1 BP2 T
			  (INPUT-FONT-NAME NIL "Change font of region to (Font ID): ")))) 

(DEFCOM COM-CHANGE-ONE-FONT-REGION "Change all characters of one font to another font.
Asks you to specify two fonts, then changes each character in the region
which is currently in the first font to be in the second font instead." ()
  (LET* ((OFONT (INPUT-FONT-NAME NIL "Change font of characters in (Font ID): "))
	 (NFONT (INPUT-FONT-NAME NIL " to (Font ID): ")))
    (REGION (BP1 BP2) (CHANGE-ONE-FONT-INTERVAL OFONT NFONT BP1 BP2))
    DIS-TEXT)) 
      
(DEFCOM COM-CHANGE-DEFAULT-FONT "Set the default font. Reads in the echo
area a letter signifying the new font (A for the first font, etc.).
Or type ESCAPE and a font name, click mouse left on a char in the buffer to use its font,
or mouse right and get a menu of fonts." ()
  (SETQ *FONT* (INPUT-FONT-NAME NIL "Set default font to (Font ID): "))
  (UPDATE-FONT-NAME)
  DIS-BPS)					;This may change the size of the blinker

;;; New command from dkm. Added by rpm on 11-25-86. 
(DEFCOM COM-FONTIFY-REGION-OR-BUFFER
  "Fontify a marked region or the whole buffer according to electric font lock rules:
     Default font      A 
     Comments          B 
     Doc Strings       C 
     Non Doc Strings   D or *ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT* if a fixnum
     Function Spec     E" ()
  (LET* ((CURRENT-FONT 0)
	 (NUMBER-OF-FONTS (LENGTH (WINDOW-FONT-ALIST *WINDOW*)))
	 (FONT-COMMENTS (>= NUMBER-OF-FONTS 2))
	 (FONT-DOC-STRING (>= NUMBER-OF-FONTS 3))
	 (FONT-FSPEC (>= NUMBER-OF-FONTS 4))
	 (FONT-NON-DOC-STRINGS (>= NUMBER-OF-FONTS 5))
	 (SAVE-FONT *FONT*)
	 FIRST-LINE MODE-LINE-EXISTS LAST-LINE SLASH IN-STRING DBL-QUOTE-CHAR
	 NEXT-CHAR-IN-STRING NEXT-CHAR-IN-COMMENT IN-COMMENT DOC-STRING-P
	 FUNC-SPEC-START FUNC-SPEC-END MAX-INDEX START-BP END-BP NEW-BP)
    (WHEN (NEQ *INTERVAL* (WINDOW-INTERVAL *MINI-BUFFER-WINDOW*))
      ;; Get the interval to fontify.  Use marked region, or do the whole buffer.
      (MULTIPLE-VALUE-SETQ (START-BP END-BP)
          (IF (WINDOW-MARK-P *WINDOW*)               ; Region is marked.
	      (REGION (A B)
		 (VALUES A B))                       ; Just return start and end of marked region.
	      (VALUES (INTERVAL-FIRST-BP *INTERVAL*) ; Else the start and end of whole buffer.
		      (INTERVAL-LAST-BP *INTERVAL*))))
      (SETQ FIRST-LINE (BP-LINE START-BP)
	    ;should be a better way to determine this
	    MODE-LINE-EXISTS (LISP:SEARCH "-*-" (THE STRING (STRING FIRST-LINE)) :TEST #'CHAR-EQUAL))
      ;;Loop thru the whole region character by character via charmap.
      (CHARMAP (START-BP END-BP)
	 ;;Some things to do each time we get a new line
         (WHEN (NEQ LINE LAST-LINE)
            (AND FONT-FSPEC            ; See if there is func spec on this line, if we're fonting them.
		 (MULTIPLE-VALUE-SETQ (FUNC-SPEC-START FUNC-SPEC-END)
		      (IN-FUNCTION-SPEC LINE)))
	    (SETQ MAX-INDEX (1- (LENGTH LINE))
		   LAST-LINE LINE
		   IN-COMMENT NIL))
	 ;;Determine where this character is and set the appropriate state variables.
	 (LET ((CH (CHARMAP-CH-CHAR)))
	   (COND (SLASH (SETQ SLASH NIL))
		 (T
		   (WHEN NEXT-CHAR-IN-STRING
		     (SETQ IN-STRING T
                           NEXT-CHAR-IN-STRING NIL))
		   (WHEN NEXT-CHAR-IN-COMMENT
		     (SETQ IN-COMMENT T
			   NEXT-CHAR-IN-COMMENT NIL))
		   (SELECT (LIST-SYNTAX-OPEN-CODED CH)
		      (LIST-SLASH (SETQ SLASH T))
		      (LIST-COMMENT (OR IN-STRING (SETQ NEXT-CHAR-IN-COMMENT T)))
		      (LIST-DOUBLE-QUOTE
			(COND (IN-COMMENT)
			      ((NOT IN-STRING)
			       (SETQ NEXT-CHAR-IN-STRING CH
			               DBL-QUOTE-CHAR CH)
			       (WHEN FONT-DOC-STRING
			         (LET ((BP (CREATE-BP LINE INDEX)))
				   (SETQ DOC-STRING-P
					 (BP-= (FORWARD-CHAR (FORWARD-DEFUN BP -1 T) 1 T)
					       (FORWARD-SEXP BP -3))))))
			      ((CHAR= CH dbl-quote-char)
			       (SETQ IN-STRING NIL
				     DOC-STRING-P NIL))))))))
	 ;;Determine which font this character should be in.
	 (LET ((NEW-FONT
		 (COND
		   ((AND FONT-COMMENTS IN-COMMENT )	                ;In a Comment.
		    (IF (AND MODE-LINE-EXISTS (EQ LINE FIRST-LINE))	;Don't font mode line.
			0 1))
		   ((AND FONT-DOC-STRING IN-STRING DOC-STRING-P)	;In a documentation string.
		    2)
		   ((AND FONT-NON-DOC-STRINGS IN-STRING)		;In a non-documentation string.
		    (IF (INTEGERP  *ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT*)
			*ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT*
			3))
		   ((AND FUNC-SPEC-START (<= FUNC-SPEC-START INDEX FUNC-SPEC-END))	;In a function spec.
		    4)		
		   (T 0))))						;In a default area.
	   ;;See if we have a new-font.  Change the previous interval to the correct font.
	   (COND ((/= *FONT* NEW-FONT)
		  (SETQ *FONT* NEW-FONT)	
		  (CHANGE-FONT-INTERVAL START-BP
					(SETQ NEW-BP (CHARMAP-BP-BEFORE)) 
					T CURRENT-FONT)
		  (SETQ CURRENT-FONT *FONT*
			START-BP NEW-BP
			LAST-LINE (BP-LINE START-BP))))))
      ;;Change the font of the last interval if necessary.
      (UNLESS (BP-= START-BP END-BP)
	  (CHANGE-FONT-INTERVAL START-BP END-BP T *FONT*))
      ;;Put the font back to the value it was when we started all this
      (SETQ *FONT* SAVE-FONT)
      (UPDATE-FONT-NAME)))
  DIS-TEXT)

;;; Used by COM-FONTIFY-REGION-OR-BUFFER. Also added on 11-25-86.
(DEFUN IN-FUNCTION-SPEC (LINE &OPTIONAL NEW-CHAR NEW-CHAR-INDEX)
  "Returns location of of the function spec or nil if not a line starting with (def."
  (DECLARE (VALUES START-LOC END-LOC) (STRING LINE))
  (LET ((LINE-SIZE (LENGTH LINE)))
    (WHEN (AND (PLUSP LINE-SIZE)
	       (STRING-EQUAL LINE "(DEF" :END1 4 :END2 4))	;does this line start with "(def" ? 
	  (LET* ((FIRST-SPACE (POSITION #\SPACE LINE :TEST #'CHAR-EQUAL))
		 (SECOND-ATOM-POSITION (AND FIRST-SPACE
					    (POSITION #\SPACE LINE
						      :TEST-NOT #'CHAR-EQUAL 
						      :START (1+ FIRST-SPACE))))
		 (START-OF-FSPEC (AND FIRST-SPACE
				      (OR
					;; This variable now can be affected by this next clause.
					;; This next clause handles the case when the next character that
					;; is about to be inserted is the beginning of the function spec
					;; although the current line doesn't have it yet.  This is specifically
					;; for font-lock-hook, where this routine is called on the current
					;; character before it is inserted in the line. -dkm
					(AND NEW-CHAR
					     (NOT (CHAR-EQUAL NEW-CHAR #\SPACE))
					     NEW-CHAR-INDEX
					     SECOND-ATOM-POSITION
					     (< FIRST-SPACE NEW-CHAR-INDEX SECOND-ATOM-POSITION)
					     NEW-CHAR-INDEX)
					SECOND-ATOM-POSITION
					(1- LINE-SIZE))))
		 (FSPEC-IN-PARENS (AND START-OF-FSPEC
				       (CHAR-EQUAL (AREF LINE START-OF-FSPEC) #\( )))
		 (ENDING-CHAR (IF (AND FSPEC-IN-PARENS
				       (NOT (STRING-EQUAL "(DEFSTRUCT" LINE :END2 10.)))
				  #\)
				  #\SPACE))
		 (ENDING-POS (AND START-OF-FSPEC
				  (1- (OR (POSITION ENDING-CHAR LINE :START (1+ START-OF-FSPEC) :TEST #'CHAR-EQUAL)
					  LINE-SIZE)))))
	    (VALUES START-OF-FSPEC
		    ENDING-POS)))))

;; COM-SET-FONTS redefined from 94.214.   ddd/gsl 3/5/84.
;; Modified to support shortening buffer font-list on 2-4-87 by rpm.
(DEFCOM COM-SET-FONTS "Change the set of fonts to use.
This sets the fonts associated with the current buffer.
Reads a list of font names, separated by spaces, from the mini-buffer." ()
  (LET ((TEM (DO ((FL (OR (WINDOW-FONT-ALIST *WINDOW*)
			  (LIST (LIST (W:FONT-NAME
					(AREF (SEND (WINDOW-SHEET *WINDOW*) :FONT-MAP)
					      0)))))
		      (CDR FL))
		  (STR (MAKE-ARRAY 100
				   :ELEMENT-TYPE 'STRING-CHAR
				   :LEADER-LIST '(0))) 
		  (FIL "" " "))
		 ((NULL FL) STR)
	       (SETQ STR (STRING-NCONC STR FIL (CAAR FL)))))
	FONTS)
    (LET ((*MINI-BUFFER-DEFAULT-STRING* TEM))
      (DO ((I 0)
	   J
	   (STRING (STRING-TRIM *BLANKS* (TYPEIN-LINE-READLINE "font1 font2 ...:"))))
	  ((OR (NULL I) (EQUAL STRING "")))
	(SETQ J (STRING-SEARCH-SET '(#\SPACE #\, #\TAB) STRING I))
	(PUSH (INTERN (STRING-UPCASE (NSUBSTRING STRING I J)) "FONTS")
	      FONTS)
	(SETQ I (AND J (STRING-SEARCH-NOT-SET '(#\SPACE #\, #\TAB) STRING (1+ J))))))
    ;; If fewer fonts have been set than previously existed, query the user and confirm
    ;; before proceeding. rpm 2-4-87
    (OR (LET* ((NEW-LENGTH (LENGTH FONTS))
	       (OLD-FONTS (SEND *INTERVAL* :GET-ATTRIBUTE :FONTS))
	       (OLD-LENGTH (IF (LISTP OLD-FONTS) (LENGTH OLD-FONTS) 1)))
	  (AND (< NEW-LENGTH OLD-LENGTH)
	       (NOT (FQUERY `(:SELECT T :BEEP T :TYPE :READLINE :CHOICES ,FORMAT:YES-OR-NO-P-CHOICES)
			    "Buffer font list has been shortened. Characters in font~:[ ~A~*~;s ~A through ~A~] will change to ~:[font A~;the default font~]. ~%Proceed? "
			    (> (- OLD-LENGTH NEW-LENGTH) 1)
			    (INT-CHAR (+ NEW-LENGTH 101))
			    (INT-CHAR (+ OLD-LENGTH 100))
			    (ZEROP NEW-LENGTH)))
	       DIS-NONE))
	(PROGN 
	  ;; Fonts are in reverse order now, but we will reverse them again as we check them.
	  (DO ((L FONTS (CDR L))
	       (FONT)
	       (FONT-NAMES)
	       (AL NIL))
	      ((NULL L)
	       (SEND *INTERVAL* :SET-ATTRIBUTE :FONTS FONT-NAMES :QUERY)
	       (SETQ TEM AL))
	    (SETQ FONT (CAR L))
	    (COND ((NOT (SYMBOLP FONT))
		   (BARF "~S is not the name of a font" FONT))
		  ((NOT (BOUNDP FONT))
		   (send (send (send *window* :superior) :superior)
						  :parse-font-descriptor FONT)
		   (OR (BOUNDP FONT) (BARF "~S is not a defined font" FONT))))
	    (PUSH FONT FONT-NAMES)
	    (PUSH (CONS (SYMBOL-NAME FONT) (SYMBOL-VALUE FONT))
		  AL))
	  (REDEFINE-FONTS *WINDOW* TEM)
	  (UPDATE-FONT-NAME)
	  DIS-ALL))))

(DEFCOM COM-LIST-FONTS "List the loaded fonts.
With an argument, also lists the font files on the file computer." ()
  (FORMAT T "Loaded fonts:~%")
  (SEND *STANDARD-OUTPUT* :ITEM-LIST 'FONT
	(LET ((LIST NIL))
	  (DECLARE (SPECIAL LIST))
	  (DO-LOCAL-SYMBOLS (X (FIND-PACKAGE 'FONTS) NIL)
	    (AND (BOUNDP X)
		 (TYPEP (SYMBOL-VALUE X) 'FONT)
		 (PUSH X LIST)))
	  (SORT LIST #'STRING-LESSP)))
  (COND (*NUMERIC-ARG-P*
	 (FORMAT T "~&Plus fonts on the file computer:~%")
	 (SEND *STANDARD-OUTPUT* :ITEM-LIST 'FONT
	       (LOOP FOR FILE IN
		     (FS:DIRECTORY-LIST (MAKE-PATHNAME :HOST "SYS"
						       :DIRECTORY "FONTS"
						       :NAME :WILD
						       :TYPE (SYSTEM:LOCAL-BINARY-FILE-TYPE)
						       :VERSION :NEWEST)
					:FAST)
		     WHEN (CAR FILE)
		     COLLECT (INTERN (STRING-UPCASE (SEND (CAR FILE) :NAME))
				     "FONTS")))))
  DIS-NONE)

;;; COM-DISPLAY-FONT modified to take a numeric arg to mean that you want to display this font 
;;; with something other than CPTFONT as the upper font in the display. Done by rpm on 7-9-87.
(DEFCOM COM-DISPLAY-FONT "Display all the characters in a font. With a numeric arg, you will
be prompted for the font to display above it for comparison. Otherwise the top font is CPTFONT." ()
  (LET ((FONT (COMPLETING-READ-FROM-MINI-BUFFER
		"Font to display:"
		(LET ((LIST NIL))
		  (DECLARE (SPECIAL LIST))
		  (DO-LOCAL-SYMBOLS (X (FIND-PACKAGE 'FONTS) LIST)
		    (AND (BOUNDP X)
			 (TYPEP (SYMBOL-VALUE X) 'FONT)
			 (PUSH (CONS (SYMBOL-NAME X) X)
			       LIST))))			       
		T))
	(COMPARISON-FONT (WHEN *NUMERIC-ARG-P*
			  (COMPLETING-READ-FROM-MINI-BUFFER
			    "Font to display above it for comparison:"
			    (LET ((LIST NIL))
			      (DECLARE (SPECIAL LIST))
			      (DO-LOCAL-SYMBOLS (X (FIND-PACKAGE 'FONTS) LIST)
				(AND (BOUNDP X)
				     (TYPEP (SYMBOL-VALUE X) 'FONT)
				     (PUSH (CONS (SYMBOL-NAME X) X)
					   LIST))))			       
			    T))))
    (SETQ FONT (IF (STRINGP FONT)
		   (INTERN (STRING-UPCASE (STRING-RIGHT-TRIM #\SPACE FONT)) "FONTS")
		   (CDR FONT)))
    (SETQ COMPARISON-FONT (COND ((NULL COMPARISON-FONT)
				 FONTS:CPTFONT)
				((STRINGP COMPARISON-FONT)
				 (INTERN (STRING-UPCASE (STRING-RIGHT-TRIM #\SPACE COMPARISON-FONT)) "FONTS"))
				(T (CDR COMPARISON-FONT))))
    (DISPLAY-FONT FONT COMPARISON-FONT))
  DIS-NONE) 

(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST* FONT "Display" DISPLAY-FONT
			  T "Display this font.") 

;;; DISPLAY-FONT now accepts optional arg as the font to compare to. Done by rpm on 7-9-87.
(DEFUN DISPLAY-FONT (FONT-SYMBOL &OPTIONAL (COMPARISON-FONT FONTS:CPTFONT) &AUX FONT)
  (COND ((NOT (TYPEP *WINDOW* 'WINDOW))
	 (FERROR NIL "DISPLAY-FONT called on a non-sheet window."))
	((FBOUNDP 'W:DISPLAY-FONT)
	 (SETQ FONT (SEND (W:SHEET-GET-SCREEN *WINDOW*)
		   :PARSE-FONT-DESCRIPTOR FONT-SYMBOL))
	 (SETQ COMPARISON-FONT (SEND (W:SHEET-GET-SCREEN *WINDOW*)
				     :PARSE-FONT-DESCRIPTOR COMPARISON-FONT))
	 (W:DISPLAY-FONT FONT :WINDOW *TERMINAL-IO* :SAMPLE-FONT COMPARISON-FONT)
	 (SEND *STANDARD-OUTPUT* :CLEAR-EOF)
	 (VALUES))
	(T
	 (BEEP)
	 (FORMAT *QUERY-IO* "Support for displaying fonts is not present in this environment."))))


;;; Diagram stuff

(DEFFLAVOR LINE-DIAGRAM-MIXIN () ()
  (:REQUIRED-METHODS :DRAW)) 

(DEFMETHOD (LINE-DIAGRAM-MIXIN :ADD-LINE) (LINE &OPTIONAL CONTENTS)
  LINE CONTENTS) 

(DEFMETHOD (LINE-DIAGRAM-MIXIN :STRING-FOR-FILE) (LINE)
  LINE "") 

(DEFMETHOD (LINE-DIAGRAM-MIXIN :NUMBER-OF-LINES) ()
  1) 

;;; Changed to fix a bug in the assumed order of the values returned by
;;; MAKE-DIAGRAM-LINE. Done by rpm (from dls) on 11-26-86.
(DEFUN INSERT-DIAGRAM (BP DIAGRAM &REST OPTIONS)
  (DO ((I 0 (1+ I))
       (AT-LINE (BP-LINE BP))
       (NUMBER-OF-LINES (SEND DIAGRAM :NUMBER-OF-LINES))
       (LINE))
      ((>= I NUMBER-OF-LINES) DIAGRAM)
    (MULTIPLE-VALUE-SETQ (LINE DIAGRAM)
      (APPLY #'MAKE-DIAGRAM-LINE DIAGRAM OPTIONS))
    (INSERT-LINE-WITH-LEADER LINE AT-LINE))) 

(DEFUN MAKE-DIAGRAM-LINE (DIAGRAM &REST OPTIONS &AUX LINE)
  (AND (SYMBOLP DIAGRAM)
       (SETQ DIAGRAM (APPLY #'MAKE-INSTANCE DIAGRAM OPTIONS)))
  (SETQ LINE (MAKE-LINE :MAKE-ARRAY (:AREA *LINE-AREA* :ELEMENT-TYPE 'STRING-CHAR)
			LINE-TICK *TICK*
			LINE-LENGTH 0))
  (SETF (GETF (LINE-PLIST LINE) :DIAGRAM) DIAGRAM)
  (SEND DIAGRAM :ADD-LINE LINE)
  (VALUES LINE DIAGRAM)) 

(DEFFLAVOR BLACK-BLOCK-DIAGRAM () (LINE-DIAGRAM-MIXIN)) 

(DEFMETHOD (BLACK-BLOCK-DIAGRAM :DRAW) (IGNORE SHEET)
  (W:PREPARE-SHEET (SHEET)
     (SYSTEM:%DRAW-RECTANGLE (W:SHEET-INSIDE-WIDTH SHEET) (W:SHEET-LINE-HEIGHT SHEET)
			     (W:SHEET-INSIDE-LEFT SHEET) (W:SHEET-CURSOR-Y SHEET)
			     (W:SHEET-CHAR-ALUF SHEET) SHEET))) 

(DEFUN INSERT-BLACK-BLOCK-DIAGRAM (BP)
  (SETQ *BLACK-BLOCK-DIAGRAM* (INSERT-DIAGRAM BP 'BLACK-BLOCK-DIAGRAM))) 

(DEFFLAVOR GREY-BLOCK-DIAGRAM () (LINE-DIAGRAM-MIXIN)) 

(DEFMETHOD (GREY-BLOCK-DIAGRAM :DRAW) (IGNORE SHEET)
  (W:PREPARE-SHEET (SHEET)
     (BITBLT W:ALU-SETA (W:SHEET-INSIDE-WIDTH SHEET) (W:SHEET-LINE-HEIGHT SHEET)
	     W:50%-GRAY 0 0
	     (W:SHEET-SCREEN-ARRAY SHEET)
	     (W:SHEET-INSIDE-LEFT SHEET) (W:SHEET-CURSOR-Y SHEET)))) 

(DEFUN INSERT-GREY-BLOCK-DIAGRAM (BP)
  (SETQ *GREY-BLOCK-DIAGRAM* (INSERT-DIAGRAM BP 'GREY-BLOCK-DIAGRAM))) 

(DEFFLAVOR RESTORABLE-LINE-DIAGRAM-MIXIN () ()
  (:INCLUDED-FLAVORS LINE-DIAGRAM-MIXIN)
  (:INIT-KEYWORDS :NUMBER-OF-LINES)
  (:DEFAULT-INIT-PLIST :NUMBER-OF-LINES 1)
  (:REQUIRED-METHODS :CONTENTS)) 

(DEFMETHOD (RESTORABLE-LINE-DIAGRAM-MIXIN :STRING-FOR-FILE) (LINE)
  (IF (SEND SELF :FIRST-LINE-P LINE)
      (FORMAT NIL "#~D ~A~%~A"
	      (SEND SELF :NUMBER-OF-LINES)
	      (TYPE-OF SELF)
	      (SEND SELF :CONTENTS LINE))
      (SEND SELF :CONTENTS LINE))) 

