;;; ZWEI keyboard macros -*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1980, Massachusetts Institute of Technology

(DEFVAR MACRO-ESCAPE-CHAR #\c-m-TERM
  "The character that tells a macro stream to define/execute a macro.") 
(DEFVAR MACRO-STREAM :UNBOUND
  "The stream to do I/O through, in a macro stream.") 
(DEFVAR MACRO-LEVEL :UNBOUND
  "Current depth in definition/execution of keyboard macros, in a macro stream.") 
(DEFVAR MACRO-UNREAD :UNBOUND
  "Holds an unread character, in a macro stream.") 
(DEFVAR MACRO-LEVEL-ARRAY :UNBOUND
  "Array of macro being defined/executed at each level, in a macro stream.") 
(DEFVAR MACRO-CURRENT-ARRAY :UNBOUND
  "Macro currently being executed.") 
(DEFVAR MACRO-PREVIOUS-ARRAY) 
(DEFVAR MACRO-READING NIL
  "T when executing a macro, in the macro stream.") 
(DEFVAR MACRO-REDIS-LEVEL -1
  "Macro level last displayed, in the macro stream.") 
(DEFVAR MACRO-OPERATIONS :UNBOUND
  "List of operations implemented by this macro stream.") 
(DEFVAR MACRO-INPUT-RECORD NIL
  "This is a circular buffer array holding the last 60. characters read from this macro stream.") 

(DEFSTRUCT (MACRO-ARRAY (:MAKE-ARRAY (:LENGTH 100))
			(:CONC-NAME NIL)
			(:CALLABLE-CONSTRUCTORS NIL)
			(:ALTERANT ALTER-MACRO-ARRAY)
			(:PREDICATE NIL)
			(:COPIER NIL)
			(:TYPE :ARRAY-LEADER))
  (MACRO-POSITION 0)				;Current position reading or writing
  (MACRO-LENGTH 0)				;Length of macro
  MACRO-COUNT					;Current repeat count for macro
  MACRO-DEFAULT-COUNT				;Initial value of MACRO-COUNT, or NIL if writing
  MACRO-NAME					;Name of macro as a string, or NIL if temporary.
  (MACRO-INSTALLED-ON-KEYS NIL)) 

;;; The following structure is used for the MACRO-A command.
;;; It is important that it be a LIST since that is how it is
;;; identified.
(DEFSTRUCT (MACRO-A (:CONC-NAME NIL)
		    (:CALLABLE-CONSTRUCTORS NIL)
		    (:ALTERANT ALTER-MACRO-A)
		    (:PREDICATE NIL)
		    (:COPIER NIL)
		    (:TYPE LIST))
  (MACRO-A-NAME '*A*)				;Symbol by which this is recognized.
  MACRO-A-VALUE					;Current value of the character.
  MACRO-A-STEP					;Number to increase VALUE by on each step.
  MACRO-A-INITIAL-VALUE)			;Initial current-value given by user.

(DEFUN MAKE-MACRO-STREAM (STREAM)
  "Create a keyboard-macro stream that does its I/O through STREAM.
The keyboard macro stream support definition and replay of macros."
  (LET-CLOSED ((MACRO-STREAM STREAM)
	       (MACRO-LEVEL -1)
	       (MACRO-UNREAD NIL)
	       (MACRO-LEVEL-ARRAY (MAKE-ARRAY 20))
	       (MACRO-CURRENT-ARRAY NIL)
	       (MACRO-PREVIOUS-ARRAY NIL)
	       (MACRO-INPUT-RECORD (MAKE-ARRAY 74 :LEADER-LIST '(74 0)))
	       (MACRO-OPERATIONS
		 (LET ((OPS (COPY-LIST (SEND STREAM :WHICH-OPERATIONS))))
		   (MAPC #'(LAMBDA (X) (SETQ OPS (DELETE X (THE LIST OPS) :TEST #'EQ)))
			 '(:TYI :UNTYI :LISTEN :CLEAR-INPUT :MACRO-LEVEL :MACRO-ERROR :MACRO-EXECUTE
				:LINE-IN :RUBOUT-HANDLER :READ-CHAR :UNREAD-CHAR))
		   `(:TYI :UNTYI :LISTEN :CLEAR-INPUT :MACRO-LEVEL :MACRO-ERROR :MACRO-EXECUTE :MACRO-PUSH
			  :MACRO-POP :MACRO-QUERY :MACRO-PREVIOUS-ARRAY :READ-CHAR :UNREAD-CHAR ,@OPS))))
    'MACRO-STREAM-IO)) 

(DEFVAR MACRO-ERROR-HOOK NIL
   "If non-NIL, funcalled by macro stream after :MACRO-ERROR operation.") 

(DEFVAR MACRO-POP-HOOK NIL
  "If non-NIL, funcalled by macro stream after :MACRO-POP operation.") 

;;;
;;; Modified :macro-execute method, to bring it in line with other code, mine and existing.
(DEFUN MACRO-STREAM-IO (OP &REST ARGS)
  ;;this was implemented previously as a defselect - rpm 6/23/86
  (CASE OP
    (:WHICH-OPERATIONS					;ARGS = ()       
     MACRO-OPERATIONS)
    ((:UNREAD-CHAR :UNREAD-ANY :UNTYI)			;ARGS = (CH)
     (DECF (ARRAY-LEADER MACRO-INPUT-RECORD 1))
     (IF (MINUSP (ARRAY-LEADER MACRO-INPUT-RECORD 1))
	 (SETF (ARRAY-LEADER MACRO-INPUT-RECORD 1)
	       (1- (ARRAY-TOTAL-SIZE MACRO-INPUT-RECORD))))
     (SETQ MACRO-UNREAD (CAR ARGS)))
    ((:READ-CHAR :READ-ANY :READ-MOUSE-OR-KBD :READ-CHAR-NO-HANG :READ-ANY-NO-HANG
		 :READ-MOUSE-OR-KBD-NO-HANG :TYI)
     (MACRO-STREAM-IO-READ OP ARGS))
    (:PLAYBACK						;ARGS = ()
     MACRO-INPUT-RECORD)
    (:WAIT-FOR-INPUT-WITH-TIMEOUT			;ARGS = (TIMEOUT)
     (COND (MACRO-UNREAD T)
	   ((OR MACRO-READING
		(NULL MACRO-CURRENT-ARRAY)
		(NULL (MACRO-DEFAULT-COUNT MACRO-CURRENT-ARRAY))
		(MEMBER (AREF MACRO-CURRENT-ARRAY (MACRO-POSITION MACRO-CURRENT-ARRAY))
			'(*SPACE* *MOUSE* *MICE* NIL) :TEST #'EQ))
	    (SEND MACRO-STREAM :WAIT-FOR-INPUT-WITH-TIMEOUT (CAR ARGS)))
	   (T T)))
    (:LISTEN						;ARGS = ()
     (COND (MACRO-UNREAD T)
	   ((OR MACRO-READING
		(NULL MACRO-CURRENT-ARRAY)
		(NULL (MACRO-DEFAULT-COUNT MACRO-CURRENT-ARRAY))
		(MEMBER (AREF MACRO-CURRENT-ARRAY (MACRO-POSITION MACRO-CURRENT-ARRAY))
			'(*SPACE* *MOUSE* *MICE* NIL) :TEST #'EQ))
	    (LISTEN MACRO-STREAM))
	   (T T)))
    (:MACRO-LEVEL					;ARGS = ()
     (1+ MACRO-LEVEL))
    (:MACRO-ERROR					;ARGS = ()
     ;; If executing a macro, stop and return T.  Else return NIL.
     (PROG1 (AND MACRO-CURRENT-ARRAY (MACRO-DEFAULT-COUNT MACRO-CURRENT-ARRAY))
	    (MACRO-STOP NIL)
	    (IF MACRO-ERROR-HOOK
		(FUNCALL MACRO-ERROR-HOOK))))
    (:CLEAR-INPUT					;ARGS = ()
     (MACRO-STOP NIL)
     (SETQ MACRO-UNREAD NIL)
     (SEND MACRO-STREAM :CLEAR-INPUT))
    (:MACRO-EXECUTE					;ARGS = (&OPTIONAL ARRAY TIMES)
     (LET ((ARRAY (FIRST ARGS))
	   (TIMES (SECOND ARGS)))
       ;; Execute the macro ARRAY, TIMES times. (0 means , nil means default)
       (COND ((OR ARRAY (SETQ ARRAY MACRO-PREVIOUS-ARRAY))
	      (MACRO-PUSH-LEVEL (MACRO-STORE ARRAY))
	      (COND ((NULL TIMES))
		    ;;made to work for both macros on keys and C-X E.
		    ((= TIMES 0)
		     (SETF (MACRO-COUNT ARRAY) '*REPEAT*))	;gsl 4-17-85
		    (T
		     (SETF (MACRO-COUNT ARRAY) TIMES))))
	     (T
	      (BEEP)))))
    (:MACRO-PUSH					;ARGS = (&OPTIONAL N APPEND-TO)
     ;; Start defining another macro.
     ;; N is number of most recent chars that were the command to do so.
     ;; APPEND-TO is a macro to append to, or NIL.
     (LET ((N (FIRST ARGS))
	   (APPEND-TO (SECOND ARGS)))
       (AND MACRO-CURRENT-ARRAY				
	    ;;Erase the command that caused this to happen
	    N
	    (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY)
		  (- (MACRO-POSITION MACRO-CURRENT-ARRAY) N)))
       (COND (APPEND-TO
	      (SETF (MACRO-DEFAULT-COUNT APPEND-TO) NIL)
	      (INCF (MACRO-POSITION APPEND-TO))))
       (MACRO-PUSH-LEVEL (MACRO-STORE (OR APPEND-TO T))
			 (NULL APPEND-TO))))
    (:MACRO-POP						;ARGS = (&OPTIONAL N TIMES)
     ;; Finished defining the innermost macro.
     ;; N is number of most recent chars that were the command to do so.
     ;; TIMES is number of times to execute (counting definition as 1).
     (LET ((N (FIRST ARGS))
	   (TIMES (SECOND ARGS)))
       (AND MACRO-CURRENT-ARRAY
	    N
	    (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY)
		  (- (MACRO-POSITION MACRO-CURRENT-ARRAY) N)))
       (PROG1
	 (MACRO-REPEAT TIMES)
	 (IF MACRO-POP-HOOK
	     (FUNCALL MACRO-POP-HOOK)))))
    (:MACRO-QUERY					;ARGS = (&OPTIONAL N)
     ;; Record that user is being queried now.
     (LET ((N (CAR ARGS)))
       (AND MACRO-CURRENT-ARRAY
	    N
	    (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY)
		  (- (MACRO-POSITION MACRO-CURRENT-ARRAY) N)))
       (MACRO-STORE '*SPACE*)))
    (:MACRO-PREVIOUS-ARRAY				;ARGS = ()
     MACRO-PREVIOUS-ARRAY)
    (OTHERWISE
     (MACRO-STREAM-DEFAULT-HANDLER OP ARGS)))) 

;;; Made this more general by adding except-at optional parameter. gsl. 4-14-85
(DEFUN QUERIED-GSL (ARRAY &OPTIONAL (EXCEPT-AT -1))
  "Checks a kbd macro for queries.  If except-at is given then a query there doesn't count."
  ;;This is specifically not used in macro-escape-char R, so that truely infinite macros may be created. gsl.
  (DO ((SIZE (MACRO-LENGTH ARRAY))
       (INDEX 0 (1+ INDEX))
       (FOUND NIL))
      ((OR (> INDEX SIZE)
	   (SETQ FOUND
		 (AND (/= INDEX EXCEPT-AT)
		      (MEMBER (AREF ARRAY INDEX) '(*SPACE* *MOUSE* *MICE*) :TEST #'EQ))))
       FOUND))) 

;;; Set to NIL so killed buffers can get garbage collected. This was probably just used 
;;; for debugging, anyway, since only PRINT-BUFFER-INPUT (which has no callers) uses it.
;;; Done 12-3-86 by rpm (from lmo).
(DEFPARAMETER RECORD-ALL-INPUT NIL
   "Non-NIL means record all input that comes through the macro stream, per buffer.") 

(DEFUN PRINT-BUFFER-INPUT (&OPTIONAL (NBUFS 12) (INT *INTERVAL*))
  "Print the last NBUFS buffers of saved input that applies to interval INT.
Each buffer is 400. characters worth.  NBUFS defaults to 10."
  (DOLIST (ARRAY (REVERSE (FIRSTN NBUFS (GET INT 'INPUT-LIST))))
    (DOTIMES (I (LENGTH ARRAY))
      (IF (CONSP (AREF ARRAY I))
	  (PRIN1 (AREF ARRAY I))
	  (FORMAT T "~:C" (AREF ARRAY I)))
      (WRITE-CHAR #\SPACE)))) 

(DEFVAR RECORD-INPUT-INTERVAL NIL) 

(DEFUN MACRO-STREAM-IO-READ (OP &OPTIONAL IGNORE)	; &optional eof for other :READs.
  (LET ((CHAR (COND (MACRO-UNREAD
		     (PROG1 MACRO-UNREAD
			    (SETQ MACRO-UNREAD NIL)))
		    (MACRO-READING
		     (MACRO-UPDATE-LEVEL)
		     (FUNCALL MACRO-STREAM OP))
		    (T
		     (MACRO-READ OP)))))
    (WHEN CHAR
      (WHEN RECORD-ALL-INPUT
	(UNLESS (EQ *INTERVAL* (WINDOW-INTERVAL *MINI-BUFFER-WINDOW*))
	  (SETQ RECORD-INPUT-INTERVAL *INTERVAL*))
	(LET ((INPUT-LIST (GET RECORD-INPUT-INTERVAL 'INPUT-LIST)))
	  (WHEN (OR (NULL INPUT-LIST)
		    (>= (LENGTH (CAR INPUT-LIST)) 620))
	    (SETF (GET RECORD-INPUT-INTERVAL 'INPUT-LIST)
		  (SETQ INPUT-LIST (CONS (MAKE-ARRAY 620 :TYPE ART-Q :FILL-POINTER 0)
					 INPUT-LIST))))
	  (VECTOR-PUSH CHAR (CAR INPUT-LIST))))
      (SETF (ARRAY-LEADER MACRO-INPUT-RECORD 1)
	    (REM (1+ (ARRAY-LEADER MACRO-INPUT-RECORD 1))
		 (ARRAY-TOTAL-SIZE MACRO-INPUT-RECORD)))
      (SETF (AREF MACRO-INPUT-RECORD (ARRAY-LEADER MACRO-INPUT-RECORD 1)) CHAR))
    CHAR)) 

(DEFUN MACRO-STREAM-DEFAULT-HANDLER (OP ARGS)
  (IF (AND (MEMBER OP MACRO-OPERATIONS :TEST #'EQ)
	   (NOT (MEMBER OP '(:SEND-IF-HANDLES :OPERATION-HANDLED-P) :TEST #'EQ)))
      (APPLY MACRO-STREAM OP ARGS)
      (STREAM-DEFAULT-HANDLER 'MACRO-STREAM-IO OP (CAR ARGS) (CDR ARGS))))

;;;
;;;  Fancy macro handling a la the Macro key on cadrs.
;;;  What a can of worms, with gems hidden among them. gsl.
;;;  -- I think it all works now.  But it really needs a thorough testing. gsl 4-17-85

(DEFUN MACRO-READ (&OPTIONAL (OP :READ-MOUSE-OR-KBD))
  (DO ((CH) (TEM) (NUMARG) (FLAG) (TEM2) (SUPPRESS))
      (NIL)
    (CATCH 'MACRO-LOOP
      (COND ((AND MACRO-CURRENT-ARRAY
		  (SETQ TEM2 (MACRO-DEFAULT-COUNT MACRO-CURRENT-ARRAY)))
	     (SETQ TEM (MACRO-POSITION MACRO-CURRENT-ARRAY)
		   CH (AREF MACRO-CURRENT-ARRAY TEM))
	     (COND ((EQ CH '*SPACE*) ;;We have found a macro query in a macro.
		    (FORMAT *TYPEIN-WINDOW* "Pausing at macro query.") ;;gsl 3-23-85
		    (CASE (W:READ-MOUSE-OR-KBD MACRO-STREAM)
			  (#\SPACE
			   (SETQ CH '*IGNORE*))
			  ((#\? #\HELP)
			   (FORMAT T "~&You are in an interactive macro.~@
                               ~2@TSpace - continues on,~@
                               ~2@TRubout - skips this one,~@
                               ~2@TPeriod - finishes this one,~@
                               ~2@T! - Eliminates this macro query this run unless~@
                               ~8@T infinite uninterrupted repetition results,~@
                               ~2@TClear-screen - refreshes the screen,~@
                               ~2@TControl-R - enters a typein macro level (~:C R exits),~@
                               ~2@TAnything else exits." MACRO-ESCAPE-CHAR)
			   (SEND *TYPEIN-WINDOW* :CLEAR-SCREEN)
			   (THROW 'MACRO-LOOP NIL))
			  (#\RUBOUT
			   (SETQ TEM (MACRO-LENGTH MACRO-CURRENT-ARRAY)
				 CH '*IGNORE*))
			  ((#\c-R #\c-\r)
			   (SETQ CH NIL)) ;;nil acts as a code to indicate we want to push a macro level, see cond below.
			  (#\PAGE         ;; Used to (return #\ff) gsl. 4-16-85
			   (REDISPLAY *WINDOW* :NONE)
			   (SEND *TYPEIN-WINDOW* :CLEAR-SCREEN)
			   (THROW 'MACRO-LOOP NIL))
			  (#\.
			   (SETF (MACRO-DEFAULT-COUNT MACRO-CURRENT-ARRAY) 0)
			   (SETF (MACRO-COUNT MACRO-CURRENT-ARRAY) 0)
			   (SETQ CH '*IGNORE*))
			  (#\!
			   (IF (OR (NOT (EQ (MACRO-COUNT MACRO-CURRENT-ARRAY) '*REPEAT*))
				   (AND (EQ (MACRO-COUNT MACRO-CURRENT-ARRAY) '*REPEAT*)
					(QUERIED-GSL MACRO-CURRENT-ARRAY TEM)))
			       (SETF (AREF MACRO-CURRENT-ARRAY TEM) '*RUN*))
			   (SETQ CH '*IGNORE*))
			  (OTHERWISE
			   (MACRO-STOP 1)
			   (SEND *TYPEIN-WINDOW* :CLEAR-SCREEN) ;;gsl 3-23-85
			   (THROW 'MACRO-LOOP NIL)))
		    (SEND *TYPEIN-WINDOW* :CLEAR-SCREEN)) ;;gsl 3-23-85
		   ((MEMBER CH '(*MOUSE* *MICE*) :TEST #'EQ)
		    ;;We have encountered a place the user used the mouse in the macro.
		    (AND (EQ CH '*MOUSE*)
			 (FORMAT *TYPEIN-WINDOW* "~&Use the mouse.~%"))
		    (SETQ CH (PROG1 (W:READ-MOUSE-OR-KBD MACRO-STREAM)
				    (IF (EQ CH '*MOUSE*)
					(SEND *TYPEIN-WINDOW* :CLEAR-SCREEN)))) ;;gsl 4-16-85
		    (COND ((CHAR-BIT CH :MOUSE)
			   (SETF (AREF MACRO-CURRENT-ARRAY TEM) '*MICE*)
			   (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY) ;;this & next 2 lines gsl. 4-14-85
				 (1+ (MACRO-POSITION MACRO-CURRENT-ARRAY)))
			   (RETURN `(:MOUSE-BUTTON ,CH ,*WINDOW* ,SYSTEM:MOUSE-X ,SYSTEM:MOUSE-Y)))
			  (T
			   (SETF (AREF MACRO-CURRENT-ARRAY TEM) '*MOUSE*)
			   (SETQ CH '*IGNORE*)))))
	     ;;This cond handles macro repetition or termination.
	     (COND ((AND (ZEROP TEM)
			 ;;Stop infinite repetitions at macro begin if either buffer end reached.
			 (EQ (MACRO-COUNT MACRO-CURRENT-ARRAY) '*REPEAT*)
			 ;; old check for tem2. gsl. 4-14-85
			 (MEMBER :MACRO-TERMINATE MACRO-OPERATIONS :TEST #'EQ)
			 (SEND MACRO-STREAM :MACRO-TERMINATE))
		    ;;check for buffer end.
		    (SETQ CH '*IGNORE*)
		    ;;Ending macro with M-C-term r needs this. gsl 4-14-85
		    (COND ((>= (SETQ MACRO-LEVEL (1- MACRO-LEVEL)) 0)
			   (SETQ MACRO-CURRENT-ARRAY (AREF MACRO-LEVEL-ARRAY MACRO-LEVEL)))
			  (T
			   (SETQ MACRO-CURRENT-ARRAY NIL))))
		   ((< TEM (MACRO-LENGTH MACRO-CURRENT-ARRAY))
		    ;;Go to next char in macro.
		    (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY) (1+ TEM)))
		   ((EQ (MACRO-COUNT MACRO-CURRENT-ARRAY) '*REPEAT*)
		    ;;Reset infinite macros. ;;gsl 4-17-85
		    (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY) 0))
		   ((> (SETQ TEM (1- (MACRO-COUNT MACRO-CURRENT-ARRAY))) 0)
		    ;;Others repeat here.
		    (SETF (MACRO-COUNT MACRO-CURRENT-ARRAY) TEM)
		    (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY) 0))
		   ((>= (SETQ MACRO-LEVEL (1- MACRO-LEVEL)) 0)
		    ;;Higher level macros pop here.
		    (SETQ MACRO-CURRENT-ARRAY (AREF MACRO-LEVEL-ARRAY MACRO-LEVEL)))
		   (T
		    ;;Done with top level macro here.
		    (SETQ MACRO-CURRENT-ARRAY NIL)))
	     (COND ((CHARACTERP CH)
		    (OR SUPPRESS (RETURN CH))) ;;Finally, do something with the macro character.
		   ((MEMBER CH '(*RUN* *IGNORE*) :TEST #'EQ))
		   ((AND (CONSP CH) (EQ (CAR CH) '*A*)) ;;Handles the macro-escape-char A command's form.
		    (LET ((X (MACRO-A-VALUE CH)))
		      (SETF (MACRO-A-VALUE CH) (INT-CHAR (+ (CHAR-INT X) (MACRO-A-STEP CH))))
		      (OR SUPPRESS (RETURN X))))
		   (T (MACRO-PUSH-LEVEL CH))))
	    ;;There is no current macro, so read a character from the keyboard.
	    (T
	     (MACRO-UPDATE-LEVEL)
	     (MULTIPLE-VALUE-SETQ (CH TEM) (FUNCALL MACRO-STREAM OP)) ;;macro stream is kbd input.
	     (COND (FLAG ;;This is only when the macro-escape-char has just been entered.
		    (OR (CHARACTERP CH)
			(MACRO-BARF))
		    (SETQ CH (CHAR-UPCASE CH))
		    (COND ((AND (CHAR<= #\0 CH #\9) ;;Numeric arg entered, stored in numarg.
				(SETQ NUMARG (+ (- (CHAR-CODE CH) (CHAR-CODE #\0))
						(* (OR NUMARG 0) 12)))))
			  (T ;;We are accepting a macro-escape-char command.
			   (IF (EQ FLAG :HELP)
			       (REDISPLAY *WINDOW* :NONE)) ;;gsl 4-14-85
			   (SETQ FLAG NIL)
			   (CASE CH
				 (#\END
				  (RETURN #\SPACE)) ;;gsl 3-23-85
				 (#\C
				  (SETQ TEM (MACRO-DO-READ "Macro to call: "))
				  (OR (SETQ TEM (GET TEM 'MACRO-STREAM-MACRO)) (MACRO-BARF))
				  (MACRO-STORE TEM)
				  (OR SUPPRESS (MACRO-PUSH-LEVEL TEM)))
				 (#\D
				  (SETQ SUPPRESS MACRO-LEVEL)
				  (MACRO-PUSH-LEVEL (MACRO-MAKE-NAMED-MACRO)))
				 (#\M
				  (MACRO-PUSH-LEVEL
				    (MACRO-STORE (MACRO-MAKE-NAMED-MACRO))))
				 (#\P
				  (MACRO-PUSH-LEVEL (MACRO-STORE)))
				 (#\R
				  (MACRO-REPEAT NUMARG)	
				  (AND (EQ SUPPRESS MACRO-LEVEL) (SETQ SUPPRESS NIL)))
				 (#\S
				  (MACRO-STOP NUMARG))
				 (#\T
				  (MACRO-PUSH-LEVEL (MACRO-STORE NIL)))
				 (#\U
				  (MACRO-PUSH-LEVEL NIL))
				 (#\SPACE
				  (MACRO-STORE '*SPACE*))
				 (#\A
				  (LET ((STR (MACRO-READ-STRING "Initial character (type a one-character string):")))
				    (OR (= (LENGTH STR) 1) (MACRO-BARF))
				    (LET ((VAL (AREF STR 0))
					  (NUM (MACRO-READ-NUMBER
						 "Amount by which to increase it (type a decimal number):")))
				      (MACRO-STORE
					(MAKE-MACRO-A MACRO-A-VALUE (INT-CHAR (+ (CHAR-INT VAL) NUM))
						      MACRO-A-STEP NUM
						      MACRO-A-INITIAL-VALUE VAL))
				      (OR SUPPRESS (RETURN VAL)))))
				 (#\HELP
				  (FORMAT
				    T "~&Macro commands are:~@
                                       ~2@TP - Push a level of macro,~@
                                       ~2@TR - End and Repeat arg times (0 means infinite times),~@
                                       ~2@TC - Call a macro by name,~@
                                       ~2@TS - Stop macro definition,~@
                                       ~2@TU - Allow typein now only, ~@
                                       ~2@TT - Allow typein in expansion too. (terminate typein with ~:C R)~@
                                       ~2@TM - Define a named macro,~@
                                       ~2@TD - Define a named macro but don't execute as building.~@
                                       ~2@TA - Enter a character,  then a numeric increment used for each macro iteration.~@
                                       ~2@TSpace - Enter macro query, ~@
                                       ~2@TEnd - Cancels ~0@*~:c prefix and refreshes screen.~@
                                       ~2@T (Arguments are digits following ~0@*~:c.)~@
                                       ~2@TNow type a macro command: "
				    MACRO-ESCAPE-CHAR)
				  (SETQ FLAG :HELP))
				 (OTHERWISE
				  (MACRO-BARF))))))
		   ((EQ CH MACRO-ESCAPE-CHAR)
		    (SETQ FLAG T
			  NUMARG NIL))
		   (T
		    (COND ((OR (NUMBERP CH) (CHARACTERP CH))
			   (MACRO-STORE (IF (CHAR-BIT CH :MOUSE) '*MOUSE* (INT-CHAR CH))))
			  ((AND (CONSP CH) (EQ (CAR CH) :MOUSE-BUTTON)) ;;added. gsl 4-14-85
			   (MACRO-STORE '*MOUSE*)))
		    (OR SUPPRESS
			(RETURN (VALUES CH TEM)))))))))) 

(DEFUN MACRO-PUSH-LEVEL (MAC &OPTIONAL (RESET-POSITION T))
  (COND (MAC
	 (AND (SYMBOLP MAC)
	      (SETQ MAC (GET MAC 'MACRO-STREAM-MACRO)))
	 (OR (ARRAYP MAC)
	     (MACRO-BARF))))
  (SETQ MACRO-LEVEL (1+ MACRO-LEVEL)
	MACRO-CURRENT-ARRAY MAC)
  (AND (>= MACRO-LEVEL (ARRAY-TOTAL-SIZE MACRO-LEVEL-ARRAY))
       (ADJUST-ARRAY MACRO-LEVEL-ARRAY (FLOOR (* MACRO-LEVEL 3) 2)))
  (SETF (AREF MACRO-LEVEL-ARRAY MACRO-LEVEL) MAC)
  (COND (MAC
	 (AND RESET-POSITION
	      (SETF (MACRO-POSITION MAC) 0))
	 (SETF (MACRO-COUNT MAC) (MACRO-DEFAULT-COUNT MAC))
	 (DO ((I 0 (1+ I))
	      (X)
	      (LIM (MACRO-LENGTH MAC)))
	     ((> I LIM))
	   (SETQ X (AREF MAC I))
	   (COND ((EQ '*RUN* X)
		  (SETF (AREF MAC I) '*SPACE*))
		 ((EQ '*MICE* X)
		  (SETF (AREF MAC I) '*MOUSE*))
		 ((AND (CONSP X) (EQ (CAR X) '*A*))
		  (SETF (MACRO-A-VALUE X) (MACRO-A-INITIAL-VALUE X)))))))) 

(DEFUN MACRO-STORE (&OPTIONAL (THING T))
  (AND (EQ THING T) (SETQ THING (MAKE-MACRO-ARRAY)))
  (AND MACRO-CURRENT-ARRAY (VECTOR-PUSH-EXTEND THING MACRO-CURRENT-ARRAY))
  THING) 

(DEFUN MACRO-BARF ()
  (BEEP)
  (THROW 'MACRO-LOOP NIL)) 

;;;
;;; I've increased arg by one throughout, to duplicate the typein when the macro was defined. gsl. 4-16-85
;;; this also makes it agree with doc. on macro-escape-char R. gsl.

(DEFUN MACRO-REPEAT (ARG &AUX (TEM -1))
  ;;This function is called only by the macro-escape-char R command, which terminates macros and typein,
  ;; and :macro-pop which is called only by com-end-kbd-macro.
  (AND (< MACRO-LEVEL 0) (MACRO-BARF))
  (COND (MACRO-CURRENT-ARRAY
	 (COND ((NULL ARG)
		(SETQ ARG 0)) ;;this cond added. gsl 4-17-85.
	       ((= ARG 0)
		(SETQ ARG '*REPEAT*))) ;;This looks stupid, but should provide better default. gsl.
	 (SETF (MACRO-DEFAULT-COUNT MACRO-CURRENT-ARRAY)
	       (IF (NUMBERP ARG)
		   (1+ ARG) ;;the 1 duplicates the typein when the macro was defined.
		   ARG))
	 (SETQ TEM (1- (MACRO-POSITION MACRO-CURRENT-ARRAY)))
	 (SETF (MACRO-LENGTH MACRO-CURRENT-ARRAY) TEM)
	 (SETQ MACRO-PREVIOUS-ARRAY MACRO-CURRENT-ARRAY)))
  (COND ((AND (>= TEM 0) (NUMBERP ARG) (> ARG 0))
	 (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY) 0)
	 (SETF (MACRO-COUNT MACRO-CURRENT-ARRAY) ARG))
	((EQ ARG '*REPEAT*)
	 (SETF (MACRO-POSITION MACRO-CURRENT-ARRAY) 0)
	 (SETF (MACRO-COUNT MACRO-CURRENT-ARRAY) '*REPEAT*))
	((>= (SETQ MACRO-LEVEL (1- MACRO-LEVEL)) 0)
	 (SETQ MACRO-CURRENT-ARRAY (AREF MACRO-LEVEL-ARRAY MACRO-LEVEL)))
	(T
	 (SETQ MACRO-CURRENT-ARRAY NIL)))) 

(DEFUN MACRO-MAKE-NAMED-MACRO (&AUX NAME MAC)
  (SETQ NAME (MACRO-DO-READ "Name of macro to define: "))
  (OR (SYMBOLP NAME) (MACRO-BARF))
  (SETQ MAC (MAKE-MACRO-ARRAY))
  (SETF (GET NAME 'MACRO-STREAM-MACRO) MAC)
  (SETF (MACRO-NAME MAC) (STRING NAME))
  ;;Let it be called as a command. gsl 4-21-85
  (LET ((MACRO-CLOSURE (MAKE-MACRO-COMMAND NAME NIL)))
    (SETF (GET NAME 'MACRO-STREAM-MACRO-COMMAND) MACRO-CLOSURE)
    (MAKE-MACRO-A-COMMAND NAME MACRO-CLOSURE))
  MAC) 

(DEFUN MAKE-MACRO-A-COMMAND (MACRO-NAME MACRO-CLOSURE)
  "This puts the macro command on the command alist under macro-name."
  (COND ((DOCUMENTATION MACRO-NAME 'FUNCTION))
	(T (SETF (DOCUMENTATION MACRO-NAME 'FUNCTION) "A Keyboard Macro")))
  (LET* ((NAME (MAKE-COMMAND-NAME MACRO-NAME))
	 (AENTRY (ASSOC MACRO-NAME *COMMAND-ALIST* :TEST 'EQUALP)))
    (SETF (GET MACRO-NAME 'COMMAND-NAME) NAME)
    (IF AENTRY
	(UNLESS (EQ MACRO-CLOSURE (CDR AENTRY))
	  (WHEN (FQUERY NIL "The command name ~S is currently set up to call ~S;
redefine it to call ~S? "
			NAME (CDR AENTRY) MACRO-NAME)
	    (SETF (CDR AENTRY) MACRO-CLOSURE)))
	(PUSH (CONS NAME MACRO-CLOSURE) *COMMAND-ALIST*)))) 

(DEFUN MACRO-READ-STRING (STR &AUX (MACRO-READING T) (MACRO-REDIS-LEVEL -1))
  (IF (MEMBER :READ-MACRO-LINE MACRO-OPERATIONS :TEST #'EQ)
      (SEND MACRO-STREAM :READ-MACRO-LINE STR)
      (PROGN
	(PRINC STR MACRO-STREAM)
	(READ-LINE MACRO-STREAM)))) 

(DEFUN MACRO-DO-READ (STR)
  (INTERN (STRING-UPCASE (STRING-TRIM '(#\SPACE #\TAB) (MACRO-READ-STRING STR)))
	  "")) 

(DEFUN MACRO-READ-NUMBER (STR)
  (LET ((NUM (READ-FROM-STRING (MACRO-READ-STRING STR) T)))
    (OR (NUMBERP NUM) (MACRO-BARF))
    NUM)) 

(DEFUN MACRO-STOP (NUM)
  "Stop defining or executing the innermost NUM macros."
  (SETQ MACRO-LEVEL (MAX -1 (- MACRO-LEVEL (OR NUM 200000)))
	MACRO-CURRENT-ARRAY (AND (>= MACRO-LEVEL 0) (AREF MACRO-LEVEL-ARRAY MACRO-LEVEL)))) 

(DEFUN MACRO-UPDATE-LEVEL ()
  "Maybe tell the i/o stream to redisplay the macro level (if it has changed)."
  (COND ((AND (/= MACRO-LEVEL MACRO-REDIS-LEVEL)
	      (MEMBER :SET-MACRO-LEVEL MACRO-OPERATIONS :TEST #'EQ))
	 (SETQ MACRO-REDIS-LEVEL MACRO-LEVEL)
	 (SEND MACRO-STREAM :SET-MACRO-LEVEL
	       (AND (NOT (MINUSP MACRO-LEVEL))
		    (FORMAT () "~D" (1+ MACRO-LEVEL))))))) 

;;; Handy things for saving out macros on disk and editing them
(DEFMACRO DEFINE-KEYBOARD-MACRO (NAME (COUNT) . EXPANSION)
  `(DEFINE-KEYBOARD-MACRO-1 ',NAME ,(OR COUNT 1) ',(COPY-LIST EXPANSION))) 

(DEFUN DEFINE-KEYBOARD-MACRO-1 (NAME COUNT EXPANSION &AUX MACRO-ARRAY (LEN 0) STRING)
  (SETQ STRING (STRING NAME)
	NAME (INTERN STRING ""))
  (DOLIST (THING EXPANSION)
    (IF (STRINGP THING)
	(SETQ LEN (+ LEN (LENGTH THING)))
	(SETQ LEN (1+ LEN))))
  (SETQ MACRO-ARRAY (MAKE-MACRO-ARRAY :MAKE-ARRAY (:LENGTH LEN)
				      MACRO-LENGTH (1- LEN)
				      MACRO-DEFAULT-COUNT COUNT
				      MACRO-NAME STRING))
  (DOLIST (THING EXPANSION)
    (IF (STRINGP THING)
	(LOOP FOR C BEING THE ARRAY-ELEMENTS OF THING
	      DO (VECTOR-PUSH C MACRO-ARRAY))
	(PROGN
	  (COND ((CHARACTERP THING))
		((STRING-EQUAL THING '*INPUT*)
		 (SETQ THING NIL))
		((STRING-EQUAL THING '*SPACE*)
		 (SETQ THING '*SPACE*))
		((STRING-EQUAL THING '*MOUSE*)
		 (SETQ THING '*MOUSE*))
		((STRING-EQUAL THING '*MICE*)
		 (SETQ THING '*MICE*))
		(T
		 (FERROR NIL "~S is not a known macro expansion element." THING)))
	  (VECTOR-PUSH THING MACRO-ARRAY))))
  (SETF (GET NAME 'MACRO-STREAM-MACRO) MACRO-ARRAY)
  NAME) 

(DEFUN PRINT-KEYBOARD-MACRO-DEFINITION (STREAM NAME &OPTIONAL MACRO-ARRAY)
  (LET ((*PACKAGE* (FIND-PACKAGE (SYMBOL-PACKAGE 'FOO)))
	(*PRINT-BASE* :CHARACTER))
    (GRIND-TOP-LEVEL (GET-KEYBOARD-MACRO-DEFINITION NAME MACRO-ARRAY) 137 STREAM))) 

(DEFUN GET-KEYBOARD-MACRO-DEFINITION (NAME MACRO-ARRAY)
  "Return a Lisp form that would define NAME as a keyboard macro.
The definition is MACRO-ARRAY or what NAME is actually defined as."
  (OR MACRO-ARRAY (SETQ MACRO-ARRAY (GET NAME 'MACRO-STREAM-MACRO)))
  (SETQ NAME (INTERN (STRING NAME) (SYMBOL-PACKAGE 'FOO)))
  (DO ((I 0 (1+ I))
       (LEN (1+ (MACRO-LENGTH MACRO-ARRAY)))
       (THING)
       (STATE NIL)
       (LIST NIL)
       (STRING (MAKE-ARRAY 12 :ELEMENT-TYPE 'STRING-CHAR :LEADER-LENGTH 1)))
      ((>= I LEN)
       `(DEFINE-KEYBOARD-MACRO ,NAME () ,@(NREVERSE LIST)))
    (SETQ THING (AREF MACRO-ARRAY I))
    (COND ((OR (SYMBOLP THING) (NOT (ZEROP (TV:CHAR-CMSH-BITS THING))))
	   (COND (STATE
		  (PUSH (STRING-APPEND STRING) LIST)
		  (SETQ STATE NIL)))
	   (COND ((CHARACTERP THING))
		 ((NULL THING)
		  (SETQ THING '*INPUT*)))
	   (PUSH THING LIST))
	  (T
	   (COND ((NOT STATE)
		  (STORE-ARRAY-LEADER 0 STRING 0)
		  (SETQ STATE T)))
	   (VECTOR-PUSH-EXTEND THING STRING))))) 

;Define CHARACTER as a radix.
(DEFPROP CHARACTER PRINC-BASE-CHARACTER SI:PRINC-FUNCTION) 
(DEFUN PRINC-BASE-CHARACTER (-N STREAM)
  (FORMAT STREAM "~@C" (- -N))) 
