;;;  -*- Mode:Common-Lisp; Package:ZWEI; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;;
;;;  The commands in this file are supported by TI, but are not considered
;;;  important to include in the band, so they're autoloading.
;;;
;;;  From modes:

(DEFMAJOR COM-MIDAS-MODE MIDAS-MODE "MIDAS"
	  "Sets things up for editing assembly language code." ()
  (SETQ *COMMENT-COLUMN* 400)
  (SETQ *COMMENT-START* ";")
  (SETQ *PARAGRAPH-DELIMITER-LIST* NIL)
  (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\.)
  (SET-COMTAB *MODE-COMTAB* '(#\TAB COM-INSERT-TAB
			      #\c-m-A COM-GO-TO-AC-FIELD
			      #\c-m-E COM-GO-TO-ADDRESS-FIELD
			      #\c-m-D COM-KILL-TERMINATED-WORD
			      #\c-m-N COM-GO-TO-NEXT-LABEL
			      #\c-m-P COM-GO-TO-PREVIOUS-LABEL))) 

(DEFPROP MIDAS-MODE T ALL-UPPERCASE) 

(DEFCOM COM-KILL-TERMINATED-WORD "Kill a word and the following character.
If the word is followed by a CRLF, the CRLF is not killed." ()
  (LET ((BP (OR (FORWARD-WORD (POINT)) (BARF))))
    (OR (CHAR= (BP-CH-CHAR BP) #\NEWLINE)
	(SETQ BP (FORWARD-CHAR BP 1 T)))
    (KILL-INTERVAL-ARG (POINT) BP 1))
  (SETQ *CURRENT-COMMAND-TYPE* 'KILL)
  DIS-TEXT) 

(DEFCOM COM-GO-TO-PREVIOUS-LABEL "Put point after last label.
With an argument, moves after the argth previous label." (KM)
  (LET ((*NUMERIC-ARG* (- *NUMERIC-ARG*)))
    (COM-GO-TO-NEXT-LABEL))) 

(DEFCOM COM-GO-TO-NEXT-LABEL "Put point after the next label.
With an argument, moves after the argth label." (KM)
  (LET ((ARG (ABS *NUMERIC-ARG*))
	(SIGN (IF (MINUSP *NUMERIC-ARG*) -1 1))
	(POINT (POINT)))
    (DO ((I 0 (1+ I))
	 (BP (BEG-LINE POINT)))
	(NIL)
      (DO NIL (NIL)
	(OR (MEMBER (BP-CH-CHAR BP) '(#\* #\SPACE #\TAB #\NEWLINE) :TEST #'EQ)
	    (STRING-EQUAL (BP-LINE BP) *COMMENT-START* :START1 0 :START2 0 :END1 (LENGTH *COMMENT-START*))
	    (RETURN NIL))
	(OR (SETQ BP (BEG-LINE BP SIGN))
	    (BARF)))
      (COND ((>= I ARG)
	     (LET ((LINE (BP-LINE BP)))
	       (MOVE-BP BP LINE (OR (STRING-SEARCH-SET *BLANKS* LINE)
				    (LINE-LENGTH LINE))))
	     (COND ((IF (MINUSP SIGN)
			(BP-< BP POINT)
			(BP-< POINT BP))
		    (MOVE-BP POINT BP)
		    (RETURN NIL)))))))
  DIS-BPS) 

(DEFCOM COM-GO-TO-ADDRESS-FIELD "Put point before the address field." (KM)
   (GO-TO-ADDRESS-OR-AC-FIELD-INTERNAL T)) 

(DEFCOM COM-GO-TO-AC-FIELD "Put point before the accumulator field." (KM)
   (GO-TO-ADDRESS-OR-AC-FIELD-INTERNAL NIL)) 

(DEFUN GO-TO-ADDRESS-OR-AC-FIELD-INTERNAL (ADDRESS-P &AUX LINE BP)
  (SETQ LINE (BP-LINE (POINT))
	BP (OR (FORWARD-WORD (BEG-LINE (POINT))) (BARF)))
  (OR (MEMBER (BP-CH-CHAR BP) '(#\: #\= #\_) :TEST #'EQ)
      (SETQ BP (BEG-LINE BP)))
  (SETQ BP (OR (FORWARD-TO-WORD BP) (BARF)))
  (MOVE-BP BP LINE (OR (STRING-SEARCH-SET *BLANKS* LINE (BP-INDEX BP))
		       (LINE-LENGTH LINE)))
  (LET ((BP1 (FORWARD-OVER *BLANKS* BP)))
    (OR (CHAR= (BP-CH-CHAR BP1) #\;)
	(SETQ BP BP1)))
  (COND ((MEMBER (MAKE-CHAR (BP-CHAR-BEFORE BP)) *BLANKS* :TEST #'EQ)
	 (AND ADDRESS-P
	      (LET ((I (STRING-SEARCH-SET '(#\SPACE #\, #\; #\/) LINE (BP-INDEX BP))))
		(AND I
		     (CHAR-EQUAL (AREF LINE I) #\,)
		     (MOVE-BP BP LINE (1+ I)))))
	 (MOVE-BP (POINT) BP)
	 DIS-BPS)
	(T
	 (MOVE-BP (POINT) (INSERT BP " "))
	 DIS-TEXT))) 

(DEFMAJOR COM-TEX-MODE TEX-MODE "TeX"
	  "Set up things for editing TeX files
Gives paren syntax to {[]}, makes \\ escape, % comment, tab like text." ()
  (SET-CHAR-SYNTAX LIST-COMMENT *MODE-LIST-SYNTAX-TABLE* #\%)
  (SET-CHAR-SYNTAX LIST-DELIMITER *MODE-LIST-SYNTAX-TABLE* #\/)
  (SET-CHAR-SYNTAX LIST-DELIMITER *MODE-LIST-SYNTAX-TABLE* #\;)
  (SET-CHAR-SYNTAX LIST-OPEN *MODE-LIST-SYNTAX-TABLE* #\[)
  (SET-CHAR-SYNTAX LIST-SLASH *MODE-LIST-SYNTAX-TABLE* #\\)
  (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\\)
  (SET-CHAR-SYNTAX LIST-CLOSE *MODE-LIST-SYNTAX-TABLE* #\])
  (SET-CHAR-SYNTAX LIST-OPEN *MODE-LIST-SYNTAX-TABLE* #\{)
  (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\{)
  (SET-CHAR-SYNTAX LIST-DELIMITER *MODE-LIST-SYNTAX-TABLE* #\|)
  (SET-CHAR-SYNTAX LIST-CLOSE *MODE-LIST-SYNTAX-TABLE* #\})
  (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\})
  (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\_)
  (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\')
  (SET-COMTAB *MODE-COMTAB* '(#\TAB COM-STUPID-TAB))
  (SETQ *COMMENT-START* "%")
  (SETQ *COMMENT-BEGIN* "% ")
  (SETQ *COMMENT-END* "")) 

(DEFPROP TEX-MODE :TEXT EDITING-TYPE) 
