;;; -*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*- 

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1980, Massachusetts Institute of Technology

;;; Change History :
;;; 01-10-89 MAY Removed c-m-space key binding to COM-WORD-ABBREV-MODE since
;;;              it conflicts with conficts with Move To Previous Point,
;;;              and is available by any chars in INITIALIZE-WORD-ABBREV-TABLE
;;;              or ctrl-meta-x Expand Only when not in Word Abbrev Mode.
;;;              SPR`s 1399, 8723, 5683, 9188.
;;; 12-13-88 MAY Changed functions to make Word Abbrev Mode work properly :
;;;              EXPAND-ABBREV, (spr 8089 and 6715), LIST-WORD-ABBREV-1, 
;;;              COM-LIST-WORD-ABBREVS, COM-DEFINE-WORD-ABBREVS, 
;;; 		 and WRITE-QWABL (spr 9065)

;;; Functions to implement major and minor modes.

;;; Two specials, bound at every ZWEI invocation, are used to keep track of
;;; modes.  *MODE-LIST* is a list, each of whose elements speaks of a mode that
;;; is in effect.  The most recent mode entered is at the head of the list, and
;;; the least recent at the end.  Modes are turned on and off in a stack-like
;;; fashion, to prevent a bug in which keys end up redefined even after all
;;; modes are turned off.

;;; Each mode is known by a symbol, such as AUTO-FILL-MODE.  The symbol has a
;;; MODE property, the value of which is a list of Lisp forms.  These forms are
;;; very restricted; they must be SETQ, ASET, SET-COMTAB, PUSH, SET-CHAR-SYNTAX,
;;; SET-SYNTAX-TABLE-INDIRECTION or COMMAND-HOOK forms.  They are interpreted
;;; by ZWEI rather than Lisp, because ZWEI figures out how to undo them, and
;;; generates a list of forms to undo their effects.  An element of the
;;; *MODE-LIST* consists of a 2-list of the mode symbol, and the ZWEI-generated
;;; list of undoing forms.

;;; The symbol *MAJOR-MODE* is the symbol for the major mode currently in
;;; effect, if any (e.g. ZETALISP-MODE).  When a new major mode is entered, the old
;;; one is exited automatically.

;;; The value of a mode symbol is a string used to display that mode, and it
;;; has a MAJOR-MODE-P property which is T if it is a major-mode symbol.

;;; Modes are defined with DEFMAJOR or DEFMINOR.
;;; The syntax is the nearly the same:

;;; (DEFMAJOR <command-name> <mode-symbol> <mode-name> <command-documentation>
;;; <command-options> . <body>)

;;; (DEFMINOR <command-name> <mode-symbol> <mode-name> <mode-line-position>
;;; <command-documentation> <command-options> . <body>)

;;; The <mode-name> is the value of the <mode-symbol>.  The
;;; <mode-line-position> is a fixnum that determines where the name goes on
;;; the mode line, it is stored on the MODE-LINE-POSITION property of the
;;; symbol.  The other three things are used to form a DEFCOM for a command
;;; to enter or exit the mode (exiting happens when the argument is zero).
;;; The reason that the <command-name> comes first, and isn't next to the
;;; <command-documentation>, etc., is so that M-. can find the command.
;;; The <body> should consist of only the restricted set of forms mentioned
;;; above.  If you need to use other forms, change the function below to
;;; know how to produce an undo-form.

(DEFUN TURN-OFF-MODE (MODE-SYMBOL)
  "Turn off mode MODE-SYMBOL.  Mode symbol is, for example, LISP-MODE.
Does nothing if the mode is not on."
  (LET ((MODE-ELEMENT (ASSOC MODE-SYMBOL *MODE-LIST* :TEST #'EQ)))
    (COND ((NOT (NULL MODE-ELEMENT))
	   ;; We must first turn off all more-recently added modes,
	   ;; then turn off this mode, and then turn the more-recently added
	   ;; modes on in reverse order.  To accomplish the last, backward step
	   ;; more elegantly, we use a recursive subfunction.
	   (TURN-OFF-MODE-UNDERSCORE *MODE-LIST* MODE-ELEMENT))))
  (SETQ *MODE-NAME-LIST* (DELETE MODE-SYMBOL (THE LIST *MODE-NAME-LIST*) :TEST #'EQ))) 

;;; Internal recursive subfunction of TURN-OFF-MODE.
(DEFUN TURN-OFF-MODE-UNDERSCORE (LIST ELEMENT)
  (COND ((EQ (CAR LIST) ELEMENT)
	 (EVAL-UNDO-LIST (SECOND ELEMENT))
	 (SETQ *MODE-LIST* (DELETE ELEMENT (THE LIST *MODE-LIST*) :TEST #'EQ)))
	(T
	 (EVAL-UNDO-LIST (SECOND (CAR LIST)))
	 (TURN-OFF-MODE-UNDERSCORE (CDR LIST) ELEMENT)
	 (SETF (SECOND (CAR LIST)) (EVALUATE-FORMING-UNDO-LIST (GET (CAAR LIST) 'MODE)))))) 

;; Macro for suggestions added by rpm 11-26-86.
(DEFUN TURN-ON-MODE (MODE-SYMBOL)
  "Turn on mode MODE-SYMBOL.  Mode symbol is, for example, LISP-MODE.
Does nothing if the mode is already on."
  (SYS:WITH-SUGGESTIONS-MENUS-FOR ZWEI:TURN-ON-MODE
    (COND ((NULL (ASSOC MODE-SYMBOL *MODE-LIST* :TEST #'EQ))
	   (COND ((GET MODE-SYMBOL 'MAJOR-MODE-P)
		  (SETQ *MAJOR-MODE* MODE-SYMBOL)))
	   (PUSH (LIST MODE-SYMBOL (EVALUATE-FORMING-UNDO-LIST (GET MODE-SYMBOL 'MODE)))
		 *MODE-LIST*)))
    (COND ((NULL (MEMBER MODE-SYMBOL *MODE-NAME-LIST* :TEST #'EQ))
	   (PUSH MODE-SYMBOL *MODE-NAME-LIST*)))
    (COND ((GET MODE-SYMBOL 'MAJOR-MODE-P)
	   (DOLIST (MINOR *INITIAL-MINOR-MODES*)
	     (IF (ATOM MINOR)
		 (TURN-ON-MODE MINOR)
		 (IF (MEMBER MODE-SYMBOL (CAR MINOR) :TEST #'EQ)
		     (DOLIST (MINOR (CDR MINOR))
		       (TURN-ON-MODE MINOR)))))))
    (LET ((HOOK (GET MODE-SYMBOL 'MODE-HOOK-SYMBOL)))
      (AND HOOK
	   (BOUNDP HOOK)
	   (FUNCALL (SYMBOL-VALUE HOOK))))
    (SORT *MODE-NAME-LIST* #'(LAMBDA (X Y) (< (GET X 'MODE-LINE-POSITION)
					      (GET Y 'MODE-LINE-POSITION)))))) 

;;; Take a list of forms, evaluate them, and produce a list of forms which,
;;; when evaluated, will undo the effects of the evaluatation of the original list.
;;; These forms are very restricted; they must be SETQ, ASET, SET-COMTAB,
;;; PUSH, SET-CHAR-SYNTAX, SET-SYNTAX-TABLE-INDIRECTION, COMMAND-HOOK,
;;; TURN-ZETALISP-ON, TURN-COMMON-LISP-ON, or SET-MOUSE-DOCUMENTATION forms.
(DEFUN EVALUATE-FORMING-UNDO-LIST (FORM-LIST)
  (LET ((RESULT NIL))
    (DOLIST (FORM FORM-LIST)
      (CASE (CAR FORM)
	(SETQ
	 (OR (GET (SECOND FORM) 'MODE-SETTABLE-P)
	     (FERROR NIL "Illegal to SETQ ~A inside a mode macro" (SECOND FORM)))
	 (PUSH `(SETQ ,(SECOND FORM) ',(SYMBOL-VALUE (SECOND FORM))) RESULT)
	 (EVAL FORM))
	(ASET
	 (PUSH `(SETF (AREF ,@(CDDR FORM)) ',(EVAL `(AREF ,@(CDDR FORM)))) RESULT)  
	 (EVAL FORM))
	(SET-COMTAB
	 ;; Knowledge of how to reverse SET-COMTAB is kept in the COMTAB > file.
	 (PUSH (MAKE-SET-COMTAB-UNDO-LIST FORM) RESULT)
	 (EVAL FORM))
	(PUSH
	 (LET ((THING (EVAL (SECOND FORM))))
	   (PUSH `(SETF ,(THIRD FORM) (DELETE ',THING ,(THE LIST (THIRD FORM)) :TEST #'EQ))
		 RESULT)	   
	   (EVAL `(PUSH ',THING ,(THIRD FORM)))))
	(COMMAND-HOOK
	 (LET ((THING (EVAL (SECOND FORM))))
	   (PUSH `(SETF ,(THIRD FORM) (DELETE ',THING ,(THE LIST (THIRD FORM)) :TEST #'EQ))
		 RESULT)
	   (COMMAND-HOOK THING (THIRD FORM))))
	(SET-CHAR-SYNTAX
	 (LET ((SYNTAX-TABLE (SYMBOL-VALUE (THIRD FORM)))
	       (CHAR (FOURTH FORM)))
	   (PUSH `(SET-CHAR-SYNTAX ,(CHAR-SYNTAX CHAR SYNTAX-TABLE) ',SYNTAX-TABLE ,CHAR)
		 RESULT))
	 (EVAL FORM))
	(SET-MODE-LINE-LIST
	 (PUSH `(SET-MODE-LINE-LIST ',*MODE-LINE-LIST*) RESULT)
	 (EVAL FORM))
	(SET-SYNTAX-TABLE-INDIRECTION
	 (LET ((OF (SYMBOL-VALUE (SECOND FORM)))
	       (TO (SYMBOL-VALUE (THIRD FORM))))
	   (PUSH `(RPLACA ',OF ',(CAR OF)) RESULT)
	   (RPLACA OF TO)))
	(PROGN
	 (EVAL FORM))
	((TURN-ZETALISP-ON TURN-COMMON-LISP-ON)
	 (PUSH `(SETF (LISP-MODE) ,(LISP-MODE)) RESULT) ; Eval-undo-list, below, recognises this setf.
	 (SI::EVAL1 FORM))
	(SET-MOUSE-DOCUMENTATION
	 (PUSH `(SET-MOUSE-DOCUMENTATION ',(SEND *WINDOW* :MOUSE-DOCUMENTATION-STRING))
	       RESULT)
	 (SET-MOUSE-DOCUMENTATION))
	(OTHERWISE
	 (FERROR NIL "The form ~S cannot be used in a mode, because I can't invert it." FORM))))
    RESULT)) 

;;;
;;;  Internal hack to run eval on most of the undo forms for a mode (since they
;;;  *are* in Zetalisp), but si:eval1 for (setf (lisp-mode) <whatever>), because
;;;  eval binds the things that the setf wants to change (lisp-mode variables).
(DEFUN EVAL-UNDO-LIST (UNDO-LIST)
  (LOOP FOR FORM IN UNDO-LIST
	IF (AND (EQ (FIRST FORM) 'SETF)
		(EQUAL (SECOND FORM) '(LISP-MODE)))
	DO (SI::EVAL1 FORM)
	ELSE DO (EVAL FORM))) 

;;;  Set mouse doc to a list, if given, or to the comtab-mouse-prompt.
(DEFUN SET-MOUSE-DOCUMENTATION (&OPTIONAL DOC-LIST)
  (IF DOC-LIST
      (SEND *WINDOW* :SET-MOUSE-DOCUMENTATION-STRING DOC-LIST)
      (SEND *WINDOW* :SET-MOUSE-DOCUMENTATION-STRING (COMTAB-MOUSE-PROMPT *COMTAB*)))) 

;;; Turn off all modes.  For ZMACS.
;;; Macro for suggestions added by rpm 11-26-86.
(DEFUN UN-SET-MODES ()
  "Turns off all modes that are on."
  (SYS:WITH-SUGGESTIONS-MENUS-FOR ZWEI:UN-SET-MODES
    (DOLIST (V *LOCAL-BOUND-VARIABLES*)
      (SYSTEM:DELETE-BINDING-FROM-CLOSURE *EDITOR* V))
    (DOLIST (V *LOCAL-VARIABLES*)
      (UNLESS (MEMBER V *LOCAL-BOUND-VARIABLES* :TEST #'EQ)
	(SET V (SYMEVAL-GLOBALLY V))))
    (DOLIST (L *MODE-LIST*)
      (EVAL-UNDO-LIST (SECOND L)))
    (SETQ *MODE-NAME-LIST* NIL
	  *MODE-LIST* NIL)
    (SETQ *LOCAL-VARIABLES* NIL)
    (SETQ *LOCAL-BOUND-VARIABLES* NIL))) 

;;; Turn on a saved set of modes.  For ZMACS.
(DEFUN SET-MODES (MODE-LIST MAJOR-MODE &OPTIONAL LOCAL-VARIABLES)
  "Turn on the modes in MODE-LIST, and major mode MAJOR-MODE.
MAJOR-MODE, and the elements of MODE-LIST, are mode symbols, like LISP-MODE."
  (SET-MODES-UNDERSCORE MODE-LIST)
  (SETQ *MAJOR-MODE* MAJOR-MODE)
  (TURN-ON-MODE *MAJOR-MODE*)
  (DOLIST (V LOCAL-VARIABLES)
    (MAKE-LOCAL-VARIABLE (CAR V) NIL (CDR V))
    (SET (CAR V) (CADR V)))) 

(DEFUN SET-MODES-UNDERSCORE (LIST)
  (COND ((NULL LIST)
	 NIL)
	(T
	 (SET-MODES-UNDERSCORE (CDR LIST))
	 (PUSH (LIST (CAAR LIST) (EVALUATE-FORMING-UNDO-LIST (GET (CAAR LIST) 'MODE)))
	       *MODE-LIST*)
	 (SETQ *MODE-NAME-LIST* (NCONC *MODE-NAME-LIST* (CONS (CAAR LIST) NIL)))))) 

(DEFUN MAKE-LOCAL-VARIABLE (VARIABLE &OPTIONAL (VALUE NIL VALUEP) XVCELL)
  "Make VARIABLE be bound locally in the editor that is running.
If XVCELL is non-NIL, we use it as the closure value cell for the variable
 if the variable does not already have one.
If XVCELL is NIL, then if VALUE was specified, that becomes the
 new value of the variable; otherwise, the value does not change."
  ;; Make sure that our editor closure includes *LOCAL-VARIABLES*
  ;; This is a kludge to get around the fact that in system 98
  ;; all the closure variable lists omit *LOCAL-VARIABLES*.
  ;; The following four lines are not needed after system 98.
  (UNLESS *LOCAL-VARIABLES*
    (SYSTEM:INSERT-BINDING-IN-CLOSURE *EDITOR* '*LOCAL-VARIABLES* '*EDITOR*))
  (UNLESS *LOCAL-BOUND-VARIABLES*
    (SYSTEM:INSERT-BINDING-IN-CLOSURE *EDITOR* '*LOCAL-BOUND-VARIABLES* '*EDITOR*))
  (UNLESS (MEMBER VARIABLE *LOCAL-VARIABLES* :TEST #'EQ)
    (PUSH VARIABLE *LOCAL-VARIABLES*)
    (IF (SYSTEM:INSERT-BINDING-IN-CLOSURE *EDITOR* VARIABLE '*EDITOR* XVCELL)
	(PUSH VARIABLE *LOCAL-BOUND-VARIABLES*)
	(WHEN XVCELL
	  (SET VARIABLE (CAR XVCELL)))))
  (UNLESS XVCELL
    (WHEN VALUEP
      (SET VARIABLE VALUE)))
  (WHEN (TYPEP *INTERVAL* 'ZMACS-BUFFER)
    ;; Put this variable, with correct vcell or value,
    ;; on the buffer's local variable list.  Don't wait for a buffer switch.
    (LET ((AELT (ASSOC VARIABLE (BUFFER-SAVED-LOCAL-VARIABLES *INTERVAL*) :TEST #'EQ)))
      (UNLESS AELT
	(SETQ AELT (CONS VARIABLE NIL))
	(PUSH AELT (BUFFER-SAVED-LOCAL-VARIABLES *INTERVAL*)))
      (IF (MEMBER VARIABLE *LOCAL-BOUND-VARIABLES* :TEST #'EQ)
	  (SETF (CDR AELT) (SYS:%P-CONTENTS-AS-LOCATIVE (VALUE-CELL-LOCATION VARIABLE)))
	  (SETF (CDR AELT) (LIST (SYMBOL-VALUE VARIABLE))))))) 

(DEFUN KILL-LOCAL-VARIABLE (VARIABLE)
  "Make VARIABLE no longer be bound locally in the editor that is running.
Its value reverts to the global value."
  (WHEN (TYPEP *INTERVAL* 'ZMACS-BUFFER)
    (SETF (BUFFER-SAVED-LOCAL-VARIABLES *INTERVAL*)
	  (DELETE VARIABLE (BUFFER-SAVED-LOCAL-VARIABLES *INTERVAL*) :KEY 'CAR)))
  (WHEN (MEMBER VARIABLE *LOCAL-BOUND-VARIABLES* :TEST #'EQ)
    (SYSTEM:DELETE-BINDING-FROM-CLOSURE *EDITOR* VARIABLE)
    (SETQ *LOCAL-BOUND-VARIABLES* (DELETE VARIABLE
					  (THE LIST *LOCAL-BOUND-VARIABLES*)
					  :TEST #'EQ)))
  (SETQ *LOCAL-VARIABLES* (DELETE VARIABLE
				  (THE LIST *LOCAL-VARIABLES*)
				  :TEST #'EQ))) 

;; For the mode line list.
(DEFUN NAME-OF-MAJOR-MODE ()
  "Return the pretty name of the current major mode."
  (SYMBOL-VALUE *MAJOR-MODE*)) 

(DEFUN GET-FILE-MAJOR-MODE (MODE-PROP)
  "Given a keyword such as :COMMON-LISP, return the major mode symbol such as COMMON-LISP-MODE."
  (WHEN (AND MODE-PROP (SYMBOLP MODE-PROP))
    (SETQ MODE-PROP (OR (CDR (ASSOC MODE-PROP *MAJOR-MODE-TRANSLATIONS* :TEST #'EQ))
			MODE-PROP))
    (LET ((MODE (FIND-SYMBOL (STRING-APPEND (SYMBOL-NAME MODE-PROP) "-MODE")
			     #.(PACKAGE-NAME *PACKAGE*)))) ;;  Why not "ZWEI"?  Beats me.
      (COND ((AND (BOUNDP MODE)               ;;  Normal case.
		  (STRING-EQUAL (SYMBOL-VALUE MODE)
				(SYMBOL-NAME MODE-PROP)))
	     MODE)
	    (:ELSE                            ;;  Abnormal case.
	     (LET ((COMMAND (FIND-SYMBOL (STRING-APPEND "COM-" (SYMBOL-NAME MODE-PROP) "-MODE")
					 #.(PACKAGE-NAME *PACKAGE*))))
	       (COND ((AND COMMAND            ;;  Autoloading function exists, load its file.
			   (GET COMMAND 'AUTOLOADING-FILENAME))
		      (IF (EQ :LIBRARY (GET COMMAND 'AUTOLOADING-SUPPORT-TYPE))
			  (FORMAT *QUERY-IO* "~&~A is a library mode;  autoloading its definition."
				  MODE-PROP)
			  (FORMAT *QUERY-IO* "~&~A is an unsupported mode;  autoloading its definition."
				  MODE-PROP))
		      (let ((inhibit-fdefine-warnings t))
			(LOAD (GET COMMAND 'AUTOLOADING-FILENAME)
			      :VERBOSE NIL
			      :SET-DEFAULT-PATHNAME NIL))
		      (SETQ MODE (FIND-SYMBOL (STRING-APPEND (SYMBOL-NAME MODE-PROP) "-MODE")
					      #.(PACKAGE-NAME *PACKAGE*)))
		      (IF (AND (BOUNDP MODE)
			       (STRING-EQUAL (SYMBOL-VALUE MODE)
					     (SYMBOL-NAME MODE-PROP)))
			  MODE))
		     (:ELSE NIL)))))))) 

(DEFUN STICKY-MODE-LIST ()
  "Return a list of minor modes currently on that should be sticky.
Sticky means not turned off by changing major mode."
  (AND (BOUNDP '*MODE-LIST*)			;Somewhat of a kludge
       (DO ((MODES *MODE-LIST* (CDR MODES))
	    (NMODES NIL))
	   ((NULL MODES)
	    (NREVERSE NMODES))
	 (AND (MEMBER (CAAR MODES) *STICKY-MINOR-MODES* :TEST #'EQ)
	      (PUSH (CAR MODES) NMODES))))) 

(DEFUN TURN-ON-EMACS-MODE ()
  (OR (MEMBER 'EMACS-MODE *INITIAL-MINOR-MODES* :TEST #'EQ)
      (LOGIN-SETQ *INITIAL-MINOR-MODES* (CONS 'EMACS-MODE *INITIAL-MINOR-MODES*)))) 

(DEFUN TURN-OFF-EMACS-MODE ()
  (LOGIN-SETQ *INITIAL-MINOR-MODES* (REMOVE 'EMACS-MODE
					    (THE LIST *INITIAL-MINOR-MODES*)
					    :TEST #'EQ))) 

;;; Major modes.

(DEFMAJOR COM-ZETALISP-MODE ZETALISP-MODE "Zetalisp"
	  "Sets things up for editing Zetalisp code." ()
  (SETQ *SPACE-INDENT-FLAG* T)
  (SETQ *PARAGRAPH-DELIMITER-LIST* '(#\. #\SPACE #\TAB #\"))
  (SETQ *COMMENT-START* 'LISP-FIND-COMMENT-START-AND-END)
  (SET-CHAR-SYNTAX LIST-SLASH *MODE-LIST-SYNTAX-TABLE* #\/)
  (SET-CHAR-SYNTAX LIST-ALPHABETIC *MODE-LIST-SYNTAX-TABLE* #\\)
  (SET-COMTAB *MODE-COMTAB*
	      '(#\TAB COM-INDENT-FOR-LISP
		#\RUBOUT COM-TAB-HACKING-RUBOUT
		#\c-RUBOUT COM-RUBOUT
		#\m-Z COM-COMPILE-AND-EXIT
		#\c-m-Z COM-EVALUATE-AND-EXIT))
  (TURN-ZETALISP-ON)) 

(DEFPROP ZETALISP-MODE T ALL-UPPERCASE) 
(DEFPROP ZETALISP-MODE :LISP EDITING-TYPE) 


(DEFMAJOR COM-COMMON-LISP-MODE COMMON-LISP-MODE "Common-Lisp"
	  "Sets things up for editing Common Lisp code." ()
  (SETQ *SPACE-INDENT-FLAG* T)
  (SETQ *PARAGRAPH-DELIMITER-LIST* '(#\. #\SPACE #\TAB #\"))
  (SETQ *COMMENT-START* 'LISP-FIND-COMMENT-START-AND-END)
  (SET-CHAR-SYNTAX LIST-SLASH *MODE-LIST-SYNTAX-TABLE* #\\)
  (SET-CHAR-SYNTAX LIST-ALPHABETIC *MODE-LIST-SYNTAX-TABLE* #\/)
  (SET-COMTAB *MODE-COMTAB*
	      '(#\TAB COM-INDENT-FOR-LISP
		#\RUBOUT COM-TAB-HACKING-RUBOUT
		#\c-RUBOUT COM-RUBOUT
		#\m-Z COM-COMPILE-AND-EXIT
		#\c-m-Z COM-EVALUATE-AND-EXIT))
  (TURN-COMMON-LISP-ON)) 

(DEFPROP COMMON-LISP-MODE T ALL-UPPERCASE) 
(DEFPROP COMMON-LISP-MODE :LISP EDITING-TYPE) 


(DEFMAJOR COM-TEXT-MODE TEXT-MODE "Text"
   "Sets things up for editing English text.
Puts Tab-To-Tab-Stop on Tab." ()
   (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\_)
   (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\')
   (SET-COMTAB *MODE-COMTAB* '(#\TAB COM-TAB-TO-TAB-STOP))) 

(DEFPROP TEXT-MODE :TEXT EDITING-TYPE) 


(DEFMAJOR COM-FUNDAMENTAL-MODE FUNDAMENTAL-MODE "Fundamental"
   "Return to ZWEI's fundamental mode." ()) 

(DEFMAJOR COM-MACSYMA-MODE MACSYMA-MODE "Macsyma"
	  "Enter a mode for editing Macsyma code.
Modifies the delimiter dispatch tables appropriately for Macsyma syntax,
makes comment delimiters /* and */.  Tab is Indent Nested." ()
  (SET-COMTAB *MODE-COMTAB*
	      '(#\TAB COM-INDENT-NESTED))
  ;; Tab hacking rubout.
  (SETQ *SPACE-INDENT-FLAG* T)
  (SETQ *PARAGRAPH-DELIMITER-LIST* NIL)
  (SETQ *COMMENT-COLUMN* (* 50 6))
  (SETQ *COMMENT-START* "/*")
  (SETQ *COMMENT-BEGIN* "/* ")
  (SETQ *COMMENT-END* "*/")
  (PROGN
    (OR (BOUNDP '*MACSYMA-LIST-SYNTAX-TABLE*)
	(SETQ *MACSYMA-LIST-SYNTAX-TABLE*
	      (MAKE-SYNTAX-TABLE *MACSYMA-LIST-SYNTAX-LIST*))))
  (SETQ *MODE-LIST-SYNTAX-TABLE* *MACSYMA-LIST-SYNTAX-TABLE*)
  (SET-CHAR-SYNTAX WORD-ALPHABETIC *MODE-WORD-SYNTAX-TABLE* #\?)
  ;; Also does something like make right bracket point at right paren?
  ) 

(SETQ *MACSYMA-LIST-SYNTAX-LIST*
      '((40 LIST-ALPHABETIC)
	LIST-DELIMITER		 ;040 space
	LIST-DELIMITER		 ;041 ! ***
	LIST-DOUBLE-QUOTE	 ;042 "     "
	LIST-DELIMITER		 ;043 # ***
	LIST-DELIMITER		 ;044 $ ***
	LIST-ALPHABETIC		 ;045 %
	LIST-DELIMITER		 ;046 & ***
	LIST-SINGLE-QUOTE	 ;047 '
	LIST-OPEN		 ;050 (
	LIST-CLOSE		 ;051 )
	LIST-DELIMITER		 ;052 * ***
	LIST-DELIMITER		 ;053 + ***
	LIST-DELIMITER		 ;054 , ***
	LIST-DELIMITER		 ;055 - ***
	LIST-DELIMITER		 ;056 . ***
	LIST-DELIMITER		 ;057 / ***
	(12 LIST-ALPHABETIC)			;;DIGITS
	LIST-DELIMITER		 ;072 : ***
	LIST-DELIMITER		 ;073 ; ***
	LIST-DELIMITER		 ;074 < ***
	LIST-DELIMITER		 ;075 = ***
	LIST-DELIMITER		 ;076 > ***
	LIST-ALPHABETIC		 ;077 ?
	LIST-DELIMITER		 ;100 @ ***
	(32 LIST-ALPHABETIC)	 		;;UPPERCASE LETTERS
	LIST-OPEN		 ;133 [ ***
	LIST-SLASH		 ;134 \ ***
	LIST-CLOSE		 ;135 ] ***
	LIST-DELIMITER		 ;136 ^ ***
	LIST-DELIMITER		 ;137 _ ***
	LIST-DELIMITER		 ;140 ` ***
	(32 LIST-ALPHABETIC)			;;LOWERCASE LETTERS
	LIST-OPEN		 ;173 { ***
	LIST-DELIMITER		 ;174 | ***        |
	LIST-CLOSE		 ;175 } ***
	LIST-DELIMITER		 ;176 ~ ***
	LIST-ALPHABETIC		 ;177 integral ???
	LIST-ALPHABETIC		 ;200 null character
	LIST-DELIMITER		 ;201 break
	LIST-DELIMITER		 ;202 clear
	LIST-DELIMITER		 ;203 call
	LIST-DELIMITER		 ;204 escape (NOT altmode!)
	LIST-DELIMITER		 ;205 backnext
	LIST-DELIMITER		 ;206 help
	LIST-DELIMITER		 ;207 rubout
	LIST-ALPHABETIC		 ;210 bs
	LIST-DELIMITER		 ;211 tab
	LIST-DELIMITER		 ;212 line
	LIST-DELIMITER		 ;213 vt
	LIST-DELIMITER		 ;214 form = newpage
	LIST-DELIMITER		 ;215 return = newline
	(162 LIST-ALPHABETIC))) 


;;; Minor modes.

(DEFMINOR COM-RETURN-INDENTS-MODE RETURN-INDENTS-MODE "Return" 1
	  "Minor mode in which Return indents and Line does not.
A positive argument turns the mode on, zero turns it off;
no argument toggles." ()
  (SETQ *INDENT-NEW-LINE-NEW-LINE-FUNCTION* 'COM-INSERT-CRS)
  (SET-COMTAB *MODE-COMTAB* '
	      (#\NEWLINE COM-INDENT-NEW-LINE
	       #\LINEFEED COM-INSERT-CRS))) 

(DEFMINOR COM-ATOM-WORD-MODE ATOM-WORD-MODE "Atom" 1
   "Minor mode in which all word commands act on Lisp atoms.
A positive argument turns the mode on, zero turns it off;
no argument toggles." ()
  (SET-SYNTAX-TABLE-INDIRECTION *MODE-WORD-SYNTAX-TABLE* *ATOM-WORD-SYNTAX-TABLE*)) 

;;; Gets a single character from the user.  If HIGHBITSP is true, does not
;;; strip the control and meta bis.
(DEFUN GET-ECHO-CHAR (PROMPT HIGHBITSP &AUX CHAR)
  (DISCARD-LAST-PROMPT)
  (ADD-PROMPT PROMPT)
  (SETQ CHAR (INPUT-WITH-PROMPTS *STANDARD-INPUT* :TYI))
  (OR HIGHBITSP (SETQ CHAR (MAKE-CHAR CHAR)))
  CHAR) 

(DEFUN PROCESS-PREFIX-COMMAND-CHAR (KEY &AUX VALUE)
  (SETQ VALUE (PROCESS-COMMAND-CHAR KEY))
  (IF (EQ VALUE :ARGUMENT)
      VALUE
      DIS-NONE)) 

(DEFUN DOCUMENT-PREFIX-CHAR (COMMAND IGNORE OP &AUX COLNUM)
  (SETQ COLNUM (CDR (ASSOC COMMAND '((COM-PREFIX-CONTROL . 1)
				     (COM-PREFIX-META . 2)
				     (COM-PREFIX-CONTROL-META . 3))
			   :TEST #'EQ)))
  (CASE OP
	(:NAME (GET COMMAND 'COMMAND-NAME))
	(:SHORT (FORMAT T "Set the ~[Control~;Meta~;Control-Meta~] prefix." (1- COLNUM)))
	(:FULL (FORMAT T "Set the ~[Control~;Meta~;Control-Meta~] prefix.
Make the next character act as if it were typed with ~[CTRL~;META~;CTRL and META~]
held down, just as if you were on a losing terminal that doesn't
support all of the wonderful keys that we cleverly provide
on these marvelous keyboards.
Type a subcommand to document (or \"*\" for all): " (1- COLNUM) (1- COLNUM))
	       (LET ((CHAR (READ-CHAR)))
		 (COND ((CHAR= CHAR #\*)
			(FORMAT T "~2%The following ~[Control~;Meta~;Control-Meta~]- commands are availible:~%"
				(1- COLNUM))
			(DO ((I 0 (1+ I)))
			    ((>= I 220))
			  (PRINT-SHORT-DOC-FOR-TABLE (CODE-CHAR I COLNUM) *COMTAB* 3)))
		       (T
			(SETQ CHAR (MAKE-CHAR CHAR COLNUM))
			(FORMAT T "~:C~2%" CHAR)
			(DOCUMENT-KEY CHAR *COMTAB*))))))) 

(DEFCOM COM-UNIVERSAL-ARGUMENT "Sets argument or multiplies it by four.
Followed by digits, uses them to specify the
argument for the command after the digits.
Not followed by digits, multiplies the argument by four." ()
  (SETQ *NUMERIC-ARG-P* :CONTROL-U)
  (DO ((FIRSTP T NIL)
       (MINUSP NIL)
       (DIGITP NIL)
       (NUM 1)
       (CHAR))
      (NIL)
    (UNLESS (LISTEN)
      (REDISPLAY-ALL-WINDOWS))
    (SETQ CHAR (INPUT-WITH-PROMPTS *STANDARD-INPUT* :TYI))
    (COND ((AND FIRSTP (CHAR= CHAR #\-))
	   (SETQ MINUSP T
		 *NUMERIC-ARG-P* :SIGN))
	  ((AND (CHAR>= CHAR #\0)
		(CHAR<= CHAR #\9))
	   (COND (DIGITP  ;extract digits from characters
		  (SETQ NUM (+ (- (CHAR-CODE CHAR) 60) (* NUM 12))))
		 (T
		  (SETQ NUM (- (CHAR-CODE CHAR) 60)
			*NUMERIC-ARG-P* :DIGITS
			DIGITP T))))
	  (T
	   (COND ((OR MINUSP DIGITP)
		  (SETQ *NUMERIC-ARG* (IF MINUSP (- NUM) NUM)))
		 (T
		  (SETQ *NUMERIC-ARG* (* 4 *NUMERIC-ARG*))))
	   (UNREAD-CHAR CHAR)
	   (DISCARD-LAST-PROMPT)
	   (RETURN :ARGUMENT))))) 


(DEFMINOR COM-ANY-BRACKET-MODE ANY-BRACKET-MODE "Bracket" 4
	  "Minor mode in which List commands treat [{}] like parens.
This makes them easier to balance, etc.
A positive argument turns the mode on, zero turns it off;
no argument toggles." ()
  (SET-CHAR-SYNTAX LIST-OPEN *MODE-LIST-SYNTAX-TABLE* #\[)
  (SET-CHAR-SYNTAX LIST-OPEN *MODE-LIST-SYNTAX-TABLE* #\{)
  (SET-CHAR-SYNTAX LIST-CLOSE *MODE-LIST-SYNTAX-TABLE* #\})
  (SET-CHAR-SYNTAX LIST-CLOSE *MODE-LIST-SYNTAX-TABLE* #\])) 

(DEFMINOR COM-AUTO-FILL-MODE AUTO-FILL-MODE "Fill" 2
   "Minor mode in which insertion fills text.
A positive argument turns the mode on, zero turns it off;
no argument toggles." ()
  (COMMAND-HOOK 'AUTO-FILL-HOOK *POST-COMMAND-HOOK*)) 

(DEFPROP AUTO-FILL-HOOK 20 COMMAND-HOOK-PRIORITY) 
(DEFUN AUTO-FILL-HOOK (CHAR &AUX BP)
  (AND (MEMBER CHAR *AUTO-FILL-ACTIVATION-CHARACTERS* :TEST #'EQ)
       (NOT *NUMERIC-ARG-P*)
       (NEQ *INTERVAL* (WINDOW-INTERVAL *MINI-BUFFER-WINDOW*))
       (LET ((LINE (BP-LINE (POINT)))
	     (FILL-COLUMN *FILL-COLUMN*))
	 (AND (CHAR= CHAR #\NEWLINE)
	      (SETQ LINE (LINE-PREVIOUS LINE)))
	 (SETQ BP (DO ((SHEET (WINDOW-SHEET *WINDOW*))
		       (LEN (1+ (OR (POSITION #\SPACE
					      (THE STRING (STRING LINE))
					      :FROM-END T
					      :TEST-NOT #'CHAR-EQUAL)
				    -1)))
		       (POS 0)
		       (CHAR-POS 0 CP)
		       (CP))
		      ((= CHAR-POS LEN)
		       NIL)
		    (SETQ CP (OR (POSITION #\SPACE
					   (THE STRING (STRING LINE))
					   :START (1+ CHAR-POS)
					   :TEST #'CHAR-EQUAL)
				 LEN)
			  POS (W:SHEET-STRING-LENGTH SHEET LINE CHAR-POS CP NIL NIL POS))
		    (AND (> POS FILL-COLUMN)
			 (> CHAR-POS 0)
			 (RETURN (CREATE-BP LINE CHAR-POS))))))
       (WITH-BP (PT (POINT) ':MOVES)	 ;Save point
	 (MOVE-BP (POINT) BP)
	 (LET ((LINE (BP-LINE BP))
	       (LINE2 NIL))
	   (COND ((OR (END-LINE-P (FORWARD-OVER *BLANKS* BP))
		      (EQ LINE (BP-LINE (INTERVAL-LAST-BP *INTERVAL*)))
		      (> (+ (STRING-WIDTH LINE (BP-INDEX BP))
			    (STRING-WIDTH (SETQ LINE2 (LINE-NEXT LINE))))
			 *FILL-COLUMN*)
		      ;; If in Lispish mode and not inside a comment.
		      (AND (EQ (GET *MAJOR-MODE* 'EDITING-TYPE) :LISP)
			   (> (OR (FIND-COMMENT-START LINE)
				  (1+ (LINE-LENGTH LINE)))
			      (BP-INDEX BP))))
		  (DELETE-INTERVAL BP (FORWARD-CHAR BP))
		  (LET ((AT-POINT-P (BP-= PT BP)))
		    (MUST-REDISPLAY *WINDOW* (COM-INDENT-NEW-COMMENT-LINE))
		    (AND AT-POINT-P (MOVE-BP PT (POINT)))))
		 (T
		  (MULTIPLE-VALUE-BIND (COMMENT-START COMMENT-END)
		      (FIND-COMMENT-START LINE)
		    ;; Move some text from the end of the line onto the next line.
		    ;; If the next line starts a paragraph,
		    ;; or is a solitary blank line,
		    ;; make a new line to move the text onto.
		    (COND ((AND COMMENT-START
				(< COMMENT-START (BP-INDEX BP)))
			   ;; If we are breaking inside a comment, make a new line
			   ;; unless the next one is also a comment line
			   ;; and the comment starts at the same hpos
			   ;; and with the same comment-starting string.
			   (MULTIPLE-VALUE-BIND (LINE2-COMMENT-START LINE2-COMMENT-END)
			       (AND LINE2 (FIND-COMMENT-START LINE2))
			     (WHEN (OR (NOT LINE2-COMMENT-START)
				       (/= (BP-VIRTUAL-INDENTATION (CREATE-BP LINE2 LINE2-COMMENT-START)
								   (WINDOW-SHEET *WINDOW*))
					   (BP-VIRTUAL-INDENTATION (CREATE-BP LINE COMMENT-START)
								   (WINDOW-SHEET *WINDOW*)))
				       (/= (- COMMENT-END COMMENT-START)
					   (- LINE2-COMMENT-END LINE2-COMMENT-START))
				       (NOT (STRING-EQUAL LINE LINE2
							  :START1 COMMENT-START
							  :START2 LINE2-COMMENT-START
							  :END1 COMMENT-END
							  :END2 LINE2-COMMENT-END)))
			       (INSERT (CREATE-BP LINE2 0) #\NEWLINE)
			       (SETQ LINE2 (LINE-PREVIOUS LINE2)))))
			  ((IF (LINE-BLANK-P LINE2)
			       (NOT (OR (EQ LINE2 (BP-LINE (INTERVAL-LAST-BP *INTERVAL*)))
					(LINE-BLANK-OR-DIAGRAM-P (LINE-NEXT LINE2))))
			       (BP-AT-PARAGRAPH-TERMINATOR (CREATE-BP LINE2 0)))
			   (INSERT (CREATE-BP LINE2 0) #\NEWLINE)
			   (SETQ LINE2 (LINE-PREVIOUS LINE2))))
		    ;; Don't include leading blanks in the stuff to move to the next line.
		    (MOVE-BP BP (FORWARD-OVER *BLANKS* BP))
		    ;; IDX is beginning of text to move to next line.
		    (LET ((IDX (BP-INDEX BP))
			  (START-BP (CREATE-BP LINE2 0)))
		      (COND ((AND COMMENT-START
				  (< COMMENT-START (BP-INDEX BP)))
			     ;; If stuff to move down is inside a comment,
			     ;; must move it to inside a comment.
			     (MULTIPLE-VALUE-BIND (ST END)
				 (FIND-COMMENT-START LINE2)
			       (IF END
				   (MOVE-BP START-BP LINE2 END)
				   (PROGN
				     (INDENT-TO START-BP (BP-INDENTATION (CREATE-BP LINE COMMENT-START)))
				     (MOVE-BP START-BP LINE2 (LENGTH LINE2))
				     (SETF (VALUES ST END) (FIND-COMMENT-START LINE))
				     (INSERT-MOVING START-BP (NSUBSTRING LINE ST END))))))
			    (T
			     ;; Make sure the following line starts with the fill prefix
			     ;; (in case we just created it and it is empty.)
			     (OR (LOOKING-AT START-BP *FILL-PREFIX*)
				 (INSERT (CREATE-BP LINE2 0) *FILL-PREFIX*))
			     (MOVE-BP START-BP (FORWARD-OVER *BLANKS* START-BP))))
		      ;; Insert the stuff from the previous line.
		      (LET* ((AFTER-BP (INSERT START-BP (NSUBSTRING LINE IDX)))
			     (BACK-ONE-BP (FORWARD-CHAR AFTER-BP -1)))
			;; Make sure there is a space between the new stuff and
			;; what was already on the following line.
			(OR (BP-LOOKING-AT-LIST AFTER-BP *BLANKS*)
			    (BP-LOOKING-AT-LIST BACK-ONE-BP *BLANKS*)
			    (END-LINE-P AFTER-BP)
			    (INSERT AFTER-BP " "))
			;; If it ends a sentence, want two spaces after it.
			(AND (BP-LOOKING-AT-LIST BACK-ONE-BP *FILL-EXTRA-SPACE-LIST*)
			     (NOT (LOOKING-AT AFTER-BP "  "))
			     (NOT (END-LINE-P AFTER-BP))
			     (INSERT AFTER-BP " ")))
		      ;; If Point was within the text moved, keep it with that text.
		      (AND (EQ (BP-LINE PT) (BP-LINE BP))
			   (>= (SETQ IDX (- (BP-INDEX PT) IDX)) 0)
			   (MOVE-BP PT LINE2 (+ (BP-INDEX START-BP) IDX))))
		    ;; Delete what we moved from the previous line, plus any blanks before it.
		    (DELETE-INTERVAL (BACKWARD-OVER *BLANKS* BP) (END-OF-LINE LINE))
		    (MUST-REDISPLAY *WINDOW* DIS-TEXT)))))
	 (MOVE-BP (POINT) PT)))) 

(DEFPROP AUTO-FILL-HOOK DOCUMENT-AUTO-FILL-HOOK HOOK-DOCUMENTATION-FUNCTION) 

(DEFUN DOCUMENT-AUTO-FILL-HOOK (IGNORE CHAR)
  (AND (MEMBER CHAR *AUTO-FILL-ACTIVATION-CHARACTERS* :TEST #'EQ)
       (PRINC "With no numeric argument, auto fill line if needed.
"))) 

(DEFMINOR COM-OVERWRITE-MODE OVERWRITE-MODE "Overwrite" 4
   "Minor mode in which normal typing replaces existing text.
A positive argument turns the mode on, zero turns it off;
no argument toggles." ()
  (SETQ *STANDARD-COMMAND* 'COM-SELF-OVERWRITE)) 

(DEFCOM COM-SELF-OVERWRITE "Replace the character at point with the character typed.
At the end of a line, inserts instead of replacing the newline." ()
  ;;Added font and numeric arg handling. -rpm 7-11-86
  (LET* ((PT (POINT))
	 (CHARS-TO-EOL (- (LINE-LENGTH (BP-LINE PT)) (BP-INDEX PT))))
    (OR (ZEROP CHARS-TO-EOL)
	(DELETE-INTERVAL PT (FORWARD-CHAR PT (MIN CHARS-TO-EOL *NUMERIC-ARG*) T)))
    (DOTIMES (I *NUMERIC-ARG*)
      (INSERT-MOVING PT (IN-CURRENT-FONT *LAST-COMMAND-CHAR*)))	
    (VALUES DIS-LINE (BP-LINE PT) (- (BP-INDEX PT) *NUMERIC-ARG*)))) 

;;; Word abbrev mode

(DEFUN INITIALIZE-WORD-ABBREV-TABLE ()
  (LET ((INIT " ~@#;$%^&*()-_=+[]\\|:'`\"{},<.>/?!

"))
    (SETQ *WORD-ABBREV-TABLE* (MAKE-ARRAY 400 :ELEMENT-TYPE 'BIT))
    (DO ((I 0 (1+ I))
	 (LIM (LENGTH INIT)))
	((>= I LIM))
      (SETF (AREF *WORD-ABBREV-TABLE* (AREF INIT I)) 1)))) 

(DEFPROP EXPAND-ABBREV-HOOK 10 COMMAND-HOOK-PRIORITY) 
(DEFUN EXPAND-ABBREV-HOOK (IGNORE)
  (AND (EXPAND-P *LAST-COMMAND-CHAR*)
       (NOT *NUMERIC-ARG-P*)
       (EXPAND-ABBREV)
       (MUST-REDISPLAY *WINDOW* DIS-TEXT))) 

(DEFPROP EXPAND-ABBREV-HOOK DOCUMENT-EXPAND-ABBREV-ITEM HOOK-DOCUMENTATION-FUNCTION) 
(DEFUN DOCUMENT-EXPAND-ABBREV-ITEM (IGNORE CHAR)
  (AND (EXPAND-P CHAR)
       (PRINC "With no numeric argument, expand preceeding word abbrev if any.
"))) 

;;; Does this character try to expand preceeding abbrev?
(DEFUN EXPAND-P (CHAR)
  "T if CHAR is a command that should expand word abbrevs."
  (AND (ZEROP (CHAR-BITS CHAR))
       (NOT (ZEROP (AREF *WORD-ABBREV-TABLE* (CHAR-CODE CHAR)))))) 

(DEFUN EXPAND-ABBREV (&AUX BP STRING SYM TEM PROP)
  "Expand the word abbrev before point, if there is one."
  (AND (NOT (DELIMCHAR-P (BP-CHAR-BEFORE (POINT))))
       (MULTIPLE-VALUE-SETQ (STRING BP)
			    (BOUND-WORD (POINT))))
  (COND ((AND STRING
	      (SETQ SYM (FIND-SYMBOL (STRING-UPCASE STRING) *UTILITY-PACKAGE*))
	      (SETQ TEM (OR (GET SYM (SETQ PROP (GET-ABBREV-MODE-NAME)))
			    (GET SYM (SETQ PROP '*-ABBREV)))))
	 (COND (*LAST-EXPANSION-BP*
		(MOVE-BP *LAST-EXPANSION-BP* BP))
	       (T
		(SETQ *LAST-EXPANSION-BP* (COPY-BP BP :NORMAL))))
	 (COND ((AND (CHAR-EQUAL (BP-CHAR-BEFORE BP) #\-)
		     (BP-= (MOVE-BP BP (FORWARD-CHAR BP -1)) *WORD-ABBREV-PREFIX-MARK*))
		(SETQ STRING (STRING-APPEND "-" STRING))
		(DELETE-INTERVAL BP (FORWARD-CHAR BP) T))
	       (T
		(SETQ STRING (STRING-APPEND STRING))))
	 (SETQ *LAST-EXPANDED* STRING
	       *LAST-EXPANSION* TEM
	       *LAST-EXPANSION-SYMBOL* SYM
	       *LAST-EXPANSION-USAGE-PROP* (GET-ABBREV-USAGE-NAME PROP))
	 (LET ((V (GET SYM *LAST-EXPANSION-USAGE-PROP*)))
	   (SETF (GET SYM *LAST-EXPANSION-USAGE-PROP*)
		 (IF V (1+ V) 1)))

	 ;; may 12-13-88 added lowercase code below modified from pf's version
	 (if (and (upper-case-p (char string 0))
		  (upper-case-p (char string (1- (length string)))))

	     ;;  However, it's lots smarter about doing the casification.

	     ;;  First we do the case-replace (without the casing overhead).
	     (with-bp (bp *last-expansion-bp* :normal)
	       (delete-interval *last-expansion-bp* (point) t)
	       (insert-moving bp tem)

	       ;;  Then we count the words in the expansion and use that as the
	       ;;  argument to com-uppercase-initial (m-c).
	       (let* ((nwords (count-objects 'forward-word *last-expansion-bp* bp))
		      (*numeric-arg* nwords))
		 (move-bp (point) *last-expansion-bp*)
		 (IF (= nwords 1)		;; may 12-13-88 This is different than pf's version 
		     (COM-UPPERCASE-WORD)	;; may 12-13-88 but does what manual says it should.
		     (com-uppercase-initial))	
		 (move-bp (point) bp)))

	     ;;  Old normal case.
	     (MOVE-BP (POINT) (CASE-REPLACE *LAST-EXPANSION-BP* (POINT) TEM))))) ) 

(DEFCOM COM-EXPAND-ONLY "Expand last word, but insert nothing after it.
If given an argument, beep unless expanded." ()
  (AND (NULL (EXPAND-ABBREV))
       *NUMERIC-ARG-P*
       (BARF))
  DIS-TEXT) 

(DEFCOM COM-UNEXPAND-LAST-WORD "Undo last expansion, leaving the abbrev." ()
  (LET (BP TEM)
    (OR *LAST-EXPANSION* (BARF "No last expansion"))
    (SETQ BP (FORWARD-CHAR *LAST-EXPANSION-BP* (ARRAY-ACTIVE-LENGTH *LAST-EXPANSION*)))
    (OR (STRING-EQUAL (STRING-INTERVAL *LAST-EXPANSION-BP* BP T) *LAST-EXPANSION*)
	(BARF "No last expansion"))
    (SETQ TEM (BP-= BP (POINT))
	  BP (INSERT (DELETE-INTERVAL *LAST-EXPANSION-BP* BP) *LAST-EXPANDED*))
    (SETF (GET *LAST-EXPANSION-SYMBOL* *LAST-EXPANSION-USAGE-PROP*)
	  (1- (GET *LAST-EXPANSION-SYMBOL* *LAST-EXPANSION-USAGE-PROP*)))
    (AND TEM (MOVE-BP (POINT) BP)))
  DIS-TEXT) 

(DEFMINOR COM-WORD-ABBREV-MODE WORD-ABBREV-MODE "Abbrev" 3
	  "Mode for expanding word abbrevs.
No arg or non-zero arg sets the mode, 0 arg clears it." ()
;  (SET-COMTAB *MODE-COMTAB*			;; may 01-10-89
;	      '(#\c-m-SPACE COM-EXPAND-ONLY))  	;; may 01-10-89  
  (SET-COMTAB *STANDARD-CONTROL-X-COMTAB*
	      '(#\U COM-UNEXPAND-LAST-WORD
		#\c-A COM-ADD-MODE-WORD-ABBREV
		#\+ COM-ADD-GLOBAL-WORD-ABBREV))
  (COMMAND-HOOK 'EXPAND-ABBREV-HOOK *COMMAND-HOOK*)
  (SETQ *LAST-EXPANSION-BP* NIL)
  (SETQ *LAST-EXPANDED* NIL)
  (SETQ *LAST-EXPANSION* NIL)
  (SETQ *WORD-ABBREV-PREFIX-MARK* NIL)) 

(DEFCOM COM-MAKE-WORD-ABBREV "Prompt for and make a new word abbrev.
An argument means make global abbrev, else local for this mode." ()
  (MAKE-WORD-ABBREV *NUMERIC-ARG-P*
		    (TYPEIN-LINE-READLINE "Define ~:[~A mode~;global~*~] abbrev for: "
					  *NUMERIC-ARG-P* (NAME-OF-MAJOR-MODE)))) 

(DEFCOM COM-ADD-MODE-WORD-ABBREV "Read mode abbrev for words before point.
A negative arg means delete the word abbrev.  (If there is no such mode
abbrev, but there is a global, ask if should kill the global.)
Positive arg means expansion if last ARG words.
If there is a region, it is used instead." ()
  (COND ((MINUSP *NUMERIC-ARG*)
	 (LET ((*NUMERIC-ARG* (- *NUMERIC-ARG*)))
	   (COM-KILL-MODE-WORD-ABBREV)))
	(T
	 (MAKE-WORD-ABBREV NIL)))) 

(DEFCOM COM-ADD-GLOBAL-WORD-ABBREV "Reads mode abbrev for words before point.
A negative arg means delete the word abbrev.
Positive arg means expansion if last ARG words.
If there is a region, it is used instead." ()
  (COND ((MINUSP *NUMERIC-ARG*)
	 (LET ((*NUMERIC-ARG* (- *NUMERIC-ARG*)))
	   (COM-KILL-GLOBAL-WORD-ABBREV)))
	(T
	 (MAKE-WORD-ABBREV T)))) 

(DEFUN MAKE-WORD-ABBREV (GLOBAL-P &OPTIONAL STRING &AUX ABBREV)
  (OR STRING
      (SETQ STRING (COND ((WINDOW-MARK-P *WINDOW*)
			  (STRING-INTERVAL (MARK) (POINT)))
			 (T
			  (OR (BOUND-WORD (POINT) (ABS *NUMERIC-ARG*))
			      (BARF))))))
  (OR (STRINGP STRING)
      (SETQ STRING (STRING-APPEND "" STRING)))
  (SETQ ABBREV (LET ((*MINI-BUFFER-DEFAULT-STRING* STRING)
		     (*COMMAND-HOOK* NIL))	 ;Don't expand abbrevs within
		 (TYPEIN-LINE-READLINE
		   (FORMAT NIL "~:[~A mode~;Global~*~] abbrev for \"~A\": "
			   GLOBAL-P (NAME-OF-MAJOR-MODE) STRING))))
  (SETF (GET (INTERN (STRING-UPCASE ABBREV) *UTILITY-PACKAGE*)
	     (COND (GLOBAL-P '*-ABBREV)
		   (T (GET-ABBREV-MODE-NAME))))
	(STRING-APPEND STRING))
  (SETQ *WORD-ABBREV-TICK* (TICK))
  DIS-NONE) 

(DEFCOM COM-KILL-MODE-WORD-ABBREV "Cause mode abbrev typed to be expunged." ()
  (KILL-ABBREV NIL)) 

(DEFCOM COM-KILL-GLOBAL-WORD-ABBREV "Cause global abbrev typed to be expunged." ()
  (KILL-ABBREV T)) 

(DEFUN KILL-ABBREV (GLOBAL-P &AUX STRING SYM MODE-NAME PROP)
  (SETQ STRING (LET ((*COMMAND-HOOK* NIL))	 ;Don't expand abbrevs within
		 (TYPEIN-LINE-READLINE "Kill ~:[~A mode~;global~*~] abbrev: "
				       GLOBAL-P (NAME-OF-MAJOR-MODE))))
  (OR (SETQ SYM (FIND-SYMBOL (STRING-UPCASE STRING) *UTILITY-PACKAGE*))
      (BARF "No such abbrev defined"))
  (COND (GLOBAL-P
	 (OR (GET SYM '*-ABBREV)
	     (BARF "No such global abbrev defined."))
	 (REMPROP SYM '*-ABBREV)
	 (REMPROP SYM '*-ABBREV-USAGE))
	((GET SYM (SETQ MODE-NAME (SETQ PROP (GET-ABBREV-MODE-NAME))))
	 (REMPROP SYM MODE-NAME)
	 (REMPROP SYM (GET-ABBREV-USAGE-NAME PROP)))
	((NOT (GET SYM '*-ABBREV))
	 (BARF "No such abbrev defined."))
	((FQUERY '(:SELECT T) "~A is not a ~A mode abbrev, but is a global one, kill it? "
		 STRING (NAME-OF-MAJOR-MODE))
	 (REMPROP SYM '*-ABBREV)
	 (REMPROP SYM '*-ABBREV-USAGE))
	(T
	 (PROMPT-LINE "~&Not killed.")))
  (SETQ *WORD-ABBREV-TICK* (TICK))
  DIS-NONE) 

(DEFCOM COM-KILL-ALL-WORD-ABBREVS "No word abbrevs are defined after this." ()
  (DO-LOCAL-SYMBOLS (SYM *UTILITY-PACKAGE* NIL)
    (KILL-ALL-ABBREVS-1 SYM))
  DIS-NONE) 

(DEFUN KILL-ALL-ABBREVS-1 (SYM)
  (DO ((L (SYMBOL-PLIST SYM) (CDDR L))
       (IND)
       (IND-NAME)
       (LEN))
      ((NULL L))
    (SETQ IND (CAR L)
	  IND-NAME (SYMBOL-NAME IND)
	  LEN (LENGTH IND-NAME))
    (AND (> LEN 7)
	 (STRING-EQUAL (NSUBSTRING IND-NAME (- (LENGTH IND-NAME) 7)) "-ABBREV")
	 (REMPROP SYM IND)
	 (REMPROP SYM (GET-ABBREV-USAGE-NAME IND-NAME))))) 

(DEFUN BOUND-WORD (BP &OPTIONAL (TIMES 1) &AUX BP1 STRING)
  (AND (SETQ BP (FORWARD-TO-WORD BP -1))
       (SETQ BP1 (FORWARD-WORD BP (- TIMES)))
       (SETQ STRING (STRING-INTERVAL BP BP1)))
  (VALUES STRING BP1)) 

(DEFCOM COM-WORD-ABBREV-PREFIX-MARK "Mark point as end of a prefix" ()
  (EXPAND-ABBREV)
  (COND (*WORD-ABBREV-PREFIX-MARK*
	 (MOVE-BP *WORD-ABBREV-PREFIX-MARK* (POINT)))
	(T
	 (SETQ *WORD-ABBREV-PREFIX-MARK* (COPY-BP (POINT) :NORMAL))))
  (INSERT-MOVING (POINT) "-")
  DIS-TEXT) 

(DEFCOM COM-LIST-WORD-ABBREVS "List all abbrevs and their expansions." ()
  (FORMAT T "~%abbrev:   (mode)             count:    \"expansion\"~3%") ;; may 12-13-88
  (LIST-WORD-ABBREV-1 *STANDARD-OUTPUT*)
  DIS-NONE) 

(DEFCOM COM-INSERT-WORD-ABBREVS "Insert all abbrevs and their expansions into the buffer." (X)
  (LIST-WORD-ABBREV-1 (INTERVAL-STREAM *INTERVAL*))
  DIS-TEXT) 

(DEFUN LIST-WORD-ABBREV-1 (STREAM)
  (DECLARE (SPECIAL STREAM))
  (DO-LOCAL-SYMBOLS (SYM *UTILITY-PACKAGE* NIL)
    (DO ((L (SYMBOL-PLIST SYM) (CDDR L))
	 (IND)
	 (IND-NAME)
	 (USAGE)
	 (LEN)
	 (STRING ""))
	((NULL L)
	 (FORMAT STREAM STRING))
      (SETQ IND (CAR L)
	    IND-NAME (SYMBOL-NAME IND)
	    LEN (LENGTH IND-NAME)
	    USAGE (OR (GET SYM (GET-ABBREV-USAGE-NAME IND-NAME)) 0))
      (AND (> LEN 7)
	   (STRING-EQUAL (NSUBSTRING IND-NAME (- LEN 7)) "-ABBREV")
	   (SETQ STRING (FORMAT NIL
				"~A~10,,2A~:[~15,,2A~;~15@T~*~] ~6D~6@T \"~A\"~%" ;; may 12-13-88
				STRING
				(STRING-APPEND SYM ":")
				(EQ IND '*-ABBREV)
				(STRING-APPEND "(" (NSUBSTRING IND-NAME 0 (- LEN 7)) ")")
				USAGE
				(CADR L))))))) 

(DEFCOM COM-DEFINE-WORD-ABBREVS "Define word abbrevs from buffer" ()
  (DO ((BP1 (COPY-BP (INTERVAL-FIRST-BP *INTERVAL*)))
       (BP2)
       (MODE "*" "*")
       (USAGE)
       (SYM)
       (TEM))
      (NIL)
    (OR (SETQ BP2 (SEARCH BP1 #\:))
	(RETURN NIL))
    (SETQ TEM (STRING-UPCASE (STRING-INTERVAL BP1 (FORWARD-CHAR BP2 -1))))
    (SETQ SYM (INTERN TEM *UTILITY-PACKAGE*))
    (SETQ BP2 (FORWARD-OVER *BLANKS* (FORWARD-CHAR BP2)))
    (COND ((CHAR-EQUAL (BP-CHAR BP2) #\()
	   (OR (SETQ BP1 (SEARCH (SETQ BP2 (FORWARD-CHAR BP2)) #\)))
	       (BARF "Unmatched paren ~A" (BP-LINE BP2)))
	   (SETQ MODE (STRING-INTERVAL BP2 (FORWARD-CHAR BP1 -1)))
	   (SETQ BP2 (PROG1 (FORWARD-OVER *BLANKS* (FORWARD-CHAR BP1))
			    (SETQ BP1 BP2)))))
    (MULTIPLE-VALUE-SETQ (USAGE TEM)
			 (PARSE-NUMBER (BP-LINE BP2) (BP-INDEX BP2) NIL 12))
    (if (null tem)
	(BARF "No usage count ~A" (BP-LINE BP2)))
    (AND (= TEM (BP-INDEX BP2))
	 (BARF "No usage count ~A" (BP-LINE BP2)))
    (SETF (BP-INDEX BP2) TEM)
    (SETQ BP2 (FORWARD-OVER *BLANKS* BP2))
    (OR (CHAR-EQUAL (BP-CHAR BP2) #\")
	(BARF "No expansion ~A" (BP-LINE BP2)))
    (OR (SETQ BP1 (SEARCH (SETQ BP2 (FORWARD-CHAR BP2)) #\"))
	(BARF "Unmatched quote ~A" (BP-LINE BP2)))
    (SETF (GET SYM (GET-ABBREV-MODE-NAME MODE))
	  (STRING-APPEND (STRING-INTERVAL BP2 (FORWARD-CHAR BP1 -1))))
    (AND (EQ (BP-LINE BP1) (BP-LINE (INTERVAL-LAST-BP *INTERVAL*)))
	 (RETURN NIL))
    (MOVE-BP BP1 (BEG-LINE BP1 1)))
  ;; fix so Edit Word Abbrev will NOT cause Save Word Abbrev to say "(Word abbrevs do not need to be saved)" 
  (SETQ *WORD-ABBREV-TICK* (TICK)) ;; may 12-13-88 
  DIS-NONE) 

(DEFUN GET-ABBREV-MODE-NAME (&OPTIONAL (MODE (STRING (NAME-OF-MAJOR-MODE))))
  (INTERN (STRING-APPEND MODE "-ABBREV") (SYMBOL-PACKAGE 'FOO))) 

;;; Given the name of a X-ABBREV prop, return the usage count property.
(DEFUN GET-ABBREV-USAGE-NAME (STR)
  (INTERN (STRING-APPEND STR "-USAGE") (SYMBOL-PACKAGE 'FOO))) 

(DEFCOM COM-EDIT-WORD-ABBREVS "Enter recursive edit on the abbrev definitions." ()
  (LET ((INTERVAL (CREATE-INTERVAL NIL NIL T)))
    (SETF (NODE-TICK INTERVAL) (TICK))
    (LET ((*INTERVAL* INTERVAL)
	  (*BATCH-UNDO-SAVE* T))
      (COM-INSERT-WORD-ABBREVS))
    (RECURSIVE-EDIT INTERVAL "Edit Word Abbrevs")
    (COM-KILL-ALL-WORD-ABBREVS)
    (LET ((*INTERVAL* INTERVAL))
      (COM-DEFINE-WORD-ABBREVS)))
  DIS-NONE) 

(DEFCOM COM-RECURSIVE-EDIT-BEEP "Exit from recursive edit without updating." ()
  (BEEP)
  (MUST-REDISPLAY *WINDOW* DIS-TEXT)
  (THROW 'TOP-LEVEL T)) 

(DEFCOM COM-LIST-SOME-WORD-ABBREVS "List abbreviations or expansions with the given string" ()
  (MULTIPLE-VALUE-BIND (APROPOS-SEARCH-FUNCTION APROPOS-KEY)
      (GET-EXTENDED-SEARCH-STRINGS "Word abbrev apropos (substring:)")
    (DO-LOCAL-SYMBOLS (SYM *UTILITY-PACKAGE* NIL)
      (WORD-ABBREV-APROPOS-INTERNAL SYM)))
  (FORMAT T "Done.~%")
  DIS-NONE) 

(DEFUN WORD-ABBREV-APROPOS-INTERNAL (SYM)
  (DO ((L (SYMBOL-PLIST SYM) (CDDR L))
       (IND)
       (IND-NAME)
       (LEN))
      ((NULL L))
    (SETQ IND (CAR L)
	  IND-NAME (SYMBOL-NAME IND)
	  LEN (LENGTH IND-NAME))
    (AND (> LEN 7)
	 (STRING-EQUAL (NSUBSTRING IND-NAME (- LEN 7)) "-ABBREV")
	 (OR (FUNCALL APROPOS-SEARCH-FUNCTION APROPOS-KEY (SYMBOL-NAME SYM))
	     (FUNCALL APROPOS-SEARCH-FUNCTION APROPOS-KEY (CADR L)))
	 (FORMAT T "~A:	~:[(~A)~;~*~]	\"~A\"~%"
		 SYM (EQ IND '*-ABBREV)
		 (NSUBSTRING IND-NAME 0 (- LEN 7)) (CADR L))))) 

(DEFCOM COM-READ-WORD-ABBREV-FILE "Load up new format word abbrev file." ()
  (AND *WORD-ABBREV-FILE-NAME*
       (FS:SET-DEFAULT-PATHNAME *WORD-ABBREV-FILE-NAME* *AUX-PATHNAME-DEFAULTS*))
  (LET ((FNAME (READ-DEFAULTED-AUX-PATHNAME "Load QWABL file:" :QWABL)))
    (WITH-OPEN-FILE (STREAM FNAME :CHARACTERS T :DIRECTION :INPUT)
      (LOAD-QWABL STREAM))
    (SETQ *WORD-ABBREV-FILE-NAME* FNAME
	  *WORD-ABBREV-FILE-TICK* (TICK)))
  DIS-NONE) 

(DEFUN PRELOAD-WORD-ABBREVS (FILE)
  "Load FILE's word abbrevs into ZMACS.  Loses under canonical file operations."
  (WITH-OPEN-FILE (STREAM FILE)
    (LOAD-QWABL STREAM))) 

(DEFUN LOAD-QWABL (STREAM)
  (SEND STREAM :LINE-IN)	 ;Flush some TECO macros
  (SEND STREAM :LINE-IN)
  (DO ((SYM)
       (USAGE)
       (MODE)
       (STR)
       (EOFP)
       (TEM)
       (TEM1))
      (NIL)
    (MULTIPLE-VALUE-SETQ (STR EOFP)
      (SEND STREAM :LINE-IN T))
    (AND EOFP (RETURN NIL))
    (OR (SETQ TEM (POSITION #\SPACE (THE STRING (STRING STR))
			    :TEST #'CHAR-EQUAL))
	(BARF "No abbrev ~S" STR))
    (SETQ TEM (1+ TEM))
    (OR (SETQ TEM1 (POSITION #\SPACE (THE STRING (STRING STR))
			     :START TEM
			     :TEST #'CHAR-EQUAL))
	(BARF "No mode ~S" STR))
    (SETQ SYM (STRING-UPCASE (NSUBSTRING STR TEM TEM1))
	  TEM1 (1+ TEM1))
    (OR (SETQ TEM (POSITION #\SPACE (THE STRING (STRING STR))
			    :START TEM1
			    :TEST #'CHAR-EQUAL))
	(BARF "No end of mode ~S" STR))
    (SETQ MODE (NSUBSTRING STR TEM1 TEM)
	  TEM (1+ TEM))
    (SETQ MODE (GET-ABBREV-MODE-NAME MODE))
    (OR (SETQ TEM1 (POSITION #\ESCAPE (THE STRING (STRING STR))
			     :START TEM
			     :TEST #'CHAR-EQUAL))
	(BARF "No expansion ~S" STR))
    (SETQ STR (NSUBSTRING STR (1+ TEM1)))
    (SETQ STR (DO ((EXPANSION "" (STRING-APPEND EXPANSION STR))
		   (STR STR (SEND STREAM :LINE-IN))
		   (POS))
		  (NIL)
		(COND ((SETQ POS (POSITION #\ESCAPE (THE STRING (STRING STR)) :TEST #'CHAR-EQUAL))
		       (SETQ USAGE (PARSE-NUMBER STR (1+ POS) NIL 12))
		       (SETQ STR (NSUBSTRING STR 0 POS))
		       (RETURN (STRING-APPEND EXPANSION STR))))))
    (SETQ SYM (INTERN SYM *UTILITY-PACKAGE*))
    (SETF (GET SYM MODE) STR)
    (SETF (GET SYM (GET-ABBREV-USAGE-NAME MODE)) USAGE))) 

(DEFUN WRITE-QWABL (STREAM &AUX L)				;; may 12-13-88 added ALL text comments 
  (DECLARE (SPECIAL L))
  (SEND STREAM :LINE-OUT "m.m& Make Usage Abbrev Variable[V") 	;; TECO macro stuff
  (SEND STREAM :LINE-OUT "q..q[..o")			 	;; TECO macro stuff
  (DO-LOCAL-SYMBOLS (SYM *UTILITY-PACKAGE* NIL)			;; symbols in this pkg also used by REGISTERS
    (PUSH SYM L))
  (SETQ L (SORT L 'STRING-LESSP))
  (DO ((SL L (CDR SL))
       (SYM))
      ((NULL SL))
    (SETQ SYM (CAR SL))						;; get a symbol from the pkg
    (DO ((LIST (SYMBOL-PLIST SYM) (CDDR LIST))			;; get the plist of the symbol
	 (PROPNAME)
	 (LPROPNAME)
;;	 (J 0)									;; may 12-13-88 deleted
	 )
	((NULL LIST))
      (SETQ PROPNAME (CAR LIST)					;; get the property and length WITHOUT "|" and pkg prefixes
;;	    LPROPNAME (flatsize PROPNAME))					;; may 12-13-88 deleted
	    LPROPNAME (length (string-upcase PROPNAME))) 			;; may 12-13-88 *ADDED*
      (COND (
;;	     (do ((I 0 (1+ i)))							;; may 12-13-88 deleted
;;		 ((>= I LPROPNAME))						;; may 12-13-88 deleted
;;	       (if (equal (aref (string-upcase propname) I)			;; may 12-13-88 deleted
;;			  (aref "-ABBREV" J))					;; may 12-13-88 deleted
;;		   (setq J (1+ J))						;; may 12-13-88 deleted
;;		   (setq J 0))							;; may 12-13-88 deleted
;;	       (if (>= J 6)							;; may 12-13-88 deleted
;;		   (return t)))							;; may 12-13-88 deleted
	     ;; If the LAST 7 chars are "-ABBREV" then the property is one of the types we need
	     ;; we do NOT want to process COMMON-LISP-ABBREV-USAGE ( as earlier versions incorrectly did )
	     (STRING-EQUAL "-ABBREV" PROPNAME :START1 0 :START2 (- LPROPNAME 7));; may 12-13-88 *UN-DELETED*
	     ;; file line format : ( followed by an example)
	     ;;MVX <abbreviation> <abbrev type WITHOUT "-ABBREV"> Abbrev<expansion><how many times expanded>
	     ;;MVX xx Common-Lisp AbbrevThis is a long word abbreviation with !@#$#%*$&^%( funny chars1
	     ;;MVX zz Common-Lisp AbbrevZetalisp-IS-oUT1
	     ;;The "MVX" is unused?
	     (FORMAT STREAM "MVX ~A ~A Abbrev~A~D~%" (STRING-DOWNCASE SYM)
		     (NSUBSTRING PROPNAME 0 (- LPROPNAME 7)) (CADR LIST)
		     (OR (GET SYM (GET-ABBREV-USAGE-NAME PROPNAME)) 0))))))) 

(DEFCOM COM-WRITE-WORD-ABBREV-FILE "Write out all word abbrevs in QWABL format." ()
  (AND *WORD-ABBREV-FILE-NAME*
       (FS:SET-DEFAULT-PATHNAME *WORD-ABBREV-FILE-NAME* *AUX-PATHNAME-DEFAULTS*))
  (LET ((FN (READ-DEFAULTED-AUX-PATHNAME "Write word abbrevs to:" :QWABL NIL :WRITE)))
    (COM-WRITE-WORD-ABBREV-FILE-INTERNAL FN))
  DIS-NONE) 

(DEFCOM COM-SAVE-WORD-ABBREV-FILE "Write out word abbrevs if changed." ()
  (COM-SAVE-WORD-ABBREV-FILE-INTERNAL :EXPLICIT)
  DIS-NONE) 

(DEFUN COM-SAVE-WORD-ABBREV-FILE-INTERNAL (WHEN)
  (COND ((<= *WORD-ABBREV-TICK* *WORD-ABBREV-FILE-TICK*)
	 (AND (EQ WHEN :EXPLICIT)
	      (FORMAT *QUERY-IO* "~&(Word abbrevs do not need to be saved)")))
	((AND (EQ WHEN :ASK)   
	      (NOT (FQUERY '(:SELECT T) "Save word abbrevs~@[on file ~A~]? "
			   *WORD-ABBREV-FILE-NAME*))))
	(T
	 (OR *WORD-ABBREV-FILE-NAME*
	     (SETQ *WORD-ABBREV-FILE-NAME* (READ-DEFAULTED-AUX-PATHNAME "Save word abbrevs to:"
									:QWABL NIL :WRITE)))
	 (COM-WRITE-WORD-ABBREV-FILE-INTERNAL *WORD-ABBREV-FILE-NAME*)))) 

;Return T if caller should save the word abbrevs.
;Any questions which need to be asked are asked now.
;The caller can then simply call COM-WRITE-WORD-ABBREV-FILE-INTERNAL.
(DEFUN WORD-ABBREVS-NEED-SAVING-P (NO-QUERY)
  (AND (> *WORD-ABBREV-TICK* *WORD-ABBREV-FILE-TICK*)
       (OR NO-QUERY
	   (FQUERY '(:SELECT T) "Save word abbrevs~@[on file ~A~]? "
		   *WORD-ABBREV-FILE-NAME*))
       (PROGN
	 (OR *WORD-ABBREV-FILE-NAME*
	     (SETQ *WORD-ABBREV-FILE-NAME* (READ-DEFAULTED-AUX-PATHNAME "Save word abbrevs to:"
									:QWABL NIL :WRITE)))
	 T))) 

(DEFUN COM-WRITE-WORD-ABBREV-FILE-INTERNAL (&OPTIONAL (FN *WORD-ABBREV-FILE-NAME*))
  (WITH-OPEN-FILE (STREAM FN :CHARACTERS T :DIRECTION :OUTPUT)
    (WRITE-QWABL STREAM)
    (CLOSE STREAM)
    (FORMAT *QUERY-IO* "~&Written: ~A" (SEND STREAM :TRUENAME)))
  (SETQ *WORD-ABBREV-FILE-NAME* FN
	*WORD-ABBREV-FILE-TICK* (TICK))) 

;;;Shift and font lock modes

(DEFMINOR COM-ELECTRIC-SHIFT-LOCK-MODE ELECTRIC-SHIFT-LOCK-MODE "Shift-lock" 5
	  "Minor mode upcasing all but comments and strings.
A positive argument turns the mode on, zero turns it off;
no argument toggles." ()
  (COMMAND-HOOK 'SHIFT-LOCK-HOOK *COMMAND-HOOK*)) 

(DEFMINOR COM-ELECTRIC-FONT-LOCK-MODE ELECTRIC-FONT-LOCK-MODE "Font-lock" 5
	  "Minor mode to fontify characters as follows:
     Default Font      A
     Comments          B 
     Doc Strings       C 
     Non Doc Strings   D or *ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT* if a fixnum
     Function Spec     E
A positive argument turns the mode on, zero turns it off; no argument toggles." ()
  (COMMAND-HOOK 'FONT-LOCK-HOOK *COMMAND-HOOK*)) 

(DEFPROP SHIFT-LOCK-HOOK 10 COMMAND-HOOK-PRIORITY) 
(DEFUN SHIFT-LOCK-HOOK (CHAR &AUX STRING SLASH COMMENT
			(POINT (POINT)) (*LISP-PARSE-PREPARSED-FLAG* T))
  (COND ((AND (OR (AND (CHAR>= CHAR #\A)
		       (CHAR<= CHAR #\Z))
		  (AND (CHAR>= CHAR #\a)
		       (CHAR<= CHAR #\z)))
	      (NEQ *INTERVAL* (WINDOW-INTERVAL *MINI-BUFFER-WINDOW*)))
	 (OR (AND (EQ *LAST-COMMAND-TYPE* 'SELF-INSERT)
		  (EQ (BP-LINE POINT) *SHIFT-LOCK-HOOK-LAST-LINE*))
	     (SETQ *SHIFT-LOCK-HOOK-DEFUN-BEGINNING* (FORWARD-DEFUN POINT -1 T)
		   *SHIFT-LOCK-HOOK-LAST-LINE* (BP-LINE POINT)
		   *LISP-PARSE-PREPARSED-FLAG* NIL))
	 (MULTIPLE-VALUE-SETQ (STRING SLASH COMMENT)
	   (LISP-BP-SYNTACTIC-CONTEXT POINT *SHIFT-LOCK-HOOK-DEFUN-BEGINNING*))
	 (OR STRING SLASH COMMENT
	     (SETQ *LAST-COMMAND-CHAR* (IF *ELECTRIC-SHIFT-LOCK-XORS*
					   (CODE-CHAR (LOGXOR (CHAR-CODE CHAR) 40))
					   (CODE-CHAR (BOOLE 4 (CHAR-CODE CHAR) 40)))))))) 

(DEFPROP FONT-LOCK-HOOK 10 COMMAND-HOOK-PRIORITY) 

;;;New Electric Font Lock Mode fixes    - 9/85 DLC
;;; The default behavior is as follows:
;;;  	- The default font is Font A
;;; 	- Comments go in Font B
;;; 	- Documentation strings (a string that is 4th element of a top-level form) go in Font C
;;; 	- All other strings go into Font D if there is a declared 4th font, else to Font A
;;; The font used for Non-Documentation strings is determined by the Zmacs Variable *ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT*.
;;; If the value of this variable is a fixnum, the corresponding font is used (eg Font A = 0; Font B = 1; etc).
;;; If the value of this variable is not a fixnum (the default) Font D or Font A is used (as described above).
;;; Now, if a buffer has no font specification in the file attribute line, the default behavior is for Electric 
;;; Font Lock Mode to act as if were really off.  This behavior is controlled by the Zmacs variable
;;; *ELECTRIC-FONT-LOCK-IF-NO-FONTS-DECLARED-P* [default = NIL].  
;;; Slashification of quotation marks is now handled correctly.
;;; The selection of which font (ie, Font A vs Font B) is intentionally not user definable.
;;; This enforces a certain level of standardization among users.  I.e., Lisp code will ALWAYS be in Font A.
;;; Still to be added are embedded documentation strings for things such as DEFFLAVOR.

;;; NEWER Electric Font Lock Mode fixes - 12/86 DKM & RPM
;;; Fonts defined as:
;;;  DEFAULT FONT     A 
;;;  COMMENTS         B 
;;;  DOC STRINGS      C 
;;;  NON DOC STRINGS  D OR *ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT* IF A FIXNUM
;;;  FUNCTION SPEC    E
;NOTE: This guy still will change the font on the mode line.  I don't know of a cheap 
;   check I can do to prevent it (like (mode-line-bp inteval) ).  It has to be cheap
;   because this takes too much time as it is.
;ALSO: This guy could be smarter.  It checks state of the current point in total, without
;   regard for what it did on the last character.  This can be very expensive on strings.
;   Maybe lisp-bp-syntactic-context tries to save state, but it doesn't appear to do so.
;   Turn electric font lock on in some very large buffer sometime ... S L O W.
(DEFUN FONT-LOCK-HOOK (CHAR)						
  (LET (COMMENT-P
	STRING-P
	(POINT (POINT))
;;;	(*LISP-PARSE-PREPARSED-FLAG* T)
	SLASHIFIED-P
	fun-depth)
    (COND ((NEQ *INTERVAL* (WINDOW-INTERVAL *MINI-BUFFER-WINDOW*))
	   (OR (AND (EQ *LAST-COMMAND-TYPE* 'SELF-INSERT)
		    (EQ (BP-LINE POINT) *SHIFT-LOCK-HOOK-LAST-LINE*))
	       (SETQ *SHIFT-LOCK-HOOK-DEFUN-BEGINNING* (FORWARD-DEFUN POINT -1 T)
		     *SHIFT-LOCK-HOOK-LAST-LINE* (BP-LINE POINT)
		     *LISP-PARSE-PREPARSED-FLAG* NIL))
	   (MULTIPLE-VALUE-SETQ (STRING-P SLASHIFIED-P COMMENT-P fun-depth)
	     (LISP-BP-SYNTACTIC-CONTEXT POINT *SHIFT-LOCK-HOOK-DEFUN-BEGINNING*))
	   (LET ((NEW-FONT (COND
			     ;;If there are no fonts declared, see if fonts should be used anyways
			     ((AND (NULL *ELECTRIC-FONT-LOCK-IF-NO-FONTS-DECLARED-P*)
				   (NULL (WINDOW-FONT-ALIST *WINDOW*)))
			      0)
			     ;;A Comment
			     ((AND COMMENT-P (>= (LENGTH (WINDOW-FONT-ALIST *WINDOW*)) 2))
			      1)
			     ;;A Non-Documentation String
			     ((AND STRING-P	;make sure not to include terminating quotes
				   (not (equal fun-depth 1))
				   (OR SLASHIFIED-P (NOT (EQL CHAR #\"))))	 
			      (COND ((INTEGERP *ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT*)
				     *ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT*)
				    ((< (LENGTH (WINDOW-FONT-ALIST *WINDOW*)) 4)
				     0)
				    (T 3)))
			     ((AND (>= (LENGTH (WINDOW-FONT-ALIST *WINDOW*)) 5)
				   (MULTIPLE-VALUE-BIND (BEG END)
				       (IN-FUNCTION-SPEC (BP-LINE POINT)
							 *LAST-COMMAND-CHAR*  
							 (BP-INDEX POINT))
				     (AND BEG END (<= BEG (BP-INDEX POINT) (1+ END)))))
			      4)
			     ;;A Documentation String
			     ((AND STRING-P (>= (LENGTH (WINDOW-FONT-ALIST *WINDOW*)) 3)
				   (OR SLASHIFIED-P (NOT (EQL CHAR #\"))))
;;;				   (BP-IN-DOC-STRING POINT CHAR SLASHIFIED-P))
			      2)
			     (T 0))))
	     (COND
	       ((/= *FONT* NEW-FONT)
		(SETQ *FONT* NEW-FONT)
		(UPDATE-FONT-NAME))))))))

(DEFUN BP-IN-DOC-STRING (BP CHAR SLASHIFIED-P)
  "Returns true if BP is within a documentation string, that is, a string
that is the fourth top-level form within a defining form.  This function
assumes that BP is already known to be within a string."
  (AND (OR SLASHIFIED-P (NOT (EQL CHAR #\")))
       (BP-= *SHIFT-LOCK-HOOK-DEFUN-BEGINNING*
	     (FORWARD-SEXP (FORWARD-UP-STRING BP T) -4 T)))) 

;It is useful to setq LISP-MODE-HOOK to this
(DEFUN ELECTRIC-SHIFT-LOCK-IF-APPROPRIATE ()
  (IF (SEND *INTERVAL* :GET-ATTRIBUTE :LOWERCASE)
      (TURN-OFF-MODE 'ELECTRIC-SHIFT-LOCK-MODE)
      (TURN-ON-MODE 'ELECTRIC-SHIFT-LOCK-MODE))) 

;It is useful to setq LISP-MODE-HOOK to this
(DEFUN ELECTRIC-FONT-LOCK-IF-APPROPRIATE ()
  "Turn on Electric Font Lock Mode if there's more than one font in the attribute list."
  (IF (LET ((FONTS (SEND *INTERVAL* :GET-ATTRIBUTE :FONTS)))
	(AND FONTS (CONSP FONTS)
	     (> (LENGTH FONTS) 1)))
      (TURN-ON-MODE 'ELECTRIC-FONT-LOCK-MODE)
      (TURN-OFF-MODE 'ELECTRIC-FONT-LOCK-MODE))) 

;It is useful to setq TEXT-MODE-HOOK to this
(DEFUN AUTO-FILL-IF-APPROPRIATE ()
  (IF (SEND *INTERVAL* :GET-ATTRIBUTE :NOFILL)
      (TURN-OFF-MODE 'AUTO-FILL-MODE)
      (TURN-ON-MODE 'AUTO-FILL-MODE))) 

;;;   Uppercase Global Functions Mode
;;;
;;;   This minor mode provides automatic uppercasing of function names that
;;;   are contained in the GLOBAL package.  When any word abbreviation mode
;;;   delimiter except '-' is typed, the editor will check the preceding atom 
;;;   to see if it needs uppercasing.
;;;
;;;   Besides the minor mode, two new commands are added.
;;;	-  com-uppercase-previous-global-function - uppercase the atom under point
;;;	     if it is the name of a function in the global package
;;;	-  com-uppercase-global-functions-region - uppercase global function names
;;;          from point to mark, using the current defun is no region is defined;
;;;	     with an argument, lowercase all non-global function names in the region
;;;
;;;		- rb 2/17/85 (this fix due to lo)

(DEFMINOR COM-UPPERCASE-GLOBAL-FUNCTIONS-MODE UPPERCASE-GLOBAL-FUNCTIONS-MODE "Globals" 3
	  "Mode for capitalizing global functions, macros, etc." ()
  (COMMAND-HOOK 'UPPERCASE-GLOBAL-FUNCTIONS-HOOK *COMMAND-HOOK*)) 

(DEFPROP UPPERCASE-GLOBAL-FUNCTIONS-HOOK 10 COMMAND-HOOK-PRIORITY) 

(DEFUN UPPERCASE-GLOBAL-FUNCTIONS-HOOK (CHAR)
  "Uppercases the atom preceding point if it is a function name in the global package."
  (IF (NOT (EQ *INTERVAL* (WINDOW-INTERVAL *MINI-BUFFER-WINDOW*)))
      (UPPERCASE-GLOBAL-FUNCTION-AT-BP (POINT) CHAR))) 

(DEFUN UPPERCASE-GLOBAL-FUNCTION-AT-BP (BP &OPTIONAL (CHAR (BP-CHAR BP)) (LOWERCASE NIL))
  "Uppercases atom before BP if it is the name of a function in the global package.
Conditions: (a) last command char must have been a WAM delimiter (except '-' or ':').
            (b) BP must not be inside a comment, string, or slash context."
  (AND (NOT (MEMBER (MAKE-CHAR CHAR) '(#\- #\:)))	; don't expand after dashes or colons
       (EXPAND-P CHAR)				; expand after WAM delimiters
       ;; don't expand in comments,comp  strings or slashes
       (NOT (MULTIPLE-VALUE-BIND (STRING SLASH COMMENT)
		(LISP-BP-SYNTACTIC-CONTEXT BP (FORWARD-DEFUN BP -1 T))
	      (OR STRING SLASH COMMENT)))	 
       ;; okay, check for function name in global package under bp
       (UPPERCASE-GLOBAL-FUNCTION BP LOWERCASE))) 

(DEFUN UPPERCASE-GLOBAL-FUNCTION (BP &OPTIONAL (LOWERCASE NIL))
  "If the atom under BP is a global function or special form, uppercase it."
  (LET (BP1 STR)
    (SETQ BP1 (FORWARD-ATOM BP -1 T)			 ; beginning of atom  
	  STR (STRING-UPCASE (STRING-INTERVAL BP1 BP)))	 ; string with atom
    (COND ((FUNCALL *UPPERCASE-GLOBAL-FUNCTION-P* BP1 BP STR)
	   (UPCASE-INTERVAL BP1 BP T)
	   ;; everything after beginning of atom needs redisplay
	   (MUST-REDISPLAY *WINDOW* DIS-LINE (BP-LINE BP1) (BP-INDEX BP1)))
	  (LOWERCASE
	   (DOWNCASE-INTERVAL BP1 BP T)
	   (MUST-REDISPLAY *WINDOW* DIS-LINE (BP-LINE BP1) (BP-INDEX BP1)))))) 

(DEFUN UPPERCASE-GLOBAL-FUNCTION-P (START-BP CURRENT-BP STRING)
  "Tests whether or not the current word should be capitalized."
  (DECLARE (VALUES T-OR-NIL) (IGNORE CURRENT-BP))
  (AND
    ;;Check to see if symbol is a global function.
    (FUNCTIONP (OR (FIND-SYMBOL STRING 'LISP) (FIND-SYMBOL STRING 'TICL)) T)
    ;;Check to see if we are typing a function name. (after a paren)
    (LET ((PREVIOUS-STRING (STRING-INTERVAL (FORWARD-CHAR START-BP -1 T) START-BP)))
      (AND (PLUSP (LENGTH PREVIOUS-STRING))
	   (OR (CHAR-EQUAL #\( (AREF PREVIOUS-STRING 0))
	       (CHAR-EQUAL #\' (AREF PREVIOUS-STRING 0)))))))

(DEFCOM COM-UPPERCASE-PREVIOUS-GLOBAL-FUNCTION
	"Uppercase the atom under point if it is the name of a global function." ()
  (UPPERCASE-GLOBAL-FUNCTION (POINT))
  DIS-NONE) 

(DEFCOM COM-UPPERCASE-GLOBAL-FUNCTIONS-REGION
	"Uppercase global function names from point to mark.
If there is no region, the current definition is used.
If run with a numeric parameter, make non-global function names lowercase" ()
  (LET (BP1 BP2)
    (IF (WINDOW-MARK-P *WINDOW*)			 ; if there is a region
	(SETQ BP1 (POINT)
	      BP2 (MARK))				 ; grab point and mark
	(SETQ BP1 (FORWARD-DEFUN (POINT) -1 T)		 ; else use begin and end 
	      BP2 (FORWARD-DEFUN (POINT) 1 T)))		 ; of current DEFUN
    (ORDER-BPS BP1 BP2)					 ; put BPs in order
    (DO ((BP-TEMP (FORWARD-CHAR BP1 -1 T)))		 ; move forward one char so we don't
							 ; catch the atom just before BP1
	((OR (BP-< BP2 BP-TEMP) (BP-= BP2 BP-TEMP)))	 ; stop at BP2
      (SETQ BP-TEMP (FORWARD-ATOM BP-TEMP 1 T))		 ; move forward one atom
      (UPPERCASE-GLOBAL-FUNCTION-AT-BP
	BP-TEMP (BP-CHAR BP-TEMP) *NUMERIC-ARG-P*)))	 ; and try uppercasing it
  DIS-TEXT) 
