;;; -*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1980, Massachusetts Institute of Technology

;;; This file provides the section specific code for ZMACS
;;; It uses the utility file stuff in ZWEI; FILES.
;;; The simple stuff is in ZWEI; ZMACS

;;; Change History:
;;; 05/08/89 	MAY	Added LIST-METHODS-THRU-MOUSE for COM-EDIT-DEFINITION to allow
;;; 			meta-. on a keyword ( method )
;;; 04/20/89    MAY     (for DNG) Redesigned function ARGLIST-TYPE-SPECIFIERS. 
;;; 03/28/89    MAY	Added CLOS changes for DNG. SYMBOL-FROM-STRING, arglist-type-specifiers (new)
;;;			GET-DEFmethod-arglist (new), and (:PROPERTY :LISP GET-SECTION-NAME).
;;; 07-12-88	MAY	Added check for (null pkg) in COM-FUNCTION-APROPOS. SPR 5973

(DEFCOM COM-EDIT-CALLERS "Edit functions that call the specified one. Searches the current
package, or all packages with control-U, or asks for a package with two control-U's." ()
  (MULTIPLE-VALUE-BIND (FUN CALLERS)
      (WHO-CALLS-INTERNAL "Edit")
    (LIST-ZMACS-CALLERS-TO-BE-EDITED "Callers of" FUN T CALLERS))) 

(DEFCOM COM-LIST-CALLERS "List functions that use the specified function. Searches the current
package, or all packages with control-U, or asks for a package with two control-U's." ()
  (MULTIPLE-VALUE-BIND (FUN CALLERS)
      (WHO-CALLS-INTERNAL "List")
    (LIST-ZMACS-CALLERS-TO-BE-EDITED "Callers of" FUN NIL CALLERS))
  DIS-NONE) 

(DEFCOM COM-MULTIPLE-EDIT-CALLERS "Edit functions that use the specified functions. Searches the
current package, or all packages with control-U, or asks for a package with two control-U's. This
is the same as Edit Callers except it keeps asking for callees until you type just a carriage return." ()
  (MULTIPLE-VALUE-BIND (FUN CALLERS)
      (MULTIPLE-WHO-CALLS-INTERNAL "Edit")
    (LIST-ZMACS-CALLERS-TO-BE-EDITED "Callers of" FUN T CALLERS))) 

(DEFCOM COM-MULTIPLE-LIST-CALLERS "List functions that use the specified functions. Searches the
current package, or all packages with control-U, or asks for a package with two control-U's. This is
the same as List Callers except it keeps asking for callees until you type just a carriage return." ()
  (MULTIPLE-VALUE-BIND (FUN CALLERS)
      (MULTIPLE-WHO-CALLS-INTERNAL "List")
    (LIST-ZMACS-CALLERS-TO-BE-EDITED "Callers of" FUN NIL CALLERS))
  DIS-NONE) 

;;Get a package and the name of it to go in the prompt string,
;;based on numeric argument.
(DEFUN GET-PACKAGE-TO-SEARCH ()
  (LET ((PKG (COND ((< *NUMERIC-ARG* 4)
		    *PACKAGE*)
		   ((< *NUMERIC-ARG* 20)
		    NIL)
		   (T
		    (LET ((X (TYPEIN-LINE-READLINE "Package to search (default ~A):" *PACKAGE*)))
		      ;;Allow for case-sensitivity in package names. -rpm 7-10-86  
		      (COND ((EQUAL X "")
			     (FIND-PACKAGE *PACKAGE*))
			    ;;Check for it first exactly as input by user, then try to help.
			    ((FIND-PACKAGE X)) 
			    ((FIND-PACKAGE (STRING-UPCASE X)))
			    ((FIND-PACKAGE (STRING-DOWNCASE X)))
			    (T
			     (BARF "Package ~A does not exist." X))))))))
    (VALUES PKG (IF (= *NUMERIC-ARG* 4)
		    "all packages"
		    (FORMAT NIL "package ~A" PKG))))) 

(DEFUN WHO-CALLS-INTERNAL (PROMPT &OPTIONAL MUST-BE-DEFINED-FLAG)
  "Read a function name and find all functions that call it.
First value is function looked for, second value is list of callers' names."
  (MULTIPLE-VALUE-BIND (PKG PKG-NAME)
      (GET-PACKAGE-TO-SEARCH)
    (LET ((FUNCTION (READ-FUNCTION-NAME (FORMAT NIL "~A callers in ~A of" PROMPT PKG-NAME)
					(RELEVANT-FUNCTION-NAME (POINT) NIL T T T)
					MUST-BE-DEFINED-FLAG
					'ALWAYS-READ)))
      (AND (CONSP FUNCTION)
	   (EQ (CAR FUNCTION) :METHOD)
	   (SETQ FUNCTION (CAR (LAST FUNCTION)))) ;;; Patch 94.115   ddd/gsl 2/26/84.
      (FORMAT *QUERY-IO* "~&~Aing callers in ~A of ~S." PROMPT PKG-NAME FUNCTION)
      (VALUES FUNCTION (SETUP-ZMACS-CALLERS-TO-BE-EDITED (LIST-CALLERS FUNCTION PKG)))))) 

(DEFUN MULTIPLE-WHO-CALLS-INTERNAL (PROMPT)
  "Read several function names and find all functions that call any of them.
First value is functions looked for,
second value is list of their callers' names."
  (MULTIPLE-VALUE-BIND (PKG PKG-NAME)
      (GET-PACKAGE-TO-SEARCH)
    (DO ((FUNCTIONS NIL)
	 (FUNCTION)
	 (END "Stop"))
	(NIL)
      (SETQ FUNCTION (READ-FUNCTION-NAME (FORMAT NIL "~A callers in ~A of" PROMPT PKG-NAME)
					 (IF (NULL FUNCTIONS)
					     (RELEVANT-FUNCTION-NAME (POINT) NIL T T T)
					     END)
					 NIL 'ALWAYS-READ))
      (AND (CONSP FUNCTION)
	   (EQ (CAR FUNCTION) :METHOD)
	   (SETQ FUNCTION (CAR (LAST FUNCTION)))) ;;; Patch 94.124   ddd/gsl 2/26/84.
      (COND ((NEQ FUNCTION END)
	     (PUSH FUNCTION FUNCTIONS))
	    (T
	     (SETQ FUNCTIONS (NREVERSE FUNCTIONS))
	     (PROMPT-LINE
	       "~&~Aing callers in ~A of~1{~#[~; ~S~; ~S and ~S~:;~@{~#[~1; and~] ~S~^,~}~]~}."
	       PROMPT PKG-NAME FUNCTIONS)
	     (RETURN (VALUES FUNCTIONS (SETUP-ZMACS-CALLERS-TO-BE-EDITED (LIST-CALLERS FUNCTIONS PKG))))))))) 

(DEFUN SETUP-ZMACS-CALLERS-TO-BE-EDITED (CALLERS)
  "Sort a list of function specs by pathname of source file.
Also discard any function specs that are for :PREVIOUS-DEFINITION properties."
  (SETQ CALLERS (DELETE-IF #'(LAMBDA (X) (AND (CONSP X)
					      (EQ (CAR X) :PROPERTY)
					      (EQ (CADDR X) :PREVIOUS-DEFINITION)))
			   CALLERS))
  (SORT CALLERS #'(LAMBDA (X Y) (LET ((XNAME (CAR (SOURCE-FILE-NAMES X NIL)))
				      (YNAME (CAR (SOURCE-FILE-NAMES Y NIL))))
				  (AND (EQ XNAME YNAME)
				       (SETQ XNAME X
					     YNAME Y))
				  (OR (SYMBOLP XNAME)
				      (TYPEP XNAME 'PATHNAME)
				      (SETQ XNAME ""))
				  (OR (SYMBOLP YNAME)
				      (TYPEP YNAME 'PATHNAME)
				      (SETQ YNAME ""))
				  (STRING-LESSP XNAME YNAME))))) 

;; Small change to this function from patch 94.206.   ddd/gsl 3/4/84.
(DEFUN LIST-ZMACS-CALLERS-TO-BE-EDITED (TYPE FUNCTION JUST-EDIT CALLERS)
  (FUNCALL (IF JUST-EDIT
	       'EDIT-FUNCTIONS-NO-DISPLAY
	       'EDIT-FUNCTIONS-DISPLAY)
	   ;; We want the symbols to show with their proper package prefixes
	   (STABLE-SORT (MAPCAR #'(LAMBDA (X) (CONS (FORMAT NIL "~S" X) X))
				CALLERS)
			#'STRING-LESSP
			:KEY #'CAR)
	   'COM-GO-TO-NEXT-TOP-LEVEL-POSSIBILITY
	   #\c-\p
	   "~A~@[ ~A~]:"
	   "No ~A~@[ ~A~] found."
	   TYPE FUNCTION)) 


;; Small change to this function too from patch 94.206. ddd/gsl 3/4/84.
(DEFCOM COM-FUNCTION-APROPOS "List functions containing the given substring.
Searches the current package, or all packages with control-U, or asks for
a package with two control-Us." ()
  (WITH-TYPEOUT-FONT-MAP-OF ((GET-SEARCH-MINI-BUFFER-WINDOW))
			    (MULTIPLE-VALUE-BIND (PKG PKG-NAME) (GET-PACKAGE-TO-SEARCH)
			      (MULTIPLE-VALUE-BIND (FUNCTION KEY STR)
				  (GET-EXTENDED-SEARCH-STRINGS
				    (FORMAT NIL "List functions in ~A containing substring:" PKG-NAME))
				(LIST-ZMACS-CALLERS-TO-BE-EDITED
				  "Functions matching" STR NIL
				  (SETUP-ZMACS-CALLERS-TO-BE-EDITED
				    (LET ((*FUNCTION* FUNCTION)
					  (*KEY* KEY)
					  (*LIST* NIL))
				      (DECLARE (SPECIAL *FUNCTION* *KEY* *LIST*))
				      (IF (OR (EQ PKG PKG-GLOBAL-PACKAGE) (NULL pkg)) ;; may 7-12-88 added null pkg check
					  (DO-ALL-SYMBOLS (SYM *LIST*)
					    (AND (FUNCALL *FUNCTION* *KEY* (STRING SYM))
								    (FBOUNDP SYM)
								    (PUSH SYM *LIST*)))
					  (DO-LOCAL-SYMBOLS (SYM PKG *LIST*)
					    (AND (FUNCALL *FUNCTION* *KEY* (STRING SYM))
								    (FBOUNDP SYM)
								    (PUSH SYM *LIST*))))))))))
  DIS-NONE) 

(DEFCOM COM-LIST-MATCHING-SYMBOLS "List symbols satisfying the given predicate.
Searches the current package, or all packages with control-U, or asks for
a package with two control-U's." ()
  (MULTIPLE-VALUE-BIND (PKG PKG-NAME) (GET-PACKAGE-TO-SEARCH)
    (LET ((FUNCTION (READ-EVALUATED-MINI-BUFFER
		      "'(LAMBDA (SYMBOL) )" 22
		      "List functions in ~A satisfying: (end with ~C)" PKG-NAME #\END))
	  (SYMBOL (GENSYM)))
      (FDEFINE SYMBOL FUNCTION T NIL)
      (COMPILE SYMBOL)
      (LET ((*PRINT-LENGTH* 3)
	    (*PRINT-LEVEL* 3))
	(LIST-ZMACS-CALLERS-TO-BE-EDITED
	  "Symbols satisfying" FUNCTION NIL
	  (SETUP-ZMACS-CALLERS-TO-BE-EDITED (LIST-MATCHING-SYMBOLS SYMBOL PKG))))))
  DIS-NONE) 

(DEFUN READ-EVALUATED-MINI-BUFFER (&OPTIONAL INITIAL-CONTENTS INITIAL-CHAR-POS FORMAT-STRING
				   &REST FORMAT-ARGS
				   &AUX INTERVAL PROMPT)
  "Read and evaluate an expression in the mini buffer, returning the value.
INITIAL-CONTENTS is a string to start off with, and INITIAL-CHAR-POS if non-NIL
 is where to put the cursor in that string.
FORMAT-STRING and FORMAT-ARGS are used for prompting."
  (SETQ PROMPT (IF (NULL FORMAT-ARGS)
		   FORMAT-STRING
		   (APPLY #'FORMAT NIL FORMAT-STRING FORMAT-ARGS)))
  (UNWIND-PROTECT ;; from patch 94.190  ddd/gsl 3/4/84.
      (PROGN
	(COMMAND-STORE 'COM-INDENT-FOR-LISP #\TAB *MINI-BUFFER-MULTI-LINE-COMTAB*)	;pardon the kludge
	(MULTIPLE-VALUE-SETQ (NIL NIL INTERVAL)
	  (EDIT-IN-MINI-BUFFER *MINI-BUFFER-MULTI-LINE-COMTAB*
			       INITIAL-CONTENTS INITIAL-CHAR-POS
			       (AND PROMPT (CONS PROMPT NIL)))))
    (COMMAND-STORE 'COM-INSERT-TAB #\TAB *MINI-BUFFER-MULTI-LINE-COMTAB*))	;patch 94.190.
  (LET ((FORM-STRING (STRING-INTERVAL INTERVAL))
	(FORM)
	(EOF '(NIL)))
    (SETQ FORM (READ-FROM-STRING FORM-STRING NIL EOF :START 0))
    (AND (EQUAL FORM EOF)
	 (BARF "Unbalanced parentheses."))
    (SI:EVAL-ABORT-TRIVIAL-ERRORS FORM))) 


(DEFUN STRING-FROM-SPEC (SPEC)
  "Return a string for how function spec SPEC would look in a definition a source file.
For example, a (:PROPERTY x y) function spec looks like (x y)."
  (COND ((STRINGP SPEC)
	 SPEC)
	((SYMBOLP SPEC)
	 (SYMBOL-NAME SPEC))
	((ATOM SPEC)
	 (FORMAT NIL "~S" SPEC))
	((MEMBER (CAR SPEC) '(:PROPERTY :METHOD :MAYBE-METHOD :HANDLER) :TEST #'EQ)
	 (DEFINITION-NAME-AS-STRING (CAR SPEC)
				    (IF (EQ (AND (CDDDR SPEC) (THIRD SPEC)) :WRAPPER)
					(LIST (SECOND SPEC) (FOURTH SPEC))
					(CDR SPEC))))
	((MEMBER (CAR SPEC) '(:DEFSTRUCT :DEFSELECT) :TEST #'EQ)
	 (DEFINITION-NAME-AS-STRING (CAR SPEC) (CADR SPEC)))
	((EQ (CAR SPEC) :PROPERTY)
	 (DEFINITION-NAME-AS-STRING :PROPERTY (CDR SPEC)))
	((EQ (CAR SPEC) :INTERNAL)
	 (STRING-FROM-SPEC (CADR SPEC)))
	;; For random function specs, use as given.
	(T
	 (DEFINITION-NAME-AS-STRING NIL SPEC)))) 

(DEFUN DEFINITION-NAME-AS-STRING (TYPE SPEC)
  "Like printing SPEC into a string, but faster, with a few quirks.
The quirk is that if TYPE is non-NIL the printing is done
in the package which the car of SPEC belongs to, if that is possible."
  (LET-IF (AND TYPE
	       (TYPEP (CAR-SAFE SPEC) '(AND SYMBOL (NOT NULL)))
	       (LET ((TEM (SYMBOL-PACKAGE (CAR SPEC))))
		 (AND TEM
		      (NEQ TEM PKG-KEYWORD-PACKAGE)
		      (NEQ TEM PKG-GLOBAL-PACKAGE)
		      (NEQ TEM PKG-SYSTEM-PACKAGE))))
	  ((*PACKAGE* (SYMBOL-PACKAGE (CAR SPEC))))
    (COND ((AND (SYMBOLP SPEC)
		(LOOP WITH PNAME = (SYMBOL-NAME SPEC)
		      FOR I FROM 0 BELOW (LENGTH PNAME)
		      AS CH = (AREF PNAME I)
		      ALWAYS (OR (CHAR<= #\A CH #\Z)
				 (CHAR= CH #\-))))
	   (SYMBOL-NAME SPEC))
	  ((AND (CONSP SPEC)
		(LOOP FOR ELT IN SPEC
		      ALWAYS (AND (SYMBOLP ELT)
				  (LET ((PNAME (SYMBOL-NAME ELT)))
				    (LOOP FOR I FROM 0 BELOW (LENGTH PNAME)
					  AS CH = (AREF PNAME I)
					  ALWAYS (OR (CHAR<= #\A CH #\Z)
						     (CHAR= CH #\-)))))))
	   (LET ((STRING (MAKE-ARRAY 50 :ELEMENT-TYPE 'STRING-CHAR :FILL-POINTER 0)))
	     (VECTOR-PUSH #\( STRING)
	     (LOOP FOR X IN SPEC
		   FOR POS FROM 0 BY 1
		   AS P = (SYMBOL-PACKAGE X)
		   AS PNAME = (SYMBOL-NAME X)
		   DO (OR (ZEROP POS)
			  (STRING-NCONC STRING #\SPACE))
		   (IF (AND (NEQ P *PACKAGE*)
			    (NOT (MEMBER P (PACKAGE-USE-LIST *PACKAGE*) :TEST #'EQ)))
		       (STRING-NCONC STRING
				     (IF (EQ P PKG-KEYWORD-PACKAGE)
					 ""
					 (SI:PKG-SHORTEST-NAME P))
				     #\: PNAME)
		       (STRING-NCONC STRING PNAME)))
	     (STRING-NCONC STRING #\))
	     STRING))
	  (T
	   ;; Not all symbols, stay on the safe side
	   (WITH-OUTPUT-TO-STRING (*STANDARD-OUTPUT*)
	     (PRIN1 SPEC)))))) 

;; may 03/28/89 Added changes to source code.
;;  1/06/89 DNG - Include fix for Scheme DEFINE.  Fix for SETF and LOCF 
;;		function specs.  Fix for DEFWHOPPER.
;;  2/11/89 DNG - Fix for :COND specs.
;;  3/23/89 DNG - Replace test (MEMBER TYPE '(SETF LOCF :COND) :TEST #'EQ) with 
;;		(NOT (MEMBER TYPE '(:METHOD :HANDLER :PROPERTY :INTERNAL) :TEST #'EQ)) so 
;;		that TICLOS:METHOD specs are handled correctly by META-. and we don't have 
;;		to keep patching this for every new kind of function spec.
(DEFUN SYMBOL-FROM-STRING (STR &OPTIONAL LINE OK-TO-ASK SYM &AUX (EOF '(NIL)) ERROR-P)
  "Given a string STR as found after DEF..., return the name of the object being defined.
LINE is the line that the string was found in.  It is used for
finding the particular defining construct used; this affects the result
since (DEFUN (FOO BAR) defines (:PROPERTY FOO BAR)
while (DEFMETHOD (FOO BAR) defines (:METHOD FOO BAR).
OK-TO-ASK means in certain circumstances
where things are not clear, ask the user.  Otherwise we guess.

The arg can also be an object; then its printed representation is used as the string.

The second value is a canonicalized string for the object
 (maybe the same string specified, maybe not).
The third is T if there was a problem
 in parsing the string (such as unbalanced parens).

You can pass the read-in form of the object as the fourth arg
if you already know it."
  (DECLARE (VALUES SYM STR ERROR-P))
  (IF (ARRAYP STR)
      (UNLESS SYM
	(MULTIPLE-VALUE-SETQ (SYM ERROR-P)
			     (CATCH-ERROR (READ-FROM-STRING STR NIL EOF) NIL))
	(SETQ ERROR-P (AND (NULL SYM) ERROR-P)))	; chk 4 error on multi-val funct
							; patch 94.68 ddd/gsl 2/28/84.
      (SETQ SYM STR
	    STR (FORMAT NIL "~S" STR)))
  (COND (ERROR-P
	 (VALUES NIL NIL ERROR-P))
	((SYMBOLP SYM)
	 (VALUES SYM (SYMBOL-NAME SYM)))
	((OR (ATOM SYM) (EQUAL SYM EOF))
	 (VALUES NIL NIL T))
	(T
	 ;; Here SYM is a list.  Certain types of function specs have two ways to
	 ;; type them, with and without the leading type keyword.  Also certain types
	 ;; of functions and other definitions do not follow the standard form
	 ;; of (DEFxxx name options...).  What we do here is to recognize and
	 ;; standardize those cases.  The variables are:
	 ;;	TYPE - the type of function spec or non-function definition
	 ;;	SYM - the function spec or definition name
	 ;;	SPEC - the variant of SYM which appears in the source code
	 ;;	STR - SPEC converted to a string
	 ;; :HANDLER doesn't appear in source files, but gets translated into
	 ;; an appropriate :METHOD here, by analyzing the combined method.
	 ;; :INTERNAL doesn't appear in source files, but might be given as the argument
	 ;; to M-X Disassemble.  The code here just tries not to destory it.
	 (LET ((TYPE (CAR SYM))
	       DELIM-IDX SPEC)
	   (IF (AND (SYMBOLP TYPE)		; Beware of pathological cases.
		    (GET TYPE 'SYSTEM:FUNCTION-SPEC-HANDLER))
	       (IF (AND (NOT (MEMBER TYPE '(:METHOD :HANDLER :PROPERTY :INTERNAL) :TEST #'EQ)) ; handled specially below
			(SYSTEM:VALIDATE-FUNCTION-SPEC SYM))
		   (RETURN-FROM SYMBOL-FROM-STRING (VALUES SYM STR))
		 (SETQ SPEC (CDR SYM)
		       STR (DEFINITION-NAME-AS-STRING TYPE SPEC)))
	       (SETQ SPEC SYM
		     DELIM-IDX (AND LINE (STRING-SEARCH-SET "( " LINE 1))
		     TYPE (COND ((NULL LINE)
				 :MAYBE-METHOD)
				((AND (EQL DELIM-IDX 12)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFMETHOD" 0 12))
				 :ALWAYS-METHOD)
				((AND (EQL DELIM-IDX 13)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFWRAPPER" 0 13))
				 (SETQ SPEC (LIST (CAR SPEC) :WRAPPER (SECOND SPEC)))
				 :ALWAYS-METHOD)
				((AND (EQL DELIM-IDX 12)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFSTRUCT" 0 12))
				 :DEFSTRUCT)
				((AND (EQL DELIM-IDX 12)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFSELECT" 0 12))
				 :DEFSELECT)
				((AND (EQL DELIM-IDX 7)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFINE" 0 7)) ; Scheme DEFINE
				 :DEFSTRUCT) ; use (CAR SPEC)
				((AND (EQL DELIM-IDX 13)
				      (SYS:%STRING-EQUAL LINE 0 "(DEFWHOPPER" 0 13))
				 :ALWAYS-METHOD)
				(T
				 :PROPERTY))))
	   (OR (CASE TYPE
		 (:ALWAYS-METHOD
		  (SETQ SYM (CONS :METHOD SPEC)))
		 ((:METHOD :HANDLER :MAYBE-METHOD)
		  (LET ((FLAVOR (CAR SPEC))
			(MESSAGE (IF (CDDR SPEC)
				     (CADDR SPEC)
				     (CADR SPEC)))
			FL)
		    (COND ((SETQ FL (GET FLAVOR 'SI:FLAVOR)))
			  (OK-TO-ASK
			   (DOLIST (SYMBOL (PACKAGE-LOOKALIKE-SYMBOLS FLAVOR NIL '(SI:FLAVOR)))
			     (IF (FQUERY '(:SELECT T) "Do you mean ~S? "
					 `(:METHOD ,SYMBOL ,@(CDR SPEC)))
				 (RETURN (SETQ FLAVOR SYMBOL
					       SPEC (CONS FLAVOR (CDR SPEC))
					       FL (GET FLAVOR 'SI:FLAVOR)))))))
		    (COND ((SYMBOLP FL)			;T or NIL
			   (AND (EQ TYPE :MAYBE-METHOD)
				(VALIDATE-2-LONG-LIST SPEC)
				(SETQ SYM (CONS :PROPERTY SPEC))))
			  ((FDEFINEDP `(:METHOD . ,SPEC))
			   (SETQ SYM `(:METHOD . ,SPEC)))
			  (OK-TO-ASK
			   (DOLIST (SYMBOL (OR (FIND-COMBINED-METHODS FLAVOR MESSAGE NIL)
					       (SI:FLAVOR-ALL-INHERITABLE-METHODS FLAVOR MESSAGE)))
			     (IF (FQUERY '(:SELECT T) "Do you mean ~S? " SYMBOL)
				 (RETURN (SETQ SYM SYMBOL))))))))
		 (:DEFSTRUCT
		  (SETQ SYM (CAR SPEC))
		  (LOOP WHILE (CONSP SYM) ; this can happen in Scheme DEFINE
			DO (SETQ SYM (CAR SYM)))
		  (SETQ STR (SYMBOL-NAME SYM)))
		 (:DEFSELECT
		  (SETQ SYM (CAR SPEC))
		   (IF (SYMBOLP SYM)
		       (SETQ STR (SYMBOL-NAME SYM))
		       (MULTIPLE-VALUE-SETQ (SYM STR)
			 (SYMBOL-FROM-STRING SYM))))
		 (:PROPERTY
		  (AND (VALIDATE-2-LONG-LIST SPEC)
		       (SETQ SYM (CONS TYPE SPEC))))
		 (:INTERNAL
		  (SETQ SYM (CONS TYPE SPEC))
		  (SETQ STR (DEFINITION-NAME-AS-STRING NIL (CAR SPEC)))))
	       ;; Something we don't understand, make a bogus symbol to use as a property
	       ;; list to remember the location of this definition
	       (SETQ SYM (INTERN STR *UTILITY-PACKAGE*))))
	 (IF (NOT (SYSTEM:VALIDATE-FUNCTION-SPEC SYM))
	     (VALUES NIL NIL T)
	     (VALUES SYM STR)))))
 

(DEFUN VALIDATE-2-LONG-LIST (L)
  "T if L is a list of exactly two symbols."
  (AND (CONSP L)
       (SYMBOLP (CAR L))
       (CONSP (CDR L))
       (SYMBOLP (CADR L))
       (NULL (CDDR L)))) 

;; may 03/28/89 Added for CLOS (for DNG)
;; DNG 04/20/89 - Redesigned to use EXTRACT-SPECIALIZER-NAMES and to fix to 
;;		evaluate EQL specializers.
(defun arglist-type-specifiers ( arglist)
  (let ((specs (ticlos:extract-specializer-names arglist)))
    (dolist (spec specs)
      (when (and (consp spec)
		 (eq (first spec) 'eql)
		 (null (cddr spec)))
	(ignore-errors
	  (setf (second spec) (eval (second spec))))))
    specs))


;; may 03/28/89 Added for CLOS (for DNG) - modified from GET-DEFUN-NAME 
(DEFUN GET-DEFmethod-arglist (BP &AUX BP1 str idx arglist x)
  (declare (values list string error-p))
  ;; The first value returned should be: (function-spec {qualifier}* arglist)
  ;;  3/9/89 DNG - Fixed to not error if the arglist can't be found.
  (AND (SETQ BP (FORWARD-ATOM BP))		;skip defmethod
       (SETQ BP (FORWARD-OVER *WHITESPACE-CHARS* BP))
       (setq bp1 bp)
       (loop
	 (unless (and (SETQ BP1 (forward-sexp bp1))
		      (setq bp1 (SKIP-OVER-BLANK-LINES-AND-COMMENTS Bp1))
		      (setq bp1 (forward-over *blanks* bp1))
		      (not (char= (BP-CH-CHAR BP1) #\()))
	   (if  (and BP1 (char= (BP-CH-CHAR BP1) #\())
		(progn
		  (setq bp1 (FORWARD-SEXP BP1)
			idx 0)
		  (or (and bp1
			    (setf str (STRING-REMOVE-FONTS (STRING-INTERVAL BP BP1)))
			    (ignore-errors
			       (return
				 (values
				   (loop (setf (values x idx)
					       (read-from-string str nil 'eof :start idx))
					 (if (neq x 'eof)
					     (push x arglist)
					   (if (listp (first arglist))
					        (return (nreverse arglist))
					      ;; else not a valid arglist
					      (return-from get-defmethod-arglist (values nil nil t)))
					      ))
				   str nil))))
		       (return (values nil nil t))))
	     (return (values nil nil t)))))))

;The properties this uses are defined in MODES.
;The mode's property is probably :LISP, :TEXT or NIL.
;Note that some things open-code this so they can do the GETs only once.
(DEFUN GET-SECTION-NAME (MODE LINE BP)
  "Determine whether LINE is a definition line, and if so return the name defined.
Determination is made according to MODE (which might be, eg, LISP-MODE).
BP should be a temporary BP that can be modified by this function.
The first value is the symbol or function spec defined, or NIL.
The second value is the string which appears in the line to specify that symbol or spec.
The third value is T if there is no definition on the line."
  (DECLARE (VALUES SYM STR ERROR-P))
  (FUNCALL (GET (GET MODE 'EDITING-TYPE) 'GET-SECTION-NAME)
	   LINE BP)) 

;And here are the functions that actually implement GET-SECTION-NAME
;for various well-known modes.

;; may 05-11-89 Provide temp fdefinition until ticlos is loaded for build.
(unless (fboundp 'ticlos:flavor-method-spec-p)
  (setf (symbol-function 'ticlos:flavor-method-spec-p) #'(lambda (sym) t)))

;; may 03/28/89 Added to source code
;; 10/17/88 DNG Changed 'GLOBAL:STRING to 'STRING to enable optimization of CONCATENATE.
;;  1/06/89 DNG Fixed DEFMETHOD handling to work for both CLOS and Flavor methods.
(DEFUN (:PROPERTY :LISP GET-SECTION-NAME) (LINE BP &AUX STR SYM ERROR-P
					   IDX END-IDX (EOF "") NON-FONT-LINE )
  (IF (NOT (AND (> (LENGTH LINE) 1)
		(CHAR= (MAKE-CHAR (AREF LINE 0)) #\()))
      (VALUES NIL NIL T)
      (PROGN
	(SETQ ERROR-P T)
	(WHEN (AND (SYS:%STRING-EQUAL LINE 0 "(DEF" 0 4)
		   (NOT (SYS:%STRING-EQUAL LINE 0 "(DEFPROP " 0 11))
		   (SETQ IDX (STRING-SEARCH-SET *WHITESPACE-CHARS* LINE))
		   (SETQ IDX (STRING-SEARCH-NOT-SET *WHITESPACE-CHARS* LINE IDX)))
	  (SETQ ERROR-P NIL)
	  (SETQ NON-FONT-LINE (STRING-REMOVE-FONTS LINE))
	  (CONDITION-CASE () (SETF (VALUES SYM END-IDX)
				   (READ-FROM-STRING NON-FONT-LINE NIL EOF :START IDX))
	    (:NO-ERROR
	     (IF (EQUAL SYM EOF)
		 (SETQ ERROR-P T)
		 (SETQ STR (SUBSEQ NON-FONT-LINE IDX (MIN (LENGTH LINE) END-IDX)))))
	    (SYSTEM:READ-ERROR
	     (SETQ STR (GET-DEFUN-NAME (MOVE-BP BP LINE 0)))))
	  (UNLESS ERROR-P
	    (if (and (sys:%string-equal NON-FONT-LINE 0 "(DEFMETHOD " 0 13)
		     (not (ticlos:flavor-method-spec-p sym)))
		;; Looks like a CLOS method definition.
		(progn 
		  (setf (values sym str error-p)
			(GET-DEFmethod-arglist (MOVE-BP BP LINE 0)))
		  (and (and sym (null error-p))
		       (progn
			 (setf (car (last sym)) (arglist-type-specifiers (car (last sym))))
			 (setf sym `(ticlos:method ,@sym)))))
		(MULTIPLE-VALUE-SETQ (SYM NIL ERROR-P)
		  (SYMBOL-FROM-STRING STR NON-FONT-LINE NIL SYM)))))
	(WHEN ERROR-P
	  (SETQ SYM (CONCATENATE 'STRING
				 (THE STRING
				      (LET ((BUFFER (NODE-TOP-LEVEL-NODE (LINE-NODE LINE))))
					(IF (BUFFER-PATHNAME BUFFER)
					    (LET ((NAME (PATHNAME-NAME (BUFFER-PATHNAME BUFFER))))
					      (IF (CONSP NAME)
						  (APPLY 'STRING-APPEND
							 (MAPCAR #'(LAMBDA (NAME-ELT) (IF (CONSP NAME-ELT)
											  (CAR NAME-ELT)
											  NAME-ELT))
								 NAME))
						  (STRING NAME)))
					    (BUFFER-NAME BUFFER))))
				 "-"
				 (THE STRING
				      (LET ((START-INDEX (POSITION #\(
								   (THE STRING (STRING LINE))
								   :TEST-NOT #'CHAR-EQUAL)))
					(SUBSEQ LINE START-INDEX
						(AND START-INDEX
						     (STRING-SEARCH-SET *WHITESPACE-CHARS*
									LINE
									START-INDEX)))))
				 "-"
				 (PRIN1-TO-STRING (INCF *SECTION-COUNT*)))
		STR SYM))
	(VALUES SYM STR NIL)))) 

(DEFUN GET-DEFUN-NAME (BP &AUX BP1)
  "Return the function spec defined by the defun starting at BP."
  (AND (SETQ BP (FORWARD-ATOM BP))
       (SETQ BP (FORWARD-OVER *WHITESPACE-CHARS* BP))
       (SETQ BP1 (FORWARD-SEXP BP))
       (STRING-REMOVE-FONTS (STRING-INTERVAL BP BP1)))) 

(DEFUN (:PROPERTY :TEXT GET-SECTION-NAME) (LINE BP &AUX STR SYM ERROR-P)
  (IF (AND (SYS:%STRING-EQUAL LINE 0 ".DEF" 0 4)
	   (SETQ STR (GET-TEXT-DEFUN-NAME (MOVE-BP BP LINE 0))))
      (SETQ SYM (INTERN STR *UTILITY-PACKAGE*))
      (SETQ ERROR-P T))
  (VALUES SYM STR ERROR-P)) 

(DEFUN GET-TEXT-DEFUN-NAME (BP &AUX BP1)
  (ATOM-WORD-SYNTAX-BIND
    ;; Now get the second word after BP.
    (AND (SETQ BP (FORWARD-WORD BP))
	 (SETQ BP (FORWARD-OVER *BLANKS* BP))
	 (SETQ BP1 (FORWARD-WORD BP))
	 (STRING-REMOVE-FONTS (STRING-INTERVAL BP BP1))))) 

(DEFUN (:PROPERTY NIL GET-SECTION-NAME) (LINE BP)
  LINE
  BP
  (VALUES NIL NIL T)) 

;;; The SECTION-P is a function which takes a line as arg
;;; and returns T if the line looks likely to be the start of a section.
;;; You could use the GET-SECTION-NAME function, but that will not be
;;; guaranteed to be right if the following line(s) are not there,
;;; and if it is wrong it will return "no section".
;;; These functions err by returning T if in doubt;
;;; then you can use the GET-SECTION-NAME function after you prepare for it.
(DEFUN (:PROPERTY :LISP SECTION-P) (LINE)
  (AND (PLUSP (LENGTH LINE))
       (CHAR= (MAKE-CHAR (AREF LINE 0)) #\())) 

(DEFUN (:PROPERTY :TEXT SECTION-P) (LINE)
  (SYS:%STRING-EQUAL LINE 0 ".DEF" 0 4)) 

(DEFUN (:PROPERTY NIL SECTION-P) (IGNORE)
  NIL) 

(DEFUN DEFINITION-LIKELY-POSITION (BP1 &OPTIONAL BP2 IN-ORDER-P SYMBOL
				   &AUX DEFUN-LINE NON-DEFUN-LINE COMMENT-LINE BUFFER)
  "Return a BP within specified interval to a line that might relate to defining SYMBOL.
SYMBOL is actually a function spec or a name being defined in any way.
The line is found by a textual search for SYMBOL.
However, lines starting with open parens get first priority.
Comment lines have lower priority than non-comment lines.
If nothing is found, the beginning of the interval is returned."
  (GET-INTERVAL BP1 BP2 IN-ORDER-P)
  (DO ((LINE (BP-LINE BP1) (LINE-NEXT LINE))
       (FROM-INDEX (BP-INDEX BP1) 0)
       (END-LINE (BP-LINE BP2))
       (KEY (STRING-FROM-SPEC SYMBOL))
       (INDEX)
       (TEM))
      (NIL)
    (AND (SETQ INDEX (GLOBAL:STRING-SEARCH KEY LINE FROM-INDEX (AND (EQ LINE END-LINE)
                                                                    (BP-INDEX BP2))))
	 (COND ((CHAR-EQUAL #\( (AREF LINE 0))
		(SETQ DEFUN-LINE LINE)
		(RETURN T))
	       ((AND (NULL COMMENT-LINE)
		     (SETQ TEM (FIND-COMMENT-START LINE))
		     (< TEM INDEX))
		(SETQ COMMENT-LINE LINE))
	       ((NULL NON-DEFUN-LINE)
		(SETQ NON-DEFUN-LINE LINE))))
    (AND (EQ LINE END-LINE)
	 (RETURN NIL)))
  (SETQ BUFFER (BP-TOP-LEVEL-NODE BP1))
  (VALUES (CREATE-BP (OR DEFUN-LINE
			 NON-DEFUN-LINE
			 COMMENT-LINE
			 (BP-LINE BP1))
		     0)
	  BUFFER)) 

(DEFUN PACKAGE-LOOKALIKE-SYMBOLS (PNAME
				  &OPTIONAL IGNORE
				  (PROPERTIES '(:SOURCE-FILE-NAME ZMACS-BUFFERS))
				  &AUX LIST)
  "Return a list of symbols with pname PNAME in various packages.
Only symbols which possess one of the properties in PROPERTIES are considered."
  (DOLIST (PKG (LIST-ALL-PACKAGES))
    (SETQ LIST (PACKAGE-LOOKALIKE-SYMBOLS-1 (STRING PNAME) PKG LIST PROPERTIES)))
  LIST) 

(DEFUN PACKAGE-LOOKALIKE-SYMBOLS-1 (PNAME PKG LIST PROPERTIES)
  (LET ((TEM (MULTIPLE-VALUE-BIND (SYMBOL STATUS) (FIND-SYMBOL PNAME PKG)
	       (AND (NEQ STATUS :INHERITED)
		    SYMBOL))))
    (AND TEM
	 (NOT (MEMBER TEM LIST :TEST #'EQ))
	 ;; Used to be GETL, but that was fooled by a property of NIL
	 (LOOP FOR PROP IN PROPERTIES
	       THEREIS (GET TEM PROP))
	 (PUSH TEM LIST)))
  LIST) 

(DEFUN DEFINITION-TEXT-LOCATION-1 (SPEC)
  "Return a BP to SPEC's definition in ZMACS, if it has a unique definition.
If SPEC has no definition or more than one, return NIL.
Will read in a source file if there is a unique source file.
Does not select the buffer, and can be called from outside the editor.
If there is more than one definition, the user should be advised to
enter ZMACS, give C-M-X Make Correspondence on the definition he prefers,
and retry your operation."
  (DECLARE (VALUES BP BUFFER))
  (LET ((CHOSEN-DEF (SI:FUNCTION-SPEC-GET SPEC 'ZMACS-CHOSEN-DEFINITION))
	(ZBP (SI:FUNCTION-SPEC-GET SPEC 'ZMACS-BUFFERS)))
    (COND (CHOSEN-DEF
	   (VALUES (CREATE-BP (CDR CHOSEN-DEF) 0) (CAR CHOSEN-DEF)))
	  (ZBP
	   (AND (NULL (CDR ZBP))
		(NOT (BUFFER-IS-NOT-ONLY-SOURCE-FILE-P (CAAR ZBP) SPEC))
		(IF (DEFINITION-STILL-REAL-P-NEW (CAAR ZBP) (CDAR ZBP) SPEC)
		    (VALUES (CREATE-BP (CDAR ZBP) 0) (CAAR ZBP))
		    (LET ((*PACKAGE* *PACKAGE*))
		      (COMPUTE-BUFFER-PACKAGE (CAAR ZBP))
		      (SECTIONIZE-BUFFER (CAAR ZBP))
		      (DEFINITION-TEXT-LOCATION-1 SPEC)))))
	  (T
	   (LET ((SOURCES (REMOVE-IF 'FIND-BUFFER-NAMED (SOURCE-FILE-NAMES SPEC))))
	     (IF (AND SOURCES (NULL (CDR SOURCES)))
		 (PROGN
		   (FIND-FILE (CAR SOURCES))
		   (DEFINITION-TEXT-LOCATION-1 SPEC)))))))) 


;;; Tag table stuff
(DEFCOM COM-VISIT-TAG-TABLE
	"Read in the specified tag table file.
Go through the tag table, and mark the name of each tag as being
a possible section of its file.  Later, the Edit Definition command
will see these marks and figure out which file to use.
Get the name of the file from the mini-buffer." ()
  (READ-TAG-TABLE (READ-DEFAULTED-PATHNAME "Tag Table:" (PATHNAME-DEFAULTS) "TAGS" :NEWEST))
  DIS-NONE) 

;;; READ-TAG-TABLE fixed to not go into EH if no name is specified for a simple-tag-table. 
;;; That is, the first line merely has to say "in " or "files in " without actually following 
;;; either string with a name for the tag table. The name that we will use for it will come
;;; from (gensym) and look odd, but business will be carried on as usual. 
(DEFUN READ-TAG-TABLE (FILE &AUX IN-POS)
  "This is called by the revert method on tag table buffers and reads in
a tag table file for the visit tag table command.

The old format tag table lists files for the tag table by file.type
only, unquoted.  The directory is assumed to be the same as that of the
tag table file.  There is no way to change this in the old format.  This
first file line should contain nothing else.  The following line should
end with a comma and a mode (i.e. lisp, text, bolio, or R).  No other
commas should appear in this line.  Following this there may 0 or more
lines of the form: '(def <symbol>' or '.def <symbol>'.
The form which begins with ( is used if lisp mode was specified.  The
form which begins with . is used for the other modes.  The 's are my
delimiters and should not appear in the file's line.  These symbols are
associated with the preceeding pathname for Meta-. (Edit Definition
command).  They are also added to the completions aarray.  The sequence
that began with the pathname is terminated with a line which contains a
 (Symbol-W) in column 0.  Any number of such sequences of lines
beginning with a file and ending with a  may appear in the file.  No
blank lines may appear in the file.  Note:  is inserted into a
buffer with the keystrokes Ctrl-Q 2 6 1.

The simpler format is recognized by the presence of the string ' in ' in
the first line of the file.  This allows Tag Table buffers to be written
to disk and then recognized as new format tag table files by the visit
tag table command.  The files may be written by hand also, of course.
The first line of the file is otherwise ignored in this format.  The
succeeding lines should each contain one file name.  The file names may
be complete, but if not will be defaulted in the standard ways.  This
format allows blank lines (lines containing only characters in
zwei:*WHITESPACE-CHARS*)."
  (WITH-OPEN-FILE-CASE (STREAM FILE)
    (:NO-ERROR
     (LET ((LINE)
	   (EOF))
       (MULTIPLE-VALUE-SETQ (LINE EOF)
	 (SEND STREAM :LINE-IN))
       (COND (EOF
	      NIL)
	     ((SETQ IN-POS (LISP:SEARCH  "in " (STRING LINE) :TEST #'CHAR-EQUAL))
	      (LET* ((NAME-BEGINS (STRING-SEARCH-NOT-SET *WHITESPACE-CHARS* LINE (+ 3 IN-POS)))
		     (NAME (AND NAME-BEGINS	; Allow for no name actually specified.
				(SUBSEQ LINE
					NAME-BEGINS
					(STRING-SEARCH-SET *WHITESPACE-CHARS* LINE NAME-BEGINS)))))
		(READ-SIMPLE-TAG-TABLE STREAM (IF (ZEROP (LENGTH NAME))
						  (STRING (GENSYM))
						  NAME))))
	     (T
	      (READ-OLD-FORMAT-TAG-TABLE STREAM LINE FILE)))))
    (FS:FILE-NOT-FOUND (BARF STREAM))
    (FS:FILE-ERROR (BARF STREAM)))) 
     
(DEFUN READ-OLD-FORMAT-TAG-TABLE (STREAM LINE FILE &AUX
				  (ADDED-COMPLETIONS (MAKE-ARRAY 1000 :TYPE 'ART-Q-LIST :LEADER-LENGTH 2)))
  (STORE-ARRAY-LEADER 0 ADDED-COMPLETIONS 0)
  (DO ((EOF)
       (FILE-LIST)
       (PATHNAME)
       (MODE))
      (NIL)
    (UNLESS LINE
      (MULTIPLE-VALUE-SETQ (LINE EOF)
			   (SEND STREAM :LINE-IN)))
    (COND (EOF
	   (SEND FILE :PUTPROP (NREVERSE FILE-LIST) 'ZMACS-TAG-TABLE-FILE-SYMBOLS)
	   (OR (RASSOC FILE *ZMACS-TAG-TABLE-ALIST* :TEST #'EQ)
	       (PUSH (CONS (STRING FILE) FILE)
		     *ZMACS-TAG-TABLE-ALIST*))
	   (RETURN)))
    (SETQ PATHNAME (FS:MERGE-PATHNAME-DEFAULTS LINE *PATHNAME-DEFAULTS*))
    (PUSH PATHNAME FILE-LIST)
    (SETQ LINE (SEND STREAM :LINE-IN))			;Length,Mode
    (SETQ MODE (GET-FILE-MAJOR-MODE
		 (INTERN (SUBSEQ LINE
				    (1+ (OR (POSITION #\, (THE STRING (STRING LINE)) :TEST #'CHAR-EQUAL)
					    (BARF "Line after file name needs comma and mode."))))
			 "USER")))
    (DO ((*PACKAGE* (FIND-PACKAGE (OR (SEND (SEND PATHNAME :GENERIC-PATHNAME)
					    :GET :PACKAGE)
				      *PACKAGE*)))
	 (SPACE-POS)
	 (RUBOUT-POS)
	 (STR)
	 (SNAME))
	((CHAR= (AREF (SETQ LINE (SEND STREAM :LINE-IN)) 0)
		#\)
	 (SETQ LINE NIL))
      (COND ((SETQ SPACE-POS (STRING-SEARCH-SET '(#\SPACE #\TAB) LINE))
	     (SETQ SPACE-POS (1+ SPACE-POS)
		   RUBOUT-POS (COND ((POSITION #\ (THE STRING (STRING LINE))
					       :START SPACE-POS
					       :TEST #'CHAR-EQUAL))
				    (T
				     (SEND STREAM :LINE-IN)
				     (1+ (LENGTH LINE))))
		   STR (SUBSEQ LINE SPACE-POS (1- RUBOUT-POS)))
	     (COND ((CASE (GET MODE 'EDITING-TYPE)
			  (:LISP
			   (AND (SYS:%STRING-EQUAL LINE 0 "(DEF" 0 4)
				(NOT (SYS:%STRING-EQUAL LINE 0 "(DEFPROP " 0 11))
				(SETQ SNAME (SYMBOL-FROM-STRING STR LINE))))
			  (:TEXT
			   (AND (SYS:%STRING-EQUAL LINE 0 ".DEF" 0 4)
				(SETQ SNAME (INTERN STR *UTILITY-PACKAGE*))))
			  (OTHERWISE NIL))
		    (SECTION-COMPLETION SNAME STR ADDED-COMPLETIONS 1000)
		    (SECTION-COMPLETION SNAME (DEFINITION-NAME-AS-STRING NIL SNAME) ADDED-COMPLETIONS 1000)
		    (OR (GET SNAME :SOURCE-FILE-NAME)
			(SETF (GET SNAME :SOURCE-FILE-NAME) PATHNAME))
		    (PUSH* PATHNAME (GET SNAME 'ZMACS-TAG-FILE-SYMBOLS))))))))
  (SORT-COMPLETION-AARRAY ADDED-COMPLETIONS)
  (MERGE-COMPLETION-AARRAY *ZMACS-COMPLETION-AARRAY* ADDED-COMPLETIONS)) 

(DEFUN READ-SIMPLE-TAG-TABLE (STREAM NAME) ;;gsl
  (IF (ASSOC NAME *ZMACS-TAG-TABLE-ALIST* :TEST #'EQUAL) ;;no duplicates please.
      (WITHOUT-INTERRUPTS
	(SETQ *ZMACS-TAG-TABLE-ALIST* (DELETE (ASSOC NAME *ZMACS-TAG-TABLE-ALIST* :TEST #'EQUAL)
					      (THE LIST *ZMACS-TAG-TABLE-ALIST*)
					      :TEST #'EQ))))
  (DO ((LINE)
       (EOF)
       (FILE-LIST)
       (PATHNAME)
       FIRST-NON-BLANK
       NEXT-BLANK)
      (EOF (SELECT-FILE-LIST-AS-TAG-TABLE (REVERSE FILE-LIST) NAME))
    (MULTIPLE-VALUE-SETQ (LINE EOF)
      (SEND STREAM :LINE-IN))
    (AND (SETQ FIRST-NON-BLANK (STRING-SEARCH-NOT-SET *WHITESPACE-CHARS* LINE))
	 (SETQ NEXT-BLANK (OR (STRING-SEARCH-SET *WHITESPACE-CHARS* LINE FIRST-NON-BLANK)
			      (LENGTH LINE)))
	 (NOT (= FIRST-NON-BLANK NEXT-BLANK))
	 (SETQ PATHNAME (MAKE-DEFAULTED-PATHNAME (NSUBSTRING LINE FIRST-NON-BLANK NEXT-BLANK)
						 *PATHNAME-DEFAULTS*))
	 (PUSH PATHNAME FILE-LIST)))) 

(DEFCOM COM-LIST-TAG-TABLES "List the names of all the tag table files read in" ()
  (DOLIST (TAG-TABLE *ZMACS-TAG-TABLE-ALIST*)
    (FORMAT T "~&~4TFiles in tag table ~A:~%" (CAR TAG-TABLE))
    (SEND *STANDARD-OUTPUT* :ITEM-LIST 'FILE
	  (SEND (CDR TAG-TABLE) :GET 'ZMACS-TAG-TABLE-FILE-SYMBOLS)))
  (FORMAT T "~&Done.~%")
  DIS-NONE) 

(DEFCOM COM-CURRENT-TAG-TABLE "Displays the name of the current tag table in the echo area." ()
  (IF *ZMACS-CURRENT-TAG-TABLE*
      (FORMAT *TYPEIN-WINDOW* "The current tag table is ~A."
	      (SEND *ZMACS-CURRENT-TAG-TABLE* :NAME))
      (FORMAT *TYPEIN-WINDOW* "There is no current tag table."))
  DIS-NONE) 

(DEFCOM COM-NEXT-FILE "Move to the next file in the tags table" ()
  (NEXT-FILE *NUMERIC-ARG-P*)
  DIS-TEXT) 

(DEFUN NEXT-FILE (RESTART &AUX PATHNAME BUFFER)
  "Select the next file in the selected tag table.
RESTART non-NIL means select the first file in the tag table
and reset the list of files to be gone through."
  (AND RESTART
       (SETQ *ZMACS-LAST-TAGS-FILE-LIST*
	     (SEND (SELECT-TAG-TABLE) :GET 'ZMACS-TAG-TABLE-FILE-SYMBOLS)))
  (OR *ZMACS-LAST-TAGS-FILE-LIST*
      (BARF "No more files"))
  (POP *ZMACS-LAST-TAGS-FILE-LIST* PATHNAME)
  (COND ((SETQ BUFFER (FIND-FILE-BUFFER PATHNAME))
	 (FORMAT *QUERY-IO* "~&~A~%" PATHNAME)
	 (MAKE-BUFFER-CURRENT BUFFER)
	 (MOVE-BP (POINT) (INTERVAL-FIRST-BP *INTERVAL*)))
	(T
	 (FIND-FILE PATHNAME)))) 

(DEFUN NEXT-FILE-BP (RESTART &AUX PATHNAME BUFFER)
  "Return BP to start of the next file in the selected tag table.
RESTART non-NIL means start again at first file in tag table."
  (AND RESTART
       (SETQ *ZMACS-LAST-TAGS-FILE-LIST*
	     (SEND (SELECT-TAG-TABLE) :GET 'ZMACS-TAG-TABLE-FILE-SYMBOLS)))
  (OR *ZMACS-LAST-TAGS-FILE-LIST*
      (BARF "No more files"))
  (POP *ZMACS-LAST-TAGS-FILE-LIST* PATHNAME)
  (COND ((SETQ BUFFER (FIND-FILE-BUFFER PATHNAME))
	 (FORMAT *QUERY-IO* "~&~A~%" PATHNAME)
	 (INTERVAL-FIRST-BP BUFFER))
	(T
	 (INTERVAL-FIRST-BP (FIND-FILE PATHNAME NIL))))) 

(DEFUN TAG-TABLE-BUFFERS (READ-IN-ALL-FILES &AUX BUFFER-LIST FILE-LIST)
  "Return a list of all buffers in the selected tag table.
READ-IN-ALL-FILES means visit all the files;
otherwise, we return only the buffers for files already read in."
  (SETQ FILE-LIST (SEND (SELECT-TAG-TABLE) :GET 'ZMACS-TAG-TABLE-FILE-SYMBOLS))
  (DOLIST (FILE FILE-LIST)
    (LET ((BUFFER (FIND-FILE-BUFFER FILE)))
      (IF BUFFER
	  (PUSH BUFFER BUFFER-LIST))
      (IF READ-IN-ALL-FILES
	  (PUSH (FIND-FILE FILE NIL)
		BUFFER-LIST))))
  (NREVERSE BUFFER-LIST)) 

(DEFCOM COM-TAGS-SEARCH "Search for the specified string within files of the tags table" ()
  (LET ((*MINI-BUFFER-DEFAULT-STRING* *ZMACS-TAGS-SEARCH-KEY-STRING*))
    (MULTIPLE-VALUE-SETQ (*ZMACS-TAGS-SEARCH-FUNCTION* *ZMACS-TAGS-SEARCH-KEY*)
      (GET-EXTENDED-STRING-SEARCH-STRINGS NIL "Tags search:" *SEARCH-MINI-BUFFER-COMTAB*)))
  (SETQ *ZMACS-TAGS-SEARCH-KEY-STRING* *EXTENDED-STRING-SEARCH-LAST-STRING*)
  (MUST-REDISPLAY *WINDOW* DIS-TEXT)
  (TAGS-SEARCH-NEXT-OCCURRENCE T)) 

(DEFUN TAGS-SEARCH-ALTERNATIVE-STRINGS (&REST STRINGS)
  "Begin a tags search looking for any of STRINGS."
  (IF (= (LENGTH STRINGS) 1)
      (SETQ *ZMACS-TAGS-SEARCH-KEY-STRING* (CAR STRINGS)
	    *ZMACS-TAGS-SEARCH-KEY* (CAR STRINGS)
	    *ZMACS-TAGS-SEARCH-FUNCTION* 'SEARCH)
      (PROGN
	(SETQ *ZMACS-TAGS-SEARCH-KEY* (COPY-LIST STRINGS))
	(SETQ *ZMACS-TAGS-SEARCH-FUNCTION* 'FSM-SEARCH)
	(SETQ *ZMACS-TAGS-SEARCH-KEY-STRING*
	      (MAKE-ARRAY 20 :ELEMENT-TYPE 'SI:FAT-CHAR :LEADER-LIST '(0)))
	(DO ((STRINGS STRINGS (CDR STRINGS)))
	    ((NULL STRINGS))
	  (STRING-NCONC *ZMACS-TAGS-SEARCH-KEY-STRING* (CAR STRINGS))
	  (OR (NULL (CDR STRINGS))
	      (STRING-NCONC *ZMACS-TAGS-SEARCH-KEY-STRING*
			    (INT-CHAR 402))))))
  (MUST-REDISPLAY *WINDOW* DIS-TEXT)
  (TAGS-SEARCH-NEXT-OCCURRENCE T)) 

(DEFCOM COM-TAGS-SEARCH-NEXT-OCCURRENCE "Search for the next occurrence of search string" ()
   (TAGS-SEARCH-NEXT-OCCURRENCE NIL)) 

(DEFUN TAGS-SEARCH-NEXT-OCCURRENCE (RESTART)
  (DO ((BP)
       (PT (IF RESTART (NEXT-FILE-BP T) (POINT))))
      (NIL)
    (LET ((*INTERVAL* (BP-TOP-LEVEL-NODE PT)))
      (SETQ BP (FUNCALL *ZMACS-TAGS-SEARCH-FUNCTION* PT *ZMACS-TAGS-SEARCH-KEY*)))
    (COND (BP
	   (POINT-PDL-PUSH (POINT) *WINDOW*)
	   (MAKE-BUFFER-CURRENT (BP-TOP-LEVEL-NODE BP))
	   (MOVE-BP (POINT) BP)
	   (WITH-BORROWED-FONT-MAP-OF ((GET-SEARCH-MINI-BUFFER-WINDOW) *TYPEIN-WINDOW*)
	     (TYPEIN-LINE "~&Use ~A to find next occurrence of \"~A\".~&"
			  (IF RESTART
			      (KEY-FOR-COMMAND-SET-C-. 'COM-TAGS-SEARCH-NEXT-OCCURRENCE)
			      (KEY-FOR-COMMAND 'COM-TAGS-SEARCH-NEXT-OCCURRENCE *COMTAB* NIL NIL #\c-.))
			  *ZMACS-TAGS-SEARCH-KEY-STRING*))	;my addition gsl 3-26-85
	   (RETURN DIS-TEXT))
	  (T
	   (SETQ PT (NEXT-FILE-BP NIL))
	   (MUST-REDISPLAY *WINDOW* DIS-TEXT))))) 

(DEFPROP COM-TAGS-QUERY-REPLACE
	 KIND-OF-QUERY-REPLACE-DOCUMENTATION
	 DOCUMENTATION-FUNCTION) 
(DEFPROP COM-TAGS-QUERY-REPLACE T CONTROL-PERIOD) 

(DEFCOM COM-TAGS-QUERY-REPLACE "Perform a query replace within the tags table files.
Does Query Replace over all the file in the selected tags table, one file at a time." ()
  (MULTIPLE-VALUE-SETQ (*TAGS-QUERY-REPLACE-FROM* *TAGS-QUERY-REPLACE-TO*)
    (QUERY-REPLACE-STRINGS NIL))
  (SETQ *TAGS-QUERY-REPLACE-DELIMITED* (AND *NUMERIC-ARG-P* *NUMERIC-ARG*))
  (CONTINUE-TAGS-QUERY-REPLACE T)) 

(DEFCOM COM-CONTINUE-TAGS-QUERY-REPLACE "Continue the last tags query replace" ()
   (CONTINUE-TAGS-QUERY-REPLACE NIL)) 

(DEFUN CONTINUE-TAGS-QUERY-REPLACE (RESTART)
  (DO ((BEGINNING RESTART)
       (VAL))
      (NIL)
    ;; Find the next buffer in the list which has an occurrence, select it.
    (DO (BP FOUND-AT) (NIL)
      (SETQ BP (NEXT-FILE-BP BEGINNING))
      (SETQ BEGINNING NIL)
      (WHEN (LET ((*INTERVAL* (BP-TOP-LEVEL-NODE BP)))
	      (SETQ FOUND-AT (SEARCH BP *TAGS-QUERY-REPLACE-FROM*)))
	(POINT-PDL-PUSH (POINT) *WINDOW* NIL NIL)
	(MAKE-BUFFER-CURRENT (BP-TOP-LEVEL-NODE BP))
	;; Move bp to just before the occurrence, so we avoid
	;; re-scanning the part of the buffer already searched over.
	(MOVE-BP (POINT) (FORWARD-CHAR FOUND-AT (- (LENGTH *TAGS-QUERY-REPLACE-FROM*))))
	(RETURN)))
    (MUST-REDISPLAY *WINDOW* DIS-TEXT)
    ;; Query replace thru that buffer.
    (SETQ VAL (QUERY-REPLACE (POINT) (INTERVAL-LAST-BP *INTERVAL*)
			     *TAGS-QUERY-REPLACE-FROM*
			     *TAGS-QUERY-REPLACE-TO*
			     *TAGS-QUERY-REPLACE-DELIMITED*))
    (WHEN (EQ VAL 'ABORTED)
      (TYPEIN-LINE "~&Use ~A to continue the Tags Query Replace."
		   (IF RESTART
		       (KEY-FOR-COMMAND-SET-C-. 'COM-CONTINUE-TAGS-QUERY-REPLACE)
		       (KEY-FOR-COMMAND 'COM-CONTINUE-TAGS-QUERY-REPLACE *COMTAB* NIL NIL #\c-.)))
      (RETURN DIS-TEXT)))) 

(DEFPROP COM-TAGS-MULTIPLE-QUERY-REPLACE
	 KIND-OF-QUERY-REPLACE-DOCUMENTATION
	 DOCUMENTATION-FUNCTION) 
(DEFPROP COM-TAGS-MULTIPLE-QUERY-REPLACE T CONTROL-PERIOD) 

(DEFCOM COM-TAGS-MULTIPLE-QUERY-REPLACE "Perform a query replace within the tags table files" ()
  (MULTIPLE-VALUE-BIND (FROM-LIST TO-LIST)
      (MULTIPLE-QUERY-REPLACE-STRINGS NIL)
    (TAGS-MULTIPLE-QUERY-REPLACE FROM-LIST TO-LIST (AND *NUMERIC-ARG-P* *NUMERIC-ARG*)))) 

(DEFPROP COM-TAGS-MULTIPLE-QUERY-REPLACE-FROM-BUFFER
	 KIND-OF-QUERY-REPLACE-DOCUMENTATION
	 DOCUMENTATION-FUNCTION) 
(DEFPROP COM-TAGS-MULTIPLE-QUERY-REPLACE-FROM-BUFFER T CONTROL-PERIOD) 

(DEFCOM COM-TAGS-MULTIPLE-QUERY-REPLACE-FROM-BUFFER
	"Perform a multiple query replace from the contents of the specified buffer" ()
  (MULTIPLE-VALUE-BIND (FROM-LIST TO-LIST)
      (PARSE-BUFFER-REPLACE-PAIRS *INTERVAL*)
    (TAGS-MULTIPLE-QUERY-REPLACE FROM-LIST TO-LIST (AND *NUMERIC-ARG-P* *NUMERIC-ARG*)))) 

(DEFPROP COM-MULTIPLE-QUERY-REPLACE-FROM-BUFFER
	 KIND-OF-QUERY-REPLACE-DOCUMENTATION
	 DOCUMENTATION-FUNCTION) 

(DEFCOM COM-MULTIPLE-QUERY-REPLACE-FROM-BUFFER
	"Perform a multiple query replace from the contents of the specified buffer" ()
  (WITH-REGION-OR-WHOLE-INTERVAL (REGION-P)
    (LET ((*MODE-WORD-SYNTAX-TABLE* (IF (AND *NUMERIC-ARG-P* (MINUSP *NUMERIC-ARG*))
					*ATOM-WORD-SYNTAX-TABLE*
					*MODE-WORD-SYNTAX-TABLE*))
	  FROM-LIST
	  TO-LIST)
      (MULTIPLE-VALUE-SETQ (FROM-LIST TO-LIST)
	(PARSE-BUFFER-REPLACE-PAIRS T))
      (QUERY-REPLACE-LIST (POINT) (INTERVAL-LAST-BP *INTERVAL*)
			  FROM-LIST TO-LIST
			  *NUMERIC-ARG-P*)))
  DIS-TEXT) 

(DEFUN TAGS-MULTIPLE-QUERY-REPLACE (FROM-LIST TO-LIST ARG)
  (SETQ *TAGS-QUERY-REPLACE-FROM* FROM-LIST
	*TAGS-QUERY-REPLACE-TO* TO-LIST
	*TAGS-QUERY-REPLACE-DELIMITED* ARG)
  (CONTINUE-TAGS-MULTIPLE-QUERY-REPLACE T)) 

(DEFCOM COM-CONTINUE-TAGS-MULTIPLE-QUERY-REPLACE "Continue the last tags query replace" ()
  (CONTINUE-TAGS-MULTIPLE-QUERY-REPLACE NIL)) 
;;;
;;;  Blurb.
;;; This one is the good one, but commented out because of compiler error.
;;; Second version works with the compiler error.
;;(DEFUN CONTINUE-TAGS-MULTIPLE-QUERY-REPLACE (RESTART)
;;  (DO ((BEGINNING RESTART)
;;       (*MODE-WORD-SYNTAX-TABLE* (IF (AND *TAGS-QUERY-REPLACE-DELIMITED*
;;					  (MINUSP *TAGS-QUERY-REPLACE-DELIMITED*))
;;				*ATOM-WORD-SYNTAX-TABLE*
;;                              *MODE-WORD-SYNTAX-TABLE*))
;;       (VAL))
;;      (NIL)
;;    ;; Find and select the next buffer that has an occurrence of any of the strings.
;;    (DO (BP) (NIL)
;;      (SETQ BP (NEXT-FILE-BP BEGINNING))
;;      (SETQ BEGINNING NIL)
;;      (WHEN (LET ((*INTERVAL* (BP-TOP-LEVEL-NODE BP)))
;;	  (DOLIST (STR *TAGS-QUERY-REPLACE-FROM*)
;;	    (IF (SEARCH BP STR)
;;              (RETURN T))))
;;	(POINT-PDL-PUSH (POINT) *WINDOW* NIL NIL)
;;	(MAKE-BUFFER-CURRENT (BP-TOP-LEVEL-NODE BP))
;;	(MOVE-BP (POINT) BP)
;;	(RETURN)))
;;    (MUST-REDISPLAY *WINDOW* DIS-TEXT)
;;    ;; Query replace thru that buffer.
;;    (SETQ VAL (QUERY-REPLACE-LIST (POINT) (INTERVAL-LAST-BP *INTERVAL*)
;;				  *TAGS-QUERY-REPLACE-FROM* *TAGS-QUERY-REPLACE-TO*
;;				  *TAGS-QUERY-REPLACE-DELIMITED*))
;;    (WHEN (EQ VAL 'ABORTED)
;;      (TYPEIN-LINE "~&Use ~A to continue the Tags Multiple Query Replace."
;;		     (IF RESTART
;;		         (KEY-FOR-COMMAND-SET-C-. 'COM-CONTINUE-TAGS-QUERY-REPLACE)
;;		         (KEY-FOR-COMMAND 'COM-CONTINUE-TAGS-QUERY-REPLACE)))
;;      (RETURN DIS-TEXT))))

(DEFUN CONTINUE-TAGS-MULTIPLE-QUERY-REPLACE (RESTART)
  (DO ((BEGINNING RESTART)
       (*MODE-WORD-SYNTAX-TABLE* (IF (AND *TAGS-QUERY-REPLACE-DELIMITED*
					  (MINUSP *TAGS-QUERY-REPLACE-DELIMITED*))
				     *ATOM-WORD-SYNTAX-TABLE*
				     *MODE-WORD-SYNTAX-TABLE*))
       (VAL))
      (NIL)
    ;; Find and select the next buffer that has an occurrence of any of the strings.
    (DO (BP) (NIL)
      (SETQ BP (NEXT-FILE-BP BEGINNING))
      (SETQ BEGINNING NIL)
      (WHEN (GARRS-IDEA (BP-TOP-LEVEL-NODE BP) BP) ;THIS IS THE ONLY CHANGE FROM ABOVE
	(POINT-PDL-PUSH (POINT) *WINDOW* NIL NIL)
	(MAKE-BUFFER-CURRENT (BP-TOP-LEVEL-NODE BP))
	(MOVE-BP (POINT) BP)
	(RETURN)))
    (MUST-REDISPLAY *WINDOW* DIS-TEXT)
    ;; Query replace thru that buffer.
    (SETQ VAL (QUERY-REPLACE-LIST (POINT) (INTERVAL-LAST-BP *INTERVAL*)
				  *TAGS-QUERY-REPLACE-FROM* *TAGS-QUERY-REPLACE-TO*
				  *TAGS-QUERY-REPLACE-DELIMITED*))
    (WHEN (EQ VAL 'ABORTED)
      (TYPEIN-LINE "~&Use ~A to continue the Tags Multiple Query Replace."
		   (IF RESTART
		       (KEY-FOR-COMMAND-SET-C-. 'COM-CONTINUE-TAGS-QUERY-REPLACE)
		       (KEY-FOR-COMMAND 'COM-CONTINUE-TAGS-QUERY-REPLACE  *COMTAB* NIL NIL #\c-.)))
      (RETURN DIS-TEXT)))) 

(DEFUN GARRS-IDEA (*INTERVAL* BP)
  ;;DELETE ME WHEN COMPILER FIXED.
  (DOLIST (STR *TAGS-QUERY-REPLACE-FROM*)
    (IF (SEARCH BP STR)
	(RETURN T)))) 

(DEFCOM COM-SELECT-TAG-TABLE "Make a tag table current for commands like tags search" ()
  (SETQ *ZMACS-CURRENT-TAG-TABLE* (SELECT-TAG-TABLE NIL))
  DIS-NONE) 

(DEFCOM COM-SELECT-SYSTEM-AS-TAG-TABLE "Make the files in a system behave like a tags file" ()
  (LET ((SYSTEM-NAME (READ-SYSTEM-NAME "System to select as tag table:")))
    (SELECT-FILE-LIST-AS-TAG-TABLE (SI:SYSTEM-SOURCE-FILES SYSTEM-NAME
							   SI:*SOURCE-FILE-TYPES*
							   NIL T)
				   SYSTEM-NAME))
  DIS-NONE) 

(DEFUN SYSTEM-OF-PATHNAME (PATHNAME &OPTIONAL ALREADY-GENERIC)
  "Return the si:system defstruct for the system PATHNAME is in.
If none can be determined, return the one for the SYSTEM system.
ALREADY-GENERIC non-NIL says assume PATHNAME is already a generic pathname."
  ;;if there wasn't a pathname, return the SYSTEM system
  (IF (NULL PATHNAME)
      (SI:FIND-SYSTEM-NAMED SI:*SYSTEM-NAME* T T)
      (LET* ((GENERIC-PATHNAME (IF ALREADY-GENERIC
				   PATHNAME
				   (SEND PATHNAME :GENERIC-PATHNAME)))
	     (SYSTEMS (SEND GENERIC-PATHNAME :GET :SYSTEMS)))
	;;if it wasn't defined as part of a system, use the SYSTEM system
	(IF (NULL SYSTEMS)
	    (SI:FIND-SYSTEM-NAMED SI:*SYSTEM-NAME* T T)      
	    (SI:FIND-SYSTEM-NAMED (CAR SYSTEMS)))))) 

(DEFUN READ-SYSTEM-NAME (PROMPT &OPTIONAL (DEFAULT (SYSTEM-OF-PATHNAME
						     (BUFFER-GENERIC-PATHNAME *INTERVAL*)
						     T)))
  "Read a system name in the mini buffer, defaulting to DEFAULT.
Prompts with PROMPT (which should end with a colon and not mention the default).
DEFAULT defaults to a guess based on the current buffer."
  (LET ((SYSTEM-NAME (COMPLETING-READ-FROM-MINI-BUFFER (IF DEFAULT
							   (FORMAT NIL "~A (Default ~A)"
								   PROMPT (SI:SYSTEM-NAME DEFAULT))
							   PROMPT)
						       (SI:ALL-SYSTEMS-NAME-ALIST))))
    (IF (CONSP SYSTEM-NAME)
	(SETQ SYSTEM-NAME (CAR SYSTEM-NAME))
	(IF DEFAULT
	    (SETQ SYSTEM-NAME (SI:SYSTEM-NAME DEFAULT))
	    (BARF)))
    SYSTEM-NAME)) 

(DEFCOM COM-SELECT-ALL-BUFFERS-AS-TAG-TABLE "Select all files currently read in as a tag table.
This causes commands such as Tags Search, Tags Query Replace, and
Tags Compile Changed Sections to look through all files now visited." ()
  (SELECT-FILE-LIST-AS-TAG-TABLE
    (LOOP FOR BUFFER IN *ZMACS-BUFFER-LIST*
	  AS FILE-ID = (BUFFER-FILE-ID BUFFER)
	  WHEN (OR (EQ FILE-ID T)
		   (AND FILE-ID
			(CONSP FILE-ID)
			(NOT (NODE-SPECIAL-TYPE BUFFER))))
	  COLLECT (BUFFER-PATHNAME BUFFER))
    "All buffers visiting files")
  DIS-NONE) 

(DEFFLAVOR TAG-TABLE-DUMMY-FILE (NAME) (SI:PROPERTY-LIST-MIXIN)
  :INITTABLE-INSTANCE-VARIABLES
  :GETTABLE-INSTANCE-VARIABLES) 

(DEFMETHOD (TAG-TABLE-DUMMY-FILE :PRINT-SELF) (STREAM IGNORE SLASHIFY-P)
  (IF SLASHIFY-P
      (SYSTEM:PRINTING-RANDOM-OBJECT (SELF STREAM) (PRINC NAME STREAM))
      (PRINC NAME STREAM))) 

(DEFUN SELECT-FILE-LIST-AS-TAG-TABLE (FILE-LIST NAME)
  "Select a phony tag table named NAME consisting of the files in FILE-LIST.
This can be used to control commands such as Tags Search."
  (SETQ *ZMACS-CURRENT-TAG-TABLE* (MAKE-INSTANCE 'TAG-TABLE-DUMMY-FILE :NAME NAME))
  (SEND *ZMACS-CURRENT-TAG-TABLE* :PUTPROP
	(MAPCAR #'(LAMBDA (X) (FS:MERGE-PATHNAME-DEFAULTS X *PATHNAME-DEFAULTS*))
		FILE-LIST)
	'ZMACS-TAG-TABLE-FILE-SYMBOLS)
  (PUSH (CONS NAME *ZMACS-CURRENT-TAG-TABLE*)
	*ZMACS-TAG-TABLE-ALIST*)) 

(DEFUN SELECT-TAG-TABLE (&OPTIONAL (DEFAULT-P T)) "Read a tag table name and return that tag table.
DEFAULT-P non-NIL (as it is if omitted) means if there is an
obvious default than just return it without asking the user at all."
  (COND ((NULL *ZMACS-TAG-TABLE-ALIST*)
	 (CASE (FQUERY '(:CHOICES (((:FILE "Tags File") #\F)
				   ((:ALL-BUFFERS "All Buffers") #\B)
				   ((:SYSTEM "System") #\S #\D)))
		       "Specify tag table how? ")
	   (:FILE (LET ((PATHNAME (READ-DEFAULTED-PATHNAME "Tag table:" (PATHNAME-DEFAULTS) "TAGS" :NEWEST)))
		    (READ-TAG-TABLE PATHNAME)
		    *ZMACS-CURRENT-TAG-TABLE*))
	   (:ALL-BUFFERS (COM-SELECT-ALL-BUFFERS-AS-TAG-TABLE)
			 *ZMACS-CURRENT-TAG-TABLE*)
	   (:SYSTEM (COM-SELECT-SYSTEM-AS-TAG-TABLE)
		    *ZMACS-CURRENT-TAG-TABLE*)))
	((AND DEFAULT-P *ZMACS-CURRENT-TAG-TABLE*)	;for internal commands who want the current one.
	 *ZMACS-CURRENT-TAG-TABLE*)
	((AND DEFAULT-P (NULL (CDR *ZMACS-TAG-TABLE-ALIST*)))
	 ;;next file in current table?
	 (CDAR *ZMACS-TAG-TABLE-ALIST*))
	(T
	 (CASE (FQUERY '(:CHOICES (((:FILE "New Tags File") #\F)
				   ((:ALL-BUFFERS "All Buffers") #\B)
				   ((:SYSTEM "System") #\S #\D)
				   ((:EXISTING "Existing Table") #\E)))
		       "Specify tag table how? ")
	   (:FILE (LET ((PATHNAME (READ-DEFAULTED-PATHNAME "Tag table:" (PATHNAME-DEFAULTS) "TAGS" :NEWEST)))
		    (READ-TAG-TABLE PATHNAME)
		    *ZMACS-CURRENT-TAG-TABLE*))
	   (:ALL-BUFFERS (COM-SELECT-ALL-BUFFERS-AS-TAG-TABLE)
			 *ZMACS-CURRENT-TAG-TABLE*)
	   (:SYSTEM (COM-SELECT-SYSTEM-AS-TAG-TABLE)
		    *ZMACS-CURRENT-TAG-TABLE*)
	   (:EXISTING (LET ((TABLE (COMPLETING-READ-FROM-MINI-BUFFER
				     "Tag table (Ctrl-? for list):"
				     *ZMACS-TAG-TABLE-ALIST* NIL NIL
				     "You are typing the name of an existing tags table.")))
			(COND ((EQUAL TABLE "")
			       (COND (*ZMACS-CURRENT-TAG-TABLE*)
				     (T (BARF))))
			      (T
			       (CDR TABLE))))))))) 

(DEFUN PARSE-BUFFER-REPLACE-PAIRS (DEFAULT &AUX *INTERVAL*)
  "Parse a buffer as a list of string replacements.
Reads a buffer name from the user and parses that buffer.
Returns two values, a list of strings to replace and a list of replacement strings."
  (SETQ *INTERVAL* (READ-BUFFER-NAME "Use replacements in buffer:" DEFAULT))
  (DO ((BP (INTERVAL-FIRST-BP *INTERVAL*))
       (END-BP (INTERVAL-LAST-BP *INTERVAL*))
       (FROM-LIST)
       (TO-LIST)
       (TEM))
      (NIL)
    (SETQ BP (FORWARD-OVER *WHITESPACE-CHARS* BP))
    (AND (BP-= BP END-BP)
	 (RETURN (VALUES (NREVERSE FROM-LIST)
			 (NREVERSE TO-LIST))))
    (IF (CHAR= (BP-CH-CHAR BP) #\;)
	(SETQ BP (BEG-LINE BP 1))
	(PROGN
	  (MULTIPLE-VALUE-SETQ (TEM BP)
	    (PARSE-BUFFER-REPLACE-PAIRS-1 BP))
	  (PUSH TEM FROM-LIST)
	  (SETQ BP (FORWARD-OVER *BLANKS* BP))
	  (AND (END-LINE-P BP)
	       (BARF "Only one item on line ~S" (BP-LINE BP)))
	  (MULTIPLE-VALUE-SETQ (TEM BP)
	    (PARSE-BUFFER-REPLACE-PAIRS-1 BP))
	  (PUSH TEM TO-LIST))))) 

(DEFUN PARSE-BUFFER-REPLACE-PAIRS-1 (BP &AUX BP1 STR)
  (OR (SETQ BP1 (FORWARD-SEXP BP))
      (BARF "Premature eof on line ~S" (BP-LINE BP)))
  (SETQ STR (STRING-INTERVAL BP BP1 T))
  (AND (CHAR= (AREF STR 0) #\")
       (SETQ STR (READ-FROM-STRING STR T)))
  (VALUES STR BP1)) 

(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST* FUNCTION-NAME "Arglist" TYPEOUT-MENU-ARGLIST
			  NIL "Print arglist for this function.") 

(DEFUN TYPEOUT-MENU-ARGLIST (FUNCTION)
  (FORMAT *QUERY-IO* "~&~S: ~A" FUNCTION (ARGLIST FUNCTION))
  T)							;Leave the typeout window there

(DEFUN TYPEOUT-YES-OR-NO-P (&REST FORMAT-ARGS)
  "Ask the user Yes or No, printing in the typeout window."
  (LET ((*QUERY-IO* *STANDARD-OUTPUT*))
    (APPLY #'FQUERY '#,`(:SELECT T
			 :TYPE :READLINE
			 :CHOICES ,FORMAT:YES-OR-NO-P-CHOICES)
	   FORMAT-ARGS))) 

(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST* BP "Move" MOVE-TO-BP
			  T "Move to this line.") 

(DEFUN MOVE-TO-BP (BP &AUX INTERVAL)
  "Move point to BP, selecting its buffer if necessary."

  (let ((move-to-bp-hook (get (send *interval* :major-mode) 'major-mode-move-bp-function)))
    (cond (move-to-bp-hook
	   ;; This is mainly to support mail, without making the editor depend on it. spr6561 wjb,nn,rpm
	   (funcall move-to-bp-hook (point) bp))
	  (:else
	   (AND (SETQ INTERVAL (BP-TOP-LEVEL-NODE BP))
		(NEQ INTERVAL *INTERVAL*)
		(PROGN
		  (TYPEOUT-ABORT-MINI-BUFFER)
		  (MAKE-BUFFER-CURRENT INTERVAL)))
	   (MOVE-BP (POINT) BP))))
  NIL)

(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST* FILE "Find" FIND-DEFAULTED-FILE
			  T "Find file this file.") 

(DEFUN FIND-DEFAULTED-FILE (STRING &AUX PATHNAME VERSION)
  (TYPEOUT-ABORT-MINI-BUFFER)
  (SETQ PATHNAME (MAKE-DEFAULTED-PATHNAME (STRING STRING) (PATHNAME-DEFAULTS)))
  ;;It we get a specific file, see if that was the newest and if so, use that instead
  (AND (NOT (MEMBER (SETQ VERSION (SEND PATHNAME :VERSION))
		    '(:NEWEST :UNSPECIFIC) :TEST #'EQ))
       (= VERSION (SEND (SEND (SEND PATHNAME :NEW-VERSION :NEWEST)
			      :TRUENAME)
			:VERSION))
       (SETQ PATHNAME (SEND PATHNAME :NEW-VERSION :NEWEST)))
  (FIND-FILE PATHNAME)) 

(W:ADD-TYPEOUT-ITEM-TYPE *TYPEOUT-COMMAND-ALIST* FILE "Load" LOAD-DEFAULTED-FILE
			  NIL "LOAD this file.") 

(DEFUN LOAD-DEFAULTED-FILE (STRING)
  (LOAD (MAKE-DEFAULTED-PATHNAME (STRING STRING) (PATHNAME-DEFAULTS)))) 

(DEFCOM COM-EDIT-METHODS "Edit all methods for specified message." ()
  (MULTIPLE-VALUE-BIND (CLASSES-AND-FUNCTION-SYMBOLS MESSAGE)
      (LIST-METHODS-INTERNAL (READ-OPERATION-NAME "Edit flavors with methods for"))
    (EDIT-FUNCTIONS-NO-DISPLAY
      ;; The printed-representation of each item is the flavor name,
      ;; including package prefix if necessary, and the method type
      ;; in parentheses.  The cdr (function-name)
      ;; of each item is the mumble-method symbol.
      (MAPCAR #'(LAMBDA (X) (CONS (FORMAT NIL "~S~@[ (:~A)~]" (CAR X) (CADR X))
				  (CADDR X)))
	      CLASSES-AND-FUNCTION-SYMBOLS)
      'COM-GO-TO-NEXT-POSSIBILITY
      #\c-\p
      "Flavors with ~S methods:"
      "No methods for ~S found."
      MESSAGE))) 

(DEFCOM COM-LIST-METHODS "List all flavors with methods for specified message." ()
  (MULTIPLE-VALUE-BIND (CLASSES-AND-FUNCTION-SYMBOLS MESSAGE)
      (LIST-METHODS-INTERNAL (READ-OPERATION-NAME "List flavors with methods for"))
    (EDIT-FUNCTIONS-DISPLAY
      ;; The printed-representation of each item is the flavor name,
      ;; including package prefix if necessary, and the method type
      ;; in parentheses.  The cdr (function-name)
      ;; of each item is the mumble-method symbol.
      (MAPCAR #'(LAMBDA (X) (CONS (FORMAT NIL "~S~@[ (:~A)~]" (CAR X) (CADR X))
				  (CADDR X)))
	      CLASSES-AND-FUNCTION-SYMBOLS)
      'COM-GO-TO-NEXT-POSSIBILITY
      #\c-\p
      "Flavors with ~S methods:"
      "No methods for ~S found."
      MESSAGE))
  DIS-NONE) 

(DEFUN LIST-METHODS-INTERNAL (OPERATION &AUX FL TEM)
  (LET ((FUNCTION-SYMBOLS NIL))
    (DOLIST (FLAVOR *ALL-FLAVOR-NAMES*)
      (AND (SETQ FL (GET FLAVOR 'SI:FLAVOR))
	   (SETQ TEM (ASSOC OPERATION (SI:FLAVOR-METHOD-TABLE FL) :TEST #'EQ))
	   (DOLIST (METH (CDDDR TEM))
	     (OR (EQ (SI:METH-METHOD-TYPE METH) :COMBINED)
		 (AND (SI:METH-DEFINEDP METH)
		      (PUSH (LIST FLAVOR
				  (SI:METH-METHOD-TYPE METH)
				  (SI:METH-FUNCTION-SPEC METH))
			    FUNCTION-SYMBOLS))))))
    (BLOCK NIL
      (RETURN (VALUES FUNCTION-SYMBOLS OPERATION))
      NIL))) 

;; may 05/08/89 NEW to support meta-. on keywords for com-edit-defintion
(DEFUN list-methods-thru-mouse (spec)
  (declare (values spec-list string))
  (LET ((choice-methods (sort (MAPCAR #'THIRD 
                                (list-methods-internal spec))
			      #'(lambda (x y) (and (eq (symbol-package (second x)) *package*)
						   (not (eq (symbol-package (second y)) *package*)))))))
    (COND ((NULL choice-methods)
           (barf "~s is not a method keyword for any flavor." spec))
          ((= (LENGTH choice-methods) 1)
           (values choice-methods (FORMAT nil "~a" (CAR choice-methods))))
          (t (LET ((meth (w:menu-choose (MAPCAR #'(lambda (x) (LIST (prin1-to-string x) :value x))
						choice-methods)
                                        :label "Select a Method to Edit:")))
               
               (IF meth
                   (VALUES (LIST meth) (prin1-to-string meth))
                   (barf "No method chosen. Aborting.")))))))

(DEFF GET-MESSAGE-NAME 'READ-OPERATION-NAME) 

(DEFUN READ-OPERATION-NAME (PROMPT)
  "Read an operation name in the minibuffer, prompting with PROMPT.
PROMPT should be a string ending in a colon, if you want a colon.
If point is inside a FUNCALL or similar function, a default
may be derived from it."
  (PKG-BIND PKG-KEYWORD-PACKAGE				;So the colon can be omitted
    (MULTIPLE-VALUE-BIND (SYM STR)
	(READ-FUNCTION-NAME
	  PROMPT
	  (LET ((FUN (RELEVANT-FUNCTION-NAME (POINT))))
	    (AND (MEMBER FUN '(FUNCALL LEXPR-FUNCALL SEND LEXPR-SEND <- FUNCALL-SELF LEXPR-FUNCALL-SELF)
			 :TEST #'EQ)
		 (RELEVANT-METHOD-NAME (POINT)
				       (IF (MEMBER FUN '(FUNCALL-SELF LEXPR-FUNCALL-SELF) :TEST #'EQ)
					   1
					   2))))
	  NIL T)
      ;; Kludge around to not get screwed by completions to funny symbols
      ;; while still working if user points with the mouse
      (IF STR
	  (READ-FROM-STRING STR T)
	  SYM)))) 

(DEFCOM COM-EDIT-COMBINED-METHODS
	"Edit all methods used for specified operation on specified flavor." ()
  (LIST-COMBINED-METHODS-INTERNAL "Edit" T)
  DIS-NONE) 


(DEFCOM COM-LIST-COMBINED-METHODS
	"List all methods used for specified operation on specified flavor." ()
  (LIST-COMBINED-METHODS-INTERNAL "List" NIL)
  DIS-NONE) 


(DEFUN LIST-COMBINED-METHODS-INTERNAL (OP JUST-EDIT &AUX MESSAGE FLAVOR METHODS)
  (SETQ MESSAGE (READ-OPERATION-NAME
		  (FORMAT NIL "~A combined methods for operation" OP)))
  (SETQ FLAVOR (READ-FLAVOR-NAME
		 (FORMAT NIL "~A combined methods for operation ~S on flavor" OP MESSAGE)
		 "You are typing the name of a flavor, to see its combined methods"))
  ;;Duplicates code from SETUP-ZMACS-CALLERS-TO-BE-EDITED in order to
  ;;put the methods in execution order rather than alphabetical order
  (SETQ METHODS (FIND-COMBINED-METHODS FLAVOR MESSAGE))
  (FUNCALL (IF JUST-EDIT
	       'EDIT-FUNCTIONS-NO-DISPLAY
	       'EDIT-FUNCTIONS-DISPLAY)
	   (MAPCAR #'(LAMBDA (X) (CONS (FORMAT NIL "~S" X) X))
		   METHODS)
	   'COM-GO-TO-NEXT-TOP-LEVEL-POSSIBILITY
	   #\c-\p
	   "Methods combined for operation ~S on flavor ~S:"
	   "No methods for operation ~S on flavor ~S."
	   MESSAGE FLAVOR)) 

(DEFUN FIND-COMBINED-METHODS (FLAVOR MESSAGE &OPTIONAL (ERROR T) &AUX FL SM METHOD)
  "Return a list of the non-combined methods involved in handling MESSAGE to FLAVOR"
  (BLOCK NIL
    (OR (SETQ FL (GET FLAVOR 'SI:FLAVOR))
	(IF ERROR
	    (BARF "~S not DEFFLAVOR'ed" FLAVOR)
	    (RETURN)))
    (OR (SI:FLAVOR-METHOD-HASH-TABLE FL)
	(IF (FQUERY NIL "~S's combined methods are not composed.  Compose them? " FLAVOR)
	    (OR (ERRSET (SI::ASSURE-FLAVOR-COMPOSED FLAVOR) NIL)
		(BARF "~S's methods cannot be composed." FLAVOR))
	    (IF ERROR
		(BARF "~S's methods are not composed" FLAVOR)
		(RETURN))))
    (SETQ METHOD (IF (SI:FLAVOR-GET FL :ABSTRACT-FLAVOR)
		     (AND (FDEFINEDP `(:METHOD ,FLAVOR :COMBINED ,MESSAGE))
			  (FDEFINITION `(:METHOD ,FLAVOR :COMBINED ,MESSAGE)))
		     (IGNORE-ERRORS (SI:GET-FLAVOR-HANDLER-FOR FLAVOR MESSAGE))))
    (OR METHOD (IF ERROR
		   (BARF "Flavor ~S does not handle message ~S" FLAVOR MESSAGE)
		   (RETURN)))
    (SETQ METHOD (FUNCTION-NAME METHOD))
    (RETURN (COND ((SETQ SM (CDDDR (OR (SI:GET-DEBUG-INFO-FIELD 
					 (SI::GET-DEBUG-INFO-STRUCT (SI::UNENCAPSULATE-FUNCTION-SPEC METHOD))
					 'SI::COMBINED-METHOD-DERIVATION)
				       (SI:FUNCTION-SPEC-GET METHOD 'SI::COMBINED-METHOD-DERIVATION))))
		   (NCONC (REVERSE (CDR (ASSOC :WRAPPER SM :TEST #'EQ)))	;Try to approximate the order
			  (REVERSE (CDR (ASSOC :BEFORE SM :TEST #'EQ)))	;in which they're called
			  (REVERSE (CDR (ASSOC NIL SM :TEST #'EQ)))
			  (COPY-LIST (CDR (ASSOC :AFTER SM :TEST #'EQ)))
			  (MAPCAN #'(LAMBDA (X) (AND (NOT (MEMBER (CAR X) '(:WRAPPER :BEFORE NIL :AFTER)
								  :TEST #'EQ))
						     (REVERSE (CDR X))))
				  SM)))
		  (T
		   (LIST METHOD))))
    NIL)) 
