;;; -*- Mode:Common-Lisp; Package:SRCCOM; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (c) 1980 Massachusetts Institute of Technology 

;; LISP Machine Source Compare

;; 02/23/89 	may	Changed COMPARE-LINES to do what it says. SPR 9312
;; 8-12-88 	may	Fixed minor bug with SRCCOM: DESCRIBE-SRCCOM-SOURCES
;; Added option to SOURCE-COMPARE to allow files to be compared by their
;; s-expressions as well as by their text.  Implemented by adding an optional
;; argument to the SOurce Compare functions which specifies the type of compare
;; either :TEXT or :FORM.  The ZMACS interface queries the user for the type
;; of compare.

(DEFVAR *OUTPUT-STREAM*)

(DEFVAR *PRINT-LABELS* T
  "T means print the \"label\" preceding each run of differences.
In Lisp code, this is the function from the most recent DEFUN line.")

(DEFVAR *PATHNAME-DEFAULTS* (FS:MAKE-PATHNAME-DEFAULTS))

(DEFVAR *DIFFERENCE-PRINTER* 'PRINT-DIFFERENCES
  "Function to print out information about one run of differences.")

(DEFVAR *LINES-NEEDED-TO-MATCH* 3
  "Number of lines in a row that must match to finish a run of differences.")

(DEFVAR *LINES-TO-PRINT-BEFORE* 0
  "Number of matching lines preceding a run of differences to print with the differences.") 

(DEFVAR *LINES-TO-PRINT-AFTER* 1
  "Number of matching lines following a run of differences to print with the differences.") 

(DEFVAR *MERGE-LINE-NO* () "Line number in FILE-1 of first line after last run of differences.")

(DEFVAR *RECORD-MERGE-BOUNDS-P* ()
  "T means record buffer pointers in *MERGE-RECORD*, describing differences.") 

(DEFVAR *MERGE-RECORD*) 

(DEFVAR *MERGE-THIS-RECORD*) 

(DEFSTRUCT (FILE (:MAKE-ARRAY (:LENGTH 144))
		 (:CONC-NAME NIL)
		 (:CALLABLE-CONSTRUCTORS NIL)
		 (:ALTERANT ALTER-FILE)
		 (:PREDICATE NIL)
		 (:COPIER NIL)
		 (:TYPE :ARRAY-LEADER))
  (FILE-LENGTH 0)				;Number of lines
  FILE-NAME
  (FILE-TYPE "File")				;What kind of source it has
  FILE-STREAM					;Input stream
  FILE-MAJOR-MODE				;Lisp Mode.
  (FILE-COMPARE :TEXT)				;Kind of comparison to use
  (FILE-VARS NIL)				;Vars set from file attributes
  (FILE-VALUES NIL)				;Values of vars from file attributes
  ) 

(DEFUN FILE-IDENTIFIER (FILE-OBJECT)
  "Return a string identifying FILE-OBJECT for the user.
FILE-OBJECT is a SRCCOM:FILE structure."
  (STRING-APPEND (FILE-TYPE FILE-OBJECT) " " (STRING (FILE-NAME FILE-OBJECT)))) 

(DEFUN CREATE-FILE (FILENAME)
  "Make a SRCCOM:FILE object for a file to be source compared.
FILENAME is opened and the SRCCOM:FILE object contains a stream for it."
  (LET ((STREAM (OPEN FILENAME :CHARACTERS T :DIRECTION :INPUT))
	(GENERIC-PATHNAME (SEND FILENAME :GENERIC-PATHNAME)))
    (FS:READ-ATTRIBUTE-LIST GENERIC-PATHNAME STREAM)
    (MULTIPLE-VALUE-BIND (VARS VALS)
	(FS:FILE-ATTRIBUTE-BINDINGS GENERIC-PATHNAME)
      (MAKE-FILE :FILE-STREAM STREAM
		 :FILE-NAME (SEND STREAM :TRUENAME)
		 :FILE-VARS VARS
		 :FILE-VALUES VALS
		 :FILE-MAJOR-MODE (OR (SEND GENERIC-PATHNAME :GET :MODE) :LISP))))) 

(DEFUN GET-FILE-LINE (FILE LINE-NO)
  "Get the line recorded in FILE, a SRCCOM:FILE structure, for line number LINE-NO.
This will cause more data to be read from the stream if necessary.
The line is simply a string containing the data on the line."
  (IF (< LINE-NO (FILE-LENGTH FILE))
    (AREF FILE LINE-NO)
    (MULTIPLE-VALUE-BIND (LINE EOF)
	(CASE (FILE-COMPARE FILE)
	  (:TEXT (SEND (FILE-STREAM FILE) :LINE-IN T))
	  (:FORM (LET ((X (PROGV (FILE-VARS FILE)
				 (FILE-VALUES FILE)
			    (READ (FILE-STREAM FILE) nil '**EOF**))))
		   (IF (EQ X '**EOF**)
		       (VALUES nil T)		;No form, end of file
		       (VALUES X nil)))))
      (COND
	((NOT (AND EOF (OR (NULL LINE) (EQUAL LINE ""))))
	 (VECTOR-PUSH-EXTEND LINE FILE)
	 LINE))))) 

(DEFUN LINE-LAST-LABEL (FILE LINE-NO)
  "Return the last \"interesting\" line preceding line number LINE-NO, in FILE."
  (DO ((I (1- LINE-NO) (1- I))
       (MODE (FILE-MAJOR-MODE FILE))
       (LINE))
      ((< I 0))
    (AND (LINE-INTERESTING-P (SETQ LINE (AREF FILE I)) MODE) (RETURN LINE)))) 

(DEFUN LINE-INTERESTING-P (LINE MODE &AUX LEN)
  (AND (PLUSP (SETQ LEN (ARRAY-ACTIVE-LENGTH LINE)))
     (CASE MODE
       ((:LISP :ZTOP :ZETALISP :COMMON-LISP)
	(CHAR= (AREF LINE 0) #\( ))
       ((:TEXT :BOLIO)
	(CHAR= (AREF LINE 0) #\.))
       (OTHERWISE (NOT (MEMBER (AREF LINE 0)
			       '(#\SPACE #\TAB)
			       :TEST #'EQ)))))) 

;;; Compare two lines
;;; *** FOR NOW NO HAIR FOR COMMENTS, WHITESPACE, ETC. ***
(DEFUN COMPARE-LINES (LINE-1 LINE-2)
;;; Strip fonts if one uses s and the other fat-string characters, for compare only.
;;; Changes from patch 94.169.   ddd/gsl, 3/7/84.
;;; removed : from :art-... gsl. 3-11-85
  (IF (AND (TYPEP LINE-1 'STRING) (TYPEP LINE-2 'STRING)
	   (NEQ (ARRAY-TYPE LINE-1) (ARRAY-TYPE line-2))) ;; may 02/23/89 added
      (LET ((COMP-LINE-1 (CASE (ARRAY-TYPE LINE-1)
			   (ART-STRING (STRING-STRIP-EPS LINE-1))
			   (ART-FAT-STRING (STRING-REMOVE-FONTS LINE-1))))
	    (COMP-LINE-2 (CASE (ARRAY-TYPE LINE-2)
			   (ART-STRING (STRING-STRIP-EPS LINE-2))
			   (ART-FAT-STRING (STRING-REMOVE-FONTS LINE-2)))))
	(STRING-EQUAL COMP-LINE-1 COMP-LINE-2))
      ;;else we don't have two strings.  Just see if they are equal.
      (EQUAL LINE-1 LINE-2))) 
 
;;; Useful interface for automatic comparison
(DEFUN SOURCE-COMPARE (FILENAME-1 FILENAME-2
		       &OPTIONAL (OUTPUT-STREAM *STANDARD-OUTPUT*) (TYPE :TEXT)
		       &AUX FILE-1 FILE-2)
  "Source compare files FILENAME-1 and FILENAME-2, output to OUTPUT-STREAM.
See SOURCE-COMPARE-FILES for some additional information."
  (ASSERT (MEMBER TYPE '(:TEXT :FORM) :TEST #'EQ)
	  (TYPE)
	  "Only the values :TEXT or :FORM may be supplied for argument TYPE")
  (SETQ FILENAME-1 (FS:MERGE-AND-SET-PATHNAME-DEFAULTS FILENAME-1 *PATHNAME-DEFAULTS*)
	FILENAME-2 (FS:MERGE-PATHNAME-DEFAULTS FILENAME-2 FILENAME-1))
  (UNWIND-PROTECT
      (PROGN
	(SETQ FILE-1 (CREATE-FILE FILENAME-1)
	      FILE-2 (CREATE-FILE FILENAME-2))
	(DESCRIBE-SRCCOM-SOURCES FILE-1 FILE-2 OUTPUT-STREAM)
	(SOURCE-COMPARE-FILES FILE-1 FILE-2 OUTPUT-STREAM TYPE))
    (AND FILE-1 (SEND (FILE-STREAM FILE-1) :CLOSE))
    (AND FILE-2 (SEND (FILE-STREAM FILE-2) :CLOSE)))) 

(DEFUN PROMPTED-SOURCE-COMPARE (&optional FILE-1 FILE-2 (OUTPUT-STREAM *STANDARD-OUTPUT*) )
  (SETQ FILE-1 (FS:MERGE-PATHNAME-DEFAULTS FILE-1 *PATHNAME-DEFAULTS*)
	FILE-2 (FS:MERGE-PATHNAMES FILE-2 FILE-1))
  (MULTIPLE-VALUE-SETQ (FILE-1 FILE-2)
     (GET-SRCCOM-FILE-NAMES FILE-1 FILE-2))
  (CATCH-ERROR-RESTART ((SYSTEM:ABORT ERROR) "Terminate source comparison now.")
    (SOURCE-COMPARE FILE-1 FILE-2 OUTPUT-STREAM (QUERY-TYPE)))) 

(DEFUN QUERY-TYPE ()
  "Ask user for the type of compare to do.  Returns value suitable for SOURCE-COMPARE-FILES"
  (FQUERY '(:CHOICES (((:TEXT "Text") #\T) ((:FORM "Form") #\F))) "Compare by Text or Forms ")) 

(DEFUN GET-SRCCOM-FILE-NAMES (FILE-1 FILE-2)
  (DECLARE (VALUES FILE-1 FILE-2)
	   (SPECIAL FILE-1 FILE-2))
  (BLOCK TOP
    (DO ((STR)
	 (COMMA-POS))
	(NIL)
      (BLOCK nil
	(CATCH-ERROR-RESTART ((SYSTEM:ABORT ERROR) "Specify no file.")
	   (RETURN
	    (SETQ STR (SEND *QUERY-IO* :RUBOUT-HANDLER
			    '((:PROMPT GET-SRCCOM-FILE-NAMES-PROMPT))
			    #'READ-LINE *QUERY-IO*))))
	(RETURN-FROM TOP nil)
	nil)
      (COND
	((EQUAL STR "") (RETURN (VALUES FILE-1 FILE-2)))
	((SETQ COMMA-POS (POSITION #\, (THE STRING (STRING STR)) :TEST #'CHAR-EQUAL))
	 (SETQ FILE-1 (FS:MERGE-PATHNAMES (SUBSEQ STR 0 COMMA-POS) FILE-1)
	       FILE-2 (FS:MERGE-PATHNAMES (SUBSEQ STR (1+ COMMA-POS)) FILE-1)))
	(T (SETQ FILE-1 (FS:MERGE-PATHNAMES STR FILE-1)))))
    )) 

(DEFUN GET-SRCCOM-FILE-NAMES-PROMPT (STREAM IGNORE)
  (DECLARE (SPECIAL FILE-1 FILE-2))
  (FORMAT STREAM
	  "~%About to compare ~A with ~A~@
	   ~2TAre these correct? (Type Return, <file-1>, <file-1,file-2>, or Abort): "
	  FILE-1 FILE-2)) 

(DEFUN DESCRIBE-SRCCOM-SOURCES (FILE-1 FILE-2 STREAM)
  (FORMAT STREAM "~&;Source Compare of ~A and ~A on ~\\datime\\"
	  (FILE-IDENTIFIER FILE-1) (FILE-IDENTIFIER FILE-2))
  (UNLESS (TYPEP STREAM 'W:SHEET)
    (SEND STREAM :STRING-OUT " -*-Fundamental-*-"))
  (TERPRI STREAM)  ;; may 8-12-88 added "stream"
  (TERPRI STREAM)) ;; may 8-12-88 added "stream"

;;Helper function for doing text compares on files with epsilon fonts. This is a
;;destructive function in that it will modify the string to remove epsilons.

(DEFUN STRING-STRIP-EPS (LINE)
  (DECLARE (STRING LINE))
  (LOOP WITH END = (LENGTH LINE)
	AS HEAD FIRST 0 THEN FOUND-INDEX
	AS FOUND-INDEX = (POSITION #\ LINE :START HEAD :END END :TEST #'CHAR=)
	UNTIL (NULL FOUND-INDEX)
	FINALLY (ADJUST-ARRAY LINE END)
	(RETURN LINE)
	DOING (LET ((DIFFERENCE (MIN 2 (- END FOUND-INDEX))))
		(COPY-ARRAY-PORTION LINE (+ FOUND-INDEX DIFFERENCE) END
				    LINE FOUND-INDEX (- END DIFFERENCE))
		(DECF END DIFFERENCE)))) 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                              MAIN LOOP                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(DEFUN SOURCE-COMPARE-FILES (FILE-1 FILE-2 &OPTIONAL (*OUTPUT-STREAM* *STANDARD-OUTPUT*) (TYPE :TEXT))
  "Source compare data from two SRCCOM:FILE objects, with output to
*OUTPUT-STREAM*.  SRCCOM:FILE objects are made with SRCCOM:CREATE-FILE,
and contain input streams.  The (opt.) TYPE argument determines how the
files are compared, ':TEXT => line by line textual compare,
                    ':FORM => compare form by form.
When one source is buffer with fonts and other is file with fonts, font info is
removed when checking if lines of textual compare are the same.

Some special variables serve as parameters:
*PRINT-LABELS* - T means print the last function name, etc.,
 preceding each run of non-matching lines.
*LINES-NEEDED-TO-MATCH* - number of lines that must match
 in order for the two files to begin to agree again.
*LINES-TO-PRINT-BEFORE* - number of matching lines to print
 before each run of non-matching lines.
*LINES-TO-PRINT-AFTER* - number of matching lines to print
 after each run of non-matching lines.
*DIFFERENCE-PRINTER* - function to print out the differences.
 See SRCCOM:PRINT-DIFFERENCES for a sample."
  (ASSERT (MEMBER TYPE '(:TEXT :FORM) :TEST #'EQ) (TYPE)
	  "Only the values :TEXT or :FORM may be supplied for argument TYPE")
  (SETF (FILE-COMPARE FILE-1) TYPE)
  (SETF (FILE-COMPARE FILE-2) TYPE)
  (IF (EQ TYPE :FORM)
      (SET-FORM-VARIABLES 'START))
  (LET ((FILES-IDENTICAL T))
    (DO ((LINE-NO-1 0 (1+ LINE-NO-1))
	 (LINE-NO-2 0 (1+ LINE-NO-2))
	 (LINE-1)
	 (LINE-2))
	(NIL)
      ;; Files are current matched up, check the next two lines
      (SETQ LINE-1 (GET-FILE-LINE FILE-1 LINE-NO-1)
	    LINE-2 (GET-FILE-LINE FILE-2 LINE-NO-2))
      (UNLESS (COMPARE-LINES LINE-1 LINE-2)
	(SETQ FILES-IDENTICAL nil)
	(MULTIPLE-VALUE-SETQ (LINE-NO-1 LINE-NO-2 LINE-1)
	   (HANDLE-DIFFERENCE FILE-1 LINE-NO-1 FILE-2 LINE-NO-2)))
      (OR LINE-1 (RETURN nil)))			;When NULL lines match both files are at EOF
    (CLOSE (FILE-STREAM FILE-1))
    (CLOSE (FILE-STREAM FILE-2))
    (IF (EQ TYPE :FORM)
	(SET-FORM-VARIABLES 'END))
    FILES-IDENTICAL)) 

(DEFUN SET-FORM-VARIABLES (OPTION)
  "OPTION = 'START => initialize free variables for form compare.
        = 'END   => reset free variables."
  (CASE OPTION
    (START (SETQ *PRINT-LABELS* ()
		 *LINES-NEEDED-TO-MATCH* 1
		 *LINES-TO-PRINT-AFTER* 0))
    (END (SETQ *PRINT-LABELS* T
	       *LINES-NEEDED-TO-MATCH* 3
	       *LINES-TO-PRINT-AFTER* 1)))) 
 
;; First difference detected, look ahead for a match
(DEFUN HANDLE-DIFFERENCE (FILE-1 DIFF-LINE-NO-1 FILE-2 DIFF-LINE-NO-2
			  &AUX (NEW-LINE-NO-1 DIFF-LINE-NO-1)
			       (NEW-LINE-NO-2 DIFF-LINE-NO-2) LINE)
  (BLOCK TOP
    (DO ()
	(NIL)
      ;; Check next line from first file against lines in the second file
      (DO ((NEW-LINE-1 (GET-FILE-LINE FILE-1 (SETQ NEW-LINE-NO-1 (1+ NEW-LINE-NO-1))))
	   (LINE-NO-2 DIFF-LINE-NO-2 (1+ LINE-NO-2)))
	  (NIL)
	(SETQ LINE (GET-FILE-LINE FILE-2 LINE-NO-2))
	(COND ((AND (COMPARE-LINES NEW-LINE-1 LINE)
		    (CHECK-POTENTIAL-MATCH FILE-1 NEW-LINE-NO-1 FILE-2 LINE-NO-2))
	       (SETQ NEW-LINE-NO-2 LINE-NO-2)
	       (RETURN-FROM TOP)))      
	(AND (= LINE-NO-2 NEW-LINE-NO-2)
	     (RETURN)))
      ;; Check next line from second file against lines from the first file
      (DO ((NEW-LINE-2 (GET-FILE-LINE FILE-2 (SETQ NEW-LINE-NO-2 (1+ NEW-LINE-NO-2))))
	   (LINE-NO-1 DIFF-LINE-NO-1 (1+ LINE-NO-1)))
	  (NIL)
	(SETQ LINE (GET-FILE-LINE FILE-1 LINE-NO-1))
	(COND ((AND (COMPARE-LINES LINE NEW-LINE-2)
		    (CHECK-POTENTIAL-MATCH FILE-1 LINE-NO-1 FILE-2 NEW-LINE-NO-2))
	       (SETQ NEW-LINE-NO-1 LINE-NO-1)
	       (RETURN-FROM TOP)))
	(AND (= LINE-NO-1 NEW-LINE-NO-1)
	     (RETURN)))))
  (FUNCALL *DIFFERENCE-PRINTER* FILE-1 DIFF-LINE-NO-1 NEW-LINE-NO-1 FILE-2 DIFF-LINE-NO-2
	   NEW-LINE-NO-2)
  (VALUES NEW-LINE-NO-1 NEW-LINE-NO-2 LINE))

;; Found a potential match, check ahead to see if it is ok
(DEFUN CHECK-POTENTIAL-MATCH (FILE-1 LINE-NO-1 FILE-2 LINE-NO-2)
  (DO ((I *LINES-NEEDED-TO-MATCH* (1- I))
       (LINE-1)
       (LINE-2))
      ((<= I 1) T)
    (SETQ LINE-NO-1 (1+ LINE-NO-1)
	  LINE-NO-2 (1+ LINE-NO-2))
    (SETQ LINE-1 (GET-FILE-LINE FILE-1 LINE-NO-1)
	  LINE-2 (GET-FILE-LINE FILE-2 LINE-NO-2))
    (OR (COMPARE-LINES LINE-1 LINE-2)
	(RETURN nil)))) 

;; We are back in synch, print the differences
(DEFUN PRINT-DIFFERENCES (FILE-1 DIFF-LINE-NO-1 SAME-LINE-NO-1 FILE-2 DIFF-LINE-NO-2 SAME-LINE-NO-2)
  (PRINT-DIFFS-1 FILE-1 DIFF-LINE-NO-1 SAME-LINE-NO-1)
  (PRINT-DIFFS-1 FILE-2 DIFF-LINE-NO-2 SAME-LINE-NO-2)
  (FORMAT *OUTPUT-STREAM* "~&***************~2%")) 

(DEFUN PRINT-DIFFS-1 (FILE DIFF-LINE-NO SAME-LINE-NO &AUX LABEL)
  (SETQ DIFF-LINE-NO (MAX 0 (- DIFF-LINE-NO *LINES-TO-PRINT-BEFORE*))
	SAME-LINE-NO (+ SAME-LINE-NO *LINES-TO-PRINT-AFTER*))
  (FORMAT *OUTPUT-STREAM* "~&**** ~A ~A, Line #~D" (FILE-TYPE FILE) (FILE-NAME FILE)
	  DIFF-LINE-NO)
  (COND
    ((SETQ LABEL (AND *PRINT-LABELS* (LINE-LAST-LABEL FILE DIFF-LINE-NO)))
     (SEND *OUTPUT-STREAM* :STRING-OUT ", After \"")
     (SEND *OUTPUT-STREAM* :STRING-OUT (STRING-REMOVE-FONTS LABEL) 0
	      (MIN (LENGTH LABEL)
		   (IF (LET ((WHICH-OPERATIONS (SEND *OUTPUT-STREAM* :WHICH-OPERATIONS)))
		      (AND (MEMBER :READ-CURSORPOS WHICH-OPERATIONS :TEST #'EQ)
			 (MEMBER :SIZE-IN-CHARACTERS WHICH-OPERATIONS :TEST #'EQ)))
		     (- (SEND *OUTPUT-STREAM* :SIZE-IN-CHARACTERS)
			(SEND *OUTPUT-STREAM* :READ-CURSORPOS :CHARACTER) 1)
		     31)))
     (WRITE-CHAR #\" *OUTPUT-STREAM*)))
  (WRITE-CHAR #\NEWLINE *OUTPUT-STREAM*)
  (PRINT-FILE-SEGMENT FILE DIFF-LINE-NO SAME-LINE-NO)) 

(DEFUN PRINT-FILE-SEGMENT (FILE START-LINE-NO END-LINE-NO)
  (DO ((LINE-NO START-LINE-NO (1+ LINE-NO))
       (LINE))
      ((= LINE-NO END-LINE-NO))
    (OR (SETQ LINE (GET-FILE-LINE FILE LINE-NO)) (RETURN))
    (CASE (FILE-COMPARE FILE)
      (:TEXT
       (SEND *OUTPUT-STREAM* :LINE-OUT
		(IF (TYPEP *OUTPUT-STREAM* 'ZWEI:INTERVAL-STREAM-WITH-FONTS)
		    LINE
		    ;; Patch 94.110, ddd/gsl 2/26/84.
		    (STRING-REMOVE-FONTS LINE))))
      (:FORM (PPRINT LINE *OUTPUT-STREAM*)
	     (TERPRI *OUTPUT-STREAM*)
	     (TERPRI *OUTPUT-STREAM*))))) 

;; Merging.

(DEFUN SOURCE-COMPARE-AUTOMATIC-MERGE (FILE-1 FILE-2 OUTPUT-STREAM)
  "Source Compare Merge two FILE objects, output to file OUTPUT-FILENAME.
The data output to the OUTPUT-STREAM contains all the text of both files.
What the files agree on is simply output.  Where they disagree,
a run of differences is printed in the usual fashion.
See SOURCE-COMPARE-FILES for some additional information."
  (UNWIND-PROTECT
      (SOURCE-COMPARE-AUTOMATIC-MERGE-1 FILE-1 FILE-2 OUTPUT-STREAM)
    (AND FILE-1 (SEND (FILE-STREAM FILE-1) :CLOSE))
    (AND FILE-2 (SEND (FILE-STREAM FILE-2) :CLOSE)))) 

(DEFUN SOURCE-COMPARE-AUTOMATIC-MERGE-1 (FILE-1 FILE-2 *OUTPUT-STREAM*)
  "Do actual source compare & merge operations."
  (LET ((*DIFFERENCE-PRINTER* 'PRINT-AUTOMATIC-MERGE)
	(*MERGE-LINE-NO* 0))
    (SOURCE-COMPARE-FILES FILE-1 FILE-2 *OUTPUT-STREAM* (QUERY-TYPE))
    (PRINT-FILE-SEGMENT FILE-1 *MERGE-LINE-NO* (FILE-LENGTH FILE-1)))) 

(DEFUN SOURCE-COMPARE-AUTOMATIC-MERGE-RECORDING (FILE-1 FILE-2 OUTPUT-STREAM &AUX
						 (*RECORD-MERGE-BOUNDS-P* T)
						 (*MERGE-RECORD* NIL)
						 (*MERGE-THIS-RECORD* NIL))
  "Source Compare Merge two FILE objects, output to OUTPUT-STREAM, record differences.
The data output to the OUTPUT-STREAM contains all the text of both files.
What the files agree on is simply output.  Where they disagree,
a run of differences is printed in the usual fashion.
See SOURCE-COMPARE-FILES for some additional information.

Recording differences:
The value returned is a record of the differences.
It is a list with one element for each difference.
The element is a list of six buffer pointers, which point to
the six identified points in this skeleton of how one difference is printed.
<1>*** in file foo
<2>data from file foo (perhaps multiple lines)
<3>*** in file bar
<4>data from file bar (perhaps multiple lines)
<5>***
<6>

This works only if OUTPUT-STREAM supports the :READ-BP operation,
as streams that write into editor buffers do."
  
  (IF (NULL (GET-HANDLER-FOR output-stream :read-bp))
      (FORMAT output-stream "~%This function requires an editor buffer stream.  An equivalent function
that skips the creation of buffer pointers is (srccom:source-compare-automatic-merge ...)")
      ;;else
      (SOURCE-COMPARE-AUTOMATIC-MERGE FILE-1 FILE-2 OUTPUT-STREAM)
      ;; Do not CREATE these BPs with status :MOVES, or they will get
      ;; relocated by the insertion of the rest of the merged data!
      (DOLIST (RECORD (SETQ *MERGE-RECORD* (NREVERSE *MERGE-RECORD*)))
	(SETF (ZWEI:BP-STATUS (FIRST RECORD)) :MOVES)
	(SETF (ZWEI:BP-STATUS (THIRD RECORD)) :MOVES)
	(SETF (ZWEI:BP-STATUS (FIFTH RECORD)) :MOVES))
      *MERGE-RECORD*)) 

(DEFUN PRINT-AUTOMATIC-MERGE (FILE-1 DIFF-LINE-NO-1 SAME-LINE-NO-1 FILE-2 DIFF-LINE-NO-2 SAME-LINE-NO-2)
  (PRINT-FILE-SEGMENT FILE-1 *MERGE-LINE-NO* DIFF-LINE-NO-1)
  (IF (EQ (FILE-COMPARE FILE-1) :FORM)
    (FORMAT *OUTPUT-STREAM* "~%"))
  (WHEN *RECORD-MERGE-BOUNDS-P*
    (SETQ *MERGE-THIS-RECORD* nil)
    (RECORD-MERGE-BOUND))
  (SEND *OUTPUT-STREAM* :LINE-OUT "*** MERGE LOSSAGE ***")
  (PRINT-AUTOMATIC-MERGE-1 FILE-1 DIFF-LINE-NO-1 SAME-LINE-NO-1)
  (PRINT-AUTOMATIC-MERGE-1 FILE-2 DIFF-LINE-NO-2 SAME-LINE-NO-2)
  (SEND *OUTPUT-STREAM* :LINE-OUT "*** END OF MERGE LOSSAGE ***")
  (WHEN *RECORD-MERGE-BOUNDS-P*
    (RECORD-MERGE-BOUND)
    (PUSH (NREVERSE *MERGE-THIS-RECORD*) *MERGE-RECORD*))
  (SETQ *MERGE-LINE-NO* SAME-LINE-NO-1)) 

(DEFUN PRINT-AUTOMATIC-MERGE-1 (FILE DIFF-LINE-NO SAME-LINE-NO)
  (FORMAT *OUTPUT-STREAM* "~&*** ~A HAS:~%" (FILE-IDENTIFIER FILE))
  (AND *RECORD-MERGE-BOUNDS-P*
       (RECORD-MERGE-BOUND))
  (PRINT-FILE-SEGMENT FILE DIFF-LINE-NO SAME-LINE-NO)
  (AND *RECORD-MERGE-BOUNDS-P*
       (RECORD-MERGE-BOUND))) 

(DEFUN RECORD-MERGE-BOUND ()
  (PUSH (ZWEI:COPY-BP (SEND *OUTPUT-STREAM* :READ-BP) :NORMAL)
	*MERGE-THIS-RECORD*)) 
