;; -*- Mode:Common-Lisp; Base:10 -*-


;(setf (arglist LIST-ALL-PACKAGES) NIL)
(setf (documentation 'LIST-ALL-PACKAGES 'function) 
  "Returns a list containing all package objects in the system.")

(setf (documentation '*STANDARD-INPUT* 'variable) 
  "Global variable whose value is the default stream for input functions such as READ.")

;(setf (arglist BOUNDP) (SYMBOL))
(setf (documentation 'BOUNDP 'function) 
  "Returns true if the value cell of SYMBOL is not empty, otherwise NIL.")

;(setf (arglist PATHNAME-NAME) (P))
(setf (documentation 'PATHNAME-NAME 'function) 
  "Returns the name component of pathname P, NIL if a name component does not exist.")

(setf (documentation 'BOOLE-ORC2 'variable) 
  "Constant used as first argument to function BOOLE to take the bitwise-or of BOOLE's second
   argument with the complement of BOOLE's third argument.")

;(setf (arglist IGNORE) (&REST REST))
(setf (documentation 'IGNORE 'function) 
  "Discards any number of arguments then returns NIL.")

;(setf (arglist READ-CHAR) (&OPTIONAL STREAM (EOF-ERRORP T) EOF-VALUE RECURSIVE-P))
(setf (documentation 'READ-CHAR 'function)
  "Returns one character object after reading that character from STREAM.  End of file
   is an error if EOF-ERRORP is true; otherwise READ-CHAR returns EOF-VALUE when end
   of file is read.  The argument RECURSIVE-P is ignored.")

;(setf (arglist APPLYHOOK) (FUNCTION ARGS *EVALHOOK* *APPLYHOOK* &OPTIONAL ENVIRONMENT))
(setf (documentation 'APPLYHOOK 'function)
  "Applies FUNCTION to ARGS with ENVIRONMENT being the lexical environment for evaluation.
   If ENVIRONMENT is NIL, no lexical bindings exist.  The variables *EVALHOOK* and *APPLYHOOK*
   are used in all but the top level evaluations.")
 
(setf (documentation 'APPLYHOOK 'variable)
  "Variable APPLYHOOK, when non-NIL, has as its value a function to be called when a function
   is to be applied to a list of arguments.  This function takes three arguments: a function,
   a list of arguments, and an environment.")    

;(setf (arglist NCONC) (&REST LISTS))
(setf (documentation 'NCONC 'function) 
  "Returns a list containing a concatenation of all arguments in LISTS.  The arguments, which
   must be lists, are destructively modified by NCONC.")

;(setf (arglist HASH-TABLE-P) (S))
(setf (documentation 'HASH-TABLE-P 'function) 
  "Returns true is S is a hash table, otherwise NIL.")

;(setf (arglist MAKE-SYMBOL) (PNAME &OPTIONAL PERMANENT-P))
(setf (documentation 'MAKE-SYMBOL 'function) 
  "Creates an new uninterned symbol whose value and function bindings are unbound, property list 
   is empty, and print name is PNAME.  If PERMANENT-P is NIL the symbol is placed in the default
   area; otherwise, it is place in an area for permanent interned symbols.  MAKE-SYMBOL returns
   the symbol object.")

;(setf (arglist PARSE-BODY) (BODY ENVIRONMENT &OPTIONAL (DOC-STRING-ALLOWED T)))
(setf (documentation 'PARSE-BODY 'function) 
  "Parses the defun-like form BODY into three parts : declare forms, doc string, and rest of body. 
   BODY is the form to be parsed and consists of everything following the formal parameter list.
   ENVIRONMENT is the lexical environment for expansion of macros.  If DOC-STRING-ALLOWED is true,
   a doc string will be parsed out of the body and returned; otherwise, a string will terminate the
   search for declarations. 
   Returns three values :
     Value1 consists of everything which follows the declarations and doc string in BODY.
     Value2 is a list of declare forms.
     Value3 is the doc string if one exist, otherwise NIL.")

;(setf (arglist DISPLACED-ARRAY-P) (ARRAY))
(setf (documentation 'DISPLACED-ARRAY-P 'function)
  "Returns two values :
     Value1 is the array to which ARRAY is offset if ARRAY is an indirect array, otherwise NIL.
     Value2 is the index to which ARRAY is offset if ARRAY is an indirect array; otherwise, it is 0.")  

;(setf (arglist ROTATEF) (&REST PLACES))
(setf (documentation 'ROTATEF 'function) 
  "Rotates the place values left one position by performing a left circular shift one position
   on all the specified PLACES.  That is, place-value-1 = place-value-2, place-value-2 =
   place-value-3, ..., place-value-last = place-value-1.  ROTATEF returns NIL.")

;(setf (arglist >) (NUMBER &REST NUMBERS))
(setf (documentation '> 'function) 
  "Returns true if the arguments are in strictly decreasing numerical order, otherwise NIL.")

;(setf (arglist ATOM) (X))
(setf (documentation 'ATOM 'function) 
  "Returns true if X is not a cons, otherwise NIL.")

;(setf (arglist ACOS) (NUMBER))
(setf (documentation 'ACOS 'function)
  "Returns the angle in radians whose cosine is NUMBER.  The argument NUMBER may be complex.") 

;(setf (arglist ADJUSTABLE-ARRAY-P) (ARRAY))
(setf (documentation 'ADJUSTABLE-ARRAY-P 'function)
  "Returns true if ARRAY is an adjustable array, otherwise NIL.") 

;(setf (arglist Y-OR-N-P) (&OPTIONAL FORMAT-STRING &REST FORMAT-ARGS))
(setf (documentation 'Y-OR-N-P 'function) 
  "Asks the user a question which can be answered with Y(y) or N(n). If FORMAT-STRING is present
   and non-NIL, FORMAT-STRING and FORMAT-ARGS are printed as if they were given to function
   FORMAT.  Y-OR-N-P returns true if the reply is Y or y, NIL if N or n.")

;(setf (arglist SHADOWING-IMPORT) (SYMBOLS &OPTIONAL PKG))
(setf (documentation 'SHADOWING-IMPORT 'function) 
  "Makes SYMBOLS, which is either a single symbol or a list of symbols, internal to package PKG.
   SYMBOLS can therefore be referred to without the package prefix.  No error occurs if a symbol
   with the same name already exists in PKG.  SHADOWING-IMPORT returns T.")

;(setf (arglist 1-) (N))
(setf (documentation '1- 'function) 
  "Returns the value of N minus 1, where N can be any type of number.")

;(setf (arglist LAST) (LIST))
(setf (documentation 'LAST 'function) 
  "Returns the last cons of LIST, not the last element. If LIST is (), LAST returns ().")

;(setf (arglist ASSERT) (TEST-FORM &OPTIONAL PLACES (FORMAT-STRING Assertion failed.) &REST ARGS))
(setf (documentation 'ASSERT 'function) 
  "Signals an error if TEST-FORM evaluates to NIL.  PLACES, which are SETF'able references, should
   be both variables upon which TEST-FORM depends and changeable by the user when attempting to
   correct the error.  FORMAT-STRING and ARGS are passed to function FORMAT to output the error
   message.  ASSERT return NIL.")

;(setf (arglist LIST) (&REST ELEMENTS))
(setf (documentation 'LIST 'function)
  "Returns a list constructed by using ELEMENTS as the elements of the returned list.") 

(setf (documentation 'MOST-NEGATIVE-SINGLE-FLOAT 'variable)
  "Constant whose value is the greatest negative magnitude representable by the single float type.")

;(setf (arglist APROPOS) (STRING &KEY PACKAGE INHERITORS (INHERITED T) PREDICATE BOUNDP FBOUNDP DONT-PRINT))
(setf (documentation 'APROPOS 'function)
  "Prints the name of all symbols in package :PACKAGE whose print name contains STRING as a
   substring.  If :PACKAGE is NIL or not specified, all packages are searched.  The remaining
   keywords are implementation extensions.  They are defined as :
      :INHERITORS - If true, all packages which use PACKAGE are also searched for STRING.
      :INHERITED  - If NIL, the packages used by PACKAGE are not searched.
      :PREDICATE  - If specified, only symbols which return true when the form
                       (PREDICATE STRING) is evaluated are printed.
      :BOUNDP     - If true, only symbols with values are printed.
      :FBOUNDP    - If true, only symbols with function definitions are printed.
      :DONT-PRINT - If true, nothing is printed but rather the results are returned as a list.")   

;(setf (arglist CHAR>=) (CHAR &REST CHARS))
(setf (documentation 'CHAR>= 'function) 
  "Returns true if the arguments are nonincreasing, considering bits, fonts, and case; otherwise,
   NIL is returned.")

;(setf (arglist CHAR-GREATERP) (&REST CHARS))
(setf (documentation 'CHAR-GREATERP 'function) 
  "Returns true if the arguments are in strictly decreasing order, ignoring font and case;
   otherwise, NIL is returned.")

;(setf (arglist GETF) (PLACE INDICATOR &OPTIONAL DEFAULT))
(setf (documentation 'GETF 'function)
  "Returns the value from the property list stored in PLACE that corresponds to the indicator which
   is EQ to INDICATOR.  If INDICATOR is not found, DEFAULT is returned.  If DEFAULT is not specified,
   NIL is returned.") 

;(setf (arglist NSUBST-IF) (NEW PREDICATE TREE &KEY KEY))
(setf (documentation 'NSUBST-IF 'function)
  "Returns the tree structure created by destructivley replacing each atom or tree in TREE 
   that satisfies PREDICATE with NEW.  If non-NIL, :KEY is a function applied to each node
   of TREE to obtain the object to test against.") 

;(setf (arglist SXHASH) (X &OPTIONAL RANDOM-OBJECT-ACTION))
(setf (documentation 'SXHASH 'function)
  "Returns the hash code for X as a positive fixnum.") 

;(setf (arglist VECTOR-PUSH-EXTEND) (DATA VECTOR &OPTIONAL EXTENSION))
(setf (documentation 'VECTOR-PUSH-EXTEND 'function)
  "Stores the new element DATA at the end of VECTOR, extending VECTOR by EXTENSION number of cells
   if required.  VECTOR must have a fill pointer.  VECTOR-PUSH-EXTEND returns the index of DATA.")  

;(setf (arglist CAAAR) (X))
(setf (documentation 'CAAAR 'function)
  "Returns the result of (CAR (CAR (CAR X))).")

;(setf (arglist ACOSH) (NUMBER))
(setf (documentation 'ACOSH 'function) 
  "Returns the angle in radians as a floating-point or complex number whose hyperbolic cosine
   is NUMBER.")

;(setf (arglist BIT-IOR) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-IOR 'function) 
  "Returns the bit-array generated from a bitwise-or of BIT-ARRAY1 with BIT-ARRAY2.  If
   RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted, the result
   is returned as a new bit-array; if a bit-array, the result is destructively placed into it.")

;(setf (arglist SOME) (PREDICATE &REST SEQUENCES))
(setf (documentation 'SOME 'function)
  "Returns the first true value obtained by applying PREDICATE to successive elements
   of SEQUENCES.  PREDICATE operates on element 0, element 1, etc., of each sequence in SEQUENCE
   and must accept as many arguments as there are sequences provided.  If one sequence is exhausted
   before a true value is obtained, SOME returns NIL.")

;(setf (arglist NREVERSE) (SEQUENCE))
(setf (documentation 'NREVERSE 'function)
  "Returns the elements of SEQUENCE in reverse order.  NREVERSE destructively alters SEQUENCE.")

;(setf (arglist CODE-CHAR) (CODE &OPTIONAL (BITS 0) (FONT 0)))
(setf (documentation 'CODE-CHAR 'function)
  "Returns the character object whose code attribute comes from CODE, bit attribute from
   BITS, and font attribute from FONT.  All three arguments are non-negative integers.  If this
   character object cannot be constructed, NIL is returned.") 

;(setf (arglist CHAR-LESSP) (CHAR &REST CHARS))
(setf (documentation 'CHAR-LESSP 'function)
  "Returns true if all arguments are in strictly increasing order, ignoring font and case;
   otherwise, NIL is returned.") 

;(setf (arglist ARGLIST) (FUNCTION &OPTIONAL REAL-FLAG))
(setf (documentation 'ARGLIST 'function) 
  "Returns two values :
     Value1 is the argument list of FUNCTION.  If REAL-FLAG is true, the real argument list is 
            returned even if an more descriptive list is given thru a DECLARE form.  Otherwise,
            if a DECLARE is included, that argument list is returned.
     Value2 is a list of the return values of FUNCTION if they were included in the function
            definition thru a VALUES declaration.  If not specified, NIL is returned as Value2.")

;(setf (arglist CHAR-NAME) (CHAR))
(setf (documentation 'CHAR-NAME 'function)
  "Returns the standard name of CHAR as a string if CHAR is a character object; otherwise, NIL
   is returned.")  

;(setf (arglist DELETE-PACKAGE) (PKG))
(setf (documentation 'DELETE-PACKAGE 'function)
  "Uninterns all symbols present in PKG, invokes function UNUSE-PACKAGE on all other packages
   used by PKG, then deletes PKG and returns NIL.  If PKG is used by a different package, an error
   is signaled and PKG is not deleted.")   

;(setf (arglist READ-PRESERVING-WHITESPACE) (&OPTIONAL (STREAM *STANDARD-INPUT*) (EOF-ERRORP T)
;                                           (EOF-VALUE NIL) (RECURSIVEP NIL)))
(setf (documentation 'READ-PRESERVING-WHITESPACE 'function)
  "Returns one Lisp object after reading the printed representation of that object from STREAM.
   Any whitespace token used to terminate the printed representation is also returned.
   End of file is an error if EOF-ERRORP is true; otherwise, READ-PRESERVING-WHITESPACE returns
   EOF-VALUE when end of file is read.  The argument RECURSIVE-P, if specified and true, 
   signals that this is not a top level call to READ-PRESERVING-WHITESPACE but an imbedded call.")
 
;(setf (arglist BLOCK) (&QUOTE NAME &REST BODY))
(setf (documentation 'BLOCK 'function)
  "Creates a construct of name NAME where all forms in BODY are evaluated sequentially.  BLOCK 
   returns the value of the last form evaluated.  Symbol NAME is not evaluated.  BLOCK can be
   exited before all forms are evaluated by explicit use of a RETURN-FROM special form.  If the
   explicit exit is used, the return value is supplied by the RETURN-FROM form.")
 
;(setf (arglist REPLACE) (SINK-SEQ SOURCE-SEQ &KEY START1 END1 START2 END2))
(setf (documentation 'REPLACE 'function)
  "Returns a sequence created by replacing the elements in SINK-SEQ with the elements from
   SOURCE-SEQ.  SINK-SEQ is destructively modified by REPLACE.  Although the entire length of both
   SINK-SEQ and SOURCE-SEQ is the default, a subsequence of SINK-SEQ starting at :START1 and ending
   at :END1 and/or a subsequence of SOURCE-SEQ starting at :START2 and ending at :END2 may be
   specified.  If the number of elements to copy out of SOURCE-SEQ is less than the number of
   elements to be copied into SINK-SEQ, the extra elements of SINK-SEQ are unchanged.  If the number
   of elements to copy out of SOURCE-SEQ is greater than the number of elements to be copied into
   SINK-SEQ, the last extra elements from SOURCE-SEQ are ignored.")

;(setf (arglist >=) (NUMBER &REST NUMBERS))
(setf (documentation '>= 'function)
  "Returns true if the arguments are in nonincreasing numerical order, otherwise NIL.")

;(setf (arglist RASSOC) (ITEM ALIST &KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'RASSOC 'function)
  "Returns the first cons in association list ALIST whose CDR matches ITEM according to test
   predicate :TEST.  If no match is found, NIL is returned.  If TEST-NOT is specified, RASSOC
   returns the first cons whose CDR does not match ITEM according to test predicate :TEST-NOT.") 

;(setf (arglist THE) (&QUOTE &REST X))
(setf (documentation 'THE 'function)
  "Special form used to declare the type of a form's return value.  Syntax of (THE value-form form)
   is defined as : Evaluate <form>, using this value as THE's return value.  In addition, compare
   <form>'s type against the type specifier <value-form> and return an error if they do no match.
   THE may be used with the VALUES type specifier to indicate types of multiple values.")

;(setf (arglist STRING) (X))
(setf (documentation 'STRING 'function)
  "Returns the value of X coerced into a string.  If X is a string, a symbol, a string character,
   or an instance that supports :string-for-printing, it is coerced; otherwise, an error is
   signaled.")

(setf (documentation '*PRINT-PRETTY* 'variable)
  "Variable whose value when true signals that extra whitespace is to be inserted into
   expressions during their printing for readability.")

;(setf (arglist HOST-NAMESTRING) (OBJECT))
(setf (documentation 'HOST-NAMESTRING 'function)
  "Returns as a string the host name of pathname OBJECT.")

(setf (documentation '*PRINT-ARRAY* 'variable)
  "Variable whose value, when true, signals that the entire contents of a non-string array
   are to be printed in a form which can be read back in.  When NIL, only enough information to
   identify the array without listing the contents is printed.  The printing of strings is not
   affected by this variable.") 

(setf (documentation '*READ-SUPPRESS* 'variable)
  "Variable whose value, when true, signals that the Lisp reader will parse the input characters
   but interrupted very little.  This feature is in support of the real time condition constructs
   #+ and #-.  When *READ-SUPPRESS* is NIL, the Lisp reader operates normally.") 

;(setf (arglist TREE-EQUAL) (X Y &KEY TEST TEST-NOT))
(setf (documentation 'TREE-EQUAL 'function)
  "Returns true if the comparison of tree X to tree Y meets the following two conditions :  atoms
   match according to the test specified by :TEST and conses are TREE-EQUAL in both the CAR and
   the CDR.  TREE-EQAUL compares conses recursively.  If :TEST-NOT is specified instead of TEST,
   two atoms match if :TEST-NOT returns NIL.")  

;(setf (arglist BYTE-POSITION) (BYTE-SPECIFIER))
(setf (documentation 'BYTE-POSITION 'function) 
  "Returns as an integer the position of the byte specified by BYTE-SPECIFIER.")

;(setf (arglist COMPILED-FUNCTION-P) (OBJECT))
(setf (documentation 'COMPILED-FUNCTION-P 'function)
  "Returns true is OBJECT is a compiled code object, otherwise NIL.  A compiled code object can
   be a function or special form.") 

;(setf (arglist ARRAYP) (X))
(setf (documentation 'ARRAYP 'function)
  "Returns true is X is an array, otherwise NIL.") 

;(setf (arglist CDAAR) (X))
(setf (documentation 'CDAAR 'function)
  "Returns the result of (CDR (CAR (CAR X))).")

;(setf (arglist NSTRING-DOWNCASE) (STRING &KEY (START 0) END))
(setf (documentation 'NSTRING-DOWNCASE 'function)
  "Returns a string created by changing all uppercase characters in STRING to lowercase.  STRING is 
   destructively modified by NSTRING-DOWNCASE.  Although the entire length of STRING is the default,
   a substring of STRING starting at :START and ending at :END may be specified.  NSTRING-DOWNCASE
   returns the entire string, even if a substring is specified.")

;(setf (arglist OPEN) (FILENAME &KEY (DIRECTION INPUT) (ERROR T) (CHARACTERS T) BYTE-SIZE ELEMENT-TYPE
;                     IF-EXISTS IF-DOES-NOT-EXIST DELETED TEMPORARY ESTIMATED-LENGTH PRESERVE-DATES RAW
;                     SUPER-IMAGE SUBMIT FLAVOR LINK-TO &ALLOW-OTHER-KEYS))
(setf (documentation 'OPEN 'function)
  "Opens a file and returns a stream that is connected to file FILENAME.  FILENAME may be 
   either a pathname or a string.  The remaining arguments are either defined by Common Lisp
   or are implementation extensions.  The Common Lisp defined keywords are :
      :DIRECTION - Defines whether the stream handles input, output, or both.  Acceptable values
                   are :input, :output, :io, :probe, or the implementation defined values
                   of :probe-link or :probe-directory.
      :ELEMENT-TYPE - The type of transaction unit the stream supports.  A subtype of character
                      or integer is acceptable.
      :IF-EXISTS - Specifies the action to take if :DIRECTION is :output or :io and the file
                   already exists.  It is ignored in all other cases.  Acceptable values are
                   :error, :new-version, :rename, :rename-and-delete, :overwrite, :append, NIL,
                   :supersede, or the implementation defined value :truncate.
      :IF-DOES-NOT-EXIST - Specifies the action to take if the file does not exist.  Acceptable
                           values are :error, :create, or NIL.
   The implementation extension keywords are :
      :ERROR - If NIL, return and error object rather that signaling an error.
      :CHARACTERS - True means transfer character data.
      :BYTE-SIZE - Specify byte size; defaults according to CHARACTERS.
      :DELETED - If true, it is acceptable to open deleted but not expunged files.
      :TEMPORARY - If true, open the file and mark it as temporary.
      :ESTIMATED-LENGTH - Estimated length of file to be written in bytes.
      :PRESERVE-DATES - If true, do not alter the file's read or write dates.
      :RAW - If true, disable the server's character set translation.
      :SUPER-IMAGE - If true, disable special treatment of #\\rubout in pdp-10 servers.
      :SUBMIT - If true, submit this file as a batch job.
      :FLAVOR - Acceptable values are NIL, :directory, or :link.
      :LINK-TO - Specifies the link target when a file is created with flavor :link.") 

;(setf (arglist PACKAGE-USED-BY-LIST) (PKG))
(setf (documentation 'PACKAGE-USED-BY-LIST 'function) 
  "Returns a list of package objects that use package PKG.")

;(setf (arglist PATHNAME-VERSION) (P))
(setf (documentation 'PATHNAME-VERSION 'function)
  "Returns the version component of pathname P, NIL if a version component does not exist.")

(setf (documentation 'LONG-FLOAT-EPSILON 'variable) 
  "Constant whose value is the smallest possible positive number of type long float that, when
   added to some value, will change the magnitude of that value.")

;(setf (arglist INSPECT) (&OPTIONAL (OBJECT NIL OBJP)))
(setf (documentation 'INSPECT 'function)
  "Selects an Inspector window to view/modify object OBJECT after invoking the Inspector.  The
   Inspector runs in its own process; thus, special variable bindings are not visible.  To view
   special bindings, use INSPECT*.")

;(setf (arglist LDB) (PPSS WORD))
(setf (documentation 'LDB 'function)
  "Returns as a non negative integer the byte specified by PPSS from integer WORD.") 

;(setf (arglist *) (&REST NUMBERS))
(setf (documentation '* 'function) 
  "Returns the product of the arguments.")

(setf (documentation '* 'variable) 
  "Variable whose value is bound to the printed result of the last executed read-eval-print loop.")

;(setf (arglist POP) (PLACE &OPTIONAL INTO-PLACE))
(setf (documentation 'POP 'function) 
  "Returns the CAR of the contents of PLACE.  As a side effect, the CDR of the contents of PLACE
   is stored back into PLACE.  If INTO-PLACE is specified, the CAR value is stored there also.")

;(setf (arglist CADAR) (X))
(setf (documentation 'CADAR 'function)
  "Returns the result of (CAR (CDR (CAR X)))")

;(setf (arglist REVERSE) (SEQUENCE))
(setf (documentation 'REVERSE 'function) 
  "Returns a sequence of the same type and length of SEQUENCE but whose elements are in reverse
   order.")

;(setf (arglist VALUES) (&REST VALUES))
(setf (documentation 'VALUES 'function)
  "Returns a variable number of values determined by the number of arguments VALUES is passed.
   VALUES returns one value for each of its arguments, this value being the result of the
   evaluation of the argument.")

;(setf (arglist PROG1) (VALUES &REST IGNORED))
(setf (documentation 'PROG1 'function) 
  "Evaluates VALUES and saves this result.  All forms in IGNORED are then evaluated sequentially.
   Finally, PROG1 returns the saved result from the evaluation of VALUES.")

;(setf (arglist RATIONALIZE) (X))
(setf (documentation 'RATIONALIZE 'function)
  "Returns a rational approximation for the floating point number X, assuming floating point
   numbers are not completely accurate but are only accurate to their precision.") 

;(setf (arglist READ-BYTE) (&OPTIONAL STREAM (EOF-ERRORP T) EOF-VALUE))
(setf (documentation 'READ-BYTE 'function)
  "Returns one byte as an integer after reading that byte from STREAM.  End of file
   is an error if EOF-ERRORP is true; otherwise READ-BYTE returns EOF-VALUE when end
   of file is read.")

;(setf (arglist FFLOOR) (DIVIDEND &OPTIONAL (DIVISOR 1)))
(setf (documentation 'FFLOOR 'function)
  "Returns two values :
     Value1 is the largest integer expressed in float format that is either not larger than
            DIVIDEND if DIVISOR is not specified or not larger than the quotient obtained by 
            dividing DIVIDEND by DIVISOR if DIVISOR is specified.
     Value2 is the number to add to Value1 (or Value1 times DIVISOR if DIVISOR is specified) in
            order to equal DIVIDEND.")

;(setf (arglist ROOM) (&REST ARGS))
(setf (documentation 'ROOM 'function) 
  "Prints to *STANDARD-OUTPUT* information about the size and remaining free space of some areas.
   If ARGS is not specified, the names in variable ROOM are used.  ARGS may take on the value of
   T to print the maximum amount of information about the system, NIL to print the minimun amount,
   or area names to print information about those areas which are of interest.")

(setf (documentation 'ROOM 'variable)
  "Variable whose value is the list of areas to be used by function ROOM if that function is
   called with no arguments.") 

;(setf (arglist STRING-NOT-LESSP) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING-NOT-LESSP 'function)
  "Returns the index into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test greater than or equal to, ignoring case and fonts.  If the test is never
   satisfied, NIL is returned.  If the strings match, the length of the shortest string is
   returned.  Although the entire length of both STRING1 and STRING2 is the default, a substring
   of STRING1 starting at :START1 and ending at :END1 and/or a substring of STRING2 starting at
   :START2 and ending at :END2 may be specified.")

(setf (documentation 'MOST-POSITIVE-SINGLE-FLOAT 'variable)
  "Constant whose value is the greatest positive magnitude representable by the single float type.") 

;(setf (arglist FLET) (&QUOTE FUNCTION-LIST &REST BODY))
(setf (documentation 'FLET 'function)
  "Special form used to define locally named functions.  Syntax of 
   (FLET ((fct1 body1) ... (fctN bodyN)) {flet-form}*) is defined as :  <fct1> thru <fctN> are
   the names of locally defined functions whose bodies <body1> thru <bodyN>, respectively, are
   similar in format to a DEFUN form.  These functions can only be called within the lexical scope
   of <flet-form> and can neither call each other nor recurse.  The forms in <flet-form> are
   evaluated as if in a PROGN, with function names matching those defined by the FLET referring not 
   to the global but rather the local function.")

;(setf (arglist COPY-LIST) (LIST &OPTIONAL FORCE-DOTTED))
(setf (documentation 'COPY-LIST 'function)
  "Returns a list that is EQUAL to but not EQ to LIST.  Dotted pair termination of the list will
   also be copied.") 

;(setf (arglist DIGIT-CHAR) (WEIGHT &OPTIONAL (RADIX 10) (FONT 0)))
(setf (documentation 'DIGIT-CHAR 'function) 
  "Returns a character object whose font attribute is FONT and whose weight is WEIGHT when 
   considered as a digit of the radix RADIX.  If this is not possible, NIL is returned.")

;(setf (arglist MACROEXPAND-1) (FORM &OPTIONAL ENVIRONMENT))
(setf (documentation 'MACROEXPAND-1 'function)
  "Returns two values :
     Value1 is FORM expanded but once in an environment where the local macro definitions are
            defined by ENVIRONMENT.
     Value2 is T if FORM is a macro, NIL if it is not.") 

;(setf (arglist FBOUNDP) (SYMBOL))
(setf (documentation 'FBOUNDP 'function)
  "Returns true is SYMBOL has a global function definition, otherwise NIL.") 

;(setf (arglist REST) (LIST))
(setf (documentation 'REST 'function)
  "Returns the CDR of LIST.") 

;(setf (arglist ARRAY-DIMENSION) (ARRAY DIMENSION))
(setf (documentation 'ARRAY-DIMENSION 'function) 
  "Returns the length of dimension DIMENSION of ARRAY.  ARRAY-DIMENSION uses zero-origin.")

;(setf (arglist ROTL) (FLAVOR-INSTANCE))
;(setf (documentation 'ROTL 'function) 
;  "Run Object Top Level - Implicit messages will be sent to FLAVOR-INSTANCE.  If FLAVOR-INSTANCE
;is NIL, implicit message sending is deactivated.  ROTL is used when you are working
;with a given flavor instance.  Messages are sent to the instance by typing the message
;name followed by any arguments.  All typed forms are evaluated with the instance's
;instance variables bound. The following is an example of using ROTL, with typed
;expressions on the left and an indication of what is retuned on the right.   (ROTL
;*terminal-io*) => <uninteresting value>  tv:name            => <*TERMINAL-IO*'s
;name>  :CLEAR-SCREEN    => <*TERMINAL-IO* is cleared>  (SETQ tv:name 'foo) => <*TERMINAL-IO*'s
;name becomes FOO>  The current FLAVOR-INSTANCE is stored in USER:TOP-LEVEL-SELF.")

;(setf (arglist GET-PROPERTIES) (PLACE INDICATOR-LIST))
(setf (documentation 'GET-PROPERTIES 'function)
  "Returns three values :
     Value1 is the first element of INDICATOR-LIST which matches an indicator in the property
            list stored in PLACE.
     Value2 is the property value of the matched indicator.
     Value3 is the tail of the property list starting with the found property value pair.
   If no match is found, all three values return NIL.")   

;(setf (arglist BOOLE) (OP ARG1 &REST ARGS))
(setf (documentation 'BOOLE 'function)
  "Returns an integer produced by performing the bitwise logical operation specified by OP
   on the integers ARG1 and ARGS.  If ARGS is more than one argument, the first two arguments
   beginning at ARG1 are operated on and the results applied to the next argument.  This is
   repeated until the argument list is exhausted.  Implementation dependent constants are
   provided for OP.")   
 
;(setf (arglist KEYWORDP) (SYMBOL))
(setf (documentation 'KEYWORDP 'function)
  "Returns true if SYMBOL is a keyword, otherwise NIL.")

(setf (documentation '*EVALHOOK* 'variable) 
  "Variable whose value, when non-NIL, should be a function accepting the two arguments of
   form and environment that is invoked whenever an evaluation is done either implicitly
   or through an explicit call to function EVAL.  This is for use with the EVALHOOK and 
   APPLYHOOK functions.")

;(setf (arglist MACROLET) (&QUOTE MACRO-LIST &REST BODY))
(setf (documentation 'MACROLET 'function)
  "Special form used to define locally named macros.  Syntax of 
   (MACROLET ((mac1 body1) ... (macN bodyN)) {macrolet-form}*) is defined as :  <mac1> thru <macN>
   are the names of locally defined macros whose bodies <body1> thru <bodyN>, respectively, are
   similar in format to a DEFMACRO form.  These macros can only be called within the lexical scope
   of <macrolet-form> and cannot refer to each other.  The forms in <macrolet-form> are evaluated
   as if in a PROGN, with macro names matching those defined by the MACROLET referring not 
   to the global but rather the local macros.")

;(setf (arglist FILE-LENGTH) (FILE))
(setf (documentation 'FILE-LENGTH 'function)
  "Returns as a non-negative integer the length of file FILE if it can be determined, otherwise NIL.")

;(setf (arglist GET-DISPATCH-MACRO-CHARACTER) (DISP-CHAR SUB-CHAR &OPTIONAL (RT *READTABLE*)))
(setf (documentation 'GET-DISPATCH-MACRO-CHARACTER 'function)
  "Returns the function associated with SUB-CHAR following dispatch macro character DISP-CHAR
   in readtable RT.  If SUB-CHAR is not defined for DISP-CHAR, NIL is returned.")     

;(setf (arglist COSH) (NUMBER))
(setf (documentation 'COSH 'function) 
  "Returns the hyperbolic cosine of NUMBER as a floating-point or complex number.")

;(setf (arglist HASH-TABLE-REHASH-SIZE) (HASH-TABLE))
(setf (documentation 'HASH-TABLE-REHASH-SIZE 'function)
  "Returns the amount to increase HASH-TABLE when it becomes full, either an integer indicating the
   number of entries to add or a floating-point number indicating the ratio of new size to old size.") 

;(setf (arglist SYMBOL-PLIST) (SYMBOL))
(setf (documentation 'SYMBOL-PLIST 'function) 
  "Returns the contents of the property list of SYMBOL.  SYMBOL may be a symbol, an instance
   supporting the :PROPERTY-LIST operation, or a locative or cons cell whose CDR is the
   property list.")

;(setf (arglist KILL-PACKAGE) (PKG))
(setf (documentation 'KILL-PACKAGE 'function) 
  "Removes PKG from the list of package names that are searched when package names are looked
   up then returns NIL.")

;(setf (arglist ENCODE-UNIVERSAL-TIME) (SECONDS MINUTES HOURS DAY MONTH YEAR &OPTIONAL TIMEZONE))
(setf (documentation 'ENCODE-UNIVERSAL-TIME 'function)
  "Returns as a bignum encoded into Universal Time the supplied arguments.  Universal Time is the
   number of seconds since midnight, January 1, 1900 GMT.") 

;(setf (arglist GET) (SYMBOL-OR-PLIST PROPERTY &OPTIONAL DEFAULT))
(setf (documentation 'GET 'function)
  "Returns the value found in property list SYMBOL-OR-PLIST whose indicator is EQ to PROPERTY.  If
   no indicator is found DEFAULT is returned, if specified, or NIL if DEFAULT is unspecified.")

;(setf (arglist CDDAR) (X))
(setf (documentation 'CDDAR 'function)
  "Returns the result of (CDR (CDR (CAR X)))")

;(setf (arglist SET) (SYMBOL VALUE))
(setf (documentation 'SET 'function)
  "Assigns the results obtained by evaluating VALUE to the value cell of SYMBOL.  This result is
   also returned by SET.") 

;(setf (arglist SEVENTH) (LIST))
(setf (documentation 'SEVENTH 'function) 
  "Returns the seventh element of LIST.  One-origin numbering is used.")

(setf (documentation 'CHAR-META-BIT 'variable) 
  "Constant whose value gives a numerical meaning to the meta bit in the bit configuration for a
   character object.")

;(setf (arglist CLEAR-INPUT) (&OPTIONAL STREAM))
(setf (documentation 'CLEAR-INPUT 'function) 
  "Discards any buffered input on STREAM if it is an interactive stream then returns NIL.")

(setf (documentation '*PACKAGE* 'variable) 
  "Variable whose value is the current package.")

(setf (documentation 'ARRAY-RANK-LIMIT 'variable)
  "Constant whose positive integer value is the exclusive upper bound on the rank of an array.") 

;(setf (arglist MAP) (RESULT-TYPE FCN &REST SEQUENCES))
(setf (documentation 'MAP 'function)
  "Returns a sequence of type RESULT-TYPE created by applying FCN to each successive element
   of SEQUENCES until some input sequence is exhausted.  The resulting sequence is as long
   as the shortest input sequence.  Function FCN must take as many arguments as there are
   sequences provided.  If RESULT-TYPE is NIL, FCN is called for effect only and returns NIL
   instead of a sequence.")

;(setf (arglist DIRECTORY) (PATHNAME))
(setf (documentation 'DIRECTORY 'function)
  "Returns a list of truenames, one for each file in the file system that matches PATHNAME.
   If no files are found, NIL is returned.") 

;(setf (arglist BIT-ANDC2) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-ANDC2 'function)
  "Returns the bit-array generated from a bitwise-and of BIT-ARRAY1 with the complement of
   BIT-ARRAY2.  If RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or
   omitted, the result is returned as a new bit-array; if a bit-array, the result is destructively
   placed into it.") 

;(setf (arglist CAAR) (X))
(setf (documentation 'CAAR 'function)
  "Returns the result of (CAR (CAR X))")

;(setf (arglist COUNT) (ITEM SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END FROM-END))
(setf (documentation 'COUNT 'function) 
  "Returns as a non negative integer the number of elements in SEQUENCE which match ITEM according
   to test predicate :TEST.  Although the entire length of SEQUENCE is the default, a subsequence
   starting a :START and ending at :END may be specified.  If :TEST-NOT is specified, the number of
   elements which do not match ITEM according to test predicate :TEST-NOT is returned.  If :KEY is
   non-NIL, it is a function accepting one argument that is applied to each element of SEQUENCE
   before the test is performed.  :FROM-END has no effect on the results.")

;(setf (arglist DPB) (VALUE PPSS WORD))
(setf (documentation 'DPB 'function)
  "Returns the integer composed by substituting the bits of VALUE into WORD according to byte 
   specifier PPSS.  VALUE is taken to be right justified.")

;(setf (arglist TIME) (&OPTIONAL FORM))
(setf (documentation 'TIME 'function)
  "If FORM is specified, TIME returns FORM's return value while printing to *TRACE-OUTPUT*
   elapsed realtime, paging time, number of page faults, and number of conses by area.
   If FORM is not specified or NIL, TIME returns time in 60ths of a second.  Only differences
   between successive calls to TIME are significant.") 

;(setf (arglist FIND-IF-NOT) (PREDICATE SEQUENCE &KEY KEY START END FROM-END))
(setf (documentation 'FIND-IF-NOT 'function)
  "Returns the leftmost element of SEQUENCE that does not satisfy the test predicate PREDICATE.
   Although the entire length of SEQUENCE is the default, a subsequence starting at :START and ending
   at :END may be specified.  If :FROM-END is true, the rightmost element that does not satisfy
   the test is returned.  If specified, :KEY is a function taking one argument that is applied to
   each element of SEQUENCE before the test is performed.") 

;(setf (arglist DIRECTORY-NAMESTRING) (OBJECT))
(setf (documentation 'DIRECTORY-NAMESTRING 'function)
  "Returns a namestring of only the device and directory components of pathname object OBJECT.") 

(setf (documentation 'BOOLE-NOR 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-nor of BOOLE's second
   argument with BOOLE's third argument.")

(setf (documentation 'LEAST-POSITIVE-SHORT-FLOAT 'variable)
  "Constant whose value is the smallest positive magnitude representable by the short float type.") 

;(setf (arglist SYMBOL-NAME) (SYMBOL))
(setf (documentation 'SYMBOL-NAME 'function) 
  "Returns the print name of symbol SYMBOL.")

;(setf (arglist NSTRING-UPCASE) (STRING &KEY (START 0) END))
(setf (documentation 'NSTRING-UPCASE 'function)
  "Returns a string created by changing all lowercase characters in STRING to uppercase.  STRING is
   destructively modified by NSTRING-UPCASE.  Although the entire length of STRING is the default,
   a substring of STRING starting at :START and ending at :END may be specified.  NSTRING-UPCASE
   returns the entire string, even if a substring is specified.")

;(setf (arglist GET-DECODED-TIME) NIL)
(setf (documentation 'GET-DECODED-TIME 'function) 
  "Returns nine values representating the present time in the following order :
     second, minute, hour, date, month, year, day-of-week, daylight-savings-time-p, and time zone.
   If the time cannot be obtained, NIL is returned.")

(setf (documentation '*MODULES* 'variable)
  "Variable whose value is a list of names of the modules currently loaded into the Lisp
   environment.")

;(setf (arglist MAKE-ARRAY) (DIMENSIONS &KEY TYPE ELEMENT-TYPE INITIAL-ELEMENT INITIAL-VALUE INITIAL-CONTENTS
;                            ADJUSTABLE FILL-POINTER DISPLACED-TO DISPLACED-TO-PHYSICAL-ADDRESS
;                            DISPLACED-INDEX-OFFSET LEADER-LENGTH LEADER-LIST AREA NAMED-STRUCTURE-SYMBOL))
(setf (documentation 'MAKE-ARRAY 'function)
  "Creates an array of size DIMENSIONS and returns the array object.  The Common 
   Lisp defined keywords are :
     :ELEMENT-TYPE - Data type of elements allowed in the array.  A value of T allows any
                     Lisp object in the array.
     :INITIAL-ELEMENT - Value to be stored in each element of the new array.
     :INITIAL-CONTENTS - Value is a sequence of sequences used to initial the array.  Top level
                         of sequence correspondings to the most slowly varying subscript.
     :ADJUSTABLE - If true, the array is made adjustable.
     :FILL-POINTER - If non-NIL, this value defines the number of active elements in the array.
                     If T, the array length is used.
     :DISPLACED-TO - If non-NIL, this symbol is an array that this new array will overlap.
     :DISPLACED-INDEX-OFFSET - When :DISPLACED-TO is specified, this nonnegative value specifies 
                               the element that corresponds to element 0 of the new array.
   The implementation extension keywords are :
     :TYPE - This data type accept both Common Lisp data types and Explorer data types.
     :DISPLACED-TO-PHYSICAL-ADDRESS - Specifies in terms of addresses the displaced value.
     :LEADER-LENGTH - If non-NIL, this fixnum specifies the number of elements of the array
                      leader.
     :LEADER-LIST - If non-NIL, this list is used to initialize the elements of the leader.
     :AREA - If non-NIL, this integer specifies the area in which to create the array.
     :NAMED-STRUCTURE-SYMBOL - If non-NIL, this symbol is stored in the named structure cell
                               of the array, thus tagging the array as a named structure.") 
    
;(setf (documentation 'PACKAGE 'variable)
;  " ")

;(setf (arglist VECTOR-PUSH) (NEW-ELEMENT VECTOR))
(setf (documentation 'VECTOR-PUSH 'function)
  "Stores the new element NEW-ELEMENT in vector VECTOR at the array position specified by the
   fill pointer.  This position should be the end of the array.  The fill pointer is incremented
   by one and the former value of the fill pointer is returned by VECTOR-PUSH.  If the element
   cannot be added, VECTOR-PUSH returns NIL.")
 
;(setf (arglist DELETE-SETF-METHOD) (ACCESS-FN))
(setf (documentation 'DELETE-SETF-METHOD 'function) 
  "Deletes the setf method associated with the accessor function ACCESS-FN")

(setf (documentation '*APPLYHOOK* 'variable)
  "Variable whose value, when non-NIL, is a function to be applied to its evaluated arguments
   inside calls to functions EVALHOOK and APPLYHOOK.")

;(setf (arglist PACKAGE-SHADOWING-SYMBOLS) (PKG))
(setf (documentation 'PACKAGE-SHADOWING-SYMBOLS 'function)
  "Returns a list of all shadowing symbols in package PKG.")

;(setf (arglist READ-DELIMITED-LIST) (ENDCHAR &OPTIONAL (INPUT-STREAM *STANDARD-INPUT*) IGNORE))
(setf (documentation 'READ-DELIMITED-LIST 'function)
  "Reads objects from stream INPUT-STREAM until the next character found after an object's
   representation is character ENDCHAR, then returns the objects read as a list.  The argument
   IGNORE should be set to true if READ-DELIMITED-LIST is called from within a read function.
   End of file is always an error.")

;(setf (arglist ARRAY-RANK) (ARRAY))
(setf (documentation 'ARRAY-RANK 'function) 
  "Returns as a nonnegative integer the number of dimensions of ARRAY.")

;(setf (arglist CAAAAR) (X))
(setf (documentation 'CAAAAR 'function)
  "Returns the result of (CAR (CAR (CAR (CAR X))))")

;(setf (arglist INTERN) (STRING &OPTIONAL PKG))
(setf (documentation 'INTERN 'function)
  "Returns two values (Common Lisp version) :
     Value1 is the symbol whose print name matches STRING in package PKG.  If the  symbol is
            not found one is created and interned in PKG.  
     Value2 takes on one of the following values :
            NIL - symbol was created
            :INTERNAL - symbol was directly present as an internal symbol
            :EXTERNAL - symbol was directly present as an external symbol
            :INHERITED - symbol was inherited from a package used by PKG
   An implementation extension third value is also returned.  This value is the name of the
   package in which the symbol was found.  This is different from PKG only if the value of
   Value2 is :INHERITED.")

;(setf (arglist COND) (&QUOTE &REST CLAUSES))
(setf (documentation 'COND 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (COND (test1 body1) ... (testN bodyN)) is defined as :  the (test# body#) construct is called
   a clause.  <test1> is evaluated and, if true, all forms in <body1> are evaluated sequentially
   and the value of the last expression in <body1> is returned by COND.  If <body1> is not specified,
   the value of <test1> is returned by COND.  If <test1> evaluates to NIL, <test2> is evaluated and
   checked for a true value, and so on, until either one <test#> evaluates to true or all clauses are
   exhausted.  COND returns NIL if no clause is selected.  An OTHERWISE or T as the <testN> always
   evalutes to true.")

;(setf (arglist GENTEMP) (&OPTIONAL (PREFIX T) (PKG *PACKAGE*)))
(setf (documentation 'GENTEMP 'function)
  "Returns a symbol whose name begins with PREFIX which is both unique and interned in package PKG.")

;(setf (arglist PLUSP) (X))
(setf (documentation 'PLUSP 'function)
  "Returns true is X is greater than zero, otherwise NIL.  PLUSP does not work on complex numbers.")

;(setf (arglist IDENTITY) (X))
(setf (documentation 'IDENTITY 'function) 
  "Returns X.")

;(setf (arglist SUBST) (NEW OLD TREE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'SUBST 'function)
  "Returns the tree structure created by replacing each subtree or element of TREE that matches OLD
   with NEW.  Object are said to match if predicate :TEST returns true or, if specified, :TEST-NOT 
   returns NIL.  If non-NIL, :KEY is a function applied to each node of TREE to obtain the object to
   test against.")
 
(setf (documentation '+++ 'variable) 
  "Variable whose value is the form evaluated three interactions ago by read-eval-print loop.")

;(setf (arglist MULTIPLE-VALUE-BIND) (&QUOTE VARLIST EXP &REST BODY))
(setf (documentation 'MULTIPLE-VALUE-BIND 'function)
  "Macro used to bind values returned from the evaluation of a form to variables.  Syntax of  
   (MULTIPLE-VALUE-BIND (var1 ... varN) (value-form)  {form}*) is defined as : evaluate the form
   <value-form> and bind these results to the respective variables <var1> thru <varN>.  If there
   are more results than variables, they are discarded; if more variables that results, the
   variables are bound to NIL.  All forms in <form> are then evaluated sequentially as if in a
   PROGN with the result of the last form evaluated returned by MULTIPLE-VALUE-BIND.") 

;(setf (arglist BREAK) (&OPTIONAL FORMAT-STRING &REST ARGS))
(setf (documentation 'BREAK 'function) 
  "Signals an error and provides a read-eval-print loop for use as a subroutine.  FORMAT-STRING is a
   control string given to FORMAT and describes the error.  ARGS are the format arguments which the 
   control string will use if required.  The read-eval-print loop uses the Universal Command Loop.
   BREAK returns NIL.")

;(setf (arglist +) (&REST NUMBERS))
(setf (documentation '+ 'function) 
  "Returns the sum of the arguments.")

(setf (documentation '+ 'variable)
  "Variable whose value is the last form evaluated by read-eval-print loop.")

;(setf (arglist APPLY) (FCT ARG &REST ARGS))
(setf (documentation 'APPLY 'function)
  "Applies the function FCT to the argument list ARG and ARGS which have been appended into
   one list.  FCT cannot be a macro or special form and the last argument must be a list.  The
   return value of APPLY is the result obtained by applying the function to the argument list.")

;(setf (arglist ECASE) (TEST-OBJECT &BODY CLAUSES))
(setf (documentation 'ECASE 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (ECASE TEST-OBJECT (test1 body1) ... (testN bodyN)) is defined as :  the (test# body#)
   construct is called a clause.  TEST-OBJECT is evaluated.  <test1> is evaluated and compared
   against TEST-OBJECT; if they are EQL, all forms in <body1> are evaluated sequentially and the
   value of the last expression in <body1> is returned by ECASE.  If <test1> does not match,
   <test2> is evaluated and compared, and so on, until either one <test#> matches TEST-OBJECT or
   all clauses are exhausted.  If no clause is selected, ECASE signals an error.  An OTHERWISE
   or T as the <test#> is not permitted.")

;(setf (arglist MACROEXPAND) (FORM &OPTIONAL ENV))
(setf (documentation 'MACROEXPAND 'function)
  "Returns two values :
     Value1 is FORM expanded until the result is not a macrocall in an environment where the local
            macro definitions are defined by ENVIRONMENT.
     Value2 is T if one or more expansions have taken place.")

;(setf (arglist CDAR) (X))
(setf (documentation 'CDAR 'function)
  "Returns the result of (CDR (CAR X)).")

;(setf (arglist CHAR) (STRING INDEX))
(setf (documentation 'CHAR 'function)
  "Returns the character from string STRING at position INDEX.") 

;(setf (arglist REALP) (X))
(setf (documentation 'REALP 'function) 
  "Returns true if X is any non complex number, otherwise NIL.")

;(setf (arglist NSET-EXCLUSIVE-OR) (LIST1 LIST2 &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'NSET-EXCLUSIVE-OR 'function)
  "Returns a list consisting of elements that are members of LIST1 but not members of LIST2 plus
   those elements that are members of LIST2 but not members of LIST1.  :TEST is a predicate that
   returns true if the elements match; :TEST-NOT is a predicate that returns NIL if the elements
   match.  If :KEY is specified, it is a function applied to elements of both lists before the 
   test is performed.  Both LIST1 and LIST2 may be destroyed.")

(setf (documentation 'BOOLE-IOR 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-or of BOOLE's second
   argument with BOOLE's third argument.")

;(setf (arglist RPLACA) (CONS NEW-CAR))
(setf (documentation 'RPLACA 'function) 
  "Destructively replaces the CAR of CONS with NEW-CAR and returns the modified CONS.")

;(setf (arglist ENDP) (LIST))
(setf (documentation 'ENDP 'function)
  "Returns true if LIST is NIL or NIL if LIST is a cons; in all other cases an error is signaled.") 

;(setf (arglist CHAR<=) (CHAR &REST CHARS))
(setf (documentation 'CHAR<= 'function)
  "Returns true if the arguments are nondecreasing, considering bits, fonts, and case;
   otherwise, NIL is returned.")

(setf (documentation 'SHORT-FLOAT-EPSILON 'variable)
  "Constant whose value is the smallest possible positive number of type short float that, when
   added to some value, will change the magnitude of that value.")

;(setf (arglist DECODE-UNIVERSAL-TIME) (UNIVERSAL-TIME &OPTIONAL TIMEZONE))
(setf (documentation 'DECODE-UNIVERSAL-TIME 'function)
  "Returns nine values representating the inputted UNIVERSAL-TIME in the following order :
     second, minute, hour, date, month, year, day-of-week, daylight-savings-time-p, and time zone.
   If TIMEZONE is specified, time is given relative to that time zone.")

;(setf (arglist STRINGP) (X))
(setf (documentation 'STRINGP 'function)
  "Returns true if X is a string, otherwise NIL.") 

(setf (documentation '*STANDARD-OUTPUT* 'variable)
  "Global variable whose value is the default stream for output functions such as PRINT.")

;(setf (arglist FLOATP) (X))
(setf (documentation 'FLOATP 'function) 
  "Returns true if X is a floating point number, otherwise NIL.")

;(setf (arglist MAKE-CHAR) (CHAR &OPTIONAL (BITS 0) (FONT 0)))
(setf (documentation 'MAKE-CHAR 'function) 
  "Returns a character with the same character code as CHAR, but with the bits attribute and
   font attribute set by BITS and FONT, respectively.  If this character cannot be made,
   NIL is returned.")

;(setf (arglist NUMERATOR) (RATIONAL))
(setf (documentation 'NUMERATOR 'function) 
  "Returns the numerator of rational number RATIONAL.  The numerator of an integer is that integer.")

;(setf (arglist DOTIMES) ((VAR LIMIT RESULTFORM) &BODY BODY))
(setf (documentation 'DOTIMES 'function)
  "Iterates BODY over a sequence of integers.  LIMIT is evaluated.  BODY is performed once for
   each integer from zero to LIMIT (exclusive) with VAR bound to this integer value.  RESULTFORM
   is then evaluated and returned as the result of DOTIMES.  NIL is returned if RESULTFORM is
   not specified.") 

;(setf (arglist DEFSTRUCT) (NAME-AND-OPTIONS &REST SLOT-OPTIONS))
(setf (documentation 'DEFSTRUCT 'function)
  "Construct used to define data structures.  The structure, named NAME-AND-OPTIONS if 
   NAME-AND-OPTIONS is a symbol or the car of the list if NAME-AND-OPTIONS is a list, is acceptable
   as a data type to function TYPEP.  SLOT-OPTIONS defines the individual components of the data
   structure.  Accessing functions are automatically defined for creating and accessing the data
   structure.  DEFSTRUCT returns the name of the data structure.")

;(setf (arglist CHAR-DOWNCASE) (CHAR))
(setf (documentation 'CHAR-DOWNCASE 'function)
  "Returns the lower case version of CHAR with the same bit and font attributes of CHAR if CHAR
   is an alphabetic character.  If CHAR is not alphabetic, CHAR-DOWNCASE returns CHAR unchanged.
   Note: if the bits attribute of CHAR is non-zero, it is not considered alphabetic.")

;(setf (arglist SUBST-IF) (NEW PREDICATE TREE &KEY KEY))
(setf (documentation 'SUBST-IF 'function)
  "Returns the tree structure created by replacing each atom or tree in TREE that satisfies PREDICATE
   with NEW.  If non-NIL, :KEY is a function applied to each node of TREE to obtain the object to
   test against.  TREE is unaltered.") 

;(setf (arglist LOCALLY) (&QUOTE &REST BODY))
(setf (documentation 'LOCALLY 'function)
  "Executes BODY within the context of the given declarations.")

;(setf (arglist DESCRIBE) (ANYTHING &OPTIONAL NO-COMPLAINTS))
(setf (documentation 'DESCRIBE 'function) 
  "Prints to the stream *STANDARD-OUTPUT* the value or components of the object ANYTHING. 
   What is printed is implementation dependent.  DESCRIBE returns no values.")

;(setf (arglist SYMBOL-VALUE) (SYMBOL))
(setf (documentation 'SYMBOL-VALUE 'function)
  "Returns the current value of the dynamic symbol SYMBOL.  An error is signaled if SYMBOL
   is unbound.")

;(setf (arglist LOG) (NUMBER &OPTIONAL *BASE))
(setf (documentation 'LOG 'function)
  "Returns the logarithm of NUMBER in the base *BASE.  *BASE defaults to e, the base of
   natural logarithms.")

(setf (documentation 'LOG 'structure) 
  "Data structure used to log namespace changes.")

;(setf (arglist CADAAR) (X))
(setf (documentation 'CADAAR 'function)
  "Returns the result of (CAR (CDR (CAR (CAR X))))")

(setf (documentation '*PRINT-LENGTH* 'variable)
  "Variable whose value, when non-NIL, is an integer specifying how many elements of a list
   are printed before the remaining list is representated by ellipses (...).")

(setf (documentation 'BOOLE-SET 'variable)
  "Constant used as first argument to function BOOLE to make all output bits one.")

;(setf (arglist CAADAR) (X))
(setf (documentation 'CAADAR 'function)
  "Returns the result of (CAR (CAR (CDR (CAR X)))).")

;(setf (arglist CHAR-CODE) (CHAR))
(setf (documentation 'CHAR-CODE 'function) 
  "Returns as a non-negative integer the code attribute of the character object CHAR.")

;(setf (arglist SET-EXCLUSIVE-OR) (LIST1 LIST2 &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'SET-EXCLUSIVE-OR 'function)
  "Returns a list consisting of elements that are members of LIST1 but not members of LIST2 plus
   those elements that are members of LIST2 but not members of LIST1.  :TEST is a predicate that
   returns true if the elements match; :TEST-NOT is a predicate that returns NIL if the elements
   match.  If :KEY is specified, it is a function applied to elements of both lists before the 
   test is performed.  Neither LIST1 and LIST2 is altered.")

;(setf (arglist CHAR-BITS) (CHAR))
(setf (documentation 'CHAR-BITS 'function)
  "Returns as a non-negative integer the bits attribute of the character object CHAR.")

;(setf (arglist LISTEN) (&OPTIONAL STREAM))
(setf (documentation 'LISTEN 'function)
  "Returns true is input is available on STREAM, otherwise NIL.") 

(setf (documentation '*READTABLE* 'variable)
  "Variable whose value is the current readtable.") 

;(setf (arglist <=) (NUMBER &REST NUMBERS))
(setf (documentation '<= 'function) 
  "Returns true if the arguments are in nondecreasing numerical order, otherwise NIL.")

;(setf (arglist FIRST) (LIST))
(setf (documentation 'FIRST 'function)
  "Returns the first element of LIST, that is, (CAR LIST).  One-origin numbering is used.")

;(setf (arglist ATAN) (Y &OPTIONAL (X (FLOAT 1 Y))))
(setf (documentation 'ATAN 'function)
  "Returns the angle in radians whose tangent is Y.  If X is specified, the tangent used is Y/X.
   The sign of Y and X determine which angle, differing by pi and with the same tangent, is
   returned.  Y may be complex if X is not specified; X may be 0 if Y is not 0.") 

;(setf (arglist TRACE) (&REST SPECS))
(setf (documentation 'TRACE 'function)
  "Traces one or more functions during execution.  The default tracing action is to print the
   function's name and arguments when called and its name and value(s) when returned from.
   Valid implementation extension TRACE options are :
      :break <pred> - Enter a breakpoint before function execution if <pred> is true.
      :exitbreak <pred> - Enter a breakpoint after function execution if <pred> is true. 
      :error - Causes the debugger to be invoked before the function is executed.
      :step - Causes the function to be single stepped during its execution.
      :entrycond <pred> - Prints trace information on entry only if <pred> is true.
      :exitcond <pred> - Prints trace information on exit only if <pred> is true.
      :cond <pred> - Specifies both :entrycond and :exitcond.
      :wherein <fun> - Causes traced function to be traced only when called from <fun>.
      :argpdl <pdl> - Specifies the symbol <pdl> will contain the recent function history.
      :entryprint <form> - Evaluates and prints <form> before the function is executed.
      :exitprint <form> - Evaluates and prints <form> after the function is executed.
      :print <form> - Specifies both :entryprint and :exitprint.
      :entry <list> - Specifies the list of forms to be evaluated and printed on entry.
      :exit <list> - Specifies the list of forms to be evaluated and printed on exit.
      :arg - Prints the function's name and arguments on entry.
      :value - Prints the function's returned value(s) on exit.
      :both - Specifies both :arg and :value (the default of TRACE).")

;(setf (arglist WARN) (TYPE SEVERITY FORMAT-STRING &REST ARGS))
;(setf (documentation 'WARN 'function) 
;  "Record and print a compiler warning. TYPE describes the particular kind of problem,
;such as FUNCTION-NOT-VALID. SEVERITY is a symbol in the keyword package giving a
;broader classification; see the source for a list of possible severities.  FORMAT-STRING
;and ARGS are used to print the warning.")

;(setf (arglist SIN) (RADIANS))
(setf (documentation 'SIN 'function) 
  "Returns the sine of angle RADIANS.  RADIANS is measured in radians and may be complex.")

(setf (documentation 'MOST-POSITIVE-SHORT-FLOAT 'variable)
  "Constant whose value is the greatest positive magnitude representable by the short float type.") 

;(setf (arglist SPECIAL-FORM-P) (SYMBOL))
(setf (documentation 'SPECIAL-FORM-P 'function)
  "Returns true if SYMBOL is a special form, otherwise NIL.") 

(setf (documentation 'BOOLE-XOR 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-xor of BOOLE's second
   argument with BOOLE's third argument.")

;(setf (documentation 'READTABLE 'variable) 
;  " ")

;(setf (arglist MEMBER) (ITEM LIST &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'MEMBER 'function)
  "Returns a list consisting of the tail of list LIST whose CAR is the first element of LIST that
   matches ITEM according to test :TEST.  If no match is found, NIL is returned.  If :TEST-NOT is
   used, the CAR of the returned list is the first element of LIST that failed test :TEST-NOT. 
   If non-NIL, :KEY is a function applied to each element of LIST before the test is performed.")   

;(setf (arglist SUBTYPEP) (X Y))
(setf (documentation 'SUBTYPEP 'function)
  "Returns two values :
     Value1 is true if X is definitely a subtype of Y, NIL if it may not be a subtype.
     Value2 is true if X is definitely a subtype of Y or definitely not a subtype of Y; otherwise,
            NIL is returned to denote the system could not determine if X is a subtype of Y.") 

;(setf (arglist COS) (RADIANS))
(setf (documentation 'COS 'function)
  "Returns the cosine of angle RADIANS.  RADIANS is measured in radians and may be complex.")

;(setf (arglist HASH-TABLE-SIZE) (HASH-TABLE))
(setf (documentation 'HASH-TABLE-SIZE 'function) 
  "Returns the size in number of entries of hash table HASH-TABLE.")

;(setf (arglist DEPOSIT-FIELD) (VALUE PPSS FIXNUM))
(setf (documentation 'DEPOSIT-FIELD 'function)
  "Returns the integer composed by substituting the bits of VALUE for the PPSS bits of FIXNUM.
   PPSS is a byte specifier.") 

;(setf (arglist SETQ) (&QUOTE &REST SYMBOLS-AND-VALUES))
(setf (documentation 'SETQ 'function)
  "Assigns values to symbols.  SYMBOLS-AND-VALUES is a list of alternating symbol-value pairs.
   The SETQ syntax of (SETQ var1 exp1 var2 exp2 ... varN expN) is defined as : evaluate <exp1>
   and assign this result to <var1>, evaluate <exp2> and assign this result to <exp2>, and so on. 
   The value of the last expression evaluated is returned by SETQ.")
 
;(setf (arglist CADDAR) (X))
(setf (documentation 'CADDAR 'function) 
  "Returns the result of (CAR (CDR (CDR (CAR X)))).")

;(setf (arglist HASH-TABLE-COUNT) (HASH-TABLE))
(setf (documentation 'HASH-TABLE-COUNT 'function) 
  "Returns the number of filled entries in hash table HASH-TABLE.")

;(setf (arglist DEFMACRO) (&REST X))
(setf (documentation 'DEFMACRO 'function)
  "Macro used to define the construct of macros.  Syntax of 
   (DEFMACRO name lambda-list {declaration | doc-string}* {forms}*) is defined as :
   <name> is the symbol whose macro definition is being created; <lambda-list> accepts most of the
   keywords of a DEFUN lambda-list in additional to the markers &body, &whole, and &environment;
   <declaration> are a list of explicit declarations; <doc-string> is the additional description of
   the macro that becomes part of the debug information; <forms> are the forms that constitute the
   body of the expander function.  DEFMACRO returns <name>.  When the macro defined by DEFMACRO is
   called, the forms in <forms> are evaluated as an implicit PROGN and the value of the last form is
   returned as the expansion of the macro call.")

;(setf (arglist LET) (&QUOTE VARLIST &REST BODY))
(setf (documentation 'LET 'function)
  "Special form used to establish bindings of variables.  Syntax of 
   (LET ((var1 exp1)...(varN expN)) [(DECLARE ...)] body) is defined as :  all expressions <exp1>
   thru <expN> are evaluated and their resulting values saved.  These values are then bound to <var1> 
   thru <varN>, respectively, in parallel.  If an expression is not specified, NIL is used.  These
   binding are in force only over the scope of the LET.    All forms in <body> are then evaluated as
   a PROGN with these binding in force; LET returns the result of the last form evaluated.")

(setf (documentation '/// 'variable) 
  "Variable whose value is a list of the results obtained from the form evaluated three
   interactions ago by the read-eval-print loop.")

;(setf (arglist NAMESTRING) (OBJECT))
(setf (documentation 'NAMESTRING 'function) 
  "Returns as a string the host specific full form pathname of OBJECT.")

;(setf (arglist STRING/=) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING/= 'function)
  "Returns the index, relative to STRING1, of the first difference found between STRING1 and
   STRING2.  Although the entire length of both STRING1 and STRING2 is the default, a substring
   of STRING1 starting at :START1 and ending at :END1 and/or a substring of STRING2 starting at
   :START2 and ending at :END2 may be specified.  Case and font are significant during the
   comparison.  STRING/= returns NIL if the strings are equal.") 

;(setf (arglist NOTEVERY) (PREDICATE &REST SEQUENCES))
(setf (documentation 'NOTEVERY 'function)
  "Returns true if PREDICATE returns NIL when applied to some element of SEQUENCES.  If every
   application of PREDICATE returns true, then NOTEVERY returns NIL.  If multiple sequences are
   supplied, element 0 of each sequence is tested, then element 1, and so on.  PREDICATE must
   accept the same number of arguments as sequences provided.  If one of the sequences is
   exhausted before NOTEVERY returns true, NOTEVERY returns NIL.")

(setf (documentation 'BOOLE-C2 'variable)
  "Constant used as first argument to function BOOLE to take the complement of BOOLE's third
   argument.")

;(setf (arglist OR) (&QUOTE &REST EXPRESSIONS))
(setf (documentation 'OR 'function) 
  "Evaluates the forms one at a time from left to right until one returns true.  Evaluation
   then stops and the true value is returned.  If no forms are included or none return true, OR
   returns NIL.  Only the last form can return multiple values and it must be the first that
   evaluates to true in order to return multiple values.")

;(setf (arglist EVAL-WHEN) (&QUOTE TIMES &REST FORMS))
(setf (documentation 'EVAL-WHEN 'function) 
  "Processes the FORMS only at the specified TIMES.  TIMES is a list which may include the symbols
   eval, load, and/or compile.  <eval> means to evaluate the FORMS if the EVAL-WHEN is processed by
   the interpreter, or to compile and evaluate the FORMS when compiling to memory.  <load> means
   the compiler, when compiling to a file, should compile the FORMS if appropriate and also cause
   them to be executed when the object file is loaded.  <compile> means the compiler should execute
   the forms at compile time.  (EVAL LOAD) is the normal operations mode.")

(setf (documentation '*READ-BASE* 'variable) 
  "Variable whose value is the default radix for reading integers and ratios.")

;(setf (arglist COMPILE) (NAME &OPTIONAL LAMBDA-EXP TARGET (MODE (QUOTE COMPILE-TO-CORE))))
(setf (documentation 'COMPILE 'function) 
  "Compiles the definition of NAME, or its previous interpreted definition if it is already
   compiled, and returns NAME.  If LAMBDA-EXP is supplied, it is compiled and made the definition
   of NAME.  If NAME is NIL, LAMBDA-EXP is compiled and the result is returned.")

;(setf (arglist SIMPLE-BIT-VECTOR-P) (OBJECT))
(setf (documentation 'SIMPLE-BIT-VECTOR-P 'function) 
  "Returns true if OBJECT is a simple bit vector, otherwise NIL.")

;(setf (arglist EVERY) (PREDICATE &REST SEQUENCES))
(setf (documentation 'EVERY 'function)
  "Returns true if PREDICATE returns true when applied to all element of SEQUENCES.  If some
   appplication of PREDICATE returns NIL, then EVERY returns NIL.  If multiple sequences are
   supplied, element 0 of each sequence is tested, then element 1, and so on.  PREDICATE must
   accept the same number of arguments as sequences provided.  If one of the sequences is
   exhausted before EVERY returns NIL, EVERY returns true.")

;(setf (arglist STRING-LESSP) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING-LESSP 'function)
  "Returns the index into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test less than ignoring case and fonts.  If the test is never satisfied, NIL
   is returned.  If the strings match, the length of the shortest string is returned.  Although
   the default string length of both STRING1 and STRING2 is the entire length, a substring of
   STRING1 starting at :START1 and ending at :END1 and/or a substring of STRING2 starting at
   :START2 and ending at :END2 may be specified.")

;(setf (arglist PRIN1-TO-STRING) (OBJECT))
(setf (documentation 'PRIN1-TO-STRING 'function)
  "Returns as a string and in the form that would be printed by PRIN1 the object OBJECT.") 

(setf (documentation 'SINGLE-FLOAT-EPSILON 'variable)
  "Constant whose value is the smallest possible positive number of type single float that,
   when added to some value, will change the magnitude of that value.")

;(setf (arglist COUNT-IF) (PRED SEQUENCE &KEY KEY START END FROM-END))
(setf (documentation 'COUNT-IF 'function)
  "Returns as a non negative integer the number of elements in SEQUENCE which, when PRED is
   applied to them, return true.  Although the entire length of SEQUENCE is the default, a
   subsequence starting at :START and ending at :END may be specified.  If non-NIL, KEY is a
   function applied to each element of SEQUENCE before the test is performed.  FROM-END has
   no effect on the results.")

;(setf (arglist PATHNAME-HOST) (P))
(setf (documentation 'PATHNAME-HOST 'function)
  "Returns the host component of pathname P.")

;(setf (arglist VECTOR-POP) (ARRAY))
(setf (documentation 'VECTOR-POP 'function) 
  "Decrements the fill pointer of ARRAY by 1 and returns the element pointed to by this new
   value.  An error is signaled if ARRAY is empty.")

;(setf (arglist SINH) (NUMBER))
(setf (documentation 'SINH 'function)
  "Returns the hyperbolic sine of NUMBER as a floating-point or complex number.")

;(setf (arglist PROGV) (VARS VALS &QUOTE &REST BODY))
(setf (documentation 'PROGV 'function)
  "Binds the variables whose names are in the list VARS with their corresponding by position values
   in the list VALS, then executes all forms in BODY with these bindings.  The elements of both
   VARS and VALS are evaluated upon entry to PROGV and their binding are undone upon exit.  The 
   results of the last form evaluated in BODY are the return values of PROGV.")

;(setf (arglist CHAR=) (CHAR &REST CHARS))
(setf (documentation 'CHAR= 'function) 
  "Returns true if all the characters are equal, considering bits, font, and case; otherwise,
   NIL is returned.")

(setf (documentation 'LEAST-NEGATIVE-SINGLE-FLOAT 'variable)
  "Constant whose value is the least negative magnitude representable by the single float type.")

;(setf (arglist REMPROP) (SYMBOL-OR-PLIST PROPERTY))
(setf (documentation 'REMPROP 'function)
  "Removes the property pair from property list SYMBOL-OR-PLIST whose indicator is EQ to PROPERTY.
   SYMBOL-OR-PLIST is destructively modified by REMPROP.  REMPROP returns true if a property is
   removed, NIL if no match is found.") 

;(setf (arglist FIND) (ITEM SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END FROM-END))
(setf (documentation 'FIND 'function)
  "Returns the leftmost element found in sequence SEQUENCE which satisfies the desired test 
   against ITEM.  Although the entire length of SEQUENCE is the default, a  subsequence starting
   at :START and ending at :END may be specified.  :TEST is a predicate that should return true in
   order to have that element of SEQUENCE returned by FIND.  :TEST-NOT is a predicate which causes
   the first element which does not satisfy the test to be returned.  If non-NIL, :KEY is a function
   applied to each element of SEQUENCE before the test is performed.  :FROM-END will cause the
   rightmost element which satisfies the test to be returned.  An implementation extension second
   value, the index of the returned element, is also returned.")

;(setf (arglist RENAME-PACKAGE) (PKG NEW-NAME &OPTIONAL NEW-NICKNAMES))
(setf (documentation 'RENAME-PACKAGE 'function)
  "Replaces the old name and all old nicknames of package PKG with NEW-NAME and NEW-NICKNAMES,
   respectively.  An error is signaled if PKG does not exist or if NEW-NAME or NEW-NICKNAMES 
   are names already in use for another package.")

;(setf (arglist CHAR-INT) (CHARACTER))
(setf (documentation 'CHAR-INT 'function) 
  "Returns a nonnegative integer whose value represents the character object CHARACTER.")

;(setf (arglist RATIONALP) (X))
(setf (documentation 'RATIONALP 'function) 
  "Returns true is X is a rational number (an integer or ratio), otherwise NIL.")

;(setf (arglist LOGANDC1) (INTEGER1 INTEGER2))
(setf (documentation 'LOGANDC1 'function)
  "Returns the bitwise-and of the complement of INTEGER1 with INTEGER2.") 

;(setf (arglist RETURN) (&QUOTE &REST VALS))
(setf (documentation 'RETURN 'function) 
  "Exits a block named NIL and returns the result from the evaluation of VALS.  Forms such as
   DO, PROG, and LOOP create blocks named NIL.")

;(setf (arglist ROUND) (DIVIDEND &OPTIONAL (DIVISOR 1)))
(setf (documentation 'ROUND 'function)
  "Returns two values :
     Value1 is DIVIDEND divided by DIVISOR, rounded to the nearest integer, or DIVIDEND rounded
            to the nearest integer if DIVISOR is not specified.  A number exactly halfway
            between two numbers rounds to the even number.
     Value2 is the number to add to Value1 (or Value1 times DIVISOR if DIVISOR is specified) in
            order to equal DIVIDEND.")

;(setf (arglist TANH) (NUMBER))
(setf (documentation 'TANH 'function) 
  "Returns the hyperbolic tangent of NUMBER as a floating-point or complex number.")

(setf (documentation 'BOOLE-ORC1 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-or of the complement
   of BOOLE's second argument with BOOLE's third argument.")

;(setf (arglist CAADR) (X))
(setf (documentation 'CAADR 'function) 
  "Returns the result of (CAR (CAR (CDR X))).")

(setf (documentation 'CHAR-BITS-LIMIT 'variable)
  "Constant whose value is the exclusive upper bound on the modifier bit attribute value of any
   character.") 

(setf (documentation '++ 'variable)
  "Variable whose value is the form evaluated two interactions ago by read-eval-print loop.")

;(setf (arglist DISASSEMBLE) (FUNCTION &KEY BASE VERBOSE START END))
(setf (documentation 'DISASSEMBLE 'function)
  "Prints in humanly readable form on *STANDARD-OUTPUT* a disassembly of FUNCTION.  Implementation
   extension keywords are :
      :BASE - specifies what base to print the numbers output by DISASSEMBLE.
      :VERBOSE - causes the printing of the numeric representation of the macrocode to be included.
      :START - specifies where in the macrocode to begin printing.
      :END - specifies where in the macrocode to end printing.")

;(setf (arglist NBUTLAST) (LIST &OPTIONAL (N 1)))
(setf (documentation 'NBUTLAST 'function)
  "Returns a list with the same elements as LIST but with the last N elements deleted.  LIST is
   destructively modified.  If the number of elements in LIST is less than or equal to N, NIL
   is returned and LIST is not modified.") 

;(setf (arglist NINTERSECTION) (LIST1 LIST2 &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'NINTERSECTION 'function)
  "Returns a list consisting of all elements that are members of both LIST1 and LIST2.  Although
   LIST2 is unmodified, LIST1 may be destructively modified.  :TEST specifies the test that returns
   true to denote intersection; :TEST-NOT specifies the test that returns NIL to denote intersection.
   If non-NIL, :KEY specifies a function to be applied to each element of each list before the test.")

;(setf (arglist SIMPLE-STRING-P) (OBJECT))
(setf (documentation 'SIMPLE-STRING-P 'function)
  "Returns true is OBJECT is a simple string (not displaced and without fill pointers),
   otherwise NIL.") 

(setf (documentation '*BREAK-ON-WARNINGS* 'variable)
  "Variable whose value, when set to true, causes the function WARN to call the function BREAK.") 

(setf (documentation '*FEATURES* 'variable)
  "Variable whose value is a list of symbols naming implementation dependent features for this Lisp
   environment.") 

;(setf (arglist MIN) (&REST NUMBERS))
(setf (documentation 'MIN 'function)
  "Returns the argument with the smallest value.")

;(setf (arglist STREAM-ELEMENT-TYPE) (STREAM))
(setf (documentation 'STREAM-ELEMENT-TYPE 'function)
  "Returns a Common Lisp type specifier describing what objects can be read from or written to STREAM.")
 
;(setf (arglist ZEROP) (NUMBER))
(setf (documentation 'ZEROP 'function) 
  "Returns true if NUMBER is equal to zero, otherwise NIL.")

;(setf (arglist GET-INTERNAL-RUN-TIME) NIL)
(setf (documentation 'GET-INTERNAL-RUN-TIME 'function) 
  "Returns time in 60ths of a second since the last boot.")

;(setf (arglist FORMAT) (STREAM CTL-STRING &REST ARGS))
(setf (documentation 'FORMAT 'function)
  "Formats arguments according to a control string and prints to a stream.  If STREAM is T,
   *STANDARD-OUTPUT* is used; if NIL, FORMAT returns a string containing the formatted text.
   The control string is copied to the stream with ~ indicating special formatting commands.
 STRING and NUMBER formatting commands :
   ~A - Print as ASCII output; complete syntax is ~mincol,colinc,minpad,padcharA.
        ~@A - Right justify the string; ~:A - Print NIL as (); ~:@A - Both justify and ().
   ~S - Same as ~A, but uses PRIN1 instead of PRINC.
   ~C - Print a character;  Control characters preceeded by h-, s-, m-, or c- as needed.
        ~:C - Print format effectors as names; ~@C - Print character in READ format;
        ~:@C - Like ~:C except names match keycaps with SYMBOL and SYMBOL-SHIFT notations included.
   ~D - Print number as a decimal integer; complete syntax is ~mincol,padchar,commacharD.
        ~:D - Print comma every three digits; ~@D - Always print sign; ~:@D - both sign and comma.
   ~F - Print float in nonexponential notation; complete syntax is ~w,d,s,overflowchar,padcharF.
   ~E - Print float in exponential notation;
        complete syntax is ~w,d,e,s,overflowchar,padchar,exptcharE.
   ~G - Same as ~E, but will print the number without exponent if possible.
   ~O - Octal equivalent of ~D.
   ~X - Hexidecimal equivalent of ~D.
   ~B - Binary equivalent of ~D.
   ~$ - Print floating point number as money amount; complete syntax is ~w,x,y,z$.
        ~@$ - Print the + sign; ~:$ - Print the + sign before the padding.
   ~R - Print number as an English cardinal number;  ~:R - English ordinal number;
        ~@R - Roman numeral; ~:@R - Old Roman numeral; ~nR - Print number in radix <n>.
 CONTROL formatting commands : 
   ~* - Ingore an argument; ~n* - Ignore <n> arguments; ~:n - Back up <n> arguments.
   ~% - Insert a newline; ~n% - Insert <n> newlines.
   ~~ - Insert a tilde; ~n~ - Insert <n> tildes.
   ~| - Insert a form feed; ~n| - Insert <n> form feeds; ~:| - Insert a form feed after a
        CLEAR-SCREEN; ~:n| - Insert <n> form feeds after a CLEAR-SCREEN.
   ~<cr> - Ignore the CR and following whitespace in control string.
           ~:<cr> - Ingore the CR but not whitespace; ~@<cr> - Ignore whitespace but not CR.
   ~& - Output a CR; ~n& - Output <n> CRs.
   ~^ - Terminate processing if no more arguments; ~n - Terminate if <n>=0; ~n,m - Terminate if
        <n>=<m>; ~n,m,p - Terminate if <n><m><p>.
   ~T - Tab to column; complete syntax is mincol,colincT.
        ~:T - Tabs in pixels instead of characters; ~@T - Insert spaces to tab column
   ~P - Pluralize the output by inserting s when argument > 1; ~:P - Use previous not next argument.
        ~@P - Plural is ies, single is y; ~:@P - Both previous argument and ies.
   ~? - Use two arguments, first as format string, the second as argument to format string.
   ~[ - Select one of the clauses to format between ~[ and ~] depending upon argument.
   ~( - Prints output between ~( and ~) in lower case;  ~:( - Capitalize each word;
        ~@( - Capitalize the first word; ~:@( - Force output to be uppercase.
   ~< - Format all strings then output all strings with padding between them; complete syntax is
        ~mincol,colinc,minpad,padchar<str0~;str1~;...~;strn~>.
   ~{ - Use this argument as the format string for each element in the argument. Terminates with ~}.
   ~Q - Calls the argument as a function with ~Q parameters being the functions argument list.
   ~ - Successive lines within ~...~ are indented to align each on the same column as first line.
   ~M - Makes the item printed in ~A format mouse sensitive if stream supports that operation.
        ~:M - Recursive call to make a sentence mouse sensitive.")

;(setf (arglist ARRAY-HAS-FILL-POINTER-P) (ARRAY))
(setf (documentation 'ARRAY-HAS-FILL-POINTER-P 'function) 
  "Returns true if ARRAY has a fill pointer, otherwise NIL.")

;(setf (arglist CHAR-EQUAL) (CHAR &REST CHARS))
(setf (documentation 'CHAR-EQUAL 'function) 
  "Returns true if all the characters are equal, ignoring font and case; otherwise,
   NIL is returned.")

;(setf (arglist COUNT-IF-NOT) (PRED SEQUENCE &KEY KEY START END FROM-END))
(setf (documentation 'COUNT-IF-NOT 'function)
  "Returns as a non negative integer the number of elements in SEQUENCE which, when PRED is
   applied to them, return NIL.  Although the entire length of SEQUENCE is the default, a
   subsequence starting at :START and ending at :END may be specified.  If non-NIL, :KEY is a
   function applied to each element of SEQUENCE before the test is performed.  :FROM-END has
   no effect on the results.")

;(setf (arglist CLOSE) (STREAM &KEY &OPTIONAL ABORT))
(setf (documentation 'CLOSE 'function) 
  "Closes the stream STREAM.  Input/output can no longer be performed to it.  If non-NIL, ABORT
   directs the system to clean up any side effects from creating the stream, including deleting 
   a newly created output file if possible.")

;(setf (arglist MAKE-RANDOM-STATE) (&OPTIONAL STATE))
(setf (documentation 'MAKE-RANDOM-STATE 'function)
  "Returns an object of type random-state.  If the value of STATE is T, MAKE-RANDOM-STATE creates
   and returns a new random-state object based on the current time; if NIL, the value of variable
   *RANDOM-STATE* is returned; if a valid random-state object, the value of that object is returned.")

;(setf (arglist NTH) (N LIST))
(setf (documentation 'NTH 'function) 
  "Returns the Nth element of LIST.  Zero-origin numbering is used.")

;(setf (arglist DO-EXTERNAL-SYMBOLS) ((VARIABLE PACK RESULT-FORM) &BODY BODY))
(setf (documentation 'DO-EXTERNAL-SYMBOLS 'function)
  "Iterates over the symbols in a package.  BODY is performed once for each external symbol in
   PACK with VARIABLE bound to that symbol.  RESULT-FORM is then evaluated and returned as the
   result of DO-EXTERNAL-SYMBOLS.  NIL is returned if RESULT-FORM is not specified.")    

;(setf (arglist THIRD) (LIST))
(setf (documentation 'THIRD 'function) 
  "Returns the third element of LIST.  One-origin numbering is used.")

;(setf (arglist PACKAGE-USE-LIST) (PKG))
(setf (documentation 'PACKAGE-USE-LIST 'function) 
  "Returns the list of package objects used by package PKG.")

;(setf (arglist BIT-NOT) (BIT-ARRAY &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-NOT 'function) 
  "Returns a bit array containing the complement of the elements of BIT-ARRAY.  If the value of
   RESULT-BIT-ARRAY is NIL or omitted, BIT-NOT returns the inverted bit array as its return
   value; if a bit array, the result is destructively placed in this array; if T, the result is
   destructively placed back into BIT-ARRAY.")

;(setf (arglist FTRUNCATE) (DIVIDEND &OPTIONAL (DIVISOR 1)))
(setf (documentation 'FTRUNCATE 'function)
  "Returns two values :
     Value1 is the integer expressed in float format that is obtained by dropping all fractional
            parts of DIVIDEND if DIVISOR is not specified or all fractional parts of the quotient
            of DIVIDEND divided by DIVISOR if DIVISOR is specified.
     Value2 is the number to add to Value1 (or Value1 times DIVISOR if DIVISOR is specified) in
            order to equal DIVIDEND.")

(setf (documentation '*RANDOM-STATE* 'variable) 
  "Variable whose value is the default random-state data type.")

;(setf (arglist COERCE) (OBJECT RESULT-TYPE))
(setf (documentation 'COERCE 'function)
  "Returns OBJECT converted into data type RESULT-TYPE.  Not are conversion are allowed.") 

;(setf (arglist DECODE-FLOAT) (FLOAT))
(setf (documentation 'DECODE-FLOAT 'function) 
  "Returns three values :
     Value1 is a number of the same type as FLOAT between .5 and 1 (but zero if FLOAT is zero).
            This value, times 2 to a suitable power, equals FLOAT (except perhaps in sign).
     Value2 is an integer whose value is the above mentioned exponent of 2.  Value1 times 2 to
            the Value2 power equals FLOAT (except perhaps in sign).
     Value3 is a float of magnitude 1 with the same sign and type of FLOAT.")

;(setf (arglist INTEGER-DECODE-FLOAT) (FLOAT))
(setf (documentation 'INTEGER-DECODE-FLOAT 'function)
  "Returns three values :
     Value1 is an integer representing the fraction part of FLOAT. This value floated,
            times two to a suitable power, equals FLOAT (except perhaps in sign).  
     Value2 is an integer whose value is the above mentioned exponent of 2.  Value1 times 2 to
            the Value2 power equals FLOAT (except perhaps in sign).
     Value3 is a float of magnitude 1 with the same sign and type as FLOAT.")

;(setf (arglist SVREF) (VECTOR INDEX))
(setf (documentation 'SVREF 'function) 
  "Returns the element of simple vector VECTOR whose index is INDEX.")

;(setf (arglist DO) (&QUOTE &REST BODY))
(setf (documentation 'DO 'function)
  "Macro used to provide a generalized iteration facility.  Syntax of   
   (DO ((var1 init1 step1) ... (varN initN stepN)) (end-test return-result) [(declare...)] body)
   is defined as :  before the first iteration, all init forms <init1> thru <initN> are evaluated
   and these results saved.  The variables <var1> thru <varN> are then bound to the value of <init1>
   thru <initN>, respectively.  After all variables are bound, <end-test> is evaluated.  If
   <end-test> returns true, the forms in <return-result> are evaluated as a PROGN and DO returns
   the result of the last form evaluated.  If <end-test> returns NIL, the forms in <body> are
   evaluated as in a PROGN.  After the last form in <body> is evaluated, the step forms <step1>
   thru <stepN> are evaluated and their return values saved.  These values are assigned in parallel
   to <var1> thru <varN>, respectively.  <end-test> is again evaluated.  This iteration continues
   until <end-test> returns true.  NIL is the default for any form not specified.")

;(setf (arglist COPY-SYMBOL) (SYMBOL &OPTIONAL COPYPROPS))
(setf (documentation 'COPY-SYMBOL 'function) 
  "Returns a new uninterned symbol with the same print name as SYMBOL. If COPYPROPS is true,
   the value, function definition, and properties of SYMBOL are also copied into the new symbol.  
   If COPYPROPS is NIL, SYMBOL is unbound and undefined.")

(setf (documentation 'LEAST-NEGATIVE-SHORT-FLOAT 'variable)
  "Constant whose value is the least negative magnitude representable by the short float type.")

;(setf (arglist CDADR) (X))
(setf (documentation 'CDADR 'function) 
  "Returns the result of (CDR (CAR (CDR X))).")

;(setf (arglist EIGHTH) (LIST))
(setf (documentation 'EIGHTH 'function)
  "Returns the eighth element of LIST.  One-origin numbering is used.")

;(setf (arglist MAKE-TWO-WAY-STREAM) (INPUT-STREAM OUTPUT-STREAM))
(setf (documentation 'MAKE-TWO-WAY-STREAM 'function)
  "Returns a stream which passes input operations to INPUT-STREAM and passes output operations
   to OUTPUT-STREAM.") 

;(setf (arglist NINTH) (LIST))
(setf (documentation 'NINTH 'function) 
  "Returns the ninth element of LIST.  One-origin numbering is used.")

;(setf (arglist DOLIST) ((VAR LIST RESULTFORM) &BODY BODY))
(setf (documentation 'DOLIST 'function) 
  "Iterates BODY over all elements of a list.  BODY is performed once for each element in LIST
   with VAR bound to that element.  When the list is exhausted RESULTFORM is evaluated and
   returned as the result of DOLIST.  NIL is returned if RESULTFORM is not specified.") 
 
;(setf (arglist MERGE-PATHNAMES) (PATHNAME &OPTIONAL DEFAULTS (DEFAULT-VERSION NEWEST)))
(setf (documentation 'MERGE-PATHNAMES 'function)
  "Returns a pathname created by defaulting all components in PATHNAME that are NIL to that
   component's value given in DEFAULTS.  DEFAULT-VERSION is used if the version cannot be
   determined any other way.")  

;(setf (arglist FIND-IF) (PREDICATE SEQUENCE &KEY KEY START END FROM-END))
(setf (documentation 'FIND-IF 'function)
  "Returns the leftmost element of SEQUENCE that satisfies the test predicate PREDICATE.
   Although the entire length of SEQUENCE is the default, a subsequence starting a :START and ending
   at :END may be specified.  If :KEY is non-NIL, it is a function accepting one argument that is
   applied to each element of SEQUENCE before the test is performed.  If :FROM-END is true, the
   rightmost element that satisfies the test is returned.") 

;(setf (arglist TRUNCATE) (DIVIDEND &OPTIONAL (DIVISOR 1)))
(setf (documentation 'TRUNCATE 'function)
  "Returns two values :
     Value1 is the integer obtained by dropping all fractional parts of DIVIDEND if DIVISOR is not
            specified or all fractional parts of the quotient of DIVIDEND divided by DIVISOR if
            DIVISOR is specified.
     Value2 is the number to add to Value1 (or Value1 times DIVISOR if DIVISOR is specified) in
            order to equal DIVIDEND.")

;(setf (arglist CADDR) (X))
(setf (documentation 'CADDR 'function)
  "Returns the result of (CAR (CDR (CDR X))).")

;(setf (arglist FCEILING) (DIVIDEND &OPTIONAL (DIVISOR 1)))
(setf (documentation 'FCEILING 'function)
  "Returns two values :
     Value1 is the integer expressed in float format that is the smallest integer value possible
            that is not smaller than DIVIDEND if DIVISOR is not specified or smaller than the 
            quotient of DIVIDEND divided by DIVISOR if DIVISOR is specified.
     Value2 is the number to add to Value1 (or Value1 times DIVISOR if DIVISOR is specified) in
            order to equal DIVIDEND.")

;(setf (arglist READ-FROM-STRING) (STRING &OPTIONAL (EOF-ERRORP T) EOF-VALUE &KEY &OPTIONAL (START 0)
;                                   END PRESERVE-WHITESPACE))
(setf (documentation 'READ-FROM-STRING 'function)
  "Returns two values :
     Value1 is the Lisp object created by reading the characters of STRING.  Macro characters and
             escaping are all in effect.  Although the default is the entire length of STRING, a
             substring starting at :START and ending at :END may be specified.  If EOF-ERRORP is
             true, an error is signaled when end of file is found; if NIL, EOF-VALUE is returned.
             If :PRESERVE-WHITESPACE is true, whitespace is treated in the same manner as by the
             function READ-PRESERVING-WHITESPACE.
     Value2 is the index into STRING of the first character not read.")

;(setf (arglist MAKE-STRING-OUTPUT-STREAM) (&OPTIONAL STRING START-INDEX EXTRA-ARG))
(setf (documentation 'MAKE-STRING-OUTPUT-STREAM 'function) 
  "Returns a stream that accumulates all output given it.  This output is retrieved by function
   GET-OUTPUT-STREAM-STRING.  The implementation extension arguments allow the existing string
   STRING to store the output and allow a starting position to be specified by START-INDEX.")

(setf (documentation '*PRINT-RADIX* 'variable) 
  "Variable whose value, when true, causes a radix specifier to be printed along with the
   integer or ratio.")

;(setf (arglist CHAR/=) (&REST CHARS))
(setf (documentation 'CHAR/= 'function) 
  "Returns true if all the characters are not equal, considering bits, font, and case; otherwise,
   NIL is returned.")

;(setf (arglist DECLARE) (&QUOTE &REST DECLARATIONS))
(setf (documentation 'DECLARE 'function)
  "Construct used for embedding declarations within executable code.  These declarations are in 
   effect throughout the construct at the head of whose body the DECLARE appears.")

;(setf (arglist FUNCALL) (FN &REST ARGS))
(setf (documentation 'FUNCALL 'function) 
  "Applies the function FN to the arguments ARGS.  The return value of FUNCALL is the value
   returned by FN.")

;(setf (documentation 'FUNCALL 'setf) 
;  "Send a message to OBJECT, with operation OPERATION and ARGUMENTS.  If the object
;is SELF,    a special fast calling sequence is generated.")

;(setf (arglist WITH-INPUT-FROM-STRING) ((STREAM STRING . KEYWORD-ARGS) &BODY BODY))
(setf (documentation 'WITH-INPUT-FROM-STRING 'function) 
  "Evaluates BODY with STREAM bound to a character input stream that supplies characters from the
   value of the form STRING.  WITH-INPUT-FROM-STRING returns the values of the last expression 
   evaluated.  Allowed keywords in KEYWORD-ARGS are :START, :END, and :INDEX.  Although the 
   default length of STRING is the entire string, a subset of STRING starting at :START and ending
   at :END may be specified.  :INDEX is a SETF'able place which stores the index of the first 
   character not read by a read operation, or the length of the string if the entire string was
   read.  :INDEX is updated when WITH-INPUT-FROM-STRING is exited; therefore, it cannot be used
   during the read operation.")   

;(setf (arglist UNION) (LIST1 LIST2 &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'UNION 'function)
  "Returns a list whose elements are a member of at least one of the lists.  :TEST specifies the
   predicate that, when true is returned, causes only one of the two elements that tested as equal
   to be added to the returned list; :TEST-NOT is the predicate that, when NIL is returned, causes
   only one of the two elements that tested as equal to be added to returned list.  If :KEY is
   specified, it is a function applied to each member of each list before the test is performed.")
     
;(setf (arglist MAKUNBOUND) (SYMBOL))
(setf (documentation 'MAKUNBOUND 'function) 
  "Removes the current binding of symbol SYMBOL.  It is now an error to evaluate SYMBOL. 
   MAKUNBOUND returns SYMBOL.")

;(setf (arglist PUSH) (VALUE PLACE))
(setf (documentation 'PUSH 'function)
  "Returns a list whose new first element is VALUE and remaining elements are the elements of
   the list stored at PLACE.  The list stored at PLACE is destructively modified by PUSH.")

;(setf (arglist DO-SYMBOLS) ((VARIABLE PACK RESULT-FORM) &BODY BODY))
(setf (documentation 'DO-SYMBOLS 'function)
  "Iterates over the symbols in a package.  BODY is performed once for each assessible symbol in
   PACK with VARIABLE bound to that symbol.  RESULT-FORM is then evaluated and returned as the
   result of DO-SYMBOLS.  NIL is returned if RESULT-FORM is not specified.")    

;(setf (arglist REVAPPEND) (LIST STARTING-TAIL))
(setf (documentation 'REVAPPEND 'function)
  "Returns a list created by appending list STARTING-TAIL to the reverse of list LIST.") 

(setf (documentation '*ERROR-OUTPUT* 'variable)
  "Global variable whose value is the default stream for unanticipated noninteractive output,
   such as warnings.")

;(setf (arglist RETURN-FROM) (&QUOTE NAME &REST VALS))
(setf (documentation 'RETURN-FROM 'function)
  "Exits a block named NAME and returns the result from the evaluation of VALS.  Forms such as
   DEFUN and BLOCK use named blocks.")

;(setf (arglist VECTOR) (&REST OBJECTS))
(setf (documentation 'VECTOR 'function) 
  "Returns a vector (1-dimensional array) whose elements are OBJECTS.")

;(setf (arglist -) (NUMBER &REST NUMBERS))
(setf (documentation '- 'function) 
  "Returns the value of the first argument minus the remaining arguments, or the negation of 
   NUMBER if only one argument is given.")

(setf (documentation '- 'variable) 
  "Variable whose value is the expression currently being evaluated by read-eval-print loop.")

(setf (documentation 'MOST-NEGATIVE-DOUBLE-FLOAT 'variable)
  "Constant whose value is the greatest negative magnitude representable by the double float type.") 

;(setf (arglist STRING-UPCASE) (STRING &KEY (START 0) END))
(setf (documentation 'STRING-UPCASE 'function)
  "Returns a copy of STRING changing all lowercase characters to uppercase.  Although the entire
   length of STRING is the default, a substring of STRING starting at :START and ending at :END 
   may be specified.  STRING-UPCASE returns the entire string, even if a substring is specified.")

;(setf (arglist NOT) (X))
(setf (documentation 'NOT 'function) 
  "Returns T if X is NIL, otherwise NIL.")

(setf (documentation '// 'variable)
  "Variable whose value is a list of the results obtained from the form evaluated two
   interactions ago by the read-eval-print loop.")

(setf (documentation 'DOUBLE-FLOAT-EPSILON 'variable)
  "Constant whose value is the smallest possible positive number of type double float that, when
   added to some value, will change the magnitude of that value.")

(setf (documentation 'LEAST-NEGATIVE-DOUBLE-FLOAT 'variable)
  "Constant whose value is the least negative magnitude representable by the double float type.")

(setf (documentation 'CALL-ARGUMENTS-LIMIT 'variable) 
  "Constant whose value is the upper exclusive limit on the number of arguments which may be
   passed to a function")

;(setf (arglist POSITION-IF-NOT) (PREDICATE SEQUENCE &KEY KEY START END FROM-END))
(setf (documentation 'POSITION-IF-NOT 'function)
  "Returns the index of the leftmost element of SEQUENCE that does not satisfy the test predicate
   PREDICATE.  Although the entire length of SEQUENCE is the default, a subsequence starting at
   :START and ending at :END may be specified.  If :KEY is non-NIL, it is a function accepting one
   argument that is applied to each element of SEQUENCE before the test is performed.  If :FROM-END
   is true, the index of the rightmost element that does not satisfy the test is returned.") 

;(setf (arglist PRIN1) (OBJECT &OPTIONAL STREAM))
(setf (documentation 'PRIN1 'function) 
  "Prints the printed representation of OBJECT on STREAM with quoting if needed.  PRIN1 returns
   OBJECT.")

;(setf (documentation 'PRIN1 'variable) 
;  NIL)

;(setf (arglist CDDDR) (X))
(setf (documentation 'CDDDR 'function)
  "Returns the result of (CDR (CDR (CDR X))).")

;(setf (arglist COPY-READTABLE) (&OPTIONAL (FROM-READTABLE *READTABLE*) TO-READTABLE))
(setf (documentation 'COPY-READTABLE 'function) 
  "Returns a copy of readtable FROM-READTABLE.  If FROM-READTABLE is omitted, the value of
   *READTABLE* is copied; if NIL, the Common Lisp readtable is copied.  If TO-READTABLE is
   specified, the copied readtable is destructively stored into it.")

;(setf (arglist FILE-POSITION) (FILE-STREAM &OPTIONAL NEW-POSITION))
(setf (documentation 'FILE-POSITION 'function)
  "Can perform two different operations.  If only argument FILE-STREAM is specified, 
   FILE-POSITION returns the current position within that stream, or NIL if it cannot be
   determined.  If NEW-POSITION is also specified, FILE-POSITION will set the current position
   within FILE-STREAM to NEW-POSITION and return T, or NIL if this cannot be done.")

;(setf (arglist MAKE-HASH-TABLE) (&KEY TEST SIZE REHASH-SIZE REHASH-THRESHOLD &EXTENSION COMPARE-FUNCTION
;                                HASH-FUNCTION AREA ACTUAL-SIZE NUMBER-OF-VALUES REHASH-FUNCTION))
(setf (documentation 'MAKE-HASH-TABLE 'function)
  "Creates a hash table and returns the hash table object.  The Common Lisp defined key
   words are:
     :TEST - Function which determines how keys are compared.
     :SIZE - Sets the initial size of the hash table in entries.
     :REHASH-SIZE - Number of entries to increase table by when it becomes full.  If a floating 
                    point number is given, it is greater than one and represents the ratio of
                    the new table size to old table size.
     :REHASH-THRESHOLD - Number of entries to be filled before the table is expanded.  Either 
                         an integer representating number of entries or a floating point number
                         less than one representing percentage size of table can be used.
   Implementation extension keywords are :
     :COMPARE-FUNCTION - Function used when comparing keys versus keys stored in the table.  This
                         is ignored if :test is specified.
     :HASH-FUNCTION - Function used to compute the hash code from a key.
     :AREA - Area where the table is stored.
     :ACTUAL-SIZE - The precise number of entries to provide space for.
     :NUMBER-OF-VALUES - Number of values to associate with each key. 
     :REHASH-FUNCTION - Function used to increase the size of the hash table when necessary.")

;(setf (arglist MINUSP) (X))
(setf (documentation 'MINUSP 'function) 
  "Returns true if X is less than zero, otherwise NIL.")

;(setf (arglist DO*) (&QUOTE &REST BODY))
(setf (documentation 'DO* 'function)
  "Macro used to provide a generalized iteration facility.  Syntax of   
   (DO* ((var1 init1 step1) ... (varN initN stepN)) (end-test return-result) [(declare...)] body)
   is defined as :  before the first iteration, <init1> is evaluated and this value bound to <var1>;
   <init2> is evaluated and this value bound to <var2>, and so on, until all variables are bound to
   a value.  After all variables are bound, <end-test> is evaluated.  If <end-test> returns true,
   the forms in <return-result> are evaluated as a PROGN and DO* returns the result of the last form
   evaluated.  If <end-test> returns NIL, the forms in <body> are evaluated as in a PROGN.  After the
   last form in <body> is evaluated, the step form <step1> is evaluated and this result assigned to
   <var1>, <step2> is evaluated and this result assigned to <var2>, and so on this all variables are
   assigned values.  <end-test> is again evaluated.  This iteration continues until <end-test> returns
   true.  NIL is the default for any form not specified.")

;(setf (arglist CHAR<) (CHAR &REST CHARS))
(setf (documentation 'CHAR< 'function) 
  "Returns true if all the characters are strictly increasing, considering bits, fonts and case;
   otherwise, NIL is returned.")

;(setf (arglist ALPHANUMERICP) (CHAR))
(setf (documentation 'ALPHANUMERICP 'function) 
  "Returns true if CHAR is a letter or digit whose bit attribute is also zero, otherwise NIL.")

;(setf (arglist LOGANDC2) (INTEGER1 INTEGER2))
(setf (documentation 'LOGANDC2 'function)
  "Returns the bitwise-and of INTEGER1 with the complement of INTEGER2.")

;(setf (arglist FLOAT-SIGN) (FLOAT1 &OPTIONAL (FLOAT2 (FLOAT 1 FLOAT1))))
(setf (documentation 'FLOAT-SIGN 'function)
  "Returns a floating-point number whose sign matches FLOAT1 and whose magnitude and type matches
   FLOAT2.  If FLOAT2 is omitted, the magnitude defaults to 1 and the type defaults to the same
   type as FLOAT1.")

;(setf (arglist ODDP) (NUMBER))
(setf (documentation 'ODDP 'function)
  "Returns true if NUMBER is odd, otherwise NIL.") 

;(setf (arglist CADR) (X))
(setf (documentation 'CADR 'function) 
  "Returns the result of (CAR (CDR X)).")

;(setf (arglist SET-MACRO-CHARACTER) (CHAR FUNCTION &OPTIONAL (NON-TERMINATINGP NIL) (RT *READTABLE*)))
(setf (documentation 'SET-MACRO-CHARACTER 'function) 
  "Makes CHAR a macro character in readtable RT such that when READ encounters CHAR the function 
   FUNCTION is invoked.  If NON-TERMINATINGP is true, CHAR can be used inside extended tokens.
   SET-MACRO-CHARACTER returns true.")

;(setf (arglist LDB-TEST) (PPSS WORD))
(setf (documentation 'LDB-TEST 'function)
  "Returns true if the field designated by PPSS in integer WORD is not zero, otherwise NIL.") 

;(setf (arglist NSUBLIS) (ALIST TREE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'NSUBLIS 'function)
  "Returns a new tree structure created by performing multiple substitution in TREE.  ALIST is
   an association list whose CAR is the object to be checked against in TREE and CDR is the 
   replacement object if the test is successful.  :TEST is a predicate whose value, when true,
   causes the substitution to be performed.  If :TEST-NOT is specified, the substitution is
   performed if :TEST-NOT returns NIL.  If specified, :KEY is a function to be applied to all
   components of TREE before the test is performed.  Each element or subtree in TREE is compared
   against each CAR in ALIST.  TREE is destructively modified by NSUBLIS.")

;(setf (arglist PATHNAME-TYPE) (P))
(setf (documentation 'PATHNAME-TYPE 'function)
  "Returns the type component of pathname P, NIL if a type component does not exist.")

;(setf (arglist SUBSEQ) (SEQUENCE START &OPTIONAL END))
(setf (documentation 'SUBSEQ 'function)
  "Returns a subsequence of SEQUENCE starting at START and ending at END.  If END is not
   specified, the entire sequence starting at START is returned.") 

(setf (documentation 'MOST-POSITIVE-FIXNUM 'variable)
  "Constant whose value is the greatest positive magnitude representable by the fixnum type.") 

;(setf (arglist STRING-RIGHT-TRIM) (CHAR-SET STRING))
(setf (documentation 'STRING-RIGHT-TRIM 'function)
  "Returns a substring of STRING with all characters in CHAR-SET removed from the end of STRING.
   CHAR-SET can be a string or list of characters.  Case is presently ignored.")

;(setf (arglist MACRO-FUNCTION) (FSPEC))
(setf (documentation 'MACRO-FUNCTION 'function) 
  "Returns the expander function for FSPEC if FSPEC has a function definition which is a macro;
   otherwise, NIL is returned.")

(setf (documentation 'MULTIPLE-VALUES-LIMIT 'variable) 
  "Constant whose value is the upper exclusive bound on the number of values a function can return.")

(setf (documentation '*PRINT-CASE* 'variable) 
  "Variable whose value controls the case used for printing characters in symbol print names.
   Allowable values are :UPCASE, :DOWNCASE, or :CAPITALIZE.")

(setf (documentation '*DEFAULT-PATHNAME-DEFAULTS* 'variable)
  "Variable whose value is the default pathname components used when no other defaults are specified.") 

;(setf (arglist DEFSETF) (ACCESS-FUNCTION ARG1 &OPTIONAL ARG2 &BODY BODY))
(setf (documentation 'DEFSETF 'function) 
  "Defines a SETF expander for ACCESS-FUNCTION.  DEFSETF has two forms.  The simple form
   (DEFSETF access-function update-function [doc-string]) is defined as :
   <update-function> takes one more argument than <access-function>.  When SETF is given a place
   that is a call on <access-function>, it expands into a call on <update-function> with the new
   update value.  <update-function>, like SETF, must return the update value.  The complex form
   (DEFSETF access-function access-lambda-list newvalue-lambda-list body...) is defined as :
   <access-lambda-list> is the argument forms of <access-function>.  <newvalue-lambda-list> has
   but one value, that being the value to be stored.  <body> must compute a replacement for the
   SETF form and return the replacement value.  When <body> is executed, the args in the
   lambda-lists will not contain the value expression or parts of the form to be set, but rather
   gensymmed variables which SETF may or may not eliminate by substitution.")

;(setf (arglist NUMBERP) (X))
(setf (documentation 'NUMBERP 'function) 
  "Returns true if X is a number, otherwise NIL.")

;(setf (arglist BIT-ANDC1) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-ANDC1 'function)
  "Returns the bit-array generated from a bitwise-and of the complement of BIT-ARRAY1 with
   BIT-ARRAY2.  If RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted,
   the result is returned as a new bit-array; if a bit-array, the result is destructively placed
   into it.")

;(setf (arglist CHAR-FONT) (CHAR))
(setf (documentation 'CHAR-FONT 'function)
  "Returns as a non-negative integer the font attribute of the character object CHAR.")

;(setf (arglist BOTH-CASE-P) (CHAR))
(setf (documentation 'BOTH-CASE-P 'function)
  "Returns true if CHAR has both an uppercase and lowercase form, otherwise NIL.") 

;(setf (arglist SEARCH) (FOR-SEQUENCE-1 IN-SEQUENCE-2 &KEY FROM-END (TEST (FUNCTION EQL)) TEST-NOT
;                        KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'SEARCH 'function)
  "Returns the index in IN-SEQUENCE-2 of the first  subsequence that matches FOR-SEQUENCE-1.  If
   no match is found, NIL is returned.  Although the entire length of both FOR-SEQUENCE-1 and 
   IN-SEQUENCE-2 is the default, a subsequence of FOR-SEQUENCE-1 starting at :START1 and ending
   at :END1 and/or a subsequence of IN-SEQUENCE-2 starting at :START2 and ending at :END2 may be
   specified.  :TEST is a predicate that determines whether or not there is a match.  :TEST-NOT,
   if specified, must return NIL in order to have SEARCH succeed in the match.  If non-NIL, :KEY
   is a function accepting one argument that is applied to each element of each list before the
   test is made.  If :FROM-END is true, the index in IN-SEQUENCE-2 of the last subsequence that
   matches FOR-SEQUENCE-1 is returned.")   

;(setf (arglist CHARACTER) (X))
(setf (documentation 'CHARACTER 'function) 
  "Returns X converted into a character object if possible.")

;(setf (arglist CAAADR) (X))
(setf (documentation 'CAAADR 'function) 
  "Returns the result of (CAR (CAR (CAR (CDR X)))).")

;(setf (arglist STREAM) (CONNECTION &KEY &OPTIONAL (DIRECTION BIDIRECTIONAL) (CHARACTERS T) (ASCII-TRANSLATION NIL)))
;(setf (documentation 'STREAM 'function) 
;  "Return a stream that does i/o to an already established chaos connection.    :ASCII-TRANSLATION
;- if non-NIL, assume the data on the connection is in ASCII     and translate to
;and from the Lisp machine character set as appropriate.    :DIRECTION, :CHARACTERS
;- as in OPEN.  :DIRECTION defaults to ':BIDIRECTIONAL.")

;(setf (arglist ARRAY-ELEMENT-TYPE) (ARRAY))
(setf (documentation 'ARRAY-ELEMENT-TYPE 'function) 
  "Returns a Common Lisp data type describing the objects that can be stored in array ARRAY.")

;(setf (arglist BIT-EQV) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-EQV 'function)
  "Returns the bit-array generated from a bitwise-equivalence of BIT-ARRAY1 with BIT-ARRAY2. 
   If RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted, the result
   is returned as a new bit-array; if a bit-array, the result is destructively placed into it.")

;(setf (arglist ASSOC-IF) (PREDICATE ALIST))
(setf (documentation 'ASSOC-IF 'function)
  "Returns the first association pair in association list ALIST whose CAR satisfies test predicate
   PREDICATE.  If no match is found, ASSOC-IF returns NIL.")

;(setf (arglist BUTLAST) (LIST &OPTIONAL (N 1)))
(setf (documentation 'BUTLAST 'function)
  "Returns a list with the same elements as LIST but with the last N elements deleted.  If the
   number of elements in LIST is less than or equal to N, NIL is returned.")

;(setf (arglist AND) (&QUOTE &REST EXPRESSIONS))
(setf (documentation 'AND 'function)
  "Evaluates the forms one at a time from left to right until one returns NIL.  Evaluation then
   stops and NIL is returned.  If the forms are exhausted before NIL is returned, AND returns
   the result of the last form evaluated.  AND will return multiple values only if they are the
   result of evaluating the last form in the sequence.")

;(setf (arglist CDDR) (X))
(setf (documentation 'CDDR 'function) 
  "Returns the result of (CDR (CDR X)).")

;(setf (arglist MAPL) (FCT &REST LISTS))
(setf (documentation 'MAPL 'function) 
  "Applies function FCT to each list in LISTS, then successivley to the CDR of each of the lists, 
   terminating when the end of the shortest list has been reached.  Function FCT must take as many
   arguments as there are lists in LISTS.  MAPL does not retain the results, being called mainly
   for side effects, and returns the first list in LISTS as its return value.")

;(setf (arglist EVALHOOK) (FORM *EVALHOOK* *APPLYHOOK* &OPTIONAL ENVIRONMENT))
(setf (documentation 'EVALHOOK 'function) 
  "Evaluates FORM, using specified *EVALHOOK* and *APPLYHOOK* except at the top level, in the
   in the lexical environment defined by ENVIRONMENT.  The value returned by EVALHOOK is the
   value returned by the hook function which is bound to *EVALHOOK*.")

(setf (documentation 'EVALHOOK 'variable)
  "Variable whose value, when non-NIL, should be a function accepting the two arguments of
   form and environment that is invoked whenever an evaluation is done either implicitly
   or through an explicit call to function EVAL.")

;(setf (arglist LENGTH) (LIST-OR-ARRAY))
(setf (documentation 'LENGTH 'function) 
  "Returns the number of elements in sequence LIST-OR-ARRAY.  If LIST-OR-ARRAY is an array,
   this is the active length of the array, which is either the value of the fill-pointer if one
   exists, or the number of elements in the array if a fill-pointer does not exist.")

;(setf (arglist MAPCAR) (FCT &REST LISTS))
(setf (documentation 'MAPCAR 'function)
  "Applies function FCT to the successively elements of the input lists LISTS.  Function MAPCAR
   must take as many arguments as there are lists provided and terminates when the shortest list
   is exhausted.  The return value of MAPCAR is a list of the results from the successive calls
   to FCT.")  

;(setf (arglist INTEGER-LENGTH) (INTEGER))
(setf (documentation 'INTEGER-LENGTH 'function)
  "Returns the minimum number of bits excluding the sign bit needed to represent the integer
   INTEGER in two's complement notation.") 

;(setf (arglist POSITION) (ITEM SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END FROM-END))
(setf (documentation 'POSITION 'function)
  "Returns the index in SEQUENCE of the first element which matches ITEM.  Although the entire
   length of SEQUENCE is the default, a subsequence starting at :START and ending at :END may be
   specified.  :TEST is a predicate which determines if ITEM matches an element in SEQUENCE.  If
   :TEST-NOT is specified, it is a predicate which must return NIL in order to match ITEM with an
   element in SEQUENCE.  If :KEY is non-NIL, it is a function accepting one argument that is applied
   to each element of SEQUENCE before the test is performed.  If :FROM-END is true, the index of
   the rightmost element in SEQUENCE that matches ITEM is returned.") 

;(setf (arglist COMPLEXP) (X))
(setf (documentation 'COMPLEXP 'function)
  "Returns true if X is a complex number, otherwise NIL.") 

;(setf (arglist ASIN) (NUMBER))
(setf (documentation 'ASIN 'function)
  "Returns the angle in radians whose sine is NUMBER.  The argument NUMBER may be complex.") 

;(setf (arglist PSETQ) (&REST REST))
(setf (documentation 'PSETQ 'function)
  "Assigns values to symbols.  REST is a list of alternating symbol-value pairs.
   The PSETQ syntax of (PSETQ var1 exp1 var2 exp2 ... varN expN) is defined as : evaluate all
   expressions exp1, exp2, ..., expN and assign these results to var1, var2, ..., varN,
   respectively.  The return value of PSETQ is NIL.")

;(setf (arglist BIGP) (X))
(setf (documentation 'BIGP 'function) 
  "Returns true if X is a bignum, otherwise NIL.")

;(setf (arglist INTERSECTION) (LIST1 LIST2 &KEY (TEST (FUNCTION EQL)) KEY TEST-NOT))
(setf (documentation 'INTERSECTION 'function)
  "Returns a list consisting of all elements that are members of both LIST1 and LIST2.  :TEST
   specifies the test that returns true to denote intersection; :TEST-NOT specifies the test that
   returns NIL to denote intersection.  If non-NIL, :KEY specifies a function to be applied to each
   element of each list before the test is performed.")

;(setf (arglist SECOND) (LIST))
(setf (documentation 'SECOND 'function)
  "Returns the second element of LIST.  One-origin numbering is used.")

;(setf (arglist CASE) (TEST-OBJECT &BODY CLAUSES))
(setf (documentation 'CASE 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (CASE TEST-OBJECT (test1 body1) ... (testN bodyN)) is defined as :  the (test# body#)
   construct is called a clause.  TEST-OBJECT is evaluated.  <test1> is evaluated and compared
   against TEST-OBJECT; if they are EQL, all forms in <body1> are evaluated sequentially and the
   value of the last expression in <body1> is returned by CASE.  If <test1> does not match,
   <test2> is evaluated and compared, and so on, until either one <test#> matches TEST-OBJECT or
   all clauses are exhausted.  CASE returns NIL if no clause is selected.  An OTHERWISE or T as
   the <testN> matches any TEST-OBJECT.")

;(setf (arglist PARSE-NAMESTRING) (THING &OPTIONAL WITH-RESPECT-TO (DEFAULTS *DEFAULT-PATHNAME-DEFAULTS*)
;                                  &KEY (START 0) END JUNK-ALLOWED))
(setf (documentation 'PARSE-NAMESTRING 'function)
  "Returns two values :
     Value1 is the pathname parsed from THING.  Although the entire length of THING is the default
            a substring starting at :START and ending at :END may be specified.
     Value2 is the index into THING of the delimiter that terminated the parse or the index 
            beyond the substring if the parse terminated at the end of the substring.") 

;(setf (arglist UPPER-CASE-P) (CHAR))
(setf (documentation 'UPPER-CASE-P 'function) 
  "Returns true if CHAR is an upper case character with no meta bits, otherwise NIL.")

;(setf (arglist PROG*) (&REST BODY))
(setf (documentation 'PROG* 'function)
  "Macro allowing local variable bindings, return statements, and GO statements.  Syntax of  
   (PROG* [name] ((var1 exp1) ... (varN expN)) [(declare...)] body) is defined as :
   <name>, if specified, is a symbol which names the PROG* construct and allows the use of the 
   (RETURN-FROM <name>) statement.  <exp1> is evaluated and this value bound to <var1>; <exp2>
   is evaluated and this value bound to <var2>, and so on till all variables are bound to a value.
   If an expression is not specified, NIL is used.  These binding are in force only over the scope
   of the PROG*.  All forms in <body>, which may include GO and RETURN statements, are then
   evaluated with these binding in force; PROG* returns the result of the last form evaluated.")

;(setf (arglist CADADR) (X))
(setf (documentation 'CADADR 'function) 
  "Returns the result of (CAR (CDR (CAR (CDR X)))).")

;(setf (arglist WRITE-BYTE) (BYTE &OPTIONAL STREAM))
(setf (documentation 'WRITE-BYTE 'function) 
  "Outputs BYTE to STREAM and returns BYTE.")

;(setf (arglist ASINH) (NUMBER))
(setf (documentation 'ASINH 'function)
  "Returns the angle in radians as a floating-point or complex number whose hyperbolic sine
   is NUMBER.")

;(setf (arglist STRING<) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING< 'function)
  "Returns the index into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test less than.  Both case and font are considered.  If the test is never
   satisfied, NIL is returned.  If the strings match, the length of the shortest string is
   returned.  Although the default string length of both STRING1 and STRING2 is the entire
   length, a substring of STRING1 starting at :START1 and ending at :END1 and/or and substring
   of STRING2 starting at :START2 and ending at :END2 may be specified.")

;(setf (arglist CAADDR) (X))
(setf (documentation 'CAADDR 'function) 
  "Returns the result of (CAR (CAR (CDR (CDR X)))).")

;(setf (arglist INT-CHAR) (FIXNUM))
(setf (documentation 'INT-CHAR 'function) 
  "Returns the character object whose integer representation is FIXNUM; if this cannot be done,
   NIL is returned.")

;(setf (arglist LOGCOUNT) (INTEGER))
(setf (documentation 'LOGCOUNT 'function)
  "Returns a non-negative integer which is a count of the number of 1 bits if INTEGER is 
   positive or the number of 0 bits if INTEGER is negative.")

;(setf (arglist ASSOC-IF-NOT) (PREDICATE ALIST))
(setf (documentation 'ASSOC-IF-NOT 'function)
  "Returns the first association pair in association list ALIST whose CAR does not satisfy test
   predicate PREDICATE.  If the test is not satisfied, ASSOC-IF-NOT returns NIL.")

;(setf (arglist ATANH) (NUMBER))
(setf (documentation 'ATANH 'function)
  "Returns the angle in radians as a floating-point or complex number whose hyperbolic tangent
   is NUMBER.")

;(setf (arglist DO-ALL-SYMBOLS) ((VARIABLE RESULT-FORM) &BODY BODY))
(setf (documentation 'DO-ALL-SYMBOLS 'function)
  "Iterates over the symbols in all packages.  BODY is performed once for every symbol present
   with VARIABLE bound to that symbol.  RESULT-FORM is then evaluated and returned as the
   result of DO-ALL-SYMBOLS.  NIL is returned if RESULT-FORM is not specified.  Since symbols
   may be in more than one package, some symbols may be processed more than once.")    

;(setf (arglist ASSOC) (ITEM ALIST &KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'ASSOC 'function)
  "Returns the first cons in association list ALIST whose CAR matches ITEM according to test
   predicate :TEST.  If no match is found, ASSOC returns NIL.  If :TEST-NOT is specified, ASSOC
   returns the first cons whose CAR does not match ITEM using predicate :TEST-NOT.") 

;(setf (arglist EQUALP) (X Y))
(setf (documentation 'EQUALP 'function)
  "Returns true if X is equal to Y.  Equal to is defined to be CHAR-EQUAL if they are 
   characters, = if they are numbers, EQ if they are symbols, and STRING-EQUAL if they 
   are strings; otherwise, NIL is returned.") 

;(setf (arglist BIT-AND) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-AND 'function)
  "Returns the bit-array generated from a bitwise-and of BIT-ARRAY1 with BIT-ARRAY2.
   If RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted, the result
   is returned as a new bit-array; if a bit-array, the result is destructively placed into it.")

;(setf (arglist REM) (DIVIDEND MODULUS))
(setf (documentation 'REM 'function)
  "Returns the remainder after dividing DIVIDEND by MODULUS.  The sign of a non-zero return value
   is the same as the sign of DIVIDEND.") 

(setf (documentation 'MOST-NEGATIVE-SHORT-FLOAT 'variable)
  "Constant whose value is the greatest negative magnitude representable by the short float type.")

;(setf (arglist EQL) (X Y))
(setf (documentation 'EQL 'function)
  "Returns true if X and Y are of the same type and are numerically equal if numbers or represent
   the same character if characters.  All other objects must be EQ to also be EQL.  NIL is
   returned if the test fails.")  

;(setf (arglist RANDOM-STATE-P) (OBJECT))
(setf (documentation 'RANDOM-STATE-P 'function)
  "Returns true if OBJECT is of type random-state, otherwise NIL.") 

;(setf (arglist SUBLIS) (ALIST TREE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'SUBLIS 'function)
  "Returns a new tree structure created by performing multiple substitution in TREE.  ALIST is
   an association list whose CAR is the object to be checked against in TREE and CDR is the 
   replacement object if the test is successful.  :TEST is a predicate whose value, when true,
   causes the substitution to be performed.  If :TEST-NOT is specified, the substitution is
   performed if :TEST-NOT returns NIL.  If specified, :KEY is a function to be applied to all
   components of TREE before the test is performed.  Each element or subtree in TREE is compared
   against each CAR in ALIST.")

;(setf (arglist CCASE) (TEST-OBJECT &BODY CLAUSES))
(setf (documentation 'CCASE 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (CCASE TEST-OBJECT (test1 body1) ... (testN bodyN)) is defined as :  the (test# body#)
   construct is called a clause.  TEST-OBJECT is evaluated.  <test1> is evaluated and compared
   against TEST-OBJECT; if they are EQL, all forms in <body1> are evaluated sequentially and the
   value of the last expression in <body1> is returned by CCASE.  If <test1> does not match,
   <test2> is evaluated and compared, and so on, until either one <test#> matches TEST-OBJECT or all
   clauses are exhausted.  If no <test#> matches TEST-OBJECT, an error is signaled with a message
   constructed from the clauses.  Continuing from the error causes CCASE to accept a new value for
   TEST-OBJECT and redo the clause tests again.  An OTHERWISE or T as the <test#> is not permitted.")

;(setf (arglist SET-CHAR-BIT) (CHAR BIT-NAME NEW-VALUE))
(setf (documentation 'SET-CHAR-BIT 'function) 
  "Returns a character equal to CHAR except that the bit BIT-NAME has been set if NEW-VALUE is
   true or reset if NEW-VALUE is NIL.  BIT-NAME accepts the values :CONTROL, :META, :SUPER,
   or :HYPER.")

;(setf (arglist FIND-ALL-SYMBOLS) (STRING))
(setf (documentation 'FIND-ALL-SYMBOLS 'function) 
  "Returns a list of all symbols whose print names match STRING.  This search, which is case 
   sensitive, is made in every package in the Lisp system.")

(setf (documentation 'INTERNAL-TIME-UNITS-PER-SECOND 'variable) 
  "Constant whose value is an integer and represents the number of internal time units per second.")

(setf (documentation 'MOST-NEGATIVE-LONG-FLOAT 'variable)
  "Constant whose value is the greatest negative magnitude representable by the long float type.")

;(setf (arglist BYTE-SIZE) (BYTE-SPECIFIER))
(setf (documentation 'BYTE-SIZE 'function)
  "Returns as an integer the size of the byte specified by BYTE-SPECIFIER.")

;(setf (arglist RASSOC-IF-NOT) (PREDICATE ALIST))
(setf (documentation 'RASSOC-IF-NOT 'function)
  "Returns the first association pair in association list ALIST whose CDR does not satisfy test
   predicate PREDICATE.  If the test is not satisfied, RASSOC-IF-NOT returns NIL.")

;(setf (arglist VALUES-LIST) (LIST))
(setf (documentation 'VALUES-LIST 'function) 
  "Returns all the elements of LIST as multiple values.") 

;(setf (arglist TRUENAME) (OBJECT))
(setf (documentation 'TRUENAME 'function)
  "Returns the truename of OBJECT.  A truename is that sequence of characters that represents the
   file to the file system that created it.") 

;(setf (arglist SLEEP) (INTERVAL-IN-SECONDS &OPTIONAL (WHOSTATE Sleep)))
(setf (documentation 'SLEEP 'function) 
  "Halts execution for approximately INTERVAL-IN-SECONDS seconds then resumes execution.")

;(setf (arglist CDAAAR) (X))
(setf (documentation 'CDAAAR 'function) 
  "Returns the result of (CDR (CAR (CAR (CAR X)))).")

;(setf (arglist CATCH) (TAG &QUOTE &REST BODY))
(setf (documentation 'CATCH 'function) 
  "Construct used for non-local exits.  TAG is evaluated to produce an object that names the
   CATCH.  BODY is then evaluated as an implicit PROGN with the results of the last form
   evaluated returned as the values of CATCH.  However, if during this evaluation of BODY a
   THROW is executed with the same tag as CATCH, evaluation of the forms in BODY immediately
   halts and CATCH returns the values specified by the THROW.")

;(setf (arglist ADJUST-ARRAY) (ARRAY NEW-DIMENSIONS &REST KEYARGS &KEY &OPTIONAL ELEMENT-TYPE
;                              (INITIAL-ELEMENT NIL INITIAL-ELEMENT-P) (INITIAL-CONTENTS NIL INITIAL-CONTENTS-P)
;                              FILL-POINTER DISPLACED-TO DISPLACED-INDEX-OFFSET))
(setf (documentation 'ADJUST-ARRAY 'function) 
  "Alters dimensions, contents, and/or displacement of ARRAY then returns ARRAY.
   The dimensions are altered to those in the list NEW-DIMENSIONS.  The keywords are
   defined as :
     :ELEMENT-TYPE - Data type of elements allowed in the array.  This is merely an error check
                     and cannot change the element type of ARRAY.
     :INITIAL-ELEMENT - Value to be stored in each new element of ARRAY that would not be
                        in bounds in the old array.
     :INITIAL-CONTENTS - Value is a sequence of sequences used to initial the array.  Top level
                         of sequence correspondings to the most slowly varying subscript.
     :FILL-POINTER - If non-NIL, this value resets the fill pointer of ARRAY.
     :DISPLACED-TO - If non-NIL, this symbol is an array that this new array will overlap.
     :DISPLACED-INDEX-OFFSET - When :DISPLACED-TO is specified, this nonnegative value specifies 
                               the element that corresponds to element 0 of the new array.")

;(setf (arglist BIT-NAND) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-NAND 'function)
  "Returns the bit-array generated from a bitwise-nand of BIT-ARRAY1 with BIT-ARRAY2.
   If RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted, the result
   is returned as a new bit-array; if a bit-array, the result is destructively placed into it.")

;(setf (arglist CADDDR) (X))
(setf (documentation 'CADDDR 'function) 
  "Returns the result of (CAR (CDR (CDR (CDR X)))).")

;(setf (arglist DELETE-IF) (PREDICATE SEQUENCE &KEY KEY START END COUNT FROM-END))
(setf (documentation 'DELETE-IF 'function)
  "Returns a sequence where those elements in sequence SEQUENCE that satisfy test PREDICATE have
   been deleted.  SEQUENCE is destructively modified by this operation.  Although the entire length
   of SEQUENCE is the default, a subsequence starting at :START and ending at :END may be specified.
   If non-NIL, :KEY is a funcion applied to each element of SEQUENCE before the test is performed.
   :COUNT is an integer specifying how many leftmost elements are to be deleted if more than :COUNT
   number satisfy the test.  :FROM-END is only meaningful if :COUNT is specified and, if true, causes
   the rightmost :COUNT elements that satisfy the test to be deleted.")

;(setf (arglist CONCATENATE) (RESULT-TYPE &REST SEQUENCES))
(setf (documentation 'CONCATENATE 'function) 
  "Returns a sequence of type RESULT-TYPE consisting of all the elements of SEQUENCES
   concatenated together.  RESULT-TYPE must be a valid sequence type such as LIST or VECTOR.")

;(setf (arglist GET-OUTPUT-STREAM-STRING) (STREAM))
(setf (documentation 'GET-OUTPUT-STREAM-STRING 'function) 
  "Returns as a string all characters output to stream STREAM thus far, then resets the stream.
   STREAM must be a stream created by MAKE-OUTPUT-STRING-STREAM.")

;(setf (arglist IMAGPART) (X))
(setf (documentation 'IMAGPART 'function) 
  "Returns the imaginary part of complex number X, or 0 if given a non-complex number.")

;(setf (arglist MAKE-STRING-INPUT-STREAM) (STRING &OPTIONAL (START 0) END))
(setf (documentation 'MAKE-STRING-INPUT-STREAM 'function) 
  "Returns a stream that can be used to read the characters of STRING.  Although the entire
   length of STRING is the default a substring starting at START and ending at END may be 
   specified.  End of file is signaled when either END is passed or end of string is found.")

;(setf (arglist NRECONC) (LIST TAIL))
(setf (documentation 'NRECONC 'function) 
  "Returns a list created by concatenating list TAIL to the reverse of list LIST.  LIST is
   destructively modified by NRECONC.")

;(setf (arglist PUTPROP) (SYMBOL-OR-PLIST VALUE PROPERTY))
(setf (documentation 'PUTPROP 'function) 
  "Updates the property value to VALUE in property list PROPERTY whose indicator is equal to 
   SYMBOL-OR-PLIST.  PUTPROP returns VALUE.  Common Lisp considers this function to be obsolete;
   preferred method is (SETF (GET SYMBOL-OR-PLIST PROPERTY) VALUE).")

;(setf (arglist DEFINE-MODIFY-MACRO) (NAME ADDITIONAL-ARGLIST ACTION-FUNCTION &OPTIONAL DOC-STRING))
(setf (documentation 'DEFINE-MODIFY-MACRO 'function) 
  "Defines a macro which modifies the value of its first argument.  NAME is the name given
   to the defined macro.  ADDITIONAL-ARGLIST is a lambda list of arguments (except the first) 
   accepted by the new macro; the first argument is the place position that will receive the
   new value.  ACTION-FUNCTION is the function applied to the original place position contents
   and the arguments to produce the new value which is stored back into the place of the 
   original value.  DOC-STRING is optional and defines the documentation string describing this
   macro.")

;(setf (arglist COMPILEDP) (FUNCTION &OPTIONAL DONT-UNENCAPSULATE))
(setf (documentation 'COMPILEDP 'function)
  "Returns true if FUNCTION is a compiled function, otherwise NIL.  If the original interpreted
   definition is known it is the return value.") 

;(setf (arglist EVENP) (NUMBER))
(setf (documentation 'EVENP 'function) 
  "Returns true if NUMBER is even, otherwise NIL.")

;(setf (arglist TYPE-OF) (OBJECT))
(setf (documentation 'TYPE-OF 'function) 
  "Returns a type specifier describing the type that OBJECT is a member of.")

;(setf (arglist LISTP) (X))
(setf (documentation 'LISTP 'function) 
  "Returns true if X is a cons or the empty list, otherwise NIL.")

;(setf (arglist WRITE-USER-OPTIONS) (ALIST STREAM))
;(setf (documentation 'WRITE-USER-OPTIONS 'function) 
;  "Write forms on STREAM to set all non-default variables on ALIST to their current
;values. That is, for each variable on ALIST whose current value is not its default,
;a LOGIN-SETQ form is output to STREAM which records the variable's current value.
;ALIST may be one list, or a list of alists. If ALIST is NIL, it defaults to *all-user-option-alists*")

;(setf (arglist NSET-DIFFERENCE) (LIST1 LIST2 &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'NSET-DIFFERENCE 'function)
  "Returns a list consisting of every element in LIST1 that is not in LIST2.  List1 may be 
   destructively modified.  An element is considered to be in a list if the test :TEST returns 
   true or the test :TEST-NOT returns NIL.  If non-NIL, :KEY specifies a function to be applied
   to each element of each list before the test is performed.") 

;(setf (arglist DRIBBLE) (&OPTIONAL FILENAME))
(setf (documentation 'DRIBBLE 'function) 
  "Opens a file named FILENAME and sends a record of input/output interactions to it.  This is
   done by rebinding  *STANDARD-OUTPUT* and *STANDARD-INPUT*.   With no argument, DRIBBLE exits
   a previously entered DRIBBLE and closes the dribble file.")

(setf (documentation 'BOOLE-EQV 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-equivalence of BOOLE's
   second argument with BOOLE's third argument.")

;(setf (arglist DELETE) (ITEM SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END COUNT FROM-END))
(setf (documentation 'DELETE 'function)
  "Returns a list where those elements in SEQUENCE that match ITEM have been deleted.  SEQUENCE
   can be destructively modified.  An element matches ITEM if the test :TEST returns a true value
   or the test :TEST-NOT returns NIL.  Although the entire length of SEQUENCE is the default, a
   subsequence starting at :START and ending at :END may be specified.  If non-NIL, :KEY specifies
   a function to be applied to each element of SEQUENCE before the test is performed.  :COUNT is
   an integer specifying how many leftmost elements are to be deleted if more than :COUNT number
   match ITEM.  :FROM-END is only meaningful if :COUNT is specified and, if true, causes the rightmost
   :COUNT elements that match ITEM to be deleted.")

;(setf (arglist STRING-DOWNCASE) (STRING &KEY (START 0) END))
(setf (documentation 'STRING-DOWNCASE 'function)
  "Returns a copy of STRING changing all uppercase characters to lowercase.  Although the entire
   length of STRING is the default, a substring of STRING starting at :START and ending at :END 
   may be specified.  STRING-DOWNCASE returns the entire string, even if a substring is specified.")

;(setf (arglist WITH-OPEN-FILE) ((STREAM FILENAME . OPTIONS) &BODY BODY))
(setf (documentation 'WITH-OPEN-FILE 'function)
  "Evaluates BODY with the variable STREAM bound to a file named by the value of FILENAME.
   When control leaves WITH-OPEN-FILE the stream is closed;  if the exit is abnormal, an output
   file is treated as if it never existed.  OPTIONS in WITH-OPEN-FILE are identical to the
   OPEN options, the Common Lisp defined keywords being :
      :DIRECTION - Defines whether the stream handles input, output, or both.  Acceptable values
                   are :input, :output, :io, :probe, or the implementation defined values
                   of :probe-link or :probe-directory.
      :ELEMENT-TYPE - The type of transaction unit the stream supports.  A subtype of character
                      or integer is acceptable.
      :IF-EXISTS - Specifies the action to take if :DIRECTION is :output or :io and the file
                   already exists.  It is ignored in all other cases.  Acceptable values are
                   :error, :new-version, :rename, :rename-and-delete, :overwrite, :append, NIL,
                   :supersede, or the implementation defined value :truncate.
      :IF-DOES-NOT-EXIST - Specifies the action to take if the file does not exist.  Acceptable
                           values are :error, :create, or NIL.
   The implementation extension keywords are :
      :ERROR - If NIL, return and error object rather that signaling an error.
      :CHARACTERS - True means transfer character data.
      :BYTE-SIZE - Specify byte size; defaults according to CHARACTERS.
      :DELETED - If true, it is acceptable to open deleted but not expunged files.
      :TEMPORARY - If true, open the file and mark it as temporary.
      :ESTIMATED-LENGTH - Estimated length of file to be written in bytes.
      :PRESERVE-DATES - If true, do not alter the file's read or write dates.
      :RAW - If true, disable the server's character set translation.
      :SUPER-IMAGE - If true, disable special treatment of #\\rubout in pdp-10 servers.
      :SUBMIT - If true, submit this file as a batch job.
      :FLAVOR - Acceptable values are NIL, :directory, or :link.
      :LINK-TO - Specifies the link target when a file is created with flavor :link.") 

;(setf (arglist DELETE-IF-NOT) (PREDICATE SEQUENCE &KEY KEY START END COUNT FROM-END))
(setf (documentation 'DELETE-IF-NOT 'function)
  "Returns a sequence where those elements in sequence SEQUENCE that do not satisfy test PREDICATE
   have been deleted.  SEQUENCE is destructively modified by this operation.  Although the entire
   length of SEQUENCE is the default, a subsequence starting at :START and ending at :END may be
   specified.  If non-NIL, :KEY is a funcion applied to each element of SEQUENCE before the test is
   performed.  :COUNT is an integer specifying how many leftmost elements are to be deleted if more
   than :COUNT number do not satisfy the test.  :FROM-END is only meaningful if :COUNT is specified
   and, if true,  causes the rightmost :COUNT elements that do not satisfy the test to be deleted.")

;(setf (arglist WRITE-STRING) (STRING &OPTIONAL STREAM &KEY &OPTIONAL (START 0) END))
(setf (documentation 'WRITE-STRING 'function) 
  "Outputs STRING to the stream STREAM.  Although the entire length of STRING is the default, a
   substring of STRING starting at :START and ending at :END may be specified.  WRITE-STRING returns
   the entire string even if a substring was specified.")

;(setf (arglist POSITION-IF) (PREDICATE SEQUENCE &KEY KEY START END FROM-END))
(setf (documentation 'POSITION-IF 'function)
  "Returns the index of the leftmost element of sequence SEQUENCE that satisfies test PREDICATE.
   Although the entire length of SEQUENCE is the default, a subsequence of SEQUENCE starting at
   :START and ending at :END may be specified.  If non-NIL, :KEY is a function applied to each
   element of SEQUENCE before the test is performed.  If true, :FROM-END specifies that the
   index of the rightmost element that satisfies test PREDICATE be returned.") 

;(setf (arglist SIMPLE-VECTOR-P) (OBJECT))
(setf (documentation 'SIMPLE-VECTOR-P 'function)
  "Returns true if OBJECT is a simple vector, otherwise NIL.") 

;(setf (arglist ARRAY-DIMENSIONS) (ARRAY))
(setf (documentation 'ARRAY-DIMENSIONS 'function) 
  "Returns a list whose elements are the dimensions of array ARRAY.")

;(setf (arglist LOWER-CASE-P) (CHAR))
(setf (documentation 'LOWER-CASE-P 'function) 
  "Returns true if CHAR is an lowercase character with no meta bits, otherwise NIL.")

;(setf (arglist CDDAAR) (X))
(setf (documentation 'CDDAAR 'function) 
  "Returns the result of (CDR (CDR (CAR (CAR X)))).")

;(setf (arglist WRITE-TO-STRING) (OBJECT &KEY &OPTIONAL ESCAPE RADIX BASE CIRCLE PRETTY LEVEL LENGTH CASE GENSYM ARRAY))
(setf (documentation 'WRITE-TO-STRING 'function) 
  "Returns OBJECT, which has been made into a string identical to that which would be printed
   by the function WRITE.  The allowed options are identical to those of WRITE and default to 
   the values of the global variables *print-escape*, *print-radix*, *print-base*, *print-circle*,
   *print-pretty*, *print-level*, *print-length*, *print-case*, *print-gensym*, and 
   *print-array*, respectively.")

;(setf (arglist SYMBOL-FUNCTION) (SYMBOL))
(setf (documentation 'SYMBOL-FUNCTION 'function) 
  "Returns the function definition of SYMBOL, otherwise signal an error if one does not exit.")

;(setf (arglist PATHNAME-DEVICE) (P))
(setf (documentation 'PATHNAME-DEVICE 'function)
  "Returns the device component of pathname P, NIL if a device component does not exist.")

;(setf (arglist REQUIRE) (MODULE &OPTIONAL PATHNAMES))
(setf (documentation 'REQUIRE 'function) 
  "Tests whether MODULE is already present in the system and loads it if it is not.  If PATHNAMES
   is specified, it should be a pathname or a list of pathnames to be loaded in the order given.
   If PATHNAMES is NIL or not specified, an implementation dependent procedure is invoked which 
   assumes MODULE to be a defined system and therefore performs a MAKE-SYSTEM on it.")

;(setf (arglist VECTORP) (OBJECT))
(setf (documentation 'VECTORP 'function) 
  "Returns true if OBJECT is a vector, otherwise NIL.")

;(setf (arglist FORCE-OUTPUT) (&OPTIONAL STREAM))
(setf (documentation 'FORCE-OUTPUT 'function) 
  "Forces buffered output on STREAM to be transmitted immediately.  FORCE-OUTPUT does not wait
   for completion or acknowledgment but returns NIL immediately.")

;(setf (arglist CDADAR) (X))
(setf (documentation 'CDADAR 'function) 
  "Returns the result of (CDR (CAR (CDR (CAR X)))).")

;(setf (arglist FOURTH) (LIST))
(setf (documentation 'FOURTH 'function)
  "Returns the fourth element of LIST.  One-origin numbering is used.")

;(setf (arglist STRING-CAPITALIZE) (STRING &KEY (START 0) END SPACES))
(setf (documentation 'STRING-CAPITALIZE 'function)
  "Returns a string where the first character of each word in STRING is uppercase and all
   remaining characters are lowercase.  Any non-alphanumeric character delineates a word.
   Although the entire length of STRING is the default, a substring starting at :START and ending
   at :END may be specified.  If true, :SPACES causes all hyphens to be converted to spaces.
   STRING-CAPITALIZE returns the entire string, even if a substring is specified.")

(setf (documentation 'BOOLE-1 'variable)
  "Constant used as first argument to function BOOLE to return BOOLE's second argument.")

;(setf (arglist COMPLEX) (REALPART &OPTIONAL (IMAGPART 0)))
(setf (documentation 'COMPLEX 'function) 
  "Returns a complex number with real part REALPART and imaginary part IMAGPART.  Note that if 
   REALPART is a rational number and IMAGPART is 0 or unspecified, a rational number is returned.")

;(setf (arglist CONSP) (OBJECT))
(setf (documentation 'CONSP 'function) 
  "Returns true if OBJECT is a cons, otherwise NIL.")

;(setf (arglist APROPOS-LIST) (STRING &OPTIONAL PACKAGE))
(setf (documentation 'APROPOS-LIST 'function) 
  "Returns a list of symbols available in package PACKAGE whose print names contain STRING.  If
   PACKAGE is NIL or not specified, all packages are searched.")

;(setf (arglist CHARACTERP) (OBJECT))
(setf (documentation 'CHARACTERP 'function) 
  "Returns true if OBJECT is a character, otherwise NIL.")

;(setf (arglist MASK-FIELD) (PPSS FIXNUM))
(setf (documentation 'MASK-FIELD 'function)
  "Returns as a non negative integer the byte specified by PPSS from integer FIXNUM.  The result
   returned has the byte extracted in the position specified by PPSS, not at position 0.")  

;(setf (arglist NSTRING-CAPITALIZE) (STRING &KEY (START 0) END SPACES))
(setf (documentation 'NSTRING-CAPITALIZE 'function)
  "Returns a string where the first character of each word in STRING is uppercase and all
   remaining characters are lowercase.  STRING is destructively modified.  Any non-alphanumeric
   character delineates a word.  Although the entire length of STRING is the default, a substring
   starting at :START and ending at :END may be specified.  If true, :SPACES causes all hyphens
   to be converted to spaces.  NSTRING-CAPITALIZE returns the entire modified string, even if a 
   substring is specified.") 
 
;(setf (arglist DELETE-DUPLICATES) (SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END FROM-END))
(setf (documentation 'DELETE-DUPLICATES 'function) 
  "Returns a sequence where all duplicates in sequence SEQUENCE have been removed.  The first
   encountered duplicate of a duplicate pair is removed.  SEQUENCE is destructively modified.
   Although the entire lenght of SEQUENCE is the default a subsequence starting at :START and
   ending at :END may be specified.  :TEST is a predicate which must return true in order for two
   elements to be regarded as duplicates; :TEST-NOT is a predicate which must return NIL in order
   for two elements to be regarded as duplicates.  If non-NIL, :KEY is a function applied to each
   element of SEQUENCE before the test is applied.  If :FROM-END is true, the later duplicate of a
   duplicate pair is deleted.")

;(setf (arglist /) (NUMBER &REST NUMBERS))
(setf (documentation '/ 'function) 
  "Returns the reciprocal of NUMBER if NUMBERS is not specified, or NUMBER divided by the
   remaining arguments if NUMBERS is specified.  Note that the division of integers that does
   not produce an exact integer returns a ratio.")

(setf (documentation '/ 'variable)
  "Variable whose value is a list of the results obtained from the last form evaluated by the
   read-eval-print loop.")

;(setf (arglist RANDOM) (&OPTIONAL HIGH ARRAY))
(setf (documentation 'RANDOM 'function) 
  "Returns a randomly chosen number of the same type as HIGH between 0 (inclusive) and 
   HIGH (exclusive).  If ARRAY is specified, it is an object of type random-state that is used
   and updated when generating the random number.")

;(setf (arglist LOGBITP) (INDEX INTEGER))
(setf (documentation 'LOGBITP 'function)
  "Returns true if the bit at position INDEX in word INTEGER is a 1, otherwise NIL.") 

;(setf (arglist IF) (&QUOTE PREDICATE THEN &REST ELSES))
(setf (documentation 'IF 'function)
  "Construct used for conditional execution.  First PREDICATE is evaluated.  If true, the THEN
   form is evaluated and that value returned by IF; if NIL, the ELSE form is evaluated and
   that value returned by IF.  An implementation extension allows multiple forms in the ELSE clause
   that are evaluated as if in a PROGN.") 

;(setf (arglist FIND-SYMBOL) (STRING &OPTIONAL PKG))
(setf (documentation 'FIND-SYMBOL 'function)
  "Returns two values (Common Lisp version) :
     Value1 is the symbol whose print name matches STRING in package PKG.  If the  symbol is
            not found, NIL is returned.
     Value2 takes on one of the following values :
            NIL - symbol was not found.
            :INTERNAL - symbol was directly present as an internal symbol
            :EXTERNAL - symbol was directly present as an external symbol
            :INHERITED - symbol was inherited from a package used by PKG
   An implementation extension third value is also returned.  This value is the name of the
   package in which the symbol was found.  This is different from PKG only if the value of
   Value2 is :INHERITED.  If the symbol is not found, NIL is returned.")

;(setf (arglist LOGNOR) (INTEGER1 INTEGER2))
(setf (documentation 'LOGNOR 'function) 
  "Returns the bitwise-nor of INTEGER1 with INTEGER2.")

;(setf (arglist PROCLAIM) (&REST DECLARATIONS))
(setf (documentation 'PROCLAIM 'function) 
  "Puts DECLARATIONS into effect globally, though it can be locally shadowed.")

;(setf (arglist STRING=) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING= 'function)
  "Returns true if STRING1 is equal to STRING2 considering case and font .  Although the default
   length for both STRING1 and STRING2 is the entire string, a substring of STRING1 starting at
   :START1 and ending at :END1 and/or a substring of STRING2 starting at :START2 and ending at :END2
   may be specified.  NIL is returned if the strings are not equal.") 

(setf (documentation 'CHAR-SUPER-BIT 'variable)
  "Constant whose value gives a numerical meaning to the super bit in the bit configuration
   for a character object.")

;(setf (arglist STRING>=) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING>= 'function)
  "Returns the index into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test greater than or equal to.  Both case and font are considered.  If the test
   is never satisfied, NIL is returned.  If the strings match, the length of the shorter string
   is returned.  Although the default string length of both STRING1 and STRING2 is the entire
   length, a substring of STRING1 starting at :START1 and ending at :END1 and/or and substring
   of STRING2 starting at :START2 and ending at :END2 may be specified.")
 
;(setf (arglist PATHNAME-DIRECTORY) (P))
(setf (documentation 'PATHNAME-DIRECTORY 'function) 
  "Returns the directory component of pathname P, NIL if a directory component does not exist.")

;(setf (arglist GET-SETF-METHOD) (FORM))
(setf (documentation 'GET-SETF-METHOD 'function)
  "Returns five values which constitute the SETF method for FORM.  These values are :
     Value1 is a list of temporary variables for the subforms of the place value used when 
            invoking the SETF.
     Value2 is a list of the subforms to which the temporary variables correspond.
     Value3 is a list of the temporary variables for the values to be stored.
     Value4 is a form that does the storing.
     Value5 is a form to retrieve the argument of the place value used when invoking the SETF.") 

;(setf (arglist WITH-OUTPUT-TO-STRING) ((STREAM STRING INDEX) &BODY BODY))
(setf (documentation 'WITH-OUTPUT-TO-STRING 'function)
  "Evaluates BODY with the variable STREAM bound to a stream.  Two different types of return
   values are possible.  If STRING is specified, the output characters are appended to the end of
   STRING, destructively modifying STRING, and WITH-OUTPUT-TO-STRING returns the results of the last
   form executed in BODY.  If STRING is not specified, a new string is created and returned as the
   return value of WITH-OUTPUT-TO-STRING.  INDEX is an implementation extension argument which allows
   the output to be stored into STRING at position INDEX instead of the end, destructively modifying
   STRING.")

;(setf (arglist SHORT-FLOAT) (NUMBER))
;(setf (documentation 'SHORT-FLOAT 'function) 
;  "Returns NUMBER converted to type short float.")

;(setf (arglist FLOOR) (DIVIDEND &OPTIONAL (DIVISOR 1)))
(setf (documentation 'FLOOR 'function)
  "Returns two values :
     Value1 is the largest integer that is either not larger than DIVIDEND if DIVISOR is not
            specified or not larger than the quotient obtained by dividing DIVIDEND by DIVISOR
            if DIVISOR is specified.
     Value2 is the number to add to Value1 (or Value1 times DIVISOR if DIVISOR is specified) in
            order to equal DIVIDEND.")

;(setf (arglist MAKE-ECHO-STREAM) (INPUT-STREAM OUTPUT-STREAM))
(setf (documentation 'MAKE-ECHO-STREAM 'function) 
  "Returns a bidirectional stream which gets its input from INPUT-STREAM and sends its output to
   OUTPUT-STREAM.  In addition, the input obtained is also echoed to the output stream.")

;(setf (arglist PROBE-FILE) (FILE))
(setf (documentation 'PROBE-FILE 'function) 
  "Returns the truename of FILE if FILE exists, otherwise NIL.")

;(setf (arglist CHAR>) (CHAR &REST CHARS))
(setf (documentation 'CHAR> 'function)
  "Returns true if all the characters are strictly decreasing, considering bits, fonts, and case;
   otherwise, NIL is returned.") 

;(setf (arglist ARRAY-IN-BOUNDS-P) (ARRAY &REST SUBSCRIPTS))
(setf (documentation 'ARRAY-IN-BOUNDS-P 'function) 
  "Returns true if the subscripts given in SUBSCRIPTS are legal subscripts for array ARRAY;
   otherwise, NIL is returned.")

(setf (documentation 'ARRAY-TOTAL-SIZE-LIMIT 'variable) 
  "Constant whose value is the upper exclusive bound on the total number of elements in an array.")

;(setf (arglist CDDDAR) (X))
(setf (documentation 'CDDDAR 'function) 
  "Returns the result of (CDR (CDR (CDR (CAR X)))).")

;(setf (arglist SUBST-IF-NOT) (NEW PREDICATE TREE &KEY KEY))
(setf (documentation 'SUBST-IF-NOT 'function)
  "Returns the tree structure created by replacing each atom or tree in TREE with NEW that does
   not satisfy PREDICATE.  If non-NIL, :KEY is a function applied to each node of TREE to obtain
   the object to test against.  TREE is unaltered.") 

;(setf (arglist DENOMINATOR) (RATIONAL))
(setf (documentation 'DENOMINATOR 'function)
  "Returns the denominator of rational number RATIONAL.  The denominator of an integer is 1.")

;(setf (arglist AREF) (ARRAY &REST INDICES))
(setf (documentation 'AREF 'function) 
  "Returns the element of array ARRAY whose subscripts are specified by INDICES.")

(setf (documentation 'MOST-POSITIVE-DOUBLE-FLOAT 'variable)
  "Constant whose value is the greatest positive magnitude representable by the double float type.") 

;(setf (arglist SUBSTITUTE) (NEW-ITEM ITEM SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END COUNT FROM-END))
(setf (documentation 'SUBSTITUTE 'function)
  "Returns a sequence created by replacing in SEQUENCE each instance of ITEM with NEW-ITEM. 
   Although the entire length of SEQUENCE is the default, a subsequence starting at :START and
   ending at :END may be specified.  If specified, :TEST-NOT is a predicate which must return
   NIL in order for ITEM to be replaced.  If non-NIL, :KEY is a funcion applied to each element of
   SEQUENCE before the test is performed.  :COUNT is an integer specifying how many leftmost
   elements are to be replaced if more than :COUNT number match ITEM.  :FROM-END is only meaningful
   if :COUNT is specified and, if true, causes the rightmost :COUNT elements that match ITEM to be
   replaced.")

;(setf (arglist CHAR-NOT-EQUAL) (&REST CHARS))
(setf (documentation 'CHAR-NOT-EQUAL 'function) 
  "Returns true if all the characters are not equal, ignoring font and case; otherwise,
   NIL is returned.")

;(setf (arglist WRITE-LINE) (STRING &OPTIONAL STREAM &KEY &OPTIONAL (START 0) END))
(setf (documentation 'WRITE-LINE 'function) 
  "Outputs the string STRING followed by a newline character to STREAM.  Although the entire 
   length of STRING is the default, a substring starting at :START and ending at :END may be 
   specified.  WRITE-LINE returns the entire string even if a substring is specified.")

;(setf (arglist GO) (&QUOTE TAGID))
(setf (documentation 'GO 'function)
  "Transfers control to the point in the body labelled by TAGID.  GO returns no value.") 

;(setf (arglist RENAME-FILE) (STRING-OR-STREAM NEW-NAME &KEY (ERROR T) QUERY))
(setf (documentation 'RENAME-FILE 'function)
  "Renames a file from STRING-OR-STREAM to NEW-NAME.  STRING-OR-STREAM can be a filename or
   stream name.  :ERROR is an implementation extension keyword that, when set to NIL, allows an
   error message to be returned to the user as a return value instead of being signaled as a Lisp
   error.  :QUERY is an implementation extension keyword which prompts the user to signal which
   files are to be renamed if wildcarding, also an implementation extension, is used.  RENAME-FILE
   returns three values :  the defaulted pathname NEW-NAME,  the truename of STRING-OR-STREAM, and
   the outcome of the operation.  Lists may be returned if wildcarding is used.  An outcome is
   either the truename of the renamed file, NIL if the user was queried an said no, or an error
   object if the rename failed and :ERROR is set to NIL.")

;(setf (arglist ARRAY-TOTAL-SIZE) (ARRAY))
(setf (documentation 'ARRAY-TOTAL-SIZE 'function) 
  "Returns the total number of elements in ARRAY, calculated as the product of ARRAY's dimensions.")

;(setf (arglist PRINC-TO-STRING) (OBJECT))
(setf (documentation 'PRINC-TO-STRING 'function)
  "Returns as a string and in the form that would be printed by PRINC the object OBJECT.") 

;(setf (arglist SORT) (X SORT-LESSP-PREDICATE &KEY &OPTIONAL KEY))
(setf (documentation 'SORT 'function)
  "Returns a sequence created by destructively modifying X in such a manner that the elements
   in X are in a sorted order.  This sorted order is determined by predicate SORT-LESSP-PREDICATE,
   a test predicate that should accept two arguments and return true when it determines that the
   first argument can be considered less that the second element.  If non-NIL, :KEY specifies a 
   function to be applied to the two arguments before the test is performed.")

;(setf (arglist TENTH) (LIST))
(setf (documentation 'TENTH 'function)
  "Returns the tenth element of LIST.  One-origin numbering is used.")

;(setf (arglist MAKE-BROADCAST-STREAM) (&REST STREAMS))
(setf (documentation 'MAKE-BROADCAST-STREAM 'function)
  "Returns a stream of type output that will resend all output sent to it to all the streams
   given in STREAMS.  If STREAMS is NIL or unspecified, it acts like a sink and discards all output.") 

;(setf (arglist BIT-ORC1) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-ORC1 'function)
  "Returns the bit-array generated from a bitwise-or of the complement of BIT-ARRAY1 with
   BIT-ARRAY2.  If RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted,
   the result is returned as a new bit-array; if a bit-array, the result is destructively placed
   into it.")

;(setf (arglist MAPCAN) (FCT &REST LISTS))
(setf (documentation 'MAPCAN 'function)
  "Applies function FCT to each element of each list in LISTS, terminating when the end of the
   shortest list has been reached.  Function FCT must take as many arguments as there are lists
   in LISTS.  The value returned by MAPCAN is a list of the results of the successive calls to FCT.
   Since this list is created by combining the results using NCONC, not only are all non-list 
   results from FCT ignored, but the lists returned by FCT may also be destructively modified.")

;(setf (arglist SIXTH) (LIST))
(setf (documentation 'SIXTH 'function) 
  "Returns the sixth element of LIST.  One-origin numbering is used.")

;(setf (arglist UNEXPORT) (SYMBOLS &OPTIONAL (PKG *PACKAGE*) FORCE-FLAG))
(setf (documentation 'UNEXPORT 'function) 
  "Makes SYMBOLS internal, not external, in package PKG.  An error is signaled if any symbol is
   not directly present in PKG or if PKG is used by another package.  An implmentation dependent
   option, FORCE-FLAG, when true, will disable the in-use error.  UNEXPORT returns T.")

;(setf (arglist SCALE-FLOAT) (FLONUM INTEGER))
(setf (documentation 'SCALE-FLOAT 'function) 
  "Returns a floating point number of the same type as FLONUM that is the product of FLONUM times
   2 to the INTEGER power.")

;(setf (arglist GET-MACRO-CHARACTER) (CHAR &OPTIONAL (RT *READTABLE*)))
(setf (documentation 'GET-MACRO-CHARACTER 'function)
  "Returns two values :
     Value1 is the function associated with macro character CHAR in readtable RT.
     Value2 is the value of the flag non-terminating-p.  This was set by the function
            SET-MACRO-CHARACTER.
   If CHAR is not a macro character in readtable RT, Value1 and Value2 are returned as NIL.") 

;(setf (arglist FLOAT) (NUMBER &OPTIONAL (OTHER NUMBER)))
(setf (documentation 'FLOAT 'function)
  "Returns as a floating point number NUMBER converted to type OTHER.  OTHER must be a floating
   point number and, if not specified, type single float is used.")

;(setf (arglist CHAR-NOT-LESSP) (CHARACTER &REST MORE-CHARACTERS))
(setf (documentation 'CHAR-NOT-LESSP 'function)
  "Returns true if all the characters are nonincreasing, ignoring font and case; otherwise,
   NIL is returned.") 

;(setf (arglist CHAR-NOT-GREATERP) (&REST CHARS))
(setf (documentation 'CHAR-NOT-GREATERP 'function)
  "Returns true is all the characters are nondecreasing, ignoring font and case; otherwise,
   NIL is returned.") 

;(setf (arglist HASH-TABLE-REHASH-THRESHOLD) (HASH-TABLE))
(setf (documentation 'HASH-TABLE-REHASH-THRESHOLD 'function) 
  "Returns the threshold at which HASH-TABLE is made larger and rehashed.")

;(setf (arglist CONS) (CAR CDR))
(setf (documentation 'CONS 'function) 
  "Returns a newly allocated cons whose CAR is CAR and CDR is CDR.")

;(setf (arglist MAKE-SYNONYM-STREAM) (STREAM-SYMBOL))
(setf (documentation 'MAKE-SYNONYM-STREAM 'function) 
  "Returns a stream which will perform all operations by redirecting those operations to the
   stream that is then the value of the variable STREAM-SYMBOL.")

;(setf (arglist STANDARD-CHAR-P) (CHAR))
(setf (documentation 'STANDARD-CHAR-P 'function)
  "Returns true if CHAR is of type standard-char, otherwise NIL.") 

(setf (documentation 'LEAST-POSITIVE-DOUBLE-FLOAT 'variable)
  "Constant whose value is the smallest positive magnitude representable by the double float type.") 

;(setf (arglist MEMBER-IF-NOT) (PREDICATE LIST &KEY KEY))
(setf (documentation 'MEMBER-IF-NOT 'function)
  "Returns a list consisting of the tail of list LIST whose CAR is the first element of LIST that
   does not satisfy test PREDICATE.  If no match is found, NIL is returned.  If non-NIL, :KEY is a
   function applied to each element of LIST before the test is performed.")
   
;(setf (arglist USE-PACKAGE) (PKGS-TO-USE &OPTIONAL (PKG *PACKAGE*)))
(setf (documentation 'USE-PACKAGE 'function) 
  "Adds PKGS-TO-USE, a list of packages or package names, to the inheritance list of PKG, thus
   making all external symbols in these packages accessible to PKG as internal symbols.  
   USE-PACKAGE returns T.") 

;(setf (arglist SHIFTF) (&REST PLACES-AND-FINAL-VALUE))
(setf (documentation 'SHIFTF 'function)
  "Shifts the place values left one position by setting the first place to the value of the
   second place, the second to the third, and so on.  That is, place-value-1 = place-value-2,
   place-value-2 = place-value-3,..., place-value-last = value-of-last-arg.  PLACES-AND-FINAL-VALUE
   is an even number of forms acceptable as a generalized variable to SETF.  SHIFTF returns the
   value of place-value-1, the place value that is shifted out.")

(setf (documentation 'BOOLE-AND 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-and of BOOLE's second
   argument with BOOLE's third argument.")

;(setf (arglist CALL) (FUNCTION ARG-DESC-1 ARG-DATA-1 ARG-DESC-2 ARG-DATA-2 ETC))
(setf (documentation 'CALL 'function)
  "Provides a generalized way of controlling what arguments are passed to a function.  FUNCTION is 
   the function to call.  The remaining arguments are in pairs, consisting of first a descriptor
   argument followed by a data argument.  The descriptor argument defines the action to take with
   the data argument.  The descriptor argument value is either a keyword, a list of keywords, or
   NIL.  NIL signifies that the data argument is to be treated as a single argument to FUNCTION.
   The allowed keywords and their definition are :
      :SPREAD - The data argument is a list and each element of the list should be passed as a
                separate agrument to FUNCTION.
      :OPTIONAL - All remaining arguments to the right are optional arguments.
   The form '(:OPTIONAL :SPREAD) is acceptable and specifies that both keywords are in effect.")

;(setf (arglist PATHNAME) (OBJECT))
(setf (documentation 'PATHNAME 'function)
  "Returns OBJECT converted into a pathname.  If OBJECT is already a pathname, it is unchanged;
   if a string or symbol, it is parsed.")

;(setf (arglist UNWIND-PROTECT) (&QUOTE BODY-FORM &REST CLEANUP-FORMS))
(setf (documentation 'UNWIND-PROTECT 'function)
  "Special form used for protection against nonlocal exits.  After the forms in BODY-FORM are 
   evaluated the forms in CLEANUP-FORMS are evaluated, even if a nonlocal exit such as a THROW,
   RETURN, or signaled error would cause execution to be transferred out of the UNWIND-PROTECT.
   The result of the last form evaluated in BODY-FORM is returned by UNWIND-PROTECT.")

;(setf (arglist NSUBST) (NEW OLD TREE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'NSUBST 'function)
  "Returns a tree created by substituting in TREE each element or subtree that matches OLD
   with NEW.  TREE is destructively modified by NSUBST.  Object are said to match if predicate :TEST
   returns true or, if specified, :TEST-NOT returns NIL.  If non-NIL, :KEY is a function applied to
   each node of TREE to obtain the object to test against.")

;(setf (arglist SET-SYNTAX-FROM-CHAR) (TO-CHAR FROM-CHAR &OPTIONAL (TO-READTABLE *READTABLE*) (FROM-READTABLE NIL)))
(setf (documentation 'SET-SYNTAX-FROM-CHAR 'function)
  "Copies the syntax of FROM-CHAR in readtable FROM-READTABLE to TO-CHAR in readtable TO-READTABLE.") 

(setf (documentation 'SINGLE-FLOAT-NEGATIVE-EPSILON 'variable)
  "Constant whose value is the smallest possible positive number of type single float that, when
   subtracted from some value, will change the magnitude of that value.")

;(setf (arglist BIT-VECTOR-P) (OBJECT))
(setf (documentation 'BIT-VECTOR-P 'function)
  "Returns true if OBJECT is a bit vector, otherwise NIL.") 

;(setf (arglist FLOAT-RADIX) (FLOAT))
(setf (documentation 'FLOAT-RADIX 'function) 
  "Returns as an integer the radix of FLOAT.  The present value returned is always 2 (binary).")

;(setf (arglist STRING-CHAR-P) (CHAR))
(setf (documentation 'STRING-CHAR-P 'function)
  "Returns true if CHAR can be stored in a Common Lisp string, otherwise NIL.") 

;(setf (arglist UNTRACE) (&REST FNS))
(setf (documentation 'UNTRACE 'function) 
  "Causes those functions listed in FNS to no longer be traced.  If FNS is not specified, all
   traced functions are untraced.")

(setf (documentation 'BOOLE-2 'variable)
  "Constant used as first argument to function BOOLE to return BOOLE's third argument.")

;(setf (arglist NSUBST-IF-NOT) (NEW PREDICATE TREE &KEY KEY))
(setf (documentation 'NSUBST-IF-NOT 'function)
  "Returns the tree structure created by replacing each atom or tree in TREE with NEW that does
   not satisfy PREDICATE.  TREE is destructively modified by NSUBST-IF-NOT.  If non-NIL, :KEY is a
   function applied to each node of TREE to obtain the object to test against.") 

;(setf (arglist ERROR) (FORMAT-STRING &REST FORMAT-ARGS))
(setf (documentation 'ERROR 'function) 
  "Reports a detectable but uncorrectable error.  FORMAT-STRING will be the message displayed as
   the reason for the error; FORMAT-ARGS are the format arguments for the format string.")

;(setf (arglist NAME-CHAR) (NAME))
(setf (documentation 'NAME-CHAR 'function)
  "Returns as a character object the character for which NAME is a name.  NIL is returned if
   NAME is not a recognizable character name.") 

;(setf (arglist FRESH-LINE) (&OPTIONAL STREAM))
(setf (documentation 'FRESH-LINE 'function) 
  "Outputs a Newline character to stream STREAM if not already at the beginning of a new line.
   FRESH-LINE returns true if a Newline character is output, otherwise NIL.")

;(setf (arglist DELETE-FILE) (STRING-OR-STREAM &KEY (ERROR T) QUERY))
(setf (documentation 'DELETE-FILE 'function)
  "Deletes file STRING-OR-STREAM.  STRING-OR-STREAM can be a filename or stream name.   If the 
   name of an open stream, the stream is closed and the file deleted.  :ERROR is an implementation
   extension keyword that, when set to NIL, allows an error message to be returned to the user as
   a return value instead of being signaled as a Lisp error.  :QUERY is an implementation extension
   keyword which causes the user to be prompted before any file is deleted.  This is particularly
   useful when wildcarding, also an implementation extension, is used.  DELETE-FILE returns a list
   of lists where each sublist consists of both the truename of the file acted upon and true if the
   file was deleted, NIL if not deleted due to user input when queried, or the error obtained if an
   error occured and :ERROR is set to NIL.")

;(setf (arglist LOGXOR) (&REST INTEGERS))
(setf (documentation 'LOGXOR 'function)
  "Returns the bitwise-exclusive-or of INTEGERS, or 0 if no arguments are given.")

;(setf (arglist MAKE-LIST) (LENGTH &KEY AREA INITIAL-ELEMENT))
(setf (documentation 'MAKE-LIST 'function)
  "Returns a list containing LENGTH number of elements, each of with are initialized to NIL or, 
   if specified, :INITIAL-ELEMENT.  The implementation extension keyword :AREA allows the list
   to be created in that specific area.")

;(setf (arglist RPLACD) (CONS NEW-CDR))
(setf (documentation 'RPLACD 'function)
  "Returns the modified cons where the CDR of cons CONS has been replaced by NEW-CDR.  CONS is
   destructively modified.") 

;(setf (arglist STREAMP) (OBJECT))
(setf (documentation 'STREAMP 'function)
  "Returns true if OBJECT is a stream, otherwise NIL.  This predicate considers the following to
   be streams:  any instance incorporating SI:STREAM or TV:SHEET, any function handling either
   :TYI or :TYO, or any symbol with a non-NIL SI:IO-STREAM-P property.")

;(setf (arglist STRING>) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING> 'function)
  "Returns the index into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test greater than.  Both case and font are considered.  If the test is never
   satisfied, NIL is returned.  If the strings match, the length of the shorter string is returned.
   Although the default string length of both STRING1 and STRING2 is the entire length, a
   substring of STRING1 starting at :START1 and ending at :END1 and/or and substring of STRING2
   starting at :START2 and ending at :END2 may be specified.")

;(setf (arglist PACKAGEP) (OBJECT))
(setf (documentation 'PACKAGEP 'function) 
  "Returns true if OBJECT is a package, otherwise NIL.")

(setf (documentation 'ARRAY-DIMENSION-LIMIT 'variable)
  "Constant whose positive integer value is the exclusive upper bound on the number of elements
   any one dimension of an array may have.") 
 
;(setf (arglist PATHNAMEP) (OBJECT))
(setf (documentation 'PATHNAMEP 'function)
  "Returns true if OBJECT is a pathname object, otherwise NIL.") 

;(setf (arglist GET-UNIVERSAL-TIME) NIL)
(setf (documentation 'GET-UNIVERSAL-TIME 'function) 
  "Returns the current time as a single integer in Universal Time format.")

;(setf (arglist MULTIPLE-VALUE-LIST) (&QUOTE EXP))
(setf (documentation 'MULTIPLE-VALUE-LIST 'function)
  "Returns as a list the multiple values returned from the evaluation of form EXP.") 

;(setf (arglist STABLE-SORT) (X LESSP-PREDICATE &KEY &OPTIONAL KEY))
(setf (documentation 'STABLE-SORT 'function)
  "Returns a sequence created by destructively modifying X in such a manner that the elements
   in X are in a sorted order.  This sorted order is determined by predicate LESSP-PREDICATE,
   a test predicate that should accept two arguments and return true when it determines that the
   first argument can be considered less that the second element.  If non-NIL, :KEY specifies a 
   function to be applied to the two arguments before the test is performed.  If two elements
   are found to be equal, their original order is preserved in the new sorted list.")

;(setf (arglist DEFINE-SETF-METHOD) (ACCESS-FUNCTION LAMBDA-LIST &BODY BODY))
(setf (documentation 'DEFINE-SETF-METHOD 'function)
  "Defines a generalized SETF expander for ACCESS-FUNCTION.  Syntax of 
   (DEFINE-SETF-METHOD access-function access-lambda-list body) is defined as :
   <access-lambda-list> is the argument forms of <access-function>.  <body> is evaluated and 
   must return five values representating everything macros need to know for manipulating 
   generalized variables.  These five values are :
     Value1 is a list of temporary variables for the subforms of the place value used when 
            invoking the SETF.
     Value2 is a list of the subforms to which the temporary variables correspond.
     Value3 is a list of the temporary variables for the values to be stored.
     Value4 is a form that does the storing.
     Value5 is a form to retrieve the argument of the place value used when invoking the SETF.
   With these values the SETF method can decide how to parse the form to be SETF'ed and which
   parts to replace with tempvars, among other things.")

;(setf (arglist FILL) (SEQUENCE ITEM &KEY START END))
(setf (documentation 'FILL 'function)
  "Returns the sequence created by destructively modifying SEQUENCE by replacing each element with
   ITEM.  Although the entire length of SEQUENCE is the default, a subsequence starting at :START
   and ending at :END may be specified.") 

(setf (documentation 'BOOLE-CLR 'variable)
  "Constant used as first argument to function BOOLE to return an output whose bits are all zero.")

;(setf (arglist LDIFF) (LIST TAIL))
(setf (documentation 'LDIFF 'function) 
  "Returns a list consisting of that part of LIST which precedes TAIL.  If TAIL is not a sublist
   of LIST, all elements of LIST are copied.")

;(setf (arglist LOGEQV) (&REST INTEGERS))
(setf (documentation 'LOGEQV 'function)
  "Returns the bitwise-equivalence of INTEGERS, or -1 if no arguments are given.")

;(setf (arglist LOGNAND) (INTEGER1 INTEGER2))
(setf (documentation 'LOGNAND 'function)
  "Returns the bitwise-not-and of INTEGER1 with INTEGER2.") 

;(setf (arglist INPUT-STREAM-P) (STREAM))
(setf (documentation 'INPUT-STREAM-P 'function)
  "Returns true if STREAM, which must be a stream object, can handle input operations;  otherwise,
   NIL is returned.") 

;(setf (arglist MERGE) (RESULT-TYPE SEQUENCE1 SEQUENCE2 PREDICATE &KEY KEY))
(setf (documentation 'MERGE 'function) 
  "Returns a sequence of type RESULT-TYPE containing the elements of SEQUENCE1 and SEQUENCE2
   interleaved in order according to predicate PREDICATE.  Since MERGE is stable, if the next
   element to merge from SEQUENCE1 is equal to the next element to merge from SEQUENCE2 according
   to PREDICATE, the element from SEQUENCE1 will precede the one from SEQUENCE2.  If non-NIL, :KEY
   specifies a function to apply to each element of each sequence before the test is performed.")

;(setf (arglist PARSE-INTEGER) (STRING &KEY (START 0) (END (LENGTH STRING)) (RADIX 10) (JUNK-ALLOWED NIL)))
(setf (documentation 'PARSE-INTEGER 'function) 
  "Returns two values :
     Value1 is a number parsed from the contents of STRING.  Although the entire length of STRING
            is the default, a substring starting at :START and ending at :END may be specified.
            Leading whitespace is ignored.
     Value2 is the index into STRING of the delimiter that terminated parsing.
   :RADIX is the radix of the number to be read and is between 2 and 36 inclusive.  If :JUNK-ALLOWED
   is NIL, the string specified must contain only a number possibly surrounded by whitespace.  The
   number may be digits up to that allowed by the radix and an optional sign.  If :JUNK-ALLOWED is
   true, parsing stops when a nondigit is found.  If no number has been found by by the end of
   the string, NIL is returned.")


;(setf (arglist RASSOC-IF) (PREDICATE ALIST))
(setf (documentation 'RASSOC-IF 'function)
  "Returns the first association pair in association list ALIST whose CDR satisfies test predicate
   PREDICATE.  If the test is not satisfied, RASSOC-IF returns NIL.")

;(setf (arglist READ-LINE) (&OPTIONAL (STREAM *STANDARD-INPUT*) (EOF-ERROR-P T) EOF-VALUE RECURSIVE-P RH-OPTIONS))
(setf (documentation 'READ-LINE 'function)
  "Returns two values :
     Value1 is a string obtained by reading a line from STREAM terminated by Newline and returning
            it as a string without the Newline.  End of file is an error if EOF-ERRORP is true;
            otherwise READ-LINE returns EOF-VALUE when end of file is read.  The argument
            RECURSIVE-P is ignored.
     Value2 is a flag whose value is NIL if the line terminated normally or true if end of file
            was found.
   RH-OPTIONS is an implementation extension argument which is a list of rubout handler options, 
   passed to WITH-INPUT-EDITING if it is used.")

;(setf (arglist TYPE-SPECIFIER-P) (TYPE))
(setf (documentation 'TYPE-SPECIFIER-P 'function) 
  "Returns true if TYPE is a  valid type specifier, otherwise NIL.")

;(setf (arglist FLOAT-PRECISION) (FLOAT))
(setf (documentation 'FLOAT-PRECISION 'function) 
  "Returns as a non-negative integer the number of significant bits used to represent FLOAT.")

;(setf (arglist MAKE-PACKAGE) (NAME &KEY NICKNAMES (USE (QUOTE (LISP TICL))) (SIZE 200) SHADOW EXPORT
;                               PREFIX-NAME AUTO-EXPORT-P IMPORT (AREA NR-SYM) SHADOWING-IMPORT PLIST
;                               STORE-FUNCTION AFTER-INTERN-DAEMON))
(setf (documentation 'MAKE-PACKAGE 'function)
  "Creates and returns a package object named NAME with nicknames :NICKNAMES.  The name and
   nicknames must be distinct from the names and nicknames for all existing packages or an error
   will be signaled.  :USE is a list of packages whose external symbols are to be inherited by NAME.
   Implementation extensions to the argument list are defined as follows :
     :SHADOW - Specifies a list of symbol names to shadow in this package.
     :EXPORT - Specifies a list of symbol names to export in this package.
     :IMPORT - Specifies a list of symbol names to import into this package.
     :SHADOWING-IMPORT  - Specifies a list of symbols to import into this package, overriding
                          any name conflicts.
     :PREFIX-NAME  - Specifies the prefix name to be used when printing the symbols.  This must be
                     a name or nickname of the package.
     :AUTO-EXPORT-P - When true it specifies that all symbols interned in this package should be
                      exported automatically.
     :SIZE - Specifies the number of symbols for which space is to be initially allocated.")

;(setf (arglist ED) (&OPTIONAL THING))
(setf (documentation 'ED 'function) 
  "Selects an editor window in order to edit THING.  If THING is T, select a newly created buffer.
   If THING is a string or pathname, find and edit that file.  If THING is a list whose car is
   CALLERS-TO-BE-EDITED  or FUNCTIONS-TO-BE-EDITED, make possibilities for them and edit the first
   one.  If THING is a function or function spec, edit that function's definition.")

;(setf (arglist PPRINT) (OBJECT &OPTIONAL (*STANDARD-OUTPUT* *STANDARD-OUTPUT*)))
(setf (documentation 'PPRINT 'function)
  "Outputs the printed representation of OBJECT to stream *STANDARD-OUTPUT* preceded by a newline.
   The *PRINT-PRETTY* flag is set to true during the execution of PPRINT.  PPRINT returns no
   values.") 

;(setf (arglist SQRT) (NUMBER))
(setf (documentation 'SQRT 'function)
  "Returns the square root of NUMBER either as a floating point or, if required, a complex number.") 

;(setf (arglist PACKAGE-NICKNAMES) (PKG))
(setf (documentation 'PACKAGE-NICKNAMES 'function) 
  "Returns the list of nicknames as strings of the package PKG.  PKG's name is not included.")

;(setf (arglist ETYPECASE) (OBJECT &BODY CLAUSES))
(setf (documentation 'ETYPECASE 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (ETYPECASE OBJECT (type1 body1) ... (typeN bodyN)) is defined as :  the (type# body#) construct
   is called a clause.  OBJECT is evaluated.  <type#> is a literal type specifier and is not
   evaluated.   <type1> is compared against OBJECT's type; if they the same, all forms in <body1>
   are evaluated sequentially and the value of the last expression in <body1> is returned by
   ETYPECASE.  If <type1> does not match OBJECT's type, <type2> is compared to OBJECT's type, and
   so on, until either one <type#> matches OBJECT's type or all clauses are exhausted.   If no <type#>
   matches the data type of OBJECT, an error is signaled with a message constructed from the clauses.
   Continuing from the error is not permitted.  An OTHERWISE or T case is not permitted.")

;(setf (arglist SET-DIFFERENCE) (LIST1 LIST2 &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'SET-DIFFERENCE 'function)
  "Returns a list of those elements in LIST1 that are not in LIST2 according to test :TEST.  If
   :TEST-NOT is specified, it is a predicate accepting two arguments that must return true in order
   for a element in LIST1 not to be considered in LIST2.  If non-NIL, :KEY is a function applied to
   each element of LIST1 and LIST2 before the test is performed.") 

;;(setf (arglist *TERMINAL-IO*) (&REST ARGS))
;(setf (documentation '*TERMINAL-IO* 'function) 
;  NIL)

(setf (documentation '*TERMINAL-IO* 'variable)
  "Global variable whose value is the default stream for I/O to the console, which is normally
   the selected window for output and the keyboard for input when the operations are interactive.") 

;(setf (arglist GRAPHIC-CHAR-P) (CHAR))
(setf (documentation 'GRAPHIC-CHAR-P 'function)
  "Returns true if CHAR is a graphic character, otherwise NIL.") 

;(setf (arglist HASH-TABLE-TEST) (HASH-TABLE))
(setf (documentation 'HASH-TABLE-TEST 'function)
  "Returns the test function used on hash table HASH-TABLE.") 

;(setf (arglist PROG) (&REST BODY))
(setf (documentation 'PROG 'function)
  "Macro allowing local variable bindings, return statements, and GO statements.  Syntax of  
   (PROG [name] ((var1 exp1) ... (varN expN)) [(declare...)] body) is defined as :
   <name>, if specified, is a symbol which names the PROG construct and allows the use of the 
   (RETURN-FROM <name>) statement.   All expressions <exp1> thru <expN> are evaluated and their
   resulting values saved.  These values are then bound to <var1> thru <varN>, respectively, in
   parallel.  These binding are in force only over the scope of the PROG.  If an expression is not
   specified, NIL is used.  All forms in <body>, which may include GO and RETURN statements, are
   then evaluated with these binding in force; PROG returns the result of the last form evaluated.")

(setf (documentation '*** 'variable)
  "Variable whose value is bound to the printed result of the read-eval-print loop evaluated
   three interaction ago.")

;(setf (arglist ACONS) (KEY DATUM ALIST))
(setf (documentation 'ACONS 'function) 
  "Returns an association list created by consing the association pair (KEY . DATUM) to the front
   of ALIST.")

;(setf (arglist LIST*) (FIRST &REST ELEMENTS))
(setf (documentation 'LIST* 'function)
  "Returns a list constructed by using FIRST as the car of the new list and ELEMENTS as the
   remaining elements of the returned list.  The last cons is constructed as a dotted pair;
   the last argument in ELEMENTS is used as the CDR of the last cons constructed.  If ELEMENTS
   is not specified, LIST* returns the argument FIRST, not a list.") 

;(setf (arglist LOGIOR) (&REST INTEGERS))
(setf (documentation 'LOGIOR 'function)
  "Returns the bitwise-or of INTEGERS, or 0 if no arguments are given.")

;(setf (arglist CIS) (ANGLE))
(setf (documentation 'CIS 'function)
  "Returns the phase of ANGLE given in radians.  This is a complex number and is equal to 
   (COMPLEX (COS ANGLE) (SIN ANGLE)).") 

;(setf (arglist SPECIAL) (&QUOTE &REST SYMBOLS))
(setf (documentation 'SPECIAL 'function) 
  "Specifies that all the SYMBOLS be marked special for compilation.  This special form is 
   considered obsolute.")

(setf (documentation 'LONG-FLOAT-NEGATIVE-EPSILON 'variable)
  "Constant whose value is the smallest possible positive number of type long float that, when
   subtracted from some value, will change the magnitude of that value.")

;(setf (arglist SUBSTITUTE-IF) (NEW-ITEM PREDICATE SEQUENCE &KEY KEY START END COUNT FROM-END))
(setf (documentation 'SUBSTITUTE-IF 'function)
  "Returns a sequence created by replacing in SEQUENCE each element that satisfies PREDICATE with
   NEW-ITEM.  Although the entire length of SEQUENCE is the default, a subsequence starting at
   :START and ending at :END may be specified.  If non-NIL, :KEY is a function applied to each
   element of SEQUENCE before the test is performed.  :COUNT is an integer specifying how many
   leftmost elements are to be replaced if more than :COUNT number satisfy PREDICATE.  :FROM-END
   is only meaningful if :COUNT is specified and, if true,  causes the rightmost :COUNT elements
   that satisfy PREDICATE to be replaced.")

;(setf (arglist BIT-ORC2) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-ORC2 'function)
  "Returns the bit-array generated from a bitwise-or of BIT-ARRAY1 with the complement of 
   BIT-ARRAY2.  If RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted,
   the result is returned as a new bit-array; if a bit-array, the result is destructively placed
   into it.")

(setf (documentation 'CHAR-FONT-LIMIT 'variable)
  "Constant whose value is the upper exclusive bound of the font attribute value of any character.") 

;(setf (arglist MEMBER-IF) (PREDICATE LIST &KEY KEY))
(setf (documentation 'MEMBER-IF 'function)
  "Returns a list consisting of the tail of list LIST whose CAR is the first element of LIST
   that satisfies test PREDICATE.  If no match is found, NIL is returned.  If non-NIL, :KEY is
   a function applied to each element of LIST before the test is performed.")

;(setf (arglist UNINTERN) (SYMBOL &OPTIONAL PKG))
(setf (documentation 'UNINTERN 'function) 
  "Removes (uninterns) the symbol SYMBOL from package PKG, or the current package if PKG is not
   specified.  The symbol is also removed from the shadowing-symbols list.  UNINTERN returns T
   if the symbol was uninterned, otherwise NIL.")

;(setf (arglist OUTPUT-STREAM-P) (STREAM))
(setf (documentation 'OUTPUT-STREAM-P 'function)
  "Returns true if STREAM, which must be a stream object, can handle output operations;  otherwise,
   NIL is returned.") 

;(setf (arglist READTABLEP) (OBJECT))
(setf (documentation 'READTABLEP 'function)
  "Returns true if OBJECT is a readtable, otherwise NIL.") 

;(setf (arglist GCD) (&REST NUMBERS))
(setf (documentation 'GCD 'function)
  "Returns the greatest common divisor of NUMBERS, or either the absolute value of NUMBERS if only one 
   argument is given or 0 if no arguments are given.")

;(setf (arglist PAIRLIS) (KEYS DATA &OPTIONAL STARTING-ALIST))
(setf (documentation 'PAIRLIS 'function)
  "Returns an association list created by associating elements of list KEYS with the corresponding
   element from list DATA.  KEYS and DATA must be the same length.  If STARTING-ALIST IS specified,
   the new pairs are added to the front of STARTING-ALIST.")

;(setf (arglist MAPLIST) (FCT &REST LISTS))
(setf (documentation 'MAPLIST 'function)
  "Returns a list created by appending together the results obtained by applying function FCT to each
   list in LISTS, then successively to the CDR of each of the lists, terminating when the end of the
   shortest list has been reached.  Function FCT must take as many arguments as there are lists
   in LISTS.") 

;(setf (arglist SYMBOLP) (X))
(setf (documentation 'SYMBOLP 'function) 
  "Returns true if X is a symbol, otherwise NIL.")

;(setf (arglist EXPT) (BASE EXPONENT))
(setf (documentation 'EXPT 'function)
  "Returns the value of BASE raised to the EXPONENT power.") 

;(setf (arglist LCM) (NUMBER &REST NUMBERS))
(setf (documentation 'LCM 'function) 
  "Returns the least common multiple of all the arguments.  At least one argument must be given and
   if any arguments are 0, LCM returns 0.")

;(setf (arglist ADJOIN) (ITEM LIST &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'ADJOIN 'function)
  "Returns a list created by adding ITEM to LIST only if ITEM is not already a member of LIST.
   :TEST is a predicate that should return NIL when tested against each element in LIST in order
   for ITEM to be added.  If specified, :TEST-NOT is a predicate which should return true when
   tested against each element in LIST in order for ITEM to be added.  If non-NIL, :KEY is a
   function applied not only to each element of LIST but also to ITEM before the test is performed.")

;(setf (arglist MAKE-STRING) (LENGTH &KEY &OPTIONAL INITIAL-ELEMENT &ALLOW-OTHER-KEYS))
(setf (documentation 'MAKE-STRING 'function) 
  "Returns a simple string of LENGTH elements, all elements initialized to INITIAL-ELEMENT.  If
   INITIAL-ELEMENT is not specified, the implementation defined initial value is the character
   with code 0.")

;(setf (arglist WRITE) (OBJECT &KEY &OPTIONAL (STREAM *STANDARD-OUTPUT*) ((ESCAPE *PRINT-ESCAPE*) *PRINT-ESCAPE*)
;                        ((RADIX *PRINT-RADIX*) *PRINT-RADIX*) ((BASE *PRINT-BASE*) *PRINT-BASE*) ((CIRCLE *PRINT-CIRCLE*)
;                         *PRINT-CIRCLE*) ((PRETTY *PRINT-PRETTY*) *PRINT-PRETTY*) ((LEVEL *PRINT-LEVEL*) *PRINT-LEVEL*)
;                        ((LENGTH *PRINT-LENGTH*) *PRINT-LENGTH*) ((CASE *PRINT-CASE*) *PRINT-CASE*) ((GENSYM *PRINT-GENSYM*)
;                        *PRINT-GENSYM*) ((ARRAY *PRINT-ARRAY*) *PRINT-ARRAY*)))
(setf (documentation 'WRITE 'function)
  "Prints OBJECT on output stream :STREAM.  WRITE returns OBJECT.  The keyword arguments control
   parameters that affect printing and default to the values of the global variables *print-escape*,
   *print-radix*, *print-base*, *print-circle*, *print-pretty*, *print-level*, *print-length*,
   *print-case*, *print-gensym*, and *print-array*, respectively.")

;(setf (arglist STRING<=) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING<= 'function)
  "Returns the index into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test less than or equal to.  Both case and font are considered.  If the test
   is never satisfied, NIL is returned.  If the strings match, the length of the shorter string
   is returned.  Although the default string length of both STRING1 and STRING2 is the entire
   length, a substring of STRING1 starting at :START1 and ending at :END1 and/or and substring
   of STRING2 starting at :START2 and ending at :END2 may be specified.")

;(setf (arglist MULTIPLE-VALUE-SETQ) (&QUOTE VAR-LIST EXP))
(setf (documentation 'MULTIPLE-VALUE-SETQ 'function) 
  "Evaluates EXP, and assigns the variables in VAR-LIST to them.  If EXP returns more values than
   there are variables, the extra values are ignored; if there are more variables than values, they
   are assigned the value NIL.  MULTIPLE-VALUE-SETQ returns the first value returned by EXP.")

(setf (documentation '*PRINT-CIRCLE* 'variable)
  "Variable whose value controls the printing of circular structures.  If true, circular
   structures are represented with #n# and #n= labels when printing; if NIL, printing is done
   through recursive descent into the structure.")

;(setf (arglist /=) (&REST NUMBERS))
(setf (documentation '/= 'function) 
  "Returns true if all arguments are not equal in magnitude, otherwise NIL.")

(setf (documentation 'DOUBLE-FLOAT-NEGATIVE-EPSILON 'variable)
  "Constant whose value is the smallest possible positive number of type double float that, when
   subtracted from some value, will change the magnitude of that value.")

;(setf (arglist PRINC) (OBJECT &OPTIONAL STREAM))
(setf (documentation 'PRINC 'function) 
  "Prints OBJECT on STREAM.  Strings and characters will print their contents with no delimiters
   or quoting.  A symbol is printed as the characters of its print name.  PRINC returns OBJECT.")

(setf (documentation 'MOST-POSITIVE-LONG-FLOAT 'variable)
  "Constant whose value is the greatest positive magnitude representable by the long float type.") 

;(setf (arglist MAKE-DISPATCH-MACRO-CHARACTER) (CHAR &OPTIONAL (NON-TERMINATING-P NIL) (RT *READTABLE*)))
(setf (documentation 'MAKE-DISPATCH-MACRO-CHARACTER 'function) 
  "Makes CHAR a dispatch macro character in readtable RT.  If NON-TERMINATING-P is true, CHAR
   will be a non-terminating macro character and may therefore be embedded within extended tokens.
   MAKE-DISPATCH-MACRO-CHARACTER returns T.")

;(setf (arglist GET-SETF-METHOD-MULTIPLE-VALUE) (FORM &OPTIONAL SHORT-CUT))
(setf (documentation 'GET-SETF-METHOD-MULTIPLE-VALUE 'function)
  "Returns five values constituting the SETF method for FORM.  These values are :
     Value1 is a list of temporary variables for the subforms of the place value used when 
            invoking the SETF.
     Value2 is a list of the subforms to which the temporary variables correspond.
     Value3 is a list of the temporary variables for the values to be stored.
     Value4 is a form that does the storing.
     Value5 is a form to retrieve the argument of the place value used when invoking the SETF.
   This differs from GET-SETF-METHOD in that the number of values to be stored is not checked.
   SHORT-CUT is an implementation extension argument that, when set to true, will return
   only the setting function and a replacement form if the method of SETF'ing was defined by
   a simple DEFSETF.  This condition is signaled by returning Value1 as a non-NIL symbol.")   

;(setf (arglist READ-CHAR-NO-HANG) (&OPTIONAL STREAM (EOF-ERRORP T) EOF-VALUE RECURSIVE-P))
(setf (documentation 'READ-CHAR-NO-HANG 'function)
  "Returns one character object after reading that character from STREAM.  READ-CHAR-NO-HANG
   does not wait for input.  On an interactive stream, if no input is currently buffered, NIL
   is returned.  End of file is an error if EOF-ERRORP is true; otherwise READ-CHAR-NO-HANG
   returns EOF-VALUE when end of file is read.  The argument RECURSIVE-P is ignored.")

;(setf (arglist CTYPECASE) (OBJECT &BODY CLAUSES))
(setf (documentation 'CTYPECASE 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (CTYPECASE OBJECT (type1 body1) ... (typeN bodyN)) is defined as :  the (type# body#) construct
   is called a clause.  OBJECT is evaluated.  <type#> is a literal type specifier and is not
   evaluated.   <type1> is compared against OBJECT's type; if they the same, all forms in <body1>
   are evaluated sequentially and the value of the last expression in <body1> is returned by
   CTYPECASE.  If <type1> does not match OBJECT's type, <type2> is compared to OBJECT's type, and
   so on, until either one <type#> matches OBJECT's type or all clauses are exhausted.   If no <type#>
   matches the data type of OBJECT, an error is signaled with a message constructed from the clauses.
   Continuing from the error causes CTYPECASE to accept a new value for OBJECT and redo the clause
   tests again.  An OTHERWISE or T case is not permitted.")

;(setf (arglist LOGTEST) (BITS WORD))
(setf (documentation 'LOGTEST 'function)
  "Returns true if any of the 1 bits in BITS are also 1's in WORD, otherwise NIL.") 

;(setf (arglist TERPRI) (&OPTIONAL STREAM))
(setf (documentation 'TERPRI 'function)
  "Outputs a Newline to STREAM.  TERPRI returns NIL.") 

;(setf (arglist COPY-TREE) (TREE))
(setf (documentation 'COPY-TREE 'function) 
  "Returns a tree which is a copy of TREE.  All cons cells in the tree are copied; objects that
   are not cons cells are EQ to those in TREE.")

;(setf (arglist GETHASH) (KEY HASH-TABLE &OPTIONAL DEFAULT-VALUE))
(setf (documentation 'GETHASH 'function)
  "Returns three values, the third being an implementation extension to Common Lisp:
     Value1 is the associated value of the entry whose key is KEY.  If KEY is not found,
            DEFAULT-VALUE is returned if specified, otherwise NIL.
     Value2 is true if KEY is found, otherwise NIL.
     Value3, an implementation extension, is a list whose CAR is KEY and whose CDR is all the
             associated values of KEY if the hash table has more than one value per entry.")

(setf (documentation 'CHAR-CODE-LIMIT 'variable)
  "Constant whose value is the upper exclusive bound on the code attribute value of any character.")  

;(setf (arglist DEFVAR) (VARIABLE . ARGS))
(setf (documentation 'DEFVAR 'function) 
  "Defines a special variable named VARIABLE.  ARGS can have two values, the first being the 
   initial value of VARIABLE and the second a documentation string for VARIABLE.  The initial
   value form is not evaluated if VARIABLE already has a value.")

;(setf (arglist RATIONAL) (X))
(setf (documentation 'RATIONAL 'function)
  "Returns a rational number numerically equal to the precise value of the floating point number X.
   If X is an integer or ratio, it is returned unchanged.")

(setf (documentation '*MACROEXPAND-HOOK* 'variable) 
  "Variable whose value is the function used by MACROEXPAND-1.  This function expects three
   arguments : the expander function, the form to be expanded, and an environment argument")

(setf (documentation 'CHAR-HYPER-BIT 'variable)
  "Constant whose value gives a numerical meaning to the hyper bit in the bit configuration
   for a character object.")

;(setf (arglist LOGNOT) (INTEGER))
(setf (documentation 'LOGNOT 'function) 
  "Returns the bitwise complement of INTEGER.")

(setf (documentation '*READ-DEFAULT-FLOAT-FORMAT* 'variable)
  "Variable whose value is the type specifier symbol for floating point format when the precise
   type is not specified by the syntax.  Default is single-float.") 

;(setf (arglist REMHASH) (KEY HASH-TABLE))
(setf (documentation 'REMHASH 'function) 
  "Deletes any entry for KEY in HASH-TABLE.  HASH-TABLE is destructively modified.  REMHASH returns
   true if there was an entry, otherwise NIL is returned.")

(setf (documentation 'SHORT-FLOAT-NEGATIVE-EPSILON 'variable)
  "Constant whose value is the smallest possible positive number of type short float that, when
   subtracted from some value, will change the magnitude of that value.")

;(setf (arglist TAGBODY) (&QUOTE &REST BODY))
(setf (documentation 'TAGBODY 'function)
  "Allows control to be transferred within BODY.  Each form in BODY is evaluated left to right.
   Numbers and symbols represent tag-ids while lists represent expressions.  Tag-ids are ignored;
   statements are evaluated and their return values ignored.  If a statement has the form
   (GO tag-id) control is transferred to the statements following <tag-id>.  TAGBODY returns NIL.")

;(setf (arglist DEFUN) (&QUOTE &REST ARG))
(setf (documentation 'DEFUN 'function)
  "Macro used to define functions that are part of a program.  Syntax of  
   (DEFUN name lambda-list [doc-string] body) is defined as :  <name> is a symbol that specifies the
   function spec defined as a function.  <lambda-list> contains the keywords and parameters to be
   used by the function.  The constant LAMBDA-LIST-KEYWORKDS defines the allowable keywords.  <body>
   is a list of forms that are executed in sequence whenever FUNCTION is called.  The value of the 
   last form evaluated is returned by FUNCTION.  DEFUN itself returns <name>.")

;(setf (arglist MAPCON) (FCT &REST LISTS))
(setf (documentation 'MAPCON 'function)
  "Returns a list created by NCONCing together the results obtained by applying function FCT to each
   list in LISTS, then successively to the CDR of each of the lists, terminating when the end of the
   shortest list has been reached.  Function FCT must take as many arguments as there are lists
   in LISTS.  The lists returned by FCT are destructively modified by MAPCON.") 

;(setf (arglist UNLESS) (PRED &BODY BODY))
(setf (documentation 'UNLESS 'function)
  "Construct used for conditional execution.  First PRED is evaluated.  If true, no forms are
   evaluated and UNLESS returns NIL.  If PRED returns NIL, all forms in BODY are evaluated as
   in a PROGN and UNLESS returns the value of the last form evaluated.")

;(setf (arglist CDAADR) (X))
(setf (documentation 'CDAADR 'function) 
  "Returns the result of (CDR (CAR (CAR (CDR X)))).")

;(setf (arglist NULL) (X))
(setf (documentation 'NULL 'function)
  "Returns true if X is NIL, otherwise NIL.") 

(setf (documentation 'MOST-NEGATIVE-FIXNUM 'variable)
  "Constant whose value is the greatest negative magnitude representable by the fixnum type.")

;(setf (arglist FUNCTIONP) (FCT &OPTIONAL ALLOW-SPECIAL-FORMS))
(setf (documentation 'FUNCTIONP 'function)
  "Returns true if FCT is a function object, otherwise NIL.  The implementation extension
   argument ALLOW-SPECIAL-FORMS, when true, causes FUNCTIONP to return true for macros
   and special forms in addition to functions.") 

;(setf (arglist PRINT) (OBJECT &OPTIONAL STREAM))
(setf (documentation 'PRINT 'function)
  "Prints the printed representation of OBJECT, preceded by a newline and followed by a space,
   on STREAM with quoting if needed.  PRINT returns OBJECT.")

;(setf (arglist STRING-NOT-GREATERP) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING-NOT-GREATERP 'function)
  "Returns the index into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test less than or equal to ignoring case and fonts.  If the test is never
   satisfied, NIL is returned.  If the strings match, the length of the shortest string is
   returned.  Although the entire length of both STRING1 and STRING2 is the default, a substring
   of STRING1 starting at :START1 and ending at :END1 and/or a substring of STRING2 starting at
   :START2 and ending at :END2 may be specified.")

(setf (documentation '*DEBUG-IO* 'variable)
  "Global variable whose value is the default stream for input and output by the error handler
   and interactive debugging.") 

;(setf (arglist USER-HOMEDIR-PATHNAME) (&OPTIONAL HOST RESET-P USER FORCE-P))
(setf (documentation 'USER-HOMEDIR-PATHNAME 'function) 
  "Returns a pathname describing the user's home directory on host HOST.  The arguments RESET-P,
   USER, and FORCE-P are implementation extensions.  RESET-P, when true, will make HOST the 
   machine that the user is logged in to; USER is the user name that is used.")

;(setf (arglist CHECK-TYPE) (ARG-NAME TYPE &OPTIONAL TYPE-STRING))
(setf (documentation 'CHECK-TYPE 'function) 
  "Generates an error if ARG-NAME is not of type TYPE.  ARG-NAME, the contents of ARG-NAME, and 
   TYPE are mentioned in the error message.  In addition, if TYPE-STRING is specified, it is a 
   string that is used in generating the error message.  This error is correctable.  Continuing
   from the error causes CHECK-TYPE to accept a new value for ARG-NAME and redo the test.
   CHECK-TYPE returns NIL.")

;(setf (arglist QUOTE) (&QUOTE X))
(setf (documentation 'QUOTE 'function)
  "Returns X without evaluating X.  QUOTE can be abbreviated by the #\' character.") 

;(setf (arglist WHEN) (PRED &BODY BODY))
(setf (documentation 'WHEN 'function)
  "Construct used for conditional execution.  First PRED is evaluated.  If NIL, no forms are
   evaluated and WHEN returns NIL.  If PRED returns true, all forms in BODY are evaluated as
   in a PROGN and WHEN returns the value of the last form evaluated.")

;(setf (arglist MULTIPLE-VALUE-CALL) (FUNCTION &QUOTE &REST FORMS))
(setf (documentation 'MULTIPLE-VALUE-CALL 'function)
  "Evaluates FORMS, passing all returned values as the arguments to FUNCTION, which is then 
   executed.  MULTIPLE-VALUE-CALL returns whatever the function FUNCTION returns.") 

(setf (documentation '** 'variable)
  "Variable whose value is bound to the printed result of the read-eval-print loop evaluated
   two interaction ago.")

;(setf (arglist SETF) (&REST PLACES-AND-VALUES))
(setf (documentation 'SETF 'function)
  "Assigns values to symbols.  PLACES-AND-VALUES is a list of alternating place-value pairs.
   The SETF syntax of (SETF place1 exp1 place2 exp2 ... placeN expN) is defined as : evaluate
   exp1 and assign this result to the location determined by place1, evaluate exp2 and assign
   this result to the location determined by place2, and so on.  The value of the last expression
   evaluated is returned by SETF")

;(setf (arglist EXPORT) (SYMBOLS &OPTIONAL PKG))
(setf (documentation 'EXPORT 'function) 
  "Makes SYMBOLS external in package PKG.  If the symbols are not already present in PKG, they are
   imported first.  An error is signaled if this causes a name conflict in any package that uses
   PKG.  EXPORT returns T.")

;(setf (arglist PROG2) (FORM RESULT &BODY BODY))
(setf (documentation 'PROG2 'function)
  "Evaluates FORM, then RESULT, and saves the return value of RESULT.  All forms in BODY are then
   evaluated sequentially.  Finally, PROG2 returns the saved result from the evaluation of RESULT.")

;(setf (arglist STRING-LEFT-TRIM) (CHAR-SET STRING))
(setf (documentation 'STRING-LEFT-TRIM 'function)
  "Returns a substring of STRING with all characters in CHAR-SET removed from the beginning of
   STRING.  CHAR-SET can be a string or list of characters.  Case is presently ignored.")

;(setf (arglist LOAD) (FILE &KEY &OPTIONAL PACKAGE VERBOSE SET-DEFAULT-PATHNAME (IF-DOES-NOT-EXIST T) PRINT))
(setf (documentation 'LOAD 'function) 
  "Loads the specified file or input stream.  If the specified filename has no type field, an attempt
   to load the binary file is first made followed by the Lisp file.  Common Lisp arguments are :
     :VERBOSE - If true, useful information concering the load is printed on *STANDARD-OUTPUT*.
     :PRINT - If true, the value of each expression loaded from the file is printed on
              *STANDARD-OUTPUT*.
     :IF-DOES-NOT-EXIST - If NIL, LOAD returns NIL if the file to be loaded does not exist instead
                          of signaling an error.
   Implementation extensions to the arguments are :
     :PACKAGE - Specifies which package to load the file into.
     :SET-DEFAULT-PATHNAME - If true, the value of the pathname defaults for LOAD are set to the
                             components of the file loaded.") 

;(setf (arglist SUBSTITUTE-IF-NOT) (NEW-ITEM PREDICATE SEQUENCE &KEY KEY START END COUNT FROM-END))
(setf (documentation 'SUBSTITUTE-IF-NOT 'function)
  "Returns a sequence created by replacing in SEQUENCE each element with NEW-ITEM that does not
   satisfies PREDICATE.  Although the entire length of SEQUENCE is the default, a subsequence
   starting at :START and ending at :END may be specified.  If non-NIL, :KEY is a function applied
   to each element of SEQUENCE before the test is performed.  :COUNT is an integer specifying how
   many leftmost elements are to be replaced if more than :COUNT number do not satisfy PREDICATE. 
   :FROM-END is only meaningful if COUNT is specified and, if true,  causes the rightmost :COUNT
   elements that do not satisfy PREDICATE to be replaced.")

(setf (documentation '*PRINT-ESCAPE* 'variable)
  "Variable whose value controls the printing of escape characters.  If *PRINT-ESCAPE* is NIL, the
   escape characters are not output when an expression is printed; if true, they are printed and
   the printed expression can be read back in.") 

;(setf (arglist FILE-AUTHOR) (FILENAME-OR-STREAM))
(setf (documentation 'FILE-AUTHOR 'function) 
  "Returns as a string the creator of the file.  NIL is returned if this cannot be determined.")

;(setf (arglist NSUBSTITUTE-IF-NOT) (NEW-ITEM PREDICATE SEQUENCE &KEY KEY START END COUNT FROM-END))
(setf (documentation 'NSUBSTITUTE-IF-NOT 'function)
  "Returns a sequence created by replacing in SEQUENCE each element that does not satisfies
   PREDICATE with NEW-ITEM.  SEQUENCE is destructively modified by this operation.  Although the
   entire length of SEQUENCE is the default, a subsequence starting at :START and ending at :END
   may be specified.  If non-NIL, :KEY is a function applied to each element of SEQUENCE before
   the test is performed.  :COUNT is an integer specifying how many leftmost elements are to be
   replaced if more than :COUNT number do not satisfy PREDICATE.   :FROM-END is only meaningful
   if :COUNT is specified and, if true, causes the rightmost :COUNT elements that do not satisfy
   PREDICATE to be replaced.")

;(setf (arglist GENSYM) (&OPTIONAL ARG (PERMANENT-P T)))
(setf (documentation 'GENSYM 'function) 
  "Returns a new uninterned symbol with a generated name.  If ARG is a non-negative integer, it is
   used to reset the counter used in generating names; if a string, it becomes the default prefix
   for generated symbols.  The implementation extension argument PERMANENT-P, when true, causes
   the generated symbol to be consed in a permanent area.")

;(setf (arglist FROUND) (DIVIDEND &OPTIONAL (DIVISOR 1)))
(setf (documentation 'FROUND 'function)
  "Returns two values :
     Value1 is DIVIDEND divided by DIVISOR, rounded to the nearest integer but expressed in float
            format, or DIVIDEND rounded to the nearest integer but expressed in float format if
            DIVISOR is not specified.  A number exactly halfway between two numbers rounds to the
            even number.
     Value2 is the number to add to Value1 (or Value1 times DIVISOR if DIVISOR is specified) in
            order to equal DIVIDEND.")

;(setf (arglist CDDADR) (X))
(setf (documentation 'CDDADR 'function) 
  "Returns the result of (CDR (CDR (CAR (CDR X)))).")

(setf (documentation 'PI 'variable)
  "Constant of type long floating-point whose value is pi.")

;(setf (arglist UNCOMPILE) (FUNCTION-SPEC &OPTIONAL DONT-UNENCAPSULATE))
(setf (documentation 'UNCOMPILE 'function) 
  "Replaces the compiled definition of FUNCTION-SPEC with the interpreted definition.  If the
   interpreted definition does not exist, an error message is returned and the definition is
   not updated.")

(setf (documentation '*LOAD-VERBOSE* 'variable)
  "Variable whose value is the default option for the :VERBOSE keyword of the LOAD function. 
   A *LOAD-VERBOSE* value of true permits LOAD to print a message saying what it is loading.")

;(setf (arglist STRING-GREATERP) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING-GREATERP 'function)
  "Returns the index into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test greater than ignoring case and fonts.  If the test is never satisfied, NIL
   is returned.  If the strings match, the length of the shortest string is returned.  Although
   the default string length of both STRING1 and STRING2 is the entire length, a substring of
   STRING1 starting at :START1 and ending at :END1 and/or a substring of STRING2 starting at
   :START2 and ending at :END2 may be specified.")

;(setf (arglist CDADDR) (X))
(setf (documentation 'CDADDR 'function) 
  "Returns the result of (CDR (CAR (CDR (CDR X)))).")

;(setf (arglist COPY-SEQ) (SEQUENCE))
(setf (documentation 'COPY-SEQ 'function) 
  "Returns a copy of SEQUENCE.")

;(setf (arglist TAILP) (TAIL LIST))
(setf (documentation 'TAILP 'function)
  "Returns true if TAIL is a sublist of LIST, otherwise NIL.")

;(setf (arglist UNREAD-CHAR) (CHAR &OPTIONAL STREAM))
(setf (documentation 'UNREAD-CHAR 'function) 
  "Places CHAR back onto the front of STREAM to be read again as the next input character.
   CHAR must be the same character last read from STREAM.  UNREAD-CHAR returns NIL.")

;(setf (arglist DOUBLE-FLOAT) (NUMBER))
;(setf (documentation 'DOUBLE-FLOAT 'function) 
;  "Convert NUMBER to a double precision float.")

;(setf (arglist PACKAGE-NAME) (PKG))
(setf (documentation 'PACKAGE-NAME 'function) 
  "Returns as a string the name of package PKG.")

;(setf (arglist LET*) (&QUOTE VARLIST &REST BODY))
(setf (documentation 'LET* 'function)
  "Special form used to establish bindings of variables.  Syntax of 
   (LET* ((var1 exp1)...(varN expN)) [(DECLARE ...)] body) is defined as : <exp1> is evaluated and
   this value bound to <var1>; <exp2> is evaluated and this value bound to <var2>, and so on till 
   all variables are bound to a value.  If an expression is not specified, NIL is used.  These
   binding are in force only over the scope of the LET*.  All forms in <body> are then evaluated as
   a PROGN with these binding in force; LET* returns the result of the last form evaluated.")

;(setf (arglist REMF) (PLACE PROPERTY))
(setf (documentation 'REMF 'function) 
  "Removes the property value and indicator from the property list stored in PLACE whose indicator
   is EQ to PROPERTY.  The property list stored at PLACE is destructively modified.  REMF returns
   true if the property is found, NIL is no indicator matches PROPERTY.") 

;(setf (arglist EXP) (NUMBER))
(setf (documentation 'EXP 'function) 
  "Returns as a floating-point or complex number e raised to the power NUMBER, where e is the 
   base of the natural logarithms.")

;(setf (arglist <) (NUMBER &REST NUMBERS))
(setf (documentation '< 'function)
  "Returns true if the arguments are in strictly increasing numerical order, otherwise NIL.")

;(setf (arglist DECF) (PLACE &OPTIONAL (DELTA 1)))
(setf (documentation 'DECF 'function) 
  "Decrements the value stored at PLACE by DELTA amount, which defaults to 1, and stores the result back
   into PLACE.")

;(setf (arglist DIGIT-CHAR-P) (CHAR &OPTIONAL (RADIX 10)))
(setf (documentation 'DIGIT-CHAR-P 'function)
  "Returns as a non-negative integer the magnitude of CHAR if CHAR is a digit of radix RADIX,
   otherwise NIL.  If the bit attribute of CHAR is nonzero, NIL is returned.") 

;(setf (arglist 1+) (N))
(setf (documentation '1+ 'function)
  "Returns the value of N plus 1.") 

;(setf (arglist STRING-EQUAL) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING-EQUAL 'function)
  "Returns true if STRING1 is equal to STRING2.  Case and font are not considered.  Although
   the default length for both STRING1 and STRING2 is the entire string, a substring of STRING1
   starting at :START1 and ending at :END1 and/or a substring of STRING2 starting at :START2 and
   ending at :END2 may be specified.  NIL is returned if the string are not equal.") 

(setf (documentation 'CHAR-CONTROL-BIT 'variable)
  "Constant whose value gives a numerical meaning to the control bit in the bit configuration for
   a character object.")

;(setf (arglist MAKE-SEQUENCE) (TYPE SIZE &KEY INITIAL-ELEMENT))
(setf (documentation 'MAKE-SEQUENCE 'function) 
  "Returns a sequence of type TYPE whose length is SIZE.  If specified, :INITIAL-ELEMENT must be
   of type TYPE and is the value that each element of the new sequence is set to.")

;(setf (arglist FUNCTION) (&QUOTE FUNCTION))
(setf (documentation 'FUNCTION 'function)
  "Returns the function definition of the argument FUNCTION if the argument is a symbol or
   functional specification.  If the argument FUNCTION is a lambda expression, FUNCTION returns
   a function that, when invoked, will execute the body of the lambda expression properly observing
   the lexical scoping rules.")

;(setf (arglist MISMATCH) ((SEQUENCE1 SEQUENCE2 &KEY FROM-END TEST TEST-NOT KEY (START1 0) END1 (START2 0) END2)))
(setf (documentation 'MISMATCH 'function)
  "Returns the index in SEQUENCE1 of the first mismatch between SEQUENCE1 and SEQUENCE2, comparing
   successive elements of each.  If the sequences match, NIL is returned.  If the sequences match
   but are of different lengths, the index relative to SEQUENCE1 where comparison stopped is
   returned.  Although the entire length of both SEQUENCE1 and SEQUENCE2 is the default, a
   subsequence of SEQUENCE1 starting at :START1 and ending at :END1 and/or a subsequence of SEQUENCE2
   starting at :START2 and ending at :END2 may be specified.  :TEST is a predicate which takes two
   arguments and returns true in order to continue processing; :TEST-NOT is a predicate which takes
   two arguments and returns NIL in order to continue processing.  If specified, :KEY is a function
   taking one argument that is applied to the elements of each sequence before the test is made. 
   If true, :FROM-END specifies that the index plus one of the rightmost element that fails the test
   should be returned.  This index is relative to SEQUENCE1.")

(setf (documentation 'BOOLE-ANDC1 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-and of the complement of 
   BOOLE's second argument with BOOLE's third argument.")

;(setf (arglist ABS) (NUMBER))
(setf (documentation 'ABS 'function) 
  "Returns the absolute value of NUMBER, which can be of any number type.")

;(setf (arglist FILL-POINTER) (ARRAY))
(setf (documentation 'FILL-POINTER 'function) 
  "Returns the fill pointer of ARRAY.  Common Lisp suggests signaling an error if ARRAY does not
   have a fill pointer.  An implementation extension will cause FILL-POINTER to return NIL if 
   ARRAY does not have a fill pointer.")

;(setf (arglist INCF) (PLACE &OPTIONAL (DELTA 1)))
(setf (documentation 'INCF 'function)
  "Increments the value stored at PLACE by DELTA amount, which defaults to 1, and stores the result
   back into PLACE.")

;(setf (arglist CLEAR-OUTPUT) (&OPTIONAL STREAM))
(setf (documentation 'CLEAR-OUTPUT 'function) 
  "Discards any buffered output and aborts any outstanding output operations on stream STREAM.
   CLEAR-OUTPUT returns NIL.")

;(setf (arglist BIT) (BITARRAY &REST SUBSCRIPTS))
(setf (documentation 'BIT 'function)
  "Returns the element of bit array BITARRAY whose subscripts are specified by SUBSCRIPTS.")

;(setf (arglist DEFTYPE) (NAME ARGLIST &BODY BODY))
(setf (documentation 'DEFTYPE 'function)
  "Construct used to define new type-specifiers.  NAME is the symbol that identifies the type
   specifier.  ARGLIST is a lambda-list which may contain the &OPTIONAL and &REST arguments.
   If a default value is not given to optional arguments, they default to *.  A documentation
   string may also be supplied and is associated with NAME.  BODY is a set of forms to be
   evaluated using the values in ARGLIST to return the new type specifier.  DEFTYPE returns NAME.") 

;(setf (arglist REDUCE) (FUNCTION SEQUENCE &KEY (INITIAL-VALUE NIL INITIAL-VALUE-P) START END FROM-END))
(setf (documentation 'REDUCE 'function)
  "Returns the result of applying FUNCTION to the first two elements of SEQUENCE, then applying
   FUNCTION to that result and the third element of SEQUENCE, and so on, until all elements of
   SEQUENCE have been processed.  FUNCTION is a function accepting two arguments.  Although the
   entire length of SEQUENCE is the default, a subsequence of SEQUENCE starting at :START and 
   ending at :END may be specified.  If true, :FROM-END specifies that FUNCTION be applied to
   the elements right to left instead of left to right.  If :INITIAL-VALUE is specified, it is
   made the first element of the sequence that FUNCTION is applied to, being either the leftmost
   element if :FROM-END is NIL or the rightmost element if :FROM-END is true.  If there is only one
   element to be processed, that element is returned and FUNCTION is not called.  If there are 
   no elements (SEQUENCE is of length zero and no :INITIAL-VALUE),  FUNCTION is called with no
   arguments and FUNCTION's return value is returned by REDUCE.")

;(setf (arglist CDDDDR) (X))
(setf (documentation 'CDDDDR 'function) 
  "Returns the result of (CDR (CDR (CDR (CDR X)))).")

;(setf (arglist SBIT) (BITARRAY &REST SUBSCRIPTS))
(setf (documentation 'SBIT 'function)
  "Returns the element of simple bit array BITARRAY whose subscripts are specified by SUBSCRIPTS.")

(setf (documentation '*QUERY-IO* 'variable)
  "Global variable whose value is the default stream for asking question of the user.")

;(setf (arglist DEFCONSTANT) (VARIABLE INITIAL-VALUE &OPTIONAL DOCUMENTATION))
(setf (documentation 'DEFCONSTANT 'function) 
  "Construct used to define a special variable whose value will never be changed.  DEFCONSTANT
   proclaims VARIABLE to be special, evaluates INITIAL-VALUE, and assigns that value to VARIABLE.
   DOCUMENTATION is a string describing VARIABLE.  DEFCONSTANT returns VARIABLE.")

;(setf (arglist FMAKUNBOUND) (SYMBOL))
(setf (documentation 'FMAKUNBOUND 'function) 
  "Causes SYMBOL to have no function definition.  FMAKUNBOUND returns SYMBOL.")

;(setf (arglist PHASE) (NUMBER))
(setf (documentation 'PHASE 'function) 
  "Returns the phase of the complex number NUMBER, in radians.  Phase is the angle in the complex
   plane from the positive real axis to the ray from the origin through NUMBER.  It is between 
   -pi (exclusive) and pi (inclusive).  For a positive real number PHASE returns 0.0; for a negative
   real number, PHASE returns pi; for zero, PHASE returns 0.0.")

;(setf (arglist REMOVE) (ITEM SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END COUNT FROM-END))
(setf (documentation 'REMOVE 'function)
  "Returns a list where those elements in SEQUENCE that match ITEM have been removed.  An element
   matches ITEM if the test :TEST returns a true value or the test :TEST-NOT returns NIL.  Although
   the entire length of SEQUENCE is the default, a subsequence starting at :START and ending at :END
   may be specified.  If non-NIL, :KEY specifies a function to be applied to each element of SEQUENCE
   before the test is performed.  :COUNT is an integer specifying how many leftmost elements are to
   be removed if more than :COUNT number match ITEM.  FROM-END is only meaningful if :COUNT is
   specified and, if true, causes the rightmost :COUNT elements that match ITEM to be removed.")

;(setf (arglist STRING-NOT-EQUAL) (STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2))
(setf (documentation 'STRING-NOT-EQUAL 'function)
  "Returns the index into STRING1 of the first element that satisfies the test not equal when
   compared to STRING2.  Case and font are not considered.  Although the default length for both
   STRING1 and STRING2 is the entire string, a substring of STRING1 starting at :START1 and ending
   at :END1 and/or a substring of STRING2 starting at :START2 and ending at :END2 may be specified.
   NIL is returned if the strings are equal.") 

;(setf (arglist COPY-ALIST) (AL))
(setf (documentation 'COPY-ALIST 'function)
  "Returns an association list that is a copy of AL.  Dotted pair termination of list will also
   be copied.")

(setf (documentation '*PRINT-BASE* 'variable)
  "Variable whose value is the radix used to print integers and ratios.") 

(setf (documentation '*PRINT-LEVEL* 'variable)
  "Variable whose value, when non-NIL, is an integer specifying how many levels deep a nested
   object will be printed.  Any structure at a level lower that *PRINT-LEVEL* is replaced by
   #.  A NIL value for *PRINT-LEVEL* specifies no control.")

(setf (documentation 'LAMBDA-LIST-KEYWORDS 'variable)
  "Constant whose value is a list of all available lambda-list keywords in the current Lisp
   environment.")

;(setf (arglist LOOP) (LOOP-SOURCE-CODE &OPTIONAL ENVIRONMENT))
(setf (documentation 'LOOP 'function)
  "Evaluates each form in LOOP-SOURCE-CODE in order from left to right.  When the last form has
   been evaluated execution starts again with the first form in LOOP-SOURCE-CODE.  This continues
   until some specific control transfer such as RETURN or THROW is executed.  LOOP returns no 
   value.") 

;(setf (arglist LOGAND) (&REST INTEGERS))
(setf (documentation 'LOGAND 'function)
  "Returns the bitwise-and of INTEGERS, or -1 if no arguments are given.")

;(setf (arglist CLRHASH) (HASH-TABLE &OPTIONAL IGNORE))
(setf (documentation 'CLRHASH 'function) 
  "Removes all entries from HASH-TABLE and returns HASH-TABLE.  HASH-TABLE is destructively modified.")

;(setf (arglist DOCUMENTATION) (SYMBOL &OPTIONAL (DOC-TYPE (QUOTE FUNCTION))))
(setf (documentation 'DOCUMENTATION 'function) 
  "Returns the documentation string for SYMBOL, NIL if none exist.  The standard values of
   DOC-TYPE are function, variable, type, structure, and setf.")

;(setf (documentation 'DOCUMENTATION 'variable) 
;  NIL)

;(setf (arglist MOD) (DIVIDEND MODULUS))
(setf (documentation 'MOD 'function) 
  "Returns the value of DIVIDEND modulo MODULUS.  It is the same as the second value returned
   by the form (FLOOR DIVIDEND MODULUS).  The result will be in the range from zero (inclusive)
   to MODULUS (exclusive) with the same sign as MODULUS.")

;(setf (arglist SIGNUM) (Z))
(setf (documentation 'SIGNUM 'function) 
  "Returns a number of the same type, of unit magnitude, and with the same sign as Z.  If Z is
   equal to 0, SIGNUM returns 0.")

;(setf (arglist FIXNUMP) (OBJECT))
(setf (documentation 'FIXNUMP 'function) 
  "Returns true if OBJECT is a fixnum, otherwise NIL.")

;(setf (arglist BYTE) (WIDTH POSITION))
(setf (documentation 'BYTE 'function) 
  "Returns a byte specifier for a byte of WIDTH bits starting at POSITION bits from the least
   signifiant bit.  This return value is suitable for use as an argument to the byte manipulation
   functions.")

;(setf (arglist DEFPARAMETER) (VARIABLE INITIAL-VALUE . ARGS))
(setf (documentation 'DEFPARAMETER 'function)
  "Defines and initializes a special variable named VARIABLE to the value INITIAL-VALUE.  ARGS can
   take on the value of the documentation string for VARIABLE.  VARIABLE is assigned the value 
   INITIAL-VALUE each time the DEFPARAMETER is executed, even if VARIABLE already has a value.
   DEFPARAMETER returns VARIABLE.")

;(setf (arglist PUSHNEW) (VALUE PLACE &REST OPTIONS))
(setf (documentation 'PUSHNEW 'function) 
  "Adds VALUE to the front of the list stored at PLACE if VALUE is not already a member of the
   list.  Not only is the new list returned by PUSHNEW, but PUSHNEW also stores this new list back
   into PLACE.  OPTIONS can assume the values :
     :TEST - A function taking two arguments which should return true when VALUE is compared
             against the elements of the list at PLACE and VALUE should not be added.
     :TEST-NOT - A function taking two arguments which should return NIL when VALUE is compared
                 against the elements of the list at PLACE and VALUE should not be added.
     :KEY - A function taking one argument that is applied not only to each element of the list
            stored at PLACE but also to VALUE before the test is performed.") 

;(setf (arglist PATHNAME-PLIST) (P))
(setf (documentation 'PATHNAME-PLIST 'function) 
  "Returns the list of properties of the parsed pathname P, NIL if none exist.")

;(setf (arglist STEP) (FORM &OPTIONAL *STEP-AUTO*))
(setf (documentation 'STEP 'function) 
  "Evaluates FORM, allowing the user to interactively step through this evaluation of FORM.") 

;(setf (arglist TYPECASE) (OBJECT &BODY CLAUSES))
(setf (documentation 'TYPECASE 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (TYPECASE OBJECT (type1 body1) ... (typeN bodyN)) is defined as :  the (type# body#) construct
   is called a clause.  OBJECT is evaluated.  <type#> is a literal type specifier and is not
   evaluated.   <type1> is compared against OBJECT's type; if they the same, all forms in <body1>
   are evaluated sequentially and the value of the last expression in <body1> is returned by
   TYPECASE.  If <type1> does not match OBJECT's type, <type2> is compared to OBJECT's type, and
   so on, until either one <type#> matches OBJECT's type or all clauses are exhausted.  An OTHERWISE
   or T as the <typeN> may be included and will always be executed if no other <type#> matches OBJECT.
   If no clause is selected, TYPECASE returns NIL.")

(setf (documentation 'LEAST-POSITIVE-SINGLE-FLOAT 'variable)
  "Constant whose value is the smallest positive magnitude representable by the single float type.") 

;(setf (arglist NSUBSTITUTE-IF) (NEW-ITEM PREDICATE SEQUENCE &KEY KEY START END COUNT FROM-END))
(setf (documentation 'NSUBSTITUTE-IF 'function)
  "Returns a sequence created by replacing in SEQUENCE each element that satisfies PREDICATE with
   NEW-ITEM.  Sequence is destructively modified by NSUBSTITUTE-IF.  Although the entire length of
   SEQUENCE is the default, a subsequence starting at :START and ending at :END may be specified. 
   If non-NIL, :KEY is a function applied to each element of SEQUENCE before the test is performed.
   :COUNT is an integer specifying how many leftmost elements are to be replaced if more than :COUNT
   number satisfy PREDICATE.  :FROM-END is only meaningful if :COUNT is specified and, if true,
   causes the rightmost :COUNT elements that satisfy PREDICATE to be replaced.")

;(setf (arglist ARRAY-ROW-MAJOR-INDEX) (ARRAY &REST SUBSCRIPTS))
(setf (documentation 'ARRAY-ROW-MAJOR-INDEX 'function) 
  "Returns as a non-negative integer the combined index in ARRAY of the element identified by
   SUBSCRIPTS.  Row-major order is used.")

;(setf (arglist REALPART) (X))
(setf (documentation 'REALPART 'function) 
  "Returns the real part of complex number X.  If X is not complex, REALPART returns X.")

;(setf (arglist ROW-MAJOR-AREF) (ARRAY INDEX))
(setf (documentation 'ROW-MAJOR-AREF 'function)
  "Returns the element of ARRAY whose index is INDEX when ARAAY is considered to be a one
   dimensional array stored in row-major order.") 

;(setf (arglist REMOVE-IF) (PREDICATE SEQUENCE &KEY KEY START END COUNT FROM-END))
(setf (documentation 'REMOVE-IF 'function)
  "Returns a sequence where those elements in sequence SEQUENCE that satisfy test PREDICATE have
   been removed.  Although the entire length of SEQUENCE is the default, a subsequence starting at
   :START and ending at :END may be specified.  If non-NIL, :KEY is a funcion applied to each element
   of SEQUENCE before the test is performed.  :COUNT is an integer specifying how many leftmost
   elements are to be removed if more than :COUNT number satisfy the test.  :FROM-END is only
   meaningful if :COUNT is specified and, if true, causes the rightmost :COUNT elements that satisfy
   the test to be removed.")

(setf (documentation '*PRINT-GENSYM* 'variable)
  "Variable whose value, when true, causes uninterned symbols to be printed with a #: prefix.") 

;(setf (arglist CERROR) (CONTINUE-FORMAT-STRING ERROR-FORMAT-STRING &REST ARGS))
(setf (documentation 'CERROR 'function)
  "Signals an error and provides a single way to proceed from the error.  CONTINUE-FORMAT-STRING
   is a control string given to FORMAT to inform the user of the effects of continuing from the
   error.  ERROR-FORMAT-STRING is a control string given to FORMAT and describes the error.  ARGS
   are the format arguments which the control strings use if required.")

;(setf (arglist CHAR-UPCASE) (CHAR))
(setf (documentation 'CHAR-UPCASE 'function)
  "Returns the upper case version of CHAR with the same bit and font attributes of CHAR if CHAR
   is an alphabetic character.  If CHAR is not alphabetic, CHAR-UPCASE returns CHAR unchanged.
   Note: if the bits attribute of CHAR is non-zero, it is not considered alphabetic.")

;(setf (arglist EQUAL) (X Y))
(setf (documentation 'EQUAL 'function)
  "Returns true if X is equal to Y.  Numbers are equal if they have the same value and type. 
   Strings are compared element by element with case being considered.  Arrays are considered
   equal only if they are EQ.  Conses are equal if the two CAR's are equal and the two CDR's
   are equal.")  

;(setf (arglist MAKE-CONCATENATED-STREAM) (&REST STREAMS))
(setf (documentation 'MAKE-CONCATENATED-STREAM 'function) 
  "Returns a stream which will read from the first stream in STREAMS until end of file is reached,
   then read from the second stream in STREAMS until end of file is reached, and so on.  When end
   of file is reached on the last stream in STREAMS, or if STREAMS is not specified, any input 
   attempt will result in end of file.")

;(setf (arglist PROVIDE) (MODULE))
(setf (documentation 'PROVIDE 'function)
  "Adds MODULE to the list of currently loaded modules contained in variable *MODULES*.  This 
   marks MODULE as being loaded.")

;(setf (arglist SUBSETP) (LIST1 LIST2 &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'SUBSETP 'function)
  "Returns true if every element of LIST1 is an element of LIST2.  :TEST is a predicate that 
   takes two arguments and should return true when an element in LIST1 is to be considered an
   element in LIST2 when compared against all elements in LIST2.  :TEST-NOT is a predicate that
   takes two arguments and should return NIL for the elements to match.  :KEY, if specified, is
   a function applied to the elements of each list before the test is performed.") 

;(setf (arglist NSUBSTITUTE) (NEW-ITEM ITEM SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END COUNT FROM-END))
(setf (documentation 'NSUBSTITUTE 'function)
  "Returns a sequence created by replacing in SEQUENCE each instance of ITEM with NEW-ITEM.
   SEQUENCE is destructively modified by NSUBSTITUTE.   Although the entire length of SEQUENCE
   is the default, a subsequence starting at :START and ending at :END may be specified.  :TEST
   is a predicate that should return true for ITEM to be replaced.  If specified, :TEST-NOT is a
   predicate which should return NIL in order for ITEM to be replaced.  If non-NIL, :KEY is a
   function applied to each element of SEQUENCE before the test is performed.  :COUNT is an integer
   specifying how many leftmost elements are to be replaced if more than :COUNT number match ITEM.
   :FROM-END is only meaningful if :COUNT is specified and, if true, causes the rightmost :COUNT
   elements that match ITEM to be replaced.")

;(setf (arglist SHADOW) (SYMBOLS &OPTIONAL PKG))
(setf (documentation 'SHADOW 'function) 
  "Makes the symbols SYMBOLS in package PKG internal symbols.  If a symbol in SYMBOLS is already
   present in PKG, no action is taken; if not present, the symbol is created, interned in PKG, and
   added to the shadowing symbols list.  SHADOW returns T.")

;(setf (arglist NOTANY) (PREDICATE &REST SEQUENCES))
(setf (documentation 'NOTANY 'function)
  "Returns true if PREDICATE returns NIL when applied to every element of SEQUENCES.  If some
   application of PREDICATE returns true, then NOTANY returns NIL.  If multiple sequences are
   supplied, element 0 of each sequence is tested, then element 1, and so on.  PREDICATE must
   accept the same number of arguments as sequences provided.  If one of the sequences is
   exhausted before PREDICATE returns true, NOTANY returns true.")

;(setf (arglist IMPORT) (SYMBOLS &OPTIONAL PKG))
(setf (documentation 'IMPORT 'function)
  "Interns each element of SYMBOLS in package PKG.  If a symbol in SYMBOLS is already in package
   PKG, no action is taken; if accessible through inheritance, it is made internal; if not 
   present, it is created and interned in PKG; if a symbol already exists with this name, a
   correctable error is signaled.  IMPORT returns T.") 

;(setf (arglist FIND-PACKAGE) (OBJECT))
(setf (documentation 'FIND-PACKAGE 'function) 
  "Return the package object whose name or one of whose nicknames is OBJECT.  If no such package
   exists, NIL is returned.  OBJECT can be a symbol, a case-sensitive string, or, as an 
   implementation extension, a package object.")

;(setf (arglist LOGORC1) (INTEGER1 INTEGER2))
(setf (documentation 'LOGORC1 'function)
  "Returns the bitwise-or of the complement of INTEGER1 with INTEGER2.") 

;(setf (arglist CONSTANTP) (FORM))
(setf (documentation 'CONSTANTP 'function) 
  "Returns true if FORM always evaluates to the same thing.  This includes but is not limited to
   keyword symbols, numbers, and characters.")

;(setf (arglist MAX) (&REST NUMBERS))
(setf (documentation 'MAX 'function) 
  "Returns the argument with the largest value.")

;(setf (arglist BIT-NOR) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-NOR 'function)
  "Returns the bit-array generated from a bitwise-nor of BIT-ARRAY1 with BIT-ARRAY2.  If
   RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted, the result is
   returned as a new bit-array; if a bit-array, the result is destructively placed into it.")

;(setf (arglist COMPILE-FILE) (INPUT-FILE &KEY OUTPUT-FILE LOAD VERBOSE SET-DEFAULT-PATHNAME PACKAGE
;                              DECLARE TARGET SUPPRESS-DEBUG-INFO MERCILESS))
(setf (documentation 'COMPILE-FILE 'function) 
  "Compiles source file INPUT-FILE into object file OUTPUT-FILE.  :OUTPUT-FILE defaults are based
   upon INPUT-FILE and *LOAD-PATHNAME-DEFAULTS*.  Implementation extension arguments are :
      :LOAD - If true, load the output file after compiling.
      :VERBOSE - If true, print the name of each function as it is compiled.
      :SET-DEFAULT-PATHNAME - If true, set the default pathname components from INPUT-FILE.
      :PACKAGE - The name of the package in which the file is to compiled.
      :DECLARE - Contains a list of declaration specifiers, such as optimization levels.
      :TARGET - The name of the machine for which code will be generated.
      :SUPPRESS-DEBUG-INFO - If true, do not record debug info in the object file.
   COMPILE-FILE returns two values.  Value1 is output file pathname.  Value2 is a status code
   equal to one of the following constants:  COMPILER:OK, COMPILER:WARNINGS, COMPILER:ERRORS,
   or COMPILER:FATAL.")

;(setf (arglist NUNION) (LIST1 LIST2 &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT))
(setf (documentation 'NUNION 'function)
  "Returns a list whose elements are a member of at least one of the lists.  Both LIST1 and LIST2
   are destructively modified by NUNION.  :TEST specifies the predicate that, when true is returned,
   causes only one of the two elements that tested as equal to be added to the returned list;
   :TEST-NOT is the predicate that, when NIL is returned, causes only one of the two elements that
   tested as equal to be added to returned list.  If :KEY is specified, it is a function applied to
   each member of each list before the test is performed.")

;(setf (arglist MULTIPLE-VALUE-PROG1) (&QUOTE VALUE-FORM &REST FORMS))
(setf (documentation 'MULTIPLE-VALUE-PROG1 'function)
  "Evaluates VALUE-FORM and saves all results.  All forms in FORMS are then evaluated sequentially.
   Finally, MULTIPLE-VALUE-PROG1 returns all saved results from the evaluation of VALUE-FORM.")

;(setf (arglist FIFTH) (LIST))
(setf (documentation 'FIFTH 'function)
  "Returns the fifth element of LIST.  One-origin numbering is used.")

;(setf (arglist THROW) (TAG &QUOTE VALUE-EXPRESSION))
(setf (documentation 'THROW 'function) 
  "Transfers control out of a matching CATCH.  After TAG is evaluated to produce a throw tag, 
   VALUE-EXPRESSION is evaluated and all of its values saved.  A search of the stack is made to
   find the innermost CATCH that is EQ to the throw tag, that CATCH construct is exited, and
   the saved values are returned as the value(s) of the CATCH.")

;(setf (arglist =) (&REST NUMBERS))
(setf (documentation '= 'function) 
  "Return true if all arguments are numerically equal, otherwise NIL.")

(setf (documentation 'LEAST-NEGATIVE-LONG-FLOAT 'variable)
  "Constant whose value is the least negative magnitude representable by the long float type.")

;(setf (arglist LIST-LENGTH) (LIST))
(setf (documentation 'LIST-LENGTH 'function) 
  "Returns as an integer the length of LIST, or NIL if LIST is circular.")

;(setf (arglist MAPC) (FCT &REST LISTS))
(setf (documentation 'MAPC 'function)
  "Applies function FCT to the successively elements of the input lists LISTS.  Function MAPC
   must take as many arguments as there are lists provided and terminates when the shortest list
   is exhausted.  MAPC does not retain the results, being called mainly for side effects, and
   returns the first list in LISTS as its return value.")

;(setf (arglist ISQRT) (N))
(setf (documentation 'ISQRT 'function) 
  "Returns the greatest integer less than or equal to the square root of N, where N is a
   non-negative integer.")

;(setf (arglist CEILING) (DIVIDEND &OPTIONAL (DIVISOR 1)))
(setf (documentation 'CEILING 'function)
  "Returns two values :
     Value1 is the integer that is the smallest integer value possible that is not smaller than
            DIVIDEND if DIVISOR is not specified or smaller than the quotient of DIVIDEND divided
            by DIVISOR if DIVISOR is specified.
     Value2 is the number to add to Value1 (or Value1 times DIVISOR if DIVISOR is specified) in
            order to equal DIVIDEND.")

;(setf (arglist FLOAT-DIGITS) (FLOAT))
(setf (documentation 'FLOAT-DIGITS 'function) 
  "Returns the number of bits used to represent the fractional part of FLOAT, including hidden
   bits.  This number depends only on the data type of FLOAT, that is, short-float, single-float,
   or double-float.")

;(setf (arglist NTHCDR) (N LIST))
(setf (documentation 'NTHCDR 'function)
  "Returns a list created by applying the function CDR N times to LIST, analogous to removing the
   first N elements of LIST.  Zero-origin numbering is used.") 

;(setf (arglist WITH-OPEN-STREAM) ((STREAM CONSTRUCTION-FORM) &BODY BODY))
(setf (documentation 'WITH-OPEN-STREAM 'function)
  "Evaluates BODY with the variable STREAM bound to a stream named by the value of CONSTRUCTION-FORM.
   When control leaves WITH-OPEN-STREAM normally, the stream is closed and the value of the last
   form in BODY is returned as WITH-OPEN-STREAM's return value;  if the exit is abnormal, the stream
   is close with the :ABORT option of CLOSE.")

(setf (documentation 'BOOLE-ANDC2 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-and of BOOLE's second
   argument with the complement of BOOLE's third argument.")

;(setf (arglist FILE-NAMESTRING) (OBJECT))
(setf (documentation 'FILE-NAMESTRING 'function)
  "Returns as a namestring the name, type, and version of OBJECT, after OBJECT has been converted
   to a pathname.") 

;(setf (arglist ASH) (N NBITS))
(setf (documentation 'ASH 'function) 
  "Returns an integer created by shifting N to the left NBITS bit positions if NBITS is positive
   or to the right -NBITS bit positions if NBITS is negative.  The sign of the result is always
   the same as the sign of N.")

;(setf (arglist MAKE-PATHNAME) (&KEY &OPTIONAL (DEFAULTS T) HOST DEVICE RAW-DEVICE DIRECTORY RAW-DIRECTORY
;                                NAME RAW-NAME TYPE RAW-TYPE VERSION CANONICAL-TYPE ORIGINAL-TYPE))
(setf (documentation 'MAKE-PATHNAME 'function)
  "Returns a pathname created by specifying the components as keyword arguments.  After the 
   components whose values are specified by :HOST, :DEVICE, :RAW-DEVICE, :DIRECTORY, :RAW-DIRECTORY,
   :NAME, :RAW-NAME, :TYPE, :RAW-TYPE, and :VERSION are filled in, the values given in :DEFAULTS are
   used.  If DEFAULTS is T, the variable *DEFAULT-PATHNAME-DEFAULTS* is used for the default value
   of the host component while the remaining components are not defaulted at all.  The keywords
   whose names are prefixed by RAW specify the raw format of those components.  :CANONICAL-TYPE
   specifies the canonical form of the pathname to return.  :ORIGINAL-TYPE specifies the original
   type of the pathname to return.")  

;(setf (arglist APPEND) (&REST LISTS))
(setf (documentation 'APPEND 'function)
  "Returns a list created by concatenating together all arguments in LISTS.  All arguments but
   the last must be lists; the last may be any Lisp object.  The top level structure of all but
   the last argument is copied.") 

;(setf (arglist REMOVE-IF-NOT) (PREDICATE SEQUENCE &KEY KEY START END COUNT FROM-END))
(setf (documentation 'REMOVE-IF-NOT 'function)
  "Returns a sequence where those elements in sequence SEQUENCE that do not satisfy test PREDICATE
   have been removed.  Although the entire length of SEQUENCE is the default, a subsequence starting
   at :START and ending at :END may be specified.  If non-NIL, :KEY is a funcion applied to each
   element of SEQUENCE before the test is performed.  :COUNT is an integer specifying how many
   leftmost elements are to be removed if more than :COUNT number do not satisfy the test.  :FROM-END
   is only meaningful if :COUNT is specified and, if true, causes the rightmost :COUNT elements that
   do not satisfy the test to be removed.")

;(setf (arglist SYMBOL-PACKAGE) (SYMBOL))
(setf (documentation 'SYMBOL-PACKAGE 'function) 
  "Returns the contents of the package cell of symbol SYMBOL.  This is a package object or NIL.")

;(setf (arglist TYPEP) (OBJECT TYPE))
(setf (documentation 'TYPEP 'function)
  "Returns true if OBJECT is of type TYPE, otherwise NIL.  An obsolete mode of use is with one
   argument; in this mode TYPEP returns the type of OBJECT.")

;(setf (arglist COMMONP) (OBJECT))
(setf (documentation 'COMMONP 'function)
  "Returns true if OBJECT is any standard Common Lisp type, otherwise NIL.") 

;(setf (arglist INTEGERP) (X))
(setf (documentation 'INTEGERP 'function) 
  "Returns true if X is an integer, otherwise NIL.")

;(setf (arglist COMPILER-LET) (&QUOTE BINDLIST &REST BODY))
(setf (documentation 'COMPILER-LET 'function) 
  "Performs bindings in BINDLIST at evaluation or compilation time.  In interpreted code,
   COMPILER-LET is the same as LET.  When compiled, however, the evaluation of the forms in BODY
   is done with the variables in BINDLIST bound at compile time in the execution environment of
   the compiler, not at execution time.")

;(setf (arglist EQ) (X Y))
(setf (documentation 'EQ 'function) 
  "Returns true if X and Y are the same identical Lisp object, otherwise NIL.")

;(setf (arglist IN-PACKAGE) (NAME &KEY NICKNAMES (USE (QUOTE (LISP TICL))) (SIZE 200) SHADOW EXPORT))
(setf (documentation 'IN-PACKAGE 'function) 
  "Creates package NAME with nicknames :NICKNAMES if that package does not already exist.  If it 
   exists, the package is reloaded; if it does not exist, package NAME is created with size :SIZE,
   inheriting the symbols from the packages specified by :USE.  If :SHADOW or :EXPORT are true, the
   arguments to these keywords are passed to the SHADOW and EXPORT function, respectively, for
   processing.  The package object created is assigned to *PACKAGE* and returned by IN-PACKAGE.")

(setf (documentation 'LAMBDA-PARAMETERS-LIMIT 'variable) 
  "Constant whose value is the upper exclusive bound on the number of different parameters allowed
   in one lambda list.")

;(setf (arglist SCHAR) (STRING INDEX))
(setf (documentation 'SCHAR 'function)
  "Returns the character from the simple string STRING whose index is INDEX.") 

;(setf (arglist GET-INTERNAL-REAL-TIME) NIL)
(setf (documentation 'GET-INTERNAL-REAL-TIME 'function)
  "Returns time in 60ths of a second since the last boot.")

(setf (documentation 'BOOLE-NAND 'variable)
  "Constant used as first argument to function BOOLE to take the bitwise-nand of BOOLE's second
   argument with BOOLE's third argument.")

;(setf (arglist MAPHASH) (FUNCTION HASH-TABLE &REST EXTRA-ARGS))
(setf (documentation 'MAPHASH 'function) 
  "Applies FUNCTION to each item in HASH-TABLE.  FUNCTION's arguments are the key, the values
   associated with that key, and the implementation extension argument EXTRA-ARGS.  Only a REMHASH
   on the entry currently being processed or a SETF of GETHASH on the current entry can modify 
   HASH-TABLE without causing unpredictable results.  MAPHASH returns NIL.")  

;(setf (arglist WRITE-CHAR) (CHAR &OPTIONAL STREAM))
(setf (documentation 'WRITE-CHAR 'function) 
  "Outputs CHAR to stream STREAM.  WRITE-CHAR returns CHAR.")

;(setf (arglist PROGN) (&QUOTE &REST BODY))
(setf (documentation 'PROGN 'function)
  "Evaluates all forms in BODY, ignoring the values returned by each form but the last.  PROGN
   returns all values returned by the last form.")

;(setf (arglist READ) (&OPTIONAL (STREAM *STANDARD-INPUT*) (EOF-ERRORP T) (EOF-VALUE NIL) (RECURSIVEP NIL)
;                      PRESERVE-WHITESPACE DISCARD-CLOSE-PARENS CHECK-INDENT))
(setf (documentation 'READ 'function)
  "Returns one Lisp object after reading the printed representation of that object from STREAM.
   End of file is an error if EOF-ERRORP is true; otherwise, READ returns EOF-VALUE when end of
   file is read.  The argument RECURSIVE-P, if specified and true, signals that this is not
   a top level call to READ but an imbedded call.  The remaining arguments are implementation
   extensions.  PRESERVE-WHITESPACE, if true, will cause any whitespace token used to terminate
   the printed representation to be retained.  DISCARD-CLOSE-PARENS, if true, specifies that
   reading is to continue past a closing parens rather than consider the expression terminated.
   CHECK-INDENT, if true, will cause indentation to be checked in the symbolic expression.")  

;(setf (arglist YES-OR-NO-P) (&OPTIONAL FORMAT-STRING &REST FORMAT-ARGS))
(setf (documentation 'YES-OR-NO-P 'function)
  "Asks the user a question which can be answered with yes or no.  YES-OR-NO-P is case insensitive.
   If FORMAT-STRING is present and non-NIL, FORMAT-STRING and FORMAT-ARGS are printed as if they
   were given to function FORMAT.  YES-OR-NO-P returns true if the answer is yes, otherwise NIL.")

(setf (documentation 'BOOLE-C1 'variable)
  "Constant used as first argument to function BOOLE to take the complement of BOOLE's second
   argument.")

;(setf (arglist SET-DISPATCH-MACRO-CHARACTER) (DISP-CHAR SUB-CHAR FUNCTION &OPTIONAL (RT *READTABLE*)))
(setf (documentation 'SET-DISPATCH-MACRO-CHARACTER 'function) 
  "Sets the syntax of the two character sequence DISP-CHAR SUB-CHAR in readtable RT.  Whenever
   the two character sequence is read, FUNCTION is called.  DISP-CHAR is assumed to be a 
   dispatch macro character; SUB-CHAR cannot be one of the ten decimal digits.
   SET-DISPATCH-MACRO-CHARACTER returns T.")

;(setf (arglist UNUSE-PACKAGE) (PKGS &OPTIONAL PKG))
(setf (documentation 'UNUSE-PACKAGE 'function) 
  "Removes PKGS from the use list of PKG; their external symbols are no longer inherited. PKGS is a
   list of packages or package names or a single package or package name.  UNUSE-PACKAGE returns T.")

;(setf (arglist PEEK-CHAR) (&OPTIONAL PEEK-TYPE STREAM (EOF-ERRORP T) EOF-VALUE RECURSIVE-P))
(setf (documentation 'PEEK-CHAR 'function)
  "Returns the next character from STREAM without removing it from the input stream.  This same
   character will be reread on the next read operation.  The type of character read depends upon
   the value of PEEK-TYPE.  If PEEK-TYPE is NIL, read the next input character on STREAM without
   removing it from STREAM.  If PEEK-TYPE is T, discard all whitespace characters and read the first
   non-whitespace character without removing it from STREAM.  If PEEK-TYPE is a character, discard
   all characters before the first one that is equal to PEEK-TYPE and read that character without
   removing it from STREM.  End of file is an error if EOF-ERRORP is true; otherwise, PEEK-CHAR
   returns EOF-VALUE when end of file is read.  The argument RECURSIVE-P, if specified and true,
   signals that this is not a top level call to PEEK-CHAR but an imbedded call.") 

(setf (documentation 'LEAST-POSITIVE-LONG-FLOAT 'variable)
  "Constant whose value is the smallest positive magnitude representable by the long float type.") 

;(setf (arglist ALPHA-CHAR-P) (CHAR))
(setf (documentation 'ALPHA-CHAR-P 'function) 
  "Returns true if CHAR is a letter of the alphabet whose bit attribute is 0, otherwise NIL.")

;(setf (arglist 'TAN) (RADIANS))
(setf (documentation 'TAN 'function)
  "Returns the tangent of angle RADIANS.  RADIANS is measured in radians and may be complex.")

;(setf (arglist EVAL) (FORM))
(setf (documentation 'EVAL 'function) 
  "Evaluates FORM in the current dynamic environment and returns its result.")

;(setf (arglist CHAR-BIT) (CHAR BIT-NAME))
(setf (documentation 'CHAR-BIT 'function)
  "Returns true if the bit BIT-NAME is set in CHAR, otherwise NIL.  Valid values for BIT-NAME are 
   :CONTROL, :META, :SUPER, or :HYPER.")

;(setf (arglist STRING-TRIM) (CHAR-SET STRING))
(setf (documentation 'STRING-TRIM 'function)
  "Returns a substring of STRING with all characters in CHAR-SET removed from both ends of STRING.
   CHAR-SET can be a string or list of characters.  Case is presently ignored.")

;(setf (arglist ENOUGH-NAMESTRING) (OBJECT &OPTIONAL (DEFAULTS *DEFAULT-PATHNAME-DEFAULTS*)))
(setf (documentation 'ENOUGH-NAMESTRING 'function) 
  "Returns a namestring that, when considered relative to the pathname given in DEFAULTS, will 
   identify file OBJECT.  That is, this is the shortest string necessary to identify OBJECT from
   DEFAULTS.")

;(setf (arglist BIT-XOR) (BIT-ARRAY1 BIT-ARRAY2 &OPTIONAL RESULT-BIT-ARRAY))
(setf (documentation 'BIT-XOR 'function)
  "Returns the bit-array generated from a bitwise-xor of BIT-ARRAY1 with BIT-ARRAY2.  If
   RESULT-BIT-ARRAY is T, the result is also stored in BIT-ARRAY1; if NIL or omitted, the result is
   returned as a new bit-array; if a bit-array, the result is destructively placed into it.")

;(setf (arglist REMOVE-DUPLICATES) (SEQUENCE &KEY KEY (TEST (FUNCTION EQL)) TEST-NOT START END FROM-END))
(setf (documentation 'REMOVE-DUPLICATES 'function)
  "Returns a sequence where all duplicates in sequence SEQUENCE have been removed.  The first
   encountered duplicate of a duplicate pair is removed.  Although the entire length of SEQUENCE
   is the default, a subsequence starting at :START and ending at :END may be specified.  :TEST is a
   predicate which must return true in order for two elements to be regarded as duplicates;
   :TEST-NOT is a predicate which must return NIL in order for two elements to be regarded as
   duplicates.  If non-NIL, :KEY is a function applied to each element of SEQUENCE before the test
   is applied.  If :FROM-END is true, the later duplicate of a duplicate pair is removed.")

;(setf (arglist LABELS) (&QUOTE FUNCTION-LIST &REST BODY))
(setf (documentation 'LABELS 'function)
  "Special form used to define locally named functions with scope closed withing LABELS.  Syntax of 
   (LABELS ((fct1 body1) ... (fctN bodyN)) {labels-form}*) is defined as :  <fct1> thru <fctN> are
   the names of locally defined functions whose bodies <body1> thru <bodyN>, respectively, are
   similar in format to a DEFUN form.  Since the scope encompasses both the function definition and
   <labels-form>, the defined functions may call each other and recurse.  The forms in <labels-form>
   are evaluated as if in a PROGN, with function names matching those defined by the LABELS referring
   not to the global but rather the local function.")

;(setf (arglist LOGORC2) (INTEGER1 INTEGER2))
(setf (documentation 'LOGORC2 'function)
  "Returns the bitwise-or of INTEGER1 with the complement of INTEGER2.")

(setf (documentation '*TRACE-OUTPUT* 'variable)
  "Global variable whose value is the default stream for output by the function TRACE.")

;(setf (arglist PSETF) (&REST REST))
(setf (documentation 'PSETF 'function)
  "Assigns values to symbols.  REST is a list of alternating place-value pairs.  The PSETF syntax
   of (PSETF place1 exp1 place2 exp2 ... placeN expN) is defined as : evaluate all expressions
   <exp1> thru <expN> and then assign those values to the place variables <place1> thru <placeN>,
   respectively, but in an unpredictable order.  PSETF returns NIL.") 
